#!/usr/bin/perl -w

$arg = shift @ARGV;
if (!defined($arg)) {
    $file = "Makefile.template";
} 
else {
    $file = $arg;
}

if(! -r $file) {
    print $file." not found, exiting.\n";
    exit 0;
}

$_ = $file;

if(s/Makefile.template//) {

    $dir = $_;
 
    open(OUT,">".$dir."Makefile.am");
    
    print OUT "## Process this file with automake to produce Makefile.in\n\n";

    open(IN,$file);

    $need = "";
    $headers = "";
    $instHeaders = "";
    $instMan = "";
    $libs = "";

    while (<IN>) {
    	if(/^NAME: *(.*)/) {
            $name=$1;
        }
        elsif (/^SOURCES: *(.*)/) {
            $sources=$1;
        }
        elsif (/^HEADERS: *(.*)/) {
            $headers=$1;
        }
        elsif (/^PACKAGES: *(.*)/) {
            $need=$need.$1." ";
        }
        elsif (/^TYPE: *(\w)/) {
            $type=$1;
        }
        elsif (/^SYSLIBS:(.*)/) {
            $libs=$1;
        }
        elsif (/^MISC:(.*)/) {
            @myMisc=split(/ /,$1);
            foreach $x (@myMisc) {
                $_ = $x;
                s/ //g;
                if (/\.h$/) {
                    $instHeaders=$instHeaders.$_." ";
                }
                elsif (/\.[0-9nm]$/) {
                    $instMan=$instMan.$_." ";
                }
            }
        }
        elsif (/^VERBATIMSTART/) {
            while (<IN>) {
                if(/^VERBATIMEND/) {
                    last;
                }
                if(/clean::/) {
                    print OUT "clean-extra:\n";
                    next;
                }
                print OUT $_;
            }
        }
    }
    close(IN);
    
    @deps = split(/ /,$need);
    
    if ($headers ne "") {
       print OUT "noinst_HEADERS = ".$headers."\n\n";
    }

    if ($instMan ne "") {
       print OUT "man_MANS = ".$instMan."\n\n";
    }
    
    if($type eq "P") {

        print OUT "include_HEADERS = ".$name.".h $instHeaders\n\n";

        print OUT "lib_LIBRARIES = lib".$name.".a\n\n";
	
        print OUT "lib".$name."_a_SOURCES = ".$sources."\n\n";
    }
    elsif($type eq "T") {
        print OUT "bin_PROGRAMS = ".$name."\n\n";
        $_ = $name;
        s/[\-\.]/_/g;
        $name = $_;
        print OUT $name."_SOURCES = ".$sources."\n\n";
	
        print OUT $name."_LDADD = ";
        foreach $x (@deps) {
            $_ = $x;
            s/ //g;
            if($_ ne "") {
                print OUT " \${prefix}/lib/lib$_.a  "
            }
        }

        print OUT "\n\n";

        print OUT "LIBS = \@X_LIBS\@ ".$libs."\n\n"
    }
    
    print OUT "INCLUDES = -I\${prefix}/include \@X_CFLAGS\@";
    
    print OUT "\n\nMAINTAINERCLEANFILES = Makefile.in\n";
    
    print OUT "\ndepend:\n\t\@echo \"".$need."\"\n";

    $file="Makefile.template";
} elsif (s/TOOL-LIST//) {
    $dir = $_;
    if ($dir ne "") {
        chdir($dir);
    }
    $file="TOOL-LIST"; 
    $list = `./$file`;
	
    open(OUT,">Makefile.am");
	
    print OUT "## Process this file with automake to produce Makefile.in\n\n";
	
    print OUT "SUBDIRS = ";

    foreach $d (split(/[ \n\t]/,$list)) {
        print OUT "\\\n".$d;
    }
	
    print OUT "\n\nMAINTAINERCLEANFILES = Makefile.in\n";
}

print OUT "\nMakefile.am: $file \${top_srcdir}/create_makefile_am.pl\n";
print OUT "\tcd \${top_srcdir} && perl -w create_makefile_am.pl $dir/$file\n\n";
close(OUT);
