\" $Header: /home/cvs/octtools/src/BearFP-1.1/doc/bearStyle.ms,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
.tl 'Bear Coding Style'''
.ce 5
\s+1\fBCoding Style Conventions For Bear\fR\s-1
.ce 0
.ds / \h'\w'/'u'
.SH
Introduction
.PP
This document describes the general conventions we have decided
on for writing code for Bear.  
This includes conventions for writing code as well as the specification of
a directory structure that supports the development
of a large software project.
These ideas have evolved from previous projects, and we
expect that they may evolve more in the future.
Coding conventions are important in large projects because
the person doing a modification or bug fix may not be the
original author of the code.
A consistent style makes it easier for the maintainer to understand the code,
and makes it less likely that important features will be overlooked
because of some intricate coding trick.
See the files bearStyle.c and bearStyle.h for examples
of how the various conventions are used.
.SH
Directory Organization of ~bear
.PP
.IP  ~bear 20
This is Bear's home directory.
It has several sub-directories for
documentation, the source code, utilities, and testing.
.IP ~bear/src 20
This is the top level source directory.
It is further organized into subdirectories for installed code,
code under development, and testing.
.IP ~bear/src/bear 20
This has a Makefile for the whole system.
A small source file
with a version number is here also,
as well as the image of the compiled system.
.IP ~bear/src/\fImodule\fP 20
Each module is developed in its own subdirectory,
the name of the subdirectory being the same as the name of the module.
The module subdirectory has the \fB.h, .c,\fP and \fB.o\fP
file for its module.
All the \fB.o\fP files of a module are linked
together to make a \fImodule\fP.o file.
Linking this file makes the final link of the whole system go faster \-
there are less files for the linker to read.
The Makefile in each module subdirectory can generate the
\fImodule\fP.o file, the lint library for the module,
and install the source files of the module.
.IP ~bear/src/Installed/\fImodule\fP 20
There is a subdirectory under Installed for each module.
When a working version of a module is ready,
``make install'' will copy its \fB.c\fP and \fB.h\fP file there.
These files are used by the debugger as the source of installed modules.
It is necessary to use a -I \fImodule\fP argument to kdbx for each
installed module directory.
We have found it useful to have a shell script in the testing directories
that enumerate all of these directories to kdbx.
.IP ~bear/src/Include 20
This subdirectory will have the installed versions of the header files.
For each module there should be one header file for
use by the other modules, \fImoduleName\fP.h.
A module may have several header files that it uses internally.
The main internal header file is called \fImoduleName\fPInt.h.
Other internal header files are distinguished by prefixing their name with
the module prefix.
For example, the filesystem module might have these header files:
\fBfs.h\fP \- the header file used
by other modules \- and \fBfsInt.h, fsLock.h\fP and
\fBfsDisk.h\fP \- the header files used internally by the filesystem module.
.IP ~bear/src/Object 20
This subdirectory has the installed \fImodule\fP.o files.
These are used to compile the whole system.
They represent stable versions of the module:
a developer can link together his new
code with these stable object files for testing.
.IP ~bear/src/Lint 20
This has lint library files for each module.
The \fB-C\fP option of lint produces a
lint library file.
This library file summarizes each module;
it is used when doing a lint on other modules
to check inter-module references.
.IP ~bear/src/\fIfoo\fPtest 20
These are testing directories used by the developers.
There is where new versions of the system are linked together and
tested.
A module developer will use his new code with the stable
\fImodule\fP.o files to test his changes in a known
environment.
.SH
Module Organization
.PP
Development for each module is done in the module's subdirectory.
A module will usually consist of several source files and header files.
To distinguish source files from different modules,
a module has a unique prefix associated with it.
File names for the module start with this prefix,
except for three exceptions as described in the \fBDirectory Organization\fP
section: \fImoduleName\fP.h,
\fImoduleName\fPInt.h, and \fImoduleName\fP.o.
Here is an example of a directory listing for the \fBsync\fP module:
.TS
l l l l l.
Makefile	sync.o	syncLock.o	syncSleep.c	syncUser.o
RCS/	syncInt.h	syncMonitor.c	syncSleep.o
sync.h	syncLock.c	syncMonitor.o	syncUser.c
.TE
(The module prefix is also used to name global procedures
and variables of the module; see the \fBNaming Convention\fP section below.
.SH
File Layout
.PP
(Prototype files are in the directory ~bear/src/proto.)
Both \fB.c\fP and \fB.h\fP files start with a special comment block.
The comment block begins with the name of the file followed
by two dashes ``-\|-''.
Following this is the file abstract:
overall documentation that any would-be maintainer should read before
attempting to decipher the file.
It is important to give brief overall information about
how the pieces of a module fit together.
The standard UC copyright notice is placed at the end of the comment block.
For header files,
the RCS id follows the copyright notice in the comment block.
For source files,
the RCS id is a static string declaration placed just after the comment block.
.PP
After the leading comment block,
the first page of a \fB.c\fP file contains
\fB#include\fP statements and any module global
or file global declarations.
\fBExtern\fP statements should never appear here,
they should only occur in header files.
After the first page,
each successive page contains a single routine,
or a few short procedures.
(A control-L is treated as white space by the compiler
and forces a page eject when printing or viewing the file.)
Large modules should be broken up into several
files in order to keep each file small;
try to keep files in the 400-800 line range.
.PP
Include statements usually follow the leading comment block
of a header file.
We allow an include file to include other include files.
This allows a client of a module to include the main
include file for the module and not worry about any
dependencies of that module.
The body of an include file should be protected against multiple inclusions.
Surround the definitions and declarations of the header file
with an \fB#ifndef\fP like this:
.DS
#ifndef _HEADER
#define _HEADER
/*
\*/* define everything here
\*/*/
#endif _HEADER
.DE
.SH
Routine Headers
.PP
Normally, each routine begins a new page and has a collection of
comments at its head.  There are always at least three sections
to the header comments.  The first section contains a line with
nothing but the routine's name and two dashes, followed
by one or two lines giving a quick summary
of what the routine does.
This is analogous to the file abstract:
as well as describing the routine, it should indicate
how the routine fits into the module.
After this comes a section describing
the results returned by the routine.
The third section describes
any side effects the routine has.  These three sections must
always be present for every routine.  If there are no results
or side effects, indicate that fact.  Where routines have special
kludges or tricks, use additional header sections to describe them.
.PP
Routines should always be typed,
even integer-valued functions and
especially void procedures.
The type is placed on the line by itself before the routine name.
It is necessary for the routine name to begin the line
for the \fIctags\fP program to work.
The declaration of the parameters follow, one per line,
indented one shift width, and each with a side by side comment.
Note the curly brace placement.
.sp 1
.in +5
.ta \w'    Type arg1;                 'u
.nf
\fBType
ProcedureName(arg1, arg2, argN)
Type arg1;	\fI/* comment for arg1 */\fP
Type arg2;	\fI/* longer comment
	\*/* for arg2 
	\*/*/\fP
Type arg3;	\fI/* comment for arg3 */\fP
{
    \fI/*
    \*/*   routine body
    \*/*/\fP
}\fP
.in -5
.fi
.SH
Naming Conventions
.PP
This section describes the conventions for all identifiers that
occur in the code.
The point of the conventions is to organize variables and procedures
by module as well as prevent naming conflicts among modules.
The conventions are presented in three parts;
a few guidelines for chosing names,
a list of rules,
and some standard names for common things.
.PP
Names should be kept reasonably short
(names longer than 10-15 characters are usually unwieldly),
but not so short that they are cryptic.
Use the singular case,
e.g. ``cursor'', except for tables and arrays where plural
names are better, e.g. ``windows''.
Try to use the same names for the same things everywhere,
i.e. use ``filePtr'' for file descriptor pointers each time you
have a variable which is a pointer to a file descriptor.
.PP
There is a system of capitalization and underscores used
to distinguish variable names, routines, types, and defined constants.
Furthermore, routines and variables that are exported from
a module are further distinguished:
.DS
.ta 1i 3i
Types and routines each start with a capital letter.
\fBtypedef int	Boolean;\fP
.DE
.DS
.ta 1i 3i
Variables start with a small letter.
\fBint	counter;\fP
.DE
.DS
.ta 1i 3i
Variables global to a module start with a module prefix of all small
letters.
\fBFile	fsFileTable[FS_NFILES];\fP
.DE
.DS
.ta 1i 3i
Routines global to a module start with a module prefix whose first
letter is capitalized.
\fBFile	*FsFileAlloc();\fP
.DE
.DS
.ta 1i 3i
In all instances of multiword names, except names of all capitals, the
first letter of each trailing word is a capital.  This is true even if
the second word comes after an underscore.
\fBint	loopCounter;\fP
.DE
.DS
.ta 1i 3i
Global variables and routines that will be used outside the module have
an underscore added between the module prefix and the rest of the name.
\fBFile	*Fs_Open();
int	vm_FreeMem;\fP
.DE
.DS
.ta 1i 3i
Local variables don't have to start with a special prefix.
Declare as static global variables that aren't used outside
a file, this prevents name conflicts with other files.
\fBstatic	internalState;\fP
.DE
.DS
.ta 1i 3i 4i
Defined constants and macros are all capitals, except for macros used
as procedures.  In that case use the naming rules for procedures.  Long
names of all caps are hard to read so we allow underscores to break up
long defined identifiers.  A prefix (followed by an underscore) isn't
always appropriate, for instance with NULL or TRUE, but in most cases a
prefix helps group constants and prevent name conflicts.
\fB#define	NULL	0
#define	Min(a, b)	(((a) < (b)) ? (a) : (b))\fP
\fB#define	FS_FREE	0
#define	FS_LOCKED	0x1
#define	FS_WANTED	0x2\fP
.DE
.DS
.ta 1i 3i
Pointer names end with ``Ptr'', and doubly indirect pointers end with
``PtrPtr''.
.B "File	*filePtr;"
.DE
.DS
.ta 1i 3i
Pointers to the heads of lists should end with ``List'' instead of
``Ptr'', this distinguishes the list header pointer from list element
pointers.
.B "List	*fileList;"
.DE
.DS
.ta 1i 3i
Pointers to procedures should end in ``Proc''.
.B "void	(*callBackProc)();"
.DE
.IP Typedefs
Use typedefs for structures, enumerated types,
and integers used as booleans.
We don't need a great many type definitions for
variations on \fBint\fR.
(In the past typedefs were used to hide the length of
integers;
we are going to assume that integers are at least 32 bit long.)
.IP ""
For global struct definitions declare a type for the struct.
There is no prefix requirement on structure member names,
but remember the capitalization rule and the use of the ``Ptr'' suffix.
The name for the type follows the same prefix/capitalization rules
for procedures: types are capitalized,
global types start with a module prefix,
and if it is exported the prefix is separated by an underscore.
.DS
.ta 2.2i 4i
\fBtypedef struct FsInode
{
    int	count;	\fI/* comment */\fP
    DiskMap	map;	\fI/* comment */\fP
    struct FsInode	*nextInodePtr;
} FsInode;\fP
.DE
.IP ""
Occasionally unions are needed.
Follow the same general guidelines for
structs, except that the union name should end be short and
end in ``Un''.
.DS
.ta 1i 2.5i 4i
\fBtypedef struct RpcFullHdr
{
    union
	{
	Net_EtherHdr	etherHdr;
	char	genericHdr[15];
    } hdrUn;
    RpcHdr	rpcHdr;
} RpcFullHdr;
.DE
.TP
Use enums instead of integers when the values are mutually exclusive.
Declare a type for the enum,
use all capital letters for the enum values,
and distinguish the values of an enum by a common prefix
just like defined constants.
Don't use enums for boolean variables, (e.g. WAIT and DONTWAIT);
use the integer type Boolean instead.
.DS
.ta 1i 3i
\fBtypedef enum
{
    S_LOOKUP,		\fI/* look for the object */\fP
    S_INSERT,		\fI/* insert for the object */\fP
    S_DELETE		\fI/* delete the object */\fP
} SearchType;
.sp 1
Boolean	wait;	\fI/* wait for children to die */\fP
\fR
.DE
.SH
Standard Names 
.PP
There are some common types of names for which we
have assigned standard names:
.IP \fBflags\fP 6
Use this as the field identifier of a word in a structure that
contains bit flags.
.IP "\fBnextPtr, prevPtr\fP"
Use these names for the pointers in linked lists.
.IP \fBAddress\fP
This is the generic type for a pointer variable.  It is a typedef to char *.
.IP \fBClientData\fP
This is the standard type for opaque parameters.
For example, the timer queue is a queue of procedures to call at
a specified time.  Along with the procedure a private argument
is placed in the queue.  When the procedure is invoked it is passed
this private argument.  The timer routines won't know the type
of this private argument so they declare it as type ClientData.
The call back procedure should cast its argument into the appropriate type.
.SH
Comments
.PP
Normally, comments are in blocks that occupy full lines.  However,
for structure, parameter, and variable declarations, it is better
to have comments on the right sides of the declaration lines.
Avoid comments to the right of other statements.
Line up the left edges of comments,
and indent them along with the code that surrounds them.
.DS
\fI
/*
\*/*  This is a comment block, note the first and last lines
\*/*  that have only the open and close comment symbols.
\*/*  Furthermore, the comment block should be separated from the
\*/*  surrounding code by a preceding and following blank line.
\*/*/

/*
\*/*  This is a one-liner and it still is in the form of a block.
\*/*/
\fP
.DE
.SH
Tabs
.PP
Tabs are four spaces wide.
In vi, set the shiftwidth to 4, and use ^T to indent and ^D to
backspace over the indentation.
.SH
Continuation Lines
.PP
With the longer identifiers that our style encourages
a statement may often span more that one line.
In these cases indent the remaining part of the statement in a manner
that is pleasing to the eye.
Some suggestions:
.IP
For procedure calls, group adjacent parameters on the same
line if they are related (and if they fit).
Align the continuation line either one shift-width to the right
of the procedure name, or right under the first parameter.
.DS
.ta \w'Win_DefineRegion('u
Win_DefineRegion(windowPtr,
	left, top, width, height,
	regionFlags, &region);
.DE
.IP
Split long \fBfor\fP statements after the semi-colons
that divide the clauses of the for statement.
.DS
.ta \w'for ('u
for (menuPtr = List_First(menuList);
	! List_IsAtEnd(menuList, menuPtr);
	menuPtr = List_Next(menuPtr))
{
}
.DE
.IP
Split long boolean expressions after major clauses of the
boolean expression.
Try to end each line (except the last) with the and, ``&&'', or or, ``||''
operators.
.DS
.ta \w'if ('u
if (((regionPtr->flags & RGN_SCROLL) == 0) ||
	( ! WinScrollBarInput(regionPtr, &wmInputEvent)))
{
}
.DE
.IP
Assignment statements can occasionally be divided after the equals sign
to keep the right hand side of the equation on one line.
Indent the right hand side one shift-width beyond the first line
of the statement.
.DS
mouseEventPtr->buttonState =
    regionPtr->lastButtonState;
.DE
.IP
Long expressions can also be divided in a number of ways.
Be consistent in any one instance of a divided expression.
For example, always end a line with an operator
and line up the left edge of leading terms.
.DS
.ta \w'totalHeight = 'u
totalHeight = windowPtr->bannerHeight + windowPtr->borderWidth +
	textRegionPtr->rect.height + optionRegionPtr->rect.height +
	cmdRegionPtr->rect.height +
	2 * windowPtr->regionSpacing;
.DE
.SH
Curly Brace Convention
.PP
It is required that you always use curly braces
with any compound statement,
even if there is only one statement in the block.
.br
.in +5
.nf
\fBif (boolean expression)
{
    \fI/*
    \*/* body of if clause
    \*/*/\fP
}
else
{
    \fI/*
    \*/* body of else clause
    \*/*/\fP
}

for (boolean expression)
{
    \fI/*
    \*/* body of for clause
    \*/*/\fP
}

switch (value)
{
case VALUE:
	\fI/*
	\*/* body of case block
	\*/*/\fP
	break;
case VALUE2:
	{
		int foo;	\fI/* a local variable */
		/*
		\*/* body of second case block
		\*/*/\fP
   	}
	break;
default:
	break;
}\fP
.in -5
.fi
.SH
Macros
.PP
Macros should generally be avoided if there is another way to
achieve the same end.  This is because macro information isn't
available at run-time, thereby making it difficult to debug.
Long macros are generally
better off as routines, unless they need to be in-line to
obtain greater performance.  Constant definitions are often
better as variables that are initialized and then never changed
(this has the advantage of making it possible to go in with the
debugger and change ``constants'' while debugging).
.PP
If you use a macro like a procedure, for 1 or 2 line operations,
use the naming conventions for procedures.
If you use a macro to produce other code constructs,
like special for loops or structure field access,
make the macro all capitals.
In the definition of a macro remember to enclose arguments
in parentheses because they may be expressions.
Similarly,
when a constant is defined as an arithmetic expression of other
constants the expression should be enclosed in parenthesis so
the constant can be used freely in other expressions.
.SH
Comments Within Routines
.PP
For short routines, there may not be any need for comments
within the routine itself.  For longer routines, there should be
occasional blocks of comments indicating the various
pieces of the routine and general flow.  For routines with
many local variables, either comment the variable
declarations, or place the declarations in the inner blocks
where the variables are used.
.SH
Automatic Variable Initialization
.PP
The value of an automatic variable is undefined upon entry into
a procedure.
C allows initialization of an automatic variable on the line
that it is declared.
This is convenient but should only be used in simple cases.
For flags or integers it is reasonable to set their value to
FALSE or to 0 in the declaration.
It is not good practice to call functions or do things with side effects
when initializing automatic variables.
.SH
Inner Blocks
.PP
It is possible to declare variables inside inner blocks in C.
There is a trade-off between scattering variable declarations
among the various blocks of a long procedure and
declaring a lot of locals at the beginning of a procedure.
Block local declarations have the advantage of placing
the declaration close to the place where the variables are used.
They may also be more efficient because these ``block local'' variables
can be declared as register variables.
A procedure may have a few major branches in which
different things are done.
In this case it is good practice to localize the variable
declarations.
Finally, use lint to catch the errors of unused outer block
variables or parameters that are hidden by inner block declarations.
.SH
Nested Header Files and the Makefile
.PP
Header files can be nested,
that is they can have \fB#include\fP statements in them.
There is an option to \fIcpp\fP, the c preprocessor,
that generates makefile dependency entries.
This option is smart about nested header files,
it generates output in the following form.
.DS
\fIfile.\fPo : \fIfile.\fPc
\fIfile.\fPo : \fIheader.\fPh
.DE
There is a ``make depend'' entry in the makefiles
that uses this option to update the makefile.
By nesting \fB#include\fP statements,
a module can insure that its users import the
other modules that the module depends on.
