
/* bear.h --
 *
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright(C) 1991 Regents of the University of California
 * All rights reserved.
 *
 * needs: /usr/include/stdio.h, magic.h, geometry.h
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Include/bear.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _BEAR
#define _BEAR /* once-only latch */

#ifndef _MAGIC
int err1 = Need_to_include_magic_header;
#endif _MAGIC

#ifndef _GEOMETRY
int err2 = Need_to_include_geometry_header;
#endif _GEOMETRY


#include "ansi.h"
#include "list.h"
#include "bearStatus.h"
#include "bearReadLine.h"


#define BEAR_PROG_NAME "bearFP"

#define BEAR_MARKED             1
#define BEAR_UNMARKED           0

#define NAME_STRING_SIZE       40

#define BEAR_MSG_SIZE        1024

#define MAGIC_STOP_SEARCH       1
#define MAGIC_CONTINUE_SEARCH   0

typedef int MagicStatus;


typedef struct _BearDummy BearDummy;
struct _BearDummy {
    int dummy;                      /* not used                             */
};

typedef char         *BearString;

typedef BearDummy    *BearGeneric;  /* Generic pointer.                     */

typedef unsigned int BearId;        /* Bear identification numbers 
                                     * are positive integers 
                                     * bigger then zero.
                                     */

typedef enum _SideDirection SideDirection;
enum _SideDirection {
    LEFT_SIDE,                      /* the left side of a tile              */
    RIGHT_SIDE,                     /* the right side of a tile             */
    TOP_SIDE,                       /* the top side of a tile               */
    BOTTOM_SIDE,                    /* the bottom side of a tile            */
    BOTTOM_LEFT_SIDE,               /* the bottom or left side of a tile    */
    TOP_LEFT_SIDE,                  /* the top or left side of a tile       */
    TOP_RIGHT_SIDE                  /* the top or right side of a tile      */
};


extern FILE *bearout;
extern FILE *bearerr;
extern char bear_ErrorMsg[];
extern char bear_Msg[];
extern char *bear_tmpDir;
extern bool bear_BlackBoxFlag;


EXTERN BearStatus Bear_Begin
    NULLARGS;

EXTERN void Bear_Print
    ARGS((BearString msg));

EXTERN void Bear_PrintWarn
    ARGS((BearString warnMsg));

EXTERN void Bear_PrintErr 
    ARGS((BearString errMsg));

EXTERN void Bear_PrintFatalErr
    ARGS((BearString fatalErrMsg));

EXTERN char *Bear_GetBooleanString
    ARGS((bool value));

EXTERN void Bear_Pause
    NULLARGS;

EXTERN void Bear_RemoveTmpFiles
    NULLARGS;


/* ---------- Macros ---------- */

#ifndef NIL
#define NIL(type)       ((type *) NULL)
#endif /* NIL */

#ifndef MAX
#define MAX(a,b)        ((a) > (b) ? (a) : (b))
#endif /* MAX */
#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif /* MIN */
#ifndef ABS
#define ABS(a)          ((a) < 0 ? -(a) : (a))
#endif /* ABS */

#define BEAR_EVEN(A)    (!BEAR_ODD(A))
#define BEAR_ODD(A)     (A & 01)

#define Bear_SetOption(optionMask, option)      ((optionMask) |= (option))
#define Bear_UnsetOption(optionMask, option)    ((optionMask) &= ~(option))
#define Bear_ToggleOption(optionMask, option)   ((optionMask) ^= (option))

#define Bear_BackeachItem(list, gen, data)             \
    for (gen = lsEnd(list);                            \
    (lsPrev(gen, (lsGeneric *) &data, LS_NH) == LS_OK) \
        || ((void) lsFinish(gen), 0); )

#endif _BEAR

