
/* macCell.h --
 *
 *     Wei-ming Dai
 *     June Wang
 *     (Professor E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Include/macCell.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _MAC_CELL
#define _MAC_CELL /* once-only latch */

#define MAC_MAX_PLANES  5               /* Maximum number of planes per cell*/
#define MAC_CELL_PLANE  0               /* The plane with cells             */

/* 
 * Cell master status.
 */
#define MAC_CLUSTERING_DONE      01  /* clustering has been done            */
#define MAC_PLACEMENT_DONE       02  /* placement has been done             */
#define MAC_GLOBAL_ROUTING_DONE  04  /* global routing has been done        */
#define MAC_CORE_ROUTING_DONE   010  /* core detailed routing has been done */
#define MAC_RING_ROUTING_DONE   020  /* ring routing has been done          */

typedef struct _Contents Contents;
struct _Contents {
    BearString          cmFileName;     /* Name of the file that contains
                                         * the description of the cell
                                         * master.
                                         */
    BearString          ctreeFileName;  /* Name of the file that contains
                                         * the description of a cluster
                                         * tree used with the ncl command.
                                         */
    BearString          flexFileName;   /* Name of the file that contains
                                         * the flex information used with the
                                         * so command.
                                         */
    BearString          pgFileName;     /* Name of the file that contains
                                         * the power ground routing
                                         * information.
                                         */
    BearString          techFileName;   /* Name of the file that contains
                                         * the technology information.
                                         */
    BearString          tmFileName;     /* Name of the file that contains
                                         * the timing information used with 
                                         * the ncl and npl commands.
                                         */
    unsigned int        status;
    struct _Pgr_WireTech *wireTechPtr;
    struct _CellInst    *cellInstList;  /* Head of the child cell 
                                         * master list   
                                         */
    struct _Net         *netList;       /* Head of the child net list       */
    struct _RouteCell   *routeCellList; /* Head of the routing cell list    */
    struct _Ring        *ringPtr;       /* The routing ring                 */
    PlaneData           *planeData[MAC_MAX_PLANES];
    bool                placedFlag;     /* TRUE if all the cells are currently
                                         * placed so that none of them overlap.
                                         */
};

typedef enum _ShapeFlag {
/* 
 * Each cell master has a shape flag and it can either be 
 * undefined, fixed, flexible, or contrained.
 */

    MAC_UNDEF_SHAPE,
    MAC_FIXED_SHAPE,
    MAC_FLEX_SHAPE,
    MAC_CONSTR_SHAPE
} ShapeFlag;

typedef struct _Shape {
    int     horzDim;                /* horizontal dimension of the shape    */
    int     vertDim;                /* vertical dimension of the shape      */
} Shape;

typedef struct _Scale {
    int numerator;                  /* numerator of a scale number          */
    int denominator;                /* denominator of a scale number        */
} Scale;

typedef struct _RectConstr {
    Scale minAspectRatio;           /* Minimum aspect ratio of the rectangle
                                     * aspect radio : x-extent/y-extent
                                     */
    Scale tarAspectRatio;           /* Target aspect ratio of rectangle     */
    Scale maxAspectRatio;           /* Maximum aspect ratio of rectangle    */
} RectConstr;

typedef struct _AreaConstr {
    int minArea;                    /* Minimum area                         */
    int maxArea;                    /* Maximum area                         */
} AreaConstr;

typedef struct _ShapeConstr {
    int     minArea;                /* Minimum area                          */
    int     tarArea;                /* Target area                           */
    int     maxArea;                /* Maximum area                          */
    Scale   minAspectRatio;         /* Minimum aspect ratio of a cell shape  */
    Scale   tarAspectRatio;         /* Target aspect ratio of a cell shape   */
    Scale   maxAspectRatio;         /* Maximum aspect ratio of a cell shape  */
} ShapeConstr;

typedef unsigned char OrientVector; /* The permission of the orientation
                                     * of a cell instance is specified 
                                     * by an 8-bit vector. If bit i is 1.
                                     * the corresponding i-th orientation
                                     * is forbidden for the instance. A
                                     * bit vector with all  0's means all
                                     * orientation are permitted. A bit
                                     * vector with all 1's is illegal.
                                     */

typedef struct _CellMaster CellMaster;
typedef struct _CellInst CellInst;

struct _CellMaster {
    BearString          name;            /* Cell master name                */
    BearId              id;              /* Cell master identification
                                          * number.
                                          */
    ShapeFlag           shapeFlag;       /* Cell master shape flag          */
    lsList              shapeList;

    ShapeConstr         *shapeConstr;    /* Shape constraints for a cell    */

    Rect                bbox;            /* The cell instances bounding box */
    LinkedRect          *horzRectList;   /* List of horizontally sliced
                                          * rectangles.
                                          */
    LinkedRect          *vertRectList;   /* List of vertically sliced
                                          * rectangles.
                                          */
    st_table            *pinInstTablePtr;
    CellInst            *instList;       /* The list of cell instances      */
    struct _PortMaster  *portMasterList; /* The list of port masters of 
                                          * the cell.
                                          */
    Contents            contents;        /* Contents of the cell            */
    ClientData          clientData;      /* This space for rent.            */
};

struct _CellInst {
    BearId             id;              /* Cell instance id                 */
    BearString         name;            /* Cell instance name.              */
    int                mark;            /* The default value for this field
                                         * is BEAR_UNMARKED.
                                         */
    Point              origin;          /* The boundary of the cell instance
                                         * and all of its pins are expressed
                                         * relative to this point.
                                         */
    Transform          transform;       /* Transformation suitable for
                                         * placing the cell instance on the 
                                         * horzizontal tile plane.
                                         */
    OrientVector       orientConstr;    /* Orientation constraints          */ 
    LinkedPolygon      *translConstr;   /* Translation constraints          */
    LinkedLineSeg      *boundary;       /* List of line segments in clock-
                                         * wise order, that define the cell.
                                         */
    struct _LinkedLinkedPort *feedThrList;  /* The list of lists of feed 
                                             * through ports.
                                             */
    struct _LinkedLinkedPort *eleEqvList;   /* The list of lists of electrical
                                             * equivalent ports.
                                             */
    CellMaster         *masterPtr;      /* The master of this cell instance. */
    struct _PortInst   *portInstList;   /* The list of port instances 
                                         * of the cell.
                                         */
    struct _CellGroup  *cellGroupPtr;   /* Pointer to cell group.            */
    CellInst           *nextPtr;        /* The next cell instance.           */
    CellMaster         *parentPtr;      /* Pointer to the parent cell.       */
    CellInst           *siblingPtr;     /* Pointer to the sibling cell.      */
    Fpg_Room           *roomPtr;
    BearGeneric        nclModulePtr;    /* Reserved for the new cluster
                                         * module.
                                         */
    ClientData         clientData;      /* This space for rent.              */
};


extern BearStatus 
Mac_CreateCell(/* parentCellPtr, masterName, instName, cellInstPtrPtr */);
/*
 * CellMaster  *parentCellPtr;
 * BearString  masterName;
 * BearString  instName;
 * CellInst    **cellInstPtrPtr;
 */

extern BearStatus Mac_DuplicateCell(/* parentCellPtr, oldCellInstPtr, 
                       newCellInstName, newCellInstPtrPtr */);
/*
 * CellMaster  *parentCellPtr;
 * CellInst    *oldCellInstPtr;
 * BearString  newCellInstName;
 * CellInst    **newCellInstPtrPtr;
 */

extern BearStatus Mac_DeleteCellMaster(/* cellMasterPtr */);
/*
 * CellMaster *cellMasterPtr;
 */

extern BearStatus Mac_DeleteCell(/* cellInstPtr */);
/*
 * CellInst   *cellInstPtr;
 */

extern BearStatus Mac_MoveCell(/* deltaX, deltaY, cellInstPtr */);
/*
 * int      deltaX;
 * int      deltaY;
 * CellInst *cellInstPtr;
 */

extern void Mac_FindBBoxForAllCells(/* cellMasterPtr, bboxPtr */);
/*
 * CellMaster   *cellMasterPtr;
 * Rect         *bboxPtr;
 */

extern BearStatus Mac_AssignShapeFlag(/* cellInstPtr */);
/*
 * CellInst *cellInstPtr;
 */

extern void     Mac_TidyUpShapeList();
extern lsList   Mac_CreateCellInstList();
extern void     Mac_SortCellInstListByName();


/* ---------- cell master macros for global status ---------- */

#define Mac_SetClusteringDone(cellMasterPtr) \
    (Bear_SetOption(cellMasterPtr->contents.status, MAC_CLUSTERING_DONE))
#define Mac_UnsetClusteringDone(cellMasterPtr) \
    (Bear_UnsetOption(cellMasterPtr->contents.status, MAC_CLUSTERING_DONE))
#define Mac_IsClusteringDone(cellMasterPtr) \
    (((cellMasterPtr)->contents.status & MAC_CLUSTERING_DONE) ? \
    TRUE : FALSE)

#define Mac_SetPlacementDone(cellMasterPtr) \
    (Bear_SetOption(cellMasterPtr->contents.status, MAC_PLACEMENT_DONE))
#define Mac_UnsetPlacementDone(cellMasterPtr) \
    (Bear_UnsetOption(cellMasterPtr->contents.status, MAC_PLACEMENT_DONE))
#define Mac_IsPlacementDone(cellMasterPtr) \
    (((cellMasterPtr)->contents.status & MAC_PLACEMENT_DONE) ? \
    TRUE : FALSE)

#define Mac_SetGlobalRoutingDone(cellMasterPtr) \
    (Bear_SetOption(cellMasterPtr->contents.status, MAC_GLOBAL_ROUTING_DONE))
#define Mac_UnsetGlobalRoutingDone(cellMasterPtr) \
    (Bear_UnsetOption(cellMasterPtr->contents.status, MAC_GLOBAL_ROUTING_DONE))
#define Mac_IsGlobalRoutingDone(cellMasterPtr) \
    (((cellMasterPtr)->contents.status & MAC_GLOBAL_ROUTING_DONE) ? \
    TRUE : FALSE)

#define Mac_SetCoreRoutingDone(cellMasterPtr) \
    (Bear_SetOption(cellMasterPtr->contents.status, MAC_CORE_ROUTING_DONE))
#define Mac_UnsetCoreRoutingDone(cellMasterPtr) \
    (Bear_UnsetOption(cellMasterPtr->contents.status, MAC_CORE_ROUTING_DONE))
#define Mac_IsCoreRoutingDone(cellMasterPtr) \
    (((cellMasterPtr)->contents.status & MAC_CORE_ROUTING_DONE) ? \
    TRUE : FALSE)

#define Mac_SetRingRoutingDone(cellMasterPtr) \
    (Bear_SetOption(cellMasterPtr->contents.status, MAC_RING_ROUTING_DONE))
#define Mac_UnsetRingRoutingDone(cellMasterPtr) \
    (Bear_UnsetOption(cellMasterPtr->contents.status, MAC_RING_ROUTING_DONE))
#define Mac_IsRingRoutingDone(cellMasterPtr) \
    (((cellMasterPtr)->contents.status & MAC_RING_ROUTING_DONE) ? \
    TRUE : FALSE)

#endif _MAC_CELL

