
/*
 * macRPin.h --
 *
 *     Wei-ming Dai
 *     June Wang
 *     (Professor E. S. Kuh)
 *
 * It declares the data structures related to routing pins and the
 * operations to create, insert, and delete.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Include/macRPin.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _MAC_RPIN
#define _MAC_RPIN

typedef struct _RoutePin
{
    int               mark;
    BearId            id;            /* ID of the routing pin            */
    BearId            intNetId;      /* Internal net id.                 */
    BearId            extNetId;      /* External net id.                 */
    BearId            horzExtNetId;  /* External net id after horzontal
                                      * unification.
                                      */
    BearId            vertExtNetId;  /* External net id after vertical
                                      * unification.
                                      */
    Flexibility       positionFlag;  /* Position flexibility flag        */
    Point             position;      /* Position relative to routing
                                      * cell origin
                                      */
    PinDirection      direction;     /* Direction of the routing pin     */
    int               order;         /* Relative ordering of the pin     */
    float             pinWidth;      /* Width of the routing pin         */
    float             segmWidth;     /* Width of the routing segment     */

    float             widthVect[8];  /* Width vector of the route pin
                                      *    0: poly-silicon; 1-7: Al
                                      */
    LayerVector       layer;         /* Layer of the routing pin         */
    RouteCell         *cellPtr;      /* Pointer to the route cell to
                                      * which the routing pin belongs 
                                      */
    Net               *netPtr;       /* Pointer to the net to which 
                                      * the routing pin belongs 
                                      */
    ClientData        pinPtr;        /* Pointer to generic Pin or pseudo pin */
    struct _RdbPinInfo        *BBLTermPtr;
    struct _RoutePin  *nextPtr;      /* Pointer to the next routing pin 
                                      * of the routing cell 
                                      */
    ClientData        clientData;    /* This space for rent              */
} RoutePin;

typedef struct _LinkedRPin
{
    RoutePin           *rPinPtr;
    struct _LinkedRPin *nextPtr;
} LinkedRPin;


extern BearStatus Mac_CreateRPin (/* rCellPtr, side, rPinPtrPtr */);
/*
 * RouteCell        *rCellPtr;
 * SideDirection    side;
 * RoutePin         **rPinPtrPtr;
 */

#endif _MAC_RPIN

