
/* tileMapping.h --
 *
 *     Tim Collins
 *     (Professor E. S. Kuh)
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 * needs: magic.h bear.h ...
 * 
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Include/tileMapping.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _TILEMAPPING
#define _TILEMAPPING

#define TM_MOD_NAME "tileMapping"

typedef struct _PinsAroundJunction {
    LinkedPinInst    *pinInstList;
    LinkedRPin       *rPinList;
    LinkedPPin       *horzPPinList;
    LinkedPPin       *vertPPinList;
} PinsAroundJunction;

typedef enum _ViewType {
    NONE_DONE,
    H_DONE,
    V_DONE,
    BOTH_DONE
} ViewType;

extern ViewType viewTypeFlag;


/* ---------- External procedures for the tileMapping module ---------- */

extern void Tm_HUnifyNoFree(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_HUnify(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_VUnifyNoFree(/* parentCellPtr */);
/* 
 * CellMaster *parentCellPtr;
 */

extern void Tm_VUnify(/* parentCellPtr */);
/* 
 * CellMaster *parentCellPtr;
 */

extern void Tm_HSplit(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_VSplit(/* parentCellPtr */);
/* 
 * CellMaster *parentCellPtr;
 */

extern void Tm_HCSplit(/* parentCellPtr, planeDataPtr, domRect */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Rect       domRect;
 */

extern void Tm_VCSplit(/* parentCellPtr, planeDataPtr, domRect */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Rect       domRect;
 */

extern void Tm_HFSplit(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_VFSplit(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_PinsAroundJunction();
/*
 * CellMaster          *parentCellPtr;
 * PlaneData           *planeDataPtr;
 * WallJunction        *junctPtr;
 * PinsAroundJunction  *pinListsPtr;
 */

extern void Tm_PinsAroundCircle();
/*
 * CellMaster          *parentCellPtr;
 * PlaneData           *planeDataPtr;
 * WallJunction        *junctPtr;
 * PinsAroundJunction  *pinListsPtr;
 */

extern void Tm_HRectUnify(/* parentCellPtr, planeDataPtr, dmntRect */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Rect       dmntRect;
 */

extern void Tm_VRectUnify(/* parentCellPtr, planeDataPtr, dmntRect */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Rect       dmntRect;
 */

extern void Tm_HVFUnify(/* parentCellPtr, planeData */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 */

extern void Tm_HFUnify(/* copyFlag */);
/*
 * int copyFlag;
 */

extern void Tm_VFUnify(/* copyFlag */);
/*
 * int copyFlag;
 */

extern void Tm_HFOnlyUnify(/* parentCellPtr, planeData */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 */

extern void Tm_VFOnlyUnify(/* parentCellPtr, planeData */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 */

extern void Tm_HBnUnify(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_VBnUnify(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void
Tm_RingUnify(/* parentCellPtr, hRect1, vRect1, hRect2, vRect2, junctPtr */);
/* CellMaster *parentCellPtr;
 * Rect       hRect1;
 * Rect       vRect1;
 * Rect       hRect2;
 * Rect       vRect2;
 * Junction   *junctPtr;
 */

extern void Tm_InsDCellOnHTP(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern void Tm_InsDCellOnVTP(/* parentCellPtr */);
/*
 * CellMaster *parentCellPtr;
 */

extern PinsAroundJunction *
Tm_PinsAroundHFace(/* parentCellPtr, planeData, rect, includeInsidePPins */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *tmGlobalPlaneDataPtr;
 * Rect       rect;
 * bool       includeInsidePPins;
 */

extern PinsAroundJunction *
Tm_PinsAroundVFace(/* parentCellPtr, planeData, rect, includeInsidePPins */);
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *tmGlobalPlaneDataPtr;
 * Rect       rect;
 * bool       includeInsidePPins;
 */

extern void Tm_GetPinsInVBottleNeckRegion();
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Tile       *bnTilePtr;
 * PinsAroundJunction *pinListsPtr;
 */

extern void Tm_GetPinsInHBottleNeckRegion();
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Tile       *bnTilePtr;
 * PinsAroundJunction *pinListsPtr;
 */

extern void Tm_VCircleUnify();
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Rect       dmntRect; (vertical Rect)
 */

extern void Tm_HCircleUnify();
/*
 * CellMaster *parentCellPtr;
 * PlaneData  *planeDataPtr;
 * Rect       dmntRect; (horz Rect)
 */

extern PinsAroundJunction *Tm_MakePinsAroundJunction();

extern void Tm_FreePinsAroundJunction(/* pinListsPtr */);
/*
 * PinsAroundJunction *pinListsPtr;
 */

/* ---------- External global variables for the tileMapping module ---------- */
extern bool tm_DupFlag;

#endif _TILEMAPPING

