
/* tpTile.h -
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Include/tpTile.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _TP_TILE
#define _TP_TILE

typedef struct _LinkedTile {
    struct _LinkedTile  *prevPtr;
    struct _LinkedTile  *nextPtr;
    Tile                *tilePtr;
} LinkedTile;

/*
 * Direction of a fragment
 */
#define     TP_BOTTOM   0           /* the bottom fragment of a side        */
#define     TP_TOP      1           /* the top fragment of a side           */

/* tile plane status */
#define TP_SPACE_TILES_READY    01  /* space tiles have been identified     */

typedef enum _BNLineType
{
    TP_VERT_R_UP,                   /* BN line of vertical tiles t1 and t2,
                                     * where t2 is to the right and above t1.
                                     */
    TP_VERT_R_DOWN,                 /* BN line of vertical tiles t1 and t2,
                                     * where t2 is to the right and below t1.
                                     */
    TP_HORZ_B_RIGHT,                /* BN line of horizontal tiles t1 and t2,
                                     * where t2 is below and to the right of
                                     * t1.
                                     */
    TP_HORZ_B_LEFT,                 /* BN line of horizontal tiles t1 and t2,
                                     * where t2 is below and to the left of t1.
                                     */
    TP_NULL_BNL
} BNLineType;                       /* bottle neck line type                */

typedef struct _BNLine
{
    BearId          id;             /* identification number                */
    BNLineType      bnLineType;     /* type                                */
    Line            line;           /* geometry                             */
    int             capacity;       /* routing capacity                     */
    int             density;        /* routing desity                       */
    Tile            *tlTilePtr;     /* top or left tile                     */
    Tile            *brTilePtr;     /* bottom or right tile                 */
    struct _WallSegment *segPtr;    /* corresponding wall segment in the 
                     * floor plan graph.
                                     */
    struct _BNLine  *nextPtr;       /* next bottleneck line on the plane    */
} BNLine;

#define TP_BN_LINE  01              /* bottleneck line                      */
#define TP_BN_TILE  02              /* bottleneck tile                      */

typedef struct _BottleNeck {
    int         unType;
    union {
        BNLine  *bnLinePtr;         /* unType = TP_BN_LINE                  */
        Tile    *bnTilePtr;         /* unType = TP_BN_TILE                  */
    }           unValue;
} BottleNeck;

typedef int PlaneType;
#define NULL_TILE_PLANE 01
#define HORZ_TILE_PLANE 02
#define VERT_TILE_PLANE 04

typedef struct _PlaneData
{
    unsigned int    pd_status;
    Rect    pd_horzPlaneBnd;        /* World coordinates of horizontal
                                     * plane boundary.
                                     */
    Rect    pd_vertPlaneBnd;        /* World coordinates of vertical
                                     * plane boundary.
                                     */
    Plane   *pd_horzPlanePtr;       /* pointer to the horziontal tile plane */
    Plane   *pd_vertPlanePtr;       /* pointer to the vertical tile plane   */
    BNLine  *pd_horzBnLineList;     /* List of all the bottleneck lines on
                                     * the horizontal tile planes.
                                     */
    BNLine  *pd_vertBnLineList;     /* List of all the bottleneck lines on
                                     * the vertical tile planes.
                                     */
    struct _DummyCell *pd_dcellList;/* pointer to dummy cell list          */
    FloorPlanGraph    *pd_fpgPtr;   /* pointer to the floor plan graph     */
} PlaneData;

/*
 * Associated to each space tile, there are properties:
 * capacity and density. Also, each space tile has two 
 * sides which adjacent to other space tiles. Attached to the 
 * bottom or left side and top or right side of a tile, there 
 * are two psuedo pin lists.
 * (currently these properties are only attached to bottleneck 
 * tiles).
 */

/*
 * Note that TT_SPACE is defined to be 0 in database.h
 */
#define TT_SOLID 1

#define TP_CELL_BASE            100

/*
 * There are currently six types of space tiles.
 */
#define TP_HORZ_BN_TILE          01     /* horizontal bottleneck tile       */
#define TP_VERT_BN_TILE          02     /* vertical bottleneck tile         */
#define TP_HORZ_DOM_TILE         04     /* horizontal dominant tile         */
#define TP_VERT_DOM_TILE        010     /* vertical dominant tile           */
#define TP_HORZ_DOM_SBN_TILE    020     /* horizontal dominant semi
                                         * bottleneck tile        
                                         */
#define TP_VERT_DOM_SBN_TILE    040     /* vertical dominant semi
                                         * bottleneck tile      
                                         */
#define TP_HORZ_GEN_TILE       0100     /* horizontal generic tile          */
#define TP_VERT_GEN_TILE       0200     /* vertical generic tile            */
#define TP_HORZ_GEN_SBN_TILE   0400     /* horizontal generic semi
                                         * bottleneck tile  
                                         */
#define TP_VERT_GEN_SBN_TILE  01000     /* vertical generic semi
                                         * bottleneck tile 
                                         */
#define TP_VALID_SPACE_TILE   01777

typedef struct _SpaceProp 
{
    int             type;           /* Type of space tile.                  */
    BearId          id;             /* The tiles identification number.     */
    int             junctTableIndex;/* If this is a dominant tile then
                                     * this is either the horizontal or
                                     * the vertical index into the external
                                     * variable 'fpg_JunctTable', otherwise
                                     * it's -1.
                                     */
    int             capacity;       /* The capacity of the tile.            */
    int             density;        /* The density of the tile.             */
    int             ridgeWidth;
    Tile            *prevTilePtr;
    struct _WallSegment *segPtr;    /* Bottleneck tiles each point to a wall
                                     * segment in the floor plan graph.
                                     */
    PseudoPin       **blPPinList[2]; /* The psuedo pin list attached to 
                                     * the bottom or left side of the tile.
                                     */
    PseudoPin       **trPPinList[2]; /* The psuedo pin list attached to 
                                      * the top or right side of the tile.
                                      */
    int             mark;           /* The default value for this field
                                     * is BEAR_UNMARKED.
                                     */
    ClientData      clientData;     /* This space for rent.                 */
} SpaceProp;

/*
 * The properties associated with a tile.
 */

typedef enum _TileKind
{
    TP_CELL_INST, TP_DUMMY_CELL, TP_ROUTE_CELL, TP_RING, TP_SPACE_PROP
} TileKind;

typedef struct _TileProp 
{
    TileKind             unType;
    union
    {
        struct _CellInst  *cellInstPtr;  /* unType = TP_CELL_INST           */
        struct _DummyCell *dummyCellPtr; /* unType = TP_DUMMY_CELL          */
        struct _RouteCell *routeCellPtr; /* unType = TP_ROUTE_CELL          */
        struct _Ring      *ringPtr;      /* unType = TP_RING                */
        SpaceProp         *spacePropPtr; /* unType = TP_SPACE_PROP          */
    }                     unValue;
    int                   mark;          /* The default value for this field
                                          * is BEAR_UNMARKED.
                                          */
    ClientData            clientData;    /* This space for rent.            */
} TileProp;

extern PlaneData *Tp_MakePlaneData();
extern SpaceProp *Tp_MakeSpaceProp();
extern TileProp  *Tp_MakeTileProp();

extern void Tp_CreateSpaceTileProp(/* tilePtr */);
/*
 * Tile *tilePtr;
 */

extern void Tp_CreateSolidTileProp(/* tilePtr, gcellPtr */);
/* 
 * Tile                *tilePtr;
 * struct _GenericCell *gcellPtr;
 */

extern void Tp_CreateSolidTilePropForRing(/* tilePtr, ringPtr */);
/* 
 * Tile    *tilePtr;
 * Ring    *ringPtr;
 */

extern BearStatus Tp_CreateBnLine(/* tlTile, brTile, bnLineType,
    bnLineList, bnLine */);
/*
 * Tile         *tlTile, *brTile;
 * BNLineType   bnLineType;
 * BNLine       **bnLineList, **bnLine;
 */

extern BearStatus Tp_FindBnLine(/* tlTile, brTile, line */);
/*
 * Tile *tlTile, *brTile;
 * Line *line;
 */

extern void Tp_FreePlaneData(/* planeDataPtr */);
/*
 * PlaneData *planeDataPtr;
 */

extern void Tp_FreeTileProp(/* tilePtr */);
/*
 * Tile *tilePtr;
 */

extern void Tp_FreeAllPPinsOnHorzPlane(/* planeDataPtr */);
/*
 * PlaneData *planeDataPtr;
 */

extern void Tp_FreeAllPPinsOnVertPlane(/* planeDataPtr */);
/*
 * PlaneData *planeDataPtr;
 */

extern void Tp_RemoveTileProp(/* tilePtr, tilePropPtrPtr */);
/*
 * Tile        *tilePtr;
 * TileProp    **tilePropPtrPtr;
 */

extern void Tp_AddTileProp(/* tilePropPtr, tilePtr */);
/*
 * TileProp    *tilePropPtr;
 * Tile        *tilePtr;
 */

extern void Tp_HorzTileToRect(/* planeData, horzTile, rect */);
/*
 * PlaneData    *planeData;
 * Tile         *horzTile;
 * Rect         *rect;
 */

extern void Tp_VertTileToRect(/* planeData, vertTile, rect */);
/*
 * PlaneData    *planeData;
 * Tile         *vertTile;
 * Rect         *rect;
 */

extern BearStatus Tp_DeleteCell(/* cellInst */);
/*
 * struct _CellInst *cellInst;
 */

extern BearStatus Tp_ResizeCellInst(/* deltaX, deltaY, cellInstPtr */);
/*
 * int         deltaX;
 * int         deltaY;
 * CellInst    *cellInstPtr;
 */


extern BearStatus Tp_ResizeParentCell(/* deltaX, deltaY, coreInstPtr */);
/*
 * int         deltaX;
 * int         deltaY;
 * CellInst    *coreInstPtr;
 */

extern BearStatus Tp_InsertCellMaster(/* cellMasterPtr, cellName,
                                       * origin, pointList
                                       */);
/*
 * PlaneData    *planeData;
 * BearString   cellName;
 * Point        *origin;
 * LinkedPoint  *pointList;
 */

extern BearStatus Tp_MoveCell(/* gcellPtr, deltaPtr */);
/*
 * GenericCell *gcellPtr;
 * Point       *deltaPtr;
 */

extern BearStatus Tp_MoveCellGroup(/* cellGroupPtr, deltaPtr */);
/*
 * CellGroup   *cellGroupPtr;
 * Point       *deltaPtr;
 */

extern BearStatus Tp_MoveCellInst(/* cellInstPtr, deltaPtr */);
/*
 * CellInst    *cellInstPtr;
 * Point       *deltaPtr;
 */

extern bool Tp_PushHorz(/* planeData, horzSpaceTileList, distance */);
/*
 * PlaneData    *planeData;
 * LinkedTile   *horzSpaceTileList;
 * int          distance;
 */

extern bool Tp_PushVert(/* planeData, vertSpaceTileList, distance */);
/*
 * PlaneData    *planeData;
 * LinkedTile   *vertSpaceTileList;
 * int          distance;
 */

extern void Tp_FreeTileList(/* tileList */);
/*
 * LinkedTile *tileList;
 */

extern BearStatus Tp_CreateDCells(/* planeData */);
/*
 * PlaneData *planeData;
 */

extern void Tp_DeleteDCells(/* planeData */);
/*
 * PlaneData *planeData;
 */

extern void Tp_FreeDCellList(/* planeData */);
/*
 * PlaneData *planeData;
 */

extern void Tp_IdentifySpaceTiles(/* planeData */);
/*
 * PlaneData *planeData;
 */

extern int Tp_FindMinHeight();
extern int Tp_FindMinWidth();

extern void Tp_FindBBoxForAllCells(/* planeData, bbox */);
/*
 * PlaneData    *planeData;
 * Rect         *bbox;
 */

extern void Tp_StretchRect(/* rect, percent */);
/*
 * Rect *rect;
 * double percent;
 */

extern BearStatus Tp_ConvertBBLModToCell();

extern BearStatus Tp_CreateTiPlaneForCellMaster(/* cellMasterPtr */);
/*
 * CellMaster *cellMasterPtr;
 */

extern void Tp_ResizeTiPlane(/* planeDataPtr, deltaX, deltaY */);
/*
 * PlaneData   *planeDataPtr;
 * int         deltaX, deltaY;
 */

extern void Tp_MoveTiPlane(/* planeDataPtr, deltaX, deltaY */);
/*
 * PlaneData   *planeDataPtr;
 * int         deltaX, deltaY;
 */
extern BearStatus Tp_RotateCell(/* planeDataPtr, gCellPtr, rotate */);
/*
 * PlaneData   *planeDataPtr;
 * GenericCell *gCellPtr;
 * int         rotate;
 */
extern BearStatus Tp_ReflectCell(/* planeDataPtr, gCellPtr, reflect */);
/*
 * PlaneData   *planeDataPtr;
 * GenericCell *gCellPtr;
 * int         reflect;
 */

extern BearStatus Tp_DuplicateCellInst(/* cellInstPtr, moveDeltaPtr */);
/*
 * CellInst    *cellInstPtr;
 * Point       *moveDeltaPtr;
 */

/* ---------- Macros ---------- */

/* note that hp and vp must be different */
#define Tp_HPointToVPoint(hp, vp) (GeoTransPoint(&Geo90Transform,(hp),(vp)))
#define Tp_VPointToHPoint(vp, hp) (GeoTransPoint(&Geo270Transform,(vp),(hp)))

/* note that hr and vr must be different */
#define Tp_HRectToVRect(hr, vr) (GeoTransRect(&Geo90Transform,(hr),(vr)))
#define Tp_VRectToHRect(vr, hr) (GeoTransRect(&Geo270Transform,(vr),(hr)))

#define Tp_AreSpaceTilesReady(planeDataPtr) \
    ((planeDataPtr)->pd_status & TP_SPACE_TILES_READY)

#endif _TP_TILE

