
/* grStrTags.h --
 *
 *     Malgorzata Marek-Sadowska
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Installed/globalRouter/grStrTags.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef GR_STR_TAGS_H
#define GR_STR_TAGS_H

typedef struct _grPrimit grPrimit;
struct _grPrimit {
    int number;
};

typedef struct _grChain grChain;
struct _grChain {
    BearId name;
    grChain *nextPtr;
};

typedef struct _grBnLine grBnLine;
struct _grBnLine {
    int type; /* 0 for vertical, 1 for horizontal */
    int base; /* x for vertical,y for horizontal*/
    int lb; /* left bottom coordinate*/
    int rt; /* right top coordinate */
    int substitute; /* 1 h left-up,2 h right-up,3 h right-bot*/
            /* 4 h left-bot,5 v bot-right 6 v left-up*/
            /* 7 v bot-left,8 v right-top*/
            /* 9 h left (bot for TP_HORZ_B_LEFT,
               up for TP_HORZ_B_RIGHT) */
            /* 10 v bot ( right for TP_VERT_R_DOWN,
               left for TP_VERT_R_UP*/
    BNLineType bntype;
    BNLine *bnLinePtr;
    Tile *tilePtr; /* pointer to tile where pseudo pins go */
    int cover; /* 0 not covered yet, 1 covered */
    grBnLine *nextBnLine;
};

typedef struct _grRoom grRoom;
struct _grRoom {
    int who; /* block number, 0 for routing */
    int xv; /*right top coordinates*/
    int yv;
    int c_x; /* central point coordinates*/
    int c_y;
    int dist; /* current distance from the start*/
    int status;/* 0 routable, 1 left start, 2 right start,
            3 left target,4 right target, 5 other
            target,6 current btn, 7 forbidden*/
    grRoom *father; /*room from where entered*/
};

typedef struct _grAddress grAddress;
struct _grAddress {
    grRoom      *gRoom;
    grAddress   *next;
};

typedef struct _grBlock grBlock;
struct _grBlock {
    int name; /*this name is for global router use*/
    Tile *tilePtr; /*pointer to the parent tile*/
    BNLine  *bnLinePtr;     /* non NIL for bottleneck tiles substituting
                   bottleneck lines */
    int xpos;   /* absolute coordinates*/
    int ypos;
    int height;
    int width;
    int xvirt; /* virtual coordinates*/
    int yvirt;
    int wvirt;
    int hvirt;
    int persona; /* 1 solid block, 2 vertical bottleneck,3 horizontal btn
            4 vertical btln line, 5 horz btln line, 6 pseudo
            block, 7 dummy block */
    int status; /* 0 completely assigned, 1 needs assignment*/
    int avCap; /* for bottlenecks, how many nets can pass through*/
    int capLeft; /* porousities */
    int capRight;
    int capTop;
    int capBot;

    int dCell;  /* 1 if it is a dummy cell, 0 otherwise */
    grBlock *top;
    grBlock *bot;
    grBlock *right;
    grBlock *left;
    int tooThin;  /* 1 if a bottleneck is too thin, 0 otherwise */

    grBlock *nextBlockPtr; /*ptr to the next block on the list */
};

typedef struct _grKeyBlock grKeyBlock;
struct _grKeyBlock {
    int     name;
    grBlock *blockPtr;
};

typedef struct _grPin grPin;
struct _grPin {
    int x;
    int y; /*absolute coordinates of the pin position*/
    int type; /* 1 regular pin, 2 pseudo(in bottleneck), 3 imaginary */
    PinDirection shade; /* marks pin position with respect to block side*/
    int netname; /* old name of a net, 0 is feedthrough*/
    BearId intname; 
    BearId extname;
    grRoom *room;
    union{
        PinInst *pinInst;
        PseudoPin *pseudoPin;
        /*if PseudoPin points to NULL it means that the pin is imaginary and does*/
        /* not need update                                                       */
    } 
    PinPtr;
    grPin *nextPtr;
};

typedef struct _grKeyToNets grKeyToNets;
struct _grKeyToNets {
    int         key;
    BearString  name;
};

typedef struct _grSortedPin grSortedPin;
struct _grSortedPin {
    grPin *pinPtr; /* pointer to a pin*/
    int side; /* 0 pin is internal, >0 at the boundary*/
    /* 1 left, 2 top, 3 right, 4 bot */
    grSortedPin *nextXPin; /*next pin on x coordinate not necesserily in the same pinRect*/
                                       /*these links are not destroyed*/
    grSortedPin *nextYPin; /*next pin on y coordinate not necesserily in the same pinRect*/
                                       /*these links are not destroyed*/
    grSortedPin *nextxPin; /*next pin with respect to the x coordinates in the same pinRect*/
    struct _grSortedPin *nextyPin; /*next pin with respect to y coordinates in the same pinRect*/
    grSortedPin *nextOnList; /*next pin on the list*/
};

typedef struct _grPinRectangle grPinRectangle;
struct _grPinRectangle {
    grSortedPin *firstxPin;
    grSortedPin *firstyPin;
    int minx;
    int maxx;
    int miny;
    int maxy;
    int count; /* number of pins in the rectangle*/
    int cutMark; /* marks if rectangle has an intersection with cutline*/
    int name; /* name given to prerouted segments*/
    grPinRectangle *nextInNet;
    grPinRectangle *nextOnList;
};

typedef struct _grNet grNet;
struct _grNet {
    int name; /* net name*/
    Net *netBear;
    int critical; /* 0 - regular net, not critical;
                         >0 critical :
                          1: 3*currentLength < maxLength,
                          2: 2*curr<maxLen<3*curr,
                          3: curr<maxLen<2*curr,
                          4: maxLen <= curr                 */
    float maxlength; /* maximum allowable net length */
    float currentLength; /* current approximation of length */
    int xfixed; /* fixed x length in btn*/
    int yfixed; /* fixed y length in btn*/
    int xapprox; /* x current approximation in swboxes*/
    int yapprox; /* y current approximation in swboxes*/
    int xLor;     /* boundary of the initial box for the net*/
    int xRor;
    int yTor;
    int yBor;
    int botHole; /* the first hole from left or bottom which
            is inside the initial box */
    int topHole; /* the last hole which is inside the initial box*/
    int noLoosePins; /* number of loose pins*/
    grSortedPin *firstXPin;
    grSortedPin *firstYPin;
    grPinRectangle *loosePins; /*pins to be connected*/
    int noComponents; /*number of components*/
    grPinRectangle *areConnected;
};

typedef struct _grSuperNode grSuperNode;
struct _grSuperNode {
    grPinRectangle *contentsPtr;
    int kind; /*0 fragment of prerouted net*/
    /*1 marks pins to be connected*/
    BearId nodeName;
    int mark; /* for marking visits*/
    int side; /* 1 sits on the left, 2 on the right*/
    grSuperNode *nextOnList;
};

typedef struct _grRootTree grRootTree;
struct _grRootTree {
    grSuperNode *nodePtr;
    int distance; /* to the best not marked yet*/
    int mark; /* marks visits in edge 0:no visit yet;1 visited;2 removed*/
    int cut;     /* 0 no intersection with cut line, 1 intersection with */
                 /* cut line and assignment, 2 intersection of prerouted */
                 /* net */
    grRootTree *bestNotMarked;
    grRootTree *childPtr;
    grRootTree *siblingPtr;
    grRootTree *fatherPtr;
    grRootTree *nextInSubnet;
    grRootTree *nextOnList;
};

typedef struct _grVirtNet grVirtNet;
struct _grVirtNet {
    grSuperNode *leftNode;
    grSuperNode *rightNode;
    int netName;
    BearId subNetName;
    int assiName;
    int hole;
    grVirtNet *nextInNet;
};

typedef struct _grIndex grIndex;
struct _grIndex {
    grVirtNet *vNetPtr;
};

typedef struct _grSubNet grSubNet;
struct _grSubNet {
    int netname;
    grPinRectangle *toBeConnected;
    grPinRectangle *areConnected;
    grPinRectangle *decomposed;
    BearId subname;
    int critical; /* 0 not critical,
                         >=1 critical*/
    int decCandidate;
    float approxLength; /* approximated length of the subnet*/
    int xapprox;
    int yapprox;
    int active; /* =0 if the subnet is unassesed, 1 needs assignment*/
                    /* 2 no assignment but intersection occurs          */
                    /* 3 sits on left , 4 sits on right                 */
                    /* 5 simple virtNet, 6 one vNet but not all sNodes  */
                    /* included, 7 multiple virtual nets, 8 no vNets    */
    int minx;
    int maxx;
    int miny;
    int maxy;
    int noLoosePins;
    grSuperNode *NodeList;
    int noComponents;
    grRootTree *tree;
    grVirtNet *children; /* virtal nets generated fron the subnet */
    grSubNet *nextOnList;
};

typedef struct _grRegion grRegion;
struct _grRegion {
    int xvirt;
    int yvirt; /* bottom left corner of the region in virtual coordinates*/
    int whvirt; /*virtual width*/
    int hvirt; /* virtual height*/
    int xcutvirt; 
    int ycutvirt; /* virtual cut point*/
    int xpos;
    int ypos; /*bottom left corner of the region in absolute coordinates*/
    int width; /* absolute*/
    int height; /*absolute*/
    int dir; /* dir=0 region does not need to be cut; 1-up direction,
            2-right*/
    grSubNet  *subNetsPtr; /*pointer to resident nets*/
    grPinRectangle *firstPinRectanglePtr;
    grPinRectangle *lastPinRectanglePtr;
    grRegion *childrenPtr;
    grRegion *siblingPtr;
    grRegion *leftPtr;
    grRegion *rightPtr;
    grRegion *nextPtr;
};

typedef struct _grLayer grLayer;
struct _grLayer {
    grRegion    *gRegion;
    grLayer     *nextInLayer;
};

typedef struct _grHole grHole;
struct _grHole {
    grBlock *leftBlockPtr;
    grBlock *rightBlockPtr;
    int status; /* 0 proper bottleneck,1 improper,2no bottleneck*/
            /* 3 bnLine */
    int topOrRight_p_x; /*absolute interval*/
    int topOrRight_p_y;
    int botOrLeft_p_x; 
    int botOrLeft_p_y;
    int leftEst;
    int rightEst;
    int bLBound;
    int rTBound;
    int width;
    int rtpins;
    int lbpins;
    int height;
    int signalCapacity; /* no of possibly passing nets if the hole was empty*/
    int chDensity;
    grChain *netsInHole;
};

typedef struct _grDeadRegion grDeadRegion;
struct _grDeadRegion {
    int xleft;
    int xright;
    int ybot;
    int ytop;
    int side;                   /* 0 left side, 1 right side */
    int botLeftHoleNr;
    int topRightHoleNr;
    int pullControl;            /* 0 needs pull out, 1 - after pull */
    grRegion *region;           /*pointer to a region created from deadRegion*/
    grDeadRegion *dr_inline;    /* next for out-assignment */
    grDeadRegion *partner;      /* assignment to that region
                                 * from the current one.
                                 */
    grDeadRegion *next;         /* next on list */
};

#endif GR_STR_TAGS_H


