
/* macGeometry.h --
 *
 *     Wei-Ming Dai
 *     (Professor E. S. Kuh)
 *
 * This file defines the functions which deal with geometry.
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 * needs: magic.h malloc.h geometry.h bear.h
 * 
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Installed/mac/macGeometry.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef _MAC_GEOMETRY
#define _MAC_GEOMETRY

extern void
Mac_CreateLinkedPoint(/* xCoord, yCoord, lnPointListPtr, pointPtrPtr */);
/*
 * int         xCoord;                             (* IN       *)
 * int         yCoord;                             (* IN       *)
 * LinkedPoint **lnPointListPtr;                   (* IN / OUT *)
 * Point       **pointPtrPtr;                      (* OUT      *)
 */

extern void Mac_CreatePointListFromRect(/* rectPtr, pointListPtr */);
/*
 * Rect         *rectPtr;
 * LinkedPoint  **pointListPtr;
 */

extern BearStatus Mac_IsBoundaryCW(/* boundaryList */);
/*
 * LinkedLineSeg *boundaryList;
 */

extern BearStatus Mac_IsPointListCW(/* pointList */);
/*
 * LinkedPoint *pointList;
 */

extern BearStatus Mac_ReversePointList(/* pointListPtr */);
/*
 * LinkedPoint **pointListPtr;
 */

extern BearStatus Mac_CopyPointList(/* oldList, newListPtr */);
/*
 * LinkedPoint *oldList;
 * LinkedPoint **newListPtr;
 */

extern BearStatus Mac_MergePointLists(/* pointList1, pointList2, 
                                        * newPointListPtr */);
/*
 * LinkedPoint *pointList1;
 * LinkedPoint *pointList2;
 * LinkedPoint **newPointListPtr;
 */

extern BearStatus Mac_DeleteLinkedPoint(/* pointPtr, pointListPtr */);
/* 
 * LinkedPoint *pointPtr;
 * LinkedPoint **pointListPtr;
 */

extern BearStatus Mac_ItrLinkedPoint(/* pointList, func, clientData */);
/*
 * LinkedPoint *pointList;
 * BearStatus  (*func)();
 * ClientData  clientData;
 */

extern BearStatus Mac_FindBBoxForPointList(/* pointList, bbox */);
/*
 * LinkedPoint  *pointList;
 * Rect         *bbox;
 */

extern BearStatus Mac_MakeLineSegList(/* pointList, lineSegListPtr */);
/*
 * LinkedPoint     *pointList;
 * LinkedLineSeg   **lineSegListPtr;
 */

extern void Mac_FreePointList(/* pointList */);
/*
 * LinkedPoint *pointList;
 */ 

extern void Mac_FreeRectList(/* rectList */);
/*
 * LinkedRect   *rectList;
 */


/* ---------- Macros ---------- */

#define Mac_CopyPoint(point1, point2)                                       \
{                                                                           \
    (point1)->p_x = (point2)->p_x;                                          \
    (point1)->p_y = (point2)->p_y;                                          \
}

#define Mac_CopyRect(rect1, rect2)                                          \
{                                                                           \
    (rect1)->r_xbot = (rect2)->r_xbot;                                      \
    (rect1)->r_xtop = (rect2)->r_xtop;                                      \
    (rect1)->r_ybot = (rect2)->r_ybot;                                      \
    (rect1)->r_ytop = (rect2)->r_ytop;                                      \
}

#define Mac_CopyTransform(trans1, trans2)                                   \
{                                                                           \
    (trans1)->t_a = (trans2)->t_a;                                          \
    (trans1)->t_b = (trans2)->t_b;                                          \
    (trans1)->t_c = (trans2)->t_c;                                          \
    (trans1)->t_d = (trans2)->t_d;                                          \
    (trans1)->t_e = (trans2)->t_e;                                          \
    (trans1)->t_f = (trans2)->t_f;                                          \
}

#endif _MAC_GEOMETRY

