
/* newC.h --
 *
 *     Bernhard Eschermann / Yasushi Ogawa / Massoud Pedram
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Global structure definitions and external declarations for the 
 * clustering program cluster.o.
 *
 * Copyright (C) 1991 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Installed/newC/newC.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */


#ifndef NEWC_EXPORT_H
#define NEWC_EXPORT_H /* once-only latch */

#define NEWC_MOD_NAME "newC"


#define MAX_CHILDREN 4 

#define NORTH   MAX_CHILDREN
#define EAST    (MAX_CHILDREN+1)
#define SOUTH   (MAX_CHILDREN+2)
#define WEST    (MAX_CHILDREN+3)

/* constants that characterize the different cluster algorithms */

#define RANDOM      0 /* no real algo, just flips coins, only for comparison */
#define GREEDY      1 /* greedy algorithm by Bernhard Eschermann */
#define MATCHING    2 /* optimal matching algorithm by Mallek Khellaf */
#define SIM_ANNEAL  3 /* simulated annealing */
#define FILE_IN     4 /* input of an old tree from a file */

/*
 * Constants that are used to specify the directions of pins and slots.
 */
#define NCL_NORTH_DIR   0
#define NCL_EAST_DIR    1
#define NCL_SOUTH_DIR   2
#define NCL_WEST_DIR    3

#define NclForeachSide(_X) \
    for (_X = 0; _X < 4 ;_X++)


typedef int Ncl_SideDir;

typedef enum _Ncl_CutDir Ncl_CutDir;
enum _Ncl_CutDir {
    VERT, HORZ
};

typedef enum _Ncl_NodeType Ncl_NodeType;
enum _Ncl_NodeType {
    NON_LEAF    = 0,
    P_LEAF      = 1,
    LEAF        = 2
};

typedef enum _Ncl_NodeStatus {
    PLACED      = 0, 
    ACTIVE      = 1,
    NOT_PLACED  = 2
} Ncl_NodeStatus;

typedef struct _Ncl_SlotData {
    Ncl_SideDir     slotDir;
    Point           tilePos;    /* Coordinates of the slot on the tile plane */
    int             col;        /* Index into 'plCostMatrix'    */
} Ncl_SlotData;

typedef struct _Ncl_LengthData Ncl_LengthData;
struct _Ncl_LengthData {
    float           a, l;
    int             pattern, *label, orient, belongsTo;
    Ncl_LengthData  *source[MAX_CHILDREN];
};

typedef struct _Ncl_LengthSet Ncl_LengthSet;
struct _Ncl_LengthSet {
    Ncl_LengthData  *length;
    int             size;
};

typedef struct _Ncl_ShapeData Ncl_ShapeData;
struct _Ncl_ShapeData {
    float           sd_x, sd_y;
    int             pattern, *label, orient, belongsTo;
    Ncl_ShapeData   *source[MAX_CHILDREN];
};

typedef struct _Ncl_ShapeSet Ncl_ShapeSet;
struct _Ncl_ShapeSet {
    Ncl_ShapeData   *shape;
    int             size;
};


typedef struct _Ncl_Node {
   int          nd_id;              /* top-down left-right increasing */
   int          nd_index;           /* index into the parentPtr->nd_children */
   int          nd_mark;
   int          nd_depth;           /* depth=0 for root */
   int          nd_numOfChildren;   /* number of direct cluster elements */
   Ncl_NodeType  nd_type;
   Ncl_NodeStatus nd_status;
   int          nd_pattern;
   int          *nd_label;
   int          nd_orient;
   double       nd_area;            /* Total area of all cells in this
                                     * cluster.
                                     */
   Point        nd_dim;             /* x=width and y=height of cluster */
   Point        nd_storeDim;        /* backup dimension for node */
   Transform    nd_nodeTrans;       /* Transform from origin to intermediate
                                     * position.
                                     */
   Transform    nd_tileTrans;   /* transfo from origin to the tile plane */
   Transform    nd_storeTrans;  /* back-up transfo for node */
   Rect         nd_tileRect;    /* Coordinates of the node on the tile plane */
   int          nd_maxDim;         /* extreme nd_dim in this cluster */
   CellInst     *nd_cellInstPtr;   /* only used for leaf level nodes */
   struct   _Ncl_Node        *nd_children[MAX_CHILDREN];   /* down in tree */
   struct   _Ncl_Node        *nd_parentPtr;                /* up in tree */
   struct   _Ncl_Edge        *nd_netList;        /* connections */
   Ncl_ShapeSet           *nd_shapeSet;      /* shape function */
   Ncl_LengthSet          *nd_lengthSet;     /* length function */
   struct   _Ncl_LinkedCritNetInNode *nd_lnkCritNetList;
   int          nd_numOfPins;

   lsList       nd_totalNetList;  /* This list contains all of the nets that
                                   * belong to this node.
                                   */
   lsList       nd_bdrNetList;    /* This list is a subset of 'nd_totalNetList'
                                   * and contains only those nets that attach
                                   * to this nodes border.
                                   */
   lsList       nd_initPinList[4];
   lsList       nd_tilePinList[4];

   int          nd_fixPinCnt;           /* Count of fixed pins. */
   int          nd_freeFloPinCnt;       /* Count of floating pins that can be
                                         * placed on any side of the node.
                                         */

   int          nd_sideFloPinCnt[4];    /* Count of floating pins for a 
                                         * particular side.
                                         */
   int          nd_totalPinCnt;         /* Total pin count.                 */

   bool         nd_hasSlotsFlag;
   Ncl_SlotData *nd_northSlot;
   int          nd_numOfNorthSlots;
   Ncl_SlotData *nd_southSlot;
   int          nd_numOfSouthSlots;
   Ncl_SlotData *nd_westSlot;
   int          nd_numOfWestSlots;
   Ncl_SlotData *nd_eastSlot;
   int          nd_numOfEastSlots;

   int          nd_connectivity[MAX_CHILDREN][MAX_CHILDREN+4];
   int          nd_spacing[MAX_CHILDREN][MAX_CHILDREN+4];

   ClientData   clientData;
} Ncl_Node;

typedef struct _Ncl_ScratchPad {
    float  intra_l, inter_l;
    bool   hasInternalEdge, hasExternalEdge, connectsChild[MAX_CHILDREN];
    Rect   rect;
    int    numEdges;
} Ncl_ScratchPad;

typedef struct _Ncl_LinkedCritNet {
    struct _Net         *netPtr;
    float               used_l;
    double              timeWeight;
    Ncl_LengthSet       tmp_ld, ld; 
    Ncl_ScratchPad      inNode;
    int                 numPseudoPins;
    struct _Ncl_LinkedCritNet *nextPtr;
} Ncl_LinkedCritNet;

typedef struct _Ncl_LinkedCritNetOnEdge {
    Ncl_LinkedCritNet               *critNetPtr;
    float                           weight2;
    struct _Ncl_LinkedCritNetOnEdge *nextPtr;
} Ncl_LinkedCritNetOnEdge;

typedef struct _Ncl_LinkedCritNetInNode Ncl_LinkedCritNetInNode;
struct _Ncl_LinkedCritNetInNode {
    Ncl_LinkedCritNet       *critNetPtr;
    Ncl_LengthSet           ld;
    float                   bu_l, inter_l, best_inter_l;
    int                     numOfChildren;
    struct  _Ncl_LinkedEdge *edgeList[MAX_CHILDREN];
    Ncl_LinkedCritNetInNode *children[MAX_CHILDREN];
    int                     connectivity[MAX_CHILDREN][MAX_CHILDREN+4];
    Ncl_LinkedCritNetInNode *nextPtr;
};

typedef struct _Ncl_Edge {
   Ncl_Node                 *connectedNode;
   int                      weight;     /* # of connections x 100 represented by edge */
   int                      edgeType;   /* BLOCK, NORTH, EAST, SOUTH, WEST */
   Ncl_LinkedCritNetOnEdge  *critOnEdgePtr;
   struct   _Ncl_Edge           *nextPtr;
} Ncl_Edge;

typedef struct _Ncl_LinkedEdge {
    Ncl_Edge                 *edgePtr;
    struct   _Ncl_LinkedEdge *nextPtr;
} Ncl_LinkedEdge;

/*
 * The following parameters may be set by the user.
 */
typedef struct _Ncl_UserParms Ncl_UserParms;
struct _Ncl_UserParms {
    int             algorithm;

    /* Begin matching only.*/ 
    int             *areaRatioPtr; /* Needs to be pointer for iv_promptVar 
                                    * to work.
                                    */
    int             *dimRatioPtr;  /* Needs to be pointer for iv_promptVar 
                                    * to work.
                                    */
    bool            promptFlag;
    /* End matching only.  */ 

    /* Begin greedy only.  */ 
    double          areaFact;
    double          areaRelax;
    double          ratioMax;
    double          ratioRelax;
    /* End greedy only.    */ 

    double          connMin;
    double          routAdjust;

    bool            quickRtEstFlag;     /* quick route estimate */
    bool            chanFreeFlag;       /* channel free */
    bool            timingFlag;
    bool            storeFlag;

    char            *clusterFileName;
    char            *storageFileName;
    char            *timingFileName;

    /* These parameters are for debugging only. */
    unsigned int    maxNumOfShapes;
};
    

/* ---------- External procedures for the newC module ---------- */

EXTERN BearStatus Ncl_BearInterface
    ARGS((CellMaster *cellMasterPtr, Ncl_UserParms *parmsPtr, 
        BearGeneric ivWinPtr));

EXTERN Ncl_ShapeData *Ncl_InitShapes
    ARGS((int shapeCount));

EXTERN BearStatus Ncl_MergeShapeSets
    ARGS((Ncl_ShapeSet *lShapeSet, Ncl_ShapeSet *rShapeSet, 
        Ncl_ShapeSet *mergedShapeSet));

EXTERN Ncl_ShapeSet *Ncl_AddShapeSets
    ARGS((Ncl_ShapeSet *lShapeSet, Ncl_ShapeSet *rShapeSet, 
        Ncl_CutDir cutDir));

EXTERN void Ncl_PrintShapeSets
    ARGS((Ncl_ShapeSet *shapeSet));

EXTERN Ncl_SideDir Ncl_GetOppositeSideDir
    ARGS((Ncl_SideDir sideDir));

EXTERN char *Ncl_GetSideDirName
    ARGS((Ncl_SideDir sideDir));

EXTERN void Ncl_ResetShapeLists
    ARGS((CellMaster *coreMasterPtr));

EXTERN BearStatus Ncl_ReadControlData
    ARGS((char *fileNamePtr, Ncl_UserParms *parmsPtr));

EXTERN BearStatus Ncl_ReadShapeData
    ARGS((CellMaster *coreMasterPtr, char *shapeFileNamePtr));

EXTERN BearStatus Ncl_WriteTiming
    ARGS((CellMaster *cellMasterPtr, char *timingFileNamePtr));

EXTERN Ncl_UserParms *Ncl_MakeUserParms
    NULLARGS;

EXTERN void Ncl_FreeUserParms
    ARGS((Ncl_UserParms **parmsPtrPtr));

EXTERN void Ncl_EchoUserParms
    ARGS((char *msgPtr, Ncl_UserParms *parmsPtr));

#ifdef DEBUG
EXTERN BearStatus Ncl_WriteBBLShapes
    ARGS((char *cmFileNamePtr, lsList ciList, char *shapeFileNamePtr));
#endif DEBUG


/* ---------- External global variables for the newC module ---------- */

extern Ncl_Node             *ncl_ClusterTree;  /* Contains the output of
                                                * NclBuildTree()
                                                */
extern Ncl_LinkedCritNet    *ncl_CritNetList;

extern int                  ncl_ClusterTreeDepth;
extern bool                 ncl_TimingFlag;
extern bool                 ncl_QuickRouteEstFlag;
extern bool                 ncl_ChannelFreeFlag;
extern double               ncl_WireWidth;
extern double               ncl_RouteAdjust;
extern unsigned int         ncl_MaxNumOfShapes;

#endif NEWC_EXPORT_H

