
/* pgRoute.h --
 *
 *     Robi Dutta
 *     (Professor E. S. Kuh)
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Installed/pgRoute/pgRoute.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef _PGROUTE
#define _PGROUTE

#define MAXWIDTH ((double) 10000.0)

typedef enum _Pgr_NodeType {
    PIN, JUNCT
} Pgr_NodeType;

typedef struct _Pgr_NodeProp {
    Pgr_NodeType            nodeType;
    int                     nodeNum;
    double                  maxVoltDrop;
    double                  maxCurrentReq;
    BearGeneric             genericPtr;
    struct _Pgr_NodeProp    *nextPtr;
} Pgr_NodeProp;

typedef struct _Pgr_EdgeProp {
    int                     edgeNum;
    Pgr_NodeProp            *node1Ptr, *node2Ptr;
    double                  length;
    double                  width;  /* MAXWIDTH is default value */
    BearGeneric             genericPtr;
    struct _Pgr_EdgeProp    *nextPtr;
} Pgr_EdgeProp;

#define PGR_ELECT_FLAG      01      /* include electromigration constraints */
#define PGR_VOLT_FLAG       02      /* include voltage constraints          */
#define PGR_FEASIBLE_FLAG   04      /* if only a feasible solution          */

typedef struct _Pgr_WireTech {
/*
 * This structure contains the technology and design style information.
 */
    float grid;         /* grid size in microns */
    float conductance;  /* conductance per square */
    float curPerMicron; /* current per micron */
    float minWidth;     /* width tolerance */
    int   flags;        
    int   timeSteps;    /* number of time steps for current
                         * required.
                         */
} Pgr_WireTech;

typedef struct _Pgr_Wire {
    int          nodeCount;
    int          edgeCount;
    Pgr_NodeProp *nodeList;
    Pgr_EdgeProp *edgeList;
} Pgr_Wire;

/* ---------- External procedures for the pgRoute module ---------- */

extern BearStatus Pgr_CalcWireWidths(/* wireTechPtr, wirePtr */);
/*
 * Pgr_WireTech *wireTechPtr;
 * Pgr_Wire     *wirePtr;
 */

extern Pgr_NodeProp *Pgr_MakeNodeProp();

extern Pgr_EdgeProp *Pgr_MakeEdgeProp();

extern Pgr_Wire *Pgr_MakeWire();

extern void Pgr_FreeWire(/* wirePtrPtr */);
/*
 * Pgr_Wire     **wirePtrPtr;
 */

extern BearStatus Pgr_ReadWireTech(/* pgFileNamePtr, wireTechPtr */);
/*
 * char         *pgFileNamePtr;
 * Pgr_WireTech *wireTechPtr;
 */

extern void Pgr_PrintPgWireData(/* fileNamePtr, pgrWirePtr */);
/*
 * char         *fileNamePtr;
 * Pgr_Wire     *pgrWirePtr;
 */

/* ---------- External global variables for the pgRoute module ---------- */
extern bool pgr_DebugFlag;

#endif _PGROUTE

