
/* tpPPin.h -
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/Installed/tileProp/tpPPin.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef _TP_PPIN
#define _TP_PPIN

/*
 * A psuedo pin pointer is attached to a side of a space tile
 * to keep track the global routing information.
 * From the point of the view local to a tile, two psuedoPins may
 * not to be connected even they belong to the same global 
 * net. So we need the notion of local nets. The internal 
 * net of a psuedoPin is the net it connects inside the space 
 * tile; the external net is the one it connects outside of 
 * the tile. We use an integer number to specify the relative
 * ordering of the psuedoPin.
 */

typedef struct _PseudoPin {
    BearId            id;           /* The pseudo pins identification
                                     * number.
                                     */
    struct _Net       *netPtr;
    BearId            intNetId;     /* The internal net identification number
                                     * of the psuedo pin.
                                     */
    BearId            extNetId;     /* The external net identification number
                                     * of the psuedo pin.
                                     */
    SideDirection     side;         /* Tells you which side of the bottleneck
                                     * tile this pseudo pin is on.
                                     */
    float             width;
    int               order;        /* The relative ordering of the psuedo
                                     * pin on one side of a tile. 
                                     */
    Tile              *tilePtr;
    struct _PseudoPin *nextPtr;     /* The next psuedo pin in the list      */
    ClientData        clientData;   /* This space for rent.                 */
} PseudoPin;

typedef struct _LinkedPPin {
    struct _LinkedPPin  *prevPtr;
    struct _LinkedPPin  *nextPtr;
    PseudoPin           *ppinPtr;
} LinkedPPin;

/* ----------- Pseudo Pin Operations ----------- */

extern BearStatus Tp_CreatePPin(/* bnTile, side, fragment, ppin */);
/*
 * Tile             *bnTile;
 * SideDirection    side;
 * int              fragment;
 * PseudoPin        **ppin;
 */

extern BearStatus Tp_CreatePPinOfBnLine(/* bnLine, ppin */);
/*
 * BNLine           *bnLine;
 * PseudoPin        **ppin;
 */

extern BearStatus Tp_CreatePPinOfBnTile(/* bnTile, side, ppin */);
/*
 * Tile             *bnTile;
 * SideDirection    side;
 * PseudoPin        **ppin;
 */

extern BearStatus Tp_CreatePPinOfDCell(/* dCellPtr, side, ppinPtrPtr */);
/*
 * struct _DummyCell    *dCellPtr;
 * SideDirection        side;
 * PseudoPin            **ppinPtrPtr;
 */

extern BearStatus Tp_InsertPPinOfBnTile(/* bnTile, side, ppin */);
/*
 * Tile             *bnTile;
 * SideDirection    side;
 * PseudoPin        *ppin;
 */

extern BearStatus Tp_InsertPPinOfBnLine(/* sbnTile, side, fragment, ppin */);
/*
 * Tile             *sbnTile;
 * SideDirection    side;
 * int              fragment;
 * PseudoPin        **ppin;
 */

extern BearStatus Tp_DeletePPinOfBnTile(/* bnTile, side, ppin */);
/*
 * Tile             *bnTile;
 * SideDirection    side;
 * PseudoPin        *ppin;
 */

extern BearStatus Tp_DeletePPinOfBnLine(/* sbnTile, side, fragment, ppin */);
/*
 * Tile           *sbnTile;
 * SideDirection  side;
 * int            fragment;
 * PseudoPin      *ppin;
 */

extern void Tp_FreePPinsOfBnLine(/* bnLinePtr */);
/*
 * BNLine *bnLinePtr;
 */

extern void Tp_FreePPinsOfTile(/* tilePtr */);
/*
 * Tile             *tilePtr;
 */

extern void Tp_FreePPinListOfSide(/* tilePtr, side */);
/*
 * Tile             *tilePtr;
 * SideDirection    side;
 */

extern void Tp_FreePPinListOfFragment(/* tilePtr, side, fragment */);
/*
 * Tile             *tilePtr;
 * SideDirection    side;
 * int              fragment;
 */

extern void Tp_RemovePPinListOfTile(/* tilePtr, side, fragment, ppinListPtr */);
/*
 * Tile             *tilePtr;
 * SideDirection    side;
 * int              fragment;
 * PseudoPin        **ppinListPtr;
 */

extern void Tp_AppendPPinListOfTile(/* ppinList, tilePtr, side, fragment */);
/*
 * PseudoPin        *ppinList;
 * Tile             *tilePtr;
 * SideDirection    side;
 * int              fragment;
 */

/* ---------- Pseudo Pin Property Primitives ---------- */

extern float Tp_GetWidthOfPPin (/* ppin */);
/* PseudoPin *ppin; */

extern void Tp_PutWidthOfPPin (/* width, ppin */);
/*
float     width;
PseudoPin *ppin;
*/

extern int Tp_GetOrderOfPPin (/*ppin*/);
/* PseudoPin *ppin; */

extern void Tp_PutOrderOfPPin (/*order, ppin*/);
/*
int       order;
PseudoPin *ppin;
*/

extern SideDirection Tp_GetSideOfPPin(/* ppin */);
/* PseudoPin *ppin; */

extern Tile *Tp_GetTileOfPPin(/* ppin */);
/* PseudoPin *ppin; */

extern BearStatus Tp_GetCenterOfSide(/* tile, side, center */);
/*
 * Tile             *tile;
 * SideDirection    side;
 * Point            *center;
 */

extern struct _Net *Tp_GetNetOfPPin(/* ppin */);
/* PseudoPin *ppin; */

extern void Tp_PutNetOfPPin(/* net, ppin */);
/* 
 * struct _Net  *net;
 * PseudoPin    *ppin;
 */

extern BearId Tp_GetIdOfPPin(/* ppin */);
/* PseudoPin *ppin; */

extern void Tp_PutIdOfPPin(/* id, ppin */);
/*
 * BearId       id;
 * PseudoPin    *ppin;
 */

extern BearId Tp_GetIntNetIdOfPPin(/* ppin */);
/* PseudoPin *ppin; */

extern void Tp_PutIntNetIdOfPPin(/* intNetId, ppin */);
/*
 * BearId    intNetId;
 * PseudoPin *ppin;
 */

extern BearId Tp_GetExtNetIdOfPPin(/* ppin */);
/* PseudoPin *ppin; */

extern void Tp_PutExtNetIdOfPPin(/* extNetId, ppin */);
/*
 * BearId    extNetId;
 * PseudoPin *ppin;
 */


/* ---------- Interators ---------- */

extern BearStatus Tp_ItrPPinsOfBnTile(/* bnTile, func, clientData */);
/*
 * Tile             *bnTile;
 * BearStatus       (*func)();
 * ClientData       clientData;
 */

extern BearStatus Tp_ItrPPinsOfSide(/* bnTile, side, func, clientData */);
/*
 * Tile             *bnTile;
 * SideDirection    side;
 * BearStatus       (*func)();
 * ClientData       clientData;
 */

extern BearStatus Tp_ItrPPinsOfFragment(/* bnTile, side, fragment,
    func, clientData */);
/*
 * Tile             *bnTile;
 * SideDirection    side;
 * int              fragment;
 * BearStatus       (*func)();
 * ClientData       clientData;
 */

extern BearStatus Tp_ItrPPinsOfBnLine(/* tile1, tile2, func, clientData*/);
/*
 * Tile            *tile1, *tile2;
 * BearStatus      (*func)();
 * ClientData      clientData;
 */

extern BearStatus Tp_ItrPPinsOfGraph(/* graph, func, clientData */);
/*
 * Gt_Graph   graph;
 * BearStatus (*func)();
 * ClientData clientData;
 */

extern BearStatus Tp_NumOfPPinsOfGraph(/* graph, numPtr */);
/*
 * Gt_Graph   graph;
 * int        *numPtr;
 */

/* ---------- Debugging stuff ---------- */

extern void Tp_PrintPPinListsOfTile(/* tilePtr, netPtr */);
/*
 * Tile *tilePtr;
 * Net  *netPtr;
 */

extern void Tp_PrintPPinListOfBnLine(/* bnLinePtr, netPtr */);
/*
 * BNLine *bnLinePtr;
 * Net    *netPtr;
 */


extern BearStatus Tp_PrintPPin(/* ppinPtr, clientData */);
/*
 * PseudoPin    *ppinPtr;
 * ClientData   clientData;
 */

extern BearStatus
Tp_GetAbsPosOfPPin(/* planeDataPtr, ppinPtr, absPosPtr */);
/*
 * PlaneData   *planeDataPtr;                      (* IN       *)
 * PseudoPin   *ppinPtr;                           (* IN       *)
 * Point       *absPosPtr;                         (* OUT      *)
 */

#endif _TP_PPIN

