
/* bdiInputCheck.h --
 *
 *    (Advising Professor - E. S. Kuh)
 *
 * Definitions internal to the input module.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/bblInterface/bdiInputCheck.h,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $
 */

#ifndef _BDI_INPUT_CHECK
#define _BDI_INPUT_CHECK

/*
 *  Assorted Utilities.
 */

/*  assert.h    4.1 83/05/03    */

#if !defined(assert) /* { */
#ifdef DEBUG
#define _assert(ex) {if (!(ex)){fprintf(stderr,"Assertion failed: file %s, line %d\n", __FILE__, __LINE__);abort();}}
#define assert(ex) {if (!(ex)){fprintf(stderr,"Assertion failed: file %s, line %d\n", __FILE__, __LINE__);abort();}}
#else
#define _assert(ex) ;
#define assert(ex) ;
#endif
#endif /* } assert */

#define bascii(b) ((b)?"true":"false")

#define MAXLINE  150
#define streq(x,y) (strcmp((x),(y))==0)

#define NETNAMEMAX (40+1)
#define NETMAX 1000
#define MODNAMEMAX (40+1)

typedef struct _Net_Tab_Entry {
    char    NTAB_net_name[NETNAMEMAX];
    int     NTAB_term_count;
} Net_Tab_Entry;

typedef enum _Mod_Type {
    REG_MOD =       0,
    ROUT_MOD =      1
} Mod_Type;

typedef enum _Rout_Dir {
    LEFT_DIR =      0,
    DOWN_DIR =      1,
    RIGHT_DIR =     2,
    UP_DIR =        3
} Rout_Dir;

typedef enum _Term_Type {
    FIX_TERM =          2,
    FLOAT_TERM =        3,
    FLOAT_NORTH_TERM =  4,
    FLOAT_EAST_TERM =   5,
    FLOAT_SOUTH_TERM =  6,
    FLOAT_WEST_TERM =   7
} Term_Type;

typedef struct _Lex_Pos {
    int LP_line;
    int LP_col;
} Lex_Pos;

typedef struct _Coord {
    int CO_x;
    int CO_y;
} Coord;

typedef struct _Edge_Node {
    Lex_Pos             position;
    Coord               EN_coord;
    struct _Edge_Node   *EN_next;
} Edge_Node;

typedef struct _Term_Node {
    Lex_Pos             position;
    int                 TN_coordDefFlag;
    Coord               TN_coord;
    char                TN_parent_net_name[NETNAMEMAX];
    Rout_Dir            TN_rout_dir;
    Term_Type           TN_term_type;
    int                 TN_pg_flag;
    double              TN_pg_currentReq;
    double              TN_pg_voltDrop;

    struct _Term_Node   *TN_next;
} Term_Node;

typedef struct _Mod_Node {
    Lex_Pos             position;
    Coord               M_coord;
    char                M_name[MODNAMEMAX];
    int                 M_trans_num;        /* transform number */
    Mod_Type            M_type;
    Edge_Node           *M_edge_list;
    Term_Node           *M_term_list;
    struct _Mod_Node    *M_next;
} Mod_Node;

typedef struct _Root_Module_Node {
    Lex_Pos             position;
    int                 RM_net_number;
    Mod_Node            *RM_mod;
    Mod_Node            *RM_sons_list;
} Root_Mod_Node;



extern void       BdiInputCheckBegin();
extern void       BdiInputCheckEnd();
extern BearStatus BdiReadRootMod();

extern void Show_Term_List( /* term_list */ );
/* Term_Node *term_list; */

extern void Show_Term( /* term */ );
/* Term_Node *term; */

extern void Show_Edge_List( /* edge_list */ );
/* Edge_Node * edge_list; */

extern void Show_Edge( /* edge */ );
/* Edge_Node * edge; */

extern void Show_Mod( /* mod, gory_detail */ );
/* Mod_Node *mod;
   int gory_detail; 
*/

extern void Show_Root_Mod ( /* root, gory_detail */ );
/* Root_Mod_Node *root;
   int goru_detail;
*/

extern void Show_Coord ( /* co */ );
/* Coord *co; */

extern void Show_Mod_Type ( /* mod_type */ );
/* Mod_Type mod_type; */

extern void Show_Rout_Dir ( /* rout_dir */ );
/* Rout_Dir rout_dir; */

extern void Show_Term_Type ( /* term_type */ );
/* Term_Type term_type; */

extern void Show_Mod_List ( /* mod_list, gory_detail */ );
/* Mod_Node *mod_list; 
   int gory_detail; 
*/

extern void Check_All_Nets_For_Single_Term();

extern void Find_Wall_Of_Term ( /* mod, term, from_edge, to_edge */ );
/* Mod_Node *mod; Term_Node *term; Edge_Node **from_edge, **to_edge; */

extern Rout_Dir Dir_Of_Mod_Wall ( /* from_edge, to_edge */ );
/* Edge_Node *from_edge, *to_edge; */

extern void Check_Term_Rout_Dir ( /* mod, term; */ );
/* Mod_Node *mod; Term_Node *term; */

extern void Check_All_Terms_Rout_Dir ( /* mod */ );
/* Mod_Node *mod; */

extern void Check_All_Terms_On_Corner( /* mod */ );
/* Mod_Node *mod; */

extern void Check_All_Terms_On_Bound( /* mod */ );
/* Mod_Node *mod; */


extern void Check_Term_On_Boundary( /* mod, term */ );
/* Mod_Node *mod;
   Term_Type *term;
*/

 
extern void Check_Mod( /* mod */ );
/* Mod_Node *mod; */

extern Rout_Dir Dir_Of_Wall( /* from_edge, to_edge */ );
/* Edge_Node *from_edge;
   Edge_Node *to_edge;
*/
extern LinkedPoint *FindHeadPoint();

extern void BdiFreeRootMod();


/* ---------- Global variables used within this module. ---------- */
extern FILE *bdiFilePtr;
extern bool bdiCheckFileFlag;
extern int  bdiErrorCount;

extern int  cur_line;
extern int  display_input_option;
extern int  check_rout_dir;
extern char line_buf[MAXLINE];

#endif _BDI_INPUT_CHECK

