
/* bfpRead.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 *     The whole interface presented here should be redone using 
 *     bison or yacc.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/bfpInterface/bfpRead.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "rectSlice.h"
#include "bfpInterface.h"
#include "bfpInterfaceInt.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */

#define MAX_LINE_SIZE 1024

#define M_NAME_FORMAT "%s_master"

#define RETURN_ERR \
{ \
    if (filePtr != NIL(FILE)) { \
        (void) fclose (filePtr); \
        filePtr = NIL(FILE); \
    } \
    return (BEAR_ERROR); \
}

#define READ_LINE \
    if (ReadLine(curLine, MAX_LINE_SIZE, filePtr, &eofFlag) == \
        BEAR_ERROR) { \
        RETURN_ERR; \
    } \
    if (eofFlag == TRUE) { \
        (void) fclose (filePtr); \
        filePtr = NIL(FILE); \
        if (okayToStopFlag == FALSE) { \
            PrintInputErr("Unexpected end of file."); \
            return (BEAR_ERROR); \
        } \
        return (BEAR_OK); \
    }

#define GET_KEYWORD_STR strtok(curLine, " ")

#define MISSING_KEYWORD_ERR(_X) \
{ \
    (void) sprintf (bear_ErrorMsg, \
        "%s key word is missing at line %d.", (_X), lineNum); \
    PrintInputErr(bear_ErrorMsg); \
    RETURN_ERR; \
}

#define LOOK_FOR_KEYWORD(_X) \
    keyWordPtr = GET_KEYWORD_STR; \
    if (strcmp(keyWordPtr, (_X)) != 0) { \
        MISSING_KEYWORD_ERR(_X); \
    }

#define LOOK_FOR_VALUE(_X) \
    if ((valuePtr = strtok(NIL(char), " ")) == NIL(char)) { \
        (void) sprintf (bear_ErrorMsg, \
            "%s's %s is missing at line %d.", keyWordPtr, (_X), lineNum); \
        PrintInputErr(bear_ErrorMsg); \
        RETURN_ERR; \
    }

#define CHECK_WIDTH(_X) \
    if ((_X) <= 0) { \
        (void) sprintf (bear_Msg, "width on line %d must be greater than 0.", \
            lineNum); \
        PrintInputErr(bear_Msg); \
        return (BEAR_ERROR); \
    }

#define CHECK_HEIGHT(_X) \
    if ((_X) <= 0) { \
        (void) sprintf (bear_Msg, "height on line %d must be greater than 0.", \
            lineNum); \
        PrintInputErr(bear_Msg); \
        return (BEAR_ERROR); \
    }

typedef struct _ReadData ReadData;
struct _ReadData {
    CellMaster  *coreMasterPtr;

    Point       bdrOrigin;          /* border origin */
    double      bdrWidth;           /* border width */
    double      bdrHeight;          /* border height */
    Point       cellOrigin;         /* cell origin */
};

typedef struct _InputBorderPin InputBorderPin;
struct _InputBorderPin {
    BearString  pinName;
    BearString  netName;
    int         sideValue;
    double      fraction;
};

typedef struct _InputPin InputPin;
struct _InputPin {
    BearString  pinName;
    BearString  netName;
    Flexibility pinType;
    bool        pinPosDefFlag;
    Point       relPos;             /* relative position */
};


/* ---------- External Procedures ---------- */
extern char         *util_strsav();
extern char         *util_tilde_expand();


/* ---------- Local Procedures ---------- */
static void         PrintInputErr();
static BearStatus   CreateDefaultChip();
static BearStatus   FillInCellInst();
static BearStatus   ReadLine();
static void         FreeInputPin();
static void         FreeInputBorderPin();

static BearStatus   ReadInputV1();
static BearStatus   ReadBorderCellV1();
static BearStatus   ReadBorderPinV1();
static BearStatus   ReadCellV1();
static BearStatus   ReadCellPinV1();
static BearStatus   ReadCellShapeV1();
static BearStatus   StoreCellPinsV1();
static BearStatus   StoreBorderPinsV1();


/* ---------- Local Global Variables ---------- */
static BearString   fileName;
static FILE         *filePtr;
static int          lineNum;
static char         curLine[MAX_LINE_SIZE];
static char         *keyWordPtr, *valuePtr;
static bool         eofFlag, okayToStopFlag;


void
BfpPrintFatalErr(message)
BearString message;
{
    errRaise(BFP_MOD_NAME, ((int) -1), message);
}


static void
PrintInputErr(message)
BearString message;
{
    (void) fprintf (bearerr, "BFP Parser : Error : %s\n", message);
}


static BearStatus
ReadLine(line, maxLineSize, filePtr, eofFlagPtr)
char    line[];             /* OUT */
int     maxLineSize;        /* IN  */
FILE    *filePtr;           /* IN  */
bool    *eofFlagPtr;        /* OUT */
/*
 * This procedure will skip over blank lines.
 */
{
    char *stringPtr;

    while (TRUE) {
        if (fgets(line, maxLineSize, filePtr) == (char *) NULL) {
            *eofFlagPtr = TRUE;
            return (BEAR_OK);
        }
        else if ((stringPtr = strchr(line, '\n')) == (char *) NULL) {
            PrintInputErr("Buffer overflow.");
            return (BEAR_ERROR);
        }

        /* Remove the new line character. */
        *stringPtr = '\0';

        lineNum++;
        if (Bear_AllWhiteSpace(line) == TRUE) {
            /* Blank line. */
            continue;
        }
        else if (line[0] == '#') {
            /* Comment line. */
            continue;
        }
        else {
            return (BEAR_OK);
        }
    }
    /* NOTREACHED */
}


static void
FreeInputPin(item)
lsGeneric item;
{
    InputPin *inputPinPtr = (InputPin *) item;

    FREE(inputPinPtr->pinName);
    FREE(inputPinPtr->netName);
    FREE(inputPinPtr);
}


static void
FreeInputBorderPin(item)
lsGeneric item;
{
    InputBorderPin *inputBPinPtr = (InputBorderPin *) item;

    FREE(inputBPinPtr->pinName);
    FREE(inputBPinPtr->netName);
    FREE(inputBPinPtr);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Bfp_BearFPToMac() --
 *
 *    Read in information that was stored using the Bear floor
 *    planning format and generate the cooresponding MAC data base.
 *
 * Results:
 *    BEAR_OK is returned if everything is spiffy.
 *    BEAR_ERROR otherwise.
 *
 * Side effects:
 *    None.
 * 
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
BearStatus
Bfp_BearFPToMac(bearFPFileName, readOptsPtr, coreMasterPtrPtr)
BearString      bearFPFileName;                 /* IN       */
Bfp_ReadOptions *readOptsPtr;                   /* IN       */
CellMaster      **coreMasterPtrPtr;             /* OUT      */
{
    fileName = bearFPFileName;

    *coreMasterPtrPtr = NIL(CellMaster);

#ifdef USE_OCT_CODE
    (void) VOVinput(VOV_UNIX_FILE, fileName);
#endif /* USE_OCT_CODE */

    filePtr = fopen (fileName, "r");
    if (filePtr != NIL(FILE)) {
        double version;

        lineNum = 0;
        okayToStopFlag = FALSE;
        eofFlag = FALSE;

        READ_LINE;
        LOOK_FOR_KEYWORD("Version");
        LOOK_FOR_VALUE("number");
        (void) sscanf (valuePtr, "%lf", &version);

        if (version == 1.0) {
            if (ReadInputV1(coreMasterPtrPtr) != BEAR_OK) {
                RETURN_ERR;
            }
        }
        else {
            PrintInputErr("unknown version number.");
            RETURN_ERR;
        }
    }
    else {
        if (CreateDefaultChip(coreMasterPtrPtr) != BEAR_OK) {
            RETURN_ERR;
        }
    }

    return (BEAR_OK);
}


static BearStatus
ReadInputV1(coreMasterPtrPtr)
CellMaster **coreMasterPtrPtr;
{
    BearStatus  bearStatus;
    ReadData    readData;
    char        *techFileNamePtr;

    okayToStopFlag = FALSE;
    *coreMasterPtrPtr = NIL(CellMaster);

    READ_LINE;
    LOOK_FOR_KEYWORD("TechFileName");
    LOOK_FOR_VALUE("argument");
    if (strcmp(valuePtr, BFP_UNDEF_STR) == 0) {
        techFileNamePtr = NIL(char);
    }
    else {
        if (strstr(valuePtr, "BEAR") == NIL(char)) {
            techFileNamePtr = util_strsav(valuePtr);
        }
        else {
            if (getenv("BEAR") == NIL(char)) {
                PrintInputErr("The environment variable $BEAR is undefined.");
                return (BEAR_ERROR);
            }
            else {
                techFileNamePtr = util_strsav(util_logical_expand(valuePtr));
            }
        }
    }

    if (ReadBorderCellV1(&readData) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    *coreMasterPtrPtr = readData.coreMasterPtr;

    Mac_PutTechFileNameOfCellMaster(techFileNamePtr, 
        readData.coreMasterPtr);

    okayToStopFlag = TRUE;
    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    while (TRUE) {
        if (strcmp(keyWordPtr, "CellName") == 0) {
            bearStatus = ReadCellV1(&readData);
            if ((bearStatus == BEAR_OK) && (eofFlag == TRUE)) {
                break;
            }
            else if ((bearStatus == BEAR_OK) && (eofFlag == FALSE)) {
                continue;
            }
            else if (bearStatus == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
        }
        else {
            MISSING_KEYWORD_ERR("CellName");
        }
    }

    return (BEAR_OK);
}


static BearStatus
ReadBorderCellV1(readDataPtr)
ReadData *readDataPtr;
/*
 * Create the top level cell, i.e. the chip cell which contains 
 * a core cell and I/O pads. 
 */
{
    BearStatus      bearStatus;
    CellMaster      *chipMasterPtr;
    CellInst        *chipInstPtr, *coreInstPtr;
    lsList          inputBPinList = (lsList) NULL;
    InputBorderPin  *inputBPinPtr;
    Rect            chipRect;
    int             x, y;
    double          width, height;
    BearString      chipName, chipMasterName;
    BearString      coreName, coreMasterName;

    okayToStopFlag = FALSE;

    READ_LINE;
    LOOK_FOR_KEYWORD("BorderCellName");
    LOOK_FOR_VALUE("arguement");
    coreName = util_strsav(valuePtr);

    Mac_GenCellName(fileName, &chipName);

    chipMasterName = ALLOC(char, strlen(M_NAME_FORMAT) + strlen(chipName) + 1);
    (void) sprintf (chipMasterName, M_NAME_FORMAT, chipName);

    if (Mac_CreateCell(NIL(CellMaster), chipMasterName, chipName,
        &chipInstPtr) != BEAR_OK) {

        PrintInputErr("Could not create the chip cell.");

        FREE(chipName);
        FREE(chipMasterName);
        return (BEAR_ERROR);
    }
    FREE(chipName);
    FREE(chipMasterName);

    coreMasterName = ALLOC(char, strlen(M_NAME_FORMAT) + strlen(coreName) + 1);
    (void) sprintf (coreMasterName, M_NAME_FORMAT, coreName);

    chipMasterPtr = Mac_GetMasterOfCell(chipInstPtr);
    if (Mac_CreateCell(chipMasterPtr,
        coreMasterName, coreName, &coreInstPtr) != BEAR_OK) {

        PrintInputErr("Could not create the core cell.");
        FREE(coreName);
        FREE(coreMasterName);
        return (BEAR_ERROR);
    }
    FREE(coreName);
    FREE(coreMasterName);

    READ_LINE;
    LOOK_FOR_KEYWORD("BorderCellOrigin");
    LOOK_FOR_VALUE("x coordinate");
    (void) sscanf (valuePtr, "%d", &x);
    LOOK_FOR_VALUE("y coordinate");
    (void) sscanf (valuePtr, "%d", &y);
    readDataPtr->bdrOrigin.p_x = x;
    readDataPtr->bdrOrigin.p_y = y;

    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    while (strcmp(keyWordPtr, "BorderPin") == 0) {
        inputBPinPtr = ALLOC(InputBorderPin, 1);

        bearStatus = ReadBorderPinV1(inputBPinPtr);
        if (bearStatus == BEAR_OK) {
            if (inputBPinList == (lsList) NULL) {
                inputBPinList = lsCreate();
            }
            lsNewBegin(inputBPinList, (lsGeneric) inputBPinPtr, LS_NH);

            continue;
        }
        else if (bearStatus == BEAR_ERROR) {
            FREE(inputBPinPtr);
            if (inputBPinList != (lsList) NULL) {
                lsDestroy(inputBPinList, FreeInputBorderPin);
            }

            return (BEAR_ERROR);
        }
    }

    if (strcmp(keyWordPtr, "BorderCellShape") != 0) {
        MISSING_KEYWORD_ERR("BorderCellShape");
    }

    LOOK_FOR_VALUE("width");
    (void) sscanf (valuePtr, "%lf", &width);
    CHECK_WIDTH(width);
    if (width > INT_MAX) {
        (void) sprintf (bear_Msg, "Width of border cell > %d\n%s", 
            INT_MAX,
            "    You must scale your data downwards.\n");
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    LOOK_FOR_VALUE("height");
    (void) sscanf (valuePtr, "%lf", &height);
    CHECK_HEIGHT(height);
    if (height > INT_MAX) {
        (void) sprintf (bear_Msg, "Height of border cell > %d\n%s", 
            INT_MAX,
            "    You must scale your data downwards.\n");
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    readDataPtr->bdrWidth = width;
    readDataPtr->bdrHeight = height;

    chipRect.r_xbot = x;
    chipRect.r_ybot = y;
    chipRect.r_xtop = x + width;
    chipRect.r_ytop = y + height;

    if (FillInCellInst(chipInstPtr, chipRect) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (FillInCellInst(coreInstPtr, chipRect) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    readDataPtr->coreMasterPtr = Mac_GetMasterOfCell(coreInstPtr);
    
    Mac_PutFileNameOfCellMaster(util_strsav(fileName), 
        readDataPtr->coreMasterPtr);

    if (inputBPinList != (lsList) NULL) {
        if (StoreBorderPinsV1(coreInstPtr, inputBPinList, readDataPtr) !=
            BEAR_OK) {
            return (BEAR_ERROR);
        }
        lsDestroy(inputBPinList, FreeInputBorderPin);
    }

    okayToStopFlag = TRUE;

    return (BEAR_OK);
}


static BearStatus
ReadCellV1(readDataPtr)
ReadData *readDataPtr;
{
    BearStatus  bearStatus;
    CellInst    *cellInstPtr;
    CellMaster  *cellMasterPtr;
    InputPin    *inputPinPtr;
    Rect        cellRect;
    lsList      inputPinList = (lsList) NULL;
    int         x, y, w, h;
    char        *cellNamePtr, *cellMasterNamePtr;
    bool        rotFlag;

    okayToStopFlag = FALSE;

    LOOK_FOR_VALUE("argument"); /* cell name */
    cellNamePtr = valuePtr;

    cellMasterNamePtr = ALLOC(char,
        strlen(M_NAME_FORMAT) + strlen(cellNamePtr) + 1);
    (void) sprintf (cellMasterNamePtr, M_NAME_FORMAT, cellNamePtr);

    if (Mac_CreateCell(readDataPtr->coreMasterPtr,
        cellMasterNamePtr, cellNamePtr, &cellInstPtr) != BEAR_OK) {

        (void) sprintf (bear_Msg, "Could not create the cell named \"%s\".", 
            cellNamePtr);
        PrintInputErr(bear_Msg);
        FREE(cellMasterNamePtr);
        return (BEAR_ERROR);
    }
    FREE(cellMasterNamePtr);
    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);

    READ_LINE;
    LOOK_FOR_KEYWORD("CellOrigin");
    LOOK_FOR_VALUE("x coordinate");
    (void) sscanf (valuePtr, "%d", &x);
    LOOK_FOR_VALUE("y coordinate");
    (void) sscanf (valuePtr, "%d", &y);
    readDataPtr->cellOrigin.p_x = x;
    readDataPtr->cellOrigin.p_y = y;
    x += readDataPtr->bdrOrigin.p_x;
    y += readDataPtr->bdrOrigin.p_y;

    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    while (strcmp(keyWordPtr, "Pin") == 0) {
        inputPinPtr = ALLOC(InputPin, 1);

        bearStatus = ReadCellPinV1(inputPinPtr);
        if (bearStatus == BEAR_OK) {
            if (inputPinList == (lsList) NULL) {
                inputPinList = lsCreate();
            }
            lsNewBegin(inputPinList, (lsGeneric) inputPinPtr, LS_NH);

            continue;
        }
        else if (bearStatus == BEAR_ERROR) {
            FREE(inputPinPtr);
            if (inputPinList != (lsList) NULL) {
                lsDestroy(inputPinList, FreeInputPin);
            }

            return (BEAR_ERROR);
        }
    }

    if (strcmp(keyWordPtr, "InitCellShape") != 0) {
        MISSING_KEYWORD_ERR("InitCellShape");
    }

    LOOK_FOR_VALUE("width");
    (void) sscanf (valuePtr, "%d", &w);
    CHECK_WIDTH(w);

    LOOK_FOR_VALUE("height");
    (void) sscanf (valuePtr, "%d", &h);
    CHECK_HEIGHT(h);

    LOOK_FOR_VALUE("rotation flag");
    if (strcmp(valuePtr, "TRUE") == 0) {
        rotFlag = TRUE;
    }
    else if (strcmp(valuePtr, "FALSE") == 0) {
        rotFlag = FALSE;
    }
    else {
        (void) sprintf (bear_Msg, "unknown rotation flag on line %d.", lineNum);
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    cellRect.r_xbot = x;
    cellRect.r_ybot = y;
    cellRect.r_xtop = x + w;
    cellRect.r_ytop = y + h;

    if (FillInCellInst(cellInstPtr, cellRect) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (rotFlag == TRUE) {
        Shape *shapePtr;

        shapePtr = ALLOC(Shape, 1);
        shapePtr->horzDim = h;
        shapePtr->vertDim = w;
        lsNewEnd(cellMasterPtr->shapeList, (lsGeneric) shapePtr, LS_NH);
    }
    if (inputPinList != (lsList) NULL) {
        if (StoreCellPinsV1(cellInstPtr, inputPinList, readDataPtr) !=
            BEAR_OK) {
            return (BEAR_ERROR);
        }
        lsDestroy(inputPinList, FreeInputPin);
    }

    if (Mac_AssignShapeFlag(cellInstPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    okayToStopFlag = TRUE;
    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    while (strcmp(keyWordPtr, "CellShape") == 0) {
        bearStatus = ReadCellShapeV1(cellInstPtr);
        if ((bearStatus == BEAR_OK) && (eofFlag == TRUE)) {
            break;
        }
        else if ((bearStatus == BEAR_OK) && (eofFlag == FALSE)) {
            continue;
        }
        else if (bearStatus == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }
    okayToStopFlag = TRUE;

    if (Mac_AssignShapeFlag(cellInstPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}


static BearStatus
ReadBorderPinV1(inputBPinPtr)
InputBorderPin *inputBPinPtr;
{
    LOOK_FOR_VALUE("pin name");
    inputBPinPtr->pinName = util_strsav(valuePtr);

    LOOK_FOR_VALUE("net name");
    inputBPinPtr->netName = util_strsav(valuePtr);

    LOOK_FOR_VALUE("side value");
    if (strcmp(valuePtr, "N") == 0) {
        inputBPinPtr->sideValue = GEO_NORTH;
    }
    else if (strcmp(valuePtr, "E") == 0) {
        inputBPinPtr->sideValue = GEO_EAST;
    }
    else if (strcmp(valuePtr, "S") == 0) {
        inputBPinPtr->sideValue = GEO_SOUTH;
    }
    else if (strcmp(valuePtr, "W") == 0) {
        inputBPinPtr->sideValue = GEO_WEST;
    }
    else {
        (void) sprintf (bear_Msg, "unknown side value on line %d.", lineNum);
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    LOOK_FOR_VALUE("fraction");
    (void) sscanf (valuePtr, "%lf", &inputBPinPtr->fraction);
    if ((inputBPinPtr->fraction < 0.0) || (inputBPinPtr->fraction > 1.0)) {
        (void) sprintf (bear_Msg,
            "fraction on line %d is not between 0.0 and 1.0", lineNum);
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    return (BEAR_OK);
}


static BearStatus
ReadCellPinV1(inputPinPtr)
InputPin *inputPinPtr;
{
    char *xValuePtr, *yValuePtr;

    LOOK_FOR_VALUE("pin name");
    inputPinPtr->pinName = util_strsav(valuePtr);

    LOOK_FOR_VALUE("net name");
    inputPinPtr->netName = util_strsav(valuePtr);

    LOOK_FOR_VALUE("pin type");
    if (strcmp(valuePtr, "fixed") == 0) {
        inputPinPtr->pinType = FIXED;
    }
    else if (strcmp(valuePtr, "floating") == 0) {
        inputPinPtr->pinType = FLEXIBLE;
    }
    else {
        (void) sprintf (bear_Msg, "unknown pin type on line %d.", lineNum);
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    LOOK_FOR_VALUE("x coordinate");
    xValuePtr = valuePtr;
    LOOK_FOR_VALUE("y coordinate");
    yValuePtr = valuePtr;
    if ((strcmp(xValuePtr, BFP_UNDEF_STR) == 0) ||
        (strcmp(yValuePtr, BFP_UNDEF_STR) == 0)) {
        inputPinPtr->pinPosDefFlag = FALSE;
    }
    else {
        inputPinPtr->pinPosDefFlag = TRUE;
        (void) sscanf (xValuePtr, "%d", &inputPinPtr->relPos.p_x);
        (void) sscanf (yValuePtr, "%d", &inputPinPtr->relPos.p_y);
    }

    if ((inputPinPtr->pinType == FIXED) &&
        (inputPinPtr->pinPosDefFlag == FALSE)) {
        (void) sprintf (bear_Msg,
            "fixed pin has undefined coordinates on line %d.", lineNum);
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    return (BEAR_OK);
}


static BearStatus
ReadCellShapeV1(cellInstPtr)
CellInst *cellInstPtr;
{
    CellMaster  *cellMasterPtr;
    Shape       *shapePtr;
    int         w, h;
    bool        rotFlag;

    LOOK_FOR_VALUE("width"); 
    (void) sscanf (valuePtr, "%d", &w);
    CHECK_WIDTH(w);

    LOOK_FOR_VALUE("height");
    (void) sscanf (valuePtr, "%d", &h);
    CHECK_HEIGHT(h);

    LOOK_FOR_VALUE("rotation flag");
    if (strcmp(valuePtr, "TRUE") == 0) {
        rotFlag = TRUE;
    }
    else if (strcmp(valuePtr, "FALSE") == 0) {
        rotFlag = FALSE;
    }
    else {
        (void) sprintf (bear_Msg, "unknown rotation flag on line %d.", lineNum);
        PrintInputErr(bear_Msg);
        return (BEAR_ERROR);
    }

    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);

    shapePtr = ALLOC(Shape, 1);
    shapePtr->horzDim = w;
    shapePtr->vertDim = h;
    lsNewEnd(cellMasterPtr->shapeList, (lsGeneric) shapePtr, LS_NH);
    if (rotFlag == TRUE) {
        shapePtr = ALLOC(Shape, 1);
        shapePtr->horzDim = h;
        shapePtr->vertDim = w;
        lsNewEnd(cellMasterPtr->shapeList, (lsGeneric) shapePtr, LS_NH);
    }

    READ_LINE;
    keyWordPtr = GET_KEYWORD_STR;

    return (BEAR_OK);
}


static BearStatus
StoreBorderPinsV1(cellInstPtr, inputBPinList, readDataPtr)
CellInst    *cellInstPtr;
lsList      inputBPinList;
ReadData    *readDataPtr;
{
    InputBorderPin  *inputBPinPtr;
    lsGen           gen;
    PinInst         *pinInstPtr;
    Net             *netPtr;
    int             x, y;
    char            portMasterName[80], portInstName[80];
    char            pinMasterName[80], *pinInstName; 
    char            tmp1[80];

    lsForeachItem(inputBPinList, gen, inputBPinPtr) {
        (void) sprintf (portMasterName, "%d", Mac_GenerateId());
        (void) sprintf (portInstName,   "%d", Mac_GenerateId());

        (void) sprintf (pinMasterName,  "%d", Mac_GenerateId());

        if (strcmp(inputBPinPtr->pinName, BFP_UNDEF_STR) == 0) {
            (void) sprintf (tmp1, "PX%d", Mac_GenerateId());
            pinInstName = tmp1;
        }
        else {
            pinInstName = inputBPinPtr->pinName;
        }

        if (Mac_CreatePin(cellInstPtr, portMasterName, portInstName, 
            pinMasterName, pinInstName, &pinInstPtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        if (Mac_IsMember(NET_HT, inputBPinPtr->netName) == FALSE) {
            if (Mac_CreateNet(readDataPtr->coreMasterPtr,
                inputBPinPtr->netName, &netPtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            Mac_PutKindOfNet(SIGNAL_NET, netPtr);
        }
        else { 
            netPtr = (Net *) Mac_LookUp(NET_HT, inputBPinPtr->netName);
        }

        Mac_PutPinPosDefFlag(TRUE, pinInstPtr);

        switch (inputBPinPtr->sideValue) {
        case GEO_NORTH:
            x = inputBPinPtr->fraction * readDataPtr->bdrWidth +
                readDataPtr->bdrOrigin.p_x;
            y = readDataPtr->bdrOrigin.p_y + readDataPtr->bdrHeight;
            break;
        case GEO_EAST:
            x = readDataPtr->bdrOrigin.p_x + readDataPtr->bdrWidth;
            y = inputBPinPtr->fraction * readDataPtr->bdrHeight +
                readDataPtr->bdrOrigin.p_y;
            break;
        case GEO_SOUTH:
            x = inputBPinPtr->fraction * readDataPtr->bdrWidth +
                readDataPtr->bdrOrigin.p_x;
            y = readDataPtr->bdrOrigin.p_y;
            break;
        case GEO_WEST:
            x = readDataPtr->bdrOrigin.p_x;
            y = inputBPinPtr->fraction * readDataPtr->bdrHeight +
                readDataPtr->bdrOrigin.p_y;
            break;
        default:
            (void) sprintf (bear_Msg, "unknown side value");
            PrintInputErr(bear_Msg);
            return (BEAR_ERROR);
        }
        pinInstPtr->masterPtr->posUn.position.p_x = x;
        pinInstPtr->masterPtr->posUn.position.p_y = y;

        Mac_PutFlexOfPin(FIXED, pinInstPtr);

        if (Mac_InsertPortInstOfNet(netPtr, pinInstPtr->portPtr) ==
            BEAR_ERROR) {
            return (BEAR_ERROR);
        }
        if (Mac_InsertPinOfLineSeg(cellInstPtr->boundary, pinInstPtr) == 
            BEAR_ERROR) {
            return (BEAR_ERROR);
        }
        (void) Mac_ConvertToBorderPin(pinInstPtr, (ClientData) NULL);
    }

    return (BEAR_OK);
}


static BearStatus
StoreCellPinsV1(cellInstPtr, inputPinList, readDataPtr)
CellInst    *cellInstPtr;
lsList      inputPinList;
ReadData    *readDataPtr;
{
    Flexibility pinType;
    InputPin    *inputPinPtr;
    lsGen       gen;
    PinInst     *pinInstPtr;
    Net         *netPtr;
    int         x, y;
    char        portMasterName[80], portInstName[80];
    char        pinMasterName[80], *pinInstName; 
    char        tmp1[80];

    lsForeachItem(inputPinList, gen, inputPinPtr) {
        (void) sprintf (portMasterName, "%d", Mac_GenerateId());
        (void) sprintf (portInstName,   "%d", Mac_GenerateId());

        (void) sprintf (pinMasterName,  "%d", Mac_GenerateId());

        if (strcmp(inputPinPtr->pinName, BFP_UNDEF_STR) == 0) {
            (void) sprintf (tmp1, "PX%d", Mac_GenerateId());
            pinInstName = tmp1;
        }
        else {
            pinInstName = inputPinPtr->pinName;
        }

        if (Mac_CreatePin(cellInstPtr, portMasterName, portInstName, 
            pinMasterName, pinInstName, &pinInstPtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        if (Mac_IsMember(NET_HT, inputPinPtr->netName) == FALSE) {
            if (Mac_CreateNet(readDataPtr->coreMasterPtr,
                inputPinPtr->netName, &netPtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            Mac_PutKindOfNet(SIGNAL_NET, netPtr);
        }
        else { 
            netPtr = (Net *) Mac_LookUp(NET_HT, inputPinPtr->netName);
        }

        if (inputPinPtr->pinPosDefFlag == TRUE) {
            Mac_PutPinPosDefFlag(TRUE, pinInstPtr);

            pinInstPtr->masterPtr->posUn.position.p_x =
                readDataPtr->bdrOrigin.p_x + readDataPtr->cellOrigin.p_x +
                inputPinPtr->relPos.p_x;
            pinInstPtr->masterPtr->posUn.position.p_y =
                readDataPtr->bdrOrigin.p_y + readDataPtr->cellOrigin.p_y +
                inputPinPtr->relPos.p_y;
        }
        else {
            Mac_PutPinPosDefFlag(FALSE, pinInstPtr);

            pinInstPtr->masterPtr->posUn.position.p_x = 0;
            pinInstPtr->masterPtr->posUn.position.p_y = 0;
        }

        Mac_PutFlexOfPin(inputPinPtr->pinType, pinInstPtr);

        if (Mac_InsertPortInstOfNet(netPtr, pinInstPtr->portPtr) ==
            BEAR_ERROR) {
            return (BEAR_ERROR);
        }
        if (Mac_InsertPinOfLineSeg(cellInstPtr->boundary, pinInstPtr) == 
            BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }

    return (BEAR_OK);
}


static BearStatus
CreateDefaultChip(coreMasterPtrPtr)
CellMaster **coreMasterPtrPtr;
/*
 * Create a chip cell, and a core cell.
 */
{
    CellMaster  *chipMasterPtr;
    CellInst    *chipInstPtr, *coreInstPtr;
    Rect        chipRect;

    *coreMasterPtrPtr = NIL(CellMaster);

    if (Mac_CreateCell(NIL(CellMaster), "Biff", "Buffy",
        &chipInstPtr) != BEAR_OK) {

        PrintInputErr("Could not create the chip cell.");

        return (BEAR_ERROR);
    }

    chipRect.r_xbot = chipRect.r_ybot = 0;
    chipRect.r_xtop = chipRect.r_ytop = 1000;
    if (FillInCellInst(chipInstPtr, chipRect) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    chipMasterPtr = Mac_GetMasterOfCell(chipInstPtr);
    if (Mac_CreateCell(chipMasterPtr, "Jerry_Lewis", "Pee_Wee_Herman",
        &coreInstPtr) != BEAR_OK) {

        PrintInputErr("Could not create the core cell.");
        return (BEAR_ERROR);
    }

    if (FillInCellInst(coreInstPtr, chipRect) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    /*
     * Fill in the technology part of the MAC database with
     * default vaules.
     */
    mac_NumOfLayer = 3;
    Mac_PutMinWidthOfLay(20, METAL1_LAYER);
    Mac_PutMinWidthOfLay(20, METAL2_LAYER);
    Mac_PutMinSpaceOfLayLay(10, METAL1_LAYER, METAL1_LAYER);
    Mac_PutMinSpaceOfLayLay(10, METAL2_LAYER, METAL2_LAYER);
    Mac_PutSizeOfHole(25, M1M2_VIA);
    Mac_PutMinOverlapOfLayHole(5, METAL1_LAYER, M1M2_VIA);

    *coreMasterPtrPtr = Mac_GetMasterOfCell(coreInstPtr);

    Mac_PutFileNameOfCellMaster(util_strsav(fileName), *coreMasterPtrPtr);

    return (BEAR_OK);
}


static BearStatus
FillInCellInst(cellInstPtr, cellRect)
CellInst    *cellInstPtr;
Rect        cellRect;
{
    CellMaster      *cellMasterPtr;
    GenericCell     gcell;
    LinkedPoint     *pointList;
    LinkedLineSeg   *boundaryPtr, *lnLineSegPtr;
    LinkedRect      *rectList;
    Point           origin;
    Shape           *shapePtr;
    
    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;

    Mac_PutBBoxOfCell(&cellRect, cellInstPtr);

    origin.p_x = origin.p_y = 0;
    Mac_PutOriginOfCell(&origin, cellInstPtr);

    Mac_PutTransformOfGCell(&GeoIdentityTransform, &gcell);

    Mac_CreatePointListFromRect(&cellRect, &pointList);

    if (Mac_MakeLineSegList(pointList, &boundaryPtr) == BEAR_OK) {
        (void) Mac_AssignDirOfLineSegments(boundaryPtr);
        Mac_PutBoundaryOfCell(boundaryPtr, cellInstPtr);
    }
    else {
        PrintInputErr("Cannot make line segment list.");
        return (BEAR_ERROR);
    }

    if (Rs_HorzRectSlice(pointList, &rectList) == BEAR_OK) {
        Mac_PutHorzRectListOfGCell(rectList, &gcell);
    }
    else {
        PrintInputErr("Cannot make horizontal rectangle list.");
        return (BEAR_ERROR);
    }
    if (Rs_VertRectSlice(pointList, &rectList) == BEAR_OK) {
        Mac_PutVertRectListOfGCell(rectList, &gcell);
    }
    else {
        PrintInputErr("Cannot make vertical rectangle list.");
        return (BEAR_ERROR);
    }

    Mac_FreePointList(pointList);

    for (lnLineSegPtr = boundaryPtr;
        lnLineSegPtr != NIL(LinkedLineSeg);
        lnLineSegPtr = lnLineSegPtr->nextPtr) {

        lnLineSegPtr->lineSeg->pinList = NIL(PinInst);
    }

    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);
    cellMasterPtr->shapeFlag = MAC_UNDEF_SHAPE;
    cellMasterPtr->shapeList = lsCreate();
    shapePtr = ALLOC(Shape, 1);
    shapePtr->horzDim = cellRect.r_xtop - cellRect.r_xbot;
    shapePtr->vertDim = cellRect.r_ytop - cellRect.r_ybot;
    lsNewEnd(cellMasterPtr->shapeList, (lsGeneric) shapePtr, LS_NH);

    return (BEAR_OK);
}


