
/* cpaMain.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Bear interface to the cpa (channel pin arrangement) module.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/cpaInterface/cpaMain.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "magic.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "st.h"
#include "cpaInterface.h"


#ifdef USE_OCT_CODE /* { */
#include "vov.h"
#endif /* } USE_OCT_CODE */

/*
 * Later on once the cpa module is used by passing data structures
 * the TMP_CODE flags can be switched to DEBUG flags.
 */
#define TMP_CODE


#ifdef TMP_CODE
#define TMP_FILE_PREFIX     "cpa"
#endif TMP_CODE


typedef struct _LocalChannel LocalChannel;
struct _LocalChannel {
    CellMaster      *coreMasterPtr;
    PlaneData       *planeDataPtr;
    int             mps;            /* minimum pin spacing */
    int             errCount;
    int             chanCount;      /* channel count */
    int             modChanCount;   /* modified channel count */
    char            *cpaProgPtr;
    char            *tmpDirPtr;
    bool            warnFlag;

    Rect            absRect;
    SideDirection   sideDir;
    char            inFileName[1024], outFileName[1024];
    bool            horzFlag;
    bool            botFixedFlag, topFixedFlag;

    int             width, height, colWidth;
    char            *chanClassPtr;
    st_table        *netHashTable;
    lsList          botPinList, topPinList;
};

typedef struct _LocalNet LocalNet;
struct _LocalNet {
    Net         *macNetPtr;
    int         leftSide;
    int         rightSide;
    lsList      pinList;
};


/* ---------- Local Procedures ---------- */
static LocalNet     *MakeLocalNet();
static void         PrintWarn();
static BearStatus   ArrPinsBegin();
static BearStatus   ArrPinsEnd();

static BearStatus   DoBottleNeck();
static BearStatus   BldChanOfBnTile();
static BearStatus   AddPinOfBnTile();
static BearStatus   LookAtPPinOfBnTile();
static BearStatus   ExecCpaModule();

#ifdef TMP_CODE
static BearStatus   ReadCpaOutput();
static BearStatus   WriteCpaInput();
#endif TMP_CODE


/*
 * ----------------------------------------------------------------------------
 *
 * MakeLocalNet() --
 *
 * Allocates memory for the structure LocalNet, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to the structure LocalNet.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static LocalNet *
MakeLocalNet()
{
    LocalNet *locNetPtr;

    locNetPtr = ALLOC(LocalNet, 1);
    locNetPtr->macNetPtr = NIL(Net);
    locNetPtr->leftSide = 0;
    locNetPtr->rightSide = 0;
    locNetPtr->pinList = lsCreate();

    return (locNetPtr);
}


/*
 * ----------------------------------------------------------------------------
 *
 * PrintWarn() --
 *
 * Print the warning contained in the string 'warnMsgPtr' if the warning flag
 * in the structure LocalChannel is TRUE.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static void
PrintWarn(warnMsgPtr, locChanPtr)
char            *warnMsgPtr;
LocalChannel    *locChanPtr;
{
    if (locChanPtr->warnFlag == TRUE) {
        Bear_PrintWarn(warnMsgPtr);
    }
}


/*
 * ----------------------------------------------------------------------------
 *
 * ExecCpaModule() --
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
ExecCpaModule(locChanPtr)
LocalChannel *locChanPtr;
{
#define CMD_FORMAT "%s -in %s -out %s"

#ifdef TMP_CODE
{
    char *commandPtr;

    if (WriteCpaInput(locChanPtr) != BEAR_OK) {
        return BEAR_ERROR;
    }

    commandPtr = ALLOC(char, (strlen(CMD_FORMAT) + 
        strlen(locChanPtr->cpaProgPtr) + 
        strlen(locChanPtr->inFileName) + 
        strlen(locChanPtr->outFileName) + 1));
    (void) sprintf (commandPtr, CMD_FORMAT,
        locChanPtr->cpaProgPtr,
        locChanPtr->inFileName,
        locChanPtr->outFileName);
    if (util_csystem(commandPtr) != 0) {
        (void) sprintf (bear_ErrorMsg, "sys cmd : \"%s\" failed.", commandPtr);
        Bear_PrintErr(bear_ErrorMsg);
        return BEAR_ERROR;
    }
    FREE(commandPtr);

#ifdef USE_OCT_CODE /* { */
    (void) VOVdelete(VOV_UNIX_FILE, locChanPtr->inFileName);
#endif /* } USE_OCT_CODE */
    if (ReadCpaOutput(locChanPtr) != BEAR_OK) {
        return BEAR_ERROR;
    }
#ifdef USE_OCT_CODE /* { */
    (void) VOVdelete(VOV_UNIX_FILE, locChanPtr->outFileName);
#endif /* } USE_OCT_CODE */
}
#else
    /*
     * Should be done by passing data structures.
     */
#endif TMP_CODE

    return BEAR_OK;

#undef CMD_FORMAT
}


/*
 * ----------------------------------------------------------------------------
 *
 * ReadCpaOutput() --
 *
 * Read in the information generated by the cpa (channel pin arrangement)
 * module and then update the relative pin position of each pin instance
 * in the channel.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

#ifdef TMP_CODE
static BearStatus
ReadCpaOutput(locChanPtr)
LocalChannel *locChanPtr;
{
#define MAX_LINE_SIZE 1024

#define RETURN_ERR \
{ \
    (void) fclose (filePtr); \
    return BEAR_ERROR; \
}

    BearId      pinInstId;
    FILE        *filePtr = (FILE *) NULL;
    PinInst     *pinInstPtr;
    Point       absPos, relPos;
    Transform   invTrans;
    int         numOfParms, colNum;
    int         colWidth, halfColWidth;
    bool        eofFlag;
    char        lineBuf[MAX_LINE_SIZE];
    char        extraParmBuf[MAX_LINE_SIZE];

#ifdef USE_OCT_CODE /* { */
    (void) VOVinput(VOV_UNIX_FILE, locChanPtr->outFileName);
#endif /* } USE_OCT_CODE */

    if ((filePtr = fopen (locChanPtr->outFileName, "r")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open file named \"%s\" for reading.", 
            locChanPtr->outFileName);
        Bear_PrintErr(bear_ErrorMsg);

#ifdef USE_OCT_CODE /* { */
    (void) VOVdelete(VOV_UNIX_FILE, locChanPtr->outFileName);
#endif /* } USE_OCT_CODE */

        return BEAR_ERROR;
    }

    colWidth = locChanPtr->colWidth;
    halfColWidth = (colWidth / 2) + 0.5;

    eofFlag = FALSE;
    while (TRUE) {
        if (Bear_ReadLine(lineBuf, MAX_LINE_SIZE, filePtr, &eofFlag) == 
            BEAR_ERROR) {
            RETURN_ERR;
        }
        if (eofFlag == TRUE) {
            (void) fclose (filePtr);
            break;
        }

        numOfParms = sscanf (lineBuf, " X%d %d %s", 
            &pinInstId, &colNum, extraParmBuf);
        if (numOfParms < 2) {
    Bear_PrintErr("ReadCpaOutput() failed : bad output data (to few parms).");
            RETURN_ERR;
        }
        else if (numOfParms == 3) {
    Bear_PrintErr("ReadCpaOutput() failed : bad output data (to many parms).");
            RETURN_ERR;
        }

        if (Mac_IsMember(PIN_INSTANCE_HT, pinInstId)) {
            pinInstPtr = (PinInst *) Mac_LookUp(PIN_INSTANCE_HT, pinInstId);
        }
        else {
            (void) sprintf (bear_ErrorMsg,
                "ReadCpaOutput() failed : can't find pin with internal id %d.", 
                pinInstId);
            Bear_PrintErr(bear_ErrorMsg);
            RETURN_ERR;
        }

        if (locChanPtr->horzFlag) {
            absPos.p_y = locChanPtr->absRect.r_ytop -
                ((ABS(colNum) * colWidth) - halfColWidth);
            if ((absPos.p_y < locChanPtr->absRect.r_ybot) ||
                (absPos.p_y > locChanPtr->absRect.r_ytop)) {
           Bear_PrintErr("ReadCpaOutput() failed : new pin position is bad.");
                RETURN_ERR;
            }
            absPos.p_x = (colNum < 0) ?
                locChanPtr->absRect.r_xbot : locChanPtr->absRect.r_xtop;
        }
        else {
            absPos.p_x = locChanPtr->absRect.r_xbot +
                ((ABS(colNum) * colWidth) - halfColWidth);
            if ((absPos.p_x < locChanPtr->absRect.r_xbot) ||
                (absPos.p_x > locChanPtr->absRect.r_xtop)) {
           Bear_PrintErr("ReadCpaOutput() failed : new pin position is bad.");
                RETURN_ERR;
            }
            absPos.p_y = (colNum < 0) ?
                locChanPtr->absRect.r_ybot : locChanPtr->absRect.r_ytop;
        }

        GeoInvertTrans(&pinInstPtr->portPtr->cellInstPtr->transform, &invTrans);
        GeoTransPoint(&invTrans, &absPos, &relPos);
        Mac_PutRelPosOfPin(&relPos, pinInstPtr);
    }

    (void) fclose (filePtr);
    return BEAR_OK;

#undef MAX_LINE_SIZE
#undef RETURN_ERR
}
#endif TMP_CODE


/*
 * ----------------------------------------------------------------------------
 *
 * WriteCpaInput() --
 *
 * Write out all of the information about the channel contained within the 
 * structure LocalChannel in the cpa (channel pin arrangement) modules format.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

#ifdef TMP_CODE
static BearStatus
WriteCpaInput(locChanPtr)
LocalChannel *locChanPtr;
{
    FILE            *filePtr = (FILE *) NULL;
    lsGen           genPin;
    st_generator    *genLocNet;
    LocalNet        *locNetPtr;
    Net             *macNetPtr;
    PinInst         *pinInstPtr;
    char            *exitPtr;

#ifdef USE_OCT_CODE /* { */
    (void) VOVoutput(VOV_UNIX_FILE, locChanPtr->inFileName);
#endif /* } USE_OCT_CODE */

    if ((filePtr = fopen (locChanPtr->inFileName, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open file named \"%s\" for writing.", 
            locChanPtr->inFileName);
        Bear_PrintErr(bear_ErrorMsg);

#ifdef USE_OCT_CODE /* { */
        (void) VOVdelete(VOV_UNIX_FILE, locChanPtr->inFileName);
#endif /* } USE_OCT_CODE */

        return BEAR_ERROR;
    }

    (void) fprintf (filePtr, "\nchannel\n");
    (void) fprintf (filePtr, "    %-12s = %d\n", "height", locChanPtr->height);
    (void) fprintf (filePtr, "    %-12s = %d\n", "top", locChanPtr->width);
    (void) fprintf (filePtr, "    %-12s = %d\n", "bottom", locChanPtr->width);
    (void) fprintf (filePtr, "    %-12s = %d\n", "spacing",
        locChanPtr->colWidth);
    (void) fprintf (filePtr, "    %s\n", locChanPtr->chanClassPtr);
    (void) fprintf (filePtr, "\n");

    lsForeachItem(locChanPtr->botPinList, genPin, pinInstPtr) {
        (void) fprintf (filePtr, "pin X%d %2s\n",
            Mac_GetIdOfPinInst(pinInstPtr), "-1");
    }
    lsForeachItem(locChanPtr->topPinList, genPin, pinInstPtr) {
        (void) fprintf (filePtr, "pin X%d %2s\n",
            Mac_GetIdOfPinInst(pinInstPtr), "1");
    }
    (void) fprintf (filePtr, "\n");

    st_foreach_item(locChanPtr->netHashTable, genLocNet, 
        (char **) &macNetPtr, (char **) &locNetPtr) {

        (void) fprintf (filePtr, "net X%s %d ", 
            Mac_GetNameOfNet(macNetPtr),
            Mac_GetCriticalityOfNet(macNetPtr));

        if ((locNetPtr->leftSide == 0) && (locNetPtr->rightSide == 0)) {
            exitPtr = "none";
        }
        else if ((locNetPtr->leftSide > 0) && (locNetPtr->rightSide > 0)) {
            exitPtr = "both";
        }
        else if ((locNetPtr->leftSide > 0) && (locNetPtr->rightSide == 0)) {
            exitPtr = "left";
        }
        else if ((locNetPtr->leftSide == 0) && (locNetPtr->rightSide > 0)) {
            exitPtr = "right";
        }
        (void) fprintf (filePtr, "%s\n", exitPtr);

        lsForeachItem(locNetPtr->pinList, genPin, pinInstPtr) {
            (void) fprintf (filePtr, "    X%d\n",
                Mac_GetIdOfPinInst(pinInstPtr));
        }
        (void) fprintf (filePtr, "\n");
    }

    if (filePtr != (FILE *) NULL) {
        (void) fclose (filePtr);
    }
    return BEAR_OK;
}
#endif TMP_CODE


/*
 * ----------------------------------------------------------------------------
 *
 * LookAtPPinOfBnTile() --
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
LookAtPPinOfBnTile(ppinPtr, locChanPtr)
PseudoPin       *ppinPtr;
LocalChannel    *locChanPtr;
{
    LocalNet    *locNetPtr;
    Net         *macNetPtr;

    macNetPtr = Tp_GetNetOfPPin(ppinPtr);
    if (st_lookup(locChanPtr->netHashTable, (char *) macNetPtr,
        (char **) &locNetPtr) == 1) {

        if (locChanPtr->sideDir == RIGHT_SIDE) {
            locNetPtr->rightSide++;
        }
        else {
            locNetPtr->leftSide++;
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPinOfBnTile(pinInstPtr, locChanPtr)
PinInst         *pinInstPtr;
LocalChannel    *locChanPtr;
{
    Point       absPos;
    LocalNet    *locNetPtr;
    Net         *macNetPtr;

    Mac_GetAbsPosOfPin(pinInstPtr, &absPos);

    if (locChanPtr->horzFlag == TRUE) {
        if (! (((absPos.p_x == locChanPtr->absRect.r_xbot) ||
            (absPos.p_x == locChanPtr->absRect.r_xtop)) &&

            (absPos.p_y <= locChanPtr->absRect.r_ytop) &&
            (absPos.p_y >= locChanPtr->absRect.r_ybot))) {
            return (BEAR_CONTINUE_SEARCH);
        }
    }
    else {
        if (! (((absPos.p_y == locChanPtr->absRect.r_ybot) ||
            (absPos.p_y == locChanPtr->absRect.r_ytop)) &&

            (absPos.p_x <= locChanPtr->absRect.r_xtop) &&
            (absPos.p_x >= locChanPtr->absRect.r_xbot))) {
            return (BEAR_CONTINUE_SEARCH);
        }
    }

    if (locChanPtr->sideDir == BOTTOM_SIDE) {
        if (Mac_GetFlexOfPin(pinInstPtr) == FIXED) {
            locChanPtr->botFixedFlag = TRUE;
        }
        lsNewEnd(locChanPtr->botPinList, (lsGeneric) pinInstPtr, LS_NH);
    } 
    else {
        if (Mac_GetFlexOfPin(pinInstPtr) == FIXED) {
            locChanPtr->topFixedFlag = TRUE;
        }
        lsNewEnd(locChanPtr->topPinList, (lsGeneric) pinInstPtr, LS_NH);
    }

    macNetPtr = Mac_GetNetOfPin(pinInstPtr);
    if (st_lookup(locChanPtr->netHashTable, (char *) macNetPtr,
        (char **) &locNetPtr) == 0) {
        locNetPtr = MakeLocalNet();
        (void) st_insert(locChanPtr->netHashTable, (char *) macNetPtr,
            (char *) locNetPtr);
    }
    lsNewEnd(locNetPtr->pinList, (lsGeneric) pinInstPtr, LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * BldChanOfBnTile() --
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
BldChanOfBnTile(bnTilePtr, locChanPtr)
Tile            *bnTilePtr;
LocalChannel    *locChanPtr;
{
    st_generator    *genLocNet;
    CellInst        *cellInstPtr;
    LocalNet        *locNetPtr;
    Net             *macNetPtr;
    PinInst         *pinInstPtr;
    Plane           *planePtr;
    Tile            *tilePtr;
    Rect            absTiRect;
    int             botPinCount, topPinCount, pinCount;
    int             requiredWidth;
    int             tmp;
    bool            errorFlag;

    errorFlag = FALSE;

    locChanPtr->chanCount++;
    locChanPtr->netHashTable = st_init_table(st_ptrcmp, st_ptrhash);
    locChanPtr->botPinList = (lsList) NULL;
    locChanPtr->topPinList = (lsList) NULL;
#ifdef TMP_CODE
    (void) sprintf (locChanPtr->inFileName, "%s/%s_%d.in",
        locChanPtr->tmpDirPtr, TMP_FILE_PREFIX, Tp_GetIdOfTile(bnTilePtr));
    (void) sprintf (locChanPtr->outFileName, "%s/%s_%d.out",
        locChanPtr->tmpDirPtr, TMP_FILE_PREFIX, Tp_GetIdOfTile(bnTilePtr));
#endif TMP_CODE

    if (Tp_IsHorzTile(bnTilePtr) == TRUE) {
        locChanPtr->horzFlag = TRUE;
        Tp_HorzTileToRect(locChanPtr->planeDataPtr, bnTilePtr,
            &locChanPtr->absRect);
        planePtr = Tp_GetHorzPlane(locChanPtr->planeDataPtr);
        locChanPtr->width =
            locChanPtr->absRect.r_ytop - locChanPtr->absRect.r_ybot;
        locChanPtr->height =
            locChanPtr->absRect.r_xtop - locChanPtr->absRect.r_xbot;
    }
    else {
        locChanPtr->horzFlag = FALSE;
        Tp_VertTileToRect(locChanPtr->planeDataPtr, bnTilePtr,
            &locChanPtr->absRect);
        planePtr = Tp_GetVertPlane(locChanPtr->planeDataPtr);
        locChanPtr->width =
            locChanPtr->absRect.r_xtop - locChanPtr->absRect.r_xbot;
        locChanPtr->height =
            locChanPtr->absRect.r_ytop - locChanPtr->absRect.r_ybot;
    }

    /*
     * Add all pin instances on the bottom (or left) sides of
     * the bottleneck tile to the bottom pin list.
     */
    locChanPtr->sideDir = BOTTOM_SIDE;
    locChanPtr->botPinList = lsCreate();
    locChanPtr->botFixedFlag = FALSE;
    tilePtr = BL(bnTilePtr);
    if ((tilePtr == planePtr->pl_left) || (tilePtr == planePtr->pl_right)) {
        cellInstPtr = Mac_GetInstOfCoreMaster(locChanPtr->coreMasterPtr);
        (void) Mac_ItrPinOfCellInst(cellInstPtr,
            AddPinOfBnTile, (ClientData) locChanPtr);
    }
    else if (Tp_IsCellInstTile(tilePtr) == TRUE) {
        (void) Mac_ItrPinOfCellInst(Tp_GetCellInstOfTile(tilePtr),
            AddPinOfBnTile, (ClientData) locChanPtr);
    }
    else if (Tp_IsDCellTile(tilePtr) == FALSE) {
        Bear_PrintErr("BldChanOfBnTile() : unknown tile.");
        errorFlag = TRUE;
        goto fin;
    }

    /*
     * Add all pin instances on the top (or right) sides of
     * the bottleneck tile to the top pin list.
     */
    locChanPtr->sideDir = TOP_SIDE;
    locChanPtr->topPinList = lsCreate();
    locChanPtr->topFixedFlag = FALSE;
    tilePtr = TR(bnTilePtr);
    if ((tilePtr == planePtr->pl_left) || (tilePtr == planePtr->pl_right)) {
        cellInstPtr = Mac_GetInstOfCoreMaster(locChanPtr->coreMasterPtr);
        (void) Mac_ItrPinOfCellInst(cellInstPtr,
            AddPinOfBnTile, (ClientData) locChanPtr);
    }
    else if (Tp_IsCellInstTile(tilePtr) == TRUE) {
        (void) Mac_ItrPinOfCellInst(Tp_GetCellInstOfTile(tilePtr),
            AddPinOfBnTile, (ClientData) locChanPtr);
    }
    else if (Tp_IsDCellTile(tilePtr) == FALSE) {
        Bear_PrintErr("BldChanOfBnTile() : unknown tile.");
        errorFlag = TRUE;
        goto fin;
    }

    botPinCount = lsLength(locChanPtr->botPinList);
    topPinCount = lsLength(locChanPtr->topPinList);
    if ((botPinCount == 0) && (topPinCount == 0)) {
        (void) sprintf (bear_Msg,
            "ignored bn tile (id = %d) (no pins).",
            Tp_GetIdOfTile(bnTilePtr));
        PrintWarn(bear_Msg, locChanPtr);
        goto fin;
    }

    pinCount = MAX(botPinCount, topPinCount);
    tmp = locChanPtr->width / (1.5 * pinCount);
    locChanPtr->colWidth = MAX(locChanPtr->mps, tmp);

    requiredWidth = pinCount * locChanPtr->colWidth;
    if (requiredWidth > locChanPtr->width) {
        /*
         * The width of this bottleneck is to small to accommodate
         * all of the pins.
         */
        (void) sprintf (bear_Msg,
            "ignored bn tile (id = %d) (pins won't fit).",
            Tp_GetIdOfTile(bnTilePtr));
        PrintWarn(bear_Msg, locChanPtr);
        goto fin;
    }

    if ((locChanPtr->botFixedFlag == TRUE) &&
        (locChanPtr->topFixedFlag == TRUE)) {
        (void) sprintf (bear_Msg,
            "ignored bn tile (id = %d) (all pins fixed).",
            Tp_GetIdOfTile(bnTilePtr));
        PrintWarn(bear_Msg, locChanPtr);
        goto fin;
    }
    else if ((locChanPtr->botFixedFlag == TRUE) &&
        (locChanPtr->topFixedFlag == FALSE)) {
        locChanPtr->chanClassPtr = "bottom_fixed";

        /*
         * (gyorgy : 11 oct 90)
         * Currently the cpa module cannot handle this case.
         */
        (void) sprintf (bear_Msg,
            "ignored bn tile (id = %d) (%s pins fixed).",
            Tp_GetIdOfTile(bnTilePtr),
            ((Tp_IsHorzTile(bnTilePtr) == TRUE) ? "left" : "bottom"));
        PrintWarn(bear_Msg, locChanPtr);
        goto fin;
    }
    else if ((locChanPtr->botFixedFlag == FALSE) &&
        (locChanPtr->topFixedFlag == TRUE)) {
        locChanPtr->chanClassPtr = "top_fixed";

        /*
         * (gyorgy : 11 oct 90)
         * Currently the cpa module cannot handle this case.
         */
        (void) sprintf (bear_Msg,
            "ignored bn tile (id = %d) (%s pins fixed).",
            Tp_GetIdOfTile(bnTilePtr),
            ((Tp_IsHorzTile(bnTilePtr) == TRUE) ? "right" : "top"));
        PrintWarn(bear_Msg, locChanPtr);
        goto fin;
    }
    else if ((locChanPtr->botFixedFlag == FALSE) &&
        (locChanPtr->topFixedFlag == FALSE)) {
        locChanPtr->chanClassPtr = "no_fixed";
    }

    locChanPtr->sideDir = LEFT_SIDE;
    (void) Tp_ItrPPinsOfSide(bnTilePtr,
        (locChanPtr->horzFlag == TRUE) ? TOP_RIGHT_SIDE : BOTTOM_LEFT_SIDE,
        LookAtPPinOfBnTile, (ClientData) locChanPtr);
    locChanPtr->sideDir = RIGHT_SIDE;
    (void) Tp_ItrPPinsOfSide(bnTilePtr,
        (locChanPtr->horzFlag == TRUE) ? BOTTOM_LEFT_SIDE : TOP_RIGHT_SIDE,
        LookAtPPinOfBnTile, (ClientData) locChanPtr);

    if (ExecCpaModule(locChanPtr) != BEAR_OK) {
        errorFlag = TRUE;
        goto fin;
    }
    else {
        locChanPtr->modChanCount++;
    }

fin:
    if (locChanPtr->netHashTable != (st_table *) NULL) {
        st_foreach_item(locChanPtr->netHashTable, genLocNet, 
            (char **) &macNetPtr, (char **) &locNetPtr) {
            lsDestroy(locNetPtr->pinList, (void (*)()) NULL);
            FREE(locNetPtr);
        }
        st_free_table(locChanPtr->netHashTable);
    }
    if (locChanPtr->botPinList != (lsList) NULL) {
        lsDestroy(locChanPtr->botPinList, (void (*)()) NULL);
    }
    if (locChanPtr->topPinList != (lsList) NULL) {
        lsDestroy(locChanPtr->topPinList, (void (*)()) NULL);
    }

    if (errorFlag == TRUE) {
        locChanPtr->errCount++;

        (void) sprintf (bear_ErrorMsg,
            "Failed to arrange pins for bn tile (id = %d).",
            Tp_GetIdOfTile(bnTilePtr));
        Bear_PrintErr(bear_ErrorMsg);
    }

    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * DoBottleNeck() --
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
DoBottleNeck(bnPtr, locChanPtr)
BottleNeck      *bnPtr;
LocalChannel    *locChanPtr;
{
    if (bnPtr->unType == TP_BN_TILE) {
        (void) BldChanOfBnTile(bnPtr->unValue.bnTilePtr, locChanPtr);
    }

    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * ArrPinsBegin() --
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
ArrPinsBegin(coreMasterPtr, optDataPtr, locChanPtr)
CellMaster      *coreMasterPtr;         /* IN                       */
Cpa_ArrPinData  *optDataPtr;            /* IN       (optional data) */
LocalChannel    *locChanPtr;            /* OUT                      */
{
#define CMD_FORMAT "mkdir %s"

    char *stringPtr;

    (void) fprintf (bearout, "\nChannel Pin Arrangement\n");
    (void) fprintf (bearout,   "=======================\n\n");
    (void) fflush (bearout);

    if (Mac_IsGlobalRoutingDone(coreMasterPtr) == FALSE) {
        Bear_PrintErr("ArrPinsBegin() : need to complete global routing.");
        return BEAR_ERROR;
    }

    locChanPtr->coreMasterPtr = coreMasterPtr;
    locChanPtr->planeDataPtr = Mac_GetCellPlaneOfCell(coreMasterPtr);
    locChanPtr->mps = Mac_CalcMinPinSpacing();
    locChanPtr->errCount = 0;
    locChanPtr->chanCount = 0;
    locChanPtr->modChanCount = 0;
#ifdef TMP_CODE

#ifdef USE_OCT_CODE /* { */
    stringPtr = util_logical_expand("$OCTTOOLS/bin/koalaBear");
#else /*  } else { USE_OCT_CODE */
    stringPtr = util_logical_expand("$BEAR/bin/koalaBear");
#endif /* } USE_OCT_CODE */

    locChanPtr->cpaProgPtr = util_strsav(stringPtr);

    locChanPtr->tmpDirPtr = tempnam(bear_tmpDir, TMP_FILE_PREFIX);
#endif TMP_CODE
    locChanPtr->warnFlag = FALSE;

    if (optDataPtr != NIL(Cpa_ArrPinData)) {
#ifdef TMP_CODE
        if (optDataPtr->cpaProgPtr != NIL(char)) {
            FREE(locChanPtr->cpaProgPtr);
            stringPtr = util_tilde_expand(optDataPtr->cpaProgPtr);
            locChanPtr->cpaProgPtr = util_strsav(stringPtr);
        }
        if (optDataPtr->tmpDirPtr != NIL(char)) {
            FREE(locChanPtr->tmpDirPtr);
            stringPtr = util_tilde_expand(optDataPtr->tmpDirPtr);
            locChanPtr->tmpDirPtr = tempnam(stringPtr, TMP_FILE_PREFIX);
        }
#endif TMP_CODE
        locChanPtr->warnFlag = optDataPtr->warnFlag;
    }

#ifdef TMP_CODE
{
    char *commandPtr;

    /*
     * Use the unix mkdir command to create the temporary directory.
     */
    commandPtr = ALLOC(char,
        (strlen(CMD_FORMAT) + strlen(locChanPtr->tmpDirPtr) + 1));
    (void) sprintf (commandPtr, CMD_FORMAT, locChanPtr->tmpDirPtr);
    if (util_csystem(commandPtr) != 0) {
        (void) sprintf (bear_ErrorMsg, "sys cmd : \"%s\" failed.", commandPtr);
        Bear_PrintErr(bear_ErrorMsg);
        return BEAR_ERROR;
    }
    FREE(commandPtr);
}
#endif TMP_CODE

    return BEAR_OK;

#undef CMD_FORMAT
}


/*
 * ----------------------------------------------------------------------------
 *
 * ArrPinsEnd() --
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
ArrPinsEnd(optDataPtr, locChanPtr)
Cpa_ArrPinData  *optDataPtr;            /* IN       (optional data) */
LocalChannel    *locChanPtr;            /* IN / OUT                 */
{
#define CMD_FORMAT "rm -rf %s"

    bool pinSpacingFlag;

#ifdef TMP_CODE
    if (locChanPtr->cpaProgPtr != NIL(char)) {
        FREE(locChanPtr->cpaProgPtr);
        locChanPtr->cpaProgPtr = NIL(char);
    }

    if (locChanPtr->tmpDirPtr != NIL(char)) {
        if ((optDataPtr == NIL(Cpa_ArrPinData)) ||

            ((optDataPtr != NIL(Cpa_ArrPinData)) &&
            (optDataPtr->saveFlag == FALSE))) {
            char *commandPtr;

            /*
             * Use the unix rm command to remove the temporary directory.
             */
            commandPtr = ALLOC(char,
                (strlen(CMD_FORMAT) + strlen(locChanPtr->tmpDirPtr) + 1));
            (void) sprintf (commandPtr, CMD_FORMAT, locChanPtr->tmpDirPtr);
            if (util_csystem(commandPtr) != 0) {
                (void) sprintf (bear_ErrorMsg,
                    "sys cmd : \"%s\" failed.", commandPtr);
                Bear_PrintErr(bear_ErrorMsg);
                return BEAR_ERROR;
            }
            FREE(commandPtr);
        }
        else {
            (void) sprintf (bear_Msg, "Cpa data saved in \"%s\".\n",
                locChanPtr->tmpDirPtr);
            Bear_Print(bear_Msg);
        }

        FREE(locChanPtr->tmpDirPtr);
        locChanPtr->tmpDirPtr = NIL(char);
    }
#endif TMP_CODE

    if (Mac_CheckPins(locChanPtr->coreMasterPtr, &pinSpacingFlag) ==
        BEAR_ERROR) {
        Bear_PrintErr("ArrPinsEnd() failed : Bad pin data.");
        return BEAR_ERROR;
    }
    if (pinSpacingFlag == FALSE) {
        Bear_PrintErr("ArrPinsEnd() failed : some pins are to close.");
        return BEAR_ERROR;
    }

    if (locChanPtr->errCount != 0) {
        return BEAR_ERROR;
    }

    (void) fprintf (bearout,
        "%d channel(s) modified out of a total of %d channel(s).\n",
        locChanPtr->modChanCount, locChanPtr->chanCount);

    if (bear_BlackBoxFlag == FALSE) {
        (void) fprintf (bearout, "\nChannel pin arrangement completed.\n");
        (void) fflush (bearout);
    }

    return BEAR_OK;

#undef CMD_FORMAT
}



/*
 * ----------------------------------------------------------------------------
 *
 * Cpa_ArrPinsForOneBnTile() --
 *
 * This procedure uses the Channel Pin Arrangement module to rearrange
 * the pin instances within one bottleneck tile.
 *
 * The optDataPtr parameter may be NULL.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The positions of some of the pin instances are altered.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Cpa_ArrPinsForOneBnTile(coreMasterPtr, bnTilePtr, optDataPtr)
CellMaster      *coreMasterPtr;         /* IN / OUT                 */
Tile            *bnTilePtr;             /* IN                       */
Cpa_ArrPinData  *optDataPtr;            /* IN       (optional data) */
{
    LocalChannel locChan;

    if (ArrPinsBegin(coreMasterPtr, optDataPtr, &locChan) != BEAR_OK) {
        return BEAR_ERROR;
    }

    (void) BldChanOfBnTile(bnTilePtr, &locChan);

    if (ArrPinsEnd(optDataPtr, &locChan) != BEAR_OK) {
        return BEAR_ERROR;
    }

    return BEAR_OK;
}



/*
 * ----------------------------------------------------------------------------
 *
 * Cpa_ArrPinsForAllBnTiles() --
 *
 * This procedure uses the Channel Pin Arrangement module to rearrange
 * the pin instances within all the bottleneck tiles on both tile planes.
 *
 * The optDataPtr parameter may be NULL.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The positions of some of the pin instances are altered.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Cpa_ArrPinsForAllBnTiles(coreMasterPtr, optDataPtr)
CellMaster      *coreMasterPtr;         /* IN / OUT                */
Cpa_ArrPinData  *optDataPtr;            /* IN       (optional data) */
{
    LocalChannel locChan;

    if (ArrPinsBegin(coreMasterPtr, optDataPtr, &locChan) != BEAR_OK) {
        return BEAR_ERROR;
    }

    (void) Fpg_ItrBottleNeckOfFpg(Tp_GetFpg(locChan.planeDataPtr),
        DoBottleNeck, (ClientData) &locChan);

    if (ArrPinsEnd(optDataPtr, &locChan) != BEAR_OK) {
        return BEAR_ERROR;
    }

    return BEAR_OK;
}


