
/* ezGetLine.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezGetLine.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"

#define XK_MISCELLANY
#include <X11/keysymdef.h>


#define BACKSPACE       '\010'
#define BELL            '\07'        /* (control-g) ring the bell !!         */
#define CARRIAGE_RETURN '
'
#define CONTROL_L       '\014'       /* Redraw the window the mouse is in.   */
#define CONTROL_U       '\025'       /* Erase to the beginning of the line.  */
#define CONTROL_W       '\027'       /* Erase the last word on the line.     */
#define DASH            '-'
#define DELETE          '\177'
#define EQUALS          '='
#define NULL_CHAR       '\0'
#define PERIOD          '.'
#define PLUS            '+'
#define SLASH           '/'
#define SPACE           ' '
#define EZ_TAB          '\t'
#define UNDER_SCORE     '_'

#define TEXT_SIZE 1024

static char ezStringBuffer[TEXT_SIZE];
static bool ezPromptFlag;
static int  ezI;


bool 
EzParseBuf(text, textLen)
char    text[];
int     textLen;
{
    static char outtext[TEXT_SIZE];

    int         j, k;
    char        c;
    bool        eraseToSpace;

    if (textLen > TEXT_SIZE) {
        Ez_RingBell();
        return FALSE;
    }
    j = 0;
    bzero(outtext, sizeof (char) * TEXT_SIZE);
    for (k = 0; k < textLen; k++) {
        c = text[k];
        if (c == BACKSPACE || c == DELETE) {
            if ((ezI > 0) && ((j + 2) <= TEXT_SIZE)) {
            ezStringBuffer[--ezI] = NULL_CHAR;
            outtext[j++] = BACKSPACE;

            }
        } 
        else if (c == BELL) {
            Ez_RingBell();
        } 
        else if (c == CONTROL_U) {
            /*
             * Erase to the beginning of the line.
             */
            while ((ezI > 0) && ((j + 2) <= TEXT_SIZE)) {
                ezStringBuffer[--ezI] = NULL_CHAR;
                outtext[j++] = BACKSPACE;
            }
            
        } 
        else if (c == CONTROL_L) {
            return TRUE;
        } 
        else if (c == CONTROL_W) {
            /*
             * Erase the last word on the line.
             */
            if (ezI > 0) {
                eraseToSpace = TRUE;
                if (isspace(ezStringBuffer[ezI - 1])) {
                    /*
                     * If the last character typed was a space then
                     * first erase to the next non-space character (skip
                     * over white space) and then erase to the
                     * next space character (skip over text).
                     */
                    eraseToSpace = FALSE;
                }
                while ((ezI > 0) && ((j + 2) <= TEXT_SIZE)) {
                    ezI--;
                    if (eraseToSpace) {
                        if (ezI == 0) {
                            ezStringBuffer[0] = NULL_CHAR;
                            outtext[j++] = BACKSPACE;
                            break;
                        } 
                        else if (isspace(ezStringBuffer[ezI])) {
                            ezI++;
                            break;
                        }
                    } 
                    else {
                        if (isspace(ezStringBuffer[ezI]) == FALSE) {
                            eraseToSpace = TRUE;
                        }
                    }
                    ezStringBuffer[ezI] = NULL_CHAR;
                    outtext[j++] = BACKSPACE;
                }
            }
        } 
        else if (c == ezEofChar || c == ezIntrChar || c == ezQuitChar) {
            ezStringBuffer[ezI++] = EOF;
            ezStringBuffer[ezI] = NULL_CHAR;
            ezPromptFlag = FALSE;
        } 
        else if (isalpha(c) || isdigit(c) || isprint(c)) {
            if (((ezI + 2) <= TEXT_SIZE) && ((ezI + 2) <= TEXT_SIZE)) {
                ezStringBuffer[ezI++] = c;
                outtext[j++] = c;
            } 
            else {
                Ez_RingBell();
            }
        } 
        else if (c == CARRIAGE_RETURN) {
            if (((ezI + 2) <= TEXT_SIZE) && ((j + 2) <= TEXT_SIZE)) {
                ezStringBuffer[ezI++] = '\n';
                outtext[j++] = '\n';
                ezStringBuffer[ezI] = outtext[j] = NULL_CHAR;
                ezPromptFlag = FALSE;
            } 
            else {
                Ez_RingBell();
            }
        } 
        else if ((ezI > 0) && (c == EZ_TAB || c == SPACE)) {
            if (((ezI + 2) <= TEXT_SIZE) && ((j + 2) <= TEXT_SIZE)) {
                ezStringBuffer[ezI++] = SPACE;
                outtext[j++] = SPACE;
            } 
            else {
                Ez_RingBell();
            }
        }
        else {
            return FALSE;
        }
    }
    Ez_Print(EZ_BOLD_FONT, outtext);

    return FALSE;
}



ezStatus
Ez_GetLine(winPtr, screenPtr, buffer, bufferSize, keepGoingPtr)
Window  *winPtr;
Point   *screenPtr;
char    *buffer;
int     bufferSize;
bool    *keepGoingPtr;
/*
 * This procedure may be stopped by setting '*keepGoingPtr' to FALSE.
 *
 * Note that if the user types his "interrupt", "end-of-file", or "quit"
 * characters 'buffer' will contain EOF.
 */
{  
    static char xtext[TEXT_SIZE];

    KeySym      keysym;
    WindowData  *winDataPtr;
    XEvent      report;
    XAnyEvent   *anyEvent = (XAnyEvent *) &report;
    XKeyEvent   *keyEvent = (XKeyEvent *) &report;
    int         length;

    ezPromptFlag = TRUE;  /* Set to false when carriage return is
                           * detected in EzConsInput() 
                           */

    ezI = 0;
    bzero(xtext, sizeof (char) * TEXT_SIZE);
    while (*keepGoingPtr && ezPromptFlag) {
        XNextEvent(ez_Display, &report);
        Ez_XEventHandler(&report);

        if (anyEvent->type == KeyPress) {
            if (anyEvent->window != ezConsoleWin) {
                length = XLookupString(keyEvent, xtext, TEXT_SIZE, &keysym,
                    (XComposeStatus *) NULL);
                if ((keysym == XK_Shift_L) || (keysym == XK_Shift_R)) {
                    continue;
                }
                if (EzParseBuf(xtext, length)) {
                    /*
                     * Redraw the window the mouse is in.
                     */
                    if (((winDataPtr = EzLookUpWin(anyEvent->window)) != 
                        NIL(WindowData)) &&
                         
                        (winDataPtr->wd_redrawProc != (void (*)()) NULL)) {

                        (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
                    }
                }
            }
        }
    }

    strncpy(buffer, ezStringBuffer, bufferSize);
    if ((strlen (ezStringBuffer) + 1) > bufferSize) {
        return EZ_ERROR;
    }

    Ez_GetPointerXY(anyEvent->window, &screenPtr->p_x, &screenPtr->p_y);
    *winPtr = anyEvent->window;

    return EZ_OK;
}


