
/* ezMain.c --
 *
 *     George Carvalho, Esq.
 *     Ben Chen
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1991 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezMain.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include <sys/ioctl.h>
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"
#include <X11/Shell.h>


/* ---------- Internal global variables for the ezPlot module ---------- */

GC                   ezFillGC;
GC                   ezTextGC;
GC                   ezDrawGC;

FILE   *ez_LogFilePtr;
Pixmap ezSmallDotBitmap;
Cursor ezNWArrowCur;        /* north west pointing arrow cursor */
Cursor ezArrowCrossCur, ezHorzArrowsCur, ezVertArrowsCur;
Cursor ezCrossCur;
Cursor ezCircleOutlineCur, ezSquareOutlineCur;
Cursor ezStopWatchCur;
Display *ez_Display;
XFontStruct *ezPlanePtr, *tb_fontInfoPtr;
XFontStruct *ezConFontPtr;
Pixmap ezGrayBdrPixmap;
Window ezConsoleWin;
Widget ezConsoleWidget, ezTopWidget;
WindowData *ezWinDataHead;          /* The head of the list that contains
                                     * all the windows.
                                     */
st_table *ezWinHashTable;
unsigned int ez_Options;

unsigned int ezWinWidth;            /* starting width of the windows         */
unsigned int ezWinHeight;           /* starting height of the windows        */
unsigned int ezBorderWidth;         /* border width of the windows           */

#ifdef USE_OLD_X_CODE
Vertex  ezVertex[MAX_NUM_OF_VERTICES];
int     ezVertexCount;
#endif /* USE_OLD_X_CODE */

char ezEofChar;                     /* users end-of-file character           */
char ezIntrChar;                    /* users interrupt character             */
char ezQuitChar;                    /* users quit character                  */
char *ezConGeometryPtr;
char *ezProgNamePtr;
char *ezAppNamePtr;
bool ezDebug;                       /* When this is TRUE the optional 
                                     * debbuging messages will be printed.
                                     */
bool ezPsPlotFlag;                  /* When TRUE, output of certain ez-plot
                                     * drawing commands will be written to a
                                     * PostScript file.
                                     */
FILE *ezOutFile;
FILE *ezErrFile;
FILE *ezPsFile;
char ezErrorMsg[EZ_MSG_SIZE];


/* ---------- External global variables for the ezPlot module ---------- */

Pixel
    ez_WhitePixel,
    ez_BlackPixel,

    ez_BorderColor,                 /* border color of the windows           */
    ez_BackColor,                   /* background color of the windows       */
    ez_MouseColor,                  /* mouse color                           */
    ez_TextColor,                   /* text color                            */
    ez_HighLightColor,              /* high light color                      */

    ez_ConBackColor,                /* background color for the console      */
    ez_ConBorderColor,              /* border color for the console          */
    ez_ConTextColor;              /* normal text color for the console     */

Pixmap
    ez_CargoNetBitmap,
    ez_LeftSlashBitmap, 
    ez_RightSlashBitmap,
    ez_BackPixmap;                  /* background pixmap for the windows     */

XtAppContext ezAppContext;


/* ---------- Local Procedures ---------- */
static void LocalXError();
static void LocalXtError();


static void 
LocalXError(theDisp, theError)
Display     *theDisp;       /* Effected display  */
XErrorEvent *theError;      /* Error description */
{
    XGetErrorText(theDisp, theError->error_code, ezErrorMsg, EZ_MSG_SIZE);
    EzPrintErr(ezErrorMsg);

    exit(1);
}


static void 
LocalXtError(msg)
char *msg;
{
    (void) sprintf (ezErrorMsg, 
        "A Fatal Xt Toolkit Error has occurred:\n  %s\n", msg);
    EzPrintErr(ezErrorMsg);

    exit(1);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_Begin() --
 *
 *   This procedure MUST be called before any of the others in the
 * ezPlot module.  It gathers all of the default X data, and then
 * initializes most of the ezPlot modules global variables.
 *
 * Results:
 *   EZ_OK is returned if everything is spiffy, EZ_ERROR otherwise.
 *
 * Side effects:
 *   Initializes most of the ezPlot modules global variables.
 *
 * ----------------------------------------------------------------------------
 */

ezStatus
Ez_Begin(appNamePtr, topWidget, argc, argv)
char    *appNamePtr;                /* The name of the application.  It is 
                                     * used to find the X defaults. 
                                     */
Widget  *topWidget;
int     argc;
char    **argv;
{
    static Cardinal xrmNumOpts = 2;
    static XrmOptionDescRec xrmOpts[] = {
        {"-display",    ".display", XrmoptionSepArg, (caddr_t) NULL},
        {"-d",          ".display", XrmoptionSepArg, (caddr_t) NULL}
    };

    Widget          shell;
    struct tchars   tty;
    char            *tmpStringPtr;

    ezOutFile = stdout;
    ezErrFile = stderr;
    ezPsFile = NIL(FILE);

    ezDebug = FALSE;
    ezPsPlotFlag = FALSE;

    ezConGeometryPtr = NIL(char);

    if (ioctl(fileno (stdin), TIOCGETC, &tty) < 0) {
        perror(ezProgNamePtr);
        return EZ_ERROR;
    }
    else {
        ezEofChar =  tty.t_eofc;
        ezIntrChar = tty.t_intrc;
        ezQuitChar = tty.t_quitc;
    }

    ezAppNamePtr = util_strsav(appNamePtr);

    if (argv != (char **) NULL) {
        if ((tmpStringPtr = strrchr(argv[0], '/')) == NIL(char)) {
            tmpStringPtr = argv[0];
        } 
        else {
            tmpStringPtr++;
        }
        ezProgNamePtr = util_strsav(tmpStringPtr);
    }
    else {
        ezProgNamePtr = util_strsav(ezAppNamePtr);
    }

    XtSetErrorHandler(LocalXtError);
/*
 *  XtToolkitInitialize();
 */
    ezTopWidget = XtInitialize(ezProgNamePtr, ezAppNamePtr, 
        xrmOpts, xrmNumOpts, &argc, argv);
    ez_Display = XtDisplay(ezTopWidget);
    ezAppContext = XtWidgetToApplicationContext(ezTopWidget);

/*
 *  shell = XtAppCreateShell(ezProgNamePtr, ezAppNamePtr,
 *      applicationShellWidgetClass, ez_Display, (Arg *) 0, 0);
 */

    *topWidget = ezTopWidget;

    /* XSynchronize(ez_Display, 1); */
    XSetErrorHandler(LocalXError);

    ez_Options = 0;
           
    if (EzGetXDefaults() != EZ_OK) {
        return EZ_ERROR;
    }
    EzSetUpCursors();
    EzSetUpTilesAndBitmaps();
    EzSetUpGCs();
    ezWinHashTable = st_init_table(st_ptrcmp, st_ptrhash);

#ifdef USE_OLD_X_CODE
    bzero((char *) ezVertex, sizeof (*ezVertex) * MAX_NUM_OF_VERTICES);
    ezVertex[0].flags = VertexDontDraw; 
#endif /* USE_OLD_X_CODE */

    ezConsoleWin = (Window) NULL;
    ez_LogFilePtr = NIL(FILE);
    ezWinDataHead = NIL(WindowData);

    XFlush(ez_Display);

    return EZ_OK;
}



void
Ez_XEnd()
{
    /* Discard all the events on the input queue. */
    XSync(ez_Display, 1); 
    XCloseDisplay(ez_Display);

    st_free_table(ezWinHashTable);
}


bool
Ez_FindOverlap(rect1Ptr, rect2Ptr, overlapPtr)
register Rect *rect1Ptr, *rect2Ptr, *overlapPtr;
{
    if (GEO_OVERLAP(rect1Ptr, rect2Ptr)) {
        overlapPtr->r_xbot = MAX(rect1Ptr->r_xbot, rect2Ptr->r_xbot);
        overlapPtr->r_xtop = MIN(rect1Ptr->r_xtop, rect2Ptr->r_xtop);
        overlapPtr->r_ybot = MAX(rect1Ptr->r_ybot, rect2Ptr->r_ybot);
        overlapPtr->r_ytop = MIN(rect1Ptr->r_ytop, rect2Ptr->r_ytop);
        return (TRUE);
    } 
    else {
        return (FALSE);
    }
    /* NOTREACHED */
}


void
Ez_Print(whichFont, stringPtr)                            
int     whichFont;  /* EZ_BOLD_FONT | EZ_ITALIC_FONT | EZ_NORMAL_FONT */
char    *stringPtr;
{                                           
    if (ez_LogFilePtr != NIL(FILE)) {
        if (whichFont == EZ_ITALIC_FONT) {
            (void) fprintf (ez_LogFilePtr, "Error : %s", stringPtr);
        }
        else {
            (void) fprintf (ez_LogFilePtr, "%s", stringPtr);
        }
        fflush (ez_LogFilePtr);
    }

    if (ezConsoleWin == (Window) NULL) {
        if (whichFont == EZ_ITALIC_FONT) {
            (void) fprintf (stdout, "Error : %s", stringPtr);
        }
        else {
            (void) fprintf (stdout, "%s", stringPtr);
        }
        fflush (stdout);
    }
    else {
        switch (whichFont) {
        case EZ_ITALIC_FONT:
            (void) tsWriteStr(ezConsoleWidget, "Error : ");
        case EZ_BOLD_FONT:
        case EZ_NORMAL_FONT:
            /*
             * Used to support multiple fonts.  AsciiText widget
             * supports only one.
             */
            (void) tsWriteStr(ezConsoleWidget, stringPtr);
            break;
        default:
            EzPrintErr("Ez_Print() : unknown font");
            (void) fprintf (stdout, "%s", stringPtr);
            fflush (stdout);
            return;
        }
    }
    Ez_EmptyXEventQue();
}


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_Pause() --
 *
 *   Waits for you to type either 'c', 'q', or 'a' in a window
 * followed by return or just plain return before going on.
 *
 * Results:
 *   Returns FALSE if you type 'q', TRUE otherwise.
 *
 * Side effects:
 *   Typing 'a' will cause the program to be aborted.
 * ----------------------------------------------------------------------------
 */

bool
Ez_Pause()
{
    Window  win;
    Point   screen;
    char    string[80];
    char    letter[10];
    bool    keepGoing = TRUE;

#ifdef DEBUG
    Ez_Print(EZ_NORMAL_FONT, 
        "Type 'c' to continue, 'q' to quit, 'a' to abort: ");
#else
    Ez_Print(EZ_NORMAL_FONT, "Type 'c' to continue, 'q' to quit");
#endif DEBUG

/*
    (void) Ez_GetLine(&win, &screen, string, 80, &keepGoing);
 */

    (void) sscanf(string, "%1s", letter);
    switch (letter[0]) {
#ifdef DEBUG
    case 'a':
        abort();
        /* NOTREACHED */
        break;
#endif DEBUG
    case 'c':
        return (TRUE);
        break;
    case 'q':
        return (FALSE);
        break;
    default:
        return (TRUE);
        break;
    }
    /* NOTREACHED */
}


void
Ez_ClearConWin()
{
/*    (void) TxtClear(ezConsoleWin);*/
}


void
Ez_Prompt(promptPtr)
char *promptPtr;
{
    Ez_Print(EZ_NORMAL_FONT, promptPtr);
}


void
Ez_ToggleDebug()
{
    ezDebug = ezDebug ? FALSE : TRUE;
}


void
Ez_Flush() 
{ 
    XFlush(ez_Display); 
}


void
Ez_RingBell() 
{ 
    XBell(ez_Display, ((int) -7)); 
    XFlush(ez_Display); 
}


