
/* ezMisc.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezMisc.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"
#include "ezPsPlot.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_ItrWindows() --
 *
 * Applies the users procedure to each window.  If more then one kind of
 * window is currently being displayed, it is up to the users procedure
 * to check that it is working on the type of window for which it was 
 * designed.
 *
 * Results:
 *   EZ_COMPLETE_SEARCH is returned if the search completed normally.
 *   EZ_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

ezItrStatus
Ez_ItrWindows(userProc, clientData)
ezItrStatus (*userProc)();
ClientData  clientData;
{
    register WindowData *winDataPtr;

    if (ezWinDataHead == NIL(WindowData)) {
        return (EZ_COMPLETE_SEARCH);
    } 
    for (winDataPtr = ezWinDataHead;
        winDataPtr != NIL(WindowData);
        winDataPtr = winDataPtr->wd_nextPtr) {

        if (((*userProc)(winDataPtr, clientData)) == EZ_STOP_SEARCH) {
            return (EZ_STOP_SEARCH);
        }
    }

    return (EZ_COMPLETE_SEARCH);
}



ezStatus
Ez_FindUserWin(button, checkWinProc, clientData, winDataPtrPtr, screenPtr)
unsigned int    button;
bool            (*checkWinProc)();
WindowData      **winDataPtrPtr;
ClientData      clientData;
Point           *screenPtr;
/*
 * The procedure 'checkWinProc' should have the following form:
 *
 * bool
 * checkWinProc(winData, clientData)
 * WindowData *winData
 * ClientData clientData;
 * {
 * }
 *
 * Its purpose is to check whether the window picked by the user is
 * acceptable or not, and howdy dowdy then to return EZ_OK or
 * EZ_ERROR based on this decision.
 */
{
    XEvent          report;
    XButtonEvent    *buttonEventPtr = (XButtonEvent *) &report.xbutton;

    while (TRUE) {
        XNextEvent(ez_Display, &report);
        if ((report.type == ButtonPress) &&
            ((buttonEventPtr->button == button))) {

            if ((*winDataPtrPtr = Ez_GetWinDataOfWin(buttonEventPtr->window))
                != NIL(WindowData)) {

                if (Ez_IsConsoleWin(*winDataPtrPtr)) {
                    Ez_RingBell();
                } 
                else if (buttonEventPtr->window ==
                    (*winDataPtrPtr)->wd_userWin) {

                    if ((*checkWinProc)(*winDataPtrPtr, clientData)) {
                        screenPtr->p_x = buttonEventPtr->x;
                        screenPtr->p_y = buttonEventPtr->y;
                        return EZ_OK;
                    } 
                    else {
                        *winDataPtrPtr = NIL(WindowData);
                        return EZ_ERROR;
                    }
                }
            }
        } 
        else {
            Ez_XEventHandler(&report);
        }
    }
    /* NOTREACHED */
}


void
Ez_GetPointerXY(win, xLocPtr, yLocPtr)
Window  win;
int     *xLocPtr, *yLocPtr;
{
    Window          rootWin, childWin;
    int             xRootLoc, yRootLoc;
    unsigned int    keysAndButtons;

    (void) XQueryPointer(ez_Display, win, &rootWin, &childWin, 
        &xRootLoc, &yRootLoc, xLocPtr, yLocPtr, &keysAndButtons);
}


void
EzSetUpGCs()
{
    XGCValues values;

    values.foreground = ez_TextColor;
    values.background = ez_BackColor;
    ezFillGC = (GC) XCreateGC(ez_Display, (Drawable) EZ_ROOT_WINDOW,
        GCForeground | GCBackground, &values);

    values.font = ezPlanePtr->fid;
    ezTextGC = (GC) XCreateGC(ez_Display, (Drawable) EZ_ROOT_WINDOW,
        GCForeground | GCBackground | GCFont, &values);

    ezDrawGC = (GC) XCreateGC(ez_Display, (Drawable) EZ_ROOT_WINDOW, 
        0, 0);

    /* Use default values */
    XSetLineAttributes(ez_Display, ezDrawGC, 0, LineSolid, CapButt,
        JoinMiter);
    XSetFunction(ez_Display, ezDrawGC, GXxor);
    XSetForeground(ez_Display, ezDrawGC, ez_WhitePixel);
    XSetPlaneMask(ez_Display, ezDrawGC, AllPlanes);
}


void
EzSetUpSignalHandler(signalName, userProc)
int     signalName;
void    (*userProc)();
{
    if (signal(signalName, userProc) == (void (*) ()) -1) {
        perror(ezProgNamePtr);
        exit(1);
    }
}



ezStatus 
Ez_OpenPsFile(fileNamePtr, width, height, translateX, translateY,
    scaleX, scaleY, rotation)

/*
 * Open the specified PostScript file for output and set the global plot
 * flag so that most ez-plot drawing is reflected into the file. 
 * Ez_ClosePsFile() should be called when finished plotting.
 * All abscissa references should be negated to transform from the X11
 * left-hand coordinate system to PostScript's right-hand coordinate system.
 */
char    *fileNamePtr;
int     width, height;
double  translateX, translateY; 
double  scaleX, scaleY;
double  rotation;
{
#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((ezPsFile = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (ezErrorMsg,
            "Could not open post script file named \"%s\" for writing.",
            fileNamePtr);
        EzPrintErr(ezErrorMsg);
        return EZ_ERROR;
    }

    ezPsPlotFlag = TRUE;

    (void) PsReadColorMapFile(".xwps");
    (void) PsOpenImage(ezPsFile, width, height, translateX, translateY,
         scaleX, scaleY, rotation);

    return EZ_OK;
}


void
Ez_ClosePsFile()
{
    ezPsPlotFlag = FALSE;
    (void) PsCloseImage(ezPsFile);
    fclose (ezPsFile);
    ezPsFile = NIL(FILE);
}


float 
EzXPsGray(pixel)
unsigned long pixel;
{
#define SCALE(n) ((float)n/65535)

    XColor colorSpec;
    float r, g, b, l;
    
    colorSpec.pixel = pixel;
    XQueryColor(ez_Display, EZ_CMAP, &colorSpec);
    r = SCALE(colorSpec.red);
    g = SCALE(colorSpec.green);
    b = SCALE(colorSpec.blue);
    l = (0.30*r) + (0.59*g) + (0.11*b);

    return l;

#undef SCALE(n)
}


void
EzPrintRect(rectPtr)
Rect *rectPtr;
{
#ifdef DEBUG
    if (rectPtr == NIL(Rect)) {
        EzPrintFatalErr("EzPrintRect() failed : no rectangle given");
    }
#endif DEBUG

    (void) fprintf (ezOutFile, "xb = %d  xt = %d  yb = %d  yt = %d\n",
        rectPtr->r_xbot,rectPtr->r_xtop,
        rectPtr->r_ybot,rectPtr->r_ytop);

    fflush (ezOutFile);
}



void
EzPrintWarn(warnMsg)
char *warnMsg;
{
    (void) fprintf (ezOutFile, "%s: Warning: %s\n", EZ_MOD_NAME, warnMsg);
}


void
EzPrintErr(errMsg)
char *errMsg;
{
    (void) fprintf (ezErrFile, "%s: Error: %s\n", EZ_MOD_NAME, errMsg);
}


void
EzPrintFatalErr(fatalErrMsg)
char *fatalErrMsg;
{
    errRaise(EZ_MOD_NAME, ((int) -1), fatalErrMsg);
}


