
/* ezPickPoints.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezPickPoints.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"


/* ---------- Local Procedures ---------- */
static bool AnyUserWin();


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_PickPointList() --
 *
 *      Allows you to interactively pick the world coordinates of a
 * list of points.  These points will be forced to align with an
 * imaginary grid, that is based on the windows boundary.
 *      The list is complete when at least five points have been picked and
 * the first and last points are equal.
 *
 * Results:
 *   EZ_OK is returned along with a list of points (in which the first
 *   and last points are NOT equal) if everything is spiffy, EZ_ERROR
 *   otherwise.
 *
 * Side effects:
 *   none.
 * ----------------------------------------------------------------------------
 */

ezStatus
Ez_PickPointList(winDataPtr, pointListPtr)
WindowData  *winDataPtr;
LinkedPoint **pointListPtr; /* RETURN */
{
#define NUM_OF_POINTS 50

    Cursor              cursor;
    LinkedPoint         *lpPtr;
    Point               screen, wPoint[NUM_OF_POINTS];
    Window              userWin;
    XEvent              report;
    XButtonEvent        *bevPtr = (XButtonEvent *) &report;
    XPointerMovedEvent  *pevPtr = (XPointerMovedEvent *) &report;
    XPoint              sPoint[NUM_OF_POINTS];
    double              (*invScrMatrix)[4];
    double              (*scrMatrix)[4];
    unsigned int        eventMask = ButtonPressMask | LeaveWindowMask |
                            PointerMotionMask;
    int                 curPoint, prevPoint, worldGridSize, i;
    bool                vertLine = TRUE;
    char                errMsg1[80];

    *pointListPtr = NIL(LinkedPoint);

    (void) Ez_FindUserWin(LeftButton, AnyUserWin, (ClientData) NULL,
        &winDataPtr, &screen);
    userWin = winDataPtr->wd_userWin;
    if (XGrabPointer(ez_Display, userWin, FALSE, eventMask, GrabModeAsync,
        GrabModeAsync, None, ezVertArrowsCur, CurrentTime) != GrabSuccess) {
        EzPrintErr("could not grab button");
        return EZ_ERROR;
    }

    invScrMatrix = winDataPtr->wd_invScrMatrix;
    scrMatrix = winDataPtr->wd_scrMatrix;
    worldGridSize = winDataPtr->wd_worldGridSize;
    (void) sprintf (errMsg1, "the point list may only have %d points", 
        NUM_OF_POINTS);
    bzero((char *) sPoint, sizeof (XPoint) * NUM_OF_POINTS);

    curPoint = prevPoint = 0;

    /* Figure out the world coordinates of the point they picked. */
    wPoint[curPoint].p_x = Ez_XTrans(invScrMatrix, screen.p_x, screen.p_y);
    wPoint[curPoint].p_y = Ez_YTrans(invScrMatrix, screen.p_x, screen.p_y);
    ALIGN_POINT(&wPoint[curPoint], worldGridSize);
#ifdef DEBUG
    if (ezDebug) {
        (void) sprintf (ezErrorMsg, "%d) x = %d y = %d\n", curPoint,
            wPoint[curPoint].p_x, wPoint[curPoint].p_y);
        EzPrintErr(ezErrorMsg);
    }
#endif DEBUG

    sPoint[curPoint].x = Ez_XTrans(scrMatrix,
        wPoint[curPoint].p_x, wPoint[curPoint].p_y);
    sPoint[curPoint].y = Ez_YTrans(scrMatrix,
        wPoint[curPoint].p_x, wPoint[curPoint].p_y);

    prevPoint = curPoint++;
    if (curPoint >= NUM_OF_POINTS) {
        Ez_Print(EZ_ITALIC_FONT, errMsg1);
        goto badExit;
    } 
    else {
        if (vertLine) {
            wPoint[curPoint].p_x = wPoint[prevPoint].p_x;
        } 
        else {
            wPoint[curPoint].p_y = wPoint[prevPoint].p_y;
        }
        sPoint[curPoint].x = sPoint[prevPoint].x;
        sPoint[curPoint].y = sPoint[prevPoint].y;
        /* Draw line from the previous point to the current point. */
        XDrawLines(ez_Display, userWin, ezDrawGC, &sPoint[prevPoint], 2,
           CoordModeOrigin);
    }

    while (TRUE) {
        XNextEvent(ez_Display, &report);
        if (bevPtr->window != userWin) {
            Ez_XEventHandler(&report);
            continue;
        }
        switch (report.type) {
        case ButtonPress:
            /* Figure out the world coordinates of the point they picked. */
            if (vertLine) {
                wPoint[curPoint].p_y = Ez_YTrans(invScrMatrix,
                    sPoint[prevPoint].x, bevPtr->y);
            } 
            else {
                wPoint[curPoint].p_x = Ez_XTrans(invScrMatrix,
                    bevPtr->x, sPoint[prevPoint].y);
            }
            ALIGN_POINT(&wPoint[curPoint], worldGridSize);

            sPoint[curPoint].x = Ez_XTrans(scrMatrix,
                wPoint[curPoint].p_x, wPoint[curPoint].p_y);
            sPoint[curPoint].y = Ez_YTrans(scrMatrix,
                wPoint[curPoint].p_x, wPoint[curPoint].p_y);

            if ((sPoint[curPoint].x == sPoint[0].x) &&
                (sPoint[curPoint].y == sPoint[0].y)) {
                /*
                 * The current point is the same as the first point, 
                 * so the list is complete.
                 */
                if (curPoint < 4) {
                    Ez_Print(EZ_ITALIC_FONT,
                        "you must pick at least five points\n");
                    goto badExit;
                } 
                else {
                    XUngrabPointer (ez_Display, CurrentTime);

                    /* Build the list one element at a time. */
                    *pointListPtr = NIL(LinkedPoint);
                    for (i = 0; i < curPoint ;i++) {
                        lpPtr = ALLOC(LinkedPoint, 1);
                        lpPtr->point.p_x = wPoint[i].p_x;
                        lpPtr->point.p_y = wPoint[i].p_y;
                        lpPtr->nextPtr = *pointListPtr;
                        *pointListPtr = lpPtr;
                    }
                    /* Erase all the lines. */
                    XDrawLines(ez_Display, userWin, ezDrawGC, sPoint, 
                   (curPoint + 1), CoordModeOrigin);

                    /* Head for home pilgrim ... */
                    return EZ_OK;
                }
            } 
            else if (sPoint[curPoint].x == sPoint[prevPoint].x &&
                sPoint[curPoint].y == sPoint[prevPoint].y) {
                /*
                 * The current point is the SAME as the previous point.
                 * so ignor it !
                 */
                Ez_RingBell();
            } 
            else {
                /*
                 * The current point is NOT the same as the previous point,
                 * so add it to the list.
                 */
#ifdef DEBUG
                if (ezDebug) {
                    (void) sprintf (ezErrorMsg, "%d) x = %d y = %d\n",
                        curPoint, wPoint[curPoint].p_x, wPoint[curPoint].p_y);
                    EzPrintErr(ezErrorMsg);
                }
#endif DEBUG
                prevPoint = curPoint++;
                if (curPoint >= NUM_OF_POINTS) {
                    Ez_Print(EZ_ITALIC_FONT, errMsg1);
                    goto badExit;
                } 
                else {
                    if (vertLine = vertLine ? FALSE : TRUE) {
                        wPoint[curPoint].p_x = wPoint[prevPoint].p_x;
                        cursor = ezVertArrowsCur;
                    } else {
                        wPoint[curPoint].p_y = wPoint[prevPoint].p_y;
                        cursor = ezHorzArrowsCur;
                    }
                    if (XGrabPointer(ez_Display, userWin, FALSE, eventMask,
                        GrabModeAsync, GrabModeAsync, None,
                        cursor, CurrentTime) != GrabSuccess) {

                        EzPrintErr("could not grab button");
                        XDrawLines(ez_Display, userWin, ezDrawGC, sPoint, 
                               (curPoint + 1), CoordModeOrigin);
                        XFlush(ez_Display);
                        return EZ_ERROR;
                    }
            
                    sPoint[curPoint].x = sPoint[prevPoint].x;
                    sPoint[curPoint].y = sPoint[prevPoint].y;
                    /*
                     * Draw line from the previous point to the current
                     * point.
                     */
                    XDrawLines(ez_Display, userWin, ezDrawGC, 
                   &sPoint[prevPoint], 2, CoordModeOrigin);
                }
            }
            break;
        case LeaveNotify:
            /*
             * Leaving the window before completing the list
             * cancels the command.
             */
            goto badExit;
            break;
        case MotionNotify:
            /* Erase line from the previous point to the current point. */
            XDrawLines(ez_Display, userWin, ezDrawGC, 
                &sPoint[prevPoint], 2, CoordModeOrigin);
            
            /* Figure out the world coordinates of their current location. */
            if (vertLine) {
                wPoint[curPoint].p_y = Ez_YTrans(invScrMatrix,
                    sPoint[prevPoint].x, pevPtr->y);
            } 
            else {
                wPoint[curPoint].p_x = Ez_XTrans(invScrMatrix,
                    pevPtr->x, sPoint[prevPoint].y);
            }
            ALIGN_POINT(&wPoint[curPoint], worldGridSize);

            sPoint[curPoint].x = Ez_XTrans(scrMatrix,
                wPoint[curPoint].p_x, wPoint[curPoint].p_y);
            sPoint[curPoint].y = Ez_YTrans(scrMatrix,
                wPoint[curPoint].p_x, wPoint[curPoint].p_y);
            /* Draw line from the previous point to the current point. */
            XDrawLines(ez_Display, userWin, ezDrawGC, 
                &sPoint[prevPoint], 2, CoordModeOrigin);
            XFlush(ez_Display);
            break;
        default:
            Ez_XEventHandler(&report);
            break;
        }
    }

badExit:
    /* Erase all the lines. */
    XDrawLines(ez_Display, userWin, ezDrawGC, sPoint, 
        (curPoint + 1), CoordModeOrigin);
    XUngrabPointer(ez_Display, CurrentTime);
    Ez_RingBell();
    XFlush(ez_Display);

    return EZ_ERROR;
}



/* ARGSUSED */
static bool
AnyUserWin(winDataPtr, clientData)
WindowData  *winDataPtr;
ClientData  clientData;
{
    return TRUE;
}


