
/* ezPlotInt.h --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Definitions internal to the ezPlot module.
 *
 * Copyright (C) 1991 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezPlotInt.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef _EZPLOTINT
#define _EZPLOTINT /* once-only latch */

#include "st.h"

#if !defined(BEAR_BLACK_BOX)

#define ALARM_SIGNAL_MASK     (1 << (SIGALRM - 1))

#define DEF_FONTNAME            "8x13" /* Must be a fixed width font.        */
#define DEF_CON_NORMAL_FONTNAME "9x15"

#define DEF_BDR_WIDTH               1   /* Default border width of a window. */

#define DEF_WIN_HEIGHT            600   /* Minimum height of a window.       */
#define DEF_WIN_WIDTH             600   /* Minimum width of a window.        */
#define DEF_CONS_WIN_HEIGHT       150   /* Minimum height of console window. */
#define DEF_CONS_WIN_WIDTH        550   /* Minimum width of console window.  */

#define DEF_TB_WIN_BDR_WIDTH        1   /* Title bar window's border width.  */

#define MAX_NUM_OF_VERTICES        20   /* Must be at least five so that boxes
                                         * may be drawn.
                                         */

#define EZ_MSG_SIZE 1024

/* ---------- Internal procedures for the ezPlot module ---------- */

extern void         EzRedrawTitleBar();
extern void         EzConfigureSubWins();
extern ezStatus     EzGetXDefaults();
extern bool         EzGetButtonEvent();
extern void         EzFillRect();
extern void         EzOutlineRect();
extern void         EzMatrixMult();
extern void         EzMkIdMatrix();
extern void         EzMkScaleMatrix();
extern void         EzMkTransMatrix();
extern int          EzInvertMatrix();
extern void         EzSetUpCursors();
extern void         EzSetUpTilesAndBitmaps();
extern void         EzSetUpSignalHandler();
extern void         EzXErrorHandler();
extern float        EzXPsGray();
extern bool         EzParseBuf();

EXTERN WindowData *EzLookUpWin
    ARGS((Window win));

EXTERN void EzPrintRect
    ARGS((Rect *rectPtr));

EXTERN void EzPrintWarning
    ARGS((char *warnMsg));

EXTERN void EzPrintErr
    ARGS((char *errMsg));

EXTERN void EzPrintFatalErr
    ARGS((char *fatalerrMsg));


/* ---------- Internal global variables for the ezPlot module ---------- */

extern GC                   ezFillGC;
extern GC                   ezTextGC;
extern GC                   ezDrawGC;
extern GC                   ezStippleGC;
extern Pixmap               ezSmallDotBitmap;
extern Cursor               ezNWArrowCur;
extern Cursor               ezArrowCrossCur, ezHorzArrowsCur, ezVertArrowsCur;
extern Cursor               ezCrossCur;
extern Cursor               ezCircleOutlineCur, ezSquareOutlineCur;
extern Cursor               ezStopWatchCur;
extern Display              *ezDisplayPtr;
extern XFontStruct          *ezPlanePtr;
extern XFontStruct          *ezConFontPtr;
extern Pixmap               ezGrayBdrPixmap;
extern Window               ezRootWindow;
extern Window               ezConsoleWin;
extern Widget               ezConsoleWidget;
extern Widget               ezTopWidget;
extern WindowData           *ezWinDataHead;
extern st_table             *ezWinHashTable;
extern XtAppContext         ezAppContext;
extern unsigned int         ezBorderWidth;               
extern unsigned int         ezWinWidth;
extern unsigned int         ezWinHeight;

extern char                 ezEofChar, ezIntrChar, ezQuitChar;
extern char                 *ezConGeometryPtr; 
extern char                 *ezAppNamePtr;
extern char                 *ezProgNamePtr;
extern bool                 ezDebug;
extern bool                 ezPsPlotFlag;
extern FILE                 *ezOutFile;
extern FILE                 *ezErrFile;
extern FILE                 *ezPsFile;
extern char                 ezErrorMsg[];

#ifdef USE_OLD_X_CODE
extern Vertex               ezVertex[];
extern int                  ezVertexCount;
#endif /* USE_OLD_X_CODE */


/* ---------- Internal macros for the ezPlot module ---------- */

#define Ez_SetOption(optionMask, option) ((optionMask) |= (option))

#define ALIGN_POINT(pt, gridSize)                                       \
{                                                                       \
    (pt)->p_x -= ((pt)->p_x % (gridSize));                              \
    (pt)->p_y -= ((pt)->p_y % (gridSize));                              \
}

#define EZ_COPY_RECT(rect1, rect2)                                      \
{                                                                       \
    (rect1)->r_xbot = (rect2)->r_xbot;                                  \
    (rect1)->r_xtop = (rect2)->r_xtop;                                  \
    (rect1)->r_ybot = (rect2)->r_ybot;                                  \
    (rect1)->r_ytop = (rect2)->r_ytop;                                  \
}

#ifdef USE_OLD_X_CODE
#define EZ_DRAW_VERTICES(win, vertex, vertexCount)                      \
    XDraw(ez_Display, (win), ezDrawGC, (vertex), (vertexCount));
#endif /* USE_OLD_X_CODE */

#endif /* BEAR_BLACK_BOX */

#endif /* _EZPLOTINT */

