
/* ezPsPlot.h --
 *
 *     Benjamin Chen
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1990 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezPsPlot.h,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $
 */

#ifndef _EZ_POSTSCRIPT 
#define _EZ_POSTSCRIPT /* once-only latch */

#define BUTT_CAP   0
#define ROUND_CAP  1
#define SQUARE_CAP 2

/* Return Codes */
#define PS_OK 0
#define PS_ERROR 1

/* Gray scale colors */
#define PS_BLACK 0.0
#define PS_WHITE 1.0
#define PS_NOFILL -1.0

extern int PsDrawSolidLine(/* psFile, x1, y1, x2, y2, 
                  capStyle, lineWidth, lineGray */);
/*
 *   FILE *psFile;
 *   int x1, y1, x2, y2, capStyle, lineWidth;
 *   flot lineGray;
 *
 * This procedure draws a line to the PostScript file.  'xn' and 'yn' determine
 * the endpoints of the line. 'capStyle' refers to the PostScript code that
 * determines how PostScript will end line segments:  butt caps, round caps,
 * or projecting square caps (see psplot.h).  'lineWidth' and 'lineGray'
 * refer to the width and gray scale of the line, respectively.
 */

extern int PsDrawCircle(/* psFile, ox, oy, radius, 
               lineWidth, lineGray, fillGray */);
/*    FILE *psFile;
 *   int ox, oy, radius, lineWidth;
 *   float lineGray, fillGray;
 *
 * This procedure draws a Circle to the PostScript file. 'ox' and 'oy'
 * are the coordinates of the origin of the circle.  'radius' refers
 * to the radius of the circle.  'lineWidth' is the weight of the line 
 * surrounding the circle.  'lineGray' and 'fillGray' specify the gray 
 * scale of the surrounding line and the fill color of the Circle,
 * respectively.  If fillGray is  set to NO_FILL, then only an outline
 * will be drawn.
 */

extern int PsDrawRectangle(/* psFile, llx, lly, width, height, lineWidth,
                  lineGray, fillGray */);
/*   FILE *psFile;
 *   int llx, lly, width, height, lineWidth;
 *   float lineGray, fillGray
 *
 * This procedure draws a rectangle to the PostScript file. 'llx' and 'lly'
 * are the coordinates of the lower left corner of the rectanle.  'width'
 * and 'height' refer to the corresponding dimensions of the rectangle.
 * 'lineWidth' is the weight of the line surrounding the rectangle.
 * 'lineGray' and 'fillGray' specify the gray scale of the surrounding
 * line and the fill color of the rectangle, respectively.  If fillGray is
 * set to NO_FILL, then only an outline will be drawn.
 */

extern int 
PsDrawPolygon(/* psFile, points, npoints, lineWidth, lineGray, fillGray */);
/*
 * FILE     *psFile;
 * XPoint   *points;
 * int      npoints;
 * int      lineWidth;
 * float    lineGray, fillGray;
 *
 * This procedure writes out a postscript description of a polygon.
 *
 * 'lineWidth' is the weight of the line surrounding the polygon.
 * 'lineGray' and 'fillGray' specify the gray scale of the surrounding
 * line and the fill color of the polygon, respectively. If fillGray is
 * set to NO_FILL, then only an outline will be drawn.
 */


/*ARGSUSED*/
extern int PsDrawText(/* psFile, text, x, y, font, pointSize */);
/*   FILE *psFile;
 *   char *text;
 *   int x, y;
 *   char *font;
 *   int pointSize;
 *
 * This procedure writes the 'text' string at location 'x', 'y' using
 * the given 'font' and 'pointSize'.
 */

/*ARGSUSED*/
extern int PsDrawRText(/* psFile, text, x, y, font, pointSize */);
/*   FILE *psFile;
 *   char *text;
 *   int x, y;
 *   char *font;
 *   int pointSize;
 *
 * This procedure writes the 'text' string at location 'x', 'y' using
 * the given 'font' and 'pointSize'.  The text is rotated 180 degrees
 * so, in effect, it is "upside down."
 */

extern int PsOpenImage(/* psFile, width, height,
           translateX, translateY, scaleX, scaleY, rotation */);
/*   FILE *psFile;
 *   int width, height; Dimensions of the image to be plotted.
 *   float scaleX; In inches 
 *   float scaleY; In inches
 *   float translateX; In inches 
 *   float translateY;  In inches 
 *   float rotation;  in degrees counterclockwise 
 *
 * This procedure writes the PostScript header, defines miscellaneous
 * PostScript functions to be used in the ps file, and writes the
 * commands for the appropriate page transformations.  If width and
 * height are non-zero (known) then a bounding box is calculated for
 * use in programs such as LaTex.  'psFile' must be a valid output
 * stream.
 */

extern int PsCloseImage(/* psFile */);

extern int PsReadColorMapFile(/* fileName */);

#endif /* _EZ_POSTSCRIPT */

