
/* ezTrans.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezTrans.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"


int
EzInvertMatrix(m, inv)
Matrix m, inv;
{
/*
 * EzInvertMatrix() is a brute force method using the formula
 * inv[][] = (1.0 / det(m[][]))*adj(m[][])
 */

    Matrix adj;
    double a, t[3][3];
    int r, c, u;
    register int i, j, v;

    /* find the matrix of cofactors and store it in inv[][] */
    for (r = 0; r < 4 ;r++)
        for (c = 0; c < 4 ;c++) {
            u = 0;
            for (i = 0; i < 4 ;i++) {
                if (i != r) {
                    v = 0;
                    for (j = 0; j < 4 ;j++)
                        if (j != c) 
                            t[u][v++] = m[i][j];
                    u++;
                }
            }
            /* find the determinant of t[][] */
            a = t[0][0]*(t[1][1]*t[2][2] - t[1][2]*t[2][1])
                + t[0][1]*(t[1][2]*t[2][0] - t[1][0]*t[2][2])
                + t[0][2]*(t[1][0]*t[2][1] - t[1][1]*t[2][0]);
            if ((r + c) & 01)
                a *= -1.0;
            inv[r][c] = a;
        }
    /* find the transpose of the cofactor matrix which is the adjoint */
    for (i = 0; i < 4 ;i++)
        for (j = 0; j < 4 ;j++) 
            adj[i][j] = inv[j][i];
    /* det(m[][]) equals the last column of m[][] times
     * the cofactors of that column, i.e. inv[3][3]) */
    if (inv[3][3] == 0.0) {
        EzPrintErr("This matrix does not have an inverse !!");
        return (FALSE);
    }
    a = 1.0 / inv[3][3];
    /* now find the inverse */
    for (i = 0; i < 4 ;i++)
        for (j = 0; j < 4 ;j++) 
            inv[i][j] = a * adj[i][j];
    return (TRUE);
}

void
EzMatrixMult(m1, m2, m3)
Matrix m1, m2, m3;
{
/* m1, m2, m3 MUST ALL BE DIFFERENT ARRAYS!!!!!!
 * 
 * m3 = m1 * m2 */

    register int i, j, k;
    double t;

    for (i = 0; i < 4 ;i++) 
        for (j = 0; j < 4 ;j++) {
            t = 0.0;
            for (k = 0; k < 4 ;k++)
                t += m1[i][k] * m2[k][j];
            m3[i][j] = t;
        }
}

void
EzMkIdMatrix(id)
register Matrix id;
{
/* initialize the matrix to look like this
 *
 *  | 1 0 0 0 |
 *  | 0 1 0 0 |
 *  | 0 0 1 0 |
 *  | 0 0 0 1 |
 *
 */
    register int i, j;

    for (i = 0; i < 4 ;i++)
        for (j = 0; j < 4 ;j++) 
            id[i][j] = (i == j) ? 1.0 : 0.0;
}

void
EzMkScaleMatrix(sx, sy, sz, scaleMatrix)
double sx, sy, sz;
Matrix scaleMatrix;
{
/*  
 *  | sx 0  0  0 |
 *  | 0  sy 0  0 |
 *  | 0  0  sz 0 |
 *  | 0  0  0  1 |
 */
    EzMkIdMatrix(scaleMatrix);

    scaleMatrix[0][0] = sx;
    scaleMatrix[1][1] = sy;
    scaleMatrix[2][2] = sz;
}

void
EzMkTransMatrix(tx, ty, tz, t)
double tx, ty, tz;
Matrix t;
{
/*  
 *  | 1  0  0  0 |
 *  | 0  1  0  0 |
 *  | 0  0  1  0 |
 *  | tx ty tz 1 |
 */
    EzMkIdMatrix(t);

    t[3][0] = tx;
    t[3][1] = ty;
    t[3][2] = tz;
}

