
/* ezView.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezView.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_CalcViewingTrans() --
 *
 *   Calculates the matrix that will handle the transformation from
 *   world coordinates to the screen coordinates as well as this matrix's
 *   inverse.
 *
 * Results:
 *   EZ_OK is returned if everything went swell, EZ_ERROR otherwise.
 *
 * Side effects:
 *   After the transformation matrix is calculated 'winDataPtr->wd_imageBnd'
 *   is recomputed so that it covers the whole window.
 * ----------------------------------------------------------------------------
 */

ezStatus
Ez_CalcViewingTrans(winDataPtr)
register WindowData *winDataPtr;    /* This must contain up to date 
                                     * information about the size of the
                                     * window !!
                                     */
{
    Matrix mat1, mat2, mat3;
    Rect *imageBndPtr;              /* Area of the world to be drawn. */
    double winAspectRatio;          /* Aspect ratio (winHeight / winWidth) of
                                     * the window on the screen where the image
                                     * will be drawn.
                                     */
    double imageAspectRatio;        /* Aspect ratio (imageHeight / imageWidth)
                                     * of the image to be drawn.
                                     */
    double tx, ty, sx, sy;
    int scrGridSize;
    int worldGridSize;
    double imageWidth, imageHeight; /* The width and height in world
                                     * coordinates of the area to be drawn.
                                     */
    double winWidth, winHeight;     /* The width and height in screen
                                     * coordinates of the window.
                                     */
    register double (*invScrMatrix)[4];

    imageBndPtr = &winDataPtr->wd_imageBnd;
    tx = (imageBndPtr->r_xbot == 0) ? 0 : -imageBndPtr->r_xbot;
    ty = (imageBndPtr->r_ytop == 0) ? 0 : -imageBndPtr->r_ytop;
    EzMkTransMatrix(tx, ty, (double) 0.0, mat1);

    imageWidth = imageBndPtr->r_xtop - imageBndPtr->r_xbot + 1;
    imageHeight = imageBndPtr->r_ytop - imageBndPtr->r_ybot + 1;

    EzMkScaleMatrix(1.0 / imageWidth, 1.0 / imageHeight, (double) 1.0, mat2);
    EzMatrixMult(mat1, mat2, mat3);

    EzMkScaleMatrix((double) 1.0, (double) -1.0, (double) 1.0, mat1);
    EzMatrixMult(mat3, mat1, mat2);

    winWidth = winDataPtr->wd_userWinInfo.width;
    winHeight = winDataPtr->wd_userWinInfo.height;
    imageAspectRatio = imageHeight / imageWidth;
    winAspectRatio = winHeight / winWidth;
    /* Fit the whole image onto the window. */
    if (((imageAspectRatio < 1.0) &&
        (imageAspectRatio <= winAspectRatio)) ||
        ((imageAspectRatio >= 1.0) &&
        (imageAspectRatio < winAspectRatio))) {

        sx = winWidth;
        sy = winWidth * imageAspectRatio;
        EzMkScaleMatrix(sx, sy, (double) 1.0, mat1);
        EzMatrixMult(mat2, mat1, mat3);
        EzMkTransMatrix((double) 0.0, (winHeight - sy) / 2.0,
            (double) 0.0, mat2);
        EzMatrixMult(mat3, mat2, winDataPtr->wd_scrMatrix);
    } else {
        sx = winHeight / imageAspectRatio;
        sy = winHeight;
        EzMkScaleMatrix(sx, sy, (double) 1.0, mat1);
        EzMatrixMult(mat2, mat1, mat3);
        EzMkTransMatrix((winWidth - sx) / 2.0, (double) 0.0,
            (double) 0.0, mat2);
        EzMatrixMult(mat3, mat2, winDataPtr->wd_scrMatrix);
    } 
    invScrMatrix = winDataPtr->wd_invScrMatrix;
    if (EzInvertMatrix(winDataPtr->wd_scrMatrix, invScrMatrix) == FALSE) {
        return EZ_ERROR;
    }
    /* Make the image boundary cover the whole window. */
    imageBndPtr->r_xbot = Ez_XTrans(invScrMatrix, 0, 0);
    imageBndPtr->r_ytop = Ez_YTrans(invScrMatrix, 0, 0);
    imageBndPtr->r_xtop = Ez_XTrans(invScrMatrix, winWidth, winHeight);
    imageBndPtr->r_ybot = Ez_YTrans(invScrMatrix, winWidth, winHeight);

    /*
     * Compute the width of the grid in world coordinates.
     */
    scrGridSize = winDataPtr->wd_scrGridSize;
    if ((worldGridSize = Ez_XTrans(invScrMatrix, scrGridSize, 0) -
            Ez_XTrans(invScrMatrix, 0, 0)) == 0) {
        worldGridSize = 1;
    }
    winDataPtr->wd_worldGridSize = ABS(worldGridSize);

    return EZ_OK;
}


void
Ez_StretchRect(rectPtr, percent)
register Rect   *rectPtr;
double          percent;
{
    register int deltaX, deltaY;

    deltaX = (rectPtr->r_xtop - rectPtr->r_xbot) * (percent - 1.0) / 2.0;
    deltaY = (rectPtr->r_ytop - rectPtr->r_ybot) * (percent - 1.0) / 2.0;
    rectPtr->r_xbot -= deltaX;
    rectPtr->r_ybot -= deltaY;
    rectPtr->r_xtop += deltaX;
    rectPtr->r_ytop += deltaY;
}


void
Ez_Zoom(winDataPtr, percent)
WindowData  *winDataPtr;
double      percent;
{
    Ez_StretchRect(&winDataPtr->wd_imageBnd, percent);
    (void) Ez_CalcViewingTrans(winDataPtr);
    (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
}


