
/* ezWinUtils.c --
 *
 *     George Carvalho, Esq.
 *     Ben Chen
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezWinUtils.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include "ts.h"                 /* For console window.       */


/* ---------- Local Procedures ---------- */
static WindowData   *MakeWindowData();
static ezStatus     SetUpWin();
static void         EzConsInput();


/* 
 * ----------------------------------------------------------------------------
 *
 * MakeWindowData() --
 *
 *    Allocates memory for a new window data structure, and initializes
 *    the fields.
 *
 * Results:
 *    A pointer to a window data structure is returned.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static WindowData *
MakeWindowData()
{
    WindowData *winDataPtr;

    winDataPtr = ALLOC(WindowData, 1);
    winDataPtr->wd_baseWin = (Window) 0;
    winDataPtr->wd_titleBarWin = (Window) 0;
    winDataPtr->wd_userWin = (Window) 0;
    winDataPtr->wd_redrawProc = (void (*)()) NULL;
    winDataPtr->wd_closeWinProc = (void (*)()) NULL;
    winDataPtr->wd_userWinButtonProc = (void (*)()) NULL;
    winDataPtr->wd_clientData = (ClientData) NULL;
    winDataPtr->wd_nextPtr = NIL(WindowData);

    return (winDataPtr);
}



static ezStatus
SetUpWin(tbTextPtr, baseFramePtr, userWinPtr, winDataPtr)
char        *tbTextPtr;         /* title bar text */
OpaqueFrame *baseFramePtr;
Window      *userWinPtr;
WindowData  *winDataPtr;
{
    XFontStruct *fontInfoPtr = ezConFontPtr;
    Window      baseWin, titleBarWin, userWin;
    int         width, height;
    int         deltaY, pixLen;
    int         tbHeight;           /* title bar window height              */

    width = baseFramePtr->width;
    height = baseFramePtr->height;
    baseWin = baseFramePtr->self;
    tbHeight = winDataPtr->wd_titleBar.tb_winHeight = 
        fontInfoPtr->ascent + fontInfoPtr->descent;
    if (tbTextPtr != NIL(char)) {
        winDataPtr->wd_titleBar.tb_visible = TRUE;
    } 
    else {
        winDataPtr->wd_titleBar.tb_visible = FALSE;
    }
    if ((titleBarWin = XCreateSimpleWindow(ez_Display, baseWin,
        -DEF_TB_WIN_BDR_WIDTH, -DEF_TB_WIN_BDR_WIDTH, width, tbHeight,
        DEF_TB_WIN_BDR_WIDTH, baseFramePtr->border,
        baseFramePtr->background)) == (Window) 0) {

        EzPrintErr("cannot create title bar window");
        XDestroyWindow(ez_Display, titleBarWin);
        return EZ_ERROR;
    }
    deltaY = tbHeight + DEF_TB_WIN_BDR_WIDTH;
    height = height - tbHeight;
    if ((userWin = XCreateSimpleWindow(ez_Display, baseWin,
        0, deltaY, width, height, 
        DEF_TB_WIN_BDR_WIDTH, baseFramePtr->border,
        baseFramePtr->background)) == (Window) 0) {

        EzPrintErr("cannot create user window");
        XDestroyWindow(ez_Display, userWin);
        return EZ_ERROR;
    }
    winDataPtr->wd_titleBar.tb_fontInfoPtr = fontInfoPtr;
    {char *s; s = "JENNIFER CONNELLY IS ONE FINE BABE";}
    winDataPtr->wd_titleBar.tb_y = fontInfoPtr->ascent;
    if (winDataPtr->wd_titleBar.tb_visible) {
        winDataPtr->wd_titleBar.tb_textPtr = util_strsav(tbTextPtr);
        pixLen = winDataPtr->wd_titleBar.tb_pixLen =
            XTextWidth(fontInfoPtr, tbTextPtr, strlen(tbTextPtr));
        winDataPtr->wd_titleBar.tb_x = ((pixLen < width) ?
            ((width - pixLen) / 2) : 0);
    }
    XDefineCursor(ez_Display, userWin, ezCrossCur);
    XDefineCursor(ez_Display, baseWin, ezNWArrowCur);
    XSelectInput(ez_Display, baseWin, ExposureMask | EnterWindowMask |
         LeaveWindowMask | StructureNotifyMask);
    if (XGetWindowAttributes(ez_Display, baseWin, 
        &winDataPtr->wd_baseWinInfo) == 0) {
        EzPrintErr("cannot query base window");
        return EZ_ERROR;
    }
    XSelectInput(ez_Display, titleBarWin, ExposureMask | ButtonReleaseMask
         | StructureNotifyMask);
    if (winDataPtr->wd_titleBar.tb_visible) {
        XMapRaised(ez_Display, titleBarWin);
    }

    winDataPtr->wd_baseWin =     baseWin;
    winDataPtr->wd_titleBarWin = titleBarWin;
    winDataPtr->wd_userWin =     userWin;
    (void) st_insert(ezWinHashTable, 
        (char *) winDataPtr->wd_baseWin,     (char *) winDataPtr);
    (void) st_insert(ezWinHashTable, 
        (char *) winDataPtr->wd_titleBarWin, (char *) winDataPtr);
    (void) st_insert(ezWinHashTable, 
        (char *) winDataPtr->wd_userWin,     (char *) winDataPtr);

    *userWinPtr = userWin;

    return EZ_OK;
}



static XWMHints xwmh = {
    (InputHint|StateHint),  /* flags */
    True,           /* input */
    NormalState,    /* initial_state */
    0,              /* icon pixmap */
    0,              /* icon window */
    0, 0,           /* icon location */
    0,              /* icon mask */
    0,              /* Window group */
};


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_CreateWin() --
 *
 * Results:
 *     EZ_OK is returned if everything went swell, EZ_ERROR otherwise.
 *
 * Side effects:
 *     The new window is added to this modules global list of windows. Also
 *     it's mapped after XSelectInput() so that the ensuing ExposeWindow event
 *     will automatically cause the users redraw procedure to be called
 *     on the window.  Therefore provided that X events are being processed
 *     you should not draw the window yourself.
 *
 * ----------------------------------------------------------------------------
 */

ezStatus
Ez_CreateWin(promptPtr, geometryPtr, tbTextPtr, redrawProc, imageBndPtr,
    winDataPtrPtr)

char                *promptPtr;         /* Text string that the X server
                                         * will use, to prompt the user with
                                         * when it creates the new window. It
                                         * will be used for the window name
                                         * as well.
                                         */
char                *geometryPtr;       /* X geometry specs (not used)      */
char                *tbTextPtr;         /* Title bar text string.  It will
                                         * be used as the icon name as well.
                                         */
void                (*redrawProc)();
Rect                *imageBndPtr;
WindowData          **winDataPtrPtr;
{
#define BAD_RETURN()                                                        \
{                                                                           \
    FREE(winDataPtr);                                                       \
    *winDataPtrPtr = NIL(WindowData);                                       \
    return EZ_ERROR;                                                        \
}

    OpaqueFrame             frame;
    Window                  baseWin, userWin, root; /* root is a dummy var */
    WindowData              *winDataPtr;
    XSetWindowAttributes    xswa; /* Temporary Set Window Attribute struct */
    XClassHint              classHint;
    XSizeHints              xsh; /* Size hints for window manager */
    unsigned int            depth; /* dummy var */
    int                     tbHeight, x, y; /* x, y dummy vars */
    char                    *iconNamePtr, *winNamePtr;

    winDataPtr = MakeWindowData();

    tbHeight = ezConFontPtr->ascent + 6;
    winDataPtr->wd_titleBar.tb_winHeight = tbHeight;
    xsh.flags = (PPosition | PSize);
    xsh.height = tbHeight + ezWinHeight;
    xsh.width = ezWinWidth;
    xsh.x = (DisplayWidth(ez_Display, DefaultScreen(ez_Display)) - 
         xsh.width)  / 2;
    xsh.y = (DisplayHeight(ez_Display, DefaultScreen(ez_Display)) - 
         xsh.height) / 2;
    if ((baseWin = XCreateSimpleWindow(ez_Display, EZ_ROOT_WINDOW,
        xsh.x, xsh.y, xsh.width, xsh.height, 
        0, ez_BorderColor, ez_BackColor)) == (Window) 0) {

        EzPrintErr("cannot create base window");
        XDestroyWindow(ez_Display, baseWin);
        return EZ_ERROR;
    }

    classHint.res_name =  ezProgNamePtr;
    classHint.res_class = ezAppNamePtr;
    XSetClassHint(ez_Display, baseWin, &classHint);

    frame.self = baseWin;
    XGetGeometry(ez_Display, (Drawable) baseWin, &root, &x, &y, &frame.width,
        &frame.height, &frame.bdrwidth, &depth);

    winNamePtr = util_strsav(promptPtr);
    iconNamePtr = util_strsav(tbTextPtr);
    XSetStandardProperties(ez_Display, baseWin, winNamePtr, iconNamePtr,
        None, NIL(char *), 0, &xsh);

    XSetWMHints(ez_Display, baseWin, &xwmh);
    
    /*
     * Ensure that the window's colormap field points to the default
     * colormap,  so that the window manager knows the correct colormap to
     * use for the window.  See Section 3.2.9. Also,  set the window's Bit
     * Gravity to reduce Expose events.
     */
    xswa.colormap = DefaultColormap(ez_Display, DefaultScreen(ez_Display));
    xswa.bit_gravity = CenterGravity;
    xswa.border_pixmap = ezGrayBdrPixmap;
    XChangeWindowAttributes(ez_Display, baseWin, (CWColormap | CWBitGravity), 
        &xswa);
    frame.bdrwidth = ezBorderWidth;
    frame.border = ez_WhitePixel;
    frame.background = ez_BackColor;
    if (SetUpWin(tbTextPtr, &frame, &userWin, winDataPtr) != EZ_OK) {
        BAD_RETURN();
    }
    winDataPtr->wd_redrawProc = redrawProc;
    if (XGetWindowAttributes(ez_Display, userWin,
        &winDataPtr->wd_userWinInfo) == 0) {
        EzPrintErr("cannot query user window");
        BAD_RETURN();
    }
    winDataPtr->wd_scrGridSize = 12;
    if (imageBndPtr != NIL(Rect)) {
        EZ_COPY_RECT(&winDataPtr->wd_imageBnd, imageBndPtr);
        if (Ez_CalcViewingTrans(winDataPtr) != EZ_OK) {
            BAD_RETURN();
        }
    }

    /*
     * Include ButtonReleased, and MouseMoved events
     * in the mask only when you need them.
     */
    XSelectInput(ez_Display, userWin, ExposureMask | LeaveWindowMask |
        KeyPressMask | ButtonPressMask | StructureNotifyMask);
    XMapWindow(ez_Display, userWin);

    /*
     * Add this window to the top of the list.
     */
    winDataPtr->wd_nextPtr = ezWinDataHead;
    ezWinDataHead = winDataPtr;

    *winDataPtrPtr = winDataPtr;
    XMapWindow(ez_Display, baseWin);
    XFlush(ez_Display);

    return EZ_OK;
}



void
EzConfigureSubWins(baseWinWidth, baseWinHeight, winDataPtr)
int         baseWinWidth, baseWinHeight;
WindowData  *winDataPtr;
/*
 * Postion all the sub windows snuggly inside of the base window.
 */
{
    int deltaY, pixLen;
    int tbHeight;       /* title bar window height */

    /* 
     * The title bar must be updated even if its not currently being
     * displayed so that it will be ready to go when it's turned
     * back on.
     */
    tbHeight = winDataPtr->wd_titleBar.tb_winHeight;
    XMoveResizeWindow(ez_Display, winDataPtr->wd_titleBarWin,
        -DEF_TB_WIN_BDR_WIDTH, -DEF_TB_WIN_BDR_WIDTH,
        baseWinWidth, tbHeight);
    pixLen = winDataPtr->wd_titleBar.tb_pixLen;
    winDataPtr->wd_titleBar.tb_x = ((pixLen < baseWinWidth) ?
        ((baseWinWidth - pixLen) / 2) : 0);

    if (winDataPtr->wd_titleBar.tb_visible) {
        deltaY = tbHeight + DEF_TB_WIN_BDR_WIDTH;
        XMoveResizeWindow(ez_Display, winDataPtr->wd_userWin,
            0, deltaY,
            baseWinWidth, baseWinHeight - tbHeight);
    } 
    else {
        XMoveResizeWindow(ez_Display, winDataPtr->wd_userWin,
            0, 0, baseWinWidth, baseWinHeight);
    }
}


void
Ez_ToggleTitleBar(winDataPtr)
WindowData *winDataPtr;
{
    XWindowAttributes  *winInfoPtr;
    int                deltaY;

    winInfoPtr = &winDataPtr->wd_baseWinInfo;
    deltaY = winDataPtr->wd_titleBar.tb_winHeight + DEF_TB_WIN_BDR_WIDTH;
    if (winDataPtr->wd_titleBar.tb_visible) {
        /* turn the title bar off */

        /*
         * Unmap the title bar window, move the user window up
         * (relative to the base window), and then move the base
         * window down and shrink it so it fits under the user window.
         */
        XUnmapWindow(ez_Display, winDataPtr->wd_titleBarWin);
        XMoveWindow(ez_Display, winDataPtr->wd_userWin, 0, 0);
        XMoveResizeWindow(ez_Display, winDataPtr->wd_baseWin,
            winInfoPtr->x, winInfoPtr->y + deltaY,
            winInfoPtr->width, winInfoPtr->height - deltaY);

        /* 
         * Update the window info field here to prevent the event
         * handler from trying to re-configure the window.
         */
        if (XGetWindowAttributes(ez_Display, winDataPtr->wd_baseWin,
            &winDataPtr->wd_baseWinInfo) == 0) {
            EzPrintErr("cannot get base window information");
            return;
        }
        winDataPtr->wd_titleBar.tb_visible = FALSE;
    } 
    else {
        /* turn the title bar on */

        /*
         * Move the base window up and enlarge it so there
         * will be room for the title bar window.
         */
        XMoveResizeWindow(ez_Display, winDataPtr->wd_baseWin,
            winInfoPtr->x, winInfoPtr->y - deltaY,
            winInfoPtr->width, winInfoPtr->height + deltaY);

        /* 
         * Update the window info field here to prevent the event
         * handler from trying to re-configure the window.
         */
        if (XGetWindowAttributes(ez_Display, winDataPtr->wd_baseWin,
            &winDataPtr->wd_baseWinInfo) == 0) {
            EzPrintErr("cannot get base window information");
            return;
        }
        /*
         * Move the user window down (relative to the base window),
         * and then map the title bar window.
         */
        XMoveWindow(ez_Display, winDataPtr->wd_userWin, 0, deltaY);
        XMapWindow(ez_Display, winDataPtr->wd_titleBarWin);
        winDataPtr->wd_titleBar.tb_visible = TRUE;
    }
}



ezStatus
Ez_DestroyWin(winDataPtr)
register WindowData *winDataPtr;
/*
 * Side effects:
 *   This window is removed from the  global window list, and from
 *   the hash tables maintained by X.
 */
{
    register WindowData *curWinDataPtr;
    char                *keyPtr;

#ifdef DEBUG
    if (winDataPtr == NIL(WindowData)) {
        EzPrintFatalErr("Ez_DestroyWindow() failed:  winDataPtr is NULL)");
    }
#endif DEBUG

    keyPtr = (char *) winDataPtr->wd_baseWin;
    (void) st_delete(ezWinHashTable, &keyPtr, NIL(char *));
    keyPtr = (char *) winDataPtr->wd_titleBarWin;
    (void) st_delete(ezWinHashTable, &keyPtr, NIL(char *));
    keyPtr = (char *) winDataPtr->wd_userWin;
    (void) st_delete(ezWinHashTable, &keyPtr, NIL(char *));

    XDestroyWindow(ez_Display, winDataPtr->wd_baseWin);

    if (winDataPtr == ezWinDataHead) {
        ezWinDataHead = winDataPtr->wd_nextPtr;
    } else {
        curWinDataPtr = ezWinDataHead;
        while (curWinDataPtr->wd_nextPtr != winDataPtr &&
            curWinDataPtr->wd_nextPtr != NIL(WindowData)) {

            curWinDataPtr = curWinDataPtr->wd_nextPtr;
        }
        if (curWinDataPtr->wd_nextPtr == winDataPtr) {
            curWinDataPtr->wd_nextPtr = winDataPtr->wd_nextPtr;
        } else {
            EzPrintErr("there's no such pointer in list");
            return EZ_ERROR;
        }
    }
    FREE(winDataPtr);

    return EZ_OK;
}


static void 
EzConsInput(w, client_data, call_data)
Widget  w;                      /* Console window widget   */
caddr_t client_data;            /* Static data from call   */
caddr_t call_data;              /* Dynamic data from call  */
/*
 * This routine handles typed input into the console window.  The
 * call data is actually a pointer to (ts_call_data) which indicates
 * what characters where typed.  This is passed on to doKeyInput.
 */
{
    ts_call_data *data = (ts_call_data *) call_data;

    if (data->len == 0) {
        return;
    }
    (void) EzParseBuf(data->buf, data->len);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_CreateConsoleWin() --
 *
 * Results:
 *   EZ_OK is returned if everything is spiffy, EZ_ERROR otherwise.
 *
 * Side effects:
 *     None.
 * ----------------------------------------------------------------------------
 */

ezStatus
Ez_CreateConsoleWin(tbTextPtr, winDataPtrPtr)
char        *tbTextPtr;         /* title bar text                   */
WindowData  **winDataPtrPtr;
{
    Arg args[20];
    int argCount;

    if (ezConsoleWin != (Window) NULL) {
        Ez_Print(EZ_ITALIC_FONT, 
            "Only one console window per customer buddy.\n");
        return EZ_ERROR;
    }

    argCount = 0;
    XtSetArg(args[argCount], XtNgeometry,    ezConGeometryPtr);    argCount++;
    XtSetArg(args[argCount], XtNinput,       TRUE);                argCount++;
    XtSetArg(args[argCount], XtNtitle,       tbTextPtr);           argCount++;
    XtSetArg(args[argCount], XtNborderWidth, ezBorderWidth);       argCount++;
    XtSetArg(args[argCount], XtNbackground,  ez_ConBackColor);     argCount++;
    XtSetArg(args[argCount], XtNforeground,  ez_ConTextColor);     argCount++;
    XtSetArg(args[argCount], XtNborderColor, ez_ConBorderColor);   argCount++;
    XtSetArg(args[argCount], XtNfont,        ezConFontPtr);        argCount++;
    XtSetArg(args[argCount], XtNwidth,       DEF_CONS_WIN_WIDTH);  argCount++;
    XtSetArg(args[argCount], XtNheight,      DEF_CONS_WIN_HEIGHT); argCount++;

    XtSetValues(ezTopWidget, args, argCount);

    ezConsoleWidget = tsCreate(ezAppContext, tbTextPtr, ezTopWidget, 8192, 
        EzConsInput, NULL, args, argCount);
    XtRealizeWidget(ezTopWidget);
    ezConsoleWin = XtWindow(ezConsoleWidget);

    XDefineCursor(ez_Display, ezConsoleWin, ezCrossCur);

    *winDataPtrPtr = MakeWindowData();

    if (XGetWindowAttributes(ez_Display, ezConsoleWin,
        &(*winDataPtrPtr)->wd_userWinInfo) == 0) {

        ezConsoleWin = (Window) NULL;
        EzPrintErr("cannot get console window information");
        return EZ_ERROR;
    }

    (*winDataPtrPtr)->wd_userWin = ezConsoleWin;
    (void) st_insert(ezWinHashTable, 
        (char *) (*winDataPtrPtr)->wd_userWin, (char *) (*winDataPtrPtr));

    return EZ_OK;
}



void
Ez_DestroyConsoleWin()
{
    WindowData  *winDataPtr;
    char        *keyPtr;

    if ((ezConsoleWin != (Window) NULL) &&

        ((winDataPtr = EzLookUpWin(ezConsoleWin)) != NIL(WindowData))) {

        keyPtr = (char *) winDataPtr->wd_userWin;
        (void) st_delete(ezWinHashTable, &keyPtr, NIL(char *));

        XtDestroyWidget(ezConsoleWidget);
        ezConsoleWin = (Window) NULL;
    }
}



WindowData *
EzLookUpWin(win)
Window win;
{
    WindowData *winDataPtr;

    if (st_lookup(ezWinHashTable, (char *) win, (char **) &winDataPtr) == 1) {
        return (winDataPtr);
    }
    return (NIL(WindowData));
}


WindowData *
Ez_GetWinDataOfWin(win)
Window win;
/*
 * Only the user window of the WindowData structure is visible outside
 * of this module.
 */
{
    WindowData *winDataPtr;

    if ((winDataPtr = EzLookUpWin(win)) == NIL(WindowData)) {
        return (NIL(WindowData));
    }
    else if (win != winDataPtr->wd_userWin) {
        return (NIL(WindowData));
    } 

    return (winDataPtr);
}



bool
Ez_IsConsoleWin(winDataPtr)
WindowData *winDataPtr;
{
    return (winDataPtr->wd_userWin == ezConsoleWin);
}



void
EzRedrawTitleBar(winDataPtr)
WindowData *winDataPtr;
{
    TitleBar *titleBarPtr;
    char     *textPtr;
    XGCValues values;
    static GC gc = (GC) 0;

    titleBarPtr = &winDataPtr->wd_titleBar;
    values.foreground = ez_BorderColor;
    values.background = ez_BackColor;
    if (!gc) {
        values.font = titleBarPtr->tb_fontInfoPtr->fid;
        gc = (GC) XCreateGC(ez_Display, (Drawable) winDataPtr->wd_titleBarWin,
            GCFont|GCForeground|GCBackground, &values);
    }
    if (titleBarPtr->tb_visible) {
        textPtr = titleBarPtr->tb_textPtr;
        if (winDataPtr->wd_userWin == ezConsoleWin) {
            /* title bar of console window */
        values.foreground = ez_ConBorderColor;
        values.background = ez_ConBackColor;
        XChangeGC(ez_Display, gc, 
              GCStipple|GCFunction|GCForeground, &values);
    }
    XDrawImageString(ez_Display, winDataPtr->wd_titleBarWin,
               gc, titleBarPtr->tb_x, titleBarPtr->tb_y,
               textPtr, strlen (textPtr));
    }
}


