
/* ezXEvent.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlot/ezXEvent.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "ezPlot.h"
#include "ezPlotInt.h"


void
Ez_EmptyXEventQue()
/*
 * Process all of the events that are currently in the queue.
 */
{
    XEvent report;

    XSync(ez_Display, FALSE);
    while (XPending(ez_Display)) {
        XNextEvent(ez_Display, &report);
        Ez_XEventHandler(&report);
    }
}


/*
 * ----------------------------------------------------------------------------
 *
 * Ez_XEventHandler() --
 *
 *   This is event handler has been designed to be used with the XMenu package
 * and as such its only arguement is a pointer to the XEvent structure.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

void
Ez_XEventHandler(reportPtr)
register XEvent *reportPtr;
{
    Rect                subWinBnd;
    Window              root, child; /* unused */
    WindowData          *winDataPtr;
    XExposeEvent        *expEventPtr;
    XConfigureEvent     *confEventPtr;
    XButtonEvent        *bEventPtr;
    double              (*invScrMatrix)[4];
    int                 xtop, ybot, winx, winy;
    int                 xRootWin, yRootWin; /* relative to the 'RootWindow' */
    int                 scrDeltaX, scrDeltaY;
    int                 wldDeltaX, wldDeltaY;
    unsigned int        keys; /* unused */

#ifdef DEBUG
    if (ezDebug) {
        EzPrintErr("in Ez_XEventHandler()");
        (void) sprintf (ezErrorMsg, "event type = 0x%x", reportPtr->type);
        EzPrintErr(ezErrorMsg);
    }
#endif DEBUG

    XtDispatchEvent(reportPtr);
    switch (reportPtr->type) {
    case ButtonPress:
        bEventPtr = (XButtonEvent *) &(reportPtr->xbutton);
        if (XQueryPointer(ez_Display, EZ_ROOT_WINDOW,
            &root, &child, &winx, &winy, 
            &xRootWin, &yRootWin, &keys) == 0) {
            /* Will probably never fail but ... */
            xRootWin = DisplayWidth(ez_Display, DefaultScreen(ez_Display))/2;
            yRootWin = DisplayHeight(ez_Display, DefaultScreen(ez_Display))/2;
        }
        else {
            if (EZ_ROOT_WINDOW != root)  {
                xRootWin = winx;
                yRootWin = winy;
            }
        }
        if ((winDataPtr = EzLookUpWin(bEventPtr->window)) != 
            NIL(WindowData)) {

            if (bEventPtr->window != winDataPtr->wd_userWin) {
                return;
            }
        } 
        else {
            /*
             * If this event is for a window that has been destroyed
             * ignor it.
             */
            return;
        }
        if (winDataPtr->wd_userWinButtonProc != (void (*)()) NULL) {
            (*winDataPtr->wd_userWinButtonProc)(winDataPtr, bEventPtr->button,
                   xRootWin, yRootWin,
                   (int) bEventPtr->x, (int) bEventPtr->y);
        }
        break;

     case ConfigureNotify:
         /*
          * Resize the base window.
          */
         confEventPtr = (XConfigureEvent *) reportPtr;
         if ((winDataPtr = EzLookUpWin(confEventPtr->window)) != 
            NIL(WindowData)) {

             if (confEventPtr->window == winDataPtr->wd_baseWin) {
                 scrDeltaX = confEventPtr->width -
                     winDataPtr->wd_baseWinInfo.width;
                 scrDeltaY = confEventPtr->height -
                     winDataPtr->wd_baseWinInfo.height;
                 EzConfigureSubWins (confEventPtr->width,
                             confEventPtr->height, winDataPtr);
                 
                 if (XGetWindowAttributes(ez_Display, winDataPtr->wd_baseWin,
                     &winDataPtr->wd_baseWinInfo) == 0) {
                     /* will probably never fail ... */
                     EzPrintFatalErr("Cannot get base window information");
                 }
                 if (XGetWindowAttributes(ez_Display, winDataPtr->wd_userWin,
                     &winDataPtr->wd_userWinInfo) == 0) {
                     /* will probably never fail ... */
                     EzPrintFatalErr("Cannot get user window information");
                 }
                 invScrMatrix = winDataPtr->wd_invScrMatrix;
                 wldDeltaX = (Ez_XTrans(invScrMatrix,
                            ABS(scrDeltaX), 0) -
                          Ez_XTrans(invScrMatrix, 0, 0)) / 2.0;
                 wldDeltaY = (Ez_YTrans(invScrMatrix, 0,
                            ABS(scrDeltaY)) -
                          Ez_YTrans(invScrMatrix, 0, 0)) / 2.0;
                 if (scrDeltaX < 0) {
                     winDataPtr->wd_imageBnd.r_xbot += ABS(wldDeltaX);
                     winDataPtr->wd_imageBnd.r_xtop -= ABS(wldDeltaX);
                 } 
                 else {
                     winDataPtr->wd_imageBnd.r_xbot -= ABS(wldDeltaX);
                     winDataPtr->wd_imageBnd.r_xtop += ABS(wldDeltaX);
                 }
                 if (scrDeltaY < 0) {
                     winDataPtr->wd_imageBnd.r_ybot += ABS(wldDeltaY);
                     winDataPtr->wd_imageBnd.r_ytop -= ABS(wldDeltaY);
                 } 
                 else {
                     winDataPtr->wd_imageBnd.r_ybot -= ABS(wldDeltaY);
                     winDataPtr->wd_imageBnd.r_ytop += ABS(wldDeltaY);
                 }
                 (void) Ez_CalcViewingTrans (winDataPtr);
             }
             break;

         case Expose:
             expEventPtr = (XExposeEvent *) reportPtr;
             if ((winDataPtr = EzLookUpWin(expEventPtr->window)) != 
                NIL(WindowData)) {

                if (expEventPtr->window == winDataPtr->wd_userWin) {
                    invScrMatrix = winDataPtr->wd_invScrMatrix;
                    subWinBnd.r_xbot = Ez_XTrans(invScrMatrix,
                        expEventPtr->x, expEventPtr->y) + 0.5;
                    subWinBnd.r_ytop = Ez_YTrans(invScrMatrix,
                        expEventPtr->x, expEventPtr->y) + 0.5;
                    
                    xtop = expEventPtr->x + expEventPtr->width;
                    ybot = expEventPtr->y + expEventPtr->height;
                    subWinBnd.r_xtop = Ez_XTrans(invScrMatrix, xtop, ybot)
                        + 0.5;
                    subWinBnd.r_ybot = Ez_YTrans(invScrMatrix, xtop, ybot)
                        + 0.5;
                    if (winDataPtr->wd_redrawProc != (void (*)()) NULL) {
                       (*winDataPtr->wd_redrawProc)(winDataPtr, &subWinBnd);
                    }
                }
                else if (expEventPtr->window == winDataPtr->wd_titleBarWin) {
                    /*
                     * Redraw the title bar sub window.
                     */
                    EzRedrawTitleBar(winDataPtr);
                } 
                else if (expEventPtr->window == winDataPtr->wd_userWin) {
                    /*
                     * Redraw the user sub window.
                     */
                    if (winDataPtr->wd_redrawProc != (void (*)()) NULL) {
                       (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
                    }
                }
            }
        }
        break;

    default:
        break;
    }
    XFlush(ez_Display);
}



