
/* sample.c --
 *
 *     George Carvalho, Esq.
 *     Benjamin Chen
 *     Howdy Dowdy
 *     (Advising Professor - E. S. Kuh)
 *
 * This is a sample of how to use the ezPlot module.
 *
 * Copyright (C) 1991 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/ezPlotTest/sample.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "magic.h"
#include "geometry.h"
#include "bearStatus.h"
#include "ezPlot.h"


#define APPL_NAME   "sample"    /* Used to retrieve the X defaults. */
#define WIN_PROMPT  "SAMPLE"


/* ---------- Local Procedures ---------- */
static BearStatus   Begin();
static BearStatus   Sample();
static void         Usage();

static void         RunCmdLoop();
static void         HelpCmd();
static void         PostScriptCmd();

static void         CloseWin();
static void         RedrawWin();
static void         HandleUserWinButton();


/* ---------- Local Global Variables ---------- */
static Widget       widget;
static char         *progNamePtr;
static bool         loopFlag;

static Pixel dotColor;
static Pixel polyColor;
static Pixel boxColorOne, boxColorTwo;
static Pixel lineColorOne, lineColorTwo, lineColorThree, lineColorFour;
static ColorData colorData[] = {
    { "line.color.one",     &lineColorOne,   &ez_BlackPixel, RED         },
    { "line.color.two",     &lineColorTwo,   &ez_BlackPixel, TEAL        },
    { "line.color.three",   &lineColorThree, &ez_WhitePixel, YELLOW      },
    { "line.color.four",    &lineColorFour,  &ez_BlackPixel, DARK_PURPLE },
 
    { "box.color.one",      &boxColorOne,    &ez_WhitePixel, LAVENDER    },
    { "box.color.two",      &boxColorTwo,    &ez_BlackPixel, MED_BLUE    },

    { "poly.color",         &polyColor,      &ez_BlackPixel, LIME        },
    { "dot.color",          &dotColor,       &ez_BlackPixel, YELLOW      }
};
static int numOfColors = sizeof (colorData) / sizeof (*colorData);

static int cursorType[] = {
    EZ_ARROW_CROSS_CUR,
    EZ_NW_ARROW_CUR,
    EZ_HORZ_ARROWS_CUR,
    EZ_VERT_ARROWS_CUR,
    EZ_CIRCLE_OUTLINE_CUR,
    EZ_SQUARE_OUTLINE_CUR,
    EZ_STOP_WATCH_CUR,
    EZ_CROSS_CUR
};
static int numOfCursors = sizeof (cursorType) / sizeof (*cursorType);


main (argc, argv)
int     argc;
char    **argv;
{
    int     i;
    char    displayName[80];
    char    *geometryPtr = NIL(char);

    if ((progNamePtr = rindex(argv[0], '/')) == (char *) NULL) {
        progNamePtr = argv[0];
    }
    else {
        progNamePtr++;
    }

    displayName[0] = '\0';
    for (i = 1; i < argc ;i++) {
        if ((strcmp (argv[i], "-h") == 0) ||
            (strcmp (argv[i], "-help") == 0) ||
            (strcmp (argv[i], "help") == 0)) {  
            Usage(argv[0]);
        }

        (void) fprintf (stderr, "unknown option %s\n", argv[i]);
        Usage(argv[0]);
    }

    if (Begin(argc, argv) == BEAR_ERROR) {
        exit (-1);
    }
    (void) Sample(geometryPtr);
    Ez_XEnd();
}


static BearStatus
Begin(argc, argv)
int     argc;
char    **argv;
{
    WindowData *winDataPtr;

    if (Ez_Begin(APPL_NAME, &widget, argc, argv) != EZ_OK) {
        return BEAR_ERROR;
    }
/*
    if (Ez_CreateConsoleWin(WIN_PROMPT, &winDataPtr) != EZ_OK) {
        return BEAR_ERROR;
    }
*/
    if (Ez_SetUpColors(colorData, numOfColors) != EZ_OK) {
        return BEAR_ERROR;
    }

    return BEAR_OK;
}


static BearStatus
Sample(geometryPtr)
char *geometryPtr;
{
    Window      win;
    WindowData  *winDataPtr;
    Point       screen;
    Rect        imageBnd;
    XEvent      report;
    char        buffer[80], tbText[80];
    bool        loopFlag = TRUE;

    imageBnd.r_xtop = imageBnd.r_ytop = 1000;
    imageBnd.r_xbot = imageBnd.r_ybot = -1000;
    (void) sprintf (tbText, "Lions and Tigers and Bears, Oh My !!");
    if (Ez_CreateWin(WIN_PROMPT, geometryPtr, tbText,
        RedrawWin, &imageBnd, &winDataPtr) != EZ_OK) {
        return BEAR_ERROR;
    }
    winDataPtr->wd_closeWinProc = CloseWin;
    winDataPtr->wd_userWinButtonProc = HandleUserWinButton;
    winDataPtr->wd_clientData = (ClientData) "COWABUNGA DUDE !!";

    XSync(ez_Display, TRUE);

/*      
 *  while (TRUE) {
 *      XNextEvent(ez_Display, &report);
 *      Ez_XEventHandler(&report);
 *
 *      Ez_GetLine(&win, &screen, buffer, 80, &loopFlag);
 *      (void) printf("sample: %s \n", buffer);
 *  }
 *  (* NOTREACHED *)
 */

    RunCmdLoop("sample > ");
}


/* ARGSUSED */
static void
CloseWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
{
    Ez_Print(EZ_NORMAL_FONT, "\nbye bye\n\n");

    (void) Ez_DestroyWin(winDataPtr);
    Ez_EmptyXEventQue();

    loopFlag = FALSE;
}


static void
HandleUserWinButton(winDataPtr, button,
    xRootWin, yRootWin, xUserWin, yUserWin)

WindowData  *winDataPtr;
int         button;
int         xRootWin, yRootWin, xUserWin, yUserWin;
{
    static int  count = 0;
    char        *textPtr;

    if ((count%3) == 0) {
        textPtr = "SNAP   ";
    } 
    else if ((count%3) == 1) {
        textPtr = "CRACKLE";
    }
    else if ((count%3) == 2) {
        textPtr = "POP    ";
    }
    Ez_WorldText(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        -800, -900, textPtr);
    Ez_RingBell();

    count++;
}


/* ARGSUSED */
static void
RedrawWin(winDataPtr, subWinBndPtr)
WindowData  *winDataPtr;
Rect        *subWinBndPtr;
/*
 * For this simple example we will redraw the entire
 * window every time.
 */
{
#define NUM_OF_POINTS 6

    Rect    rect;
    Point   points[NUM_OF_POINTS];
    int     from_xw, from_yw, to_xw, to_yw;
    char    *textPtr;

    rect.r_xtop = rect.r_ytop = 900;
    rect.r_xbot = rect.r_ybot = -50;
    Ez_FillRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &rect, boxColorOne);

    points[0].p_x = 100;
    points[0].p_y = 100;
    points[1].p_x = 200;
    points[1].p_y = 50;
    points[2].p_x = 350;
    points[2].p_y = 125;
    points[3].p_x = 400;
    points[3].p_y = 400;
    points[4].p_x = 125;
    points[4].p_y = 350;
    points[5].p_x = 50;
    points[5].p_y = 200;
    Ez_FillPolygon(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        points, NUM_OF_POINTS, polyColor);

    from_xw = -800;
    from_yw =  900;
    to_xw = -300;
    to_yw =  900;
    Ez_DrawWorldLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 
        EZ_SOLIDLINE2 | EZ_OUTLINEDARROW, GXcopy, lineColorFour, boxColorOne);

    from_xw = -800;
    from_yw =  800;
    to_xw = -300;
    to_yw =  800;
    Ez_DrawWorldLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 
        EZ_SOLIDLINE1 | EZ_SOLIDARROW, GXcopy, boxColorTwo, boxColorOne);

    from_xw = -800;
    from_yw =  700;
    to_xw = -300;
    to_yw =  700;
    Ez_DrawWorldLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 
        EZ_SOLIDLINE2, GXcopy, lineColorOne, boxColorOne);

    from_xw = -800;
    from_yw =  600;
    to_xw = -300;
    to_yw =  600;
    Ez_DrawWorldLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 
        EZ_DASHEDLINE1, GXcopy, lineColorTwo, boxColorOne);

    from_xw = -800;
    from_yw =  500;
    to_xw = -300;
    to_yw =  500;
    Ez_DrawWorldLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 
        EZ_DOTTEDLINE2, GXcopy, lineColorThree, boxColorOne);

    from_xw = 100;
    from_yw = 800;
    to_xw = 800;
    to_yw = 600;
    Ez_DrawFatLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 20, GXcopy, lineColorThree);

    from_xw = 800;
    from_yw = 100;
    to_xw = 600;
    to_yw = 800;
    Ez_DrawFatLine(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        from_xw, from_yw, to_xw, to_yw, 40, GXcopy, boxColorTwo);

    rect.r_xbot = -800;
    rect.r_xtop = -300;
    rect.r_ybot = -100;
    rect.r_ytop =  400;
    Ez_StippleRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &rect, lineColorTwo, ez_CargoNetBitmap, GXcopy);

    rect.r_xbot = -800;
    rect.r_xtop = -300;
    rect.r_ybot = -600;
    rect.r_ytop = -200;
    Ez_FillRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &rect, boxColorTwo);

    textPtr = (char *) winDataPtr->wd_clientData;
    Ez_WorldText(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        -800, -700, textPtr);

    Ez_DrawWorldSmallDot(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        -800, -800, dotColor);        

    rect.r_xbot = 100;
    rect.r_xtop = 600;
    rect.r_ybot = -600;
    rect.r_ytop = -100;
    Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &rect, EZ_SOLIDLINE2, lineColorOne, boxColorOne);

    rect.r_xbot = 200;
    rect.r_xtop = 700;
    rect.r_ybot = -700;
    rect.r_ytop = -200;
    Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &rect, EZ_DASHEDLINE1, lineColorTwo, boxColorOne);

    rect.r_xbot = 300;
    rect.r_xtop = 800;
    rect.r_ybot = -800;
    rect.r_ytop = -300;
    Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &rect, EZ_DOTTEDLINE2, lineColorThree, boxColorOne);

    Ez_Flush();

#undef NUM_OF_POINTS
}


static void
RunCmdLoop(userMsgPtr)
char *userMsgPtr;
{
#define MAX_LINE_SIZE 128

    Point       screen;
    Window      win;
    WindowData  *winDataPtr;
    int         i = 0;
    char        line[MAX_LINE_SIZE], token[MAX_LINE_SIZE];

    HelpCmd();
    loopFlag = TRUE;
    while (loopFlag == TRUE) {
        Ez_Print(EZ_NORMAL_FONT, userMsgPtr);
        (void) Ez_GetLine(&win, &screen, line, MAX_LINE_SIZE, &loopFlag);
    
        if ((winDataPtr = Ez_GetWinDataOfWin(win)) == NIL(WindowData)) {
            continue;
        }

        (void) sscanf (line, "%1s", token);
        if (strcmp(token, "p") == 0) {
            PostScriptCmd(winDataPtr);
        }
        else if (strcmp(token, "c") == 0) {
            Ez_DefineCursor(winDataPtr->wd_userWin, cursorType[i]);
            i = (i + 1) % numOfCursors;
        }
        else if (strcmp(token, "r") == 0) {
            Point   llwPoint;
            Rect    rect;

            if (Ez_PickRect(winDataPtr, &llwPoint, &rect) == EZ_OK) {
                (void) fprintf (stdout,
                    "\nxb = %d  xt = %d  yb = %d  yt = %d\n\n",
                    rect.r_xbot, rect.r_xtop,
                    rect.r_ybot, rect.r_ytop);
                (void) fflush (stdout);
            }
        }
        else if (strcmp(token, "h") == 0) {
            HelpCmd();
        }
        else if ((index(token, EOF) != 0) || (strcmp(token, "q") == 0)) {
            if (winDataPtr->wd_closeWinProc == (void (*)()) NULL) {
                (void) fprintf (stderr,
                    "RunCmdLoop() : close window procedure is missing.");
            }
            else {
                (*winDataPtr->wd_closeWinProc)(winDataPtr, NIL(CloseWinData));
            }
        }
    }

#undef MAX_LINE_SIZE
}


static void
HelpCmd()
{
    Ez_Print(EZ_NORMAL_FONT, 
        "Be sure mouse is in the sample window while typing commands.\n\n");

    Ez_Print(EZ_NORMAL_FONT, "Commands:\n\n");
    Ez_Print(EZ_NORMAL_FONT, "    h     print this message\n");
    Ez_Print(EZ_NORMAL_FONT, "    c     change the mouse cursor\n");
    Ez_Print(EZ_NORMAL_FONT, 
        "    p     write window out in postscript format\n");
    Ez_Print(EZ_NORMAL_FONT, "    r     pick a rectangle\n");
    Ez_Print(EZ_NORMAL_FONT, "    q     quit\n");
    Ez_Print(EZ_NORMAL_FONT, "    ^L    redraw the window\n");
    Ez_Print(EZ_NORMAL_FONT, "\n");
    Ez_Print(EZ_NORMAL_FONT,
      "    Also each time a mouse button is pressed, a message is printed.\n");
    Ez_Print(EZ_NORMAL_FONT, "\n");
}


static void
PostScriptCmd(winDataPtr)
WindowData *winDataPtr;
{
    double   tx, ty, sx, sy, r;
    char     *fileNamePtr;

    tx = 0.0;
    ty = 10.5;
    sx = 0.0139;
    sy = 0.0139;
    r = 180.0;
    fileNamePtr = "sample.ps";
    if (Ez_OpenPsFile(fileNamePtr,
        winDataPtr->wd_userWinInfo.width,
        winDataPtr->wd_userWinInfo.height,
        tx, ty, sx, sy, r) != EZ_OK) {

        return;
    }

    (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));

    Ez_ClosePsFile();

    (void) fprintf (stdout,
        "\nThe postscript description is in a file named \"%s\"\n\n",
        fileNamePtr);
    fflush (stdout);
}


static void
Usage(progNamePtr)
char *progNamePtr;
{
    (void) fprintf (stderr, "Usage: %s [-options ...]\n\n", progNamePtr);

    (void) fprintf (stderr, "where options include:\n");
    (void) fprintf (stderr,
        "    -help                        print out this message\n");

    exit (-1);
}


