
/* fpgGenRooms.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/fpg/fpgGenRooms.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "fpgInt.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"


typedef struct _LocCellData {
    Fpg_Room        *roomPtr;
    bool            count;
} LocCellData;

typedef struct _LocRoomData {
    PlaneData       *planeDataPtr;
    FloorPlanGraph  *fpgPtr;
} LocRoomData;


/* ---------- Local Procedures ---------- */
static Fpg_Room     *MakeRoom();
static BearStatus   LocateRoom();
static MagicStatus  LocateCellInst();
static BearStatus   GenRoomList();
static BearStatus   GenMainRoom();


static Fpg_Room *
MakeRoom()
{
    Fpg_Room *roomPtr;

    MALLOC(Fpg_Room *, roomPtr, sizeof (Fpg_Room));
    MALLOC(GenericCell *, roomPtr->gcellPtr, sizeof (GenericCell));
    roomPtr->gcellPtr->unType = UNDEF_CELL;
    roomPtr->wallList[FPG_NORTH] = (lsList) NULL;
    roomPtr->wallList[FPG_EAST] = (lsList) NULL;
    roomPtr->wallList[FPG_SOUTH] = (lsList) NULL;
    roomPtr->wallList[FPG_WEST] = (lsList) NULL;

    return (roomPtr);
}


static MagicStatus
LocateCellInst(tilePtr, locCellPtr)
Tile        *tilePtr;
LocCellData *locCellPtr;
{
    CellInst    *cellInstPtr;
    DummyCell   *dummyCellPtr;

    if (Tp_IsSpaceTile(tilePtr)) {
        return (MAGIC_CONTINUE_SEARCH);
    }
    else if (Tp_IsCellInstTile(tilePtr)) {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        cellInstPtr->roomPtr = locCellPtr->roomPtr;

        locCellPtr->roomPtr->gcellPtr->unType = CELL_INST;
        locCellPtr->roomPtr->gcellPtr->unValue.cellInstPtr = cellInstPtr;
        locCellPtr->count++;
    }
    else if (Tp_IsDCellTile(tilePtr)) {
        dummyCellPtr = Tp_GetDCellOfTile(tilePtr);
        dummyCellPtr->roomPtr = locCellPtr->roomPtr;

        locCellPtr->roomPtr->gcellPtr->unType = DUMMY_CELL;
        locCellPtr->roomPtr->gcellPtr->unValue.dummyCellPtr = dummyCellPtr;
        locCellPtr->count++;
    }
    else {
        Bear_PrintErr("LocateCellInst() : unknown tile.");
        return (MAGIC_STOP_SEARCH);
    }

    return (MAGIC_CONTINUE_SEARCH);
}

  
static BearStatus
LocateRoom(junctPtr, locRoomPtr)
WallJunction    *junctPtr;
LocRoomData     *locRoomPtr;
{
    lsList      wallList;
    Fpg_Room    *roomPtr;
    LocCellData locCell;
    Point       *pointPtr;
    WallSegment *segPtr;

    if ((Fpg_GetNorthSegOfWallJunct(junctPtr) != NIL(WallSegment)) &&
        (Fpg_GetEastSegOfWallJunct(junctPtr) != NIL(WallSegment))) {

        roomPtr = MakeRoom();
        lsNewEnd(locRoomPtr->fpgPtr->roomList, (lsGeneric) roomPtr, LS_NH);

        junctPtr->roomPtr = roomPtr;

        wallList = roomPtr->wallList[FPG_WEST] = lsCreate();
        while (TRUE) {
            segPtr = Fpg_GetNorthSegOfWallJunct(junctPtr);
            lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

            junctPtr = Fpg_GetNWJunctOfWallSeg(segPtr);
            if (Fpg_GetEastSegOfWallJunct(junctPtr) != NIL(WallSegment)) {
                pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
                roomPtr->rect.r_xbot = pointPtr->p_x;
                roomPtr->rect.r_ytop = pointPtr->p_y;
                break;
            }
        }
        wallList = roomPtr->wallList[FPG_NORTH] = lsCreate();
        while (TRUE) {
            segPtr = Fpg_GetEastSegOfWallJunct(junctPtr);
            lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

            junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
            if (Fpg_GetSouthSegOfWallJunct(junctPtr) != NIL(WallSegment)) {
                pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
                roomPtr->rect.r_xtop = pointPtr->p_x;
                break;
            }
        }
        wallList = roomPtr->wallList[FPG_EAST] = lsCreate();
        while (TRUE) {
            segPtr = Fpg_GetSouthSegOfWallJunct(junctPtr);
            lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

            junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
            if (Fpg_GetWestSegOfWallJunct(junctPtr) != NIL(WallSegment)) {
                pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
                roomPtr->rect.r_ybot = pointPtr->p_y;
                break;
            }
        }
        wallList = roomPtr->wallList[FPG_SOUTH] = lsCreate();
        while (TRUE) {
            segPtr = Fpg_GetWestSegOfWallJunct(junctPtr);
            lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

            junctPtr = Fpg_GetNWJunctOfWallSeg(segPtr);
            if (Fpg_GetNorthSegOfWallJunct(junctPtr) != NIL(WallSegment)) {
                break;
            }
        }

        locCell.roomPtr = roomPtr;
        locCell.count = 0;
        if (TiSrArea(NIL(Tile), Tp_GetHorzPlane(locRoomPtr->planeDataPtr),  
            &roomPtr->rect, LocateCellInst,
            (ClientData) &locCell) == MAGIC_STOP_SEARCH) {

            return (BEAR_STOP_SEARCH);
        }
        if (locCell.count > 1) {
            Bear_PrintErr("LocateRoom() : too many objects in the room.");
            return (BEAR_STOP_SEARCH);
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
GenRoomList(cellMasterPtr)
CellMaster *cellMasterPtr;
{
    LocRoomData locRoom;

    locRoom.planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    locRoom.fpgPtr = Tp_GetFpg(locRoom.planeDataPtr);
    if (locRoom.fpgPtr == NIL(FloorPlanGraph)) {
        Bear_PrintErr("GenRoomList() : no floor plan graph.");
        return (BEAR_ERROR);
    }

    locRoom.fpgPtr->roomList = lsCreate();
    if (Fpg_ItrWallJunctSeq(locRoom.fpgPtr, LocateRoom,
        (ClientData) &locRoom) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}


static BearStatus
GenMainRoom(cellMasterPtr)
CellMaster *cellMasterPtr;
{
    lsList         wallList;
    CellInst       *coreInstPtr;
    Fpg_Room       *mainRoomPtr;
    FloorPlanGraph *fpgPtr;
    PlaneData      *planeDataPtr;
    Point          *pointPtr;
    WallJunction   *junctPtr;
    WallSegment    *segPtr;

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    if (Fpg_IsFpgBuilt(planeDataPtr) == FALSE) {
        Bear_PrintErr("GenMainRoom() : no floor plan graph.");
        return (BEAR_ERROR);
    }
    fpgPtr = Tp_GetFpg(planeDataPtr);
    junctPtr = Fpg_GetLLJunctOfFpg(fpgPtr);

    junctPtr = Fpg_GetLLJunctOfFpg(fpgPtr);

    mainRoomPtr = MakeRoom();
    fpgPtr->mainRoomPtr = mainRoomPtr;

    wallList = mainRoomPtr->wallList[FPG_WEST] = lsCreate();
    while (TRUE) {
        segPtr = Fpg_GetNorthSegOfWallJunct(junctPtr);
        lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

        junctPtr = Fpg_GetNWJunctOfWallSeg(segPtr);
        if (Fpg_GetNorthSegOfWallJunct(junctPtr) == NIL(WallSegment)) {
            pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
            mainRoomPtr->rect.r_xbot = pointPtr->p_x;
            mainRoomPtr->rect.r_ytop = pointPtr->p_y;

            fpgPtr->ulJunctPtr = junctPtr;
            break;
        }
    }
    wallList = mainRoomPtr->wallList[FPG_NORTH] = lsCreate();
    while (TRUE) {
        segPtr = Fpg_GetEastSegOfWallJunct(junctPtr);
        lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

        junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
        if (Fpg_GetEastSegOfWallJunct(junctPtr) == NIL(WallSegment)) {
            pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
            mainRoomPtr->rect.r_xtop = pointPtr->p_x;

            fpgPtr->urJunctPtr = junctPtr;
            break;
        }
    }
    wallList = mainRoomPtr->wallList[FPG_EAST] = lsCreate();
    while (TRUE) {
        segPtr = Fpg_GetSouthSegOfWallJunct(junctPtr);
        lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

        junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
        if (Fpg_GetSouthSegOfWallJunct(junctPtr) == NIL(WallSegment)) {
            pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
            mainRoomPtr->rect.r_ybot = pointPtr->p_y;

            fpgPtr->lrJunctPtr = junctPtr;
            break;
        }
    }
    wallList = mainRoomPtr->wallList[FPG_SOUTH] = lsCreate();
    while (TRUE) {
        segPtr = Fpg_GetWestSegOfWallJunct(junctPtr);
        lsNewEnd(wallList, (lsGeneric) segPtr, LS_NH);

        junctPtr = Fpg_GetNWJunctOfWallSeg(segPtr);
        if (Fpg_GetWestSegOfWallJunct(junctPtr) == NIL(WallSegment)) {
            break;
        }
    }

    coreInstPtr = Mac_GetInstOfCoreMaster(cellMasterPtr);
    coreInstPtr->roomPtr = mainRoomPtr;

    mainRoomPtr->gcellPtr->unType = CELL_INST;
    mainRoomPtr->gcellPtr->unValue.cellInstPtr = coreInstPtr;

    return (BEAR_OK);
}


void
FpgFreeRoom(item)
lsGeneric item;
{
    Fpg_Room    *roomPtr = (Fpg_Room *) item;
    int         i;

    if (roomPtr != NIL(Fpg_Room)) {
        FREE((char *) roomPtr->gcellPtr);
        for (i = 0; i < 4; i++) {
            if (roomPtr->wallList[i] != (lsList) NULL) {
                lsDestroy(roomPtr->wallList[i], (void (*)()) NULL);
            }
        }
        FREE((char *) roomPtr);
    }
}


BearStatus
Fpg_GenRooms(cellMasterPtr)
CellMaster *cellMasterPtr;
{
    if (GenRoomList(cellMasterPtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }
    if (GenMainRoom(cellMasterPtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}


