
/* fpgItr.c --
 *
 *     George Carvalho, Esq.    (sequential interators)
 *     Wei-Ming Dai             (recursive interators)
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/fpg/fpgItr.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "fpgInt.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"

/* ---------- Local Procedures ---------- */
static BearStatus ItrWallJunctInt();
static BearStatus ItrWallSegInt();


static BearStatus 
ItrWallJunctInt(junctPtr, funcPtr, oldItrMark, clientData)
register WallJunction   *junctPtr;
BearStatus              (**funcPtr)();
int                     oldItrMark;
ClientData              clientData;
{
    WallJunction    *northJunctPtr;
    WallJunction    *southJunctPtr;
    WallJunction    *eastJunctPtr;
    WallJunction    *westJunctPtr;

    if (*funcPtr != (BearStatus (*)()) NULL) 
    {
        if ((**funcPtr)(junctPtr, clientData) == BEAR_STOP_SEARCH)
        {
            *funcPtr = (BearStatus (*)()) NULL;
        }
    }

    /*
     * Mark the junction with the opposite of 'oldItrMark' so
     * that we know that we've been here.
     */
    junctPtr->itrMark = (oldItrMark == BEAR_MARKED) ?
        BEAR_UNMARKED : BEAR_MARKED;

    if ((northJunctPtr = Fpg_GetNorthJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (northJunctPtr->itrMark == oldItrMark) 
        {
            (void) ItrWallJunctInt(northJunctPtr, funcPtr, oldItrMark,
                clientData);
        }
    }

    if ((southJunctPtr = Fpg_GetSouthJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (southJunctPtr->itrMark == oldItrMark) 
        {
            (void) ItrWallJunctInt(southJunctPtr, funcPtr, oldItrMark, 
                clientData);
        }
    }

    if ((westJunctPtr = Fpg_GetWestJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (westJunctPtr->itrMark == oldItrMark) 
        {
            (void) ItrWallJunctInt(westJunctPtr, funcPtr, oldItrMark,
                clientData);
        }
    }

    if ((eastJunctPtr = Fpg_GetEastJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (eastJunctPtr->itrMark == oldItrMark) 
        {
            (void) ItrWallJunctInt(eastJunctPtr, funcPtr, oldItrMark,
                clientData);
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus 
ItrWallSegInt(junctPtr, func, oldItrMark, clientData)

register WallJunction   *junctPtr;
BearStatus              (*func)();
int                     oldItrMark;
ClientData              clientData;
{
    WallJunction    *northJunctPtr;
    WallJunction    *southJunctPtr;
    WallJunction    *eastJunctPtr;
    WallJunction    *westJunctPtr;
    WallSegment     *eastSegPtr;
    WallSegment     *southSegPtr;

    if ((eastSegPtr = Fpg_GetEastSegOfWallJunct(junctPtr)) !=
        NIL(WallSegment))
    {
        if ((*func)(eastSegPtr, clientData) == BEAR_STOP_SEARCH)
        {
            return (BEAR_STOP_SEARCH);
        }
    }
    
    if ((southSegPtr = Fpg_GetSouthSegOfWallJunct(junctPtr)) !=
        NIL(WallSegment))
    {
        if ((*func)(southSegPtr, clientData) == BEAR_STOP_SEARCH)
        {
            return (BEAR_STOP_SEARCH);
        }
    }

    /*
     * Mark the junction with the opposite of 'oldItrMark' so
     * that we know that we've been here.
     */
    junctPtr->itrMark = (oldItrMark == BEAR_MARKED) ?
                BEAR_UNMARKED : BEAR_MARKED;

    if ((northJunctPtr = Fpg_GetNorthJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (northJunctPtr->itrMark == oldItrMark) 
        {
            if (ItrWallSegInt(northJunctPtr, func, oldItrMark, clientData) 
                == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    if ((southJunctPtr = Fpg_GetSouthJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (southJunctPtr->itrMark == oldItrMark) 
        {
            if (ItrWallSegInt(southJunctPtr, func, oldItrMark, clientData) 
                == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    if ((westJunctPtr = Fpg_GetWestJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (westJunctPtr->itrMark == oldItrMark) 
        {
            if (ItrWallSegInt(westJunctPtr, func, oldItrMark, clientData) 
                == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
        }
    }


    if ((eastJunctPtr = Fpg_GetEastJunctOfWallJunct(junctPtr)) !=
        NIL(WallJunction)) 
    {
        if (eastJunctPtr->itrMark == oldItrMark) 
        {
            if (ItrWallSegInt(eastJunctPtr, func, oldItrMark, clientData) 
                == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Fpg_ItrWallJunct() --
 *
 * NOTE THIS PROCEDURE USES A RECURSIVE ALGORITHUM.
 *
 * Visit each wall junction in the floor plan graph, and apply the given
 * procedure to it.  The procedure should be of the following form:
 *
 *  BearStatus
 *  func(junctPtr, clientData)
 *  WallJunction    *junctPtr;
 *  ClientData      clientData;
 *  {
 *  }
 *
 * Func normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Fpg_ItrWallJunct(llJunctPtr, func, clientData)
WallJunction    *llJunctPtr;     /* The lower left wall junction in
                                  * the floor plan.
                                  */
BearStatus      (*func)();
ClientData      clientData;
{
    BearStatus (**funcPtr)();

#ifdef DEBUG
    if (llJunctPtr == NIL(WallJunction)) {
    FpgPrintFatalErr("Fpg_ItrWallJunct() failed: no floor plan graph is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
        FpgPrintFatalErr("Fpg_ItrWallJunct() failed: no func is given");
    }
#endif DEBUG

    funcPtr = &func;
    (void) ItrWallJunctInt(llJunctPtr, funcPtr, llJunctPtr->itrMark,
        clientData);
    return ((*funcPtr == (BearStatus (*)()) NULL) ? BEAR_STOP_SEARCH :
        BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Fpg_ItrWallSeg() --
 *
 * NOTE THIS PROCEDURE USES A RECURSIVE ALGORITHUM.
 *
 * Visit each wall segment in the floor plan graph, and apply the given
 * procedure to it.  The procedure should be of the following form:
 *
 *  BearStatus
 *  func(segPtr, clientData)
 *  WallSegment *segPtr;
 *  ClientData  clientData;
 *  {
 *  }
 *
 * Func normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Fpg_ItrWallSeg(llJunctPtr, func, clientData)
WallJunction    *llJunctPtr;     /* The lower left wall junction in
                                  * the floor plan.
                                  */
BearStatus      (*func)();
ClientData      clientData;
{
#ifdef DEBUG
    if (llJunctPtr == NIL(WallJunction)) {
    FpgPrintFatalErr("Fpg_ItrWallSeg() failed: no floor plan graph is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
        FpgPrintFatalErr("Fpg_ItrWallSeg() failed: no func is given");
    }
#endif DEBUG

    if (ItrWallSegInt(llJunctPtr, func, llJunctPtr->itrMark, clientData) ==
        BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    } 
    else {
        return (BEAR_COMPLETE_SEARCH);
    }
}



/*
 * ----------------------------------------------------------------------------
 *
 * Fpg_ItrWallJunctSeq() --
 *
 * Visit each wall junction in the floor plan graph, and apply the given
 * procedure to it.  The procedure should be of the following form:
 *
 *  BearStatus
 *  func(junctPtr, clientData)
 *  WallJunction    *junctPtr;
 *  ClientData      clientData;
 *  {
 *  }
 *
 * Func normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Fpg_ItrWallJunctSeq(fpgPtr, func, clientData)
FloorPlanGraph  *fpgPtr;  
BearStatus      (*func)();
ClientData      clientData;
{
    WallJunction    ***junctTable, **junctArray, *junctPtr;
    register int    r, c, rmax, cmax;

#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
FpgPrintFatalErr("Fpg_ItrWallJunctSeq() failed: no floor plan graph is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
        FpgPrintFatalErr("Fpg_ItrWallJunctSeq() failed: no func is given");
    }
#endif DEBUG

    if (fpgPtr->junctTable == (WallJunction ***) NULL) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        rmax = fpgPtr->numHorzDomTiles;
        cmax = fpgPtr->numVertDomTiles;
        junctTable = fpgPtr->junctTable;
        for (r = 0; r < rmax ;r++) {
            junctArray = junctTable[r];
            for (c = 0; c < cmax ;c++) {
                junctPtr = junctArray[c];
                if (junctPtr != NIL(WallJunction)) {
                    if (((*func)(junctPtr, clientData)) ==
                        BEAR_STOP_SEARCH) {
                        return (BEAR_STOP_SEARCH);
                    }
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}

    

/*
 * ----------------------------------------------------------------------------
 *
 * Fpg_ItrWallSegSeq() --
 *
 * Visit each wall segment in the floor plan graph, and apply the given
 * procedure to it.  The procedure should be of the following form:
 *
 *  BearStatus
 *  func(segPtr, clientData)
 *  WallSegment *segPtr;
 *  ClientData  clientData;
 *  {
 *  }
 *
 * Func normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Fpg_ItrWallSegSeq(fpgPtr, func, clientData)
FloorPlanGraph  *fpgPtr;  
BearStatus      (*func)();
ClientData      clientData;
{
    WallJunction    ***junctTable, **junctArray, *junctPtr;
    WallSegment     *eastSegPtr, *southSegPtr;
    register int    r, c, rmax, cmax;

#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
FpgPrintFatalErr("Fpg_ItrWallSegSeq() failed: no floor plan graph is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
        FpgPrintFatalErr("Fpg_ItrWallSegSeq() failed: no func is given");
    }
#endif DEBUG

    if (fpgPtr->junctTable == (WallJunction ***) NULL) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        rmax = fpgPtr->numHorzDomTiles;
        cmax = fpgPtr->numVertDomTiles;
        junctTable = fpgPtr->junctTable;
        for (r = 0; r < rmax ;r++) {
            junctArray = junctTable[r];
            for (c = 0; c < cmax ;c++) {
                junctPtr = junctArray[c];
                if (junctPtr != NIL(WallJunction)) {
                    if ((eastSegPtr = Fpg_GetEastSegOfWallJunct(junctPtr)) !=
                        NIL(WallSegment)) {
                        if ((*func)(eastSegPtr, clientData) ==
                            BEAR_STOP_SEARCH) {
                            return (BEAR_STOP_SEARCH);
                        }
                    }
                    
                    if ((southSegPtr = Fpg_GetSouthSegOfWallJunct(junctPtr)) !=
                        NIL(WallSegment)) {
                        if ((*func)(southSegPtr, clientData) ==
                            BEAR_STOP_SEARCH) {
                            return (BEAR_STOP_SEARCH);
                        }
                    }
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}

    

/*
 * ----------------------------------------------------------------------------
 *
 * Fpg_ItrBottleNeckOfFpg() --
 *
 * Visit each bottleneck (tile or line) in the floor plan graph, and apply
 * the given procedure to it.  The procedure should be of the following form:
 *
 *  BearStatus
 *  func(bnPtr, clientData)
 *  BottleNeck  *bnPtr;
 *  ClientData  clientData;
 *  {
 *  }
 *
 * Func normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Fpg_ItrBottleNeckOfFpg(fpgPtr, func, clientData)
FloorPlanGraph  *fpgPtr;  
BearStatus      (*func)();
ClientData      clientData;
{
    WallJunction    ***junctTable, **junctArray, *junctPtr;
    WallSegment     *eastSegPtr, *southSegPtr;
    register int    r, c, rmax, cmax;

#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) 
    {
FpgPrintFatalErr("Fpg_ItrBottleNeckOfFpg() failed: no floor plan graph is given");
    } 
    else if (func == (BearStatus (*)()) NULL) 
    {
        FpgPrintFatalErr("Fpg_ItrBottleNeckOfFpg() failed: no func is given");
    }
#endif DEBUG

    if (fpgPtr->junctTable == (WallJunction ***) NULL) {
        return (BEAR_COMPLETE_SEARCH);
    } else {
        rmax = fpgPtr->numHorzDomTiles;
        cmax = fpgPtr->numVertDomTiles;
        junctTable = fpgPtr->junctTable;
        for (r = 0; r < rmax ;r++) {
            junctArray = junctTable[r];
            for (c = 0; c < cmax ;c++) {
                junctPtr = junctArray[c];
                if (junctPtr != NIL(WallJunction)) {
                    if ((eastSegPtr = Fpg_GetEastSegOfWallJunct(junctPtr)) !=
                        NIL(WallSegment)) {
                        if ((*func)(Fpg_GetBottleNeckOfWallSeg(eastSegPtr),
                            clientData) == BEAR_STOP_SEARCH) {
                            return (BEAR_STOP_SEARCH);
                        }
                    }
                    
                    if ((southSegPtr = Fpg_GetSouthSegOfWallJunct(junctPtr)) !=
                        NIL(WallSegment)) {
                        if ((*func)(Fpg_GetBottleNeckOfWallSeg(southSegPtr),
                            clientData) == BEAR_STOP_SEARCH) {
                            return (BEAR_STOP_SEARCH);
                        }
                    }
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}

    

