
/* fpgPrim.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/fpg/fpgPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "fpgInt.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"


WallJunction *
Fpg_GetLLJunctOfFpg(fpgPtr)
FloorPlanGraph *fpgPtr;
{
#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_GetLLJunctOfFpg() failed");
    }
#endif DEBUG

    return (fpgPtr->llJunctPtr);
}



bool
Fpg_IsFpgBuilt(planeDataPtr)
PlaneData *planeDataPtr;
/*
 * Is the floor plan graph built.
 */
{
    FloorPlanGraph *fpgPtr;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        FpgPrintFatalErr("Fpg_IsFpgBuilt() failed");
    }
#endif DEBUG

    fpgPtr = Tp_GetFpg(planeDataPtr);
    if (fpgPtr == NIL(FloorPlanGraph)) {
        return (FALSE);
    }
    return ((Fpg_GetLLJunctOfFpg(fpgPtr) != NIL(WallJunction)) ? TRUE : FALSE);
}



int
Fpg_GetNumOfWallJunctsOfFpg(fpgPtr)
FloorPlanGraph *fpgPtr;
{
#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_GetNumOfWallJunctsOfFpg() failed");
    }
#endif DEBUG

    return (fpgPtr->numJuncts);
}



int
Fpg_GetNumOfWallSegsOfFpg(fpgPtr)
FloorPlanGraph *fpgPtr;
{
#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_GetNumOfWallSegsOfFpg() failed");
    }
#endif DEBUG

    return (fpgPtr->numSegs);
}



BearId
Fpg_GetIdOfFpg(fpgPtr)
FloorPlanGraph *fpgPtr;
{
#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_GetIdOfFpg() failed");
    }
#endif DEBUG

    return (fpgPtr->id);
}



ClientData
Fpg_GetClientDataOfFpg(fpgPtr)
FloorPlanGraph *fpgPtr;
{
#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_GetClientDataOfFpg() failed");
    }
#endif DEBUG

    return (fpgPtr->clientData);
}



void
Fpg_PutClientDataOfFpg(clientData, fpgPtr)
ClientData     clientData;
FloorPlanGraph *fpgPtr;
{
#ifdef DEBUG
    if (fpgPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_PutClientDataOfFpg() failed");
    }
#endif DEBUG

    fpgPtr->clientData = clientData;
}



void
Fpg_MarkWallJunct(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_MarkWallJunct() failed");
    }
#endif DEBUG

    junctPtr->mark = BEAR_MARKED;
}



void
Fpg_UnmarkWallJunct(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_UnmarkWallJunct() failed");
    }
#endif DEBUG

    junctPtr->mark = BEAR_UNMARKED;
}



bool
Fpg_IsWallJunctMarked(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_IsWallJunctMarked() failed");
    }
#endif DEBUG

    return (junctPtr->mark == BEAR_MARKED);
}



Point *
Fpg_GetPointOfWallJunct(whichPoint, junctPtr)
int             whichPoint;
WallJunction    *junctPtr;
{
#ifdef DEBUG
    if (whichPoint != FPG_NW && 
        whichPoint != FPG_CENTER &&
        whichPoint != FPG_SE) {
        FpgPrintFatalErr("Fpg_GetPointOfWallJunct() failed : no such point");
    }
    else if (junctPtr == NIL(WallJunction)) {
       FpgPrintFatalErr("Fpg_GetPointOfWallJunct() failed : no junction given");
    }
#endif DEBUG

    switch (whichPoint) {
    case FPG_NW:
        return (&junctPtr->nwPoint);
    case FPG_CENTER:
        return (&junctPtr->centerPoint);
    case FPG_SE:
        return (&junctPtr->sePoint);
    }
    /* NOTREACHED */
}



void
FpgPutPointOfWallJunct(whichPoint, pointPtr, junctPtr)
int             whichPoint;
Point           *pointPtr;
WallJunction    *junctPtr;
{
#ifdef DEBUG
    if (whichPoint != FPG_NW && 
        whichPoint != FPG_CENTER &&
        whichPoint != FPG_SE) {
        FpgPrintFatalErr("Fpg_GetPointOfWallJunct() failed : no such point");
    } else if (pointPtr == NIL(Point)) {
        FpgPrintFatalErr("Fpg_GetPointOfWallJunct() failed : no point given");
    } else if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetPointOfWallJunct() failed : no junction given");
    }
#endif DEBUG

    switch (whichPoint) {
    case FPG_NW:
        junctPtr->nwPoint.p_x = pointPtr->p_x;
        junctPtr->nwPoint.p_y = pointPtr->p_y;
        break;
    case FPG_CENTER:
        junctPtr->centerPoint.p_x = pointPtr->p_x;
        junctPtr->centerPoint.p_y = pointPtr->p_y;
        break;
    case FPG_SE:
        junctPtr->sePoint.p_x = pointPtr->p_x;
        junctPtr->sePoint.p_y = pointPtr->p_y;
        break;
    }
}



int
Fpg_GetDegreeOfWallJunct(junctPtr)
WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetDegreeOfWallJunct() failed");
    }
#endif DEBUG

    return (junctPtr->degree);
}



void
FpgPutDegreeOfWallJunct(degree, junctPtr)
int          degree;
WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("FpgPutDegreeOfWallJunct() failed");
    }
#endif DEBUG

    junctPtr->degree = degree;
}



WallSegment *
Fpg_GetNorthSegOfWallJunct(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetNorthSegOfWallJunct() failed");
    }
#endif DEBUG

    return (junctPtr->wallSegPtr[FPG_NORTH]);
}



WallSegment *
Fpg_GetSouthSegOfWallJunct(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetSouthSegOfWallJunct() failed");
    }
#endif DEBUG

    return (junctPtr->wallSegPtr[FPG_SOUTH]);
}



WallSegment *
Fpg_GetWestSegOfWallJunct(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetWestSegOfWallJunct() failed");
    }
#endif DEBUG

    return (junctPtr->wallSegPtr[FPG_WEST]);
}



WallSegment *
Fpg_GetEastSegOfWallJunct(junctPtr)
register WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetEastSegOfWallJunct() failed");
    }
#endif DEBUG

    return (junctPtr->wallSegPtr[FPG_EAST]);
}



ClientData
Fpg_GetClientDataOfWallJunct(junctPtr)
WallJunction *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetClientDataOfWallJunct() failed");
    }
#endif DEBUG

    return (junctPtr->clientData);
}

void
Fpg_PutClientDataOfWallJunct(clientData, junctPtr)
ClientData      clientData;
WallJunction    *junctPtr;
{
#ifdef DEBUG
    if (junctPtr == NIL(WallJunction)) {
        FpgPrintFatalErr("Fpg_GetClientDataOfWallJunct() failed");
    }
#endif DEBUG

    junctPtr->clientData = clientData;
}



WallJunction *
Fpg_GetWallJunctWhereDomTilesCross(fpGraphPtr, horzDomTilePtr, vertDomTilePtr)
FloorPlanGraph  *fpGraphPtr;
Tile            *horzDomTilePtr, *vertDomTilePtr;
{
    int horzIndex, vertIndex;

#ifdef DEBUG
    if (fpGraphPtr == NIL(FloorPlanGraph)) {
        FpgPrintFatalErr("Fpg_GetWallJunctWhereDomTilesCross() failed 1");
    } else if (horzDomTilePtr == NIL(Tile)) {
        FpgPrintFatalErr("Fpg_GetWallJunctWhereDomTilesCross() failed 2");
    } else if (vertDomTilePtr == NIL(Tile)) {
        FpgPrintFatalErr("Fpg_GetWallJunctWhereDomTilesCross() failed 3");
    }
#endif DEBUG

    if ((Tp_IsHorzTile(horzDomTilePtr) == FALSE) ||
        (Tp_IsDominantTile(horzDomTilePtr) == FALSE) ||

        (Tp_IsVertTile(vertDomTilePtr) == FALSE) ||
        (Tp_IsDominantTile(vertDomTilePtr) == FALSE)) {

        return (NIL(WallJunction));
    }

    horzIndex = Tp_GetJunctTableIndexOfDomTile(horzDomTilePtr);
    vertIndex = Tp_GetJunctTableIndexOfDomTile(vertDomTilePtr);

    return (fpGraphPtr->junctTable[horzIndex][vertIndex]);
}



void
Fpg_MarkWallSeg(wallSegPtr)
register WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
        FpgPrintFatalErr("Fpg_MarkWallSeg() failed");
    }
#endif DEBUG

    wallSegPtr->mark = BEAR_MARKED;
}



void
Fpg_UnmarkWallSeg(wallSegPtr)
register WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
        FpgPrintFatalErr("Fpg_UnmarkWallSeg() failed");
    }
#endif DEBUG

    wallSegPtr->mark = BEAR_UNMARKED;
}



bool
Fpg_IsWallSegMarked(segPtr)
register WallSegment *segPtr;
{
#ifdef DEBUG
    if (segPtr == NIL(WallSegment)) {
        FpgPrintFatalErr("Fpg_IsWallSegMarked() failed");
    }
#endif DEBUG

    return (segPtr->mark == BEAR_MARKED);
}



BottleNeck *
Fpg_GetBottleNeckOfWallSeg(wallSegPtr)
WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
        FpgPrintFatalErr("Fpg_GetBottleNeckOfWallSeg() failed");
    }
#endif DEBUG

    return (wallSegPtr->bnPtr);
}



ClientData
Fpg_GetClientDataOfWallSeg(wallSegPtr)
WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
        FpgPrintFatalErr("Fpg_GetClientDataOfWallSeg() failed");
    }
#endif DEBUG

    return (wallSegPtr->clientData);
}

void
Fpg_PutClientDataOfWallSeg(clientData, wallSegPtr)
ClientData  clientData;
WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
        FpgPrintFatalErr("Fpg_GetClientDataOfWallSeg() failed");
    }
#endif DEBUG

    wallSegPtr->clientData = clientData;
}



void
FpgPutBnLineOfWallSeg(bnLinePtr, wallSegPtr)
BNLine      *bnLinePtr;
WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if ((bnLinePtr == NIL(BNLine)) || (wallSegPtr == NIL(WallSegment))) {
        FpgPrintFatalErr("FpgPutBnLineOfWallSeg() failed");
    }
#endif DEBUG

    wallSegPtr->bnPtr->unType = TP_BN_LINE;
    wallSegPtr->bnPtr->unValue.bnLinePtr = bnLinePtr;
}



void
FpgPutBnTileOfWallSeg(bnTilePtr, wallSegPtr)
Tile        *bnTilePtr;
WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if ((bnTilePtr == NIL(Tile)) || (wallSegPtr == NIL(WallSegment))) {
        FpgPrintFatalErr("FpgPutBnTileOfWallSeg() failed");
    }
#endif DEBUG

    wallSegPtr->bnPtr->unType = TP_BN_TILE;
    wallSegPtr->bnPtr->unValue.bnTilePtr = bnTilePtr;
}



WallJunction *
Fpg_GetSEJunctOfWallSeg(wallSegPtr)
register WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
      FpgPrintFatalErr("Fpg_GetSEJunctOfWallSeg() failed: no segment is given");
    }
#endif DEBUG

    return (wallSegPtr->SEJunctPtr);
}



WallJunction *
Fpg_GetNWJunctOfWallSeg(wallSegPtr)
register WallSegment *wallSegPtr;
{
#ifdef DEBUG
    if (wallSegPtr == NIL(WallSegment)) {
      FpgPrintFatalErr("Fpg_GetNWJunctOfWallSeg() failed: no segment is given");
    }
#endif DEBUG

    return (wallSegPtr->NWJunctPtr);
}



void
FpgInsertWallSeg(NWJunctPtr, SEJunctPtr, segmentDir, segmentPtr)
register WallJunction *NWJunctPtr, *SEJunctPtr;
SegmentDir segmentDir;
register WallSegment *segmentPtr;
{
#ifdef DEBUG
    if ((NWJunctPtr == NIL(WallJunction)) ||
        (SEJunctPtr == NIL(WallJunction)) ||
        (segmentPtr == NIL(WallSegment))) {
        FpgPrintFatalErr("FpgInsertWallSeg() failed");
    }
#endif DEBUG

    segmentPtr->NWJunctPtr = NWJunctPtr;
    segmentPtr->SEJunctPtr = SEJunctPtr;
    segmentPtr->direction = segmentDir;
    if (segmentDir == FPG_HORIZONTAL) {
        NWJunctPtr->wallSegPtr[FPG_EAST] = segmentPtr;
        SEJunctPtr->wallSegPtr[FPG_WEST] = segmentPtr;
    } else {
        NWJunctPtr->wallSegPtr[FPG_SOUTH] = segmentPtr;
        SEJunctPtr->wallSegPtr[FPG_NORTH] = segmentPtr;
    }
    NWJunctPtr->degree++;
    SEJunctPtr->degree++;
}



int
Fpg_OppositeDir(direction)
int direction; 
{
    switch (direction) {
    case FPG_NORTH:
        return (FPG_SOUTH);
    case FPG_SOUTH:
        return (FPG_NORTH);
    case FPG_WEST:
        return (FPG_EAST);
    case FPG_EAST:
        return (FPG_WEST);
    default:
        FpgPrintFatalErr("Fpg_OppositeDir() failed");
        break;
    }
    /* NOTREACHED */
}



int
Fpg_MacPinDirToFpgDir(pinDir)
PinDirection pinDir;
{
    switch (pinDir) {
    case NORTH_PIN:
        return (FPG_NORTH);
    case EAST_PIN:
        return (FPG_EAST);
    case SOUTH_PIN:
        return (FPG_SOUTH);
    case WEST_PIN:
        return (FPG_WEST);
    default:
        FpgPrintFatalErr("Fpg_MacPinDirToFpgDir() failed");
        break;
    }
    /* NOTREACHED */
}



PinOrJunct *
Fpg_MakePinOrJunct()
{
    PinOrJunct *pinOrJunctPtr;

    MALLOC(PinOrJunct *, pinOrJunctPtr, sizeof (PinOrJunct));
    pinOrJunctPtr->bnTilePtr = NIL(Tile);
    pinOrJunctPtr->genericPtr = (BearGeneric) NULL;

    return (pinOrJunctPtr);
}



PseudoJunct *
Fpg_MakePseudoJunct()
{
    PseudoJunct *pjunctPtr;

    MALLOC(PseudoJunct *, pjunctPtr, sizeof (PseudoJunct));
    pjunctPtr->id = 0;
    pjunctPtr->clientData = (ClientData) NULL;

    return (pjunctPtr);
}



