
/* grBnHole.c --
 *
 *    Malgorzata Marek-Sadowska
 *    (Advising Professor - E. S. Kuh)
 *
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grBnHole.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint




#include "grInt.h"


/*
*---------------------------------------------------------------------------
*   Functions in this file perform various tasks related to filling
*   in bottleneck tiles after assignment at a cutline has been
*   made.
*--------------------------------------------------------------------------
*/



/*
*--------------------------------------------------------------------------
*   Gr_CollectDataAfterAssignment(kPtr)
*   Results       : none.
*   Side effects :                                                     
*   If there was a hole in the cut line then this function             
*   puts all the data decided during the assignment process into the   
*   proper positions in the data structures. Action depends on net     
*   activity.                                                          
*-------------------------------------------------------------------------
*/


void Gr_CollectDataAfterAssignment(kPtr)

grRegion *kPtr;


{

    grSubNet *sNet,*previous;
    int dir;
    int d1,d2,cnt,i;
    int pom1,pom2,hn;
    grPrimit *xwhois,*ywhois,*xwhere,*ywhere;
    grVirtNet *vNet,*cNet;
    grHole *gHole;
    float total;


    dir=kPtr->dir;
    /* get position of the cut */

    sNet=kPtr->subNetsPtr;

    while (sNet!=NULL)
    {
        if (sNet->active==7)  /*more than 1 virtual net */
        {
            Gr_CombineSuperNodesOfSubNet(sNet);
            Gr_MergeVNetsAssignedToSameHoles(sNet,dir);
        }
        if ((sNet->active==7)||(sNet->active==6))
        {
            Gr_CombineSuperNodesOfSubNet(sNet);

        }
        if ((sNet->active==5)||(sNet->active==6)||(sNet->active==7))
        {

            if (gr_Timing ==1)
            {
                /*subtract approx length*/
                ((gr_HeadGrNets + sNet->netname)->currentLength) -=
                    sNet->approxLength;
                ((gr_HeadGrNets + sNet->netname)->xapprox) -=
                    sNet->xapprox;
                ((gr_HeadGrNets + sNet->netname)->yapprox) -=
                    sNet->yapprox;
                /* end subtraction */

                /* clear holes  */
                for (i=1;i<=n;i++)
                {
                    gHole= (grHeadHolePtr+i);
                    gHole->leftEst=0;
                    gHole->rightEst=0;
                    gHole->lbpins=0;
                    gHole->rtpins=0;
                }
            }

            vNet=sNet->children;

            while (vNet!=NULL)
            {
                Gr_UpdateBoundingBox(vNet->leftNode->
                    contentsPtr);
                Gr_UpdateBoundingBox(vNet->rightNode->
                    contentsPtr);
                d1=vNet->leftNode->contentsPtr->count;
                d2=vNet->rightNode->contentsPtr->count;
                cnt=d1+d2;
                MALLOC(grPrimit*,xwhois,
                sizeof (grPrimit)*(cnt+1));
                MALLOC(grPrimit*,ywhois,
                sizeof (grPrimit)*(cnt+1));
                MALLOC(grPrimit*,xwhere,
                sizeof (grPrimit)*(cnt+1));
                MALLOC(grPrimit*,ywhere,
                sizeof (grPrimit)*(cnt+1));
                for (i=1;i<=cnt;i++)
                {
                    (xwhois+i)->number=0;
                    (ywhois+i)->number=0;
                    (xwhere+i)->number=0;
                    (ywhere+i)->number=0;
                }

                Gr_DetermineWhichPinsAreInBtn
                (vNet,xwhois,xwhere,0,dir,1,n);
                Gr_DetermineWhichPinsAreInBtn
                (vNet,ywhois,ywhere,1,dir,1,n);

                /* check if the virtual net 
                vNet has multiple nodes */

                cNet=sNet->children;
                while (cNet!=NULL)
                {
                    if (cNet!=vNet)
                    {
                        if ((vNet->leftNode == 
                        cNet->leftNode)||
                        (vNet->rightNode ==
                         cNet->rightNode))
                        {
                            hn=cNet->hole;
                            for (i=1;i<=cnt;i++)
                            {
                                if ((xwhere+i)->
                                number == hn)
                                    {
                                (xwhere+i)->
                                number = 0;
                                (xwhois+i)->
                                number = 0;
                                }
                                if ((ywhere+i)->
                                number == hn)
                                    {
                                (ywhere+i)->
                                number =0;
                                (ywhois+i)->
                                number = 0;
                                }
                            }
                        }
                    }
                    cNet=cNet->nextInNet;
                }



                Gr_PullOutPinsFromBtnForAssignment
                (kPtr,NIL(grDeadRegion),sNet,vNet,dir,
                xwhois,ywhois,xwhere,ywhere,d1,d2);

                Gr_UpdateBoundingBox
                ((vNet->leftNode->contentsPtr));
                Gr_UpdateBoundingBox
                ((vNet->rightNode->contentsPtr));

                GR_FREE((char*) xwhere);
                GR_FREE((char*) ywhere);
                GR_FREE((char*) xwhois);
                GR_FREE((char*) ywhois);

                vNet=vNet->nextInNet;
            }
            if (gr_Timing  > 0)
            {
                pom1=0;
                pom2=0;
                for (i=1;i<=n;i++)
                {
                gHole=(grHeadHolePtr+i);
                  if (gHole->status ==0)
                  {
                    pom1 +=(gHole->leftEst);
                    pom1 +=(gHole->rightEst);

                    if ((gHole->lbpins)*(gHole->rtpins)>0)
                    {
                    pom2 +=(gHole->height);
                    }
                    else
                    {
                            if ((gHole->lbpins + gHole->rtpins)>0)
                        {
                        pom2 +=(gHole->height)/2;
                        }
                    }
                  }
                    }
                  if (dir == 1)
                  {
                   ((gr_HeadGrNets+sNet->netname)->xfixed) += pom1;
                   ((gr_HeadGrNets+sNet->netname)->yfixed) += pom2;
                   total=pom1+pom2*gr_HorVertWire;
                   ((gr_HeadGrNets + sNet->netname)->currentLength) += total;
                  }
                  if (dir == 2)
                  {
                   ((gr_HeadGrNets+sNet->netname)->xfixed) += pom2;
                   ((gr_HeadGrNets+sNet->netname)->yfixed) += pom1;
                   total=pom2+pom1*gr_HorVertWire;
                   ((gr_HeadGrNets+sNet->netname)->currentLength) +=
                total;
                  }

            }
            Gr_PushDownwardMultipleHoleAssignedNet(kPtr,sNet);
        }
        
        sNet=sNet->nextOnList;
    }
}

