
/* grDropNet.c --
 *
 *     Malgorzata Marek-Sadowska
 *     (Advising Professor - E. S. Kuh)
 *
 * This file contains functions called when a net is pushed from
 * the present level of hierarchy to the lower.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grDropNet.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "grInt.h"


/*
 *--------------------------------------------------------------------------
 *  Gr_PrepareNetForDrop() --
 *
 *  Results      : none
 *  Side effects :
 *  For a sNet which sits on one side of the cut line pins which are co-
 *  vered by the currently active bottlenecks are removed and
 *  pseudopins are set at bottleneck boundaries. 
 *  Net length approximations are updated.
 *--------------------------------------------------------------------------
 */

void
Gr_PrepareNetForDrop(kPtr, sNet, dir, side, bLHole, rTHole)
grRegion    *kPtr;      /* pointer to the parent region              */
grSubNet    *sNet;      /* pointer to a net which will be pushed down*/
int         dir;        /* direction of a cut line                   */
int         side;       /* 0 : left, 1 : right                       */
int         bLHole;     /* bottom available hole                     */
int         rTHole;     /* top available hole                        */
{
    grHole          *gHole;
    grPrimit        *xwhois, *xwhere, *ywhois, *ywhere;
    grSuperNode     *current;
    float           total;
    unsigned int    numOfBytes;
    int             i, cnt, pom1, pom2;

    current = sNet->NodeList;

    cnt = current->contentsPtr->count;
    numOfBytes = (cnt + 1) * sizeof (grPrimit);
    MALLOC(grPrimit *, xwhois, numOfBytes);
    MALLOC(grPrimit *, ywhois, numOfBytes);
    MALLOC(grPrimit *, xwhere, numOfBytes);
    MALLOC(grPrimit *, ywhere, numOfBytes);

    for (i=1;i<=cnt;i++)
    {
        (xwhois+i)->number = 0;
        (ywhois+i)->number = 0;
        (xwhere+i)->number = 0;
        (ywhere+i)->number = 0;
    }

    Gr_DeterminePinsOfSuperNodeToPullOutFromBtn(current,
        xwhois,xwhere,0,dir,bLHole,rTHole);
    Gr_DeterminePinsOfSuperNodeToPullOutFromBtn(current,
        ywhois,ywhere,1,dir,bLHole,rTHole);

    if (gr_Timing ==1)  /* all nets lengths are calculated*/
    {
        for (i=1;i<=n;i++)
        {
            gHole=(grHeadHolePtr+i);
            if (gHole->status == 0)
            {
                gHole->leftEst =0;
                gHole->rightEst=0;
                gHole->lbpins  =0;
                gHole->rtpins  =0;
            }
        }
        for (i=1;i<=cnt;i++)
        {
            if ((xwhois+i)->number ==1)
            {
                if (side == 0)
                {
                    (xwhois+i)->number=3;
                }
                else
                {
                    (xwhois+i)->number=4;
                }
            }
            if ((ywhois+i)->number ==1)
            {
                if (side == 0)
                {
                    (ywhois+i)->number=3;
                }
                else
                {
                    (ywhois+i)->number=4;
                }
            }
        }
    }
    Gr_PullOutPinsOfSuperNodeFromBtn(kPtr,sNet,current,
        xwhois,xwhere,ywhois,ywhere,dir,side);
    if (gr_Timing > 0 )
    {
        pom1=0;
        pom2=0;
        for (i=1;i<=n;i++)
        {
            gHole=(grHeadHolePtr+i);

            if (gHole->status ==0)
            {
                pom1 += (gHole->leftEst);
                pom1 += (gHole->rightEst);
                if ((gHole->lbpins)*(gHole->rtpins) >0)
                {
                    pom2 +=(gHole->height);
                }
                else
                {
                    if ((gHole->lbpins + gHole->rtpins) > 0)
                    {
                        pom2 +=(gHole->height)/2;
                    }
                }
            }
        }
        ((gr_HeadGrNets+sNet->netname)->currentLength) -=
            sNet->approxLength; /*subtract current approximation*/
        ((gr_HeadGrNets+sNet->netname)->xapprox) -=
            sNet->xapprox;
        ((gr_HeadGrNets+sNet->netname)->yapprox) -=
            sNet->yapprox;


        if ((pom1+pom2) >0)
        {
            if (dir==1)
            {
                ((gr_HeadGrNets+sNet->netname)->xfixed) += pom1;
                ((gr_HeadGrNets+sNet->netname)->yfixed) += pom2;
                total=pom1+pom2*gr_HorVertWire;
                ((gr_HeadGrNets+sNet->netname)->currentLength) += total;
            }
            if (dir==2)
            {
            ((gr_HeadGrNets+sNet->netname)->xfixed) += pom2;
                ((gr_HeadGrNets+sNet->netname)->yfixed) += pom1;
                total=pom2+pom1*gr_HorVertWire;
                ((gr_HeadGrNets+sNet->netname)->currentLength) += total;
            }
        }
    }

    Gr_UpdateBoundingBox((current->contentsPtr));

    GR_FREE((char *) xwhere);
    GR_FREE((char *) ywhere);
    GR_FREE((char *) xwhois);
    GR_FREE((char *) ywhois);
}


/*
*---------------------------------------------------------------------------
*   Gr_DeterminePinsOfSuperNodeToPullOutFromBtn
*   Results       :none;
*   Side effects  :
*   pins which sit inside current are marked if they are covered by
*   bottlenecks in the present cutline.
*---------------------------------------------------------------------------
*/


void Gr_DeterminePinsOfSuperNodeToPullOutFromBtn(current,whois,where,par,dir,
    bLHole,tRHole)

grSuperNode *current;       /* pointer to a superNode               */
grPrimit    *whois;     /* vector of marked pin positions       */
grPrimit    *where;     /* vector of hole numbers where pins sit*/
int          par;       /* 0 marks for x-links, 1 for y-links   */
int          dir;       /* direction of cutline                 */
int          bLHole;        /* bot or left-most feasible hole       */
int          tRHole;        /* top or right-most feasible hole      */


{

    int i,j;
    grSortedPin *gSPin;
    grBlock *block;

    for (j=bLHole;j<=tRHole;j++)  /* iterate over available holes */
    {
        if (((grHeadHolePtr+j)->status == 0)||
        ((grHeadHolePtr+j)->status==3))
                 /* only bottlenecks are considered */
        {
            block=(grHeadHolePtr+j)->leftBlockPtr;
            if (block==NULL) 
                block=(grHeadHolePtr+j)->rightBlockPtr;

            i=0;
            if (par==0)
            {
                /* determine for x links*/

                gSPin=current->contentsPtr->firstxPin;
            }
            else
            {
                /* determine for y links */

                gSPin=current->contentsPtr->firstyPin;
            }

            while (gSPin!=NULL)
            {
                i++;
                if (dir==1)
                {
                    if (Gr_IsSortedPinInsideBlock(
                    block,gSPin,0)>0)
                    {
                        (whois+i)->number=1;
                        (where+i)->number=j;
                    }
                }
                if (dir==2)
                {
                    if (Gr_IsSortedPinInsideBlock(
                    block,gSPin,1)>0)
                    {
                        (whois+i)->number=1;
                        (where+i)->number=j;
                    }
                }
                if (par==0)
                    gSPin=gSPin->nextxPin;
                else
                    gSPin=gSPin->nextyPin;
            }
        }
    }
}

/*
*---------------------------------------------------------------------------
*   Gr_PutPseudoPinsToOneSideOfBtn
*   Results      : none;
*   Side effects : pseudo pins are added to appropriate bottlenecks.
*--------------------------------------------------------------------------
*/


void Gr_PutPseudoPinsToOneSideOfBtn(cnt,where,sNet,current,gChain,dir,side)

int          cnt;
grPrimit    *where;     /* Vector of marked bottlenecks        */
grSubNet    *sNet;      /* Pointer to subNet                   */
grSuperNode     *current;   /* Pointer to superNode            */
grChain     *gChain;    /* Vector of BearIds to name new pseudoPins*/
int          dir;       /* cut direction               */
int      side;      /* 0 left, 1 right             */


{
    int i,j;

    for (i=1;i<=n;i++)  /* itr over all holes of the cut */
    {
        for (j=1;j<=cnt;j++)  
        /* itr over all marks for previously present pins*/
            if (((where+j)->number) == i)
            {
                if (side==0)
                 /*insert to the left_top side of btn*/
                {
                    if (dir==1)
                    {
                        Gr_InsertPseudoPinToVertHoleBtn
                        (i,sNet,current,
                        ((gChain+j)->name),0);
                    }
                    else
                    {
                        Gr_InsertPseudoPinToHorzHoleBtn
                        (i,sNet,current,
                        ((gChain+j)->name),0);
                    }
                }
                else
                {
                    /* insert to the right_bottom*/
                    if (dir==1)
                    {
                    Gr_InsertPseudoPinToVertHoleBtn
                    (i,sNet,current,((gChain+j)->name),1);
                    }
                    else
                    {
                    Gr_InsertPseudoPinToHorzHoleBtn
                    (i,sNet,current,((gChain+j)->name),1);
                    }
                }
                goto lab1;
            }
lab1:;
    }
}


/*
*-------------------------------------------------------------------------
*   Gr_InsertPseudoPinToHorzHoleBtn
*   Results       : none;
*   Side effects  : pseudoPins are inserted to a bottleneck
*           corresponding to holeNr.
*-------------------------------------------------------------------------
*/


void Gr_InsertPseudoPinToHorzHoleBtn(holeNr,sNet,sNode,name,side)

int          holeNr;    /* Hole number where to insert pPins    */
grSubNet    *sNet;      /* Pointer to subNet            */
grSuperNode     *sNode;     /* Pointer to super node        */
BearId       name;      /* Internal name for that pPin      */
int          side;      /* 0 ppin inserted on the top side 1
                   at the bottom            */

{

    grBlock *left,*right;
    int xo,ytop,ybot,xleft,xright;
    Tile *tilePtr;
    BNLine  *bnLinePtr;

    left=(grHeadHolePtr+holeNr)->leftBlockPtr;
    right=(grHeadHolePtr+holeNr)->rightBlockPtr;
    xright=(grHeadHolePtr+holeNr)->topOrRight_p_x;
    xleft=(grHeadHolePtr+holeNr)->botOrLeft_p_x;

    if (left!=NULL)
    {
        if (left->persona != 6)
        {
            ybot=left->ypos;
            ytop=ybot+left->height;
            tilePtr=left->tilePtr;
            bnLinePtr=left->bnLinePtr;
        }
        else
        {
            ybot=right->ypos;
            ytop=ybot+right->height;
            tilePtr=right->tilePtr;
            bnLinePtr=right->bnLinePtr;
        }
    }
    else
    {
        ybot=right->ypos;
        ytop=ybot+right->height;
        tilePtr=right->tilePtr;
        bnLinePtr=right->bnLinePtr;
    }

    xo=xleft+(xright-xleft)/2;

    if (side==0)
    {
        if ((grHeadHolePtr+holeNr)->status==0)
        {
            Gr_CreatePseudoPin(sNet,sNode,name,xo,ytop,4,tilePtr);
        }
        if ((grHeadHolePtr+holeNr)->status==3)
        {
            Gr_CreateBnPseudoPin(sNet,sNode,name,xo,ytop,4,tilePtr,
            bnLinePtr);
        }
    }
    else
    {
        if ((grHeadHolePtr+holeNr)->status==0)
        {
            Gr_CreatePseudoPin(sNet,sNode,name,xo,ybot,2,tilePtr);
        }
        if ((grHeadHolePtr+holeNr)->status==3)
        {
            Gr_CreateBnPseudoPin(sNet,sNode,name,xo,ybot,2,tilePtr,
            bnLinePtr);
        }
    }
}
/*
*-------------------------------------------------------------------------
*   Gr_InsertPseudoPinToVertHoleBtn
*   Results       : none;
*   Side effects  : pseudoPins are inserted to a bottleneck
*           corresponding to holeNr.
*-------------------------------------------------------------------------
*/


void Gr_InsertPseudoPinToVertHoleBtn(holeNr,sNet,sNode,name,side)

int          holeNr;    /* Hole number where to insert pPins    */
grSubNet    *sNet;      /* Pointer to subNet            */
grSuperNode     *sNode;     /* Pointer to super node        */
BearId       name;      /* Internal name for that pPin      */
int          side;      /* 0 ppin inserted on the left side
                   1 on the right           */



{

    grBlock *left,*right;
    int yo,ytop,ybot,xleft,xright;
    Tile *tilePtr;
    BNLine *bnLinePtr;

    left=(grHeadHolePtr+holeNr)->leftBlockPtr;
    right=(grHeadHolePtr+holeNr)->rightBlockPtr;
    ytop=(grHeadHolePtr+holeNr)->topOrRight_p_y;
    ybot=(grHeadHolePtr+holeNr)->botOrLeft_p_y;

    if (left!=NULL)
    {
        if (left->persona != 6 )
        {
            xleft=left->xpos;
            xright=xleft+left->width;
            tilePtr=left->tilePtr;
            bnLinePtr=left->bnLinePtr;
        }
        else
        {
            xleft=right->xpos;
            xright=xleft+right->width;
            tilePtr=right->tilePtr;
            bnLinePtr=right->bnLinePtr;
        }
    }
    else
    {
        xleft=right->xpos;
        xright=xleft+right->width;
        tilePtr=right->tilePtr;
        bnLinePtr=right->bnLinePtr;
    }


    yo=ybot+(ytop-ybot)/2;

    if (side==0)
    {
        if ((grHeadHolePtr+holeNr)->status==0)
        {
            Gr_CreatePseudoPin(sNet,sNode,name,xleft,yo,3,tilePtr);
        }
        if ((grHeadHolePtr+holeNr)->status==3)
        {
               Gr_CreateBnPseudoPin(sNet,sNode,name,xleft,yo,3,tilePtr,
            bnLinePtr);
        }
    }
    else
    {
        if ((grHeadHolePtr+holeNr)->status==0)
        {
            Gr_CreatePseudoPin(sNet,sNode,name,xright,yo,1,tilePtr);
        }
        if ((grHeadHolePtr+holeNr)->status==3)
        {
              Gr_CreateBnPseudoPin(sNet,sNode,name,xright,yo,1,tilePtr,
            bnLinePtr);
        }
    }
}

/*
*----------------------------------------------------------------------------
*   Gr_DeletePinsOfSuperNodeWhichSitInBtn
*   Results         : none.
*   Side effests    : marked pins are removed from supernodes and
*   given proper external names.
*----------------------------------------------------------------------------
*/



void Gr_DeletePinsOfSuperNodeWhichSitInBtn(current,xwhois,ywhois,gChain)

grSuperNode *current;   /* Pointer to superNode         */
grPrimit    *xwhois;    /* Vector of marks for x-links      */
grPrimit    *ywhois;    /* Vector of marks for y-links      */
grChain     *gChain;    /* Vector of names to use inside btn    */


{

    int j;
    grSortedPin *pending,*before;


    j=0;

    /* remove along x links and give external names to deleted pins*/
    before=NULL;
    pending=current->contentsPtr->firstxPin;
    while (pending!=NULL)
    {
        j++;
        if (((xwhois+j)->number) > 0 )
        {
            pending->pinPtr->extname = (gChain+j)->name;

            if (gr_BearConnection == 1)
            {
                /* link to bear*/
                Mac_PutExtNetIdOfPin((gChain+j)->name,
                pending->pinPtr->PinPtr.pinInst);
                /* end link to bear*/
            }

            pending=Gr_DeleteSortedPinFromXLinksInPinRectangle
            (current,pending,before);
        }
        else
        {
            before=pending;
            pending=pending->nextxPin;
        }
    }

    /* remove along y links */
    j=0;
    before=NULL;
    pending=current->contentsPtr->firstyPin;
    while (pending!=NULL)
    {
        j++;
        if (((ywhois+j)->number) > 0)
        {
            pending=Gr_DeleteSortedPinFromYLinksInPinRectangle
            (current,pending,before);
        }
        else
        {
            before=pending;
            pending=pending->nextyPin;
        }
    }
}

/*
*---------------------------------------------------------------------------
*   Gr_PullOutPinsOfSuperNodeFromBtn
*   Results       : none;
*   Side effects  : marked pins  are removed and pseudo pins
*   are added instead on btn boundary.
*---------------------------------------------------------------------------
*/



void Gr_PullOutPinsOfSuperNodeFromBtn(kPtr,sNet,current,xwhois,xwhere,ywhois
,ywhere,dir,side)

grRegion    *kPtr;      /* Pointer to the parent region     */
grSubNet    *sNet;      /* Pointer to the subNet        */
grSuperNode     *current;   /* pointer to the super Node        */
grPrimit    *xwhois;    /* Vector of marks along x-links    */
grPrimit    *xwhere;    /* Vector of hole numbers for x-links   */
grPrimit    *ywhois;    /* Vector of marks along y-links    */
grPrimit    *ywhere;    /* Vector of hole numbers along y-links */
int          dir;       /* cut direction            */
int      side;      /* 0 : left; 1: right           */

{
    int i,cnt;
    int differ;
    grChain *gChain;

    differ=0;
    differ=(xwhere+1)->number;
    cnt=current->contentsPtr->count;
    for (i=1;i<=cnt;i++)
    {
        if (differ!=(xwhere+i)->number) differ=0;
    }
    MALLOC(grChain*,gChain,(cnt+1)*sizeof (grChain));
    if (cnt > 0)
    {
        for (i=1;i<=cnt;i++)
        {
            (gChain+i)->name=0;
        }
    }
        


    if (gr_Timing > 0)
    {
        if (differ != 0)
        {
            Gr_AddFixedLengthsForBtnNets(sNet);
        }
        else
        {
            if (dir == 1)
            {
        Gr_UpdateBtnEstimates(kPtr,dir,xwhere,xwhois,current,0,0,0,0);
            }
            if ( dir == 2)
            {
        Gr_UpdateBtnEstimates(kPtr,dir,ywhere,ywhois,current,0,0,0,0);
            }
        }
    }
    Gr_DetermineInternalNamesForBtn(cnt,cnt,0,xwhere,xwhois,gChain);
    Gr_DeletePinsOfSuperNodeWhichSitInBtn(current,xwhois,ywhois,gChain);
    Gr_UpdateBoundingBox((current->contentsPtr));
    if ((current->contentsPtr!= NULL)&&(differ==0))
    {
        Gr_PutPseudoPinsToOneSideOfBtn(cnt,xwhere,sNet,current,
        gChain,dir,side);
        Gr_UpdateBoundingBox((current->contentsPtr));
    }

    GR_FREE((char*) gChain);
}

/*
*----------------------------------------------------------------------------
*   Gr_DetermineInternalNamesForBtn
*   Results      : none.
*   Side effects : determines names for pins covered by btncks
*   and puts them into vector gChain.
*----------------------------------------------------------------------------
*/


void Gr_DetermineInternalNamesForBtn(cnt,lcnt,hNumber,whereis,whois,gChain)

int         cnt;        /* Total number of pins         */
int         lcnt;       /* Number of pins on left hand side */
int         hNumber;    /* Hole number              */
grPrimit    *whereis;   /* Pin marks                */
grPrimit    *whois;     /* Hole numbers             */
grChain     *gChain;    /* Vector of names          */


{
    int i,lflag,rflag,j;
    BearId lnew,rnew;

    for (i=1;i<=cnt;i++)
    {
        (gChain+i)->name=0;
        if (((whois+i)->number == 1)&&(i<=lcnt))
            (whois+i)->number=3;
        if (((whois+i)->number == 1)&&(i>lcnt))
            (whois+i)->number=4;
    }

    for (i=1;i<=n;i++)
    {
        lflag=0;
        rflag=0;
        for (j=1;j<=cnt;j++)
        {
            if (((whereis+j)->number ==i)&&((whois+j)->number==3)
            &&(i!=hNumber))
            {
                if (lflag==0)
                {
                    lflag=1;
                    lnew=Gr_GetNewSubNetName();
                }
                (gChain+j)->name=lnew;
            }
            if (((whereis+j)->number ==i)&&((whois+j)->number==4)
            &&(i!=hNumber))
            {
                if (rflag==0)
                {
                    rflag=1;
                    rnew=Gr_GetNewSubNetName();
                }
                (gChain+j)->name=rnew;
            }
            if (((whereis+j)->number ==i)&&(i==hNumber))
            {
                if ((lflag+rflag)==0)
                {
                    rflag=1;
                    lflag=1;
                    lnew=Gr_GetNewSubNetName();
                }
                (gChain+j)->name=lnew;
            }
        }
    }
}

/*
*--------------------------------------------------------------------------
*   Gr_IsSortedPinInsideBlock
*   Results     : 1 if pin pointed to by gSPin is inside block :
*           par =0 on horz walls, par=1 on vert walls.
*             0 otherwise.
*   Only regular or imaginary pins are considered.
*---------------------------------------------------------------------------
*/


int Gr_IsSortedPinInsideBlock(block,gSPin,par)

grBlock     *block;
grSortedPin     *gSPin;
int          par;       /* 0 : horz walls, 1 vert walls */


/*
 Need change when Imaginary pins will be handled. Instead of a block
 should be hole .
*/

{

    int x1,y1,h,w;
    int x,y;

    if (gSPin->pinPtr->type == 2) return  (0);

    x1=block->xpos;
    y1=block->ypos;
    h=block->height;
    w=block->width;

    x=gSPin->pinPtr->x;
    y=gSPin->pinPtr->y;

    if (par==0)
    {
        if ((y!=y1)&&(y!=(y1+h))) return (0);
        if ((x<x1)||(x>(x1+w)))  return (0);
        return (1);
    }
    else
    {
        if ((x!=x1)&&(x!=(x1+w))) return (0);
        if ((y<y1)||(y>(y1+h))) return (0);
        return (1);
    }
}



