
/* grHole.c --
 *
 *     Malgorzata Marek-Sadowska
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grHole.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "grInt.h"


/*
*-----------------------------------------------------------------------
*   Gr_SetMemoryForHoles()
*   Results     : none.
*   Side effects    : allocates and initializes memory for the 
*             vector of n holes.
*-----------------------------------------------------------------------
*/



void Gr_SetMemoryForHoles()

{
    int i;

    MALLOC(grHole*,grHeadHolePtr,(n+1)*sizeof (grHole));

    for (i=0;i<(n+1);i++)
    {
                (grHeadHolePtr+i)->leftBlockPtr=NULL;
                (grHeadHolePtr+i)->rightBlockPtr=NULL;
                (grHeadHolePtr+i)->signalCapacity=0;
                (grHeadHolePtr+i)->chDensity=0;
                (grHeadHolePtr+i)->netsInHole=NULL;
    }
}

/*
*----------------------------------------------------------------------
*   Gr_GetBlockPtrFromBitValue(bitValue)
*   Results     : pointer to block whose bitValue was specified.
*   Side effects    : none.
*----------------------------------------------------------------------
*/



grBlock *Gr_GetBlockPtrFromBitValue(bitValue)
int bitValue;       /* Cell value in Bitmap */


{
    grBlock *current;

    if (bitValue==0) return (NULL);
    if (bitValue<0) bitValue = -bitValue;
    current=grHeadBlockPtr;
    while (current!=NULL)
    {
        if (current->name==bitValue)  
        {

            return (current);

        }
        current=current->nextBlockPtr;
    }


        return (NULL);
}

/*
*-------------------------------------------------------------------------
*   Gr_VertHoleAffiliation
*   Results     :none.
*   Vertical cut case.
*   Side effects    : fills out proper fields on grHole structures:
*   1.left and right block names for the holes;                         
*   2.bottom and top boundary points;                                    
*   3.determine status of hole 
*       0:regular bottleneck, 1: improper (bottleneck in 
*         wrong direction), 2: no btlnck,
*       3 :btln line
*-------------------------------------------------------------------------
*/


void Gr_VertHoleAffiliation(kPtr)
grRegion *kPtr;     /* Pointer to the parent region         */


{
    int posx,posy1,posy2,j1,cleft,cright,lleft,lright;
    int cnt,startHole;
    grBlock *lBlock,*rBlock;

    posx=kPtr->xcutvirt;
    posy1=kPtr->ycutvirt;
    posy2=posy1+(kPtr->hvirt)-1;
    lleft=0;
    lright=0;
    cnt=0;
    startHole=0;

    for (j1=posy1;j1<=posy2;j1++)
    {
        cleft=Gr_UpLeft(posx,j1);
        cright=Gr_UpRight(posx,j1);

        lBlock=Gr_GetBlockPtrFromBitValue(cleft);
        rBlock=Gr_GetBlockPtrFromBitValue(cright);
        if ((lBlock!=NULL)&&(rBlock!=NULL))
        {
            if ((lBlock->persona == 6)&&(rBlock->persona!= 6))
            {
                lBlock=NULL;
                cleft=0;
            }
            if ((rBlock->persona == 6)&&(lBlock->persona!=6))
            {
                rBlock=NULL;
                cright=0;
            }
        }
        if (cnt==0)  /* no hole yet*/
        {
            if (Gr_IsHorzHole(cleft,cright))
            {
                startHole=1;
                cnt++;
                (grHeadHolePtr+cnt)->leftBlockPtr=
                lBlock;
                (grHeadHolePtr+cnt)->rightBlockPtr=
                rBlock;
                if (lBlock != NULL)
                {
                    if (lBlock->persona==2)
                    {
                        (grHeadHolePtr+cnt)->status=0;
                        (grHeadHolePtr+cnt)->bLBound=
                        lBlock->xpos;
                        (grHeadHolePtr+cnt)->rTBound
                        =(grHeadHolePtr+cnt)->bLBound +
                        lBlock->width;
                    }
                }
    
                if (rBlock != NULL)
                {
                    if (rBlock->persona==2)
                    {
                        (grHeadHolePtr+cnt)->status=0;
                        (grHeadHolePtr+cnt)->bLBound
                        =rBlock->xpos;
                        (grHeadHolePtr+cnt)->rTBound=
                        (grHeadHolePtr+cnt)->bLBound +
                        rBlock->width;
                    }
                }
                if ((rBlock != NULL)&&(lBlock!=NULL))
                {
                    if ((rBlock->persona !=2)&&(lBlock->
                    persona !=2))
                    {
                        if ((lBlock->
                        persona==3)&&(
                        rBlock->persona==3))
                        {
                            (grHeadHolePtr+cnt)->
                            status = 2;
                        }
                        else
                        {
                            (grHeadHolePtr+cnt)->
                            status = 1;
                        }
                    }
                }

                if (lBlock!=NULL)
                {
                    if (lBlock->persona==4)
                    {
                        (grHeadHolePtr+cnt)->status=3;
                    }
                }
                if (rBlock!=NULL)
                {
                    if (rBlock->persona==4)
                    {
                        (grHeadHolePtr+cnt)->status=3;
                    }
                }



                (grHeadHolePtr+cnt)->botOrLeft_p_x=
                (gr_xcoord+posx)->number;
                (grHeadHolePtr+cnt)->botOrLeft_p_y=
                (gr_ycoord+j1)->number;
            }
            else
            {
                startHole=0;
            }
        }
        else
        {
            if ((cleft!=lleft)||(cright!=lright))
            {
                if (startHole==1)
                {
                    startHole=0; /*end previous hole*/
                    (grHeadHolePtr+cnt)->topOrRight_p_x=
                    (gr_xcoord+posx)->number;
                    (grHeadHolePtr+cnt)->topOrRight_p_y=
                    (gr_ycoord+j1)->number;
                }
                if (Gr_IsHorzHole(cleft,cright))
                {
                  /* new hole starts*/
                    startHole=1;
                    cnt++;
                    (grHeadHolePtr+cnt)->leftBlockPtr=
                    lBlock;
                    (grHeadHolePtr+cnt)->rightBlockPtr=
                    rBlock;
                    if (lBlock != NULL)
                    {
                        if (lBlock->persona == 2)
                        {
                        (grHeadHolePtr+cnt)->status=0;
                        (grHeadHolePtr+cnt)->bLBound
                            =lBlock->xpos;
                        (grHeadHolePtr+cnt)->rTBound
                        =(grHeadHolePtr+cnt)->bLBound +
                            lBlock->width;
                        }
                    }
                    if (rBlock != NULL)
                    {
                        if (rBlock->persona == 2)
                        {
                        (grHeadHolePtr+cnt)->status=0;
                        (grHeadHolePtr+cnt)->bLBound
                            =rBlock->xpos;
                        (grHeadHolePtr+cnt)->rTBound=
                        (grHeadHolePtr+cnt)->bLBound +
                            rBlock->width;
                        }
                    }
                    if ((lBlock !=NULL)&&(rBlock != NULL))
                    {
                      if ((lBlock->persona!=2)&&(rBlock->
                      persona!=2))
                      {
                        if ((lBlock->persona==3)&&(rBlock->
                        persona == 3))
                        {
                        (grHeadHolePtr+cnt)->status = 2;
                        }
                        else
                        {
                        (grHeadHolePtr+cnt)->status = 1;
                        }
                      }
                    }

                    if (lBlock != NULL)
                    {
                        if (lBlock->persona==4)
                        {
                        (grHeadHolePtr+cnt)->status=3;
                        }
                    }
                    if (rBlock != NULL)
                    {
                        if (rBlock->persona==4)
                        {
                        (grHeadHolePtr+cnt)->status=3;
                        }
                    }

                    (grHeadHolePtr+cnt)->botOrLeft_p_x=
                    (gr_xcoord+posx)->number;
                    (grHeadHolePtr+cnt)->botOrLeft_p_y=
                    (gr_ycoord+j1)->number;
                }
            }
        }
        lleft=cleft;
        lright=cright;
    }
    if (startHole==1)  /*end the last hole*/
    {
        (grHeadHolePtr+cnt)->topOrRight_p_x=(gr_xcoord+posx)->number;
        (grHeadHolePtr+cnt)->topOrRight_p_y=
        (gr_ycoord+(posy2+1))->number;
        startHole=0;
    }
}

/*
*-------------------------------------------------------------------------
*   Gr_HorzHoleAffiliation
*   Results     :none.
*   Horizontal cut case.
*   Side effects    : fills out proper fields on grHole structures:
*   1.left and right block names for the holes;                         
*   2.bottom and top boundary points;                                    
*   3.determine status of hole 
*       0:regular bottleneck, 1: improper (bottleneck in 
*         wrong direction), 2: no btlnck,
*       3 :btln line
*-------------------------------------------------------------------------
*/


void Gr_HorzHoleAffiliation(kPtr)
grRegion *kPtr;     /* Pointer to the parent region.        */


{
    int posx1,posx2,posy,j1,cleft,cright,lleft,lright;
    int cnt,startHole;
    grBlock *lBlock, *rBlock;

    posx1=kPtr->xcutvirt;
    posx2=posx1+(kPtr->whvirt)-1;
    posy=kPtr->ycutvirt;
    lleft=0;
    lright=0;
    cnt=0;
    startHole=0;

    for (j1=posx1;j1<=posx2;j1++)
    {
        cleft=Gr_UpRight(j1,posy);
        cright=Gr_BotRight(j1,posy);
        lBlock=Gr_GetBlockPtrFromBitValue(cleft);
        rBlock=Gr_GetBlockPtrFromBitValue(cright);
        if ((lBlock!=NULL)&&(rBlock!=NULL))
        {
            if ((lBlock->persona == 6)&&(rBlock->persona!=6))
            {
                lBlock=NULL;
                cleft=0;
            }
            if ((rBlock->persona == 6)&&(lBlock->persona!=6))
            {
                rBlock=NULL;
                cright=0;
            }
        }


        if (cnt==0)  /* no hole yet*/
        {
            if (Gr_IsVertHole(cleft,cright))
            {
                startHole=1;
                cnt++;
                (grHeadHolePtr+cnt)->leftBlockPtr=lBlock;
                (grHeadHolePtr+cnt)->rightBlockPtr=rBlock;
                if (lBlock != NULL)
                {
                    if (lBlock->persona == 3)
                    {
                        (grHeadHolePtr+cnt)->status=0;
                        (grHeadHolePtr+cnt)->bLBound
                        =lBlock->ypos;
                        (grHeadHolePtr+cnt)->rTBound
                        =(grHeadHolePtr+cnt)->bLBound +
                        lBlock->height;
                    }
                }
                if (rBlock != NULL)
                {
                    if (rBlock->persona == 3)
                    {
                        (grHeadHolePtr+cnt)->status=0;
                        (grHeadHolePtr+cnt)->bLBound
                        =rBlock->ypos;
                        (grHeadHolePtr+cnt)->rTBound
                        =(grHeadHolePtr+cnt)->bLBound +
                        rBlock->height;
                    }
                }



                if ((lBlock!=NULL)&&(rBlock!=NULL))
                {
                    if ((lBlock->persona!=3)&&(rBlock->persona
                    !=3))
                    {
                    if ((lBlock->persona==2)&&
                    (rBlock->persona ==2))
                    {
                        (grHeadHolePtr+cnt)->status = 2;
                    }
                    else
                    {
                        (grHeadHolePtr+cnt)->status = 1;
                    }
                    }
                }
                if (lBlock!=NULL)
                {
                    if (lBlock->persona==5)
                    {
                        (grHeadHolePtr+cnt)->status=3;
                    }
                }
                if (rBlock!=NULL)
                {
                    if (rBlock->persona==5)
                    {
                        (grHeadHolePtr+cnt)->status=3;
                    }
                }
                (grHeadHolePtr+cnt)->botOrLeft_p_x
                =(gr_xcoord+j1)->number;
                (grHeadHolePtr+cnt)->botOrLeft_p_y
                =(gr_ycoord+posy)->number;
            }
            else
            {
                startHole=0;
            }
        }
        else
        {
            if ((cleft!=lleft)||(cright!=lright))
            {
                if (startHole==1)
                {
                    startHole=0;  /* end previous hole*/
                    (grHeadHolePtr+cnt)->topOrRight_p_x
                    =(gr_xcoord+j1)->number;
                    (grHeadHolePtr+cnt)->topOrRight_p_y
                    =(gr_ycoord+posy)->number;
                }
                if (Gr_IsVertHole(cleft,cright))
                 /* new hole starts */
                {
                    startHole=1;
                    cnt++;
                    (grHeadHolePtr+cnt)->leftBlockPtr
                    =lBlock;
                    (grHeadHolePtr+cnt)->rightBlockPtr
                    =rBlock;
                    if (lBlock != NULL)
                    {
                        if (lBlock->persona == 3)
                        {
                        (grHeadHolePtr+cnt)->status = 0;
                        (grHeadHolePtr+cnt)->bLBound
                            =lBlock->ypos;
                        (grHeadHolePtr+cnt)->rTBound
                        =(grHeadHolePtr+cnt)->bLBound +
                        lBlock->height;
                        }
                    }
                    if (rBlock != NULL)
                    {
                        if (rBlock->persona == 3)
                        {
                        (grHeadHolePtr+cnt)->status = 0;
                        (grHeadHolePtr+cnt)->bLBound
                            =rBlock->ypos;
                        (grHeadHolePtr+cnt)->rTBound
                        =(grHeadHolePtr+cnt)->bLBound +
                            rBlock->height;
                        }
                    }
                    if ((lBlock!=NULL)&&(rBlock!=NULL))
                    {
                        if ((lBlock->persona!=3)&&(rBlock
                        ->persona!=3))
                        {
                        if ((lBlock->persona==2)&&
                        (rBlock->persona==2))
                        {
                            (grHeadHolePtr+cnt)->
                            status=2;
                        }
                        else
                        {
                            (grHeadHolePtr+cnt)->
                            status=1;
                        }
                        }
                    }
                    if (lBlock !=NULL)
                    {
                        if (lBlock->persona==5)
                        {
                        (grHeadHolePtr+cnt)->status = 3;
                        }
                    }
                    if (rBlock !=NULL)
                    {
                        if (rBlock->persona==5)
                        {
                        (grHeadHolePtr+cnt)->status = 3;
                        }
                    }
                    (grHeadHolePtr+cnt)->botOrLeft_p_x=
                    (gr_xcoord+j1)->number;
                    (grHeadHolePtr+cnt)->botOrLeft_p_y=
                    (gr_ycoord+posy)->number;
                }
            }
        }
        lleft=cleft;
        lright=cright;
    }
    if (startHole==1) /*end the last hole*/
    {
        (grHeadHolePtr+cnt)->topOrRight_p_x=
        (gr_xcoord+(posx2+1))->number;
        (grHeadHolePtr+cnt)->topOrRight_p_y=(gr_ycoord+posy)->number;
        startHole=0;
    }
}


/*
*---------------------------------------------------------------------------
*   Gr_CalculateSignalCapacityForHoles
*   Results     : capacity of a cut.
*   If the total capacity of a cut is
*   less than the number of crossing nets then latter the placement needs
*   to be decompressed by means of overfloving some holes i.e. artificially
*   increasing their capacities. 
*   Side effects     : the capacity vector for the assignment procedure is
*   initially calculated and for each hole the field signalCapacity is 
*   filled.    
*-----------------------------------------------------------------------------
*/



int Gr_CalculateSignalCapacityForHoles(kPtr)

grRegion *kPtr;     /* Pointer to the parent region         */


/*
*----------------------------------------------------------------------------
*   REMARK : capacities need to be modified if prerouted nets are present 
*----------------------------------------------------------------------------
*/


{

    int par,dir,length,i,sum,capacity;


    dir=kPtr->dir;
    sum=0;

    if (dir==1)  /* vertical cut */
    {
        for (i=1;i<=n;i++)
        {
            capacity=0;
            if ((grHeadHolePtr+i)->status == 0)
            {
                if ((grHeadHolePtr+i)->leftBlockPtr != NULL)
                {
                if (((grHeadHolePtr+i)->leftBlockPtr->persona
                     ==2)
                ||((grHeadHolePtr+i)->leftBlockPtr->persona
                     ==3))
                    {
                            capacity=(grHeadHolePtr+i)->
                        leftBlockPtr->avCap;
                    }
                }
                else
                {
                    if ((grHeadHolePtr+i)->
                    rightBlockPtr != NULL)
                    {
                        if (((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==2)
                        ||((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==3))
                        {
                            capacity=(grHeadHolePtr+i)->
                            rightBlockPtr->avCap;
                        }
                    }
                }
            }
            if ((grHeadHolePtr+i)->status == 3)
            {
                if ((grHeadHolePtr+i)->leftBlockPtr != NULL)
                {
                    if (((grHeadHolePtr+i)->
                    leftBlockPtr->persona ==4)
                    ||((grHeadHolePtr+i)->leftBlockPtr->
                    persona ==5))
                    {
                            capacity=(grHeadHolePtr+i)->
                        leftBlockPtr->avCap;
                    }
                }
                else
                {
                    if ((grHeadHolePtr+i)->
                    rightBlockPtr != NULL)
                    {
                        if (((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==4)
                        ||((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==5))
                        {
                            capacity=(grHeadHolePtr+i)->
                            rightBlockPtr->avCap;
                        }
                    }
                }
            }
            if (((grHeadHolePtr+i)->status != 3)&&
            ((grHeadHolePtr+i)->status != 0))
            {

                length=(grHeadHolePtr+i)->
                topOrRight_p_y-(grHeadHolePtr+i)->botOrLeft_p_y;
                capacity=length/gr_horzDesRule;
            }
            par=Gr_ClipExternalHole(i,dir);
            if (par!=0)
            {
                capacity=Gr_CalculateBoundaryCapacity(
                    par,kPtr);
                if (capacity==0)
                {
                    capacity++;
                }
            }

            sum += capacity;
            (grHeadHolePtr+i)->signalCapacity=capacity;
            (grACap+i)->number=capacity;

        }
    }  /* end vertical cut*/

    if (dir==2)  /*horz cut */
    {
        for (i=1;i<=n;i++)
        {
            capacity=0;
            if ((grHeadHolePtr+i)->status == 0)
            {
                if ((grHeadHolePtr+i)->leftBlockPtr != NULL)
                {
                    if (((grHeadHolePtr+i)->
                    leftBlockPtr->persona ==2)
                    ||((grHeadHolePtr+i)->
                    leftBlockPtr->persona ==3))
                    {
                            capacity=(grHeadHolePtr+i)->
                        leftBlockPtr->avCap;
                    }
                }
                else
                {
                    if ((grHeadHolePtr+i)->rightBlockPtr
                     != NULL)
                    {
                        if (((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==2)
                        ||((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==3))
                        {
                            capacity=(grHeadHolePtr+i)->
                            rightBlockPtr->avCap;
                        }
                    }
                }
            }
            if ((grHeadHolePtr+i)->status == 3)
            {
                if ((grHeadHolePtr+i)->leftBlockPtr != NULL)
                {
                    if (((grHeadHolePtr+i)->leftBlockPtr->
                    persona ==4)
                    ||((grHeadHolePtr+i)->leftBlockPtr->
                    persona ==5))
                    {
                            capacity=(grHeadHolePtr+i)->
                        leftBlockPtr->avCap;
                    }
                }
                else
                {
                    if ((grHeadHolePtr+i)->rightBlockPtr 
                    != NULL)
                    {
                        if (((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==4)
                        ||((grHeadHolePtr+i)->
                        rightBlockPtr->persona ==5))
                        {
                            capacity=(grHeadHolePtr+i)->
                            rightBlockPtr->avCap;
                        }
                    }
                }
            }
            if (((grHeadHolePtr+i)->status != 3)&&
            ((grHeadHolePtr+i)->status!=0))
            {
                length=(grHeadHolePtr+i)->topOrRight_p_x
                -(grHeadHolePtr+i)->botOrLeft_p_x;
                capacity=length/gr_vertDesRule;
            }
            par=Gr_ClipExternalHole(i,dir);
            if (par!=0)
            {
                capacity=Gr_CalculateBoundaryCapacity(par,kPtr);
                if (capacity==0)
                {
                    capacity++;
                }
            }
            (grACap+i)->number=capacity;
            sum += capacity;
            (grHeadHolePtr+i)->signalCapacity=capacity;

        }
    }  /* end horz cut */
    return (sum);
}



/*
*----------------------------------------------------------------------------
*   Gr_ClipExternalHole(i,dir)
*   Results     : 0 if hole does not touch external boundary;
*            -1 if hole touches bottom or left side;
*             1 if hole touches top or right side.
*   Side effects    : none.
*---------------------------------------------------------------------------
*/



int Gr_ClipExternalHole(i,dir)

int i;          /* Hole number.             */
int dir;        /* Cut direction.           */

{

    if (dir==1)
    {
        if (i==1)
        {
            if ((grHeadHolePtr+i)->botOrLeft_p_y==
            (gr_ycoord+0)->number)
            {
                return (-1);
            }
        }
        if (i==n)
        {
            if ((grHeadHolePtr+i)->topOrRight_p_y==
            (gr_ycoord+gr_totygrid)->number)
            {
                return (1);
            }
        }
    }
    if (dir==2)
    {
        if (i==1)
        {
            if ((grHeadHolePtr+i)->botOrLeft_p_x==
            (gr_xcoord+0)->number)
            {
                return (-1);
            }
        }
        if (i==n)
        {
            if ((grHeadHolePtr+i)->topOrRight_p_x==
            (gr_xcoord+gr_totxgrid)->number)
            {
                return (1);
            }
        }
    }
            
    return (0);
}

/*
*----------------------------------------------------------------------------
*   Gr_RelaxCapacities()
*   Results     : none.
*   Side effects    : increases signalCapacities for holes proportionally
*   to congestionFactor.
*----------------------------------------------------------------------------
*/

void
Gr_RelaxCapacities()
{
    int capacity, i;
    
    if ((gr_congestionFactor > 0.0) && (n>1))
    {
        for (i=1;i<=n;i++)
        {
            (grHeadHolePtr+i)->signalCapacity *= (1.0 + gr_congestionFactor);
        }
    }
}


/*
*----------------------------------------------------------------------------
*   Gr_DecompressProportionallyCutCapacity
*   Results     : 1 if decompaction occured, 0 otherwise.
*   Side effects    : if sum < m (total number of virtual nets
*   in the assignment) it adds capacity to holes which are
*   proper bottlenecks proportionally to their initial capcity.
*---------------------------------------------------------------------------
*/



int  Gr_DecompressProportionallyCutCapacity(sum)

int sum;        /* Capacity of a cut line. */


{

    int i,exsum,noexp,delta,proc,d;

    if (sum < m)  /* only if sum < m decompression is made */
    {

        exsum=0;
        noexp=0;

        for (i=1;i<=n;i++)
        {
            if (((grHeadHolePtr+i)->status==0)||
            ((grHeadHolePtr+i)->status == 3))
            {
                noexp++;
                exsum += (grHeadHolePtr+i)->signalCapacity;
            }
        }

        if (noexp>0) /* there are bottlenecks which can be overflown */
        {
            delta=m-sum;
            if (exsum > 0)
            {
                proc=(delta*100)/exsum +1;
                exsum=0;
                for (i=1;i<=n;i++)
                {
                    if (exsum<delta)
                        {
                        d=(((grHeadHolePtr+i)->
                        signalCapacity)*proc)/100;
                        exsum += d;
                        (grHeadHolePtr+i)->
                        signalCapacity += d;
                    }
                }
                if (exsum<delta) {
                    (grHeadHolePtr+n)->signalCapacity += (delta-exsum);
                }
                return (1);
            }
            else
            {
                proc=delta/noexp +1;
                for (i=1;i<=n;i++)
                {
                    (grHeadHolePtr+i)->signalCapacity=proc;
                }
                return (1);
            }
        }  /* end overflown bottlenecks */
        return (0); 
        /* there is no such a hole which could be expanded */
    }
    return (1); /* no need to decompress*/
}

/*
*---------------------------------------------------------------------------
*   Gr_CalculateBoundaryCapacity
*   Results : calculates, how many subnets touches the extarnal
*   boundary.
*   Side effects : none.
*---------------------------------------------------------------------------
*/



int Gr_CalculateBoundaryCapacity(par,kPtr)

int       par;      /* Parameter refering to a particular side  */ 
grRegion *kPtr;     /* Pointer to parent region.            */

{

    int dir,count;
    grSubNet *sNet;

    count=0;
    dir=kPtr->dir;

    sNet=kPtr->subNetsPtr;

    while (sNet!=NULL)
    {
        count=count+Gr_IsSubNetAtBoundary(sNet,par,dir);
        sNet=sNet->nextOnList;
    }
    return (count);
}

/*
*--------------------------------------------------------------------------
*   Gr_IsSubNetAtBoundary(sNet,par,dir)
*   Results     : 1 if sNet touches boundary of the bounding box,
*             0 otherwise.
*   Side effects    : none.
*--------------------------------------------------------------------------
*/



int Gr_IsSubNetAtBoundary(sNet,par,dir)

grSubNet *sNet;     /* Pointer to a subnet              */
int       par;      /* Parameter refering to a particular side  */
int       dir;      /* Direction of a cutline           */

{

    int left,right,top,bot;

    left=(gr_xcoord+0)->number;
    right=(gr_xcoord+gr_totxgrid)->number;
    top=(gr_ycoord+gr_totygrid)->number;
    bot=(gr_ycoord+0)->number;

    if (dir==0)
    {
        return (0);
    }

    if (dir ==1)
    {
        if (par==(-1))
        {
            if (sNet->miny==bot)
            {
                return (1);
            }
            else
            {
                return (0);
            }
        }
        if (par==1)
        {
            if (sNet->maxy==top)
            {
                return (1);
            }
            else
            {
                return (0);
            }
        }
    }
    if (dir==2)
    {
        if (par==(-1))
        {
            if (sNet->minx==left)
            {
                return (1);
            }
            else
            {
                return (0);
            }
        }
        if (par==1)
        {
            if (sNet->maxx==right)
            {
                return (1);
            }
            else
            {
                return (0);
            }
        }
    }
    return (0);
}
