/*
 * grLinAssign.c --
 *
 *     Author
 *  Malgorzata Marek-Sadowska
 *
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grLinAssign.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint


/*
*-----------------------------------------------------------------------
*   Functions in this file are for linear assignment problem.
*-----------------------------------------------------------------------
*/ 

#include "grInt.h"
#define printlength  80

/*
*---------------------------------------------------------------------------
*   Gr_PrtSoln() 
*   Results     : none.
*   Side effects    : print outs of the results to the assignment problem
*---------------------------------------------------------------------------
*/



void Gr_PrtSoln() 
{
    int ij, i, j, sum,max, cij;

    sum = 0; 
    max = 0;
    ij = 0;
    printf("\n");
    for (i = 1 ; i <= m ; i++ ) 
    {
        cij = (c +ij +(map+(col + i)->number)->number )->number;
            printf("row %3d   col  %3d   cij  %3d\n",i,(col+i)->number,cij);
        sum = sum + cij;
        if (cij > max) 
        {
            max = cij;
        }
        ij = ij + n;
    }
    printf("\n col after mapping back \n :");
    for (i = 1 ; i <= m ; i++ ) 
    {
            printf(" %4d",(map+(col+i)->number)->number);
    }
    printf("\n");
    printf("Total cost of edges = %d   max cost = %d\n", sum, max);
    printf("\n");
    if ( (n+1)*4 <= printlength) 
    {
        for (j = 1; j <= n; j++) printf("%3d ",(grACap + j)->number);
        printf("\n\n");
        ij = 0;
        for (i = 1; i <= m; i++) 
        {
                for (j = 1; j <= n; j++) 
            {
                    ij = ij + 1;
                if ((map+(col+i)->number)->number == j)
                {
                    printf("[%3d]", (c+ij)->number);
                }
                    else
                {
                    printf("%3d ",(c + ij)->number);
                }
                }
                printf("\n");
        }
        printf("\n");
    }
}
/*
*----------------------------------------------------------------------------
*   Gr_LSAssign ()
*   Results     : none.
*   Side effects    :
*   For a bipartite graph with m sources (i-th row) and n sinks
*   (j-th column) connected by edges with cost  c[i,j] >= 0, an assignment 
*   of rows i  to columns j  is calculated which minimizes the cost
*   z = sum c[i,j] * x[i,j]
*   i,j
*   subject to
*   sum x[i,j] <= cap[j];  sum x[i,j] = 1;  x[i,j] >= 0
*     i                      j
*   x[i,j] = 0 (open) or 1 (edge is connected)
*
*   input:   capacity vector cap[1 .. n] of n sinks.
*    m by n  cost matrix  c[1 .. m*n] stored by rows
*   problemsize  m, n.  m <= sum cap[j]
*                 j
*   CAUTION: cap is destroyed during by running this procedure.
*   output:  two vectors row[1:n], col[1:m] describing rows 
*   and columns of the selected matrix-entries.
*   row[j] points to one of the rows assigned to colum j; others
*   may be accessed by nexts[row[j]] ...
*   scratch: h1[1:m], h2[1:n], h3[1:m], h4[1:n], h5[1:n], h6[1:n]
*   method:  First an optimal assignment for a subset of rows and columns 
*   is constructed heuristically. Then shortest augmenting paths
*   are generated starting from unassigned sourcenodes ending at
*   sinknodes with free capacity, to determine the solution of
*   the complete problem.  For details see burkhard & derigs, 
*   "assignment and matching problems: solution methods with 
*   fortran programs",springer 1980.
*   Problems with m < n are handled as an  n by n  problem with 
*   c[i,j] = highvalue for i > m.
*   Sinknodes j with cap[j] are handled like a bunch of cap[j]
*   nodes with capacity 1 each.
*-----------------------------------------------------------------------------
*/

void Gr_LSAssign()
{
    int i, j, d;

    Gr_InitAssign();

    /* initialize candidate-list for shortestpath */
    (h6 + 1)->number = neff;
    for (j = 2; j <= neff; j++) (h6+j)->number = j-1;

    for (i = 1; i <= m; i++) {
        /* all free sourcenodes */
        if ((col+i)->number == 0) {
            Gr_ShortestPath(h1,h2,h3,h4,h5,h6,i,&j,&d);
            Gr_Augmentation(h5,i,j);
            Gr_Transformation(h1,h2,h3,h4,d);
        }
    }
    Gr_FillHolesInVirtualNetsAfterAssignment();
}


/*
*--------------------------------------------------------------------
*   Gr_Augmentation 
*   Results     : none.
*   Side effects    :
*   tracing back the augmenting path from index back to u, 
*   assignments are updated accordingly.
*--------------------------------------------------------------------
*/


void Gr_Augmentation (back,u,ind) 
grPrimit *back;     /* poiner to the path .         */
int       u;        /* start                */
int       ind;      /* end                  */
{
    int oldind, w;

    (cap+ind)->number = (cap+ind)->number - 1;
    do {
        oldind = ind;
        w      = (back+ind)->number;
        if (w != u) 
        {
                   /*remove edge w,col[w] from list associated
             with col[w]*/
                ind      = (col+w)->number;
                (row+ind)->number = (nexts+w)->number;
        }
        /* add edge w,index to list associated with index */
        (nexts+w)->number    = (row+oldind)->number;
        (row+oldind)->number = w;
        (col+w)->number      = oldind;
    } 
    while (w != u);
}



/*
*--------------------------------------------------------------------------
*   Gr_InitAssign()
*   Results     : none.
*   Side effects    :
*   Initializes, checks feasability and constructs an optimal assignment 
*   for a subset of rows and columns :
*   First we calculate h1[i] = min c[i,j] and  h2[j] = min (c[i,j] - h1[i])
*                             j                      i
*   then "admissible" assignments (i,j) with c[i,j] = h1[i] + h2[j] 
*   are made.
*----------------------------------------------------------------------------
*/



void Gr_InitAssign()
{
    int i,j,ij,j0,ui,cij;
    for (j = 1; j <=n; j++) 
    {
            (row+j)->number = 0;
    }
    /* calculate h1 and make trivial assignments */
    ij = 0;
    for (i = 1; i <= m; i++) 
    {
            (col+i)->number = 0;
            ui     = maxint;
            for (j = 1; j <= n; j++) 
        {
                ij = ij + 1;
                if ((c+ij)->number < ui) 
            {
                        ui = (c+ij)->number;
                        j0 = j;
            }
            }
            (h1+i)->number = ui;
            if ((cap+(mback+j0)->number)->number > 0) 
        { 
            /* trivial assignment */
                (cap+(mback+j0)->number)->number =
            (cap+(mback+j0)->number)->number - 1;
                (nexts+i)->number =
            (row+(mback+j0)->number)->number;
                (row+(mback+j0)->number)->number  = i;
                (col+i)->number   = (mback+j0)->number;
                (h2+(mback+j0)->number)->number  = 0;
            }
    }

    /* calculate h2 */
    if (m < neff) 
    {
        for (j = 1; j <= neff; j++)  
        {
            (h2+j)->number = 0;
        }
    }
    else
    {
        for (j = 1; j <= neff; j++) 
        {
                if ((row+j)->number == 0) 
            {
                    ij = (map+j)->number; 
                ui = maxint;
                    for (i = 1; i <= m; i++ ) 
                {
                        cij = (c+ij)->number - (h1+i)->number;
                        if (cij < ui)
                    {
                         ui = cij;
                    }
                        ij = ij + n;
                    }
                    (h2+j)->number = ui;
                }
        }
    }
    /* make admissible assignments */
    for (i = 1; i <= m; i++) 
    {
        if ((col+i)->number == 0) 
        {
                ui = (h1+i)->number; 
            ij = (i-1)*n;
                for (j = 1; j <= neff; j++) 
            {
                    if ((cap+j)->number > 0) 
                {
                        if ((c+ij+(map+j)->number)->number 
                    == (ui + (h2+j)->number)) 
                    {
                              (cap+j)->number=(cap+j)->number-1;
                              (nexts+i)->number=(row+j)->number;
                              (row+j)->number   = i;
                              (col+i)->number   = j;
                              break;
                        }
                    }
                }
        }
    }
}


/*
*---------------------------------------------------------------------------
*   Gr_Transformation 
*   Results     : none.
*   Side effects    : updating of ys and yt.
*---------------------------------------------------------------------------
*/


void Gr_Transformation(ys,yt,dplus,dminus,d) 

grPrimit *ys,*yt,*dplus,*dminus;
int d;

{
    int i;
    for (i = 1; i <= m; i++) 
    {
            if ((dplus +i)->number != maxint) 
        {
                (ys+i)->number =
            (ys+i)->number + d - (dplus+i)->number;
        }
    }
    for (i = 1; i <= neff; i++) 
    {
            if ((dminus+i)->number <  d)
        {
                (yt+i)->number =
             (yt+i)->number - d + (dminus+i)->number;
        }
    }
}
/*
*-------------------------------------------------------------------------
*   Results     : none.
*   Side effects    :
*   Finds a shortest augmenting path from startnode i to sinknode index
*   using dijkstra's algorithm and modified costs.
*********************************************************/


void Gr_ShortestPath (ys,yt,dplus,dminus,back,next,i,index,d)

grPrimit    *ys,*yt,*dplus,*dminus;
grPrimit    *back;      /* points back to the path      */
grPrimit    *next;
                /*next is a cyclic list of all sinknodes. 
                    next[headcand] ... lastcand 
                    are the active candidate nodes.*/
int          i, *index, *d;

{
    int lastcand,headcand,is,v,vgl,j,dalt,ysi;
    for (j = 1; j <= neff; j++)  
    {
        (dminus+j)->number = maxint; 
    }
    for (j = 1; j <= m; j++)  
    {
        (dplus +j)->number = maxint;
    }
    lastcand = neff;    
    headcand = neff;    
    *d = 0;
    do { /* till sinknode with free capacity is reached */
        (dplus+i)->number = *d; 
        dalt = *d; 
        *d = maxint;
        v = headcand; 
        ysi = (ys+i)->number; 
        is = (i-1)*n;
        do {  /* for all active and feasible sinknodes */
                j = (next+v)->number;
                if (((cap+j)->number > 0) || ((row+j)->number > 0)) 
            {
                    /* update distances */
                    vgl = dalt + (c+is+(map+j)->number)->number 
                -ysi -(yt+j)->number;
                    if (vgl < (dminus+j)->number) 
                {
                        (dminus+j)->number = vgl;
                        (back+j)->number   = i;
                    }
                    /*  select nearest node */
                    if ((dminus+j)->number < *d) 
                {
                        *d     = (dminus+j)->number;
                        (*index) = v;
                    }
                }
                v = j;
        } while (v != lastcand);
        /* remove (next+v)->number from candidate set */
        v = (*index); 
        (*index) = (next+v)->number;
        if ((*index) == lastcand) 
        {
            lastcand = v;
        }
        else 
        {
                (next+v)->number   = (next+(*index))->number;
                (next+(*index))->number = (next+headcand)->number;
                (next+headcand)->number = (*index);
                headcand  = (*index);
        }
        i = (row+(*index))->number;
    } while ((cap+(*index))->number <= 0);
}
