/*
 * grLoop.c --
 *
 *     Author
 *	Malgorzata Marek-Sadowska
 *	(Advising Professor - E. S. Kuh)
 *
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grLoop.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint





#include "grInt.h"

/*
*--------------------------------------------------------------------------
*	This file contains functions for breaking loops in nets
*	on which decomposition was performed. It is a remote
*	possibility, but real.
*	Loops occur when portions of the same net are assigned to 
*	same bottleneck. In such a situation a net has multiple
*	pseudopins on the same side of a bottleneck. Since
*	multiple pseudpins in the same bottleneck are illegal,
*	the functions in this file perform various tasks which
*	in effect delete such illegal pseudopins and perhaps some other 
*	pseudopins. The result is that a net which had loops
*	is clipped to a tree.
*--------------------------------------------------------------------------
*/



/*--------------------------------------------------------------------------
*	Gr_CrackLoops
*       This routine is called after all nets have been routed.
*	Runs only on automatically routed nets on which decompisi-
*	tion has been performed. Loops never occur in nets which
*	were not decomposed.
*       Results      : none.
*       Side effects : nets with loops due to multiple pseudo pins
*	in same bottlenecks are clipped to trees.
*---------------------------------------------------------------------------
*/

void Gr_CrackLoops()

{

	int i;

	for (i=1;i<=gr_NetCount;i++)
	{
		if ((gr_HeadGrNets+i)->noLoosePins > 2)
		{
			while (Gr_BreakLoops(i) !=0)
			{
			}
		}
	}
}


/*--------------------------------------------------------------------------
*	Gr_BreakLoops
*       This routine is called for each routed net to break a single
*	loop due to multiple pseudopins.
*	Results      : 1 if a break occured, 0 if did not.
*	Side effects : loops are removed along with execess pseudo pins.
*---------------------------------------------------------------------------
*/

int Gr_BreakLoops(num)

int num;	/* Internal net number.				*/


{
	grSortedPin *previous,*current;
	int i,count,pos1,pos2,sc1,sc2,pos3,pos4,pom;
	BearId int1Name,int2Name;

	sc1=0;
	sc2=0;
	pos3=0;
	pos4=0;
	previous=(gr_HeadGrNets + num)->firstXPin;
	pos1=1;
	count=1;
	current=previous->nextXPin;
	pos2=2;

	while (current!=NULL)
	{
		count++;
		if (previous->pinPtr->x == current->pinPtr->x)
		{
			if (previous->pinPtr->y == current->pinPtr->y)
			{


				int1Name=previous->pinPtr->intname;
				int2Name=current->pinPtr->intname;
				sc1=1;
				goto lab1;
			}
		}
		pos1=pos2;
		pos2++;
		previous=current;
		current=previous->nextXPin;
	}
	lab1:;

	if (sc1==0)
	{
		/* no multiple pins present */
		return (0);
	}
	/*
	if (num == 172)
	{
		printf("\n Break Loops");
		printf("\n pos1=%4d pos2=%4d",pos1,pos2);
	}
	*/

	count=0;
	current=(gr_HeadGrNets+num)->firstXPin;
	pos3=0;
	pos4=0;
	while (current!=NULL)
	{
		count++;
		if ((current->pinPtr->intname==int1Name)&&
		(current->pinPtr->type!=1)&&(count!=pos1))
		{
			pos3=count;
		}
		if ((current->pinPtr->intname==int2Name)&&
		(current->pinPtr->type!=1)&&(count!=pos2))
		{
			pos4=count;
		}
		current=current->nextXPin;
	}
	if ((pos3!=0)&&(pos4!=0))
	{
		sc2=1;
	}
	else
	{
		sc2=0;
	}
	/*
	if (num == 172)
	{
		printf("\n pos3=%4d pos4=%4d",pos3,pos4);
	}
	*/
	if (pos3==0)
	{
		pos3=pos1;
		pos1=pos2;
		pos2=pos3;
		pos3=pos4;
		pos4=0;
	}
	/*
	if (num == 172)
	{
printf("\n pos1=%4d pos2=%4d pos3=%4d pos4=%4d",pos1,pos2,pos3,pos4);
	}
	*/


	if (sc2==1)
	{
		pom=Gr_IsPairActive(pos1,pos2,pos3,pos4,count,num);
		if (pom>0)
		{
		/*
		if (num ==172)
		{
		printf("\n A");
		}
		*/
			Gr_UndoDoublePair(pos1,pos2,pos3,pos4,pom,count,num);
			return (1);
		}
		else
		{
			pom=Gr_IsPairActive(pos3,pos4,pos1,pos2,count,num);

			if (pom > 0)
			{
		/*
		if (num ==172)
		{
		printf("\n B");
		}
		*/
				Gr_UndoDoublePair(pos3,pos4,pos1,
				pos2,pom,count,num);
				return (1);
			}
			else
			{
				pom=Gr_IsPairActive(pos1,pos4,
				pos2,pos3,count,num);
				if (pom>0)
				{
		/*
		if (num ==172)
		{
		printf("\n C");
		}
		*/
					Gr_UndoDoublePair(pos3,pos4,pos1,
					pos2,pom,count,num);
					return (1);
				}
				else
				{
					pom=Gr_IsPairActive(pos2,pos3,
					pos1,pos4,count,num);
					if (pom>0)
					{
		/*
		if (num ==172)
		{
		printf("\n D");
		}
		*/
					      Gr_UndoDoublePair(pos4,pos3,pos2,
						pos1,pom,count,num);
						return (1);
					}
				}
			}
		}
	}
	else
	{
		if (pos3==0)
		{
		/*
		if (num ==172)
		{
		printf("\n E");
		}
		*/
			pom=Gr_IsPairActive(pos1,pos2,0,0,count,num);
			Gr_UndoDoublePair(pos1,pos2,0,0,pom,count,num);
			return (1);
		}
		else
		{
			pom=Gr_IsPairActive(pos1,pos2,pos3,0,count,num);
			if (pom > 0)
			{
		/*
		if (num ==172)
		{
		printf("\n F");
		}
		*/
				Gr_UndoDoublePair(pos1,pos2,pos3,
				0,pom,count,num);
				return (1);
			}
			else
			{
				pom=Gr_IsPairActive(pos1,pos3,pos2,0,
				count,num);
				if (pom > 0)
				{
		/*
		if (num ==172)
		{
		printf("\n G");
		}
		*/
					Gr_UndoDoublePair(pos1,pos3,pos2,
					0,pom,count,num);
					return (1);
				}
				else
				{
					pom=Gr_IsPairActive(pos2,pos3,pos1,0,
					count,num);
					if (pom > 0)
					{
		/*
		if (num ==172)
		{
		printf("\n H");
		}
		*/
					       Gr_UndoDoublePair(pos2,pos3,pos1,
						0,pom,count,num);
						return (1);
					}
				}

			}	
		}
	}
	return (0);
}

/*-------------------------------------------------------------------------
*	Gr_IsPairActive
*	This routine is called when there are 2 pseudopins
*	of the same net in the same side of a bottleneck.
*	Positions are locations of pins on X-sorted lists of
*	gr_HeadGrNets structures.
*	For a pair of double-pins at positions pos1,pos2 it checks
*	if there is a path between them which does not go through
*	pins at positions pos3,pos4.
*	Results      : 0 - if no such a path exists, >0 otherwise.
*	Side effects : none.
*-------------------------------------------------------------------------
*/


int Gr_IsPairActive(pos1,pos2,pos3,pos4,count,num)

int pos1,pos2,pos3,pos4;	/* Pin positions.		*/
int count;			/* # pinsin net.  		*/
int num;			/* Internal net number.		*/


{
	grSortedPin *gSPin,*runPtr;
	BearId iname,ename;
	grPrimit *marks;
	int i,last,current,j,j1;

	MALLOC(grPrimit*,marks,(count+1)*sizeof (grPrimit));
	for (i=1;i<=count;i++)
	{
		(marks+i)->number=0;
	}
	(marks+pos1)->number =   1;   /* start */
	(marks+pos2)->number = (-1);  /* trget */
	(marks+pos3)->number = (-2);  /* obstacle */
	(marks+pos4)->number = (-2);  /* obstacle */

	last=1;
	current=1;

	/*
	if (num == 172)
	{
	printf("\n Is pair active");
	printf("\n pos1=%4d,pos2=%4d,pos3=%4d,pos4=%4d",pos1,pos2,pos3,pos4);
	}
	*/

	lab1:;
	gSPin = (gr_HeadGrNets+num)->firstXPin;
	j=0;
	while (gSPin!=NULL)
	{
		j++;
		if ((marks+j)->number==last)
		{
			iname=gSPin->pinPtr->intname;
			ename=gSPin->pinPtr->extname;
			if (iname==0) 
			{
				iname=ename;
			}
			/* change Apr29 1989 */
			if ( j == pos1)
			{
				iname = ename;
			}
			/* end change */
			/*
			if (num == 172)
			{
			printf("\n j=%4d,iname=%4d,ename=%4d",j,iname,ename);
			}
			*/
			runPtr=(gr_HeadGrNets+num)->firstXPin;
			j1=0;
			while (runPtr!=NULL)
			{
				j1++;
				if ((marks+j1)->number==0)
				{
					if ((runPtr->pinPtr->intname==iname)||
					(runPtr->pinPtr->intname==ename)||
					(runPtr->pinPtr->extname==iname)||
					(runPtr->pinPtr->extname==ename))
					{
						if (last==current)
						{
							current++;
						}
						(marks+j1)->number=current;
						/*
						if (num ==172)
						{
						printf("\n j1=%4d",j1);
						}
						*/
					}
				}
				if ((marks+j1)->number==(-1))
				{
					if ((runPtr->pinPtr->intname==iname)||
					(runPtr->pinPtr->intname==ename)||
					(runPtr->pinPtr->extname==iname)||
					(runPtr->pinPtr->extname==ename))
					{
						GR_FREE((char*) marks);
						/*
						if (num == 172)
						{
						printf("\n j=%4d,",j);
						}
						*/
						return (j);
					}
				}
				runPtr=runPtr->nextXPin;
			}
		}
		gSPin=gSPin->nextXPin;
	}
	if (current!=last)
	{
		last=current;
		goto lab1;
	}
	GR_FREE((char*) marks);
	return (0);
}
	

/*-------------------------------------------------------------------------
*	Gr_ReconnectPins
*	This routine is called when there are multiple pseudopins
*	at positions pos3 and pos4 and there is no path between
*	them which does not cross other side of the bottleneck.
*	Results      : none.
*	Side effects : pins which were connected to the pseudopin at pos4
*	became reconnected to the pseudopin at pos3.
*	It works only for pins outside bottlenecks.
*-------------------------------------------------------------------------
*/

void Gr_ReconnectPins(pos3,pos4,count,num)

int pos3,pos4;	/* Positions of active pins.			*/
int count;	/* # pins in net.				*/
int num;	/* Internal number of net.			*/


{
	BearId ename,iname,chename,chiname;
	grSortedPin *gSPin;
	int i,j;

	i=0;
	gSPin=(gr_HeadGrNets+num)->firstXPin;

	while (gSPin!=NULL)
	{
		i++;
		if (i==pos4)
		{
			iname=gSPin->pinPtr->intname;
			ename=gSPin->pinPtr->extname;
		}
		if (i==pos3)
		{
			chiname=gSPin->pinPtr->intname;
			chename=gSPin->pinPtr->extname;
		}
		gSPin=gSPin->nextXPin;
	}
	/*
	if (num == 172)
	{
	printf("\n iname=%4d,ename=%4d",iname,ename);
	printf("\n chiname=%4d,chename=%4d",chiname,chename);
	}
	*/
	j=0;
	gSPin=(gr_HeadGrNets+num)->firstXPin;
	while (gSPin!=NULL)
	{
		j++;
		if ((gSPin->pinPtr->extname == ename)&&
		(j!=pos4))
		{
			gSPin->pinPtr->extname = chename;
			if (gr_BearConnection==1)
			{
				if (gSPin->pinPtr->type==1)
				{
					Mac_PutExtNetIdOfPin(chename,gSPin->
					pinPtr->PinPtr.pinInst);
				}
				if (gSPin->pinPtr->type==2)
				{
					Tp_PutExtNetIdOfPPin(chename,gSPin->
					pinPtr->PinPtr.pseudoPin);
				}
				if (gSPin->pinPtr->type==5)
				{
					Tp_PutIntNetIdOfPPin(chename,gSPin->
					pinPtr->PinPtr.pseudoPin);
				}
				if (gSPin->pinPtr->type==4)
	
				{
					Tp_PutExtNetIdOfPPin(chename,gSPin->
					pinPtr->PinPtr.pseudoPin);
				}
			}
		}
		if (gSPin->pinPtr->extname==iname)
		{
			gSPin->pinPtr->extname=chiname;
			if ((gSPin->pinPtr->type==1)&&(gr_BearConnection==1))
			{
				Mac_PutExtNetIdOfPin(chiname,gSPin->pinPtr->
				PinPtr.pinInst);
			}
		}
		gSPin=gSPin->nextXPin;
	}
}


/*---------------------------------------------------------------------------
*	Gr_UndoInternalPins(pos1,pos2,num)
*	This routine is called when there are multiple pseudopins.
*	Results      : none.
*	Side effects : pins which were previously connected to pseudo
*	pin at position pos2  become connected to pseudopin at pos1.
*	Reconnected pins are inside bottleneck.
*----------------------------------------------------------------------------
*/


void Gr_UndoInternalPins(pos1,pos2,num)

int pos1,pos2;	/* Positions of active pins.			*/
int num;	/* Internal number of net.			*/

{

	BearId iname,chiname;
	grSortedPin *gSPin;
	int i;

	/*
	if ( num ==172)
	{
	printf("UndoInternalPins");
	printf("\n pos1=%4d,pos2+%4d",pos1,pos2);
	}
	*/
	i=0;
	gSPin=(gr_HeadGrNets+num)->firstXPin;
	while (gSPin!=NULL)
	{
		i++;
		if (i==pos2)
		{
			iname=gSPin->pinPtr->intname;
		}
		if (i==pos1)
		{
			chiname=gSPin->pinPtr->intname;
		}
		gSPin=gSPin->nextXPin;
	}
	gSPin=(gr_HeadGrNets+num)->firstXPin;
	while (gSPin!=NULL)
	{
		if ((gSPin->pinPtr->extname==iname)&&(gSPin->pinPtr->type==1))
		{
			gSPin->pinPtr->extname=chiname;
			if (gr_BearConnection==1)
			{
				Mac_PutExtNetIdOfPin(chiname,gSPin->pinPtr->
				PinPtr.pinInst);
			}
		}
		gSPin=gSPin->nextXPin;
	}
}


/*---------------------------------------------------------------------------
*	Gr_DetermineXPinsToDelete
*	This function is called when a cycle due to multiple pseudo
*	pins has been detected. 
*	Results      : none,
*	Side effects : pins to be deleted from X -links 
*	are marked 1, those which stay are marked 0.
*	Vector marks contains the information.
-----------------------------------------------------------------------------
*/


void Gr_DetermineXPinsToDelete(pos,marks,num)

int 	  	 pos;
int 		 num;	/* Internal net number.				*/
grPrimit  	*marks; /* Vector of marked pins to delete.		*/

{

	int  count,save,last;
	grSortedPin *gSPin,*search;
	BearId checkname;

	gSPin=(gr_HeadGrNets+num)->firstXPin;
	count=0;
	last=pos;
	while (gSPin!=NULL)
	{
		count++;
		if (count==last)
		{
			(marks+count)->number=1;
			search=gSPin;
			goto lab1;
		}
		gSPin=gSPin->nextXPin;
	}

	lab1:;

}

/*---------------------------------------------------------------------------
*	Gr_DetermineYPinsToDelete
*	This function is called when a cycle due to multiple pseudo
*	pins has been detected. 
*	Results      : none,
*	Side effects : pins to be deleted from Y -links 
*	are marked 1, those which stay are marked 0.
*	Vector ymarks contains the information.
-----------------------------------------------------------------------------
*/

void Gr_DetermineYPinsToDelete(marks,ymarks,num)

int 		 num;		/* Internal net number.			*/
grPrimit 	*ymarks,*marks; /* Vectors of marked pins along X and Y
				   links; respectively.			*/ 

{

	grSortedPin *gSPin,*run;
	int i,j;

	gSPin=(gr_HeadGrNets+num)->firstXPin;
	i=0;
	while (gSPin!=NULL)
	{
		i++;
		if ((marks+i)->number==1)
		{
			run=(gr_HeadGrNets+num)->firstYPin;
			j=0;
			while (run!=NULL)
			{
				j++;
				if (run==gSPin)
				{
					(ymarks+j)->number=1;
					goto lab1;
				}
				run=run->nextYPin;
			}
			lab1:;
		}
		gSPin=gSPin->nextXPin;
	}
}


/*----------------------------------------------------------------------
*	Gr_DeletePinsFromLinks
*	This procedure is called when it is known which pseudopins
*	are to be deleted from X- and Y- links.
*	Results     : none.
*	Side effects: pseudopins marked 1 in marks and ymarks
*	are deleted from lists.
*-----------------------------------------------------------------------
*/


void Gr_DeletePinsFromLinks(marks,ymarks,num,par)

grPrimit *marks,*ymarks;	/* Vectors of marked pins along X and Y
				   links; respectively.			*/
int 	  num;			/* Internal net number.			*/
int 	  par;

{
	grSortedPin *gSPin,*previous,*next;
	int i;
	Tile *tile;
	SideDirection side;

	i=0;
	if (par==0)
	{
		gSPin=(gr_HeadGrNets+num)->firstXPin;
	}
	else
	{
		gSPin=(gr_HeadGrNets+num)->firstYPin;
	}
	previous=NULL;
	while (gSPin!=NULL)
	{
		if (par==0)
		{
			next=gSPin->nextXPin;
		}
		else
		{
			next=gSPin->nextYPin;
		}
		i++;
		if ((par==0)&&((marks+i)->number==1))
		{
			if (previous==NULL)
			{
				(gr_HeadGrNets+num)->firstXPin=next;
			}
			else
			{
				previous->nextXPin=next;
			}
			if (gr_BearConnection==1)
			{
				if (gSPin->pinPtr->type==2)
				{
					tile=Tp_GetTileOfPPin(gSPin->pinPtr->
					PinPtr.pseudoPin);
					side=Tp_GetSideOfPPin(gSPin->pinPtr->
					PinPtr.pseudoPin);
					Tp_DeletePPinOfBnTile(tile,side,
					gSPin->pinPtr->PinPtr.pseudoPin);
				}
				if (gSPin->pinPtr->type==5)
				{
					tile=Tp_GetTileOfPPin(gSPin->pinPtr->
					PinPtr.pseudoPin);
					side=Tp_GetSideOfPPin(gSPin->pinPtr->
					PinPtr.pseudoPin);
					Tp_DeletePPinOfBnLine(tile,side,
					TP_BOTTOM,gSPin->pinPtr->
					PinPtr.pseudoPin);
				}
			}
		}
		if ((par==1)&&((ymarks+i)->number==1))
		{
			if (previous==NULL)
			{
				(gr_HeadGrNets+num)->firstYPin=next;
			}
			else
			{	
				previous->nextYPin=next;
			}
		}
		if ((par==0)&&((marks+i)->number!=1))
		{
			previous=gSPin;
		}
		if ((par==1)&&((ymarks+i)->number!=1))
		{
			previous=gSPin;
		}
		gSPin=next;
	}
}

/*----------------------------------------------------------------------
*	Gr_UndoDoublePair
*	This procedure is called when there are multiple pseudo
*	pins on both sides of the same bottleneck. 
*	Results     : none.
*	Side effects: one set of multiple pseudopins are eliminated.
*-----------------------------------------------------------------------
*/



void Gr_UndoDoublePair(pos1,pos2,pos3,pos4,pom,count,num)

int pos1,pos2,pos3,pos4;	/* Positions of active pins.	*/
int pom;
int count;			/* # pins in net.		*/
int num;			/* Internal net number.		*/

{
	grPrimit *marks,*ymarks;
	int i;
	char string[10];

/*
if (num == 172)
{
printf("\n UndoDoublePair");
printf("\n pos1=%4d,pos2=%4d,pos3=%4d,pos4=%4d",pos1,pos2,pos3,pos4);
printf("\n pom=%4d",pom);
}
*/
	if ((pos3!=0)&&(pos4==0)&&((pos1-pos2)*(pos1-pos2)>1))
	{
		Gr_ReconnectPins(pos1,pos3,count,num);
	}
	if ((pos3!=0)&&(pos4!=0))
	{
		Gr_ReconnectPins(pos3,pos4,count,num);
	}

	Gr_UndoInternalPins(pos1,pos2,num);

	MALLOC(grPrimit*,marks,(count+1)*sizeof (grPrimit));
	MALLOC(grPrimit*,ymarks,(count+1)*sizeof (grPrimit));
	for (i=1;i<=count;i++)
	{
		(marks+i)->number=0;
		(ymarks+i)->number=0;
	}

	if (pos4==0)
	{
		Gr_DetermineXPinsToDelete(pos2,marks,num);
	}
	else
	{
		Gr_DetermineXPinsToDelete(pos4,marks,num);
	}

	Gr_ClipHangingBranches(marks,num);
	Gr_DetermineYPinsToDelete(marks,ymarks,num);

	Gr_DeletePinsFromLinks(marks,ymarks,num,0);
	Gr_DeletePinsFromLinks(marks,ymarks,num,1);
	GR_FREE((char*) marks);
	GR_FREE((char*) ymarks);
}

/*-----------------------------------------------------------------------
*	Gr_ClipHangingBranches
*	This procedure is called when multiple pseudopins have been
*	deleted. 
*	Results      : none.
*	Side effects : branches which terminate at pseudopins are
*	removed.
*-----------------------------------------------------------------------
*/



void Gr_ClipHangingBranches(marks,num)

grPrimit *marks;	/* Vector of marked pins.		*/
int num;		/* Internal net number.			*/

{
	BearId ename,iname;
	grSortedPin  *gSPin,*rPin;
	int i,cntre,cntri,count,save_e,save_i;


	lab2:;
	gSPin=(gr_HeadGrNets+num)->firstXPin;
	count=0;

	while (gSPin!=NULL)
	{
		count++;
		cntri=0;
		cntre=0;

		if ((marks+count)->number == 1)
		{
			ename=gSPin->pinPtr->extname;
			iname=gSPin->pinPtr->intname;
			i=0;
			
			rPin=(gr_HeadGrNets+num)->firstXPin;

			while (rPin!=NULL)
			{
				i++;
				if ((marks+i)->number==0)
				{
					if ((rPin->pinPtr->extname==ename)||
					(rPin->pinPtr->intname==ename))
					
					{
						if (cntre==0)
						{
							cntre=1;
							save_e=i;
						}
						else
						{
							cntre=2;
						}
					}
					if ((rPin->pinPtr->intname==iname)||
					(rPin->pinPtr->extname==iname))
					{
						if (cntri==0)
						{
							cntri=1;
							save_i=i;
						}
						else
						{
							cntri=2;
						}
					}
				}
				rPin=rPin->nextXPin;
			}
		}
		if (cntre==1)
		{
			(marks+save_e)->number=1;
			goto lab2;
		}
		if (cntri==1)
		{
			(marks+save_i)->number=1;
			goto lab2;
		}
		gSPin=gSPin->nextXPin;
	}
}
