
/* grOnTiles.c --
 *
 *     Malgorzata Marek-Sadowska
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grOnTiles.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint


#include "bearCopyRight.h"
#include "grInt.h"


/*
*-----------------------------------------------------------------------------
*   This file contains functions used for reading geometrical
*   objects from tile plane into internal data base of global
*   router.
*-----------------------------------------------------------------------------
*/


/* ---------- Local Procedures ---------- */
static int GetBlocks();
static int GetHBottlenecks(), GetVBottlenecks();


/*
*----------------------------------------------------------------------------
*   Function used by iterator running on tile plane.
*   GetBlocks
*   Results     : 0
*   Side effects    : one solid  block is read in, its memory assigned
*             and proper fields in grBlock filled in.
*---------------------------------------------------------------------------
*/

/* ARGSUSED */
static int
GetBlocks(tilePtr, clientData)

register Tile *tilePtr;     /* Pointer to a tile.           */
ClientData clientData;      /* Not used, sits there because this
                   is expected by iterator.     */

{
    Rect rect;

    if (Tp_IsSolidTile(tilePtr))
    {
        grSolidTileCount++;
        if (grSolidTileCount==1)
        {
            MALLOC(grBlock*,grHeadBlockPtr,sizeof (grBlock));
            grLastBlockPtr=grHeadBlockPtr;
        }
        else
        {
            MALLOC(grBlock *, grLastBlockPtr->nextBlockPtr, sizeof (grBlock));
            grLastBlockPtr=grLastBlockPtr->nextBlockPtr;
        }

        grLastBlockPtr->dCell=0; 
        grLastBlockPtr->top=NULL; 
        grLastBlockPtr->bot=NULL; 
        grLastBlockPtr->right=NULL; 
        grLastBlockPtr->left=NULL; 
        grLastBlockPtr->tooThin=0; 
        grLastBlockPtr->nextBlockPtr=NULL;
        grLastBlockPtr->name=grSolidTileCount;

        Tp_HorzTileToRect(grPlaneDataPtr,tilePtr,&rect);

        grLastBlockPtr->xpos=rect.r_xbot + gr_xShift;
        grLastBlockPtr->ypos=rect.r_ybot + gr_yShift;
        grLastBlockPtr->height=rect.r_ur.p_y - rect.r_ll.p_y;
        grLastBlockPtr->width=rect.r_ur.p_x - rect.r_ll.p_x;
        grLastBlockPtr->persona=1; /*solid block*/
        grLastBlockPtr->tilePtr=tilePtr; 
        grLastBlockPtr->bnLinePtr=NULL; 
        grLastBlockPtr->avCap=0; /* no net can pass through*/

        if( Tp_IsDCellTile(tilePtr))
        {
            grLastBlockPtr->dCell = 1;
        }

    }
    return (0); /* continue search */
}


/*
 *----------------------------------------------------------------------------
 *  Function used by iterator running on tile plane.
 *  GetHBottlenecks()
 *  Results     : 0 or 1
 *  Side effects    : one horizontal bottleneck is read in,
 *            its memory assigned
 *            and proper fields in grBlock filled in.
 *---------------------------------------------------------------------------
 */

/* ARGSUSED */
static int
GetHBottlenecks(tilePtr, clientData)
register Tile   *tilePtr;       /* Pointer to a tile.                       */
ClientData      clientData;     /* Not used, sits there because iterator
                                 * expects it.
                                 */
{
    Rect    rect;
    int     delta;

    if (Tp_IsBottleneckTile(tilePtr))
    {
        grHorzBottlNumber++;
        grBottleneckTileCount++;
        MALLOC(grBlock *, grLastBlockPtr->nextBlockPtr,
        sizeof (grBlock));
        grLastBlockPtr=grLastBlockPtr->nextBlockPtr;
        grLastBlockPtr->nextBlockPtr=NULL;
        grLastBlockPtr->name=grBottleneckTileCount;


        Tp_HorzTileToRect(grPlaneDataPtr,tilePtr,&rect);

        grLastBlockPtr->xpos=rect.r_xbot + gr_xShift;
        grLastBlockPtr->ypos=rect.r_ybot + gr_yShift;
        grLastBlockPtr->height=rect.r_ur.p_y - rect.r_ll.p_y;
        grLastBlockPtr->width=rect.r_ur.p_x - rect.r_ll.p_x;
        
        grLastBlockPtr->dCell=0;
        grLastBlockPtr->top=NULL;
        grLastBlockPtr->bot=NULL;
        grLastBlockPtr->left=NULL;
        grLastBlockPtr->right=NULL;
        grLastBlockPtr->tooThin=0;

        if (grLastBlockPtr->width < gr_horzDesRule)
        {
            (void) fprintf(bearerr, " horizontal tile has too thin width\n");
    /*
    Bear_PrintErr("global router does not accept tiles thinner than 1 unit");
            exit(-1);
            */
            grLastBlockPtr->tooThin = 1;
        }
        grLastBlockPtr->persona=3; /* horz bottleneck*/
        grLastBlockPtr->tilePtr=tilePtr; 
        grLastBlockPtr->bnLinePtr=NULL; 

#ifdef DEBUG
{
        int tileCap, tileDen;

        tileCap = Tp_GetCapacityOfTile(tilePtr);
        tileDen = Tp_GetDensityOfTile(tilePtr);
        delta = tileCap - tileDen;
}
#else
        delta = Tp_GetCapacityOfTile(tilePtr) - Tp_GetDensityOfTile(tilePtr);
#endif DEBUG
        if (delta <= 0) {
            Bear_PrintErr("GetHBottlenecks() failed : (cap - den) < 0");
            return (MAGIC_STOP_SEARCH);
        }
        grLastBlockPtr->avCap = delta / gr_vertDesRule;
    }

    return (MAGIC_CONTINUE_SEARCH);
}


/*
 *----------------------------------------------------------------------------
 *  Function used by iterator running on tile plane.
 *  GetVBottlenecks()
 *  Results     : 0 or 1
 *  Side effects    : one vertical bottleneck is read in,
 *            its memory assigned
 *            and proper fields in grBlock filled in.
 *---------------------------------------------------------------------------
 */

/* ARGSUSED */
static int 
GetVBottlenecks(tilePtr, clientData)
register Tile   *tilePtr;       /* Pointer to a tile.                       */
ClientData      clientData;     /* Not used, sits there because iterator
                                 * expects it.
                                 */
{
    int     delta;
    Rect    rect;

    if (Tp_IsBottleneckTile(tilePtr))
    {
        grVertBottlNumber++;
        grBottleneckTileCount++;
        MALLOC(grBlock *, grLastBlockPtr->nextBlockPtr,
        sizeof (grBlock));

        grLastBlockPtr->dCell=0; 
        grLastBlockPtr->top=NULL; 
        grLastBlockPtr->bot=NULL; 
        grLastBlockPtr->right=NULL; 
        grLastBlockPtr->left=NULL; 
        grLastBlockPtr->tooThin=0; 

        grLastBlockPtr=grLastBlockPtr->nextBlockPtr;
        grLastBlockPtr->nextBlockPtr=NULL;
        grLastBlockPtr->name=grBottleneckTileCount;

        Tp_VertTileToRect(grPlaneDataPtr,tilePtr,&rect);

        grLastBlockPtr->xpos=rect.r_xbot + gr_xShift;
        grLastBlockPtr->ypos=rect.r_ybot + gr_yShift;
        grLastBlockPtr->height=rect.r_ur.p_y - rect.r_ll.p_y;
        if (grLastBlockPtr->height < gr_horzDesRule) {
            (void) fprintf (bearerr, " vertical tile has too thin height\n");
            /*
    Bear_PrintErr("global router does not accept tiles thinner than 1 unit");
            exit(-1);
            */
            grLastBlockPtr->tooThin = 1;
        }
        grLastBlockPtr->width=rect.r_ur.p_x - rect.r_ll.p_x;
        grLastBlockPtr->persona=2; /*vertical bottleneck*/
        grLastBlockPtr->tilePtr=tilePtr; 
        grLastBlockPtr->bnLinePtr=NULL; 

#ifdef DEBUG
{
        int tileCap, tileDen;

        tileCap = Tp_GetCapacityOfTile(tilePtr);
        tileDen = Tp_GetDensityOfTile(tilePtr);
        delta = tileCap - tileDen;
}
#else
        delta = Tp_GetCapacityOfTile(tilePtr) - Tp_GetDensityOfTile(tilePtr);
#endif DEBUG
        if (delta <= 0) {
            Bear_PrintErr("GetVBottlenecks() failed : (cap - den) < 0");
            return (MAGIC_STOP_SEARCH);
        }
        grLastBlockPtr->avCap = delta / gr_horzDesRule;
    }

    return (MAGIC_CONTINUE_SEARCH);
}


/*
*----------------------------------------------------------------------------
*   Iterator calling function.
*   Gr_FillSolidTiles()
*   Results     : none.
*   Side effects    : solid blocks are read in from the tile plane.
*----------------------------------------------------------------------------
*/

void Gr_FillSolidTiles()

{
    planeBndPtr=Tp_GetHorzPlaneBnd(grPlaneDataPtr);

    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(grPlaneDataPtr),
        planeBndPtr, GetBlocks, (ClientData) NULL);
}


/*
 *----------------------------------------------------------------------------
 *   Iterator calling function.
 *   GrFillVBottlenecks()
 *   Results :
 *       BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *   Side effects    : vertical bottlenecks are read in from the tile plane.
 *----------------------------------------------------------------------------
 */

BearStatus
GrFillVBottlenecks()
{
    int magicStatus;

    grVertBottlNumber = 0;
    planeBndPtr = Tp_GetVertPlaneBnd(grPlaneDataPtr);

    magicStatus = TiSrArea(NIL(Tile), Tp_GetVertPlane(grPlaneDataPtr),
        planeBndPtr, GetVBottlenecks, (ClientData) NULL);

    return ((magicStatus == 0) ? BEAR_OK : BEAR_ERROR);
}


/*
 *----------------------------------------------------------------------------
 *   Iterator calling function.
 *   GrFillHBottlenecks()
 *   Results :
 *       BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *   Side effects    :horizontal bottlenecks are read in from the tile plane.
 *----------------------------------------------------------------------------
 */

BearStatus
GrFillHBottlenecks()
{
    int magicStatus;

    grHorzBottlNumber = 0;
    planeBndPtr = Tp_GetHorzPlaneBnd(grPlaneDataPtr);

    magicStatus = TiSrArea(NIL(Tile), Tp_GetHorzPlane(grPlaneDataPtr),
        planeBndPtr, GetHBottlenecks, (ClientData) NULL);

    return ((magicStatus == 0) ? BEAR_OK : BEAR_ERROR);
}


/*
*----------------------------------------------------------------------------
*   Function used by iterator running on tile plane.
*   Gr_ReadVertBnLine
*   Results     : BearStatus
*   Side effects    : one vertical bottleneck line is read in,
*             its memory assigned
*             and proper fields in gBline filled in.
*---------------------------------------------------------------------------
*/



BearStatus Gr_ReadVertBnLine(bnLine,clientData)

BNLine *bnLine;     /* Pointer to btn line.             */
ClientData clientData;  /* Not used, sits because iterator expects it.  */

{
    Line *line;
    grBnLine *gBLine;
    int a,b;

    gBLine=Gr_CreateBnLine();
    line=Tp_GetLineOfBnLine(bnLine);

    gBLine->type=0;
    gBLine->base=line->point1.p_x + gr_xShift;
    gBLine->bnLinePtr = bnLine;
    a=line->point1.p_y + gr_yShift;
    b=line->point2.p_y + gr_yShift;

    if (a > b)
    {
        gBLine->lb=b;
        gBLine->rt=a;
    }
    else
    {
        gBLine->lb=a;
        gBLine->rt=b;
    }
    if ((gBLine->rt - gBLine->lb) < (gr_horzDesRule + 2))
    {
        (void) fprintf(bearerr, " vertical bnLine has too thin height");
Bear_PrintErr("global router does not accept bnLines thinner than 2 units");
            printf("\n");
            exit(-1);
    }

    /*
     * CHANGE HERE . THE SAME RULE AS IN tpPPin.c file.
     * Select tile so that the bottom fragment (TP_BOTTOM)
     * will always be used when creating pseudo pins.
     */
    gBLine->bntype = Tp_GetBnLineTypeOfBnLine(bnLine);

/* NEW RULE */
    gBLine->tilePtr=Tp_GetTLTileOfBnLine(bnLine);

/*
    if (gBLine->bntype==TP_VERT_R_UP)
    {
        gBLine->tilePtr=Tp_GetBRTileOfBnLine(bnLine);
    }
    else
    {
        gBLine->tilePtr=Tp_GetTLTileOfBnLine(bnLine);
    }
*/

    gBLine->cover = 0;

    return  (BEAR_CONTINUE_SEARCH);

}


/*
*----------------------------------------------------------------------------
*   Function used by iterator running on tile plane.
*   Gr_ReadHorzBnLine
*   Results     : BearStatus
*   Side effects    : one horizontal bottleneck line is read in,
*             its memory assigned
*             and proper fields in gBline filled in.
*---------------------------------------------------------------------------
*/

BearStatus Gr_ReadHorzBnLine(bnLine,clientData)

BNLine *bnLine;     /*  Pointer to bottleneck line.         */
ClientData clientData;  /*  Not used. Sits there because iterator expects
                it.                     */

{
    Line *line;
    grBnLine *gBLine;
    int a,b;

    gBLine=Gr_CreateBnLine();
    line=Tp_GetLineOfBnLine(bnLine);

    gBLine->type=1;
    gBLine->base=line->point1.p_y + gr_yShift;
    gBLine->bnLinePtr = bnLine;
    a=line->point1.p_x + gr_xShift;
    b=line->point2.p_x + gr_xShift;

    if (a > b)
    {
        gBLine->lb=b;
        gBLine->rt=a;
    }
    else
    {
        gBLine->lb=a;
        gBLine->rt=b;
    }
    if ((gBLine->rt - gBLine->lb) < (gr_horzDesRule + 2))
    {
        (void) fprintf(bearerr, " horz bnLine has too thin width");
Bear_PrintErr("global router does not accept bnLines thinner than 3 units");
            printf("\n");
            exit(-1);
    }

    /*
     * CHANGE HERE . THE SAME RULE AS IN tpPPin.c file.
     * Select tile so that the bottom fragment (TP_BOTTOM)
     * will always be used when creating pseudo pins.
     */
    gBLine->bntype=Tp_GetBnLineTypeOfBnLine(bnLine);

/* NEW RULE */
    gBLine->tilePtr=Tp_GetTLTileOfBnLine(bnLine);

/*
    if (gBLine->bntype==TP_HORZ_B_RIGHT)
    {
        gBLine->tilePtr=Tp_GetBRTileOfBnLine(bnLine);
    }
    else
    {
        gBLine->tilePtr=Tp_GetTLTileOfBnLine(bnLine);
    }
*/

    gBLine->cover = 0;

    return  (BEAR_CONTINUE_SEARCH);

}

/*
*-----------------------------------------------------------------------------
*   Gr_CreateBnLine()
*   Results     : pointer to  newly created grBnLine structure.
*   Side effects    : memory is assigned and cleaned up for
*             new grBnLine.
*----------------------------------------------------------------------------
*/



grBnLine *Gr_CreateBnLine()

{

    grBnLine *next;


    bnLineCount++;
    if (bnLineCount == 1)
    {
        MALLOC(grBnLine*,grBnLineList,sizeof (grBnLine));
        grBnLineList->nextBnLine=NULL;
    }
    else
    {
        MALLOC(grBnLine*,next,sizeof (grBnLine));
        next->nextBnLine=grBnLineList;
        grBnLineList = next;
    }

    grBnLineList->type=0;
    grBnLineList->base=0;
    grBnLineList->lb=0;
    grBnLineList->rt=0;
    grBnLineList->tilePtr=NULL;
    grBnLineList->bnLinePtr=NULL; 
    grBnLineList->cover =0;

    return (grBnLineList);

}


/*
*----------------------------------------------------------------------------
*   Iterator calling function.
*   Gr_ReadAllBnLines
*   Results     : none.
*   Side effects    : bottleneck lines are read in from the tile plane.
*----------------------------------------------------------------------------
*/



void Gr_ReadAllBnLines(planeDataPtr)

PlaneData *planeDataPtr;    /* Pointer to plane data.       */

{

    bnLineCount=0;
    hbnLCount =0;
    vbnLCount =0;
    grBnLineList = NULL;

    (void) Tp_ItrBnLine(Tp_GetHorzBnLineList(planeDataPtr),
        Gr_ReadHorzBnLine, (ClientData) NULL);
    hbnLCount=bnLineCount;

    (void) Tp_ItrBnLine(Tp_GetVertBnLineList(planeDataPtr),
        Gr_ReadVertBnLine, (ClientData) NULL);
    vbnLCount=bnLineCount - hbnLCount;


}


/*--------------------------------------------------------------------------
*  Gr_SetLinksForDummyCells ()  fills the pointers to neighboring bottlenecks
*  for a dummy cell. 
*  It is used to shrink dummy cells.
-----------------------------------------------------------------------------*/


void Gr_SetLinksForDummyCells(dCellPtr)

grBlock  *dCellPtr;

{
    grBlock  *run;
    int xleft,xright,ybot,ytop;
    int rxleft, rxright, rybot, rytop;

    run = grHeadBlockPtr;

    xleft = dCellPtr->xpos;
    xright= xleft + dCellPtr->width;
    ybot = dCellPtr->ypos;
    ytop = ybot + dCellPtr->height;

    while( run != NULL)
    {

        if((xleft != run->xpos)||(ybot != run->ypos))
        {
            rxleft = run->xpos;
            rxright= rxleft + run->width;
            rybot = run->ypos;
            rytop = rybot + run->height;
        
            /* Check if is top tile */

            if(( xleft == rxleft)&&(ytop == rybot))
            {
                dCellPtr->top = run;
            }

            if(( xleft == rxright)&&(ytop == rytop))
            {
                dCellPtr->left = run;
            }

            if(( xright == rxleft)&&(ybot == rybot))
            {
                dCellPtr->right = run;
            }

            if(( xright == rxright)&&(ybot == rytop))
            {
                dCellPtr->bot = run;
            }
        }
        run = run->nextBlockPtr;
    }
    if(dCellPtr->top == NULL)
    {
        printf("\n top is missing");
        dCellPtr->dCell =0;
    }
    if(dCellPtr->bot == NULL)
    {
        printf("\n bot is missing");
        dCellPtr->dCell =0;
    }
    if(dCellPtr->left == NULL)
    {
        printf("\n left is missing");
        dCellPtr->dCell =0;
    }
    if(dCellPtr->right == NULL)
    {
        printf("\n right is missing");
        dCellPtr->dCell =0;
    }
}


/*----------------------------------------------------------------------------
*   Gr_ConnectDummyCells()
*   Goes through a list of dummy cells and establishes connections
*   to bottlenecks.
*----------------------------------------------------------------------------*/

void Gr_ConnectDummyCells()

{

    grBlock *run;

    run = grHeadBlockPtr;

    while(run != NULL)
    {
        if ( run-> dCell == 1 )
        {
            Gr_SetLinksForDummyCells(run);
        }

        run = run->nextBlockPtr;
    }
}


/*
 * ---------------------------------------------------------------------------
 *   Gr_SqueezDCells()
 *   Results : 0 - thin bottlenecks can not be eliminated; 
 *             1 - thin botthenecks are eliminated.
 *   If there are tooThinBottl resulting from bottleneck lines that this
 *   procedure does not attempt to fix this.
 *   Side effects :
 *   If dummy cell is agjacent to a tooThinBottl then if is squeezed and
 *   bottleneck fattened.
 * ----------------------------------------------------------------------------
 */

int
Gr_SqueezDCells()
{
    grBlock  *run, *top, *bot, *left, *right ;
    int       delta;

    run = grHeadBlockPtr;

    while(run != NULL)
    {
        if ( run-> dCell == 1 )
        {
            top   = run->top;
            bot   = run->bot;
            left  = run->left;
            right = run->right;
            if(top->tooThin == 1)
            {
                /* check if dCell can be squeezed */

                if( run->height < (gr_horzDesRule + 3))
                {
                    return (0);
                }

                top->height = top->height + gr_horzDesRule;
                top->ypos = top->ypos - gr_horzDesRule;
                top->tooThin = 0;

                run->height = run->height - gr_horzDesRule;

                delta = left->ypos + left->height - ( run->ypos + run->height);

                if( delta > 0 )
                {

                    left->height = left->height - delta ;

                }

                delta = right->ypos + right->height - (run->ypos + run->height);

                if( delta > 0 )
                {

                    right->height = right->height - delta ;

                }
            }
            if(bot->tooThin == 1)
            {
                /* check if dCell can be squeezed */

                if( run->height < (gr_horzDesRule + 3))
                {
                    return (0);
                }

                bot->height = bot->height + gr_horzDesRule;
                bot->tooThin = 0;

                run->ypos = run->ypos + gr_horzDesRule;
                run->height = run->height - gr_horzDesRule;

                delta = run->ypos - left->ypos;
                if( delta > 0 )
                {
                    left->ypos = run->ypos;
                    left->height = left->height - delta;
                }

                delta = run->ypos - right->ypos;
                if( delta > 0 )
                {
                    right->ypos = run->ypos;
                    right->height = right->height - delta;
                }
            }

            if(left->tooThin == 1)
            {
                /* check if dCell can be squeezed */

                if( run->width < (gr_horzDesRule + 3))
                {
                    return (0);
                }

                left->width = left->width + gr_horzDesRule;
                left->tooThin = 0;

                run->xpos = run->xpos + gr_horzDesRule;
                run->width = run->width - gr_horzDesRule;

                delta =  run->xpos - top->xpos;
                if( delta > 0 )
                {
                    top->xpos = run->xpos ;
                    top->width = top->width - delta;
                }

                delta = run->xpos - bot->xpos ;
                if( delta > 0 )
                {
                    bot->xpos = run->xpos ;
                    bot->width = bot->width - delta;
                }
            }
            if(right->tooThin == 1)
            {
                /* check if dCell can be squeezed */

                if( run->width < (gr_horzDesRule + 3))
                {
                    return (0);
                }

                right->width = right->width + gr_horzDesRule;
                right->xpos = right->xpos - gr_horzDesRule;
                right->tooThin = 0;

                run->width = run->width - gr_horzDesRule;

                delta = top->xpos + top->width - (run->xpos + run->width);
                if( delta > 0 )
                {
                    top->width = top->width - delta;
                }

                delta = bot->xpos + bot->width - (run->xpos + run->width);
                if( delta > 0 )
                {
                    bot->width = bot->width - delta;
                }
            }
        }
        run = run->nextBlockPtr;
    }

    return (1) ;
}
