/*
 * grRootTree.c --
 *
 *     Author
 *	Malgorzata Marek-Sadowska
 *
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grRootTree.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint




#include "grInt.h"

/*
*---------------------------------------------------------------------------
*	This file contains functions used for decomposition purposes.
*---------------------------------------------------------------------------
*/



/*
*----------------------------------------------------------------------------
*	Gr_CreateNewRootTree
*	Results		: none.
*	Side effects	:
*	Allocate memory for a new grRootTree for a sub net pointed to by
*	subNetPtr . It is attached at the head of subNetPtr->tree
*----------------------------------------------------------------------------
*/



void Gr_CreateNewRootTree(subNetPtr)

grSubNet  *subNetPtr;	/* Pointer to a subnet.			*/


{

	grRootTree *save;

	save=subNetPtr->tree;
	MALLOC(grRootTree*,subNetPtr->tree,sizeof (grRootTree));

	subNetPtr->tree->nextOnList = save;
	subNetPtr->tree->nextInSubnet = NULL;
	subNetPtr->tree->mark=0;
	subNetPtr->tree->cut=0;
	subNetPtr->tree->distance=0;
	subNetPtr->tree->fatherPtr=NULL;
	subNetPtr->tree->siblingPtr=NULL;
	subNetPtr->tree->childPtr=NULL;
	subNetPtr->tree->bestNotMarked=NULL;

}

/*
*-----------------------------------------------------------------------------
*	Gr_ClearMarksOfRootTrees
*	Results		: none.
*	Side effects	: clears marks of root tree .
*----------------------------------------------------------------------------
*/


void Gr_ClearMarksOfRootTrees(subNetPtr)

grSubNet *subNetPtr;	/* Pointer fo subnet.				*/


{

	grRootTree *current,*pending;

	current=subNetPtr->tree;
        
        pending=current;

        if (current!=NULL)
        {

          while(pending->nextInSubnet!=NULL)
           {
              pending=pending->nextInSubnet;
           }
        }
        current=pending;

	while(current!=NULL)
	{
		current->mark =0;
		current = current->nextOnList;
	}

}

/*
*-----------------------------------------------------------------------------
*	Gr_PutBestUnmarkedRootTree
*	Results		: none.
*	Side effects	: For a grRootTree pointed to by rootTreePtr
*	and being a child of a subnet
*	pointed to by subNetPtr it finds the closest
*	grRootTree which has not 
*	been marked yet (i.e. included in a tree)
*	and puts it as bestNotMarked 
*	for that rootTreePtr.                                                   
*-----------------------------------------------------------------------------
*/


void Gr_PutBestUnmarkedRootTree(rootTreePtr,subNetPtr)

grRootTree *rootTreePtr;	/* Pointer to rootTree.			*/
grSubNet *subNetPtr;		/* Pointer to a subnet.			*/


{

	grRootTree *pending,*best;
	int dist,bestdist;
	best=NULL;
	bestdist=0;
	if (rootTreePtr->mark != 0)
	{
	pending=subNetPtr->tree;
	while(pending!=NULL)
	{
		if ((pending!=rootTreePtr)&&(pending->mark!=1))
		{
			dist=Gr_GetDistanceBetweenBoxes(
                        pending->nodePtr->contentsPtr,
                        rootTreePtr->nodePtr->contentsPtr);
			if (best==NULL)
			{
				best=pending;
				bestdist=dist;
			}
			else
			{
				if (bestdist>dist)
				{
					best=pending;
					bestdist=dist;
				}
			}
		}
		pending=pending->nextOnList;
	}
	}
	rootTreePtr->bestNotMarked = best;
	rootTreePtr->distance = bestdist;
}


/*
*----------------------------------------------------------------------------
*	Gr_SameSideBestUnmarkedRootTree
*	Resulte		: none.
*	Side effects	:
*	For a grRootTree pointed to by rootTreePtr and being a child of a sub
*	net pointed to by subNetPtr it finds the closest grRootTree which 
*	has not been marked yet (i.e. included in a tree) and such that either
*	it is on the same side of the cut line or is a part of the same pre-
*	routed segment                                                        
*-----------------------------------------------------------------------------
*/



void Gr_SameSideBestUnmarkedRootTree(rootTreePtr,subNetPtr)

grRootTree *rootTreePtr;	/* Pointer to rootTree.			*/
grSubNet *subNetPtr;		/* Pointer to subnet.			*/


{

	grRootTree *pending,*best;
	int dist,bestdist;

	best=NULL;
	bestdist=0;

	pending=subNetPtr->tree;

	while(pending!=NULL)
	{
		if ((pending!=rootTreePtr)&&(pending->mark!=1))  /*standard*/
		{
			if ((rootTreePtr->nodePtr->side)==
				(pending->nodePtr->side))
			{
				dist=Gr_GetDistanceBetweenBoxes(
				pending->nodePtr->contentsPtr,
				rootTreePtr->nodePtr->contentsPtr);
				if (best==NULL)
				{
					best=pending;
					bestdist=dist;
				}
				else
				{
					if (bestdist>dist)
					{
						best=pending;
						bestdist=dist;
					}
				}
			} /*end are on the same side*/

			if ((rootTreePtr->nodePtr->contentsPtr->name!=0)&&
			(rootTreePtr->nodePtr->contentsPtr->name ==
			 pending->nodePtr->contentsPtr->name))
			{
				best=pending;
				bestdist=0;
			}  /*end the same prerouted net*/

			pending=pending->nextOnList;
		}
	}
	rootTreePtr->bestNotMarked=best;
	rootTreePtr->distance = bestdist;

}


/*
*-----------------------------------------------------------------------------
*	Results		: none.
*	Side effects	:
*	Sets root trees only with nodePtr and bestNotMarked filled in         
*	Elements are linked along nextOnList .
*	It is initial tree and it is assumed that there is just one such a tree
*	If opti=0 the best unmarked is used, opti=1 the best on the same side  
*	of the cut line or of the same prerouted portion.                       
*-----------------------------------------------------------------------------
*/



void Gr_InitializeRootTrees(subNetPtr,opti)

grSubNet *subNetPtr;	/* Pointer to subnet.				*/
int opti;		/* Option parameter. See comments above.	*/


{

	grSuperNode *current;
	grRootTree *running;

/*
	Allocate memory and set nodePtr for root trees                         
*/

	current=subNetPtr->NodeList;

	while(current!=NULL)
	{
		Gr_CreateNewRootTree(subNetPtr);
		subNetPtr->tree->nodePtr = current;
		current=current->nextOnList;
	}

	/*
		 Calculate best not marked and distances    
	*/

/* 	running=subNetPtr->tree;

	while(running!=NULL)
	{
		if (opti==0)
		{
			Gr_PutBestUnmarkedRootTree(running,subNetPtr);
		}
		else
		{
			Gr_SameSideBestUnmarkedRootTree(running,subNetPtr);
		}

		running=running->nextOnList;
	}
*/
}


/*
*----------------------------------------------------------------------------
*	Gr_MkeRootTree
*	Results		: none.
*	Side effects	:
*	For a subnet pointed to bt subNetPtr it builds the initial rootTree 
*	The tree contains all the nodes and all links are established       
*	If opti=0 the best tree is sought, otherwise a tree which does not    
*	introduce more intersections with cutline .                            
*-----------------------------------------------------------------------------
*/



void Gr_MkeRootTree(subNetPtr,opti)

grSubNet *subNetPtr;	/* Pointer to a subnet.				*/
int opti;		/* Option parameter, see comments above.	*/


{

	grRootTree *current,*best,*parent,*child,*save;
	int name1,name2;
	int bestdist;

	bestdist=0;
	Gr_InitializeRootTrees(subNetPtr,opti);

	if (subNetPtr->tree!=NULL)
	{
		/* Initialize first root tree  */

		subNetPtr->tree->mark=1;
		Gr_PutBestUnmarkedRootTree(subNetPtr->tree,subNetPtr);
		subNetPtr->tree->childPtr = subNetPtr->tree->bestNotMarked;
		subNetPtr->tree->childPtr->mark=1;
		subNetPtr->tree->childPtr->bestNotMarked=NULL;
		subNetPtr->tree->childPtr->distance=0;
		subNetPtr->tree->fatherPtr=NULL;
		subNetPtr->tree->bestNotMarked->fatherPtr = subNetPtr->tree;
		subNetPtr->tree->bestNotMarked = NULL;
		subNetPtr->tree->distance = 0;
		Gr_PutBestUnmarkedRootTree(subNetPtr->tree,subNetPtr);
		Gr_PutBestUnmarkedRootTree(subNetPtr->tree->childPtr,subNetPtr);

		if (subNetPtr->tree->nodePtr->side!=
		subNetPtr->tree->childPtr->nodePtr->side)
		{
			subNetPtr->tree->childPtr->cut=1;
		}
		current=subNetPtr->tree;
		while(current!=NULL)
		{
			if (opti==0)
			{
				Gr_PutBestUnmarkedRootTree(current,subNetPtr);
			}
			else
			{
				Gr_SameSideBestUnmarkedRootTree(
				current,subNetPtr);
			}
			current=current->nextOnList;
		}
		/* End initialization */

lab1: 
		current=subNetPtr->tree;
		best=NULL;
		bestdist=0;
		while(current!=NULL)
		{
			if (current->mark == 1)
			{
				if (best == NULL)
				{
					best=current->bestNotMarked;
					parent = current;
					bestdist = current->distance;
				}
				else
				{
					if ((current->distance < bestdist)&&
					(current->bestNotMarked!=NULL))
					{
						bestdist = current->distance;
						parent = current;
						best = current->bestNotMarked;
					}
				}
			}
			current=current->nextOnList;
		}

		if (best!=NULL)
		{
			/*growth of the tree continues */
			if (parent->childPtr != NULL)
			{
				/* best is a sibling of a child */
				child=parent->childPtr;
				save=child->siblingPtr;
				child->siblingPtr = best;
				best->siblingPtr = save;
				best->fatherPtr = parent;
				if (best->nodePtr->side !=parent->nodePtr->side)
				{
					best->cut = 1;
				}
				if (best->cut == 1 )
				{
				      name1=best->nodePtr->contentsPtr->name;
				      name2=parent->nodePtr->contentsPtr->name;
				      if ((name1==name2)&&(name1!=NULL))
				      {
					      best->cut = 2;  
					      /* prerouted net intersects */
				      }
				}
				best->mark=1;
			}
			else
			{
				/* best is a child */

				parent->childPtr = best;
				best->fatherPtr = parent;
				if (best->nodePtr->side!=parent->nodePtr->side)
				{
					best->cut = 1;
				}
				if (best->cut == 1)
				{
					name1=best->nodePtr->contentsPtr->name;
					name2=parent->nodePtr->contentsPtr->
					name;
					if ((name1==name2)&&(name1!=NULL))
					{
						best->cut = 2; 
						/* prerouted net intersects*/
					}
				}
				best->mark = 1;
			}
			current=subNetPtr->tree;
			while(current!=NULL)
			{
				if (opti==0)
				{
					Gr_PutBestUnmarkedRootTree(
					current,subNetPtr);
				}
				else
				{
					Gr_SameSideBestUnmarkedRootTree(
					current,subNetPtr);
				}
				current=current->nextOnList;
			}
			goto lab1;
		}
	}
}


/*
*----------------------------------------------------------------------------
*	Gr_SplitRootTreeAtCutBranches
*	Results		: none.
*	Side effects	:
*	For a subnet pointed to by subNetPtr it generates a set of subtrees    
*	induced by branches which intersect with the cut line                 
*	Trees are linked along nextInSubnet pointers,  missing branches are    
*	not phisically removed but are marked 2 instead.                       
*	Latter the missing branches are used when subnets at the next level    
*	of hierarchy are generated                                            
*----------------------------------------------------------------------------
*/


void Gr_SplitRootTreeAtCutBranches(subNetPtr)

grSubNet *subNetPtr;	/* Pointer to a subnet.				*/


{

	grRootTree *current,*treePtr,*leftSibling,*previous;

	Gr_ClearMarksOfRootTrees(subNetPtr);
	treePtr=subNetPtr->tree;
        previous=NULL;

lab1: 
	current=treePtr;
	leftSibling=NULL;

	while(current!=NULL)  /*make visits */
	{
		if ((current->cut >= 1)&&(current->mark != 2)) /* make split*/
		{
			current->mark = 2;
			if (leftSibling!=NULL)
			{
				leftSibling->siblingPtr = current->siblingPtr;
			}
			else
			{
				current->fatherPtr->childPtr =
				 current->siblingPtr;
			}

			current->siblingPtr=NULL;
			current->nextInSubnet = treePtr;
                       if (previous==NULL)
                       {
				subNetPtr->tree=current;
                       }
                       else
                       {
                        	current->nextInSubnet=previous->nextInSubnet;
                        	previous->nextInSubnet=current;
                       }
                       treePtr=current;
		}
		if (current->mark==2) /*marked already*/
		{
                  if (current->childPtr!=NULL)
                  {
			current=current->childPtr;
			leftSibling = NULL;
                  }
                  else
                  {
                        goto lab2;
                  }
		}
	    if (current!=NULL)
	    {
		if (current->cut == 0)
		{
			/* mark and visit next */
			current->mark =1;
			if (current->childPtr != NULL)
			{
				current = current->childPtr;
				leftSibling = NULL;
			}
			else
			{
				if (current->siblingPtr != NULL)
				{
					leftSibling = current;
					current=current->siblingPtr;
				}
				else
				{
					while((current->siblingPtr == NULL)
					&&(current->mark!=2))
					{
						current=current->fatherPtr;
						if (current == NULL)
						{
/* CHANGE AUG 19 th */
							goto lab2;  
							/* get next tree */
						}
					}
					if (current->mark == 2)
					{
/* CHANGE AUG 19 th */
						goto lab2;  /* get next tree */
					}
					leftSibling=current;
					current=current->siblingPtr;
				}
			}
		}
	    }
	}
lab2: 
        previous=treePtr;
	treePtr = treePtr->nextInSubnet;
	if (treePtr != NULL) goto lab1;

}
