/*
 * grVirtNet.c --
 *
 *     Author
 *	Malgorzata Marek-Sadowska
 *
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalRouter/grVirtNet.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint



/*
*---------------------------------------------------------------------------
*	This file contains functions used when virtual nets are created
*	so that assignment can be made. Each virtual net has one super
*	node on the left hand side and one node on the right hand side.
*--------------------------------------------------------------------------
*/




#include "grInt.h"


/*
*-------------------------------------------------------------------------
*	Gr_DecomposeNets
*	This function performes first cut of nets after cut position and    
*	direction have been determined.
*	Results		: none.
*	Side effects	:   
*	the field "active" in grSubNet-s is filled out and if a net is     
*	separated by the cut line the new grPinRectangles are attached to   
*	the "decomposed" field and super nodes are created and attached to 
*	"NodeList".                 
*--------------------------------------------------------------------------
*/




void Gr_DecomposeNets(kPtr)

grRegion *kPtr;		/* Pointer to the parent region.		*/

{

	int dir,cutPos;
	grSubNet *pending;

	dir=kPtr->dir;
	if (dir!=0)
	{
		if (dir==1)
		{
			cutPos=kPtr->xcutvirt;
			cutPos=(gr_xcoord+cutPos)->number;
		}
		if (dir==2)
		{
			cutPos=kPtr->ycutvirt;
			cutPos=(gr_ycoord+cutPos)->number;
		}

		pending=kPtr->subNetsPtr;

		while(pending!=NULL)
		{
			Gr_CreateInitialSuperNodes(pending,cutPos,dir);

			pending=pending->nextOnList;
		}
	}
}




/*
*----------------------------------------------------------------------------
*	Gr_CreateSimpleVirtualNet
*	Results		: none.
*	Side effects	: virtual nets are created for subNets which
*	have exactly 2 superNodes on their NodeList.
*	No decomposition is attempted.
*----------------------------------------------------------------------------
*/



void Gr_CreateSimpleVirtualNet(sNet)

grSubNet *sNet;		/* Parent subNet.			*/

{
	grSuperNode *current;
	grVirtNet *vNet;

        vNet= Gr_CreateNewVirtNet(sNet);

	current=sNet->NodeList;

	if (current->side == 1)
	{
		vNet->leftNode=current;
		vNet->rightNode=current->nextOnList;
	}
	else
	{
		vNet->rightNode = current;
		vNet->leftNode = current->nextOnList;
	}

	vNet->netName = sNet->netname;
	vNet->subNetName=sNet->subname;
	m++;
	vNet->leftNode->nodeName=sNet->subname; /*left side gets the old name*/
	vNet->rightNode->nodeName=Gr_GetNewSubNetName();
}


/*
*-----------------------------------------------------------------------------
*	Gr_CreateGreedyVirtualNet(sNet)
*	For a net with prerouted segments which needs single crossing finds 
*	left and right super nodes with the closest distance
*	Results		: none.
*	Side effects	:
*	A virtual net is created, subNet activity is set to 1
*	left Super Nodes get the current sNet subname, right get new name.
*----------------------------------------------------------------------------
*/



void Gr_CreateGreedyVirtualNet(sNet)

grSubNet *sNet;		/* Pointer to the parent subNet.		*/

{

	grSuperNode *bestleft, *bestright, *current, *pending;
	int dist,bestdist;
        grVirtNet *vNet;

	bestdist=0;
	bestleft=NULL;
	bestright=NULL;

	current=sNet->NodeList;

	while(current!=NULL)
	{
		if (current->side == 2)
		{
			pending=sNet->NodeList;

			while(pending!=NULL)
			{
				if (pending->side == 1)
				{
					pending=pending->nextOnList;
				}
				else
				{
					dist=Gr_GetDistanceBetweenBoxes(
					current->contentsPtr,
					pending->contentsPtr);
					if ((bestdist == 0)||(dist<bestdist))
					{
						bestdist=dist;
						bestleft=current;
						bestright=pending;
					}
					pending=pending->nextOnList;
				}
			}
                	current=current->nextOnList;
		}
		else
		{
               		current=current->nextOnList;
            	}
	}

        vNet=Gr_CreateNewVirtNet(sNet);

	vNet->rightNode=bestright;
	vNet->leftNode=bestleft;
	vNet->netName=sNet->netname;
	vNet->subNetName=sNet->subname;
	m++;

}


/*
*---------------------------------------------------------------------------
*	Gr_DropNetToOneRegion
*	All pins of the subnet are on one side of a cut line 
*	If side = 0 : left , = 1 : right.   
*	Results		: none.
*	Side effects	:
*	A net is dropped to the corresponding child region   
*---------------------------------------------------------------------------
*/



void Gr_DropNetToOneRegion(sNet,region)

grSubNet *sNet;
grRegion *region;

{

	grSubNet *newNet;
	grSuperNode *current;
	grPinRectangle *new, *old, *del, *last;
        int hpt,vpt;
	float total;

	Gr_CreateNewSubnet(region);
	newNet=region->subNetsPtr;

	newNet->netname=sNet->netname;
	newNet->subname=sNet->subname;
	newNet->critical=sNet->critical;
	newNet->decCandidate=sNet->decCandidate;
	newNet->xapprox=sNet->xapprox;
	newNet->yapprox=sNet->yapprox;
	newNet->active = 0;
        newNet->minx=sNet->minx;
        newNet->maxx=sNet->maxx;
        newNet->miny=sNet->miny;
        newNet->maxy=sNet->maxy;
	newNet->noComponents=sNet->noComponents;

	current=sNet->NodeList;

	while(current!=NULL)
	{

		new=current->contentsPtr;

		if (current->kind>0) /* to be connected*/
		{
	        	newNet->noLoosePins=new->count;

			if (gr_Timing ==1)
			{
			/* update currentLength and set approxlength for
			   newNet                                       */

				hpt=Gr_CalculateHorizontalPortionOfTree(new);
				vpt=Gr_CalculateVerticalPortionOfTree(new);
				total=hpt+vpt*gr_HorVertWire;
				newNet->approxLength=total;
				((gr_HeadGrNets+sNet->netname)->currentLength)
				+= total;
				((gr_HeadGrNets+sNet->netname)->xapprox)+=
				hpt;
				((gr_HeadGrNets+sNet->netname)->yapprox)+=
				vpt;
				newNet->xapprox=hpt;
				newNet->yapprox=vpt;
			}
		}
		/*link pin rectangle */

		if (region->firstPinRectanglePtr == NULL)
		{
			region->firstPinRectanglePtr = new;
			region->lastPinRectanglePtr  = new;
			new->nextOnList = NULL;
		}
		else
		{
			new->nextOnList=region->firstPinRectanglePtr;
			region->firstPinRectanglePtr = new;
		}


		/* end link*/

			if (current->kind ==0) /*prerouted net*/
				newNet->areConnected = new;

		if (current->kind==1) /* to be connected*/
			newNet->toBeConnected = new;

		current = current->nextOnList;
	}

	/*update nextInNet links */

	del=newNet->toBeConnected;
	last=NULL;

	if (del!=NULL)  /*remove del from the list areConnected */
		old=newNet->areConnected;

	while((old!=NULL)&&(old!=del))
	{
		last=old;
		old=old->nextInNet;
	}
	if (old!=NULL)
	{
		if (last==NULL)
		{
			newNet->areConnected = old->nextInNet;
		}
		else
		{
			last->nextInNet = old->nextInNet;
		}
	}
	newNet->toBeConnected->nextInNet = NULL;
	sNet->netname = 0; /* sNet to be freed latter */
}

