
/* gsDecompact.c --
 *
 *     Wei-Ming Dai
 *     (Professor E. S. Kuh)
 *
 * This module performs global decompaction on a given placement.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalSpacer/gsDecompact.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"         /*temp*/
#include "tileProp.h"
#include "mac.h"
#include "globalSpacer.h"
#include "globalSpacerInt.h"

#include "ezPlot.h"
#include "../polarBear/polarBear.h"

#ifndef BEAR_BLACK_BOX
WindowData          *globalWinDataPtr;
#endif BEAR_BLACK_BOX


static Plane        *planePtr;
static PlaneData    *globalPlaneDataPtr;
static Rect         planeSearchRect;
static Rect         tailRect;
static Rect         updateTileRect;
static Tile         *leftmostTilePtr;
static Tile         *topmostTilePtr;


/*
 * ----------------------------------------------------------------------------
 *
 * UpdateHorzDecompactRidge() --
 *
 *     Update horizontal decompaction ridge. 
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *    Update the best decompaction ridge found so far.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
UpdateHorzDecompactRidge(tilePtr, ridgeInfoPtr)

register Tile      *tilePtr;
register RidgeInfo *ridgeInfoPtr;
{
    int  curWidth;
    Rect tileRect;

    TiToRect (tilePtr, &tileRect);
    GeoClip (&tileRect, &planeSearchRect);

    if (Tp_IsSpaceTile (tilePtr) == TRUE 
       && GEO_TOUCH (&tileRect, &updateTileRect) == TRUE)
    {
    curWidth = Tp_GetRidgeWidthOfTile (tilePtr);
    if (curWidth < ridgeInfoPtr->ridgeWidth)
    {
        ridgeInfoPtr->ridgeWidth = curWidth;
        ridgeInfoPtr->prevTilePtr = tilePtr;
/*
printf ("new ridge width = %d\n", ridgeInfoPtr->ridgeWidth);
fflush (stdout);
*/
    }
    if (ridgeInfoPtr->prevTilePtr == NIL(Tile))
    {
        if (Tp_GetPrevTileOfTile(tilePtr) != NIL(Tile) ||
        tilePtr == topmostTilePtr)
        {
        ridgeInfoPtr->ridgeWidth = curWidth;
        ridgeInfoPtr->prevTilePtr = tilePtr;
        }
    }
    }
    return (BEAR_CONTINUE_SEARCH); 
}


/*
 * ----------------------------------------------------------------------------
 *
 * UpdateVertDecompactRidge --
 *
 *     Update vertical decompaction ridge. 
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *    Update the best decompaction ridge found so far.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
UpdateVertDecompactRidge(tilePtr, ridgeInfoPtr)

register Tile      *tilePtr;
register RidgeInfo *ridgeInfoPtr;
{
    int  curWidth;
    Rect tileRect;

    TiToRect (tilePtr, &tileRect);
    GeoClip (&tileRect, &planeSearchRect);

    if (Tp_IsSpaceTile (tilePtr) == TRUE 
       && GEO_TOUCH (&tileRect, &updateTileRect) == TRUE)
    {
    curWidth = Tp_GetRidgeWidthOfTile (tilePtr);
    if (curWidth < ridgeInfoPtr->ridgeWidth)
    {
        ridgeInfoPtr->ridgeWidth = curWidth;
        ridgeInfoPtr->prevTilePtr = tilePtr;
/*
printf ("new ridge width = %d\n", ridgeInfoPtr->ridgeWidth);
fflush (stdout);
*/
    }
    if (ridgeInfoPtr->prevTilePtr == NIL(Tile))
    {
        if (Tp_GetPrevTileOfTile(tilePtr) != NIL(Tile) ||
        tilePtr == leftmostTilePtr)
        {
        ridgeInfoPtr->ridgeWidth = curWidth;
        ridgeInfoPtr->prevTilePtr = tilePtr;
        }
    }
    }
    return (BEAR_CONTINUE_SEARCH); 
}


/* ----------------------------------------------------------------------------
 *
 * FindDecompactRidge --
 *
 *     Find the last tile on the decompaction ridge 
 *
 * Results:
 *     Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *     aborting the search.
 *     Return the ridge width and the last tile of the ridge.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
FindDecompactRidge (tilePtr, ridgeInfoPtr)

register Tile      *tilePtr;
register RidgeInfo *ridgeInfoPtr;
{
    int  curWidth;
    Rect tileRect;

/*
printf ("Enter FindDecompactRidge - \n"); 
fflush (stdout);
*/

    TiToRect (tilePtr, &tileRect);
    GeoClip (&tileRect, &planeSearchRect);

    if (Tp_IsSpaceTile (tilePtr) && GEO_TOUCH (&tileRect, &planeSearchRect))
    {
    curWidth = Tp_GetRidgeWidthOfTile (tilePtr);
    if (curWidth < ridgeInfoPtr->ridgeWidth)
    {
        ridgeInfoPtr->ridgeWidth = curWidth;
        ridgeInfoPtr->prevTilePtr = tilePtr;
/*
printf ("new ridge width = %d\n", ridgeInfoPtr->ridgeWidth);
fflush (stdout);
*/
    }
    }
    return (BEAR_CONTINUE_SEARCH); 
}


/*
 * ----------------------------------------------------------------------------
 *
 * SearchHorzDecompactRidge --
 *
 *    Search for a horizontal decompaction ridge using dynamic 
 *    programming. 
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *
 * Side effects:
 *    Record the decompaction ridge width so far and 
 *    correponding back tracing pointer of the tile.
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus
SearchHorzDecompactRidge (horzTilePtr, clientData)

register Tile *horzTilePtr;
ClientData    clientData;
{
    RidgeInfo ridgeInfo;
    int       ridgeWidth;
    Rect      searchRect;
/*
printf ("Enter SearchVertDecompactRidge - \n"); 
fflush (stdout);
*/
    TITORECT (horzTilePtr, &updateTileRect);
    GeoClip (&updateTileRect, &planeSearchRect);

    if (Tp_IsSpaceTile (horzTilePtr)) 
    {
/*
printf ("Is space tile \n"); 
fflush (stdout);
*/
    ridgeWidth = Tp_GetDensityOfTile (horzTilePtr)
             - Tp_GetCapacityOfTile (horzTilePtr); 

    if (ridgeWidth > 0)
    {
        Tp_PutRidgeWidthOfTile(ridgeWidth, horzTilePtr);
    }
    else
    {
        Tp_PutRidgeWidthOfTile(INFINITY, horzTilePtr);
    }

    /*
    if (ridgeWidth > 0)
    {
    */
        TITORECT (horzTilePtr, &updateTileRect);
        GeoClip (&updateTileRect, &planeSearchRect);
/*
Mac_PrintRect (&updateTileRect);
*/
        Mac_CopyRect(&searchRect, &updateTileRect);
        searchRect.r_ybot = searchRect.r_ytop;
        searchRect.r_ytop += 1;
/*
Mac_PrintRect (&searchRect);
*/

        if (GEO_SURROUND(&planeSearchRect, &searchRect) == TRUE)
        {
        ridgeInfo.ridgeWidth = INFINITY;
        ridgeInfo.prevTilePtr = NIL (Tile);
        (void) TiSrArea((Tile *) NULL, 
            planePtr,
            &searchRect,
            UpdateHorzDecompactRidge, 
            (ClientData) &ridgeInfo);

        ridgeWidth = MIN(Tp_GetRidgeWidthOfTile (horzTilePtr), ridgeInfo.ridgeWidth);

        Tp_PutRidgeWidthOfTile (ridgeWidth, horzTilePtr);
        Tp_PutPrevTileOfTile (ridgeInfo.prevTilePtr, horzTilePtr);
        }
    /*
    }
    */
    }
    return (BEAR_CONTINUE_SEARCH); 
}


/*
 * ----------------------------------------------------------------------------
 *
 * SearchVertDecompactRidge --
 *
 *    Search for a vertical decompaction ridge using dynamic 
 *    programming. 
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *
 * Side effects:
 *    Record the decompaction ridge width so far and 
 *    correponding back tracing pointer of the tile.
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus 
SearchVertDecompactRidge (vertTilePtr, clientData)

register Tile *vertTilePtr;
ClientData    clientData;
{
    RidgeInfo ridgeInfo;
    int       ridgeWidth;
    Rect      searchRect;

/*
printf ("Enter SearchVertDecompactRidge - \n"); 
fflush (stdout);
*/

    if (Tp_IsSpaceTile (vertTilePtr)) 
    {
/*
printf ("Is space tile \n"); 
fflush (stdout);
*/
    ridgeWidth = Tp_GetDensityOfTile (vertTilePtr) - 
             Tp_GetCapacityOfTile (vertTilePtr); 

    if (ridgeWidth > 0)
    {
        Tp_PutRidgeWidthOfTile(ridgeWidth, vertTilePtr);
    }
    else
    {
        Tp_PutRidgeWidthOfTile(INFINITY, vertTilePtr);
    }

    /*
    if (ridgeWidth > 0)
    {
    */
        TiToRect(vertTilePtr, &updateTileRect);
        GeoClip(&updateTileRect, &planeSearchRect);
/*
Mac_PrintRect(&updateTileRect);
*/
        Mac_CopyRect(&searchRect, &updateTileRect);
        searchRect.r_ybot = searchRect.r_ytop;
        searchRect.r_ytop += 1;
/*
Mac_PrintRect(&searchRect);
*/

        if (GEO_SURROUND(&planeSearchRect, &searchRect) == TRUE)
        {
        ridgeInfo.ridgeWidth = INFINITY;
        ridgeInfo.prevTilePtr = NIL (Tile);
        (void) TiSrArea((Tile *) NULL, 
                planePtr,
                &searchRect,
                UpdateVertDecompactRidge, 
                (ClientData) &ridgeInfo);

        ridgeWidth = MIN(Tp_GetRidgeWidthOfTile(vertTilePtr), 
                 ridgeInfo.ridgeWidth);

        Tp_PutRidgeWidthOfTile(ridgeWidth, vertTilePtr);
        Tp_PutPrevTileOfTile(ridgeInfo.prevTilePtr, vertTilePtr);
        }
    /*
    }
    */
    }
    return (BEAR_CONTINUE_SEARCH); 
}


/* ----------------------------------------------------------------------------
 *
 * Gs_HorzDecompact --
 *
 *     The main routine to perform horizontal decompaction,
 *
 * Results:
 *     None.
 *
 * Side effects:
 *     Whatever the side effects resulted from the decompaction.
 *
 * ----------------------------------------------------------------------------
 */

SpacingStatus
Gs_HorzDecompact(planeDataPtr)
PlaneData *planeDataPtr;
{
#ifndef BEAR_BLACK_BOX
    ChipWin     *chipWinPtr;
    DrawRidge   drawRidge;
#endif BEAR_BLACK_BOX

    RidgeInfo   ridgeInfo;
    LinkedTile  *ridgeHead;
    Rect        *horzPlaneBndPtr;
    Point       topPoint;
    CellMaster  *cellMasterPtr;
    CellInst    *coreInstPtr;

/*
printf ("Enter Gs_HorzDecompact - \n"); 
fflush (stdout);
*/

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        chipWinPtr = PbGetChipWin(globalWinDataPtr);
        cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    }
    else
#endif BEAR_BLACK_BOX
    {
        cellMasterPtr = pbCellMasterPtr;
        if (cellMasterPtr == NIL(CellMaster)) {
  Bear_PrintFatalErr("Gs_HorzDecompact() failed : pbCellMasterPtr is NULL.\n");
        }
    }
    coreInstPtr = Mac_GetInstOfCoreMaster(cellMasterPtr);

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && gs_DrawRidge) {
        drawRidge.displayFunc = GXcopy;
        drawRidge.planeDataPtr = planeDataPtr;
        drawRidge.ridgeColor = ez_HighLightColor;
    }
#endif BEAR_BLACK_BOX

    gs_Quit = FALSE;

    globalPlaneDataPtr = planeDataPtr;
    planePtr = Tp_GetHorzPlane (planeDataPtr);
    horzPlaneBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    Mac_CopyRect(&planeSearchRect, horzPlaneBndPtr);

    /*
    Tp_FindBBoxForAllCells (planeDataPtr, &planeSearchRect);
    planeSearchRect.r_ytop = planeSearchRect.r_ytop + 1;
    */
    Mac_CopyRect(&tailRect, &planeSearchRect);
    tailRect.r_ytop = tailRect.r_ybot + 1;

    topPoint.p_x = planeSearchRect.r_xbot;
    topPoint.p_y = planeSearchRect.r_ytop;
    topmostTilePtr = TiSrPointNoHint(planePtr, &topPoint);

/*
printf ("Horizontal Decompaction ...\n ");
fflush (stdout);
 */

    /* Search the horizontal decompaction ridge */
    (void) TiSrArea(NIL(Tile),
             planePtr, 
             &planeSearchRect,
             SearchHorzDecompactRidge, 
             NIL (ClientData));

    /* Find the decompaction ridge */
    ridgeInfo.ridgeWidth = INFINITY;
    ridgeInfo.prevTilePtr = NIL(Tile);
    (void) TiSrArea(NIL(Tile),
            planePtr, 
            &tailRect,
            FindDecompactRidge, 
            (ClientData) &ridgeInfo);

/*
printf ("ridge width =  %d\n", ridgeInfo.ridgeWidth); 
fflush (stdout);
 */

    /* Trace the decompaction ridge */
    if (ridgeInfo.ridgeWidth > 0 && ridgeInfo.ridgeWidth < INFINITY)
    {

    GsBuildRidgeTileList (ridgeInfo.prevTilePtr, &ridgeHead);

/*
printf ("Drawing the ridge ...\n");
fflush (stdout);
*/

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && gs_DrawRidge) {
        /* Draw the decompaction ridge */
        drawRidge.tileList = ridgeHead;
        (void) PbItrChipWindows(PbDrawHorzRidge, (ClientData) &drawRidge);
     
        /*
         * Store the horizontal tile list with the window so that
         * the ridge can be redrawn automatically if the window
         * needs to be updated.
         */
        PbPutHorzTileListOfCHW(ridgeHead, chipWinPtr);

        Gs_Pause(); 
        PbPutHorzTileListOfCHW(NIL(LinkedTile), chipWinPtr);
    }
    else
#endif BEAR_BLACK_BOX
    {
        Gs_Pause(); 
    }

    if (gs_Quit == TRUE)
    {
        return (GS_QUIT);
    }

/*
printf ("Spacing the horizontal decompaction ridge ...\n ");
fflush (stdout);
*/

    /*
    PbResizeCoreCell(planeDataPtr, coreInstPtr, ridgeInfo.ridgeWidth, 0,0,0);
    */

    (void) Tp_ResizeParentCell(0,0,ridgeInfo.ridgeWidth,0,
        coreInstPtr, planeDataPtr);
    if (Tp_PushHorz(globalPlaneDataPtr, ridgeHead, 
            ridgeInfo.ridgeWidth) == FALSE) {
#ifdef DEBUG
        printf ("Tp_PushHorz failed.\n");
        fflush (stdout);
#endif DEBUG
    }

    /*
    planeSearchRect.r_xtop += ridgeInfo.ridgeWidth;
    globalWinDataPtr->wd_imageBnd.r_xtop += ridgeInfo.ridgeWidth;
    (void) Ez_CalcViewingTrans (globalWinDataPtr);
    */

        /* Update tile capacity */
    GsUpdateCapOfTileList (ridgeHead, ridgeInfo.ridgeWidth);
    return (GS_HDC_NOT_FINISHED);
    }
    else
    {
    return (GS_HDC_FINISHED);
    }

/*
printf ("Exit Gs_HorzDecompact - \n"); 
fflush (stdout);
*/
}


/* ----------------------------------------------------------------------------
 *
 * Gs_VertDecompact --
 *
 *     The main routine to perform vertical decompaction,
 *
 * Results:
 *     None.
 *
 * Side effects:
 *     Whatever the side effects resulted from the decompaction.
 *
 * ----------------------------------------------------------------------------
 */

SpacingStatus
Gs_VertDecompact(planeDataPtr)
PlaneData *planeDataPtr;
{
#ifndef BEAR_BLACK_BOX
    ChipWin     *chipWinPtr;
    DrawRidge  drawRidge;
#endif BEAR_BLACK_BOX

    RidgeInfo  ridgeInfo;
    LinkedTile *ridgeHead;
    Rect       *horzPlaneBndPtr;
    Rect       tailRectH;
    Rect       tailRect;
    Point      leftPoint;
    CellMaster  *cellMasterPtr;
    CellInst    *coreInstPtr;

/*
printf ("Enter Gs_VertDecompact - \n"); 
fflush (stdout);
*/
#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        chipWinPtr = PbGetChipWin(globalWinDataPtr);
        cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    }
    else
#endif BEAR_BLACK_BOX
    {
        cellMasterPtr = pbCellMasterPtr;
        if (cellMasterPtr == NIL(CellMaster)) {
  Bear_PrintFatalErr("Gs_VertDecompact() failed : pbCellMasterPtr is NULL.\n");
        }
    }
    coreInstPtr = Mac_GetInstOfCoreMaster(cellMasterPtr);

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && gs_DrawRidge) {
        drawRidge.displayFunc = GXcopy;
        drawRidge.planeDataPtr = planeDataPtr;
        drawRidge.ridgeColor = ez_HighLightColor;
    }
#endif BEAR_BLACK_BOX

    gs_Quit = FALSE;

    globalPlaneDataPtr = planeDataPtr;
    planePtr = Tp_GetVertPlane (planeDataPtr);
    horzPlaneBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    Tp_HRectToVRect(horzPlaneBndPtr, &planeSearchRect);
    /*
    Tp_FindBBoxForAllCells (planeDataPtr, &bbox);
    bbox.r_xbot = bbox.r_xbot - 1;
    Tp_HRectToVRect(&bbox, &planeSearchRect);
    Mac_CopyRect(&tailRect, &planeSearchRect);
    tailRect.r_ytop = tailRect.r_ybot + 1;
    */
    Mac_CopyRect(&tailRectH, horzPlaneBndPtr);
    tailRectH.r_xbot = tailRectH.r_xtop - 1;
    Tp_HRectToVRect(&tailRectH, &tailRect);

    leftPoint.p_x = planeSearchRect.r_xbot;
    leftPoint.p_y = planeSearchRect.r_ytop;
    leftmostTilePtr = TiSrPointNoHint(planePtr, &leftPoint);

/*
 *  printf ("Vertical decompaction ...\n ");
 *  fflush (stdout);
 */

    /* Search the vertical decompaction ridge */
    (void) TiSrArea(NIL(Tile),
             planePtr, 
             &planeSearchRect,
             SearchVertDecompactRidge, 
             NIL (ClientData));

    /* Find the decompaction ridge */
    ridgeInfo.ridgeWidth = INFINITY;
    ridgeInfo.prevTilePtr = NIL(Tile);
    (void) TiSrArea(NIL(Tile),
            planePtr, 
            &tailRect,
            FindDecompactRidge, 
            (ClientData) &ridgeInfo);


/*
 *  printf ("ridge width =  %d\n", ridgeInfo.ridgeWidth); 
 *  fflush (stdout);
 */

    /* Trace the decompaction ridge */
    if (ridgeInfo.ridgeWidth > 0 && ridgeInfo.ridgeWidth < INFINITY)
    {

    GsBuildRidgeTileList (ridgeInfo.prevTilePtr, &ridgeHead);
/*
printf ("Drawing the ridge ...\n");
fflush (stdout);
*/

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && gs_DrawRidge) {
        /* Draw the decompaction ridge */
        drawRidge.tileList = ridgeHead;
        (void) PbItrChipWindows(PbDrawVertRidge, (ClientData) &drawRidge);

        /*
         * Store the vertical tile list with the window so that
         * the ridge can be redrawn automatically if the window
         * needs to be updated.
         */
        PbPutVertTileListOfCHW(ridgeHead, chipWinPtr);

        Gs_Pause(); 
        PbPutVertTileListOfCHW(NIL(LinkedTile), chipWinPtr);
    }
    else
#endif BEAR_BLACK_BOX
    {
        Gs_Pause(); 
    }

    if (gs_Quit == TRUE)
    {
        return (GS_QUIT);
    }

/*
printf ("Spacing the vertical decompaction ridge ...\n ");
fflush (stdout);
*/

    /*
        PbResizeCoreCell(planeDataPtr, coreInstPtr, 0,ridgeInfo.ridgeWidth,0,0);
    */
    (void) Tp_ResizeParentCell(0,0,0,ridgeInfo.ridgeWidth,
        coreInstPtr, planeDataPtr);
    if (Tp_PushVert(globalPlaneDataPtr, ridgeHead, 
        ridgeInfo.ridgeWidth)== FALSE) {
#ifdef DEBUG
        printf ("Tp_PushVert failed.\n");
        fflush (stdout);
#endif DEBUG
    }
    /*
    planeSearchRect.r_ytop += ridgeInfo.ridgeWidth;
    globalWinDataPtr->wd_imageBnd.r_ytop += ridgeInfo.ridgeWidth;
    (void) Ez_CalcViewingTrans(globalWinDataPtr);
    */

    /* Update tile capacity */
    GsUpdateCapOfTileList (ridgeHead, ridgeInfo.ridgeWidth);
    return (GS_VDC_NOT_FINISHED);
    }
    else
    {
        return (GS_VDC_FINISHED);
    }
/*
printf ("Exit Gs_VertDecompact - \n"); 
fflush (stdout);
*/
}


