
/* gsMisc.c --
 *
 *     Wei-Ming Dai
 *     (Professor E. S. Kuh)
 *
 * This module performs global compaction on a given placement.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/globalSpacer/gsMisc.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"         /*temp*/
#include "tileProp.h"
#include "mac.h"
#include "globalSpacer.h"
#include "ezPlot.h"  /*temp */


/* ------- External global variables for the globalSpacer module ------- */
bool gs_Quit = FALSE;
bool gs_DrawRidge = TRUE;


/* ---------- Local Procedures ---------- */
static int CheckBoundaryTiles();


/* ---------- Local Global Variables ---------- */
static bool pauseFlag = FALSE;


/* ----------------------------------------------------------------------------
 *
 * GsBuildRidgeTileList() --
 *
 *     Build a list of tiles of the ridge starting
 *     from the `ridgePtr' in order.
 *
 * Results:
 *     A list of tiles of the ridge. 
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

void
GsBuildRidgeTileList(tilePtr, ridgeHead)

Tile       *tilePtr;
LinkedTile **ridgeHead;
{
    register Tile       *curTilePtr;
    register LinkedTile *curLinkedTilePtr;

/*
printf ("Enter GsBuildRidgeTileList - \n"); 
fflush (stdout);
*/

    MALLOC(LinkedTile *, curLinkedTilePtr, sizeof (LinkedTile));
    curLinkedTilePtr->tilePtr = tilePtr;
    curLinkedTilePtr->nextPtr = NIL (LinkedTile);
    *ridgeHead = curLinkedTilePtr;

    curTilePtr =  Tp_GetPrevTileOfTile (tilePtr);

    while (curTilePtr != NIL (Tile))
    {
    MALLOC(LinkedTile *, curLinkedTilePtr, sizeof (LinkedTile));
        curLinkedTilePtr->tilePtr = curTilePtr;
    curLinkedTilePtr->nextPtr = *ridgeHead;
    *ridgeHead = curLinkedTilePtr; 
    curTilePtr =  Tp_GetPrevTileOfTile (curTilePtr);
    }
}


/* ----------------------------------------------------------------------------
 *
 * GsUpdateCapOfTileList() --
 *
 *     Update the capacity of tiles in a list
 *     using the specified offset
 *
 * Results:
 *     None
 *
 * Side effects:
 *     The capacities of the tiles in the list 
 *     are updated
 *
 * ----------------------------------------------------------------------------
 */

void 
GsUpdateCapOfTileList(tileList, capOffset)

LinkedTile *tileList;
int        capOffset;
{
    LinkedTile *tempLinkedTile;
    int        newCap;

    tempLinkedTile = tileList; 
    while (tempLinkedTile != NIL (LinkedTile))
    {
     newCap = Tp_GetCapacityOfTile (tempLinkedTile->tilePtr) - capOffset;
     Tp_PutCapacityOfTile (newCap, tempLinkedTile->tilePtr);
     tempLinkedTile = tempLinkedTile->nextPtr;
    }

}



void
Gs_SetPause(flag)
bool flag;
{
    pauseFlag = flag;
}



/* 
 * ----------------------------------------------------------------------------
 *
 * Gs_Pause() --
 *
 *     Pause for giving user the choises
 *     of continuing, quitting or aborting
 *
 * Results:
 *     None
 *
 * Side effects:
 *     Continue: none
 *     Quit:     Gs_quit = TRUE
 *     Abort:    A "core" file after abort
 *
 * ----------------------------------------------------------------------------
 */

void
Gs_Pause()
{
#ifndef BEAR_BLACK_BOX
    Window  win;
    char    string[80];
    char    letter[10];
    Point   screen;
    bool    keepGoing = TRUE;

    if (bear_BlackBoxFlag || !pauseFlag) {
        return;
    }

#ifdef DEBUG
    Ez_Print(EZ_NORMAL_FONT,
        "Type 'c' to continue, 'q' to quit, 'a' to abort: ");
#else
    Ez_Print(EZ_NORMAL_FONT,
        "Type 'c' to continue, 'q' to quit: ");
#endif DEBUG

    (void) Ez_GetLine(&win, &screen, string, 80, &keepGoing);
    sscanf(string, "%1s", letter);
    switch (letter[0]) {
#ifdef DEBUG
    case 'a':
        abort();
        break;
#endif DEBUG
    case 'c':
        break;
    case 'q':
        gs_Quit = TRUE;
        break;
    default:
        break;
    }
#endif BEAR_BLACK_BOX
}



void
Gs_HResizeParentCell(planeDataPtr, coreInstPtr)
PlaneData     *planeDataPtr;
CellInst      *coreInstPtr;
{
    Rect      *bboxPtr;
    Rect      searchRect;
    int       tempNum = INFINITY;

    bboxPtr = Mac_GetBBoxOfCell(coreInstPtr);
    searchRect.r_xbot = bboxPtr->r_xtop - 1;
    searchRect.r_ybot = bboxPtr->r_ybot;
    searchRect.r_xtop = bboxPtr->r_xtop;
    searchRect.r_ytop = bboxPtr->r_ytop;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
            &searchRect, CheckBoundaryTiles, (ClientData) &tempNum);
            /*
    PbResizeCoreCell(planeDataPtr, coreInstPtr, -tempNum, 0,0,0);
    */
    Tp_ResizeParentCell(0,0,-tempNum,0,coreInstPtr, planeDataPtr);
}



void
Gs_VResizeParentCell(planeDataPtr, coreInstPtr)
PlaneData     *planeDataPtr;
CellInst      *coreInstPtr;
{
    Rect      *bboxPtr;
    Rect      searchRect;
    Rect      vSearchRect;
    int       tempNum = INFINITY;

    bboxPtr = Mac_GetBBoxOfCell(coreInstPtr);
    searchRect.r_xbot = bboxPtr->r_xbot;
    searchRect.r_ybot = bboxPtr->r_ytop - 1;
    searchRect.r_xtop = bboxPtr->r_xtop;
    searchRect.r_ytop = bboxPtr->r_ytop;
    (void) Tp_HRectToVRect(&searchRect, &vSearchRect);
    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
            &vSearchRect, CheckBoundaryTiles, (ClientData) &tempNum);
            /*
    PbResizeCoreCell(planeDataPtr, coreInstPtr, 0, -tempNum,0,0);
    */
    Tp_ResizeParentCell(0,0,0,-tempNum,coreInstPtr, planeDataPtr);
}



static int
CheckBoundaryTiles(tilePtr, tempNum)
Tile     *tilePtr;
int      *tempNum;
{
    int      density;
    int      capacity;
    int      number;

    if (Tp_IsBottleneckTile(tilePtr))
    {
        density = Tp_GetDensityOfTile(tilePtr);
        capacity = Tp_GetCapacityOfTile(tilePtr);
    number = capacity - density;
        if (*tempNum > 0)
        {
        if (number >= 0)
        {
        if (*tempNum > number)
        {
                *tempNum = number;
        }
        }
        else
        {
        *tempNum = number;
        }
        }
    else
    {
        if (number <= 0)
        {
        if (*tempNum > number)
        {
                *tempNum = number;
        }
        }
    }
    }
    return (MAGIC_CONTINUE_SEARCH);
}


