
/* greBldGraph.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/grEditor/greBldGraph.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "grEditor.h"
#include "grEditorInt.h"
#include "routeDB.h"
#include "tileMapping.h"
#include "ezPlot.h"

#include "../polarBear/polarBear.h"


typedef struct _BldGraphDataInt {
    Net             *netPtr;
    Gt_Graph        graph;

    bool            warnFlag;
    bool            placeFlag;
    bool            pgFlag;
    bool            checkRSegFlag;

    CellMaster      *cellMasterPtr;
    PlaneData       *planeDataPtr;
    FloorPlanGraph  *fpgPtr;
    RouteCell       *rcellList;
    DummyCell       *dcellList;

    st_table        *hashTable;
    lsList          nodeList;
    int             nodeNum;
    int             edgeNum;

    lsList          bnTileList; /* only applies if placeFlag is TRUE        */
    lsList          junctList;  /* only applies if placeFlag is TRUE        */

    Point           *centerPtr; /* switch box center                        */
    int             count;
    bool            horzFlag;
    bool            beginFlag;  /* only applies to AddPPinOfBnTile()        */
    Rect            absRect;    /* rectangle given in absolute coordinates  */
} BldGraphDataInt;


typedef struct _ConnData {
    Gre_BldGraphDataPtr bldDataPtr;
    bool                connFlag;
    bool                printMsgFlag;
} ConnData;


/* ---------- Local Procedures ---------- */

static Gt_GraphNode InsertKey();
static Gt_GraphNode LookUpKey();

static int          CmpHorzNodes();
static int          CmpVertNodes();

static bool         AddGraphEdge();
static void         ConnNodesOne();
static void         ConnNodesTwo();
static void         ConnNodesThree();
static BearId       GetBnTileId();
static BearId       GetDCellId();
static BearId       GetSwitchBoxId();

static BearStatus   AddPinOfNet();

static BearStatus   AddPPinOfBnLine();

static BearStatus   AddPinOfBnTile();
static BearStatus   AddPPinOfBnTile();
static BearStatus   AddRPinOfBnTile();

static BearStatus   AddPPinOfDCell();

static int          SearchSideOfBnTile();

static BearStatus   DoBottleNeck();
static BearStatus   DoBnLine();
static BearStatus   DoBnTile();
static BearStatus   DoDCell();
static BearStatus   DoSwitchBox();

static BearStatus   StoreSwitchBoxData();
static BearStatus   RemoveSwitchBoxData();

#ifdef USE_DR_CODE
static BearStatus   AddRSegmOfRRoot();
static BearStatus   DoRRoot();
#endif /* USE_DR_CODE */

#ifdef DEBUG
static void         PrintGraph();
static BearStatus   PrintEdge();
static BearStatus   PrintNode();
#endif DEBUG

/* ---------- Local Global Variables ---------- */
static PlaneData    *globalPlaneDataPtr;


#ifdef DEBUG
static bool genFlag = FALSE;
static bool dupFlag = FALSE;
#endif DEBUG


static Gt_GraphNode
InsertKey(keyPtr, type, intBldDataPtr)
char            *keyPtr;
int             type;   /* FPG_PIN || FPG_PSEUDO_PIN || FPG_ROUTE_PIN
                         * FPG_BBL_TERM || FPG_BBL_JUNCT
                         */
BldGraphDataInt *intBldDataPtr;
{
    Gt_GraphNode    node;
    PinOrJunct      *pinOrJunctPtr;

    pinOrJunctPtr = Fpg_MakePinOrJunct();
    switch (type) {
    case FPG_PSEUDO_JCT:
        pinOrJunctPtr->unType = FPG_PSEUDO_JCT;
        pinOrJunctPtr->unValue.pjunctPtr = (PseudoJunct *) keyPtr;
        break;
    case FPG_PIN:
        pinOrJunctPtr->unType = FPG_PIN;
        pinOrJunctPtr->unValue.pinPtr = (PinInst *) keyPtr;
        break;
    case FPG_PSEUDO_PIN:
        pinOrJunctPtr->unType = FPG_PSEUDO_PIN;
        pinOrJunctPtr->unValue.ppinPtr = (PseudoPin *) keyPtr;
        break;
    case FPG_ROUTE_PIN:
        pinOrJunctPtr->unType = FPG_ROUTE_PIN;
        pinOrJunctPtr->unValue.rPinPtr = (RoutePin *) keyPtr;
        break;
    case FPG_BBL_TERM:
        pinOrJunctPtr->unType = FPG_BBL_TERM;
        pinOrJunctPtr->unValue.bblTermPtr = (term *) keyPtr;
        break;
    case FPG_BBL_JUNCT:
        pinOrJunctPtr->unType = FPG_BBL_JUNCT;
        pinOrJunctPtr->unValue.bblJunctPtr = (rjun *) keyPtr;
        break;
    default:
        Bear_PrintFatalErr("InsertKey() : unexpected pin or junction type");
        break;
    }
    node = Gt_AddGraphNode(intBldDataPtr->graph,
        (BearId) intBldDataPtr->nodeNum,
        (BearGeneric) pinOrJunctPtr);
    (intBldDataPtr->nodeNum)++;

    (void) st_insert(intBldDataPtr->hashTable, keyPtr, (char *) node);

    return (node);
}


static Gt_GraphNode
LookUpKey(keyPtr, type, intBldDataPtr)
char            *keyPtr;
int             type;   /* FPG_PIN || FPG_PSEUDO_PIN || FPG_ROUTE_PIN ||
                         * FPG_BBL_TERM || FPG_BBL_JUNCT
                         */
BldGraphDataInt *intBldDataPtr;
{
    Gt_GraphNode node;

    if (st_lookup(intBldDataPtr->hashTable, keyPtr, (char **) &node) == 0) {
        return (InsertKey(keyPtr, type, intBldDataPtr));
    }
    return (node);
}


static int
CmpHorzNodes(item1, item2)
lsGeneric item1, item2;
/*
 * Compare horizontal nodes.
 */
{
    PinOrJunct      *pinOrJunctPtr;
    Point           absPos;
    register int    x1, x2;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode((Gt_GraphNode) item1);
    (void) Fpg_GetAbsPosOfPinOrJunct(globalPlaneDataPtr, 
        pinOrJunctPtr, &absPos);
    x1 = absPos.p_x;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode((Gt_GraphNode) item2);
    (void) Fpg_GetAbsPosOfPinOrJunct(globalPlaneDataPtr, 
        pinOrJunctPtr, &absPos);
    x2 = absPos.p_x;

    return ((x1 == x2) ? 0 : ((x1 < x2) ? -1 : 1));
}


static int
CmpVertNodes(item1, item2)
lsGeneric item1, item2;
/*
 * Compare vertical nodes.
 */
{
    PinOrJunct      *pinOrJunctPtr;
    Point           absPos;
    register int    y1, y2;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode((Gt_GraphNode) item1);
    (void) Fpg_GetAbsPosOfPinOrJunct(globalPlaneDataPtr,
        pinOrJunctPtr, &absPos);
    y1 = absPos.p_y;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode((Gt_GraphNode) item2);
    (void) Fpg_GetAbsPosOfPinOrJunct(globalPlaneDataPtr,
        pinOrJunctPtr, &absPos);
    y2 = absPos.p_y;

    return ((y1 == y2) ? 0 : ((y1 < y2) ? -1 : 1));
}


/* ARGSUSED */
static BearId
GetBnTileId(intBldDataPtr, node)
BldGraphDataInt *intBldDataPtr;
Gt_GraphNode    node;
{
    PinOrJunct  *pinOrJunctPtr;
    PseudoPin   *ppinPtr;
    Tile        *tilePtr;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    switch (pinOrJunctPtr->unType) {
    case FPG_PIN:
        return (Mac_GetExtNetIdOfPin(pinOrJunctPtr->unValue.pinPtr));
    case FPG_PSEUDO_PIN:
        ppinPtr = pinOrJunctPtr->unValue.ppinPtr;
        tilePtr = Tp_GetTileOfPPin(ppinPtr);
        if (Tp_IsBottleneckTile(tilePtr) == TRUE) {
            /* bottleneck tile*/
            return (Tp_GetIntNetIdOfPPin(ppinPtr));
        }
        else {
            /* dummy cell tile */
            return (Tp_GetExtNetIdOfPPin(ppinPtr));
        }
    case FPG_ROUTE_PIN:
        return (Mac_GetExtNetIdOfRPin(pinOrJunctPtr->unValue.rPinPtr));
    default:
        Bear_PrintFatalErr("GetBnTileId() : unexpected pin or junction type");
        break;
    }
    /* NOTREACHED */
}


/* ARGSUSED */
static BearId
GetDCellId(intBldDataPtr, node)
BldGraphDataInt *intBldDataPtr;
Gt_GraphNode    node;
{
    PinOrJunct *pinOrJunctPtr;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    switch (pinOrJunctPtr->unType) {
    case FPG_PSEUDO_PIN:
        return (Tp_GetIntNetIdOfPPin(pinOrJunctPtr->unValue.ppinPtr));
    default:
        Bear_PrintFatalErr("GetDCellId() : unexpected pin or junction type");
        break;
    }
    /* NOTREACHED */
}


static BearId
GetSwitchBoxId(intBldDataPtr, node)
BldGraphDataInt *intBldDataPtr;
Gt_GraphNode    node;
{
    PinOrJunct  *pinOrJunctPtr;
    PseudoPin   *ppinPtr;
    Tile        *tilePtr;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    switch (pinOrJunctPtr->unType) {
    case FPG_PIN:
        return (Mac_GetExtNetIdOfPin(pinOrJunctPtr->unValue.pinPtr));
    case FPG_PSEUDO_PIN:
        ppinPtr = pinOrJunctPtr->unValue.ppinPtr;
        tilePtr = Tp_GetTileOfPPin(ppinPtr);
        if (Tp_IsBottleneckTile(tilePtr) == TRUE) {
            /* bottleneck tile */
            return (Tp_GetExtNetIdOfPPin(ppinPtr));
        }
        else {
            /* bottleneck line */
            Point absPos;
/*
 * The following picture corresponds to a junction surrounded by
 * four bottleneck lines, and shows how the id's of pseudo pins
 * attached to those lines are defined.
 *
 *          e
 *    -------------
 *   |      i      |
 *   |             |
 *   |             |
 * i | e    J    i | e
 *   |             |
 *   |             |
 *   |      e      |
 *    -------------
 *          i      
 */
            Tp_GetAbsPosOfPPin(intBldDataPtr->planeDataPtr, ppinPtr, &absPos);
            if (Tp_IsHorzTile(tilePtr)) {
                return ((absPos.p_y > intBldDataPtr->centerPtr->p_y) ?
                    Tp_GetIntNetIdOfPPin(ppinPtr) :
                    Tp_GetExtNetIdOfPPin(ppinPtr));
            }
            else {
                return ((absPos.p_x > intBldDataPtr->centerPtr->p_x) ?
                    Tp_GetIntNetIdOfPPin(ppinPtr) :
                    Tp_GetExtNetIdOfPPin(ppinPtr));
            }
        }
    case FPG_ROUTE_PIN:
        return (Mac_GetExtNetIdOfRPin(pinOrJunctPtr->unValue.rPinPtr));
    default:
      Bear_PrintFatalErr("GetSwitchBoxId() : unexpected pin or junction type");
        break;
    }
    /* NOTREACHED */
}


static bool
AddGraphEdge(intBldDataPtr, node1, node2, edgePtr)
BldGraphDataInt *intBldDataPtr;
Gt_GraphNode    node1, node2;
Gt_GraphEdge    *edgePtr;
{
    Mac_EdgeData *edgeDataPtr;

    edgeDataPtr = (intBldDataPtr->pgFlag == TRUE) ?
        Mac_MakeEdgeData() : (Mac_EdgeData *) NULL;

    if (Gt_AddGraphEdge(node1, node2, (BearId) intBldDataPtr->edgeNum,
        (BearGeneric) edgeDataPtr, edgePtr) != BEAR_OK) {
        return (FALSE);
    }

    (intBldDataPtr->edgeNum)++;

    return (TRUE);
}


static void
ConnNodesOne(intBldDataPtr, getIdProc)
BldGraphDataInt *intBldDataPtr;
register BearId (*getIdProc)();
/*
 * Method one for connecting together nodes in intBldDataPtr->nodeList.
 */
{
    BearId          id1;
    Gt_GraphNode    node1, node2;
    lsGen           gen;

    while (lsDelBegin(intBldDataPtr->nodeList, (lsGeneric *) &node1) == LS_OK) {
        id1 = (*getIdProc)(intBldDataPtr, node1);

        gen = lsStart(intBldDataPtr->nodeList);
        while (lsNext(gen, (lsGeneric *) &node2, LS_NH) == LS_OK) {
            if (node1 == node2) {
                lsDelBefore(gen, (lsGeneric *) NULL);
                continue;
            }

            if (id1 == (*getIdProc)(intBldDataPtr, node2)) {
                lsDelBefore(gen, (lsGeneric *) NULL);
                (void) AddGraphEdge(intBldDataPtr, node1, node2,
                    (Gt_GraphEdge *) NULL);
            }
        }
        lsFinish(gen);
    }
}


static void
ConnNodesTwo(intBldDataPtr, getIdProc)
BldGraphDataInt *intBldDataPtr;
register BearId (*getIdProc)();
/*
 * Method two for connecting together nodes in intBldDataPtr->nodeList.
 */
{
    BearId          id1;
    Gt_GraphNode    node1, node2;
    PseudoJunct     *pjunctPtr;
    lsGen           gen;

    while (lsDelBegin(intBldDataPtr->nodeList, (lsGeneric *) &node1) == LS_OK) {
        id1 = (*getIdProc)(intBldDataPtr, node1);

        pjunctPtr = Fpg_MakePseudoJunct();
        node2 = InsertKey((char *) pjunctPtr, FPG_PSEUDO_JCT, intBldDataPtr);
        pjunctPtr->absPos.p_x = intBldDataPtr->centerPtr->p_x;
        pjunctPtr->absPos.p_y = intBldDataPtr->centerPtr->p_y;
        pjunctPtr->id = id1;
        (void) AddGraphEdge(intBldDataPtr, node1, node2, (Gt_GraphEdge *) NULL);
        node1 = node2;

        gen = lsStart(intBldDataPtr->nodeList);
        while (lsNext(gen, (lsGeneric *) &node2, LS_NH) == LS_OK) {
            if (node1 == node2) {
                lsDelBefore(gen, (lsGeneric *) NULL);
                continue;
            }

            if (id1 == (*getIdProc)(intBldDataPtr, node2)) {
                lsDelBefore(gen, (lsGeneric *) NULL);
                (void) AddGraphEdge(intBldDataPtr, node1, node2,
                    (Gt_GraphEdge *) NULL);
            }
        }
        lsFinish(gen);
    }
}


static void
ConnNodesThree(intBldDataPtr, getIdProc)
BldGraphDataInt *intBldDataPtr;
register BearId (*getIdProc)();
/*
 * Method three for connecting together nodes in intBldDataPtr->nodeList.
 */
{
    BearId          id1;
    Gt_GraphNode    node1, node2;
    lsGen           gen;

    while (lsDelBegin(intBldDataPtr->nodeList, (lsGeneric *) &node1) == LS_OK) {
        id1 = (*getIdProc)(intBldDataPtr, node1);

        gen = lsStart(intBldDataPtr->nodeList);
        while (lsNext(gen, (lsGeneric *) &node2, LS_NH) == LS_OK) {
            if (node1 == node2) {
                lsDelBefore(gen, (lsGeneric *) NULL);
                continue;
            }

            if (id1 == (*getIdProc)(intBldDataPtr, node2)) {
                lsDelBefore(gen, (lsGeneric *) NULL);
                (void) AddGraphEdge(intBldDataPtr, node1, node2,
                    (Gt_GraphEdge *) NULL);
                node1 = node2;
            }
        }
        lsFinish(gen);
    }
}


static BearStatus
AddPinOfNet(pinPtr, intBldDataPtr)
PinInst         *pinPtr;
BldGraphDataInt *intBldDataPtr;
{
    (void) InsertKey((char *) pinPtr, FPG_PIN, intBldDataPtr);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPPinOfBnLine(ppinPtr, intBldDataPtr)
PseudoPin       *ppinPtr;
BldGraphDataInt *intBldDataPtr;
{
    if (intBldDataPtr->netPtr == Tp_GetNetOfPPin(ppinPtr)) {
        (void) LookUpKey((char *) ppinPtr, FPG_PSEUDO_PIN, intBldDataPtr);

        if (intBldDataPtr->warnFlag == TRUE) {
            if (++(intBldDataPtr->count) > 1) {
                sprintf (bear_ErrorMsg, "Warning : net named \"%s\"\n",
                    Mac_GetNameOfNet(intBldDataPtr->netPtr));
                sprintf (&bear_ErrorMsg[strlen (bear_ErrorMsg)],
                    "    has two ppins on same side of bn line\n");
                Bear_PrintErr(bear_ErrorMsg);
            }
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPinOfBnTile(pinPtr, intBldDataPtr)
PinInst         *pinPtr;
BldGraphDataInt *intBldDataPtr;
{
    if (intBldDataPtr->netPtr == Mac_GetNetOfPin(pinPtr)) {
        Gt_GraphNode    node;
        Point           absPos;

        Mac_GetAbsPosOfPin(pinPtr, &absPos);

        if (intBldDataPtr->horzFlag == TRUE) {
            if (! (((absPos.p_x == intBldDataPtr->absRect.r_xbot) ||
                (absPos.p_x == intBldDataPtr->absRect.r_xtop)) &&

                (absPos.p_y <= intBldDataPtr->absRect.r_ytop) &&
                (absPos.p_y >= intBldDataPtr->absRect.r_ybot))) {
                return (BEAR_CONTINUE_SEARCH);
            }
        }
        else {
            if (! (((absPos.p_y == intBldDataPtr->absRect.r_ybot) ||
                (absPos.p_y == intBldDataPtr->absRect.r_ytop)) &&

                (absPos.p_x <= intBldDataPtr->absRect.r_xtop) &&
                (absPos.p_x >= intBldDataPtr->absRect.r_xbot))) {
                return (BEAR_CONTINUE_SEARCH);
            }
        }

        node = LookUpKey((char *) pinPtr, FPG_PIN, intBldDataPtr);
        lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPPinOfBnTile(ppinPtr, intBldDataPtr)
PseudoPin       *ppinPtr;
BldGraphDataInt *intBldDataPtr;
{
    if (intBldDataPtr->netPtr == Tp_GetNetOfPPin(ppinPtr)) {
        Gt_GraphNode node;

        node = LookUpKey((char *) ppinPtr, FPG_PSEUDO_PIN, intBldDataPtr);
        if (intBldDataPtr->beginFlag == TRUE) {
            lsNewBegin(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
        }
        else {
            lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
        }

        if (intBldDataPtr->warnFlag == TRUE) {
            if (++(intBldDataPtr->count) > 1) {
                sprintf (bear_ErrorMsg, "Warning : net named \"%s\"\n",
                    Mac_GetNameOfNet(intBldDataPtr->netPtr));
                sprintf (&bear_ErrorMsg[strlen (bear_ErrorMsg)],
                    "    has two ppins on same side of bn (or dc) tile 0x%x\n",
                    Tp_GetTileOfPPin(ppinPtr));
                Bear_PrintErr(bear_ErrorMsg);
            }
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddRPinOfBnTile(rpinPtr, intBldDataPtr)
RoutePin        *rpinPtr;
BldGraphDataInt *intBldDataPtr;
{
    if (intBldDataPtr->netPtr == Mac_GetNetPtrOfRPin(rpinPtr)) {
        Gt_GraphNode    node;
        Point           absPos;

        Mac_GetAbsPosOfRPin(rpinPtr, &absPos);

        if (intBldDataPtr->horzFlag == TRUE) {
            if (! (((absPos.p_x == intBldDataPtr->absRect.r_xbot) ||
                (absPos.p_x == intBldDataPtr->absRect.r_xtop)) &&

                (absPos.p_y <= intBldDataPtr->absRect.r_ytop) &&
                (absPos.p_y >= intBldDataPtr->absRect.r_ybot))) {
                return (BEAR_CONTINUE_SEARCH);
            }
        }
        else {
            if (! (((absPos.p_y == intBldDataPtr->absRect.r_ybot) ||
                (absPos.p_y == intBldDataPtr->absRect.r_ytop)) &&

                (absPos.p_x <= intBldDataPtr->absRect.r_xtop) &&
                (absPos.p_x >= intBldDataPtr->absRect.r_xbot))) {
                return (BEAR_CONTINUE_SEARCH);
            }
        }

        node = LookUpKey((char *) rpinPtr, FPG_ROUTE_PIN, intBldDataPtr);
        lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPPinOfDCell(ppinPtr, intBldDataPtr)
PseudoPin       *ppinPtr;
BldGraphDataInt *intBldDataPtr;
{
    if (intBldDataPtr->netPtr == Tp_GetNetOfPPin(ppinPtr)) {
        Gt_GraphNode node;

        node = LookUpKey((char *) ppinPtr, FPG_PSEUDO_PIN, intBldDataPtr);
        lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


#ifdef USE_DR_CODE
static BearStatus
AddRSegmOfRRoot(rsegPtr, intBldDataPtr)
rseg            *rsegPtr;
BldGraphDataInt *intBldDataPtr;
{
    GenericPin      *gpinPtr;
    Gt_GraphNode    node0, node1, node2;
    Gt_GraphEdge    edge;
    Mac_EdgeData    *edgeDataPtr;
    PinOrJunct      *pinOrJunctPtr;
    Point           absPos0, absPos1;
    term            *bblTermPtr;
    int             hwidth;
    bool            pgFlag;

    pgFlag = intBldDataPtr->pgFlag;

    if (rsegPtr->t0fsr) {
        if (rsegPtr->j0sr.rt->mp->ansmp == NIL(module)) {
            /*
             * This terminal belongs to a straight channel.
             */
            bblTermPtr = rsegPtr->j0sr.rt;
        }
        else {
            /*
             * This terminal belongs to an L-channel.
             */
            bblTermPtr = Mac_LinkTermToTerm(rsegPtr->j0sr.rt,
                rsegPtr->j0sr.rt->mp->ansmp);
        }

        if ((gpinPtr = (GenericPin *) bblTermPtr->genericPinPtr) !=
            NIL(GenericPin)) {

            switch (gpinPtr->pinType) {
            case PIN_INST:
                node0 = LookUpKey((char *) Mac_GetPinInstOfGPin(gpinPtr),
                    FPG_PIN, intBldDataPtr);
                break;
            case ROUTE_PIN:
                node0 = LookUpKey((char *) Mac_GetRoutePinOfGPin(gpinPtr),
                    FPG_ROUTE_PIN, intBldDataPtr);
                break;
            default:
             Bear_PrintFatalErr("AddRSegmOfRRoot() : illegal generic pin type");
                break;
            }
            if (pgFlag == TRUE) {
                hwidth = bblTermPtr->rsp->hwidth;
            }
        }
        else {
            /*
             * As far as I can tell nodes of the type FPG_BBL_TERM will
             * only appear in the graph when you have an L-channel.  They
             * form the connection between the two straight channels that
             * make up the L-channel.
             */

            node0 = LookUpKey((char *) rsegPtr->j0sr.rt, FPG_BBL_TERM,
                intBldDataPtr);
            if (pgFlag == TRUE) {
                hwidth = rsegPtr->j0sr.rt->rsp->hwidth;
            }

            node2 = LookUpKey((char *) bblTermPtr, FPG_BBL_TERM,
                intBldDataPtr);
            (void) AddGraphEdge(intBldDataPtr, node0, node2, &edge);
            if (pgFlag == TRUE) {
                edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);
                edgeDataPtr->width = 2 * hwidth;
            }
        }
    }
    else {
        node0 = LookUpKey((char *) rsegPtr->j0sr.rj, FPG_BBL_JUNCT,
            intBldDataPtr);
        if (pgFlag == TRUE) {
            hwidth = rsegPtr->j0sr.rj->sljr->hwidth;
        }
    }

    if (rsegPtr->t1fsr) {
        if (rsegPtr->j1sr.rt->mp->ansmp == NIL(module)) {
            bblTermPtr = rsegPtr->j1sr.rt;
        }
        else {
            bblTermPtr = Mac_LinkTermToTerm(rsegPtr->j1sr.rt,
                rsegPtr->j1sr.rt->mp->ansmp);
        }

        if ((gpinPtr = (GenericPin *) bblTermPtr->genericPinPtr) !=
            NIL(GenericPin)) {

            switch (gpinPtr->pinType) {
            case PIN_INST:
                node1 = LookUpKey((char *) Mac_GetPinInstOfGPin(gpinPtr),
                    FPG_PIN, intBldDataPtr);
                break;
            case ROUTE_PIN:
                node1 = LookUpKey((char *) Mac_GetRoutePinOfGPin(gpinPtr),
                    FPG_ROUTE_PIN, intBldDataPtr);
                break;
            default:
             Bear_PrintFatalErr("AddRSegmOfRRoot() : illegal generic pin type");
                break;
            }
        }
        else {
            node1 = LookUpKey((char *) rsegPtr->j1sr.rt, FPG_BBL_TERM,
                intBldDataPtr);

            node2 = LookUpKey((char *) bblTermPtr, FPG_BBL_TERM,
                intBldDataPtr);
            (void) AddGraphEdge(intBldDataPtr, node1, node2, &edge);
            if (pgFlag == TRUE) {
                edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);
                edgeDataPtr->width = 2 * hwidth;
            }
        }
    }
    else {
        node1 = LookUpKey((char *) rsegPtr->j1sr.rj, FPG_BBL_JUNCT,
            intBldDataPtr);
    }

    if (intBldDataPtr->checkRSegFlag == TRUE) {
        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node0);
        if (Fpg_GetAbsPosOfPinOrJunct(intBldDataPtr->planeDataPtr,
            pinOrJunctPtr, &absPos0) == BEAR_ERROR) {
            sprintf (bear_ErrorMsg, "Warning : net named \"%s\"\n",
                Mac_GetNameOfNet(intBldDataPtr->netPtr));
            sprintf (&bear_ErrorMsg[strlen (bear_ErrorMsg)],
                "    has a bad route segment for point 0\n");
            Bear_PrintErr(bear_ErrorMsg);
            return (BEAR_CONTINUE_SEARCH);
        }
        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node1);
        if (Fpg_GetAbsPosOfPinOrJunct(intBldDataPtr->planeDataPtr,
            pinOrJunctPtr, &absPos1) == BEAR_ERROR) {
            sprintf (bear_ErrorMsg, "Warning : net named \"%s\"\n",
                Mac_GetNameOfNet(intBldDataPtr->netPtr));
            sprintf (&bear_ErrorMsg[strlen (bear_ErrorMsg)],
                "    has a bad route segment for point 1\n");
            Bear_PrintErr(bear_ErrorMsg);
            return (BEAR_CONTINUE_SEARCH);
        }

        if ((absPos0.p_x != absPos1.p_x) &&
            (absPos0.p_y != absPos1.p_y)) {
            sprintf (bear_ErrorMsg, "Warning : net named \"%s\"\n",
                Mac_GetNameOfNet(intBldDataPtr->netPtr));
            sprintf (&bear_ErrorMsg[strlen (bear_ErrorMsg)],
                "    has a non-manhattan route segment\n");
            Bear_PrintErr(bear_ErrorMsg);

            return (BEAR_CONTINUE_SEARCH);
        }
    }

    (void) AddGraphEdge(intBldDataPtr, node0, node1, &edge);
    if (pgFlag == TRUE) {
        edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);
        edgeDataPtr->width = 2 * hwidth;
    }

    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */


static int
SearchSideOfBnTile(tilePtr, intBldDataPtr)
Tile            *tilePtr;
BldGraphDataInt *intBldDataPtr;
{
    if (Tp_IsCellInstTile(tilePtr) == TRUE) {
        (void) Mac_ItrPinOfCellInst(Tp_GetCellInstOfTile(tilePtr),
            AddPinOfBnTile, (ClientData) intBldDataPtr);
    }
    else if (Tp_IsRCellTile(tilePtr) == TRUE) {
        RouteCell *rcellPtr;

        rcellPtr = Tp_GetRCellOfTile(tilePtr);

        (void) Mac_ItrRSidePinOfRCell(rcellPtr, AddRPinOfBnTile,
            (ClientData) intBldDataPtr);
        (void) Mac_ItrTLSidePinOfRCell(rcellPtr, AddRPinOfBnTile,
            (ClientData) intBldDataPtr);
    }

    return (MAGIC_CONTINUE_SEARCH);
}


static BearStatus
DoDCell(dcellPtr, intBldDataPtr)
DummyCell       *dcellPtr;
BldGraphDataInt *intBldDataPtr;
{
    intBldDataPtr->nodeList = lsCreate();

    (void) Mac_ItrPPinOfDCell(dcellPtr, TOP_SIDE, AddPPinOfDCell,
        (ClientData) intBldDataPtr);
    
    (void) Mac_ItrPPinOfDCell(dcellPtr, BOTTOM_SIDE, AddPPinOfDCell,
        (ClientData) intBldDataPtr);
    
    (void) Mac_ItrPPinOfDCell(dcellPtr, LEFT_SIDE, AddPPinOfDCell,
        (ClientData) intBldDataPtr);
    
    (void) Mac_ItrPPinOfDCell(dcellPtr, RIGHT_SIDE, AddPPinOfDCell,
        (ClientData) intBldDataPtr);

    if (intBldDataPtr->pgFlag == FALSE) {
        ConnNodesOne(intBldDataPtr, GetDCellId);
    }
    else {
        ConnNodesTwo(intBldDataPtr, GetDCellId);
    }


    lsDestroy(intBldDataPtr->nodeList, (void (*)()) NULL);
    intBldDataPtr->nodeList = (lsList) NULL;

    return (BEAR_CONTINUE_SEARCH);
}


#ifdef USE_DR_CODE
static BearStatus
DoRRoot(rrootPtr, intBldDataPtr)
root            *rrootPtr;
BldGraphDataInt *intBldDataPtr;
{
    (void) Rdb_ItrRSegmOfRRoot(rrootPtr, AddRSegmOfRRoot,
        (ClientData) intBldDataPtr);

    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */


static BearStatus
DoBnLine(bnPtr, intBldDataPtr)
BottleNeck      *bnPtr;
BldGraphDataInt *intBldDataPtr;
{
    BNLine *bnLinePtr;

    bnLinePtr = bnPtr->unValue.bnLinePtr;
    intBldDataPtr->count = 0;
    (void) Tp_ItrPPinsOfBnLine(Tp_GetTLTileOfBnLine(bnLinePtr),
        Tp_GetBRTileOfBnLine(bnLinePtr), AddPPinOfBnLine,
        (ClientData) intBldDataPtr);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
DoBnTile(bnPtr, intBldDataPtr)
BottleNeck      *bnPtr;
BldGraphDataInt *intBldDataPtr;
{
    CellInst        *cellInstPtr;
    DummyCell       *dcellPtr;
    Plane           *planePtr;
    Tile            *bnTilePtr, *tilePtr;
    SideDirection   side;

    bnTilePtr = bnPtr->unValue.bnTilePtr;

    if (Tp_IsHorzTile(bnTilePtr) == TRUE) {
        intBldDataPtr->horzFlag = TRUE;
        Tp_HorzTileToRect(intBldDataPtr->planeDataPtr, bnTilePtr,
            &intBldDataPtr->absRect);
        planePtr = Tp_GetHorzPlane(intBldDataPtr->planeDataPtr);
    }
    else {
        intBldDataPtr->horzFlag = FALSE;
        Tp_VertTileToRect(intBldDataPtr->planeDataPtr, bnTilePtr,
            &intBldDataPtr->absRect);
        planePtr = Tp_GetVertPlane(intBldDataPtr->planeDataPtr);
    }

    intBldDataPtr->nodeList = lsCreate();
    intBldDataPtr->beginFlag = FALSE;

    /*
     * Add all nodes on the bottom (or left) sides of the bottleneck
     * tile to the nodeList.
     */
    tilePtr = BL(bnTilePtr);
    if ((tilePtr == planePtr->pl_left) || (tilePtr == planePtr->pl_right)) {
        cellInstPtr = Mac_GetInstOfCoreMaster(intBldDataPtr->cellMasterPtr);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, AddPinOfBnTile,
            (ClientData) intBldDataPtr);
    }
    else if (Tp_IsDCellTile(tilePtr) == TRUE) {
        dcellPtr = Tp_GetDCellOfTile(tilePtr);
        side = (intBldDataPtr->horzFlag == TRUE) ? RIGHT_SIDE : TOP_SIDE;
        intBldDataPtr->count = 0;
        (void) Mac_ItrPPinOfDCell(dcellPtr, side, AddPPinOfBnTile,
            (ClientData) intBldDataPtr);
    }
    else if ((intBldDataPtr->rcellList == NIL(RouteCell)) &&
        (Tp_IsCellInstTile(tilePtr) == TRUE)) {
        (void) Mac_ItrPinOfCellInst(Tp_GetCellInstOfTile(tilePtr),
            AddPinOfBnTile, (ClientData) intBldDataPtr);
    }
    else {
        Rect searchArea;

        TITORECT(bnTilePtr, &searchArea);
        searchArea.r_xtop = searchArea.r_xbot;
        searchArea.r_xbot -= 1;
        TiSrArea(NIL(Tile), planePtr, &searchArea, SearchSideOfBnTile,
            (ClientData) intBldDataPtr);
    }

    /*
     * Add all nodes on the top (or right) sides of the bottleneck
     * tile to the nodeList.
     */
    tilePtr = TR(bnTilePtr);
    if ((tilePtr == planePtr->pl_left) || (tilePtr == planePtr->pl_right)) {
        cellInstPtr = Mac_GetInstOfCoreMaster(intBldDataPtr->cellMasterPtr);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, AddPinOfBnTile,
            (ClientData) intBldDataPtr);
    }
    else if (Tp_IsDCellTile(tilePtr) == TRUE) {
        dcellPtr = Tp_GetDCellOfTile(tilePtr);
        side = (intBldDataPtr->horzFlag == TRUE) ? LEFT_SIDE : BOTTOM_SIDE;
        intBldDataPtr->count = 0;
        (void) Mac_ItrPPinOfDCell(dcellPtr, side, AddPPinOfBnTile,
            (ClientData) intBldDataPtr);
    }
    else if ((intBldDataPtr->rcellList == NIL(RouteCell)) &&
        (Tp_IsCellInstTile(tilePtr) == TRUE)) {
        (void) Mac_ItrPinOfCellInst(Tp_GetCellInstOfTile(tilePtr),
            AddPinOfBnTile, (ClientData) intBldDataPtr);
    }
    else {
        Rect searchArea;

        TITORECT(bnTilePtr, &searchArea);
        searchArea.r_xbot = searchArea.r_xtop;
        searchArea.r_xtop += 1;
        TiSrArea(NIL(Tile), planePtr, &searchArea, SearchSideOfBnTile,
            (ClientData) intBldDataPtr);
    }

    if ((intBldDataPtr->pgFlag == TRUE) ||
        (intBldDataPtr->placeFlag == TRUE)) {

        lsSort(intBldDataPtr->nodeList,
            (intBldDataPtr->horzFlag == TRUE) ? CmpHorzNodes : CmpVertNodes);
    }

    intBldDataPtr->count = 0;
    intBldDataPtr->beginFlag = TRUE;
    (void) Tp_ItrPPinsOfSide(bnTilePtr, BOTTOM_LEFT_SIDE, AddPPinOfBnTile,
        (ClientData) intBldDataPtr);

    intBldDataPtr->count = 0;
    intBldDataPtr->beginFlag = FALSE;
    (void) Tp_ItrPPinsOfSide(bnTilePtr, TOP_RIGHT_SIDE, AddPPinOfBnTile,
        (ClientData) intBldDataPtr);

    if (intBldDataPtr->placeFlag == TRUE) {
        lsGen           nodeGen;
        Gt_GraphNode    node;
        PinOrJunct      *pinOrJunctPtr;

        lsForeachItem(intBldDataPtr->nodeList, nodeGen, node) {
            pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
            pinOrJunctPtr->bnTilePtr = bnTilePtr;
        }

        if (lsLength(intBldDataPtr->nodeList) != 0) {
            /*
             * The graph of the given net intersects this bottleneck tile
             * so we include it in the bottleneck tile list.
             */
            lsNewEnd(intBldDataPtr->bnTileList, (lsGeneric) bnTilePtr, LS_NH);
        }
    }

    if ((intBldDataPtr->pgFlag == TRUE) ||
        (intBldDataPtr->placeFlag == TRUE)) {
        ConnNodesThree(intBldDataPtr, GetBnTileId);
    }
    else {
        ConnNodesOne(intBldDataPtr, GetBnTileId);
    }

    lsDestroy(intBldDataPtr->nodeList, (void (*)()) NULL);
    intBldDataPtr->nodeList = (lsList) NULL;

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
DoBottleNeck(bnPtr, intBldDataPtr)
BottleNeck      *bnPtr;
BldGraphDataInt *intBldDataPtr;
{
    if (bnPtr->unType == TP_BN_TILE) {
        if (DoBnTile(bnPtr, intBldDataPtr) == BEAR_STOP_SEARCH) {
            return (BEAR_STOP_SEARCH);
        }
    }
    else {
        (void) DoBnLine(bnPtr, intBldDataPtr);
    }

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
DoSwitchBox(junctPtr, intBldDataPtr)
WallJunction                *junctPtr;
register BldGraphDataInt    *intBldDataPtr;
{
    LinkedPinInst       *lnPinInstPtr;  
    LinkedPPin          *lnPPinPtr;  
    LinkedRPin          *lnRPinPtr;  
    PinsAroundJunction  *pajPtr;
    PinInst             *pinPtr;  
    PinOrJunct          *pinOrJunctPtr;
    PseudoPin           *ppinPtr;
    RoutePin            *rpinPtr;
    Gt_GraphNode        node;
    int                 numOfNodes;

    pajPtr = (PinsAroundJunction *) junctPtr->greModulePtr;

    intBldDataPtr->nodeList = lsCreate();
    intBldDataPtr->centerPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);

    for (lnPPinPtr = pajPtr->horzPPinList;
        lnPPinPtr != NIL(LinkedPPin);
        lnPPinPtr = lnPPinPtr->nextPtr) {

        ppinPtr = lnPPinPtr->ppinPtr;
        if (intBldDataPtr->netPtr == Tp_GetNetOfPPin(ppinPtr)) {
            node = LookUpKey((char *) ppinPtr, FPG_PSEUDO_PIN, intBldDataPtr);
            lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
        }
    }
    for (lnPPinPtr = pajPtr->vertPPinList;
        lnPPinPtr != NIL(LinkedPPin);
        lnPPinPtr = lnPPinPtr->nextPtr) {

        ppinPtr = lnPPinPtr->ppinPtr;
        if (intBldDataPtr->netPtr == Tp_GetNetOfPPin(ppinPtr)) {
            node = LookUpKey((char *) ppinPtr, FPG_PSEUDO_PIN, intBldDataPtr);
            lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
        }
    }
    for (lnRPinPtr = pajPtr->rPinList;
        lnRPinPtr != NIL(LinkedRPin);
        lnRPinPtr = lnRPinPtr->nextPtr) {

        rpinPtr = lnRPinPtr->rPinPtr;
        if (intBldDataPtr->netPtr == Mac_GetNetPtrOfRPin(rpinPtr)) {
            node = LookUpKey((char *) rpinPtr, FPG_ROUTE_PIN, intBldDataPtr);
            lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
        }
    }
    for (lnPinInstPtr = pajPtr->pinInstList;
        lnPinInstPtr != NIL(LinkedPinInst);
        lnPinInstPtr = lnPinInstPtr->nextPtr) {

        pinPtr = lnPinInstPtr->pinInstPtr;
        if (intBldDataPtr->netPtr == Mac_GetNetOfPin(pinPtr)) {
            node = LookUpKey((char *) pinPtr, FPG_PIN, intBldDataPtr);
            lsNewEnd(intBldDataPtr->nodeList, (lsGeneric) node, LS_NH);
        }
    }

#ifdef DEBUG
    if (dupFlag == TRUE) {
        Gt_GraphNode    node;
        lsGen           gen;

        if (lsLength(intBldDataPtr->nodeList) != 0) {
            fprintf (stderr, "---\n");
            gen = lsStart(intBldDataPtr->nodeList);
            while (lsNext(gen, (lsGeneric *) &node, LS_NH) == LS_OK) {
                (void) PrintNode(node, stderr);
            }
            lsFinish(gen);
            fprintf (stderr, "---\n");
        }
    }
#endif DEBUG

    numOfNodes = lsLength(intBldDataPtr->nodeList);
    if ((intBldDataPtr->placeFlag == TRUE) && (numOfNodes != 0)) {
        /*
         * The graph of the given net intersects this switch box area
         * so we include it in the junction list.
         */
        lsNewEnd(intBldDataPtr->junctList, (lsGeneric) junctPtr, LS_NH);
    }

    if ((intBldDataPtr->pgFlag == TRUE) ||
        ((intBldDataPtr->placeFlag == TRUE) && (numOfNodes > 2))) {
        ConnNodesTwo(intBldDataPtr, GetSwitchBoxId);
    }
    else {
        ConnNodesOne(intBldDataPtr, GetSwitchBoxId);
    }

    lsDestroy(intBldDataPtr->nodeList, (void (*)()) NULL);
    intBldDataPtr->nodeList = (lsList) NULL;

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
StoreSwitchBoxData(junctPtr, intBldDataPtr)
WallJunction    *junctPtr;
BldGraphDataInt *intBldDataPtr;
{
    PinsAroundJunction *pajPtr;

    pajPtr = Tm_MakePinsAroundJunction();
    Tm_PinsAroundJunction(intBldDataPtr->cellMasterPtr,
        intBldDataPtr->planeDataPtr, junctPtr, pajPtr);
    junctPtr->greModulePtr = (BearGeneric) pajPtr;

    return (BEAR_CONTINUE_SEARCH);
}


/* ARGSUSED */
static BearStatus
RemoveSwitchBoxData(junctPtr, clientData)
WallJunction    *junctPtr;
ClientData      clientData;
{
    PinsAroundJunction *pajPtr;

    pajPtr = (PinsAroundJunction *) junctPtr->greModulePtr;
    Tm_FreePinsAroundJunction(pajPtr);
    FREE((char *) pajPtr);

    junctPtr->greModulePtr = (BearGeneric) NULL;

    return (BEAR_CONTINUE_SEARCH);
}


#ifdef DEBUG
static void 
PrintGraph(fileNamePtr, graph)
char        *fileNamePtr;
Gt_Graph    graph;
{
    FILE *filePtr;

    if ((filePtr = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg, "could not open %s\n", fileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return;
    }
                                                        
    fprintf (filePtr, "\n------------- edge list -------------\n\n");
    fprintf (filePtr, "%8s%12s%8s%8s\n\n", "edge", "width", "node1", "node2");
    (void) Gt_ItrEdgesOfGraph(graph, PrintEdge, (ClientData) filePtr);

    fprintf (filePtr, "\n------------- node list -------------\n\n");
    (void) Gt_ItrNodesOfGraph(graph, PrintNode, (ClientData) filePtr);
    fprintf (filePtr, "\n");

    (void) fclose (filePtr);
}
#endif DEBUG



#ifdef DEBUG
static BearStatus 
PrintNode(node, filePtr)
Gt_GraphNode    node;
FILE            *filePtr;
{
    PinOrJunct  *pinOrJunctPtr;
    PseudoPin   *ppinPtr;
    Tile        *tilePtr;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    
    fprintf (filePtr, "node %d : ", Gt_GetIdOfNode(node));
    switch (pinOrJunctPtr->unType) {
    case (int) FPG_PIN:
        fprintf (filePtr, "%-15s (ext id = %d)\n", "pin",
            Mac_GetExtNetIdOfPin(pinOrJunctPtr->unValue.pinPtr));
        break;
    case (int) FPG_PSEUDO_PIN:
        ppinPtr = pinOrJunctPtr->unValue.ppinPtr;
        tilePtr = Tp_GetTileOfPPin(ppinPtr);

        fprintf (filePtr, "%-15s (%s ", "pseudo pin",
            (Tp_IsVertTile(tilePtr) == TRUE) ? "vert" : "horz");

        if (Tp_IsBottleneckTile(tilePtr) == TRUE) {
            fprintf (filePtr, "bn tile 0x%x) (side %d)\n", tilePtr,
                Tp_GetSideOfPPin(ppinPtr));
        }
        else if (Tp_IsSBNTile(tilePtr) == TRUE) {
            fprintf (filePtr, "sbn tile 0x%x)\n", tilePtr);
        }
        else if (Tp_IsDCellTile(tilePtr) == TRUE) {
            fprintf (filePtr, "dc tile 0x%x)\n", tilePtr);
        }
        else {
            fprintf (filePtr, "UNKNOWN tile 0x%x)", tilePtr);
        }
        fprintf (filePtr, "    (int id = %d) (ext id = %d)\n",
            Tp_GetIntNetIdOfPPin(ppinPtr),
            Tp_GetExtNetIdOfPPin(ppinPtr));
        break;
    case (int) FPG_ROUTE_PIN:
        fprintf (filePtr, "%-15s (ext id = %d)\n", "route pin",
            Mac_GetExtNetIdOfRPin(pinOrJunctPtr->unValue.rPinPtr));
        break;
    case (int) FPG_WALL_JUNCT:
        fprintf (filePtr, "%-15s\n", "wall junction");
        break;
    case (int) FPG_PSEUDO_JCT:
        fprintf (filePtr, "%-15s\n", "pseudo junction");
        break;
    case FPG_BBL_TERM:
        fprintf (filePtr, "%-15s\n", "bbl terminal");
        break;
    case FPG_BBL_JUNCT:
        fprintf (filePtr, "%-15s\n", "bbl junction");
        break;
    default:
        Bear_PrintErr("PrintNode() : unexpected pin or junction type");
        break;
    }

    return (BEAR_CONTINUE_SEARCH);
}
#endif DEBUG



#ifdef DEBUG
static BearStatus 
PrintEdge(edge, filePtr)
Gt_GraphEdge    edge;
FILE            *filePtr;
{
    Gt_GraphNode node1, node2;
    Mac_EdgeData *edgeDataPtr;

    node1 = Gt_GetNodeOneOfEdge(edge);
    node2 = Gt_GetNodeTwoOfEdge(edge);

    if ((edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge)) ==
        NIL(Mac_EdgeData)) {

        fprintf (filePtr, "%8d%12s%8d%8d\n", Gt_GetIdOfEdge(edge),
            "          --", Gt_GetIdOfNode(node1), Gt_GetIdOfNode(node2));
    }
    else {
        fprintf (filePtr, "%8d%12.3f%8d%8d\n", Gt_GetIdOfEdge(edge),
            edgeDataPtr->width, Gt_GetIdOfNode(node1), Gt_GetIdOfNode(node2));
    }

    return (BEAR_CONTINUE_SEARCH);
}
#endif DEBUG



BearStatus
Gre_BuildAutoGraphOfNet(cellMasterPtr, netPtr, warnFlag, placeFlag, pgFlag)
CellMaster  *cellMasterPtr;                     /* IN       */
Net         *netPtr;                            /* IN       */
bool        warnFlag;                           /* IN       */
bool        placeFlag;                          /* IN       */
bool        pgFlag;                             /* IN       */
{
    BearStatus          bearStatus;
    Gre_BldGraphDataPtr bldDataPtr;
    Gt_Graph            autoGraph;

    Mac_DeleteAutoGraphOfNet(netPtr);

    bldDataPtr = Gre_BldGraphBegin(cellMasterPtr, warnFlag, placeFlag, pgFlag);

    if ((bearStatus = Gre_BldGraph(bldDataPtr, netPtr, &autoGraph)) ==
        BEAR_OK) {
        Mac_PutAutoGraphOfNet(autoGraph, netPtr);
    }

    Gre_BldGraphEnd(bldDataPtr);

    return (bearStatus);
}


Gre_BldGraphDataPtr
Gre_BldGraphBegin(cellMasterPtr, warnFlag, placeFlag, pgFlag)
CellMaster          *cellMasterPtr;             /* IN       */
bool                warnFlag;                   /* IN       */
bool                placeFlag;                  /* IN       */
bool                pgFlag;                     /* IN       */
/*
 * Begin to build an auto-graph.
 */
{
    BldGraphDataInt *intBldDataPtr;

    MALLOC(BldGraphDataInt *, intBldDataPtr, sizeof (BldGraphDataInt));

    intBldDataPtr->warnFlag = warnFlag;
    intBldDataPtr->placeFlag = placeFlag;
    intBldDataPtr->pgFlag = pgFlag;
    intBldDataPtr->checkRSegFlag = warnFlag;

    intBldDataPtr->cellMasterPtr = cellMasterPtr;
    intBldDataPtr->planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    intBldDataPtr->fpgPtr = Tp_GetFpg(intBldDataPtr->planeDataPtr);
    intBldDataPtr->dcellList = Tp_GetDCellList(intBldDataPtr->planeDataPtr);
    intBldDataPtr->rcellList = Mac_GetRCellListOfCell(cellMasterPtr);

    intBldDataPtr->bnTileList = (lsList) NULL;
    intBldDataPtr->junctList = (lsList) NULL;

    if (intBldDataPtr->fpgPtr != NIL(FloorPlanGraph)) {
        (void) Fpg_ItrWallJunctSeq(intBldDataPtr->fpgPtr,
            StoreSwitchBoxData, (ClientData) intBldDataPtr);
    }

    return ((Gre_BldGraphDataPtr) intBldDataPtr);
}


BearStatus
Gre_BldGraph(bldDataPtr, netPtr, graphPtr)
Gre_BldGraphDataPtr bldDataPtr;                 /* IN       */
Net                 *netPtr;                    /* IN       */
Gt_Graph            *graphPtr;                  /* OUT      */
/*
 * Build Graph.
 */
{
    BearStatus      bearStatus = BEAR_OK;
    BldGraphDataInt *intBldDataPtr;
    Mac_GraphData   *graphDataPtr;

    intBldDataPtr = (BldGraphDataInt *) bldDataPtr;
    intBldDataPtr->netPtr = netPtr;

    globalPlaneDataPtr = intBldDataPtr->planeDataPtr;

    graphDataPtr = Mac_MakeGraphData();
    if (intBldDataPtr->fpgPtr != NIL(FloorPlanGraph)) {
        graphDataPtr->fpgId = Fpg_GetIdOfFpg(intBldDataPtr->fpgPtr);
    }

    if (intBldDataPtr->pgFlag == TRUE) {
        graphDataPtr->pgFlag = TRUE;
        graphDataPtr->edgeDataFlag = TRUE;

        intBldDataPtr->graph = Gt_CreateGraph(Mac_GenerateId(),
            (BearGeneric) graphDataPtr);
    }
    else if (intBldDataPtr->placeFlag == TRUE) {
        graphDataPtr->pgFlag = FALSE;
        graphDataPtr->edgeDataFlag = FALSE;

        intBldDataPtr->graph = Gt_CreateGraph(Mac_GenerateId(),
            (BearGeneric) graphDataPtr);
        intBldDataPtr->bnTileList = lsCreate();
        intBldDataPtr->junctList = lsCreate();
    }
    else {
        graphDataPtr->pgFlag = FALSE;
        graphDataPtr->edgeDataFlag = FALSE;

        intBldDataPtr->graph = Gt_CreateGraph((BearId) 0,
            (BearGeneric) graphDataPtr);
    }

    intBldDataPtr->hashTable = st_init_table(st_ptrcmp, st_ptrhash);
    intBldDataPtr->nodeNum = 1;
    intBldDataPtr->edgeNum = 1;

    (void) Mac_ItrAllPinOfNet(intBldDataPtr->netPtr, AddPinOfNet, 
        (ClientData) intBldDataPtr);

    if (intBldDataPtr->fpgPtr == NIL(FloorPlanGraph)) {
        goto end1;
    }

    (void) Fpg_ItrWallJunctSeq(intBldDataPtr->fpgPtr, DoSwitchBox,
        (ClientData) intBldDataPtr);
    if (Fpg_ItrBottleNeckOfFpg(intBldDataPtr->fpgPtr, DoBottleNeck,
        (ClientData) intBldDataPtr) == BEAR_STOP_SEARCH) {
        bearStatus = BEAR_ERROR;
        goto end1;
    }
    (void) Mac_ItrDCell(intBldDataPtr->dcellList, DoDCell,
        (ClientData) intBldDataPtr);
#ifdef USE_DR_CODE
    (void) Mac_ItrRRootOfNet(intBldDataPtr->netPtr, DoRRoot,
        (ClientData) intBldDataPtr);
#endif /* USE_DR_CODE */

end1:
    st_free_table(intBldDataPtr->hashTable);

#ifdef DEBUG
    if (genFlag == TRUE) {
        PrintGraph("graph.debug", intBldDataPtr->graph);
    }
#endif DEBUG

    if (intBldDataPtr->placeFlag == TRUE) {
        graphDataPtr->bnTileList = intBldDataPtr->bnTileList;
        graphDataPtr->junctList = intBldDataPtr->junctList;

        intBldDataPtr->bnTileList = (lsList) NULL;
        intBldDataPtr->junctList = (lsList) NULL;
    }

    if (bearStatus == BEAR_ERROR) {
        Mac_DeleteGraph(intBldDataPtr->graph);
        *graphPtr = (Gt_Graph) NULL;
    }
    else {
        *graphPtr = intBldDataPtr->graph;
    }
    intBldDataPtr->graph = (Gt_Graph) NULL;

    return (bearStatus);
}


void
Gre_BldGraphEnd(bldDataPtr)
Gre_BldGraphDataPtr bldDataPtr;                       /* IN       */
/*
 * Free all resources used while building the auto-graph.
 */
{
    BldGraphDataInt *intBldDataPtr;

    intBldDataPtr = (BldGraphDataInt *) bldDataPtr;

    if (intBldDataPtr->fpgPtr != NIL(FloorPlanGraph)) {
        (void) Fpg_ItrWallJunctSeq(intBldDataPtr->fpgPtr,
            RemoveSwitchBoxData, (ClientData) NULL);
    }
    FREE((char *) intBldDataPtr);
}


/*
 * ----------------------------------------------------------------------------
 *
 * CheckConnForOneNet() --
 *
 * This procedure is designed to be used by an "interator".  It builds a
 * graph of the given net, checks the graphs connectivity, and then deletes
 * the graph.  If a net is not connected then 'connDataPtr->connFlag' will be
 * set to FALSE;
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion, even if theres an
 *   error so that the interator should continue on to the next net.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
CheckConnForOneNet(netPtr, connDataPtr)
Net         *netPtr;
ConnData    *connDataPtr;
{
    Gt_Graph    autoGraph;
    char        *netNamePtr;
    bool        loopFlag, connFlag;

    netNamePtr = Mac_GetNameOfNet(netPtr);

    if (Mac_GetAutoGraphOfNet(netPtr) != (Gt_Graph) NULL) {
        Mac_DeleteAutoGraphOfNet(netPtr);
    }

    if (Gre_BldGraph(connDataPtr->bldDataPtr, netPtr, &autoGraph) ==
        BEAR_ERROR) {
        sprintf (bear_ErrorMsg, "Cannot build graph for net named \"%s\".\n",
            netNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        connDataPtr->connFlag = FALSE;
        return (BEAR_CONTINUE_SEARCH);
    }
    
    Mac_PutAutoGraphOfNet(autoGraph, netPtr);
    if (Mac_CheckAutoGraph(netPtr, &loopFlag, &connFlag) == BEAR_ERROR) {
        sprintf (bear_ErrorMsg, "Cannot check graph for net named \"%s\".\n",
            netNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        connDataPtr->connFlag = FALSE;
        Mac_DeleteAutoGraphOfNet(netPtr);
        return (BEAR_CONTINUE_SEARCH);
    }
    
    if (connDataPtr->printMsgFlag == TRUE) {
        sprintf (bear_Msg, "Net named \"%s\" %s\n", netNamePtr,
            (connFlag == TRUE) ? 
                ((loopFlag == TRUE) ? "is connected but has a loop" :
                    "is connected and has no loops") :
                ((loopFlag == TRUE) ? "is not connected but has a loop" :
                    "is not connected and has no loops"));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
    }
    else {
        if (connFlag == FALSE) {
            sprintf (bear_ErrorMsg, "Net named \"%s\" is not connected\n",
                netNamePtr);
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        }
        else if (loopFlag == TRUE) {
            sprintf (bear_Msg,
                "Warning: Net named \"%s\" has a loop\n", netNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
    }

    if (connFlag == FALSE) {
        connDataPtr->connFlag = FALSE;
    }
    Mac_DeleteAutoGraphOfNet(netPtr);

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gre_CheckConnForOneNet() --
 *
 * This procedure checks the connectivity of the given net.  If the net is
 * connected it draws the graph, and prints out a message.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Gre_CheckConnForOneNet(cellMasterPtr, netPtr, warnFlag)
CellMaster  *cellMasterPtr;
Net         *netPtr;
bool        warnFlag;
{
    ConnData connData;

    connData.bldDataPtr = Gre_BldGraphBegin(cellMasterPtr,
        warnFlag, FALSE, FALSE);

    connData.connFlag = TRUE;
    connData.printMsgFlag = TRUE;
    (void) CheckConnForOneNet(netPtr, &connData);

    Gre_BldGraphEnd(connData.bldDataPtr);

    if (connData.connFlag == FALSE) {
        return (BEAR_ERROR);
    }
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gre_CheckConnForAllNets() --
 *
 * This procedure checks the connectivity of all the nets.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Gre_CheckConnForAllNets(cellMasterPtr, warnFlag)
CellMaster  *cellMasterPtr;
bool        warnFlag;
{
    ConnData connData;

    connData.bldDataPtr = Gre_BldGraphBegin(cellMasterPtr,
        warnFlag, FALSE, FALSE);

    connData.connFlag = TRUE;
    connData.printMsgFlag = FALSE;
    (void) Mac_ItrNetOfCell(cellMasterPtr, CheckConnForOneNet, 
        (ClientData) &connData);

    Gre_BldGraphEnd(connData.bldDataPtr);

    if (connData.connFlag == FALSE) {
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}



