
/* grePgWiring.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/grEditor/grePgWiring.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#ifdef USE_WIRE_CODE

#include "bearCopyRight.h"
#include <stdio.h>
#include <math.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "routeDB.h"
#include "ezPlot.h"
#include "grEditor.h"
#include "grEditorInt.h"

#include "../polarBear/polarBear.h"

typedef struct _AddNodeData {
    Pgr_NodeProp    *srcNodePtr;
    Pgr_Wire        *wirePtr;
} AddNodeData;

typedef struct _AddEdgeData {
    PlaneData       *planeDataPtr;
    Pgr_Wire        *wirePtr;
} AddEdgeData;

typedef struct _PortData {
    PortKind        portKind;
    double          maxVoltDrop, maxCurrentReq;
} PortData;

typedef struct _CalcPgPinWidthData {
    Gt_GraphNode    node;           /* Graph node that contains a pin. */
    Point           pinPoint;       /* Coordinates of the pin.  */
    PinDirection    pinDir;         /* Direction of the pin.  */
    float           metal1Width;
    float           metal2ESWidth;
    float           metal2WNWidth;
    bool            hasESNode;      /* TRUE if there is a node to the east
                                     * (south) of the pin.
                                     */
} CalcPgPinWidthData;


typedef struct _WireData {
    Gre_BldGraphDataPtr bldDataPtr;
    CellMaster          *cellMasterPtr;
    PlaneData           *planeDataPtr;
} WireData;


/* ---------- Local Procedures ---------- */
static BearStatus   AddEdgeToList(), AddNodeToList();
static BearStatus   LookForSrcPort();
static BearStatus   StoreNodeProp();
static BearStatus   StorePortData();
static BearStatus   StorePgWireInTiPlane();
static BearStatus   CalcPgPinWidthOne(), CalcPgPinWidthTwo();
static BearStatus   DoPgWiringForOneNet();


static BearStatus
AddNodeToList(node, addNodeDataPtr)
Gt_GraphNode    node;
AddNodeData     *addNodeDataPtr;
/*
 * The node list may only contain one source node.
 */
{
    Pgr_NodeProp    *pgrNodePtr;
    PinOrJunct      *pinOrJunctPtr;
    PortInst        *portInstPtr;
    PortKind        portKind;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    if (pinOrJunctPtr->unType == FPG_PIN) {
        portInstPtr = Mac_GetPortOfPin(pinOrJunctPtr->unValue.pinPtr);
        portKind = Mac_GetKindOfPort(portInstPtr);
        if (portKind == (PortKind) PG_SRC_PORT) {
            if (addNodeDataPtr->srcNodePtr == NIL(Pgr_NodeProp)) {
                addNodeDataPtr->srcNodePtr = Pgr_MakeNodeProp();
                addNodeDataPtr->srcNodePtr->nextPtr =
                    addNodeDataPtr->wirePtr->nodeList;
                addNodeDataPtr->wirePtr->nodeList = addNodeDataPtr->srcNodePtr;
                if (StoreNodeProp(addNodeDataPtr->srcNodePtr, pinOrJunctPtr, 
                    (BearGeneric) NULL,
                    addNodeDataPtr->wirePtr) == BEAR_ERROR) {
                    return (BEAR_STOP_SEARCH);
                }
            }
            pinOrJunctPtr->genericPtr =
                (BearGeneric) addNodeDataPtr->srcNodePtr;
            return (BEAR_CONTINUE_SEARCH);
        }
    }

    pgrNodePtr = Pgr_MakeNodeProp();
    pgrNodePtr->nextPtr = addNodeDataPtr->wirePtr->nodeList;
    addNodeDataPtr->wirePtr->nodeList = pgrNodePtr;

    if (StoreNodeProp(pgrNodePtr, pinOrJunctPtr, 
        (BearGeneric) pinOrJunctPtr, addNodeDataPtr->wirePtr) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    pinOrJunctPtr->genericPtr = (BearGeneric) pgrNodePtr;

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
AddEdgeToList(edge, addEdgeDataPtr)
Gt_GraphEdge    edge;
AddEdgeData     *addEdgeDataPtr;
{
    Gt_GraphNode    node1, node2;
    Mac_EdgeData    *edgeDataPtr;
    PinOrJunct      *pinOrJunct1Ptr, *pinOrJunct2Ptr;
    Pgr_EdgeProp    *pgrEdgePtr;
    Pgr_NodeProp    *pgrTmpNodePtr, *pgrNode1Ptr, *pgrNode2Ptr;
    Point           absPos1, absPos2;

    node1 = Gt_GetNodeOneOfEdge(edge);
    node2 = Gt_GetNodeTwoOfEdge(edge);

    edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);

    pinOrJunct1Ptr = (PinOrJunct *) Gt_GetNodeDataOfNode(node1);
    pinOrJunct2Ptr = (PinOrJunct *) Gt_GetNodeDataOfNode(node2);
    
    pgrNode1Ptr = (Pgr_NodeProp *) pinOrJunct1Ptr->genericPtr;
    pgrNode2Ptr = (Pgr_NodeProp *) pinOrJunct2Ptr->genericPtr;

    if ((pgrNode1Ptr->nodeNum == 0) && (pgrNode2Ptr->nodeNum == 0)) {
        return (BEAR_CONTINUE_SEARCH);
    }
    else if (pgrNode1Ptr->nodeNum == pgrNode2Ptr->nodeNum) {
        PbPrint(EZ_ITALIC_FONT, "Each edge must contain two distinct nodes.");
        return (BEAR_STOP_SEARCH);
    }

    /* Add edge between the nodes to the edge list. */
    pgrEdgePtr = Pgr_MakeEdgeProp();
    pgrEdgePtr->nextPtr = addEdgeDataPtr->wirePtr->edgeList;
    addEdgeDataPtr->wirePtr->edgeList = pgrEdgePtr;

    pgrEdgePtr->edgeNum = ++(addEdgeDataPtr->wirePtr->edgeCount);
    pgrEdgePtr->genericPtr = (BearGeneric) Gt_GetEdgeDataOfEdge(edge);

    /* 
     * Flip pointers so that Robi's program does not blow.
     */
    if (pgrNode1Ptr->nodeNum == 0) {
        pgrTmpNodePtr = pgrNode1Ptr;
        pgrNode1Ptr = pgrNode2Ptr;
        pgrNode2Ptr = pgrTmpNodePtr;
    }
    pgrEdgePtr->node1Ptr = pgrNode1Ptr;
    pgrEdgePtr->node2Ptr = pgrNode2Ptr;

    /*
     * Calculate wire length between the nodes.
     *
     * Note that length is defined as the so called manhatten
     * distance between the two nodes.
     */

    if (Fpg_GetAbsPosOfPinOrJunct(addEdgeDataPtr->planeDataPtr,
        pinOrJunct1Ptr, &absPos1) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    if (Fpg_GetAbsPosOfPinOrJunct(addEdgeDataPtr->planeDataPtr,
        pinOrJunct2Ptr, &absPos2) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }
    pgrEdgePtr->length = (ABS(absPos2.p_y - absPos1.p_y) +
        ABS(absPos2.p_x - absPos1.p_x));
    pgrEdgePtr->width = edgeDataPtr->width;

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
CalcPgPinWidthOne(node)
Gt_GraphNode node;                  /* Node must be a pin. */
/*
 * Calculate the width of a power/ground pin.
 */
{
    CalcPgPinWidthData  calcData;
    PinInst             *pinPtr;
    PinOrJunct          *pinOrJunctPtr;
    float               *widthVect;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    if (pinOrJunctPtr->unType != FPG_PIN) {
        return (BEAR_ERROR);
    }

    pinPtr = pinOrJunctPtr->unValue.pinPtr;

    calcData.node = node;
    Mac_GetAbsPosOfPin(pinPtr, &calcData.pinPoint);
    calcData.pinDir = Mac_GetDirectionOfPin(pinPtr);
    calcData.metal1Width = 0.0;
    calcData.metal2ESWidth = 0.0;
    calcData.metal2WNWidth = 0.0;
    calcData.hasESNode = FALSE;

    if (Gt_ItrEdgesOfNode(node, CalcPgPinWidthTwo,
        (ClientData) &calcData) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

    widthVect = Mac_GetWidthVectOfPin(pinPtr);
    widthVect[METAL1_LAYER] = calcData.metal1Width;
    widthVect[METAL2_LAYER] = calcData.hasESNode ?
        calcData.metal2ESWidth : calcData.metal2WNWidth;

    return (BEAR_OK);
}



static BearStatus
CalcPgPinWidthTwo(edge, calcDataPtr)
Gt_GraphEdge        edge;
CalcPgPinWidthData  *calcDataPtr;
/*
 * CalcPgPinWidthTwo() is called once for each edge attached to
 * 'calcDataPtr->node' in an effort to visit each node attached to
 * 'calcDataPtr->node'.
 */
{
    Mac_EdgeData    *edgeDataPtr;
    PinOrJunct      *pinOrJunctPtr;
    Point           *junctPointPtr, nodePoint;
    Gt_GraphNode    node1, node2;

    node1 = Gt_GetNodeOneOfEdge(edge);
    node2 = Gt_GetNodeTwoOfEdge(edge);

    if (node1 != calcDataPtr->node) {
        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node1);
    }
    else {
        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node2);
    }
    
    edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);

    /*
     * Calculate the pin width for metal layer one.
     */
    calcDataPtr->metal1Width += edgeDataPtr->width;

    /*
     * Calculate the pin width for metal layer two.
     */
    if (pinOrJunctPtr->unType == FPG_PIN) {
        Mac_GetAbsPosOfPin(pinOrJunctPtr->unValue.pinPtr, &nodePoint);
    }
    else {
        junctPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER,
            pinOrJunctPtr->unValue.junctPtr);
        nodePoint.p_x = junctPointPtr->p_x;
        nodePoint.p_y = junctPointPtr->p_y;
    }
    if (calcDataPtr->pinDir == (PinDirection) NORTH_PIN ||
        calcDataPtr->pinDir == (PinDirection) SOUTH_PIN) {
        if (nodePoint.p_x > calcDataPtr->pinPoint.p_x) {

            /* There is a node to the east of the pin. */
            calcDataPtr->hasESNode = TRUE;

            calcDataPtr->metal2ESWidth += edgeDataPtr->width;
        }
        else {
            calcDataPtr->metal2WNWidth += edgeDataPtr->width;
        }
    }
    else if (calcDataPtr->pinDir == (PinDirection) EAST_PIN ||
        calcDataPtr->pinDir == (PinDirection) WEST_PIN) {
        if (nodePoint.p_y < calcDataPtr->pinPoint.p_y) {

            /* There is a node to the south of the pin. */
            calcDataPtr->hasESNode = TRUE;

            calcDataPtr->metal2ESWidth += edgeDataPtr->width;
        }
        else {
            calcDataPtr->metal2WNWidth += edgeDataPtr->width;
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
LookForSrcPort(portInstPtr, clientData)
PortInst    *portInstPtr;
ClientData  clientData;
{
    if (Mac_GetKindOfPort(portInstPtr) == (PortKind) PG_SRC_PORT) {
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
StoreNodeProp(pgrNodePtr, pinOrJunctPtr, genericPtr, wirePtr)
Pgr_NodeProp    *pgrNodePtr;
PinOrJunct      *pinOrJunctPtr;
BearGeneric     genericPtr;
Pgr_Wire        *wirePtr;
/*
 * Store the information about the pin or junction in the Pgr_NodeProp
 * structure that's used by Robbi Dutta's wiring code.
 */
{
    PortInst    *portInstPtr;
    PortKind    portKind;

    if (pinOrJunctPtr->unType == FPG_PIN) {
        pgrNodePtr->nodeType = (Pgr_NodeType) PIN;

        portInstPtr = Mac_GetPortOfPin(pinOrJunctPtr->unValue.pinPtr);
        portKind = Mac_GetKindOfPort(portInstPtr);
        if (portKind != (PortKind) PG_PORT &&
            portKind != (PortKind) PG_SRC_PORT) {

            sprintf (bear_ErrorMsg, "net may only have pg ports\n");
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
            return (BEAR_ERROR);
        }

        if (portKind == (PortKind) PG_PORT) {
            pgrNodePtr->nodeNum = wirePtr->nodeCount++;
        }
        else if (portKind == (PortKind) PG_SRC_PORT) {
            pgrNodePtr->nodeNum = 0;
        }

        pgrNodePtr->maxVoltDrop = Mac_GetMaxVoltDropOfPort(portInstPtr);
        pgrNodePtr->maxCurrentReq = Mac_GetMaxCurrentReqOfPort(portInstPtr);
    } 
    else {
        pgrNodePtr->nodeType = (Pgr_NodeType) JUNCT;
        pgrNodePtr->nodeNum = wirePtr->nodeCount++;
    }

    pgrNodePtr->genericPtr = genericPtr;

    return (BEAR_OK);
}



/* ARGSUSED */
static BearStatus
StorePgWireInTiPlane(edge, clientData)
Gt_GraphEdge    edge;
ClientData      clientData;
{
    Mac_EdgeData    *edgeDataPtr;
    Gt_GraphNode    node1, node2;
    PinOrJunct      *pinOrJunctPtr;

    node1 = Gt_GetNodeOneOfEdge(edge);
    node2 = Gt_GetNodeTwoOfEdge(edge);
    
    edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node1);
    if (pinOrJunctPtr->unType == FPG_PIN) {
        (void) CalcPgPinWidthOne(node1);
    }
    else if (pinOrJunctPtr->unType == FPG_PSEUDO_PIN) {
        Tp_PutWidthOfPPin(edgeDataPtr->width * 1.5,
            pinOrJunctPtr->unValue.ppinPtr);
    }

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node2);
    if (pinOrJunctPtr->unType == FPG_PIN) {
        (void) CalcPgPinWidthOne(node2);
    }
    else if (pinOrJunctPtr->unType == FPG_PSEUDO_PIN) {
        Tp_PutWidthOfPPin(edgeDataPtr->width * 1.5,
            pinOrJunctPtr->unValue.ppinPtr);
    }

    /*
     * (gyorgy : 4 sept 90) Currently pseudo pins might be stored in the
     * edges of a "user graph" but never an "auto graph".  In the future
     * if the "auto graph" is pruned this may change.
     */
    if (edgeDataPtr->ppin1Ptr != NIL(PseudoPin)) {
        Tp_PutWidthOfPPin(edgeDataPtr->width, edgeDataPtr->ppin1Ptr);
    }
    if (edgeDataPtr->ppin2Ptr != NIL(PseudoPin)) {
        Tp_PutWidthOfPPin(edgeDataPtr->width, edgeDataPtr->ppin2Ptr);
    }

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
StorePortData(portInstPtr, portDataPtr)
PortInst    *portInstPtr;
PortData    *portDataPtr;
/*
 * Store the information about the port in the PortInst structure.
 */
{
    PortKind portKind;

    portKind = Mac_GetKindOfPort(portInstPtr);
    if ((portKind != PG_PORT) && (portKind != PG_SRC_PORT)) {
        Mac_PutKindOfPort(portDataPtr->portKind, portInstPtr);
    }

    Mac_PutMaxVoltDropOfPort(portDataPtr->maxVoltDrop, portInstPtr);
    Mac_PutMaxCurrentReqOfPort(portDataPtr->maxCurrentReq, portInstPtr);

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * CalcWireWidths() --
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

BearStatus
CalcWireWidths(cellMasterPtr, netPtr)
CellMaster  *cellMasterPtr;
Net         *netPtr;
{
    Pgr_WireTech            *wireTechPtr;
    register Pgr_EdgeProp   *pgrEdgePtr;
    double                  m1Width, m2Width, minWidth;

    if (netPtr->wirePtr == NIL(Pgr_Wire)) {
        return (BEAR_ERROR);
    }

    if ((wireTechPtr = cellMasterPtr->contents.wireTechPtr) ==  
        NIL(Pgr_WireTech)) {
        Bear_PrintErr("CalcWireWidths() failed : no pg technology given\n");
        return (BEAR_ERROR);
    }
    if (Pgr_CalcWireWidths(wireTechPtr, netPtr->wirePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    m1Width = (double) Mac_GetMinWidthOfLay(METAL1_LAYER);
    m2Width = (double) Mac_GetMinWidthOfLay(METAL2_LAYER);
    minWidth = MAX(m1Width, m2Width);

    for (pgrEdgePtr = netPtr->wirePtr->edgeList;
        pgrEdgePtr != NIL(Pgr_EdgeProp);
        pgrEdgePtr = pgrEdgePtr->nextPtr) {

        if (pgrEdgePtr->width < minWidth) {
            pgrEdgePtr->width = minWidth;
        }
    }

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * StoreNetsPgWireData() --
 *
 * Put the power/ground wire data for this net into the global
 * routing tree and the tile plane.  Do the wire calculations 
 * before you call this procedure.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

BearStatus
StoreNetsPgWireData(netPtr)
Net *netPtr;
{
    Pgr_EdgeProp    *pgrEdgePtr;

    if (netPtr->wirePtr == NIL(Pgr_Wire)) {
        Bear_PrintErr("StoreNetsPgWireData() error : net has no wire data");
        return (BEAR_ERROR);
    }

    /*
     * Store the wire data in the global routing tree.
     */
    for (pgrEdgePtr = netPtr->wirePtr->edgeList;
        pgrEdgePtr != NIL(Pgr_EdgeProp);
        pgrEdgePtr = pgrEdgePtr->nextPtr) {

        ((Mac_EdgeData *) pgrEdgePtr->genericPtr)->width =
            (float) pgrEdgePtr->width;
    }

    /*
     * Store the power ground wire data in the tile plane.
     */
    if (Mac_GetStatusOfNetInTiPlane(netPtr) == TRUE) {
        if (Gt_ItrEdgesOfGraph(Mac_GetAutoGraphOfNet(netPtr),
            StorePgWireInTiPlane, (ClientData) NULL) == BEAR_STOP_SEARCH) {
            return (BEAR_STOP_SEARCH);
        }
    }

    return (BEAR_OK);
}



static BearStatus
DoPgWiringForOneNet(netPtr, wireDataPtr)
Net         *netPtr;
WireData    *wireDataPtr;
{
    AddNodeData addNodeData;
    AddEdgeData addEdgeData;
    Gt_Graph    autoGraph;
    NetKind     netKind;
    char        *netNamePtr;
    bool        loopFlag, connFlag;

    netKind = Mac_GetKindOfNet(netPtr);
    if ((netKind != POWER_NET) && (netKind != GROUND_NET)) {
        return (BEAR_CONTINUE_SEARCH);
    }

    netNamePtr = Mac_GetNameOfNet(netPtr);

    if (Mac_ItrAllPortOfNet(netPtr, LookForSrcPort, (ClientData) NULL) != 
        BEAR_STOP_SEARCH) {
        sprintf (bear_ErrorMsg, "Net named \"%s\" has no PG source port.\n",
            netNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        return (BEAR_STOP_SEARCH);
    }

    if (Mac_GetAutoGraphOfNet(netPtr) != (Gt_Graph) NULL) {
        Mac_DeleteAutoGraphOfNet(netPtr);
    }

    if (Gre_BldGraph(wireDataPtr->bldDataPtr, netPtr, &autoGraph) ==
        BEAR_ERROR) {
        sprintf (bear_ErrorMsg, "Cannot build graph for net named \"%s\".\n",
            netNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        return (BEAR_STOP_SEARCH);
    }
    Mac_PutAutoGraphOfNet(autoGraph, netPtr);
    
    if (Mac_CheckAutoGraph(netPtr, &loopFlag, &connFlag) == BEAR_ERROR) {
        sprintf (bear_ErrorMsg, "Cannot check graph for net named \"%s\".\n",
            netNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        Mac_DeleteAutoGraphOfNet(netPtr);
        return (BEAR_STOP_SEARCH);
    }
    if (connFlag == FALSE) {
        sprintf (bear_ErrorMsg, "Net named \"%s\" is not connected.\n",
            netNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        return (BEAR_STOP_SEARCH);
    }

    netPtr->wirePtr = Pgr_MakeWire();
    netPtr->wirePtr->nodeCount = 1;

    /*
     * Go through each node in the graph of the given net, so
     * you can build up the power/ground node list.
     */
    addNodeData.srcNodePtr = NIL(Pgr_NodeProp);
    addNodeData.wirePtr = netPtr->wirePtr;
    if (Gt_ItrNodesOfGraph(autoGraph, AddNodeToList,
        (ClientData) &addNodeData) == BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }

    /* 
     * Go through each node in the graph of the given net to get
     * access to each of its edges, so you can build up the
     * power/ground edge list.
     */
    addEdgeData.planeDataPtr = wireDataPtr->planeDataPtr;
    addEdgeData.wirePtr = netPtr->wirePtr;
    if (Gt_ItrEdgesOfGraph(autoGraph, AddEdgeToList,
        (ClientData) &addEdgeData) == BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }

    if (CalcWireWidths(wireDataPtr->cellMasterPtr, netPtr) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }

    if (StoreNetsPgWireData(netPtr) == BEAR_ERROR) {
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Gre_InitPortsOfNet() --
 *
 * Initialize each port in the given net with the given values.
 *
 * Results:
 *   none
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

void
Gre_InitPortsOfNet(netPtr, maxVoltDrop, maxCurrentReq)
Net     *netPtr;
double  maxVoltDrop, maxCurrentReq;
{
    PortData    portData;
    NetKind     netKind;

    netKind = Mac_GetKindOfNet(netPtr);
    if ((netKind != POWER_NET) && (netKind != GROUND_NET)) {
        Mac_PutKindOfNet(POWER_NET, netPtr);
    }

    portData.portKind = PG_PORT;
    portData.maxVoltDrop = maxVoltDrop;
    portData.maxCurrentReq = maxCurrentReq;
    (void) Mac_ItrAllPortOfNet(netPtr, StorePortData, (ClientData) &portData);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gre_DoPgWiringForOneNet() --
 *
 * Calculate the power/ground wire data for all of the given net.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

BearStatus
Gre_DoPgWiringForOneNet(cellMasterPtr, netPtr)
CellMaster   *cellMasterPtr;
Net          *netPtr;
{
    BearStatus  bearStatus;
    WireData    wireData;
  
    wireData.bldDataPtr = Gre_BldGraphBegin(cellMasterPtr, FALSE, FALSE, TRUE);
  
    wireData.cellMasterPtr = cellMasterPtr;
    wireData.planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    bearStatus = DoPgWiringForOneNet(netPtr, (ClientData) &wireData);

    Gre_BldGraphEnd(wireData.bldDataPtr);

    return ((bearStatus == BEAR_STOP_SEARCH) ? BEAR_ERROR : BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gre_DoPgWiringForAllNets() --
 *
 * Calculate the power/ground wire data for all of the nets.
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

BearStatus
Gre_DoPgWiringForAllNets(cellMasterPtr)
CellMaster *cellMasterPtr;
{
    BearStatus  bearStatus;
    WireData    wireData;
  
    wireData.bldDataPtr = Gre_BldGraphBegin(cellMasterPtr, FALSE, FALSE, TRUE);
  
    wireData.cellMasterPtr = cellMasterPtr;
    wireData.planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    bearStatus = Mac_ItrNetOfCell(cellMasterPtr, DoPgWiringForOneNet, 
        (ClientData) &wireData);
  
    Gre_BldGraphEnd(wireData.bldDataPtr);
  
    return ((bearStatus == BEAR_STOP_SEARCH) ? BEAR_ERROR : BEAR_OK);
}


#endif USE_WIRE_CODE

