
/* greTiPlane.c --
 *
 *     Malgorzata Marek-Sadowska
 *     (Advising Professor - E. S. Kuh)
 *
 * File contains procedures to store and remove global routing
 * informaition in the tile plane.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/grEditor/greTiPlane.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "grEditor.h"
#include "grEditorInt.h"
#include "grStorage.h"


/* Here are global variables for files which insert data to Mac */

grEd_Node           *greNodeListHeadPtr;
grEd_Btn            *greBtnListHeadPtr;
grEd_BtnLine        *greBtnLineHeadPtr;
greLinkedDomTile    *greHeadLinkedDomTiles;
FloorPlanGraph      *greFpgPtr;
Net                 *greNetPtr;
PlaneData           *grePlaneDataPtr;


/* ---------- Local Procedures ---------- */
static BearStatus RemoveEdgeOfNet();


/*
 * ----------------------------------------------------------------------------
 *
 * GreAreTheseNodesLegal() --
 *
 * Check to see whether it's legal to add an edge between these
 * two nodes (where a node is a pin or a wall junction).
 *
 * Results:
 *   BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

BearStatus
GreAreTheseNodesLegal(pinOrJunct1Ptr, pinOrJunct2Ptr)
PinOrJunct *pinOrJunct1Ptr, *pinOrJunct2Ptr;
{
    if (pinOrJunct1Ptr->unType == FPG_WALL_JUNCT &&
        pinOrJunct2Ptr->unType == FPG_WALL_JUNCT) {

        if (pinOrJunct1Ptr->unValue.junctPtr ==
            pinOrJunct2Ptr->unValue.junctPtr) {

            PbPrint(EZ_ITALIC_FONT, "Looping on a node is not allowed.\n");
            return (BEAR_ERROR);
        }

        if (GreJunctTouchesJunct(pinOrJunct1Ptr->unValue.junctPtr,
            pinOrJunct2Ptr->unValue.junctPtr) == FALSE) {
            PbPrint(EZ_ITALIC_FONT,
                "Only wall junctions that touch may be connected.\n");
            return (BEAR_ERROR);
        }
    }

    if (pinOrJunct1Ptr->unType == FPG_PIN &&
        pinOrJunct2Ptr->unType == FPG_PIN) {

        if (pinOrJunct1Ptr->unValue.pinPtr ==
            pinOrJunct2Ptr->unValue.pinPtr) {
            PbPrint(EZ_ITALIC_FONT, "Looping on a node is not allowed.\n");
            return (BEAR_ERROR);
        }

/*
        grEd_GetCoveringWallSegment(pinOrJunct1Ptr->unValue.pinPtr);
        grEd_GetPtrOfDomTileCoveringPin(pinOrJunct1Ptr->unValue.pinPtr);
*/
    }

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gre_RemoveNet() --
 *
 *   First free the memory associated with the power ground wire data.
 *   Then remove all of the pseudo pins cooresponding to the given net
 *   from the tile plane, reset the width of all the real pins, and
 *   reset the properties of all the edges in the global routing graph
 *   to their default values.
 *
 * Results:
 *   none
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

void
Gre_RemoveNet(netPtr)
Net *netPtr;
{
/*
#ifdef BETTY_BOOP
    if (netPtr->clientData != (ClientData) NULL) {
        Pgr_FreeWire(&netPtr->wirePtr);
    }
#endif BETTY_BOOP
 */
netPtr->clientData = (ClientData) NULL;

    if (Mac_GetStatusOfNetInTiPlane(netPtr) == TRUE) {
        (void) Gt_ItrEdgesOfGraph(Mac_GetUserGraphOfNet(netPtr),
            RemoveEdgeOfNet, (ClientData) NULL);
        Mac_PutStatusOfNetInTiPlane((bool) FALSE, netPtr);
    }
}



/* ARGSUSED */
static BearStatus
RemoveEdgeOfNet(edge, clientData)
Gt_GraphEdge    edge;
ClientData      clientData;
{
    BNLineType      bnLineType;
    BNLine          *bnLinePtr;
    Gt_GraphNode    node1, node2;
    Mac_EdgeData    *edgeDataPtr;
    PinOrJunct      *pinOrJunct1Ptr;
    PinOrJunct      *pinOrJunct2Ptr;
    float           *widthVect;

    node1 = Gt_GetNodeOneOfEdge(edge);
    node2 = Gt_GetNodeTwoOfEdge(edge);

    pinOrJunct1Ptr = (PinOrJunct *) Gt_GetNodeDataOfNode(node1);
    pinOrJunct2Ptr = (PinOrJunct *) Gt_GetNodeDataOfNode(node2);
    
    edgeDataPtr = (Mac_EdgeData *) Gt_GetEdgeDataOfEdge(edge);

    if (edgeDataPtr->bnPtr != NIL(BottleNeck)) {
        if (edgeDataPtr->bnPtr->unType == TP_BN_TILE) {
            if (edgeDataPtr->ppin1Ptr != NIL(PseudoPin)) {
             (void) Tp_DeletePPinOfBnTile(edgeDataPtr->bnPtr->unValue.bnTilePtr,
                (SideDirection) BOTTOM_LEFT_SIDE,
                edgeDataPtr->ppin1Ptr);

             (void) Tp_DeletePPinOfBnTile(edgeDataPtr->bnPtr->unValue.bnTilePtr,
                (SideDirection) TOP_RIGHT_SIDE,
                edgeDataPtr->ppin1Ptr);
            }
            if (edgeDataPtr->ppin2Ptr != NIL(PseudoPin)) {
             (void) Tp_DeletePPinOfBnTile(edgeDataPtr->bnPtr->unValue.bnTilePtr,
                (SideDirection) BOTTOM_LEFT_SIDE,
                edgeDataPtr->ppin2Ptr);

             (void) Tp_DeletePPinOfBnTile(edgeDataPtr->bnPtr->unValue.bnTilePtr,
                (SideDirection) TOP_RIGHT_SIDE,
                edgeDataPtr->ppin2Ptr);
            }
        }
        else {
            if (edgeDataPtr->ppin1Ptr != NIL(PseudoPin)) {
                bnLinePtr = edgeDataPtr->bnPtr->unValue.bnLinePtr;
                bnLineType = Tp_GetBnLineTypeOfBnLine(bnLinePtr);

                if (bnLineType == (BNLineType) TP_HORZ_B_RIGHT) {
                  (void) Tp_DeletePPinOfBnLine(Tp_GetBRTileOfBnLine(bnLinePtr),
                    (SideDirection) TOP_RIGHT_SIDE,
                    TP_BOTTOM, edgeDataPtr->ppin1Ptr);
                }
                else if (bnLineType == (BNLineType) TP_HORZ_B_LEFT) {
                  (void) Tp_DeletePPinOfBnLine(Tp_GetTLTileOfBnLine(bnLinePtr),
                    (SideDirection) BOTTOM_LEFT_SIDE,
                    TP_BOTTOM, edgeDataPtr->ppin1Ptr);
                }
                else if (bnLineType == (BNLineType) TP_VERT_R_UP) {
                  (void) Tp_DeletePPinOfBnLine(Tp_GetBRTileOfBnLine(bnLinePtr),
                    (SideDirection) BOTTOM_LEFT_SIDE,
                    TP_BOTTOM, edgeDataPtr->ppin1Ptr);
                }
                else if (bnLineType == (BNLineType) TP_VERT_R_DOWN) {
                  (void) Tp_DeletePPinOfBnLine(Tp_GetTLTileOfBnLine(bnLinePtr),
                    (SideDirection) TOP_RIGHT_SIDE,
                    TP_BOTTOM, edgeDataPtr->ppin1Ptr);
                }
            }

            /*
             * edgeDataPtr->ppin2Ptr is expected to be NULL for the bottle
             * neck line case.
             */
        }
    }

    if (pinOrJunct1Ptr->unType == FPG_PIN) {
        widthVect = Mac_GetWidthVectOfPin(pinOrJunct1Ptr->unValue.pinPtr);
        widthVect[METAL1_LAYER] = MAXWIDTH;
        widthVect[METAL2_LAYER] = MAXWIDTH;
    }

    if (pinOrJunct2Ptr->unType == FPG_PIN) {
        widthVect = Mac_GetWidthVectOfPin(pinOrJunct2Ptr->unValue.pinPtr);
        widthVect[METAL1_LAYER] = MAXWIDTH;
        widthVect[METAL2_LAYER] = MAXWIDTH;
    }

    edgeDataPtr->width = MAXWIDTH;
    edgeDataPtr->bnPtr = NIL(BottleNeck);
    edgeDataPtr->ppin1Ptr = NIL(PseudoPin);
    edgeDataPtr->ppin2Ptr = NIL(PseudoPin);

    return (BEAR_CONTINUE_SEARCH);
}



BearStatus
Gre_StoreNetInTiPlane(netPtr, planeDataPtr)
Net         *netPtr;
PlaneData   *planeDataPtr;
{
    Gt_Graph userGraph;

    userGraph = Mac_GetUserGraphOfNet(netPtr);
    if (userGraph == (Gt_Graph) NULL) {
        return (BEAR_ERROR);
    }

    greNetPtr = netPtr;
    grePlaneDataPtr = planeDataPtr;
    greFpgPtr = Tp_GetFpg(planeDataPtr);

    grEd_FillInBtncks();
    grEd_FillInBtnLines();
    grEd_StoreJunctData();

    /* end building internal data structure  */

    (void) Gt_ItrEdgesOfGraph(userGraph, grEd_InsertEdgeOfGraph,
        (ClientData) NULL);

    Mac_PutStatusOfNetInTiPlane((bool) TRUE, netPtr);

    /* 
     * Free internal data structures.
     */
    grEd_FreeJunctData();

    return (BEAR_OK);
}



