
/* gtGraphCheck.c --
 *
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/gtGraph/gtGraphCheck.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include <stdio.h>
#include "gtBearStuff.h"
#include "gtGraph.h"
#include "gtGraphInt.h"


/* ---------- Local Procedures ---------- */
static void CheckInt();


/* ---------- Local Global Variables ---------- */
static int  count;
static bool loopFlag;


/*
 * ----------------------------------------------------------------------------
 *
 * Gt_CheckGraph() --
 *
 * Check the graph for loops and connectivity.
 *
 * Results:
 *   The loop and connectivity flags are filled in.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

void
Gt_CheckGraph(graph, loopFlagPtr, connFlagPtr)
Gt_Graph    graph;                          /* IN       */
bool        *loopFlagPtr;                   /* OUT      */
bool        *connFlagPtr;                   /* OUT      */
{
    edge_t      e;
    graph_t     g;
    lsGen       gen;
    lsList      edgeList, vertList;
    vertex_t    startVert, v;
    int         numOfVert;

    g = ((Gt_GraphInt *) graph)->g;

#ifdef DEBUG
    if (g == (graph_t) NULL) {
        Bear_PrintFatalErr("Gt_CheckGraph() failed : no graph given");
    }

    if (gt_DebugFlag == TRUE) {
        GraStatus graStatus;

        graStatus = g_check(g);
        (void) printf("Graph %d (is %s)\n", Gt_GetIdOfGraph(graph),
            (graStatus == GRA_OK) ? "okay" :
                ((graStatus == GRA_DISCONN) ? "not connected" : "NOT okay"));
    }
#endif DEBUG

    edgeList = g_get_edges(g);
    gen = lsStart(edgeList);
    while (lsNext(gen, (lsGeneric *) &e, LS_NH) == LS_OK) {
        ((Gt_GraphEdgeInt *) e->user_data)->mark = BEAR_UNMARKED;
    }
    lsFinish(gen);

    vertList = g_get_vertices(g);
    if (lsFirstItem(vertList, (lsGeneric *) &startVert, LS_NH) == LS_NOMORE) {
        *loopFlagPtr = FALSE;
        *connFlagPtr = FALSE;
        return;
    }

    gen = lsStart(vertList);
    numOfVert = 0;
    while (lsNext(gen, (lsGeneric *) &v, LS_NH) == LS_OK) {
        ((Gt_GraphNodeInt *) v->user_data)->mark = BEAR_UNMARKED;
        numOfVert++;
    }
    lsFinish(gen);

    loopFlag = FALSE;
    count = 0;
    CheckInt(startVert);
    *loopFlagPtr = loopFlag;
    *connFlagPtr = (numOfVert == count);

#ifdef DEBUG
    if (gt_DebugFlag == TRUE) {
        printf ("\n");
        printf ("numOfVert = %d\n", numOfVert);
        printf ("count = %d\n", count);
        printf ("Graph %s a loop.\n",
            (*loopFlagPtr == TRUE) ? "has" : "does not have");
        printf ("Graph %s connected.\n",
            (*connFlagPtr == TRUE) ? "is" : "is not");
    }
#endif DEBUG
}


static void
CheckInt(v)
vertex_t v;
{
    if (((Gt_GraphNodeInt *) v->user_data)->mark == BEAR_UNMARKED) {
        lsGen   gen;
        edge_t  e;

#ifdef DEBUG
        if (gt_DebugFlag == TRUE) {
            printf ("%d ", ((Gt_GraphNodeInt *) v->user_data)->id);
            fflush (stdout);
        }
#endif DEBUG

        ((Gt_GraphNodeInt *) v->user_data)->mark = BEAR_MARKED;
        count++;

        foreach_out_edge(v, gen, e) {
            if (((Gt_GraphEdgeInt *) e->user_data)->mark == BEAR_UNMARKED) {

                ((Gt_GraphEdgeInt *) e->user_data)->mark = BEAR_MARKED;
                CheckInt(g_e_dest(e));
            }
        }
        foreach_in_edge(v, gen, e) {
            if (((Gt_GraphEdgeInt *) e->user_data)->mark == BEAR_UNMARKED) {

                ((Gt_GraphEdgeInt *) e->user_data)->mark = BEAR_MARKED;
                CheckInt(g_e_source(e));
            }
        }
    }
    else {
#ifdef DEBUG
        if (gt_DebugFlag == TRUE) {
        printf ("L\n");
        fflush (stdout);
        }
#endif DEBUG

        loopFlag = TRUE;
    }
}


