
/* gtGraphPrim.c --
 *
 *     George Carvalho, Esq.
 *     Wei-Ming Dai
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * This file contains the primitive operations for a graph.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/gtGraph/gtGraphPrim.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include <stdio.h>
#include "gtBearStuff.h"
#include "gtGraph.h"
#include "gtGraphInt.h"


int
Gt_GetDegreeOfNode(node)
Gt_GraphNode node;                           /* IN       */
{
    vertex_t v;

    v = ((Gt_GraphNodeInt *) node)->v;

    return (lsLength(g_get_in_edges(v)) + lsLength(g_get_out_edges(v)));
}


bool
Gt_IsNodeUnconnected(node)
Gt_GraphNode node;                           /* IN       */
{
    return ((Gt_GetDegreeOfNode(node) == 0) ? TRUE : FALSE);
}


BearGeneric
Gt_GetGraphDataOfGraph(graph)
Gt_Graph graph;                              /* IN       */
{
    return (((Gt_GraphInt *) graph)->graphData);
}


BearGeneric
Gt_GetNodeDataOfNode(node)
Gt_GraphNode node;                           /* IN       */
{
#ifdef DEBUG
    if (node == (Gt_GraphNode) NULL) {
        Bear_PrintFatalErr("Gt_GetNodeDataOfNode() failed : no node given");
    }
#endif DEBUG

    return (((Gt_GraphNodeInt *) node)->nodeData);
}


BearGeneric
Gt_PutNodeDataOfNode(nodeData, node)
BearGeneric     nodeData;                    /* IN       */
Gt_GraphNode    node;                        /* IN       */
{
#ifdef DEBUG
    if (node == (Gt_GraphNode) NULL) {
        Bear_PrintFatalErr("Gt_PutNodeDataOfNode() failed : no node given");
    }
#endif DEBUG

    ((Gt_GraphNodeInt *) node)->nodeData = nodeData;
}


BearGeneric
Gt_GetEdgeDataOfEdge(edge)
Gt_GraphEdge edge;                           /* IN       */
{
#ifdef DEBUG
    if (edge == (Gt_GraphEdge) NULL) {
        Bear_PrintFatalErr("Gt_GetEdgeDataOfEdge() failed : no edge given");
    }
#endif DEBUG

    return (((Gt_GraphEdgeInt *) edge)->edgeData);
}


BearGeneric
Gt_PutEdgeDataOfEdge(edgeData, edge)
BearGeneric     edgeData;                    /* IN       */
Gt_GraphEdge    edge;                        /* IN       */
{
#ifdef DEBUG
    if (edge == (Gt_GraphEdge) NULL) {
        Bear_PrintFatalErr("Gt_PutEdgeDataOfEdge() failed : no edge given");
    }
#endif DEBUG

    ((Gt_GraphEdgeInt *) edge)->edgeData = edgeData;
}


Gt_GraphNode
Gt_GetNodeOneOfEdge(edge)
Gt_GraphEdge edge;                           /* IN       */
{
    edge_t e;

#ifdef DEBUG
    if (edge == (Gt_GraphEdge) NULL) {
        Bear_PrintFatalErr("Gt_GetNodeOneOfEdge() failed : no edge given");
    }
#endif DEBUG

    e = ((Gt_GraphEdgeInt *) edge)->e;
    return ((Gt_GraphNode) (g_e_source(e)->user_data));
}


Gt_GraphNode
Gt_GetNodeTwoOfEdge(edge)
Gt_GraphEdge edge;                           /* IN       */
{
    edge_t e;

#ifdef DEBUG
    if (edge == (Gt_GraphEdge) NULL) {
        Bear_PrintFatalErr("Gt_GetNodeTwoOfEdge() failed : no edge given");
    }
#endif DEBUG

    e = ((Gt_GraphEdgeInt *) edge)->e;
    return ((Gt_GraphNode) (g_e_dest(e)->user_data));
}


BearId
Gt_GetIdOfGraph(graph)
Gt_Graph graph;                              /* IN       */
{
    return (((Gt_GraphInt *) graph)->id);
}


BearId
Gt_GetIdOfNode(node)
Gt_GraphNode node;                           /* IN       */
{
    return (((Gt_GraphNodeInt *) node)->id);
}


BearId
Gt_GetIdOfEdge(edge)
Gt_GraphEdge edge;                           /* IN       */
{
    return (((Gt_GraphEdgeInt *) edge)->id);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Gt_GetNumOfGraphNodes() --
 *
 * Return the number of nodes in the graph.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

int
Gt_GetNumOfGraphNodes(graph)
Gt_Graph graph;                                 /* IN       */
{
#ifdef DEBUG
    if (graph == (Gt_Graph) NULL) {
        char *s = "Wettin' you tink tday ?";
        Bear_PrintFatalErr("Gt_GetNumOfGraphNodes() failed : no graph given");
    }
#endif DEBUG

    return (((Gt_GraphInt *) graph)->numOfNodes);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Gt_GetNumOfGraphEdges() --
 *
 * Return the number of edges in the graph.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

int
Gt_GetNumOfGraphEdges(graph)
Gt_Graph graph;                                 /* IN       */
{
#ifdef DEBUG
    if (graph == (Gt_Graph) NULL) {
        Bear_PrintFatalErr("Gt_GetNumOfGraphEdges() failed : no graph given");
    }
#endif DEBUG

    return (((Gt_GraphInt *) graph)->numOfEdges);
}



