
/* cpaUtils.c --
 *
 *     Benjamin Chen
 *     Electronics Research Laboratory
 *     University of California, Berkeley
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/koalaBear/cpaUtils.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#ifdef CPA_BEAR
#include "bearCopyRight.h"
#endif CPA_BEAR
#include <stdio.h>
#include <string.h>
#include "cpa.h"
#include "cpaInt.h"
#include "cpaList.h"
#include "sort.h"


/*
 * Routines to create and arrange data structures
 * for the Channel Pin Arrangement algorithm.
 */

static void OldPrintChannel(channel)
     BChannel *channel;
{
    BNet *net;
    Pin *pin;
    List *pinList;

    net = C_NETS(channel);
    while(net != (BNet *) NULL) {
    (void) printf("Net (%s): \n", NAME(net));
    pinList = N_PINS(net);
    while (pinList != (List *) NULL) {
        pin = (Pin *) L_ITEM(pinList);
        (void) printf("%s %d ", NAME(pin), P_COLUMN(pin));
        pinList = L_NEXT(pinList);
    }
    (void) printf("\n");
    net = NEXT(net);
    }
}

void PrintChannel(channel)
     BChannel *channel;
{
    FILE    *file = stdout;
    BNet    *net;
    Pin     *pin;
    List    *pinList;

    if (cpaOutFiName != (char *) NULL) {

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, cpaOutFiName);
#endif /* USE_OCT_CODE */

        if ((file = fopen (cpaOutFiName, "w")) == (FILE *) NULL) {
            (void) fprintf (stderr,
                "Could not open file named \"%s\" for writing.\n",
                cpaOutFiName);
            KB_EXIT(-1);
        }
    }

    net = C_NETS(channel);
    while(net != (BNet *) NULL) {
        pinList = N_PINS(net);
        while (pinList != (List *) NULL) {
            pin = (Pin *) L_ITEM(pinList);
            (void) fprintf (file, "%8s %8d\n", NAME(pin), P_COLUMN(pin));
            pinList = L_NEXT(pinList);
        }
        net = NEXT(net);
    }

    if (file != stdout) {
        fclose (file);
    }
}


/*
 * SortNetList --
 *     Sorts each of the net lists in order of decreasing criticality.
 */
void SortNetList(netlist)
     NetClassList *netlist;
{
    MergeSort((Generic **) &LS_NETS(netlist));
    MergeSort((Generic **) &RS_NETS(netlist));
    MergeSort((Generic **) &BS_NETS(netlist));
    MergeSort((Generic **) &LP_NETS(netlist));
    MergeSort((Generic **) &RP_NETS(netlist));
    MergeSort((Generic **) &BP_NETS(netlist));
    MergeSort((Generic **) &IP_NETS(netlist));
}

/*
 * AddPinGlobal --
 *    Adds a pin to the head of the global pin list.
 */
void AddPinGlobal(pin)
     Pin *pin;
{
    List *entry;

    entry = NewEntry((char *) pin);
    L_NEXT(entry) = globalPinList;
    globalPinList = entry;
}


/*
 * SearchGlobalPin --
 *     Searches the global pin list for the given name and returns the
 *     pin associated with the name.  If the name does not exist, it returns
 *     NULL.
 */
Pin *SearchGlobalPin(pinName)
     char *pinName;
{
    List *current;
    Pin *pin;

    current = globalPinList;
    while(current != NULL) {
    pin = (Pin *) L_ITEM(current);
    if (!strcmp(NAME(pin), pinName)) return pin;
    current = L_NEXT(current);
    }
    return NULL;
}

/*
 * MakeChannel --
 *     This procedure creates and returns a channel structure with the
 *     given fields set.
 */
BChannel *MakeChannel(height, topLength, bottomLength, spacing, class)
     int height, topLength, bottomLength, spacing, class;
{
    BChannel *newChannel;

    MALLOC(BChannel *, newChannel, sizeof(BChannel));
    C_HEIGHT(newChannel) = height;
    C_TOP(newChannel) = topLength;
    C_BOTTOM(newChannel) = bottomLength;
    C_SPACING(newChannel) = spacing;
    C_CLASS(newChannel) = (ChannelClass) class;
    return newChannel;
}

/*
 * MakePin --
 *     This procedure creates and returns a pin structure with the
 *     given fields set.
 */
Pin *MakePin(name, slotnumber)
     char *name;
     int slotnumber;
{
    Pin *newPin;

    MALLOC(Pin *, newPin, sizeof(Pin));
    NAME(newPin) = name;
    /* slots will correspond to columns during the pin arrangement */
    P_COLUMN(newPin) = slotnumber;
    NEXT(newPin) = (Pin *) NULL;
    AddPinGlobal(newPin);
    return newPin;
}

/*
 * AddPin --
 *    This procedure attaches pin2 to the front of pin1's list, and returns
 *    the modified list.
 */
Pin *AddPin(pin1, pin2)
     Pin *pin1, *pin2;
{
    NEXT(pin2) = pin1;
    return pin2;
}

/*
 * MakePinItem --
 *    This procedure searches the globalPinList for pin entry with the given
 *    pin name and allocates a list item for it.
 */
List *MakePinItem(pinName)
     char *pinName;
{
    List *entry;
    Pin *pin;

    if ((pin = SearchGlobalPin(pinName)) == (Pin *) NULL) {
        fprintf(stderr, "Net refers to undefined pin. \n");
        KB_EXIT(-1);
    }
    return NewEntry((char *) pin);
}

/*
 * AddPinItem --
 *    This procedure adds the pin specified by pinName to the front
 *    of the given pin list.
 */
List *AddPinItem(pinList, pinName)
     List *pinList;
     char *pinName;
{
    List *newItem;

    newItem = MakePinItem(pinName);
    L_NEXT(newItem) = pinList;
    return newItem;
}

/*
 * Return 1 if the pin is a located on the top edge of the channel and
 * 0 if the pin is on the bottom edge of the channel.  If the pin has a
 * location of 0, this is an error condition..
 */
int top_pin(pin)
     Pin *pin;
{
    if (P_COLUMN(pin) > 0) return 1;
    else if (P_COLUMN(pin) < 0) return 0;
    else {
    fprintf(stderr,
        "top_pin(): Pin location is zero for pin %s.\n", NAME(pin));
    KB_EXIT(-1);
    }
}

#define PIN_ITEM(net) L_ITEM(N_PINS(net))

/*
 * AddNetClassList --
 *     Adds the net to the appropriate netclass list.  There exists only
 *     two such lists: one for nets whose pins are located in the top
 *     edge of the channel, and one for those on the bottom.  Single pin
 *     nets exist in only one such list (i.e. an LS net entry exists only
 *     in the bottomNetList if its single pin is located on the bottom edge
 *     of the channel.   Pin locations are determined by the sign of its
 *     slotnumber (see cpa.h).  Two pin nets will exist in both netclass
 *     lists.
 */
void AddNetClassList(net)
     BNet *net;
{
    NetClassList *targetNetList;

    if (topNetList == (NetClassList *) NULL) {
    MALLOC(NetClassList *, topNetList, sizeof(NetClassList));
    }
    if (bottomNetList == (NetClassList *) NULL) {
    MALLOC(NetClassList *, bottomNetList, sizeof(NetClassList));
    }
    if (top_pin(PIN_ITEM(net))) targetNetList = topNetList;
    else targetNetList = bottomNetList;

    switch(N_CLASS(net)) {
    case LS:
    LS_NETS(targetNetList) = AddEntry(LS_NETS(targetNetList), net);
    break;
    case RS:
    RS_NETS(targetNetList) = AddEntry(RS_NETS(targetNetList), net);
    break;
    case BS:
    BS_NETS(targetNetList) = AddEntry(BS_NETS(targetNetList), net);
    break;
    case LP:
    LP_NETS(bottomNetList) = LP_NETS(topNetList) = 
        AddEntry(LP_NETS(topNetList), net);
    break;
    case RP:
    RP_NETS(bottomNetList) = RP_NETS(topNetList) = 
        AddEntry(RP_NETS(topNetList), net);
    break;
    case BP:
    BP_NETS(bottomNetList) = BP_NETS(topNetList) = 
        AddEntry(BP_NETS(topNetList), net);
    break;
    case IP:
    IP_NETS(bottomNetList) = IP_NETS(topNetList) = 
        AddEntry(IP_NETS(topNetList), net);
    break;
    default:
    break;
    }
}

/*
 * CalcNetClass --
 *     Determine the netclass based upon the number of pins in the net
 *     and direction of the exits. (see cpa.h)
 */
NetClass CalcNetClass(exits, pinCount)
     Exit exits;
     int pinCount;
{
    switch(exits) {
    case LEFT_EXIT:
    if (pinCount == 1) {
        return LS;
    } 
    else if (pinCount == 2) {
        return LP;
    }
    else {
        fprintf(stderr,
        "CalcNetClass(): Basic nets must have either 1 or 2 pins.\n");
        KB_EXIT(-1);
    }
    case RIGHT_EXIT:
    if (pinCount == 1) {
        return RS;
    } 
    else if (pinCount == 2) {
        return RP;
    }
    else {
        fprintf(stderr,
        "CalcNetClass(): Basic nets must have either 1 or 2 pins.\n");
    }
    case BOTH_EXITS:
    if (pinCount == 1) {
        return BS;
    } 
    else if (pinCount == 2) {
        return BP;
    }
    else {
        fprintf(stderr, 
        "CalcNetClass(): Basic nets must have either 1 or 2 pins.\n");
        KB_EXIT(-1);       
    }
    case NO_EXITS:
    if (pinCount == 2) {
        return IP;
    }
    else {
        fprintf(stderr,
       "CalcNetClass(): Net with no exits must have at least two pins.\n");
        KB_EXIT(-1);       
    }   
    default:
    fprintf(stderr,
        "CalcNetClass(): Exit type not left, right, none, or both.\n");
    KB_EXIT(-1);
    }
}

/*
 * AttachNetToPins --
 *      Sets the net field of each pin in the list.
 */
void AttachNetToPins(net, pinList)
     BNet *net;
     List *pinList;
{
    Pin *pin;
    while(pinList != (List *) NULL) {
    pin = (Pin *) L_ITEM(pinList);
    P_NET(pin) = net;
    pinList = L_NEXT(pinList);
    }
}

/*
 * NewNet --
 *     This procedure creates and returns a net structure with the
 *     given fields set.
 */
BNet *NewNet(name, netclass, criticality, pinList) 
     char *name;
     NetClass netclass;
     int criticality;
     List *pinList;
{
    BNet *newNet;

    MALLOC(BNet *, newNet, sizeof(BNet));
    NAME(newNet) = name;
    N_CLASS(newNet) = netclass;
    N_CRITICAL(newNet) = criticality;
    N_PINS(newNet) = pinList;
    NEXT(newNet) = (BNet *) NULL;
    return newNet;
}

/*
 * MakeNet --
 *     This procedure creates and returns a net structure with the
 *     given fields set.  In addition, it attaches the net to each pin in the
 *     net for future reference.  Finally, it adds the new net to the
 *     global netclass list.
 */
BNet *MakeNet(name, criticality, exits, pinList)
     char *name;
     int criticality;
     int exits;
     List *pinList;
{
    BNet *newNet;
    Pin *pin1, *pin2;
    int pinCount;
    int nonBasic = 0;

    pinCount = listLength(pinList);
    
    if(pinCount == 2) {
        pin1 = (Pin *) L_ITEM(pinList);
        pin2 = (Pin *) L_ITEM(L_NEXT(pinList));
        if((P_COLUMN(pin1) > 0 && P_COLUMN(pin2) > 0)
           || (P_COLUMN(pin1) < 0 && P_COLUMN(pin2) < 0))
            nonBasic = 1;
    }
    nonBasic = nonBasic || (pinCount > 2);
    if(nonBasic) {
    BNet *bnets;

    newNet = NewNet(name, NB, criticality, pinList);
    bnets = NetToBNets(newNet, exits);
    while(bnets != (BNet *) NULL) {
        AttachNetToPins(bnets, N_PINS(bnets));
        AddNetClassList(bnets);
        bnets = NEXT(bnets); 
    }
    }
    else {
    newNet = NewNet(name, CalcNetClass((Exit) exits, pinCount),
            criticality, pinList);
    AttachNetToPins(newNet, pinList);
    AddNetClassList(newNet);
    }

    return newNet;
}

/*
 * AddNet --
 *    This procedure attaches net2 to the front of net1's list, and returns
 *    the modified list.
 */
BNet *AddNet(net1, net2)
     BNet *net1, *net2;
{
    NEXT(net2) = net1;
    return net2;
}
