
/* macCellItr.c --
 *
 *    Wei-Ming Dai
 *    June Wang
 *    (Advising Professor - E. S. Kuh)
 *
 * It has all the iterators related to cell masters or cell instances.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macCellItr.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


typedef struct _LocalData LocalData;
struct _LocalData {
    BearStatus (*func)();
    ClientData clientData;
    Line       *interval;
};


extern BearStatus CheckIntervalOfPin();


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrInstOfCell() --
 *
 *    Given a cell master pointer, it iterates all the cell 
 *    instances under that cell master.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful.
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrInstOfCell(cellMasterPtr, func, clientData)
CellMaster *cellMasterPtr;
BearStatus (*func)();
ClientData clientData;
{
    register CellInst *curCellInstPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))  {
        MacPrintFatalErr("Mac_ItrInstOfCell() failed: no cell is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrInstOfCell() failed: no func is given");
    }
#endif DEBUG

    if (cellMasterPtr->instList == NIL(CellInst))
    {
        return (BEAR_COMPLETE_SEARCH);
    }
    else
    {
        curCellInstPtr = cellMasterPtr->instList;
        while (curCellInstPtr != NIL(CellInst))
        {
            if (((*func)(curCellInstPtr, clientData)) == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
            curCellInstPtr = curCellInstPtr->nextPtr;
        }
    }
    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPortOfCellMaster() --
 *
 *    Given a cell master, it iterates all of the port masters
 *    of that cell master.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful.
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrPortOfCellMaster(cellMasterPtr, func, clientData)
CellMaster  *cellMasterPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortMaster *curPortMasterPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_ItrPortOfCellMaster() failed: no cell master is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrPortOfCellMaster() failed: no func is given");
    }
#endif DEBUG

    if (cellMasterPtr->portMasterList == NIL(PortMaster))
    {
        return (BEAR_COMPLETE_SEARCH);
    }
    else
    {
        curPortMasterPtr = cellMasterPtr->portMasterList;
        while (curPortMasterPtr != NIL(PortMaster))
        {
            if (((*func)(curPortMasterPtr, clientData)) == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
            curPortMasterPtr = curPortMasterPtr->nextPtr;
        }
    }
    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPinOfCellMaster() --
 *
 *    Given a cell master, it iterates all the pin masters 
 *    of that cell master.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful.
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrPinOfCellMaster(cellMasterPtr, func, clientData)
CellMaster  *cellMasterPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortMaster   *curPortMasterPtr;
    register PinMaster    *curPinMasterPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))  {
    MacPrintFatalErr("Mac_ItrPinOfCellMaster() failed: no cell instance is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrPinOfCellMaster() failed: no func is given");
    }
#endif DEBUG

    if (cellMasterPtr->portMasterList == NIL(PortMaster)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (curPortMasterPtr = cellMasterPtr->portMasterList;
             curPortMasterPtr != NIL(PortMaster);
             curPortMasterPtr = curPortMasterPtr->nextPtr) {

            for (curPinMasterPtr = curPortMasterPtr->pinMasterList;
                 curPinMasterPtr != NIL(PinMaster);
                 curPinMasterPtr = curPinMasterPtr->nextPtr) {

                if (((*func)(curPinMasterPtr, clientData)) ==
                    BEAR_STOP_SEARCH) {
                    return (BEAR_STOP_SEARCH);
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPortOfCellInst() --
 *
 *  Given a cell instance, it iterates all the port instances
 *  of that cell instance.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrPortOfCellInst(cellInstPtr, func, clientData)
CellInst    *cellInstPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortInst *portPtr;

#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) {
 MacPrintFatalErr("Mac_ItrPortOfCellInst() failed: no cell instance is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrPortOfCellInst() failed: no func is given");
    }
#endif DEBUG

    if (cellInstPtr->portInstList == NIL(PortInst)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (portPtr = cellInstPtr->portInstList;
            portPtr != NIL(PortInst);
            portPtr = portPtr->nextCellPortPtr) {

            if (((*func)(portPtr, clientData)) == BEAR_STOP_SEARCH) {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPinOfCellInst() --
 *
 *    Every pin instance of every port instance in the cell
 *    instance is visited.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful.
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrPinOfCellInst(cellInstPtr, func, clientData)
CellInst    *cellInstPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortInst   *portPtr;
    register PinInst    *pinPtr;

#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))  {
   MacPrintFatalErr("Mac_ItrPinOfCellInst() failed: no cell instance is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrPinOfCellInst() failed: no func is given");
    }
#endif DEBUG

    if (cellInstPtr->portInstList == NIL(PortInst)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (portPtr = cellInstPtr->portInstList;
            portPtr != NIL(PortInst);
            portPtr = portPtr->nextCellPortPtr) {

            for (pinPtr = portPtr->pinList;
                pinPtr != NIL(PinInst);
                pinPtr = pinPtr->nextPtr) {

                if (((*func)(pinPtr, clientData)) == BEAR_STOP_SEARCH) {
                    return (BEAR_STOP_SEARCH);
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * CheckIntervalOfPin() --
 *
 *    Given a pin instance, it checks that the pin instance is within the
 *    interval specified. And then applies the users function to the pin.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
CheckIntervalOfPin(pinPtr, localDataPtr)
PinInst     *pinPtr;
LocalData   *localDataPtr;
{
    Point absPos;

    Mac_GetAbsPosOfPin(pinPtr, &absPos);
    if ((absPos.p_x >= localDataPtr->interval->point1.p_x) &&
        (absPos.p_x <= localDataPtr->interval->point2.p_x) &&
        (absPos.p_y >= localDataPtr->interval->point1.p_y) &&
        (absPos.p_y <= localDataPtr->interval->point2.p_y)) {

        if ((*(localDataPtr->func))(pinPtr,
            localDataPtr->clientData) == BEAR_STOP_SEARCH) {
            return (BEAR_STOP_SEARCH);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPinOfIntervalOnCellBnd() --
 *
 *    Given a cell instance, it iterates all the pin instances of
 *    that cell instance that are within the interval specified.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful.
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrPinOfIntervalOnCellBnd (cellInstPtr, interval, func, clientData)
CellInst   *cellInstPtr;
Line       *interval;
BearStatus (*func)();
ClientData clientData;
{
    LocalData localData;

#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))  {
MacPrintFatalErr("Mac_ItrPinOfIntervalOnCellBnd() failed: no cell instance is given");
    } 
    else if (func == (BearStatus (*)()) NULL) {
    MacPrintFatalErr("Mac_ItrPinOfIntervalOnCellBnd() failed: no func is given");
    }
#endif DEBUG

    localData.func = func;
    localData.interval = interval;
    localData.clientData = clientData;
    if (Mac_ItrPinOfCellInst(cellInstPtr, CheckIntervalOfPin,
        (ClientData) &localData) == BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrChildrenOfCell() --
 *
 *    It iterates all of the cell instances, of all of the cell masters
 *    which are the children of the given cell master.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful;
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrChildrenOfCell(cellMasterPtr, func, clientData)
CellMaster  *cellMasterPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register CellInst    *curCellInstPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster)) {
       MacPrintFatalErr("Mac_ItrChildrenOfCell() failed: no parent cell is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrChildrenOfCell() failed: no func is given");
    }
#endif DEBUG

    if (cellMasterPtr->contents.cellInstList == NIL(CellInst)) {
        return (BEAR_COMPLETE_SEARCH);
    } 
    else {
        curCellInstPtr = cellMasterPtr->contents.cellInstList;
        while (curCellInstPtr != NIL(CellInst)) {
            if (((*func)(curCellInstPtr, clientData)) == BEAR_STOP_SEARCH) {
                return (BEAR_STOP_SEARCH);
            }
            curCellInstPtr = curCellInstPtr->siblingPtr;
        }
    }
    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrNetOfCell() --
 *
 *    It iterates all of the nets which are inside
 *    the given cell master.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful;
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrNetOfCell(cellMasterPtr, func, clientData)
CellMaster  *cellMasterPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register Net *netPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_ItrNetOfCell() failed: no cell is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrNetOfCell() failed: no func is given");
    }
#endif DEBUG

    if (cellMasterPtr->contents.netList == NIL(Net)) {
        return (BEAR_COMPLETE_SEARCH);
    } 
    else {
        for (netPtr = cellMasterPtr->contents.netList;
            netPtr != NIL(Net);
            netPtr = netPtr->nextPtr) {

            if (((*func)(netPtr, clientData)) == BEAR_STOP_SEARCH) {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRCellOfCell() --
 *
 *    It iterates all of the route cells which are inside 
 *    the given cell master.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful;
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrRCellOfCell(cellMasterPtr, func, clientData)
CellMaster  *cellMasterPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register RouteCell *curRCellPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))
    {
        MacPrintFatalErr("Mac_ItrRCellOfCell() failed: no cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrRCellOfCell() failed: no func is given");
    }
#endif DEBUG

    if (cellMasterPtr->contents.routeCellList == NIL(RouteCell))
    {
        return (BEAR_COMPLETE_SEARCH);
    } 
    else
    {
        curRCellPtr = cellMasterPtr->contents.routeCellList;
        while (curRCellPtr != NIL(RouteCell))
        {
            if (((*func)(curRCellPtr, clientData)) == BEAR_STOP_SEARCH)
            {
                return (BEAR_STOP_SEARCH);
            }
            curRCellPtr = curRCellPtr->nextPtr;
        }
    }
    return (BEAR_COMPLETE_SEARCH);
}


