
/* macCellPrim.c --
 *
 *    Wei-Ming Dai
 *    June Wang
 *    (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macCellPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"

typedef struct _MovePinData {
    Point vector;
    Rect  *bboxPtr;
    double xRatio, yRatio;
} MovePinData;


/* ---------- Local Procedures ---------- */
static BearStatus MovePin();
static BearStatus PinMirrorXSwitch();
static BearStatus PinMirrorYSwitch();
static BearStatus Pin90Switch();
static BearStatus Pin180Switch();
static BearStatus Pin270Switch();
static BearStatus PinRef45Switch();
static BearStatus PinRef135Switch();
static BearStatus LineSegMirrorXSwitch();
static BearStatus LineSegMirrorYSwitch();
static BearStatus LineSeg90Switch();
static BearStatus LineSeg180Switch();
static BearStatus LineSeg270Switch();
static BearStatus LineSegRef45Switch();
static BearStatus LineSegRef135Switch();
static BearStatus MoveResizedBoundary();
static int        WhatTransform();


BearString 
Mac_GetFileNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->contents.cmFileName);
}


  
BearString 
Mac_GetCTreeFileNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->contents.ctreeFileName);
}


  
BearString 
Mac_GetFlexFileNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->contents.flexFileName);
}


  
BearString 
Mac_GetPgFileNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->contents.pgFileName);
}


  
BearString 
Mac_GetTechFileNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->contents.techFileName);
}



BearString 
Mac_GetTimingFileNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->contents.tmFileName);
}


  
void
Mac_PutFileNameOfCellMaster(fileName, cellMasterPtr) 
BearString  fileName;
CellMaster  *cellMasterPtr;
{
    cellMasterPtr->contents.cmFileName = fileName;
}



void
Mac_PutCTreeFileNameOfCellMaster(ctreeFileName, cellMasterPtr) 
BearString  ctreeFileName;
CellMaster  *cellMasterPtr;
{
    cellMasterPtr->contents.ctreeFileName = ctreeFileName;
}



void
Mac_PutFlexFileNameOfCellMaster(flexFileName, cellMasterPtr) 
BearString  flexFileName;
CellMaster  *cellMasterPtr;
{
    cellMasterPtr->contents.flexFileName = flexFileName;
}



void
Mac_PutPgFileNameOfCellMaster(pgFileName, cellMasterPtr) 
BearString  pgFileName;
CellMaster  *cellMasterPtr;
{
    cellMasterPtr->contents.pgFileName = pgFileName;
}



void
Mac_PutTechFileNameOfCellMaster(techFileName, cellMasterPtr) 
BearString  techFileName;
CellMaster  *cellMasterPtr;
{
    cellMasterPtr->contents.techFileName = techFileName;
}



void
Mac_PutTimingFileNameOfCellMaster(timingFileName, cellMasterPtr) 
BearString  timingFileName;
CellMaster  *cellMasterPtr;
{
    cellMasterPtr->contents.tmFileName = timingFileName;
}



BearString 
Mac_GetNameOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster)) {
    MacPrintFatalErr("Mac_GetNameOfCellMaster() failed: no cell master is given");
    }
#endif DEBUG

    return (cellMasterPtr->name);
}


  
BearStatus
Mac_PutNameOfCellMaster(cellMastName, cellMasterPtr) 
BearString  cellMastName;
CellMaster  *cellMasterPtr;
{
#ifdef DEBUG
    if (cellMastName == (BearString) NULL) 
    {
        MacPrintFatalErr("Mac_PutNameOfCellMaster() failed: no name is given");
    }
    else if (cellMasterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_PutNameOfCellMaster() failed: no cell master is given");
    }
#endif DEBUG

    if (Mac_IsMember(CELL_MASTER_HT, cellMastName)) {
        sprintf(bear_ErrorMsg, "Cannot have two cell masters named \"%s\"",
            cellMastName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    } 
    cellMasterPtr->name = cellMastName;
    Mac_Insert(CELL_MASTER_HT, cellMastName, (char *) cellMasterPtr);

    return (BEAR_OK);
}



BearId
Mac_GetIdOfCellMaster(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
    return (cellMasterPtr->id);
}

  

BearString
Mac_GetNameOfCellInst(cellInstPtr) 
CellInst *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) {
  MacPrintFatalErr("Mac_GetNameOfCellInst() failed: no cell instance is given");
    }
#endif DEBUG

    return (cellInstPtr->name);
}



BearStatus
Mac_PutNameOfCellInst(cellInstName, cellInstPtr) 
BearString cellInstName;
CellInst   *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstName == (BearString) NULL) {
        MacPrintFatalErr("MacPutNameOfCellInst() failed: no name is given");
    }
    else if (cellInstPtr == NIL(CellInst)) {
MacPrintFatalErr("Mac_PutNameOfCellInst() failed: no cell instance is given");
    }
#endif DEBUG

    if (Mac_IsMember(CELL_INSTANCE_HT, cellInstName)) {
        sprintf(bear_ErrorMsg, "Cannot have two cell instances named \"%s\"",
            cellInstName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    } 
    cellInstPtr->name = cellInstName;
    Mac_Insert(CELL_INSTANCE_HT, cellInstName, (char *) cellInstPtr);

    return (BEAR_OK);
}



void
Mac_MarkCell(cellInstPtr)
CellInst    *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_MarkCell() failed: no cell is given");
    }
#endif DEBUG

    cellInstPtr->mark = BEAR_MARKED;
}

void
Mac_UnmarkCell(cellInstPtr)
CellInst *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_UnmarkCell() failed: no cell is given");
    }
#endif DEBUG

    cellInstPtr->mark = BEAR_UNMARKED;
}



bool
Mac_IsCellMarked(cellInstPtr)
CellInst *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_IsCellMarked() failed: no cell instance is given");
    }
#endif DEBUG

    return (cellInstPtr->mark == BEAR_MARKED);
}

bool
Mac_IsFixedCell(cellInstPtr)
CellInst *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_IsFixedCell() failed: no cell instance is given");
    }
#endif DEBUG

    return (cellInstPtr->masterPtr->shapeFlag == MAC_FIXED_SHAPE);
}



Point *
Mac_GetOriginOfCell(cellInstPtr)
CellInst    *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) {
        MacPrintFatalErr("Mac_GetOriginOfCell() failed: no cell is given");
    }
#endif DEBUG

    return (&cellInstPtr->origin);
}

void
Mac_PutOriginOfCell(originPtr, cellInstPtr)
Point       *originPtr;
CellInst    *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) {
        MacPrintFatalErr("Mac_PutOriginOfCell() failed: no cell is given");
    }
#endif DEBUG

    Mac_CopyPoint(&cellInstPtr->origin, originPtr);
}



Rect *
Mac_GetBBoxOfCell(cellInstPtr)  
CellInst    *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
    MacPrintFatalErr("Mac_GetBBoxOfCell() failed: no cell is given");
    }
    else if (cellInstPtr->masterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_GetBBoxOfCell() failed: the cell is not connected to a master");
    }
#endif DEBUG

    return (&cellInstPtr->masterPtr->bbox);
}



void
Mac_PutBBoxOfCell(bboxPtr, cellInstPtr)  
Rect        *bboxPtr;
CellInst    *cellInstPtr;
{
#ifdef DEBUG
    if (bboxPtr == NIL(Rect))
    {
        MacPrintFatalErr("Mac_PutBBoxOfCell() failed: no BBox is given");
    }
    else if (cellInstPtr == NIL(CellInst))
    {
    MacPrintFatalErr("Mac_PutBBoxOfCell() failed: no cell instance is given");
    }
    else if (cellInstPtr->masterPtr == NIL(CellMaster)) 
    {
MacPrintFatalErr("Mac_PutBBoxOfCell() failed: the cell is not connected to a master");
    }
#endif DEBUG

    Mac_CopyRect(&cellInstPtr->masterPtr->bbox, bboxPtr);
}



CellInst *
Mac_GetInstOfChipMaster(chipMasterPtr)
CellMaster *chipMasterPtr;
{
#ifdef DEBUG
    if (chipMasterPtr == NIL(CellMaster))
    {
        MacPrintFatalErr("Mac_GetInstOfChipMaster() failed: no chip cell is given");
    }
#endif DEBUG

    return (chipMasterPtr->instList);
}



CellInst *
Mac_GetInstOfCoreMaster(coreMasterPtr)
CellMaster *coreMasterPtr;
{
#ifdef DEBUG
    if (coreMasterPtr == NIL(CellMaster))
    {
        MacPrintFatalErr("Mac_GetInstOfCoreMaster() failed: no core cell is given");
    }
#endif DEBUG

    return (coreMasterPtr->instList);
}



Rect *
Mac_GetBBoxOfCellMaster(cellMasterPtr)  
CellMaster *cellMasterPtr;
{
#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))
    {
        MacPrintFatalErr("Mac_GetBBoxOfCellMaster() failed: no cell is given");
    }
#endif DEBUG

    return (&cellMasterPtr->bbox);
}



void
Mac_PutBBoxOfCellMaster(bboxPtr, cellMasterPtr)  
Rect        *bboxPtr;
CellMaster  *cellMasterPtr;
{
#ifdef DEBUG
    if (bboxPtr == NIL(Rect))
    {
        MacPrintFatalErr("Mac_PutBBoxOfCellMaster() failed: no BBox is given");
    }
    else if (cellMasterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_PutBBoxOfCellMaster() failed: no cell master is given");
    }
#endif DEBUG

    Mac_CopyRect(&cellMasterPtr->bbox, bboxPtr);
}



OrientVector
Mac_GetOrientConstrOfCell(cellInstPtr)
CellInst  *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_GetOrientConstrOfCell() failed: no cell is given");
    }
#endif DEBUG

    return (cellInstPtr->orientConstr);
}

void
Mac_PutOrientConstrOfCell(orientConstr, cellInstPtr)
OrientVector orientConstr;
CellInst     *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_GetOrientConstrOfCell() failed: no cell is given");
    }
#endif DEBUG

    cellInstPtr->orientConstr = orientConstr;
}


LinkedPolygon *
Mac_GetTranslConstrOfCell(cellInstPtr)
CellInst    *cellInstPtr;
{
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_GetTranslConstrOfCell() failed: no cell instance is given");
    }
    return (cellInstPtr->translConstr);
}

void
Mac_PutTranslConstrOfCell(translConstr, cellInstPtr)
LinkedPolygon   *translConstr;
CellInst        *cellInstPtr;
{
#ifdef DEBUG
    if (translConstr == NIL(LinkedPolygon))
    {
        MacPrintFatalErr("Mac_PutTranslConstrOfCell() failed: no translConstr is given");
    }
    if (cellInstPtr == NIL(CellInst))
    {
        MacPrintFatalErr("Mac_PutTranslConstrOfCell() failed: no cell is given");
    }
#endif DEBUG

    cellInstPtr->translConstr = translConstr;
}


LinkedLineSeg *
Mac_GetBoundaryOfCell(cellInstPtr)
CellInst  *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) 
    {
        MacPrintFatalErr("Mac_GetBoundaryOfCell() failed: no cell is given");
    }
#endif DEBUG

    return (cellInstPtr->boundary);
}

void
Mac_PutBoundaryOfCell(boundary, cellInstPtr)
LinkedLineSeg   *boundary;
CellInst        *cellInstPtr;
{
#ifdef DEBUG
    if (boundary == NIL(LinkedLineSeg)) 
    {
        MacPrintFatalErr("Mac_PutBoundaryOfCell() failed: no boundary is given");
    }
    if (cellInstPtr == NIL(CellInst)) 
    {
        MacPrintFatalErr("Mac_PutBoundaryOfCell() failed: no cell is given");
    }
#endif DEBUG

    cellInstPtr->boundary = boundary;
}


static BearStatus
MoveBoundary(lineSegPtr, transPtr)
register LineSeg *lineSegPtr;
Transform *transPtr;
{
    Point tmpPoint;

    GeoTransPoint(transPtr, &lineSegPtr->line.point1, &tmpPoint);
    Mac_CopyPoint(&lineSegPtr->line.point1, &tmpPoint);
    GeoTransPoint(transPtr, &lineSegPtr->line.point2, &tmpPoint);
    Mac_CopyPoint(&lineSegPtr->line.point2, &tmpPoint);

    return (BEAR_CONTINUE_SEARCH);
}


BearStatus
Mac_MoveBoundaryOfCell(transPtr, cellInstPtr)
Transform       *transPtr;
CellInst        *cellInstPtr;
{
#ifdef DEBUG
    if (transPtr == NIL(Transform))  {
        MacPrintFatalErr("Mac_MoveBoundaryOfCell() failed: no transform is given");
    }
    else if (cellInstPtr == NIL(CellInst)) {
        MacPrintFatalErr("Mac_MoveBoundaryOfCell() failed: no cell is given");
    }
    else if (cellInstPtr->boundary == NIL(LinkedLineSeg)) {
      MacPrintFatalErr("Mac_MoveBoundaryOfCell() failed: no boundary in the cell");
    }
#endif DEBUG

    (void) Mac_ItrLineSeg(cellInstPtr->boundary, MoveBoundary,
        (ClientData) transPtr);

    return (BEAR_OK);
}



CellGroup *
Mac_GetGroupOfCell(cellInstPtr)  
CellInst    *cellInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst))
    {
    MacPrintFatalErr("Mac_GetGoupOfCell() failed: no cell is given");
    }
#endif DEBUG

    return (cellInstPtr->cellGroupPtr);
}



RouteCell *
Mac_GetRCellListOfCell(cellMasterPtr)
CellMaster  *cellMasterPtr;
{
#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster)) {
MacPrintFatalErr("Mac_GetRCellListOfCell() failed: no cell master is given");
    }
#endif DEBUG

    return (cellMasterPtr->contents.routeCellList);
}



Net *
Mac_GetNetListOfCell(cellMasterPtr)
CellMaster *cellMasterPtr;
{
#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster)) {
    MacPrintFatalErr("Mac_GetNetListOfCell() failed: no cell master given");
    }
#endif DEBUG

    return (cellMasterPtr->contents.netList);
}



PlaneData *
Mac_GetCellPlaneOfCell(cellMasterPtr)
CellMaster *cellMasterPtr;
{
#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_GetCellPlaneOfCell() failed : no cell master given");
    }
#endif DEBUG

    return (cellMasterPtr->contents.planeData[MAC_CELL_PLANE]);
}

void
Mac_PutCellPlaneOfCell(planeDataPtr, cellMasterPtr)
PlaneData   *planeDataPtr;
CellMaster  *cellMasterPtr;
{
#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_PutCellPlaneOfCell() failed : no cell master given");
    }
#endif DEBUG

    cellMasterPtr->contents.planeData[MAC_CELL_PLANE] = planeDataPtr;
}



bool
Mac_GetPlacedFlagOfCell(cellMasterPtr)
CellMaster *cellMasterPtr;
{
#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster)) {
   MacPrintFatalErr("Mac_GetPlacedFlagOfCell() failed : no cell master given");
    }
#endif DEBUG

    return (cellMasterPtr->contents.placedFlag ? TRUE : FALSE);
}



void
Mac_PutPlacedFlagOfCell(placedFlag, cellMasterPtr)
bool        placedFlag;
CellMaster  *cellMasterPtr;
{
#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster)) {
   MacPrintFatalErr("Mac_PutPlacedFlagOfCell() failed : no cell master given");
    }
#endif DEBUG

    cellMasterPtr->contents.placedFlag = placedFlag;
}



WallJunction *
Mac_GetFpgOfCell(cellMasterPtr)
CellMaster *cellMasterPtr;
/*
 * Returns a pointer to the lower left wall junction of the
 * floor plan graph.
 */
{
    FloorPlanGraph  *fpgPtr;
    PlaneData       *planeDataPtr;

#ifdef DEBUG 
    if (cellMasterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_GetFpgOfCell() failed: no cell master is given");
    }
#endif DEBUG

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    fpgPtr = Tp_GetFpg(planeDataPtr);
    return ((fpgPtr != NIL(FloorPlanGraph)) ?
        Fpg_GetLLJunctOfFpg(fpgPtr) : NIL(WallJunction));
}



CellMaster *
Mac_GetMasterOfCell(cellInstPtr)
CellInst *cellInstPtr;
{
#ifdef DEBUG 
    if (cellInstPtr == NIL(CellInst))  
    {
        MacPrintFatalErr("Mac_GetMasterOfCell() failed: no cell instance is given");
    }
    if (cellInstPtr->masterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_GetMasterOfCell() failed: the cell instance is not connected to a master");
    }
#endif DEBUG

    return (cellInstPtr->masterPtr);
}


CellMaster *
Mac_GetParentOfCell(cellInstPtr)
CellInst *cellInstPtr;
{
#ifdef DEBUG 
    if (cellInstPtr == NIL(CellInst))  {
      MacPrintFatalErr("Mac_GetParentOfCell() failed : no cell instance is given");
    }
    if (cellInstPtr->parentPtr == NIL(CellMaster)) {
MacPrintFatalErr("Mac_GetParentOfCell() failed : gadzooks this cell instance is an orphan !");
    }
#endif DEBUG

    return (cellInstPtr->parentPtr);
}



CellMaster *
Mac_GetCellMasterByName(cellMastName)
BearString cellMastName;
{
    return ((CellMaster *) Mac_LookUp(CELL_MASTER_HT, cellMastName));
}



CellInst *
Mac_GetCellInstByName(cellInstName)
BearString cellInstName;
{
    return ((CellInst *) Mac_LookUp(CELL_INSTANCE_HT, cellInstName));
}



bool 
Mac_IsNumOfCellsZero(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_IsNumOfCellsZero() failed : no cell master given");
    }
#endif DEBUG

    return ((cellMasterPtr->contents.cellInstList == NIL(CellInst)) ?
        TRUE : FALSE);
}



/* ARGSUSED */
static BearStatus
MacCountNumOfCells(cellInstPtr, numPtr)
CellInst    *cellInstPtr;
int         *numPtr;
{
    (*numPtr)++;

    return (BEAR_CONTINUE_SEARCH);
}


BearStatus 
Mac_NumOfCells(parentCellPtr, numPtr) 
CellMaster  *parentCellPtr;
int         *numPtr;
{
#ifdef DEBUG
    if (parentCellPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_NumOfCells() failed: no cell master is given");
    }
#endif DEBUG

    *numPtr = 0;
    (void) Mac_ItrChildrenOfCell(parentCellPtr, MacCountNumOfCells,
        (ClientData) numPtr);

    return (BEAR_OK);
}



void
Mac_ResizeCell(deltaX, deltaY, cellInstPtr)
int         deltaX, deltaY;
CellInst    *cellInstPtr;
{
    LinkedLineSeg   *boundary; 
    LinkedLineSeg   *alineSeg;
    CellInst        *curCellInstPtr;
    Rect            bbox;
    Rect            *rectPtr;
    MovePinData     movePinData;
    int             bndDeltaX, bndDeltaY;
    int             moveX, moveY;

    for (curCellInstPtr = cellInstPtr->masterPtr->instList;
        curCellInstPtr != NIL(CellInst);
        curCellInstPtr = curCellInstPtr->nextPtr) {
    switch (WhatTransform(&curCellInstPtr->transform)) {
        case 1: /* GeoIdentityTransform */
        
        bndDeltaX = deltaX;
        bndDeltaY = deltaY;
        moveX = 0;
        moveY = 0;
        
                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; 
                    alineSeg != NIL(LinkedLineSeg);
                    alineSeg = alineSeg->nextPtr) {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);

            break;

        case 2: /* GeoUpsideDownTransform */
    
        bndDeltaX = deltaX;
        bndDeltaY = deltaY;
        moveX = 0;
        moveY = deltaY;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(moveX, -moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);

        break;
        case 3: /* GeoSidewaysTransform */
    
        bndDeltaX = deltaX;
        bndDeltaY = deltaY;
        moveX = deltaX;
        moveY = 0;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(-moveX, moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);

        break;
        case 4: /* Geo90Transform */

        bndDeltaX = deltaY;
        bndDeltaY = deltaX;
        moveX = 0;
        moveY = deltaY;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(moveX, -moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);

        break;
        case 5: /* Geo180Transform */
    
        bndDeltaX = deltaX;
        bndDeltaY = deltaY;
        moveX = deltaX;
        moveY = deltaY;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(-moveX, -moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);

        break;
        case 6: /* Geo270Transform */
    
        bndDeltaX = deltaY;
        bndDeltaY = deltaX;
        moveX = deltaX;
        moveY = 0;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(-moveX, -moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);

        break;
        case 7: /* GeoRef45Transform */
    
        bndDeltaX = deltaY;
        bndDeltaY = deltaX;
        moveX = 0;
        moveY = 0;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(moveX, moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);
        break;
        case 8: /* GeoRef135Transform */
    
        bndDeltaX = deltaY;
        bndDeltaY = deltaX;
        moveX = deltaX;
        moveY = deltaY;

                boundary = Mac_GetBoundaryOfCell(curCellInstPtr);
                bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
                bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
                for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
            {
                    if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
                {
                        bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
                {
                        bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
                {
                        bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
                {
                        bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
                    }
                }
                for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
                     alineSeg = alineSeg->nextPtr) 
                {
                    if (alineSeg->lineSeg->line.point1.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point1.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point1.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point1.p_y += deltaY;
                    }
                    if (alineSeg->lineSeg->line.point2.p_x == bbox.r_xtop) 
                {
                        alineSeg->lineSeg->line.point2.p_x += deltaX;
                    }
                    if (alineSeg->lineSeg->line.point2.p_y == bbox.r_ytop) 
                {
                        alineSeg->lineSeg->line.point2.p_y += deltaY;
                    }
                }
            Mac_PutBoundaryOfCell(boundary, curCellInstPtr);
            (void) MoveResizedBoundary(-moveX, -moveY, curCellInstPtr);
                (void) Mac_MoveCell(moveX, moveY, curCellInstPtr);

            break;
        default:
            break;
        } /* switch */
    }

    movePinData.vector.p_x = bndDeltaX;
    movePinData.vector.p_y = bndDeltaY;
    movePinData.bboxPtr = Mac_GetBBoxOfCell(cellInstPtr);

    movePinData.xRatio =
        (double) (movePinData.bboxPtr->r_xtop - movePinData.bboxPtr->r_xbot
        + movePinData.vector.p_x)
        / (double) (movePinData.bboxPtr->r_xtop - movePinData.bboxPtr->r_xbot);
    movePinData.yRatio =
        (double) (movePinData.bboxPtr->r_ytop - movePinData.bboxPtr->r_ybot
        + movePinData.vector.p_y)
        / (double) (movePinData.bboxPtr->r_ytop - movePinData.bboxPtr->r_ybot);
    (void) Mac_ItrPinOfCellInst(cellInstPtr, MovePin,
        (ClientData) &movePinData);
        
    rectPtr = movePinData.bboxPtr;
    rectPtr->r_xtop += bndDeltaX;
    rectPtr->r_ytop += bndDeltaY;
    Mac_PutBBoxOfCell(rectPtr, cellInstPtr);

    cellInstPtr->masterPtr->horzRectList->r_r.r_xtop += bndDeltaX;
    cellInstPtr->masterPtr->horzRectList->r_r.r_ytop += bndDeltaY;
    cellInstPtr->masterPtr->vertRectList->r_r.r_xtop += bndDeltaX;
    cellInstPtr->masterPtr->vertRectList->r_r.r_ytop += bndDeltaY; 
}



static BearStatus
MovePin(pinPtr, movePinDataPtr)
PinInst *pinPtr;
MovePinData  *movePinDataPtr;
{
    Point *pinPosPtr;

/*
    if ((movePinDataPtr->bboxPtr->r_xtop 
             - movePinDataPtr->bboxPtr->r_xbot) == pinPosPtr->p_x) {
        pinPosPtr->p_y *= movePinDataPtr->yRatio;
        pinPosPtr->p_x += movePinDataPtr->vector.p_x;
    } else if ((movePinDataPtr->bboxPtr->r_ytop 
             - movePinDataPtr->bboxPtr->r_ybot) == pinPosPtr->p_y) {
        pinPosPtr->p_x *= movePinDataPtr->xRatio;
        pinPosPtr->p_y += movePinDataPtr->vector.p_y;
    } else if (pinPosPtr->p_x == movePinDataPtr->bboxPtr->r_xbot) {
        pinPosPtr->p_y *= movePinDataPtr->yRatio;
    } else if (pinPosPtr->p_y == movePinDataPtr->bboxPtr->r_ybot) {
        pinPosPtr->p_x *= movePinDataPtr->xRatio;
    }
fprintf(stderr,"%d,%d -> ",pinPosPtr->p_x,pinPosPtr->p_y);
fprintf(stderr,"%d,%d: v=%d,%d r=%f,%f\n",pinPosPtr->p_x,pinPosPtr->p_y,
    movePinDataPtr->vector.p_x,movePinDataPtr->vector.p_y,
    movePinDataPtr->xRatio,movePinDataPtr->yRatio);
*/
    pinPosPtr = Mac_GetRelPosOfPin(pinPtr);
    if (pinPosPtr->p_x == movePinDataPtr->bboxPtr->r_xbot) {
        pinPosPtr->p_y = (pinPosPtr->p_y - movePinDataPtr->bboxPtr->r_ybot)
            * movePinDataPtr->yRatio + movePinDataPtr->bboxPtr->r_ybot;
    } else if (pinPosPtr->p_y == movePinDataPtr->bboxPtr->r_ybot) {
        pinPosPtr->p_x = (pinPosPtr->p_x - movePinDataPtr->bboxPtr->r_xbot)
            * movePinDataPtr->xRatio + movePinDataPtr->bboxPtr->r_xbot;
    } else if (pinPosPtr->p_x == movePinDataPtr->bboxPtr->r_xtop) {
        pinPosPtr->p_y = (pinPosPtr->p_y - movePinDataPtr->bboxPtr->r_ybot)
            * movePinDataPtr->yRatio + movePinDataPtr->bboxPtr->r_ybot;
        pinPosPtr->p_x += movePinDataPtr->vector.p_x;
    } else if (pinPosPtr->p_y == movePinDataPtr->bboxPtr->r_ytop) {
        pinPosPtr->p_x = (pinPosPtr->p_x - movePinDataPtr->bboxPtr->r_xbot)
            * movePinDataPtr->xRatio + movePinDataPtr->bboxPtr->r_xbot;
        pinPosPtr->p_y += movePinDataPtr->vector.p_y;
    }
    Mac_PutRelPosOfPin(pinPosPtr, pinPtr);
    return (BEAR_CONTINUE_SEARCH);
}



void
Mac_MirrorCell(XorY, cellInstPtr)

int         XorY;
CellInst    *cellInstPtr;

{
    Rect *rectPtr;
    Point pt1, pt2;
    Transform trans1, trans2;
    Transform tTrans, tTrans1, tTrans2;
    int x, y;

    rectPtr = Mac_GetBBoxOfCell(cellInstPtr);
    pt1.p_x = (rectPtr->r_xtop + rectPtr->r_xbot) / 2;
    pt1.p_y = (rectPtr->r_ytop + rectPtr->r_ybot) / 2;
    GeoTransPoint(&cellInstPtr->transform,&pt1,&pt2);
    x = pt2.p_x;
    y = pt2.p_y;
    GeoTranslateTrans(&GeoIdentityTransform,-x,-y,&tTrans1);
    GeoTranslateTrans(&cellInstPtr->transform,-x,-y,&trans1);
    if (XorY == 0) {
        GeoTransTrans(&tTrans1,&GeoSidewaysTransform,&tTrans2);
        GeoTransTrans(&trans1,&GeoSidewaysTransform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr,PinMirrorXSwitch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSegMirrorXSwitch, (ClientData) NULL);
    } else {
        GeoTransTrans(&tTrans1,&GeoUpsideDownTransform,&tTrans2);
        GeoTransTrans(&trans1,&GeoUpsideDownTransform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr,PinMirrorYSwitch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSegMirrorYSwitch, (ClientData) NULL);
    }
    GeoTranslateTrans(&tTrans2, x, y, &tTrans);
    GeoTranslateTrans(&trans2, x, y, &cellInstPtr->transform);
/*
    pt1.p_x = cellInstPtr->origin.p_x;
    pt1.p_y = cellInstPtr->origin.p_y;
    GeoTransPoint(&tTrans,&pt1,&cellInstPtr->origin);
*/
    (void) Mac_MoveBoundaryOfCell(&tTrans, cellInstPtr);
}



void
Mac_RotateCell(rot90or180or270, cellInstPtr)

int         rot90or180or270;
CellInst    *cellInstPtr;

{
    Rect *rectPtr;
    Point pt1, pt2;
    Transform trans1, trans2;
    Transform tTrans, tTrans1, tTrans2;
    int x, y;

    rectPtr = Mac_GetBBoxOfCell(cellInstPtr);
    pt1.p_x = (rectPtr->r_xtop + rectPtr->r_xbot) / 2;
    pt1.p_y = (rectPtr->r_ytop + rectPtr->r_ybot) / 2;
    GeoTransPoint(&cellInstPtr->transform,&pt1,&pt2);
    x = pt2.p_x;
    y = pt2.p_y;
    GeoTranslateTrans(&GeoIdentityTransform,-x,-y,&tTrans1);
    GeoTranslateTrans(&cellInstPtr->transform,-x,-y,&trans1);
    if (rot90or180or270 == 90) {
        GeoTransTrans(&tTrans1,&Geo90Transform,&tTrans2);
        GeoTransTrans(&trans1,&Geo90Transform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin90Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSeg90Switch, (ClientData) NULL);
    } else if (rot90or180or270 == 180) {
        GeoTransTrans(&tTrans1,&Geo180Transform,&tTrans2);
        GeoTransTrans(&trans1,&Geo180Transform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin180Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSeg180Switch, (ClientData) NULL);
    } else if (rot90or180or270 == 270) {
        GeoTransTrans(&tTrans1,&Geo270Transform,&tTrans2);
        GeoTransTrans(&trans1,&Geo270Transform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin270Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSeg270Switch, (ClientData) NULL);
    }
    GeoTranslateTrans(&tTrans2, x, y, &tTrans);
    GeoTranslateTrans(&trans2, x, y, &cellInstPtr->transform);
/*
    pt1.p_x = cellInstPtr->origin.p_x;
    pt1.p_y = cellInstPtr->origin.p_y;
    GeoTransPoint(&tTrans,&pt1,&cellInstPtr->origin);
*/
    (void) Mac_MoveBoundaryOfCell(&tTrans, cellInstPtr);
}


void
Mac_RotateCellAboutOrigin(rot90or180or270, cellInstPtr)

int         rot90or180or270;
CellInst    *cellInstPtr;

{
    Transform    trans;
    Transform    invTrans;
    Transform    trans1;

    GeoTransTrans(&GeoIdentityTransform, &cellInstPtr->transform, &trans);
    GeoInvertTrans(&cellInstPtr->transform, &invTrans);
    switch (rot90or180or270)
    {
    case 90:  GeoTransTrans(&Geo90Transform, &trans, &trans1); 
          break;
    case 180: GeoTransTrans(&Geo180Transform, &trans, &trans1); 
          break;
    case 270: GeoTransTrans(&Geo270Transform, &trans, &trans1); 
          break;
    }
    Mac_CopyTransform(&cellInstPtr->transform, &trans1);
    switch (rot90or180or270)
    {
    case 90:
        (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin90Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr), 
            LineSeg90Switch, (ClientData) NULL);
        break;
    case 180:
        (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin180Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr), 
            LineSeg180Switch, (ClientData) NULL);
        break;
    case 270:
        (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin270Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr), 
            LineSeg270Switch, (ClientData) NULL);
        break;
    }
    (void) Mac_MoveBoundaryOfCell(&invTrans, cellInstPtr);
    (void) Mac_MoveBoundaryOfCell(&trans1, cellInstPtr);
}



void
Mac_ReflectCell(ref45or135, cellInstPtr)

int         ref45or135;
CellInst    *cellInstPtr;

{
    Rect *rectPtr;
    Point pt1, pt2;
    Transform trans1, trans2;
    Transform tTrans, tTrans1, tTrans2;
    int x, y;

    rectPtr = Mac_GetBBoxOfCell(cellInstPtr);
    pt1.p_x = (rectPtr->r_xtop + rectPtr->r_xbot) / 2;
    pt1.p_y = (rectPtr->r_ytop + rectPtr->r_ybot) / 2;
    GeoTransPoint(&cellInstPtr->transform,&pt1,&pt2);
    x = pt2.p_x;
    y = pt2.p_y;
    GeoTranslateTrans(&GeoIdentityTransform,-x,-y,&tTrans1);
    GeoTranslateTrans(&cellInstPtr->transform,-x,-y,&trans1);
    if (ref45or135 == 45) {
        GeoTransTrans(&tTrans1,&GeoRef45Transform,&tTrans2);
        GeoTransTrans(&trans1,&GeoRef45Transform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, PinRef45Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSegRef45Switch, (ClientData) NULL);
    } else if (ref45or135 == 135) {
        GeoTransTrans(&tTrans1,&GeoRef135Transform,&tTrans2);
        GeoTransTrans(&trans1,&GeoRef135Transform,&trans2);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, PinRef135Switch,
            (ClientData) NULL);
        (void) Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr),
            LineSegRef135Switch, (ClientData) NULL);
    }
    GeoTranslateTrans(&tTrans2, x, y, &tTrans);
    GeoTranslateTrans(&trans2, x, y, &cellInstPtr->transform);
/*
    pt1.p_x = cellInstPtr->origin.p_x;
    pt1.p_y = cellInstPtr->origin.p_y;
    GeoTransPoint(&tTrans,&pt1,&cellInstPtr->origin);
*/
    (void) Mac_MoveBoundaryOfCell(&tTrans, cellInstPtr);
}



void
Mac_TransPinDirOfCell(trans, cellInstPtr)

Transform   trans;
CellInst    *cellInstPtr;

{
    if (trans.t_a == 0) {
        if (trans.t_d < 0) {
            if (trans.t_b < 0) {
                (void) Mac_ItrPinOfCellInst(cellInstPtr, PinRef135Switch,
                    (ClientData) NULL);
            } else {
                (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin90Switch,
                    (ClientData) NULL);
            }
        } else {
            if (trans.t_b < 0) {
                (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin270Switch,
                    (ClientData) NULL);
            } else {
                (void) Mac_ItrPinOfCellInst(cellInstPtr, PinRef45Switch,
                    (ClientData) NULL);
            }
        }
    } else if (trans.t_a < 0) {
        if (trans.t_e < 0) {
            (void) Mac_ItrPinOfCellInst(cellInstPtr, Pin180Switch,
                (ClientData) NULL);
        } else {
            (void) Mac_ItrPinOfCellInst(cellInstPtr, PinMirrorXSwitch,
                (ClientData) NULL);
        }
    } else {
        if (trans.t_e < 0) {
            (void) Mac_ItrPinOfCellInst(cellInstPtr, PinMirrorYSwitch,
                (ClientData) NULL);
        }
    }
}



/* ARGSUSED */
static BearStatus
PinMirrorXSwitch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case EAST_PIN:
        Mac_PutDirectionOfPin(WEST_PIN, pinPtr);
        break;
    case WEST_PIN:
        Mac_PutDirectionOfPin(EAST_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
PinMirrorYSwitch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        Mac_PutDirectionOfPin(SOUTH_PIN, pinPtr);
        break;
    case SOUTH_PIN:
        Mac_PutDirectionOfPin(NORTH_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
Pin90Switch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        Mac_PutDirectionOfPin(EAST_PIN, pinPtr);
        break;
    case SOUTH_PIN:
        Mac_PutDirectionOfPin(WEST_PIN, pinPtr);
        break;
    case EAST_PIN:
        Mac_PutDirectionOfPin(SOUTH_PIN, pinPtr);
        break;
    case WEST_PIN:
        Mac_PutDirectionOfPin(NORTH_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
Pin180Switch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        Mac_PutDirectionOfPin(SOUTH_PIN, pinPtr);
        break;
    case SOUTH_PIN:
        Mac_PutDirectionOfPin(NORTH_PIN, pinPtr);
        break;
    case EAST_PIN:
        Mac_PutDirectionOfPin(WEST_PIN, pinPtr);
        break;
    case WEST_PIN:
        Mac_PutDirectionOfPin(EAST_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
Pin270Switch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        Mac_PutDirectionOfPin(WEST_PIN, pinPtr);
        break;
    case SOUTH_PIN:
        Mac_PutDirectionOfPin(EAST_PIN, pinPtr);
        break;
    case EAST_PIN:
        Mac_PutDirectionOfPin(NORTH_PIN, pinPtr);
        break;
    case WEST_PIN:
        Mac_PutDirectionOfPin(SOUTH_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
PinRef45Switch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        Mac_PutDirectionOfPin(EAST_PIN, pinPtr);
        break;
    case SOUTH_PIN:
        Mac_PutDirectionOfPin(WEST_PIN, pinPtr);
        break;
    case EAST_PIN:
        Mac_PutDirectionOfPin(NORTH_PIN, pinPtr);
        break;
    case WEST_PIN:
        Mac_PutDirectionOfPin(SOUTH_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
PinRef135Switch(pinPtr, clientData)

PinInst     *pinPtr;
ClientData  clientData;

{
    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        Mac_PutDirectionOfPin(WEST_PIN, pinPtr);
        break;
    case SOUTH_PIN:
        Mac_PutDirectionOfPin(EAST_PIN, pinPtr);
        break;
    case EAST_PIN:
        Mac_PutDirectionOfPin(SOUTH_PIN, pinPtr);
        break;
    case WEST_PIN:
        Mac_PutDirectionOfPin(NORTH_PIN, pinPtr);
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSeg90Switch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_EAST;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_WEST;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_NORTH;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSeg180Switch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_NORTH;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_WEST;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_EAST;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSeg270Switch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_WEST;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_EAST;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_NORTH;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSegRef45Switch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_EAST;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_WEST;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_NORTH;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSegRef135Switch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_WEST;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_EAST;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_NORTH;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSegMirrorXSwitch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_NORTH;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_WEST;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_EAST;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
static BearStatus
LineSegMirrorYSwitch(lineSegPtr, clientData)

LineSeg     *lineSegPtr;
ClientData  clientData;

{
    switch (lineSegPtr->dir) {
    case GEO_NORTH:
        lineSegPtr->dir = GEO_SOUTH;
        break;
    case GEO_SOUTH:
        lineSegPtr->dir = GEO_NORTH;
        break;
    case GEO_EAST:
        lineSegPtr->dir = GEO_EAST;
        break;
    case GEO_WEST:
        lineSegPtr->dir = GEO_WEST;
        break;
    default:
        break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



static int
WhatTransform(transform)
Transform *transform;
{
    if (transform->t_a == 1 && transform->t_b == 0 && 
    transform->t_d == 0 && transform->t_e == 1 )
    {
    return (1); /* GeoIdentityTransform */
    }

    if (transform->t_a == 1 && transform->t_b == 0 && 
    transform->t_d == 0 && transform->t_e == -1 )
    {
    return (2); /* GeoUpsideDownTransform */
    }

    if (transform->t_a == -1 && transform->t_b == 0 && 
    transform->t_d == 0 && transform->t_e == 1 )
    {
    return (3); /* GeoSidewaysTransform */
    }

    if (transform->t_a == 0 && transform->t_b == 1 && 
    transform->t_d == -1 && transform->t_e == 0 )
    {
    return (4); /* Geo90Transform */
    }

    if (transform->t_a == -1 && transform->t_b == 0 && 
    transform->t_d == 0 && transform->t_e == -1 )
    {
    return (5); /* Geo180Transform */
    }

    if (transform->t_a == 0 && transform->t_b == -1 && 
    transform->t_d == 1 && transform->t_e == 0 )
    {
    return (6); /* Geo270Transform */
    }

    if (transform->t_a == 0 && transform->t_b == 1 &&
    transform->t_d == 1 && transform->t_e == 0 )
    {
    return (7); /* GeoRef45Transform */
    }

    if (transform->t_a == 0 && transform->t_b == -1 && 
    transform->t_d == -1 && transform->t_e == 0 )
    {
    return (8); /* GeoRef135Transform */
    }
    return (0);
}


static BearStatus
MoveResizedBoundary(moveDeltaX, moveDeltaY, cellInstPtr)
int      moveDeltaX;
int      moveDeltaY;
CellInst *cellInstPtr;
{
    LinkedLineSeg   *boundary;
    LinkedLineSeg   *alineSeg;

    boundary = Mac_GetBoundaryOfCell(cellInstPtr);
    for (alineSeg = boundary; alineSeg != NIL(LinkedLineSeg);
         alineSeg = alineSeg->nextPtr) 
    {
        alineSeg->lineSeg->line.point1.p_x += moveDeltaX;
        alineSeg->lineSeg->line.point1.p_y += moveDeltaY;
        alineSeg->lineSeg->line.point2.p_x += moveDeltaX;
        alineSeg->lineSeg->line.point2.p_y += moveDeltaY;
    }
    Mac_PutBoundaryOfCell(boundary, cellInstPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetAbsRectOfCellInst() --
 *
 * Results:
 *    BEAR_OK is returned if everything is spiffy.
 *    BEAR_ERROR is returned if the given cell instance is not a rectangle.
 *    The Rect structure supplied by the user is filled in.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_GetAbsRectOfCellInst(cellInstPtr, absRectPtr) 
CellInst    *cellInstPtr;  /* IN       */
Rect        *absRectPtr;   /* OUT      */
{
    GenericCell gcell;
    LinkedRect  *horzRectList;
    Transform   *horzTransPtr;

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;

    horzRectList = Mac_GetHorzRectListOfGCell(&gcell);
    if (horzRectList->r_next != NIL(LinkedRect)) {
        (void) sprintf (bear_ErrorMsg,
         "Mac_GetAbsRectOfCellInst() : cell instance %s is not a rectangle.\n",
            Mac_GetNameOfCellInst(cellInstPtr));
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    }
    horzTransPtr = Mac_GetTransformOfGCell(&gcell);
    GeoTransRect(horzTransPtr, &horzRectList->r_r, absRectPtr);

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetCurrentShape() --
 *
 * Results:
 *    BEAR_OK is returned if everything is spiffy.
 *    BEAR_ERROR is returned if the given cell instance is not a rectangle.
 *    The Shape structure supplied by the user is filled in.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_GetCurrentShape(cellInstPtr, shapePtr) 
CellInst    *cellInstPtr;  /* IN       */
Shape       *shapePtr;     /* OUT      */
{
    Rect absRect;

    if (Mac_GetAbsRectOfCellInst(cellInstPtr, &absRect) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    shapePtr->horzDim = absRect.r_xtop - absRect.r_xbot;
    shapePtr->vertDim = absRect.r_ytop - absRect.r_ybot;

    return (BEAR_OK);
}


