
/* macGCell.c --
 *
 *    Wei-Ming Dai
 *    (Advising Professor - E. S. Kuh)
 *
 *    Basic operations related to generic cells.
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macGCell.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"

/* ---------- Local Procedures ---------- */

GenericCell      *MakeGCell();

/* 
 * ----------------------------------------------------------------------------
 *
 * MakeGCell() --
 *
 *    Allocates memory for a new generic cell, and initializes
 *    the fields.
 *
 * Results:
 *    a generic cell.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static GenericCell *
MakeGCell()
{
    GenericCell *gCellPtr;

    MALLOC(GenericCell *, gCellPtr, sizeof (GenericCell));
    gCellPtr->unType = CELL_INST;
    gCellPtr->unValue.cellInstPtr = NIL(CellInst);

    return (gCellPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_CreateGCell() --
 *
 *    Create a generic cell of specified cell class. 
 *
 * Results:
 *    Newly created generic cell.
 *    BEAR_OK, if successful; BEAR_ERROR, otherwise.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_CreateGCell(cellClass, cellPtr, gcellPtrPtr)
CellClass   cellClass;
ClientData  cellPtr;
GenericCell **gcellPtrPtr;
{
    GenericCell *gcellPtr;

#ifdef DEBUG
    if (cellPtr == (ClientData) NULL) 
    {
        MacPrintFatalErr("Mac_CreateGCell() failed: no cell is given");
    }
#endif DEBUG

    gcellPtr = MakeGCell();
    switch (cellClass) 
    {
    case CELL_INST:
        gcellPtr->unType = CELL_INST;
        gcellPtr->unValue.cellInstPtr = (CellInst *) cellPtr;
        break;

    case DUMMY_CELL:
        gcellPtr->unType = DUMMY_CELL;
        gcellPtr->unValue.dummyCellPtr = (DummyCell *) cellPtr;
        break;

    case ROUTE_CELL:
        gcellPtr->unType = ROUTE_CELL;
        gcellPtr->unValue.routeCellPtr = (RouteCell *) cellPtr;
        break;

    default:
        FREE((char *) gcellPtr);
        Bear_PrintErr("Mac_CreateGCell() failed: unknown cell class");
        return (BEAR_ERROR);
    }
    *gcellPtrPtr = gcellPtr;
    return (BEAR_OK);
}


