
/* MacGCellPrim.c --
 *
 *     June Wong
 *     Wei-Ming Dai
 *     (Advising Professor - E. S. Kuh)
 *
 * Accessors, modifier, and presicates related to generic cells.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macGCellPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


LinkedRect *
Mac_GetHorzRectListOfGCell(gcellPtr)  
GenericCell *gcellPtr;
{
    CellInst *cellInstPtr;

#ifdef DEBUG
    if (gcellPtr == NIL(GenericCell)) {
MacPrintFatalErr("Mac_GetHorzRectListOfGCell() failed: no generic cell given");
    }
#endif DEBUG

    switch (gcellPtr->unType) {
    case CELL_INST:
        cellInstPtr = gcellPtr->unValue.cellInstPtr;
        if (cellInstPtr->masterPtr == NIL(CellMaster)) {
MacPrintFatalErr("Mac_GetHorzRectListOfGCell() failed: cell instance is not connected to a master");
        }
        return (cellInstPtr->masterPtr->horzRectList);
    case DUMMY_CELL:
        return (gcellPtr->unValue.dummyCellPtr->horzRectList);
    case ROUTE_CELL:
        return (gcellPtr->unValue.routeCellPtr->horzRectList);
    default:
       MacPrintFatalErr("Mac_GetHorzRectListOfGCell() failed: unknown cell class");
        break;
    }
    /* NOTREACHED */
}



void
Mac_PutHorzRectListOfGCell(horzRectList, gcellPtr)  
LinkedRect  *horzRectList;
GenericCell *gcellPtr;
{
    CellInst *cellInstPtr;

#ifdef DEBUG
    if (horzRectList == NIL(LinkedRect)) {
    MacPrintFatalErr("Mac_PutHorzRectListOfGCell() failed: no rectlist given");
    }
    else if (gcellPtr == NIL(GenericCell)) {
    MacPrintFatalErr("Mac_PutHorzRectListOfGCell() failed: no generic cell given");
    }
#endif DEBUG

    switch (gcellPtr->unType) {
    case CELL_INST:
        cellInstPtr = gcellPtr->unValue.cellInstPtr;
        if (cellInstPtr->masterPtr == NIL(CellMaster)) 
        {
        MacPrintFatalErr("Mac_PutHorzRectListOfGCell() failed: cell instance is not connected to a master");
        }
        cellInstPtr->masterPtr->horzRectList = horzRectList;
        break;
    case DUMMY_CELL:
        gcellPtr->unValue.dummyCellPtr->horzRectList = horzRectList;
        break;
    case ROUTE_CELL:
        gcellPtr->unValue.routeCellPtr->horzRectList = horzRectList;
        break;
    default:
        MacPrintFatalErr("Mac_PutHorzRectListOfGCell() failed: unknown cell class");
        break;
    }
}



LinkedRect *
Mac_GetVertRectListOfGCell(gcellPtr)  
GenericCell *gcellPtr;
{
    CellInst *cellInstPtr;

#ifdef DEBUG
    if (gcellPtr == NIL(GenericCell)) {
MacPrintFatalErr("Mac_GetVertRectListOfGCell() failed: no generic cell given");
    }
#endif DEBUG

    switch (gcellPtr->unType) {
    case CELL_INST:
        cellInstPtr = gcellPtr->unValue.cellInstPtr;
        if (cellInstPtr->masterPtr == NIL(CellMaster)) 
        {
MacPrintFatalErr("Mac_GetVertRectListOfGCell() failed: cell instance is not connected to a master");
        }
        return (cellInstPtr->masterPtr->vertRectList);
    case DUMMY_CELL:
        return (gcellPtr->unValue.dummyCellPtr->vertRectList);
    case ROUTE_CELL:
        return (gcellPtr->unValue.routeCellPtr->vertRectList);
    default:
       MacPrintFatalErr("Mac_GetVertRectListOfGCell() failed: unknown cell class");
        break;
    }
    /* NOTREACHED */
}



void
Mac_PutVertRectListOfGCell(vertRectList, gcellPtr)  
LinkedRect  *vertRectList;
GenericCell *gcellPtr;
{
    CellInst *cellInstPtr;

#ifdef DEBUG
    if (vertRectList == NIL(LinkedRect)) {
    MacPrintFatalErr("Mac_PutVertRectListOfGCell() failed: no rectlist given");
    }
    else if (gcellPtr == NIL(GenericCell)) {
    MacPrintFatalErr("Mac_PutVertRectListOfGCell() failed: no generic cell given");
    }
#endif DEBUG

    switch (gcellPtr->unType) {
    case CELL_INST:
        cellInstPtr = gcellPtr->unValue.cellInstPtr;
        if (cellInstPtr->masterPtr == NIL(CellMaster)) 
        {
    MacPrintFatalErr("Mac_PutVertRectListOfGCell() failed: cell instance is not connected to a master");
        }
        cellInstPtr->masterPtr->vertRectList = vertRectList;
        break;
    case DUMMY_CELL:
        gcellPtr->unValue.dummyCellPtr->vertRectList = vertRectList;
        break;
    case ROUTE_CELL:
        gcellPtr->unValue.routeCellPtr->vertRectList = vertRectList;
        break;
    default:
        MacPrintFatalErr("Mac_PutVertRectListOfGCell() failed: unknown cell class");
        break;
    }
}



Transform *
Mac_GetTransformOfGCell(gcellPtr)
GenericCell *gcellPtr;
{
#ifdef DEBUG
    if (gcellPtr == NIL(GenericCell)) {
       MacPrintFatalErr("Mac_GetTransformOfGCell() failed: no generic cell given");
    }
#endif DEBUG

    switch (gcellPtr->unType) {
    case CELL_INST:
        return (&(gcellPtr->unValue.cellInstPtr->transform));
    case DUMMY_CELL:
        return (&(gcellPtr->unValue.dummyCellPtr->transform));
    case ROUTE_CELL:
        return (&(gcellPtr->unValue.routeCellPtr->transform));
    default:
        MacPrintFatalErr("Mac_GetTransformOfGCell() failed: unknown cell class");
        break;
    }
    /* NOTREACHED */
}



void
Mac_PutTransformOfGCell(transPtr, gcellPtr)
Transform   *transPtr; 
GenericCell *gcellPtr;
{
#ifdef DEBUG
    if (transPtr == NIL(Transform))  {
       MacPrintFatalErr("Mac_PutTransformOfGCell() failed: no transform is given");
    }
    else if (gcellPtr == NIL(GenericCell)) {
       MacPrintFatalErr("Mac_PutTransformOfGCell() failed: no generic cell given");
    }
#endif DEBUG

    switch (gcellPtr->unType) {
    case CELL_INST:
        Mac_CopyTransform(&(gcellPtr->unValue.cellInstPtr->transform),
            transPtr);
        break;
    case DUMMY_CELL:
        Mac_CopyTransform(&(gcellPtr->unValue.dummyCellPtr->transform),
            transPtr);
        break;
    case ROUTE_CELL:
        Mac_CopyTransform(&(gcellPtr->unValue.routeCellPtr->transform),
            transPtr);
        break;
    default:
        MacPrintFatalErr("Mac_PutTransformOfGCell() failed: unknown cell class");
        break;
    }
}



BearId
Mac_GetIdOfGCell(gcellPtr)
GenericCell *gcellPtr;
{
    switch (gcellPtr->unType) {
    case CELL_INST:
        return (gcellPtr->unValue.cellInstPtr->id);
    case DUMMY_CELL:
        return (gcellPtr->unValue.dummyCellPtr->id);
    case ROUTE_CELL:
        return (gcellPtr->unValue.routeCellPtr->id);
    default:
        MacPrintFatalErr("Mac_GetIdOfGCell() failed : unknown cell class");
        break;
    }
    /* NOTREACHED */
}



void
Mac_PutIdOfGCell(bearId, gcellPtr)
BearId      bearId;
GenericCell *gcellPtr;
{
    switch (gcellPtr->unType) {
    case CELL_INST:
        gcellPtr->unValue.cellInstPtr->id = bearId;
        break;
    case DUMMY_CELL:
        gcellPtr->unValue.dummyCellPtr->id = bearId;
        break;
    case ROUTE_CELL:
        gcellPtr->unValue.routeCellPtr->id = bearId;
        break;
    default:
        MacPrintFatalErr("Mac_PutIdOfGCell() failed : unknown cell class");
        break;
    }
}



