
/* macGraph.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macGraph.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint


#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "list.h"
#include "mac.h"
#include "macInt.h"


static BearStatus   DeleteGraphOfNet();
static void         FreeEdgeData();


void
Mac_FreeGraphData(generic)
BearGeneric generic;
{
    if (generic != (BearGeneric) NULL) {
        Mac_GraphData *graphDataPtr = (Mac_GraphData *) generic;

        if (graphDataPtr->bnTileList != (lsList) NULL) {
            lsDestroy((lsList) graphDataPtr->bnTileList, (void (*)()) NULL);
        }
        if (graphDataPtr->junctList != (lsList) NULL) {
            lsDestroy((lsList) graphDataPtr->junctList, (void (*)()) NULL);
        }
        FREE((char *) graphDataPtr); /* Free Mac_GraphData */
    }
}


void
Mac_FreeNodeData(generic)
BearGeneric generic;
{
    if (generic != (BearGeneric) NULL) {
        PinOrJunct *pinOrJunctPtr = (PinOrJunct *) generic;

        if (pinOrJunctPtr->unType == FPG_PSEUDO_JCT) {
            FREE((char *) pinOrJunctPtr->unValue.pjunctPtr);
        }
        FREE((char *) pinOrJunctPtr); /* Free PinOrJunct */
    }
}


static void
FreeEdgeData(generic)
BearGeneric generic;
{
    if (generic != (BearGeneric) NULL) {
        FREE((char *) generic);
    }
}


/* ARGSUSED */
static BearStatus
DeleteGraphOfNet(netPtr, clientData)
Net         *netPtr;
ClientData  clientData;
{
    Gt_Graph graph;

#ifdef USE_DR_CODE
    Pgr_FreeWire(&netPtr->wirePtr);
#endif /* USE_DR_CODE */
    Mac_DeleteAutoGraphOfNet(netPtr);
    Mac_DeleteUserGraphOfNet(netPtr);

    return (BEAR_CONTINUE_SEARCH);
}


void
Mac_DeleteAutoGraphOfNet(netPtr)
Net *netPtr;
{
    Gt_Graph graph;

#ifdef USE_DR_CODE
    Pgr_FreeWire(&netPtr->wirePtr);
#endif /* USE_DR_CODE */
    if ((graph = Mac_GetAutoGraphOfNet(netPtr)) != (Gt_Graph) NULL) {
        Mac_DeleteGraph(graph);
        Mac_PutAutoGraphOfNet((Gt_Graph) NULL, netPtr);
    }
}


void
Mac_DeleteUserGraphOfNet(netPtr)
Net *netPtr;
{
    Gt_Graph graph;

    if ((graph = Mac_GetUserGraphOfNet(netPtr)) != (Gt_Graph) NULL) {
        Mac_DeleteGraph(graph);
        Mac_PutUserGraphOfNet((Gt_Graph) NULL, netPtr);
    }
}


void
Mac_DeleteGraph(graph)
Gt_Graph graph;
{
    Gt_DeleteGraph(graph, Mac_FreeGraphData, Mac_FreeNodeData, FreeEdgeData);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteAllGraphs() --
 *
 * Go through the given net list and delete each nets user generated graph,
 * as well as its automatically generated graph.
 *
 * Results:
 *   None
 *
 * Side effects:
 *   All of the graphs are deleted.
 * ----------------------------------------------------------------------------
 */

void
Mac_DeleteAllGraphs(netList)
Net *netList;
{
    (void) Mac_ItrNet(netList, DeleteGraphOfNet, (ClientData) NULL);
}



