
/* macNet.h --
 *
 *     Wei-ming Dai
 *     June Wang
 *     (Professor E. S. Kuh)
 *
 *     It defines the data structures related to Net and all 
 *     the subroutines to create, insert, and delete.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macNet.h,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
 */

#ifndef _MAC_NET
#define _MAC_NET

typedef enum {
    SIGNAL_NET,                     /* Signal net                           */
    CLOCK_NET,                      /* Clock net                           */
    POWER_NET,                      /* Power net                            */
    GROUND_NET,                     /* Ground net                           */
    DUMMY_NET                       /* Dummy net                            */
} NetKind;

typedef struct _Net {
    BearString  name;               /* Net name                             */
    BearId      id;                 /* Net identification number            */
    int         maxLength;          /* Net maximum length                   */
    int         minLength;          /* Net minimum length                   */
    int         criticality;        /* Net criticality: may be
                                     * positive or negative;
                                     * more positive, more critial;
                                     */
    NetKind     kind;               /* Net Kind                             */
    LayerVector layer;              /* Net layer constraints                */
    PortInst    *portList;          /* The list of port instances           */
    struct _LinkedRPin *routePinList; /* Head of the routing pin list       */
    bool        storedInTiPlane;    /* TRUE if the net has been stored in the
                                     * tile plane.
                                     */
    Gt_Graph    autoGraph;          /* Automatically generated graph        */
    Gt_Graph    userGraph;          /* User generated graph                 */
/*
 * A graph is dependant on the following things:
 *     1.) Real pins that are part of a net.
 *     2.) Wall junctions that are part of the floor plan graph. (user
 *         graph only)
 *
 * The following things are dependant on a graph:
 *     1.) Pseudo pins which are stored in the tile plane if they belong
 *         to a manually routed net.
 *     2.) The wire structure that contains the power ground information.
 */

    struct _Pgr_Wire *wirePtr;      /* Pointer containing the power ground
                                     * information.
                                     */
    root             *routeRootList;/* List of the BBL routing root
                                     * of the net.
                                     */
    CellMaster       *cellMasterPtr;/* The cell master within which the net
                                     * is defined.
                                     */
   
    struct _Net      *nextPtr;      /* The next net                         */
    int              numOfPin;
    ClientData       clientData;    /* This space for rent                  */
} Net;

typedef struct _LinkedNet {
    Net               net;          /* A port instance                      */
    struct _LinkedNet *nextPtr;     /* Pointer to another linked net        */
} LinkedNet;

typedef struct _Mac_GraphData {
    BearId          fpgId;
    bool            pgFlag;         /* TRUE if this is a graph of a power or
                                     * ground net.
                                     */
    bool            edgeDataFlag;   /* TRUE if the each edge's data field
                                     * is filled in.
                                     */
    lsList          bnTileList;     /* This list is used by the placement
                                     * code.
                                     */
    lsList          junctList;      /* This list is used by the placement
                                     * code.
                                     */
} Mac_GraphData;

typedef struct _Mac_EdgeData {
    float       width;
    BottleNeck  *bnPtr;
    PseudoPin   *ppin1Ptr; 
    PseudoPin   *ppin2Ptr;
} Mac_EdgeData;


extern BearStatus Mac_CreateNet(/* cellMasterPtr, netName, netPtrPtr */);
/*
 * CellMaster   *cellMasterPtr;
 * BearString   netName;
 * Net          **netPtrPtr;
 */

extern BearStatus Mac_DeleteNet(/* netPtr */);
/*
 * Net      *netPtr;
 */

extern BearStatus Mac_InsertRPinOfNet(/* netPtr, rPinPtr */);
/*
 * Net       *netPtr;
 * RoutePin  *rPinPtr;
 */

extern BearStatus Mac_DeleteRPinOfNet(/* netPtr, rPinPtr */);
/*
 * Net          *netPtr;
 * RoutePin     *rPinPtr;
 */

extern BearStatus Mac_InsertRRootOfNet(/* netPtr, rRootPtr */);
/*
 * Net      *netPtr;
 * root     *rRootPtr;
 */

extern BearStatus Mac_DeleteRRootOfNet(/* netPtr, rRootPtr */);
/*
 * Net      *netPtr;
 * root     *rRootPtr;
 */

extern BearStatus Mac_CheckUserGraph(/* netPtr, loopFlagPtr, connFlagPtr */);
/*
 * Net      *netPtr;                            (* IN       *)
 * bool     *loopFlagPtr;                       (* OUT      *)
 * bool     *connFlagPtr;                       (* OUT      *)
 */

extern BearStatus Mac_CheckAutoGraph(/* netPtr, loopFlagPtr, connFlagPtr */);
/*
 * Net      *netPtr;                            (* IN       *)
 * bool     *loopFlagPtr;                       (* OUT      *)
 * bool     *connFlagPtr;                       (* OUT      *)
 */

extern Mac_EdgeData *Mac_MakeEdgeData();

extern Mac_GraphData *Mac_MakeGraphData();

extern void Mac_DeleteGraph(/* graph */);
/*
 * Gt_Graph graph;                              (* IN       *)
 */

extern void Mac_DeleteAutoGraphOfNet(/* netPtr */);
/*
 * Net      *netPtr;                            (* IN       *)
 */

extern void Mac_DeleteUserGraphOfNet(/* netPtr */);
/*
 * Net      *netPtr;                            (* IN       *)
 */

extern void Mac_DeleteAllGraphs(/* netList */);
/*
 * Net     *netList;                            (* IN       *)
 */

extern void Mac_FreeGraphData(/* generic */);
/*
 * BearGeneric generic;
 */

extern void Mac_FreeNodeData(/* generic */);
/*
 * BearGeneric generic;
 */

#endif _MAC_NET


