
/* macNetItr.c --
 *
 *     George Carvalho, Esq.
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains all the iterators related to Net.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macNetItr.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrNet() --
 *
 * Takes a list of nets, and iterates every net in the list.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrNet(netList, func, clientData)
Net         *netList;
BearStatus  (*func)();
ClientData  clientData;
{
    Net *netPtr;

#ifdef DEBUG
    if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrNet() failed: no function is given");
    }
#endif DEBUG

    if (netList == NIL(Net)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (netPtr = netList;
            netPtr != NIL(Net);
            netPtr = netPtr->nextPtr) {

            if ((*func)(netPtr, clientData) == BEAR_STOP_SEARCH) {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH); 
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPortOfNet() --
 *
 * Iterates the unrouted port instances connected to the given net.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrPortOfNet(netPtr, func, clientData)
Net         *netPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortInst *portPtr;

#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_ItrPortOfNet() failed: no net is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrPortOfNet() failed: no function is given");
    }
#endif DEBUG

    if (netPtr->portList == NIL(PortInst)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (portPtr = netPtr->portList;
            portPtr != NIL(PortInst);
            portPtr = portPtr->nextNetPortPtr) {

            if (portPtr->mark == BEAR_UNMARKED) {
                if (((*func)(portPtr, clientData)) == BEAR_STOP_SEARCH) {
                    return (BEAR_STOP_SEARCH);
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPinOfNet() --
 *
 * The assigned pin instance of every port in the given net
 * is visited if it's unrouted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrPinOfNet(netPtr, func, clientData)
Net         *netPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortInst *portPtr;

#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_ItrPinOfNet() failed: no net is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrPinOfNet() failed: no function is given");
    }
#endif DEBUG

    if (netPtr->portList == NIL(PortInst)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (portPtr = netPtr->portList;
            portPtr != NIL(PortInst);
            portPtr = portPtr->nextNetPortPtr) {

            if (portPtr->mark == BEAR_UNMARKED) {
                if (((*func)(Mac_GetAssignedPinOfPort(portPtr), clientData)) ==
                    BEAR_STOP_SEARCH) {
                    return (BEAR_STOP_SEARCH);
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrAllPinOfNet() --
 *
 * The assigned pin instance of every port in the given net is visited.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrAllPinOfNet(netPtr, func, clientData)
Net         *netPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    register PortInst *portPtr;

#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_ItrAllPinOfNet() failed: no net is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrAllPinOfNet() failed: no function is given");
    }
#endif DEBUG

    if (netPtr->portList == NIL(PortInst)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (portPtr = netPtr->portList;
            portPtr != NIL(PortInst);
            portPtr = portPtr->nextNetPortPtr) {

            if (((*func)(Mac_GetAssignedPinOfPort(portPtr), clientData)) ==
                BEAR_STOP_SEARCH) {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrAllPortOfNet() --
 *
 * Iterates the routed and the unrouted port instances
 * connected to the given net.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrAllPortOfNet(netPtr, func, clientData)
Net         *netPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    PortInst *portPtr;

#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_ItrAllPortOfNet() failed: no net is given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrAllPortOfNet() failed: no function is given");
    }
#endif DEBUG

    if (netPtr->portList == NIL(PortInst)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (portPtr = netPtr->portList;
            portPtr != NIL(PortInst);
            portPtr = portPtr->nextNetPortPtr) {

            if (((*func)(portPtr, clientData)) == BEAR_STOP_SEARCH) {
                return (BEAR_STOP_SEARCH);
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRPinOfNet() --
 *
 * Iterates the unmarked route pins connected to the given net.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrRPinOfNet(netPtr, func, clientData)
Net         *netPtr;
BearStatus  (*func)();
ClientData  clientData;
{
    LinkedRPin *lnRPinPtr;

#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_ItrRPinOfNet() failed: no net given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrRPinOfNet() failed: no function given");
    }
#endif DEBUG

    if (netPtr->routePinList == NIL(LinkedRPin)) {
        return (BEAR_COMPLETE_SEARCH);
    }
    else {
        for (lnRPinPtr = netPtr->routePinList;
            lnRPinPtr != NIL(LinkedRPin);
            lnRPinPtr = lnRPinPtr->nextPtr) {

            if (lnRPinPtr->rPinPtr->mark == BEAR_UNMARKED) {
                if (((*func)(lnRPinPtr->rPinPtr, clientData)) == 
                    BEAR_STOP_SEARCH) {
                    return (BEAR_STOP_SEARCH);
                }
            }
        }
    }

    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRRootOfNet() --
 *
 * It iterates every routing root of the net.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrRRootOfNet(net, func, clientData)
Net        *net;
BearStatus (*func)();
ClientData clientData;

{
    root    *curRRootPtr;

#ifdef DEBUG
    if (net == NIL(Net))
    {
        MacPrintFatalErr("Mac_ItrRRootOfNet() failed: no net is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrRRootOfNet() failed: no function is given");
    }
#endif DEBUG

    if (net->routeRootList == NIL(root))
    {
        return (BEAR_COMPLETE_SEARCH);
    }
    else
    {
        curRRootPtr = net->routeRootList;
        while (curRRootPtr != NIL(root))
        {
        if (((*func)(curRRootPtr, clientData)) == BEAR_STOP_SEARCH)
        {
            return (BEAR_STOP_SEARCH);
        }
        curRRootPtr = curRRootPtr->nextInNet;
        }
    }
    return (BEAR_COMPLETE_SEARCH);
}



