
/* macNetPrim.c --
 *
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains all the operations used to retrieve and update the
 * fields of Net.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macNetPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Net Primitives ---------- */


BearString
Mac_GetNameOfNet(netPtr)
Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetNameOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->name);
}


BearStatus
Mac_PutNameOfNet(netName, netPtr)
BearString netName;
Net        *netPtr;
{
#ifdef DEBUG
    if (netName == (BearString) NULL) {
        MacPrintFatalErr("MacPutNameOfNet() failed : no name is given");
    }
    else if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_PutNameOfNet() failed : no net given");
    }
#endif DEBUG

    if (Mac_IsMember(NET_HT, netName)) {
        sprintf(bear_ErrorMsg, "Cannot have two nets named \"%s\"", netName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    }
    netPtr->name = netName;
    Mac_Insert(NET_HT, netName, (char *) netPtr);

    return (BEAR_OK);
}


BearId
Mac_GetIdOfNet(netPtr) 

Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetIdOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->id);
}
  

void
Mac_PutIdOfNet(id, netPtr)

BearId  id;
Net     *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_PutIdOfNet() failed : no net given");
    }
    else if (id == 0)
    {
        MacPrintFatalErr("Mac_PutIdOfNet() failed : id value cannot be 0");
    }
#endif DEBUG

    netPtr->id = id;
}
  

int
Mac_GetMaxLengthOfNet(netPtr)

Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetMaxLengthOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->maxLength);
}


void
Mac_PutMaxLengthOfNet(maxLength, netPtr)

int maxLength;
Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_PutMaxLengthOfNet() failed : no net given");
    }
#endif DEBUG

    netPtr->maxLength = maxLength;
}



int
Mac_GetMinLengthOfNet(netPtr)

Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetMinLengthOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->minLength);
}


void
Mac_PutMinLengthOfNet(minLength, netPtr)

int minLength;
Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_PutMinLengthOfNet() failed : no net given");
    }
#endif DEBUG

    netPtr->minLength = minLength;
}



int
Mac_GetCriticalityOfNet(netPtr)
Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_GetCriticalityOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->criticality);
}


void
Mac_PutCriticalityOfNet(criticality, netPtr)
int criticality;
Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_PutCriticalityOfNet() failed : no net given");
    }
#endif DEBUG

    netPtr->criticality = criticality;
}


NetKind 
Mac_GetKindOfNet(netPtr)

Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetKindOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->kind);
}


void
Mac_PutKindOfNet(kind, netPtr)
NetKind kind;
Net     *netPtr;

{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_PutKindOfNet() failed : no net given");
    }
#endif DEBUG

    netPtr->kind = kind;
}
  

LayerVector 
Mac_GetLayerOfNet(netPtr)
Net *netPtr;

{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetLayerOfNet() failed : no net given");
    }
#endif DEBUG

    return (netPtr->layer);
}
  

void
Mac_PutLayerOfNet(layer, netPtr)
LayerVector layer;
Net         *netPtr;

{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_PutLayerOfNet() failed : no net given");
    }
#endif DEBUG

    netPtr->layer = layer;
}



CellMaster *
Mac_GetCellOfNet(netPtr) 
Net *netPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net))
    {
        MacPrintFatalErr("Mac_GetCellOfNet() failed : no net is given");
    }
#endif DEBUG

    return (netPtr->cellMasterPtr);
}


  
bool 
Mac_IsNumOfNetsZero(cellMasterPtr) 
CellMaster *cellMasterPtr;
{
#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))
    {
MacPrintFatalErr("Mac_IsNumOfNetsZero() failed: no cell master is given");
    }
#endif DEBUG

    return (cellMasterPtr->contents.netList == NIL(Net));
}



/* ARGSUSED */
BearStatus
MacCountNumOfNets(netPtr, numPtr)
Net  *netPtr;
int  *numPtr;
{
    if (netPtr == NIL(Net)) {
        return (BEAR_STOP_SEARCH);
    }

    (*numPtr)++;

    return (BEAR_CONTINUE_SEARCH);
}



BearStatus 
Mac_NumOfNets(cellMasterPtr, numPtr)
CellMaster  *cellMasterPtr;
int         *numPtr;
{
#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_NumOfNets() failed: no cell master is given");
    }
#endif DEBUG

    *numPtr = 0;
    if (Mac_ItrNet(cellMasterPtr->contents.netList,
        MacCountNumOfNets, (ClientData) numPtr) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}



bool
Mac_GetStatusOfNetInTiPlane(netPtr)
Net *netPtr;
{
    return (netPtr->storedInTiPlane);
}

  

void
Mac_PutStatusOfNetInTiPlane(storedInTiPlane, netPtr)
bool    storedInTiPlane;
Net     *netPtr;
{
    netPtr->storedInTiPlane = storedInTiPlane;
}

  

Net *
Mac_GetNetByName(name)
BearString name;
{
    return ((Net *) Mac_LookUp(NET_HT, name));
}


bool
Mac_IsGraphValid(cellMasterPtr, graph)
CellMaster  *cellMasterPtr;
Gt_Graph    graph;
{
    FloorPlanGraph  *fpgPtr;
    Mac_GraphData   *graphDataPtr;
    PlaneData       *planeDataPtr;

    if ((planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr)) ==
        NIL(PlaneData)) {
        return (FALSE);
    }
    if ((fpgPtr = Tp_GetFpg(planeDataPtr)) == NIL(FloorPlanGraph)) {
        return (FALSE);
    }
    graphDataPtr = (Mac_GraphData *) Gt_GetGraphDataOfGraph(graph);

    return ((graphDataPtr->fpgId == Fpg_GetIdOfFpg(fpgPtr)) ? TRUE : FALSE);
}


Gt_Graph
Mac_GetAutoGraphOfNet(netPtr)
Net *netPtr;
{
    return (netPtr->autoGraph);
}

void 
Mac_PutAutoGraphOfNet(autoGraph, netPtr)
Gt_Graph    autoGraph;
Net         *netPtr;
{
    netPtr->autoGraph = autoGraph;
}

Gt_Graph
Mac_GetUserGraphOfNet(netPtr)
Net *netPtr;
{
    return (netPtr->userGraph);
}

void 
Mac_PutUserGraphOfNet(userGraph, netPtr)
Gt_Graph    userGraph;
Net         *netPtr;
{
    netPtr->userGraph = userGraph;
}

