
/* macPinItr.c --
 *
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the iterators related to pins.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macPinItr.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


typedef struct _UsersData UsersData;
struct _UsersData {
    BearStatus (*func)();
    ClientData clientData;
};


/* ---------- Local Procedures ---------- */
static BearStatus AddNumOfPinsInPortToTotal();
static BearStatus DoPin();


static BearStatus
AddNumOfPinsInPortToTotal(portPtr, totalCountPtr)
PortInst    *portPtr;
int         *totalCountPtr;
/*
 * Add the number of pins in this port to the total count.
 */
{
    int pinCount;

    (void) Mac_NumOfPinsOfPort(portPtr, &pinCount);
    *totalCountPtr = *totalCountPtr + pinCount;

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * DoPin() --
 *
 * Apply the user function to a pin attached to the node if any.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   Whatever the side effects the user function has.
 * ----------------------------------------------------------------------------
 */

static BearStatus
DoPin(node, userDataPtr)
Gt_GraphNode    node;
UsersData       *userDataPtr;
{
    PinOrJunct *pinOrJunctPtr;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    if (pinOrJunctPtr->unType == FPG_PIN) {
        if ((*(userDataPtr->func))(pinOrJunctPtr->unValue.pinPtr,
            userDataPtr->clientData) == BEAR_STOP_SEARCH) {
            return (BEAR_STOP_SEARCH);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrInstOfPin() --
 *
 * Given a pin master, it iterates all the pin instances connected to
 * that pin master.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, when successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus 
Mac_ItrInstOfPin(master, func, clientData)
PinMaster  *master;
BearStatus (*func)();
ClientData clientData;

{
    register PinInst *inst;

#ifdef DEBUG
    if (master == NIL(PinMaster))
    {
    MacPrintFatalErr("Mac_ItrInstOfPin() failed: no master is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrInstOfPin() failed: no function is given");
    }
#endif DEBUG

    if (master->instList == NIL(PinInst))
    {
    return (BEAR_COMPLETE_SEARCH);
    }
    else
    {
        inst = master->instList;
        while (inst != NIL(PinInst))
        {
        if (((*func)(inst, clientData)) == BEAR_STOP_SEARCH)
        {
            return (BEAR_STOP_SEARCH);
        }
        inst = inst->nextPtr;
        }
    }
    return (BEAR_COMPLETE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrPinOfGraph() --
 *
 * Visit each pin in the graph, and apply the given procedure to it. 
 * The procedure should be of the following form:
 *
 *  BearStatus
 *  func(pinPtr, clientData)
 *  PinInst     *pinPtr;
 *  ClientData  clientData;
 *  {
 *  }
 *
 * Func normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrPinOfGraph(graph, func, clientData)
Gt_Graph    graph;
BearStatus  (*func)();
ClientData  clientData;
{
    UsersData userData;

#ifdef DEBUG
    if (graph == (Gt_Graph) NULL) {
        MacPrintFatalErr("Mac_ItrPinOfGraph() failed: no graph given");
    }
    else if (func == (BearStatus (*)()) NULL) {
        MacPrintFatalErr("Mac_ItrPinOfGraph() failed: no function given");
    }
#endif DEBUG

    userData.func = func;
    userData.clientData = clientData;
    if (Gt_ItrNodesOfGraph(graph, DoPin, (ClientData) &userData) ==
        BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_COMPLETE_SEARCH);
}


BearStatus
Mac_NumOfPinsOfGraph(graph, numPtr)
Gt_Graph    graph;
int         *numPtr;
{
    int pinCount = 0;

    if (Mac_ItrPinOfGraph(graph, MacCountNumOfPins,
        (ClientData) &pinCount) == BEAR_STOP_SEARCH) {
        *numPtr = 0;
        return (BEAR_ERROR);
    }

    *numPtr = pinCount;
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_NumOfPinsInPortListOfNet() --
 *
 * Calculate the number of pins that are in this net's port list.
 *
 * Results:
 *   BEAR_OK is returned if everything went swell, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_NumOfPinsInPortListOfNet(netPtr, numPtr)
Net *netPtr;
int *numPtr;
{
    int pinCount;

    *numPtr = pinCount = 0;
    if (Mac_ItrAllPortOfNet(netPtr, AddNumOfPinsInPortToTotal,
        (ClientData) &pinCount) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }
    *numPtr = pinCount;

    return (BEAR_OK);
}



