 
/* macPort.c --
 *
 *      George Carvalho, Esq.
 *      Sangjin Hong
 *      June Wang
 *      (Advising Professor - E. S. Kuh)
 *
 * It contains all the operations to create, insert, and delete.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macPort.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Port Primitives ---------- */


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_MakePortMaster() --
 *
 * Allocates a space for a pointer to PortMaster.
 *
 * Results:
 *   The pointer to PortMaster is returned.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

PortMaster *
Mac_MakePortMaster()
{
    PortMaster *portMasterPtr;

    MALLOC(PortMaster *, portMasterPtr, sizeof (PortMaster));
    portMasterPtr->name = (BearString) NULL;
    portMasterPtr->direction = INPUT;
    portMasterPtr->kind = SIGNAL_PORT;
    portMasterPtr->instList = NIL(PortInst);
    portMasterPtr->cellMasterPtr = NIL(CellMaster);
    portMasterPtr->pinMasterList = NIL(PinMaster);
    portMasterPtr->nextPtr = NIL(PortMaster);
    portMasterPtr->clientData = (ClientData) NULL;

    return (portMasterPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_MakePortInst() --
 *
 * Allocates a space for a pointer to PortInst.
 *
 * Results:
 *   The pointer to PortInst is returned.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

PortInst *
Mac_MakePortInst()
{
    PortInst *portPtr;

    MALLOC(PortInst *, portPtr, sizeof (PortInst));
    portPtr->name = (BearString) NULL;
    portPtr->feedThrListPtr = NIL(LinkedLinkedPort);
    portPtr->eleEqvListPtr = NIL(LinkedLinkedPort);
    portPtr->masterPtr = NIL(PortMaster);
    portPtr->netPtr = NIL(Net);
    portPtr->intNetPtr = NIL(Net);
    portPtr->cellInstPtr = NIL(CellInst);
    portPtr->pinList = NIL(PinInst);
    portPtr->assignedPinPtr = NIL(PinInst);
    portPtr->nextCellPortPtr = NIL(PortInst);
    portPtr->nextNetPortPtr = NIL(PortInst);
    portPtr->nextPtr = NIL(PortInst);
    portPtr->mark = BEAR_UNMARKED;
    portPtr->clientData = (ClientData) NULL;

    return (portPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DuplicatePort() --
 *
 * Makes a new port instance and attach it to the cell instance and the
 * port master of the original port instances.
 *
 * Results:
 *   BEAR_OK, if successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DuplicatePort(cellInstPtr,origPortInstPtr, copyPortInstName, 
                  copyPortInstPtrPtr)
PortInst      *origPortInstPtr;
CellInst      *cellInstPtr;
BearString    copyPortInstName;
PortInst      **copyPortInstPtrPtr;

{
    PortMaster *portMaster;
    char       *stringPtr;

#ifdef DEBUG
    if (origPortInstPtr == NIL(PortInst))
    {
    MacPrintFatalErr("Mac_DuplicatePort() failed: no original port instance is given");
    }
#endif DEBUG

/* Make new port instance */

    *copyPortInstPtrPtr = Mac_MakePortInst();

/* Insert new port instance into its port master */

   portMaster = origPortInstPtr->masterPtr;
   if (Mac_InsertPortInstOfPortMaster(portMaster,*copyPortInstPtrPtr)
        == BEAR_ERROR)
   {
       Bear_PrintErr("Mac_InsertPortInstOfPortMaster() failed in Mac_DuplicatePort()\n");
       return (BEAR_ERROR);
   }


/* Attach port instance to its cell instances */

    if (Mac_InsertPortInstOfCellInst(cellInstPtr,*copyPortInstPtrPtr)
        == BEAR_ERROR)
    {
        Bear_PrintErr("Mac_DuplicatePort() failed in Mac_InsertPortInstOfCellInst\n");
        return (BEAR_ERROR);
    }

/* Put new port instance name */

    MALLOC(char *, stringPtr, (strlen(copyPortInstName) + 1) * sizeof (char));
    strcpy(stringPtr, copyPortInstName);

    if (Mac_PutNameOfPortInst(stringPtr,*copyPortInstPtrPtr) 
        == BEAR_ERROR)
    {
        Bear_PrintErr("Mac_DuplicatePort() failed in Mac_PutNameOfPortInst()\n");
        return (BEAR_ERROR);
    }



/* Copy port instance contents */


    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertPortMaster() --
 *
 * Takes the list of port masters from the given cell master, and
 * inserts the given port master at the beginning of that list.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertPortMaster(cellMasterPtr, portMasterPtr)
CellMaster  *cellMasterPtr;
PortMaster  *portMasterPtr;
{
#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster)) {
        MacPrintFatalErr("Mac_InsertPortMaster() failed: cellMaster is NIL");
    }
    else if (portMasterPtr == NIL(PortMaster)) {
        MacPrintFatalErr("Mac_InsertPortMaster() failed: portMaster is NIL");
    }  
#endif DEBUG

    portMasterPtr->nextPtr = cellMasterPtr->portMasterList;
    cellMasterPtr->portMasterList = portMasterPtr;
    portMasterPtr->cellMasterPtr = cellMasterPtr;

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeletePortMaster() --
 *
 * Deletes the given port master from the given cell master.  Also
 * deletes all the port instances and pin masters connected to 
 * the port master.
 * Finally, updates the hash table and frees the memory space.
 *
 * Results:
 *   BEAR_OK, if successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeletePortMaster(cellMasterPtr, portMasterPtr)
CellMaster  *cellMasterPtr;
PortMaster  *portMasterPtr;

{
    PortMaster  *curPortMaster;
    PortInst    *curPortInst;
    PinMaster   *curPinMaster;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))
    {
        MacPrintFatalErr("Mac_DeletePortMaster() failed: no cell master is given");
    }
    else if (cellMasterPtr->portMasterList == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_DeletePortMaster() failed: the given cell master has no port master");
    }
    else if (portMasterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_DeletePortMaster() failed: no port master is given");
    }
#endif DEBUG

    if (portMasterPtr == cellMasterPtr->portMasterList) {
        cellMasterPtr->portMasterList = portMasterPtr->nextPtr;
    }
    else {
        curPortMaster = cellMasterPtr->portMasterList;
        while (curPortMaster->nextPtr != NIL(PortMaster) &&
               curPortMaster->nextPtr != portMasterPtr) {
            curPortMaster = curPortMaster->nextPtr;
        }
        if (curPortMaster->nextPtr == portMasterPtr) {
            curPortMaster->nextPtr = curPortMaster->nextPtr->nextPtr;
        }
        else {
            return (BEAR_ERROR);
        }
    }
    curPortInst = portMasterPtr->instList;
    while (curPortInst != NIL(PortInst))
    {
        if (Mac_DeletePortInstOfPortMaster(portMasterPtr, curPortInst) == BEAR_ERROR)
        {
            Bear_PrintErr("Mac_DeletePortInstOfPortMaster() failed in Mac_DeletePortMaster()");
            return (BEAR_ERROR);
        }
        curPortInst = curPortInst->nextPtr;
    }

    curPinMaster = portMasterPtr->pinMasterList;
    while (curPinMaster != NIL(PinMaster))
    {
        if (Mac_DeletePinMaster(curPinMaster) == BEAR_ERROR) {
           Bear_PrintErr("Mac_DeletePinMaster() failed in Mac_DeletePortMaster()");
            return (BEAR_ERROR);
        }
        curPinMaster = curPinMaster->nextPtr;
    }
    Mac_Delete(PORT_MASTER_HT, portMasterPtr->name);
    FREE((char *) portMasterPtr);

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertPortInstOfCellInst() --
 *
 * Takes the list of port instances from the given cell instance, and
 * inserts the given port instance at the beginning of the list.
 *
 * Results:
 *   BEAR_OK, when done.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertPortInstOfCellInst(cellInstPtr, portInstPtr)
CellInst    *cellInstPtr;
PortInst    *portInstPtr;
{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) {
   MacPrintFatalErr("Mac_InsertPortInstOfCellInst() failed: cellInst is NIL");
    }
    else if (portInstPtr == NIL(PortInst)) {
     MacPrintFatalErr("Mac_InsertPortInstOfCellInst() failed: porInst is NIL");
    }  
#endif DEBUG

    portInstPtr->nextCellPortPtr = cellInstPtr->portInstList;
    cellInstPtr->portInstList = portInstPtr;
    portInstPtr->cellInstPtr = cellInstPtr;

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertPortInstOfNet() --
 *
 * Takes the list of port instances from the given net, and inserts
 * the given port instance at the beginning of the list.
 *
 * Results:
 *   BEAR_OK, when done.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertPortInstOfNet(netPtr, portPtr)
Net      *netPtr;
PortInst *portPtr;
{
#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_InsertPortInstOfNet() failed : net is NULL");
    }
    else if (portPtr == NIL(PortInst)) {
  MacPrintFatalErr("Mac_InsertPortInstOfNet() failed : port instance is NULL");
    }  
#endif DEBUG

    portPtr->netPtr = netPtr;

    portPtr->nextNetPortPtr = netPtr->portList;
    netPtr->portList = portPtr;

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertPortInstOfPortMaster() --
 *
 * Takes the list of port instances from the given port master and
 * inserts the given port instance at the beginning of the list.
 *
 * Results:
 *   BEAR_OK, when successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertPortInstOfPortMaster(portMasterPtr, portInstPtr)
PortMaster  *portMasterPtr;
PortInst    *portInstPtr;
{
#ifdef DEBUG
    if (portMasterPtr == NIL(PortMaster)) {
MacPrintFatalErr("Mac_InsertPortInstOfPortMaster() failed: no cell master given");
    }
    else if (portInstPtr == NIL(PortInst)) {
MacPrintFatalErr("Mac_InsertPortInstOfPortMaster() failed: no port inst given");
    }
#endif DEBUG

    portInstPtr->masterPtr = portMasterPtr;

    portInstPtr->nextPtr = portMasterPtr->instList;
    portMasterPtr->instList = portInstPtr; 

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeletePortInstOfPortMaster() --
 *
 * Deletes the given port instance from the given port master as well 
 * as the pin instances from the port instance.
 *
 * Results:
 *   BEAR_OK, if successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeletePortInstOfPortMaster(masterPtr, instPtr)
PortMaster  *masterPtr;
PortInst    *instPtr;
{
    PortInst    *curInstPtr;
    PinInst     *currentPinPtr;

#ifdef DEBUG
    if (instPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_DeletePortInstOfPortMaster() failed: no port instance is given");
    }
    else if (instPtr->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_DeletePortInstOfPortMaster() failed:  the given port instance is not connected to a master");
    }   
    else if (masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_DeletePortInstOfPortMaster() failed:  no master is given");
    }
    else if (masterPtr->instList == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_DeletePortInstOfPortMaster() failed:  the given master has no port instances");
    }
#endif DEBUG

    if (masterPtr->instList != instPtr)
    {
        curInstPtr = masterPtr->instList;
        while (curInstPtr->nextPtr != NIL(PortInst) && 
               curInstPtr->nextPtr != instPtr)
        {
            curInstPtr = curInstPtr->nextPtr;
        }
        if (curInstPtr->nextPtr == NIL(PortInst))
        {
            Bear_PrintErr("Mac_DeletePortInstOfPortMaster() failed: the given port instance is not connected to the given master");
            return (BEAR_ERROR);
        }
        else if (curInstPtr->nextPtr == instPtr)
        {
            curInstPtr->nextPtr = instPtr->nextPtr;
        }
        else
        {
            Bear_PrintErr("Mac_DeletePortInstOfPortMaster(): the given port instance can't be found");
            return (BEAR_ERROR);
        }
    }
    else
    {
        masterPtr->instList = masterPtr->instList->nextPtr;
    }
    currentPinPtr = instPtr->pinList;
    while (currentPinPtr != NIL(PinInst))
    {
        if (Mac_DeletePinInstOfPort(instPtr, currentPinPtr) ==
            BEAR_ERROR)
        {
            Bear_PrintErr("Mac_DeletePinInstOfPort() failed in MacDeletePortInstOfCell");
            return (BEAR_ERROR);
        }
        currentPinPtr = currentPinPtr->nextPtr;
    }
    Mac_Delete(PORT_INSTANCE_HT, instPtr->name);
    FREE((char *) instPtr);
    return (BEAR_OK);
}


        
/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeletePortInstOfCell() --
 *
 * Deletes the given port instance from its port master.
 * Since it is to delete the port instance from its cell link, the 
 * nextCellPortPtr links of the port instances adjacent to the deleted 
 * port instance have to be updated.
 * Also, the pin instances of the deleted port instance have to be
 * deleted.
 *
 * Results:
 *   BEAR_OK, if successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeletePortInstOfCell(portInstPtr)
PortInst   *portInstPtr;

{
    PortMaster  *masterPtr;
    PortInst    *currentInstPtr;
    PinInst     *currentPinPtr;

#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_DeletePortInstOfCell() failed: no port is given");
    }
    else if (portInstPtr->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_DeletePortInstOfCell() failed: the given port \
        is not connected to a master");
    }
    else if (portInstPtr->masterPtr->instList == NIL(PortInst))
    {
        Bear_PrintErr("Mac_DeletePortInstOfCell() failed: the port master has not instances");
        return (BEAR_ERROR);
    }
#endif DEBUG


    masterPtr = portInstPtr->masterPtr;
    currentInstPtr = masterPtr->instList;

    if (currentInstPtr == portInstPtr)
    {
        masterPtr->instList = portInstPtr->nextPtr;
    }
    else
    {
        while (currentInstPtr->nextPtr != NIL(PortInst) &&
               currentInstPtr->nextPtr != portInstPtr)
        {
            currentInstPtr = currentInstPtr->nextPtr;
        }
        if (currentInstPtr->nextPtr == portInstPtr)
        {
            currentInstPtr->nextPtr = portInstPtr->nextPtr;
        }
        else
        {
            Bear_PrintErr("The specified port instance can't be found\n");
            return (BEAR_ERROR);
        }
    }
    currentPinPtr = portInstPtr->pinList;
    while (currentPinPtr != NIL(PinInst))
    {
        if (Mac_DeletePinInstOfPort(portInstPtr, currentPinPtr) ==
            BEAR_ERROR)
        {
            Bear_PrintErr("Mac_DeletePinInstOfPort() failed in \
            Mac_DeletePortInstOfCell()");
            return (BEAR_ERROR);
        }

        if (Mac_DeletePinInstOfPinMaster(currentPinPtr,currentPinPtr->masterPtr)
            == BEAR_ERROR)
        {
            Bear_PrintErr("Mac_DeletePinInstOfPinMaster() failed in Mac_DeletePortInstOfCell()");
            return (BEAR_ERROR);
        }

        currentPinPtr = currentPinPtr->nextPtr;
     }
    Mac_Delete(PORT_INSTANCE_HT, portInstPtr->name);
    FREE((char *) portInstPtr);
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeletePortInstOfNet() --
 *
 * Deletes the given port instance from its net.  Since it is
 * to be deleted from its net link, we have to update the nextNetPortPtr
 * links of the port instances adjacent to the deleted port instance.
 *
 * Results:
 *   BEAR_OK, if successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeletePortInstOfNet(portPtr)
PortInst *portPtr;
{
    PortInst *curPortPtr;

#ifdef DEBUG
    if (portPtr == NIL(PortInst)) {
        MacPrintFatalErr("Mac_DeletePortInstOfNet() failed: no port is given");
    }
#endif DEBUG

    curPortPtr = portPtr->netPtr->portList;
    if (curPortPtr == portPtr) {
        portPtr->netPtr->portList = portPtr->nextNetPortPtr;
    }
    else {
        while ((curPortPtr->nextNetPortPtr != NIL(PortInst)) && 
            (curPortPtr->nextNetPortPtr != portPtr)) {
            curPortPtr = curPortPtr->nextNetPortPtr;
        }  
        if (curPortPtr->nextNetPortPtr == portPtr) {
            curPortPtr->nextNetPortPtr =
                curPortPtr->nextNetPortPtr->nextNetPortPtr;
        }
        else {
  MacPrintFatalErr("Mac_DeletePortInstOfNet() failed: port is not in the net");
        }
    }
    portPtr->netPtr = NIL(Net);
    portPtr->nextNetPortPtr = NIL(PortInst);

    return (BEAR_OK);
}


