 
/* macPortPrim.c --
 *
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the operations to retrieve and update the information
 * in ports.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macPortPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Port Primitives ---------- */


BearString
Mac_GetNameOfPortMaster(portMastPtr)
PortMaster *portMastPtr;
{
#ifdef DEBUG
    if (portMastPtr == NIL(PortMaster))
    {
   MacPrintFatalErr("Mac_GetNamdOfPortMaster() failed: no port master is given");
    }
#endif DEBUG

    return (portMastPtr->name);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutNameOfPortMaster() --
 *
 * Each port master must have a unique name.
 *
 * Results:
 *   BEAR_OK is returned if the name is not already in the hash table,
 *   BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The name of the port master is stored in the hash table.
 *
 * ----------------------------------------------------------------------------
 */


BearStatus
Mac_PutNameOfPortMaster(portMastName, portMastPtr)
BearString  portMastName;
PortMaster  *portMastPtr;
{
#ifdef DEBUG
    if (portMastPtr == NIL(PortMaster)) {
     MacPrintFatalErr("Mac_PutNameOfPortMaster() failed: no port master is given");
    }
#endif DEBUG

    if (Mac_IsMember(PORT_MASTER_HT, portMastName)) {
        sprintf(bear_ErrorMsg, "Cannot have two port masters named \"%s\"", 
            portMastName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    } 
    portMastPtr->name = portMastName;
    Mac_Insert(PORT_MASTER_HT, portMastName, (char *) portMastPtr);

    return (BEAR_OK);
}



BearString
Mac_GetNameOfPortInst(portInstPtr)
PortInst *portInstPtr;
{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))
    {
MacPrintFatalErr("Mac_GetNameOfPortInst() failed: no port instance is given");
    }
#endif DEBUG

    return (portInstPtr->name);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutNameOfPortInst() --
 *
 * Each port instance must have a unique name.
 *
 * Results:
 *   BEAR_OK is returned if the name is not already in the hash table,
 *   BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The name of the port instance is stored in the hash table.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_PutNameOfPortInst(portInstName, portInstPtr)
BearString   portInstName;
PortInst     *portInstPtr;
{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst)) {
     MacPrintFatalErr("Mac_PutNameOfPortInst() failed: no port instance is given");
    }
#endif DEBUG

    if (Mac_IsMember(PORT_INSTANCE_HT, portInstName)) {
        sprintf(bear_ErrorMsg, "Cannot have two port instances named \"%s\"", 
            portInstName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    } 
    portInstPtr->name = portInstName;
    Mac_Insert(PORT_INSTANCE_HT, portInstName, (char *) portInstPtr);

    return (BEAR_OK);
}



PortKind
Mac_GetKindOfPort(portInstPtr)
PortInst *portInstPtr;

{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))  
    {
       MacPrintFatalErr("Mac_GetKindOfPort() failed: no port instance is given");
    }
    if (portInstPtr->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetKindOfPort() failed: the port instance is not connected to a port master");
    }
#endif DEBUG

    return (portInstPtr->masterPtr->kind);
}


void
Mac_PutKindOfPort(kind, portInst)
PortKind kind;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
       MacPrintFatalErr("Mac_PutKindOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutKindOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    portInst->masterPtr->kind = kind;
}


PortDir
Mac_GetDirectionOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_GetDirectionOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetDirectionOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    return (portInst->masterPtr->direction);
}


void
Mac_PutDirectionOfPort(direction, portInst) 
PortDir  direction;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutDirectionOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutDirectionOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    portInst->masterPtr->direction = direction;
}


MinNomMax 
Mac_GetDCFanInLoadOfPort(portInst)
PortInst  *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_GetDCFanInLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetDCFanInLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.sigProp.DCFanInLoad);
}


void
Mac_PutDCFanInLoadOfPort(DCFanInLoad, portInst)
MinNomMax DCFanInLoad;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutDCFanInLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutDCFanInLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    portInst->masterPtr->PropUn.sigProp.DCFanInLoad = DCFanInLoad;
}


MinNomMax 
Mac_GetDCFanOutLoadOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_GetDCFanOutLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetDCFanOutLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.sigProp.DCFanOutLoad);
}


void
Mac_PutDCFanOutLoadOfPort(DCFanOutLoad, portInst)
MinNomMax DCFanOutLoad;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutDCFanOutLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutDCFanOutLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    portInst->masterPtr->PropUn.sigProp.DCFanOutLoad = DCFanOutLoad;
}


MinNomMax 
Mac_GetDCMaxFanInOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_GetDCMaxFanInLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetDCMaxFanInLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.sigProp.DCMaxFanIn);
}


void
Mac_PutDCMaxFanInOfPort(DCMaxFanIn, portInst)
MinNomMax DCMaxFanIn;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutDCMaxFanInLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutDCMaxFanInLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    portInst->masterPtr->PropUn.sigProp.DCMaxFanIn = DCMaxFanIn;
}


MinNomMax 
Mac_GetDCMaxFanOutOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_GetDCMaxFanOutLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetDCMaxFanOutLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.sigProp.DCMaxFanOut);
}


void
Mac_PutDCMaxFanOutOfPort(DCMaxFanOut, portInst)
MinNomMax DCMaxFanOut;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutDCMaxFanOutLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutDCMaxFanOutLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
#endif DEBUG

    portInst->masterPtr->PropUn.sigProp.DCMaxFanOut = DCMaxFanOut;
}


MinNomMax 
Mac_GetACLoadOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_GetACLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetACLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
    if (portInst->masterPtr->kind != (PortKind) SIGNAL_PORT) {
        MacPrintFatalErr("Mac_GetACLoadOfPort() failed: not a sig-port");
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.sigProp.ACLoad);
}


void
Mac_PutACLoadOfPort(ACLoad, portInst)
MinNomMax ACLoad;
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutACLoadOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutACLoadOfPort() failed: the port instance is not connected to a port master");
          
    }
    if (portInst->masterPtr->kind != (PortKind) SIGNAL_PORT) {
        MacPrintFatalErr("Mac_PutACLoadOfPort() failed: not a sig-port");
    }
#endif DEBUG

    portInst->masterPtr->PropUn.sigProp.ACLoad = ACLoad;
}


double
Mac_GetMaxVoltDropOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  {
        MacPrintFatalErr("Mac_GetMaxVoltDropOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster)) {
        MacPrintFatalErr("Mac_GetMaxVoltDropOfPort() failed: the port instance is not connected to a port master");
    }
    if (portInst->masterPtr->kind == SIGNAL_PORT) {
        MacPrintFatalErr("Mac_GetMaxVoltDropOfPort() failed: not a pg-port");
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.PGProp.maxVoltDrop);
}



void
Mac_PutMaxVoltDropOfPort(maxVoltDrop, portInst)
double  maxVoltDrop;
PortInst  *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
        MacPrintFatalErr("Mac_PutMaxVoltDropOfPort() failed: no port instance is given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutMaxVoltDropOfPort() failed: the port instance is not connected to a port master");
          
    }
    if (portInst->masterPtr->kind == SIGNAL_PORT)
    {
        MacPrintFatalErr("Mac_PutMaxVoltDropOfPort() failed: not a pg-port");
    }
#endif DEBUG

    portInst->masterPtr->PropUn.PGProp.maxVoltDrop = maxVoltDrop;
}


double
Mac_GetMaxCurrentReqOfPort(portInst)
PortInst *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
MacPrintFatalErr("Mac_GetMaxCurrentReqOfPort() failed: no port instance given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_GetMaxCurrentReqOfPort() failed: the port instance is not connected to a port master");
    }
    if (portInst->masterPtr->kind ==  SIGNAL_PORT)
    {
        MacPrintFatalErr("Mac_GetMaxCurrentReqOfPort() failed: not a pg-port");
    }
#endif DEBUG

    return (portInst->masterPtr->PropUn.PGProp.maxCurrentReq);
}


void
Mac_PutMaxCurrentReqOfPort(maxCurrentReq, portInst)
double      maxCurrentReq;
PortInst    *portInst;

{
#ifdef DEBUG
    if (portInst == NIL(PortInst))  
    {
MacPrintFatalErr("Mac_PutMaxCurrentReqOfPort() failed: no port instance given");
    }
    if (portInst->masterPtr == NIL(PortMaster))
    {
        MacPrintFatalErr("Mac_PutMaxCurrentReqOfPort() failed: the port instance is not connected to a port master");
          
    }
    if (portInst->masterPtr->kind == SIGNAL_PORT)
    {
        MacPrintFatalErr("Mac_PutMaxCurrentReqOfPort() failed: not a pg-port");
    }
#endif DEBUG

    portInst->masterPtr->PropUn.PGProp.maxCurrentReq = maxCurrentReq;
}


LinkedPort *
Mac_GetEleEqvOfPort(port)
PortInst *port;

{
#ifdef DEBUG
    if (port == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_GetEleEqvOfPort() failed: no port given");
    }
    if (port->eleEqvListPtr == NIL(LinkedLinkedPort))
    {
        MacPrintFatalErr("Mac_GetEleEqvOfPort() failed: no EleEqv list of the port");
    }
#endif DEBUG

    return (port->eleEqvListPtr->portList);       
}


LinkedPort * 
Mac_GetFeedThrOfPort(portInstPtr)
PortInst *portInstPtr;

{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_GetFeedThrOfPort() failled : no port given");
    }
    if (portInstPtr->feedThrListPtr == NIL(LinkedLinkedPort))
    {
        MacPrintFatalErr("Mac_GetFeedThrOfPort() failled: no feedThr list in the given port");
          
    }
#endif DEBUG

    return (portInstPtr->feedThrListPtr->portList);
}



PortMaster *
Mac_GetMasterOfPort(portInstPtr)
PortInst *portInstPtr;

{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))
    {
    MacPrintFatalErr("Mac_GetMasterOfPort() failed : no port given");
    }
#endif DEBUG

    return (portInstPtr->masterPtr);
}



CellInst *
Mac_GetCellInstOfPort(portInstPtr)
PortInst *portInstPtr;

{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_GetCellInstOfPort() failed : no port given");
    }
#endif DEBUG

    return (portInstPtr->cellInstPtr);
}



Net *
Mac_GetNetOfPort(portInstPtr)
PortInst *portInstPtr;

{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_GetNetOfPort() failed : no port given");
    }
#endif DEBUG

    return (portInstPtr->netPtr);
}



/* ARGSUSED */
BearStatus
MacCountNumOfPorts(portPtr, numPtr)
PortInst    *portPtr;
int         *numPtr;
{
    (*numPtr)++;

    return (BEAR_CONTINUE_SEARCH);
}



BearStatus
Mac_NumOfPortsOfCell(cellInstPtr, numPtr)
CellInst    *cellInstPtr;
int         *numPtr;

{
#ifdef DEBUG
    if (cellInstPtr == NIL(CellInst)) {
  MacPrintFatalErr("Mac_NumOfPortsOfCell() failed: no cell instance is given"); 
    }
#endif DEBUG
    *numPtr = 0;  
    return ((Mac_ItrPortOfCellInst(cellInstPtr, MacCountNumOfPorts, 
        (ClientData) numPtr) == BEAR_COMPLETE_SEARCH) ? BEAR_OK : BEAR_ERROR);
}



BearStatus
Mac_NumOfPortsOfNet(netPtr, numPtr)
Net  *netPtr;
int  *numPtr;
/*
 * Counts the routed ports as well as the unrouted ports of the net.
 */
{
#ifdef DEBUG
    if (netPtr == NIL(Net)) {
        MacPrintFatalErr("Mac_NumOfPortsOfNet() failed: no net is given"); 
    }
#endif DEBUG

    *numPtr = 0;  
    return ((Mac_ItrAllPortOfNet(netPtr, MacCountNumOfPorts, 
        (ClientData) numPtr) == BEAR_COMPLETE_SEARCH) ? BEAR_OK : BEAR_ERROR);
}



void
Mac_MarkPort(portInstPtr)
PortInst    *portInstPtr;
{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst)) {
        MacPrintFatalErr("Mac_MarkPort() failed: no port is given");
    }
#endif DEBUG

    portInstPtr->mark = BEAR_MARKED;
}

void
Mac_UnmarkPort(portInstPtr)
PortInst *portInstPtr;
{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst)) {
        MacPrintFatalErr("Mac_UnmarkPort() failed: no port is given");
    }
#endif DEBUG

    portInstPtr->mark = BEAR_UNMARKED;
}



BearStatus
Mac_IsPortMarked(portInstPtr)
PortInst *portInstPtr;
{
#ifdef DEBUG
    if (portInstPtr == NIL(PortInst)) {
      MacPrintFatalErr("Mac_IsPortMarked() failed: no port instance is given");
    }
#endif DEBUG

    return ((portInstPtr->mark == BEAR_MARKED) ? BEAR_TRUE : BEAR_FALSE);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetAssignedPinOfPort() --
 *
 * Given a port instance, it fetches the assigned pin if any, otherwise
 * the first pin on the pin list of the port is returned.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

PinInst *
Mac_GetAssignedPinOfPort(portPtr)
PortInst *portPtr;
{
#ifdef DEBUG
    if (portPtr == NIL(PortInst))
    {
        MacPrintFatalErr("Mac_GetAssignedPinOfPort() failed: No port given");
    }
#endif DEBUG

    return ((portPtr->assignedPinPtr != NIL(PinInst)) ?
        portPtr->assignedPinPtr : portPtr->pinList);
}



