
/* macRCell.c --
 *
 *     Wei-Ming Dai
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the creation, insertion, and deletion subroutines.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macRCell.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#ifdef USE_DR_CODE

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"
#include "routeDB.h"


typedef struct _RRootInfo
{
    Net   *netPtr;
    root  *rRootPtr;
} RRootInfo;


/* ---------- Local Procedures ---------- */
static BearStatus DeleteRRoot();
static BearStatus DeleteRRootFromAllNet();
static BearStatus DeleteRRootFromNet();
static BearStatus SetMinMax();


/*
 * ----------------------------------------------------------------------------
 *
 * MakeRCell() --
 *
 *    Allocates memory for a new route cell, and initializes
 *    the fields.
 *
 * Results:
 *    A pointer a route cell.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

/*
Channel interface testing need MakeRCell()
static RouteCell *
MakeRCell()
*/

RouteCell *
MakeRCell()
{
    register RouteCell *routeCellPtr;

    MALLOC(RouteCell *, routeCellPtr, sizeof (RouteCell));

    routeCellPtr->id = 0;
    routeCellPtr->type = HORZ_RCELL;
    routeCellPtr->mark = BEAR_UNMARKED;
    routeCellPtr->blBoundary = NIL(LinkedPoint);
    routeCellPtr->trBoundary = NIL(LinkedPoint);
    Mac_CopyRect(&routeCellPtr->bbox, &GeoNullRect);
    routeCellPtr->horzRectList = NIL(LinkedRect);
    routeCellPtr->vertRectList = NIL(LinkedRect);
    routeCellPtr->origin.p_x = 0;
    routeCellPtr->origin.p_y = 0;
    Mac_CopyTransform(&routeCellPtr->transform, &GeoIdentityTransform);
    routeCellPtr->routeModPtr = NIL(module);
    routeCellPtr->routeRootList = NIL(root);
    routeCellPtr->blBoundPinList = NIL(LinkedGPin);
    routeCellPtr->trBoundPinList = NIL(LinkedGPin);
    routeCellPtr->tlSidePinList = NIL(RoutePin);
    routeCellPtr->rSidePinList = NIL(RoutePin);
    routeCellPtr->cellMasterPtr = NIL(CellMaster);
    routeCellPtr->cellGroupPtr = NIL(CellGroup);
    routeCellPtr->cifFileNamePtr = NIL(char);
    routeCellPtr->nextPtr = NIL(RouteCell);
    routeCellPtr->clientData = (ClientData) NULL;

    return (routeCellPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_CreateRCell() --
 *
 *    Creates a route cell. The newly created route cell is attached to
 *    the given cell master.    
 *
 * Results:
 *    BEAR_OK, when done.
 *
 * Side effects:
 *    The new route cell is attached to the cell master.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_CreateRCell(cellMasterPtr, rcellPtrPtr)
CellMaster  *cellMasterPtr;
RouteCell   **rcellPtrPtr;
{
    RouteCell   *rcellPtr;
    RouteCell   *tmpRCellPtr;

#ifdef DEBUG
    if (cellMasterPtr == NIL(CellMaster))
    {
    MacPrintFatalErr("Mac_CreateRCell() failed: no cell master is given.");
    }
#endif DEBUG

    rcellPtr = MakeRCell();

    rcellPtr->cellMasterPtr = cellMasterPtr;
    rcellPtr->id = Mac_GenerateId();

    /* Add routeCell to end of list - to maintain ordering */
    if (cellMasterPtr->contents.routeCellList == NIL(RouteCell))
    {
       cellMasterPtr->contents.routeCellList = rcellPtr;
    }
    else
    {
       tmpRCellPtr = cellMasterPtr->contents.routeCellList;
       while (tmpRCellPtr->nextPtr != NIL(RouteCell))
       {
          tmpRCellPtr = tmpRCellPtr->nextPtr;
       }
       tmpRCellPtr->nextPtr = rcellPtr;
    }
    rcellPtr->nextPtr = NIL(RouteCell);

    *rcellPtrPtr = rcellPtr;
    return (BEAR_OK);
}



static BearStatus 
UnMarkBoundPin(gPinPtr, clientData)

GenericPin *gPinPtr;
ClientData clientData;
{

    if (gPinPtr->pinType == PIN_INST)
    {
       gPinPtr->pinPtrUn.pinInstPtr->portPtr->mark = BEAR_UNMARKED;
    }
    else
    {
       gPinPtr->pinPtrUn.routePinPtr->mark = BEAR_UNMARKED;
    }
    return(BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteRCell() --
 *
 *    Delete a routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The routing cell gets deleted from the cell master.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeleteRCell(rcellPtr)
RouteCell *rcellPtr;
{
    RouteCell   *curRCellPtr;
    CellMaster  *cellMasterPtr;

#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell)) {
        MacPrintFatalErr("Mac_DeleteRCell() failed : no route cell is given");
    }
#endif DEBUG

    cellMasterPtr = rcellPtr->cellMasterPtr;
    if (rcellPtr == cellMasterPtr->contents.routeCellList) {
        cellMasterPtr->contents.routeCellList = 
            cellMasterPtr->contents.routeCellList->nextPtr;
    }
    else {
        curRCellPtr = cellMasterPtr->contents.routeCellList;
        while (curRCellPtr->nextPtr != NIL(RouteCell) &&
            curRCellPtr->nextPtr != rcellPtr) {
            curRCellPtr = curRCellPtr->nextPtr;
        }
        if (curRCellPtr->nextPtr == rcellPtr) {
            curRCellPtr->nextPtr = curRCellPtr->nextPtr->nextPtr;
        }
        else {
            return (BEAR_ERROR);
        }
    }

    (void) Mac_ItrBoundPinOfRCell(rcellPtr, UnMarkBoundPin, NULL);

    (void) Rdb_ItrRRootOfRMod(rcellPtr->routeModPtr, DeleteRRoot, 
        (ClientData) Mac_GetCellOfRCell(rcellPtr));
/*
 * MEMORY LEAK : george : 15 nov 89
 *
 * Should free the entire RouteCell structure.
 *
 */
    Rdb_FreeRMod(rcellPtr->routeModPtr);
    FREE((char *) rcellPtr);

    return (BEAR_OK);
}



static BearStatus 
DeleteRRoot(rRootPtr, cellMasterPtr)

root         *rRootPtr;
CellMaster   *cellMasterPtr;
{
    (void) Mac_ItrNet(cellMasterPtr->contents.netList, 
        DeleteRRootFromAllNet, (ClientData) rRootPtr);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
DeleteRRootFromAllNet(netPtr, rRootPtr)

Net *netPtr;
root    *rRootPtr;
{
    RRootInfo   rRootInfo;

    rRootInfo.netPtr = netPtr;
    rRootInfo.rRootPtr = rRootPtr;
    (void) Mac_ItrRRootOfNet(netPtr, 
        DeleteRRootFromNet, (ClientData) &rRootInfo);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
DeleteRRootFromNet(rRootPtr, rRootInfoPtr)

root        *rRootPtr;
RRootInfo   *rRootInfoPtr;
{
    if (rRootPtr == rRootInfoPtr->rRootPtr)
    {
        (void) Mac_DeleteRRootOfNet(rRootInfoPtr->netPtr, rRootPtr);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertBoundPinOfRCell() --
 *
 * Insert a pin instance or a routing pin to the side of 
 * the routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The pin gets inserted to the side of the routing cell.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertBoundPinOfRCell(rcellPtr, boundary, gPinPtr)
RouteCell   *rcellPtr;
RCellBound  boundary;
GenericPin  *gPinPtr;
{
    LinkedGPin *newGPinPtr;

#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell)) 
    {
        MacPrintFatalErr("Mac_InsertBoundPinOfRCell() failed: no routing cell is given");
    }
    else if (gPinPtr == NIL(GenericPin))
    {
        MacPrintFatalErr("Mac_InsertBoundPinOfRCell() failed: no pin is given");
    }
#endif DEBUG

    MALLOC(LinkedGPin *, newGPinPtr, sizeof (LinkedGPin));
    newGPinPtr->gPinPtr = gPinPtr;
    newGPinPtr->nextPtr = NIL(LinkedGPin);

    switch (boundary)
    {

    case BOTTOM_LEFT_BOUND:
        newGPinPtr->nextPtr = rcellPtr->blBoundPinList;
        rcellPtr->blBoundPinList = newGPinPtr;
        break;

    case  TOP_RIGHT_BOUND:
        newGPinPtr->nextPtr = rcellPtr->trBoundPinList;
        rcellPtr->trBoundPinList = newGPinPtr;
        break;

    default:
        MacPrintFatalErr("Mac_InsertBoundPinOfRCell(): Illegal boundary type."); 
        break;
    }
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteBoundPinOfRCell() --
 *
 * Delete a pin instance or a routing pin from the boundary of 
 * the routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The pin gets deleted from the boundary of the routing cell.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeleteBoundPinOfRCell(rcellPtr, boundary, gPinPtr)
RouteCell   *rcellPtr;
RCellBound  boundary;
GenericPin *gPinPtr;
{
    LinkedGPin *curLinkedGPinPtr;

#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell)) 
    {
        MacPrintFatalErr("Mac_DeleteBoundPinOfRCell() failed: no routing cell is given");
    }
    else if (gPinPtr == NIL(GenericPin))
    {
        MacPrintFatalErr("Mac_DeleteBoundPinOfRCell() failed: no pin is given");
    }
#endif DEBUG

    switch (boundary)
    {

    case BOTTOM_LEFT_BOUND:
        if (gPinPtr == rcellPtr->blBoundPinList->gPinPtr)
        {
        rcellPtr->blBoundPinList = rcellPtr->blBoundPinList->nextPtr;
        }
        else
        {
        curLinkedGPinPtr = rcellPtr->blBoundPinList;  
        while (curLinkedGPinPtr->nextPtr != NIL(LinkedGPin) &&
               curLinkedGPinPtr->nextPtr->gPinPtr != gPinPtr)
        {
             curLinkedGPinPtr = curLinkedGPinPtr->nextPtr;
        }
        if (curLinkedGPinPtr->nextPtr->gPinPtr == gPinPtr)
        {
            curLinkedGPinPtr->nextPtr = curLinkedGPinPtr->nextPtr->nextPtr;
        }
        else
        {
            return (BEAR_ERROR);
        }
        }
        FREE((char *) gPinPtr);
        return (BEAR_OK);

    case TOP_RIGHT_BOUND:
        if (gPinPtr == rcellPtr->trBoundPinList->gPinPtr)
        {
        rcellPtr->trBoundPinList = rcellPtr->trBoundPinList->nextPtr;
        }
        else
        {
        curLinkedGPinPtr = rcellPtr->trBoundPinList;  
        while (curLinkedGPinPtr->nextPtr != NIL(LinkedGPin) &&
               curLinkedGPinPtr->nextPtr->gPinPtr != gPinPtr)
        {
             curLinkedGPinPtr = curLinkedGPinPtr->nextPtr;
        }
        if (curLinkedGPinPtr->nextPtr->gPinPtr == gPinPtr)
        {
            curLinkedGPinPtr->nextPtr = curLinkedGPinPtr->nextPtr->nextPtr;
        }
        else
        {
            return (BEAR_ERROR);
        }
        }
        FREE((char *) gPinPtr);
        return (BEAR_OK);

    default:
        MacPrintFatalErr("Mac_DeleteBoundPinOfRCell(): Illegal boundary type."); 
    }
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertSidePinOfRCell() --
 *
 * Insert a routing pin to the side of the routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The pin gets inserted to the side of the routing cell.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertSidePinOfRCell(rcellPtr, side, rPinPtr)
RouteCell     *rcellPtr;
SideDirection side;
RoutePin      *rPinPtr;
{
#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell)) 
    {
        MacPrintFatalErr("Mac_Mac_InsertSidePinOfRCell() failed: no routing cell is given");
    }
    else if (rPinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_Mac_InsertSidePinOfRCell() failed: no pin is given");
    }
#endif DEBUG

    switch (side)
    {

    case TOP_LEFT_SIDE:
        rPinPtr->nextPtr = rcellPtr->tlSidePinList;
        rcellPtr->tlSidePinList = rPinPtr;
        break;

    case  RIGHT_SIDE:
        rPinPtr->nextPtr = rcellPtr->rSidePinList;
        rcellPtr->rSidePinList = rPinPtr;
        break;

    default:
        MacPrintFatalErr("Mac_InsertSidePinOfRCell(): Illegal side type."); 
    }
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteSidePinOfRCell() --
 *
 * Delete a routing pin from the side of the routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The pin gets deleted from the side of the routing cell.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeleteSidePinOfRCell(rcellPtr, side, rPinPtr)
RouteCell     *rcellPtr;
SideDirection side;
RoutePin      *rPinPtr;
{
    RoutePin *curRPinPtr;

#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell))
    {
        MacPrintFatalErr("Mac_DeleteSidePinOfRCell() failed: no routing cell is given");
    }
    else if (rPinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_DeleteSidePinOfRCell() failed: no pin is given");
    }
#endif DEBUG

    switch (side)
    {

    case BOTTOM_LEFT_SIDE:
        if (rPinPtr == rcellPtr->tlSidePinList)
        {
        rcellPtr->tlSidePinList = rPinPtr->nextPtr;
        }
        else
        {
        curRPinPtr = rcellPtr->tlSidePinList;  
        while (curRPinPtr->nextPtr != NIL(RoutePin) &&
               curRPinPtr->nextPtr != rPinPtr)
        {
             curRPinPtr = curRPinPtr->nextPtr;
        }
        if (curRPinPtr->nextPtr == rPinPtr)
        {
            curRPinPtr->nextPtr = curRPinPtr->nextPtr->nextPtr;
        }
        else
        {
            return (BEAR_ERROR);
        }
        }
        FREE((char *) rPinPtr);
        return (BEAR_OK);

    case RIGHT_SIDE:
        if (rPinPtr == rcellPtr->rSidePinList)
        {
        rcellPtr->rSidePinList = rPinPtr->nextPtr;
        }
        else
        {
        curRPinPtr = rcellPtr->rSidePinList;  
        while (curRPinPtr->nextPtr != NIL(RoutePin) &&
               curRPinPtr->nextPtr != rPinPtr)
        {
             curRPinPtr = curRPinPtr->nextPtr;
        }
        if (curRPinPtr->nextPtr == rPinPtr)
        {
            curRPinPtr->nextPtr = curRPinPtr->nextPtr->nextPtr;
        }
        else
        {
            return (BEAR_ERROR);
        }
        }
        FREE((char *) rPinPtr);
        return (BEAR_OK);

    default:
        MacPrintFatalErr("Mac_DeleteSidePinOfRCell(): Illegal side direction."); 
    }
    return (BEAR_OK);
}

/*
 * ----------------------------------------------------------------------------
 *
 * Mac_InsertRRootOfRCell() --
 *
 * Insert a routing root to the routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The routing root gets inserted to the routing cell.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_InsertRRootOfRCell (rcellPtr, rRootPtr)

RouteCell *rcellPtr;
root      *rRootPtr;
{
#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell)) 
    {
        MacPrintFatalErr("Mac_InsertRRootOfRCell() failed: no routing cell is given");
    }
    else if (rRootPtr == NIL(root))
    {
        MacPrintFatalErr("Mac_InsertRRootOfRCell() failed: no root is given");
    }
#endif DEBUG

    rRootPtr->nextInRCell = rcellPtr->routeRootList;
    rcellPtr->routeRootList = rRootPtr;
    return (BEAR_OK);
}

/*
 * ----------------------------------------------------------------------------
 *
 * Mac_DeleteRRootOfRCell() --
 *
 * Delete a routing root from the routing cell.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   The routing root gets deleted from the routing cell.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_DeleteRRootOfRCell (rcellPtr, rRootPtr)

RouteCell *rcellPtr;
root      *rRootPtr;
{
    root *curRRootPtr;

#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell))
    {
        MacPrintFatalErr("Mac_DeleteRRootOfRCell() failed: no routing cell is given");
    }
    else if (rRootPtr == NIL(root))
    {
        MacPrintFatalErr("Mac_DeleteRRootOfRCell() failed: no root is given");
    }
#endif DEBUG

    if (rRootPtr == rcellPtr->routeRootList)
    {
    rcellPtr->routeRootList = rRootPtr->nextInRCell;
    }
    else
    {
    curRRootPtr = rcellPtr->routeRootList;  
    while (curRRootPtr->nextInRCell != NIL(root) &&
           curRRootPtr->nextInRCell != rRootPtr)
    {
         curRRootPtr = curRRootPtr->nextInRCell;
    }
    if (curRRootPtr->nextInRCell == rRootPtr)
    {
        curRRootPtr->nextInRCell = curRRootPtr->nextInRCell->nextInRCell;
    }
    else
    {
        return (BEAR_ERROR);
    }
    }
    FREE((char *) rRootPtr);
    return (BEAR_OK);
}



void
Mac_MoveRouteCell(deltaX, deltaY, rcellPtr)

int     deltaX;
int     deltaY;
RouteCell   *rcellPtr;
{
    Transform   newTrans;
    Transform   trans;
    Point       *originPtr;
    Point       newOrigin;

    GeoTranslateTrans(&rcellPtr->transform,
              deltaX, deltaY, &newTrans);
    rcellPtr->transform = newTrans;
    GeoTranslateTrans(&GeoIdentityTransform, deltaX,
              deltaY, &trans);
    originPtr = Mac_GetOriginOfRCell(rcellPtr);
    GeoTransPoint(&trans, originPtr, &newOrigin);
    Mac_PutOriginOfRCell(&newOrigin, rcellPtr);
}



void
Mac_CreateBBoxOfRCell(rcellPtr)  

RouteCell  *rcellPtr;
{
    LinkedPoint  *linkedPointPtr;
    Rect         bbox;

#ifdef DEBUG
    if (rcellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_CreateBBoxOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    bbox.r_xbot = INFINITY;
    bbox.r_ybot = INFINITY;
    bbox.r_xtop = MINFINITY;
    bbox.r_ytop = MINFINITY;
    
    linkedPointPtr = Mac_GetBLBoundOfRCell(rcellPtr);
    (void) Mac_ItrLinkedPoint(linkedPointPtr, SetMinMax, (ClientData) &bbox);

    linkedPointPtr = Mac_GetTRBoundOfRCell(rcellPtr);
    (void) Mac_ItrLinkedPoint(linkedPointPtr, SetMinMax, (ClientData) &bbox);

    Mac_CopyRect(&rcellPtr->bbox, &bbox);
}



static BearStatus
SetMinMax(linkedPointPtr, bboxPtr)
LinkedPoint     *linkedPointPtr;
register Rect   *bboxPtr;
{
    register int x, y;

    x = linkedPointPtr->point.p_x;
    y = linkedPointPtr->point.p_y;

    bboxPtr->r_xbot = MIN(bboxPtr->r_xbot, x);
    bboxPtr->r_ybot = MIN(bboxPtr->r_ybot, y);
    bboxPtr->r_xtop = MAX(bboxPtr->r_xtop, x);
    bboxPtr->r_ytop = MAX(bboxPtr->r_ytop, y);
    
    return (BEAR_CONTINUE_SEARCH);
}

#endif /* USE_DR_CODE */


