
/* macRCellItr.c --
 *
 *     Wei-Ming Dai
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * This file contains all the iterators related to routing cells.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macRCellItr.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"

typedef struct _LocalData
{
    BearStatus (*func)();
    ClientData clientData;
    Line       *interval;
} LocalData;

/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRCell() --
 *
 * Given a list of routing cells, iterates the routing cells on the list.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrRCell(rCellList, func, clientData)

RouteCell  *rCellList;
BearStatus (*func)();
ClientData clientData;
{
    RouteCell  *curRCellPtr;
#ifdef DEBUG
    if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrRouteCell() failed: no function is given");
    }
#endif DEBUG
    
    if (rCellList == NIL(RouteCell))
    {
        return (BEAR_COMPLETE_SEARCH);
    }

    curRCellPtr = rCellList;
    while (curRCellPtr != NIL(RouteCell))
    {
        if (((*func)(curRCellPtr, clientData)) == BEAR_STOP_SEARCH)
        {
            return (BEAR_STOP_SEARCH);
        }
        curRCellPtr = curRCellPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrBoundPinOfRCell() --
 *
 * Given a routing cell, it iterates the pin instances or routing pins
 * on the boundary of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrBoundPinOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    BearStatus bearStatus;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrBoundPinOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrBoundPinOfRCell() failed: no function is given");
    }
#endif DEBUG

    bearStatus = Mac_ItrBLBoundPinOfRCell(rCellPtr, func, clientData);
    if (bearStatus == BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }

    bearStatus = Mac_ItrTRBoundPinOfRCell(rCellPtr, func, clientData);
    return (bearStatus);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrBLBoundPinOfRCell() --
 *
 * Given a routing cell, it iterates the pin instances or routing pins
 * on the bottom-left boundary of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrBLBoundPinOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    LinkedGPin  *curLinkedGPinPtr;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrBLBoundPinOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrBLBoundPinOfRCell() failed: no function is given");
    }
#endif DEBUG

    if (rCellPtr->blBoundPinList == NIL(LinkedGPin)) {
        return (BEAR_COMPLETE_SEARCH);
    }

    curLinkedGPinPtr = rCellPtr->blBoundPinList;
    while (curLinkedGPinPtr != NIL(LinkedGPin)) {
        if (((*func)(curLinkedGPinPtr->gPinPtr, clientData)) == 
            BEAR_STOP_SEARCH) {
            return (BEAR_STOP_SEARCH);
        }
        curLinkedGPinPtr = curLinkedGPinPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrTRBoundPinOfRCell() --
 *
 * Given a routing cell, it iterates the pin instances or routing pins
 * on the top-right boundary of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrTRBoundPinOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    LinkedGPin  *curLinkedGPinPtr;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrTRBoundPinOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrTRBoundPinOfRCell() failed: no function is given");
    }
#endif DEBUG

    if (rCellPtr->trBoundPinList == NIL(LinkedGPin))
    {
    return (BEAR_COMPLETE_SEARCH);
    }

    curLinkedGPinPtr = rCellPtr->trBoundPinList;
    while (curLinkedGPinPtr != NIL(LinkedGPin))
    {
    if (((*func)(curLinkedGPinPtr->gPinPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    curLinkedGPinPtr = curLinkedGPinPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}

/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrSidePinOfRCell() --
 *
 * Given a routing cell, it iterates the routing pins on the open side 
 * of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrSidePinOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    BearStatus bearStatus;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrSidePinOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
        MacPrintFatalErr("Mac_ItrSidePinOfRCell() failed: no function is given");
    }
#endif DEBUG

    bearStatus = Mac_ItrTLSidePinOfRCell(rCellPtr, func, clientData); 
    if (bearStatus == BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    
    bearStatus = Mac_ItrRSidePinOfRCell(rCellPtr, func, clientData);
    return (bearStatus);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrTLSidePinOfRCell() --
 *
 * Given a routing cell, it iterates the routing pins on the top or left
 * open side of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrTLSidePinOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    RoutePin  *curRoutePinPtr;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrTLSidePinOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrTLSidePinOfRCell() failed: no function is given");
    }
#endif DEBUG

    if (rCellPtr->tlSidePinList == NIL(RoutePin))
    {
    return (BEAR_COMPLETE_SEARCH);
    }

    curRoutePinPtr = rCellPtr->tlSidePinList;
    while (curRoutePinPtr != NIL(RoutePin))
    {
    if (((*func)(curRoutePinPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    curRoutePinPtr = curRoutePinPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}

/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRSidePinOfRCell() --
 *
 * Given a routing cell, it iterates the routing pins on the right
 * open side of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrRSidePinOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    RoutePin  *curRoutePinPtr;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrRSidePinOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrRSidePinOfRCell() failed: no function is given");
    }
#endif DEBUG

    if (rCellPtr->rSidePinList == NIL(RoutePin))
    {
    return (BEAR_COMPLETE_SEARCH);
    }

    curRoutePinPtr = rCellPtr->rSidePinList;
    while (curRoutePinPtr != NIL(RoutePin))
    {
    if (((*func)(curRoutePinPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    curRoutePinPtr = curRoutePinPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}

/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRRootOfRCell() --
 *
 * Given a routing cell, it iterates the routing roots of the routing cell.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrRRootOfRCell(rCellPtr, func, clientData)

RouteCell  *rCellPtr;
BearStatus (*func)();
ClientData clientData;
{
    root  *curRRootPtr;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_ItrRRootOfRCell() failed: no routing cell is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrRRootOfRCell() failed: no function is given");
    }
#endif DEBUG

    if (rCellPtr->routeRootList == NIL(root))
    {
    return (BEAR_COMPLETE_SEARCH);
    }

    curRRootPtr = rCellPtr->routeRootList;
    while (curRRootPtr != NIL(root))
    {
    if (((*func)(curRRootPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    curRRootPtr = curRRootPtr->nextInRCell;
    }
    return (BEAR_COMPLETE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * CheckIntervalOfRPin() --
 *
 *    Given a routepin , it checks that the routepin is within the
 *    interval specified. And then applies the users function to the routepin.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
CheckIntervalOfRPin(rPinPtr, localDataPtr)
RoutePin    *rPinPtr;
LocalData   *localDataPtr;
{
    Point   absPos;

    Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if ((absPos.p_x >= localDataPtr->interval->point1.p_x) &&
        (absPos.p_x <= localDataPtr->interval->point2.p_x) &&
        (absPos.p_y >= localDataPtr->interval->point1.p_y) &&
        (absPos.p_y <= localDataPtr->interval->point2.p_y))
    {
        if ((*(localDataPtr->func))(rPinPtr, localDataPtr->clientData) == BEAR_STOP_SEARCH)
        {
            return (BEAR_STOP_SEARCH);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrSidePinOfIntervalOnRCell() --
 *
 *    Given a routecell , it iterates all the routepin of
 *    that routecell that are within the interval specified.
 *
 * Results:
 *    BEAR_COMPLETE_SEARCH, if successful.
 *    BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *    Whatever side effects result from application of the 
 *    supplied procedure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrSidePinOfIntervalOnRCell (rCellPtr, interval, func, clientData)
RouteCell  *rCellPtr;
Line       *interval;
BearStatus (*func)();
ClientData clientData;
{
    LocalData localData;

#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))  
    {
MacPrintFatalErr("Mac_ItrSidePinOfIntervalOnRCell() failed: no routecell is given");
    } 
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrSidePinOfIntervalOnRCell() failed: no func is given");
    }
#endif DEBUG

    localData.func = func;
    localData.interval = interval;
    localData.clientData = clientData;
    if (Mac_ItrSidePinOfRCell(rCellPtr, CheckIntervalOfRPin, (ClientData) &localData) == BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_COMPLETE_SEARCH);
}


