
/* macRCellPrim.c --
 *
 *     Wei-Ming Dai
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 *     It contains the operations to retrieve and update the attributes
 *     of routing cells.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macRCellPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Local Procedures ---------- */
static BearStatus UnMarkSidePin();
static BearStatus UnMarkBoundPin();
static BearStatus SearchSidePin();
static BearStatus SearchBoundPin();
static BearStatus MarkSidePin();
static BearStatus MarkBoundPin();


/* ---------- RouteCell Primitives ---------- */

BearId
Mac_GetIdOfRCell(rCellPtr)

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetIdOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (rCellPtr->id);
}

void
Mac_PutIdOfRCell(id, rCellPtr)

BearId     id;
RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
        MacPrintFatalErr("Mac_PutIdOfRCell() failed: no routing cell is given");
    }
    else if (id == 0)
    {
        MacPrintFatalErr("Mac_PutIdOfRCell() failed: id cannot be 0");
    }
#endif DEBUG

    rCellPtr->id = id;
}


RCellType
Mac_GetTypeOfRCell(rCellPtr)

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetTypeOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (rCellPtr->type);
}

BearStatus
Mac_PutTypeOfRCell(type, rCellPtr)

RCellType type;
RouteCell *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_PutTypeOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    rCellPtr->type = type;
    return (BEAR_OK);
}


void
Mac_MarkRCell(rCellPtr)

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_MarkRCell() failed: no routing cell is given");
    }
#endif DEBUG

    rCellPtr->mark = BEAR_MARKED;
}

void
Mac_UnmarkRCell(rCellPtr)

RouteCell *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_UnmarkRCell() failed: no routing cell is given");
    }
#endif DEBUG

    rCellPtr->mark = BEAR_UNMARKED;
}

BearStatus
Mac_IsRCellMarked(rCellPtr)

RouteCell *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_IsRCellMarked() failed: no routing cell is given");
    }
#endif DEBUG

    return ((rCellPtr->mark == BEAR_MARKED) ? BEAR_TRUE : BEAR_FALSE);
}


LinkedPoint *
Mac_GetBLBoundOfRCell(rCellPtr)

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetBLBoundOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (rCellPtr->blBoundary);
}


LinkedPoint *
Mac_GetTRBoundOfRCell(rCellPtr)

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetTRBoundOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (rCellPtr->trBoundary);
}



Rect *
Mac_GetBBoxOfRCell(rCellPtr)  

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetBBoxOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (&rCellPtr->bbox);
}


Point *
Mac_GetOriginOfRCell(rCellPtr)

RouteCell *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetOriginOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (&rCellPtr->origin);
}

void
Mac_PutOriginOfRCell(originPtr, rCellPtr)

Point    *originPtr;
RouteCell *rCellPtr;
{
#ifdef DEBUG
    if (originPtr == NIL(Point))
    {
    MacPrintFatalErr("Mac_PutOriginOfRCell() failed: no origin is given");
    }
    else if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_PutOriginOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    Mac_CopyPoint(&(rCellPtr->origin), originPtr);
}

Transform *
Mac_GetTransformOfRCell(rCellPtr)

RouteCell *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetTransformOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (&rCellPtr->transform);
}

void
Mac_PutTransformOfRCell(transPtr, rCellPtr)

Transform   *transPtr; 
RouteCell   *rCellPtr;
{
#ifdef DEBUG
    if (transPtr == NIL(Transform))  
    {
    Bear_PrintErr("Mac_PutTransformOfRCell() failed: no transform is given");
    } 
    else if (rCellPtr == NIL(RouteCell)) 
    {
    Bear_PrintErr("Mac_PutTransformOfRCell() failed: no routing cell given");
    }
#endif DEBUG

    Mac_CopyTransform(&(rCellPtr->transform), transPtr);
}

CellMaster *
Mac_GetCellOfRCell(rCellPtr)

RouteCell  *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetCellMasterOfRCell() failed: no routing cell is given");
    }
#endif DEBUG

    return (rCellPtr->cellMasterPtr);
}


CellGroup *
Mac_GetGroupOfRCell(rCellPtr)  

RouteCell    *rCellPtr;
{
#ifdef DEBUG
    if (rCellPtr == NIL(RouteCell))
    {
    MacPrintFatalErr("Mac_GetGroupOfRCell() failed: no route cell is given");
    }
#endif DEBUG

    return (rCellPtr->cellGroupPtr);
}


typedef struct _LocalData
{
    RouteCell  *rCellPtr;
    int        numOfNet;
} LocalData;

int
Mac_GetNumOfNetsOfRCell(rCellPtr)
RouteCell    *rCellPtr;
{
    LocalData   localData;

    (void) Mac_ItrSidePinOfRCell(rCellPtr, UnMarkSidePin,
        (ClientData)NULL);
    (void) Mac_ItrBoundPinOfRCell(rCellPtr, UnMarkBoundPin,
        (ClientData)NULL);

    localData.rCellPtr = rCellPtr;
    localData.numOfNet = 0;
    (void) Mac_ItrSidePinOfRCell(rCellPtr, SearchSidePin, 
        (ClientData)&localData);
    (void) Mac_ItrBoundPinOfRCell(rCellPtr, SearchBoundPin, 
        (ClientData)&localData);
    return (localData.numOfNet);
}



static BearStatus
SearchSidePin(rPinPtr,localDataPtr)
RoutePin   *rPinPtr;
LocalData  *localDataPtr;
{
    RouteCell  *rCellPtr;
    BearId     netId;

    if (rPinPtr->clientData != (ClientData)NULL)
    {
    return (BEAR_CONTINUE_SEARCH);
    }
    rCellPtr = localDataPtr->rCellPtr;
    netId = Mac_GetIntNetIdOfRPin(rPinPtr);
    (void) Mac_ItrSidePinOfRCell(rCellPtr, MarkSidePin,
        (ClientData)&netId);
    (void) Mac_ItrBoundPinOfRCell(rCellPtr, MarkBoundPin,
        (ClientData)&netId);
    localDataPtr->numOfNet++;
    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
SearchBoundPin(gPinPtr,localDataPtr)
GenericPin *gPinPtr;
LocalData  *localDataPtr;
{
    RouteCell   *rCellPtr;
    BearId      netId;
    PinInst     *pinInstPtr;
    RoutePin    *rPinPtr;
    BearStatus  MarkSidePin();  
    BearStatus  MarkBoundPin();

    rCellPtr = localDataPtr->rCellPtr;
    switch(Mac_GetTypeOfGPin(gPinPtr))
    {
    case PIN_INST:
    {
        pinInstPtr = Mac_GetPinInstOfGPin(gPinPtr);
        if (pinInstPtr->clientData != (ClientData)NULL)
        {
        return (BEAR_CONTINUE_SEARCH);
        }
        netId = Mac_GetExtNetIdOfPin(pinInstPtr);
        (void) Mac_ItrSidePinOfRCell(rCellPtr, MarkSidePin,
            (ClientData)&netId);
        (void) Mac_ItrBoundPinOfRCell(rCellPtr, MarkBoundPin,
            (ClientData)&netId);
        localDataPtr->numOfNet++;
    }
    break;
    case ROUTE_PIN:
    {
        rPinPtr = Mac_GetRoutePinOfGPin(gPinPtr);
        if (rPinPtr->clientData != (ClientData)NULL)
        {
        return (BEAR_CONTINUE_SEARCH);
        }
        netId = Mac_GetExtNetIdOfRPin(rPinPtr);
        (void) Mac_ItrSidePinOfRCell(rCellPtr, MarkSidePin,
            (ClientData)&netId);
        (void) Mac_ItrBoundPinOfRCell(rCellPtr, MarkBoundPin,
            (ClientData)&netId);
        localDataPtr->numOfNet++;
    }
    break;
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
MarkSidePin(rPinPtr,netIdPtr)
RoutePin   *rPinPtr;
BearId     *netIdPtr;
{
    if (*netIdPtr == Mac_GetIntNetIdOfRPin(rPinPtr))
    {
    rPinPtr->clientData = (ClientData)rPinPtr;
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
MarkBoundPin(gPinPtr,netIdPtr)
GenericPin *gPinPtr;
BearId     *netIdPtr;
{
    switch(Mac_GetTypeOfGPin(gPinPtr))
    {
    case PIN_INST:
    {
        if (*netIdPtr==Mac_GetExtNetIdOfPin(Mac_GetPinInstOfGPin(gPinPtr)))
        {
        Mac_GetPinInstOfGPin(gPinPtr)->clientData = (ClientData)gPinPtr;
        }
    }
    break;
    case ROUTE_PIN:
    {
        if (*netIdPtr==Mac_GetExtNetIdOfRPin(Mac_GetRoutePinOfGPin(gPinPtr)))
        {
        Mac_GetRoutePinOfGPin(gPinPtr)->clientData =(ClientData)gPinPtr;
        }
    }
    break;
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
UnMarkSidePin(rPinPtr,clientData)
RoutePin   *rPinPtr;
ClientData clientData;
{
    rPinPtr->clientData = clientData;
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
UnMarkBoundPin(gPinPtr,clientData)
GenericPin *gPinPtr;
ClientData clientData;
{
    switch(Mac_GetTypeOfGPin(gPinPtr))
    {
    case PIN_INST:
    {
        Mac_GetPinInstOfGPin(gPinPtr)->clientData = clientData;
    }
    break;
    case ROUTE_PIN:
    {
        Mac_GetRoutePinOfGPin(gPinPtr)->clientData = clientData;
    }
    break;
    }
    return (BEAR_CONTINUE_SEARCH);
}


