
/* macRPinPrim.c --
 *
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the subroutines used to retrieve and update
 * the information in routing pins.
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macRPinPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include <strings.h>
#include "magic.h"
#include "geometry.h"
#include "tile.h"
#include "malloc.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Pin Primitives ---------- */


void
Mac_MarkRPin(rpinPtr)
RoutePin *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin)) {
        MacPrintFatalErr("Mac_MarkRPin() failed: no routing pin given");
    }
#endif DEBUG

    rpinPtr->mark = BEAR_MARKED;
}

  
BearId
Mac_GetIdOfRPin(rpinPtr) 
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetIdOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->id);
}

void
Mac_PutIdOfRPin(id, rpinPtr) 
BearId  id;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutIdOfRPin() failed: no routing pin is given");
    }
    else if (id == 0)
    {
    MacPrintFatalErr("Mac_PutIdOfRPin() failed: id cannot be 0"); 
    }
#endif DEBUG

    rpinPtr->id = id;
}



BearId
Mac_GetIntNetIdOfRPin(rpinPtr) 
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetIdOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->intNetId);
}


void
Mac_PutIntNetIdOfRPin(intNetId, rpinPtr) 
BearId  intNetId;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutIntNetIdOfRPin() failed: no routing pin is given");
    }
    else if (intNetId == 0)
    {
    MacPrintFatalErr("Mac_PutIntNetIdOfRPin() failed: id cannot be 0"); 
    }
#endif DEBUG

    rpinPtr->intNetId = intNetId;
}


BearId
Mac_GetExtNetIdOfRPin(rpinPtr) 
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetIdOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->extNetId);
}


void
Mac_PutExtNetIdOfRPin(extNetId, rpinPtr) 
BearId  extNetId;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutExtNetIdOfRPin() failed: no routing pin is given");
    }
    else if (extNetId == 0)
    {
    MacPrintFatalErr("Mac_PutExtNetIdOfRPin() failed: id cannot be 0"); 
    }
#endif DEBUG

    rpinPtr->extNetId = extNetId;
}

BearId 
Mac_GetHExtNetIdOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_GetHExtNetIdOfRPin() failed: no pin is given");
    }
#endif DEBUG

    return (rpinPtr->horzExtNetId);
}


void
Mac_PutHExtNetIdOfRPin(id, rpinPtr)
BearId  id;
RoutePin *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_PutHExtNetIdOfRPin() failed: no pin is given");
    }
#endif DEBUG

    rpinPtr->horzExtNetId = id;
}


BearId 
Mac_GetVExtNetIdOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_GetIntNetIdOfRPin() failed: no pin is given");
    }
#endif DEBUG

    return (rpinPtr->vertExtNetId);
}


void
Mac_PutVExtNetIdOfRPin(id, rpinPtr)
BearId  id;
RoutePin *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_PutIntNetIdOfRPin() failed: no pin is given");
    }
#endif DEBUG

    rpinPtr->vertExtNetId = id;
}

Point *
Mac_GetRelPosOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetRelPosOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (&rpinPtr->position);
}


void
Mac_PutRelPosOfRPin(relPosPtr, rpinPtr)
Point   *relPosPtr;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutRelPosOfRPin() failed: no routing pin is given");
    }
    else if (relPosPtr == NIL(Point))
    {
    MacPrintFatalErr("Mac_PutRelPosOfPin() failed: no point is given");
    }
#endif DEBUG

    Mac_CopyPoint(&rpinPtr->position, relPosPtr);
}


void
Mac_GetAbsPosOfRPin(rpinPtr, absPosPtr)
RoutePin  *rpinPtr;
Point *absPosPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin)) {
    MacPrintFatalErr("Mac_GetAbsPosOfRPin() failed: no routing pin is given");
    }
    else if (rpinPtr->cellPtr == NIL(RouteCell)) {
    MacPrintFatalErr("Mac_GetAbsPosOfRPin() failed: the routing pin is not connected to a routing cell");
    }
#endif DEBUG

    GeoTransPoint(&rpinPtr->cellPtr->transform, &rpinPtr->position, absPosPtr);
}


PinDirection
Mac_GetDirectionOfRPin(rpinPtr)
RoutePin  *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin)) {
    MacPrintFatalErr("Mac_GetDirectionOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->direction);
}


void
Mac_PutDirectionOfRPin(direction, rpinPtr)
PinDirection direction;
RoutePin     *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutDirectionOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    rpinPtr->direction = direction;
}


LayerVector
Mac_GetLayerOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetLayerVectorOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->layer);
}


void
Mac_PutLayerOfRPin(layer, rpinPtr)
LayerVector  layer;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutLayerVectorOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    rpinPtr->layer = layer;
}


float
Mac_GetPinWidthOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_GetPinWidthOfRPin() failed:  no pin is given");
    }
#endif DEBUG

    return (rpinPtr->pinWidth);
}


void
Mac_PutPinWidthOfRPin(pinWidth, rpinPtr)
float    pinWidth;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_PutPinWidthOfRPin() failed:  no pin is given");
    }
#endif DEBUG

    rpinPtr->pinWidth = pinWidth;
}


float
Mac_GetSegmWidthOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_GetSegmWidthOfRPin() failed:  no pin is given");
    }
#endif DEBUG

    return (rpinPtr->segmWidth);
}


void
Mac_PutSegmWidthOfRPin(segmWidth, rpinPtr)
float    segmWidth;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_PutSegmWidthOfRPin() failed:  no pin is given");
    }
#endif DEBUG

    rpinPtr->segmWidth = segmWidth;
}


float *
Mac_GetWidthVectOfRPin(rpinPtr)

RoutePin  *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_GetWidthVectOfRPin() failed:  no pin is given");
    }
#endif DEBUG

    return (rpinPtr->widthVect);
}


void
Mac_PutWidthVectOfRPin(widthVect, rpinPtr)

float    *widthVect;
RoutePin *rpinPtr;
{
    int i;

#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
        MacPrintFatalErr("Mac_PutWidthVectOfRPin() failed:  no pin is given");
    }
#endif DEBUG

    for (i = 0; i < 8; i++)
    {
        rpinPtr->widthVect[i] = widthVect[i];
    }
}


Flexibility
Mac_GetFlexOfRPin(rpinPtr)
RoutePin *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin)) {
        MacPrintFatalErr("Mac_GetFlexOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->positionFlag);
}


void
Mac_PutFlexOfRPin(positionFlag, rpinPtr)
Flexibility positionFlag;
RoutePin    *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin)) {
        MacPrintFatalErr("Mac_PutFlexOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    rpinPtr->positionFlag = positionFlag;
}


Net *
Mac_GetNetPtrOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetNetPtrOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->netPtr);
}


void
Mac_PutNetPtrOfRPin(netPtr, rpinPtr)
Net       *netPtr;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutNetPtrOfRPin() failed: no routing pin is given");
    }
    if (netPtr == NIL(Net))
    {
    MacPrintFatalErr("Mac_PutNetPtrOfRPin() failed: no netPtr is given");
    }
#endif DEBUG

    rpinPtr->netPtr = netPtr;
}



int
Mac_GetOrderOfRPin(rpinPtr)
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetOrderOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->order);
}


void
Mac_PutOrderOfRPin(order, rpinPtr)
int       order;
RoutePin  *rpinPtr;

{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutOrderOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    rpinPtr->order = order;
}



ClientData
Mac_GetPinOfRPin(rpinPtr)

RoutePin    *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_GetPinOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    return (rpinPtr->pinPtr);
}

void
Mac_PutPinOfRPin(pinPtr, rpinPtr)

ClientData    pinPtr;
RoutePin      *rpinPtr;
{
#ifdef DEBUG
    if (rpinPtr == NIL(RoutePin))
    {
    MacPrintFatalErr("Mac_PutPinOfRPin() failed: no routing pin is given");
    }
#endif DEBUG

    rpinPtr->pinPtr = pinPtr;
}


