
/* macRingItr.c --
 *
 *     Wei-Ming Dai
 *     (Advising Professor - E. S. Kuh)
 *
 * This file contains all the iterators related to routing ring.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macRingItr.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#ifdef USE_DR_CODE

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"

typedef struct _LocalData
{
    BearStatus (*func)();
    ClientData clientData;
    Line       *interval;
} LocalData;


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrBoundPinOfRing() --
 *
 * Given a routing ring, it iterates the pin instances or routing pins
 * on the boundary of the routing ring.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrBoundPinOfRing(ringPtr, func, clientData)

Ring  *ringPtr;
BearStatus (*func)();
ClientData clientData;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_ItrBoundPinOfRing() failed: no routing ring is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrBoundPinOfRing() failed: no function is given");
    }
#endif DEBUG

    (void) Mac_ItrInBoundPinOfRing(ringPtr, func, clientData);
    (void) Mac_ItrOutBoundPinOfRing(ringPtr, func, clientData);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrInBoundPinOfRing() --
 *
 * Given a routing ring, it iterates the pin instances or routing pins
 * on the inner boundary of the routing ring.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrInBoundPinOfRing(ringPtr, func, clientData)

Ring  *ringPtr;
BearStatus (*func)();
ClientData clientData;
{
    LinkedGPin  *curLinkedGPinPtr;

#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_ItrInBoundPinOfRing() failed: no routing ring is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrInBoundPinOfRing() failed: no function is given");
    }
#endif DEBUG

    if (ringPtr->inBoundPinList == NIL(LinkedGPin))
    {
        return (BEAR_COMPLETE_SEARCH);
    }

    curLinkedGPinPtr = ringPtr->inBoundPinList;
    while (curLinkedGPinPtr != NIL(LinkedGPin))
    {
        if (((*func)(curLinkedGPinPtr->gPinPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
        curLinkedGPinPtr = curLinkedGPinPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrOutBoundPinOfRing() --
 *
 * Given a routing ring, it iterates the pin instances or routing pins
 * on the outer boundary of the routing ring.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrOutBoundPinOfRing(ringPtr, func, clientData)

Ring  *ringPtr;
BearStatus (*func)();
ClientData clientData;
{
    LinkedGPin  *curLinkedGPinPtr;

#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_ItrOutBoundPinOfRing() failed: no routing ring is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrOutBoundPinOfRing() failed: no function is given");
    }
#endif DEBUG

    if (ringPtr->outBoundPinList == NIL(LinkedGPin))
    {
        return (BEAR_COMPLETE_SEARCH);
    }

    curLinkedGPinPtr = ringPtr->outBoundPinList;
    while (curLinkedGPinPtr != NIL(LinkedGPin))
    {
    if (((*func)(curLinkedGPinPtr->gPinPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    curLinkedGPinPtr = curLinkedGPinPtr->nextPtr;
    }
    return (BEAR_COMPLETE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_ItrRRootOfRing() --
 *
 * Given a routing ring, it iterates the routing roots of the routing ring.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, if the function returns the same status.
 *
 * Side effects:
 *   Whatever the side effects of the user function if any.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Mac_ItrRRootOfRing (ringPtr, func, clientData)

Ring  *ringPtr;
BearStatus (*func)();
ClientData clientData;
{
    root  *curRRootPtr;

#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_ItrRRootOfRing() failed: no routing ring is given");
    }
    else if (func == (BearStatus (*)()) NULL)
    {
    MacPrintFatalErr("Mac_ItrRRootOfRing() failed: no function is given");
    }
#endif DEBUG

    if (ringPtr->routeRootList == NIL(root))
    {
        return (BEAR_COMPLETE_SEARCH);
    }

    curRRootPtr = ringPtr->routeRootList;
    while (curRRootPtr != NIL(root))
    {
    if (((*func)(curRRootPtr, clientData)) == 
            BEAR_STOP_SEARCH)
    {
        return (BEAR_STOP_SEARCH);
    }
    curRRootPtr = curRRootPtr->nextInRCell;
    }
    return (BEAR_COMPLETE_SEARCH);
}

#endif /* USE_DR_CODE */


