
/* macRingPrim.c --
 *
 *     Wei-Ming Dai
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the operations to retrieve and update the attributes
 * of routing rings.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macRingPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#ifdef USE_DR_CODE

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"


/* ---------- Local Procedures ---------- */
static BearStatus   SearchBoundPin();
static BearStatus   MarkBoundPin();
static BearStatus   UnMarkBoundPin();


/* ---------- Ring Primitives ---------- */


BearId
Mac_GetIdOfRing(ringPtr)
Ring  *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetIdOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (ringPtr->id);
}


LinkedPoint *
Mac_GetInBoundOfRing(ringPtr)

Ring  *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetInBoundOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (ringPtr->inBoundary);
}

LinkedPoint *
Mac_GetOutBoundOfRing(ringPtr)

Ring  *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetOutBoundOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (ringPtr->outBoundary);
}


Rect *
Mac_GetInBBoxOfRing(ringPtr)  

Ring    *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetInBBoxOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (&ringPtr->inBBox);
}

void
Mac_PutInBBoxOfRing(bboxPtr, ringPtr)  

Rect    *bboxPtr;
Ring    *ringPtr;
{
#ifdef DEBUG
    if (bboxPtr == NIL(Rect))
    {
    MacPrintFatalErr("Mac_PutInBBoxOfRing() failed: no BBox is given");
    }
    else if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_PutInBBoxOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    Mac_CopyRect(&(ringPtr->inBBox), bboxPtr);
}

Rect *
Mac_GetOutBBoxOfRing(ringPtr)  

Ring    *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetOutBBoxOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (&ringPtr->outBBox);
}

void
Mac_PutOutBBoxOfRing(bboxPtr, ringPtr)  

Rect    *bboxPtr;
Ring    *ringPtr;
{
#ifdef DEBUG
    if (bboxPtr == NIL(Rect))
    {
    MacPrintFatalErr("Mac_PutOutBBoxOfRing() failed: no BBox is given");
    }
    else if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_PutOutBBoxOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    Mac_CopyRect(&(ringPtr->outBBox), bboxPtr);
}


Point *
Mac_GetOriginOfRing(ringPtr)

Ring    *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetOriginOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (&ringPtr->origin);
}

void
Mac_PutOriginOfRing(originPtr, ringPtr)

Point   *originPtr;
Ring    *ringPtr;
{
#ifdef DEBUG
    if (originPtr == NIL(Point))
    {
    MacPrintFatalErr("Mac_PutOriginOfRing() failed: no origin is given");
    }
    else if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_PutOriginOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    Mac_CopyPoint(&(ringPtr->origin), originPtr);
}


Transform *
Mac_GetTransformOfRing(ringPtr)

Ring *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetTransformOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (&ringPtr->transform);
}

void
Mac_PutTransformOfRing(transPtr, ringPtr)

Transform   *transPtr; 
Ring   *ringPtr;
{
#ifdef DEBUG
    if (transPtr == NIL(Transform))  
    {
    Bear_PrintErr("Mac_PutTransformOfRing() failed: no transform is given");
    } 
    else if (ringPtr == NIL(Ring)) 
    {
    Bear_PrintErr("Mac_PutTransformOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    Mac_CopyTransform(&(ringPtr->transform), transPtr);
}


module *
Mac_GetRModOfRing(ringPtr)

Ring    *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetRModOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (ringPtr->routeModPtr);
}

void
Mac_PutRModOfRing(rModPtr, ringPtr)

module   *rModPtr;
Ring     *ringPtr;
{
#ifdef DEBUG
    if (rModPtr == NIL(module))
    {
    MacPrintFatalErr("Mac_PutRModOfRing() failed: no routing module is given");
    }
    else if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_PutRModOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    ringPtr->routeModPtr = rModPtr;
}


CellMaster *
Mac_GetCellOfRing(ringPtr)

Ring  *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetCellMasterOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (ringPtr->cellMasterPtr);
}


char *
Mac_GetCifFileNameOfRing(ringPtr)

Ring    *ringPtr;
{
#ifdef DEBUG
    if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_GetCifFileNameOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    return (ringPtr->cifFileName);
}

void
Mac_PutCifFileNameOfRing(cifFileName, ringPtr)

char    *cifFileName;
Ring    *ringPtr;
{
#ifdef DEBUG
    if (cifFileName == NIL(char))
    {
    MacPrintFatalErr("Mac_PutCifFileNameOfRing() failed: no file name is given");
    }
    else if (ringPtr == NIL(Ring))
    {
    MacPrintFatalErr("Mac_PutCifFileNameOfRing() failed: no routing ring is given");
    }
#endif DEBUG

    ringPtr->cifFileName = cifFileName;
}

typedef struct _LocalData
{
    Ring       *ringPtr;
    int        numOfNet;
} LocalData;


int
Mac_GetNumOfNetsOfRing(ringPtr)
Ring    *ringPtr;
{
    LocalData   localData;

    Mac_ItrBoundPinOfRing(ringPtr,UnMarkBoundPin,(ClientData)NULL);

    localData.ringPtr  = ringPtr;
    localData.numOfNet = 0;
    Mac_ItrBoundPinOfRing(ringPtr,SearchBoundPin,(ClientData)&localData);
    return (localData.numOfNet);
}



static BearStatus
SearchBoundPin(gPinPtr,localDataPtr)
GenericPin *gPinPtr;
LocalData  *localDataPtr;
{
    Ring        *ringPtr;
    BearId      netId;
    PinInst     *pinInstPtr;
    RoutePin    *rPinPtr;

    ringPtr = localDataPtr->ringPtr;
    switch(Mac_GetTypeOfGPin(gPinPtr))
    {
    case PIN_INST:
    {
        pinInstPtr = Mac_GetPinInstOfGPin(gPinPtr);
        if (pinInstPtr->clientData != (ClientData)NULL)
        {
        return (BEAR_CONTINUE_SEARCH);
        }
        netId = Mac_GetExtNetIdOfPin(pinInstPtr);
        Mac_ItrBoundPinOfRing(ringPtr,MarkBoundPin,(ClientData)&netId);
        localDataPtr->numOfNet++;
    }
    break;
    case ROUTE_PIN:
    {
        rPinPtr = Mac_GetRoutePinOfGPin(gPinPtr);
        if (rPinPtr->clientData != (ClientData)NULL)
        {
        return (BEAR_CONTINUE_SEARCH);
        }
        netId = Mac_GetExtNetIdOfRPin(rPinPtr);
        Mac_ItrBoundPinOfRing(ringPtr,MarkBoundPin,(ClientData)&netId);
        localDataPtr->numOfNet++;
    }
    break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
MarkBoundPin(gPinPtr,netIdPtr)
GenericPin *gPinPtr;
BearId     *netIdPtr;
{
    switch(Mac_GetTypeOfGPin(gPinPtr))
    {
    case PIN_INST:
    {
        if (*netIdPtr==Mac_GetExtNetIdOfPin(Mac_GetPinInstOfGPin(gPinPtr)))
        {
        Mac_GetPinInstOfGPin(gPinPtr)->clientData = (ClientData)gPinPtr;
        }
    }
    break;
    case ROUTE_PIN:
    {
        if (*netIdPtr==Mac_GetExtNetIdOfRPin(Mac_GetRoutePinOfGPin(gPinPtr)))
        {
        Mac_GetRoutePinOfGPin(gPinPtr)->clientData =(ClientData)gPinPtr;
        }
    }
    break;
    }
    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
UnMarkBoundPin(gPinPtr,clientData)
GenericPin *gPinPtr;
ClientData clientData;
{
    switch(Mac_GetTypeOfGPin(gPinPtr))
    {
    case PIN_INST:
    {
        Mac_GetPinInstOfGPin(gPinPtr)->clientData = clientData;
    }
    break;
    case ROUTE_PIN:
    {
        Mac_GetRoutePinOfGPin(gPinPtr)->clientData = clientData;
    }
    break;
    }
    return (BEAR_CONTINUE_SEARCH);
}

#endif /* USE_DR_CODE */


