
/* macTech.c --
 *
 *     Masahiro Kobayashi
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the subroutines to read or write a technilogy file.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macTech.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"
#include "macTechInt.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */


#define NUM_OF_LAYERS       3           /* Number of metal layers.           */

#define NUM_OF_LAY_VECT     4           /* Number of layer vectors.          */

#define NUM_OF_HOLE_VECT    3           /* Number of hole vectors.           */

#define NUM_OF_LLR_ROWS     4           /* Number of rows in the
                                         * layer layer rule matrix.
                                         */
#define NUM_OF_LLR_COLS     4           /* Number of columns in the
                                         * layer layer rule matrix.
                                         */

#define NUM_OF_HHR_ROWS     3           /* Number of rows in the
                                         * hole hole rule matrix.
                                         */
#define NUM_OF_HHR_COLS     3           /* Number of columns in the
                                         * hole hole rule matrix.
                                         */

#define NUM_OF_LHR_ROWS     4           /* Number of rows in the
                                         * layer hole rule matrix.
                                         */
#define NUM_OF_LHR_COLS     3           /* Number of columns in the
                                         * layer hole rule matrix.
                                         */


/* ---------- Global variables used outside of this module. ---------- */
int             mac_NumOfLayer;


/* ---------- Global variables used within this module. ---------- */
LayerRule       *macLayVect;
HoleRule        *macHoleVect;
LayerLayerRule  **macLayLayMtrx;
HoleHoleRule    **macHoleHoleMtrx;
LayerHoleRule   **macLayHoleMtrx;


/* ---------- Local Procedures ---------- */
void InitTechData();


static void
InitTechData()
{
    int i, j;

    /*
     *  Layer Rule
     */
    for (i = 0 ; i < NUM_OF_LAY_VECT ; i++) {
        macLayVect[i].minWidth = 0;
        macLayVect[i].maxCurrent = 0.0;
        macLayVect[i].sheetRes = 0.0;
        macLayVect[i].plateCap = 0.0;
        macLayVect[i].perimCap = 0.0;
    }

    /*
     *  Hole Rule
     */
    for (i = 0 ; i < NUM_OF_HOLE_VECT ; i++) {
        macHoleVect[i].size = 0;
        macHoleVect[i].current = 0.0;
        macHoleVect[i].res = 0.0;
        macHoleVect[i].cap = 0.0;
    }

    /*
     *  Layer Layer Rule
     */
    for (i = 0 ; i < NUM_OF_LLR_ROWS ; i++) {
        for (j = 0 ; j < NUM_OF_LLR_COLS ; j++) {
            macLayLayMtrx[i][j].minSpacing = 0;
            macLayLayMtrx[i][j].maxParallel = 0;
        }
    }

    /*
     *  Hole Hole Rule
     */
    for (i = 0 ; i < NUM_OF_HHR_ROWS ; i++) {
        for (j = 0 ; j < NUM_OF_HHR_COLS ; j++) {
            macHoleHoleMtrx[i][j].minSpacing = 0;
        }
    }

    /*
     *  Layer Hole Rule
     */
    for (i = 0 ; i < NUM_OF_LHR_ROWS ; i++) {
        for (j = 0 ; j < NUM_OF_LHR_COLS ; j++) {
            macLayHoleMtrx[i][j].minSpacing = 0;
            macLayHoleMtrx[i][j].minOverlap = 0;
        }
    }
}



void
MacMakeTechData()
{
    int i;

    /*
     *  Layer Rule
     */
    MALLOC(LayerRule *, macLayVect, NUM_OF_LAY_VECT * sizeof (LayerRule));

    /*
     *  Hole Rule
     */
    MALLOC(HoleRule *, macHoleVect, NUM_OF_HOLE_VECT * sizeof (HoleRule));

    /*
     *  Layer Layer Rule
     */
    MALLOC(LayerLayerRule **, macLayLayMtrx, 
        NUM_OF_LLR_ROWS * sizeof (LayerLayerRule *));
    for (i = 0 ; i < NUM_OF_LLR_ROWS ; i++) {
        MALLOC(LayerLayerRule *, macLayLayMtrx[i], 
            NUM_OF_LLR_COLS * sizeof (LayerLayerRule));
    }

    /*
     *  Hole Hole Rule
     */
    MALLOC(HoleHoleRule **, macHoleHoleMtrx, 
        NUM_OF_HHR_ROWS * sizeof (HoleHoleRule *));
    for (i = 0 ; i < NUM_OF_HHR_ROWS ; i++) {
        MALLOC(HoleHoleRule *, macHoleHoleMtrx[i], 
            NUM_OF_HHR_COLS * sizeof (HoleHoleRule));
    }

    /*
     *  Layer Hole Rule
     */
    MALLOC(LayerHoleRule **, macLayHoleMtrx, 
        NUM_OF_LHR_ROWS * sizeof (LayerHoleRule *));
    for (i = 0 ; i < NUM_OF_LHR_ROWS ; i++) {
        MALLOC(LayerHoleRule *, macLayHoleMtrx[i], 
            NUM_OF_LHR_COLS * sizeof (LayerHoleRule));
    }

    InitTechData();
}



/*
 * Mac_ReadTech() --
 *
 *   Read in design rule data from technology file.
 *
 * Results:
 *   BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   A bunch of global variables are filled.
 */

BearStatus
Mac_ReadTech(fileName)
BearString fileName;
{
    FILE *filePtr;
    int  count, value;
    char keyWord[80];

#ifdef DEBUG
    if (fileName == (BearString) NULL) {
        MacPrintFatalErr("Mac_ReadTech() failed : no technology file given");
    }
#endif DEBUG

#ifdef USE_OCT_CODE
    (void) VOVinput(VOV_UNIX_FILE, fileName);
#endif /* USE_OCT_CODE */

    /*
     * Open technology file for reading.
     */
    if ((filePtr = fopen (fileName, "r")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Cannot read from technology file named \"%s\"\n", fileName);
        Bear_PrintErr(bear_ErrorMsg);

        return (BEAR_ERROR);
    }

    InitTechData();

    mac_NumOfLayer = 3;

    count = fscanf (filePtr, "%s %d", keyWord, &value);
    if (count != 2 || strcmp(keyWord, "M1_MinWireWidth") != 0) {
       Bear_PrintErr("Mac_ReadTech() : Problem with  M1_MinWireWidth line.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    if (value < 2) {
   Bear_PrintErr("Mac_ReadTech() : M1_MinWireWidth is smaller than 2.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    Mac_PutMinWidthOfLay(value, METAL1_LAYER);

    count = fscanf (filePtr, "%s %d", keyWord, &value);
    if (count != 2 || strcmp(keyWord, "M2_MinWireWidth") != 0) {
        Bear_PrintErr("Mac_ReadTech() : Problem with M2_MinWireWidth line.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    if (value < 2) {
   Bear_PrintErr("Mac_ReadTech() : M2_MinWireWidth is smaller than 2.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    Mac_PutMinWidthOfLay(value, METAL2_LAYER);

    count = fscanf (filePtr, "%s %d", keyWord, &value);
    if (count != 2 || strcmp(keyWord, "M1_MinWireSpacing") != 0) {
      Bear_PrintErr("Mac_ReadTech() : Problem with M1_MinWireSpacing line.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    Mac_PutMinSpaceOfLayLay(value, METAL1_LAYER, METAL1_LAYER);

    count = fscanf (filePtr, "%s %d", keyWord, &value);
    if (count != 2 || strcmp(keyWord, "M2_MinWireSpacing") != 0) {
      Bear_PrintErr("Mac_ReadTech() : Problem with  M2_MinWireSpacing line.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    Mac_PutMinSpaceOfLayLay(value, METAL2_LAYER, METAL2_LAYER);

    count = fscanf (filePtr, "%s %d", keyWord, &value);
    if (count != 2 || strcmp(keyWord, "ViaSize") != 0) {
        Bear_PrintErr("Mac_ReadTech() : Problem with ViaSize line.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    if (value < 2) {
        Bear_PrintErr("Mac_ReadTech() : ViaSize is smaller than 2.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    Mac_PutSizeOfHole(value, M1M2_VIA);

    count = fscanf (filePtr, "%s %d", keyWord, &value);
    if (count != 2 || strcmp(keyWord, "MinOverLap") != 0) {
        Bear_PrintErr("Mac_ReadTech() : Problem with MinOverLap line.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    Mac_PutMinOverlapOfLayHole(value, METAL1_LAYER, M1M2_VIA);

    fclose (filePtr);
    return (BEAR_OK);
}



/*
 * Mac_WriteTech() --
 *
 *   Write design rule data to technology file.
 *
 * Results:
 *   BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   none.
 */

BearStatus
Mac_WriteTech(fileName, scale)
BearString  fileName;
double      scale;
{
    FILE *filePtr;
    int  value;
     
#ifdef DEBUG
    if ( fileName == (BearString) NULL ) {
        MacPrintFatalErr("Mac_WriteTech() failed: no technology file given");
    }
#endif DEBUG

    if (scale < 0.0) {
        Bear_PrintErr("Mac_WriteTech() : Scale must be greater then 0.0");
        return (BEAR_ERROR);
    }

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileName);
#endif /* USE_OCT_CODE */

    /*
     * Open technology file for writing.
     */
    if ((filePtr = fopen (fileName, "w")) == NIL(FILE)) {
        sprintf (bear_ErrorMsg,
            "Cannot write to technology file named \"%s\"\n",
            fileName);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR);
    }

    (void) fprintf (filePtr,"\n");
    /*
     * Print out the minimum width of a wire for the layer
     * containing metal one.
     */
    value = Mac_GetMinWidthOfLay(METAL1_LAYER) * scale;
    if (value < 2) {
   Bear_PrintErr("Mac_WriteTech() : M1_MinWireWidth is smaller than 2.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    (void) fprintf (filePtr, "%-20s %d\n\n", "M1_MinWireWidth", value);

    /*
     * Print out the minimum width of a wire for the layer
     * containing metal two.
     */
    value = Mac_GetMinWidthOfLay(METAL2_LAYER) * scale;
    if (value < 2) {
   Bear_PrintErr("Mac_WriteTech() : M2_MinWireWidth is smaller than 2.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    (void) fprintf (filePtr, "%-20s %d\n\n", "M2_MinWireWidth", value);

    /*
     * Print out the minimum spacing between two wires on the layer
     * containing metal one.
     */
    value = Mac_GetMinSpaceOfLayLay(METAL1_LAYER, METAL1_LAYER) * scale;
    (void) fprintf (filePtr, "%-20s %d\n\n", "M1_MinWireSpacing", value);

    /*
     * Print out the minimum spacing between two wires on the layer
     * containing metal two.
     */
    value = Mac_GetMinSpaceOfLayLay(METAL2_LAYER, METAL2_LAYER) * scale;
    (void) fprintf (filePtr, "%-20s %d\n\n", "M2_MinWireSpacing", value);

    /*
     * Print out the via size.
     */
    value = Mac_GetSizeOfHole(M1M2_VIA) * scale;
    if (value < 2) {
        Bear_PrintErr("Mac_WriteTech() : ViaSize is smaller than 2.\n");

        fclose (filePtr);
        return (BEAR_ERROR);
    }
    (void) fprintf (filePtr, "%-20s %d\n\n", "ViaSize", value);

    /*
     * Print out the minimum overlap.
     */
    value = Mac_GetMinOverlapOfLayHole(METAL1_LAYER, M1M2_VIA) * scale;
    (void) fprintf (filePtr, "%-20s %d\n\n", "MinOverLap", value);

    (void) fclose(filePtr);

    return (BEAR_OK);
}



int
Mac_CalcMinPinSpacing()
{
    int minWidth1, minWidth2;
    int minSpacing1, minSpacing2;
    int minViaSize, minOverLap;
    int d1, d2, d3;

    minWidth1 = Mac_GetMinWidthOfLay(METAL1_LAYER);
    minWidth2 = Mac_GetMinWidthOfLay(METAL2_LAYER);

    minSpacing1 = Mac_GetMinSpaceOfLayLay(METAL1_LAYER, METAL1_LAYER);
    minSpacing2 = Mac_GetMinSpaceOfLayLay(METAL2_LAYER, METAL2_LAYER);

    minViaSize = Mac_GetSizeOfHole(M1M2_VIA);
    minOverLap = Mac_GetMinOverlapOfLayHole(METAL1_LAYER, M1M2_VIA);
    
    d1 = minWidth1 + minSpacing1;
    d2 = minWidth2 + minSpacing2;
    d3 = minViaSize + 2*minOverLap + MAX(minSpacing1, minSpacing2);

    return (MAX(MAX(d1, d2), d3));
}


