
/* macTechPrim.c --
 *
 *     Masahiro Kobayashi
 *     (Advising Professor - E. S. Kuh)
 *
 * Get or put design rule value
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/mac/macTechPrim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "macInt.h"
#include "macTechInt.h"


/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetMinWidthOfLay()  --
 *
 *    Get minimum width of layer
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

int
Mac_GetMinWidthOfLay(layer)
int layer;
{
#ifdef DEBUG
    /*
    printf("mac_NumOfLayer = %d\n", mac_NumOfLayer); 
    fflush(stdout);
    */
    if (layer < 0 || layer > mac_NumOfLayer) {
        MacPrintFatalErr("Mac_GetMinWidthOfLay(): invalid layer number");
    }
#endif DEBUG

    return (macLayVect[layer].minWidth);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetSizeOfHole()  --
 *
 *   Get size of hole.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

int
Mac_GetSizeOfHole(hole)
int hole;
{
#ifdef DEBUG
    if (hole < 0 || hole >= mac_NumOfLayer ) {
        MacPrintFatalErr("Mac_GetSizeOfHole(): invalid hole number");
    }
#endif DEBUG

    return (macHoleVect[hole].size);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetMinSpaceOfLayLay()  --
 *
 *    Get minimum spacing between layers.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

int
Mac_GetMinSpaceOfLayLay(layer1,layer2)
int layer1;
int layer2;
{
#ifdef DEBUG
    if (layer1 < 0 || layer1 > mac_NumOfLayer ||
        layer2 < 0 || layer2 > mac_NumOfLayer )
    {
        MacPrintFatalErr("Mac_GetMinSpaceOfLayLay(): invalid layer number");
    }
#endif DEBUG

    return (macLayLayMtrx[layer1][layer2].minSpacing);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_GetMinOverlapOfLayHole()  --
 *
 *    Get minimum overlaping between layer and hole.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

int
Mac_GetMinOverlapOfLayHole(layer,hole)
int layer;
int hole;
{
#ifdef DEBUG
    if (layer < 0 || layer > mac_NumOfLayer)
    {
        MacPrintFatalErr("Mac_GetMinOverlapOfLayHole(): invalid layer number");
    }
    if (hole  < 0 || hole >= mac_NumOfLayer )
    {
        MacPrintFatalErr("Mac_GetMinOverlapOfLayHole(): invalid hole number");
    }
#endif DEBUG

    return (macLayHoleMtrx[layer][hole].minOverlap);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutMinWidthOfLay()  --
 *
 *    Put minimum width of layer
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Mac_PutMinWidthOfLay(minWidth, layer)
int minWidth;
int layer;
{
#ifdef DEBUG
    if (layer < 0 || layer > mac_NumOfLayer )
    {
        MacPrintFatalErr("Mac_PutMinWidthOfLay(): invalid layer number");
    }
#endif DEBUG

    macLayVect[layer].minWidth = minWidth;
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutSizeOfHole()  --
 *
 *   Put size of hole.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Mac_PutSizeOfHole(size, hole)
int size;
int hole;
{
#ifdef DEBUG
    if (hole < 0 || hole >= mac_NumOfLayer )
    {
        MacPrintFatalErr("Mac_PutSizeOfHole(): invalid hole number");
    }
#endif DEBUG

    macHoleVect[hole].size = size;
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutMinSpaceOfLayLay()  --
 *
 *    Put minimum spacing between layers.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Mac_PutMinSpaceOfLayLay(minSpace, layer1, layer2)
int minSpace;
int layer1;
int layer2;
{
#ifdef DEBUG
    if (layer1 < 0 || layer1 > mac_NumOfLayer ||
        layer2 < 0 || layer2 > mac_NumOfLayer )
    {
        MacPrintFatalErr("Mac_PutMinSpaceOfLayLay(): invalid layer number");
    }
#endif DEBUG

    macLayLayMtrx[layer1][layer2].minSpacing = minSpace;
}



/*
 * ----------------------------------------------------------------------------
 *
 * Mac_PutMinOverlapOfLayHole()  --
 *
 *    Put minimum overlaping between layer and hole.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Mac_PutMinOverlapOfLayHole(minOverlap, layer, hole)
int minOverlap;
int layer;
int hole;
{
#ifdef DEBUG
    if (layer < 0 || layer > mac_NumOfLayer)
    {
        MacPrintFatalErr("Mac_PutMinOverlapOfLayHole(): invalid layer number");
    }
    if (hole  < 0 || hole >= mac_NumOfLayer )
    {
        MacPrintFatalErr("Mac_PutMinOverlapOfLayHole(): invalid hole number");
    }
#endif DEBUG

    macLayHoleMtrx[layer][hole].minOverlap = minOverlap;
}



