
/* nclEvaluate.c --
 *
 *    Bernhard Eschermann / Massoud Pedram
 *    (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1988 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclEvaluate.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "newCInt.h"



/*
 * ----------------------------------------------------------------------------
 *
 * NclEvaluateClustering --
 *  
 *    Computes and prints out information to assess the quality of the
 *    produced clustering.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The global variables interConn and intraConn are changed.
 *
 * ----------------------------------------------------------------------------
 */

void
NclEvaluateClustering (NumClusters, clusterNodes)
int         NumClusters;
GraphNode   *clusterNodes;
{
    extern double interConn, intraConn;
    extern int ncl_ClusterTreeDepth;

    GraphNode   *gNodePtr;
    int         intraNum = 0, interNum = 0;
    int         numChildren;

    if (interConn+intraConn == 0) {
        return;
    }

    fprintf(stdout, "Percentage of intra-cluster connections: %.1f%%\n",
            intraConn / (interConn+intraConn) * 100);
    fflush(stdout);

    if (NumClusters > 1) {
        for (gNodePtr = clusterNodes; gNodePtr != NIL(GraphNode); 
                NEXT(gNodePtr)) {
            numChildren = gNodePtr->gn_numOfChildren;
            intraNum += numChildren * (numChildren-1) / 2;
            interNum += numChildren;
        }

        interNum = interNum * (interNum-1) / 2 - intraNum;
        intraConn /= intraNum;
        interConn /= interNum;

        printf("Average intra-cluster connectivity: %.2f\n", intraConn);
        printf("Average inter-cluster connectivity: %.2f\n", interConn);
        printf("Average connectivity score: %.1f%%\n\n",
                intraConn / (interConn+intraConn) * 100);
        fflush(stdout);
    }
}

