
/* nclFShapes.c --
 * 
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1989 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclFShapes.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclFShapes.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newCInt.h"



static void
AssignIDs(shapeSet, label, pattern, orient)
Ncl_ShapeSet *shapeSet;
int       *label;
int       pattern, orient;
{
    int i;
    for (i=0; i < shapeSet->size; i++) {
        shapeSet->shape[i].pattern = pattern;
        shapeSet->shape[i].label = label;
        shapeSet->shape[i].orient = orient;
    }
}


/*
 * --------------------------------------------------------------------------
 *
 * NclShapeFunc --
 *
 *    Computes the composite shape function of each floorplan template.
 *    'label' specifies the assignment of blocks to rooms, and
 *    'orient' specifies the orientation of template.
 *
 * Results:
 *    Returns non-adjacency cost.
 *
 * Side Effects:
 *    None.
 *
 * --------------------------------------------------------------------------
 */


Ncl_ShapeSet *
NclShapeFunc0(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    Ncl_ShapeSet    *mShapeSet, *sumShapeSet;
    int             i;

    mShapeSet = (clNode->gn_children[0])->gn_shapeSet;
    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = 0;
    }

    MALLOC(Ncl_ShapeSet *, sumShapeSet, sizeof (Ncl_ShapeSet));
    sumShapeSet->shape = Ncl_InitShapes(mShapeSet->size);
    for (i=0; i < mShapeSet->size; i++) {
        sumShapeSet->shape[i].sd_x = mShapeSet->shape[i].sd_x;
        sumShapeSet->shape[i].sd_y = mShapeSet->shape[i].sd_y;
        sumShapeSet->shape[i].pattern = 0;
        sumShapeSet->shape[i].belongsTo = 0;
        sumShapeSet->shape[i].orient = orient;
        sumShapeSet->shape[i].label = label;
        sumShapeSet->shape[i].source[0] = &(mShapeSet->shape[i]);
    }
    sumShapeSet->size = mShapeSet->size;

    return (sumShapeSet);
}


Ncl_ShapeSet *
NclShapeFunc1(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    Ncl_CutDir      cut_horz, cut_vert;
    Ncl_ShapeSet    *sumShapeSet;
    Ncl_ShapeSet    *mShapeSet, *nShapeSet;
    int             m=label[0];
    int             n=label[1];
    int             i;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }
    sumShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);

    AssignIDs(sumShapeSet, label, 1, orient);

    return (sumShapeSet);
}
    

Ncl_ShapeSet *
NclShapeFunc2(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    Ncl_CutDir      cut_horz, cut_vert;
    Ncl_ShapeSet    *lShapeSet, *sumShapeSet;
    Ncl_ShapeSet    *mShapeSet, *nShapeSet, *pShapeSet;
    int             m=label[0];
    int             n=label[1];
    int             p=label[2];
    int             i;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_horz);
    sumShapeSet = Ncl_AddShapeSets(lShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 2, orient);

    return (sumShapeSet);
}


Ncl_ShapeSet *
NclShapeFunc3(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);
    sumShapeSet = Ncl_AddShapeSets(lShapeSet, pShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 3, orient);

    return (sumShapeSet);
}


Ncl_ShapeSet *
NclShapeFunc4(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

        

    lShapeSet = Ncl_AddShapeSets(mShapeSet, pShapeSet, cut_horz);
    rShapeSet = Ncl_AddShapeSets(nShapeSet, qShapeSet, cut_horz);
    sumShapeSet = Ncl_AddShapeSets(lShapeSet, rShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 4, orient);

    return (sumShapeSet);
}


Ncl_ShapeSet *
NclShapeFunc5(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, qShapeSet, cut_horz);
    sumShapeSet = Ncl_AddShapeSets(rShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 5, orient);

    return (sumShapeSet);
}
    

Ncl_ShapeSet *
NclShapeFunc6(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(pShapeSet, qShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, nShapeSet, cut_horz);
    sumShapeSet = Ncl_AddShapeSets(rShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 6, orient);

    return (sumShapeSet);
}


Ncl_ShapeSet *
NclShapeFunc7(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_horz);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, mShapeSet, cut_vert);
    sumShapeSet = Ncl_AddShapeSets(rShapeSet, qShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 7, orient);

    return (sumShapeSet);
}
    

Ncl_ShapeSet *
NclShapeFunc8(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(pShapeSet, qShapeSet, cut_horz);
    sumShapeSet = Ncl_AddShapeSets(lShapeSet, rShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 8, orient);

    return (sumShapeSet);
}
    

Ncl_ShapeSet *
NclShapeFunc9(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_horz);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, qShapeSet, cut_horz);
    sumShapeSet = Ncl_AddShapeSets(rShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 9, orient);

    return (sumShapeSet);
}
    

Ncl_ShapeSet *
NclShapeFunc10(clNode, label, orient)
GraphNode  *clNode;
int        *label;
int        orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sumShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (clNode->gn_children[m])->gn_shapeSet;
    nShapeSet = (clNode->gn_children[n])->gn_shapeSet;
    pShapeSet = (clNode->gn_children[p])->gn_shapeSet;
    qShapeSet = (clNode->gn_children[q])->gn_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(pShapeSet, qShapeSet, cut_vert);
    sumShapeSet = Ncl_AddShapeSets(lShapeSet, rShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sumShapeSet, label, 10, orient);

    return (sumShapeSet);
}
    
