
/* nclInputTree.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Read in a cluster tree.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclInputTree.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "newC.h"
#include "newCInt.h"
#include "st.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */


/* ---------- Local Procedures ---------- */
static Ncl_InputTree    *MakeInputTree();
static Ncl_InputNode    *MakeInputNode();
static int              CalcInputNodeLevel();
static void             FreeInputNode();

#ifdef DEBUG
static void             PrintTable();
#endif DEBUG

/* ---------- Local Global Variables ---------- */
#ifdef DEBUG
static bool             printFlag = FALSE;
#endif DEBUG


static Ncl_InputTree *
MakeInputTree()
{
    Ncl_InputTree *inputTreePtr;

    MALLOC(Ncl_InputTree *, inputTreePtr, sizeof (Ncl_InputTree));
    inputTreePtr->it_rootPtr = NIL(Ncl_InputNode);
    inputTreePtr->it_depth = 0;

    return (inputTreePtr);
}


static Ncl_InputNode *
MakeInputNode()
{
    Ncl_InputNode   *inputNodePtr;
    int             i;

    MALLOC(Ncl_InputNode *, inputNodePtr, sizeof (Ncl_InputNode));
    inputNodePtr->in_pos = 0;
    inputNodePtr->in_name = NIL(char);
    inputNodePtr->in_level = 0;

    inputNodePtr->in_isRootFlag = FALSE;
    inputNodePtr->in_isParentFlag = FALSE;
    inputNodePtr->in_isChildFlag = FALSE;
    inputNodePtr->in_parentPtr = NIL(Ncl_InputNode);

    inputNodePtr->in_numOfChildren = 0;
    for (i = 0; i < MAX_CHILDREN; i++) {
        inputNodePtr->in_child[i] = NIL(Ncl_InputNode);
    }
    inputNodePtr->in_cellInstPtr = NIL(CellInst);

    return (inputNodePtr);
}


static void
FreeInputNode(item)
lsGeneric item;
{
    Ncl_InputNode *inputNodePtr = (Ncl_InputNode *) item;

    if (inputNodePtr->in_name != NIL(char)) {
        FREE((char *) inputNodePtr->in_name);
    }

    FREE((char *) inputNodePtr);
}


void
NclFreeInputTree(inputTreePtr)
Ncl_InputTree *inputTreePtr;
{
    Ncl_InputNode   *nodePtr, *childPtr, *rootPtr;
    lsList          bfsList;    /* List of nodes in bread first search order. */
    lsList          levelList;
    int             i;

    if (inputTreePtr == NIL(Ncl_InputTree)) {
        return;
    }

    if ((rootPtr = inputTreePtr->it_rootPtr) != NIL(Ncl_InputNode)) {
        levelList = lsCreate();
        lsNewEnd(levelList, (lsGeneric) rootPtr, LS_NH);
        bfsList = lsCreate();
        while (TRUE) {
            if (lsDelBegin(levelList, (lsGeneric *) &nodePtr) != LS_OK) {
                break;
            }
            lsNewEnd(bfsList, (lsGeneric) nodePtr, LS_NH);

            for (i = 0; i < nodePtr->in_numOfChildren; i++) {
                childPtr = nodePtr->in_child[i];
                lsNewEnd(levelList, (lsGeneric) childPtr, LS_NH);
            }
        }
        lsDestroy(levelList, (void (*)()) NULL);

        lsDestroy(bfsList, FreeInputNode);
    }

    FREE((char *) inputTreePtr);
}


static int
CalcInputNodeLevel(inputNodePtr)
Ncl_InputNode *inputNodePtr;
{
    if (inputNodePtr->in_parentPtr == NIL(Ncl_InputNode)) {
        return (0);
    }

    return (CalcInputNodeLevel(inputNodePtr->in_parentPtr) + 1);
}


BearStatus
NclGenLevelLists(inputTreePtr, graphNodeList, treeListPtr)
Ncl_InputTree   *inputTreePtr;
GraphNode       *graphNodeList;
lsList          *treeListPtr;
/*
 * The position field of the structure Ncl_InputNode (in_pos) is
 * filled in here.
 */
{
    GraphNode       *graphNodePtr;
    Ncl_InputNode   *inputNodePtr, *childPtr, *rootPtr;
    lsGen           levelGen, treeGen;
    lsList          levelList, newLevelList;
    lsList          treeList;
    int             i, pos, count;

    if ((rootPtr = inputTreePtr->it_rootPtr) == NIL(Ncl_InputNode)) {
        Bear_PrintErr("GenLevelLists() : Input node cluster tree is empty.");
        return (BEAR_ERROR);
    }

    for (graphNodePtr = graphNodeList, pos = 0;
        graphNodePtr != NIL(GraphNode); 
        NEXT(graphNodePtr), pos++) {

        graphNodePtr->gn_cellInstPtr->nclModulePtr = (BearGeneric) pos;
    }

    levelList = lsCreate();
    lsNewEnd(levelList, (lsGeneric) rootPtr, LS_NH);

    treeList = lsCreate();
    while (TRUE) {
        lsNewBegin(treeList, (lsGeneric) levelList, LS_NH);

        pos = 0;
        newLevelList = lsCreate();
        lsForeachItem(levelList, levelGen, inputNodePtr) {
            inputNodePtr->in_pos = pos++;

            for (i = 0; i < inputNodePtr->in_numOfChildren; i++) {
                childPtr = inputNodePtr->in_child[i];

                if (childPtr->in_numOfChildren != 0) {
                    lsNewEnd(newLevelList, (lsGeneric) childPtr, LS_NH);
                }
                else {
                    childPtr->in_pos = 
                        (int) childPtr->in_cellInstPtr->nclModulePtr;
                }
            }
        }
        if (lsLength(newLevelList) == 0) {
            lsDestroy(newLevelList, (void (*)()) NULL);
            break;
        }
        levelList = newLevelList;
    }

    for (graphNodePtr = graphNodeList;
        graphNodePtr != NIL(GraphNode);
        NEXT(graphNodePtr)) {

        graphNodePtr->gn_cellInstPtr->nclModulePtr = (BearGeneric) NULL;
    }
    *treeListPtr = treeList;

    return (BEAR_OK);
}


#ifdef DEBUG
static void
PrintTable(tablePtr)
st_table *tablePtr;
{
    Ncl_InputNode   *inputNodePtr;
    st_generator    *genPtr;
    int             i;
    char            *keyPtr, *typePtr;

    st_foreach_item(tablePtr, genPtr, &keyPtr, (char **) &inputNodePtr) {
        if (inputNodePtr->in_numOfChildren != 0) {
            if (inputNodePtr->in_isRootFlag == TRUE) {
                (void) fprintf (bearout, "ROOT ");
            }
            (void) fprintf (bearout, "cluster : %s\n", keyPtr);

            for (i = 0; i < inputNodePtr->in_numOfChildren; i++) {
                (void) fprintf (bearout, "%s\n",
                    inputNodePtr->in_child[i]->in_name);
            }
        }
        else {
            (void) fprintf (bearout, "LEAF %s\n", inputNodePtr->in_name);
        }

        (void) fprintf (bearout, "\n");
    }
    fflush (bearout);
}
#endif DEBUG


/*
 * ----------------------------------------------------------------------------
 *
 * NclReadCTree() --
 *
 * The data is checked to make sure that the tree contains no loops
 * or disconnected nodes and that it is in fact a tree.  For an example
 * of its exemplary efforts along these lines you can run the new cluster
 * command with the file in option using the data in badCTree1. 
 *
 * The position field of the structure Ncl_InputNode (in_pos) is not
 * filled in here.
 *
 * Results:
 *    BEAR_OK is returned if everything is spiffy.
 *    BEAR_ERROR_2 if the file could not be opened for reading.
 *    BEAR_ERROR otherwise.
 *
 * Side effects:
 *    None.
 * 
 * ----------------------------------------------------------------------------
 */

BearStatus 
NclReadCTree(numOfCells, ctreeFileNamePtr, inputTreePtrPtr)
int             numOfCells;         /* Number of cells expected to be found
                                     * in the file.
                                     */
char            *ctreeFileNamePtr;
Ncl_InputTree   **inputTreePtrPtr;
{
#define MAX_LINE_SIZE   1024

#define RETURN_ERR \
{ \
    bearStatus = BEAR_ERROR; \
    goto fin; \
}

    FILE            *filePtr = NIL(FILE);
    BearStatus      bearStatus = BEAR_OK;
    CellInst        *cellInstPtr;
    Ncl_InputNode   *rootNodePtr = NIL(Ncl_InputNode);
    Ncl_InputNode   *parentPtr, *childPtr;
    Ncl_InputNode   *inputNodePtr;
    lsGen           gen;
    lsList          cellList = (lsList) NULL;
    st_generator    *genPtr;
    st_table        *tablePtr = (st_table *) NULL;
    int             i, errCount, cellLevel;
    bool            clusterFlag, eofFlag, isRootFlag;
    char            line[MAX_LINE_SIZE];
    char            *parentNamePtr, *childNamePtr;
    char            *keyWordPtr;
    char            *keyPtr;

    *inputTreePtrPtr = NIL(Ncl_InputTree);

#ifdef USE_OCT_CODE
    (void) VOVinput(VOV_UNIX_FILE, ctreeFileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (ctreeFileNamePtr, "r")) == NIL(FILE)) {
        (void) sprintf(bear_ErrorMsg,
            "Could not open cluster tree file named \"%s\" for reading.",
            ctreeFileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR_2);
    }

    errCount = 0;
    tablePtr = st_init_table(strcmp, st_strhash);
    eofFlag = FALSE;
    while (TRUE) {
        if (Bear_ReadLine(line, MAX_LINE_SIZE, filePtr, &eofFlag) ==
            BEAR_ERROR) {
            RETURN_ERR;
        }
        if (eofFlag == TRUE) {
            break;
        }

        clusterFlag = FALSE;

        keyWordPtr = strtok (line, " ");
        if (strcmp (keyWordPtr, "Version") == 0) {
            /* Someday we may want to verify the version number. */
            continue;
        }
        else if (strcmp (keyWordPtr, "RootCluster") == 0) {
            if (rootNodePtr != NIL(Ncl_InputNode)) {
           Bear_PrintErr("NclReadCTree() : Cannot have two root nodes.");
                RETURN_ERR;
            }

            isRootFlag = TRUE;
            clusterFlag = TRUE;
        }
        else if (strcmp (keyWordPtr, "Cluster") == 0) {
            isRootFlag = FALSE;
            clusterFlag = TRUE;
        }
        else {
            Bear_PrintErr("NclReadCTree() failed : unknown keyword.");
            RETURN_ERR;
        }

        if (clusterFlag == TRUE) {
            if ((parentNamePtr = strtok (NIL(char), " ")) == NIL(char)) {
              Bear_PrintErr("NclReadCTree() : Cluster name is missing.");
                RETURN_ERR;
            }

            if (st_lookup(tablePtr, parentNamePtr, (char **) &parentPtr) == 1) {
                if (parentPtr->in_isParentFlag == TRUE) {
                    (void) sprintf (bear_ErrorMsg, 
                        "\"%s\" is used as a parent more than once.",
                        parentNamePtr);
                    Bear_PrintErr(bear_ErrorMsg);
                    errCount++;
                }
            }
            else {
                parentPtr = MakeInputNode();
                parentPtr->in_isRootFlag = isRootFlag;
                parentPtr->in_name = util_strsav(parentNamePtr);
                (void) st_insert(tablePtr, parentPtr->in_name, 
                    (char *) parentPtr);
            }
            parentPtr->in_isParentFlag = TRUE;

            i = 0;
            while (TRUE) {
                if (Bear_ReadLine(line, MAX_LINE_SIZE, filePtr, &eofFlag) ==
                    BEAR_ERROR) {
                    RETURN_ERR;
                }
                if (eofFlag == TRUE) {
           Bear_PrintErr("NclReadCTree() failed : eof detected to soon.");
                    RETURN_ERR;
                }

                if (line[0] == ';') {
                    break;
                }

                childNamePtr = strtok (line, " ");
                if (i >= MAX_CHILDREN) {
                    (void) sprintf (bear_ErrorMsg,
                        "\"%s\" has too many children.", parentNamePtr);
                    Bear_PrintErr(bear_ErrorMsg);
                    errCount++;
                    continue;
                }

                if (st_lookup(tablePtr, childNamePtr,
                    (char **) &childPtr) == 1) {

                    if (childPtr->in_isRootFlag == TRUE) {
                   Bear_PrintErr("The root node may not be used as a child !");
                        errCount++;
                    }
                    else if (childPtr->in_isChildFlag == TRUE) {
                        (void) sprintf (bear_ErrorMsg, 
                            "\"%s\" is used as a child more than once.",
                            childNamePtr);
                        Bear_PrintErr(bear_ErrorMsg);
                        errCount++;
                    }
                }
                else {
                    childPtr = MakeInputNode();
                    childPtr->in_isRootFlag = FALSE;
                    childPtr->in_name = util_strsav(childNamePtr);
                    childPtr->in_parentPtr = parentPtr;
                    (void) st_insert(tablePtr, childPtr->in_name, 
                        (char *) childPtr);
                }
                childPtr->in_isChildFlag = TRUE;

                parentPtr->in_child[i] = childPtr;
                i++;
            }
            if ((parentPtr->in_numOfChildren = i) == 0) {
                (void) sprintf (bear_ErrorMsg, "\"%s\" has no children.",
                    parentNamePtr);
                Bear_PrintErr(bear_ErrorMsg);
                errCount++;
            }

            if (isRootFlag == TRUE) {
                rootNodePtr = parentPtr;
            }
        }
    }

    if (st_count(tablePtr) == 0) {
        (void) sprintf (bear_Msg, "No cluster tree in file named \"%s\"\n",
            ctreeFileNamePtr);
        Bear_Print(bear_Msg);
        RETURN_ERR;
    }

    cellList = lsCreate();
    st_foreach_item(tablePtr, genPtr, &keyPtr, (char **) &inputNodePtr) {
        if (inputNodePtr->in_numOfChildren == 0) {
            if (Mac_IsMember(CELL_INSTANCE_HT, inputNodePtr->in_name) == 
                FALSE) {
                (void) sprintf (bear_ErrorMsg, "Cannot find cell named \"%s\"", 
                    inputNodePtr->in_name);
                Bear_PrintErr(bear_ErrorMsg);
                errCount++;
            }
            else { 
                inputNodePtr->in_cellInstPtr =
                    (CellInst *) Mac_LookUp(CELL_INSTANCE_HT,
                    inputNodePtr->in_name);

                lsNewEnd(cellList, (lsGeneric) inputNodePtr, LS_NH);
            }
        }
        else {
            if ((inputNodePtr->in_isChildFlag == FALSE) &&
                (inputNodePtr != rootNodePtr)) {
                (void) sprintf (bear_ErrorMsg, "\"%s\" has no parent.",
                    inputNodePtr->in_name);
                Bear_PrintErr(bear_ErrorMsg);
                errCount++;
            }
        }
    }

    if (lsLength(cellList) != numOfCells) {
        Bear_PrintErr("Not all of the cells were used.");
        errCount++;
    }
    if (errCount > 0) {
        RETURN_ERR;
    }

    /*
     * Must not do the level calculation until we have verified that the
     * tree has no loops.
     */
    lsForeachItem(cellList, gen, inputNodePtr) {
        cellLevel = inputNodePtr->in_level = CalcInputNodeLevel(inputNodePtr);
    }
    lsForeachItem(cellList, gen, inputNodePtr) {
        if (inputNodePtr->in_level != cellLevel) {
            Bear_PrintErr("Cells are not all at the same level.");
            lsFinish(gen);
            RETURN_ERR;
        }
    }

fin:
    if (filePtr != NIL(FILE)) {
        (void) fclose (filePtr);
    }
    if (cellList != (lsList) NULL) {
        lsDestroy(cellList, (void (*)()) NULL);
    }

    if (bearStatus == BEAR_OK) {
#ifdef DEBUG
        if (printFlag == TRUE) {
            PrintTable(tablePtr);
        }
#endif DEBUG

        (void) sprintf (bear_Msg,
            "Reading cluster tree from file named \"%s\"\n", ctreeFileNamePtr);
        Bear_Print(bear_Msg);
        fflush (bearout);

        *inputTreePtrPtr = MakeInputTree();
        (*inputTreePtrPtr)->it_rootPtr = rootNodePtr;
        (*inputTreePtrPtr)->it_depth = cellLevel;
    }
    else {
        (void) fprintf (stderr, "\n");
        (void) sprintf (bear_ErrorMsg,
            "Unable to generate cluster tree for file named \"%s\"\n",
            ctreeFileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
    }

    if (tablePtr != (st_table *) NULL) {
        st_free_table(tablePtr);
    }

    return (bearStatus);

#undef RETURN_ERR
#undef MAX_LINE_SIZE
}


