
/* nclLengthsUtil.c --
 *
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1989 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclLengthsUtil.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "newCInt.h"

#include "bearCopyRight.h"
/* Global local variables */



/* 
 * This function sums up wire length `used' within nd_children of the 
 * gNodePtr. In addition, it calculates the inNode field of the critical 
 * nets. 
 */

float
NclGetIntraBlockL(gNodePtr, shapePtr)
GraphNode       *gNodePtr;
Ncl_ShapeData   *shapePtr;
{
    GraphNode               *childGNodePtr;
    Ncl_LengthSet           *childLengthSet;
    Ncl_LinkedCritNetInNode *critInNodePtr;
    Ncl_LinkedCritNet       *lnkCritNetPtr;
    Ncl_ShapeData           *childShapePtr;
    float                   minArea, maxArea, interval, shapeArea, intraLength;
    int                     i, j, numOfLengths;

    intraLength = 0;

    if (ncl_TimingFlag == TRUE) {

        for (lnkCritNetPtr = ncl_CritNetList;
            lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
            NEXT(lnkCritNetPtr)) {

            lnkCritNetPtr->inNode.intra_l = 0;
        }

    } /* end if ncl_TimingFlag */

    for (i = 0; i < gNodePtr->gn_numOfChildren; i++) {
        childShapePtr = shapePtr->source[i];
        childGNodePtr = gNodePtr->gn_children[childShapePtr->belongsTo];
        childLengthSet = childGNodePtr->gn_lengthSet;
        numOfLengths = childLengthSet->size;

        minArea = childLengthSet->length[0].a;
        maxArea = childLengthSet->length[numOfLengths - 1].a;
        shapeArea = childShapePtr->sd_x * childShapePtr->sd_y;

        interval = (maxArea - minArea) / numOfLengths;
        j = (interval==0) ? 
            0 : (int) ((shapeArea - minArea) / interval);

        if (j > numOfLengths - 1) {
            j = numOfLengths - 1;
        } 
        else if (j < 0) {
            j = 0;
        }

        intraLength += childLengthSet->length[j].l;

        if (ncl_TimingFlag == TRUE) {
            for (critInNodePtr = childGNodePtr->gn_lnkCritNetList;
                    critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                    NEXT(critInNodePtr)) {

                childLengthSet = &(critInNodePtr->ld);

                if (numOfLengths == 1) {

                   critInNodePtr->critNetPtr->inNode.intra_l += 
                            childLengthSet->length[0].l;

                } else if (numOfLengths > 1) {

                    minArea = childLengthSet->length[0].a;
                    maxArea = childLengthSet->length[numOfLengths - 1].a;
                    interval = (maxArea - minArea) / numOfLengths;
                    j = (interval==0) ? 
                            0 : (int) ((shapeArea - minArea) / interval);

                    if (j > numOfLengths - 1) {
                        j = numOfLengths - 1;
                    } else if (j < 0) {
                        j = 0;
                    }

                    critInNodePtr->critNetPtr->inNode.intra_l += 
                            childLengthSet->length[j].l;
                } /* else do nothing. */

            } /* for critical nets in node */

        } /* end if ncl_TimingFlag */

    } /* for all children of gNodePtr */

    return (intraLength);
}


/* 
 * This function adds length of wires that connect children of the gNodePtr.
 * In addition, it calculates the inNode field of the critical nets.
 * we use HALF PERIMETER LENGTH estimates for CRITICAL NETS.
 */

float
NclGetInterBlockL(gNodePtr, holder)
GraphNode       *gNodePtr;
MatrixHolder    *holder;
{
    ConnEdge                *connEdgePtr;
    Ncl_LinkedCritNet       *lnkCritNetPtr;
    Ncl_LinkedCritNetOnEdge *critOnEdgePtr, *lnkOnEdgePtr;
    Point                   (*matrix)[3];
    Rect                    *critRectPtr; 
    Rect                    nodeRect, touchRect, newRect, area, newArea;
    float                   interLength, len;
    int                     i, j, fnd1, fnd2, tempStor, 
                            extension[4], horzExtension, vertExtension;

    matrix = holder->matrix;

    if (ncl_TimingFlag == TRUE) {
        /* 
         * We initialize inNode fields of critical nets. 
         */
        for (lnkCritNetPtr = ncl_CritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {

            lnkCritNetPtr->inNode.inter_l = 0;

            for (i=0; i < gNodePtr->gn_numOfChildren; i++) {
                lnkCritNetPtr->inNode.connectsChild[i] = FALSE;
            }

            lnkCritNetPtr->inNode.hasInternalEdge = 
                    lnkCritNetPtr->inNode.hasExternalEdge = FALSE;
        }
    } /* end if ncl_TimingFlag */

    /*
     * Enclosing rect calculation.
     */
    nodeRect.r_xbot = nodeRect.r_ybot = INFINITY;
    nodeRect.r_xtop = nodeRect.r_ytop = MINFINITY;
    for (i = 0; i < gNodePtr->gn_numOfChildren; i++) {
        nodeRect.r_xbot = MIN(nodeRect.r_xbot, matrix[i][LOWER].p_x);
        nodeRect.r_ybot = MIN(nodeRect.r_ybot, matrix[i][LOWER].p_y);
        nodeRect.r_xtop = MAX(nodeRect.r_xtop, matrix[i][UPPER].p_x);
        nodeRect.r_ytop = MAX(nodeRect.r_ytop, matrix[i][UPPER].p_y);
    }

    interLength = 0;
    for (i = 0; i < gNodePtr->gn_numOfChildren; i++) {

        for (connEdgePtr = gNodePtr->gn_children[i]->edgeList; 
                connEdgePtr != NIL(ConnEdge); 
                NEXT(connEdgePtr)) {
            GraphEdge *edgePtr = connEdgePtr->edgePtr;

            if (edgePtr->edgeType == BLOCK &&
                    edgePtr->node1Ptr->belongsToPtr == gNodePtr &&
                    edgePtr->node2Ptr->belongsToPtr == gNodePtr) {

                for (j=0; j < gNodePtr->gn_numOfChildren; j++) {

                    if (gNodePtr->gn_children[j] == edgePtr->node1Ptr) {
                        fnd1 = j;
                    }

                    if (gNodePtr->gn_children[j] == edgePtr->node2Ptr) {
                        fnd2 = j;
                    }

                }

                if (ncl_ClusterTreeDepth > 1) {
                    /*
                     * Touching rect calculation for non_leaf nodes.
                     */
                    touchRect.r_xbot = 
                            MIN(matrix[fnd1][UPPER].p_x, matrix[fnd2][UPPER].p_x);
                    touchRect.r_xtop = 
                            MAX(matrix[fnd1][LOWER].p_x, matrix[fnd2][LOWER].p_x);
                    touchRect.r_ybot = 
                            MIN(matrix[fnd1][UPPER].p_y, matrix[fnd2][UPPER].p_y);
                    touchRect.r_ytop = 
                            MAX(matrix[fnd1][LOWER].p_y, matrix[fnd2][LOWER].p_y);
                    GeoCanonicalRect(&touchRect, &newRect);

                    area.r_xbot = matrix[fnd1][MIDDLE].p_x;
                    area.r_xtop = matrix[fnd2][MIDDLE].p_x;
                    area.r_ybot = matrix[fnd1][MIDDLE].p_y;
                    area.r_ytop = matrix[fnd2][MIDDLE].p_y;
                    GeoCanonicalRect(&area, &newArea);

                    GeoClip(&newRect, &newArea);

                    len = ABS(newRect.r_xtop - newRect.r_xbot) +
                            ABS(newRect.r_ytop - newRect.r_ybot);

                } else {
                    /* 
                     * Center-to-center calculation for pseudo-leaf nodes.
                     */
                    len = ABS(matrix[fnd1][MIDDLE].p_x - matrix[fnd2][MIDDLE].p_x) + 
                            ABS(matrix[fnd1][MIDDLE].p_y - matrix[fnd2][MIDDLE].p_y);
                }
                interLength += edgePtr->connect2 * len;

if (ncl_TimingFlag == TRUE) {
    /* 
     * We check whether node contains > 1 cells from each crit net. 
     */
    for (lnkOnEdgePtr = edgePtr->critOnEdgePtr;
            lnkOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
            NEXT(lnkOnEdgePtr)) {
        lnkCritNetPtr = lnkOnEdgePtr->critNetPtr;

        if (! lnkCritNetPtr->inNode.connectsChild[fnd1]) {
            lnkCritNetPtr->inNode.connectsChild[fnd1] = TRUE;
        }

        if (! lnkCritNetPtr->inNode.connectsChild[fnd2]) {
            lnkCritNetPtr->inNode.connectsChild[fnd2] = TRUE;
        }

        if (! lnkCritNetPtr->inNode.hasInternalEdge) {
            lnkOnEdgePtr->critNetPtr->inNode.hasInternalEdge = TRUE;
        }
    }
} /* end if ncl_TimingFlag */

            } else {

if (ncl_TimingFlag == TRUE) {
    /* 
     * We check if node contains 1 cell from each crit net. 
     */
    for (lnkOnEdgePtr = edgePtr->critOnEdgePtr;
            lnkOnEdgePtr != NIL(Ncl_LinkedCritNetOnEdge); 
            NEXT(lnkOnEdgePtr)) {

        if (! lnkOnEdgePtr->critNetPtr->inNode.hasExternalEdge) {
            lnkOnEdgePtr->critNetPtr->inNode.hasExternalEdge = TRUE;
        }

    }
} /* end if ncl_TimingFlag */

            }

        } /* end for each connEdgePtr */

    } /* end for each child of gNodePtr */

    if (ncl_TimingFlag == TRUE) {

        for (lnkCritNetPtr = ncl_CritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {
            int numConnChildren = 0;

            critRectPtr = &(lnkCritNetPtr->inNode.rect);

            critRectPtr->r_xbot = critRectPtr->r_ybot = INFINITY;
            critRectPtr->r_xtop = critRectPtr->r_ytop = MINFINITY;

            area.r_xbot = area.r_ybot = INFINITY;
            area.r_xtop = area.r_ytop = MINFINITY;

            for (i = 0; i < gNodePtr->gn_numOfChildren; i++) {

                if (lnkCritNetPtr->inNode.connectsChild[i]) {
                    numConnChildren++;

                    /*
                     * Touching rect calculation for all node types.
                     * We assume that during the final optimization, critical
                     * net optimization will dictate the block orientations.
                     */
                    critRectPtr->r_xbot = 
                        MIN (critRectPtr->r_xbot, matrix[i][UPPER].p_x);
                    critRectPtr->r_ybot = 
                        MIN (critRectPtr->r_ybot, matrix[i][UPPER].p_y);
                    critRectPtr->r_xtop = 
                        MAX (critRectPtr->r_xtop, matrix[i][LOWER].p_x);
                    critRectPtr->r_ytop = 
                        MAX (critRectPtr->r_ytop, matrix[i][LOWER].p_y);

                    area.r_xbot = MIN(area.r_xbot, matrix[i][MIDDLE].p_x);
                    area.r_ybot = MIN(area.r_ybot, matrix[i][MIDDLE].p_y);
                    area.r_xtop = MAX(area.r_xtop, matrix[i][MIDDLE].p_x);
                    area.r_ytop = MAX(area.r_ytop, matrix[i][MIDDLE].p_y);
                }

            } /* for each connected child of gNodePtr. */

            if (numConnChildren > 0) {
                GeoCanonicalRect(critRectPtr, &newRect);
                GeoCanonicalRect(&area, &newArea);

                GeoClip(&newRect, &newArea);

                extension[WEST] = ABS(newRect.r_xbot - nodeRect.r_xbot);
                extension[EAST] = ABS(nodeRect.r_xtop - newRect.r_xtop);
                horzExtension = MIN(extension[WEST], extension[EAST]);

                extension[SOUTH] = ABS(newRect.r_ybot - nodeRect.r_ybot);
                extension[NORTH] = ABS(nodeRect.r_ytop - newRect.r_ytop);
                vertExtension = MIN(extension[SOUTH], extension[NORTH]);

                if (numConnChildren > 1) {
                    lnkCritNetPtr->inNode.inter_l += 
                        ABS(newRect.r_xtop - newRect.r_xbot);
                    lnkCritNetPtr->inNode.inter_l += 
                        ABS(newRect.r_ytop - newRect.r_ybot);
                }

                if (lnkCritNetPtr->inNode.hasExternalEdge) {
                    lnkCritNetPtr->inNode.inter_l += 
                        MIN(horzExtension, vertExtension);
                }
            }

        } /* end for each critical net */

    } /* end if ncl_TimingFlag */

    /* 
     * We must divide by 2 since each internal edge is been visited twice. 
     * We should not, however, divide crit net lengths by 2 since these
     * calculations are done on a net basis and not edge basis.
     */

    return (interLength / 2.0);
}


