
/* nclLowerBnds.c --
 *
 *    Massoud Pedram
 *    (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclLowerBnds.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newCInt.h"
#include "nclArrays.h"
#include "../newP/newP.h"
#include <math.h>


/* ---------- Local Procedures ---------- */
static BearStatus   InitLeafLowerBounds();
static void         AssignFieldsOfNode();
static BearStatus   AddWiringSpread();
static void         CalculateCorners();


/* ---------- Local Global Variables ---------- */
static CellInst     *checkCellInstPtr;

#ifdef DEBUG
static bool cluster_printStat = FALSE;
static bool combin_printStat = FALSE;
#endif /* DEBUG */


static BearStatus
CheckMultiPins(portPtr, inNodePtr)
PortInst        *portPtr;
Ncl_ScratchPad  *inNodePtr;
{
    CellInst *connCell = Mac_GetCellInstOfPort(portPtr);

    if (connCell == checkCellInstPtr) {
        if (inNodePtr->hasExternalEdge) {
            inNodePtr->hasInternalEdge = TRUE;
        } 
        else {
            inNodePtr->hasExternalEdge = TRUE;
        }
    }
}



static BearStatus
InitLeafLowerBounds(gNodePtr)
GraphNode *gNodePtr;
{
    Ncl_LinkedCritNetInNode *lnkCritNetList, *critInNodePtr; /* timing-driven */
    Ncl_LinkedCritNet   *lnkCritNetPtr;  /* timing-driven */
    CellMaster          *cellMasterPtr;
    CellInst            *cellInstPtr;
    GraphNode           *gChildPtr;
    Ncl_ShapeSet        *shapeSet;
    Ncl_LengthSet       *lengthSet;
    int                 numOfChildren = gNodePtr->gn_numOfChildren;
    int                 i, shapeCount;
    lsGen               shapeGen;
    Shape               macShape, *macShapePtr;
    int                 shapeNr;

    for (i = 0; i < numOfChildren; i++) {
        gChildPtr = gNodePtr->gn_children[i];
        cellInstPtr = gChildPtr->gn_cellInstPtr;
        cellMasterPtr = cellInstPtr->masterPtr;

        MALLOC(Ncl_ShapeSet *, shapeSet, sizeof (Ncl_ShapeSet));
        MALLOC(Ncl_LengthSet *, lengthSet, sizeof (Ncl_LengthSet));
        lengthSet->length = NclInitLengths();

        if (cellMasterPtr->shapeList == (lsList) NULL) {
            Bear_PrintErr("InitLeafLowerBounds() failed: NULL shapeList.");
            return (BEAR_ERROR);
        }

        shapeCount = lsLength(cellMasterPtr->shapeList);
        shapeSet->shape = Ncl_InitShapes(shapeCount);

        shapeNr = 0;
        lsForeachItem(cellMasterPtr->shapeList, shapeGen, macShapePtr) {
            shapeSet->shape[shapeNr].sd_x = macShapePtr->horzDim;
            shapeSet->shape[shapeNr].sd_y = macShapePtr->vertDim;
            shapeSet->shape[shapeNr].pattern = 0;
            if (cellMasterPtr->shapeFlag == MAC_FLEX_SHAPE) {
                shapeSet->shape[shapeNr].orient = 0;
            }
            shapeNr++;
        }
        shapeSet->size = shapeNr;

        if (cellMasterPtr->shapeFlag != MAC_FLEX_SHAPE) {

            if (shapeCount == 1) {
                shapeSet->shape[0].orient = 0;
            } 
            else {
                Mac_GetCurrentShape(cellInstPtr, &macShape);
                if ((macShape.horzDim == (int) shapeSet->shape[0].sd_x) &&
                        (macShape.vertDim == (int) shapeSet->shape[0].sd_y)) {
                    shapeSet->shape[0].orient = 0;
                    shapeSet->shape[1].orient = 1;
                }
                else if ((macShape.horzDim == (int) shapeSet->shape[1].sd_x) &&
                        (macShape.vertDim == (int) shapeSet->shape[1].sd_y)) {
                    shapeSet->shape[1].orient = 0;
                    shapeSet->shape[0].orient = 1;
                }
                else {
       Bear_PrintErr("InitLeafLowerBounds() failed: fixed cell shape mix-up.");
                    return (BEAR_ERROR);
                }
            }
        } /* end if (cellMasterPtr->shapeFlag) */

        lengthSet->size = 1;
        lengthSet->length[0].pattern = 0;
        lengthSet->length[0].a = ((float) gChildPtr->gn_dim.p_x) *
                ((float) gChildPtr->gn_dim.p_y);
        /* 
         * This is to account for nets which have multi-pins on blocks. 
         */
        lengthSet->length[0].l = 4.0 * (gChildPtr->gn_dim.p_x + 
                gChildPtr->gn_dim.p_y);
        lengthSet->length[0].orient = 0;

        if (ncl_TimingFlag) {
            lnkCritNetList = NIL(Ncl_LinkedCritNetInNode);

            checkCellInstPtr = cellInstPtr;
            for (lnkCritNetPtr = ncl_CritNetList;
                    lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                    NEXT(lnkCritNetPtr)) {

                lnkCritNetPtr->inNode.hasInternalEdge = FALSE;
                lnkCritNetPtr->inNode.hasExternalEdge = FALSE;
                (void) Mac_ItrPortOfNet(lnkCritNetPtr->netPtr, CheckMultiPins, 
                        &(lnkCritNetPtr->inNode));

                if (lnkCritNetPtr->inNode.hasInternalEdge) {
                    GetStorMc(critInNodePtr, 1, Ncl_LinkedCritNetInNode);
                    critInNodePtr->ld.length = NclInitLengths();
                    critInNodePtr->nextPtr = lnkCritNetList;
                    lnkCritNetList = critInNodePtr;
                }
            }

            for (lnkCritNetPtr = ncl_CritNetList, 
                    critInNodePtr = lnkCritNetList;
                critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                NEXT(lnkCritNetPtr)) {

                if (lnkCritNetPtr->inNode.hasInternalEdge) {
                    critInNodePtr->critNetPtr = lnkCritNetPtr;
                    critInNodePtr->ld.size = 1;
                    critInNodePtr->ld.length[0].a = 
                            ((float) gChildPtr->gn_dim.p_x) *
                            ((float) gChildPtr->gn_dim.p_y);
                    critInNodePtr->ld.length[0].l = (gChildPtr->gn_dim.p_x + 
                            gChildPtr->gn_dim.p_y) / 4.0; 
                    NEXT(critInNodePtr);
                }
            }
        } 
        else {
            lnkCritNetList = NIL(Ncl_LinkedCritNetInNode);
        }

        gChildPtr->gn_shapeSet = shapeSet;

        gChildPtr->gn_lengthSet = lengthSet;

        gChildPtr->gn_lnkCritNetList = lnkCritNetList;

        cellInstPtr->clientData = (ClientData) gNodePtr->gn_members[i];
    }

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * NclGetLowerBnds() --
 *
 *    Evaluates all the geometrical combinations to place the blocks in
 *    the given clusters. Picks the template, room assignment and orientation
 *    with the smallest area and records the nd_dim in the tree 
 *    data structure. 
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The nd_dim field of all gNodes is set.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
NclGetLowerBnds(gNodes)
GraphNode *gNodes;
{
    GraphNode       *gNodePtr;
    Ncl_Node        *nodePtr;
    int             numOfChildren, i, j, pattern, *label, labelID,
                    orient, shapeNr;
    float           intraLength, interLength, wireLength;
    Ncl_ShapeSet    shapeSetStruct;
    Ncl_LengthSet   lengthSetStruct;
    MatrixHolder    *holder;

    MALLOC(MatrixHolder *, holder, sizeof (MatrixHolder));

    nodePtr = NclMakeNode();

    if (ncl_ClusterTreeDepth == 0) {
        nodePtr->nd_type = LEAF;
    }
    else if (ncl_ClusterTreeDepth == 1) {
        nodePtr->nd_type = P_LEAF;
    }
    else {
        nodePtr->nd_type = NON_LEAF;
    }

    nodePtr->nd_status = NOT_PLACED;

    for (i = 0; i < MAX_CHILDREN; i++) {
        for (j = 0; j < MAX_CHILDREN + 4; j++) {
            nodePtr->nd_connectivity[i][j] =
            nodePtr->nd_spacing[i][j] = 0;
        }
    }

    /* initialize LB_Data fields of leaf nodes */
    for (gNodePtr = gNodes; gNodePtr != NIL(GraphNode); NEXT(gNodePtr)) {

        if (gNodePtr->gn_children[0]->gn_cellInstPtr != NIL(CellInst)) {
            if (InitLeafLowerBounds(gNodePtr) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
        }
    }

/* Generate the area, length, and critical net length lower bounds */
for (gNodePtr = gNodes; gNodePtr != NIL(GraphNode); NEXT(gNodePtr)) {

    Ncl_LinkedCritNetInNode  *glnkCritNetList, *critInNodePtr;
    Ncl_LinkedCritNet        *lnkCritNetPtr;
    Ncl_ShapeSet             *globalShapeSet, *iShapeSet, *nShapeSet;
    Ncl_LengthSet            *globalLengthSet;

    glnkCritNetList = NIL(Ncl_LinkedCritNetInNode);

    MALLOC(Ncl_ShapeSet *, globalShapeSet, sizeof (Ncl_ShapeSet));
    globalShapeSet->shape = Ncl_InitShapes(ncl_MaxNumOfShapes);
    globalShapeSet->size = 0;

    MALLOC(Ncl_LengthSet *, globalLengthSet, sizeof (Ncl_LengthSet));
    globalLengthSet->length = NclInitLengths();
    globalLengthSet->size = 0;

    shapeSetStruct.shape = Ncl_InitShapes(ncl_MaxNumOfShapes);
    shapeSetStruct.size = 0;

    lengthSetStruct.length = NclInitLengths();
    lengthSetStruct.size = 0;

    if (ncl_TimingFlag == TRUE) {
        /* 
         * We re-initialize tmp_ld and ld fields of critical nets. 
         */
        for (lnkCritNetPtr = ncl_CritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {
            lnkCritNetPtr->tmp_ld.size = 0;
            NclReInitLengths(lnkCritNetPtr->tmp_ld.length);
            lnkCritNetPtr->ld.size = 0;
            NclReInitLengths(lnkCritNetPtr->ld.length);
            lnkCritNetPtr->inNode.intra_l = 0;
        }
    } /* end if ncl_TimingFlag */

    nodePtr->nd_numOfChildren = numOfChildren = gNodePtr->gn_numOfChildren;

    for (i = 0; i < numOfChildren; i++) {
        gNodePtr->gn_members[i]->nd_id = 0;
        nodePtr->nd_children[i] = gNodePtr->gn_members[i];
    }

    for (pattern = 0; pattern < NPL_MAX_PATTERNS; pattern++) {

        if (npl_FloorplanSizeArray[pattern] != numOfChildren) continue;

        for (labelID = 0; labelID < npl_FactArray[numOfChildren-1]; labelID++) {
            label = npl_LableTable[labelID];

            /* 
             * We generate shape lower bounds. 
             */
            nShapeSet = (*(nclShapeFuncArray[pattern])) (gNodePtr, label, 0);
            iShapeSet = (gNodePtr->gn_numOfChildren == 1) ?
                    (Ncl_ShapeSet *) NULL :
                    (*(nclShapeFuncArray[pattern])) (gNodePtr, label, 1);
            if (Ncl_MergeShapeSets(iShapeSet, nShapeSet, &shapeSetStruct) == 
                BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            if (nShapeSet != NULL) {
                FREE((char *) nShapeSet->shape);
                FREE((char *) nShapeSet);
            }
            if (iShapeSet != NULL) {
                FREE((char *) iShapeSet->shape);
                FREE((char *) iShapeSet);
            }

#ifdef DEBUG
if (combin_printStat) {
    printf("\n");
    printf("Printing shape (normal template) ---");
    Ncl_PrintShapeSets(nShapeSet);
    printf("\n");
    printf("Printing shape (inverted template) ---");
    Ncl_PrintShapeSets(iShapeSet);
}
#endif DEBUG

            /* 
             * We generate length lower bounds. 
             */
            for (shapeNr=0; shapeNr < shapeSetStruct.size; shapeNr++) {
                register Ncl_ShapeData *shapePtr;

                shapePtr = &(shapeSetStruct.shape[shapeNr]);
                /*
                 * The shapeSetStruct contains both inverted and normal 
                 * template lower bounds.
                 */
                orient = shapePtr->orient;

                AssignFieldsOfNode(shapePtr, pattern, label, orient, nodePtr);

                /*
                 * We add current level connections to get the correct 
                 * nd_dim for the node.
                 */

                if (AddWiringSpread(shapePtr, gNodePtr, nodePtr) ==
                        BEAR_ERROR) {
                    return (BEAR_ERROR);
                }
                shapePtr->sd_x = nodePtr->nd_dim.p_x;
                shapePtr->sd_y = nodePtr->nd_dim.p_y;

                intraLength = NclGetIntraBlockL(gNodePtr, shapePtr);

                if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
                    return (BEAR_ERROR);
                }

                CalculateCorners(nodePtr, holder);
                interLength = NclGetInterBlockL(gNodePtr, holder);
                wireLength = intraLength + interLength;

                if (ncl_TimingFlag) {
                    for (lnkCritNetPtr = ncl_CritNetList;
                            lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                            NEXT(lnkCritNetPtr)){

                        if (lnkCritNetPtr->inNode.hasInternalEdge) {
                            Ncl_LengthData *lengthInfoPtr;

                            lengthInfoPtr = 
                                &(lnkCritNetPtr->tmp_ld.length[shapeNr]);
                            lengthInfoPtr->a = ((float) nodePtr->nd_dim.p_x) *
                                ((float) nodePtr->nd_dim.p_y);
                            lengthInfoPtr->l = lnkCritNetPtr->inNode.inter_l + 
                                lnkCritNetPtr->inNode.intra_l;
                            lengthInfoPtr->pattern = shapePtr->pattern;
                            lengthInfoPtr->label = shapePtr->label;
                            lengthInfoPtr->orient = shapePtr->orient;
                            lengthInfoPtr->belongsTo = shapePtr->belongsTo;
                        }

                    }

                } /* end if ncl_TimingFlag */

                lengthSetStruct.length[shapeNr].a = 
                        ((float) nodePtr->nd_dim.p_x) *
                        ((float) nodePtr->nd_dim.p_y);
                lengthSetStruct.length[shapeNr].l = wireLength;
                lengthSetStruct.length[shapeNr].pattern = shapePtr->pattern;
                lengthSetStruct.length[shapeNr].label = shapePtr->label;
                lengthSetStruct.length[shapeNr].orient = shapePtr->orient;
                lengthSetStruct.length[shapeNr].belongsTo = 
                        shapePtr->belongsTo;

            } /* for each shape */

            lengthSetStruct.size = shapeSetStruct.size;

            /* 
             * We MUST sort and prune lengthSetStruct which we currently 
             * don't.  MP
             */

            /* input and output sets could be the same. 13 July 90 . */
            if (Ncl_MergeShapeSets(globalShapeSet, &shapeSetStruct,
                globalShapeSet) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (NclMergeLengthSets(globalLengthSet, &lengthSetStruct,
                    globalLengthSet) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }

            if (ncl_TimingFlag == TRUE) {
                /* 
                 * We merge wire length for critical nets found within node. 
                 */
                for (lnkCritNetPtr = ncl_CritNetList;
                        lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                        NEXT(lnkCritNetPtr)) {

                    if (lnkCritNetPtr->inNode.hasInternalEdge) {

                        lnkCritNetPtr->tmp_ld.size = shapeSetStruct.size;
                        if (NclMergeLengthSets(&lnkCritNetPtr->ld, 
                                &lnkCritNetPtr->tmp_ld, 
                                &lnkCritNetPtr->ld) == BEAR_ERROR) {
                            return (BEAR_ERROR);
                        }
                    }

                } /* end for each critical net */

            } /* end if ncl_TimingFlag */

#ifdef DEBUG
    if (combin_printStat) {
        printf("\n");
        printf("Printing net shape and length ---");
        Ncl_PrintShapeSets(globalShapeSet);
        NclPrintLengthSet(globalLengthSet);
    }
#endif /* DEBUG */

        } /* end "for (label .. " */

    } /* end "for (pattern ..." */

    if (ncl_TimingFlag) {
        /* 
         * We create lnkCritNetList structure. 
         */
        for (lnkCritNetPtr = ncl_CritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {

            if (lnkCritNetPtr->inNode.hasInternalEdge ||
                    lnkCritNetPtr->inNode.hasExternalEdge) {

                GetStorMc(critInNodePtr, 1, Ncl_LinkedCritNetInNode);
                critInNodePtr->ld.length = NclInitLengths();
                critInNodePtr->nextPtr = glnkCritNetList;
                glnkCritNetList = critInNodePtr;
            }
        } /* end for each critical net */

        for (lnkCritNetPtr = ncl_CritNetList, critInNodePtr = glnkCritNetList;
                lnkCritNetPtr != NIL(Ncl_LinkedCritNet); 
                NEXT(lnkCritNetPtr)) {

            if (lnkCritNetPtr->inNode.hasInternalEdge) {

                critInNodePtr->critNetPtr = lnkCritNetPtr;
                NclCopyLengthSet(&(critInNodePtr->ld), &(lnkCritNetPtr->ld));
                NEXT(critInNodePtr);

            } else if (lnkCritNetPtr->inNode.hasExternalEdge) {

                critInNodePtr->critNetPtr = lnkCritNetPtr;
                critInNodePtr->ld.size = 1;
                /* 
                 * It is non-zero if there are multi pins of the net on the 
                 * block. 
                 */
                critInNodePtr->ld.length[0].l = 0.0;
                NEXT(critInNodePtr);

            } /* else do nothing. */

        } /* end for each critical net */

        gNodePtr->gn_lnkCritNetList = glnkCritNetList;

    } else {
        gNodePtr->gn_lnkCritNetList = NIL(Ncl_LinkedCritNetInNode);
    }

    gNodePtr->gn_shapeSet = globalShapeSet;

    gNodePtr->gn_lengthSet = globalLengthSet;

#ifdef DEBUG
    if (cluster_printStat) {
        printf("\n");
        printf("Printing net shape and length ---");
        Ncl_PrintShapeSets(globalShapeSet);
    }

    if (cluster_printStat) {
        NclPrintLengthSet(globalLengthSet);
    }
#endif /* DEBUG */

} /* end "for (gNodePtr .." */

    FREE((char *) holder);
    FREE((char *) nodePtr);

    return (BEAR_OK);
}



static BearStatus
AddWiringSpread(shapePtr, gNodePtr, nodePtr)
Ncl_ShapeData   *shapePtr;
GraphNode       *gNodePtr;
Ncl_Node        *nodePtr;
{
    Ncl_Node    *childPtr;
    Point       wSpread;
    int         i, numOfChildren = gNodePtr->gn_numOfChildren;

#ifdef DEBUG
    if (numOfChildren != nodePtr->nd_numOfChildren) {
        Bear_PrintErr("AddWiringSpread failed: inconsistent node data!");
        return (BEAR_ERROR);
    }
#endif /* DEBUG */

    /*
     * Adds pin space; similar to NplAddPinSpace.
     */
    if (! ncl_ChannelFreeFlag) {
        GraphNode   *gChildPtr;
        Point       pinCount, pinSpace;
        double      pinSpaceAdjust;

        if (nodePtr->nd_type == P_LEAF) {
            for (i = 0; i < numOfChildren; i++) {
                childPtr = nodePtr->nd_children[i];
                childPtr->nd_storeDim = childPtr->nd_dim;
            }
        }

        if (! ncl_QuickRouteEstFlag) {
            double  routeAdjust = ncl_RouteAdjust / 1.4;
            int     j, numEdges, numExtern, inConn, exConn;

            numEdges = (numOfChildren > 2) ? 
                    (numOfChildren * (numOfChildren - 1)) / 2 : 1;
            inConn = TIMES100(gNodePtr->intraConn / (double) numEdges);
            numExtern = 4 * numOfChildren;
            exConn = TIMES100(gNodePtr->interConn / (double) numExtern);

            /*
             * Only the upper right of children submatrix is filled. 
             * All entries corresaponding to IO submatrix are filled.
             */
            for (i = 0; i < MAX_CHILDREN; i++) {
                for (j = i + 1; j < MAX_CHILDREN; j++) {
                    if (j >= numOfChildren) {
                        nodePtr->nd_connectivity[i][j] = 0;
                    }
                    else {
                        nodePtr->nd_connectivity[i][j] = 
                                (int) (inConn * routeAdjust);
                    }
                }
                for (j = MAX_CHILDREN; j < MAX_CHILDREN + 4; j++) {
                    nodePtr->nd_connectivity[i][j] = 
                            (int) (exConn * routeAdjust);
                }
            }

            pinSpaceAdjust = ncl_WireWidth * ncl_RouteAdjust / 1.75;
        }
        else {
            pinSpaceAdjust = ncl_WireWidth * ncl_RouteAdjust / 1.325;
        }

        if (nodePtr->nd_type == P_LEAF) {
            for (i = 0; i < numOfChildren; i++) {
                gChildPtr = gNodePtr->gn_children[i];
                childPtr = nodePtr->nd_children[i];

                pinCount.p_x = gChildPtr->pinCount[WEST-MAX_CHILDREN] +
                        gChildPtr->pinCount[EAST-MAX_CHILDREN];
                pinCount.p_y = gChildPtr->pinCount[SOUTH-MAX_CHILDREN] +
                        gChildPtr->pinCount[NORTH-MAX_CHILDREN];

#ifdef DEBUG
    if ((childPtr->nd_nodeTrans.t_c != 0) || 
        (childPtr->nd_nodeTrans.t_f != 0)) {
        Bear_PrintErr(
            "AddWiringSpread failed: invalid nd_nodeTrans.");
        return (BEAR_ERROR);
    }
#endif /* DEBUG */

                GeoTransPoint(&childPtr->nd_nodeTrans, &pinCount, &pinSpace);

#ifdef DEBUG
{
    float x, y;

    x = ((float) pinSpace.p_x) * ((float) pinSpaceAdjust);
    if (x > 2147483647) {
        sprintf (bear_Msg, "(y = %f) is too BIG!\n", x);
        Bear_PrintFatalErr(bear_Msg);
    }

    y = ((float) pinSpace.p_y) * ((float) pinSpaceAdjust);
    if (y > 2147483647) {
        sprintf (bear_Msg, "(y = %f) is too BIG!\n", x);
        Bear_PrintFatalErr(bear_Msg);
    }
}
#endif /* DEBUG */

                childPtr->nd_dim.p_x += ((int) (pinSpace.p_x * pinSpaceAdjust));
                childPtr->nd_dim.p_y += ((int) (pinSpace.p_y * pinSpaceAdjust));
            }
        }
    }

    /*
     * Set spacing matrix according to global options.
     */
    if (Npl_GenSpace(nodePtr, (double) ncl_ClusterTreeDepth) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (! ncl_ChannelFreeFlag) {
        if (nodePtr->nd_type == P_LEAF) {
            for (i = 0; i < numOfChildren; i++) {
                childPtr = nodePtr->nd_children[i];
                childPtr->nd_dim = childPtr->nd_storeDim;
            }
        }
    }

    wSpread.p_x = nodePtr->nd_dim.p_x - shapePtr->sd_x;
    wSpread.p_y = nodePtr->nd_dim.p_y - shapePtr->sd_y;

    if ((wSpread.p_x < 0) || (wSpread.p_y < 0)) {
  Bear_PrintErr("AddWiringSpread() failed: wiring spread cannot be negative.");
        return (BEAR_ERROR);
    }

    return (BEAR_OK);

}



/*
 * Due to bottom-up procedure, nodePtr->nd_nodeTrans MUST BE initialized 
 * to GeoIdentityTransform as well.
 */
static void
AssignFieldsOfNode(shapePtr, pattern, label, orient, nodePtr)
Ncl_ShapeData   *shapePtr;
int             pattern, *label, orient;
Ncl_Node        *nodePtr;
{
    CellMaster      *cellMasterPtr;
    Ncl_Node        *childPtr;
    Ncl_ShapeData   *sourcePtr;
    int             i;

    nodePtr->nd_pattern = pattern;
    nodePtr->nd_label = label;
    nodePtr->nd_orient = orient;
    nodePtr->nd_dim.p_x = shapePtr->sd_x;
    nodePtr->nd_dim.p_y = shapePtr->sd_y;
    nodePtr->nd_orient = shapePtr->orient;
    nodePtr->nd_nodeTrans = GeoIdentityTransform;

    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        sourcePtr = shapePtr->source[i];
        childPtr = nodePtr->nd_children[sourcePtr->belongsTo];

        if (sourcePtr->orient == 0) {
            childPtr->nd_dim.p_x = sourcePtr->sd_x;
            childPtr->nd_dim.p_y = sourcePtr->sd_y;
            childPtr->nd_nodeTrans = GeoIdentityTransform;

        } 
        else {
            /*
             * sourcePtr has previously been rotated.  This fliping will 
             * restore the correct dimension for the childPtr.
             */
            childPtr->nd_dim.p_x = sourcePtr->sd_y;
            childPtr->nd_dim.p_y = sourcePtr->sd_x;
            childPtr->nd_nodeTrans = GeoRef45Transform;
        }
    }
}



static void
CalculateCorners(nodePtr, holder)
Ncl_Node        *nodePtr;
MatrixHolder    *holder;
{
    Ncl_Node    *childPtr;
    Point       relCorner, tmpUpperRight, tmpLowerLeft;
    Point       (*matrix)[3];
    int         i;

    matrix = holder->matrix;
    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        childPtr = nodePtr->nd_children[i];

        relCorner.p_x = childPtr->nd_dim.p_x / 2;
        relCorner.p_y = childPtr->nd_dim.p_y / 2;
        GeoTransPoint(&childPtr->nd_nodeTrans, &relCorner, 
            &(matrix[i][MIDDLE]));

        relCorner.p_x = relCorner.p_y = 0;
        GeoTransPoint(&childPtr->nd_nodeTrans, &relCorner, &tmpLowerLeft);

        relCorner.p_x = childPtr->nd_dim.p_x;
        relCorner.p_y = childPtr->nd_dim.p_y;
        GeoTransPoint(&childPtr->nd_nodeTrans, &relCorner, &tmpUpperRight);

        if (tmpUpperRight.p_x > tmpLowerLeft.p_x) {
            matrix[i][LOWER].p_x = tmpLowerLeft.p_x;
            matrix[i][UPPER].p_x = tmpUpperRight.p_x;
        } else {
            matrix[i][LOWER].p_x = tmpUpperRight.p_x;
            matrix[i][UPPER].p_x = tmpLowerLeft.p_x;
        }

        if (tmpUpperRight.p_y > tmpLowerLeft.p_y) {
            matrix[i][LOWER].p_y = tmpLowerLeft.p_y;
            matrix[i][UPPER].p_y = tmpUpperRight.p_y;
        } else {
            matrix[i][LOWER].p_y = tmpUpperRight.p_y;
            matrix[i][UPPER].p_y = tmpLowerLeft.p_y;
        }
    }
}

