
/* nclSort.c --
 *
 *    Bernhard Eschermann
 *    (Advising Professor - E. S. Kuh)
 *
 *    This file is necessary to be compiled with clMain.c. It contains 
 *    the procedures that do the sorting for ClusterGen.
 *
 * Copyright (C) 1987 Regents of the University of California
 * All rights reserved.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclSort.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "newCInt.h"



/* ---------- Local Procedures ---------- */
static void SortMerge();


/*
 * ----------------------------------------------------------------------------
 *
 * NclSortEdges() --
 *
 *    Called by the procedure ClusterGen in cluster.c this procedure sorts
 *    the edges for that procedure with respect to the field .sortingCrit in
 *    decreasing order. The first parameter contains the number of list elements
 *    the second is a pointer to the list that should be sorted. 
 *    The sorting is done as mergesort recursively with help of the procedure
 *    SortMerge.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The content of edgePtrPtr afterwards points to the head of the sorted
 *    list.
 *
 * ----------------------------------------------------------------------------
 */

void
NclSortEdges(edgeNumber, edgePtrPtr)
int         edgeNumber;       /* in */
GraphEdge   **edgePtrPtr;     /* in/out */
{
    GraphEdge   *part1,  *part2, *divider;
    int         i;

    if (edgeNumber <= 1) {
        /* end of recursion */
        return;
    }

    for (divider = *edgePtrPtr, i = 1; i < edgeNumber / 2; NEXT(divider), i++) ;
    /* divider points now to element in middle of list */

    part2 = divider->nextPtr;
    divider->nextPtr = NIL(GraphEdge);
    part1 = *edgePtrPtr;

    NclSortEdges(edgeNumber/2, &part1);  /* sort first half */
    NclSortEdges(edgeNumber - edgeNumber/2, &part2);  /* sort second half */
    SortMerge(part1, &part2);  /* merge the two parts again */

    *edgePtrPtr = part2;  /* result of mergesort */
}



/*
 * ----------------------------------------------------------------------------
 *
 * SortMerge() --
 *
 *    Called by NclSortEdges. Merges two sorted parts of a list to a single
 *    sorted list.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    The list to which list2Ptr points is altered to be the sorted
 *    concatenation of both input lists.
 *
 * ----------------------------------------------------------------------------
 */

static void
SortMerge(list1, list2Ptr)
GraphEdge   *list1;       /* in */
GraphEdge   **list2Ptr;   /* in/out */
{
    GraphEdge *old2Ptr, *current2Ptr, *new1Ptr;

    for (current2Ptr = *list2Ptr; 
        current2Ptr && list1;
        old2Ptr = current2Ptr, NEXT(current2Ptr)) {
        for (; list1 && list1->sortingCrit > current2Ptr->sortingCrit;
            old2Ptr = list1, list1 = new1Ptr) {

            new1Ptr = list1->nextPtr;
            list1->nextPtr = current2Ptr;

            (current2Ptr == *list2Ptr) ?
                (*list2Ptr = list1) : (old2Ptr->nextPtr = list1);
        }
    }

    if (list1) {
        old2Ptr->nextPtr = list1;
    }
}


