
/* nclUtilities.c --
 *
 *     George Carvalho, Esq.
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newC/nclUtilities.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newCInt.h"
#include "nclArrays.h"



typedef struct _CheckPinData CheckPinData;
struct _CheckPinData {
    CellInst    *cellInstPtr;
    ShapeFlag   cellFlexType;
    int         fixPinCnt;              /* fixed pin count          */
    int         floatPinCnt;            /* floating pin count       */
    int         nFloatPinCnt;           /* north floating pin count */
    int         eFloatPinCnt;
    int         sFloatPinCnt;
    int         wFloatPinCnt;
    int         totalPinCnt;            /* total pin count          */
};

typedef struct _SortData SortData;
struct _SortData {
    int         area;
    int         pinCount;
};


/* ---------- Local Procedures ---------- */
static BearStatus   CheckPin();
static BearStatus   TidyUpChildCell();
static BearStatus   CountNumOfPins();
static int          CompareCells();

#ifdef DEBUG
static void         PrintLengthData();
#endif /* DEBUG */


/* ---------- Local Global Variables ---------- */
#ifdef DEBUG
static bool echoUserParmsFlag = FALSE;
#endif DEBUG


static void
FreeShapeList(shapeList)
Ncl_LinkedBinaryMerged *shapeList;
{
    Ncl_LinkedBinaryMerged *lnkShapePtr, *nextPtr;

    lnkShapePtr = shapeList;
    while (lnkShapePtr != NIL(Ncl_LinkedBinaryMerged)) {
        nextPtr = lnkShapePtr->nextPtr;
        FREE((char *) lnkShapePtr);
        lnkShapePtr = NIL(Ncl_LinkedBinaryMerged);
        lnkShapePtr = nextPtr;
    }
}


Ncl_ShapeData *
Ncl_InitShapes(shapeCount)
int shapeCount;
{
    Ncl_ShapeData   *shape;
    int             i, j;

    MALLOC(Ncl_ShapeData *, shape, shapeCount * sizeof (Ncl_ShapeData));

    for (j = 0; j < shapeCount; j++) {
        shape[j].sd_x = shape[j].sd_y = 0.0;
        shape[j].orient = 0;
        shape[j].pattern = shape[j].belongsTo = -1;

        for (i = 0; i < MAX_CHILDREN; i++) {
            shape[j].source[i] = NIL(Ncl_ShapeData);
        }
    }

    return (shape);
}


static void
CopyShapeData(to, from)
Ncl_ShapeData *to, *from;
{
    int i;

    to->sd_x = from->sd_x;
    to->sd_y = from->sd_y;
    to->pattern = from->pattern;
    to->label = from->label;
    to->orient = from->orient;
    to->belongsTo = from->belongsTo;

    for (i = 0; i < MAX_CHILDREN; i++) {
        to->source[i] = from->source[i];
    }
}


static void
PrintShapeData(shapePtr)
Ncl_ShapeData *shapePtr;
{
    int i, labelForPrint = 0;

    if (shapePtr->label != NIL(int)) {
        for (i=0; i < MAX_CHILDREN; i++) {
            labelForPrint = 10*labelForPrint + shapePtr->label[i];
        }
    }

    (void) fprintf(bearout,
      "\nx: %.0f, y: %.0f, pattern: %d, orient: %d, label: %d, belongsTo: %d",
        shapePtr->sd_x, shapePtr->sd_y, 
        shapePtr->pattern, 
        shapePtr->orient, labelForPrint, shapePtr->belongsTo);
}


void
Ncl_PrintShapeSets(shapeSet)
Ncl_ShapeSet *shapeSet;
{
    int i, j;

    for (j = 0; j < shapeSet->size; j++) {
        PrintShapeData(&(shapeSet->shape[j]));

        /*
        printf("\nPrinting source shape ---");
        for (i=0; i < MAX_CHILDREN; i++) {
            if (shapeSet->shape[j].source[i]==NIL(Ncl_ShapeData))
                continue;
            PrintShapeData(shapeSet->shape[j].source[i]);
        }
        printf("\nEnd printing source shape ---\n");
        */

        fflush(stdout);
    }
    printf("\n");
}


#define ADD_FLINK(nShapeList, lastShape, lnkShapePtr) \
    lnkShapePtr->nextPtr = NIL(Ncl_LinkedBinaryMerged); \
    if (nShapeList == NIL(Ncl_LinkedBinaryMerged)) { \
        nShapeList = lastShape = lnkShapePtr; \
    } \
    else { \
        lastShape->nextPtr = lnkShapePtr; \
        lastShape = lastShape->nextPtr; \
    } 

#define ADD_BLINK(nShapeList, lnkShapePtr) \
    lnkShapePtr->nextPtr = nShapeList; \
    nShapeList = lnkShapePtr; 

#define ADD_LSHAPEX(nShapeList, lastShape, lnkShapePtr, lShape, rShape, li, ri) \
    MALLOC(Ncl_LinkedBinaryMerged *, \
        lnkShapePtr, sizeof (Ncl_LinkedBinaryMerged)); \
    lnkShapePtr->bm_x = lShape[li].sd_x; \
    lnkShapePtr->bm_y = lShape[li].sd_y + rShape[ri].sd_y; \
    lnkShapePtr->src1 = &(lShape[li]); \
    lnkShapePtr->src2 = &(rShape[ri]); \

#define ADD_RSHAPEX(nShapeList, lastShape, lnkShapePtr, lShape, rShape, li, ri) \
    MALLOC(Ncl_LinkedBinaryMerged *, \
        lnkShapePtr, sizeof (Ncl_LinkedBinaryMerged)); \
    lnkShapePtr->bm_x = rShape[ri].sd_x; \
    lnkShapePtr->bm_y = rShape[ri].sd_y + lShape[li].sd_y; \
    lnkShapePtr->src1 = &(lShape[li]); \
    lnkShapePtr->src2 = &(rShape[ri]); \

#define ADD_LSHAPEY(nShapeList, lastShape, lnkShapePtr, lShape, rShape, li, ri) \
    MALLOC(Ncl_LinkedBinaryMerged *, \
        lnkShapePtr, sizeof (Ncl_LinkedBinaryMerged)); \
    lnkShapePtr->bm_y = lShape[li].sd_y; \
    lnkShapePtr->bm_x = lShape[li].sd_x + rShape[ri].sd_x; \
    lnkShapePtr->src1 = &(lShape[li]); \
    lnkShapePtr->src2 = &(rShape[ri]); \

#define ADD_RSHAPEY(nShapeList, lastShape, lnkShapePtr, lShape, rShape, li, ri) \
    MALLOC(Ncl_LinkedBinaryMerged *, \
        lnkShapePtr, sizeof (Ncl_LinkedBinaryMerged)); \
    lnkShapePtr->bm_y = rShape[ri].sd_y; \
    lnkShapePtr->bm_x = rShape[ri].sd_x + lShape[li].sd_x; \
    lnkShapePtr->src1 = &(lShape[li]); \
    lnkShapePtr->src2 = &(rShape[ri]);


static Ncl_LinkedBinaryMerged *
GetShapeList(lShapeSet, rShapeSet, cutDir, shapeCount)
Ncl_ShapeSet    *lShapeSet;
Ncl_ShapeSet    *rShapeSet;
Ncl_CutDir      cutDir;
int             *shapeCount;
{
    Ncl_LinkedBinaryMerged  *lnkShapePtr, *nShapeList, *lastShape;
    Ncl_ShapeData           *lShape, *rShape, *tShapes;
    int                     lSize, rSize, tSize;
    int                     li, ri, count;

    /* build the linked list */
    lShape = lShapeSet->shape;
    rShape = rShapeSet->shape;
    lSize = lShapeSet->size;
    rSize = rShapeSet->size;

    nShapeList = lastShape = NIL(Ncl_LinkedBinaryMerged);
    count = 0;

    if (cutDir == HORZ) {
    /* want x of 1st entry in lShape to be bigger than 
       that that in rShape */

        if (rShape[0].sd_x > lShape[0].sd_x) {
            tShapes = lShape; 
            lShape = rShape; 
            rShape = tShapes;
            tSize = lSize;
            lSize = rSize;
            rSize = tSize;
        }

        li = 0;
        for (ri = 0; ri < rSize && rShape[ri].sd_x < lShape[0].sd_x; ++ri) {} 

        while (li < lSize && lShape[li].sd_x < rShape[ri].sd_x) {
            ADD_LSHAPEX(nShapeList, lastShape, lnkShapePtr, 
                lShape, rShape, li, ri-1);
            ADD_FLINK(nShapeList, lastShape, lnkShapePtr);
            count++; 
            li++;
        }

        while (li < lSize && ri < rSize) {
            if (lShape[li].sd_x < rShape[ri].sd_x) {
                ADD_LSHAPEX(nShapeList, lastShape, lnkShapePtr,
                        lShape, rShape, li, ri-1);
                ADD_FLINK(nShapeList, lastShape, lnkShapePtr);
                count++; 
                li++;
            } 
            else if (lShape[li].sd_x == rShape[ri].sd_x) {
                ADD_LSHAPEX(nShapeList, lastShape, lnkShapePtr,
                    lShape, rShape, li, ri);
                ADD_FLINK(nShapeList, lastShape, lnkShapePtr);
                count++; 
                li++;  
                ri++;
            } 
            else {
                ADD_RSHAPEX(nShapeList, lastShape, lnkShapePtr,
                    lShape, rShape, li-1, ri);
                ADD_FLINK(nShapeList, lastShape, lnkShapePtr);
                count++;
                ri++;
            }
        }

        if (li == lSize) {
            while (ri < rSize) {
        ADD_RSHAPEX(nShapeList, lastShape, lnkShapePtr,
               lShape, rShape, lSize-1, ri);
        ADD_FLINK(nShapeList, lastShape, lnkShapePtr);
        count++; ri++;
            }
        } else if (ri == rSize) {
            while (li < lSize) {
        ADD_LSHAPEX(nShapeList, lastShape, lnkShapePtr, 
               lShape, rShape, li, rSize - 1);
        ADD_FLINK(nShapeList, lastShape, lnkShapePtr);
        count++; li++;
        }    
    }
    } else {
        /* want y of 1st entry in lShape to be bigger than 
           that in rShape */

        if (rShape[rSize-1].sd_y > lShape[lSize-1].sd_y) {
            tShapes = lShape;
            lShape = rShape;
            rShape = tShapes;
            tSize = lSize;
            lSize = rSize;
            rSize = tSize;
        }

        li = lSize - 1;
    ri = rSize - 1;
    while (rShape[ri].sd_y < lShape[li].sd_y) {
        if (ri < 0) break;
        else ri--;
    }

        while (li >= 0 && lShape[li].sd_y < rShape[ri].sd_y) {
        ADD_LSHAPEY(nShapeList, lastShape, lnkShapePtr,
               lShape, rShape, li, ri + 1);
        ADD_BLINK(nShapeList, lnkShapePtr);
        count++; li--;
        }

        while (li >= 0 && ri >= 0) {
            if (lShape[li].sd_y < rShape[ri].sd_y) {
        ADD_LSHAPEY(nShapeList, lastShape, lnkShapePtr,
               lShape, rShape, li, ri + 1);
        ADD_BLINK(nShapeList, lnkShapePtr);
        count++; li--;
            } else if (lShape[li].sd_y == rShape[ri].sd_y) {
        ADD_LSHAPEY(nShapeList, lastShape, lnkShapePtr,
               lShape, rShape, li, ri);
        ADD_BLINK(nShapeList, lnkShapePtr);
        count++; li--; ri--;
            } else {
        ADD_RSHAPEY(nShapeList, lastShape, lnkShapePtr,
               lShape, rShape, li+1, ri);
        ADD_BLINK(nShapeList, lnkShapePtr);
        count++; ri--;
            }
        }

        if (li < 0) {
            while (ri >= 0) {
        ADD_RSHAPEY(nShapeList, lastShape, lnkShapePtr,
                lShape, rShape, 0, ri);
        ADD_BLINK(nShapeList, lnkShapePtr);
        count++; ri--;
            }
        } else if (ri < 0) {
            while (li >= 0) {
        ADD_LSHAPEY(nShapeList, lastShape, lnkShapePtr,
                lShape, rShape, li, 0);
        ADD_BLINK(nShapeList, lnkShapePtr);
        count++; li--;
            }
        }    
    }
    *shapeCount = count;
    return (nShapeList);
}



void
put_sources(source1, source2, count)
Ncl_ShapeData **source1, **source2;
int *count;
{
    register int i1, i2;

    i1 = *count;
    i2 = 0;
    do {  
        source1[i1] = source2[i2]; 
    i1++; i2++;
    } while (source2[i2] != NIL(Ncl_ShapeData)); 
    *count = i1;
}


static void
PutActualSources(shape, src1, src2)
Ncl_ShapeData       *shape, *src1, *src2;
{
    int    count = 0;

    if (src1->pattern == -2 && src2->pattern == -2) {
    put_sources(shape->source, src1->source, &count);
    put_sources(shape->source, src2->source, &count);
    } else if (src1->pattern == -2) {
    put_sources(shape->source,src1->source, &count);
        shape->source[count] = src2;
    } else if (src2->pattern == -2) {
    put_sources(shape->source, src2->source, &count);
        shape->source[count] = src1;
    } else {
        shape->source[count++] = src1;
        shape->source[count] = src2;
    }
}



static Ncl_ShapeSet *
BuildShapeSet(shapeList, shapeCount)
Ncl_LinkedBinaryMerged  *shapeList;
int                     shapeCount;
{
    Ncl_LinkedBinaryMerged  *current;
    Ncl_ShapeData           *shape;
    Ncl_ShapeSet            *shapeSet;
    int                     i, j, interval;

    shape = Ncl_InitShapes(shapeCount);
    current = shapeList;
    i = 0;
    while (current != NIL(Ncl_LinkedBinaryMerged)) {
        shape[i].sd_x = current->bm_x;
        shape[i].sd_y = current->bm_y;
        shape[i].pattern = -2;
        shape[i].belongsTo = -2;
        PutActualSources(&(shape[i]), current->src1, current->src2);
        current = current->nextPtr;
        i++;
    }
    
    MALLOC(Ncl_ShapeSet *, shapeSet, sizeof (Ncl_ShapeSet));
    shapeSet->shape = shape;
    shapeSet->size = i;

    return (shapeSet);
}



Ncl_ShapeSet *
Ncl_AddShapeSets(lShapeSet, rShapeSet, cutDir)
Ncl_ShapeSet    *lShapeSet, *rShapeSet;
Ncl_CutDir      cutDir;
{
    Ncl_LinkedBinaryMerged  *nShapeList;
    Ncl_ShapeData           *shapePtr;
    Ncl_ShapeSet            *addShapeSet;
    int                     i, j, shapeCount;
    
    nShapeList = GetShapeList(lShapeSet, rShapeSet, cutDir, &shapeCount);
    addShapeSet = BuildShapeSet(nShapeList, shapeCount);
    FreeShapeList(nShapeList);

    for (i = 0; i < addShapeSet->size; i++) {
        shapePtr = &(addShapeSet->shape[i]);

        for (j = 0; j < MAX_CHILDREN; j++) {
            if (shapePtr->source[j] != NIL(Ncl_ShapeData)) {
                if (shapePtr->source[j]->belongsTo < 0) {
      Bear_PrintErr("Ncl_AddShapeSets() failed: invalid source for shape.");
                    return ((Ncl_ShapeSet *) NULL);
                }
            }
        }
    }

    return (addShapeSet);
}



BearStatus
Ncl_MergeShapeSets(lShapeSet, rShapeSet, mergedShapeSet)
Ncl_ShapeSet *lShapeSet, *rShapeSet, *mergedShapeSet;
{
    Ncl_ShapeData   *lShape, *rShape, *mShape;
    Ncl_ShapeData   *shapePtr;
    float           rangeBot, rangeTop;
    int             li, ri, lSize, rSize;
    int             interval, i, j, newSize;

    if ((lShapeSet == NIL(Ncl_ShapeSet)) && 
        (rShapeSet == NIL(Ncl_ShapeSet))) {
        return (BEAR_OK);
    }
    
    if (lShapeSet == NIL(Ncl_ShapeSet)) {
        lSize = 0;
    }
    else {
        lSize = lShapeSet->size;
        lShape = lShapeSet->shape;  
    }

    if (rShapeSet == NIL(Ncl_ShapeSet)) {
        rSize = 0;
    }
    else {
        rSize = rShapeSet->size;
        rShape = rShapeSet->shape;  
    }

    if (lSize > 0) {
        rangeBot = rangeTop = lShape[0].sd_x;
    }
    else if (rSize > 0) {
        rangeBot = rangeTop = rShape[0].sd_x;
    }
    else {
        Bear_PrintErr("Ncl_MergeShapeSets() failed");
        return (BEAR_ERROR);
    }

    for (li = 0; li < lSize; li++) {
        rangeBot = MIN(rangeBot, lShape[li].sd_x);
        rangeTop = MAX(rangeTop, lShape[li].sd_x);
    }
    for (ri = 0; ri < rSize; ri++) {
        rangeBot = MIN(rangeBot, rShape[ri].sd_x);
        rangeTop = MAX(rangeTop, rShape[ri].sd_x);
    }

    mShape = Ncl_InitShapes(ncl_MaxNumOfShapes);

    if ((interval = (rangeTop - rangeBot) / ncl_MaxNumOfShapes) == 0) {
        if (lSize == 0) {
            CopyShapeData(&(mShape[0]), &(rShape[0]));
        }
        else if (rSize == 0) {
            CopyShapeData(&(mShape[0]), &(lShape[0]));
        }
        else if (lShape[0].sd_y < rShape[0].sd_y) {
            CopyShapeData(&(mShape[0]), &(lShape[0]));
        }
        else {
            CopyShapeData(&(mShape[0]), &(rShape[0]));
        }
    } 
    else {
        for (li = 0; li < lSize; li++) {
            j = (lShape[li].sd_x - rangeBot) / interval;
            if (j > ncl_MaxNumOfShapes - 1) {
                j = ncl_MaxNumOfShapes - 1;
            }

            if ((mShape[j].pattern == -1) || 
                ((lShape[li].sd_x * lShape[li].sd_y) < 
                    (mShape[j].sd_x * mShape[j].sd_y))) {
                CopyShapeData(&(mShape[j]), &(lShape[li]));
            }
        }

        for (ri = 0; ri < rSize; ri++) {
            j = (rShape[ri].sd_x - rangeBot) / interval;
            if (j > ncl_MaxNumOfShapes - 1) {
                j = ncl_MaxNumOfShapes - 1;
            }

            if ((mShape[j].pattern == -1) || 
                ((rShape[ri].sd_x * rShape[ri].sd_y) < 
                    (mShape[j].sd_x * mShape[j].sd_y))) {
                CopyShapeData(&(mShape[j]), &(rShape[ri]));
            }
        }
    }

    newSize = 0;
    for (j = 0; j < ncl_MaxNumOfShapes; j++) {
        if (mShape[j].pattern != -1) {
            CopyShapeData(&(mergedShapeSet->shape[newSize]), &(mShape[j]));
            newSize++;
        }
    }
    mergedShapeSet->size = newSize;

#ifdef DEBUG
    for (i=0; i < mergedShapeSet->size; i++) {
        shapePtr = &(mergedShapeSet->shape[i]);

        for (j=0; j < MAX_CHILDREN; j++) {
            if (shapePtr->source[j] != NIL(Ncl_ShapeData)) {
                if (shapePtr->source[j]->belongsTo < 0) {
      Bear_PrintErr("Ncl_AddShapeSets() failed: invalid source for shape.");
                    return (BEAR_ERROR);
                }
            }
        }
    }
#endif DEBUG

    FREE((char *) mShape);
    mShape = NIL(Ncl_ShapeData);

    return (BEAR_OK);
}



Ncl_LengthData *
NclInitLengths()
{
    Ncl_LengthData  *length;
    int             j, i;

    MALLOC(Ncl_LengthData *, length, 
        ncl_MaxNumOfShapes * sizeof (Ncl_LengthData));

    for (j = 0; j < ncl_MaxNumOfShapes; j++) {
        length[j].a = length[j].l = length[j].orient = 0;
        length[j].pattern = length[j].belongsTo = -1;

        for (i = 0; i < MAX_CHILDREN; i++) {
            length[j].source[i] = NIL(Ncl_LengthData);
        }
    }

    return (length);
}



void
NclReInitLengths(length)
Ncl_LengthData *length;
{
    int j, i;

    for (j = 0; j < ncl_MaxNumOfShapes; j++) {

        length[j].a = length[j].l = 
        length[j].orient = 0;
        length[j].pattern = length[j].belongsTo = -1;

        for (i=0; i < MAX_CHILDREN; i++) {
            length[j].source[i] = NIL(Ncl_LengthData);
        }

    }
}



void
NclCopyLengthData(to, from)
Ncl_LengthData    *to, *from;
{
    int i;

    to->a = from->a;
    to->l = from->l;
    to->pattern = from->pattern;
    to->orient = from->orient;
    to->label = from->label;
    to->belongsTo = from->belongsTo;

    for (i=0; i < MAX_CHILDREN; i++) {
        to->source[i] = from->source[i];
    }

}



void
NclCopyLengthSet(to, from)
Ncl_LengthSet *to, *from;
{
    int i, j;

    for (j = 0; j < from->size; j++) {
        NclCopyLengthData(&(to->length[j]), &(from->length[j]));
    }

    for (j = from->size; j < ncl_MaxNumOfShapes; j++) {
        to->length[j].pattern = -1;
        to->length[j].belongsTo = -1;
        to->length[j].a = to->length[j].l =
        to->length[j].orient = 0;

        for (i=0; i < MAX_CHILDREN; i++) {
            to->length[j].source[i] = NIL(Ncl_LengthData);
        }

    }

    to->size = from->size;
}



#ifdef DEBUG
static void
PrintLengthData(lengthPtr)
Ncl_LengthData *lengthPtr;
{
    int i, labelForPrint = 0;

    if (lengthPtr->label != NIL(int)) {
        for (i = 0; i < MAX_CHILDREN; i++) {
            labelForPrint = 10*labelForPrint + lengthPtr->label[i];
        }
    }

    (void) fprintf (bearout,
        "\na: %.2f, l: %.2f, pattern: %d, orient: %d, label: %d, belongsTo: %d",
        lengthPtr->a, lengthPtr->l, lengthPtr->pattern, 
        lengthPtr->orient, labelForPrint, lengthPtr->belongsTo);
}
#endif /* DEBUG */


#ifdef DEBUG
void
NclPrintLengthSet(lengthSet)
Ncl_LengthSet *lengthSet;
{
    int i, j;

    for (j = 0; j < lengthSet->size; j++) {
        PrintLengthData(&(lengthSet->length[j]));

        /*
        printf("\nPrinting source length ---");
        for (i=0; i < MAX_CHILDREN; i++) {
            if (lengthSet->length[j].source[i]==NIL(Ncl_LengthData))
                continue;
            PrintLengthData(lengthSet->length[j].source[i]);
        }
        printf("\nEnd printing source length ---\n");
        */

        fflush (bearout);
    }
    (void) fprintf (bearout, "\n");
}
#endif /* DEBUG */



BearStatus
NclMergeLengthSets(lLengthSet, rLengthSet, mergeLengthSet)
Ncl_LengthSet *lLengthSet, *rLengthSet, *mergeLengthSet;
{
    Ncl_LengthData  *lLength, *rLength, *mLength;
    float           rangeBot, rangeTop;
    float           interval, a_last;
    int             li, ri, lSize, rSize, j;

    if (lLengthSet == NIL(Ncl_LengthSet) && rLengthSet == NIL(Ncl_LengthSet)) {
        return (BEAR_OK);
    } 
    else if (lLengthSet == NIL(Ncl_LengthSet)) {
        for (ri = 0; ri < rLengthSet->size; ri++) {
            NclCopyLengthData(&(mergeLengthSet->length[ri]), 
                &(rLengthSet->length[ri]));
        }
        mergeLengthSet->size = rLengthSet->size; 
        return (BEAR_OK);

    } 
    else if (rLengthSet == NIL(Ncl_LengthSet)) {
        for (li = 0; li < lLengthSet->size; li++) {
            NclCopyLengthData(&(mergeLengthSet->length[li]), 
                &(lLengthSet->length[li]));
        }
        mergeLengthSet->size = lLengthSet->size; 
        return (BEAR_OK);
    } 

    if (lLengthSet->size == 0 && rLengthSet->size == 0) {
        return (BEAR_OK);
    } 
    else if (lLengthSet->size == 0) {
        for (ri = 0; ri < rLengthSet->size; ri++) {
            NclCopyLengthData(&(mergeLengthSet->length[ri]), 
                &(rLengthSet->length[ri]));
        }
        mergeLengthSet->size = rLengthSet->size; 

        return (BEAR_OK);
    } 
    else if (rLengthSet->size == 0) {
        for (li = 0; li < lLengthSet->size; li++) {
            NclCopyLengthData(&(mergeLengthSet->length[li]), 
                &(lLengthSet->length[li]));
        }
        mergeLengthSet->size = lLengthSet->size; 

        return (BEAR_OK);
    } 

    lLength = lLengthSet->length;
    rLength = rLengthSet->length;
    lSize = lLengthSet->size;
    rSize = rLengthSet->size;

    rangeBot = NCL_REAL_MAX;
    rangeTop = -NCL_REAL_MAX;
    for (li = 0; li < lSize; li++) {
        rangeBot = MIN(rangeBot, lLength[li].a);
        rangeTop = MAX(rangeTop, lLength[li].a);
    }
    for (ri = 0; ri < rSize; ri++) {
        rangeBot = MIN(rangeBot, rLength[ri].a);
        rangeTop = MAX(rangeTop, rLength[ri].a);
    }

    interval = (rangeTop - rangeBot) / ncl_MaxNumOfShapes; 

    mLength = NclInitLengths();
    if (interval == 0.0) {
        if (lLength[0].l < rLength[0].l) {
            NclCopyLengthData(&(mLength[0]), &(lLength[0]));
        } 
        else {
            NclCopyLengthData(&(mLength[0]), &(rLength[0]));
        }
    } 
    else {
        for (li = 0; li < lSize; li++) {
            j = MIN(ncl_MaxNumOfShapes - 1, 
                (lLength[li].a-rangeBot)/interval);
            if (mLength[j].pattern == -1 || lLength[li].a * lLength[li].l <
                    mLength[j].a * mLength[j].l) {
                NclCopyLengthData(&(mLength[j]), &(lLength[li]));
            }
        }

        for (ri = 0; ri < rSize; ri++) {
            j = MIN(ncl_MaxNumOfShapes - 1, 
                (rLength[ri].a - rangeBot) / interval);
            if ((mLength[j].pattern == -1) || 
                ((rLength[ri].a * rLength[ri].l) <
                    (mLength[j].a * mLength[j].l))) {

                NclCopyLengthData(&(mLength[j]), &(rLength[ri]));
            }
        }
    }

    mergeLengthSet->size = 0;
    a_last = (float) INFINITY;
    for (j = 0; j < ncl_MaxNumOfShapes; j++) {
        if ((mLength[j].pattern != -1) && (mLength[j].a != a_last)) {
            NclCopyLengthData(&(mergeLengthSet->length[mergeLengthSet->size]),
                &(mLength[j]));
            a_last = mLength[j].a;
            (mergeLengthSet->size)++;
        }
    }

    FREE((char *) mLength);
    mLength = NIL(Ncl_LengthData);
    /*
    FREE((char *) lLengthSet->shape);
    FREE((char *) lLengthSet);
    FREE((char *) rLengthSet->shape);
    FREE((char *) rLengthSet);
    */

    return (BEAR_OK);
}



Ncl_SideDir
Ncl_GetOppositeSideDir(sideDir)
Ncl_SideDir sideDir;
{
    switch (sideDir) {
    case NCL_NORTH_DIR:
        return (NCL_SOUTH_DIR);
    case NCL_EAST_DIR:
        return (NCL_WEST_DIR);
    case NCL_SOUTH_DIR:
        return (NCL_NORTH_DIR);
    case NCL_WEST_DIR:
        return (NCL_EAST_DIR);
    default:
     Bear_PrintFatalErr("Ncl_GetOppositeSideDir() failed : illegal direction");
        break;
    }
    /* NOTREACHED */
}


char *
Ncl_GetSideDirName(sideDir)
Ncl_SideDir sideDir;
{
    switch (sideDir) {
    case NCL_NORTH_DIR:
        return ("North");
    case NCL_EAST_DIR:
        return ("East");
    case NCL_SOUTH_DIR:
        return ("South");
    case NCL_WEST_DIR:
        return ("West");
    default:
    Bear_PrintFatalErr("Ncl_GetSideDirName() failed : illegal side direction");
        break;
    }
    /* NOTREACHED */
}


static BearStatus
CheckPin(pinInstPtr, checkDataPtr)
PinInst         *pinInstPtr;
CheckPinData    *checkDataPtr;
{
    switch (checkDataPtr->cellFlexType) {
    case MAC_FIXED_SHAPE:
        if (Mac_IsPinPosDef(pinInstPtr) == FALSE) {
            sprintf(bear_ErrorMsg,
                "Cell named %s is fixed but has a pin with no position.\n",
                Mac_GetNameOfCellInst(checkDataPtr->cellInstPtr));
            Bear_PrintErr(bear_ErrorMsg);
            return (BEAR_STOP_SEARCH);
        }
        break;
    case MAC_FLEX_SHAPE:
        break;
    default:
        sprintf(bear_ErrorMsg,
            "Cell named %s has an illegal flexibility type.\n",
            Mac_GetNameOfCellInst(checkDataPtr->cellInstPtr));
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_STOP_SEARCH);
    }

    switch (Mac_GetFlexOfPin(pinInstPtr)) {
    case (Flexibility) FIXED:
        checkDataPtr->fixPinCnt++;
        break;
    case (Flexibility) FLEXIBLE:
        checkDataPtr->floatPinCnt++;
        break;
    case (Flexibility) FLEXIBLE_NORTH:
        checkDataPtr->nFloatPinCnt++;
        break;
    case (Flexibility) FLEXIBLE_EAST:
        checkDataPtr->eFloatPinCnt++;
        break;
    case (Flexibility) FLEXIBLE_SOUTH:
        checkDataPtr->sFloatPinCnt++;
        break;
    case (Flexibility) FLEXIBLE_WEST:
        checkDataPtr->wFloatPinCnt++;
        break;
    default:
      Bear_PrintErr("CheckPin() failed : pin has an illegal flexibility type");
        return (BEAR_STOP_SEARCH);
    }
    checkDataPtr->totalPinCnt++;

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CountNumOfPins(pinPtr, pinCountPtr)
PinInst *pinPtr;
int     *pinCountPtr;
{
    (*pinCountPtr)++;

    return (BEAR_CONTINUE_SEARCH);
}


static int
CompareCells(item1, item2)
lsGeneric item1, item2;
{
    SortData *sortData1Ptr = (SortData *) ((CellInst *) item1)->clientData;
    SortData *sortData2Ptr = (SortData *) ((CellInst *) item2)->clientData;

    if (sortData1Ptr->area < sortData2Ptr->area) {
        return (-1);
    }
    else if (sortData1Ptr->area == sortData2Ptr->area) {
        if (sortData1Ptr->pinCount < sortData2Ptr->pinCount) {
            return (-1);
        }
        else if (sortData1Ptr->pinCount == sortData2Ptr->pinCount) {
            return (0);
        }
        else {
            return (1);
        }
    }
    else {
        return (1);
    }
    /* NOTREACHED */
}


/* ARGSUSED */
static BearStatus
TidyUpChildCell(cellInstPtr, clientData)
CellInst    *cellInstPtr;
ClientData  clientData;
{
    lsGen           gen;
    CellMaster      *cellMasterPtr;
    CheckPinData    checkData;
    Shape           *shapePtr;
    int             maxPinCnt;  /* maximum pin count    */
    int             mps;        /* minimum pin spacing  */

    if (Mac_AssignShapeFlag(cellInstPtr) != BEAR_OK) {
        return (BEAR_STOP_SEARCH);
    }

    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);

    checkData.cellInstPtr = cellInstPtr;
    checkData.cellFlexType = cellMasterPtr->shapeFlag;
    checkData.fixPinCnt = 0;
    checkData.floatPinCnt = 0;
    checkData.nFloatPinCnt = 0;
    checkData.eFloatPinCnt = 0;
    checkData.sFloatPinCnt = 0;
    checkData.wFloatPinCnt = 0;
    checkData.totalPinCnt = 0;
    if (Mac_ItrPinOfCellInst(cellInstPtr, CheckPin,
        (ClientData) &checkData) == BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }

    if (cellMasterPtr->shapeFlag == MAC_FIXED_SHAPE) {
        return (BEAR_CONTINUE_SEARCH);
    }
    else if (cellMasterPtr->shapeFlag == MAC_FLEX_SHAPE) {
        mps = Mac_CalcMinPinSpacing();
        lsForeachItem(cellMasterPtr->shapeList, gen, shapePtr) {
            maxPinCnt =
                2 * ((shapePtr->horzDim / mps) + (shapePtr->vertDim / mps));
            if (checkData.totalPinCnt > maxPinCnt) {
                (void) sprintf (bear_ErrorMsg,
                    "Shape (%d %d) for cell %s cannot hold %d pins.\n",
                    shapePtr->horzDim, shapePtr->vertDim,
                    Mac_GetNameOfCellInst(cellInstPtr),
                    checkData.totalPinCnt);
                Bear_PrintErr(bear_ErrorMsg);
                lsFinish(gen);
                return (BEAR_STOP_SEARCH);
            }
        }
    }
    else {
        (void) sprintf (bear_ErrorMsg, 
            "The shape of the cell named \"%s\" is undefined.",
            Mac_GetNameOfCellInst(cellInstPtr));
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


BearStatus
NclTidyUpCells(coreMasterPtr)
CellMaster *coreMasterPtr;
/*
 * For cells whose shape flag is fixed :
 *     all pins must be fixed
 *     a defined position for each pin is mandatory
 *
 * For cells whose shape flag is flexible :
 *     all pins must be flexible
 *     a defined position for each pin is optional
 *
 * Additionally all of the cell's pins must fit on each of the
 * given shapes.
 */
{
    if (Mac_ItrChildrenOfCell(coreMasterPtr, TidyUpChildCell,
        (ClientData) NULL) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

    if (nclAlgorithum != FILE_IN) {
        CellInst    *cellInstPtr;
        lsGen       gen;
        lsList      ciList;
        Shape       shape;
        SortData    *sortDataPtr;
        int         pinCount;

        /*
         * We sort the cells by area and number of pins, because the
         * final placement is indirectly dependant on the original order
         * of the cells and we want the user to get the same result no
         * matter how he orders the cells in the input file.
         */
        ciList = Mac_CreateCellInstList(coreMasterPtr);
        lsForeachItem(ciList, gen, cellInstPtr) {
            MALLOC(SortData *, sortDataPtr, sizeof (SortData));

            (void) Mac_GetCurrentShape(cellInstPtr, &shape);
            sortDataPtr->area = shape.horzDim * shape.vertDim;

            pinCount = 0;
            (void) Mac_ItrPinOfCellInst(cellInstPtr, CountNumOfPins,
                (ClientData) &pinCount);
            sortDataPtr->pinCount = pinCount;

            cellInstPtr->clientData = (ClientData) sortDataPtr;
        }
        lsSort(ciList, CompareCells);

        coreMasterPtr->contents.cellInstList = NIL(CellInst);
        lsForeachItem(ciList, gen, cellInstPtr) {
            cellInstPtr->siblingPtr = coreMasterPtr->contents.cellInstList;
            coreMasterPtr->contents.cellInstList = cellInstPtr;

            FREE((char *) cellInstPtr->clientData);
            cellInstPtr->clientData = (ClientData) NULL;
        }
        lsDestroy(ciList, (void (*)()) NULL);
    }

    return (BEAR_OK);
}


Ncl_Node *
NclMakeNode()
{
    Ncl_Node    *nodePtr;
    int         s;

    MALLOC(Ncl_Node *, nodePtr, sizeof (Ncl_Node));
    nodePtr->nd_id = 0;
    nodePtr->nd_index = 0;
    nodePtr->nd_mark = BEAR_UNMARKED;
    nodePtr->nd_depth = 0;
    nodePtr->nd_numOfChildren = 0;

    nodePtr->nd_totalNetList = (lsList) NULL;
    nodePtr->nd_bdrNetList = (lsList) NULL;
    NclForeachSide(s) {
        nodePtr->nd_initPinList[s] = (lsList) NULL;
        nodePtr->nd_tilePinList[s] = (lsList) NULL;
    }

    nodePtr->nd_fixPinCnt = 0;
    nodePtr->nd_freeFloPinCnt = 0;
    NclForeachSide(s) {
        nodePtr->nd_sideFloPinCnt[s] = 0;
    }
    nodePtr->nd_totalPinCnt = 0;

    nodePtr->nd_numOfNorthSlots = 0;
    nodePtr->nd_numOfSouthSlots = 0;
    nodePtr->nd_numOfWestSlots = 0;
    nodePtr->nd_numOfEastSlots = 0;

    nodePtr->clientData = (ClientData) NULL;

    return (nodePtr);
}


Ncl_UserParms *
Ncl_MakeUserParms()
{
    register Ncl_UserParms *parmsPtr;

    MALLOC(Ncl_UserParms *, parmsPtr, sizeof (Ncl_UserParms));

    parmsPtr->algorithm = MATCHING;

    /* Begin matching only. */ 
    MALLOC(int *, parmsPtr->dimRatioPtr, sizeof (int));
    *(parmsPtr->dimRatioPtr) = 10;
    MALLOC(int *, parmsPtr->areaRatioPtr, sizeof (int));
    *(parmsPtr->areaRatioPtr) = 10;
    parmsPtr->promptFlag = FALSE;
    /* End matching only.   */ 
    
    /* Begin greedy only.   */ 
    parmsPtr->areaFact = 1.5;
    parmsPtr->areaRelax = 2.0;
    parmsPtr->ratioMax = 5.0;
    parmsPtr->ratioRelax = 2.0;
    /* End greedy only.     */

    parmsPtr->connMin = 0.0;
    parmsPtr->routAdjust = 1.0; /* routFlag */

    parmsPtr->quickRtEstFlag = TRUE;
    parmsPtr->chanFreeFlag = FALSE;
    parmsPtr->timingFlag = FALSE;
    parmsPtr->storeFlag = FALSE;

    parmsPtr->clusterFileName = NIL(char);
    parmsPtr->storageFileName = NIL(char);
    parmsPtr->timingFileName = NIL(char);

    parmsPtr->maxNumOfShapes = NCL_DEF_NUM_OF_SHAPES;

    Ncl_EchoUserParms("Cluster parameters values set by program.", parmsPtr);

    return (parmsPtr);
}



void
Ncl_FreeUserParms(parmsPtrPtr)
Ncl_UserParms **parmsPtrPtr;
{
    FREE((char *) *parmsPtrPtr);
    *parmsPtrPtr = NIL(Ncl_UserParms);
}



void
Ncl_EchoUserParms(msgPtr, parmsPtr)
char            *msgPtr;
Ncl_UserParms   *parmsPtr;
{
#ifdef DEBUG
    if (echoUserParmsFlag == FALSE) {
        return;
    }

    (void) fprintf (bearout, "----------\n");
    (void) fprintf (bearout, "%s\n\n", msgPtr);

    (void) fprintf (bearout, "%-33s : %d\n", "algorithm",
        parmsPtr->algorithm);

    (void) fprintf (bearout, "%-33s : %d\n", "areaRatioPtr", 
        *(parmsPtr->areaRatioPtr));
    (void) fprintf (bearout, "%-33s : %d\n", "dimRatioPtr",
        *(parmsPtr->dimRatioPtr));
    (void) fprintf (bearout, "%-33s : %d\n", "promptFlag",
        parmsPtr->promptFlag);

    (void) fprintf (bearout, "%-33s : %f\n", "areaFact", 
        parmsPtr->areaFact);
    (void) fprintf (bearout, "%-33s : %f\n", "areaRelax",
        parmsPtr->areaRelax);
    (void) fprintf (bearout, "%-33s : %f\n", "ratioMax",
        parmsPtr->ratioMax);
    (void) fprintf (bearout, "%-33s : %f\n", "ratioRelax",
        parmsPtr->ratioRelax);
    (void) fprintf (bearout, "%-33s : %f\n", "connMin",
        parmsPtr->connMin);
    (void) fprintf (bearout, "%-33s : %f\n", "routAdjust",
        parmsPtr->routAdjust);

    (void) fprintf (bearout, "%-33s : %d\n", "quickRtEstFlag",
        parmsPtr->quickRtEstFlag);
    (void) fprintf (bearout, "%-33s : %d\n", "chanFreeFlag",
        parmsPtr->chanFreeFlag);
    (void) fprintf (bearout, "%-33s : %d\n", "timingFlag",
        parmsPtr->timingFlag);
    (void) fprintf (bearout, "%-33s : %d\n", "storeFlag",
        parmsPtr->storeFlag);

    (void) fprintf (bearout, "%-33s : \"%s\"\n", "clusterFileName",  
        (parmsPtr->clusterFileName != NIL(char)) ?  
            parmsPtr->clusterFileName : "No file given");

    (void) fprintf (bearout, "%-33s : \"%s\"\n", "storageFileName",  
        (parmsPtr->storageFileName != NIL(char)) ?  
            parmsPtr->storageFileName : "No file given");

    (void) fprintf (bearout, "%-33s : \"%s\"\n", "timingFileName",  
        (parmsPtr->timingFileName != NIL(char)) ?  
            parmsPtr->timingFileName : "No file given");

    (void) fprintf (bearout, "%-33s : %d\n", "maxNumOfShapes",
        parmsPtr->maxNumOfShapes);

    (void) fprintf (bearout, "\n");
    (void) fflush (bearout);
#endif DEBUG
}


/*
 * ----------------------------------------------------------------------------
 *
 * NclPrintFatalErr() --
 *
 *  If the program can't be terminated because of some illegal condition,
 *  this procedure is called and the program aborted.
 *
 * Results:
 *   Message is printed and program is terminated
 *
 * Side effects:
 *   none
 *
 * ----------------------------------------------------------------------------
 */

void
NclPrintFatalErr(errorCode)
int errorCode;
{
    switch (errorCode) {
    default:
        sprintf (bear_ErrorMsg, "Sorry, no error explanation available.\n");
        break;
    }
    errRaise(NEWC_MOD_NAME, ((int) -1), bear_ErrorMsg);
}


char *
NclGetAlgString(algorithum)
int algorithum;
{
    switch (algorithum) {
    case RANDOM:
        return ("random");
    case GREEDY:
        return ("greedy");
    case MATCHING:
        return ("matching");
    case SIM_ANNEAL:
        return ("annealing");
    case FILE_IN:
        return ("file in");
    default:
        break;
    }
    return ("unknown");
}


