
/* ngrCompGraph.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newG/ngrCompGraph.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "newG.h"
#include "newGInt.h"


typedef struct _CompData CompData;
struct _CompData {
    Gt_Graph        baseGraph;
    Gt_Graph        compGraph;
    Gt_AlgData      algData;
    Gt_GraphNode    srcNode;
    int             edgeNum;
    bool            pathListFlag;
};


/* ---------- Local procedures ---------- */
static BearStatus   AddPinToCompGraph();
static BearStatus   ConnectNodeToOtherNodes();
static BearStatus   AddEdgeToCompGraph();


static BearStatus
ConnectNodeToOtherNodes(node, compDataPtr)
Gt_GraphNode    node;
CompData        *compDataPtr;
{
    compDataPtr->srcNode = node;
    (void) Gt_ItrNodesOfGraph(compDataPtr->compGraph, AddEdgeToCompGraph,
        (ClientData) compDataPtr);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddEdgeToCompGraph(node, compDataPtr)
Gt_GraphNode    node;
CompData        *compDataPtr;
{
    Gt_GraphEdge edge;

    if (Gt_AddGraphEdge(compDataPtr->srcNode, node,
        (BearId) compDataPtr->edgeNum, (BearGeneric) NULL, &edge) == BEAR_OK) {

        Gt_GraphNode        baseNode1, baseNode2;
        Gt_Path             *pathPtr;
        Ngr_CompEdgeData    *compEdgePtr;
        PinOrJunct          *pinOrJunctPtr;

        pinOrJunctPtr =
            (PinOrJunct *) Gt_GetNodeDataOfNode(compDataPtr->srcNode);
        baseNode1 = BASE_NODE_OF_PIN(pinOrJunctPtr->unValue.pinPtr);

        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
        baseNode2 = BASE_NODE_OF_PIN(pinOrJunctPtr->unValue.pinPtr);

        (void) Gt_ShortestPath(compDataPtr->baseGraph, compDataPtr->algData,
            baseNode1, baseNode2, compDataPtr->pathListFlag, &pathPtr);

    	compEdgePtr = ALLOC(Ngr_CompEdgeData, 1);
        compEdgePtr->pathPtr = pathPtr;
        Gt_PutEdgeDataOfEdge((BearGeneric) compEdgePtr, edge);

        (compDataPtr->edgeNum)++;
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPinToCompGraph(pinPtr, compDataPtr)
PinInst     *pinPtr;
CompData    *compDataPtr;
{
    Gt_GraphNode    baseNode;
    PinOrJunct      *pinOrJunctPtr;

    pinOrJunctPtr = Fpg_MakePinOrJunct();
    pinOrJunctPtr->unType = FPG_PIN;
    pinOrJunctPtr->unValue.pinPtr = pinPtr;

    baseNode = BASE_NODE_OF_PIN(pinPtr);

    COMP_NODE_OF_PIN(pinPtr) =
        (BearGeneric) Gt_AddGraphNode(compDataPtr->compGraph,
            (BearId) Gt_GetIdOfNode(baseNode), (BearGeneric) pinOrJunctPtr);

    return (BEAR_CONTINUE_SEARCH);
}


BearStatus
Ngr_BuildCompGraph(bldTreePtr, compGraphPtr)
Ngr_BldTreeDataPtr  bldTreePtr;             /* IN       */
Gt_Graph            *compGraphPtr;          /* OUT      */
/*
 * pinCount = The number of pins in the given net.
 *
 * If (pinCount <= NGR_LOWER_PIN_CUT_OFF) then
 *     The complete graph will be built.
 *
 * If ((pinCount > NGR_LOWER_PIN_CUT_OFF) and
 *     (pinCount <= NGR_UPPER_PIN_CUT_OFF)) then
 *     The complete graph will be built, but in an effort to conserve
 *     memory only the length of the path will be retained.  That is
 *     the edge lists for the paths generated in the base graph between every
 *     two nodes in the complete graph will not be saved and as such the 
 *     final step will take a little longer because the paths that belong
 *     to edges in the complete graph that are kept will have to be re-done.
 *    
 * If (pinCount > NGR_UPPER_PIN_CUT_OFF) then
 *     The complete graph will NOT be built.  In this case we will be
 *     unable to obtain a steiner tree.
 */
{
    BldTreeDataInt  *intBldTreePtr = (BldTreeDataInt *) bldTreePtr;
    CompData        compData;
    Mac_GraphData   *graphDataPtr;

    *compGraphPtr = intBldTreePtr->compGraph = (Gt_Graph) NULL;
    if (intBldTreePtr->pinCount > NGR_UPPER_PIN_CUT_OFF) {
        return (BEAR_OK);
    }

    graphDataPtr = Mac_MakeGraphData();
    graphDataPtr->fpgId = Fpg_GetIdOfFpg(intBldTreePtr->fpgPtr);
    graphDataPtr->pgFlag = FALSE;
    graphDataPtr->edgeDataFlag = FALSE;

    compData.compGraph = Gt_CreateGraph(Mac_GenerateId(),
        (BearGeneric) graphDataPtr);
    compData.srcNode = (Gt_GraphNode) NULL;
    compData.edgeNum = 0;

    (void) Mac_ItrAllPinOfNet(intBldTreePtr->netPtr, AddPinToCompGraph, 
        (ClientData) &compData);

    compData.pathListFlag =
        (intBldTreePtr->pinCount > NGR_LOWER_PIN_CUT_OFF) ? FALSE : TRUE;
    compData.baseGraph = intBldTreePtr->baseGraph;
    compData.algData = Gt_GenAlgData(intBldTreePtr->baseGraph,
        Ngr_GetBaseEdgeWeight);
    (void) Gt_ItrNodesOfGraph(compData.compGraph, ConnectNodeToOtherNodes,
        (ClientData) &compData);

    Gt_FreeAlgData(intBldTreePtr->baseGraph, &compData.algData);

    *compGraphPtr = intBldTreePtr->compGraph = compData.compGraph;

    return (BEAR_OK);
}


