
/* ngrFindMST.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newG/ngrFindMST.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "newG.h"
#include "newGInt.h"


typedef struct _MBPData MBPData;
struct _MBPData {
    Gt_Graph    baseGraph;
    Gt_AlgData  algData;
};


/* ---------- Local procedures ---------- */
static BearStatus   InitBaseEdgeMark();
static BearStatus   MarkBasePath();
static BearStatus   AddUnmarkedBaseEdgeToList();


/* ARGSUSED */
static BearStatus
InitBaseEdgeMark(baseEdge, clientData)
Gt_GraphEdge    baseEdge;           /* Edge from the base graph. */
ClientData      clientData;
{
    ((Ngr_BaseEdgeData *) Gt_GetEdgeDataOfEdge(baseEdge))->mark = BEAR_UNMARKED;

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
MarkBasePath(compEdge, mbpDataPtr)
Gt_GraphEdge    compEdge;           /* Edge from the complete graph. */
MBPData         *mbpDataPtr;
{
    Gt_Path     *pathPtr;
    Gt_PathEdge *pathEdgePtr;
    lsGen       gen;

    pathPtr = ((Ngr_CompEdgeData *) Gt_GetEdgeDataOfEdge(compEdge))->pathPtr;

    if (pathPtr->pathList == (lsList) NULL) {
        Gt_GraphNode    baseNode1, baseNode2;
        Gt_GraphNode    compNode;
        Gt_Path         *newPathPtr;
        PinOrJunct      *pinOrJunctPtr;

        compNode = Gt_GetNodeOneOfEdge(compEdge);
        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(compNode);
        baseNode1 = BASE_NODE_OF_PIN(pinOrJunctPtr->unValue.pinPtr);

        compNode = Gt_GetNodeTwoOfEdge(compEdge);
        pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(compNode);
        baseNode2 = BASE_NODE_OF_PIN(pinOrJunctPtr->unValue.pinPtr);

        (void) Gt_ShortestPath(mbpDataPtr->baseGraph, mbpDataPtr->algData,
            baseNode1, baseNode2, TRUE, &newPathPtr);

        pathPtr->pathList = newPathPtr->pathList;
        newPathPtr->pathList = (lsList) NULL;
        Gt_FreePath(&newPathPtr);
    }

    lsForeachItem(pathPtr->pathList, gen, pathEdgePtr) {
        ((Ngr_BaseEdgeData *) Gt_GetEdgeDataOfEdge(pathEdgePtr->edge))->mark =
            BEAR_MARKED;
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddUnmarkedBaseEdgeToList(baseEdge, baseEdgeList)
Gt_GraphEdge    baseEdge;           /* Edge from the base graph. */
lsList          baseEdgeList;
{
    if (((Ngr_BaseEdgeData *) Gt_GetEdgeDataOfEdge(baseEdge))->mark ==
        BEAR_UNMARKED) {
        lsNewEnd(baseEdgeList, (lsGeneric) baseEdge, LS_NH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


BearStatus
Ngr_FindMST(bldTreePtr)
Ngr_BldTreeDataPtr bldTreePtr;              /* IN       */
/*
 * 1.) Find a minimum spanning tree for the complete graph if it has
 *     more then 2 nodes.
 * 2.) Visit each tree edge and traverse the path in the base graph
 *     it represents, marking each edge in the path as you go.
 * 3.) Eliminate unused edges in the base graph.
 *
 * Note that upon completion of these steps the base graph may still
 * have some loops in it.
 */
{
    BldTreeDataInt      *intBldTreePtr = (BldTreeDataInt *) bldTreePtr;
    Gt_GraphEdge        baseEdge;
    Gt_GraphNode        baseNode1, baseNode2;
    lsGen               gen;
    lsList              baseEdgeList;
    Ngr_BaseEdgeData    *baseEdgePtr;
    PinOrJunct          *pinOrJunctPtr;
    int                 numOfEdges;

    if (intBldTreePtr->compGraph == (Gt_Graph) NULL) {
        return (BEAR_OK);
    }

    numOfEdges = Gt_GetNumOfGraphEdges(intBldTreePtr->compGraph);
    if (numOfEdges > 1) {
        (void) Gt_MinSpanningTree(intBldTreePtr->compGraph, 
            Ngr_GetCompEdgeWeight, Mac_FreeNodeData, NgrFreeCompEdgeData);
    }
    else {
        intBldTreePtr->haveTreeFlag = TRUE;
    }

    (void) Gt_ItrEdgesOfGraph(intBldTreePtr->baseGraph, InitBaseEdgeMark,
        (ClientData) NULL);

    if (intBldTreePtr->pinCount > NGR_LOWER_PIN_CUT_OFF) {
        MBPData mbpData;

        mbpData.baseGraph = intBldTreePtr->baseGraph;
        mbpData.algData = Gt_GenAlgData(intBldTreePtr->baseGraph,
            Ngr_GetBaseEdgeWeight);

        (void) Gt_ItrEdgesOfGraph(intBldTreePtr->compGraph, MarkBasePath,
            (ClientData) &mbpData);

        Gt_FreeAlgData(intBldTreePtr->baseGraph, &mbpData.algData);
    }
    else {
        (void) Gt_ItrEdgesOfGraph(intBldTreePtr->compGraph, MarkBasePath,
            (ClientData) NULL);
    }

    baseEdgeList = lsCreate();
    (void) Gt_ItrEdgesOfGraph(intBldTreePtr->baseGraph,
        AddUnmarkedBaseEdgeToList, (ClientData) baseEdgeList);
    lsForeachItem(baseEdgeList, gen, baseEdge) {
        baseNode1 = Gt_GetNodeOneOfEdge(baseEdge);
        baseNode2 = Gt_GetNodeTwoOfEdge(baseEdge);

        Gt_DeleteGraphEdge(baseEdge, (BearGeneric *) &baseEdgePtr);
        NgrFreeBaseEdgeData(baseEdgePtr);

        if (Gt_IsNodeUnconnected(baseNode1) == TRUE) {
            (void) Gt_DeleteGraphNode(baseNode1, &pinOrJunctPtr);
            Mac_FreeNodeData(pinOrJunctPtr);
        }
        if (Gt_IsNodeUnconnected(baseNode2) == TRUE) {
            (void) Gt_DeleteGraphNode(baseNode2, &pinOrJunctPtr);
            Mac_FreeNodeData(pinOrJunctPtr);
        }
    }
    lsDestroy(baseEdgeList, (void (*)()) NULL);

    return (BEAR_OK);
}


