
/* ngrWrite.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newG/ngrWrite.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "newG.h"
#include "newGInt.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */


typedef struct _WriteData WriteData;
struct _WriteData {
    FILE    *filePtr;
    int     (*weightProc)();
};


/* ---------- Local Procedures ---------- */
static BearStatus WriteEdge();
static BearStatus WriteNode();


void 
Ngr_WriteGraph(fileNamePtr, graph, weightProc)
char        *fileNamePtr;
Gt_Graph    graph;
int         (*weightProc)();
{
    FILE        *filePtr;
    WriteData   printData;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg, "could not open %s\n", fileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return;
    }
                                                        
    (void) fprintf (filePtr, "\n------------- edge list -------------\n\n");
    (void) fprintf (filePtr, "%8s%8s%8s%12s\n\n", "edge", "node1", "node2",
        "weight");
    printData.filePtr = filePtr;
    printData.weightProc = weightProc;
    (void) Gt_ItrEdgesOfGraph(graph, WriteEdge, (ClientData) &printData);

    (void) fprintf (filePtr, "\n------------- node list -------------\n\n");
    (void) fprintf (filePtr, "%8s  %-15s %8s\n\n", "node", "type", "degree");
    (void) Gt_ItrNodesOfGraph(graph, WriteNode, (ClientData) filePtr);
    (void) fprintf (filePtr, "\n");

    (void) fflush (filePtr);
    (void) fclose (filePtr);
}



static BearStatus 
WriteNode(node, filePtr)
Gt_GraphNode    node;
FILE            *filePtr;
{
    PinOrJunct  *pinOrJunctPtr;
    PseudoPin   *ppinPtr;
    Tile        *tilePtr;

    pinOrJunctPtr = (PinOrJunct *) Gt_GetNodeDataOfNode(node);
    
    (void) fprintf (filePtr, "%8d", Gt_GetIdOfNode(node));

    switch (pinOrJunctPtr->unType) {
    case (int) FPG_PIN:
        (void) fprintf (filePtr, "  %-15s ", "pin");
        break;
    case (int) FPG_PSEUDO_PIN:
        (void) fprintf (filePtr, "  %-15s ", "pseudo pin");
        break;
    case (int) FPG_ROUTE_PIN:
        (void) fprintf (filePtr, "  %-15s ", "route pin");
        break;
    case (int) FPG_WALL_JUNCT:
        (void) fprintf (filePtr, "  %-15s ", "wall junction");
        break;
    case (int) FPG_PSEUDO_JCT:
        (void) fprintf (filePtr, "  %-15s ", "pseudo junction");
        break;
    case FPG_BBL_TERM:
        (void) fprintf (filePtr, "  %-15s ", "bbl terminal");
        break;
    case FPG_BBL_JUNCT:
        (void) fprintf (filePtr, "  %-15s ", "bbl junction");
        break;
    default:
        Bear_PrintErr("WriteNode() : unexpected pin or junction type");
        break;
    }

    (void) fprintf (filePtr, "%8d\n", Gt_GetDegreeOfNode(node));

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus 
WriteEdge(edge, printDataPtr)
Gt_GraphEdge    edge;
WriteData       *printDataPtr;
{
    Gt_GraphNode node1, node2;

    node1 = Gt_GetNodeOneOfEdge(edge);
    node2 = Gt_GetNodeTwoOfEdge(edge);

    if (Gt_GetEdgeDataOfEdge(edge) == (BearGeneric) NULL) {
        (void) fprintf (printDataPtr->filePtr, "%8d%8d%8d%-12s\n",
            Gt_GetIdOfEdge(edge),
            Gt_GetIdOfNode(node1), 
            Gt_GetIdOfNode(node2),
            "--");
    }
    else {
        (void) fprintf (printDataPtr->filePtr, "%8d%8d%8d%12d\n",
            Gt_GetIdOfEdge(edge),
            Gt_GetIdOfNode(node1), 
            Gt_GetIdOfNode(node2),
            (*(printDataPtr->weightProc))(edge));
    }

    return (BEAR_CONTINUE_SEARCH);
}


