
/* laLinAssign.c --
 *
 *     (Advising Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/laLinAssign.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "bearStatus.h"
#include "laLinAssign.h"


/*
 * Grody to the max but then that's what you get when you borrow
 * code from other people.
 */
#define LA_ASSERT(cond, mesg)                               \
{                                                           \
    if (!(cond)) {                                          \
        (void) fprintf (stderr, "%s\n", mesg);              \
        (void) fprintf (stderr, "program btest failed\n");  \
        exit(100);                                          \
    }                                                       \
}

#ifndef bool
#define bool int
#endif bool

#ifndef TRUE
#define TRUE 1
#endif TRUE

#ifndef FALSE
#define FALSE 0
#endif FALSE


/* ---------- Local Procedures ---------- */
static BearStatus   CheckLASolution();
static void         Augmentation();
static void         InitAssign();
static void         SetUpInternalArrays();
static void         Transformation();

#ifdef DEBUG
static void         PrintLASolution();
#endif DEBUG


/* ---------- Local Global Variables ---------- */
static La_CostEntry    *c, *slotCap;
static La_CostEntry    *mback, *map, *cap, *row, *col, *nexts;
static La_CostEntry    *h1, *h2, *h3, *h4, *h5, *h6;
static int              m, n, neff;

bool printFlag = FALSE;



#ifdef DEBUG
static void
PrintLASolution() 
/*
 * Print out the solution of the linear assignment problem.
 */
{
#define druckerbreite 100

    int i, j, cij, pj;
    int sum = 0;
    int max = 0;
    int ij = 0;

    (void) fprintf (stdout, "\n");

    for (i = 1; i <= m ;i++) {
        cij = (c + ij + (map + (col + i)->value)->value)->value;

        (void) fprintf (stdout,
            "row: %4d, col: %4d,    cij: %d \n", i, (col+i)->value, cij);

        sum += cij;
        if (cij > max) {
            max = cij;
        }
        ij += n;
    }

    (void) fprintf (stdout, "\nAssigned slot:");
    for (i = 1 ; i <= m ; i++) {
        (void) fprintf (stdout, " %d", (map+(col+i)->value)->value);
    }
    (void) fprintf (stdout, "\n\n");

    if ((n+1)*4 <= druckerbreite) {
        /* read CAUTION on DoLinearAssign  */
        for (j = 1; j <= n; j++) {
            (void) fprintf (stdout, "%d ", (slotCap + j)->value);
        }
        (void) fprintf (stdout, "\n\n");
        ij = 0;
        for (i = 1; i <= m; i++) {
            for (j = 1; j <= n; j++) {
                ij++;
                if ((map+(col+i)->value)->value == j) {
                    (void) fprintf (stdout, "[%d] ", (c+ij)->value);
                } else {
                    (void) fprintf (stdout, "%d ",(c+ij)->value);
                }
            }
            (void) fprintf (stdout, "\n");
        }
        (void) fprintf (stdout, "\n");
    }
    (void) fflush (stdout);
}
#endif DEBUG


static BearStatus
CheckLASolution() 
{
    int i, j, cij, pj;
    int sum = 0;
    int max = 0;
    int ij = 0;

    for (i = 1; i <= m ;i++) {
        cij = (c + ij + (map + (col + i)->value)->value)->value;

        if (cij == LA_INFINITE_COST) {
            (void) fprintf (stderr,
                "Linear assignment failed : cannot find feasible solution.\n");
            return (BEAR_ERROR);
        }

        sum += cij;
        if (cij > max) {
            max = cij;
        }
        ij += n;
    }

    if (printFlag == TRUE) {
        (void) fprintf (stdout,
            "Total cost of edges = %d  Maximum edge cost = %d\n", sum, max);
    }

    return (BEAR_OK);
}



static void
SetUpInternalArrays() 
{
    int j, licz, pom, i;

    col =   (La_CostEntry *) calloc((m + 1), sizeof (La_CostEntry));
    nexts = (La_CostEntry *) calloc((m + 1), sizeof (La_CostEntry));
    h1 =    (La_CostEntry *) calloc((m + 1), sizeof (La_CostEntry));
    h3 =    (La_CostEntry *) calloc((m + 1), sizeof (La_CostEntry));

    mback = (La_CostEntry *) calloc((n + 1), sizeof (La_CostEntry));

    cap =   (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));
    row =   (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));
    h2 =    (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));
    h4 =    (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));
    h5 =    (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));
    h6 =    (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));
    map =   (La_CostEntry *) calloc((neff + 1), sizeof (La_CostEntry));

/* All capacities are 1 */

    for (j=1;j<=neff;j++) {
        (cap+j)->value=1;
    }

/* mback maps columns with  > 1 capacity into first corresponding unit 
capacity column */

    (mback+1)->value = 1;

    if (n>1) {
        for (j=2;j<=n;j++) {
            (mback+j)->value=(mback+j-1)->value+(slotCap+j)->value;
        }
    }
/* map maps unit capacity columns into original columns with > 1 capacity*/

    licz=0;
    for (j=1;j<=n;j++) {
        pom=(slotCap+j)->value;
        for (i=1;i<=pom;i++) {
            licz++;
            (map+licz)->value=j;
        }
    }

#ifdef DEBUG
    if (printFlag == TRUE) {
        (void) fprintf (stdout, "\nMback:");
        for (i=1;i<=n;i++) {
            (void) fprintf (stdout, "%d ", (mback+i)->value);
        }

        (void) fprintf (stdout, "\nMap:");
        for (i=1;i<=neff;i++) {
            (void) fprintf (stdout, "%d ", (map+i)->value);
        }
        (void) fprintf (stdout, "\n");

        (void) fflush (stdout);
    }
#endif DEBUG
}



static void
FreeInternalArrays()
{
    cfree((char *) mback);
    cfree((char *) map);
    cfree((char *) cap);
    cfree((char *) col);
    cfree((char *) row);
    cfree((char *) nexts);
    mback = map = cap = col = row = nexts = (La_CostEntry *) NULL;

    cfree((char *) h1);
    cfree((char *) h2);
    cfree((char *) h3);
    cfree((char *) h4);
    cfree((char *) h5);
    cfree((char *) h6);
    h1 = h2 = h3 = h4 = h5 = h6 = (La_CostEntry *) NULL;
}



/**********************
    updating of ys and yt.
**********************/

static void
Transformation(ys, yt, dplus, dminus, d) 
La_CostEntry    *ys, *yt, *dplus, *dminus;
int             d;
{
    int i;

    for (i = 1; i <= m; i++) {
        if ((dplus +i)->value != LA_INFINITE_COST) 
            (ys+i)->value = (ys+i)->value + d - (dplus+i)->value;
    }
    for (i = 1; i <= neff; i++) {
        if ((dminus+i)->value <    d)
            (yt+i)->value = (yt+i)->value - d + (dminus+i)->value;
    }
}



/********************************************************************
 tracing back the augmenting path from index back to u, 
 assignments are updated accordingly.
********************************************************************/

static void
Augmentation(back,u,ind) 
La_CostEntry    *back;
int             u, ind;
{
    int oldind, w;

    (cap+ind)->value = (cap+ind)->value - 1;
    do {
        oldind = ind;
        w = (back+ind)->value;
        if (w != u) {
            /* remove edge w,col[w] from list associated with col[w] */
            ind = (col+w)->value;
            LA_ASSERT((row+ind)->value == w, "not top of stack#");
            (row+ind)->value = (nexts+w)->value;
        }
        /* add edge w,index to list associated with index */
        (nexts+w)->value    = (row+oldind)->value;
        (row+oldind)->value = w;
        (col+w)->value      = oldind;
    } while (w != u);
}



/*********************************************************
  Finds a shortest augmenting path from startnode i to sinknode index
  using dijkstra's algorithm and modified costs.
  next is a cyclic list of all sinknodes. next[headcand] ... lastcand 
  are the active candidate nodes. back points back in the path.
*********************************************************/

static
ShortestPath(ys,yt,dplus,dminus,back,next,i,index,d)
La_CostEntry    *ys,*yt,*dplus,*dminus,*back,*next;
int             i, *index, *d;
{
    int lastcand,headcand,is,v,vgl,j,dalt,ysi;
    for (j = 1; j <= neff; j++) (dminus+j)->value = LA_INFINITE_COST; 
    for (j = 1; j <= m; j++)    (dplus +j)->value = LA_INFINITE_COST;
    lastcand = neff;    headcand = neff;    *d = 0;
    do { /* till sinknode with free capacity is reached */
        (dplus+i)->value = *d; dalt = *d; *d = LA_INFINITE_COST;
        v = headcand; ysi = (ys+i)->value; is = (i-1)*n;
        do { /* for all active and feasible sinknodes */
            j = (next+v)->value;
            if (((cap+j)->value > 0) || ((row+j)->value > 0)) {
                /* update distances */
                vgl = dalt + (c+is+(map+j)->value)->value -ysi -(yt+j)->value;
                if (vgl < (dminus+j)->value) {
                    (dminus+j)->value  = vgl;
                    (back+j)->value    = i;
                }
                /*  select nearest node */
                if ((dminus+j)->value < *d) {
                    *d       = (dminus+j)->value;
                    (*index) = v;
                }
            }
            v = j;
        } while (v != lastcand);
        /* remove (next+v)->value from candidate set */
        v = (*index); (*index) = (next+v)->value;
        if ((*index) == lastcand) lastcand = v;
        else {
            (next+v)->value        = (next+(*index))->value;
            (next+(*index))->value = (next+headcand)->value;
            (next+headcand)->value = (*index);
            headcand                = (*index);
        }
        i = (row+(*index))->value;
    } while ((cap+(*index))->value <= 0);
}




/*****************************************************************
 Initializes, checks feasability and constructs an optimal assignment 
 for a subset of rows and columns :
 First we calculate h1[i] = min c[i,j] and  h2[j] = min (c[i,j] - h1[i])
                              j                      i
 then "admissible" assignments (i,j) with c[i,j] = h1[i] + h2[j] 
   are made.
*****************************************************************/

static void
InitAssign()
{
    int i,j,ij,j0,ui,cij;

    for (j = 1; j <=n; j++) {
        (row+j)->value = 0;
    }
    LA_ASSERT(neff >= m, "total cap too small#"); 
    /* calculate h1 and make trivial assignments */
    ij = 0;
    for (i = 1; i <= m; i++) {
        (col+i)->value = 0;
        ui = LA_INFINITE_COST;
        for (j = 1; j <= n; j++) {
            ij++;
            if ((c+ij)->value < ui) {
                ui = (c+ij)->value;
                j0 = j;
            }
        }
        (h1+i)->value = ui;
        if ((cap+(mback+j0)->value)->value > 0) { 
            /* trivial assignment */
            (cap+(mback+j0)->value)->value = 
                    (cap+(mback+j0)->value)->value - 1;
            (nexts+i)->value = (row+(mback+j0)->value)->value;
            (row+(mback+j0)->value)->value  = i;
            (col+i)->value   = (mback+j0)->value;
            (h2+(mback+j0)->value)->value  = 0;
        }
    }

    /* calculate h2 */
    if (m < neff) {
       for (j = 1; j <= neff; j++)  (h2+j)->value = 0;
    } else {
        for (j = 1; j <= neff; j++) {
            if ((row+j)->value == 0) {
                ij = (map+j)->value; ui = LA_INFINITE_COST;
                for (i = 1; i <= m; i++ ) {
                    cij = (c+ij)->value - (h1+i)->value;
                    if (cij < ui) ui = cij;
                    ij = ij + n;
                }
            (h2+j)->value = ui;
            }
        }
    }

    /* make admissible assignments */
    for (i = 1; i <= m; i++) {
        if ((col+i)->value == 0) {
            ui = (h1+i)->value; ij = (i-1)*n;
            for (j = 1; j <= neff; j++) {
                if ((cap+j)->value > 0) {
                    if ((c+ij+(map+j)->value)->value == (ui + (h2+j)->value)) {
                        (cap+j)->value   = (cap+j)->value - 1;
                        (nexts+i)->value = (row+j)->value;
                        (row+j)->value   = i;
                        (col+i)->value   = j;
                        break;
                    }
                }
            }
        }
    }
}




/********************************************************************
For a bipartite graph with m source (i-th row) and n sinks
(j-th column) connected by edges with cost  c[i,j] >= 0, an assignment 
of rows i  to columns j  is calculated which minimizes the cost
     z = sum c[i,j] * x[i,j]
             i,j
 subject to
   sum x[i,j] <= cap[j];  sum x[i,j] = 1;  x[i,j] >= 0
    i                      j
 x[i,j] = 0 (open) or 1 (edge is connected)

 input:   capacity vector cap[1 .. n] of n sinks.
      m by n  cost matrix  c[1 .. m*n] stored by rows
          problemsize  m, n.  m <= sum cap[j]
                    j
          CAUTION: cap is destroyed during by running this procedure.
 output:  two vectors row[1:n], col[1:m] describing rows 
      and columns of the selected matrix-entries.
      row[j] points to one of the rows assigned to colum j; others
      may be accessed by nexts[row[j]] ...
 scratch: h1[1:m], h2[1:n], h3[1:m], h4[1:n], h5[1:n], h6[1:n]
 method:  First an optimal assignment for a subset of rows and columns 
          is constructed heuristically. Then shortest augmenting paths
      are generated starting from unassigned sourcenodes ending at
      sinknodes with free capacity, to determine the solution of
      the complete problem.  For details see burkhard & derigs, 
      "assignment and matching problems: solution methods with 
      fortran programs",springer 1980.
          Problems with m < n are handled as an  n by n  problem with 
      c[i,j] = highvalue for i > m.
          Sinknodes j with cap[j] are handled like a bunch of cap[j]
      nodes with capacity 1 each.
*********************************************************************/

static void
DoLinearAssign()
{
    int i,j, d;
    InitAssign();

    /* initialize candidate-list for ShortestPath */
    (h6 + 1)->value = neff;
    for (j = 2; j <= neff; j++) {
        (h6+j)->value = j-1;
    }
    for (i = 1; i <= m; i++) { 
        /* all free sourcenodes */
        if ((col+i)->value == 0) {
            ShortestPath(h1,h2,h3,h4,h5,h6,i,&j,&d);
            Augmentation(h5,i,j);
            Transformation(h1,h2,h3,h4,d);
        }
    }
}


/* 
 * ----------------------------------------------------------------------------
 *
 * La_LinAssign() --
 *
 * Note that the cost matrix is really an array in which the 0th
 * element is never used.
 *
 * As an example consider the following 3 by 5 matrix.
 *
 * numRows = 3    (* number of rows    *)
 * numCols = 5    (* number of columns *)
 *
 * r goes from 0 to 2
 * c goes from 0 to 4
 *
 *         0  1  2  3  4
 *   --------------------
 * 0 |  0  1  2  3  4  5
 * 1 |     6  7  8  9 10
 * 2 |    11 12 13 14 15
 *
 * The formula of the array index of matrix element (r, c) is
 * ((r * numCols) + c + 1)
 * 
 * Therefore the matrix element (2, 3) is equivalent to array element 14.
 *
 * The solution vector and the capacity vector are also arrays in which
 * the 0th element is never used.
 *
 * Results:
 *   BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
La_LinAssign(numRows, numCols, capVec, costMatrix, solVec)
int             numRows;        /* Number of rows in cost matrix.       */
int             numCols;        /* Number of columns in cost matrix.    */
La_CostEntry    *capVec;        /* Capacity vector.  It should have
                                 * numCols + 1 elements.
                                 */
La_CostEntry    *costMatrix;    /* Cost matrix.  It should have
                                 * numRows * numCols + 1 elements.
                                 */
La_CostEntry    *solVec;        /* Solution vector.  It should have
                                 * numRows + 1 elements.
                                 */
{
    int     i, j, ij, k, licz,pom;
    double  t1;

    m = numRows;
    n = numCols;

    if (printFlag == TRUE) {
        (void) fprintf (stdout, "\nNumber of rows    = %d", m);
        (void) fprintf (stdout, "\nNumber of columns = %d", n);
        (void) fflush (stdout);
    }

    slotCap = capVec;
    c       = costMatrix;

    neff = 0;
    for (j = 1; j <= n ;j++) {
        neff += (slotCap+j)->value;
    }

    if (printFlag == TRUE) {
        (void) fprintf (stdout, "\nNeff = %d", neff);
        (void) fflush (stdout);
    }

    SetUpInternalArrays();

    t1 = NplPtime();
    DoLinearAssign();
    if (printFlag == TRUE) {
        (void) fprintf (stdout, "\nCPU-time (Linear Assignment) = %g sec \n",
            (NplPtime()-t1));
    }

    ij = k = 1;
    for (i = 1; i <= m; i++) {
        for (j = 1; j <= n; j++) {
            if ((map+(col+i)->value)->value == j) {
                (solVec+k)->value = ij;
                k += 1;
            }
            ij++;
        }
    }

#ifdef DEBUG
    if (printFlag == TRUE) {
        PrintLASolution();
    }
#endif DEBUG

    if (CheckLASolution() == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (printFlag == TRUE) {
        (void) fprintf (stdout, "Linear assignment is finished.\n");
        (void) fflush (stdout);
    }

    FreeInternalArrays();

    return (BEAR_OK);
}


