
/* newP.h --
 *
 *     Bernhard Eschermann / Massoud Pedram
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1991 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/newP.h,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
 */

#ifndef NEWP_EXPORT_H
#define NEWP_EXPORT_H /* once-only latch */

#include "newC.h"

/*
 * Default name of the default data file !
 */
#define NPL_CONTROL_DATA    "controlData"

#define NPL_MAX_PATTERNS    11  /* # of templates currently in the library */

typedef enum _Npl_PlaceMode Npl_PlaceMode;
enum _Npl_PlaceMode {
    NPL_UNDEF_MODE, NPL_FIXED_X, NPL_FIXED_Y, NPL_ASPECT_RATIO
};

typedef struct _Npl_FullEnumData Npl_FullEnumData;
struct _Npl_FullEnumData {
    double          tolerance;
    double          areaWeight;
    double          wireWeight;
    double          critWeight;
};

typedef struct _Npl_PartialEnumData Npl_PartialEnumData;
struct _Npl_PartialEnumData {
    double          tolerance;
    double          wireWeight;
    double          critWeight;
};

/*
 * The following parameters may be set by the user.
 */
typedef struct _Npl_UserParms Npl_UserParms;
struct _Npl_UserParms {
    Npl_PlaceMode       p_mode;

    bool                p_quickPlanFlag;
    bool                p_finalOptFlag;
    int                 p_maxLoopCnt;
    double              p_borderSize;
    char                *p_tmpDir;

    bool                p_fullEnumFlag;
    Npl_FullEnumData    p_fullEnum;
    Npl_PartialEnumData p_partEnum;

    /* These parameters are for debugging only. */
    double              p_adjustSlots;
    bool                p_arrPinsFlag;
    bool                p_saveDataFlag;
    bool                p_interactiveFlag;
    bool                p_createWinFlag;
};


/* ---------- External procedures for the newP module ---------- */

EXTERN BearStatus Npl_ReadControlData
    ARGS((char *fileNamePtr, Npl_UserParms *parmsPtr));

EXTERN void Npl_ReadSpaceInfo
    NULLARGS;

EXTERN BearStatus Npl_GenSpace
    ARGS((Ncl_Node *nodePtr, double levelAdjust));

EXTERN BearStatus Npl_DimOfNode
    ARGS((Ncl_Node *nodePtr));

EXTERN BearStatus Npl_TransOfChildren
    ARGS((Ncl_Node *nodePtr));

EXTERN BearStatus Npl_Placement
    ARGS((double aspectR, CellMaster *coreMasterPtr, Npl_UserParms *parmsPtr));

EXTERN Npl_UserParms *Npl_MakeUserParms
    NULLARGS;

EXTERN void Npl_FreeUserParms
    ARGS((Npl_UserParms **parmsPtrPtr));


/* ---------- External global variables for the newP module ---------- */

extern int  npl_FactArray[];
extern int  npl_FloorplanSizeArray[];
extern int  npl_LableTable[][5];

#endif NEWP_EXPORT_H

