
/* newPInt.h --
 *
 *     Massoud Pedram
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 *     Global structure definitions and external declarations for the 
 *     new placement program.
 *
 * Copyright (C) 1990 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/newPInt.h,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
 */

#ifndef _NEW_P_INT_H
#define _NEW_P_INT_H /* once-only latch */

#ifndef _NEW_C_INT_H
#define _NEW_C_INT_H
#include <port.h>
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "list.h"
#endif _NEW_C_INT_H

#include "newP.h"
#include "geofast.h"
#include "deque.h"
#include "laLinAssign.h"



/* ------------- Definition of global structures ----------- */

typedef enum _Npl_PinKind Npl_PinKind;
enum _Npl_PinKind {
    NPL_BORDER_PIN,      /* Pins on the border of a parent cell. */
    NPL_INSIDE_PIN       /* Pins on the border of a child cell.  */
};

typedef struct _Npl_NetData Npl_NetData;
struct _Npl_NetData {
    Net             *netPtr;
    lsList          pinList;
    lsList          bakPinList;         /* Back-up pin list. */
    float           criticality; 

    bool            allBdrPinFlag;      /* TRUE if all of the pins belonging
                                         * to this net are on the border of the
                                         * chip.
                                         */
    bool            chipBdrPinFlag;     /* TRUE if one or more pins belonging
                                         * to this net are on the border of the
                                         * chip.
                                         */

    int             count;
};

typedef struct _Npl_PinData Npl_PinData;
struct _Npl_PinData {
    char            pinName[80];
    Npl_PinKind     pinKind;
    Flexibility     pinFlexType;
    Npl_NetData     *netDataPtr;

    Ncl_SideDir     initDir;     /* Direction of the pin initially           */
    Point           initPos;     /* Position of the pin initially            */

    Ncl_SideDir     tileDir;     /* Direction of pin on the tile plane       */
    Point           tilePos;     /* Coordinates of the pin on the tile plane */

    Ncl_Node        *nodePtr;
    int             row;         /* Index into 'nplCostMatrix'               */

    PinInst         *pinPtr;     /* Pointer to pin in MAC database           */

    ClientData      clientData;  /* space for rent */
};

typedef struct _Npl_JunctData Npl_JunctData;
struct _Npl_JunctData {
    Rect junctArea;
};


/* ------------- Definition of global macros --------------- */

#define NPLMAX(X,Y) ((X) >= (Y) ? (X) : (Y))
#define NPLMIN(X,Y) ((X) < (Y) ? (X) : (Y))
#define NPLMAX3(X,Y,Z) NPLMAX(NPLMAX(X,Y),Z)
#define NPLMAX4(X,Y,Z,T) NPLMAX(NPLMAX(X,Y),NPLMAX(Z,T))
#define NPLSQUARE(X) ((X) * (X))
#define NPLDIVIDE100(x) ((float) (0.01 * (x)))
#define NPLNEXT(ptr) (ptr = (ptr)->nextPtr)
#define NPLISLEAF(node) ((node)->nd_numOfChildren == 0)
#define NPLISPLEAF(node) ((node)->nd_children[0]->nd_numOfChildren == 0)

#define NplCostMatrixValue(ROW, COL) \
    (nplCostMatrix[NplCostMatrixIndex(ROW, COL)].value)


/* -------------- Definition of global constants ----------- */

#define NPLMAX_INT         2147483647  /* "infinity" */
#define NPLMAX_FLOAT       ((float) 1.0e20)
#define NPLMAX_ROOMS       5      /* max. # of rooms per template */
#define NPLMAX_ROOM_ASSIGN 120    /* factorial of NPLMAX_ROOMS */
#define NPLMAX_GOALS       4      /* # of directions (N, E, S, W) */
#define NPLMAX_ORIENT      4      /* # of possible orientations of blocks */
#define NPLSENS_TOT        1.01   /* sensitivity for total cost */
#define NPLNUM_OF_TRACKS   40


/* ---------- Internal global variables for the newP module ---------- */
extern lsList           nplTotalNodeList, nplCoreNodeList;
extern lsList           nplNetList, nplBdrPinList[4];
extern CellMaster       *nplCoreMasterPtr;
extern Point            nplDelta;
extern Transform        nplTrans[4];
extern Rect             nplInsideOutRect;

extern bool             nplSaveDataFlag;
extern bool             nplUseMSTCalcFlag;
extern int              nplMaxPinsForMST;
extern double           nplBorderSize;

extern int              nplMaxLoopCnt;
extern double           nplAdjustSlots;
extern La_CostEntry     *nplCostMatrix;
extern La_CostEntry     *nplCostOfNet;
extern int              nplNumOfCMRows, nplNumOfCMCols;
extern Npl_PinData      **nplPinData;
extern Ncl_SlotData     **nplSlot;

extern bool             nplFinalOptFlag;
extern bool             nplFullEnumFlag;
extern double           nplTolerance;
extern double           nplAreaWeight;
extern double           nplWireWeight;
extern double           nplCritWeight;

extern char             *nplDefFileNamePtr;
extern char             *nplTmpDirPtr;

extern bool             nplArrPinsFlag;         /* arrange pins flag */
extern bool             nplInteractiveFlag;
extern bool             nplCreateWinFlag;



/* ---------- Declarations of external procedures ---------- */

#include "nplTemplates.h"

extern int          util_csystem();
extern char         *util_strsav();

/*
 * nplPinArranger.c
 */
extern BearStatus   NplDoPinArrangement();

/*
 * nplBegin.c
 */
extern BearStatus   NplBeginPlacerInt();

/*
 * nplPTime.c
 */
extern double       NplPtime();

/*
 * nplF*.c
 */
extern BearStatus   NplRoomExtOfNode();

/*
 * nplmst.c
 */
extern float        NplFindMST();

/*
 * nplEnum.c
 */
extern BearStatus   NplEnumTPs();
extern BearStatus   NplUsePartialEnumTPs();
extern BearStatus   NplEnumLabels();
extern BearStatus   NplFinalOpt();

/*
 * nplIOGoals.c
 */
extern BearStatus   NplCalcIOGoals();
extern BearStatus   NplCalcIOConnectivities();
extern BearStatus   NplCalcCritNetIOConnectivities();
extern BearStatus   NplCalcLeafConnectivities();

/*
 * nplRouteEstimate.c
 */
extern BearStatus   NplAddPinSpace();

/*
 * nplEnumUtils.c
 */
extern BearStatus   NplGetCNCost();
extern BearStatus   NplGetIntraBlockL();
extern BearStatus   NplGetInterBlockL();
extern BearStatus   NplGetInterLeafL();
extern BearStatus   NplCalcCNIntraBlockL();
extern BearStatus   NplCalcCNInterBlockL();
extern BearStatus   NplCalcCNInterLeafL();
extern BearStatus   NplAssignShapeToNode();

/*
 * nplPinAssign.c
 */
extern BearStatus   NplDoPinAssignment();
extern int          NplCostMatrixIndex();

/*
 * nplBottlenecks.c
 */
extern BearStatus   NplRecalcSpace();

extern BearStatus   NplDoSpacing();

/*
 * nplUtilities.c
 */
extern BearStatus   NplCalcTileDataOfNode();
extern BearStatus   NplCalcTileDataOfLeafNode();
extern BearStatus   NplGetAbsPosOfPin();
extern BearStatus   NplCalcTouchRect();
extern BearStatus   NplCalcEncloseRect();
extern BearStatus   NplTileDirOfRect();
extern void         NplCalcInitPinLists();
extern void         NplCalcTilePinLists();
extern int          NplDistanceFromRect();
extern void         NplSwapDimOfNode();
extern void         NplSwapDimOfChildren();
extern int          NplNodeDepth();
extern Npl_PinData  *NplMakePinData();
extern void         NplFreePinData();

#ifdef DEBUG
extern void         NplCheckNetList();
extern void         NplSaveCellMaster();
#endif DEBUG

/*
 * nplGeoPen.c
 */
extern void         NplGeoPen();


/*
 * nplPrint.c
 */
extern void         NplPrintResults();
extern void         NplPrintData();
extern void         NplPrintLevelStat();
extern void         NplPrintInitPinListsOfNode();
extern void         NplPrintTilePinListsOfNode();
extern void         NplCheckCritNetConstraints();


extern bool gr_FreeMemFlag;

#endif _NEW_P_INT_H

