
/* nplBegin.c --
 *
 *     Massoud Pedram
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplBegin.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplWriteToMac.h"
#include "nplDraw.h"


typedef struct _GenPinData GenPinData;
struct _GenPinData {
    Ncl_Node    *nodePtr;
    int         dx, dy;
};


/* ---------- Local Procedures ---------- */
static Npl_NetData  *MakeNetData();
static BearStatus   DoPinOfNode();
static BearStatus   IsBorderPin();
static BearStatus   GenNetDataOfNet();
static BearStatus   GenRootNodePins();
static void         GenNetLists();
static void         InitLeafNode();
static int          CompareNet();
static lsGeneric    CopyPinData();
static BearStatus   InitGlobalVars();



/* ---------- Local Global Variables ---------- */
#ifdef DEBUG
static bool printFlag = FALSE;
#endif DEBUG


/*
 * ----------------------------------------------------------------------------
 *
 * MakeNetData() --
 *
 * Allocates memory for the structure Npl_NetData, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to the structure Npl_NetData.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static Npl_NetData *
MakeNetData()
{
    Npl_NetData *netDataPtr;

    MALLOC(Npl_NetData *, netDataPtr, sizeof (Npl_NetData));
    netDataPtr->netPtr = NIL(Net);
    netDataPtr->pinList = (lsList) NULL;
    netDataPtr->bakPinList = (lsList) NULL;
    netDataPtr->criticality = 0;
    netDataPtr->allBdrPinFlag = FALSE;
    netDataPtr->chipBdrPinFlag = FALSE;
    netDataPtr->count = 0;

    return (netDataPtr);
}


static lsGeneric
CopyPinData(item)
lsGeneric item;
{
    Npl_PinData *oldPinDataPtr = (Npl_PinData *) item;
    Npl_PinData *newPinDataPtr;

    newPinDataPtr = NplMakePinData();
    (void) sprintf (newPinDataPtr->pinName, "%d", Mac_GenerateId());
    newPinDataPtr->pinKind = NPL_INSIDE_PIN;
    newPinDataPtr->pinFlexType = oldPinDataPtr->pinFlexType;
    newPinDataPtr->tileDir = Ncl_GetOppositeSideDir(oldPinDataPtr->tileDir);
    newPinDataPtr->tilePos = oldPinDataPtr->tilePos;
    newPinDataPtr->netDataPtr = oldPinDataPtr->netDataPtr;

    return ((lsGeneric) newPinDataPtr);
}


static BearStatus
GenRootNodePins()
{
    lsGen       pinGen;
    lsList      pinList;
    Npl_PinData *pinDataPtr;
    Point       tmpPos;
    Rect        *bboxPtr;
    Transform   trans;
    double      xScale, yScale;
    int         s, deltaX, deltaY;
    int         xbot, ybot, xtop, ytop;
    int         width, height, mps, sidePinCnt;

    mps = Mac_CalcMinPinSpacing();

    if (NplCalcTileDataOfNode(ncl_ClusterTree) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    NclForeachSide(s) {
        pinList = lsCopy(nplBdrPinList[s], CopyPinData);
        lsForeachItem(pinList, pinGen, pinDataPtr) {
            if (pinDataPtr->netDataPtr->allBdrPinFlag) {
                lsGeneric item;
 
                lsDelBefore(pinGen, (lsGeneric *) &item);
                NplFreePinData(item);
            }
            else {
                pinDataPtr->nodePtr = ncl_ClusterTree;
            }
        }

        ncl_ClusterTree->nd_initPinList[s] = (lsList) NULL;
        if (ncl_ClusterTree->nd_tilePinList[s] != (lsList) NULL) {
            lsDestroy(ncl_ClusterTree->nd_tilePinList[s], (void (*)()) NULL);
        }
        ncl_ClusterTree->nd_tilePinList[s] = pinList;
    }

    bboxPtr = Mac_GetBBoxOfCellMaster(nplCoreMasterPtr);

    xScale = ((float) ncl_ClusterTree->nd_dim.p_x) /
        ((float) (bboxPtr->r_xtop - bboxPtr->r_xbot));
    yScale = ((float) ncl_ClusterTree->nd_dim.p_y) /
        ((float) (bboxPtr->r_ytop - bboxPtr->r_ybot));

    GeoTranslateTrans(&GeoIdentityTransform,
        -bboxPtr->r_xbot, -bboxPtr->r_ybot, &trans);

    height = ncl_ClusterTree->nd_tileRect.r_ytop -
        ncl_ClusterTree->nd_tileRect.r_ybot;
    width = ncl_ClusterTree->nd_tileRect.r_xtop -
        ncl_ClusterTree->nd_tileRect.r_xbot;

    xbot = ncl_ClusterTree->nd_tileRect.r_xbot;
    ybot = ncl_ClusterTree->nd_tileRect.r_ybot;
    sidePinCnt = lsLength(ncl_ClusterTree->nd_tilePinList[NCL_WEST_DIR]);
    if ((height / mps) < sidePinCnt) {
 Bear_PrintErr("GenRootNodePins() : to many pins on west side of root node.");
        return (BEAR_ERROR);
    }
    lsForeachItem(ncl_ClusterTree->nd_tilePinList[NCL_WEST_DIR], 
        pinGen, pinDataPtr) {
        GeoTransPoint(&trans, &pinDataPtr->tilePos, &tmpPos);
        pinDataPtr->tilePos.p_x = xbot;
        pinDataPtr->tilePos.p_y = ybot + tmpPos.p_y * yScale;
    }
    sidePinCnt = lsLength(ncl_ClusterTree->nd_tilePinList[NCL_SOUTH_DIR]);
    if ((width / mps) < sidePinCnt) {
 Bear_PrintErr("GenRootNodePins() : to many pins on south side of root node.");
        return (BEAR_ERROR);
    }
    lsForeachItem(ncl_ClusterTree->nd_tilePinList[NCL_SOUTH_DIR], 
        pinGen, pinDataPtr) {
        GeoTransPoint(&trans, &pinDataPtr->tilePos, &tmpPos);
        pinDataPtr->tilePos.p_x = xbot + tmpPos.p_x * xScale;
        pinDataPtr->tilePos.p_y = ybot;
    }

    GeoTranslateTrans(&GeoIdentityTransform,
        -bboxPtr->r_xbot, -bboxPtr->r_ytop, &trans);
    xbot = ncl_ClusterTree->nd_tileRect.r_xbot;
    ytop = ncl_ClusterTree->nd_tileRect.r_ytop;
    sidePinCnt = lsLength(ncl_ClusterTree->nd_tilePinList[NCL_NORTH_DIR]);
    if ((width / mps) < sidePinCnt) {
 Bear_PrintErr("GenRootNodePins() : to many pins on north side of root node.");
        return (BEAR_ERROR);
    }
    lsForeachItem(ncl_ClusterTree->nd_tilePinList[NCL_NORTH_DIR],
        pinGen, pinDataPtr) {
        GeoTransPoint(&trans, &pinDataPtr->tilePos, &tmpPos);
        pinDataPtr->tilePos.p_x = xbot + tmpPos.p_x * xScale;
        pinDataPtr->tilePos.p_y = ytop;
    }

    GeoTranslateTrans(&GeoIdentityTransform,
        -bboxPtr->r_xtop, -bboxPtr->r_ybot, &trans);
    xtop = ncl_ClusterTree->nd_tileRect.r_xtop;
    ybot = ncl_ClusterTree->nd_tileRect.r_ybot;
    sidePinCnt = lsLength(ncl_ClusterTree->nd_tilePinList[NCL_EAST_DIR]);
    if ((height / mps) < sidePinCnt) {
 Bear_PrintErr("GenRootNodePins() : to many pins on east side of root node.");
        return (BEAR_ERROR);
    }
    lsForeachItem(ncl_ClusterTree->nd_tilePinList[NCL_EAST_DIR],
        pinGen, pinDataPtr) {
        GeoTransPoint(&trans, &pinDataPtr->tilePos, &tmpPos);
        pinDataPtr->tilePos.p_x = xtop;
        pinDataPtr->tilePos.p_y = ybot + tmpPos.p_y * yScale;
    }

    NplCalcInitPinLists(ncl_ClusterTree);

    return (BEAR_OK);
}


static BearStatus 
IsBorderPin(pinPtr, allBdrPinFlagPtr)
PinInst *pinPtr;
bool    *allBdrPinFlagPtr;
{
    if (pinPtr->pinKind != MAC_BORDER_PIN) {
        *allBdrPinFlagPtr = FALSE;
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus 
GenNetDataOfNet(netPtr, clientData)
Net         *netPtr;
ClientData  clientData;
{
    Npl_NetData *netDataPtr;

    netDataPtr = MakeNetData();
    netDataPtr->netPtr = netPtr;
    netPtr->clientData = (ClientData) netDataPtr;

    netDataPtr->allBdrPinFlag = TRUE;
    (void) Mac_ItrAllPinOfNet(netPtr, IsBorderPin, &netDataPtr->allBdrPinFlag);

    lsNewEnd(nplNetList, (lsGeneric) netDataPtr, LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static int
CompareNet(item1, item2)
lsGeneric item1, item2;
{
    register Net *net1Ptr = ((Npl_NetData *) item1)->netPtr;
    register Net *net2Ptr = ((Npl_NetData *) item2)->netPtr;

    return ((net1Ptr == net2Ptr) ? 0 : ((net1Ptr < net2Ptr) ? -1 : 1));
}


static BearStatus
DoPinOfNode(pinPtr, genPinDataPtr)
PinInst     *pinPtr;
GenPinData  *genPinDataPtr;
{
    lsList      pinList;
    Ncl_SideDir sideDir;
    Net         *netPtr;
    Ncl_Node    *nodePtr;
    Npl_PinData *pinDataPtr;
    Npl_NetData *netDataPtr;

    netPtr = Mac_GetNetOfPin(pinPtr);
    netDataPtr = (Npl_NetData *) netPtr->clientData;

    pinDataPtr = NplMakePinData();
    pinDataPtr->pinPtr = pinPtr;
    pinDataPtr->netDataPtr = netDataPtr;

    switch (Mac_GetDirectionOfPin(pinPtr)) {
    case NORTH_PIN:
        sideDir = NCL_NORTH_DIR;
        break;
    case EAST_PIN:
        sideDir = NCL_EAST_DIR;
        break;
    case SOUTH_PIN:
        sideDir = NCL_SOUTH_DIR;
        break;
    case WEST_PIN:
        sideDir = NCL_WEST_DIR;
        break;
    default:
        Bear_PrintFatalErr("GenPinDataOfPin() failed : pin has bad direction");
        break;
    }

    pinDataPtr->pinFlexType = Mac_GetFlexOfPin(pinPtr);

    if (pinPtr->pinKind == MAC_BORDER_PIN) {
        netDataPtr->chipBdrPinFlag = TRUE;

        pinDataPtr->pinKind = NPL_BORDER_PIN;
        Mac_GetAbsPosOfPin(pinPtr, &pinDataPtr->tilePos);
        pinDataPtr->tileDir = sideDir;

        pinDataPtr->nodePtr = NIL(Ncl_Node);

        pinList = nplBdrPinList[Ncl_GetOppositeSideDir(sideDir)];
    }
    else {
        pinDataPtr->pinKind = NPL_INSIDE_PIN;
        Mac_GetAbsPosOfPin(pinPtr, &pinDataPtr->initPos);
        pinDataPtr->initPos.p_x += genPinDataPtr->dx;
        pinDataPtr->initPos.p_y += genPinDataPtr->dy;
        pinDataPtr->initDir = sideDir;

        nodePtr = pinDataPtr->nodePtr = genPinDataPtr->nodePtr;

        switch (pinDataPtr->pinFlexType) {
        case (Flexibility) FIXED:
            nodePtr->nd_fixPinCnt++;
            break;
        case (Flexibility) FLEXIBLE:
            nodePtr->nd_freeFloPinCnt++;
            break;
        case (Flexibility) FLEXIBLE_NORTH:
            nodePtr->nd_sideFloPinCnt[NCL_NORTH_DIR]++;
            break;
        case (Flexibility) FLEXIBLE_EAST:
            nodePtr->nd_sideFloPinCnt[NCL_EAST_DIR]++;
            break;
        case (Flexibility) FLEXIBLE_SOUTH:
            nodePtr->nd_sideFloPinCnt[NCL_SOUTH_DIR]++;
            break;
        case (Flexibility) FLEXIBLE_WEST:
            nodePtr->nd_sideFloPinCnt[NCL_WEST_DIR]++;
            break;
        default:
            Bear_PrintErr("DoPinOfNode() failed : illegal flexibility type");
            break;
        }
        nodePtr->nd_totalPinCnt++;

        lsNewEnd(nodePtr->nd_bdrNetList, (lsGeneric) netDataPtr, LS_NH);

        pinList = nodePtr->nd_initPinList[sideDir];
    }
    lsNewEnd(pinList, (lsGeneric) pinDataPtr, LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static void
InitLeafNode(leafNodePtr)
Ncl_Node *leafNodePtr;
{
    CellInst    *cellInstPtr;
    CellMaster  *cellMasterPtr;
    GenPinData  genPinData;
    Point       delta;
    int         s;

    /*
     * nd_cellInstPtr has been assigned in clustering stage. 
     */
    leafNodePtr->nd_depth = ncl_ClusterTreeDepth;
    leafNodePtr->nd_status = NOT_PLACED;
    leafNodePtr->nd_nodeTrans = GeoIdentityTransform;
    leafNodePtr->nd_tileTrans = GeoIdentityTransform;
    leafNodePtr->nd_storeTrans = GeoIdentityTransform;
    leafNodePtr->nd_type = LEAF;
    leafNodePtr->nd_pattern = 0;
    leafNodePtr->nd_orient = 0;
    NclForeachSide(s) {
        leafNodePtr->nd_tilePinList[s] =
            leafNodePtr->nd_initPinList[s] = lsCreate();
    }
    leafNodePtr->nd_totalNetList = (lsList) NULL;
    leafNodePtr->nd_bdrNetList = lsCreate();

    cellInstPtr = leafNodePtr->nd_cellInstPtr;
    cellInstPtr->clientData = (ClientData) leafNodePtr;

    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);

    GeoTransPoint(&cellInstPtr->transform,
        &(cellMasterPtr->horzRectList->r_r.r_ll), &delta);

    leafNodePtr->nd_fixPinCnt = 0;
    leafNodePtr->nd_freeFloPinCnt = 0;
    NclForeachSide(s) {
        leafNodePtr->nd_sideFloPinCnt[s] = 0;
    }
    leafNodePtr->nd_totalPinCnt = 0;

    genPinData.nodePtr = leafNodePtr;
    genPinData.dx = -(delta.p_x);
    genPinData.dy = -(delta.p_y);
    (void) Mac_ItrPinOfCellInst(cellInstPtr, 
        DoPinOfNode, (ClientData) &genPinData);

    lsSort(leafNodePtr->nd_bdrNetList, CompareNet);
    lsUniq(leafNodePtr->nd_bdrNetList, CompareNet, (void (*)()) NULL);
}


static void
GenNetLists(nodeList, netList)
lsList nodeList;
lsList netList;
{
    Ncl_Node                *nodePtr;
    Ncl_LinkedCritNetInNode  *critInNodePtr;
    Npl_NetData             *netDataPtr;
    lsGen                   netGen, nodeGen;
    int                     depth, i;

    for (depth = ncl_ClusterTreeDepth - 1; depth >= 0 ; depth--) {
        lsForeachItem(netList, netGen, netDataPtr) {
            netDataPtr->count = 0;
        }

        /* 
         * Pass 1 : make the total net list and the critical net list
         *          for each non-leaf node.
         */
        lsForeachItem(nodeList, nodeGen, nodePtr) {
            if (nodePtr->nd_depth == depth) {
                nodePtr->nd_totalNetList = lsCreate();

                for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
                    lsForeachItem(nodePtr->nd_children[i]->nd_bdrNetList,
                        netGen, netDataPtr) {
                        lsNewEnd(nodePtr->nd_totalNetList,
                            (lsGeneric) netDataPtr, LS_NH);
                    }
                }

                lsSort(nodePtr->nd_totalNetList, CompareNet);
                lsUniq(nodePtr->nd_totalNetList, CompareNet, (void (*)()) NULL);

                if (ncl_TimingFlag == TRUE) {
                    lsForeachItem(nodePtr->nd_totalNetList, netGen,
                        netDataPtr) {
                        netDataPtr->count++;
                    }

                    for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                        critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                        NPLNEXT(critInNodePtr)) {

                        netDataPtr = ((Npl_NetData *)
                            critInNodePtr->critNetPtr->netPtr->clientData);
                    }
                }
                else {
                    lsForeachItem(nodePtr->nd_totalNetList, netGen,
                        netDataPtr) {
                        netDataPtr->count++;
                    }
                }
            }
        }

        /* 
         * Pass 2 : make the border net list for each non-leaf node.
         */
        lsForeachItem(nodeList, nodeGen, nodePtr) {
            if (nodePtr->nd_depth == depth) {
                nodePtr->nd_bdrNetList = lsCreate();

                lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
                    if ((netDataPtr->chipBdrPinFlag == TRUE) ||
                        (netDataPtr->count > 1)) {
                        lsNewEnd(nodePtr->nd_bdrNetList,
                            (lsGeneric) netDataPtr, LS_NH);
                    }
                }
            }
        }
    }
}


static BearStatus
InitGlobalVars(coreMasterPtr, parmsPtr)
CellMaster      *coreMasterPtr;
Npl_UserParms   *parmsPtr;
/*
 * Initialize global variables.
 */
{
    int i;

    ncl_WireWidth = Mac_CalcMinPinSpacing();

    nplTmpDirPtr = util_strsav(parmsPtr->p_tmpDir);

    nplBorderSize = parmsPtr->p_borderSize;

    nplNumOfCMRows = nplNumOfCMCols = 0;
    nplSlot = (Ncl_SlotData **) NULL;
    nplPinData = (Npl_PinData **) NULL;
    nplCostMatrix = NIL(La_CostEntry);
    nplCostOfNet = NIL(La_CostEntry);

    nplTotalNodeList = lsCreate();
    nplCoreNodeList = (lsList) NULL;
    nplNetList = lsCreate();
    NclForeachSide(i) {
        nplBdrPinList[i] = lsCreate();
    }
    nplInsideOutRect.r_xbot = INFINITY;
    nplInsideOutRect.r_xtop = MINFINITY;
    nplInsideOutRect.r_ybot = INFINITY;
    nplInsideOutRect.r_ytop = MINFINITY;

    nplTrans[0] = GeoIdentityTransform;
    nplTrans[1] = GeoUpsideDownTransform;
    GeoTransTrans(&GeoSidewaysTransform, &GeoUpsideDownTransform, &nplTrans[2]);
    nplTrans[3] = GeoSidewaysTransform;

    nplDelta.p_x = nplDelta.p_y = 0;

    nplUseMSTCalcFlag = FALSE;
    nplMaxPinsForMST = 8;

    nplCoreMasterPtr = coreMasterPtr;

#ifndef BEAR_BLACK_BOX
    nplPlaceWinPtr = NIL(Npl_PlaceWin);
#endif BEAR_BLACK_BOX

    nplFullEnumFlag = parmsPtr->p_fullEnumFlag;
    nplFinalOptFlag = parmsPtr->p_finalOptFlag;
    nplMaxLoopCnt = parmsPtr->p_maxLoopCnt;
    if (parmsPtr->p_quickPlanFlag == TRUE) {
        nplFullEnumFlag = FALSE;
        nplFinalOptFlag = FALSE;
        nplMaxLoopCnt = 0;
    }

    if (parmsPtr->p_fullEnumFlag == TRUE) {
        nplTolerance = parmsPtr->p_fullEnum.tolerance;
        nplAreaWeight = parmsPtr->p_fullEnum.areaWeight;
        nplWireWeight = parmsPtr->p_fullEnum.wireWeight;
        nplCritWeight = parmsPtr->p_fullEnum.critWeight;
    }
    else {
        nplTolerance = parmsPtr->p_partEnum.tolerance;
        nplAreaWeight = 0.0;    /* Should not be used in this case. */
        nplWireWeight = parmsPtr->p_partEnum.wireWeight;
        nplCritWeight = parmsPtr->p_partEnum.critWeight;
    }

    if (ncl_TimingFlag == FALSE) {
        nplCritWeight = 0.0;    /* Should not be used in this case. */
    }

    /* These parameters are for debugging only. */
    nplAdjustSlots = parmsPtr->p_adjustSlots;
    nplArrPinsFlag = parmsPtr->p_arrPinsFlag;
    nplSaveDataFlag = parmsPtr->p_saveDataFlag;
    nplInteractiveFlag = parmsPtr->p_interactiveFlag;
    nplCreateWinFlag = parmsPtr->p_createWinFlag;

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * NplBeginPlacerInt() --
 *
 *     This procedure initializes some variables, partly with the help
 *     of information supplied by the user.
 *
 * Results:
 *     BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *     ncl_ClusterTree->nd_dim is set, ncl_ClusterTree->nd_nodeTrans is
 *     initialized to the identity matrix.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
NplBeginPlacerInt(aspectR, coreMasterPtr, parmsPtr)
double          aspectR;
CellMaster      *coreMasterPtr;
Npl_UserParms   *parmsPtr;          /* IN */
{
    Ncl_ShapeSet        *shapeSet;
    Deque               *initQueue = MakeDeque(); /* Yet another FIFO. */
    CellInst            *cellInstPtr;
    CellMaster          *cellMasterPtr;
    Ncl_LinkedCritNet   *lnkCritNetPtr;
    LinkedLineSeg       *lnLineSegPtr;
    LinkedPoint         *pointList;
    LinkedRect          *rectList;
    Ncl_Node            *nodePtr, *childNodePtr;
    Npl_NetData         *netDataPtr;
    PlaneData           *corePlaneDataPtr;
    float               cur_aspect, cur_diff, min_diff, chipHeight;
    int                 bestShapeNr, nodeCnt, i;
    char                blockName[80];
    char                blockMasterName[80];
    float               weightSum;

    if (InitGlobalVars(coreMasterPtr, parmsPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    /*
     * Check validity of global variables.
     */
    if ((nplMaxLoopCnt < 0) || (nplMaxLoopCnt > 2)) {
Bear_PrintErr("NplBeginPlacerInt(): max loop count out of range (0, 1, or 2)");
        return (BEAR_ERROR);
    }

    if (nplTolerance < 0.05) {
   Bear_PrintErr("NplBeginPlacerInt(): Aspect tolerance low.  Reset to 0.05.");
        nplTolerance = 0.05;
    }

    /* 
     * We normalize weight sum to 1.0
     */
    if ((nplFullEnumFlag == TRUE) && (nplAreaWeight < 0.0)) {
        Bear_PrintErr("NplBeginPlacerInt() : AreaWeight < 0.0");
        return (BEAR_ERROR);
    }
    if ((ncl_TimingFlag == TRUE) && (nplCritWeight < 0.0)) {
        Bear_PrintErr("NplBeginPlacerInt() : CritWeight < 0.0");
        return (BEAR_ERROR);
    }
    if (nplWireWeight < 0.0) {
        Bear_PrintErr("NplBeginPlacerInt() : WireWeight < 0.0");
        return (BEAR_ERROR);
    }

    weightSum = nplAreaWeight + nplCritWeight + nplWireWeight;

    if (weightSum <= 0) {
        Bear_PrintErr("NplBeginPlacerInt() : sum of weights <= 0");
        return (BEAR_ERROR);
    }
    else {
        nplAreaWeight /= weightSum;
        nplWireWeight /= weightSum;
        if (ncl_TimingFlag == TRUE) {
            nplCritWeight /= weightSum;
        }
    }

#ifdef DEBUG
    if (nplSaveDataFlag) {
        char command[BEAR_MSG_SIZE];

        (void) sprintf (command, "rm -f %s/%s* >& /dev/null", nplTmpDirPtr,
            "bear");
        (void) util_csystem(command);
/*
 * (gyorgy : 24 oct 90) This can fail because the directory is bad
 *    or because the directory is empty.  Since the latter is acceptable
 *    we ignor the error status.
 *
 *      if (util_csystem(command) != 0) {
 *          (void) sprintf (bear_ErrorMsg, "sys cmd : \"%s\" failed.", command);
 *          Bear_PrintErr(bear_ErrorMsg);
 *          return (BEAR_ERROR);
 *      }
 */
    }
#endif DEBUG

    nplCoreMasterPtr->contents.cellInstList = NIL(CellInst);
    if ((corePlaneDataPtr = Mac_GetCellPlaneOfCell(nplCoreMasterPtr)) !=
        NIL(PlaneData)) {
        Tp_FreePlaneData(corePlaneDataPtr);
        Mac_PutCellPlaneOfCell(NIL(PlaneData), nplCoreMasterPtr);
    }

    (void) Mac_ItrNetOfCell(nplCoreMasterPtr, GenNetDataOfNet,
        (ClientData) NULL);
    nplCoreMasterPtr->contents.netList = NIL(Net); 

    (void) fprintf (bearout, "\nNew Placement\n");
    (void) fprintf (bearout,   "=============\n");
    (void) fflush (bearout);

    if (ncl_TimingFlag) {
        /* 
         * We set the used_l of crit nets to zero for the root node.
         */
        for (lnkCritNetPtr = ncl_CritNetList;
            lnkCritNetPtr != NIL(Ncl_LinkedCritNet);
            NPLNEXT(lnkCritNetPtr)) {

            lnkCritNetPtr->used_l = 0;
        }
    }
    else {
        ncl_CritNetList = NIL(Ncl_LinkedCritNet);
    }

    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(nplCoreMasterPtr),
        DoPinOfNode, (ClientData) NULL);

    InjectDeque(initQueue, (ClientData) ncl_ClusterTree);
    ncl_ClusterTree->nd_id = nodeCnt = 0;
    ncl_ClusterTree->nd_index = 0;
    nodeCnt++;

    while (! NPLISLEAF(nodePtr = (Ncl_Node *) LookUpHead(initQueue))) {

        nodePtr = (Ncl_Node *) PopDeque(initQueue);

        lsNewEnd(nplTotalNodeList, (lsGeneric) nodePtr, LS_NH);

        nodePtr->nd_depth = NplNodeDepth(nodePtr);
        if (nodePtr != ncl_ClusterTree) {
            nodePtr->nd_dim.p_x = nodePtr->nd_dim.p_y = 0;
            nodePtr->nd_storeDim.p_x = nodePtr->nd_storeDim.p_y = 0;
            nodePtr->nd_area = 0;
        }
        nodePtr->nd_orient = 0;
        nodePtr->nd_cellInstPtr = NIL(CellInst);
        NclForeachSide(i) {
            nodePtr->nd_tilePinList[i] =
                nodePtr->nd_initPinList[i] = lsCreate();
        }
        nodePtr->nd_totalNetList = (lsList) NULL;
        nodePtr->nd_bdrNetList = (lsList) NULL;

        nodePtr->nd_nodeTrans = GeoIdentityTransform;
        nodePtr->nd_storeTrans = GeoIdentityTransform;
        if (nodePtr == ncl_ClusterTree) {
            nodePtr->nd_status = PLACED;
            if (NplCalcTileDataOfNode(nodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
        else {
            nodePtr->nd_status = NOT_PLACED;
            nodePtr->nd_tileTrans = GeoIdentityTransform;
        }

        if (NPLISPLEAF(nodePtr)) {
            nodePtr->nd_type = P_LEAF;

            for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
                childNodePtr = nodePtr->nd_children[i];
                InitLeafNode(childNodePtr);
                lsNewEnd(nplTotalNodeList, (lsGeneric) childNodePtr, LS_NH);
            }
        }
        else {
            nodePtr->nd_type = NON_LEAF;
        }

        for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
            childNodePtr = nodePtr->nd_children[i];
            InjectDeque(initQueue, (ClientData) (childNodePtr));
            childNodePtr->nd_id = nodeCnt;
            childNodePtr->nd_index = i;
            nodeCnt++;
        }
    }
    DeleteDeque(initQueue);

    GenNetLists(nplTotalNodeList, nplNetList);
#ifdef DEBUG
    if (printFlag == TRUE) {
        NplPrintData(nplTotalNodeList, nplNetList);
    }
#endif DEBUG

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && (nplInteractiveFlag == TRUE) &&
        (nplCreateWinFlag == TRUE)) {

        (void) NplCreatePlaceWinOne();
    }
#endif /* BEAR_BLACK_BOX */

    min_diff = NPLMAX_FLOAT;
    shapeSet = ncl_ClusterTree->nd_shapeSet;

    for (i = 0; i < shapeSet->size; i++) {
        cur_aspect = shapeSet->shape[i].sd_x / shapeSet->shape[i].sd_y;
        cur_diff = ABS(cur_aspect - aspectR);
        if (cur_diff < min_diff) {
            min_diff = cur_diff;
            bestShapeNr = i;
        }
    }

    ncl_ClusterTree->nd_area = shapeSet->shape[bestShapeNr].sd_x *
        shapeSet->shape[bestShapeNr].sd_y;

    switch (parmsPtr->p_mode) {
    /* x-extent fixed */
    case NPL_FIXED_X: 
        ncl_ClusterTree->nd_dim.p_x = aspectR;
        ncl_ClusterTree->nd_dim.p_y = ncl_ClusterTree->nd_area / aspectR;
        break;
    /* y-extent fixed */
    case NPL_FIXED_Y: 
        ncl_ClusterTree->nd_dim.p_y = aspectR;
        ncl_ClusterTree->nd_dim.p_x = ncl_ClusterTree->nd_area / aspectR;
        break;
    /* aspect ratio '= y-extent / x-extent' is given */
    case NPL_ASPECT_RATIO: 
        chipHeight = sqrt(ncl_ClusterTree->nd_area * aspectR);
        ncl_ClusterTree->nd_dim.p_y = chipHeight;
        ncl_ClusterTree->nd_dim.p_x = chipHeight / aspectR;
        break;
    default: 
        Bear_PrintErr("NplBeginPlacerInt() : unknown mode");
        return (BEAR_ERROR);
        break;
    }
    ncl_ClusterTree->nd_storeDim = ncl_ClusterTree->nd_dim;

    if (GenRootNodePins() != BEAR_OK) {
        return (BEAR_ERROR);
    }

    /*
     * Make first child cell of the core cell (which is the root of the tree).
     */
    if (NplWriteNodeToMac(ncl_ClusterTree) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    for (lnLineSegPtr = ncl_ClusterTree->nd_cellInstPtr->boundary;
        lnLineSegPtr != NIL(LinkedLineSeg);
        lnLineSegPtr = lnLineSegPtr->nextPtr) {

        if (NplAttachPinListOne(lnLineSegPtr->lineSeg, ncl_ClusterTree) !=
            BEAR_OK) {
            return (BEAR_ERROR);
        }
    }

    nplCoreNodeList = lsCreate();
    lsNewEnd(nplCoreNodeList, (lsGeneric) ncl_ClusterTree, LS_NH);
    if (NplSetUpCoreMaster(nplCoreMasterPtr, nplCoreNodeList) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (NplSetUpNetList(nplCoreMasterPtr, nplCoreNodeList) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    lsDestroy(nplCoreNodeList, (void (*)()) NULL);
    nplCoreNodeList = (lsList) NULL;

#ifdef DEBUG
    NplSaveCellMaster(nplCoreMasterPtr);
#endif /* DEBUG */

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        NplInitChipWinView(globalWinDataPtr, nplCoreMasterPtr);
        NplDrawChipWin();

        if (nplInteractiveFlag) {
            if (NplKeepGoing("After placing core - Keep going") == FALSE) {
                return (BEAR_ERROR);
            }
        }
    }
#endif /* BEAR_BLACK_BOX */

    return (BEAR_OK);
}


