
/* nplBottlenecks.c --
 *
 *      Massoud Pedram / George Carvalho, Esq.
 *
 *      (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplBottlenecks.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"

typedef struct _SpacingData {
    bool        horzFlag;
    Ncl_Node    *nodePtr;
    PlaneData   *planeDataPtr;
} SpacingData;

static BearStatus
DoBnLine(bnPtr, spacingDataPtr)
BottleNeck      *bnPtr;
SpacingData     *spacingDataPtr;
{
    BNLine *bnLinePtr;

    bnLinePtr = bnPtr->unValue.bnLinePtr;
    Bear_PrintErr("DoBnLine() failed: encountered bottleneck line.");

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
DoBnTile(bnPtr, spacingDataPtr)
BottleNeck  *bnPtr;
SpacingData *spacingDataPtr;
{
    CellInst    *cellInstPtr;
    DummyCell   *dcellPtr;
    Plane       *planePtr;
    Tile        *bnTilePtr, *tilePtr;
    Ncl_Node    *nodePtr, *childNodePtr;
    int         numOfChildren, density, rowIndex, colIndex;

    bnTilePtr = bnPtr->unValue.bnTilePtr;

    nodePtr = spacingDataPtr->nodePtr;
    numOfChildren = nodePtr->nd_numOfChildren;
        
    if (Tp_IsHorzTile(bnTilePtr) == TRUE) {
        spacingDataPtr->horzFlag = TRUE;
        planePtr = Tp_GetHorzPlane(spacingDataPtr->planeDataPtr);
    }
    else {
        spacingDataPtr->horzFlag = FALSE;
        planePtr = Tp_GetVertPlane(spacingDataPtr->planeDataPtr);
    }

    /*
     * Add all nodes on the bottom (or left) sides of the bottleneck
     * tile to the list.
     */
    tilePtr = BL(bnTilePtr);
    if (tilePtr == planePtr->pl_left) {
        if (spacingDataPtr->horzFlag == TRUE) {
            rowIndex = WEST;
        } else {
            rowIndex = SOUTH;
        }
    } else if (tilePtr == planePtr->pl_right) {
        if (spacingDataPtr->horzFlag == TRUE) {
            rowIndex = EAST;
        } else {
            rowIndex = NORTH;
        }
    } else if (Tp_IsDCellTile(tilePtr) == TRUE) {
        Bear_PrintErr("DoBnTile() failed: encountered a dummy cell.");
        return (BEAR_CONTINUE_SEARCH);
    } else if (Tp_IsCellInstTile(tilePtr) == TRUE) {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        childNodePtr = (Ncl_Node *) cellInstPtr->clientData;
        if (childNodePtr == NIL (Ncl_Node)) {
            Bear_PrintErr("DoBnTile() failed: nil childNodePtr.");
            return (BEAR_STOP_SEARCH);
        }
        rowIndex = childNodePtr->nd_index;
    } else {
        Bear_PrintErr("DoBnTile() failed: encountered a route cell.");
        return (BEAR_STOP_SEARCH);
    }

    /*
     * Add all nodes on the top (or right) sides of the bottleneck
     * tile to the list.
     */
    tilePtr = TR(bnTilePtr);
    if (tilePtr == planePtr->pl_left) {
        if (spacingDataPtr->horzFlag == TRUE) {
            colIndex = WEST;
        } else {
            colIndex = SOUTH;
        }
    } else if (tilePtr == planePtr->pl_right) {
        if (spacingDataPtr->horzFlag == TRUE) {
            colIndex = EAST;
        } else {
            colIndex = NORTH;
        }
    } else if (Tp_IsDCellTile(tilePtr) == TRUE) {
        Bear_PrintErr("DoBnTile() failed: encountered a dummy cell.");
        return (BEAR_CONTINUE_SEARCH);
    } else if (Tp_IsCellInstTile(tilePtr) == TRUE) {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        childNodePtr = (Ncl_Node *) cellInstPtr->clientData;
        if (childNodePtr == NIL (Ncl_Node)) {
            Bear_PrintErr("DoBnTile() failed: nil childNodePtr.");
            return (BEAR_STOP_SEARCH);
        }
        colIndex = childNodePtr->nd_index;
    } else {
        Bear_PrintErr("DoBnTile() failed: encountered a route cell.");
        return (BEAR_STOP_SEARCH);
    }

    if (rowIndex > colIndex) {
        int temp;

        temp = rowIndex;
        rowIndex = colIndex;
        colIndex = temp;
    }

    if (rowIndex >= numOfChildren && colIndex >= numOfChildren) {
        Bear_PrintErr("DoBnTile() failed: bad indices for nd_spacing.");
        return (BEAR_STOP_SEARCH);
    }

    density = Tp_GetDensityOfTile(bnTilePtr);
    nodePtr->nd_spacing[rowIndex][colIndex] = NPLMAX(density,
            nodePtr->nd_spacing[rowIndex][colIndex]);

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
DoBottleNeck(bnPtr, spacingDataPtr)
BottleNeck   *bnPtr;
SpacingData     *spacingDataPtr;
{
    if (spacingDataPtr == NIL(SpacingData)) {
        Bear_PrintErr("DoBottleNeck() failed: spacingDataPtr is nil.");
        return (BEAR_STOP_SEARCH);
    }

    if (bnPtr->unType == TP_BN_TILE) {
        if (DoBnTile(bnPtr, spacingDataPtr) == BEAR_STOP_SEARCH) {
            return (BEAR_STOP_SEARCH);
        }
    }
    else {
        (void) DoBnLine(bnPtr, spacingDataPtr);
    }

    return (BEAR_CONTINUE_SEARCH);
}



BearStatus 
NplRecalcSpace(nodePtr)
Ncl_Node        *nodePtr;
{
    CellMaster  *cellMasterPtr;
    PlaneData   *planeDataPtr;
    int         width;

    if (nodePtr == NIL(Ncl_Node)) {
        Bear_PrintErr("NplRecalcSpace() failed: nodePtr is nil.");
        return (BEAR_ERROR);
    }

    cellMasterPtr = Mac_GetMasterOfCell(nodePtr->nd_cellInstPtr);
    if (cellMasterPtr == NIL(CellMaster)) {
        Bear_PrintErr("NplRecalcSpace() failed: cellMasterPtr is nil.");
        return (BEAR_ERROR);
    }

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    if (planeDataPtr == NIL(PlaneData)) {
        Bear_PrintErr("NplRecalcSpace() failed: planeDataPtr is nil.");
        return (BEAR_ERROR);
    }

    if ((Tp_GetDCellList(planeDataPtr) == NIL(DummyCell)) &&
        (Tp_GetHorzBnLineList(planeDataPtr) == NIL(BNLine)) &&
        (Tp_GetVertBnLineList(planeDataPtr) == NIL(BNLine))) {

        SpacingData     spacingData;
        FloorPlanGraph  *fpgPtr;
        int             i, j;

        spacingData.nodePtr = nodePtr;
        spacingData.planeDataPtr = planeDataPtr;

        for (i = 0; i < MAX_CHILDREN; i++) {
            for (j = i+1; j < MAX_CHILDREN+4; j++) {
                nodePtr->nd_spacing[i][j] = 0;
            }
        }

        fpgPtr = Tp_GetFpg(planeDataPtr);
        if (Fpg_ItrBottleNeckOfFpg(fpgPtr, DoBottleNeck,
                (ClientData) &spacingData) == BEAR_STOP_SEARCH) {
            return (BEAR_ERROR);
        }

        width = 2 * ncl_WireWidth;
        for (i = 0; i < MAX_CHILDREN; i++) {
            for (j = i+1; j < MAX_CHILDREN+4; j++) {
                if (nodePtr->nd_spacing[i][j] < width) {
                    nodePtr->nd_spacing[i][j] = width;
                }
            }
        }

    }

    return (BEAR_OK);
}

