
/* nplDraw.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplDraw.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#ifndef BEAR_BLACK_BOX

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "grEditor.h"
#include "list.h"
#include "newP.h"
#include "newPInt.h"
#include "nplDraw.h"


Npl_PlaceWin *nplPlaceWinPtr;


/* ---------- Local Procedures ---------- */
static Npl_PlaceWin *MakePlaceWin();


/* ---------- Local Global Variables ---------- */
static bool writePsFlag;


/* 
 * ----------------------------------------------------------------------------
 *
 * MakePlaceWin() --
 *
 * Allocates memory for the structure Npl_PlaceWin, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to the structure Npl_PlaceWin.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static Npl_PlaceWin *
MakePlaceWin()
{
    register Npl_PlaceWin *placeWinPtr;

    MALLOC(Npl_PlaceWin *, placeWinPtr, sizeof (Npl_PlaceWin));
    placeWinPtr->nodePtr = NIL(Ncl_Node);
    placeWinPtr->graph = (Gt_Graph) NULL;
    placeWinPtr->planeDataPtr = NIL(PlaneData);
    placeWinPtr->costOfNet = NIL(La_CostEntry);

    return (placeWinPtr);
}



BearStatus
NplCreatePlaceWinOne()
{
    PbWindowType    *winTypePtr;
    Npl_PlaceWin    *placeWinPtr;
    Rect            *imageBndPtr = NIL(Rect);
    WindowData      *winDataPtr;
    char            *geometryPtr = NIL(char);
    char            prompt[80];
    char            tbText[80]; /* title bar text */

    sprintf (prompt, "BEAR ( %s ) : PLACEMENT", pbHostName);
    sprintf (tbText, "BEAR ( %s ) : PLACEMENT", pbHostName);
    if (Ez_CreateWin(prompt, geometryPtr, tbText,
        (void (*)()) NplDrawPlacementRegion,
        imageBndPtr, &winDataPtr) != EZ_OK) {
        return (BEAR_ERROR);
    }

    placeWinPtr = MakePlaceWin();
    placeWinPtr->winDataPtr = winDataPtr;
    nplPlaceWinPtr = placeWinPtr;

    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_PLACEMENT_WIN;
    winTypePtr->unValue.placeWinPtr = (BearGeneric) placeWinPtr;

    winDataPtr->wd_closeWinProc = NplClosePlaceWin;
    winDataPtr->wd_clientData = (ClientData) winTypePtr;

    return (BEAR_OK);
}



/* ARGSUSED */
void
NplClosePlaceWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
{
    Npl_PlaceWin *placeWinPtr;

    nplPlaceWinPtr = NIL(Npl_PlaceWin);

    placeWinPtr = (Npl_PlaceWin *) PbGetPlaceWin(winDataPtr);
    FREE((char *) placeWinPtr);
    FREE((char *) winDataPtr->wd_clientData);
    (void) Ez_DestroyWin(winDataPtr);
}



void
NplOutlineNodeList(nodeList)
lsList nodeList;
{
    lsGen       nodeGen;
    Ncl_Node    *nodePtr;

    lsForeachItem(nodeList, nodeGen, nodePtr) {
        if (nodePtr->nd_status == PLACED) {
            Ez_OutlineRect(globalWinDataPtr->wd_userWin,
                globalWinDataPtr->wd_scrMatrix,
                &nodePtr->nd_tileRect, EZ_SOLIDLINE1,
                ez_HighLightColor, pbChipColor);
        }
    }
    Ez_Flush();
}


void
NplShowPinDir(nodePtr)
Ncl_Node *nodePtr;
{
    lsGen       netGen, pinGen;
    Npl_PinData *pinDataPtr;
    Npl_NetData *netDataPtr;
    char        *letterPtr;

    lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
        lsForeachItem(netDataPtr->pinList, pinGen, pinDataPtr) {
            switch (pinDataPtr->tileDir) {
            case NCL_NORTH_DIR:
                letterPtr = "N";
                break;
            case NCL_EAST_DIR:
                letterPtr = "E";
                break;
            case NCL_SOUTH_DIR:
                letterPtr = "S";
                break;
            case NCL_WEST_DIR:
                letterPtr = "W";
                break;
            default:
         Bear_PrintFatalErr("NplShowPinDir() failed : illegal side direction");
                break;
            }

            Ez_WorldText(globalWinDataPtr->wd_userWin,
                globalWinDataPtr->wd_scrMatrix,
                pinDataPtr->tilePos.p_x, pinDataPtr->tilePos.p_y,
                letterPtr);
        }
    }
    Ez_Flush();
}


void
NplInitChipWinView(winDataPtr, cellMasterPtr)
WindowData  *winDataPtr;
CellMaster  *cellMasterPtr;
{
    ChipWin     *chipWinPtr;
    PlaneData   *planeDataPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    if (PbGetGCellOfCHW(chipWinPtr) != NIL(GenericCell)) {
        PbPutGCellOfCHW(NIL(GenericCell), chipWinPtr);
    }
    PbPutCellMasterOfCHW(cellMasterPtr, chipWinPtr);

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    PbViewEntireObject(winDataPtr, 
        Tp_GetHorzPlaneBnd(planeDataPtr), TRUE);
}


void
NplInitPlaceWinView()
{
    PbViewEntireObject(nplPlaceWinPtr->winDataPtr,
        Tp_GetHorzPlaneBnd(nplPlaceWinPtr->planeDataPtr), TRUE);
    Ez_StretchRect(&nplPlaceWinPtr->winDataPtr->wd_imageBnd, 1.1);
    (void) Ez_CalcViewingTrans(nplPlaceWinPtr->winDataPtr);
}


void
NplDrawSlotsOfNode(winDataPtr, nodePtr, costOfNet,
    regSlotColor, zeroSlotColor)

WindowData      *winDataPtr;
Ncl_Node        *nodePtr;
La_CostEntry    *costOfNet;
unsigned int    regSlotColor;
unsigned int    zeroSlotColor;
{
    Ncl_SlotData     *slotDataPtr;
    Ncl_Node        *childNodePtr;
    unsigned int    dotColor;
    int             i, j;

    
    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];

        for (j = 0; j < childNodePtr->nd_numOfNorthSlots ;j++) {
            slotDataPtr = &childNodePtr->nd_northSlot[j];

            if (costOfNet != NIL(La_CostEntry)) {
                dotColor = (costOfNet[1 + slotDataPtr->col].value == 0)
                    ? zeroSlotColor : regSlotColor;
            }
            else {
                dotColor = regSlotColor;
            }
            Ez_DrawWorldSmallDot(winDataPtr->wd_userWin,
                winDataPtr->wd_scrMatrix,
                slotDataPtr->tilePos.p_x,
                slotDataPtr->tilePos.p_y,
                dotColor);
        }
        for (j = 0; j < childNodePtr->nd_numOfEastSlots ;j++) {
            slotDataPtr = &childNodePtr->nd_eastSlot[j];

            if (costOfNet != NIL(La_CostEntry)) {
                dotColor = (costOfNet[1 + slotDataPtr->col].value == 0)
                    ? zeroSlotColor : regSlotColor;
            }
            else {
                dotColor = regSlotColor;
            }
            Ez_DrawWorldSmallDot(winDataPtr->wd_userWin,
                winDataPtr->wd_scrMatrix,
                slotDataPtr->tilePos.p_x,
                slotDataPtr->tilePos.p_y,
                dotColor);
        }
        for (j = 0; j < childNodePtr->nd_numOfSouthSlots ;j++) {
            slotDataPtr = &childNodePtr->nd_southSlot[j];

            if (costOfNet != NIL(La_CostEntry)) {
                dotColor = (costOfNet[1 + slotDataPtr->col].value == 0)
                    ? zeroSlotColor : regSlotColor;
            }
            else {
                dotColor = regSlotColor;
            }
            Ez_DrawWorldSmallDot(winDataPtr->wd_userWin,
                winDataPtr->wd_scrMatrix,
                slotDataPtr->tilePos.p_x,
                slotDataPtr->tilePos.p_y,
                dotColor);
        }
        for (j = 0; j < childNodePtr->nd_numOfWestSlots ;j++) {
            slotDataPtr = &childNodePtr->nd_westSlot[j];

            if (costOfNet != NIL(La_CostEntry)) {
                dotColor = (costOfNet[1 + slotDataPtr->col].value == 0)
                    ? zeroSlotColor : regSlotColor;
            }
            else {
                dotColor = regSlotColor;
            }
            Ez_DrawWorldSmallDot(winDataPtr->wd_userWin,
                winDataPtr->wd_scrMatrix,
                slotDataPtr->tilePos.p_x,
                slotDataPtr->tilePos.p_y,
                dotColor);
        }
    }
}


void
NplDrawNameOfNode(winDataPtr, nodePtr)
WindowData  *winDataPtr;
Ncl_Node    *nodePtr;
{
    Rect    rect;
    int     x, y;
    char    nodeName[80];

    rect = nodePtr->nd_tileRect;
    x = (rect.r_xbot + rect.r_xtop) / 2;
    y = (rect.r_ybot + rect.r_ytop) / 2;

    sprintf (nodeName, "B%d", nodePtr->nd_id);
    Ez_WorldText(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        x, y, nodeName);
}


bool
NplKeepGoing(userMsgPtr)
char *userMsgPtr;
{
#define MAX_LINE_SIZE 128

    Point       screen;
    Window      win;
    WindowData  *winDataPtr;
    bool        keepGoingFlag;
    char        msg[BEAR_MSG_SIZE];
    char        line[MAX_LINE_SIZE], token[MAX_LINE_SIZE];

    (void) sprintf (msg, "%s (y or n) ? ", userMsgPtr);

start:
    PbPrint(EZ_NORMAL_FONT, msg);
    (void) Ez_GetLine(&win, &screen, line, MAX_LINE_SIZE, &pbCmdLoop);
    if (pbCmdLoop == FALSE) {
        return (FALSE);
    }
    if ((winDataPtr = Ez_GetWinDataOfWin(win)) == NIL(WindowData)) {
        PbPrint(EZ_ITALIC_FONT, "Cannot find window.\n");
        goto start;
    }

    keepGoingFlag = TRUE;
    sscanf (line, "%1s", token);
    if ((strcmp(token, "Y") == 0) ||
        (strcmp(token, "g") == 0)) {
        nplInteractiveFlag = FALSE;
    }
    else if (strcmp(token, "n") == 0) {
        keepGoingFlag = FALSE;
    }
    else if (strcmp(token, "z") == 0) {
        Ez_Zoom(winDataPtr, 1.15);
        goto start;
    }
    else if (strcmp(token, "Z") == 0) {
        Ez_Zoom(winDataPtr, 1.0 / 1.15);
        goto start;
    }
    else if (strcmp(token, "p") == 0) {
        double  tx, ty, sx, sy, r1, r2;
        char    fileName[40];

        if (Ez_IsConsoleWin(winDataPtr)) {
            PbPrint(EZ_ITALIC_FONT, "You cannot print the console window.\n");
            goto start;
        }

#ifdef USE_XDM_CODE
        PbPsPlotDialog(winDataPtr, fileName, &tx, &ty, &sx, &sy, &r1);
#else  /* USE_XDM_CODE */
        strcpy(fileName, "chip.ps");
        tx = 0.0;
        ty = 10.5;
        sx = sy = 0.0139;
        r1 = 0.0;
#endif /* USE_XDM_CODE */

        r2 = r1 + 180.0;
        if (Ez_OpenPsFile(fileName,
            winDataPtr->wd_userWinInfo.width,
            winDataPtr->wd_userWinInfo.height,
            tx, ty, sx, sy, r2) != EZ_OK) {

            return (TRUE);
        }

        if (PbIsChipWin(winDataPtr) == TRUE) {
            NplDrawChipWin();
        }
        else if (PbIsPlaceWin(winDataPtr) == TRUE) {
            writePsFlag = TRUE;
            (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
            writePsFlag = FALSE;
        }
        else {
            (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
        }

        Ez_ClosePsFile();

        printf ("\nThe postscript description is in a file named \"%s\"\n\n",
            fileName);
        fflush (stdout);

        goto start;
    }

    if (keepGoingFlag == FALSE) {
        PbPrint(EZ_ITALIC_FONT, "Placer aborted (by user).");
        return (FALSE);
    }

    return (TRUE);

#undef MAX_LINE_SIZE
}


void
NplDrawChipWin()
{
    lsGen       nodeGen;
    Ncl_Node    *nodePtr;

    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

    if (nplCoreNodeList != (lsList) NULL) {
        lsForeachItem(nplCoreNodeList, nodeGen, nodePtr) {
            NplDrawNameOfNode(globalWinDataPtr, nodePtr);
        }
    }
}


/*
 * ----------------------------------------------------------------------------
 *
 * NplDrawPlacementRegion() --
 *
 * This procedure clears and redraws a placement window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion, even if there is
 *   an error because the iterator should continue on to the next window.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
NplDrawPlacementRegion(winDataPtr, subWinBndPtr)
WindowData  *winDataPtr;
Rect        *subWinBndPtr;
{
    lsGen           nodeGen;
    DrawGraph       drawGraph;
    Ncl_Node        *nodePtr, *coreNodePtr;
    Npl_PlaceWin    *placeWinPtr;
    PlaneData       *planeDataPtr;
    unsigned int    backColor, graphColor, outlineColor, slotColor;
    unsigned int    regSlotColor, zeroSlotColor;
    int             i;

    if (PbIsPlaceWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    /* 
     * Redraw the whole window.
     */
    XClearWindow(ez_Display, winDataPtr->wd_userWin);

    placeWinPtr = (Npl_PlaceWin *) PbGetPlaceWin(winDataPtr);
    if ((nodePtr = placeWinPtr->nodePtr) == NIL(Ncl_Node)) {
        return (BEAR_CONTINUE_SEARCH);
    }

    backColor = ez_BackColor;
    if (writePsFlag == TRUE) {
        graphColor = ez_BlackPixel;
        outlineColor = ez_BlackPixel;
        regSlotColor = ez_WhitePixel;
        zeroSlotColor = pbNet3Color;
    }
    else {
        graphColor = pbNet2Color;
        outlineColor = ez_HighLightColor;
        regSlotColor = ez_HighLightColor;
        zeroSlotColor = pbNet1Color;
    }

    planeDataPtr = placeWinPtr->planeDataPtr;
    Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        Tp_GetHorzPlaneBnd(planeDataPtr), EZ_SOLIDLINE1,
        outlineColor, backColor);

    lsForeachItem(nplCoreNodeList, nodeGen, coreNodePtr) {
        Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
            &coreNodePtr->nd_tileRect, EZ_SOLIDLINE1,
            outlineColor, backColor);

        NplDrawNameOfNode(winDataPtr, coreNodePtr);
    }

    if (placeWinPtr->graph == (Gt_Graph) NULL) {
        NplDrawSlotsOfNode(winDataPtr,
            nodePtr, NIL(La_CostEntry), regSlotColor, regSlotColor);
    }
    else {
        lsGen           junctGen;
        Npl_JunctData   *junctDataPtr;
        Mac_GraphData   *graphDataPtr;
        WallJunction    *junctPtr;

        if (writePsFlag == FALSE) {
            graphDataPtr = (Mac_GraphData *)
                Gt_GetGraphDataOfGraph(placeWinPtr->graph);
            lsForeachItem(graphDataPtr->junctList, junctGen, junctPtr) {
                junctDataPtr = (Npl_JunctData *) junctPtr->nplModulePtr;
                Ez_StippleRect(winDataPtr->wd_userWin,
                    winDataPtr->wd_scrMatrix,
                    &junctDataPtr->junctArea,
                    ez_HighLightColor, ez_CargoNetBitmap, GXcopy);
                Ez_OutlineRect(winDataPtr->wd_userWin,
                    winDataPtr->wd_scrMatrix,
                    &junctDataPtr->junctArea,
                    EZ_SOLIDLINE1, ez_HighLightColor, pbChipColor);
            }
        }

        NplDrawSlotsOfNode(winDataPtr,
            nodePtr, placeWinPtr->costOfNet, regSlotColor, zeroSlotColor);

        drawGraph.graph = placeWinPtr->graph;
        drawGraph.planeDataPtr = planeDataPtr;
        drawGraph.graphColor = graphColor;
        drawGraph.dotFlag = TRUE;
        drawGraph.widthFlag = FALSE;
        (void) Gre_DrawGraph(winDataPtr, &drawGraph);
    }

    Ez_Flush();

    return (BEAR_CONTINUE_SEARCH);
}

#endif BEAR_BLACK_BOX


