
/* nplEnumUtils.c --
 *
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1990 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplEnumUtils.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplArrays.h"
#include "mst.h"

static bool step_like_cost = FALSE;



static void
CalcRects(nodePtr, rectArray)
Ncl_Node    *nodePtr;
Rect        rectArray[];
{
    Ncl_Node    *childPtr;
    Rect        initRect, initBnd;
    int         i;

    /* 
     * Notice that the floorplan pattern, label, and orient effects have 
     * been captured by the node->nd_nodeTrans. 
     */

    for (i=0; i < nodePtr->nd_numOfChildren; i++) {
        childPtr = nodePtr->nd_children[i];
        initRect.r_xbot = initRect.r_ybot = 0;
        initRect.r_xtop = childPtr->nd_dim.p_x;
        initRect.r_ytop = childPtr->nd_dim.p_y;
        GeoTransRect(&childPtr->nd_nodeTrans, &initRect, &rectArray[i]);
    }

#ifdef DEBUG
    if (nodePtr->nd_status != PLACED) {
        Bear_PrintErr("CalcRects failed: nodePtr is not placed.");
        return;
    }
#endif DEBUG

    /*
     * node->nd_tileRect is already been set.  However, this was more convenient.
     */
    initBnd.r_xbot = 0;
    initBnd.r_xtop = nodePtr->nd_dim.p_x;
    initBnd.r_ybot = initBnd.r_ytop = nodePtr->nd_dim.p_y;
    GeoTransRect(&nodePtr->nd_nodeTrans, &initBnd, &rectArray[NORTH]);
    initBnd.r_xbot = initBnd.r_xtop = nodePtr->nd_dim.p_x;
    initBnd.r_ybot = 0;
    initBnd.r_ytop = nodePtr->nd_dim.p_y;
    GeoTransRect(&nodePtr->nd_nodeTrans, &initBnd, &rectArray[EAST]);
    initBnd.r_xbot = 0;
    initBnd.r_xtop = nodePtr->nd_dim.p_x;
    initBnd.r_ybot = initBnd.r_ytop = 0;
    GeoTransRect(&nodePtr->nd_nodeTrans, &initBnd, &rectArray[SOUTH]);
    initBnd.r_xbot = initBnd.r_xtop = 0;
    initBnd.r_ybot = 0;
    initBnd.r_ytop = nodePtr->nd_dim.p_y;
    GeoTransRect(&nodePtr->nd_nodeTrans, &initBnd, &rectArray[WEST]);

}



BearStatus
NplGetIntraBlockL(nodePtr, shapePtr, lengthPtr)
Ncl_Node        *nodePtr;
Ncl_ShapeData   *shapePtr;
float           *lengthPtr;
{
    Ncl_LengthSet   *childLengthSet;
    Ncl_Node        *childNodePtr;
    Ncl_ShapeData   *childShapePtr;
    float           minArea, maxArea, interval;
    int             i, bucketNr;

    *lengthPtr = 0;
    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {

        childShapePtr = shapePtr->source[i];
        childNodePtr = nodePtr->nd_children[childShapePtr->belongsTo];
        childLengthSet = childNodePtr->nd_lengthSet;

        minArea = childLengthSet->length[0].a;
        maxArea = childLengthSet->length[childLengthSet->size - 1].a;

        interval = (maxArea - minArea) / childLengthSet->size;
        bucketNr = (interval == 0) ? 
            0 : 
            ((childShapePtr->sd_x * childShapePtr->sd_y - minArea) / interval);

        if (bucketNr > childLengthSet->size - 1) {
            bucketNr = childLengthSet->size - 1;
        } 
        else if (bucketNr < 0) {
            bucketNr = 0;
        }

        *lengthPtr += childLengthSet->length[bucketNr].l;
    }

#ifdef DEBUG
    if (*lengthPtr <= 0 && nodePtr->nd_type == NON_LEAF) {
 Bear_PrintErr("NplGetIntraBlockL() failed: bottom-up length is non-positive.");
        return (BEAR_ERROR);
    }
#endif DEBUG

    return (BEAR_OK);
}



BearStatus
NplGetInterBlockL(nodePtr, lengthPtr)
Ncl_Node    *nodePtr;
float       *lengthPtr;
{
    int     row, col;
    float   length;
    Rect    rectArray[MAX_CHILDREN + 4];
    Rect    touchRect, newRect, area, newArea;

    CalcRects(nodePtr, rectArray);

    *lengthPtr = 0;
    for (row=0; row < nodePtr->nd_numOfChildren; row++) {

        for (col=row+1; col < MAX_CHILDREN + 4; col++) {

            if (col >= nodePtr->nd_numOfChildren && col < MAX_CHILDREN) continue;

            touchRect.r_xbot = 
                    NPLMIN(rectArray[row].r_xtop, rectArray[col].r_xtop);
            touchRect.r_xtop = 
                    NPLMAX(rectArray[row].r_xbot, rectArray[col].r_xbot);
            touchRect.r_ybot = 
                    NPLMIN(rectArray[row].r_ytop, rectArray[col].r_ytop);
            touchRect.r_ytop = 
                    NPLMAX(rectArray[row].r_ybot, rectArray[col].r_ybot);
            GeoCanonicalRect(&touchRect, &newRect);

            area.r_xbot = (rectArray[row].r_xbot + rectArray[row].r_xtop) / 2;
            area.r_xtop = (rectArray[col].r_xbot + rectArray[col].r_xtop) / 2;
            area.r_ybot = (rectArray[row].r_ybot + rectArray[row].r_ytop) / 2;
            area.r_ytop = (rectArray[col].r_ybot + rectArray[col].r_ytop) / 2;
            GeoCanonicalRect(&area, &newArea);

            GeoClip(&newRect, &newArea);

            length = ABS(newRect.r_xtop - newRect.r_xbot) +
                    ABS(newRect.r_ytop - newRect.r_ybot);
            *lengthPtr += nodePtr->nd_connectivity[row][col] * length;
        }
    }
            
    *lengthPtr = NPLDIVIDE100(*lengthPtr);

    return (BEAR_OK);
}



BearStatus
NplGetInterLeafL(nodePtr, lengthPtr)
Ncl_Node    *nodePtr;
float       *lengthPtr;
{
    lsGen       netGen;
    Npl_NetData *netDataPtr;
    Rect        encloseRect;
    float       length;

    *lengthPtr = 0;
    lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
        if (lsLength(netDataPtr->pinList) < 2) {
     Bear_PrintErr("NplGetInterLeafL() failed: number of pins is less than 2.");
            return (BEAR_ERROR);
        }
        if ((nplUseMSTCalcFlag == TRUE) && 
            (lsLength(netDataPtr->pinList) <= nplMaxPinsForMST)) {

            length = NplFindMST(netDataPtr->pinList);
        } 
        else {
            if (NplCalcEncloseRect(netDataPtr, &encloseRect) == BEAR_ERROR) {
                lsFinish(netGen);
                return (BEAR_ERROR);
            }
            length = (encloseRect.r_xtop - encloseRect.r_xbot +
                      encloseRect.r_ytop - encloseRect.r_ybot);
        }

#ifdef DEBUG
        /* length=0 is not necessarily an error. MP 14 Dec. 19
        if (length == 0) {
    Bear_PrintErr("NplGetInterLeafL() failed: interconnection length is zero.");
            return (BEAR_ERROR);
        }
        */
#endif DEBUG

        *lengthPtr += length;
    }

    return (BEAR_OK);
}



BearStatus
NplCalcCNIntraBlockL(nodePtr)
Ncl_Node *nodePtr;
{
    Ncl_LinkedCritNetInNode *critInNodePtr;
    Ncl_LengthSet            *childLDPtr;
    Ncl_Node                *childPtr;
    float                   interval, buLength, childArea, minArea,  maxArea;
    int                     i, bucketNr;

    for (critInNodePtr = nodePtr->nd_lnkCritNetList;
            critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
            NPLNEXT(critInNodePtr)) {
        
        /* 
         * We add up lower bounds on length within children of the node. 
        */
        buLength = 0;

        for (i = 0; i < critInNodePtr->numOfChildren; i++) {
        
            childPtr = nodePtr->nd_children[i];
            childLDPtr = &(critInNodePtr->children[i]->ld);

            if (childLDPtr->size <= 0) continue;

            childArea = childPtr->nd_dim.p_x * childPtr->nd_dim.p_y;
            minArea = childLDPtr->length[0].a;
            maxArea = childLDPtr->length[childLDPtr->size - 1].a;

            interval = (maxArea - minArea) / childLDPtr->size;
            bucketNr = (interval == 0) ? 
                    0 : (int) ((childArea - minArea) / interval);

            if (bucketNr > childLDPtr->size - 1) {
                bucketNr = childLDPtr->size - 1;
            } else if (bucketNr < 0) {
                bucketNr = 0;
            }

            buLength += childLDPtr->length[bucketNr].l;
        }

#ifdef DEBUG
        /*
         * This check fails on net NRLFTA in xerox example.
         *
        if (buLength <= 0 && nodePtr->nd_type == NON_LEAF) {
            Bear_PrintErr("NplCalcCNIntraBlockL failed: bottom-up length is non-positive.");
            return (BEAR_ERROR);
        }
         *
         */
#endif DEBUG

        /* for book-keeping */
        critInNodePtr->bu_l = buLength; 
    }

    return (BEAR_OK);
}


BearStatus
NplCalcCNInterBlockL(nodePtr)
Ncl_Node *nodePtr;
{
    Ncl_LinkedCritNetInNode *critInNodePtr;
    Rect                    rectArray[MAX_CHILDREN + 4];
    Rect                    touchRect, newRect, area, newArea;
    float                   length, interLength;
    int                     row, col;

    CalcRects(nodePtr, rectArray);

    for (critInNodePtr = nodePtr->nd_lnkCritNetList;
            critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
            NPLNEXT(critInNodePtr)) {
        
        interLength = 0;
        for (row=0; row < nodePtr->nd_numOfChildren; row++) {

            for (col=row+1; col < MAX_CHILDREN + 4; col++) {

                if (col >= nodePtr->nd_numOfChildren && col < MAX_CHILDREN) continue;

                touchRect.r_xbot = 
                        NPLMIN(rectArray[row].r_xtop, rectArray[col].r_xtop);
                touchRect.r_xtop = 
                        NPLMAX(rectArray[row].r_xbot, rectArray[col].r_xbot);
                touchRect.r_ybot = 
                        NPLMIN(rectArray[row].r_ytop, rectArray[col].r_ytop);
                touchRect.r_ytop = 
                        NPLMAX(rectArray[row].r_ybot, rectArray[col].r_ybot);
                GeoCanonicalRect(&touchRect, &newRect);

                area.r_xbot =
                    (rectArray[row].r_xbot + rectArray[row].r_xtop) / 2;
                area.r_xtop =
                    (rectArray[col].r_xbot + rectArray[col].r_xtop) / 2;
                area.r_ybot =
                    (rectArray[row].r_ybot + rectArray[row].r_ytop) / 2;
                area.r_ytop =
                    (rectArray[col].r_ybot + rectArray[col].r_ytop) / 2;
                GeoCanonicalRect(&area, &newArea);

                GeoClip(&newRect, &newArea);

                length = ABS(newRect.r_xtop - newRect.r_xbot) +
                        ABS(newRect.r_ytop - newRect.r_ybot);
                interLength += critInNodePtr->connectivity[row][col] * length;
            }

        }

        /* for book-keeping */
        critInNodePtr->inter_l = NPLDIVIDE100(interLength);
    }

    return (BEAR_OK);
}



BearStatus
NplCalcCNInterLeafL(nodePtr)
Ncl_Node *nodePtr;
{
    Ncl_LinkedCritNetInNode *critInNodePtr;
    Npl_NetData  *netDataPtr;
    Rect        encloseRect;
    float       interLength = 0;
    int         numOfCritNets;

    for (critInNodePtr = nodePtr->nd_lnkCritNetList;
            critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
            NPLNEXT(critInNodePtr)) {

        netDataPtr = (Npl_NetData *) 
            critInNodePtr->critNetPtr->netPtr->clientData;
        if (lsLength(netDataPtr->pinList) < 2) {
  Bear_PrintErr("NplCalcCNInterLeafL() failed: number of pins is less than 2.");
            return (BEAR_ERROR);
        }
        if (lsLength(netDataPtr->pinList) <= nplMaxPinsForMST) {
            interLength = NplFindMST(netDataPtr->pinList);
        } else {
            if (NplCalcEncloseRect(netDataPtr, &encloseRect) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
            interLength = (encloseRect.r_xtop - encloseRect.r_xbot +
                   encloseRect.r_ytop - encloseRect.r_ybot);
        }

#ifdef DEBUG
        if (interLength == 0) {
  Bear_PrintErr("NplCalcCNInterLeafL()failed: interconnection length is zero.");
            return (BEAR_ERROR);
        }
#endif DEBUG

        /* for book-keeping */
        critInNodePtr->inter_l = interLength;
    }

    return (BEAR_OK);
}



BearStatus
NplGetCNCost(nodePtr, costPtr)
Ncl_Node    *nodePtr;
float       *costPtr;
{
    Ncl_LinkedCritNetInNode *critInNodePtr;
    float tdLength, totLength, maxLength, ratio, cost;
    int numCritNets = 0;
    
    *costPtr = cost = 0;
    for (critInNodePtr = nodePtr->nd_lnkCritNetList;
            critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
            NPLNEXT(critInNodePtr)) {
    
        numCritNets++;
        /* add up length used to reach the node. */
        tdLength = critInNodePtr->critNetPtr->used_l;
        totLength = critInNodePtr->bu_l + critInNodePtr->inter_l + tdLength;

        maxLength = critInNodePtr->critNetPtr->netPtr->maxLength;
        ratio = totLength / maxLength;
        if (ratio <= 1.0) {
            if (step_like_cost == FALSE) {
                cost += ratio;
            }
        } else {
            cost += ratio*ratio;
        }
    }

    if (numCritNets > 0) cost /= numCritNets;

    *costPtr = cost;

    return (BEAR_OK);
}




/*
 *
 * ----------------------------------------------------------------------------
 *
 * NplAssignShapeToNode()
 *
 *      Matches nodePtr to shapePtr.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Assigns temporary nd_pattern, nd_label, nd_orient, nd_dim to the 
 *      nodePtr.  In addition, this procedure assigns dimensions to children
 *      of the node and sets their nd_nodeTrans to GeoIdentityTransform if 
 *      children are not leaf nodes.  Otherwise, sets nd_nodeTrans to either
 *      GeoIdentityTransform or GeoRef45Transform depending on the
 *      orientation of sourcePtr.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
NplAssignShapeToNode(shapePtr, nodePtr)
Ncl_ShapeData   *shapePtr; /* IN */
Ncl_Node        *nodePtr;  /* IN-OUT */
{
    CellMaster      *cellMasterPtr;
    Ncl_ShapeData   *sourcePtr;
    Ncl_Node        *childNodePtr;
    int             i;

    nodePtr->nd_pattern = shapePtr->pattern;
    nodePtr->nd_label = shapePtr->label;
    nodePtr->nd_orient = shapePtr->orient;
    nodePtr->nd_dim.p_x = shapePtr->sd_x;
    nodePtr->nd_dim.p_y = shapePtr->sd_y;

    if ((shapePtr->orient < 0) || (shapePtr->orient > 3)) {
     Bear_PrintErr("NplAssignShapeToNode() failed: shape orientation is bad.");
        return (BEAR_ERROR);
    }
        
    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {

        sourcePtr = shapePtr->source[i];
        if (sourcePtr->belongsTo < 0) {
          Bear_PrintErr("NplAssignShapeToNode() failed: bad bottom-up shape.");
            return (BEAR_ERROR);
        }
        childNodePtr = nodePtr->nd_children[sourcePtr->belongsTo];

        if (nodePtr->nd_type == P_LEAF) {
            cellMasterPtr = childNodePtr->nd_cellInstPtr->masterPtr;
            if (cellMasterPtr == NIL (CellMaster)) {
          Bear_PrintErr("NplAssignShapeToNode() failed: cellMaster is NIL.");
                return (BEAR_ERROR);
            }

            if (cellMasterPtr->shapeFlag == MAC_FIXED_SHAPE) {
                if (sourcePtr->orient == 0) {
                    childNodePtr->nd_nodeTrans = GeoIdentityTransform;
                } else {
                    childNodePtr->nd_nodeTrans = GeoRef45Transform;
                }
            } 
            else {
                childNodePtr->nd_dim.p_x = sourcePtr->sd_x;
                childNodePtr->nd_dim.p_y = sourcePtr->sd_y;
                childNodePtr->nd_nodeTrans = GeoIdentityTransform;
            }

        }
        else {
            childNodePtr->nd_dim.p_x = sourcePtr->sd_x;
            childNodePtr->nd_dim.p_y = sourcePtr->sd_y;
            childNodePtr->nd_nodeTrans = GeoIdentityTransform;

        }
    }

    return (BEAR_OK);
}

