
/* nplFDimOfNode.c --
 *
 *   Massoud Pedram
 *   (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1988 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplFDim.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplArrays.h"

#define SPACE nodePtr->nd_spacing


/*
 * --------------------------------------------------------------------------
 *
 * Npl_DimOfNode() --
 *
 *  Computes the nd_dims of the node based on the nd_pattern, nd_label,
 *  and nd_orient and the nd_dim's of nd_children of the node.
 *  If nd_type = P_LEAF, the child transformation vector may be either
 *  GeoIdentityTrans or GeoRef45Trans.  Otherwise, the child transformation 
 *  vector must be GeoIdentityTrans.
 *
 * Results:
 *  None.
 *
 * Side Effects:
 *  Sets the nd_dim field of node.
 *
 * --------------------------------------------------------------------------
 */

BearStatus
Npl_DimOfNode(nodePtr)
Ncl_Node *nodePtr;
{
    Point    dimSet[MAX_CHILDREN];
    Ncl_Node *childPtr;
    int      roomNr, numOfRooms = nodePtr->nd_numOfChildren;
    int      doSwap = nodePtr->nd_orient%2;

    if (nodePtr->nd_type == LEAF) {
        Bear_PrintErr("Npl_DimOfNode failed: nodePtr cannot be a leaf.");
        return (BEAR_ERROR);
    }

    if (doSwap) {
        NplSwapDimOfChildren(nodePtr);
    }

    for (roomNr = 0; roomNr < numOfRooms; roomNr++) {

        /*
         * Note that childNr = nd_label[roomNr]. 
         */
        childPtr = nodePtr->nd_children[nodePtr->nd_label[roomNr]];

        if (childPtr->nd_nodeTrans.t_c != 0 || childPtr->nd_nodeTrans.t_f != 0) {
        Bear_PrintErr("Npl_DimOfNode failed: invalid nd_nodeTrans on childPtr.");
            return (BEAR_ERROR);
        }

        GeoTransPoint(&childPtr->nd_nodeTrans, &childPtr->nd_dim, 
                &dimSet[roomNr]);

    }

    (*(nplDimOfNodeFuncArray[nodePtr->nd_pattern]))(nodePtr, dimSet);

    if (doSwap) {
        NplSwapDimOfChildren(nodePtr);
        NplSwapDimOfNode(nodePtr);
    }

    return (BEAR_OK);
}



void
Npl_DimOfNodeFunc0(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    nodePtr->nd_dim.p_x = p[0].p_x + SPACE[0][WEST] + 
            SPACE[0][EAST]; 
    nodePtr->nd_dim.p_y = p[0].p_y + SPACE[0][NORTH] + 
            SPACE[0][SOUTH];
}
  


void
Npl_DimOfNodeFunc1(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a,b;

    nodePtr->nd_dim.p_x = p[0].p_x + p[1].p_x + SPACE[0][1] + 
            SPACE[0][WEST] + SPACE[1][EAST];

    a = p[0].p_y + SPACE[0][SOUTH] + SPACE[0][NORTH];
    b = p[1].p_y + SPACE[1][SOUTH] + SPACE[1][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX(a, b);
}
  


void
Npl_DimOfNodeFunc3(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c;

    nodePtr->nd_dim.p_x = p[0].p_x + SPACE[0][1] + p[1].p_x + 
            SPACE[1][2] + p[2].p_x + SPACE[0][WEST] + 
            SPACE[2][EAST];

    a = p[0].p_y + SPACE[0][NORTH] + SPACE[0][SOUTH];
    b = p[1].p_y + SPACE[1][NORTH] + SPACE[1][SOUTH];
    c = p[2].p_y + SPACE[2][NORTH] + SPACE[2][SOUTH];

    nodePtr->nd_dim.p_y = NPLMAX3(a, b, c);
}
  


void
Npl_DimOfNodeFunc2(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b;

    a = p[1].p_x + SPACE[0][1] + SPACE[1][EAST];
    b = p[2].p_x + SPACE[0][2] + SPACE[2][EAST];

    nodePtr->nd_dim.p_x = p[0].p_x + NPLMAX(a, b) + SPACE[0][WEST]; 

    a = p[0].p_y + SPACE[0][SOUTH] + SPACE[0][NORTH];
    b = p[1].p_y + p[2].p_y + SPACE[1][2] + SPACE[1][SOUTH] + 
            SPACE[2][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX(a, b);
}

    

void
Npl_DimOfNodeFunc10(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c, d;

    nodePtr->nd_dim.p_x = p[0].p_x + p[1].p_x + p[2].p_x + p[3].p_x + 
            SPACE[0][1] + SPACE[1][2] + SPACE[2][3] + SPACE[0][WEST] + 
            SPACE[3][EAST];

    a = p[0].p_y + SPACE[0][NORTH] + SPACE[0][SOUTH];
    b = p[1].p_y + SPACE[1][NORTH] + SPACE[1][SOUTH];
    c = p[2].p_y + SPACE[2][NORTH] + SPACE[2][SOUTH];
    d = p[3].p_y + SPACE[3][NORTH] + SPACE[3][SOUTH];

    nodePtr->nd_dim.p_y = NPLMAX4(a, b, c, d);
}
  


void
Npl_DimOfNodeFunc8(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c;

    a = p[0].p_x + p[1].p_x + SPACE[0][1] + SPACE[0][WEST];
    b = p[2].p_x + SPACE[1][2] + SPACE[2][EAST];
    c = p[3].p_x + SPACE[1][3] + SPACE[3][EAST];

    nodePtr->nd_dim.p_x = a + NPLMAX(b, c);

    a = p[0].p_y + SPACE[0][NORTH] + SPACE[0][SOUTH];
    b = p[1].p_y + SPACE[1][NORTH] + SPACE[1][SOUTH];
    c = p[2].p_y + p[3].p_y + SPACE[2][3] + SPACE[2][SOUTH] + 
            SPACE[3][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX3(a, b, c);
}



void
Npl_DimOfNodeFunc7(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c;

    a = p[0].p_x + p[3].p_x + SPACE[0][WEST] + SPACE[3][EAST];
    b = p[1].p_x + SPACE[0][1] + SPACE[1][3];
    c = p[2].p_x + SPACE[0][2] + SPACE[2][3];

    nodePtr->nd_dim.p_x = a + NPLMAX(b, c);

    a = p[0].p_y + SPACE[0][NORTH] + SPACE[0][SOUTH];
    b = p[3].p_y + SPACE[3][NORTH] + SPACE[3][SOUTH];
    c = p[1].p_y + p[2].p_y + SPACE[1][2] + SPACE[1][SOUTH] + 
            SPACE[2][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX3(a, b, c);
}



void
Npl_DimOfNodeFunc9(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c;

    a = p[1].p_x + SPACE[0][1] + SPACE[1][EAST];
    b = p[2].p_x + SPACE[0][2] + SPACE[2][EAST];
    c = p[3].p_x + SPACE[0][3] + SPACE[3][EAST];

    nodePtr->nd_dim.p_x = p[0].p_x + SPACE[0][WEST] + NPLMAX3(a, b, c);

    a = p[0].p_y + SPACE[0][NORTH] + SPACE[0][SOUTH];
    b = p[1].p_y + p[2].p_y + p[3].p_y + SPACE[1][2] + SPACE[2][3] + 
            SPACE[1][SOUTH] + SPACE[3][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX(a, b);
}



void
Npl_DimOfNodeFunc5(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b;

    a = p[3].p_x + SPACE[0][3] + SPACE[3][EAST];
    b = p[1].p_x + p[2].p_x + SPACE[0][1] + SPACE[1][2] + 
             SPACE[2][EAST];

    nodePtr->nd_dim.p_x = p[0].p_x + SPACE[0][WEST] + NPLMAX(a, b);

    a = p[1].p_y + SPACE[1][3] + SPACE[1][SOUTH];
    b = p[2].p_y + SPACE[2][3] + SPACE[2][SOUTH];
    a = NPLMAX(a, b) + p[3].p_y + SPACE[3][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX(a, p[0].p_y + SPACE[0][NORTH] + 
            SPACE[0][SOUTH]);
}



void
Npl_DimOfNodeFunc4(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c, d, e;

    a = p[0].p_x + SPACE[0][1] + SPACE[0][WEST];
    b = p[2].p_x + SPACE[2][3] + SPACE[2][WEST];

    if (p[0].p_y + SPACE[0][SOUTH] + SPACE[0][2] <
        p[1].p_y + SPACE[1][SOUTH]) {

        c = p[2].p_x + SPACE[2][WEST] + SPACE[1][2];
    } else if (p[0].p_y + SPACE[0][SOUTH] >
            p[1].p_y + SPACE[1][SOUTH] + SPACE[1][3]) {

        c = p[0].p_x + SPACE[0][WEST] + SPACE[0][3];
    } else {
        c = 0;
    }

    d = p[1].p_x + SPACE[1][EAST];
    e = p[3].p_x + SPACE[3][EAST];
    nodePtr->nd_dim.p_x = NPLMAX3(a, b, c) + NPLMAX(d, e);

    a = p[1].p_y + p[3].p_y + SPACE[1][3] + SPACE[1][SOUTH] + 
            SPACE[3][NORTH];
    b = p[0].p_y + p[2].p_y + SPACE[0][2] + SPACE[0][SOUTH] +
            SPACE[2][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX(a , b);
}



void
Npl_DimOfNodeFunc6(nodePtr, p)

    register Ncl_Node *nodePtr;
    Point p[];

{
    int a, b, c, d;

    a = p[1].p_x + SPACE[0][1] + SPACE[1][EAST];
    b = p[2].p_x + p[3].p_x + SPACE[0][2] + SPACE[2][3] + 
            SPACE[3][EAST];

    nodePtr->nd_dim.p_x = p[0].p_x + SPACE[0][WEST] + NPLMAX(a, b);

    a = p[2].p_y + SPACE[1][2] + SPACE[1][SOUTH];
    b = p[3].p_y + SPACE[1][3] + SPACE[3][NORTH];
    c = NPLMAX(a, b) + p[1].p_y + SPACE[1][SOUTH];
    d = p[0].p_y + SPACE[0][SOUTH] + SPACE[0][NORTH];

    nodePtr->nd_dim.p_y = NPLMAX(c, d);
}

#undef SPACE
