
/* nplFRoomExt.c --
 *
 *     Benjamin Chen / Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplFRoomExt.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplArrays.h"

void
ExtendChildren(nodePtr, p)
     Ncl_Node *nodePtr;
     Point p[];
{
    Ncl_Node     *childPtr;
    int      roomNr, numOfRooms = nodePtr->nd_numOfChildren;

    for (roomNr = 0; roomNr < numOfRooms; roomNr++) {
        childPtr = nodePtr->nd_children[nodePtr->nd_label[roomNr]];
        GeoTransPoint(&childPtr->nd_nodeTrans, &p[roomNr], &childPtr->nd_dim);
    }
}


/*
 * --------------------------------------------------------------------------
 *
 * NplRoomExtOfNode() --
 *
 *  This procedure
 * Results:
 *  None.
 *
 * Side Effects:
 *  Sets the nd_dim field of the all children of the node such that their
 *  dimensions are extended to fill unused spaces in the rooms.
 *
 * --------------------------------------------------------------------------
 */

BearStatus
NplRoomExtOfNode(nodePtr)
Ncl_Node    *nodePtr;
{
    Point    dimSet[MAX_CHILDREN];
    Ncl_Node *childPtr;
    int      roomNr, numOfRooms = nodePtr->nd_numOfChildren;
    int      doSwap = nodePtr->nd_orient%2;
    
    if (nodePtr->nd_type == LEAF) {
        Bear_PrintErr("NplRoomExtOfNode failed: nodePtr cannot be a leaf.");
        return (BEAR_ERROR);
    }
    
    if (doSwap == 1) {
        NplSwapDimOfChildren(nodePtr);
        NplSwapDimOfNode(nodePtr);
    }
    
    for (roomNr = 0; roomNr < numOfRooms; roomNr++) {

        childPtr = nodePtr->nd_children[nodePtr->nd_label[roomNr]];
    
        if (childPtr->nd_nodeTrans.t_c != 0 
            || childPtr->nd_nodeTrans.t_f != 0) {
            Bear_PrintErr("NplRoomExtOfNode failed: invalid nodeTrans on child.");
            return (BEAR_ERROR);
        }
    
        GeoTransPoint(&childPtr->nd_nodeTrans, &childPtr->nd_dim, 
              &dimSet[roomNr]);
    
    }
    
    (*(nplRoomExtOfNodeFuncArray[nodePtr->nd_pattern]))(nodePtr, dimSet);
    ExtendChildren(nodePtr, dimSet);
    
    if (doSwap == 1) {
        NplSwapDimOfChildren(nodePtr);
        NplSwapDimOfNode(nodePtr);
    }
    
    return (BEAR_OK);
}

void
ExtendVerticalRooms(nodePtr, p, n)
     Ncl_Node *nodePtr;
     Point p[];
     int n;  /* Number of rooms */
{
    int (*spacing)[MAX_CHILDREN+4];
    int path[MAX_CHILDREN];
    int i, maxpath = 0, maxi;

    spacing = nodePtr->nd_spacing;
    /* Calculate the slack from SOUTH to NORTH for each room */
    for (i = 0; i < n; i++) {
        if ((path[i] = spacing[i][SOUTH] + p[i].p_y + spacing[i][NORTH])
            > maxpath) {
            maxpath = path[i];
            maxi = i;
        }
    }
    for (i = 0; i < n; i++) {
        if (i == maxi) continue;
        p[i].p_y += maxpath - path[i];
    }
}


void
AdjustTwoRoomPath(slack, r1dim, r2dim)
     int slack;
     int *r1dim, *r2dim;
{
    int delta;

    if ((*r1dim) < (*r2dim)) {
        delta = (int) (((float) (*r1dim)/ (float) (*r2dim)) 
            * ((float) slack));
        (*r1dim) +=  delta;
        (*r2dim) +=  slack - delta;
    }
    else {
        delta = (int) (((float) (*r2dim)/ (float) (*r1dim)) 
            * ((float) slack));
        (*r1dim) +=  slack - delta;
        (*r2dim) +=  delta;
    }

}

void
ExtendPattern2(nodePtr, p, m)
     Ncl_Node *nodePtr;
     Point p[];
     int m;      /* Index to the child that begins pattern 2 */
{
    int (*spacing)[MAX_CHILDREN+4];
    int path1;  /* Top or left path (see template) */
    int path2;  /* Bottom or right path  (see template) */
    int slack, nr, pr;  /* n-room and p-room */
    
    nr = m + 1;
    pr = m + 2;
    spacing = nodePtr->nd_spacing;
    /* Calculate slack from WEST to EAST */
    path1 = spacing[m][pr] + p[pr].p_x + spacing[pr][EAST];
    path2 = spacing[m][nr] + p[nr].p_x + spacing[nr][EAST];
    slack = ABS(path2 - path1);
    if (path1 < path2) p[pr].p_x += slack;
    else p[nr].p_x += slack;

    /* Calculate slack from NORTH to SOUTH*/
    path1 = spacing[m][NORTH] + p[m].p_y + spacing[m][SOUTH];
    path2 = spacing[pr][NORTH] + p[pr].p_y + spacing[nr][pr]
        + p[nr].p_y + spacing[nr][SOUTH];
    slack = ABS(path2 - path1);
    if (path1 < path2)  p[m].p_y += slack;
    else AdjustTwoRoomPath(slack, &p[nr].p_y, &p[pr].p_y);
}


void
NplRoomExtOfNodeFunc0 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{   
    return;
}
  
void
NplRoomExtOfNodeFunc1 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    ExtendVerticalRooms(nodePtr, p, 2);  
}
  
void
NplRoomExtOfNodeFunc3 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    ExtendVerticalRooms(nodePtr, p, 3);  
}
  
void
NplRoomExtOfNodeFunc2 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    ExtendPattern2(nodePtr, p, 0);
}

void
NplRoomExtOfNodeFunc10 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    ExtendVerticalRooms(nodePtr, p, 4);
}
  
void
NplRoomExtOfNodeFunc8 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    ExtendVerticalRooms(nodePtr, p, 2);  
    ExtendPattern2(nodePtr, p, 1);
    /* 
     *  Do another vertical pass in case vertical height of q+p (updated in n)
     *  was greater than the vertical height of m.
     */
    ExtendVerticalRooms(nodePtr, p, 2); 
}

void
NplRoomExtOfNodeFunc7 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    int (*spacing)[MAX_CHILDREN+4];
    int pathm, pathq;

    spacing = nodePtr->nd_spacing;

    ExtendPattern2(nodePtr, p, 0);
    pathm = spacing[0][SOUTH] + p[0].p_y + spacing[0][NORTH];
    pathq = spacing[3][SOUTH] + p[3].p_y + spacing[3][NORTH];
    if (pathq < pathm) {
        p[3].p_y += pathm - pathq;
    }
    else {
        int slack = pathq - pathm;
        p[0].p_y += slack;
        AdjustTwoRoomPath(slack, &p[1].p_y, &p[2].p_y);
    }
}

void
NplRoomExtOfNodeFunc9 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    int (*spacing)[MAX_CHILDREN+4];
    int path[MAX_CHILDREN];
    int i, maxpath = 0, maxi;

    spacing = nodePtr->nd_spacing;

    /* Extend horizontal rooms */
    for (i = 1; i < MAX_CHILDREN; i++) {
        if ((path[i] = spacing[0][i] + p[i].p_x + spacing[i][EAST])
            > maxpath) {
            maxpath = path[i];
            maxi = i;
        }
    }
    for (i = 1; i < MAX_CHILDREN; i++) {
        if (i == maxi) continue;
        p[i].p_x += maxpath - path[i];
    }
    /* Calculate vertical paths */
    path[1] = spacing[0][SOUTH] + p[0].p_y + spacing[0][NORTH];
    path[2] = spacing[1][SOUTH] + p[1].p_y + spacing[1][2] + p[2].p_y 
        + spacing[2][3] + p[3].p_y + spacing[3][NORTH]; 
    if (path[1] < path[2]) {
        p[0].p_y += path[2] - path[1];
    }
    else {
        int nd, pd;
        int slack = path[1] - path[2];
        nd = (int) ((((float) p[1].p_y / (float) path[2]) * (float) slack));
        pd = (int) ((((float) p[2].p_y / (float) path[2]) * (float) slack));
        p[1].p_y += nd;
        p[2].p_y += pd;
        p[3].p_y += slack - nd - pd;
    }
}

void
NplRoomExtOfNodeFunc5 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    int (*spacing)[MAX_CHILDREN+4];
    int path1;  /* Top or left path (see template) */
    int path2;  /* Bottom or middle path  (see template) */
    int path3;  /* Right path  (see template) */
    int longPath;
    int slack;
    
    spacing = nodePtr->nd_spacing;
    /* Calculate slack from WEST to EAST */
    path1 = spacing[0][3] + p[3].p_x + spacing[3][EAST];
    path2 = spacing[0][1] + p[1].p_x + spacing[1][2]
        + p[2].p_x + spacing[2][EAST];
    slack = ABS(path2 - path1);
    if (path1 < path2) p[3].p_x += slack;
    else AdjustTwoRoomPath(slack, &p[1].p_x, &p[2].p_x);

    /* Calculate slack from NORTH to SOUTH*/
    longPath = path1 = spacing[0][NORTH] + p[0].p_y + spacing[0][SOUTH];
    path2 = spacing[3][NORTH] + p[3].p_y + spacing[1][3] 
        + p[1].p_y + spacing[1][SOUTH];
    if (path2 > longPath) longPath = path2;
    path3 = spacing[3][NORTH] + p[3].p_y + spacing[2][3] 
        + p[2].p_y + spacing[2][SOUTH];
    if (path3 > longPath) longPath = path3;
    
    if (longPath == path1) {
        p[1].p_y += path1-path2;
        p[2].p_y += path1-path3;
    }
    else if (longPath == path2) {
        /* path2 is longest path */
        p[0].p_y += path2-path1;
        p[2].p_y += path2-path3;
    }
    else if (longPath == path3) {
        /* path3 is longest path */
        p[0].p_y += path3-path1;
        p[1].p_y += path3-path2;
    }
}

void
NplRoomExtOfNodeFunc4 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    int (*spacing)[MAX_CHILDREN+4];
    int path1;  /* Top or left path (see template) */
    int path2;  /* Bottom or right path  (see template) */
    
    spacing = nodePtr->nd_spacing;
    /* Calculate slack from WEST to EAST */
    path1 = spacing[2][WEST] + p[2].p_x + spacing[2][3] + p[3].p_x 
        + spacing[3][EAST];
    path2 = spacing[0][WEST] + p[0].p_x + spacing[0][1] + p[1].p_x 
        + spacing[1][EAST];
    if (path1 < path2) AdjustTwoRoomPath(path2-path1, &p[2].p_x, &p[3].p_x);
    else AdjustTwoRoomPath(path1-path2, &p[0].p_x, &p[1].p_x);

    /* Calculate slack from NORTH to SOUTH */
    path1 = spacing[2][NORTH] + p[2].p_y + spacing[0][2] + p[0].p_y 
        + spacing[0][SOUTH];
    path2 = spacing[3][NORTH] + p[3].p_y + spacing[1][3] + p[1].p_y 
        + spacing[1][SOUTH];
    if (path1 < path2) AdjustTwoRoomPath(path2-path1, &p[0].p_y, &p[2].p_y);
    else AdjustTwoRoomPath(path1-path2, &p[1].p_y, &p[3].p_y);
}

void
NplRoomExtOfNodeFunc6 (nodePtr, p)
    Ncl_Node *nodePtr;
    Point p[];
{
    int (*spacing)[MAX_CHILDREN+4];
    int path1;  /* Top or left path (see template) */
    int path2;  /* Bottom or middle path  (see template) */
    int path3;  /* Right path  (see template) */
    int longPath;
    int slack;
    
    spacing = nodePtr->nd_spacing;
    /* Calculate slack from WEST to EAST */
    path1 = spacing[0][1] + p[1].p_x + spacing[1][EAST];
    path2 = spacing[0][2] + p[2].p_x + spacing[2][3]
        + p[3].p_x + spacing[3][EAST];
    slack = ABS(path2 - path1);
    if (path1 < path2) p[1].p_x += slack;
    else AdjustTwoRoomPath(slack, &p[2].p_x, &p[3].p_x);

    /* Calculate slack from NORTH to SOUTH*/
    longPath = path1 = spacing[0][NORTH] + p[0].p_y + spacing[0][SOUTH];
    path2 = spacing[2][NORTH] + p[2].p_y + spacing[1][2] 
        + p[1].p_y + spacing[1][SOUTH];
    if (path2 > longPath) longPath = path2;
    path3 = spacing[3][NORTH] + p[3].p_y + spacing[1][3] 
        + p[1].p_y + spacing[1][SOUTH];
    if (path3 > longPath) longPath = path3;
    
    if (longPath == path1) {
        p[2].p_y += path1-path2;
        p[3].p_y += path1-path3;
    }
    else if (longPath == path2) {
        /* path2 is longest path */
        p[0].p_y += path2-path1;
        p[3].p_y += path2-path3;
    }
    else if (longPath == path3) {
        /* path3 is longest path */
        p[0].p_y += path3-path1;
        p[2].p_y += path3-path2;
    }
}
