
/* nplFShapes.c --
 * 
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1989 Regents of the University of California
 * All rights reserved.
 *
 * rcsid $Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplFShapes.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplFShapes.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"



static void
AssignIDs(shapeSet, label, pattern, orient)
register Ncl_ShapeSet   *shapeSet;
int                     *label;
int                     pattern, orient;
{
    register int i;

    for (i = 0; i < shapeSet->size; i++) {
        shapeSet->shape[i].label = label;
        shapeSet->shape[i].pattern = pattern;
        shapeSet->shape[i].orient = orient;
    }
}



/*
 * --------------------------------------------------------------------------
 *
 * NplShapeFunc0() --
 *
 *  Computes the composite shape function of each floorplan template.
 *  'label' specifies the assignment of blocks to rooms, and
 *  'orient' specifies the orientation of template.
 *
 * Results:
 *  Returns non-adjacency cost.
 *
 * Side Effects:
 *  None.
 *
 * --------------------------------------------------------------------------
 */

Ncl_ShapeSet *
NplShapeFunc0(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    Ncl_ShapeSet    *lShapeSet, *sShapeSet;
    Ncl_ShapeSet    *mShapeSet;
    int             i;

    mShapeSet = (node->nd_children[0])->nd_shapeSet;
    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = 0;
    }

    MALLOC(Ncl_ShapeSet *, sShapeSet, sizeof (Ncl_ShapeSet));
    sShapeSet->shape = Ncl_InitShapes(mShapeSet->size);
    for (i = 0; i < mShapeSet->size; i++) {
        sShapeSet->shape[i].sd_x = mShapeSet->shape[i].sd_x;
        sShapeSet->shape[i].sd_y = mShapeSet->shape[i].sd_y;

        sShapeSet->shape[i].pattern = 0;
        sShapeSet->shape[i].label = label;
        sShapeSet->shape[i].orient = orient;
        sShapeSet->shape[i].belongsTo = 0;

        sShapeSet->shape[i].source[0] = &(mShapeSet->shape[i]);
    }
    sShapeSet->size = mShapeSet->size;

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc1(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    Ncl_CutDir      cut_horz, cut_vert;
    Ncl_ShapeSet    *sShapeSet;
    Ncl_ShapeSet    *mShapeSet, *nShapeSet;
    int             m = label[0];
    int             n = label[1];
    register int    i;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }
    sShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);

    AssignIDs(sShapeSet, label, 1, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc2(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    Ncl_CutDir      cut_horz, cut_vert;
    Ncl_ShapeSet    *lShapeSet, *sShapeSet;
    Ncl_ShapeSet    *mShapeSet, *nShapeSet, *pShapeSet;
    int             m=label[0];
    int             n=label[1];
    int             p=label[2];
    int             i;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_horz);
    sShapeSet = Ncl_AddShapeSets(lShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 2, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc3(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);
    sShapeSet = Ncl_AddShapeSets(lShapeSet, pShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 3, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc4(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    Ncl_CutDir      cut_horz, cut_vert;
    Ncl_ShapeSet    *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet    *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;
    int             m = label[0];
    int             n = label[1];
    int             p = label[2];
    int             q = label[3];
    register int    i;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, pShapeSet, cut_horz);
    rShapeSet = Ncl_AddShapeSets(nShapeSet, qShapeSet, cut_horz);
    sShapeSet = Ncl_AddShapeSets(lShapeSet, rShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 4, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc5(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, qShapeSet, cut_horz);
    sShapeSet = Ncl_AddShapeSets(rShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 5, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc6(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(pShapeSet, qShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, nShapeSet, cut_horz);
    sShapeSet = Ncl_AddShapeSets(rShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 6, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc7(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_horz);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, mShapeSet, cut_vert);
    sShapeSet = Ncl_AddShapeSets(rShapeSet, qShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 7, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc8(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } 
    else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(pShapeSet, qShapeSet, cut_horz);
    sShapeSet = Ncl_AddShapeSets(lShapeSet, rShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 8, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc9(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(nShapeSet, pShapeSet, cut_horz);
    rShapeSet = Ncl_AddShapeSets(lShapeSet, qShapeSet, cut_horz);
    sShapeSet = Ncl_AddShapeSets(rShapeSet, mShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 9, orient);

    return (sShapeSet);
}



Ncl_ShapeSet *
NplShapeFunc10(node, label, orient)
Ncl_Node    *node;
int         *label;
int         orient;
{
    int m=label[0];
    int n=label[1];
    int p=label[2];
    int q=label[3];
    int i;
    Ncl_CutDir cut_horz, cut_vert;
    Ncl_ShapeSet *lShapeSet, *rShapeSet, *sShapeSet;
    Ncl_ShapeSet *mShapeSet, *nShapeSet, *pShapeSet, *qShapeSet;

    mShapeSet = (node->nd_children[m])->nd_shapeSet;
    nShapeSet = (node->nd_children[n])->nd_shapeSet;
    pShapeSet = (node->nd_children[p])->nd_shapeSet;
    qShapeSet = (node->nd_children[q])->nd_shapeSet;

    for (i = 0; i < mShapeSet->size; i++) {
        mShapeSet->shape[i].belongsTo = m;
    }
    for (i = 0; i < nShapeSet->size; i++) {
        nShapeSet->shape[i].belongsTo = n;
    }
    for (i = 0; i < pShapeSet->size; i++) {
        pShapeSet->shape[i].belongsTo = p;
    }
    for (i = 0; i < qShapeSet->size; i++) {
        qShapeSet->shape[i].belongsTo = q;
    }

    if (orient == 0) {
        cut_horz = HORZ;
        cut_vert = VERT;
    } else {
        cut_horz = VERT;
        cut_vert = HORZ;
    }

    lShapeSet = Ncl_AddShapeSets(mShapeSet, nShapeSet, cut_vert);
    rShapeSet = Ncl_AddShapeSets(pShapeSet, qShapeSet, cut_vert);
    sShapeSet = Ncl_AddShapeSets(lShapeSet, rShapeSet, cut_vert);

    FREE((char *) lShapeSet->shape);
    FREE((char *) lShapeSet);
    lShapeSet = NIL(Ncl_ShapeSet);

    FREE((char *) rShapeSet->shape);
    FREE((char *) rShapeSet);
    rShapeSet = NIL(Ncl_ShapeSet);

    AssignIDs(sShapeSet, label, 10, orient);

    return (sShapeSet);
}

