
/* nplFTrans.c --
 *
 *    Massoud Pedram
 *    (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1988 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplFTrans.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "newPInt.h"
#include "nplArrays.h"

#define NO_THIN_BOTTLENECK

/* ---------- Local Global Variables ---------- */
static Point        pM, pN, pP, pQ;
static Transform    trans;

#ifdef NO_THIN_BOTTLENECK
static int  minSpacing;
static int  topOfM, topOfN, topOfQ;
#endif NO_THIN_BOTTLENECK


/*
 * --------------------------------------------------------------------------
 *
 * Npl_TransOfChildren() --
 *
 *    Computes the absolute transformation of nd_children of node in the 
 *    coordinate system of the chip.  Computation is performed for each 
 *    floorplan template so as to eliminate overlaps between nd_children
 *    which may be non-leaf or leaf nodes themselves.  If no routing area
 *    estimation is attempted, nodePtr->nd_spacing[][] elements will be all
 *    zero.  The nodePtr->nd_nodeTrans and the childNodePtr->nd_dim must have
 *    been set previously.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    Assigns absolute transformations to nd_children of node.
 *
 * --------------------------------------------------------------------------
 */

BearStatus
Npl_TransOfChildren(nodePtr)
Ncl_Node *nodePtr;
{
    Transform   firstTrans, secondTrans, thirdTrans, finalTrans;
    Ncl_Node    *childPtr;
    Point       actualDim;
    int         i, doSwap = nodePtr->nd_orient%2;

#ifdef NO_THIN_BOTTLENECK
    minSpacing = 2 * ncl_WireWidth;
#endif NO_THIN_BOTTLENECK

    if (doSwap == 1) {
        NplSwapDimOfNode(nodePtr);
        NplSwapDimOfChildren(nodePtr);
    }

    (*(nplTransOfChildrenFuncArray[nodePtr->nd_pattern]))(nodePtr);

    /*
     *    Given the relative transformation of nd_children of the node with
     *    respect to the node, computes the absolute transformation of the
     *    nd_children taking into account the orientation on the floorplan
     *    template attached to the node.  
     */
    for (i = 0; i < nodePtr->nd_numOfChildren ;i++ ) {
        childPtr = nodePtr->nd_children[i];

        firstTrans = childPtr->nd_nodeTrans;
        childPtr->nd_nodeTrans.t_c = childPtr->nd_nodeTrans.t_f = 0;
        GeoTransPoint(&childPtr->nd_nodeTrans, &childPtr->nd_dim, 
                &actualDim);

        switch (nodePtr->nd_orient) {
        case 0:
            secondTrans = GeoIdentityTransform;
             break;
        case 1:
             secondTrans = Geo90Transform;
             secondTrans.t_c = 0;
             secondTrans.t_f = nodePtr->nd_dim.p_x - actualDim.p_x;
             break;
        case 2:
            secondTrans = Geo180Transform;
             secondTrans.t_c = nodePtr->nd_dim.p_x - actualDim.p_x;
             secondTrans.t_f = nodePtr->nd_dim.p_y - actualDim.p_y;
             break;
        case 3:
            secondTrans = Geo270Transform;
             secondTrans.t_c =  nodePtr->nd_dim.p_y - actualDim.p_y;
             secondTrans.t_f = 0;
             break;
        default:
    Bear_PrintErr("Npl_TransOfChildren() failed: unexpected orientation on node.");
             return (BEAR_ERROR);
        }

        GEOTRANSTRANS(&firstTrans, &secondTrans, &thirdTrans);
        GEOTRANSTRANS(&thirdTrans, &nodePtr->nd_nodeTrans, &finalTrans);
        childPtr->nd_nodeTrans.t_c = finalTrans.t_c;
        childPtr->nd_nodeTrans.t_f = finalTrans.t_f;
    }

    if (doSwap == 1) {
        NplSwapDimOfChildren(nodePtr);
        NplSwapDimOfNode(nodePtr);
    }

    return (BEAR_OK);
}



void
Npl_TransOfChildrenFunc0(nodePtr)
register Ncl_Node *nodePtr;
{
    trans = nodePtr->nd_children[0]->nd_nodeTrans;
    GeoTranslateTrans(&trans,
       nodePtr->nd_spacing[0][WEST], nodePtr->nd_spacing[0][SOUTH],
        &nodePtr->nd_children[0]->nd_nodeTrans);
}
 


void
Npl_TransOfChildrenFunc1(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    Ncl_Node   *childMPtr = nodePtr->nd_children[m];
    Ncl_Node   *childNPtr = nodePtr->nd_children[n];
    
    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][1] + nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[1][SOUTH],
             &childNPtr->nd_nodeTrans);
}
 

   
void
Npl_TransOfChildrenFunc2(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    Ncl_Node   *childMPtr = nodePtr->nd_children[m];
    Ncl_Node   *childNPtr = nodePtr->nd_children[n];
    Ncl_Node   *childPPtr = nodePtr->nd_children[p];
 
    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    pM.p_x += nodePtr->nd_spacing[0][WEST];
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][1],
            nodePtr->nd_spacing[1][SOUTH],
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
         pM.p_x + nodePtr->nd_spacing[0][2],
         pN.p_y + nodePtr->nd_spacing[1][2] + nodePtr->nd_spacing[1][SOUTH],
         &childPPtr->nd_nodeTrans);
 
#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childPPtr->nd_nodeTrans.t_f &&
            topOfM > childPPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}



void
Npl_TransOfChildrenFunc3(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    pM.p_x += nodePtr->nd_spacing[0][WEST] + nodePtr->nd_spacing[0][1];
    pN.p_x += nodePtr->nd_spacing[1][2];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
             nodePtr->nd_spacing[0][SOUTH],
             &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x, 
            nodePtr->nd_spacing[1][SOUTH],
             &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + pN.p_x, 
            nodePtr->nd_spacing[2][SOUTH],
             &childPPtr->nd_nodeTrans);
}



void
Npl_TransOfChildrenFunc4(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    int x1, x2, extraSpacing;
    Ncl_Node    *childMPtr = nodePtr->nd_children[m]; /* Blk assigned to rm 0 */
    Ncl_Node    *childNPtr = nodePtr->nd_children[n]; /* Blk assigned to rm 1 */
    Ncl_Node    *childPPtr = nodePtr->nd_children[p]; /* Blk assigned to rm 2 */
    Ncl_Node    *childQPtr = nodePtr->nd_children[q]; /* Blk assigned to rm 3 */

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    pM.p_x += nodePtr->nd_spacing[0][WEST];
    pP.p_x += nodePtr->nd_spacing[2][WEST];
    pM.p_y += nodePtr->nd_spacing[0][2] + nodePtr->nd_spacing[0][SOUTH];
    pN.p_y += nodePtr->nd_spacing[1][3] + nodePtr->nd_spacing[1][SOUTH];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[2][WEST], 
            pM.p_y,
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    x1 = pM.p_x + nodePtr->nd_spacing[0][1];
    x2 = pP.p_x + nodePtr->nd_spacing[2][3];

#ifdef NO_THIN_BOTTLENECK
    extraSpacing = minSpacing;
#else
    extraSpacing = 0;
#endif NO_THIN_BOTTLENECK

    if (pM.p_y - extraSpacing < pN.p_y - nodePtr->nd_spacing[1][3]) {
       x1 = NPLMAX(pM.p_x + nodePtr->nd_spacing[0][1], 
               pP.p_x + nodePtr->nd_spacing[1][2]);
    }
    else if (pM.p_y - nodePtr->nd_spacing[0][2] > pN.p_y - extraSpacing) {
       x2 = NPLMAX(pP.p_x + nodePtr->nd_spacing[2][3], 
               pM.p_x + nodePtr->nd_spacing[0][3]);
    }
    GeoTranslateTrans(&trans, 
            x1, 
            nodePtr->nd_spacing[1][SOUTH],
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            x2, 
            pN.p_y,
            &childQPtr->nd_nodeTrans);

}



void
Npl_TransOfChildrenFunc5(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    int x, y;
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];
    Ncl_Node    *childQPtr = nodePtr->nd_children[q];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    pM.p_x += nodePtr->nd_spacing[0][WEST];
    pN.p_x += nodePtr->nd_spacing[1][2];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH], 
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][1],
            nodePtr->nd_spacing[1][SOUTH], 
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    x = pM.p_x + nodePtr->nd_spacing[0][1] + pN.p_x;
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[2][SOUTH],
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    y = NPLMAX(pN.p_y + nodePtr->nd_spacing[1][3] + nodePtr->nd_spacing[1][SOUTH],
            pP.p_y + nodePtr->nd_spacing[2][3] + nodePtr->nd_spacing[2][SOUTH]);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][3], 
            y,
            &childQPtr->nd_nodeTrans);

#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childQPtr->nd_nodeTrans.t_f &&
            topOfM > childQPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}

 

void
Npl_TransOfChildrenFunc6(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    int x;
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];
    Ncl_Node    *childQPtr = nodePtr->nd_children[q];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    pM.p_x += nodePtr->nd_spacing[0][WEST];
    pP.p_x += nodePtr->nd_spacing[2][3];
    pN.p_y += nodePtr->nd_spacing[1][SOUTH];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][1],
            nodePtr->nd_spacing[1][SOUTH], 
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][2],
            pN.p_y + nodePtr->nd_spacing[1][2], 
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    x = pM.p_x + nodePtr->nd_spacing[0][2] + pP.p_x;
    GeoTranslateTrans(&trans, 
            x, 
            pN.p_y + nodePtr->nd_spacing[1][3],
            &childQPtr->nd_nodeTrans);

#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childPPtr->nd_nodeTrans.t_f &&
            topOfM > childPPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}
 

void
Npl_TransOfChildrenFunc7(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    int x;
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];
    Ncl_Node    *childQPtr = nodePtr->nd_children[q];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    GeoTransPoint(&childQPtr->nd_nodeTrans, &childQPtr->nd_dim, &pQ);
    pM.p_x += nodePtr->nd_spacing[0][WEST];
    pN.p_y += nodePtr->nd_spacing[1][2] + nodePtr->nd_spacing[1][SOUTH];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][1],
            nodePtr->nd_spacing[1][SOUTH],
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][2],
            pN.p_y, 
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    x = pM.p_x + 
       NPLMAX(pN.p_x + nodePtr->nd_spacing[0][1] + nodePtr->nd_spacing[1][3],
            pP.p_x + nodePtr->nd_spacing[0][2] + nodePtr->nd_spacing[2][3]);
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[3][SOUTH],
            &childQPtr->nd_nodeTrans);

#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childPPtr->nd_nodeTrans.t_f &&
            topOfM > childPPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
    topOfQ = childQPtr->nd_nodeTrans.t_f + pQ.p_y;
    if (topOfQ <= childPPtr->nd_nodeTrans.t_f &&
            topOfQ > childPPtr->nd_nodeTrans.t_f - minSpacing) {
       childQPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}
 


void
Npl_TransOfChildrenFunc8(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    int x;
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];
    Ncl_Node    *childQPtr = nodePtr->nd_children[q];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    pM.p_x += nodePtr->nd_spacing[0][1] + nodePtr->nd_spacing[0][WEST];
    pP.p_y += nodePtr->nd_spacing[2][3] + nodePtr->nd_spacing[2][SOUTH];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    x = pM.p_x;
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[1][SOUTH],
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    x += pN.p_x + nodePtr->nd_spacing[1][2];
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[2][SOUTH],
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    x += nodePtr->nd_spacing[1][3] - nodePtr->nd_spacing[1][2];
    GeoTranslateTrans(&trans, 
            x, 
            pP.p_y, 
            &childQPtr->nd_nodeTrans);

#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childQPtr->nd_nodeTrans.t_f &&
            topOfM > childQPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
    topOfN = childNPtr->nd_nodeTrans.t_f + pN.p_y;
    if (topOfN <= childQPtr->nd_nodeTrans.t_f &&
            topOfN > childQPtr->nd_nodeTrans.t_f - minSpacing) {
       childNPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}
 


void
Npl_TransOfChildrenFunc9(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];
    Ncl_Node    *childQPtr = nodePtr->nd_children[q];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    pM.p_x += nodePtr->nd_spacing[0][WEST];
    pN.p_y += nodePtr->nd_spacing[1][2] + nodePtr->nd_spacing[1][SOUTH];
    pP.p_y += nodePtr->nd_spacing[2][3];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH], 
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][1],
            nodePtr->nd_spacing[1][SOUTH], 
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][2],
            pN.p_y, 
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            pM.p_x + nodePtr->nd_spacing[0][3],
            pN.p_y + pP.p_y, 
            &childQPtr->nd_nodeTrans);

#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childPPtr->nd_nodeTrans.t_f &&
            topOfM > childPPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childQPtr->nd_nodeTrans.t_f &&
            topOfM > childQPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}
 


void
Npl_TransOfChildrenFunc10(nodePtr)
register Ncl_Node *nodePtr;
{
    int m = nodePtr->nd_label[0];
    int n = nodePtr->nd_label[1];
    int p = nodePtr->nd_label[2];
    int q = nodePtr->nd_label[3];
    int x;
    Ncl_Node    *childMPtr = nodePtr->nd_children[m];
    Ncl_Node    *childNPtr = nodePtr->nd_children[n];
    Ncl_Node    *childPPtr = nodePtr->nd_children[p];
    Ncl_Node    *childQPtr = nodePtr->nd_children[q];

    GeoTransPoint(&childMPtr->nd_nodeTrans, &childMPtr->nd_dim, &pM);
    GeoTransPoint(&childNPtr->nd_nodeTrans, &childNPtr->nd_dim, &pN);
    GeoTransPoint(&childPPtr->nd_nodeTrans, &childPPtr->nd_dim, &pP);
    pM.p_x += nodePtr->nd_spacing[0][WEST] + nodePtr->nd_spacing[0][1];
    pN.p_x += nodePtr->nd_spacing[1][2];
    pP.p_x += nodePtr->nd_spacing[2][3];
    Mac_CopyTransform(&trans, &childMPtr->nd_nodeTrans);
    GeoTranslateTrans(&trans, 
            nodePtr->nd_spacing[0][WEST],
            nodePtr->nd_spacing[0][SOUTH],
            &childMPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childNPtr->nd_nodeTrans);
    x = pM.p_x;
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[1][SOUTH],
            &childNPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childPPtr->nd_nodeTrans);
    x += pN.p_x;
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[2][SOUTH],
            &childPPtr->nd_nodeTrans);
    Mac_CopyTransform(&trans, &childQPtr->nd_nodeTrans);
    x += pP.p_x;
    GeoTranslateTrans(&trans, 
            x, 
            nodePtr->nd_spacing[3][SOUTH],
            &childQPtr->nd_nodeTrans);

#ifdef NO_THIN_BOTTLENECK
    topOfM = childMPtr->nd_nodeTrans.t_f + pM.p_y;
    if (topOfM <= childPPtr->nd_nodeTrans.t_f &&
            topOfM > childPPtr->nd_nodeTrans.t_f - minSpacing) {
       childMPtr->nd_nodeTrans.t_f += minSpacing;
    }
#endif NO_THIN_BOTTLENECK

}

