
/* nplMain.c --
 *
 *     Massoud Pedram
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplMain.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "rectSlice.h"
#include "globalRouter.h"
#include "grEditor.h"
#include "list.h"
#include "cpaInterface.h"
#include "newP.h"
#include "newPInt.h"
#include "nplDraw.h"


static bool mytest = TRUE;

/* ---------- Internal global variables for the newP module ---------- */
lsList              nplCoreNodeList, nplTotalNodeList;
lsList              nplNetList, nplBdrPinList[4];
CellMaster          *nplCoreMasterPtr;
Point               nplDelta;
Transform           nplTrans[4];
int                 plDimension;
Rect                nplInsideOutRect;
bool                nplSaveDataFlag;
bool                nplUseMSTCalcFlag;
int                 nplMaxPinsForMST;

La_CostEntry        *nplCostMatrix, *nplCostOfNet;
int                 nplNumOfCMRows, nplNumOfCMCols;
Npl_PinData         **nplPinData;
Ncl_SlotData        **nplSlot;
double              nplAdjustSlots;
double              nplBorderSize;

bool                nplFinalOptFlag;
bool                nplFullEnumFlag;
double              nplTolerance;
double              nplAreaWeight;
double              nplWireWeight;
double              nplCritWeight;

int                 nplMaxLoopCnt;
char                *nplDefFileNamePtr;
char                *nplTmpDirPtr;

bool                nplArrPinsFlag;
bool                nplInteractiveFlag;
bool                nplCreateWinFlag;


/* ---------- Local Procedures ---------- */
static BearStatus   EndPlacer();
static void         FreeNetData();
static void         QueueSort();

static BearStatus   ChooseTopology();
static BearStatus   ChooseLeafTopology();

static BearStatus   DoPlacement();
static BearStatus   GenPinListsForNetsOfNode();

static BearStatus   AddNodeToList();
static BearStatus   StoreNodeData();

static int          CompArea();
static void         CalcCriticalityOfNets();

static void         BackUpPinListForNets();
static void         RestorePinListsForNetsOfNode();
static void         MergePinLists();



static BearStatus
DoPlacement(nodePtr)
Ncl_Node *nodePtr;
{
    Ncl_Node    *childNodePtr;
    double      levelAdj, t1;
    int         i;

    t1 = NplPtime();

    (void) fprintf (bearout, "\nFloorplanning Node %2d --\n", nodePtr->nd_id);
    (void) fflush (bearout);

    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        nodePtr->nd_children[i]->nd_status = ACTIVE;
    }

/*
 *  (void) printf("Global netlist before local picture.\n");
 *  NplCheckNetList(nplNetList, FALSE);
 */
    BackUpPinListForNets(nodePtr->nd_totalNetList);

    if (GenPinListsForNetsOfNode(nodePtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

/*
 *  (void) fprintf(bearout, "local net list for node %d -- \n", nodePtr->nd_id);
 *  NplCheckNetList(nodePtr->nd_totalNetList, FALSE);
 */

    if (NplCalcIOConnectivities(nodePtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (ncl_TimingFlag) {
        if (NplCalcCritNetIOConnectivities(nodePtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }
    }

    if (nodePtr->nd_type == P_LEAF) {
        if (NplCalcLeafConnectivities(nodePtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }
    }

    RestorePinListsForNetsOfNode(nodePtr);
/*
 *  (void) fprintf(bearout,
 *      "global net list while processing node %d -- \n", nodePtr->nd_id);
 *  NplCheckNetList(nplNetList, FALSE);
 */
    if (nplFullEnumFlag == TRUE) {
        if (NplEnumTPs(nodePtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }
    }
    else {
        if (mytest) {
            if (NplEnumLabels(nodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        } else {
            if (NplUsePartialEnumTPs(nodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
    }

    (void) fprintf (bearout,
        "CPU-time (Placement) = %g sec\n", (NplPtime() - t1));
    (void) fflush (bearout);

    if (nplFinalOptFlag == TRUE) {
        if (nodePtr->nd_type == P_LEAF) {
            if (NplFinalOpt(nodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
    }

    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];

        childNodePtr->nd_status = PLACED;
        childNodePtr->nd_area = ((double) childNodePtr->nd_dim.p_x) *
            ((double) childNodePtr->nd_dim.p_y);
    }

    return (BEAR_OK);
}



static int
CompArea(node1, node2)
Ncl_Node **node1, **node2;
{
    static bool sort_decreasing = TRUE;

    if (sort_decreasing) {
        if ((*node1)->nd_area < (*node2)->nd_area) return (1);
        else if ((*node1)->nd_area > (*node2)->nd_area) return (-1);
        else return (0);
    } else {
        if ((*node1)->nd_area < (*node2)->nd_area) return (-1);
        else if ((*node1)->nd_area > (*node2)->nd_area) return (1);
        else return (0);
    }
}


/*
 * ----------------------------------------------------------------------------
 *
 * QueueSort() --
 *
 *     Sorts the elements in the queue (nodes on one level of the hierarchy)
 *     according to their area. This sorting influences the IO goals of
 *     the nodes. Nodes later in the sorted queue can use the knowledge
 *     about the elements of the clusters earlier in the queue that are
 *     already placed.
 *
 * Results:
 *     None.
 * 
 * Side effects:
 *     The argument that is passed to the procedure is sorted.
 *
 * ----------------------------------------------------------------------------
 */

static void
QueueSort(q)
Deque *q;
{
    Ncl_Node *sortList[1000];
    int      sortElts = 0;

    while (! IsEmptyDeque(q)) {
        sortList[sortElts++] = (Ncl_Node *) PopDeque(q);
    }
    qsort((char *) sortList, sortElts, sizeof (Ncl_Node *), CompArea);
    while (sortElts > 0) {
        PushDeque(q, (ClientData) sortList[--sortElts]);
    }
}



/*
 * ----------------------------------------------------------------------------
 *
 * ChooseTopology() --
 *
 *     Does the initial phase of floorplanning by enumerating all possible
 *     templates, all assignments, and orientations of blocks (at leaf levels)
 *     so as to minimize the objective function.
 *
 * Results:
 *     BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 * 
 * Side effects:
 *     Creates an initial placement of modules by specifying transformations
 *     of the lower left corners of modules with respect to a fixed coordinate
 *     system.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
ChooseTopology(enumQueue, finishedQueue)
Deque *enumQueue, *finishedQueue;
{
    lsGen               netGen;
    Ncl_LinkedCritNet   *lnkCritNetPtr;
    GenericCell         gcell;
    Ncl_Node            *childNodePtr, *nodePtr;
    Npl_NetData         *netDataPtr;
    int                 i, oldDepth = 0;
    int                 minPinSpacing;

    minPinSpacing = ncl_WireWidth;
    while (! NPLISPLEAF(nodePtr = (Ncl_Node *) LookUpHead(enumQueue))) {

        if (oldDepth != nodePtr->nd_depth) {
            oldDepth = nodePtr->nd_depth;

            /*
             * We sort nodes in the enumQueue so that one with the
             * largest area is in front of queue (if sort_decreasing == TRUE). 
             * This is heuristics.
             */
            QueueSort(enumQueue);
        }

        nodePtr = (Ncl_Node *) PopDeque(enumQueue);
        for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
            InjectDeque(enumQueue, (ClientData)(nodePtr->nd_children[i]));
        }

        nplCoreNodeList = lsCreate();
        ItrDeque(enumQueue, AddNodeToList, (ClientData) nplCoreNodeList);

        PushDeque(finishedQueue, (ClientData) nodePtr);
        ItrDeque(finishedQueue, StoreNodeData, (ClientData) NULL);

        if (DoPlacement(nodePtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (ncl_TimingFlag) {
            CalcCriticalityOfNets(nodePtr->nd_lnkCritNetList);
        }

        if (NplDoSpacing(finishedQueue) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (NplDoPinAssignment(nodePtr, minPinSpacing) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (ncl_TimingFlag == TRUE) {
            Ncl_LinkedCritNetInNode  *critInNodePtr;

            if (NplCalcCNInterLeafL(nodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
            for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                NPLNEXT(critInNodePtr)) {

                critInNodePtr->critNetPtr->used_l = critInNodePtr->inter_l;
                (void) fprintf (bearout, 
                    "  Length of critical net %s in node %d = %f\n",
                    critInNodePtr->critNetPtr->netPtr->name, 
                    nodePtr->nd_id, 
                    critInNodePtr->critNetPtr->used_l);
            }

        }

/*
 *      ItrDeque(finishedQueue, StoreNodeData, (ClientData) NULL);
 *
 *      if (NplDoSpacing(finishedQueue) != BEAR_OK) {
 *          return (BEAR_ERROR);
 *      }
 *      if (NplSetUpCore(nplCoreNodeList, FALSE) != BEAR_OK) {
 *          return (BEAR_ERROR);
 *      }
 */

        lsDestroy(nplCoreNodeList, (void (*)()) NULL);
        nplCoreNodeList = (lsList) NULL;

#ifndef BEAR_BLACK_BOX
        if (bear_BlackBoxFlag == FALSE) {
            NplInitChipWinView(globalWinDataPtr, nplCoreMasterPtr);
            NplDrawChipWin();
        }
#endif BEAR_BLACK_BOX
    }

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * ChooseLeafTopology() --
 *
 *     Does the initial phase of floorplanning by enumerating all possible
 *     templates, all assignments, and orientations of blocks (at leaf levels)
 *     so as to minimize the objective function.
 *
 * Results:
 *     BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 * 
 * Side effects:
 *     Creates an initial placement of modules by specifying transformations
 *     of the lower left corners of modules with respect to a fixed coordinate
 *     system.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
ChooseLeafTopology(enumQueue, finishedQueue)
Deque *enumQueue, *finishedQueue;
{
    lsGen               netGen;
    Ncl_LinkedCritNet    *lnkCritNetPtr;
    Ncl_Node            *nodePtr, *childNodePtr;
    Npl_NetData         *netDataPtr;
    int                 i, j;
    int                 minPinSpacing;

    minPinSpacing = ncl_WireWidth;
    while (! NPLISLEAF(nodePtr = (Ncl_Node *) LookUpHead(enumQueue))) {

        nodePtr = (Ncl_Node *) PopDeque(enumQueue);
        for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
            InjectDeque(enumQueue, (ClientData) (nodePtr->nd_children[i]));
        }

        nplCoreNodeList = lsCreate();
        ItrDeque(enumQueue, AddNodeToList, (ClientData) nplCoreNodeList);

        PushDeque(finishedQueue, (ClientData) nodePtr);
        ItrDeque(finishedQueue, StoreNodeData, (ClientData) NULL);
   
        if (DoPlacement(nodePtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (NplDoSpacing(finishedQueue) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (NplDoPinAssignment(nodePtr, minPinSpacing) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (ncl_TimingFlag == TRUE) {
            Ncl_LinkedCritNetInNode  *critInNodePtr;

            if (NplCalcCNInterLeafL(nodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
            for (critInNodePtr = nodePtr->nd_lnkCritNetList;
                critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
                NPLNEXT(critInNodePtr)) {

                critInNodePtr->critNetPtr->used_l = critInNodePtr->inter_l;
                (void) fprintf (bearout, 
                    "  Length of critical net %s in node %d = %f\n",
                    critInNodePtr->critNetPtr->netPtr->name, 
                    nodePtr->nd_id, 
                    critInNodePtr->critNetPtr->used_l);
            }

        }

/*
 *      ItrDeque(finishedQueue, StoreNodeData, (ClientData) NULL);
 *
 *      if (NplDoSpacing(finishedQueue) != BEAR_OK) {
 *          return (BEAR_ERROR);
 *      }
 *      if (NplSetUpCore(nplCoreNodeList, FALSE) != BEAR_OK) {
 *          return (BEAR_ERROR);
 *      }
 */

        lsDestroy(nplCoreNodeList, (void (*)()) NULL);
        nplCoreNodeList = (lsList) NULL;

#ifndef BEAR_BLACK_BOX
        if (bear_BlackBoxFlag == FALSE) {
            NplInitChipWinView(globalWinDataPtr, nplCoreMasterPtr);
            NplDrawChipWin();
        }
#endif BEAR_BLACK_BOX
    }

    return (BEAR_OK);
}



/* ARGSUSED */
static BearStatus
StoreNodeData(nodePtr, clientData)
Ncl_Node    *nodePtr;
ClientData  clientData;
{
    Ncl_Node    *childNodePtr;
    int         i;

    if (nodePtr == ncl_ClusterTree) {
        nodePtr->nd_storeDim = nodePtr->nd_dim;
        nodePtr->nd_storeTrans = nodePtr->nd_nodeTrans;
    }

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (childNodePtr->nd_status == PLACED) {
            childNodePtr->nd_storeDim = childNodePtr->nd_dim;
            childNodePtr->nd_storeTrans = childNodePtr->nd_nodeTrans;
        }
    }

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddNodeToList(nodePtr, nodeList)
Ncl_Node    *nodePtr;
lsList      nodeList;
{
    lsNewEnd(nodeList, (lsGeneric) nodePtr, LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static void
FreeNetData(item)
lsGeneric item;
{
    Npl_NetData *netDataPtr = (Npl_NetData *) item;

    netDataPtr->netPtr->clientData = (ClientData) NULL;

    FREE((char *) netDataPtr);
}


static BearStatus
EndPlacer()
/*      
 * In the future all of the MAC data such as pin inst, pin mast, port inst
 * port mast, cell inst, and cell mast etc. that were created should be freed.
 */     
{
    lsGen       gen, pinGen;
    Net         *netPtr;
    Ncl_Node    *nodePtr;
    Npl_PinData *pinDataPtr;
    int         s;

    NclForeachSide(s) {
        if (nplBdrPinList[s] != (lsList) NULL) {
            lsDestroy(nplBdrPinList[s], NplFreePinData);
            nplBdrPinList[s] = (lsList) NULL;
        }
    }

    if (nplCoreNodeList != (lsList) NULL) {
        lsDestroy(nplCoreNodeList, (void (*)()) NULL);
        nplCoreNodeList = (lsList) NULL;
    }

    if (nplTotalNodeList != (lsList) NULL) {
        lsForeachItem(nplTotalNodeList, gen, nodePtr) {
            NclForeachSide(s) {
                if (nodePtr->nd_initPinList[s] != (lsList) NULL) {
                    lsDestroy(nodePtr->nd_initPinList[s], NplFreePinData);
                    nodePtr->nd_initPinList[s] = (lsList) NULL;
                }
                nodePtr->nd_tilePinList[s] = (lsList) NULL;
            }

            if (nodePtr->nd_totalNetList != (lsList) NULL) {
                lsDestroy(nodePtr->nd_totalNetList, (void (*)()) NULL);
            }
            if (nodePtr->nd_bdrNetList != (lsList) NULL) {
                lsDestroy(nodePtr->nd_bdrNetList, (void (*)()) NULL);
            }

            if (nodePtr->nd_type == LEAF) {
                nodePtr->nd_cellInstPtr->clientData = (ClientData) NULL;
            }
        }
        lsDestroy(nplTotalNodeList, (void (*)()) NULL);
        nplTotalNodeList = (lsList) NULL;
    }

    if (nplNetList != (lsList) NULL) {
        lsDestroy(nplNetList, FreeNetData);
        nplNetList = (lsList) NULL;
    }

    if (nplDefFileNamePtr != NIL(char)) {
        FREE(nplDefFileNamePtr);
        nplDefFileNamePtr = NIL(char);
    }
    if (nplTmpDirPtr != NIL(char)) {
        FREE(nplTmpDirPtr);
        nplTmpDirPtr = NIL(char);
    }

    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Npl_Placement() --
 *
 *   Top level procedure for the placement.
 *   Note the following:
 *      void        PushDeque = DQPushFront();
 *      ClientData  PopDeque() = DQPopFront();
 *      void        InjectDeque = DQPushRear();
 *      ClientData  EjectDeque() = DQPopRear();
 *
 * Results:
 *   BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   The placement information is stored in the ncl_ClusterTree data structure.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Npl_Placement(aspectR, coreMasterPtr, parmsPtr)
double          aspectR;
CellMaster      *coreMasterPtr;
Npl_UserParms   *parmsPtr;          /* IN */
{
    Deque   *enumQueue;         /* FIFO queue for tree enumeration. */
    Deque   *finishedQueue;     /* FIFO queue for placed nodes.     */
    bool    pinSpacingFlag;

    if (NplBeginPlacerInt(aspectR, coreMasterPtr, parmsPtr) != BEAR_OK) {
      Bear_PrintErr("Npl_Placement() failed : could not initialize properly.");
#ifndef BEAR_BLACK_BOX
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
#endif BEAR_BLACK_BOX
        return (BEAR_ERROR);
    }

    enumQueue = MakeDeque();
    finishedQueue = MakeDeque();

    PushDeque(enumQueue, (ClientData) ncl_ClusterTree);

    if (ChooseTopology(enumQueue, finishedQueue) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    QueueSort(enumQueue);

    if (ChooseLeafTopology(enumQueue, finishedQueue) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (Mac_CheckPins(nplCoreMasterPtr, &pinSpacingFlag) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (pinSpacingFlag == FALSE) {
        Bear_PrintWarn("Some pins are too close.");
    }

    NplPrintResults();

    if (ncl_TimingFlag == TRUE) {
        NplCheckCritNetConstraints();
    }

#ifdef DEBUG
    NplSaveCellMaster(nplCoreMasterPtr);
#endif DEBUG

    if (bear_BlackBoxFlag == FALSE) {
        (void) fprintf (bearout, "\nPlacement completed.\n\n");
        (void) fflush (bearout);
    }

#ifdef BETTY_BOOP
/*
 * (gyorgy 8 jan 90) taken out until new global router is capable
 *                   of putting pseudo pins in.
 */

    if (nplArrPinsFlag == TRUE) {
        Cpa_ArrPinData optData; /* optional data */
        
        optData.cpaProgPtr = NIL(char);
        optData.tmpDirPtr = nplTmpDirPtr;
        optData.saveFlag = nplSaveDataFlag;
        optData.warnFlag = TRUE;
        if (Cpa_ArrPinsForAllBnTiles(nplCoreMasterPtr, &optData) ==
            BEAR_ERROR) {
            Bear_PrintErr("Unable to do pin arrangement for all the bn tiles.");
            return (BEAR_ERROR);
        }
  
#ifndef BEAR_BLACK_BOX
        if (bear_BlackBoxFlag == FALSE) {
            NplInitChipWinView(globalWinDataPtr, nplCoreMasterPtr);
            NplDrawChipWin();
        }
#endif BEAR_BLACK_BOX

#ifdef DEBUG
        NplSaveCellMaster(nplCoreMasterPtr);
#endif DEBUG
    }
#endif BETTY_BOOP

    DeleteDeque(finishedQueue);
    DeleteDeque(enumQueue);

    (void) EndPlacer();

    return (BEAR_OK);
}



/*
 * This function generates the local pl net list for the node, i.e., the 
 * nd_totalNetList.  It uses the initPinList on the node and either the 
 * initPinList (for leaf children) or the bdrNetList (for p-leaf and non-leaf 
 * children). In addition, it sets nplNumOfCMRows and nplPinData.
 */
static BearStatus
GenPinListsForNetsOfNode(nodePtr)
Ncl_Node *nodePtr;
{
    lsGen       netGen, pinGen;
    Ncl_SideDir initDir;
    Ncl_Node    *childNodePtr;
    Npl_PinData *pinDataPtr;
    Npl_NetData *netDataPtr;
    int         i, row, s;

    lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
        netDataPtr->pinList = lsCreate();
    }

    NclForeachSide(s) {
        initDir = Ncl_GetOppositeSideDir(s);
        lsForeachItem(nodePtr->nd_initPinList[s], pinGen, pinDataPtr) {
            pinDataPtr->pinKind = NPL_BORDER_PIN;
            pinDataPtr->pinFlexType = FIXED;
            pinDataPtr->initDir = initDir;

            lsNewEnd(pinDataPtr->netDataPtr->pinList,
                (lsGeneric) pinDataPtr, LS_NH);
        }
    }
    if (NplCalcTileDataOfNode(nodePtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    /*
     * Note that the children's pins do not have a position or a 
     * direction at this point.
     */
    nplNumOfCMRows = 0;
    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (childNodePtr->nd_type == LEAF) {
            NclForeachSide(s) {
                lsForeachItem(childNodePtr->nd_initPinList[s], pinGen,
                    pinDataPtr) {

                    lsNewEnd(pinDataPtr->netDataPtr->pinList,
                        (lsGeneric) pinDataPtr, LS_NH);
                    if (pinDataPtr->pinFlexType != FIXED) {
                        nplNumOfCMRows++;
                    }
                }
            }
        }
        else {
            lsForeachItem(childNodePtr->nd_bdrNetList, netGen, netDataPtr) {
                pinDataPtr = NplMakePinData();
                (void) sprintf (pinDataPtr->pinName, "%d", Mac_GenerateId());
                pinDataPtr->pinKind = NPL_INSIDE_PIN;
                pinDataPtr->pinFlexType = FLEXIBLE;
                pinDataPtr->netDataPtr = netDataPtr;
                pinDataPtr->nodePtr = childNodePtr;

                lsNewEnd(netDataPtr->pinList, (lsGeneric) pinDataPtr, LS_NH);
                nplNumOfCMRows++;
            }
        }
    }

    MALLOC(Npl_PinData **, nplPinData, nplNumOfCMRows * sizeof (Npl_PinData *));
    row = 0;
    lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
        lsForeachItem(netDataPtr->pinList, pinGen, pinDataPtr) {
            if (pinDataPtr->pinKind == NPL_INSIDE_PIN) {
                if (pinDataPtr->pinFlexType != FIXED) {
                    pinDataPtr->row = row;
                    nplPinData[row++] = pinDataPtr;
                }
            }
        }
    }

    return (BEAR_OK);
}



static void
BackUpPinListForNets(netList)
lsList netList;
{
    lsGen       netGen;
    Npl_NetData *netDataPtr;

    lsForeachItem(netList, netGen, netDataPtr) {
        netDataPtr->bakPinList = netDataPtr->pinList;
    }
}


static void
RestorePinListsForNetsOfNode(nodePtr)
Ncl_Node *nodePtr;
{
    lsGen       pinGen;
    lsList      bakPinList;
    Npl_PinData *pinDataPtr;
    int         s;

    NclForeachSide(s) {
        lsForeachItem(nodePtr->nd_initPinList[s], pinGen, pinDataPtr) {
            bakPinList = pinDataPtr->netDataPtr->bakPinList;
            if (bakPinList != (lsList) NULL) {
                MergePinLists(pinDataPtr->netDataPtr->pinList, bakPinList);
                pinDataPtr->netDataPtr->bakPinList = (lsList) NULL;
            }
        }
    }
}



static void
MergePinLists(target, source)
lsList target, source;
{
    lsGen       pinTGen, pinSGen;
    Npl_PinData	*pinTDataPtr, *pinSDataPtr;

    lsForeachItem(target, pinTGen, pinTDataPtr) {
        lsForeachItem(source, pinSGen, pinSDataPtr) {
            if (pinTDataPtr->pinPtr == pinSDataPtr->pinPtr) {
                lsDelBefore(pinSGen, (lsGeneric *) NULL);
                lsDelBefore(pinTGen, (lsGeneric *) NULL);
                lsFinish(pinSGen);
                break;
            }
        }
    }
    lsForeachItem(source, pinSGen, pinSDataPtr) {
        lsNewEnd(target, (lsGeneric) pinSDataPtr, LS_NH);
    }
    lsDestroy(source, (void (*)()) NULL);
}



static void
CalcCriticalityOfNets(critNetList)
Ncl_LinkedCritNetInNode *critNetList;
{
    Npl_NetData             *netDataPtr;
    Ncl_LinkedCritNetInNode *critInNodePtr;
    float                   maxLength, usedLength, ratio;

    for (critInNodePtr = critNetList;
        critInNodePtr != NIL(Ncl_LinkedCritNetInNode); 
        NPLNEXT(critInNodePtr)) {

        netDataPtr = (Npl_NetData *) 
            critInNodePtr->critNetPtr->netPtr->clientData;
        maxLength = (float) critInNodePtr->critNetPtr->netPtr->maxLength;
        usedLength = (float) critInNodePtr->best_inter_l;
        ratio = usedLength / maxLength;
        if (ratio <= 1.0) {
            netDataPtr->criticality = ratio;
        }
        else {
            netDataPtr->criticality = ratio*ratio;
        }
    }
}


