
/* nplPinAssign.c --
 *
 *     Massoud Pedram / George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplPinAssign.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "list.h"
#include "newP.h"
#include "newG.h"
#include "newPInt.h"
#include "nplWriteToMac.h"
#include "nplDraw.h"


typedef struct _BNSearchData BNSearchData;
struct _BNSearchData {
    int         delta;
    PlaneData   *planeDataPtr;
    Rect        *horzPlaneBndPtr;
    Ncl_Node    *nodePtr;
    Ncl_SideDir  searchDir;
};

typedef struct _HandleData HandleData;
struct _HandleData {
    Ngr_BldTreeDataPtr  bldTreePtr;
    Ncl_Node            *nodePtr;
    PlaneData           *planeDataPtr;
};


/* ---------- Local Procedures ---------- */
static void         FreeSlotsForChildrenOfNode();
static BearStatus   GenSlotsForChildrenOfNode();
static BearStatus   SetUpSlotsBeforeLA();
static BearStatus   SetUpNodeAfterLA();
static lsStatus     HandleNet();
static BearStatus   FillSearchData();
static int          CalcSlotSpacingForBnTile();
static BearStatus   CalcSlotSpacingForBnLine();
static void         CalcJunctArea();
static void         PutCostOfSlotsWithinRect();
static void         CalcCostOfSlotsFromRect();
static void         CalcCostOfSlotsFromJunctList();
static BearStatus   StoreJunctData();
static BearStatus   RemoveJunctData();
static BearStatus   CalcSlotCostForPin();
static BearStatus   CalcSlotCostForNet();



int
NplCostMatrixIndex(row, col)
int row, col;
{
    int i;

    i = row * nplNumOfCMCols + col + 1;
    if (i > (nplNumOfCMCols*nplNumOfCMRows + 1)) {
        Bear_PrintFatalErr("BLOW OUT");
    }
    return (i);
}



static BearStatus
StoreJunctData(junctPtr, planeDataPtr)
WallJunction    *junctPtr;
PlaneData       *planeDataPtr;
{
    Npl_JunctData *junctDataPtr;

    MALLOC(Npl_JunctData *, junctDataPtr, sizeof (Npl_JunctData));

    junctPtr->nplModulePtr = (BearGeneric) junctDataPtr;

    CalcJunctArea(planeDataPtr, junctPtr, &junctDataPtr->junctArea);
        
    return (BEAR_CONTINUE_SEARCH);
}


/* ARGSUSED */
static BearStatus
RemoveJunctData(junctPtr, clientData)
WallJunction    *junctPtr;
ClientData      clientData;
{
    FREE((char *) junctPtr->nplModulePtr);

    junctPtr->nplModulePtr = (BearGeneric) NULL;

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
GenSlotsForChildrenOfNode(nodePtr, minPinSpacing)
Ncl_Node    *nodePtr;
int         minPinSpacing;
{
    static int      minNumOfSlotsPerSide = 4;

    BNSearchData    bnSearchData;
    CellMaster      *cellMasterPtr;
    lsGen           nodeGen;
    Ncl_SlotData    *slot;
    Ncl_Node        *childNodePtr, *tmpNodePtr;
    Rect            tileRect;
    double          adjustFactor;
    int             a, pinCount;
    int             i, j, k, x, y, s, delta, totNumOfPins, totNumOfSlots, 
                    numOfSlots[MAX_CHILDREN][4], maxNumOfSlots, lenOfSide;

    nplNumOfCMCols = 0;

    if (nodePtr->nd_type == NON_LEAF) {
        /* So that all pins can be placed on the same side, if necessary. */
        adjustFactor = 4.0 * nplAdjustSlots;
    }
    else {
        adjustFactor = nplAdjustSlots;
    }
    bnSearchData.planeDataPtr = Mac_GetCellPlaneOfCell(nplCoreMasterPtr);
    bnSearchData.horzPlaneBndPtr =
        Tp_GetHorzPlaneBnd(bnSearchData.planeDataPtr);

    lsForeachItem(nplTotalNodeList, nodeGen, tmpNodePtr) {
        tmpNodePtr->nd_hasSlotsFlag = FALSE;
    }

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == TRUE) {
            continue;
        }

        childNodePtr->nd_hasSlotsFlag = TRUE;

        if (! ncl_ChannelFreeFlag && ! ncl_QuickRouteEstFlag) {

            tileRect = childNodePtr->nd_tileRect;
            bnSearchData.nodePtr = childNodePtr;

            /* 
             * Calculate number of north slots.
             */
            bnSearchData.searchDir = NCL_NORTH_DIR;
            if (FillSearchData(&bnSearchData) != BEAR_OK) {
                return (BEAR_ERROR);
            }
            numOfSlots[i][NCL_NORTH_DIR] = adjustFactor *
                    (tileRect.r_xtop - tileRect.r_xbot) / bnSearchData.delta;

            /* 
             * Calculate number of east slots.
             */
            bnSearchData.searchDir = NCL_EAST_DIR;
            if (FillSearchData(&bnSearchData) != BEAR_OK) {
                return (BEAR_ERROR);
            }
            numOfSlots[i][NCL_EAST_DIR] = adjustFactor *
                    (tileRect.r_ytop - tileRect.r_ybot) / bnSearchData.delta;

            /* 
             * Calculate number of south slots.
             */
            bnSearchData.searchDir = NCL_SOUTH_DIR;
            if (FillSearchData(&bnSearchData) != BEAR_OK) {
                return (BEAR_ERROR);
            }
            numOfSlots[i][NCL_SOUTH_DIR] = adjustFactor *
                    (tileRect.r_xtop - tileRect.r_xbot) / bnSearchData.delta;

            /* 
             * Calculate number of west slots.
             */
            bnSearchData.searchDir = NCL_WEST_DIR;
            if (FillSearchData(&bnSearchData) != BEAR_OK) {
                return (BEAR_ERROR);
            }
            numOfSlots[i][NCL_WEST_DIR] = adjustFactor *
                    (tileRect.r_ytop - tileRect.r_ybot) / bnSearchData.delta;

        }
        else {
            numOfSlots[i][NCL_NORTH_DIR] = numOfSlots[i][NCL_EAST_DIR] = 
            numOfSlots[i][NCL_SOUTH_DIR] = numOfSlots[i][NCL_WEST_DIR] = 1;
        }

    }

    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == TRUE) {
            continue;
        }

        totNumOfSlots = 0;
        NclForeachSide(s) {
            totNumOfSlots += numOfSlots[i][s];
        }

        if (nodePtr->nd_type == NON_LEAF) {
            totNumOfPins = lsLength(childNodePtr->nd_bdrNetList);
        }
        else {
            totNumOfPins = 0;
            NclForeachSide(s) {
                totNumOfPins += lsLength(childNodePtr->nd_initPinList[s]);
            }
        }

        if (totNumOfSlots < totNumOfPins) {
            NclForeachSide(s) {
                if (! ncl_ChannelFreeFlag && ! ncl_QuickRouteEstFlag) {
                    numOfSlots[i][s] += (1 + (totNumOfPins - totNumOfSlots) / 4);
                }
                else {
                    /* So that half pins can be placed on the same side, if necessary. */
                    numOfSlots[i][s] += (totNumOfPins - totNumOfSlots) / 2;
                }
            }
        }
        else if (totNumOfSlots > 4 * totNumOfPins) {
            NclForeachSide(s) {
                numOfSlots[i][s] = totNumOfPins;
            }
        }
    }

    k = 0;
    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == TRUE) {
            continue;
        }

        tileRect = childNodePtr->nd_tileRect;
        /* 
         * Generate the north slots.
         */
        lenOfSide = tileRect.r_xtop - tileRect.r_xbot;
        maxNumOfSlots = lenOfSide / minPinSpacing;
        if (maxNumOfSlots == 0) maxNumOfSlots = 1;
        pinCount = childNodePtr->nd_sideFloPinCnt[NCL_NORTH_DIR] +
            ((childNodePtr->nd_freeFloPinCnt/4) + 1);
        a = NPLMAX(pinCount, numOfSlots[i][NCL_NORTH_DIR]);
        maxNumOfSlots = NPLMIN(maxNumOfSlots, a);
        delta = lenOfSide / maxNumOfSlots;
        maxNumOfSlots += ((lenOfSide - delta*maxNumOfSlots) / delta);

#ifdef DEBUG
        if (lenOfSide < delta * maxNumOfSlots) {
          Bear_PrintErr("GenSlotsForChildrenOfNode failed: programming error");
            return (BEAR_ERROR);
        }
#endif DEBUG

        MALLOC(Ncl_SlotData *, slot, maxNumOfSlots * sizeof (Ncl_SlotData));
        y = tileRect.r_ytop;
        for (j = 0, x = tileRect.r_xbot + delta / 2;
            j < maxNumOfSlots;
            j++, x += delta) {

            slot[j].slotDir = NCL_NORTH_DIR;
            slot[j].tilePos.p_x = x;
            slot[j].tilePos.p_y = y;
            slot[j].col = k++;
        }

        childNodePtr->nd_northSlot = slot;
        childNodePtr->nd_numOfNorthSlots = maxNumOfSlots;
        nplNumOfCMCols += maxNumOfSlots;
        
        /* 
         * Generate the east slots.
         */
        lenOfSide = tileRect.r_ytop - tileRect.r_ybot;
        maxNumOfSlots = lenOfSide / minPinSpacing;
        if (maxNumOfSlots == 0) maxNumOfSlots = 1;
        pinCount = childNodePtr->nd_sideFloPinCnt[NCL_EAST_DIR] +
            ((childNodePtr->nd_freeFloPinCnt/4) + 1);
        a = NPLMAX(pinCount, numOfSlots[i][NCL_EAST_DIR]);
        maxNumOfSlots = NPLMIN(maxNumOfSlots, a);
        delta = lenOfSide / maxNumOfSlots;
        maxNumOfSlots += ((lenOfSide - delta*maxNumOfSlots) / delta);

#ifdef DEBUG
        if (lenOfSide < delta * maxNumOfSlots) {
          Bear_PrintErr("GenSlotsForChildrenOfNode failed: programming error");
            return (BEAR_ERROR);
        }
#endif DEBUG

        MALLOC(Ncl_SlotData *, slot, maxNumOfSlots * sizeof (Ncl_SlotData));
        x = tileRect.r_xtop;
        for (j = 0, y = tileRect.r_ybot + delta / 2;
            j < maxNumOfSlots;
            j++, y += delta) {

            slot[j].slotDir = NCL_EAST_DIR;
            slot[j].tilePos.p_x = x;
            slot[j].tilePos.p_y = y;
            slot[j].col = k++;
        }

        childNodePtr->nd_eastSlot = slot;
        childNodePtr->nd_numOfEastSlots = maxNumOfSlots;
        nplNumOfCMCols += maxNumOfSlots;

        /* 
         * Generate the south slots.
         */
        lenOfSide = tileRect.r_xtop - tileRect.r_xbot;
        maxNumOfSlots = lenOfSide / minPinSpacing;
        if (maxNumOfSlots == 0) maxNumOfSlots = 1;
        pinCount = childNodePtr->nd_sideFloPinCnt[NCL_SOUTH_DIR] +
            ((childNodePtr->nd_freeFloPinCnt/4) + 1);
        a = NPLMAX(pinCount, numOfSlots[i][NCL_SOUTH_DIR]);
        maxNumOfSlots = NPLMIN(maxNumOfSlots, a);
        delta = lenOfSide / maxNumOfSlots;
        maxNumOfSlots += ((lenOfSide - delta*maxNumOfSlots) / delta);

#ifdef DEBUG
        if (lenOfSide < delta * maxNumOfSlots) {
            Bear_PrintErr("GenSlotsForChildrenOfNode failed: programming error");
            return (BEAR_ERROR);
        }
#endif DEBUG

        MALLOC(Ncl_SlotData *, slot, maxNumOfSlots * sizeof (Ncl_SlotData));
        y = tileRect.r_ybot;
        for (j = 0, x = tileRect.r_xbot + delta / 2;
            j < maxNumOfSlots;
            j++, x += delta) {

            slot[j].slotDir = NCL_SOUTH_DIR;
            slot[j].tilePos.p_x = x;
            slot[j].tilePos.p_y = y;
            slot[j].col = k++;
        }

        childNodePtr->nd_southSlot = slot;
        childNodePtr->nd_numOfSouthSlots = maxNumOfSlots;
        nplNumOfCMCols += maxNumOfSlots;

        /* 
         * Generate the west slots.
         */
        lenOfSide = tileRect.r_ytop - tileRect.r_ybot;
        maxNumOfSlots = lenOfSide / minPinSpacing;
        if (maxNumOfSlots == 0) maxNumOfSlots = 1;
        pinCount = childNodePtr->nd_sideFloPinCnt[NCL_WEST_DIR] +
            ((childNodePtr->nd_freeFloPinCnt/4) + 1);
        a = NPLMAX(pinCount, numOfSlots[i][NCL_WEST_DIR]);
        maxNumOfSlots = NPLMIN(maxNumOfSlots, a);
        delta = lenOfSide / maxNumOfSlots;
        maxNumOfSlots += ((lenOfSide - delta*maxNumOfSlots) / delta);

#ifdef DEBUG
        if (lenOfSide < delta * maxNumOfSlots) {
            Bear_PrintErr("GenSlotsForChildrenOfNode failed: programming error");
            return (BEAR_ERROR);
        }
#endif DEBUG

        MALLOC(Ncl_SlotData *, slot, maxNumOfSlots * sizeof (Ncl_SlotData));
        x = tileRect.r_xbot;
        for (j = 0, y = tileRect.r_ybot + delta / 2;
            j < maxNumOfSlots;
            j++, y += delta) {

            slot[j].slotDir = NCL_WEST_DIR;
            slot[j].tilePos.p_x = x;
            slot[j].tilePos.p_y = y;
            slot[j].col = k++;
        }

        childNodePtr->nd_westSlot = slot;
        childNodePtr->nd_numOfWestSlots = maxNumOfSlots;
        nplNumOfCMCols += maxNumOfSlots;
    }

    MALLOC(Ncl_SlotData **, nplSlot, nplNumOfCMCols * sizeof (Ncl_SlotData *));
    k = 0;
    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];
    
        if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == TRUE) {
            continue;
        }
    
        for (j = 0; j < childNodePtr->nd_numOfNorthSlots; j++) {
            nplSlot[k++] = &childNodePtr->nd_northSlot[j];
        }
        for (j = 0; j < childNodePtr->nd_numOfEastSlots; j++) {
            nplSlot[k++] = &childNodePtr->nd_eastSlot[j];
        }
        for (j = 0; j < childNodePtr->nd_numOfSouthSlots; j++) {
            nplSlot[k++] = &childNodePtr->nd_southSlot[j];
        }
        for (j = 0; j < childNodePtr->nd_numOfWestSlots; j++) {
            nplSlot[k++] = &childNodePtr->nd_westSlot[j];
        }
    }

    return (BEAR_OK);
}



static lsStatus 
HandleNet(item, arg)
lsGeneric   item;
lsGeneric   arg;
{
    lsGen           junctGen, pinGen;
    lsList          junctList;
    Gt_Graph        baseGraph, compGraph;
    HandleData      *handleDataPtr;
    Npl_JunctData   *junctDataPtr;
    Mac_GraphData   *graphDataPtr;
    Net             *netPtr;
    Ncl_Node        *nodePtr, *pinNodePtr;
    Npl_PinData     *pinDataPtr;
    Npl_NetData     *netDataPtr;
    WallJunction    *junctPtr;
    Tile            *bnTilePtr;
    Rect            encloseRect, junctArea;
    char            *netNamePtr;
    int             i, row, col;
    float           criticality;
    bool            loopFlag, connFlag, horzTileDirFlag;

    netDataPtr = (Npl_NetData *) item;
    handleDataPtr = (HandleData *) arg;

    netPtr = netDataPtr->netPtr;
    nodePtr = handleDataPtr->nodePtr;

    criticality = netDataPtr->criticality + 1;

    for (i = 1; i < nplNumOfCMCols + 1; i++) {
        nplCostOfNet[i].value = LA_INFINITE_COST;
    }

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        nodePtr->nd_children[i]->nd_mark = BEAR_UNMARKED;
    }

    lsForeachItem(netDataPtr->pinList, pinGen, pinDataPtr) {
        if (pinDataPtr->pinKind != NPL_BORDER_PIN) {
            pinDataPtr->nodePtr->nd_mark = BEAR_MARKED;
        }
    }

    netNamePtr = Mac_GetNameOfNet(netPtr);
    if (Ngr_BuildBaseGraph(handleDataPtr->bldTreePtr,   
        netPtr, &baseGraph) != BEAR_OK) {
        return (LS_STOP);
    }
    if (Ngr_BuildCompGraph(handleDataPtr->bldTreePtr,
        &compGraph) != BEAR_OK) {
        return (LS_STOP);
    }
    if (Ngr_FindMST(handleDataPtr->bldTreePtr) != BEAR_OK) {
        return (LS_STOP);
    }
    if (Ngr_FindSteinerTree(handleDataPtr->bldTreePtr) != BEAR_OK) {
        return (LS_STOP);
    }
    Ngr_DeleteCompGraph(&compGraph);

    graphDataPtr = (Mac_GraphData *) Gt_GetGraphDataOfGraph(baseGraph);
    junctList = graphDataPtr->junctList;

    if (lsLength(junctList) > 0) {
        CalcCostOfSlotsFromJunctList(nodePtr, junctList, criticality);
    }
    else {
        if (NplCalcEncloseRect(netDataPtr, &encloseRect) != BEAR_OK) {
            return (LS_STOP);
        }
        if (NplTileDirOfRect(netDataPtr, &horzTileDirFlag) != BEAR_OK) {
            return (LS_STOP);
        }

        if (horzTileDirFlag) {
            encloseRect.r_ybot = 
                (encloseRect.r_ybot + encloseRect.r_ytop) / 2;
            encloseRect.r_ytop = encloseRect.r_ybot + 1;
        } 
        else {
            encloseRect.r_xbot = 
                (encloseRect.r_xbot + encloseRect.r_xtop) / 2;
            encloseRect.r_xtop = encloseRect.r_xbot + 1;
        }
        CalcCostOfSlotsFromRect(nodePtr, &encloseRect, criticality);
    }

    lsForeachItem(junctList, junctGen, junctPtr) {
        junctDataPtr = (Npl_JunctData *) junctPtr->nplModulePtr;
        PutCostOfSlotsWithinRect(nodePtr, &junctDataPtr->junctArea, 0);
    }

    lsForeachItem(netDataPtr->pinList, pinGen, pinDataPtr) {
        if (pinDataPtr->pinKind == NPL_BORDER_PIN) {
            continue;
        }
        pinNodePtr = pinDataPtr->nodePtr;
        if (pinNodePtr->nd_hasSlotsFlag == FALSE) {
            continue;
        }

        row = pinDataPtr->row;

        if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
            (pinDataPtr->pinFlexType == FLEXIBLE_NORTH)) {
            for (i = 0; i < pinNodePtr->nd_numOfNorthSlots ;i++) {
                col = pinNodePtr->nd_northSlot[i].col;
                nplCostMatrix[NplCostMatrixIndex(row, col)].value =
                    nplCostOfNet[col + 1].value;
            }
        }
        if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
            (pinDataPtr->pinFlexType == FLEXIBLE_EAST)) {
            for (i = 0; i < pinNodePtr->nd_numOfEastSlots ;i++) {
                col = pinNodePtr->nd_eastSlot[i].col;
                nplCostMatrix[NplCostMatrixIndex(row, col)].value =
                    nplCostOfNet[col + 1].value;
            }
        }
        if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
            (pinDataPtr->pinFlexType == FLEXIBLE_SOUTH)) {
            for (i = 0; i < pinNodePtr->nd_numOfSouthSlots ;i++) {
                col = pinNodePtr->nd_southSlot[i].col;
                nplCostMatrix[NplCostMatrixIndex(row, col)].value =
                    nplCostOfNet[col + 1].value;
            }
        }
        if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
            (pinDataPtr->pinFlexType == FLEXIBLE_WEST)) {
            for (i = 0; i < pinNodePtr->nd_numOfWestSlots ;i++) {
                col = pinNodePtr->nd_westSlot[i].col;
                nplCostMatrix[NplCostMatrixIndex(row, col)].value =
                    nplCostOfNet[col + 1].value;
            }
        }
    }

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && (nplInteractiveFlag == TRUE)) {
        char msg[80];
        
        if (nplPlaceWinPtr != NIL(Npl_PlaceWin)) {
            nplPlaceWinPtr->nodePtr = nodePtr;
            nplPlaceWinPtr->graph = baseGraph;
            nplPlaceWinPtr->planeDataPtr = handleDataPtr->planeDataPtr;
            nplPlaceWinPtr->costOfNet = nplCostOfNet;

            NplInitPlaceWinView();
           (void) NplDrawPlacementRegion(nplPlaceWinPtr->winDataPtr, NIL(Rect));
        }
        Ez_Flush();

        sprintf (msg, "Slots for net %8s - Keep going", netNamePtr);
        if (NplKeepGoing(msg) == FALSE) {
            if (nplPlaceWinPtr != NIL(Npl_PlaceWin)) {
                nplPlaceWinPtr->nodePtr = NIL(Ncl_Node);
            }
            return (LS_STOP);
        }

        NplDrawChipWin();

        if (nplPlaceWinPtr != NIL(Npl_PlaceWin)) {
            nplPlaceWinPtr->nodePtr = NIL(Ncl_Node);
        }
    }
#endif BEAR_BLACK_BOX

    Ngr_DeleteBaseGraph(&baseGraph);

    return (LS_OK);
}



static BearStatus
SetUpSlotsBeforeLA(nodePtr, loopIndex, solVec)
Ncl_Node        *nodePtr;
int             loopIndex;
La_CostEntry    *solVec;
{
    int i;

    for (i = 1; i < nplNumOfCMRows * nplNumOfCMCols + 1; i++) {
        nplCostMatrix[i].value = LA_INFINITE_COST;
    }
    for (i = 1; i < nplNumOfCMRows + 1; i++) {
        solVec[i].value = 0;
    }

    if (loopIndex == 1) {
        lsGen       netGen;
        Npl_NetData *netDataPtr;
        Rect        touchRect;

        lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
            if (NplCalcTouchRect(netDataPtr, &touchRect) != BEAR_OK) {
                lsFinish(netGen);
                return (BEAR_ERROR);
            }
            if (CalcSlotCostForNet(netDataPtr, &touchRect) != BEAR_OK) {
                lsFinish(netGen);
                return (BEAR_ERROR);
            }
        }
    }
    else {
        FloorPlanGraph  *fpgPtr;
        HandleData      handleData;
        PlaneData       *planeDataPtr;

        planeDataPtr = Mac_GetCellPlaneOfCell(nplCoreMasterPtr);
        fpgPtr = Tp_GetFpg(planeDataPtr);

        (void) Fpg_ItrWallJunctSeq(fpgPtr, StoreJunctData,
            (ClientData) planeDataPtr);

        if (Ngr_BldTreeBegin(nplCoreMasterPtr, NIL(Ngr_BldOptions),
            &handleData.bldTreePtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }
        handleData.nodePtr = nodePtr;
        handleData.planeDataPtr = planeDataPtr;
        if (lsForeach(nodePtr->nd_totalNetList, HandleNet,
            (lsGeneric) &handleData) == LS_STOP) {
            return (BEAR_ERROR);
        }

        (void) Ngr_BldTreeEnd(handleData.bldTreePtr);

        (void) Fpg_ItrWallJunctSeq(fpgPtr, RemoveJunctData, (ClientData) NULL);
    }

    return (BEAR_OK);
}


static void
FreeSlotsForChildrenOfNode(nodePtr)
Ncl_Node *nodePtr;
{
    Ncl_Node    *childNodePtr;
    int         i;

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];

        FREE((char *) childNodePtr->nd_northSlot);
        childNodePtr->nd_northSlot = NIL(Ncl_SlotData);
        childNodePtr->nd_numOfNorthSlots = 0;

        FREE((char *) childNodePtr->nd_eastSlot);
        childNodePtr->nd_eastSlot = NIL(Ncl_SlotData);
        childNodePtr->nd_numOfEastSlots = 0;

        FREE((char *) childNodePtr->nd_southSlot);
        childNodePtr->nd_southSlot = NIL(Ncl_SlotData);
        childNodePtr->nd_numOfSouthSlots = 0;

        FREE((char *) childNodePtr->nd_westSlot);
        childNodePtr->nd_westSlot = NIL(Ncl_SlotData);
        childNodePtr->nd_numOfWestSlots = 0;
    }
}



static void
PutCostOfSlotsWithinRect(nodePtr, rectPtr, slotCost)
Ncl_Node    *nodePtr;
Rect        *rectPtr;
int         slotCost;
{
    int         i, j;
    Ncl_SlotData slotData;
    Ncl_Node    *childNodePtr;

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];
        if (childNodePtr->nd_mark == BEAR_MARKED) {

            for (j = 0; j < childNodePtr->nd_numOfNorthSlots; j++) {
                slotData = childNodePtr->nd_northSlot[j];
                if (GEO_ENCLOSE(&slotData.tilePos, rectPtr) == TRUE) {
                    nplCostOfNet[slotData.col + 1].value = slotCost;
                }
            }
            for (j = 0; j < childNodePtr->nd_numOfEastSlots; j++) {
                slotData = childNodePtr->nd_eastSlot[j];
                if (GEO_ENCLOSE(&slotData.tilePos, rectPtr) == TRUE) {
                    nplCostOfNet[slotData.col + 1].value = slotCost;
                }
            }
            for (j = 0; j < childNodePtr->nd_numOfSouthSlots; j++) {
                slotData = childNodePtr->nd_southSlot[j];
                if (GEO_ENCLOSE(&slotData.tilePos, rectPtr) == TRUE) {
                    nplCostOfNet[slotData.col + 1].value = slotCost;
                }
            }
            for (j = 0; j < childNodePtr->nd_numOfWestSlots; j++) {
                slotData = childNodePtr->nd_westSlot[j];
                if (GEO_ENCLOSE(&slotData.tilePos, rectPtr) == TRUE) {
                    nplCostOfNet[slotData.col + 1].value = slotCost;
                }
            }
        }
    }
}



static void
CalcCostOfSlotsFromRect(nodePtr, rectPtr, criticality)
Ncl_Node    *nodePtr;
Rect        *rectPtr;
float       criticality;
{
    int         i, j;
    Ncl_SlotData slotData;
    Ncl_Node    *childNodePtr;

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];
        if (childNodePtr->nd_mark == BEAR_MARKED) {

            for (j = 0; j < childNodePtr->nd_numOfNorthSlots; j++) {
                slotData = childNodePtr->nd_northSlot[j];
                nplCostOfNet[slotData.col + 1].value = 
                    NplDistanceFromRect(&(slotData.tilePos), rectPtr) *
                        criticality;
            }
            for (j = 0; j < childNodePtr->nd_numOfEastSlots; j++) {
                slotData = childNodePtr->nd_eastSlot[j];
                nplCostOfNet[slotData.col + 1].value = 
                    NplDistanceFromRect(&(slotData.tilePos), rectPtr) * 
                        criticality;
            }
            for (j = 0; j < childNodePtr->nd_numOfSouthSlots; j++) {
                slotData = childNodePtr->nd_southSlot[j];
                nplCostOfNet[slotData.col + 1].value = 
                    NplDistanceFromRect(&(slotData.tilePos), rectPtr) *
                        criticality;
            }
            for (j = 0; j < childNodePtr->nd_numOfWestSlots; j++) {
                slotData = childNodePtr->nd_westSlot[j];
                nplCostOfNet[slotData.col + 1].value = 
                    NplDistanceFromRect(&(slotData.tilePos), rectPtr) *
                        criticality;
            }
        }
    }
}



static BearStatus
SetUpNodeAfterLA(nodePtr, loopIndex, solVec)
Ncl_Node        *nodePtr;
int             loopIndex;
La_CostEntry    *solVec;
{
    CellMaster      *cellMasterPtr;
    Ncl_SlotData    *slotDataPtr;
    LinkedLineSeg   *lnLineSegPtr;
    Ncl_Node        *childNodePtr;
    Npl_PinData     *pinDataPtr;
    int             cmIndex;            /* cost matrix index */
    int             i, s, row, col;

    if ((loopIndex == 1) && (nodePtr->nd_type == NON_LEAF)) {
        for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
            NclForeachSide(s) {
                nodePtr->nd_children[i]->nd_initPinList[s] = (lsList) NULL;
            }
        }
    }
    else {
        for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
            childNodePtr = nodePtr->nd_children[i];
            if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == TRUE) {
                continue;
            }
            NclForeachSide(s) {
                if (childNodePtr->nd_initPinList[s] != (lsList) NULL) {
                    lsDestroy(childNodePtr->nd_initPinList[s],
                        (void (*)()) NULL);
                    childNodePtr->nd_initPinList[s] = (lsList) NULL;
                }

                childNodePtr->nd_tilePinList[s] = lsCreate();
            }
        }
    }

    for (i = 1; i < nplNumOfCMRows + 1; i++) {
        cmIndex = solVec[i].value - 1;
        col = cmIndex % nplNumOfCMCols;
        row = (cmIndex - col) / nplNumOfCMCols;

        if ((nplSlot[col]->col != col) || (nplPinData[row]->row != row)) {
      Bear_PrintErr("SetUpNodeAfterLA() failed: bad decoding of solVec.");
            return (BEAR_ERROR);
        }

        slotDataPtr = nplSlot[col];
        pinDataPtr = nplPinData[row];

        pinDataPtr->tilePos = slotDataPtr->tilePos;
        pinDataPtr->tileDir = slotDataPtr->slotDir;

        lsNewEnd(pinDataPtr->nodePtr->nd_tilePinList[slotDataPtr->slotDir],
            (lsGeneric) pinDataPtr, LS_NH);
    }

    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == TRUE) {
            continue;
        }

        NplCalcInitPinLists(childNodePtr);

        for (lnLineSegPtr = Mac_GetBoundaryOfCell(childNodePtr->nd_cellInstPtr);
            lnLineSegPtr != NIL(LinkedLineSeg);
            lnLineSegPtr = lnLineSegPtr->nextPtr) {

            if ((loopIndex == 1) && (nodePtr->nd_type == NON_LEAF)) {
                if (NplAttachPinListOne(lnLineSegPtr->lineSeg, childNodePtr) == 
                    BEAR_ERROR) {

                    return (BEAR_ERROR);
                }
            }
            else {
                if (Mac_IsFixedCell(childNodePtr->nd_cellInstPtr) == FALSE) {
                    NplAttachPinListFour(lnLineSegPtr->lineSeg, childNodePtr);
                }
            }
        }
    }

    return (BEAR_OK);
}



static void
CalcJunctArea(planeDataPtr, junctPtr, junctAreaPtr)
PlaneData       *planeDataPtr;
WallJunction    *junctPtr;
Rect            *junctAreaPtr;
{
    BottleNeck  *bnPtr;
    Point       *horzPointPtr, vertPoint;
    Rect        bnTileRect, domTileRect;
    Tile        *domTilePtr;
    WallSegment *segPtr;
    int         a;

    horzPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    Tp_HPointToVPoint(horzPointPtr, &vertPoint);

    if ((segPtr = Fpg_GetNorthSegOfWallJunct(junctPtr)) != NIL(WallSegment)) {
        bnPtr = Fpg_GetBottleNeckOfWallSeg(segPtr);
        if (bnPtr->unType == TP_BN_TILE) {
            Tp_HorzTileToRect(planeDataPtr, bnPtr->unValue.bnTilePtr,
                &bnTileRect);
            a = bnTileRect.r_ybot;
        }
        else {
            a = bnPtr->unValue.bnLinePtr->line.point1.p_y;
        }
    }
    else {
        domTilePtr = TiSrPoint(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
            &vertPoint);
        Tp_VertTileToRect(planeDataPtr, domTilePtr, &domTileRect);
        a = domTileRect.r_ytop;
    }
    junctAreaPtr->r_ytop = a;

    if ((segPtr = Fpg_GetEastSegOfWallJunct(junctPtr)) != NIL(WallSegment)) {
        bnPtr = Fpg_GetBottleNeckOfWallSeg(segPtr);
        if (bnPtr->unType == TP_BN_TILE) {
            Tp_VertTileToRect(planeDataPtr, bnPtr->unValue.bnTilePtr,
                &bnTileRect);
            a = bnTileRect.r_xbot;
        }
        else {
            a = bnPtr->unValue.bnLinePtr->line.point1.p_x;
        }
    }
    else {
        domTilePtr = TiSrPoint(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
            horzPointPtr);
        Tp_HorzTileToRect(planeDataPtr, domTilePtr, &domTileRect);
        a = domTileRect.r_xtop;
    }
    junctAreaPtr->r_xtop = a;

    if ((segPtr = Fpg_GetSouthSegOfWallJunct(junctPtr)) != NIL(WallSegment)) {
        bnPtr = Fpg_GetBottleNeckOfWallSeg(segPtr);
        if (bnPtr->unType == TP_BN_TILE) {
            Tp_HorzTileToRect(planeDataPtr, bnPtr->unValue.bnTilePtr,
                &bnTileRect);
            a = bnTileRect.r_ytop;
        }
        else {
            a = bnPtr->unValue.bnLinePtr->line.point1.p_y;
        }
    }
    else {
        domTilePtr = TiSrPoint(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
            &vertPoint);
        Tp_VertTileToRect(planeDataPtr, domTilePtr, &domTileRect);
        a = domTileRect.r_ybot;
    }
    junctAreaPtr->r_ybot = a;

    if ((segPtr = Fpg_GetWestSegOfWallJunct(junctPtr)) != NIL(WallSegment)) {
        bnPtr = Fpg_GetBottleNeckOfWallSeg(segPtr);
        if (bnPtr->unType == TP_BN_TILE) {
            Tp_VertTileToRect(planeDataPtr, bnPtr->unValue.bnTilePtr,
                &bnTileRect);
            a = bnTileRect.r_xtop;
        }
        else {
            a = bnPtr->unValue.bnLinePtr->line.point1.p_x;
        }
    }
    else {
        domTilePtr = TiSrPoint(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
            horzPointPtr);
        Tp_HorzTileToRect(planeDataPtr, domTilePtr, &domTileRect);
        a = domTileRect.r_xbot;
    }
    junctAreaPtr->r_xbot = a;
}



static BearStatus
FillSearchData(bnSearchDataPtr)
BNSearchData *bnSearchDataPtr;
{
    Rect    hSearchRect, searchRect, tileRect;
    Plane   *planePtr;
    BNLine  *bnLineList;

    if (bnSearchDataPtr->planeDataPtr == NIL(PlaneData)) {
        Bear_PrintErr("FillSearchData() failed: nil planeDataPtr.");
        return (BEAR_ERROR);
    }
    tileRect = bnSearchDataPtr->nodePtr->nd_tileRect;

    switch (bnSearchDataPtr->searchDir) {
    case NCL_NORTH_DIR:
        bnLineList = Tp_GetVertBnLineList(bnSearchDataPtr->planeDataPtr),
        hSearchRect = tileRect;
        hSearchRect.r_ybot = tileRect.r_ytop;
        hSearchRect.r_ytop++;
        Tp_HRectToVRect(&hSearchRect, &searchRect);
        planePtr = Tp_GetVertPlane(bnSearchDataPtr->planeDataPtr);
        break;
    case NCL_EAST_DIR:
        bnLineList = Tp_GetHorzBnLineList(bnSearchDataPtr->planeDataPtr),
        searchRect = tileRect;
        searchRect.r_xbot = tileRect.r_xtop;
        searchRect.r_xtop++;
        planePtr = Tp_GetHorzPlane(bnSearchDataPtr->planeDataPtr);
        break;
    case NCL_SOUTH_DIR:
        bnLineList = Tp_GetVertBnLineList(bnSearchDataPtr->planeDataPtr),
        hSearchRect = tileRect;
        hSearchRect.r_ybot--;
        hSearchRect.r_ytop = tileRect.r_ybot;
        Tp_HRectToVRect(&hSearchRect, &searchRect);
        planePtr = Tp_GetVertPlane(bnSearchDataPtr->planeDataPtr);
        break;
    case NCL_WEST_DIR:
        bnLineList = Tp_GetHorzBnLineList(bnSearchDataPtr->planeDataPtr),
        searchRect = tileRect;
        searchRect.r_xbot--;
        searchRect.r_xtop = tileRect.r_xbot;
        planePtr = Tp_GetHorzPlane(bnSearchDataPtr->planeDataPtr);
        break;
    default:
        Bear_PrintErr("FillSearchData() failed: bad search direction.");
        return (BEAR_ERROR);
    }

    bnSearchDataPtr->delta = - NPLMAX_INT;
    if (TiSrArea(NIL(Tile), planePtr, &searchRect, CalcSlotSpacingForBnTile,
            (ClientData) bnSearchDataPtr) == MAGIC_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

    if (bnSearchDataPtr->delta == - NPLMAX_INT) {
        if (Tp_ItrBnLine(bnLineList, CalcSlotSpacingForBnLine,
                (ClientData) bnSearchDataPtr) == BEAR_STOP_SEARCH) {
            return (BEAR_ERROR);
        }
        if (bnSearchDataPtr->delta == - NPLMAX_INT) {
            Bear_PrintErr("FillSearchData() failed : no bn tiles or bn lines found.");
            return (BEAR_ERROR);
        }
    }

    /* 
     * We multiply the minimum pin spacing by 2 so that the
     * channel pin arrangement algorithum will have room to
     * accommodate all the pins while satisfying the spacing 
     * requirement from the bottleneck edge. MP
     */
    if (bnSearchDataPtr->delta < 2*ncl_WireWidth) {
        bnSearchDataPtr->delta = 2*ncl_WireWidth;
    }

    return (BEAR_OK);
}



static int
CalcSlotSpacingForBnTile(tilePtr, bnSearchDataPtr)
register Tile   *tilePtr;
BNSearchData    *bnSearchDataPtr;
{
    int     bnHeight, bnWidth, slotToSlot;
    Rect    rect;

    if (Tp_IsBottleneckTile(tilePtr)) {
        if (Tp_IsHorzTile(tilePtr)) {
            Tp_HorzTileToRect(bnSearchDataPtr->planeDataPtr, tilePtr, &rect);
            bnWidth = rect.r_xtop - rect.r_xbot;
/*
 * NEW CODE FOR DUMMY CELLS HERE (optional)
 */
            bnHeight = rect.r_ytop - rect.r_ybot;
  
            if (bnWidth < ncl_WireWidth) {
                Bear_PrintErr(
                    "CalcSlotSpacingForBnTile() failed : horz bottleneck too thin.");
                return (MAGIC_STOP_SEARCH);
            }

            slotToSlot = (bnHeight * ncl_WireWidth) / bnWidth;
        }
        else {
            Tp_VertTileToRect(bnSearchDataPtr->planeDataPtr, tilePtr, &rect);
            bnWidth = rect.r_xtop - rect.r_xbot;
/*
 * NEW CODE HERE
 */
            bnHeight = rect.r_ytop - rect.r_ybot;

            if (bnHeight < ncl_WireWidth) {
                Bear_PrintErr(
                    "CalcSlotSpacingForBnTile() failed: vert bottleneck too thin.");
                return (MAGIC_STOP_SEARCH);
            }
   
            slotToSlot = (bnWidth * ncl_WireWidth) / bnHeight;
        }

        if (slotToSlot > bnSearchDataPtr->delta) {
            bnSearchDataPtr->delta = slotToSlot;
        }
    }

    return (MAGIC_CONTINUE_SEARCH);
}


static BearStatus 
CalcSlotSpacingForBnLine(bnLinePtr, bnSearchDataPtr)
BNLine          *bnLinePtr;
BNSearchData    *bnSearchDataPtr;
{
    BNLineType  bnLineType;
    Line        *linePtr;
    Point       blSearchPoint, trSearchPoint;
    Rect        tileRect;

    if (bnLinePtr == NIL(BNLine)) {
        Bear_PrintFatalErr(
            "CalcSlotSpacingForBnLine() failed: no bottleneck line is given.");
    }

    tileRect = bnSearchDataPtr->nodePtr->nd_tileRect;

    bnLineType = Tp_GetBnLineTypeOfBnLine(bnLinePtr);
    linePtr = Tp_GetLineOfBnLine(bnLinePtr);

    if (bnLineType == TP_HORZ_B_RIGHT)
    {
        blSearchPoint.p_x = linePtr->point1.p_x - 1;
        blSearchPoint.p_y = linePtr->point1.p_y - 1;
        trSearchPoint.p_x = linePtr->point2.p_x + 1;
        trSearchPoint.p_y = linePtr->point2.p_y + 1;
    }
    else if (bnLineType == TP_HORZ_B_LEFT)
    {
        blSearchPoint.p_x = linePtr->point1.p_x - 1;
        blSearchPoint.p_y = linePtr->point1.p_y + 1;
        trSearchPoint.p_x = linePtr->point2.p_x + 1;
        trSearchPoint.p_y = linePtr->point2.p_y - 1;
    }
    else if (bnLineType == TP_VERT_R_UP)
    {
        blSearchPoint.p_x = linePtr->point1.p_x - 1;
        blSearchPoint.p_y = linePtr->point1.p_y + 1;
        trSearchPoint.p_x = linePtr->point2.p_x + 1;
        trSearchPoint.p_y = linePtr->point2.p_y - 1;
    }
    else if (bnLineType == TP_VERT_R_DOWN)
    {
        blSearchPoint.p_x = linePtr->point1.p_x - 1;
        blSearchPoint.p_y = linePtr->point1.p_y - 1;
        trSearchPoint.p_x = linePtr->point2.p_x + 1;
        trSearchPoint.p_y = linePtr->point2.p_y + 1;
    }
    else 
    {
        Bear_PrintFatalErr("CalcSlotSpacingForBnLine(): Illegal BNLineType!");
    }
    
    if (GEO_ENCLOSE(&trSearchPoint, &tileRect) || 
            GEO_ENCLOSE(&blSearchPoint, &tileRect)) {
        
        /*
        int capacity;

        if (bnLineType == TP_VERT_R_UP || bnLineType == TP_VERT_R_DOWN) {
            capacity = ABS(linePtr->point1.p_y - linePtr->point2.p_y);
        }
        else if (bnLineType == TP_HORZ_B_RIGHT || bnLineType == TP_HORZ_B_LEFT) {
            capacity = ABS(linePtr->point1.p_x - linePtr->point2.p_x);
        }
        */

        bnSearchDataPtr->delta = 2 * ncl_WireWidth;
    }

    return (BEAR_CONTINUE_SEARCH);
}



static void
CalcCostOfSlotsFromJunctList(nodePtr, junctList, criticality)
Ncl_Node    *nodePtr;
lsList      junctList;
float       criticality;
{
    lsGen           junctGen;
    Ncl_SlotData     slotData;
    Ncl_Node        *childNodePtr;
    Npl_JunctData   *junctDataPtr;
    WallJunction    *junctPtr;
    int             i, j;
    int             value;

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];
        if (childNodePtr->nd_mark == BEAR_MARKED) {

            for (j = 0; j < childNodePtr->nd_numOfNorthSlots; j++) {
                slotData = childNodePtr->nd_northSlot[j];

                lsForeachItem(junctList, junctGen, junctPtr) {
                    junctDataPtr = (Npl_JunctData *) junctPtr->nplModulePtr;
                    value = (int) NplDistanceFromRect(&(slotData.tilePos),
                            &junctDataPtr->junctArea) * criticality;
                    nplCostOfNet[slotData.col + 1].value =
                            NPLMIN(nplCostOfNet[slotData.col + 1].value, value);
                }
            }
            for (j = 0; j < childNodePtr->nd_numOfEastSlots; j++) {
                slotData = childNodePtr->nd_eastSlot[j];
                lsForeachItem(junctList, junctGen, junctPtr) {
                    junctDataPtr = (Npl_JunctData *) junctPtr->nplModulePtr;
                    value = (int) NplDistanceFromRect(&(slotData.tilePos),
                            &junctDataPtr->junctArea) * criticality;
                    nplCostOfNet[slotData.col + 1].value =
                            NPLMIN(nplCostOfNet[slotData.col + 1].value, value);
                }
            }
            for (j = 0; j < childNodePtr->nd_numOfSouthSlots; j++) {
                slotData = childNodePtr->nd_southSlot[j];
                lsForeachItem(junctList, junctGen, junctPtr) {
                    junctDataPtr = (Npl_JunctData *) junctPtr->nplModulePtr;
                    value = (int) NplDistanceFromRect(&(slotData.tilePos),
                            &junctDataPtr->junctArea) * criticality;
                    nplCostOfNet[slotData.col + 1].value =
                            NPLMIN(nplCostOfNet[slotData.col + 1].value, value);
                }
            }
            for (j = 0; j < childNodePtr->nd_numOfWestSlots; j++) {
                slotData = childNodePtr->nd_westSlot[j];
                lsForeachItem(junctList, junctGen, junctPtr) {
                    junctDataPtr = (Npl_JunctData *) junctPtr->nplModulePtr;
                    value = (int) NplDistanceFromRect(&(slotData.tilePos),
                            &junctDataPtr->junctArea) * criticality;
                    nplCostOfNet[slotData.col + 1].value =
                        NPLMIN(nplCostOfNet[slotData.col + 1].value, value);
                }
            }
        }
    }
}



static BearStatus
CalcSlotCostForPin(pinDataPtr, netRectPtr, criticality)
Npl_PinData     *pinDataPtr;
register Rect   *netRectPtr;
float           criticality;
{
    Ncl_SlotData slotData;
    Ncl_Node    *childNodePtr = pinDataPtr->nodePtr;
    int         j, row = pinDataPtr->row;

    if (pinDataPtr->pinKind == NPL_BORDER_PIN) {
        return (BEAR_OK);
    }

    if (childNodePtr == NIL(Ncl_Node)) {
        Bear_PrintErr("CalcSlotCostForPin() failed: childNodePtr is NIL.");
        return (BEAR_ERROR);
    }

    if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
        (pinDataPtr->pinFlexType == FLEXIBLE_NORTH)) {
        for (j = 0; j < childNodePtr->nd_numOfNorthSlots ;j++) {
            slotData = childNodePtr->nd_northSlot[j];
            nplCostMatrix[NplCostMatrixIndex(row, slotData.col)].value = 
                NplDistanceFromRect(&(slotData.tilePos), netRectPtr) *
                criticality;
        }
    }
    if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
        (pinDataPtr->pinFlexType == FLEXIBLE_EAST)) {
        for (j = 0; j < childNodePtr->nd_numOfEastSlots ;j++) {
            slotData = childNodePtr->nd_eastSlot[j];
            nplCostMatrix[NplCostMatrixIndex(row, slotData.col)].value = 
                NplDistanceFromRect(&(slotData.tilePos), netRectPtr) *
                criticality;
        }
    }
    if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
        (pinDataPtr->pinFlexType == FLEXIBLE_SOUTH)) {
        for (j = 0; j < childNodePtr->nd_numOfSouthSlots ;j++) {
            slotData = childNodePtr->nd_southSlot[j];
            nplCostMatrix[NplCostMatrixIndex(row, slotData.col)].value = 
                NplDistanceFromRect(&(slotData.tilePos), netRectPtr) *
                criticality;
        }
    }
    if ((pinDataPtr->pinFlexType == FLEXIBLE) ||
        (pinDataPtr->pinFlexType == FLEXIBLE_WEST)) {
        for (j = 0; j < childNodePtr->nd_numOfWestSlots ;j++) {
            slotData = childNodePtr->nd_westSlot[j];
            nplCostMatrix[NplCostMatrixIndex(row, slotData.col)].value = 
                NplDistanceFromRect(&(slotData.tilePos), netRectPtr) *
                criticality;
        }
    }

    return (BEAR_OK);
}



/* 
 * ---------------------------------------------------------------------------
 * CalcSlotCostForNet() --
 *
 *      This function updates the cost matrix entries for the net.
 *
 * Results:
 *      netRectPtr is assigned fields.
 *
 * Side Effects: 
 *      None.
 *
 * ---------------------------------------------------------------------------
 */

static BearStatus
CalcSlotCostForNet(netDataPtr, netRectPtr)
Npl_NetData *netDataPtr;
Rect        *netRectPtr;
{
    lsGen       pinGen;
    Npl_PinData *pinDataPtr;
    float       criticality;

    lsForeachItem(netDataPtr->pinList, pinGen, pinDataPtr) {
        criticality = netDataPtr->criticality + 1;
        if (pinDataPtr->pinKind == NPL_INSIDE_PIN) {
            if (CalcSlotCostForPin(pinDataPtr, netRectPtr, criticality) ==
                BEAR_ERROR) {
                lsFinish(pinGen);
                return (BEAR_ERROR);
            }
        }
    }

    return (BEAR_OK);
}


BearStatus
NplDoPinAssignment(nodePtr, minPinSpacing)
Ncl_Node    *nodePtr;
int         minPinSpacing;
{
    lsGen           netGen;
    CellMaster      *cellMasterPtr;
    La_CostEntry    *solVec;            /* solution vector */
    La_CostEntry    *capVec;            /* capacity vector */
    Ncl_Node        *childNodePtr;
    Npl_NetData     *netDataPtr;
    int             i, s, loopIndex;
    double          t1;

    t1 = NplPtime();

    for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
        childNodePtr = nodePtr->nd_children[i];
        if (childNodePtr->nd_cellInstPtr == NIL(CellInst)) {
            /*
             * These child nodes don't have any fixed pins.
             */
            if (NplWriteNodeToMac(childNodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
        else {
            /*
             * These child nodes may be hard blocks (i.e., have fixed pins).
             */
            if (NplRewriteNodeToMac(childNodePtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
    }

    if (NplSetUpCoreMaster(nplCoreMasterPtr, nplCoreNodeList) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    if (GenSlotsForChildrenOfNode(nodePtr, minPinSpacing) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    
    if ((nplNumOfCMRows == 0) && (nplNumOfCMCols == 0)) {
        /* No flexible pins.  Therefore, we don't assign pins. */

        if (NplSetUpMacNetList(nplCoreMasterPtr, nplNetList) != BEAR_OK) {
            return (BEAR_ERROR);
        }   
#ifdef DEBUG
        NplSaveCellMaster(nplCoreMasterPtr);
#endif DEBUG

        FREE((char *) nplSlot);
        FREE((char *) nplPinData);

        return (BEAR_OK);
    }

    /*
     * Inputs / Output for LinAssign package.
     */
    MALLOC(La_CostEntry *, nplCostMatrix,
        (nplNumOfCMRows * nplNumOfCMCols + 1) * sizeof (La_CostEntry));
    MALLOC(La_CostEntry *, nplCostOfNet,
        (nplNumOfCMCols + 1) * sizeof (La_CostEntry));
    MALLOC(La_CostEntry *, capVec,
        (nplNumOfCMCols + 1) * sizeof (La_CostEntry));
    MALLOC(La_CostEntry *, solVec,
        (nplNumOfCMRows + 1) * sizeof (La_CostEntry));

    for (i = 1; i < (nplNumOfCMCols + 1) ;i++) {
        capVec[i].value = LA_MAX_CAP;
    }

    for (loopIndex = 1; loopIndex <= nplMaxLoopCnt+1; loopIndex++) {

        if (SetUpSlotsBeforeLA(nodePtr, loopIndex, solVec) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (La_LinAssign(nplNumOfCMRows, nplNumOfCMCols, capVec, 
            nplCostMatrix, solVec) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (SetUpNodeAfterLA(nodePtr, loopIndex, solVec) != BEAR_OK) {
            return (BEAR_ERROR);
        }
/*
 *      fprintf(stderr, "After linear assignment.\n");
 *      NplCheckNetList(nplNetList, TRUE);
 */
        if (NplSetUpCoreMaster(nplCoreMasterPtr, nplCoreNodeList) != BEAR_OK) {
            return (BEAR_ERROR);
        }
        if (NplSetUpMacNetList(nplCoreMasterPtr, nplNetList) != BEAR_OK) {
            return (BEAR_ERROR);
        }

#ifdef DEBUG
        NplSaveCellMaster(nplCoreMasterPtr);
#endif /* DEBUG */
/*
 *      fprintf(stderr, "After updating MAC net lists.\n");
 *      NplCheckNetList(nplNetList, TRUE);
 */
#ifndef BEAR_BLACK_BOX
        if (bear_BlackBoxFlag == FALSE) {
            NplInitChipWinView(globalWinDataPtr, nplCoreMasterPtr);

            if (nplInteractiveFlag == TRUE) {
                char msg[80];

                NplDrawChipWin();

                (void) sprintf (msg, "After %d iteration(s) - Keep going", 
                    loopIndex);
                if (NplKeepGoing(msg) == FALSE) {
                    return (BEAR_ERROR);
                }
            }
        }
#endif /* BEAR_BLACK_BOX */
    }

    FreeSlotsForChildrenOfNode(nodePtr);

    FREE((char *) nplSlot);
    FREE((char *) nplPinData);
    FREE((char *) nplCostMatrix);
    FREE((char *) nplCostOfNet);
    FREE((char *) capVec);
    FREE((char *) solVec);

    nplNumOfCMRows = nplNumOfCMCols = 0;
    nplSlot = (Ncl_SlotData **) NULL;
    nplPinData = (Npl_PinData **) NULL;
    nplCostMatrix = NIL(La_CostEntry);
    nplCostOfNet = NIL(La_CostEntry);
    capVec = NIL(La_CostEntry);
    solVec = NIL(La_CostEntry);

    (void) fprintf (stdout,
        "CPU-time (Pin Assignment) = %g sec\n", (NplPtime() - t1));
    fflush(stdout);

    return (BEAR_OK);
}


