
/* nplPrint.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplPrint.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "list.h"
#include "newP.h"
#include "newPInt.h"
#include "mst.h"

#include "ezPlot.h"
#include "../polarBear/polarBear.h"


/*
 * ----------------------------------------------------------------------------
 *
 * NplPrintResults() --
 *
 *     Prints out the 
 *     - area usage (if no routing area estimation)
 *     - the final area (if with routing area estimation)
 *
 * Results:
 *     None.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

void
NplPrintResults()
/*
 * bug fix (George Carvalho 18 April 91)
 *     The variables used to calculate the width, height, and area
 *     have been changed from integer to double in order to 
 *     accomodate a larger range of values.
 */
{
    double w, h, area;

    w = ncl_ClusterTree->nd_dim.p_x; 
    h = ncl_ClusterTree->nd_dim.p_y;
    area = w * h;

    (void) fprintf (bearout, "\nChip size: %.0f x %.0f = %.0f\n", 
        w, h, area);

    fflush (bearout);
}



void
NplPrintLevelStat(nodePtr, stat)
Ncl_Node    *nodePtr;
int         stat;
{
    static int number = 0;

    if (stat == 2) {
        fprintf(stderr, "level %d  (number of clusters %d)\n\n",
                NplNodeDepth(nodePtr), number);
        number = 0;
    } else if (stat == 1) {
        number++;
    }
}


void
NplPrintData(nodeList, netList)
lsList nodeList;
lsList netList;
{
    Ncl_Node    *nodePtr;
    Npl_NetData *netDataPtr;
    lsGen       netGen, nodeGen;
    int         depth, i;

    printf ("Number of Nodes = %d\n", lsLength(nodeList));
    printf ("Number of Nets  = %d\n\n", lsLength(netList));
    fflush (stdout);

    for (depth = 0; depth < ncl_ClusterTreeDepth ; depth++) {
        lsForeachItem(nodeList, nodeGen, nodePtr) {
            if (nodePtr->nd_depth == depth) {
                printf ("node %d (%d children)\n", nodePtr->nd_id,
                    nodePtr->nd_numOfChildren);
                for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
                    printf ("    node %d\n", nodePtr->nd_children[i]->nd_id);
                }
                printf ("\n");
            }
        }
    }
    printf ("----------\n\n");
    fflush (stdout);

    lsForeachItem(nodeList, nodeGen, nodePtr) {
        printf ("node %d %s\n", nodePtr->nd_id,
            (nodePtr->nd_depth == ncl_ClusterTreeDepth) ? "(leaf)" : "");
        fflush (stdout);

        if (nodePtr->nd_depth == ncl_ClusterTreeDepth) {
            printf ("cell instance name = %s\n", 
                Mac_GetNameOfCellInst(nodePtr->nd_cellInstPtr));
            fflush (stdout);
        }

        if (nodePtr->nd_totalNetList != (lsList) NULL) {
            printf ("  total net list (length = %d)\n",
                lsLength(nodePtr->nd_totalNetList));
            fflush (stdout);
            i = 1;
            lsForeachItem(nodePtr->nd_totalNetList, netGen, netDataPtr) {
                printf ("    %3d) %s\n", i,
                    Mac_GetNameOfNet(netDataPtr->netPtr));
                fflush (stdout);
                i++;
            }
        }

        if (nodePtr->nd_bdrNetList != (lsList) NULL) {
            printf ("  border net list (length = %d)\n",
                lsLength(nodePtr->nd_bdrNetList));
            fflush (stdout);
            i = 1;
            lsForeachItem(nodePtr->nd_bdrNetList, netGen, netDataPtr) {
                printf ("    %3d) %s\n", i,
                Mac_GetNameOfNet(netDataPtr->netPtr));
                fflush (stdout);
                i++;
            }
        }

        printf ("\n");
        fflush (stdout);
    }
    printf ("----------\n\n");
    fflush (stdout);
}


void
NplPrintInitPinListsOfNode(nodePtr)
Ncl_Node *nodePtr;
{
    lsGen       pinGen;
    Npl_PinData  *pinDataPtr;
    int         s;

    NclForeachSide(s) {
        printf ("%s side (length = %d)\n", Ncl_GetSideDirName(s),
            lsLength(nodePtr->nd_initPinList[s]));
        fflush (stdout);

        lsForeachItem(nodePtr->nd_initPinList[s], pinGen, pinDataPtr) {
            printf ("%8s : init pos = %5d %5d  init dir = %s\n",
                pinDataPtr->pinName,
                pinDataPtr->initPos.p_x, pinDataPtr->initPos.p_y,
                Ncl_GetSideDirName(pinDataPtr->initDir));
            fflush (stdout);
        }
    }
    printf ("----------\n\n");
    fflush (stdout);
}


void
NplPrintTilePinListsOfNode(nodePtr)
Ncl_Node *nodePtr;
{
    lsGen       pinGen;
    Npl_PinData  *pinDataPtr;
    int         s;

    NclForeachSide(s) {
        printf ("%s side (length = %d)\n", Ncl_GetSideDirName(s),
            lsLength(nodePtr->nd_tilePinList[s]));
        fflush (stdout);

        lsForeachItem(nodePtr->nd_tilePinList[s], pinGen, pinDataPtr) {
            printf ("%8s : tile pos = %5d %5d  tile dir = %s\n",
                pinDataPtr->pinName,
                pinDataPtr->tilePos.p_x, pinDataPtr->tilePos.p_y,
                Ncl_GetSideDirName(pinDataPtr->tileDir));
            fflush (stdout);
        }
    }
    printf ("----------\n\n");
    fflush (stdout);
}


void
NplCheckCritNetConstraints() 
{
    Ncl_LinkedCritNet   *critNetPtr;
    Npl_NetData         *netDataPtr;
    float               netLength;
    int                 numCritNets=0, numViolatedNets=0;
    Rect                encloseRect;

    for (critNetPtr = ncl_CritNetList;
        critNetPtr != NIL(Ncl_LinkedCritNet);
        critNetPtr = critNetPtr->nextPtr) {

        numCritNets++;
        netDataPtr = (Npl_NetData *) critNetPtr->netPtr->clientData;

        if (lsLength(netDataPtr->pinList) <= MAX_PINS_ON_NET) {
            netLength = NplFindMST(netDataPtr->pinList);
        } else {
            (void) NplCalcEncloseRect(netDataPtr, &encloseRect);
            netLength = (encloseRect.r_xtop - encloseRect.r_xbot +
                      encloseRect.r_ytop - encloseRect.r_ybot);
        }

        if (netLength > critNetPtr->netPtr->maxLength) {
            numViolatedNets++;
            (void) printf ("Critical net %s has constraint violation :\n",
                  Mac_GetNameOfNet(critNetPtr->netPtr));
            (void) printf ("    max length = %d  actual length= %.4f\n",
                  critNetPtr->netPtr->maxLength, netLength);
        }
    }

    if (numViolatedNets > 0) {
        (void) sprintf (bear_Msg,
            "Timing : %d out of %d upper bound constraints violated.\n", 
            numViolatedNets, numCritNets);
    } else {
        (void) sprintf (bear_Msg,
            "Timing : all upper bound constraints satisfied.\n");
    }

    (void) fprintf (bearout, bear_Msg);
    fflush (bearout);

    PbPrint(EZ_NORMAL_FONT, bear_Msg);
}


