
/* nplSpacing.c --
 *
 *     Massoud Pedram
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/newP/nplSpacing.c,v 1.1.1.1 2002/12/31 15:51:08 saj Exp $";
#endif not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "globalSpacer.h"
#include "list.h"
#include "newP.h"
#include "newPInt.h"
#include "nplWriteToMac.h"
#include "nplDraw.h"


/*
 * ----------------------------------------------------------------------------
 *
 * NplDoSpacing() --
 *
 *   Removes overlaps by moving cells apart from each other. 
 *   Computes the final position of all blocks. Does some compaction
 *   within clusters of blocks, but none between clusters.
 *   The queue contains internal nodes of the cluster where the deepest node is
 *   on the front of the queue and the root is on the rear of queue.
 *
 * Results:
 *   Assigns final values to the transformation fields of ClusterTree
 *   nodes. Also assigns values to nd_dim fields of non-leaf ClusterTree
 *   nodes.  The procedure does not change the queue.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
NplDoSpacing(nodeQueue)
Deque *nodeQueue;
{
    Deque       *tempQueue = MakeDeque();
    Ncl_Node    *nodePtr, *childNodePtr;
    int         i;
    double      t1;

    t1 = NplPtime();

    ncl_ClusterTree->nd_nodeTrans.t_c = 0;
    ncl_ClusterTree->nd_nodeTrans.t_f = 0;

    while (! IsEmptyDeque(nodeQueue)) {
        /* nodeQueue: front = ..., 2, 1, 0  = rear */

        /* top-down traversal of the cluster tree */
        nodePtr = (Ncl_Node *) EjectDeque(nodeQueue);

        for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
            childNodePtr = nodePtr->nd_children[i];

            childNodePtr->nd_nodeTrans.t_c = 0;
            childNodePtr->nd_nodeTrans.t_f = 0;
        }

        PushDeque(tempQueue, (ClientData) nodePtr);
    }

    while (! IsEmptyDeque(tempQueue)) {
        /* tempQueue: front = ..., 2, 1, 0  = rear */

        /* bottom-up traversal of the cluster tree */
        nodePtr = (Ncl_Node *) PopDeque(tempQueue);

        if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        PushDeque(nodeQueue, (ClientData) nodePtr);
    }

    while (! IsEmptyDeque(nodeQueue)) {
        /* nodeQueue: front = 0, 1, 2, ... = rear */

        /* top-down traversal of the cluster tree */
        nodePtr = (Ncl_Node *) PopDeque(nodeQueue);

        if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }

        PushDeque(tempQueue, (ClientData) nodePtr);
    }

    if (NplReformatCell(ncl_ClusterTree) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    while (! IsEmptyDeque(tempQueue)) {
        /* tempQueue: front = ..., 2, 1, 0  = rear */

        /* top-down traversal of the cluster tree */
        nodePtr = (Ncl_Node *) EjectDeque(tempQueue);

        for (i = 0; i < nodePtr->nd_numOfChildren ;i++) {
            if (NplReformatCell(nodePtr->nd_children[i]) == BEAR_ERROR) {
                return (BEAR_ERROR);
            }
        }

        PushDeque(nodeQueue, (ClientData) nodePtr);
    }

    DeleteDeque(tempQueue);
    /* nodeQueue: front = ..., 2, 1, 0  = rear */

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        NplOutlineNodeList(nplTotalNodeList);
    }
#endif BEAR_BLACK_BOX

    fprintf (stdout, "CPU-time (Spacing) = %g sec\n", (NplPtime() - t1));
    fflush(stdout);

    return (BEAR_OK);
}
  

#ifdef BETTY_BOOP
/*
 * (gyorgy 18 dec 90) NplCompactNodeAfterGR() is not currently needed.
 */
BearStatus
NplCompactNodeAfterGR(nodePtr)
Ncl_Node *nodePtr;
{
    CellMaster  *cellMasterPtr;
    Ncl_Node    *childNodePtr;
    int         i;

    cellMasterPtr = Mac_GetMasterOfCell(nodePtr->nd_cellInstPtr);
    if (cellMasterPtr == NIL(CellMaster)) {
        Bear_PrintErr("NplCompactNodeAfterGR() failed: cellMasterPtr is nil.");
        return (BEAR_ERROR);
    }

    if (Gs_CalcCapDenOfTilePlane(cellMasterPtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    nodePtr->nd_storeDim = nodePtr->nd_dim;
    nodePtr->nd_storeTrans = nodePtr->nd_nodeTrans;

    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        childNodePtr = nodePtr->nd_children[i];

        if (childNodePtr->nd_status == PLACED) {
            childNodePtr->nd_storeDim = childNodePtr->nd_dim;
            childNodePtr->nd_storeTrans = childNodePtr->nd_nodeTrans;
        }
        childNodePtr->nd_nodeTrans.t_c = 0;
        childNodePtr->nd_nodeTrans.t_f = 0;
    }

    if (NplRecalcSpace(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_DimOfNode(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (Npl_TransOfChildren(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    if (NplReformatCell(nodePtr) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }
    for (i = 0; i < nodePtr->nd_numOfChildren; i++) {
        if (NplReformatCell(nodePtr->nd_children[i]) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }
    if (Mac_SetUpTiPlane(cellMasterPtr, FALSE) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }
    if (Fpg_GenRooms(cellMasterPtr) == BEAR_ERROR) {
       Bear_PrintErr("NplCompactNodeAfterGR() : could not generate rooms.");
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}
#endif BETTY_BOOP


