
/* writeToOct.c --
 *      
 *     Denis Lee
 *     George Carvalho, Esq. -- conversion to Octtools 5.0 and cleanup
 *
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1988 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/octInterface/writeToOct.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "vov.h"
#include "vulcan.h"
#include "octInterface.h"
#include "oh.h"
#include "octInterfaceInt.h"


/* ---------- Local Procedures ---------- */
static BearStatus   WriteCellsToOct();
static BearStatus   WriteIOPadsToOct();
static BearStatus   WritePinsToOct();
static BearStatus   WriteNetPinConnectionsToOct();


/*
 * ----------------------------------------------------------------------------
 *
 * WritePinsToOct() --
 *
 * This procedure is designed to be used by an "interator".
 *
 * Writes each pin of a cell from MAC to OCT.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
WritePinsToOct(pinPtr, cellPtr)
PinInst     *pinPtr;
octObject   *cellPtr;
{
    octObject   layers;
    octObject   properties;
    octObject   terminal;
    octObject   termBox;
    Point       *coordinate;

    /* Create terminal (pin). */
    if (ohCreateTerm(cellPtr, &terminal, 
        (char *) Mac_GetNameOfPinInst(pinPtr)) != OCT_OK) {
        octError("WritePinsToOct(): cannot create term");
/*
, %s.", terminal.contents.term.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    /* Get coordinates of pin. */
    if (Mac_IsPinPosDef(pinPtr) == FALSE) {
        Bear_PrintErr("WritePinsToOct(): a pin's position is undefined.");
        return (BEAR_STOP_SEARCH);
    }
    coordinate = Mac_GetRelPosOfPin(pinPtr);

/*
size =
    if (Mac_GetKindOfPort(Mac_GetPortOfPin(pinPtr)) == SIGNAL_PORT) {
    } else {
    }
fprintf(stderr,"pin width = %.1f, segment width = %.1f, hwidth = %d, widsr = %d.\n",pinPtr->pinWidth,pinPtr->segmWidth,pinPtr->BBLTermPtr->termPtr->rsp->hwidth,pinPtr->BBLTermPtr->termPtr->rsp->widsr);
*/
    /* create box representing the terminal (pin) */
    termBox.type = OCT_BOX;
    switch (Mac_GetDirectionOfPin(pinPtr)) {
        case NORTH_PIN:
            termBox.contents.box.lowerLeft.x = coordinate->p_x - 2;
            termBox.contents.box.lowerLeft.y = coordinate->p_y - 4;
            termBox.contents.box.upperRight.x = coordinate->p_x + 2;
            termBox.contents.box.upperRight.y = coordinate->p_y;
            break;
        case SOUTH_PIN:
            termBox.contents.box.lowerLeft.x = coordinate->p_x - 2;
            termBox.contents.box.lowerLeft.y = coordinate->p_y;
            termBox.contents.box.upperRight.x = coordinate->p_x + 2;
            termBox.contents.box.upperRight.y = coordinate->p_y + 4;
            break;
        case WEST_PIN:
            termBox.contents.box.lowerLeft.x = coordinate->p_x;
            termBox.contents.box.lowerLeft.y = coordinate->p_y - 2;
            termBox.contents.box.upperRight.x = coordinate->p_x + 4;
            termBox.contents.box.upperRight.y = coordinate->p_y + 2;
            break;
        case EAST_PIN:
            termBox.contents.box.lowerLeft.x = coordinate->p_x - 4;
            termBox.contents.box.lowerLeft.y = coordinate->p_y - 2;
            termBox.contents.box.upperRight.x = coordinate->p_x;
            termBox.contents.box.upperRight.y = coordinate->p_y + 2;
            break;
        default:
            Bear_PrintErr("WritePinsToOct(): illegal pin direction.");
            return (BEAR_STOP_SEARCH);
    }
    /* convert coordinates using OCT_PER_LAMBDA */
    termBox.contents.box.lowerLeft.x *= OCT_PER_LAMBDA;
    termBox.contents.box.lowerLeft.y *= OCT_PER_LAMBDA;
    termBox.contents.box.upperRight.x *= OCT_PER_LAMBDA;
    termBox.contents.box.upperRight.y *= OCT_PER_LAMBDA;
    if (octCreate(&terminal, &termBox) != OCT_OK) {
        octError("WritePinsToOct(): cannot create term box.");
/*
, %s.", terminal.contents.term.name);
*/
        return (BEAR_STOP_SEARCH);
    }
    /* the put layers of the terminal (pin) */
    layers.type = OCT_LAYER;
    layers.contents.layer.name = "MET1";
    if (octGetByName(cellPtr, &layers) != OCT_OK) {
        octError("WritePinsToOct(): cannot get MET1 layer.");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&layers,&termBox) != OCT_OK) {
        octError("WritePinsToOct(): cannot attach term to MET1 layer.");
        return (BEAR_STOP_SEARCH);
    }
    layers.contents.layer.name = "MET2";
    if (octGetByName(cellPtr, &layers) != OCT_OK) {
        octError("WritePinsToOct(): cannot get MET2 layer.");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&layers,&termBox) != OCT_OK) {
        octError("WritePinsToOct(): cannot attach term to MET2 layer.");
        return (BEAR_STOP_SEARCH);
    }

    /* create terminal's (pin's) properties */
    properties.type = OCT_PROP;
/*
    properties.contents.prop.type = OCT_STRING;
    properties.contents.prop.name = "MOBILITY";
    properties.contents.prop.name = "TERM_EDGE";
    properties.contents.prop.type = OCT_REAL;
    properties.contents.prop.name = "TERM_RELATIVE_POSITION";
*/
    /* put direction property */
    properties.contents.prop.type = OCT_STRING;
    properties.contents.prop.name = "DIRECTION";
    switch (Mac_GetDirectionOfPort(Mac_GetPortOfPin(pinPtr))) {
        case INPUT:
            properties.contents.prop.value.string = "INPUT";
            break;
        case OUTPUT:
            properties.contents.prop.value.string = "OUTPUT";
            break;
        case INOUT:
            properties.contents.prop.value.string = "INOUT";
            break;
        default:
            Bear_PrintErr("WritePinsToOct(): illegal port direction.");
            return (BEAR_STOP_SEARCH);
    }
    if (octCreate(&terminal, &properties) != OCT_OK) {
        octError("WritePinsToOct(): cannot create DIRECTION property.");
        return (BEAR_STOP_SEARCH);
    }
    /* put terminal (pin) type property */
    switch (Mac_GetKindOfNet(Mac_GetNetOfPin(pinPtr))) {
        case SIGNAL_NET:
            properties.contents.prop.value.string = "SIGNAL";
            break;
        case POWER_NET:
            /* put peak current property for power */
            properties.contents.prop.type = OCT_REAL;
            properties.contents.prop.name = "PEAKCURRENT";
            properties.contents.prop.value.real = 
                Mac_GetMaxCurrentReqOfPort(Mac_GetPortOfPin(pinPtr));
            if (octCreate(&terminal, &properties) != OCT_OK) {
             octError("WritePinsToOct(): cannot create PEAKCURRENT property.");
                return (BEAR_STOP_SEARCH);
            }
            properties.contents.prop.type = OCT_STRING;
            properties.contents.prop.value.string = "SUPPLY";
            break;
        case GROUND_NET:
            /* put peak current property for ground */
            properties.contents.prop.type = OCT_REAL;
            properties.contents.prop.name = "PEAKCURRENT";
            properties.contents.prop.value.real = 
                Mac_GetMaxCurrentReqOfPort(Mac_GetPortOfPin(pinPtr));
            if (octCreate(&terminal, &properties) != OCT_OK) {
             octError("WritePinsToOct(): cannot create PEAKCURRENT property.");
                return (BEAR_STOP_SEARCH);
            }
            properties.contents.prop.type = OCT_STRING;
            properties.contents.prop.value.string = "GROUND";
            break;
        case CLOCK_NET:
            properties.contents.prop.value.string = "CLOCK";
            break;
        default:
            Bear_PrintErr("WritePinsToOct(): illegal kind of net.");
            return (BEAR_STOP_SEARCH);
    }
    properties.contents.prop.name = "TERMTYPE";
    if (octCreate(&terminal, &properties) != OCT_OK) {
        octError("WritePinsToOct(): cannot create TERMTYPE property.");
        return (BEAR_STOP_SEARCH);
    }
/*
 *  properties.contents.prop.name = "PLACEMENT_CLASS";
 *  properties.contents.prop.value.string = "PAD";
 *  if (octCreate(&terminst, &properties) != OCT_OK) {
 *      octError("WritePinsToOct(): cannot create PLACEMENT_CLASS property.");
 *  }
 *  properties.contents.prop.name = "MOBILITY";
 *  properties.contents.prop.value.string = "";
 *  if (octCreate(&terminst, &properties) != OCT_OK) {
 *      octError("WritePinsToOct(): cannot create MOBILITY property.");
 *  }
 */

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * WriteNetPinConnectionsToOct() --
 *
 * This procedure is designed to be used by an "interator".
 *
 * Connect each pin to its net from MAC to OCT.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
WriteNetPinConnectionsToOct(pinPtr, instance)
PinInst     *pinPtr;
octObject   *instance;
{
    octObject chip;
    octObject net;
    octObject properties;
    octObject terminal;

/*
pass cell
and get chip with octGenFirstContainer
    instance.type = OCT_INSTANCE;
    instance.contents.instance.name = 
        (char *) Mac_GetNameOfCellInst(Mac_GetCellInstOfPort(
                                               Mac_GetPortOfPin(pinPtr)));
    if (octGetByName(chip, &instance) == OCT_NOT_FOUND) {
        octError("WriteNetPinConnectionsToOct(): cannot get instance, %s.",
            instance.contents.instance.name);
    }
*/
    /* get the chip facet */
    if (octGenFirstContainer(instance, OCT_FACET_MASK, &chip) != OCT_OK) {
        octError("WriteNetPinConnectionsToOct(): cannot generate first FACET container of instance.");
/*
, %s.", instance->contents.instance.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    /* Get OCT's terminal representation of MAC's pin. */
    if (ohGetByTermName(instance, &terminal,
        (char *) Mac_GetNameOfPinInst(pinPtr)) == OCT_NOT_FOUND) {
        octError("WriteNetPinConnectionsToOct(): cannot get term.");
/*
, %s.", terminal.contents.term.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    /* Get the net and attach the terminal to the net. */
    if (ohGetOrCreateNet(&chip, &net, 
        (char *) Mac_GetNetNameOfPin(pinPtr)) != OCT_OK) {
        octError("WriteNetPinConnectionsToOct(): cannot get or create net");
/*
, %s.", net.contents.net.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    if (octAttach(&net, &terminal) != OCT_OK) {
        octError("WriteNetPinConnectionsToOct(): cannot attach term.");
/*
, %s.", terminal.contents.term.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    /* create the net's properties */
    properties.type = OCT_PROP;
/*
    properties.contents.prop.type = OCT_INTEGER;
    properties.contents.prop.name = "NET_MAX_LENGTH";
*/
    /* put net type property */
    properties.contents.prop.type = OCT_STRING;
    switch (Mac_GetKindOfNet(Mac_GetNetOfPin(pinPtr))) {
        case SIGNAL_NET:
            properties.contents.prop.value.string = "SIGNAL";
            break;
        case POWER_NET:
            properties.contents.prop.value.string = "SUPPLY";
            break;
        case GROUND_NET:
            properties.contents.prop.value.string = "GROUND";
            break;
        case CLOCK_NET:
            properties.contents.prop.value.string = "CLOCK";
            break;
        default:
          Bear_PrintErr("WriteNetPinConnectionsToOct(): illegal kind of net.");
            break;
    }
    properties.contents.prop.name = "NETTYPE";
    if (octGetOrCreate(&net, &properties) != OCT_OK) {
   octError("WriteNetPinConnectionsToOct(): cannot create NETTYPE property.");
        return (BEAR_STOP_SEARCH);
    }
/*
    properties.contents.prop.type = OCT_INTEGER;
    properties.contents.prop.name = "NETCRITICALITY";
    properties.contents.prop.value.integer = 
        Mac_GetCriticalityOfNet(Mac_GetNetOfPin(pinPtr));
    if (octGetOrCreate(&net, &properties) != OCT_OK) {
        octError("WriteNetPinConnectionsToOct(): cannot create NETCRITICALITY property.");
    }
*/

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * WriteCellsToOct() --
 *
 * This procedure is designed to be used by an "interator".
 *
 * Writes each cell of a chip design from MAC to OCT.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
WriteCellsToOct(cellInstPtr, chip)
CellInst    *cellInstPtr;
octObject   *chip;
{
    octObject   cell;
    octObject   properties;
    octObject   bags;
    octObject   layers;
    octObject   cellBox;
    octObject   interface;
    octObject   instance;
    Rect        *cellBBoxPtr;

    /* create contents facet for the cell */
    cell.type = OCT_FACET;
    cell.contents.facet.cell = cellInstPtr->name;
    cell.contents.facet.view = "physical";
    cell.contents.facet.version = OCT_CURRENT_VERSION;
    cell.contents.facet.facet = "contents";
    cell.contents.facet.mode = "w";

    (void) VOVoutputFacet(&cell);

    if (octOpenFacet(&cell) < OCT_OK) {
        octError("WriteCellsToOct(): cannot open cell.");
/*
, %s.", cellInstPtr->name);
*/
        return (BEAR_STOP_SEARCH);
    }
    /* create cell's properties */
    properties.type = OCT_PROP;
    /* put edit style, view type, and technology properties */
    properties.contents.prop.type = OCT_STRING;
    properties.contents.prop.name = "EDITSTYLE";
    properties.contents.prop.value.string = "PHYSICAL";
    if (octCreate(&cell, &properties) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create EDITSTYLE property.");
        return (BEAR_STOP_SEARCH);
    }
    properties.contents.prop.name = "VIEWTYPE";
    if (octCreate(&cell, &properties) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create VIEWTYPE property.");
        return (BEAR_STOP_SEARCH);
    }
    properties.contents.prop.name = "TECHNOLOGY";
    properties.contents.prop.value.string = "scmos";
    if (octCreate(&cell, &properties) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create TECHNOLOGY property.");
        return (BEAR_STOP_SEARCH);
    }
/*
    properties.contents.prop.name = "CELLCLASS";
    properties.contents.prop.value.string = "MODULE";
    if (octCreate(&cell, &properties) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create CELLCLASS property.");
    }
*/
    /* create joined terminals (pins) bag for the cell */
    bags.type = OCT_BAG;
    bags.contents.bag.name = "JOINEDTERMS";
    if (octCreate(&cell, &bags) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create JOINEDTERMS bag.");
        return (BEAR_STOP_SEARCH);
    }
/*
    bags.contents.bag.name = "INSTANCES";
    if (octCreate(&cell, &bags) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create INSTANCES bag.");
    }
    bags.contents.bag.name = "CONNECTORS";
    if (octCreate(&cell, &bags) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create CONNECTORS bag.");
    }
*/
    /* create available layers for the cell */
    layers.type = OCT_LAYER;
    layers.contents.layer.name = "MET1";
    if (octCreate(&cell,&layers) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create MET1 layer.");
        return (BEAR_STOP_SEARCH);
    }
    layers.contents.layer.name = "MET2";
    if (octCreate(&cell,&layers) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create MET2 layer.");
        return (BEAR_STOP_SEARCH);
    }
    layers.contents.layer.name = "PLACE";
    if (octCreate(&cell,&layers) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create PLACE layer.");
        return (BEAR_STOP_SEARCH);
    }
    /* create box representing the cell */
    cellBBoxPtr = Mac_GetBBoxOfCell(cellInstPtr);
    cellBox.type = OCT_BOX;
    cellBox.contents.box.lowerLeft.x = cellBBoxPtr->r_xbot * OCT_PER_LAMBDA;
    cellBox.contents.box.lowerLeft.y = cellBBoxPtr->r_ybot * OCT_PER_LAMBDA;
    cellBox.contents.box.upperRight.x = cellBBoxPtr->r_xtop * OCT_PER_LAMBDA;
    cellBox.contents.box.upperRight.y = cellBBoxPtr->r_ytop * OCT_PER_LAMBDA;
    if (octCreate(&layers,&cellBox) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create cell box on PLACE layer.");
        return (BEAR_STOP_SEARCH);
    }
/*
    if (octAttach(&cell,&cellBox) != OCT_OK) {
        octError("WriteCellsToOct(): cannot attach on PLACE layer.");
    }
*/

    /* Go through and create all pins of the cell in OCT. */
    if (Mac_ItrPinOfCellInst(cellInstPtr, WritePinsToOct, 
        (ClientData) &cell) == BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }

    /* flush data to disk (just in case) */
    if (octFlushFacet(&cell) < OCT_OK) {
        octError("WriteCellsToOct(): cannot save cell by flushing to disk.");
        return (BEAR_STOP_SEARCH);
    }
    
    /* create interface facet for the cell */
    interface.type = OCT_FACET;
    interface.contents.facet.cell = cell.contents.facet.cell;
    interface.contents.facet.view = cell.contents.facet.view;
    interface.contents.facet.version = OCT_CURRENT_VERSION;
    interface.contents.facet.facet = "interface";
    interface.contents.facet.mode = "w";

    (void) VOVoutputFacet(&interface);

    if (octOpenFacet(&interface) < OCT_OK) {
        octError("WriteCellsToOct(): cannot open cell interface.");
        return (BEAR_STOP_SEARCH);
    }
    
    vulcan(&cell, &interface);

    /* create the instance of the cell on the chip facet */
    instance.type = OCT_INSTANCE;
    instance.contents.instance.facet = "contents";
    instance.contents.instance.name = cell.contents.facet.cell;
    instance.contents.instance.master = cell.contents.facet.cell;
    instance.contents.instance.view = cell.contents.facet.view;
    instance.contents.instance.version = OCT_CURRENT_VERSION;
    instance.contents.instance.transform.translation.x 
        = cellInstPtr->transform.t_c * OCT_PER_LAMBDA;
    instance.contents.instance.transform.translation.y 
        = cellInstPtr->transform.t_f * OCT_PER_LAMBDA;
    instance.contents.instance.transform.transformType 
        = TransMagicToOct(cellInstPtr->transform);
    if (octCreate(chip, &instance) != OCT_OK) {
        octError("WriteCellsToOct(): cannot create instance.");
/*
, %s.", instance.contents.instance.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    /* go through and connect all the pins
       of the cell to their respective nets */
    if (Mac_ItrPinOfCellInst(cellInstPtr, WriteNetPinConnectionsToOct,
        (ClientData) &instance) == BEAR_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }

    /* get the INSTANCES and CORECELLS bags
       and put the cell instance into the bags */
    bags.contents.bag.name = "INSTANCES";
    if (octGetByName(chip, &bags) == OCT_NOT_FOUND) {
        octError("WriteCellsToOct(): cannot get INSTANCES bag.");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&bags,&instance) != OCT_OK) {
        octError("WriteCellsToOct(): cannot attach instance to INSTANCES bag.");
        return (BEAR_STOP_SEARCH);
    }
    bags.contents.bag.name = "CORECELLS";
    if (octGetByName(chip, &bags) == OCT_NOT_FOUND) {
        octError("WriteCellsToOct(): cannot get CORECELLS bag.");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&bags,&instance) != OCT_OK) {
        octError("WriteCellsToOct(): cannot attach instance to CORECELLS bag.");
        return (BEAR_STOP_SEARCH);
    }

    /* close contents and interface facets of the cell */
    if (octCloseFacet(&cell) < OCT_OK) {
        octError("WriteCellsToOct(): cannot close cell.");
/*
, %s.", cellInstPtr->name);
*/
        return (BEAR_STOP_SEARCH);
    }
    if (octCloseFacet(&interface) < OCT_OK) {
        octError("WriteCellsToOct(): cannot close cell interface.");
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * WriteIOPadsToOct() --
 *
 * This procedure is designed to be used by an "interator".
 *
 * Writes each I/O pad of the chip design from MAC to OCT
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH, if successful.
 *   BEAR_STOP_SEARCH, otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static BearStatus
WriteIOPadsToOct(pinPtr, chipPtr)
PinInst     *pinPtr;
octObject   *chipPtr;
{
    octObject   bags;
    octObject   net;
    octObject   properties;
    octObject   terminal, termInst, terminalt;
    Point       coordinate;
    int         size;
    char        fullpadname[512];

    /* Create (i/o pad) terminal. */
    if (ohCreateTerm(chipPtr, &terminal,
        (char *) Mac_GetNameOfPinInst(pinPtr)) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create terminal");
/*
, %s.", terminal.contents.term.name);
*/
        return (BEAR_STOP_SEARCH);
    }
    /* create the instance of the (i/o pad) terminal */
    termInst.type = OCT_INSTANCE;
    termInst.contents.instance.facet = "contents";
    termInst.contents.instance.name = terminal.contents.term.name;
/*
primitives
    instance, formal term
*/
/*
    termInst.contents.instance.master 
        = /cad/lib/technology/scmos/pads/_2uFamily;
    termInst.contents.instance.view =  _2u;
*/
/*
fprintf(stderr,"pin width = %.1f, segment width = %.1f, hwidth = %d, widsr = %d.\n",pinPtr->pinWidth,pinPtr->segmWidth,pinPtr->BBLTermPtr->termPtr->rsp->hwidth,pinPtr->BBLTermPtr->termPtr->rsp->widsr);
*/

    /* 
     * Use a fixed standard (i/o pad) terminal size because of technology. 
     */
    if (Mac_GetKindOfPort(Mac_GetPortOfPin(pinPtr)) == SIGNAL_PORT) {
        /* terminal and size for signals */
        (void) sprintf (fullpadname, 
            "$OCTTOOLS/tech/scmos/contacts/com2.4x4");
        termInst.contents.instance.master = fullpadname;
        size = 2;
    } 
    else {
        /* terminal and size for others like power and ground */
        (void) sprintf (fullpadname,
            "$OCTTOOLS/tech/scmos/contacts/com2.30x30");
        termInst.contents.instance.master = fullpadname;
        size = 15;
    }
    termInst.contents.instance.view = "physical";
    termInst.contents.instance.version = OCT_CURRENT_VERSION;

    /* Get coordinates of the (i/o pad) terminal. */
    if (Mac_IsPinPosDef(pinPtr) == FALSE) {
        Bear_PrintErr("WriteIOPadsToOct(): a pin's position is undefined.");
        return (BEAR_STOP_SEARCH);
    }
    Mac_GetAbsPosOfPin(pinPtr, &coordinate);

    /* convert coordinates using OCT_PER_LAMBDA */
    termInst.contents.instance.transform.translation.x = 
        coordinate.p_x * OCT_PER_LAMBDA;
    termInst.contents.instance.transform.translation.y = 
        coordinate.p_y * OCT_PER_LAMBDA;

    /* center the (i/o pad) terminal */
    switch (Mac_GetDirectionOfPin(pinPtr)) {
        case NORTH_PIN:
            termInst.contents.instance.transform.translation.y -= size;
            break;
        case SOUTH_PIN:
            termInst.contents.instance.transform.translation.y += size;
            break;
        case WEST_PIN:
            termInst.contents.instance.transform.translation.x += size;
            break;
        case EAST_PIN:
            termInst.contents.instance.transform.translation.x -= size;
            break;
        default:
            Bear_PrintErr("WriteIOPadsToOct(): illegal pin direction.");
            return (BEAR_STOP_SEARCH);
    }
    termInst.contents.instance.transform.transformType = OCT_NO_TRANSFORM;
    if (octCreate(chipPtr, &termInst) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create termInst");
/*
, %s.", termInst.contents.instance.name);
*/
        return (BEAR_STOP_SEARCH);
    }
    /* Get and attach psuedo terminal to the (i/o pad) terminal. */
    if (ohGetByTermName(&termInst, &terminalt,
        "t") == OCT_NOT_FOUND) {
        octError("WriteIOPadsToOct(): cannot get term named t");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&terminal, &terminalt) != OCT_OK) {
     octError("WriteIOPadsToOct(): cannot attach term named t, of termInst.");
        return (BEAR_STOP_SEARCH);
    }

    /* Get or create the net and attach the (i/o pad) terminal to the net. */
    if (ohGetOrCreateNet(chipPtr, &net, 
        (char *) Mac_GetNetNameOfPin(pinPtr)) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot get or create net");
/*
, %s.", net.contents.net.name);
*/
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&net, &terminal) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot attach term.");
/*
, %s.", terminal.contents.term.name);
*/
        return (BEAR_STOP_SEARCH);
    }

    /* create (i/o pad) terminal's properties */
    properties.type = OCT_PROP;
/*
 *  properties.contents.prop.type = OCT_STRING;
 *  properties.contents.prop.name = "MOBILITY";
 *  properties.contents.prop.name = "TERM_EDGE";
 *  properties.contents.prop.type = OCT_REAL;
 *  properties.contents.prop.name = "TERM_RELATIVE_POSITION";
 *
 *  properties.contents.prop.type = OCT_INTEGER;
 *  properties.contents.prop.name = "NET_MAX_LENGTH";
 */
    /* put direction property */
    properties.contents.prop.type = OCT_STRING;
    properties.contents.prop.name = "DIRECTION";
    switch (Mac_GetDirectionOfPort(Mac_GetPortOfPin(pinPtr))) {
        case INPUT:
            properties.contents.prop.value.string = "INPUT";
            break;
        case OUTPUT:
            properties.contents.prop.value.string = "OUTPUT";
            break;
        case INOUT:
            properties.contents.prop.value.string = "INOUT";
            break;
        default:
            Bear_PrintErr("WriteIOPadsToOct(): illegal port direction.");
            return (BEAR_STOP_SEARCH);
    }
    if (octCreate(&terminal, &properties) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create DIRECTION property.");
        return (BEAR_STOP_SEARCH);
    }
    /* put (i/o pad) terminal type property */
    switch (Mac_GetKindOfNet(Mac_GetNetOfPin(pinPtr))) {
        case SIGNAL_NET:
            properties.contents.prop.value.string = "SIGNAL";
            break;
        case POWER_NET:
            /* put peak current property for power */
            properties.contents.prop.type = OCT_REAL;
            properties.contents.prop.name = "PEAKCURRENT";
            properties.contents.prop.value.real = 
                Mac_GetMaxCurrentReqOfPort(Mac_GetPortOfPin(pinPtr));
            if (octCreate(&terminal, &properties) != OCT_OK) {
          octError("WriteIOPadsToOct(): cannot create PEAKCURRENT property.");
                return (BEAR_STOP_SEARCH);
            }
            properties.contents.prop.type = OCT_STRING;
            properties.contents.prop.value.string = "SUPPLY";
            break;
        case GROUND_NET:
            /* put peak current property for ground */
            properties.contents.prop.type = OCT_REAL;
            properties.contents.prop.name = "PEAKCURRENT";
            properties.contents.prop.value.real = 
                Mac_GetMaxCurrentReqOfPort(Mac_GetPortOfPin(pinPtr));
            if (octCreate(&terminal, &properties) != OCT_OK) {
          octError("WriteIOPadsToOct(): cannot create PEAKCURRENT property.");
                return (BEAR_STOP_SEARCH);
            }
            properties.contents.prop.type = OCT_STRING;
            properties.contents.prop.value.string = "GROUND";
            break;
        case CLOCK_NET:
            properties.contents.prop.value.string = "CLOCK";
            break;
        default:
            Bear_PrintErr("WriteIOPadsToOct(): illegal kind of net.");
            return (BEAR_STOP_SEARCH);
    }
    properties.contents.prop.name = "TERMTYPE";
    if (octCreate(&terminal, &properties) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create TERMTYPE property.");
        return (BEAR_STOP_SEARCH);
    }
    /* also put net's type property */
    properties.contents.prop.name = "NETTYPE";
    if (octGetOrCreate(&net, &properties) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create NETTYPE property.");
        return (BEAR_STOP_SEARCH);
    }
    /* and put (i/o pad) terminal instance's placement class property */
    properties.contents.prop.name = "PLACEMENT_CLASS";
    properties.contents.prop.value.string = "PAD";
    if (octCreate(&termInst, &properties) != OCT_OK) {
      octError("WriteIOPadsToOct(): cannot create PLACEMENT_CLASS property.");
        return (BEAR_STOP_SEARCH);
    }
/*
    properties.contents.prop.name = "MOBILITY";
    properties.contents.prop.value.string = "";
    if (octCreate(&termInst, &properties) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create MOBILITY property.");
    }
    properties.contents.prop.type = OCT_INTEGER;
    properties.contents.prop.name = "NETCRITICALITY";
    properties.contents.prop.value.integer = Mac_GetCriticalityOfNet(
                                             Mac_GetNetOfPin(pinPtr));
    if (octGetOrCreate(&net, &properties) != OCT_OK) {
        octError("WriteIOPadsToOct(): cannot create NETCRITICALITY property.");
    }
*/
/*
name or create term
attach and move term to edge
attach net
properties net_max_length
*/
    /* 
     * Get CONNECTORS and IOPADS bags and put the (i/o pad) terminal 
     * instance into the bags.
     */
    bags.type = OCT_BAG;
    bags.contents.bag.name = "CONNECTORS";
    if (octGetByName(chipPtr, &bags) == OCT_NOT_FOUND) {
        octError("WriteIOPadsToOct(): cannot get CONNECTORS bag.");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&bags,&termInst) != OCT_OK) {
octError("WriteIOPadsToOct(): cannot attach term instance to CONNECTORS bag.");
        return (BEAR_STOP_SEARCH);
    }
    bags.contents.bag.name = "IOPADS";
    if (octGetByName(chipPtr, &bags) == OCT_NOT_FOUND) {
        octError("WriteIOPadsToOct(): cannot get IOPADS bag.");
        return (BEAR_STOP_SEARCH);
    }
    if (octAttach(&bags,&termInst) != OCT_OK) {
   octError("WriteIOPadsToOct(): cannot attach term instance to IOPADS bag.");
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * WriteToOct() - iterate through all the children of the
cell name : physical (contents, interface)
 */
/*
 * Odi_WriteToOct() --
 *
 * Iterate through all the children of the
 * chip core master while calling WriteCellsToOct()
 * for each child to write out all the cells in OCT
 * and also iterate through all the nets while calling
 * WriteRoutingOfNetsToOct() for each net to write out
 * the routing information if routing is specified.
 *
 */

BearStatus
Odi_WriteToOct(cellMasterPtr, cell, view, routingFlag)
CellMaster  *cellMasterPtr;
char        *cell, *view;
bool        routingFlag;
{
    octObject   chip;
    octObject   bags;
    octObject   interface;
    octObject   layers;
    octObject   properties;
    
    if (Mac_GetPlacedFlagOfCell(cellMasterPtr) == FALSE) {
        Bear_PrintErr("Odi_WriteToOct(): cells are not placed.");
        return (BEAR_ERROR);
    }

    /* create the chip facet */
    chip.type = OCT_FACET;
    chip.contents.facet.cell = cell;
    chip.contents.facet.view = view;
    chip.contents.facet.version = OCT_CURRENT_VERSION;
    chip.contents.facet.facet = "contents";
    chip.contents.facet.mode = "w";

    (void) VOVoutputFacet(&chip);

    if (octOpenFacet(&chip) < OCT_OK) {
        octError("Odi_WriteToOct(): cannot open chip.");
        return (BEAR_ERROR);
    }
    /* create the chip's properties */
    properties.type = OCT_PROP;
    /* put edit style, view type, technology, and cell class properties */
    properties.contents.prop.type = OCT_STRING;
    properties.contents.prop.name = "EDITSTYLE";
    properties.contents.prop.value.string = "SYMBOLIC";
    if (octCreate(&chip, &properties) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create EDITSTYLE property.");
        return (BEAR_ERROR);
    }
    properties.contents.prop.name = "VIEWTYPE";
    if (octCreate(&chip, &properties) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create VIEWTYPE property.");
        return (BEAR_ERROR);
    }
    properties.contents.prop.name = "TECHNOLOGY";
    properties.contents.prop.value.string = "scmos";
    if (octCreate(&chip, &properties) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create TECHNOLOGY property.");
        return (BEAR_ERROR);
    }
    properties.contents.prop.name = "CELLCLASS";
    properties.contents.prop.value.string = "MODULE";
    if (octCreate(&chip, &properties) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create CELLCLASS property.");
        return (BEAR_ERROR);
    }
    /* create the INSTANCES, CONNECTORS, CORECELLS, and IOPADS bags */
    bags.type = OCT_BAG;
    bags.contents.bag.name = "INSTANCES";
    if (octCreate(&chip, &bags) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create INSTANCES bag.");
        return (BEAR_ERROR);
    }
    bags.contents.bag.name = "CONNECTORS";
    if (octCreate(&chip, &bags) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create CONNECTORS bag.");
        return (BEAR_ERROR);
    }
    bags.contents.bag.name = "CORECELLS";
    if (octCreate(&chip, &bags) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create CORECELLS bag.");
        return (BEAR_ERROR);
    }
    bags.contents.bag.name = "IOPADS";
    if (octCreate(&chip, &bags) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create IOPADS bag.");
        return (BEAR_ERROR);
    }
    /* create the layers available on the chip */
    layers.type = OCT_LAYER;
    layers.contents.layer.name = "PLACE";
    if (octCreate(&chip,&layers) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create PLACE layer.");
        return (BEAR_ERROR);
    }
    layers.contents.layer.name = "MET1";
    if (octCreate(&chip,&layers) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create MET1 layer.");
        return (BEAR_ERROR);
    }
    layers.contents.layer.name = "MET2";
    if (octCreate(&chip,&layers) != OCT_OK) {
        octError("Odi_WriteToOct(): cannot create MET2 layer.");
        return (BEAR_ERROR);
    }

    /* 
     * Set technology directory used by OCT.
     */
    tapRootDirectory("$OCTTOOLS/tech");

    /* 
     * Go through all the (i/o pads) pins of the chip and write the 
     * pins to OCT.
     */
    if (Mac_ItrPinOfCellInst(Mac_GetInstOfChipMaster(cellMasterPtr),
        WriteIOPadsToOct, (ClientData) &chip) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

    /* 
     * Go through all the children cells of the chip and write the 
     * cells to OCT.
     */
    if (Mac_ItrChildrenOfCell(cellMasterPtr, 
        WriteCellsToOct, (ClientData) &chip) == BEAR_STOP_SEARCH) {
        return (BEAR_ERROR);
    }

#ifdef USE_DR_CODE
    /* Write out all the routing if specified. */
    if (routingFlag) {
        if (OdiWriteRoutingToOct(cellMasterPtr, &chip) == BEAR_ERROR) {
            return (BEAR_ERROR);
        }
    }
#endif /* USE_DR_CODE */

    /* flush data to disk */
    if (octFlushFacet(&chip) < OCT_OK) {
        octError("Odi_WriteToOct(): cannot save chip by flushing to disk.");
        return (BEAR_ERROR);
    }
    
    /* create interface facet for chip */
    interface.type = OCT_FACET;
    interface.contents.facet.cell = cell;
    interface.contents.facet.view = view;
    interface.contents.facet.version = OCT_CURRENT_VERSION;
    interface.contents.facet.facet = "interface";
    interface.contents.facet.mode = "w";

    (void) VOVoutputFacet(&interface);

    if (octOpenFacet(&interface) < OCT_OK) {
        octError("Odi_WriteToOct(): cannot open chip interface.");
        return (BEAR_ERROR);
    }
    
    vulcan(&chip, &interface);

    /* close the chip and interface facets */
    if (octCloseFacet(&chip) < OCT_OK) {
        octError("Odi_WriteToOct(): cannot close chip.");
        return (BEAR_ERROR);
    }
    if (octCloseFacet(&interface) < OCT_OK) {
        octError("Odi_WriteToOct(): cannot close chip interface.");
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}


