
/* pbAutoRoute.c --
 *
 *    Sherry Solden
 *    (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbAutoRoute.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#ifdef USE_DR_CODE

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "deque.h"
#include "grEditor.h"
#include "ezPlot.h"
#include "geoChDecomposer.h"
#include "topChDecomposer.h"
#include "globalSpacer.h"
#include "routeDB.h"
#include "localRouter.h"
#include "tileMapping.h"
#include "polarBear.h"
#include "stack.h"
#include "pbDetour.h"

#ifndef BEAR_BLACK_BOX
#include "bearIV.h"
#include "pbDialogs.h"
#endif /* BEAR_BLACK_BOX */


#ifdef DEBUG
#define USE_COMP_CODE 1     /* Use compaction code. */
#endif /* DEBUG */


typedef struct _LocalData LocalData;
struct _LocalData {
    Point        *pointPtr;
    WallJunction *junctPtr;
};


/*  inserted by Sherry for fixBadNets during unification */
typedef struct _LocalData2 LocalData2;
struct _LocalData2 {
   BearId intId;
   BearId extId;
};

typedef struct _AdjacentData AdjacentData;
struct _AdjacentData {
   PlaneData *planeDataPtr;
   Tile      *tilePtr;
   RouteCell *rCellPtr;
};


/* Status bits for type of channel */
#define alive 1
#define halfDead 2
#define allDead 3
#define leftDead 4
#define rightDead 5
#define topDead 6
#define botDead 7
#define bothDead 8


/* ---------- Local Procedures ---------- */
extern BearStatus      MarkGPin();
extern BearStatus      InsertRPin();
extern BearStatus      GenNewId();
extern BearStatus      ReverseDir();
extern PinDirection    OppositeDir();
extern void            CreateNewCellGroup();
extern void            LCreateNewCellGroup();
extern BearStatus      DefineAdjacentTile();
extern int             FindAdjacentTile();
extern char            *GetReply();
extern WallJunction    *GetJunction();
extern BearStatus      CheckPoint();
extern int             GetStatusOfChannel();
extern int             GetStatusOfWallSeg();
extern void            GetLAndRTiles();
extern void            GetBAndTTiles();
extern void            PinsAroundVWallSect();
extern void            PinsAroundHWallSect();
extern void            SearchTAndBJunct();
extern void            SearchRAndLJunct();
extern void            FixAllBadNets();
extern BearStatus      ChangeBadPinIds();
extern void            CleanPinList();
extern BearStatus      FindPPin();
extern bool            PPinOnTRSideOfTile();
extern BearStatus      CheckTopLeftSidePins();
extern BearStatus      CheckBottomRightSidePins();
extern void            RemoveSingleSidePins();
extern void            AbortRouteChannel();
extern void            AbortLChannel();


/*
 * ----------------------------------------------------------------------------
 *
 * PbAutoRouteCmd() --
 *  
 *    Automatically does the detail routing and ring routing for the chip.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
BearStatus
PbAutoRouteCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;     /* not used */
int         argCount;
char        **argList;
/*
 * BEAR_OK is returned only if the auto-routing is completed.
 */
{
    static int      rcCount = 1;            /* routed channel count */

#ifndef BEAR_BLACK_BOX
    ChipWin         *chipWinPtr;
#endif BEAR_BLACK_BOX

    BearStatus      bearStatus;
    CellMaster      *cellMasterPtr;
    Channel         *channelPtr;
    Deque           *channelQue;
    PlaneData       *planeDataPtr;
    Point           *point1Ptr; 
    Point           *point2Ptr;
    Point           point1; 
    Point           point2;
    Stack           *channelStackPtr;
    WallJunction    *llJunctPtr;
    WallJunction    *junct1Ptr;
    WallJunction    *junct2Ptr;
    WallJunction    *junct3Ptr;
    WallJunction    *leftJunctPtr; 
    WallJunction    *botJunctPtr; 
    WallJunction    *rightJunctPtr; 
    WallJunction    *topJunctPtr; 
    WallJunction    *junctPtr; 
    int             i;
    int             delta;
    int             ridgeCnt;
    bool            checkConnFlag = FALSE;
    bool            debugFlag = FALSE;
    bool            dialogWinFlag = TRUE;
    bool            doCompactionFlag = FALSE;
    bool            forceRouteFlag;
    bool            interactiveFlag = FALSE;
    bool            nutcrackerFlag = FALSE;
    bool            printMsgFlag = TRUE;
    bool            showSpacingFlag = FALSE;
    bool            viaReduceFlag = TRUE;
    bool            LToggle = FALSE;
    bool            warnFlag = FALSE;

#ifndef BEAR_BLACK_BOX
    IVWindow        *IVwin;
    XEvent          theEvent;
#endif BEAR_BLACK_BOX
    bool            doRoute = FALSE;
    bool            exitFlag = FALSE;

    if (lr_singleTermFlag == TRUE) {
        PbPrint(EZ_ITALIC_FONT,
            "Database has been corrupted cannot continue.\n");
        PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
        return (BEAR_ERROR);
    }

    /* I added the line below this. Massoud */
    Gs_SetPause(FALSE);

    /* parse arguments */
    for (i = 1; i < argCount; i++) {
        /* 
         * Print out this commands help message.
         */
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        }

        /* 
         * Don't pop up any dialog windows.  Used when you want to run
         * this command from a script.
         */
        if (strcmp (argList[i], "-NDW") == 0) {
            dialogWinFlag = FALSE;
            continue;
        }

#ifdef USE_COMP_CODE
        /*
         * Set compaction mode.
         */
        if (strcmp (argList[i], "-c") == 0) {
            doCompactionFlag = TRUE;
            continue;
        }
#endif /* USE_COMP_CODE */

        /*
         * Set check connectivity mode which will verify that
         * all of the nets are connected after each step.
         */
        if (strcmp (argList[i], "-C") == 0) {
            checkConnFlag = TRUE;
            warnFlag = FALSE;
            continue;
        }
        if (strcmp (argList[i], "-CW") == 0) {
            checkConnFlag = TRUE;
            warnFlag = TRUE;
            continue;
        }

        /* 
         * Set interactive mode.
         */
        if (strcmp (argList[i], "-i") == 0) {
            if (bear_BlackBoxFlag == TRUE) {
                PbPrint(EZ_ITALIC_FONT,
                    "-i option not used in black box mode.\n");
                return (BEAR_ERROR);
            }
            interactiveFlag = TRUE;
            continue;
        }

#ifdef USE_NUT_CODE
        /*
         * Set nutcracker mode.
         */
        if (strcmp (argList[i], "-nut") == 0) {
            nutcrackerFlag = TRUE;
            continue;
        }
#endif USE_NUT_CODE

        /*
         * Set reduce via's mode.
         */
        if (strcmp (argList[i], "-via") == 0) {
            viaReduceFlag = TRUE;
            continue;
        }

        goto usage;
    }

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        chipWinPtr = PbGetChipWin(winDataPtr);
        cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    }
    else 
#endif BEAR_BLACK_BOX
    {
        if (pbCellMasterPtr == NIL(CellMaster)) {
            PbPrint(EZ_ITALIC_FONT, "Must read in a cell master first.\n");
            return (BEAR_ERROR);
        }
        else {
            cellMasterPtr = pbCellMasterPtr;
        }
    }
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    if (Mac_IsGlobalRoutingDone(cellMasterPtr) == FALSE) {
        PbPrint(EZ_ITALIC_FONT, "Complete global routing and try again.\n");
        return (BEAR_ERROR);
    }

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && dialogWinFlag) {
        IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME, 
            "Auto-Router", 13);
        
#ifdef USE_NUT_CODE
        (void) iv_AddVar(IVwin, "Use Nutcracker", 
              &nutcrackerFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
#endif USE_NUT_CODE
        
        (void) iv_AddVar(IVwin, "Via Reduction", 
              &viaReduceFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);

#ifdef USE_COMP_CODE
        (void) iv_AddVar(IVwin, "Do Compaction",
              &doCompactionFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
#endif /* USE_COMP_CODE */

        (void) iv_AddVar(IVwin, "Check Connectivity",
              &checkConnFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);

        (void) iv_AddVar(IVwin, "Interactive Mode",
              &interactiveFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Route Chip", 
              &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

        (void) iv_AddVar(IVwin, "Exit",
            &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

        (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

        while (!doRoute) {
            XNextEvent(ez_Display, &theEvent);
            (void) iv_HandleEvent(&theEvent);;
            Ez_XEventHandler(&theEvent);
            if (exitFlag) {
                iv_FreeWindow(IVwin);
                PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (by user).\n");
                return (BEAR_ERROR);
            }
        }
        iv_FreeWindow(IVwin);
        Ez_EmptyXEventQue();
        showSpacingFlag = interactiveFlag;
    }
    else 
#endif BEAR_BLACK_BOX
    {
        showSpacingFlag = interactiveFlag = FALSE;
    }
    
    if (Fpg_IsFpgBuilt(planeDataPtr) == FALSE) {
        BuildFloorPlan(cellMasterPtr, FALSE);
    }

    forceRouteFlag = !doCompactionFlag;

    while (TRUE) {
    /* Get a channel to be routed */
    channelStackPtr = StackNew(1);
    (void) Tcd_RRDO(Tp_GetFpg(planeDataPtr), channelStackPtr);
    channelQue = MakeDeque();
    Tcd_FindIndepChannel(cellMasterPtr, channelStackPtr, channelQue);

    if (IsEmptyDeque(channelQue)) {
        /* No channels remain to be routed */
        break;
    }

    while (IsEmptyDeque(channelQue) == FALSE) {
        channelPtr = (Channel *)PopDeque(channelQue);
        
        if ((channelPtr->type == HORZ_CHANNEL) ||
        (channelPtr->type == VERT_CHANNEL))
        {
        break;
        }
        /* Make sure the L-channel is not adjacent to an empty cell */
        junct1Ptr = channelPtr->channelUn.lShapedChannel.hEndJunct;
        junct2Ptr = channelPtr->channelUn.lShapedChannel.cornerJunct;
        junct3Ptr = channelPtr->channelUn.lShapedChannel.vEndJunct;

        if (Fpg_GetDegreeOfWallJunct(junct2Ptr) > 2)
        {
        channelPtr = NIL(Channel);
        continue;
        }

        if (junct2Ptr->centerPoint.p_y < junct3Ptr->centerPoint.p_y)
        {
        topJunctPtr = junct3Ptr;
        botJunctPtr = junct2Ptr;
        }
        else
        {
        topJunctPtr = junct2Ptr;
        botJunctPtr = junct3Ptr;
        }
        for (junctPtr  = Fpg_GetSouthJunctOfWallJunct(topJunctPtr);
         junctPtr != botJunctPtr;
         junctPtr  = Fpg_GetSouthJunctOfWallJunct(junctPtr))
        {
        if (Fpg_GetDegreeOfWallJunct(junctPtr) > 2)
        {
            channelPtr = NIL(Channel);
            continue;
        }
        }

        if (junct2Ptr->centerPoint.p_x < junct1Ptr->centerPoint.p_x)
        {
        leftJunctPtr = junct2Ptr;
        rightJunctPtr = junct1Ptr;
        }
        else
        {
        leftJunctPtr = junct1Ptr;
        rightJunctPtr = junct2Ptr;
        }
        for (junctPtr  = Fpg_GetEastJunctOfWallJunct(leftJunctPtr);
         junctPtr != rightJunctPtr;
         junctPtr  = Fpg_GetEastJunctOfWallJunct(junctPtr))
        {
        if (Fpg_GetDegreeOfWallJunct(junctPtr) > 2)
        {
            channelPtr = NIL(Channel);
            continue;
        }
        }
    }

    if (channelPtr == NIL(Channel))
    {
        /* No channels remain to be routed */
        break;
    }

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        /* Highlight Channel */
        PbViewEntireObject(winDataPtr, Tp_GetHorzPlaneBnd(planeDataPtr), FALSE);

        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
        PbHighlightChannel(cellMasterPtr, winDataPtr, channelPtr);
    }
#endif BEAR_BLACK_BOX

    if ((channelPtr->type == HORZ_CHANNEL) ||
        (channelPtr->type == VERT_CHANNEL)) {
        junct1Ptr = channelPtr->channelUn.straightChannel.tlEndJunct;
        junct2Ptr = channelPtr->channelUn.straightChannel.brEndJunct;
    }
    else {
        junct1Ptr = channelPtr->channelUn.lShapedChannel.hEndJunct;
        junct2Ptr = channelPtr->channelUn.lShapedChannel.cornerJunct;
        junct3Ptr = channelPtr->channelUn.lShapedChannel.vEndJunct;
    }

    if (checkConnFlag == TRUE) {
        if (Gre_CheckConnForAllNets(cellMasterPtr, warnFlag) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT,
                "Auto Route aborted (disconnected nets).\n");
            return (BEAR_ERROR);
        }
    }

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && interactiveFlag) {
        char string[10];
        bool keepGoing = TRUE;

/*
 *      PbAutoRoutePauseDialog(winDataPtr,
 *          (checkConnFlag == TRUE) ? FALSE : TRUE,
 *          &keepGoing);
 */
        fprintf (stdout, "\nKeep going (y or n) ? ");
        fflush (stdout);
        fscanf (stdin, "%1s", string);
        keepGoing = (strcmp (string, "y") == 0) ? TRUE : FALSE;

        if (keepGoing == FALSE) {
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (by user).\n");
            return (BEAR_ERROR);
        }
    }
#endif BEAR_BLACK_BOX

    /*
     * Save the point locations of the junctions since Unify/Split
     * will recreate floor plan graph.
     */
    point1Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct1Ptr);
    point2Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct2Ptr);
    point1.p_x = point1Ptr->p_x;
    point1.p_y = point1Ptr->p_y;
    point2.p_x = point2Ptr->p_x;
    point2.p_y = point2Ptr->p_y;

    if ((channelPtr->type == VERT_CHANNEL) ||
        (channelPtr->type == HORZ_CHANNEL))
    {
        delta = 0;
        bearStatus = PbRouteChannel(cellMasterPtr,
                    junct1Ptr, 
                    junct2Ptr, 
                    FALSE,
                    nutcrackerFlag,
                    viaReduceFlag,
                    forceRouteFlag,
                    &delta,
                    debugFlag);
    }
    else {
        bearStatus = PbRouteLChannel(cellMasterPtr, 
                     junct1Ptr, 
                     junct2Ptr, 
                     junct3Ptr,
                     FALSE,
                     viaReduceFlag,
                     debugFlag);
    }
    switch (bearStatus) {
    case BEAR_OK:
        if (printMsgFlag) {
            sprintf (bear_Msg, "Completed channel number %d.\n", rcCount);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
            rcCount++;
        }
        break;
    case BEAR_ERROR_2:
#ifndef BEAR_BLACK_BOX
        if (bear_BlackBoxFlag == FALSE) {
            (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
        }
#endif BEAR_BLACK_BOX
        PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
        return (BEAR_ERROR);
    default:
        break;
    }

    /* Adjust Placement based on if routing was successful */
    if (((bearStatus == BEAR_OK) && (channelPtr->type == VERT_CHANNEL)) ||
        ((delta > 0) && (channelPtr->type == HORZ_CHANNEL)))
    {
        /* Adjust Placement - Vert Decompact */
        if (PbDoVertDecompact(winDataPtr, &ridgeCnt,
            showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
            return (BEAR_ERROR);
        }
    } 
    else if (((delta > 0) && (channelPtr->type == VERT_CHANNEL)) ||
         ((bearStatus == BEAR_OK) && (channelPtr->type == HORZ_CHANNEL)))
    {
        /* Adjust Placement - Horz Decompact */
        if (PbDoHorzDecompact(winDataPtr, &ridgeCnt, 
            showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
            return (BEAR_ERROR);
        }
    }
    else if ((delta < 0) && (channelPtr->type == VERT_CHANNEL)) {
        /* Adjust Placement - Horz Compact */
        if (PbDoHorzCompact(winDataPtr, &ridgeCnt,
            showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
            return (BEAR_ERROR);
        }
        /*
         * If no compaction was made (wouldn't reduce chip area)
         * then force channel route even with extra space.
         */
        if (ridgeCnt == 1) {
            /* Get the new junction pointers from the latest fpg */
            llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
            junct1Ptr = GetJunction(llJunctPtr, &point1);
            junct2Ptr = GetJunction(llJunctPtr, &point2);
            delta = 0;
            bearStatus = PbRouteChannel(cellMasterPtr,
                junct1Ptr, 
                junct2Ptr,
                FALSE,
                nutcrackerFlag,
                viaReduceFlag,
                TRUE,
                &delta,
                debugFlag);
        }
        switch (bearStatus) {
        case BEAR_OK:
            if (printMsgFlag) {
                sprintf (bear_Msg, "Completed channel number %d.\n", rcCount);
                PbPrint(EZ_NORMAL_FONT, bear_Msg);
                rcCount++;
            }
            break;
        case BEAR_ERROR_2:
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
            return (BEAR_ERROR);
        default:
            break;
        }
    }
    else if ((delta < 0) && (channelPtr->type == HORZ_CHANNEL))
    {
        /* Adjust Placement - Vert Compact */
        if (PbDoVertCompact(winDataPtr, &ridgeCnt,
            showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
            return (BEAR_ERROR);
        }
        /*
         * If no compaction was made (wouldn't reduce chip area)
         * then force channel route even with extra space.
         */
        if (ridgeCnt == 1) {
            /* Get the new junction pointers from the latest fpg */
            llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
            junct1Ptr = GetJunction(llJunctPtr, &point1);
            junct2Ptr = GetJunction(llJunctPtr, &point2);
            delta = 0;
            bearStatus = PbRouteChannel(cellMasterPtr,
                junct1Ptr, 
                junct2Ptr,
                FALSE,
                nutcrackerFlag,
                viaReduceFlag,
                TRUE,
                &delta,
                debugFlag);
        }
        switch (bearStatus) {
        case BEAR_OK:
            if (printMsgFlag) {
                sprintf (bear_Msg, "Completed channel number %d.\n", rcCount);
                PbPrint(EZ_NORMAL_FONT, bear_Msg);
                rcCount++;
            }
            break;
        case BEAR_ERROR_2:
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
#endif BEAR_BLACK_BOX
            PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
            return (BEAR_ERROR);
        default:
            break;
        }
    }
    else {
        if (LToggle == TRUE) {
            /* Adjust Placement - Horz Decompact */
            if (PbDoHorzDecompact(winDataPtr, &ridgeCnt,
                showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
                if (bear_BlackBoxFlag == FALSE) {
                    (void) PbItrChipWindows(PbDrawChipRegion,
                        (ClientData) NULL);
                }
#endif BEAR_BLACK_BOX
                PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
                return (BEAR_ERROR);
            }
            /* Adjust Placement - Vert Decompact */
            if (PbDoVertDecompact(winDataPtr, &ridgeCnt,
                showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
                if (bear_BlackBoxFlag == FALSE) {
                    (void) PbItrChipWindows(PbDrawChipRegion,
                        (ClientData) NULL);
                }
#endif BEAR_BLACK_BOX
                PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
                return (BEAR_ERROR);
            }
            LToggle = FALSE;
        }
        else {
            /* Adjust Placement - Vert Decompact */
            if (PbDoVertDecompact(winDataPtr, &ridgeCnt,
                showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
                if (bear_BlackBoxFlag == FALSE) {
                    (void) PbItrChipWindows(PbDrawChipRegion,
                        (ClientData) NULL);
                }
#endif BEAR_BLACK_BOX
                PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
                return (BEAR_ERROR);
            }
            /* Adjust Placement - Horz Decompact */
            if (PbDoHorzDecompact(winDataPtr, &ridgeCnt,
                showSpacingFlag, printMsgFlag, FALSE) == BEAR_ERROR) {
#ifndef BEAR_BLACK_BOX
                if (bear_BlackBoxFlag == FALSE) {
                    (void) PbItrChipWindows(PbDrawChipRegion,
                        (ClientData) NULL);
                }
#endif BEAR_BLACK_BOX
                PbPrint(EZ_NORMAL_FONT, "Auto Route aborted (error).\n");
                return (BEAR_ERROR);
            }
            LToggle = TRUE;
        }
    }
    }

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    }
#endif BEAR_BLACK_BOX

    /* ring Route */

    Mac_SetCoreRoutingDone(cellMasterPtr);
    PbPrint(EZ_NORMAL_FONT, "Auto Route is done.\n");

    return (BEAR_OK);

usage: 
    PbPrint(EZ_NORMAL_FONT,
        "auto route the chip (ok in script) (ok in black box)\n");

    PbPrint(EZ_NORMAL_FONT, "Usage: ar [-options ...]\n");
    PbPrint(EZ_NORMAL_FONT, "\nwhere options include:\n");

    PbPrint(EZ_NORMAL_FONT, "    -C check connectivity after each step\n");
    PbPrint(EZ_NORMAL_FONT, "    -CW  check connectivity after each step\n");
    PbPrint(EZ_NORMAL_FONT, "         and print warnings\n");
    PbPrint(EZ_NORMAL_FONT, "    -via reduce via's\n");

#ifdef USE_COMP_CODE
    PbPrint(EZ_NORMAL_FONT, "    -c compaction mode\n");
#endif /* USE_COMP_CODE */

#ifdef USE_NUT_CODE
    PbPrint(EZ_NORMAL_FONT, "    -nut use nutcracker\n");
#endif /* USE_NUT_CODE */

    PbPrint(EZ_NORMAL_FONT, "    -NDW no dialog windows\n");

    return (BEAR_ERROR);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbRouteChannel() --
 *
 * Results:
 *   returns either BEAR_OK, BEAR_ERROR, or BEAR_ERROR_2.
 *
 *   If BEAR_ERROR_2 is returned the routing process should
 *   be stopped.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbRouteChannel(cellMasterPtr, junct1Ptr, junct2Ptr, interactiveFlag,
    nutcrackerFlag, viaReduceFlag, forceRouteFlag, deltaPtr, debug)

CellMaster   *cellMasterPtr;
WallJunction *junct1Ptr;
WallJunction *junct2Ptr;
bool         interactiveFlag;
bool         nutcrackerFlag;
bool         viaReduceFlag;
bool         forceRouteFlag;
int          *deltaPtr;
bool         debug;
{
    AdjPar             *adjParPtr;
    DCellsAroundRCell  *dCellListsPtr;
    GenericCell        *gCellPtr;
    PinsAroundJunction *pinListPtr;
    PinsAroundJunction *trListPtr;
    PinsAroundJunction *blListPtr;
    PinsAroundJunction trList;
    PinsAroundJunction blList;
    PinsAroundJunction tmptrList;
    PinsAroundJunction tmpblList;
    PlaneData          *planeDataPtr;
    Point              *point1Ptr; 
    Point              *point2Ptr;
    Point              point1;
    Point              point2;
    Point              vPoint;
    Rect               dmntRect;
    Rect               vdmntRect;
    RouteCell          *rCellPtr;
    Tile               *horzDomTile1Ptr; 
    Tile               *vertDomTile1Ptr;
    Tile               *horzDomTile2Ptr; 
    Tile               *vertDomTile2Ptr;
    WallJunction       *llJunctPtr;
    module             *modulePtr;
    bool               doRoute=FALSE;
    bool               exitFlag=FALSE;
    int                status;
    int                heightRequired;
    int                heightAvail;
    int                heightTarget;
    int                nut_compaction;
    int                compactPar = 0;
    char               *reply;
#ifndef BEAR_BLACK_BOX
    IVWindow           *IVwin;
    XEvent             theEvent;
#endif BEAR_BLACK_BOX

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    point1Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct1Ptr);
    point2Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct2Ptr);

    /*
     * IMPORTANT : The variable 'point1Ptr' contains an address that is
     * found in the variable 'junct1Ptr' and as such it's valid only until
     * a new floor plan is built.  Since this procedure causes (either
     * directly or indirectly) the floor plan to be rebuilt a number of
     * times, we need to save the coordinates of point one rather than
     * its address.  Likewise for 'point2Ptr'.
     */
    point1.p_x = point1Ptr->p_x;
    point1.p_y = point1Ptr->p_y;
    point2.p_x = point2Ptr->p_x;
    point2.p_y = point2Ptr->p_y;


    horzDomTile1Ptr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr),&point1);
    horzDomTile2Ptr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr),&point2);
    Tp_HPointToVPoint(&point1, &vPoint);
    vertDomTile1Ptr = TiSrPointNoHint(Tp_GetVertPlane(planeDataPtr),&vPoint);
    Tp_HPointToVPoint(&point2, &vPoint);
    vertDomTile2Ptr = TiSrPointNoHint(Tp_GetVertPlane(planeDataPtr),&vPoint);

    (void) Mac_CreateRCell(cellMasterPtr, &rCellPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);


    if (point1.p_x == point2.p_x)
    {
        /* The channel is vertical. */
        (void) Mac_PutTypeOfRCell(VERT_RCELL, rCellPtr);

        /* Must save the rect for tmHCircleUnify - lose it in VSplit */
        Tp_VertTileToRect(planeDataPtr,vertDomTile1Ptr,&dmntRect);
        dmntRect.r_ytop = point1.p_y;
        dmntRect.r_ybot = point2.p_y;

        /* Define the geometrical channel */
        
        /* Must delete dummy cells first */
        (void) Tm_VUnify(cellMasterPtr);
        /* Reassign the junctPtrs since Unify rebuilt floorplan graph */
        llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
        junct1Ptr = GetJunction(llJunctPtr, &point1);
        junct2Ptr = GetJunction(llJunctPtr, &point2);
        (void) Gcd_DefineVertBound(cellMasterPtr, junct1Ptr, junct2Ptr,
                               vertDomTile1Ptr, &rCellPtr);
        (void) Tm_VSplit(cellMasterPtr);

        planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
        (void) Mac_CreateBBoxOfRCell(rCellPtr);

        /* Reassign the junctPtrs since floorplan graph has changed */
        llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
        junct1Ptr = GetJunction(llJunctPtr, &point1);
        junct2Ptr = GetJunction(llJunctPtr, &point2);

    status = GetStatusOfChannel(planeDataPtr, junct1Ptr, 
        junct2Ptr, VERT_CHANNEL);
    if (status == halfDead) {
        /* Detour PseudoPins that might enter the boundary of the channel */
        /* so they are rerouted to the side of the channel */
        DefineAdjacentDummyCells(planeDataPtr, Mac_GetTypeOfRCell(rCellPtr), 
                                 dmntRect, &dCellListsPtr);
        DetourAroundDCells(cellMasterPtr, planeDataPtr, rCellPtr,
                           &dCellListsPtr);
        Tm_PinsAroundCircle(cellMasterPtr, planeDataPtr, junct1Ptr, 
                            &tmptrList, TRUE);
        Tm_PinsAroundCircle(cellMasterPtr, planeDataPtr, junct2Ptr, 
                            &tmpblList, TRUE);
    }

    /* Define Boundary Pins Of Channel */
#ifdef USE_WIRE_CODE
    if (Gre_DoPgWiringForAllNets(cellMasterPtr) == BEAR_ERROR) {
        return (BEAR_ERROR_2);
    }
#endif USE_WIRE_CODE

    /* Horizontal Unify No Free */
    Tm_HUnifyNoFree(cellMasterPtr);
    /* Tm_HUnify(cellMasterPtr);*/

    /* Circle Unify - to make netIds match across multiple faces */
    (void) Tm_HCircleUnify(cellMasterPtr, planeDataPtr, dmntRect);

    /* Reassign the junctPtrs since floorplan graph has changed */
    llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
    junct1Ptr = GetJunction(llJunctPtr, &point1);
    junct2Ptr = GetJunction(llJunctPtr, &point2);


    /* Define Bound Pin */
    Gcd_DefineBoundPin(planeDataPtr,rCellPtr);
    Mac_ItrBLBoundPinOfRCell(rCellPtr,ReverseDir,(ClientData)NULL);
    Mac_ItrTRBoundPinOfRCell(rCellPtr,ReverseDir,(ClientData)NULL);
      
    /* For debugging purposes only
       printf("Bound Pins on LEFT_SIDE\n");
       Mac_ItrBLBoundPinOfRCell(rCellPtr,Mac_PrintGPin,(ClientData)NULL);
       printf("Bound Pins on RIGHT_SIDE\n");
       Mac_ItrTRBoundPinOfRCell(rCellPtr,Mac_PrintGPin,(ClientData)NULL);
        */

    /* Define Side Pin */

    /* Get all pins adjacent to each edge of channel */
    Tp_HorzTileToRect(planeDataPtr, horzDomTile1Ptr, &dmntRect);
    trListPtr = Tm_PinsAroundHFace(cellMasterPtr, planeDataPtr, 
                       dmntRect, FALSE);
    trList = *trListPtr;

    Tp_HorzTileToRect(planeDataPtr,horzDomTile2Ptr,&dmntRect);
    blListPtr = Tm_PinsAroundHFace(cellMasterPtr, planeDataPtr, 
                       dmntRect, FALSE);
    blList = *blListPtr;

    if (status == halfDead)
    {
        /* Append the List from PinsAroundCircle */
        CleanPinList(tmptrList);
        AppendLists(&tmptrList, &trList);
        CleanPinList(tmpblList);
        AppendLists(&tmpblList, &blList);
    }

    Gcd_DefineSidePin(planeDataPtr, junct2Ptr, &blList,
              rCellPtr, RIGHT_SIDE, NORTH_PIN);
    Gcd_DefineSidePin(planeDataPtr, junct1Ptr, &trList,
              rCellPtr, TOP_LEFT_SIDE, SOUTH_PIN);

        /* remove "single terminal" entries from sidePinList */
        if (status != alive)
    {
        RemoveSingleSidePins(rCellPtr);
    }

    /* For debugging purposes only
       printf("Side Pins on TOP_SIDE\n");
       Mac_ItrTLSidePinOfRCell(rCellPtr,Mac_PrintRPin,(ClientData)NULL);
       printf("Side Pins on BOTTOM_SIDE\n");
       Mac_ItrRSidePinOfRCell(rCellPtr,Mac_PrintRPin,(ClientData)NULL);
        */

    /* Route Channel */
    if (Rdb_MacToBBL(rCellPtr, &modulePtr) == BEAR_ERROR) {
        Mac_DeleteRCell(rCellPtr);
        return (BEAR_ERROR_2);
    }
    rCellPtr->routeModPtr = modulePtr;
    modulePtr->placg=2;
    MALLOC(AdjPar *, adjParPtr, sizeof (AdjPar));
    if (Rdb_RouteChannelPass1(modulePtr, rCellPtr, adjParPtr) == BEAR_ERROR) {
        AbortRouteChannel(cellMasterPtr, planeDataPtr, rCellPtr,
            point1, point2, adjParPtr, dCellListsPtr, status);
        return (BEAR_ERROR_2);
    }
    heightRequired = Lr_GetChannelHeight(0);
    heightAvail = Lr_GetChannelHeight(1);
    adjParPtr->xAdj = adjParPtr->yAdj = heightRequired  - heightAvail;

#ifndef BEAR_BLACK_BOX
    if (interactiveFlag == TRUE) {
        heightTarget = heightAvail;
        IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
            "Route Channel", 13);
        (void) iv_AddVar(IVwin, "Available Channel Height", 
              &heightAvail, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightAvail, FALSE);
        (void) iv_AddVar(IVwin, "Required Channel Height", 
              &heightRequired, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightRequired, FALSE);
        (void) iv_AddVar(IVwin, "Target Channel Height",
              &heightTarget, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        
#ifdef USE_NUT_CODE
        (void) iv_AddVar(IVwin, "Use Nutcracker", 
              &nutcrackerFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
#endif USE_NUT_CODE
        
        (void) iv_AddVar(IVwin, "Via Reduction", 
              &viaReduceFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Route Channel", 
              &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Exit",
            &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);
        while (!doRoute) {
            XNextEvent(ez_Display, &theEvent);
            (void) iv_HandleEvent(&theEvent);;
            Ez_XEventHandler(&theEvent);
            if (exitFlag) {
                iv_FreeWindow(IVwin);
                AbortRouteChannel(cellMasterPtr, planeDataPtr, rCellPtr,
                    point1, point2, adjParPtr, dCellListsPtr, status);
                *deltaPtr = adjParPtr->xAdj;
                return (BEAR_ERROR);
            }
        }
        iv_FreeWindow(IVwin);
        Ez_EmptyXEventQue();
        
        doRoute = FALSE;
        exitFlag = FALSE;
    }
#endif BEAR_BLACK_BOX
    
#ifdef USE_NUT_CODE
        if (nutcrackerFlag == TRUE) {
            if (heightAvail > heightRequired) {
                adjParPtr->xAdj = adjParPtr->yAdj = 0;
            }
            nut_compaction = Ls_nutcracker(modulePtr);
            heightRequired = MAX(heightRequired, heightAvail) - nut_compaction;
           
#ifndef BEAR_BLACK_BOX
        if (interactiveFlag == TRUE) {
        IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
            "Nutcracker", 13);
        (void) iv_AddVar(IVwin, "Available Channel Height", 
               &heightAvail, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightAvail, FALSE);
        (void) iv_AddVar(IVwin, "Required Channel Height", 
               &heightRequired, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightRequired, FALSE);
        (void) iv_AddVar(IVwin, "Target Channel Height",
            &heightTarget, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Continue", 
               &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Exit",
            &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);
        adjParPtr->xAdj = adjParPtr->yAdj = heightRequired - heightAvail;
        while (!doRoute) 
        {
            XNextEvent(ez_Display, &theEvent);
            (void) iv_HandleEvent(&theEvent);;
            Ez_XEventHandler(&theEvent);
            if (exitFlag) 
            {
                iv_FreeWindow(IVwin);
                AbortRouteChannel(cellMasterPtr, planeDataPtr, rCellPtr,
                    point1, point2, 
                    adjParPtr, dCellListsPtr, status);
                *deltaPtr = adjParPtr->xAdj;
                return (BEAR_ERROR);
            }
        }
        iv_FreeWindow(IVwin);
        Ez_EmptyXEventQue();
        }
#endif BEAR_BLACK_BOX

        Ls_nutcrackerPass2(modulePtr, &compactPar,
            heightRequired, viaReduceFlag);
    }

    if ((viaReduceFlag == TRUE) && (nutcrackerFlag == TRUE)) {
        /* via reduction, if TRUE, is done in nutcracker module */
        Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr, compactPar,
            FALSE);
    }
    else
#endif USE_NUT_CODE
    {
        Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr, compactPar,
                  viaReduceFlag);
    }

    /* Abort the channel route if: */
    /* 1) not enough space (need to decompact) */
    /* 2) too much space and not forced to route (can attempt to compact) */
    if ((heightAvail < heightRequired) ||
        ((heightAvail > heightRequired) && (forceRouteFlag == FALSE)))
    {
        AbortRouteChannel(cellMasterPtr, planeDataPtr,
            rCellPtr, point1, point2, adjParPtr,
            dCellListsPtr, status);
        *deltaPtr = adjParPtr->xAdj;
        return (BEAR_ERROR);
    }

    /* Generate new Id */
    reply = debug ? GetReply("generate new id ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {   
        Tp_HorzTileToRect(planeDataPtr,horzDomTile1Ptr,&dmntRect);
        pinListPtr = Tm_PinsAroundHFace(cellMasterPtr, planeDataPtr, 
            dmntRect, TRUE);
        if (status == halfDead) {
            trListPtr = Pb_PinsAroundDCells(cellMasterPtr, planeDataPtr, 
                Mac_GetTypeOfRCell(rCellPtr), dCellListsPtr, point1);
            AppendLists(trListPtr, pinListPtr);
        }
        Mac_ItrTLSidePinOfRCell(rCellPtr,GenNewId,(ClientData)pinListPtr);
        Tm_FreePinsAroundJunction(pinListPtr);
    }

    /* Reassign floor plan graph */
    llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
    junct1Ptr = GetJunction(llJunctPtr, &point1);
    junct2Ptr = GetJunction(llJunctPtr, &point2);

    /* Create a new cell group */
    CreateNewCellGroup(planeDataPtr, rCellPtr);

    /* Update Net Id Of Side Pin */
    reply = debug ? GetReply("update net id of side pin ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
        Gcd_UpdateSidePinOfRCell(rCellPtr,TOP_LEFT_SIDE);
        Gcd_UpdateSidePinOfRCell(rCellPtr,RIGHT_SIDE);
    }

    /* Create Generic Cell */
    reply = debug ? GetReply("create generic cell ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
        MALLOC(GenericCell *,gCellPtr,sizeof (GenericCell));
        Mac_CreateGCell(ROUTE_CELL,rCellPtr,&gCellPtr);
        Tp_PaintGCell(gCellPtr,planeDataPtr);
        Tp_IdentifySpaceTiles(planeDataPtr);
        (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
        Mac_ItrBoundPinOfRCell(rCellPtr,MarkGPin,(ClientData)NULL);
        Mac_ItrSidePinOfRCell(rCellPtr,InsertRPin,(ClientData)NULL);
    }

    /* Horizontal Face Split */
    reply = debug ? GetReply("split horizontal face ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
        if (status == halfDead)
        {
        /* Need to split the circle  Tm_HCircleSplit */
        if (dCellListsPtr->BRptr)
        {
            dmntRect.r_xbot = dCellListsPtr->BRpoint->p_x;
            dmntRect.r_xtop = dmntRect.r_xbot + 1;
            dmntRect.r_ytop = point1.p_y;
            dmntRect.r_ybot = dCellListsPtr->BRpoint->p_y;
            (void) Tm_HCSplit(cellMasterPtr, planeDataPtr, dmntRect);
        }
        if (dCellListsPtr->BLptr)
        {
            dmntRect.r_xbot = dCellListsPtr->BLpoint->p_x;
            dmntRect.r_xtop = dmntRect.r_xbot + 1;
            dmntRect.r_ybot = point2.p_y;
            dmntRect.r_ytop = dCellListsPtr->BLpoint->p_y;
            (void) Tm_HCSplit(cellMasterPtr, planeDataPtr, dmntRect);
        }
        if (dCellListsPtr->TRptr)
        {
            dmntRect.r_xtop = dCellListsPtr->TRpoint->p_x;
            dmntRect.r_xbot = dmntRect.r_xtop - 1;
            dmntRect.r_ytop = point1.p_y;
            dmntRect.r_ybot = dCellListsPtr->TRpoint->p_y;
            (void) Tm_HCSplit(cellMasterPtr, planeDataPtr, dmntRect);
        }
        if (dCellListsPtr->TLptr)
        {
            dmntRect.r_xtop = dCellListsPtr->TLpoint->p_x;
            dmntRect.r_xbot = dmntRect.r_xtop - 1;
            dmntRect.r_ybot = point2.p_y;
            dmntRect.r_ytop = dCellListsPtr->TLpoint->p_y;
            (void) Tm_HCSplit(cellMasterPtr, planeDataPtr, dmntRect);
            }   
        }
        Tm_InsDCellOnHTP(cellMasterPtr);
        Tm_HFSplit(cellMasterPtr);

    }

        /*
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
    */

    } 
    else if (point1.p_y == point2.p_y) 
    {
        /* The channel is horizontal. */
        (void) Mac_PutTypeOfRCell(HORZ_RCELL, rCellPtr);
        
        /* Must get the rect for VCircleUnify now (lose it in HSplit) */
    Tp_HorzTileToRect(planeDataPtr,horzDomTile1Ptr,&dmntRect);
    dmntRect.r_xbot = point1.p_x;
        dmntRect.r_xtop = point2.p_x;
    Tp_HRectToVRect(&dmntRect, &vdmntRect);


    /* Define the geometrical channel */

    /* Must delete dummy cells first */
    (void) Tm_HUnify(cellMasterPtr);
    /* Reassign the junctPtrs since Unify rebuilt floorplan graph */
    llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
    junct1Ptr = GetJunction(llJunctPtr, &point1);
    junct2Ptr = GetJunction(llJunctPtr, &point2);
    (void) Gcd_DefineHorzBound(cellMasterPtr, junct1Ptr, junct2Ptr,
                                   horzDomTile1Ptr, &rCellPtr);
    (void) Tm_HSplit(cellMasterPtr);

        planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    (void) Mac_CreateBBoxOfRCell(rCellPtr);

    /* Reassign the junctPtrs since floorplan graph was rebuilt */
    llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
    junct1Ptr = GetJunction(llJunctPtr, &point1);
    junct2Ptr = GetJunction(llJunctPtr, &point2);

    status = GetStatusOfChannel(planeDataPtr, junct1Ptr, 
        junct2Ptr, HORZ_CHANNEL);
    if (status == halfDead) {
        /* Detour PseudoPins that might enter the boundary of the channel */
        /* so they are rerouted to the side of the channel */
        DefineAdjacentDummyCells(planeDataPtr, Mac_GetTypeOfRCell(rCellPtr), 
                     vdmntRect, &dCellListsPtr);
        DetourAroundDCells(cellMasterPtr, planeDataPtr, rCellPtr,
                   &dCellListsPtr);
        Tm_PinsAroundCircle(cellMasterPtr, planeDataPtr, junct1Ptr, 
                &tmpblList, TRUE);
        Tm_PinsAroundCircle(cellMasterPtr, planeDataPtr, junct2Ptr, 
                &tmptrList, TRUE);
    }

#ifdef USE_WIRE_CODE
    if (Gre_DoPgWiringForAllNets(cellMasterPtr) == BEAR_ERROR) {
        return (BEAR_ERROR_2);
    }
#endif USE_WIRE_CODE

    /* Vertical Unify No Free */
    Tm_VUnifyNoFree(cellMasterPtr);
    /* Tm_VUnify(cellMasterPtr);*/

    /* Circle Unify - to make netIds match across multiple faces */
    (void) Tm_VCircleUnify(cellMasterPtr, planeDataPtr, vdmntRect);

    /* Reassign the junctPtrs since floorplan graph has changed */
    llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
    junct1Ptr = GetJunction(llJunctPtr, &point1);
    junct2Ptr = GetJunction(llJunctPtr, &point2);

    /* Define Bound Pin */
    Gcd_DefineBoundPin(planeDataPtr,rCellPtr);
    Mac_ItrBLBoundPinOfRCell(rCellPtr,ReverseDir,(ClientData)NULL);
    Mac_ItrTRBoundPinOfRCell(rCellPtr,ReverseDir,(ClientData)NULL);
      
    /* For debugging only
       printf("Bound Pins on BOTTOM_SIDE\n");
       Mac_ItrBLBoundPinOfRCell(rCellPtr,Mac_PrintGPin,(ClientData)NULL);
       printf("Bound Pins on TOP_SIDE\n");
       Mac_ItrTRBoundPinOfRCell(rCellPtr,Mac_PrintGPin,(ClientData)NULL);
       */
 
    /* Define Side Pin */

    /* Determine pins at each edge of the channel */
    Tp_VertTileToRect(planeDataPtr,vertDomTile1Ptr,&dmntRect);
    blListPtr = Tm_PinsAroundVFace(cellMasterPtr, planeDataPtr, 
                       dmntRect, FALSE);
        blList = *blListPtr;

    Tp_VertTileToRect(planeDataPtr,vertDomTile2Ptr,&dmntRect);
    trListPtr = Tm_PinsAroundVFace(cellMasterPtr, planeDataPtr, 
                       dmntRect, FALSE);
        trList = *trListPtr;

        if (status == halfDead)
    {
        /* Append list collected by PinsAroundCircle */
        AppendLists(&tmpblList, &blList);
        CleanPinList(blList);
        AppendLists(&tmptrList, &trList);
        CleanPinList(trList);
    }

    
    /* Define Boundary Pins Of Channel */
        Gcd_DefineSidePin(planeDataPtr, junct2Ptr, &trList,
              rCellPtr, RIGHT_SIDE, WEST_PIN);
        Gcd_DefineSidePin(planeDataPtr, junct1Ptr, &blList,
              rCellPtr, TOP_LEFT_SIDE, EAST_PIN);
        
        /* remove "single terminal" entries from sidePinList */
        if (status != alive)
    {
        RemoveSingleSidePins(rCellPtr);
    }

    /* For debugging purposes only 
       printf("Side Pins on LEFT_SIDE\n");
       Mac_ItrTLSidePinOfRCell(rCellPtr,Mac_PrintRPin,(ClientData)NULL);
       printf("Side Pins on RIGHT_SIDE\n");
       Mac_ItrRSidePinOfRCell(rCellPtr,Mac_PrintRPin,(ClientData)NULL);
        */

    /* Route Channel */
    if (Rdb_MacToBBL(rCellPtr, &modulePtr) == BEAR_ERROR) {
        Mac_DeleteRCell(rCellPtr);
        return (BEAR_ERROR_2);
    }
    rCellPtr->routeModPtr = modulePtr;
    modulePtr->placg = 2;
    MALLOC(AdjPar *, adjParPtr, sizeof (AdjPar));
    if (Rdb_RouteChannelPass1(modulePtr, rCellPtr, adjParPtr) == BEAR_ERROR) {
        AbortRouteChannel(cellMasterPtr, planeDataPtr,
            rCellPtr, point1, point2, adjParPtr, dCellListsPtr, status);
        return (BEAR_ERROR_2);
    }
    heightRequired = Lr_GetChannelHeight(0);
    heightAvail = Lr_GetChannelHeight(1);
    adjParPtr->xAdj = adjParPtr->yAdj = heightRequired - heightAvail;

#ifndef BEAR_BLACK_BOX
    if (interactiveFlag == TRUE)  {
        heightTarget = heightAvail;
        IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
            "Route Channel", 13);
        (void) iv_AddVar(IVwin, "Available Channel Height", 
            &heightAvail, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightAvail, FALSE);
        (void) iv_AddVar(IVwin, "Required Channel Height", 
            &heightRequired, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightRequired, FALSE);
        (void) iv_AddVar(IVwin, "Target Channel Height",
            &heightTarget, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);

#ifdef USE_NUT_CODE
           (void) iv_AddVar(IVwin, "Use Nutcracker", 
              &nutcrackerFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
#endif USE_NUT_CODE

       (void) iv_AddVar(IVwin, "Via Reduction", 
            &viaReduceFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_AddVar(IVwin, "Route Channel", 
            &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_AddVar(IVwin, "Exit",
            &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);
       while (!doRoute) {
           XNextEvent(ez_Display, &theEvent);
           (void) iv_HandleEvent(&theEvent);;
           Ez_XEventHandler(&theEvent);
           if (exitFlag) {
               iv_FreeWindow(IVwin);
               AbortRouteChannel(cellMasterPtr, planeDataPtr,
                   rCellPtr, point1, point2, adjParPtr, dCellListsPtr, status);
               *deltaPtr = adjParPtr->yAdj;
               return (BEAR_ERROR);
           }
       }
       iv_FreeWindow(IVwin);
       Ez_EmptyXEventQue();
       doRoute = FALSE;
       exitFlag = FALSE;
    }
#endif BEAR_BLACK_BOX

#ifdef USE_NUT_CODE
    if (nutcrackerFlag) {
        if (heightAvail > heightRequired) {
            adjParPtr->xAdj = adjParPtr->yAdj = 0;
        }
        nut_compaction = Ls_nutcracker(modulePtr);
        heightRequired = MAX(heightRequired, heightAvail) - nut_compaction;

#ifndef BEAR_BLACK_BOX
        if (interactiveFlag == TRUE) {
        IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
            "Nutcracker", 13);
        (void) iv_AddVar(IVwin, "Available Channel Height", 
            &heightAvail, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightAvail, FALSE);
        (void) iv_AddVar(IVwin, "Required Channel Height", 
            &heightRequired, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightRequired, FALSE);
        (void) iv_AddVar(IVwin, "Target Channel Height", &heightTarget,
            IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Continue", 
            &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Exit",
            &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);
        adjParPtr->xAdj = adjParPtr->yAdj = heightRequired - heightAvail;
        while (!doRoute) {
            XNextEvent(ez_Display, &theEvent);
            (void) iv_HandleEvent(&theEvent);;
            Ez_XEventHandler(&theEvent);
            if (exitFlag) {
            iv_FreeWindow(IVwin);
            AbortRouteChannel(cellMasterPtr, planeDataPtr,
                rCellPtr, point1, point2, 
                adjParPtr, dCellListsPtr, status);
            *deltaPtr = adjParPtr->yAdj;
            return (BEAR_ERROR);
            }
        }
        iv_FreeWindow(IVwin);
        Ez_EmptyXEventQue();
        }
#endif BEAR_BLACK_BOX

        Ls_nutcrackerPass2(modulePtr, &compactPar,
            heightRequired, viaReduceFlag);
    }
    if (viaReduceFlag && nutcrackerFlag) {
        /* via reduction, if TRUE, is done in nutcracker module */
        Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr, compactPar,
                  viaReduceFlag);
    }
    else
#endif USE_NUT_CODE
    {
        Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr, compactPar,
            viaReduceFlag);
    }

    /* Abort the channel route if: */
    /* 1) not enough space (need to decompact) */
    /* 2) too much space and not forced to route (can attempt to compact) */
    if ((heightAvail < heightRequired) ||
        ((heightAvail > heightRequired) && (forceRouteFlag == FALSE)))
    {
        AbortRouteChannel(cellMasterPtr, planeDataPtr,
            rCellPtr, point1, point2, adjParPtr, dCellListsPtr, status);
        *deltaPtr = adjParPtr->yAdj;
        return (BEAR_ERROR);
    }

    /* Generate new Id */
    reply = debug ? GetReply("generate new id ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0) { 
        Tp_VertTileToRect(planeDataPtr,vertDomTile1Ptr,&dmntRect);
        pinListPtr = Tm_PinsAroundVFace(cellMasterPtr, planeDataPtr, 
            dmntRect, TRUE);
        if (status == halfDead) {
            blListPtr = Pb_PinsAroundDCells(cellMasterPtr, planeDataPtr, 
                Mac_GetTypeOfRCell(rCellPtr), dCellListsPtr, point1);
            AppendLists(blListPtr, pinListPtr);
        }
        Mac_ItrTLSidePinOfRCell(rCellPtr,GenNewId,(ClientData)pinListPtr);
        Tm_FreePinsAroundJunction(pinListPtr);
    }

    /* Reassign floor plan graph */
    llJunctPtr = Mac_GetFpgOfCell(cellMasterPtr);
    junct1Ptr = GetJunction(llJunctPtr, &point1);
    junct2Ptr = GetJunction(llJunctPtr, &point2);
    
    /* Create a new cell group */
    CreateNewCellGroup(planeDataPtr, rCellPtr);

    /* Update Net Id Of Side Pin */
    reply = debug ? GetReply("update net id of side pin ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
        Gcd_UpdateSidePinOfRCell(rCellPtr,TOP_LEFT_SIDE);
        Gcd_UpdateSidePinOfRCell(rCellPtr,RIGHT_SIDE);
    }

    /* Create Generic Cell */
    reply = debug ? GetReply("create generic cell ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
        MALLOC(GenericCell *,gCellPtr,sizeof (GenericCell));
        Mac_CreateGCell(ROUTE_CELL,rCellPtr,&gCellPtr);
        Tp_PaintGCell(gCellPtr,planeDataPtr);
        Tp_IdentifySpaceTiles(planeDataPtr);
        (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
        Mac_ItrBoundPinOfRCell(rCellPtr,MarkGPin,(ClientData)NULL);
        Mac_ItrSidePinOfRCell(rCellPtr,InsertRPin,(ClientData)NULL);
    }

    /* Vertical Face Split */
    reply = debug ? GetReply("split vertical face ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
        if (status == halfDead)
        {
        /* Need to split the circle - Tm_VCircleSplit */
        if (dCellListsPtr->BRptr)
        {
            dmntRect.r_xbot = dCellListsPtr->BRpoint->p_x;
            dmntRect.r_xtop = point2.p_x;
            dmntRect.r_ytop = dCellListsPtr->BRpoint->p_y;
            dmntRect.r_ybot = dmntRect.r_ytop - 1;
            Tp_HRectToVRect(&dmntRect, &vdmntRect);
            (void) Tm_VCSplit(cellMasterPtr, planeDataPtr, vdmntRect);
        }
        if (dCellListsPtr->BLptr)
        {
            dmntRect.r_xtop = dCellListsPtr->BLpoint->p_x;
            dmntRect.r_xbot = point1.p_x;
            dmntRect.r_ytop = dCellListsPtr->BLpoint->p_y;
            dmntRect.r_ybot = dmntRect.r_ytop - 1;
            Tp_HRectToVRect(&dmntRect, &vdmntRect);
            (void) Tm_VCSplit(cellMasterPtr, planeDataPtr, vdmntRect);
        }
        if (dCellListsPtr->TRptr)
        {
            dmntRect.r_xbot = dCellListsPtr->TRpoint->p_x;
            dmntRect.r_xtop = point2.p_x;
            dmntRect.r_ybot = dCellListsPtr->TRpoint->p_y;
            dmntRect.r_ytop = dmntRect.r_ybot + 1;
            Tp_HRectToVRect(&dmntRect, &vdmntRect);
            (void) Tm_VCSplit(cellMasterPtr, planeDataPtr, vdmntRect);
        }
        if (dCellListsPtr->TLptr)
        {
            dmntRect.r_xtop = dCellListsPtr->TLpoint->p_x;
            dmntRect.r_xbot = point1.p_x;
            dmntRect.r_ybot = dCellListsPtr->TLpoint->p_y;
            dmntRect.r_ytop = dmntRect.r_ybot + 1;
            Tp_HRectToVRect(&dmntRect, &vdmntRect);
            (void) Tm_VCSplit(cellMasterPtr, planeDataPtr, vdmntRect);
        }
        }
        Tm_InsDCellOnVTP(cellMasterPtr);
        Tm_VFSplit(cellMasterPtr);
    }
    /*
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
    */
    } 

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        /* Redraw the window with the route cell. */
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    }
#endif BEAR_BLACK_BOX

    *deltaPtr = 0;

    if (lr_singleTermFlag == TRUE) {
        PbPrint(EZ_ITALIC_FONT, "Channel has single terminal.\n");
        return (BEAR_ERROR_2);
    }
    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbRouteLChannel() --
 *
 * Results:
 *   returns either BEAR_OK, BEAR_ERROR, or BEAR_ERROR_2.
 *
 *   If BEAR_ERROR_2 is returned the routing process should
 *   be stopped.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbRouteLChannel(cellMasterPtr, junct1Ptr, junct2Ptr, junct3Ptr, 
    interactiveFlag, viaReduceFlag, debug)

CellMaster   *cellMasterPtr;
WallJunction *junct1Ptr;
WallJunction *junct2Ptr;
WallJunction *junct3Ptr;
bool         interactiveFlag;
bool         viaReduceFlag;
bool         debug;
{
    AdjPar              *adjParPtr;
    GenericCell         *gCellPtr;
    LinkedRect          *hRectList;
    LinkedRect          *vRectList;
    LinkedRect          *rectList;
    PinsAroundJunction  pinList;
    PinsAroundJunction  *faceListPtr;
    PlaneData           *planeDataPtr;
    Point               *point1Ptr; 
    Point               *point2Ptr; 
    Point               *point3Ptr;
    Point               point1; 
    Point               point2; 
    Point               point3;
    Point               vPoint;
    RCellType           rCellType;
    Rect                vDmntRect;
    Rect                hDmntRect;
    RouteCell           *rCellPtr;
    Tile                *horzDomTile2Ptr;
    Tile                *vertDomTile2Ptr;
    WallJunction        *llJunctPtr;
    WallSegment         *wallSegPtr;
    module              *modulePtr;
    bool                doRoute=FALSE;
    bool                exitFlag=FALSE;
    char                *reply;
    int                 compactPar = 0;
#ifndef BEAR_BLACK_BOX
    IVWindow            *IVwin;
    XEvent              theEvent;
#endif BEAR_BLACK_BOX

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    point1Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct1Ptr);
    point2Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct2Ptr);
    point3Ptr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct3Ptr);
    point1.p_x = point1Ptr->p_x;
    point1.p_y = point1Ptr->p_y;
    point2.p_x = point2Ptr->p_x;
    point2.p_y = point2Ptr->p_y;
    point3.p_x = point3Ptr->p_x;
    point3.p_y = point3Ptr->p_y;

    horzDomTile2Ptr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr),point2Ptr);
    Tp_HPointToVPoint(&point1, &vPoint);
    Tp_HPointToVPoint(&point2, &vPoint);
    vertDomTile2Ptr = TiSrPointNoHint(Tp_GetVertPlane(planeDataPtr),&vPoint);
    Tp_HPointToVPoint(&point3, &vPoint);

    (void) Mac_CreateRCell(cellMasterPtr, &rCellPtr);

    if (point1.p_x > point2.p_x && point3.p_y > point2.p_y)
    { 
    rCellType = TOP_RIGHT_L_RCELL;
    (void) Gcd_DefineTopRightLBound(cellMasterPtr, junct1Ptr, junct2Ptr,
                   junct3Ptr, horzDomTile2Ptr, vertDomTile2Ptr, &rCellPtr);
    }
    else if (point1.p_x > point2.p_x && point3.p_y < point2.p_y)
    { 
    rCellType = BOTTOM_RIGHT_RCELL;
    (void) Gcd_DefineBottomRightLBound(cellMasterPtr, junct1Ptr, junct2Ptr,
                   junct3Ptr, horzDomTile2Ptr, vertDomTile2Ptr, &rCellPtr);
    }
    else if (point1.p_x < point2.p_x && point3.p_y > point2.p_y)
    { 
    rCellType = TOP_LEFT_L_RCELL;
    (void) Gcd_DefineTopLeftLBound(cellMasterPtr, junct1Ptr, junct2Ptr,
                   junct3Ptr, horzDomTile2Ptr, vertDomTile2Ptr, &rCellPtr);
    }
    else
    { 
    rCellType = BOTTOM_LEFT_RCELL;
    (void) Gcd_DefineBottomLeftLBound(cellMasterPtr, junct1Ptr, junct2Ptr,
                   junct3Ptr, horzDomTile2Ptr, vertDomTile2Ptr, &rCellPtr);
    }
    (void) Mac_PutTypeOfRCell(rCellType, rCellPtr);
    (void) Mac_CreateBBoxOfRCell(rCellPtr);

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    /* Rect Unify */
#ifdef USE_WIRE_CODE
    if (Gre_DoPgWiringForAllNets(cellMasterPtr) == BEAR_ERROR) {
        return (BEAR_ERROR_2);
    }
#endif USE_WIRE_CODE
    Tp_VertTileToRect(planeDataPtr,vertDomTile2Ptr,&vDmntRect);
    Tp_HorzTileToRect(planeDataPtr,horzDomTile2Ptr,&hDmntRect);
    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == BOTTOM_RIGHT_RCELL)
    {
    hDmntRect.r_xbot = point2.p_x;
    hDmntRect.r_xtop = point1.p_x;
    }
    else
    {
    hDmntRect.r_xbot = point1.p_x;
    hDmntRect.r_xtop = point2.p_x;
    }
    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == TOP_LEFT_L_RCELL)
    {
    vDmntRect.r_ytop = point3.p_y;
    vDmntRect.r_ybot = point2.p_y;
    Tm_HRectUnify(cellMasterPtr, planeDataPtr, hDmntRect);
    Tm_VRectUnify(cellMasterPtr,planeDataPtr, vDmntRect);
    }
    else 
    {
    vDmntRect.r_ytop = point2.p_y;
    vDmntRect.r_ybot = point3.p_y;
    Tm_VRectUnify(cellMasterPtr,planeDataPtr, vDmntRect);
    Tm_HRectUnify(cellMasterPtr, planeDataPtr, hDmntRect);
    /* Unification of channel yields conflicting ids in corner */
    (void) Tm_PinsAroundJunction(cellMasterPtr,planeDataPtr, 
                     junct2Ptr, &pinList);
    faceListPtr = Tm_PinsAroundVFace(cellMasterPtr, planeDataPtr, 
                     vDmntRect, TRUE);
    (void) FixAllBadNets(pinList.horzPPinList, faceListPtr);
    }

    /* Define Bound Pin */
    Gcd_DefineBoundPin(planeDataPtr,rCellPtr);
    Mac_ItrBLBoundPinOfRCell(rCellPtr,ReverseDir,(ClientData)NULL);
    Mac_ItrTRBoundPinOfRCell(rCellPtr,ReverseDir,(ClientData)NULL);

    /* Define Side Pin */
    if (rCellType == TOP_RIGHT_L_RCELL)
    {
        PinsAroundVWallSect(cellMasterPtr, rCellPtr, junct1Ptr,
            TOP_RIGHT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct1Ptr, &pinList,
            rCellPtr, RIGHT_SIDE, WEST_PIN);
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
            TOP_RIGHT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct3Ptr, &pinList,
            rCellPtr, TOP_LEFT_SIDE, SOUTH_PIN);
    }
    else if (rCellType == TOP_LEFT_L_RCELL)
    {
        PinsAroundVWallSect(cellMasterPtr, rCellPtr, junct1Ptr,
            BOTTOM_LEFT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct1Ptr, &pinList,
            rCellPtr, TOP_LEFT_SIDE,EAST_PIN);
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
            TOP_RIGHT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct3Ptr, &pinList,
            rCellPtr, RIGHT_SIDE, SOUTH_PIN);
    }
    else if (rCellType == BOTTOM_LEFT_RCELL)
    {
        PinsAroundVWallSect(cellMasterPtr, rCellPtr, junct1Ptr,
            BOTTOM_LEFT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct1Ptr, &pinList,
            rCellPtr, RIGHT_SIDE,EAST_PIN);
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
            BOTTOM_LEFT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct3Ptr, &pinList,
            rCellPtr, TOP_LEFT_SIDE, NORTH_PIN);
    }
    else
    {
        PinsAroundVWallSect(cellMasterPtr, rCellPtr, junct1Ptr,
            TOP_RIGHT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct1Ptr, &pinList,
            rCellPtr, TOP_LEFT_SIDE,WEST_PIN);
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
            BOTTOM_LEFT_SIDE, &pinList);
        Gcd_DefineSidePin(planeDataPtr, junct3Ptr, &pinList,
            rCellPtr, RIGHT_SIDE, NORTH_PIN);
    }

    /* Route Channel */
    if (Rdb_MacToBBL(rCellPtr, &modulePtr) == BEAR_ERROR) {
        Mac_DeleteRCell(rCellPtr);
        return (BEAR_ERROR_2);
    }
    rCellPtr->routeModPtr = modulePtr;
    modulePtr->placg = 2;
    MALLOC(AdjPar *, adjParPtr, sizeof (AdjPar));
    if (Rdb_RouteChannelPass1(modulePtr, rCellPtr, adjParPtr) == BEAR_ERROR) {
        AbortLChannel(rCellPtr, junct1Ptr, junct2Ptr, junct3Ptr, adjParPtr);
        return (BEAR_ERROR_2);
    }

#ifndef BEAR_BLACK_BOX
    if (interactiveFlag == TRUE) {
       IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
            "Route L-Channel", 13);
       (void) iv_AddVar(IVwin, "X Adjustment",
            &(adjParPtr->xAdj), IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_SetEnableFlag(IVwin, &(adjParPtr->xAdj), FALSE);
       (void) iv_AddVar(IVwin, "Y Adjustment",
            &(adjParPtr->yAdj), IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_SetEnableFlag(IVwin, &(adjParPtr->yAdj), FALSE);
       (void) iv_AddVar(IVwin, "Via Reduction", 
            &viaReduceFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_AddVar(IVwin, "Route Channel", 
            &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
       (void) iv_AddVar(IVwin, "Exit",
            &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

       (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

       while (!doRoute) {
       XNextEvent(ez_Display, &theEvent);
       (void) iv_HandleEvent(&theEvent);;
       Ez_XEventHandler(&theEvent);
       if (exitFlag)
       {
           iv_FreeWindow(IVwin);
           AbortLChannel(rCellPtr, junct1Ptr, junct2Ptr, junct3Ptr, adjParPtr);
           return (BEAR_ERROR);
       }
       }
       iv_FreeWindow(IVwin);
       Ez_EmptyXEventQue();
    }
#endif BEAR_BLACK_BOX

    Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr,
        compactPar, viaReduceFlag);

    if (adjParPtr->yAdj > 0 || adjParPtr->xAdj > 0) 
    {
    AbortLChannel(rCellPtr, junct1Ptr, junct2Ptr, junct3Ptr, adjParPtr);
    return (BEAR_ERROR);
    }

    /* Generate new Id */
    reply = debug ? GetReply("generate new id ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    if (rCellType == TOP_RIGHT_L_RCELL)
    {
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
                TOP_RIGHT_SIDE, &pinList);
        Mac_ItrTLSidePinOfRCell(rCellPtr,GenNewId,(ClientData)&pinList);
    }
    else if (rCellType == TOP_LEFT_L_RCELL)
    {
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
                TOP_RIGHT_SIDE, &pinList);
        Mac_ItrRSidePinOfRCell(rCellPtr,GenNewId,(ClientData)&pinList);
    }
    else if (rCellType == BOTTOM_LEFT_RCELL)
    {
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
                BOTTOM_LEFT_SIDE, &pinList);
        Mac_ItrTLSidePinOfRCell(rCellPtr,GenNewId,(ClientData)&pinList);
    }
    else
    {
        PinsAroundHWallSect(cellMasterPtr, rCellPtr, junct3Ptr,
                BOTTOM_LEFT_SIDE, &pinList);
        Mac_ItrRSidePinOfRCell(rCellPtr,GenNewId,(ClientData)&pinList);
    }
    Tm_FreePinsAroundJunction(&pinList);
    }

    /* Create a new cell group */
    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == BOTTOM_RIGHT_RCELL)
    {
    wallSegPtr = Fpg_GetWestSegOfWallJunct(junct1Ptr);
    }
    else
    {
    wallSegPtr = Fpg_GetEastSegOfWallJunct(junct1Ptr);
    }
    LCreateNewCellGroup(planeDataPtr, rCellPtr, wallSegPtr);

    /* Horizontal Face Unify */
    reply = debug ? GetReply("unify horizontal face ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    Tm_HFOnlyUnify(cellMasterPtr, planeDataPtr);
    Tm_HBnUnify(cellMasterPtr);
    }

    /* Update Net Id Of Side Pin */
    reply = debug ? GetReply("update net id of TL side pin ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == BOTTOM_LEFT_RCELL)
    {
        Gcd_UpdateSidePinOfRCell(rCellPtr,TOP_LEFT_SIDE);
    }
    else
    {
        Gcd_UpdateSidePinOfRCell(rCellPtr,RIGHT_SIDE);
    }
    }

    /* Create & Paint Generic Cell */
    MALLOC(GenericCell *, gCellPtr, sizeof (GenericCell));
    Mac_CreateGCell(ROUTE_CELL, rCellPtr, &gCellPtr);

    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == TOP_LEFT_L_RCELL)
    {
    hRectList = Mac_GetHorzRectListOfGCell(gCellPtr)->r_next;
    vRectList = Mac_GetVertRectListOfGCell(gCellPtr);
    Mac_PutVertRectListOfGCell(Mac_GetHorzRectListOfGCell(gCellPtr), gCellPtr);
    Mac_GetHorzRectListOfGCell(gCellPtr)->r_next = NIL(LinkedRect);
    }
    else
    {
    hRectList = rectList = Mac_GetHorzRectListOfGCell(gCellPtr);
    vRectList = Mac_GetVertRectListOfGCell(gCellPtr);
    while (rectList->r_next->r_next != NIL(LinkedRect))
    {
        rectList = rectList->r_next;
    }
    Mac_PutHorzRectListOfGCell(rectList->r_next, gCellPtr);
    Mac_PutVertRectListOfGCell(rectList->r_next, gCellPtr);
    }

    reply = debug ? GetReply("paint generic cell (1 rect) ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    Tp_PaintGCell(gCellPtr,planeDataPtr);
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
    }
    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == TOP_LEFT_L_RCELL)
    {
    Mac_GetHorzRectListOfGCell(gCellPtr)->r_next = hRectList;
    Mac_PutVertRectListOfGCell(vRectList, gCellPtr);
    }
    else
    {
    Mac_PutHorzRectListOfGCell(hRectList, gCellPtr);
    Mac_PutVertRectListOfGCell(vRectList, gCellPtr);
    }

    /* Horizontal Face Split */
    reply = debug ? GetReply("split horizontal face ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    Tm_InsDCellOnHTP(cellMasterPtr);
    Tm_HFSplit(cellMasterPtr);
    }


    /* Vertical Face Unify */
    reply = debug ? GetReply("unify vertical face ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    Tm_VFOnlyUnify(cellMasterPtr, planeDataPtr);
    Tm_VBnUnify(cellMasterPtr);
    }

    /* Update Net Id Of Side Pin */
    reply = debug ? GetReply("update net id of R side pin ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    if (rCellType == TOP_RIGHT_L_RCELL || rCellType == BOTTOM_LEFT_RCELL)
    {
        Gcd_UpdateSidePinOfRCell(rCellPtr,RIGHT_SIDE);
    }
    else
    {
        Gcd_UpdateSidePinOfRCell(rCellPtr,TOP_LEFT_SIDE);
    }
    }

    /* Paint Generic Cell */
    reply = debug ? GetReply("paint generic cell ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    Tp_PaintGCellOnHorzPlane(gCellPtr,planeDataPtr);
    Tp_PaintGCellOnVertPlane(gCellPtr,planeDataPtr);
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
    Mac_ItrBoundPinOfRCell(rCellPtr,MarkGPin,(ClientData)NULL);
    Mac_ItrSidePinOfRCell(rCellPtr,InsertRPin,(ClientData)NULL);
    }

    /* Vertical Face Split */
    reply = debug ? GetReply("split vertical face ? (y/n): ") : "y";
    if (strcmp(reply,"y") == 0)
    {
    Tm_InsDCellOnVTP(cellMasterPtr);
    Tm_VFSplit(cellMasterPtr);
    }

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        /* Redraw the window with the route cell. */
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    }
#endif BEAR_BLACK_BOX

    if (lr_singleTermFlag == TRUE) {
        PbPrint(EZ_ITALIC_FONT, "L-channel has single terminal.\n");
        return (BEAR_ERROR_2);
    }
    return (BEAR_OK);
}


#endif /* USE_DR_CODE */

