
/* pbCellCmds.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbCellCmds.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "utility.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "globalSpacer.h"

#ifdef USE_WIRE_CODE
#include "grEditor.h"
#endif USE_WIRE_CODE

#include "polarBear.h"

#ifdef USE_XDM_CODE
#include "vemDM.h"
#endif USE_XDM_CODE


extern WindowData *globalWinDataPtr;

#ifdef USE_OLD_X_CODE
typedef struct _LocalData LocalData;
struct _LocalData {
    int     maxVertices;
    int     *vertexCountPtr;
    Vertex  *vertex;
    double  (*scrMatrix)[4];
};
#endif /* USE_OLD_X_CODE */


/* ---------- Local Procedures ---------- */
static void         UpdateShapeList();

#ifdef USE_OLD_X_CODE
static BearStatus   StoreVertex();
#endif /* USE_OLD_X_CODE */


#ifdef USE_OLD_X_CODE
static BearStatus
StoreVertex(lineSegPtr, dataPtr)
LineSeg     *lineSegPtr;
LocalData   *dataPtr;
{
    register int wx, wy, *countPtr;

    countPtr = dataPtr->vertexCountPtr;
    if (*countPtr < dataPtr->maxVertices) {
        wx = lineSegPtr->line.point1.p_x;
        wy = lineSegPtr->line.point1.p_y;
        dataPtr->vertex[*countPtr].x = Ez_XTrans(dataPtr->scrMatrix, wx, wy);
        dataPtr->vertex[*countPtr].y = Ez_YTrans(dataPtr->scrMatrix, wx, wy);
        (*countPtr)++;
        return (BEAR_CONTINUE_SEARCH);
    } 
    else {
        return (BEAR_STOP_SEARCH);
    }
}
#endif /* USE_OLD_X_CODE */



/* ARGSUSED */
void
PbCreateDCellsCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
/*
 * Create the dummy cell list and then redraw every window.
 */
{
    CellMaster   *cellMasterPtr;
    ChipWin      *chipWinPtr;
    PlaneData    *planeDataPtr;
    WallJunction *llJunctPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    if (Mac_IsNumOfCellsZero(cellMasterPtr) == TRUE) {
        Ez_RingBell();
        return;
    }
    if (Tp_GetDCellList(planeDataPtr) == NIL(DummyCell)) {
        BuildDCells(cellMasterPtr);
    } else {
        PbPrint(EZ_ITALIC_FONT,
            "Delete the old dummy cells and try again.\n");
        return;
    }
    (void) Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
}



/* ARGSUSED */
void
PbDeleteCellCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellInst    *cellInstPtr;
    CellMaster  *cellMasterPtr;
    ChipWin     *chipWinPtr;
    GenericCell *gcellPtr;
    PlaneData   *planeDataPtr;
    Tile        *tilePtr;
    bool        dummyCells;

    PbPrint(EZ_NORMAL_FONT, "click mouse (left button)\n");
    if (Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
        &winDataPtr, screenPtr) != EZ_OK) {
        PbPrint(EZ_ITALIC_FONT, "Must pick a point in a chip window.");
        return;
    }
    if (PbFindTile(winDataPtr, screenPtr, &tilePtr, NULL_TILE_PLANE) == TRUE) {

        chipWinPtr = PbGetChipWin(winDataPtr);
        cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
        planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

        if (Tp_IsCellInstTile(tilePtr)) {
            cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
            if (Tp_DeleteCell(cellInstPtr) == BEAR_ERROR) {
Bear_PrintErr("PbDeleteCellCmd() : deleting cell caused an error in data base.");
            }

            dummyCells = Tp_GetDCellList(planeDataPtr) != NIL(DummyCell);
            if (Fpg_IsFpgBuilt(planeDataPtr)) {
                /* Rebuild the floor plan graph. */
                BuildFloorPlan(cellMasterPtr, dummyCells);
            } else if (dummyCells) {
                BuildDCells(cellMasterPtr);
            } else {
                Tp_IdentifySpaceTiles(planeDataPtr);
            }

            /*
             * Check to see if they are deleting the cell instance
             * that is currently being high-lighted, and if they are
             * then stop high-lighting it.
             */
            gcellPtr = PbGetGCellOfCHW(chipWinPtr);
            if (gcellPtr->unType == CELL_INST) {
                if (gcellPtr->unValue.cellInstPtr == cellInstPtr) {
                    PbPutGCellOfCHW(NIL(GenericCell), chipWinPtr);
                }
            }

            /*
             * Set the pointer that keeps track of the pin that is being
             * high-lighted to NULL, because it might contain a pin from
             * this cell.
             */
            PbPutPinOfCHW(NIL(PinInst), chipWinPtr);

            PbPutUpdateStatusOfCHW(TRUE, chipWinPtr);
            (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
        } else {
            PbPrint(EZ_ITALIC_FONT, "You may only delete cell instances.\n");
            Ez_RingBell();
        }
    } else {
        Ez_RingBell();
    }
}



/* ARGSUSED */
void
PbDeleteDCellsCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellMaster  *cellMasterPtr;
    ChipWin     *chipWinPtr;
    PlaneData   *planeDataPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    if (Tp_GetDCellList(planeDataPtr) != NIL(DummyCell)) {
        Tp_DeleteDCells(planeDataPtr);

        if (Fpg_IsFpgBuilt(planeDataPtr)) {
            /* Rebuild the floor plan graph without the dummy cells. */
            BuildFloorPlan(cellMasterPtr, FALSE);
        }
        else {
            Tp_IdentifySpaceTiles(planeDataPtr);
        }
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    }
}



/* ARGSUSED */
void
PbCreateCellCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellMaster  *cellMasterPtr;
    ChipWin     *chipWinPtr;
    LinkedPoint *lpPtr, *pointList;
    PlaneData   *planeDataPtr;
    Point       origin;
    Rect        rect;
    int         i;
    char        *cellNamePtr= NIL(char);
    bool        dummyCells, floorPlan;
    bool        complexCell = FALSE;

    /* parse arguments */
    for (i = 1; i < argCount; i++) {
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        }

        if (strcmp (argList[i], "-c") == 0) {
            complexCell = TRUE;
            continue;
        } 
        else {
            cellNamePtr = argList[i];
            continue;
        }
    }

    if (cellNamePtr == NIL(char)) {
        goto usage;
    }

    if (complexCell) {
        /*
         * Create a complicated cell master, i.e. one whose boundary
         * has more then four sides.
         */

        if (Ez_PickPointList(winDataPtr, &pointList) == BEAR_OK) { 
            /*
             * The origin will be the lower left corner of the rectangle
             * that encloses all of the points.
             */
            if (Mac_FindBBoxForPointList(pointList, &rect) == BEAR_ERROR) {
                return;
            }
            origin.p_x = rect.r_xbot;
            origin.p_y = rect.r_ybot;

            /*
             * At the moment all of the points are absolute, so we
             * subtract 'origin' from each point in 'pointList' thus
             * making them relative to 'origin'.
             */
            for (lpPtr = pointList;
                lpPtr != NIL(LinkedPoint);
                lpPtr = lpPtr->nextPtr) {
                
                lpPtr->point.p_x -= origin.p_x;
                lpPtr->point.p_y -= origin.p_y;
            }
        } 
        else {
            return;
        }
    } 
    else {
        /*
         * Create a simple cell master, i.e. one with a rectangular shape.
         */
        if (Ez_PickRect(winDataPtr, &origin, &rect) == BEAR_OK) { 
            Mac_CreatePointListFromRect(&rect, &pointList);
        } 
        else {
            return;
        }
    }

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    dummyCells = Tp_GetDCellList(planeDataPtr) != NIL(DummyCell);
    floorPlan = Fpg_IsFpgBuilt(planeDataPtr);

    if (dummyCells) {
        Tp_DeleteDCells(planeDataPtr);
    }
    if (Tp_InsertCellMaster(cellMasterPtr,
        cellNamePtr, &origin, pointList) == BEAR_OK) {
        if (floorPlan) {
            /* Rebuild the floor plan graph. */
            BuildFloorPlan(cellMasterPtr, dummyCells);
        } 
        else if (dummyCells) {
            BuildDCells(cellMasterPtr);
        } 
        else {
            Tp_IdentifySpaceTiles(planeDataPtr);
        }
        PbPutUpdateStatusOfCHW(TRUE, chipWinPtr);
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    } 
    else {
        if (floorPlan) {
            /* Rebuild the floor plan graph. */
            BuildFloorPlan(cellMasterPtr, dummyCells);
        } 
        else if (dummyCells) {
            BuildDCells(cellMasterPtr);
        }
        PbPrint(EZ_ITALIC_FONT, "Cannot insert cell.\n");
    }
    Mac_FreePointList(pointList);
    return;

usage:
    PbPrint(EZ_NORMAL_FONT, "create cell\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: cc cellname [-c]\n");
    PbPrint(EZ_NORMAL_FONT, "options:\n");
    PbPrint(EZ_NORMAL_FONT, "    -c  create a complex cell\n");
}



static void
UpdateShapeList(cellInstPtr)
CellInst *cellInstPtr;
{
    CellMaster *cellMasterPtr;
    lsGen       gen;
    Shape       curShape, *shapePtr;
    int         lenBefore, lenAfter;
    int         w, h;
    bool        foundFlag;
    
    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);
    if (cellMasterPtr->shapeList == (lsList) NULL) {
        (void) sprintf (bear_ErrorMsg,
            "The shape list of the cell named \"%s\" is NULL.",
            Mac_GetNameOfCellInst(cellInstPtr));
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        return;
    }

    Mac_TidyUpShapeList(cellMasterPtr->shapeList);
    lenBefore = lsLength(cellMasterPtr->shapeList);

    (void) Mac_GetCurrentShape(cellInstPtr, &curShape);
    w = curShape.horzDim;
    h = curShape.vertDim;

    /* Look for the current shape. */
    foundFlag = FALSE;
    lsForeachItem(cellMasterPtr->shapeList, gen, shapePtr) {
        if ((shapePtr->horzDim == w) && (shapePtr->vertDim == h)) {
            foundFlag = TRUE;
            lsFinish(gen);
            break;
        }
    }
    if (foundFlag == FALSE) {
        shapePtr = ALLOC(Shape, 1);
        shapePtr->horzDim = w;
        shapePtr->vertDim = h;
        lsNewEnd(cellMasterPtr->shapeList, (lsGeneric) shapePtr, LS_NH);
    }

    /* Look for the rotated version of the current shape. */
    foundFlag = FALSE;
    lsForeachItem(cellMasterPtr->shapeList, gen, shapePtr) {
        if ((shapePtr->horzDim == h) && (shapePtr->vertDim == w)) {
            foundFlag = TRUE;
            lsFinish(gen);
            break;
        }
    }
    if (foundFlag == FALSE) {
        shapePtr = ALLOC(Shape, 1);
        shapePtr->horzDim = h;
        shapePtr->vertDim = w;
        lsNewEnd(cellMasterPtr->shapeList, (lsGeneric) shapePtr, LS_NH);
    }

    Mac_TidyUpShapeList(cellMasterPtr->shapeList);
    lenAfter = lsLength(cellMasterPtr->shapeList);

    (void) sprintf (bear_Msg,
        "%d shape(s) were added to this cells shape list.\n",
        (lenAfter - lenBefore));
    PbPrint(EZ_NORMAL_FONT, bear_Msg);
}



#ifdef USE_XDM_CODE
static dmWhichItem tcDialogPtr[] = {
    {"Mirror", 0, 0},
    {"Move", 0, 0},
    {"Reflect", 0, 0},
    {"Resize", 0, 0},
    {"Rotate", 0, 0}};
#endif USE_XDM_CODE


/* ARGSUSED */
void
PbTransformCellCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
/*
 * Moving a cell involves the following 5 step procedure:
 *
 *   1.) Place mouse inside the cell you want to move.
 *   2.) Type the name of the move command in the console
 *       window and press return.
 *   3.) Press the left mouse button.
 *   4.) Move the cell (while holding the left mouse button down).
 *   5.) Release the left mouse button.
 */
{
#define MAX_VERTICES      25     /* Must be at least five so that boxes
                                  * may be drawn.
                                  */
#ifdef USE_OLD_X_CODE
    LocalData       localData;
    Vertex          vertex[MAX_VERTICES], *vertexArray;
    int             vertexCount;
#endif /* USE_OLD_X_CODE */

    CellInst        *cellInstPtr;
    CellMaster      *coreMasterPtr;
    ChipWin         *chipWinPtr;
    ezStatus        findStatus;
    GenericCell     gcell;
    Pb_DefCurData   defCurData;
    PlaneData       *planeDataPtr;
    Point           moveDelta, resizeDelta;
    Rect            bndBox;
    Tile            *tilePtr = NIL(Tile);
    WallJunction    *llJunctPtr;
    double          (*scrMatrix)[4];
    int             selected = 0;
    int             notSelected = 0;
    int             atoi ();
    int             i;
    int             llx, lly, urx, ury;    
    bool            dummyCells, floorPlan;
    bool            needDeltaFlag = FALSE;
    bool            moveFlag = FALSE;
    bool            resizeFlag = FALSE;
    bool            interactiveFlag = FALSE;              
    int             mirror = -1;
    int             rotateAngle = -1;
    int             reflect = -1;
    
    scrMatrix = winDataPtr->wd_scrMatrix;
    moveDelta.p_x = moveDelta.p_y = 0;
    resizeDelta.p_x = resizeDelta.p_y = 0;

#ifdef USE_OLD_X_CODE
    bzero((char *) vertex, sizeof (*vertex) * MAX_VERTICES);
    vertex[0].flags = VertexDontDraw;
#endif /* USE_OLD_X_CODE */
    
#ifdef USE_XDM_CODE
    if (argCount == 1) {
        dmWhichOne("Transform Cell",
               sizeof (tcDialogPtr)/sizeof (dmWhichItem),
               tcDialogPtr, &selected, 0, 0);
        if (!(selected == -1)) {
            notSelected = PbTcDialog(selected,argList,&argCount);
            if (!notSelected) {
            return; /* Aborted dialog. */
            }
        }
        else {
            return;     /* Aborted dialog. */
        }
    }
#endif USE_XDM_CODE

    /* parse arguments */
    i = 1;
    if (strcmp (argList[i], "-help") == 0) {
        goto usage; 
    }
    else if (strcmp (argList[i], "-dx") == 0) {
        if (++i >= argCount) {
            goto usage;
        }
        moveDelta.p_x = atoi (argList[i++]);
        moveFlag = TRUE;
        if (i < argCount) {
            if (strcmp (argList[i], "-dy") == 0) {
                if (++i >= argCount) {
                    goto usage;
                }
                moveDelta.p_y = atoi (argList[i]);
            }
        }
    }

#ifdef USE_OLD_X_CODE
    else if (strcmp (argList[i], "-im") == 0) {
        interactiveFlag = TRUE;
        needDeltaFlag = TRUE;
        moveFlag = TRUE;
    }
    else if (strcmp (argList[i], "-ir") == 0) {
        interactiveFlag = TRUE;
        resizeFlag = TRUE;
    }
#endif /* USE_OLD_X_CODE */

    else if (strcmp (argList[i], "-dy") == 0) {
        if (++i >= argCount) {
            goto usage;
        }
        moveDelta.p_y = atoi (argList[i++]);
        moveFlag = TRUE;
        if (i < argCount) {
            if (strcmp (argList[i], "-dx") == 0) {
                if (++i >= argCount) {
                    goto usage;
                }
                moveDelta.p_x = atoi (argList[i++]);
            }
        }
    }
    else if (strcmp (argList[i], "-my") == 0) {
        mirror = MIRROR_X;
        rotateAngle = -1;
        reflect = -1;
    }
    else if (strcmp (argList[i], "-mx") == 0) {
        mirror = MIRROR_Y;
        rotateAngle = -1;
        reflect = -1;
    }
    else if (strcmp (argList[i], "-rt") == 0) {
        if (++i >= argCount) {
            goto usage;
        }
        rotateAngle = atof (argList[i]);
        if (rotateAngle % 90 != 0) {
            goto usage;
        }
        rotateAngle = rotateAngle % 360;
        mirror = -1;
        reflect = -1;
    }
    else if (strcmp (argList[i], "-rf") == 0) {
        if (++i >= argCount) {
            goto usage;
        }
        reflect = atof (argList[i]);
        if (reflect != 45 && reflect != 135) {
            goto usage;
        }
        mirror = -1;
        rotateAngle = -1;
    }
    else if (strncmp (argList[i], "-rs", 3) == 0) {
        if ((i + 1) >= argCount) {
            goto usage;
        }
        for (i = 1; i < argCount; i++) {
            if (strcmp (argList[i], "-rsx") == 0) {
                resizeDelta.p_x = atoi (argList[++i]);
                resizeFlag = TRUE;
                continue;
            }

            if (strcmp (argList[i], "-rsy") == 0) {
                if (++i >= argCount) {
                    goto usage;
                }
                resizeDelta.p_y = atoi (argList[i]);
                resizeFlag = TRUE;
                continue;
            }

            if (strcmp (argList[i], "UL") == 0) {
                moveDelta.p_x = -resizeDelta.p_x;
                break;
            }
            if (strcmp (argList[i], "LL") == 0) {
                moveDelta.p_x = -resizeDelta.p_x;
                moveDelta.p_y = -resizeDelta.p_y;
                break;
            }
            if (strcmp (argList[i], "UR") == 0) {
                moveDelta.p_x = 0;
                moveDelta.p_y = 0;
                break;
            }
            if (strcmp (argList[i], "LR") == 0) {
                moveDelta.p_y = -resizeDelta.p_y;
                break;
            }

            goto usage; /* Invalid resize option. */
        }
    }
    else {
        goto usage;
    }
    
    chipWinPtr = PbGetChipWin(winDataPtr);
               
    /*
     * Change the cursor to a circle in all of the chip windows.
     */
    defCurData.idWinProc = PbIsChipWin;
    defCurData.cursorType = (int) EZ_CIRCLE_OUTLINE_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();
                  
    /*
     * Wait for the user to click the mouse in a chip window, which
     * will indicate the location of the square.
     */
    findStatus = Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
        &winDataPtr, screenPtr);
                  
    /*
     * Change the cursor back to a cross in all of the chip windows.
     */
    defCurData.cursorType = (int) EZ_CROSS_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();
                  
    if (findStatus != EZ_OK) {
        PbPrint(EZ_ITALIC_FONT, "must pick a point in a chip window\n");
        return;
    }

    if (! PbFindTile(winDataPtr, screenPtr, &tilePtr, NULL_TILE_PLANE)) {
        Ez_RingBell();
        return;
    }
    else if (! Tp_IsCellInstTile(tilePtr)) {
        PbPrint(EZ_ITALIC_FONT, "Point must be within a cell instance.\n");
        Ez_RingBell();
        return;
    } 
    
    cellInstPtr = Tp_GetCellInstOfTile(tilePtr);

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;
    coreMasterPtr = Mac_GetParentOfCell(cellInstPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(coreMasterPtr);

    if (resizeFlag == TRUE) {
        Point   llwPoint, urwPoint;
        Rect    absRect;
        
        /* 
         * Only rectangular cells may be resized.
         */
        if (Mac_GetAbsRectOfCellInst(cellInstPtr, &absRect) == BEAR_ERROR) {
            PbPrint(EZ_ITALIC_FONT, "You cannot resize a complex cell.\n");
            return;
        }

#ifdef USE_OLD_X_CODE
        if (interactiveFlag == TRUE) {
            /*
             * Convert the cell rectangle to screen coordintaes.
             */
            llwPoint.p_x = Ez_XTrans(scrMatrix, absRect.r_xbot, absRect.r_ybot);
            llwPoint.p_y = Ez_YTrans(scrMatrix, absRect.r_xbot, absRect.r_ybot);
            urwPoint.p_x = Ez_XTrans(scrMatrix, absRect.r_xtop, absRect.r_ytop);
            urwPoint.p_y = Ez_YTrans(scrMatrix, absRect.r_xtop, absRect.r_ytop);
            
            if (Ez_ResizeRect(winDataPtr, &llwPoint, &urwPoint, screenPtr,
                &resizeDelta, &moveDelta) == BEAR_ERROR) {
                return;
            }
        }
#endif /* USE_OLD_X_CODE */

        (void) Tp_ResizeCellInst(moveDelta.p_x, moveDelta.p_y,
            resizeDelta.p_x, resizeDelta.p_y, cellInstPtr);

        UpdateShapeList(cellInstPtr);

        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

        return;
    }
                  
    if (mirror != -1) {
        Tp_UnpaintGCell(&gcell, planeDataPtr);
        Mac_MirrorCell(mirror,cellInstPtr);
        Tp_PaintGCell(&gcell, planeDataPtr);
        (void) Tp_IdentifySpaceTiles(planeDataPtr);
        (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
        return;
    }

    if (rotateAngle != -1) {
        if (Mac_IsGlobalRoutingDone(coreMasterPtr) == FALSE) {
            if (Tp_RotateCell(coreMasterPtr, cellInstPtr, rotateAngle) == 
                BEAR_FALSE) {
                PbPrint(EZ_ITALIC_FONT, "Area under the cell is not clear.\n");
                return;
            }
            else {
                (void) Tp_IdentifySpaceTiles(planeDataPtr);
                (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
                return;
            }
        }
        else {
            PbPrint(EZ_ITALIC_FONT, 
                "You may not rotate a cell after global routing is done.\n");
            return;
        }
    }

    if (reflect != -1) {
        if (Mac_IsGlobalRoutingDone(coreMasterPtr) == FALSE) {
            if (Tp_ReflectCell(coreMasterPtr, cellInstPtr, reflect) == 
                BEAR_FALSE) {
                PbPrint(EZ_ITALIC_FONT, "Area under the cell is not clear.\n");
                return;
            }
            else {
                (void) Tp_IdentifySpaceTiles(planeDataPtr);
                (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
                return;
            }
        }
        else {
            PbPrint(EZ_ITALIC_FONT,
                "You may not reflect a cell after global routing is done.\n");
            return;
        }
    }

#ifdef USE_OLD_X_CODE
    if (needDeltaFlag) {
        ezStatus dragStatus;

        if (PbGetOutlineVert(cellInstPtr, scrMatrix,
            &vertexArray, &vertexCount) != BEAR_OK) {

            vertexCount = 0;
            localData.vertexCountPtr = &vertexCount;
            localData.maxVertices = MAX_VERTICES;
            localData.vertex = vertex;
            localData.scrMatrix = scrMatrix;
            if ((Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr), 
                StoreVertex, (ClientData) &localData) == 
                BEAR_COMPLETE_SEARCH) &&

                (vertexCount < MAX_VERTICES)) {

                vertex[vertexCount].x = vertex[0].x;
                vertex[vertexCount].y = vertex[0].y;
                vertexCount++;
            }
            else {
                /*
                 * Convert the cells bounding box to screen coordintaes.
                 */

                GeoTransRect(Mac_GetTransformOfGCell(&gcell),
                    Mac_GetBBoxOfCell(cellInstPtr), &bndBox);
                llx = Ez_XTrans(scrMatrix, bndBox.r_xbot, bndBox.r_ybot);
                lly = Ez_YTrans(scrMatrix, bndBox.r_xbot, bndBox.r_ybot);
                urx = Ez_XTrans(scrMatrix, bndBox.r_xtop, bndBox.r_ytop);
                ury = Ez_YTrans(scrMatrix, bndBox.r_xtop, bndBox.r_ytop);
                Ez_StoreBox(vertex, vertexCount, llx, lly, urx, ury);
            }
            dragStatus = Ez_DragBoundary(winDataPtr, screenPtr, 
                vertex, vertexCount, &moveDelta);
        }
        else {
            vertexArray[0].flags = VertexDontDraw;
            dragStatus = Ez_DragBoundary(winDataPtr, screenPtr, 
                vertexArray, vertexCount, &moveDelta);
        }

        if (dragStatus != EZ_OK) {
            return;
        }
    }
#endif /* USE_OLD_X_CODE */

    if (moveFlag == TRUE) {
        coreMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
        planeDataPtr = Mac_GetCellPlaneOfCell(coreMasterPtr);
        dummyCells = Tp_GetDCellList(planeDataPtr) != NIL(DummyCell);
        floorPlan = Fpg_IsFpgBuilt(planeDataPtr);

        /*
        if (Tp_MoveCellInst(cellInstPtr, &moveDelta) == BEAR_OK)
        */
        if (Tp_MoveCell(&gcell, &moveDelta) == BEAR_OK) {
            if (Mac_IsGlobalRoutingDone(coreMasterPtr) == FALSE) {
               if (floorPlan) {
                   BuildFloorPlan(coreMasterPtr, dummyCells);
               }    
               else if (dummyCells) {
                   BuildDCells(coreMasterPtr);
               }
               else {
                   Tp_IdentifySpaceTiles(planeDataPtr);
               }
            }

            PbPutUpdateStatusOfCHW(TRUE, chipWinPtr);
            (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

            return;
        }
        else {
            (void) sprintf (bear_ErrorMsg, "%s%s",
                 "The cell must stay within the boundary and\n",
                 "the area under the cell must be free.\n");
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

            return;
        }
    }

    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

usage:
    PbPrint(EZ_NORMAL_FONT,
        "Usage: tc [[-dx #] [-dy #]] %s\n",
#ifdef USE_OLD_X_CODE
        "[-im]"
#else
        ""
#endif /* USE_OLD_X_CODE */
        );

    PbPrint(EZ_NORMAL_FONT,
        "          [[-rsx #] [-rsy #] [<UL,LL,UR,LR>]] %s\n",
#ifdef USE_OLD_X_CODE
        "[-ir]"
#else
        ""
#endif /* USE_OLD_X_CODE */
        );

    PbPrint(EZ_NORMAL_FONT,
        "          [-mx] [-my] [-rt <X90>] [-rf <45,135>]\n");

    PbPrint(EZ_NORMAL_FONT, "\nwhere options include:\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -dx #             move cell in x direction\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -dy #             move cell in y direction\n");
#ifdef USE_OLD_X_CODE
    PbPrint(EZ_NORMAL_FONT,
        "    -im               interactive move\n");
#endif /* USE_OLD_X_CODE */

    PbPrint(EZ_NORMAL_FONT, 
        "    -rsx #            resize cell in x direction\n");
    PbPrint(EZ_NORMAL_FONT, 
        "    -rsy #            resize cell in y direction\n");
#ifdef USE_OLD_X_CODE
    PbPrint(EZ_NORMAL_FONT, 
        "    -ir               interactive resize\n");
#endif /* USE_OLD_X_CODE */

    PbPrint(EZ_NORMAL_FONT,
        "    -mx               mirror cell across x-axis\n");
    PbPrint(EZ_NORMAL_FONT, 
        "    -my               mirror cell across y-axis\n");
    PbPrint(EZ_NORMAL_FONT, 
        "    -rt #             rotate cell by multiples of 90 degrees\n");
    PbPrint(EZ_NORMAL_FONT, 
        "    -rf #             reflect cell across the line y = x or\n");
    PbPrint(EZ_NORMAL_FONT, 
        "                      the line y = -x\n");
}



#ifdef USE_OLD_X_CODE
/* ARGSUSED */
void
PbDuplicateCellCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
#define MAX_VERTICES      25     /* Must be at least five so that boxes
                                  * may be drawn.
                                  */
    CellInst        *cellInstPtr;
    CellMaster      *cellMasterPtr;
    ChipWin         *chipWinPtr;
    ezStatus        findStatus;
    GenericCell     gcell;
    Pb_DefCurData   defCurData;
    LocalData       localData;
    PlaneData       *planeDataPtr;
    Point           moveDelta;
    Rect            bndBox;
    Tile            *tilePtr = NIL(Tile);
    Vertex          vertex[MAX_VERTICES];
    double          (*scrMatrix)[4];
    int             vertexCount;
    int             llx, lly, urx, ury;     /* Screen coordinates of the
                                             * bounding box.
                                             */
    bool            dummyCells, floorPlan;

    moveDelta.p_x = moveDelta.p_y = 0;
    scrMatrix = winDataPtr->wd_scrMatrix;
    bzero((char *) vertex, sizeof (*vertex) * MAX_VERTICES);
    vertex[0].flags = VertexDontDraw;
    
    chipWinPtr = PbGetChipWin(winDataPtr);
    PbPrint(EZ_NORMAL_FONT, "pick a cell.\n");
               
    /*
     * Change the cursor to a circle in all of the chip windows.
     */
    defCurData.idWinProc = PbIsChipWin;
    defCurData.cursorType = (int) EZ_CIRCLE_OUTLINE_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();

    /*
     * Wait for the user to click the mouse in a chip window, which
     * will indicate the location of the square.
     */
    findStatus = Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
        &winDataPtr, screenPtr);
                  
    /*
     * Change the cursor back to a cross in all the chip windows.
     */
    defCurData.cursorType = (int) EZ_CROSS_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();
                  
    if (findStatus != EZ_OK) {
        PbPrint(EZ_ITALIC_FONT, "must pick a point in a chip window\n");
        return;
    }

    if (PbFindTile(winDataPtr, screenPtr,
        &tilePtr, NULL_TILE_PLANE) && Tp_IsCellInstTile(tilePtr)) {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
    } 
    else {
        PbPrint(EZ_ITALIC_FONT, "You may only duplicate cell instances.\n");
        Ez_RingBell();
        return;
    }
    
    vertexCount = 0;
    localData.vertexCountPtr = &vertexCount;
    localData.maxVertices = MAX_VERTICES;
    localData.vertex = vertex;
    localData.scrMatrix = scrMatrix;
    if ((Mac_ItrLineSeg(Mac_GetBoundaryOfCell(cellInstPtr), StoreVertex,
        (ClientData) &localData) == BEAR_COMPLETE_SEARCH) &&

        (vertexCount < MAX_VERTICES)) {

        vertex[vertexCount].x = vertex[0].x;
        vertex[vertexCount].y = vertex[0].y;
        vertexCount++;
    }
    else {
    /*
     * Convert the cells bounding box to screen coordintaes.
     */

        gcell.unType = CELL_INST;
        gcell.unValue.cellInstPtr = cellInstPtr;

        GeoTransRect(Mac_GetTransformOfGCell(&gcell),
            Mac_GetBBoxOfCell(cellInstPtr), &bndBox);
        llx = Ez_XTrans(scrMatrix, bndBox.r_xbot, bndBox.r_ybot);
        lly = Ez_YTrans(scrMatrix, bndBox.r_xbot, bndBox.r_ybot);
        urx = Ez_XTrans(scrMatrix, bndBox.r_xtop, bndBox.r_ytop);
        ury = Ez_YTrans(scrMatrix, bndBox.r_xtop, bndBox.r_ytop);
        Ez_StoreBox(vertex, vertexCount, llx, lly, urx, ury);
    }
       
    PbPrint(EZ_NORMAL_FONT, "with the mouse(left button) down,\n");
    PbPrint(EZ_NORMAL_FONT, "move the cell to a new location.\n");

    if (Ez_DragBoundary(winDataPtr, screenPtr, vertex,
        vertexCount, &moveDelta) != EZ_OK) {
        return;
    }

    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    dummyCells = Tp_GetDCellList(planeDataPtr) != NIL(DummyCell);
    floorPlan = Fpg_IsFpgBuilt(planeDataPtr);

    if (dummyCells) {
        Tp_DeleteDCells(planeDataPtr);
    }
    if (Tp_DuplicateCellInst(cellInstPtr, &moveDelta) == BEAR_OK) {
        if (floorPlan) {
            /* Rebuild the floor plan graph. */
            BuildFloorPlan(cellMasterPtr, dummyCells);
        } 
        else if (dummyCells) {
            BuildDCells(cellMasterPtr);
        } 
        else {
            Tp_IdentifySpaceTiles(planeDataPtr);
        }
        PbPutUpdateStatusOfCHW(TRUE, chipWinPtr);
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    } 
    else {
        if (floorPlan) {
            /* Rebuild the floor plan graph. */
            BuildFloorPlan(cellMasterPtr, dummyCells);
        } 
        else if (dummyCells) {
            BuildDCells(cellMasterPtr);
        }
        PbPrint(EZ_ITALIC_FONT, "can not duplicate cell\n");
    }

    return;
}
#endif /* USE_OLD_X_CODE */



#ifdef USE_OLD_X_CODE
/* ARGSUSED */
void
PbResizeParentCellCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellInst        *coreInstPtr;
    CellMaster      *coreMasterPtr;
    ChipWin         *chipWinPtr;
    ezStatus        findStatus;
    GenericCell     gcell;
    Pb_DefCurData   defCurData;
    PlaneData       *planeDataPtr;
    WallJunction    *llJunctPtr;
    Point           moveDelta, resizeDelta, urwPoint, llwPoint;
    Rect            bndBox;
    double          (*scrMatrix)[4];
    int             i;
    
    if (argCount > 1) {
        goto usage;
    }

    globalWinDataPtr = winDataPtr;
    moveDelta.p_x = moveDelta.p_y = 0;
    resizeDelta.p_x = resizeDelta.p_y = 0;
    scrMatrix = winDataPtr->wd_scrMatrix;
    
    chipWinPtr = PbGetChipWin(winDataPtr);
    coreMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    coreInstPtr = Mac_GetInstOfCoreMaster(coreMasterPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(coreMasterPtr);

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = coreInstPtr;

    /*
     * Change the cursor to a circle in all of the chip windows.
     */
    defCurData.idWinProc = PbIsChipWin;
    defCurData.cursorType = (int) EZ_CIRCLE_OUTLINE_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();
                  
    /*
     * Wait for the user to click the mouse in a chip window, which
     * will indicate the location of the square.
     */
    findStatus = Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
        &winDataPtr, screenPtr);
                  
    /*
     * Change the cursor back to a cross in all of the chip windows.
     */
    defCurData.cursorType = (int) EZ_CROSS_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();
                  
    if (findStatus != EZ_OK) {
        PbPrint(EZ_ITALIC_FONT, "must pick a point in a chip window\n");
        return;
    }

    /*
     * Convert the cells bounding box to screen coordintaes.
     */
    
    GeoTransRect(Mac_GetTransformOfGCell(&gcell),
        Mac_GetBBoxOfCell(coreInstPtr), &bndBox);
    
    llwPoint.p_x = Ez_XTrans(scrMatrix, bndBox.r_xbot, bndBox.r_ybot);
    llwPoint.p_y = Ez_YTrans(scrMatrix, bndBox.r_xbot, bndBox.r_ybot);
    urwPoint.p_x = Ez_XTrans(scrMatrix, bndBox.r_xtop, bndBox.r_ytop);
    urwPoint.p_y = Ez_YTrans(scrMatrix, bndBox.r_xtop, bndBox.r_ytop);
    
    if (Ez_ResizeRect(winDataPtr, &llwPoint, &urwPoint, screenPtr,
        &resizeDelta, &moveDelta) == BEAR_ERROR) {
        return;
    }
    (void) Tp_ResizeParentCell(moveDelta.p_x, moveDelta.p_y, resizeDelta.p_x,
        resizeDelta.p_y, coreInstPtr, planeDataPtr);
    (void) Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
#ifdef USE_WIRE_CODE
    (void) Gre_DoPgWiringForAllNets(coreMasterPtr);
#endif USE_WIRE_CODE
    (void) Gs_CalcCapDenOfTilePlane(coreMasterPtr);

    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

usage:
    PbPrint(EZ_NORMAL_FONT, "resize parent cell\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: rp\n");
}
#endif /* USE_OLD_X_CODE */


