
/* pbCmdList.c --
 *
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbCmdList.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "deque.h"
#include "polarBear.h"

/*
 * Only one procedure may be associated with each command as follows:
 *
 * if (pbCommand[i].co_propList & (PB_SCRIPT_CP | PB_BLACK_BOX_CP)) then
 *     pbCommand[i].co_statusProc must be filled in
 * else 
 *     pbCommand[i].co_voidProc must be filled in
 *
 * WARNING - Be carefull when changing the "long name" of a command because
 * some of the code is dependant on these names.  For example the code in
 * pbWinCmds.c that handles the creation of new chip windows scans for
 * commands by there "long names".
 *   
 */

Command pbCommand[] = {

#if defined(DEBUG)
    { "crash", "Crash",
      (void (*)()) NULL, 
      PbCrashProgramCmd,
      PB_ANY_WIN,
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "D", "Debug",
      (void (*)()) NULL, 
      PbDebugCmd,
      PB_ANY_WIN,
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#if !defined(BEAR_BLACK_BOX)
    { "t", "Test",
      PbTestCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN,
      PB_MOUSE_LOC_CP
    },
#endif /* BEAR_BLACK_BOX */
#endif /* DEBUG */

/*
 * Commands that may be used in any window your heart desires.
 */

    { "log", "LogConsole",
      (void (*)()) NULL, 
      PbLogConsoleCmd, 
      PB_ANY_WIN,
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "v", "Version",
      (void (*)()) NULL, 
      PbVersionCmd, 
      PB_ANY_WIN,
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "help", "Help",
      (void (*)()) NULL, 
      PbHelpCmd, 
      PB_ANY_WIN,
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#if !defined(BEAR_BLACK_BOX)
    { "so", "Source",
      PbSourceCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN,
      PB_NULL_CP
    },
#endif /* BEAR_BLACK_BOX */

    { "mail", "Mail",         
      (void (*)()) NULL, 
      PbMailCmd, 
      PB_ANY_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#if !defined(BEAR_BLACK_BOX)
    { "ow", "OpenWin",
      PbOpenWindowCmd, 
      (BearStatus (*)()) NULL, 
      PB_ANY_WIN,
      PB_MOUSE_LOC_CP
    },

    { "cw", "CloseWin",
      PbCloseWindowCmd, 
      (BearStatus (*)()) NULL, 
      PB_ANY_WIN,
      PB_MOUSE_LOC_CP
    },
#endif /* BEAR_BLACK_BOX */

    { "g", "Generic",
      (void (*)()) NULL, 
      PbGenericCmd, 
      PB_ANY_WIN,
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#if !defined(BEAR_BLACK_BOX) /* { */
    { "CC", "ClearConsole",
      PbClearConsoleCmd, 
      (BearStatus (*)()) NULL, 
      PB_ANY_WIN,
      PB_MOUSE_LOC_CP
    },

/*
 * Commands that may be used in cif and chip windows.
 */
    { "c", "Center",
      PbCenterCmd,
      (BearStatus (*)()) NULL, 
      (PB_CHIP_WIN | PB_CIF_WIN),
      PB_MOUSE_LOC_CP
    },

    { "mag", "MagArea",
      PbMagAreaCmd,
      (BearStatus (*)()) NULL, 
      (PB_CHIP_WIN | PB_CIF_WIN),
      PB_MOUSE_LOC_CP
    },

    { "P", "Pan",
      PbPanCmd,
      (BearStatus (*)()) NULL, 
      (PB_CHIP_WIN | PB_CIF_WIN),
      PB_MOUSE_LOC_CP
    },

    { "Z", "ZoomIn",
      PbZoomInCmd,
      (BearStatus (*)()) NULL, 
#if defined(USE_OLDP_CODE)
      (PB_CHIP_WIN | PB_CIF_WIN | PB_FLOOR_WIN | PB_TREE_WIN),
#else
      (PB_CHIP_WIN | PB_CIF_WIN),
#endif /* USE_OLDP_CODE */
      PB_NULL_CP
    },

    { "z", "ZoomOut",
      PbZoomOutCmd,
      (BearStatus (*)()) NULL, 
#if defined(USE_OLDP_CODE)
      (PB_CHIP_WIN | PB_CIF_WIN | PB_FLOOR_WIN | PB_TREE_WIN),
#else
      (PB_CHIP_WIN | PB_CIF_WIN),
#endif /* USE_OLDP_CODE */
      PB_NULL_CP
    },

/*
 * Commands that may be used in chip windows.
 */
#ifdef USE_XDM_CODE
    { "cr", "Create",
      PbCreateDialog, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "d", "Delete",
      PbDeleteDialog, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "du",     "Duplicate",
      PbDuplicateDialog, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },
#endif /* USE_XDM_CODE */

    { "cd", "CreateDCells",
      PbCreateDCellsCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "dd", "DeleteDCells",
      PbDeleteDCellsCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "cc", "CreateCell",
      PbCreateCellCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "dc", "DeleteCell",
      PbDeleteCellCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "tc", "TransformCell",
      PbTransformCellCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

#ifdef USE_OLD_X_CODE
    { "duc", "DuplicateCell",
      PbDuplicateCellCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "rp", "ResizeParentCell",
      PbResizeParentCellCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },
#endif /* USE_OLD_X_CODE */
#endif /* } BEAR_BLACK_BOX */

#if defined(USE_OLDP_CODE)
    { "ncl", "NewCluster",
      (void (*)()) NULL, 
      PbReminderCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "npl", "NewPlacer",           
      (void (*)()) NULL, 
      PbReminderCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "cl", "Cluster",
      (void (*)()) NULL, 
      PbClusterCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "pl", "Placer",           
      (void (*)()) NULL, 
      PbPlacerCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "wpl", "WritePlacement",   
      (void (*)()) NULL, 
      PbWritePlacementCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },
#endif /* USE_OLDP_CODE */

#if defined(USE_NEWP_CODE)
    { "ncl", "NewCluster",
      (void (*)()) NULL, 
      PbNewClusterCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "npl", "NewPlacer",           
      (void (*)()) NULL, 
      PbNewPlacerCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#if !defined(BEAR_BLACK_BOX)
    { "ngr", "NewGlobalRouter",           
      PbNewGlobalRouterCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_NULL_CP
    },
#endif /* BEAR_BLACK_BOX */
#endif /* USE_NEWP_CODE */

#if !defined(BEAR_BLACK_BOX)
    { "ap", "AddPinToNet",
      PbAddPinToNetCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "cp", "CreatePin",        
      PbCreatePinCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "dp", "DeletePin",        
      PbDeletePinCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "cn", "CreateNet",        
      PbCreateNetCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "dn", "DeleteNet",        
      PbDeleteNetCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "dup", "DuplicatePin",     
      PbDuplicatePinCmd,
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "scp", "ShowCellProp",     
      PbShowCellPropCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "sjp", "ShowJunctProp",
      PbShowJunctPropCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN,
      PB_MOUSE_LOC_CP
    },

    { "snp", "ShowNetProp",      
      PbShowNetPropCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "spp", "ShowPinProp",      
      PbShowPinPropCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "stp", "ShowTileProp",     
      PbShowTilePropCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

#if defined(USE_WIRE_CODE)
    { "rn", "RouteNet",         
      PbRouteNetCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "wn", "Wire",             
      PbWireCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },
#endif /* USE_WIRE_CODE */

    { "pu", "Push",             
      PbPushCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },
#endif /* BEAR_BLACK_BOX */

    { "bg", "BuildGraph",        
      (void (*)()) NULL, 
      PbBuildGraphCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "r", "ReadChip",         
      (void (*)()) NULL, 
      PbReadChipCmd, 
      PB_NULL_WIN, 
      PB_BLACK_BOX_CP
    },

    { "w", "WriteChip",         
      (void (*)()) NULL, 
      PbWriteChipCmd, 
      (PB_CHIP_WIN | PB_CIF_WIN),
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#if !defined(BEAR_BLACK_BOX)
#ifdef USE_DR_CODE
    { "dch", "DefChannel",       
      PbDefChannelCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "dlch", "DefLChannel",      
      PbDefLChannelCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },
#endif /* USE_DR_CODE */

    { "hcm", "HorzCompact",      
      PbHorzCompactCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "vcm", "VertCompact",      
      PbVertCompactCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "hdcm", "HorzDecompact",    
      PbHorzDecompactCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },

    { "vdcm", "VertDecompact",    
      PbVertDecompactCmd, 
      (BearStatus (*)()) NULL, 
      PB_CHIP_WIN, 
      PB_MOUSE_LOC_CP
    },
#endif /* BEAR_BLACK_BOX */

    { "cm", "Compact",          
      (void (*)()) NULL, 
      PbCompactCmd,
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "dcm", "Decompact",        
      (void (*)()) NULL, 
      PbDecompactCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "sho", "ShapeOpt",         
      (void (*)()) NULL, 
      PbShapeOptCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "gr", "GlobalRouter",     
      (void (*)()) NULL, 
      PbGlobalRouterCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

    { "cpa", "ChanPinArr",     
      (void (*)()) NULL, 
      PbChanPinArrCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },

#ifdef USE_DR_CODE
    { "ar", "AutoRoute",
      (void (*)()) NULL, 
      PbAutoRouteCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    },
#endif /* USE_DR_CODE */

#ifdef USE_DR_CODE
    { "rr", "RingRouter",
      (void (*)()) NULL, 
      PbRingRouterCmd, 
      PB_CHIP_WIN, 
      (PB_SCRIPT_CP | PB_BLACK_BOX_CP)
    }
#endif /* USE_DR_CODE */
};
int pbNumOfCmds = sizeof (pbCommand) / sizeof (Command);

