
/* pbDetour.c --
 *
 *     Sherry Solden
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbDetour.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#ifdef USE_DR_CODE

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"         
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "pbDetour.h"

typedef struct _LocalData
{
    PlaneData         *planeDataPtr;
    DCellsAroundRCell *dCellListsPtr;
    Tile              *TRtile;
    Tile              *TLtile;
    Tile              *BRtile;
    Tile              *BLtile;
    RCellType         rType;
    bool              startTR;
    bool              startBL;
} LocalData;

typedef struct _DetourData
{
    Tile          *tile1;
    Tile          *tile2;
    Tile          *tile3;
    Tile          *tile4;
    PlaneData     *planeDataPtr;
    CellMaster    *parentCellPtr;
    RCellType     rType;
    SideDirection side;
} DetourData;

typedef struct _MatchData
{
    PseudoPin *ppinPtr;
    BearId    ppinId;
    Net       *netPtr;
    float     width;
} MatchData;

typedef struct _DCellPinData
{
    PinsAroundJunction  *pinListPtr;
    bool                semiFoundFlag;
} DCellPinData;


/* ---------- Local Procedures ---------- */
void              AddTileToList();
static void       GetBLAndTRTiles();
static BearStatus ReRouteNet();
BearId            ProcessJunct();
BearId            ProcessBnTile();
int               IsMatchInJunctRegion();
int               IsMatchingPinInBnTile();
static BearStatus FindMatchPPinOnBnTile();
static BearStatus FindMatchPPinInJunct();
WallJunction *    GetNEJunctOfWallSeg();
WallJunction *    GetSWJunctOfWallSeg();
void              GetAcrossTile();
static BearStatus AddPPinsToHPinList();
static BearStatus AddPPinsToVPinList();
static int        FindDCells();
static int        PinsAroundDCells();


/* ---------- Local Global Variables ---------- */
static CellMaster *detourCellMasterPtr;
static PlaneData  *detourPlaneDataPtr;



/*
 * ----------------------------------------------------------------------------
 *
 * DefineAdjacentDummyCells() --
 *
 * Results:  Creates 4 lists of dummy cells that are adjacent to
 *           the routeCell. 
 *
 *                TL <------      -----> TR              ---------
 *                                                      |         |
 *                 ------------------------     (BL) TR |  TOP    | BR (TL)
 *                |                        |            |         |
 *                |                        |            |         |
 *                |                        |            |         |
 *                 ------------------------             |         |
 *                BL <------      -----> BR     (BR) TL |  BOT    | BL (TR)
 *                                                      |         |
 *                                                       ---------
 *
 *     Look for dummy cells by scanning across/down the center of the channel
 *          for bottleneckTiles.  Then for each bottleneck tile found determine
 *          if adjacentTiles are dummy cells.
 *     For horizontal channel the bnTile search proceeds left to right, but
 *           vertical channel bnTiles proceed top-down. So at the end
 *           need to shuffle around the vertical pointers.
 *     The order of empty cells should proceed from center of the channel
 *       towards the edge.
 *
 * ----------------------------------------------------------------------------
 */

void
DefineAdjacentDummyCells(planeDataPtr, rType, searchRect, dCellListsPtrPtr)
PlaneData          *planeDataPtr;
RCellType          rType;
Rect               searchRect;
DCellsAroundRCell  **dCellListsPtrPtr;
{
    LocalData           localData;
    DCellsAroundRCell   *dCellListsPtr;
    Rect                hRect;
    LinkedTile          *tmpTileList;
    Tile                *tmpTile;
    Plane               *planePtr;

    MALLOC(DCellsAroundRCell *, dCellListsPtr, sizeof (DCellsAroundRCell));
    localData.dCellListsPtr = dCellListsPtr;

    dCellListsPtr->TLptr = NIL(LinkedTile);
    dCellListsPtr->TRptr = NIL(LinkedTile);
    dCellListsPtr->BLptr = NIL(LinkedTile);
    dCellListsPtr->BRptr = NIL(LinkedTile);
 
    localData.startTR = localData.startBL = TRUE;
    localData.planeDataPtr = planeDataPtr;
    localData.rType = rType;
  
    if (rType == HORZ_RCELL)
    {
       planePtr = Tp_GetVertPlane(planeDataPtr);
    }
    else
    {
       planePtr = Tp_GetHorzPlane(planeDataPtr);
    }
    (void) TiSrArea(NIL(Tile), planePtr, &searchRect, FindDCells,
        (ClientData) &localData);

    /* Switch the entries */
    if (rType == VERT_RCELL)
    {
    tmpTileList = dCellListsPtr->TLptr;
    dCellListsPtr->TLptr = dCellListsPtr->BRptr;
    dCellListsPtr->BRptr = tmpTileList;

    tmpTileList = dCellListsPtr->TRptr;
    dCellListsPtr->TRptr = dCellListsPtr->BLptr;
    dCellListsPtr->BLptr = tmpTileList;

        tmpTile = localData.TLtile;
    localData.TLtile = localData.BRtile;
    localData.BRtile = tmpTile;

    tmpTile = localData.TRtile;
    localData.TRtile = localData.BLtile;
    localData.BLtile = tmpTile;
    }

    /* Assign the boundary points */
    if (rType == HORZ_RCELL)
    {
    if (dCellListsPtr->BLptr)
    {
        Tp_HorzTileToRect(planeDataPtr, localData.BLtile, &hRect);
        MALLOC(Point *, dCellListsPtr->BLpoint, sizeof (Point));
        dCellListsPtr->BLpoint->p_x = hRect.r_xbot;
        dCellListsPtr->BLpoint->p_y = hRect.r_ytop;
    }
    if (dCellListsPtr->BRptr)
    {
        Tp_HorzTileToRect(planeDataPtr, localData.BRtile, &hRect);
        MALLOC(Point *, dCellListsPtr->BRpoint, sizeof (Point));
        dCellListsPtr->BRpoint->p_x = hRect.r_xtop;
        dCellListsPtr->BRpoint->p_y = hRect.r_ytop;
    }
    if (dCellListsPtr->TLptr)
    {
        Tp_HorzTileToRect(planeDataPtr, localData.TLtile, &hRect);
        MALLOC(Point *, dCellListsPtr->TLpoint, sizeof (Point));
        dCellListsPtr->TLpoint->p_x = hRect.r_xbot;
        dCellListsPtr->TLpoint->p_y = hRect.r_ybot;
    }
    if (dCellListsPtr->TRptr)
    {
        Tp_HorzTileToRect(planeDataPtr, localData.TRtile, &hRect);
        MALLOC(Point *, dCellListsPtr->TRpoint, sizeof (Point));
        dCellListsPtr->TRpoint->p_x = hRect.r_xtop;
        dCellListsPtr->TRpoint->p_y = hRect.r_ybot;
    }
    }
    else
    {
    if (dCellListsPtr->BLptr)
    {
        Tp_VertTileToRect(planeDataPtr, localData.BLtile, &hRect);
        MALLOC(Point *, dCellListsPtr->BLpoint, sizeof (Point));
        dCellListsPtr->BLpoint->p_x = hRect.r_xbot;
        dCellListsPtr->BLpoint->p_y = hRect.r_ybot;
    }
    if (dCellListsPtr->BRptr)
    {
        Tp_VertTileToRect(planeDataPtr, localData.BRtile, &hRect);
        MALLOC(Point *, dCellListsPtr->BRpoint, sizeof (Point));
        dCellListsPtr->BRpoint->p_x = hRect.r_xbot;
        dCellListsPtr->BRpoint->p_y = hRect.r_ytop;
    }
    if (dCellListsPtr->TLptr)
    {
        Tp_VertTileToRect(planeDataPtr, localData.TLtile, &hRect);
        MALLOC(Point *, dCellListsPtr->TLpoint, sizeof (Point));
        dCellListsPtr->TLpoint->p_x = hRect.r_xtop;
        dCellListsPtr->TLpoint->p_y = hRect.r_ybot;
    }
    if (dCellListsPtr->TRptr)
    {
        Tp_VertTileToRect(planeDataPtr, localData.TRtile, &hRect);
        MALLOC(Point *, dCellListsPtr->TRpoint, sizeof (Point));
        dCellListsPtr->TRpoint->p_x = hRect.r_xtop;
        dCellListsPtr->TRpoint->p_y = hRect.r_ytop;
    }
    }
    
    *dCellListsPtrPtr = dCellListsPtr;
}



static int
FindDCells(tilePtr, localDataPtr)
Tile       *tilePtr;
LocalData  *localDataPtr;
{
    Tile *BLTilePtr;
    Tile *TRTilePtr;

    if (Tp_IsBottleneckTile(tilePtr))
    {
    
    GetBLAndTRTiles(localDataPtr->planeDataPtr, tilePtr, 
            localDataPtr->rType, &BLTilePtr, &TRTilePtr);

    if (Tp_IsDCellTile(BLTilePtr))
    {
        AddTileToList(localDataPtr->dCellListsPtr, BLTilePtr, 
              BOTTOM_LEFT_SIDE, localDataPtr->startBL);
    }
    else if ((Tp_IsCellInstTile(BLTilePtr)) ||
         (Tp_IsRCellTile(BLTilePtr)))
    {
            if (localDataPtr->startBL == TRUE)
        {
                localDataPtr->BLtile = BLTilePtr;
                localDataPtr->BRtile = BLTilePtr;
        localDataPtr->startBL = FALSE;
        }
            else
        {
                localDataPtr->BRtile = BLTilePtr;
        }

    }

    if (Tp_IsDCellTile(TRTilePtr))
    {
        AddTileToList(localDataPtr->dCellListsPtr, TRTilePtr, 
              TOP_RIGHT_SIDE, localDataPtr->startTR);
    }
    else if ((Tp_IsCellInstTile(TRTilePtr)) ||
         (Tp_IsRCellTile(TRTilePtr)))
    {
            if (localDataPtr->startTR == TRUE)
        {
                localDataPtr->TLtile = TRTilePtr;
                localDataPtr->TRtile = TRTilePtr;
        localDataPtr->startTR = FALSE;
        }
            else
        {
                localDataPtr->TRtile = TRTilePtr;
        }
    }
    }
    return (MAGIC_CONTINUE_SEARCH);
}



void
AddTileToList(dCellListsPtr, tilePtr, side, addToFrontOfList)
DCellsAroundRCell *dCellListsPtr;
Tile              *tilePtr;
SideDirection     side;
bool              addToFrontOfList;

{
    LinkedTile  *tmpTileList;
    LinkedTile  *newTileList;

    MALLOC(LinkedTile *, newTileList, sizeof (LinkedTile));
    newTileList->tilePtr = tilePtr;

    if (addToFrontOfList == TRUE)
    {
        if (side == BOTTOM_LEFT_SIDE)
    {
        newTileList->nextPtr = dCellListsPtr->BLptr;
        dCellListsPtr->BLptr = newTileList;
    }
        else
    {
        newTileList->nextPtr = dCellListsPtr->TLptr;
        dCellListsPtr->TLptr = newTileList;
    }
    }
    else
    {
    if (side == BOTTOM_LEFT_SIDE)
    {
        tmpTileList = dCellListsPtr->BRptr;
    }
    else
    {
        tmpTileList = dCellListsPtr->TRptr;
    }

        if (tmpTileList == NIL(LinkedTile))
    {
        if (side == BOTTOM_LEFT_SIDE)
        {
        dCellListsPtr->BRptr = newTileList;
        }
        else
        {
        dCellListsPtr->TRptr = newTileList;
        }
    }
    else
    {
        while (tmpTileList->nextPtr != NIL(LinkedTile))
        {
        tmpTileList = tmpTileList->nextPtr;
        }
        tmpTileList->nextPtr = newTileList;

    }
    newTileList->nextPtr = NIL(LinkedTile);
    }
}


static void
GetBLAndTRTiles(planeDataPtr, tilePtr, rType, blTilePtrPtr, trTilePtrPtr)
PlaneData *planeDataPtr;
Tile      *tilePtr;
RCellType rType;
Tile      **blTilePtrPtr;
Tile      **trTilePtrPtr;
{
    Point         searchPnt;
    Point         vSearchPnt;
    Rect          rect;
    Plane         *planePtr;
    register Tile *blTilePtr;
    register Tile *trTilePtr;

    if (rType == HORZ_RCELL)
    {
        planePtr = Tp_GetHorzPlane(planeDataPtr);
        /* Get Bottom Tile */
        (void) Tp_VertTileToRect(planeDataPtr, tilePtr, &rect);
        searchPnt.p_x = rect.r_ll.p_x;
        searchPnt.p_y = rect.r_ll.p_y - 1;
        blTilePtr = TiSrPoint(NIL(Tile), planePtr, &searchPnt);

        /* Get Top Tile */
        searchPnt.p_x = rect.r_ll.p_x;
        searchPnt.p_y = rect.r_ur.p_y;
        trTilePtr = TiSrPoint(NIL(Tile), planePtr, &searchPnt);
    }
    else
    {
        planePtr = Tp_GetVertPlane(planeDataPtr);
        /* Get Left Tile */
        (void) Tp_HorzTileToRect(planeDataPtr, tilePtr, &rect);
        searchPnt.p_x = rect.r_ll.p_x;
        searchPnt.p_y = rect.r_ll.p_y;
        Tp_HPointToVPoint(&searchPnt, &vSearchPnt);
        blTilePtr = TiSrPoint(NIL(Tile), planePtr, &vSearchPnt);

        /* Get Right Tile */
        searchPnt.p_x = rect.r_ur.p_x + 1;
        searchPnt.p_y = rect.r_ll.p_y;
        Tp_HPointToVPoint(&searchPnt, &vSearchPnt);
        trTilePtr = TiSrPoint(NIL(Tile), planePtr, &vSearchPnt);
    }
    *blTilePtrPtr = blTilePtr;
    *trTilePtrPtr = trTilePtr;
}


/*
 * ----------------------------------------------------------------------------
 *
 * DetourAroundDCells() --
 *
 *      Given 4 lists of dummy cells that are adjacent to
 *      the routeCell.
 *      Reroute the nets on "farthest-edge" of dummy cell
 *                    towards the side of the rCell;
 *
 *                                    tile2
 *                                 ___________
 *                                |///////////|
 *                           -->--*----->-----*-->--  
 *                          |     |           |     |
 *                          |     |///////////|     |
 *                    ------*-----------------------*-------
 *                   |//    |   //|XXXXXXXXXXX|//   |     //|
 *            tile1  |//    |   //|XXXXXXXXXXX|//   |     //| tile3
 *                   |//    ^   //|XemptycellX|//   V     //|
 *                   |//    |   //|XXXXXXXXXXX|//   |     //|
 *                   |//    |   //|XXXXXXXXXXX|//   |     //|
 *     ---------------------*-----------------------*-------
 *    |    boundary of channel    |///////////|   | |
 *    |                           *-----<-------<-*- 
 *    |    HORIZONTAL CHANNEL     |///////////|   |     side of channel
 *     -------------------------------------------
 *                                    tile4
 *
 *    In this example of a horizontal channel - with empty cell
 *       on top boundary near the right side.  Reroute the ppins
 *       on the botSide of tile1 up and around the emptyCell by
 *       inserting ppins on tile1, tile3, tile3 and tile4, as needed.
 *
 *    This procedure modifies the ppins on bnTiles to insure that
 *    all nets that intersected the "boundary" of the rCell can
 *    be routed through the nearest "side" of the channel.
 *
 * Side effects:
 *  None.
 *
 * ----------------------------------------------------------------------------
 */
void
DetourAroundDCells(parentCellPtr, planeDataPtr,rCellPtr, dCellListsPtrPtr)
CellMaster    *parentCellPtr;
PlaneData     *planeDataPtr;
RouteCell     *rCellPtr;
DCellsAroundRCell  **dCellListsPtrPtr;
{
    LinkedTile        *dCellPtr;
    DetourData        detourData;
    DCellsAroundRCell *dCellListsPtr;
    SideDirection     side;

    dCellListsPtr = *dCellListsPtrPtr;
    detourData.planeDataPtr = planeDataPtr;
    detourData.parentCellPtr = parentCellPtr;
    detourData.rType = Mac_GetTypeOfRCell(rCellPtr);

    for (dCellPtr = dCellListsPtr->TRptr; 
     dCellPtr != NIL(LinkedTile);
     dCellPtr = dCellPtr->nextPtr)
    {
    detourData.tile1 = BL(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, TOP_RIGHT_SIDE, 
              &detourData.tile2, detourData.rType);
    detourData.tile3 = TR(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, BOTTOM_LEFT_SIDE, 
              &detourData.tile4, detourData.rType);
    detourData.side = TOP_RIGHT_SIDE;
    
    if (detourData.rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    Tp_ItrPPinsOfSide(detourData.tile1, side, ReRouteNet, 
              (ClientData)&detourData);
    }

    for (dCellPtr = dCellListsPtr->TLptr; 
     dCellPtr != NIL(LinkedTile);
     dCellPtr = dCellPtr->nextPtr)
    {
    detourData.tile1 = TR(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, TOP_RIGHT_SIDE, 
              &detourData.tile2, detourData.rType);
    detourData.tile3 = BL(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, BOTTOM_LEFT_SIDE, 
              &detourData.tile4, detourData.rType);
    detourData.side = TOP_LEFT_SIDE;

    if (detourData.rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    Tp_ItrPPinsOfSide(detourData.tile1, side, ReRouteNet, 
              (ClientData)&detourData);
    }

    for (dCellPtr = dCellListsPtr->BRptr; 
     dCellPtr != NIL(LinkedTile);
     dCellPtr = dCellPtr->nextPtr)
    {
    detourData.tile1 = BL(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, BOTTOM_LEFT_SIDE, 
              &detourData.tile2, detourData.rType);
    detourData.tile3 = TR(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, TOP_RIGHT_SIDE, 
              &detourData.tile4, detourData.rType);
    detourData.side = RIGHT_SIDE;
    
    if (detourData.rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    Tp_ItrPPinsOfSide(detourData.tile1, side, ReRouteNet, 
              (ClientData)&detourData);
    }

    for (dCellPtr = dCellListsPtr->BLptr; 
     dCellPtr != NIL(LinkedTile);
     dCellPtr = dCellPtr->nextPtr)
    {
    detourData.tile1 = TR(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, BOTTOM_LEFT_SIDE, 
              &detourData.tile2, detourData.rType);
    detourData.tile3 = BL(dCellPtr->tilePtr);
    GetAcrossTile(planeDataPtr, dCellPtr->tilePtr, TOP_RIGHT_SIDE, 
              &detourData.tile4, detourData.rType);
    detourData.side = BOTTOM_LEFT_SIDE;

    if (detourData.rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    Tp_ItrPPinsOfSide(detourData.tile1, side, ReRouteNet, 
              (ClientData)&detourData);
    }
}


static BearStatus
ReRouteNet(pinPtr, detourData)
PseudoPin  *pinPtr;
DetourData *detourData;
{
    BearId        intId;
    BearId        extId;
    BearId        newId;
    PseudoPin     *ppinPtr;
    Net           *netPtr;
    float         width;
    Tile          *tile1;
    Tile          *tile2;
    Tile          *tile3;
    Tile          *tile4;
    RCellType     rType;
    SideDirection side;
    WallSegment   *segPtr;
    CellMaster    *parentCellPtr;
    PlaneData     *planeDataPtr;
    MatchData     matchData;


    netPtr = Tp_GetNetOfPPin(pinPtr);
    width  = Tp_GetWidthOfPPin(pinPtr);
    tile1 = detourData->tile1;
    tile2 = detourData->tile2;
    tile3 = detourData->tile3;
    tile4 = detourData->tile4;
    rType = detourData->rType;
    parentCellPtr = detourData->parentCellPtr;
    planeDataPtr = detourData->planeDataPtr;

    intId = Tp_GetIntNetIdOfPPin(pinPtr);
    newId = Tp_GetExtNetIdOfPPin(pinPtr);

    if (detourData->side == TOP_RIGHT_SIDE)
    {
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    Tp_DeletePPinOfBnTile(tile1, side, pinPtr);
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile1, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile1);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetNWJunctOfWallSeg(segPtr), 
                 extId, width, netPtr, 
                 tile2, BOTTOM_LEFT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile2, intId, 
                  width, netPtr, TOP_RIGHT_SIDE);
    segPtr = Tp_GetWallSegmentOfBnTile(tile2);
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 GetNEJunctOfWallSeg(segPtr),
                 extId, width, netPtr,
                 tile3, side);
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile3, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile3);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetSEJunctOfWallSeg(segPtr),
                 extId, width, netPtr,
                 tile4, TOP_RIGHT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile4, intId,
                  width, netPtr, BOTTOM_LEFT_SIDE);
    matchData.ppinPtr = NIL(PseudoPin);
    matchData.ppinId = intId;
    Tp_ItrPPinsOfSide(tile4, BOTTOM_LEFT_SIDE, FindMatchPPinOnBnTile, 
              (ClientData)&matchData);
    if (matchData.ppinPtr == NIL(PseudoPin))
    {
        (void) Tp_CreatePPin(tile4, BOTTOM_LEFT_SIDE, TP_BOTTOM, &ppinPtr);
        (void) Tp_PutExtNetIdOfPPin(newId, ppinPtr);
        (void) Tp_PutIntNetIdOfPPin(intId, ppinPtr);
        (void) Tp_PutNetOfPPin(netPtr, ppinPtr);
        (void) Tp_PutWidthOfPPin(width, ppinPtr);
    }
    else
    {
        Tp_PutExtNetIdOfPPin(newId, matchData.ppinPtr);
    }
    }

    else if (detourData->side == TOP_LEFT_SIDE)
    {
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    Tp_DeletePPinOfBnTile(tile1, side, pinPtr);
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile1, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile1);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetNWJunctOfWallSeg(segPtr), 
                 extId, width, netPtr,
                 tile2, TOP_RIGHT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile2, intId,
                  width, netPtr, BOTTOM_LEFT_SIDE);
    
    segPtr = Tp_GetWallSegmentOfBnTile(tile2);
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 GetSWJunctOfWallSeg(segPtr),
                 extId, width, netPtr,
                 tile3, side);
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile3, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile3);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetSEJunctOfWallSeg(segPtr),
                 extId, width, netPtr,
                 tile4, BOTTOM_LEFT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile4, intId,
                  width, netPtr, TOP_RIGHT_SIDE);
    matchData.ppinPtr = NIL(PseudoPin);
    matchData.ppinId = intId;
    Tp_ItrPPinsOfSide(tile4, TOP_RIGHT_SIDE, FindMatchPPinOnBnTile, 
              (ClientData)&matchData);
    if (matchData.ppinPtr == NIL(PseudoPin))
    {
        (void) Tp_CreatePPin(tile4, TOP_RIGHT_SIDE, TP_BOTTOM, &ppinPtr);
        (void) Tp_PutExtNetIdOfPPin(newId, ppinPtr);
        (void) Tp_PutIntNetIdOfPPin(intId, ppinPtr);
        (void) Tp_PutNetOfPPin(netPtr, ppinPtr);
        (void) Tp_PutWidthOfPPin(width, ppinPtr);
    }
    else
    {
        Tp_PutExtNetIdOfPPin(newId, matchData.ppinPtr);
    }
    }
    else if (detourData->side == BOTTOM_LEFT_SIDE)
    {
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    Tp_DeletePPinOfBnTile(tile1, side, pinPtr);
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile1, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile1);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetSEJunctOfWallSeg(segPtr), 
                 extId, width, netPtr,
                 tile2, TOP_RIGHT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile2, intId,
                  width, netPtr, BOTTOM_LEFT_SIDE);
    segPtr = Tp_GetWallSegmentOfBnTile(tile2);
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 GetSWJunctOfWallSeg(segPtr), 
                 extId, width, netPtr,
                 tile3, side);
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile3, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile3);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetNWJunctOfWallSeg(segPtr),
                 extId, width, netPtr,
                 tile4, BOTTOM_LEFT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile4, intId,
                  width, netPtr, TOP_RIGHT_SIDE);
    matchData.ppinPtr = NIL(PseudoPin);
    matchData.ppinId = intId;
    Tp_ItrPPinsOfSide(tile4, TOP_RIGHT_SIDE, FindMatchPPinOnBnTile, 
              (ClientData)&matchData);
    if (matchData.ppinPtr == NIL(PseudoPin))
    {
        (void) Tp_CreatePPin(tile4, TOP_RIGHT_SIDE, TP_BOTTOM, &ppinPtr);
        (void) Tp_PutExtNetIdOfPPin(newId, ppinPtr);
        (void) Tp_PutIntNetIdOfPPin(intId, ppinPtr);
        (void) Tp_PutNetOfPPin(netPtr, ppinPtr);
        (void) Tp_PutWidthOfPPin(width, ppinPtr);
    }
    else
    {
        Tp_PutExtNetIdOfPPin(newId, matchData.ppinPtr);
    }
    }
    else if (detourData->side == RIGHT_SIDE)
    {
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    Tp_DeletePPinOfBnTile(tile1, side, pinPtr);
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile1, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile1);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetSEJunctOfWallSeg(segPtr), 
                 extId, width, netPtr,
                 tile2, BOTTOM_LEFT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile2, intId,
                  width, netPtr, TOP_RIGHT_SIDE);
    segPtr = Tp_GetWallSegmentOfBnTile(tile2);
    if (rType == HORZ_RCELL)
        side = BOTTOM_LEFT_SIDE;
    else side = TOP_RIGHT_SIDE;
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 GetNEJunctOfWallSeg(segPtr), 
                 extId, width, netPtr,
                 tile3, side);
    if (rType == HORZ_RCELL)
        side = TOP_RIGHT_SIDE;
    else side = BOTTOM_LEFT_SIDE;
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile3, intId,
                  width, netPtr, side);
    segPtr = Tp_GetWallSegmentOfBnTile(tile3);
    intId = ProcessJunct(parentCellPtr, planeDataPtr,
                 Fpg_GetNWJunctOfWallSeg(segPtr),
                 extId, width, netPtr,
                 tile4, TOP_RIGHT_SIDE);
    extId = ProcessBnTile(parentCellPtr, planeDataPtr, tile4, intId,
                  width, netPtr, BOTTOM_LEFT_SIDE);
    matchData.ppinPtr = NIL(PseudoPin);
    matchData.ppinId = intId;
    Tp_ItrPPinsOfSide(tile4, BOTTOM_LEFT_SIDE, FindMatchPPinOnBnTile, 
              (ClientData)&matchData);
    if (matchData.ppinPtr == NIL(PseudoPin))
    {
        (void) Tp_CreatePPin(tile4, BOTTOM_LEFT_SIDE, TP_BOTTOM, &ppinPtr);
        (void) Tp_PutExtNetIdOfPPin(newId, ppinPtr);
        (void) Tp_PutIntNetIdOfPPin(intId, ppinPtr);
        (void) Tp_PutNetOfPPin(netPtr, ppinPtr);
        (void) Tp_PutWidthOfPPin(width, ppinPtr);
    }
    else
    {
        Tp_PutExtNetIdOfPPin(newId, matchData.ppinPtr);
    }
    }
}

BearId
ProcessJunct(parentCellPtr, planeDataPtr, junctPtr, 
             newId, width, netPtr, bnTilePtr, side)
CellMaster    *parentCellPtr;
PlaneData     *planeDataPtr;
WallJunction  *junctPtr;
BearId        newId;
float         width;
Net           *netPtr;
Tile          *bnTilePtr;
SideDirection side;
{
    int       matchCnt;
    PseudoPin *ppinPtr;
    MatchData matchData;
    BearId    ppinId;

    matchCnt = IsMatchInJunctRegion(parentCellPtr, planeDataPtr, 
                    newId, junctPtr);
    matchData.ppinPtr = NIL(PseudoPin);
    matchData.ppinId = newId;
    Tp_ItrPPinsOfSide(bnTilePtr, side, FindMatchPPinInJunct, 
              (ClientData)&matchData);
    if (matchData.ppinPtr != NIL(PseudoPin))
    {
    ppinId = Tp_GetIntNetIdOfPPin(matchData.ppinPtr);
    if (matchCnt == 1)
    {
        Tp_DeletePPinOfBnTile(bnTilePtr, side, matchData.ppinPtr);
    }
    }
    else if (matchCnt > 0)
    {
    (void) Tp_CreatePPin(bnTilePtr, side, TP_BOTTOM, &ppinPtr);
    (void) Tp_PutExtNetIdOfPPin(newId, ppinPtr);
    ppinId = Mac_GenerateId();
    (void) Tp_PutIntNetIdOfPPin(ppinId, ppinPtr);
    (void) Tp_PutNetOfPPin(netPtr, ppinPtr);
    (void) Tp_PutWidthOfPPin(width, ppinPtr);
    }
    return (ppinId);
}



BearId
ProcessBnTile(parentCellPtr, planeDataPtr, bnTilePtr, newId, 
              width, netPtr, side)
CellMaster    *parentCellPtr;
PlaneData     *planeDataPtr;
Tile          *bnTilePtr;
BearId        newId;
float         width;
Net           *netPtr;
SideDirection side;
{
    int       matchCnt;
    PseudoPin *ppinPtr;
    PseudoPin *newPPinPtr;
    MatchData matchData;
    BearId    ppinId;
    SideDirection otherside;

    ppinId = newId;
    matchCnt = IsMatchingPinInBnTile(parentCellPtr, planeDataPtr, 
                                   newId, bnTilePtr);
    matchData.ppinId = newId;
    matchData.ppinPtr = NIL(PseudoPin);
    Tp_ItrPPinsOfSide(bnTilePtr, side, FindMatchPPinOnBnTile, 
              (ClientData)&matchData);
    newPPinPtr = matchData.ppinPtr;

    matchData.ppinPtr = NIL(PseudoPin);
    if (side == TOP_RIGHT_SIDE)
    otherside = BOTTOM_LEFT_SIDE;
    else otherside = TOP_RIGHT_SIDE;
    Tp_ItrPPinsOfSide(bnTilePtr, otherside, FindMatchPPinOnBnTile, 
              (ClientData)&matchData);

    if (newPPinPtr != NIL(PseudoPin))
    {
    ppinId = Tp_GetExtNetIdOfPPin(newPPinPtr);
    if ((matchCnt == 0) && (matchData.ppinPtr == NIL(PseudoPin)))
    {
        Tp_DeletePPinOfBnTile(bnTilePtr, side, newPPinPtr);
    }
    }
    else if ((matchCnt > 0) || (matchData.ppinPtr != NIL(PseudoPin)))
    {
    (void) Tp_CreatePPin(bnTilePtr, side, TP_BOTTOM, &ppinPtr);
    (void) Tp_PutIntNetIdOfPPin(newId, ppinPtr);
    ppinId = Mac_GenerateId();
    (void) Tp_PutExtNetIdOfPPin(ppinId, ppinPtr);
    (void) Tp_PutNetOfPPin(netPtr, ppinPtr);
    (void) Tp_PutWidthOfPPin(width, ppinPtr);
    }
    return (ppinId);
}


int
IsMatchInJunctRegion(parentCellPtr, planeDataPtr, ppinId, junctPtr)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
BearId   ppinId;
WallJunction *junctPtr;
{
    PinsAroundJunction *pinListPtr;
    LinkedPPin         *curLinkedPPinPtr;
    LinkedPinInst      *curLinkedPinInstPtr;
    LinkedRPin         *curLinkedRPinPtr;
    int                count;

    count = 0;
    MALLOC(PinsAroundJunction *, pinListPtr, sizeof (PinsAroundJunction));
    Tm_PinsAroundJunction(parentCellPtr, planeDataPtr, junctPtr, pinListPtr);

    for (curLinkedPPinPtr  = pinListPtr->horzPPinList;
     curLinkedPPinPtr != NIL(LinkedPPin);
     curLinkedPPinPtr  = curLinkedPPinPtr->nextPtr)
    {
    if (Tp_GetExtNetIdOfPPin(curLinkedPPinPtr->ppinPtr) == ppinId)
    {
        count++;
    }
    }
    for (curLinkedPPinPtr  = pinListPtr->vertPPinList;
     curLinkedPPinPtr != NIL(LinkedPPin);
     curLinkedPPinPtr  = curLinkedPPinPtr->nextPtr)
    {
    if (Tp_GetExtNetIdOfPPin(curLinkedPPinPtr->ppinPtr) == ppinId)
    {
        count++;
    }
    }

    for (curLinkedPinInstPtr  = pinListPtr->pinInstList;
     curLinkedPinInstPtr != NIL(LinkedPinInst);
     curLinkedPinInstPtr  = curLinkedPinInstPtr->nextPtr)
    {
    if (Mac_GetExtNetIdOfPin(curLinkedPinInstPtr->pinInstPtr) == ppinId)
    {
        count++;
    }
    }

    for (curLinkedRPinPtr  = pinListPtr->rPinList;
     curLinkedRPinPtr != NIL(LinkedRPin);
     curLinkedRPinPtr  = curLinkedRPinPtr->nextPtr)
    {
    if (Mac_GetExtNetIdOfRPin(curLinkedRPinPtr->rPinPtr) == ppinId)
    {
        count++;
    }
    }
    return (count);
}


int             
IsMatchingPinInBnTile(parentCellPtr, planeDataPtr, ppinId, bnTilePtr)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
BearId ppinId;
Tile   *bnTilePtr;
{
    int count;
    PinsAroundJunction *pinListPtr;
    LinkedPPin         *curLinkedPPinPtr;
    LinkedPinInst      *curLinkedPinInstPtr;
    LinkedRPin         *curLinkedRPinPtr;
    
    count = 0;
    MALLOC(PinsAroundJunction *, pinListPtr, sizeof (PinsAroundJunction));
    if (Tp_IsHorzTile(bnTilePtr))
    {
    Tm_GetPinsInHBottleNeckRegion(parentCellPtr, planeDataPtr, 
                       bnTilePtr, pinListPtr);
        /* Might find pins on side of dummy cell */
        for (curLinkedPPinPtr  = pinListPtr->vertPPinList;
         curLinkedPPinPtr != NIL(LinkedPPin);
         curLinkedPPinPtr  = curLinkedPPinPtr->nextPtr)
    {
        if (Tp_GetExtNetIdOfPPin(curLinkedPPinPtr->ppinPtr) == ppinId)
        {
        count++;
        }
    }
    }
    else
    {
    Tm_GetPinsInVBottleNeckRegion(parentCellPtr, planeDataPtr, 
                      bnTilePtr, pinListPtr);
    /* Might find pins on side of dummy cell */
    for (curLinkedPPinPtr  = pinListPtr->horzPPinList;
         curLinkedPPinPtr != NIL(LinkedPPin);
         curLinkedPPinPtr  = curLinkedPPinPtr->nextPtr)
    {
        if (Tp_GetExtNetIdOfPPin(curLinkedPPinPtr->ppinPtr) == ppinId)
        {
        count++;
        }
    }
    }
    for (curLinkedPinInstPtr  = pinListPtr->pinInstList;
     curLinkedPinInstPtr != NIL(LinkedPinInst);
     curLinkedPinInstPtr  = curLinkedPinInstPtr->nextPtr)
    {
    if (Mac_GetExtNetIdOfPin(curLinkedPinInstPtr->pinInstPtr) == ppinId)
    {
        count++;
    }
    }
    for (curLinkedRPinPtr  = pinListPtr->rPinList;
     curLinkedRPinPtr != NIL(LinkedRPin);
     curLinkedRPinPtr  = curLinkedRPinPtr->nextPtr)
    {
    if (Mac_GetExtNetIdOfRPin(curLinkedRPinPtr->rPinPtr) == ppinId)
    {
        count++;
    }
    }
    return (count);
}

static BearStatus
FindMatchPPinOnBnTile(ppinPtr, matchDataPtr)
PseudoPin  *ppinPtr;
MatchData  *matchDataPtr;
{

    if (Tp_GetIntNetIdOfPPin(ppinPtr) == matchDataPtr->ppinId)
    {
    matchDataPtr->ppinPtr = ppinPtr;
    return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}

static BearStatus
FindMatchPPinInJunct(ppinPtr, matchDataPtr)
PseudoPin  *ppinPtr;
MatchData  *matchDataPtr;
{

    if (Tp_GetExtNetIdOfPPin(ppinPtr) == matchDataPtr->ppinId)
    {
    matchDataPtr->ppinPtr = ppinPtr;
    return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}

WallJunction *
GetNEJunctOfWallSeg(wallSegPtr)
WallSegment *wallSegPtr;
{
    if (wallSegPtr == NULL)
    {
    Bear_PrintFatalErr("GetNEJunctOfWallSeg() failed: no segment given");
    }
    if (wallSegPtr->direction == FPG_HORIZONTAL)
    {
    return (wallSegPtr->SEJunctPtr);
    }
    else
    {
    return (wallSegPtr->NWJunctPtr);
    }
}

WallJunction *
GetSWJunctOfWallSeg(wallSegPtr)
WallSegment *wallSegPtr;
{
    if (wallSegPtr == NULL)
    {
    Bear_PrintFatalErr("GetSWJunctOfWallSeg() failed: no segment given");
    }
    if (wallSegPtr->direction == FPG_HORIZONTAL)
    {
    return (wallSegPtr->NWJunctPtr);
    }
    else
    {
    return (wallSegPtr->SEJunctPtr);
    }
}

void
GetAcrossTile(planeDataPtr, dCellTilePtr, side, newTilePtrPtr, rType)
PlaneData *planeDataPtr;
Tile *dCellTilePtr;
SideDirection side;
Tile **newTilePtrPtr;
RCellType rType;
{
    Point point;
    Point vPoint;
    Rect  dRect;
    register Tile *newTilePtr;

    if (rType == HORZ_RCELL)
    {
    Tp_HorzTileToRect(planeDataPtr, dCellTilePtr, &dRect);
    if (side == TOP_RIGHT_SIDE)
    {
        point.p_x = dRect.r_xtop;
        point.p_y = dRect.r_ytop;
    }
    else
    {
        point.p_x = dRect.r_xtop;
        point.p_y = dRect.r_ybot - 1;
    }
    Tp_HPointToVPoint(&point, &vPoint);
    newTilePtr = TiSrPoint(NIL(Tile), Tp_GetVertPlane(planeDataPtr), &vPoint);
    /* When bnLine crosses the empty cell */
    if (!Tp_IsBottleneckTile(newTilePtr))
    {
        point.p_x = dRect.r_xbot;
        Tp_HPointToVPoint(&point, &vPoint);
        newTilePtr = TiSrPoint(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
            &vPoint);
    }
    }
    else
    {
    Tp_VertTileToRect(planeDataPtr, dCellTilePtr, &dRect);
    if (side == TOP_RIGHT_SIDE)
    {
        point.p_x = dRect.r_xbot - 1;
        point.p_y = dRect.r_ybot;
    }
    else
    {
        point.p_x = dRect.r_xtop;
        point.p_y = dRect.r_ybot;
    }
    newTilePtr = TiSrPoint(NIL(Tile), Tp_GetHorzPlane(planeDataPtr), &point);
    /* When bnLine crosses the empty cell */
    if (!Tp_IsBottleneckTile(newTilePtr))
    {
        point.p_y = dRect.r_ybot;
        newTilePtr = TiSrPoint(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
            &point);
    }
    }
    
    *newTilePtrPtr = newTilePtr;
}



PinsAroundJunction *
Pb_PinsAroundDCells(cellMasterPtr, planeDataPtr, rType, dCellListsPtr, jPoint)
CellMaster        *cellMasterPtr;
PlaneData         *planeDataPtr;
RCellType         rType;
DCellsAroundRCell *dCellListsPtr;
Point             jPoint;

{
    Plane           *planePtr;
    Rect            rect, vRect;
    DCellPinData    dcellPinData;

    detourCellMasterPtr = cellMasterPtr;
    detourPlaneDataPtr = planeDataPtr;
 
/*
 * code change : by george : on 25 oct 89
 *
    MALLOC(PinsAroundJunction *, pinListPtr, sizeof (PinsAroundJunction));
    pinListPtr->horzPPinList = NIL(LinkedPPin);
    pinListPtr->vertPPinList = NIL(LinkedPPin);
    pinListPtr->pinInstList = NIL(LinkedPinInst);
    pinListPtr->rPinList = NIL(LinkedRPin);
 */
    dcellPinData.pinListPtr = Tm_MakePinsAroundJunction();

    if (rType == VERT_RCELL) 
    {
        planePtr = Tp_GetHorzPlane(planeDataPtr);
        if (dCellListsPtr->BRptr) 
        {
            rect.r_xbot = dCellListsPtr->BRpoint->p_x;
            rect.r_xtop = rect.r_xbot + 1;
            rect.r_ytop = jPoint.p_y;
            rect.r_ybot = dCellListsPtr->BRpoint->p_y;
            dcellPinData.semiFoundFlag = FALSE;
            (void) TiSrArea(NIL(Tile),
                planePtr,
                &rect,
                PinsAroundDCells,
                (ClientData) &dcellPinData);
        }
        if (dCellListsPtr->TRptr) 
        {
            rect.r_xtop = dCellListsPtr->TRpoint->p_x;
            rect.r_xbot = rect.r_xtop - 1;
            rect.r_ytop = jPoint.p_y;
            rect.r_ybot = dCellListsPtr->TRpoint->p_y;
            dcellPinData.semiFoundFlag = FALSE;
            (void) TiSrArea(NIL(Tile),
                planePtr,
                &rect,
                PinsAroundDCells,
                (ClientData) &dcellPinData);
        }
    }
    else
    {
        planePtr = Tp_GetVertPlane(planeDataPtr);
        if (dCellListsPtr->BLptr) 
        {
            rect.r_xtop = dCellListsPtr->BLpoint->p_x;
            rect.r_xbot = jPoint.p_x;
            rect.r_ytop = dCellListsPtr->BLpoint->p_y;
            rect.r_ybot = rect.r_ytop - 1;
            Tp_HRectToVRect(&rect, &vRect);
            dcellPinData.semiFoundFlag = FALSE;
            (void) TiSrArea(NIL(Tile),
                planePtr,
                &vRect,
                PinsAroundDCells,
                (ClientData) &dcellPinData);
        }
        if (dCellListsPtr->TLptr) 
        {
            rect.r_xtop = dCellListsPtr->TLpoint->p_x;
            rect.r_xbot = jPoint.p_x;
            rect.r_ybot = dCellListsPtr->TLpoint->p_y;
            rect.r_ytop = rect.r_ybot + 1;
            Tp_HRectToVRect(&rect, &vRect);
            dcellPinData.semiFoundFlag = FALSE;
            (void) TiSrArea(NIL(Tile),
                 planePtr,
                 &vRect,
                 PinsAroundDCells,
                 (ClientData) &dcellPinData);
        }
    }

    return (dcellPinData.pinListPtr);
}



static int
PinsAroundDCells(tilePtr, dcellPinDataPtr)
Tile            *tilePtr;
DCellPinData    *dcellPinDataPtr;
{
    BNLineType          bnType;
    PinsAroundJunction  *pinListPtr;
    Rect                domRect;
    Tile                *semiTile1Ptr;
    Tile                *semiTile2Ptr;

    if (Tp_IsDominantTile(tilePtr))
    {
/*
 * code change : by george : on 25 oct 89
 *
 * Not needed here because Tm_PinsAround?Face() returns memory.
 *
       MALLOC(PinsAroundJunction *, pinListPtr, sizeof (PinsAroundJunction));
 */
       /* Get all pins in face region */
       if (Tp_IsHorzTile(tilePtr))
       {
           Tp_HorzTileToRect(detourPlaneDataPtr, tilePtr, &domRect);
           pinListPtr = Tm_PinsAroundHFace(detourCellMasterPtr, 
                        detourPlaneDataPtr, 
                        domRect,
                        TRUE);
       }
       else
       {
           Tp_VertTileToRect(detourPlaneDataPtr, tilePtr, &domRect);
           pinListPtr = Tm_PinsAroundVFace(detourCellMasterPtr, 
                        detourPlaneDataPtr, 
                        domRect,
                        TRUE);
       }

       AppendLists(pinListPtr, dcellPinDataPtr->pinListPtr);
       FREE((char *) pinListPtr);
    }

    /* If tile is bottleneck tile */
    if (Tp_IsBottleneckTile(tilePtr))
    {
       dcellPinDataPtr->semiFoundFlag = FALSE;

/*
 * code change : by george : on 25 oct 89
 *
       MALLOC(PinsAroundJunction *, pinListPtr, sizeof (PinsAroundJunction));
 */
       pinListPtr = Tm_MakePinsAroundJunction();

       /* Get all pins in bottleneck region */
       if (Tp_IsHorzTile(tilePtr))
       {
           /* Add pseudo pins on top and bottom sides of bnTile */
           (void) Tp_ItrPPinsOfSide(tilePtr, 
                    TOP_RIGHT_SIDE,
                    AddPPinsToHPinList,
                    (ClientData) dcellPinDataPtr->pinListPtr);

           (void) Tp_ItrPPinsOfSide(tilePtr, 
                    BOTTOM_LEFT_SIDE,
                    AddPPinsToHPinList,
                    (ClientData) dcellPinDataPtr->pinListPtr);

           /* Add pinInst and rPins from side of bnTile */
           (void) Tm_GetPinsInHBottleNeckRegion(detourCellMasterPtr,
                            detourPlaneDataPtr, 
                            tilePtr, 
                            pinListPtr);
           AppendLists(pinListPtr, dcellPinDataPtr->pinListPtr);
       }
       else
       {
           /* Add pseudo pins on top and bottom sides of bnTile */
           (void) Tp_ItrPPinsOfSide(tilePtr, 
                    TOP_RIGHT_SIDE,
                    AddPPinsToVPinList,
                    (ClientData) dcellPinDataPtr->pinListPtr);

           (void) Tp_ItrPPinsOfSide(tilePtr, 
                    BOTTOM_LEFT_SIDE,
                    AddPPinsToVPinList,
                    (ClientData) dcellPinDataPtr->pinListPtr);

           /* Add pinInst and rPins from side of bnTile */
           (void) Tm_GetPinsInVBottleNeckRegion(detourCellMasterPtr,
                            detourPlaneDataPtr, 
                            tilePtr, 
                            pinListPtr);
           AppendLists(pinListPtr, dcellPinDataPtr->pinListPtr);
       }
       FREE((char *) pinListPtr);
    }
    /* Semi Bottleneck Tile */
    else if (Tp_IsSBNTile(tilePtr))
    {
       if (dcellPinDataPtr->semiFoundFlag == FALSE)
       {
          semiTile1Ptr = tilePtr;
          dcellPinDataPtr->semiFoundFlag = TRUE;
       }
       else
       {
          semiTile2Ptr = tilePtr;
          if (!Tp_IsPairOfSBNTiles(&semiTile1Ptr, &semiTile2Ptr, &bnType))
          {
              /* Didn't find a match */
              semiTile1Ptr = tilePtr;
              return (MAGIC_CONTINUE_SEARCH);
          }
          if (bnType != TP_NULL_BNL)
          {
/*
 * code change : by george : on 25 oct 89
 *
             MALLOC(PinsAroundJunction *, pinListPtr,
                 sizeof (PinsAroundJunction));
 */
             pinListPtr = Tm_MakePinsAroundJunction();
         
             /* Add pseudo pins from bnLine */
             if (Tp_IsHorzTile(semiTile1Ptr))
             {
                 (void) Tp_ItrPPinsOfBnLine(semiTile1Ptr, semiTile2Ptr, 
                                            AddPPinsToHPinList,
                                            (ClientData) pinListPtr);
             }
             else
             {
                 (void) Tp_ItrPPinsOfBnLine(semiTile1Ptr, semiTile2Ptr, 
                                            AddPPinsToVPinList,
                                            (ClientData) pinListPtr);
             }
             AppendLists(pinListPtr, dcellPinDataPtr->pinListPtr);
             FREE((char *) pinListPtr);

             dcellPinDataPtr->semiFoundFlag = FALSE;
          }
       }
    }
    else
    {
       dcellPinDataPtr->semiFoundFlag = FALSE;
    }
    return (MAGIC_CONTINUE_SEARCH);
}



void
AppendLists(pinListPtr1, pinListPtr2)
PinsAroundJunction *pinListPtr1;
PinsAroundJunction *pinListPtr2;
{
    LinkedPPin    *pPinPtr;
    LinkedRPin    *rPinPtr;
    LinkedPinInst *pinPtr;

    if (pinListPtr2->horzPPinList == NIL(LinkedPPin))
    {
    pinListPtr2->horzPPinList = pinListPtr1->horzPPinList;
    pinListPtr1->horzPPinList = NIL(LinkedPPin);
    }
    else if (pinListPtr1->horzPPinList != NIL(LinkedPPin))
    {
    for (pPinPtr = pinListPtr1->horzPPinList;
         pPinPtr->nextPtr != NIL(LinkedPPin); pPinPtr = pPinPtr->nextPtr);
    pPinPtr->nextPtr = pinListPtr2->horzPPinList;
    pinListPtr2->horzPPinList = pinListPtr1->horzPPinList;
    }

    if (pinListPtr2->vertPPinList == NIL(LinkedPPin))
    {
    pinListPtr2->vertPPinList = pinListPtr1->vertPPinList;
    pinListPtr1->vertPPinList = NIL(LinkedPPin);
    }
    else if (pinListPtr1->vertPPinList != NIL(LinkedPPin))
    {
    for (pPinPtr = pinListPtr1->vertPPinList;
         pPinPtr->nextPtr != NIL(LinkedPPin); pPinPtr = pPinPtr->nextPtr);
    pPinPtr->nextPtr = pinListPtr2->vertPPinList;
    pinListPtr2->vertPPinList = pinListPtr1->vertPPinList;
    }

    if (pinListPtr2->pinInstList == NIL(LinkedPinInst))
    {
    pinListPtr2->pinInstList = pinListPtr1->pinInstList;
    pinListPtr1->pinInstList = NIL(LinkedPinInst);
    }
    else if (pinListPtr1->pinInstList != NIL(LinkedPinInst))
    {
    for (pinPtr = pinListPtr1->pinInstList;
         pinPtr->nextPtr != NIL(LinkedPinInst); pinPtr = pinPtr->nextPtr);
    pinPtr->nextPtr = pinListPtr2->pinInstList;
    pinListPtr2->pinInstList = pinListPtr1->pinInstList;
    }

    if (pinListPtr2->rPinList == NIL(LinkedRPin))
    {
    pinListPtr2->rPinList = pinListPtr1->rPinList;
    pinListPtr1->rPinList = NIL(LinkedRPin);
    }
    else if (pinListPtr1->rPinList != NIL(LinkedRPin))
    {
    for (rPinPtr = pinListPtr1->rPinList;
         rPinPtr->nextPtr != NIL(LinkedRPin); rPinPtr = rPinPtr->nextPtr);
    rPinPtr->nextPtr = pinListPtr2->rPinList;
    pinListPtr2->rPinList = pinListPtr1->rPinList;
    }
}



static BearStatus
AddPPinsToHPinList(pPinPtr, pinListPtr)
PseudoPin *pPinPtr;
PinsAroundJunction *pinListPtr;
{
   LinkedPPin *pinPtr;

   MALLOC(LinkedPPin *, pinPtr, sizeof (LinkedPPin));
   pinPtr->ppinPtr = pPinPtr;
   pinPtr->nextPtr = pinListPtr->horzPPinList;
   pinListPtr->horzPPinList = pinPtr;
   return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPPinsToVPinList(pPinPtr, pinListPtr)
PseudoPin *pPinPtr;
PinsAroundJunction *pinListPtr;
{
   LinkedPPin *pinPtr;

   MALLOC(LinkedPPin *, pinPtr, sizeof (LinkedPPin));
   pinPtr->ppinPtr = pPinPtr;
   pinPtr->nextPtr = pinListPtr->vertPPinList;
   pinListPtr->vertPPinList = pinPtr;
   return (BEAR_CONTINUE_SEARCH);
}

#endif /* USE_DR_CODE */


