
/* pbIVFpWins.c --
 *
 *     Benjamin Chen
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbIVFpWins.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#if !defined(BEAR_BLACK_BOX)
/*
 * This file does not need to be compiled when the black box version of
 * the BEAR-FP system is built.
 */

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "shapeOptimizer.h"

#include "bearIV.h"
#include "pbDialogs.h"
#include "polarBear.h"


static ivWhichItem soOpts[] = {
    {"Best Slack", 1},
    {"Half Slack", 0},
    {"Full Slack", 0}
};

static ivWhichItem soPref[] = {
    {"None", 1},
    {"Horz", 0},
    {"Vert", 0}
};

static ivWhichItem soAlgs[] = {
    {"2-D", 1},
    {"1-D", 0}
};

static ivWhichItem soDesign[] = {
    {"BBL", 1},
    {"SC",  0},
    {"GA",  0}
};


BearStatus
PbShapeOptDialog(parmsPtr)
So_UserParms *parmsPtr;
{
#define FILE_NAME_SIZE 40

    IVWindow    *ivWinPtr;
    XEvent      theEvent;
    char        *flFileName;
    bool        doOpt = FALSE;
    bool        abortFlag;

    switch (parmsPtr->algorithm) {
    case ONE_D:
        soAlgs[0].select = 0;
        soAlgs[1].select = 1;
        break;
    case TWO_D:
        soAlgs[0].select = 1;
        soAlgs[1].select = 0;
        break;
    default:
        PbPrint(EZ_ITALIC_FONT, "Illegal choice of algorithum.\n");
        return (BEAR_ERROR);
    }

    switch (parmsPtr->prefDirect) {
    case SO_HORZ:
        soPref[0].select = 0;
        soPref[1].select = 1;
        soPref[2].select = 0;
        break;
    case SO_VERT:
        soPref[0].select = 0;
        soPref[1].select = 0;
        soPref[2].select = 1;
        break;
    case SO_NONE:
        soPref[0].select = 1;
        soPref[1].select = 0;
        soPref[2].select = 0;
        break;
    default:
        PbPrint(EZ_ITALIC_FONT, "Illegal choice of prefered direction.\n");
        return (BEAR_ERROR);
    }

    if (parmsPtr->flexFileName != NIL(char)) {
        flFileName = parmsPtr->flexFileName;
    }
    else {
        MALLOC(char *, flFileName, FILE_NAME_SIZE * sizeof (char));
        flFileName[0] = '\0';
    }

    ivWinPtr = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
        "Shape Optimizer", FILE_NAME_SIZE);

    (void) iv_AddVar(ivWinPtr, "Algorithm               ",
        soAlgs, IV_WHICH, sizeof (soAlgs)/sizeof (ivWhichItem),
        IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Preferred Direction     ",
        soPref, IV_WHICH, sizeof (soPref)/sizeof (ivWhichItem),
        IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Options                 ",
        soOpts, IV_WHICH, sizeof (soDesign)/sizeof (ivWhichItem),
        IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Layout Style            ",
        soDesign, IV_WHICH, sizeof (soDesign)/sizeof (ivWhichItem),
        IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Maximum Iteration Count ",
        &parmsPtr->maxCnt, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Minimum Slack Size      ",
        &parmsPtr->gridSize, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Flex Filename           ",
        (int *) &flFileName, IV_STRING, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Optimize Placement      ",
        &doOpt, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    abortFlag = FALSE;
    (void) iv_AddVar(ivWinPtr, "Exit                    ",
        &abortFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_MapWindow(ivWinPtr, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);    

    while (!doOpt) {
        XNextEvent(ez_Display, &theEvent);
        (void) iv_HandleEvent(&theEvent);
        Ez_XEventHandler(&theEvent);

        if (abortFlag == TRUE) {
            iv_FreeWindow(ivWinPtr);
            ivWinPtr = NIL(IVWindow);
            return (BEAR_ERROR);
        }
    }

    switch (iv_WhichSelect(ivWinPtr, soOpts)) {
    case 0:
        parmsPtr->option = BEST_SLACK;
        break;
    case 1:
        parmsPtr->option = HALF_SLACK;
        break;
    case 2:
        parmsPtr->option = FULL_SLACK;
        break;
    default:
        PbPrint(EZ_ITALIC_FONT, "Illegal choice of option parameter.\n");
        iv_FreeWindow(ivWinPtr);
        return (BEAR_ERROR);
    }

    switch (iv_WhichSelect(ivWinPtr, soAlgs)) {
    case 0:
        parmsPtr->algorithm = TWO_D;
        break;
    case 1:
        parmsPtr->algorithm = ONE_D;
        break;
    default:
        PbPrint(EZ_ITALIC_FONT, "Illegal choice of algorithum parameter.\n");
        iv_FreeWindow(ivWinPtr);
        return (BEAR_ERROR);
    }

    switch (iv_WhichSelect(ivWinPtr, soPref)) {
    case 0:
        parmsPtr->prefDirect = SO_NONE;
        break;
    case 1:
        parmsPtr->prefDirect = SO_HORZ;
        break;
    case 2:
        parmsPtr->prefDirect = SO_VERT;
        break;
    default:
        PbPrint(EZ_ITALIC_FONT,
            "Illegal choice of prefered direction parameter.\n");
        iv_FreeWindow(ivWinPtr);
        return (BEAR_ERROR);
    }

    switch (iv_WhichSelect(ivWinPtr, soDesign)) {
    case 0:
        parmsPtr->style = BBL;
        break;
    case 1:
        parmsPtr->style = SC;
        break;
    case 2:
        parmsPtr->style = GA;
        break;
    default:
        PbPrint(EZ_ITALIC_FONT, "Illegal choice of design style parameter.\n");
        iv_FreeWindow(ivWinPtr);
        return (BEAR_ERROR);
    }
    iv_FreeWindow(ivWinPtr);

    parmsPtr->flexFileName = flFileName;

    return (BEAR_OK);

#undef FILE_NAME_SIZE
}

#endif /* BEAR_BLACK_BOX */

    
