
/* pbIVMiscWins.c --
 *
 *     George Carvalho, Esq.
 *     Benjamin Chen
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbIVMiscWins.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#if !defined(BEAR_BLACK_BOX)
/*
 * This file does not need to be compiled when the black box version of
 * the BEAR-FP system is built.
 */

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "routeDB.h"
#include "ezPlot.h"
#include "grEditor.h"
#include "cpaInterface.h"
#include "geoChDecomposer.h"
#include "localRouter.h"
#include "localVia.h"
#include "newG.h"
#include "ringRouter.h"
#include "tileMapping.h"
#include "bearIV.h"
#include "pbDialogs.h"
#include "polarBear.h"


typedef struct _FindData FindData;
struct _FindData {
    BearId          nodeId;
    Gt_GraphNode    node;
};


/* ---------- Local Procedures ---------- */
static BearStatus   CompareNode();
static void         ShowGraph(), DrawPath();
static void         PickBasePath();
static void         HiLiteTile();


void
PbAutoRoutePauseDialog(winDataPtr, checkConnFlag, keepGoingFlagPtr)
WindowData  *winDataPtr;
bool        checkConnFlag;
bool        *keepGoingFlagPtr;
{
    CellMaster  *cellMasterPtr;
    ChipWin     *chipWinPtr;
    Net         *netPtr;
    PinInst     *pinPtr;
    IVWindow    *ivWinPtr;
    XEvent      theEvent;
    bool        checkAllNetsFlag = FALSE;
    bool        checkOneNetFlag = FALSE;
    bool        contAutoRouteFlag = FALSE;
    bool        exitAutoRouteFlag = FALSE;

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);

    ivWinPtr = iv_InitWindow(ez_Display, BEAR_PROG_NAME, "Options", 6);
    if (checkConnFlag == TRUE) {
        (void) iv_AddVar(ivWinPtr, "Check All Nets", 
            &checkAllNetsFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(ivWinPtr, "Check One Net",
            &checkOneNetFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    }
    (void) iv_AddVar(ivWinPtr, "Continue",
        &contAutoRouteFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Exit",
        &exitAutoRouteFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_MapWindow(ivWinPtr, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

    while (TRUE) {
        XNextEvent(ez_Display, &theEvent);
        (void) iv_HandleEvent(&theEvent);
        Ez_XEventHandler(&theEvent);

        if ((checkConnFlag == TRUE) && (checkAllNetsFlag == TRUE))  {
            if (Gre_CheckConnForAllNets(cellMasterPtr, TRUE) == BEAR_OK) {
                PbPrint(EZ_NORMAL_FONT, "Every net is connected.\n");
            }

            checkAllNetsFlag = FALSE;
            continue;
        } 

        if ((checkConnFlag == TRUE) && (checkOneNetFlag == TRUE))  {
            if ((netPtr = PbGetNetOfCHW(FPG_CENTER, chipWinPtr)) == NIL(Net)) {
                /* 
                 * Use the mouse to locate the net, and then display it.
                 */
                if (Gre_PickNet(winDataPtr, &netPtr, &pinPtr) == BEAR_ERROR) {
                    checkOneNetFlag = FALSE;
                    continue;
                }
            }
            (void) Gre_CheckConnForOneNet(cellMasterPtr, netPtr, TRUE);

            checkOneNetFlag = FALSE;
            continue;
        }

        if (contAutoRouteFlag) {
            *keepGoingFlagPtr = TRUE;
            break;
        } 

        if (exitAutoRouteFlag) {
            *keepGoingFlagPtr = FALSE;
            break;
        }
    }
    iv_FreeWindow(ivWinPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * RouteChannelDialog() --
 *
 * Results:
 *   returns either BEAR_OK, BEAR_ERROR, or BEAR_ERROR_2.
 *
 *   If BEAR_ERROR_2 is returned the routing process should
 *   be stopped.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
#ifdef USE_DR_CODE
BearStatus 
RouteChannelDialog(rCellPtr, adjParPtr, LShapeFlag)
RouteCell   *rCellPtr;
AdjPar      *adjParPtr;  
bool        LShapeFlag;
{
    module *modulePtr;
    IVWindow *IVwin;
    XEvent theEvent;
    int heightRequired;
    int heightAvail;
    int heightTarget;
    int nut_compaction;
    static bool nutcrackerFlag = FALSE;
    static bool viaReduceFlag = TRUE;
    bool doRoute = FALSE;
    bool exitFlag = FALSE;
    
    int compactPar = 0;

    if (Rdb_MacToBBL(rCellPtr, &modulePtr) == BEAR_ERROR) {
        return (BEAR_ERROR_2);
    }
    rCellPtr->routeModPtr = modulePtr;
    modulePtr->placg = 2;
    if (Rdb_RouteChannelPass1(modulePtr, rCellPtr, adjParPtr) == BEAR_ERROR) {
        return (BEAR_ERROR_2);
    }
    heightRequired = Lr_GetChannelHeight(0);
    heightAvail = Lr_GetChannelHeight(1);
    heightTarget = heightAvail;

    IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME, "Channel Router", 13);
    if (LShapeFlag) {
        (void) iv_AddVar(IVwin, "X Adjustment",
            &(adjParPtr->xAdj), IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &(adjParPtr->xAdj), FALSE);
        (void) iv_AddVar(IVwin, "Y Adjustment",
            &(adjParPtr->yAdj), IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &(adjParPtr->yAdj), FALSE);
        nutcrackerFlag = FALSE;
    }
    else {
        (void) iv_AddVar(IVwin, "Available Channel Height", 
            &heightAvail, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightAvail, FALSE);
        (void) iv_AddVar(IVwin, "Required Channel Height", 
            &heightRequired, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightRequired, FALSE);
        (void) iv_AddVar(IVwin, "Target Channel Height",
            &heightTarget, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);

#ifdef USE_NUT_CODE
        (void) iv_AddVar(IVwin, "Use Nutcracker", 
            &nutcrackerFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
#endif USE_NUT_CODE
    }
    (void) iv_AddVar(IVwin, "Via Reduction", 
         &viaReduceFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(IVwin, "Okay", 
         &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(IVwin, "Exit",
        &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

    while (!doRoute) {
        XNextEvent(ez_Display, &theEvent);
        (void) iv_HandleEvent(&theEvent);
        Ez_XEventHandler(&theEvent);
        if (exitFlag) {
            iv_FreeWindow(IVwin);
            return BEAR_ERROR;
        }
    }
    iv_FreeWindow(IVwin);
    doRoute = FALSE;
    exitFlag = FALSE;
#ifdef USE_NUT_CODE
    if (nutcrackerFlag) {
        nut_compaction = Ls_nutcracker(modulePtr);
        heightRequired = MAX(heightRequired, heightAvail) - nut_compaction;
        IVwin = iv_InitWindow(ez_Display, BEAR_PROG_NAME, "Nutcracker", 13);
        (void) iv_AddVar(IVwin, "Available Channel Height", 
            &heightAvail, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightAvail, FALSE);
        (void) iv_AddVar(IVwin, "Required Channel Height", 
            &heightRequired, IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_SetEnableFlag(IVwin, &heightRequired, FALSE);
        (void) iv_AddVar(IVwin, "Target Channel Height", &heightTarget,
            IV_INT, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Continue", 
            &doRoute, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_AddVar(IVwin, "Exit", &exitFlag,
            IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
        (void) iv_MapWindow(IVwin, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);
        adjParPtr->xAdj = adjParPtr->yAdj = heightRequired - heightAvail;
        while (!doRoute) {
            XNextEvent(ez_Display, &theEvent);
            (void) iv_HandleEvent(&theEvent);
            Ez_XEventHandler(&theEvent);
            if (exitFlag) {
                iv_FreeWindow(IVwin);
                return BEAR_ERROR;
            }
        }
        iv_FreeWindow(IVwin);
        Ls_nutcrackerPass2(modulePtr, &compactPar, 
                           heightTarget, viaReduceFlag);
    }
    if (viaReduceFlag && nutcrackerFlag) {
        /* via reduction, if TRUE, is done in nutcracker module */
        Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr, compactPar,
                              FALSE);
    }
    else
#endif USE_NUT_CODE
    {
       Rdb_RouteChannelPass2(modulePtr, rCellPtr, adjParPtr, compactPar,
                              viaReduceFlag);
    }
    return BEAR_OK;
}
#endif /* USE_DR_CODE */



#ifdef DEBUG
void
PbSetDebugFlags()
{
    IVWindow    *ivWinPtr;
    XEvent      theEvent;
    bool        ivGeoChDecompFlag;
    bool        ivGtGraphFlag;
    bool        ivLocalRouterFlag;
    bool        ivLocalViaFlag;
    bool        ivPgRouteFlag;
    bool        ivPolarBearFlag;
    bool        ivRingRouterFlag;
    bool        ivRouteDBFlag;
    bool        ivTilePropFlag;
    bool        ivTmDupFlag;
    bool        allFlag;
    bool        okayFlag;
    bool        abortFlag;

    ivGeoChDecompFlag = lr_DebugFlag;
    ivGtGraphFlag = gt_DebugFlag;
    ivLocalRouterFlag = lr_DebugFlag;
    ivLocalViaFlag = lv_DebugFlag;
    ivPgRouteFlag = pgr_DebugFlag;
    ivPolarBearFlag = pbDebugFlag;
    ivRingRouterFlag = rr_DebugFlag;
    ivRouteDBFlag = rdb_DebugFlag;
    ivTilePropFlag = tp_DebugFlag;
    ivTmDupFlag = tm_DupFlag;
    allFlag = FALSE;
    okayFlag = FALSE;
    abortFlag = FALSE;

    ivWinPtr = iv_InitWindow(ez_Display, BEAR_PROG_NAME, "Debug Flags", 12);

    (void) iv_AddVar(ivWinPtr, "geoChDecomposer", 
        &ivGeoChDecompFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "gtGraph", 
        &ivGtGraphFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "localRouter", 
        &ivLocalRouterFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "localVia", 
        &ivLocalViaFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "pgRoute", 
        &ivPgRouteFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "polarBear", 
        &ivPolarBearFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "ringRouter", 
        &ivRingRouterFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "routeDB", 
        &ivRouteDBFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "tileProp",
        &ivTilePropFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "tmDupFlag",
        &ivTmDupFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "all",
        &allFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Okay",
        &okayFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Exit",
        &abortFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_MapWindow(ivWinPtr, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

    while (TRUE) {
        XNextEvent(ez_Display, &theEvent);
        (void) iv_HandleEvent(&theEvent);
        Ez_XEventHandler(&theEvent);

        if (okayFlag == TRUE) {
            gcd_DebugFlag = ivGeoChDecompFlag;
            gt_DebugFlag = ivGtGraphFlag;
            lr_DebugFlag = ivLocalRouterFlag;
            lv_DebugFlag = ivLocalViaFlag;
            pgr_DebugFlag = ivPgRouteFlag;
            pbDebugFlag = ivPolarBearFlag;
            rr_DebugFlag = ivRingRouterFlag;
            rdb_DebugFlag = ivRouteDBFlag;
            tp_DebugFlag = ivTilePropFlag;
            tm_DupFlag = ivTmDupFlag;
            break;
        }

        if (abortFlag == TRUE) {
            break;
        }

        if (allFlag == TRUE) {
            ivGeoChDecompFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivGeoChDecompFlag);

            ivGtGraphFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivGtGraphFlag);

            ivLocalRouterFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivLocalRouterFlag);

            ivLocalViaFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivLocalViaFlag);

            ivPgRouteFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivPgRouteFlag);

            ivPolarBearFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivPolarBearFlag);

            ivRingRouterFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivRingRouterFlag);

            ivRouteDBFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivRouteDBFlag);

            ivTilePropFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivTilePropFlag);

            ivTmDupFlag = TRUE;
            (void) iv_UpdateVar(ivWinPtr, &ivTmDupFlag);

            allFlag = FALSE;
            (void) iv_UpdateVar(ivWinPtr, &allFlag);
            continue;
        } 
    }
    iv_FreeWindow(ivWinPtr);
}

#endif DEBUG


static BearStatus
CompareNode(node, findDataPtr)
Gt_GraphNode    node;
FindData        *findDataPtr;
{
    if (findDataPtr->nodeId == Gt_GetIdOfNode(node)) {
        findDataPtr->node = node;
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


static void
ShowGraph(winDataPtr, planeDataPtr, graph, netPtr, labelFlag)
WindowData  *winDataPtr;
PlaneData   *planeDataPtr;
Gt_Graph    graph;
Net         *netPtr;
bool        labelFlag;
{
    Mac_GraphData *graphDataPtr;

    if (graph == (Gt_Graph) NULL) {
        return;
    }

    Mac_PutAutoGraphOfNet(graph, netPtr);

    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

    graphDataPtr = (Mac_GraphData *) Gt_GetGraphDataOfGraph(graph);
    if (graphDataPtr->junctList != (lsList) NULL) {
        DrawData        drawData;
        lsGen           gen;
        WallJunction    *junctPtr;

        drawData.winDataPtr = winDataPtr;
        drawData.color = ez_HighLightColor;
        drawData.width = MAXWIDTH;
        lsForeachItem(graphDataPtr->junctList, gen, junctPtr) {
            (void) Gre_DrawWallJunct(junctPtr, &drawData);
        }
    }

    if (labelFlag == TRUE) {
        LabelNet labelNet;

        labelNet.netPtr = netPtr;
        labelNet.planeDataPtr = planeDataPtr;
        labelNet.whichNet = FPG_CENTER;
        Gre_LabelNet(winDataPtr, (ClientData) &labelNet);
    }
}


static void
DrawPath(winDataPtr, planeDataPtr, pathPtr)
WindowData  *winDataPtr;
PlaneData   *planeDataPtr;
Gt_Path     *pathPtr;
{
    DrawGraphEdge   drawEdge;
    DrawGraphNode   drawNode;
    Gt_PathEdge     *pathEdgePtr;
    lsGen           gen;

    drawNode.planeDataPtr = planeDataPtr;
    drawNode.nodeColor = ez_HighLightColor;
    drawNode.dotFlag = FALSE;

    drawEdge.planeDataPtr = planeDataPtr;
    drawEdge.edgeColor = ez_HighLightColor;
    drawEdge.dotFlag = FALSE;
    drawEdge.widthFlag = FALSE;

    lsForeachItem(pathPtr->pathList, gen, pathEdgePtr) {
        drawEdge.edge = pathEdgePtr->edge;
        (void) Gre_DrawGraphEdge(winDataPtr, &drawEdge);

        drawNode.node = pathEdgePtr->node1;
        (void) Gre_DrawGraphNode(winDataPtr, &drawNode);
        drawNode.node = pathEdgePtr->node2;
        (void) Gre_DrawGraphNode(winDataPtr, &drawNode);
    }
}


static void
PickBasePath(winDataPtr, planeDataPtr, baseGraph)
WindowData  *winDataPtr;
PlaneData   *planeDataPtr;
Gt_Graph    baseGraph;
{
    DrawGraphNode   drawNode;
    FindData        findData;
    Gt_AlgData      algData;
    Gt_GraphNode    node1, node2;
    Gt_Path         *pathPtr;
    Point           screen;
    Window          win;
    char            answer[80];

    PbPrint(EZ_NORMAL_FONT, "Enter id of node 1 : ");
    (void) Ez_GetLine(&win, &screen, answer, 80, &pbCmdLoop);
    sscanf (answer, "%d", &findData.nodeId);
    findData.node = (Gt_GraphNode) NULL;
    (void) Gt_ItrNodesOfGraph(baseGraph, CompareNode,
        (ClientData) &findData);
    if ((node1 = findData.node) == (Gt_GraphNode) NULL) {
        PbPrint(EZ_ITALIC_FONT, "Cannot find that node.\n");
        return;
    }
    drawNode.node = node1;
    drawNode.planeDataPtr = planeDataPtr;
    drawNode.nodeColor = PbGetColorOfNet(FPG_CENTER);
    drawNode.dotFlag = FALSE;
    (void) Gre_DrawGraphNode(winDataPtr, &drawNode);

    PbPrint(EZ_NORMAL_FONT, "Enter id of node 2 : ");
    (void) Ez_GetLine(&win, &screen, answer, 80, &pbCmdLoop);
    sscanf (answer, "%d", &findData.nodeId);
    findData.node = (Gt_GraphNode) NULL;
    (void) Gt_ItrNodesOfGraph(baseGraph, CompareNode,
        (ClientData) &findData);
    if ((node2 = findData.node) == (Gt_GraphNode) NULL) {
        PbPrint(EZ_ITALIC_FONT, "Cannot find that node.\n");
        return;
    }
    drawNode.node = node2;
    (void) Gre_DrawGraphNode(winDataPtr, &drawNode);

    algData = Gt_GenAlgData(baseGraph, Ngr_GetBaseEdgeWeight);
    (void) Gt_ShortestPath(baseGraph, algData, node1, node2, TRUE, &pathPtr);
    DrawPath(winDataPtr, planeDataPtr, pathPtr);
    Gt_FreePath(&pathPtr);
    Gt_FreeAlgData(baseGraph, &algData);
}


BearStatus
PbSteinerTreeDialog(winDataPtr, netPtr)
WindowData  *winDataPtr;
Net         *netPtr;
{
    Ngr_BldTreeDataPtr  bldTreePtr;
    CellMaster          *cellMasterPtr;
    ChipWin             *chipWinPtr;
    Gt_Graph            baseGraph, compGraph;
    PinInst             *pinPtr;
    PlaneData           *planeDataPtr;
    IVWindow            *ivWinPtr;
    XEvent              theEvent;
    bool                netFlag;
    bool                pathFlag;
    bool                labelBaseFlag, labelCompFlag;
    bool                printBaseFlag, printCompFlag;
    bool                step1Flag, step2Flag, step3Flag, step4Flag;
    bool                all4Flag;
    bool                doneFlag;

    netFlag = FALSE;
    step1Flag = FALSE;
    step2Flag = FALSE;
    step3Flag = FALSE;
    step4Flag = FALSE;
    all4Flag = FALSE;
    pathFlag = FALSE;
    labelBaseFlag = FALSE;
    labelCompFlag = FALSE;
    printBaseFlag = FALSE;
    printCompFlag = FALSE;
    doneFlag = FALSE;

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    ivWinPtr = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
        "Steiner Tree Package", 6);

    (void) iv_AddVar(ivWinPtr, "Choose net with mouse         ",
        &netFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "1.) Base graph                ", 
        &step1Flag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "2.) Complete graph            ", 
        &step2Flag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "3.) Find minimum spanning tree", 
        &step3Flag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "4.) Find steiner tree         ", 
        &step4Flag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Do all four steps             ", 
        &all4Flag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Label base graph              ", 
        &labelBaseFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Label complete graph          ", 
        &labelCompFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Shortest path in base graph   ", 
        &pathFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Print base graph              ", 
        &printBaseFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Print complete graph          ", 
        &printCompFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_AddVar(ivWinPtr, "Exit                          ", 
        &doneFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_SetEnableFlag(ivWinPtr, (int *) &step1Flag, 1);
    (void) iv_SetEnableFlag(ivWinPtr, (int *) &step2Flag, 0);
    (void) iv_SetEnableFlag(ivWinPtr, (int *) &step3Flag, 0);
    (void) iv_SetEnableFlag(ivWinPtr, (int *) &step4Flag, 0);
    (void) iv_SetEnableFlag(ivWinPtr, (int *) &all4Flag,  1);

    (void) iv_MapWindow(ivWinPtr, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

    if (Ngr_BldTreeBegin(cellMasterPtr, NIL(Ngr_BldOptions), &bldTreePtr) !=
        BEAR_OK) {
        return (BEAR_ERROR);
    }
    baseGraph = (Gt_Graph) NULL;
    compGraph = (Gt_Graph) NULL;
    while (TRUE) {
        XNextEvent(ez_Display, &theEvent);
        (void) iv_HandleEvent(&theEvent);
        Ez_XEventHandler(&theEvent);

        if (netFlag == TRUE) {
            netFlag = FALSE;

            Mac_PutAutoGraphOfNet((Gt_Graph) NULL, netPtr);
            Ngr_DeleteBaseGraph(&baseGraph);
            Ngr_DeleteCompGraph(&compGraph);

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step1Flag, 1);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step2Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step3Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step4Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &all4Flag,  1);

            /*
             * At this point baseGraph = compGraph = NULL, and we
             * use the mouse to locate a net, and then display it.
             */
            if (Gre_PickNet(winDataPtr, &netPtr, &pinPtr) == BEAR_ERROR) {
                Ez_RingBell();
                netPtr = NIL(Net);
            }
            continue;
        }

        if (step1Flag == TRUE) {
            step1Flag = FALSE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step1Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &all4Flag,  0);
            if (Ngr_BuildBaseGraph(bldTreePtr, netPtr, &baseGraph) == BEAR_OK) {
                ShowGraph(winDataPtr, planeDataPtr, baseGraph, netPtr, FALSE);

                (void) iv_SetEnableFlag(ivWinPtr, (int *) &step2Flag, 1);
                (void) iv_SetEnableFlag(ivWinPtr, (int *) &step3Flag, 1);
                (void) iv_SetEnableFlag(ivWinPtr, (int *) &step4Flag, 1);
                PbPrint(EZ_NORMAL_FONT, "Step 1 is done.\n");
            }
            else {
                PbPrint(EZ_ITALIC_FONT, "Cannot build the base graph.\n");
            }

            continue;
        }

        if (step2Flag == TRUE) {
            step2Flag = FALSE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step2Flag, 0);
            if (Ngr_BuildCompGraph(bldTreePtr, &compGraph) == BEAR_OK) {
                ShowGraph(winDataPtr, planeDataPtr, compGraph, netPtr, FALSE);

                PbPrint(EZ_NORMAL_FONT, "Step 2 is done.\n");
            }
            else {
                PbPrint(EZ_ITALIC_FONT, "Cannot build the complete graph.\n");
            }
            continue;
        }

        if (step3Flag == TRUE) {
            step3Flag = FALSE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step3Flag, 0);
            if (Ngr_FindMST(bldTreePtr) == BEAR_OK) {
                ShowGraph(winDataPtr, planeDataPtr, compGraph, netPtr, FALSE);

                PbPrint(EZ_NORMAL_FONT, "Step 3 is done.\n");
            }
            continue;
        }

        if (step4Flag == TRUE) {
            step4Flag = FALSE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step4Flag, 0);
            if (Ngr_FindSteinerTree(bldTreePtr) == BEAR_OK) {
                ShowGraph(winDataPtr, planeDataPtr, baseGraph, netPtr, FALSE);
                PbPrint(EZ_NORMAL_FONT, "Step 4 is done.\n");
            }
            continue;
        }

        if (all4Flag == TRUE) {
            all4Flag = FALSE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step1Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step2Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step3Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &step4Flag, 0);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &all4Flag,  0);

            if (Ngr_BuildBaseGraph(bldTreePtr, netPtr, &baseGraph) ==
                BEAR_ERROR) {
                continue;
            }
            PbPrint(EZ_NORMAL_FONT, "Step 1 is done.\n");

            if (Ngr_BuildCompGraph(bldTreePtr, &compGraph) == BEAR_ERROR) {
                continue;
            }
            PbPrint(EZ_NORMAL_FONT, "Step 2 is done.\n");

            if (Ngr_FindMST(bldTreePtr) == BEAR_ERROR) {
                continue;
            }
            PbPrint(EZ_NORMAL_FONT, "Step 3 is done.\n");

            if (Ngr_FindSteinerTree(bldTreePtr) == BEAR_ERROR) {
                continue;
            }
            PbPrint(EZ_NORMAL_FONT, "Step 4 is done.\n");

            ShowGraph(winDataPtr, planeDataPtr, baseGraph, netPtr, FALSE);

            continue;
        }

        if (labelBaseFlag == TRUE) {
            labelBaseFlag = FALSE;

            if (baseGraph == (Gt_Graph) NULL) {
                PbPrint(EZ_ITALIC_FONT, "Must build base graph first.\n");
                continue;
            }
            ShowGraph(winDataPtr, planeDataPtr, baseGraph, netPtr, TRUE);

            continue;
        }

        if (labelCompFlag == TRUE) {
            LabelNet labelNet;

            labelCompFlag = FALSE;

            if (compGraph == (Gt_Graph) NULL) {
                PbPrint(EZ_ITALIC_FONT, "Must build complete graph first.\n");
                continue;
            }
            ShowGraph(winDataPtr, planeDataPtr, compGraph, netPtr, TRUE);

            continue;
        }

        if (pathFlag == TRUE) {
            pathFlag = FALSE;

            if (baseGraph == (Gt_Graph) NULL) {
                PbPrint(EZ_ITALIC_FONT, "Must build base graph first.\n");
                continue;
            }
            PickBasePath(winDataPtr, planeDataPtr, baseGraph);

            continue;
        }

        if (printBaseFlag == TRUE) {
            printBaseFlag = FALSE;

            if (baseGraph == (Gt_Graph) NULL) {
                PbPrint(EZ_ITALIC_FONT, "Must build base graph first.\n");
                continue;
            }

            Ngr_WriteGraph("base.graph", baseGraph, Ngr_GetBaseEdgeWeight);
            PbPrint(EZ_NORMAL_FONT, 
                "Base graph printed to file named \"base.graph\"\n");
            continue;
        }

        if (printCompFlag == TRUE) {
            printCompFlag = FALSE;

            if (compGraph == (Gt_Graph) NULL) {
                PbPrint(EZ_ITALIC_FONT, "Must build complete graph first.\n");
                continue;
            }

            Ngr_WriteGraph("comp.graph", compGraph, Ngr_GetCompEdgeWeight);
            PbPrint(EZ_NORMAL_FONT, 
                "Complete graph printed to file named \"comp.graph\"\n");
            continue;
        }

        if (doneFlag == TRUE) {
            Ngr_DeleteCompGraph(&compGraph);
            break;
        }
    }
    (void) Ngr_BldTreeEnd(bldTreePtr);
    iv_FreeWindow(ivWinPtr);

    return (BEAR_OK);
}


static void
HiLiteTile(winDataPtr, planeDataPtr, tilePtr)
WindowData  *winDataPtr;
PlaneData   *planeDataPtr;
Tile        *tilePtr;
{
    Rect tileRect;

    if (Tp_IsHorzTile(tilePtr) == TRUE) {
        Tp_HorzTileToRect(planeDataPtr, tilePtr, &tileRect);
    }
    else {
        Tp_VertTileToRect(planeDataPtr, tilePtr, &tileRect);
    }

    Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &tileRect, EZ_SOLIDLINE1, ez_HighLightColor, pbChipColor);
    Ez_StippleRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        &tileRect, ez_HighLightColor, ez_CargoNetBitmap, GXcopy);
}


void
PbCpaDialog(winDataPtr)
WindowData *winDataPtr;
{
    CellMaster      *cellMasterPtr;
    ChipWin         *chipWinPtr;
    Cpa_ArrPinData  optData;        /* optional data */
    PlaneData       *planeDataPtr;
    Tile            *bnTilePtr;
    IVWindow        *ivWinPtr;
    XEvent          theEvent;
    int             oldOptions, newOptions;
    bool            pickFlag, oneFlag, allFlag, saveFlag;
    bool            rectVisibleFlag;
    bool            exitFlag;

    pickFlag = FALSE;
    oneFlag = FALSE;
    allFlag = FALSE;
    saveFlag = FALSE;
    exitFlag = FALSE;

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    oldOptions = chipWinPtr->chw_options;
    if (((chipWinPtr->chw_options & PB_SHOW_HORZ_PLANE_OPT) == FALSE) ||
        ((chipWinPtr->chw_options & PB_SHOW_VERT_PLANE_OPT) == FALSE) ||
        ((chipWinPtr->chw_options & PB_SHOW_BN_TILES_OPT) == FALSE)) {
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_HORZ_PLANE_OPT);
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_VERT_PLANE_OPT);
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_BN_TILES_OPT);
        (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
    }
    newOptions = chipWinPtr->chw_options;

    ivWinPtr = iv_InitWindow(ez_Display, BEAR_PROG_NAME,
        "Channel Pin Arrangement Package", 12);

/*
 * Posible additions:
 *    1.) A button to change all the pins to floating.
 */

    (void) iv_AddVar(ivWinPtr, "Choose tile with mouse ",
        &pickFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Arrange pins in bn tile",
        &oneFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
    (void) iv_AddVar(ivWinPtr, "Do all bn tiles        ",
        &allFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);
#ifdef DEBUG
    (void) iv_AddVar(ivWinPtr, "Save data              ",
        &saveFlag, IV_BOOLEAN, IV_NO_OPT, IV_NULL_FUNCTION);
#endif DEBUG

    (void) iv_AddVar(ivWinPtr, "Exit                   ", 
        &exitFlag, IV_TOGGLE, IV_NO_OPT, IV_NULL_FUNCTION);

    (void) iv_SetEnableFlag(ivWinPtr, (int *) &oneFlag,  0);

    (void) iv_MapWindow(ivWinPtr, IV_NO_OPT, IV_NO_OPT, IV_MOUSE);

    bnTilePtr = NIL(Tile);
    rectVisibleFlag = FALSE;
    optData.cpaProgPtr = NIL(char);
    optData.tmpDirPtr = bear_tmpDir;
    optData.warnFlag = TRUE;
    while (TRUE) {
        XNextEvent(ez_Display, &theEvent);
        (void) iv_HandleEvent(&theEvent);
        Ez_XEventHandler(&theEvent);

        if (pickFlag == TRUE) {
            ezStatus        findStatus;
            Pb_DefCurData   defCurData;
            Point           screen;

            pickFlag = FALSE;

            bnTilePtr = NIL(Tile);
            (void) iv_SetEnableFlag(ivWinPtr, (int *) &oneFlag,  0);

            /*
             * Change the cursor to a circle in all of the chip windows.
             */
            defCurData.idWinProc = PbIsChipWin;
            defCurData.cursorType = (int) EZ_CIRCLE_OUTLINE_CUR;
            (void) PbItrChipWindows(PbDefCurInUserSubWin,
                (ClientData) &defCurData);
            Ez_Flush();
                          
           /*
            * Wait for the user to click the mouse in a chip window, which
            * will indicate the location of the tile.
            */
            findStatus = Ez_FindUserWin(LeftButton, PbIsChipWin,
                (ClientData) NULL, &winDataPtr, &screen);
                          
           /*
            * Change the cursor back to a cross in all of the chip windows.
            */
            defCurData.cursorType = (int) EZ_CROSS_CUR;
            (void) PbItrChipWindows(PbDefCurInUserSubWin,
                (ClientData) &defCurData);
            Ez_Flush();
                          
            if (findStatus != EZ_OK) {
                PbPrint(EZ_ITALIC_FONT,
                    "Must pick a point in a chip window.\n");
                continue;
            }

            if (PbFindTile(winDataPtr, &screen, &bnTilePtr,
                HORZ_TILE_PLANE) == FALSE) {
                PbPrint(EZ_ITALIC_FONT, 
                    "You may only choose bottleneck tiles.\n");
                Ez_RingBell();
                continue;
            }
            else if (Tp_IsBottleneckTile(bnTilePtr) == FALSE) {
                if ((PbFindTile(winDataPtr, &screen, &bnTilePtr,
                    VERT_TILE_PLANE) == FALSE) ||
                    (Tp_IsBottleneckTile(bnTilePtr) == FALSE)) {
                    PbPrint(EZ_ITALIC_FONT, 
                        "You may only choose bottleneck tiles.\n");
                    Ez_RingBell();
                    continue;
                }
            }

            if (rectVisibleFlag == TRUE) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
            }
            HiLiteTile(winDataPtr, planeDataPtr, bnTilePtr);
            rectVisibleFlag = TRUE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &oneFlag,  1);
                    
            continue;
        }

        if (oneFlag == TRUE) {
            oneFlag = FALSE;

            (void) iv_SetEnableFlag(ivWinPtr, (int *) &oneFlag,  0);

            if ((bnTilePtr == NIL(Tile)) ||
                (Tp_IsBottleneckTile(bnTilePtr) == FALSE)) {
                PbPrint(EZ_ITALIC_FONT, "Bad tile.\n");
                continue;
            }

            optData.saveFlag = saveFlag;
            if (Cpa_ArrPinsForOneBnTile(cellMasterPtr, bnTilePtr, &optData) !=
                BEAR_OK) {
                PbPrint(EZ_ITALIC_FONT,
                    "Unable to do pin arrangement for the bn tile.\n");
                bnTilePtr = NIL(Tile);
                continue;
            }
            (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

            HiLiteTile(winDataPtr, planeDataPtr, bnTilePtr);
            rectVisibleFlag = TRUE;

            continue;
        }

        if (allFlag == TRUE) {
            allFlag = FALSE;

            optData.saveFlag = saveFlag;
            if (Cpa_ArrPinsForAllBnTiles(cellMasterPtr, &optData) != BEAR_OK) {
                PbPrint(EZ_ITALIC_FONT,
                    "Unable to do pin arrangement for all the bn tiles.\n");
                continue;
            }
            (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

            PbPrint(EZ_NORMAL_FONT, "Channel pin arrangement is done.\n");
            continue;
        }

        if (exitFlag == TRUE) {
            break;
        }
    }
    iv_FreeWindow(ivWinPtr);

    if (newOptions != oldOptions) {
        chipWinPtr->chw_options = oldOptions;
    }
    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
}

#endif /* BEAR_BLACK_BOX */

