
/* pbMisc.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbMisc.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include <sys/stat.h>
#include <sys/time.h>
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "bfpInterface.h"

#ifdef USE_BBL_CODE
#include "bblInterface.h"
#endif /* USE_BBL_CODE */

#ifdef USE_OCT_CODE
#include "octInterface.h"
#endif /* USE_OCT_CODE */

#include "ezPlot.h"

#ifdef USE_OLDP_CODE
#include "placer.h"
#endif /* USE_OLDP_CODE */

#ifdef USE_NEWP_CODE
#include "newC.h"
#endif /* USE_NEWP_CODE */

#include "routeDB.h"
#include "localRouter.h"
#include "polarBear.h"


/* ---------- Local Procedures ---------- */
static BearStatus   AddTilePlanes();



static BearStatus
AddTilePlanes(coreMasterPtr)
CellMaster *coreMasterPtr;
{
    CellInst    *coreInstPtr;
    CellMaster  *chipMasterPtr;
    PlaneData   *chipPlaneDataPtr, *corePlaneDataPtr;

   /*
    * Get the pointer to the core cell instance from
    * the core cell master.
    */
    coreInstPtr = Mac_GetInstOfCoreMaster(coreMasterPtr);

    /*
     * Get the pointer to the chip cell master from
     * the core cell instance.
     */
    chipMasterPtr = Mac_GetParentOfCell(coreInstPtr);

    Mac_PutPlacedFlagOfCell(TRUE, chipMasterPtr);
    if (Mac_SetUpTiPlane(chipMasterPtr, TRUE) != BEAR_OK) {
        PbPrint(EZ_ITALIC_FONT,
      "diasaster struck while building the tile plane for the chip\n");
        return (BEAR_ERROR);
    }

    if (Mac_SetUpTiPlane(coreMasterPtr, TRUE) != BEAR_OK) {
        PbPrint(EZ_ITALIC_FONT,
      "diasaster struck while building the tile plane for the core\n");
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}



int
PbCompareCommands(cmd1Ptr, cmd2Ptr)
Command *cmd1Ptr, *cmd2Ptr;
{
    return (strcmp(cmd1Ptr->co_shortName, cmd2Ptr->co_shortName));
}



Command *
PbFindCommand(cmdHashTablePtr, cmdNamePtr)
HashTable   *cmdHashTablePtr;
char        *cmdNamePtr;
{
    HashEntry *hashEntryPtr;

    return (((hashEntryPtr = HashLookOnly(cmdHashTablePtr, cmdNamePtr)) ==
            NIL(HashEntry)) ? NIL(Command) :
            (Command *) hashEntryPtr->h_pointer);
}



char *
PbGetDate()
/*
 * Returns a string containing the current date.
 */
{
    struct timeval  timeValue;
    struct timezone timeZone;
    int             i, length;
    char            *datePtr, *ctime ();

    if (gettimeofday (&timeValue, &timeZone) != 0) {
        fprintf (stderr, "%s: gettimeofday() failed\n", BEAR_PROG_NAME);
        perror (BEAR_PROG_NAME);
        return (NIL(char));
    } 

#if defined(__STDC__)
    datePtr = ctime ((time_t *) &timeValue.tv_sec);
#else
    datePtr = ctime ((long *) &timeValue.tv_sec);
#endif /* __STDC__ */

    length = strlen (datePtr);
    for (i = 0; i < length ;i++) {
        if (datePtr[i] == '\n') {
            datePtr[i] = '\0';
        }
    }
    return (datePtr);
}



BearStatus
PbStoreCommand(cmdHashTablePtr, cmdPtr)
HashTable   *cmdHashTablePtr;
Command     *cmdPtr;
{
    if (cmdPtr->co_shortName == NIL(char)) {

        Bear_PrintErr("PbStoreCommand() : command has no name !");
        return (BEAR_ERROR);
    } 
    if (PbFindCommand(cmdHashTablePtr, cmdPtr->co_shortName) !=
        NIL(Command)) {

        sprintf (bear_ErrorMsg, "%s command already exists",
            cmdPtr->co_shortName);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR);
    } 
    if (cmdPtr->co_propList & (PB_SCRIPT_CP | PB_BLACK_BOX_CP)) {
        if (cmdPtr->co_statusProc == (BearStatus (*)()) NULL) { 
            sprintf (bear_ErrorMsg,
                "%s command must have a procedure that returns status",
                cmdPtr->co_shortName);
            Bear_PrintErr(bear_ErrorMsg);
            return (BEAR_ERROR);
        }
    }
    else {
        if (cmdPtr->co_voidProc == (void (*)()) NULL) { 
            sprintf (bear_ErrorMsg,
                "%s command must have a procedure that returns void",
                cmdPtr->co_shortName);
            Bear_PrintErr(bear_ErrorMsg);
            return (BEAR_ERROR);
        }
    }
    if ((cmdPtr->co_voidProc != (void (*)()) NULL) &&
        (cmdPtr->co_statusProc != (BearStatus (*)()) NULL)) { 

        sprintf (bear_ErrorMsg,
            "%s command may only have one procedure to handle it",
            cmdPtr->co_shortName);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR);
    }
    (HashFind(cmdHashTablePtr, cmdPtr->co_shortName))->h_pointer =
        (char *) cmdPtr;
    return (BEAR_OK);
}



void
BuildFloorPlan(cellMasterPtr, dummyCells)
CellMaster  *cellMasterPtr;
bool        dummyCells;
{
    PlaneData *planeDataPtr;

    if (Mac_IsNumOfCellsZero(cellMasterPtr) == FALSE) {
        Mac_DeleteAllGraphs(Mac_GetNetListOfCell(cellMasterPtr));

        planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
        if (dummyCells || (Tp_GetDCellList(planeDataPtr) != NIL(DummyCell))) {
            Tp_DeleteDCells(planeDataPtr);

            Tp_IdentifySpaceTiles(planeDataPtr);
            if (Tp_CreateDCells(planeDataPtr) != BEAR_OK) {
                Bear_PrintFatalErr("BuildDCells() failed");
            } 

            /*
             * Paint all dummy cells onto the tile planes.
             */
            (void) Mac_ItrDCell(Tp_GetDCellList(planeDataPtr),
                Tp_PaintDCell, (ClientData) planeDataPtr);

            /*
             * Painting the dummy cells onto the tile plane disrupts
             * most of the space tiles, so memory must be re-allocated
             * for each one of them and then they must all be
             * re-identified.
             */
            Tp_IdentifySpaceTiles(planeDataPtr);
        }
        else {
            Tp_IdentifySpaceTiles(planeDataPtr);
        }

        if (Fpg_CreateFloorPlanGraph(planeDataPtr, (WallJunction **) NULL) ==
            BEAR_ERROR) {

            PbPrint(EZ_ITALIC_FONT, "Could not make floor plan graph\n");
            return;
        }
    }
}



void
BuildDCells(cellMasterPtr)
CellMaster *cellMasterPtr;
{
    PlaneData *planeDataPtr;

    Mac_DeleteAllGraphs(Mac_GetNetListOfCell(cellMasterPtr));

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    Tp_DeleteDCells(planeDataPtr);

    Tp_IdentifySpaceTiles(planeDataPtr);
    if (Tp_CreateDCells(planeDataPtr) != BEAR_OK) {
        Bear_PrintFatalErr("BuildDCells() failed");
    } 

    /*
     * Paint all dummy cells onto the tile planes.
     */
    (void) Mac_ItrDCell(Tp_GetDCellList(planeDataPtr), Tp_PaintDCell,
        (ClientData) planeDataPtr);

    /*
     * Painting the dummy cells onto the tile plane disrupts
     * most of the space tiles, so memory must be re-allocated
     * for each one of them and then they must all be
     * re-identified.
     */
    Tp_IdentifySpaceTiles(planeDataPtr);
}



BearStatus
PbRunBlackBox(scriptFileNamePtr)
char *scriptFileNamePtr;
{
    BearStatus bearStatus;

    bearStatus = PbSource(NIL(WindowData), scriptFileNamePtr,
        FALSE, FALSE);
    Bear_RemoveTmpFiles();
    
    return (bearStatus);
}



#ifndef BEAR_BLACK_BOX
BearStatus
PbRunCommandLoop()
{
    Point       screen;
    Window      win;
    WindowData  *winDataPtr;
    char        tbText[80]; /* title bar text */
    char        consoleLine[PB_CMD_LINE_LEN];

    Ez_EmptyXEventQue();
    (void) sprintf (tbText, "BEAR-FP ( %s ) : CONSOLE", pbHostName);
    if (Ez_CreateConsoleWin(tbText, &winDataPtr) != EZ_OK) {
        return (BEAR_ERROR);
    }
    winDataPtr->wd_userWinButtonProc = PbHandleUserWinButton;
    PbPrint(EZ_NORMAL_FONT, PbGetVersion());

    pbCmdLoop = TRUE;
    while (pbCmdLoop) {
        pbBusyFlag = FALSE;
        PbPrompt(pbConsPrompt);
        while (Ez_GetLine(&win, &screen,
            consoleLine, PB_CMD_LINE_LEN, &pbCmdLoop) != EZ_OK) {
            ;
        }
        if ((winDataPtr = Ez_GetWinDataOfWin(win)) != NIL(WindowData)) {
            if (index (consoleLine, EOF) != 0) {
                PbCloseWindowCmd(winDataPtr, &screen, 0, (char **) NULL);
                PbPrint(EZ_NORMAL_FONT, "\n");
                continue;
            }
            (void) PbExecCommand(winDataPtr, &screen, consoleLine);
        }
    }
    return (BEAR_OK);
}
#endif BEAR_BLACK_BOX



BearStatus
PbExecCommand(winDataPtr, screenPtr, line)
WindowData  *winDataPtr;
Point       *screenPtr;
char        line[];
/*
 * Execute command.
 */
{
#define MAX_NUM_OF_ARGS 25

    BearStatus  bearStatus;
    Command     *cmdPtr;
    int         argCount, i, j, length;
    char        *argList[MAX_NUM_OF_ARGS];
    bool        start;

    pbBusyFlag = TRUE;

    i = 0;
    start = TRUE;
    length = strlen (line);
    for (j = 0; j < length ;j++) {
        if (start && (isspace(line[j]) == FALSE)) {
            argList[i++] = &line[j];
            if ((i < MAX_NUM_OF_ARGS) == FALSE) {
                Bear_PrintErr("PbExecCommand() : too many args");
                exit (1);
            }
            start = FALSE;
        } else if ((start == FALSE) && isspace(line[j])) {
            line[j] = '\0';
            start = TRUE;
        }
    }
    argCount = i;
    if (argCount == 0) {
        pbBusyFlag = FALSE;
        return (BEAR_OK); /* No input */
    }

    if ((cmdPtr = PbFindCommand(&pbCmdHashTable, argList[0])) ==
        NIL(Command)) {

        (void) sprintf (bear_ErrorMsg, "%s : Command not found.\n", argList[0]);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
#ifndef BEAR_BLACK_BOX
        Ez_RingBell();
#endif BEAR_BLACK_BOX
        return (BEAR_ERROR);
    }

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        if (cmdPtr->co_validWins == PB_NULL_WIN) {
            sprintf (bear_ErrorMsg,
                "%s: %s%s",
                argList[0],
                "command may only be used when the program\n",
                "    is run in the black box mode.\n");
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
            pbBusyFlag = FALSE;
            return (BEAR_ERROR);
        }

        if (pbScriptFlag) {
            if ((cmdPtr->co_propList & PB_SCRIPT_CP) == FALSE) {
                sprintf (bear_ErrorMsg,
                    "%s: command may not be used in a script\n", argList[0]);
                PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            } 
        }

        if ((screenPtr == NIL(Point)) &&
            (cmdPtr->co_propList & PB_MOUSE_LOC_CP)) {

            sprintf (bear_ErrorMsg,
                "%s: command needs location of mouse\n", argList[0]);
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
            pbBusyFlag = FALSE;
            return (BEAR_ERROR);
        }

        if (PbIsCifWin(winDataPtr)) {
            /* Mouse is in a cif window. */
        
            if ((cmdPtr->co_validWins & PB_CIF_WIN) == FALSE) {
                PbPrint(EZ_ITALIC_FONT,
                    "Command not valid in cif window.\n");
                Ez_RingBell();
                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            }
        }
        else if (PbIsChipWin(winDataPtr)) {
            /* Mouse is in a chip window. */
            
            if ((cmdPtr->co_validWins & PB_CHIP_WIN) == FALSE) {
                PbPrint(EZ_ITALIC_FONT,
                    "Command not valid in chip window.\n");
                Ez_RingBell();
                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            }
        }
        else if (Ez_IsConsoleWin(winDataPtr)) {
            /* Mouse is in the console window. */
        
            if ((cmdPtr->co_validWins & PB_CONSOLE_WIN) == FALSE) {
                PbPrint(EZ_ITALIC_FONT,    
                    "Command not valid in console window.\n");
                Ez_RingBell();
                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            }
        }
#ifdef USE_OLDP_CODE
        else if (PbIsFloorWin(winDataPtr)) {
            /* Mouse is in a floor window. */
        
            if ((cmdPtr->co_validWins & PB_FLOOR_WIN) == FALSE) {
                PbPrint(EZ_ITALIC_FONT,
                    "Command not valid in floor window.\n");
                Ez_RingBell();
                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            }
        }
        else if (PbIsTreeWin(winDataPtr)) {
            /* Mouse is in a tree window. */
        
            if ((cmdPtr->co_validWins & PB_TREE_WIN) == FALSE) {
                PbPrint(EZ_ITALIC_FONT,
                    "Command not valid in tree window.\n");
                Ez_RingBell();
                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            }
        }
#endif USE_OLDP_CODE
        else if (PbIsPlaceWin(winDataPtr)) {
            /* Mouse is in a placement window. */
        
            if ((cmdPtr->co_validWins & PB_PLACEMENT_WIN) == FALSE) {
                PbPrint(EZ_ITALIC_FONT,
                    "Command not valid in placement window.\n");
                Ez_RingBell();
                pbBusyFlag = FALSE;
                return (BEAR_ERROR);
            }
        }
        else {
            Bear_PrintErr("PbExecCommand() : unknown window type");
            pbBusyFlag = FALSE;
            return (BEAR_ERROR);
        }

        if (pbScriptFlag) {
            bearStatus = (*cmdPtr->co_statusProc)(winDataPtr, screenPtr,
                argCount, argList);
        }
        else {
            if (cmdPtr->co_voidProc != (void (*)()) NULL) {
                (*cmdPtr->co_voidProc)(winDataPtr, screenPtr, argCount,
                    argList);
                bearStatus = BEAR_OK;
            }
            if (cmdPtr->co_statusProc != (BearStatus (*)()) NULL) {
                bearStatus = (*cmdPtr->co_statusProc)(winDataPtr, screenPtr,
                    argCount, argList);
            }
        }
        cmdPtr->co_usageCount++;

        pbBusyFlag = FALSE;
        return (bearStatus);
    }
    else
#endif BEAR_BLACK_BOX
    {
        /*
         * Black box mode.
         */

        if ((cmdPtr->co_propList & PB_BLACK_BOX_CP) == FALSE) {

            sprintf (bear_ErrorMsg,
                "%s: command may not be used in the black box mode\n",
                argList[0]);
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

            pbBusyFlag = FALSE;
            return (BEAR_ERROR);
        } 

        bearStatus = (*cmdPtr->co_statusProc)(NIL(WindowData), NIL(Point),
            argCount, argList);
        cmdPtr->co_usageCount++;

        pbBusyFlag = FALSE;
        return (bearStatus);
    }
    /* NOTREACHED */

#undef MAX_NUM_OF_ARGS
}



BearStatus
PbCreateChipFromBFPData(bfpFileName,
    placedFlag, checkFlag, coreMasterPtrPtr)

BearString  bfpFileName;
bool        placedFlag;
bool        checkFlag;               /* currently we check no matter what */
CellMaster  **coreMasterPtrPtr;
{
    FILE            *filePtr;
    BearString      ctreeFileName, flexFileName, techFileName, tmFileName;
    CellMaster      *coreMasterPtr;
    bool            pinSpacingFlag;

    if (Bfp_BearFPToMac(bfpFileName, NIL(Bfp_ReadOptions), &coreMasterPtr) !=
        BEAR_OK) {
        PbPrint(EZ_ITALIC_FONT,
            "Cannot create database for the given BFP file.\n");
        return (BEAR_ERROR);
    }
    Mac_PutPlacedFlagOfCell(placedFlag, coreMasterPtr);

    /*
     * Generate the name of the cluster tree file based on the name of the
     * BFP file, and if it exists store it for later use.
     */
    Mac_GenFileName(bfpFileName, ".bfp", ".ctree", &ctreeFileName);
    filePtr = fopen (ctreeFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        Mac_PutCTreeFileNameOfCellMaster(ctreeFileName, coreMasterPtr);
    }
    else {
        FREE((char *) ctreeFileName);
    }

    /*
     * Generate the name of the flex file based on the name of the
     * BFP file, and if it exists store it for later use.
     */
    Mac_GenFileName(bfpFileName, ".bfp", ".flex", &flexFileName);
    filePtr = fopen (flexFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        Mac_PutFlexFileNameOfCellMaster(flexFileName, coreMasterPtr);
    }
    else {
        FREE((char *) flexFileName);
    }

    /*
     * Generate the name of the timing file based on the name of the
     * BFP file, and if it exists store it for later use.
     */
    Mac_GenFileName(bfpFileName, ".bfp", ".timing", &tmFileName);
    filePtr = fopen (tmFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        Mac_PutTimingFileNameOfCellMaster(tmFileName, coreMasterPtr);
    }
    else {
        FREE((char *) tmFileName);
    }

    /*
     * Read in the technology part of the MAC database from
     * the file they specified.
     */
    techFileName = Mac_GetTechFileNameOfCellMaster(coreMasterPtr);
    if (techFileName != (BearString) NULL) {
        if (Mac_ReadTech(techFileName) != BEAR_OK) {
            return (BEAR_ERROR);
        }

        if (Mac_CheckPins(coreMasterPtr, &pinSpacingFlag) != BEAR_OK) {
            PbPrint(EZ_ITALIC_FONT, "The pin data is not legal.\n");
            return (BEAR_ERROR);
        }
        if (pinSpacingFlag == FALSE) {
            PbPrint(EZ_ITALIC_FONT, "Some pins are not spaced properly.\n");
            return (BEAR_ERROR);
        }
    }

    if (AddTilePlanes(coreMasterPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    *coreMasterPtrPtr = coreMasterPtr;
    return (BEAR_OK);
}



#ifdef USE_BBL_CODE
BearStatus
PbCreateChipFromBBLData(bblFileName, techFileName,
    placedFlag, checkModFlag, coreMasterPtrPtr)

BearString  bblFileName;
BearString  techFileName;
bool        placedFlag;
bool        checkModFlag;       /* check all of the bbl modules' syntax */
CellMaster  **coreMasterPtrPtr;
{
    FILE            *filePtr;
    BearStatus      bearStatus;
    BearString      string;
    BearString      ctreeFileName, flexFileName, rulesFileName, tmFileName;
    CellMaster      *coreMasterPtr;
    Pgr_WireTech    *wireTechPtr;
    bool            pinSpacingFlag;
    char            *pgFileNamePtr = NIL(char);
    char            *dotPtr, *tmpPtr;

    if (Bdi_BBLToMac(bblFileName, checkModFlag, &coreMasterPtr) ==
        BEAR_ERROR) {
        PbPrint(EZ_ITALIC_FONT,
            "Cannot create database for the given BBL file.\n");
        return (BEAR_ERROR);
    }
    Mac_PutPlacedFlagOfCell(placedFlag, coreMasterPtr);

    /*
     * Try to read in the cell shapes from a file whose name
     * is based on the name of the BBL file.
     */
    Mac_GenFileName(bblFileName, ".r", ".rules", &rulesFileName);
    filePtr = fopen (rulesFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        bearStatus = Ncl_ReadShapeData(coreMasterPtr, rulesFileName);
        if (bearStatus != BEAR_OK) {
            FREE((char *) rulesFileName);
            return (BEAR_ERROR);
        }
    }
    else {
        Ncl_ResetShapeLists(coreMasterPtr);
    }
    FREE((char *) rulesFileName);

    /*
     * Generate the name of the cluster tree file based on the name of the
     * BBL file, and if it exists store it for later use.
     */
    Mac_GenFileName(bblFileName, ".r", ".ctree", &ctreeFileName);
    filePtr = fopen (ctreeFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        Mac_PutCTreeFileNameOfCellMaster(ctreeFileName, coreMasterPtr);
    }
    else {
        FREE((char *) ctreeFileName);
    }

    /*
     * Generate the name of the flex file based on the name of the
     * BBL file, and if it exists store it for later use.
     */
    Mac_GenFileName(bblFileName, ".r", ".flex", &flexFileName);
    filePtr = fopen (flexFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        Mac_PutFlexFileNameOfCellMaster(flexFileName, coreMasterPtr);
    }
    else {
        FREE((char *) flexFileName);
    }

    /*
     * Generate the name of the timing file based on the name of the
     * BBL file, and if it exists store it for later use.
     */
    Mac_GenFileName(bblFileName, ".r", ".timing", &tmFileName);
    filePtr = fopen (tmFileName, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        Mac_PutTimingFileNameOfCellMaster(tmFileName, coreMasterPtr);
    }
    else {
        FREE((char *) tmFileName);
    }

    if (techFileName != (BearString) NULL) {
        /*
         * Read in the technology part of the MAC database from
         * the file they specified.
         */
        if (Mac_ReadTech(techFileName) != BEAR_OK) {
            return (BEAR_ERROR);
        }
        MALLOC(char *, string,
            (strlen (techFileName) + 1) * sizeof (char));
        strcpy (string, techFileName);
        Mac_PutTechFileNameOfCellMaster(string, coreMasterPtr);
    }
    else {
        filePtr = fopen (bblFileName, "r");
        if (filePtr != NIL(FILE)) {
            fclose (filePtr);

            /*
             * Try to read in the technology part of the MAC database from
             * a file whose name is based on the name of the BBL file.
             */
            Mac_GenFileName(bblFileName, ".r", ".tech", &string);
            if (Mac_ReadTech(string) != BEAR_OK) {
                FREE((char *) string);
                return (BEAR_ERROR);
            }
            Mac_PutTechFileNameOfCellMaster(string, coreMasterPtr);
        }
        else {
            /*
             * Fill in the technology part of the MAC database with
             * default vaules.
             */
            mac_NumOfLayer = 3;
            Mac_PutMinWidthOfLay(20, METAL1_LAYER);
            Mac_PutMinWidthOfLay(20, METAL2_LAYER);
            Mac_PutMinSpaceOfLayLay(10, METAL1_LAYER, METAL1_LAYER);
            Mac_PutMinSpaceOfLayLay(10, METAL2_LAYER, METAL2_LAYER);
            Mac_PutSizeOfHole(25, M1M2_VIA);
            Mac_PutMinOverlapOfLayHole(5, METAL1_LAYER, M1M2_VIA);
        }
    }

#ifdef USE_WIRE_CODE
    pgFileNamePtr = Mac_GetPgFileNameOfCellMaster(coreMasterPtr);
    if (pgFileNamePtr == NIL(char)) {
        /*
         * Generate the name for the wire technology file.
         *
         * fileName.r -> fileName.wt
         */
        tmpPtr = Mac_GetFileNameOfCellMaster(coreMasterPtr);
        if ((dotPtr = rindex(tmpPtr, '.')) != NIL(char)) {
            *dotPtr = '\0';
            MALLOC(char *, pgFileNamePtr,
                (strlen(tmpPtr) + 4) * sizeof (char));
            sprintf(pgFileNamePtr, "%s.wt", tmpPtr);
            *dotPtr = '.';
        }
        else {
            /* cell master file name does not have a dot in it */
            MALLOC(char *, pgFileNamePtr,
                (strlen(tmpPtr) + 4) * sizeof (char));
            sprintf(pgFileNamePtr, "%s.wt", tmpPtr);
        }
        Mac_PutPgFileNameOfCellMaster(pgFileNamePtr, coreMasterPtr);
    }

    filePtr = fopen (pgFileNamePtr, "r");
    if (filePtr != NIL(FILE)) {
        fclose (filePtr);
        MALLOC(Pgr_WireTech *, wireTechPtr, sizeof (Pgr_WireTech));
        if (Pgr_ReadWireTech(pgFileNamePtr, wireTechPtr) != BEAR_OK) {
            FREE((char *) wireTechPtr);
            return;
        }
        coreMasterPtr->contents.wireTechPtr = wireTechPtr;
    }
#endif USE_WIRE_CODE

    if (Mac_CheckPins(coreMasterPtr, &pinSpacingFlag) != BEAR_OK) {
        PbPrint(EZ_ITALIC_FONT, "Bad pin data.\n");
        return (BEAR_ERROR);
    }
    if (pinSpacingFlag == FALSE) {
        PbPrint(EZ_ITALIC_FONT, "Some pins are not spaced properly.\n");
        return (BEAR_ERROR);
    }

    if (AddTilePlanes(coreMasterPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    *coreMasterPtrPtr = coreMasterPtr;
    return (BEAR_OK);
}
#endif USE_BBL_CODE



#ifdef USE_OCT_CODE
BearStatus
PbCreateChipFromOCTData(fileNamePtr, cellViewPtr, techFileNamePtr,  
    placedFlag, checkFlag, coreMasterPtrPtr)

char        *fileNamePtr;
char        *cellViewPtr;
char        *techFileNamePtr;
bool        placedFlag;
bool        checkFlag;           /* check data for validity */
CellMaster  **coreMasterPtrPtr;
{
    BearStatus  bearStatus;
    struct stat statBuf;
    char        *stringPtr = NIL(char);
    bool        pinSpacingFlag;

    if (Odi_ReadFromOct(fileNamePtr, cellViewPtr, coreMasterPtrPtr) == 
        BEAR_ERROR) {

        return (BEAR_ERROR);
    }
    Mac_PutPlacedFlagOfCell(placedFlag, *coreMasterPtrPtr);
    
    Ncl_ResetShapeLists(*coreMasterPtrPtr);

    if (techFileNamePtr == NIL(char)) {
        Mac_GenFileName(fileNamePtr, ".r", ".tech", &stringPtr);
        if (stat(stringPtr, &statBuf)) {
            FREE((char *) stringPtr);
            Mac_GenFileName("oct", ".r", ".tech", &stringPtr);
            if (stat(stringPtr, &statBuf)) {
                Bear_PrintErr("The proper technology file does not exist.");

                FREE((char *) stringPtr);
                return (BEAR_ERROR);
            }
        }
        techFileNamePtr = stringPtr;
    }

    if (Mac_ReadTech(techFileNamePtr) != BEAR_OK) {
        if (stringPtr != NIL(char)) {
            FREE((char *) stringPtr);
        }
        return (BEAR_ERROR);
    }
    if (stringPtr == NIL(char)) {
        MALLOC(char *, stringPtr,
            (strlen (techFileNamePtr) + 1) * sizeof (char));
        strcpy (stringPtr, techFileNamePtr);
    }
    Mac_PutTechFileNameOfCellMaster(stringPtr, *coreMasterPtrPtr);

    if (Mac_CheckPins(*coreMasterPtrPtr, &pinSpacingFlag) != BEAR_OK) {
        PbPrint(EZ_ITALIC_FONT, "Bad pin data.\n");
        return (BEAR_ERROR);
    }
    if (pinSpacingFlag == FALSE) {
        PbPrint(EZ_ITALIC_FONT, "Some pins are not spaced properly.\n");
        return (BEAR_ERROR);
    }

    if (AddTilePlanes(*coreMasterPtrPtr) != BEAR_OK) {
        return (BEAR_ERROR);
    }

    return (BEAR_OK);
}
#endif /* USE_OCT_CODE */



void
PbMailUser(mailDataPtr)
MailData *mailDataPtr;
{
/*
 * mail -s "subject" user < fileName
 */
#define SPRINTF_FORMAT "mail -s \"%s\" %s < %s"

    char *subjectPtr, *commandPtr;
    int  len;

    if (mailDataPtr->subjectPtr == NIL(char)) {
        subjectPtr = "A letter from Bear";
    }
    else {
        subjectPtr = mailDataPtr->subjectPtr;
    }

    len = strlen (SPRINTF_FORMAT) +
        strlen (subjectPtr) +
        strlen (mailDataPtr->userPtr) +
        strlen (mailDataPtr->fileNamePtr) + 1;
    MALLOC(char *, commandPtr, len * sizeof (char));

    sprintf (commandPtr, SPRINTF_FORMAT,
        subjectPtr, mailDataPtr->userPtr, mailDataPtr->fileNamePtr);
    system (commandPtr);

    FREE((char *) commandPtr);

    if (mailDataPtr->userPtr != NIL(char)) {
        FREE((char *) mailDataPtr->userPtr);
        mailDataPtr->userPtr = NIL(char);
    }
    if (mailDataPtr->subjectPtr != NIL(char)) {
        FREE((char *) mailDataPtr->subjectPtr);
        mailDataPtr->subjectPtr = NIL(char);
    }
    if (mailDataPtr->fileNamePtr != NIL(char)) {
        FREE((char *) mailDataPtr->fileNamePtr);
        mailDataPtr->fileNamePtr = NIL(char);
    }

#undef SPRINTF_FORMAT
}



BearStatus
PbBeginChip()
{
#ifdef USE_OLDP_CODE
    Pl_SetNoPlacement();
#endif USE_OLDP_CODE

    if (Mac_Begin() != BEAR_OK) {
        return (BEAR_ERROR);
    }

#ifdef USE_DR_CODE
    if (Rdb_Begin() != BEAR_OK) {
        return (BEAR_ERROR);
    }

    lr_singleTermFlag = FALSE;
#endif /* USE_DR_CODE */

    return (BEAR_OK);
}



BearStatus
PbEndChip(cellMasterPtrPtr)
CellMaster **cellMasterPtrPtr;
{
    PlaneData *planeDataPtr;

    planeDataPtr = Mac_GetCellPlaneOfCell(*cellMasterPtrPtr);
    Tp_FreePlaneData(planeDataPtr);
    /*
     * Should also free:
     *   Mac database (Well it's a thought ..)
     */

    *cellMasterPtrPtr = NIL(CellMaster);

    pbCellMasterPtr = NIL(CellMaster);

    return (BEAR_OK);
}


void
PbPrint(whichFont, stringPtr)                            
int     whichFont;  /* EZ_BOLD_FONT | EZ_ITALIC_FONT | EZ_NORMAL_FONT */
char    *stringPtr;
{                                           
#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        Ez_Print(whichFont, stringPtr);
    } 
    else 
#endif /* BEAR_BLACK_BOX */
    {
        if (pbLogFilePtr != NIL(FILE)) {
            if (whichFont == EZ_ITALIC_FONT) {
                (void) fprintf (pbLogFilePtr, "Error : %s", stringPtr);
            }
            else {
                (void) fprintf (pbLogFilePtr, "%s", stringPtr);
            }
            fflush (pbLogFilePtr);
        }

        (void) fprintf (bearout, stringPtr);
    }
}


void
PbPrompt(promptPtr)
char *promptPtr;
{
    PbPrint(EZ_NORMAL_FONT, promptPtr);
}


