
/* pbMiscDraw.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbMiscDraw.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#if !defined(BEAR_BLACK_BOX)
/*
 * This file does not need to be compiled when the black box version of
 * the BEAR-FP system is built.
 */

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "stack.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "grEditor.h"
#include "deque.h"
#include "topChDecomposer.h"
#include "polarBear.h"


/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawBnLineOne --
 *
 * This procedure is designed to be used by an "interator".
 * It draws the given bottleneck line in the given chip window.
 *
 * The following fields from the structure DrawData must be filled in:
 *   winDataPtr
 *   color
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus 
PbDrawBnLineOne(bnLinePtr, drawDataPtr)
BNLine      *bnLinePtr;
DrawData    *drawDataPtr;
/*
 * Draw bottleneck line.
 */
{
    Line *linePtr;

    /*
     * If this is not a chip window return BEAR_CONTINUE_SEARCH.
     */
    if (PbIsChipWin(drawDataPtr->winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    linePtr = Tp_GetLineOfBnLine(bnLinePtr);
    Ez_DrawWorldLine(drawDataPtr->winDataPtr->wd_userWin,
        drawDataPtr->winDataPtr->wd_scrMatrix,
        linePtr->point1.p_x, linePtr->point1.p_y,
        linePtr->point2.p_x, linePtr->point2.p_y,
        EZ_SOLIDLINE2, GXcopy, drawDataPtr->color, pbChipColor);

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawBnLineTwo --
 *
 * This procedure is designed to be used by an "interator".
 * It draws the given bottleneck tile in the chip window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus 
PbDrawBnLineTwo(winDataPtr, drawBnLinePtr)
WindowData  *winDataPtr;
DrawBnLine  *drawBnLinePtr;
/*
 * Draw bottleneck line.
 */
{
    DrawData drawData;

    /*
     * If this is not a chip window return BEAR_CONTINUE_SEARCH.
     */
    if (PbIsChipWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    drawData.winDataPtr = winDataPtr;
    drawData.color = drawBnLinePtr->bnLineColor;
    (void) PbDrawBnLineOne(drawBnLinePtr->bnLinePtr, &drawData);

    return (BEAR_CONTINUE_SEARCH);
}



void
PbDrawFloorPlanGraph(winDataPtr)
WindowData *winDataPtr;
{
    CellMaster      *cellMasterPtr;
    ChipWin         *chipWinPtr;
    DrawData        drawData;
    FloorPlanGraph  *fpgPtr;
    PlaneData       *planeDataPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    fpgPtr = Tp_GetFpg(planeDataPtr);
           
    if (fpgPtr != NIL(FloorPlanGraph)) {
        drawData.winDataPtr = winDataPtr;
        drawData.color = pbCellBorderColor;
        drawData.width = MAXWIDTH;
        Fpg_ItrWallSegSeq(fpgPtr, Gre_DrawWallSeg, &drawData);
        Ez_Flush();
    }
}



void
PbDrawRoom(winDataPtr, roomPtr)
WindowData  *winDataPtr;
Fpg_Room    *roomPtr;
{
    lsGen           gen;
    DrawData        drawData;
    WallSegment     *segPtr;
    int             i;

    if (roomPtr != NIL(Fpg_Room)) {
        drawData.winDataPtr = winDataPtr;
        drawData.color = ez_HighLightColor;
        drawData.width = MAXWIDTH;
        for (i = 0; i < 4 ;i++) {
            lsForeachItem(roomPtr->wallList[i], gen, segPtr) {
                Gre_DrawWallSeg(segPtr, &drawData);
            }
        }
        Ez_Flush();
    }
}



void
PbHighlightChannel(cellMasterPtr, winDataPtr, channelPtr)
CellMaster *cellMasterPtr;
WindowData *winDataPtr;
Channel *channelPtr;
{
#define TLENDJUNCTPTR   channelPtr->channelUn.straightChannel.tlEndJunct
#define BRENDJUNCTPTR   channelPtr->channelUn.straightChannel.brEndJunct
#define CORNERJUNCTPTR  channelPtr->channelUn.lShapedChannel.cornerJunct
#define VENDJUNCTPTR    channelPtr->channelUn.lShapedChannel.vEndJunct
#define HENDJUNCTPTR    channelPtr->channelUn.lShapedChannel.hEndJunct

    DrawNode     drawNode;
    DrawData     *drawDataPtr;
    DrawData     drawData;
    PinOrJunct   pinOrJunct;
    PlaneData    *planeDataPtr;
    WallJunction *junct1Ptr;
    WallJunction *junct2Ptr;

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    if (Fpg_IsFpgBuilt(planeDataPtr) == FALSE)
    {
        BuildFloorPlan(cellMasterPtr, FALSE);
        (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
    }
    drawData.winDataPtr = winDataPtr;
    drawData.color = ez_HighLightColor;
    drawData.width = MAXWIDTH;
    drawDataPtr = &drawData;

    pinOrJunct.unType = FPG_WALL_JUNCT;
    drawNode.pinOrJunctPtr = &pinOrJunct;
    drawNode.pinColor = drawNode.junctColor = pbCellBorderColor;

    if (channelPtr->type == VERT_CHANNEL)
    {
        drawDataPtr->color = pbVChannelColor;
    junct1Ptr = TLENDJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetSouthJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != BRENDJUNCTPTR);

    junct1Ptr = TLENDJUNCTPTR;
    junct2Ptr = BRENDJUNCTPTR;
    }
    else if (channelPtr->type == HORZ_CHANNEL)
    {
    drawDataPtr->color = pbHChannelColor;
    junct1Ptr = TLENDJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetEastJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != BRENDJUNCTPTR);

    junct1Ptr = TLENDJUNCTPTR;
    junct2Ptr = BRENDJUNCTPTR;
    }
    else if (channelPtr->type == TOP_LEFT_L_CHANNEL)
    {
        drawDataPtr->color = pbLChannelColor;
    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetNorthJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != VENDJUNCTPTR);

    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetWestJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != HENDJUNCTPTR);

    junct1Ptr = VENDJUNCTPTR;
    junct2Ptr = HENDJUNCTPTR;
    }
    else if (channelPtr->type == TOP_RIGHT_L_CHANNEL)
    {
    drawDataPtr->color = pbLChannelColor;
    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetNorthJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != VENDJUNCTPTR);

    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetEastJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != HENDJUNCTPTR);

    junct1Ptr = VENDJUNCTPTR;
    junct2Ptr = HENDJUNCTPTR;
    }
    else if (channelPtr->type == BOTTOM_RIGHT_L_CHANNEL)
    {   
    drawDataPtr->color = pbLChannelColor;
    junct1Ptr = CORNERJUNCTPTR;
    do
    {
       junct2Ptr  = Fpg_GetSouthJunctOfWallJunct(junct1Ptr);
       Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
       junct1Ptr = junct2Ptr;
    } while (junct1Ptr != VENDJUNCTPTR);

    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetEastJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != HENDJUNCTPTR);

    junct1Ptr = VENDJUNCTPTR;
    junct2Ptr = HENDJUNCTPTR;
    }
    else if (channelPtr->type == BOTTOM_LEFT_L_CHANNEL)
    {
    drawDataPtr->color = pbLChannelColor;
    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetSouthJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != VENDJUNCTPTR);

    junct1Ptr = CORNERJUNCTPTR;
    do
    {
        junct2Ptr  = Fpg_GetWestJunctOfWallJunct(junct1Ptr);
        Gre_DrawFromJunctToJunct(drawDataPtr, junct1Ptr, junct2Ptr);
        junct1Ptr = junct2Ptr;
    } while (junct1Ptr != HENDJUNCTPTR);

    junct1Ptr = VENDJUNCTPTR;
    junct2Ptr = HENDJUNCTPTR;
     }
     pinOrJunct.unValue.junctPtr = junct1Ptr;
     (void) Gre_DrawNode(winDataPtr, &drawNode);
     pinOrJunct.unValue.junctPtr = junct2Ptr;
     (void) Gre_DrawNode(winDataPtr, &drawNode);
     Ez_Flush();

#undef TLENDJUNCTPTR
#undef BRENDJUNCTPTR
#undef CORNERJUNCTPTR
#undef VENDJUNCTPTR
#undef HENDJUNCTPTR
}



#ifdef USE_DR_CODE
void
PbHighlightAllChannels(winDataPtr)
WindowData *winDataPtr;
{
    Deque           *channelQue;
    CellMaster      *cellMasterPtr;
    Channel         *channelPtr;
    ChipWin         *chipWinPtr;
    FloorPlanGraph  *fpgPtr;
    PlaneData       *planeDataPtr;
    Stack           *channelStackPtr;
    
    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    if (Fpg_IsFpgBuilt(planeDataPtr) == FALSE)
    {
        BuildFloorPlan(cellMasterPtr, FALSE);
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_FPG_OPT);
        (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
    }
    else if ((chipWinPtr->chw_options & PB_SHOW_FPG_OPT) == FALSE) {
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_FPG_OPT);
        (void) PbDrawChipRegion(winDataPtr, NIL(Rect));
    }

    fpgPtr = Tp_GetFpg(planeDataPtr);
    channelStackPtr = StackNew(1);
    (void) Tcd_RRDO(fpgPtr, channelStackPtr);

    channelQue = MakeDeque();

    Tcd_FindIndepChannel(cellMasterPtr, channelStackPtr, channelQue);

    while (IsEmptyDeque(channelQue) == FALSE)
    {
        channelPtr = (Channel *) PopDeque(channelQue);
        PbHighlightChannel(cellMasterPtr, winDataPtr, channelPtr);
    }
    DeleteDeque(channelQue);
    while (StackEmpty(channelStackPtr) == FALSE)
    {
        channelPtr = (Channel *) StackPop(channelStackPtr);
        FREE((char *) channelPtr);
    }
    StackFree(channelStackPtr);
}
#endif /* USE_DR_CODE */

#endif /* BEAR_BLACK_BOX */


