
/* pbNewPlaceCmds.c --
 *
 *    Benjamin Chen
 *    George Carvalho, Esq.
 *    (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbNewPlaceCmds.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#ifdef USE_NEWP_CODE

#include "bearCopyRight.h"
#include <stdio.h>
#include <math.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "globalSpacer.h"
#include "tileMapping.h"
#include "newC.h"
#include "newP.h"
#include "ezPlot.h"
#include "polarBear.h"
#ifndef BEAR_BLACK_BOX
#include "bearIV.h"
#include "pbDialogs.h"
#endif BEAR_BLACK_BOX


#define STDSTEP 0.2
#define MAXSTEP 100


extern WindowData *globalWinDataPtr;


/* ---------- Local Procedures ---------- */
static BearStatus   GetNewClusterParms();
static BearStatus   GetNewPlacerParms();


static BearStatus
GetNewClusterParms(argCount, argList, parmsPtr, dialogWinFlagPtr)
int             argCount;
char            **argList;
Ncl_UserParms   *parmsPtr;
bool            *dialogWinFlagPtr;
{
    BearString  controlFileName = (BearString) NULL;
    int         i;
    bool        haveAlgFlag = FALSE;
    
    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        /* 
         * Print out this commands help message.
         */
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        }

        /* 
         * Don't pop up any dialog windows.  Used when you want to run
         * this command from a script.
         */
        if (strcmp (argList[i], "-NDW") == 0) {
            *dialogWinFlagPtr = FALSE;
            continue;
        }

        if (strcmp (argList[i], "-alg") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            switch (*argList[i]) {
            case 'R':
            case 'r': 
                parmsPtr->algorithm = RANDOM;
                break;

            case 'G':
            case 'g':
                parmsPtr->algorithm = GREEDY;
                break;

            case 'M':
            case 'm': 
                parmsPtr->algorithm = MATCHING;
                break;

            case 'A':
            case 'a': 
                parmsPtr->algorithm = SIM_ANNEAL;
                break;

            case 'F':
            case 'f':
                parmsPtr->algorithm = FILE_IN;

                if (++i >= argCount) {
                    goto usage;
                }
                MALLOC(char *, parmsPtr->clusterFileName,
                    (strlen (argList[i]) + 1) * sizeof (char));
                strcpy (parmsPtr->clusterFileName, argList[i]);
                break;

            default:
                goto usage;
            }
            haveAlgFlag = TRUE;
            continue;
        }

        /* Only used with matching and annealing cluster algorithums. */
        if (strcmp (argList[i], "-mdr") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            MALLOC(int *, parmsPtr->dimRatioPtr, sizeof (int));
            *(parmsPtr->dimRatioPtr) = atoi (argList[i]);
            continue;
        }

        /* Only used with matching and annealing cluster algorithums. */
        if (strcmp (argList[i], "-mar") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            MALLOC(int *, parmsPtr->areaRatioPtr, sizeof (int));
            *(parmsPtr->areaRatioPtr) = atoi (argList[i]);
            continue;
        }

        /* Storage file name. */
        if (strcmp (argList[i], "-sf") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            parmsPtr->storeFlag = TRUE;
            MALLOC(char *, parmsPtr->storageFileName,
                (strlen (argList[i]) + 1) * sizeof (char));
            strcpy (parmsPtr->storageFileName, argList[i]);
            continue;
        }

        /* Timing file name. */
        if (strcmp (argList[i], "-tf") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            parmsPtr->timingFlag = TRUE;
            MALLOC(char *, parmsPtr->timingFileName,
                (strlen (argList[i]) + 1) * sizeof (char));
            strcpy (parmsPtr->timingFileName, argList[i]);
            continue;
        }

        /* Control file name. */
        if (strcmp (argList[i], "-cf") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            MALLOC(char *, controlFileName,
                (strlen (argList[i]) + 1) * sizeof (char));
            strcpy (controlFileName, argList[i]);
            continue;
        }

        (void) sprintf (bear_Msg, "unknown option %s\n", argList[i]);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        goto usage;
    }
    Ncl_EchoUserParms("Cluster parameters values after command line.", 
        parmsPtr);

    if (controlFileName == NIL(char)) {
        FILE *filePtr;

        filePtr = fopen (NPL_CONTROL_DATA, "r");
        if (filePtr != NIL(FILE)) {
            (void) fclose (filePtr);
            MALLOC(char *, controlFileName,
                (strlen (NPL_CONTROL_DATA) + 1) * sizeof (char));
            strcpy (controlFileName, NPL_CONTROL_DATA);
        }
    }
    if (controlFileName != NIL(char)) {
        if (Ncl_ReadControlData(controlFileName, parmsPtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }
        FREE((char *) controlFileName);
    }

    if (haveAlgFlag == FALSE) {
#ifndef BEAR_BLACK_BOX
        if ((bear_BlackBoxFlag == FALSE) && (*dialogWinFlagPtr == TRUE)) {
            if (PbNewClusterAlgDialog(parmsPtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
        else
#endif BEAR_BLACK_BOX
        {
            PbPrint(EZ_ITALIC_FONT, "Must specify an algorithum.\n");
            return (BEAR_ERROR);
        }
    }

    return (BEAR_OK);

usage:
    PbPrint(EZ_NORMAL_FONT, "new cluster (ok in script) (ok in black box)\n");
    PbPrint(EZ_NORMAL_FONT,
        "Usage: ncl [-alg < r || g || m || a || f filename >] [-NDW]\n");
    PbPrint(EZ_NORMAL_FONT,
        "       [-mar <integer>] [-mdr <integer>] [-sf fileName]\n");
    PbPrint(EZ_NORMAL_FONT,
        "       [-tf fileName] [-cf fileName]\n");

    PbPrint(EZ_NORMAL_FONT, "\noptions:\n");
    PbPrint(EZ_NORMAL_FONT, "    -alg algorithms:\n");
    PbPrint(EZ_NORMAL_FONT, "        r)andom\n");
    PbPrint(EZ_NORMAL_FONT, "        g)reedy\n");
    PbPrint(EZ_NORMAL_FONT, "        m)atching\n");
    PbPrint(EZ_NORMAL_FONT, "        a)nnealing\n");
    PbPrint(EZ_NORMAL_FONT, "        f)ile-in\n");

    PbPrint(EZ_NORMAL_FONT,
  "    -mar #             max area ratio (matching and annealing only)\n");
    PbPrint(EZ_NORMAL_FONT,
  "    -mdr #             max dimension ratio (matching and annealing only)\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -sf filename       storage data file \n");
    PbPrint(EZ_NORMAL_FONT,
        "    -tf filename       timing data file \n");
    PbPrint(EZ_NORMAL_FONT,
        "    -cf filename       control data file \n");

    PbPrint(EZ_NORMAL_FONT,
        "    -NDW               no dialog windows\n");
    PbPrint(EZ_NORMAL_FONT, "\n");

    return (BEAR_ERROR);
}



static BearStatus
GetNewPlacerParms(argCount, argList, modeValuePtr, placeParmsPtr)
int             argCount;
char            **argList;
double          *modeValuePtr;
Npl_UserParms   *placeParmsPtr;
{
#define ANS_SIZE 15

    static double   arValue, dimValue;

    BearString      controlFileName = (BearString) NULL;
    int             i, value;
    char            message[25], answer[ANS_SIZE];
    bool            dialogWinFlag = TRUE;
    bool            haveModeFlag = FALSE;

    *modeValuePtr = 0.0;

    /* Round the dimension value off to the nearest integer. */
    dimValue = (int) (sqrt((double) ncl_ClusterTree->nd_area) * 1.05);
    arValue = 1.0;

    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        /* 
         * Print out this commands help message.
         */
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        }

        /* 
         * Don't pop up any dialog windows.  Used when you want to run
         * this command from a script.
         */
        if (strcmp (argList[i], "-NDW") == 0) {
            dialogWinFlag = FALSE;
            continue;
        }

        if (strcmp (argList[i], "-m") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            switch (*argList[i]) {
            case 'X':
            case 'x':
                placeParmsPtr->p_mode = NPL_FIXED_X;
                break;
            case 'Y':
            case 'y':
                placeParmsPtr->p_mode = NPL_FIXED_Y;
                break;
            case 'R':
            case 'r':
                placeParmsPtr->p_mode = NPL_ASPECT_RATIO;
                break;
            default:
                goto usage;
            }
            if (++i >= argCount) {
                goto usage;
            }
            *modeValuePtr = atof (argList[i]);
            haveModeFlag = TRUE;
            continue;
        }

        if (strcmp (argList[i], "-cf") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            MALLOC(char *, controlFileName,
                (strlen (argList[i]) + 1) * sizeof (char));
            strcpy (controlFileName, argList[i]);
            continue;
        }

        (void) sprintf (bear_Msg, "unknown option %s\n", argList[i]);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        goto usage;
    }

    if (controlFileName == NIL(char)) {
        FILE *filePtr;

        filePtr = fopen (NPL_CONTROL_DATA, "r");
        if (filePtr != NIL(FILE)) {
            (void) fclose (filePtr);
            MALLOC(char *, controlFileName,
                (strlen (NPL_CONTROL_DATA) + 1) * sizeof (char));
            strcpy (controlFileName, NPL_CONTROL_DATA);
        }
    }
    if (controlFileName != NIL(char)) {
        if (Npl_ReadControlData(controlFileName, placeParmsPtr) != BEAR_OK) {
            return (BEAR_ERROR);
        }
        FREE((char *) controlFileName);
    }

    if (haveModeFlag == FALSE) {
#ifndef BEAR_BLACK_BOX
        if ((bear_BlackBoxFlag == FALSE) && (dialogWinFlag == TRUE)) {
            if (PbNewPlacerVarDialog(modeValuePtr, dimValue, arValue,
                placeParmsPtr) != BEAR_OK) {
                return (BEAR_ERROR);
            }
        }
        else
#endif BEAR_BLACK_BOX
        {
            PbPrint(EZ_ITALIC_FONT, "Must specify a mode.\n");
            return (BEAR_ERROR);
        }
    }

    return (BEAR_OK);

usage:
    PbPrint(EZ_NORMAL_FONT, "new placer (ok in script) (ok in black box)\n");
    PbPrint(EZ_NORMAL_FONT,
        "Usage: npl [-m < x || y || r > <float>] [-cf fileName] [-NDW]\n");

    PbPrint(EZ_NORMAL_FONT, "\noptions:\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -m                 indicate mode and value\n");
    PbPrint(EZ_NORMAL_FONT,
        "        The allowed modes are x, y, and r followed by a\n");
    PbPrint(EZ_NORMAL_FONT,
        "    fixed x dimension, a fixed y dimension or an X to Y\n");
    PbPrint(EZ_NORMAL_FONT,
        "    aspect ratio.\n\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -cf filename       control data file \n");
    PbPrint(EZ_NORMAL_FONT,
        "    -NDW               no dialog windows\n");
    PbPrint(EZ_NORMAL_FONT, "\n");

    return BEAR_ERROR;

#undef ANS_SIZE
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbNewClusterCmd() --
 *  
 * The Ncl_UserParms structure is filled in four steps as follows:
 *     1.) Initial values are assigned by the program.
 *     2.) Values are assigned from the command line.
 *     3.) Values are read out of the control data file.
 *     4.) Values are selected using the iv window interface.
 * 
 * Results:
 *     BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
BearStatus
PbNewClusterCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr; /* not used */
int         argCount;
char        **argList;
{
/* Name of the default storage file */
#define DEF_STORAGE_FILE    "cluster.tmp"

#ifndef BEAR_BLACK_BOX
    ChipWin         *chipWinPtr;
#endif BEAR_BLACK_BOX

    BearString      ctreeFileName, tmFileName;
    CellMaster      *cellMasterPtr;
    Ncl_UserParms   *parmsPtr;
    bool            dialogWinFlag = TRUE;

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        chipWinPtr = PbGetChipWin(winDataPtr);
        cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    }
    else
#endif BEAR_BLACK_BOX
    {
        if (pbCellMasterPtr == NIL(CellMaster)) {
            PbPrint(EZ_ITALIC_FONT, "Must read in a cell master first.\n");
            return (BEAR_ERROR);
        }
        else {
            cellMasterPtr = pbCellMasterPtr;
        }
    }
    Mac_UnsetClusteringDone(cellMasterPtr);

    parmsPtr = Ncl_MakeUserParms();
    if ((ctreeFileName = Mac_GetCTreeFileNameOfCellMaster(cellMasterPtr)) !=
        (BearString) NULL) {
        parmsPtr->clusterFileName = ctreeFileName;
    }
    if ((tmFileName = Mac_GetTimingFileNameOfCellMaster(cellMasterPtr)) !=
        (BearString) NULL) {
        parmsPtr->timingFileName = tmFileName;
    }
    MALLOC(char *, parmsPtr->storageFileName,
        (strlen (DEF_STORAGE_FILE) + 1) * sizeof (char));    
    strcpy (parmsPtr->storageFileName, DEF_STORAGE_FILE);

    if (GetNewClusterParms(argCount, argList, parmsPtr, &dialogWinFlag) !=
        BEAR_OK) {

        Ncl_FreeUserParms(&parmsPtr);
        return (BEAR_ERROR);
    }

    if ((bear_BlackBoxFlag == TRUE) || (dialogWinFlag == FALSE)) {
        if (Ncl_BearInterface(cellMasterPtr, parmsPtr, 
            (BearGeneric) NULL) != BEAR_OK) {
            return (BEAR_ERROR);
        }
    }
#ifndef BEAR_BLACK_BOX
    else {
        if (PbNewClusterVarDialog(cellMasterPtr, parmsPtr) != BEAR_OK) {
            PbPrint(EZ_ITALIC_FONT, "Clustering aborted (error).\n");
            Ncl_FreeUserParms(&parmsPtr);
            return (BEAR_ERROR);
        }
    }
#endif BEAR_BLACK_BOX

    Ncl_FreeUserParms(&parmsPtr); Mac_SetClusteringDone(cellMasterPtr);
    PbPrint(EZ_NORMAL_FONT, "Clustering is done.\n");

    return (BEAR_OK);
#undef DEF_STORAGE_FILE
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbNewPlacerCmd() --
 *  
 * The Npl_UserParms structure is filled in four steps as follows:
 *     1.) Initial values are assigned by the program.
 *     2.) Values are assigned from the command line.
 *     3.) Values are read out of the control data file.
 *     4.) Values are selected using the iv window interface.
 * 
 * Results:
 *     BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
BearStatus
PbNewPlacerCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
/*
 * The Npl_UserParms structure is filled in four steps as follows:
 *     1.) Initial values are assigned by the program.
 *     2.) Values are assigned from the command line.
 *     3.) Values are read out of the control data file.
 *     4.) Values are selected using the iv window interface.
 */
{
#ifndef BEAR_BLACK_BOX
    ChipWin         *chipWinPtr;
#endif BEAR_BLACK_BOX

    CellMaster      *cellMasterPtr;
    Npl_UserParms   *placeParmsPtr;
    double          aspectR;

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        globalWinDataPtr = winDataPtr;
        chipWinPtr = PbGetChipWin(winDataPtr);
        cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    }
    else
#endif BEAR_BLACK_BOX
    {
        if (pbCellMasterPtr == NIL(CellMaster)) {
            PbPrint(EZ_ITALIC_FONT, "Must read in a cell master first.\n");
            return (BEAR_ERROR);
        }
        else {
            cellMasterPtr = pbCellMasterPtr;
        }
    }
    Mac_UnsetPlacementDone(cellMasterPtr);

    if (Mac_IsClusteringDone(cellMasterPtr) == FALSE) {
        PbPrint(EZ_ITALIC_FONT, "Cannot do placement before clustering.\n");
        return (BEAR_ERROR);
    }

    placeParmsPtr = Npl_MakeUserParms();
    if (GetNewPlacerParms(argCount, argList, &aspectR, placeParmsPtr) ==
        BEAR_ERROR) {

        return (BEAR_ERROR);
    }

    if (Npl_Placement(aspectR, cellMasterPtr, placeParmsPtr) != BEAR_OK) {
        Npl_FreeUserParms(&placeParmsPtr);
        PbPrint(EZ_NORMAL_FONT, "Could not place cells.\n");
        return (BEAR_ERROR);
    }
    Npl_FreeUserParms(&placeParmsPtr);
    Mac_SetPlacementDone(cellMasterPtr);
    PbPrint(EZ_NORMAL_FONT, "Placement is done.\n");

    return (BEAR_OK);
}



#endif USE_NEWP_CODE


