
/* pbPickRidge.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbPickRidge.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#if !defined(BEAR_BLACK_BOX)
/*
 * This file does not need to be compiled when the black box version of
 * the BEAR-FP system is built.
 */

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "deque.h"
#include "polarBear.h"
#ifdef USE_XDM_CODE
#include "vemDM.h"
#endif USE_XDM_CODE

typedef struct _DrawPartOfRidge {
    int     displayFunc;    /* display function                     */
    Pixel   ridgeColor;     /* ridge color                          */
    Tile    *prevTilePtr;   /* previous tile                        */
    Tile    *curTilePtr;    /* current tile                         */
} DrawPartOfRidge;


/* ---------- Local Procedures ---------- */
BearStatus DrawPartOfHorzRidge(), DrawPartOfVertRidge();
int FindSpan();
bool UserWantsToQuit();

/* ---------- Local Global Variables ---------- */
static Rect     *planeBndPtr, searchArea;
static Tile     *topPseudoTilePtr, *botPseudoTilePtr;
static Tile     *rightPseudoTilePtr, *leftPseudoTilePtr;
static int      spanEnd1, spanEnd2;
static int      spaceTileEnd1, spaceTileEnd2;

/*
 * Macros for this file.
 *
 * These macros make it posible to handle rectangles without constantly
 * having to clip them against the planes boundary, which of course saves
 * time !!
 */

#define TOP_SIDE(tp) ((RT(tp) == topPseudoTilePtr) ? \
    planeBndPtr->r_ytop : TOP(tp))

#define BOTTOM_SIDE(tp) ((LB(tp) == botPseudoTilePtr) ? \
    planeBndPtr->r_ybot : BOTTOM(tp))

#define LEFT_SIDE(tp) ((BL(tp) == leftPseudoTilePtr) ? \
    planeBndPtr->r_xbot : LEFT(tp))

#define RIGHT_SIDE(tp) ((TR(tp) == rightPseudoTilePtr) ? \
    planeBndPtr->r_xtop : RIGHT(tp))


static bool
UserWantsToQuit()
/*
 * Double check if the user really wants to quit the pick ridge command.
 */
{
#ifdef USE_XDM_CODE
    sprintf (bear_Msg, "%s%s%s",
        "Picking a solid tile\n",
        "terminates this command.\n",
        "Do you really want to quit?\n");
    return (dmConfirm("Warning", bear_Msg, "yes", "no") == VEM_OK ? TRUE : FALSE);
#else
    return (TRUE);
#endif USE_XDM_CODE
}



/*
 * ----------------------------------------------------------------------------
 *
 * DrawPartOfHorzRidge() --
 *
 * This procedure is designed to be used by an "interator".
 * It draws the part (one tiles worth) of a horizontal ridge in the
 * chip window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

static BearStatus
DrawPartOfHorzRidge(winDataPtr, drawPartOfRidgePtr)
WindowData      *winDataPtr;
DrawPartOfRidge *drawPartOfRidgePtr;
{
    Rect    rect;
    Window  chipWin;
    double  (*scrMatrix)[4];
    int     wp1x, wp1y, wp2x, wp2y, wp3x, wp3y; /* world coordinates */

    /*
     * If this is not a chip window return BEAR_CONTINUE_SEARCH.
     */
    if (PbIsChipWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    chipWin = winDataPtr->wd_userWin;
    scrMatrix = winDataPtr->wd_scrMatrix;

    wp3x = wp2x = RIGHT_SIDE(drawPartOfRidgePtr->curTilePtr);
    if (drawPartOfRidgePtr->prevTilePtr == NIL(Tile)) {
        wp2y = planeBndPtr->r_ytop;
    } else {
        wp1y = wp2y = TOP_SIDE(drawPartOfRidgePtr->curTilePtr);
        wp1x = RIGHT_SIDE(drawPartOfRidgePtr->prevTilePtr);

        /*
         * Draw a line from the bottom right corner of the previous
         * tile to the top right corner of the current tile.
         */
        if (wp1x != wp2x) {
            Ez_DrawWorldLine(chipWin, scrMatrix, wp1x, wp1y,
                wp2x, wp2y, EZ_SOLIDLINE1,
                drawPartOfRidgePtr->displayFunc,
                drawPartOfRidgePtr->ridgeColor,
                pbChipColor);
        }
    }
    /*
     * Draw a line from the top right corner of the current
     * tile to the bottom right corner of the current tile.
     */
    Ez_DrawWorldLine(chipWin, scrMatrix, wp2x, wp2y,
        wp3x, wp3y = BOTTOM_SIDE(drawPartOfRidgePtr->curTilePtr),
        EZ_SOLIDLINE1, 
        drawPartOfRidgePtr->displayFunc,
        drawPartOfRidgePtr->ridgeColor,
        pbChipColor);

    if ((rect.r_xbot = wp2x) != (rect.r_xtop = planeBndPtr->r_xtop)) {
        rect.r_ytop = wp2y;
        rect.r_ybot = wp3y;
        Ez_StippleRect(chipWin, scrMatrix, &rect,
            drawPartOfRidgePtr->ridgeColor,
            ez_CargoNetBitmap, drawPartOfRidgePtr->displayFunc);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * DrawPartOfVertRidge() --
 *
 * This procedure is designed to be used by an "interator".
 * It draws the part (one tiles worth) of a vertical ridge in the
 * chip window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

static BearStatus
DrawPartOfVertRidge(winDataPtr, drawPartOfRidgePtr)
WindowData      *winDataPtr;
DrawPartOfRidge *drawPartOfRidgePtr;
{
    Point   wp1, wp2, wp3, wtop, tmp; /* world coordinates */
    Window  chipWin;
    Rect    rect;
    double  (*scrMatrix)[4];

    /*
     * If this is not a chip window return BEAR_CONTINUE_SEARCH.
     */
    if (PbIsChipWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    chipWin = winDataPtr->wd_userWin;
    scrMatrix = winDataPtr->wd_scrMatrix;

    wp3.p_x = wp2.p_x = RIGHT_SIDE(drawPartOfRidgePtr->curTilePtr);
    if (drawPartOfRidgePtr->prevTilePtr == NIL(Tile)) {
        wp2.p_y = planeBndPtr->r_ytop;
        Tp_VPointToHPoint(&wp2, &tmp);
        Mac_CopyPoint(&wp2, &tmp);
    } else {
        wp1.p_y = wp2.p_y = TOP_SIDE(drawPartOfRidgePtr->curTilePtr);
        wp1.p_x = RIGHT_SIDE(drawPartOfRidgePtr->prevTilePtr);

        Tp_VPointToHPoint(&wp1, &tmp);
        Mac_CopyPoint(&wp1, &tmp);

        Tp_VPointToHPoint(&wp2, &tmp);
        Mac_CopyPoint(&wp2, &tmp);

        /*
         * Draw a line from the top right corner of the previous
         * tile to the top left corner of the current tile.
         */
        if (wp1.p_y != wp2.p_y) {
            Ez_DrawWorldLine(chipWin, scrMatrix, wp1.p_x, wp1.p_y,
                wp2.p_x, wp2.p_y, EZ_SOLIDLINE1,
                drawPartOfRidgePtr->displayFunc,
                drawPartOfRidgePtr->ridgeColor,
                pbChipColor);
        }
    }
    /*
     * Draw a line from the top left corner of the current
     * tile to the top right corner of the current tile.
     */
    wp3.p_y = BOTTOM_SIDE(drawPartOfRidgePtr->curTilePtr);
    Tp_VPointToHPoint(&wp3, &tmp);
    Mac_CopyPoint(&wp3, &tmp);
    Ez_DrawWorldLine(chipWin, scrMatrix, wp2.p_x, wp2.p_y,
        wp3.p_x, wp3.p_y, EZ_SOLIDLINE1,
        drawPartOfRidgePtr->displayFunc,
        drawPartOfRidgePtr->ridgeColor,
        pbChipColor);

    wtop.p_x = planeBndPtr->r_xtop;
    wtop.p_y = planeBndPtr->r_ytop;
    Tp_VPointToHPoint(&wtop, &tmp);
    Mac_CopyPoint(&wtop, &tmp);
    if ((rect.r_ybot = wp3.p_y) != (rect.r_ytop = wtop.p_y)) {
        rect.r_xbot = wp2.p_x;
        rect.r_xtop = wp3.p_x;
        Ez_StippleRect(chipWin, scrMatrix, &rect,
            drawPartOfRidgePtr->ridgeColor,
            ez_CargoNetBitmap, drawPartOfRidgePtr->displayFunc);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
static int
FindSpan(tilePtr, clientData)
register Tile   *tilePtr;
ClientData      clientData;
{
    register int x;

    if (Tp_IsSolidTile(tilePtr) && (TOP_SIDE(tilePtr) > searchArea.r_ytop)) {

        x = RIGHT_SIDE(tilePtr);
        if ((x <= spaceTileEnd1) && (x > spanEnd1)) {
            spanEnd1 = x;
        }
        x = LEFT_SIDE(tilePtr);
        if ((x >= spaceTileEnd2) && (x < spanEnd2)) {
            spanEnd2 = x;
        }
    }
    return (0); /* continue search */
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawHorzRidge() --
 *
 * This procedure is designed to be used by an "interator".
 * It draws the entire horizontal ridge in a chip window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbDrawHorzRidge(winDataPtr, drawRidgePtr)
WindowData  *winDataPtr;
DrawRidge   *drawRidgePtr;
{
    DrawPartOfRidge     drawPartOfRidge;
    Plane               *horzPlanePtr;
    PlaneData           *planeDataPtr;
    Tile                *prevTilePtr;
    register LinkedTile *tmpPtr;

    /*
     * If this is not a chip window return BEAR_CONTINUE_SEARCH.
     */
    if (PbIsChipWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    drawPartOfRidge.displayFunc = drawRidgePtr->displayFunc;
    drawPartOfRidge.ridgeColor = drawRidgePtr->ridgeColor;

    planeDataPtr = drawRidgePtr->planeDataPtr;

    planeBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    horzPlanePtr = Tp_GetHorzPlane(planeDataPtr);

    topPseudoTilePtr = horzPlanePtr->pl_top;
    botPseudoTilePtr = horzPlanePtr->pl_bottom;
    leftPseudoTilePtr = horzPlanePtr->pl_left;
    rightPseudoTilePtr = horzPlanePtr->pl_right;

    for (prevTilePtr = NIL(Tile), tmpPtr = drawRidgePtr->tileList;
        tmpPtr != NIL(LinkedTile);
        prevTilePtr = tmpPtr->tilePtr, tmpPtr = tmpPtr->nextPtr) {

        drawPartOfRidge.prevTilePtr = prevTilePtr;
        drawPartOfRidge.curTilePtr = tmpPtr->tilePtr;
        (void) DrawPartOfHorzRidge(winDataPtr, &drawPartOfRidge);
    }
    Ez_Flush();

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawVertRidge() --
 *
 * This procedure is designed to be used by an "interator".
 * It draws the entire vertical ridge in a chip window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbDrawVertRidge(winDataPtr, drawRidgePtr)
WindowData  *winDataPtr;
DrawRidge   *drawRidgePtr;
{
    DrawPartOfRidge     drawPartOfRidge;
    Plane               *vertPlanePtr;
    PlaneData           *planeDataPtr;
    Tile                *prevTilePtr;
    register LinkedTile *tmpPtr;

    /*
     * If this is not a chip window return BEAR_CONTINUE_SEARCH.
     */
    if (PbIsChipWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    drawPartOfRidge.displayFunc = drawRidgePtr->displayFunc;
    drawPartOfRidge.ridgeColor = drawRidgePtr->ridgeColor;

    planeDataPtr = drawRidgePtr->planeDataPtr;

    planeBndPtr = Tp_GetVertPlaneBnd(planeDataPtr);
    vertPlanePtr = Tp_GetVertPlane(planeDataPtr);

    topPseudoTilePtr = vertPlanePtr->pl_top;
    botPseudoTilePtr = vertPlanePtr->pl_bottom;
    leftPseudoTilePtr = vertPlanePtr->pl_left;
    rightPseudoTilePtr = vertPlanePtr->pl_right;

    for (prevTilePtr = NIL(Tile), tmpPtr = drawRidgePtr->tileList;
        tmpPtr != NIL(LinkedTile);
        prevTilePtr = tmpPtr->tilePtr, tmpPtr = tmpPtr->nextPtr) {

        drawPartOfRidge.prevTilePtr = prevTilePtr;
        drawPartOfRidge.curTilePtr = tmpPtr->tilePtr;
        (void) DrawPartOfVertRidge(winDataPtr, &drawPartOfRidge);
    }
    Ez_Flush();

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbFindTile() --
 *
 * Results:
 *   TRUE is returned if everything is spiffy, FALSE otherwise.
 *
 * Side effects:
 *   Upon successfull completion 'tilePtrPtr' will point to the
 *   tile that contains the point whose screen coordinates are
 *   given by 'screenPtr'.
 * ----------------------------------------------------------------------------
 */

bool
PbFindTile(winDataPtr, screenPtr, tilePtrPtr, planeType)
WindowData  *winDataPtr;
Point       *screenPtr;             /* screen coordinates                   */
Tile        **tilePtrPtr;           /* out parameter                        */
PlaneType   planeType;              /* Tells you which plane to search.     */
{
    CellMaster      *cellMasterPtr;
    Plane           *planePtr;
    PlaneData       *planeDataPtr;
    ChipWin         *chipWinPtr;
    Point           wPoint, tmp;
    register double (*invScrMatrix)[4]; /* inverse screen matrix */

    *tilePtrPtr = NIL(Tile);

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    invScrMatrix = winDataPtr->wd_invScrMatrix;
    wPoint.p_x = Ez_XTrans(invScrMatrix, screenPtr->p_x, screenPtr->p_y);
    wPoint.p_y = Ez_YTrans(invScrMatrix, screenPtr->p_x, screenPtr->p_y);
    if (GEO_ENCLOSE(&wPoint, Tp_GetHorzPlaneBnd(planeDataPtr)) == FALSE) {

        PbPrint(EZ_ITALIC_FONT, "Point must be within the tile plane.\n");
        return (FALSE);
    }
    if (planeType & VERT_TILE_PLANE) {
        planePtr = Tp_GetVertPlane(planeDataPtr);
        Tp_HPointToVPoint(&wPoint, &tmp);
        Mac_CopyPoint(&wPoint, &tmp);
    } 
    else {
        planePtr = Tp_GetHorzPlane(planeDataPtr);
    }
    *tilePtrPtr = TiSrPointNoHint(planePtr, &wPoint);

    return (TRUE);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbPickRidge() --
 *
 * This procedure allows you to pick consecutive space tiles from a tile
 * plane (horizontal or vertical) yippie with the mouse to form a ridge.
 * If your picking horizontal tiles you start at the top edge of the
 * plane, otherwise you start at the left edge.  Note that at any time
 * you can stop picking the tile list by selecting a solid tile. A double
 * linked linear list is made to keep track of the tiles as they are picked.
 *
 * SPECIAL NOTE ABOUT CODE:
 *   The reason there are if statements to differentiate between horzontal
 *   and vertical planes (both of which SEEM to be the same) is because
 *   in the future the vertical plane will have the same coordinates as the
 *   horizontal in which case the current top of the vertical plane will be
 *   the left, the left will be the bottom and so on, at which point
 *   these if's will be necessary!  Yeah right.
 *
 * Results:
 *   BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   Upon successfull completion 'spaceTileListPtr' will point to a linked
 *   list of the chosen tiles.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbPickRidge(spaceTileListPtr, planeType)
LinkedTile  **spaceTileListPtr;  /* out parameter                           */
int         planeType;           /* Tells you which plane to pick from.     */
{
    CellMaster          *cellMasterPtr;
    ChipWin             *chipWinPtr;
    DrawRidge           drawRidge;
    DrawPartOfRidge     drawPartOfRidge;
    LinkedTile          *headPtr, *tmpPtr;
    PlaneData           *planeDataPtr;
    Plane               *planePtr;
    Point               screen;
    Tile                *tilePtr;
    WindowData          *winDataPtr;
    BearStatus          (*DrawLinesProc)();
    register LinkedTile *tailPtr;

    *spaceTileListPtr = NIL(LinkedTile);
    headPtr = NIL(LinkedTile);
    pbUserIsPickingRidge = TRUE;

    if (Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
        &winDataPtr, &screen) != EZ_OK) {

        PbPrint(EZ_ITALIC_FONT, "must pick a point in a chip window");
        chipWinPtr = PbGetChipWin(winDataPtr);
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
    }

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    if (PbFindTile(winDataPtr, &screen, &tilePtr, planeType) == FALSE) {
        Ez_RingBell();
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
    }

    drawPartOfRidge.displayFunc = GXxor;
    drawPartOfRidge.ridgeColor = (Pixel) 0xfd;

    if (planeType & HORZ_TILE_PLANE) {
        planePtr = Tp_GetHorzPlane(planeDataPtr);
        planeBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
        searchArea.r_xtop = planeBndPtr->r_xtop;
        searchArea.r_xbot = planeBndPtr->r_xbot;
        DrawLinesProc = DrawPartOfHorzRidge;
    } 
    else if (planeType & VERT_TILE_PLANE) {
        planePtr = Tp_GetVertPlane(planeDataPtr);
        planeBndPtr = Tp_GetVertPlaneBnd(planeDataPtr);
        searchArea.r_xtop = planeBndPtr->r_xtop;
        searchArea.r_xbot = planeBndPtr->r_xbot;
        DrawLinesProc = DrawPartOfVertRidge;
    } 
    else {
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
    }

    topPseudoTilePtr = planePtr->pl_top;
    botPseudoTilePtr = planePtr->pl_bottom;
    leftPseudoTilePtr = planePtr->pl_left;
    rightPseudoTilePtr = planePtr->pl_right;

    if (Tp_IsSolidTile(tilePtr)) {
        /*
         * If the first tile they select is a solid tile then
         * return BEAR_ERROR.
         */
         pbUserIsPickingRidge = FALSE;
         chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
         chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
         return (BEAR_ERROR);
    } 
    else if ((planeType & HORZ_TILE_PLANE) &&
        (TOP_SIDE(tilePtr) != planeBndPtr->r_ytop)) {

        PbPrint(EZ_ITALIC_FONT,
            "the first tile must touch the planes top edge\n");
        Ez_RingBell();
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
    } 
    else if ((planeType & VERT_TILE_PLANE) &&
        (TOP_SIDE(tilePtr) != planeBndPtr->r_ytop)) {

        PbPrint(EZ_ITALIC_FONT,
            "the first tile must touch the planes left edge\n");
        Ez_RingBell();
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
    } 
    else {
        MALLOC(LinkedTile *, headPtr, sizeof (LinkedTile));
        headPtr->prevPtr = NIL(LinkedTile);
        headPtr->nextPtr = NIL(LinkedTile);
        headPtr->tilePtr = tilePtr;
        tailPtr = headPtr;

        /* draw */
        drawPartOfRidge.prevTilePtr = NIL(Tile);
        drawPartOfRidge.curTilePtr = tailPtr->tilePtr;
        (void) PbItrChipWindows(DrawLinesProc, (ClientData) &drawPartOfRidge);
        Ez_Flush();
    }
    while (TRUE) {
        if ((planeType & HORZ_TILE_PLANE) ?
            (BOTTOM_SIDE(tailPtr->tilePtr) == planeBndPtr->r_ybot) :
            (BOTTOM_SIDE(tailPtr->tilePtr) == planeBndPtr->r_ybot)) {

            break;
        }

        spanEnd1 = searchArea.r_xbot;
        spanEnd2 = searchArea.r_xtop;
        searchArea.r_ytop = BOTTOM_SIDE(tailPtr->tilePtr);
        searchArea.r_ybot = BOTTOM_SIDE(LB(tailPtr->tilePtr));
        spaceTileEnd1 = LEFT_SIDE(tailPtr->tilePtr);
        spaceTileEnd2 = RIGHT_SIDE(tailPtr->tilePtr);
        (void) TiSrArea(NIL(Tile), planePtr, &searchArea,
            FindSpan, (ClientData) NULL);

        if (Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
            &winDataPtr, &screen) != EZ_OK) {
            PbPrint(EZ_ITALIC_FONT, "must pick a point in a chip window");
            continue;
        }
        if (PbFindTile(winDataPtr, &screen, &tilePtr, planeType) == FALSE) {
            Ez_RingBell();
            continue;
        } 

        if (Tp_IsSolidTile(tilePtr)) {
            /*
             * Selecting a solid tile is the signal to stop picking a
             * ridge.  To be on the safe side though, ask the user if
             * this is what they really want.
             */
            if (UserWantsToQuit()) {
                (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
                Tp_FreeTileList(headPtr);
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
            }
        } else if (tilePtr == tailPtr->tilePtr) {
            /*
             * If they want to delete the first tile return BEAR_ERROR
             * else delete the tile from the list.
             */
            if (tailPtr->prevPtr == NIL(LinkedTile)) {
                /* erase */
                drawPartOfRidge.prevTilePtr = NIL(Tile);
                drawPartOfRidge.curTilePtr = tailPtr->tilePtr;
                (void) PbItrChipWindows(DrawLinesProc,
                    (ClientData) &drawPartOfRidge);
                Ez_Flush();

                Tp_FreeTileList(headPtr);
        pbUserIsPickingRidge = FALSE;
        chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
        chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
        return (BEAR_ERROR);
            } else {
                /* draw */
                drawPartOfRidge.prevTilePtr = tailPtr->prevPtr->tilePtr;
                drawPartOfRidge.curTilePtr = tailPtr->tilePtr;
                (void) PbItrChipWindows(DrawLinesProc,
                    (ClientData) &drawPartOfRidge);
                Ez_Flush();

                tailPtr = tailPtr->prevPtr;
            }
        } else if ((planeType & HORZ_TILE_PLANE) ?

            (((TOP_SIDE(tilePtr) == BOTTOM_SIDE(tailPtr->tilePtr)) &&
             (LEFT_SIDE(tilePtr) >= spanEnd1) &&
             (RIGHT_SIDE(tilePtr) <= spanEnd2)) == FALSE) :

            (((TOP_SIDE(tilePtr) == BOTTOM_SIDE(tailPtr->tilePtr)) &&
             (LEFT_SIDE(tilePtr) >= spanEnd1) &&
             (RIGHT_SIDE(tilePtr) <= spanEnd2)) == FALSE)) {

            PbPrint(EZ_ITALIC_FONT,
                "the next tile must touch the previous one\n");
            Ez_RingBell();
        } else {
            /*
             * If you have some space at the bottom of the list use it,
             * otherwise get some memory and enlarge the list.  Either
             * way add the new space tile onto the bottom of the list.
             */
            if (tailPtr->nextPtr != NIL(LinkedTile)) {
                tailPtr = tailPtr->nextPtr;
                tailPtr->tilePtr = tilePtr;
            } else {
                MALLOC(LinkedTile *, tmpPtr, sizeof (LinkedTile));
                tmpPtr->prevPtr = tailPtr;
                tmpPtr->nextPtr = tailPtr->nextPtr;
                tailPtr->nextPtr = tmpPtr;
                tmpPtr->tilePtr = tilePtr;
                tailPtr = tmpPtr;
            }
            /* draw */
            drawPartOfRidge.prevTilePtr = tailPtr->prevPtr->tilePtr;
            drawPartOfRidge.curTilePtr = tailPtr->tilePtr;
            (void) PbItrChipWindows(DrawLinesProc,
                (ClientData) &drawPartOfRidge);
            Ez_Flush();
        }
    }

    if (tailPtr->nextPtr != NIL(LinkedTile)) {
        Tp_FreeTileList(tailPtr->nextPtr);
        tailPtr->nextPtr = NIL(LinkedTile);
    }

    /*
     * Redraw the ridge with the high-light color in all of the
     * chip windows so that the unix command xwd (window image
     * dumper) will work.
     */
    drawRidge.displayFunc = GXcopy;
    drawRidge.planeDataPtr = planeDataPtr;
    drawRidge.ridgeColor = ez_HighLightColor;
    drawRidge.tileList = headPtr;
    if (planeType & HORZ_TILE_PLANE) {
        (void) PbItrChipWindows(PbDrawHorzRidge, (ClientData) &drawRidge);
    } else {
        (void) PbItrChipWindows(PbDrawVertRidge, (ClientData) &drawRidge);
    }
                
    *spaceTileListPtr = headPtr;

    pbUserIsPickingRidge = FALSE;
    chipWinPtr->chw_sharedMemPtr->horzTileList = NIL(LinkedTile);
    chipWinPtr->chw_sharedMemPtr->vertTileList = NIL(LinkedTile);
    return (BEAR_OK);
}

#endif /* BEAR_BLACK_BOX */


