
/* pbPinCmds.c --
 *
 *     Sangjin Hong
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbPinCmds.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "grEditor.h"
#include "polarBear.h"


/* ---------- Local Procedures ---------- */
BearStatus CheckPinPos();
PinInst    *GetPinOfCell();


/* ARGSUSED */
void
PbAddPinToNetCmd(winDataPtr, screenPtr, argCount, argList)
WindowData      *winDataPtr;
Point           *screenPtr;
int             argCount;
char            **argList;
{
    ChipWin          *chipWinPtr;
    Net              *netPtr;
    PortInst         *portInstPtr;
    PinInst          *pinInstPtr;
    Window           win;
    char             answer[80];   
    bool             done;

    chipWinPtr = PbGetChipWin(winDataPtr);
    PbPutNetOfCHW(FPG_NW, NIL(Net), chipWinPtr);
    PbPutNetOfCHW(FPG_CENTER, NIL(Net), chipWinPtr);
    PbPutNetOfCHW(FPG_SE, NIL(Net), chipWinPtr);
    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

    /* Pick a net */
    if (Gre_PickNet(winDataPtr, &netPtr, &pinInstPtr) == BEAR_ERROR)
    {
        Ez_RingBell();
        return;
    }

    /* Insert a port a net until done == TRUE */

    done = FALSE;
    while (done == FALSE) {
        PbPrint(EZ_NORMAL_FONT, "Pick a port.\n");
        if (Gre_GetPin(winDataPtr, (BearString) NULL, (BearString) NULL,
            ez_HighLightColor, &pinInstPtr) != BEAR_OK) {
            PbPrint(EZ_ITALIC_FONT, "Cannot insert port to a net.\n");
            return;
        }

    portInstPtr = Mac_GetPortOfPin(pinInstPtr);
    if (Mac_GetNetOfPin(pinInstPtr) != NIL(Net)) {
        PbPrint(EZ_ITALIC_FONT, "this port already belongs to a net.\n");
            do {
            PbPrint(EZ_NORMAL_FONT,
                    "delete this port from the previous net? (y,n) ");
            
            } while (Ez_GetLine(&win, screenPtr, answer, 80, &pbCmdLoop) != 
                EZ_OK);
        if (answer[0] == 'y') {
            if (Mac_DeletePortInstOfNet(portInstPtr) == BEAR_ERROR) {
                PbPrint(EZ_ITALIC_FONT, "cannot delete this port.\n");
                }
            if (Mac_InsertPortInstOfNet(netPtr, portInstPtr) == BEAR_ERROR) {
                PbPrint(EZ_ITALIC_FONT, "cannot insert port to a net\n");
                return;
            }
        }
    }
    else if (Mac_InsertPortInstOfNet(netPtr, portInstPtr) == BEAR_ERROR) {
        PbPrint(EZ_ITALIC_FONT, "cannot insert port to a net\n");
        return;
    }
        do {
            PbPrint(EZ_NORMAL_FONT, "insert more port to a net? (y,n) ");
        } while (Ez_GetLine(&win, screenPtr, answer, 80, &pbCmdLoop) != EZ_OK);
        done  = (answer[0] == 'n') ? TRUE : FALSE;
    }
}



/* ARGSUSED */
void
PbCreatePinCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellMaster  *cellMasterPtr;
    CellInst    *cellInstPtr, *curCellInstPtr;
    ChipWin     *chipWinPtr;
    DrawPin     drawPin;
    Line        span;
    LineSeg     *lineSegPtr;
    PinInst     *pinInstPtr;
    PinMaster   *pinMasterPtr;
    PortMaster  *portMasterPtr;
    Point       absPos, relPos;
    Transform   invTransform;
    char        pinMasterName[80], pinInstName[80];
    char        portMasterName[80], portInstName[80];

    PbPrint(EZ_NORMAL_FONT, "Pick location for new pin.\n");
    if (Gre_PickPinArea(&cellInstPtr, &lineSegPtr, &absPos, &span)
        == BEAR_OK)
    { 

        cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);
        GeoInvertTrans(&cellInstPtr->transform, &invTransform);
        GeoTransPoint(&invTransform, &absPos, &relPos);

        if (Mac_ItrPinOfLineSeg(lineSegPtr, CheckPinPos,
            (ClientData) &relPos) == BEAR_STOP_SEARCH) 
        {
            PbPrint(EZ_ITALIC_FONT, "You already have a pin there !\n");
            return;
        }

        sprintf (portMasterName, "%d", Mac_GenerateId());
        sprintf (pinMasterName, "%d", Mac_GenerateId());
        if (Mac_CreatePinMaster(cellMasterPtr, &portMasterPtr, portMasterName,
            pinMasterName, &pinMasterPtr) == BEAR_ERROR)
        {
            PbPrint(EZ_ITALIC_FONT, "can not create pin.\n");
            return;
        }
        Mac_CopyPoint(&pinMasterPtr->posUn.position, &relPos);
        for (curCellInstPtr = cellMasterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr)
        {
            sprintf(portInstName, "%d", Mac_GenerateId());
            sprintf(pinInstName, "%d", Mac_GenerateId());
            if (Mac_CreatePinOfMaster(curCellInstPtr, portMasterPtr,
                portInstName, pinMasterPtr, pinInstName, &pinInstPtr) ==
                BEAR_ERROR)
            {
                PbPrint(EZ_ITALIC_FONT, "Cannot create pin.\n");
                return;
            }
            
            if (Mac_InsertPinOfLineSeg(Mac_GetBoundaryOfCell(curCellInstPtr),
                pinInstPtr) == BEAR_ERROR)
            {
                PbPrint(EZ_ITALIC_FONT, "Cannot insert pin.\n");
                return;
            }

            chipWinPtr = PbGetChipWin(winDataPtr);
            PbPutUpdateStatusOfCHW(TRUE, chipWinPtr);

            drawPin.pinInstPtr = pinInstPtr;
            drawPin.pinColor = pbCellBorderColor;;
            (void) PbItrChipWindows(Gre_DrawPinTwo, (ClientData) &drawPin);
        }
    }
    Ez_Flush();
}



/* ARGSUSED */
void
PbDuplicatePinCmd(winDataPtr,screenPtr,argCount,argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellMaster  *cellMasterPtr;
    CellInst    *cellInstPtr, *curCellInstPtr;
    ChipWin     *chipWinPtr;
    DrawPin     drawPin;
    Line        span;
    LineSeg     *lineSegPtr;
    Transform   invTransform;
    PortInst    *portInstPtr;
    PinMaster   *pinMasterPtr;
    PinInst     *pinInstPtr,*curPinInstPtr,*oldPinInstPtr;
    Point       absPos,relPos;
    char        pinInstName[80], pinMasterName[80];


    PbPrint(EZ_NORMAL_FONT, "pick an original pin.\n");
    if (Gre_GetPin(winDataPtr, (BearString) NULL, (BearString) NULL,
        ez_HighLightColor, &oldPinInstPtr) != BEAR_OK) {
        return;
    }

    PbPrint(EZ_NORMAL_FONT, "pick a location of duplicated pin.\n");
    if (Gre_PickPinArea(&cellInstPtr,&lineSegPtr,&absPos,&span) == BEAR_ERROR)
    {
    return;
    }

    cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);
    GeoInvertTrans(&cellInstPtr->transform, &invTransform);
    GeoTransPoint(&invTransform, &absPos, &relPos);
    if (Mac_ItrPinOfLineSeg(lineSegPtr,CheckPinPos,(ClientData) &relPos) ==
        BEAR_STOP_SEARCH)
    {
        PbPrint(EZ_ITALIC_FONT, "you already have a pin there!\n");
        return;
    }

    sprintf(pinMasterName, "%d", Mac_GenerateId());
    if (Mac_DuplicatePinMaster(oldPinInstPtr->masterPtr, pinMasterName, 
        &pinMasterPtr) == BEAR_ERROR)
    {
        PbPrint(EZ_ITALIC_FONT, "can not duplicate pin.\n");
        return;
    }
    Mac_CopyPoint(&pinMasterPtr->posUn.position, &relPos);
    for (curCellInstPtr = cellMasterPtr->instList;
        curCellInstPtr != NIL(CellInst);
        curCellInstPtr = curCellInstPtr->nextPtr)
    {
        curPinInstPtr = GetPinOfCell(oldPinInstPtr, curCellInstPtr);
        if (curPinInstPtr == NIL(PinInst))
        {
            PbPrint(EZ_ITALIC_FONT, "can not duplicate pin..\n");
            return;
        }
        sprintf(pinInstName, "%d", Mac_GenerateId());
        portInstPtr = curPinInstPtr->portPtr;
        if (Mac_DuplicatePin(portInstPtr,pinMasterPtr,pinInstName,
        &pinInstPtr) == BEAR_ERROR)
        {
            PbPrint(EZ_ITALIC_FONT, "can not duplicate pin\n");
            return;
        }

        if (Mac_InsertPinOfLineSeg(Mac_GetBoundaryOfCell(curCellInstPtr),
        pinInstPtr) == BEAR_ERROR)
        {
            PbPrint(EZ_ITALIC_FONT, "can not duplicate pin...\n");
            return;
        }

        chipWinPtr=PbGetChipWin(winDataPtr);
        PbPutUpdateStatusOfCHW(TRUE,chipWinPtr);
        drawPin.pinInstPtr=pinInstPtr;
        drawPin.pinColor=pbCellBorderColor;;
        (void) PbItrChipWindows(Gre_DrawPinTwo,(ClientData) &drawPin);
    }
    Ez_Flush();
}

PinInst *
GetPinOfCell (oldPinInstPtr,curCellInstPtr)
PinInst    *oldPinInstPtr;
CellInst   *curCellInstPtr;
{
    PinInst        *curPinInstPtr;
    LinkedLineSeg  *curLinkedLineSeg;

    for (curLinkedLineSeg = Mac_GetBoundaryOfCell(curCellInstPtr);
        curLinkedLineSeg != NIL (LinkedLineSeg);
        curLinkedLineSeg = curLinkedLineSeg->nextPtr)
    {
        for (curPinInstPtr = curLinkedLineSeg->lineSeg->pinList;
             curPinInstPtr != NIL (PinInst);
             curPinInstPtr = curPinInstPtr->nextInCellPtr)
        {
            if ((oldPinInstPtr->masterPtr->posUn.position.p_x ==
                curPinInstPtr->masterPtr->posUn.position.p_x) &&
               (oldPinInstPtr->masterPtr->posUn.position.p_y ==
                curPinInstPtr->masterPtr->posUn.position.p_y))
            {
                return (curPinInstPtr);
            }
        }
    }
    return (curPinInstPtr=NIL(PinInst));
}



static BearStatus
CheckPinPos(pinInstPtr, pointPtr)
PinInst   *pinInstPtr;
Point     *pointPtr;      /* relative position */
{
    Point *relPosPtr;

    relPosPtr = Mac_GetRelPosOfPin(pinInstPtr);
    return (((pointPtr->p_x == relPosPtr->p_x) &&
             (pointPtr->p_y == relPosPtr->p_y)) ? BEAR_STOP_SEARCH :
                                                  BEAR_CONTINUE_SEARCH);
}




/* ARGSUSED */
void
PbDeletePinCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    ChipWin     *chipWinPtr;
    DrawData    drawData;
    PinMaster   *pinMasterPtr;
    PinInst     *pinInstPtr, *curPinInstPtr, *hiLightPinPtr;

    PbPrint(EZ_NORMAL_FONT, "Pick a pin.\n");
    if (Gre_GetPin(winDataPtr, (BearString) NULL, (BearString) NULL,
        ez_HighLightColor, &pinInstPtr) != BEAR_OK) {
        return;
    }

    pinMasterPtr = pinInstPtr->masterPtr;
    for (curPinInstPtr = pinMasterPtr->instList;
        curPinInstPtr != NIL(PinInst);
        curPinInstPtr = curPinInstPtr->nextInMListPtr ) {
        /*
         * Check to see if they are deleting the pin that is
         * currently being high-lighted, and if they are then
         * stop high-lighting it.
         */
        chipWinPtr = PbGetChipWin(winDataPtr);
        hiLightPinPtr = PbGetPinOfCHW(chipWinPtr);
        if (hiLightPinPtr == curPinInstPtr) {
            PbPutPinOfCHW(NIL(PinInst), chipWinPtr);
        }
        drawData.winDataPtr = winDataPtr;
        Gre_ErasePin(curPinInstPtr, &drawData);
        if (Mac_DeletePin(curPinInstPtr) == BEAR_ERROR)
        {
            PbPrint(EZ_ITALIC_FONT, "can not delete pin.\n");
            return;
        }
    }
    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);
}



