
/* pbPrintProp.c --
 *
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbPrintProp.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "stack.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "list.h"
#include "polarBear.h"


typedef struct _PinCount {
    int fixed;
    int floating;
} PinCount;


/* ---------- Local Procedures ---------- */
static BearStatus   AddCellToList();
static BearStatus   CountPinOfCell();
static int          CompareCells();
static void         PrintCellList();


void
PbPrintBnLineProp(bnLinePtr)
BNLine *bnLinePtr;
{
    Line    *linePtr;
    int     i;

    i = 0;
    (void) sprintf (&bear_Msg[i],
        "bottleneck line type: %s\n", Tp_GetTypeNameOfBnLine(bnLinePtr)); 
    i += strlen (&bear_Msg[i]);
    (void) sprintf (&bear_Msg[i], "id: %d\n", Tp_GetIdOfBnLine(bnLinePtr));
    i += strlen (&bear_Msg[i]);

    linePtr = Tp_GetLineOfBnLine(bnLinePtr);
    (void) sprintf (&bear_Msg[i], "p1 (x, y) p2 (x, y): (%d, %d) (%d, %d)\n",
        linePtr->point1.p_x, linePtr->point1.p_y,
        linePtr->point2.p_x, linePtr->point2.p_y);
    i += strlen (&bear_Msg[i]);

    (void) sprintf (&bear_Msg[i], "capacity: %d\n",
        Tp_GetCapacityOfBnLine(bnLinePtr));
    i += strlen (&bear_Msg[i]);
    (void) sprintf (&bear_Msg[i], "density: %d\n",
        Tp_GetDensityOfBnLine(bnLinePtr));
    i += strlen (&bear_Msg[i]);

    PbPrint(EZ_NORMAL_FONT, bear_Msg);
}



static BearStatus
CountNumOfPins(pinPtr, pinCountPtr)
PinInst     *pinPtr;
PinCount    *pinCountPtr;
{
    if (Mac_IsFixedPin(pinPtr) == TRUE) {
        pinCountPtr->fixed++;
    }
    else {
        pinCountPtr->floating++;
    }

    return (BEAR_CONTINUE_SEARCH);
}


void
PbPrintGCellProp(gcellPtr)
GenericCell *gcellPtr;
{
    CellInst    *cellInstPtr;
    DummyCell   *dcellPtr;
    Rect        *bndBoxPtr;
    Transform   *transPtr;
    int         i;

    if (gcellPtr->unType == CELL_INST) {
        CellMaster  *cellMasterPtr;
        PinCount    pinCount;
        char        *stringPtr;

        cellInstPtr = gcellPtr->unValue.cellInstPtr;
        cellMasterPtr = Mac_GetMasterOfCell(cellInstPtr);

        (void) sprintf (bear_Msg, "cell instance name: %s\n",
            Mac_GetNameOfCellInst(cellInstPtr));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        (void) sprintf (bear_Msg, "cell master name:   %s\n",
            Mac_GetNameOfCellMaster(Mac_GetMasterOfCell(cellInstPtr)));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        (void) sprintf (bear_Msg, "cell instance id:   %d\n",
            Mac_GetIdOfGCell(gcellPtr));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        PbPrint(EZ_NORMAL_FONT, "bounding box of cell instance:\n");

        bndBoxPtr = Mac_GetBBoxOfCell(cellInstPtr);
        (void) sprintf (bear_Msg,
            "left = %d  right = %d  bottom = %d  top = %d\n",
            bndBoxPtr->r_xbot, bndBoxPtr->r_xtop,
            bndBoxPtr->r_ybot, bndBoxPtr->r_ytop);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        transPtr = Mac_GetTransformOfGCell(gcellPtr);
        PbPrint(EZ_NORMAL_FONT, "transform:\n");
        (void) sprintf (bear_Msg, "    a = %8d  d = %8d\n",
            transPtr->t_a, transPtr->t_d);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        (void) sprintf (bear_Msg, "    b = %8d  e = %8d\n",
            transPtr->t_b, transPtr->t_e);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        (void) sprintf (bear_Msg, "    c = %8d  f = %8d\n",
            transPtr->t_c, transPtr->t_f);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        switch (cellMasterPtr->shapeFlag) {
        case MAC_FIXED_SHAPE:
            stringPtr = "fixed";
            break;
        case MAC_FLEX_SHAPE:
            stringPtr = "flexible";
            break;
        default:
            stringPtr = "unknown";
            break;
        }
        (void) sprintf (bear_Msg, "cell type:               %s\n", stringPtr);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        if (cellMasterPtr->shapeList != (lsList) NULL) {
            (void) sprintf (bear_Msg, "number of cell shapes:   %4d\n", 
                lsLength(cellMasterPtr->shapeList));
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }

        pinCount.fixed = pinCount.floating = 0;
        Mac_ItrPinOfCellInst(cellInstPtr, CountNumOfPins,
            (ClientData) &pinCount);
        (void) sprintf (bear_Msg, "number of fixed pins:    %4d\n",
            pinCount.fixed);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        (void) sprintf (bear_Msg, "number of floating pins: %4d\n",
            pinCount.floating);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
    }
    else if (gcellPtr->unType == DUMMY_CELL) {
        i = 0;

        dcellPtr = gcellPtr->unValue.dummyCellPtr;
        sprintf (&bear_Msg[i], "dummy cell id: %d\n",
            Mac_GetIdOfGCell(gcellPtr));
        i += strlen(&bear_Msg[i]);
        sprintf (&bear_Msg[i], "bounding box of dummy cell:\n");
        i += strlen(&bear_Msg[i]);
        bndBoxPtr = Mac_GetBBoxOfDCell(dcellPtr);
        sprintf (&bear_Msg[i], "left = %d  right = %d  bottom = %d  top = %d\n",
            bndBoxPtr->r_xbot, bndBoxPtr->r_xtop,
            bndBoxPtr->r_ybot, bndBoxPtr->r_ytop);
        i += strlen(&bear_Msg[i]);

        PbPrint(EZ_NORMAL_FONT, bear_Msg);
    }
    else if (gcellPtr->unType == ROUTE_CELL) {
        i = 0;

        sprintf (&bear_Msg[i], "route cell id: %d\n",
            Mac_GetIdOfGCell(gcellPtr));
        i += strlen(&bear_Msg[i]);

        PbPrint(EZ_NORMAL_FONT, bear_Msg);
    }
    else {
        Bear_PrintErr("PbPrintGCellProp() failed : unknown generic cell type\n");
        return;
    }
}



void
PbPrintJunctProp(junctPtr)
WallJunction *junctPtr;
{
    ClientData  clientData;
    Point       *pointPtr;
    WallSegment *segPtr;
    int         i = 0;

    pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    sprintf (&bear_Msg[i], "center point (x, y) : (%d, %d)\n", pointPtr->p_x,
        pointPtr->p_y);
    i += strlen (&bear_Msg[i]);

    sprintf (&bear_Msg[i], "mark : %s\n",
        Fpg_IsWallJunctMarked(junctPtr) ? "BEAR_MARKED" : "BEAR_UNMARKED");
    i += strlen (&bear_Msg[i]);

    sprintf (&bear_Msg[i], "degree : %d\n", Fpg_GetDegreeOfWallJunct(junctPtr));
    i += strlen (&bear_Msg[i]);

    sprintf (&bear_Msg[i], "junct = 0x%x\n", junctPtr); 
    i += strlen (&bear_Msg[i]);

    /*
     * Print the pointer to this junctions west segment.
     */
    segPtr = Fpg_GetWestSegOfWallJunct(junctPtr);
    if (segPtr == NIL(WallSegment)) {
        sprintf (&bear_Msg[i], "west seg : NIL\n");
    } else {
        sprintf (&bear_Msg[i], "west seg : 0x%x\n", segPtr);
    }
    i += strlen (&bear_Msg[i]);

    /*
     * Print the pointer to this junctions east segment.
     */
    segPtr = Fpg_GetEastSegOfWallJunct(junctPtr);
    if (segPtr == NIL(WallSegment)) {
        sprintf (&bear_Msg[i], "east seg : NIL\n");
    } else {
        sprintf (&bear_Msg[i], "east seg : 0x%x\n", segPtr);
    }
    i += strlen (&bear_Msg[i]);

    /*
     * Print the pointer to this junctions north segment.
     */
    segPtr = Fpg_GetNorthSegOfWallJunct(junctPtr);
    if (segPtr == NIL(WallSegment)) {
        sprintf (&bear_Msg[i], "north seg : NIL\n");
    } else {
        sprintf (&bear_Msg[i], "north seg : 0x%x\n", segPtr);
    }
    i += strlen (&bear_Msg[i]);

    /*
     * Print the pointer to this junctions south segment.
     */
    segPtr = Fpg_GetSouthSegOfWallJunct(junctPtr);
    if (segPtr == NIL(WallSegment)) {
        sprintf (&bear_Msg[i], "south seg : NIL\n");
    } else {
        sprintf (&bear_Msg[i], "south seg : 0x%x\n", segPtr);
    }
    i += strlen (&bear_Msg[i]);

    clientData = Fpg_GetClientDataOfWallJunct(junctPtr);
    if (clientData == (ClientData) NULL) {
        sprintf (&bear_Msg[i], "clientData : NIL\n");
    } else {
        sprintf (&bear_Msg[i], "clientData : 0x%x\n", clientData);
    }
    i += strlen (&bear_Msg[i]);

    PbPrint(EZ_NORMAL_FONT, bear_Msg);
}



void
PbPrintRoomProp(roomPtr)
Fpg_Room *roomPtr;
{
    char    *contentsPtr;
    int     i = 0;

    if (roomPtr != NIL(Fpg_Room)) {
        switch (roomPtr->gcellPtr->unType) {
        case CELL_INST:
            contentsPtr = "cell instance";
            break;
        case DUMMY_CELL:
            contentsPtr = "dummy cell";
            break;
        case UNDEF_CELL:
            contentsPtr = "empty";
            break;
        default:
            contentsPtr = "unknown";
            break;
        }
        sprintf (&bear_Msg[i], "room contents      : %s\n", contentsPtr);
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i], "north seg count    : %d\n",
            lsLength(roomPtr->wallList[FPG_NORTH]));
        i += strlen (&bear_Msg[i]);
        sprintf (&bear_Msg[i], "east seg count     : %d\n",
            lsLength(roomPtr->wallList[FPG_EAST]));
        i += strlen (&bear_Msg[i]);
        sprintf (&bear_Msg[i], "south seg count    : %d\n",
            lsLength(roomPtr->wallList[FPG_SOUTH]));
        i += strlen (&bear_Msg[i]);
        sprintf (&bear_Msg[i], "west seg count     : %d\n",
            lsLength(roomPtr->wallList[FPG_WEST]));
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i],
            "left = %d  right = %d  bottom = %d  top = %d\n",
            roomPtr->rect.r_xbot, roomPtr->rect.r_xtop,
            roomPtr->rect.r_ybot, roomPtr->rect.r_ytop);
        i += strlen(&bear_Msg[i]);

        PbPrint(EZ_NORMAL_FONT, bear_Msg);
    }
}



static BearStatus
AddCellToList(portInstPtr, cellList)
PortInst    *portInstPtr;
lsList      cellList;
{
    lsNewEnd(cellList, (lsGeneric) Mac_GetCellInstOfPort(portInstPtr), LS_NH);

    return (BEAR_CONTINUE_SEARCH);
}


static int
CompareCells(item1, item2)
lsGeneric item1, item2;
{
    CellInst *cellInst1Ptr = (CellInst *) item1;
    CellInst *cellInst2Ptr = (CellInst *) item2;

    return (strcmp (Mac_GetNameOfCellInst(cellInst1Ptr),
        Mac_GetNameOfCellInst(cellInst2Ptr)));
}


static void
PrintCellList(cellList)
lsList cellList;
{
    lsGen       gen;
    CellInst    *cellInstPtr;
    int         i;

    i = 1;
    (void) sprintf (bear_Msg, "list of cells :\n");
    PbPrint(EZ_NORMAL_FONT, bear_Msg);
    lsForeachItem(cellList, gen, cellInstPtr) {
        (void) sprintf (bear_Msg, 
            "%3d) %s\n", i, Mac_GetNameOfCellInst(cellInstPtr));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        i++;
    }
}


void
PbPrintNetProp(netPtr)
Net *netPtr;
{
    if (netPtr == NIL(Net)) {
        int i = 0;

        (void) sprintf (&bear_Msg[i],
            "net name        : UNDEFINED\n");
        i += strlen (&bear_Msg[i]);
        (void) sprintf (&bear_Msg[i],
            "net id          : UNDEFINED\n");
        i += strlen (&bear_Msg[i]);

        PbPrint(EZ_NORMAL_FONT, bear_Msg);
    }
    else {
        lsList  cellList;
        int     netPinCount;
        bool    connFlag, loopFlag;

        (void) sprintf (bear_Msg, 
            "net name        : %s\n", Mac_GetNameOfNet(netPtr));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
        (void) sprintf (bear_Msg, 
            "net id          : %d\n", Mac_GetIdOfNet(netPtr));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        cellList = lsCreate();
        (void) Mac_ItrAllPortOfNet(netPtr, AddCellToList,
            (ClientData) cellList);

        /*
         * Sort the cells and then remove all duplicates.
         */
        lsSort(cellList, CompareCells);
        lsUniq(cellList, CompareCells, (void (*)()) NULL);

        (void) sprintf (bear_Msg, 
            "number of cells : %d\n", lsLength(cellList));
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        /*
         * Count up all of the pins that are in this net's port list.
         */
        netPinCount = 0;
        (void) Mac_NumOfPinsInPortListOfNet(netPtr, &netPinCount);
        (void) sprintf (bear_Msg, 
            "number of pins  : %d\n", netPinCount);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        PrintCellList(cellList);

        lsDestroy(cellList, (void (*)()) NULL);

        if (Mac_GetUserGraphOfNet(netPtr) != (Gt_Graph) NULL) {
            if (Mac_CheckUserGraph(netPtr, &loopFlag, &connFlag) == BEAR_OK) {
                sprintf (bear_Msg, "user's graph is connected: %s\n",
                    connFlag ? "TRUE" : "FALSE");
                PbPrint(EZ_NORMAL_FONT, bear_Msg);
                sprintf (bear_Msg, "user's graph has a loop: %s\n",
                    loopFlag ? "  TRUE" : "  FALSE");
                PbPrint(EZ_NORMAL_FONT, bear_Msg);
            }
        }
    }
}



void
PbPrintPinProp(pinInstPtr)
PinInst *pinInstPtr;
{
    CellInst    *cellInstPtr = NIL(CellInst);
    Net         *netPtr = NIL(Net);
    PinKind     pinKind;
    PortInst    *portInstPtr = NIL(PortInst);
    PortKind    portKind;
    int         i = 0;
    float       *widthVect;
    Point       absPos;

    netPtr = Mac_GetNetOfPin(pinInstPtr);
    portInstPtr = Mac_GetPortOfPin(pinInstPtr);
    cellInstPtr = Mac_GetCellInstOfPort(portInstPtr);

    sprintf (&bear_Msg[i], "%-18s : %s\n", "pin name",
        Mac_GetNameOfPinInst(pinInstPtr));
    i += strlen (&bear_Msg[i]);

    if (netPtr == NIL(Net)) {
        sprintf (&bear_Msg[i], "%-18s : UNDEFINED\n", "net name");
    }
    else {
        sprintf (&bear_Msg[i], "%-18s : %s\n", "net name",
            Mac_GetNameOfNet(netPtr));
    }
    i += strlen (&bear_Msg[i]);

    if (cellInstPtr == NIL(CellInst)) {
        sprintf (&bear_Msg[i], "%-18s : UNDEFINED\n", "cell name");
    }
    else {
        sprintf (&bear_Msg[i], "%-18s : %s\n", "cell name",
            Mac_GetNameOfCellInst(cellInstPtr));
    }
    i += strlen (&bear_Msg[i]);

    sprintf (&bear_Msg[i], "%-18s : ", "pin kind");
    i += strlen(&bear_Msg[i]);
    switch (Mac_GetKindOfPin(pinInstPtr)) {
    case MAC_BORDER_PIN:
        sprintf (&bear_Msg[i], "border\n");
        break;
    case MAC_INSIDE_PIN:
        sprintf (&bear_Msg[i], "inside\n");
        break;
    default:
        Bear_PrintErr("PbPrintPinProp() failed : illegal pin kind");
        break;
    }
    i += strlen(&bear_Msg[i]);

    sprintf (&bear_Msg[i], "%-18s : ", "pin flex type");
    i += strlen(&bear_Msg[i]);
    switch (Mac_GetFlexOfPin(pinInstPtr)) {
    case (Flexibility) FIXED:
        sprintf (&bear_Msg[i], "fixed\n");
        break;
    case (Flexibility) FLEXIBLE:
        sprintf (&bear_Msg[i], "floating\n");
        break;
    case (Flexibility) FLEXIBLE_NORTH:
        sprintf (&bear_Msg[i], "floating (north)\n");
        break;
    case (Flexibility) FLEXIBLE_EAST:
        sprintf (&bear_Msg[i], "floating (east)\n");
        break;
    case (Flexibility) FLEXIBLE_SOUTH:
        sprintf (&bear_Msg[i], "floating (south)\n");
        break;
    case (Flexibility) FLEXIBLE_WEST:
        sprintf (&bear_Msg[i], "floating (west)\n");
        break;
    default:
        Bear_PrintErr("PbPrintPinProp() failed : illegal flexibility type");
        break;
    }
    i += strlen(&bear_Msg[i]);

    if (Mac_IsPinPosDef(pinInstPtr) == TRUE) {
        (void) Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
        sprintf (&bear_Msg[i], "%-18s : %d %d\n", "abs pin pos",
            absPos.p_x, absPos.p_y);
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i], "%-18s : ", "pin dir");
        i += strlen(&bear_Msg[i]);
        switch (Mac_GetDirectionOfPin(pinInstPtr)) {
        case (PinDirection) NORTH_PIN:
            sprintf (&bear_Msg[i], "north\n");
            break;
        case (PinDirection) EAST_PIN:
            sprintf (&bear_Msg[i], "east\n");
            break;
        case (PinDirection) SOUTH_PIN:
            sprintf (&bear_Msg[i], "south\n");
            break;
        case (PinDirection) WEST_PIN:
            sprintf (&bear_Msg[i], "west\n");
            break;
        default:
            Bear_PrintErr("PbPrintPinProp() failed : illegal pin direction");
            break;
        }
        i += strlen(&bear_Msg[i]);
    }
    else {
        sprintf (&bear_Msg[i], "%-18s : UNDEFINED\n", "abs pin pos");
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i], "%-18s : UNDEFINED\n", "pin dir");
        i += strlen (&bear_Msg[i]);
    }

    sprintf (&bear_Msg[i], "%-18s : %d\n", "layer of pin",
       Mac_GetLayerOfPin(pinInstPtr));
    i += strlen(&bear_Msg[i]);

    widthVect = Mac_GetWidthVectOfPin(pinInstPtr);
    sprintf (&bear_Msg[i], "%-18s : %1.3f\n", "layer 1 pin width",
        widthVect[METAL1_LAYER]);
    i += strlen(&bear_Msg[i]);
    sprintf (&bear_Msg[i], "%-18s : %1.3f\n", "layer 2 pin width",
        widthVect[METAL2_LAYER]);
    i += strlen(&bear_Msg[i]);

    if (netPtr != NIL(Net)) {
        sprintf (&bear_Msg[i], "-----\n");
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i], "%-18s : %d\n", "net id",
            Mac_GetIdOfNet(netPtr));
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i], "%-18s : %d\n", "int net id",
            Mac_GetIntNetIdOfPin(pinInstPtr));
        i += strlen(&bear_Msg[i]);
        sprintf (&bear_Msg[i], "%-18s : %d\n", "ext net id",
            Mac_GetExtNetIdOfPin(pinInstPtr));
        i += strlen(&bear_Msg[i]);
        sprintf (&bear_Msg[i], "%-18s : %d\n", "H ext net id",
            Mac_GetHExtNetIdOfPin(pinInstPtr));
        i += strlen(&bear_Msg[i]);
        sprintf (&bear_Msg[i], "%-18s : %d\n", "V ext net id",
            Mac_GetVExtNetIdOfPin(pinInstPtr));
        i += strlen(&bear_Msg[i]);
    }

    if (portInstPtr != NIL(PortInst)) {
        sprintf (&bear_Msg[i], "-----\n");
        i += strlen (&bear_Msg[i]);

        sprintf (&bear_Msg[i], "%-18s : %s\n", "port name",
            Mac_GetNameOfPortInst(portInstPtr));
        i += strlen (&bear_Msg[i]);

        portKind = Mac_GetKindOfPort(portInstPtr);
        if (portKind == (PortKind) SIGNAL_PORT) {
            sprintf (&bear_Msg[i], "%-18s : signal\n", "port kind");
            i += strlen (&bear_Msg[i]);
        }
        else if (portKind == (PortKind) PG_PORT ||
            portKind == (PortKind) PG_SRC_PORT) {

            sprintf (&bear_Msg[i], "%-18s : p/g %s\n", "port kind",
                (portKind == (PortKind) PG_SRC_PORT) ? "(source)" : "");
            i += strlen (&bear_Msg[i]);

            sprintf (&bear_Msg[i], "%-18s : %1.3f\n", "port's max volt drop",
                Mac_GetMaxVoltDropOfPort(portInstPtr));
            i += strlen (&bear_Msg[i]);
            sprintf (&bear_Msg[i], "%-18s : %1.3f\n", 
                "port's max current requirement",
                Mac_GetMaxCurrentReqOfPort(portInstPtr));
            i += strlen (&bear_Msg[i]);
        }
        else {
            sprintf (&bear_Msg[i], "%-18s : UNDEFINED\n", "port kind");
        }
    }

    PbPrint(EZ_NORMAL_FONT, bear_Msg);
}



void
PbPrintTileProp(planeDataPtr, tilePtr)
PlaneData   *planeDataPtr;
Tile        *tilePtr;
{
    GenericCell gcell;
    Rect        rect;
    int         i;

    if (Tp_IsSolidTile(tilePtr)) {
        i = 0;
        sprintf (&bear_Msg[i], "%s\n", Tp_GetTypeNameOfTile(tilePtr)); 
        i += strlen(&bear_Msg[i]);
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

        if (Tp_IsCellInstTile(tilePtr)) {
            gcell.unType = CELL_INST;
            gcell.unValue.cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        } else if (Tp_IsDCellTile(tilePtr)) {
            gcell.unType = DUMMY_CELL;
            gcell.unValue.dummyCellPtr = Tp_GetDCellOfTile(tilePtr);
        } else if (Tp_IsRCellTile(tilePtr)) {
            gcell.unType = ROUTE_CELL;
            gcell.unValue.routeCellPtr = Tp_GetRCellOfTile(tilePtr);
        }
        if (Tp_IsRingTile(tilePtr) == TRUE) {
            /*
             PbPrintRingProp(Tp_GetRingOfTile(tilePtr));
             */
        }
        else {
            PbPrintGCellProp(&gcell);
        }
    } else {
        /* space tile */
        i = 0;
        sprintf (&bear_Msg[i], "%s\n", Tp_GetTypeNameOfTile(tilePtr)); 
        i += strlen(&bear_Msg[i]);
        sprintf (&bear_Msg[i], "tile id: %d\n", Tp_GetIdOfTile(tilePtr));
        i += strlen(&bear_Msg[i]);
        if (Tp_IsHorzTile(tilePtr)) {
            Tp_HorzTileToRect(planeDataPtr, tilePtr, &rect);
        } else {
            Tp_VertTileToRect(planeDataPtr, tilePtr, &rect);
        }
        sprintf (&bear_Msg[i],
            "left = %d  right = %d  bottom = %d  top = %d\n",
            rect.r_xbot, rect.r_xtop,
            rect.r_ybot, rect.r_ytop);
        i += strlen(&bear_Msg[i]);

    /*
        if (Tp_IsBottleneckTile(tilePtr)) {
    */
            sprintf (&bear_Msg[i], "capacity %d\n",
                Tp_GetCapacityOfTile(tilePtr));
            i += strlen(&bear_Msg[i]);
            sprintf (&bear_Msg[i], "density %d\n",
                Tp_GetDensityOfTile(tilePtr));
            i += strlen(&bear_Msg[i]);
            sprintf (&bear_Msg[i], "ridgeWidth %d\n",
                Tp_GetRidgeWidthOfTile(tilePtr));
            i += strlen(&bear_Msg[i]);
    /*
        }
    */
        PbPrint(EZ_NORMAL_FONT, bear_Msg);

#ifdef DEBUG
        sprintf (bear_Msg, "prevTilePtr = 0x%x\n",
            Tp_GetPrevTileOfTile(tilePtr)); 
        PbPrint(EZ_NORMAL_FONT, bear_Msg);
#endif DEBUG
    }
#ifdef DEBUG
    sprintf (bear_Msg, "tilePtr = 0x%x\n", tilePtr); 
    PbPrint(EZ_NORMAL_FONT, bear_Msg);
#endif DEBUG
}



