
/* pbWinCmds.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 * Contains interactive commands the user may call from the
 * console window to manipulate windows.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbWinCmds.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#if !defined(BEAR_BLACK_BOX)
/*
 * This file does not need to be compiled when the black box version of
 * the BEAR-FP system is built.
 */

#include "bearCopyRight.h"
#include "port.h"
#include <sys/time.h>
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"

#ifdef USE_OCT_CODE
#include "octInterface.h"
#endif USE_OCT_CODE

#include "ezPlot.h"
#include "bearGrUtils.h"
#include "deque.h"
#include "polarBear.h"
#include "../bearGrUtils/bguCifPlotInt.h"
#include "stack.h"


/* ARGSUSED */
void
PbCenterCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CifWin      *cifWinPtr;
    ChipWin     *chipWinPtr;
    CellMaster  *parentCellPtr;
    PlaneData   *planeDataPtr;
    Rect        *objBndPtr;

    if (PbIsCifWin(winDataPtr)) {
        cifWinPtr = PbGetCifWin(winDataPtr);
        objBndPtr = PbGetCifBndOfCW(cifWinPtr);
    } 
    else if (Ez_IsConsoleWin(winDataPtr)) {
        PbPrint(EZ_ITALIC_FONT,
            "center command is invalid in the console window\n");
        return;
    } 
    else if (PbIsChipWin(winDataPtr)) {
        chipWinPtr = PbGetChipWin(winDataPtr);
        parentCellPtr = PbGetCellMasterOfCHW(chipWinPtr);
        planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
        objBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    } 
    else {
        Bear_PrintErr("PbCenterCmd() : unknown window type");
        return;
    }

    PbCenterView(winDataPtr, objBndPtr);

    (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
}



/* ARGSUSED */
void
PbClearConsoleCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    Ez_ClearConWin();
}



/* ARGSUSED */
void
PbCloseWindowCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CloseWinData    closeWinData;
    int             i;

    closeWinData.force = FALSE;

    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        } 
        else if (strcmp (argList[i], "-f") == 0) {
            closeWinData.force = TRUE;
            continue;
        }           
        goto usage;
    }

    if (Ez_IsConsoleWin(winDataPtr)) {
        PbCloseConsoleWin(winDataPtr, (CloseWinData *) &closeWinData);
    }
    else {
        if (winDataPtr->wd_closeWinProc == (void (*)()) NULL) {
       Bear_PrintErr("PbCloseWindowCmd() : close window procedure is missing.");
        }
        else {
            (*winDataPtr->wd_closeWinProc)(winDataPtr,
                (CloseWinData *) &closeWinData);
        }
    }

    return;

usage:
    PbPrint(EZ_NORMAL_FONT, "close window\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: cw [-f]\n");
    PbPrint(EZ_NORMAL_FONT, "options:\n");
    PbPrint(EZ_NORMAL_FONT, "    -f force closure\n");
}



/* ARGSUSED */
void
PbMagAreaCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
/*
 * Magnify area command.
 */
{
    CifWin      *cifWinPtr;
    ChipWin     *chipWinPtr;
    CellMaster  *parentCellPtr;
    PlaneData   *planeDataPtr;
    Point       point;
    Rect        *objBndPtr, newImageBnd, rect;
    bool        isCifWinFlag, isChipWinFlag;

    if (Ez_IsConsoleWin(winDataPtr)) {
        PbPrint(EZ_ITALIC_FONT,
            "mag command is invalid in the console window\n");
        return;
    }

    isCifWinFlag = PbIsCifWin(winDataPtr);
    isChipWinFlag = PbIsChipWin(winDataPtr);
    if (isCifWinFlag == FALSE && isChipWinFlag == FALSE) {
        Bear_PrintErr("PbMagAreaCmd() : unknown window type");
        return;
    }

    if (Ez_PickRect(winDataPtr, &point, &rect) == EZ_OK) {
        newImageBnd.r_xbot = point.p_x;
        newImageBnd.r_ybot = point.p_y;
        newImageBnd.r_xtop = point.p_x + rect.r_xtop;
        newImageBnd.r_ytop = point.p_y + rect.r_ytop;

        if (isChipWinFlag == TRUE) {
            chipWinPtr = PbGetChipWin(winDataPtr);
            parentCellPtr = PbGetCellMasterOfCHW(chipWinPtr);
            planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
            objBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
        }
        else if (isCifWinFlag == TRUE) {
            cifWinPtr = PbGetCifWin(winDataPtr);
            objBndPtr = PbGetCifBndOfCW(cifWinPtr);
        }

        if (GEO_OVERLAP(objBndPtr, &newImageBnd)) {
            Mac_CopyRect(&winDataPtr->wd_imageBnd, &newImageBnd);
            (void) Ez_CalcViewingTrans(winDataPtr);
            (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
        }
        else {
            Ez_RingBell();
        }
    } 
}



/*
int color[5], colorNode, colorEdge;
*/


/* ARGSUSED */
void
PbOpenWindowCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
#ifdef USE_BBL_CODE
    bool        readBBLFlag = FALSE;
#endif /* USE_BBL_CODE */

#ifdef USE_OCT_CODE
    bool        readOctFlag = FALSE;
    char        *cellViewPtr = NIL(char);
#endif /* USE_OCT_CODE */

#ifdef USE_OLDP_CODE
    static int  xImageDim = -1;
    static int  yImageDim = -1;

    FloorWin    *floorWinPtr;
    TreeWin     *treeWinPtr;
    char        response[80];
#endif /* USE_OLDP_CODE */

    CellMaster  *coreMasterPtr;
    ChipWin     *chipWinPtr;
    CifWin      *cifWinPtr;
    PlaneData   *corePlaneDataPtr;
    Point       point, p1, p2;
    Rect        imageBnd, *imageBndPtr;
    Window      win;
    WindowData  *newWinDataPtr, *winDataPtr1;
    double      exPerGrowth;
    int         i, winType, chipWinCount;
    char        tbText[80], tbText1[80];     /* title bar text */
    char        *tmpStringPtr;
    char        *fileNamePtr = NIL(char);
    char        *bfpFileNamePtr = NIL(char);
    char        *techFileNamePtr = NIL(char);
    char        *geometryPtr = NIL(char);
    bool        sameViewFlag = FALSE;
    bool        readBFPFlag = FALSE;
    bool        checkFlag = FALSE;
    bool        placedFlag = TRUE;

    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        /* 
         * Print out this commands help message.
         */
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        }

        if (strcmp (argList[i], "-geom") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            geometryPtr = argList[i];
            continue;
        }
        if (strcmp (argList[i], "-cif") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            fileNamePtr = argList[i];
            winType = PB_CIF_WIN;
            continue;
        }

        if (strcmp (argList[i], "-bfp") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            bfpFileNamePtr = argList[i];
            winType = PB_CHIP_WIN;
            readBFPFlag = TRUE;
            continue;
        }

#ifdef USE_BBL_CODE
        if (strcmp (argList[i], "-bbl") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            fileNamePtr = argList[i];
            winType = PB_CHIP_WIN;
            readBBLFlag = TRUE;
            continue;
        }
#endif USE_BBL_CODE

#ifdef USE_OCT_CODE
        if (strcmp (argList[i], "-oct") == 0) {
            if (i + 3 == argCount) {
                i++;
                fileNamePtr = argList[i];
                i++;
                cellViewPtr = argList[i];
                odi_outchip.contents.facet.cell = "macout";
            }
            else if (i + 4 == argCount) {
                i++;
                fileNamePtr = argList[i];
                i++;
                cellViewPtr = argList[i];
                i++;
                odi_outchip.contents.facet.cell = argList[i];
            }
            else {
                goto usage;
            }
            winType = PB_CHIP_WIN;
            readOctFlag = TRUE;
            continue;
        }
#endif USE_OCT_CODE

        if (strcmp (argList[i], "-C") == 0) {
            checkFlag = TRUE;
            continue;
        }

        if (strcmp (argList[i], "-tech") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            techFileNamePtr = argList[i];
            continue;
        }

        /*
         * The "not placed" option is intended to allow the user
         * to read in data that needs to be placed, i.e. data inwhich
         * the cells overlap one another and as such cannot be painted
         * onto a tile plane.
         */
        if (strcmp (argList[i], "-np") == 0) {
            placedFlag = FALSE;
            continue;
        }

        if (strcmp (argList[i], "-s") == 0) {
            sameViewFlag = TRUE;
            continue;
        }

#ifdef USE_OLDP_CODE
        if (strcmp (argList[i], "-tf") == 0) {
            winType = PB_FLOOR_WIN | PB_TREE_WIN;
            continue;
        }
#endif /* USE_OLDP_CODE */

        goto usage;
    }


#ifdef USE_OLDP_CODE
    if ((fileNamePtr == NIL(char)) && (bfpFileNamePtr == NIL(char)) &&
        !(winType == (PB_FLOOR_WIN | PB_TREE_WIN))) {
#else  /* USE_OLDP_CODE */
    if ((fileNamePtr == NIL(char)) && (bfpFileNamePtr == NIL(char))) {
#endif /* USE_OLDP_CODE */

        /* 
         * Duplicate an existing window.
         */

        if (Ez_IsConsoleWin(winDataPtr)) {
            PbPrint(EZ_ITALIC_FONT, "cannot duplicate the console window\n");
            Ez_RingBell();
            return;
        }
        else if (PbIsPlaceWin(winDataPtr)) {
            PbPrint(EZ_ITALIC_FONT, "cannot duplicate a placement window\n");
            Ez_RingBell();
            return;
        }

        if (sameViewFlag) {
            imageBnd.r_xbot = winDataPtr->wd_imageBnd.r_xbot;
            imageBnd.r_ybot = winDataPtr->wd_imageBnd.r_ybot;
            imageBnd.r_xtop = winDataPtr->wd_imageBnd.r_xtop;
            imageBnd.r_ytop = winDataPtr->wd_imageBnd.r_ytop;
        }
        else {
            if (Ez_PickRect(winDataPtr, &point, &imageBnd) != EZ_OK) {
                return;
            }
            imageBnd.r_xbot = point.p_x;
            imageBnd.r_ybot = point.p_y;
            imageBnd.r_xtop += point.p_x;
            imageBnd.r_ytop += point.p_y;
        }
        if (PbIsCifWin(winDataPtr)) {
            (void) PbDupCifWin(winDataPtr, &imageBnd, &newWinDataPtr);
        }
        else if (PbIsChipWin(winDataPtr)) {
            (void) PbDupChipWin(winDataPtr, &imageBnd, &newWinDataPtr);
        }
        else {
            Bear_PrintErr("PbOpenWindowCmd() failed : unknown window type");
            return;
        }
    }
    else {
        if (winType == PB_CIF_WIN) {
            /*
             * Create a new cif window.
             */
            (void) sprintf (tbText, "%s", fileNamePtr);
            if (Bgu_GetCifPlotBnd(fileNamePtr, &imageBnd) == BEAR_ERROR) {
                return;
            }
            (void) PbCreateCifWinOne(geometryPtr, tbText,
                &imageBnd, &winDataPtr);
            winDataPtr->wd_userWinButtonProc = PbHandleUserWinButton;

            cifWinPtr = PbGetCifWin(winDataPtr);
            MALLOC(char *, tmpStringPtr,
                (strlen (fileNamePtr) + 1) * sizeof (char));
            strcpy (tmpStringPtr, fileNamePtr);
            PbPutFileNameOfCW(tmpStringPtr, cifWinPtr);
            PbPutCifBndOfCW(&imageBnd, cifWinPtr);
        }
        else if (winType == PB_CHIP_WIN) {
            /*
             * Create a new chip window.
             */
            bool startOverFlag = FALSE;

            for (i = 0; i < pbNumOfCmds ;i++) {
                if ((strcmp (pbCommand[i].co_longName, "AutoRoute") == 0) &&
                    (pbCommand[i].co_usageCount > 0)) {
                    startOverFlag = TRUE;
                    break;
                }
                else if ((strcmp (pbCommand[i].co_longName,
                        "RingRouter") == 0) &&
                    (pbCommand[i].co_usageCount > 0)) {
                    startOverFlag = TRUE;
                    break;
                }
            }
            if (startOverFlag == TRUE) {
                /*
                 * After using polarBear on one example the user must
                 * exit because the code of many of the key modules
                 * (for example the ring router) is not re-entrant and so
                 * the results will become unpredictable.
                 */
                PbPrint(EZ_ITALIC_FONT, "Exit the system and begin again.\n");
                return;
            }

            chipWinCount = 0;
            if (PbNumOfChipWindows(&chipWinCount) == BEAR_ERROR) {
                return;
            }
            if (chipWinCount > 0) {
                PbPrint(EZ_ITALIC_FONT,
                    "Only one file may be open at a time.\n");
                return;
            }

            if (PbBeginChip() == BEAR_ERROR) {
                return;
            }

            if (readBFPFlag == TRUE) {
                if (techFileNamePtr != NIL(char)) {
                    PbPrint(EZ_ITALIC_FONT,
                        "Technology file must be specified in bfp file.");
                    return;
                }

                if (PbCreateChipFromBFPData(bfpFileNamePtr,
                    placedFlag, checkFlag, &coreMasterPtr) == BEAR_ERROR) {
                    return;
                }

                fileNamePtr = bfpFileNamePtr;
            }

#ifdef USE_BBL_CODE
            if (readBBLFlag == TRUE) {
                if (PbCreateChipFromBBLData(fileNamePtr,
                    techFileNamePtr,
                    placedFlag, checkFlag, &coreMasterPtr) == BEAR_ERROR) {
                    return;
                }
            }
#endif USE_BBL_CODE

#ifdef USE_OCT_CODE
            if (readOctFlag == TRUE) {
                if (PbCreateChipFromOCTData(fileNamePtr, cellViewPtr,
                    techFileNamePtr,
                    placedFlag, checkFlag, &coreMasterPtr) == BEAR_ERROR) {
                    return;
                }
            }
#endif USE_OCT_CODE

            if (checkFlag == TRUE) {
                if (Mac_CheckCellMaster(coreMasterPtr, "errorList") == 
                    BEAR_ERROR) {
                    (void) sprintf (bear_ErrorMsg,
                        "See list of errors in the file named \"%s\".\n", 
                        "errorList");
                    PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
                    return;
                }
            }

            (void) sprintf (tbText, "%s", fileNamePtr);
            corePlaneDataPtr = Mac_GetCellPlaneOfCell(coreMasterPtr);
            imageBndPtr = Tp_GetHorzPlaneBnd(corePlaneDataPtr),
            (void) PbCreateChipWinOne(geometryPtr, tbText,
                imageBndPtr, &winDataPtr);
            winDataPtr->wd_userWinButtonProc = PbHandleUserWinButton;

            chipWinPtr = PbGetChipWin(winDataPtr);
            PbPutCellMasterOfCHW(coreMasterPtr, chipWinPtr);
        }
#ifdef USE_OLDP_CODE
        else if (winType == (PB_FLOOR_WIN | PB_TREE_WIN)) {
            /*
             * Create a new floor window and a new tree window.
             */

/*
            color[0] = pbNet1Color;
            color[1] = pbNet2Color;
            color[2] = pbNet3Color;
            color[3] = pbPinColor;
            color[4] = pbChipColor;
*/
/*
            color[0] = pbChipColor;
            color[1] = pbChipColor;
            color[2] = pbChipColor;
            color[3] = pbChipColor;
            color[4] = pbChipColor;
*/
/*
            colorNode = pbCellColor;
*/
/*
            colorNode = pbChipColor;
            colorEdge = pbCellBorderColor;
*/

            sprintf (tbText, "BEAR-FP ( %s ) : PLACEMENT", pbHostName);
            if (xImageDim == -1) {
                /*
                PbPrint(EZ_NORMAL_FONT,
                    "Please enter expected percent growth: ");
                (void) Ez_GetLine(&win, screenPtr, response, 80, &pbCmdLoop);
                sscanf(response, "%f", &exPerGrowth);
                exPerGrowth = 100 / exPerGrowth;
                */
                exPerGrowth = 5;
                p1.p_x = 0;
                p1.p_y = 0;
                GeoTransPoint(&cl_ClusterTree->transform, &p1, &p2);
                imageBnd.r_xbot = p2.p_x 
                    - (cl_ClusterTree->dimension.p_x / exPerGrowth);
                imageBnd.r_ybot = p2.p_y 
                    - (cl_ClusterTree->dimension.p_y / exPerGrowth);
                p1.p_x = cl_ClusterTree->dimension.p_x;
                p1.p_y = cl_ClusterTree->dimension.p_y;
                GeoTransPoint(&cl_ClusterTree->transform, &p1, &p2);
                imageBnd.r_xtop = p2.p_x  
                    + (cl_ClusterTree->dimension.p_x / exPerGrowth);
                imageBnd.r_ytop = p2.p_y  
                    + (cl_ClusterTree->dimension.p_y / exPerGrowth);
                xImageDim = imageBnd.r_xtop - imageBnd.r_xbot;
                yImageDim = imageBnd.r_ytop - imageBnd.r_ybot;
                /*
        printf("The image boundary is %d, %d, %d, %d.\n",
                    imageBnd.r_xbot,imageBnd.r_ybot,
                    imageBnd.r_xtop,imageBnd.r_ytop);
                printf("Do you want to change the image boundary? ");
                scanf("%s",response);
                if (response[0] == 'y' || response[0] == 'Y') {
                    printf("Please enter image boundary: ");
                    scanf("%d %d %d %d",
                        &imageBnd.r_xbot,&imageBnd.r_ybot,
                        &imageBnd.r_xtop,&imageBnd.r_ytop);
                }
                */
    /* Else clause can cause problems when we read a new file. MP */
        xImageDim = -1; yImageDim = -1;
            }
            else {
                p1.p_x = 0;
                p1.p_y = 0;
                GeoTransPoint(&cl_ClusterTree->transform, &p1, &p2);
                imageBnd.r_xbot = p2.p_x;
                imageBnd.r_ybot = p2.p_y;
                p1.p_x = cl_ClusterTree->dimension.p_x;
                p1.p_y = cl_ClusterTree->dimension.p_y;
                GeoTransPoint(&cl_ClusterTree->transform, &p1, &p2);
                imageBnd.r_xtop = p2.p_x +
                    (xImageDim - (p2.p_x - imageBnd.r_xbot)) / 2;
                imageBnd.r_ytop = p2.p_y +
                    (yImageDim - (p2.p_y - imageBnd.r_ybot)) / 2;
                imageBnd.r_xbot = imageBnd.r_xbot -
                    (xImageDim - (p2.p_x - imageBnd.r_xbot)) / 2;
                imageBnd.r_ybot = imageBnd.r_ybot -
                    (yImageDim - (p2.p_y - imageBnd.r_ybot)) / 2;
                printf("The image boundary is %d, %d, %d, %d.\n",
                    imageBnd.r_xbot,imageBnd.r_ybot,
                    imageBnd.r_xtop,imageBnd.r_ytop);
                printf("Do you want to change the image boundary? ");
                scanf("%s",response);
                if (response[0] == 'y' || response[0] == 'Y') {
                    printf("Please enter image boundary: ");
                    scanf("%d %d %d %d",
                        &imageBnd.r_xbot,&imageBnd.r_ybot,
                        &imageBnd.r_xtop,&imageBnd.r_ytop);
                }
            }
            (void) PbCreateFloorWinOne(geometryPtr, tbText,
                &imageBnd, &winDataPtr);
            winDataPtr->wd_userWinButtonProc = PbHandleUserWinButton;

            floorWinPtr = PbGetFloorWin(winDataPtr);
            floorWinPtr->level = 0;
            floorWinPtr->tree = cl_ClusterTree;
            floorWinPtr->depth = cl_ClusterTreeDepth;

            sprintf (tbText1, "BEAR-FP ( %s ) : TREE", pbHostName);
            imageBnd.r_xbot = imageBnd.r_ybot = -100;
            imageBnd.r_xtop = (cl_ClusterTree->numOfLeaves) * 100;
            imageBnd.r_ytop = (cl_ClusterTreeDepth * 220) + 100;  
            (void) PbCreateTreeWinOne(geometryPtr, tbText1,
                &imageBnd, &winDataPtr1);
            winDataPtr1->wd_userWinButtonProc = PbHandleUserWinButton;

            treeWinPtr = PbGetTreeWin(winDataPtr1);
            treeWinPtr->level = 0;
            treeWinPtr->tree = cl_ClusterTree;
            treeWinPtr->depth = cl_ClusterTreeDepth;

            PbPrint(EZ_NORMAL_FONT, "Next tree and floorplan level? ");
            (void) Ez_GetLine(&win, screenPtr, response, 80, &pbCmdLoop);
            while ((response[0] == 'y') || (response[0] == 'Y')) {
                floorWinPtr->level++;
                treeWinPtr->level++;
                (void) PbDrawFloorRegion(winDataPtr, NIL(Rect));
                (void) PbDrawTreeRegion(winDataPtr1, NIL(Rect));
                PbPrint(EZ_NORMAL_FONT, "Next tree and floorplan level? ");
                (void) Ez_GetLine(&win, screenPtr, response, 80, &pbCmdLoop);
            }
        }
#endif USE_OLDP_CODE
        else {
            Bear_PrintErr("PbOpenWindowCmd() : unknown window type");
            return;
        }
    }
    return;

usage:
    PbPrint(EZ_NORMAL_FONT, "open window\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: ow [-options ...]\n");
    PbPrint(EZ_NORMAL_FONT, "\nwhere options include:\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -s                duplicate window with same view\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -bfp filename     cell data in BFP format\n");

#ifdef USE_BBL_CODE
    PbPrint(EZ_NORMAL_FONT,
        "    -bbl filename     cell data in BBL format\n");
#endif USE_BBL_CODE

#ifdef USE_OCT_CODE
    PbPrint(EZ_NORMAL_FONT,
        "    -oct cellname viewname [[output_cellname] [output_viewname]]\n");
#endif USE_OCT_CODE

    PbPrint(EZ_NORMAL_FONT,
        "    -tech filename    technology data (bbl and oct only)\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -cif filename     cell data in CIF format\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -np               cells aren't placed\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -C attempt to verify the validity of the data\n");

#ifdef USE_OLDP_CODE
    PbPrint(EZ_NORMAL_FONT,
        "    -tf show tree and corresponding floorplan\n");
#endif USE_OLDP_CODE
}



/* ARGSUSED */
void
PbPanCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
{
    CellMaster  *parentCellPtr;
    CifWin      *cifWinPtr;
    ChipWin     *chipWinPtr;
    PlaneData   *planeDataPtr;
    Point       wPoint;                 /* world coordinate point   */
    Rect        *objBndPtr, *imageBndPtr;
    int         deltaX, deltaY;

    wPoint.p_x = Ez_XTrans(winDataPtr->wd_invScrMatrix, screenPtr->p_x,
        screenPtr->p_y);
    wPoint.p_y = Ez_YTrans(winDataPtr->wd_invScrMatrix, screenPtr->p_x,
        screenPtr->p_y);
    if (PbIsCifWin(winDataPtr)) {
        cifWinPtr = PbGetCifWin(winDataPtr);
        objBndPtr = PbGetCifBndOfCW(cifWinPtr);
    }
    else if (Ez_IsConsoleWin(winDataPtr)) {
        PbPrint(EZ_ITALIC_FONT,
            "pan command is invalid in the console window\n");
        return;
    }
    else if (PbIsChipWin(winDataPtr)) {
        chipWinPtr = PbGetChipWin(winDataPtr);
        parentCellPtr = PbGetCellMasterOfCHW(chipWinPtr);
        planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
        objBndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    }
    else {
        Bear_PrintErr("PbPanCmd() : unknown window type");
        return;
    }
    if (GEO_ENCLOSE(&wPoint, objBndPtr) == FALSE) {
        PbPrint(EZ_ITALIC_FONT, "point is not within the boundary\n");
        Ez_RingBell();
        return;
    }

    /*
     * Center the image boundary (whats in view) around the point
     * they picked.
     */
    imageBndPtr = &winDataPtr->wd_imageBnd;
    deltaX = (imageBndPtr->r_xtop - imageBndPtr->r_xbot) / 2;
    deltaY = (imageBndPtr->r_ytop - imageBndPtr->r_ybot) / 2;
    imageBndPtr->r_xbot = wPoint.p_x - deltaX;
    imageBndPtr->r_xtop = wPoint.p_x + deltaX;
    imageBndPtr->r_ybot = wPoint.p_y - deltaY;
    imageBndPtr->r_ytop = wPoint.p_y + deltaY;
    (void) Ez_CalcViewingTrans(winDataPtr);
    (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));
}



/* ARGSUSED */
void
PbZoomInCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;     /* not used */
int         argCount;
char        **argList;
{
    double  atof ();
    double  mag = 1.0 / PB_DEF_ZOOM_MAG;
    int     i;

    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        } 
        else {
            mag = atof (argList[i]);
            if ((mag = atof (argList[i])) == 0) {
                mag = 1.0 / PB_DEF_ZOOM_MAG;
            } 
            else {
                mag = 1.0 / mag;
            }
            continue;
        }           
    }
    Ez_Zoom(winDataPtr, mag);
    return;

usage:
    PbPrint(EZ_NORMAL_FONT, "zoom in\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: Z [amount]\n");
}



/* ARGSUSED */
void
PbZoomOutCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;     /* not used */
int         argCount;
char        **argList;
{
    double  atof ();
    double  mag = PB_DEF_ZOOM_MAG;
    int     i;

    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        } 
        else {
            if ((mag = atof (argList[i])) == 0) {
                mag = PB_DEF_ZOOM_MAG;
            }
            continue;
        }
    }
    Ez_Zoom(winDataPtr, mag);
    return;

usage:
    PbPrint(EZ_NORMAL_FONT, "zoom out\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: z [amount]\n");
}

#endif /* BEAR_BLACK_BOX */


