
/* pbWinUtils.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbWinUtils.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#if !defined(BEAR_BLACK_BOX)
/*
 * This file does not need to be compiled when the black box version of
 * the BEAR-FP system is built.
 */

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "bearGrUtils.h"
#include "polarBear.h"
#ifdef USE_XDM_CODE
#include "vemDM.h"
#endif USE_XDM_CODE

typedef struct _ApplyData ApplyData;
struct _ApplyData {
    bool        (*idWinProc)();         /* Procedure that identifies the
                                         * correct window.
                                         */
    BearStatus  (*applyProc)();
    ClientData  clientData;
};

typedef struct _BnLineSearch BnLineSearch;
struct _BnLineSearch {
    Rect    box;            /* search area */
    BNLine  *bnLinePtr;
};


/* ---------- Local Procedures ---------- */
static ezItrStatus      ApplyProcToWin();
static BearStatus       CountChipWindows();
static BearStatus       BnLineCrossesBox();
static SharedChipWinMem *MakeSharedChipWinMem();

#ifdef USE_OLDP_CODE
static void DrawRootAndSubtreeFloorPlan();
static void DrawTreeRootAndSubtree();
#endif USE_OLDP_CODE


/* ---------- Local Global Variables ---------- */
#ifdef USE_OLDP_CODE
static int     pbDraw_level, pbDraw_depth;
static Window  pbDraw_window;
static Matrix  pbDraw_matrix;
#endif USE_OLDP_CODE


static ezItrStatus
ApplyProcToWin(winDataPtr, applyDataPtr)
WindowData  *winDataPtr;
ApplyData   *applyDataPtr;
{
    /*
     * If this is the correct type of window then apply the
     * supplied procedure.
     */
    if ((*applyDataPtr->idWinProc)(winDataPtr) == TRUE) {
        if ((*applyDataPtr->applyProc)(winDataPtr,
            applyDataPtr->clientData) == BEAR_STOP_SEARCH) {
            return (EZ_STOP_SEARCH);
        }
    }

    return (EZ_CONTINUE_SEARCH);
}



/* ARGSUSED */
static BearStatus
CountChipWindows(winDataPtr, numPtr)
WindowData  *winDataPtr;
int         *numPtr;
{
    (*numPtr)++;

    return (BEAR_CONTINUE_SEARCH);
}



/* 
 * ----------------------------------------------------------------------------
 *
 * MakeSharedChipWinMem() --
 *
 * Allocates memory for the structure SharedChipWinMem, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to the new structure is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

static SharedChipWinMem *
MakeSharedChipWinMem()
{
    register SharedChipWinMem *sharedMemPtr;

    MALLOC(SharedChipWinMem *, sharedMemPtr, sizeof (SharedChipWinMem));
    sharedMemPtr->cellMasterPtr = NIL(CellMaster);
    sharedMemPtr->upDate = FALSE;
    sharedMemPtr->nwNetPtr = NIL(Net);
    sharedMemPtr->centerNetPtr = NIL(Net);
    sharedMemPtr->seNetPtr = NIL(Net);
    sharedMemPtr->bnLinePtr = NIL(BNLine);
    sharedMemPtr->gcellPtr = NIL(GenericCell);
    sharedMemPtr->pinInstPtr = NIL(PinInst);
    sharedMemPtr->horzTileList = NIL(LinkedTile);
    sharedMemPtr->vertTileList = NIL(LinkedTile);

    return (sharedMemPtr);
}



void
PbCloseChipWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
{
    CellMaster  *cellMasterPtr;
    ChipWin     *chipWinPtr;
    PlaneData   *planeDataPtr;
    int         chipWinCount;

    chipWinPtr = PbGetChipWin(winDataPtr);
    chipWinCount = 0;
    if (PbNumOfChipWindows(&chipWinCount) == BEAR_ERROR) {
        Bear_PrintErr("PbCloseChipWin() failed : couldn't count chip windows.");
        return;
    }

    if (closeWinDataPtr->force ||
        (chipWinCount > 1) ||
        (PbGetUpdateStatusOfCHW(chipWinPtr) == FALSE)) {

        chipWinCount--;

        if (chipWinCount == 0) {
            cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
            (void) PbEndChip(&cellMasterPtr);
        }
        FREE((char *) chipWinPtr);
        FREE((char *) winDataPtr->wd_clientData);
        (void) Ez_DestroyWin(winDataPtr);
    }
    else {
        PbPrint(EZ_NORMAL_FONT, "No write since last change\n");
        return;
    }
}



/* ARGSUSED */
void
PbCloseCifWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
{
    CifWin *cifWinPtr;

    cifWinPtr = PbGetCifWin(winDataPtr);
    FREE((char *) cifWinPtr->cw_fileNamePtr);
    FREE((char *) cifWinPtr);
    FREE((char *) winDataPtr->wd_clientData);
    (void) Ez_DestroyWin(winDataPtr);
}



/* ARGSUSED */
void
PbCloseConsoleWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
/*
 * Side Effects:
 *   When the console window is closed the global variable
 *   'pbCmdLoop' is set to FALSE, which will cause the command
 *   loop to be terminated.
 */
{
    if (closeWinDataPtr->force) {
        PbPrint(EZ_NORMAL_FONT, "\nbye bye\n\n");
        Ez_DestroyConsoleWin();
        pbCmdLoop = FALSE;
        Bear_RemoveTmpFiles();
    }
    else {
#ifdef USE_XDM_CODE
        sprintf (bear_Msg, "%s%s%s",
            "Closing the console window\n",
            "will terminate the program.\n",
            "Do you really want to exit?\n");
        if (dmConfirm("Warning", bear_Msg, "yes", "no") == VEM_OK) {
            PbPrint(EZ_NORMAL_FONT, "\nbye bye\n\n");
            Ez_DestroyConsoleWin();
            pbCmdLoop = FALSE;
            Bear_RemoveTmpFiles();
        }
#else
        PbPrint(EZ_NORMAL_FONT, "\nbye bye\n\n");
        Ez_DestroyConsoleWin();
        pbCmdLoop = FALSE;
        Bear_RemoveTmpFiles();
#endif USE_XDM_CODE
    }
}



/* ARGSUSED */
#ifdef USE_OLDP_CODE
void
PbCloseFloorWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
{
    FloorWin *floorWinPtr;

    floorWinPtr = PbGetFloorWin(winDataPtr);
    FREE((char *) floorWinPtr);
    FREE((char *) winDataPtr->wd_clientData);
    (void) Ez_DestroyWin(winDataPtr);
}
#endif USE_OLDP_CODE



/* ARGSUSED */
#ifdef USE_OLDP_CODE
void
PbCloseTreeWin(winDataPtr, closeWinDataPtr)
WindowData      *winDataPtr;
CloseWinData    *closeWinDataPtr;
{
    TreeWin *treeWinPtr;

    treeWinPtr = PbGetTreeWin(winDataPtr);
    FREE((char *) treeWinPtr);
    FREE((char *) winDataPtr->wd_clientData);
    (void) Ez_DestroyWin(winDataPtr);
}
#endif USE_OLDP_CODE



BearStatus
PbCreateChipWinOne(geometryPtr, tbTextPtr, imageBndPtr, winDataPtrPtr)
char        *geometryPtr;
char        *tbTextPtr;             /* title bar text */
Rect        *imageBndPtr;
WindowData  **winDataPtrPtr;
{
    ChipWin         *chipWinPtr;
    PbWindowType    *winTypePtr;
    char            prompt[80];

    (void) sprintf (prompt, "BEAR-FP ( %s ) : CHIP", pbHostName);
    if (Ez_CreateWin(prompt, geometryPtr, tbTextPtr,
        (void (*)()) PbDrawChipRegion,
        imageBndPtr, winDataPtrPtr) != EZ_OK) {
        return (BEAR_ERROR);
    }

    chipWinPtr = PbMakeChipWin();
    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_CHIP_WIN;
    winTypePtr->unValue.chipWinPtr = chipWinPtr;

    (*winDataPtrPtr)->wd_closeWinProc = PbCloseChipWin;
    (*winDataPtrPtr)->wd_clientData = (ClientData) winTypePtr;

    chipWinPtr->chw_options = PB_SHOW_PINS_OPT;
    chipWinPtr->chw_sharedMemPtr = MakeSharedChipWinMem();

    return (BEAR_OK);
}



BearStatus
PbCreateCifWinOne(geometryPtr, tbTextPtr, imageBndPtr, winDataPtrPtr)
char        *geometryPtr;
char        *tbTextPtr;             /* title bar text */
Rect        *imageBndPtr;
WindowData  **winDataPtrPtr;
{
    CifWin          *cifWinPtr;
    PbWindowType    *winTypePtr;
    char            prompt[80];

    sprintf (prompt, "BEAR-FP ( %s ) : CIF", pbHostName);
    if (Ez_CreateWin(prompt, geometryPtr, tbTextPtr,
        (void (*)()) PbDrawCifRegion,
        imageBndPtr, winDataPtrPtr) != EZ_OK) {
        return (BEAR_ERROR);
    }

    cifWinPtr = PbMakeCifWin();
    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_CIF_WIN;
    winTypePtr->unValue.cifWinPtr = cifWinPtr;

    (*winDataPtrPtr)->wd_closeWinProc = PbCloseCifWin;
    (*winDataPtrPtr)->wd_clientData = (ClientData) winTypePtr;

    return (BEAR_OK);
}



#ifdef USE_OLDP_CODE
BearStatus
PbCreateFloorWinOne(geometryPtr, tbTextPtr, imageBndPtr, winDataPtrPtr)
char        *geometryPtr;
char        *tbTextPtr;             /* title bar text */
Rect        *imageBndPtr;
WindowData  **winDataPtrPtr;
{
    FloorWin        *floorWinPtr;
    PbWindowType    *winTypePtr;
    char            prompt[80];

    sprintf (prompt, "BEAR-FP ( %s ) : PLACEMENT", pbHostName);
    if (Ez_CreateWin(prompt, geometryPtr, tbTextPtr,
        (void (*)()) PbDrawFloorRegion,
        imageBndPtr, winDataPtrPtr) != EZ_OK) {
        return (BEAR_ERROR);
    }

    floorWinPtr = PbMakeFloorWin();
    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_FLOOR_WIN;
    winTypePtr->unValue.floorWinPtr = floorWinPtr;

    (*winDataPtrPtr)->wd_closeWinProc = PbCloseFloorWin;
    (*winDataPtrPtr)->wd_clientData = (ClientData) winTypePtr;

    return (BEAR_OK);
}
#endif USE_OLDP_CODE



#ifdef USE_OLDP_CODE
BearStatus
PbCreateTreeWinOne(geometryPtr, tbTextPtr, imageBndPtr, winDataPtrPtr)
char        *geometryPtr;
char        *tbTextPtr;             /* title bar text */
Rect        *imageBndPtr;
WindowData  **winDataPtrPtr;
{
    TreeWin         *treeWinPtr;
    PbWindowType    *winTypePtr;
    char            prompt[80];

    sprintf (prompt, "BEAR-FP ( %s ) : TREE", pbHostName);
    if (Ez_CreateWin(prompt, geometryPtr, tbTextPtr,
        (void (*)()) PbDrawTreeRegion,
        imageBndPtr, winDataPtrPtr) != EZ_OK) {
        return (BEAR_ERROR);
    }

    treeWinPtr = PbMakeTreeWin();
    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_TREE_WIN;
    winTypePtr->unValue.treeWinPtr = treeWinPtr;

    (*winDataPtrPtr)->wd_closeWinProc = PbCloseTreeWin;
    (*winDataPtrPtr)->wd_clientData = (ClientData) winTypePtr;

    return (BEAR_OK);
}
#endif USE_OLDP_CODE



/*
 * ----------------------------------------------------------------------------
 *
 * PbDefCurInUserSubWin() --
 *
 * This procedure is designed to be used by an "interator".
 * If this is the correct type of window, then the cursor in the "user 
 * sub-window" is changed.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion.
 *
 * Side effects:
 *   The cursor shape (visible when the mouse is in the user
 *   window) is changed.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbDefCurInUserSubWin(winDataPtr, defCurDataPtr)
WindowData      *winDataPtr;
Pb_DefCurData   *defCurDataPtr;
{
    /*
     * If this is the correct type of window change the cursor.
     */
    if ((*defCurDataPtr->idWinProc)(winDataPtr) == TRUE) {
        Ez_DefineCursor(winDataPtr->wd_userWin, defCurDataPtr->cursorType);
    }

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawCifRegion() --
 *
 * This procedure clears and redraws a cif window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion, even if theres
 *   an error because the interator should continue on to the next window.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbDrawCifRegion(winDataPtr, subWinBndPtr)
WindowData  *winDataPtr;
Rect        *subWinBndPtr;              /* subWinBndPtr contains the world
                                         * coordinates of a portion of the
                                         * window that is to be drawn, if it
                                         * is NULL the whole window is drawn.
                                         */
{
    DrawCif drawCif;
    CifWin  *cifWinPtr;
    Rect    *cifBndPtr, clipBnd;

    if (PbIsCifWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    cifWinPtr = PbGetCifWin(winDataPtr);
    drawCif.fileNamePtr = PbGetFileNameOfCW(cifWinPtr);
    cifBndPtr = PbGetCifBndOfCW(cifWinPtr);

    drawCif.win = winDataPtr->wd_userWin;
    drawCif.scrMatrix = winDataPtr->wd_scrMatrix;
    drawCif.options = BGU_ENTIRE_CHIP | BGU_CALC_VIEW_TRANS;

    if (subWinBndPtr == NIL(Rect)) {
        /* 
         * Redraw the whole window.
         */
/*
        if (Ez_FindOverlap(cifBndPtr,
            &winDataPtr->wd_imageBnd,
            &clipBnd) == FALSE) {
            return (BEAR_CONTINUE_SEARCH);
        }
*/
        XClearWindow(ez_Display, winDataPtr->wd_userWin);
    } 
	else {
        /*
         * Redraw part of the window.
         */
/*
        if (Ez_FindOverlap(&cifBndPtr,
            subWinBndPtr,
            &clipBnd) == FALSE) {
            return (BEAR_CONTINUE_SEARCH);
        }
*/
/*
 * The X Server always clears an exposed region to the 
 * background color, so we don't have to.
 */
    }
    (void) Bgu_DrawCifPlot(&drawCif);
    Ez_OutlineRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
        cifBndPtr, EZ_SOLIDLINE1, pbCellBorderColor, pbChipColor);

    Ez_Flush();

    return (BEAR_CONTINUE_SEARCH);
}



#ifdef USE_OLDP_CODE
static void
DrawTreeRootAndSubtree(rootX, rootY, curLevel, numOfChildren,
    children, endColor)

int rootX, rootY;
int curLevel;
int numOfChildren;
Node *children[MAX_CHILDREN];
int endColor;
{
    int i;
    int x[MAX_CHILDREN+1], y;

    if ((curLevel == pbDraw_level) || (numOfChildren == 0)) {
        Ez_DrawWorldSmallDot(pbDraw_window, pbDraw_matrix,
            rootX, rootY, endColor);
    } 
    else {
        curLevel++;
        y = (pbDraw_depth - curLevel) * 220;
        x[0] = MAX(((children[0]->numOfLeaves - 1) * 50),0);
        x[1] = 2 * x[0];
        for (i = 1; i < numOfChildren; i++) {
            x[i] = x[i] + 100 + MAX(((children[i]->numOfLeaves - 1) * 50),0);
            x[i+1] = x[i] + MAX(((children[i]->numOfLeaves - 1) * 50),0);
        }
        x[numOfChildren] = rootX - (x[numOfChildren] / 2);
        for (i = 0; i < numOfChildren; i++) {
            x[i] = x[i] + x[numOfChildren];
            Ez_DrawWorldLine(pbDraw_window, pbDraw_matrix, rootX, rootY,
                x[i], y, EZ_SOLIDLINE2, GXcopy, pbTFEdgeColor, ez_BackColor);
            DrawTreeRootAndSubtree(x[i], y, curLevel,
                children[i]->numOfChildren, children[i]->children,
                pbTFLeafColor[i]);
        }
        Ez_DrawWorldSmallDot(pbDraw_window, pbDraw_matrix, rootX, rootY,
            pbTFNodeColor);
    }
}
#endif USE_OLDP_CODE


/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawTreeRegion() --
 *
 * This procedure clears and redraws a tree window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion, even if there is
 *   an error because the iterator should continue on to the next window.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

#ifdef USE_OLDP_CODE
BearStatus
PbDrawTreeRegion(winDataPtr, subWinBndPtr)
WindowData  *winDataPtr;
Rect        *subWinBndPtr;              /* subWinBndPtr contains the world
                                         * coordinates of a portion of the
                                         * window that is to be drawn, if it
                                         * is NULL the whole window is drawn.
                                         */
{
    TreeWin  *treeWinPtr;
    int rootX, rootY;
    int i, j;

    if (PbIsTreeWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    treeWinPtr = PbGetTreeWin(winDataPtr);
    if (treeWinPtr->tree == NIL(Node)) {
        return (BEAR_CONTINUE_SEARCH);
    }

    if (subWinBndPtr == NIL(Rect)) {
        /* 
         * Redraw the whole window.
         */
        XClearWindow(ez_Display, winDataPtr->wd_userWin);
    }
    else {
        /*
         * Redraw part of the window.
         */
/*
 * The X Server always clears an exposed region to the 
 * background color, so we don't have to.
 */
    }
    pbDraw_level = treeWinPtr->level;
    pbDraw_depth = treeWinPtr->depth;
    pbDraw_window = winDataPtr->wd_userWin;
    for (i = 0; i < 4; i++) {
        for (j = 0; j < 4; j++) {
            pbDraw_matrix[i][j] = winDataPtr->wd_scrMatrix[i][j];
        }
    }
    rootX = (treeWinPtr->tree->numOfLeaves - 1) * 50;
    rootY = pbDraw_depth * 220;
    DrawTreeRootAndSubtree(rootX, rootY, (int) 0, 
        treeWinPtr->tree->numOfChildren, treeWinPtr->tree->children,
        pbTFLeafColor[0]);

    Ez_Flush();

    return (BEAR_CONTINUE_SEARCH);
}
#endif USE_OLDP_CODE



#ifdef USE_OLDP_CODE
static void
DrawRootAndSubtreeFloorPlan(transform, dimension, curLevel,
    numOfChildren, children, endColor)

Transform transform;
Point dimension;
int curLevel;
int numOfChildren;
Node *children[MAX_CHILDREN];
int endColor;
{
    Point p1, p2;
    Rect rect;
    int i;

    p1.p_x = 0;
    p1.p_y = 0;
    GeoTransPoint(&transform, &p1, &p2);
    rect.r_xbot = p2.p_x;
    rect.r_ybot = p2.p_y;
    p1.p_x = dimension.p_x;
    p1.p_y = dimension.p_y;
    GeoTransPoint(&transform, &p1, &p2);
    rect.r_xtop = p2.p_x;
    rect.r_ytop = p2.p_y;
    if (rect.r_xbot > rect.r_xtop) {
    rect.r_xtop = rect.r_xbot;
    rect.r_xbot = p2.p_x;
    }
    if (rect.r_ybot > rect.r_ytop) {
    rect.r_ytop = rect.r_ybot;
    rect.r_ybot = p2.p_y;
    }
    if ((curLevel == pbDraw_level) || (numOfChildren == 0)) {
        Ez_FillRect(pbDraw_window, pbDraw_matrix, &rect, endColor);
/*
        Ez_StippleRect(pbDraw_window, pbDraw_matrix, &rect, endColor,
            ez_CargoNetBitmap, GXcopy);
*/
/*
        Ez_OutlineRect(pbDraw_window, pbDraw_matrix, &rect, EZ_SOLIDLINE1,
            colorNode, ez_BackColor);
*/
        Ez_OutlineRect(pbDraw_window, pbDraw_matrix, &rect, EZ_SOLIDLINE1,
            pbTFEdgeColor, ez_BackColor);
    } else {
        curLevel++;
        for (i = 0; i < numOfChildren; i++) {
            DrawRootAndSubtreeFloorPlan(children[i]->transform,
                children[i]->dimension, curLevel, children[i]->numOfChildren,
                children[i]->children, pbTFLeafColor[i]);
        }
/*
        Ez_OutlineRect(pbDraw_window, pbDraw_matrix, &rect, EZ_SOLIDLINE2,
            colorEdge, ez_BackColor);
*/
    }
}
#endif USE_OLDP_CODE


/*
 * ----------------------------------------------------------------------------
 *
 * PbDrawFloorRegion() --
 *
 * This procedure clears and redraws a floorplan window.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned upon completion, even if there is
 *   an error because the iterator should continue on to the next window.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

#ifdef USE_OLDP_CODE
BearStatus
PbDrawFloorRegion(winDataPtr, subWinBndPtr)
WindowData  *winDataPtr;
Rect        *subWinBndPtr;              /* subWinBndPtr contains the world
                                         * coordinates of a portion of the
                                         * window that is to be drawn, if it
                                         * is NULL the whole window is drawn.
                                         */
{
    FloorWin  *floorWinPtr;
    int i, j;
    Point p1, p2;
    Rect rect;

    if (PbIsFloorWin(winDataPtr) == FALSE) {
        return (BEAR_CONTINUE_SEARCH);
    }

    floorWinPtr = PbGetFloorWin(winDataPtr);
    if (floorWinPtr->tree == NIL(Node)) {
        return (BEAR_CONTINUE_SEARCH);
    }

    if (subWinBndPtr == NIL(Rect)) {
        /* 
         * Redraw the whole window.
         */
        XClearWindow(ez_Display, winDataPtr->wd_userWin);
    } else {
        /*
         * Redraw part of the window.
         */
/*
 * The X Server always clears an exposed region to the 
 * background color, so we don't have to.
 */
    }

    pbDraw_level = floorWinPtr->level;
    pbDraw_depth = floorWinPtr->depth;
    pbDraw_window = winDataPtr->wd_userWin;
    for (i = 0; i < 4; i++) {
        for (j = 0; j < 4; j++) {
            pbDraw_matrix[i][j] = winDataPtr->wd_scrMatrix[i][j];
        }
    }

    p1.p_x = 0;
    p1.p_y = 0;
    GeoTransPoint(&floorWinPtr->tree->transform, &p1, &p2);
    rect.r_xbot = p2.p_x - 2;
    rect.r_ybot = p2.p_y;
    p1.p_x = floorWinPtr->tree->dimension.p_x;
    p1.p_y = floorWinPtr->tree->dimension.p_y;
    GeoTransPoint(&floorWinPtr->tree->transform, &p1, &p2);
    rect.r_xtop = p2.p_x;
    rect.r_ytop = p2.p_y + 2;
    Ez_OutlineRect(pbDraw_window, pbDraw_matrix, &rect, EZ_SOLIDLINE2,
        pbTFEdgeColor, ez_BackColor);
    DrawRootAndSubtreeFloorPlan(floorWinPtr->tree->transform,
        floorWinPtr->tree->dimension, (int) 0, floorWinPtr->tree->numOfChildren,
        floorWinPtr->tree->children, pbTFLeafColor[0]);

    Ez_Flush();

    return (BEAR_CONTINUE_SEARCH);
}
#endif USE_OLDP_CODE



BearStatus
PbDupChipWin(oldWinDataPtr, newImageBndPtr, newWinDataPtrPtr)
WindowData  *oldWinDataPtr;
Rect        *newImageBndPtr;
WindowData  **newWinDataPtrPtr;
/*
 * duplicate chip window
 */
{
    CellMaster      *parentCellPtr;
    ChipWin         *oldChipWinPtr, *newChipWinPtr;
    PlaneData       *planeDataPtr;
    PbWindowType    *winTypePtr;
    WindowData      *newWinDataPtr;
    char            prompt[80];

    *newWinDataPtrPtr = NIL(WindowData);
    if (PbIsChipWin(oldWinDataPtr) == FALSE) {
        return (BEAR_ERROR);
    }

    oldChipWinPtr = PbGetChipWin(oldWinDataPtr);
    parentCellPtr = PbGetCellMasterOfCHW(oldChipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    if (GEO_OVERLAP(Tp_GetHorzPlaneBnd(planeDataPtr), newImageBndPtr) ==
        FALSE) {
        return (BEAR_ERROR);
    }

    sprintf (prompt, "BEAR-FP ( %s ) : CHIP", pbHostName);
    if (Ez_CreateWin(prompt, NIL(char),
        oldWinDataPtr->wd_titleBar.tb_textPtr,
        oldWinDataPtr->wd_redrawProc, newImageBndPtr,
        &newWinDataPtr) != EZ_OK) {

        return (BEAR_ERROR);
    }
    newChipWinPtr = PbMakeChipWin();
    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_CHIP_WIN;
    winTypePtr->unValue.chipWinPtr = newChipWinPtr;

    newWinDataPtr->wd_closeWinProc = oldWinDataPtr->wd_closeWinProc;
    newWinDataPtr->wd_userWinButtonProc = oldWinDataPtr->wd_userWinButtonProc;
    newWinDataPtr->wd_clientData = (ClientData) winTypePtr;

    newChipWinPtr->chw_options = oldChipWinPtr->chw_options;
    newChipWinPtr->chw_sharedMemPtr = oldChipWinPtr->chw_sharedMemPtr;

    *newWinDataPtrPtr = newWinDataPtr;
    return (BEAR_OK);
}



BearStatus
PbDupCifWin(oldWinDataPtr, newImageBndPtr, newWinDataPtrPtr)
WindowData  *oldWinDataPtr;
Rect        *newImageBndPtr;
WindowData  **newWinDataPtrPtr;
/*
 * duplicate cif window
 */
{
    CifWin          *oldCifWinPtr, *newCifWinPtr;
    Rect            *cifBndPtr;
    PbWindowType    *winTypePtr;
    WindowData      *newWinDataPtr;
    char            *tmpStringPtr;
    char            prompt[80];

    *newWinDataPtrPtr = NIL(WindowData);
    if (PbIsCifWin(oldWinDataPtr) == FALSE) {
        return (BEAR_ERROR);
    }

    oldCifWinPtr = PbGetCifWin(oldWinDataPtr);
    cifBndPtr = PbGetCifBndOfCW(oldCifWinPtr);
    if (GEO_OVERLAP(cifBndPtr, newImageBndPtr) == FALSE) {
        return (BEAR_ERROR);
    }

    sprintf (prompt, "BEAR-FP ( %s ) : CIF", pbHostName);
    if (Ez_CreateWin(prompt, NIL(char),
        oldWinDataPtr->wd_titleBar.tb_textPtr,
        oldWinDataPtr->wd_redrawProc, newImageBndPtr,
        &newWinDataPtr) != EZ_OK) {

        return (BEAR_ERROR);
    }
    newCifWinPtr = PbMakeCifWin();
    MALLOC(PbWindowType *, winTypePtr, sizeof (PbWindowType));
    winTypePtr->unType = PB_CIF_WIN;
    winTypePtr->unValue.cifWinPtr = newCifWinPtr;

    newWinDataPtr->wd_closeWinProc = oldWinDataPtr->wd_closeWinProc;
    newWinDataPtr->wd_userWinButtonProc = oldWinDataPtr->wd_userWinButtonProc;
    newWinDataPtr->wd_clientData = (ClientData) winTypePtr;

    MALLOC(char *, tmpStringPtr,
        (strlen (oldCifWinPtr->cw_fileNamePtr) + 1) * sizeof (char));
    strcpy (tmpStringPtr, oldCifWinPtr->cw_fileNamePtr);
    PbPutFileNameOfCW(tmpStringPtr, newCifWinPtr);
    PbPutCifBndOfCW(cifBndPtr, newCifWinPtr);

    *newWinDataPtrPtr = newWinDataPtr;
    return (BEAR_OK);
}



void
PbHandleUserWinButton(winDataPtr, button,
    xRootWin, yRootWin, xUserWin, yUserWin)

WindowData  *winDataPtr;
int         button;
int         xRootWin, yRootWin, xUserWin, yUserWin;
{
    return;  /* do nothing, for now */
}



bool
PbIsCifWin(winDataPtr)
WindowData *winDataPtr;
{
    PbWindowType *winTypePtr = (PbWindowType *) winDataPtr->wd_clientData;

    if (winTypePtr == NIL(PbWindowType)) {
        return (FALSE);
    }

    return ((winTypePtr->unType == PB_CIF_WIN) ? TRUE : FALSE);
}



bool
PbIsChipWin(winDataPtr)
WindowData *winDataPtr;
{
    PbWindowType *winTypePtr = (PbWindowType *) winDataPtr->wd_clientData;

    if (winTypePtr == NIL(PbWindowType)) {
        return (FALSE);
    }

    return ((winTypePtr->unType == PB_CHIP_WIN) ? TRUE : FALSE);
}



#ifdef USE_OLDP_CODE
bool
PbIsFloorWin(winDataPtr)
WindowData *winDataPtr;
{
    PbWindowType *winTypePtr = (PbWindowType *) winDataPtr->wd_clientData;

    if (winTypePtr == NIL(PbWindowType)) {
        return (FALSE);
    }

    return ((winTypePtr->unType == PB_FLOOR_WIN) ? TRUE : FALSE);
}
#endif USE_OLDP_CODE



#ifdef USE_OLDP_CODE
bool
PbIsTreeWin(winDataPtr)
WindowData *winDataPtr;
{
    PbWindowType *winTypePtr = (PbWindowType *) winDataPtr->wd_clientData;

    if (winTypePtr == NIL(PbWindowType)) {
        return (FALSE);
    }

    return ((winTypePtr->unType == PB_TREE_WIN) ? TRUE : FALSE);
}
#endif USE_OLDP_CODE



bool
PbIsPlaceWin(winDataPtr)
WindowData *winDataPtr;
{
    PbWindowType *winTypePtr = (PbWindowType *) winDataPtr->wd_clientData;

    if (winTypePtr == NIL(PbWindowType)) {
        return (FALSE);
    }

    return ((winTypePtr->unType == PB_PLACEMENT_WIN) ? TRUE : FALSE);
}



void
PbPutGCellOfCHW(gcellPtr, chipWinPtr)
GenericCell  *gcellPtr;
ChipWin      *chipWinPtr;
{
    GenericCell *newGCellPtr;

    if (gcellPtr == NIL(GenericCell)) {
        FREE((char *) chipWinPtr->chw_sharedMemPtr->gcellPtr);
        chipWinPtr->chw_sharedMemPtr->gcellPtr = NIL(GenericCell);
    }
    else {
        if (chipWinPtr->chw_sharedMemPtr->gcellPtr == NIL(GenericCell)) {
            MALLOC(GenericCell *, newGCellPtr, sizeof (GenericCell));
            chipWinPtr->chw_sharedMemPtr->gcellPtr = newGCellPtr;
        }
        else {
            newGCellPtr = chipWinPtr->chw_sharedMemPtr->gcellPtr;
        }

        newGCellPtr->unType = gcellPtr->unType;
        switch (newGCellPtr->unType) {
        case CELL_INST:
            newGCellPtr->unValue.cellInstPtr = gcellPtr->unValue.cellInstPtr;
            break;
        case DUMMY_CELL:
            newGCellPtr->unValue.dummyCellPtr = gcellPtr->unValue.dummyCellPtr;
            break;
        case ROUTE_CELL:
            newGCellPtr->unValue.routeCellPtr = gcellPtr->unValue.routeCellPtr;
            break;
        default:
        Bear_PrintFatalErr("PbPutGCellOfCHW() failed: unknown cell class");
            break;
        }
    }
}


   
/* 
 * ----------------------------------------------------------------------------
 *
 * PbMakeCifWin() --
 *
 * Allocates memory for a new cif window, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to a cif window is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

CifWin *
PbMakeCifWin()
{
    register CifWin *cifWinPtr;

    MALLOC(CifWin *, cifWinPtr, sizeof (CifWin));
    cifWinPtr->cw_fileNamePtr = NIL(char);

    return (cifWinPtr);
}



/* 
 * ----------------------------------------------------------------------------
 *
 * PbMakeChipWin() --
 *
 * Allocates memory for a new chip window, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to a chip window is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

ChipWin *
PbMakeChipWin()
{
    register ChipWin *chipWinPtr;

    MALLOC(ChipWin *, chipWinPtr, sizeof (ChipWin));
    chipWinPtr->chw_options = PB_NULL_OPT;
    chipWinPtr->chw_sharedMemPtr = NIL(SharedChipWinMem);

    return (chipWinPtr);
}



/* 
 * ----------------------------------------------------------------------------
 *
 * PbMakeFloorWin() --
 *
 * Allocates memory for a new floorplan window, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to a floor window is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

#ifdef USE_OLDP_CODE
FloorWin *
PbMakeFloorWin()
{
    register FloorWin *floorWinPtr;

    MALLOC(FloorWin *, floorWinPtr, sizeof (FloorWin));
    floorWinPtr->level = 0;
    floorWinPtr->tree = (Node *) NULL;
    floorWinPtr->depth = 0;

    return (floorWinPtr);
}
#endif USE_OLDP_CODE



/* 
 * ----------------------------------------------------------------------------
 *
 * PbMakeTreeWin() --
 *
 * Allocates memory for a new tree window, and initializes
 * the fields.
 *
 * Results:
 *   A pointer to a tree window is returned.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

#ifdef USE_OLDP_CODE
TreeWin *
PbMakeTreeWin()
{
    register TreeWin *treeWinPtr;

    MALLOC(TreeWin *, treeWinPtr, sizeof (TreeWin));
    treeWinPtr->level = 0;
    treeWinPtr->tree = (Node *) NULL;
    treeWinPtr->depth = 0;

    return (treeWinPtr);
}
#endif USE_OLDP_CODE



/* 
 * ----------------------------------------------------------------------------
 *
 * PbNumOfChipWindows() --
 *
 * Count up all the chip windows.
 *
 * Results:
 *   BEAR_OK is returned if everything is okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus 
PbNumOfChipWindows(numPtr)
int *numPtr;
{
    int chipWinCount = 0;

    if (PbItrChipWindows(CountChipWindows, (ClientData) &chipWinCount) ==
        BEAR_STOP_SEARCH) {

        Bear_PrintErr("PbNumOfChipWindows() failed");
        *numPtr = 0;
        return (BEAR_ERROR);
    }

    *numPtr = chipWinCount;
    return (BEAR_OK);
}



Rect *
PbGetCifBndOfCW(cifWinPtr)  
CifWin *cifWinPtr;
{
    if (cifWinPtr == NIL(CifWin))
    {
        Bear_PrintFatalErr("PbGetCifBndOfCW() failed: no cif window given");
    }

    return (&cifWinPtr->cw_cifBnd);
}



void
PbPutCifBndOfCW(cifBndPtr, cifWinPtr)  
Rect    *cifBndPtr;
CifWin  *cifWinPtr;
{
    if (cifBndPtr == NIL(Rect))
    {
        Bear_PrintFatalErr("PbPutCifBndOfCW() failed: cif boundary not given");
    }
    else if (cifWinPtr == NIL(CifWin))
    {
        Bear_PrintFatalErr("PbGetCifBndOfCW() failed: cif window not given");
    }

    Mac_CopyRect(&cifWinPtr->cw_cifBnd, cifBndPtr);
}



void
PbPutNetOfCHW(whichNet, netPtr, chipWinPtr)
int         whichNet;
Net         *netPtr;
ChipWin     *chipWinPtr;
{
#ifdef DEBUG
    if (chipWinPtr == NIL(ChipWin)) {
        Bear_PrintFatalErr("PbPutNetOfCHW() failed : no chip window given");
    }
#endif DEBUG

    switch (whichNet) {
    case FPG_NW:
        chipWinPtr->chw_sharedMemPtr->nwNetPtr = netPtr;
        break;
    case FPG_CENTER:
        chipWinPtr->chw_sharedMemPtr->centerNetPtr = netPtr;
        break;
    case FPG_SE:
        chipWinPtr->chw_sharedMemPtr->seNetPtr = netPtr;
        break;
    default:
        Bear_PrintFatalErr("PbPutNetOfCHW() failed : no such net");
        break;
    }
}



Net *
PbGetNetOfCHW(whichNet, chipWinPtr)
int         whichNet;
ChipWin     *chipWinPtr;
{
#ifdef DEBUG
    if (chipWinPtr == NIL(ChipWin)) {
        Bear_PrintFatalErr("PbPutNetOfCHW() failed : no chip window given");
    }
#endif DEBUG

    switch (whichNet) {
    case FPG_NW:
        return (chipWinPtr->chw_sharedMemPtr->nwNetPtr);
    case FPG_CENTER:
        return (chipWinPtr->chw_sharedMemPtr->centerNetPtr);
    case FPG_SE:
        return (chipWinPtr->chw_sharedMemPtr->seNetPtr);
    default:
        Bear_PrintFatalErr("PbGetNetOfCHW() failed : no such net");
        break;
    }
    /* NOTREACHED */
}



Pixel
PbGetColorOfNet(whichNet)
int whichNet;    /* FPG_NW | FPG_CENTER | FPG_SE */
{
    switch (whichNet) {
    case FPG_NW:
        return (pbNet1Color);
    case FPG_CENTER:
        return (pbNet2Color);
    case FPG_SE:
        return (pbNet3Color);
    default:
        Bear_PrintFatalErr("PbGetColorOfNet() failed : no such net");
        break;
    }
    /* NOTREACHED */
}



/*
 * ----------------------------------------------------------------------------
 *
 * PbItrChipWindows() --
 *
 * Applies the users procedure to each chip window.
 *
 * The procedure should be of the following form:
 *
 *  BearStatus
 *  userProc(winDataPtr, clientData)
 *  WindowData  *winDataPtr;
 *  ClientData  clientData;
 *  {
 *  }
 *
 * userProc normally should return BEAR_CONTINUE_SEARCH.  If it returns 
 * BEAR_STOP_SEARCH then the search will be aborted.
 *
 * Results:
 *   BEAR_COMPLETE_SEARCH is returned if the search completed normally.
 *   BEAR_STOP_SEARCH is returned if it stopped.
 *
 * Side effects:
 *   Whatever side effects result from application of the
 *   supplied procedure.
 * ----------------------------------------------------------------------------
 */

BearStatus
PbItrChipWindows(userProc, clientData)
BearStatus  (*userProc)();
ClientData  clientData;
{
    ApplyData applyData;

    applyData.idWinProc = PbIsChipWin;
    applyData.applyProc = userProc;
    applyData.clientData = clientData;
    if (Ez_ItrWindows(ApplyProcToWin, (ClientData) &applyData) == 
        EZ_STOP_SEARCH) {
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_COMPLETE_SEARCH);
}



bool
PbHorzTilesOn(winDataPtr)
WindowData *winDataPtr;
{
    ChipWin *chipWinPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    if (((chipWinPtr->chw_options & PB_SHOW_HORZ_PLANE_OPT) == FALSE) ||
        (chipWinPtr->chw_options & PB_SHOW_VERT_PLANE_OPT)) {
        /*
         * Adjust the chip windows option flag so that the next time
         * it's redraw only the horizontal tile plane will be shown.
         */
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_HORZ_PLANE_OPT);
        Bear_UnsetOption(chipWinPtr->chw_options, PB_SHOW_VERT_PLANE_OPT);

        return (TRUE); /* Needs to be redrawn. */
    }
    return (FALSE); /* Does not need to be redrawn. */
}



bool
PbVertTilesOn(winDataPtr)
WindowData *winDataPtr;
{
    ChipWin *chipWinPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    if ((chipWinPtr->chw_options & PB_SHOW_HORZ_PLANE_OPT) ||
        ((chipWinPtr->chw_options & PB_SHOW_VERT_PLANE_OPT) == FALSE)) {
        /*
         * Adjust the chip windows option flag so that the next time
         * it's redraw only the vertical tile plane will be shown.
         */
        Bear_UnsetOption(chipWinPtr->chw_options, PB_SHOW_HORZ_PLANE_OPT);
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_VERT_PLANE_OPT);

        return (TRUE); /* Needs to be redrawn. */
    }
    return (FALSE); /* Does not need to be redrawn. */
}



bool
PbBothTilesOn(winDataPtr)
WindowData *winDataPtr;
{
    ChipWin *chipWinPtr;

    chipWinPtr = PbGetChipWin(winDataPtr);
    if (((chipWinPtr->chw_options & PB_SHOW_HORZ_PLANE_OPT) == FALSE) ||
        ((chipWinPtr->chw_options & PB_SHOW_VERT_PLANE_OPT) == FALSE)) {
        /*
         * Adjust the chip windows option flag so that the next time
         * it's redraw both tile planes will be shown.
         */
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_HORZ_PLANE_OPT);
        Bear_SetOption(chipWinPtr->chw_options, PB_SHOW_VERT_PLANE_OPT);

        return (TRUE); /* Needs to be redrawn. */
    }
    return (FALSE); /* Does not need to be redrawn. */
}



void
PbCenterView(winDataPtr, objBndPtr)
WindowData  *winDataPtr;
Rect        *objBndPtr;
{
    Point   wPoint;                 /* world coordinate point   */
    Rect    *imageBndPtr;
    int     deltaX, deltaY;

    /*
     * Find the center of the objects boundary.
     */
    wPoint.p_x = objBndPtr->r_xbot +
        ((objBndPtr->r_xtop - objBndPtr->r_xbot) / 2);
    wPoint.p_y = objBndPtr->r_ybot +
        ((objBndPtr->r_ytop - objBndPtr->r_ybot) / 2);

    /*
     * Now center the image boundary (whats in view) around this point.
     */
    imageBndPtr = &winDataPtr->wd_imageBnd;
    deltaX = (imageBndPtr->r_xtop - imageBndPtr->r_xbot) / 2;
    deltaY = (imageBndPtr->r_ytop - imageBndPtr->r_ybot) / 2;
    imageBndPtr->r_xbot = wPoint.p_x - deltaX;
    imageBndPtr->r_xtop = wPoint.p_x + deltaX;
    imageBndPtr->r_ybot = wPoint.p_y - deltaY;
    imageBndPtr->r_ytop = wPoint.p_y + deltaY;
    (void) Ez_CalcViewingTrans(winDataPtr);
}


void
PbViewEntireObject(winDataPtr, objBndPtr, forceFlag)
WindowData  *winDataPtr;
Rect        *objBndPtr;
bool        forceFlag;
{
    Rect *imageBndPtr;

    imageBndPtr = &winDataPtr->wd_imageBnd;

    if ((forceFlag == TRUE) ||
        (GEO_SURROUND(imageBndPtr, objBndPtr) == FALSE)) {

        imageBndPtr->r_xbot = objBndPtr->r_xbot;
        imageBndPtr->r_xtop = objBndPtr->r_xtop;
        imageBndPtr->r_ybot = objBndPtr->r_ybot;
        imageBndPtr->r_ytop = objBndPtr->r_ytop;
        (void) Ez_CalcViewingTrans(winDataPtr);
    }
}


BearStatus
PbGetSquare(winDataPtrPtr, screenPtr, squarePtr)
WindowData  **winDataPtrPtr;
Point       *screenPtr;
Rect        *squarePtr;
{
    ezStatus        findStatus;
    Pb_DefCurData   defCurData;
    WindowData      *winDataPtr;

    *winDataPtrPtr = NIL(WindowData);

    /*
     * Change the cursor to a square in all of the chip windows.
     */
    defCurData.idWinProc = PbIsChipWin;
    defCurData.cursorType = (int) EZ_SQUARE_OUTLINE_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();

    /*
     * Wait for the user to click the mouse in a chip window, which
     * will indicate the location of the square.
     */
    findStatus = Ez_FindUserWin(LeftButton, PbIsChipWin, (ClientData) NULL,
        &winDataPtr, screenPtr);

    /*
     * Change the cursor back to a cross in all of the chip windows.
     */
    defCurData.cursorType = (int) EZ_CROSS_CUR;
    (void) PbItrChipWindows(PbDefCurInUserSubWin, (ClientData) &defCurData);
    Ez_Flush();

    if (findStatus != EZ_OK) {
        PbPrint(EZ_ITALIC_FONT, "must pick a point in a chip window\n");
        return (BEAR_ERROR);
    }

    PbCalcSquare(winDataPtr, screenPtr, squarePtr);

    *winDataPtrPtr = winDataPtr;

    return (BEAR_OK);
}



void
PbCalcSquare(winDataPtr, screenPtr, wSquarePtr)
WindowData  *winDataPtr;
Point       *screenPtr;     /* Point in screen coordinates. */
Rect        *wSquarePtr;    /* Square in world coordinates. */
/*
 * Calculate the world coordinates of the lower left and upper right
 * corners of a square that is sixteen pixels wide, centered
 * yabba dabba doo about the point thats picked.
 */
{
    Point           world;
    register double (*invScrMatrix)[4];
    int             halfWldPinWidth;    /* Half the width of a pin in world
                                         * coordinates.
                                         */

    invScrMatrix = winDataPtr->wd_invScrMatrix;
    halfWldPinWidth = (Ez_XTrans(invScrMatrix, 16, 0) -
        Ez_XTrans(invScrMatrix, 0, 0)) / 2;
    halfWldPinWidth = ABS(halfWldPinWidth);

    world.p_x = Ez_XTrans(invScrMatrix, screenPtr->p_x, screenPtr->p_y);
    world.p_y = Ez_YTrans(invScrMatrix, screenPtr->p_x, screenPtr->p_y);
    wSquarePtr->r_xbot = world.p_x - halfWldPinWidth;
    wSquarePtr->r_ybot = world.p_y - halfWldPinWidth;
    wSquarePtr->r_xtop = world.p_x + halfWldPinWidth;
    wSquarePtr->r_ytop = world.p_y + halfWldPinWidth;

#ifdef DEBUG
    if (pbDebugFlag == TRUE) {
        int x, y, height, width;

        Ez_FillRect(winDataPtr->wd_userWin, winDataPtr->wd_scrMatrix,
            wSquarePtr, ez_HighLightColor);

        x = wSquarePtr->r_xbot;
        y = wSquarePtr->r_ytop;
        width = wSquarePtr->r_xtop - x;
        height = y - wSquarePtr->r_ybot;
        printf ("world point x = %d y = %d\n", world.p_x, world.p_y);
        printf ("world square x = %d y = %d width = %d height = %d\n",
            x, y, width, height);
        fflush (stdout);
    }
#endif DEBUG
}



static BearStatus
BnLineCrossesBox(bnLinePtr, bnLineSearchPtr)
BNLine          *bnLinePtr;
BnLineSearch    *bnLineSearchPtr;
{
    int     a, endPoint1, endPoint2;
    bool    horzBnLine;

    if (bnLinePtr->line.point1.p_y == bnLinePtr->line.point2.p_y) {
        /* horizontal bottleneck line */
        horzBnLine = TRUE;
        endPoint1 = bnLinePtr->line.point1.p_x;
        endPoint2 = bnLinePtr->line.point2.p_x;
    } else {
        /* vertical bottleneck line */
        horzBnLine = FALSE;
        endPoint1 = bnLinePtr->line.point1.p_y;
        endPoint2 = bnLinePtr->line.point2.p_y;
    }
    if (endPoint1 > endPoint2) {
        /* switch them */
        a = endPoint1;
        endPoint1 = endPoint2;
        endPoint2 = a;
    }

    if (horzBnLine) {
        if ((bnLinePtr->line.point1.p_y < bnLineSearchPtr->box.r_ytop) &&
            (bnLinePtr->line.point1.p_y > bnLineSearchPtr->box.r_ybot)) {

            if ((endPoint1 < bnLineSearchPtr->box.r_xbot) &&
                (endPoint2 > bnLineSearchPtr->box.r_xbot)) {
                bnLineSearchPtr->bnLinePtr = bnLinePtr;
                return (BEAR_STOP_SEARCH);
            }
            if ((endPoint1 < bnLineSearchPtr->box.r_xtop) &&
                (endPoint2 > bnLineSearchPtr->box.r_xtop)) {
                bnLineSearchPtr->bnLinePtr = bnLinePtr;
                return (BEAR_STOP_SEARCH);
            }
        }
    } else {
        if ((bnLinePtr->line.point1.p_x < bnLineSearchPtr->box.r_xtop) &&
            (bnLinePtr->line.point1.p_x > bnLineSearchPtr->box.r_xbot)) {

            if ((endPoint1 < bnLineSearchPtr->box.r_ybot) &&
                (endPoint2 > bnLineSearchPtr->box.r_ybot)) {
                bnLineSearchPtr->bnLinePtr = bnLinePtr;
                return (BEAR_STOP_SEARCH);
            }
            if ((endPoint1 < bnLineSearchPtr->box.r_ytop) &&
                (endPoint2 > bnLineSearchPtr->box.r_ytop)) {
                bnLineSearchPtr->bnLinePtr = bnLinePtr;
                return (BEAR_STOP_SEARCH);
            }
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}



BearStatus
PbPickBnLine(winDataPtr, bnLinePtrPtr)
WindowData  *winDataPtr;
BNLine      **bnLinePtrPtr;
{
    BnLineSearch    bnLineSearch;
    CellMaster      *cellMasterPtr;
    ChipWin         *chipWinPtr;
    PlaneData       *planeDataPtr;
    Point           screen;

    *bnLinePtrPtr = NIL(BNLine);

    if (PbGetSquare(&winDataPtr, &screen, &bnLineSearch.box) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    chipWinPtr = PbGetChipWin(winDataPtr);
    cellMasterPtr = PbGetCellMasterOfCHW(chipWinPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);

    bnLineSearch.bnLinePtr = NIL(BNLine);
    (void) Tp_ItrBnLine(Tp_GetHorzBnLineList(planeDataPtr),
        BnLineCrossesBox, &bnLineSearch);
    if (bnLineSearch.bnLinePtr == NIL(BNLine)) {
        (void) Tp_ItrBnLine(Tp_GetVertBnLineList(planeDataPtr),
            BnLineCrossesBox, &bnLineSearch);
    }
    if (bnLineSearch.bnLinePtr == NIL(BNLine)) {
        PbPrint(EZ_ITALIC_FONT, "Can't find a bottleneck line there.\n");
        return (BEAR_ERROR);
    } 

    PbPutBnLineOfCHW(bnLineSearch.bnLinePtr, chipWinPtr);
    (void) PbItrChipWindows(PbDrawChipRegion, (ClientData) NULL);

    *bnLinePtrPtr = bnLineSearch.bnLinePtr;

    return (BEAR_OK);
}

#endif /* BEAR_BLACK_BOX */


