
/* pbWriteChip.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbWriteChip.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"

#ifdef USE_NEWP_CODE
#include "newC.h"
#endif /* USE_NEWP_CODE */

#include "ezPlot.h"
#include "bfpInterface.h"

#ifdef USE_BBL_CODE
#include "bblInterface.h"
#endif /* USE_BBL_CODE */

#ifdef USE_OCT_CODE
#include "octInterface.h"
#include "vov.h"
#endif /* USE_OCT_CODE */

#include "magInterface.h"
#include "polarBear.h"


extern int Lr_OutputChannel(/*modulePtr,cifFilePtr,scale*/);


typedef struct _WriteData WriteData;
struct _WriteData {
    char        *fileNamePtr;
    FILE        *filePtr;
    CellMaster  *cellMasterPtr;
    CellInst    *cellInstPtr;
    Point       *originPtr;
    LinkedPoint *pointList;
    double      scale;
    double      xScale, yScale;     /* bbl format only */
    bool        scaleFlag;          /* bbl format only : TRUE if both xScale
                                     * and yScale do NOT equal one.
                                     */
    bool        pinFlag;
    bool        floatingPinFlag;    /* bbl format only */
    bool        mirrorYAxisFlag;    /* bbl format only */
    bool        rootModFlag;        /* bbl format only */

    bool        mixFlag;            /* bbl format only */
    lsList      ciList;             /* bbl format only */
};


/* ---------- Local Procedures ---------- */
#ifdef USE_BBL_CODE
static BearStatus WriteChipInBBLFormat();
static lsStatus   WriteCellInBBLFormat();
static BearStatus WritePinInBBLFormat();
#endif /* USE_BBL_CODE */

static BearStatus WriteChipInCifFormat();
static BearStatus WriteCellInCifFormat();
static BearStatus WritePinInCifFormat();

static BearStatus WriteChipTransform();
static BearStatus WriteCellTransform();

static BearStatus AddPointToPointList();
static BearStatus WritePointInBBLFormat();
static BearStatus WritePointOfLineSegInBBLFormat();

#ifdef USE_DR_CODE
static BearStatus WriteChipWireInMagFormat();
static BearStatus WriteRCellInCifFormat();
static BearStatus WriteRCellInMagFormat();
#endif /* USE_DR_CODE */


#ifdef USE_BBL_CODE
static BearStatus
WriteChipInBBLFormat(writeDataPtr)
WriteData *writeDataPtr;
{
    FILE    *filePtr;
    lsList  ciList;
    int     numOfNets;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, writeDataPtr->fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (writeDataPtr->fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open BBL file named \"%s\" for writing.\n", 
            writeDataPtr->fileNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        return (BEAR_ERROR);
    }
    writeDataPtr->filePtr = filePtr;

    if (Mac_NumOfNets(writeDataPtr->cellMasterPtr, &numOfNets) != BEAR_OK) {
        return (BEAR_ERROR);
    }
    if (writeDataPtr->pinFlag == TRUE) {
        (void) fprintf (filePtr, "SN %d\n", numOfNets);
    }
    else {
        (void) fprintf (filePtr, "SN 0\n");
    }

    /*
     * Write out the root module into the file.
     */
    writeDataPtr->rootModFlag = TRUE;
    if (WriteCellInBBLFormat(writeDataPtr->cellMasterPtr->instList,
        writeDataPtr) != LS_OK) {

        return (BEAR_ERROR);
    }

    /*
     * Write out all the children cells into the file in alphabetical
     * order by cell instance name.
     */
    writeDataPtr->rootModFlag = FALSE;
    ciList = Mac_CreateCellInstList(writeDataPtr->cellMasterPtr);
    Mac_SortCellInstListByName(ciList);
    if (lsForeach(ciList, WriteCellInBBLFormat, (lsGeneric) writeDataPtr) ==
        LS_STOP) {
        return (BEAR_ERROR);
    }
    lsDestroy(ciList, (void (*)()) NULL);

    (void) fprintf (filePtr, "$\n");

    (void) fclose (filePtr);
    writeDataPtr->filePtr = NIL(FILE);

    return (BEAR_OK);
}
#endif /* USE_BBL_CODE */



static BearStatus
WriteChipInCifFormat(writeDataPtr)
WriteData *writeDataPtr;
{
    FILE        *filePtr;
    PlaneData   *planeDataPtr;
    Rect        *bboxPtr, printBBox;
    double      scale;
    Ring        *ringPtr;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, writeDataPtr->fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (writeDataPtr->fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open CIF file named \"%s\" for writing.\n", 
            writeDataPtr->fileNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        return (BEAR_ERROR);
    }
    writeDataPtr->filePtr = filePtr;

    fprintf (filePtr, "( Silly Comment !);\n");
    fprintf (filePtr, "L BND1;\n");

    planeDataPtr = Mac_GetCellPlaneOfCell(writeDataPtr->cellMasterPtr);
    bboxPtr = Tp_GetHorzPlaneBnd(planeDataPtr);

    scale = writeDataPtr->scale;
    printBBox.r_xbot = bboxPtr->r_xbot * scale;
    printBBox.r_xtop = bboxPtr->r_xtop * scale;
    printBBox.r_ybot = bboxPtr->r_ybot * scale;
    printBBox.r_ytop = bboxPtr->r_ytop * scale;

    fprintf (filePtr, "P \n");
    fprintf (filePtr, "    %d %d\n", printBBox.r_xbot, printBBox.r_ybot);
    fprintf (filePtr, "    %d %d\n", printBBox.r_xtop, printBBox.r_ybot);
    fprintf (filePtr, "    %d %d\n", printBBox.r_xtop, printBBox.r_ytop);
    fprintf (filePtr, "    %d %d;\n", printBBox.r_xbot, printBBox.r_ytop);
    fprintf (filePtr, "L BND0;\n");

/* Temp */
writeDataPtr->pinFlag = FALSE;

    if (writeDataPtr->pinFlag) {
        /*
         * Go through each pin of the core cell.
         */
        (void) Mac_ItrPinOfCellInst(writeDataPtr->cellMasterPtr->instList,
            WritePinInCifFormat, (ClientData) writeDataPtr);
    }

    Mac_ItrChildrenOfCell(writeDataPtr->cellMasterPtr,
        WriteCellInCifFormat, (ClientData) writeDataPtr);

#ifdef USE_DR_CODE
    Mac_ItrRCellOfCell(writeDataPtr->cellMasterPtr,
        WriteRCellInCifFormat, (ClientData) writeDataPtr);

    ringPtr = writeDataPtr->cellMasterPtr->contents.ringPtr;
    if (ringPtr != NIL(Ring)) {
        Lr_OutputChannel(ringPtr->routeModPtr, writeDataPtr->filePtr,
            (int) writeDataPtr->scale, Mac_GetTransformOfRing(ringPtr));
    }
#endif /* USE_DR_CODE */

    fprintf (filePtr, "E\n");

    fclose (filePtr);
    writeDataPtr->filePtr = NIL(FILE);

    return (BEAR_OK);
}



#ifdef USE_DR_CODE
static BearStatus
WriteChipWireInMagFormat(writeDataPtr)
WriteData *writeDataPtr;
/* writeDataPtr->filePtr must hold complete name for new magic file */
{
    FILE    *filePtr, *magFilePtr;
    Rect    *bboxPtr, printBBox;
    double  scale;
    Ring    *ringPtr;
    char    *magFileNamePtr = NIL(char);
    char    buffer[256], word[256];

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, writeDataPtr->fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (writeDataPtr->fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open file named \"%s\" for writing.\n", 
            writeDataPtr->fileNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        return (BEAR_ERROR);
    }
    writeDataPtr->filePtr = filePtr;

    MALLOC(char *, magFileNamePtr, 
        (strlen(writeDataPtr->fileNamePtr) + 5) * sizeof (char));
    strcpy(magFileNamePtr, writeDataPtr->fileNamePtr);
    removeExt(magFileNamePtr);
    addExt(magFileNamePtr, ".mag");

#ifdef USE_OCT_CODE
    (void) VOVinput(VOV_UNIX_FILE, magFileNamePtr);
#endif /* USE_OCT_CODE */

    if ((magFilePtr = fopen (magFileNamePtr, "r")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open file named \"%s\" for reading.\n", 
            magFileNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);

        FREE((char *) magFileNamePtr);
        return (BEAR_ERROR);
    }

    fgets(buffer,256,magFilePtr);
    sscanf(buffer,"%s",word);

    /* should add a magic file format checker for the first three lines */
    if (strcmp("magic",word)==0) {
        fprintf(filePtr,buffer);

        fgets(buffer,256,magFilePtr);
        fprintf(filePtr,buffer);

        fgets(buffer,256,magFilePtr);
        fprintf(filePtr,buffer);
    }
    else {
        (void) sprintf (bear_ErrorMsg,
            "Error in magic file: \"%s\". ", magFileNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        
        FREE((char *) magFileNamePtr);
        return (BEAR_ERROR);
    }

    fprintf (filePtr, "<< checkpaint >>\n");
    bboxPtr = Mac_GetBBoxOfCellMaster(writeDataPtr->cellMasterPtr);
    scale = writeDataPtr->scale;
    printBBox.r_xbot = bboxPtr->r_xbot * scale;
    printBBox.r_xtop = bboxPtr->r_xtop * scale;
    printBBox.r_ybot = bboxPtr->r_ybot * scale;
    printBBox.r_ytop = bboxPtr->r_ytop * scale;
   
    fprintf (filePtr, "rect %d %d %d %d\n", 
        printBBox.r_xbot, printBBox.r_ybot,
        printBBox.r_xtop, printBBox.r_ytop);

    Mac_ItrRCellOfCell(writeDataPtr->cellMasterPtr,
        WriteRCellInMagFormat, (ClientData) writeDataPtr);

    ringPtr = writeDataPtr->cellMasterPtr->contents.ringPtr;
    if (ringPtr != NIL(Ring)) {
        Mag_OutputMagicWire(ringPtr->routeModPtr,
             writeDataPtr->filePtr,
             (int) writeDataPtr->scale, 
             Mac_GetTransformOfRing(ringPtr));
    }

    Mag_WriteMagic(filePtr, magFilePtr);

    fclose(magFilePtr);
    fclose (filePtr);
    writeDataPtr->filePtr = NIL(FILE);
    FREE((char *) magFileNamePtr);

    return (BEAR_OK);
}
#endif /* USE_DR_CODE */



static BearStatus
WriteChipTransform(writeDataPtr)
WriteData *writeDataPtr;
{
    FILE *filePtr;

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, writeDataPtr->fileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (writeDataPtr->fileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "Could not open file named \"%s\" for writing.\n", 
            writeDataPtr->fileNamePtr);
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        return (BEAR_ERROR);
    }
    writeDataPtr->filePtr = filePtr;

    fprintf (filePtr, "Transform file\n\n");

    /*
     * Write out all modules into the file.
     */
    Mac_ItrChildrenOfCell(writeDataPtr->cellMasterPtr,
        WriteCellTransform, (ClientData) writeDataPtr);

    fprintf (filePtr, "$\n");

    fclose (filePtr);
    writeDataPtr->filePtr = NIL(FILE);

    return (BEAR_OK);
}



#ifdef USE_BBL_CODE
static lsStatus
WriteCellInBBLFormat(item, arg)
lsGeneric   item;
lsGeneric   arg;
{
    FILE            *filePtr;
    BearStatus      bndCW;
    CellInst        *cellInstPtr;
    LinkedLineSeg   *boundaryList;
    Transform       cellTrans;
    WriteData       *writeDataPtr;
    int             x, y;

    cellInstPtr = (CellInst *) item;
    writeDataPtr = (WriteData *) arg;

    filePtr = writeDataPtr->filePtr;

    fprintf (filePtr, "MOD\n");
    writeDataPtr->originPtr = Mac_GetOriginOfCell(cellInstPtr);

    if (writeDataPtr->scaleFlag) {
        x = writeDataPtr->originPtr->p_x * writeDataPtr->xScale;
        y = writeDataPtr->originPtr->p_y * writeDataPtr->yScale;
    }
    else {
        x = writeDataPtr->originPtr->p_x;
        y = writeDataPtr->originPtr->p_y;
    }
    fprintf (filePtr, "%d %d\n", x, y);

    if (writeDataPtr->rootModFlag == TRUE) {
        fprintf (filePtr, "bound\n");
        fprintf (filePtr, "1\n");
    }
    else {
        fprintf (filePtr, "%s\n", Mac_GetNameOfCellInst(cellInstPtr));
        fprintf (filePtr, "0\n");
    }

    /*
     * Note that bbl expects the points to be in counter clockwise order.
     */
    boundaryList = Mac_GetBoundaryOfCell(cellInstPtr);
    bndCW = Mac_IsBoundaryCW(boundaryList);

    if (((bndCW == BEAR_TRUE) && !writeDataPtr->mirrorYAxisFlag) ||
        ((bndCW == BEAR_FALSE) && writeDataPtr->mirrorYAxisFlag)) {

        /* boundary is clockwise */
        writeDataPtr->pointList = NIL(LinkedPoint);
        (void) Mac_ItrLineSeg(boundaryList, AddPointToPointList,
            (ClientData) writeDataPtr);
        (void) Mac_ItrLinkedPoint(writeDataPtr->pointList,
            WritePointInBBLFormat, (ClientData) writeDataPtr);
        Mac_FreePointList(writeDataPtr->pointList);
    }
    else {
        /* boundary is counter clockwise */
        (void) Mac_ItrLineSeg(boundaryList, WritePointOfLineSegInBBLFormat,
            (ClientData) writeDataPtr);
    }

    fprintf (filePtr, "$\n");
    fprintf (filePtr, "T\n");

    if (writeDataPtr->pinFlag) {
        /*
         * Go through each pin of the cell instance.
         */
        writeDataPtr->cellInstPtr = cellInstPtr;
        if (Mac_ItrPinOfCellInst(cellInstPtr, WritePinInBBLFormat,
            (ClientData) writeDataPtr) == BEAR_STOP_SEARCH) {

            return (LS_STOP);
        }
    }
    fprintf (filePtr, "$\n");

    return (LS_OK);
}
#endif /* USE_BBL_CODE */



static BearStatus
WriteCellTransform(cellInstPtr, writeDataPtr)
CellInst    *cellInstPtr;
WriteData   *writeDataPtr;
{
    FILE        *filePtr;
    Transform   cellTrans;

    filePtr = writeDataPtr->filePtr;

    fprintf (filePtr, "MOD %s\n", Mac_GetNameOfCellInst(cellInstPtr));

    cellTrans = cellInstPtr->transform;
    fprintf (filePtr, "transform:\n");
    fprintf (filePtr, "    %8d  %8d\n", cellTrans.t_a, cellTrans.t_d);
    fprintf (filePtr, "    %8d  %8d\n", cellTrans.t_b, cellTrans.t_e);
    fprintf (filePtr, "    %8d  %8d\n", cellTrans.t_c, cellTrans.t_f);
    fprintf (filePtr, "\n");

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
WriteCellInCifFormat(cellInstPtr, writeDataPtr)
CellInst    *cellInstPtr;
WriteData   *writeDataPtr;
/*
 * Write a cif box command for each rectangle in the cell's
 * horizontal rectangle list.  And write a cif name command.
 */
{
    FILE        *filePtr;
    GenericCell gcell;
    LinkedRect  *lrPtr;
    Rect        horzRect;
    Transform   *transPtr;
    double      scale;
    int         width, height, xCenter, yCenter;

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;
    transPtr = Mac_GetTransformOfGCell(&gcell);
    filePtr =  writeDataPtr->filePtr;
    scale =  writeDataPtr->scale;
    for (lrPtr = Mac_GetHorzRectListOfGCell(&gcell);
        lrPtr != NIL(LinkedRect);
        lrPtr = lrPtr->r_next) {

        GeoTransRect(transPtr, &lrPtr->r_r, &horzRect);
        horzRect.r_xbot *= scale;
        horzRect.r_xtop *= scale;
        horzRect.r_ybot *= scale;
        horzRect.r_ytop *= scale;

        width = horzRect.r_xtop - horzRect.r_xbot;
        height = horzRect.r_ytop - horzRect.r_ybot;
        xCenter = horzRect.r_xbot + width / 2;
        yCenter = horzRect.r_ybot + height / 2;
        fprintf (filePtr, "B %d %d %d %d;\n", width, height,
            xCenter, yCenter);
    }
    fprintf (filePtr, "94 %s %d %d;\n", Mac_GetNameOfCellInst(cellInstPtr),
        xCenter, yCenter);

    if (writeDataPtr->pinFlag) {
        /*
         * Go through each pin of the cell instance.
         */
        (void) Mac_ItrPinOfCellInst(cellInstPtr, WritePinInCifFormat,
            (ClientData) writeDataPtr);
    }

    return (BEAR_CONTINUE_SEARCH);
}



#ifdef USE_DR_CODE
static BearStatus
WriteRCellInMagFormat(routeCellPtr, writeDataPtr)
RouteCell *routeCellPtr;
WriteData *writeDataPtr;
{
    Mag_OutputMagicWire(routeCellPtr->routeModPtr, writeDataPtr->filePtr,
        (int) writeDataPtr->scale, Mac_GetTransformOfRCell(routeCellPtr));
    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */



#ifdef USE_DR_CODE
static BearStatus
WriteRCellInCifFormat(routeCellPtr, writeDataPtr)
RouteCell *routeCellPtr;
WriteData *writeDataPtr;
{
    Lr_OutputChannel(routeCellPtr->routeModPtr, writeDataPtr->filePtr,
        (int) writeDataPtr->scale, Mac_GetTransformOfRCell(routeCellPtr));
    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */



#ifdef USE_BBL_CODE
static BearStatus
WritePinInBBLFormat(pinInstPtr, writeDataPtr)
PinInst     *pinInstPtr;
WriteData   *writeDataPtr;
/*
 * Write out the pin's relative position in BBL format
 */
{
    FILE        *filePtr;
    LayerVector pinLayer;   
    Net         *netPtr;
    NetKind     netKind;
    Point       absPos, relPos, printRelPos, *originPtr;
    PortInst    *portInstPtr;
    int         bblPinDir;          /* bbl pin direction */

    filePtr = writeDataPtr->filePtr;

    if (Mac_IsPinPosDef(pinInstPtr) == TRUE) {
        /*
         * Note that you must start with the absolute pin position
         * (by calling Mac_GetAbsPosOfPin()) to take into account
         * the cells transformation matrix.
         */
        Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
        originPtr = Mac_GetOriginOfCell(writeDataPtr->cellInstPtr);
        relPos.p_x = absPos.p_x - originPtr->p_x;
        relPos.p_y = absPos.p_y - originPtr->p_y;

        if (writeDataPtr->scaleFlag) {
            printRelPos.p_x = relPos.p_x * writeDataPtr->xScale;
            printRelPos.p_y = relPos.p_y * writeDataPtr->yScale;
        }
        else {
            printRelPos.p_x = relPos.p_x;
            printRelPos.p_y = relPos.p_y;
        }
        fprintf (filePtr, "%8d %8d ", printRelPos.p_x, printRelPos.p_y);
    }
    else {
        fprintf (filePtr, "%8s %8s ", "undef", "undef");
    }

    netPtr = Mac_GetNetOfPin(pinInstPtr);
    if (netPtr == NIL(Net)) {
        (void) sprintf (bear_ErrorMsg,
            "Pin named %s is not connected to a net.\n",
            Mac_GetNameOfPinInst(pinInstPtr));
        PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        return (BEAR_STOP_SEARCH);
    }
    else {
        fprintf (filePtr, "%8s ", Mac_GetNameOfNet(netPtr));
    }

    /*
     * The BBL pin direction is indicated as follows:
     *     0        means west
     *     1        means south
     *     2        means east
     *     3        means north
     *     undef    means undefined
     */
    if (Mac_IsPinPosDef(pinInstPtr) == TRUE) {
        switch (Mac_GetDirectionOfPin(pinInstPtr)) {
        case (PinDirection) WEST_PIN:
            bblPinDir = (writeDataPtr->mirrorYAxisFlag) ? 2 : 0;
            break;
        case (PinDirection) SOUTH_PIN:
            bblPinDir = 1;
            break;
        case (PinDirection) EAST_PIN:
            bblPinDir = (writeDataPtr->mirrorYAxisFlag) ? 0 : 2;
            break;
        case (PinDirection) NORTH_PIN:
            bblPinDir = 3;
            break;
        default:
          Bear_PrintErr("WritePinInBBLFormat() failed : illegal pin direction");
            return (BEAR_STOP_SEARCH);
        }
        fprintf (filePtr, "%8d ", bblPinDir);
    }
    else {
        fprintf (filePtr, "%8s ", "undef");
    }

    /*
     * The BBL pin type is indicated as follows:
     *     2 means fixed
     *     3 means flexible
     *     4 means flexible (north)
     *     5 means flexible (east)
     *     6 means flexible (south)
     *     7 means flexible (west)
     */

    if (writeDataPtr->rootModFlag == TRUE) {
        fprintf (filePtr, "%d ", 2);
    }
    else if (writeDataPtr->mixFlag == TRUE) {
        CellInst    *cellInstPtr;
        lsGen       gen;
        bool        fixedFlag;

        fixedFlag = TRUE;
        lsForeachItem(writeDataPtr->ciList, gen, cellInstPtr) {
            if (writeDataPtr->cellInstPtr == cellInstPtr) {
                fixedFlag = FALSE;
                lsFinish(gen);
                break;
            }
        }
        fprintf (filePtr, "%d ", (fixedFlag == TRUE) ? 2 : 3);
    }
    else if (writeDataPtr->floatingPinFlag == TRUE) {
        fprintf (filePtr, "%d ", 3);
    }
    else {
        switch (Mac_GetFlexOfPin(pinInstPtr)) {
        case (Flexibility) FIXED:
            fprintf (filePtr, "2 ");        /* fixed pin type               */
            break;
        case (Flexibility) FLEXIBLE:
            fprintf (filePtr, "3 ");        /* floating pin type            */
            break;
        case (Flexibility) FLEXIBLE_NORTH:
            fprintf (filePtr, "4 ");
            break;
        case (Flexibility) FLEXIBLE_EAST:
            fprintf (filePtr, "5 ");
            break;
        case (Flexibility) FLEXIBLE_SOUTH:
            fprintf (filePtr, "6 ");
            break;
        case (Flexibility) FLEXIBLE_WEST:
            fprintf (filePtr, "7 ");
            break;
        default:
       Bear_PrintErr("WritePinInBBLFormat() failed : illegal flexibility type");
            break;
        }
    }

/*
 *  taken out by george
 *
 *  pinLayer = Mac_GetLayerOfPin(pinInstPtr);
 *  netKind = Mac_GetKindOfNet(netPtr);
 *
 *  if ((netKind != SIGNAL_NET) || (pinLayer != 12)) {
 *      fprintf (filePtr, "%d ", pinLayer);
 *  }
 *
 *  if (netKind != SIGNAL_NET) {
 *      fprintf (filePtr, "%d ", (int) pinInstPtr->widthVect[1]); 
 *      fprintf (filePtr, "%d ", (int) pinInstPtr->widthVect[2]); 
 *  }
 *
 *  if (netPtr != NIL(Net)) {
 *      switch (netKind) {
 *      case (NetKind) DUMMY_NET:
 *      case (NetKind) SIGNAL_NET:
 *          break;
 *      case (NetKind) POWER_NET:
 *          fprintf (filePtr, "2 ");
 *          break;
 *      case (NetKind) GROUND_NET:
 *          fprintf (filePtr, "1 ");
 *          break;
 *      default:
 *          Bear_PrintErr("WritePinInBBLFormat() failed : illegal net kind");
 *          return (BEAR_STOP_SEARCH);
 *          break;
 *      }
 *      
 *      if ((netKind == POWER_NET) || (netKind == GROUND_NET)) {
 *          portInstPtr = Mac_GetPortOfPin(pinInstPtr);
 *          fprintf (filePtr, "%lf ", Mac_GetMaxCurrentReqOfPort(portInstPtr));
 *          fprintf (filePtr, "%lf ", Mac_GetMaxVoltDropOfPort(portInstPtr));
 *      }
 *  }
 */

    fprintf (filePtr, "\n");

    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_BBL_CODE */



static BearStatus
WritePinInCifFormat(pinInstPtr, writeDataPtr)
PinInst     *pinInstPtr;
WriteData   *writeDataPtr;
/*
 * Write out the pin's absolute position in cif format.
 *
 * Write out the name of the pins's net.
 */
{
    Net     *netPtr;
    Point   absPos;

    Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
    absPos.p_x *= writeDataPtr->scale;
    absPos.p_y *= writeDataPtr->scale;

    netPtr = Mac_GetNetOfPin(pinInstPtr);
    fprintf (writeDataPtr->filePtr, "94 %s %d %d;\n",
        Mac_GetNameOfNet(netPtr), absPos.p_x, absPos.p_y);

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
WritePointOfLineSegInBBLFormat(lineSegPtr, writeDataPtr)
LineSeg     *lineSegPtr;
WriteData   *writeDataPtr;
{
    Point relPos;       /* relative position */

    if (writeDataPtr->scaleFlag) {
        relPos.p_x =
            (lineSegPtr->line.point1.p_x - writeDataPtr->originPtr->p_x) * 
            writeDataPtr->xScale;
        relPos.p_y =
            (lineSegPtr->line.point1.p_y - writeDataPtr->originPtr->p_y) * 
            writeDataPtr->yScale;
    }
    else {
        relPos.p_x =
            (lineSegPtr->line.point1.p_x - writeDataPtr->originPtr->p_x); 
        relPos.p_y =
            (lineSegPtr->line.point1.p_y - writeDataPtr->originPtr->p_y); 
    }
    fprintf (writeDataPtr->filePtr, "%d %d\n", relPos.p_x, relPos.p_y);

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
WritePointInBBLFormat(pointPtr, writeDataPtr)
Point       *pointPtr;
WriteData   *writeDataPtr;
{
    Point relPos;       /* relative position */

    if (writeDataPtr->scaleFlag) {
        relPos.p_x = (pointPtr->p_x - writeDataPtr->originPtr->p_x) * 
            writeDataPtr->xScale;
        relPos.p_y = (pointPtr->p_y - writeDataPtr->originPtr->p_y) *
            writeDataPtr->yScale;
    }
    else {
        relPos.p_x = (pointPtr->p_x - writeDataPtr->originPtr->p_x);
        relPos.p_y = (pointPtr->p_y - writeDataPtr->originPtr->p_y);
    }
    fprintf (writeDataPtr->filePtr, "%d %d\n", relPos.p_x, relPos.p_y);

    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
AddPointToPointList(lineSegPtr, writeDataPtr)
LineSeg     *lineSegPtr;
WriteData   *writeDataPtr;
{
    Mac_CreateLinkedPoint(lineSegPtr->line.point1.p_x,
        lineSegPtr->line.point1.p_y, &writeDataPtr->pointList,
        (Point **) NULL);

    return (BEAR_CONTINUE_SEARCH);
}



/* ARGSUSED */
BearStatus
PbWriteChipCmd(winDataPtr, screenPtr, argCount, argList)
WindowData  *winDataPtr;
Point       *screenPtr;
int         argCount;
char        **argList;
/*
 * BEAR_OK is returned if the command is successfully completed.
 */
{
#define DEFAULT_ERR_FILE_NAME "errorList"

    extern BearStatus PbWriteLily();

#ifdef USE_OCT_CODE
    bool        writeOctFlag = FALSE;
    bool        writeRoutingFlag = FALSE;
#endif /* USE_OCT_CODE */

#ifndef BEAR_BLACK_BOX
    ChipWin     *chipWinPtr;
    bool        chipWinFlag = FALSE;
#endif BEAR_BLACK_BOX

    BearStatus  bearStatus;
    CellMaster  *cellMasterPtr;
    WriteData   writeData;
    double      atof ();
    double      scale;
    lsList      ciList = (lsList) NULL;
    int         i;
    char        *fileNamePtr = NIL(char);
    char        *bfpFileNamePtr = NIL(char);
    char        *errFileNamePtr = NIL(char);
    char        *rst0FileNamePtr = NIL(char);
    char        *rst1FileNamePtr = NIL(char);
    char        *rst2FileNamePtr = NIL(char);
    char        *rrStFileNamePtr = NIL(char);
    char        *techFileNamePtr = NIL(char);
    char        *timingFileNamePtr = NIL(char);
    char        *shapeFileNamePtr = NIL(char);
    char        *mdFileNamePtr = NIL(char);
    char        *cellViewPtr = NIL(char);
    bool        writeBBLFlag = FALSE;
    bool        writeBFPFlag = FALSE;
    bool        writeCifFlag = FALSE;
    bool        writeRST0Flag = FALSE;      /* write net list */
    bool        writeRST1Flag = FALSE;
    bool        writeRST2Flag = FALSE;
    bool        writeRrStFlag = FALSE;      /* write ring routing statistics */
    bool        writeMagFlag = FALSE;
    bool        writeTechFlag = FALSE;
    bool        writeTimingFlag = FALSE;
    bool        writeXFormFlag = FALSE;
    bool        writePsFlag = FALSE;        /* write postscript              */
    bool        writeShapeFlag = FALSE;
    bool        writeLilyFlag = FALSE;

#ifndef BEAR_BLACK_BOX
    if (bear_BlackBoxFlag == FALSE) {
        if (PbIsChipWin(winDataPtr) == TRUE) {
            /* Mouse is in a chip window. */

            chipWinFlag = TRUE;

            chipWinPtr = PbGetChipWin(winDataPtr);
            writeData.cellMasterPtr = cellMasterPtr =
                PbGetCellMasterOfCHW(chipWinPtr);
        }
        else if (PbIsCifWin(winDataPtr) == FALSE) {
            /* Mouse is not in a cif window, which is the only
             * other posiblity.
             */
            return (BEAR_ERROR);
        }
    }
    else
#endif /* BEAR_BLACK_BOX */
    {
        writeData.cellMasterPtr = cellMasterPtr = pbCellMasterPtr;
    }
    writeData.pinFlag = TRUE;
    writeData.mirrorYAxisFlag = FALSE;
    writeData.floatingPinFlag = FALSE;
    writeData.scaleFlag = FALSE;
    writeData.xScale = writeData.yScale = writeData.scale = 1.0;
    writeData.mixFlag = FALSE;

    /* parse arguments */
    for (i = 1; i < argCount ;i++) {
        /* 
         * Print out this commands help message.
         */
        if (strcmp (argList[i], "-help") == 0) {
            goto usage;
        }

#ifdef USE_BBL_CODE
        if (strcmp (argList[i], "-bbl") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            fileNamePtr = argList[i];
            writeBBLFlag = TRUE;
            continue;
        }
#endif /* USE_BBL_CODE */

        if (strcmp (argList[i], "-bfp") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            bfpFileNamePtr = argList[i];
            writeBFPFlag = TRUE;
            continue;
        }

        if (strcmp (argList[i], "-err") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            errFileNamePtr = argList[i];
            continue;
        }

        if (strcmp (argList[i], "-cif") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            fileNamePtr = argList[i];
            writeCifFlag = TRUE;
            continue;
        }
        if (strcmp (argList[i], "-xform") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            fileNamePtr = argList[i];
            writeXFormFlag = TRUE;
            continue;
        }
        if (strcmp (argList[i], "-ps") == 0) {
            if (bear_BlackBoxFlag == TRUE) {
                PbPrint(EZ_ITALIC_FONT,
                    "-ps option not used in black box mode.\n");
                return (BEAR_ERROR);
            }
            writePsFlag = TRUE;
            continue;
        }
        if (strcmp (argList[i], "-rst0") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeRST0Flag = TRUE;
            rst0FileNamePtr = argList[i];
            continue;
        }
        if (strcmp (argList[i], "-rst1") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeRST1Flag = TRUE;
            rst1FileNamePtr = argList[i];
            continue;
        }
        if (strcmp (argList[i], "-rst2") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeRST2Flag = TRUE;
            rst2FileNamePtr = argList[i];
            continue;
        }
        if (strcmp (argList[i], "-tech") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeTechFlag = TRUE;
            techFileNamePtr = argList[i];
            continue;
        }

        if (strcmp (argList[i], "-np") == 0) {
            writeData.pinFlag = FALSE;
            continue;
        }

        if (strcmp (argList[i], "-s") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeData.scaleFlag = TRUE;
            writeData.scale = atof (argList[i]);
            continue;
        }

        if (strcmp (argList[i], "-my") == 0) { 
            /*
             * The entire chip will be mirrored across the y-axis,
             * when it's written out. (bbl only)
             */
            writeData.scaleFlag = TRUE;
            writeData.mirrorYAxisFlag = TRUE;
            continue;
        }

#ifdef USE_OCT_CODE
        if (strcmp (argList[i], "-oct") == 0) {
            writeOctFlag = TRUE;
            ++i;
            if (i == argCount) {
                continue;
            } 
            else if (i < argCount) {
                if (strcmp(argList[i],"-r") == 0) {
                    writeRoutingFlag = TRUE;
                    i++;
                }

                fileNamePtr = argList[i];
                if (++i < argCount) {
                    cellViewPtr = argList[i];
                    continue;
                } 
                else {
                    goto usage;
                }
            } 
            else {
                goto usage;
            }
        }
#endif /* USE_OCT_CODE */

#ifdef USE_NEWP_CODE
        /*
         * When the mix option is used with the bbl option
         * the data will be "mixed" so that half the cells will be 
         * designated flexible with floating pins and the other 
         * half will be designated fixed with fixed pins.
         */
        if (strcmp (argList[i], "-mix") == 0) {
            writeData.mixFlag = TRUE;
            continue;
        }

        if (strcmp (argList[i], "-timing") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeTimingFlag = TRUE;
            timingFileNamePtr = argList[i];
            continue;
        }

#ifdef DEBUG
        if (strcmp (argList[i], "-shape") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeShapeFlag = TRUE;
            shapeFileNamePtr = argList[i];
            continue;
        }
#endif /* DEBUG */
#endif /* USE_NEWP_CODE */

#ifdef USE_DR_CODE
        if (strcmp (argList[i], "-mag") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            fileNamePtr = argList[i];
            writeMagFlag = TRUE;
            continue;
        }
        if (strcmp (argList[i], "-rrst") == 0) {
            if (++i >= argCount) {
                goto usage;
            }
            writeRrStFlag = TRUE;
            rrStFileNamePtr = argList[i];
            continue;
        }
#endif /* USE_DR_CODE */

        if (strcmp (argList[i], "-lily") == 0) { 
            /* 
             * The -lily (massoud data) option is used to write out a list
             * of the cell instances sorted by the x coordinate of there
             * lower left hand corner.
             */
            if (++i >= argCount) {
                goto usage;
            }
            writeLilyFlag = TRUE;
            mdFileNamePtr = argList[i];
            continue;
        }

#ifdef DEBUG
        if (strcmp (argList[i], "-fp") == 0) { 
            /*
             * All of the pins will be marked flexible as they are
             * written out. (bbl only)
             */
            writeData.floatingPinFlag = TRUE;
            continue;
        }
#endif /* DEBUG */

        goto usage;
    }

    if ((writeCifFlag == FALSE) &&

        (writeBFPFlag == FALSE) &&

#ifdef USE_BBL_CODE
        (writeBBLFlag == FALSE) &&
#endif /* USE_BBL_CODE */

#ifdef USE_OCT_CODE
        (writeOctFlag == FALSE) &&
#endif /* USE_OCT_CODE */

        (writeTechFlag == FALSE) &&

        (writeRST0Flag == FALSE) &&
        (writeRST1Flag == FALSE) &&
        (writeRST2Flag == FALSE) &&
        (writeXFormFlag == FALSE) &&

#ifdef USE_NEWP_CODE
        (writeShapeFlag == FALSE) &&
        (writeTimingFlag == FALSE) &&
#endif /* USE_NEWP_CODE */

#ifdef USE_DR_CODE
        (writeMagFlag == FALSE) &&
        (writeRrStFlag == FALSE) &&
#endif /* USE_DR_CODE */

        (writeLilyFlag == FALSE) &&
        (writePsFlag == FALSE)) {

        PbPrint(EZ_ITALIC_FONT, "Must indicate a format and a file.\n");
        goto usage;
    }

#ifndef BEAR_BLACK_BOX
    if ((bear_BlackBoxFlag == FALSE) && (writePsFlag == FALSE) && 
        (chipWinFlag == FALSE)) {
        PbPrint(EZ_ITALIC_FONT,
            "Place mouse in a chip window and try again.\n");
        return (BEAR_ERROR);
    }
#endif BEAR_BLACK_BOX

    if ((writeData.mirrorYAxisFlag == TRUE) && (writeBBLFlag == FALSE)) {
        PbPrint(EZ_ITALIC_FONT,    
            "The mirror option is for bbl files only.\n");
        goto usage;
    }

    if ((writeBBLFlag == TRUE) || (writeShapeFlag == TRUE)) {
        ciList = Mac_CreateCellInstList(cellMasterPtr);
        if (writeData.mixFlag == TRUE)  {
            int i, num;

            /*
             * Remove half of the cell instances from the list.
             */
            num = lsLength(ciList) / 2;
            for (i = 0; i < num; i++) {
                (void) lsDelEnd(ciList, (lsGeneric *) NULL);
            }
        }
    }

    if ((writeBFPFlag == TRUE) 

#ifdef USE_BBL_CODE
        || (writeBBLFlag == TRUE)
#endif USE_BBL_CODE

#ifdef USE_OCT_CODE
        || (writeOctFlag == TRUE)
#endif USE_OCT_CODE

        ) {

        if (errFileNamePtr == NIL(char)) {
            errFileNamePtr = DEFAULT_ERR_FILE_NAME;
        }
        if (Mac_CheckCellMaster(cellMasterPtr, errFileNamePtr) == BEAR_ERROR) {
            (void) sprintf (bear_ErrorMsg,
                "See list of errors in the file named \"%s\".\n", 
                errFileNamePtr);
            PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
        }
    }

    if (writeBFPFlag == TRUE) {
        Bfp_WriteOptions writeOpts;

        writeOpts.pinFlag = writeData.pinFlag;

        if ((writeOpts.scaleFlag = writeData.scaleFlag) == TRUE) {
            writeOpts.xScale = writeOpts.yScale = writeData.scale;
        }

        if (Bfp_MacToBearFP(bfpFileNamePtr, &writeOpts, cellMasterPtr) ==
            BEAR_OK) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                PbPutUpdateStatusOfCHW(FALSE, chipWinPtr);
            }
#endif BEAR_BLACK_BOX
            (void) sprintf (bear_Msg,
                "Cell data written in BFP format to file named\n    \"%s\"\n",
                bfpFileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else {
            PbPrint(EZ_ITALIC_FONT, "Unable to write data in BFP format.\n");
            return (BEAR_ERROR);
        }
    }

#ifdef USE_OCT_CODE
    if (writeOctFlag == TRUE) {
        if (writeRoutingFlag && !(Mac_IsRingRoutingDone(cellMasterPtr))) {
            PbPrint(EZ_ITALIC_FONT, 
                "Complete the ring routing and try again.\n");
            return (BEAR_ERROR);
        }
        else {
            if (fileNamePtr == NIL(char)) {
                if (odi_outchip.contents.facet.cell == NIL(char)) {
                    (void) sprintf (bear_ErrorMsg, 
                        "Must be the first save after an \"ow -oct\".\n");
                    PbPrint(EZ_ITALIC_FONT, bear_ErrorMsg);
                    return (BEAR_ERROR);
                }
                else {
                    if (Odi_WriteBackToOct(cellMasterPtr) == BEAR_ERROR) {
                        PbPrint(EZ_ITALIC_FONT,
                            "Unable to write back OCT data.\n");
                        return (BEAR_ERROR);
                    }
                }
            }
            else {
                if (Odi_WriteToOct(cellMasterPtr, fileNamePtr, cellViewPtr,
                    writeRoutingFlag) == BEAR_ERROR) {
                    PbPrint(EZ_ITALIC_FONT,
                        "Unable to write data in OCT format.\n");
                    return (BEAR_ERROR);
                }
            }
            PbPrint(EZ_NORMAL_FONT, "OCT data written.\n");
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                PbPutUpdateStatusOfCHW(FALSE, chipWinPtr);
            }
#endif BEAR_BLACK_BOX
        }
    }
#endif USE_OCT_CODE

    writeData.fileNamePtr = fileNamePtr;

#ifdef USE_BBL_CODE
    if (writeBBLFlag == TRUE) { 
        if (writeData.scaleFlag == TRUE) {
            writeData.xScale = writeData.yScale = writeData.scale;
            if (writeData.mirrorYAxisFlag == TRUE) {
                writeData.xScale *= -1.0;
            }
        }

        if (writeData.mixFlag == TRUE) {
            writeData.ciList = ciList;
        }

        if (WriteChipInBBLFormat(&writeData) == BEAR_OK) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                PbPutUpdateStatusOfCHW(FALSE, chipWinPtr);
            }
#endif BEAR_BLACK_BOX
            (void) sprintf (bear_Msg,
                "Cell data written in BBL format to file named\n    \"%s\"\n",
                fileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else {
            PbPrint(EZ_ITALIC_FONT, "Unable to write data in BBL format.\n");
            return (BEAR_ERROR);
        }
    }
#endif USE_BBL_CODE

    if (writeCifFlag == TRUE) {
        if (WriteChipInCifFormat(&writeData) == BEAR_OK) {
#ifndef BEAR_BLACK_BOX
            if (bear_BlackBoxFlag == FALSE) {
                PbPutUpdateStatusOfCHW(FALSE, chipWinPtr);
            }
#endif BEAR_BLACK_BOX
            (void) sprintf (bear_Msg, 
                "Cell data written in CIF format to file named \"%s\"\n",
                fileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else {
            PbPrint(EZ_ITALIC_FONT, "Unable to write data in CIF format.\n");
            return (BEAR_ERROR);
        }
    } 

#ifdef USE_DR_CODE
    if (writeMagFlag == TRUE) {
        (void) WriteChipWireInMagFormat(&writeData);
    }
#endif /* USE_DR_CODE */

    if (writeRST0Flag == TRUE) {
        bearStatus = Mac_WriteStats(0, cellMasterPtr, rst0FileNamePtr);
        if (bearStatus == BEAR_OK) {
            (void) sprintf (bear_Msg, 
                "Net list written to file named\n     \"%s\"\n",
                rst0FileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else if (bearStatus == BEAR_ERROR_3) {
            (void) sprintf (bear_Msg,  
                "Could not open file named \"%s\" for writing.\n",
                rst0FileNamePtr);
            PbPrint(EZ_ITALIC_FONT, bear_Msg);
            return (BEAR_ERROR);
        }
        else {
            return (BEAR_ERROR);
        }
    }

    if (writeRST1Flag == TRUE) {
        bearStatus = Mac_WriteStats(1, cellMasterPtr, rst1FileNamePtr);
        if (bearStatus == BEAR_OK) {
            (void) sprintf (bear_Msg, "%s\n%s\"%s\"\n",
               "Routing statistics based on bounding boxes",
               "written to file named ",
                rst1FileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else if (bearStatus == BEAR_ERROR_3) {
            (void) sprintf (bear_Msg,  
                "Could not open file named \"%s\" for writing.\n",
                rst1FileNamePtr);
            PbPrint(EZ_ITALIC_FONT, bear_Msg);
            return (BEAR_ERROR);
        }
        else {
            return (BEAR_ERROR);
        }
    }

    if (writeRST2Flag == TRUE) {
        bearStatus = Mac_WriteStats(2, cellMasterPtr, rst2FileNamePtr);
        if (bearStatus == BEAR_OK) {
            (void) sprintf (bear_Msg, "%s\n%s\"%s\"\n",
               "Routing statistics based on steiner trees",
               "written to file named ",
                rst2FileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else if (bearStatus == BEAR_ERROR_3) {
            (void) sprintf (bear_Msg,  
                "Could not open file named \"%s\" for writing.\n",
                rst2FileNamePtr);
            PbPrint(EZ_ITALIC_FONT, bear_Msg);
            return (BEAR_ERROR);
        }
        else {
            return (BEAR_ERROR);
        }
    }

#ifdef USE_DR_CODE
    if (writeRrStFlag == TRUE) {
        bearStatus = Mac_WriteStats(3, cellMasterPtr, rrStFileNamePtr);
        if (bearStatus == BEAR_OK) {
            (void) sprintf (bear_Msg,  
               "Statistics for detailed routing written to file named \"%s\"\n",
                rrStFileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else if (bearStatus == BEAR_ERROR) {
            PbPrint(EZ_ITALIC_FONT, "Complete ring routing and try again.\n");
            return (BEAR_ERROR);
        }
        else if (bearStatus == BEAR_ERROR_2) {
            return (BEAR_ERROR);
        }
        else if (bearStatus == BEAR_ERROR_3) {
            (void) sprintf (bear_Msg,  
                "Could not open file named \"%s\" for writing.\n",
                rrStFileNamePtr);
            PbPrint(EZ_ITALIC_FONT, bear_Msg);
            return (BEAR_ERROR);
        }
    }
#endif /* USE_DR_CODE */

#ifdef USE_NEWP_CODE
#ifdef DEBUG
    if (writeShapeFlag == TRUE) {
        char *cmFileNamePtr;

        cmFileNamePtr = Mac_GetFileNameOfCellMaster(cellMasterPtr);
        if (Ncl_WriteBBLShapes(cmFileNamePtr, ciList, shapeFileNamePtr) ==
            BEAR_OK) {
            (void) sprintf (bear_Msg,
                "Shape data written to file named\n    \"%s\"\n", 
                shapeFileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else {
            PbPrint(EZ_ITALIC_FONT, "Unable to write shape data.\n");
            return (BEAR_ERROR);
        }
    }
#endif DEBUG

    if (writeTimingFlag == TRUE) {
        if (Ncl_WriteTiming(cellMasterPtr, timingFileNamePtr) ==
            BEAR_OK) {
            (void) sprintf (bear_Msg,
                "Timing data written to file named\n    \"%s\"\n",
                timingFileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else {
            PbPrint(EZ_ITALIC_FONT, "Unable to write timing data.\n");
            return (BEAR_ERROR);
        }
    }
#endif /* USE_NEWP_CODE */

    if (writeTechFlag == TRUE) {
        scale = ABS(writeData.scale);
        if (Mac_WriteTech(techFileNamePtr, scale) == BEAR_OK) {
            char *stringPtr;

            (void) sprintf (bear_Msg,
                "Technology data written to file named\n    \"%s\"\n",
                techFileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);

            MALLOC(char *, stringPtr,
                (strlen (techFileNamePtr) + 1) * sizeof (char));
            strcpy (stringPtr, techFileNamePtr);
            Mac_PutTechFileNameOfCellMaster(stringPtr, cellMasterPtr);
        }
        else {
            PbPrint(EZ_ITALIC_FONT,
                "Write aborted because technology data is illegal.\n");
            return (BEAR_ERROR);
        }
    }

    if (writeXFormFlag == TRUE) {
        (void) WriteChipTransform(&writeData);
    }
    
    if (writeLilyFlag == TRUE)  {
        if (PbWriteLily(cellMasterPtr, mdFileNamePtr) == BEAR_OK) {
            (void) sprintf (bear_Msg,
                "Lily data written to file named\n    \"%s\"\n",
                mdFileNamePtr);
            PbPrint(EZ_NORMAL_FONT, bear_Msg);
        }
        else {
            PbPrint(EZ_ITALIC_FONT, "Unable to write Lily data.\n");
            return (BEAR_ERROR);
        }
    }

#ifndef BEAR_BLACK_BOX
    if (writePsFlag == TRUE) {
        /*
         * The user may generate a postscript file of either a chip
         * or a cif window.
         */

        double  tx, ty, sx, sy, r;
#ifdef USE_XDM_CODE
        char    filename[40];

        /*
         * (gyorgy : 14 sept 90) This dialog still returns a filename
         * filled with garbage characters.
         */
        PbPsPlotDialog(winDataPtr, filename, &tx, &ty, &sx, &sy, &r);
#else  /* USE_XDM_CODE */
        tx = 0.0;
        ty = 10.5;
        sx = sy = 0.0139;
        r = 0.0;
#endif /* USE_XDM_CODE */

        r = r + 180.0;
        if (Ez_OpenPsFile("chip.ps",
            winDataPtr->wd_userWinInfo.width,
            winDataPtr->wd_userWinInfo.height,
            tx, ty, sx, sy, r) != EZ_OK) {

            return (BEAR_ERROR);
        }

        (*winDataPtr->wd_redrawProc)(winDataPtr, NIL(Rect));

        Ez_ClosePsFile();
    }
#endif BEAR_BLACK_BOX

    if (ciList != (lsList) NULL) {
        lsDestroy(ciList, (void (*)()) NULL);
    }

    return (BEAR_OK);

usage:
    PbPrint(EZ_NORMAL_FONT, "write chip (ok in script) (ok in black box)\n");
    PbPrint(EZ_NORMAL_FONT, "Usage: w [-options ...]\n");
    PbPrint(EZ_NORMAL_FONT, "\nwhere options include:\n");

    PbPrint(EZ_NORMAL_FONT, "    -bfp fileName use BearFP format\n");

#ifdef USE_BBL_CODE
    PbPrint(EZ_NORMAL_FONT, "    -bbl fileName use BBL format\n");
#endif USE_BBL_CODE

#ifdef USE_OCT_CODE
    PbPrint(EZ_NORMAL_FONT,
        "    -oct [[-r] output_cellname output_viewname]\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -oct writes back to OCT only if the data was previously\n");
    PbPrint(EZ_NORMAL_FONT,
        "         entered as follows \"ow -oct cellname viewname\".\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -oct output_cellname output_viewname writes back to OCT.\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -r  write the routing data back to OCT.\n");
#endif USE_OCT_CODE

    PbPrint(EZ_NORMAL_FONT,
        "    -err  Any errors that occur while saving the chip will be\n");
    PbPrint(EZ_NORMAL_FONT,
        "          listed here. (bfp, bbl and oct only)\n");

    PbPrint(EZ_NORMAL_FONT, "    -tech fileName write out technology data\n");

    PbPrint(EZ_NORMAL_FONT, "    -cif fileName use CIF format\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -rst0 fileName write net list\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -rst1 fileName write routing stats based on bounding boxes\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -rst2 fileName write routing stats based on steiner trees\n");

#ifdef USE_DR_CODE
    PbPrint(EZ_NORMAL_FONT,
        "    -rrst fileName write statistics after ring router\n");

    PbPrint(EZ_NORMAL_FONT,
        "    -mag filename use magic format (wiring only)\n");
#endif /* USE_DR_CODE */
     
#ifdef USE_NEWP_CODE
#ifdef DEBUG
    PbPrint(EZ_NORMAL_FONT, "    -shape fileName write out shape data\n");
#endif DEBUG
    PbPrint(EZ_NORMAL_FONT, "    -timing fileName write out timing data\n");
#endif /* USE_NEWP_CODE */

    /*
     * test: grabs transform by iterating over all cellInstances of 
     *       parentcell.
     * xform: get transform by emptying queue in plDatabase.c
     */
    PbPrint(EZ_NORMAL_FONT, "    -xform fileName write out transform\n");

#ifdef DEBUG
    PbPrint(EZ_NORMAL_FONT, "    -mix\n");
    PbPrint(EZ_NORMAL_FONT, "    -fp   mark all pins flexible (bbl only)\n");
    PbPrint(EZ_NORMAL_FONT, "    -lily filename\n");
#endif DEBUG

    PbPrint(EZ_NORMAL_FONT, "    -np   no pin information will be printed\n");
    PbPrint(EZ_NORMAL_FONT, "    -s    scale (float)\n");
    PbPrint(EZ_NORMAL_FONT,
        "    -my   mirror about y axis (bbl only)\n");

    PbPrint(EZ_NORMAL_FONT, "    -ps   output window to PostScript\n");

    return (BEAR_ERROR);
    
#undef DEFAULT_ERR_FILE_NAME
}



