
/* pbWriteLily.c --
 *
 *     George Carvalho, Esq.
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/polarBear/pbWriteLily.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include "port.h"
#include "malloc.h"
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "ezPlot.h"
#include "polarBear.h"

#ifdef USE_OCT_CODE
#include "vov.h"
#endif /* USE_OCT_CODE */


/* ---------- Local Procedures ---------- */
static int CompareCI();


static int
CompareCI(item1, item2)
lsGeneric item1, item2;
{
    CellInst    *cellInst1Ptr = (CellInst *) item1;
    CellInst    *cellInst2Ptr = (CellInst *) item2;
    GenericCell gcell;
    LinkedRect  *horzRectList;
    Rect        absRect;
    Transform   *transPtr;
    int         x1, x2;

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInst1Ptr;
    transPtr = Mac_GetTransformOfGCell(&gcell);
    horzRectList = Mac_GetHorzRectListOfGCell(&gcell);
    GeoTransRect(transPtr, &horzRectList->r_r, &absRect);
    x1 = absRect.r_xbot;

    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInst2Ptr;
    transPtr = Mac_GetTransformOfGCell(&gcell);
    horzRectList = Mac_GetHorzRectListOfGCell(&gcell);
    GeoTransRect(transPtr, &horzRectList->r_r, &absRect);
    x2 = absRect.r_xbot;

    return ((x1 == x2) ? 0 : ((x1 < x2) ? -1 : 1));
}


/* 
 * ----------------------------------------------------------------------------
 *
 * PbWriteLily() --
 *
 * This procedure is used to write out a list of the cell instances
 * sorted by the x coordinate of there lower left hand corner.
 * All of the cell instances must be rectangular.
 * 
 * Results:
 *   BEAR_OK is returned if everything is spiffy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
PbWriteLily(coreMasterPtr, lilyFileNamePtr)
CellMaster  *coreMasterPtr;
char        *lilyFileNamePtr;
{
    FILE        *filePtr;
    CellInst    *cellInstPtr;
    GenericCell gcell;
    LinkedRect  *horzRectList;
    lsList      ciList;
    lsGen       gen;

    ciList = Mac_CreateCellInstList(coreMasterPtr);
    gcell.unType = CELL_INST;
    lsForeachItem(ciList, gen, cellInstPtr) {
        gcell.unValue.cellInstPtr = cellInstPtr;

        horzRectList = Mac_GetHorzRectListOfGCell(&gcell);
        if (horzRectList->r_next != NIL(LinkedRect)) {
            (void) sprintf (bear_ErrorMsg,
                "PbWriteLily() : cell instance %s is not a rectangle.\n",
                Mac_GetNameOfCellInst(cellInstPtr));
            Bear_PrintErr(bear_ErrorMsg);

            lsFinish(gen);
            lsDestroy(ciList, (void (*)()) NULL);
            return (BEAR_ERROR);
        }
    }
    lsSort(ciList, CompareCI);

#ifdef USE_OCT_CODE
    (void) VOVoutput(VOV_UNIX_FILE, lilyFileNamePtr);
#endif /* USE_OCT_CODE */

    if ((filePtr = fopen (lilyFileNamePtr, "w")) == NIL(FILE)) {
        (void) sprintf (bear_ErrorMsg,
            "PbWriteLily() : could not open \"%s\" for writing\n", 
            lilyFileNamePtr);
        Bear_PrintErr(bear_ErrorMsg);
        return (BEAR_ERROR);
    }

    lsForeachItem(ciList, gen, cellInstPtr) {
        (void) fprintf (filePtr, "%s\n", Mac_GetNameOfCellInst(cellInstPtr));
    }
    lsDestroy(ciList, (void (*)()) NULL);

    fclose (filePtr);

    return (BEAR_OK);
}


