
/* rsHorzRectSlice.c --
 *
 *     Tim Collins
 *     (Advising Professor - E. S. Kuh)
 *
 * It contains the subroutines to horizontally dissect
 * the polygon into rectangles.
 *
 */

#ifndef lint
static char rcsid[] = 
    "$Header: /home/cvs/octtools/src/BearFP-1.1/src/rectSlice/rsHorzRectSlice.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "rectSlice.h"
#include "rectSliceInt.h"


/* ---------- Local Procedures ---------- */
static LinkedLine   *FindSliceLine();
static LinkedRect   *SliceRect();
static BearStatus   ConvertPointToHLine();
static int          DetermineSliceType();


static BearStatus
ConvertPointToHLine(pointList, lineListPtr, countPtr)
LinkedPoint   *pointList;
LinkedLine    **lineListPtr;
int           *countPtr;
{
    LinkedPoint   *pointPtr;
    LinkedLine    *linePtr;
    int           needFirstPoint;

    if (pointList == NIL(LinkedPoint))
    {
        Bear_PrintErr("RsHorzRectSlice() failed");
        return (BEAR_ERROR);
    }

    *countPtr = 0;
    needFirstPoint = 0;
    pointPtr = pointList;
    do
    {
    pointPtr = pointPtr->nextPtr;
    (*countPtr)++;
    } while(pointPtr != NIL(LinkedPoint));
    *countPtr /= 2;
    MALLOC(LinkedLine *, *lineListPtr, *countPtr * sizeof (LinkedLine));
    pointPtr = pointList;
    linePtr = *lineListPtr;
    if (pointPtr->point.p_y != pointPtr->nextPtr->point.p_y)
    {
    needFirstPoint = 1;
    }
    do
    {
    if (pointPtr->point.p_x != pointPtr->nextPtr->point.p_x)
    {
        if (pointPtr->point.p_y != pointPtr->nextPtr->point.p_y) 
        {
        Bear_PrintErr("RsHorzRectSlice() illegally shaped polygon");
        return (BEAR_ERROR);
        }
        else
        {
        if (pointPtr->point.p_x < pointPtr->nextPtr->point.p_x)
        {
            linePtr->line.point1.p_x = pointPtr->point.p_x;
            linePtr->line.point1.p_y = pointPtr->point.p_y;
            linePtr->line.point2.p_x = pointPtr->nextPtr->point.p_x;
            linePtr->line.point2.p_y = pointPtr->nextPtr->point.p_y;
        }
        else
        {
            linePtr->line.point1.p_x = pointPtr->nextPtr->point.p_x;
            linePtr->line.point1.p_y = pointPtr->nextPtr->point.p_y;
            linePtr->line.point2.p_x = pointPtr->point.p_x;
            linePtr->line.point2.p_y = pointPtr->point.p_y;
        }
        linePtr->nextPtr = linePtr + 1;
        linePtr = linePtr->nextPtr;
        }
    }
    pointPtr = pointPtr->nextPtr;
    } while (pointPtr->nextPtr != NIL(LinkedPoint));
    if (needFirstPoint)
    {
    if (pointPtr->point.p_y != pointList->point.p_y)
    {
        Bear_PrintErr("RsHorzRectSlice() illegal polygon");
        return (BEAR_ERROR);
    }
    else
    {
        if (pointPtr->point.p_x < pointList->point.p_x)
        {
        linePtr->line.point1.p_x = pointPtr->point.p_x;
        linePtr->line.point1.p_y = pointPtr->point.p_y;
        linePtr->line.point2.p_x = pointList->point.p_x;
        linePtr->line.point2.p_y = pointList->point.p_y;
        }
        else
        {
        linePtr->line.point1.p_x = pointList->point.p_x;
        linePtr->line.point1.p_y = pointList->point.p_y;
        linePtr->line.point2.p_x = pointPtr->point.p_x;
        linePtr->line.point2.p_y = pointPtr->point.p_y;
        }
    }
    }
    return (BEAR_OK);
}



static int
Compare(line1, line2)
LinkedLine *line1, *line2;
{
    if (line1->line.point1.p_y > line2->line.point1.p_y)
    {
    return (-1);
    }
    else if (line1->line.point1.p_y < line2->line.point1.p_y)
    {
    return (1);
    }
    else if (line1->line.point1.p_x > line2->line.point1.p_x)
    {
    return (1);
    }
    else
    {
    return (-1);
    }
}



static void
SortLines(count, lineList, orderedLineListPtr)
int          count;
LinkedLine   *lineList;
OrderedLine  **orderedLineListPtr;
{
    int          loop;
    int          Compare();
    LinkedLine   *linePtr;
    LinkedLine   *linePlacePtr;
    OrderedLine  *orderedLinePtr;

    qsort(lineList, count, sizeof (LinkedLine), (int (*)()) Compare);
    { char *s; s = "CAN AIRPLANES LAND ON CLOUDS ?"; }
    linePtr = lineList;
    MALLOC(OrderedLine *, *orderedLineListPtr, sizeof (OrderedLine));
    orderedLinePtr = *orderedLineListPtr;
    orderedLinePtr->linePtr = lineList;
    linePlacePtr = lineList;
    linePtr = orderedLinePtr->linePtr;
    for (loop = 1; loop < count; loop++)
    {
    linePlacePtr++;
    if (linePtr->line.point1.p_y == linePlacePtr->line.point1.p_y)
    {
        linePtr->nextPtr = linePlacePtr;
        linePtr = linePtr->nextPtr;
    }
    else
    {
        linePtr->nextPtr = NIL(LinkedLine);
        MALLOC(OrderedLine *, orderedLinePtr->nextPtr, sizeof (OrderedLine));
        orderedLinePtr = orderedLinePtr->nextPtr;
        orderedLinePtr->linePtr = linePlacePtr;
        linePtr = orderedLinePtr->linePtr;
    }
    }
    orderedLinePtr->nextPtr = NIL(OrderedLine);
    linePtr->nextPtr = NIL(LinkedLine);
}



static LinkedLine *
FindSliceLine(orderedLineList, firstLinePtr, prevOrdLinePtrPtr, prevLinePtrPtr)
OrderedLine *orderedLineList;
LinkedLine  *firstLinePtr;
OrderedLine **prevOrdLinePtrPtr;
LinkedLine  **prevLinePtrPtr;
{
    OrderedLine *orderedLinePtr;
    LinkedLine  *linePtr;
    int         sliceType;
    int         newLine;
    int         firstLine;

    newLine = 1;
    *prevOrdLinePtrPtr = orderedLineList;
    if (orderedLineList->linePtr->line.point1.p_y ==
        firstLinePtr->line.point1.p_y)
    {
    orderedLinePtr = orderedLineList->nextPtr;
    firstLine = 0;
    }
    else
    {
    orderedLinePtr = orderedLineList;
    firstLine = 1;
    }
    linePtr = orderedLinePtr->linePtr;
    *prevLinePtrPtr = orderedLinePtr->linePtr;
    while(1)
    {
    sliceType = DetermineSliceType(firstLinePtr, linePtr);
    if (sliceType == NO_SLICE)
    {
        if (linePtr->nextPtr == NIL(LinkedLine))
        {
        orderedLinePtr = orderedLinePtr->nextPtr;
        if (!(firstLine))
        {
            *prevOrdLinePtrPtr = (*prevOrdLinePtrPtr)->nextPtr;
        }
        firstLine = 0;
        linePtr = orderedLinePtr->linePtr;
        *prevLinePtrPtr = orderedLinePtr->linePtr;
        newLine = 1;
        }
        else
        {
        linePtr = linePtr->nextPtr;
        if (newLine) 
        {
            newLine = 0;
        }
        else
        {
            *prevLinePtrPtr = (*prevLinePtrPtr)->nextPtr;
        }
        }
    }
    else
    {
        break;
    }
    }
    return (linePtr);
}



static int
DetermineSliceType(firstLinePtr, sliceLinePtr)
LinkedLine *firstLinePtr;
LinkedLine *sliceLinePtr;
{
    if (sliceLinePtr->line.point2.p_x == firstLinePtr->line.point1.p_x)
    {
    return (LEFT_SLICE);
    }
    if (sliceLinePtr->line.point1.p_x == firstLinePtr->line.point2.p_x)
    {
    return (RIGHT_SLICE);
    }
    if (sliceLinePtr->line.point1.p_x == firstLinePtr->line.point1.p_x)
    {
    if (sliceLinePtr->line.point2.p_x == firstLinePtr->line.point2.p_x)
    {
        return (WHOLE_SLICE);
    }
    else 
    {
        return (LEFT_MID_SLICE);
    }
    }
    if (sliceLinePtr->line.point1.p_x > firstLinePtr->line.point1.p_x)
    {
    if (sliceLinePtr->line.point2.p_x < firstLinePtr->line.point2.p_x)
    {
        return (MID_SLICE);
    }
    else if (sliceLinePtr->line.point2.p_x == firstLinePtr->line.point2.p_x)
    {
        return (RIGHT_MID_SLICE);
    }
    }
    return (NO_SLICE);
}



static LinkedRect *
SliceRect(firstLinePtr, sliceLinePtr)
LinkedLine  *firstLinePtr;
LinkedLine  *sliceLinePtr;
{
    LinkedRect  *rectPtr;

    MALLOC(LinkedRect  *, rectPtr, sizeof (LinkedRect));
    rectPtr->r_r.r_ur.p_x = firstLinePtr->line.point2.p_x;
    rectPtr->r_r.r_ur.p_y = firstLinePtr->line.point2.p_y;
    rectPtr->r_r.r_ll.p_x = firstLinePtr->line.point1.p_x;
    rectPtr->r_r.r_ll.p_y = sliceLinePtr->line.point1.p_y;
    rectPtr->r_next = NIL(LinkedRect);
    return (rectPtr);
}

static void
ExtendLineEnd(newEndpointLinePtr, currentLinePtr)
LinkedLine  *newEndpointLinePtr;
LinkedLine  *currentLinePtr;
{
    currentLinePtr->line.point2.p_x = newEndpointLinePtr->line.point2.p_x;
}

static void
ExtendLineStart(newStartpointLinePtr, currentLinePtr)
LinkedLine  *newStartpointLinePtr;
LinkedLine  *currentLinePtr;
{
    currentLinePtr->line.point1.p_x = newStartpointLinePtr->line.point1.p_x;
}



static void
RemoveLine(orderedLineList, prevOrdLinePtr, prevLinePtr, sliceLinePtr)
OrderedLine  **orderedLineList;
OrderedLine  *prevOrdLinePtr;
LinkedLine   *prevLinePtr;
LinkedLine   *sliceLinePtr;
{
    if (prevLinePtr == sliceLinePtr)
    {
    if (prevOrdLinePtr->linePtr == prevLinePtr)
    {
        if (prevOrdLinePtr->linePtr->nextPtr == NIL(LinkedLine))
        {
        *orderedLineList = (*orderedLineList)->nextPtr;
        }
        else
        {
        prevOrdLinePtr->linePtr = prevOrdLinePtr->linePtr->nextPtr;
        }
    }
    else
    {
        if (prevOrdLinePtr->nextPtr->linePtr->nextPtr == NIL(LinkedLine))
        {
        prevOrdLinePtr->nextPtr = prevOrdLinePtr->nextPtr->nextPtr;
        }
        else
        {
        prevOrdLinePtr->nextPtr->linePtr = 
                prevOrdLinePtr->nextPtr->linePtr->nextPtr;
        }
    }
    }
    else
    {
    prevLinePtr->nextPtr = sliceLinePtr->nextPtr;
    }
}



static void
SearchLineFix(firstLinePtr, sliceLinePtr)
LinkedLine  *firstLinePtr;
LinkedLine  *sliceLinePtr;
{
    LinkedLine  *currentLinePtr; 
    LinkedLine  *searchLinePtr; 
    int         sliceType;
    int         done;

    done = 0;
    currentLinePtr = sliceLinePtr;
    searchLinePtr = sliceLinePtr->nextPtr;
    do
    {
    if (searchLinePtr != NIL(LinkedLine))
    {
        sliceType = DetermineSliceType(firstLinePtr, searchLinePtr);
        switch(sliceType)
        {
        case RIGHT_SLICE:
            ExtendLineEnd(searchLinePtr, currentLinePtr);
            currentLinePtr->nextPtr = searchLinePtr->nextPtr;
            done = 1;
            break;
        case RIGHT_MID_SLICE:
            currentLinePtr->line.point2.p_x =
                searchLinePtr->line.point1.p_x;
            currentLinePtr->nextPtr = searchLinePtr->nextPtr;
            done = 1;
            break;
        case NO_SLICE:
            ExtendLineEnd(firstLinePtr, currentLinePtr);
            done = 1;
            break;
        case MID_SLICE:
            currentLinePtr->line.point2.p_x =
                searchLinePtr->line.point1.p_x;
            searchLinePtr->line.point1.p_x =
                searchLinePtr->line.point2.p_x;
            currentLinePtr = searchLinePtr;
            searchLinePtr = searchLinePtr->nextPtr;
            break;
        }
    }
    else
    {
        ExtendLineEnd(firstLinePtr, currentLinePtr);
        done = 1;
    }
    } while(!(done));
}



static void
FixSlicingLines(orderedLineList, prevOrdLinePtr, firstLinePtr, prevLinePtr,
    sliceLinePtr)

OrderedLine  **orderedLineList;
OrderedLine  *prevOrdLinePtr;
LinkedLine   *firstLinePtr;
LinkedLine   *prevLinePtr;
LinkedLine   *sliceLinePtr;
{
    LinkedLine  *newLinePtr;
    int         sliceType;

    sliceType = DetermineSliceType(firstLinePtr, sliceLinePtr);
    switch(sliceType)
    {
    case WHOLE_SLICE:
        RemoveLine(orderedLineList, prevOrdLinePtr, prevLinePtr,
            sliceLinePtr);
        break;
    case RIGHT_MID_SLICE:
        sliceLinePtr->line.point2.p_x = sliceLinePtr->line.point1.p_x;
    case RIGHT_SLICE:
        ExtendLineStart(firstLinePtr, sliceLinePtr);
        break;
    case LEFT_MID_SLICE:
        sliceLinePtr->line.point1.p_x = sliceLinePtr->line.point2.p_x;
    case LEFT_SLICE:
        SearchLineFix(firstLinePtr, sliceLinePtr);
        break;
    case MID_SLICE:
        MALLOC(LinkedLine  *, newLinePtr, sizeof (LinkedLine));
        newLinePtr->line.point1.p_x = sliceLinePtr->line.point2.p_x;
        newLinePtr->line.point1.p_y = sliceLinePtr->line.point1.p_y;
        newLinePtr->line.point2.p_y = sliceLinePtr->line.point1.p_y;
        sliceLinePtr->line.point2.p_x = sliceLinePtr->line.point1.p_x;
        sliceLinePtr->line.point1.p_x = firstLinePtr->line.point1.p_x;
        newLinePtr->nextPtr = sliceLinePtr->nextPtr;
        sliceLinePtr->nextPtr = newLinePtr;
        SearchLineFix(firstLinePtr, newLinePtr);
        break;
    }
}



BearStatus 
Rs_HorzRectSlice(pointList, rectListPtr)
LinkedPoint   *pointList;
LinkedRect    **rectListPtr;
{
    OrderedLine  *orderedLineList;
    OrderedLine  *prevOrdLinePtr;
    LinkedLine   *lineList;
    LinkedLine   *firstLinePtr;
    LinkedLine   *sliceLinePtr;
    LinkedLine   *prevLinePtr;
    LinkedRect   *rectPtr;
    int          count;

    *rectListPtr = NIL(LinkedRect);
    (void) ConvertPointToHLine(pointList, &lineList, &count);
    SortLines(count, lineList, &orderedLineList);
    while (orderedLineList != NIL(OrderedLine) &&
       orderedLineList->nextPtr != NIL(OrderedLine))
    {
    firstLinePtr = RsGetFirstLine(&orderedLineList);
    sliceLinePtr = FindSliceLine(orderedLineList, firstLinePtr,
        &prevOrdLinePtr, &prevLinePtr);
    rectPtr = SliceRect(firstLinePtr, sliceLinePtr);
    RsInsertRectList(rectListPtr, rectPtr);
    FixSlicingLines(&orderedLineList, prevOrdLinePtr, firstLinePtr,
        prevLinePtr, sliceLinePtr);
    }
    FREE((char *) lineList);
    return (BEAR_OK);
}
