/* soHorzSlack.c --
 *
 *     Massoud Pedram
 *
 * This module computes horizontal slacks and horizontal critical paths.
 *
 * Copyright (C) 1988 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/shapeOptimizer/soHorzSlack.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "soInt.h"
#include "soUtil.h"

/* ---------- Local Global Variables ---------- */
static Plane        *horzPlanePtr;
static Deque        *stack = NIL(Deque);
static int            curTileLeft, curTileRight, chipX;
static Tile            *leftPseudoTilePtr, *rightPseudoTilePtr;


/*
 * ----------------------------------------------------------------------------
 *
 * FindMaxNodeVal() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
FindMaxLeftVal(horzTilePtr, valPtr)

register Tile    *horzTilePtr;
int                *valPtr;

{
    int        posSlack;

    if (RIGHT(horzTilePtr) == curTileLeft  && Tp_IsBottleneckTile(horzTilePtr))
    {
        posSlack = SoGetHorzPosSlack(horzTilePtr);
        *valPtr = (posSlack > *valPtr) ? posSlack : *valPtr;
    }
    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * ComputeHorzPosSlack() --
 *
 *    Compute horz positive slack for all tiles.
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *
 * Side effects:
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus
ComputeHorzPosSlack (horzTilePtr, clientData)

register Tile    *horzTilePtr;
ClientData    clientData;

{
    Rect        planeSearchRect;
    int        tileWidth, maxLeftVal;
    CellInst    *cellInstPtr;
    TileInfo    *tileInfoPtr;

    if (Tp_IsCellInstTile (horzTilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile (horzTilePtr);
        PushDeque(stack, (ClientData) horzTilePtr);
        maxLeftVal = 0;
        tileWidth = RIGHT (horzTilePtr) - LEFT (horzTilePtr);
        planeSearchRect.r_xbot = LEFT (horzTilePtr) - 2;
        planeSearchRect.r_ybot = BOTTOM (horzTilePtr);
        planeSearchRect.r_xtop = LEFT (horzTilePtr) - 1;
        planeSearchRect.r_ytop = TOP (horzTilePtr);
        curTileLeft = LEFT (horzTilePtr);

        if (BL(horzTilePtr)  != leftPseudoTilePtr)
        {
            (void) TiSrArea(NIL(Tile),
                    horzPlanePtr,
                    &planeSearchRect,
                    FindMaxLeftVal,
                    (ClientData) &maxLeftVal);
        }
        SoPutHorzPosSlack (maxLeftVal + tileWidth, horzTilePtr);
        tileInfoPtr = (TileInfo *) cellInstPtr->clientData;
        tileInfoPtr->hPtr = horzTilePtr;
    }
    else if (Tp_IsBottleneckTile (horzTilePtr))
    {
        PushDeque(stack, (ClientData) horzTilePtr);
        tileWidth = Tp_GetDensityOfTile(horzTilePtr);
        maxLeftVal = (BL(horzTilePtr) != leftPseudoTilePtr) ?
                SoGetHorzPosSlack (BL(horzTilePtr)) : 0;
        SoPutHorzPosSlack (maxLeftVal + tileWidth, horzTilePtr);
    }
    else if (Tp_IsDCellTile (horzTilePtr))
    {
        PushDeque(stack, (ClientData) horzTilePtr);
        tileWidth = 0;
        /* tileWidth = RIGHT (horzTilePtr) - LEFT (horzTilePtr); */
        maxLeftVal = (BL(horzTilePtr) != leftPseudoTilePtr) ?
                SoGetHorzPosSlack (BL(horzTilePtr)) : 0;
        SoPutHorzPosSlack (maxLeftVal + tileWidth, horzTilePtr);
    }

    return (BEAR_CONTINUE_SEARCH); 
}



/*
 * ----------------------------------------------------------------------------
 *
 * FindMinRightVal() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
FindMinRightVal(horzTilePtr, valPtr)

register Tile    *horzTilePtr;
int                *valPtr;

{
    int        negSlack, tileWidth;

    if (LEFT(horzTilePtr) == curTileRight  && Tp_IsBottleneckTile(horzTilePtr))
    {
        negSlack = SoGetHorzNegSlack (horzTilePtr);
        tileWidth = Tp_GetDensityOfTile (horzTilePtr);
        *valPtr = (negSlack-tileWidth < *valPtr) ? negSlack-tileWidth : *valPtr;
    }
    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * FindMaxPosSlack() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
FindMaxPosSlack(tilePtr, valPtr)

register Tile    *tilePtr;
int                *valPtr;

{
    int            posSlack;

    if (Tp_IsCellInstTile (tilePtr) || Tp_IsBottleneckTile(tilePtr) ||
        Tp_IsDCellTile (tilePtr))
    {
        posSlack = SoGetHorzPosSlack (tilePtr);
        *valPtr = (posSlack > *valPtr) ? posSlack : *valPtr;
    }
    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * FindMinNonZeroSlack() --
 *
 *    finds minimum difference between length of 1st and 2nd critical paths
 *
 * Results:
 *    returns the value of the difference
 *
 * Side Effects:
 *    None.
 * ----------------------------------------------------------------------------
 */

static int
FindMinNonZeroSlack(tilePtr, diffPtr)

register Tile    *tilePtr;
int                *diffPtr;

{
    int            slack;

    if (Tp_IsCellInstTile(tilePtr) || Tp_IsBottleneckTile(tilePtr) ||
        Tp_IsDCellTile(tilePtr))
    {
        slack = SoGetHorzNegSlack(tilePtr) - SoGetHorzPosSlack (tilePtr);
        *diffPtr = (slack < *diffPtr && slack != 0) ? slack : *diffPtr;
    }

    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * ComputeHorzNegSlack() --
 *
 *    Compute horz negative slack for one tile.
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *
 * Side effects:
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus
ComputeHorzNegSlack (horzTilePtr)

register Tile    *horzTilePtr;
{
    Rect        planeSearchRect;
    int            tileWidth, minRightVal;

    minRightVal = chipX;
    if (Tp_IsCellInstTile (horzTilePtr))
    {
        planeSearchRect.r_xbot = RIGHT (horzTilePtr) + 1;
        planeSearchRect.r_ybot = BOTTOM (horzTilePtr);
        planeSearchRect.r_xtop = RIGHT (horzTilePtr) + 2;
        planeSearchRect.r_ytop = TOP (horzTilePtr);
        curTileRight = RIGHT (horzTilePtr);
        if (TR (horzTilePtr) != rightPseudoTilePtr)
        {
            (void) TiSrArea(NIL(Tile),
                    horzPlanePtr,
                    &planeSearchRect,
                    FindMinRightVal,
                    (ClientData) &minRightVal);
        }
        SoPutHorzNegSlack (minRightVal, horzTilePtr);
    }
    else if (Tp_IsBottleneckTile (horzTilePtr))
    {
        if (TR(horzTilePtr) != rightPseudoTilePtr)
        {
            /* TR(tp) is either cellInst tile or dummyCell tile */
            if (Tp_IsDCellTile (TR(horzTilePtr)))
            {
                tileWidth = 0;
            }
            else
            {
                tileWidth = RIGHT (TR(horzTilePtr)) - LEFT (TR(horzTilePtr));
            }
            minRightVal = SoGetHorzNegSlack (TR (horzTilePtr));
            SoPutHorzNegSlack (minRightVal - tileWidth, horzTilePtr);
        }
        else
        {
            SoPutHorzNegSlack (chipX, horzTilePtr);
        }
    }
    else if (Tp_IsDCellTile (horzTilePtr))
    {
        if (! Tp_IsBottleneckTile (TR (horzTilePtr)))
        {
            Bear_PrintFatalErr ("ComputeHorzNegSlack() failed, tile adjacent to DUMMY_CELL tile is not a Bottleneck!");
        }
        /* TR(tp) is a bottleneck tile. */
        tileWidth = Tp_GetDensityOfTile (TR (horzTilePtr));
        minRightVal = SoGetHorzNegSlack (TR (horzTilePtr));
        SoPutHorzNegSlack (minRightVal - tileWidth, horzTilePtr);
    }
    else
    {
        Bear_PrintFatalErr("ComputeHorzNegSlack() should not happen, call maintainer. ");
    }

    return (BEAR_CONTINUE_SEARCH); 
}



/*
 * ----------------------------------------------------------------------------
 *
 * PushHorzTile() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static int
PushHorzTile(tilePtr, diffPtr)

register Tile    *tilePtr;
int                *diffPtr;
{
    register LinkedTile        *lnkTilePtr;

    if (Tp_IsCellInstTile (tilePtr) &&
        SoGetHorzPosSlack (tilePtr) == SoGetHorzNegSlack (tilePtr))
    {
        /* Add this tile into `soH1TileList'. */
        MALLOC(LinkedTile *, lnkTilePtr, sizeof (LinkedTile));
        lnkTilePtr->tilePtr = tilePtr;
        lnkTilePtr->nextPtr = soH1TileList;
        soH1TileList = lnkTilePtr;
    } 
    
    if (Tp_IsCellInstTile (tilePtr) &&
        (SoGetHorzNegSlack (tilePtr) == SoGetHorzPosSlack (tilePtr) + *diffPtr))
    {
        /* Add this tile into `soH2TileList'. */
        MALLOC(LinkedTile *, lnkTilePtr, sizeof (LinkedTile));
        lnkTilePtr->tilePtr = tilePtr;
        lnkTilePtr->nextPtr = soH2TileList;
        soH2TileList = lnkTilePtr;
    }
        
    return (MAGIC_CONTINUE_SEARCH);
}



/* ----------------------------------------------------------------------------
 *
 * SoFindHorzCritTiles() --
 *
 *     The main routine to find all cellInst tiles which correspond to soft
 *     modules and lie on horz critical path(s).
 *
 * Results:
 *     Fills in horzTileList.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

int 
SoFindHorzCritTiles()

{
    Tile        *horzTilePtr;
    Rect        searchRect, *horzPlaneBndPtr;
    int            diff = INFINITY;
    int            dimX;

    /*
    printf ("\n");
    printf ("Enter SoFindHorzCritTiles - \n"); 
    fflush (stdout);
    */

    stack = MakeDeque();
    chipX = 0;

    horzPlanePtr = Tp_GetHorzPlane (soGlobalPlaneDataPtr);
    horzPlaneBndPtr = Tp_GetHorzPlaneBnd (soGlobalPlaneDataPtr);
    leftPseudoTilePtr = horzPlanePtr->pl_left;
    rightPseudoTilePtr = horzPlanePtr->pl_right;

    /*
    printf ("Compute Horizontal Slack ...\n");
    fflush (stdout);
    */

    /* Compute positive horz slack of tiles */
    (void) TiSrArea((Tile *) NULL, 
             horzPlanePtr, 
             horzPlaneBndPtr,
             ComputeHorzPosSlack, 
             NIL (ClientData));

    /* Compute horz chip dimension. */
    searchRect.r_xbot = horzPlaneBndPtr->r_xtop - 2;
    searchRect.r_ybot = horzPlaneBndPtr->r_ybot;
    searchRect.r_xtop = horzPlaneBndPtr->r_xtop - 1;
    searchRect.r_ytop = horzPlaneBndPtr->r_ytop;
    (void) TiSrArea((Tile *) NULL,
            horzPlanePtr,
            &searchRect,
            FindMaxPosSlack,
            (ClientData) &chipX);

    /* Compute negative horz slack of tiles */
    while (! IsEmptyDeque(stack))
    {
        horzTilePtr = (Tile *) PopDeque (stack);
        (void) ComputeHorzNegSlack (horzTilePtr);
    }

    /* Compute difference between length of first and second critical paths */
    (void) TiSrArea((Tile *) NULL,
            horzPlanePtr,
            horzPlaneBndPtr,
            FindMinNonZeroSlack,
            (ClientData) &diff);

    if (diff == 0)
    {
        Bear_PrintFatalErr ("SoFindHorzCritTiles() failed, diff is zero");
    }

    diff = (diff == INFINITY) ? 0 : diff;

    /* CONS all cellInst tiles which lie on first or second critical horz 
       paths and correspond to soft modules into soH1TileList, soH2TileList. */

    (void) TiSrArea((Tile *) NULL,
            horzPlanePtr,
            horzPlaneBndPtr,
            PushHorzTile,
            (ClientData) &diff);
    /*
    printf ("Exit So_FindHorzCritTiles - \n"); 
    fflush (stdout);
    */

    DeleteDeque (stack);
    dimX = chipX;

    return (dimX);
}
