
/* soMain.c --
 *
 *     Massoud Pedram
 *
 * This module performs shape optimization on a given placement.
 *
 * Copyright(C) 1988 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/shapeOptimizer/soMain.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "soInt.h"
#include "soPrivate.h"
#include "soUtil.h"

/* global variables */
char soDebug;
bool soQuit;
int  soCount;
double      soChipAreaArray[HIST_PARM];
LinkedTile  *soH1TileList, *soH2TileList;
LinkedTile  *soV1TileList, *soV2TileList;
PlaneData   *soGlobalPlaneDataPtr;
CellMaster  *soGlobalParentCellPtr;
double      soInitChipArea, soFinalChipArea;



void
So_Init()
{
    int i;

    soCount = 1; 
    soInitChipArea = 0;
    for(i = 0; i < HIST_PARM; i++) soChipAreaArray[i] = 0;
}


static SlackInfo *
MakeSlackInfo()
{
    SlackInfo *slackInfoPtr;
    MALLOC(SlackInfo *, slackInfoPtr, sizeof(SlackInfo));
    slackInfoPtr->horzSlack.pos = slackInfoPtr->horzSlack.neg = -1;
    slackInfoPtr->vertSlack.pos = slackInfoPtr->vertSlack.neg = -1;
    return (slackInfoPtr);
}

static int
AttachSlackInfoToTile(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
    TileProp    *tilePropPtr;

    if(Tp_IsCellInstTile(tilePtr) || Tp_IsBottleneckTile(tilePtr) ||
        Tp_IsDCellTile(tilePtr)) {
        tilePropPtr = (TileProp *) tilePtr->ti_client;
        tilePropPtr->clientData = (ClientData) MakeSlackInfo();
    }

    return (MAGIC_CONTINUE_SEARCH);
}



static int
FreeSIOnBNTile(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
    TileProp    *tilePropPtr;

    if(Tp_IsBottleneckTile(tilePtr)) {
        tilePropPtr = (TileProp *) tilePtr->ti_client;
        FREE((char *) tilePropPtr->clientData);
    }

    return (MAGIC_CONTINUE_SEARCH);
}


static int
FreeSIOnCITile(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
    TileProp    *tilePropPtr;

    if(Tp_IsCellInstTile(tilePtr)) {
        tilePropPtr = (TileProp *) tilePtr->ti_client;
        FREE((char *) tilePropPtr->clientData);
    }

    return (MAGIC_CONTINUE_SEARCH);
}



static void
AttachHorzSIs(parentCellPtr)
CellMaster        *parentCellPtr;
{

    PlaneData    *planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetHorzPlane(planeDataPtr),
        Tp_GetHorzPlaneBnd(planeDataPtr),
        AttachSlackInfoToTile,
        (ClientData) NULL);
}

static void
FreeHorzSIsOnBNs(parentCellPtr)
CellMaster        *parentCellPtr;
{

    PlaneData    *planeDataPtr=Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetHorzPlane(planeDataPtr),
        Tp_GetHorzPlaneBnd(planeDataPtr),
        FreeSIOnBNTile,
        (ClientData) NULL);
}


static void
AttachVertSIs(parentCellPtr)
CellMaster        *parentCellPtr;
{

    PlaneData    *planeDataPtr=Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetVertPlane(planeDataPtr),
        Tp_GetVertPlaneBnd(planeDataPtr),
        AttachSlackInfoToTile,
        (ClientData) NULL);
}

static void
FreeVertSIsOnBNs(parentCellPtr)
CellMaster        *parentCellPtr;
{

    PlaneData    *planeDataPtr=Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetVertPlane(planeDataPtr),
        Tp_GetVertPlaneBnd(planeDataPtr),
        FreeSIOnBNTile,
        (ClientData) NULL);
}

static void
FreeSIsOnCIs(parentCellPtr)
CellMaster        *parentCellPtr;
{

    PlaneData    *planeDataPtr=Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetHorzPlane(planeDataPtr),
        Tp_GetHorzPlaneBnd(planeDataPtr),
        FreeSIOnCITile,
        (ClientData) NULL);

    (void) TiSrArea(NIL(Tile),
        Tp_GetVertPlane(planeDataPtr),
        Tp_GetVertPlaneBnd(planeDataPtr),
        FreeSIOnCITile,
        (ClientData) NULL);
}



/* ARGSUSED */
static int
PrintSIOnCITile(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
    TileProp    *tilePropPtr;
    SlackInfo   *slackInfoPtr;
    CellInst    *cellInstPtr;
    Tile        *vertTilePtr;
    TileInfo    *tileInfoPtr;

    if(Tp_IsCellInstTile(tilePtr)) {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        tileInfoPtr =(TileInfo *) cellInstPtr->clientData;
        vertTilePtr = tileInfoPtr->vPtr;
        if(tileInfoPtr == NIL(TileInfo)) {
            Bear_PrintFatalErr("PrintSIOnCITile() failed, nil tileInfo");
        }
        printf("\n");
        tilePropPtr = (TileProp *) tilePtr->ti_client;
        slackInfoPtr =(SlackInfo *) tilePropPtr->clientData;
        printf("cell instance name: %s, plane type: SO_HORZ \n",
            Mac_GetNameOfCellInst(Tp_GetCellInstOfTile(tilePtr)));
        printf("HorzPosSlack: %d, HorzNegSlack: %d \n",
            slackInfoPtr->horzSlack.pos, slackInfoPtr->horzSlack.neg);
        tilePropPtr =(TileProp *) vertTilePtr->ti_client;
        slackInfoPtr =(SlackInfo *) tilePropPtr->clientData;
        printf("VertPosSlack: %d, VertNegSlack: %d \n",
            slackInfoPtr->vertSlack.pos, slackInfoPtr->vertSlack.neg);
    } 
    return (MAGIC_CONTINUE_SEARCH);
}



static int
PrintSIOnBNTile(tilePtr, planePtr)
Tile        *tilePtr;
Plane       *planePtr;
{
    TileProp    *tilePropPtr;
    SlackInfo   *slackInfoPtr;

    if(Tp_IsBottleneckTile(tilePtr)) {
        printf("\n");
        tilePropPtr = (TileProp *) tilePtr->ti_client;
        slackInfoPtr =(SlackInfo *) tilePropPtr->clientData;
        if(planePtr == Tp_GetHorzPlane(soGlobalPlaneDataPtr)) {
            printf("Bottleneck plane tile: SO_HORZ \n");
            printf("HorzPosSlack: %d, HorzNegSlack: %d \n",
                slackInfoPtr->horzSlack.pos, slackInfoPtr->horzSlack.neg);
        } else {
            printf("Bottleneck plane tile: SO_VERT \n");
            printf("VertPosSlack: %d, VertNegSlack: %d \n",
                slackInfoPtr->vertSlack.pos, slackInfoPtr->vertSlack.neg);
        }
    }

    return (MAGIC_CONTINUE_SEARCH);
}


static void
PrintSIsOnCIs(parentCellPtr)
CellMaster        *parentCellPtr;
{
    PlaneData    *planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetHorzPlane(planeDataPtr),
        Tp_GetHorzPlaneBnd(planeDataPtr),
        PrintSIOnCITile,
        (ClientData) NULL);
}

static void
PrintHorzSIsOnBNs(parentCellPtr)
CellMaster        *parentCellPtr;
{
    PlaneData    *planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    Plane        *hPlanePtr = Tp_GetHorzPlane(planeDataPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetHorzPlane(planeDataPtr),
        Tp_GetHorzPlaneBnd(planeDataPtr),
        PrintSIOnBNTile,
        (ClientData) hPlanePtr);
}

static void
PrintVertSIsOnBNs(parentCellPtr)
CellMaster        *parentCellPtr;
{
    PlaneData    *planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    Plane        *vPlanePtr = Tp_GetVertPlane(planeDataPtr);
    (void) TiSrArea(NIL(Tile),
        Tp_GetVertPlane(planeDataPtr),
        Tp_GetVertPlaneBnd(planeDataPtr),
        PrintSIOnBNTile,
        (ClientData) vPlanePtr);
}


static TileInfo *
MakeTileInfo()
{
    TileInfo *tileInfoPtr;

    MALLOC(TileInfo *, tileInfoPtr, sizeof(TileInfo));
    tileInfoPtr->hPtr = NIL(Tile);
    tileInfoPtr->vPtr = NIL(Tile);

    return (tileInfoPtr);
}

/* ARGSUSED */
static BearStatus
AttachTileInfoToCell(cellInstPtr, clientData)
CellInst    *cellInstPtr;
ClientData  clientData;
{
    cellInstPtr->clientData = (ClientData) MakeTileInfo();
    return (BEAR_CONTINUE_SEARCH);
}

/* ARGSUSED */
static BearStatus
FreeTileInfoOnCell(cellInstPtr, clientData)
CellInst    *cellInstPtr;
ClientData  clientData;
{
    FREE((char *) cellInstPtr->clientData);
    cellInstPtr->clientData = (ClientData) NULL;

    return (BEAR_CONTINUE_SEARCH);
}



    
/*
 * ----------------------------------------------------------------------
 *
 * So_ShapeOptimize() --
 *
 *   Main proc which performs shape optimization on a given placement. 
 *   
 * Results:
 *   A shape optimized placement.
 *  
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------
 */

SoStatus
So_ShapeOptimize(parentCellPtr, rsDataPtr, parmsPtr)
CellMaster      *parentCellPtr;
ResizeData      *rsDataPtr;
So_UserParms    *parmsPtr; 
{
    CritTileInfo    critTileInfo;
    Plane           *horzPlanePtr, *vertPlanePtr;
    Rect            bBox, newBox;
    SoStatus        soStatus = SO_NOT_FINISHED;
    int             i, dimX, dimY, xDim, yDim;
    double          diff, chipArea=1;
    static int      chipX, chipY; /* must be static */
    /* char            string[4]; */

    /* Initializations */
    soDebug = 'n';
    soQuit = FALSE;
    soH1TileList = soV1TileList = NIL(LinkedTile);
    soH2TileList = soV2TileList = NIL(LinkedTile);
    soGlobalParentCellPtr = parentCellPtr;
    soGlobalPlaneDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    rsDataPtr->move.p_x = rsDataPtr->move.p_y = 0; 
    rsDataPtr->resize.p_x = rsDataPtr->resize.p_y = 0;
    rsDataPtr->bnDensity = 0;
    rsDataPtr->bnTilePtr = NIL(Tile);

    if(! Mac_IsGlobalRoutingDone(soGlobalParentCellPtr)) {
        Bear_Print("Do global routing and try again.");
        soQuit = TRUE; soStatus = SO_FINISHED;
        goto finish;
    }

    if (Gre_CheckConnForAllNets(soGlobalParentCellPtr, FALSE) == BEAR_ERROR) {
        Bear_PrintErr("Shape optimizer aborted (disconnected nets).");
        soQuit = TRUE; soStatus = SO_FINISHED;
        goto finish;
    }

    /*
    printf("\nDo you want to continue?(y/n) ");
    scanf("%1s",string);
    if(strcmp(string, "n") == 0) {
        soQuit = TRUE; soStatus = SO_FINISHED;
        goto finish;
    }
    */

    if(soCount > parmsPtr->maxCnt) {
        soQuit = TRUE; soStatus = SO_FINISHED;
        goto finish;
    }

    Mac_ItrChildrenOfCell(soGlobalParentCellPtr, AttachTileInfoToCell,
                   (ClientData) NULL);

    if(parmsPtr->algorithm == ONE_D) {
         if(parmsPtr->option == BEST_SLACK) {
            printf("algorithm: %s, option: %s, iteration: %d --- \n", 
                "ONE_D", "BEST_SLACK", soCount);
        } else if(parmsPtr->option == HALF_SLACK) {
            printf("algorithm: %s, option: %s, iteration: %d --- \n", 
                "ONE_D", "HALF_SLACK", soCount);
        } else {
            printf("algorithm: %s, option: %s, iteration: %d --- \n", 
                "ONE_D", "FULL_SLACK", soCount);
        }
    } else {
         if(parmsPtr->option == BEST_SLACK) {
            printf("algorithm: %s, option: %s, iteration: %d --- \n", 
                "TWO_D", "BEST_SLACK", soCount);
        } else if(parmsPtr->option == HALF_SLACK) {
            printf("algorithm: %s, option: %s, iteration: %d --- \n", 
                "TWO_D", "HALF_SLACK", soCount);
        } else {
            printf("algorithm: %s, option: %s, iteration: %d --- \n", 
                "TWO_D", "FULL_SLACK", soCount);
        }
    }

    horzPlanePtr = Tp_GetHorzPlane(soGlobalPlaneDataPtr);
    Tm_HUnify(soGlobalParentCellPtr);
    (void) Gs_CalcCapDenOfTilePlane(soGlobalParentCellPtr);
    AttachHorzSIs(soGlobalParentCellPtr);
    chipX = SoFindHorzCritTiles();
    /* PrintHorzSIsOnBNs(soGlobalParentCellPtr); */
    FreeHorzSIsOnBNs(soGlobalParentCellPtr);
    Tm_HSplit(soGlobalParentCellPtr);

    vertPlanePtr = Tp_GetVertPlane(soGlobalPlaneDataPtr);
    Tm_VUnify(soGlobalParentCellPtr);
    (void) Gs_CalcCapDenOfTilePlane(soGlobalParentCellPtr);
    AttachVertSIs(soGlobalParentCellPtr);
    chipY = SoFindVertCritTiles();
    /* PrintVertSIsOnBNs(soGlobalParentCellPtr); */
    FreeVertSIsOnBNs(soGlobalParentCellPtr);
    Tm_VSplit(soGlobalParentCellPtr);

    printf("Chip Width: %d, Chip Height: %d \n", chipX, chipY);
    chipArea = ((double) chipX) *((double) chipY);
    soInitChipArea = (soCount == 1) ? chipArea : soInitChipArea;
    if(parmsPtr->algorithm == TWO_D) {
        soQuit = TRUE;
        for(i=0; i < HIST_PARM; i++) {
            diff = (chipArea == 0) ? 0 :
                100.0*(chipArea-soChipAreaArray[i])/chipArea;
            diff = (diff >= 0) ? diff :(-1 * diff);
            if(diff < parmsPtr->stopParm) continue;
            soQuit = FALSE;
            break; 
        }
        for(i=0; i < HIST_PARM-1; i++)
            soChipAreaArray[i] = soChipAreaArray[i+1];
        soChipAreaArray[HIST_PARM-1] = chipArea;
    }
    critTileInfo = SoGetCritTileInfo(parmsPtr);

    /* PrintSIsOnCIs(soGlobalParentCellPtr); */
    FreeSIsOnCIs(soGlobalParentCellPtr);

    if(critTileInfo.slk <= parmsPtr->gridSize || 
        critTileInfo.tilePtr == NIL(Tile)) soQuit = TRUE;
    if(! soQuit) {
        rsDataPtr->cellInstPtr = Tp_GetCellInstOfTile(critTileInfo.tilePtr);
        bBox = *Mac_GetBBoxOfCell(rsDataPtr->cellInstPtr);
        GeoTransRect(&(rsDataPtr->cellInstPtr->transform), &bBox, &newBox);
        dimX = newBox.r_ur.p_x - newBox.r_ll.p_x;
        dimY = newBox.r_ur.p_y - newBox.r_ll.p_y;
        if(dimX < 0) dimX *= -1; if(dimY < 0) dimY *= -1;
        if(critTileInfo.planeDir == SO_HORZ) {
            SoPrintCellInstTile(critTileInfo.tilePtr, horzPlanePtr);
            xDim = (int)((double) -(dimX * critTileInfo.slk) / 
                   (double)(dimY + critTileInfo.slk));
            if (-xDim >= dimX)
                Bear_PrintFatalErr("So_ShapeOptimize() failed, programming error.");
        } else {
            SoPrintCellInstTile(critTileInfo.tilePtr, vertPlanePtr);
            yDim = (int)((double) -(dimY * critTileInfo.slk) / 
                   (double)(dimX + critTileInfo.slk));
            if (-yDim >= dimY)
                Bear_PrintFatalErr("So_ShapeOptimize() failed, programming error.");
        }
        soStatus = SoShapeFunc(&critTileInfo, rsDataPtr, parmsPtr);
        if(rsDataPtr->resize.p_x == 0 && rsDataPtr->resize.p_y == 0) {
            soQuit = TRUE; soStatus = SO_FINISHED;
        }
    } else {
        soStatus = SO_FINISHED; 
    }
    Tp_FreeTileList(soH1TileList); Tp_FreeTileList(soV1TileList);
    Tp_FreeTileList(soH2TileList); Tp_FreeTileList(soV2TileList);

    Mac_ItrChildrenOfCell(soGlobalParentCellPtr, FreeTileInfoOnCell, 
        (ClientData) NULL);

    soCount++;

finish:

    if(soStatus == SO_FINISHED) {
        soFinalChipArea = ((double) chipX) *((double) chipY);
    }

    return (soStatus);
}


/*
 * ----------------------------------------------------------------------
 *
 * So_ResizeCell() --
 *
 *   Resizes cell according to resizeData.
 *   
 * Results:
 *  
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------
 */

void
So_ResizeCell(rsDataPtr, soStatus)
ResizeData  *rsDataPtr;
SoStatus    soStatus;
{
    switch(soStatus) {
    case SO_NOT_FINISHED:
        printf("move.x: %d, move.y: %d \n", 
            rsDataPtr->move.p_x, rsDataPtr->move.p_y);
        printf("resize.x: %d, resize.y: %d \n",
        rsDataPtr->resize.p_x, rsDataPtr->resize.p_y);
        Tp_ResizeCellInst(rsDataPtr->move.p_x, rsDataPtr->move.p_y, 
            rsDataPtr->resize.p_x, rsDataPtr->resize.p_y, 
            rsDataPtr->cellInstPtr);
        break;
    case SO_FINISHED:
        printf ("\nShape optimization done.\n");
        if (soInitChipArea > 0 && soFinalChipArea > 0) {
            printf ("Reduced Chip Area by %lf percent.\n\n",
               (100*(soInitChipArea-soFinalChipArea)/soInitChipArea));
        }
        break;
    default:
        break;
    }
}


