
/* soVertSlack.c --
 *
 *     Massoud Pedram
 *     (Professor E. S. Kuh)
 *
 * This module computes vertical slacks and vertical critical paths.
 *
 * Copyright (C) 1988 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/shapeOptimizer/soVertSlack.c,v 1.1.1.1 2002/12/31 15:51:09 saj Exp $";
#endif  not lint

#include "soInt.h"
#include "soUtil.h"

/* ---------- Local Global Variables ---------- */
static Plane        *vertPlanePtr;
static Deque        *stack = NIL(Deque);
static int            curTileLeft, curTileRight, chipY;
static Tile            *leftPseudoTilePtr, *rightPseudoTilePtr;


/*
 * ----------------------------------------------------------------------------
 *
 * FindMaxNodeVal() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
FindMaxLeftVal(vertTilePtr, valPtr)

register Tile    *vertTilePtr;
int                *valPtr;

{
    int        posSlack;

    if (RIGHT(vertTilePtr) == curTileLeft  && Tp_IsBottleneckTile(vertTilePtr))
    {
        posSlack = SoGetVertPosSlack(vertTilePtr);
        *valPtr = (posSlack > *valPtr) ? posSlack : *valPtr;
    }
    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * ComputeVertPosSlack() --
 *
 *    Compute vert positive slack for all tiles.
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *
 * Side effects:
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus
ComputeVertPosSlack (vertTilePtr, clientData)

register Tile    *vertTilePtr;
ClientData    clientData;

{
    Rect        planeSearchRect;
    int        tileWidth, maxLeftVal;
    CellInst    *cellInstPtr;
    TileInfo    *tileInfoPtr;

    if (Tp_IsCellInstTile (vertTilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile (vertTilePtr);
        PushDeque(stack, (ClientData) vertTilePtr);
        maxLeftVal = 0;
        tileWidth = RIGHT (vertTilePtr) - LEFT (vertTilePtr);
        planeSearchRect.r_xbot = LEFT (vertTilePtr) - 2;
        planeSearchRect.r_ybot = BOTTOM (vertTilePtr);
        planeSearchRect.r_xtop = LEFT (vertTilePtr) - 1;
        planeSearchRect.r_ytop = TOP (vertTilePtr);
        curTileLeft = LEFT (vertTilePtr);

        if (BL(vertTilePtr)  != leftPseudoTilePtr)
        {
            (void) TiSrArea(NIL(Tile),
                    vertPlanePtr,
                    &planeSearchRect,
                    FindMaxLeftVal,
                    (ClientData) &maxLeftVal);
        }
        SoPutVertPosSlack (maxLeftVal + tileWidth, vertTilePtr);
        tileInfoPtr = (TileInfo *) cellInstPtr->clientData;
        tileInfoPtr->vPtr = vertTilePtr;
    }
    else if (Tp_IsBottleneckTile (vertTilePtr))
    {
        PushDeque(stack, (ClientData) vertTilePtr);
        tileWidth = Tp_GetDensityOfTile(vertTilePtr);
        maxLeftVal = (BL(vertTilePtr) != leftPseudoTilePtr) ?
                SoGetVertPosSlack (BL(vertTilePtr)) : 0;
        SoPutVertPosSlack (maxLeftVal + tileWidth, vertTilePtr);
    }
    else if (Tp_IsDCellTile (vertTilePtr))
    {
        PushDeque(stack, (ClientData) vertTilePtr);
        tileWidth = 0;
        /* tileWidth = RIGHT (vertTilePtr) - LEFT (vertTilePtr); */
        maxLeftVal = (BL(vertTilePtr) != leftPseudoTilePtr) ?
        SoGetVertPosSlack (BL(vertTilePtr)) : 0;
                SoPutVertPosSlack (maxLeftVal + tileWidth, vertTilePtr);
    }

    return (BEAR_CONTINUE_SEARCH); 
}



/*
 * ----------------------------------------------------------------------------
 *
 * FindMinRightVal() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
FindMinRightVal(vertTilePtr, valPtr)

register Tile    *vertTilePtr;
int                *valPtr;

{
    int        negSlack, tileWidth;

    if (LEFT(vertTilePtr) == curTileRight  && Tp_IsBottleneckTile(vertTilePtr))
    {
        negSlack = SoGetVertNegSlack (vertTilePtr);
        tileWidth = Tp_GetDensityOfTile (vertTilePtr);
        *valPtr = (negSlack-tileWidth < *valPtr) ? negSlack-tileWidth : *valPtr;
    }
    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * FindMaxPosSlack() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
FindMaxPosSlack(tilePtr, valPtr)

register Tile    *tilePtr;
int                *valPtr;

{
    int            posSlack;

    if (Tp_IsCellInstTile(tilePtr) || Tp_IsBottleneckTile(tilePtr) ||
        Tp_IsDCellTile(tilePtr))
    {
        posSlack = SoGetVertPosSlack (tilePtr);
        *valPtr = (posSlack > *valPtr) ? posSlack : *valPtr;
    }
    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * FindMinNonZeroSlack() --
 *
 *    finds minimum difference between length of 1st and 2nd critical paths
 *
 * Results:
 *    returns the value of the difference
 *
 * Side Effects:
 *    None.
 * ----------------------------------------------------------------------------
 */

static int
FindMinNonZeroSlack(tilePtr, diffPtr)

register Tile   *tilePtr;
int             *diffPtr;

{
    int         slack;

    if (Tp_IsCellInstTile (tilePtr) || Tp_IsBottleneckTile(tilePtr) ||
        Tp_IsDCellTile (tilePtr))
    {
        slack = SoGetVertNegSlack (tilePtr) - SoGetVertPosSlack (tilePtr);
        *diffPtr = (slack < *diffPtr && slack != 0) ? slack : *diffPtr;
    }

    return (MAGIC_CONTINUE_SEARCH);
}


/*
 * ----------------------------------------------------------------------------
 *
 * ComputeVertNegSlack() --
 *
 *    Compute vert negative slack for one tile.
 *
 * Results:
 *    Return BEAR_CONTINUE_SEARCH to keep TiSrArea from 
 *    aborting the search.
 *
 * Side effects:
 *
 * ----------------------------------------------------------------------------
 */

/* ARGSUSED */
static BearStatus
ComputeVertNegSlack(vertTilePtr)

register Tile    *vertTilePtr;
{
    Rect        planeSearchRect;
    int            tileWidth, minRightVal;

    minRightVal = chipY;
    if (Tp_IsCellInstTile(vertTilePtr))
    {
        planeSearchRect.r_xbot = RIGHT (vertTilePtr) + 1;
        planeSearchRect.r_ybot = BOTTOM (vertTilePtr);
        planeSearchRect.r_xtop = RIGHT (vertTilePtr) + 2;
        planeSearchRect.r_ytop = TOP (vertTilePtr);
        curTileRight = RIGHT (vertTilePtr);
        if (TR(vertTilePtr) != rightPseudoTilePtr)
        {
            (void) TiSrArea(NIL(Tile),
                    vertPlanePtr,
                    &planeSearchRect,
                    FindMinRightVal,
                    (ClientData) &minRightVal);
        }
        SoPutVertNegSlack (minRightVal, vertTilePtr);
    }
    else if (Tp_IsBottleneckTile (vertTilePtr))
    {
        if (TR (vertTilePtr) != rightPseudoTilePtr)
        {
            /* TR (tp) is either cellInst tile or dummyCell tile */
            if (Tp_IsDCellTile (TR(vertTilePtr)))
            {
                tileWidth = 0;
            }
            else
            {
                tileWidth = RIGHT (TR(vertTilePtr)) - LEFT (TR(vertTilePtr));
            }
            minRightVal = SoGetVertNegSlack (TR (vertTilePtr));
            SoPutVertNegSlack (minRightVal - tileWidth, vertTilePtr);
        }
        else
        {
            SoPutVertNegSlack (chipY, vertTilePtr);
        }
    }
    else if (Tp_IsDCellTile (vertTilePtr))
    {
        if (! Tp_IsBottleneckTile (TR (vertTilePtr)))
        {
            Bear_PrintFatalErr ("ComputeVertNegSlack() failed, tile adjacent to DUMMY_CELL tile is not a Bottleneck!");
        }
        /* TR(tp) is a bottleneck tile. */
        tileWidth = Tp_GetDensityOfTile (TR (vertTilePtr));
        minRightVal = SoGetVertNegSlack (TR (vertTilePtr));
        SoPutVertNegSlack (minRightVal - tileWidth, vertTilePtr);
    }
    else
    {
        Bear_PrintFatalErr("ComputeVertNegSlack() should not happen, call maintainer. ");
    }

    return (BEAR_CONTINUE_SEARCH); 
}



/*
 * ----------------------------------------------------------------------------
 *
 * PushVertTile() --
 *
 * Results:
 *
 * Side Effects:
 *
 * ----------------------------------------------------------------------------
 */

static int
PushVertTile(tilePtr, diffPtr)

register Tile    *tilePtr;
int                *diffPtr;
{
    register LinkedTile        *lnkTilePtr;

    if (Tp_IsCellInstTile(tilePtr) && 
        (SoGetVertPosSlack(tilePtr) == SoGetVertNegSlack(tilePtr)))
    {
        /* Add this tile into `soV1TileList'. */
        MALLOC(LinkedTile *, lnkTilePtr, sizeof (LinkedTile));
        lnkTilePtr->tilePtr = tilePtr;
        lnkTilePtr->nextPtr = soV1TileList;
        soV1TileList = lnkTilePtr;
    }

    if (Tp_IsCellInstTile (tilePtr) && 
        (SoGetVertNegSlack (tilePtr) == SoGetVertPosSlack (tilePtr) + *diffPtr))
    {
        /* Add this tile into `soV2TileList'. */
        MALLOC(LinkedTile *, lnkTilePtr, sizeof (LinkedTile));
        lnkTilePtr->tilePtr = tilePtr;
        lnkTilePtr->nextPtr = soV2TileList;
        soV2TileList = lnkTilePtr;
    }

    return (MAGIC_CONTINUE_SEARCH);
}



/* ----------------------------------------------------------------------------
 *
 * SoFindVertCritTiles() --
 *
 *     The main routine to find all cellInst tiles which correspond to soft
 *     modules and lie on vert critical path(s).
 *
 * Results:
 *     Fills in soV1TileList, soV2TileList.
 *
 * Side effects:
 *     None.
 *
 * ----------------------------------------------------------------------------
 */

int
SoFindVertCritTiles()

{
    Tile        *vertTilePtr;
    Rect        vertSearchRect, searchRect, *horzPlaneBndPtr;
    int            diff = INFINITY;
    int            dimY;

    /*
    printf ("\n");
    printf ("Enter SoFindVertCritTiles - \n"); 
    fflush (stdout);
    */

    stack = MakeDeque();
    chipY = 0;

    vertPlanePtr = Tp_GetVertPlane (soGlobalPlaneDataPtr);
    horzPlaneBndPtr = Tp_GetHorzPlaneBnd (soGlobalPlaneDataPtr);
    Tp_HRectToVRect (horzPlaneBndPtr, &vertSearchRect);
    leftPseudoTilePtr = vertPlanePtr->pl_left;
    rightPseudoTilePtr = vertPlanePtr->pl_right;

    /*
    printf ("Compute Vertical Slack ...\n");
    fflush (stdout);
    */

    /* Compute positive vert slack of tiles */
    (void) TiSrArea((Tile *) NULL, 
             vertPlanePtr, 
             &vertSearchRect,
             ComputeVertPosSlack, 
             NIL (ClientData));

    /* Compute vert chip dimension. */
    searchRect.r_xbot = vertSearchRect.r_xtop - 2;
    searchRect.r_ybot = vertSearchRect.r_ybot;
    searchRect.r_xtop = vertSearchRect.r_xtop - 1;
    searchRect.r_ytop = vertSearchRect.r_ytop;
    (void) TiSrArea((Tile *) NULL,
            vertPlanePtr,
            &searchRect,
            FindMaxPosSlack,
            (ClientData) &chipY);

    /* Compute negative vert slack of tiles */
    while (! IsEmptyDeque(stack))
    {
        vertTilePtr = (Tile *) PopDeque (stack);
        (void) ComputeVertNegSlack (vertTilePtr);
    }

    /* Compute difference between length of first and second critical paths */

    (void) TiSrArea((Tile *) NULL,
            vertPlanePtr,
            &vertSearchRect,
            FindMinNonZeroSlack,
            (ClientData) &diff);

    if (diff == 0)
    {
        Bear_PrintFatalErr ("SoFindVertCritTiles() failed, diff is zero");
    }

    diff = (diff == INFINITY) ? 0 : diff;

    /* CONS all cellInst tiles which lie on first or second critical vert 
       paths and correspond to soft modules into soV1TileList, soV2TileList. */

    (void) TiSrArea((Tile *) NULL,
            vertPlanePtr,
            &vertSearchRect,
            PushVertTile,
            (ClientData) &diff);
    
    /*
    printf ("Exit So_FindVertCritTiles - \n"); 
    fflush (stdout);
    */

    DeleteDeque (stack);
    dimY = chipY;

    return (dimY);
}
