
/* tmCUnify.c --
 *
 *         Sherry Solden
 *     (Professor E. S. Kuh)
 *
 * Commands that unify a circle.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmCUnify.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


typedef struct _LinkedPList
{
  PinsAroundJunction  *pinListsPtr;
  struct _LinkedPList *nextPtr;
} LinkedPList;

typedef struct _UnifyData
{
  bool               horzUnify;    /* if TRUE update HExtId else VExtId */
  Line               bnLine;
  Rect               domRect;
  SideDirection      sideDir;
  BearId             pinId;
  PseudoPin          *pinPtr1;
  PseudoPin          *pinPtr2;
  LinkedPList        *bnListsPtr;
  LinkedPList        *faceListsPtr;
  LinkedTile         *bnListTilePtr;
  PlaneData          *planeDataPtr;
  CellMaster         *parentCellPtr;
} UnifyData;


/* ---------- Local Procedures ---------- */
static int             tmVCircleUnify();
static int             tmHCircleUnify();
static BearStatus      UpdateNetIdInBNRegion();
static BearStatus      UpdateNetIdInFace();
static void            FixNetIdInCircle();
static BearStatus      FixNetIdInBNRegion();
static BearStatus      FixNetIdInFace();
static void            UpdateNetIdInList();
static BearStatus      AddPPinsToHList();
static BearStatus      AddPPinsToVList();
static BearStatus      FindPinsForBnLine();


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_VCircleUnify() --
 *
 *  This procedure looks at list of adjacent vertical tiles one by one, 
 *  and unifies the netIds across bottleneck tiles, lines and vert faces.
 *
 *  The circle is a set of monotonically increasing in x-direction set of
 *  vertical space tiles.  The circle must not contain any dummy cells.
 *
 *  Results:  All pins in the circle that belong to same net: have
 *            same extIds.  In addition, all ppins in the circle are
 *            also consistent.
 *
 * ----------------------------------------------------------------------------
*/

void
Tm_VCircleUnify(parentCellPtr, planeDataPtr, dmntRect)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
Rect       dmntRect;   /* vertical rect */
{
   UnifyData unifyData;

   unifyData.horzUnify = FALSE;
   unifyData.parentCellPtr = parentCellPtr;
   unifyData.planeDataPtr = planeDataPtr;
   unifyData.bnListsPtr = NIL(LinkedPList);
   unifyData.faceListsPtr = NIL(LinkedPList);
   unifyData.bnListTilePtr = NIL(LinkedTile);
   unifyData.sideDir = RIGHT_SIDE;
   (void) TiSrArea(NIL(Tile),
           Tp_GetVertPlane(planeDataPtr),
           &dmntRect,
           tmVCircleUnify,
           (ClientData) &unifyData);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tm_HCircleUnify() --
 *
 *  This procedure looks at list of adjacent horizontal tiles one by one, 
 *  and unifies the netIds across bottleneck tiles, lines and horz faces.
 *
 *  The circle is a set of monotonically decreasing in y-direction set of
 *  horizontal space tiles.  The circle must not contain any dummy cells.
 *
 *  Results:  All pins in the circle that belong to same net: have
 *            same extIds.  In addition, all ppins in the circle are
 *            also consistent.
 *
 * ----------------------------------------------------------------------------
 */

void
Tm_HCircleUnify(parentCellPtr, planeDataPtr, dmntRect)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
Rect       dmntRect;   /* horz rect */
{
   UnifyData unifyData;

   unifyData.horzUnify = TRUE;
   unifyData.parentCellPtr = parentCellPtr;
   unifyData.planeDataPtr = planeDataPtr;
   unifyData.bnListsPtr = NIL(LinkedPList);
   unifyData.faceListsPtr = NIL(LinkedPList);
   unifyData.bnListTilePtr = NIL(LinkedTile);
   unifyData.sideDir = TOP_SIDE;
   (void) TiSrArea(NIL(Tile),
           Tp_GetHorzPlane(planeDataPtr),
           &dmntRect,
           tmHCircleUnify,
           (ClientData) &unifyData);
}



static int
tmVCircleUnify(tilePtr, unifyData)
Tile        *tilePtr;
UnifyData   *unifyData;
{
    LinkedPList  *bnListsPtr;
    LinkedPList  *faceListsPtr;
    LinkedTile   *tileListPtr;
    WallJunction *junctPtr;
    WallSegment  *segPtr;
    Tile         *semiTile1Ptr;
    Tile         *semiTile2Ptr;
    Tile         *domTilePtr;
    Point        *domPointPtr;
    Point        vdomPoint;
    BNLineType   bnType;
    static bool  semiFound;


    if ((Tp_IsDominantTile(tilePtr)) && 
        (unifyData->bnListTilePtr == NIL(LinkedTile)))
    {
       /* Get all pins in first face region */
       semiFound = FALSE;
       Tp_VertTileToRect(unifyData->planeDataPtr, 
                 tilePtr, &(unifyData->domRect));
       MALLOC(LinkedPList *, faceListsPtr, sizeof (LinkedPList));
       faceListsPtr->nextPtr = unifyData->faceListsPtr;
       unifyData->faceListsPtr = faceListsPtr;
       unifyData->faceListsPtr->pinListsPtr = 
                 Tm_PinsAroundVFace(unifyData->parentCellPtr, 
                        unifyData->planeDataPtr, 
                        unifyData->domRect,
                        TRUE);
       /* Semi Bottleneck Tile */
       if (Tp_IsSBNTile(tilePtr))
       {
          semiTile1Ptr = tilePtr;
          semiFound = TRUE;
           }
       return (MAGIC_CONTINUE_SEARCH);
    }

    /* If tile is bottleneck tile */
        if (Tp_IsBottleneckTile(tilePtr))
    {
       semiFound = FALSE;
       MALLOC(LinkedTile *, tileListPtr, sizeof (LinkedTile));
       tileListPtr->tilePtr = tilePtr;
       tileListPtr->nextPtr = unifyData->bnListTilePtr;
       unifyData->bnListTilePtr = tileListPtr;

       /* Get all pins in bottleneck region */
       MALLOC(LinkedPList *, bnListsPtr, sizeof (LinkedPList));
       bnListsPtr->nextPtr = unifyData->bnListsPtr;
/*
 * code change : by george : on 24 oct 89
 *
        MALLOC(PinsAroundJunction *, bnListsPtr->pinListsPtr,
            sizeof (PinsAroundJunction));
 */
        bnListsPtr->pinListsPtr = Tm_MakePinsAroundJunction();
       unifyData->bnListsPtr = bnListsPtr;
       (void) Tm_GetPinsInVBottleNeckRegion(
                        unifyData->parentCellPtr,
                        unifyData->planeDataPtr, 
                            tilePtr, 
                            bnListsPtr->pinListsPtr);
           /* Add pseudo pins on left and right sides of bnTile */
           (void) Tp_ItrPPinsOfSide(tilePtr, 
                 TOP_RIGHT_SIDE,
                 AddPPinsToVList,
                 (ClientData) unifyData);
           (void) Tp_ItrPPinsOfSide(tilePtr, 
                 BOTTOM_LEFT_SIDE,
                 AddPPinsToVList,
                 (ClientData) unifyData);

       /* Update nets in bottleneck region by examining */
       /* pseudo pins on leftside of bnTile */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 BOTTOM_LEFT_SIDE,
                 UpdateNetIdInBNRegion, 
                 (ClientData) unifyData);

       /* Get Pins in next face region */
       segPtr = Tp_GetWallSegmentOfBnTile(tilePtr);
       junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
       domPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
       Tp_HPointToVPoint(domPointPtr, &vdomPoint);
       domTilePtr = TiSrPointNoHint(Tp_GetVertPlane(unifyData->planeDataPtr), 
                    &vdomPoint);
       Tp_VertTileToRect(unifyData->planeDataPtr, domTilePtr, 
                 &(unifyData->domRect));
       MALLOC(LinkedPList *, faceListsPtr, sizeof (LinkedPList));
       faceListsPtr->nextPtr = unifyData->faceListsPtr;
       unifyData->faceListsPtr = faceListsPtr;
       if ((Fpg_GetNorthSegOfWallJunct(junctPtr) == NIL(WallSegment)) &&
           (Fpg_GetSouthSegOfWallJunct(junctPtr) == NIL(WallSegment)))
       {
/*
 * code change : by george : on 24 oct 89
 *
           MALLOC(PinsAroundJunction *, 
              unifyData->faceListsPtr->pinListsPtr,
              sizeof (PinsAroundJunction));
 */
           unifyData->faceListsPtr->pinListsPtr = Tm_MakePinsAroundJunction();
           Tm_PinsAroundJunction(unifyData->parentCellPtr, 
                     unifyData->planeDataPtr, 
                     junctPtr, 
                     unifyData->faceListsPtr->pinListsPtr);
       }
       else
       {
           unifyData->faceListsPtr->pinListsPtr = 
                 Tm_PinsAroundVFace(unifyData->parentCellPtr, 
                        unifyData->planeDataPtr, 
                            unifyData->domRect,
                        TRUE);
           }

       /* Update nets in face region by examining the */
       /* pseudo pins on right-side of bnTile */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 TOP_RIGHT_SIDE,
                 UpdateNetIdInFace, 
                 (ClientData) unifyData);
       /* Needs a second pass - since the fixNets may cause changes */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 TOP_RIGHT_SIDE,
                 UpdateNetIdInFace, 
                 (ClientData) unifyData);
    }
    /* Semi Bottleneck Tile */
    else if (Tp_IsSBNTile(tilePtr))
    {
       if (semiFound == FALSE)
       {
          semiTile1Ptr = tilePtr;
          semiFound = TRUE;
       }
       else
       {
          semiTile2Ptr = tilePtr;
          if (!Tp_IsPairOfSBNTiles(&semiTile1Ptr, 
                       &semiTile2Ptr, 
                       &bnType))
          {
              /* Didn't find a match */
              semiTile1Ptr = tilePtr;
              return (MAGIC_CONTINUE_SEARCH);
          }
          if (bnType != TP_NULL_BNL)
          {
              MALLOC(LinkedTile *, tileListPtr, sizeof (LinkedTile));
              tileListPtr->tilePtr = tilePtr;
              tileListPtr->nextPtr = unifyData->bnListTilePtr;
              unifyData->bnListTilePtr = tileListPtr;

              MALLOC(LinkedPList *, bnListsPtr, sizeof (LinkedPList));
              bnListsPtr->nextPtr = unifyData->bnListsPtr;  
/*
 * code change : by george : on 24 oct 89
 *
              MALLOC(PinsAroundJunction *, bnListsPtr->pinListsPtr,
                  sizeof (PinsAroundJunction));
 */
              bnListsPtr->pinListsPtr = Tm_MakePinsAroundJunction();
              unifyData->bnListsPtr = bnListsPtr;

              /* Add pseudo pins on bnLine */
              (void) Tp_ItrPPinsOfBnLine(semiTile1Ptr, semiTile2Ptr, 
                      AddPPinsToVList,
                      (ClientData)unifyData);

              /* Get Pins in next face region */
              Tp_FindBnLine(semiTile1Ptr, semiTile2Ptr, 
                    &(unifyData->bnLine));
              (void) Tp_ItrVertBnLine(unifyData->planeDataPtr, 
                       FindPinsForBnLine, 
                       (ClientData)unifyData);

              /* Update pins across the bnLine */
              (void) Tp_ItrPPinsOfBnLine(semiTile1Ptr, semiTile2Ptr, 
                      UpdateNetIdInFace,
                      (ClientData)unifyData);

              semiFound = FALSE;
          }
       }
    }
    else
    {
       semiFound = FALSE;
    }
    return (MAGIC_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * tmHCircleUnify () --
 *                           ext = 3     ext = 4
 *                    ---------*----------*-------------
 *                   |                                  |   horz face
 *             ext=3 *     ext=3         ext=4          |
 *                    -------*------------*-------------
 *                   |     int=5         int=6          |
 *    BEFORE         * /////////////////////////////////|   horz bnTile
 *                   |//////////////////////////////////|
 *                   |     int=5         int=6          |
 *                    -------*-------------*------------
 *             ext=7 *     ext=7         ext=8          |
 *                   |                                  |   horz face
 *                    -----------*--------------*-------
 *                             ext=7           ext=8
 * ----------------------------------------------------------------------------
 *                           ext = 3     ext = 4
 *                    ---------*----------*-------------
 *                   |                                  |   horz face
 *             ext=3 *     ext=3         ext=4          |
 *                    -------*------------*-------------
 *                   |     int=3         int=4          |
 *    AFTER          * /////////////////////////////////|   horz bnTile
 *                   |//////////////////////////////////|
 *                   |     int=3         int=4          |
 *                    -------*-------------*------------
 *             ext=3 *     ext=3         ext=4          |
 *                   |                                  |   horz face
 *                    ------*--------------*------------
 *                         ext=3           ext=4
 *
 *   Result:   All pins and ppins have been unified across bnTiles and faces.
 *
 * ----------------------------------------------------------------------------
 */

static int
tmHCircleUnify(tilePtr, unifyData)
Tile        *tilePtr;
UnifyData   *unifyData;
{
    WallJunction *junctPtr;
    WallSegment  *segPtr;
    LinkedPList  *bnListsPtr;
    LinkedPList  *faceListsPtr;
    LinkedTile   *tileListPtr;
    Tile         *semiTile1Ptr;
    Tile         *semiTile2Ptr;
    Tile         *domTilePtr;
    Point        *domPointPtr;
    BNLineType   bnType;
    static bool  semiFound;

    if ((Tp_IsDominantTile(tilePtr)) &&
        (unifyData->bnListTilePtr == NIL(LinkedTile)))
    {
       /* Get all pins in first face region */
       semiFound = FALSE;
       Tp_HorzTileToRect(unifyData->planeDataPtr, tilePtr,
                 &(unifyData->domRect));
       MALLOC(LinkedPList *, faceListsPtr, sizeof (LinkedPList));
       faceListsPtr->nextPtr = unifyData->faceListsPtr;
       unifyData->faceListsPtr = faceListsPtr;
       unifyData->faceListsPtr->pinListsPtr = 
                 Tm_PinsAroundHFace(unifyData->parentCellPtr, 
                        unifyData->planeDataPtr, 
                        unifyData->domRect,
                        TRUE);
       if (Tp_IsSBNTile(tilePtr))
       {
          semiTile1Ptr = tilePtr;
          semiFound = TRUE;
       }
       return (MAGIC_CONTINUE_SEARCH);
    }

    /* If tile is bottleneck tile */
        if (Tp_IsBottleneckTile(tilePtr))
    {
       semiFound = FALSE;
       MALLOC(LinkedTile *, tileListPtr, sizeof (LinkedTile));
                  
       tileListPtr->tilePtr = tilePtr;
       tileListPtr->nextPtr = unifyData->bnListTilePtr;
       unifyData->bnListTilePtr = tileListPtr;

       /* Get all pins in bottleneck region */
       MALLOC(LinkedPList *, bnListsPtr, sizeof (LinkedPList));
              
       bnListsPtr->nextPtr = unifyData->bnListsPtr;
/*
 * code change : by george : on 24 oct 89
 *
       MALLOC(PinsAroundJunction *, 
              bnListsPtr->pinListsPtr,
              sizeof (PinsAroundJunction));
 */
        bnListsPtr->pinListsPtr = Tm_MakePinsAroundJunction();
       unifyData->bnListsPtr = bnListsPtr;
       (void) Tm_GetPinsInHBottleNeckRegion(
                        unifyData->parentCellPtr,
                        unifyData->planeDataPtr, 
                            tilePtr, 
                            bnListsPtr->pinListsPtr);
       /* Add pseudo pins on top and bottom sides of bnTile */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 TOP_RIGHT_SIDE,
                 AddPPinsToHList,
                 (ClientData) unifyData);
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 BOTTOM_LEFT_SIDE,
                 AddPPinsToHList,
                 (ClientData) unifyData);

       /* Update nets in bnRegion by examining pseudopins on top */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 TOP_RIGHT_SIDE,
                 UpdateNetIdInBNRegion, 
                 (ClientData) unifyData);
       /* Get Pins in next face region */
       segPtr = Tp_GetWallSegmentOfBnTile(tilePtr);
       junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
       domPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
       domTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(unifyData->planeDataPtr), 
                    domPointPtr);
       Tp_HorzTileToRect(unifyData->planeDataPtr, domTilePtr, 
                 &(unifyData->domRect));
       MALLOC(LinkedPList *, faceListsPtr, sizeof (LinkedPList));
       faceListsPtr->nextPtr = unifyData->faceListsPtr;
       unifyData->faceListsPtr = faceListsPtr;
       if ((Fpg_GetEastSegOfWallJunct(junctPtr) == NIL(WallSegment)) &&
           (Fpg_GetWestSegOfWallJunct(junctPtr) == NIL(WallSegment)))
       {
/*
 * code change : by george : on 24 oct 89
 *
           MALLOC(PinsAroundJunction *, 
              unifyData->faceListsPtr->pinListsPtr,
              sizeof (PinsAroundJunction));
 */
           unifyData->faceListsPtr->pinListsPtr =
                Tm_MakePinsAroundJunction();
           Tm_PinsAroundJunction(unifyData->parentCellPtr, 
                     unifyData->planeDataPtr, 
                     junctPtr, 
                     unifyData->faceListsPtr->pinListsPtr);
       }
       else
       {
           unifyData->faceListsPtr->pinListsPtr = 
                 Tm_PinsAroundHFace(unifyData->parentCellPtr, 
                        unifyData->planeDataPtr, 
                        unifyData->domRect,
                        TRUE);
       }

       /* Update nets in face region by examining pseudo pins*/
       /*  on bottom of bottleneck tile */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 BOTTOM_LEFT_SIDE,
                 UpdateNetIdInFace, 
                 (ClientData) unifyData);
       /* Needs a second pass - since the fixNets may cause changes */
       (void) Tp_ItrPPinsOfSide(tilePtr, 
                 BOTTOM_LEFT_SIDE,
                 UpdateNetIdInFace, 
                 (ClientData) unifyData);
    }
    /* Semi Bottleneck Tile */
    else if (Tp_IsSBNTile(tilePtr))
    {
       if (semiFound == FALSE)
       {
          semiTile1Ptr = tilePtr;
          semiFound = TRUE;
       }
       else
       {
          semiTile2Ptr = tilePtr;
          if (!Tp_IsPairOfSBNTiles(&semiTile1Ptr, &semiTile2Ptr, &bnType))
          {
         /* Didn't find a match */
         semiTile1Ptr = tilePtr;
         return (MAGIC_CONTINUE_SEARCH);
          }
          if (bnType != TP_NULL_BNL)
          {
             MALLOC(LinkedTile *, tileListPtr, sizeof (LinkedTile));
             tileListPtr->tilePtr = tilePtr;
         tileListPtr->nextPtr = unifyData->bnListTilePtr;
             unifyData->bnListTilePtr = tileListPtr;

             MALLOC(LinkedPList *, bnListsPtr, sizeof (LinkedPList));
         bnListsPtr->nextPtr = unifyData->bnListsPtr;
/*
 * code change : by george : on 24 oct 89
 *
             MALLOC(PinsAroundJunction *, 
            bnListsPtr->pinListsPtr,
            sizeof (PinsAroundJunction));
 */
            bnListsPtr->pinListsPtr = Tm_MakePinsAroundJunction();
             unifyData->bnListsPtr = bnListsPtr;

         /* Add pseudo pins on bnLine */
         (void) Tp_ItrPPinsOfBnLine(semiTile1Ptr, semiTile2Ptr, 
                     AddPPinsToHList,
                     (ClientData)unifyData);

         /* Get Pins in next face region */
         Tp_FindBnLine(semiTile1Ptr,semiTile2Ptr,&(unifyData->bnLine));
         (void) Tp_ItrHorzBnLine(unifyData->planeDataPtr, FindPinsForBnLine, 
                  (ClientData)unifyData);

         /* Update pins across the bnLine */
         (void) Tp_ItrPPinsOfBnLine(semiTile1Ptr, semiTile2Ptr, 
                     UpdateNetIdInFace,
                     (ClientData)unifyData);
         
         semiFound = FALSE;
           }
       }
    }
    else
    {
       semiFound = FALSE;
    }
    return (MAGIC_CONTINUE_SEARCH);
}



static BearStatus
FindPinsForBnLine(bnLinePtr, unifyData)
BNLine *bnLinePtr;
UnifyData *unifyData;
{
   WallSegment  *segPtr;
   WallJunction *junctPtr;
   Point        *domPointPtr;
   Point        vdomPoint;
   Tile         *domTilePtr;
   LinkedPList  *faceListsPtr;
   Line         line1;
   Line         line2;

   line1 = bnLinePtr->line;
   line2 = unifyData->bnLine;
   if ((line1.point1.p_x == line2.point1.p_x) &&
       (line1.point1.p_y == line2.point1.p_y) &&
       (line1.point2.p_x == line2.point2.p_x) &&
       (line1.point2.p_y == line2.point2.p_y))
   {
      MALLOC(LinkedPList *, faceListsPtr, sizeof (LinkedPList));
      faceListsPtr->nextPtr = unifyData->faceListsPtr;
      unifyData->faceListsPtr = faceListsPtr;
      segPtr = Tp_GetWallSegmentOfBnLine(bnLinePtr);
      junctPtr = Fpg_GetSEJunctOfWallSeg(segPtr);
      domPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);

      if ((bnLinePtr->bnLineType == TP_HORZ_B_RIGHT) ||
      (bnLinePtr->bnLineType == TP_HORZ_B_LEFT))
      {
         domTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(unifyData->planeDataPtr),
                      domPointPtr);
         Tp_HorzTileToRect(unifyData->planeDataPtr, domTilePtr, 
               &(unifyData->domRect));
     if ((Fpg_GetEastSegOfWallJunct(junctPtr) == NIL(WallSegment)) &&
         (Fpg_GetWestSegOfWallJunct(junctPtr) == NIL(WallSegment)))
     {
/*
 * code change : by george : on 24 oct 89
 *
         MALLOC(PinsAroundJunction *, unifyData->faceListsPtr->pinListsPtr,
            sizeof (PinsAroundJunction));
 */
         unifyData->faceListsPtr->pinListsPtr =
            Tm_MakePinsAroundJunction();

         Tm_PinsAroundJunction(unifyData->parentCellPtr, 
                   unifyData->planeDataPtr, 
                   junctPtr, 
                   unifyData->faceListsPtr->pinListsPtr);
      }
      else
      {
              unifyData->faceListsPtr->pinListsPtr = 
            Tm_PinsAroundHFace(unifyData->parentCellPtr, 
                       unifyData->planeDataPtr, 
                       unifyData->domRect,
                       TRUE);
       }
    }
    else
    {
       Tp_HPointToVPoint(domPointPtr, &vdomPoint);
       domTilePtr = TiSrPointNoHint(Tp_GetVertPlane(unifyData->planeDataPtr),
                    &vdomPoint);
       Tp_VertTileToRect(unifyData->planeDataPtr, domTilePtr, 
                 &(unifyData->domRect));
       if ((Fpg_GetNorthSegOfWallJunct(junctPtr) == NIL(WallSegment)) &&
           (Fpg_GetSouthSegOfWallJunct(junctPtr) == NIL(WallSegment)))
       {
/*
 * code change : by george : on 24 oct 89
 *
           MALLOC(PinsAroundJunction *, 
              unifyData->faceListsPtr->pinListsPtr,
              sizeof (PinsAroundJunction));
 */
           unifyData->faceListsPtr->pinListsPtr =
                Tm_MakePinsAroundJunction();

           Tm_PinsAroundJunction(unifyData->parentCellPtr, 
                     unifyData->planeDataPtr, 
                     junctPtr, 
                     unifyData->faceListsPtr->pinListsPtr);
        }
        else
        {
           unifyData->faceListsPtr->pinListsPtr = 
              Tm_PinsAroundVFace(unifyData->parentCellPtr, 
                     unifyData->planeDataPtr, 
                     unifyData->domRect,
                     TRUE);
            }
    }
    return (BEAR_STOP_SEARCH);
   }
   return (BEAR_CONTINUE_SEARCH);
}

static BearStatus
UpdateNetIdInBNRegion(tlPPinPtr, unifyData)
PseudoPin *tlPPinPtr;
UnifyData *unifyData;
{
    BearId  oldId;
    BearId  newId;

    newId = Tp_GetExtNetIdOfPPin(tlPPinPtr);
    oldId = Tp_GetIntNetIdOfPPin(tlPPinPtr);

    if (newId != oldId)
    {
       UpdateNetIdInList(unifyData->bnListsPtr->pinListsPtr,
                         oldId, newId, TRUE, unifyData->horzUnify);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
UpdateNetIdInFace(brPPinPtr, unifyData)
PseudoPin *brPPinPtr;
UnifyData *unifyData;
{
    BearId  oldId;
    BearId  newId;

    if (Tp_IsSBNTile(brPPinPtr->tilePtr) && unifyData->horzUnify)
    {
       oldId = Tp_GetIntNetIdOfPPin(brPPinPtr);
       newId = Tp_GetExtNetIdOfPPin(brPPinPtr);
    }
    else
        {
       newId = Tp_GetIntNetIdOfPPin(brPPinPtr);
       oldId = Tp_GetExtNetIdOfPPin(brPPinPtr);
    }

    if (oldId != newId)
    {
       /* Update All Pins in Face (pinInst, rPin, pseudoPins) */
       UpdateNetIdInList(unifyData->faceListsPtr->pinListsPtr, 
                 oldId, newId, FALSE, unifyData->horzUnify);
       unifyData->pinId = newId;
           if (unifyData->faceListsPtr->nextPtr != NIL(LinkedPList))
       {
          FixNetIdInCircle(unifyData);
       }
    }
    return (BEAR_CONTINUE_SEARCH);
}

static void
FixNetIdInCircle(unifyData)
UnifyData *unifyData;
{
   Tile *tilePtr;
   LinkedPList        *HeadbnListsPtr;
   LinkedPList        *HeadfaceListsPtr;
   LinkedTile         *HeadbnListTilePtr;
   SideDirection      side;
   SideDirection      side2;


   /* Set pointer to head of list */
   HeadbnListsPtr    = unifyData->bnListsPtr;
   HeadfaceListsPtr  = unifyData->faceListsPtr;
   HeadbnListTilePtr = unifyData->bnListTilePtr;

   /* Ignore first set of face pins */
   unifyData->faceListsPtr = unifyData->faceListsPtr->nextPtr;

   for (tilePtr = unifyData->bnListTilePtr->tilePtr; 
        unifyData->bnListTilePtr != NIL(LinkedTile);)
   {
       if (unifyData->sideDir == TOP_SIDE)
       {
          side = BOTTOM_LEFT_SIDE;
          side2 = TOP_RIGHT_SIDE;
       }
       else
       {
          side = TOP_RIGHT_SIDE;
          side2 = BOTTOM_LEFT_SIDE;
       }
       if (!Tp_IsSBNTile(tilePtr))
       {
          (void) Tp_ItrPPinsOfSide(tilePtr,side,FixNetIdInBNRegion,
                       (ClientData)unifyData);
       }
       (void) Tp_ItrPPinsOfSide(tilePtr,side2,FixNetIdInFace,
                 (ClientData)unifyData);

       unifyData->bnListsPtr = unifyData->bnListsPtr->nextPtr;
       unifyData->faceListsPtr = unifyData->faceListsPtr->nextPtr;
       unifyData->bnListTilePtr = unifyData->bnListTilePtr->nextPtr;
       if (unifyData->bnListTilePtr)
       {
            tilePtr = unifyData->bnListTilePtr->tilePtr;
       }
   }

   /* Restore UnifyData pointers to head of list */
   unifyData->bnListsPtr    = HeadbnListsPtr;
   unifyData->faceListsPtr  = HeadfaceListsPtr;
   unifyData->bnListTilePtr = HeadbnListTilePtr;

}

static BearStatus
FixNetIdInBNRegion(brPPinPtr, unifyData)
PseudoPin *brPPinPtr;
UnifyData *unifyData;

{
    BearId  newId;
    BearId  oldId;
    
    oldId = Tp_GetIntNetIdOfPPin(brPPinPtr);
    newId = Tp_GetExtNetIdOfPPin(brPPinPtr);
    if ((newId == unifyData->pinId) && (newId != oldId))
    {
       /* Change all pinInst and rPins in bnRegion to newId */
       UpdateNetIdInList(unifyData->bnListsPtr->pinListsPtr, 
                 oldId, newId, TRUE, unifyData->horzUnify);
    }
    return (BEAR_CONTINUE_SEARCH);
}

static BearStatus
FixNetIdInFace(tlPPinPtr, unifyData)
PseudoPin *tlPPinPtr;
UnifyData *unifyData;
{
    BearId  newId;
    BearId  oldId;
    
    if (Tp_IsSBNTile(tlPPinPtr->tilePtr) && (unifyData->horzUnify == FALSE))
    {
       oldId = Tp_GetIntNetIdOfPPin(tlPPinPtr);
       newId = Tp_GetExtNetIdOfPPin(tlPPinPtr);
    }
        else
    {
      oldId = Tp_GetExtNetIdOfPPin(tlPPinPtr);
      newId = Tp_GetIntNetIdOfPPin(tlPPinPtr);
    }
    if ((newId == unifyData->pinId) && (newId != oldId))
    {
       /* Change all pinInst and rPins in bnRegion to newId */
       UpdateNetIdInList(unifyData->faceListsPtr->pinListsPtr, 
                 oldId, newId, FALSE, unifyData->horzUnify);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static void
UpdateNetIdInList(pinListsPtr, oldId, newId, matchIntPPin, horzUnify)
PinsAroundJunction *pinListsPtr;
BearId oldId;
BearId newId;
bool   matchIntPPin;
bool   horzUnify;
{
    LinkedPPin *pPinPtr;
    LinkedRPin *rPinPtr;
    LinkedPinInst *pinInstPtr;

    /* Update values of all matching pins in pinList to newId */
    for (pPinPtr = pinListsPtr->horzPPinList; pPinPtr != NIL(LinkedPPin);
        pPinPtr = pPinPtr->nextPtr)
    {
      /* Match Internal Id also when ppin belongs to semi-bnTile */
      if (((matchIntPPin == TRUE) ||
           ((pPinPtr->ppinPtr->tilePtr != NIL(Tile)) && 
        (Tp_IsSBNTile(pPinPtr->ppinPtr->tilePtr)))) &&
          (Tp_GetIntNetIdOfPPin(pPinPtr->ppinPtr) == oldId))
      {
              Tp_PutIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
          }
          /* Match External Id also when ppin belongs to dummyCell */
      else if (((matchIntPPin == FALSE) ||
            ((pPinPtr->ppinPtr->tilePtr != NIL(Tile)) && 
             (Tp_IsDCellTile(pPinPtr->ppinPtr->tilePtr)))) &&
           (Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr) == oldId))
      {
        Tp_PutExtNetIdOfPPin(newId, pPinPtr->ppinPtr);
          }
        }

    for (pPinPtr = pinListsPtr->vertPPinList; pPinPtr != NIL(LinkedPPin);
         pPinPtr = pPinPtr->nextPtr)
    {
        /* Match Internal Id also when ppin belongs to semi-bnTile */
      if (((matchIntPPin == TRUE) ||
           ((pPinPtr->ppinPtr->tilePtr != NIL(Tile)) && 
        (Tp_IsSBNTile(pPinPtr->ppinPtr->tilePtr)))) &&
          (Tp_GetIntNetIdOfPPin(pPinPtr->ppinPtr) == oldId))
      {
              Tp_PutIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
          }
          /* Match External Id also when ppin belongs to dummyCell */
          else if (((matchIntPPin == FALSE) ||
            ((pPinPtr->ppinPtr->tilePtr != NIL(Tile)) && 
             (Tp_IsDCellTile(pPinPtr->ppinPtr->tilePtr)))) &&
           (Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr) == oldId))
          {
             Tp_PutExtNetIdOfPPin(newId, pPinPtr->ppinPtr);
       }
    }

    for (rPinPtr = pinListsPtr->rPinList; rPinPtr != NIL(LinkedRPin);
         rPinPtr = rPinPtr->nextPtr)
    {
        if ((horzUnify == TRUE) &&
            (Mac_GetHExtNetIdOfRPin(rPinPtr->rPinPtr) == oldId))
        {
            Mac_PutExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
            Mac_PutHExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
        }
        if ((horzUnify == FALSE) &&
            (Mac_GetVExtNetIdOfRPin(rPinPtr->rPinPtr) == oldId))
        {
            Mac_PutExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
            Mac_PutVExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
        }
    }

    for (pinInstPtr = pinListsPtr->pinInstList; 
         pinInstPtr != NIL(LinkedPinInst);
         pinInstPtr = pinInstPtr->nextPtr)
    {
      if ((horzUnify == TRUE) &&
          ((Mac_GetHExtNetIdOfPin(pinInstPtr->pinInstPtr) == oldId) ||
           (Mac_GetExtNetIdOfPin(pinInstPtr->pinInstPtr) == oldId)))
      {
         Mac_PutExtNetIdOfPin(newId, pinInstPtr->pinInstPtr);
         Mac_PutHExtNetIdOfPin(newId, pinInstPtr->pinInstPtr);
      }
      if ((horzUnify == FALSE) &&
          ((Mac_GetVExtNetIdOfPin(pinInstPtr->pinInstPtr) == oldId) ||
           (Mac_GetExtNetIdOfPin(pinInstPtr->pinInstPtr) == oldId)))
      {
         Mac_PutExtNetIdOfPin(newId, pinInstPtr->pinInstPtr);
         Mac_PutVExtNetIdOfPin(newId, pinInstPtr->pinInstPtr);
      }
    }
}

static BearStatus
AddPPinsToHList(pPinPtr, unifyData)
PseudoPin *pPinPtr;
UnifyData *unifyData;
{
   LinkedPPin *pinPtr;

   MALLOC(LinkedPPin *, pinPtr, sizeof (LinkedPPin));
   pinPtr->ppinPtr = pPinPtr;
   pinPtr->nextPtr = unifyData->bnListsPtr->pinListsPtr->horzPPinList;
   unifyData->bnListsPtr->pinListsPtr->horzPPinList = pinPtr;
   return (BEAR_CONTINUE_SEARCH);
}

static BearStatus
AddPPinsToVList(pPinPtr, unifyData)
PseudoPin *pPinPtr;
UnifyData *unifyData;
{
   LinkedPPin *pinPtr;

   MALLOC(LinkedPPin *, pinPtr, sizeof (LinkedPPin));
   pinPtr->ppinPtr = pPinPtr;
   pinPtr->nextPtr = unifyData->bnListsPtr->pinListsPtr->vertPPinList;
   unifyData->bnListsPtr->pinListsPtr->vertPPinList = pinPtr;
   return (BEAR_CONTINUE_SEARCH);
}

