
/* tmHBnSplit.c --
 *
 *          Tim Collins
 *     (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmHBnSplit.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


typedef struct _BnLineData
{
    PseudoPin          *pPinList;
    DummyCell          *dCellPtr;
    struct _BnLineData *nextPtr;
} BnLineData;

    
typedef struct _LocalData
{
    Rect          *area;
    Table         *table;
    int           zone;
    SideDirection side;
    LinkedPinInst **pinList;
    LinkedPPin    **pPinList;
    LinkedRPin    **rPinList;
} LocalData;


/* ---------- Local Procedures ---------- */
static void         HBnLineSetup();
static void         HBnTileSetup();
static void         HBnLineSplit();
static void         HBnTileSplit();
static void         AddPPinToBnLineTile();
static void         HandleZonePins();
static void         OperateOnRightPins();
static void         OperateOnRightBn();
static void         OperateOnDCell();
static void         OperateOnLeftBn();
static void         OperateOnLeftPins();
static void         DividePPins();
static int          SearchSolidTile();
static BearStatus   GetPinsInRect();
static BearStatus   GetRPinsInRect();
static BearStatus   AddPPinsToTable();
static BearStatus   AddDCellPPinsToTable();


/* ---------- Local Global Variables ---------- */
static Table *tableHeadPtr;


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_InsDCellOnHTP() --
 *
 *    This procedure inserts all the dummy cells, and distributes the pseudo
 *    pins that were on the horizontal bottleneck tile, evenly among the two
 *    new bottleneck tiles, and the dummy cell.
 *
 * Results:
 *    All the dummy cells are inserted.
 *
 * Side effects:
 *    The dummy cells have horizontal pseudo pins.
 *
 * ----------------------------------------------------------------------------
 */

void
Tm_InsDCellOnHTP(parentCellPtr)
CellMaster *parentCellPtr;
{
    BnLineData      *bnLineDataList;
    DummyCell       *dCellPtr;
    DCellClass      class;
    PlaneData       *planeDataPtr;
    WallJunction    *llJunctPtr;

    tmGlobalParentCellPtr = parentCellPtr;
    tmGlobalPlaneDataPtr = planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    bnLineDataList = NIL(BnLineData);

    if (planeDataPtr->pd_dcellList == NIL(DummyCell))
    {
        (void) Tp_CreateDCells(planeDataPtr);
    }
    dCellPtr = planeDataPtr->pd_dcellList;
    while (dCellPtr != NIL(DummyCell))
    {
        class = Mac_GetClassOfDCell(dCellPtr);
        if ((class == HBNTI_AND_VBNTI) || (class == HBNTI_AND_VBNLN))
        {
            (void) HBnTileSetup(dCellPtr);
        }
        dCellPtr = dCellPtr->nextPtr;
    }
    dCellPtr = planeDataPtr->pd_dcellList;
    while (dCellPtr != NIL(DummyCell))
    {
        tableHeadPtr = NIL(Table);
        class = Mac_GetClassOfDCell(dCellPtr);
        if ((class != HBNLN_AND_VBNTI) && (class != HBNLN_AND_VBNLN))
        {
            (void) HBnTileSplit(dCellPtr);
        }
        dCellPtr = dCellPtr->nextPtr;
    }

    dCellPtr = planeDataPtr->pd_dcellList;
    while (dCellPtr != NIL(DummyCell))
    {
        tableHeadPtr = NIL(Table);
        class = Mac_GetClassOfDCell(dCellPtr);
        if ((class == HBNLN_AND_VBNTI) || (class == HBNLN_AND_VBNLN))
        {
            (void) HBnLineSetup(dCellPtr, &bnLineDataList);
        }
        dCellPtr = dCellPtr->nextPtr;
    }
    while (bnLineDataList != NIL(BnLineData))
    {
        HBnLineSplit(bnLineDataList);
        bnLineDataList = bnLineDataList->nextPtr;
    }
/*
 * PROPOSED CODE
 *
    while ((bnLineDataPtr = bnLineDataList) != NIL(BnLineData))
    {
        HBnLineSplit(bnLineDataPtr);
        bnLineDataList = bnLineDataPtr->nextPtr;
        FREE((char *) bnLineDataPtr);
    }
 */
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
}


static void
HBnLineSetup(dCellPtr, bnLineDataListPtr)
DummyCell  *dCellPtr;
BnLineData **bnLineDataListPtr;
{
    /* 
     * Since by inserting the dummy cell the tile plane is changed, we must
     * first save all the information from the semi-bottleneck tiles, and
     * then paint the dummy cells.
     */

    Tile       *tSBnTilePtr;
    Tile       *bSBnTilePtr;
    BnLineData *bnLineDataPtr;
    BNLineType bnType;
    Line       line;

    MALLOC(BnLineData *, bnLineDataPtr, sizeof (BnLineData));
    (void) TmGetHSBnTilesOfDCell(dCellPtr, &tSBnTilePtr, &bSBnTilePtr);
    (void) Tp_FindBnLine(tSBnTilePtr, bSBnTilePtr, &line);
    (void) Tp_IsPairOfSBNTiles(&tSBnTilePtr, &bSBnTilePtr, &bnType);
    /*
     * Save the pseudo pins from the correct fragment of
     * the semi-bottleneck tile.
     */
    if (bnType == TP_HORZ_B_RIGHT)
    {
       (void) Tp_RemovePPinListOfTile(tSBnTilePtr, BOTTOM_LEFT_SIDE, 
                      TP_TOP, &bnLineDataPtr->pPinList);
    }
    else
    {
       (void) Tp_RemovePPinListOfTile(tSBnTilePtr, BOTTOM_LEFT_SIDE, 
                      TP_BOTTOM, &bnLineDataPtr->pPinList);
    }
    (void) Tp_FreeTileProp(tSBnTilePtr);
    (void) Tp_FreeTileProp(bSBnTilePtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    bnLineDataPtr->dCellPtr = dCellPtr;
    bnLineDataPtr->nextPtr = *bnLineDataListPtr;
    *bnLineDataListPtr = bnLineDataPtr;
}


static void
HBnLineSplit(bnLineDataPtr)
BnLineData *bnLineDataPtr;
{
    /*
     * Paint the dummy cell and then put the pseudo pins evenly among
     * the two horizontal bottleneck tiles, and the dummy cell.
     */
    Tile       *leftBnTilePtr;
    Tile       *rightBnTilePtr;
/* Tile       *topBnTilePtr; */
/* Tile       *bottomBnTilePtr; */
    WallJunction *llJunctPtr;

    (void) Tp_PaintDCellOnHorzPlane(bnLineDataPtr->dCellPtr, 
      tmGlobalPlaneDataPtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    /* probably don't need to create a fpg here 
    (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &llJunctPtr);
    */
    (void) TmGetLAndRBnTilesOfDCell(bnLineDataPtr->dCellPtr, &leftBnTilePtr, 
      &rightBnTilePtr);
    /*
    (void) Tp_CreateSpaceTileProp(leftBnTilePtr);
    (void) Tp_PutTypeOfSpaceTile(TP_HORZ_BN_TILE, leftBnTilePtr);
    (void) Tp_CreateSpaceTileProp(rightBnTilePtr);
    (void) Tp_PutTypeOfSpaceTile(TP_HORZ_BN_TILE, rightBnTilePtr);
    */
    (void) Tp_PaintDCellOnVertPlane(bnLineDataPtr->dCellPtr, 
      tmGlobalPlaneDataPtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &llJunctPtr);
    /*
    (void) TmGetBAndTBnTilesOfDCell(bnLineDataPtr->dCellPtr, &bottomBnTilePtr, 
      &topBnTilePtr);
    (void) Tp_CreateSpaceTileProp(topBnTilePtr);
    (void) Tp_PutTypeOfSpaceTile(TP_VERT_BN_TILE, topBnTilePtr);
    (void) Tp_CreateSpaceTileProp(bottomBnTilePtr);
    (void) Tp_PutTypeOfSpaceTile(TP_VERT_BN_TILE, bottomBnTilePtr);
    */
    while (bnLineDataPtr->pPinList != NIL(PseudoPin))
    {
        (void) AddPPinToBnLineTile(leftBnTilePtr, bnLineDataPtr);
    }
}


static void
AddPPinToBnLineTile(bnTilePtr, bnLineDataPtr)
Tile       *bnTilePtr;
BnLineData *bnLineDataPtr;
{
    PseudoPin *pPinPtr;
    PseudoPin *newPPinPtr;
    BearId    newId;
    BearId    intId;
    Net       *netPtr;
    float     width;

        /*        extId                                        */
        /*    -----0------ pPinPtr                             */
        /*    |   newId  |                 extId               */
        /*    |          |      <----  ------0------- pPinPtr  */
        /*    |   newId  |                 intId               */
        /*    -----0------ newPPinPtr                          */
        /*        intId                                        */

    pPinPtr = bnLineDataPtr->pPinList;
    bnLineDataPtr->pPinList = bnLineDataPtr->pPinList->nextPtr;

    pPinPtr->nextPtr = NIL(PseudoPin);
    Tp_InsertPPinOfBnTile(bnTilePtr, TOP_RIGHT_SIDE, pPinPtr);
    Tp_CreatePPin(bnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM, &newPPinPtr);
    newId = Mac_GenerateId();
    intId = Tp_GetIntNetIdOfPPin(pPinPtr);
    netPtr = Tp_GetNetOfPPin(pPinPtr);
    width  = Tp_GetWidthOfPPin(pPinPtr);

        /* SS: Changed the pin assignment */
    (void) Tp_PutIntNetIdOfPPin(newId, pPinPtr);
    (void) Tp_PutIntNetIdOfPPin(newId, newPPinPtr);
    (void) Tp_PutExtNetIdOfPPin(intId, newPPinPtr);
    (void) Tp_PutNetOfPPin(netPtr, newPPinPtr);
    (void) Tp_PutWidthOfPPin(width, newPPinPtr);
}


static void
HBnTileSetup(dCellPtr)
DummyCell *dCellPtr;
{
    Tile        *vbnTilePtr;
    Point       searchPnt;
    Point       vSearchPnt;
    WallJunction *llJunctPtr;

   /* Paint dummy Cell on vertical plane */

    searchPnt.p_x = dCellPtr->bbox.r_ur.p_x;
    searchPnt.p_y = dCellPtr->bbox.r_ll.p_y;
    (void) Tp_HPointToVPoint(&searchPnt, &vSearchPnt);
    vbnTilePtr = TiSrPointNoHint(Tp_GetVertPlane(tmGlobalPlaneDataPtr),
                                &vSearchPnt);
    (void) Tp_FreeTileProp(vbnTilePtr);
    (void) Tp_PaintDCellOnVertPlane(dCellPtr, tmGlobalPlaneDataPtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &llJunctPtr);
}



static void
HBnTileSplit(dCellPtr)
DummyCell *dCellPtr;
{
    Tile         *vbnTilePtr;
    Tile         *hbnTilePtr;
    Tile         *leftBnTilePtr;
    Tile         *rightBnTilePtr;
    PseudoPin    *topList;
    PseudoPin    *bottomList;
    Point        searchPnt;
    Point        vSearchPnt;
    ZoneData     *zoneDataPtr;
    Rect         rect;
    WallJunction *llJunctPtr;
    WallSegment  *wallSegPtr;
    WallJunction *junct1Ptr;
    WallJunction *junct2Ptr;
    Point        *j1PointPtr;
    Point        *j2PointPtr;

    zoneDataPtr = TmMakeZoneData();
    zoneDataPtr->nextPtr = TmMakeZoneData();
    zoneDataPtr->nextPtr->nextPtr = TmMakeZoneData();
    zoneDataPtr->nextPtr->nextPtr->nextPtr = NIL(ZoneData);

    searchPnt.p_x = dCellPtr->bbox.r_ll.p_x;
    searchPnt.p_y = dCellPtr->bbox.r_ll.p_y;
    hbnTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
                                &searchPnt);
    searchPnt.p_x = dCellPtr->bbox.r_ur.p_x;
    searchPnt.p_y = dCellPtr->bbox.r_ll.p_y;
    (void) Tp_HPointToVPoint(&searchPnt, &vSearchPnt);
    vbnTilePtr = TiSrPointNoHint(Tp_GetVertPlane(tmGlobalPlaneDataPtr),
                                &vSearchPnt);
    Tp_VertTileToRect(tmGlobalPlaneDataPtr, vbnTilePtr, &rect);
    wallSegPtr = Tp_GetWallSegmentOfBnTile(hbnTilePtr);
    junct1Ptr = Fpg_GetNWJunctOfWallSeg(wallSegPtr);
    junct2Ptr = Fpg_GetSEJunctOfWallSeg(wallSegPtr);
    j1PointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct1Ptr);
    j2PointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junct2Ptr);
    rect.r_ytop = j1PointPtr->p_y;
    rect.r_ybot = j2PointPtr->p_y;
    /* Unify the circle (top face, horzBnTile, bot face) */
    Tm_HCircleUnify(tmGlobalParentCellPtr, tmGlobalPlaneDataPtr, rect);
    /* Delete old PPins from horzBnTile */
    (void) Tp_FreePPinsOfTile(hbnTilePtr);
        /* Reassign pins across the 3 zones (top Face, horzBnTile, bot Face) */
        /* by calling Circle Split */
    Tm_HCSplit(tmGlobalParentCellPtr, tmGlobalPlaneDataPtr, rect);

        /* Painting will result in losing list of ppins */           
    (void) Tp_RemovePPinListOfTile(hbnTilePtr, TOP_RIGHT_SIDE,
                                   TP_BOTTOM, &topList);
    (void) Tp_RemovePPinListOfTile(hbnTilePtr, BOTTOM_LEFT_SIDE,
                                   TP_BOTTOM, &bottomList);

        /* Paint dummy cell on Horz Plane */
    (void) Tp_FreeTileProp(hbnTilePtr);
    (void) Tp_PaintDCellOnHorzPlane(dCellPtr, tmGlobalPlaneDataPtr);
    searchPnt.p_x = dCellPtr->bbox.r_ll.p_x;
    searchPnt.p_y = dCellPtr->bbox.r_ll.p_y;
    hbnTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
                                &searchPnt);
    (void) Mac_PutTileOfDCell(hbnTilePtr, dCellPtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    (void) TmGetLAndRBnTilesOfDCell(dCellPtr, &leftBnTilePtr,
                    &rightBnTilePtr);

    /* Put the ppin lists back on the bnTile - proportionally divide the */
    /* list between the left and right bnTiles */
    (void) DividePPins(leftBnTilePtr,rightBnTilePtr,
               TOP_RIGHT_SIDE,topList);
    (void) DividePPins(leftBnTilePtr,rightBnTilePtr,
               BOTTOM_LEFT_SIDE,bottomList);
    (void) HandleZonePins(leftBnTilePtr, dCellPtr, rightBnTilePtr,
        zoneDataPtr);
    (void) TmHSplitByZones(3, tableHeadPtr, zoneDataPtr); 
    (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &llJunctPtr);
}



static void
DividePPins(leftBnTilePtr,rightBnTilePtr,side,pPinList)
Tile           *leftBnTilePtr;
Tile           *rightBnTilePtr;
SideDirection  side;
PseudoPin      *pPinList;
{
    Rect     rect;
    int      leftDX;
    int      rightDX;
    int      pPinNum;
    int      leftPPinNum;
    PseudoPin   *pPinPtr;

    pPinNum = 0;
    leftPPinNum = 0;
    for(pPinPtr=pPinList ; pPinPtr!=NIL(PseudoPin) ; pPinPtr=pPinPtr->nextPtr)
    {
    pPinNum++;
    }
    (void)Tp_HorzTileToRect(tmGlobalPlaneDataPtr, leftBnTilePtr, &rect);
    leftDX = rect.r_ur.p_x - rect.r_ll.p_x;
    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, rightBnTilePtr, &rect);
    rightDX = rect.r_ur.p_x - rect.r_ll.p_x;

    leftPPinNum = pPinNum * ((float)leftDX/(leftDX + rightDX)) + 0.5;
    if (leftPPinNum <= 0)
    {
    (void) Tp_AppendPPinListOfTile(NIL(PseudoPin),leftBnTilePtr,
                       side,TP_BOTTOM);
    (void) Tp_AppendPPinListOfTile(pPinList,rightBnTilePtr,side,
                       TP_BOTTOM);
    }
    else
    {
    (void) Tp_AppendPPinListOfTile(pPinList,leftBnTilePtr,side,TP_BOTTOM);
    pPinPtr = pPinList;
    for (leftPPinNum-- ; leftPPinNum > 0 ; leftPPinNum--)
    {
        pPinPtr = pPinPtr->nextPtr;
    }
    (void) Tp_AppendPPinListOfTile(pPinPtr->nextPtr,rightBnTilePtr,
                       side,TP_BOTTOM);
    pPinPtr->nextPtr = NIL(PseudoPin);
    }
}


static void
HandleZonePins(lBnTilePtr, dCellPtr, rBnTilePtr, zoneDataPtr)
Tile        *lBnTilePtr;
DummyCell   *dCellPtr;
Tile        *rBnTilePtr;
ZoneData    *zoneDataPtr;
{
    /*
     * This procedure operates on the pins in each of the 3 zones :
     * adding them to the table, and saving the pin insts and the pseudo
     * pins so that their id's can be changed if needed.
     */
    ZoneData *zonePlacePtr;

    zonePlacePtr = zoneDataPtr;

    zonePlacePtr->pinInstList = NIL(LinkedPinInst);
    zonePlacePtr->rPinList = NIL(LinkedRPin);
    zonePlacePtr->pPinList = NIL(LinkedPPin);
    (void) OperateOnLeftPins(lBnTilePtr, zonePlacePtr);
    (void) OperateOnLeftBn(lBnTilePtr, &(zonePlacePtr->pPinList));
    zonePlacePtr->zoneType = DJUNCTION;
    MALLOC(BottleNeck *, zonePlacePtr->bnPtr, sizeof (BottleNeck));
    zonePlacePtr->bnPtr->unValue.bnTilePtr = lBnTilePtr;

    zonePlacePtr = zoneDataPtr->nextPtr;

    zonePlacePtr->pinInstList = NIL(LinkedPinInst);
    zonePlacePtr->rPinList = NIL(LinkedRPin);
    zonePlacePtr->pPinList = NIL(LinkedPPin);
    (void) OperateOnDCell(dCellPtr, &(zonePlacePtr->pPinList));
    zonePlacePtr->zoneType = DCELL;
    zonePlacePtr->dCellPtr = dCellPtr;

    zonePlacePtr = zoneDataPtr->nextPtr->nextPtr;

    zonePlacePtr->pPinList = NIL(LinkedPPin);
    (void) OperateOnRightBn(rBnTilePtr, &(zonePlacePtr->pPinList));
    zonePlacePtr->pinInstList = NIL(LinkedPinInst);
    zonePlacePtr->rPinList = NIL(LinkedRPin);
    (void) OperateOnRightPins(rBnTilePtr, zonePlacePtr);
    zonePlacePtr->zoneType = DJUNCTION;
    MALLOC(BottleNeck *, zonePlacePtr->bnPtr, sizeof (BottleNeck));
    zonePlacePtr->bnPtr->unValue.bnTilePtr = rBnTilePtr;

    zonePlacePtr->nextPtr = NIL(ZoneData);
}


static void
OperateOnRightPins(rBnTilePtr, zonePlacePtr)
Tile       *rBnTilePtr;
ZoneData   *zonePlacePtr;
{
    /* 
     * Add the pins to the table, and save them for future access.
     */
    Rect      searchRect;
    LocalData localData;

    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, rBnTilePtr, &searchRect);
    searchRect.r_xbot = searchRect.r_xtop - 1;
    searchRect.r_xtop += 1;
    localData.area = &searchRect;
    localData.pinList = &(zonePlacePtr->pinInstList);
    localData.rPinList = &(zonePlacePtr->rPinList);
    localData.pPinList = &(zonePlacePtr->pPinList);
    localData.side = LEFT_SIDE;
    localData.zone = 3;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &searchRect,
      SearchSolidTile,
      (ClientData) &localData);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(tmGlobalParentCellPtr),
      GetPinsInRect, (ClientData) &localData);
}


static void
OperateOnRightBn(rBnTilePtr, pPinListPtr)
Tile          *rBnTilePtr;
LinkedPPin    **pPinListPtr;
{
    /* 
     * Add the pseudo pins to the table, and save them for future access.
     */
    LocalData localData;

    localData.pPinList = pPinListPtr;
    localData.zone = 3;
    (void) Tp_ItrPPinsOfSide(rBnTilePtr, TOP_RIGHT_SIDE, AddPPinsToTable, 
      (ClientData) &localData);
    (void) Tp_ItrPPinsOfSide(rBnTilePtr, BOTTOM_LEFT_SIDE, AddPPinsToTable,
      (ClientData) &localData);
}


static void
OperateOnDCell(dCellPtr, pPinListPtr)
DummyCell     *dCellPtr;
LinkedPPin    **pPinListPtr;
{
    /* 
     * Add the pseudo pins to the table, and save them for future access.
     */
    LocalData localData;

    localData.pPinList = pPinListPtr;
    localData.zone = 2;
/*
 * code change : by george : on 18 oct 89
 *
    (void) Mac_ItrPPinOfDCell(dCellPtr, TOP_SIDE, AddPPinsToTable, 
        (ClientData) &localData);
    (void) Mac_ItrPPinOfDCell(dCellPtr, BOTTOM_SIDE, AddPPinsToTable, 
        (ClientData) &localData);
*/
    (void) Mac_ItrPPinOfDCell(dCellPtr, TOP_SIDE, AddDCellPPinsToTable, 
        (ClientData) &localData);
    (void) Mac_ItrPPinOfDCell(dCellPtr, BOTTOM_SIDE, AddDCellPPinsToTable, 
        (ClientData) &localData);
}


static void
OperateOnLeftBn(lBnTilePtr, pPinListPtr)
Tile          *lBnTilePtr;
LinkedPPin    **pPinListPtr;
{
    /* 
     * Add the pseudo pins to the table, and save them for future access.
     */
    LocalData localData;

    localData.pPinList = pPinListPtr;
    localData.zone = 1;
    (void) Tp_ItrPPinsOfSide(lBnTilePtr, TOP_RIGHT_SIDE, AddPPinsToTable,
      (ClientData) &localData);
    (void) Tp_ItrPPinsOfSide(lBnTilePtr, BOTTOM_LEFT_SIDE, AddPPinsToTable,
      (ClientData) &localData);
}


static void
OperateOnLeftPins(lBnTilePtr, zonePlacePtr)
Tile      *lBnTilePtr;
ZoneData  *zonePlacePtr;
{
    /* 
     * Add the pins to the table, and save them for future access.
     */
    Rect      searchRect;
    LocalData localData;

    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, lBnTilePtr, &searchRect);
    searchRect.r_xtop = searchRect.r_xbot + 1;
    searchRect.r_xbot -= 1;
    localData.area = &searchRect;
    localData.pinList = &(zonePlacePtr->pinInstList);
    localData.rPinList = &(zonePlacePtr->rPinList);
    localData.pPinList = &(zonePlacePtr->pPinList);
    localData.side = RIGHT_SIDE;
    localData.zone = 1;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &searchRect,
      SearchSolidTile,
      (ClientData) &localData);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(tmGlobalParentCellPtr),
        GetPinsInRect, (ClientData) &localData);
}


static BearStatus
AddDCellPPinsToTable(pPinPtr, localDataPtr)
PseudoPin *pPinPtr;
LocalData *localDataPtr;
{
    LinkedPPin *pinPtr;
    BearId     id;
    Net        *netPtr;
    float      width;

    MALLOC(LinkedPPin *, pinPtr, sizeof (LinkedPPin));
    pinPtr->ppinPtr = pPinPtr;
    pinPtr->nextPtr = *(localDataPtr->pPinList);
    *(localDataPtr->pPinList) = pinPtr;
    id = Tp_GetExtNetIdOfPPin(pPinPtr);
    netPtr = Tp_GetNetOfPPin(pPinPtr);
    width = Tp_GetWidthOfPPin(pPinPtr);
    (void) TmNumberTable(&tableHeadPtr, id, width, netPtr, localDataPtr->zone);

    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
AddPPinsToTable(pPinPtr, localDataPtr)
PseudoPin *pPinPtr;
LocalData *localDataPtr;
{
    LinkedPPin *pinPtr;
    BearId     id;
    Net        *netPtr;
    float      width;

    MALLOC(LinkedPPin *, pinPtr, sizeof (LinkedPPin));
    pinPtr->ppinPtr = pPinPtr;
    pinPtr->nextPtr = *(localDataPtr->pPinList);
    *(localDataPtr->pPinList) = pinPtr;
    id = Tp_GetIntNetIdOfPPin(pPinPtr);
    netPtr = Tp_GetNetOfPPin(pPinPtr);
    width = Tp_GetWidthOfPPin(pPinPtr);
    (void) TmNumberTable(&tableHeadPtr, id, width, netPtr, localDataPtr->zone);

    return (BEAR_CONTINUE_SEARCH);
}


static int
SearchSolidTile(tilePtr, localDataPtr)
Tile      *tilePtr;
LocalData *localDataPtr;
{
    CellInst    *cellInstPtr;
    RouteCell   *routeCellPtr;
    DummyCell   *dCellPtr;

    if (Tp_IsCellInstTile(tilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        (void) Mac_ItrPinOfCellInst (cellInstPtr, GetPinsInRect,
          (ClientData) localDataPtr);
    }
#ifdef USE_DR_CODE
    if (Tp_IsRCellTile(tilePtr))
    {
        routeCellPtr = Tp_GetRCellOfTile(tilePtr);
        (void) Mac_ItrSidePinOfRCell(routeCellPtr, GetRPinsInRect,
          (ClientData) localDataPtr);
    }
#endif /* USE_DR_CODE */
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, localDataPtr->side, 
          AddDCellPPinsToTable, (ClientData) localDataPtr);
    }

    return (MAGIC_CONTINUE_SEARCH);
}


static BearStatus
GetRPinsInRect(rPinPtr, localDataPtr)
RoutePin  *rPinPtr;
LocalData *localDataPtr;
{
    LinkedRPin *linkRPinPtr;
    Point      absPos;
    BearId     hId;
    Net        *netPtr;
    float       width;
    float       *widthVect;

    (void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if ((absPos.p_x >= localDataPtr->area->r_xbot) &&
      (absPos.p_x <= localDataPtr->area->r_xtop) &&
      (absPos.p_y >= localDataPtr->area->r_ybot) &&
      (absPos.p_y <= localDataPtr->area->r_ytop))
    {
        MALLOC(LinkedRPin *, linkRPinPtr, sizeof (LinkedRPin));
        linkRPinPtr->rPinPtr = rPinPtr;
        linkRPinPtr->nextPtr = *(localDataPtr->rPinList);
        *(localDataPtr->rPinList) = linkRPinPtr;
        hId = Mac_GetHExtNetIdOfRPin(rPinPtr);
        netPtr = Mac_GetNetPtrOfRPin(rPinPtr);
        widthVect = Mac_GetWidthVectOfRPin(rPinPtr);
        width = widthVect[METAL2_LAYER];
        (void) TmNumberTable(&tableHeadPtr, hId, width, netPtr, 
          localDataPtr->zone);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
GetPinsInRect(pinInstPtr, localDataPtr)
PinInst   *pinInstPtr;
LocalData *localDataPtr;
{
    LinkedPinInst *pinPtr;
    Point         absPos;
    BearId        hId;
    Net           *netPtr;
    float         width;
    float         *widthVect;

    (void) Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
    if ((absPos.p_x >= localDataPtr->area->r_xbot) &&
      (absPos.p_x <= localDataPtr->area->r_xtop) &&
      (absPos.p_y >= localDataPtr->area->r_ybot) &&
      (absPos.p_y <= localDataPtr->area->r_ytop))
    {
        MALLOC(LinkedPinInst *, pinPtr, sizeof (LinkedPinInst));
        pinPtr->pinInstPtr = pinInstPtr;
        pinPtr->nextPtr = *(localDataPtr->pinList);
        *(localDataPtr->pinList) = pinPtr;
        hId = Mac_GetHExtNetIdOfPin(pinInstPtr);
        netPtr = Mac_GetNetOfPin(pinInstPtr);
        widthVect = Mac_GetWidthVectOfPin(pinInstPtr);
        width = widthVect[METAL2_LAYER];
        (void) TmNumberTable(&tableHeadPtr, hId, width,  netPtr, 
          localDataPtr->zone);
    }
    return (BEAR_CONTINUE_SEARCH);
}


/*
 * old code
 *

static void AddPPinToTopOfBnTile();
static void AddPPinToBottomOfBnTile();

static void
AddPPinToTopOfBnTile(bnTilePtr, topList)
Tile      *bnTilePtr;
PseudoPin **topList;
{
    PseudoPin *pPinPtr;

    pPinPtr = *topList;
    *topList = pPinPtr->nextPtr;
    pPinPtr->nextPtr = NIL(PseudoPin);
    Tp_InsertPPinOfBnTile(bnTilePtr, TOP_RIGHT_SIDE, pPinPtr);
}


static void
AddPPinToBottomOfBnTile(bnTilePtr, bottomList)
Tile      *bnTilePtr;
PseudoPin **bottomList;
{
    PseudoPin *pPinPtr;

    pPinPtr = *bottomList;
    *bottomList = pPinPtr->nextPtr;
    pPinPtr->nextPtr = NIL(PseudoPin);
    Tp_InsertPPinOfBnTile(bnTilePtr, BOTTOM_LEFT_SIDE, pPinPtr);
}
*/



