
/* tmHBnUnify.c --
 *
 *          Tim Collins
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmHBnUnify.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


typedef struct _LocalInfo
{
    Rect          *area;
    PseudoPin     *pseudoPin;
    Tile          *tilePtr;
    SideDirection side;
} LocalInfo;


typedef struct _DCellUpdateInfo
{
    DummyCell               *dCellPtr; 
    Rect                    searchRect;
    PseudoPin               *trPPinList;
    PseudoPin               *blPPinList;
    struct _DCellUpdateInfo *nextPtr;
} DCellUpdateInfo;

typedef struct _FixInfo
{
    DummyCell     *dCellPtr;
    PseudoPin     *pPinPtr;
    BearId        oldId;
    BearId        newId;
    SideDirection side;
} FixInfo;

static BearStatus CheckLeftPPins();
static BearStatus CheckRightPPins();
static BearStatus CheckRPins();
static BearStatus CheckPins();
static BearStatus CheckBPPinsOnDCell();
static BearStatus CheckBPPinsOfBnLine();
static BearStatus ChangeTPPinsOfBnLine();
static BearStatus ChangeTBPinsOfBn();
static BearStatus ChangeDCellPPins();
static BearStatus FixDCellSide();
static BearStatus TmHBnUnify();
static void FixInternalPins();
static void AttachHBnPPinsAndDensity();
static void AttachHBnLinePPins();
static void FixBnTileCase();
static void FixHPPinsOnBnLine();
static void FixHPPins();
static void AppendTopPPinLists();
static void AppendBottomPPinLists();
static void FixDCellPPins();
static int SearchBnSides();
static int LastDCellOnBNLine();


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_HBnUnify () --
 *
 *    This procedure deletes all the dummy cells, and transfers their pseudo
 *    pins to the appropriate horizontal bottleneck tiles.
 *
 * Results:
 *    All the dummy cells are deleted.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Tm_HBnUnify(parentCellPtr)
CellMaster *parentCellPtr;
{
    PlaneData       *planeDataPtr;
    WallJunction    *fpgPtr;
    DCellUpdateInfo *dCellUpdateInfoList;
    DCellUpdateInfo *dCellUpdatePtr;
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;
    Tile            *bottomBnTilePtr;
    Tile            *topBnTilePtr;

    tmGlobalParentCellPtr = parentCellPtr;
    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    tmGlobalPlaneDataPtr = planeDataPtr;
    dCellUpdateInfoList = NIL(DCellUpdateInfo);
    /*
     * All the pseudo pin information on the dummy cell and the bottleneck
     * tiles must be saved, in the case of a dummy cell on a bottle neck line
     * or it will be lost when the tile plane is updated!
     */
    (void) Mac_ItrDCell(planeDataPtr->pd_dcellList, TmHBnUnify, 
      (ClientData) &dCellUpdateInfoList);
    dCellUpdatePtr = dCellUpdateInfoList;
    /* unpaint all the dummy cells at once that have to be dealt with in a
       special manner */
    while (dCellUpdatePtr != NIL(DCellUpdateInfo))
    {
        (void) TmGetLAndRBnTilesOfDCell(dCellUpdatePtr->dCellPtr, 
          &leftBnTilePtr, &rightBnTilePtr);
        (void) TmGetBAndTBnTilesOfDCell(dCellUpdatePtr->dCellPtr, 
          &bottomBnTilePtr, &topBnTilePtr);
        (void) Tp_FreeTileProp(leftBnTilePtr);
        (void) Tp_FreeTileProp(rightBnTilePtr);
        (void) Tp_FreeTileProp(bottomBnTilePtr);
        (void) Tp_FreeTileProp(topBnTilePtr);
        /* new change: move from the place before the free */
        (void) Tp_UnpaintDCell(dCellUpdatePtr->dCellPtr, planeDataPtr);
        dCellUpdatePtr = dCellUpdatePtr->nextPtr;
    }

    /* Must identify space tiles before restoring the pseudopins */
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &fpgPtr);

    dCellUpdatePtr = dCellUpdateInfoList;
    while (dCellUpdatePtr != NIL(DCellUpdateInfo))
    {
        if ((dCellUpdatePtr->dCellPtr->class == HBNTI_AND_VBNTI) ||
          (dCellUpdatePtr->dCellPtr->class == HBNTI_AND_VBNLN))
        {
        /*
         * Find the new horizontal bottleneck tile and attach 
         * their pseudo pin lists.
         */
            /* (void) AttachHBnPPins(dCellUpdatePtr); */
        }
        else
        {
        /*
         * Find the new horizontal semi-bottleneck tiles and attach 
         * their pseudo pin lists.
         */
            (void) AttachHBnLinePPins(dCellUpdatePtr);
        }
        dCellUpdatePtr = dCellUpdatePtr->nextPtr;
    }
    (void) Tp_FreeDCellList(planeDataPtr);
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &fpgPtr);
}


static void 
AttachHBnPPinsAndDensity(dummyCellPtr, tPPinList, bPPinList, density)
DummyCell  *dummyCellPtr;
PseudoPin  *tPPinList;
PseudoPin  *bPPinList;
int        density;
{
    Point    hSearchPnt;
    Point    vSearchPnt;
    Tile     *bnTilePtr;

    /*
     * create new space tile properties and attach the ppins
     */
    hSearchPnt.p_x = dummyCellPtr->bbox.r_ll.p_x;
    hSearchPnt.p_y = dummyCellPtr->bbox.r_ll.p_y;
    Tp_HPointToVPoint(&hSearchPnt, &vSearchPnt);
    bnTilePtr = TiSrPointNoHint(Tp_GetVertPlane(tmGlobalPlaneDataPtr),
      &vSearchPnt);
    if (bnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(bnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_VERT_BN_TILE, bnTilePtr);
    }
    bnTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &hSearchPnt);
    if (bnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(bnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_HORZ_BN_TILE, bnTilePtr);
    }
    if (tPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(tPPinList,
          bnTilePtr, TOP_RIGHT_SIDE, TP_BOTTOM); 
    }
    if (bPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(bPPinList,
          bnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM);
    }
    Tp_PutDensityOfTile(density, bnTilePtr);

}


static void
AttachHBnLinePPins(dCellUpdatePtr)
DCellUpdateInfo *dCellUpdatePtr;
{
    Tile         *topSBnTilePtr;
    Tile         *bottomSBnTilePtr;
    Tile         *rightSBnTilePtr;
    Tile         *leftSBnTilePtr;
    BNLineType   bnType;

    (void) TmGetVSBnTilesOfDCell(dCellUpdatePtr->dCellPtr,
      &rightSBnTilePtr, &leftSBnTilePtr);
    if (rightSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(rightSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_VERT_DOM_SBN_TILE, rightSBnTilePtr);
    }
    if (leftSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(leftSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_VERT_DOM_SBN_TILE, leftSBnTilePtr);
    }
    (void) TmGetHSBnTilesOfDCell(dCellUpdatePtr->dCellPtr,
      &topSBnTilePtr, &bottomSBnTilePtr);
    if (topSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(topSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_HORZ_DOM_SBN_TILE, topSBnTilePtr);
    }
    if (bottomSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(bottomSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_HORZ_DOM_SBN_TILE, bottomSBnTilePtr);
    }
    (void) Tp_IsPairOfSBNTiles(&topSBnTilePtr, &bottomSBnTilePtr, &bnType);
    if (bnType == TP_HORZ_B_RIGHT)
    {
        if (dCellUpdatePtr->trPPinList != NIL(PseudoPin))
        {
            (void) Tp_AppendPPinListOfTile(dCellUpdatePtr->trPPinList,
              topSBnTilePtr, BOTTOM_LEFT_SIDE, TP_TOP);
        }
    }
    else
    {
        if (dCellUpdatePtr->trPPinList != NIL(PseudoPin))
        {
            (void) Tp_AppendPPinListOfTile(dCellUpdatePtr->trPPinList,
              topSBnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM);
        }
    }
}



static BearStatus
TmHBnUnify(dummyCellPtr, dCellUpdateInfoListPtr)
DummyCell       *dummyCellPtr;
DCellUpdateInfo **dCellUpdateInfoListPtr;
{
    /*
     * Fix and get the pseudo pins off the bottleneck tiles and the dummy cell.
     */
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;
    Tile            *bottomBnTilePtr;
    Tile            *topBnTilePtr;
    DCellUpdateInfo *updateInfoPtr;
    WallJunction    *fpgPtr;

    (void) FixInternalPins(dummyCellPtr);
    if ((dummyCellPtr->class == HBNTI_AND_VBNTI) || 
      (dummyCellPtr->class == HBNTI_AND_VBNLN))
    {
        (void) FixBnTileCase(dummyCellPtr);
    }
    else
    {
        /* 
         * Bottleneck line case.
         * There are two cases
         * 1. this is the only dummy cell on a horizontal bottleneck line
         *    we must create the save structure, because when this dcell
         *    is removed, a bnline will result.
         * 2. there are other dcells on this bnline, so the removal of this
         *    dcell will not cause damage to the bnline.
         */
        (void) TmGetLAndRBnTilesOfDCell(dummyCellPtr, &leftBnTilePtr, 
          &rightBnTilePtr);
        (void) FixHPPinsOnBnLine(leftBnTilePtr, dummyCellPtr, 
          rightBnTilePtr);
        (void) AppendTopPPinLists(leftBnTilePtr, dummyCellPtr, 
          rightBnTilePtr);
        /* new change: MALLOC moved before Tp_Remove */
        MALLOC(DCellUpdateInfo *, updateInfoPtr, sizeof (DCellUpdateInfo));
        (void) Tp_RemovePPinListOfTile(leftBnTilePtr, TOP_RIGHT_SIDE, 
          TP_BOTTOM, &updateInfoPtr->trPPinList);
        updateInfoPtr->dCellPtr = dummyCellPtr;
        updateInfoPtr->blPPinList = NIL(PseudoPin);
        if (LastDCellOnBNLine(leftBnTilePtr, rightBnTilePtr))
        {
            updateInfoPtr->nextPtr = *dCellUpdateInfoListPtr;
            *dCellUpdateInfoListPtr = updateInfoPtr;
        }
        else
        {
            (void) TmGetBAndTBnTilesOfDCell(dummyCellPtr, &bottomBnTilePtr, 
              &topBnTilePtr);
            (void) Tp_FreeTileProp(leftBnTilePtr);
            (void) Tp_FreeTileProp(rightBnTilePtr);
            (void) Tp_FreeTileProp(bottomBnTilePtr);
            (void) Tp_FreeTileProp(topBnTilePtr);
            /* new change: move from the place before the free */
            (void) Tp_UnpaintDCell(dummyCellPtr, tmGlobalPlaneDataPtr);
            Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
                (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &fpgPtr);
            /* The bnLine is still artificially defined as 1-unit bnTile */
            /* So reGet the leftBnTilePtr and put the PPins back on top */
            (void) TmGetLAndRBnTilesOfDCell(dummyCellPtr, 
                                            &leftBnTilePtr, 
                                            &rightBnTilePtr);
            (void) Tp_AppendPPinListOfTile(updateInfoPtr->trPPinList,
                                           leftBnTilePtr, 
                                           TOP_RIGHT_SIDE, TP_BOTTOM);
            FREE((char *) updateInfoPtr);
        }
    }
}


static int
LastDCellOnBNLine(leftBnTilePtr, rightBnTilePtr)
Tile            *leftBnTilePtr;
Tile            *rightBnTilePtr;
{
    Rect  rect;
    Point lPoint;
    Point rPoint;
    
    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, leftBnTilePtr, &rect);
    lPoint.p_x = rect.r_xbot - 1;
    lPoint.p_y = rect.r_ybot + 1;
    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, rightBnTilePtr, &rect);
    rPoint.p_x = rect.r_xtop + 1;
    rPoint.p_y = rect.r_ybot + 1;
    if (Tp_IsDCellTile(
      TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr), &lPoint)) ||
      Tp_IsDCellTile(
      TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr), &rPoint)))
    {
        return (0);
    }
    else
    {
        return (1);
    }
}


static void
FixBnTileCase(dummyCellPtr)
DummyCell       *dummyCellPtr;
{
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;
    Tile            *bottomBnTilePtr;
    Tile            *topBnTilePtr;
    PseudoPin       *tPPinList;
    PseudoPin       *bPPinList;
    WallJunction    *fpgPtr;
    int             densityTotal;
    int             densityLeft;
    int             densityRight;
    Rect            dummyBox;

    (void) TmGetLAndRBnTilesOfDCell(dummyCellPtr, &leftBnTilePtr, 
      &rightBnTilePtr);
        /* Calculate Density of Old Tiles */
    densityRight = Tp_GetDensityOfTile(rightBnTilePtr);
    densityLeft = Tp_GetDensityOfTile(leftBnTilePtr);
    if (densityRight < 0 || densityLeft  < 0)
    {
       if (densityRight >= 0) 
          densityRight = -(Tp_GetCapacityOfTile(rightBnTilePtr));
       if (densityLeft >= 0) 
          densityLeft = -(Tp_GetCapacityOfTile(leftBnTilePtr));
           dummyBox = dummyCellPtr->bbox;
           densityTotal = -(dummyBox.r_xtop - dummyBox.r_xbot)
                      + densityRight + densityLeft;
    }
    else
    {
       densityTotal = 0;
    } 
    (void) TmGetBAndTBnTilesOfDCell(dummyCellPtr, &bottomBnTilePtr, 
      &topBnTilePtr);
    (void) AppendTopPPinLists(leftBnTilePtr, dummyCellPtr, rightBnTilePtr);
    (void) AppendBottomPPinLists(leftBnTilePtr, dummyCellPtr, rightBnTilePtr);
    (void) Tp_RemovePPinListOfTile(leftBnTilePtr, TOP_RIGHT_SIDE, TP_BOTTOM,
      &tPPinList);
    (void) Tp_RemovePPinListOfTile(leftBnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM,
      &bPPinList);
    (void) Tp_FreeTileProp(leftBnTilePtr);
    (void) Tp_FreeTileProp(rightBnTilePtr);
    (void) Tp_FreeTileProp(bottomBnTilePtr);
    (void) Tp_FreeTileProp(topBnTilePtr);
    /* new change: move from the place before the free */
    (void) Tp_UnpaintDCell(dummyCellPtr, tmGlobalPlaneDataPtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &fpgPtr);
    (void) AttachHBnPPinsAndDensity(dummyCellPtr, tPPinList, bPPinList,
                        densityTotal);
}


static void
FixInternalPins(dummyCellPtr)
DummyCell       *dummyCellPtr;
{
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;

    (void) TmGetLAndRBnTilesOfDCell(dummyCellPtr, &leftBnTilePtr, 
      &rightBnTilePtr);
    if ((dummyCellPtr->class == HBNTI_AND_VBNTI) ||
      (dummyCellPtr->class == HBNTI_AND_VBNLN))
    {
        (void) FixHPPins(leftBnTilePtr, dummyCellPtr, rightBnTilePtr);
    }
}


static void
FixHPPinsOnBnLine(leftBnTilePtr, dCellPtr, rightBnTilePtr)
Tile      *leftBnTilePtr;
DummyCell *dCellPtr;
Tile      *rightBnTilePtr;
{
    /*
     * Since when the dummy cell is unpainted, the bottleneck tiles will turn
     * into a bottleneck line, we have to make the pseudo pin id's consistent.
     * (two pseudo pins turn into one pseudo pin)
     */
    (void) Tp_ItrPPinsOfSide(rightBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      CheckBPPinsOfBnLine, (ClientData) rightBnTilePtr);
    (void) Tp_ItrPPinsOfSide(leftBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      CheckBPPinsOfBnLine, (ClientData) leftBnTilePtr);
    (void) Mac_ItrPPinOfDCell(dCellPtr, TOP_SIDE, CheckBPPinsOnDCell,
      (ClientData) dCellPtr);
}


static BearStatus
CheckBPPinsOfBnLine(pPinPtr, bnTilePtr)
PseudoPin *pPinPtr;
Tile      *bnTilePtr;
{

    (void) Tp_ItrPPinsOfSide(bnTilePtr,
      (SideDirection) BOTTOM_LEFT_SIDE,
      ChangeTPPinsOfBnLine, (ClientData) pPinPtr);
    return (BEAR_CONTINUE_SEARCH);
}
    

static BearStatus
ChangeTPPinsOfBnLine(pPinPtr, tPPinPtr)
PseudoPin *pPinPtr;
PseudoPin *tPPinPtr;
{
    BearId tIntId;
    BearId tExtId;
    BearId bIntId;
    BearId bExtId;


        /*       tExtId                                          */
        /*    -----0------ tPPinPtr                              */
        /*    |   tIntId |               extId = tExtId          */
        /*    |          |         ----> ------0------- tPPinPtr */
        /*    |   bIntId |               intId = bExtId          */
        /*    -----0------ pPinPtr                               */
        /*        bExtId                                         */

    tIntId = Tp_GetIntNetIdOfPPin(tPPinPtr);
    tExtId = Tp_GetExtNetIdOfPPin(tPPinPtr);
    bIntId = Tp_GetIntNetIdOfPPin(pPinPtr);
    bExtId = Tp_GetExtNetIdOfPPin(pPinPtr);
    if (bIntId == tIntId)
    {
        /* SS: Changed the pin-assignment */
        (void) Tp_PutIntNetIdOfPPin(bExtId, tPPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckBPPinsOnDCell(pPinPtr, dCellPtr)
PseudoPin *pPinPtr;
DummyCell *dCellPtr;
{
    (void) Mac_ItrPPinOfDCell(dCellPtr, BOTTOM_SIDE, ChangeTPPinsOfBnLine,
      (ClientData) pPinPtr);
    return (BEAR_CONTINUE_SEARCH);
}


static void
AppendTopPPinLists(leftBnTilePtr, dCellPtr, rightBnTilePtr)
Tile      *leftBnTilePtr;
DummyCell *dCellPtr;
Tile      *rightBnTilePtr;
{
    PseudoPin       *dummyTopPPinList;
    PseudoPin       *rightBnTopPPinList;

    Mac_RemovePPinListOfDCell(dCellPtr, TOP_SIDE, &dummyTopPPinList);
    Tp_RemovePPinListOfTile(rightBnTilePtr, TOP_RIGHT_SIDE, TP_BOTTOM,
      &rightBnTopPPinList);
    if (dummyTopPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(dummyTopPPinList, leftBnTilePtr,
          TOP_RIGHT_SIDE, TP_BOTTOM);
    }
    if (rightBnTopPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(rightBnTopPPinList, leftBnTilePtr,
          TOP_RIGHT_SIDE, TP_BOTTOM);
    }
}


static void
AppendBottomPPinLists(leftBnTilePtr, dCellPtr, rightBnTilePtr)
Tile      *leftBnTilePtr;
DummyCell *dCellPtr;
Tile      *rightBnTilePtr;
{
    PseudoPin       *dummyBotPPinList;
    PseudoPin       *rightBnBotPPinList;

    Mac_RemovePPinListOfDCell(dCellPtr, BOTTOM_SIDE, &dummyBotPPinList);
    Tp_RemovePPinListOfTile(rightBnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM,
      &rightBnBotPPinList);
    if (dummyBotPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(dummyBotPPinList, leftBnTilePtr,
          BOTTOM_LEFT_SIDE, TP_BOTTOM);
    }
    if (rightBnBotPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(rightBnBotPPinList, leftBnTilePtr,
          BOTTOM_LEFT_SIDE, TP_BOTTOM);
    }
}


static void
FixHPPins(leftBnTilePtr, dCellPtr, rightBnTilePtr)
Tile      *leftBnTilePtr;
DummyCell *dCellPtr;
Tile      *rightBnTilePtr;
{
    /*
     * Make the id's consistent across the bottleneck tiles so that all
     * pins and pseudo pins on the same net have the same internal id.
     */
    (void) Mac_ItrPPinOfDCell(dCellPtr, LEFT_SIDE, CheckLeftPPins,
      (ClientData) leftBnTilePtr);
    (void) Mac_ItrPPinOfDCell(dCellPtr, RIGHT_SIDE, CheckRightPPins,
      (ClientData) rightBnTilePtr);
}


static BearStatus
CheckLeftPPins(pPinPtr, lBnTilePtr)
PseudoPin *pPinPtr;
Tile      *lBnTilePtr;
{
    /* 
     * Propagate consistent id value across all sides of the left bottleneck
     * tile and the part of the cell that abuts it.
     */
    Rect      searchRect;
    LocalInfo localInfo;

    (void) Tp_ItrPPinsOfSide(lBnTilePtr,
      (SideDirection) BOTTOM_LEFT_SIDE,
      ChangeTBPinsOfBn, (ClientData) pPinPtr);
    (void) Tp_ItrPPinsOfSide(lBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      ChangeTBPinsOfBn, (ClientData) pPinPtr);
    /*
     * Get adjacent cell instance and check pin which are on the edge of the
     * right bottleneck tile.
     */
    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, lBnTilePtr, &searchRect);
    searchRect.r_ur.p_x = searchRect.r_ll.p_x + 1;
    searchRect.r_ll.p_x -= 1;
    localInfo.area = &searchRect;
    localInfo.pseudoPin = pPinPtr;
    localInfo.side = RIGHT_SIDE;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &searchRect,
      SearchBnSides,
      (ClientData) &localInfo);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
      tmGlobalParentCellPtr), CheckPins, (ClientData) &localInfo);
    return (BEAR_CONTINUE_SEARCH);
}


static int
SearchBnSides(tilePtr, localInfoPtr)
Tile      *tilePtr;
LocalInfo *localInfoPtr;
{
    CellInst      *cellInstPtr;
    RouteCell     *routeCellPtr;
    FixInfo       fixInfo;

    if (Tp_IsCellInstTile(tilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        (void) Mac_ItrPinOfCellInst (cellInstPtr, CheckPins,
          (ClientData) localInfoPtr);
    }
    else
    {
#ifdef USE_DR_CODE
        if (Tp_IsRCellTile(tilePtr))
        {
            routeCellPtr = Tp_GetRCellOfTile(tilePtr);
            (void) Mac_ItrSidePinOfRCell(routeCellPtr, CheckRPins,
              (ClientData) localInfoPtr);
        }
        else
#endif /* USE_DR_CODE */
        {
            if (Tp_IsDCellTile(tilePtr))
            {
                fixInfo.dCellPtr = Tp_GetDCellOfTile(tilePtr);
                fixInfo.pPinPtr = localInfoPtr->pseudoPin;
                fixInfo.side = localInfoPtr->side;
                (void) Mac_ItrPPinOfDCell(fixInfo.dCellPtr, 
                  localInfoPtr->side,
                  ChangeDCellPPins, (ClientData) &fixInfo);
            }
        }
    }
    return (0);
}


static BearStatus
CheckRightPPins(pPinPtr, rBnTilePtr)
PseudoPin *pPinPtr;
Tile      *rBnTilePtr;
{
    /* 
     * Propagate consistent id values across all sides of the right bottleneck
     * tile and the part of the cell that abuts it.
     */
    Rect      searchRect;
    LocalInfo localInfo;

    (void) Tp_ItrPPinsOfSide(rBnTilePtr,
      (SideDirection) BOTTOM_LEFT_SIDE,
      ChangeTBPinsOfBn, (ClientData) pPinPtr);
    (void) Tp_ItrPPinsOfSide(rBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      ChangeTBPinsOfBn, (ClientData) pPinPtr);
    /*
     * Get adjacent cell instance and check pin which are on the edge of the
     * right bottleneck tile.
     */
    (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, rBnTilePtr, &searchRect);
    searchRect.r_ll.p_x = searchRect.r_ur.p_x - 1;
    searchRect.r_ur.p_x += 1;
    localInfo.area = &searchRect;
    localInfo.pseudoPin = pPinPtr;
    localInfo.side = LEFT_SIDE;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &searchRect,
      SearchBnSides,
      (ClientData) &localInfo);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
      tmGlobalParentCellPtr), CheckPins, (ClientData) &localInfo);
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
ChangeDCellPPins(pPinPtr, fixInfoPtr)
PseudoPin *pPinPtr;
FixInfo   *fixInfoPtr;
{
    BearId pExtId;
    BearId extId;
    BearId intId;

    pExtId = Tp_GetExtNetIdOfPPin(pPinPtr);
    extId = Tp_GetExtNetIdOfPPin(fixInfoPtr->pPinPtr);
    intId = Tp_GetIntNetIdOfPPin(fixInfoPtr->pPinPtr);
    if (pExtId == extId)
    {
        (void) Tp_PutExtNetIdOfPPin(intId, pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}

/*
static void
FixDCellPPins(fixInfoPtr, side)
FixInfo       *fixInfoPtr;
SideDirection side;
{
    if (side != RIGHT_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr, RIGHT_SIDE, 
          FixDCellSide, (ClientData) fixInfoPtr);
    }
    if (side != LEFT_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr, LEFT_SIDE, 
          FixDCellSide, (ClientData) fixInfoPtr);
    }
    if (side != TOP_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr, TOP_SIDE, 
          FixDCellSide, (ClientData) fixInfoPtr);
    }
    if (side != BOTTOM_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr, BOTTOM_SIDE, 
          FixDCellSide, (ClientData) fixInfoPtr);
    }
}


static BearStatus
FixDCellSide(pPinPtr, fixInfoPtr)
PseudoPin  *pPinPtr;
FixInfo    *fixInfoPtr;
{
    BearId intId;

    intId = Tp_GetIntNetIdOfPPin(fixInfoPtr->pPinPtr);
    if (intId == fixInfoPtr->oldId)
    {
        (void) Tp_PutIntNetIdOfPPin(fixInfoPtr->newId, pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}
*/


static BearStatus
ChangeTBPinsOfBn(pPinPtr, lrPPinPtr)
PseudoPin *pPinPtr;
PseudoPin *lrPPinPtr;
{
    BearId lrIntId;
    BearId intId;
    BearId lrExtId;

    lrIntId = Tp_GetIntNetIdOfPPin(lrPPinPtr);
    intId = Tp_GetIntNetIdOfPPin(pPinPtr);
    lrExtId = Tp_GetExtNetIdOfPPin(lrPPinPtr);
    if (intId == lrExtId)
    {
        (void) Tp_PutIntNetIdOfPPin(lrIntId, pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckPins(pinPtr,localInfoPtr)
PinInst   *pinPtr;
LocalInfo *localInfoPtr;
{
    Point  absPos;
    BearId pIntId;
    BearId pExtId;
    BearId extId;

    (void) Mac_GetAbsPosOfPin(pinPtr, &absPos);
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
      (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
      (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        pIntId = Tp_GetIntNetIdOfPPin(localInfoPtr->pseudoPin);
        extId = Mac_GetHExtNetIdOfPin(pinPtr);
        pExtId = Tp_GetExtNetIdOfPPin(localInfoPtr->pseudoPin);
        if (extId == pExtId)
        {
            (void) Mac_PutHExtNetIdOfPin(pIntId, pinPtr);
            /******NEW*******/
            (void) Mac_PutExtNetIdOfPin(pIntId, pinPtr);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckRPins(rPinPtr,localInfoPtr)
RoutePin  *rPinPtr;
LocalInfo *localInfoPtr;
{
    Point  absPos;
    BearId pIntId;
    BearId pExtId;
    BearId extId;

    (void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
      (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
      (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        pIntId = Tp_GetIntNetIdOfPPin(localInfoPtr->pseudoPin);
        extId = Mac_GetHExtNetIdOfRPin(rPinPtr);
        pExtId = Tp_GetExtNetIdOfPPin(localInfoPtr->pseudoPin);
        if (extId == pExtId)
        {
            (void) Mac_PutHExtNetIdOfRPin(pIntId, rPinPtr);
            /******NEW*******/
            (void) Mac_PutExtNetIdOfRPin(pIntId, rPinPtr);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}




