
/* tmHFSplit.c --
 *
 *          Tim Collins
 *     (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmHFSplit.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


extern ViewType viewTypeFlag;


/* ---------- Local Procedures ---------- */
static int DeleteVPPins();
static int SplitEachDominantTile();
static int JunctInRect();
static int CheckPinInstOfSolidTile();
static int CheckBPinsOfDummyCell();
static int CheckTPinsOfDummyCell();
static BearStatus GetLeftmostWallJunction();
static BearStatus DoPinInstsInRect();
static BearStatus DoRPinsInRect();
static BearStatus GetPPinsOfDCell();
static BearStatus FreePPinsOfBnLine();
static void DoBnTileOrLine();
static void DoJunctionRegion();
static bool PinOnTopSideOfTile();
static BearStatus FindPPin();


/* ---------- Local Global Variables ---------- */
static ZoneData *zonePtr;
static WallJunction *fpgPtr;
static Rect dominantRect;
static Table *tableHeadPtr;
static int zoneNum;


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_HFSplit () --
 *
 * This procedure looks at the horizontal dominant tiles one by one, and
 * translates the global routing information that is on the horizontal plane
 * and puts it on both planes.
 *
 * Results:
 *	 All the routing information is on both planes.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

void
Tm_HFSplit(parentCellPtr)
CellMaster *parentCellPtr;
{
	PlaneData *planeDataPtr;

	tmGlobalParentCellPtr = parentCellPtr;
	fpgPtr = Mac_GetFpgOfCell(parentCellPtr);
	planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
	tmGlobalPlaneDataPtr = planeDataPtr;

	/* Delete all vertical ppins so that all vert bn tiles are empty */
	(void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
	  Tp_GetVertPlaneBnd(planeDataPtr),
	  DeleteVPPins,
	  (ClientData) NULL);
	(void) Tp_ItrVertBnLine(planeDataPtr, FreePPinsOfBnLine, 
	  (ClientData) NULL);

	(void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
	  Tp_GetHorzPlaneBnd(planeDataPtr),
	  SplitEachDominantTile,
	  (ClientData) planeDataPtr);

	(void) Fpg_CreateFloorPlanGraph(planeDataPtr, &fpgPtr);
	viewTypeFlag = BOTH_DONE;
}


/* ARGSUSED */
static BearStatus
FreePPinsOfBnLine(bnLinePtr, clientData)
BNLine     *bnLinePtr;
ClientData clientData;
{
	(void) Tp_FreePPinsOfBnLine(bnLinePtr);
	return (BEAR_CONTINUE_SEARCH);
}


/* ARGSUSED */
static int
DeleteVPPins(tilePtr, clientData)
Tile       *tilePtr;
ClientData clientData;
{
	if (Tp_IsBottleneckTile(tilePtr))
	{
		(void) Tp_FreePPinsOfTile(tilePtr);
	}
	return (0);
}


static int
SplitEachDominantTile(tilePtr, planeDataPtr)
Tile      *tilePtr;
PlaneData *planeDataPtr;
{
	if (Tp_IsDominantTile(tilePtr))
	{
		(void)TmSplitDTile(tilePtr, planeDataPtr);
	}
	return (0);
}


void
TmSplitDTile(dmntTilePtr, planeDataPtr)
Tile      *dmntTilePtr;
PlaneData *planeDataPtr;
{
	/*
	 * By search the dominant tile for vertical bottlenecks tiles and lines
	 * from left to right, we propogate across the face the new horizontal 
	 * values for the pseudo pin id's
	 */

	WallSegment        *wallSegPtr;
	WallJunction       *junctPtr;
	ZoneData           *zoneDataHeadPtr;

	zoneNum = 1;
	tableHeadPtr = NIL(Table);
	(void) Tp_HorzTileToRect(planeDataPtr, dmntTilePtr, &dominantRect);
	(void) Fpg_ItrWallJunct(fpgPtr, GetLeftmostWallJunction, 
		(ClientData) &junctPtr);
/*
 * code change : by george : on 23 oct 89
 *
	MALLOC(ZoneData *, zonePtr, sizeof (ZoneData));
 */
	zonePtr = TmMakeZoneData();
	zoneDataHeadPtr = zonePtr;
	DoJunctionRegion(junctPtr);
	while ((wallSegPtr = Fpg_GetEastSegOfWallJunct(junctPtr)) 
	  != NIL(WallSegment))
	{
		zoneNum++;
/*
 * code change : by george : on 23 oct 89
 *
		MALLOC(ZoneData *, zonePtr->nextPtr, sizeof (ZoneData));
 */
		zonePtr->nextPtr = TmMakeZoneData();
		zonePtr = zonePtr->nextPtr;
		DoBnTileOrLine(wallSegPtr);
		zoneNum++;
		junctPtr = Fpg_GetSEJunctOfWallSeg(wallSegPtr);
/*
 * code change : by george : on 23 oct 89
 *
		MALLOC(ZoneData *, zonePtr->nextPtr, sizeof (ZoneData));
 */
		zonePtr->nextPtr = TmMakeZoneData();
		zonePtr = zonePtr->nextPtr;
		DoJunctionRegion(junctPtr);
	}
	zonePtr->nextPtr = NIL(ZoneData);
	TmHSplitByZones(zoneNum, tableHeadPtr, zoneDataHeadPtr);
}


static BearStatus
GetLeftmostWallJunction(junctPtr, leftJunctPtrPtr)
WallJunction *junctPtr;
WallJunction **leftJunctPtrPtr;
{
	WallSegment *wallSegPtr;
	Point       *pointPtr;

	pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
	if (JunctInRect(pointPtr))
	{
		*leftJunctPtrPtr = junctPtr;
		while ((wallSegPtr = 
		  Fpg_GetWestSegOfWallJunct(*leftJunctPtrPtr)) 
	  	  != NIL(WallSegment))
		{
			*leftJunctPtrPtr = Fpg_GetNWJunctOfWallSeg(wallSegPtr);
		}
		return (BEAR_STOP_SEARCH);
	}
	return (BEAR_CONTINUE_SEARCH);
}
	

static int
JunctInRect(point)
Point *point;
{
	if ((point->p_x >= dominantRect.r_ll.p_x) &&
	  (point->p_x <= dominantRect.r_ur.p_x) &&
	  (point->p_y >= dominantRect.r_ll.p_y) &&
	  (point->p_y <= dominantRect.r_ur.p_y))
	{
		return (1);
	}
	else
	{
		return (0);
	}
}


static void
DoBnTileOrLine(wallSegPtr)
WallSegment *wallSegPtr;
{
	BottleNeck *bnPtr;
	Rect       bnRect;

	bnPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
	zonePtr->bnPtr = bnPtr;
	if (bnPtr->unType == TP_BN_LINE) /* bottleneck line */
	{
		zonePtr->zoneType = V_LINE;
	}
	else
	{                               /* bottleneck tile */
		zonePtr->zoneType = V_TILE;
		zonePtr->pinInstList = NIL(LinkedPinInst);
		zonePtr->pPinList = NIL(LinkedPPin);
		zonePtr->rPinList = NIL(LinkedRPin);
		(void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, 
		  bnPtr->unValue.bnTilePtr, &bnRect);
		bnRect.r_ur.p_y += 1;
		bnRect.r_ll.p_y -= 1;
		(void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
		  &bnRect,
		  CheckPinInstOfSolidTile,
		  (ClientData) &bnRect);
		/* for i/o pins */
		(void) Mac_ItrPinOfCellInst(
			Mac_GetInstOfCoreMaster(tmGlobalParentCellPtr),
			DoPinInstsInRect, (ClientData) &bnRect);
		bnRect.r_ll.p_y += 2;
		(void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
		  &bnRect,
		  CheckBPinsOfDummyCell,
		  (ClientData) NULL);
		bnRect.r_ll.p_y -= 2;
		bnRect.r_ur.p_y -= 2;
		(void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
		  &bnRect,
		  CheckTPinsOfDummyCell,
		  (ClientData) NULL);
	}
}


static void
DoJunctionRegion(junctPtr)
WallJunction *junctPtr;
{
	PinsAroundJunction pinLists;
	LinkedPPin         *pPinPtr;
	LinkedRPin         *rPinPtr;
	LinkedPinInst      *pinInstPtr;
	Net                *netPtr;
	BearId             id;
	Tile               *tilePtr;
	Point              *pointPtr;
	Rect               rect;
	float              width;
	float              *widthVect;

	(void)Tm_PinsAroundJunction(tmGlobalParentCellPtr, tmGlobalPlaneDataPtr, 
	  junctPtr, &pinLists);
	pPinPtr = pinLists.horzPPinList;
	zonePtr->pPinList = pinLists.horzPPinList;
	pinInstPtr = pinLists.pinInstList;
	zonePtr->pinInstList = pinLists.pinInstList;
	rPinPtr = pinLists.rPinList;
	zonePtr->rPinList = pinLists.rPinList;
	zonePtr->zoneType = JUNCTION;
	while (pinInstPtr != NIL(LinkedPinInst))
	{
		id = Mac_GetHExtNetIdOfPin(pinInstPtr->pinInstPtr);
		netPtr = Mac_GetNetOfPin(pinInstPtr->pinInstPtr);
		widthVect = Mac_GetWidthVectOfPin(pinInstPtr->pinInstPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id, width, netPtr, zoneNum);
		pinInstPtr = pinInstPtr->nextPtr;
	}
	while (rPinPtr != NIL(LinkedRPin))
	{
		id = Mac_GetHExtNetIdOfRPin(rPinPtr->rPinPtr);
		netPtr = Mac_GetNetPtrOfRPin(rPinPtr->rPinPtr);
		widthVect = Mac_GetWidthVectOfRPin(rPinPtr->rPinPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id, width, netPtr, zoneNum);
		rPinPtr = rPinPtr->nextPtr;
	}
	while (pPinPtr != NIL(LinkedPPin))
	{
		tilePtr = Tp_GetTileOfPPin(pPinPtr->ppinPtr);
		(void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, tilePtr, &rect);
		pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
		if (Tp_IsSBNTile(tilePtr) == TRUE)
		{
		    if (rect.r_ybot > pointPtr->p_y)
		    {
			id = Tp_GetIntNetIdOfPPin(pPinPtr->ppinPtr);
		    }
		    else if (rect.r_ytop < pointPtr->p_y)
		    {
			id = Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr);
		    }
		    else if (PinOnTopSideOfTile(pPinPtr->ppinPtr, tilePtr))
		    {
			id = Tp_GetIntNetIdOfPPin(pPinPtr->ppinPtr);
		    }
		    else
		    {
			id = Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr);
		    }

		}
		else
		{
			id = Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr);
		}
		netPtr = Tp_GetNetOfPPin(pPinPtr->ppinPtr);
		width = Tp_GetWidthOfPPin(pPinPtr->ppinPtr);
		(void) TmNumberTable(&tableHeadPtr, id, width, netPtr, zoneNum);
		pPinPtr = pPinPtr->nextPtr;
	}
}


static int
CheckPinInstOfSolidTile(tilePtr, searchRectPtr)
Tile      *tilePtr;
Rect      *searchRectPtr;
{
	CellInst      *cellInstPtr;
	RouteCell     *routeCellPtr;

	if (Tp_IsCellInstTile(tilePtr))
	{
		cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
		(void) Mac_ItrPinOfCellInst(cellInstPtr, DoPinInstsInRect,
		  (ClientData) searchRectPtr);
	}
	if (Tp_IsRCellTile(tilePtr))
	{
		routeCellPtr = Tp_GetRCellOfTile(tilePtr);
		(void) Mac_ItrSidePinOfRCell(routeCellPtr, DoRPinsInRect,
		(ClientData) searchRectPtr);
	}
	return (0);
}


/* ARGSUSED */
static int
CheckBPinsOfDummyCell(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
	DummyCell *dCellPtr;

	if (Tp_IsDCellTile(tilePtr))
	{
		dCellPtr = Tp_GetDCellOfTile(tilePtr);
		(void) Mac_ItrPPinOfDCell(dCellPtr, BOTTOM_SIDE, GetPPinsOfDCell,
		  (ClientData) NULL);
	}
	return (0);
}


/* ARGSUSED */
static int
CheckTPinsOfDummyCell(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
	DummyCell *dCellPtr;

	if (Tp_IsDCellTile(tilePtr))
	{
		dCellPtr = Tp_GetDCellOfTile(tilePtr);
		(void) Mac_ItrPPinOfDCell(dCellPtr, TOP_SIDE, GetPPinsOfDCell,
		  (ClientData) NULL);
	}
	return (0);
}


static BearStatus
DoRPinsInRect(rPinPtr,searchRectPtr)
RoutePin  *rPinPtr;
Rect      *searchRectPtr;
{
	Point      absPos;
	LinkedRPin *linkRPinPtr;
	BearId     id;
	Net        *netPtr;
	float      width;
	float      *widthVect;

	(void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
	if ((absPos.p_x >= searchRectPtr->r_ll.p_x) &&
	  (absPos.p_x <= searchRectPtr->r_ur.p_x) &&
	  (absPos.p_y >= searchRectPtr->r_ll.p_y + 1) &&
	  (absPos.p_y <= searchRectPtr->r_ur.p_y - 1))
	{
		id = Mac_GetHExtNetIdOfRPin(rPinPtr);
		netPtr = Mac_GetNetPtrOfRPin(rPinPtr);
		widthVect = Mac_GetWidthVectOfRPin(rPinPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id, width, netPtr, zoneNum);
		MALLOC(LinkedRPin *, linkRPinPtr, sizeof (LinkedRPin));
		linkRPinPtr->rPinPtr = rPinPtr;
		linkRPinPtr->nextPtr = zonePtr->rPinList;
		zonePtr->rPinList = linkRPinPtr;
	}
	return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
DoPinInstsInRect(pinPtr,searchRectPtr)
PinInst   *pinPtr;
Rect      *searchRectPtr;
{
	Point         absPos;
	LinkedPinInst *pinInstPtr;
	BearId        id;
	Net           *netPtr;
	float         width;
	float         *widthVect;

	(void) Mac_GetAbsPosOfPin(pinPtr, &absPos);
	if ((absPos.p_x >= searchRectPtr->r_ll.p_x) &&
	  (absPos.p_x <= searchRectPtr->r_ur.p_x) &&
	  (absPos.p_y >= searchRectPtr->r_ll.p_y + 1) &&
	  (absPos.p_y <= searchRectPtr->r_ur.p_y - 1))
	{
		id = Mac_GetHExtNetIdOfPin(pinPtr);
		netPtr = Mac_GetNetOfPin(pinPtr);
		widthVect = Mac_GetWidthVectOfPin(pinPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id, width, netPtr, zoneNum);
		MALLOC(LinkedPinInst *, pinInstPtr, sizeof (LinkedPinInst));
		pinInstPtr->pinInstPtr = pinPtr;
		pinInstPtr->nextPtr = zonePtr->pinInstList;
		zonePtr->pinInstList = pinInstPtr;
	}
	return (BEAR_CONTINUE_SEARCH);
}


/* ARGSUSED */
static BearStatus
GetPPinsOfDCell(ppinPtr,clientData)
PseudoPin  *ppinPtr;
ClientData clientData;
{
	BearId     id;
	LinkedPPin *lnPPinPtr;
	Net        *netPtr;
	float      width;

	id = Tp_GetExtNetIdOfPPin(ppinPtr);
	netPtr = Tp_GetNetOfPPin(ppinPtr);
	width = Tp_GetWidthOfPPin(ppinPtr);
	(void) TmNumberTable(&tableHeadPtr, id, width, netPtr, zoneNum);

	MALLOC(LinkedPPin *, lnPPinPtr, sizeof (LinkedPPin));
	lnPPinPtr->ppinPtr = ppinPtr;
	lnPPinPtr->nextPtr = zonePtr->pPinList;
	zonePtr->pPinList = lnPPinPtr;

	return (BEAR_CONTINUE_SEARCH);
}


static bool
PinOnTopSideOfTile(pPinPtr, tilePtr)
PseudoPin *pPinPtr;
Tile *tilePtr;
{
     BearStatus status;

     status = Tp_ItrPPinsOfSide(tilePtr, TOP_RIGHT_SIDE, FindPPin, 
				(ClientData) pPinPtr);

     return (status == BEAR_STOP_SEARCH);
}


static BearStatus
FindPPin(tmpPinPtr, pPinPtr)
PseudoPin *tmpPinPtr;
PseudoPin *pPinPtr;
{
     return ((tmpPinPtr == pPinPtr) ? BEAR_STOP_SEARCH : BEAR_CONTINUE_SEARCH);
}


