
/* tmHFUnify.c --
 *
 *           Tim Collins
 *      (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmHFUnify.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"

typedef struct _LocalData
{
    Rect         rect;
    WallJunction *junctPtr;
} LocalData;

typedef struct _LocalInfo
{
    Rect               *area;
    PseudoPin          *pseudoPin;
    Tile               *tilePtr;
    PinsAroundJunction *pinListsPtr;
} LocalInfo;

typedef struct _TileQueue
{
    Tile              *tilePtr;
    int               left;
    int               paired;
    struct _TileQueue *nextPtr;
} TileQueue;

#define bnTileTopDead  1
#define bnTileBotDead  2
#define bnTileBothDead 3
#define bnTileAlive    4


/* ---------- Local Procedures ---------- */
static int UnifyEachDominantTile();
static int ItrSolidTile();
static int GetPinsOfSolidTile();
static BearStatus CheckAgainstJunctPins();
static BearStatus CheckAgainstLeftPin();
static BearStatus CheckTopBottomPin();
static BearStatus CheckTopBottomRPin();
static BearStatus CheckLeftRightPin();
static int GetAndSortBN();
static BearStatus CheckDCellPPins();
static BearStatus GetPPinsOfDCell();
static BearStatus GetRPinsInRect();
static BearStatus GetPinsInRect();
static BearStatus GetPPins();
static BearStatus FindBnLine();
static void ItrPinsOnTopAndBot();
static void InsertBnInList();
static void PropagateExtToHExt();
static BearStatus PropagateExtIdOfIOPads();
static void CopyPinListsToFaceLists();
static void FixTileList();
static BNLine * TmGetVBnLineOfSBnTiles();
int    GetStatusOfVBnTile();
void Tm_GetPinsInVBottleNeckRegion2();
static BearStatus FindJunctInRect();


/* ---------- Local Global Variables ---------- */
static int globalCopyFlag;
static Rect tmGlobalDmntRect;


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_HFUnify () --
 *
 * This procedure looks at the horizontal dominant tiles one by one, and
 * translates the global routing information that is on both planes, and puts
 * it on the horizontal plane.
 *
 * Results:
 *   All the routing information is on the horizontal plane.
 *
 * Side effects:
 *   All the routing information on the vertical plane is no good anymore.
 * ----------------------------------------------------------------------------
 */

void
Tm_HFUnify(copyFlag)
int copyFlag;
{

    tmGlobalPlaneDataPtr = Mac_GetCellPlaneOfCell(tmGlobalParentCellPtr);
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      Tp_GetHorzPlaneBnd(tmGlobalPlaneDataPtr),
      UnifyEachDominantTile,
      (ClientData) copyFlag);
}


static int
UnifyEachDominantTile(tilePtr, copyFlag)
Tile      *tilePtr;
int       copyFlag;
{
    Rect dmntRect;

    if (Tp_IsDominantTile(tilePtr))
    {
        (void) Tp_HorzTileToRect(tmGlobalPlaneDataPtr, tilePtr, 
          &dmntRect);
        (void) TmHRectUnify(tmGlobalPlaneDataPtr, dmntRect, copyFlag);
    }
    return (0);
}



PinsAroundJunction *
Tm_PinsAroundHFace(parentCellPtr, planeDataPtr, rect, includeInsidePPins)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
Rect       rect;
bool       includeInsidePPins;
{
    Point              *jPoint;
    Rect               vertRect;
    TileQueue          *bnTileList;
    TileQueue          *bnTileQue;
    WallSegment        *wallSegPtr;
    WallJunction       *junctPtr;
    WallJunction       *tmpJunctPtr;
    BNLine             *bnLinePtr;
    PinsAroundJunction pinLists;
    PinsAroundJunction *faceListsPtr;
    BottleNeck         *bottleNeckPtr;
    Tile               *tlTilePtr;
    Tile               *brTilePtr;
    Tile               *bnTilePtr;
    LocalData          localData;

    tmGlobalPlaneDataPtr = planeDataPtr;
    bnTileList = NIL(TileQueue);
    (void) Tp_HRectToVRect(&rect, &vertRect);
    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
                    &vertRect,
                    GetAndSortBN,
                    (ClientData) &bnTileList);
    FixTileList(&bnTileList);
/*
 * code change : by george : on 24 oct 89
 *
    MALLOC(PinsAroundJunction *, faceListsPtr, sizeof (PinsAroundJunction));
    faceListsPtr->horzPPinList = NIL(LinkedPPin);
    faceListsPtr->vertPPinList = NIL(LinkedPPin);
    faceListsPtr->pinInstList = NIL(LinkedPinInst);
    faceListsPtr->rPinList = NIL(LinkedRPin);
 */
    faceListsPtr = Tm_MakePinsAroundJunction();
    if (bnTileList == NIL(TileQueue))
    {
        localData.rect = rect;
/*
 * code change : by george : on 25 oct 89
 *
        Fpg_ItrWallJunct(Mac_GetFpgOfCell(parentCellPtr),
            FindJunctInRect, (ClientData)&localData);
 */
        (void) Fpg_ItrWallJunctSeq(Tp_GetFpg(planeDataPtr), FindJunctInRect,
            (ClientData) &localData);
        Tm_PinsAroundJunction(parentCellPtr, planeDataPtr,
            localData.junctPtr, faceListsPtr);
        return (faceListsPtr);
    }
    /*
     * Do the first junction region
     */
    if (Tp_IsSBNTile(bnTileList->tilePtr))
    {
    bnLinePtr = TmGetVBnLineOfSBnTiles(bnTileList->tilePtr,
    bnTileList->nextPtr->tilePtr);
    wallSegPtr = Tp_GetWallSegmentOfBnLine(bnLinePtr);
    }
    else
    {
    wallSegPtr = Tp_GetWallSegmentOfBnTile(bnTileList->tilePtr);
    }
    junctPtr = Fpg_GetNWJunctOfWallSeg(wallSegPtr);

    /* Circle Unify/Split calls this procedure when empty cells */
    /* were already deleted - therefore could get ambiguous junction */
    jPoint = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    while (GEO_ENCLOSE(jPoint, &rect) == FALSE)
    {
       tmpJunctPtr = Fpg_GetNorthJunctOfWallJunct(junctPtr);
       junctPtr = tmpJunctPtr;
       jPoint = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    }
    Tm_PinsAroundJunction(parentCellPtr, planeDataPtr, junctPtr, &pinLists);
    CopyPinListsToFaceLists(pinLists, faceListsPtr, includeInsidePPins);
    /*
     * Get the pPins on the right side of the left end vert bn tile
     */
    wallSegPtr = Fpg_GetWestSegOfWallJunct(junctPtr);
    if (wallSegPtr != NIL(WallSegment))
    {
    bottleNeckPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
    if (bottleNeckPtr->unType == TP_BN_LINE)
    {
        tlTilePtr = Tp_GetTLTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        brTilePtr = Tp_GetBRTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        (void) Tp_ItrPPinsOfBnLine(tlTilePtr, brTilePtr, GetPPins, 
                           (ClientData)faceListsPtr);
    }
    else
    {
        bnTilePtr = bottleNeckPtr->unValue.bnTilePtr;
        (void) Tp_ItrPPinsOfSide(bnTilePtr, TOP_RIGHT_SIDE, GetPPins,
                           (ClientData)faceListsPtr);
    }
    }
    /*
     * Get the pins, rpins, and ppins (except the ones on the vertical 
     * bottleneck tiles) by doing pinarndjunction and by getting the pins
     * at the ends of the vertical bottleneck tiles.
     */
    while (bnTileList != NIL(TileQueue))
    {
    if (Tp_IsSBNTile(bnTileList->tilePtr))
    {
        bnLinePtr = TmGetVBnLineOfSBnTiles(bnTileList->tilePtr,
        bnTileList->nextPtr->tilePtr);
        wallSegPtr = Tp_GetWallSegmentOfBnLine(bnLinePtr);
    }
    else
    {
        wallSegPtr = Tp_GetWallSegmentOfBnTile(bnTileList->tilePtr);
    }
    junctPtr = Fpg_GetSEJunctOfWallSeg(wallSegPtr);

        /* Could get ambiguous junction when empty cells are deleted */
    jPoint = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    while (GEO_ENCLOSE(jPoint, &rect) == FALSE)
    {
          tmpJunctPtr = Fpg_GetNorthJunctOfWallJunct(junctPtr);
      junctPtr = tmpJunctPtr;
      jPoint = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    }
    Tm_PinsAroundJunction(parentCellPtr, planeDataPtr, junctPtr, &pinLists);
    CopyPinListsToFaceLists(pinLists, faceListsPtr, includeInsidePPins);
    if (Tp_IsBottleneckTile(bnTileList->tilePtr))
    {
            /* If empty cells were deleted then both edges of the */
            /*   bottleneck might not belong to this HFace */
        (void) Tm_GetPinsInVBottleNeckRegion2(parentCellPtr, 
                         planeDataPtr,
                         bnTileList->tilePtr, 
                         junctPtr,
                         &pinLists);
        CopyPinListsToFaceLists(pinLists, faceListsPtr, includeInsidePPins);
    }
    else
    /* semi-bottleneck tile */
    {
        /* no pins on bnline, so go on to next region */
        if (bnTileList->nextPtr->paired == 0)
        {
        bnTileList = bnTileList->nextPtr;
        }
    }
    bnTileQue = bnTileList->nextPtr;
    FREE((char *)bnTileList);
    bnTileList = bnTileQue;
    }
    /*
     * Get the pPins on the left side of the right end vert bn tile
     */
    wallSegPtr = Fpg_GetEastSegOfWallJunct(junctPtr);
    if (wallSegPtr != NIL(WallSegment))
    {
    bottleNeckPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
    if (bottleNeckPtr->unType == TP_BN_LINE)
    {
        tlTilePtr = Tp_GetTLTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        brTilePtr = Tp_GetBRTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        (void) Tp_ItrPPinsOfBnLine(tlTilePtr, brTilePtr, GetPPins, 
                        (ClientData)faceListsPtr);
    }
    else
    {
        bnTilePtr = bottleNeckPtr->unValue.bnTilePtr;
        (void) Tp_ItrPPinsOfSide(bnTilePtr, BOTTOM_LEFT_SIDE, GetPPins,
                        (ClientData)faceListsPtr);
    }
    }
    return (faceListsPtr);
}



static void
CopyPinListsToFaceLists(pinLists, faceListsPtr, includeInsidePPins)
PinsAroundJunction pinLists;
PinsAroundJunction *faceListsPtr;
bool includeInsidePPins;
{
    LinkedPPin    *pPinPtr;
    LinkedRPin    *rPinPtr;
    LinkedPinInst *pinPtr;

    if (faceListsPtr->horzPPinList == NIL(LinkedPPin))
    {
       faceListsPtr->horzPPinList = pinLists.horzPPinList;
       pinLists.horzPPinList = NIL(LinkedPPin);
    }
    if (includeInsidePPins)
    {
       if (faceListsPtr->vertPPinList == NIL(LinkedPPin))
       {
          faceListsPtr->vertPPinList = pinLists.vertPPinList;
          pinLists.vertPPinList = NIL(LinkedPPin);
       }
    }
    if (faceListsPtr->pinInstList == NIL(LinkedPinInst))
    {
       faceListsPtr->pinInstList = pinLists.pinInstList;
       pinLists.pinInstList = NIL(LinkedPinInst);
    }
    if (faceListsPtr->rPinList == NIL(LinkedRPin))
    {
       faceListsPtr->rPinList = pinLists.rPinList;
       pinLists.rPinList = NIL(LinkedRPin);
    }

    if (pinLists.horzPPinList != NIL(LinkedPPin))
    {
       for (pPinPtr = pinLists.horzPPinList;
            pPinPtr->nextPtr != NIL(LinkedPPin); pPinPtr = pPinPtr->nextPtr);
            pPinPtr->nextPtr = faceListsPtr->horzPPinList;
          faceListsPtr->horzPPinList = pinLists.horzPPinList;
    }
    if (includeInsidePPins)
    {
       if (pinLists.vertPPinList != NIL(LinkedPPin))
       {
          for (pPinPtr = pinLists.vertPPinList;
               pPinPtr->nextPtr != NIL(LinkedPPin);pPinPtr = pPinPtr->nextPtr);
               pPinPtr->nextPtr = faceListsPtr->vertPPinList;
             faceListsPtr->vertPPinList = pinLists.vertPPinList;
       }
    }
    if (pinLists.pinInstList != NIL(LinkedPinInst))
    {
       for (pinPtr = pinLists.pinInstList;
            pinPtr->nextPtr != NIL(LinkedPinInst); pinPtr = pinPtr->nextPtr);
            pinPtr->nextPtr = faceListsPtr->pinInstList;
          faceListsPtr->pinInstList = pinLists.pinInstList;
    }
    if (pinLists.rPinList != NIL(LinkedRPin))
    {
       for (rPinPtr = pinLists.rPinList;
            rPinPtr->nextPtr != NIL(LinkedRPin); rPinPtr = rPinPtr->nextPtr);
            rPinPtr->nextPtr = faceListsPtr->rPinList;
          faceListsPtr->rPinList = pinLists.rPinList;
    }
}


void
Tm_GetPinsInVBottleNeckRegion2(parentCellPtr, 
                  planeDataPtr, 
                  bnTilePtr, 
                  junctPtr,
                  pinListsPtr)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
Tile       *bnTilePtr;
WallJunction *junctPtr;
PinsAroundJunction *pinListsPtr;
{
    /* Get pins on the cell instances that bound a vertical bottleneck tile */

    Rect      bnRect;
    Point     point;
    Tile      *tilePtr;
    DummyCell *dCellPtr;
    LocalInfo *localInfoPtr;
        int       bnTileStatus;

    MALLOC(LocalInfo *, localInfoPtr, sizeof (LocalInfo));
    localInfoPtr->pinListsPtr = pinListsPtr;
        pinListsPtr->horzPPinList = NIL(LinkedPPin);
        pinListsPtr->vertPPinList = NIL(LinkedPPin);
        pinListsPtr->pinInstList = NIL(LinkedPinInst);
        pinListsPtr->rPinList = NIL(LinkedRPin);

    (void) Tp_VertTileToRect(planeDataPtr, bnTilePtr, &bnRect);
    localInfoPtr->area = &bnRect;
    point.p_x = bnRect.r_ll.p_x;
    point.p_y = bnRect.r_ll.p_y - 1;
    tilePtr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr), &point);
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, 
                      TOP_SIDE,
                          GetPPinsOfDCell, 
                      (ClientData) pinListsPtr);
    }
    point.p_x = bnRect.r_ll.p_x;
    point.p_y = bnRect.r_ur.p_y;
    tilePtr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr), &point);
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, 
                      BOTTOM_SIDE,
                          GetPPinsOfDCell, 
                      (ClientData) pinListsPtr);
    }
        bnTileStatus = GetStatusOfVBnTile(bnTilePtr, junctPtr);
        /* Look at pins on top-edge of bnTile */
        if ((bnTileStatus != bnTileTopDead) && 
        (bnTileStatus != bnTileBothDead))
    {
       bnRect.r_ur.p_y += 1;
        }
        else
    {
       bnRect.r_ur.p_y -= 1;
        }
        /* Look at pins on bot-edge of bnTile */
        if ((bnTileStatus != bnTileBotDead) && 
        (bnTileStatus != bnTileBothDead))
    {
       bnRect.r_ll.p_y -= 1;
        }
        else
    {
       bnRect.r_ll.p_y += 1;
        }

    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
            &bnRect,
            GetPinsOfSolidTile,
            (ClientData) localInfoPtr);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(parentCellPtr),
                             GetPinsInRect, (ClientData) localInfoPtr);
}

void
Tm_GetPinsInVBottleNeckRegion(parentCellPtr, 
                  planeDataPtr, 
                  bnTilePtr, 
                  pinListsPtr)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
Tile       *bnTilePtr;
PinsAroundJunction *pinListsPtr;
{
    /* Get pins on the cell instances that bound a vertical bottleneck tile */




    Rect      bnRect;
    Point     point;
    Tile      *tilePtr;
    DummyCell *dCellPtr;
    LocalInfo *localInfoPtr;

    MALLOC(LocalInfo *, localInfoPtr, sizeof (LocalInfo));
    localInfoPtr->pinListsPtr = pinListsPtr;
        pinListsPtr->horzPPinList = NIL(LinkedPPin);
        pinListsPtr->vertPPinList = NIL(LinkedPPin);
        pinListsPtr->pinInstList = NIL(LinkedPinInst);
        pinListsPtr->rPinList = NIL(LinkedRPin);

    (void) Tp_VertTileToRect(planeDataPtr, bnTilePtr, &bnRect);
    localInfoPtr->area = &bnRect;
    point.p_x = bnRect.r_ll.p_x;
    point.p_y = bnRect.r_ll.p_y - 1;
    tilePtr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr), &point);
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, 
                      TOP_SIDE,
                          GetPPinsOfDCell, 
                      (ClientData) pinListsPtr);
    }
    point.p_x = bnRect.r_ll.p_x;
    point.p_y = bnRect.r_ur.p_y;
    tilePtr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr), &point);
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, 
                      BOTTOM_SIDE,
                          GetPPinsOfDCell, 
                      (ClientData) pinListsPtr);
    }
    bnRect.r_ur.p_y += 1;
    bnRect.r_ll.p_y -= 1;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
            &bnRect,
            GetPinsOfSolidTile,
            (ClientData) localInfoPtr);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(parentCellPtr),
                             GetPinsInRect, (ClientData) localInfoPtr);
}


static BearStatus
GetPPinsOfDCell(pPinPtr, faceListsPtr)
PseudoPin          *pPinPtr;
PinsAroundJunction *faceListsPtr;
{
    LinkedPPin *linkPPinPtr;

    MALLOC(LinkedPPin *, linkPPinPtr, sizeof (LinkedPPin));
    linkPPinPtr->ppinPtr = pPinPtr;
    linkPPinPtr->nextPtr = faceListsPtr->horzPPinList;
    faceListsPtr->horzPPinList = linkPPinPtr;
    return (BEAR_CONTINUE_SEARCH);
}


static int
GetPinsOfSolidTile(tilePtr, localInfoPtr)
Tile      *tilePtr;
LocalInfo *localInfoPtr;
{
    CellInst    *cellInstPtr;
    RouteCell   *routeCellPtr;

    if (Tp_IsCellInstTile(tilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        (void) Mac_ItrPinOfCellInst (cellInstPtr, GetPinsInRect,
          (ClientData) localInfoPtr);
    }
#ifdef USE_DR_CODE
    if (Tp_IsRCellTile(tilePtr))
    {
        routeCellPtr = Tp_GetRCellOfTile(tilePtr);
        (void) Mac_ItrSidePinOfRCell(routeCellPtr, GetRPinsInRect,
          (ClientData) localInfoPtr);
    }
#endif /* USE_DR_CODE */

    return (0);
}


static BearStatus
GetRPinsInRect(rPinPtr, localInfoPtr)
RoutePin  *rPinPtr;
LocalInfo *localInfoPtr;
{
    LinkedRPin *linkRPinPtr;
    Point      absPos;

    (void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if ((absPos.p_x >= localInfoPtr->area->r_xbot) &&
      (absPos.p_x <= localInfoPtr->area->r_xtop) &&
      (absPos.p_y >= localInfoPtr->area->r_ybot) &&
      (absPos.p_y <= localInfoPtr->area->r_ytop))
    {
        MALLOC(LinkedRPin *, linkRPinPtr, sizeof (LinkedRPin));
        linkRPinPtr->rPinPtr = rPinPtr;
        linkRPinPtr->nextPtr = localInfoPtr->pinListsPtr->rPinList;
        localInfoPtr->pinListsPtr->rPinList = linkRPinPtr;
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
GetPinsInRect(pinInstPtr, localInfoPtr)
PinInst   *pinInstPtr;
LocalInfo *localInfoPtr;
{
    LinkedPinInst *linkPinPtr;
    Point         absPos;

    (void) Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
    if ((absPos.p_x >= localInfoPtr->area->r_xbot) &&
      (absPos.p_x <= localInfoPtr->area->r_xtop) &&
      (absPos.p_y >= localInfoPtr->area->r_ybot) &&
      (absPos.p_y <= localInfoPtr->area->r_ytop))
    {
        MALLOC(LinkedPinInst *, linkPinPtr, sizeof (LinkedPinInst));
        linkPinPtr->pinInstPtr = pinInstPtr;
        linkPinPtr->nextPtr = localInfoPtr->pinListsPtr->pinInstList;
        localInfoPtr->pinListsPtr->pinInstList = linkPinPtr;
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
GetPPins(pPinPtr,pinListPtr)
PseudoPin          *pPinPtr;
PinsAroundJunction *pinListPtr;
{
    LinkedPPin *linkedPPinPtr;

    MALLOC(LinkedPPin *, linkedPPinPtr, sizeof (LinkedPPin));
    linkedPPinPtr->ppinPtr = pPinPtr;
    linkedPPinPtr->nextPtr = pinListPtr->vertPPinList;
    pinListPtr->vertPPinList = linkedPPinPtr;
    return (BEAR_CONTINUE_SEARCH);
}



static void
FixBadNetUnification(oldId, newId, curTilePtr)
BearId oldId;
BearId newId;
Tile   *curTilePtr;
{
    PinsAroundJunction *faceListsPtr;
    BearId             extId;
    BearId             intId;
    LinkedPPin         *pPinPtr;
    LinkedRPin         *rPinPtr;
    LinkedPinInst      *pinPtr;
    Rect               vertRect;
    TileQueue          *bnTileList;
    TileQueue          *bnTileQue;
    Tile               *pinTilePtr;

    faceListsPtr =
      Tm_PinsAroundHFace(tmGlobalParentCellPtr, tmGlobalPlaneDataPtr, 
        tmGlobalDmntRect, FALSE);

    for (pPinPtr = faceListsPtr->horzPPinList; pPinPtr != NIL(LinkedPPin);
      pPinPtr = faceListsPtr->horzPPinList)
    {
    extId = TmGetTempExtNetIdOfPPin(pPinPtr->ppinPtr);
    intId = TmGetTempIntNetIdOfPPin(pPinPtr->ppinPtr);
    if (extId == oldId)
    {
        (void) TmPutTempExtNetIdOfPPin(newId, pPinPtr->ppinPtr);
    }
    if ((intId == oldId) && 
      Tp_IsSBNTile(Tp_GetTileOfPPin(pPinPtr->ppinPtr)))
    {
        (void) TmPutTempIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
    }
    faceListsPtr->horzPPinList = pPinPtr->nextPtr;
    FREE((char *)pPinPtr);
    }
    for (pinPtr = faceListsPtr->pinInstList;
      pinPtr != NIL(LinkedPinInst); pinPtr = faceListsPtr->pinInstList)
    {
    extId = Mac_GetHExtNetIdOfPin(pinPtr->pinInstPtr);
    if (extId == oldId)
    {
        (void) Mac_PutHExtNetIdOfPin(newId, pinPtr->pinInstPtr);
            if (globalCopyFlag)
        {
        (void) Mac_PutExtNetIdOfPin(newId, pinPtr->pinInstPtr);
        }

    }
    faceListsPtr->pinInstList = pinPtr->nextPtr;
    FREE((char *)pinPtr);
    }
    for (rPinPtr = faceListsPtr->rPinList;
      rPinPtr != NIL(LinkedRPin); rPinPtr = faceListsPtr->rPinList)
    {
    extId = Mac_GetHExtNetIdOfRPin(rPinPtr->rPinPtr);
    if (extId == oldId)
    {
        (void) Mac_PutHExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
        if (globalCopyFlag)
        {
        (void) Mac_PutExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
        }

    }
    faceListsPtr->rPinList = rPinPtr->nextPtr;
    FREE((char *)rPinPtr);
    }

    bnTileList = NIL(TileQueue);
    (void) Tp_HRectToVRect(&tmGlobalDmntRect, &vertRect);
    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(tmGlobalPlaneDataPtr),
      &vertRect,
      GetAndSortBN,
      (ClientData) &bnTileList);
    FixTileList(&bnTileList);
    while (bnTileList != NIL(TileQueue))
    {
    if (Tp_IsSBNTile(bnTileList->tilePtr))
    {
        (void) Tp_ItrPPinsOfBnLine(bnTileList->tilePtr,
                bnTileList->nextPtr->tilePtr, GetPPins, 
                (ClientData)faceListsPtr);
        if (bnTileList->nextPtr->paired == 0)
        {
        bnTileList = bnTileList->nextPtr;
        }
    }
    else
    {
        (void) Tp_ItrPPinsOfBnTile(bnTileList->tilePtr, GetPPins,
                        (ClientData)faceListsPtr);
    }
    bnTileQue = bnTileList->nextPtr;
    FREE((char *)bnTileList);
    bnTileList = bnTileQue;
    }

    for (pPinPtr = faceListsPtr->vertPPinList; pPinPtr != NIL(LinkedPPin);
      pPinPtr = faceListsPtr->vertPPinList)
    {
    extId = TmGetTempExtNetIdOfPPin(pPinPtr->ppinPtr);
    intId = TmGetTempIntNetIdOfPPin(pPinPtr->ppinPtr);
    if (extId == oldId)
    {
        (void) TmPutTempExtNetIdOfPPin(newId, pPinPtr->ppinPtr);
        /* Sherry added these 4 lines - to make sure the intId is also changed*/
        if (intId == oldId)
        {
           (void) TmPutTempIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
        }
    }
    if (intId == oldId)
    {
       pinTilePtr = Tp_GetTileOfPPin(pPinPtr->ppinPtr);
       if ((pinTilePtr == curTilePtr) ||
           (Tp_IsSBNTile(pinTilePtr)))
       {
          (void) TmPutTempIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
       }
    }
    faceListsPtr->vertPPinList = pPinPtr->nextPtr;
    FREE((char *)pPinPtr);
    }
    FREE((char *)faceListsPtr);
}

 
void
TmHRectUnify(planeDataPtr, rect, copyFlag)
PlaneData *planeDataPtr;
Rect      rect;
int       copyFlag;
{
    /*
     * By search the dominant tile for vertical bottlenecks tiles and lines
     * from left to right, we propogate across the face the new horizontal 
     * values for the pseudo pin id's
     */

    Rect               vertRect;
    Rect               bnRect;
    TileQueue          *bnTileList;
    WallSegment        *wallSegPtr;
    WallJunction       *junctPtr;
    PinsAroundJunction pinLists;
    BNLine             *bnLinePtr;
    LocalInfo          localInfo;

    tmGlobalDmntRect.r_xbot = rect.r_xbot;
    tmGlobalDmntRect.r_ybot = rect.r_ybot;
    tmGlobalDmntRect.r_xtop = rect.r_xtop;
    tmGlobalDmntRect.r_ytop = rect.r_ytop;
    tmGlobalPlaneDataPtr = planeDataPtr;
    globalCopyFlag = copyFlag;
    bnTileList = NIL(TileQueue);
    (void) Tp_HRectToVRect(&tmGlobalDmntRect, &vertRect);
    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
      &vertRect,
      GetAndSortBN,
      (ClientData) &bnTileList);
    (void) FixTileList(&bnTileList);
    if (bnTileList != NIL(TileQueue))
    {
        if (Tp_IsSBNTile(bnTileList->tilePtr))
        {
        bnLinePtr = TmGetVBnLineOfSBnTiles(bnTileList->tilePtr,
                           bnTileList->nextPtr->tilePtr);
        wallSegPtr = Tp_GetWallSegmentOfBnLine(bnLinePtr);
        }
        else
        {
        wallSegPtr = Tp_GetWallSegmentOfBnTile(bnTileList->tilePtr);
        }
        junctPtr = Fpg_GetNWJunctOfWallSeg(wallSegPtr);
        (void)Tm_PinsAroundJunction(tmGlobalParentCellPtr, planeDataPtr, 
                    junctPtr, &pinLists);
        (void)PropagateExtToHExt(&pinLists);
    }
    while (bnTileList != NIL(TileQueue))
    {
        if (Tp_IsSBNTile(bnTileList->tilePtr))
        {
        bnLinePtr = TmGetVBnLineOfSBnTiles(bnTileList->tilePtr,
                           bnTileList->nextPtr->tilePtr);
        wallSegPtr = Tp_GetWallSegmentOfBnLine(bnLinePtr);
        }
        else
        {
        wallSegPtr = Tp_GetWallSegmentOfBnTile(bnTileList->tilePtr);
        }
        junctPtr = Fpg_GetSEJunctOfWallSeg(wallSegPtr);
        (void)Tm_PinsAroundJunction(tmGlobalParentCellPtr, planeDataPtr, 
                    junctPtr, &pinLists);
        (void)PropagateExtToHExt(&pinLists);
        if (Tp_IsBottleneckTile(bnTileList->tilePtr))
        {
        (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr,
                     bnTileList->tilePtr,
                     &bnRect);
        localInfo.tilePtr = bnTileList->tilePtr;
        localInfo.area = &bnRect;
        /* Sherry added 1/26/89 to propagate the vExtIds for I/O Pads */
        (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
                        tmGlobalParentCellPtr),
                        PropagateExtIdOfIOPads,
                        (ClientData) &localInfo);

        (void) Tp_ItrPPinsOfSide(bnTileList->tilePtr,
                     (SideDirection) BOTTOM_LEFT_SIDE,
                     CheckAgainstLeftPin, 
                     (ClientData) &localInfo);
        (void) Tp_ItrPPinsOfSide(bnTileList->tilePtr,
                     (SideDirection) TOP_RIGHT_SIDE,
                     CheckAgainstJunctPins, 
                     (ClientData) &pinLists);
        }
        else
        /* semi-bottleneck tile */
        {
            (void) Tp_ItrPPinsOfBnLine(bnTileList->tilePtr,
                           bnTileList->nextPtr->tilePtr, 
                           CheckAgainstJunctPins, 
                           (ClientData) &pinLists);
            /* the next element in the list is the other sbntile */
            if (bnTileList->nextPtr->paired == 0)
            {
            bnTileList = bnTileList->nextPtr;
            }
        }
        bnTileList = bnTileList->nextPtr;
    }
    }


static void
FixTileList(bnTileListPtr)
TileQueue **bnTileListPtr;
{
    TileQueue  *bnTileList;
    TileQueue  *placePtr;
    BNLineType bnLineType;
    int        removeElt;

    bnTileList = *bnTileListPtr;
    while (bnTileList != NIL(TileQueue))
    {
    if (Tp_IsSBNTile(bnTileList->tilePtr))
    {
        if (bnTileList->nextPtr == NIL(TileQueue))
        {
        removeElt = 1;
        }
        else
        {
        if (Tp_IsSBNTile(bnTileList->nextPtr->tilePtr))
        {
            if (Tp_IsPairOfSBNTiles(&(bnTileList->tilePtr), 
              &(bnTileList->nextPtr->tilePtr), &bnLineType))
            {
            removeElt = 0;
            bnTileList->paired = 1;
            /*
             * there may be a string of sbntiles that are
             * consecutively pairs.
             */
            if (bnTileList->nextPtr->nextPtr == NIL(TileQueue))
            {
                bnTileList = bnTileList->nextPtr;
            }
            else
            {
                if (!(Tp_IsPairOfSBNTiles(
                  &(bnTileList->nextPtr->tilePtr),
                  &(bnTileList->nextPtr->nextPtr->tilePtr),
                  &bnLineType)))
                {
                /*
                 * these two are not consecutive, so we are done
                 * with the first one.
                 */
                    bnTileList = bnTileList->nextPtr;
                }
            }
            }
            else
            {
            removeElt = 1;
            bnTileList->paired = 0;
            }
        }
        else
        {
            removeElt = 1;
        }
        }
    }
    else
    {
        removeElt = 0;
    }
    if (removeElt == 1)
    {
        if (*bnTileListPtr == bnTileList)
        {
        *bnTileListPtr = (*bnTileListPtr)->nextPtr;
        }
        else
        {
        for (placePtr = *bnTileListPtr;placePtr->nextPtr != bnTileList;
          placePtr = placePtr->nextPtr);
        placePtr->nextPtr = placePtr->nextPtr->nextPtr;
        }
    }
    bnTileList = bnTileList->nextPtr;
    }
}


typedef struct _BnLineInfo {
    Tile *lTilePtr;
    Tile *rTilePtr;
    BNLine *bnLinePtr;
} BnLineInfo;


static BNLine *
TmGetVBnLineOfSBnTiles(lTilePtr, rTilePtr)
Tile *lTilePtr;
Tile *rTilePtr;
{
    BnLineInfo bnLineInfo;

    bnLineInfo.lTilePtr = lTilePtr;
    bnLineInfo.rTilePtr = rTilePtr;
    (void) Tp_ItrVertBnLine(tmGlobalPlaneDataPtr, FindBnLine, 
        (ClientData) &bnLineInfo);
    return (bnLineInfo.bnLinePtr);
}


static BearStatus
FindBnLine(bnLinePtr, bnLineInfoPtr)
BNLine     *bnLinePtr;
BnLineInfo *bnLineInfoPtr;
{
    if ((Tp_GetTLTileOfBnLine(bnLinePtr) == bnLineInfoPtr->lTilePtr) &&
      (Tp_GetBRTileOfBnLine(bnLinePtr) == bnLineInfoPtr->rTilePtr))
    {
        bnLineInfoPtr->bnLinePtr = bnLinePtr;
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static void
PropagateExtToHExt(pinListsPtr)
PinsAroundJunction *pinListsPtr;
{
    LinkedPinInst *pinPosPtr;
    LinkedRPin    *rPinPtr;

    pinPosPtr = pinListsPtr->pinInstList;
    while (pinPosPtr != NIL(LinkedPinInst))
    {
        (void) Mac_PutHExtNetIdOfPin(
          Mac_GetExtNetIdOfPin(pinPosPtr->pinInstPtr), 
          pinPosPtr->pinInstPtr);
        pinPosPtr = pinPosPtr->nextPtr;
    }
    rPinPtr = pinListsPtr->rPinList;
    while (rPinPtr != NIL(LinkedRPin))
    {
        (void) Mac_PutHExtNetIdOfRPin(
          Mac_GetExtNetIdOfRPin(rPinPtr->rPinPtr), 
          rPinPtr->rPinPtr);
        rPinPtr = rPinPtr->nextPtr;
    }
}

static BearStatus 
PropagateExtIdOfIOPads(pinPtr, localInfoPtr)
PinInst *pinPtr;
LocalInfo *localInfoPtr;
{
   /*
    * Make sure that pin is on the boundary of the bottleneck tile
    * and then propagate the external id to the vExt id
    */

    Point absPos;

    (void) Mac_GetAbsPosOfPin(pinPtr, &absPos);
    
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
        (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
        (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
        (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        (void) Mac_PutHExtNetIdOfPin(Mac_GetExtNetIdOfPin(pinPtr), pinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}

static BearStatus
CheckAgainstJunctPins(pPinPtr, pinListsPtr)
PseudoPin          *pPinPtr;
PinsAroundJunction *pinListsPtr;
{
    /* Propogates id values across a junction region */

    LinkedPinInst *pinPosPtr;
    LinkedPPin    *pPinPosPtr;
    LinkedRPin    *rPinPosPtr;
    BearId        rIntId;
    BearId        rExtId;
    BearId        intId;
    BearId        extId;
    Tile          *curTilePtr;

    rIntId = TmGetTempIntNetIdOfPPin(pPinPtr);
    rExtId = TmGetTempExtNetIdOfPPin(pPinPtr);
    curTilePtr = Tp_GetTileOfPPin(pPinPtr);

    pinPosPtr = pinListsPtr->pinInstList;
    while (pinPosPtr != NIL(LinkedPinInst))
    {
    extId = Mac_GetExtNetIdOfPin(pinPosPtr->pinInstPtr);
    if (extId == rExtId)
    {
        (void) Mac_PutHExtNetIdOfPin(rIntId, pinPosPtr->pinInstPtr);
        if (globalCopyFlag)
        {
        (void) Mac_PutExtNetIdOfPin(rIntId, pinPosPtr->pinInstPtr);
        }
    }
    pinPosPtr = pinPosPtr->nextPtr;
    }
    rPinPosPtr = pinListsPtr->rPinList;
    while (rPinPosPtr != NIL(LinkedRPin))
    {
    extId = Mac_GetExtNetIdOfRPin(rPinPosPtr->rPinPtr);
    if (extId == rExtId)
    {
        (void) Mac_PutHExtNetIdOfRPin(rIntId, rPinPosPtr->rPinPtr);
        if (globalCopyFlag)
        {
        (void) Mac_PutExtNetIdOfRPin(rIntId, rPinPosPtr->rPinPtr);
        }
    }
    rPinPosPtr = rPinPosPtr->nextPtr;
    }
    pPinPosPtr = pinListsPtr->horzPPinList;
    while (pPinPosPtr != NIL(LinkedPPin))
    {
    intId = TmGetTempIntNetIdOfPPin(pPinPosPtr->ppinPtr);
    extId = TmGetTempExtNetIdOfPPin(pPinPosPtr->ppinPtr);
    if (extId == rExtId)
    {
        (void) TmPutTempExtNetIdOfPPin(rIntId, pPinPosPtr->ppinPtr);
    }
    /* This is needed for bottle neck lines */
    if ((intId == rExtId) && 
      Tp_IsSBNTile(Tp_GetTileOfPPin(pPinPosPtr->ppinPtr)))
    {
        (void) TmPutTempIntNetIdOfPPin(rIntId, pPinPosPtr->ppinPtr);
    }
    pPinPosPtr = pPinPosPtr->nextPtr;
    }
    pPinPosPtr = pinListsPtr->vertPPinList;
    while (pPinPosPtr != NIL(LinkedPPin))
    {
    if (pPinPosPtr->ppinPtr != pPinPtr)
    {
        intId = TmGetTempIntNetIdOfPPin(pPinPosPtr->ppinPtr);
        extId = TmGetTempExtNetIdOfPPin(pPinPosPtr->ppinPtr);
        if (extId == rExtId)
        {
        if ((Tp_GetTileOfPPin(pPinPosPtr->ppinPtr) == curTilePtr)
            && (intId != rIntId))
        {
            (void) FixBadNetUnification(intId, rIntId, curTilePtr);
        }
        else
        {
            (void) TmPutTempExtNetIdOfPPin(rIntId, pPinPosPtr->ppinPtr);
        }
        }
        /* This is needed for bottle neck lines */
        if ((intId == rExtId) && 
          Tp_IsSBNTile(Tp_GetTileOfPPin(pPinPosPtr->ppinPtr)))
        {
        (void) TmPutTempIntNetIdOfPPin(rIntId, pPinPosPtr->ppinPtr);
        }
    }
    pPinPosPtr = pPinPosPtr->nextPtr;
    }
    (void) TmPutTempExtNetIdOfPPin(rIntId, pPinPtr);
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckAgainstLeftPin(pPinPtr, localInfoPtr)
PseudoPin *pPinPtr;
LocalInfo *localInfoPtr;
{
    /* Propogates id values across a vertical bottleneck tile */
    
    (void) Tp_ItrPPinsOfSide(localInfoPtr->tilePtr,
                 (SideDirection) BOTTOM_LEFT_SIDE,
                 CheckLeftRightPin, 
                 (ClientData) pPinPtr);

    (void) Tp_ItrPPinsOfSide(localInfoPtr->tilePtr,
                 (SideDirection) TOP_RIGHT_SIDE,
                 CheckLeftRightPin, 
                 (ClientData) pPinPtr);

    (void) ItrPinsOnTopAndBot(localInfoPtr, pPinPtr);
    (void) TmPutTempIntNetIdOfPPin(TmGetTempExtNetIdOfPPin(pPinPtr), pPinPtr);
    return (BEAR_CONTINUE_SEARCH);
}


static void
ItrPinsOnTopAndBot(localInfoPtr, pPinPtr)
LocalInfo *localInfoPtr;
PseudoPin *pPinPtr;
{
    /* Get pins on the cell instances that bound a vertical bottleneck tile */

    Rect      bnRect;
    Point     point;
    Tile      *tilePtr;
    DummyCell *dCellPtr;

    (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, localInfoPtr->tilePtr,
      &bnRect);
    localInfoPtr->area = &bnRect;
    localInfoPtr->pseudoPin = pPinPtr;
    point.p_x = bnRect.r_ll.p_x;
    point.p_y = bnRect.r_ll.p_y - 1;
    tilePtr = TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &point);
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, TOP_SIDE,
                      CheckDCellPPins, 
                      (ClientData) localInfoPtr);
    }
    point.p_x = bnRect.r_ll.p_x;
    point.p_y = bnRect.r_ur.p_y;
    tilePtr = TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &point);
    if (Tp_IsDCellTile(tilePtr))
    {
        dCellPtr = Tp_GetDCellOfTile(tilePtr);
        (void) Mac_ItrPPinOfDCell(dCellPtr, BOTTOM_SIDE,
                      CheckDCellPPins,
                      (ClientData) localInfoPtr);
    }
    bnRect.r_ur.p_y += 1;
    bnRect.r_ll.p_y -= 1;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &bnRect,
      ItrSolidTile,
      (ClientData) localInfoPtr);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
      tmGlobalParentCellPtr), CheckTopBottomPin, (ClientData) localInfoPtr);
}


static int
ItrSolidTile(tilePtr, localInfoPtr)
Tile      *tilePtr;
LocalInfo *localInfoPtr;
{
    CellInst    *cellInstPtr;
    RouteCell   *routeCellPtr;

    if (Tp_IsCellInstTile(tilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        (void) Mac_ItrPinOfCellInst (cellInstPtr, CheckTopBottomPin,
          (ClientData) localInfoPtr);
    }
#ifdef USE_DR_CODE
    if (Tp_IsRCellTile(tilePtr))
    {
        routeCellPtr = Tp_GetRCellOfTile(tilePtr);
        (void) Mac_ItrSidePinOfRCell(routeCellPtr, CheckTopBottomRPin,
          (ClientData) localInfoPtr);
    }
#endif /* USE_DR_CODE */

    return (0);
}


static BearStatus
CheckTopBottomRPin(rPinPtr,localInfoPtr)
RoutePin  *rPinPtr;
LocalInfo *localInfoPtr;
{
    /*
     * Make sure that pin is on the boundary area, and then change the
     * external id value so that it is consistent
     */

    Point  absPos;
    BearId lIntId;
    BearId lExtId;
    BearId extId;

    (void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if (Mac_GetHExtNetIdOfRPin(rPinPtr) == 0)
    {
        (void) Mac_PutHExtNetIdOfRPin(Mac_GetExtNetIdOfRPin(rPinPtr), rPinPtr);
    }
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
      (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
      (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        lIntId = TmGetTempIntNetIdOfPPin(localInfoPtr->pseudoPin);
        extId = Mac_GetExtNetIdOfRPin(rPinPtr);
        lExtId = TmGetTempExtNetIdOfPPin(localInfoPtr->pseudoPin);
        if (extId == lIntId)
        {
            (void) Mac_PutHExtNetIdOfRPin(lExtId, rPinPtr);
            if (globalCopyFlag)
            {
                (void) Mac_PutExtNetIdOfRPin(lExtId, rPinPtr);
            }
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckTopBottomPin(pinPtr,localInfoPtr)
PinInst   *pinPtr;
LocalInfo *localInfoPtr;
{
    /*
     * Make sure that pin is on the boundary area, and then change the
     * external id value so that it is consistent
     */

    Point  absPos;
    BearId lIntId;
    BearId lExtId;
    BearId extId;

    (void) Mac_GetAbsPosOfPin(pinPtr, &absPos);
    if (Mac_GetHExtNetIdOfPin(pinPtr) == 0)
    {
        (void) Mac_PutHExtNetIdOfPin(Mac_GetExtNetIdOfPin(pinPtr), pinPtr);
    }
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
      (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
      (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        lIntId = TmGetTempIntNetIdOfPPin(localInfoPtr->pseudoPin);
        extId = Mac_GetExtNetIdOfPin(pinPtr);
        lExtId = TmGetTempExtNetIdOfPPin(localInfoPtr->pseudoPin);
        if (extId == lIntId)
        {
            (void) Mac_PutHExtNetIdOfPin(lExtId, pinPtr);
            if (globalCopyFlag)
            {
                (void) Mac_PutExtNetIdOfPin(lExtId, pinPtr);
            }
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckDCellPPins(pPinPtr,localInfoPtr)
PseudoPin *pPinPtr;
LocalInfo *localInfoPtr;
{
    BearId lIntId;
    BearId extId;

    extId = TmGetTempExtNetIdOfPPin(pPinPtr);
    lIntId = TmGetTempIntNetIdOfPPin(localInfoPtr->pseudoPin);
    if (extId == lIntId)
    {
        (void) TmPutTempExtNetIdOfPPin(
          TmGetTempExtNetIdOfPPin(localInfoPtr->pseudoPin), pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
CheckLeftRightPin(pPinPtr,leftPPinPtr)
PseudoPin *pPinPtr;
PseudoPin *leftPPinPtr;
{
    BearId lIntId;
    BearId intId;
    BearId extId;
    BearId lExtId;

    lIntId = TmGetTempIntNetIdOfPPin(leftPPinPtr);
    lExtId = TmGetTempExtNetIdOfPPin(leftPPinPtr);
    intId = TmGetTempIntNetIdOfPPin(pPinPtr);
    extId = TmGetTempExtNetIdOfPPin(pPinPtr);
    if (pPinPtr != leftPPinPtr)
    {
    if (extId != lExtId)
    {
        if (intId == lIntId)
        {
        if (Tp_GetSideOfPPin(pPinPtr) == TOP_RIGHT_SIDE)
        {
            (void) TmPutTempIntNetIdOfPPin(lExtId, pPinPtr);
        }
        /* pPinPtr on the left side */
        else
        {
            (void) TmPutTempExtNetIdOfPPin(lExtId, pPinPtr);
            (void) TmPutTempIntNetIdOfPPin(lExtId, pPinPtr);
            /* Backward unification */
            (void) FixBadNetUnification(extId, lExtId, 
                                        Tp_GetTileOfPPin(leftPPinPtr));
/*
printf("Backward unification for (%d, %d) \n", intId, extId);
fflush(stdout);
*/
        }
        }
    }
    }
    return (BEAR_CONTINUE_SEARCH);
}

static BearStatus
FindJunctInRect(junctPtr, localDataPtr)
WallJunction *junctPtr;
LocalData *localDataPtr;
{
    Point *pointPtr;

    pointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
    if ((pointPtr->p_x < localDataPtr->rect.r_xtop) &&
    (pointPtr->p_y < localDataPtr->rect.r_ytop) &&
    (pointPtr->p_x > localDataPtr->rect.r_xbot) &&
    (pointPtr->p_y > localDataPtr->rect.r_ybot))
    {
    localDataPtr->junctPtr = junctPtr;
    return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static int
GetAndSortBN(tilePtr, bnTileList)
Tile       *tilePtr;
TileQueue  **bnTileList;
{
    if (Tp_IsBottleneckTile(tilePtr))
    {
        (void) InsertBnInList(tilePtr, bnTileList);
    }
    if (Tp_IsSBNTile(tilePtr))
    {
        (void) InsertBnInList(tilePtr, bnTileList);
    }
    return (MAGIC_CONTINUE_SEARCH);
}


static void
InsertBnInList(bnTilePtr, bnTileList)
Tile       *bnTilePtr;
TileQueue  **bnTileList;
{
    TileQueue *tilePtr;
    TileQueue *currentTilePtr;
    Rect      bnRect;

        MALLOC(TileQueue *, tilePtr, sizeof (TileQueue));
    (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, bnTilePtr, &bnRect);
    tilePtr->tilePtr = bnTilePtr;
    tilePtr->paired = 0;
    tilePtr->left = bnRect.r_xbot;
    if ((*bnTileList == NIL(TileQueue)) ||
        (bnRect.r_xbot < (*bnTileList)->left))
    {
        tilePtr->nextPtr = *bnTileList;
        *bnTileList = tilePtr;
    }
    else
    {
        currentTilePtr = *bnTileList;
        while ((currentTilePtr->nextPtr != NIL(TileQueue)) && 
               (bnRect.r_xbot > currentTilePtr->nextPtr->left))
        {
            currentTilePtr = currentTilePtr->nextPtr;
        }
    tilePtr->nextPtr = currentTilePtr->nextPtr;
        currentTilePtr->nextPtr = tilePtr;
    }
}

int
GetStatusOfVBnTile(bnTilePtr, junctPtr)
Tile *bnTilePtr;
WallJunction *junctPtr;
{
    WallSegment  *wallSegPtr2;
    WallJunction *junctPtr2;
    BottleNeck   *bnPtr;
    int          bnTileStatus;

    bnTileStatus = bnTileAlive;

    /* Check the top of the bnTile */
    junctPtr2 = Fpg_GetNorthJunctOfWallJunct(junctPtr);
    if (junctPtr2)
    {
       wallSegPtr2 = Fpg_GetWestSegOfWallJunct(junctPtr2);
       if (wallSegPtr2)
       {
           bnPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr2);
       if (bnPtr->unType == TP_BN_TILE)
           {
              if (bnTilePtr == bnPtr->unValue.bnTilePtr)
              {
                bnTileStatus = bnTileTopDead;
              }
           }
       }
    }

    /* Check the bot of the bnTile */
    junctPtr2 = Fpg_GetSouthJunctOfWallJunct(junctPtr);
    if (junctPtr2)
    {
        wallSegPtr2 = Fpg_GetWestSegOfWallJunct(junctPtr2);
        if (wallSegPtr2)
        {
           bnPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr2);
       if (bnPtr->unType == TP_BN_TILE)
           {
          if (bnTilePtr == bnPtr->unValue.bnTilePtr)
              {
         if (bnTileStatus == bnTileTopDead)
                    bnTileStatus = bnTileBothDead;
                 else bnTileStatus = bnTileBotDead;
              }
           }
        }
    }
   return (bnTileStatus);

}
