
/* tmPinsArndJunction.c --
 *
 *          Tim Collins
 *     (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmPinsArndJunction.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"
#include "list.h"


typedef struct _LocalData
{
    Rect          *bound;
    Rect          *area;
    LinkedPinInst *pinList;
    LinkedRPin    *rPinList;
    PlaneData     *planeDataPtr;
} LocalData;


/* ---------- Local Procedures ---------- */
static int SearchSpaceTile();
static int SearchSolidTile();
static BearStatus GetPinsInRect();
static BearStatus AddPins();
static void DoNorthWallSeg();
static void DoSouthWallSeg();
static void DoEastWallSeg();
static void DoWestWallSeg();
static void SearchRectArea();
static void AddPPinsToList();
static void RemoveDuplicates();
static int ComparePin();
static int ComparePPin();
static int CompareRPin();
static void RemoveDupPin();
static void RemoveDupPPin();
static void RemoveDupRPin();

#ifdef USE_DR_CODE
static BearStatus GetRPinsInRect();
#endif /* USE_DR_CODE */


bool tm_DupFlag = FALSE;


static int
ComparePin(item1, item2)
lsGeneric item1, item2;
{
    register LinkedPinInst *lnPin1Ptr = (LinkedPinInst *) item1;
    register LinkedPinInst *lnPin2Ptr = (LinkedPinInst *) item2;

    return ((lnPin1Ptr->pinInstPtr == lnPin2Ptr->pinInstPtr) ? 0 :
        ((lnPin1Ptr->pinInstPtr < lnPin2Ptr->pinInstPtr) ? -1 : 1));
}


static int
ComparePPin(item1, item2)
lsGeneric item1, item2;
{
    register LinkedPPin *lnPPin1Ptr = (LinkedPPin *) item1;
    register LinkedPPin *lnPPin2Ptr = (LinkedPPin *) item2;

    return ((lnPPin1Ptr->ppinPtr == lnPPin2Ptr->ppinPtr) ? 0 :
        ((lnPPin1Ptr->ppinPtr < lnPPin2Ptr->ppinPtr) ? -1 : 1));
}


static int
CompareRPin(item1, item2)
lsGeneric item1, item2;
{
    register LinkedRPin *lnRPin1Ptr = (LinkedRPin *) item1;
    register LinkedRPin *lnRPin2Ptr = (LinkedRPin *) item2;

    return ((lnRPin1Ptr->rPinPtr == lnRPin2Ptr->rPinPtr) ? 0 :
        ((lnRPin1Ptr->rPinPtr < lnRPin2Ptr->rPinPtr) ? -1 : 1));
}


static void
RemoveDupPin(item)
lsGeneric item;
{
    if (tm_DupFlag == TRUE) {
        printf ("duplicate :\n");
        Mac_PrintPinInst(((LinkedPinInst *) item)->pinInstPtr,
            (ClientData) NULL);
    }
    FREE((char *) item);
}


static void
RemoveDupPPin(item)
lsGeneric item;
{
    if (tm_DupFlag == TRUE) {
        printf ("duplicate :\n");
        Tp_PrintPPin(((LinkedPPin *) item)->ppinPtr, (ClientData) NULL);
    }
    FREE((char *) item);
}


static void
RemoveDupRPin(item)
lsGeneric item;
{
    if (tm_DupFlag == TRUE) {
        printf ("duplicate :\n");
        Mac_PrintRPin(((LinkedRPin *) item)->rPinPtr, (ClientData) NULL);
    }
    FREE((char *) item);
}


static void
RemoveDuplicates(pinListsPtr)
register PinsAroundJunction *pinListsPtr;
{
    LinkedPinInst   *lnPinInstPtr;  
    LinkedPPin      *lnPPinPtr;  
    LinkedRPin      *lnRPinPtr;  
    lsList          list;

    if (pinListsPtr->pinInstList != NIL(LinkedPinInst)) {
        list = lsCreate();
        for (lnPinInstPtr = pinListsPtr->pinInstList;
            lnPinInstPtr != NIL(LinkedPinInst);
            lnPinInstPtr = lnPinInstPtr->nextPtr) {

            lsNewEnd(list, (lsGeneric) lnPinInstPtr, LS_NH);
        }
        lsSort(list, ComparePin);
        lsUniq(list, ComparePin, RemoveDupPin);

        lsDelBegin(list, (lsGeneric *) &lnPinInstPtr);
        lnPinInstPtr->nextPtr = NIL(LinkedPinInst);
        pinListsPtr->pinInstList = lnPinInstPtr;
        while (lsDelBegin(list, (lsGeneric *) &lnPinInstPtr) == LS_OK) {
            lnPinInstPtr->nextPtr = pinListsPtr->pinInstList;
            pinListsPtr->pinInstList = lnPinInstPtr;
        }

        lsDestroy(list, (void (*)()) NULL);
    }

    if (pinListsPtr->horzPPinList != NIL(LinkedPPin)) {
        list = lsCreate();
        for (lnPPinPtr = pinListsPtr->horzPPinList;
            lnPPinPtr != NIL(LinkedPPin);
            lnPPinPtr = lnPPinPtr->nextPtr) {

            lsNewEnd(list, (lsGeneric) lnPPinPtr, LS_NH);
        }
        lsSort(list, ComparePPin);
        lsUniq(list, ComparePPin, RemoveDupPPin);

        lsDelBegin(list, (lsGeneric *) &lnPPinPtr);
        lnPPinPtr->nextPtr = NIL(LinkedPPin);
        pinListsPtr->horzPPinList = lnPPinPtr;
        while (lsDelBegin(list, (lsGeneric *) &lnPPinPtr) == LS_OK) {
            lnPPinPtr->nextPtr = pinListsPtr->horzPPinList;
            pinListsPtr->horzPPinList = lnPPinPtr;
        }

        lsDestroy(list, (void (*)()) NULL);
    }

    if (pinListsPtr->vertPPinList != NIL(LinkedPPin)) {
        list = lsCreate();
        for (lnPPinPtr = pinListsPtr->vertPPinList;
            lnPPinPtr != NIL(LinkedPPin);
            lnPPinPtr = lnPPinPtr->nextPtr) {

            lsNewEnd(list, (lsGeneric) lnPPinPtr, LS_NH);
        }
        lsSort(list, ComparePPin);
        lsUniq(list, ComparePPin, RemoveDupPPin);

        lsDelBegin(list, (lsGeneric *) &lnPPinPtr);
        lnPPinPtr->nextPtr = NIL(LinkedPPin);
        pinListsPtr->vertPPinList = lnPPinPtr;
        while (lsDelBegin(list, (lsGeneric *) &lnPPinPtr) == LS_OK) {
            lnPPinPtr->nextPtr = pinListsPtr->vertPPinList;
            pinListsPtr->vertPPinList = lnPPinPtr;
        }

        lsDestroy(list, (void (*)()) NULL);
    }

    if (pinListsPtr->rPinList != NIL(LinkedRPin)) {
        list = lsCreate();
        for (lnRPinPtr = pinListsPtr->rPinList;
            lnRPinPtr != NIL(LinkedRPin);
            lnRPinPtr = lnRPinPtr->nextPtr) {

            lsNewEnd(list, (lsGeneric) lnRPinPtr, LS_NH);
        }
        lsSort(list, CompareRPin);
        lsUniq(list, CompareRPin, RemoveDupRPin);

        lsDelBegin(list, (lsGeneric *) &lnRPinPtr);
        lnRPinPtr->nextPtr = NIL(LinkedRPin);
        pinListsPtr->rPinList = lnRPinPtr;
        while (lsDelBegin(list, (lsGeneric *) &lnRPinPtr) == LS_OK) {
            lnRPinPtr->nextPtr = pinListsPtr->rPinList;
            pinListsPtr->rPinList = lnRPinPtr;
        }

        lsDestroy(list, (void (*)()) NULL);
    }
}


void
Tm_PinsAroundJunction(parentCellPtr, planeDataPtr, junctPtr, pinListsPtr)
CellMaster         *parentCellPtr;
PlaneData          *planeDataPtr;
WallJunction       *junctPtr;
PinsAroundJunction *pinListsPtr;
{
    Rect         searchRect;
    LocalData    localData;

    tmGlobalParentCellPtr = parentCellPtr;
    tmGlobalPlaneDataPtr = planeDataPtr;
    MALLOC(Rect *, localData.bound, sizeof (Rect));
    pinListsPtr->horzPPinList = NIL(LinkedPPin);
    pinListsPtr->vertPPinList = NIL(LinkedPPin);
    searchRect.r_ur.p_x = junctPtr->centerPoint.p_x + 1;
    searchRect.r_ll.p_x = junctPtr->centerPoint.p_x - 1;
    localData.pinList = NIL(LinkedPinInst);
    localData.rPinList = NIL(LinkedRPin);
    localData.planeDataPtr = planeDataPtr;

    (void) DoNorthWallSeg(planeDataPtr, junctPtr, pinListsPtr, &searchRect);
    (void) DoSouthWallSeg(planeDataPtr, junctPtr, pinListsPtr, &searchRect);
    (void) DoEastWallSeg(planeDataPtr, junctPtr, pinListsPtr, &localData);
    (void) DoWestWallSeg(planeDataPtr, junctPtr, pinListsPtr, &localData);
        localData.bound->r_ll.p_y = searchRect.r_ll.p_y;
        localData.bound->r_ur.p_y = searchRect.r_ur.p_y;
    (void) SearchRectArea(&searchRect, &localData);
    pinListsPtr->pinInstList = localData.pinList;
    pinListsPtr->rPinList = localData.rPinList;

    /*
     * george ( 8 feb 90 )
     *
     * Every once in a while the the lists of pins returned by this
     * procedure via the structure PinsAroundJunction contain duplicates,
     * so the procedure RemoveDuplicates() has been placed here to filter
     * those duplicates out.
     */
    RemoveDuplicates(pinListsPtr);
}



static void
DoNorthWallSeg(planeDataPtr, junctPtr, pinListsPtr, searchRectPtr)
PlaneData          *planeDataPtr;
WallJunction       *junctPtr;
PinsAroundJunction *pinListsPtr;
Rect               *searchRectPtr;
{
    WallSegment  *wallSegPtr;
    Tile         *bnTilePtr;
    Tile         *dmntTilePtr;
    Rect         bnRect;
    Rect         dmntRect;
    Point        searchPoint;
    Point        *jPointPtr;
    BottleNeck   *bottleNeckPtr;

    if ((wallSegPtr = Fpg_GetNorthSegOfWallJunct(junctPtr)) != NIL(WallSegment))
    {
        bottleNeckPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
        if (bottleNeckPtr->unType == TP_BN_LINE)
        {
            bnTilePtr = Tp_GetTLTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        }
        else
        {
            bnTilePtr = bottleNeckPtr->unValue.bnTilePtr;
        }
        (void) Tp_HorzTileToRect(planeDataPtr, bnTilePtr, &bnRect);
        if (Tp_IsSBNTile(bnTilePtr))
        {
            (void) Tp_ItrPPinsOfBnLine(bnTilePtr,
              Tp_GetBRTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr), 
              AddPins, (ClientData) &pinListsPtr->horzPPinList);
            searchRectPtr->r_ur.p_y = bnRect.r_ll.p_y;
        }
        else
        {
            AddPPinsToList(bnTilePtr, BOTTOM_LEFT_SIDE, 
              &pinListsPtr->horzPPinList);
            searchRectPtr->r_ur.p_y = bnRect.r_ll.p_y;
        }
    }
    else
    {
        jPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
        (void) Tp_HPointToVPoint(jPointPtr, &searchPoint);
        dmntTilePtr = TiSrPointNoHint(Tp_GetVertPlane(planeDataPtr),
          &searchPoint);
        (void) Tp_VertTileToRect(planeDataPtr, dmntTilePtr, &dmntRect);
        searchRectPtr->r_ur.p_y = dmntRect.r_ur.p_y;
    }
}


static void
DoSouthWallSeg(planeDataPtr, junctPtr, pinListsPtr, searchRectPtr)
PlaneData          *planeDataPtr;
WallJunction       *junctPtr;
PinsAroundJunction *pinListsPtr;
Rect               *searchRectPtr;
{
    WallSegment  *wallSegPtr;
    Tile         *bnTilePtr;
    Tile         *dmntTilePtr;
    Rect         bnRect;
    Rect         dmntRect;
    Point        searchPoint;
    Point        *jPointPtr;
    BottleNeck   *bottleNeckPtr;

    if ((wallSegPtr = Fpg_GetSouthSegOfWallJunct(junctPtr)) != NIL(WallSegment))
    {
        bottleNeckPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
        if (bottleNeckPtr->unType == TP_BN_LINE)
        {
            bnTilePtr = Tp_GetTLTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        }
        else
        {
            bnTilePtr = bottleNeckPtr->unValue.bnTilePtr;
        }
        (void) Tp_HorzTileToRect(planeDataPtr, bnTilePtr, &bnRect);
        if (Tp_IsSBNTile(bnTilePtr))
        {
            searchRectPtr->r_ll.p_y = bnRect.r_ll.p_y;
            (void) Tp_ItrPPinsOfBnLine(bnTilePtr,
              Tp_GetBRTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr), 
              AddPins, (ClientData) &pinListsPtr->horzPPinList);
        }
        else
        {
            searchRectPtr->r_ll.p_y = bnRect.r_ur.p_y;
            AddPPinsToList(bnTilePtr, TOP_RIGHT_SIDE, 
              &pinListsPtr->horzPPinList);
        }
    }
    else
    {
        jPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
        (void) Tp_HPointToVPoint(jPointPtr, &searchPoint);
        dmntTilePtr = TiSrPointNoHint(Tp_GetVertPlane(planeDataPtr),
          &searchPoint);
        (void) Tp_VertTileToRect(planeDataPtr, dmntTilePtr, &dmntRect);
        searchRectPtr->r_ll.p_y = dmntRect.r_ll.p_y;
    }
}


static void
DoEastWallSeg(planeDataPtr, junctPtr, pinListsPtr, localDataPtr)
PlaneData          *planeDataPtr;
WallJunction       *junctPtr;
PinsAroundJunction *pinListsPtr;
LocalData          *localDataPtr;
{
    WallSegment  *wallSegPtr;
    Tile         *bnTilePtr;
    Tile         *dmntTilePtr;
    Rect         bnRect;
    Rect         dmntRect;
    Point        *jPointPtr;
    BottleNeck   *bottleNeckPtr;

    if ((wallSegPtr = Fpg_GetEastSegOfWallJunct(junctPtr)) != NIL(WallSegment))
    {
        bottleNeckPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
        if (bottleNeckPtr->unType == TP_BN_LINE)
        {
            bnTilePtr = Tp_GetTLTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        }
        else
        {
            bnTilePtr = bottleNeckPtr->unValue.bnTilePtr;
        }
        (void) Tp_VertTileToRect(planeDataPtr, bnTilePtr, &bnRect);
        if (Tp_IsSBNTile(bnTilePtr))
        {
            (void) Tp_ItrPPinsOfBnLine(bnTilePtr,
              Tp_GetBRTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr), 
              AddPins, (ClientData) &pinListsPtr->vertPPinList);
            localDataPtr->bound->r_ur.p_x = bnRect.r_ur.p_x;
        }
        else
        {
            localDataPtr->bound->r_ur.p_x = bnRect.r_ll.p_x;
            AddPPinsToList(bnTilePtr, BOTTOM_LEFT_SIDE, 
              &pinListsPtr->vertPPinList);
        }
    }
    else
    {
        jPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
        dmntTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr),
          jPointPtr);
        (void) Tp_HorzTileToRect(planeDataPtr, dmntTilePtr, &dmntRect);
        localDataPtr->bound->r_ur.p_x = dmntRect.r_ur.p_x;
    }
}


static void
DoWestWallSeg(planeDataPtr, junctPtr, pinListsPtr, localDataPtr)
PlaneData          *planeDataPtr;
WallJunction       *junctPtr;
PinsAroundJunction *pinListsPtr;
LocalData          *localDataPtr;
{
    WallSegment  *wallSegPtr;
    Tile         *bnTilePtr;
    Tile         *dmntTilePtr;
    Rect         bnRect;
    Rect         dmntRect;
    Point        *jPointPtr;
    BottleNeck   *bottleNeckPtr;

    if ((wallSegPtr = Fpg_GetWestSegOfWallJunct(junctPtr)) != NIL(WallSegment))
    {
        bottleNeckPtr = Fpg_GetBottleNeckOfWallSeg(wallSegPtr);
        if (bottleNeckPtr->unType == TP_BN_LINE)
        {
            bnTilePtr = Tp_GetTLTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr);
        }
        else
        {
            bnTilePtr = bottleNeckPtr->unValue.bnTilePtr;
        }
        (void) Tp_VertTileToRect(planeDataPtr, bnTilePtr, &bnRect);
        if (Tp_IsSBNTile(bnTilePtr))
        {
            (void) Tp_ItrPPinsOfBnLine(bnTilePtr,
              Tp_GetBRTileOfBnLine(bottleNeckPtr->unValue.bnLinePtr), 
              AddPins, (ClientData) &pinListsPtr->vertPPinList);
            localDataPtr->bound->r_ll.p_x = bnRect.r_ur.p_x;
        }
        else
        {
            AddPPinsToList(bnTilePtr, TOP_RIGHT_SIDE, 
              &pinListsPtr->vertPPinList);
            localDataPtr->bound->r_ll.p_x = bnRect.r_ur.p_x;
        }
    }
    else
    {
        jPointPtr = Fpg_GetPointOfWallJunct(FPG_CENTER, junctPtr);
        dmntTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(planeDataPtr),
          jPointPtr);
        (void) Tp_HorzTileToRect(planeDataPtr, dmntTilePtr, &dmntRect);
        localDataPtr->bound->r_ll.p_x = dmntRect.r_ll.p_x;
    }
}


static void
SearchRectArea(searchRectPtr, localDataPtr)
Rect      *searchRectPtr;
LocalData *localDataPtr;
{
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(localDataPtr->planeDataPtr),
      searchRectPtr,
      SearchSpaceTile,
      (ClientData) localDataPtr);
}


static int
SearchSpaceTile(tilePtr, localDataPtr)
Tile      *tilePtr;
LocalData *localDataPtr;
{
    Rect  spaceRect;

    if (Tp_IsSpaceTile(tilePtr))
    {
        (void) Tp_HorzTileToRect(localDataPtr->planeDataPtr, tilePtr,
          &spaceRect);
        if (spaceRect.r_ur.p_x > localDataPtr->bound->r_ur.p_x)
        {
            spaceRect.r_ur.p_x = localDataPtr->bound->r_ur.p_x;
        }
        if (spaceRect.r_ll.p_x < localDataPtr->bound->r_ll.p_x)
        {
            spaceRect.r_ll.p_x = localDataPtr->bound->r_ll.p_x;
        }
        spaceRect.r_ur.p_x += 1;
        spaceRect.r_ll.p_x -= 1;

                MALLOC(Rect *, localDataPtr->area, sizeof (Rect));
        Mac_CopyRect(localDataPtr->area, &spaceRect);
                /* Exclude Horz PinInst on boundary of the junction region */
                if (localDataPtr->area->r_ll.p_y == 
                                         localDataPtr->bound->r_ll.p_y)
            {
            localDataPtr->area->r_ll.p_y += 1;
        }
        if (localDataPtr->area->r_ur.p_y == 
                                         localDataPtr->bound->r_ur.p_y)
        {
            localDataPtr->area->r_ll.p_y -= 1;
        }

        (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(localDataPtr->planeDataPtr),
          &spaceRect,
          SearchSolidTile,
          (ClientData) localDataPtr);
        /* for i/o pins */
        (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
          tmGlobalParentCellPtr), GetPinsInRect, (ClientData) localDataPtr);

        spaceRect.r_ur.p_x -= 1;
        spaceRect.r_ll.p_x += 1;

        spaceRect.r_ur.p_y += 1;
        spaceRect.r_ll.p_y -= 1;
        Mac_CopyRect(localDataPtr->area, &spaceRect);

        (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(localDataPtr->planeDataPtr),
          &spaceRect,
          SearchSolidTile,
          (ClientData) localDataPtr);
        /* for i/o pins */
        (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
          tmGlobalParentCellPtr), GetPinsInRect, (ClientData) localDataPtr);
    }
    return (0);
}


static int
SearchSolidTile(tilePtr, localDataPtr)
Tile      *tilePtr;
LocalData *localDataPtr;
{
    CellInst      *cellInstPtr;
    RouteCell     *routeCellPtr;

    if (Tp_IsCellInstTile(tilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        (void) Mac_ItrPinOfCellInst(cellInstPtr, GetPinsInRect,
          (ClientData) localDataPtr);
    }
#ifdef USE_DR_CODE
    if (Tp_IsRCellTile(tilePtr))
    {
        routeCellPtr = Tp_GetRCellOfTile(tilePtr);
        (void) Mac_ItrSidePinOfRCell(routeCellPtr, GetRPinsInRect,
          (ClientData) localDataPtr);
    }
#endif /* USE_DR_CODE */

    return (0);
}


#ifdef USE_DR_CODE
static BearStatus
GetRPinsInRect(rPinPtr, localDataPtr)
RoutePin  *rPinPtr;
LocalData *localDataPtr;
{
    LinkedRPin *linkRPinPtr;
    Point      absPos;

    (void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if ((absPos.p_x > localDataPtr->area->r_ll.p_x) &&
      (absPos.p_x < localDataPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localDataPtr->area->r_ll.p_y) &&
      (absPos.p_y < localDataPtr->area->r_ur.p_y))
    {
        MALLOC(LinkedRPin *, linkRPinPtr, sizeof (LinkedRPin));
        linkRPinPtr->rPinPtr = rPinPtr;
        linkRPinPtr->nextPtr = localDataPtr->rPinList;
        localDataPtr->rPinList = linkRPinPtr;
    }
    return (BEAR_CONTINUE_SEARCH);
}
#endif /* USE_DR_CODE */


static BearStatus
GetPinsInRect(pinInstPtr, localDataPtr)
PinInst   *pinInstPtr;
LocalData *localDataPtr;
{
    LinkedPinInst *pinPtr;
    Point         absPos;

    (void) Mac_GetAbsPosOfPin(pinInstPtr, &absPos);
    if ((absPos.p_x > localDataPtr->area->r_ll.p_x) &&
      (absPos.p_x < localDataPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localDataPtr->area->r_ll.p_y) &&
      (absPos.p_y < localDataPtr->area->r_ur.p_y))
    {
        MALLOC(LinkedPinInst *, pinPtr, sizeof (LinkedPinInst));
        pinPtr->pinInstPtr = pinInstPtr;
        pinPtr->nextPtr = localDataPtr->pinList;
        localDataPtr->pinList = pinPtr;
    }
    return (BEAR_CONTINUE_SEARCH);
}


static void
AddPPinsToList(bnTilePtr, direction, ppinListPtr)
Tile            *bnTilePtr;
SideDirection   direction;
LinkedPPin      **ppinListPtr;
{
    (void) Tp_ItrPPinsOfSide(bnTilePtr, direction, AddPins,
        (ClientData) ppinListPtr);
}


static BearStatus
AddPins(ppinPtr, ppinListPtr)
PseudoPin   *ppinPtr;
LinkedPPin  **ppinListPtr;
{
    LinkedPPin *lnPPinPtr;

    MALLOC(LinkedPPin *, lnPPinPtr, sizeof (LinkedPPin));
    lnPPinPtr->ppinPtr = ppinPtr;
    lnPPinPtr->nextPtr = *ppinListPtr;
    *ppinListPtr = lnPPinPtr;

    return (BEAR_CONTINUE_SEARCH);
}


