
/* tmRingUnify.c --
 *
 *         Tim Collins
 *      (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmRingUnify.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


typedef struct _LocalData
{
    BearId intId;
    BearId extId;
} LocalData;

/* ---------- Local Procedures ---------- */
static void FixAllBadNets();
static BearStatus ChangeBadPinIds();
static BearStatus ChangeBadRPinIds();


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_RingUnify () --
 *
 *    This procedure unifies for the ring router, all route pins, pin insts,
 *    and i/o pins externals now match.
 *
 * Results:
 *    All nets are unified.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */


void
Tm_RingUnify(parentCellPtr, hRect1, vRect1, hRect2, vRect2, junctPtr)
CellMaster   *parentCellPtr;
Rect         hRect1;
Rect         vRect1;
Rect         hRect2;
Rect         vRect2;
WallJunction *junctPtr;
{
    PinsAroundJunction pinLists;
    LinkedPPin         *vPPinList;

    tmGlobalParentCellPtr = parentCellPtr;
    tmGlobalPlaneDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    (void) Tm_HRectUnify(parentCellPtr, tmGlobalPlaneDataPtr, hRect1);
    (void) Tm_VRectUnify(parentCellPtr, tmGlobalPlaneDataPtr, vRect1);
    (void) Tm_HRectUnify(parentCellPtr, tmGlobalPlaneDataPtr, hRect2);
    (void) Tm_VRectUnify(parentCellPtr, tmGlobalPlaneDataPtr, vRect2);
    (void) Tm_PinsAroundJunction(tmGlobalParentCellPtr, 
      tmGlobalPlaneDataPtr, junctPtr, &pinLists);
    vPPinList = pinLists.vertPPinList;
    (void) FixAllBadNets(vPPinList);
}


static void 
FixAllBadNets(vPPinList)
LinkedPPin *vPPinList;
{
    BearId    intId;
    BearId    extId;
    Net       *netPtr;
    LocalData localData;

    while (vPPinList != NIL(LinkedPPin))
    {
        intId = Tp_GetIntNetIdOfPPin(vPPinList->ppinPtr);
        extId = Tp_GetExtNetIdOfPPin(vPPinList->ppinPtr);
        if (intId != extId)
        {
            localData.intId = intId;
            localData.extId = extId;
            netPtr = Tp_GetNetOfPPin(vPPinList->ppinPtr);
            (void) Mac_ItrPinOfNet(netPtr, ChangeBadPinIds, 
                (ClientData) &localData);
            (void) Mac_ItrRPinOfNet(netPtr, ChangeBadRPinIds,
                (ClientData) &localData);
        }
        vPPinList = vPPinList->nextPtr;
    }
}


static BearStatus
ChangeBadPinIds(pinPtr, localDataPtr)
PinInst   *pinPtr;
LocalData *localDataPtr;
{
    BearId extId;

    extId = Mac_GetExtNetIdOfPin(pinPtr);
    if ((extId == localDataPtr->intId) || (extId == localDataPtr->extId))
    {
        (void) Mac_PutExtNetIdOfPin(localDataPtr->extId, pinPtr);
    }
}


static BearStatus
ChangeBadRPinIds(rPinPtr, localDataPtr)
RoutePin  *rPinPtr;
LocalData *localDataPtr;
{
    BearId extId;

    extId = Mac_GetExtNetIdOfRPin(rPinPtr);
    if ((extId == localDataPtr->intId) || (extId == localDataPtr->extId))
    {
        (void) Mac_PutExtNetIdOfRPin(localDataPtr->extId, rPinPtr);
    }
}

