
/* tmVBnUnify.c --
 *
 *     Tim Collins
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmVBnUnify.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"

typedef struct _LocalInfo {
    Rect          *area;
    PseudoPin     *pseudoPin;
    Tile          *tilePtr;
    SideDirection side;
} LocalInfo;

typedef struct _DCellUpdateInfo {
    DummyCell               *dCellPtr; 
    Rect                    searchRect;
    PseudoPin               *trPPinList;
    PseudoPin               *blPPinList;
    struct _DCellUpdateInfo *nextPtr;
} DCellUpdateInfo;

typedef struct _FixInfo {
    DummyCell     *dCellPtr;
    PseudoPin     *pPinPtr;
    BearId        oldId;
    BearId        newId;
    SideDirection side;
} FixInfo;


/* ---------- Local Procedures ---------- */
static BearStatus   CheckTopPPins();
static BearStatus   CheckBottomPPins();
static BearStatus   CheckRPins();
static BearStatus   CheckPins();
static BearStatus   CheckLPPinsOfBnLine();
static BearStatus   CheckLPPinsOnDCell();
static BearStatus   ChangeRPPinsOfBnLine();
static BearStatus   ChangeRLPinsOfBn();
static BearStatus   ChangeDCellPPins();
static BearStatus   FixDCellSide();
static BearStatus   TmVBnUnify();
static void         FixInternalPins();
static void         AttachVBnPPinsAndDensity();
static void         AttachVBnLinePPins();
static void         FixBnTileCase();
static void         FixVPPinsOnBnLine();
static void         FixVPPins();
static void         AppendRightPPinLists();
static void         AppendLeftPPinLists();
static void         FixDCellPPins();
static int          SearchBnSides();
static int          LastDCellOnBNLine();


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_VBnUnify () --
 *
 *    This procedure deletes all the dummy cells, and transfers their pseudo
 *    pins to the appropriate vertical bottleneck tiles.
 *
 * Results:
 *    All the dummy cells are deleted.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

void
Tm_VBnUnify(parentCellPtr)
CellMaster *parentCellPtr;
{
    PlaneData    *planeDataPtr;
    WallJunction *fpgPtr;
    DCellUpdateInfo *dCellUpdateInfoList;
    DCellUpdateInfo *dCellUpdatePtr;
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;
    Tile            *topBnTilePtr;
    Tile            *bottomBnTilePtr;

    tmGlobalParentCellPtr = parentCellPtr;
    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    tmGlobalPlaneDataPtr = planeDataPtr;
    dCellUpdateInfoList = NIL(DCellUpdateInfo);
    /*
     * All the pseudo pin information on the dummy cell and the bottleneck
     * tiles must be saved, in the case of a dummy cell on a bottle neck line
     * or it will be lost when the tile plane is updated!
     */
    (void) Mac_ItrDCell(planeDataPtr->pd_dcellList, TmVBnUnify, 
      (ClientData) &dCellUpdateInfoList);
    dCellUpdatePtr = dCellUpdateInfoList;
    while (dCellUpdatePtr != NIL(DCellUpdateInfo))
    {
        (void) TmGetLAndRBnTilesOfDCell(dCellUpdatePtr->dCellPtr,
          &leftBnTilePtr, &rightBnTilePtr);
        (void) TmGetBAndTBnTilesOfDCell(dCellUpdatePtr->dCellPtr,
          &bottomBnTilePtr, &topBnTilePtr);
        (void) Tp_FreeTileProp(leftBnTilePtr);
        (void) Tp_FreeTileProp(rightBnTilePtr);
        (void) Tp_FreeTileProp(bottomBnTilePtr);
        (void) Tp_FreeTileProp(topBnTilePtr);
        /* new change: move from the place before the free */
        (void) Tp_UnpaintDCell(dCellUpdatePtr->dCellPtr, planeDataPtr);
        dCellUpdatePtr = dCellUpdatePtr->nextPtr;
    }
        /* newer change: move these 3 lines to identify space tiles before */
    /* attaching the ppins - since idenifySpaceTile creates new bnLine */
    /* new change: added Tp_Iden */
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &fpgPtr);

    dCellUpdatePtr = dCellUpdateInfoList;
    while (dCellUpdatePtr != NIL(DCellUpdateInfo))
    {
        if ((dCellUpdatePtr->dCellPtr->class == HBNTI_AND_VBNTI) ||
          (dCellUpdatePtr->dCellPtr->class == HBNLN_AND_VBNTI))
        {
        /*
         * Find the new vertical bottleneck tile and attach its pseudo
         * pin lists.
         */
        /*  (void) AttachVBnPPins(dCellUpdatePtr); */
        }
        else
        {
        /*
         * Find the new vertical semi-bottleneck tiles and attach their 
         * pseudo pin lists.
         */
            (void) AttachVBnLinePPins(dCellUpdatePtr);
        }
        dCellUpdatePtr = dCellUpdatePtr->nextPtr;
    }
    (void) Tp_FreeDCellList(planeDataPtr);
    Tp_IdentifySpaceTiles(planeDataPtr);
    (void) Fpg_CreateFloorPlanGraph(planeDataPtr, &fpgPtr);
}


static void 
AttachVBnPPinsAndDensity(dummyCellPtr, rPPinList, lPPinList, density)
DummyCell  *dummyCellPtr;
PseudoPin  *rPPinList;
PseudoPin  *lPPinList;
int        density;
{
    Point    hSearchPnt, vSearchPnt;
    Tile     *bnTilePtr;

    hSearchPnt.p_x = dummyCellPtr->bbox.r_ur.p_x;
    hSearchPnt.p_y = dummyCellPtr->bbox.r_ll.p_y;
    bnTilePtr = TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &hSearchPnt);
    if (bnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(bnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_HORZ_BN_TILE, bnTilePtr);
    }
    Tp_HPointToVPoint(&hSearchPnt, &vSearchPnt);
    bnTilePtr = TiSrPointNoHint(Tp_GetVertPlane(tmGlobalPlaneDataPtr),
      &vSearchPnt);
    if (bnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(bnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_VERT_BN_TILE, bnTilePtr);
    }
    (void) Tp_AppendPPinListOfTile(rPPinList,
      bnTilePtr, TOP_RIGHT_SIDE, TP_BOTTOM);
    (void) Tp_AppendPPinListOfTile(lPPinList,
      bnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM);

        Tp_PutDensityOfTile(density, bnTilePtr);
}


static void
AttachVBnLinePPins(dCellUpdatePtr)
DCellUpdateInfo *dCellUpdatePtr;
{
    Tile         *rightSBnTilePtr;
    Tile         *leftSBnTilePtr;
    Tile         *topSBnTilePtr;
    Tile         *bottomSBnTilePtr;
    BNLineType   bnType;

    (void) TmGetHSBnTilesOfDCell(dCellUpdatePtr->dCellPtr,
      &topSBnTilePtr, &bottomSBnTilePtr);
    if (topSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(topSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_HORZ_DOM_SBN_TILE, topSBnTilePtr);
    }
    if (bottomSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(bottomSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_HORZ_DOM_SBN_TILE, bottomSBnTilePtr);
    }
    (void) TmGetVSBnTilesOfDCell(dCellUpdatePtr->dCellPtr,
      &rightSBnTilePtr, &leftSBnTilePtr);
    if (rightSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(rightSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_VERT_DOM_SBN_TILE, rightSBnTilePtr);
    }
    if (leftSBnTilePtr->ti_client == (ClientData) MINFINITY)
    {
        (void) Tp_CreateSpaceTileProp(leftSBnTilePtr);
        (void) Tp_PutTypeOfSpaceTile(TP_VERT_DOM_SBN_TILE, leftSBnTilePtr);
    }
    (void) Tp_IsPairOfSBNTiles(&leftSBnTilePtr, &rightSBnTilePtr,
      &bnType);
    if (bnType == TP_VERT_R_UP)
    {
        if (dCellUpdatePtr->trPPinList != NIL(PseudoPin))
        {
        (void) Tp_AppendPPinListOfTile(dCellUpdatePtr->trPPinList,
          rightSBnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM);
        }
    }
    else
    {
        if (dCellUpdatePtr->trPPinList != NIL(PseudoPin))
        {
        (void) Tp_AppendPPinListOfTile(dCellUpdatePtr->trPPinList,
          rightSBnTilePtr, BOTTOM_LEFT_SIDE, TP_TOP);
        }
    }
}


static BearStatus
TmVBnUnify(dummyCellPtr, dCellUpdateInfoListPtr)
DummyCell       *dummyCellPtr;
DCellUpdateInfo **dCellUpdateInfoListPtr;
{
    /*
     * Fix and get the pseudo pins off the bottleneck tiles and the dummy cell.
     */
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;
    Tile            *bottomBnTilePtr;
    Tile            *topBnTilePtr;
    DCellUpdateInfo *updateInfoPtr;
    WallJunction    *fpgPtr;

    (void) FixInternalPins(dummyCellPtr);
    if ((dummyCellPtr->class == HBNTI_AND_VBNTI) ||
      (dummyCellPtr->class == HBNLN_AND_VBNTI))
    {
        (void) FixBnTileCase(dummyCellPtr);
    }
    else
    {
        /* 
         * Bottleneck line case.
         * There are two cases
         * 1. this is the only dummy cell on a horizontal bottleneck line
         *    we must create the save structure, because when this dcell
         *    is removed, a bnline will result.
         * 2. there are other dcells on this bnline, so the removal of this
         *    dcell will not cause damage to the bnline.
         */
        (void) TmGetBAndTBnTilesOfDCell(dummyCellPtr, &bottomBnTilePtr,
          &topBnTilePtr);
        (void) FixVPPinsOnBnLine(bottomBnTilePtr, dummyCellPtr, topBnTilePtr);
        (void) AppendRightPPinLists(bottomBnTilePtr, dummyCellPtr, 
          topBnTilePtr);
        /* new change: MALLOC moved before Tp_Remove */
        MALLOC(DCellUpdateInfo *, updateInfoPtr, sizeof (DCellUpdateInfo));
        (void) Tp_RemovePPinListOfTile(bottomBnTilePtr, TOP_RIGHT_SIDE,
          TP_BOTTOM, &updateInfoPtr->trPPinList);
        updateInfoPtr->dCellPtr = dummyCellPtr;
        updateInfoPtr->blPPinList = NIL(PseudoPin);
        if (LastDCellOnBNLine(bottomBnTilePtr, topBnTilePtr))
        {
            updateInfoPtr->nextPtr = *dCellUpdateInfoListPtr;
            *dCellUpdateInfoListPtr = updateInfoPtr;
        }
        else
        {
            (void) TmGetLAndRBnTilesOfDCell(dummyCellPtr, &leftBnTilePtr, 
              &rightBnTilePtr);
            (void) Tp_FreeTileProp(leftBnTilePtr);
            (void) Tp_FreeTileProp(rightBnTilePtr);
            (void) Tp_FreeTileProp(bottomBnTilePtr);
            (void) Tp_FreeTileProp(topBnTilePtr);
            /* new change: move from the place before the free */
            (void) Tp_UnpaintDCell(dummyCellPtr, tmGlobalPlaneDataPtr);
            Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
            (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &fpgPtr);

            (void) TmGetBAndTBnTilesOfDCell(dummyCellPtr, 
                            &bottomBnTilePtr,
                            &topBnTilePtr);
            (void) Tp_AppendPPinListOfTile(updateInfoPtr->trPPinList,
                               bottomBnTilePtr,
                               TOP_RIGHT_SIDE, TP_BOTTOM);
            FREE((char *) updateInfoPtr);
        }
    }
}


static int
LastDCellOnBNLine(bottomBnTilePtr, topBnTilePtr)
Tile            *bottomBnTilePtr;
Tile            *topBnTilePtr;
{
    Rect  rect;
    Point bPoint;
    Point tPoint;

    (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, bottomBnTilePtr, &rect);
    bPoint.p_x = rect.r_xbot + 1;
    bPoint.p_y = rect.r_ybot - 1;
    (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, topBnTilePtr, &rect);
    tPoint.p_x = rect.r_xbot + 1;
    tPoint.p_y = rect.r_ytop + 1;
    if (Tp_IsDCellTile(
      TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr), &bPoint)) ||
      Tp_IsDCellTile(
      TiSrPointNoHint(Tp_GetHorzPlane(tmGlobalPlaneDataPtr), &tPoint)))
    {
        return (0);
    }
    else
    {
        return (1);
    }
}


static void
FixBnTileCase(dummyCellPtr)
DummyCell       *dummyCellPtr;
{
    Tile            *rightBnTilePtr;
    Tile            *leftBnTilePtr;
    Tile            *bottomBnTilePtr;
    Tile            *topBnTilePtr;
    PseudoPin       *rPPinList;
    PseudoPin       *lPPinList;
    WallJunction    *fpgPtr;
    int             densityTotal;
    int             densityTop;
    int             densityBot;
    Rect            dummyBox;


    (void) TmGetLAndRBnTilesOfDCell(dummyCellPtr, &leftBnTilePtr, 
      &rightBnTilePtr);
    (void) TmGetBAndTBnTilesOfDCell(dummyCellPtr, &bottomBnTilePtr, 
      &topBnTilePtr);
        /* Calculate Density of Old Tiles */
    densityTop = Tp_GetDensityOfTile(topBnTilePtr);
    densityBot = Tp_GetDensityOfTile(bottomBnTilePtr);
    if (densityTop < 0 || densityBot < 0)
    {
       if (densityTop >= 0)
          densityTop = -(Tp_GetCapacityOfTile(topBnTilePtr));
       if (densityBot >= 0)
          densityBot = -(Tp_GetCapacityOfTile(bottomBnTilePtr));
           dummyBox = dummyCellPtr->bbox;
       densityTotal = -(dummyBox.r_ytop - dummyBox.r_ybot) 
              + densityTop + densityBot;
    }
    else
    {
       densityTotal = 0;
    }


    (void) AppendRightPPinLists(bottomBnTilePtr, dummyCellPtr, topBnTilePtr);
    (void) AppendLeftPPinLists(bottomBnTilePtr, dummyCellPtr, topBnTilePtr);
    (void) Tp_RemovePPinListOfTile(bottomBnTilePtr, TOP_RIGHT_SIDE, TP_BOTTOM,
      &rPPinList);
    (void) Tp_RemovePPinListOfTile(bottomBnTilePtr, BOTTOM_LEFT_SIDE,
      TP_BOTTOM, &lPPinList);
    (void) Tp_FreeTileProp(leftBnTilePtr);
    (void) Tp_FreeTileProp(rightBnTilePtr);
    (void) Tp_FreeTileProp(bottomBnTilePtr);
    (void) Tp_FreeTileProp(topBnTilePtr);
    /* new change: move from the place before the free */
    (void) Tp_UnpaintDCell(dummyCellPtr, tmGlobalPlaneDataPtr);
    Tp_IdentifySpaceTiles(tmGlobalPlaneDataPtr);
    (void) Fpg_CreateFloorPlanGraph(tmGlobalPlaneDataPtr, &fpgPtr);
    (void) AttachVBnPPinsAndDensity(dummyCellPtr, rPPinList, lPPinList,
                    densityTotal);
}



static void
FixInternalPins(dummyCellPtr)
DummyCell       *dummyCellPtr;
{
    Tile            *bottomBnTilePtr;
    Tile            *topBnTilePtr;

    (void) TmGetBAndTBnTilesOfDCell(dummyCellPtr, &bottomBnTilePtr, 
      &topBnTilePtr);
    if ((dummyCellPtr->class == HBNTI_AND_VBNTI) ||
      (dummyCellPtr->class == HBNLN_AND_VBNTI))
    {
        (void) FixVPPins(bottomBnTilePtr, dummyCellPtr, topBnTilePtr);
    }
}


static void
FixVPPinsOnBnLine(bottomBnTilePtr, dCellPtr, topBnTilePtr)
Tile      *bottomBnTilePtr;
DummyCell *dCellPtr;
Tile      *topBnTilePtr;
{
    /*
     * Since when the dummy cell is unpainted, the bottleneck tiles will turn
     * into a bottleneck line, we have to make the pseudo pin id's consistent.
     * (two pseudo pins turn into one pseudo pin)
     */
    (void) Tp_ItrPPinsOfSide(bottomBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      CheckLPPinsOfBnLine, (ClientData) bottomBnTilePtr);
    (void) Tp_ItrPPinsOfSide(topBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      CheckLPPinsOfBnLine, (ClientData) topBnTilePtr);
    (void) Mac_ItrPPinOfDCell(dCellPtr, (SideDirection) RIGHT_SIDE,
      CheckLPPinsOnDCell,
      (ClientData) dCellPtr);
}


static BearStatus
CheckLPPinsOfBnLine(pPinPtr, bnTilePtr)
PseudoPin *pPinPtr;
Tile      *bnTilePtr;
{

    (void) Tp_ItrPPinsOfSide(bnTilePtr,
      (SideDirection) BOTTOM_LEFT_SIDE,
      ChangeRPPinsOfBnLine, (ClientData) pPinPtr);
    return (BEAR_CONTINUE_SEARCH);
}
    

static BearStatus
ChangeRPPinsOfBnLine(pPinPtr, rPPinPtr)
PseudoPin *pPinPtr;
PseudoPin *rPPinPtr;
{
    BearId rIntId;
    BearId intId;
    BearId extId;

    rIntId = Tp_GetIntNetIdOfPPin(rPPinPtr);
    intId = Tp_GetIntNetIdOfPPin(pPinPtr);
    extId = Tp_GetExtNetIdOfPPin(pPinPtr);
    if (intId == rIntId)
    {
        (void) Tp_PutIntNetIdOfPPin(extId, rPPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckLPPinsOnDCell(pPinPtr, dCellPtr)
PseudoPin *pPinPtr;
DummyCell *dCellPtr;
{
    (void) Mac_ItrPPinOfDCell(dCellPtr, (SideDirection) LEFT_SIDE,
      ChangeRPPinsOfBnLine,
      (ClientData) pPinPtr);
    return (BEAR_CONTINUE_SEARCH);
}


static void
AppendRightPPinLists(bottomBnTilePtr, dCellPtr, topBnTilePtr)
Tile      *bottomBnTilePtr;
DummyCell *dCellPtr;
Tile      *topBnTilePtr;
{
    PseudoPin       *dummyRightPPinList;
    PseudoPin       *topBnRightPPinList;

    Mac_RemovePPinListOfDCell(dCellPtr, RIGHT_SIDE, &dummyRightPPinList);
    Tp_RemovePPinListOfTile(topBnTilePtr, TOP_RIGHT_SIDE, TP_BOTTOM,
      &topBnRightPPinList);
    if (dummyRightPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(dummyRightPPinList, bottomBnTilePtr,
          TOP_RIGHT_SIDE, TP_BOTTOM); 
    }
    if (topBnRightPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(topBnRightPPinList, bottomBnTilePtr,
           TOP_RIGHT_SIDE, TP_BOTTOM);
    }
}


static void
AppendLeftPPinLists(bottomBnTilePtr, dCellPtr, topBnTilePtr)
Tile      *bottomBnTilePtr;
DummyCell *dCellPtr;
Tile      *topBnTilePtr;
{
    PseudoPin       *dummyLeftPPinList;
    PseudoPin       *topBnLeftPPinList;

    Mac_RemovePPinListOfDCell(dCellPtr, LEFT_SIDE, &dummyLeftPPinList);
    Tp_RemovePPinListOfTile(topBnTilePtr, BOTTOM_LEFT_SIDE, TP_BOTTOM,
      &topBnLeftPPinList);
    if (dummyLeftPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(dummyLeftPPinList, bottomBnTilePtr,
          BOTTOM_LEFT_SIDE, TP_BOTTOM);
    }
    if (topBnLeftPPinList != NIL(PseudoPin))
    {
        (void) Tp_AppendPPinListOfTile(topBnLeftPPinList, bottomBnTilePtr,
          BOTTOM_LEFT_SIDE, TP_BOTTOM);
    }
}


static void
FixVPPins(bottomBnTilePtr, dCellPtr, topBnTilePtr)
Tile      *bottomBnTilePtr;
DummyCell *dCellPtr;
Tile      *topBnTilePtr;
{
    /*
     * Make the id's consistent across the bottleneck tiles so that all
     * pins and pseudo pins on the same net have the same internal id.
     */
    (void) Mac_ItrPPinOfDCell(dCellPtr, (SideDirection) BOTTOM_SIDE,
      CheckBottomPPins,
      (ClientData) bottomBnTilePtr);
    (void) Mac_ItrPPinOfDCell(dCellPtr, (SideDirection) TOP_SIDE,
      CheckTopPPins,
      (ClientData) topBnTilePtr);
}


static BearStatus
CheckBottomPPins(pPinPtr, bBnTilePtr)
PseudoPin *pPinPtr;
Tile      *bBnTilePtr;
{
    /* 
     * Propagate consistent id values across all sides of the bottom bottleneck
     * tile and the part of the cell that abuts it.
     */
    Rect      searchRect;
    LocalInfo localInfo;

    (void) Tp_ItrPPinsOfSide(bBnTilePtr,
      (SideDirection) BOTTOM_LEFT_SIDE,
      ChangeRLPinsOfBn, (ClientData) pPinPtr);
    (void) Tp_ItrPPinsOfSide(bBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      ChangeRLPinsOfBn, (ClientData) pPinPtr);
    /*
     * Get adjacent cell instance and check pin which are on the edge of the
     * right bottleneck tile.
     */
    (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, bBnTilePtr, &searchRect);
    searchRect.r_ur.p_y = searchRect.r_ll.p_y + 1;
    searchRect.r_ll.p_y -= 1;
    localInfo.area = &searchRect;
    localInfo.pseudoPin = pPinPtr;
    localInfo.side = TOP_SIDE;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &searchRect,
      SearchBnSides,
      (ClientData) &localInfo);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
      tmGlobalParentCellPtr), CheckPins, (ClientData) &localInfo);
    return (BEAR_CONTINUE_SEARCH);
}


static int
SearchBnSides(tilePtr, localInfoPtr)
Tile      *tilePtr;
LocalInfo *localInfoPtr;
{
    CellInst      *cellInstPtr;
    RouteCell     *routeCellPtr;
    FixInfo       fixInfo;

    if (Tp_IsCellInstTile(tilePtr))
    {
        cellInstPtr = Tp_GetCellInstOfTile(tilePtr);
        (void) Mac_ItrPinOfCellInst (cellInstPtr, CheckPins,
          (ClientData) localInfoPtr);
    }
    else
    {
#ifdef USE_DR_CODE
        if (Tp_IsRCellTile(tilePtr))
        {
            routeCellPtr = Tp_GetRCellOfTile(tilePtr);
            (void) Mac_ItrSidePinOfRCell(routeCellPtr, CheckRPins,
              (ClientData) localInfoPtr);
        }
        else
#endif /* USE_DR_CODE */
        {
            if (Tp_IsDCellTile(tilePtr))
            {
                fixInfo.dCellPtr = Tp_GetDCellOfTile(tilePtr);
                fixInfo.pPinPtr = localInfoPtr->pseudoPin;
                fixInfo.side = localInfoPtr->side;
                (void) Mac_ItrPPinOfDCell(fixInfo.dCellPtr, 
                  localInfoPtr->side,
                  ChangeDCellPPins, (ClientData) &fixInfo);
            }
        }
    }
    return (0);
}


static BearStatus
CheckTopPPins(pPinPtr, tBnTilePtr)
PseudoPin *pPinPtr;
Tile      *tBnTilePtr;
{
    /* 
     * Propagate consistent id values across all sides of the right bottleneck
     * tile and the part of the cell that abuts it.
     */
    Rect      searchRect;
    LocalInfo localInfo;

    (void) Tp_ItrPPinsOfSide(tBnTilePtr,
      (SideDirection) BOTTOM_LEFT_SIDE,
      ChangeRLPinsOfBn, (ClientData) pPinPtr);
    (void) Tp_ItrPPinsOfSide(tBnTilePtr,
      (SideDirection) TOP_RIGHT_SIDE,
      ChangeRLPinsOfBn, (ClientData) pPinPtr);
    /*
     * Get adjacent cell instance and check pin which are on the edge of the
     * right bottleneck tile.
     */
    (void) Tp_VertTileToRect(tmGlobalPlaneDataPtr, tBnTilePtr, &searchRect);
    searchRect.r_ll.p_y = searchRect.r_ur.p_y - 1;
    searchRect.r_ur.p_y += 1;
    localInfo.area = &searchRect;
    localInfo.pseudoPin = pPinPtr;
    localInfo.side = BOTTOM_SIDE;
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(tmGlobalPlaneDataPtr),
      &searchRect,
      SearchBnSides,
      (ClientData) &localInfo);
    /* for i/o pins */
    (void) Mac_ItrPinOfCellInst(Mac_GetInstOfCoreMaster(
      tmGlobalParentCellPtr), CheckPins, (ClientData) &localInfo);
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
ChangeDCellPPins(pPinPtr, fixInfoPtr)
PseudoPin *pPinPtr;
FixInfo   *fixInfoPtr;
{
    BearId pExtId;
    BearId intId;
    BearId extId;

    pExtId = Tp_GetExtNetIdOfPPin(pPinPtr);
    intId = Tp_GetIntNetIdOfPPin(fixInfoPtr->pPinPtr);
    extId = Tp_GetExtNetIdOfPPin(fixInfoPtr->pPinPtr);
    if (pExtId == extId)
    {
        (void) Tp_PutExtNetIdOfPPin(intId, pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}

/*
static void
FixDCellPPins(fixInfoPtr, side)
FixInfo       *fixInfoPtr;
SideDirection side;
{
    if (side != (SideDirection) RIGHT_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr,
          (SideDirection) RIGHT_SIDE,
          FixDCellSide, (ClientData) fixInfoPtr);
    }
    if (side != (SideDirection) LEFT_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr,
          (SideDirection) LEFT_SIDE,
          FixDCellSide, (ClientData) fixInfoPtr);
    }
    if (side != (SideDirection) TOP_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr,
          (SideDirection) TOP_SIDE,
          FixDCellSide, (ClientData) fixInfoPtr);
    }
    if (side != (SideDirection) BOTTOM_SIDE)
    {
        (void) Mac_ItrPPinOfDCell(fixInfoPtr->dCellPtr,
          (SideDirection) BOTTOM_SIDE, 
          FixDCellSide, (ClientData) fixInfoPtr);
    }
}


static BearStatus
FixDCellSide(pPinPtr, fixInfoPtr)
PseudoPin  *pPinPtr;
FixInfo    *fixInfoPtr;
{
    BearId intId;

    intId = Tp_GetIntNetIdOfPPin(fixInfoPtr->pPinPtr);
    if (intId == fixInfoPtr->oldId)
    {
        (void) Tp_PutIntNetIdOfPPin(fixInfoPtr->newId, pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}
*/


static BearStatus
CheckPins(pinPtr,localInfoPtr)
PinInst   *pinPtr;
LocalInfo *localInfoPtr;
{
    Point  absPos;
    BearId pIntId;
    BearId pExtId;
    BearId extId;

    (void) Mac_GetAbsPosOfPin(pinPtr, &absPos);
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
      (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
      (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        pIntId = Tp_GetIntNetIdOfPPin(localInfoPtr->pseudoPin);
        extId = Mac_GetVExtNetIdOfPin(pinPtr);
        pExtId = Tp_GetExtNetIdOfPPin(localInfoPtr->pseudoPin);
        if (extId == pExtId)
        {
            (void) Mac_PutVExtNetIdOfPin(pIntId, pinPtr);
            /******NEW*****/
            (void) Mac_PutExtNetIdOfPin(pIntId, pinPtr);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
CheckRPins(rPinPtr,localInfoPtr)
RoutePin  *rPinPtr;
LocalInfo *localInfoPtr;
{
    Point  absPos;
    BearId pIntId;
    BearId pExtId;
    BearId extId;

    (void) Mac_GetAbsPosOfRPin(rPinPtr, &absPos);
    if ((absPos.p_x >= localInfoPtr->area->r_ll.p_x) &&
      (absPos.p_x <= localInfoPtr->area->r_ur.p_x) &&
      (absPos.p_y >= localInfoPtr->area->r_ll.p_y) &&
      (absPos.p_y <= localInfoPtr->area->r_ur.p_y))
    {
        pIntId = Tp_GetIntNetIdOfPPin(localInfoPtr->pseudoPin);
        extId = Mac_GetVExtNetIdOfRPin(rPinPtr);
        pExtId = Tp_GetExtNetIdOfPPin(localInfoPtr->pseudoPin);
        if (extId == pExtId)
        {
            (void) Mac_PutVExtNetIdOfRPin(pIntId, rPinPtr);
            /******NEW*****/
            (void) Mac_PutExtNetIdOfRPin(pIntId, rPinPtr);
        }
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
ChangeRLPinsOfBn(pPinPtr, btPPinPtr)
PseudoPin *pPinPtr;
PseudoPin *btPPinPtr;
{
    BearId btIntId;
    BearId intId;
    BearId btExtId;

    btIntId = Tp_GetIntNetIdOfPPin(btPPinPtr);
    intId = Tp_GetIntNetIdOfPPin(pPinPtr);
    btExtId = Tp_GetExtNetIdOfPPin(btPPinPtr);
    if (intId == btExtId)
    {
        (void) Tp_PutIntNetIdOfPPin(btIntId, pPinPtr);
    }
    return (BEAR_CONTINUE_SEARCH);
}



