
/* tmVCSplit.c --
 *
 *         Sherry Solden
 *     (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmVCSplit.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "gtGraph.h"
#include "fpg.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


typedef struct _VBnLineData {
	Tile *bTilePtr;
	Tile *tTilePtr;
	BNLine *bnLinePtr;
} VBnLineData;


/* ---------- Local Procedures ---------- */
static int        VCSplit();
static BNLine     *GetBnLineOfSBnTiles2();
static BearStatus MatchVertBnLine();

static void       DoBnRegion();
static void       DoFaceRegion();
static BearId     GetNetIdOfPPin2();
static BearStatus FindPPin();
static bool       IsPinOnTRSideOfTile2();


/* ---------- Local Global Variables ---------- */
static ZoneData   *zonePtr;
static Table      *tableHeadPtr;
static int        zoneNum;


/*
 * ----------------------------------------------------------------------------
 *
 * Tm_VCSplit () --
 *
 * This procedure looks at list of adjacent vertical tiles one by one, and
 * translates the global routing information that is on the surrounding
 * circle and puts it on the vertical plane.
 *
 * The circle is a set of monotonically decreasing in x-direction set
 * of vertical space tiles.  Pins are previously unified around this
 * circle.  Where pins exist on top/bottom edges of bottleneck tiles
 * and left/right edges of faces.
 *
 * Results:
 *	 All the routing information is on vertical plane.
 *
 * Side effects:
 *   none
 * ----------------------------------------------------------------------------
 */

void
Tm_VCSplit(parentCellPtr, planeDataPtr, domRect)
CellMaster *parentCellPtr;
PlaneData  *planeDataPtr;
Rect       domRect;
{
    WallJunction    *llJunctPtr;
    ZoneData        *zoneDataHeadPtr;
    bool            sbnTileFoundFlag;  

	tmGlobalParentCellPtr = parentCellPtr;
	tmGlobalPlaneDataPtr = planeDataPtr;
/*
 * code change : by george : on 18 oct 89
 *
	llJunctPtr = Mac_GetFpgOfCell(parentCellPtr);
 */

	zoneNum = 1;
	tableHeadPtr = NIL(Table);
/*
 * code change : by george : on 23 oct 89
 *
    MALLOC(ZoneData *, zonePtr, sizeof (ZoneData));
    zonePtr->rPinList = NIL(LinkedRPin);
	zonePtr->pinInstList = NIL(LinkedPinInst);
	zonePtr->pPinList = NIL(LinkedPPin);
 */
    zonePtr = TmMakeZoneData();
	zoneDataHeadPtr = zonePtr;

	sbnTileFoundFlag = FALSE;
    (void) TiSrArea(NIL(Tile), 
        Tp_GetVertPlane(planeDataPtr),
        &domRect,
        VCSplit,
        (ClientData) &sbnTileFoundFlag);
	zonePtr->nextPtr = NIL(ZoneData);
	TmHSplitByZones(zoneNum, tableHeadPtr, zoneDataHeadPtr);

	(void) Fpg_CreateFloorPlanGraph(planeDataPtr, &llJunctPtr);
}



static int
VCSplit(tilePtr, sbnTileFoundFlagPtr)
Tile    *tilePtr;
bool    *sbnTileFoundFlagPtr;  
{
/*
 * code change : by george : on 18 oct 89
 *
    static bool SBNTileFound = FALSE;
 */
 
    BNLine      *bnLinePtr;
    BNLineType  bnType;
    Tile        *semiTile1Ptr;
    Tile        *semiTile2Ptr;

    if (Tp_IsSBNTile(tilePtr))
    {
	if (*sbnTileFoundFlagPtr == FALSE)
	{
	    *sbnTileFoundFlagPtr = TRUE;
	    semiTile1Ptr = tilePtr;
	}
	else
	{
	    semiTile2Ptr = tilePtr;
            if (!Tp_IsPairOfSBNTiles(&semiTile1Ptr,
				     &semiTile2Ptr,
				     &bnType))
	    {
                /* Didn't find a match */
		semiTile1Ptr = tilePtr;
	    }
            if (bnType != TP_NULL_BNL)
	    {
		zoneNum++;
/*
 * code change : by george : on 23 oct 89
 *
		MALLOC(ZoneData *, zonePtr->nextPtr, sizeof (ZoneData));
 */
		zonePtr->nextPtr = TmMakeZoneData();
		zonePtr = zonePtr->nextPtr;
		zonePtr->zoneType = C_LINE;
		bnLinePtr = GetBnLineOfSBnTiles2(semiTile1Ptr, semiTile2Ptr);
        zonePtr->bnPtr = Fpg_GetBottleNeckOfWallSeg(bnLinePtr->segPtr);
		*sbnTileFoundFlagPtr = FALSE;
	    }
	}
    }
	else 
	{
	   *sbnTileFoundFlagPtr = FALSE;
	} 

   if (Tp_IsBottleneckTile(tilePtr))
   {
      zoneNum++;
/*
 * code change : by george : on 23 oct 89
 *
      MALLOC(ZoneData *, zonePtr->nextPtr, sizeof (ZoneData));
 */
      zonePtr->nextPtr = TmMakeZoneData();
      zonePtr = zonePtr->nextPtr;
      /* Delete old PPins from bnTile */
      (void) Tp_FreePPinsOfTile(tilePtr);
      DoBnRegion(tilePtr);
   }
   else if (Tp_IsDominantTile(tilePtr))
   {
      zoneNum++;
/*
 * code change : by george : on 23 oct 89
 *
      MALLOC(ZoneData *, zonePtr->nextPtr, sizeof (ZoneData));
 */
      zonePtr->nextPtr = TmMakeZoneData();
      zonePtr = zonePtr->nextPtr;
      DoFaceRegion(tilePtr);
   }
   return (MAGIC_CONTINUE_SEARCH);
}



static BNLine *
GetBnLineOfSBnTiles2(tTilePtr, bTilePtr)
Tile *tTilePtr;
Tile *bTilePtr;
{
    VBnLineData vertBnLineData;

/*
 * code change : by george : on 18 oct 89
 *
    vertBnLineData.bTilePtr = bTilePtr;
    vertBnLineData.tTilePtr = tTilePtr;
    (void) Tp_ItrHorzBnLine(tmGlobalPlaneDataPtr, MatchVertBnLine,
		(ClientData) &vertBnLineData);
 */

    vertBnLineData.bTilePtr = bTilePtr;
    vertBnLineData.tTilePtr = tTilePtr;
    vertBnLineData.bnLinePtr = NIL(BNLine);
    (void) Tp_ItrVertBnLine(tmGlobalPlaneDataPtr, MatchVertBnLine,
		(ClientData) &vertBnLineData);

    return (vertBnLineData.bnLinePtr);

}



static BearStatus
MatchVertBnLine(bnLinePtr, vertBnLineDataPtr)
BNLine      *bnLinePtr;
VBnLineData *vertBnLineDataPtr;
{
	if ((Tp_GetBRTileOfBnLine(bnLinePtr) == vertBnLineDataPtr->bTilePtr) &&
		(Tp_GetTLTileOfBnLine(bnLinePtr) == vertBnLineDataPtr->tTilePtr))
	{
		vertBnLineDataPtr->bnLinePtr = bnLinePtr;
		return (BEAR_STOP_SEARCH);
	}
	return (BEAR_CONTINUE_SEARCH);
}



static void
DoBnRegion(tilePtr)
Tile *tilePtr;
{
	WallSegment        *segPtr;
	BottleNeck         *bnPtr;
	PinsAroundJunction *pinListPtr;
	LinkedPPin         *pPinPtr;
	LinkedRPin         *rPinPtr;
	LinkedPinInst      *pinInstPtr;
	Net                *netPtr;
	BearId             id;
	float              width;
	float              *widthVect;

	segPtr = Tp_GetWallSegmentOfBnTile(tilePtr);
	bnPtr = Fpg_GetBottleNeckOfWallSeg(segPtr);
	zonePtr->bnPtr = bnPtr;
	if (bnPtr->unType == TP_BN_LINE) /* bottleneck line */
	{
		zonePtr->zoneType = V_LINE;
		zonePtr->pinInstList = NIL(LinkedPinInst);
		zonePtr->pPinList = NIL(LinkedPPin);
		zonePtr->rPinList = NIL(LinkedRPin);
	}
	else
	{                               /* bottleneck tile */
	        zonePtr->zoneType = V_FACE;
		zonePtr->pinInstList = NIL(LinkedPinInst);
		zonePtr->pPinList = NIL(LinkedPPin);
		zonePtr->rPinList = NIL(LinkedRPin);
		MALLOC(PinsAroundJunction *, pinListPtr, 
		       sizeof (PinsAroundJunction));
		(void) Tm_GetPinsInVBottleNeckRegion(tmGlobalParentCellPtr,
			                             tmGlobalPlaneDataPtr,
						     tilePtr,
						     pinListPtr);
		zonePtr->pinInstList = pinListPtr->pinInstList;
		zonePtr->rPinList = pinListPtr->rPinList;
		zonePtr->pPinList = pinListPtr->horzPPinList;
	}

	pinInstPtr = zonePtr->pinInstList;
	rPinPtr = zonePtr->rPinList;
	pPinPtr = pinListPtr->horzPPinList;

	while (pinInstPtr != NIL(LinkedPinInst))
	{
		id = Mac_GetVExtNetIdOfPin(pinInstPtr->pinInstPtr);
		netPtr = Mac_GetNetOfPin(pinInstPtr->pinInstPtr);
		widthVect = Mac_GetWidthVectOfPin(pinInstPtr->pinInstPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id,width, netPtr, zoneNum);
		pinInstPtr = pinInstPtr->nextPtr;
	}
	while (rPinPtr != NIL(LinkedRPin))
	{
		id = Mac_GetVExtNetIdOfRPin(rPinPtr->rPinPtr);
		netPtr = Mac_GetNetPtrOfRPin(rPinPtr->rPinPtr);
		widthVect = Mac_GetWidthVectOfRPin(rPinPtr->rPinPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id,width, netPtr, zoneNum);
		rPinPtr = rPinPtr->nextPtr;
	}
	while (pPinPtr != NIL(LinkedPPin))
	{
            /* Check for ppins of dummy cell */
            if ((pPinPtr->ppinPtr->tilePtr != NIL(Tile)) && 
   	        (Tp_IsDCellTile(pPinPtr->ppinPtr->tilePtr)))
	    {
		id = Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr);
		netPtr = Tp_GetNetOfPPin(pPinPtr->ppinPtr);
		width = Tp_GetWidthOfPPin(pPinPtr->ppinPtr);
		(void) TmNumberTable(&tableHeadPtr, id,width, netPtr, zoneNum);
	    }
   	    pPinPtr = pPinPtr->nextPtr;
	}
}


static void
DoFaceRegion(tilePtr)
Tile *tilePtr;
{
	Rect               domRect;
	PinsAroundJunction *pinListPtr;
	LinkedPPin         *pPinPtr;
	LinkedRPin         *rPinPtr;
	LinkedPinInst      *pinInstPtr;
	Net                *netPtr;
	BearId             id;
	float              width;
	float              *widthVect;
        Point              centerPos;

	zonePtr->zoneType = JUNCTION;
	zonePtr->pinInstList = NIL(LinkedPinInst);
	zonePtr->rPinList = NIL(LinkedRPin);
	zonePtr->pPinList = NIL(LinkedPPin);
	Tp_VertTileToRect(tmGlobalPlaneDataPtr, tilePtr, &domRect);
	pinListPtr = Tm_PinsAroundVFace(tmGlobalParentCellPtr,
					tmGlobalPlaneDataPtr,
					domRect,
					FALSE);

	zonePtr->pinInstList = pinListPtr->pinInstList;
	zonePtr->rPinList = pinListPtr->rPinList;
	zonePtr->pPinList = pinListPtr->vertPPinList;

	pinInstPtr = pinListPtr->pinInstList;
	rPinPtr = pinListPtr->rPinList;
	pPinPtr = pinListPtr->vertPPinList;

	while (pinInstPtr != NIL(LinkedPinInst))
	{
		id = Mac_GetVExtNetIdOfPin(pinInstPtr->pinInstPtr);
		netPtr = Mac_GetNetOfPin(pinInstPtr->pinInstPtr);
		widthVect = Mac_GetWidthVectOfPin(pinInstPtr->pinInstPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id,width, netPtr, zoneNum);
		pinInstPtr = pinInstPtr->nextPtr;
	}
	while (rPinPtr != NIL(LinkedRPin))
	{
		id = Mac_GetVExtNetIdOfRPin(rPinPtr->rPinPtr);
		netPtr = Mac_GetNetPtrOfRPin(rPinPtr->rPinPtr);
		widthVect = Mac_GetWidthVectOfRPin(rPinPtr->rPinPtr);
		width = widthVect[METAL2_LAYER];
		(void) TmNumberTable(&tableHeadPtr, id,width, netPtr, zoneNum);
		rPinPtr = rPinPtr->nextPtr;
	}
	while (pPinPtr != NIL(LinkedPPin))
	{
	    /* Must compare the PPin to the domRect to determine the correct id */
   	    centerPos.p_x = domRect.r_xbot + ((domRect.r_xtop - domRect.r_xbot) / 2);
	    centerPos.p_y = domRect.r_ybot + ((domRect.r_ytop - domRect.r_ybot) / 2);
	    id = GetNetIdOfPPin2(pPinPtr->ppinPtr, centerPos);
	    netPtr = Tp_GetNetOfPPin(pPinPtr->ppinPtr);
	    width = Tp_GetWidthOfPPin(pPinPtr->ppinPtr);
	    (void) TmNumberTable(&tableHeadPtr, id,width, netPtr, zoneNum);
	    pPinPtr = pPinPtr->nextPtr;
	}
}



static BearId
GetNetIdOfPPin2(ppinPtr, point)
PseudoPin *ppinPtr;
Point point;
{
    Tile *tilePtr;
    

    tilePtr = Tp_GetTileOfPPin(ppinPtr);
    if (Tp_IsSBNTile(tilePtr) == TRUE)
    {
	if (((TiBottom(tilePtr) > point.p_y) && Tp_IsHorzTile(tilePtr)) ||
	    ((TiLeft(tilePtr) > point.p_x) && Tp_IsVertTile(tilePtr)))
	{
	    return (Tp_GetIntNetIdOfPPin(ppinPtr));
	}
	else if (((TiTop(tilePtr) < point.p_y) && Tp_IsHorzTile(tilePtr)) ||
		 ((TiRight(tilePtr) < point.p_x) && Tp_IsVertTile(tilePtr)))
        {
	    return (Tp_GetExtNetIdOfPPin(ppinPtr));
	}
	else if (IsPinOnTRSideOfTile2(ppinPtr, tilePtr) == TRUE)
	{
	    return (Tp_GetIntNetIdOfPPin(ppinPtr));
	}
	else
	{
	    return (Tp_GetExtNetIdOfPPin(ppinPtr));
	}
	
    }
    else
    {
	return (Tp_GetExtNetIdOfPPin(ppinPtr));
    }
    
}



static bool
IsPinOnTRSideOfTile2(pPinPtr, tilePtr)
PseudoPin *pPinPtr;
Tile *tilePtr;
{
     BearStatus status;

     status = Tp_ItrPPinsOfSide(tilePtr, TOP_RIGHT_SIDE, FindPPin, 
				(ClientData) pPinPtr);
     if (status == BEAR_STOP_SEARCH)
     {
	 return (TRUE);
     }
     else
     {
	 return (FALSE);
     }
}

static BearStatus
FindPPin(tmpPinPtr, pPinPtr)
PseudoPin *tmpPinPtr;
PseudoPin *pPinPtr;
{            
     if (tmpPinPtr == pPinPtr)
     {
	 return (BEAR_STOP_SEARCH);
     }
     else 
     {
	 return (BEAR_CONTINUE_SEARCH);
     }
}

