		
/* tmVSplitByZones.c --
 *
 *          Tim Collins
 *     (Professor E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileMapping/tmVSplitByZones.c,v 1.1.1.1 2002/12/31 15:51:10 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "tile.h"
#include "bear.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "mac.h"
#include "tileMapping.h"
#include "tileMappingInt.h"


static void DoBottomDCell();
static void DoTopDCell();
static void DoBottomHFace();
static void DoTopHFace();
static void DoBottomHTile();
static void DoTopHTile();
static void DoHLine();
static void DoCLine();
static void PropagateNewId();
static void CopyVExtToExt();

void
TmVSplitByZones(numOfZones, tablePtr, zoneDataPtr)
int      numOfZones;
Table    *tablePtr;
ZoneData *zoneDataPtr;
{
    /*
     * This procedure uses a table of starting and ending zones of pins and
     * pseudo pins to determine the pseudo pins to be put on the dummy cells,
     * and their correct id's.
     */
    int           i;
    Table         *entryPtr;
    Table         *placePtr;
    BearId        newId;
    ZoneData      *zonePlacePtr;

    if (numOfZones <= 1)
    {
	TmFreeZoneData(&zoneDataPtr);
	return;
    }
    /* 
     * The first one is skipped, because we allocated one too many for the 
     * sake of homogeneity in the code.
     */
    /* make first region consisten */
    CopyVExtToExt(zoneDataPtr);
    zonePlacePtr = zoneDataPtr->nextPtr;
	TmFreeZoneData(&zoneDataPtr);
    for (i = 2; i <= numOfZones; i++)
    {
	entryPtr = tablePtr;
	while (entryPtr != NIL(Table))
	{
	    /* 
	     * This means that a net crosses the side of a zone
	     * therefore a pseudo pin must be put there.
	     */
	    if ((entryPtr->startZone <= i) && (entryPtr->endZone > 1))
	    {
		   newId = Mac_GenerateId();
		   /* If bottleneck line */
		   if ((zonePlacePtr->zoneType == H_LINE) &&
		       (entryPtr->startZone < i))
		   {
		       DoHLine(entryPtr, newId, zonePlacePtr);
		       entryPtr->currentId = newId;
		   }
		   else if ((zonePlacePtr->zoneType == C_LINE) &&
			    (entryPtr->startZone < i))
		   {
		       DoCLine(entryPtr, newId, zonePlacePtr);
		       entryPtr->currentId = newId;
		   }
		   else
		   {
		       if (entryPtr->startZone < i)
		       {
				  /* If net started before this zone - create ppin
				   * on an edge of tile */
			      if (zonePlacePtr->zoneType == DCELL)
			      {
			          DoBottomDCell(entryPtr, newId, zonePlacePtr);
			      }
			      if (zonePlacePtr->zoneType == H_TILE)
			      {
			          DoBottomHTile(entryPtr, newId, zonePlacePtr);
			      }
			      if (zonePlacePtr->zoneType == H_FACE)
			      {
			          DoTopHFace(entryPtr, newId, zonePlacePtr);
			      }

				  /* Propagate the id into the zone */
			      if ((zonePlacePtr->zoneType == JUNCTION) ||
			          (zonePlacePtr->zoneType == DJUNCTION))
			      {
			          PropagateNewId(entryPtr,
			                     entryPtr->currentId, zonePlacePtr);
			      }
			      else
			      {
			          PropagateNewId(entryPtr, newId, zonePlacePtr);
			      }
		       }
		       else
		       {
			   /*
			    * this is the starting zone so just keep the
			    * current id, and propagate into the zone.
			    */
			      newId = entryPtr->currentId;
			      PropagateNewId(entryPtr, newId, zonePlacePtr);
		       }
		       /*
		        * If net ends in this zone, delete the entry from the table.
		        */
		       if (entryPtr->endZone == i)
		       {
			      placePtr = tablePtr;
			      if (placePtr == entryPtr)
			      {
			          tablePtr = tablePtr->nextPtr;
			          FREE((char *)entryPtr);
			          entryPtr = tablePtr;
			      }
			      else
			      {
			          while (placePtr->nextPtr != entryPtr)
			          {
				         placePtr = placePtr->nextPtr;
			          }
			          placePtr->nextPtr = placePtr->nextPtr->nextPtr;
			          FREE((char *)entryPtr);
			          entryPtr = placePtr->nextPtr;
			      }
			      continue;
		       }
		       else
		       {
				  /* Net continues into next zone */
				  /* Create ppins on "other" edge of zone */
			      if (zonePlacePtr->zoneType == DCELL)
			      {
			          entryPtr->currentId = newId;
			          newId = Mac_GenerateId();
			          DoTopDCell(entryPtr, newId, zonePlacePtr);
			          entryPtr->currentId = newId;
			      }
			      if (zonePlacePtr->zoneType == H_TILE)
			      {
			          entryPtr->currentId = newId;
			          newId = Mac_GenerateId();
			          DoTopHTile(entryPtr, newId, zonePlacePtr);
			          entryPtr->currentId = newId;
			      }
			      if (zonePlacePtr->zoneType == H_FACE)
			      {
			          entryPtr->currentId = newId;
			          newId = Mac_GenerateId();
			          DoBottomHFace(entryPtr, newId, zonePlacePtr);
			          entryPtr->currentId = newId;
			      }
		       }
		   }
	    }
	    entryPtr = entryPtr->nextPtr;
	}
	zoneDataPtr  = zonePlacePtr;
	zonePlacePtr = zonePlacePtr->nextPtr;
	TmFreeZoneData(&zoneDataPtr);
    }
}


static void
DoBottomDCell(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;

	(void) Tp_CreatePPinOfDCell(zonePlacePtr->dCellPtr,
	  BOTTOM_SIDE, &newPPinPtr);
	(void) Tp_PutIntNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutExtNetIdOfPPin(entryPtr->currentId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}


static void
DoTopDCell(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;

	(void) Tp_CreatePPinOfDCell(zonePlacePtr->dCellPtr,
	  TOP_SIDE, &newPPinPtr);
	(void) Tp_PutIntNetIdOfPPin(entryPtr->currentId, newPPinPtr);
	(void) Tp_PutExtNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}


static void
DoBottomHFace(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;

	(void) Tp_CreatePPin(zonePlacePtr->bnPtr->unValue.bnTilePtr, 
	  BOTTOM_LEFT_SIDE, TP_BOTTOM, &newPPinPtr);
	(void) Tp_PutIntNetIdOfPPin(entryPtr->currentId, newPPinPtr);
	(void) Tp_PutExtNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}


static void
DoTopHFace(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;

	(void) Tp_CreatePPin(zonePlacePtr->bnPtr->unValue.bnTilePtr,
	  TOP_RIGHT_SIDE, TP_BOTTOM, &newPPinPtr);
	(void) Tp_PutExtNetIdOfPPin(entryPtr->currentId, newPPinPtr);
	(void) Tp_PutIntNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}

static void
DoBottomHTile(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;

	(void) Tp_CreatePPin(zonePlacePtr->bnPtr->unValue.bnTilePtr, 
	  BOTTOM_LEFT_SIDE, TP_BOTTOM, &newPPinPtr);
	(void) Tp_PutIntNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutExtNetIdOfPPin(entryPtr->currentId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}


static void
DoTopHTile(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;

	(void) Tp_CreatePPin(zonePlacePtr->bnPtr->unValue.bnTilePtr,
	  TOP_RIGHT_SIDE, TP_BOTTOM, &newPPinPtr);
	(void) Tp_PutIntNetIdOfPPin(entryPtr->currentId, newPPinPtr);
	(void) Tp_PutExtNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}


static void
DoHLine(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;
        BNLine        *bnLinePtr;

        bnLinePtr = zonePlacePtr->bnPtr->unValue.bnLinePtr;
	(void) Tp_CreatePPinOfBnLine(bnLinePtr, &newPPinPtr);
        (void) Tp_PutIntNetIdOfPPin(entryPtr->currentId, newPPinPtr);
        (void) Tp_PutExtNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}

static void
DoCLine(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
	PseudoPin     *newPPinPtr;
        BNLine        *bnLinePtr;

        bnLinePtr = zonePlacePtr->bnPtr->unValue.bnLinePtr;
	(void) Tp_CreatePPinOfBnLine(bnLinePtr, &newPPinPtr);
        (void) Tp_PutExtNetIdOfPPin(entryPtr->currentId, newPPinPtr);
        (void) Tp_PutIntNetIdOfPPin(newId, newPPinPtr);
	(void) Tp_PutNetOfPPin(entryPtr->netPtr, newPPinPtr);
	(void) Tp_PutWidthOfPPin(entryPtr->width, newPPinPtr);
}


static void
CopyVExtToExt(zonePlacePtr)
ZoneData      *zonePlacePtr;
{
	BearId        vExtId;
	LinkedPinInst *pinInstPtr;
	LinkedRPin    *rPinPtr;

	/*
	 * Update the ext from the vext of pins in this zone.
	 */
	pinInstPtr = zonePlacePtr->pinInstList;
	while (pinInstPtr != NIL(LinkedPinInst))
	{
		(void) Mac_PutExtNetIdOfPin(
		  Mac_GetVExtNetIdOfPin(pinInstPtr->pinInstPtr),
		  pinInstPtr->pinInstPtr);
		pinInstPtr = pinInstPtr->nextPtr; 
	}
	rPinPtr = zonePlacePtr->rPinList;
	while (rPinPtr != NIL(LinkedRPin))
	{
	    if ((vExtId = Mac_GetVExtNetIdOfRPin(rPinPtr->rPinPtr)) != 0)
	    {
		(void) Mac_PutExtNetIdOfRPin(vExtId, rPinPtr->rPinPtr);
	    }
	    rPinPtr = rPinPtr->nextPtr;
	}
}


static void
PropagateNewId(entryPtr, newId, zonePlacePtr)
Table         *entryPtr;
BearId        newId;
ZoneData      *zonePlacePtr;
{
    BearId        intId;
    BearId        extId;
    BearId        vExtId;
    LinkedPPin    *pPinPtr;
    LinkedRPin    *rPinPtr;
    LinkedPinInst *pinInstPtr;
    float          width;
    float          *widthVect;

    /* 
     * Update the id of pins in this zone that are on the
     * current net.
     */
    pinInstPtr = zonePlacePtr->pinInstList;
    while (pinInstPtr != NIL(LinkedPinInst))
    {
	vExtId = Mac_GetVExtNetIdOfPin(pinInstPtr->pinInstPtr);
	extId = Mac_GetExtNetIdOfPin(pinInstPtr->pinInstPtr);
	if ((extId == entryPtr->id) || (vExtId == entryPtr->id))
	{
	    (void) Mac_PutExtNetIdOfPin(newId, pinInstPtr->pinInstPtr);
	    (void) Mac_PutHExtNetIdOfPin(newId, pinInstPtr->pinInstPtr);
	    widthVect = Mac_GetWidthVectOfPin(pinInstPtr->pinInstPtr);
	    width = widthVect[METAL2_LAYER];
	    if (width > entryPtr->width)
	    {
		entryPtr->width = width;
	    }
	}
	pinInstPtr = pinInstPtr->nextPtr;
    }
    rPinPtr = zonePlacePtr->rPinList;
    while (rPinPtr != NIL(LinkedRPin))
    {
	vExtId = Mac_GetVExtNetIdOfRPin(rPinPtr->rPinPtr);
	extId = Mac_GetExtNetIdOfRPin(rPinPtr->rPinPtr);
	if ((extId == entryPtr->id) || (vExtId == entryPtr->id))
	{
	    (void) Mac_PutExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
	    (void) Mac_PutHExtNetIdOfRPin(newId, rPinPtr->rPinPtr);
	    widthVect = Mac_GetWidthVectOfRPin(rPinPtr->rPinPtr);
	    width = widthVect[METAL2_LAYER];
	    if (width > entryPtr->width)
	    {
		entryPtr->width = width;
	    }
	}
	rPinPtr = rPinPtr->nextPtr;
    }
    /* 
     * Update the id of pseudo pins in this zone that are on the
     * current net.
     */
    pPinPtr = zonePlacePtr->pPinList;
    while (pPinPtr != NIL(LinkedPPin))
    {
	intId = Tp_GetIntNetIdOfPPin(pPinPtr->ppinPtr);
	extId = Tp_GetExtNetIdOfPPin(pPinPtr->ppinPtr);
	if (zonePlacePtr->zoneType == JUNCTION)
	{
	  if (extId == entryPtr->id)
	  {
		(void) Tp_PutExtNetIdOfPPin(newId, pPinPtr->ppinPtr);
		(void) Tp_PutNetOfPPin(entryPtr->netPtr, pPinPtr->ppinPtr);
	  }
	  if ((intId == entryPtr->id) &&
	      (pPinPtr->ppinPtr->tilePtr != NIL(Tile)) &&
	      (Tp_IsSBNTile(pPinPtr->ppinPtr->tilePtr)))
	  {
	    (void) Tp_PutIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
	    (void) Tp_PutNetOfPPin(entryPtr->netPtr, pPinPtr->ppinPtr);
	  }
	}
	else
	{
	  if (intId == entryPtr->id)
	  {
	    (void) Tp_PutIntNetIdOfPPin(newId, pPinPtr->ppinPtr);
	    (void) Tp_PutNetOfPPin(entryPtr->netPtr, pPinPtr->ppinPtr);
	  }
          else if ((extId == entryPtr->id) &&
		   (pPinPtr->ppinPtr->tilePtr != NIL(Tile)) &&
		   (Tp_IsDCellTile(pPinPtr->ppinPtr->tilePtr)))
	  {
	    (void) Tp_PutExtNetIdOfPPin(newId, pPinPtr->ppinPtr);
	    (void) Tp_PutNetOfPPin(entryPtr->netPtr, pPinPtr->ppinPtr);
	  }
	}
	pPinPtr = pPinPtr->nextPtr;
    }
}

