
/* tpIdentify.c --
 *
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpIdentify.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"

typedef struct _IdData {
    BNLine      *bnLineList;
    Plane       *planePtr;
    PlaneType   planeType;
} IdData;

/* ---------- Local Procedures ---------- */
int         IdSpaceTile(), IdSBNSpaceTile();
int         IncTopCount(), IncBottomCount();
int         SBNTileSearch();
int         GetSBNType();
int         AttachSpacePropToTile();
void        FillInSpaceTiles();

/* ---------- Local Global Variables ---------- */
static Tile     *topTilePtr, *botTilePtr;
static int      startTileBot, startTileTop;


/*
 * ----------------------------------------------------------------------------
 *
 * Tp_IdentifySpaceTiles() --
 *
 * First the memory associated the old floor plan graph is freed. 
 * Then the type of each space tile on both the horizontal and
 * vertical tile planes is determined, and new bottleneck line
 * lists are generated.
 *
 * Results:
 *     None.
 *
 * Side effects:
 *     None.
 * ----------------------------------------------------------------------------
 */

void
Tp_IdentifySpaceTiles(planeDataPtr)
PlaneData *planeDataPtr;
{
    IdData idData;

    if (Fpg_IsFpgBuilt(planeDataPtr)) {
/*
 *      (void) Tp_ItrHorzBnLine(planeDataPtr, ResetWallSegmentOfBnLine,
 *          (ClientData) NULL);
 *      (void) Tp_ItrVertBnLine(planeDataPtr, ResetWallSegmentOfBnLine,
 *          (ClientData) NULL);
 */

        /*
         * Free the memory associated with the floor plan graph.
         */
        Fpg_DeleteFloorPlanGraph(planeDataPtr);
    }

    FillInSpaceTiles(planeDataPtr);

    /*
     * Go through the horizontal tile plane and identify the following
     * tiles:
     *   TP_HORZ_BN_TILE, TP_HORZ_DOM_TILE, and TP_HORZ_GEN_TILE
     */
    idData.planePtr = Tp_GetHorzPlane(planeDataPtr);
    idData.planeType = HORZ_TILE_PLANE;
    idData.bnLineList = NIL(BNLine);
    (void) TiSrArea(NIL(Tile), idData.planePtr, &TiPlaneRect, IdSpaceTile,
        (ClientData) &idData);

    /*
     * Go through the horizontal tile plane and identify the following
     * tiles:
     *   TP_HORZ_DOM_SBN_TILE and TP_HORZ_GEN_SBN_TILE
     */
    (void) TiSrArea(NIL(Tile), idData.planePtr, &TiPlaneRect, IdSBNSpaceTile,
        (ClientData) &idData);
    Tp_PutHorzBnLineList(planeDataPtr, idData.bnLineList);

    /*
     * Go through the vertical tile plane and identify the following
     * tiles:
     *   TP_VERT_BN_TILE, TP_VERT_DOM_TILE, and TP_VERT_GEN_TILE
     */
    idData.planePtr = Tp_GetVertPlane(planeDataPtr);
    idData.planeType = VERT_TILE_PLANE;
    idData.bnLineList = NIL(BNLine);
    (void) TiSrArea(NIL(Tile), idData.planePtr, &TiPlaneRect, IdSpaceTile,
        (ClientData) &idData);

    /*
     * Go through the vertical tile plane and identify the following
     * tiles:
     *   TP_VERT_DOM_SBN_TILE and TP_VERT_GEN_SBN_TILE
     */
    (void) TiSrArea(NIL(Tile), idData.planePtr, &TiPlaneRect, IdSBNSpaceTile,
        (ClientData) &idData);
    Tp_PutVertBnLineList(planeDataPtr, idData.bnLineList);

    Bear_SetOption(planeDataPtr->pd_status, TP_SPACE_TILES_READY);
}



static int
IdSBNSpaceTile(startTilePtr, idDataPtr)
register Tile   *startTilePtr;
IdData          *idDataPtr;
/*
 * Identify the semi-bottleneck space tiles.  Note that 'IdSpaceTile()' must
 * be called first.
 */
{
    BNLineType  bnLineType;
    BNLine      *bnLinePtr;                 /* not used */
    Tile        *tlTilePtr, *brTilePtr;     /* These variables are needed
                                             * because Tp_IsPairOfSBNTiles()
                                             * may switch them.
                                             */
    Rect        spaceTileSearchArea;

    if (LB(startTilePtr) == idDataPtr->planePtr->pl_bottom) {
        return (0); /* continue search */
    }
    if (Tp_IsDominantTile(startTilePtr) || Tp_IsGenericTile(startTilePtr)) {
        TiToRect(startTilePtr, &spaceTileSearchArea);

        startTileBot = spaceTileSearchArea.r_xbot;
        startTileTop = spaceTileSearchArea.r_xtop;
        if (BL(startTilePtr) != idDataPtr->planePtr->pl_left) {
            spaceTileSearchArea.r_xbot -= 1;
        }
        if (TR(startTilePtr) != idDataPtr->planePtr->pl_right) {
            spaceTileSearchArea.r_xtop += 1;
        }
        spaceTileSearchArea.r_ytop = spaceTileSearchArea.r_ybot;
        spaceTileSearchArea.r_ybot -= 1;
        topTilePtr = botTilePtr = NIL(Tile);
        (void) TiSrArea(NIL(Tile), idDataPtr->planePtr, &spaceTileSearchArea,
            SBNTileSearch, (ClientData) NULL);

        if (botTilePtr != NIL(Tile)) {
            tlTilePtr = startTilePtr;
            brTilePtr = botTilePtr;
            if (Tp_IsPairOfSBNTiles(&tlTilePtr, &brTilePtr, &bnLineType)) {
                Tp_PutTypeOfSpaceTile(GetSBNType(tlTilePtr), tlTilePtr);
                Tp_PutTypeOfSpaceTile(GetSBNType(brTilePtr), brTilePtr);
                (void) Tp_CreateBnLine(tlTilePtr, brTilePtr, bnLineType, 
                    &idDataPtr->bnLineList, &bnLinePtr);
            }
        }
        if (topTilePtr != NIL(Tile)) {
            tlTilePtr = topTilePtr;
            brTilePtr = startTilePtr;
            if (Tp_IsPairOfSBNTiles(&tlTilePtr, &brTilePtr, &bnLineType)) {
                Tp_PutTypeOfSpaceTile(GetSBNType(tlTilePtr), tlTilePtr);
                Tp_PutTypeOfSpaceTile(GetSBNType(brTilePtr), brTilePtr);
                (void) Tp_CreateBnLine(tlTilePtr, brTilePtr, bnLineType,
                    &idDataPtr->bnLineList, &bnLinePtr);
            }
        }
    }
    return (0); /* continue search */
}



/* ARGSUSED */
static int
SBNTileSearch(tilePtr, clientData)
register Tile   *tilePtr;
ClientData      clientData;
{
    register int bot, top;

    if (Tp_IsGenericTile(tilePtr) || Tp_IsDominantTile(tilePtr)) {
        bot = LEFT(tilePtr);
        top = RIGHT(tilePtr);
        if (bot < startTileBot && top > startTileBot) {
            botTilePtr = tilePtr;
        }
        if (bot < startTileTop && top > startTileTop) {
            topTilePtr = tilePtr;
        }
    }
    return (0); /* continue search */
}



static int
GetSBNType(tilePtr)
Tile *tilePtr;
{
    switch (Tp_GetTypeOfSpaceTile(tilePtr)) {
    case TP_HORZ_DOM_TILE:
    case TP_HORZ_DOM_SBN_TILE:
        return (TP_HORZ_DOM_SBN_TILE);

    case TP_VERT_DOM_TILE:
    case TP_VERT_DOM_SBN_TILE:
        return (TP_VERT_DOM_SBN_TILE);

    case TP_HORZ_GEN_TILE:
    case TP_HORZ_GEN_SBN_TILE:
        return (TP_HORZ_GEN_SBN_TILE);

    case TP_VERT_GEN_TILE:
    case TP_VERT_GEN_SBN_TILE:
        return (TP_VERT_GEN_SBN_TILE);
    }
    /* NOTREACHED */
}



static int
IdSpaceTile(startTilePtr, idDataPtr)
register Tile   *startTilePtr;
IdData          *idDataPtr;
/*
 * Identify each space tile as bottleneck, dominant, you be illin or
 * generic.  Note that the class semi-bottleneck is not determined
 * by this function.
 */
{
    Rect    rect, spaceTileSearchArea;
    int     bottomEdgeCount, topEdgeCount;

    if (Tp_IsSpaceTile(startTilePtr)) {
        TiToRect(startTilePtr, &rect);

        topEdgeCount = 0;
        if (RT(startTilePtr) != idDataPtr->planePtr->pl_top) {
            startTileTop = TOP(startTilePtr);
            Mac_CopyRect(&spaceTileSearchArea, &rect);
            spaceTileSearchArea.r_ytop += 10;
            (void) TiSrArea(NIL(Tile), idDataPtr->planePtr,
                &spaceTileSearchArea, IncTopCount, (ClientData) &topEdgeCount);
        }

        bottomEdgeCount = 0;
        if (LB(startTilePtr) != idDataPtr->planePtr->pl_bottom) {
            startTileBot = BOTTOM(startTilePtr);
            Mac_CopyRect(&spaceTileSearchArea, &rect);
            spaceTileSearchArea.r_ybot -= 10;
            (void) TiSrArea(NIL(Tile), idDataPtr->planePtr,
                &spaceTileSearchArea, IncBottomCount,
                (ClientData) &bottomEdgeCount);
        }

        /*
         * Note that the topEdgeCount (bottomEdgeCount) can be zero if
         * the the tile above (below) 'startTilePtr' is a psuedo tile.
         */
        if (((topEdgeCount == 0) ||
            (topEdgeCount == 1 && Tp_IsSpaceTile(RT(startTilePtr)))) &&

            ((bottomEdgeCount == 0) ||
            (bottomEdgeCount == 1 && Tp_IsSpaceTile(LB(startTilePtr))))) {
            /* bottleneck tile */

            Tp_PutTypeOfSpaceTile(((idDataPtr->planeType == HORZ_TILE_PLANE) ?
                TP_HORZ_BN_TILE : TP_VERT_BN_TILE), startTilePtr);
        } else if (((topEdgeCount == 0) || (topEdgeCount > 1)) &&
            ((bottomEdgeCount == 0) || (bottomEdgeCount > 1))) {
            /* dominant tile */

            Tp_PutTypeOfSpaceTile(((idDataPtr->planeType == HORZ_TILE_PLANE) ?
                TP_HORZ_DOM_TILE : TP_VERT_DOM_TILE), startTilePtr);
        } else {
            /* generic tile */

            Tp_PutTypeOfSpaceTile(((idDataPtr->planeType == HORZ_TILE_PLANE) ?
                TP_HORZ_GEN_TILE : TP_VERT_GEN_TILE), startTilePtr);
        }

    }
    return (0); /* continue search */
}



static int
IncBottomCount(tilePtr, countPtr)
register Tile   *tilePtr;
register int    *countPtr;
{
    /*
     * Only those tiles directly below 'startTilePtr' will be counted.
     */
    if (TOP(tilePtr) == startTileBot) {
        *countPtr += 1; 
    }
    return (0); /* continue search */
}



static int
IncTopCount(tilePtr, countPtr)
register Tile   *tilePtr;
register int    *countPtr;
{
    /*
     * Only those tiles directly above 'startTilePtr' will be counted.
     */
    if (BOTTOM(tilePtr) == startTileTop) {
        *countPtr += 1; 
    }
    return (0); /* continue search */
}



static void
FillInSpaceTiles(planeDataPtr)
PlaneData *planeDataPtr;
{
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr), &TiPlaneRect,
        AttachSpacePropToTile, (ClientData) NULL);

    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr), &TiPlaneRect,
        AttachSpacePropToTile, (ClientData) NULL);
}



/* ARGSUSED */
static int
AttachSpacePropToTile(tilePtr, clientData)
Tile        *tilePtr;
ClientData  clientData;
{
/*
    if (Tp_IsSpaceTile(tilePtr) &&
        (Tp_IsTileFilled(tilePtr) == FALSE)) {

        Tp_CreateSpaceTileProp(tilePtr);
    }
*/
    if (Tp_IsSpaceTile(tilePtr)) {
        if (Tp_IsTileFilled(tilePtr)) {
            TILE_PROP_PTR->unValue.spacePropPtr->segPtr = NIL(WallSegment);
        } 
        else {
            Tp_CreateSpaceTileProp(tilePtr);
        }
    }
    return (0); /* continue search */
}



