
/* tpInsert.c --
 *
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpInsert.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "rectSlice.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"

/*
 * ----------------------------------------------------------------------------
 *
 * Tp_InsertCellMaster() --
 *
 * Insert a cell master.
 *
 * Results:
 *    BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *    The cell is painted onto the horizontal and vertical
 *    tile planes and entered into the data base.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_InsertCellMaster(parentCellPtr, namePtr, originPtr, pointList)
CellMaster      *parentCellPtr;
BearString      namePtr;
Point           *originPtr;
LinkedPoint     *pointList;         /* List of points describing a 
                                     * "manhattan polygon".
                                     */
{
    CellInst        *cellInstPtr;
    CheckData       checkData;
    GenericCell     gcell;
    LinkedLineSeg   *boundaryPtr;
    LinkedRect      *horzRectList = NIL(LinkedRect);
    LinkedRect      *vertRectList = NIL(LinkedRect);
    PlaneData       *planeDataPtr;
    Rect            bbox;           /* cell master's bounding box           */
    Transform       horzTrans;      /* horizontal transform                 */
    char            masterName[80]; /* cell master's name                   */

    if (Rs_HorzRectSlice(pointList, &horzRectList) == BEAR_ERROR) {
        goto putMemBack;
    }
    GeoTranslateTrans(&GeoIdentityTransform,
        originPtr->p_x, originPtr->p_y, &horzTrans);

    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    checkData.strong = TRUE;
    checkData.tileType = TT_SPACE;
    checkData.namePtr = NIL(char);
    checkData.planePtr = Tp_GetHorzPlane(planeDataPtr);
    checkData.bndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    checkData.rectList = horzRectList;
    checkData.transPtr = &horzTrans;
    if (TpCheckPlaneUnderRectList(&checkData)) {

        sprintf(masterName, "%s_master", namePtr);

        if (Mac_CreateCell(parentCellPtr, 
            masterName, namePtr, &cellInstPtr) == BEAR_OK) {
            gcell.unType = CELL_INST;
            gcell.unValue.cellInstPtr = cellInstPtr;
        } else {
            goto putMemBack;
        }

        if (Rs_VertRectSlice(pointList, &vertRectList) == BEAR_ERROR) {
            goto putMemBack;
        }
        Mac_PutHorzRectListOfGCell(horzRectList, &gcell);
        Mac_PutVertRectListOfGCell(vertRectList, &gcell);
        Mac_PutTransformOfGCell(&horzTrans, &gcell);
        if (Mac_FindBBoxForPointList(pointList, &bbox) == BEAR_ERROR) {
            goto putMemBack;
        } else {
            Mac_PutBBoxOfCell(&bbox, cellInstPtr);
        }
        if (Mac_MakeLineSegList(pointList, &boundaryPtr) == BEAR_ERROR) {
            goto putMemBack;
        } else {
            (void) Mac_AssignDirOfLineSegments(boundaryPtr);
            Mac_PutBoundaryOfCell(boundaryPtr, cellInstPtr);
            (void) Mac_MoveBoundaryOfCell(&horzTrans, cellInstPtr);
            Mac_PutOriginOfCell(originPtr, cellInstPtr);
        }

        (void) Tp_PaintGCell(&gcell, planeDataPtr);

        return (BEAR_OK);
    } else {
        goto putMemBack;
    }

putMemBack:
    /* The cell could not be inserted, so the memory we got is freed up. */
    Mac_FreeRectList(horzRectList);
    Mac_FreeRectList(vertRectList);

    return (BEAR_ERROR);
}


