
/* tpPPin.c --
 *
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 * This file contains the subroutines to create, insert, and delete
 * pseudo pins.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpPPin.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"


typedef struct _SearchData {
    PseudoPin   *ppinPtr;
    Line        *linePtr;
} SearchData;


/* ---------- Local Procedures ---------- */
static PseudoPin    *MakePPin();
static void         AddPPinToTile();
static BearStatus   GetBnLineInfo();
static BearStatus   BnLineSearch();
static BearStatus   IsSamePPin();


static BearStatus
IsSamePPin(ppinPtr, targetPPinPtr)
PseudoPin   *ppinPtr;
PseudoPin   *targetPPinPtr;
{
    if (ppinPtr == targetPPinPtr) {
        return (BEAR_STOP_SEARCH);
    }
    return (BEAR_CONTINUE_SEARCH);
}


static BearStatus
BnLineSearch(bnLinePtr, searchDataPtr)
BNLine      *bnLinePtr;
SearchData  *searchDataPtr;
{
    if (Tp_ItrPPinsOfBnLine(Tp_GetTLTileOfBnLine(bnLinePtr),
        Tp_GetBRTileOfBnLine(bnLinePtr), IsSamePPin,
        (ClientData) searchDataPtr->ppinPtr) == BEAR_STOP_SEARCH) {

        searchDataPtr->linePtr = Tp_GetLineOfBnLine(bnLinePtr);
        return (BEAR_STOP_SEARCH);
    }

    return (BEAR_CONTINUE_SEARCH);
}


static void
AddPPinToTile(ppinPtr, tilePtr, side, fragment)
PseudoPin       *ppinPtr;
Tile            *tilePtr;
SideDirection   side;
int             fragment;
{
    if (side == (SideDirection) BOTTOM_LEFT_SIDE) {
        ppinPtr->nextPtr = 
            *(Tp_GetSpacePropOfTile(tilePtr)->blPPinList[fragment]);
        *(Tp_GetSpacePropOfTile(tilePtr)->blPPinList[fragment]) = ppinPtr;
    }
    else if (side == (SideDirection) TOP_RIGHT_SIDE) {
        ppinPtr->nextPtr =
            *(Tp_GetSpacePropOfTile(tilePtr)->trPPinList[fragment]);
        *(Tp_GetSpacePropOfTile(tilePtr)->trPPinList[fragment]) = ppinPtr;
    }
    else {
        TpPrintFatalErr("AddPPinToTile() failed : illegal side value");
    }
}



static BearStatus
GetBnLineInfo(bnLinePtr, tilePtrPtr, sidePtr, fragmentPtr)
BNLine          *bnLinePtr;
Tile            **tilePtrPtr;
SideDirection   *sidePtr;
int             *fragmentPtr;
{
#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) {
        TpPrintFatalErr("GetBnLineInfo() failed: no bn line is given");
    }
#endif DEBUG

    switch (Tp_GetBnLineTypeOfBnLine(bnLinePtr)) {
    case TP_NULL_BNL:
        return (BEAR_ERROR);

    case TP_VERT_R_UP:
        /* 
         * vertical plane : left tile : right side : top fragment
         */
        *sidePtr = TOP_RIGHT_SIDE;
        *fragmentPtr = TP_TOP;
        break;

    case TP_VERT_R_DOWN:
        /* 
         * vertical plane : left tile : right side : bottom fragment
         */
        *sidePtr = TOP_RIGHT_SIDE;
        *fragmentPtr = TP_BOTTOM;
        break;

    case TP_HORZ_B_RIGHT:
        /* 
         * horizontal plane : top tile : bottom side : top fragment
         */
        *sidePtr = BOTTOM_LEFT_SIDE;
        *fragmentPtr = TP_TOP;
        break;

    case TP_HORZ_B_LEFT:
        /* 
         * horizontal plane : top tile : bottom side : bottom fragment
         */
        *sidePtr = BOTTOM_LEFT_SIDE;
        *fragmentPtr = TP_BOTTOM;
        break;

    default:
        TpPrintFatalErr("GetBnLineInfo() failed : illegal bn line type");
        break;
    }
    *tilePtrPtr = Tp_GetTLTileOfBnLine(bnLinePtr);
    return (BEAR_OK);
}



/* 
 * ----------------------------------------------------------------------------
 *
 * MakePPin() --
 *
 *    Allocates memory for a new pseudo pin, and initializes
 *    the fields.
 *
 * Results:
 *    A pointer to a pseudo pin is returned.
 *
 * Side effects:
 *    None.
 *
 * ----------------------------------------------------------------------------
 */

static PseudoPin *
MakePPin()
{
    register PseudoPin *ppinPtr;

    MALLOC(PseudoPin *, ppinPtr, sizeof (PseudoPin));
    ppinPtr->id = Mac_GenerateId();
    ppinPtr->netPtr = NIL(Net);
    ppinPtr->intNetId = 0;
    ppinPtr->extNetId = 0;
    ppinPtr->width = MAXWIDTH;
    ppinPtr->order = 0;
    ppinPtr->tilePtr = NIL(Tile);
    ppinPtr->nextPtr = NIL(PseudoPin);
    return (ppinPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_CreatePPin() --
 *
 * Loads the 'side' and 'bnTilePtr' fields of the pseudo pin structure
 * with the given information.  Finally the pseudo
 * pin is inserted into the appropiate list within the bottleneck tile.
 *
 * Results:
 *   BEAR_OK is returned if everything is peachy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_CreatePPin(tilePtr, side, fragment, ppinPtrPtr)
Tile            *tilePtr;
SideDirection   side;
int             fragment;
PseudoPin       **ppinPtrPtr;
{
    register PseudoPin *ppinPtr;

#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_CreatePPin() failed: no tile is given");
    } else if ((fragment != TP_BOTTOM) && (fragment != TP_TOP)) {
        TpPrintFatalErr("Tp_CreatePPin() failed : fragment has illegal value");
    }
#endif DEBUG

    ppinPtr = MakePPin();
    ppinPtr->tilePtr = tilePtr;
    ppinPtr->side = side;

    AddPPinToTile(ppinPtr, tilePtr, side, fragment);

    *ppinPtrPtr = ppinPtr;

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_CreatePPinBnLine() --
 *
 * Loads the 'side' and 'bnTilePtr' fields of the pseudo pin structure
 * based on the given information.  Then the pseudo pin is inserted into
 * the appropiate list.
 *
 * Results:
 *   BEAR_OK is returned if everything is peachy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_CreatePPinOfBnLine(bnLinePtr, ppinPtrPtr)
BNLine      *bnLinePtr;
PseudoPin   **ppinPtrPtr;
{
    SideDirection       side;
    Tile                *tilePtr;
    int                 fragment;
    register PseudoPin  *ppinPtr;

#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) {
        TpPrintFatalErr("Tp_CreatePPinOfBnLine() failed: no bnLine given");
    }
#endif DEBUG

    if (GetBnLineInfo(bnLinePtr, &tilePtr, &side, &fragment) == BEAR_ERROR) {
        return (BEAR_ERROR);
    }

    ppinPtr = MakePPin();
    ppinPtr->tilePtr = tilePtr;
    ppinPtr->side = side;

    AddPPinToTile(ppinPtr, tilePtr, side, fragment);

    *ppinPtrPtr = ppinPtr;

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_CreatePPinBnTile() --
 *
 * Loads the 'side' and 'tilePtr' fields of the pseudo pin structure
 * based on the given information.  Then the pseudo pin is inserted
 * into a list within the given bottleneck tile.  The list is determined
 * using the side thats given and the TP_BOTTOM fragment.
 *
 * Results:
 *   BEAR_OK is returned if everything is peachy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_CreatePPinOfBnTile(bnTilePtr, side, ppinPtrPtr)
Tile            *bnTilePtr;
SideDirection   side;
PseudoPin       **ppinPtrPtr;
{
    register PseudoPin *ppinPtr;

#ifdef DEBUG
    if (bnTilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_CreatePPinOfBnTile() failed : no tile given");
    } else if (Tp_IsBottleneckTile(bnTilePtr) == FALSE) {
        TpPrintFatalErr("Tp_CreatePPinOfBnTile() failed : not a bn tile");
    }
#endif DEBUG

    ppinPtr = MakePPin();
    ppinPtr->side = side;
    ppinPtr->tilePtr = bnTilePtr;

    AddPPinToTile(ppinPtr, bnTilePtr, side, TP_BOTTOM);

    *ppinPtrPtr = ppinPtr;

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_CreatePPinOfDCell() --
 *
 * The pseudo pin is inserted into the specified side of the 
 * dummy cell.
 *
 * Results:
 *   BEAR_OK is returned if everything is peachy, BEAR_ERROR otherwise.
 *
 * Side effects:
 *   pseudo pin list of the dummy cell is updated.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_CreatePPinOfDCell(dCellPtr, side, ppinPtrPtr)
DummyCell       *dCellPtr;
SideDirection   side;
PseudoPin       **ppinPtrPtr;
{
    register PseudoPin *ppinPtr;

#ifdef DEBUG
    if (dCellPtr == NIL(DummyCell)) 
    {
    TpPrintFatalErr("Tp_CreatePPinOfDCell() failed: no dummy cell is given");
    } 
    else if ((side != (SideDirection) LEFT_SIDE) &&
        (side != (SideDirection) TOP_SIDE) &&
        (side != (SideDirection) RIGHT_SIDE) &&
        (side != (SideDirection) BOTTOM_SIDE)) 
    {
      TpPrintFatalErr("Tp_CreatePPinOfDCell() failed : side has illegal value");
    }
#endif DEBUG

    ppinPtr = MakePPin();
    ppinPtr->tilePtr = Mac_GetTileOfDCell(dCellPtr);
    ppinPtr->side = side;

    (void) Mac_InsertPPinOfDCell(dCellPtr, side, ppinPtr);
    *ppinPtrPtr = ppinPtr;

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_InsertPPinOfBnTile() --
 *
 * Inserts a ppin to the given bn tile on the designated side.
 * Notice that it's always inserted at the bottom fragment of a side.
 *
 * Results:
 *   BEAR_OK, when done.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_InsertPPinOfBnTile(bnTilePtr, side, ppinPtr)
Tile            *bnTilePtr;
SideDirection   side;
PseudoPin       *ppinPtr;
{
    PseudoPin   *ppinList;

#ifdef DEBUG
    if (bnTilePtr == NIL(Tile) || ppinPtr== NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_InsertPPinOfBnTile() failed");
    }
    else if (Tp_IsBottleneckTile(bnTilePtr) == FALSE)
    {
        TpPrintFatalErr("Tp_InsertPPinOfBnTile() failed: bnTilePtr is not a bottleneck tile");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) {
        TpPrintFatalErr("Tp_InsertPPinOfBnTile() failed : side has illegal value");
    }
#endif DEBUG

    ppinList = ((side == (SideDirection) BOTTOM_LEFT_SIDE) ?
        *((Tp_GetSpacePropOfTile(bnTilePtr))->blPPinList[TP_BOTTOM]) :
        *((Tp_GetSpacePropOfTile(bnTilePtr))->trPPinList[TP_BOTTOM]));
    ppinPtr->nextPtr = ppinList;
    ppinPtr->tilePtr = bnTilePtr;

    if (side == (SideDirection) BOTTOM_LEFT_SIDE) {
        *((Tp_GetSpacePropOfTile(bnTilePtr))->blPPinList[TP_BOTTOM]) = ppinPtr; 
    }
    else {
        *((Tp_GetSpacePropOfTile(bnTilePtr))->trPPinList[TP_BOTTOM]) = ppinPtr;
    }
    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Tp_DeletePPinOfBnTile() --
 *
 * Deletes the given ppin from the designated side of a bn tile.
 * Assume that the ppin is always located in the bottom fragment of the side
 *
 * Results:
 *   BEAR_OK, when done.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_DeletePPinOfBnTile(bnTilePtr, side, ppinPtr)
Tile            *bnTilePtr;
SideDirection   side;
PseudoPin       *ppinPtr;
{
    PseudoPin           *ppinList;
    register PseudoPin  *curPPinPtr;
    
#ifdef DEBUG
    if (bnTilePtr == NIL(Tile) || ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_DeletePPinOfBnTile() failed");
    }
    else if (Tp_IsBottleneckTile(bnTilePtr) == FALSE) {
        TpPrintFatalErr("Tp_DeletePPinOfBnTile() failed: bnTilePtr is not a bottleneck tile");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) {
        TpPrintFatalErr("Tp_DeletePPinOfBnTile() failed : side has illegal value");
    }
#endif DEBUG

    if (ppinPtr->tilePtr == NIL(Tile)) {
        Bear_PrintErr("Tp_DeletePPinOfBnTile() failed: the pseudo pin is not connected to a bottleneck tile");
        
        return (BEAR_ERROR);
    }

    /* 
     * Look for the pseudo pin in the bottom fragment list.
     */
    ppinList = (side == (SideDirection) BOTTOM_LEFT_SIDE) ?
        *((Tp_GetSpacePropOfTile(bnTilePtr))->blPPinList[TP_BOTTOM]) :
        *((Tp_GetSpacePropOfTile(bnTilePtr))->trPPinList[TP_BOTTOM]);
    if (ppinList != NIL(PseudoPin)) {
        if (ppinPtr == ppinList) {
            if (side == (SideDirection) BOTTOM_LEFT_SIDE) {
                *((Tp_GetSpacePropOfTile(bnTilePtr))->blPPinList[TP_BOTTOM]) =
                    ppinPtr->nextPtr;
            }
            else {
                *((Tp_GetSpacePropOfTile(bnTilePtr))->trPPinList[TP_BOTTOM]) = 
                    ppinPtr->nextPtr;
            }
        } 
        else {
            curPPinPtr = ppinList;
            while ((curPPinPtr->nextPtr != ppinPtr) && 
                (curPPinPtr->nextPtr != NIL(PseudoPin))) {
                curPPinPtr = curPPinPtr->nextPtr;
            }
            if (curPPinPtr->nextPtr != ppinPtr) {
                /* The psuedo pin is not in the list. */
                goto part2;
            } 
            else {
                curPPinPtr->nextPtr = ppinPtr->nextPtr;
            }
        }
        FREE((char *) ppinPtr);
    }

part2:
    /* 
     * Look for the pseudo pin in the top fragment list.
     */
    ppinList = (side == (SideDirection) BOTTOM_LEFT_SIDE) ?
        *((Tp_GetSpacePropOfTile(bnTilePtr))->blPPinList[TP_TOP]) :
        *((Tp_GetSpacePropOfTile(bnTilePtr))->trPPinList[TP_TOP]);
    if (ppinList != NIL(PseudoPin)) {
        if (ppinPtr == ppinList) {
            if (side == (SideDirection) BOTTOM_LEFT_SIDE) {
                *((Tp_GetSpacePropOfTile(bnTilePtr))->blPPinList[TP_TOP]) =
                    ppinPtr->nextPtr;
            }
            else {
                *((Tp_GetSpacePropOfTile(bnTilePtr))->trPPinList[TP_TOP]) = 
                    ppinPtr->nextPtr;
            }
        } 
        else {
            curPPinPtr = ppinList;
            while ((curPPinPtr->nextPtr != ppinPtr) && 
                (curPPinPtr->nextPtr != NIL(PseudoPin))) {
                curPPinPtr = curPPinPtr->nextPtr;
            }
            if (curPPinPtr->nextPtr != ppinPtr) {
                /* The psuedo pin is not in the list. */
                return (BEAR_OK);
            } 
            else {
                curPPinPtr->nextPtr = ppinPtr->nextPtr;
            }
        }
        FREE((char *) ppinPtr);
    }

    return (BEAR_OK);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_InsertPPinOfBnLine() --
 *
 * Inserts the ppin onto the designated fragment of the given bn tile.
 * The newly inserted ppin will become head of the list of ppins.
 *
 * Results:
 *   BEAR_OK, when done. 
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_InsertPPinOfBnLine(sbnTilePtr, side, fragment, ppinPtr)
Tile          *sbnTilePtr;
SideDirection side;
int         fragment;
PseudoPin   *ppinPtr;

{
    PseudoPin   *ppinList;

#ifdef DEBUG
    if (sbnTilePtr == NIL(Tile) || ppinPtr == NIL(PseudoPin)) 
    {
        TpPrintFatalErr("Tp_InsertPPinOfBnLine() failed");
    } 
    else if (Tp_IsSBNTile(sbnTilePtr) == FALSE)
    {
        TpPrintFatalErr("Tp_InsertPPinOfBnLine() failed: sbnTilePtr is not a semi bottleneck tile");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) 
    {
        TpPrintFatalErr("Tp_InsertPPinOfBnLine() failed : side has illegal value");
    }
    else if ((fragment != TP_BOTTOM) && (fragment != TP_TOP))
    {
        TpPrintFatalErr("Tp_InsertPPinOfBnLine() failed : fragment has illegal value");
    }
#endif DEBUG

    ppinList = (side == (SideDirection) BOTTOM_LEFT_SIDE) ?
        (*((Tp_GetSpacePropOfTile(sbnTilePtr))->blPPinList[fragment]))->nextPtr :
        (*((Tp_GetSpacePropOfTile(sbnTilePtr))->trPPinList[fragment]))->nextPtr;
    ppinPtr->nextPtr = ppinList;
    if (side == (SideDirection) BOTTOM_LEFT_SIDE) 
    {
        (*((Tp_GetSpacePropOfTile(sbnTilePtr))->blPPinList[TP_BOTTOM]))->nextPtr
            = ppinPtr; 
    }
    else
    {
        (*((Tp_GetSpacePropOfTile(sbnTilePtr))->trPPinList[TP_BOTTOM]))->nextPtr
            = ppinPtr;
    }
    return (BEAR_OK);
}


/*
 * ----------------------------------------------------------------------------
 *
 * Tp_DeletePPinOfBnLine() --
 *
 * Dletes the given ppin from the specified fragment on the given 
 * bn tile.
 *
 * Results:
 *   BEAR_OK, when successful.
 *   BEAR_ERROR, otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_DeletePPinOfBnLine(sbnTilePtr, side, fragment, ppinPtr)
Tile            *sbnTilePtr;
SideDirection   side;
int             fragment;
PseudoPin       *ppinPtr;

{
    PseudoPin   *curPPinPtr, *ppinList;

#ifdef DEBUG
    if (sbnTilePtr == NIL(Tile) || ppinPtr == NIL(PseudoPin)) 
    {
        TpPrintFatalErr("Tp_DeletePPinOfBnLine() failed");
    } 
    else if (Tp_IsSBNTile(sbnTilePtr) == FALSE)
    {
        TpPrintFatalErr("Tp_DeletePPinOfBnLine() failed: sbnTilePtr is not a semi bottleneck tile");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) 
    {
        TpPrintFatalErr("Tp_DeletePPinOfBnLine() failed : side has illegal value");
    }
    else if ((fragment != TP_BOTTOM) && (fragment != TP_TOP))
    {
        TpPrintFatalErr("Tp_DeletePPinOfBnLine() failed : fragment has illegal value");
    }
#endif DEBUG

    if (ppinPtr->tilePtr == NIL(Tile)) 
    {
Bear_PrintErr("Tp_DeletePPin() failed: the pseudo pin is not connected to a bottleneck tile");
        return (BEAR_ERROR);
    }
    /* Get the head of the pseudo pin list. */
    ppinList = (side == (SideDirection) BOTTOM_LEFT_SIDE) ?
        (*((Tp_GetSpacePropOfTile(sbnTilePtr))->blPPinList[fragment]))->nextPtr :
        (*((Tp_GetSpacePropOfTile(sbnTilePtr))->trPPinList[fragment]))->nextPtr;
    if (ppinList == NIL(PseudoPin)) 
    {
        return (BEAR_OK);
    }

    if (ppinPtr == ppinList) 
    {
        ppinList = ppinPtr->nextPtr;
    } 
    else 
    {
        curPPinPtr = ppinList;
        while (curPPinPtr->nextPtr != ppinPtr && 
            curPPinPtr->nextPtr != NIL(PseudoPin)) 
        {
            curPPinPtr = curPPinPtr->nextPtr;
        }
        if (curPPinPtr->nextPtr != ppinPtr) 
        {
            /* The psuedo pin is not in the list. */
            return (BEAR_OK);
        } 
        else 
        {
            curPPinPtr->nextPtr = ppinPtr->nextPtr;
        }
    }
    FREE((char *) ppinPtr);
    return (BEAR_OK);
}


void
Tp_RemovePPinListOfTile(tilePtr, side, fragment, ppinListPtr)
Tile            *tilePtr;
SideDirection   side;
int             fragment;
PseudoPin       **ppinListPtr;
/*
 * Return the head of the pseudo pin list to the user and then
 * make the head of the list stored in the tile NULL.
 */
{
    SpaceProp       *spacePropPtr;

#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_RemovePPinList() failed: no tile given");
    } else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) {
        TpPrintFatalErr("Tp_RemovePPinList() failed : side has illegal value");
    } else if ((fragment != TP_BOTTOM) && (fragment != TP_TOP)) {
        TpPrintFatalErr(
            "Tp_RemovePPinList() failed : fragment has illegal value");
    }
#endif DEBUG

    spacePropPtr = Tp_GetSpacePropOfTile(tilePtr);
    if (side == (SideDirection) BOTTOM_LEFT_SIDE) 
    {
        *ppinListPtr = *(spacePropPtr->blPPinList[fragment]);
        *(spacePropPtr->blPPinList[fragment]) = NIL(PseudoPin);
    } 
    else 
    {
        *ppinListPtr = *(spacePropPtr->trPPinList[fragment]);
        *(spacePropPtr->trPPinList[fragment]) = NIL(PseudoPin);
    }
}



void
Tp_FreePPinsOfBnLine(bnLinePtr)
BNLine *bnLinePtr;
/*
 * Free all the pseudo pins of the given bottleneck line.
 */
{
    Tile            *tilePtr;
    SideDirection   side;
    int             fragment;

#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) 
    {
        TpPrintFatalErr("Tp_FreePPinsOfBnLine() failed: no bn. line given");
    }
#endif DEBUG

    if (GetBnLineInfo(bnLinePtr, &tilePtr, &side, &fragment) == BEAR_OK) {
        Tp_FreePPinListOfFragment(tilePtr, side, fragment);
    }
}



void
Tp_FreePPinsOfTile(tilePtr)
Tile *tilePtr;
/*
 * Free all the pseudo pins of a tile.
 */
{

#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_FreePPinListOfTile() failed: no tile given");
    }
    else if (Tp_IsSpacePropTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_FreePPinListOfTile() failed: not a space prop tile");
    }
#endif DEBUG

    Tp_FreePPinListOfSide(tilePtr, BOTTOM_LEFT_SIDE);
    Tp_FreePPinListOfSide(tilePtr, TOP_RIGHT_SIDE);
}



void
Tp_FreePPinListOfSide(tilePtr, side)
Tile            *tilePtr;
SideDirection   side;
/*
 * Free the pseudo pin list of the side of a tile.
 */
{

#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_FreePPinListOfTile() failed: no tile given");
    }
    else if (Tp_IsSpacePropTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_FreePPinListOfSide() failed: not a space prop tile");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) {
        TpPrintFatalErr("Tp_FreePPinListOfTile() failed : side has illegal value");
    } 
#endif DEBUG

    Tp_FreePPinListOfFragment(tilePtr, side, TP_BOTTOM);
    Tp_FreePPinListOfFragment(tilePtr, side, TP_TOP);
}



void
Tp_FreePPinListOfFragment(tilePtr, side, fragment)
Tile            *tilePtr;
SideDirection   side;
int             fragment;
/*
 * Free the pseudo pin list of the fragment of a side of a tile.
 */
{
    PseudoPin *ppinPtr, *ppinList;

#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_FreePPinListOfTile() failed: no tile given");
    }
    else if (Tp_IsSpacePropTile(tilePtr) == FALSE) {
   TpPrintFatalErr("Tp_FreePPinListOfFragment() failed: not a space prop tile");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) {
        TpPrintFatalErr("Tp_FreePPinListOfTile() failed : illegal side value");
    }
    else if ((fragment != TP_BOTTOM) && (fragment != TP_TOP)) {
        TpPrintFatalErr(
            "Tp_FreePPinListOfTile() failed : fragment has illegal value");
    }
#endif DEBUG

    if (side == (SideDirection) BOTTOM_LEFT_SIDE) {
        ppinList = *(Tp_GetSpacePropOfTile(tilePtr)->blPPinList[fragment]);
        *(Tp_GetSpacePropOfTile(tilePtr)->blPPinList[fragment]) =
            NIL(PseudoPin);
    }
    else {
        ppinList = *(Tp_GetSpacePropOfTile(tilePtr)->trPPinList[fragment]);
        *(Tp_GetSpacePropOfTile(tilePtr)->trPPinList[fragment]) =
            NIL(PseudoPin);
    }

    while ((ppinPtr = ppinList) != NIL(PseudoPin)) {
        ppinList = ppinPtr->nextPtr;
        FREE((char *) ppinPtr);
    }
}



void
Tp_AppendPPinListOfTile(ppinList, tilePtr, side, fragment)
PseudoPin       *ppinList;
Tile            *tilePtr;
SideDirection   side;
int             fragment;
/*
 * Append a pseudo pin list onto the given side and fragment of a tile.
 */
{
    PseudoPin *ppinPtr, **ppinPtrPtr;

#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_AppendPPinList() failed: no tile given");
    }
    else if ((side != (SideDirection) BOTTOM_LEFT_SIDE) &&
        (side != (SideDirection) TOP_RIGHT_SIDE)) {
        TpPrintFatalErr("Tp_AppendPPinList() failed : side has illegal value");
    }
    else if ((fragment != TP_BOTTOM) && (fragment != TP_TOP)) {
    TpPrintFatalErr("Tp_AppendPPinList() failed : fragment has illegal value");
    }
#endif DEBUG

    if (ppinList != NIL(PseudoPin)) 
    {
    ppinPtrPtr = (side == (SideDirection) BOTTOM_LEFT_SIDE) ?
        Tp_GetSpacePropOfTile(tilePtr)->blPPinList[fragment] :
        Tp_GetSpacePropOfTile(tilePtr)->trPPinList[fragment];

    if (*ppinPtrPtr == NIL(PseudoPin)) {
        *ppinPtrPtr = ppinList;
    } else {
        /*
         * Find the end of the pseudo pin list that is currently 
         * attached to the given side and fragment of the tile.
         */
         for (ppinPtr = *ppinPtrPtr;
        /* infinite loop */;
        ppinPtr = ppinPtr->nextPtr) {

        if (ppinPtr->nextPtr == NIL(PseudoPin)) {
            break;
        }
        }
        /* Append the new list onto the end of the old one.  */
        ppinPtr->nextPtr = ppinList;
    }

    for (ppinPtr = ppinList;ppinPtr != NIL(PseudoPin);
      ppinPtr = ppinPtr->nextPtr) 
    {
        ppinPtr->tilePtr = tilePtr;
    }
    }
}



BearStatus
Tp_GetAbsPosOfPPin(planeDataPtr, ppinPtr, absPosPtr)
PlaneData   *planeDataPtr;                      /* IN       */
PseudoPin   *ppinPtr;                           /* IN       */
Point       *absPosPtr;                         /* OUT      */
{
    DummyCell   *dcellPtr;
    Tile        *tilePtr;

    tilePtr = Tp_GetTileOfPPin(ppinPtr);
    if (Tp_IsDCellTile(tilePtr) == TRUE) {
        dcellPtr = Tp_GetDCellOfTile(tilePtr);

        switch (Tp_GetSideOfPPin(ppinPtr)) {
        case TOP_SIDE:
            absPosPtr->p_x = dcellPtr->bbox.r_xbot +
                (dcellPtr->bbox.r_xtop - dcellPtr->bbox.r_xbot) / 2;
            absPosPtr->p_y = dcellPtr->bbox.r_ytop;
            break;
        case BOTTOM_SIDE:
            absPosPtr->p_x = dcellPtr->bbox.r_xbot +
                (dcellPtr->bbox.r_xtop - dcellPtr->bbox.r_xbot) / 2;
            absPosPtr->p_y = dcellPtr->bbox.r_ybot;
            break;
        case LEFT_SIDE:
            absPosPtr->p_x = dcellPtr->bbox.r_xbot;
            absPosPtr->p_y = dcellPtr->bbox.r_ybot +
                (dcellPtr->bbox.r_ytop - dcellPtr->bbox.r_ybot) / 2;
            break;
        case RIGHT_SIDE:
            absPosPtr->p_x = dcellPtr->bbox.r_xtop;
            absPosPtr->p_y = dcellPtr->bbox.r_ybot +
                (dcellPtr->bbox.r_ytop - dcellPtr->bbox.r_ybot) / 2;
            break;
        default:
    TpPrintFatalErr("GreGetAbsPosOfPPin() : ppin on dcell has illegal side value");
            break;
        }

        return (BEAR_OK);
    }

    if (Tp_IsBottleneckTile(tilePtr) == TRUE) {
        Rect            rect;
        SideDirection   side;

        side = Tp_GetSideOfPPin(ppinPtr);
        if (Tp_IsVertTile(tilePtr) == TRUE) {
            Tp_VertTileToRect(planeDataPtr, tilePtr, &rect);
            absPosPtr->p_x = (side == BOTTOM_LEFT_SIDE) ?
                rect.r_xbot : rect.r_xtop;
            absPosPtr->p_y = rect.r_ybot + (rect.r_ytop - rect.r_ybot) / 2;
        } 
        else {
            Tp_HorzTileToRect(planeDataPtr, tilePtr, &rect);
            absPosPtr->p_x = rect.r_xbot + (rect.r_xtop - rect.r_xbot) / 2;
            absPosPtr->p_y = (side == BOTTOM_LEFT_SIDE) ?
                rect.r_ybot : rect.r_ytop;
        }

        return (BEAR_OK);
    }

    if (Tp_IsSBNTile(tilePtr) == TRUE) {
        SearchData      searchData;
        register int    p1, p2;

        searchData.ppinPtr = ppinPtr;
        if (Tp_IsVertTile(tilePtr) == TRUE) {
            if (Tp_ItrBnLine(Tp_GetVertBnLineList(planeDataPtr),
                BnLineSearch, (ClientData) &searchData) == BEAR_STOP_SEARCH) {
                absPosPtr->p_x = searchData.linePtr->point1.p_x;
                p1 = searchData.linePtr->point1.p_y;
                p2 = searchData.linePtr->point2.p_y;
                absPosPtr->p_y = (p1 > p2) ? p2 + (p1 - p2) / 2 :
                    p1 + (p2 - p1) / 2;
                return (BEAR_OK);
            }
        } 
        else {
            if (Tp_ItrBnLine(Tp_GetHorzBnLineList(planeDataPtr),
                BnLineSearch, (ClientData) &searchData) == BEAR_STOP_SEARCH) {
                p1 = searchData.linePtr->point1.p_x;
                p2 = searchData.linePtr->point2.p_x;
                absPosPtr->p_x = (p1 > p2) ? p2 + (p1 - p2) / 2 :
                    p1 + (p2 - p1) / 2;
                absPosPtr->p_y = searchData.linePtr->point1.p_y;
                return (BEAR_OK);
            }
        }
    }

    Bear_PrintErr("GreGetAbsPosOfPPin() : failed");
    return (BEAR_ERROR);
}



