
/* tpPPinPrim.c --
 *
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpPPinPrim.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"

/* ---------- Pseudo Pin Property Primitives ---------- */

Net *
Tp_GetNetOfPPin(ppinPtr)
PseudoPin *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetNetOfPPin() failed : no pseudo pin given");
    }
    return (ppinPtr->netPtr);
}

void
Tp_PutNetOfPPin(netPtr, ppinPtr)
Net         *netPtr;
PseudoPin   *ppinPtr;
{
    if (netPtr == NIL(Net)) {
        TpPrintFatalErr("Tp_PutNetOfPPin() failed : no net given");
    } else if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_PutNetOfPPin() failed : no pseudo pin given");
    }
    ppinPtr->netPtr = netPtr;
}

BearId
Tp_GetIdOfPPin(ppinPtr)
PseudoPin *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetIdOfPPin() failed : no pseudo pin given");
    }
    return (ppinPtr->id);
}

void
Tp_PutIdOfPPin(id, ppinPtr)
BearId      id;
PseudoPin   *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_PutIdOfPPin() failed : no pseudo pin given");
    } else if (id == 0) {
        TpPrintFatalErr("Tp_PutIdOfPPin() failed : id value cannot be 0");
    }
    ppinPtr->id = id;
}

BearId
Tp_GetIntNetIdOfPPin(ppinPtr)
PseudoPin  *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetIntNetIdOfPPin() failed : no pseudo pin given");
    }
    return (ppinPtr->intNetId);
}

void
Tp_PutIntNetIdOfPPin(intNetId, ppinPtr)
BearId      intNetId;
PseudoPin   *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_PutIntNetIdOfPPin() failed");
    } else if (intNetId == 0) {
    TpPrintFatalErr("Tp_PutIntNetIdOfPPin() failed : id value cannot be 0");
    }
    ppinPtr->intNetId = intNetId;
}

BearId
Tp_GetExtNetIdOfPPin(ppinPtr)
PseudoPin  *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetExtNetIdOfPPin() failed");
    }
    return (ppinPtr->extNetId);
}

void
Tp_PutExtNetIdOfPPin(extNetId, ppinPtr)
BearId      extNetId;
PseudoPin   *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_PutExtNetIdOfPPin() failed");
    } else if (extNetId == 0) {
    TpPrintFatalErr("Tp_PutExtNetIdOfPPin() failed : id value cannot be 0");
    }
    ppinPtr->extNetId = extNetId;
}

float
Tp_GetWidthOfPPin(ppinPtr)
PseudoPin *ppinPtr;
{  
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetWidthOfPPin() failed");
    }
    return (ppinPtr->width);
}

void
Tp_PutWidthOfPPin(width, ppinPtr)
float     width;
PseudoPin *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_PutWidthOfPPin() failed");
    }
    ppinPtr->width = width;
}

int
Tp_GetOrderOfPPin(ppinPtr)
PseudoPin   *ppinPtr;
{  
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetOrderOfPPin() failed");
    }
    return (ppinPtr->order);
}

void
Tp_PutOrderOfPPin(order, ppinPtr)
int       order;
PseudoPin *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_PutOrderOfPPin() failed");
    }
    ppinPtr->order = order;
}

SideDirection
Tp_GetSideOfPPin(ppinPtr)
PseudoPin  *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetSideOfPPin() failed");
    }
    return (ppinPtr->side);
}

Tile *
Tp_GetTileOfPPin(ppinPtr)
PseudoPin *ppinPtr;
{
    if (ppinPtr == NIL(PseudoPin)) {
        TpPrintFatalErr("Tp_GetTileOfPPin() failed");
    }
    return (ppinPtr->tilePtr);
}

BearStatus
Tp_GetCenterOfSide(bnTilePtr, side, centerPtr)
Tile          *bnTilePtr;               /* Pointer to a bottle neck tile. */
SideDirection side;
Point         *centerPtr;
{
    Rect horzRect, vertRect;

#ifdef DEBUG
    if (bnTilePtr == NIL(Tile) ||
        (Tp_IsBottleneckTile(bnTilePtr) == FALSE) ||
        centerPtr == NIL(Point)) {
        TpPrintFatalErr("Tp_GetCenterOfSide() failed");
    }
#endif DEBUG

    if (Tp_IsBottleneckTile(bnTilePtr) == FALSE) {
        return (BEAR_ERROR);
    }
    if (Tp_IsVertTile(bnTilePtr)) {
        /* vertical bottleneck tile */
        TITORECT(bnTilePtr, &vertRect);
        Tp_VRectToHRect(&vertRect, &horzRect);
        centerPtr->p_x = (horzRect.r_xtop - horzRect.r_xbot) / 2;
        centerPtr->p_y = (side == BOTTOM_LEFT_SIDE) ?
            horzRect.r_ybot : horzRect.r_ytop;
    } 
    else {
        /* horizontal bottleneck tile */
        TITORECT(bnTilePtr, &horzRect);
        centerPtr->p_y = (horzRect.r_ytop - horzRect.r_ybot) / 2;
        centerPtr->p_x = (side == BOTTOM_LEFT_SIDE) ?
            horzRect.r_xbot : horzRect.r_xtop;
    }
    return (BEAR_OK);
}

