
/* tpPaint.c --
 *
 *      Andrew Purshottam
 *      George Carvalho
 *      (Advising Professor - E. S. Kuh)
 *
 * Routines to paint cells onto a given plane.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpPaint.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"


/* ---------- Local Procedures ---------- */
#ifdef BETTY_BOOP
static void ShrinkDCell();
#endif BETTY_BOOP


/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PaintGCell() --
 *
 * Check to make sure that the rectangle list is within the horizontal
 * tile planes boundary, and that the area under the rectangle list on
 * the horizontal tile plane contains only space tiles, and then paint
 * the generic cell on BOTH the horizontal and vertical tile planes.
 *
 * Results:
 *   TRUE is returned if everything is spiffy, FALSE otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

bool
Tp_PaintGCell(gcellPtr, planeDataPtr)
GenericCell *gcellPtr;
PlaneData   *planeDataPtr;
{
    CheckData   checkData;
    LinkedRect  *horzRectList, *vertRectList;
    Transform   *horzTransPtr, vertTrans;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintGCell() failed : no plane data given");
    }
#endif DEBUG

    horzRectList = Mac_GetHorzRectListOfGCell(gcellPtr);
    vertRectList = Mac_GetVertRectListOfGCell(gcellPtr);
    horzTransPtr = Mac_GetTransformOfGCell(gcellPtr);

    checkData.strong = TRUE;
    checkData.tileType = TT_SPACE;
    checkData.namePtr = NIL(char);
    checkData.planePtr = Tp_GetHorzPlane(planeDataPtr);
    checkData.bndPtr = Tp_GetHorzPlaneBnd(planeDataPtr);
    checkData.rectList = horzRectList;
    checkData.transPtr = horzTransPtr;
    if (TpCheckPlaneUnderRectList(&checkData)) {
        TpPaintRectList(Tp_GetHorzPlane(planeDataPtr),
            horzRectList, horzTransPtr, gcellPtr);

        GeoTransTrans(horzTransPtr, &Geo90Transform, &vertTrans);

        TpPaintRectList(Tp_GetVertPlane(planeDataPtr),
            vertRectList, &vertTrans, gcellPtr);
    }
    else {
        return (FALSE);
    }

    return (TRUE);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PaintGCellOnHorzPlane() --
 *
 * Paint the generic cell on ONLY the horizontal tile plane.
 * Note that checking will NOT be done.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

void
Tp_PaintGCellOnHorzPlane(gcellPtr, planeDataPtr)
GenericCell *gcellPtr;
PlaneData   *planeDataPtr;
{
    Transform *horzTransPtr;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintGCellOnHorzPlane() failed : no plane data given");
    }
#endif DEBUG

    horzTransPtr = Mac_GetTransformOfGCell(gcellPtr);
/* Mac_PrintRectList(&GeoIdentityTransform, Mac_GetHorzRectListOfGCell(gcellPtr)); */

    TpPaintRectList(Tp_GetHorzPlane(planeDataPtr),
        Mac_GetHorzRectListOfGCell(gcellPtr),
        horzTransPtr, gcellPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PaintGCellOnVertPlane() --
 *
 * Paint the generic cell on ONLY the vertical tile plane.
 * Note that checking will NOT be done.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

void
Tp_PaintGCellOnVertPlane(gcellPtr, planeDataPtr)
GenericCell *gcellPtr;
PlaneData   *planeDataPtr;
{
    Transform *horzTransPtr, vertTrans;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintGCellOnVertPlane() failed : no plane data given");
    }
#endif DEBUG

    horzTransPtr = Mac_GetTransformOfGCell(gcellPtr);
    GeoTransTrans(horzTransPtr, &Geo90Transform, &vertTrans);
/* Mac_PrintRectList(&GeoIdentityTransform, Mac_GetVertRectListOfGCell(gcellPtr)); */

    TpPaintRectList(Tp_GetVertPlane(planeDataPtr),
        Mac_GetVertRectListOfGCell(gcellPtr),
        &vertTrans, gcellPtr);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_UnpaintGCell() --
 *
 * Unpaint the generic cell on the BOTH horizontal and vertical
 * tile planes.
 *
 * Results:
 *   None.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

void
Tp_UnpaintGCell(gcellPtr, planeDataPtr)
GenericCell *gcellPtr;
PlaneData   *planeDataPtr;
{
    LinkedRect  *horzRectList, *vertRectList;
    Transform   *horzTransPtr, vertTrans;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_UnpaintGCell() failed : no plane data given");
    }
#endif DEBUG

    horzRectList = Mac_GetHorzRectListOfGCell(gcellPtr);
    vertRectList = Mac_GetVertRectListOfGCell(gcellPtr);
    horzTransPtr = Mac_GetTransformOfGCell(gcellPtr);

    TpUnpaintRectList(Tp_GetHorzPlane(planeDataPtr),
        horzRectList, horzTransPtr, TT_SPACE);

    GeoTransTrans(horzTransPtr, &Geo90Transform, &vertTrans);
    TpUnpaintRectList(Tp_GetVertPlane(planeDataPtr),
        vertRectList, &vertTrans, TT_SPACE);
}



#ifdef BETTY_BOOP
static void
ShrinkDCell(dcellPtr, length, halfLength)
DummyCell   *dcellPtr;
int         length, halfLength;
{
    LinkedRect  *horzRectList, *vertRectList;
    GenericCell gcell;
    Rect        absRect;
    Transform   *horzTransPtr;

    gcell.unType = (CellClass) DUMMY_CELL;
    gcell.unValue.dummyCellPtr = dcellPtr;

    horzRectList = Mac_GetHorzRectListOfGCell(&gcell);
    vertRectList = Mac_GetVertRectListOfGCell(&gcell);
    horzTransPtr = Mac_GetTransformOfGCell(&gcell);

    GeoTransRect(horzTransPtr, &horzRectList->r_r, &absRect);
    if ((absRect.r_xtop - absRect.r_xbot) > length) {
        absRect.r_xbot += ((absRect.r_xtop - absRect.r_xbot) / 2 - halfLength);
        absRect.r_xtop = absRect.r_xbot + length;
    }
    if ((absRect.r_ytop - absRect.r_ybot) > length) {
        absRect.r_ybot += ((absRect.r_ytop - absRect.r_ybot) / 2 - halfLength);
        absRect.r_ytop = absRect.r_ybot + length;
    }
    horzRectList->r_r = vertRectList->r_r = absRect;
    Mac_PutBBoxOfDCell(&absRect, dcellPtr);
    Mac_PutTransformOfGCell(&GeoIdentityTransform, &gcell);
}
#endif BETTY_BOOP


/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PaintDCell() --
 *
 * Paint the dummy cell on the BOTH horizontal and vertical
 * tile planes.  Note that checking will be done.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned if everything is spiffy, 
 *   BEAR_STOP_SEARCH otherwise.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_PaintDCell(dcellPtr, planeDataPtr)
DummyCell   *dcellPtr;
PlaneData   *planeDataPtr;
{
    GenericCell gcell;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintDCell() failed : no plane data given");
    }
#endif DEBUG

#ifdef BETTY_BOOP
{
    int length, halfLength;

    length = Mac_CalcMinPinSpacing();
    halfLength = length / 2;
    ShrinkDCell(dcellPtr, length, halfLength);
}
#endif BETTY_BOOP

    gcell.unType = (CellClass) DUMMY_CELL;
    gcell.unValue.dummyCellPtr = dcellPtr;
    return (Tp_PaintGCell(&gcell, planeDataPtr) ?
        BEAR_CONTINUE_SEARCH : BEAR_STOP_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PaintDCellOnHorzPlane() --
 *
 * Paint the dummy cell on ONLY the horizontal tile plane.
 * Note that checking will NOT be done.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_PaintDCellOnHorzPlane(dcellPtr, planeDataPtr)
DummyCell   *dcellPtr;
PlaneData   *planeDataPtr;
{
    GenericCell gcell;
    Tile        *tilePtr;
    Rect        *bBoxOfDCell;
    Point       searchPoint;
    Plane       *horzPlanePtr;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintDCellOnHorzPlane() failed : no plane data given");
    }
#endif DEBUG

    gcell.unType = (CellClass) DUMMY_CELL;
    gcell.unValue.dummyCellPtr = dcellPtr;
    Tp_PaintGCellOnHorzPlane(&gcell, planeDataPtr);
    horzPlanePtr = Tp_GetHorzPlane(planeDataPtr);
    bBoxOfDCell = Mac_GetBBoxOfDCell(dcellPtr);
    searchPoint.p_x = (bBoxOfDCell->r_xbot + bBoxOfDCell->r_xtop) / 2;
    searchPoint.p_y = (bBoxOfDCell->r_ybot + bBoxOfDCell->r_ytop) / 2;
    tilePtr = TiSrPoint(NIL(Tile),horzPlanePtr,&searchPoint);
    Mac_PutTileOfDCell(tilePtr, dcellPtr);
    Tp_PutDCellOfTile(dcellPtr, tilePtr);
    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PaintDCellOnVertPlane() --
 *
 * Paint the dummy cell on ONLY the vertical tile plane.
 * Note that checking will NOT be done.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_PaintDCellOnVertPlane(dcellPtr, planeDataPtr)
DummyCell   *dcellPtr;
PlaneData   *planeDataPtr;
{
    GenericCell gcell;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintDCellOnVertPlane() failed : no plane data given");
    }
#endif DEBUG

    gcell.unType = (CellClass) DUMMY_CELL;
    gcell.unValue.dummyCellPtr = dcellPtr;
    Tp_PaintGCellOnVertPlane(&gcell, planeDataPtr);

    return (BEAR_CONTINUE_SEARCH);
}



/*
 * ----------------------------------------------------------------------------
 *
 * Tp_UnpaintDCell() --
 *
 * Unpaint the dummy cell on the BOTH horizontal and vertical
 * tile planes.
 *
 * Results:
 *   BEAR_CONTINUE_SEARCH is returned.
 *
 * Side effects:
 *   None.
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_UnpaintDCell(dcellPtr, planeDataPtr)
DummyCell   *dcellPtr;
PlaneData   *planeDataPtr;
{
    GenericCell gcell;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_UnpaintDCell() failed : no plane data given");
    }
#endif DEBUG

    gcell.unType = (CellClass) DUMMY_CELL;
    gcell.unValue.dummyCellPtr = dcellPtr;
    Tp_UnpaintGCell(&gcell, planeDataPtr);

    return (BEAR_CONTINUE_SEARCH);
}



#ifdef USE_DR_CODE
void
Tp_PaintRing(ringPtr, planeDataPtr)
Ring        *ringPtr;
PlaneData   *planeDataPtr;
{
    Plane               *horzPlanePtr, *vertPlanePtr;
    Transform           *horzTransPtr, vertTrans;
    Rect                paintRect;
    TileType            tileType;
    register Tile       *tilePtr;
    register LinkedRect *rectPtr;

#ifdef DEBUG
    if (planeDataPtr == NIL(PlaneData)) {
        TpPrintFatalErr("Tp_PaintRing() failed : no plane data given");
    }
#endif DEBUG

    tileType = TP_CELL_BASE + Mac_GetIdOfRing(ringPtr);
    horzPlanePtr = Tp_GetHorzPlane(planeDataPtr);
    vertPlanePtr = Tp_GetVertPlane(planeDataPtr);
    horzTransPtr = Mac_GetTransformOfRing(ringPtr);
    (void) GeoTransTrans(horzTransPtr, &Geo90Transform, &vertTrans);

    for (rectPtr = ringPtr->horzRectList;
        rectPtr != NIL(LinkedRect);
        rectPtr = rectPtr->r_next) {

        GeoTransRect(horzTransPtr, &rectPtr->r_r, &paintRect);
        DBPaintPlaneByProc(horzPlanePtr, &paintRect,
            TpPaintProc, tileType,
            NIL(PaintUndoInfo));
        tilePtr = TiSrPointNoHint(horzPlanePtr, &paintRect.r_ll);
        Tp_CreateSolidTilePropForRing(tilePtr, ringPtr);
        Tp_PutRingOfTile(ringPtr, tilePtr);
    }

    for (rectPtr = ringPtr->vertRectList;
        rectPtr != NIL(LinkedRect);
        rectPtr = rectPtr->r_next) {

        GeoTransRect(&vertTrans, &rectPtr->r_r, &paintRect);
        DBPaintPlaneByProc(vertPlanePtr, &paintRect,
            TpPaintProc, tileType,
            NIL(PaintUndoInfo));
        tilePtr = TiSrPointNoHint(vertPlanePtr, &paintRect.r_ll);
        Tp_CreateSolidTilePropForRing(tilePtr, ringPtr);
        Tp_PutRingOfTile(ringPtr, tilePtr);
    }
}
#endif /* USE_DR_CODE */


