
/* tpPushVert.c -
 *
 *     George Carvalho
 *     (Advising Professor - E. S. Kuh)
 *
 * Copyright (C) 1986 Regents of the University of California
 * All rights reserved.
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpPushVert.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"

/* ---------- Local Procedures ---------- */
int PushVertTile();

/* ---------- Local Global Variables ---------- */
static Rect          searchArea;
static LinkedTile    *tileList;
static LinkedCell    *gCellList;


/* ARGSUSED */
static int
PushVertTile(tilePtr, clientData)
register Tile   *tilePtr;
ClientData      clientData;
{
    GenericCell        *gCellPtr;
    CellInst           *cellInstPtr;
    RouteCell          *rCellPtr; 
    register LinkedTile    *lnkTilePtr;
    register LinkedCell    *lnkGCellPtr;

    if (BOTTOM(tilePtr) >= searchArea.r_ybot) 
    {
        /* Add this tile onto `tileList'. */
        MALLOC(LinkedTile *, lnkTilePtr, sizeof (LinkedTile));
        lnkTilePtr->tilePtr = tilePtr;
        lnkTilePtr->nextPtr = tileList;
        tileList = lnkTilePtr;

    /*
     * Add the corresponding generic cell onto `gCellList' 
     * if it is not already in.
     */
        if (Tp_IsCellInstTile(tilePtr) && 
            (Mac_IsCellMarked(cellInstPtr = Tp_GetCellInstOfTile(tilePtr)) ==
            FALSE)) 
    {
            Mac_MarkCell(cellInstPtr);
        (void) Mac_CreateGCell(CELL_INST, (ClientData) cellInstPtr, &gCellPtr);
        lnkGCellPtr = Mac_MakeLinkedCell();
        lnkGCellPtr->cellPtr = gCellPtr;
        lnkGCellPtr->nextPtr = gCellList;
        gCellList = lnkGCellPtr;
        }
    else if (Tp_IsRCellTile(tilePtr) &&
         (Mac_IsRCellMarked(rCellPtr = Tp_GetRCellOfTile(tilePtr)) ==
         BEAR_FALSE))
    {
        Mac_MarkRCell(rCellPtr);
        (void) Mac_CreateGCell(ROUTE_CELL, (ClientData) rCellPtr, &gCellPtr);
        lnkGCellPtr = Mac_MakeLinkedCell();
        lnkGCellPtr->cellPtr = gCellPtr;
        lnkGCellPtr->nextPtr = gCellList;
        gCellList = lnkGCellPtr;
    }
    }
    return (0); /* continue search */
}
/*
 * ----------------------------------------------------------------------------
 *
 * Tp_PushVert() --
 *
 * A valid vertical space tile list has the following properties:
 *   1.) The first tile in the list must touch the (horizontal) plane
 *       boundaries left edge.
 *       Note that a plane boundary is defined to be the smallest rectangle
 *       that encloses all of the cells.
 *   2.) The last tile in the list must touch the (horizontal) plane
 *       boundaries right edge.
 *
 * Results:
 *   TRUE is returned if everything is spiffy, FALSE otherwise.
 *
 * Side effects:
 *   None.
 *
 * ----------------------------------------------------------------------------
 */

bool
Tp_PushVert(planeDataPtr, vertSpaceTileList, distance)
PlaneData   *planeDataPtr;
LinkedTile  *vertSpaceTileList;
int         distance;
{
    CellInst    *cellInstPtr;
    GenericCell *gCellPtr;
    Point       delta, *originPtr, newOrigin;
    Plane       *horzPlanePtr, *vertPlanePtr;
    Rect        *vertPlaneBndPtr;
    Tile        *topPseudoTilePtr, *botPseudoTilePtr;
    Tile        *rightPseudoTilePtr;
    Transform   *horzTransPtr, trans;
    int         minHeight;
    RouteCell   *rCellPtr;
    register LinkedTile *ltPtr;
    register LinkedCell    *lnkCellPtr;

/*
printf("distance(input) = %d;\n", distance);
fflush(stdout);
*/

    if (distance < 0) 
    {
    minHeight = Tp_FindMinHeight(planeDataPtr, vertSpaceTileList);
        if (minHeight <= 0) 
    {
            Bear_PrintErr("you cannot compact any further along this ridge");
            return (FALSE);
        } 
    else if (ABS(distance) > minHeight) 
    {
            Bear_PrintErr("you cannot compact that much along this ridge");
            return (FALSE);
        }
    }
    delta.p_x = 0;
    delta.p_y = distance;

/*
printf("after assign: delta.p_x = %d; delta.p_y = %d;\n", delta.p_x, delta.p_y);
fflush(stdout);
*/

    horzPlanePtr = Tp_GetHorzPlane(planeDataPtr);
    vertPlanePtr = Tp_GetVertPlane(planeDataPtr);
    vertPlaneBndPtr = Tp_GetVertPlaneBnd(planeDataPtr);
    searchArea.r_xtop = vertPlaneBndPtr->r_xtop;
    topPseudoTilePtr = vertPlanePtr->pl_top;
    botPseudoTilePtr = vertPlanePtr->pl_bottom;
    rightPseudoTilePtr = vertPlanePtr->pl_right;
    tileList = NIL(LinkedTile);
    gCellList = NIL(LinkedCell);

    ltPtr = vertSpaceTileList;
    do {
        searchArea.r_xbot = (TR(ltPtr->tilePtr) == rightPseudoTilePtr) ?
            vertPlaneBndPtr->r_xtop : RIGHT(ltPtr->tilePtr);
        searchArea.r_ytop = (RT(ltPtr->tilePtr) == topPseudoTilePtr) ?
            vertPlaneBndPtr->r_ytop : TOP(ltPtr->tilePtr);
        while (ltPtr->nextPtr != NIL(LinkedTile)) {
            if (TR(ltPtr->tilePtr) == TR(ltPtr->nextPtr->tilePtr)) {
                ltPtr = ltPtr->nextPtr;
            } else {
                break;
            }
        }
        searchArea.r_ybot = (LB(ltPtr->tilePtr) == botPseudoTilePtr) ?
            vertPlaneBndPtr->r_ybot : BOTTOM(ltPtr->tilePtr);
        if (GEO_RECTNULL(&searchArea) == FALSE) {
            (void) TiSrArea(NIL(Tile), vertPlanePtr, &searchArea,
                PushVertTile, (ClientData) NULL);
        }
    } while (ltPtr = ltPtr->nextPtr);

    /* Go through 'tileList' and move each tile. */
    for (ltPtr = tileList; ltPtr != NIL(LinkedTile) ;ltPtr = ltPtr->nextPtr) {
        LEFT(ltPtr->tilePtr) += distance;
    }
    Tp_FreeTileList(tileList);

    /*
     * Go through 'gCellList' and unpaint each generic cell
     * from the horizontal tile plane.
     */
    for (lnkCellPtr = gCellList;
         lnkCellPtr != NIL(LinkedCell);
         lnkCellPtr = lnkCellPtr->nextPtr) 
    {
    gCellPtr = lnkCellPtr->cellPtr;
        TpUnpaintRectList(horzPlanePtr,
            Mac_GetHorzRectListOfGCell(gCellPtr),
            Mac_GetTransformOfGCell(gCellPtr),
            TT_SPACE);
    /*
    Tp_UnPaintGCellOnHorzPlane(gCellPtr, planeDataPtr);
    */
    }

    /*
     * Now go through 'gCellList' and paint each cell
     * onto the horizontal tile plane in its new postion.
     */
    for (lnkCellPtr = gCellList;
         lnkCellPtr != NIL(LinkedCell);
         lnkCellPtr = lnkCellPtr->nextPtr) 
    {
    gCellPtr = lnkCellPtr->cellPtr;
    if (gCellPtr->unType == CELL_INST)
    {
        cellInstPtr = gCellPtr->unValue.cellInstPtr;
        Mac_UnmarkCell(cellInstPtr);
        /* Move and paint cell (trans.) */
        horzTransPtr = Mac_GetTransformOfGCell(gCellPtr);
        GeoTranslateTrans(horzTransPtr, delta.p_x, delta.p_y, horzTransPtr);
        Tp_PaintGCellOnHorzPlane(gCellPtr, planeDataPtr);
        /* Move cell boundary */
        GeoTranslateTrans(&GeoIdentityTransform, delta.p_x, delta.p_y, &trans);
        (void) Mac_MoveBoundaryOfCell(&trans, cellInstPtr);
        /* Move cell origin */
        originPtr = Mac_GetOriginOfCell(cellInstPtr);
        GeoTransPoint(&trans, originPtr, &newOrigin);
        Mac_PutOriginOfCell(&newOrigin, cellInstPtr);
    }
    else if (gCellPtr->unType == ROUTE_CELL)
    {
        rCellPtr = gCellPtr->unValue.routeCellPtr;
        Mac_UnmarkRCell(rCellPtr);
        /* Move and paint route cell (trans.) */
        horzTransPtr = Mac_GetTransformOfGCell(gCellPtr);
        GeoTranslateTrans(horzTransPtr, delta.p_x, delta.p_y, horzTransPtr);
        Tp_PaintGCellOnHorzPlane(gCellPtr, planeDataPtr);
        /* Move route cell origin */
        originPtr = Mac_GetOriginOfRCell(rCellPtr);
        GeoTransPoint(&trans, originPtr, &newOrigin);
        Mac_PutOriginOfRCell(&newOrigin, rCellPtr);
    }
    }
    /* Free 'gCellList' */
    while ((lnkCellPtr = gCellList) != NIL(LinkedCell))
    {
        gCellList = lnkCellPtr->nextPtr;
        FREE((char *) lnkCellPtr);
    }
    return (TRUE);
}



