
/* tpRefRot.c --
 *
 *     Sangjin Hong
 *     (Advising Professor - E. S. Kuh)
 *
 * rotation and reflection  of cell .
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpRefRot.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"


/* ---------- Local Procedures ---------- */
static int CheckTileType();


BearStatus 
Tp_RotateCell(cellMasterPtr,cellInstPtr, rotateAngle)
CellMaster     *cellMasterPtr;
CellInst       *cellInstPtr;
int            rotateAngle;
{
    GenericCell   genericCellPtr;
    PlaneData      *planeDataPtr;
    CellInst   *coreInstPtr;
    Rect      bbox;
    int       x,y;
    int       dx,dy;
    int       delta;
    Rect      checkArea;
    LinkedLineSeg *boundary;
    LinkedLineSeg *alineSeg;

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    coreInstPtr = Mac_GetInstOfCoreMaster(cellMasterPtr);
    genericCellPtr.unType = CELL_INST;
    genericCellPtr.unValue.cellInstPtr = cellInstPtr;
    boundary = Mac_GetBoundaryOfCell(cellInstPtr);
    bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
    bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
    for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
         alineSeg = alineSeg->nextPtr) 
    {
        if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
    {
            bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
    {
            bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
    {
            bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
        }
        if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
    {
            bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
        }
    }

    x = (bbox.r_xtop + bbox.r_xbot) / 2;
    y = (bbox.r_ytop + bbox.r_ybot) / 2;
    dx = (bbox.r_xtop - bbox.r_xbot) / 2;
    dy = (bbox.r_ytop - bbox.r_ybot) / 2;
    checkArea.r_xbot = x - dy;
    checkArea.r_ybot = y - dx;
    checkArea.r_xtop = x + dy;
    checkArea.r_ytop = y + dx;

    boundary = Mac_GetBoundaryOfCell(coreInstPtr);
    bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
    bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
    for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
         alineSeg = alineSeg->nextPtr) 
    {
        if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
    {
            bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
    {
            bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
    {
            bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
        }
        if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
    {
            bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
        }
    }
    if (rotateAngle != 180)
    {
        if (TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr), &checkArea, 
             CheckTileType, (ClientData) cellInstPtr) != 0)
        {
            return (BEAR_FALSE);
        }
    }

    if (checkArea.r_xtop > bbox.r_xtop)
    {
    delta = checkArea.r_xtop - bbox.r_xtop+10;
    (void) Tp_ResizeParentCell(0,0,delta, 0, coreInstPtr,planeDataPtr);
    }
    if (checkArea.r_xbot < bbox.r_xbot)
    {
    delta = bbox.r_xbot - checkArea.r_xbot+10;
    (void) Tp_ResizeParentCell(-delta, 0, delta,0, coreInstPtr,planeDataPtr);
    }
    if (checkArea.r_ytop > bbox.r_ytop)
    {
    delta = checkArea.r_ytop - bbox.r_ytop+10;
    (void) Tp_ResizeParentCell(0,0,0,delta, coreInstPtr,planeDataPtr);
    }
    if (checkArea.r_ybot < bbox.r_ybot)
    {
    delta = bbox.r_ybot - checkArea.r_ybot+10;
    (void) Tp_ResizeParentCell(0,-delta,0, delta,coreInstPtr,planeDataPtr);
    }

    Tp_UnpaintGCell(&genericCellPtr, planeDataPtr);
    Mac_RotateCell(rotateAngle,cellInstPtr);
    (void) Tp_PaintGCell(&genericCellPtr, planeDataPtr);
    return (BEAR_TRUE);
}



BearStatus 
Tp_ReflectCell(cellMasterPtr,cellInstPtr, reflect)
CellMaster     *cellMasterPtr;
CellInst       *cellInstPtr;
int            reflect;
{
    GenericCell    genericCellPtr;
    PlaneData      *planeDataPtr;
    CellInst       *coreInstPtr;
    Rect      bbox;
    int       x,y;
    int       dx,dy;
    int       delta;
    Rect      checkArea;
    LinkedLineSeg *boundary;
    LinkedLineSeg *alineSeg;

    planeDataPtr = Mac_GetCellPlaneOfCell(cellMasterPtr);
    coreInstPtr = Mac_GetInstOfCoreMaster(cellMasterPtr);
    genericCellPtr.unType = CELL_INST;
    genericCellPtr.unValue.cellInstPtr = cellInstPtr;
    boundary = Mac_GetBoundaryOfCell(cellInstPtr);
    bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
    bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
    for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
         alineSeg = alineSeg->nextPtr) 
    {
        if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
    {
            bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
    {
            bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
    {
            bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
        }
        if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
    {
            bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
        }
    }

    x = (bbox.r_xtop + bbox.r_xbot) / 2;
    y = (bbox.r_ytop + bbox.r_ybot) / 2;
    dx = (bbox.r_xtop - bbox.r_xbot) / 2;
    dy = (bbox.r_ytop - bbox.r_ybot) / 2;
    checkArea.r_xbot = x - dy;
    checkArea.r_ybot = y - dx;
    checkArea.r_xtop = x + dy;
    checkArea.r_ytop = y + dx;

    boundary = Mac_GetBoundaryOfCell(coreInstPtr);
    bbox.r_xbot = bbox.r_xtop = boundary->lineSeg->line.point1.p_x;
    bbox.r_ybot = bbox.r_ytop = boundary->lineSeg->line.point1.p_y;
    for (alineSeg = boundary->nextPtr; alineSeg != NIL(LinkedLineSeg);
         alineSeg = alineSeg->nextPtr) 
    {
        if (alineSeg->lineSeg->line.point1.p_x < bbox.r_xbot)
    {
            bbox.r_xbot = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_x > bbox.r_xtop) 
    {
            bbox.r_xtop = alineSeg->lineSeg->line.point1.p_x;
        }
        if (alineSeg->lineSeg->line.point1.p_y < bbox.r_ybot) 
    {
            bbox.r_ybot = alineSeg->lineSeg->line.point1.p_y;
        }
        if (alineSeg->lineSeg->line.point1.p_y > bbox.r_ytop) 
    {
            bbox.r_ytop = alineSeg->lineSeg->line.point1.p_y;
        }
    }
    if (TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr), &checkArea, 
         CheckTileType, (ClientData) cellInstPtr) != 0)
    {
        return (BEAR_FALSE);
    }
    if (checkArea.r_xtop > bbox.r_xtop)
    {
    delta = checkArea.r_xtop - bbox.r_xtop+10;
    (void) Tp_ResizeParentCell(0,0,delta, 0, coreInstPtr,planeDataPtr);
    }
    if (checkArea.r_xbot < bbox.r_xbot)
    {
    delta = bbox.r_xbot - checkArea.r_xbot+10;
    (void) Tp_ResizeParentCell(-delta, 0, delta,0, coreInstPtr,planeDataPtr);
    }
    if (checkArea.r_ytop > bbox.r_ytop)
    {
    delta = checkArea.r_ytop - bbox.r_ytop+10;
    (void) Tp_ResizeParentCell(0,0,0,delta, coreInstPtr,planeDataPtr);
    }
    if (checkArea.r_ybot < bbox.r_ybot)
    {
    delta = bbox.r_ybot - checkArea.r_ybot+10;
    (void) Tp_ResizeParentCell(0,-delta,0, delta,coreInstPtr,planeDataPtr);
    }
    Tp_UnpaintGCell(&genericCellPtr, planeDataPtr);
    Mac_ReflectCell(reflect,cellInstPtr);
    (void) Tp_PaintGCell(&genericCellPtr, planeDataPtr);
    return (BEAR_TRUE);
}



static int
CheckTileType(tilePtr, cellInstPtr)
Tile       *tilePtr;
CellInst   *cellInstPtr;
{
    if (Tp_IsSpaceTile(tilePtr) || Tp_IsDCellTile(tilePtr)) {
        return (0); /* continue search */
    }

    if (Tp_IsCellInstTile(tilePtr)) {
        if (cellInstPtr == (Tp_GetCellInstOfTile(tilePtr))) {
            return (0); /* continue search */
        }
        else {
            return (1); /* stop search */
        }
    }
    else if (Tp_IsRCellTile(tilePtr)) {
        return (1); /* stop search */
    }
    return (1); /* stop search */
}


