
/* tpResize.c --
 *
 *     Masahiro Kobayashi
 *     (Advising Professor - E. S. Kuh)
 *
 * Resizing of cells.
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpResize.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"
#include "tileMapping.h"

typedef struct _PPinInfo PPinInfo;
struct _PPinInfo {
    PseudoPin         *trPPinList;
    PseudoPin         *blPPinList;
    int               fragment;
    int               yPos;
    struct _PPinInfo  *nextPtr;
};

typedef struct _SavePPinInfo SavePPinInfo;
struct _SavePPinInfo {
    PPinInfo    *blPPinInfoList;
    PPinInfo    *trPPinInfoList;
};


static   int  FindBnTile();
static   void TpGetLAndRBnTilesOfCell();
static   void TpGetBAndTBnTilesOfCell();
static   void TpSavePPinList();
static   void TpRestorePPinList();
static   void TpFreePPinInfoList();
static   BearStatus MoveCell();
static   BearStatus MoveBoundary();


/*
 * ----------------------------------------------------------------------------
 *
 * Tp_ResizeCellInst() --
 *
 *    Resize a cell instance.
 *
 * Results:
 *    BEAR_OK is returned if everything went down okay, BEAR_ERROR otherwise.
 *
 * Side effects:
 *    The tile planes and the data base are updated.
 *
 * ----------------------------------------------------------------------------
 */

BearStatus
Tp_ResizeCellInst(moveDeltaX,moveDeltaY,resizeDeltaX,resizeDeltaY,cellInstPtr)
int         moveDeltaX;
int         moveDeltaY;
int         resizeDeltaX;
int         resizeDeltaY;
CellInst    *cellInstPtr;
{
    PlaneData    *planeDataPtr;
    CellMaster   *parentCellPtr;
    CellInst     *curCellInstPtr;
    GenericCell  gcell;
    LinkedTile   *trBnTileList;
    LinkedTile   *blBnTileList;
    SavePPinInfo *savePPinInfo;
    SavePPinInfo *tempPPinInfo;
    PPinInfo     *blPPinInfoList;
    PPinInfo     *trPPinInfoList;
    Point        deltaPtr;

    parentCellPtr = Mac_GetParentOfCell(cellInstPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    gcell.unType = CELL_INST;
  
    /* Horizontal */
    if (Mac_IsGlobalRoutingDone(parentCellPtr)) {
        Tm_HUnify(parentCellPtr);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            MALLOC(SavePPinInfo *, savePPinInfo, sizeof (SavePPinInfo));
            blBnTileList = trBnTileList = NIL(LinkedTile);
            (void) TpGetLAndRBnTilesOfCell(curCellInstPtr, &blBnTileList,
                &trBnTileList);
            blPPinInfoList = trPPinInfoList = NIL(PPinInfo);
            TpSavePPinList(blBnTileList, BOTTOM_LEFT_SIDE, &blPPinInfoList);
            TpSavePPinList(trBnTileList, TOP_RIGHT_SIDE,   &trPPinInfoList);
            savePPinInfo->blPPinInfoList = blPPinInfoList;
            savePPinInfo->trPPinInfoList = trPPinInfoList;
            curCellInstPtr->clientData = (ClientData) savePPinInfo;
            Tp_FreeTileList(blBnTileList);
            Tp_FreeTileList(trBnTileList);
        }

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            gcell.unValue.cellInstPtr = curCellInstPtr;
            Tp_UnpaintGCell(&gcell, planeDataPtr);
        }
        deltaPtr.p_x = moveDeltaX;
        deltaPtr.p_y = 0;
        (void) Mac_ItrInstOfCell(cellInstPtr->masterPtr, MoveCell,
            (ClientData) &deltaPtr);
        Mac_ResizeCell(resizeDeltaX, 0, cellInstPtr);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            gcell.unValue.cellInstPtr = curCellInstPtr;
            (void) Tp_PaintGCell(&gcell, planeDataPtr);
        }
        (void) Tp_IdentifySpaceTiles(planeDataPtr);
        (void) Fpg_CreateFloorPlanGraph(planeDataPtr, (WallJunction **) NULL);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            blBnTileList = trBnTileList = NIL(LinkedTile);
            (void) TpGetLAndRBnTilesOfCell(curCellInstPtr, &blBnTileList,
                &trBnTileList);
            tempPPinInfo = (SavePPinInfo *) curCellInstPtr->clientData;
            blPPinInfoList = tempPPinInfo->blPPinInfoList;
            trPPinInfoList = tempPPinInfo->trPPinInfoList;
            TpRestorePPinList(blBnTileList, blPPinInfoList);
            TpRestorePPinList(trBnTileList, trPPinInfoList);
            Tp_FreeTileList(blBnTileList);
            Tp_FreeTileList(trBnTileList);
            TpFreePPinInfoList(blPPinInfoList);
            TpFreePPinInfoList(trPPinInfoList);
            FREE((char *) tempPPinInfo);
        }

        Tm_HSplit(parentCellPtr);
        
        /* Vertical */
        Tm_VUnify(parentCellPtr);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            MALLOC(SavePPinInfo *, savePPinInfo, sizeof (SavePPinInfo));
            blBnTileList = trBnTileList = NIL(LinkedTile);
            (void) TpGetBAndTBnTilesOfCell(curCellInstPtr, &blBnTileList,
                &trBnTileList);
            blPPinInfoList = trPPinInfoList = NIL(PPinInfo);
            TpSavePPinList(blBnTileList, BOTTOM_LEFT_SIDE, &blPPinInfoList);
            TpSavePPinList(trBnTileList, TOP_RIGHT_SIDE,   &trPPinInfoList);
            savePPinInfo->blPPinInfoList = blPPinInfoList;
            savePPinInfo->trPPinInfoList = trPPinInfoList;
            curCellInstPtr->clientData = (ClientData) savePPinInfo;
            Tp_FreeTileList(blBnTileList);
            Tp_FreeTileList(trBnTileList);
        }

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            gcell.unValue.cellInstPtr = curCellInstPtr;
            Tp_UnpaintGCell(&gcell, planeDataPtr);
        }
        deltaPtr.p_x = 0;
        deltaPtr.p_y = moveDeltaY;
        (void) Mac_ItrInstOfCell(cellInstPtr->masterPtr, MoveCell,
            (ClientData) &deltaPtr);
        Mac_ResizeCell(0, resizeDeltaY, cellInstPtr);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            gcell.unValue.cellInstPtr = curCellInstPtr;
            (void) Tp_PaintGCell(&gcell, planeDataPtr);
        }
        (void) Tp_IdentifySpaceTiles(planeDataPtr);
        (void) Fpg_CreateFloorPlanGraph(planeDataPtr, (WallJunction **) NULL);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            blBnTileList = trBnTileList = NIL(LinkedTile);
            (void) TpGetBAndTBnTilesOfCell(curCellInstPtr, &blBnTileList,
                &trBnTileList);
            tempPPinInfo = (SavePPinInfo *) curCellInstPtr->clientData;
            blPPinInfoList = tempPPinInfo->blPPinInfoList;
            trPPinInfoList = tempPPinInfo->trPPinInfoList;
            TpRestorePPinList(blBnTileList, blPPinInfoList);
            TpRestorePPinList(trBnTileList, trPPinInfoList);
            Tp_FreeTileList(blBnTileList);
            Tp_FreeTileList(trBnTileList);
            TpFreePPinInfoList(blPPinInfoList);
            TpFreePPinInfoList(trPPinInfoList);
            FREE((char *) tempPPinInfo);
        }

        Tm_VSplit(parentCellPtr);
    }
    else {
        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            gcell.unValue.cellInstPtr = curCellInstPtr;
            Tp_UnpaintGCell(&gcell, planeDataPtr);
        }

        deltaPtr.p_x = moveDeltaX;
        deltaPtr.p_y = moveDeltaY;
        (void) Mac_ItrInstOfCell(cellInstPtr->masterPtr, MoveCell,
            (ClientData) &deltaPtr);
        Mac_ResizeCell(resizeDeltaX, resizeDeltaY, cellInstPtr);

        for (curCellInstPtr = cellInstPtr->masterPtr->instList;
            curCellInstPtr != NIL(CellInst);
            curCellInstPtr = curCellInstPtr->nextPtr) {

            gcell.unValue.cellInstPtr = curCellInstPtr;
            (void) Tp_PaintGCell(&gcell, planeDataPtr);
        }
        (void) Tp_IdentifySpaceTiles(planeDataPtr);
        (void) Fpg_CreateFloorPlanGraph(planeDataPtr, (WallJunction **) NULL);
    }

    return (BEAR_OK);
}

 

static BearStatus
MoveCell(cellInstPtr, deltaPtr)
CellInst       *cellInstPtr;
Point          *deltaPtr;

{
    (void) MoveBoundary(deltaPtr->p_x, deltaPtr->p_y, cellInstPtr);
    return (BEAR_CONTINUE_SEARCH);
}



static BearStatus
MoveBoundary(moveDeltaX, moveDeltaY, cellInstPtr)
int      moveDeltaX;
int      moveDeltaY;
CellInst *cellInstPtr;
{
    Transform trans;
    Transform newTrans;

    GeoTranslateTrans(&cellInstPtr->transform,
              moveDeltaX, moveDeltaY, &newTrans);
    cellInstPtr->transform = newTrans;
    GeoTranslateTrans(&GeoIdentityTransform, moveDeltaX,
                      moveDeltaY, &trans);
    (void) Mac_MoveBoundaryOfCell(&trans, cellInstPtr);
}



static void
TpGetLAndRBnTilesOfCell(cellInstPtr, leftBnTileListPtr, rightBnTileListPtr)
CellInst   *cellInstPtr;
LinkedTile **leftBnTileListPtr;
LinkedTile **rightBnTileListPtr;
{
    CellMaster  *parentCellPtr;
    PlaneData   *planeDataPtr;
    GenericCell gcell;
    Rect        searchRect;
    Rect        bbox;

    parentCellPtr = Mac_GetParentOfCell(cellInstPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;
    GeoTransRect(Mac_GetTransformOfGCell(&gcell),
             Mac_GetBBoxOfCell(cellInstPtr),&bbox);
    searchRect.r_ll.p_x = bbox.r_ur.p_x;
    searchRect.r_ll.p_y = bbox.r_ll.p_y;
    searchRect.r_ur.p_x = bbox.r_ur.p_x + 1;
    searchRect.r_ur.p_y = bbox.r_ur.p_y;
    *rightBnTileListPtr = NIL(LinkedTile);
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
         &searchRect, FindBnTile, (ClientData)rightBnTileListPtr);

    searchRect.r_ll.p_x = bbox.r_ll.p_x -1;
    searchRect.r_ll.p_y = bbox.r_ll.p_y;
    searchRect.r_ur.p_x = bbox.r_ll.p_x;
    searchRect.r_ur.p_y = bbox.r_ur.p_y;
    *leftBnTileListPtr = NIL(LinkedTile);
    (void) TiSrArea(NIL(Tile), Tp_GetHorzPlane(planeDataPtr),
         &searchRect, FindBnTile, (ClientData)leftBnTileListPtr);
}


static void
TpGetBAndTBnTilesOfCell(cellInstPtr, bottomBnTileListPtr, topBnTileListPtr)
CellInst   *cellInstPtr;
LinkedTile **bottomBnTileListPtr;
LinkedTile **topBnTileListPtr;
{
    CellMaster  *parentCellPtr;
    PlaneData   *planeDataPtr;
    GenericCell gcell;
    Rect        hSearchRect;
    Rect        vSearchRect;
    Rect        bbox;

    parentCellPtr = Mac_GetParentOfCell(cellInstPtr);
    planeDataPtr = Mac_GetCellPlaneOfCell(parentCellPtr);
    gcell.unType = CELL_INST;
    gcell.unValue.cellInstPtr = cellInstPtr;
    GeoTransRect(Mac_GetTransformOfGCell(&gcell),
         Mac_GetBBoxOfCell(cellInstPtr),&bbox);
    hSearchRect.r_ll.p_x = bbox.r_ll.p_x;
    hSearchRect.r_ll.p_y = bbox.r_ur.p_y;
    hSearchRect.r_ur.p_x = bbox.r_ur.p_x;
    hSearchRect.r_ur.p_y = bbox.r_ur.p_y + 1;
    (void) Tp_HRectToVRect(&hSearchRect, &vSearchRect);
    *topBnTileListPtr = NIL(LinkedTile);
    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
         &vSearchRect, FindBnTile, (ClientData)topBnTileListPtr);

    hSearchRect.r_ll.p_x = bbox.r_ll.p_x;
    hSearchRect.r_ll.p_y = bbox.r_ll.p_y -1;
    hSearchRect.r_ur.p_x = bbox.r_ur.p_x;
    hSearchRect.r_ur.p_y = bbox.r_ll.p_y;
    (void) Tp_HRectToVRect(&hSearchRect, &vSearchRect);
    *bottomBnTileListPtr = NIL(LinkedTile);
    (void) TiSrArea(NIL(Tile), Tp_GetVertPlane(planeDataPtr),
         &vSearchRect, FindBnTile, (ClientData)bottomBnTileListPtr);
}



static int
FindBnTile(tilePtr,bnTileListPtr)
Tile         *tilePtr;
LinkedTile   **bnTileListPtr;
{
    LinkedTile *newLinkedTilePtr;

    if (Tp_IsBottleneckTile(tilePtr) || Tp_IsSBNTile(tilePtr)) {
        MALLOC(LinkedTile *, newLinkedTilePtr, sizeof (LinkedTile));
        newLinkedTilePtr->nextPtr = *bnTileListPtr;
        newLinkedTilePtr->tilePtr = tilePtr;
        *bnTileListPtr = newLinkedTilePtr;
    }
    return (0);
}



static void
TpSavePPinList(bnTileList, side, pPinInfoListPtr)
LinkedTile    *bnTileList;
SideDirection side;
PPinInfo      **pPinInfoListPtr;
{
    Tile    *tilePtr;
    PPinInfo    *newPPinInfoPtr;
    LinkedTile  *curLinkedTilePtr;

    curLinkedTilePtr = bnTileList;
    while (curLinkedTilePtr != NIL(LinkedTile))
    {
    MALLOC(PPinInfo *, newPPinInfoPtr, sizeof (PPinInfo));
    tilePtr = curLinkedTilePtr->tilePtr;
    if (Tp_IsBottleneckTile(tilePtr))
    {
        newPPinInfoPtr->fragment = TP_BOTTOM;
    }
    else if (Tp_IsSBNTile(tilePtr))
    {
        if (side == BOTTOM_LEFT_SIDE)
        {
        newPPinInfoPtr->fragment = TP_TOP;
        }
        else
        {
        newPPinInfoPtr->fragment = TP_BOTTOM;
        }
    }
    else
    {
        TpPrintFatalErr("TpSavePPinList(): not a BN or SBN tile");
    }
    (void) Tp_RemovePPinListOfTile(tilePtr, TOP_RIGHT_SIDE,
                       newPPinInfoPtr->fragment,
                       &newPPinInfoPtr->trPPinList);
    (void) Tp_RemovePPinListOfTile(tilePtr, BOTTOM_LEFT_SIDE,
                       newPPinInfoPtr->fragment,
                       &newPPinInfoPtr->blPPinList);
    newPPinInfoPtr->yPos = BOTTOM(tilePtr);
    newPPinInfoPtr->nextPtr = *pPinInfoListPtr;
    *pPinInfoListPtr = newPPinInfoPtr;
    curLinkedTilePtr = curLinkedTilePtr->nextPtr;
    }
    return;
}



static void
TpRestorePPinList(bnTileList, pPinInfoList)
LinkedTile  *bnTileList;
PPinInfo    *pPinInfoList;
{
    Tile       *tilePtr;
    PPinInfo   *curPPinInfoPtr;
    LinkedTile *curLinkedTilePtr;

    curLinkedTilePtr = bnTileList;
    while (curLinkedTilePtr != NIL(LinkedTile))
    {
    tilePtr = curLinkedTilePtr->tilePtr;
    curPPinInfoPtr = pPinInfoList;
    while (curPPinInfoPtr != NIL(PPinInfo))
    {
        if (BOTTOM(tilePtr) == curPPinInfoPtr->yPos)
        {
        break;
        }
        curPPinInfoPtr = curPPinInfoPtr->nextPtr;
    }
    if (curPPinInfoPtr == NIL(PPinInfo))
    {
        TpPrintFatalErr("TpRestorePPinList():can't find any bn tile");
    }
    (void) Tp_AppendPPinListOfTile(curPPinInfoPtr->trPPinList, tilePtr,
                    TOP_RIGHT_SIDE,
                    curPPinInfoPtr->fragment); 
    (void) Tp_AppendPPinListOfTile(curPPinInfoPtr->blPPinList, tilePtr,
                    BOTTOM_LEFT_SIDE,
                    curPPinInfoPtr->fragment); 
    curLinkedTilePtr = curLinkedTilePtr->nextPtr;
    }
    return;
}



static void
TpFreePPinInfoList(pPinInfoList)
PPinInfo  *pPinInfoList;
{
    PPinInfo  *curPPinInfoPtr;

    curPPinInfoPtr = pPinInfoList;
    while (curPPinInfoPtr != NIL(PPinInfo))
    {
    pPinInfoList = curPPinInfoPtr->nextPtr;
    FREE((char *)curPPinInfoPtr);
    curPPinInfoPtr = pPinInfoList;
    }
    return;
}


