
/* tpTilePrim.c --
 *
 *     George Carvalho
 *     June Wang
 *     (Advising Professor - E. S. Kuh)
 *
 */

#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/BearFP-1.1/src/tileProp/tpTilePrim.c,v 1.1.1.1 2002/12/31 15:51:11 saj Exp $";
#endif  not lint

#include "bearCopyRight.h"
#include <stdio.h>
#include "magic.h"
#include "malloc.h"
#include "geometry.h"
#include "bear.h"
#include "tile.h"
#include "fpg.h"
#include "gtGraph.h"
#include "tileProp.h"
#include "tilePropInt.h"
#include "mac.h"

/* ---------- Tile Property Primitives ---------- */

bool
Tp_IsSpaceTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsSpaceTile() failed : no tile is given");
    }
#endif DEBUG

    return (TiGetType(tilePtr) == (TileType) TT_SPACE);
}

bool
Tp_IsSolidTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsSolidTile() failed : no tile is given");
    }
#endif DEBUG

    return (TiGetType(tilePtr) > (TileType) 0);
/*
    return (TiGetType(tilePtr) == (TileType) TT_SOLID);
*/
}

bool
Tp_IsTileFilled(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsTileFilled() failed : no tile given");
    } 
#endif DEBUG

    return ((tilePtr->ti_client != (ClientData) MINFINITY) ? TRUE : FALSE);
}

bool
Tp_IsSpacePropTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsSpacePropTile() failed : no tile given");
    }
    else if (Tp_IsTileFilled(tilePtr) == FALSE) {
       TpPrintFatalErr("Tp_IsSpacePropTile() failed: tile has not been filled in");
    }
#endif DEBUG

    return ((TILE_PROP_PTR->unType == TP_SPACE_PROP) ? TRUE : FALSE);
}

bool
Tp_IsCellInstTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsCellInstTile() failed : no tile is given");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unType == TP_CELL_INST);
}

bool
Tp_IsDCellTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsDCellTile() failed : no tile is given");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unType == TP_DUMMY_CELL);
}

bool
Tp_IsRCellTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsRCellTile() failed : no tile is given");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unType == TP_ROUTE_CELL);
}

bool
Tp_IsRingTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsRingTile() failed : no tile is given");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unType == TP_RING);
}

bool
Tp_IsHorzTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsHorzTile() failed : no tile is given");
    }
#endif DEBUG

    return ((Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) & (TP_HORZ_BN_TILE |
        TP_HORZ_DOM_TILE | TP_HORZ_DOM_SBN_TILE |
        TP_HORZ_GEN_TILE | TP_HORZ_GEN_SBN_TILE))) ? TRUE : FALSE);
}

bool
Tp_IsVertTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsVertTile() failed : no tile is given");
    }
#endif DEBUG

    return ((Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) & (TP_VERT_BN_TILE |
        TP_VERT_DOM_TILE | TP_VERT_DOM_SBN_TILE |
        TP_VERT_GEN_TILE | TP_VERT_GEN_SBN_TILE))) ? TRUE : FALSE);
}

bool
Tp_IsBottleneckTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsBottleneckTile() failed : no tile is given");
    }
#endif DEBUG

    if (Tp_IsSpaceTile(tilePtr) == FALSE) {
        return (FALSE);
    }
    else if (Tp_IsSpacePropTile(tilePtr) == FALSE) {
        return (FALSE);
    }

    return ((Tp_GetTypeOfSpaceTile(tilePtr) &
        (TP_HORZ_BN_TILE | TP_VERT_BN_TILE)) ? TRUE : FALSE);
}

bool
Tp_IsSBNTile(tilePtr)
Tile *tilePtr; 
/*
 * Returns TRUE if this is a semi bottleneck tile, FALSE
 * otherwise.
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsSBNTile() failed : no tile is given");
    }
#endif DEBUG

    return ((Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) &
        (TP_HORZ_DOM_SBN_TILE | TP_VERT_DOM_SBN_TILE |
        TP_HORZ_GEN_SBN_TILE | TP_VERT_GEN_SBN_TILE))) ? TRUE : FALSE);
}

bool
Tp_IsDominantTile(tilePtr)
Tile *tilePtr; 
/*
 * Returns TRUE if this is any type of dominant tile, FALSE
 * otherwise.
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsDominantTile() failed : no tile is given");
    }
#endif DEBUG

    return ((Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) &
        (TP_HORZ_DOM_TILE | TP_VERT_DOM_TILE |
        TP_HORZ_DOM_SBN_TILE | TP_VERT_DOM_SBN_TILE))) ? TRUE : FALSE);
}

bool
Tp_IsDominantSBNTile(tilePtr)
Tile *tilePtr; 
/*
 * Returns TRUE if this is a dominant semi bottleneck tile, FALSE
 * otherwise.
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsDominantSBNTile() failed : no tile is given");
    }
#endif DEBUG

    return (Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) &
        (TP_HORZ_DOM_SBN_TILE | TP_VERT_DOM_SBN_TILE)));
}



bool
Tp_IsGenericTile(tilePtr)
Tile *tilePtr; 
/*
 * Returns TRUE if this is any type of generic tile, FALSE
 * otherwise.
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsGenericTile() failed : no tile is given");
    }
#endif DEBUG

    return (Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) &
        (TP_HORZ_GEN_TILE | TP_VERT_GEN_TILE |
        TP_HORZ_GEN_SBN_TILE | TP_VERT_GEN_SBN_TILE)));
}



bool
Tp_IsGenericSBNTile(tilePtr)
Tile *tilePtr; 
/*
 * Returns TRUE if this is a generic semi bottleneck tile, FALSE
 * otherwise.
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_IsGenericSBNTile() failed : no tile is given");
    }
#endif DEBUG

    return (Tp_IsSpacePropTile(tilePtr) &&
        (Tp_GetTypeOfSpaceTile(tilePtr) &
        (TP_HORZ_GEN_SBN_TILE | TP_VERT_GEN_SBN_TILE)));
}



bool
Tp_IsPairOfSBNTiles(tlTilePtrPtr, brTilePtrPtr, bnLineTypePtr)

Tile        **tlTilePtrPtr, **brTilePtrPtr; 
BNLineType  *bnLineTypePtr;
/*
 * Check to see if '*tlTilePtrPtr' and '*brTilePtrPtr' are a pair of
 * semi-bottleneck tiles.
 *
 * Note:
 *     As long as '*tlTilePtrPtr' and '*brTilePtrPtr' are semi-bottleneck
 *     tiles they don't have to be handed to this routine in any
 *     particular order.
 *
 * Results:
 *    If ('*tlTilePtrPtr' and '*brTilePtrPtr' are in fact a pair of
 *        semi-bottleneck tiles) {
 *
 *       The tiles will be switched if necessary so that '*tlTilePtrPtr'
 *       will be to the top (left) of '*brTilePtrPtr'.
 *
 *       '*bnLineTypePtr' will then have one of the following values
 *       TP_HORZ_B_LEFT, TP_HORZ_B_RIGHT, TP_VERT_R_DOWN, or TP_VERT_R_UP.
 *
 *       TRUE will be returned.
 *    } 
 *    else {
 *
 *       The tiles will not be switched.
 *
 *       '*bnLineTypePtr' will have the value TP_NULL_BNL.
 *
 *       FALSE will be returned.
 *    }
 *
 * Side effects:
 *    None.
 *
 */
{
#define RETURN(returnValue) \
{ \
    if (returnValue) { \
        if (switchTiles) { \
            tmpTilePtr = *tlTilePtrPtr; \
            *tlTilePtrPtr = *brTilePtrPtr; \
            *brTilePtrPtr = tmpTilePtr; \
        } \
        return (TRUE); \
    } \
    else { \
        return (FALSE); \
    } \
}

    Rect rightRect, leftRect, topRect, botRect, tmpRect; 
    Tile *tmpTilePtr;
    bool horzTiles, vertTiles;
    bool switchTiles = FALSE;

#ifdef DEBUG
    if (tlTilePtrPtr == (Tile **) NULL || brTilePtrPtr == (Tile **) NULL ||
        *tlTilePtrPtr == NIL(Tile) || *brTilePtrPtr == NIL(Tile) ||
        bnLineTypePtr == NIL(BNLineType)) {
        TpPrintFatalErr("Tp_IsPairOfSBNTiles() failed");
    }
#endif DEBUG

    *bnLineTypePtr = TP_NULL_BNL;
    horzTiles = Tp_IsHorzTile(*tlTilePtrPtr) && Tp_IsHorzTile(*brTilePtrPtr);
    vertTiles = Tp_IsVertTile(*tlTilePtrPtr) && Tp_IsVertTile(*brTilePtrPtr);
    if (horzTiles == FALSE && vertTiles == FALSE) {
         RETURN(FALSE);
    }

    if (horzTiles) {
        /*
         * The tiles are on the horizontal plane.
         */
        TITORECT(*tlTilePtrPtr, &topRect);
        TITORECT(*brTilePtrPtr, &botRect);
        if (topRect.r_ybot != botRect.r_ytop) {
            if (topRect.r_ytop == botRect.r_ybot) {
                Mac_CopyRect(&tmpRect, &topRect);
                Mac_CopyRect(&topRect, &botRect);
                Mac_CopyRect(&botRect, &tmpRect);
                switchTiles = TRUE;
            } 
            else {
                RETURN(FALSE);
            }
        }

        if (botRect.r_xbot < topRect.r_xbot &&
            botRect.r_xtop > topRect.r_xbot &&
            botRect.r_xtop < topRect.r_xtop) {
            *bnLineTypePtr = TP_HORZ_B_LEFT;
            RETURN(TRUE);
        } 
        else if (botRect.r_xbot > topRect.r_xbot &&
            botRect.r_xbot < topRect.r_xtop &&
            botRect.r_xtop > topRect.r_xtop) {
            *bnLineTypePtr = TP_HORZ_B_RIGHT;
            RETURN(TRUE);
        }
    } 
    else {
        /*
         * The tiles are on the vertical plane.
         */
        TITORECT(*tlTilePtrPtr, &tmpRect);
        Tp_VRectToHRect(&tmpRect, &leftRect);
        TITORECT(*brTilePtrPtr, &tmpRect);
        Tp_VRectToHRect(&tmpRect, &rightRect);

        if (leftRect.r_xtop != rightRect.r_xbot) {
            if (leftRect.r_xbot == rightRect.r_xtop) {
                Mac_CopyRect(&tmpRect, &leftRect);
                Mac_CopyRect(&leftRect, &rightRect);
                Mac_CopyRect(&rightRect, &tmpRect);
                switchTiles = TRUE;
            } 
            else {
                RETURN(FALSE);
            }
        }

        if (rightRect.r_ybot < leftRect.r_ybot &&
            rightRect.r_ytop > leftRect.r_ybot &&
            rightRect.r_ytop < leftRect.r_ytop) {
            *bnLineTypePtr = TP_VERT_R_DOWN;
            RETURN(TRUE);
        } 
        else if (rightRect.r_ybot > leftRect.r_ybot &&
            rightRect.r_ybot < leftRect.r_ytop &&
            rightRect.r_ytop > leftRect.r_ytop) {
            *bnLineTypePtr = TP_VERT_R_UP;
            RETURN(TRUE);
        }
    }
    RETURN(FALSE);
}



void
Tp_MarkTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_MarkTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->mark = BEAR_MARKED;
}

void
Tp_UnmarkTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_UnmarkTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->mark = BEAR_UNMARKED;
}



CellInst *
Tp_GetCellInstOfTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSpaceTile(tilePtr) ||
        (TILE_PROP_PTR->unType != TP_CELL_INST)) {
        TpPrintFatalErr("Tp_GetCellInstOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.cellInstPtr);
}

void
Tp_PutCellInstOfTile(cellInstPtr, tilePtr)
CellInst    *cellInstPtr;
Tile        *tilePtr;
{
#ifdef DEBUG
    if ((cellInstPtr == NIL(CellInst)) || (tilePtr == NIL(Tile)) ||
        Tp_IsSpaceTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutCellInstOfTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->unType = TP_CELL_INST;
    TILE_PROP_PTR->unValue.cellInstPtr = cellInstPtr;
}



DummyCell *
Tp_GetDCellOfTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSpaceTile(tilePtr) ||
        (TILE_PROP_PTR->unType != TP_DUMMY_CELL)) {
        TpPrintFatalErr("Tp_GetDCellOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.dummyCellPtr);
}

void
Tp_PutDCellOfTile(dcellPtr, tilePtr)
DummyCell   *dcellPtr;
Tile        *tilePtr;
{
#ifdef DEBUG
    if ((dcellPtr == NIL(DummyCell)) || (tilePtr == NIL(Tile)) ||
        Tp_IsSpaceTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutDCellOfTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->unType = TP_DUMMY_CELL;
    TILE_PROP_PTR->unValue.dummyCellPtr = dcellPtr;
}



RouteCell *
Tp_GetRCellOfTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSpaceTile(tilePtr) ||
        (TILE_PROP_PTR->unType != TP_ROUTE_CELL)) {
        TpPrintFatalErr("Tp_GetRCellOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.routeCellPtr);
}

void
Tp_PutRCellOfTile(rcellPtr, tilePtr)
RouteCell   *rcellPtr;
Tile        *tilePtr;
{
#ifdef DEBUG
    if ((rcellPtr == NIL(RouteCell)) || (tilePtr == NIL(Tile)) ||
        Tp_IsSpaceTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutRCellOfTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->unType = TP_ROUTE_CELL;
    TILE_PROP_PTR->unValue.routeCellPtr = rcellPtr;
}



Ring *
Tp_GetRingOfTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSpaceTile(tilePtr) ||
        (TILE_PROP_PTR->unType != TP_RING)) 
    {
        TpPrintFatalErr("Tp_GetRingOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.ringPtr);
}

void
Tp_PutRingOfTile(ringPtr, tilePtr)
Ring    *ringPtr;
Tile    *tilePtr;
{
#ifdef DEBUG
    if ((ringPtr == NIL(Ring)) || (tilePtr == NIL(Tile)) ||
        Tp_IsSpaceTile(tilePtr)) 
    {
        TpPrintFatalErr("Tp_PutRingOfTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->unType = TP_RING;
    TILE_PROP_PTR->unValue.ringPtr = ringPtr;
}



SpaceProp *
Tp_GetSpacePropOfTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_GetSpacePropOfTile() failed : no tile given");
    } 
    else if (Tp_IsSpaceTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_GetSpacePropOfTile() failed : not a space tile");
    } 
    else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_GetSpacePropOfTile() failed: tile has not been filled in");
    } 
    else if (TILE_PROP_PTR->unType != TP_SPACE_PROP) {
        TpPrintFatalErr("Tp_GetSpacePropOfTile() failed : no space properties");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr);
}

void
Tp_PutSpacePropOfTile(spacePropPtr, tilePtr)
SpaceProp   *spacePropPtr;
Tile        *tilePtr;
{
#ifdef DEBUG
    if ((spacePropPtr == NIL(SpaceProp)) || (tilePtr == NIL(Tile)) ||
        Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutSpacePropOfTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->unType = TP_SPACE_PROP;
    TILE_PROP_PTR->unValue.spacePropPtr = spacePropPtr;
}



int
Tp_GetTypeOfSpaceTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_GetTypeOfSpaceTile() failed : no tile is given");
    } else if (TILE_PROP_PTR->unType != TP_SPACE_PROP) {
   TpPrintFatalErr("Tp_GetTypeOfSpaceTile() failed : tile is not a space tile");
    }
#endif DEBUG

    return ((Tp_GetSpacePropOfTile(tilePtr))->type);
}

void
Tp_PutTypeOfSpaceTile(type, tilePtr)
int     type;
Tile    *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile) || Tp_IsSolidTile(tilePtr) ||
        (TILE_PROP_PTR->unType != TP_SPACE_PROP)) {
        TpPrintFatalErr("Tp_PutTypeOfSpaceTile() failed");
    } else if ((type & TP_VALID_SPACE_TILE) == FALSE) {
    TpPrintFatalErr("Tp_PutTypeOfSpaceTile() failed : unknown space tile type");
    }
#endif DEBUG

    (Tp_GetSpacePropOfTile(tilePtr))->type = type;
}



BearString
Tp_GetTypeNameOfTile(tilePtr)
Tile *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_GetTypeNameOfTile() failed");
    }
#endif DEBUG

    switch (TILE_PROP_PTR->unType) {
    case TP_CELL_INST:
        return ("cell instance tile");
    case TP_DUMMY_CELL:
        return ("dummy cell tile");
    case TP_ROUTE_CELL:
        return ("route cell tile");
    case TP_RING:
        return ("ring tile");

    case TP_SPACE_PROP:
        switch (Tp_GetTypeOfSpaceTile(tilePtr)) {
        case TP_HORZ_BN_TILE:
            return ("horizontal bottleneck tile");
        case TP_VERT_BN_TILE:
            return ("vertical bottleneck tile");
        case TP_HORZ_DOM_TILE:
            return ("horizontal dominant tile");
        case TP_VERT_DOM_TILE:
            return ("vertical dominant tile");
        case TP_HORZ_DOM_SBN_TILE:
            return ("horizontal dominant semi bottleneck tile");
        case TP_VERT_DOM_SBN_TILE:
            return ("vertical dominant semi bottleneck tile");
        case TP_HORZ_GEN_TILE:
            return ("horizontal generic tile");
        case TP_VERT_GEN_TILE:
            return ("vertical generic tile");
        case TP_HORZ_GEN_SBN_TILE:
            return ("horizontal generic semi bottleneck tile");
        case TP_VERT_GEN_SBN_TILE:
            return ("vertical generic semi bottleneck tile");
        default:
            return (NIL(char));
        }

    default:
        return (NIL(char));
    }
    /* NOTREACHED */
}



BearString
Tp_GetTypeNameOfBnLine(bnLinePtr)
BNLine *bnLinePtr;
{
#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) {
        TpPrintFatalErr("Tp_GetTypeNameOfBnLine() failed : no bn line given");
    }
#endif DEBUG

    switch (bnLinePtr->bnLineType) {
    case TP_VERT_R_UP:
        return ("vertical-right-up bottleneck line");
    case TP_VERT_R_DOWN:
        return ("vertical-right-down bottleneck line");
    case TP_HORZ_B_RIGHT:
        return ("horizontal-bottom-right bottleneck line");
    case TP_HORZ_B_LEFT:
        return ("horizontal-bottom-left bottleneck line");
    case TP_NULL_BNL:
        return ("undefined bottleneck line");
    default:
        return (NIL(char));
    }
    /* NOTREACHED */
}



void
Tp_PutHorzPlaneBnd(planeDataPtr, horzPlaneBndPtr)
PlaneData *planeDataPtr;
Rect *horzPlaneBndPtr;
{
#ifdef DEBUG
    if ((planeDataPtr == NIL(PlaneData)) || (horzPlaneBndPtr == NIL(Rect))) {
        TpPrintFatalErr("Tp_PutHorzPlaneBnd() failed");
    }
#endif DEBUG

    Mac_CopyRect(&planeDataPtr->pd_horzPlaneBnd, horzPlaneBndPtr);
}

void
Tp_PutVertPlaneBnd(planeDataPtr, vertPlaneBndPtr)
PlaneData *planeDataPtr;
Rect *vertPlaneBndPtr;
{
#ifdef DEBUG
    if ((planeDataPtr == NIL(PlaneData)) || (vertPlaneBndPtr == NIL(Rect))) {
        TpPrintFatalErr("Tp_StorePlaneBnd() failed");
    }
#endif DEBUG

    Mac_CopyRect(&planeDataPtr->pd_vertPlaneBnd, vertPlaneBndPtr);
}

int
Tp_GetIdOfTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_GetIdOfTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_GetIdOfTile() failed: tile is not a space tile");
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_GetIdOfTile() failed: tile has not been filled in");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->id);
}

void
Tp_PutIdOfTile(id, tilePtr)
BearId  id;
Tile    *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_PutIdOfTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutIdOfTile() failed: tile is not a space tile");
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_PutIdOfTile() failed: tile has not been filled in");
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->id = id;
}

int
Tp_GetJunctTableIndexOfDomTile(tilePtr)
Tile *tilePtr;
/*
 * get the junction table index of a dominant tile. 
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
  TpPrintFatalErr("Tp_GetJunctTableIndexOfDomTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_GetJunctTableIndexOfDomTile() failed: tile is not a space tile");
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_GetJunctTableIndexOfDomTile() failed: tile has not been filled in");
    } else if (Tp_IsDominantTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_GetJunctTableIndexOfDomTile() failed: must be a dominant tile");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->junctTableIndex);
}

void
Tp_PutJunctTableIndexOfDomTile(junctTableIndex, tilePtr)
int  junctTableIndex;
Tile *tilePtr;
/*
 * Put the junction table index of a dominant tile. 
 */
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
  TpPrintFatalErr("Tp_PutJunctTableIndexOfDomTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutJunctTableIndexOfDomTile() failed: tile is not a space tile");
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_PutJunctTableIndexOfDomTile() failed: tile has not been filled in");
    } else if (Tp_IsDominantTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_PutJunctTableIndexOfDomTile() failed: must be a dominant tile");
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->junctTableIndex = junctTableIndex;
}



int      
Tp_GetCapacityOfTile(tilePtr)
Tile    *tilePtr; 
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSolidTile(tilePtr))
    {
        TpPrintFatalErr("Tp_GetCapacityOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->capacity);
}

void
Tp_PutCapacityOfTile(capacity, tilePtr)
int     capacity;
Tile    *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_PutCapacityOfTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        Bear_PrintErr("Tp_PutCapacityOfTile() failed: tile is not a space tile");
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        sprintf(bear_ErrorMsg, "%s",
            "Tp_PutCapacityOfTile() failed: tile has not been filled in");
        TpPrintFatalErr(bear_ErrorMsg);
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->capacity = capacity;
}



int
Tp_GetDensityOfTile(tilePtr)
Tile    *tilePtr; 
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSolidTile(tilePtr))
    {
        TpPrintFatalErr("Tp_GetDensityOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->density);
}


void
Tp_PutDensityOfTile(density, tilePtr)
int     density;
Tile    *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_PutDensityOfTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        sprintf(bear_ErrorMsg, "%s",
            "Tp_PutDensityOfTile() failed: tile is not a space tile");
        TpPrintFatalErr(bear_ErrorMsg);
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        sprintf(bear_ErrorMsg, "%s",
            "Tp_PutDensityOfTile() failed: tile has not been filled in");
        TpPrintFatalErr(bear_ErrorMsg);
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->density = density;
}



int
Tp_GetRidgeWidthOfTile(tilePtr)
Tile    *tilePtr; 
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_GetRidgeWidthOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->ridgeWidth);
}

void
Tp_PutRidgeWidthOfTile(ridgeWidth, tilePtr)
int     ridgeWidth;
Tile    *tilePtr; 
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_PutRidgeWidthOfTile() failed: tilePtr is NULL");
    } else if (Tp_IsSolidTile(tilePtr)) {
        sprintf(bear_ErrorMsg, "%s",
            "Tp_PutRidgeWidthOfTile() failed: tile is not a space tile");
        TpPrintFatalErr(bear_ErrorMsg);
    } else if (Tp_IsTileFilled(tilePtr) == FALSE) {
        sprintf(bear_ErrorMsg, "%s",
            "Tp_PutRidgeWidthOfTile() failed: tile has not been filled in");
        TpPrintFatalErr(bear_ErrorMsg);
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->ridgeWidth = ridgeWidth;   
}



Tile *
Tp_GetPrevTileOfTile(tilePtr)
Tile    *tilePtr; 
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_GetPrevTileOfTile() failed");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->prevTilePtr);
}

void
Tp_PutPrevTileOfTile(prevTilePtr, tilePtr)
Tile    *prevTilePtr, *tilePtr; 
{
#ifdef DEBUG
    if ((tilePtr == NIL(Tile)) || Tp_IsSolidTile(tilePtr)) {
        TpPrintFatalErr("Tp_PutPrevTileOfTile() failed");
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->prevTilePtr = prevTilePtr;   
}



WallSegment *
Tp_GetWallSegmentOfBnTile(tilePtr)
Tile *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_GetWallSegmentOfBnTile() failed : no tile given");
    } else if (Tp_IsBottleneckTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_GetWallSegmentOfBnTile() failed : not a bn tile");
    }
#endif DEBUG

    return (TILE_PROP_PTR->unValue.spacePropPtr->segPtr);   
}

void
Tp_PutWallSegmentOfBnTile(segPtr, tilePtr)
WallSegment *segPtr;
Tile        *tilePtr;
{
#ifdef DEBUG
    if (tilePtr == NIL(Tile)) {
        TpPrintFatalErr("Tp_PutWallSegmentOfBnTile() failed : no tile given");
    } else if (Tp_IsBottleneckTile(tilePtr) == FALSE) {
        TpPrintFatalErr("Tp_PutWallSegmentOfBnTile() failed : not a bn tile");
    }
#endif DEBUG

    TILE_PROP_PTR->unValue.spacePropPtr->segPtr = segPtr;
}



int
Tp_GetCapacityOfBnLine(bnLinePtr)
BNLine *bnLinePtr;
{
#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) 
    {
    TpPrintFatalErr("Tp_GetCapacityOfBnLine() failed : no bn line given");
    }
#endif DEBUG

    return (bnLinePtr->capacity);
}


void
Tp_PutCapacityOfBnLine(capacity, bnLinePtr)
int    capacity;
BNLine *bnLinePtr;
{
#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) 
    {
    TpPrintFatalErr("Tp_PutCapacityOfBnLine() failed : no bn line given");
    }
#endif DEBUG

    bnLinePtr->capacity = capacity;
}



int
Tp_GetDensityOfBnLine(bnLinePtr)
BNLine *bnLinePtr;
{
#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) 
    {
    TpPrintFatalErr("Tp_GetDensityOfBnLine() failed : no bn line given");
    }
#endif DEBUG

    return (bnLinePtr->density);
}

void
Tp_PutDensityOfBnLine(density, bnLinePtr)
int    density;
BNLine *bnLinePtr;
{
#ifdef DEBUG
    if (bnLinePtr == NIL(BNLine)) 
    {
    TpPrintFatalErr("Tp_PutDensityOfBnLine() failed : no bn line given");
    }
#endif DEBUG

    bnLinePtr->density = density;
}



