#include "port.h"
#include "mag2oct.h"
#include "oct.h"

subcell(facet, filename, fp, buf)
	octObject	*facet;
	char		*filename;
	FILE		*fp;
	char		*buf;
{
	char		*p;
	char		nbuf[BUFSIZ];
	int		array;
	octCoord	arrayinfo[6];
	octCoord	a, b, c, d, e, f;
	octObject	instance;
	octObject	ibag;
	octObject	timestamp;
	char		*octfile;
	extern	char	*getinput();
	extern	char	*input();

/****/	buf += 4;			/* Through away the "use " */
	if ((p = strchr(buf, ' ')) || (p = strchr(buf, '\n')))
		(void) strcpy(p, ".mag");

	if (getinput(nbuf, sizeof nbuf, fp) == NULL) {
		logmsg(-1, "%s: Premature EOF\n", filename);
		return;
	}

	if (! strncmp("array", nbuf, 5)) {
		if (sscanf(nbuf, "array %d %d %d %d %d %d",
				&arrayinfo[0],
				&arrayinfo[1],
				&arrayinfo[2],
				&arrayinfo[3],
				&arrayinfo[4],
				&arrayinfo[5]) != 6) {
			logmsg(-1, "%s: Malformed array line ignored: %s",
					filename, nbuf);
			array = 0;
		} else {
			array = 1;
			arrayinfo[2] *= LAM_SCALE;
			arrayinfo[5] *= LAM_SCALE;
		}

		if (getinput(nbuf, sizeof nbuf, fp) == NULL) {
			logmsg(-1, "%s: Premature EOF\n", filename);
			return;
		}

	} else {
		array = 0;
	}

	if (! strncmp("timestamp", nbuf, 9)) {
		timestamp.type = OCT_PROP;
		timestamp.contents.prop.name = "TIMESTAMP";
		timestamp.contents.prop.type = OCT_INTEGER;
		timestamp.contents.prop.value.integer = atoi(&buf[9]);

		if (getinput(nbuf, sizeof nbuf, fp) == NULL) {
			logmsg(-1, "%s: Premature EOF\n", filename);
			return;
		}
	} else {
		timestamp.type = OCT_UNDEFINED_OBJECT;
	}

	if (strncmp("transform", nbuf, 9)) {
		logmsg(-1, "%s: Expecting transform line after use %s -- got %s\n",
						filename, buf, nbuf);
		return;
	}

	/* Process transform line */
	if (sscanf(nbuf, "transform %d %d %d %d %d %d",
			&a, &b, &c, &d, &e, &f) != 6) {
		logmsg(-1, "%s: Malformed transform line: %s",
				filename, nbuf);
		return;
	}

	if (getinput(nbuf, sizeof nbuf, fp) == NULL) {
		logmsg(-1, "%s: Premature EOF\n", filename);
		return;
	}

	if (strncmp("box", nbuf, 3)) {
		logmsg(-1, "%s: Expecting box line after use %s -- got %s\n",
						filename, buf, nbuf);
		return;
	}
/****/	/* Ignore box line */

	/* Make new cell */
	if ((octfile = input(buf)) == NULL)
		return;

	/* Attach the new cell */
	instance.type = OCT_INSTANCE;
	instance.contents.instance.transform.translation.x = c * LAM_SCALE;
	instance.contents.instance.transform.translation.y = f * LAM_SCALE;

	instance.contents.instance.transform.generalTransform[0][0] = a;
/****/	instance.contents.instance.transform.generalTransform[0][1] = d; /****/
/****/	instance.contents.instance.transform.generalTransform[1][0] = b; /****/
	instance.contents.instance.transform.generalTransform[1][1] = e;

#define ITT instance.contents.instance.transform.transformType

	ITT = OCT_FULL_TRANSFORM;			/* the default */
	if (a && e && !(b || d)) {
		switch (a) {
		case 1:
			switch (e) {
			case 1:				/*  x ->  x, y ->  y */
				ITT = OCT_NO_TRANSFORM;
				break;
			case -1:			/*  x ->  x, y -> -y */
				ITT = OCT_MIRROR_Y;
				break;
			}
			break;
		case -1:
			switch (e) {
			case 1:				/*  x -> -x, y ->  y */
				ITT = OCT_MIRROR_X;
				break;
			case -1:			/*  x -> -x, y -> -y */
				ITT = OCT_ROT180;
				break;
			}
			break;
		}
	} else if (b && d && !(a || e)) {
		switch (b) {
		case 1:
			switch (d) {
			case 1:				/*  x ->  y, y ->  x */
				ITT = OCT_MY_ROT90;
				break;
			case -1:			/*  x ->  y, y -> -x */
				ITT = OCT_ROT270;
				break;
			}
			break;
		case -1:
			switch (d) {
			case 1:				/*  x -> -y, y ->  x */
				ITT = OCT_ROT90;
				break;
			case -1:			/*  x -> -y, y -> -x */
				ITT = OCT_MX_ROT90;
				break;
			}
			break;
		}
	}

/**	instance.contents.instance.name = octfile; **/
	instance.contents.instance.name = (char *)NULL;
	instance.contents.instance.master = octfile;
	instance.contents.instance.view = "physical";
	instance.contents.instance.facet = "contents";
	instance.contents.instance.version = OCT_CURRENT_VERSION;

	ibag.type = OCT_BAG;
	ibag.contents.bag.name = "INSTANCES";

	if (octGetOrCreate(facet, &ibag) != OCT_OK) {
		logmsg(-1, "Can not create instance bag: %s\n", octErrorString());
		return;
	}

	if (octCreate(&ibag, &instance) != OCT_OK) {
		logmsg(-1, "Can't create instance in INSTANCES bag: %s\n",
		    octErrorString());
		return;
	}

	if (octAttach(facet, &instance) != OCT_OK) {
		logmsg(-1, "Can not attach oct instance to %s: %s\n",
		    octfile, octErrorString());
		return;
	}

	if (timestamp.type == OCT_PROP)
		if (octCreate(&instance, &timestamp) != OCT_OK)
			logmsg(-1, "Instance TIMESTAMP: %s\n", octErrorString());

	if (array) {
		octObject	prop;
		octCoord	xstart;
		int		x, y;

		prop.type = OCT_PROP;
		prop.contents.prop.name = "array";
		prop.contents.prop.type = OCT_STRANGER;
		prop.contents.prop.value.stranger.length = sizeof arrayinfo;
		prop.contents.prop.value.stranger.contents = (char *) arrayinfo;

		if (octCreate(&instance, &prop) != OCT_OK) {
			logmsg(-1, "Can not create array prop: %s\n",
					octErrorString());
			/* Continue on anyway... */
		}

#define	IT instance.contents.instance.transform.translation

		xstart = IT.x;
		IT.x += arrayinfo[2];		/* Skip the first element */
		x = arrayinfo[0] + 1;		/* Skip the first element */

		for (y = arrayinfo[3]; y <= arrayinfo[4]; y++) {
			while (x <= arrayinfo[1]) {

/* Start of code stolen from above */
	if (octCreate(&ibag, &instance) != OCT_OK) {
		logmsg(-1, "Can't create instance in ITNSTANCES bag: %s\n",
		    octErrorString());
		return;
	}

	if (octAttach(facet, &instance) != OCT_OK) {
		logmsg(-1, "Can not attach oct instance to %s: %s\n",
		    octfile, octErrorString());
		return;
	}

	if (timestamp.type == OCT_PROP)
		if (octCreate(&instance, &timestamp) != OCT_OK)
			logmsg(-1, "ITnstance TIMESTAMP: %s\n", octErrorString());
/* End of code stolen from above */

				if (octAttach(&instance, &prop) != OCT_OK) {
					logmsg(-1,
					    "Can not create array prop: %s\n",
					    octErrorString());
				}

				x++;
				IT.x += arrayinfo[2];
			}
			x = arrayinfo[0];
			IT.x = xstart;
			IT.y += arrayinfo[5];
		}
	}
}
