#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "options.h"
#include "oct.h"
#include "oh.h"
#include "octmm.h"
#include "vov.h"

#include <errno.h>

#define USAGE 

#define FPRINTF	(void) fprintf

FILE *output; 
char *optProgName;
static char *scale = NIL(char);
static char *fname = "scmos/octmm-scripts/to-magic-gen";

int next_cell = 1;

int force_bad_timestamp = 0;
int hierarchy_level = 1;
int max_hierarchy = 1000;
int strip_octflat_instances = 0;
char *layout_directory = NIL(char);

optionStruct optionList[] = {
    { "f",	"mod_file",	"octmm modification script" },
    { "s",	"scale",	"amount to scale the design" },
    { "t",        0,              "Force a mismatch of MAGIC timestamps"},
    { "n",        0,              "Produces only the top level of the cell"},
    { "i",        0,              "Strip compound octflat instance names"},
    { "l",        "layout dir",   "Put the resulting MAGIC files here"},
    { OPT_RARG,	"cell:view[:facet]",	"facet defaults to `contents'" },
    { 0,	0,		0},
};


main(argc, argv)
int argc;
char **argv;
{
    extern int optind;
    extern char *optarg;
    int option;
    octObject top_level;
    octStatus status;

    output = stdout;
    
    while ((option = optGetOption(argc, argv)) != EOF) {
	switch (option) {
	case 'f': fname = util_strsav(optarg); break;
	case 's': scale = util_strsav(optarg); break;
	case 't': force_bad_timestamp = 1; break;
	case 'n': max_hierarchy = 1; break;
	case 'i': strip_octflat_instances = 1; break;
	case 'l': layout_directory = util_strsav(optarg); break;
	default: optUsage();
	}
    }

    if (argc != optind+1) optUsage();

    ohUnpackDefaults(&top_level, "r", ":physical:contents");
    if (ohUnpackFacetName(&top_level, argv[optind]) != OCT_OK) {
	optUsage();
    }

    VOVbegin( argc, argv );

    if ( layout_directory && mkdir(layout_directory, 0777) < 0) {
	if (errno != EEXIST) {
	    FPRINTF(stderr, "oct2mag: Couldn't make the layout directory");
	    VOVend(-1);
	}
    }
    
    init_cell_table();

    VOVinputFacet( &top_level );
    if (octOpenFacet(&top_level) < OCT_OK) {
	FPRINTF(stderr, "%s: Can't open %s", optProgName,
		ohFormatName( &top_level ) );
	octError("");
	optUsage();
    }

    if (status == OCT_INCONSISTENT) {
	ohPrintInconsistent(&top_level, stdout);
	VOVend(-1);
    }

    add_cell(&top_level, next_cell);

    convert_facet(&top_level, next_cell++);

    VOVend(0);
}


void do_mask_mod(facet)
    octObject *facet;
{
    oct_mask_mod(optProgName, fname, facet, scale);
    return;
}
