/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSBldEdge.c

Build initial channel graph.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSBldEdge.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  90/02/07  16:10:27  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:39:26  octtools
 * Initial revision
 * 
 * Revision 1.1  88/03/31  11:37:23  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "MIInc.h"
#include "CDSInc.h"

/* Construct maximal edge line segments.
----------------------------------------------------------------------------*/
CDSBldMaxEdge()
{
CDSEdgeDef	*e, *e2;
char		d, d2;
int		l;
CDSLocDef	range, eRange, sRange;
CDSGEdgeDef	*pGEdge, *ge;
Boolean		alreadyDoneFlg;

MI_MOD_START("CDSBldMaxEdge");

for (d=CDS_V; ; d=CDS_H) {
    MIMSG(1, (MIDBG,"Generating graph edges for %s", CDS_DIR_TOSTR(d)));
    d2     = !d;
    pGEdge = NULL;
    CDSHead.numGEdge[d] = 0;

    for (e=CDSEdge[d]; e; e=e->next) {
	if (e->hiLo==CDS_LO || e->doneFlg) continue;

	CDSGetEdgeRange(e, eRange, d2);
	CDS_COPY_LOC(e->extRange, eRange);
	MI_QUEUE(e);
	e->queueFlg = TRUE;

	while (MIQueueNum) {

	    MI_DEQUEUE(CDSEdgeDef*, e2);
	    MIMSG(3, (MIDBG,"Dequeing %d l=%d (%d %d) %s",
		e2->id, e2->loc, e2->range[CDS_LO], e2->range[CDS_HI],
		e2->doneFlg ? "done" : "notDone"));
	    if (e2->doneFlg) continue;

	    e2->doneFlg    = TRUE;
	    alreadyDoneFlg = TRUE;
	    CDS_COPY_LOC(range, e2->extRange);
	    sRange[CDS_LO] = e2->loc;
					/* Scan edges until hit lo bdy */
	    for (e2=e2->next; e2; e2=e2->next) {
					/* Compute range at edge: */
		MIMSG(3, (MIDBG,"On edge %d l=%d (%d %d)",
		    e2->id, e2->loc, e2->range[CDS_LO], e2->range[CDS_HI]));
		l = e2->loc;
		if (e2->queueFlg) {
		    CDS_COPY_LOC(eRange, e2->extRange);
		}
		else
		    CDSGetEdgeRange(e2, eRange, d2);
					/* Out of range */
		if (eRange[CDS_HI]<=range[CDS_LO] ||
		    eRange[CDS_LO]>=range[CDS_HI]) {
		    if (!e2->doneFlg && e2->hiLo==CDS_HI && !e2->queueFlg) {
			CDS_COPY_LOC(e2->extRange, eRange);
			MIMSG(3, (MIDBG,"Queuing this edge"));
			e2->queueFlg = TRUE;
			MI_QUEUE(e2);
		    }
		    continue;
		}
		else {
		    if (e2->doneFlg) break;
		    if (eRange[CDS_LO]<range[CDS_LO]) {
			DASSERT(eRange[CDS_HI]>=range[CDS_HI]);
			CDS_COPY_LOC(range, eRange);
			sRange[CDS_LO] = l;
		    }
		    else if (eRange[CDS_HI]>range[CDS_HI]) {
			DASSERT(eRange[CDS_LO]<=range[CDS_LO]);
			CDS_COPY_LOC(range, eRange);
			sRange[CDS_LO] = l;
		    }
		    else if (eRange[CDS_LO]==range[CDS_LO] &&
			     eRange[CDS_HI]==range[CDS_HI])
			sRange[CDS_HI] = l;
		    else DASSERT(0);
		}

		e2->doneFlg = TRUE;
		if (e2->hiLo==CDS_LO) {
		    alreadyDoneFlg = FALSE;
		    break;
		}
	    }

	    if (!alreadyDoneFlg) {	/* Save maximal extent edge */
		ge = ALLOC_TYPE(CDSGEdgeDef);
		CDSHead.numGEdge[d]++;
		CDS_COPY_LOC(ge->range, range);
		ge->loc  = (sRange[CDS_LO] + sRange[CDS_HI])/2;
		ge->next = NULL;
		if (!pGEdge)
		then CDSGEdge[d]  = ge;
		else pGEdge->next = ge;
		pGEdge = ge;
		MIMSG(2, (MIDBG,"Maximal edge l=%d rg=(%d %d)",
		    ge->loc, ge->range[CDS_LO], ge->range[CDS_HI]));
	    }
	}
    }
    CDSOrderGEdge(&CDSGEdge[d], d);
    if (d==CDS_H) break;
}

MI_MOD_END();
} /* ............................. End of procedure CDSBldMaxEdge */
/* Given edge e:  Find the longest extension of this edge  (find longest line
segment containing edge e that does not pass through the interior of any
object).
----------------------------------------------------------------------------*/
CDSGetEdgeRange(e, eRange, d2)

CDSEdgeDef	*e;			/* Find range of this edge */
CDSLocDef	eRange;			/* Initial range of edge */
char		d2;			/* Orthogonal direction of edge */
{
CDSEdgeDef	*e2, *e3;
int		l, l2;
short		o2;
short		b, nb;
char		d = !d2;

MI_MOD_START("CDSGetEdgeRange");

CDS_COPY_LOC(eRange, e->range);
l = e->loc;
e3 = NULL;
					/* Scan hi orthogonal edges */
MIMSG(4, (MIDBG,"scanning hi edges"));
for (e2=CDSEdge[d2]; e2; e2=e2->next) {
    l2 = e2->loc;
    if (!e3 && l2>eRange[CDS_LO]) e3 = e2;
    MIMSG(5, (MIDBG,"e2: %s %s l=%d (%d %d)",
	CDS_DIR_TOSTR(e2->dir), CDS_HILO_TOSTR(e2->hiLo),
	l2, e2->range[CDS_LO], e2->range[CDS_HI]));
    if (e2->hiLo==CDS_HI ||
	l2<eRange[CDS_HI] ||
	e2->range[CDS_LO]>l ||
	e2->range[CDS_HI]<l) continue;
					/* Pertinent orthogonal edges: */
    MIMSG(5, (MIDBG,"got an edge"));
    if (e2->range[d2]==l) {		/* Hit a corner of obstruction */
	o2 = e2->objNum;		/* Check if edge crossed line at l */
	nb = CDSObj[o2].numBdy;
	b  = (e2->bdyIdx + 1) % nb;
	DASSERT(CDSObj[o2].bdy[b][d]==l);
	if (CDSObj[o2].bdy[b][d2]>l2)
	    continue;
    }

    else if (e2->range[d]==l) {		/* Hit a corner of obstruction */
	o2 = e2->objNum;		/* Check if edge crossed line at l */
	nb = CDSObj[o2].numBdy;
	b  = (e2->bdyIdx - 2 + nb) % nb;
	DASSERT(CDSObj[o2].bdy[b][d]==l);
	if (CDSObj[o2].bdy[b][d2]>l2)
	    continue;
    }

    eRange[CDS_HI] = l2;
    MIMSG(4, (MIDBG,"Hi: Hit edge %d at %d rg=(%d %d)",
	e2->id, l2, e2->range[CDS_LO], e2->range[CDS_HI]));
    break;
}
DASSERT(e2);
DASSERT(e3);
					/* Scan lo orthogonal edges */
MIMSG(4, (MIDBG,"scanning lo edges"));
for (e2=e3->prev; e2; e2=e2->prev) {
    l2 = e2->loc;
    MIMSG(5, (MIDBG,"e2: %s %s l=%d (%d %d)",
	CDS_DIR_TOSTR(e2->dir), CDS_HILO_TOSTR(e2->hiLo),
	l2, e2->range[CDS_LO], e2->range[CDS_HI]));
    if (e2->hiLo==CDS_LO ||
	e2->range[CDS_LO]>l ||
	e2->range[CDS_HI]<l) continue;
					/* Pertinent edges */
    MIMSG(5, (MIDBG,"got an edge"));
    if (e2->range[d2]==l) {
	o2 = e2->objNum;
	nb = CDSObj[o2].numBdy;
	b  = (e2->bdyIdx - 2 + nb) % nb;
	DASSERT(CDSObj[o2].bdy[b][d]==l);
	if (CDSObj[o2].bdy[b][d2]<l2)
	    continue;
    }

    else if (e2->range[d]==l) {
	o2 = e2->objNum;
	nb = CDSObj[o2].numBdy;
	b  = (e2->bdyIdx + 1) % nb;
	DASSERT(CDSObj[o2].bdy[b][d]==l);
	if (CDSObj[o2].bdy[b][d2]<l2)
	    continue;
    }

    eRange[CDS_LO] = l2;
    MIMSG(4, (MIDBG,"Lo: Hit edge %d at %d rg=(%d %d)",
	e2->id, l2, e2->range[CDS_LO], e2->range[CDS_HI]));
    break;
}
DASSERT(e2);

MIMSG(3, (MIDBG,"Range for edge %d: l=%d (%d %d)",
    e->id, l, eRange[CDS_LO], eRange[CDS_HI]));

MI_MOD_END();
} /* ............................. End of procedure CDSGetEdgeRange */
/* Order edges from left to right & bottom to top (for scan algorithm)
----------------------------------------------------------------------------*/
CDSOrderGEdge(startEdge, d)

CDSGEdgeDef	**startEdge;		/* Edge list to be ordered */
char		d;			/* Direction of edges */
{
int		numEdge;
int		e;
CDSGEdgeDef	**edgeOrder;
CDSGEdgeDef	*sEdge = *startEdge;
CDSGEdgeDef	*pEdge = NULL;
CDSGEdgeDef	*edge;
int		CDSQsortOrderGEdge();

MI_MOD_START("CDSOrderGEdge");

MIMSG(1, (MIDBG,"Sorting %s edges", CDS_DIR_TOSTR(d)));

numEdge = CDSHead.numGEdge[d];

edgeOrder = ALLOC_ARRAY(numEdge, CDSGEdgeDef*);
for (edge=sEdge, e=0; edge; edge=edge->next, e++)
    edgeOrder[e] = edge;

qsort((char *)edgeOrder, numEdge, sizeof(CDSGEdgeDef*), CDSQsortOrderGEdge);

for (e=0; e<numEdge; e++) {
    edge = edgeOrder[e];
    MIMSG(3, (MIDBG,"%d:edge = %d (%d %d)",
	e, edge->loc, edge->range[CDS_LO], edge->range[CDS_HI]));
    if (!pEdge)
	*startEdge = edge;
    else {
	DASSERT(pEdge->loc<=edge->loc);
	pEdge->next = edge;
    }
    pEdge = edge;
}
edge->next = NULL;

FREE_ARRAY(numEdge, edgeOrder, CDSGEdgeDef*);

MI_MOD_END();
} /* ............................. End of procedure CDSOrderGEdge */
/* Edge ordering program for qsort (see CDSOrderGEdge)
----------------------------------------------------------------------------*/
int
CDSQsortOrderGEdge(e1, e2)

CDSGEdgeDef	**e1, **e2;
{
int		d;

d = (*e1)->loc - (*e2)->loc;
if (d) return(d);
if ((*e1)->range[CDS_HI]<=(*e2)->range[CDS_LO]) return(-1);
if ((*e2)->range[CDS_HI]<=(*e1)->range[CDS_LO]) return(1);
DASSERT(0);
return(0);
} /* ............................. End of procedure CDSQsortOrderGEdge */
