/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSGetSlice.c

Slice channels for cut lines

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSGetSlice.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  90/02/07  16:10:29  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:39:27  octtools
 * Initial revision
 * 
 * Revision 1.7  88/05/07  19:44:16  mitch
 * Fixed bug with disappearing real rooms (merging of real rooms with empty
 * ones - should be the other way around).  Bug appears in weird examples
 * where there are L shaped channels separating empty rooms.
 * 
 * Revision 1.1  88/03/31  11:38:08  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "MIInc.h"
#include "CDSInc.h"

#define	COSTB

#define	NODEDLEN(n1, n2) \
    (DELTA((n1)->loc[CDS_X],(n2)->loc[CDS_X])+ \
     DELTA((n1)->loc[CDS_Y],(n2)->loc[CDS_Y]))

#define	CROSSCOST(c, n, d) { \
    short	dA, dB; \
    dA = CDS_NEXTDIR(d, CDS_CCW); \
    dB = CDS_REVDIR(dA); \
    c  = 0; \
    if (((n)->bch[dA].node && !(n)->bch[dB].node) || \
	((n)->bch[dB].node && !(n)->bch[dA].node)) \
	 c = -10; \
    if (n->bch[dA].node) c -= 5; \
    if (n->bch[dB].node) c -= 5; \
}

/* Find best slice through channel graph.
Return codes:	0 found slice through graph,
		1 z or l shaped slice through graph,
		2 no slice found through graph.
----------------------------------------------------------------------------*/
int
CDSFindBestSlice(lowerLeftGNode, numSliceNode, sliceNode, ioHiLoFlg, ioIndex)

CDSGNodeDef	*lowerLeftGNode;
short		*numSliceNode;
CDSGNodeDef	**sliceNode;
Boolean		*ioHiLoFlg;		/* TRUE if ios are on hi side */
short		*ioIndex;		/* -1 if no io side found */
{
CDSGNodeDef	*startGn, *gn, *gn2, *gn3, *gn4, *gnp;
short		startD, d, d2, d3, d4;
short		ori = CDS_CW;
short		rori = CDS_CCW;
short		maxNumSliceNode = *numSliceNode;
short		nsn = 0;
short		nsn2;
short		objNum, objNum2;
Boolean		zShapedFlag = FALSE;
Boolean		extObjFlg1, extObjFlg2;

int		cost, cost0, ccost;
int		dLen, tLen;
int		dLen1, dLen2;
short		k;
int		rc = 0;

int		bestCost;
short		bestSliceDim = -1;
CDSGNodeDef	*bestSliceNode[100];

short		io;			/* Variables for io ring */
Boolean		noMoreIoFlg = TRUE;
Boolean		ioSliceCheck;
short		numIo;

CDSTreeDef	*t;
int		tId;

static Boolean	zChannelsFoundFlg = FALSE;

MI_MOD_START("CDSFindBestSlice");

DASSERT(lowerLeftGNode->type=='e');

CDSStartFindSlice:

*ioIndex = -1;
for (io=0; io<4; io++)			/* Get next ioRingSide to process */
    if (!CDSIo[io].sliceTriedFlg) break;
if (io!=4 && (io==0 || lowerLeftGNode==CDSIo[io-1].bodyGNode)) {
    CDSIo[io].sliceTriedFlg = TRUE;
    CDSIo[io].sliceFoundFlg = FALSE;
    CDSIo[io].bodyGNode = lowerLeftGNode;
    noMoreIoFlg = (io==3);
    *ioIndex = io;
}

startGn = lowerLeftGNode;
startD  = CDS_N;
if (*ioIndex>=0)
    CDSGetStartIoGNode(io, &startGn, &startD, ioHiLoFlg);

bestCost = MAXINT;
gn = startGn->bch[startD].node;		/* Search straight through paths */
					/* Go around perimeter clockwise */
for (d=CDS_NEXTDIR(startD,rori); gn!=startGn; d=CDS_NEXTDIR(d,rori)) {
    for ( ; !gn->bch[d].node; d=CDS_NEXTDIR(d,ori));
    DASSERT(gn->bch[d].type=='e');
					/* Scan internal stuff */
    for (d2=CDS_NEXTDIR(d ,ori); gn->bch[d2].type!='e';
	 d2=CDS_NEXTDIR(d2,ori)) {
	MIMSG(4, (MIDBG,"Scanning on gn=%d (%d %d) dir=%s",
	    CDS_ID(gn), gn->loc[CDS_X], gn->loc[CDS_Y], CDS_CDIR_TOSTR(d2)));
					/* Scan along dir d2 */
	cost = 0;
	for (gn2=gn->bch[d2].node; gn2 && gn2->type!='e';
	     gn2=gn2->bch[d2].node) {
	    CROSSCOST(ccost, gn2, d2);
	    cost += ccost;
	}
	if (gn2 && cost<bestCost) {	/* Split found! */
	    MIMSG(2, (MIDBG,"Straight split dir=%s cost=%d",
		CDS_CDIR_TOSTR(d2), cost));
	    bestCost     = cost;
	    bestSliceDim = 2;
	    bestSliceNode[0] = gn;
	    bestSliceNode[1] = gn2;

	    if (*ioIndex<0 || gn!=startGn->bch[startD].node) continue;

	    MIMSG(0, (MIDBG,"Possible io slice, ioSide=%s",
		CDS_IO_TOSTR(*ioIndex)));

	    /* check if io slice has just io pads on that side */
	    gn2 = gn;
	    ioSliceCheck = TRUE;
	    numIo = 0;
	    do {			/* Go along io slice boundary */
		objNum = gn2->bch[d2].objNum;
		if (objNum>=0) {	/* Object is not an empty room */
					/*   check if it is an io pad */
		    if (!CDSObj[objNum].ioPadFlag)
		    then ioSliceCheck = FALSE;
		    else numIo++;
		}
		gn2 = gn2->bch[d2].node;
	    } while(gn2 && gn2->type!='e');

	    if (numIo==0 || !ioSliceCheck) {
		MIMSG(0, (MIDBG,"Not io slice since numIo=%d and/or nonios=%s",
		    numIo, (ioSliceCheck ? "none" : "some")));
		continue;
	    }

	    MIMSG(0, (MIDBG,"Found io slice"));
	    CDSIo[*ioIndex].sliceFoundFlg = TRUE;
	    goto CDSEndFindSlice;
	}
    }

    gn = gn->bch[d].node;
}

if (*ioIndex>=0) {
    MIMSG(0, (MIDBG,"INFO Did not find %s io side channel slice",
	CDS_IO_TOSTR(*ioIndex)));
    switch(*ioIndex) {
    case 0: tId = CDS_S_EXT_TREEID; break;
    case 1: tId = CDS_N_EXT_TREEID; break;
    case 2: tId = CDS_W_EXT_TREEID; break;
    case 3: tId = CDS_E_EXT_TREEID; break;
    default: DASSERT(0); 
    }
    for (t=CDSTree; t->id!=tId && t; ) {
	if (t->child[CDS_LO]) t = t->child[CDS_LO];
	else		      t = t->child[CDS_HI];
    }
    DASSERT(t);
    t->useExtFlag = TRUE;
}
if (!noMoreIoFlg) goto CDSStartFindSlice;
if (bestCost<MAXINT) goto CDSEndFindSlice;

gn = startGn->bch[startD].node;		/* Search L shape paths */
for (d=CDS_NEXTDIR(startD,rori); gn!=startGn; d=CDS_NEXTDIR(d,rori)) {
    for ( ; !gn->bch[d].node; d=CDS_NEXTDIR(d,ori));

    for (d2=CDS_NEXTDIR(d ,ori); gn->bch[d2].type!='e';
	 d2=CDS_NEXTDIR(d2,ori)) {
					/* Scan along dir d2 */
	cost0 = 5000;
	for (gn2=gn->bch[d2].node;
	     gn2; gn2=gn2->bch[d2].node) {
	    cost = cost0;
	    for (d3=CDS_NEXTDIR(d2,ori); ; d3=CDS_REVDIR(d3)) {
		for (gn3=gn2->bch[d3].node;
		     gn3 && gn3->type!='e'; gn3=gn3->bch[d3].node) {
		    CROSSCOST(ccost, gn3, d3);
		    cost += ccost;
		}
		if (gn3) {
		    dLen1 = NODEDLEN(gn2, gn3);
		    dLen2 = NODEDLEN(gn, gn2);
		    dLen = MIN(dLen1, dLen2);
		    tLen = MAX(dLen1, dLen2);
		    cost += dLen*1000/tLen;
		}
		if (gn3 && cost<bestCost) {	/* Path found */
		    MIMSG(2, (MIDBG,"L shape split dir=%s dir2=%s cost=%d",
			CDS_CDIR_TOSTR(d2), CDS_CDIR_TOSTR(d3), cost));
		    bestCost     = cost;
		    bestSliceDim = 3;
		    bestSliceNode[0] = gn;
		    bestSliceNode[1] = gn2;
		    bestSliceNode[2] = gn3;
		}

		if (d3==CDS_NEXTDIR(d2,rori)) break;
	    }
	    CROSSCOST(ccost, gn2, d2);
	    cost0 += ccost;
	}
    }

    gn = gn->bch[d].node;
}

gn = startGn->bch[startD].node;		/* Search Z shape paths */
for (d=CDS_NEXTDIR(startD,rori); gn!=startGn; d=CDS_NEXTDIR(d,rori)) {
    for ( ; !gn->bch[d].node; d=CDS_NEXTDIR(d,ori));

    for (d2=CDS_NEXTDIR(d ,ori); gn->bch[d2].type!='e';
	 d2=CDS_NEXTDIR(d2,ori)) {
					/* Scan along dir d2 */
	cost0 = 0;
	for (gn2=gn->bch[d2].node;
	     gn2; gn2=gn2->bch[d2].node) {
	    for (d3=CDS_NEXTDIR(d2,ori); ; d3=CDS_REVDIR(d3)) {
		for (gn3=gn2->bch[d3].node;
		     gn3 && gn3->type!='e'; gn3=gn3->bch[d3].node) {
		    d4 = d2;
		    cost = cost0;
		    for (gn4=gn3->bch[d4].node;
			 gn4 && gn4->type!='e'; gn4=gn4->bch[d4].node) {
			CROSSCOST(ccost, gn4, d4);
			cost += ccost;
		    }
		    if (gn4) {
			dLen  = NODEDLEN(gn2, gn3);
			tLen  = NODEDLEN(gn , gn2);
			tLen += NODEDLEN(gn3, gn4);
			cost += dLen*1000/tLen;
		    }
		    if (gn4 && cost<bestCost) {	/* Path found */
			MIMSG(2, (MIDBG,
			    "Z shape split dir=%s dir2=%s dir3=%s cost=%d",
			    CDS_CDIR_TOSTR(d2), CDS_CDIR_TOSTR(d3),
			    CDS_CDIR_TOSTR(d3), cost));
			bestCost     = cost;
			bestSliceDim = 4;
			bestSliceNode[0] = gn;
			bestSliceNode[1] = gn2;
			bestSliceNode[2] = gn3;
			bestSliceNode[3] = gn4;
		    }
		}

		if (d3==CDS_NEXTDIR(d2,rori)) break;
	    }
	    CROSSCOST(ccost, gn2, d2);
	    cost0 += ccost;
	}
    }

    gn = gn->bch[d].node;
}
if (bestCost<MAXINT) goto CDSEndFindSlice;

gn = startGn->bch[startD].node;		/* Search U shaped paths */
for (d=CDS_NEXTDIR(startD,rori); gn!=startGn; d=CDS_NEXTDIR(d,rori)) {
    for ( ; !gn->bch[d].node; d=CDS_NEXTDIR(d,ori));

    bestSliceDim = 1;
    bestSliceNode[0] = gn;

    for (d2=CDS_NEXTDIR(d,ori); d2!=d; d2=CDS_NEXTDIR(d2,ori))
	if (gn->bch[d2].node && gn->bch[d2].type!='e') break;

    if (d2!=d) {
	gn2 = gn;
	MIMSG(0, (MIDBG,"Search U: Slice0 on node %d: (%d %d)",
	    CDS_ID(gn2), gn2->loc[CDS_X], gn2->loc[CDS_Y]));

	while (bestSliceDim<100-1) {
	    gn3 = gn2->bch[d2].node;	/* Scan along dir d2 */
	    for (gn2=gn3; gn2 && gn3->type!='e'; gn2=gn3->bch[d2].node)
		gn3 = gn2;
	    MIMSG(0, (MIDBG,"Search U: Slice%d on node %d: (%d %d)",
		bestSliceDim, CDS_ID(gn3), gn3->loc[CDS_X], gn3->loc[CDS_Y]));
	    bestSliceNode[bestSliceDim++] = gn3;
	    if (gn3->type=='e') break;
					/* Set d3 to next direction */
	    d2 = CDS_REVDIR(d2);
	    for (d3=CDS_NEXTDIR(d2, ori); !gn3->bch[d3].node; d3=CDS_NEXTDIR(d3, ori));
	    if (d3==d2) {
		MIfprintf(MIERR,"ERROR Dangling node in graph\n");
		CDSPrintNGraph(startGn);
	    }

	    gn2 = gn3->bch[d3].node;	/* Scan along dir d3 */
	    for (gn3=gn2; gn3 && gn2->type!='e'; gn3=gn2->bch[d3].node)
		gn2 = gn3;
	    MIMSG(0, (MIDBG,"Search U:  Slice%d on node %d: (%d %d)",
		bestSliceDim, CDS_ID(gn2), gn2->loc[CDS_X], gn2->loc[CDS_Y]));
	    bestSliceNode[bestSliceDim++] = gn2;
	    if (gn2->type=='e') break;
					/* Set d2 to next direction */
	    d3 = CDS_REVDIR(d3);
	    for (d2=CDS_NEXTDIR(d3,ori); !gn2->bch[d2].node; d2=CDS_NEXTDIR(d2, ori));
	    if (d2==d3) {
		MIfprintf(MIERR,"ERROR Dangling node in graph\n");
		CDSPrintNGraph(startGn);
	    }
	}

	if (bestSliceDim>=2) break;
    }

    gn = gn->bch[d].node;
}

if (bestSliceDim>=2) {
    MIMSG(0, (MIDBG,"U slice found"));
}
else
    bestSliceDim = 0;

CDSEndFindSlice:

if (bestSliceDim>0) {			/* Reconstruct best slice */
    for (k=0; k<bestSliceDim-1; k++) {
	gn  = bestSliceNode[k];
	gn2 = bestSliceNode[k+1];
	CDS_GETDIR(d, gn->loc, gn2->loc);
	for (gn3=gn; gn3!=gn2; gn3=gn3->bch[d].node) {
	    DASSERT(gn3);
	    URANGE(nsn, maxNumSliceNode);
	    sliceNode[nsn++] = gn3;
	}
    }
    sliceNode[nsn++] = bestSliceNode[bestSliceDim-1];

    for (nsn2=0; nsn2<nsn; nsn2++) {
	if (nsn2==0 || nsn2==nsn-1) {
	     DASSERT(sliceNode[nsn2]->type=='e');
	}
	else DASSERT(sliceNode[nsn2]->type!='e');
    }

    if (bestSliceDim>2)
	zShapedFlag = TRUE;
}

if (*ioIndex>=0 && !CDSIo[*ioIndex].sliceFoundFlg)
    *ioIndex = -1;

if (nsn) {				/* Print out slice */
    MIMSG(1, (MIDBG,"Slice found numSlice=%d", nsn));
    for (nsn2=0; nsn2<nsn; nsn2++) {
	if ((*MIDebugFlag>=2) ||
	    (*MIDebugFlag>=1 && (nsn2==0 || nsn2==nsn-1)))
	    MIprintf("Slice %d: nid=%d loc=(%d %d)\n",
		nsn2, CDS_ID(sliceNode[nsn2]),
		sliceNode[nsn2]->loc[CDS_X], sliceNode[nsn2]->loc[CDS_Y]);
    }

    extObjFlg1 = extObjFlg2 = FALSE;	/* Check if empty rooms are adjacent */
    for (gn=sliceNode[0], nsn2=1; nsn2<nsn; nsn2++) {
	gnp = gn;
	gn  = sliceNode[nsn2];
	CDS_GETDIR(d, gnp->loc, gn->loc);
	if (gnp->bch[d].objNum>=CDS_NBOBJ_EXTERNAL)
	    extObjFlg1 = TRUE;
	if (gn->bch[CDS_REVDIR(d)].objNum>=CDS_NBOBJ_EXTERNAL)
	    extObjFlg2 = TRUE;
    }

    if ((!extObjFlg1 || !extObjFlg2) && (*ioIndex<0)) {
    	CDSMergeEmptyRooms(nsn, sliceNode, (extObjFlg1 ? 2 : 1));
	nsn = -1;
    }
}
else {					/* No slices found - check if leaf */
    objNum  = startGn->bch[startD].objNum;
    objNum2 = CDS_NBOBJ_NOT_SET;
    gn = startGn->bch[startD].node;
    for (d=CDS_NEXTDIR(startD,rori); gn!=startGn; d=CDS_NEXTDIR(d,rori)) {
	for ( ; !gn->bch[d].node; d=CDS_NEXTDIR(d,ori));
	if (gn->bch[d].objNum!=objNum)
	    objNum2 = gn->bch[d].objNum;
	gn = gn->bch[d].node;
    }

    if (objNum2!=CDS_NBOBJ_NOT_SET) {
	MIfprintf(MIERR,
	    "ERROR Slice not found through tree objNum=%d objNum2=%d\n",
	    objNum, objNum2);
	if (*MIDebugFlag>= -1)
	    CDSPrintNGraph(startGn);
	rc = 2;
    }
}

if (zShapedFlag && nsn>0 && !zChannelsFoundFlg) {
    zChannelsFoundFlg = TRUE;
    MIfprintf(MIDBG,"WARN This chip has \"L\" or \"Z\" shaped channels\n");
    if (*MIDebugFlag>= 0)
	for (k=0; k<bestSliceDim; k++) {
	    gn = bestSliceNode[k];
	    MIprintf("LZSlice %d: octLoc=(%d %d)\n",
		k, gn->loc[CDS_X], gn->loc[CDS_Y]);
	}
    rc = 1;
}

*numSliceNode = nsn;
MI_MOD_END();
return(rc);
} /* ............................. End of procedure CDSFindBestSlice */
/* Start slice search for io ring at proper position.  We assume gn starts
at the lower left hand corner (startGNode) and pd = North.  We go around the
external part of the graph going clockwise and find the proper starting point
for the io ring slice search at a corner determined by xy location.
----------------------------------------------------------------------------*/
CDSGetStartIoGNode(io, startGn, startD, ioHiLoFlg)

short		io;
CDSGNodeDef	**startGn;
short		*startD;
Boolean		*ioHiLoFlg;
{
short		d;
int		xy;
int		minXY = MAXINT;
CDSGNodeDef	*gn, *minGn;

MI_MOD_START("CDSGetStartIoGNode");

minGn = NULL;
gn = *startGn;				/* Go around external part of graph */
for (d= *startD; ; d=CDS_NEXTDIR(d, CDS_CCW)) {
    DASSERT(gn->type=='e');
    for ( ; !gn->bch[d].node; d=CDS_NEXTDIR(d,CDS_CW));

    switch(io) {
    case 0: xy =  gn->loc[CDS_X]+gn->loc[CDS_Y]; break;
    case 1: xy = -gn->loc[CDS_X]-gn->loc[CDS_Y]; break;
    case 2: xy =  gn->loc[CDS_X]-gn->loc[CDS_Y]; break;
    case 3: xy = -gn->loc[CDS_X]+gn->loc[CDS_Y]; break;
    default: DASSERT(0);
    }
    if (xy<minXY) {
	minXY = xy;
	minGn = gn;
	*startD = d;
    }

    MIMSG(4, (MIDBG,"On gn=%d (%d %d) nextDir=%s",
	CDS_ID(gn), gn->loc[CDS_X], gn->loc[CDS_Y], CDS_CDIR_TOSTR(d)));

    gn = gn->bch[d].node;
    if (gn== *startGn) break;
}

MIMSG(2, (MIDBG,"Io side %s: Picked gn=%d (%d %d) startDir=%s",
    CDS_IO_TOSTR(io), CDS_ID(minGn),
    minGn->loc[CDS_X], minGn->loc[CDS_Y], CDS_CDIR_TOSTR(*startD)));

DASSERT(minGn);

switch(io) {				/* Attempt to find io slice: */
case 0:					/* Bottom side */
    DASSERT(minGn== *startGn);
    *ioHiLoFlg = FALSE;
    break;
case 1:					/* Top side */
    *startGn = minGn;
    *ioHiLoFlg = TRUE;
    break;
case 2:					/* Left side */
    *startGn = minGn;
    *ioHiLoFlg = FALSE;
    break;
case 3:					/* Right side */
    *startGn = minGn;
    *ioHiLoFlg = TRUE;
    break;
default: ASSERT(0);
}
MI_MOD_END();
} /* ............................. End of procedure CDSGetStartIoGNode */
/* Merge empty rooms in channel graph (faces of planar graph that do not
have a corresponding cell are called empty rooms)
----------------------------------------------------------------------------*/
CDSMergeEmptyRooms(numSliceNode, sliceNode, emptySide)

short		numSliceNode;
CDSGNodeDef	**sliceNode;
int		emptySide;
{
CDSGNodeDef	*gn, *gnp, *gn1, *gn2, *gn3;
CDSGNodeDef	*gn0, *gnp0;
short		rd, d, d2, d0;
short		objNum, objNumd;
short		nsn;
short		cnt;

MI_MOD_START("CDSMergeEmptyRooms");

gnp0 = sliceNode[0];			/* Check if empty rooms are adjacent */
for (nsn=1; nsn<numSliceNode; nsn++) {
    gn0 = sliceNode[nsn];
    CDS_GETDIR(d0, gnp0->loc, gn0->loc);
    for (gn=gn0, gnp=gnp0, d=d0; nsn<numSliceNode; ) {
	rd = CDS_REVDIR(d);
	gnp->bch[d].node = NULL;
	gn->bch[rd].node = NULL;
	for (d2=0, cnt=0; d2<4; d2++)
	    if (gn->bch[d2].node) cnt++;
	DASSERT(cnt>0);
	if (cnt>1) break;
	DASSERT(nsn<numSliceNode-1);
	gnp = gn;
	gn  = sliceNode[++nsn];
	CDS_GETDIR(d, gnp->loc, gn->loc);
    }

    if (emptySide==1) {
	gn1 = gnp0;
	gn2 = gn;
	rd  = CDS_REVDIR(d);
	d   = d0;
    }
    else {
	gn1 = gn;
	gn2 = gnp0;
	rd  = d0;
	d   = CDS_REVDIR(d);
    }
    objNumd = gn1->bch[d].objNum;

    for (d2=d; !gn1->bch[d2].node; d2=CDS_NEXTDIR(d2,CDS_CCW));
    objNum = gn1->bch[d2].objNum;
    DASSERT(objNum!=objNumd);
    if (objNumd>=CDS_NBOBJ_EXTERNAL) {
	gn3 = gn1; gn1 = gn2; gn2 = gn3;/* Exchange gn1<->gn2 */
	d2 = d; d  = rd; rd = d2;	/* Exchange d<->rd */
	objNumd = gn1->bch[d].objNum;
	for (d2=d; !gn1->bch[d2].node; d2=CDS_NEXTDIR(d2,CDS_CCW));
	objNum = gn1->bch[d2].objNum;
    }
    DASSERT(objNumd<CDS_NBOBJ_EXTERNAL);
    MIMSG(2, (MIDBG,"Merging ext obj %d to obj %d", objNumd, objNum));

    for ( ; gn2!=gn1; gn2=gn3) {
	for (d2=CDS_NEXTDIR(rd,CDS_CCW); !gn2->bch[d2].node;
	     d2=CDS_NEXTDIR(d2,CDS_CCW));
	gn3 = gn2->bch[d2].node;
	rd = CDS_REVDIR(d2);
	DASSERT(gn2->bch[d2].objNum==objNumd);
	MIMSG(2, (MIDBG,"On edge %d:(%d %d) to %d:(%d %d)",
	    CDS_ID(gn2), gn2->loc[CDS_X], gn2->loc[CDS_Y],
	    CDS_ID(gn3), gn3->loc[CDS_X], gn3->loc[CDS_Y]));
	if (gn3->bch[rd].objNum==objNum) {
	    MIMSG(2, (MIDBG,"Deleting edge (%d %d) to (%d %d)",
		gn2->loc[CDS_X], gn2->loc[CDS_Y],
		gn3->loc[CDS_X], gn3->loc[CDS_Y]));
	    gn2->bch[d2].node = NULL;
	    gn3->bch[rd].node = NULL;
	    continue;
	}
	gn2->bch[d2].objNum = objNum;
    }

    gnp0 = gn;
}

MI_MOD_END();
} /* ............................. End of procedure CDSMergeEmptyRooms */
