/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSInc.h

Globals & macros for CDSlicexxx.c routines.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSInc.h,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  89/01/27  22:39:43  octtools
 * Initial revision
 * 
 * Revision 1.2  88/04/05  12:21:55  mitch
 * Added CDSMaxNumCh to split off CDSPrintSeqFile from CDSOctOutput
 * 
 * Revision 1.1  88/03/31  11:39:57  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
typedef struct	CDSObjDef	CDSObjDef;
typedef	struct	CDSCornDef	CDSCornDef;
typedef struct	CDSEdgeDef	CDSEdgeDef;
typedef struct	CDSTreeDef	CDSTreeDef;
typedef struct	CDSHeadDef	CDSHeadDef;
typedef	struct	CDSGEdgeDef	CDSGEdgeDef;
typedef	struct	CDSGNodeDef	CDSGNodeDef;
typedef	struct	CDSIoDef	CDSIoDef;
typedef	struct	CDSParmDef	CDSParmDef;
typedef int	CDSLocDef[2];

#define		CDS_X  0	/* X coordinate index */
#define		CDS_Y  1	/* Y coordinate index */
#define		CDS_V  0	/* Vertical cut */
#define		CDS_H  1	/* Horizontal cut */
#define		CDS_LO 0
#define		CDS_HI 1

#define		CDS_TREE_NOT_DONE	-2
#define		CDS_TREE_ELEMENT	-1
#define		CDS_TREE_EXT_DONE	 0

#define		CDS_UNORDERED		-1

#define		CDS_NULL_TREEID		-1
#define		CDS_W_EXT_TREEID	 0
#define		CDS_E_EXT_TREEID	 1
#define		CDS_S_EXT_TREEID	 2
#define		CDS_N_EXT_TREEID	 3

struct CDSObjDef {		/* Object definition: */
    RefAny	dbRef;		/* Database object reference (by name) */
    short	numBdy;		/* Number of boundary pts */
    short	numCorn;	/* Number of corners */
    short	bdyOrient;	/* Orientation of original boundary: */
				/*   clockwise or counterclockwise ? */
    short	minLocBdyIdx;	/* Minimum boundary location index */
    CDSLocDef	*bdy;		/* Boundary of object: */
				/*   pts are listed counter clockwise */
    CDSCornDef	*corn;		/* Corners of obj */
    Boolean	ioPadFlag;	/* Io pad flag */
};

struct CDSCornDef {		/* Corner definition: */
    short	bdyIdx;		/* Boundary index of corner */
    short	treeId;		/* Channel id of edges from bdyIdx */
};

struct CDSEdgeDef {
    int		id;
    short	objNum;		/* Object number */
    short	bdyIdx;		/* Starting boundary pt index */
    char	dir;		/* Direction ? */
    char	hiLo;		/* Low or hi boundary of object ? */
    int		loc;		/* Location of edge */
    Boolean	doneFlg;	/* Edge is Done */
    Boolean	queueFlg;	/* Edge is Queued */
    CDSLocDef	range;		/* Range of edge */
    CDSLocDef	extRange;	/* Extended range */
    CDSEdgeDef	*next;
    CDSEdgeDef	*prev;
};

struct CDSTreeDef {
    short	id;
    short	type;		/* Type of object: */
				/*   -2 = not done yet */
				/*   -1 = element obj */
				/*    0 = external channel */
				/*   >0 = number of subchannels+1 */
    short	objNum;		/* Element obj num */
    short	order;		/* Channel routing order */
    CDSLocDef	*slice;		/* Slice boundary */
    CDSGNodeDef	*node;		/* To node tree */
    CDSTreeDef	*child[2];	/* Leaf elements */
    Boolean	useExtFlag;	/* Use external channel flag */
    short	parentId;	/* Parent tree */
};

struct CDSHeadDef {
    short	numObj;
    int		numEdge;
    int		numGEdge[2];
    int		numGNode;
    short	numTree;
    short	numChan;
};

CDSHeadDef	CDSHead;	/* Header */
CDSEdgeDef	*CDSEdgeArray;	/* Edge list */
CDSEdgeDef	*CDSEdge[2];
CDSObjDef	*CDSObj;	/* List of objects to split */
CDSTreeDef	*CDSTree;	/* Header of split tree */
/* CDSInc.h: General globals for slicing algorithm
----------------------------------------------------------------------------*/
#define	CDS_N	0
#define	CDS_E	1
#define	CDS_S	2
#define CDS_W	3

#define CDS_NBOBJ_NOT_SET  -1
#define CDS_NBOBJ_EXTERNAL -2

struct CDSGEdgeDef {			/* Maximal edges: */
    int		loc;			/* Location */
    CDSLocDef	range;			/* Range */
    CDSGEdgeDef	*next;
};

struct CDSGNodeDef {			/* Node graph: */
    int		id;
    CDSLocDef	loc;			/* Location */
    char	type;			/* Node type: */
					/*   'e' = external */
					/*   's' = split */
    struct {
	char		type;		/* Edge type: */
					/*   'e' = external */
	short		objNum;		/* Object associated with branch */
	short		treeId;		/* Tree id number */
	CDSGNodeDef	*node;		/* Adjacent node */
    } bch[4];				/* Branches from node (index by NESW) */

    CDSGNodeDef	*next;
};

CDSGEdgeDef	*CDSGEdge[2];
CDSGNodeDef	*CDSGNode;

#define	CDS_ID(n) ((n) ? (n)->id : -99)
#define CDS_COPY_LOC(to, from) {to[0]=from[0]; to[1]=from[1];}
#define CDS_DIR_TOSTR(d) \
    (d==CDS_V ? "Vert dir" : (d==CDS_H ? "Hor dir" : "ERROR dir"))
#define CDS_HILO_TOSTR(hl) \
    (hl==CDS_HI ? "Hi" : (hl==CDS_LO ? "Lo" : "ERROR hilo"))
#define CDS_CDIR_TOSTR(d) \
    (d==CDS_N ? "N" : (d==CDS_E ? "E" : \
    (d==CDS_S ? "S" : (d==CDS_W ? "W" : "?"))))
#define CDS_IO_TOSTR(io) \
    (io==0 ? "bottom"	: (io==1 ? "top" : \
    (io==2 ? "left"	: (io==3 ? "right" : "?"))))

#define	CDS_CW   1
#define	CDS_CCW -1
#define CDS_REVORI(ori) (-(ori))
#define CDS_NEXTDIR(dir, ori) (dir+4+ori)%4
#define CDS_REVDIR(dir)       (dir+2)%4
#define CDS_GETDIR(dir, loc0, loc1) \
{ \
if (loc0[CDS_X]==loc1[CDS_X]) { \
    if (loc0[CDS_Y]<loc1[CDS_Y]) \
	 dir = CDS_N; \
    else dir = CDS_S; \
} \
else { \
    if (loc0[CDS_X]<loc1[CDS_X]) \
         dir = CDS_E; \
    else dir = CDS_W; \
} }
/* CDSInc.h: General globals for slicing algorithm
----------------------------------------------------------------------------*/

struct CDSParmDef {
    String	cellName;
    String	inputView;
    String	outputView;
    String	seqFileName;
    Boolean	showChanFlg;
    Boolean	mergeIoFlg;
    Boolean	zChanFlg;
};

struct CDSIoDef {
    Boolean	sliceTriedFlg;
    Boolean	sliceFoundFlg;
    CDSTreeDef	*sideTree;
    CDSGNodeDef	*bodyGNode;
    short	numObj;
    short	*objNum;
    String	mergedInstName;
};

CDSParmDef	CDSParm;
CDSIoDef	CDSIo[4];
Boolean		CDSOctMode;
int		CDSMaxNumCh;
