/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSMergeIo 		Merge io side instances into 4 instances.
CDSExpandInst		Expand instance locations
CDSChkFTermAttachedToInstanceBag

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSMergeIo.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.4  90/02/07  16:10:33  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.3  89/02/06  19:07:39  octtools
 * upsdated
 * 
 * Revision 1.2  89/02/01  17:19:57  octtools
 * writeable version existed, checked back in
 * 
 * Revision 1.1  89/01/27  22:39:30  octtools
 * Initial revision
 * 
 * Revision 1.2  88/04/05  12:29:04  mitch
 * Made all operations off of one oct facet instead of several like the
 * cdmerge view (for merged ios) and cdexpandxxx view (for expanded views).
 * 
 * Revision 1.1  88/03/31  11:38:41  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "MIInc.h"
#include "CDSInc.h"
#include "oct.h"
#include "nle.h"
#include "oh.h"

/* Merge io sides
----------------------------------------------------------------------------*/
int
CDSMergeIo(octMain, expFactor, expFlg)

octObject	*octMain;
double		expFactor;
Boolean		expFlg;
{
#define		CDS_NUM_IOSIDE_DIM 1000
short		io, i, o;
int		rc = 16;
int		mergeRc;
CDSTreeDef	*t;
short		objNum[CDS_NUM_IOSIDE_DIM];
Boolean		somethingToMergeFlg = FALSE;
Boolean		badIoSideFlg;
octObject	octInst, octMergedInst;
octObject	octMergeInstBag;
octObject	octSubStuff;
octObject	octMergedFacet;
octObject	octInstPadProp;
octGenerator	genSubStuff;

MI_MOD_START("CDSMergeIo");

DASSERT(!MIStackPtr);
for (io=0; io<4; io++) {		/* Check if each io side has stuff */
    if (!CDSIo[io].sliceFoundFlg) continue; /* to merge */

    t = CDSIo[io].sideTree;		/* Scan tree for all leaf objects */
    MI_STACK_PUSH(t);
    for (i=0; MIStackPtr; ) {
	MI_STACK_POP(CDSTreeDef*, t);
	if (!t) continue;
	if (t->type!=CDS_TREE_ELEMENT) {
	    MI_STACK_PUSH(t->child[CDS_LO]);
	    MI_STACK_PUSH(t->child[CDS_HI]);
	    continue;
	}
	if (t->objNum<0) continue;
					/* Record leaf objects in temp array */
	if (i<CDS_NUM_IOSIDE_DIM)
	    objNum[i] = t->objNum;
	i++;
    }
    CDSIo[io].numObj = i;
					/* Check dimensions */
    if (CDSIo[io].numObj>=CDS_NUM_IOSIDE_DIM) {
	MIfprintf(MIERR,
	    "ERROR Array exceeded %d (maximum = %d) io instances on one side\n",
	    CDSIo[io].numObj, CDS_NUM_IOSIDE_DIM);
	goto CDSMergeIoEnd;
    }
					/* If more than one object then */
					/*   copy the stuff over */
    if (CDSIo[io].numObj>1) {
	somethingToMergeFlg = TRUE;
	CDSIo[io].objNum = ALLOC_ARRAY(CDSIo[io].numObj, short);
	badIoSideFlg = FALSE;
	for (i=0; i<CDSIo[io].numObj; i++) {
	    o = objNum[i];
	    if (!CDSObj[o].ioPadFlag) {
		MIfprintf(MIERR,
		    "ERROR Thought that inst %s on io side %s is a pad\n",
		    CDSObj[o].dbRef, CDS_IO_TOSTR(io));
		badIoSideFlg = TRUE;
	    }
	    CDSIo[io].objNum[i] = o;
	}
	if (badIoSideFlg) {
	    MIfprintf(MIERR,
		"Not merging any instances on io side %s\n", CDS_IO_TOSTR(io));
	    FREE_ARRAY(CDSIo[io].numObj, CDSIo[io].objNum, short);
	    CDSIo[io].numObj = 0;
	}
    }
    else if (CDSIo[io].numObj==1) {
	o = objNum[0];
	if (CDSObj[o].ioPadFlag)
	    CDSIo[io].mergedInstName = MI_STR_COPY((String)CDSObj[o].dbRef);
    }
}

if (!somethingToMergeFlg) goto CDSMergeIoEnd;

OH_ASSERT(ohGetOrCreateBag(octMain, &octMergeInstBag, "MERGE_INSTANCE"));

CDSChkFTermAttachedToInstanceBag("before nleMakeMacro", octMain);

for (io=0; io<4; io++) {
    if (CDSIo[io].numObj<=1) continue;

    MIMSG(0, (MIDBG,"On io side %s (number of objects = %d)",
	CDS_IO_TOSTR(io), CDSIo[io].numObj));
					/* Get list of objects to be merged */
					/*   attach them to merge bag */
    for (i=0; i<CDSIo[io].numObj; i++) {
	o = CDSIo[io].objNum[i];
	MIMSG(1, (MIDBG,"%2d: objNum=%d name=<%s>", i, o, CDSObj[o].dbRef));
	OH_ASSERT(ohGetByInstName(octMain, &octInst, (String)CDSObj[o].dbRef));
	OH_ASSERT(octAttach(&octMergeInstBag, &octInst));
    }

    (void) sprintf(MIBuf,"%s.io%s", octMain->contents.facet.cell, CDS_IO_TOSTR(io));
    rc = ohOpenFacet(&octMergedFacet, MIBuf, octMain->contents.facet.view,
		     "contents", "w");
    if (rc<OCT_OK) {
	MIfprintf(MIERR,
	    "ERROR Could not open facet %s:%s:%s\n",
	    octMergedFacet.contents.facet.cell,
	    octMergedFacet.contents.facet.view,
	    octMergedFacet.contents.facet.facet);
	octError("");
	goto CDSMergeIoEnd;
    }

    MIMSG(0, (MIDBG,"Calling merge program"));
    mergeRc = nleMakeMacro(&octMergedFacet, &octMergeInstBag, &octMergedInst,
			   NLE_HARD);
    if (mergeRc!=NLE_OK) {
	(void) fprintf(stderr,"%s: merging io side %s (rc=%d)",
		       nleErrorString(), CDS_IO_TOSTR(io), mergeRc);
	rc = 8;
	goto CDSMergeIoEnd;
    }
    CDSIo[io].mergedInstName = MI_STR_COPY(octMergedInst.contents.instance.name);
    OH_ASSERT(ohGetOrCreatePropStr(&octMergedInst, &octInstPadProp,
	"PLACEMENT_CLASS", "PAD"));
					/* Clean merge bag contents */
    OH_ASSERT(octInitGenContents(&octMergeInstBag, OCT_ALL_MASK, &genSubStuff));
    while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK) {
	OH_ASSERT(octDetach(&octMergeInstBag, &octSubStuff));
    }

    FREE_ARRAY(CDSIo[io].numObj, CDSIo[io].objNum, short);
}

CDSChkFTermAttachedToInstanceBag("after CDSMergeIo", octMain);

OH_ASSERT(octDelete(&octMergeInstBag));
if (expFlg)
    CDSExpandInst(octMain, expFactor);
rc = 0;

CDSMergeIoEnd:
MI_MOD_END();
return(rc);
} /* ............................. End of procedure CDSMergeIo */
/* Check if fterms are attached to the instance bag.
----------------------------------------------------------------------------*/
CDSChkFTermAttachedToInstanceBag(msg, octMain)

octObject	*octMain;
String		msg;
{
octObject	octFTerm;
octGenerator	genFTerm;
octObject	octInstBag;
octGenerator	genInstBag;

OH_ASSERT(octInitGenContents(octMain,OCT_TERM_MASK,&genFTerm));
while (octGenerate(&genFTerm, &octFTerm)==OCT_OK) {
    octInitGenContainers(&octFTerm, OCT_BAG_MASK, &genInstBag);
    while (octGenerate(&genInstBag, &octInstBag)==OCT_OK) {
	if (!strcmp(octInstBag.contents.bag.name,"INSTANCES")) {
	    MIfprintf(MIERR,"ERROR (%s) fTerm %s is contained in INSTANCES bag - atlas will detach it\n",
		msg, octFTerm.contents.term.name);
	    OH_ASSERT(octDetach(&octInstBag, &octFTerm));
	}
    }
}
} /* ............................. End of procedure CDSChkFTermAttachedToInstanceBag */
/* Expand instance locations
----------------------------------------------------------------------------*/
CDSExpandInst(octMain, expFactor)

octObject	*octMain;
double		expFactor;
{
octObject	octInst;
octGenerator	genInst;
struct octBox	box, newBox;

struct octPoint	instLoc;
struct octPoint	newInstLoc;

struct octBox	intBox, newIntBox;
int		io;

MI_MOD_START("CDSExpandInst");

MIprintf("Expanding all instance locations by %lf\n", expFactor);

intBox.lowerLeft .x = intBox.lowerLeft .y =  MAXINT;
intBox.upperRight.x = intBox.upperRight.y = -MAXINT;
newIntBox = intBox;

octInitGenContents(octMain, OCT_INSTANCE_MASK, &genInst);
while (octGenerate(&genInst, &octInst)==OCT_OK) {
    OH_ASSERT(octBB(&octInst, &box));
    instLoc = box.lowerLeft;
    newInstLoc.x = (int)((double)instLoc.x * expFactor);
    newInstLoc.y = (int)((double)instLoc.y * expFactor);

    newBox = box;
    newBox.lowerLeft .x += newInstLoc.x - instLoc.x;
    newBox.lowerLeft .y += newInstLoc.y - instLoc.y;
    newBox.upperRight.x += newInstLoc.x - instLoc.x;
    newBox.upperRight.y += newInstLoc.y - instLoc.y;

    for (io=0; io<4; io++)
	if (CDSIo[io].mergedInstName &&
	    !strcmp(CDSIo[io].mergedInstName, octInst.contents.instance.name)) break;

    if (io==4) {
	intBox.lowerLeft .x = MIN(intBox.lowerLeft .x, box.lowerLeft .x);
	intBox.lowerLeft .y = MIN(intBox.lowerLeft .y, box.lowerLeft .y);
	intBox.upperRight.x = MAX(intBox.upperRight.x, box.upperRight.x);
	intBox.upperRight.y = MAX(intBox.upperRight.y, box.upperRight.y);
	newIntBox.lowerLeft .x = MIN(newIntBox.lowerLeft .x, box.lowerLeft .x);
	newIntBox.lowerLeft .y = MIN(newIntBox.lowerLeft .y, box.lowerLeft .y);
	newIntBox.upperRight.x = MAX(newIntBox.upperRight.x, box.upperRight.x);
	newIntBox.upperRight.y = MAX(newIntBox.upperRight.y, box.upperRight.y);
    }

    octInst.contents.instance.transform.translation.x += newInstLoc.x - instLoc.x;
    octInst.contents.instance.transform.translation.y += newInstLoc.y - instLoc.y;
    OH_ASSERT(octModify(&octInst));
}
					/* TODO readjust the io locations */

MI_MOD_END();
} /* ............................. End of procedure CDSExpandInst */
