/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSInitObjCorn.c

Slice channels for cut lines

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSObjCorn.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  90/02/07  16:10:34  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:39:31  octtools
 * Initial revision
 * 
 * Revision 1.2  88/04/05  12:30:32  mitch
 * Fixed bug in z-shaped channels.
 * 
 * Revision 1.1  88/03/31  11:38:49  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "MIInc.h"
#include "CDSInc.h"

/* Initialize object corners.
----------------------------------------------------------------------------*/
CDSInitObjCornNew()
{
#define CDS_CORN_TOSTR(d) \
    (d==CDS_E ? "lower left" : (d==CDS_N ? "lower right" : \
    (d==CDS_S ? "upper left" :             "upper right")))
CDSTreeDef	*t;
CDSGNodeDef	*g, *gn, *startGn;
short		k, rpd;
short		ti, pti;
short		o, c;
short		ori = CDS_CCW;

short		lastDir, cornDir, tempDir;
Boolean		straightFlg = FALSE;

short		cornDim;
short		numCorn;
CDSCornDef	corn[40];

short		b, minB;
int		x, x0;
int		y, y0;
Boolean		xFlg, yFlg;
int		xy, minXY;

MI_MOD_START("CDSInitObjCornNew");

DASSERT(!MIStackPtr);

MI_STACK_PUSH(CDSTree);

while (MIStackPtr) {			/* Push until all nodes sliced */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    if (t->type!=CDS_TREE_ELEMENT) {
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	continue;
    }
    o = t->objNum;
    if (o<0) continue;
					/* Init corners for object o */
    MIMSG(1, (MIDBG,
	"Finding corners of object%d name=%s", o, CDSObj[o].dbRef));
    DRANGE(0,o,CDSHead.numObj-1);

    startGn = t->node;
    DASSERT(startGn->bch[CDS_E].node && startGn->bch[CDS_N].node);

    g   = startGn;			/* Go around graph ctrClkWise */
    rpd = CDS_N;			/*   starting in lower left corner */
    pti = g->bch[rpd].treeId;		/*   until valid corner is found. */
					/*   Setting up g, rpd, pti and */
					/*   lastDir. */
    for (gn=NULL; gn!=startGn; g=gn) {
	DASSERT(g->bch[rpd].objNum==o);
	for (k=CDS_NEXTDIR(rpd,ori); !g->bch[k].node; k=CDS_NEXTDIR(k,ori));
	gn = g->bch[k].node;
	ti = g->bch[k].treeId;
	if (ti!=pti && rpd==CDS_NEXTDIR(k,ori)) {
	    lastDir = k;
	    startGn = gn;
	}
	pti = ti;
	rpd = CDS_REVDIR(k);
    }
    startGn = g;
    DASSERT(startGn!=t->node);

    numCorn = 0;
    cornDim = 4;
					/* Go all the way around graph */
    for (gn=NULL; gn!=startGn; g=gn) {	/* Go around instance channel graph */

	DASSERT(g->bch[rpd].objNum==o);
	for (k=CDS_NEXTDIR(rpd,ori); !g->bch[k].node; k=CDS_NEXTDIR(k,ori));

	gn = g->bch[k].node;
	ti = g->bch[k].treeId;

	if (ti!=pti) {			/* New corner (different channels) */
	    cornDir = k;
	    tempDir = CDS_NEXTDIR(lastDir, ori);
					/* Check if direction ok */
	    if (rpd!=CDS_NEXTDIR(k,ori)) {
					/* Corner is on straight line segment */
		straightFlg = TRUE;
		if (rpd==CDS_REVDIR(k)) {
		    MIMSG(-1, (MIDBG,
			"Corner node of obj %s is on straight line",
			CDSObj[o].dbRef));
		    cornDir = tempDir;

		    if (numCorn==cornDim) {
			MIMSG(-1, (MIDBG,
			    "WARN Already enough corners - ignoring this"));
			break;
		    }
		}
		else {			/* Bad news - check it out */
		    MIfprintf(MIERR,
			"ERROR Cannot figure out corners of instance %s\n",
			CDSObj[o].dbRef);
		    abort();
		}
	    }
					/* Corner repeated - take this one */
	    else if (cornDir==lastDir) {
		MIMSG(-1, (MIDBG, "Repeating corner %s of obj %s",
		    CDS_CORN_TOSTR(cornDir), CDSObj[o].dbRef));
		numCorn--;
		if (!straightFlg) {	/* Was not straight before then */
					/*   something might be wrong */
		    MIfprintf(MIERR, "WARN Corner %s of inst %s repeated\n",
			CDS_CORN_TOSTR(cornDir), CDSObj[o].dbRef);
		}
		straightFlg = FALSE;
	    }

	    else if (cornDir!=tempDir) {
		MIfprintf(MIERR,"WARN Skipped %s corner of inst %s\n",
		    CDS_CORN_TOSTR(tempDir), CDSObj[o].dbRef);
		cornDim--;
		straightFlg = FALSE;
	    }
					/* These coordinates have problems */
					/*   when dealing with straights */
	    x0 = g->loc[CDS_X];
	    y0 = g->loc[CDS_Y];

	    MIMSG(3, (MIDBG,
		"New corner %s node=%d (%d %d) dirPrev=%s dirNext=%s",
		CDS_CORN_TOSTR(cornDir), CDS_ID(g), x0, y0,
		CDS_CDIR_TOSTR(rpd), CDS_CDIR_TOSTR(k)));

	    xFlg = yFlg = TRUE;		/* Direction cornDir determines which */
	    switch (cornDir) {		/*   which corner we are at */
	    case CDS_E:	break;		/* Corner at min( x+y) x>x0 & y>y0 */
	    case CDS_N:	xFlg = FALSE;	/* Corner at min(-x+y) x<x0 & y>y0 */
			break;
	    case CDS_W:	xFlg = FALSE;	/* Corner at min(-x-y) x<x0 & y<y0 */
	    default   :	yFlg = FALSE;	/* Corner at min( x-y) x>x0 & y<y0 */
	    }

	    minXY = MAXINT;		/* Find corner of instance polygon */
	    for (b=0; b<CDSObj[o].numBdy; b++) {
		x = CDSObj[o].bdy[b][CDS_X];
		if (xFlg) {
		    /* if (x<x0) continue; */
		    xy = x;
		}
		else {
		    /* if (x>x0) continue; */
		    xy = -x;
		}

		y = CDSObj[o].bdy[b][CDS_Y];
		if (yFlg) {
		    /* if (y<y0) continue; */
		    xy += y;
		}
		else {
		    /* if (y>y0) continue; */
		    xy -= y;
		}

		if (xy<minXY) {		/* Save best bdy point */
		    minXY = xy;
		    minB  = b;
		}
	    }
	    DASSERT(minXY<MAXINT);
	    MIMSG(3, (MIDBG,"Corner%d of bdy %d: (%d %d) treeId=%d",
		numCorn, minB,
		CDSObj[o].bdy[minB][CDS_X], CDSObj[o].bdy[minB][CDS_Y], ti));
					/* Update corn array */
	    corn[numCorn].bdyIdx = minB;
	    corn[numCorn].treeId = ti;
	    numCorn++;

	    lastDir = cornDir;
	}

	pti = ti;
	rpd = CDS_REVDIR(k);
    }
					/* Copy corn from temp array */
    if (numCorn!=cornDim) {
	MIfprintf(MIERR,
	    "WARN Inst %s has %d surrounding channels (should be %d)\n",
	    CDSObj[o].dbRef, numCorn, cornDim);
    }
    CDSObj[o].numCorn = numCorn;
    CDSObj[o].corn    = ALLOC_ARRAY(numCorn, CDSCornDef);
    for (c=0; c<numCorn; c++) {
	CDSObj[o].corn[c].bdyIdx = corn[c].bdyIdx;
	CDSObj[o].corn[c].treeId = corn[c].treeId;
    }

    if (*MIDebugFlag>=5)
	CDSPrintNGraph(t->node);
}

MI_MOD_END();
} /* ............................. End of procedure CDSInitObjCornNew */
/* Initialize object corners.
----------------------------------------------------------------------------*/
CDSInitObjCorn()
{
#define		CDS_DIM_CORN 40
CDSTreeDef	*t;
CDSGNodeDef	*g, *gn;
short		k, rpd;
short		ti, pti;
short		o, c;
short		ori = CDS_CCW;

CDSGNodeDef	*lastGCorn;
short		lastDir, lastDir2;

short		numCorn;
CDSCornDef	corn[CDS_DIM_CORN];

short		b, minB;
int		x, x0;
int		y, y0;
Boolean		xFlg, yFlg;
int		xy, minXY;

MI_MOD_START("CDSInitObjCorn");

DASSERT(!MIStackPtr);

MI_STACK_PUSH(CDSTree);

while (MIStackPtr) {			/* Push until all nodes sliced */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    if (t->type!=CDS_TREE_ELEMENT) {
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	continue;
    }
    o = t->objNum;
    if (o<0) continue;
					/* Init corners for object o */
    MIMSG(1, (MIDBG,"Finding corners of object %d", o));
    DRANGE(0,o,CDSHead.numObj-1);
    numCorn = 0;
    g   = t->node;			/* Go around graph counterClkWise */
    rpd = CDS_N;
    pti = g->bch[CDS_N].treeId;
    DASSERT(g->bch[CDS_E].node && g->bch[CDS_N].node);
    lastGCorn = g;
    lastDir   = CDS_E;

    for (gn=NULL; gn!=t->node; g=gn) {
	DASSERT(g->bch[rpd].objNum==o);
	for (k=CDS_NEXTDIR(rpd,ori); !g->bch[k].node; k=CDS_NEXTDIR(k,ori));

	gn = g->bch[k].node;
	ti = g->bch[k].treeId;

	if (ti!=pti) {			/* New corner */
	    if (rpd!=CDS_NEXTDIR(k,ori)) {
		DASSERT(rpd==CDS_REVDIR(k));
		MIMSG(2, (MIDBG,"Corner node of obj %d is on straight line", o));
		lastDir2 = CDS_NEXTDIR(lastDir,ori);
		if (lastGCorn->bch[lastDir ].treeId==
		    lastGCorn->bch[lastDir2].treeId) {
		    g = lastGCorn;
		    if (k!=lastDir)
			MIfprintf(MIERR,
			    "ERROR Corners will be screwed up in obj %d\n", o);
		}
		else continue;
	    }

	    x0 = g->loc[CDS_X];
	    y0 = g->loc[CDS_Y];
	    MIMSG(3, (MIDBG,"New corner node=%d (%d %d) dirPrev=%s dirNext=%s",
		CDS_ID(g), x0, y0, CDS_CDIR_TOSTR(rpd), CDS_CDIR_TOSTR(k)));

	    xFlg = yFlg = TRUE;
	    switch (k) {
	    case CDS_E:	break;		/* Corner at min( x+y) x>x0 & y>y0 */
	    case CDS_N:	xFlg = FALSE;	/* Corner at min(-x+y) x<x0 & y>y0 */
			break;
	    case CDS_W:	xFlg = FALSE;	/* Corner at min(-x-y) x<x0 & y<y0 */
	    default   :	yFlg = FALSE;	/* Corner at min( x-y) x>x0 & y<y0 */
	    }

	    minXY = MAXINT;		/* Find corner */
	    for (b=0; b<CDSObj[o].numBdy; b++) {
		x = CDSObj[o].bdy[b][CDS_X];
		if (xFlg) {
		    if (x<x0) continue;
		    xy = x;
		}
		else {
		    if (x>x0) continue;
		    xy = -x;
		}

		y = CDSObj[o].bdy[b][CDS_Y];
		if (yFlg) {
		    if (y<y0) continue;
		    xy += y;
		}
		else {
		    if (y>y0) continue;
		    xy -= y;
		}

		if (xy<minXY) {		/* Save best bdy point */
		    minXY = xy;
		    minB  = b;
		}
	    }
	    DASSERT(minXY<MAXINT);
	    MIMSG(3, (MIDBG,"Corner%d of bdy %d: (%d %d) treeId=%d",
		numCorn, minB,
		CDSObj[o].bdy[minB][CDS_X], CDSObj[o].bdy[minB][CDS_Y], ti));
					/* Update corn array */
	    corn[numCorn].bdyIdx = minB;
	    corn[numCorn].treeId = ti;
	    numCorn++;
	}

	if (rpd==CDS_NEXTDIR(k,ori)) {
	    lastGCorn = g;
	    lastDir   = k;
	}

	pti = ti;
	rpd = CDS_REVDIR(k);
    }
					/* Copy corn from temp array */
    CDSObj[o].numCorn = numCorn;
    CDSObj[o].corn    = ALLOC_ARRAY(numCorn, CDSCornDef);
    for (c=0; c<numCorn; c++) {
	CDSObj[o].corn[c].bdyIdx = corn[c].bdyIdx;
	CDSObj[o].corn[c].treeId = corn[c].treeId;
    }

    if (*MIDebugFlag>=5)
	CDSPrintNGraph(t->node);
}

MI_MOD_END();
} /* ............................. End of procedure CDSInitObjCorn */
