/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSObjRing.c

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSObjRing.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  90/02/07  16:10:35  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:39:32  octtools
 * Initial revision
 * 
 * Revision 1.1  88/03/31  11:38:57  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "MIInc.h"
#include "CDSInc.h"

/* Initialize gNodes
----------------------------------------------------------------------------*/
CDSInitGNObj()
{
int		numTNode = 0;
CDSGNodeDef	**tNode;
CDSGNodeDef	*gn, *ngn;
int		*loc;
int		CDSQsortTNode();
short		nullObjNum = CDS_NBOBJ_EXTERNAL-1;
short		o;
int		n;

MI_MOD_START("CDSInitGNObj");
					/* Init tNode[] */
tNode = ALLOC_ARRAY(CDSHead.numGNode, CDSGNodeDef*);
for (gn=CDSGNode; gn; gn=gn->next) {	/* Fill with nodes with North bch */
    if (!gn->bch[CDS_N].node) continue;
    tNode[numTNode++] = gn;
}
					/* Sort lowest x then y first */
qsort((char *)tNode, numTNode, sizeof(CDSGNodeDef*), CDSQsortTNode);

for (o=0; o<CDSHead.numObj-1; o++) {	/* Set obj ring for each obj */
					/* Get lo starting Norch bch */
    loc = CDSObj[o].bdy[CDSObj[o].minLocBdyIdx];

    for (n=numTNode-1; n>=0; n--) {
	gn = tNode[n];
	if (gn->loc[CDS_X]<=loc[CDS_X] &&
	    gn->loc[CDS_Y]<=loc[CDS_Y] &&
	    gn->bch[CDS_N].node->loc[CDS_Y]>loc[CDS_Y]) break;
    }
    DASSERT(n>=0);
    ngn = gn->bch[CDS_N].node;

    CDSInitNObjRing(gn, ngn, o);	/* Init ring bchs with obj num */
}

gn = CDSGNode->bch[CDS_N].node;		/* Init external obj ring */
CDSInitNObjRing(gn, CDSGNode, (short)CDS_NBOBJ_EXTERNAL);

for (gn=CDSGNode; gn; gn=gn->next) {	/* Init null obj ring bchs */
    if (!gn->bch[CDS_N].node) continue;
    if (gn->bch[CDS_N].objNum==CDS_NBOBJ_NOT_SET)
	CDSInitNObjRing(gn, gn->bch[CDS_N].node, nullObjNum--);
}

MIMSG(1, (MIDBG,"%d empty rooms", -nullObjNum+CDS_NBOBJ_EXTERNAL-1));

FREE_ARRAY(CDSHead.numGNode, tNode, CDSGNodeDef*);

MI_MOD_END();
} /* ............................. End of procedure CDSInitGNObj */
/* Sort North branch nodes by lowest x then lowest y first.
----------------------------------------------------------------------------*/
int
CDSQsortTNode(n1, n2)

CDSGNodeDef	**n1, **n2;
{
int		l1, l2;

l1 = (*n1)->loc[CDS_X];
l2 = (*n2)->loc[CDS_X];
if (l1!=l2) return(l1-l2);
l1 = (*n1)->loc[CDS_Y];
l2 = (*n2)->loc[CDS_Y];
DASSERT(l1!=l2);
return(l1-l2);
} /* ............................. End of procedure CDSQsortTNode */
/* Starting from gStart and going towards gNext, initialize internal node ring
objNum.  Go around clockwise for northbound vector (internal ring), counter-
clockwise for southbound vector (external ring).
----------------------------------------------------------------------------*/
CDSInitNObjRing(gStart, gNext, objNum)

CDSGNodeDef	*gStart, *gNext;
short		objNum;
{
short		rpd, d, k;
CDSGNodeDef	*g, *gn;

MI_MOD_START("CDSInitNObjRing");

MIMSG(1, (MIDBG,"Init node ring objNum: gs=%d (%d %d), gn=%d (%d %d), o=%d",
    CDS_ID(gStart), gStart->loc[CDS_X], gStart->loc[CDS_Y],
    CDS_ID(gNext ), gNext ->loc[CDS_X], gNext ->loc[CDS_Y], objNum));

CDS_GETDIR(d, gStart->loc, gNext->loc);

gn = NULL;
rpd = CDS_NEXTDIR(d, CDS_CW);
for (g=gStart; gn!=gStart; g=gn) {	/* Go around graph face */
    for (k=CDS_NEXTDIR(rpd,CDS_CCW);	/* Get next internal node */
	!g->bch[k].node; k=CDS_NEXTDIR(k,CDS_CCW));

    gn  = g->bch[k].node;
    MIMSG(3, (MIDBG,"Next node=%d (%d %d) dir=%s, o=%d",
        CDS_ID(gn), gn->loc[CDS_X], gn->loc[CDS_Y],
	CDS_CDIR_TOSTR(k), gn->bch[k].objNum));
    DASSERT(g->bch[k].objNum==CDS_NBOBJ_NOT_SET);
    g->bch[k].objNum = objNum;

    DASSERT(g!=gStart || gn==gNext);
    rpd = CDS_REVDIR(k);
}

MI_MOD_END();
} /* ............................. End of procedure CDSInitNObjRing */
