/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSUtil.c

CDSCleanUp		Free atlas internal data structures
CDSPrintSeqFile		Print sequencing file

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSUtil.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  90/02/07  16:10:42  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:39:36  octtools
 * Initial revision
 * 
 * Revision 1.5  88/04/15  14:11:13  mitch
 * Free slice for external channels.
 * 
 * Revision 1.2  88/04/05  12:32:21  mitch
 * Added CDSPrintSeqFile and added new function for CDSCleanUp to handle
 * "CDSObj" mode.
 * 
 * Revision 1.1  88/03/31  13:55:01  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "MIInc.h"
#include "CDSInc.h"
#include "oct.h"

/* Clean up all temporary arrays in CDS routines.
----------------------------------------------------------------------------*/
CDSCleanUp(mode)

String		mode;
{
CDSGNodeDef	*gn, *gnNext;
CDSGEdgeDef	*ge, *geNext;
short		o, d;
CDSTreeDef	*t;

MI_MOD_START("CDSCleanUp");

if (!strcmp(mode, "CDSEdge")) {
    FREE_ARRAY(CDSHead.numEdge, CDSEdgeArray, CDSEdgeDef);
    goto CDSEndCleanUp;
}

for (o=0; o<CDSHead.numObj; o++) {
    if (CDSObj[o].dbRef)
	FREE_STRING(CDSObj[o].dbRef);
    FREE_ARRAY(CDSObj[o].numBdy, CDSObj[o].bdy, CDSLocDef);
    FREE_ARRAY(CDSObj[o].numCorn, CDSObj[o].corn, CDSCornDef);
}
FREE_ARRAY(CDSHead.numObj, CDSObj, CDSObjDef);
if (!strcmp(mode, "CDSObj")) goto CDSEndCleanUp;

for (gn=CDSGNode; gn; gn=gnNext) {
    gnNext = gn->next;
    FREE_TYPE(gn, CDSGNodeDef);
}

for (d=CDS_V; ; d=CDS_H) {
    for (ge=CDSGEdge[d]; ge; ge=geNext) {
	geNext = ge->next;
	FREE_TYPE(ge, CDSGEdgeDef);
    }
    if (d==CDS_H) break;
}

MI_STACK_PUSH(CDSTree);
while (MIStackPtr) {
    MI_STACK_POP(CDSTreeDef*, t);
    if (t->child[CDS_LO])
	MI_STACK_PUSH(t->child[CDS_LO]);
    if (t->child[CDS_HI])
	MI_STACK_PUSH(t->child[CDS_HI]);
    if (t->type>0)
	FREE_ARRAY(t->type, t->slice, CDSLocDef);
    if (t->type==CDS_TREE_EXT_DONE)
	FREE_ARRAY(2, t->slice, CDSLocDef);
    FREE_TYPE(t, CDSTreeDef);
}
CDSTree = (CDSTreeDef *)NULL;

CDSEndCleanUp: NOTHING
MI_MOD_END();
} /* ............................. End of procedure CDSCleanUp */
/* Clean up all temporary arrays in CDS routines.
----------------------------------------------------------------------------*/
CDSPrintSeqFile(seqFileName, inputView, maxChOrd, octMain)

String		seqFileName;
String		inputView;
int		maxChOrd;
octObject	*octMain;
{
FILE		*seqFile;
String		cellName;
int		c;

MI_MOD_START("CDSPrintSeqFile");

cellName = octMain->contents.facet.cell;
seqFile = MIfopen(seqFileName,"a","mosaicoDemoRpc sequence file");
fprintf(seqFile,"%s:%s\n", cellName, inputView);
fprintf(seqFile,"%s:%s.cd\n", cellName, inputView);
for (c=0; c<=maxChOrd; c++)
    fprintf(seqFile,"RTstuff/%s.ch%d\n", cellName, c);
fprintf(seqFile,"%s:%s.flat\n", cellName, inputView);
(void) fclose(seqFile);

MI_MOD_END();
} /* ............................. End of procedure CDSPrintSeqFile */
