/*
MIInc.h:

General C-language support defines, varables, and macros.

Author(s):		Mitsuru Igusa
Revision history:	$Log: MIInc.h,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  90/02/07  16:10:50  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:37:00  octtools
 * Initial revision
 * 
 * Revision 1.3  88/04/24  16:54:29  mitch
 * Put in package prefix.
 * 
 * Revision 1.1  88/03/31  15:10:30  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/

#include "utility.h"

typedef char	Boolean;
typedef char	*RefAny;
typedef char	*SelfIdent;
typedef char	*String;
typedef char	Byte;

extern FILE	*MIfopen();

#define TRUE 1
#define FALSE 0
#define MIDBG	stdout
#define MIINFO	stdout
#define MIWARN	stderr
#define MIERR	stderr
#define MI_DUMMY_STUB(fn) int fn() {MIfprintf(MIDBG,"WARN fn stubbed out\n");}

#ifdef MIInc_MAIN
short		MIDefaultDebugFlag = -1;
short		*MIDebugFlag = &MIDefaultDebugFlag;
Boolean		MIAllocDebugFlag = FALSE;
char		MIBuf[1000];
#endif
extern short	MIDefaultDebugFlag;
extern short	*MIDebugFlag;
extern Boolean	MIAllocDebugFlag;
extern char	MIBuf[];
Boolean		MIPrtHeaderFlag;

#define then
#define NOTHING {}
#define MAXSHORT (32767)
#define MAXINT   (2147483647)
#ifndef MAXFLOAT
#   define MAXFLOAT (3.402823e38)
#endif
#define MINMAX(MIN,MAX,A,B) {if ((A)<(B)) {MIN=(A); MAX=(B);} \
			     else	  {MIN=(B); MAX=(A);}}
#define DELTA(A,B) ((A)>(B)?(A)-(B):(B)-(A))
#define ROUND(A) ((A)>=0?((A)+.5):((A)-.5))
#define MI_ALT_DIR(dir) (1-(dir))
#define MI_STR_COPY(S) strcpy(ALLOC_STRING(strlen(S)),S)
#define MI_ASK MIprintf
#define MI_GET scanf
#define MI_YORN(flag) {char t999[10]; scanf("%s",t999); flag = (*t999!='n'); }
#define MI_NORY(flag) {char t999[10]; scanf("%s",t999); flag = (*t999=='y'); }
#define MI_CONV_CHAR_TO_INT(c) (int)((c)&0x80 ? ((c)&0x7f)+0x80 : (c))
#define MI_CONV_BOOLEAN_TO_CHAR(b) (char)((b) ? 'y' : 'n')
#define MI_CONV_CHAR_TO_BOOLEAN(c) ((c)=='y' || (c)=='Y')
#define MI_PRT_BOOLEAN(x) ((x) ? "Y" : "N")
#define MI_PRT_STRING(x)  ((x) ? (x) : "NULL")
#define MI_BEEP() MIprintf("");
#define MI_STUB(subrout) subrout() {MIprintf("subrout is stubbed out\n");}
#define MI_XCHANGE(x, y, z) { z=y; y=x; x=z; }
#define MI_MAXMEM(level, msg) { \
    char	*sbrk(); \
    if (*MIDebugFlag>=level) \
	MIfprintf(MIDBG,"Max memory size (%s) = %.2fM\n", \
	    msg, ((unsigned)sbrk(0))/1048576.); \
}
/* MIInc.h: General globals for Mitch's routines
----------------------------------------------------------------------------*/
#ifdef MIInc_MAIN
String MIAssertFmt  = "ERROR Line %d of file \"%s\" Assert failed\n";
String MIRangeFmt   = "ERROR Line %d of file \"%s\" Variable out of range lo=%d value=%d hi=%d\n";
String MIURangeFmt  = "ERROR Line %d of file \"%s\" Variable exceeded range value=%d limit=%d\n";
String MIAssertSFmt = "ERROR Line %d of file \"%s\" Variable %s=%d but should be %d\n";
String MIAllocChkFmt= "ERROR Line %d of file \"%s\" Could not allocate variable %s\nCurrent memory = %.2fM and trying to allocate %d bytes\n";
String MIAllocFmt   = "%8d + %7d %4d %s\n";
String MIFreeFmt    = "%8d - %7d %4d %s\n";
String MIPrefix     = NULL;
#endif
extern String	MIAssertFmt;
extern String	MIRangeFmt;
extern String	MIURangeFmt;
extern String	MIAssertSFmt;
extern String	MIAllocChkFmt;
extern String	MIAllocFmt;
extern String	MIFreeFmt;
extern String	MIPrefix;
FILE		*MIAllocFile;
String		MIAllocPtr;

#define ASSERT(cond99) \
    { if (!(cond99)) \
	MIfprintf(MIERR, MIAssertFmt,__LINE__,__FILE__); }
#define FASSERT(cond99) { ASSERT(cond99) exit(99); }
#define RANGE(lo99, val99, hi99) \
    { if ((val99)<(lo99) || (hi99)<=(val99)) \
    	MIfprintf(MIERR, MIRangeFmt,__LINE__,__FILE__,lo99,val99,hi99); } 
#define URANGE(val99, limit99) \
    { if ((limit99)<=(val99)) \
    	MIfprintf(MIERR, MIURangeFmt,__LINE__,__FILE__,val99,limit99); } 
#define ASSERTSAME(val99, val98) \
    { if ((val99)!=(val98)) \
    	MIfprintf(MIERR, MIAssertSFmt,__LINE__,__FILE__,"val99",val99,val98); } 
#define ALLOC_CHK(ptr, number, type) \
    { if (!(ptr)) { \
	char	*sbrk(); \
	MIfprintf(MIERR, MIAllocChkFmt,__LINE__,__FILE__,"ptr", \
	    ((unsigned)sbrk(0))/1048576., (number)*sizeof(type)); \
	abort(); \
    } }

#ifdef MI_DEBUG
#   define MIMSG(level, msg)	{if (*MIDebugFlag>=level) \
				    {fprintf msg; MIfprintf(stdout,"\n");}}
#   define MIPGM(level, pgm)	{if (*MIDebugFlag>=level) pgm;}
#   define DASSERT(cond)	 ASSERT(cond)
#   define DRANGE(lo, value, hi) RANGE(lo, value, hi)
#   define DURANGE(value, limit) URANGE(value, limit)
#   define DASSERTSAME(v1, v2)   ASSERTSAME(v1, v2)
#   define ALLOC_ARRAY(number, type) \
	   (MIAllocDebugFlag ? \
		(fprintf(MIAllocFile, MIAllocFmt, \
		    (MIAllocPtr = malloc((unsigned)(number)*sizeof(type))), \
		    number*sizeof(type), __LINE__, __FILE__), \
		 (type *) MIAllocPtr) : \
		((type *) malloc((unsigned)(number)*sizeof(type))) )
#   define FREE_ARRAY(number, ptr, type) { \
	    free((char *)ptr); \
	    if (MIAllocDebugFlag) \
		fprintf(MIAllocFile, MIFreeFmt, \
		    ptr, number*sizeof(type), __LINE__, __FILE__); }
#   define ALLOC_TYPE(type) 		 ALLOC_ARRAY(1, type)
#   define ALLOC_STRING(slen)		 ALLOC_ARRAY(slen+1, char)
#   define FREE_TYPE(ptr, type)		 FREE_ARRAY(1, ptr, type)
#   define FREE_STRING(str)		 FREE_ARRAY(strlen(str)+1, str, char)
#else
#   define MIMSG(level, msg)		 NOTHING
#   define MIPGM(level, pgm)		 NOTHING
#   define DASSERT(cond)		 NOTHING
#   define DRANGE(lo, value, hi)	 NOTHING
#   define DURANGE(value, limit)	 NOTHING
#   define DASSERTSAME(v1, v2)   ASSERTSAME(v1, v2)
#   define ALLOC_ARRAY(number, type)	 (type *) malloc((unsigned)(number)*sizeof(type))
#   define ALLOC_STRING(slen)		 (String) malloc((unsigned)slen+1)
#   define ALLOC_TYPE(type)		 (type *) malloc((unsigned)sizeof(type))
#   define FREE_ARRAY(number, ptr, type) free((char *)(ptr))
#   define FREE_STRING(str)		 free((char *)(str))
#   define FREE_TYPE(ptr, type)		 FREE_ARRAY(1, ptr, type)
#endif
/* MIInc.h: Variables and macros to time subroutines.
----------------------------------------------------------------------------*/
typedef struct	MIModStatDef	MIModStatDef;
#define	MI_MOD_STAT_DIM   200
#define	MI_TRACE_BACK_DIM 500

struct MIModStatDef {
    String	name;		/* Name of module */
    short	debugFlag;	/* Debug number */
    Boolean	nameAlloFlag;	/* Did a copy of the name get allocated ? */
    Boolean	stopFlag;	/* Stop in module flag */
    Boolean	traceFlag;	/* Trace entry/exit of module ? */
    Boolean	timeFlag;	/* Time module ? */
    long	cTime;		/* Cpu time */
    int		numTime;	/* Number of module invokations */
};

#ifdef MIInc_MAIN
Boolean		MITraceBackFlag	   = FALSE;
Boolean		MIDefaultTraceFlag = FALSE;
Boolean		MIDefaultTimeFlag  = FALSE;
int		MIModThreshold     = 25;
#endif

extern Boolean	MITraceBackFlag;
extern Boolean	MIDefaultTraceFlag;
extern Boolean	MIDefaultTimeFlag;
extern int	MIModThreshold;

short		MINumMod;
MIModStatDef	*MIModStat;
short		MINumTraceBack;
short		*MITraceBack;
short		MICurrentMod;

#ifdef lint
#define MI_MOD_TEST_DEBUG_FLAG(f) if (f) NOTHING
#define MI_REV_STAMP(header, rev) \
    MIPrintRev(header, (String)NULL);
#else
#define MI_MOD_TEST_DEBUG_FLAG(f)
#define MI_REV_STAMP(header, rev) { \
    MIPrintRev(header, rev); \
}
#endif

#ifdef MI_DEBUG_INLINE
/* Use this macro to time any subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_START(name99) \
\
/* String	name99; */ \
{ \
static MIModStatDef *modStat99 = NULL; \
static short	modNum99; \
static short	*MIDebugFlag; \
long 		startTime99, endTime99; \
short		MIFindModIdx(); \
\
if (!modStat99) { \
    modNum99	= MIFindModIdx(name99, FALSE); \
    modStat99	= &MIModStat[modNum99]; \
    MIDebugFlag = &(modStat99->debugFlag); \
    MI_MOD_TEST_DEBUG_FLAG(MIDebugFlag) \
} \
if (MITraceBackFlag) \
then MITraceBack[MINumTraceBack++] = modNum99; \
else MICurrentMod = modNum99; \
if (modStat99->timeFlag) \
    *startTime99 = util_cpu_time(); \
if (modStat99->traceFlag && modStat99->numTime<MIModThreshold) \
    MIfprintf(MIDBG,"--Start %s\n", modStat99->name); \
if (modStat99->stopFlag) { \
    MI_BEEP() \
    MIInitEnv(); \
}

/* ............................... End of macro MI_MOD_START */

/* Use this macro to end any timed subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_END() \
\
if (modStat99->timeFlag) { \
    *endTime99 = util_cpu_time();
    modStat99->cTime += *endTime99;
} \
if (modStat99->traceFlag && modStat99->numTime<MIModThreshold) \
    MIfprintf(MIDBG,"--End   %s\n", modStat99->name); \
modStat99->numTime++; \
if (MITraceBackFlag) \
    MINumTraceBack--; \
} /* ............................. End of macro MI_MOD_END */

#else
#ifdef MI_DEBUG
/* Use this macro to time any subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_START(name99) \
\
/* String	name99; */ \
{ \
static short	modNum99 = -1; \
static short	*MIDebugFlag; \
long 		startTime99; \
\
MIModStart(name99, &modNum99, &MIDebugFlag, &startTime99);
/* ............................... End of macro MI_MOD_START */

/* Use this macro to end any timed subroutine.
----------------------------------------------------------------------------*/
/* void */
#define MI_MOD_END() \
\
MIModEnd(modNum99, &startTime99); \
} /* ............................. End of macro MI_MOD_END */
#else

#define MI_MOD_START(name)
#define MI_MOD_END() NOTHING

#endif
#endif
/* Variables and macros to implement a FIFO queue.
----------------------------------------------------------------------------*/
int		MIQueuePtr;	/* Queue pointer */
int		MIQueueMax;	/* Maximum depth of queue */
int		MIQueueMaxM1;	/* Maximum depth of queue - 1 */
int		MIQueueNum;	/* Number of pointers queued */
RefAny		*MIQueue;	/* Pointer queue */

/* Allocate an n deep pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_QUEUE_ALLOC(n) \
\
/* int		n;	 * Size of queue desired */ \
{ \
MIQueue = ALLOC_ARRAY(n, RefAny); \
MIQueueMax   = n; \
MIQueueMaxM1 = n-1; \
MIQueuePtr   = 0; \
MIQueueNum   = 0; \
} /* ............................. End of macro MI_QUEUE_ALLOC */

/* Free pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_QUEUE_FREE() \
\
FREE_ARRAY(MIQueueMax, MIQueue, RefAny)
/* ............................... End of macro MI_QUEUE_FREE */

/* Put p onto pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_QUEUE(p) \
\
/* int		p;	 * Pointer to be pushed onto queue */ \
{ \
URANGE(MIQueueNum,MIQueueMax) \
MIQueueNum++; \
MIQueue[MIQueuePtr] = (RefAny)(p); \
if (MIQueuePtr<MIQueueMaxM1) \
     MIQueuePtr++; \
else MIQueuePtr = 0; \
} /* ............................. End of macro MI_QUEUE */

/* Pop p from pointer queue.
----------------------------------------------------------------------------*/
/* void */
#define MI_DEQUEUE(type, p) \
\
/* typedef	type;	 * Typedef of pointer p */ \
/* int		p;	 * Pointer to be dequeued */ \
{ \
int		lastQ; \
\
DASSERT(MIQueueNum) \
if (MIQueuePtr>=MIQueueNum) \
     lastQ = MIQueuePtr-MIQueueNum; \
else lastQ = MIQueuePtr-MIQueueNum+MIQueueMax; \
(p) = (type)MIQueue[lastQ]; \
MIQueueNum--; \
} /* ............................. End of macro MI_DEQUEUE */
/* Variables and macros to implement a push down stack.
----------------------------------------------------------------------------*/
int		MIStackPtr;	/* Stack pointer */
int		MIStackMax;	/* Maximum depth of stack */
RefAny		*MIStack;	/* Pointer stack */

/* Allocate an n deep pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_ALLOC(n) \
\
/* int		n;	 * Size of stack desired */ \
{ \
MIStack = ALLOC_ARRAY(n, RefAny); \
MIStackMax = n; \
MIStackPtr = 0; \
} /* ............................. End of macro MI_STACK_ALLOC */

/* Free pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_FREE() \
\
FREE_ARRAY(MIStackMax, MIStack, RefAny)
/* ............................... End of macro MI_STACK_FREE */

/* Push p onto pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_PUSH(p) \
\
/* int		p;	 * Pointer to be pushed onto stack */ \
{ \
URANGE(MIStackPtr,MIStackMax) \
MIStack[MIStackPtr++] = (RefAny)(p); \
} /* ............................. End of macro MI_STACK_PUSH */

/* Pop p from pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_POP(type, p) \
\
/* typedef	type;	 * Typedef of pointer p */ \
/* int		p;	 * Pointer to be popped from stack */ \
{ \
DASSERT(MIStackPtr) \
(p) = (type)MIStack[--MIStackPtr]; \
} /* ............................. End of macro MI_STACK_POP */

/* Get last p from pointer stack.
----------------------------------------------------------------------------*/
/* void */
#define MI_STACK_LAST(type, p) \
\
/* typedef	type;	 * Typedef of pointer p */ \
/* int		p;	 * Pointer to be popped from stack */ \
(p) = (type)MIStack[MIStackPtr];
/* ............................... End of macro MI_STACK_LAST */

/* ............................... End of include MIInc.h */
