/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

atlasRpcMain.c

atlasRpc main interface to rpc.

Author(s):		Mitsuru Igusa
Revision history:	$Log: atlasRpcMain.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:52  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.7  90/02/07  16:10:49  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.6  89/03/12  21:59:12  octtools
 * int -> long for UserMain
 * 
 * Revision 1.5  89/02/07  11:42:07  octtools
 * fixed some lint errors
 * 
 * Revision 1.4  89/02/06  19:09:55  octtools
 * added key bindings
 * 
 * Revision 1.3  89/02/01  17:19:49  octtools
 * added 'NIL' definition
 * 
 * Revision 1.2  89/02/01  11:33:07  octtools
 * added key bindings
 * 
 * Revision 1.1  89/01/27  22:39:42  octtools
 * Initial revision
 * 
 * Revision 1.4  88/04/25  13:07:51  mitch
 * Reset CDSParm.mergeIoFlg=TRUE after channel definition.
 * 
 * Revision 1.3  88/04/05  16:20:44  mitch
 * Fixed bug in initializing CDSMaxNumCh used in printing the sequence file.
 * 
 * Revision 1.2  88/04/05  15:29:57  mitch
 * Used updated procedures in atlas so that atlasRpc now works.
 * Added atlasRpcShow3 which sequences through the channels and atlasRpcOptions
 * which allows the user to turn on/off options.
 * 
 * Revision 1.1  88/03/31  14:21:16  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#ifndef lint
static char RCSId[] = "@(#) $Header: /home/cvs/octtools/src/Layout/atlasRpc/atlasRpcMain.c,v 1.1.1.1 2002/12/31 15:50:52 saj Exp $";
#endif
#include "copyright.h"
#include "port.h"
#include "rpc.h"
#include "oct.h"
#include "MIInc.h"
#include "CDSInc.h"

#ifndef NIL
#define NIL(type)	((type *) 0)
#endif

int		atlasRpcMain(), atlasRpcClear(), atlasRpcShow(),
		atlasRpcShow2(), atlasRpcShow3(), atlasRpcShowC(),
		atlasRpcOptions(), atlasRpcExit();

int		atlasCurChan, atlasLastChan;
struct octBox	atlasBBox;
octId		*atlasChanId;

RPCFunction CommandArray[] = {
    {atlasRpcMain,	"atlas", "start channel definition", NIL(char), 0},
    {atlasRpcClear,	"atlas", "clear channel definition", NIL(char), 0},
    {atlasRpcShow,	"atlas", "show channel definition", NIL(char), 0},
    {atlasRpcShow2,	"atlas", "show channels one by one", NIL(char), 0},
    {atlasRpcShow3,	"atlas", "show next channel", NIL(char), 0},
    {atlasRpcShowC,	"atlas", "clear show", NIL(char), 0},
    {atlasRpcOptions,	"atlas", "channel definition options", NIL(char), 0},
    {atlasRpcExit,	"atlas", "exit", NIL(char), 0},
};

/*ARGSUSED*/
long
UserMain(display, spot, cmdList, uow, array)
char		*display;
RPCSpot		*spot;
lsList		cmdList;
long		uow;
RPCFunction	**array;
{
char		buf[1024];
String		toolName = "atlasRpc";

MIModStatStart(toolName, "debugAtlas", FALSE);

MI_STACK_ALLOC(10000);
MI_QUEUE_ALLOC(10000);

atlasCurChan  = 0;
atlasLastChan = -1;
atlasChanId   = NULL;
atlasBBox.upperRight.x = -1;
atlasBBox.lowerLeft .x = 0;

CDSParm.mergeIoFlg  = TRUE;
CDSParm.zChanFlg    = FALSE;
CDSParm.showChanFlg = FALSE;

MI_REV_STAMP("Channel definition and ordering", RCSId);
sprintf(buf,"Atlas (Channel definition & ordering rev %s) ready.\n", MIBuf);
VEMMSG(buf);
*array = CommandArray;
return(sizeof(CommandArray) / sizeof(RPCFunction));

} /* ............................. End of UserMain procedure */

/*ARGSUSED*/
atlasRpcExit(spot, cmdList, uow)
RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
MI_STACK_FREE();
MI_QUEUE_FREE();

MIModStatEnd();

RPCExit(RPC_OK);
} /* ............................. End of atlasRpcExit procedure */

/* to get lint to shut up */
int dummy()
{
}

/*ARGSUSED*/
atlasRpcOptions(spot, cmdList, uow)
RPCSpot		*spot;
lsList		cmdList;
long		uow;
{
int		numItem = 2;
dmWhichItem	items[2];

MI_MOD_START("atlasRpcOptions");

items[0].itemName = "Allow z or l shaped channels";
items[1].itemName = "Show channels flag";
items[0].flag = (int)CDSParm.zChanFlg;
items[1].flag = (int)CDSParm.showChanFlg;

MIprintf("zChanFlg = %c, showFlg = %c\n",
    MI_CONV_BOOLEAN_TO_CHAR(CDSParm.zChanFlg),
    MI_CONV_BOOLEAN_TO_CHAR(CDSParm.showChanFlg));
if (dmMultiWhich("atlasOptions", numItem, items, dummy, (char *) 0)!=VEM_NOSELECT) {
    CDSParm.zChanFlg    = (items[0].flag!=0);
    CDSParm.showChanFlg = (items[1].flag!=0);
}

MIprintf("zChanFlg = %c, showFlg = %c\n",
    MI_CONV_BOOLEAN_TO_CHAR(CDSParm.zChanFlg),
    MI_CONV_BOOLEAN_TO_CHAR(CDSParm.showChanFlg));

MI_MOD_END();
return(RPC_OK);
} /* ............................. End of atlasRpcOptions procedure */
