/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSInitObjCorn.c

Slice channels for cut lines

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSObjCorn.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:51  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  90/02/07  16:12:18  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/27  22:36:49  octtools
 * Initial revision
 * 
 * Revision 1.2  88/04/05  12:30:32  mitch
 * Fixed bug in z-shaped channels.
 * 
 * Revision 1.1  88/03/31  11:38:49  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "miutil.h"
#include "CDSInc.h"

#define CDS_CORNDIR_TOSTR(d) \
    (d==CDS_E ? "lower left" : (d==CDS_N ? "lower right" : \
    (d==CDS_S ? "upper left" :             "upper right")))

/* Initialize object corners.  This time try to make all z-shaped channels
look straight, since cds/router does not know how to handle terminals that
are located on the side.  This mapping of terminals occurs in the "kink"
of the z-shaped channel.   ALWAYS map 4 DIFFERENT channels to each side of
the instance.   Make sure each instance has 4 sides.
----------------------------------------------------------------------------*/
CDSInitObjCornNew2()
{
CDSTreeDef	*t;
short		ti;
short		o, c;

short		b, minB;
int		x, y, xy, minXY;

CDSTreeDef	*t2, *tp, *tb;
CDSTreeDef	**treeList;
short		k;
Boolean		horChanFlag;
Boolean		onTopFlag;
Boolean		CDSWhosOnTop();

MI_MOD_START("CDSInitObjCornNew2");

DASSERT(!MIStackPtr);

					/* Initialize treeList */
treeList = ALLOC_ARRAY(CDSHead.numTree, CDSTreeDef *);
for (ti=0; ti<CDSHead.numTree; ti++)
    treeList[ti] = NULL;
MI_STACK_PUSH(CDSTree);
while (MIStackPtr) {
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    DRANGE(0, t->id, CDSHead.numTree);
    treeList[t->id] = t;
    if (t->type!=CDS_TREE_ELEMENT) {
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	continue;
    }
}

MI_STACK_PUSH(CDSTree);
while (MIStackPtr) {			/* Push until all nodes sliced */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    if (t->type!=CDS_TREE_ELEMENT) {
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	continue;
    }
    o = t->objNum;
    if (o<0) continue;
					/* Init corners for object o */
    MSG(1, ("Finding corners of object%d name=%s\n", o, CDSObj[o].dbRef));
    DRANGE(0,o,CDSHead.numObj-1);

    CDSObj[o].numCorn = 4;
    CDSObj[o].corn = ALLOC_ARRAY(CDSObj[o].numCorn, CDSCornDef);

    for (c=0; c<CDSObj[o].numCorn; c++)
	CDSObj[o].corn[c].treeId = CDS_NULL_TREEID;

    x = CDSObj[o].bdy[0][CDS_X];
    y = CDSObj[o].bdy[0][CDS_Y];
					/* Init corn[].treeId */
    tp = t;
    ti = tp->parentId;
    ASSERT(ti!=CDS_NULL_TREEID);
    t2 = treeList[ti];

    for (k=0; k<4; k++) {

	while (t2) {

	    switch(ti) {
	    case CDS_N_EXT_TREEID:
		c = 2;
		break;
	    case CDS_S_EXT_TREEID:
		c = 0;
		break;
	    case CDS_W_EXT_TREEID:
		c = 1;
		break;
	    case CDS_E_EXT_TREEID:
		c = 3;
		break;
	    default:
		ASSERT(t2->type>=2);

		if (t2->slice[0][CDS_X]!=t2->slice[1][CDS_X])
		     horChanFlag = TRUE;
		else horChanFlag = FALSE;
					/* Find what the fuck is below what */
		if (horChanFlag) {
		    onTopFlag = CDSWhosOnTop(t2, horChanFlag, x, y);
		    if (onTopFlag)
			 c = 2;
		    else c = 0;
		}
		else {
		    onTopFlag = CDSWhosOnTop(t2, horChanFlag, x, y);
		    if (onTopFlag)
			 c = 1;
		    else c = 3;
		}
	    }

	    if (CDSObj[o].corn[c].treeId==CDS_NULL_TREEID) {
		CDSObj[o].corn[c].treeId = ti;
		break;
	    }

	    tp = t2;
	    ti = tp->parentId;
	    ASSERT(ti!=CDS_NULL_TREEID);
	    if (tp!=CDSTree)
		 t2 = treeList[ti];
	    else t2 = NULL;
	}

	if (!t2) {
	    for (c=0; c<CDSObj[o].numCorn; c++)
		if (CDSObj[o].corn[c].treeId==CDS_NULL_TREEID) break;
	    DASSERT(c<CDSObj[o].numCorn);
	    MIerror("Cannot find channel on side %s of instance %s\n",
		CDS_CORN_TOSTR(c), CDSObj[o].dbRef);
	    exit(4);
	}
    }

    for (c=0; c<CDSObj[o].numCorn; c++) {

	minXY = MAXINT;			/* Find corner of instance polygon */
	for (b=0; b<CDSObj[o].numBdy; b++) {
	    x = CDSObj[o].bdy[b][CDS_X];
	    y = CDSObj[o].bdy[b][CDS_Y];
	    switch(c) {
	    case 0:  xy =  x+y; break;
	    case 1:  xy =  x-y; break;
	    case 2:  xy = -x-y; break;
	    default: xy = -x+y; break;
	    }
	    if (xy<minXY) {		/* Save best bdy point */
		minXY = xy;
		minB  = b;
	    }
	}
	DASSERT(minXY<MAXINT);
	MSG(3, ("Corner%d of bdy %d: (%d %d) treeId=%d\n",
	    c, minB, CDSObj[o].bdy[minB][CDS_X], CDSObj[o].bdy[minB][CDS_Y],
	    CDSObj[o].corn[c].treeId));
					/* Update corn array */
	CDSObj[o].corn[c].bdyIdx = minB;
    }

    if (*debugFlag>=5)
	CDSPrintNGraph(t->node);
}

FREE_ARRAY(CDSHead.numTree, treeList, CDSTreeDef *);

MI_MOD_END();
} /* ............................. End of procedure CDSInitObjCornNew2 */
/* Trying to figure out which child is ABOVE the other child of node t.
This should be trivial ... but due to z-shaped channels ... its a bitch.
----------------------------------------------------------------------------*/
Boolean
CDSWhosOnTop(t, horFlag, x, y)

CDSTreeDef	*t;
Boolean		horFlag;
int		x, y;
{
int		s;
int		x1, x2, xl, xh, xMin, xMax, xt;
int		y1, y2, yl, yh, yMin, yMax, yt;
Boolean		chanOnTopFlag = TRUE;

if (horFlag) {
    xMin =  MAXINT;
    xMax = -MAXINT;
    for (s=0; s<t->type; s++) {
	x1 = t->slice[s  ][CDS_X];
	x2 = t->slice[s+1][CDS_X];
	if (x1==x2) continue;

	yt = t->slice[s][CDS_Y];

	xl = MIN(x1, x2);
	xh = MAX(x1, x2);

	if (xl<xMin) {
	    xMin = xl;
	    yMin = yt;
	}
	if (xh>xMax) {
	    xMax = xh;
	    yMax = yt;
	}

	if (x>=xl && x<=xh)
	    chanOnTopFlag = (y<=yt);
    }
    if (x<xMin)
	chanOnTopFlag = (y<=yMin);
    if (x>xMax)
	chanOnTopFlag = (y<=yMax);
}
else {
    yMin =  MAXINT;
    yMax = -MAXINT;
    for (s=0; s<t->type-1; s++) {
	y1 = t->slice[s  ][CDS_Y];
	y2 = t->slice[s+1][CDS_Y];
	if (y1==y2) continue;

	xt = t->slice[s][CDS_X];

	yl = MIN(y1, y2);
	yh = MAX(y1, y2);

	if (yl<yMin) {
	    yMin = yl;
	    xMin = xt;
	}
	if (yh>yMax) {
	    yMax = yh;
	    xMax = xt;
	}

	if (y>=yl && y<=yh)
	    chanOnTopFlag = (x<=xt);
    }
    if (y<yMin)
	chanOnTopFlag = (x<=xMin);
    if (y>yMax)
	chanOnTopFlag = (x<=xMax);
}

return(chanOnTopFlag);
} /* ............................. End of procedure CDSWhosOnTop */
/* Initialize object corners.
----------------------------------------------------------------------------*/
int
CDSInitObjCornNew()
{
CDSTreeDef	*t;
CDSGNodeDef	*g, *gn, *startGn;
short		k, rpd;
short		ti, pti;
short		o, c;
short		ori = CDS_CCW;

short		lastDir, cornDir, tempDir;
Boolean		straightFlg = FALSE;

short		cornDim;
short		numCorn;
CDSCornDef	corn[40];

short		b, minB;
int		x, x0;
int		y, y0;
Boolean		xFlg, yFlg;
int		xy, minXY;

int		rc = 0;

MI_MOD_START("CDSInitObjCornNew");

DASSERT(!MIStackPtr);

MI_STACK_PUSH(CDSTree);

while (MIStackPtr) {			/* Push until all nodes sliced */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    if (t->type!=CDS_TREE_ELEMENT) {
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	continue;
    }
    o = t->objNum;
    if (o<0) continue;
					/* Init corners for object o */
    MSG(1, ("Finding corners of object%d name=%s\n", o, CDSObj[o].dbRef));
    DRANGE(0,o,CDSHead.numObj-1);

    startGn = t->node;
    DASSERT(startGn->bch[CDS_E].node && startGn->bch[CDS_N].node);

    g   = startGn;			/* Go around graph ctrClkWise */
    rpd = CDS_N;			/*   starting in lower left corner */
    pti = g->bch[rpd].treeId;		/*   until valid corner is found. */
					/*   Setting up g, rpd, pti and */
					/*   lastDir. */
    for (gn=NULL; gn!=startGn; g=gn) {
	DASSERT(g->bch[rpd].objNum==o);
	for (k=CDS_NEXTDIR(rpd,ori); !g->bch[k].node; k=CDS_NEXTDIR(k,ori));
	gn = g->bch[k].node;
	ti = g->bch[k].treeId;
	if (ti!=pti && rpd==CDS_NEXTDIR(k,ori)) {
	    lastDir = k;
	    startGn = gn;
	}
	pti = ti;
	rpd = CDS_REVDIR(k);
    }
    startGn = g;
    DASSERT(startGn!=t->node);

    numCorn = 0;
    cornDim = 4;
					/* Go all the way around graph */
    for (gn=NULL; gn!=startGn; g=gn) {	/* Go around instance channel graph */

	DASSERT(g->bch[rpd].objNum==o);
	for (k=CDS_NEXTDIR(rpd,ori); !g->bch[k].node; k=CDS_NEXTDIR(k,ori));

	gn = g->bch[k].node;
	ti = g->bch[k].treeId;

	if (ti!=pti) {			/* New corner (different channels) */
	    cornDir = k;
	    tempDir = CDS_NEXTDIR(lastDir, ori);
					/* Check if direction ok */
	    if (rpd!=CDS_NEXTDIR(k,ori)) {
					/* Corner is on straight line segment */
		straightFlg = TRUE;
		if (rpd==CDS_REVDIR(k)) {
		    MSG(1, ("Corner node of obj %s is on straight line\n",
			CDSObj[o].dbRef));
		    cornDir = tempDir;

		    if (numCorn==cornDim) {
			MSG(0, ("WARNING Already enough corners - ignoring this\n"));
			rc = 1;
			break;
		    }
		}
		else {			/* Bad news - check it out */
		    MIerror("Cannot figure out corners of instance %s\n",
			CDSObj[o].dbRef);
		    abort();
		}
	    }
					/* Corner repeated - take this one */
	    else if (cornDir==lastDir) {
		MSG(0, ("Repeating corner %s of obj %s\n",
		    CDS_CORNDIR_TOSTR(cornDir), CDSObj[o].dbRef));
		numCorn--;
		if (!straightFlg) {	/* Was not straight before then */
					/*   something might be wrong */
		    MSG(0, ("WARNING Corner %s of inst %s repeated\n",
			CDS_CORNDIR_TOSTR(cornDir), CDSObj[o].dbRef));
		    rc = 1;
		}
		straightFlg = FALSE;
	    }

	    else if (cornDir!=tempDir) {
		MSG(0, ("WARNING Skipped %s corner of inst %s\n",
		    CDS_CORNDIR_TOSTR(tempDir), CDSObj[o].dbRef));
		rc = 1;
		cornDim--;
		straightFlg = FALSE;
	    }
					/* These coordinates have problems */
					/*   when dealing with straights */
	    x0 = g->loc[CDS_X];
	    y0 = g->loc[CDS_Y];

	    MSG(3, ("New corner %s node=%d (%d %d) dirPrev=%s dirNext=%s\n",
		CDS_CORNDIR_TOSTR(cornDir), CDS_ID(g), x0, y0,
		CDS_CDIR_TOSTR(rpd), CDS_CDIR_TOSTR(k)));

	    xFlg = yFlg = TRUE;		/* Direction cornDir determines which */
	    switch (cornDir) {		/*   which corner we are at */
	    case CDS_E:	break;		/* Corner at min( x+y) x>x0 & y>y0 */
	    case CDS_N:	xFlg = FALSE;	/* Corner at min(-x+y) x<x0 & y>y0 */
			break;
	    case CDS_W:	xFlg = FALSE;	/* Corner at min(-x-y) x<x0 & y<y0 */
	    default   :	yFlg = FALSE;	/* Corner at min( x-y) x>x0 & y<y0 */
	    }

	    minXY = MAXINT;		/* Find corner of instance polygon */
	    for (b=0; b<CDSObj[o].numBdy; b++) {
		x = CDSObj[o].bdy[b][CDS_X];
		if (xFlg) {
		    /* if (x<x0) continue; */
		    xy = x;
		}
		else {
		    /* if (x>x0) continue; */
		    xy = -x;
		}

		y = CDSObj[o].bdy[b][CDS_Y];
		if (yFlg) {
		    /* if (y<y0) continue; */
		    xy += y;
		}
		else {
		    /* if (y>y0) continue; */
		    xy -= y;
		}

		if (xy<minXY) {		/* Save best bdy point */
		    minXY = xy;
		    minB  = b;
		}
	    }
	    DASSERT(minXY<MAXINT);
	    MSG(3, ("Corner%d of bdy %d: (%d %d) treeId=%d\n",
		numCorn, minB,
		CDSObj[o].bdy[minB][CDS_X], CDSObj[o].bdy[minB][CDS_Y], ti));
					/* Update corn array */
	    corn[numCorn].bdyIdx = minB;
	    corn[numCorn].treeId = ti;
	    numCorn++;

	    lastDir = cornDir;
	}

	pti = ti;
	rpd = CDS_REVDIR(k);
    }
					/* Copy corn from temp array */
    if (numCorn!=cornDim) {
	MSG(0, ("WARNING Inst %s has %d surrounding channels (should be %d)\n",
	    CDSObj[o].dbRef, numCorn, cornDim));
	rc = 1;
    }
    CDSObj[o].numCorn = numCorn;
    CDSObj[o].corn    = ALLOC_ARRAY(numCorn, CDSCornDef);
    for (c=0; c<numCorn; c++) {
	CDSObj[o].corn[c].bdyIdx = corn[c].bdyIdx;
	CDSObj[o].corn[c].treeId = corn[c].treeId;
    }

    if (*debugFlag>=5)
	CDSPrintNGraph(t->node);
}

MI_MOD_END();
return(rc);
} /* ............................. End of procedure CDSInitObjCornNew */
/* Initialize object corners.
----------------------------------------------------------------------------*/
CDSInitObjCorn()
{
#define		CDS_DIM_CORN 40
CDSTreeDef	*t;
CDSGNodeDef	*g, *gn;
short		k, rpd;
short		ti, pti;
short		o, c;
short		ori = CDS_CCW;

CDSGNodeDef	*lastGCorn;
short		lastDir, lastDir2;

short		numCorn;
CDSCornDef	corn[CDS_DIM_CORN];

short		b, minB;
int		x, x0;
int		y, y0;
Boolean		xFlg, yFlg;
int		xy, minXY;

MI_MOD_START("CDSInitObjCorn");

DASSERT(!MIStackPtr);

MI_STACK_PUSH(CDSTree);

while (MIStackPtr) {			/* Push until all nodes sliced */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    if (t->type!=CDS_TREE_ELEMENT) {
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	continue;
    }
    o = t->objNum;
    if (o<0) continue;
					/* Init corners for object o */
    MSG(1, ("Finding corners of object %d\n", o));
    DRANGE(0,o,CDSHead.numObj-1);
    numCorn = 0;
    g   = t->node;			/* Go around graph counterClkWise */
    rpd = CDS_N;
    pti = g->bch[CDS_N].treeId;
    DASSERT(g->bch[CDS_E].node && g->bch[CDS_N].node);
    lastGCorn = g;
    lastDir   = CDS_E;

    for (gn=NULL; gn!=t->node; g=gn) {
	DASSERT(g->bch[rpd].objNum==o);
	for (k=CDS_NEXTDIR(rpd,ori); !g->bch[k].node; k=CDS_NEXTDIR(k,ori));

	gn = g->bch[k].node;
	ti = g->bch[k].treeId;

	if (ti!=pti) {			/* New corner */
	    if (rpd!=CDS_NEXTDIR(k,ori)) {
		DASSERT(rpd==CDS_REVDIR(k));
		MSG(2, ("Corner node of obj %d is on straight line\n", o));
		lastDir2 = CDS_NEXTDIR(lastDir,ori);
		if (lastGCorn->bch[lastDir ].treeId==
		    lastGCorn->bch[lastDir2].treeId) {
		    g = lastGCorn;
		    if (k!=lastDir)
			MIerror("Corners will be screwed up in obj %d\n", o);
		}
		else continue;
	    }

	    x0 = g->loc[CDS_X];
	    y0 = g->loc[CDS_Y];
	    MSG(3, ("New corner node=%d (%d %d) dirPrev=%s dirNext=%s\n",
		CDS_ID(g), x0, y0, CDS_CDIR_TOSTR(rpd), CDS_CDIR_TOSTR(k)));

	    xFlg = yFlg = TRUE;
	    switch (k) {
	    case CDS_E:	break;		/* Corner at min( x+y) x>x0 & y>y0 */
	    case CDS_N:	xFlg = FALSE;	/* Corner at min(-x+y) x<x0 & y>y0 */
			break;
	    case CDS_W:	xFlg = FALSE;	/* Corner at min(-x-y) x<x0 & y<y0 */
	    default   :	yFlg = FALSE;	/* Corner at min( x-y) x>x0 & y<y0 */
	    }

	    minXY = MAXINT;		/* Find corner */
	    for (b=0; b<CDSObj[o].numBdy; b++) {
		x = CDSObj[o].bdy[b][CDS_X];
		if (xFlg) {
		    if (x<x0) continue;
		    xy = x;
		}
		else {
		    if (x>x0) continue;
		    xy = -x;
		}

		y = CDSObj[o].bdy[b][CDS_Y];
		if (yFlg) {
		    if (y<y0) continue;
		    xy += y;
		}
		else {
		    if (y>y0) continue;
		    xy -= y;
		}

		if (xy<minXY) {		/* Save best bdy point */
		    minXY = xy;
		    minB  = b;
		}
	    }
	    DASSERT(minXY<MAXINT);
	    MSG(3, ("Corner%d of bdy %d: (%d %d) treeId=%d\n",
		numCorn, minB,
		CDSObj[o].bdy[minB][CDS_X], CDSObj[o].bdy[minB][CDS_Y], ti));
					/* Update corn array */
	    corn[numCorn].bdyIdx = minB;
	    corn[numCorn].treeId = ti;
	    numCorn++;
	}

	if (rpd==CDS_NEXTDIR(k,ori)) {
	    lastGCorn = g;
	    lastDir   = k;
	}

	pti = ti;
	rpd = CDS_REVDIR(k);
    }
					/* Copy corn from temp array */
    CDSObj[o].numCorn = numCorn;
    CDSObj[o].corn    = ALLOC_ARRAY(numCorn, CDSCornDef);
    for (c=0; c<numCorn; c++) {
	CDSObj[o].corn[c].bdyIdx = corn[c].bdyIdx;
	CDSObj[o].corn[c].treeId = corn[c].treeId;
    }

    if (*debugFlag>=5)
	CDSPrintNGraph(t->node);
}

MI_MOD_END();
} /* ............................. End of procedure CDSInitObjCorn */
