
/*
  #ifndef lint
  static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/dijkstra.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
  #endif
  */

/*
 *  dijkstra.c
 *  **********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for obtaining a shortest path between a point
 *	and a target, which is a set of points.
 *
 *  functions contained in this file:
 *	dijkstra()
 *	dijkstra_FH()
 *	askReach()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"




/*
 *  dijkstra()
 *
 *  DESCRIPTION
 *  -----------
 *	Obtain a shortest path from a node, v, to a set of nodes, target.
 *
 *  RETURNS
 *  -------
 *	node		when a shortest path is found
 *	NOT_FOUND	otherwise
 *
 *	The node retuned is the node that was reached first during the search.
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: starting node
 *	target	<input>	: a set of nodes to be reached
 *	nodeP	<input>	: list of pointers to adjacency lists
 *			  (represents a graph)
 *
 *  Local variables:
 *  ---------------
 *	mind		: min distance among candidate nodes
 *	miny		: node with min distance among candidates
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Uses dijkstraPointer[] and fills in values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	dijkstraPointer[]	: struct for shortest path algorithm
 *
 */


dijkstra(v,target,nodeP)

int		v;
netNodeType	*target;
adjNodeType	*nodeP[];

{
    
    int		i,w,x,y;
    int		node[G_MAXNODE];
    int		mind,miny;
    int		distance;
    adjNodeType	*nextNode;
    
    
    
    /* Initialization */
    /*----------------*/
    x = v;
    v2link0 = NULL;
    for (i = 0; i <= G_MAXNODE - 1; ++i) {
	dijkstraPointer[i]->v2link = NULL;
	dijkstraPointer[i]->d = MAXINT;
	dijkstraPointer[i]->parent = 0;
	dijkstraPointer[i]->vset = 3;
	dijkstraPointer[i]->adjlist = NULL;
    }
    dijkstraPointer[v-1]->d = 0;
    dijkstraPointer[v-1]->vset = 1;
    
    
    /* Get a shortest path */
    /*---------------------*/
    /* The following is a single node target version and not used */
    /* while (x != w) { */
    
    for ( ; askReach(x,target) != YES; x = miny) {
	
	/* Maintain V2 and V3 set */
	/*------------------------*/
	for (nextNode = nodeP[x-1]; nextNode != NULL; nextNode = nextNode->next) {
	    
	    /* Sometimes the edges might be blocked and not usable */
	    /*-----------------------------------------------------*/
	    if (nodeBlock[x-1] == YES) {
		break;
	    }
	    if (nextNode->blocked == YES) {
		continue;
	    }
	    
	    y = nextNode->id;
	    if (dijkstraPointer[y-1]->vset == 2 &&
		dijkstraPointer[x-1]->d + nextNode->weight 
		< dijkstraPointer[y-1]->d) {
		dijkstraPointer[y-1]->parent = x;
		dijkstraPointer[y-1]->d = dijkstraPointer[x-1]->d 
		  + nextNode->weight;
	    }
	    if (dijkstraPointer[y-1]->vset == 3) {
		dijkstraPointer[y-1]->vset = 2;
		dijkstraPointer[y-1]->v2link = v2link0;
		v2link0 = dijkstraPointer[y-1];
		dijkstraPointer[y-1]->parent = x;
		dijkstraPointer[y-1]->d = dijkstraPointer[x-1]->d 
		  + nextNode->weight;
	    }
	}
	
	if (v2link0 == NULL) {
	    
#ifdef DIJKSTRA_DEBUG
	    printf("No path was found from %d to %d\n\n",v,target->id);
#endif
	    
	    return(NOT_FOUND);
	}
	
	/* Get the candidate with the smallest distance */
	/*----------------------------------------------*/
	mind = MAXINT;
	for (i = 0; i <= G_MAXNODE - 1; ++i) {
	    if (dijkstraPointer[i]->vset == 2 && 
		mind > dijkstraPointer[i]->d) {
		mind = dijkstraPointer[i]->d;
		miny = i + 1;
	    }
	}
	
#ifdef DIJKSTRA_DEBUG
	printf("miny = %d(%d)\n", miny, mind);
#endif
	
	/* Move miny from V2 to V1 */
	/*-------------------------*/
	dijkstraPointer[miny-1]->vset = 1;
	if (v2link0 == dijkstraPointer[miny-1]) {
	    v2link0 = dijkstraPointer[miny-1]->v2link;
	    dijkstraPointer[miny-1]->v2link = NULL;
	    continue;
	}
	for (i = 0; i <= G_MAXNODE - 1; ++i) {
	    if (dijkstraPointer[i]->v2link == dijkstraPointer[miny-1]) {
		dijkstraPointer[i]->v2link = dijkstraPointer[miny-1]->v2link;
		dijkstraPointer[miny-1]->v2link = NULL;
		break;
	    }
	}
    }
    w = x;
    distance = dijkstraPointer[w-1]->d;
    
    
    /* Print the nodes in the shortest path obtained */
    /*-----------------------------------------------*/
    
#ifdef DIJKSTRA_DEBUG
    
    for (i = 0; i <= G_MAXNODE - 1; ++i) {
	node[i] = 0;
    }
    i = 0;
    while (x != v) {
	node[i++] = x;
	x = dijkstraPointer[x-1]->parent;
    }
    node[i] = v;
    
    printf("A shortest path from %d to %d  length = %d\n",v,w,distance);
    for (i = G_MAXNODE - 1; i >= 0; --i) {
	if (node[i] != 0) {
	    printf("%d ",node[i]);
	}
    }
    printf("\n\n");
    
#endif
    
    
    return(w);
    
    
}



/*
 *  dijkstra_FH()
 *
 *  DESCRIPTION
 *  -----------
 *	Obtain a shortest path from a node, v, to a set of nodes, target.
 *	This is a Fibonacci heap version of the algorithm.
 *
 *  RETURNS
 *  -------
 *	node		when a shortest path is found
 *	NOT_FOUND	otherwise
 *
 *	The node returned is the node that was reached first during the search.
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: starting node
 *	target	<input>	: a set of nodes to be reached
 *	nodeP	<input>	: list of pointers to adjacency lists
 *			  (represents a graph)
 *
 *  Local variables:
 *  ---------------
 *	mind		: min distance among candidate nodes
 *	miny		: node with min distance among candidates
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Uses dijkstraPointer[] and fills in values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	dijkstraPointer[i]	: struct for shortest path algorithm
 *	id2heapNode[i]		: pointer to a heap node with id i
 *	nodeBlock[i]		: flag, if YES, no edge is usable from this node
 *
 */


dijkstra_FH(v,target,nodeP)

int		v;
netNodeType	*target;
adjNodeType	*nodeP[];

{
    
    int		i,w,x,y;
    int		node[G_MAXNODE];
    int		distance;
    adjNodeType	*nextNode;
    
    FH_header	*FH_headerP;
    FH_node		*FH_nodeP;
    netNodeType	*nextNetNode;
    nodeGroupType	*nextEqNode;
    short		targetNode[G_MAXNODE];
    
    
    /* Initialization */
    /*----------------*/
    x = v;
    for (i = 0; i <= G_MAXNODE - 1; ++i) {
	dijkstraPointer[i]->v2link = NULL;
	dijkstraPointer[i]->d = MAXINT;
	dijkstraPointer[i]->parent = 0;
	dijkstraPointer[i]->vset = 3;
	dijkstraPointer[i]->adjlist = NULL;
	targetNode[i] = NO;
    }
    dijkstraPointer[v-1]->d = 0;
    dijkstraPointer[v-1]->vset = 1;
    
    for (nextNetNode = target; nextNetNode != NULL; 
	 nextNetNode = nextNetNode->next) {
	targetNode[nextNetNode->id - 1] = YES;
	for (nextEqNode = nextNetNode->equiv; nextEqNode != NULL;
	     nextEqNode = nextEqNode->next) {
	    targetNode[nextEqNode->id - 1] = YES;
	}
    }
    
    
    /* Get a shortest path */
    /*---------------------*/
    FH_headerP = FH_makeheap();
    FH_insert(FH_headerP,v,0.0);
    
    for (;;) {
	
	/* Get the candidate with the smallest distance */
	/*----------------------------------------------*/
	if ((FH_nodeP = FH_findmin(FH_headerP)) == NULL) {
	    
#ifdef DIJKSTRA_DEBUG
	    printf("No path was found from %d to %d\n\n",v,target->id);
#endif
	    
	    FH_free();
	    return(NOT_FOUND);
	}
	
	x = FH_nodeP->id;
	dijkstraPointer[x - 1]->vset = 1;
	FH_deletemin(FH_headerP);
	if (targetNode[x - 1] == YES) {
	    w = x;
	    break;
	}
	for (nextNode = nodeP[x-1]; nextNode != NULL; nextNode = nextNode->next) {
	    
	    /* Sometimes the edges might be blocked and not usable */
	    /*-----------------------------------------------------*/
	    if (nodeBlock[x-1] == YES) {
		break;
	    }
	if (nextNode->blocked == YES) {
	    continue;
	}

	y = nextNode->id;
	if (dijkstraPointer[y-1]->vset == 2 &&
	    dijkstraPointer[x-1]->d + nextNode->weight 
				< dijkstraPointer[y-1]->d) {
	    dijkstraPointer[y-1]->parent = x;
	    FH_decreasekey(FH_headerP, id2heapNode[y-1],
		(float)(dijkstraPointer[y-1]->d 
			- (dijkstraPointer[x-1]->d + nextNode->weight)));
	    dijkstraPointer[y-1]->d = dijkstraPointer[x-1]->d 
				+ nextNode->weight;
	}
	if (dijkstraPointer[y-1]->vset == 3) {
	    dijkstraPointer[y-1]->vset = 2;
	    dijkstraPointer[y-1]->parent = x;
	    dijkstraPointer[y-1]->d = dijkstraPointer[x-1]->d 
				+ nextNode->weight;
	    FH_insert(FH_headerP,y,
		(float)(dijkstraPointer[x-1]->d + nextNode->weight));
	}
    }
}
FH_free();


/* Print the nodes in the shortest path obtained */
/*-----------------------------------------------*/
#ifdef DIJKSTRA_DEBUG

distance = dijkstraPointer[w-1]->d;
for (i = 0; i <= G_MAXNODE - 1; ++i) {
    node[i] = 0;
}
i = 0;
while (x != v) {
    node[i++] = x;
    x = dijkstraPointer[x-1]->parent;
}
node[i] = v;

printf("A shortest path from %d to %d  length = %d\n",v,w,distance);
for (i = G_MAXNODE - 1; i >= 0; --i) {
    if (node[i] != 0) {
	printf("%d ",node[i]);
    }
}
printf("\n\n");

#endif


return(w);


}




/*
 *  askReach()
 *
 *  DESCRIPTION
 *  -----------
 *	Check if the node belongs to target nodes to see if the search has
 *	reached any of the target nodes.
 *
 *  RETURNS
 *  -------
 *	YES	: when the node reached belongs to target
 *	NO	: otherwise
 *
 *  Arguments:
 *  ---------
 *	x	<input>	: the node the search has reached
 *	target	<input>	: a set of destination nodes to be reached
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


askReach(x,target)

int		x;
netNodeType	*target;

{

netNodeType	*nextNode;
nodeGroupType	*nextEqNode;

for (nextNode = target; nextNode != NULL; nextNode = nextNode->next) {
    if (x == nextNode->id) {
	return(YES);
    }
    for (nextEqNode = nextNode->equiv; nextEqNode != NULL;
		nextEqNode = nextEqNode->next) {
	if (x == nextEqNode->id) {
	    return(YES);
	}
    }
}
return(NO);
	

}





/*
 *  shortestPath()
 *
 *  DESCRIPTION
 *  -----------
 *	Driver routine to get a shortest path, using either a regular dijkstra
 *	algorithm or Fibonacci heap version of the algorithm.
 *
 *  RETURNS
 *  -------
 *	node		when a shortest path is found
 *	NOT_FOUND	otherwise
 *
 *	The node retuned is the node that was reached first during the search.
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: starting node
 *	target	<input>	: a set of nodes to be reached
 *	nodeP	<input>	: list of pointers to adjacency lists
 *			  (represents a graph)
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Uses dijkstraPointer[] and fills in values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


shortestPath(v,target,nodeP)

int		v;
netNodeType	*target;
adjNodeType	*nodeP[];

{


if (parm.FH_dijkstra == ON) {
    return(dijkstra_FH(v,target,nodeP));
} else {
    return(dijkstra(v,target,nodeP));
}



}



