
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/fiboHeap.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  fiboHeap.c
 *  **********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for implementing Fibonacci heap
 *
 *  functions contained in this file:
 *	FH_makeheap()
 *	FH_findmin()
 *	FH_insert()
 *	FH_meld()
 *	FH_deletemin()
 *	FH_decreasekey()
 *	FH_delete()
 *	FH_free()
 *
 *	Not visible from users
 *
 *	FH_disconnect()
 *	FH_concatenate()
 *	FH_binaryadd()
 *	FH_binaryadd1()
 *	FH_link()
 *
 *  Import descriptions:
 *      initGlb.h       : macros for mercury global router
 *	fiboHeap.h	: macros for Fibonacci heap
 *
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"



/*
 *  FH_makeheap()
 *
 *  DESCRIPTION
 *  -----------
 *	Make a header to the new heap
 *
 *  RETURNS
 *  -------
 *	heapP		: pointer to a new heap
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_header	*FH_makeheap()


{


    FH_header	*headerP;
    int		i;


    headerP = MALLOC(FH_header,1);
    *headerP = NULL;
    for (i = 0; i <= MAXHEAPSIZE - 1; ++i) {
	id2heapNode[i] = NULL;
    }

    return(headerP);

}




/*
 *  FH_findmin()
 *
 *  DESCRIPTION
 *  -----------
 *	Find the node with a minimum key value in the heap
 *
 *  RETURNS
 *  -------
 *	minNode		: pointer to a minimum key node
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */


FH_node		*FH_findmin(headerP)

FH_header	*headerP;


{


#ifdef FH_DEBUG
    if (*headerP != NULL) {
	printf("\nFindmin %d (key = %.1f)\n", (*headerP)->id, (*headerP)->key);
    } else {
	printf("Heap is empty\n");
    }
#endif

    return *headerP;

}




/*
 *  FH_insert()
 *
 *  DESCRIPTION
 *  -----------
 *	Create a new node in the specified heap
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *	id	<input>	: id of the new node
 *	key	<input>	: key of the new node
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_insert(headerP,id_in,key_in)

FH_header	*headerP;
int		id_in;
float		key_in;


{


    FH_node		*nodeP;


    id2heapNode[id_in - 1] = nodeP = MALLOC(FH_node,1);
    nodeP->id = id_in;
    nodeP->key = key_in;
    nodeP->rank = 0;
    nodeP->counter = -1;
    nodeP->parent = NULL;
    nodeP->child = NULL;
    nodeP->nextSibling = nodeP;
    nodeP->preSibling = nodeP;

    FH_concatenate(*headerP,nodeP);

    if (*headerP == NULL) {
	*headerP = nodeP;
    } else {
	if ((*headerP)->key > key_in) {
	    *headerP = nodeP;
	}
    }

#ifdef FH_DEBUG
    printf("Insert node %d (key = %.1f) to heap\n", id_in, key_in);
    FH_print(headerP, NULL);
#endif

    return;

}




/*
 *  FH_meld()
 *
 *  DESCRIPTION
 *  -----------
 *	Meld or merge two heaps.
 *	Result is stored in *headerP1. And *headerP2 is freed.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP1<input>	: pointer to the header of heap1
 *	headerP2<input>	: pointer to the header of heap2
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */


FH_meld(headerP1,headerP2)


FH_header	*headerP1;
FH_header	*headerP2;


{


    if (*headerP2 == NULL) {
	FREE(headerP2);
	return;
    }
    if (*headerP1 == NULL) {
	*headerP1 = *headerP2;
	FREE(headerP2);
	return;
    }

    FH_concatenate(*headerP1,*headerP2);
    if ((*headerP1)->key > (*headerP2)->key) {
	*headerP1 = *headerP2;
    }

    FREE(headerP2);
    return;


}




/*
 *  FH_deletemin()
 *
 *  DESCRIPTION
 *  -----------
 *	Delete a node with the smallest key value from the heap.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_deletemin(headerP)


FH_header	*headerP;


{


    FH_node		*saveMin;


    if (*headerP == NULL) {
	return;
    }

    saveMin = *headerP;

    /* Update header */
    /*---------------*/
    if ((*headerP)->nextSibling != *headerP) {
	*headerP = (*headerP)->nextSibling;
    } else if ((*headerP)->child != NULL) {
	*headerP = (*headerP)->child;
    } else {
	*headerP = NULL;
    }

    /* Delete the body of the min node */
    /*---------------------------------*/
    FH_destroy(headerP,saveMin);

    /* Do binary addition */
    /*--------------------*/
    FH_binaryadd(headerP);


}




/*
 *  FH_disconnect()
 *
 *  DESCRIPTION
 *  -----------
 *	Disconnect the specified node in the heap.
 *	This does not mean that the node is deleted and freed.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	nodeP<input>	: pointer to the node to be deleted
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */

FH_disconnect(headerP,nodeP)

FH_header	*headerP;
FH_node		*nodeP;

{

    FH_node		*saveParent;


    /* Update pointers in this node and its parent */
    /*---------------------------------------------*/
    saveParent = nodeP->parent;
    if ((saveParent != NULL) && (saveParent->child == nodeP)) {
	if (nodeP->nextSibling == nodeP) {
	    saveParent->child = NULL;
	} else {
	    saveParent->child = nodeP->nextSibling;
	}
    }
    if (saveParent != NULL) {
	--(saveParent->rank);
    }
    nodeP->parent = NULL;
    nodeP->counter = -1;

    nodeP->nextSibling->preSibling = nodeP->preSibling;
    nodeP->preSibling->nextSibling = nodeP->nextSibling;
    nodeP->nextSibling = nodeP;
    nodeP->preSibling = nodeP;


    /* Increment the counter if appropriate and apply the cascade cut rule */
    /*---------------------------------------------------------------------*/
    if ((saveParent != NULL) && (saveParent->counter >= 0)) {
	if (++(saveParent->counter) == 2) {
	    FH_disconnect(headerP,saveParent);
	    FH_concatenate(*headerP,saveParent);
	}
    }

    return;


}




/*
 *  FH_concatenate()
 *
 *  DESCRIPTION
 *  -----------
 *	Concatenate the two specified doubly linked list.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	nodeP1	<input>	: pointer to one node in doubly linked list 1
 *	nodeP2	<input>	: pointer to one node in doubly linked list 2
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */


FH_concatenate(nodeP1,nodeP2)


FH_node		*nodeP1;
FH_node		*nodeP2;


{


    FH_node		*lastNode1;
    FH_node		*lastNode2;



    if ((nodeP1 == NULL) || (nodeP2 == NULL)) {
	return;
    }

    lastNode1 = nodeP1->preSibling;
    lastNode2 = nodeP2->preSibling;

    lastNode1->nextSibling = nodeP2;
    lastNode2->nextSibling = nodeP1;
    nodeP1->preSibling = lastNode2;
    nodeP2->preSibling = lastNode1;


}



/*
 *  FH_binaryadd()
 *
 *  DESCRIPTION
 *  -----------
 *	Binary link the trees until rank duplication disappears.
 *	This is similar to binary addition.
 *	At the same time, search the node with minimum key in the root list 
 *	and update the header.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	rank2tree[i]	: pointer to a tree with rank i
 *
 *
 */


FH_binaryadd(headerP)

FH_header	*headerP;

{

    FH_node		*nextNode;
    FH_node		*save_header;
    short		flag;
    int		i;
    FH_node		*save_sequence[MAXHEAPSIZE];;



    if (*headerP == NULL) {
	return;
    }

    /* Initialization */
    /*----------------*/
    for (i = 0; i <= MAXRANK + 1; ++i) {
	rank2tree[i] = NULL;
    }

    /* Update the min node and save the original sequence of nodes */
    /*-------------------------------------------------------------*/
    save_header = *headerP;
    for (i = 0, nextNode = save_header, flag = OFF; 
	 (nextNode != save_header) || (flag != ON);
	 nextNode = nextNode->nextSibling, flag = ON) {

	if ((*headerP)->key > nextNode->key) {
	    *headerP = nextNode;
	}
	save_sequence[i++] = nextNode;
    }
    save_sequence[i] = NULL;

    /* Binary add each tree */
    /*----------------------*/
    for (i = 0; ; ++i) {

	if (save_sequence[i] == NULL) {
	    break;
	} else {
	    nextNode = save_sequence[i];
	}

#ifdef FH_BINADD_DEBUG
	printf("Processing node %d\n", nextNode->id);
#endif

	if (rank2tree[nextNode->rank] == NULL) {
	    rank2tree[nextNode->rank] = nextNode;

#ifdef FH_BINADD_DEBUG
	    printf("rank2tree[%d] = %d\n", nextNode->rank, nextNode->id);
#endif

	} else {
	    FH_binaryadd1(headerP, rank2tree[nextNode->rank], nextNode);
	}
    }


}



/*
 *  FH_decreasekey()
 *
 *  DESCRIPTION
 *  -----------
 *	Decrease the key of the specified node by the specified amount.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *	nodeP	<input>	: pointer to the node of the heap
 *	delta	<input>	: the amount of decrease of the key
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */


FH_decreasekey(headerP,nodeP,delta)

FH_header	*headerP;
FH_node		*nodeP;
float		delta;


{


#ifdef FH_DEBUG
    printf("Decreasing key of node %d from %.1f to %.1f\n", nodeP->id,
	   nodeP->key, nodeP->key - delta);
#endif

    nodeP->key -= delta;
    if (nodeP->parent != NULL) {
	FH_disconnect(headerP,nodeP);
	FH_concatenate(*headerP,nodeP);
    }
    if (*headerP == NULL) {
	*headerP = nodeP;
    } else {
	if ((*headerP)->key > nodeP->key) {
	    *headerP = nodeP;
	}
    }

#ifdef FH_DEBUG
    FH_print(headerP, NULL);
#endif


}



/*
 *  FH_delete()
 *
 *  DESCRIPTION
 *  -----------
 *	Delete the specified node from the heap.
 *	If it is the min node, call FH_deletemin.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *	nodeP	<input>	: pointer to the node of the heap
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_delete(headerP,nodeP)

FH_header	*headerP;
FH_node		*nodeP;


{


    if (nodeP == *headerP) {
	FH_deletemin(headerP);
	return;
    } else {
	FH_destroy(headerP,nodeP);
	return;
    }


}



/*
 *  FH_destroy()
 *
 *  DESCRIPTION
 *  -----------
 *	Delete the specified node from the heap.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *	nodeP	<input>	: pointer to the node of the heap
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_destroy(headerP,nodeP)

FH_header	*headerP;
FH_node		*nodeP;


{


    FH_node		*nextNode;
    short		flag;


    if (nodeP->child != NULL) {
	for (nextNode = nodeP->child, flag = OFF; 
	     (nextNode != nodeP->child) || (flag != ON);
	     nextNode = nextNode->nextSibling, flag = ON) {
	    nextNode->parent = NULL;
	    nextNode->counter = -1;
	}
    }
    FH_disconnect(headerP,nodeP);
    FH_concatenate(*headerP,nodeP->child);
    id2heapNode[nodeP->id - 1] = NULL;
    FREE(nodeP);


}




/*
 *  FH_binaryadd1()
 *
 *  DESCRIPTION
 *  -----------
 *	Recursively link the given tree with queue until no rank duplication.
 *	This corresponds to one-bit binary addition.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *	rootP1	<input>	: pointer to the root of the tree1
 *	rootP2	<input>	: pointer to the root of the tree2
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	rank2tree[i]	: pointer to a tree with rank i
 *
 *
 */


FH_binaryadd1(headerP,rootP1,rootP2)

FH_header	*headerP;
FH_node		*rootP1;
FH_node		*rootP2;


{


    FH_node		*linkedTree;


#ifdef FH_BINADD_DEBUG
    printf("Binary add node %d and node %d\n", rootP1->id, rootP2->id);
    FH_print(headerP, NULL);
#endif

    rank2tree[rootP1->rank] = NULL;

#ifdef FH_BINADD_DEBUG
    printf("rank2tree[%d] = 0\n", rootP1->rank);
#endif

    linkedTree = FH_link(headerP,rootP1,rootP2);
    if (rank2tree[linkedTree->rank] == NULL) {
	rank2tree[linkedTree->rank] = linkedTree;

#ifdef FH_BINADD_DEBUG
	printf("rank2tree[%d] = %d\n", linkedTree->rank, linkedTree->id);
#endif

    } else {
	FH_binaryadd1(headerP, rank2tree[linkedTree->rank], linkedTree);
    }


}



/*
 *  FH_link()
 *
 *  DESCRIPTION
 *  -----------
 *	Link the two specified tree into a new one.
 *
 *  RETURNS
 *  -------
 *	linkedTree	: pointer to the root of the resulting tree
 *
 *  Arguments:
 *  ---------
 *	headerP	<input>	: pointer to the header of the heap
 *	rootP1	<input>	: pointer to the root of the tree1
 *	rootP2	<input>	: pointer to the root of the tree2
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */


FH_node		*FH_link(headerP,rootP1,rootP2)

FH_header	*headerP;
FH_node		*rootP1;
FH_node		*rootP2;


{


    FH_node		*parentNode;
    FH_node		*childNode;


    if (rootP1->key <= rootP2->key) {
	parentNode = rootP1;
	childNode = rootP2;
    } else {
	parentNode = rootP2;
	childNode = rootP1;
    }
    FH_disconnect(headerP,childNode);
    FH_concatenate(parentNode->child,childNode);
    ++(parentNode->rank);
    childNode->counter = 0;
    childNode->parent = parentNode;
    parentNode->child = childNode;

#ifdef FH_BINADD_DEBUG
    FH_print(headerP, NULL);
#endif

    return(parentNode);


}



/*
 *  FH_free()
 *
 *  DESCRIPTION
 *  -----------
 *	Free the memory for each heap node still allocated.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_free()


{


    int		i;


    for (i = 0; i <= MAXHEAPSIZE - 1; ++i) {
	FREE(id2heapNode[i]);
    }


}



/*
 *  FH_print()
 *
 *  DESCRIPTION
 *  -----------
 *	Print the children of the specified Fibonacci heap node.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	headerP<input>	: pointer to the header of heap
 *	parent<input>	: pointer to the parent
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	id2heapNode[i]	: pointer to the node with id i
 *
 *
 */


FH_print(headerP, parent)

FH_header	*headerP;
FH_node		*parent;

{


    int		i;
    int		flag;
    FH_node		*headNode;
    FH_node		*nextNode;


    if (parent == NULL) {
	headNode = *headerP;
	printf("Roots of heap : ");
    } else {
	headNode = parent->child;
	printf("Children of node %d(%.1f) : ", parent->id, parent->key);
    }
    if (headNode == NULL) {
	printf("\n");
	return;
    }

    for (nextNode = headNode, flag = OFF; (nextNode != headNode) || (flag != ON);
	 nextNode = nextNode->nextSibling, flag = ON) {

	printf("%d(%.1f) ", nextNode->id, nextNode->key);
    }
    printf("\n");

    for (nextNode = headNode, flag = OFF; (nextNode != headNode) || (flag != ON);
	 nextNode = nextNode->nextSibling, flag = ON) {
	FH_print(headerP, nextNode);
    }



}


