 
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/intRtn.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  intRtn.c
 *  ********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains supporting routines for formulating and solving
 *	a global router problem
 *
 *  functions contained in this file:
 *	askCritical()
 *	OverCellError()
 *	makeNet()
 *	deleteNet()
 *	copyRoute()
 *	getLindoResult()
 *	assignRouteToNet()
 *	printStat()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"





/*
 *  askCritical()
 *
 *  DESCRIPTION
 *  -----------
 *	Determine if the given net is critical or not
 *
 *  RETURNS
 *  -------
 *	YES		if it is critical
 *	NO		if it is not critical
 *
 *  Arguments:
 *  ---------
 *	netP	<input>	: net for which to judge criticality
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


askCritical(netP)


GNetDef *netP;

{

if (netP->criticality > parm.criticality_threshold) {
    return(YES);
} else {
    return(NO);
}


}





/*
 *  overCellError()
 *
 *  DESCRIPTION
 *  -----------
 *	Error exit for capacity constraint violation on over the cell routing
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


overCellError(regionP)

GRegionDef	*regionP;


{

printf("Over the cell region capacity error: region = %d\n", regionP->id);

}






/*
 *  makeNet()
 *
 *  DESCRIPTION
 *  -----------
 *	Make a new graph net struct for multiLawler algorithm
 *
 *  RETURNS
 *  -------
 *	netP		pointer to a new net struct
 *
 *  Arguments:
 *  ---------
 *	netGrP	<input>	: pointer to a net whose net struct is created
 *
 *  Local variables:
 *  ---------------
 *	node1		: pointer to one node adjacent to G_DOM_XSEC region
 *	node2		: pointer to other node adjacent to G_DOM_XSEC region
 *	distance1	: distance between node1 and G_DOM_XSEC region
 *	distance2	: distance between node2 and G_DOM_XSEC region
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Add pins to graph
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


netType *makeNet(netGrP)

GNetGrDef	*netGrP;

{



netType		*netP;
GNodeDef	*nextNode, *nextEqNode;
netNodeType	*nextNetNode;
nodeGroupType	*nextNetEqNode;


netP = MALLOC(netType,1);
netP->next = NULL;
netP->node = NULL;

for (nextNode = netGrP->node; nextNode != NULL; nextNode = nextNode->next) {

    if (netP->node == NULL) {
	nextNetNode = netP->node = MALLOC(netNodeType,1);
    } else {
	nextNetNode->next = MALLOC(netNodeType,1);
	nextNetNode = nextNetNode->next;
    }
    nextNetNode->id = nextNode->region->id;
    nextNetNode->next = NULL;
    nextNetNode->equiv = NULL;

    for (nextEqNode = nextNode->nextEq; nextEqNode != NULL;
		nextEqNode = nextEqNode->nextEq) {

	if (nextNetNode->equiv == NULL) {
	    nextNetEqNode = nextNetNode->equiv
			= MALLOC(nodeGroupType,1);
	} else {
	    nextNetEqNode->next
			= MALLOC(nodeGroupType,1);
	    nextNetEqNode = nextNetEqNode->next;
	}
	nextNetEqNode->id = nextEqNode->region->id;
	nextNetEqNode->next = NULL;
    }
}

return(netP);



}




/*
 *  deleteNet()
 *
 *  DESCRIPTION
 *  -----------
 *	Delete a net struct used for multiLawler algorithm
 *
 *  RETURNS
 *  -------
 *
 *  Arguments:
 *  ---------
 *	netP	<input>	: pointer to a graph net to be deleted
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


deleteNet(netP)

netType	*netP;

{


netNodeType	*nextNode, *saveNode;
nodeGroupType	*nextEqNode, *saveEqNode;


for (nextNode = netP->node; nextNode != NULL; nextNode = saveNode) {

    for (nextEqNode = nextNode->equiv; nextEqNode != NULL;
		nextEqNode = saveEqNode) {

	saveEqNode = nextEqNode->next;
	FREE(nextEqNode);
    }
    saveNode = nextNode->next;
    FREE(nextNode);
}
FREE(netP);


}





/*
 *  copyRoute()
 *
 *  DESCRIPTION
 *  -----------
 *	Copy M shortest routes obtained to the specified net group
 *		This erases all the existent candRoute[]
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	netGrP	<input>	: pointer to net group to whom routes should be copied
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	This erases all the existent candRoute[]
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	routePointer[]	: pointers to M shortest routes obtained by multiLawler
 *
 */


copyRoute(netGrP)

GNetGrDef	*netGrP;

{


int		i;


for (i =0; i <= G_MAXROUTE - 1; ++i) {
    netGrP->candRoute[i] = NULL;
}
for (i =0; i <= G_MAXROUTE - 1; ++i) {
    if (routePointer[i] == NULL) {
	break;
    }
    netGrP->candRoute[i] = routePointer[i];
}


}






/*
 *  getLindoResult()
 *
 *  DESCRIPTION
 *  -----------
 *	Get lindo results from file and write them to route struct
 *
 *  RETURNS
 *  -------
 *
 *  Arguments:
 *  ---------
 *	M	: number of candidate route
 *
 *  Local variables:
 *  ---------------
 *	endFlg		: flag, if ON, EOF encountered
 *	infeasibleFlg	: flag, if ON, no feasible solution obtained
 *	word		: storage of the string read from lindo output file
 *	LPresult1[i][j]	: array of # nets taking route j for net group i
 *	LPresult2[i][j]	: array of x,y coodinates for region j+1 (i: 0=x, 1=y)
 *	netGr		: index representing net group
 *	route		: index representing route No
 *	regID		: index representing region id
 *	direction	: representing x(=0) or y(=1)
 *	result		: result value of LINDO package
 *	mh1,mh2		: mh1 - mh2 = mh = improvement margin in height
 *	mw1,mw2		: mw1 - mw2 = mw = improvement margin in width
 *	netSum		: total # of nets given by the rounded result of Lindo
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


getLindoResult(M)

int		M;


{

int		i,j;
int		endFlg, infeasibleFlg;
char		word[50];
int		LPresult1[G_MAXNETGR][G_MAXROUTE];
int		LPresult2[G_NHV][G_MAXREGION];
int		regID;
int		direction;
int		netGr, route;
float		result;
float		mh1,mh2,mw1,mw2;
GNetGrDef	*nextNetGr;
int		netSum;


/* Initialization */
/*----------------*/
for (i = 0; i <= G_MAXNETGR - 1; ++i) {
    for (j = 0; j <= G_MAXROUTE - 1; ++j) {
	LPresult1[i][j] = 0;
    }
}
for (i = 0; i <= G_NHV - 1; ++i) {
    for (j = 0; j <= G_MAXREGION - 1; ++j) {
	LPresult2[i][j] = 0;
    }
}

lindoFile = fopen("lindo.out","r");

/* Scan each set of solution in the report
 * "REDUCED COST" is the keyword appearing just before the solution set
 * "NO FEASIBLE" is the keyword indicating no feasible solution obtained
 *----------------------------------------------------------------------*/
for (endFlg = infeasibleFlg = OFF;;) {
    for (;;) {
	if (fscanf(lindoFile,"%s",word) == EOF) {
	    endFlg = ON;
	    break;
	}
	if (strcmp(word,"NO") == 0) {
	    fscanf(lindoFile,"%s",word);
	    if (strcmp(word,"FEASIBLE") == 0) {
		endFlg = ON;
		infeasibleFlg = ON;
		break;
	    }
	}
	if (strcmp(word,"REDUCED") == 0) {
	    fscanf(lindoFile,"%s",word);
	    break;
	}
    }
    if (endFlg == ON) {
	break;
    }

    /* Now seek the result
     * Note that variable names are changed to upper case letters in the result 
     *	ex) x5v2 (lindo.in) --> X5V2 (lindo.out)
     *-----------------------------------------------------------------------*/
    for (; fscanf(lindoFile,"%s",word) != EOF; ) {

	/* If "ROW" appears, solution set section ends */
	/*---------------------------------------------*/
	if (strcmp(word,"ROW") == 0) {
	    break;
	}
	if (word[0] == 'X') {
	    sscanf(word,"X%dV%d",&netGr,&route);
	    fscanf(lindoFile,"%f",&result);
	    LPresult1[netGr - 1][route - 1] = RINT(result);
	} else if (word[0] == 'P') {
	    if (word[1] == 'X') {
		sscanf(word,"PX%d",&regID);
		direction = G_HOR;
	    } else {
		sscanf(word,"PY%d",&regID);
		direction = G_VER;
	    }
	    fscanf(lindoFile,"%f",&result);
	    LPresult2[direction][regID - 1] = RINT(result);
	} else if (word[0] == 'M') {
	    fscanf(lindoFile,"%f",&result);
	    if (strcmp(word,"MW1") == 0) {
		mw1 = result;
	    } else if (strcmp(word,"MW2") == 0) {
		mw2 = result;
	    } else if (strcmp(word,"MH1") == 0) {
		mh1 = result;
	    } else if (strcmp(word,"MH2") == 0) {
		mh2 = result;
	    }
	} else {
	    continue;
	}
    }
}
fclose(lindoFile);

/* If no feasible solution, exit */
/*-------------------------------*/
if (infeasibleFlg == ON) {
    printf("NO FEASIBLE SOLUTION OBTAINED\n");
    exit(8);
}

statis.LP_improv_dim[G_HOR] = RINT((mw1 - mw2));
statis.LP_improv_dim[G_VER] = RINT((mh1 - mh2));
printf("Width improvement = %d oct_units\n", statis.LP_improv_dim[G_HOR]);
printf("Height improvement = %d oct_units\n", statis.LP_improv_dim[G_VER]);
printf("\n");

/* Copy the result into each net group and each region */
/*-----------------------------------------------------*/
for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

    if ((askTrivialRoute(nextNetGr) == YES)
		|| (nextNetGr->NoCritNet == nextNetGr->NoNet)) {
	continue;
    }
    for (netSum = 0, i = M - 1; i >= 0; --i) {
	nextNetGr->IPresult[i] = LPresult1[nextNetGr->id-1][i];
	netSum += LPresult1[nextNetGr->id-1][i];
    }

    /* Adjust the difference in total # nets between netGr and Lindo result
     *	if (Lindo netSum > real net sum)
     *	    scan candidates from head and increment IPresult by 1 if it != 0
     *	    until both are equal.
     *	if (Lindo netSum < real net sum)
     *	    scan candidates from tail and decrement IPresult by 1 if it != 0
     *	    until both are equal.
     *----------------------------------------------------------------------*/
    if (netSum < nextNetGr->NoNet - nextNetGr->NoCritNet) {
	for (i = 0; ;++i) {
	    if (i > M - 1) {
		i = 0;
	    }
	    if (nextNetGr->IPresult[i] != 0) {
		++(nextNetGr->IPresult[i]);
		++netSum;
	    }
	    if (netSum == nextNetGr->NoNet - nextNetGr->NoCritNet) {
		break;
	    }
	}
    } else if (netSum > nextNetGr->NoNet - nextNetGr->NoCritNet) {
	for (i = M - 1; ;--i) {
	    if (i < 0) {
		i = M - 1;
	    }
	    if (nextNetGr->IPresult[i] != 0) {
		--(nextNetGr->IPresult[i]);
		--netSum;
	    }
	    if (netSum == nextNetGr->NoNet - nextNetGr->NoCritNet) {
		break;
	    }
	}
    }
}
if (parm.zero_compact == ON || parm.placeRefMode != 0) {
    for (i = 0; i <= G_MAXREGION - 1; ++i) {
	if (GID2Reg[i] != NULL) {
	    if (parm.zero_compact == ON) {
		GID2Reg[i]->newLocX = -945;
		GID2Reg[i]->newLocY = -17;
	    } else {
		GID2Reg[i]->newLocX = LPresult2[G_HOR][i];
		GID2Reg[i]->newLocY = LPresult2[G_VER][i];
	    }
	}
    }
}

/* Collect statistics */
/*--------------------*/
for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

    if (askTrivialRoute(nextNetGr) == YES) {
	statis.num_net_cand_i[0] += nextNetGr->NoNet;
	continue;
    }

    statis.num_net_cand_i[0] += nextNetGr->NoCritNet;

    for (i = 0; i < M; ++i) {
	statis.num_net_cand_i[i] += nextNetGr->IPresult[i];
    }
}


}





/*
 *  assignRouteToNet()
 *
 *  DESCRIPTION
 *  -----------
 *	Assign to each net one of candidate routes of the net group.
 *	Also it creates a global routing result file.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	assignedNet[i]	: number of nets assigned so far to candidate route i
 *	route		: index to candidate route selected
 *	saveRoute	: index to candidate route selected for preceding net
 *	NoPin		: number of pin in this net
 *	pinName[i]	: names of pins in the net
 *	NoChannel	: number of channels this net passes
 *	channel[i]	: channels which a net passes
 *	saveNode	: id of the preceding node in the path
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


assignRouteToNet()


{

struct channelEdge {
    int	node1;
    int	node2;
    int	startPoint;
    int	endPoint;
};

int		i;
GNetGrDef	*nextNetGr;
GNetDef		*nextNet;
int		assignedNet[G_MAXROUTE];
int		route, saveRoute;
int		NoPin;
char		*pinName[G_MAXPIN_PER_NET];
int		NoChannel;
struct channelEdge	channel[G_MAXREGION];
pathType	*nextPath;
GPinDef		*nextPin;
nodeGroupType	*nextNode;
int		saveNode;


resultFile = fopen("mercury.result","w");

/* Process each net group */
/*------------------------*/
for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

    if (askTrivialRoute(nextNetGr) == YES) {
	continue;
    }

    /* Initialization */
    /*----------------*/
    for (i = 0; i <= G_MAXROUTE - 1; ++i) {
	assignedNet[i] = 0;
    }

    /* Process each net in this net group */
    /*------------------------------------*/
    for (saveRoute = -1, nextNet = nextNetGr->net; nextNet != NULL;
		saveRoute = route, nextNet = nextNet->nextGrMem) {

	/* If it is a critical net, assign the shortest route */

	if (nextNet->critical == YES) {
	    nextNet->routeNo = 0;
	    continue;
	} else {

	/* Otherwise, for the present, assign candidate sequentially */

	    for (route = 0; route <= G_MAXROUTE - 1; ++route) {
		if (assignedNet[route] < nextNetGr->IPresult[route]) {
		    nextNet->routeNo = route;
		    ++(assignedNet[route]);
		    break;
		}
	    }
	}

	/* Get the number and names of pin for this net */

	for (NoPin = 0, nextPin = nextNet->pin; nextPin != NULL;
		++NoPin, nextPin = nextPin->next) {
	    if (NoPin>=G_MAXPIN_PER_NET) {
		fprintf(stderr,
		    "ERROR Too many pins in net >=G_MAXPIN_PER_NET=%d\n",
		    G_MAXPIN_PER_NET);
		abort();
	    }
	    pinName[NoPin] = nextPin->name;
	}

	/* Get the number and incident nodes of channels for this net */

	if (route != saveRoute) {

	    for (i = 0; i <= G_MAXREGION - 1; ++i) {

		channel[i].node1 = channel[i].node2 = 0;
		channel[i].startPoint = channel[i].endPoint = 0;
	    }
	    for (NoChannel = 0, nextPath = nextNetGr->candRoute[route]->path;
			nextPath != NULL; nextPath = nextPath->next) {

		for (--NoChannel, saveNode = 0, nextNode = nextPath->node;
			nextNode != NULL; ++NoChannel, saveNode = nextNode->id,
			nextNode = nextNode->next) {

		    if (saveNode == 0) {
			continue;
		    }
		    channel[NoChannel].node1 = saveNode;
		    channel[NoChannel].node2 = nextNode->id;
		    channel[NoChannel].startPoint = 0;
		    channel[NoChannel].endPoint = 0;
		}
	    }
	}

	/* Output a file which describes net info */
	/*----------------------------------------*/
	fprintf(resultFile,"\n\n");
	fprintf(resultFile,"net %s\tlength %d\n", nextNet->name,
		nextNetGr->candRoute[route]->length);
	fprintf(resultFile,"pins %d\n", NoPin);
	for (i = 0; i <= NoPin - 1; ++i) {
	    fprintf(resultFile,"\t%s\n", pinName[i]);
	}
	fprintf(resultFile,"channels %d\n", NoChannel);
	for (i = 0; i <= NoChannel - 1; ++i) {
	    fprintf(resultFile,"\t%d %d %d %d\n",
			channel[i].node1, channel[i].node2,
			channel[i].startPoint, channel[i].endPoint);
	}
    }
}

fclose(resultFile);


}






/*
 *  printStat()
 *
 *  DESCRIPTION
 *  -----------
 *	Print the statistics on the global routing results
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


printStat()


{

int	i;


/* Compute some data */
/*-------------------*/
statis.ave_net_per_netgroup = 
	(float)statis.num_net / (float)statis.num_netgroup;
statis.init_area = statis.critical_dim[G_HOR] * statis.critical_dim[G_VER];
statis.final_dim[G_HOR] = 
	RINT(((float)(statis.LP_init_dim[G_HOR] - statis.LP_improv_dim[G_HOR])
		/ (float)OCT_PER_MICRON));
statis.final_dim[G_VER] = 
	RINT(((float)(statis.LP_init_dim[G_VER] - statis.LP_improv_dim[G_VER])
		/ (float)OCT_PER_MICRON));
statis.final_area = statis.final_dim[G_HOR] * statis.final_dim[G_VER];
statis.dim_ratio[G_HOR] = 
	(float)statis.final_dim[G_HOR] / (float)statis.critical_dim[G_HOR];
statis.dim_ratio[G_VER] = 
	(float)statis.final_dim[G_VER] / (float)statis.critical_dim[G_VER];
statis.area_ratio = (float)statis.final_area / (float)statis.init_area;

for (i = 0; i < parm.M; ++i) {
    statis.percent_net_cand_i[i] = 
	(float)statis.num_net_cand_i[i] / (float)statis.num_net * 100.0;
}

/* Print the statistics */
/*----------------------*/
printf("*** Mercury statistics ***\n\n");
printf("    -- Chip information\n");
printf("\t# of blocks\t\t= %d\n", statis.num_block);
printf("\t# of nets\t\t= %d\n", statis.num_net);
printf("\t# of critical nets\t= %d\n", statis.num_crit_net);
printf("\t# of trivial nets\t= %d\n", statis.num_trivial_net);
printf("\n");

printf("    -- Net group information\n");
printf("\t# of net groups\t\t\t= %d\n", statis.num_netgroup);
printf("\t# of non-trivial net groups\t= %d\n", statis.num_nontriv_netgroup);
printf("\tAverage # of nets / net group\t= %3.2f\n", 
	statis.ave_net_per_netgroup);
printf("\n");

printf("    -- Tile information\n");
printf("\t# of hor. bottleneck tiles\t= %d\n", statis.num_hor_bot);
printf("\t# of ver. bottleneck tiles\t= %d\n", statis.num_ver_bot);
printf("\t# of overcell routing regions\t= %d\n", statis.num_overcell);
printf("\n");

printf("    -- Size of global routing graph\n");
printf("\t# of nodes\t\t= %d\n", statis.num_GRgraph_node);
printf("\t# of edges\t\t= %d\n", statis.num_GRgraph_edge);
printf("\n");

printf("    -- Polar graph information\n");
printf("\t# of hor. polar paths\t= %d\n", statis.num_polarpath[G_HOR]);
printf("\t# of ver. polar paths\t= %d\n", statis.num_polarpath[G_VER]);
printf("\n");

printf("    -- Size of linear programming\n");
printf("\t# of variables\t\t= %d\n", statis.num_variable);
printf("\t# of constraints\t= %d\n", statis.num_constraint);
printf("\n");

printf("    -- Constitution of linear programming solution\n");
for (i = 0; i < parm.M; ++i) {
    printf("\t# of nets routed by %dth shortest route\t= %d (%3.1f%%)\n", 
		i+1, statis.num_net_cand_i[i], statis.percent_net_cand_i[i]);
}
printf("\n");

printf("    -- Reduction in chip core size (w/o pads)\n");
printf("\testimated shortest total net length (u)\t= %d\n", 
	statis.init_netlength);
printf("\twidth  (u)\t: %d --> %d (%3.1f%%)\n", 
	statis.critical_dim[G_HOR], statis.final_dim[G_HOR],
	statis.dim_ratio[G_HOR] * 100.0);
printf("\theight (u)\t: %d --> %d (%3.1f%%)\n", 
	statis.critical_dim[G_VER], statis.final_dim[G_VER],
	statis.dim_ratio[G_VER] * 100.0);
printf("\tarea (10^6 u^2)\t: %3.2f --> %3.2f (%3.1f%%)\n", 
	(float)statis.init_area / 1000000.0, 
	(float)statis.final_area / 1000000.0, 
	statis.area_ratio * 100.0);
printf("\n*** End of statistics ***\n\n");



}


