
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/lawler.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  lawler.c
 *  ********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for obtaining M shortest paths between a point
 *	and a target, which is a set of points.
 *
 *  functions contained in this file:
 *	lawler()
 *	modifyGraph1()
 *	modifyGraph2()
 *	insertLawler()
 *	buildLawler()
 *	mergeLawler()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */


#include <stdio.h>
#include <strings.h>
#include "initGlb.h"


/*
 *  lawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Obtain M shortest paths from a node, v, to a set of nodes, target.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: starting node
 *	target	<input>	: a set of nodes to be reached
 *	M	<input>	: number of shortest paths to be obtained
 *
 *  Local variables:
 *  ---------------
 *	n		: one of target nodes reached by the algorithm
 *	p		: constrained node of a path
 *	q		: second last node on the path
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Uses dijkstraPointer[] and fills in values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	pathPointer[]	: pointers to paths obtained by this algorithm
 *	adjList[]	: pointers to adjacency list for graph
 *
 */

lawler(v,target,M)

int		v,M;
netNodeType	*target;

{

int		i,m,p,q,n;
adjNodeType	*nextAdjNode;
lawlerNodeType	*nextNode;
lawlerPathType	*buildLawler(),*pathP;



/* Initialization */
/*----------------*/
for (i = 0; i <= G_MAXPATH - 1; ++i) {
    pathPointer[i] = NULL;
}
for (i = 0; i <= G_MAXNODE - 1; ++i) {
    nodeBlock[i] = NO;
    for (nextAdjNode = adjList[i]; nextAdjNode != NULL; 
		nextAdjNode = nextAdjNode->next) {
	nextAdjNode->blocked = NO;
    }
}


/* Obtain a shortest path from v to target */
n = shortestPath(v,target,adjList);

/* Record all nodes in the path */
/*------------------------------*/
pathPointer[0] = MALLOC(lawlerPathType,1);
pathPointer[0]->constrainedNode = v;
pathPointer[0]->constrainedLength = 0;
pathPointer[0]->node = MALLOC(lawlerNodeType,1);
pathPointer[0]->node->id = v;
pathPointer[0]->node->next = NULL;
pathPointer[0]->node->closedNode = NULL;
mergeLawler(pathPointer[0],n);


/* Obtain the second to Mth shortest paths */
/*-----------------------------------------*/
for (m = 2; m <= M; ++m) {

/* If the previous path is empty, no more paths can be generated */

    if (pathPointer[m-2] == NULL) {
	break;
    }

/* Get the constrained node, p , of the previous path */
/* Get the second last node, q , of the previous merged path */
    p = pathPointer[m-2]->constrainedNode;
    nextNode = pathPointer[m-2]->node;
    while (nextNode->next->next != NULL) {
	nextNode = nextNode->next;
    }
    q = nextNode->id;

/* If the previous path has an edge between constained node and target,
 *	block this edge and search another path
 * --------------------------------------------------------------------
 */
    if (p == q) {
	modifyGraph1(q,n,m);
	modifyGraph2(q,m);
	pathP = buildLawler(m,q,n,ON);
	if ((n = shortestPath(q,target,adjList)) >= FOUND) {
	    mergeLawler(pathP,n);
	    insertLawler(pathPointer,m,M,pathP);
	}
/* Otherwise find a path from each of those nodes from p to q */
/*------------------------------------------------------------*/
    } else {
	nextNode = pathPointer[m-2]->node;
	while (nextNode->id != p) {
	    nextNode = nextNode->next;
	}

	for ( ; nextNode != NULL; nextNode = nextNode->next) {

	    modifyGraph1(nextNode->id,nextNode->next->id,m);
	    if (nextNode->id == p) {
		modifyGraph2(p,m);
	        pathP = buildLawler(m,
			nextNode->id,nextNode->next->id,ON);
	    } else {
	        pathP = buildLawler(m,
			nextNode->id,nextNode->next->id,OFF);
	    }
	    if ((n = shortestPath(nextNode->id,target,adjList)) >= FOUND) {
		mergeLawler(pathP,n);
		insertLawler(pathPointer,m,M,pathP);
	    }
	    if (nextNode->id == q) {
		break;
	    }
	}
    }
}

#ifdef LAWLER_DEBUG
printLawler(pathPointer,v,target->id,M);
#endif


}






/*
 *  modifyGraph1()
 *
 *
 *  DESCRIPTION
 *  -----------
 *	Delete the following from the adjacency data of a graph
 *		all edges from those nodes which precedes x1 on the m-1 th path
 *		all edges to   those nodes which precedes x1 on the m-1 th path
 *		all edges between x1 and x2
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	x1	<input>	: constrained node for a modified graph
 *	x2	<input>	: following node of x1 on the m-1 th shortest path
 *	m	<input>	: path index. processing m th shortest path now.
 *
 *  Local variables:
 *  ---------------
 *	node[i]		: if ON, node i is on the m-1 th path which precedes x1
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Modifies adjacency list
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	pathPointer[]	: pointers to paths obtained by this algorithm
 *	adjList[]	: pointers to adjacency list for graph
 *
 */


modifyGraph1(x1,x2,m)

int x1,x2,m;

{

int		i;
short		node[G_MAXNODE];
adjNodeType	*nextNode;
lawlerNodeType	*nextLawler;


/* Record all nodes in the path preceding x1 */
/*-------------------------------------------*/
for (i = 0; i <= G_MAXNODE - 1; ++i) {
    node[i] = OFF;
}
for (nextLawler = pathPointer[m-2]->node; ; nextLawler = nextLawler->next) {
    if (nextLawler->id == x1) {
	break;
    } else {
	node[nextLawler->id - 1] = ON;
    }
}


#ifdef LAWLER_DEBUG
printf("Reset all the block information\n");
#endif

for (i = 0; i <= G_MAXNODE - 1; ++i) {

/* Reset the flags to represent blocking status */
/*----------------------------------------------*/
    nodeBlock[i] = NO;
    for (nextNode = adjList[i]; nextNode != NULL; nextNode = nextNode->next) {
	nextNode->blocked = NO;
    }

/* Mark if flag is ON for this node in node[] (precedes x1 in m-1 th path) */
/*-------------------------------------------------------------------------*/
    if (node[i] == ON) {
	nodeBlock[i] = YES;

#ifdef LAWLER_DEBUG
	printf("Block a node %d\n", i+1);
#endif

	continue;
    }

    for (nextNode = adjList[i]; nextNode != NULL; nextNode = nextNode->next) {

/* Mark if it is an edge to some node in node[] or edge between x1 and x2 */
/*------------------------------------------------------------------------*/
	if (node[nextNode->id - 1] == ON || (i+1 == x1 && nextNode->id == x2)
			|| (i+1 == x2 && nextNode->id == x1)) {
	    nextNode->blocked = YES;

#ifdef LAWLER_DEBUG
	    printf("Block an edge from node %d to %d\n", i+1, nextNode->id);
#endif

	}
    }
}




}




/*
 *  modifyGraph2()
 *
 *  DESCRIPTION
 *  -----------
 *	Reflect closed (blocked) edge information about p to modified graph
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	p	<input>	: node whose blocked edge should be copied
 *	m	<input>	: path index. processing m th shortest path now.
 *
 *  Local variables:
 *  ---------------
 *	closedNode[]	: array of closed node from p
 *	flag		: flag to show the search has hit something
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Modifies adjList[]
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	pathPointer[]	: pointers to paths obtained by this algorithm
 *	adjList[]	: pointers to adjacency list
 *
 */


modifyGraph2(p,m)


int p,m;

{

int		i,flag,closedNode[G_NDIR];
lawlerNodeType	*nextLawlerNode;
nodeGroupType	*nextClosedNode;
adjNodeType	*nextAdjNode;


/* Record in closedNode[] all closed node information incident to node p */
/*-----------------------------------------------------------------------*/
nextLawlerNode = pathPointer[m-2]->node;
while (nextLawlerNode->id != p) {
    nextLawlerNode = nextLawlerNode->next;
}
i = 0;
nextClosedNode = nextLawlerNode->closedNode;
while (nextClosedNode != NULL) {
    closedNode[i++] = nextClosedNode->id;
    nextClosedNode = nextClosedNode->next;
}
closedNode[i] = 0;


/* Reflect closed node information to modified graph */
/*---------------------------------------------------*/
for (nextAdjNode = adjList[p-1]; nextAdjNode != NULL;
	nextAdjNode = nextAdjNode->next) {

/* Check if the current node is included in closedNode[] */
    flag = NOT_FOUND;
    for (i = 0; i < G_MAXNODE - 1; ++i) {
	if (closedNode[i] == 0) {
	    break;
	}
	if (nextAdjNode->id == closedNode[i]) {
	    flag = FOUND;
	    break;
	}
    }

/* If included in closedNode[], delete the corresponding adjacency info */
/*----------------------------------------------------------------------*/
    if (flag == FOUND) {
	nextAdjNode->blocked = YES;

#ifdef LAWLER_DEBUG
	printf("Block an edge from node %d to %d\n", p, nextAdjNode->id);
#endif

    }
}



}




/*
 *  insertLawler()
 *
 *
 *  DESCRIPTION
 *  -----------
 *	Insert the given path in the path list according to its length.
 *	The path list is sorted by the path length in non-decreasing order.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathPtr[]	<in/out> : pointers to list of paths obtained so far
 *	m	<input>	: path index. processing m th shortest path now.
 *	M	<input>	: number of shortest paths to be obtained
 *	pathP	<input>	: path to be inserted in the list
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Insert a new member in the path list.
 *	Free space for overflowing struct if list exceeds the limit.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


insertLawler(pathPtr,m,M,pathP)


lawlerPathType *pathPtr[];
lawlerPathType *pathP;
int		m,M;

{

int i,j;


for (i = m - 1; i <= G_MAXPATH - 1; ++i) {

/* If come to the tail, add this path if # of elements does not exceed M */
    if (pathPtr[i] == NULL) {
	if (i <= M - 1) {
	    pathPtr[i] = pathP;
	} else {
	    freeLawlerPath(pathP);
	}
	break;
    }

/* If find a path whose length is greater, insert before this element */
    if (pathP->length < pathPtr[i]->length) {
	for (j = G_MAXPATH - 2; j >= i; --j) {
	    if ((j == M - 1) && (pathPtr[j] != NULL)) {
		freeLawlerPath(pathPtr[j]);
		pathPtr[j] = NULL;
	    }
	    if (j + 1 <= M - 1) {
		pathPtr[j+1] = pathPtr[j];
	    }
	}
	pathPtr[i] = pathP;
	break;
    }
}


}





/*
 *  buildLawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Build a path struct for a new path to be obtained
 *
 *  RETURNS
 *  -------
 *	Pointer to a new path struct
 *
 *  Arguments:
 *  ---------
 *	m	<input>	: path index. processing m th shortest path now.
 *	x1	<input>	: constrained node for the new path
 *	x2	<input>	: new blocked (closed) node from x1
 *	close	<input>	: flag, if ON, copy the closed node info about x1
 *
 *  Local variables:
 *  ---------------
 *	pathP		: pointer to the new path struct
 *	nextNode	: pointers to adjacency list of graph
 *	...S		: pointer to source of copying of the data
 *	...D		: pointer to destinatin of copying of the data
 *	distance	: length of edge
 *	flag		: flag, if ON, already passed constrained node
 *				of previous path
 *	parent	: preceding node of the current node in the source nodes scan
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	pathPointer[]	: pointers to paths obtained by this algorithm
 *	adjList[]	: pointers to adjacency list for graph
 *
 */


lawlerPathType *buildLawler(m,x1,x2,close)


int m,x1,x2,close;

{

lawlerPathType	*pathP;
adjNodeType	*nextNode;
lawlerNodeType	*curNodeS,*curNodeD;
nodeGroupType	*curClosedNodeS,*curClosedNodeD;
int		distance,flag,parent;



/* Allocate memory for the new struct and fill in some items */

pathP = MALLOC(lawlerPathType,1);
pathP->constrainedNode = x1;
pathP->constrainedLength = pathPointer[m-2]->constrainedLength;
pathP->length = 0;
pathP->node = NULL;


/* Copy information of each node from source (= previous path) */
/*-------------------------------------------------------------*/
flag = OFF;
for (curNodeS = pathPointer[m-2]->node; ;
	parent = curNodeS->id,curNodeS = curNodeS->next) {

    if (pathP->node == NULL) {
	curNodeD = pathP->node = MALLOC(lawlerNodeType,1);
    } else {
	curNodeD->next = MALLOC(lawlerNodeType,1); 
	curNodeD = curNodeD->next;
    }
    curNodeD->id = curNodeS->id;
    curNodeD->closedNode = NULL;
    curNodeD->next = NULL;

/* After passing the constrained node of previous path,
 *	should add edge length to the constrained length of current path
 * Length up to the constrained node of previous path has already been copied */
/*----------------------------------------------------------------------------*/
    if (flag == ON) {
	nextNode = adjList[parent-1];
	while (nextNode != NULL) {
	    if (nextNode->id == curNodeS->id) {
		distance = nextNode->weight;
		break;
	    }
	    nextNode = nextNode->next;
	}
	pathP->constrainedLength += distance;
    }

/* Check if passing the constrained node of previous path */
    if (flag == OFF && curNodeS->id == pathPointer[m-2]->constrainedNode) {
	flag = ON;
    }

/* When reach the constrained node x1, add info of closed node x2 to x1
 *	and get out of loop */
/*-------------------------------------------------------------------*/
    if (curNodeS->id == x1) {
	curNodeD->closedNode = curClosedNodeD = MALLOC(nodeGroupType,1);
	curClosedNodeD->id = x2;
	curClosedNodeD->next = NULL;

/* If close flag is ON, copy existent closed node info about x1 */
/*--------------------------------------------------------------*/
	if (close == ON) {

	    curClosedNodeS = curNodeS->closedNode;
	    while (curClosedNodeS != NULL) {
		curClosedNodeD->next = MALLOC(nodeGroupType,1);
		curClosedNodeD = curClosedNodeD->next;
		curClosedNodeD->id = curClosedNodeS->id;
		curClosedNodeD->next = NULL;

		curClosedNodeS = curClosedNodeS->next;
	    }
	}
	break;
    }
}
return(pathP);


}




/*
 *  mergeLawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Merge a constrained path and a shortest path from the constrained
 *	node to the target, into one path.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathP	<input>	: pointer to path to be merged
 *			  it already contains the constrained path portion
 *	w	<input>	: node that was reached by dijkstra's algorithm
 *
 *  Local variables:
 *  ---------------
 *	x	: node along the shortest path obtained
 *	parent	: parent node (preceding node of x along the shortest path)
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Write through pathP, which is usually a global variable
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	dijkstraPointer[]	: pointers to struct for dijkstra's algorithm
 *
 */

mergeLawler(pathP,w)


lawlerPathType *pathP;
int		w;

{


lawlerNodeType	*curNode,*nextNode;
int		x;


/* Add length of dijkstra path to length of constrained path */

pathP->length = pathP->constrainedLength + dijkstraPointer[w-1]->d;
curNode = pathP->node;
while (curNode->next != NULL) {
    curNode = curNode->next;
}

/* Add node info from target to source (reverse order) along dijkstra path */

x = w;
nextNode = NULL;
while (x != curNode->id) {
    curNode->next = MALLOC(lawlerNodeType,1);
    curNode->next->id = x;
    curNode->next->next = nextNode;
    curNode->next->closedNode = NULL;
    nextNode = curNode->next;
    x = dijkstraPointer[x-1]->parent;
}


}



