 
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/main.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  main.c
 *  ******
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains main routines
 *
 *  functions contained in this file:
 *	main()
 *	parseCommand()
 *	printParm()
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */


#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"



/*
 *  main()
 *
 *  DESCRIPTION
 *  -----------
 *	Main program of Mercury
 *
 */

main(argc,argv)

int	argc;
char	*argv[];

{
    octObject	octChip;
    octObject	octChipNewView;
    char	buf_command[G_BSIZE];
    char	buf1[G_BSIZE];
    char	buf2[G_BSIZE];


    /* For core dumping in case of failure
    errCore(1);
     */

    /* Parse the input command and set parameters */
    /*--------------------------------------------*/
    parseCommand(argc, argv);
    if (parm.outview1 == NULL) {
	sprintf(buf1,"%s.cd",parm.view);
	parm.outview1 = buf1;
    }
    if (parm.outview2 == NULL) {
	sprintf(buf2,"%s.cd_after_gr",parm.view);
        parm.outview2 = buf2;
    }
    printParm();

    /* Initilaize data */
    dataAlloc();

    /* Tile partitioning */
    /*-------------------*/
    CDSMain(parm.cell, parm.view, parm.outview1, 
		ON, &octChip, parm.chipExpand, parm.Zflag);
    printCPUtime("\nTile data read and global routing begins\n", YES);

    /* Global routing */
    /*----------------*/
    intProg();

    /* Create an interface to detailed router */
    /*----------------------------------------*/
    if (parm.placeref_only == ON) {
   	CDSHgOutMain(&octChip, OFF);
    } else {
   	CDSHgOutMain(&octChip, ON);
    }

    /* Copy the first view to the second view */
    octChipNewView = octChip;
    octChipNewView.contents.facet.view = parm.outview2;
    OH_ASSERT(octCopyFacet(&octChipNewView, &octChip));

    /* Create the interface facet */
    printf("\nCreating interface facets\n");
    sprintf(buf_command, "vulcan %s:%s\n", parm.cell, parm.outview1);
    system(buf_command);
    sprintf(buf_command, "vulcan %s:%s\n", parm.cell, parm.outview2);
    system(buf_command);

    printCPUtime("\nDetailed router interface completed\n", NO);
    printStat();
    exit(0);

}





/*
 *  parseCommand()
 *
 *  DESCRIPTION
 *  -----------
 *	Parse the input command and reads the parameters.
 *
 */

parseCommand(argc, argv)

int		argc;
char		*argv[];

{

    short	arg1;
    short	flg;
    int		i,idx;
    octObject	facet;
    octObject	facet1;
    octObject	facet2;


    /* Initialize some default parameters */
    /*------------------------------------*/
    parm.placeref_only = OFF;
    parm.zero_compact = OFF;
    parm.form_netgroup = -1;

    /* expandFactor = 1.02 is for the least amount of placement expansion */
    parm.chipExpand = 1.02;

    parm.outcell1 = NULL;
    parm.outview1 = NULL;
    parm.outcell2 = NULL;
    parm.outview2 = NULL;

    parm.Zflag = OFF;
    parm.criticality_threshold = 0;

    parm.M = 4;
    parm.m = 3;
    parm.netSizeLimit = 6;
    parm.FH_dijkstra = ON;

    parm.placeRefMode = PR_MODE1;
    parm.netlength_weight = 0.0001;
    parm.blockloc_weight = 0.0;
    parm.modelDiagonal = OFF;
    parm.tile_threshold = 320;
    parm.alpha = 1.0;
    parm.expandFactor = 1.0;
    parm.INTEGER = ON;
    parm.BIP = OFF;

    if (argc <= 1) {

parseError:

	printf("Usage: %s [options] cell:view\n", argv[0]);
	printf("  -a #1\t\t: unit usage of terminating net is #1 as to pass-thru net [1.0]\n");
	printf("  -b #1\t\t: block location factor in LP = \n");
	printf("\t\t\tinit_area/#block(init_width+init_height) x #1 [0.0]\n");
	printf("\t\t\trecommended value = 0.0002\n");
	printf("  -B\t\t: flag to use BIP declaration in Lindo [OFF]\n");
	printf("  -c #1\t\t: expand the chip by #1 to get a slicing structure [1.02]\n");
	printf("  -d\t\t: flag to model diagonal effect [OFF]\n");
	printf("  -e #1\t\t: expand estimated initial chip demension by #1 [1.0]\n");
	printf("  -f\t\t: flag NOT to use faster Dijkstra algorithm [OFF]\n");
	printf("  -g 0/1\t: form net groups(1) / not(0) [if #net >= 750, form]\n");
	printf("  -I\t\t: flag NOT to use INTEGER declaration in Lindo [OFF]\n");
	printf("  -l #1\t\t: netlength factor in LP = \n");
	printf("\t\t\tinit_area/init_netlength x #1 [0.0001]\n");
	printf("  -M #1\t\t: use #1 candidate routes in LP [4]\n");
	printf("  -m #1\t\t: select #1 paths per pin connection [3]\n");
	printf("  -n #1\t\t: regard as critical net if NET_CRITICALITY > #1 [0]\n");
	printf("  -o #1\t\t: 1st output cell:view is #1 [input_cell:input_view.cd] \n");
	printf("  -O #1\t\t: 2nd output cell:view is #1 [input_cell:input_view.cd_after_gr]\n");
	printf("  -p #1\t\t: use placement refinement mode #1 [1]\n");
	printf("\t\t\t0 : no placement refinement\n");
	printf("\t\t\t1 : preserve every adjacent bottlenecks\n");
	printf("\t\t\t2 : just prevent block overlaps\n");
	printf("\t\t\t3 : move all blocks to the origin\n");
	printf("  -r\t\t: flag to only refine placement (do NOT create subnet)\n");
	printf("\t\t\t.cd view is bogus in this case [OFF]\n");
	printf("  -t #1\t\t: set tiling threshold at #1 in Oct units [320]\n");
	printf("  -u #1\t\t: use 1 candidate route if #pin/net >= #1 [6]\n");
	printf("  -z\t\t: flag to compact bottlenecks to 0 width [OFF]\n");
	printf("  -Z\t\t: flag to use Z shaped channel in tile partition [OFF]\n");
	exit(8);
    }

    /* Read options */
    /*--------------*/
    for (arg1 = 1; arg1 <= argc - 2; ++arg1) {

	for (i = 1, flg = OFF; (argv[arg1])[i] != '\0'; ++i) {

	    switch((argv[arg1])[i]) {
	    case 'a' :
		parm.alpha = (float)atof(argv[++arg1]);
		flg = ON;
		break;
	    case 'b' :
		parm.blockloc_weight = (float)atof(argv[++arg1]);
		flg = ON;
		break;
	    case 'B' :
		parm.BIP = ON;
		flg = ON;
		break;
	    case 'c' :
		parm.chipExpand = atof(argv[++arg1]);
		flg = ON;
		break;
	    case 'd' :
		parm.modelDiagonal = ON;
		flg = ON;
		break;
	    case 'e' :
		parm.expandFactor = (float)atof(argv[++arg1]);
		flg = ON;
		break;
	    case 'f' :
		parm.FH_dijkstra = OFF;
		flg = ON;
		break;
	    case 'g' :
		parm.form_netgroup = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'I' :
		parm.INTEGER = OFF;
		flg = ON;
		break;
	    case 'l' :
		parm.netlength_weight = (float)atof(argv[++arg1]);
		flg = ON;
		break;
	    case 'M' :
		parm.M = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'm' :
		parm.m = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'n' :
		parm.criticality_threshold = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'o' :
		ohUnpackFacetName(&facet1, argv[++arg1]);
		parm.outcell1 = facet1.contents.facet.cell;
		parm.outview1 = facet1.contents.facet.view;
		flg = ON;
		break;
	    case 'O' :
		ohUnpackFacetName(&facet2, argv[++arg1]);
		parm.outcell2 = facet2.contents.facet.cell;
		parm.outview2 = facet2.contents.facet.view;
		flg = ON;
		break;
	    case 'p' :
		parm.placeRefMode = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'r' :
		parm.placeref_only = ON;
		flg = ON;
		break;
	    case 't' :
		parm.tile_threshold = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'u' :
		parm.netSizeLimit = atoi(argv[++arg1]);
		flg = ON;
		break;
	    case 'z' :
		parm.zero_compact = ON;
		flg = ON;
		break;
	    case 'Z' :
		parm.Zflag = ON;
		flg = ON;
		break;
	    default :
		printf("%s: illegal option %c\n", argv[0], (argv[arg1])[i]);
		exit(8);
	    }
	    if (flg == ON) {
		break;
	    }
	}
    }

    /* Read cell and view name */
    /*-------------------------*/
    ohUnpackFacetName(&facet, argv[arg1]);
    parm.cell = facet.contents.facet.cell;
    parm.view = facet.contents.facet.view;


}



/*
 *  printParm()
 *
 *  DESCRIPTION
 *  -----------
 *	Print the Mercury parameters.
 *
 */

printParm()

{

     printf("*** Mercury parameter information ***\n");
     printf("\n    -- Input/output cell\n");
     printf("\tinput     cell:view\t\t= %s:%s\n", parm.cell, parm.view);
     printf("\toutput(1) cell:view\t\t= %s:%s\n", parm.outcell1, parm.outview1);
     printf("\toutput(2) cell:view\t\t= %s:%s\n", parm.outcell2, parm.outview2);

     printf("\n    -- Overall framework\n");
     printf("\tform net group\t\t\t= %s\n", 
		(parm.form_netgroup < 0) ? "AUTOMATIC" : 
		((parm.form_netgroup == ON) ? "YES" : "NO"));
     printf("\tcompact bottlenecks to 0 width\t= %s\n", 
		(parm.zero_compact == ON) ? "YES" : "NO");
     printf("\tplacement refinement only\t= %s\n", 
		(parm.placeref_only == ON) ? "YES" : "NO");
     printf("\tthreshold of net criticality\t= %d\n", 
		parm.criticality_threshold);

     printf("\n    -- Tile creation\n");
     printf("\texpansion factor to get a slicing structure\t= %.2f\n", 
		parm.chipExpand);
     printf("\tuse of Z shaped channel\t\t= %s\n", 
		(parm.Zflag == ON) ? "YES" : "NO");

     printf("\n    -- Candidate route selection\n");
     printf("\t# candidate routes in LP\t= %d\n", parm.M);
     printf("\t# routes per pin connection\t= %d\n", parm.m);
     printf("\tnet size limit for huge nets\t= %d\n", parm.netSizeLimit);
     printf("\tuse of Fibonacci heap\t\t= %s\n", 
		(parm.FH_dijkstra == ON) ? "YES" : "NO");

     printf("\n    -- Area minimization\n");
     printf("\tplacement refinement mode\t= %d\n", parm.placeRefMode);
     printf("\tweight factor of net length\t= %f\n", parm.netlength_weight);
     printf("\tweight factor of block location\t= %f\n", parm.blockloc_weight);
     printf("\tmodel of diagonal effect\t= %s\n", 
		(parm.modelDiagonal == ON) ? "YES" : "NO");
     printf("\tthreshold of tile recognition\t= %d\n", parm.tile_threshold);
     printf("\tterminating net / pass-thru net in usage\t= %.2f\n", 
		parm.alpha);
     printf("\texpansion factor for initial dimension of LP\t= %.2f\n", 
		parm.expandFactor);
     printf("\tuse of INTEGER declaration\t= %s\n", 
		(parm.INTEGER == ON) ? "YES" : "NO");
     printf("\tuse of BIP declaration\t\t= %s\n", 
		(parm.BIP == ON) ? "YES" : "NO");

     printf("\n*** End of Mercury parameter information ***\n\n");


}



