
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/misc.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif

/*
 *  misc.c
 *  ******
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains miscellaneous routines and utilities
 *
 *  functions contained in this file:
 *	dataAlloc()
 *	freeLawlerPath()
 *	freePath()
 *	freeRoute()
 *	printAdjMatrix()
 *	printLawler()
 *	printMultiLawler()
 *	printCPUtime()
 *	printFIX()
 *	printINC()
 *	printWGT()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#ifdef hpux
#include <sys/times.h>
#else
#include <sys/time.h>
#endif

#include <stdio.h>
#include <strings.h>
#include <sys/resource.h>
#include "initGlb.h"


/*
 *  dataAlloc()
 *
 *  DESCRIPTION
 *  -----------
 *	Allocates and initializes some of the global variables
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Some variables are initialized here.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	dijkstraPointer[]	: pointers to a struct for dijkstra's
 *				  shortest path algorithm
 *				  lawler's algorithm
 *
 */

dataAlloc()


{
    
    int i;
    
    
    GRegList = NULL;
    GPGraph[G_HOR] = NULL;
    GPGraph[G_VER] = NULL;
    GNet = NULL;
    GNetGr = NULL;
    GNetGrTail = NULL;
    
    for (i = 0; i <= G_MAXNODE - 1; ++i) {
	dijkstraPointer[i] = MALLOC(dijkstraNodeType,1);
	adjList[i] = NULL;
	modAdjList[i] = NULL;
    }
    for (i = 0; i <= G_MAXROUTE - 1; ++i) {
	routePointer[i] = NULL;
    }
    for (i = 0; i <= G_MAXNET - 1; ++i) {
	GNetPointer[i] = NULL;
	netPointer[i] = NULL;
    }
    for (i = 0; i <= G_MAXREGION - 1; ++i) {
	GID2Reg[i] = NULL;
    }
    
    
}




/*
 *  freeLawlerPath()
 *
 *
 *  DESCRIPTION
 *  -----------
 *	Free the memory allocated for lawlerPath struct and its node struct
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathP	<input>	: pointer to the path to be freed
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

freeLawlerPath(pathP)

lawlerPathType	*pathP;

{
    
    lawlerNodeType	*nextNode;
    lawlerNodeType	*saveNode;
    nodeGroupType	*nextClosedNode;
    nodeGroupType	*saveClosedNode;
    
    
    for (nextNode = pathP->node; nextNode != NULL; nextNode = saveNode) {
	
	for (nextClosedNode = nextNode->closedNode; nextClosedNode != NULL;
	     nextClosedNode = saveClosedNode) {
	    saveClosedNode = nextClosedNode->next;
	    FREE(nextClosedNode);
	}
	saveNode = nextNode->next;
	FREE(nextNode);
    }
    FREE(pathP);
    
    
}




/*
 *  freePath()
 *
 *
 *  DESCRIPTION
 *  -----------
 *	Free the memory allocated for path struct and its node struct
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathP	<input>	: pointer to the path to be freed
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

freePath(pathP)

pathType	*pathP;

{
    
    nodeGroupType	*nextNode;
    nodeGroupType	*saveNode;
    
    
    for (nextNode = pathP->node; nextNode != NULL; nextNode = saveNode) {
	saveNode = nextNode->next;
	FREE(nextNode);
    }
    FREE(pathP);
    
    
}




/*
 *  freeRoute()
 *
 *
 *  DESCRIPTION
 *  -----------
 *	Free the memory allocated for route struct and its path & node struct
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	routeP	<input>	: pointer to the route to be freed
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

freeRoute(routeP)

routeType	*routeP;

{
    
    pathType	*nextPath;
    pathType	*savePath;
    
    
    if (routeP == NULL) {
	return;
    }
    for (nextPath = routeP->path; nextPath != NULL; nextPath = savePath) {
	savePath = nextPath->next;
	freePath(nextPath);
    }
    FREE(routeP);
    
    
}




/*
 *  printAdjMatrix()
 *
 *  DESCRIPTION
 *  -----------
 *	Prints an adjacency matrix from an adjacency list
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	adjP	<input>	: pointers to adjacency list
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

printAdjMatrix(adjP)


adjNodeType *adjP[];

{
    
    int		i,j,edgeWeight[G_MAXNODE];
    adjNodeType	*nextNode;
    
    
    printf("adjacency matrix\n");
    for (i = 0; i <= G_MAXNODE - 1; ++i) {
	if (adjP[i] == NULL ) {
	    continue;
	}
	
	for (j = 0; j <= G_MAXNODE - 1; ++j) {
	    edgeWeight[j] = 0;
	}
	
	/* Record all weights of edges adjacent to this node */
	for (nextNode = adjP[i]; nextNode != NULL; nextNode = nextNode->next) {
	    edgeWeight[nextNode->id - 1] = nextNode->weight;
	}
	
	/* Print only weights of those edges that really exist */
	printf("node %d : ", i+1);
	for (j = 0; j <= G_MAXNODE - 1; ++j) {
	    if (adjP[j] != NULL ) {
		printf("%d ",edgeWeight[j]);
	    }
	}
	printf("\n");
    }
    printf("\n");
    
    
}





/*
 *  printLawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Print M shortest paths pointed to by pathPointer[]
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathPtr[]	: pointers to paths to be printed
 *	v	<input>	: starting node
 *	w	<input>	: node to be reached
 *				(representative node in case of multiple target)
 *	M	<input>	: number of shortest paths to be printed
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


printLawler(pathPtr,v,w,M)

lawlerPathType	*pathPtr[];
int		v,w,M;

{
    
    int		i;
    lawlerNodeType	*nextNode;
    
    
    printf("%d shortest paths between %d and %d\n",M,v,w);
    for (i = 0; i <= M - 1; ++i) {
	if (pathPtr[i] == NULL) {
	    break;
	}
	printf("%d length = %d\t",i+1,pathPtr[i]->length);
	for (nextNode = pathPtr[i]->node; nextNode != NULL;
	     nextNode = nextNode->next) {
	    printf(" %d",nextNode->id);
	}
	printf("\n");
    }
    printf("\n");
    
    
}




/*
 *  printMultiLawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Print M shortest routes pointed to by routeP[]
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	routeP	<input> : pointers to routes to be printed
 *	M	<input>	: number of shortest routes to be printed
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


printMultiLawler(routeP,M)

routeType *routeP[G_MAXROUTE];
int M;

{
    
    int		i,j;
    nodeGroupType	*nextNode;
    pathType 	*nextPath;
    
    
    printf("%d shortest routes\n",M);
    for (i = 0; i <= M - 1; ++i) {
	
	if (routeP[i] == NULL) {
	    break;
	}
	
	printf("%d length = %d\n",i+1,routeP[i]->length);
	for (nextPath = routeP[i]->path, j = 0; nextPath != NULL;
	     nextPath = nextPath->next) {
	    
	    printf("\t%d length = %d\t",++j,nextPath->length);
	    for (nextNode = nextPath->node; nextNode != NULL;
		 nextNode = nextNode->next) {
		
		printf("%d ",nextNode->id);
	    }
	    printf("\n");
	}
	printf("\n");
    }
    printf("\n");
    
    
}





/*
 *  printCPUtime()
 *
 *  DESCRIPTION
 *  -----------
 *	Print two kinds of CPU time (user + system time) as a sum of parent
 *	process time and child process time.
 *		1. time spent between the last call and this call
 *		2. total time spent for the execution
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	massage	<input>	: message to be printed before CPU time information
 *	initFlg	<input>	: flag, if YES, this is the first call to this routine
 *
 *  Local variables:
 *  ---------------
 *	usageSelf	: struct for usage data of the parent process
 *	usageChild	: struct for usage data of the child process
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


printCPUtime(message, initFlg)

char	*message;
int	initFlg;

{
    
#ifdef hpux
    struct tms usage;
    long tps;
#else
    struct rusage	usageSelf;
    struct rusage	usageChild;
#endif
    static long	lastCPUtime;
    static long	totalTime;
    
    printf("%s", message);
    
    if (initFlg == YES) {
	lastCPUtime = 0;
    }
#ifdef hpux
    times(&usage);
    tps = sysconf(CLK_TCK);
    totalTime = (usage.tms_utime + usage.tms_stime
		+ usage.tms_cutime + usage.tms_cstime) * (1000/tps);
	/* note that it is unclear from the documentation whether
		cutime and cstime include utime and stime */
#else
    getrusage(RUSAGE_SELF, &usageSelf);
    getrusage(RUSAGE_CHILDREN, &usageChild);
    totalTime = usageSelf.ru_utime.tv_sec + usageSelf.ru_stime.tv_sec
      + usageChild.ru_utime.tv_sec + usageChild.ru_stime.tv_sec;
#endif
    printf("interval time = %d  total time = %d\n\n",
	   totalTime - lastCPUtime, totalTime);
    lastCPUtime = totalTime;
    
    
}




/*
 *  printFIX()
 *
 *  DESCRIPTION
 *  -----------
 *	Print FIX matrix.
 *
 */

printFIX()

{
    int	i,j;
    
    printf("Printing FIX matrix .....\n");
    for (i = 0; i <= NO_REGIONTYPE-1; ++i) {
	printf("i = %d\n", i);
	for (j = 0; j <= G_MAXREGION - 1; ++j) {
	    printf(" %d",FIX[i][j]);
	    if ((j+1) % 10 == 0) {
		printf("\n");
	    }
	}
	printf("\n");
    }
    printf("\n");
}




/*
 *  printINC()
 *
 *  DESCRIPTION
 *  -----------
 *	Print INC matrix.
 *
 */

printINC()

{
    
    int	i,j,k;
    
    printf("Printing INC matrix .....\n");
    for (i = 0; i <= NO_REGIONTYPE-2; ++i) {
	printf("i = %d\n", i);
	for (j = 0; j <= G_MAXPOLAR - 1; ++j) {
	    printf("j = %d\n", j);
	    for (k = 0; k <= G_MAXREGION - 1; ++k) {
		printf(" %d",INC[i][j][k]);
		if ((k+1) % 10 == 0) {
		    printf("\n");
		}
	    }
	    printf("\n");
	}
	printf("\n");
    }
    printf("\n");
}



/*
 *  printWGT()
 *
 *  DESCRIPTION
 *  -----------
 *	Print WGT matrix.
 *
 */

printWGT()

{
    
    int	i,j;
    
    for (i = 0; i <= G_MAXREGION - 1; ++i) {
	printf("i = %d\n", i);
	for (j = 0; j <= G_MAXROUTE * G_MAXNETGR - 1; ++j) {
	    printf(" %.1f",WGT[i][j]);
	    if ((j+1) % 10 == 0) {
		printf("\n");
	    }
	}
    }
    printf("\n");
    
}


