
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/multiLawler.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  multiLawler.c
 *  *************
 *	Routines for obtaining M shortest routes for n point net
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for obtaining M shortest routes for a multi
 *	pin net.
 *
 *  functions contained in this file:
 *	multiLawler()
 *	selectPin()
 *	connectNet()
 *	moveNode()
 *	movePath()
 *	removeNode()
 *	removePath()
 *	mergePathList()
 *	buildRoute()
 *	insertMultiLawler()
 *	askSameRoute()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "initGlb.h"



/*
 *  multiLawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Obtain M shortest routes for a multiple pin net
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	netP	<input>	: pointer to the net to be routed
 *	numPinGr	<input>	: # of equivalent pin groups in the net
 *	M	<input>	: number of shortest route to be obtained
 *	Mcon	<input>	: # of routes to be selected per pin connection
 *
 *  Local variables:
 *  ---------------
 *	v		: the first node to connect route
 *	initial		: flag, if ON, this is the first node of the net
 *	count		: counter, connecting count+1 - th node now
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Uses pathList[][] and fills in values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	C		: pointer to the 1st node in the connected node group
 *	U		: pointer to the 1st node in the unconnected node group
 *	pathList[][]	: pointers to list of paths obtained at each iteration
 *
 */


multiLawler(netP,numPinGr,M,Mcon)


netType *netP;
int	numPinGr;
int	M;
int	Mcon;

{

netNodeType	*nextNodeU;
netNodeType	*nextNetNode;
int v;
int initial;
int i,j,count;


/* Some initialization */
/*---------------------*/

for (i = 0; i <= G_MAXNODE - 1; ++i) {
    for (j = 0; j <= G_MAXPATH - 1; ++j) {
	pathList[i][j] = NULL;
    }
    pathNo[i] = 0;
}
for (i = 0; i <= G_MAXROUTE - 1; ++i) {
    routePointer[i] = NULL;
}
C = NULL;


/* Copy the net node information to U group */

initial = ON;
for (nextNetNode = netP->node; nextNetNode != NULL;
	nextNetNode = nextNetNode->next) {

    if (initial == ON) {
	U = nextNodeU = MALLOC(netNodeType,1);
	initial = OFF;
    } else {
	nextNodeU->next = MALLOC(netNodeType,1);
	nextNodeU = nextNodeU->next;
    }
    nextNodeU->id = nextNetNode->id;
    nextNodeU->next = NULL;
    nextNodeU->equiv = nextNetNode->equiv;
}


/* Choose the first node */

v = selectPin(netP);

/* Move the node and its electrically equivalent nodes from U to C */

moveNode(v);

/* Start to get M shortest routes */

count = 1;
connectNet(netP,numPinGr,M,Mcon,count);

printMultiLawler(routePointer,M);


}




/*
 *  selectPin()
 *
 *  DESCRIPTION
 *  -----------
 *	Choose the first pin to start connection with.
 *		For present, it is the first pin in net.
 *
 *  RETURNS
 *  -------
 *	node	 the first pin to start connection with
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	v		: the first node to connect route
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	None
 *
 */


selectPin(netP)

netType *netP;


{

int v;

v = netP->node->id;

#ifdef MULTI_LAWLER_DEBUG
printf("choosing pin = %d\n\n",v);
#endif

return(v);

}




/*
 *  connectNet()
 *
 *  DESCRIPTION
 *  -----------
 *	Obtain M shortest routes for a multiple pin net
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	netP	<input>		: pointer to the net to be routed
 *	numPinGr	<input>		: # equivalent pin groups in net group
 *	M	<input>		: number of shortest route to be obtained
 *	Mcon	<input>		: # of routes to be selected per pin connection
 *	count	<in/out>	: counter, connecting count+1 - th node now
 *
 *  Local variables:
 *  ---------------
 *	routeP		: pointer to a new route struct
 *	Mlaw		: # paths to be obtained by Lawler's algorithm
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Uses pathList[][] and fills in values
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	C		: pointer to the 1st node in the connected node group
 *	U		: pointer to the 1st node in the unconnected node group
 *	pathList[][]	: pointers to list of paths obtained at each iteration
 *
 */


connectNet(netP,numPinGr,M,Mcon,count)


netType	*netP;
int	numPinGr;
int	M,Mcon,count;

{


netNodeType	*nextNode;
nodeGroupType	*nextEqNode;
routeType	*routeP,*buildRoute();
int		i;
int		Mlaw;


/* Initialization */
/*----------------*/
for (i = 0; i <= G_MAXPATH - 1; ++i) {
    pathList[count-1][i] = NULL;
}


if (U != NULL) {

/* For each node x in U and its electrically equivalent pin
 *	obtain M shortest paths from x to target C */

    /* Compute Mlaw */
    /*--------------*/
    if (numPinGr == 2) {
	Mlaw = M;
    } else {
  	Mlaw = Mcon;
    }

    for (nextNode = U; nextNode != NULL; nextNode = nextNode->next) {

	lawler(nextNode->id,C,Mlaw);
	mergePathList(count,Mcon);

	for (nextEqNode = nextNode->equiv; nextEqNode != NULL;
		nextEqNode = nextEqNode->next) {

	    lawler(nextEqNode->id,C,Mlaw);
	    mergePathList(count,Mcon);
	}
    }

#ifdef MULTI_LAWLER_DEBUG
    printf("count = %d : list of %d shortest paths\n",count,Mcon);
    printLawler(&pathList[count-1][0],0,0,Mcon);
#endif

/*
 * For each of Mcon shortest paths in the pathList[] for that count,
 *	add the path and all the electrically equivalent nodes to C
 *	connectNet again and
 *	remove the path and all the electrically equivalent nodes from C
 */

    for (i = 0; i <= Mcon - 1; ++i) {
	if (pathList[count-1][i] == NULL) {
	    break;
	}
	movePath(pathList[count-1][i],netP);
	pathNo[count-1] = i + 1;

#ifdef MULTI_LAWLER_DEBUG
	printf("count = %d  new count = %d  processing pathList[%d][%d]\n",
		count,count+1,count-1,i);
#endif

	connectNet(netP,numPinGr,M,Mcon,count+1);
	removePath(pathList[count-1][i],netP);
    }
    pathNo[count-1] = 0;
    return;

} else {

/* All the pins connected.
 *	Build a route struct and fill in data
 *	And insert it in an appropriate position in net struct by its length */

    routeP = buildRoute();
    insertMultiLawler(1,M,routePointer,routeP);
}



}




/*
 *  moveNode()
 *
 *  DESCRIPTION
 *  -----------
 *	Move node v and its equivalent nodes from U to the tail of C
 *	v should be the parent node of the equivalent node group
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: node to be moved
 *
 *  Local variables:
 *  ---------------
 *	initial		: flag, if ON, this is the first node of U
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Changes U and C
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	C	: pointer to the first node in the connected node group
 *	U	: pointer to the first node in the unconnected node group
 *
 */


moveNode(v)


int v;

{

netNodeType *nextNodeC,*nextNodeU;
netNodeType *preNodeU;
int initial;


/* Search node v in U (unconnected node list) */
/*--------------------------------------------*/
nextNodeU = U;
initial = ON;
while (nextNodeU->id != v) {
    preNodeU = nextNodeU;
    nextNodeU = nextNodeU->next;
    initial = OFF;
}

/* Disconnect node v from U */
/*--------------------------*/
if (initial == ON) {
    U = nextNodeU->next;
} else {
    preNodeU->next = nextNodeU->next;
}
nextNodeU->next = NULL;


/* Insert node v in C */
/*--------------------*/
if (C == NULL) {
    C = nextNodeU;
} else {
    nextNodeC = C;
    while (nextNodeC->next != NULL) {
	nextNodeC = nextNodeC->next;
    }
    nextNodeC->next = nextNodeU;
}


}




/*
 *  movePath()
 *
 *  DESCRIPTION
 *  -----------
 *	Move all nodes in the given path to C
 *		If it is a net node, move all its equivalent nodes from U to C
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathP	<input>	: pointer to the path to be moved
 *	netP	<input>	: pointer to the net to be routed
 *
 *  Local variables:
 *  ---------------
 *	flag		: flag, if FOUND, the current node is included in net
 *	processed	: flag, if ON, already processed by its equivalent
 *	parent		: parent node of the equivalent node group in the net
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Change U and C by moving some nodes.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	C	: pointer to the first node in the connected node group
 *	U	: pointer to the first node in the unconnected node group
 *
 */


movePath(pathP,netP)


lawlerPathType	*pathP;
netType		*netP;

{

netNodeType	*nextNodeC;
netNodeType	*nextNetNode;
nodeGroupType	*nextEqNode;
netNodeType	*preNodeC;
lawlerNodeType	*nextNodeP;
int		flag,processed;
int		parent;


/* Move nodes in the path to C, except the last node (it is already in C) */
/*------------------------------------------------------------------------*/
for (nextNodeP = pathP->node; nextNodeP->next != NULL;
	nextNodeP = nextNodeP->next) {

/* Check if the node belongs to the net */
/*--------------------------------------*/
    flag = NOT_FOUND;
    parent = NOT_FOUND;
    for (nextNetNode = netP->node; nextNetNode != NULL;
		nextNetNode = nextNetNode->next) {

	if (nextNetNode->id == nextNodeP->id) {
	    parent = nextNetNode->id;
	    flag = FOUND;
	    break;
	}
/* Also check the equivalent nodes */
	for (nextEqNode = nextNetNode->equiv; nextEqNode != NULL;
		nextEqNode= nextEqNode->next) {

	    if (nextEqNode->id == nextNodeP->id) {
		parent = nextNetNode->id;
		flag = FOUND;
		break;
	    }
	}

	if (flag == FOUND) {
	    break;
	}
    }

/* Check if the node has already been processed 
 *	by its preceding equivalent node
 * If this is not a net node, it just takes you to the tail of C */
/*---------------------------------------------------------------*/
    processed = NO;
    nextNodeC = C;
    while (nextNodeC != NULL) {
	if (nextNodeC->id == parent) {
	    processed = YES;
	    break;
	}
	preNodeC = nextNodeC;
	nextNodeC = nextNodeC->next;
    }

/* If it is already processed, skip this node */

    if (processed == YES) {
	continue;
    }

/* If the node belongs to the net, call the routine, otherwise manually */
/*----------------------------------------------------------------------*/
    if (flag == FOUND) {
	moveNode(parent);
    } else {

/* Need to allocate memory for a node and attach it to C */
/* You already have the tail node of C in preNodeC */
	preNodeC->next = MALLOC(netNodeType,1);
	preNodeC->next->id = nextNodeP->id;
	preNodeC->next->next = NULL;
	preNodeC->next->equiv = NULL;
    }
}


}




/*
 *  removeNode()
 *
 *  DESCRIPTION
 *  -----------
 *	Move v and its equivalent nodes from C to the head of U
 *	v should be the parent node of the equivalent node group
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: node to be moved
 *
 *  Local variables:
 *  ---------------
 *	initial		: flag, if ON, this is the first node of C
 *
 *  Possible errors:
 *  ---------------
 *
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Changes U and C
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	C	: pointer to the first node in the connected node group
 *	U	: pointer to the first node in the unconnected node group
 *
 */


removeNode(v)


int v;

{

netNodeType	*nextNodeC;
netNodeType	*preNodeC;
int initial;


nextNodeC = C;
initial = ON;
while (nextNodeC->id != v) {
    preNodeC = nextNodeC;
    nextNodeC = nextNodeC->next;
    initial = OFF;
}
if (initial == ON) {
    C = nextNodeC->next;
} else {
    preNodeC->next = nextNodeC->next;
}

nextNodeC->next = U;
U = nextNodeC;


}




/*
 *  removePath()
 *
 *  DESCRIPTION
 *  -----------
 *	Remove all nodes in the given path from C
 *		If it is a net node, move all its equivalent nodes from C to U
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	pathP	<input> : pointer to the path to be moved
 *	netP	<input> : pointer to the net to be routed
 *
 *  Local variables:
 *  ---------------
 *	flag		: flag, if FOUND, the current node is included in net
 *	processed	: flag, if ON, already processed by its equivalent
 *	initial		: flag, if ON, first node in the C group
 *	parent		: parent node of the equivalent node group in the net
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Change U and C by moving some nodes.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	C	: pointer to the first node in the connected node group
 *	U	: pointer to the first node in the unconnected node group
 *
 */


removePath(pathP,netP)


lawlerPathType	*pathP;
netType 	*netP;

{

netNodeType	*nextNodeC,*nextNodeU;
netNodeType	*nextNetNode;
netNodeType	*preNodeC;
lawlerNodeType	*nextNodeP;
nodeGroupType	*nextEqNode;
int		flag,processed,initial;
int		parent;


/* Remove nodes in the path from C, except the last node (it should be in C) */
/*---------------------------------------------------------------------------*/
for (nextNodeP = pathP->node; nextNodeP->next != NULL;
	nextNodeP = nextNodeP->next) {

/* Check if the node belongs to the net */
/*--------------------------------------*/
    flag = NOT_FOUND;
    parent = NOT_FOUND;
    for (nextNetNode = netP->node; nextNetNode != NULL;
		nextNetNode = nextNetNode->next) {

	if (nextNetNode->id == nextNodeP->id) {
	    parent = nextNetNode->id;
	    flag = FOUND;
	    break;
	}
/* Also check the equivalent nodes */
	for (nextEqNode = nextNetNode->equiv; nextEqNode != NULL;
		nextEqNode= nextEqNode->next) {

	    if (nextEqNode->id == nextNodeP->id) {
		parent = nextNetNode->id;
		flag = FOUND;
		break;
	    }
	}

	if (flag == FOUND) {
	    break;
	}
    }

/* Check if the node has already been processed 
 *	by its preceding equivalent node
 * If this is not a net node, it just takes you to the tail of U */
/*---------------------------------------------------------------*/
    processed = NO;
    nextNodeU = U;
    while (nextNodeU != NULL) {
	if (nextNodeU->id == parent) {
	    processed = YES;
	    break;
	}
	nextNodeU = nextNodeU->next;
    }
    if (processed == YES) {
	continue;
    }

/* If the node belongs to the net, call the routine, otherwise manually */
/*----------------------------------------------------------------------*/
    if (flag == FOUND) {
	removeNode(parent);
    } else {

/* If not a net node, remove it from C and free the memory for it */
	nextNodeC = C;
	initial = ON;
	while (nextNodeC != NULL) {
	    if (nextNodeC->id == nextNodeP->id) {
		break;
	    }
	    initial = OFF;
	    preNodeC = nextNodeC;
	    nextNodeC = nextNodeC->next;
	}
	if (initial == ON) {
	    C = nextNodeC->next;
	} else {
	    preNodeC->next = nextNodeC->next;
	}
	FREE(nextNodeC);
    }
}


}



/*
 *  mergePathList()
 *
 *  DESCRIPTION
 *  -----------
 *	Insert each of M shortest paths in the pathList of current count
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	count	<input>	: counter, connecting count+1 - th node now
 *	M	<input>	: number of shortest route to be obtained
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Changes pathList
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	pathList[][]	: pointers to list of paths obtained for that count
 *
 */


mergePathList(count,M)


int count,M;

{

int i,j,k;


/* Do below for each of M shortest paths from current source node in U to C */
/*--------------------------------------------------------------------------*/
for (i = 0; i <= M - 1; ++i) {

    if (pathPointer[i] == NULL) {
	break;
    }

/* Scan every path in the list for that "count" */
/*----------------------------------------------*/
    for (j = 0; j <= G_MAXPATH - 1; ++j) {

/* If come to the tail, add this path if # of elements does not exceed M */
	if (pathList[count-1][j] == NULL) {
	    if (j <= M - 1) {
		pathList[count-1][j] = pathPointer[i];
	    } else {
		freeLawlerPath(pathPointer[i]);
		pathPointer[i] = NULL;
	    }
	    break;
	}
/* If find a path whose length is greater, insert before this element */
	if (pathPointer[i]->length < pathList[count-1][j]->length) {
	    for (k = G_MAXPATH - 2; k >= j; --k) {
		if ((k == M - 1) && (pathList[count-1][k] != NULL)) {
		    freeLawlerPath(pathList[count-1][k]);
		    pathList[count-1][k] = NULL;
		}
		if (k + 1 <= M - 1) {
		    pathList[count-1][k+1] = pathList[count-1][k];
		}
	    }
	    pathList[count-1][j] = pathPointer[i];
	    break;
	}
    }
}


}




/*
 *  buildRoute()
 *
 *  DESCRIPTION
 *  -----------
 *	Build a new route struct
 *
 *  RETURNS
 *  -------
 *	routeP		pointer to a new struct for a route
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	routeP		: pointer to a new struct for a route
 *	i		: index, represents i-th iteration of multiLawler
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	pathList[][]	: pointers to list of paths obtained
 *	pathNo[]	: indicates which path of M shortest paths should be
 *				used for route construction
 *
 */


routeType *buildRoute()


{

int		i;
pathType	*nextPath;
lawlerNodeType	*nextNodeS;
nodeGroupType	*nextNodeD;
routeType	*routeP;


/* Allocate memory for new struct */
/*--------------------------------*/
routeP = MALLOC(routeType,1);
routeP->length = 0;
routeP->path = NULL;
routeP->next = NULL;
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    routeP->regionFlg[i] = OFF;
}


/* Do below for each path constructed at each phase of multiLawler */
/*-----------------------------------------------------------------*/
for (i = 0; i <= G_MAXNODE - 1; ++i) {

    if (pathNo[i] == 0) {
	break;
    }
/* Create a new path struct and increment route length */
/*-----------------------------------------------------*/
    if (routeP->path == NULL) {
	nextPath = routeP->path = MALLOC(pathType,1);
    } else {
	nextPath->next = MALLOC(pathType,1);
	nextPath = nextPath->next;
    }
    routeP->length += pathList[i][pathNo[i]-1]->length;
    nextPath->length = pathList[i][pathNo[i]-1]->length;
    nextPath->node = NULL;
    nextPath->next = NULL;

/* Allocate memory for each node and copy the node information */
/*-------------------------------------------------------------*/
    for (nextNodeS = pathList[i][pathNo[i]-1]->node; nextNodeS != NULL;
		nextNodeS = nextNodeS->next) {

	if (nextPath->node == NULL) {
	    nextNodeD = nextPath->node = MALLOC(nodeGroupType,1);
	} else {
	    nextNodeD->next = MALLOC(nodeGroupType,1);
	    nextNodeD = nextNodeD->next;
	}
	nextNodeD->id = nextNodeS->id;
	nextNodeD->next = NULL;

/* Check every node included in the path */
/*---------------------------------------*/
	routeP->regionFlg[nextNodeS->id - 1] = ON;
    }
}
return(routeP);


}




/*
 *  insertMultiLawler()
 *
 *  DESCRIPTION
 *  -----------
 *	Insert the given route in the route list according to its length.
 *	The route list is sorted by the route length in non-decreasing order.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	m	<input>	: route index, search from m th route in the list.
 *	M	<input>	: number of shortest route to be obtained
 *	routePtr<in/out> : pointers to routes obtained by this algorithm
 *	routeP	<input>	: route to be inserted in the list
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Changes route list.
 *	Free space for overflowing struct if list exceeds the limit.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


insertMultiLawler(m,M,routePtr,routeP)


int		m,M;
routeType	*routePtr[G_MAXROUTE],*routeP;

{

int i,j;


/* scan the routes between m-1 th and G_MAXROUTE-1 th in the list */

for (i = m - 1; i <= G_MAXROUTE - 1; ++i) {

/* if come to the tail, add this path if # of elements does not exceed M */
    if (routePtr[i] == NULL) {
	if (i <= M - 1) {
	    routePtr[i] = routeP;
	} else {
	    freeRoute(routeP);
	}
	break;
    }

/* if find a route whose length is greater, insert before this element */
    if (routeP->length < routePtr[i]->length) {
	for (j = G_MAXROUTE - 2; j >= i; --j) {
	    if ((j == M - 1) && (routePtr[j] != NULL)) {
		freeRoute(routePtr[j]);
		routePtr[j] = NULL;
	    }
	    if (j + 1 <= M - 1) {
		routePtr[j+1] = routePtr[j];
	    }
	}
	routePtr[i] = routeP;
	break;
    }
/* if find a route whose length is equal, check if these are the same routes */
    if (routeP->length == routePtr[i]->length) {
	if (askSameRoute(routeP,routePtr[i]) == YES) {
	    freeRoute(routeP);
	    break;
	}
    }
}


}





/*
 *  askSameRoute()
 *
 *  DESCRIPTION
 *  -----------
 *	Check if the given two routes are the same by matching the sets of
 *		regions that constiute the route
 *
 *  RETURNS
 *  -------
 *	YES	if the routes are the same
 *	NO	Otherewise
 *
 *  Arguments:
 *  ---------
 *	routeP1 <input> : pointer to the first route
 *	routeP2	<input>	: pointer to the second route
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


askSameRoute(routeP1,routeP2)

routeType	*routeP1, *routeP2;

{

int	i;


for (i = 0; i <= G_MAXREGION - 1; ++i) {
    if (routeP1->regionFlg[i] == routeP2->regionFlg[i]) {
	continue;
    } else {
	return(NO);
    }
}
return(YES);


}



