 
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/netGroup.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  netGroup.c
 *  **********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for forming net groups
 *
 *  functions contained in this file:
 *	makeNetGroup()
 *	matchNetGroup()
 *	buildNetGroup()
 *	makeRegionPin()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"

typedef struct nodeChain {
    int			id;
    struct nodeChain	*pre;
    struct nodeChain	*next;
    }
    nodeChain;





/*
 *  makeNetGroup()
 *
 *  DESCRIPTION
 *  -----------
 *	Check for each net the combination of regions to which pins of net
 *		belong and form groups of nets with the same combination.
 *	Critical net forms a net group by itself.
 *	Also attach pin struct to region for calculating max column density.
 *		Only pins w/ max or min position in the region
 *			for the net are attached.
 *
 *  RETURNS
 *  -------
 *	netGrCount	# of net groups formed
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	regChain[i]	: pointers to nodeChain corresponding to region i
 *	regionCount	: number of the regions above
 *	newNetGr	: pointer to a new struct for a net group
 *	maxOffset	: max offset among pins of a net in this region
 *	minOffset	: min offset among pins of a net in this region
 *	regionOfPin	: ID of the only one region where net pins exist
 *	oneRegionFlg	: flag, if ON, the net has only one region
 *	matchFlg	: flag, if ON, the pattern matches
 *
 *  Possible errors:
 *  ---------------
 *	Works correctly only if regions in existent net groups are sorted by ID
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Form net groups and fill in value in GNetGr
 *	Especially GNetDef and GRegionDef are tied to GNetGrDef by pointers
 *	Creates pin struct and attach it to region w/ some useful infomation
 *		for max column density calculation.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GNet		: pointer to first net
 *	GNetGr		: pointer to first net group
 *	GID2Reg[i]	: pointers to region whose id is i + 1
 *
 *
 */


makeNetGroup()


{

nodeChain	*regChain[G_MAXREGION];
nodeChain	*nextNodeChain;

GNetGrDef	*netGroup[G_MAXREGION * G_MAXREGION / 2];
GNetGrDef	*matchNetGroup();
int		regionCount;
int		i;
int		regionOfPin;
int		oneRegionFlg;
int		maxOffset, minOffset;
GNetDef		*nextNet, *saveNet;
GPinDef		*nextPin, *nextEqPin;
GNetGrDef	*nextNetGr;
int		netGrCount;



/* Initialization */
/*----------------*/
for (i = 0; i <= G_MAXREGION - 1; ++i) {
    regChain[i] = NULL;
}
for (i = 0; i <= G_MAXREGION * G_MAXREGION / 2 - 1; ++i) {
    netGroup[i] = NULL;
}

/* Scan each net */
/*---------------*/
for (netGrCount = 0, nextNet = GNet; nextNet != NULL; nextNet = nextNet->next) {

    /* Initialization */
    /*----------------*/
    for (i = 0; i <= G_MAXREGION - 1; ++i) {
	if (regChain[i] != NULL) {
	    FREE(regChain[i]);
	    regChain[i] = NULL;
	}
    }

    /* Sort id's of regions in this net by constructing nodeChain's */
    /*--------------------------------------------------------------*/
    regionCount = 0;
    regionOfPin = nextNet->pin->region->id;

    if (GID2Reg[nextNet->pin->region->id - 1] == NULL) {
	printf("pin->region is strange\n");
    }

    oneRegionFlg = ON;
    minOffset = MAXINT;
    maxOffset = 0;

    /* Scan each pin */
    /*---------------*/
    for (nextPin = nextNet->pin; nextPin != NULL; nextPin = nextPin->next) {

	/* If this is a new region, allocate space */
	if (regChain[nextPin->region->id - 1] == NULL) {
	    ++regionCount;
	    regChain[nextPin->region->id - 1] = MALLOC(nodeChain,1);
	    regChain[nextPin->region->id - 1]->id = nextPin->region->id;
	    regChain[nextPin->region->id - 1]->pre = NULL;
	    regChain[nextPin->region->id - 1]->next = NULL;
	}

	/* Scan equivalent pins */
	/*----------------------*/
	for (nextEqPin = nextPin->equiv; nextEqPin != NULL;
		nextEqPin = nextEqPin->equiv) {

	    /* If this is a new region, allocate space */
	    if (regChain[nextEqPin->region->id - 1] == NULL) {
		++regionCount;
		regChain[nextEqPin->region->id - 1] = MALLOC(nodeChain,1);
		regChain[nextEqPin->region->id - 1]->id = nextEqPin->region->id;
		regChain[nextEqPin->region->id - 1]->pre = NULL;
		regChain[nextEqPin->region->id - 1]->next = NULL;
	    }
	    /* Locate the current head of nodeChain to which this EqPin 
	       should belong */
	    nextNodeChain = regChain[nextPin->region->id - 1];
	    while (nextNodeChain->pre != NULL) {
		nextNodeChain = nextNodeChain->pre;
	    }
	    /* Scan the chain to find the place where this EqPin should be */
	    while (nextNodeChain != NULL) {
		if (nextEqPin->region->id < nextNodeChain->id) {
		    break;
		}
		nextNodeChain = nextNodeChain->next;
	    }
	    if (nextNodeChain->pre != NULL) {
		nextNodeChain->pre->next = regChain[nextEqPin->region->id - 1];
	    }
	    regChain[nextEqPin->region->id - 1]->next = nextNodeChain;
	    regChain[nextEqPin->region->id - 1]->pre = nextNodeChain->pre;

	    /* Get the min and max offset of pin locations if the net has 
	       only one region */
	    if ((oneRegionFlg == ON) &&
			(regionOfPin != nextEqPin->region->id)) {
		oneRegionFlg = OFF;
	    }
	    if (oneRegionFlg == ON) {
		if (minOffset > nextEqPin->offset) {
		    minOffset = nextPin->offset;
		}
		if (maxOffset < nextPin->offset) {
		    maxOffset = nextPin->offset;
		}
	    }
	}

	/* Get the min and max offset of pin locations if the net has only 
	   one region */
	if ((oneRegionFlg == ON) && (regionOfPin != nextPin->region->id)) {
	    oneRegionFlg = OFF;
	}
	if (oneRegionFlg == ON) {
	    if (minOffset > nextPin->offset) {
		minOffset = nextPin->offset;
	    }
	    if (maxOffset < nextPin->offset) {
		maxOffset = nextPin->offset;
	    }
	}
    }

    /* Create new pin struct and attach it to region if the net has 1 region */
    /*-----------------------------------------------------------------------*/
    if (oneRegionFlg == ON) {
	makeRegionPin(GID2Reg[regionOfPin -  1], minOffset, ON);
	makeRegionPin(GID2Reg[regionOfPin -  1], maxOffset, OFF);
    }

    /* Compare the combination of regions between each net group and this net */
    /*------------------------------------------------------------------------*/
    if ((nextNetGr = 
		matchNetGroup(nextNet,regionCount,regChain,netGroup)) != NULL 
		&& parm.form_netgroup == ON
		&& nextNet->critical == NO) {

	/* Add the net to head of this net group */

	saveNet = nextNetGr->net;
	nextNetGr->net = nextNet;
	nextNetGr->net->nextGrMem = saveNet;
	++(nextNetGr->NoNet);
	nextNet->netGr = nextNetGr;

    } else {

	/* Otherwise, construct a new net group */
	buildNetGroup(nextNet,++netGrCount,regionCount,regChain,netGroup);
    }
}

return(netGrCount);


}





/*
 *  matchNetGroup()
 *
 *  DESCRIPTION
 *  -----------
 *	Find a net group which has the same net topology as the given regChain
 *		or return NULL otherwise.
 *
 *  RETURNS
 *  -------
 *	*netGr	: pointer to the matched net group, if no match, NULL
 *
 *  Arguments:
 *  ---------
 *	nextNet	<input>	: pointer to the net to be matched
 *	regionCount	<input>	: # of regions included in this net group
 *	regChain<input>	: pointers to region chains sorted by id's
 *	netGroup[i]<input>: pointers to net groups which has 2-pin combination i
 *
 *  Local variables:
 *  ---------------
 *	newNetGr	: pointer to a new struct for a net group
 *      nextPin		: used to scan pins in nextNet
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Form net groups and fill in value in GNetGr
 *	Especially GNetDef and GRegionDef are tied to GNetGrDef by pointers
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GNetGr		: pointer to first net group
 *	GID2Reg[i]	: pointers to region whose id is i + 1
 *
 *
 */

GNetGrDef
*matchNetGroup(nextNet,regionCount,regChain,netGroup)

GNetDef		*nextNet;
int		regionCount;
nodeChain	*regChain[G_MAXREGION];
GNetGrDef	*netGroup[G_MAXREGION * G_MAXREGION / 2];


{


    GNetGrDef	*nextNetGr;
    int		matchFlg;
    GNodeDef	*nextNode, *nextEqNode;
    int		i,n;
    int		idx;
    nodeChain	*nextNodeChain;
    int		reg1,reg2,saveReg;
    GPinDef	*nextPin;


    /* If the net is a 2-pin net, match by table lookup */
    /*--------------------------------------------------*/
    n = G_MAXREGION;
    if (regionCount == 2) {
	reg1 = nextNet->pin->region->id;
	for (nextPin=nextNet->pin; nextPin!=NULL; nextPin=nextPin->next)
	    if (nextPin->region->id!=reg1) break;
	if (nextPin==NULL) {
	    fprintf(stderr, "ERROR cannot find the second pin for net %s\n",
		nextNet->name);
	    abort();
	}
	reg2 = nextPin->region->id;
	if (reg1 > reg2) {
	    saveReg = reg1;
	    reg1 = reg2;
	    reg2 = saveReg;
	}
	idx = ((reg1 - 1) * (2 * n - reg1)) / 2 + (reg2 - reg1);
	return(netGroup[idx - 1]);
    }

    /* Compare the combination of regions between each net group and this net */
    /*------------------------------------------------------------------------*/
    for (nextNetGr = GNetGr; nextNetGr != NULL; nextNetGr = nextNetGr->next) {

	matchFlg = ON;

	/* If #region is different, obviously does not match */
	if (nextNetGr->NoRegion != regionCount) {
	    continue;
	}
	/* Skip the net group of critical net */
	if (nextNetGr->net->critical == YES) {
	    continue;
	}

	/* Compare each region in this net group with that of current net */
	/*	Regions in existent net groups are already sorted */

	/* Scan each node */
	nextNode = nextNetGr->node;
	for (i = 0; i <= G_MAXREGION - 1; ++i) {

	    if ((regChain[i] == NULL) || (regChain[i]->pre != NULL)) {
		continue;
	    }
	    if (nextNode == NULL) {
		matchFlg = OFF;
		break;
	    }
	    if (nextNode->region->id != regChain[i]->id) {
		matchFlg = OFF;
		break;
	    }

	    /* Scan each equivalent node */
	    nextEqNode = nextNode->nextEq;
	    nextNodeChain = regChain[i]->next;
	    while ((nextEqNode != NULL) && (nextNodeChain != NULL)) {
		if (nextEqNode->region->id != nextNodeChain->id) {
		    break;
		}
		nextEqNode = nextEqNode->nextEq;
		nextNodeChain = nextNodeChain->next;
	    }
	    if ((nextEqNode != NULL) || (nextNodeChain != NULL)) {
		matchFlg = OFF;
		break;
	    }

	    nextNode = nextNode->next;
	}

	/* If both regions match, found the net group, otherwise try next */
	/*----------------------------------------------------------------*/
	if ((i == G_MAXREGION) && (nextNode == NULL)) {
	    return(nextNetGr);
	}
    }

    return(NULL);


}




/*
 *  buildNetGroup()
 *
 *  DESCRIPTION
 *  -----------
 *	Create a new net group struct
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	nextNet	<input>	: pointer to net which belongs to this net group
 *	regionCount	<input>	: # of regions included in this net group
 *	regChain<input>	: pointers to region chains sorted by id's
 *	netGroup[i]<out>: pointers to net groups which has 2-pin combination i
 *
 *  Local variables:
 *  ---------------
 *	newNetGr	: pointer to a new struct for a net group
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Form net groups and fill in value in GNetGr
 *	Especially GNetDef and GRegionDef are tied to GNetGrDef by pointers
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GNetGr		: pointer to first net group
 *	GNetGrTail	: pointer to last net group
 *	GID2Reg[i]	: pointers to region whose id is i + 1
 *
 *
 */

buildNetGroup(nextNet,netGrCount,regionCount,regChain,netGroup)

GNetDef		*nextNet;
int		netGrCount;
int		regionCount;
nodeChain	*regChain[G_MAXREGION];
GNetGrDef	*netGroup[G_MAXREGION * G_MAXREGION / 2];


{


    GNetGrDef	*newNetGr;
    int		i,n;
    int		idx;
    int		reg1,reg2,saveReg;
    GNodeDef	*nextNode;
    GNodeDef	*nextEqNode;
    nodeChain	*nextNodeChain;



    if (GNetGr == NULL) {
	newNetGr = GNetGr = GNetGrTail = MALLOC(GNetGrDef,1);
    } else {
	newNetGr = GNetGrTail->next = MALLOC(GNetGrDef,1);
	GNetGrTail = newNetGr;
    }

    /* In case of 2-pin net, obtain serial number in all possible 2-pin 
     * combinations of regions and use it as index
     *---------------------------------------------------------------*/
    n = G_MAXREGION;
    if (regionCount == 2 && nextNet->critical == NO) {
	reg1 = nextNet->pin->region->id;
	reg2 = nextNet->pin->next->region->id;
	if (reg1 > reg2) {
	    saveReg = reg1;
	    reg1 = reg2;
	    reg2 = saveReg;
	}
	idx = ((reg1 - 1) * (2 * n - reg1)) / 2 + (reg2 - reg1);
	netGroup[idx - 1] = newNetGr;
    }

    newNetGr->id = netGrCount;
    newNetGr->NoNet = 1;
    if (nextNet->critical == YES) {
	newNetGr->NoCritNet = 1;
    } else {
	newNetGr->NoCritNet = 0;
    }
    newNetGr->NoRegion = regionCount;
    newNetGr->net = nextNet;
    newNetGr->node = NULL;
    newNetGr->next = NULL;

    /* Construct region info in this net group from nodeChain struct */
    for (i = 0; i <= G_MAXREGION - 1; ++i) {
	if ((regChain[i] == NULL) || (regChain[i]->pre != NULL)) {
	    continue;
	}
	if (newNetGr->node == NULL) {
	    nextNode = newNetGr->node = MALLOC(GNodeDef,1);
	} else {
	    nextNode->next = MALLOC(GNodeDef,1);
	    nextNode = nextNode->next;
	}
	nextNode->region = GID2Reg[i];
	nextNode->next = NULL;
	nextNode->nextEq = NULL;

	/* Construct equivalent region info */
	for (nextNodeChain = regChain[i]->next; nextNodeChain != NULL;
		nextNodeChain = nextNodeChain->next) {

	    if (nextNode->nextEq == NULL) {
		    nextEqNode = nextNode->nextEq = MALLOC(GNodeDef,1);
	    } else {
		nextEqNode->nextEq = MALLOC(GNodeDef,1);
		nextEqNode = nextEqNode->nextEq;
	    }
	    nextEqNode->region = GID2Reg[nextNodeChain->id - 1];
	    nextEqNode->next = NULL;
	    nextEqNode->nextEq = NULL;
	}
    }

    for (i = 0; i <= G_MAXROUTE - 1; ++i) {
	newNetGr->candRoute[i] = NULL;
	newNetGr->IPresult[i] = 0;
    }

    /* Maintain the net group information in net */
    /*-------------------------------------------*/
    nextNet->netGr = newNetGr;


}





/*
 *  makeRegionPin()
 *
 *  DESCRIPTION
 *  -----------
 *	Attach pin struct to region for calculating max column density.
 *		They are attached in nondecreasing order of offset.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	regionP	<input>	: pointer to region where pin exists.
 *	offset	<input>	: offset of the new pin to be inserted
 *	flag	<input>	: flag, if ON, represents min pin location of the net
 *			        otherwise, max pin location of the net
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	Creates pin struct and attach it to region w/ some useful infomation
 *		for max column density calculation.
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 *
 */


makeRegionPin(regionP,offset,flag)

GRegionDef	*regionP;
int		offset;
Boolean		flag;

{

GPinDef		*nextPin,*prePin;


/* Create a new pin struct and attach it to region */
nextPin = regionP->pin;
while (nextPin != NULL) {
    if (offset <= nextPin->offset) {
	break;
    }
    prePin = nextPin;
    nextPin = nextPin->next;
}
if (regionP->pin == nextPin) {
    regionP->pin = MALLOC(GPinDef,1);
    regionP->pin->next = nextPin;
    nextPin = regionP->pin;
} else {
    prePin->next = MALLOC(GPinDef,1);
    prePin->next->next = nextPin;
    nextPin = prePin->next;
}
nextPin->region = regionP;
nextPin->offset = offset;
nextPin->netStat = flag;
nextPin->mutExFlg = NO;
nextPin->equiv = NULL;


}





