
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/readFile.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  readFile.c
 *  **********
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains routines for reading data from files (test use)
 *
 *  functions contained in this file:
 *	readGraph()
 *	readNet()
 *	readGraphNet()
 *	readTile()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"



/*
 *  readGraph()
 *
 *  DESCRIPTION
 *  -----------
 *	Read an adjacency matrix from a file and constructs an adjacency list
 *	for a graph.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	w	: weight for an edge
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */

readGraph()

{

int	i, j;
int	w;
int	NoNode;

adjNodeType	*nextNode;
adjNodeType	*preNode;


graphFile = fopen("graph.in","r");
fscanf(graphFile,"%d\n",&NoNode);
for (i = 0; i <= NoNode - 1; ++i) {
    nextNode = adjList[i] = MALLOC(adjNodeType,1);
    for (j = 0; j <= NoNode - 1; ++j) {
	fscanf(graphFile,"%d",&w);
	if (w != 0) {
	    nextNode->id = j + 1;
	    nextNode->weight = w;
	    nextNode->next = MALLOC(adjNodeType,1);
	    preNode = nextNode;
	    nextNode = nextNode->next;
	}
    }
    FREE(nextNode);
    preNode->next = NULL;
}

printf("Printing the input graph data .....\n");
printAdjMatrix(adjList);
fclose(graphFile);

}




/*
 *  readNet()
 *
 *  DESCRIPTION
 *  -----------
 *	Read a net information from a file.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	pin	: node id corresponging to each pin of this net
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GNetPointer[]	: array of pointers to nets
 *
 */

readNet()


{

int	i, j, k;
int	pin;

GNetDef	*nextNet;
GPinDef	*nextPin;
char	line[MAXLINE];


netFile = fopen("net.in","r");
for (i = 0; i <= G_MAXNET - 1; ++i) {
    if (fgets(line,MAXLINE,netFile) == NULL) {
	break;
    }

    nextNet = GNetPointer[i] = MALLOC(GNetDef,1);
    if (i >= 1) {
	GNetPointer[i-1]->next = nextNet;
    }
    nextNet->critical = NO;
    nextNet->maxLength = -1;
    nextNet->trivial = NO;
    nextNet->routeNo = 0;
    nextNet->next = NULL;
    nextNet->nextGrMem = NULL;
    nextNet->netGr = NULL;
    nextNet->term = NULL;

    k = 0;
    for (j = 0; j <= G_MAXPIN_PER_NET - 1; ++j) {
	if (j == 0) {
	    nextPin = GNetPointer[i]->pin = MALLOC(GPinDef,1);
	} else {
	    nextPin->next = MALLOC(GPinDef,1);
	    nextPin = nextPin->next;
	}

/* Read pin info, skipping the spaces */
	for ( ; line[k] == ' '; ++k);
	sscanf(line + k,"%d",&pin);
	
/* this is a temporary treatment */
	nextPin->region = GID2Reg[pin - 1];
	nextPin->offset = 0;
	nextPin->next = NULL;
	nextPin->equiv = NULL;

/* Skip to the next space, or break if you see the carriage return */
	for ( ; (line[k] != ' ') && (line[k] != '\n'); ++k);
	if (line[k] == '\n') {
	    break;
	}
    }
}
GNet = GNetPointer[0];

fclose(netFile);

}




/*
 *  readGraphNet()
 *
 *  DESCRIPTION
 *  -----------
 *	Read a graph net information from a file.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	pin	: node id corresponging to each pin of this net
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	netPointer[]	: array of pointers to nets
 *
 */

readGraphNet()


{

int i, j;
int pin;

netType		*nextNet;
netNodeType	*nextNode;


netFile = fopen("net.in","r");
for (i = 0; i <= 0; ++i) {
    nextNet = netPointer[i] = MALLOC(netType,1);
    nextNet->critical = NO;
    nextNet->maxLength = -1;
    nextNet->trivial = NO;
    nextNet->next = NULL;

    for (j = 0; j <= 2; ++j) {
	if (j == 0) {
	    nextNode = netPointer[i]->node = MALLOC(netNodeType,1);
	} else {
	    nextNode->next = MALLOC(netNodeType,1);
	    nextNode = nextNode->next;
	}
	fscanf(netFile,"%d",&pin);
/* this is a temporary treatment */
	nextNode->id = pin;
	nextNode->next = NULL;
	nextNode->equiv = NULL;
    }
}

fclose(netFile);

}




/*
 *  readTile()
 *
 *  DESCRIPTION
 *  -----------
 *	Read tile info from a file and constructs an GRegionDef.
 *
 *  RETURNS
 *  -------
 *	None
 *
 *  Arguments:
 *  ---------
 *	None
 *
 *  Local variables:
 *  ---------------
 *	w	: weight for an edge
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *	None
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GRegList	: pointer to first GRegionDef
 *
 */

readTile()

{

int		i, j;
int		NoRegion;
int		id, adjID, capacity;
int		xdim, ydim, xloc, yloc;
int		factor = 1000;
char		type[20];
GRegionDef	*nextRegion;


tileFile = fopen("tile.in","r");
fscanf(tileFile,"%d\n",&NoRegion);
for (i = 0; i <= NoRegion - 1; ++i) {

/* Allocate memory for the tile */

    if (i == 0) {
	GRegList = nextRegion = MALLOC(GRegionDef,1);
    } else {
	nextRegion->next = MALLOC(GRegionDef,1);
	nextRegion = nextRegion->next;
    }
    fscanf(tileFile,"%d",&id);
    nextRegion->id = id;
    nextRegion->estUsage = 0;
    nextRegion->pin = NULL;
    nextRegion->next = NULL;
    GID2Reg[id - 1] = nextRegion;

/* According to its type, fill in items */
/*--------------------------------------*/
    fscanf(tileFile,"%s",type);
    if (!strcmp(type,"G_HOR_BN")) {
	nextRegion->type = G_HOR_BN;
	nextRegion->expandableFlg = YES;
	nextRegion->orient = G_HOR;
    } else if (!strcmp(type,"G_VER_BN")) {
	nextRegion->type = G_VER_BN;
	nextRegion->expandableFlg = YES;
	nextRegion->orient = G_VER;
    } else if (!strcmp(type,"G_DOM_XSEC")) {
	nextRegion->type = G_DOM_XSEC;
	nextRegion->expandableFlg = YES;
	nextRegion->orient = NO;
    } else if (!strcmp(type,"G_INST")) {
	nextRegion->type = G_INST;
	nextRegion->expandableFlg = NO;
	nextRegion->orient = NO;
    }

/* Read other info */
/*-----------------*/
    fscanf(tileFile,"%s",buf);
    fscanf(tileFile,"%d",&capacity);
    nextRegion->capacity[G_HOR] = capacity * factor;
    nextRegion->capacity[G_VER] = capacity * factor;

    fscanf(tileFile,"%s",buf);
    fscanf(tileFile,"%d",&xdim);
    nextRegion->dimension[G_HOR] = xdim * factor;
    fscanf(tileFile,"%s",buf);
    fscanf(tileFile,"%d",&ydim);
    nextRegion->dimension[G_VER] = ydim * factor;

    fscanf(tileFile,"%s",buf);
    fscanf(tileFile,"%d",&xloc);
    nextRegion->centerLocX = xloc * factor;
    fscanf(tileFile,"%d\n",&yloc);
    nextRegion->centerLocY = yloc * factor;
}

/* Read adjacency relations */
/*--------------------------*/
for (i = 0; i <= NoRegion - 1; ++i) {
    fscanf(tileFile,"%d",&id);
    fscanf(tileFile,"%s",buf);
    for (j = 0; j <= G_NDIR - 1; ++j) {
	fscanf(tileFile,"%d",&adjID);
	GID2Reg[id - 1]->adjReg[j] = GID2Reg[adjID - 1];
    }
}

fclose(tileFile);

}



