 
/*
#ifndef lint
static char rcsid[] = "$Header: /home/cvs/octtools/src/Layout/mercury/util.c,v 1.1.1.1 2002/12/31 15:50:50 saj Exp $";
#endif
*/

/*
 *  util.c
 *  ******
 *
 *  Copyright (C) 1988  Yoshihito Nishizaki
 *	University of California, Berkeley
 *
 *  Author:			Advising professor:
 *	Yoshihito Nishizaki	    Alberto Sangiovanni-Vincentelli
 *
 *  --------------------------------------------------------------------------
 *
 *  This file contains various utility routines
 *
 *  functions contained in this file:
 *	askPathDir()
 *	askTrivialRoute()
 *	getBotNeck()
 *	getNextRegion()
 *	getNeighbor()
 *
 *
 *  Import descriptions:
 *	initGlb.h	: macros for mercury global router
 *
 */

#include <stdio.h>
#include <strings.h>
#include "initGlb.h"





/*
 *  askPathDir()
 *
 *  DESCRIPTION
 *  -----------
 *	Determine the direction of the given path in the given region
 *
 *  RETURNS
 *  -------
 *	G_LEFT		when the path goes to left in the region
 *	G_RIGHT		when the path goes to right in the region
 *	G_UP		when the path goes up in the region
 *	G_DOWN		when the path goes down in the region
 *	-1		Otherwise
 *
 *  Arguments:
 *  ---------
 *	node1	<input>	: base node
 *	node2	<input>	: the node this path is going towards
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


askPathDir(node1,node2)


int		node1,node2;

{

int		i;
GRegionDef	*neighbor;
GRegionDef	*baseReg;



for (i = 0; i <= G_NDIR - 1; ++i) {

    for (baseReg = GID2Reg[node1 - 1];
		(neighbor = getNeighbor(baseReg,i)) != NULL; baseReg = NULL) {

	if (neighbor->id == node2) {
	    return(i);
	}
    }
}

return(-1);


}






/*
 *  askTrivialRoute()
 *
 *  DESCRIPTION
 *  -----------
 *	Determine if the given net topology has only one graph node or not
 *
 *  RETURNS
 *  -------
 *	YES		when this net topology is trivial
 *	NO		when this net topology is not trivial
 *
 *  Arguments:
 *  ---------
 *	v	<input>	: starting node
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


askTrivialRoute(netGrP)


GNetGrDef	*netGrP;

{

if (netGrP->NoRegion == 1) {
    return(YES);
} else {
    return(NO);
}


}





/*
 *  getBotNeck()
 *
 *  DESCRIPTION
 *  -----------
 *	Get a bottleneck region corresponding to the edge between node1 & node2
 *
 *  RETURNS
 *  -------
 *	region		pointer to a region requested
 *
 *  Arguments:
 *  ---------
 *	node1	<input>	: one end of the given edge
 *	node2	<input>	: the other end of the given edge
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GID2Reg[i]	: pointers to tile node for region id i
 *
 */


GRegionDef *getBotNeck(node1,node2)


int		node1,node2;


{

int		i;
int		id1, id2;
GRegionDef	*nextReg;
GRegionDef	*reg1;
GRegionDef	*reg2;


reg1 = GID2Reg[node1 - 1];
reg2 = GID2Reg[node2 - 1];

/* If node1 is a bottleneck region, return it */
if ((reg1->type == G_HOR_BN) || (reg1->type == G_VER_BN)) {
    return(reg1);
}

/* If node2 is a bottleneck region, return it */
if ((reg2->type == G_HOR_BN) || (reg2->type == G_VER_BN)) {
    return(reg2);
}

/* Otherwise search a region which makes a straight line with node1 and node2 */
/*----------------------------------------------------------------------------*/
for (i = 0; i <= G_NDIR - 1; ++i) {
    if (reg1->adjReg[i] == NULL) {
	continue;
    }
    if (reg1->adjReg[i]->adjReg[i] == NULL) {
	continue;
    }
    if (reg1->adjReg[i]->adjReg[i]->id == node2) {
	if ((reg1->adjReg[i]->type == G_HOR_BN) ||
		(reg1->adjReg[i]->type == G_VER_BN)) {
	    return(reg1->adjReg[i]);
	}
    }
}

/* Take a different approach if we cannot find them by the above */
/*---------------------------------------------------------------*/
for (nextReg = GRegList; nextReg != NULL; nextReg = nextReg->next) {
    if ((nextReg->type != G_HOR_BN) && (nextReg->type != G_VER_BN)) {
	continue;
    }
    id1 = (nextReg->adjReg[G_UP] == NULL) ? 0 : nextReg->adjReg[G_UP]->id;
    id2 = (nextReg->adjReg[G_DOWN] == NULL) ? 0 : nextReg->adjReg[G_DOWN]->id;
    if (((node1 == id1) && (node2 == id2)) ||
		((node2 == id1) && (node1 == id2))) {
	return(nextReg);
    }
    id1 = (nextReg->adjReg[G_LEFT] == NULL) ? 0 : nextReg->adjReg[G_LEFT]->id;
    id2 = (nextReg->adjReg[G_RIGHT] == NULL) ? 0 : nextReg->adjReg[G_RIGHT]->id;
    if (((node1 == id1) && (node2 == id2)) ||
		((node2 == id1) && (node1 == id2))) {
	return(nextReg);
    }
}


return(NULL);

    
}




/*
 *  getNextRegion()
 *
 *  DESCRIPTION
 *  -----------
 *	Get the next region in the given path, including bottle neck regions
 *		between two G_DOM_XSEC nodes
 *	This program memorizes the preNode and midRegion as static variables
 *
 *  RETURNS
 *  -------
 *	regionP		pointer to a region requested
 *
 *  Arguments:
 *  ---------
 *	curNode	<input>	: pointer to the current node
 *
 *  Local variables:
 *  ---------------
 *    static variables
 *	preNode		: the most recently threaded path node before this call
 *	midRegion	: bottle neck region between preNode & preNode->next
 *			  used to pick up BN that is not a node in the path
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GID2Reg[i]	: pointers to tile node for region id i
 *
 */


GRegionDef *getNextRegion(curNode)

nodeGroupType	*curNode;

{


static nodeGroupType	*preNode;
static GRegionDef	*midRegion;


/* If the starting node is specified from caller, return it as it is */
/*-------------------------------------------------------------------*/
if (curNode != NULL) {
    preNode = curNode;
    midRegion = NULL;
    return(GID2Reg[curNode->id - 1]);
}

/* If the previous node is the last node in path, return NULL */
/*------------------------------------------------------------*/
if (preNode->next == NULL) {
    return(NULL);
}

/* If the previous region returned is the middle bottle neck region
 *	between two G_DOM_XSEC regions, return the next node in path */
/*-------------------------------------------------------------------*/
if (midRegion != NULL) {
    midRegion = NULL;
    preNode = preNode->next;
    return(GID2Reg[preNode->id - 1]);
}

/* If preNode or preNode->next is not G_DOM_XSEC, return the next node */
/*---------------------------------------------------------------------*/
if ((GID2Reg[preNode->id - 1]->type != G_DOM_XSEC)
		|| (GID2Reg[preNode->next->id - 1]->type != G_DOM_XSEC)) {
    preNode = preNode->next;
    return(GID2Reg[preNode->id - 1]);
}

/* Otherwise, get the middle bottle neck region between preNode and
 *	preNode->next */
/*---------------------------------------------------------------*/
midRegion = getBotNeck(preNode->id,preNode->next->id);
return(midRegion);


}



/*
 *  getNeighbor()
 *
 *  DESCRIPTION
 *  -----------
 *	Generate neighbor regions of the given region sequentially.
 *	This program memorizes the preNode and midRegion as static variables
 *
 *  RETURNS
 *  -------
 *	regionP		pointer to a region generated
 *
 *  Arguments:
 *  ---------
 *	baseReg	<input>	: pointer to the reference region
 *			  Should be set only for the 1st call, otherwise NULL.
 *
 *  Local variables:
 *  ---------------
 *    static variables
 *	saveBase	: base region for which to obtain neighbors
 *	preReg		: region returned at the previous call 
 *	moveDir		: orthogonal direction to scan in order to get every
 *			  neighbor in the given direction
 *	foundLastReg	: flag, if ON, found the last neighbor region
 *
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *	GID2Reg[i]	: pointers to tile node for region id i
 *
 */


GRegionDef *getNeighbor(baseReg,direction)

GRegionDef	*baseReg;
int		direction;

{


    static GRegionDef	*saveBase;
    static GRegionDef	*preReg;
    static int		moveDir;
    static int		foundLastReg;
    GRegionDef		*neighbor;


    /* If the base region is specified from caller, choose the first neighbor */
    /*------------------------------------------------------------------------*/
    if (baseReg != NULL) {
	foundLastReg = OFF;
	saveBase = baseReg;
	preReg = baseReg->adjReg[direction];

	/* Choose the orthogonal search direction for the "direction" requested */
	/*----------------------------------------------------------------------*/
	switch (direction) {
	case G_UP :
	    moveDir = G_LEFT;
	    break;
	case G_LEFT :
	    moveDir = G_UP;
	    break;
	case G_DOWN :
	    moveDir = G_RIGHT;
	    break;
	case G_RIGHT :
	    moveDir = G_DOWN;
	    break;
	    default :
	    break;
	}
    } else {
	preReg = preReg->adjReg[moveDir];
    }

    /* #define GET_NEIGHBOR_BY_LOC */

#ifndef GET_NEIGHBOR_BY_LOC
    {
	int		revDir;

	switch (direction) {
	case G_UP:	revDir = G_DOWN; break;
	case G_DOWN:	revDir = G_UP; break;
	case G_LEFT:	revDir = G_RIGHT; break;
	case G_RIGHT:	revDir = G_LEFT; break;
	}

	if (baseReg || (preReg && preReg->adjReg[revDir]==saveBase))
	  return(preReg);
	else
	  return(NULL);
    }

#else

    /* If the previous region was the last neighbor region, return NULL */
    /*------------------------------------------------------------------*/
    if ((preReg == NULL) || (foundLastReg == ON)) {
	return(NULL);
    }

    /* Check if the region that is going to be returned is the last neighbor */
    /*-----------------------------------------------------------------------*/

    neighbor = preReg;

    switch (direction) {

    case G_UP :
	if (neighbor->adjReg[moveDir] == NULL) {
	    foundLastReg = ON;
	    break;
	}

	if (neighbor->centerLocX - (neighbor->dimension[G_HOR] / 2) <=
	    saveBase->centerLocX - (saveBase->dimension[G_HOR] / 2)) {

	    /* The next check is important not to overlook 0 width bottlenecks */

	    if (neighbor->adjReg[moveDir]->centerLocX <
		saveBase->centerLocX - (saveBase->dimension[G_HOR] / 2)) {
		foundLastReg = ON;
	    }
	}
	break;

    case G_LEFT :
	if (neighbor->adjReg[moveDir] == NULL) {
	    foundLastReg = ON;
	    break;
	}

	if (neighbor->centerLocY + (neighbor->dimension[G_VER] / 2) >=
	    saveBase->centerLocY + (saveBase->dimension[G_VER] / 2)) {

	    if (neighbor->adjReg[moveDir]->centerLocY >
		saveBase->centerLocY + (saveBase->dimension[G_VER] / 2)) {
		foundLastReg = ON;
	    }
	}
	break;

    case G_DOWN :
	if (neighbor->adjReg[moveDir] == NULL) {
	    foundLastReg = ON;
	    break;
	}

	if (neighbor->centerLocX + (neighbor->dimension[G_HOR] / 2) >=
	    saveBase->centerLocX + (saveBase->dimension[G_HOR] / 2)) {

	    if (neighbor->adjReg[moveDir]->centerLocX >
		saveBase->centerLocX + (saveBase->dimension[G_HOR] / 2)) {
		foundLastReg = ON;
	    }
	}
	break;

    case G_RIGHT :
	if (neighbor->adjReg[moveDir] == NULL) {
	    foundLastReg = ON;
	    break;
	}

	if (neighbor->centerLocY - (neighbor->dimension[G_VER] / 2) <=
	    saveBase->centerLocY - (saveBase->dimension[G_VER] / 2)) {

	    if (neighbor->adjReg[moveDir]->centerLocY <
		saveBase->centerLocY - (saveBase->dimension[G_VER] / 2)) {
		foundLastReg = ON;
	    }
	}
	break;

	default :
	break;
    }

    return(preReg);

#endif


}






/*
 *  askPlusJunction()
 *
 *  DESCRIPTION
 *  -----------
 *	Check if the given G_DOM_XSEC tile is a + junction.
 *
 *  RETURNS
 *  -------
 *	answer		YES/NO
 *
 *  Arguments:
 *  ---------
 *	region	<input>	: pointer to the region
 *
 *  Local variables:
 *  ---------------
 *
 *  Possible errors:
 *  ---------------
 *
 *  SIDE-EFFECTS
 *  ------------
 *
 *  GLOBAL VARIABLES
 *  ----------------
 *
 */


int askPlusJunction(region)

GRegionDef	*region;

{


int		i, direction, num_bottleneck;
GRegionDef	*baseReg, *nextAdjReg;


for (num_bottleneck = 0, i = 0; i < G_NDIR; ++i) {
    for (baseReg = region, direction = i;
	(nextAdjReg = getNeighbor(baseReg,direction)) != NULL; baseReg = NULL) {
	if (nextAdjReg->type == G_HOR_BN || nextAdjReg->type == G_VER_BN) {
	    ++num_bottleneck;
	    break;
	}
    }
}
if (num_bottleneck == G_NDIR) {
    return(YES);
} else {
    return(NO);
}


}





