/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSHgConvNet.c

Convert oct nets def to GNet definition.

CDSHgConvNet		Convert net data representation
CDSHgPrintNet		Print net data.
CDSHgNewNet		Allocate & intialize new net.
CDSHgNewPin		Allocate & intialize new pin.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSHgConvNet.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:49  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  90/02/07  17:03:19  octtools
 * Checkin for 3.5 distribution
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CDSInc.h"
#include "mercury.h"

#define HG_DEBUG

/* Create interface for hierarchical global router
----------------------------------------------------------------------------*/
CDSHgConvtNet(octMain)

octObject	*octMain;
{
octObject	octInstMaster, octInst, octATerm, octNet, octProp;
octGenerator	genATerm;
String		instName;
int		numNet = 0;
int		pb, b0, b1, o, c;
int		x, y, bx, bx2, by, by2, rx, ry;
int		minDist, dist;
int		pinOffset;
GNetDef		*prevNet, *net, *CDSHgNewNet();
GPinDef		*pin, *CDSHgNewPin();
GRegionDef	*reg, *pReg, *instReg;
CDSObjDef	*obj;
CDSLocDef	*bdy;
int		io;
int		restrictCornSide;
Boolean		ioPadFlag;
int		CDSGetCornSide();

MI_MOD_START("CDSHgConvtNet");

GNet = NULL;
prevNet = NULL;

for (o=0; o<CDSHead.numObj-1; o++) {
    obj = &CDSObj[o];
    bdy = obj->bdy;
    instName = (String)obj->dbRef;
    instReg  = (GRegionDef *)obj->ref;
    OH_ASSERT(ohGetByInstName(octMain, &octInst, instName));
    OH_ASSERT(ohOpenMaster(&octInstMaster, &octInst, "interface", "r"));

    MSG(1, ("On inst %s reg %d (%d %d) (%d %d)\n",
	ohGetName(&octInst), instReg->id,
	instReg->centerLocX - instReg->dimension[G_HOR]/2,
	instReg->centerLocY - instReg->dimension[G_VER]/2,
	instReg->centerLocX + instReg->dimension[G_HOR]/2,
	instReg->centerLocY + instReg->dimension[G_VER]/2));

    ioPadFlag = FALSE;
    if (obj->ioPadFlag) {
	for (io=0; io<4; io++) {
	    if (CDSIo[io].numObj==0) continue;
	    if (CDSIo[io].objNum[0]==o) break;
	}
	ioPadFlag = TRUE;
	switch(io) {
	case CDS_IO_BOTTOM:  restrictCornSide = CDS_CORN_UR; break;
	case CDS_IO_TOP   :  restrictCornSide = CDS_CORN_LL; break;
	case CDS_IO_LEFT  :  restrictCornSide = CDS_CORN_LR; break;
	case CDS_IO_RIGHT :  restrictCornSide = CDS_CORN_UL; break;
	default: MSG(-1, ("Skipping io instance %s\n", instName));
		 ioPadFlag = FALSE;
	}
	if (ioPadFlag)
	    MSG(-1, ("Restricting aterms of inst %s to side %s\n",
		instName, CDS_CORN_TOSTR(restrictCornSide)));
    }

    /* look at all aterms of inst */
    octInitGenContents(&octInst, OCT_TERM_MASK, &genATerm);
    while (octGenerate(&genATerm, &octATerm)==OCT_OK) {

	/* Get net */
	if (octGenFirstContainer(&octATerm, OCT_NET_MASK, &octNet)==OCT_GEN_DONE) {
	    MSG(1, ("Skipping aterm %s\n", ohGetName(&octATerm)));
	    continue;
	}
	for (net=GNet; net; net=net->next)
	    if ((octId)net->dbRef==octNet.objectId) break;
	if (!net) {
	    net = CDSHgNewNet(&numNet, &prevNet, (char *)octNet.objectId,
		    ohGetName(&octNet));
	    if (ohGetByPropName(&octNet,&octProp,"NET_CRITICALITY")==OCT_OK) {
		if (octProp.contents.prop.type!=OCT_INTEGER) {
		    fprintf(stderr,
			"ERROR %s property should be of type integer\n",
			octProp.contents.prop.name);
		}
		else
		    net->criticality = octProp.contents.prop.value.integer;
	    }
	    if (ohGetByPropName(&octNet,&octProp,"NET_MAX_LENGTH")==OCT_OK) {
		if (octProp.contents.prop.type!=OCT_INTEGER) {
		    fprintf(stderr,
			"ERROR %s property should be of type integer\n",
			octProp.contents.prop.name);
		}
		else
		    net->maxLength = octProp.contents.prop.value.integer;
	    }
	}
	pin = CDSHgNewPin(net, (char *)octATerm.objectId, ohGetName(&octATerm));

	/* Get pin location (x, y) */
	CDSHgGetTermLoc(&octInstMaster, &octInst, &octATerm, &x, &y);
	MSG(1, ("Loc of pin %s of net %s = (%d %d)\n",
	    ohGetName(&octATerm), ohGetName(&octNet), x, y));

	/* Find out which boundary (pb) the pin projects (is closest) to */
	pb = -1;
	minDist = MAXINT;
	for (b0=0; b0<obj->numBdy; b0++) {
					/* Skip invalid boundaries for ios */
	    if (ioPadFlag &&
		restrictCornSide!=CDSGetCornSide(obj, b0)) continue;

	    b1 = (b0+1)%obj->numBdy;
	    bx = bdy[b0][CDS_X];
	    by = bdy[b0][CDS_Y];
	    if (bx==bdy[b1][CDS_X]) {	/* Vertical boundary */
		by2 = bdy[b1][CDS_Y];
		if (y<MIN(by, by2) || y>MAX(by, by2)) continue;
		dist = ABS(x-bx);
	    }
	    else {			/* Horizontal boundary */
		bx2 = bdy[b1][CDS_X];
		if (x<MIN(bx, bx2) || x>MAX(bx, bx2)) continue;
		dist = ABS(y-by);
	    }
	    if (dist<minDist) {
		pb = b0;
		minDist = dist;
	    }
	}

	if (pb<0) {
	    pb = 0;
	    fprintf(stdout,
		"ERROR Cannot find boundary to project pin %s of inst %s\n",
		ohGetName(&octATerm), ohGetName(&octInst));
	    fprintf(stdout,"Pin location computed to be (%d %d)\n", x, y);
	    fprintf(stdout,
		"Assumed to be projected to bdy segment: (%d %d) (%d %d)\n",
		bdy[pb][CDS_X], bdy[pb][CDS_Y],
		bdy[(pb+1)%obj->numBdy][CDS_X],
		bdy[(pb+1)%obj->numBdy][CDS_Y]);
	}

	/* Map boundary index (pb) to side (c) */
	c = CDSGetCornSide(obj, pb);

	MSG(1, ("On side %s of inst %s\n",
	    CDS_CORN_TOSTR(c), ohGetName(&octInst)));

	/* Map side (c) to region (reg) and determine pin offset */
	pinOffset = -1;			/* pin offset in region reg */
					/*     from bottom or left side */
	switch (c) {
	case CDS_CORN_LL:		/* Bottom side */
	    for (reg=instReg->adjReg[G_DOWN];
		 reg && (reg->adjReg[G_UP]==instReg ||
		         G_REGBOX_UR_X(reg)<=G_REGBOX_UR_X(instReg));
		 reg=reg->adjReg[G_RIGHT]) {
		 rx = G_REGBOX_UR_X(reg);
		 if (x<=rx) {
		     pinOffset = MAX(0, x - rx + reg->dimension[G_HOR]);
		     break;
		 }
		 pReg = reg;
	    }
	    if (pinOffset== -1) {
		MSG(-1, ("reg could not be found\n"));
		reg = pReg;
		pinOffset = reg->dimension[G_HOR];
	    }
	    break;
	case CDS_CORN_LR:		/* Right side */
	    for (reg=instReg->adjReg[G_RIGHT];
		 reg && reg->adjReg[G_LEFT]==instReg;
		 reg=reg->adjReg[G_DOWN]) {
		 ry = G_REGBOX_LL_Y(reg);
		 if (y>=ry) {
		     pinOffset = MIN(reg->dimension[G_VER], y - ry);
		     break;
		 }
		 pReg = reg;
	    }
	    if (pinOffset== -1) {
		MSG(-1, ("reg could not be found\n"));
		reg = pReg;
		pinOffset = 0;
	    }
	    break;
	case CDS_CORN_UR:		/* Top side */
	    for (reg=instReg->adjReg[G_UP];
		 reg && reg->adjReg[G_DOWN]==instReg;
		 reg=reg->adjReg[G_LEFT]) {
		 rx = G_REGBOX_LL_X(reg);
		 if (x>=rx) {
		     pinOffset = MIN(reg->dimension[G_HOR], x - rx);
		     break;
		 }
		 pReg = reg;
	    }
	    if (pinOffset== -1) {
		MSG(-1, ("reg could not be found\n"));
		reg = pReg;
		pinOffset = 0;
	    }
	    break;
	default:			/* Left side */
	    for (reg=instReg->adjReg[G_LEFT];
		 reg && reg->adjReg[G_RIGHT]==instReg;
		 reg=reg->adjReg[G_UP]) {
		 ry = G_REGBOX_UR_Y(reg);
		 if (y<=ry) {
		     pinOffset = MAX(0, y - ry + reg->dimension[G_VER]);
		     break;
		 }
		 pReg = reg;
	    }
	    if (pinOffset== -1) {
		MSG(-1, ("reg could not be found\n"));
		reg = pReg;
		pinOffset = reg->dimension[G_VER];
	    }
	    break;
	}
	MSG(2, ("pin offset = %d\n", pinOffset));

	pin->region = reg;
	pin->offset = pinOffset;
    }

    OH_ASSERT(octCloseFacet(&octInstMaster));
}

MI_MOD_END();
} /* ............................. End of procedure CDSHgConvtNet */
/* Get location  of aterm
----------------------------------------------------------------------------*/
CDSHgGetTermLoc(octInstMaster, octInst, octATerm, x, y)

octObject	*octInstMaster, *octInst;
octObject	*octATerm;
int		*x, *y;
{
octObject	octFTerm;
octObject	octGeom, octLayer;
octGenerator	genGeom;
int		tx, ty;
Boolean		setPinLocFlag = FALSE;

MI_MOD_START("CDSHgGetTermLoc");

OH_ASSERT(ohGetByTermName(octInstMaster, &octFTerm, ohGetName(octATerm)));

octInitGenContents(&octFTerm, OCT_GEO_MASK, &genGeom);
while (octGenerate(&genGeom, &octGeom)==OCT_OK) {
    switch (octGeom.type) {
    case OCT_BOX:
	tx = (octGeom.contents.box.lowerLeft.x+
	      octGeom.contents.box.upperRight.x)/2;
	ty = (octGeom.contents.box.lowerLeft.y+
	      octGeom.contents.box.upperRight.y)/2;
	break;
    default: fprintf(stderr,
	"Cannot process pins of geometry type = %d (not an OCT_BOX)\n",
	octGeom.type);
    }
    tr_oct_transform(&octInst->contents.instance.transform, &tx, &ty);

    OH_ASSERT(octGenFirstContainer(&octGeom, OCT_LAYER_MASK, &octLayer));
    if (!strcmp(ohGetName(&octLayer),"POLY") ||
	!strcmp(ohGetName(&octLayer),"MET1") ||
	!strcmp(ohGetName(&octLayer),"MET2")) {
	setPinLocFlag = TRUE;
	*x = tx;
	*y = ty;
	break;
    }
}

if (!setPinLocFlag) {
    fprintf(stderr,
	"ERROR fTerm %s of inst %s has no geometry on POLY, MET1 or MET2\n",
	ohGetName(octATerm), ohGetName(octInst));
    fprintf(stderr,"You must put a contact to one of these layers\n");
    exit(1);
}

MI_MOD_END();
} /* ............................. End of procedure CDSHgGetTermLoc */
/* Print nets in Nishizaki format
----------------------------------------------------------------------------*/
CDSHgPrintNet(netFileName)

String		netFileName;
{
FILE		*f;
GNetDef		*n;
GPinDef		*p;
int		nIdx, pIdx;
Boolean		myNetFormatFlag = TRUE;

MI_MOD_START("CDSHgPrintNet");

f = MIfopen(netFileName,"w","Nishizaki nets");

nIdx = 0;
for (n=GNet; n; n=n->next) {
    if (myNetFormatFlag) {
	nIdx++;
	fprintf(f,"net %d %d %d %d\n",
	    nIdx, n->critical, n->maxLength, n->trivial);
	pIdx = 0;
	for (p=n->pin; p; p=p->next) {
	    pIdx++;
	    fprintf(f,"pin %d region %d offset %d\n",
		pIdx, p->region->id, p->offset);
	}
    }
    else {
	for (p=n->pin; p; p=p->next)
	    fprintf(f,"%d ", p->region->id);
	if (n->pin)
	    fprintf(f,"\n");
    }
}

fclose(f);

MI_MOD_END();
} /* ............................. End of procedure CDSHgPrintNet */
/* Allocate new net + do linked list stuff.
----------------------------------------------------------------------------*/
GNetDef *
CDSHgNewNet(numNet, prevNet, dbRef, name)

int		*numNet;
GNetDef		**prevNet;
char		*dbRef;
String		name;
{
GNetDef		*n;

MI_MOD_START("CDSHgNewNet");

n = ALLOC_TYPE(GNetDef);
if (!GNet)				/* This should be at ll corner */
    GNet = n;
if (*prevNet)				/* Linked list */
    (*prevNet)->next = n;
*prevNet = n;

n->dbRef = dbRef;
n->next = NULL;

n->criticality = 0;			/* 0 Means not critical */
n->critical = FALSE;
n->maxLength = 0;			/* Means this net is not critical */
n->trivial = FALSE;
n->pin = NULL;
n->nextGrMem = NULL;			/* Grouping of nets */
n->netGr = NULL;
n->name = STR_COPY(name);

(*numNet)++;

MI_MOD_END();
return(n);
} /* ............................. End of procedure CDSHgNewNet */
/* Allocate new net + do linked list stuff.
----------------------------------------------------------------------------*/
GPinDef *
CDSHgNewPin(net, dbRef, name)

GNetDef		*net;
char		*dbRef;
String		name;
{
GPinDef		*p;

MI_MOD_START("CDSHgNewPin");

p = ALLOC_TYPE(GPinDef);
if (net->pin)				/* Linked list */
    p->next = net->pin;
else
    p->next = NULL;
net->pin = p;

p->dbRef = dbRef;

p->region = NULL;			/* Defaults: */
p->offset = 0;
p->netStat = 0;
p->mutExFlg = FALSE;
p->equiv = NULL;
p->name = STR_COPY(name);

MI_MOD_END();
return(p);
} /* ............................. End of procedure CDSHgNewPin */
/* Compute corner side of boundary index
----------------------------------------------------------------------------*/
int
CDSGetCornSide(obj, pb)

CDSObjDef	*obj;
int		pb;
{
int		c;
int		b0, b1;

for (c=0; c<obj->numCorn; c++) {
    b0 = obj->corn[c].bdyIdx;
    b1 = obj->corn[(c+1)%obj->numCorn].bdyIdx;
    if (b1>b0) {			/* Normal */
	if (pb>=b0 && pb<b1) break;
    }
    else {				/* Boundary index wraps around */
	if (pb>=b0 || pb<b1) break;
    }
}

return(c);
} /* ............................. End of procedure CDSGetCornSide */
/* Compute corner side of boundary index
----------------------------------------------------------------------------*/
CDSHgSetIoObjNum()
{
int		io;
CDSTreeDef	*t;

DASSERT(!MIStackPtr);
for (io=0; io<4; io++) {		/* Check if each io side has stuff */
    if (!CDSIo[io].sliceFoundFlg) continue; /* to merge */

    t = CDSIo[io].sideTree;		/* Scan tree for all leaf objects */
    MI_STACK_PUSH(t);
    while (MIStackPtr) {
	MI_STACK_POP(CDSTreeDef*, t);
	if (!t) continue;
	if (t->type!=CDS_TREE_ELEMENT) {
	    MI_STACK_PUSH(t->child[CDS_LO]);
	    MI_STACK_PUSH(t->child[CDS_HI]);
	    continue;
	}
	if (t->objNum<0) continue;
					/* Record leaf objects in temp array */
	CDSIo[io].numObj = 1;
	CDSIo[io].objNum = ALLOC_ARRAY(CDSIo[io].numObj, short);
	CDSIo[io].objNum[0] = t->objNum;
    }
}
} /* ............................. End of procedure CDSHgSetIoObjNum */
