/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CDSHgConvReg.c

Convert CDSGNode def to GRegion definition.

CDSHgConvGraph		Convert graph data representation.
CDSHgPrintReg		Print regions data.
CDSHgNewRegion		Allocate & intialize new region.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSHgConvReg.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:49  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  90/02/07  17:03:20  octtools
 * Checkin for 3.5 distribution
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "miutil.h"
#include "CDSInc.h"
#include "mercury.h"

/* #define HG_DEBUG */

/* Create interface for hierarchical global router
----------------------------------------------------------------------------*/
CDSHgConvtGraph(headNode)

CDSGNodeDef	*headNode;
{
int		numTNode = 0;
int		numSkipTNode = 0;
CDSGNodeDef	**tNode, **skipTNode;
CDSGNodeDef	*gn, *gn0, *gn2;
int		k, kp, t, o, rpd, c, c2, d, d2, dr;
int		stkRef;
int		numRegion;
GRegionDef	*r, *r0, *r2, *r3, *pr;
GRegionDef	*CDSHgNewRegion();
CDSBoxDef	box;
int		maxId;
Boolean		CDSHgSkipExtNode();
int		io;
CDSGNodeDef	*cornGNode[4];

MI_MOD_START("CDSHgConvtGraph");

tNode     = ALLOC_ARRAY(CDSHead.numGNode, CDSGNodeDef*);
skipTNode = ALLOC_ARRAY(CDSHead.numGNode, CDSGNodeDef*);

stkRef = MIStackPtr;
MI_STACK_PUSH(headNode);
if (headNode->type!='e' ||
    !CDSHgSkipExtNode(headNode))
    tNode[numTNode++] = headNode;
else {
    MSG(-1, ("Skipping node %d\n", headNode->id));
    headNode->ref = NULL;
    skipTNode[numSkipTNode++] = headNode;
}

while (stkRef!=MIStackPtr) {		/* Make list of all nodes except */
					/*   "external nodes" */
    MI_STACK_POP(CDSGNodeDef*, gn);
    for (k=0; k<4; k++) {
	gn2 = gn->bch[k].node;
	if (!gn2) continue;

	for (t=0; t<numTNode; t++)
	    if (gn2==tNode[t]) break;
	if (t<numTNode) continue;

	for (t=0; t<numSkipTNode; t++)
	    if (gn2==skipTNode[t]) break;
	if (t<numSkipTNode) continue;

	MI_STACK_PUSH(gn2);
	gn2->ref = NULL;

	if (gn2->type=='e' &&		/* Skip "external nodes" that are */
	    CDSHgSkipExtNode(gn2)) {	/*   adjacent to ios */
	    MSG(-1, ("Skipping node %d\n", gn2->id));
	    ASSERT(numSkipTNode<CDSHead.numGNode);
	    skipTNode[numSkipTNode++] = gn2;
	    continue;
	}

	ASSERT(numTNode<CDSHead.numGNode);
	tNode[numTNode++] = gn2;
    }
}

FREE_ARRAY(CDSHead.numGNode, skipTNode, CDSGNodeDef*);
					/* Redefine "external" node to */
					/*    include ones that border ios */
for (t=0; t<numTNode; t++) {
    gn = tNode[t];
    for (k=0; k<4; k++) {
	gn2 = gn->bch[k].node;
	if (!gn2) continue;
	o = gn->bch[k].objNum;		
	if (o>=0 && CDSObj[o].ioPadFlag) {
	    gn ->type = 'e';
	    gn2->type = 'e';
	}
    }
}

numRegion = 0;
GRegList = NULL;
pr = NULL;
for (t=0; t<numTNode; t++) {		/* Nodes become XSECtions */
    gn = tNode[t];

    r = CDSHgNewRegion(&numRegion, &pr);
    r->type = G_DOM_XSEC;
    r->externFlg = (gn->type=='e');

    gn->ref = (RefAny)r;
}

for (k=0; k<4; k++)			/* Find the corner nodes */
    cornGNode[k] = NULL;

for (t=0; t<numTNode; t++) {
    gn = tNode[t];
    if (gn->type!='e') continue;

    if (!gn->bch[CDS_N].node ||
	!gn->bch[CDS_N].node->ref) {
	if (!gn->bch[CDS_W].node ||
	    !gn->bch[CDS_W].node->ref)
	    cornGNode[CDS_CORN_UL] = gn;
	if (!gn->bch[CDS_E].node ||
	    !gn->bch[CDS_E].node->ref)
	    cornGNode[CDS_CORN_UR] = gn;
    }

    if (!gn->bch[CDS_S].node ||
	!gn->bch[CDS_S].node->ref) {
	if (!gn->bch[CDS_W].node ||
	    !gn->bch[CDS_W].node->ref)
	    cornGNode[CDS_CORN_LL] = gn;
	if (!gn->bch[CDS_E].node ||
	    !gn->bch[CDS_E].node->ref)
	    cornGNode[CDS_CORN_LR] = gn;
    }
}

for (k=0; k<4; k++) {			/* Check if all corner nodes found */
    if (!cornGNode[k]) break;
    MSG(-1, ("Corner node %s = %d\n", CDS_CORN_TOSTR(k), cornGNode[k]->id));
}
if (k<4) {
    MIerror("Cannot find corner node %s\n", CDS_CORN_TOSTR(k));
    abort();
}

for (t=0; t<numTNode; t++) {		/* Add in the BOTtlenecks */
    gn  = tNode[t];
    r0 = (GRegionDef *)gn->ref;

    gn2 = gn->bch[CDS_N].node;		/* Above junction: */
    if (gn2 && (r2=(GRegionDef *)gn2->ref)) {
	r = CDSHgNewRegion(&numRegion, &pr);
	r->type = G_HOR_BN;
	r->externFlg = (gn->bch[CDS_N].type=='e');
	if (r->externFlg)
	    DASSERT(r0->externFlg && r2->externFlg);
	r->adjReg[G_DOWN] = r0;
	r->adjReg[G_UP] = r2;
	r0->adjReg[G_UP] = r;
	r2->adjReg[G_DOWN] = r;
    }

    gn2 = gn->bch[CDS_E].node;		/* Right junction: */
    if (gn2 && (r2=(GRegionDef *)gn2->ref)) {
	r = CDSHgNewRegion(&numRegion, &pr);
	r->type = G_VER_BN;
	r->externFlg = (gn->bch[CDS_E].type=='e');
	if (r->externFlg)
	    DASSERT(r0->externFlg && r2->externFlg);
	r->adjReg[G_LEFT] = r0;
	r->adjReg[G_RIGHT] = r2;
	r0->adjReg[G_RIGHT] = r;
	r2->adjReg[G_LEFT] = r;
    }
}

for (o=0; o<CDSHead.numObj-1; o++) {	/* Add INSTances */

    if (CDSObj[o].ioPadFlag) continue;	/* Skip ios for now - do them later */

    r = CDSHgNewRegion(&numRegion, &pr);
    r->type = G_INST;
    r->dbRef = (char *)&CDSObj[o];

    gn0 = (CDSGNodeDef *)CDSObj[o].ref;	/* Go around clockwise */
    kp  = CDS_E;
    gn2 = NULL;
    for (gn=gn0, rpd=CDS_E; gn2!=gn0; gn=gn2, rpd=CDS_REVDIR(k)) {
	for (k=CDS_NEXTDIR(rpd,CDS_CCW); !gn->bch[k].node;
	     k=CDS_NEXTDIR(k,CDS_CCW));
	gn2 = gn->bch[k].node;
	DASSERT(gn->bch[k].objNum==o);
	r0 = (GRegionDef *)gn->ref;
	DASSERT(r0);

	switch(k) {
	case CDS_N:
	    d = G_RIGHT; d2 = G_UP;
	    break;
	case CDS_S:
	    d = G_LEFT; d2 = G_DOWN;
	    break;
	case CDS_E:
	    d = G_DOWN; d2 = G_RIGHT;
	    break;
	default:
	    d = G_UP; d2 = G_LEFT;
	    break;
	}

	if (!r0->adjReg[d])
	    r0->adjReg[d] = r;
	r2 = r0->adjReg[d2];
	r2->adjReg[d] = r;

	dr = G_REVDIR(d);
	if (k==CDS_E || k==CDS_W)
	    r->adjReg[dr] = r2;
	else if (kp!=k && !r->adjReg[dr])
	    r->adjReg[dr] = r2;

	kp = k;
    }

    CDSObj[o].ref = (RefAny)r;

    r->dimension[G_HOR] = (CDSObj[o].box.ur.x-CDSObj[o].box.ll.x);
    r->dimension[G_VER] = (CDSObj[o].box.ur.y-CDSObj[o].box.ll.y);
    r->centerLocX = (CDSObj[o].box.ur.x+CDSObj[o].box.ll.x)/2;
    r->centerLocY = (CDSObj[o].box.ur.y+CDSObj[o].box.ll.y)/2;
}

for (io=0; io<4; io++) {

    r = CDSHgNewRegion(&numRegion, &pr);
    r->type = G_INST;

    if (CDSIo[io].numObj==1) {		/* Real io pad side */
	o = CDSIo[io].objNum[0];
	DASSERT(CDSObj[o].ioPadFlag);

	r->dbRef = (char *)&CDSObj[o];

	CDSObj[o].ref = (RefAny)r;
	r->dimension[G_HOR] = (CDSObj[o].box.ur.x-CDSObj[o].box.ll.x);
	r->dimension[G_VER] = (CDSObj[o].box.ur.y-CDSObj[o].box.ll.y);
	r->centerLocX = (CDSObj[o].box.ur.x+CDSObj[o].box.ll.x)/2;
	r->centerLocY = (CDSObj[o].box.ur.y+CDSObj[o].box.ll.y)/2;

	if (io==CDS_IO_BOTTOM || io==CDS_IO_TOP) {
	    r->dimension[G_HOR] = (CDSExtBox.ur.x-CDSExtBox.ll.x);
	    r->centerLocX = (CDSExtBox.ur.x+CDSExtBox.ll.x)/2;
	}
	else {
	    r->dimension[G_VER] = (CDSExtBox.ur.y-CDSExtBox.ll.y);
	    r->centerLocY = (CDSExtBox.ur.y+CDSExtBox.ll.y)/2;
	}
    }

    else {				/* Create dummy io pad */
	MSG(-1, ("Creating dummy io pad on side %s\n", CDS_IO_TOSTR(io)));
	r->dbRef = "dummyIo";
	if (io==CDS_IO_BOTTOM ||
	    io==CDS_IO_TOP) {
	    r->dimension[G_HOR] = (CDSExtBox.ur.x-CDSExtBox.ll.x);
	    r->dimension[G_VER] = 0;
	    r->centerLocX = (CDSExtBox.ur.x+CDSExtBox.ll.x)/2;
	}
	else {
	    r->dimension[G_HOR] = 0;
	    r->dimension[G_VER] = (CDSExtBox.ur.y-CDSExtBox.ll.y);
	    r->centerLocY = (CDSExtBox.ur.y+CDSExtBox.ll.y)/2;
	}
	switch(io) {
	case CDS_IO_BOTTOM: r->centerLocY = CDSExtBox.ll.y; break;
	case CDS_IO_TOP   : r->centerLocY = CDSExtBox.ur.y; break;
	case CDS_IO_LEFT  : r->centerLocX = CDSExtBox.ll.x; break;
	case CDS_IO_RIGHT : r->centerLocX = CDSExtBox.ur.x; break;
	default: DASSERT(0);
	}
    }

    switch(io) {
    case CDS_IO_BOTTOM:
	c = CDS_CORN_LR; c2 = CDS_CORN_LL;
	d = G_LEFT; d2 = G_DOWN;
	break;
    case CDS_IO_TOP:
	c = CDS_CORN_UL; c2 = CDS_CORN_UR;
	d = G_RIGHT; d2 = G_UP;
	break;
    case CDS_IO_LEFT:
	c = CDS_CORN_UL; c2 = CDS_CORN_LL;
	d = G_DOWN; d2 = G_LEFT;
	break;
    case CDS_IO_RIGHT:
	c = CDS_CORN_LR; c2 = CDS_CORN_UR;
	d = G_UP; d2 = G_RIGHT;
	break;
    default: DASSERT(0);
    }

    r2 = (GRegionDef *)cornGNode[c]->ref;
    r->adjReg[G_REVDIR(d2)] = r2->adjReg[d];
    for ( ; r2; r2=r2->adjReg[d]) {	/* This is not consistent but */
	r2->adjReg[d2] = r;		/*   I need adj ptr on the corners */
					/*   for computing its dimension */
	if (r2==(GRegionDef *)cornGNode[c2]->ref)
	    break;
    }
}

for (r=GRegList; r; r=r->next) {	/* Compute dimension + capacity */
					/*   of dominant cross sections */
    if (r->type==G_INST) break;

    box = CDSExtBox;

    switch(r->type) {
    case G_VER_BN:
	r2 = r->adjReg[G_UP];
	DASSERT(r2);
	DASSERT(r2->type==G_INST);
	box.ur.y = G_REGBOX_LL_Y(r2);
	if (r2->adjReg[G_DOWN]==r)
	    box.ll.x = G_REGBOX_LL_X(r2);
	if ((r3=r->adjReg[G_RIGHT]) &&
	    r3->adjReg[G_UP]!=r2)
	    box.ur.x = G_REGBOX_UR_X(r2);

	r2 = r->adjReg[G_DOWN];
	DASSERT(r2);
	DASSERT(r2->type==G_INST);
	box.ll.y = G_REGBOX_UR_Y(r2);
	if (r2->adjReg[G_UP]==r &&
	    box.ur.x>G_REGBOX_UR_X(r2))
	    box.ur.x = G_REGBOX_UR_X(r2);
	if ((r3=r->adjReg[G_LEFT]) &&
	    r3->adjReg[G_DOWN]!=r2 &&
	    box.ll.x<G_REGBOX_LL_X(r2))
	    box.ll.x = G_REGBOX_LL_X(r2);

	r->dimension[G_HOR] = box.ur.x-box.ll.x;
	r->dimension[G_VER] = box.ur.y-box.ll.y;
	r->centerLocX = (box.ll.x+box.ur.x)/2;
	r->centerLocY = (box.ll.y+box.ur.y)/2;

	r->capacity[0] = r->dimension[G_VER];
	break;
    case G_HOR_BN:
	r2 = r->adjReg[G_RIGHT];
	DASSERT(r2);
	DASSERT(r2->type==G_INST);
	box.ur.x = G_REGBOX_LL_X(r2);
	if (r2->adjReg[G_LEFT]==r)
	    box.ll.y = G_REGBOX_LL_Y(r2);
	if ((r3=r->adjReg[G_UP]) &&
	    r3->adjReg[G_RIGHT]!=r2)
	    box.ur.y = G_REGBOX_UR_Y(r2);

	r2 = r->adjReg[G_LEFT];
	DASSERT(r2);
	DASSERT(r2->type==G_INST);
	box.ll.x = G_REGBOX_UR_X(r2);
	if (r2->adjReg[G_RIGHT]==r &&
	    box.ur.y>G_REGBOX_UR_Y(r2))
	    box.ur.y = G_REGBOX_UR_Y(r2);
	if ((r3=r->adjReg[G_DOWN]) &&
	    r3->adjReg[G_LEFT]!=r2 &&
	    box.ll.y<G_REGBOX_LL_Y(r2))
	    box.ll.y = G_REGBOX_LL_Y(r2);

	r->dimension[G_HOR] = box.ur.x-box.ll.x;
	r->dimension[G_VER] = box.ur.y-box.ll.y;
	r->centerLocX = (box.ll.x+box.ur.x)/2;
	r->centerLocY = (box.ll.y+box.ur.y)/2;

	r->capacity[0] = r->dimension[G_HOR];
	break;
    case G_DOM_XSEC:
	break;
    default:
	MIerror("Cannot compute dimension of reg type %d\n",
	    r->type);
	break;
    }
}

for (r=GRegList; r; r=r->next) {	/* Compute dimension + capacity */
    if (r->type!=G_DOM_XSEC) break;

    box = CDSExtBox;

    r2 = r->adjReg[G_UP];
    if (r2)
	box.ur.y = G_REGBOX_LL_Y(r2);
/*  else DASSERT(r->externFlg); */

    r2 = r->adjReg[G_DOWN];
    if (r2)
	box.ll.y = G_REGBOX_UR_Y(r2);
/*  else DASSERT(r->externFlg); */

    r2 = r->adjReg[G_RIGHT];
    if (r2)
	box.ur.x = G_REGBOX_LL_X(r2);
/*  else DASSERT(r->externFlg); */

    r2 = r->adjReg[G_LEFT];
    if (r2)
	box.ll.x = G_REGBOX_UR_X(r2);
/*  else DASSERT(r->externFlg); */

    r->dimension[G_HOR] = box.ur.x-box.ll.x;
    r->dimension[G_VER] = box.ur.y-box.ll.y;
    r->centerLocX = (box.ll.x+box.ur.x)/2;
    r->centerLocY = (box.ll.y+box.ur.y)/2;
}
					/* Wipe out inconsistent corner reg */
					/*   pointers to the io regions */
((GRegionDef *)cornGNode[CDS_CORN_LL]->ref)->adjReg[G_DOWN ] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_LL]->ref)->adjReg[G_LEFT ] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_LR]->ref)->adjReg[G_DOWN ] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_LR]->ref)->adjReg[G_RIGHT] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_UL]->ref)->adjReg[G_UP   ] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_UL]->ref)->adjReg[G_LEFT ] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_UR]->ref)->adjReg[G_UP   ] = NULL;
((GRegionDef *)cornGNode[CDS_CORN_UR]->ref)->adjReg[G_RIGHT] = NULL;

gn = cornGNode[CDS_CORN_LL];		/* Init offset to be lower left */
r = (GRegionDef *)gn->ref;		/*   corner of lower left XSEC reg */
CDSHgOffsetX = G_REGBOX_LL_X(r);
CDSHgOffsetY = G_REGBOX_LL_Y(r);

maxId = 0;
for (r=GRegList; r; r=r->next) {
    /* Ids start from 1 instead of 0 */
    r->id++;

    /* Check number of regions */
    maxId =  MAX(r->id, maxId);

    /* Initialize GReg->orient */
    if (r->type==G_HOR_BN)
	r->orient = G_HOR;
    else if (r->type==G_VER_BN)
	r->orient = G_VER;

    /* Initialize GReg->newLocX and newLocY */
    r->newLocX = G_REGBOX_LL_X(r) - CDSHgOffsetX;
    r->newLocY = G_REGBOX_LL_Y(r) - CDSHgOffsetY;
}

if (maxId>=G_MAXREGION) {
    MIerror("too many regions %s > G_MAXREGION=%d\n",
	r->id, G_MAXREGION);
    abort();
}

FREE_ARRAY(CDSHead.numGNode, tNode, CDSGNodeDef*);

MI_MOD_END();
} /* ............................. End of procedure CDSHgConvtGraph */
/* Print regions in Nishizaki format
----------------------------------------------------------------------------*/
CDSHgDeleteIos()
{
GRegionDef	*r,  *r2, *rPrev;
int		k;
Boolean		foundFlg;

MI_MOD_START("CDSHgDeleteIos");

for (k=0; k<4; k++) {			/* k = io side */

    foundFlg = FALSE;			/* Exaustive search to find io reg */
    rPrev = NULL;
    for (r=GRegList; r; r=r->next) {
	switch(k) {
	case CDS_IO_BOTTOM:
	    if (!r->adjReg[G_DOWN] &&
		!r->adjReg[G_LEFT] && !r->adjReg[G_RIGHT])
		foundFlg = TRUE;
	    break;
	case CDS_IO_TOP:
	    if (!r->adjReg[G_UP] &&
		!r->adjReg[G_LEFT] && !r->adjReg[G_RIGHT])
		foundFlg = TRUE;
	    break;
	case CDS_IO_LEFT:
	    if (!r->adjReg[G_LEFT] &&
		!r->adjReg[G_DOWN] && !r->adjReg[G_UP])
		foundFlg = TRUE;
	    break;
	case CDS_IO_RIGHT:
	    if (!r->adjReg[G_RIGHT] &&
		!r->adjReg[G_DOWN] && !r->adjReg[G_UP])
		foundFlg = TRUE;
	    break;
	default: DASSERT(0);
	}
	if (foundFlg) break;
	rPrev = r;
    }
    if (!foundFlg) {
	MIerror("Cannot find side %s io reg\n", CDS_IO_TOSTR(k));
	abort();
    }
    MSG(-1, ("Deleting io reg on side %s\n", CDS_IO_TOSTR(k)));

    switch(k) {				/* Get rid of pointer references to r */
    case CDS_IO_BOTTOM:
	for (r2=r->adjReg[G_UP]; r2; r2=r2->adjReg[G_LEFT])
	    if (r2->adjReg[G_DOWN]==r)
		r2->adjReg[G_DOWN] = NULL;
	break;
    case CDS_IO_TOP:
	for (r2=r->adjReg[G_DOWN]; r2; r2=r2->adjReg[G_RIGHT])
	    if (r2->adjReg[G_UP]==r)
		r2->adjReg[G_UP] = NULL;
	break;
    case CDS_IO_LEFT:
	for (r2=r->adjReg[G_RIGHT]; r2; r2=r2->adjReg[G_DOWN])
	    if (r2->adjReg[G_LEFT]==r)
		r2->adjReg[G_LEFT] = NULL;
	break;
    case CDS_IO_RIGHT:
	for (r2=r->adjReg[G_LEFT]; r2; r2=r2->adjReg[G_UP])
	    if (r2->adjReg[G_RIGHT]==r)
		r2->adjReg[G_RIGHT] = NULL;
	break;
    default: DASSERT(0);
    }

    CDSHgDelRegion(r, rPrev);
}

MI_MOD_END();
} /* ............................. End of procedure CDSHgDeleteIos */
/* Print regions in Nishizaki format
----------------------------------------------------------------------------*/
CDSHgPrintReg(regFileName)

String		regFileName;
{
GRegionDef	*r;
int		cap, l;
String		regType[7];
FILE		*f;

MI_MOD_START("CDSHgPrintReg");

regType[G_HOR_BN]	= "HOR_BN";
regType[G_VER_BN]	= "VER_BN";
regType[G_OVERCELL]	= "OVERCELL";
regType[G_INST]		= "INST";
regType[G_SWBOX]	= "SBOX";
regType[G_CHANNEL]	= "CHANNEL";
regType[G_DOM_XSEC]	= "DOM_XSEC";

f = MIfopen(regFileName,"w","Nishizaki regions");

for (r=GRegList; r; r=r->next) {
    cap = 0;
    for (l=0; l<G_NLAY; l++)
	cap += r->capacity[l];
    fprintf(f, "%d %s%s cap %d dim %d x %d center %d %d\n",
	r->id, regType[r->type], r->externFlg ? "[e]" : "", cap,
	r->dimension[G_HOR], r->dimension[G_VER],
	r->centerLocX, r->centerLocY);
    fprintf(f, "		box=(%d %d) (%d %d)\n",
	r->centerLocX - r->dimension[G_HOR]/2,
	r->centerLocY - r->dimension[G_VER]/2,
	r->centerLocX + r->dimension[G_HOR]/2,
	r->centerLocY + r->dimension[G_VER]/2);
    if (r->type==G_INST) {
	int	o;
	for (o=0; o<CDSHead.numObj; o++)
	    if (CDSObj[o].ref==(RefAny)r) break;
	DASSERT(o<CDSHead.numObj);
	fprintf(f,"		name=%s\n",
	    CDSObj[o].dbRef);
    }
}

for (r=GRegList; r; r=r->next) {
    fprintf(f, "%d adj %d %d %d %d\n",
	r->id,
	(r->adjReg[0] ? r->adjReg[0]->id : 0),
	(r->adjReg[1] ? r->adjReg[1]->id : 0),
	(r->adjReg[2] ? r->adjReg[2]->id : 0),
	(r->adjReg[3] ? r->adjReg[3]->id : 0));
}

fclose(f);

MI_MOD_END();
} /* ............................. End of procedure CDSHgPrintReg */
/* Allocate new region + do linked list stuff.
----------------------------------------------------------------------------*/
GRegionDef *
CDSHgNewRegion(numReg, prevReg)

int		*numReg;
GRegionDef	**prevReg;
{
GRegionDef	*r;
int		l;

MI_MOD_START("CDSHgNewRegion");

r = ALLOC_TYPE(GRegionDef);
if (!GRegList)				/* This should be at ll corner */
    GRegList = r;
if (*prevReg)				/* Linked list */
    (*prevReg)->next = r;
*prevReg = r;

r->next = NULL;
r->id = *numReg;			/* Id >= 0 (will be +1 later) */
					/* Defaults: */
r->externFlg = FALSE;			/* Not external */
r->orient = -1;				/* Orientation of bottleneck */
r->overcellDir = -1;			/* No over-the-cell routing */
r->expandableFlg = TRUE;		/* Expandable channel or switch box */
r->parent = NULL;			/* To parent object */
r->sibling = NULL;
r->child = NULL;

r->adjReg[G_UP] = NULL;			/* Adjacent regions */
r->adjReg[G_DOWN] = NULL;
r->adjReg[G_LEFT] = NULL;
r->adjReg[G_RIGHT] = NULL;

for (l=0; l<G_NLAY; l++) {
    r->capacity[l] = 0;
    r->usage[l] = 0;
}
r->estUsage = 0;

r->dimension[G_VER] = 0;		/* Dimensions */
r->dimension[G_HOR] = 0;
r->centerLocX = 0;
r->centerLocY = 0;
r->newLocX = 0;
r->newLocY = 0;

r->pin  = NULL;

(*numReg)++;

MI_MOD_END();
return(r);
} /* ............................. End of procedure CDSHgNewRegion */
/* Allocate new region + do linked list stuff.
----------------------------------------------------------------------------*/
CDSHgDelRegion(delReg, prevReg)

GRegionDef	*delReg;		/* Region to be deleted */
GRegionDef	*prevReg;
{
GRegionDef	*r;

MI_MOD_START("CDSHgDelRegion");

if (prevReg) {
    DASSERT(prevReg->next==delReg);
    prevReg->next = delReg->next;
}
else
    GRegList = delReg->next;

for (r=delReg->next; r; r=r->next) r->id--;

FREE_TYPE(delReg, GRegionDef);

MI_MOD_END();
} /* ............................. End of procedure CDSHgDelRegion */
/* Check if external node should be skipped.
----------------------------------------------------------------------------*/
Boolean
CDSHgSkipExtNode(gn)

CDSGNodeDef	*gn;
{

/* Only consider keeping external nodes that have no associated io */
/* TRUE means skip this node, FALSE means keep this node in glb route graph */

if (!CDSIoFlag[CDS_IO_BOTTOM] &&
    gn->bch[CDS_N].node && !gn->bch[CDS_S].node) {
    if ((!gn->bch[CDS_W].node && CDSIoFlag[CDS_IO_LEFT]) ||
	(!gn->bch[CDS_E].node && CDSIoFlag[CDS_IO_RIGHT]))
	 return(TRUE);
    else return(FALSE);
}

if (!CDSIoFlag[CDS_IO_TOP] &&
    gn->bch[CDS_S].node && !gn->bch[CDS_N].node) {
    if ((!gn->bch[CDS_W].node && CDSIoFlag[CDS_IO_LEFT]) ||
	(!gn->bch[CDS_E].node && CDSIoFlag[CDS_IO_RIGHT]))
	 return(TRUE);
    else return(FALSE);
}

if (!CDSIoFlag[CDS_IO_LEFT] &&
    gn->bch[CDS_E].node && !gn->bch[CDS_W].node) {
    if ((!gn->bch[CDS_N].node && CDSIoFlag[CDS_IO_TOP]) ||
	(!gn->bch[CDS_S].node && CDSIoFlag[CDS_IO_BOTTOM]))
	 return(TRUE);
    else return(FALSE);
}

if (!CDSIoFlag[CDS_IO_RIGHT] &&
    gn->bch[CDS_W].node && !gn->bch[CDS_E].node) {
    if ((!gn->bch[CDS_N].node && CDSIoFlag[CDS_IO_TOP]) ||
	(!gn->bch[CDS_S].node && CDSIoFlag[CDS_IO_BOTTOM]))
	 return(TRUE);
    else return(FALSE);
}

return(TRUE);
} /* ............................. End of procedure CDSHgSkipExtNode */
