/*
Copyright (C) 1988 University of California (Berkeley). All rights reserved.

CDSHgInterface.c

Interface of channel definition and ordering to OCT database.

CDSHgInterface		Create interface for hierarchical global router
CDSRegenGraph		Regenerate graph from components
CDSHgCleanGraph		Clean up unnecessary nodes in graph
CDSCompObjBox		Compute (minimum) instance boxes

Author(s):		Mitsuru Igusa
Revision history:	$Log: CDSHgInter.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:49  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  90/02/07  17:03:23  octtools
 * Checkin for 3.5 distribution
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CDSInc.h"

/* Create interface for hierarchical global router
----------------------------------------------------------------------------*/
void
CDSHgInterface(octMain)

octObject	*octMain;
{
CDSGNodeDef	*headNode;
int		o;

MI_MOD_START("CDSHgInterface");

for (o=0; o<CDSHead.numObj; o++)
    CDSCompObjBox(o);

CDSRegenGraph(&headNode);
CDSHgCleanGraph(headNode);
if (*debugFlag>=1)
    CDSPrintNGraph(headNode);

CDSHgSetIoObjNum();
CDSHgConvtGraph(headNode);
if (*debugFlag>=0)
    CDSHgPrintReg("reg");

CDSHgConvtNet(octMain);
if (*debugFlag>=0)
    CDSHgPrintNet("regNet");

CDSHgDeleteIos();

MI_MOD_END();
} /* ............................. End of procedure CDSHgInterface */
/* Regenerate graph
----------------------------------------------------------------------------*/
CDSRegenGraph(headNode)

CDSGNodeDef	**headNode;
{
#define	CDS_DIM_SLICENODE 2000
int		stkRef;
CDSTreeDef	**ch;
CDSTreeDef	*t, *t1, *t2;
CDSGNodeDef	*g1, *gn1;
CDSGNodeDef	*g2, *gn2, *gm2, *gm2b;
int		c, numChan, chkNumChan;
int		k1, k2, k, rpd, o, o1, o2, o2b;
int		nn;
int		numEdgeInCommon;
Boolean		firstTimeFlg;
short		numSliceNode, s;
CDSGNodeDef	*sliceNode[CDS_DIM_SLICENODE];

MI_MOD_START("CDSRegenGraph");

numChan = CDSTree->order-3;
ch = ALLOC_ARRAY(numChan, CDSTreeDef*);
for (c=0; c<numChan; c++) ch[c] = NULL;

chkNumChan = 0;
stkRef = MIStackPtr;
MI_STACK_PUSH(CDSTree);

while (MIStackPtr!=stkRef) {
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;

    DASSERT(t->type!=CDS_TREE_NOT_DONE);

    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
	if (t->type!=CDS_TREE_EXT_DONE) {
	    c = t->order;
	    DRANGE(0, c, numChan);
	    DASSERT(!ch[c]);
	    ch[c] = t;
	    chkNumChan++;
	}
    }

    else {				/* Is a leaf cell */
	o = t->objNum;
	DRANGE(0, o, CDSHead.numObj-1);
	CDSObj[o].ref = (RefAny)t->node;
    }
}
DASSERTSAME(chkNumChan, numChan);

for (c=0; c<numChan; c++) {		/* Merge channels from bottom up */
    t = ch[c];
    DASSERT(!t->node);
    t1 = t->child[CDS_LO];
    t2 = t->child[CDS_HI];
    if (!t1 || !t1->node || !t2 || !t2->node) abort();
    DASSERT(t1 && t1->node && t2 && t2->node);

    MSG(0, ("######## On channel %d treeId=%d ########\n", c, t->id));

    if (*debugFlag>=1)
	CDSPrintNGraph(t1->node);
    if (*debugFlag>=1)
	CDSPrintNGraph(t2->node);

    if (t1->node->loc[CDS_X]+t1->node->loc[CDS_Y]<
	t2->node->loc[CDS_X]+t2->node->loc[CDS_Y])
    then t->node = t1->node;
    else t->node = t2->node;

    numSliceNode = 0;			/* Go around clockwise */
    gn1 = NULL;
    firstTimeFlg = TRUE;
    for (g1=t1->node, rpd=CDS_W; gn1!=t1->node; g1=gn1, rpd=CDS_REVDIR(k1)) {
	for (k1=CDS_NEXTDIR(rpd,CDS_CW); !g1->bch[k1].node;
	     k1=CDS_NEXTDIR(k1,CDS_CW));
	gn1 = g1->bch[k1].node;
	if (g1->bch[k1].treeId!=t->id) continue;

	if (firstTimeFlg)
	    sliceNode[numSliceNode++] = g1;
	URANGE(numSliceNode+1, CDS_DIM_SLICENODE);
	sliceNode[numSliceNode++] = gn1;

	firstTimeFlg = FALSE;
    }

    gn1 = NULL;
    gn2 = NULL;
    firstTimeFlg = TRUE;
    for (g2=t2->node, rpd=CDS_W; gn2!=t2->node; g2=gn2, rpd=CDS_REVDIR(k2)) {
	for (k2=CDS_NEXTDIR(rpd,CDS_CW); !g2->bch[k2].node;
	     k2=CDS_NEXTDIR(k2,CDS_CW));
	gn2 = g2->bch[k2].node;
	if (gn2==gn1) break;		/* Extra stop criteria */
	if (g2->bch[k2].treeId!=t->id) continue;

	for (nn=0; nn<(firstTimeFlg ? 2 : 1); nn++) {
	    gm2 = (nn ? g2 : gn2);

	    for (s=0; s<numSliceNode; s++) {
		if (!(g1=sliceNode[s])) continue;
		if (g1->loc[CDS_X]==gm2->loc[CDS_X] &&
		    g1->loc[CDS_Y]==gm2->loc[CDS_Y]) break;
	    }
	    if (s==numSliceNode) {
		fprintf(stderr,
		    "ERROR Matching node not found for ch %d at (%d %d)\n",
		    c, gm2->loc[CDS_X], gm2->loc[CDS_Y]);
		abort();
	    }
	    sliceNode[s] = NULL;

	    if (nn==1) gn1 = g1;	/* Kludge since g2 graph is changing */

	    numEdgeInCommon = 0;	/* Merge gm2 into g1 */
	    for (k=0; k<4; k++) {
		gm2b = gm2->bch[k].node;
		if (!gm2b) continue;
		rpd = CDS_REVDIR(k);
		if (g1->bch[k].node) {
		    numEdgeInCommon++;
		    o1 = g1->bch[k].objNum;
		    o2 = gm2->bch[k].objNum;
		    o2b = gm2b->bch[rpd].objNum;

		    if (o1<CDS_NBOBJ_EXTERNAL) {
			g1->bch[k].objNum = o2;
			DASSERT(o2>CDS_NBOBJ_EXTERNAL);
		    }
		    else if (o2>=CDS_NBOBJ_EXTERNAL) {
			DASSERT(o1==o2);
		    }

		    if (o2>CDS_NBOBJ_EXTERNAL &&
			CDSObj[o2].ref==(RefAny)gm2) {
			DRANGE(0, o2, CDSHead.numObj-1);
			CDSObj[o2].ref = (RefAny)g1;
		    }
		    if (o2b>CDS_NBOBJ_EXTERNAL &&
			CDSObj[o2b].ref==(RefAny)gm2) {
			DRANGE(0, o2b, CDSHead.numObj-1);
			CDSObj[o2b].ref = (RefAny)g1;
		    }
					/* Not an external edge */
		    g1->bch[k].type = ' ';
		    g1->bch[k].node->bch[rpd].type = ' ';
		}
		else {
		    g1->bch[k] = gm2->bch[k];
		    gm2b->bch[rpd].node = g1;
		}
	    }

	    if (numEdgeInCommon==2)	/* Not an external node */
		g1->type = ' ';

	    if (t->node==gm2)
		t->node = g1;
	}

	firstTimeFlg = FALSE;
    }

    if (*debugFlag>=1)
	CDSPrintNGraph(t->node);
}

*headNode = t->node;

FREE_ARRAY(numChan, ch, CDSTreeDef*);

MI_MOD_END();
} /* ............................. End of procedure CDSRegenGraph */
/* 
----------------------------------------------------------------------------*/
CDSHgCleanGraph(headNode)

CDSGNodeDef	*headNode;
{
int		numTNode = 0;
int		numClrNode = 0;
CDSGNodeDef	**tNode;
CDSGNodeDef	*gn, *gn2, *gnn, *gne, *gns, *gnw;
int		stkRef;
int		k, t;

MI_MOD_START("CDSHgCleanGraph");

tNode = ALLOC_ARRAY(CDSHead.numGNode, CDSGNodeDef*);

stkRef = MIStackPtr;
MI_STACK_PUSH(headNode);
tNode[numTNode++] = headNode;

while (stkRef!=MIStackPtr) {
    MI_STACK_POP(CDSGNodeDef*, gn);
    for (k=0; k<4; k++) {
	gn2 = gn->bch[k].node;
	if (!gn2) continue;
	for (t=0; t<numTNode; t++)
	    if (gn2==tNode[t]) break;
	if (t<numTNode) continue;
	ASSERT(numTNode<CDSHead.numGNode);
	MI_STACK_PUSH(gn2);
	tNode[numTNode++] = gn2;
    }
}

for (t=0; t<numTNode; t++) {
    /* Don't free gn since it gets freed later in CDSCleanUp as a linked list */
    gn = tNode[t];
    gnn = gn->bch[CDS_N].node;
    gns = gn->bch[CDS_S].node;
    gne = gn->bch[CDS_E].node;
    gnw = gn->bch[CDS_W].node;
    if (gnn && gns && !gne && !gnw) {
	gnn->bch[CDS_S].node = gns;
	gns->bch[CDS_N].node = gnn;
	numClrNode++;
    }
    if (gne && gnw && !gnn && !gns) {
	gne->bch[CDS_W].node = gnw;
	gnw->bch[CDS_E].node = gne;
	numClrNode++;
    }
}

FREE_ARRAY(CDSHead.numGNode, tNode, CDSGNodeDef*);
CDSHead.numGNode = numTNode-numClrNode;

MI_MOD_END();
} /* ............................. End of procedure CDSHgCleanGraph */
/* Compute box of instance.
----------------------------------------------------------------------------*/
CDSCompObjBox(o)

int		o;
{
CDSObjDef	*obj;
int		x, y, xy, minXY, m;
int		b, bestB;

MI_MOD_START("CDSCompObjBox");

obj = &CDSObj[o];

minXY = MAXINT;
for (b=0; b<obj->numBdy; b++) {
    x = obj->bdy[b][CDS_X];
    y = obj->bdy[b][CDS_Y];
    xy = x+y;				/* Lower left corner */
    if (xy<minXY) {
	minXY = xy;
	bestB = b;
    }
}

obj->box.ll.x = obj->bdy[bestB][CDS_X];
obj->box.ll.y = obj->bdy[bestB][CDS_Y];

minXY = MAXINT;
for (b=0; b<obj->numBdy; b++) {
    x = obj->bdy[b][CDS_X];
    y = obj->bdy[b][CDS_Y];
    xy = x-y;				/* Upper left corner */
    if (xy<minXY) {
	minXY = xy;
	bestB = b;
    }
}

obj->box.ll.x = MAX(obj->box.ll.x, obj->bdy[bestB][CDS_X]);
obj->box.ur.y = obj->bdy[bestB][CDS_Y];

minXY = MAXINT;
for (b=0; b<obj->numBdy; b++) {
    x = obj->bdy[b][CDS_X];
    y = obj->bdy[b][CDS_Y];
    xy = -x-y;				/* Upper right corner */
    if (xy<minXY) {
	minXY = xy;
	bestB = b;
    }
}

obj->box.ur.x = obj->bdy[bestB][CDS_X];
obj->box.ur.y = MIN(obj->box.ur.y, obj->bdy[bestB][CDS_Y]);

minXY = MAXINT;
for (b=0; b<obj->numBdy; b++) {
    x = obj->bdy[b][CDS_X];
    y = obj->bdy[b][CDS_Y];
    xy = -x+y;				/* Lower right corner */
    if (xy<minXY) {
	minXY = xy;
	bestB = b;
    }
}

obj->box.ur.x = MIN(obj->box.ur.x, obj->bdy[bestB][CDS_X]);
obj->box.ll.y = MAX(obj->box.ll.y, obj->bdy[bestB][CDS_Y]);

if (obj->box.ll.x>obj->box.ur.x) {
    m = (obj->box.ll.x+obj->box.ur.x)/2;
    obj->box.ll.x = obj->box.ur.x = m;
}
if (obj->box.ll.y>obj->box.ur.y) {
    m = (obj->box.ll.y+obj->box.ur.y)/2;
    obj->box.ll.y = obj->box.ur.y = m;
}

MI_MOD_END();
} /* ............................. End of procedure CDSHgCompObjBox */
