/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

hgCreateChan.c

hgCreateChan	Create channels

Author(s):		Mitsuru Igusa
Revision history:	$Log: hgCreateChan.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:49  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  90/02/07  17:03:25  octtools
 * Checkin for 3.5 distribution
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CDSInc.h"
#include "mercury.h"

int		numReg, numChan;
octId		*chanOctId;
short		*regChanNum;

/* Infer where channels are from CDSTree hierarchical data structure.
----------------------------------------------------------------------------*/
hgCreateChan(octFacet)

octObject	*octFacet;		/* Assume this has been opened */
{
GRegionDef	*reg0, *reg;
int		ti, ch;
int		r, c, o, d, k;
CDSTreeDef	**treeIdx;
CDSTreeDef	*t;
int		stkRef;
short		numAdjChan;
short		adjChan[G_NDIR];
Boolean		deleteEmptyRoomFlag = TRUE;

MI_MOD_START("hgCreateChan");

numReg = 0;				/* Alloc and init regChanNum[] */
for (reg=GRegList; reg; reg=reg->next)
    numReg++;
regChanNum = ALLOC_ARRAY(numReg+1, short);
for (r=0; r<numReg+1; r++)
    regChanNum[r] = -1;

					/* Alloc and init treeIdx[] */
treeIdx = ALLOC_ARRAY(CDSHead.numTree, CDSTreeDef*);
for (ti=0; ti<CDSHead.numTree; ti++)
    treeIdx[ti] = NULL;

stkRef = MIStackPtr;
MI_STACK_PUSH(CDSTree);

numChan = 0;
while (MIStackPtr!=stkRef) {		/* Traverse tree to init ordTreeList */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    treeIdx[t->id] = t;
    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	numChan = MAX(t->order+1, numChan);
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
    }
}
					/* Go through each object and look */
					/*   at its neighbors */
for (o=0; o<CDSHead.numObj-1; o++) {
    for (c=0; c<4; c++) {
	ti  = CDSObj[o].corn[c].treeId;
	ch  = treeIdx[ti]->order;
	reg0 = (GRegionDef *)CDSObj[o].ref;
	switch(c) {
	case CDS_CORN_LL:		/* Bottom channel */
	    for (reg=reg0->adjReg[G_DOWN];
		 reg && reg->adjReg[G_UP]==reg0;
		 reg=reg->adjReg[G_RIGHT]) {
		r = reg->id;
		if (regChanNum[r]>=0 &&
		    regChanNum[r]!=ch)
		    fprintf(stdout,
			"ERROR Reg %d bottom channel %d init to %d not %d\n",
			r, reg0->id, regChanNum[r], ch);
		regChanNum[r] = ch;
	    }
	    break;
	case CDS_CORN_LR:		/* Right channel */
	    for (reg=reg0->adjReg[G_RIGHT];
		 reg && reg->adjReg[G_LEFT]==reg0;
		 reg=reg->adjReg[G_DOWN]) {
		r = reg->id;
		if (regChanNum[r]>=0 &&
		    regChanNum[r]!=ch)
		    fprintf(stdout,
			"ERROR Reg %d right channel %d init to %d not %d\n",
			r, reg0->id, regChanNum[r], ch);
		regChanNum[r] = ch;
	    }
	    break;
	case CDS_CORN_UR:		/* Top channel */
	    for (reg=reg0->adjReg[G_UP];
		 reg && reg->adjReg[G_DOWN]==reg0;
		 reg=reg->adjReg[G_LEFT]) {
		r = reg->id;
		if (regChanNum[r]>=0 &&
		    regChanNum[r]!=ch)
		    fprintf(stdout,
			"ERROR Reg %d top channel %d init to %d not %d\n",
			r, reg0->id, regChanNum[r], ch);
		regChanNum[r] = ch;
	    }
	    break;
	case CDS_CORN_UL:		/* Left channel */
	    for (reg=reg0->adjReg[G_LEFT];
		 reg && reg->adjReg[G_RIGHT]==reg0;
		 reg=reg->adjReg[G_UP]) {
		r = reg->id;
		if (regChanNum[r]>=0 &&
		    regChanNum[r]!=ch)
		    fprintf(stdout,
			"ERROR Reg %d left channel %d init to %d not %d\n",
			r, reg0->id, regChanNum[r], ch);
		regChanNum[r] = ch;
	    }
	default: break;
	}
    }
}

for (reg0=GRegList; reg0; reg0=reg0->next) {
    r = reg0->id;
    if (regChanNum[r]>=0) continue;
    if (reg0->type!=G_DOM_XSEC) continue;

    numAdjChan = 0;
    for (d=0; d<G_NDIR; d++) {
	reg = reg0->adjReg[d];
	if (reg && regChanNum[reg->id]>=0) {
	    ch = regChanNum[reg->id];
	    for (k=0; k<numAdjChan; k++)
		if (adjChan[k]==ch) break;
	    if (k==numAdjChan)
		adjChan[numAdjChan++] = ch;
	    else {
		regChanNum[r] = ch;
		break;
	    }
	}
    }

    if (regChanNum[r]<0) {
	for (k=0; k<numAdjChan; k++)
	    regChanNum[r] = MAX(regChanNum[r], adjChan[k]);
    }
}

if (deleteEmptyRoomFlag) {
    hgDeleteEmptyRooms(octFacet, treeIdx);
}

hgChanToOct(octFacet);

MI_MOD_END();
} /* ............................. End of procedure hgCreateChan */
/* Convert net to net group
----------------------------------------------------------------------------*/
hgDeleteEmptyRooms(octFacet, treeIdx)

octObject	*octFacet;		/* Assume this has been opened */
CDSTreeDef	**treeIdx;
{
short		*oldToNewChanId;
short		*newChanId;
short		ti, oldId;
short		numNewChan;
short		numEmptyRoom;
short		k, k2, d, d2, o, c, r, ch;
char		dir;
CDSTreeDef	*t, *t2, *tp, *tp2, *tpLast;
octObject	octInst;
octGenerator	genInst;

MI_MOD_START("hgDeleteEmptyRooms");

if (*debugFlag>=2) {
    printf("Before delete empty rooms\n");
    CDSPrintTree();
}

oldToNewChanId = ALLOC_ARRAY(numChan, short);

for (k=0; k<numChan; k++)
    oldToNewChanId[k] = k;

/* Tag empty rooms by setting tree->dir = 'e' */
numEmptyRoom = 0;
for (ti=0; ti<CDSHead.numTree; ti++) {
    t = treeIdx[ti];
    if (t->type==CDS_TREE_ELEMENT) {
	DRANGE(0, t->objNum, CDSHead.numObj);
	if (!strncmp(CDSObj[t->objNum].dbRef, "emptyRoom", 9)) {
	    t->dir = 'e';
	    numEmptyRoom++;
	}
    }
}
MSG(1, ("Number of emptyRooms to merge = %d\n", numEmptyRoom));
numNewChan = numChan-numEmptyRoom;

for ( ; numEmptyRoom; numEmptyRoom--) {
    for (ti=0; ti<CDSHead.numTree; ti++) {
	t = treeIdx[ti];
	if (t->dir!='e') continue;

	DASSERT(t->id==ti);
	t->dir = 'x';			/* Delete t */

	tp = treeIdx[t->parentId];
	if (tp->child[CDS_LO]==t) d = CDS_LO;
	else			  d = CDS_HI;
	DASSERT(tp->child[d]==t);

	/* Delete emptyRoom instance corresponding to t */

	OH_ASSERT(ohGetByInstName(octFacet, &octInst, CDSObj[t->objNum].dbRef));
	OH_ASSERT(octDelete(&octInst));

	d2 = (d==CDS_LO ? CDS_HI : CDS_LO);
	t2 = tp->child[d2];
	DASSERT(t2);
	ch = tp->order;

#ifdef BAD_MERGING
	if (t2->dir=='e') {		/* 2 emptyRooms side by side */
	    MSG(2, ("2 emptyRooms %d %d\n", ti, t2->id));
	    tp->child[d2] = NULL;
	    dir = (tp->dir=='v' ? 'h' : 'v');
	    for (tp2=treeIdx[tp->parentId]; tp2; tp2=treeIdx[tp2->parentId])
		if (tp2->dir==dir) break;
	    DASSERT(tp2);
	}
	else {				/* 1 emptyRoom next to normal obj */
	    MSG(2, ("1 emptyRoom %d next to %d\n", ti, t2->id));
	    dir = tp->dir;
	    tpLast = tp;
	    for (tp2=treeIdx[tp->parentId]; tp2; tp2=treeIdx[tp2->parentId]) {
		if (tp2->dir==dir && tp2->child[d2]==tpLast) break;
		tpLast = tp2;
	    }
	    DASSERT(tp2);
	}
#else
	if (t2->dir=='e') {
	    MSG(2, ("2 emptyRooms %d %d\n", ti, t2->id));
	}
	else {
	    MSG(2, ("1 emptyRoom %d next to %d\n", ti, t2->id));
	}
	dir = tp->dir;
	tpLast = tp;
	for (tp2=treeIdx[tp->parentId]; tp2; tp2=treeIdx[tp2->parentId]) {
	    if (tp2->dir==dir && tp2->child[d2]==tpLast) break;
	    tpLast = tp2;
	}
	DASSERT(tp2);
#endif
	MSG(2, ("Mapping chan %d to %d\n", ch, tp2->order));
	MSG(2, ("Mapping treeId %d to %d\n", tp->id, tp2->id));

	for (k=0; k<numChan; k++)
	    if (oldToNewChanId[k]==ch)
		oldToNewChanId[k] = tp2->order;

	tp->dir = 'x';			/* Delete tp */
	oldId = tp->id;
	for (k=0; k<CDSHead.numTree; k++)
	    if (treeIdx[k]->id==oldId) treeIdx[k]->id = tp2->id;

	tp2 = treeIdx[tp->parentId];
	DASSERT(tp2);
	if (tp2->child[CDS_LO]==tp)
	then tp2->child[CDS_LO] = t2;
	else tp2->child[CDS_HI] = t2;
	t2->parentId = tp2->id;

	break;
    }

    DASSERT(ti<CDSHead.numTree);
}

if (*debugFlag>=2) {
    printf("After delete empty rooms\n");
    CDSPrintTree();
}

octInitGenContents(octFacet, OCT_INSTANCE_MASK, &genInst);
while (octGenerate(&genInst, &octInst)==OCT_OK) {
    if (!strncmp(ohGetName(&octInst), "emptyRoom", 9)) {
	fprintf(stdout,"WARN Deleting left over %s\n", ohGetName(&octInst));
	OH_ASSERT(octDelete(&octInst));
    }
}
					/* Adjust oldToNewChanId */
newChanId = ALLOC_ARRAY(numChan, short);
for (k=0; k<numChan; k++)
    newChanId[k] = k;
for (k=ch=0; k<numChan; k++) {
    if (oldToNewChanId[k]!=k) {
	newChanId[k] = -1;
	for (k2=0; k2<numChan; k2++)
	    if (newChanId[k2]>ch) newChanId[k2]--;
    }
    else ch++;
}
for (k=0; k<numChan; k++) {
    k2 = oldToNewChanId[k];
    DRANGE(0, newChanId[k2], numNewChan);
    oldToNewChanId[k] = newChanId[k2];
    MSG(2, ("OldChanId = %d, newId = %d\n", k, oldToNewChanId[k]));
}
FREE_ARRAY(numChan, newChanId, short);

/* Convert all references from old chanId to new chanId */

for (r=1; r<numReg+1; r++) {
    ch = regChanNum[r];
    if (ch<0) continue;
    DRANGE(0, ch, numChan);
    regChanNum[r] = oldToNewChanId[ch];
}

for (ti=0; ti<CDSHead.numTree; ti++) {
    t = treeIdx[ti];
    if (t->type==CDS_TREE_ELEMENT) continue;
    if (t->id!=ti) {
	DASSERT(t->dir=='x');
	DASSERT(treeIdx[t->id]->dir=='v' || treeIdx[t->id]->dir=='h');
	continue;
    }
    DASSERT(t->dir!='e');
    ch = t->order;
    DRANGE(0, ch, numChan);
    t->order = oldToNewChanId[ch];
}

/* Convert all references from old treeId to new treeId */

for (o=0; o<CDSHead.numObj; o++) {
    for (c=0; c<CDSObj[o].numCorn; c++) {
	oldId = CDSObj[o].corn[c].treeId;
	DASSERT(oldId!=CDS_NULL_TREEID);
	CDSObj[o].corn[c].treeId = treeIdx[oldId]->id;
    }
}

FREE_ARRAY(numChan, oldToNewChanId, short);
numChan = numNewChan;

MI_MOD_END();
} /* ............................. End of procedure hgDeleteEmptyRooms */
/* Convert net to net group
----------------------------------------------------------------------------*/
hgChanToOct(octFacet)

octObject	*octFacet;		/* Assume this has been opened */
{
octObject	octParent;
octObject	octChan, octInst, octChOrd;
octObject	octCorner;		/* Oct objects for corners */
octObject	octCEdge, octCEdgeChan;
octObject	octCProp, octECProp;
octObject	octPath;

octId		*treeOctId;

#define		MAX_OCT_PTS 1000
int		numOctPts;
struct octPoint	octPts[MAX_OCT_PTS];

int		stkRef;
short		tId, bdyIdx;
short		o, c, s;
CDSTreeDef	*t;

MI_MOD_START("hgChanToOct");

treeOctId = ALLOC_ARRAY(CDSHead.numTree, octId);

stkRef = MIStackPtr;
MI_STACK_PUSH(CDSTree);

while (MIStackPtr!=stkRef) {		/* Traverse tree just to create chan */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    MSG(1, ("Tree %d: type=%d \n", CDS_ID(t), t->type));
    DASSERT(t->type!=CDS_TREE_NOT_DONE);

    if (t->parentId==CDS_NULL_TREEID) {
	if (ohGetByBagName(octFacet, &octChan, "CDChanDef")==OCT_OK)
	    OH_ASSERT(octDelete(&octChan));
	OH_ASSERT(ohCreateBag(octFacet, &octChan, "CDChanDef"));
    }
    else {
	OH_ASSERT(ohGetById(&octParent, treeOctId[t->parentId]));
	OH_ASSERT(ohCreateBag(&octParent, &octChan, "CDChanDef"));
    }
    treeOctId[t->id] = octChan.objectId;

    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);
    }
    if (t->type==CDS_TREE_EXT_DONE) {
	if (!t->useExtFlag) {		/* External channel - no routing */
	    OH_ASSERT(ohCreatePropInt(&octChan, &octECProp,
		"CDOuterChannel", TRUE));
	}
	else {				/* External channel - routing allowed */
	    printf("Use %s external channel\n",
		(t->id==CDS_W_EXT_TREEID ? "left" :
		(t->id==CDS_E_EXT_TREEID ? "right" :
		(t->id==CDS_S_EXT_TREEID ? "bottom" : "top"))));
	}
    }
}

chanOctId = ALLOC_ARRAY(numChan, octId);

stkRef = MIStackPtr;
MI_STACK_PUSH(CDSTree);

while (MIStackPtr!=stkRef) {		/* Traverse tree again */
    MI_STACK_POP(CDSTreeDef*, t);
    if (!t) continue;
    MSG(1, ("Tree %d: type=%d order=%d\n", CDS_ID(t), t->type, t->order));
    DASSERT(t->type!=CDS_TREE_NOT_DONE);

    OH_ASSERT(ohGetById(&octChan, treeOctId[t->id]));

    if (t->type!=CDS_TREE_ELEMENT) {	/* Not a leaf cell */
	MI_STACK_PUSH(t->child[CDS_LO]);
	MI_STACK_PUSH(t->child[CDS_HI]);

	OH_ASSERT(ohCreatePropInt(&octChan, &octChOrd,
	    "CDChanOrder", t->order));
	chanOctId[t->order] = octChan.objectId;

	if (t->type>0 || t->useExtFlag) { /* Init split path */
	    if (t->useExtFlag)
	    then numOctPts = 2;
	    else numOctPts = (int)t->type;
	    URANGE(numOctPts, MAX_OCT_PTS);
	    for (s=0; s<numOctPts; s++) {
		octPts[s].x = (octCoord)t->slice[s][CDS_X];
		octPts[s].y = (octCoord)t->slice[s][CDS_Y];
		MSG(3, ("Slice %2d: (%d %d)\n", s, octPts[s]));
	    }
					/* Attach path to channel */
	    OH_ASSERT(ohCreatePath(&octChan, &octPath, 3));
	    OH_ASSERT(octPutPoints(&octPath, numOctPts, octPts));
	}
    }
    else {				/* Leaf cell: */
	o = t->objNum;
	if (o>=0) {			/* Attach instance to leaf channel */
	    MSG(1, ("Put corners on inst %s\n", CDSObj[o].dbRef));
	    OH_ASSERT(ohGetByInstName(octFacet, &octInst,
		(String)CDSObj[o].dbRef));
	    OH_ASSERT(octAttach(&octChan, &octInst));
					/* Attach corner info to inst */
	    if (ohGetByBagName(&octInst, &octCorner, "CDCorners")==OCT_OK)
		OH_ASSERT(octDelete(&octCorner));
	    OH_ASSERT(ohCreateBag(&octInst, &octCorner, "CDCorners"));
					/* Each corner */
	    for (c=0; c<CDSObj[o].numCorn; c++) {
		if (CDSObj[o].bdyOrient==CDS_CW)
		then bdyIdx = CDSObj[o].numBdy-CDSObj[o].corn[c].bdyIdx-1;
		else bdyIdx = CDSObj[o].corn[c].bdyIdx;
		MSG(2, ("Corner %d = bdy %d\n", c, bdyIdx));
		OH_ASSERT(ohCreatePropInt((c==0 ? &octCorner : &octCEdge),
		    &octCProp, "CDCorner", bdyIdx));
					/* Attach to channel */
		tId = CDSObj[o].corn[c].treeId;
		DASSERT(tId!=CDS_NULL_TREEID);
		OH_ASSERT(ohGetById(&octCEdgeChan, treeOctId[tId]));
		OH_ASSERT(ohCreateBag(&octCEdgeChan, &octCEdge, "CDCornerEdge"));
		OH_ASSERT(octAttach(&octCEdge, &octCProp));
	    }
	}
    }
}

FREE_ARRAY(CDSHead.numTree, treeOctId, octId);

MI_MOD_END();
} /* ............................. End of procedure hgChanToOct */
