/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

hgOutOct.c

hgOutOct	Output to oct facet

Author(s):		Mitsuru Igusa
Revision history:	$Log: hgOutOct.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:49  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.1  90/02/07  17:03:29  octtools
 * Checkin for 3.5 distribution
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CDSInc.h"
#include "mercury.h"

typedef	struct	hgNetGrpDef	hgNetGrpDef;
typedef	struct	hgPTermDef	hgPTermDef;
typedef	struct	hgATermDef	hgATermDef;
typedef	struct	hgNetRegDef	hgNetRegDef;

struct hgNetGrpDef {
    short	chanNum;	/* Channel associated with net */
    hgPTermDef	*pTerm;		/* Used initially for group to group links */
    hgATermDef	*aTerm;
    hgNetGrpDef	*next;		/* Doubly linked list of net groups */
    hgNetGrpDef	*prev;
};

#define	DIM_ADJ_NET_REG 10
struct hgNetRegDef {
    int		regId;		/* Region id */
    short	chanNum;
    short	numAdjNetReg;
    hgNetGrpDef	*netGrp;
    hgATermDef	*aTerm;
    hgNetRegDef	*adjNetReg[DIM_ADJ_NET_REG];
    hgNetRegDef	*next;		/* Doubly linked list for easy deleting */
    hgNetRegDef	*prev;
};

struct hgPTermDef {
    octId	dbRef;		/* Data base reference */
    hgNetGrpDef	*parentNetGrp;	/* Parent net group */
    hgPTermDef	*otherPTerm;	/* Other channel pterm */
    hgPTermDef	*next;		/* Linked list of pterms */
};

struct hgATermDef {
    octId	dbRef;		/* Data base reference */
    hgATermDef	*next;		/* Linked list of aterms */
};

#define HG_NEW_NETGRP(grp, grpHead) { \
    grp = ALLOC_TYPE(hgNetGrpDef); \
    if (grpHead) \
	(grpHead)->prev = grp; \
    grp->next = grpHead; \
    grp->prev = NULL; \
    grp->aTerm = NULL; \
    grp->pTerm = NULL; \
    grp->chanNum = -1; \
    grpHead = grp; \
}

#define HG_DELETE_NETGRP(grp, grpHead) { \
    if (grp->prev) \
	 grp->prev->next = grp->next; \
    else grpHead         = grp->next; \
    if (grp->next) \
	 grp->next->prev = grp->prev; \
    FREE_TYPE(grp, hgNetGrpDef); \
}

#define HG_DELETE_PTERM(pTermX) { \
    hgNetGrpDef	*netGrp99; \
    hgPTermDef	*prevPTerm99, *pTerm99; \
\
    netGrp99 = pTermX->parentNetGrp;	/* Find PTermX in netGrp pTerm list */ \
    prevPTerm99 = NULL; \
    for (pTerm99=netGrp99->pTerm; pTerm99; pTerm99=pTerm99->next) { \
	if (pTerm99==pTermX) break; \
	prevPTerm99 = pTerm99; \
    } \
    DASSERT(pTerm99); \
					/* Delete PTerm from list */ \
    if (prevPTerm99) \
    then prevPTerm99->next = pTerm99->next; \
    else netGrp99->pTerm = pTerm99->next; \
\
    FREE_TYPE(pTermX, hgPTermDef); \
}

#define HG_NEW_ATERM(grp, aTermDbRef) { \
    hgATermDef	*hgATerm99; \
\
    hgATerm99 = ALLOC_TYPE(hgATermDef); \
    hgATerm99->dbRef = (octId)aTermDbRef; \
    hgATerm99->next = (grp)->aTerm; \
    (grp)->aTerm = hgATerm99; \
}

#define HG_NEW_PTERM(grp1, grp2) { \
    hgPTermDef	*hgPTerm91,  *hgPTerm92; \
\
    hgPTerm91 = ALLOC_TYPE(hgPTermDef); \
    hgPTerm92 = ALLOC_TYPE(hgPTermDef); \
    hgPTerm91->dbRef = oct_null_id; \
    hgPTerm91->parentNetGrp = grp1; \
    hgPTerm91->next = (grp1)->pTerm; \
    hgPTerm91->otherPTerm = hgPTerm92; \
    hgPTerm92->dbRef = oct_null_id; \
    hgPTerm92->parentNetGrp = grp2; \
    hgPTerm92->next = (grp2)->pTerm; \
    hgPTerm92->otherPTerm = hgPTerm91; \
    (grp1)->pTerm = hgPTerm91; \
    (grp2)->pTerm = hgPTerm92; \
}

/* TODO: Initialize these two arrays */
int		numReg, numChan;
octId		*chanOctId;
short		*regChanNum;

/* Mercury output
----------------------------------------------------------------------------*/
hgOutOct(octFacet)

octObject	*octFacet;		/* Assume this has been opened */
{
hgATermDef	*hgATerm, *nextHgATerm;
hgPTermDef	*hgPTerm, *nextHgPTerm, *hgPTerm2;
hgNetGrpDef	*netGrp, *nextNetGrp, *netGrpHead;
octObject	octNet, octPTerm, octATerm, octSubNet, octChan;
GNetDef		*net;

int		numNet, numTrivialNet, numOnePinNet;

MI_MOD_START("hgOutOct");

numNet = numTrivialNet = numOnePinNet = 1;

for (net=GNet; net; net=net->next) {
    OH_ASSERT(ohGetById(&octNet, (octId)net->dbRef));
    MSG(-1, ("Converting net %s\n", ohGetName(&octNet)));
    numNet++;

    /* Create net groups */
    hgConvtToNetGrp(net, &netGrpHead);

    /* Trim unnecessary stuff from net groups */
    hgTrimNetGrp(&netGrpHead);

    if (netGrpHead && !netGrpHead->next)
	numTrivialNet++;

    /* Create oct data structures for each net sub group */
    for (netGrp=netGrpHead; netGrp; netGrp=netGrp->next) {
	if (!netGrp->pTerm && !netGrp->aTerm) continue;
	if (!netGrp->pTerm && !netGrp->aTerm->next) {
	    numOnePinNet++;
	    continue;
	}

	OH_ASSERT(ohCreateNet(&octNet, &octSubNet, "CDSubnet"));
	OH_ASSERT(ohGetById(&octChan, chanOctId[netGrp->chanNum]));
	OH_ASSERT(octAttach(&octChan, &octSubNet));
					/* Attach aTerms */
	for (hgATerm=netGrp->aTerm; hgATerm; hgATerm=hgATerm->next) {
	    octATerm.objectId = hgATerm->dbRef;
	    OH_ASSERT(octAttach(&octSubNet, &octATerm));
	}
					/* Create and/or attach pTerms */
	for (hgPTerm=netGrp->pTerm; hgPTerm; hgPTerm=hgPTerm->next) {
	    if (!hgPTerm->dbRef) {
		OH_ASSERT(ohCreateBag(&octSubNet, &octPTerm, "CDPseudoTerm"));
		hgPTerm2 = hgPTerm->otherPTerm;
		ASSERT(!hgPTerm2->dbRef);
		hgPTerm ->dbRef = octPTerm.objectId;
		hgPTerm2->dbRef = octPTerm.objectId;
	    }
	    else {
		OH_ASSERT(ohGetById(&octPTerm, hgPTerm->dbRef));
		OH_ASSERT(octAttach(&octSubNet, &octPTerm));
	    }
	}
    }

    /* Free netGrp data structures */
    for (netGrp=netGrpHead; netGrp; netGrp=nextNetGrp) {
	for (hgATerm=netGrp->aTerm; hgATerm; hgATerm=nextHgATerm) {
	    nextHgATerm = hgATerm->next;
	    FREE_TYPE(hgATerm, hgATermDef);
	}
	for (hgPTerm=netGrp->pTerm; hgPTerm; hgPTerm=nextHgPTerm) {
	    nextHgPTerm = hgPTerm->next;
	    FREE_TYPE(hgPTerm, hgPTermDef);
	}
	nextNetGrp = netGrp->next;
	FREE_TYPE(netGrp, hgNetGrpDef);
    }
}

FREE_ARRAY(numReg, regChanNum, short);
FREE_ARRAY(numChan, chanOctId, octId);

numTrivialNet -= numOnePinNet;
MSG(-1, ("Num net = %d, num trivial = %d, num one pin = %d\n",
    numNet, numTrivialNet, numOnePinNet));

MI_MOD_END();
} /* ............................. End of procedure hgOutOct */
/* Convert net to net group
----------------------------------------------------------------------------*/
hgConvtToNetGrp(net, netGrpHead)

GNetDef		*net;
hgNetGrpDef	**netGrpHead;
{
GPinDef		*pin;
routeType	*rt;
pathType	*path;
nodeGroupType	*node;
int		regId;

int		a, a1, a2, a3;
hgNetRegDef	*netRegHead, *netReg, *adjNetReg, *adjNetReg2;
hgNetRegDef	*lastNetReg, *prevNetReg;
hgNetRegDef 	*hgNewNetReg();
hgNetGrpDef	*netGrp, *netGrp2;
hgATermDef	*hgATerm;
GRegionDef	*reg;
GRegionDef	*getNextRegion();

MI_MOD_START("hgConvtToNetGrp");

*netGrpHead = NULL;

if (net->trivial ||			/* Trivial one channel net */
    net->netGr->NoRegion==1) {		/* Another way to check if trivial */
    HG_NEW_NETGRP(netGrp, *netGrpHead);
    pin = net->pin;
    DASSERT(pin);
    regId = pin->region->id;
    netGrp->chanNum = regChanNum[regId];

    for ( ; pin; pin=pin->next)
	HG_NEW_ATERM(netGrp, pin->dbRef);
    goto hgConvtToNetGrpEnd;
}

DRANGE(0, net->routeNo, G_MAXROUTE);
rt = net->netGr->candRoute[net->routeNo];
DASSERT(rt);

netRegHead = NULL;			/* Make a tree with the paths */
prevNetReg = NULL;
for (path=rt->path; path; path=path->next) {
    lastNetReg = NULL;

#define GET_NEXT_REG
#ifdef GET_NEXT_REG
    for (node=path->node; reg=getNextRegion(node); node=NULL) {
	regId = reg->id;
					/* Check for intersection */
	for (netReg=netRegHead; netReg; netReg=netReg->next)
	    if (regId==netReg->regId) break;

	if (!netReg) {			/* No intersection: create new reg */
	    netReg = hgNewNetReg(regId, &prevNetReg, &netRegHead);
	}

	if (lastNetReg) {		/* Connect to last reg in path */
	    DURANGE(lastNetReg->numAdjNetReg, DIM_ADJ_NET_REG);
	    lastNetReg->adjNetReg[lastNetReg->numAdjNetReg++] = netReg;
	    DURANGE(netReg->numAdjNetReg, DIM_ADJ_NET_REG);
	    netReg->adjNetReg[netReg->numAdjNetReg++] = lastNetReg;
	}
	lastNetReg = netReg;
    }
#else
{
GRegionDef	*adjReg;
nodeGroupType	*prevNode;
GRegionDef	*lastReg;
Boolean		foundFlg;

    prevNode = NULL;
    for (node=path->node; node; node=node->next) {

	regId = node->id;
					/* Check for intersection */
	for (netReg=netRegHead; netReg; netReg=netReg->next)
	    if (regId==netReg->regId) break;

	if (!netReg) {			/* No intersection: create new reg */
	    netReg = hgNewNetReg(regId, &prevNetReg, &netRegHead);
	}

	if (lastNetReg) {		/* Connect to last reg in path */
	    reg     = GID2Reg[netReg    ->regId-1];
	    lastReg = GID2Reg[lastNetReg->regId-1];
	    DASSERT(reg->id==regId);

	    foundFlg = TRUE;
	    for (k=0; k<G_NDIR; k++)
		if (lastReg->adjReg[k]==reg) break;
	    if (k==G_NDIR) {
		for (k=0; k<G_NDIR; k++)
		    if (reg->adjReg[k]==lastReg) break;
		if (k==G_NDIR)
		    foundFlg = FALSE;
	    }

	    if (!foundFlg) {		/* Fill in adjacent bottleneck reg */
		for (k=0; k<G_NDIR; k++) {
		    adjReg = lastReg->adjReg[k];
		    if (adjReg==reg->adjReg[G_REVDIR(k)] &&
			(adjReg->type==G_HOR_BN || adjReg->type==G_VER_BN))
			break;
		}
		if (k==G_NDIR) {
		    MIerror("Cannot find inbetween reg %d %d\n",
			regId, lastNetReg->regId);
		}
		adjNetReg = hgNewNetReg(adjReg->id, &prevNetReg, &netRegHead);
		DURANGE(lastNetReg->numAdjNetReg, DIM_ADJ_NET_REG);
		lastNetReg->adjNetReg[lastNetReg->numAdjNetReg++] = adjNetReg;
		adjNetReg ->adjNetReg[adjNetReg ->numAdjNetReg++] = lastNetReg;
		lastNetReg = adjNetReg;

	    }
	    DURANGE(lastNetReg->numAdjNetReg, DIM_ADJ_NET_REG);
	    lastNetReg->adjNetReg[lastNetReg->numAdjNetReg++] = netReg;
	    DURANGE(netReg->numAdjNetReg, DIM_ADJ_NET_REG);
	    netReg->adjNetReg[netReg->numAdjNetReg++] = lastNetReg;
	}
	lastNetReg = netReg;

	prevNode = node;
    }
}
#endif
}

for (pin=net->pin; pin; pin=pin->next) { /* Add aterms to reg */
    regId = pin->region->id;
    for (netReg=netRegHead; netReg; netReg=netReg->next)
	if (netReg->regId==regId) break;
    if (netReg) {
	HG_NEW_ATERM(netReg, pin->dbRef);
    }
    else {
	MIerror("Cannot find reg %d for aterm\n", regId);
	abort();
    }
}

do {
    for (netReg=netRegHead; netReg; netReg=netReg->next) {
	for (a=0; a<netReg->numAdjNetReg; a++) {
	    adjNetReg = netReg->adjNetReg[a];
	    if (netReg->chanNum==adjNetReg->chanNum) break;
	}
	if (a==netReg->numAdjNetReg) continue;
					/* Adjacent region of same chan found */
					/*   merge adjNetReg into netReg */

	netReg->adjNetReg[a] = netReg->adjNetReg[--netReg->numAdjNetReg];

					/* Merge aTerms */
	hgATerm = adjNetReg->aTerm;
	if (hgATerm) {
	    for ( ; hgATerm->next; hgATerm=hgATerm->next);
	    hgATerm->next = netReg->aTerm;
	    netReg->aTerm = adjNetReg->aTerm;
	}
					/* Merge adjNetReg */
	for (a2=0; a2<adjNetReg->numAdjNetReg; a2++) {
	    adjNetReg2 = adjNetReg->adjNetReg[a2];
	    if (adjNetReg2==netReg) continue;
	    DASSERT(adjNetReg!=adjNetReg2);

	    for (a1=0; a1<netReg->numAdjNetReg; a1++)
		if (netReg->adjNetReg[a1]==adjNetReg2) break;
	    if (a1<netReg->numAdjNetReg)
		fprintf(stderr,
		    "ERROR Reg %d adjacent to both %d and %d\n",
		    adjNetReg2->regId, netReg->regId, adjNetReg->regId);
	    else {
		DURANGE(netReg->numAdjNetReg, DIM_ADJ_NET_REG);
		netReg->adjNetReg[netReg->numAdjNetReg++] = adjNetReg2;
	    }

	    for (a3=0; a3<adjNetReg2->numAdjNetReg; a3++)
		if (adjNetReg2->adjNetReg[a3]==adjNetReg) break;
	    if (a3<adjNetReg2->numAdjNetReg)
		adjNetReg2->adjNetReg[a3] = netReg;
	    else
		fprintf(stderr,
		    "ERROR Reg %d adjacent to %d but not vice versa\n",
		    adjNetReg->regId, adjNetReg2->regId);
	}

					/* Unlink and free adjNetReg */
	DASSERT(adjNetReg->prev);
	adjNetReg->prev->next = adjNetReg->next;
	if (adjNetReg->next)
	    adjNetReg->next->prev = adjNetReg->prev;
	FREE_TYPE(adjNetReg, hgNetRegDef);
	break;
    }
} while (netReg);			/* While some regions were merged */

for (netReg=netRegHead; netReg; netReg=netReg->next) {
    HG_NEW_NETGRP(netGrp, *netGrpHead);
    netReg->netGrp = netGrp;
    netGrp->chanNum = netReg->chanNum;
    netGrp->aTerm = netReg->aTerm;
    netGrp->pTerm = NULL;
    for (a=0; a<netReg->numAdjNetReg; a++) {
	netGrp2 = netReg->adjNetReg[a]->netGrp;
	if (netGrp2)
	    HG_NEW_PTERM(netGrp, netGrp2);
    }
}
					/* Free up temp data structures */
for (netReg=netRegHead; netReg; netReg=lastNetReg) {
    lastNetReg = netReg->next;
    FREE_TYPE(lastNetReg, hgNetRegDef);
}

hgConvtToNetGrpEnd:
MI_MOD_END();
} /* ............................. End of procedure hgConvtToNetGrp */
/* Trim net group.
----------------------------------------------------------------------------*/
hgTrimNetGrp(netGrpHead)

hgNetGrpDef	**netGrpHead;
{
hgPTermDef	*hgPTerm, *nextHgPTerm, *prevHgPTerm;
hgPTermDef	*hgPTerm2, *hgPTermX, *pTerm;
hgATermDef	*aTerm;
hgNetGrpDef	*netGrp, *netGrp2, *netGrpX, *nextNetGrp;
hgPTermDef	**chanPTerm;
short		c, numPTerm;
Boolean		changeFlg, changedFlg;

MI_MOD_START("hgTrimNetGrp");

chanPTerm = ALLOC_ARRAY(numChan, hgPTermDef *);

for (netGrp= *netGrpHead; netGrp; netGrp=netGrp->next) {
    MSG(3, ("Printing net group %d ch=%d\n", netGrp, netGrp->chanNum));
    for (pTerm=netGrp->pTerm; pTerm; pTerm=pTerm->next) {
	hgPTermX = pTerm->otherPTerm;
	DASSERT(hgPTermX->otherPTerm==pTerm);
	netGrpX = hgPTermX->parentNetGrp;
	MSG(3, ("PTerm %d other pterm=%d netGrp=%d ch=%d\n",
	    pTerm, hgPTermX, netGrpX, netGrpX->chanNum));
    }
}

changedFlg = FALSE;
do {
    changeFlg  = FALSE;
    for (netGrp= *netGrpHead; netGrp; netGrp=nextNetGrp) {
	nextNetGrp = netGrp->next;

	for (c=0; c<numChan; c++)
	    chanPTerm[c] = NULL;

	prevHgPTerm = NULL;
	for (hgPTerm=netGrp->pTerm; hgPTerm; hgPTerm=nextHgPTerm) {
	    hgPTermX = hgPTerm->otherPTerm;
	    netGrpX = hgPTermX->parentNetGrp;
	    DASSERT(hgPTermX->otherPTerm==hgPTerm);
	    hgPTerm2 = chanPTerm[netGrpX->chanNum];

	    if (!hgPTerm2)
		chanPTerm[netGrpX->chanNum] = hgPTermX;
	    else {
		MSG(-1, ("Deleting duplicate pterm on chan %d - %d\n",
		    netGrp->chanNum, netGrpX->chanNum));
		DASSERT(hgPTerm2!=hgPTermX);
		netGrp2 = hgPTerm2->parentNetGrp;

		changeFlg = TRUE;
		changedFlg = TRUE;
		HG_DELETE_PTERM(hgPTerm);
		hgPTerm = prevHgPTerm;
		HG_DELETE_PTERM(hgPTermX);
					    /* Merge netGrpX into netGrp2 */
		if (netGrpX->aTerm) {
		    for (aTerm=netGrpX->aTerm; aTerm->next; aTerm=aTerm->next);
		    aTerm->next = netGrp2->aTerm;
		    netGrp2->aTerm = netGrpX->aTerm;
		}
		if (netGrpX->pTerm) {
		    for (pTerm=netGrpX->pTerm; pTerm->next; pTerm=pTerm->next)
			pTerm->parentNetGrp = netGrp2;
		    pTerm->parentNetGrp = netGrp2;
		    pTerm->next = netGrp2->pTerm;
		    netGrp2->pTerm = netGrpX->pTerm;
		}
		if (netGrpX==nextNetGrp)
		    nextNetGrp = netGrpX->next;
		HG_DELETE_NETGRP(netGrpX, *netGrpHead);
	    }

	    prevHgPTerm = hgPTerm;
	    if (hgPTerm)
	    then nextHgPTerm = hgPTerm->next;
	    else nextHgPTerm = netGrp->pTerm;
	}

	if (netGrp->aTerm) continue;

	numPTerm = 0;
	for (hgPTerm=netGrp->pTerm; hgPTerm; hgPTerm=hgPTerm->next) numPTerm++;

	if (numPTerm==1) {			/* Degenerate pTerm */
	    MSG(-1, ("Deleting degenerate pterm on chan %d\n", netGrp->chanNum));

	    hgPTerm = netGrp->pTerm;
	    hgPTermX = hgPTerm->otherPTerm;
	    DASSERT(hgPTermX->otherPTerm==hgPTerm);
					    /* Delete hgPTerm from pTerm list */
	    HG_DELETE_PTERM(hgPTerm);
	    HG_DELETE_PTERM(hgPTermX);
	}
    }
} while(changeFlg);

if (changedFlg)
    for (netGrp= *netGrpHead; netGrp; netGrp=netGrp->next) {
	MSG(2, ("Printing net group %d ch=%d\n", netGrp, netGrp->chanNum));
	for (pTerm=netGrp->pTerm; pTerm; pTerm=pTerm->next) {
	    hgPTermX = pTerm->otherPTerm;
	    DASSERT(hgPTermX->otherPTerm==pTerm);
	    netGrpX = hgPTermX->parentNetGrp;
	    MSG(2, ("PTerm %d other pterm=%d netGrp=%d ch=%d\n",
		pTerm, hgPTermX, netGrpX, netGrpX->chanNum));
	}
    }

FREE_ARRAY(numChan, chanPTerm, hgPTermDef *);

MI_MOD_END();
} /* ............................. End of procedure hgTrimNetGrp */
/* Create and initialize new net reg data structure.
----------------------------------------------------------------------------*/
hgNetRegDef *
hgNewNetReg(regId, prevNetReg, netRegHead)

int		regId;
hgNetRegDef	**prevNetReg;
hgNetRegDef	**netRegHead;
{
hgNetRegDef	*netReg;

netReg = ALLOC_TYPE(hgNetRegDef);

netReg->regId = regId;
netReg->netGrp = NULL;
netReg->chanNum = regChanNum[regId];
netReg->aTerm = NULL;
netReg->numAdjNetReg = 0;

netReg->next = NULL;
netReg->prev = *prevNetReg;
if (*prevNetReg)
    (*prevNetReg)->next = netReg;
else
    *netRegHead = netReg;

*prevNetReg = netReg;
return(netReg);
} /* ............................. End of procedure hgNewNetReg */
/* Create and initialize new net reg data structure.
----------------------------------------------------------------------------*/
hgUpdateLoc(octFacet)

octObject	*octFacet;
{
GRegionDef	*r;
octObject	octInst;
String		instName;
int		deltaLocX, deltaLocY;
Boolean		moveFlg;
CDSObjDef	*obj;

MI_MOD_START("hgUpdateLoc");

for (r=GRegList; r; r=r->next) {
    if (r->type!=G_INST) continue;

    r->newLocX += CDSHgOffsetX;
    r->newLocY += CDSHgOffsetY;
    deltaLocX = r->newLocX - G_REGBOX_LL_X(r);
    deltaLocY = r->newLocY - G_REGBOX_LL_Y(r);

					/* Check if change in location */
    if (deltaLocX==0 && deltaLocY==0) continue;
    moveFlg = TRUE;

					/* Check for special instances */
    obj = (CDSObjDef *)r->dbRef;
    instName = (String)obj->dbRef;
    if (!instName ||
	!strcmp (instName,"dummyIo") ||
	!strncmp(instName,"emptyRoom",9))
	moveFlg = FALSE;

    MSG(-1, ("%s inst %s from (%d %d) to (%d %d)\n",
	(moveFlg ? "Moving" : "Not moving"), instName,
	G_REGBOX_LL_X(r), G_REGBOX_LL_Y(r), r->newLocX, r->newLocY));

    if (!moveFlg) continue;

    OH_ASSERT(ohGetByInstName(octFacet, &octInst, instName));
    octInst.contents.instance.transform.translation.x += deltaLocX;
    octInst.contents.instance.transform.translation.y += deltaLocY;
    OH_ASSERT(octModify(&octInst));
}

MI_MOD_END();
} /* ............................. End of procedure hgUpdateLoc */
