/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPOct.c

Interface of channel preparation for channel router to OCT database.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CPOct.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:51  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.5  90/02/07  16:13:34  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.3  89/02/07  13:21:05  octtools
 * got rid of routeChan
 * 
 * Revision 1.2  89/02/06  16:44:09  octtools
 * fixed instance naming problem (added ../)
 * 
 * Revision 1.1  89/01/26  14:20:12  octtools
 * Initial revision
 * 
 * Revision 1.10  88/06/15  14:06:16  mitch
 * Consider also OCT_BAGs when copying the contents of aTerms
 * 
 * Revision 1.6  88/04/24  16:45:39  mitch
 * Changed from MIOctOpenMasterOfInst to octOpenMaster.
 * Attach instance to CONNECTOR bag if originally a connector.
 * 
 * Revision 1.4  88/04/08  16:34:00  mitch
 * Deleted fterm has no net error message.
 * Fixed strange bug in pin direction initialization.
 * 
 * Revision 1.3  88/04/08  12:18:32  mitch
 * Moved call to octBegin to main procedure.
 * 
 * Revision 1.2  88/04/05  15:59:34  mitch
 * Added pseudo pin direction stuff.
 * 
 * Revision 1.1  88/03/31  11:22:20  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CPInc.h"
#include "vov.h"

/* Get oct data
----------------------------------------------------------------------------*/
CPOctInput(cellName, view, newView)

String		cellName;
String		view;
String		newView;
{
    octObject	octOldMaster;
    octObject	octChan;

    octObject	octLayers[3];	/* This is to call set_up_width...() */
    /* Used in new pin location alg. */
    MI_MOD_START("CPOctInput");
    /* Copy contents facet */
    if (strcmp(view,newView)) {
	MSG(1, ("Copying contents facet <%s> view=<%s> to view=<%s>\n",
		cellName, view, newView));
	octOldMaster.type = OCT_FACET;
	octOldMaster.contents.facet.cell  = cellName;
	octOldMaster.contents.facet.view  = view;
	octOldMaster.contents.facet.facet = "contents";
	octOldMaster.contents.facet.version = OCT_CURRENT_VERSION;
	CPOctMaster = octOldMaster;
	CPOctMaster.contents.facet.view = newView;
	VOVinputFacet( &octOldMaster );
	VOVoutputFacet( &CPOctMaster );
	OH_ASSERT(octCopyFacet(&CPOctMaster, &octOldMaster));
	

	MSG(1, ("Copying interface facet <%s> view=<%s> to view=<%s>\n",
		cellName, view, newView));
	octOldMaster.contents.facet.facet = "interface";
	CPOctMaster = octOldMaster;
	CPOctMaster.contents.facet.view = newView;
	VOVinputFacet( &octOldMaster );
	VOVoutputFacet( &CPOctMaster );
	OH_ASSERT(octCopyFacet(&CPOctMaster, &octOldMaster));
    }
    /* Open new contents facet */
    OH_ASSERT(ohOpenFacet(&CPOctMaster, cellName, newView, "contents", "a"));
    /* Init CPChanId[] in routing order */
    OH_ASSERT(ohGetByBagName(&CPOctMaster, &octChan, "CDChanDef"));
    CPChanOrder(&octChan);

    if (ohGetByPropName(&CPOctMaster, &CPOctTech, "TECHNOLOGY")!=OCT_OK) {
	MIerror("no %s property on cell to be routed %s %s\n",
		CPOctTech.contents.prop.name, cellName, view);
	exit(8);
    }

    OH_ASSERT(ohGetOrCreateLayer(&CPOctMaster, &octLayers[0], "POLY"));
    OH_ASSERT(ohGetOrCreateLayer(&CPOctMaster, &octLayers[1], "MET1"));
    OH_ASSERT(ohGetOrCreateLayer(&CPOctMaster, &octLayers[2], "MET2"));

    SPinit( &CPOctMaster );
    set_up_width_and_spacing(octLayers);

    MI_MOD_END();
}				/* ............................. End of procedure CPOctInput */
/* Initialize new submaster composing of two instances + channel inbetween.
----------------------------------------------------------------------------*/
CPInitSubMaster(octMaster, octChanObj)

octObject	*octMaster;
octObject	*octChanObj;
{
    octObject	octParIInstBag, octProp;
    octObject	octCInstBag, octIInstBag;
    octObject	octInst, octInstCopy;
    octObject	octChanLeaf;
    octObject	octATerm, octATermCopy;
    octObject	octFTerm, octFTermCopy, octFTermNet;
    octObject	octNet, octSubNet, octSubNet2;
    octObject	octPTerm, octSideTermOrder;
    octObject	*octInstMaster;
    octGenerator	genChanLeaf;
    octGenerator	genATerm;
    octGenerator	genSubNet;
    octGenerator	genPTerm;

#define		MAX_OCT_PTS 1000
    int32		numOctPts;
    struct octPoint	octPts[MAX_OCT_PTS];
    Boolean		orientFlg;

    octObject	octCornHead, octCorn, octCEdge;
    octObject	octCornChan;
    octGenerator	genCorn;
    short		b, c, numCorn;
    octId		prevCornOctId;
    Boolean		foundFlg;
    CPCornDef	corn[CP_MAX_NUM_CORN];

    octObject	octSubNetChan;
    short		k, p, n, sn;
    short		chanNum;
    CPPinDef	*pin;
    short		*numSubNetPins;

    MI_MOD_START("CPInitSubMaster");

    OH_ASSERT(ohGetOrCreatePropStr(octMaster, &octProp, "EDITSTYLE", "SYMBOLIC"));
    OH_ASSERT(ohGetOrCreatePropStr(octMaster, &octProp, "VIEWTYPE" , "SYMBOLIC"));
    OH_ASSERT(ohGetOrCreatePropStr(octMaster, &octProp, "CELLCLASS", "MODULE"));
    OH_ASSERT(octGetOrCreate(octMaster, &CPOctTech));
    OH_ASSERT(ohGetOrCreateBag(octMaster, &octCInstBag, "CONNECTORS"));
    OH_ASSERT(ohGetOrCreateBag(octMaster, &octIInstBag, "INSTANCES"));
    octParIInstBag = octIInstBag;

    MSG(0, ("Channel preparation for channel %s\n", octMaster->contents.facet.cell));
    /* Init subnet/net array */
    OH_ASSERT(octInitGenContents(octChanObj, OCT_NET_MASK, &genSubNet));
    for (sn=0; octGenerate(&genSubNet, &octSubNet)==OCT_OK; sn++)
      DASSERT(!strcmp(octSubNet.contents.net.name, "CDSubnet"));
    MSG(2, ("Number of subnets in channel = %d\n", sn));
    CPNumSubNet = sn;
    CPNet = ALLOC_ARRAY(CPNumSubNet, CPNetDef);

    OH_ASSERT(octInitGenContents(octChanObj, OCT_NET_MASK, &genSubNet));
    CPNumNet = 0;
    for (sn=0; octGenerate(&genSubNet, &octSubNet)==OCT_OK; sn++) {
	OH_ASSERT(octGenFirstContainer(&octSubNet, OCT_NET_MASK, &octNet));
	for (n=0; n<CPNumNet; n++) /* Subnet <--> net not 1 to 1 */
	  if (octIdsEqual(octNet.objectId, CPNet[n].netId)) break;
	CPNet[sn].subNetId = octSubNet.objectId;
	CPNet[sn].netNum   = n;
	CPNet[n].dupFlg = (n!=CPNumNet); /* Multiple subnets under net */
	if (n==CPNumNet) {
	    CPNet[CPNumNet++].netId = octNet.objectId;
	    CPNet[n].name = STR_COPY(octNet.contents.net.name);
	}
	MSG(3, ("Subnet %d: net=%d (octSubNetId=%d, octNetId=%d)\n",
		sn, n, CPNet[sn].subNetId, CPNet[n].netId));
    }
    octFreeGenerator( &genSubNet );

    octCEdge.type = OCT_BAG;
    octCEdge.contents.bag.name = "CDCornerEdge";
    octCornChan.type = OCT_BAG;
    octCornChan.contents.bag.name = "CDChanDef";
    octSubNet.type = OCT_NET;
    octSubNet.contents.net.name = "CDSubnet";

    OH_ASSERT(octInitGenContents(octChanObj, OCT_BAG_MASK, &genChanLeaf));

    /* Each Instance bordering channel: */
    for (k=0; octGenerate(&genChanLeaf, &octChanLeaf)==OCT_OK; k++) {
	/* Skip non-channel bags */
	if (strcmp(octChanLeaf.contents.bag.name, "CDChanDef")) {
	    k--;
	    continue;
	}

	OH_ASSERT(octGenFirstContent(&octChanLeaf, OCT_INSTANCE_MASK, &octInst));
	/* Create new instance */
	OH_ASSERT(ohGetById(&octInstCopy, octInst.objectId));

	ohBestName(octMaster, &octInstCopy, NIL(octObject));

	octInstCopy.contents.instance.facet = "interface"; /* To speed up all */
	OH_ASSERT(octCreate(octMaster, &octInstCopy));

	if (MIOctGetByNameContainers(&octInst, &octParIInstBag)==OCT_OK) {
	    OH_ASSERT(octAttach(&octIInstBag, &octInstCopy));
	}
	else
	  OH_ASSERT(octAttach(&octCInstBag, &octInstCopy));
	/* Init CPInst[].xxxInstId */
	CPInst[k].name = STR_COPY(octInst.contents.instance.name);
	CPInst[k].oldInstId = octInst    .objectId;
	CPInst[k].newInstId = octInstCopy.objectId;
	MSG(2, ("Copying instance%d %s\n", k, CPInst[k].name));
	/* Init CPInst[].bdy */
	octInstMaster = &CPInst[k].octInstMaster;
	OH_ASSERT(ohOpenMaster(octInstMaster, &octInst, "interface", "r"));
	VOVinputFacet( octInstMaster );
	numOctPts = MAX_OCT_PTS;
	MIOctGetInstBdy(octInstMaster, &octInst, octPts, &numOctPts, &orientFlg);

	CPInst[k].numBdy    = numOctPts;
	CPInst[k].bdy       = ALLOC_ARRAY(numOctPts, CPLocDef);
	CPInst[k].bdyOrient = (orientFlg ? CP_CW : CP_CCW);
	CPInst[k].loc.x     = octInst.contents.instance.transform.translation.x;
	CPInst[k].loc.y     = octInst.contents.instance.transform.translation.y;
	for (p=0; p<numOctPts; p++) {
	    CPInst[k].bdy[p].x = octPts[p].x;
	    CPInst[k].bdy[p].y = octPts[p].y;
	}
	/* Get corner stuff from oct: */
	MSG(2, ("Retrieving corners of instance from oct\n"));
	/* Get first corner */
	OH_ASSERT(ohGetByBagName(&octInst, &octCornHead, "CDCorners"));
	OH_ASSERT(ohGetByPropName(&octCornHead, &octCorn, "CDCorner"));
	numCorn = 0;
	corn[numCorn].bdyNum = octCorn.contents.prop.value.integer;

	while (TRUE) {		/* Go through rest of corner chain */
	    OH_ASSERT(MIOctGetByNameContainers(&octCorn, &octCEdge));
	    if (MIOctGetByNameContainers(&octCEdge, &octCornChan)!=OCT_OK)
	      then chanNum = CP_NULL_CHANNEL;
	    else CP_CHAN_TO_CHANNUM(&chanNum, &octCornChan);
	    if (chanNum!=CP_NULL_CHANNEL)
	      DRANGE(0, chanNum, CPNumChan);
	    MSG(5, ("Got corner %d: bdyNum=%d chanNum=%d\n",
		    numCorn, corn[numCorn].bdyNum, chanNum));
	    corn[numCorn++].chanNum = chanNum;
	    prevCornOctId = octCorn.objectId;
	    OH_ASSERT(octInitGenContents(&octCEdge, OCT_PROP_MASK, &genCorn));
	    for (foundFlg=FALSE; octGenerate(&genCorn, &octCorn)==OCT_OK; ) {
		if (!octIdsEqual(octCorn.objectId, prevCornOctId)) {
		    foundFlg = TRUE;
		    break;
		}
	    }
	    octFreeGenerator( &genCorn );
	    if (!foundFlg) break;
	    URANGE(numCorn, CP_MAX_NUM_CORN);
	    corn[numCorn].bdyNum = octCorn.contents.prop.value.integer;
	    DRANGE(0, corn[numCorn].bdyNum, CPInst[k].numBdy);
	}

	CPInst[k].numCorn = numCorn; /* Copy corn stuff into CPInst[] */
	CPInst[k].corn = ALLOC_ARRAY(numCorn, CPCornDef);
	for (c=0; c<numCorn; c++) {
	    if (CPInst[k].bdyOrient!=CP_CW)
	      then b = corn[c].bdyNum;
	    else b = CPInst[k].numBdy-corn[c].bdyNum-1;
	    chanNum = corn[c].chanNum;
	    MSG(2, ("Corner %d: bdyNum=%d (%d %d) chanNum=%d\n",
		    c, b, CPInst[k].bdy[b], chanNum));
	    CPInst[k].corn[c].bdyNum  = b;
	    CPInst[k].corn[c].chanNum = chanNum;
	}
	/* Process aTerms: */
	OH_ASSERT(octInitGenContents(&octInst, OCT_TERM_MASK, &genATerm));
	while (octGenerate(&genATerm, &octATerm)==OCT_OK) {
	    CP_TERM_TO_CHANNUM(&chanNum, &octATerm);
	    MSG(2, ("On aTerm %s bordering channel %d\n",
		    octATerm.contents.term.name, chanNum));

	    OH_ASSERT(ohGetByTermName(&octInstCopy, &octATermCopy,
				      octATerm.contents.term.name));
	    CPOctCopyContents(&octATerm, &octATermCopy,
			      OCT_BAG_MASK | OCT_PROP_MASK);
	    /* Init local CPPin aTerm data */
	    pin = ALLOC_TYPE(CPPinDef);
	    pin->next = CPPin;
	    CPPin = pin;
	    pin->oldTermId = octATerm    .objectId;
	    pin->newTermId = octATermCopy.objectId;
	    pin->instNum   = k;
	    pin->chanNum   = chanNum;
	    pin->dir       = 'C';
	    pin->subNetNum = CP_NULL_NET;
	    pin->name      = STR_COPY(octATerm.contents.term.name);
	    pin->newName   = pin->name;
	    pin->isAFTermFlg = FALSE;

	    if (pin->chanNum==CPCurChan) { /* ATerm in channel (get subNetNum) */
		OH_ASSERT(MIOctGetByNameContainers(&octATerm, &octSubNet));
		CPNetToNetNum(&pin->subNetNum, &octSubNet);
	    }
	    /* ATerm in different channel */
	    else if (pin->chanNum!=CP_NULL_CHANNEL) {
		/* Create an fTerm for the ATerm */
		OH_ASSERT(MIOctGetByNameContainers(&octATerm, &octSubNet));
		CPCreateFTerm(octMaster, &octFTermCopy, pin, &octInstCopy);
		OH_ASSERT(octAttach(&octFTermCopy, &octATermCopy));
		/* Create and attach net for fTerm */
		if (octGenFirstContainer(&octSubNet, OCT_NET_MASK, &octNet)!=OCT_OK)
		  MIerror("No net found for subNet of aTerm %s inst %s chan %d\n",
			  pin->name, CPInst[k].name, pin->chanNum);
		OH_ASSERT(ohGetOrCreateNet(octMaster, &octFTermNet,
					   octNet.contents.net.name));
		OH_ASSERT(octAttach(&octFTermNet, &octFTermCopy));
		OH_ASSERT(octAttach(&octFTermNet, &octATermCopy));
		/* Copy properties under original aTerm */
		OH_ASSERT(ohGetByTermName(octInstMaster, &octFTerm,
					  octATermCopy.contents.term.name));
		CPOctCopyContents(&octFTerm, &octFTermCopy,
				  OCT_BAG_MASK | OCT_PROP_MASK);
		continue;
	    }
	    /* ATerm is an fTerm ? */
	    if (octGenFirstContainer(&octATerm, OCT_TERM_MASK, &octFTerm)==OCT_OK &&
		octIdIsNull(octFTerm.contents.term.instanceId)) {
		/* Create an fTerm for the ATerm */
		CPCreateFTerm(octMaster, &octFTermCopy, pin, &octInstCopy);
		OH_ASSERT(octAttach(&octFTermCopy, &octATermCopy));
		pin->isAFTermFlg = TRUE;
		if (octGenFirstContainer(&octFTerm, OCT_NET_MASK, &octFTermNet)==OCT_OK) {
		    OH_ASSERT(octGetOrCreate(octMaster, &octFTermNet));
		    OH_ASSERT(octAttach(&octFTermNet, &octFTermCopy));
		}
	    }
	}
    }

    if (k==1) {			/* External channel */
	CPInst[k].numBdy  = 0;
	CPInst[k].bdy     = NULL;
	CPInst[k].numCorn = 0;
	CPInst[k].loc.x   = 0;
	CPInst[k].loc.y   = 0;
	CPInst[k].corn    = NULL;
	CPInst[k].name    = CP_EXT_INST_NAME;
	CPExtInstNum      = 1;
	MSG(1, ("External channel instance %s\n", CPInst[k].name));
    }
    else CPExtInstNum     = -1;

    for (sn=0, p=0; sn<CPNumSubNet; sn++) { /* Each subNet */
	MSG(3, ("On subnet %d (octSubNetId=%d)\n", sn, CPNet[sn].subNetId));
	n = CPNet[sn].netNum;
	OH_ASSERT(ohGetById(&octSubNet, CPNet[sn].subNetId));
	if (*debugFlag>=2) {
	    OH_ASSERT(octInitGenContents(&octSubNet, OCT_TERM_MASK, &genATerm));
	    while (octGenerate(&genATerm, &octATerm)==OCT_OK) {
		CP_TERM_TO_CHANNUM(&chanNum, &octATerm);
		OH_ASSERT(octGenFirstContainer(&octATerm, OCT_INSTANCE_MASK, &octInst));
		MSG(-1, ("Aterm %s of inst %s of subnet %d borders chan %d\n",
			 octATerm.contents.term.name, octInst.contents.instance.name,
			 sn, chanNum));
	    }
	}
	/* Process pTerms */
	OH_ASSERT(octInitGenContents(&octSubNet, OCT_BAG_MASK, &genPTerm));
	while (octGenerate(&genPTerm, &octPTerm)==OCT_OK) {
	    if (strcmp(octPTerm.contents.bag.name, "CDPseudoTerm")) continue;
	    MSG(3, ("On pTerm%d (octId=%s) of subnet %d\n",
		    p, octFormatId(octPTerm.objectId), sn));
	    /* Get "2nd" subnet on pTerm */
	    OH_ASSERT(octInitGenContainers(&octPTerm, OCT_NET_MASK, &genSubNet));
	    for (foundFlg=FALSE; octGenerate(&genSubNet, &octSubNet2)==OCT_OK; ) {
		if (!octIdsEqual(octSubNet2.objectId, octSubNet.objectId)) {
		    foundFlg = TRUE;
		    break;
		}
	    }
	    octFreeGenerator( &genSubNet );
	    if (!foundFlg)
	      MIerror("pTerm has only one subnet - needs 2\n");
	    octSubNetChan.type = OCT_BAG;
	    octSubNetChan.contents.bag.name = "CDChanDef";
	    OH_ASSERT(MIOctGetByNameContainers(&octSubNet2, &octSubNetChan));
	    CP_CHAN_TO_CHANNUM(&chanNum, &octSubNetChan);
	    if (chanNum==CP_NULL_CHANNEL)
	      MIerror("pTerm subnet not attached to any channel\n");
	    MSG(3, ("2nd subnet chanNum of pTerm = %d\n", chanNum));

	    pin = ALLOC_TYPE(CPPinDef);
	    pin->next = CPPin;
	    CPPin = pin;
	    pin->oldTermId = octPTerm.objectId;
	    pin->instNum = CP_PTERM_INSTNUM;
	    pin->chanNum = chanNum;
	    pin->dir = 'C';

	    if (ohGetByPropName(&octPTerm, &octSideTermOrder, "GRSideTermOrder")==OCT_OK) {
		DASSERT(octSideTermOrder.contents.prop.type==OCT_STRING);
		pin->dir = *(octSideTermOrder.contents.prop.value.string);
		OH_ASSERT(octDelete(&octSideTermOrder));
	    }
	    if (!strchr("UDLRC",pin->dir)) pin->dir = 'C';

	    OH_ASSERT(ohGetById(&octNet, CPNet[n].netId));
	    (void) sprintf(CPTempBuf,"pTerm%s.%d.%d", octNet.contents.net.name, CPCurChan, p);
	    pin->name = STR_COPY(CPTempBuf);
	    pin->newName = pin->name;
	    pin->isAFTermFlg = TRUE;
	    /* All pTerms become fTerms */
	    CPCreateFTerm(octMaster, &octFTerm, pin, NIL(octObject));
	    pin->newTermId = octFTerm.objectId;
	    CPOctCopyContents(&octPTerm, &octFTerm, OCT_BAG_MASK | OCT_PROP_MASK);
	    /* Get netNum of pTerm */
	    CPNetToNetNum(&pin->subNetNum, &octSubNet);
	    DASSERT(pin->subNetNum!=CP_NULL_NET);

	    p++;
	}
    }
    /* Check subnet pin count */
    numSubNetPins = ALLOC_ARRAY(CPNumSubNet, short);
    for (sn=0; sn<CPNumSubNet; sn++)
      numSubNetPins[sn] = 0;
    for (pin=CPPin; pin; pin=pin->next) {
	if (pin->subNetNum==CP_NULL_NET) continue;
	RANGE(0, pin->subNetNum, CPNumSubNet);
	numSubNetPins[pin->subNetNum]++;
    }
    for (sn=0; sn<CPNumSubNet; sn++) {
	if (numSubNetPins[sn]>=2) continue;
	OH_ASSERT(ohGetById(&octNet, CPNet[CPNet[sn].netNum].netId));
	MIerror("num of pins on subnet of net %s has %d pin(s)\n",
		octNet.contents.net.name, numSubNetPins[sn]);
	MSG(-1, ("subnet octId=%d\n", CPNet[sn].subNetId));
	for (pin=CPPin; pin; pin=pin->next)
	  if (pin->subNetNum==sn)
	    MSG(-1, ("pin %s octId=%d inst=%d chan=%d\n",
		     pin->name, pin->oldTermId, pin->instNum, pin->chanNum));
    }
    FREE_ARRAY(CPNumSubNet, numSubNetPins, short);

    MI_MOD_END();
}				/* ............................. End of procedure CPInitSubMaster */
/* Remove channel definition data structures
----------------------------------------------------------------------------*/
CPCleanOctChan(octChanObj, deleteInstanceFlag)

octObject	*octChanObj;
Boolean		deleteInstanceFlag;
{
    octObject	octChanLeaf;
    octObject	octInst;
    octObject	octPTerm;
    octObject	octSubNet;
    octObject	octSubStuff;

    octGenerator	genSubStuff;
    octGenerator	genPTerm;
    octGenerator	genChanLeaf;
    octGenerator	genSubNet;

    octObject	octCHead;
    octObject	octCorn;
    octObject	octCEdge;
    octGenerator	genCorn;

    octObject	octGPoint;
    octObject	octGEdge;

    MI_MOD_START("CPCleanOctChan");

    OH_ASSERT(octInitGenContents(octChanObj, OCT_BAG_MASK, &genChanLeaf));
    /* Each Instance bordering channel: */
    while (octGenerate(&genChanLeaf, &octChanLeaf)==OCT_OK) {

	if (!strcmp(octChanLeaf.contents.bag.name, "CDGraph")) {
	    OH_ASSERT(octGenFirstContent(&octChanLeaf, OCT_POINT_MASK, &octGPoint));
	    OH_ASSERT(octDelete(&octChanLeaf));
	    while (!octIdIsNull(octGPoint.objectId)) {
		OH_ASSERT(octGenFirstContainer(&octGPoint, OCT_BAG_MASK, &octGEdge));
		DASSERT(!strcmp(octGEdge.contents.bag.name, "CDGraphEdge"));
		OH_ASSERT(octDelete(&octGPoint));
		OH_ASSERT(octInitGenContents(&octGEdge, OCT_ALL_MASK, &genSubStuff));
		octGPoint.objectId = oct_null_id;
		while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK) {
		    if (octSubStuff.type==OCT_POINT)
		      octGPoint = octSubStuff;
		    else OH_ASSERT(octDelete(&octSubStuff));
		}
		OH_ASSERT(octDelete(&octGEdge));
	    }
	}

	if (!strcmp(octChanLeaf.contents.bag.name, "CDChanDef")) {
	    if (octGenFirstContent(&octChanLeaf, OCT_INSTANCE_MASK, &octInst)!=OCT_OK)
	      continue;
	    /* Get rid of corner stuff */
	    if (ohGetByBagName(&octInst, &octCHead, "CDCorners")!=OCT_OK)
	      continue;
	    OH_ASSERT(ohGetByPropName(&octCHead, &octCorn, "CDCorner"));
	    OH_ASSERT(octDelete(&octCHead));
	    while (TRUE) {	/* Go through rest of corner chain */
		octCEdge.type = OCT_BAG;
		octCEdge.contents.bag.name = "CDCornerEdge";
		OH_ASSERT(MIOctGetByNameContainers(&octCorn, &octCEdge));
		OH_ASSERT(octDelete(&octCorn));
		if (octGenFirstContent(&octCEdge, OCT_PROP_MASK, &octCorn)!=OCT_OK) break;
		OH_ASSERT(octDelete(&octCEdge));
	    }

	    if (deleteInstanceFlag) {
		OH_ASSERT(octInitGenContents(&octInst, OCT_ALL_MASK, &genSubStuff));
		while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK) {
		    /* DO NOT delete stuff under terminals! */
		    if (octSubStuff.type!=OCT_TERM) {
			ohDeleteContents(&octSubStuff, OCT_ALL_MASK);
			OH_ASSERT(octDelete(&octSubStuff));
		    }
		}
		OH_ASSERT(octDelete(&octInst));	/* Get rid of Instance */
	    }
	}
    }

    OH_ASSERT(octInitGenContents(octChanObj, OCT_NET_MASK, &genSubNet));
    while (octGenerate(&genSubNet, &octSubNet)==OCT_OK) {
	/* Get rid of all PTerms */
	OH_ASSERT(octInitGenContents(&octSubNet, OCT_BAG_MASK, &genPTerm));
	while (octGenerate(&genPTerm, &octPTerm)==OCT_OK) {
	    DASSERT(!strcmp(octPTerm.contents.bag.name, "CDPseudoTerm"));
	    OH_ASSERT(octDelete(&octPTerm));
	}
    }
    /* Get rid of all other stuff */
    OH_ASSERT(octInitGenContents(octChanObj, OCT_ALL_MASK, &genSubStuff));
    while (octGenerate(&genSubStuff, &octSubStuff)==OCT_OK) {
	if (octSubStuff.type==OCT_INSTANCE) continue;
	OH_ASSERT(octDelete(&octSubStuff));
    }

    MI_MOD_END();
}				/* ............................. End of procedure CPCleanOctChan */
/* Remove old instances, channels, etc... to be replaced by a new instance
of the merged instance+channel+instance.
----------------------------------------------------------------------------*/
CPCleanChan(octChanObj)

octObject	*octChanObj;
{
octObject	octNet, octNetTerm;
short		k, n;
CPPinDef	*p, *pn;

MI_MOD_START("CPCleanChan");

CPCleanOctChan(octChanObj, TRUE);

FREE_STRING(CPSubMasterName);

for (p=CPPin; p; p=pn) {		/* Free pins */
    pn = p->next;
    if (p->newName!=p->name)
	FREE_STRING(p->newName);
    FREE_STRING(p->name);
    FREE_TYPE(p, CPPinDef);
}
CPPin = NULL;

for (n=0; n<CPNumNet; n++) {
    OH_ASSERT(ohGetById(&octNet, CPNet[n].netId));
    if (octGenFirstContent(&octNet, OCT_TERM_MASK, &octNetTerm)!=OCT_OK) {
	MSG(2, ("Deleting net%d name=%s\n", n, octNet.contents.net.name));
	OH_ASSERT(octDelete(&octNet));
    }
    FREE_STRING(CPNet[n].name);
}
FREE_ARRAY(CPNumSubNet, CPNet, CPNetDef);

for (k=0; k<2; k++) {			/* Free instance stuff */
    if (CPInst[k].name && k!=CPExtInstNum)
	FREE_STRING(CPInst[k].name);
    if (CPInst[k].numBdy>0)
	FREE_ARRAY(CPInst[k].numBdy, CPInst[k].bdy, CPLocDef);
    if (CPInst[k].numCorn>0)
	FREE_ARRAY(CPInst[k].numCorn, CPInst[k].corn, CPCornDef);
}

MI_MOD_END();
} /* ............................. End of procedure CPCleanChan */
/* Bind aTerms & pTerms to new subMaster instance.  Bind subMaster instance
to leaf channel object.  Recompute corners and bind new corners to subMaster
instance.
----------------------------------------------------------------------------*/
CPRebind(octChanObj)

octObject	*octChanObj;
{
octObject	octOldTerm;
octObject	octNewTerm;
octObject	octInstBag;
octObject	octNet;
int		n;
CPPinDef	*pin;

MI_MOD_START("CPRebind");

MSG(1, ("Creating new subMasterInst %s\n", CPOctChRouted.contents.facet.cell));
OH_ASSERT(ohCreateInst(&CPOctMaster, &CPOctSubMasterInst,
    CPOctChRouted.contents.facet.cell, CPOctChRouted.contents.facet.cell,
    CP_OCT_POSTROUTE_VIEW));
					/* Attach new inst to instance bag */
OH_ASSERT(ohGetByBagName(&CPOctMaster, &octInstBag, "INSTANCES"));
OH_ASSERT(octAttach(&octInstBag, &CPOctSubMasterInst));

for (pin=CPPin; pin; pin=pin->next) {	/* Bind aTerms & pTerms to new inst */
    if ((!pin->isAFTermFlg) &&
	(pin->chanNum==CPCurChan || pin->chanNum==CP_NULL_CHANNEL)) continue;
    MSG(2, ("Rebinding term %s\n", pin->newName));
					/* Get old and new terms */
    OH_ASSERT(ohGetById(&octOldTerm, pin->oldTermId));
    OH_ASSERT(ohGetByTermName(&CPOctSubMasterInst, &octNewTerm, pin->newName));
					/* Copy attachments from old to new */
    CPOctSwapTerm(&octOldTerm, &octNewTerm);
    if (pin->subNetNum!=CP_NULL_NET) {
	n = CPNet[pin->subNetNum].netNum;
	OH_ASSERT(ohGetById(&octNet, CPNet[n].netId));
	if (MIOctGetByNameContainers(&octNewTerm, &octNet)!=OCT_OK) {
	    MSG(3, ("Attaching net%d %s to term %s\n",
		n, octNet.contents.net.name, octNewTerm.contents.term.name));
	    OH_ASSERT(octAttach(&octNet, &octNewTerm));
	}
    }
}
					/* Bind new inst to channel leaf */
OH_ASSERT(octAttach(octChanObj, &CPOctSubMasterInst));

CPRemapCorners();

MI_MOD_END();
} /* ............................. End of procedure CPRebind */
