/*
Copyright (C) 1989 University of California (Berkeley). All rights reserved.

CPPinLoc.c

Smart pin location routines.  First obtain legal ranges for each pin.
Then try to assign pin locations so that they don't overlap.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CPPinLoc.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:51  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CPInc.h"
					/* See set_up_width_and_spacing() */
extern int	layer_width[3];
extern int	layer_spacing[3];

/* Assign smart locations for the "fixed" (i.e. top and bottom) pins
in channel.
----------------------------------------------------------------------------*/
CPSmartPinLoc(vertMode)

Boolean		vertMode;		/* Is this a vertical channel ? */
{
CPPinDef	*p, *p2;
double		SPcurrentDensity();
int		SPquantizeWidth();
					/* Ranges of terminal locations */
					/*   on top and bottom boundaries */
					/* Computed but not used in the */
					/*   pin movement algorithm */
int		minTopPin, maxTopPin;
int		minBotPin, maxBotPin;

CPLocDef	origLoc;
int		lo, hi, lo2, hi2;
int		space;
short		layer, k;
int		lastIteration = 5;
Boolean		locOkFlag, locLoOkFlag, locHiOkFlag;
Boolean		overlapFlag;

MI_MOD_START("CPSmartPinLoc");

/* First get the ranges of each terminal */
/* We initialize the data structures p->locLo and p->locHi */

minTopPin = minBotPin =  MAXINT;
maxTopPin = maxBotPin = -MAXINT;

for (p=CPPin; p; p=p->next)
    p->exitDir = '?';

for (p=CPPin; p; p=p->next) {
    if ((p->chanNum!=CPCurChan &&	/* Skip pins not on this channel */
	 p->instNum!=CP_PTERM_INSTNUM) ||
					/* Skip channel end terminals */
	(p->instNum==CP_PTERM_INSTNUM)) continue;

					/* Get layer of pin first */
    CPGetTermLoc(p, &p->loc, &p->layer, (vertMode ? 2 : 1));

					/* Compute width of pin */
    p->width = ABS(p->peakCurrent)/SPcurrentDensity(p->layer);
    p->width = MAX(SPquantizeWidth(p->width), layer_width[p->layer]);

    CPGetTermRange(p, vertMode);
					/* Determine entire pin range */
    if (p->instNum==CPWhere.instOnTop) {
	minTopPin = MIN(minTopPin, p->locLo - p->width/2);
	maxTopPin = MAX(maxTopPin, p->locHi + p->width/2);
    }
    if (p->instNum==CPWhere.instOnBot) {
	minBotPin = MIN(minBotPin, p->locLo - p->width/2);
	maxBotPin = MAX(maxBotPin, p->locHi + p->width/2);
    }
					/* Set exitDir */
    if (p->exitDir!='?') continue;

    for (p2=CPPin; p2; p2=p2->next) {
	if (p2->subNetNum!=p->subNetNum ||
	    p2->instNum!=CP_PTERM_INSTNUM) continue;
	if ((p2->chanNum==CPWhere.chanOnRight &&  vertMode) ||
	    (p2->chanNum==CPWhere.chanOnLeft  && !vertMode)) {
	    if      (p->exitDir=='?') p->exitDir = 'L';
	    else if (p->exitDir!='L') p->exitDir = 'X';
	}
	else {
	    if      (p->exitDir=='?') p->exitDir = 'H';
	    else if (p->exitDir!='H') p->exitDir = 'X';
	}
    }
    for (p2=CPPin; p2; p2=p2->next)
	if (p2->subNetNum==p->subNetNum && p2!=p)
	    p2->exitDir = p->exitDir;
					/* ... Done setting exitDir */
}

/* Move the remaining terminals around */

layer = (vertMode ? 2 : 1);		/* This is the layer which we */
					/*   will assume the routing is on */
					/*   we use the spacing rules on */
					/*   this layer */
space = layer_spacing[layer];
MSG(1, ("spacing for layer %d = %d\n", layer, space));

overlapFlag = TRUE;
for (k=0; k<=lastIteration && overlapFlag; k++) {

    overlapFlag = FALSE;
    for (p=CPPin; p; p=p->next) {
	if ((p->chanNum!=CPCurChan &&
	     p->instNum!=CP_PTERM_INSTNUM) ||
	    (p->instNum==CP_PTERM_INSTNUM)) continue;
	if (p->locLo==p->locHi) continue;

	origLoc.x = p->loc.x;	
	origLoc.y = p->loc.y;	
	if (vertMode) {
	    lo = origLoc.y - p->width/2 - space;
	    hi = origLoc.y + p->width/2 + space;
	}
	else {
	    lo = origLoc.x - p->width/2 - space;
	    hi = origLoc.x + p->width/2 + space;
	}

	locOkFlag = TRUE;
	locLoOkFlag = TRUE;
	locHiOkFlag = TRUE;

	for (p2=CPPin; p2; p2=p2->next) {
	    if (p2->chanNum!=CPCurChan ||
		p2->instNum!=p->instNum ||
		p==p2) continue;

	    if (vertMode) {
		hi2 = p2->loc.y+p2->width/2;
		lo2 = p2->loc.y-p2->width/2;
	    }
	    else {
		hi2 = p2->loc.x+p2->width/2;
		lo2 = p2->loc.x-p2->width/2;
	    }
					/* See if p->loc is ok */
	    if (hi2>lo && lo2<hi)
		locOkFlag = FALSE;
					/* See if p->locLo is ok */	
	    if (hi2>p->locLo-p->width/2-space &&
		lo2<p->locLo+p->width/2-space)
		locLoOkFlag = FALSE;
					/* See if p->locHi is ok */
	    if (hi2>p->locHi-p->width/2-space &&
		lo2<p->locHi+p->width/2-space)
		locHiOkFlag = FALSE;
	}
	if (locOkFlag) continue;

	if (locLoOkFlag && p->exitDir!='H') {
	    if (vertMode) 
		 p->loc.y = p->locLo;
	    else p->loc.x = p->locLo;

	    MSG(-1, ("Pin %s = moved from (%d %d) to (%d %d)\n",
		p->name, origLoc.x, origLoc.y, p->loc.x, p->loc.y));
	    continue;
	}
	if (locHiOkFlag && p->exitDir!='L') {
	    if (vertMode) 
		 p->loc.y = p->locHi;
	    else p->loc.x = p->locHi;

	    MSG(-1, ("Pin %s = moved from (%d %d) to (%d %d)\n",
		p->name, origLoc.x, origLoc.y, p->loc.x, p->loc.y));
	    continue;
	}

	if (k==lastIteration)
	    MIerror("Pin %s too close to another pin detected\n", p->name);
	else {				/* Change pin range */
	    overlapFlag = TRUE;
	    if (p->exitDir!='L') {
		p->locLo += p->width;
		p->locLo = MIN(p->locLo, p->locHi);
	    }
	    if (p->exitDir!='H') {
		p->locHi -= p->width;
		p->locHi = MAX(p->locLo, p->locHi);
	    }
	}
    }
}
MI_MOD_END();
} /* ............................. End of procedure CPSmartPinLoc */
/* Get location of pin from mid point of oct master fTerm box.
----------------------------------------------------------------------------*/
CPGetTermRange(pin, vertMode)
    CPPinDef	*pin;
    Boolean	vertMode;
{
    octObject	*octInstMaster;
    octObject	octFTerm, octInst, octATerm, octGeom, octLayer;
    octGenerator genGeom;
    int		lo, hi;
    short	l, CPCompLayerNum();
    Boolean	pinLocSet = FALSE;

    MI_MOD_START("CPGetTermRange");

    pin->locLo =  MAXINT;
    pin->locHi = -MAXINT;

    octInstMaster = &CPInst[pin->instNum].octInstMaster;
    OH_ASSERT(ohGetByTermName(octInstMaster, &octFTerm, pin->name));
    OH_ASSERT(ohGetById(&octInst, CPInst[pin->instNum].newInstId));
    OH_ASSERT(ohGetById(&octATerm, pin->oldTermId));

    OH_ASSERT(octInitGenContents(&octFTerm, OCT_GEO_MASK, &genGeom));
    while (octGenerate(&genGeom, &octGeom)==OCT_OK) {
	OH_ASSERT(octGenFirstContainer(&octGeom, OCT_LAYER_MASK, &octLayer));
	if (octGeom.type!=OCT_BOX) continue;

	l = CPCompLayerNum(&octLayer);
	if (l==pin->layer) {
	    tr_oct_transform(&octInst.contents.instance.transform,
			     &octGeom.contents.box.lowerLeft.x,
			     &octGeom.contents.box.lowerLeft.y);

	    if (vertMode) {
		lo = hi = octGeom.contents.box.lowerLeft.y;
	    }
	    else {
		lo = hi = octGeom.contents.box.lowerLeft.x;
	    }

	    tr_oct_transform(&octInst.contents.instance.transform,
			     &octGeom.contents.box.upperRight.x,
			     &octGeom.contents.box.upperRight.y);

	    if (vertMode) {
		lo = MIN(lo, octGeom.contents.box.upperRight.y);
		hi = MAX(hi, octGeom.contents.box.upperRight.y);
	    }
	    else {
		lo = MIN(lo, octGeom.contents.box.upperRight.x);
		hi = MAX(hi, octGeom.contents.box.upperRight.x);
	    }

	    lo += pin->width/2;
	    hi -= pin->width/2;

	    if (lo<pin->locLo) pin->locLo = lo;
	    if (hi>pin->locHi) pin->locHi = hi;

	    pinLocSet = TRUE;
	}
    }

    if (pinLocSet) {
	if (pin->locLo<pin->locHi) {
	    MSG(1, ("pin loc range of term %s =(%d %d) width=%d\n",
		    pin->name, pin->locLo, pin->locHi, pin->width));
	}

	if (vertMode)
	  then pin->loc.y = (pin->locLo + pin->locHi)/2;
	else pin->loc.x = (pin->locLo + pin->locHi)/2;
    }

    MI_MOD_END();
}				/* ............................. End of procedure CPGetTermRange */
