/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPRInter.c

Interface of channel preparation for channel router file interface.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CPRInter.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:51  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.2  89/01/30  15:56:18  octtools
 * lint errors
 * 
 * Revision 1.1  89/01/26  14:20:13  octtools
 * Initial revision
 * 
 * Revision 1.8  88/04/27  13:04:19  mitch
 * Updated message about non POLY, MET1 or MET2 terminal contacts.
 * 
 * Revision 1.6  88/04/24  16:51:37  mitch
 * Print out aTerm name instead of fTerm name.
 * 
 * Revision 1.5  88/04/11  12:16:02  mitch
 * Changed CPCompLayerNum to return CP_UNKNOWN_LAYER_NUM for everything
 * except POLY, MET1 and MET2.
 * 
 * Revision 1.4  88/04/08  16:34:55  mitch
 * Put in pin direction as an option to mosaicoDR.
 * 
 * Revision 1.2  88/04/05  16:00:35  mitch
 * Added pseudo pin direction entry in the pin records of the interface file.
 * 
 * Revision 1.1  88/03/31  11:23:01  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CPInc.h"
/* Create routing interface file from oct sub master contents view.
----------------------------------------------------------------------------*/
CPRouteInter(tiFileName)

String		tiFileName;		/* Termite file name */
{
FILE		*f;
CPPinDef	*p;

short		layerNum = 1;		/* Layer number */
short		numSideBdy;		/* Number of side bdy points */
short		sBdyNum, eBdyNum;
short		chanWidth;
int		tempCoord;
CPLocDef	sExtLoc, eExtLoc;
CPInstDef	*iRef, *iTop, *iBot;

octObject	octNet;
octObject	octTerm;
octObject	octPWidthProp;
octObject	octPeakCProp;

short		totNumPin = 0;		/* Total number of pins */
short		chkTotNumPin = 0;	/* Check sum for totNumPin */
short		numSidePin[4];		/* Number of pins on each side */
short		pinSubNet, pinNet;	/* Pin subNet and net numbers */
short		pinInst;		/* Pin instance number: */
					/*   -1 = pseudo terminal */
short		pinType;		/* Type of pin: */
					/*    0 = floating */
					/*    1 = relative floating */
					/*    2 = fixed */
short		netType;		/* Type of net: */
					/*    0 = normal signal */
					/*    1 = VDD */
					/*    2 = GND */
short		s;			/* Side of channel */
short		k;
Boolean		vertMode;		/* Vertical channel mode */

Boolean		smartPinLocMode = TRUE;	/* Flag for the new pin location */
					/*   routine that TRIES to avoid */
					/*   overlapping terminals */

MI_MOD_START("CPRouteInter");

f = MIfopen(tiFileName,"w","pre-router (Termite) interface file");
if (!f) goto CPEndRouteInter;
					/* Print CHANNEL header record */
vertMode = (CPWhere.chanDir==CP_V);
fprintf(f, "CHANNEL %s %d %d\n",
    CPSubMasterName, (vertMode ? 1 : 0), CPCurChan);
					/* Print NET record */
fprintf(f, "NET %d\n", CPNumSubNet);
					/* Print Boundary records */
for (s=0; s<4; s=s+2) {			/* Top and bottom boundaries */
    fprintf(f, "%s ", (s==0 ? "TOP" : "BOTTOM"));
    k = (s==0 ? CPWhere.instOnTop : CPWhere.instOnBot);

    if (k==CPExtInstNum) {		/* External boundaries */
        int bi;
	numSideBdy = 2;
	fprintf(f, "%d %d\n", numSideBdy, layerNum);
	chanWidth = 100;		/* TODO put in better channel width */

	DASSERT(k==1);
	iRef = &CPInst[0];
	sBdyNum = iRef->startBdyNum;
	eBdyNum = iRef->endBdyNum;
					/* Extremes of other inst boundary */
	sExtLoc.x = MIN(iRef->bdy[sBdyNum].x, iRef->bdy[eBdyNum].x);
	sExtLoc.y = MIN(iRef->bdy[sBdyNum].y, iRef->bdy[eBdyNum].y);
	eExtLoc.x = MAX(iRef->bdy[sBdyNum].x, iRef->bdy[eBdyNum].x);
	eExtLoc.y = MAX(iRef->bdy[sBdyNum].y, iRef->bdy[eBdyNum].y);
	if (vertMode) {
	    if (k==CPWhere.instOnBot) {	/* West side */
		for (bi=0; bi<iRef->numBdy; bi++)
		  sExtLoc.x = MIN(sExtLoc.x, iRef->bdy[bi].x);
		sExtLoc.x -= chanWidth;
	    }
	    else {			/* East side */
		tempCoord = eExtLoc.y;
		eExtLoc.y = sExtLoc.y;
		sExtLoc.y = tempCoord;
		for (bi=0; bi<iRef->numBdy; bi++)
		  sExtLoc.x = MAX(sExtLoc.x, iRef->bdy[bi].x);
		sExtLoc.x += chanWidth;
	    }
	    eExtLoc.x = sExtLoc.x;
	}
	else {
	    if (k==CPWhere.instOnTop) {	/* North side */
		for (bi=0; bi<iRef->numBdy; bi++)
		  sExtLoc.y = MAX(sExtLoc.y, iRef->bdy[bi].y);
		sExtLoc.y += chanWidth;
	    }
	    else {			/* South side */
		tempCoord = eExtLoc.x;
		eExtLoc.x = sExtLoc.x;
		sExtLoc.x = tempCoord;
		for (bi=0; bi<iRef->numBdy; bi++)
		  sExtLoc.y = MAX(sExtLoc.y, iRef->bdy[bi].y);
		sExtLoc.y -= chanWidth;
	    }
	    eExtLoc.y = sExtLoc.y;
	}
	fprintf(f, "%d %d %d %d\n", sExtLoc.x, sExtLoc.y, eExtLoc.x, eExtLoc.y);
    }

    else {				/* Internal instance boundary */
	iRef = &CPInst[k];
	sBdyNum = iRef->startBdyNum;
	eBdyNum = iRef->endBdyNum;
	CPShadowChBdy(iRef->bdy, sBdyNum, eBdyNum, iRef->numBdy,
	    vertMode, ((vertMode && !s) || (!vertMode && s)), f, layerNum);
    }
}

iTop = &CPInst[CPWhere.instOnTop];
iBot = &CPInst[CPWhere.instOnBot];

fprintf(f, "RIGHT 2 %d\n", layerNum);	/* Right side boundary */
if (CPExtInstNum<0)
    fprintf(f, "%d %d %d %d\n",
	iTop->bdy[iTop->endBdyNum].x,
	iTop->bdy[iTop->endBdyNum].y,
	iBot->bdy[iBot->startBdyNum].x,
	iBot->bdy[iBot->startBdyNum].y);
else if (CPExtInstNum==CPWhere.instOnTop)
    fprintf(f, "%d %d %d %d\n",
	eExtLoc.x,
	eExtLoc.y,
	iBot->bdy[iBot->startBdyNum].x,
	iBot->bdy[iBot->startBdyNum].y);
else
    fprintf(f, "%d %d %d %d\n",
	iTop->bdy[iTop->endBdyNum].x,
	iTop->bdy[iTop->endBdyNum].y,
	sExtLoc.x,
	sExtLoc.y);

fprintf(f, "LEFT 2 %d\n", layerNum);	/* Left side boundary */
if (CPExtInstNum<0)
    fprintf(f, "%d %d %d %d\n",
	iBot->bdy[iBot->endBdyNum].x,
	iBot->bdy[iBot->endBdyNum].y,
	iTop->bdy[iTop->startBdyNum].x,
	iTop->bdy[iTop->startBdyNum].y);
else if (CPExtInstNum==CPWhere.instOnTop)
    fprintf(f, "%d %d %d %d\n",
	iBot->bdy[iBot->endBdyNum].x,
	iBot->bdy[iBot->endBdyNum].y,
	sExtLoc.x,
	sExtLoc.y);
else
    fprintf(f, "%d %d %d %d\n",
	eExtLoc.x,
	eExtLoc.y,
	iTop->bdy[iTop->startBdyNum].x,
	iTop->bdy[iTop->startBdyNum].y);

for (s=0; s<4; s++)
    numSidePin[s] = 0;

for (p=CPPin; p; p=p->next) {		/* Count number of pins */
    if (p->chanNum!=CPCurChan &&	/* Skip pins not on this channel */
	p->instNum!=CP_PTERM_INSTNUM) continue;
    totNumPin++;
    if (p->instNum==CPWhere.instOnTop  ) numSidePin[0]++;
    if (p->chanNum==CPWhere.chanOnRight) numSidePin[1]++;
    if (p->instNum==CPWhere.instOnBot  ) numSidePin[2]++;
    if (p->chanNum==CPWhere.chanOnLeft ) numSidePin[3]++;
    if ((p->instNum==CPWhere.instOnTop && p->chanNum==CPWhere.chanOnRight) ||
	(p->instNum==CPWhere.instOnBot && p->chanNum==CPWhere.chanOnLeft)) {
	OH_ASSERT(ohGetById(&octNet, CPNet[CPNet[p->subNetNum].netNum].netId));
	MIerror("Pin assigned to two sides of channel. Pin is on Inst%d (of net %s)\n",
	    p->instNum, octNet.contents.net.name);
    }
    else if (p->instNum!=CPWhere.instOnTop &&
	     p->instNum!=CPWhere.instOnBot &&
	     p->chanNum!=CPWhere.chanOnRight &&
	     p->chanNum!=CPWhere.chanOnLeft) {
	OH_ASSERT(ohGetById(&octNet, CPNet[CPNet[p->subNetNum].netNum].netId));
	MIerror("Pin(%d) on Inst%d Chan%d of net %s is not assigned to any side of channel\n",
	    p->oldTermId, p->instNum, p->chanNum, octNet.contents.net.name);
	MSG(-1, ("topInst=%d botInst=%d chanRight=%d chanLeft=%d\n",
	    CPWhere.instOnTop, CPWhere.instOnBot,
	    CPWhere.chanOnRight, CPWhere.chanOnLeft));
    }
					/* Get pin width from property */
    OH_ASSERT(ohGetById(&octTerm, p->oldTermId));
    if (p->instNum!=CP_PTERM_INSTNUM) {
	MSG(2, ("On pin %s of inst %d\n",
	    octTerm.contents.term.name, p->instNum));
	DASSERT(p->instNum==0 || p->instNum==1);
	DASSERT(octTerm.type==OCT_TERM);
#ifdef CP_OLDTERMPROP_FROM_MASTER
	OH_ASSERT(octGetByName(&CPInst[p->instNum].octInstMaster, &octTerm));
#endif
    }
    if (ohGetByPropName(&octTerm, &octPWidthProp, "PorGWidth")==OCT_OK) {
	p->widthForSpider = octPWidthProp.contents.prop.value.integer;
	MSG(3, ("Width of pin %s = %d\n", p->newName, p->widthForSpider));
    }
    else p->widthForSpider = 0;
    if (ohGetByPropName(&octTerm, &octPeakCProp, "PEAKCURRENT")==OCT_OK) {
	p->peakCurrent = octPeakCProp.contents.prop.value.real;
	MSG(1, ("Peak current of pin %s = %lf\n",
	    p->newName, p->peakCurrent));
    }
    else p->peakCurrent = 0.0;
}

for (s=0; s<4; s++)			/* Check sum total number of pins */
    chkTotNumPin += numSidePin[s];
ASSERT(chkTotNumPin==totNumPin);
					/* Print PIN record */
fprintf(f, "PIN %d\n", chkTotNumPin);

if (smartPinLocMode)
    CPSmartPinLoc(vertMode);

for (s=0; s<4; s++) {			/* Pins on each side */

    pinType = ((s==0 || s==2) ? 2 : 0);

    switch(s) {
    case 0 : fprintf(f,"PINLIST TOP %d %d\n"   , numSidePin[s], pinType); break;
    case 1 : fprintf(f,"PINLIST RIGHT %d %d\n" , numSidePin[s], pinType); break;
    case 2 : fprintf(f,"PINLIST BOTTOM %d %d\n", numSidePin[s], pinType); break;
    default: fprintf(f,"PINLIST LEFT %d %d\n"  , numSidePin[s], pinType); break;
    }
    for (p=CPPin; p; p=p->next) {	/* Print PINLIST records */

	if (p->chanNum!=CPCurChan &&	/* Skip pins not on this channel */
	    p->instNum!=CP_PTERM_INSTNUM) continue;

	if (s==0 && p->instNum!=CPWhere.instOnTop  ) continue;
	if (s==1 && p->chanNum!=CPWhere.chanOnRight) continue;
	if (s==2 && p->instNum!=CPWhere.instOnBot  ) continue;
	if (s==3 && p->chanNum!=CPWhere.chanOnLeft ) continue;

	pinInst   = (p->instNum==CP_PTERM_INSTNUM ? -1 :
		    (p->instNum==CPWhere.instOnTop ? 0 : 1));
	pinSubNet = p->subNetNum;
	pinNet    = CPNet[p->subNetNum].netNum;

	if (p->instNum==CP_PTERM_INSTNUM) {
	    DASSERT(s==1 || s==3);
	    p->loc.x = p->loc.y = 0;
	    p->layer = (vertMode ? 1 : 2);
	}
	else if (!smartPinLocMode)
	    CPGetTermLoc(p, &p->loc, &p->layer, (vertMode ? 2 : 1));

	netType = 0;

	if (CPParm.pinDirMode)
	    fprintf(f, "%d %d %d %s %d %s %d %d %d %lf %c\n",
		pinSubNet+1, p->loc.x, p->loc.y, p->name, p->layer,
		CPNet[pinNet].name, netType, p->widthForSpider, pinInst,
		p->peakCurrent, p->dir);
	else
	    fprintf(f, "%d %d %d %s %d %s %d %d %d %lf C\n",
		pinSubNet+1, p->loc.x, p->loc.y, p->name, p->layer,
		CPNet[pinNet].name, netType, p->widthForSpider, pinInst,
		p->peakCurrent);
    }
}
					/* Print CELL records */
for (k=0; k<2; k++)
    fprintf(f, "CELL %s %d %d %d\n",
	CPInst[k].name, (k==CPWhere.instOnTop ? 0 : 1),
	CPInst[k].loc.x, CPInst[k].loc.y);

(void) fclose(f);

CPEndRouteInter:
MI_MOD_END();
} /* ............................. End of procedure CPRouteInter */
/* Get location of pin from mid point of oct master fTerm box.
----------------------------------------------------------------------------*/
CPGetTermLoc(pin, pinLoc, layerNum, prefLayerNum)

CPPinDef	*pin;
CPLocDef 	*pinLoc;
short		*layerNum;
int		prefLayerNum;
{
    octObject	octATerm,  octFTerm;
    octObject	octTermBag;
    octObject	octGeom, octLayer, octProp;
    octObject	*octInstMaster;
    octObject	octInst;
    octGenerator	genGeom, genProp;
    struct octPoint	tempPinLoc;
    short		l;
    short		CPCompLayerNum();
    short		accessLayer;
    Boolean		setPinLocFlag;
#define CP_UNKNOWN_LAYER_NUM (-100)

    MI_MOD_START("CPGetTermLoc");

    octInstMaster = &CPInst[pin->instNum].octInstMaster;
    OH_ASSERT(ohGetByTermName(octInstMaster, &octFTerm, pin->name));
    OH_ASSERT(ohGetById(&octInst, CPInst[pin->instNum].newInstId));
    OH_ASSERT(ohGetById(&octATerm, pin->oldTermId));

    /* The Bag name "MosTermInfo" is suppose to be obsolete */

    accessLayer = CP_UNKNOWN_LAYER_NUM;
    if ((ohGetByBagName(&octATerm, &octTermBag, "MosTermInfo")==OCT_OK) ||
	(ohGetByBagName(&octATerm, &octTermBag, "MosaicoTermInfo")==OCT_OK)) {
	OH_ASSERT(octInitGenContents(&octTermBag, OCT_PROP_MASK, &genProp));
	while (octGenerate(&genProp, &octProp)==OCT_OK) {
	    if (!strcmp(octProp.contents.prop.name,"LayerName")) {
		octLayer.contents.layer.name = octProp.contents.prop.value.string;
		l = CPCompLayerNum(&octLayer);
		if (l==prefLayerNum)
		  accessLayer = l;
		else if (accessLayer==CP_UNKNOWN_LAYER_NUM)
		  accessLayer = l;
	    }
	    /* Other property is "Direction" (int) - currently not used */
	    /* 0=Up, 1=Left, 2=Down and 3=Right */
	}
    }
    *layerNum = accessLayer;

    setPinLocFlag = FALSE;
    OH_ASSERT(octInitGenContents(&octFTerm, OCT_GEO_MASK, &genGeom));
    while (octGenerate(&genGeom, &octGeom)==OCT_OK) {
	switch (octGeom.type) {
	case OCT_BOX:
	    tempPinLoc.x = (octGeom.contents.box.lowerLeft.x+
			    octGeom.contents.box.upperRight.x)/2;
	    tempPinLoc.y = (octGeom.contents.box.lowerLeft.y+
			    octGeom.contents.box.upperRight.y)/2;
	    break;
	default:
	    MIerror("Cannot process pins of geometry type = %d (not an OCT_BOX)\n",
		    octGeom.type);
	}
	tr_oct_transform(&octInst.contents.instance.transform,
			 &tempPinLoc.x, &tempPinLoc.y);

	OH_ASSERT(octGenFirstContainer(&octGeom, OCT_LAYER_MASK, &octLayer));
	l = CPCompLayerNum(&octLayer);
	setPinLocFlag = TRUE;
	if (accessLayer!=CP_UNKNOWN_LAYER_NUM) {
	    pinLoc->x = tempPinLoc.x;
	    pinLoc->y = tempPinLoc.y;
	    if (l==accessLayer) {
		*layerNum = l;
		break;
	    }
	}
	if (*layerNum==CP_UNKNOWN_LAYER_NUM || l==prefLayerNum) {
	    pinLoc->x = tempPinLoc.x;
	    pinLoc->y = tempPinLoc.y;
	    *layerNum = l;
	}
    }
    octFreeGenerator( &genGeom );

    MSG(2, ("pinLoc of term %s =(%d %d)\n",
	    pin->name, pinLoc->x, pinLoc->y));

    if (*layerNum==CP_UNKNOWN_LAYER_NUM) {
	*layerNum = prefLayerNum;
    }
    if (!setPinLocFlag) {
	MIerror("fTerm %s of inst %s has no geometry on POLY, MET1 or MET2\n",
		pin->name, octInst.contents.instance.name);
	MSG(-1, ("You must put a contact to one of these layers\n"));
    }

    MI_MOD_END();
}				/* ............................. End of procedure CPGetTermLoc */

/* Compute a layer number given oct layer object.
----------------------------------------------------------------------------*/
short
CPCompLayerNum(layer)
    octObject	*layer;
{
    short		layerNum = CP_UNKNOWN_LAYER_NUM;

    MI_MOD_START("CPCompLayerNum");

    if      (!strcmp(layer->contents.layer.name,"POLY")) layerNum = 0;
    else if (!strcmp(layer->contents.layer.name,"MET1")) layerNum = 1;
    else if (!strcmp(layer->contents.layer.name,"MET2")) layerNum = 2;
    else if (!strcmp(layer->contents.layer.name,"COPO")) layerNum = CP_UNKNOWN_LAYER_NUM;
    else if (!strcmp(layer->contents.layer.name,"COM2")) layerNum = CP_UNKNOWN_LAYER_NUM;
    else if (!strcmp(layer->contents.layer.name,"COPS")) layerNum = CP_UNKNOWN_LAYER_NUM;
    else if (!strcmp(layer->contents.layer.name,"CONS")) layerNum = CP_UNKNOWN_LAYER_NUM;
    else {
	octObject facet;
	octGetFacet( layer, &facet );
	MIerror("Do not know what layer %s  in %s is (I assumed layer%d)\n",
		 layer->contents.layer.name, ohFormatName( &facet ), layerNum);
	
    }

    MI_MOD_END();
    return layerNum;
}				/* ............................. End of procedure CPCompLayerNum */
