/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

CPUtil.c

Utilities for interface of channel preparation for channel router.

Author(s):		Mitsuru Igusa
Revision history:	$Log: CPUtil.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:51  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.3  90/02/07  16:13:55  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.2  89/02/07  13:21:43  octtools
 * got rid of routeChan
 * 
 * Revision 1.1  89/01/26  14:20:17  octtools
 * Initial revision
 * 
 * Revision 1.10  88/06/15  14:09:00  mitch
 * Consider also OCT_BAGs when copying the contents of aTerms
 * 
 * Revision 1.9  88/04/27  16:58:38  mitch
 * Forgot to get rid of ASSERT.
 * 
 * Revision 1.7  88/04/25  15:18:23  mitch
 * Fixed bug in corner remapping of z-shaped channels.
 * 
 * Revision 1.6  88/04/24  16:53:05  mitch
 * Changed pin->newName to always mean the formal terminal name and
 * pin->name to the aTerm name.
 * 
 * Revision 1.1  88/03/31  11:23:35  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "miutil.h"
#include "CPInc.h"
/* Find instance on top and bottom of channel.  Also find the left and right
channels bordering the routing channel.  Instance on top means the top if
channel is horizontal and right instance if channel is vertical. Instance on
bottom is opposite of top instance.  Right channel is channel that is next
along the boundary of the top instance (if you go around the top instance
counterclockwise).
----------------------------------------------------------------------------*/
CPWhosOnTop()
{
short		k, c, z;
short		b0, b1;
short		ds, de;
CPLocDef	*bdy;
CPInstDef	*i;
short		bdyDir[2][2];		/* Direction of start and end pt of */
					/*   corners of instance bdy facing */
					/*   channel. */
					/*   1 = +x,  -1 = -x direction */
					/*   2 = +y,  -2 = -y direction */
short		nextChan[2];		/* Next channel from current channel */
					/*   of inst going clockwise around */
					/*   its boundary */
short		prevChan[2];		/* Previous channel ... */
CPLocDef	bdyMax[2];		/* Maximum corner point */

#define CP_WHOSTART	0
#define CP_WHOEND	1

MI_MOD_START("CPWhosOnTop");

for (k=0; k<2; k++) {			/* Gather some information */
    i = &CPInst[k];
    i->skipCornFlg = TRUE;

    if (k==CPExtInstNum) continue;	/* Skip external instances */

    for (c=0; c<i->numCorn; c++)
	if (i->corn[c].chanNum==CPCurChan) break;
    if (c>=i->numCorn) {
	MIerror("cannot find channel%d boundary of instance %s\n",
	    CPCurChan, CPInst[k].name);
	MSG(-1, ("This is probably due to an error in the way the program\n"));
	MSG(-1, ("handles z-shaped channels.  Try re-running the channel\n"));
	MSG(-1, ("definition tool without z-shaped channels\n"));
	exit(10);
    }
    nextChan[k] = i->corn[(c+1           )%i->numCorn].chanNum;
    prevChan[k] = i->corn[(c-1+i->numCorn)%i->numCorn].chanNum;
    bdy = i->bdy;
    bdyMax[k].x = bdyMax[k].y = -(MAXINT);
    for (z=0; z<2; z++) {
	if (!z) {
	    b0 = i->corn[c].bdyNum;
	    b1 = (b0+1)%i->numBdy;
	    CPInst[k].startBdyNum = b0;
	}
	else {
	    b0 = i->corn[(c+1)%i->numCorn].bdyNum;
	    b1 = (b0-1+i->numBdy)%i->numBdy;
	    CPInst[k].endBdyNum = b0;
	}
	if (bdy[b0].x==bdy[b1].x)
	then bdyDir[k][z] = (bdy[b0].y<bdy[b1].y ? 2 : -2);
	else bdyDir[k][z] = (bdy[b0].x<bdy[b1].x ? 1 : -1);
	bdyMax[k].x = MAX(bdy[b0].x, bdyMax[k].x);
	bdyMax[k].y = MAX(bdy[b0].y, bdyMax[k].y);
    }
}

if (CPExtInstNum<0) {			/* Not an external channel */
    ds = ABS(bdyDir[0][CP_WHOSTART]);
    de = ABS(bdyDir[0][CP_WHOEND  ]);
    DASSERT(ds==ABS(bdyDir[1][CP_WHOSTART]));
    DASSERT(de==ABS(bdyDir[1][CP_WHOEND  ]));
    CPWhere.instOnTop = 0;
    if (ds==de) {			/* Easy to find whos on top: */
	if (ds==1) {			/* Horizontal direction */
	    if (bdyDir[0][CP_WHOSTART]>0) {
		 if (bdyDir[1][CP_WHOSTART]>0 && bdyMax[1].y>bdyMax[0].y)
		     CPWhere.instOnTop = 1;
	    }
	    else if (bdyDir[1][CP_WHOSTART]<0 && bdyMax[1].y>bdyMax[0].y)
		     CPWhere.instOnTop = 1;
	}
	else {				/* Vertical direction */
	    if (bdyDir[0][CP_WHOSTART]<0) {
		 if (bdyDir[1][CP_WHOSTART]<0 && bdyMax[1].x>bdyMax[0].x)
		     CPWhere.instOnTop = 1;
	    }
	    else if (bdyDir[1][CP_WHOSTART]>0 && bdyMax[1].x>bdyMax[0].x)
		     CPWhere.instOnTop = 1;
	}
    }
    else NOTHING			/* Weird shaped channel */

    CPWhere.instOnBot   = (CPWhere.instOnTop==0 ? 1 : 0);
    CPWhere.chanOnRight = nextChan[CPWhere.instOnTop];
    CPWhere.chanOnLeft  = prevChan[CPWhere.instOnTop];

    if (ABS(bdyDir[CPWhere.instOnTop][CP_WHOSTART])==1)
    then CPWhere.chanDir = CP_H;
    else CPWhere.chanDir = CP_V;

    if (CPWhere.chanOnRight!=prevChan[CPWhere.instOnBot]) {
	MSG(-1, ("Right channels of chan %d are %d and %d (not same)\n",
	    CPCurChan, CPWhere.chanOnRight, prevChan[CPWhere.instOnBot]));
	CPInst[CPWhere.instOnTop].skipCornFlg = FALSE;
    }
    if (CPWhere.chanOnLeft!=nextChan[CPWhere.instOnBot]) {
	MSG(-1, ("Left channels of chan %d are %d and %d (not same)\n",
	    CPCurChan, CPWhere.chanOnLeft, nextChan[CPWhere.instOnBot]));
	CPInst[CPWhere.instOnBot].skipCornFlg = FALSE;
    }
}

else {					/* External channel */
    DASSERT(CPExtInstNum==1);
    CPWhere.instOnTop = 0;
    switch (bdyDir[0][CP_WHOSTART]) {
    case  2: CPWhere.instOnTop = 1;	/* Up    (left vert channel) */
    case -2: CPWhere.chanDir   = CP_V;	/* Down  (right vert channel) */
	     break;
    case -1: CPWhere.instOnTop = 1;	/* Left  (top hor channel) */
    default: CPWhere.chanDir   = CP_H;	/* Right (bottom hor channel) */
	     break;
    }

    if (CPWhere.instOnTop==0) {
	CPWhere.instOnBot   = 1;
	CPWhere.chanOnRight = nextChan[CPWhere.instOnTop];
	CPWhere.chanOnLeft  = prevChan[CPWhere.instOnTop];
    }
    else {
	CPWhere.instOnBot   = 0;
	CPWhere.chanOnRight = prevChan[CPWhere.instOnBot];
	CPWhere.chanOnLeft  = nextChan[CPWhere.instOnBot];
    }
}

MSG(1, ("iTop=%d iBot=%d cRight=%d cLeft=%d Dir=%d\n",
    CPWhere.instOnTop, CPWhere.instOnBot, CPWhere.chanOnRight,
    CPWhere.chanOnLeft, CPWhere.chanDir));

MI_MOD_END();
} /* ............................. End of procedure CPWhosOnTop */
/* Adjust top and bottom instance locations to avoid overlap.
----------------------------------------------------------------------------*/
CPAdjustInstLoc()
{
short		k;
int		botHi, topLo, delta;
int		b, eBdyNum, sBdyNum;
CPInstDef	*iTop, *iBot;
octObject	octInst;

MI_MOD_START("CPAdjustInstLoc");

if (CPExtInstNum<0) {			/* Not an external channel */
    botHi = -MAXINT;
    topLo =  MAXINT;
    for (k=0; k<2; k++) {
	if (k==CPWhere.instOnTop) {	/* Top chan: compute topLo */
	    iTop = &CPInst[k];
	    sBdyNum = iTop->startBdyNum;
	    eBdyNum = iTop->endBdyNum;
	    for (b=sBdyNum; ; b=(b+1)%iTop->numBdy) {
		if (CPWhere.chanDir==CP_V)
		then topLo = MIN(topLo, iTop->bdy[b].x);
		else topLo = MIN(topLo, iTop->bdy[b].y);
		if (b==eBdyNum) break;
	    }
	}
	else {				/* Bot chan: compute botHi */
	    iBot = &CPInst[k];
	    sBdyNum = iBot->startBdyNum;
	    eBdyNum = iBot->endBdyNum;
	    for (b=sBdyNum; ; b=(b+1)%iBot->numBdy) {
		if (CPWhere.chanDir==CP_V)
		then botHi = MAX(botHi, iBot->bdy[b].x);
		else botHi = MAX(botHi, iBot->bdy[b].y);
		if (b==eBdyNum) break;
	    }
	}
    }

    if (botHi>topLo) {			/* Overlap found: */
	delta = botHi-topLo;
	MSG(-1, ("Expanding channel by %d\n", delta));
					/* Get top inst */
	OH_ASSERT(ohGetById(&octInst, iTop->newInstId));

	if (CPWhere.chanDir==CP_V) {	/* Move stuff */
	    octInst.contents.instance.transform.translation.x += delta;
	    iTop->loc.x += delta;
	}
	else {
	    octInst.contents.instance.transform.translation.y += delta;
	    iTop->loc.y += delta;
	}
	OH_ASSERT(octModify(&octInst));

	for (b=0; b<iTop->numBdy; b++) { /* Move bdy */
	    if (CPWhere.chanDir==CP_V)
	    then iTop->bdy[b].x += delta;
	    else iTop->bdy[b].y += delta;
	}
    }
}
MI_MOD_END();
} /* ............................. End of procedure CPAdjustInstLoc */
/* Create fTerm for terminal (under octMaster) discribed by CPPin pin.
----------------------------------------------------------------------------*/
CPCreateFTerm(octMaster, octFTerm, pin, octTermInst)

octObject	*octMaster;
octObject	*octFTerm;
CPPinDef	*pin;
octObject	*octTermInst;
{
CPPinDef	*pin2;
short		t;
char		tempName[1024];

MI_MOD_START("CPCreateFTerm");
					/* Check if name matches with other */
					/*   formal terminal names */
for (pin2=CPPin->next; pin2; pin2=pin2->next)
    if (!strcmp(pin2->newName, pin->name)) break;

if (pin2) {				/* Not unique: */
    MSG(2, ("Finding new fTerm name. oldName = %s\n", pin->name));
    for (t=0; ; t++) {			/* Generate new fTerm name */
	if (octTermInst)
	     (void) sprintf(tempName,"%s.%s_%d",
		 octTermInst->contents.instance.name, pin->name, t);
	else (void) sprintf(tempName,"%s.%d", pin->name, t);
	for (pin2=CPPin->next; pin2; pin2=pin2->next)
	    if (!strcmp(pin2->name, tempName) &&
		pin2->chanNum!=CPCurChan) break;
	if (!pin2) break;
    }
    pin->newName = STR_COPY(tempName);
    if (!octTermInst) {			/* PTerm aTerm name always the same */
	FREE_STRING(pin->name);		/*   as fTerm name. */
	pin->name = pin->newName;
    }
}
			   		 /* Creating fTerm */
MSG(2, ("Creating new fTerm %s\n", pin->newName));
OH_ASSERT(ohCreateTerm(octMaster, octFTerm, pin->newName));

MI_MOD_END();
} /* ............................. End of procedure CPCreateFTerm */
/* Attach child objects of fromObj to toObj.
----------------------------------------------------------------------------*/
CPOctCopyContents(fromObj, toObj, mask)

octObject	*fromObj;
octObject	*toObj;
octObjectMask	mask;
{
    octObject	octObj;
    octObject	octObjCopy;
    octGenerator	genContents;

    MI_MOD_START("CPOctCopyContents");

    OH_ASSERT(octInitGenContents(fromObj, mask, &genContents));
    while (octGenerate(&genContents, &octObj)==OCT_OK) {
	octObjCopy.objectId = octObj.objectId;
	OH_ASSERT(octCreate(toObj, &octObj));
	OH_ASSERT(octGetById(&octObjCopy));
	CPOctCopyContents(&octObjCopy, &octObj, mask);
    }
    MI_MOD_END();
}				/* ............................. End of procedure CPOctCopyContents */
/* Attach (all except OCT_INST) contents and containers of oldTerm to newTerm.
----------------------------------------------------------------------------*/
CPOctSwapTerm(oldTerm, newTerm)

octObject	*oldTerm;
octObject	*newTerm;
{
    octObject	octObj;
    octGenerator	genContents;
    octGenerator	genContainers;

    MI_MOD_START("CPOctSwapTerm");

    DASSERT(oldTerm->type==OCT_TERM || oldTerm->type==OCT_BAG);
    DASSERT(newTerm->type==OCT_TERM);

    MSG(2, ("Swapping attachments of old %cTerm %s to aTerm %s\n",
	    (oldTerm->type==OCT_TERM ? 'a' : 'p'),
	    (oldTerm->type==OCT_TERM ? oldTerm->contents.term.name : "xxx"),
	    newTerm->contents.term.name));

    OH_ASSERT(octInitGenContents(oldTerm, OCT_ALL_MASK, &genContents));
    while (octGenerate(&genContents, &octObj)==OCT_OK)
      OH_ASSERT(octAttach(newTerm, &octObj));

    OH_ASSERT(octInitGenContainers(oldTerm, OCT_ALL_MASK, &genContainers));
    while (octGenerate(&genContainers, &octObj)==OCT_OK)
      if (octObj.type!=OCT_INSTANCE)
	OH_ASSERT(octAttach(&octObj, newTerm));

    MI_MOD_END();
}				/* ............................. End of procedure CPOctSwapTerm */
/* Given octSubNet find subNetNumber.
----------------------------------------------------------------------------*/
CPNetToNetNum(subNetNum, octSubNet)

short		*subNetNum;
octObject	*octSubNet;
{
short		sn;

MI_MOD_START("CPTermToNet");

*subNetNum = CP_NULL_NET;
for (sn=0; sn<CPNumSubNet; sn++)
    if (octIdsEqual(CPNet[sn].subNetId, octSubNet->objectId)) break;
DASSERT(sn<CPNumSubNet);
*subNetNum = sn;

MSG(2, ("Determined subNetNum = %d (from subNetOctId=%s)\n",
    sn, octFormatId(octSubNet->objectId)));

MI_MOD_END();
} /* ............................. End of procedure CPNetToNetNum */
/* Recursive routine to initialize channel routing order.
----------------------------------------------------------------------------*/
CPChanOrder(octChan)

octObject	*octChan;
{
octObject	octChOrd;
short		c;

MI_MOD_START("CPChanOrder");

OH_ASSERT(ohGetByPropName(octChan, &octChOrd, "CDChanOrder"));
CPNumChan = octChOrd.contents.prop.value.integer+1;

CPChanOctId = ALLOC_ARRAY(CPNumChan, octId);
for (c=0; c<CPNumChan; c++)
    CPChanOctId[c] = oct_null_id;

CPInitChanOctId(octChan);

MI_MOD_END();
} /* ............................. End of procedure CPChanOrder */
/* Initialize channel oct ids.
----------------------------------------------------------------------------*/
CPInitChanOctId(octChan)

octObject	*octChan;
{
octObject	octChOrd;
octObject	octSubChan;
octGenerator	genSubChan;
int		c;

MI_MOD_START("CPInitChanOctId");

if (ohGetByPropName(octChan, &octChOrd, "CDChanOrder")==OCT_OK) {
    c = octChOrd.contents.prop.value.integer;
    DRANGE(0, c, CPNumChan);
    DASSERT(octIdIsNull(CPChanOctId[c]));
    CPChanOctId[c] = octChan->objectId;

    OH_ASSERT(octInitGenContents(octChan, OCT_BAG_MASK, &genSubChan));
    while (octGenerate(&genSubChan, &octSubChan)==OCT_OK) {
	if (strcmp(octSubChan.contents.bag.name, "CDChanDef")) continue;
	CPInitChanOctId(&octSubChan);
    }
}

MI_MOD_END();
} /* ............................. End of procedure CPInitChanOctId */
