/*LINTLIBRARY*/
/*
Copyright (C) 1986 University of California (Berkeley). All rights reserved.

MIOctUtil.c

Utility programs for oct database.

Author(s):		Mitsuru Igusa
Revision history:	$Log: MIOctUtil.c,v $
Revision history:	Revision 1.1.1.1  2002/12/31 15:50:51  saj
Revision history:	octtools 5.2 in a GNU format
Revision history:	
 * Revision 1.3  90/02/07  16:13:56  octtools
 * Checkin for 3.5 distribution
 * 
 * Revision 1.1  89/01/26  14:20:18  octtools
 * Initial revision
 * 
 * Revision 1.3  88/04/24  16:55:50  mitch
 * Use octOpenMaster instead of MIOctOpenMasterOfInst.
 * 
 * Revision 1.1  88/03/31  14:25:01  mitch
 * Initial revision
 * 
----------------------------------------------------------------------------*/
#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "tr.h"
#include "miutil.h"
/* Given instance and master, return bounding polygon of instance.  Always
returns points in counter clockwise direction.  orientFlg is set when original
transformation of master gives clockwise orientation.
----------------------------------------------------------------------------*/
MIOctGetInstBdy(octMaster, octInst, octPts, numOctPts, orientFlg)

octObject	*octMaster;
octObject	*octInst;
struct octPoint	*octPts;
int32		*numOctPts;		/* Init to max num oct pts initially */
Boolean		*orientFlg;
{
#define		MAX_OCT_PTS 1000
octObject	octLayer;
octObject	octGeom, octNGeom;
octGenerator	genGeom;
String		MIOrientToString();
int		minP;
int		xy, minXY;
int		p, pr;
octCoord	t;
Boolean		nonManhattanFlg = FALSE;

MI_MOD_START("MIOctGetInstBdy");

if (ohGetByLayerName(octMaster, &octLayer, "PLACE")!=OCT_OK) {
    MSG(2, ("WARN Cannot find layer PLACE in cell=%s view=%s\n",
	octMaster->contents.facet.cell, octMaster->contents.facet.view));
    MSG(2, ("Using bounding box for channel definition"));
    OH_ASSERT(octBB(octInst, &octGeom.contents.box));
    octPts[0].x = octPts[3].x = octGeom.contents.box.lowerLeft.x;
    octPts[0].y = octPts[1].y = octGeom.contents.box.lowerLeft.y;
    octPts[2].x = octPts[1].x = octGeom.contents.box.upperRight.x;
    octPts[2].y = octPts[3].y = octGeom.contents.box.upperRight.y;
    DASSERT(octPts[0].x<octPts[2].x && octPts[0].y<octPts[2].y)
    *numOctPts = 4;
    *orientFlg = FALSE;
    goto MIOctGetInstBdyEnd;
}
OH_ASSERT(ohGetByLayerName(octMaster, &octLayer, "PLACE"));
OH_ASSERT(octInitGenContents(&octLayer, OCT_GEO_MASK, &genGeom));
OH_ASSERT(octGenerate(&genGeom, &octGeom));
					/* More than one geometry */
if (octGenerate(&genGeom, &octNGeom)==OCT_OK) {
    MIerror("More than 1 PLACE geom - using polygon routine\n");
    printf(
"The oct facet \"%s:%s\" has more than one geometry on the PLACE layer\n\
and this cannot be handled by the program at this time.  To correct this,\n\
you must change your placement or macrocell or whatever such that you don't\n\
have something that looks like this:\n\
+--------+\n\
|	 |\n\
| inst A |\n\
|	 |\n\
+--------+\n\
\n\
		+--------+\n\
		|	 |\n\
		| inst B |\n\
		|	 |\n\
		+--------+\n\
which after running vulcan will result in two separate boxes.\n",
    octMaster->contents.facet.cell, octMaster->contents.facet.view);
    exit(96);

#ifdef MULT_GEOM
    OH_ASSERT(octInitGenContents(&octLayer, OCT_GEO_MASK, &genGeom));
    tempGeom = NULL;
    readGeometry(&genGeom, &tempGeom);
    *numOctPts = perimetralPolygon(tempGeom, &tempOctPts);
    for (p=0; p<*numOctPts; p++) {
	MSG(3, ("Poly pt %2d: (%d %d)\n", p, tempOctPts[p]));
	octPts[p].x = tempOctPts[p].x;
	octPts[p].y = tempOctPts[p].y;
    }
#endif
}
/* else ...........................   <---- was deleted */

switch(octGeom.type) {
    case OCT_BOX:			/* Make sure pts are CCW! */
	*numOctPts = 4;
	octPts[0].x = octPts[3].x = octGeom.contents.box.lowerLeft.x;
	octPts[0].y = octPts[1].y = octGeom.contents.box.lowerLeft.y;
	octPts[2].x = octPts[1].x = octGeom.contents.box.upperRight.x;
	octPts[2].y = octPts[3].y = octGeom.contents.box.upperRight.y;
	DASSERT(octPts[0].x<octPts[2].x && octPts[0].y<octPts[2].y);
	break;
    case OCT_POLYGON:
	*numOctPts = MAX_OCT_PTS;
	OH_ASSERT(octGetPoints(&octGeom, numOctPts, octPts));
	DURANGE(*numOctPts, MAX_OCT_PTS);
	break;
    default:
	MIerror("Program deals with boxes & polygons not geomType=%d\n",
	    octGeom.type);
	break;
}
octFreeGenerator( &genGeom );


MSG(3, ("transform type %s + (%d %d)\n",
    MIOrientToString(octInst->contents.instance.transform.transformType),
    octInst->contents.instance.transform.translation));

minXY = MAXINT;				/* Use minXY pt to find orientation */
for (p=0; p<*numOctPts; p++) {		/* Transform oct coordinates */
    MSG(3, ("master pt %d: (%d %d)\n", p, octPts[p]))
    tr_oct_transform(&octInst->contents.instance.transform,
	&octPts[p].x, &octPts[p].y);
    MSG(3, ("trans  pt  (%d %d)\n", octPts[p]));
					/* Manhattan ? */
    if (p>0 && octPts[p].x!=octPts[p-1].x && octPts[p].y!=octPts[p-1].y)
	nonManhattanFlg = TRUE;

    xy = (int)(octPts[p].x+octPts[p].y);
    if (xy<minXY) {
	minXY = xy;
	minP  = p;
    }
}
if (octPts[0].x!=octPts[p-1].x && octPts[0].y!=octPts[p-1].y)
    nonManhattanFlg = TRUE;
if (nonManhattanFlg) {
    MIerror("Polygon pts are not manhattan.  Check %s %s %s facet\n",
	octMaster->contents.facet.cell,
	octMaster->contents.facet.view,
	octMaster->contents.facet.facet);
    pr = *numOctPts-1;
    for (p=0; p<*numOctPts; p++) {
	MSG(-1, ("Polygon pt %d: (%d %d)%s\n",
	    p, octPts[p],
	    (octPts[p].x!=octPts[pr].x && octPts[p].y!=octPts[pr].y ?
	     "non manhattan with prev pt" : "")));
	pr = p;
    }
}

if (octPts[minP].y<octPts[(minP+1)%(*numOctPts)].y) {
    MSG(2, ("Bdy is clockwise\n")); /* Clockwise - reverse direction */
    for (p=0, pr=(*numOctPts)-1; p<pr; p++, pr--) {
	t            = octPts[pr].x;
	octPts[pr].x = octPts[p ].x;
	octPts[p ].x = t;
	t            = octPts[pr].y;
	octPts[pr].y = octPts[p ].y;
	octPts[p ].y = t;
    }
    *orientFlg  = TRUE;
}
else *orientFlg = FALSE;

MIOctGetInstBdyEnd:
MI_MOD_END();
} /* ............................. End of procedure MIOctGetInstBdy */
/* Create a segment
----------------------------------------------------------------------------*/
MIOctCreateSegment(net, term1, term2, x1, y1, x2, y2, layer, width)

octObject	*net;
octObject	*term1, *term2;
int		x1, y1, x2, y2;
octObject	*layer;
octCoord	width;
{
octObject	path;
struct octPoint	octPts[2];
int		numOctPts = 2;

MI_MOD_START("MIOctCreateSegment")

MSG(1, ("Creating segment of net %s\n", net->contents.net.name));

OH_ASSERT(ohCreatePath(net, &path, width));
OH_ASSERT(octAttach(&path, term1));
OH_ASSERT(octAttach(&path, term2));
octPts[0].x = x1;
octPts[0].y = y1;
octPts[1].x = x2;
octPts[1].y = y2;
OH_ASSERT(octAttach(layer, &path));
OH_ASSERT(octPutPoints(&path, numOctPts, octPts));

MI_MOD_END();
} /* ............................. End of procedure MIOctCreateSegment */
/* Create a contact or via
----------------------------------------------------------------------------*/
MIOctCreateVia(master, techDir, rInstBag, viaInst, viaTerm, viaName, x, y)

octObject	*master;
String		techDir;
octObject	*rInstBag;
octObject	*viaInst;
octObject	*viaTerm;
String		viaName;
int		x, y;
{
static		viaNum = 0;
char		viaInstName[16];
char		viaMasterName[248];

MI_MOD_START("MIOctCreateVia");

MSG(1, ("Creating via%d: %s at (%d %d)\n", viaNum, viaName, x, y));

(void) sprintf(viaInstName,"via%d", viaNum++);
(void) sprintf(viaMasterName,"%s/%s", techDir, viaName);
OH_ASSERT(ohCreateInstance(master, viaInst,
    viaInstName, viaMasterName, "physical", x, y, OCT_NO_TRANSFORM));
OH_ASSERT(octAttach(rInstBag, viaInst));
OH_ASSERT(octGenFirstContent(viaInst, OCT_TERM_MASK, viaTerm));

MI_MOD_END();
} /* ............................. End of procedure MIOctCreateVia */
/* Get box of instance terminal.  Master of instance is given,  however if
the master's objectId is null then open master facet.
----------------------------------------------------------------------------*/
MIOctGetTermBox(box, term, inst, master)

octObject	*box;
octObject	*term;
octObject	*inst;
octObject	*master;
{
    octGenerator	genBox;
    octObject	mTerm;
    int		x, y;

    MI_MOD_START("MIOctGetTermBox");

    if (octIdIsNull(master->objectId)) {
	OH_ASSERT(ohOpenMaster(master, inst, "interface", "r"));
	VOVinputFacet( master );
    }
    OH_ASSERT(ohGetByTermName(master, &mTerm, term->contents.term.name));
    OH_ASSERT(octGenFirstContent(&mTerm, OCT_BOX_MASK, box));
    /* Transform coordinates of box */
    tr_oct_transform(&inst->contents.instance.transform,
		     &box->contents.box.lowerLeft.x , &box->contents.box.lowerLeft.y);
    tr_oct_transform(&inst->contents.instance.transform,
		     &box->contents.box.upperRight.x , &box->contents.box.upperRight.y);

    x = box->contents.box.lowerLeft.x;
    box->contents.box.lowerLeft .x = MIN(x, box->contents.box.upperRight.x);
    box->contents.box.upperRight.x = MAX(x, box->contents.box.upperRight.x);
    y = box->contents.box.lowerLeft.y;
    box->contents.box.lowerLeft .y = MIN(y, box->contents.box.upperRight.y);
    box->contents.box.upperRight.y = MAX(y, box->contents.box.upperRight.y);

    MSG(1, ("Term %s of inst %s at (%d %d) (%d %d)\n",
	    term->contents.term.name, inst->contents.instance.name,
	    box->contents.box.lowerLeft, box->contents.box.upperRight));

    MI_MOD_END();
}				/* ............................. End of procedure MIOctGetTermBox */
/* From octChild get octParent by type and name of octParent.
----------------------------------------------------------------------------*/
MIOctGetByNameContainers(octChild, octParent)

octObject	*octChild;
octObject	*octParent;
{
octGenerator	genPar;
octObject	octPar;
int		rc = OCT_ERROR;
Boolean		foundFlg = FALSE;
String		name;

switch (octParent->type) {
case OCT_BAG :
    name = octParent->contents.bag.name;
    OH_ASSERT(octInitGenContainers(octChild, OCT_BAG_MASK, &genPar));
    while (octGenerate(&genPar, &octPar)==OCT_OK) {
	if (!strcmp(octPar.contents.bag.name, name)) {
	    foundFlg = TRUE;
	    break;
	}
    }
    octFreeGenerator( &genPar );
    break;

case OCT_PROP:
    name = octParent->contents.prop.name;
    OH_ASSERT(octInitGenContainers(octChild, OCT_PROP_MASK, &genPar));
    while (octGenerate(&genPar, &octPar)==OCT_OK) {
	if (!strcmp(octPar.contents.prop.name, name)) {
	    foundFlg = TRUE;
	    break;
	}
    }
    octFreeGenerator( &genPar );

    break;

case OCT_NET:
    name = octParent->contents.net.name;
    OH_ASSERT(octInitGenContainers(octChild, OCT_NET_MASK, &genPar));
    while (octGenerate(&genPar, &octPar)==OCT_OK) {
	if (!strcmp(octPar.contents.net.name, name)) {
	    foundFlg = TRUE;
	    break;
	}
    }
    octFreeGenerator( &genPar );

    break;

default      :
    MIerror("Cannot handle container type %d\n", octParent->type);
}

if (foundFlg) {
    rc = OCT_OK;
    OH_ASSERT(ohGetById(octParent, octPar.objectId));
}

return(rc);
} /* ............................. End of procedure MIOctGetByNameContainers */
/* Multiply orients (o1 & o2) and return resultant orientation (o1 x o2).
If o1 or o2 is negative then use o1 or o2's inverse orientation.
----------------------------------------------------------------------------*/
short
MIMultOrient(o1, o2)

short		o1;		/* 1st orientation transform performed */
short		o2;		/* 2nd orientation transform performed */
{
static short	orient[8][4] = {{ 1, 0, 0, 1}, /* OCT_NO_TRANSFORM */
				{-1, 0, 0, 1}, /* OCT_MIRROR_X */
				{ 1, 0, 0,-1}, /* OCT_MIRROR_Y */
				{ 0,-1, 1, 0}, /* OCT_ROT90 */
				{-1, 0, 0,-1}, /* OCT_ROT180 */
				{ 0, 1,-1, 0}, /* OCT_ROT270 */
				{ 0,-1,-1, 0}, /* OCT_MX_ROT90 */
				{ 0, 1, 1, 0}};/* OCT_MY_ROT90 */
static short	invOri[8] = { (short)OCT_NO_TRANSFORM,
			      (short)OCT_MIRROR_X,
			      (short)OCT_MIRROR_Y,
			      (short)OCT_ROT270,
			      (short)OCT_ROT180,
			      (short)OCT_ROT90,
			      (short)OCT_MY_ROT90,
			      (short)OCT_MX_ROT90};
short		newOrient[4];
short		o1xo2;
String		MIOrientToString();

MI_MOD_START("MIMultOrient")

if (o1<0) {
    DURANGE(-o1, 8);
    o1 = invOri[-o1];
}
if (o2<0) {
    DURANGE(-o2, 8);
    o2 = invOri[-o2];
}
DURANGE(o1, 8);
DURANGE(o2, 8);
					/* Matrix multiply for new orient */
newOrient[0] = orient[o2][0]*orient[o1][0] + orient[o2][1]*orient[o1][2];
newOrient[1] = orient[o2][0]*orient[o1][1] + orient[o2][1]*orient[o1][3];
newOrient[2] = orient[o2][2]*orient[o1][0] + orient[o2][3]*orient[o1][2];
newOrient[3] = orient[o2][2]*orient[o1][1] + orient[o2][3]*orient[o1][3];

for (o1xo2=0; o1xo2<8; o1xo2++)		/* Match orientation with primitive */
    if (newOrient[0]==orient[o1xo2][0] &&
	newOrient[1]==orient[o1xo2][1] &&
	newOrient[2]==orient[o1xo2][2] &&
	newOrient[3]==orient[o1xo2][3]) break;
DASSERT(o1xo2<8);

MSG(1, ( "Orient %s x %s = %s\n",
    MIOrientToString(o1), MIOrientToString(o2), MIOrientToString(o1xo2)));

MI_MOD_END();
return(o1xo2);
} /* ............................. End of procedure MIMultOrient */

#ifdef MI_OCT_TRANSFORM
/* Apply oct linear transform to points x and y.
NOTE: Rotation is CLOCKWISE and Mirroring is about opposite axis.
----------------------------------------------------------------------------*/
MIOctTransform(x, y, t)

octCoord		*x, *y;
struct octTransform	*t;
{
octCoord		tx, ty;

MI_MOD_START("MIOctTransform");

switch (t->transformType) {
case OCT_NO_TRANSFORM : break;
case OCT_MIRROR_X     : *x = -(*x); break;
case OCT_ROT180       : *x = -(*x);
case OCT_MIRROR_Y     : *y = -(*y); break;
case OCT_ROT90        : ty = *y; *y =   *x ; *x = -ty; break;
case OCT_ROT270       : tx = *x; *x =   *y ; *y = -tx; break;
case OCT_MY_ROT90     : ty = *y; *y =   *x ; *x =  ty; break;
case OCT_MX_ROT90     : ty = *y; *y = -(*x); *x = -ty; break;
case OCT_FULL_TRANSFORM: tx = t->generalTransform[0][0]*(*x)+
			      t->generalTransform[0][1]*(*y);
			 ty = t->generalTransform[1][0]*(*x)+
			      t->generalTransform[1][1]*(*y); break;
default	              :
    MIerror("Cannot process transformType %d\n",
	t->transformType);
}
MSG(2, ("Transformation = %d, offset=(%d %d)\n",
    t->transformType, t->translation.x, t->translation.y));

*x += t->translation.x;
*y += t->translation.y;
MI_MOD_END();
} /* ............................. End of procedure MIOctTransform */
#endif

/* Convert orientation into string readable form.
----------------------------------------------------------------------------*/
String
MIOrientToString(orient)

short		orient;
{
static char	s[10];

switch(orient) {
case OCT_NO_TRANSFORM : return("IDENT");
case OCT_MIRROR_X     : return("MX");
case OCT_MIRROR_Y     : return("MY");
case OCT_ROT90        : return("R90");
case OCT_ROT180       : return("R180");
case OCT_ROT270       : return("R270");
case OCT_MX_ROT90     : return("MX90");
case OCT_MY_ROT90     : return("MY90");
default	              : (void) sprintf(s,"%d ?", orient);
			return(s);
}
} /* ............................. End of procedure MIOrientToString */
