#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "tap.h"
#include "wolfe.h"

static void drop_stub_path();
static void drop_power_stub();
static void drop_stub_path_not_attached();
static int find_power_pins();
octStatus get_connector();

double peak_vdd_i = 0.0, peak_gnd_i = 0.0;    /* Peak power & gnd currents */
int supply_reqd, ground_reqd;    /* Num of power & gnd terms needed */

struct power_terms{
    octObject left_vdd;
    struct octBox left_vdd_bb;
    octObject right_vdd;
    struct octBox right_vdd_bb;
    octObject left_gnd;
    struct octBox left_gnd_bb;
    octObject right_gnd;
    struct octBox right_gnd_bb;
} 
*power_array;

int connect_vdd_gnd(facet)
octObject *facet;
{
    octObject interface, power_conn,instance, aterm1, aterm2;
    octObject conn_bag, last_vdd_term, last_gnd_term;
    octObject power_form,gnd_form,power_form1, gnd_form1,inter_power;
    octObject power_t,prev_vdd_t, prev_gnd_t, prev_vdd_t1, prev_gnd_t1;
    octObject vdd, gnd, layer,vertlayer;
    octObject power_inst, ground_inst;
    octObject path, power_prop;
    octGenerator genPowerTerm;
    struct octBox bb1, bb2, last_vdd_bb, last_gnd_bb;
    struct octBox prev_vdd_bb1, prev_gnd_bb1;
    block_record_t *p;
    int i,index, first, power_impl, ground_impl;
    struct octBox prev_vdd_bb , prev_gnd_bb;

    OH_ASSERT(ohGetOrCreateLayer(facet, &layer, hlayer_name));
    OH_ASSERT(ohGetOrCreateLayer(facet, &vertlayer, vlayer_name));

    find_power_net(facet,&vdd,"SUPPLY");
    find_power_net(facet,&gnd,"GROUND");

    power_t.type = OCT_TERM;
    path.type = OCT_PATH;

    supply_reqd = (power_position == BOTH)? 2:1;
    ground_reqd = (power_position == BOTH)? 2:1;

    if (ohOpenInterface(&interface,facet,"a") < OCT_OK){
        (void) printf("Unable to open interface of cell\n");
        return 1;
    }
    OH_ASSERT(octInitGenContents(&interface,OCT_TERM_MASK,&genPowerTerm));
    while (octGenerate(&genPowerTerm,&inter_power) == OCT_OK){
        if (ohGetByPropName(&inter_power,&power_prop,"TERMTYPE") == OCT_OK){
            if (!strcmp(power_prop.contents.prop.value.string,"SUPPLY")){
                supply_reqd--;
                power_form1 = power_form;
                OH_ASSERT(ohGetByTermName(facet,&power_form,
                	inter_power.contents.term.name));
            } else if(!strcmp(power_prop.contents.prop.value.string,"GROUND")){
                ground_reqd--;
                gnd_form1 = gnd_form;
                OH_ASSERT(ohGetByTermName(facet,&gnd_form,
                	inter_power.contents.term.name));
            }
        }
    }

    /* Create the required number of terminals if enuff were not found */
    if(supply_reqd == 1){
	if (power_position == BOTH){
	    OH_ASSERT(ohCreateTerm(facet,&power_form1,"wolfe_Vdd1") );
	} else {
	    OH_ASSERT(ohCreateTerm(facet,&power_form,"wolfe_Vdd") );
	}
    } else if( supply_reqd == 2){
        OH_ASSERT(ohCreateTerm(facet,&power_form,"wolfe_Vdd") );
        OH_ASSERT(ohCreateTerm(facet,&power_form1,"wolfe_Vdd1") );
    }

    if( ground_reqd == 1){
	if (power_position == BOTH){
	    OH_ASSERT(ohCreateTerm(facet,&gnd_form1,"wolfe_Gnd1") );
	} else {
	    OH_ASSERT(ohCreateTerm(facet,&gnd_form,"wolfe_Gnd") );
	}
    } else if( ground_reqd == 2){
        OH_ASSERT(ohCreateTerm(facet,&gnd_form,"wolfe_Gnd") );
        OH_ASSERT(ohCreateTerm(facet,&gnd_form1,"wolfe_Gnd1") );
    }

    /* 
     * IF we already have an instantiation of the formal terminal 
     * (can happen only if the power/ground net was connected to
     * some signal terminals and then the net must have been
     * brought out to a side and the appropraite connector added)
     * THEN 
     * Record the instance and add an appropriate stub after
     * constructing the power/ground rail
     */
    power_impl = (octGenFirstContent(&power_form, OCT_TERM_MASK, &power_inst) == OCT_OK ? 1 : 0);
    ground_impl = (octGenFirstContent(&gnd_form, OCT_TERM_MASK, &ground_inst) == OCT_OK ? 1 : 0);


    /* Allocate the memory for the power array */
    power_array = ALLOC(struct power_terms, num_rows+1);

    /*
     * Loop for each row of cells --- connect the vdd and gnd terminals
     */
    for(i = 1; i <= num_rows; i++) {

        /* Loop for each cell on the row in reverse order */
        first = 1;
        power_array[i].left_vdd.objectId =
          power_array[i].right_vdd.objectId =
          power_array[i].left_gnd.objectId =
          power_array[i].right_gnd.objectId = oct_null_id;
        
        for(p = block_list[i]; p != NIL(block_record_t); p = p->next) {

            instance.objectId = p->objectId;
            OH_ASSERT(octGetById(&instance));

            /* Do the power pins */
            if (find_power_pins(&instance, &aterm1, &aterm2, &bb1, 
            &bb2, "SUPPLY",&peak_vdd_i) != 0)
                return 1;
            if (! first) {
                drop_stub_path(&vdd, &layer, &last_vdd_term, &aterm2, 
                &last_vdd_bb, &bb2,LEFT,&path);
            }
            else{
                /* FIRST terminal in the row -- add to the power array */
                power_array[i].left_vdd = aterm1;
                power_array[i].left_vdd_bb = bb1;
            }
            if ( p->next == NIL(block_record_t)){
                /* Last terminal in the row -- Add to power array */
                power_array[i].right_vdd = aterm2;
                power_array[i].right_vdd_bb  = bb2;
            }
            last_vdd_term = aterm1;
            last_vdd_bb = bb1;

            /* Do the ground pins */
            if (find_power_pins(&instance, &aterm1, &aterm2, &bb1, 
            &bb2, "GROUND",&peak_gnd_i) != 0)
                return 1;
            if (! first) {
                drop_stub_path(&gnd, &layer, &last_gnd_term, &aterm2, 
                &last_gnd_bb, &bb2,LEFT,&path);
            }
            else{
                /* FIRST terminal in the row -- add to the power array */
                power_array[i].left_gnd = aterm1;
                power_array[i].left_gnd_bb = bb1;
            }
            if ( p->next == NIL(block_record_t)){
                /* Last terminal in the row -- Add to power array */
                power_array[i].right_gnd = aterm2;
                power_array[i].right_gnd_bb = bb2;
            }
            last_gnd_term = aterm1;
            last_gnd_bb = bb1;

            first = 0;
        }
    }

    if ( debug > 2){
        /* dump the end positions of the the power terminals */
        (void) printf("POSITIONS OF THE POWER TERMINALS \n");
        for( i=1; i <= num_rows; i++){
            (void) octPrintObject(stdout,&(power_array[i].left_vdd), 0);
            (void) octPrintObject(stdout,&(power_array[i].right_vdd), 0);
            (void) octPrintObject(stdout,&(power_array[i].left_gnd), 0);
            (void) octPrintObject(stdout,&(power_array[i].right_gnd), 0);
        }
    }
    /* Connect the formal terminals to the net */
    OH_ASSERT(octAttachOnce(&vdd,&power_form));
    OH_ASSERT(octAttachOnce(&gnd,&gnd_form));
    if( power_position == BOTH){
        OH_ASSERT(octAttachOnce(&vdd,&power_form1));
        OH_ASSERT(octAttachOnce(&gnd,&gnd_form1));
    }

    /* Get the connector using a call to tapGetConnector 
        -- the instances are power_conn and ground_conn    */
    if (get_connector(vertlayer, layer, power_width * units_per_lambda, 
	    (last_vdd_bb.upperRight.y - last_vdd_bb.lowerLeft.y),
	    &power_conn) != OCT_OK){
	(void)fprintf(stdout,"WARNING: Connnector of size %dx%d not available in the technology\n", power_width,
	(last_vdd_bb.upperRight.y - last_vdd_bb.lowerLeft.y)/units_per_lambda);
	(void)fprintf(stdout,"\tSUPPLY and GROUND nets will appear in disconnected segments.\n");
	return 0;
    }

    if ( debug){
        (void) printf("Vdd RAIL IS ON %s SIDE\n",print_location[power_position]);
    }
    /* The instances of the power terminals are to be
        put in the connectors bag */

    OH_ASSERT(ohGetOrCreateBag(facet,&conn_bag,"CONNECTORS"));
    first = 1;
    for( i =1,index = 1; i <= num_rows; i++,index++){
	if (octIdIsNull(power_array[i].left_vdd.objectId)) {
	    /* no cells on this row */
	    continue;
        }
        if(power_position == LEFT){
            drop_power_stub(facet,&power_form,&vdd,&layer,
            &power_conn,"vdd_connector",index,&(power_array[i].left_vdd),
            &(power_array[i].left_vdd_bb), &power_t,&bb2,0,LEFT,&conn_bag);
            /* Add a path to the previous point */
            if(!first){
                drop_stub_path_not_attached(&vdd,&vertlayer,&prev_vdd_t,&power_t,&prev_vdd_bb,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&power_form,&path));
            }
            prev_vdd_t = power_t;
            prev_vdd_bb = bb2;

            drop_power_stub(facet,&gnd_form,&gnd,&layer,
            &power_conn,"gnd_connector",index,&(power_array[i].right_gnd),
            &(power_array[i].right_gnd_bb), &power_t,&bb2,0,RIGHT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&gnd,&vertlayer,&prev_gnd_t,&power_t,&prev_gnd_bb,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&gnd_form,&path));
            }
            prev_gnd_t = power_t;
            prev_gnd_bb = bb2;

        }
        else if (power_position == RIGHT){
            drop_power_stub(facet,&power_form,&vdd,&layer,
            &power_conn,"vdd_connector",index,&(power_array[i].right_vdd),
            &(power_array[i].right_vdd_bb), &power_t,&bb2,0,RIGHT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&vdd,&vertlayer,&prev_vdd_t,&power_t,&prev_vdd_bb,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&power_form,&path));
            }
            prev_vdd_t = power_t;
            prev_vdd_bb = bb2;

            drop_power_stub(facet,&gnd_form,&gnd,&layer,
            &power_conn,"gnd_connector",index,&(power_array[i].left_gnd),
            &(power_array[i].left_gnd_bb), &power_t,&bb2,0,LEFT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&gnd,&vertlayer,&prev_gnd_t,&power_t,&prev_gnd_bb,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&gnd_form,&path));
            }
            prev_gnd_t = power_t;
            prev_gnd_bb = bb2;
        }
        else {
            drop_power_stub(facet,&power_form,&vdd,&layer,
            &power_conn,"vdd_connector",index,&(power_array[i].left_vdd),
            &(power_array[i].left_vdd_bb), &power_t,&bb2,0,LEFT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&vdd,&vertlayer,&prev_vdd_t,&power_t,&prev_vdd_bb,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&power_form,&path));
            }
            prev_vdd_t = power_t;
            prev_vdd_bb = bb2;

            drop_power_stub(facet,&gnd_form,&gnd,&layer,
            &power_conn,"gnd_connector",index,&(power_array[i].right_gnd),
            &(power_array[i].right_gnd_bb), &power_t,&bb2,0,RIGHT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&gnd,&vertlayer,&prev_gnd_t,&power_t,&prev_gnd_bb,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&gnd_form,&path));
            }
            prev_gnd_t = power_t;
            prev_gnd_bb = bb2;

            index++;

            drop_power_stub(facet,&power_form1,&vdd,&layer,
            &power_conn,"vdd_connector",index,&(power_array[i].right_vdd),
            &(power_array[i].right_vdd_bb), &power_t,&bb2,1,RIGHT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&vdd,&vertlayer,&prev_vdd_t1,&power_t,&prev_vdd_bb1,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&power_form1,&path));
            }
            prev_vdd_t1 = power_t;
            prev_vdd_bb1 = bb2;

            drop_power_stub(facet,&gnd_form1,&gnd,&layer,
            &power_conn,"gnd_connector",index,&(power_array[i].left_gnd),
            &(power_array[i].left_gnd_bb), &power_t,&bb2,1,LEFT,&conn_bag);
            if(!first){
                drop_stub_path_not_attached(&gnd,&vertlayer,&prev_gnd_t1,&power_t,&prev_gnd_bb1,
                &bb2, BOTTOM,&path);
                OH_ASSERT( octAttach(&gnd_form1,&path));
            }
            prev_gnd_t1 = power_t;
            prev_gnd_bb1 = bb2;
        }
        first = 0;
    }

    /*
     * Now see if there was a connection from some signal terminals --- If
     * so then an extra stub needs to be dropped... drop it to the nearest
     * connector
     */
    if (power_impl){
	connect_data_and_supply(&power_form, &power_inst, &vdd, &vertlayer);
    }
    if (ground_impl){
	connect_data_and_supply(&gnd_form, &ground_inst, &gnd, &vertlayer);
    }
    return 0;
}

static int find_power_pins(instance, aterm1, aterm2, bb1, bb2, string,peak)
octObject *instance;
octObject *aterm1, *aterm2;
struct octBox *bb1, *bb2;
char *string;
double *peak;
{
    octObject master, fterm, real_prop, prop, temp;
    octGenerator gen;
    octStatus status;
    struct octBox tempbb;
    char *supply1=NULL, *supply2=NULL;
    int cnt;

    /* Generate over all terminals in the master */
    if (ohOpenMaster(&master, instance, "interface", "r") < OCT_OK) {
        octError("Unable to open interface facet for instance\n");
        exit(1);
    }

    cnt = 0;

    OH_ASSERT(octInitGenContents(&master, OCT_TERM_MASK, &gen));
    while((status = octGenerate(&gen, &fterm)) == OCT_OK) {

        /* Look for TERMTYPE property */
        status = ohGetByPropName(&fterm, &prop, "TERMTYPE");
        if (status != OCT_NOT_FOUND && status != OCT_OK) {
            OH_ASSERT(status);
        }
        if (status == OCT_OK) {
            if (strcmp(prop.contents.prop.value.string, string) == 0) {
                if (cnt == 0) {
                    supply1 = util_strsav(fterm.contents.term.name);
                    if (ohGetByPropName(&fterm,&real_prop,"PEAKCURRENT") == OCT_OK)
                        *peak += real_prop.contents.prop.value.real;
                } 
                else if (cnt == 1) {
                    supply2 = util_strsav(fterm.contents.term.name);
                    if (ohGetByPropName(&fterm,&real_prop,"PEAKCURRENT") == OCT_OK)
                        *peak += real_prop.contents.prop.value.real;
                }
                cnt++;
            }
        }
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

    if (cnt < 1) {
        (void) fprintf(stderr, "%s doesn't have any %s pins\n",
        ohFormatName(instance), string);
        return 1;
    } 
    else if (cnt == 1) {
        supply2 = util_strsav(supply1);
    } 
    else if (cnt > 2) {
        (void) fprintf(stderr, "%s has more than 2 %s pins\n",
        ohFormatName(instance), string);
        return 1;
    }

    /* Go back and get the actual terminals and their positions */
    OH_ASSERT(ohGetByTermName(instance, aterm1, supply1));
    OH_ASSERT(ohGetByTermName(instance, aterm2, supply2));
    OH_ASSERT(ohTerminalBB(aterm1, bb1));
    OH_ASSERT(ohTerminalBB(aterm2, bb2));

    /* Swap them to get the leftmost on the left */
    if (bb1->lowerLeft.x > bb2->lowerLeft.x) {
        temp = *aterm1;
        *aterm1 = *aterm2;
        *aterm2 = temp;
        tempbb = *bb1;
        *bb1 = *bb2;
        *bb2 = tempbb;
    }
    FREE(supply1);
    FREE(supply2);
    return 0;
}

/*
 *  This routine should:
 *    check layer for each terminal and use that layer
 *    check that the resulting path is manhattan
 */
static void drop_stub_path(net, layer, left, right, leftbb, rightbb,dir,path)
octObject *net;
octObject *layer;
octObject *left;
octObject *right;
octObject *path;
struct octBox *leftbb;
struct octBox *rightbb;
int dir;
{
    struct octPoint points[2];

    /* create a path and attach it to the net */
    path->type = OCT_PATH;
    if( dir == LEFT) 
        path->contents.path.width = ABS(leftbb->upperRight.y - leftbb->lowerLeft.y);
    else if ( dir == RIGHT)
        path->contents.path.width = ABS(rightbb->upperRight.y - rightbb->lowerLeft.y);
    else
        path->contents.path.width = ABS(leftbb->upperRight.x - leftbb->lowerLeft.x);
    OH_ASSERT(octCreate(net, path));

    /* Drop the path */
    points[0].x = (leftbb->upperRight.x + leftbb->lowerLeft.x) / 2;
    points[0].y = (leftbb->upperRight.y + leftbb->lowerLeft.y) / 2;
    points[1].x = (rightbb->upperRight.x + rightbb->lowerLeft.x) / 2;
    points[1].y = (rightbb->upperRight.y + rightbb->lowerLeft.y) / 2;
    OH_ASSERT(octPutPoints(path, 2, points));

    /* attach the terminals to the net */
    if (octAttachOnce(net,left) < OCT_OK) {
        octError("Attaching terminal to net ");
        return ;
    }
    if (octAttachOnce(net,right) < OCT_OK) {
        octError("Attaching terminal to net ");
        return ;
    }

    /* attach the endpoints to the path segment */
    OH_ASSERT(octAttach(path, left));
    OH_ASSERT(octAttach(path, right));

    /* Attach the path to a layer */
    OH_ASSERT(octAttach(layer, path));
}

/*
 *  This routine should:
 *    check layer for each terminal and use that layer
 *    check that the resulting path is manhattan
 *    Does not attach terminals to net since routing termianls
 */
static void drop_stub_path_not_attached(net, layer, left, right, leftbb, rightbb,dir,path)
octObject *net;
octObject *layer;
octObject *left;
octObject *right;
octObject *path;
struct octBox *leftbb;
struct octBox *rightbb;
int dir;
{
    struct octPoint points[2];

    /* create a path and attach it to the net */
    path->type = OCT_PATH;
    if( dir == LEFT) 
        path->contents.path.width = ABS(leftbb->upperRight.y - leftbb->lowerLeft.y);
    else if ( dir == RIGHT)
        path->contents.path.width = ABS(rightbb->upperRight.y - rightbb->lowerLeft.y);
    else
        path->contents.path.width = ABS(leftbb->upperRight.x - leftbb->lowerLeft.x);
    OH_ASSERT(octCreate(net, path));

    /* Drop the path */
    points[0].x = (leftbb->upperRight.x + leftbb->lowerLeft.x) / 2;
    points[0].y = (leftbb->upperRight.y + leftbb->lowerLeft.y) / 2;
    points[1].x = (rightbb->upperRight.x + rightbb->lowerLeft.x) / 2;
    points[1].y = (rightbb->upperRight.y + rightbb->lowerLeft.y) / 2;
    OH_ASSERT(octPutPoints(path, 2, points));

    /* attach the endpoints to the path segment */
    OH_ASSERT(octAttach(path, left));
    OH_ASSERT(octAttach(path, right));

    /* Attach the path to a layer */
    OH_ASSERT(octAttach(layer, path));
}

/*
 * Creates a connector instance of the connector and drops the stub to the
 * end terminal. Also drops a stub to the previous terminal and attaches 
 * the terminal of the connector and the vertical paths to the formal terminal
 */

static void drop_power_stub(facet,form_term,net,layer,inst_temp,name,
index, old_term, old_bb,new_term,new_bb,pos,side,bag)

octObject *facet,*net,*layer,*inst_temp;
octObject *form_term,*old_term,*new_term,*bag;
struct octBox *old_bb,*new_bb;
char *name;
int pos;    /* INNER=0  or OUTER=1 rail on a side */
int side;    
{
    octObject inst, path;
    char temp_name[20];

    inst.type = OCT_INSTANCE;
    path.type = OCT_PATH;
    inst.contents.instance.master = inst_temp->contents.instance.master;
    inst.contents.instance.view = inst_temp->contents.instance.view;
    inst.contents.instance.facet = "contents";
    inst.contents.instance.transform.transformType = inst_temp->contents.instance.transform.transformType;
    inst.contents.instance.version = OCT_CURRENT_VERSION;
    /* Place the connector instance at the point */
    (void)strcpy(temp_name,"");
    (void) sprintf(temp_name,"%s_%d",name,index);
    inst.contents.instance.name = temp_name;
    inst.contents.instance.transform.translation.y = 
        (old_bb->lowerLeft.y + old_bb->upperRight.y)/2;

    if (side == LEFT)
        inst.contents.instance.transform.translation.x = 
            (0 - pos * (power_width+vlayer_pitch) - power_width / 2) * units_per_lambda;
    else if ( side == RIGHT)
        inst.contents.instance.transform.translation.x =
            max_xposition + (pos *(power_width + vlayer_pitch) + power_width / 2) * units_per_lambda;

    OH_ASSERT( octCreate(facet,&inst) );
    OH_ASSERT( octAttach(bag,&inst) );
    OH_ASSERT( octBB(&inst,new_bb) );

    /* Make sure that the connector is aligned on a Lambda grid  (AC April 1991) */
    { 
	int offsetx = new_bb->lowerLeft.x % units_per_lambda;
	int offsety = new_bb->lowerLeft.y % units_per_lambda;
	if ( offsetx || offsety ) {
	    inst.contents.instance.transform.translation.x -= offsetx;
	    inst.contents.instance.transform.translation.y -= offsety;
	    if ( debug ) {
		printf( "Aligning connector to lambda grid (offset %d,%d)\n", offsetx,offsety );
	    }
	    OH_ASSERT( octModify( &inst ) );
	    OH_ASSERT(octBB(&inst,new_bb));	/* May 91.  */
	}
    }
    OH_ASSERT(octGenFirstContent(&inst,OCT_TERM_MASK,new_term));
    /* make a call to "drop_stub_path" to create the stub */

    if(side == LEFT)
        drop_stub_path_not_attached(net,layer,new_term,old_term,new_bb,
        old_bb, RIGHT,&path);
    else if(side == RIGHT)
        drop_stub_path_not_attached(net,layer,old_term,new_term,old_bb,
        new_bb, LEFT,&path);


    /* Attach the terminal of the power connector
        and the path to the formal terminal     */
    OH_ASSERT( octAttach(form_term,new_term));
}

/* find_power_net returns the net with a NETTYPE
 * propperty equal to "string". Failing this search
 * for a net attached to formal terminal with property
 * TERMTYPE  equal to "string". In case more than one 
 * net is found the first one is returned */

void
find_power_net(facet,net,string)
octObject *facet, *net;
char *string;
{
    int interface_open=1;
    octGenerator genNet, genTerm;
    octObject new_net,term,interface,interm,prop;
    octStatus status;

    if (ohOpenInterface(&interface,facet,"r") < OCT_OK){
        octError("Unable to open interface to determine TERMTYPE \n");
        (void) printf("Creating a net for %s \n",string);
	interface_open = 0;
    }

    OH_ASSERT(octInitGenContents(facet,OCT_NET_MASK,&genNet));
    while(octGenerate(&genNet,&new_net) == OCT_OK){
	/* Check for the property NETTYPE = "string" */
	status = ohGetByPropName(&new_net,&prop,"NETTYPE");
	if ((status == OCT_OK) && (prop.contents.prop.type == OCT_STRING)){
	    if(!strcmp(prop.contents.prop.value.string,string)){
		/* Found the net  ---- great */
		*net = new_net;
		return;
	    }
	}

	/* Now see if there is a formal terminal with TERMTYPE = "string" */
	if (interface_open){
	    OH_ASSERT(octInitGenContents(&new_net,OCT_TERM_MASK,&genTerm));
	    while( octGenerate(&genTerm,&term) == OCT_OK){
		if (octIdIsNull(term.contents.term.instanceId)) {
		    if (ohGetByTermName(&interface,
			    &interm, term.contents.term.name) == OCT_OK){
			if (ohGetByPropName(&interm,&prop,"TERMTYPE")==OCT_OK){
			    if(!strcmp(prop.contents.prop.value.string,string)){
				/* Might as well add the NETTYPE property */
				*net = new_net;
				(void)ohCreatePropStr(net, &prop, "NETTYPE", string);
				return;
			    }
			}
		    }
		}
	    }
	}
    }
    /* No net found create one -- Add the NETTYPE property */
    OH_ASSERT(ohGetOrCreateNet(facet,net,string));
    (void)ohCreatePropStr(net, &prop, "NETTYPE", string);
    return;
}

/*
 * Get a connector from tap --- 
 */
octStatus
get_connector(vlayer, hlayer, vwidth, hwidth, inst)
octObject vlayer, hlayer;
octCoord vwidth, hwidth;
octObject *inst;
{
    static tapLayerListElement conn_spec[2];

    conn_spec[0].layer = vlayer;
    conn_spec[0].width = vwidth;
    conn_spec[0].direction = TAP_VERTICAL;
    conn_spec[1].layer = hlayer;
    conn_spec[1].width = hwidth;
    conn_spec[1].direction = TAP_HORIZONTAL;
    return(tapGetConnector(2,conn_spec,inst));
}

#define VERY_LARGE 10000000
#define HALF_VERY_LARGE 5000000
/*
 * Connect the instance "dat_inst" to the nearest terminal
 * (in the vertical direction) of instances on net... One may
 * Have to rip up part of the net and insert tow paths if the
 * "dat_term" lies between two power connectors....
 */

void
connect_data_and_supply(fterm, data_term, net, layer)
octObject *fterm, *data_term, *net, *layer;
{
    octGenerator gen, gen1;
    octCoord xref, yref, diff;
    struct octBox dataBB, newBB;
    octObject path, segment, connector_segment;
    octCoord best_top_diff = VERY_LARGE;
    octCoord best_bot_diff = VERY_LARGE;
    octObject best_top_term, best_bot_term, data_inst, endpoint, term;

    path.type = OCT_PATH;
    OH_ASSERT(ohBB(data_term, &dataBB));
    xref = dataBB.lowerLeft.x;
    yref = dataBB.lowerLeft.y;

    OH_ASSERT(octInitGenContents(fterm, OCT_TERM_MASK, &gen));
    while (octGenerate(&gen, &term) == OCT_OK){
	if (octIdIsNull(term.contents.term.instanceId)) continue;
	OH_ASSERT(ohActualTerminalBB(&term, &newBB));
	if (newBB.lowerLeft.x == xref){
	    if (newBB.lowerLeft.y < yref){
		diff = yref - newBB.upperRight.y;
		if (diff < best_bot_diff){
		    best_bot_diff = diff;
		    best_bot_term = term;
		}
	    } else if (newBB.lowerLeft.y > yref){
		diff = newBB.lowerLeft.y - dataBB.upperRight.y;
		if (diff < best_top_diff){
		    best_top_diff = diff;
		    best_top_term = term;
		}
	    } else {
		continue;	/* It is the same instance */
	    }
	}
    }
    if (best_bot_diff < HALF_VERY_LARGE || best_top_diff < HALF_VERY_LARGE){
	/*
	 * There is a terminal on the power/gnd net to connect to 
	 */
	if (best_bot_diff > HALF_VERY_LARGE){
	    OH_ASSERT(ohActualTerminalBB(&best_top_term, &newBB));
	    drop_stub_path_not_attached(net, layer, data_term, &best_top_term,
		    &dataBB, &newBB, BOTTOM, &path);
	    OH_ASSERT(octAttach(fterm,&path));
	} else if (best_top_diff > HALF_VERY_LARGE){
	    OH_ASSERT(ohActualTerminalBB(&best_bot_term, &newBB));
	    drop_stub_path_not_attached(net, layer, &best_bot_term, data_term,
		    &newBB, &dataBB, BOTTOM, &path);
	    OH_ASSERT(octAttach(fterm,&path));
	} else { /* Connectors on top and bottom*/
	    /* Rip the existing path segments between top and bot */
	    OH_ASSERT(octInitGenContainers(&best_top_term,OCT_PATH_MASK, &gen));
	    while(octGenerate(&gen, &segment) == OCT_OK){
		OH_ASSERT(octInitGenContents(&segment, OCT_TERM_MASK, &gen1));
		while (octGenerate(&gen1, &endpoint) == OCT_OK){
		    if (octIdsEqual(endpoint.objectId, best_bot_term.objectId)) {
			/* The current segment is the one joining both terms */
			connector_segment = segment;
		    }
		}
	    }
	    OH_ASSERT(octDelete(&connector_segment));
	    /* Now add the two new stubs */
	    OH_ASSERT(ohActualTerminalBB(&best_top_term, &newBB));
	    drop_stub_path_not_attached(net, layer, data_term, &best_top_term,
		    &dataBB, &newBB, BOTTOM, &path);
	    OH_ASSERT(octAttach(fterm,&path));
	    OH_ASSERT(ohActualTerminalBB(&best_bot_term, &newBB));
	    drop_stub_path_not_attached(net, layer, &best_bot_term, data_term,
		    &newBB, &dataBB, BOTTOM, &path);
	    OH_ASSERT(octAttach(fterm,&path));
	}
    } else {
	(void)fprintf(stdout,"wolfe:INFO: User will have to connect terminal %s on instance %s to net %s\n", 
	    data_term->contents.term.name, data_inst.contents.instance.name,
	    net->contents.net.name);
    }
	return;
}
