#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "wolfe.h"

int write_block_file(prefix, facet)
char *prefix;
octObject *facet;
{
    struct octBox bb;
    octObject rowclass, bag, prop, instances, inst;
    octGenerator gen, gen1;
    octStatus status;
    char filename[1024];
    FILE *fp;
    int i, row, class, *map_row_to_class;
    double tot_x_width, min_y_width, max_y_width, x_width, y_width;

    /*
     *  Open the filename describing the blocks
     */
    (void) sprintf(filename, "%s.blk", prefix);
    if ((fp = fopen(filename, "w")) == NULL) {
	(void) fprintf(stderr, "Unable to open \"%s\"\n", filename);
	return 1;
    }


    /*
     *  Choose the number of rows based on Aspect Ratio (if not yet specified)
     */
    tot_x_width = 0.0;
    min_y_width = 1.0e20;
    max_y_width = -1;
    OH_FAIL(ohGetByBagName(facet, &instances, "INSTANCES"),
	"Failed looking for INSTANCES bag on", facet);
    OH_ASSERT(octInitGenContents(&instances, OCT_INSTANCE_MASK, &gen));
    while ((status = octGenerate(&gen, &inst)) == OCT_OK) {

 	OH_FAIL(octTB(&inst, &bb),
	    "Unable to find bounding box of cell", &inst);

	x_width = (double)(bb.upperRight.x - bb.lowerLeft.x) / units_per_lambda;
	y_width = (double)(bb.upperRight.y - bb.lowerLeft.y) / units_per_lambda;

	tot_x_width += x_width;
	if (y_width > max_y_width) {
	    max_y_width = y_width;
	} else if (y_width < min_y_width) {
	    min_y_width = y_width;
	}
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);

    if (tot_x_width <= 0.0) {
	(void) fprintf(stderr, "No instances found in INSTANCES bag\n");
	return 1;
    }
    /* XXX ARGH */
    if (num_rows < 1) {
	double temp = tot_x_width / (TargetAspectRatio * rowSep * max_y_width);
	num_rows = (int) (sqrt(temp) + 0.5);
	if (num_rows <= 0) num_rows = 1;
    }


    /*
     *  Check for a row-class definition bag
     *
     *  (FACET foo:bar:contents
     *	    (BAG WOLFE-CLASS-BAG
     *	        (BAG WOLFE-CLASS
     *		    (IPROP WOLFE-ROW 1)
     *		    (IPROP WOLFE-ROW 2)
     *		    (IPROP WOLFE-ROW 5)
     *		    (INSTANCE inst1)
     *		    (INSTANCE inst2)
     *	        )
     *	        (BAG WOLFE-CLASS
     *		    (IPROP WOLFE-ROW 3)
     *		    (IPROP WOLFE-ROW 4)
     *		    (INSTANCE inst3)
     *		    (INSTANCE inst4)
     *	        )
     *	    )
     *  )
     *
     *  Things work best if either all instances (or no instances)
     *  are placed in class bags.
     */

    class = 1;
    map_row_to_class = ALLOC(int, num_rows+2);
    for(i = 1; i <= num_rows; i++) {
	map_row_to_class[i] = 0;
    }

    status = ohGetByBagName(facet, &rowclass, "WOLFE-CLASS-BAG");
    if (status == OCT_NOT_FOUND) {
	goto skip_class;
    }
    OH_ASSERT(status);

    OH_ASSERT(octInitGenContents(&rowclass, OCT_BAG_MASK, &gen));
    while ((status = octGenerate(&gen, &bag)) == OCT_OK) {
	OH_ASSERT(octInitGenContents(&bag, OCT_PROP_MASK, &gen1));
	while ((status = octGenerate(&gen1, &prop)) == OCT_OK) {
	    if (prop.contents.prop.type == OCT_INTEGER) {
		if (strcmp(prop.contents.prop.name, "WOLFE-ROW") == 0) {
		    row = prop.contents.prop.value.integer;
		    if (row <= num_rows) {
			map_row_to_class[row] = class;
		    }
		}
	    }
	}
	OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);
	class++;
    }
    OH_ASSERT(status == OCT_GEN_DONE ? OCT_OK : status);


    /*
     *  Write out the row information
     */
skip_class:
    for(i = 1; i <= num_rows; i++) {
	(void) fprintf(fp, "block height %d class %d\n", 
	    (int) max_y_width, map_row_to_class[i]);
    }

    if (debug > 1) {
	(void) printf("rows=%d, tot_width=%2.0f, max_height=%2.0f, classes=%d\n",
	    num_rows, tot_x_width, max_y_width, class);
    }

    FREE(map_row_to_class);

    /* Return true if any errors writing the file */
    return ferror(fp) || (fclose(fp) == EOF);
}
