/* @(#)write_param.c	1.5 wolfe (ITD) 3/21/90   */
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "vov.h"
#include "wolfe.h"

static void check_for_complex_feeds();
static int write_complex_feeds (FILE *);
static octStatus GetTerminalBB(octObject *, struct octBox *); 

/* XXX this is REALY BAD */
static double check_prop(facet, name, defvalue)
octObject *facet;
char *name;
double defvalue;
{
    octObject prop;
    octStatus status;

    status = ohGetByPropName(facet, &prop, name);
    if (status == OCT_OK) {
	if (prop.contents.prop.type == OCT_REAL) {
		return prop.contents.prop.value.real;
	}
	if (prop.contents.prop.type == OCT_INTEGER) {
		return (double) prop.contents.prop.value.integer;
	}
    } else if (status != OCT_NOT_FOUND) {
	OH_ASSERT(status);
    }
	return defvalue;
}
    

static void check_position(facet, name, value)
octObject *facet;
char *name;
int *value;
{
    octObject prop;
    octStatus status;

    status = ohGetByPropName(facet, &prop, name);
    if (status == OCT_OK) {
	if(prop.contents.prop.type == OCT_STRING) {
		if ( strcmp(prop.contents.prop.value.string,"LEFT") == 0){
			*value = LEFT;
		}
		else if ( strcmp(prop.contents.prop.value.string,"RIGHT") == 0){
			*value = RIGHT;
		}
		else if ( strcmp(prop.contents.prop.value.string,"BOTH") == 0){
			*value = BOTH;
		}
		else{
		(void) printf("Unknown side for Power Terminal - Using LEFT \n");
		}
	}
    } else if (status != OCT_NOT_FOUND) {
	OH_ASSERT(status);
    }
}

/* Read the parameters from the contents facet */

int read_parameters(facet)
octObject *facet;
{
    octObject interface, bag, box;
    struct octBox *bb;

#ifdef TW3_3
    att_per_cell = check_prop(facet, "WOLFE-att.per.cell", 
		(double) att_per_cell);
    indent = check_prop(facet, "WOLFE-indent",
		(double) indent);
    do_global_route_cell_swaps = check_prop(facet, 
		"WOLFE-do.global.route.cell.swaps", 
		(double) do_global_route_cell_swaps);
#endif
    fast_mode = check_prop(facet, "WOLFE-fast", (double) fast_mode);
    rowSep = check_prop(facet, "WOLFE-rowSep", (double) rowSep);
#ifndef TW5_0
    implicit_feed_thru_range = check_prop(facet, 
		"WOLFE-implicit.feed.thru.range", (double) implicit_feed_thru_range);
#endif
    do_global_route = check_prop(facet, "WOLFE-do.global.route", 
		(double) do_global_route);
    Add_implicit = check_prop(facet, "WOLFE-add_implicit_feedthrus",
		(double) Add_implicit);
    h_net_weight = check_prop(facet, "WOLFE-h_net_weight", 
		(double) h_net_weight);
    v_net_weight = check_prop(facet, "WOLFE-v_net_weight",
		(double) v_net_weight);
    num_rows = check_prop(facet, "WOLFE-num_rows", (double) num_rows);
    power_width = check_prop(facet, "WOLFE-power_width", (double) power_width);
    check_position(facet, "WOLFE-power_position", &power_position);

    /* Fish around for an aspect ratio in the interface */
    if (ohOpenInterface(&interface, facet, "a") < OCT_OK) {
	(void) fprintf(stderr, "Unable to open interface facet of %s\n",
	    ohFormatName(&interface));
	octError(NIL(char));
	exit(1);
    }
    if (ohGetByBagName(&interface, &bag, "FLOORPLAN") == OCT_OK) {
	if (octGenFirstContent(&bag, OCT_BOX_MASK, &box) == OCT_OK) {
	    bb = &box.contents.box;
	    TargetAspectRatio = 
		(double) (bb->upperRight.x - bb->lowerLeft.x) /
		(double) (bb->upperRight.y - bb->lowerLeft.y);
	}
    }
	return 1;
}

/* ARGSUSED */
int write_param_file(prefix, facet)
char *prefix;
octObject *facet;		/* May need this eventually */
{
    char filename[1024];
    int retval;
    FILE *fp;

    (void) sprintf(filename, "%s.par", prefix);
    if ((fp = fopen(filename, "w")) == NULL) {
	(void) fprintf(stderr, "Unable to open \"%s\"\n", filename);
	return 1;
    }

#ifdef TW3_3
    (void) fprintf(fp, "att.per.cell %d\n", att_per_cell);
    (void) fprintf(fp, "indent %4.2lf\n", indent);
#endif
#ifdef TW4_2
#ifndef TW5_0
    if (!fallback){
	if (fast_mode) {
	    (void) fprintf(fp, "fast\n");
	} 
    } else{
	if ( fast_mode){
	(void) fprintf(fp, "att.per.cell 1\n");
	} else {
	(void) fprintf(fp, "att.per.cell 0\n");
	}
	(void) fprintf(fp, "indent 0.96\n");
    }
#endif
#endif
#ifdef TW5_0
    if (fast_mode) {
	(void) fprintf(fp, "fast\n");
    } 
#endif
    (void) fprintf(fp, "rowSep %4.2f\n", rowSep);
    if (addFeeds) {
    (void) fprintf(fp, "addFeeds\n");
    (void) fprintf(fp, "feedThruWidth %d\n", feedThruWidth);
#ifndef TW5_0
    (void) fprintf(fp, "implicit.feed.thru.range %4.2f\n", 
	implicit_feed_thru_range);
    check_for_complex_feeds();
    if (complex_feeds)
    {
      retval = write_complex_feeds(fp);
	if (retval != 0) return(retval);
    }
#endif
}
if (do_global_route) {
    (void) fprintf(fp, "do.global.route\n");
#ifdef TW3_3
    (void) fprintf(fp, "do.global.route.cell.swaps %4.2lf\n", 
	do_global_route_cell_swaps);
#endif
#ifdef TW4_2
#ifndef TW5_0
    if (!fallback) {
    (void) fprintf(fp, "do.global.route.cell.swaps %s\n",
	fast_mode ? "" : "full");
    } else {
	if ( fast_mode){
	    (void) fprintf(fp, "do.global.route.cell.swaps 0.0\n");
	} else{
	    (void) fprintf(fp, "do.global.route.cell.swaps 1\n");
	}
    }
#endif
#endif
    }

    /* Return true if any errors writing the file */
    return ferror(fp) || (fclose(fp) == EOF);
}

/* checks if this a complex feedthru cell */
static void
check_for_complex_feeds ()
{
  octObject feed, prop;
  int status;


  feed.type = OCT_FACET;
  feed.contents.facet.cell = feedthru_cell;		/* default cell */
  feed.contents.facet.view = feedthru_view;	/* default view */
  feed.contents.facet.facet = "contents";	/* default facet */
  feed.contents.facet.version = OCT_CURRENT_VERSION; /* default version */
  feed.contents.facet.mode = "r";
  status = octOpenFacet(&feed);

  VOVinputFacet( &feed );

  prop.type = OCT_PROP;
  prop.contents.prop.name = "COMPLEX_FEED";

  if (octGetByName (&feed, &prop) == OCT_OK)
    complex_feeds = 1;
  else
    complex_feeds = 0;
  octCloseFacet(&feed);
}




/* have to write the pins of the feedthru cell to parameters file */

static int
write_complex_feeds (fp)
FILE *fp;
{

  octObject feed, joinedterms, joinbag, aterm;
  octGenerator baggen, termgen;
  struct octBox cell_bb, pin_bb;
  int numterm, status, bottom, top;
  int x_center, y_center, pin_width, xc, yc;

  fprintf(fp, "multiple.feed.info\n");

  feed.type = OCT_FACET;
  feed.contents.facet.cell = feedthru_cell;		/* default cell */
  feed.contents.facet.view = feedthru_view;	/* default view */
  feed.contents.facet.facet = "interface";	/* default facet */
  feed.contents.facet.version = OCT_CURRENT_VERSION; /* default version */
  feed.contents.facet.mode = "r";
  status = octOpenFacet(&feed);
  
  VOVinputFacet( &feed );

  status = octBB (&feed, &cell_bb);   /*  Get the bounding box for this cell  */
  if (status != OCT_OK) { 
    (void) fprintf(stderr, "Failed on %s\n", ohFormatName(&feed));
    octError("need bounding box for inst\n");
    return 1;
  }
  x_center = (cell_bb.upperRight.x + cell_bb.lowerLeft.x) / 2;
  y_center = (cell_bb.upperRight.y + cell_bb.lowerLeft.y) / 2;
  bottom = (cell_bb.lowerLeft.y - y_center)/units_per_lambda;
  top = (cell_bb.upperRight.y - cell_bb.lowerLeft.y)/units_per_lambda + bottom;


  
  OH_FAIL(ohGetByBagName(&feed, &joinedterms, "JOINEDTERMS"),
    "Can't find JOINEDTERMS bag connected to", &feed);
  OH_ASSERT(octInitGenContents(&joinedterms, OCT_BAG_MASK, &baggen));
  while ((status = octGenerate(&baggen, &joinbag)) == OCT_OK) {
    /* now lets look for terminals attached to this bag */
    OH_ASSERT(octInitGenContents(&joinbag, OCT_TERM_MASK, &termgen));
    numterm = 0;
    while ((status = octGenerate(&termgen, &aterm)) == OCT_OK) {
      /* got a terminal */
      status = GetTerminalBB(&aterm, &pin_bb);
      if (status <  OCT_OK) {
        (void) fprintf(stderr, "Failed finding bounding box for %s\n",
		       ohFormatName(&aterm));
        OH_ASSERT(status);
      }
      /* output the pin information */
      pin_width = pin_bb.upperRight.x - pin_bb.lowerLeft.x;
      xc = (pin_bb.lowerLeft.x - x_center + pin_width/2)/units_per_lambda;
      yc = (pin_bb.lowerLeft.y > y_center) ? top : bottom;
      if (numterm == 0) 
	fprintf(fp, "pin name %s %d %d\n", aterm.contents.term.name,
		xc, yc);
      else
	fprintf(fp, "equiv name %s %d %d\n", aterm.contents.term.name,
		xc, yc);
      numterm++;
    }
    if (numterm != 2) {
      (void) fprintf(stderr,  "There must be exactly two terminals in each join bag\n");
      (void) fprintf(stderr,  "of the feedthru interface facet.\n");
       return 1;
    }
  }
  fprintf(fp, "multiple.feed.info.stop\n");
  return(0);
}







octStatus GetTerminalBB(fterm, bb)
octObject *fterm;
struct octBox *bb;
{
    octObject box;
    octStatus status;
    octCoord t;

    /* Get the first 'terminal frame' (which must be a box) */
    status = octGenFirstContent(fterm, OCT_BOX_MASK, &box);
    if (status != OCT_OK) return status;
    *bb = box.contents.box;
    if (bb->lowerLeft.x > bb->upperRight.x) {
	t = bb->lowerLeft.x;
	bb->lowerLeft.x = bb->upperRight.x;
	bb->upperRight.x = t;
    }
    if (bb->lowerLeft.y > bb->upperRight.y) {
	t = bb->lowerLeft.y;
	bb->lowerLeft.y = bb->upperRight.y;
	bb->upperRight.y = t;
    }
    return OCT_OK;
}

