#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLCase - case
 *
 *    (case key
 *     (keylist forms ... ) ... )
 *
 *    return result
 */
LLobject *
LLCase(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLCase_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLCase_tr - case
 *
 *    (case key
 *     (keylist forms ... ) ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLCase_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *sexps, *condition, *key, *keylist, *subkey;
	LLstring t, otherwise;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: (case key (keylist forms ... ) ... )\n\
\tnot enough arguments to case\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	t = LLInternSymbol("t");
	otherwise = LLInternSymbol("otherwise");

	sexps = *arguments;

	key = LLEvaluator(LL_first(sexps), *environment);

	/* cdr over the key */
	for (sexps=LL_cdr(sexps); LL_consp(sexps); sexps=LL_cdr(sexps)) {

	    condition = LL_car(sexps);
	    if ( ! LL_consp(condition) ) {
		LLErrorF("\
Form: (case key (keylist forms ... ) ... )\n\
\tbadly formed condition\n\
\toffender: ~S\n\
\tform:     ~S\n\
", condition, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    keylist = LL_car(condition);

	    /*
	     *    DON'T TODO - check to insure that the keylists are disjoint.
	     *    (fuck 'em - its just too slow to do such a thing).
	     */
	    if ( ! LL_consp(keylist) && ! LL_null(keylist) ) {
		/*
		 *    key
		 */
		if ((LL_symbolp(keylist) &&
				(LL_pname(keylist) == t ||
				 LL_pname(keylist) == otherwise)) ||
			 LLEql(keylist, key)) {
		    *arguments = LL_rest(condition);
		    if (LL_null(*arguments)) {
			*arguments = LL_nil;
			return (LL_nil);
		    }

		    return (LLProgn_tr(arguments, environment));
		}
	    } else {
		/*
		 *    (key key key key)
		 */
		while ( ! LL_null(keylist) ) {
		    subkey = LL_car(keylist);
		    if (LLEql(subkey, key)) {
			*arguments = LL_rest(condition);
			if (LL_null(*arguments))
			    return (LL_nil);

			return (LLProgn_tr(arguments, environment));
		    }
		    keylist = LL_cdr(keylist);
		}
	    }
	}

	*arguments = LL_nil;
	return (LL_nil);
}
