#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLCatch - catch
 *
 *    (catch tag
 *           forms ... )
 *
 *    return the result
 */
LLobject *
LLCatch(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* valuet */
{
	if (LL_null(LLCatch_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLCatch_tr - catch
 *
 *    (catch tag
 *           forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLCatch_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *args, *tag;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: (catch tag forms ... )\n\
\tinvalid number of arguments to catch\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	args = *arguments;

	tag = LLEvaluator(LL_first(args), *environment);

	/* arguments is not valid from here ... (LLCatchValue sets it to nil) */
	if (LLCatchValue(tag, LL_nil, LL_environment, arguments))
	    return (LL_nil);

	/*
	 *    evaluate it ALL inside the catcher's scope
	 */
	*arguments = LLProgn(LL_rest(args), *environment);
	/* ... till here */

	LLEndCatch();

	return (LL_nil);
}
