#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLCond - cond
 *
 *    (cond 
 *     (test forms ... ) ... )
 *
 *    return result
 */
LLobject *
LLCond(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLCond_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLCond_tr - cond
 *
 *    (cond 
 *     (test forms ... ) ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLCond_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *sexps, *condition, *ret;

	sexps = *arguments;

	while ( ! LL_null(sexps) ) {
	    condition = LL_car(sexps);

	    if ( ! LL_consp(condition) ) {
		LLErrorF("\
Form: (cond (test forms ... ) ... )\n\
\tbadly formed condition\n\
\tform: ~S\n\
", *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    ret = LLEvaluator(LL_car(condition), *environment);
	    if ( ! LL_null(ret) ) {
		*arguments = LL_rest(condition);
		if (LL_null(*arguments)) {
		    *arguments = ret;
		    return (LL_nil);
		}

		return (LLProgn_tr(arguments, environment));
	    }

	    sexps = LL_cdr(sexps);
	}

	*arguments = LL_nil;
	return (LL_nil);
}
