#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *defun_or_defmacro();

/*
 *    LLDefun - defun
 *
 *    (defun name lambda-list
 *       [ declaration ]
 *       forms ... )
 *
 *    return result
 */
LLobject *
LLDefun(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDefun_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLDefun_tr - defun
 *
 *    (defun name lambda-list
 *       [ declaration ]
 *       forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDefun_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (defun_or_defmacro(arguments, environment, /* defmacro */ FALSE));
}

/*
 *    LLDefmacro - demacro
 *
 *    (defmacro name lambda-list
 *       [ declaration ]
 *       forms ... )
 *
 *    return result
 */
LLobject *
LLDefmacro(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDefmacro_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLDefmacro_tr - defmacro
 *
 *    (defmacro name lambda-list
 *       [ declaration ]
 *       forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDefmacro_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (defun_or_defmacro(arguments, environment, /* defmacro */ TRUE));
}

/*
 *    WorkHorse ----------------------------------------
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
static LLobject *
defun_or_defmacro(arguments, environment, defmacroP)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
Boolean defmacroP;
{
	LLobject *name, *value, *old;

	if (LLGetLength(*arguments) < 2) {
	    LLErrorF("\
Form: (~A name lambda-list [ declaration ] forms ... )\n\
\tinvalid number of arguments to ~A\n\
\tform: ~S\n\
", defmacroP ? "defmacro": "defun", defmacroP ? "defmacro": "defun", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	name = LL_first(*arguments);

	if ( ! LL_symbolp(name) ) {
	    LLErrorF("\
Form: (~A name lambda-list [ declaration ] forms ... )\n\
\tname must be a symbol\n\
\tform: ~S\n\
", defmacroP ? "defmacro": "defun", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LLFboundp(name)) {
	    old = LLGetValue(name, LL_functionSlot, LL_environment);
	    if (LL_special_formp(old)) {
		LLErrorF("\
Form: (~A name lambda-list [ declaration ] forms ... )\n\
\tcannot use ~A to redefine a special form\n\
\tspecial form: ~S\n\
\tform:       ~S\n\
", defmacroP ? "defmacro": "defun", defmacroP ? "defmacro": "defun", old, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	}

	name = LL_first(*arguments);
	value = defmacroP ? 
		LLMacroClosure(LL_rest(*arguments), *environment):
		LLFunctionClosure(LL_rest(*arguments), *environment);
	LLSetValue(name, LL_functionSlot, value, LL_environment);

	*arguments = name;
	return (LL_nil);
}
