#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

typedef enum define_op define_op;
enum define_op {
    defvar,
    defparameter,
    defconstant,
};
static LLobject *define();

/*
 *    LLDefvar - defvar
 *
 *    (defvar var [ form [ documentation ] ] )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return result
 */
LLobject *
LLDefvar(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDefvar_tr(&arguments, &environment);

	return (arguments);
}

/*
 *    LLDefvar_tr - defvar
 *
 *    (defvar var [ form [ documentation ] ] )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDefvar_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (define(arguments, environment, "defvar", defvar));
}

/*
 *    LLDefparameter - defparameter
 *
 *    (defparameter var form [ documentation ] )
 *
 *    return result
 */
LLobject *
LLDefparameter(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDefparameter_tr(&arguments, &environment);

	return (arguments);
}

/*
 *    LLDefparameter_tr - defparameter
 *
 *    (defparameter var form [ documentation ] )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDefparameter_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (define(arguments, environment, "defparameter", defparameter));
}


/*
 *    LLDefconstant - defconstant
 *
 *    (defconstant var form [ documentation ] )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return result
 */
LLobject *
LLDefconstant(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDefconstant_tr(&arguments, &environment);

	return (arguments);
}

/*
 *    LLDefconstant_tr - defconstant
 *
 *    (defconstant var form [ documentation ] )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDefconstant_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (define(arguments, environment, "defconstant", defconstant));
}

/*
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
static LLobject *
define(arguments, environment, name, op)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
String name;
define_op op;
{
	LLobject *sexps, *symbol, *value, *documentation;
	int len;

	if ((len = LLGetLength(*arguments)) < (op == defvar ? 1: 2) && 3 < len) {
	    LLErrorF("\
Form: (~A var form [ documentation ] )\n\
\tnot enough arguments to ~A\n\
\targuments: ~S\n\
", name, name, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	sexps = *arguments;

	symbol = LL_first(sexps);
	value = LL_second(sexps);
	documentation = LL_third(sexps);

	if ( ! LL_symbolp(symbol) ) {
	    LLErrorF("\
Form: (~A var form [ documentation ] )\n\
\tvar is not a symbol\n\
\tvar:   ~S\n\
\tvalue: ~S\n\
\tform:  ~S\n\
", name, symbol, value, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if ( ! LL_null(documentation) && ! LL_stringp(documentation) ) {
	    LLErrorF("\
Form: (~A var form [ documentation ] )\n\
\tdocumentation must be a string if it is provided\n\
\tvar:            ~S\n\
\tvalue:          ~S\n\
\tdocumentation:  ~S\n\
\tform:           ~S\n\
", name, symbol, value, documentation, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	switch (op) {
	case defconstant:
	    if ( ! LL_null(LLBoundp(symbol)) ) {
		LLErrorF("\
Form: (~A var form [ documentation ] )\n\
\tcannot redefine a constant value\n\
\tvar:            ~S\n\
\tvalue:          ~S\n\
\tdocumentation:  ~S\n\
\tform:           ~S\n\
", name, symbol, value, documentation, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    LLSetValue(symbol, LL_valueSlot, 
		    LLEvaluator(value, *environment), LL_environment);
	    break;
	case defparameter:
	    LLSetValue(symbol, LL_valueSlot, 
		    LLEvaluator(value, *environment), LL_environment);
	    break;
	case defvar:
	    if (LL_null(LLBoundp(symbol))) {
		LLSetValue(symbol, LL_valueSlot, 
			LLEvaluator(value, *environment), LL_environment);
	    }
	    break;
	}

	*arguments = symbol;
	return (LL_nil);
}
