#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *do_or_dostar();

/*
 *    LLDo - do
 *
 *    (do ( (var [ init [ step ] ]) ... )
 *        (end-test . result)
 *        forms ... )
 *
 *    return the result
 */
LLobject *
LLDo(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDo_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLDo_tr - do
 *
 *    (do ( (var [ init [ step ] ]) ... )
 *        (end-test . result)
 *        forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDo_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (do_or_dostar(arguments, environment, /* starP */ FALSE));
}

/*
 *    LLDostar - do*
 *
 *    (do* ( (var [ init [ step ] ]) ... )
 *         (end-test . result)
 *         forms ... )
 *
 *    return the result
 */
LLobject *
LLDostar(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDostar_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLDostar_tr - do*
 *
 *    (do* ( (var [ init [ step ] ]) ... )
 *         (end-test . result)
 *         forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDostar_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (do_or_dostar(arguments, environment, /* starP */ TRUE));
}

/*
 *    WorkHorse ------------------------
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
static LLobject *
do_or_dostar(arguments, environment, starP)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
Boolean starP;
{
	LLobject *initializer, *ender, *i, *init, *end_test, *result,
		*var, *initform, *stepform, *value, *ret, *body,
		*doenv, **evalenv;
	int len;

	if (LLGetLength(*arguments) < 2) {
	    LLErrorF("\
Form: (do~A ( (var [ init [ step ] ]) ... ) (end-test . result) forms ... )\n\
\tillegal number of arguments to do\n\
\tform: ~S\n\
", starP ? "*": "", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	initializer = LL_first(*arguments);
	if ( ! LL_listp(initializer) ) {
	    LLErrorF("\
Form: (do~A ( (var [ init [ step ] ]) ... ) (end-test . result) forms ... )\n\
\tbadly formed initializer\n\
\tinitializer: ~S\n\
\tform:        ~S\n\
", starP ? "*": "", initializer, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	ender = LL_second(*arguments);
	if ( ! LL_consp(ender)) {
	    LLErrorF("\
Form: (do~A ( (var [ init [ step ] ]) ... ) (end-test . result) forms ... )\n\
\tbadly formed ender\n\
\tender: ~S\n\
\tform:  ~S\n\
", starP ? "*": "", ender, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	body = LLGetNthcdr(2, *arguments);
	end_test = LL_car(ender);
	result = LL_cdr(ender);

	doenv = *environment;
	evalenv = starP ? &doenv: environment;

	for (i=initializer; ! LL_null(i); i=LL_cdr(i)) {
	    init = LL_car(i);

	    if ( ! LL_consp(init) || LLGetLength(init) > 3) {
		LLErrorF("\
Form: (do~A ( (var [ init [ step ] ]) ... ) (end-test . result) forms ... )\n\
\tbadly formed initializer member\n\
\tmember: ~S\n\
\tform:   ~S\n\
", starP ? "*": "", init, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    var = LL_first(init);
	    initform = LL_second(init);	/* may be nil (the default) */

	    if ( ! LL_symbolp(var) ) {
		LLErrorF("\
Form: (do~A ( (var [ init [ step ] ]) ... ) (end-test . result) forms ... )\n\
\tvariable is not a symbol\n\
\tsymbol:    ~S\n\
\tform:      ~S\n\
", starP ? "*": "", var, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    value = LLEvaluator(initform, *evalenv);
	    doenv = LLExtendEnvironment(var, LL_valueSlot, value, doenv);
	}

	do {
	    ret = LLEvaluator(end_test, doenv);
	    if ( ! LL_null(ret) )
		break;
	    LLProgn(body, doenv);

	    if (starP) {
		/*
		 *    If do*, set the value directly
		 *    so that each successive evaluation
		 *    can have access to the new value
		 *    and successive old values
		 */
		for (i=initializer; ! LL_null(i); i=LL_cdr(i)) {
		    init = LL_car(i);
		    var = LL_first(init);

		    if (LLGetLength(init) < 3) {
			/* do nothing */
		    } else {
			stepform = LL_third(init);
			value = LLEvaluator(stepform, doenv);
			LLSetValue(var, LL_valueSlot, value, doenv);
		    }
		}
	    } else {
		LLobject *next, *last, *tmp;

		/*
		 *    If do, save the values until done
		 *    so that each successive evaluation
		 *    can have access to all the old values
		 *    until the end.
		 *
		 *    don't (setq next (cons value next))
		 *    because the list will be backwards then
		 */
		for (i=initializer, next=LL_nil, last=LL_nil; 
			! LL_null(i); i=LL_cdr(i)) {
		    init = LL_car(i);
		    var = LL_first(init);

		    if (LLGetLength(init) < 3) {
			value = LLGetValue(var, LL_valueSlot, doenv);
		    } else {
			stepform = LL_third(init);
			value = LLEvaluator(stepform, doenv);
		    }

		    tmp = LLCons(value, LL_nil);
		    if (LL_null(next)) {
			next = tmp;
		    } else {
			LL_rplacd(last, tmp);
		    }
		    last = tmp;
		}
		for (i=initializer; ! LL_null(i);
			next=LL_cdr(next), i=LL_cdr(i)) {
		    init = LL_car(i);
		    var = LL_first(init);
		    value = LL_car(next);

		    LLSetValue(var, LL_valueSlot, value, doenv);
		}
	    }
	} while (1);

	if (LL_null(result)) {
	    *arguments = LL_nil;
	} else {
	    *arguments = LLProgn(result, doenv);
	}
	return (LL_nil);
}
