#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLDolist - dolist
 *
 *    (dolist (var listform [ resultform ] )
 *      forms ... )
 *
 *    return result
 */
LLobject *
LLDolist(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDolist_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLDolist_tr - dolist
 *
 *    (dolist (var listform [ resultform ] )
 *      forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDolist_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *init, *var, *list, *subenv, *result, *body;
	int len;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: (dolist (var listform [ resultform ] ) forms ... )\n\
\tinvalid number of arguments to dolist\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	init = LL_first(*arguments);
	body = LL_rest(*arguments);

	if ( ! LL_consp(init) || (((len = LLGetLength(init)) != 2) && len != 3) ) {
	    LLErrorF("\
Form: (dolist (var listform [ resultform ] ) forms ... )\n\
\tbadly formed initialization\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	var = LL_first(init);
	if ( ! LL_symbolp(var) ) {
	    LLErrorF("\
Form: (dolist (var listform [ resultform ] ) forms ... )\n\
\tvar is not a symbol\n\
\tform: ~S\n\
\tvar:  ~S\n\
", *arguments, var);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	list = LLEvaluator(LL_second(init), *environment);
	if ( ! LL_listp(list) ) {
	    LLErrorF("\
Form: (dolist (var listform [ resultform ] ) forms ... )\n\
\tlistform does not evaluate to a cons\n\
\tform:     ~S\n\
\tlistform: ~S\n\
\tresult:   ~S\n\
", *arguments, LL_second(init), list);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	subenv = LLExtendEnvironment(var, LL_valueSlot, LL_nil, *environment);
	while ( ! LL_null(list) ) {
	    LLSetValue(var, LL_valueSlot, LL_car(list), subenv);
	    LLProgn(body, subenv);
	    list = LL_cdr(list);
	}

	result = LL_third(init);
	if (LL_null(result)) {
	    *arguments = LL_nil;
	} else {
	    LLSetValue(var, LL_valueSlot, LL_nil, subenv);
	    *arguments = LLEvaluator(result, subenv);
	}

	return (LL_nil);
}
