#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLDotimes - dotimes
 *
 *    (dotimes (var count [ resultform ] )
 *      forms ... )
 *
 *    return result
 */
LLobject *
LLDotimes(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLDotimes_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLDotimes_tr - dotimes
 *
 *    (dotimes (var count [ resultform ] )
 *      forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLDotimes_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *init, *var, *times, *subenv, *result, *body;
	int len, i, ntimes;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: (dotimes (var count [ resultform ] ) forms ... )\n\
\tinvalid number of arguments to dotimes\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	init = LL_first(*arguments);
	body = LL_rest(*arguments);

	if ( ! LL_consp(init) || (((len = LLGetLength(init)) != 2) && len != 3) ) {
	    LLErrorF("\
Form: (dotimes (var count [ resultform ] ) forms ... )\n\
\tbadly formed initialization\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	var = LL_first(init);
	if ( ! LL_symbolp(var)) {
	    LLErrorF("\
Form: (dotimes (var count [ resultform ] ) forms ... )\n\
\tvar is not a symbol\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	times = LLEvaluator(LL_second(init), *environment);
	if ( ! LL_integerp(times)) {
	    LLErrorF("\
Form: (dotimes (var count [ resultform ] ) forms ... )\n\
\tcount does not evaluate to an integer\n\
\tform:  ~S\n\
\tcount: ~S\n\
", *arguments, times);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	subenv = LLExtendEnvironment(var, LL_valueSlot, LL_nil, *environment);

	ntimes = times->value.a_integer;
	for (i=0; i<ntimes; i++) {
	    LLSetValue(var, LL_valueSlot, LLMakeInteger(i), subenv);
	    LLProgn(body, subenv);
	}

	result = LL_third(init);
	if (LL_null(result)) {
	    *arguments = LL_nil;
	} else {
	    LLSetValue(var, LL_valueSlot, LLMakeInteger(i), subenv);
	    *arguments = LLEvaluator(result, subenv);
	}

	return (LL_nil);
}
