#include <sys/types.h>
#include <setjmp.h>
#include <math.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static double trunc();
static double round();
static LLobject *func();

/*
 *    LLFfloor - ffloor
 *
 *    (ffloor number &optional divisor)
 *
 *    return the floating-point floor
 */
LLobject *
LLFfloor(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(floor, "ffloor", number, divisor, /* floatP */ TRUE));
}

/*
 *    LLFceiling - fceiling
 *
 *    (fceiling number &optional divisor)
 *
 *    return the floating-point ceiling
 */
LLobject *
LLFceiling(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(ceil, "fceiling", number, divisor, /* floatP */ TRUE));
}

/*
 *    LLFtruncate - ftruncate
 *
 *    (ftruncate number &optional divisor)
 *
 *    return the floating-point truncation
 */
LLobject *
LLFtruncate(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(trunc, "ftruncate", number, divisor, /* floatP */ TRUE));
}

/*
 *    LLFround - fround
 *
 *    (fround number &optional divisor)
 *
 *    return the floating-poitn round
 */
LLobject *
LLFround(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(round, "fround", number, divisor, /* floatP */ TRUE));
}

/*
 *    LLFloor - floor
 *
 *    (floor number &optional divisor)
 *
 *    return the floor
 */
LLobject *
LLFloor(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(floor, "floor", number, divisor, /* floatP */ FALSE));
}

/*
 *    LLCeiling - ceiling
 *
 *    (ceiling number &optional divisor)
 *
 *    return the ceiling
 */
LLobject *
LLCeiling(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(ceil, "ceiling", number, divisor, /* floatP */ FALSE));
}

/*
 *    LLTruncate - truncate
 *
 *    (truncate number &optional divisor)
 *
 *    return the truncate
 */
LLobject *
LLTruncate(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(trunc, "truncate", number, divisor, /* floatP */ FALSE));
}

/*
 *    LLRound - round
 *
 *    (round number &optional divisor)
 *
 *    return the round
 */
LLobject *
LLRound(number, divisor)
LLobject *number;
LLobject *divisor;		/* &optional */
{
	return (func(round, "round", number, divisor, /* floatP */ FALSE));
}

/*
 *    double
 *    function(d)
 *    double d;
 */
static LLobject *
func(function, name, number, divisor, floatP)
double (*function)();
String name;
LLobject *number;
LLobject *divisor;	/* &optional */
Boolean floatP;
{
	double division;

	if ( ! LL_numberp(number) ) {
	    LLErrorF("\
Function: (~A number &optional divisor)\n\
\tnumber is not a number\n\
\tnumber: ~S\n\
\tdivisor: ~S\n\
", name, number, divisor);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LL_null(divisor)) {
	    if (floatP) {
		if (LL_integerp(number))
		    return (LLMakeReal(number->value.a_integer));
		return (LLMakeReal((*function)(number->value.a_real)));
	    } else {
		if (LL_integerp(number))
		    return (number);

		return (LLMakeInteger((int) (*function)(number->value.a_real)));
	    }
	    /*NOTREACHED*/
	}

	if ( ! LL_numberp(divisor) ) {
	    LLErrorF("\
Function: (~A number &optional divisor)\n\
\tdivisor is not a number\n\
\tnumber: ~S\n\
\tdivisor: ~S\n\
", name, number, divisor);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LL_realp(divisor) ? (divisor->value.a_real == 0.0): (divisor->value.a_integer == 0)) {
	    LLErrorF("\
Function: (~A number &optional divisor)\n\
\tdivisor is zero\n\
\tnumber: ~S\n\
\tdivisor: ~S\n\
", name, number, divisor);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LL_realp(number) && LL_realp(divisor)) {
	    division  = number->value.a_real / divisor->value.a_integer;
	} else if (LL_realp(number)) {
	    division  = number->value.a_real / divisor->value.a_integer;
	} else if (LL_realp(divisor)) {
	    division  = number->value.a_integer / divisor->value.a_real;
	} else {
	    division  = ((double) number->value.a_integer) / divisor->value.a_integer;
	}

	if (floatP)
	    return (LLMakeReal((*function)(division)));
	return (LLMakeInteger((int) (*function)(division)));
}

static double
trunc(d)
double d;
{
	int id;

	id = (int) d;
	return ((double) id);
}

static double
round(d)
double d;
{
	int id;

	id = d < 0.0 ? (int) (d - 0.5): (int) (d + 0.5);
	return ((double) id);
}
