#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLFormat - format
 *
 *    (format destination control-string &rest args)
 *
 *    return {nil, string}
 */
LLobject *
LLFormat(destination, control_string, args)
LLobject *destination;
LLobject *control_string;		
LLobject *args;		/* &rest */
{
	LLobject *stream, *promise, *ret;
	STREAM *s;
	Boolean nilP;

	if ( ! LL_stringp(control_string) ) {
	    LLErrorF("\
Function: (format destination control-string &rest args)\n\
\tcontrol-string is not a string\n\
\tdestination:    ~S\n\
\tcontrol-string: ~S\n\
\targs:           ~S\n\
", destination, control_string, args);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (destination == LL_t) {
	    stream = LLGetValue(LLMakeSymbol("*standard-output*"),
		    LL_valueSlot, LL_environment);
	    promise = LL_nil;
	    nilP = TRUE;	/* return nil */
	} else if (LL_null(destination)) {
	    s = STREAMCreate(STREAM_void, NULL, NULL,
		    STREAM_string, "", "format");
	    if (s == NULL) {
		    LLErrorF("\
Function: (format destination control-string &rest args)\n\
\tcould not create a string stream to write into\n\
\tdestination:    ~S\n\
\tcontrol-string: ~S\n\
\targs:           ~S\n\
", destination, control_string, args);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    stream = LLMakeStream(s);
	    promise = LLCons(LLMakeSymbol("close"), LLCons(stream, LL_nil));
	    nilP = FALSE;	/* return the string */
	} else {
	    /* stream will be checked to insure its a stream later */
	    stream = destination;
	    promise = LL_nil;
	    nilP = TRUE;	/* return nil */
	}

	/*
	 *    The promise must be an implicit progn.
	 *    Catch-void just to eval the promise.
	 */
	promise = LLCons(promise, LL_nil);
	if (LLCatchVoid(LL_uncatchable, promise, LL_environment))
	    return (LL_nil);

	if ( ! LL_streamp(stream) ) {
	    LLErrorF("\
Function: (format destination control-string &rest args)\n\
\tdestination cannot be interpreted as a stream\n\
\tdestination:    ~S\t; interpretation: ~A\n\
\tcontrol-string  ~S\n\
\targs:           ~S\n\
", destination, LL_type_of(stream)->pname, control_string, args);
	    LLThrow(LL_lightlisp, LL_nil);
	}
	if ( ! LLOutputstreamp(stream) ) {
	    LLErrorF("\
Function: (format destination control-string &rest args)\n\
\tdestination is not an output-capable stream\n\
\tdestination:    ~S\t; interpretation: ~A\n\
\tcontrol-string  ~S\n\
\targs:           ~S\n\
", destination, LL_type_of(stream)->pname, control_string, args);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	LLFormatS(stream->value.a_stream, control_string->value.a_string, args);
	ret = nilP ? LL_nil:
		LLMakeString(STREAMGetString(stream->value.a_stream, STREAM_output));

	LLEndCatch();		/* eval the promise (maybe kill the stream) */
	return (ret);
}
