#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "machine.h"
#include "internal.h"

/*
 *    LLFuncall - funcall
 *
 *    (funcall function &rest args)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
LLFuncall(function, args)
LLobject *function;
LLobject *args;		/* &rest */
{
	LLobject *environment, *sexp, *ret;

	if (LL_symbolp(function))
	    function = LLGetValue(function, LL_functionSlot, LL_environment);

	if (LL_functionp(function)) {
	    ret = LLApplyC(function, args,
		    LL_environment, /* evaluator */ NULL);
	} else if (LL_function_closurep(function)) {
	    environment = LLLambdaArgs(
		    LL_first(LL_code(function)), LL_env(function),
		    args, LL_environment,
		    /* evaluator */ NULL, /* destructuringP */ FALSE);

	    sexp = LL_rest(LL_code(function));

	    /*
	     *   Walk over the optional documentation string
	     *   See also LLEvaluator and LLFuncall
	     */
	    if (LL_stringp(LL_first(sexp)) && ! LL_null(LL_cdr(sexp)))
		sexp = LL_cdr(sexp);

	    ret = LLProgn(sexp, environment);
	} else {
	    LLErrorF("\
Function: (funcall function &rest args)\n\
\tfunction is not a function\n\
\tform: ~S\n\
", function);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	return (ret);
}
