#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <math.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "reader.h"
#include "printer.h"
#include "version.h"
#include "internal.h"

static struct {
    String pname;
    LLobject *(*special_form)();
} special_forms[] = {
    { "and",		LLAnd_tr		},	/* macro */
    { "case",		LLCase_tr		},	/* macro */
    { "catch",		LLCatch_tr		},
    { "cond",		LLCond_tr		},	/* macro */
    { "defconstant",	LLDefconstant_tr	},	/* macro */
    { "defmacro",	LLDefmacro_tr		},	/* macro */
    { "defparameter",	LLDefparameter_tr	},	/* macro */
    { "defun",		LLDefun_tr		},	/* macro */
    { "defvar",		LLDefvar_tr		},	/* macro */
    { "do",		LLDo_tr			},	/* macro */
    { "do*",		LLDostar_tr		},	/* macro */
    { "dolist",		LLDolist_tr		},	/* macro */
    { "dotimes",	LLDotimes_tr		},	/* macro */
    { "flet",		LLFlet_tr		},
    { "if",		LLIf_tr			},
    { "labels",		LLLabels_tr		},
    { "let",		LLLet_tr		},
    { "let*",		LLLetstar_tr		},
    { "macrolet",	LLMacrolet_tr		},
    { "or",		LLOr_tr			},	/* macro */
    { "prog1",		LLProg1_tr		},	/* macro */
    { "prog2",		LLProg2_tr		},	/* macro */
    { "progn",		LLProgn_tr		},
    { "time",		LLTime_tr,		},	/* macro */
    { "setq",		LLSetq_tr,		},
    { "subshell",	LLSubshell_tr,		},
    { "unless",		LLUnless_tr		},	/* macro */
    { "unwind-protect",	LLUnwindprotect_tr	},
    { "when",		LLWhen_tr		},	/* macro */
};

static struct {
    String pname;
    LLobject *(*function)();
    int nrequired;
    int noptional;
    Boolean restP;
} functions[] = {
    { "1+",			LL1Plus,		1,	0,	FALSE },
    { "1-",			LL1Minus,		1,	0,	FALSE },
    { "abs",			LLAbs,			1,	0,	FALSE },
    { "acos",			LLAcos,			1,	0,	FALSE },
    { "asin",			LLAsin,			1,	0,	FALSE },
    { "atan",			LLAtan,			1,	1,	FALSE },
    { "atom",			LLAtom,			1,	0,	FALSE },
    { "append",			LLAppend,		0,	0,	TRUE },
    { "apply",			LLApply,		2,	0,	TRUE },
    { "boundp",			LLBoundp,		1,	0,	FALSE },
    { "caaaar",			LLCaaaar,		1,	0,	FALSE },
    { "caaadr",			LLCaaadr,		1,	0,	FALSE },
    { "caaar",			LLCaaar,		1,	0,	FALSE },
    { "caadar",			LLCaadar,		1,	0,	FALSE },
    { "caaddr",			LLCaaddr,		1,	0,	FALSE },
    { "caadr",			LLCaadr,		1,	0,	FALSE },
    { "caar",			LLCaar,			1,	0,	FALSE },
    { "cadaar",			LLCadaar,		1,	0,	FALSE },
    { "cadadr",			LLCadadr,		1,	0,	FALSE },
    { "cadar",			LLCadar,		1,	0,	FALSE },
    { "caddar",			LLCaddar,		1,	0,	FALSE },
    { "cadddr",			LLCadddr,		1,	0,	FALSE },
    { "caddr",			LLCaddr,		1,	0,	FALSE },
    { "cadr",			LLCadr,			1,	0,	FALSE },
    { "car",			LLCar,			1,	0,	FALSE },
    { "cdaaar",			LLCdaaar,		1,	0,	FALSE },
    { "cdaadr",			LLCdaadr,		1,	0,	FALSE },
    { "cdaar",			LLCdaar,		1,	0,	FALSE },
    { "cdadar",			LLCdadar,		1,	0,	FALSE },
    { "cdaddr",			LLCdaddr,		1,	0,	FALSE },
    { "cdadr",			LLCdadr,		1,	0,	FALSE },
    { "cdar",			LLCdar,			1,	0,	FALSE },
    { "cddaar",			LLCddaar,		1,	0,	FALSE },
    { "cddadr",			LLCddadr,		1,	0,	FALSE },
    { "cddar",			LLCddar,		1,	0,	FALSE },
    { "cdddar",			LLCdddar,		1,	0,	FALSE },
    { "cddddr",			LLCddddr,		1,	0,	FALSE },
    { "cdddr",			LLCdddr,		1,	0,	FALSE },
    { "cddr",			LLCddr,			1,	0,	FALSE },
    { "cdr",			LLCdr,			1,	0,	FALSE },
    { "ceiling",		LLCeiling,		1,	1,	FALSE },
    { "clear-output",		LLClearoutput,		1,	0,	FALSE },
    { "close",			LLClose,		1,	0,	FALSE },
    { "coerce",			LLCoerce,		2,	0,	FALSE },
    { "concatenate",		LLConcatenate,		1,	0,	TRUE },
    { "consp",			LLConsp,		1,	0,	FALSE },
    { "copy-list",		LLCopylist,		1,	0,	FALSE },
    { "copy-seq",		LLCopyseq,		1,	0,	FALSE },
    { "copy-tree",		LLCopytree,		1,	0,	FALSE },
    { "cos",			LLCos,			1,	0,	FALSE },
    { "/",			LLDiv,			1,	0,	TRUE },
    { "eighth",			LLEighth,		1,	0,	FALSE },
    { "elt",			LLElt,			2,	0,	FALSE },
    { "eq",			LLEq,			2,	0,	FALSE },
    { "eql",			LLEql,			2,	0,	FALSE },
    { "equal",			LLEqual,		2,	0,	FALSE },
    { "equalp",			LLEqualp,		2,	0,	FALSE },
    { "eval",			LLEval,			1,	0,	FALSE },
    { "evenp",			LLEvenp,		1,	0,	FALSE },
    { "exit",			LLExit,			0,	0,	FALSE },
    { "exp",			LLExp,			1,	0,	FALSE },
    { "expt",			LLExpt,			2,	0,	FALSE },
    { "fboundp",		LLFboundp,		1,	0,	FALSE },
    { "fceiling",		LLFceiling,		1,	1,	FALSE },
    { "ffloor",			LLFfloor,		1,	1,	FALSE },
    { "fifth",			LLFifth,		1,	0,	FALSE },
    { "finish-output",		LLFinishoutput,		1,	0,	FALSE },
    { "first",			LLFirst,		1,	0,	FALSE },
    { "floatp",			LLFloatp,		1,	0,	FALSE },
    { "floor",			LLFloor,		1,	1,	FALSE },
    { "fmakunbound",		LLFmakunbound,		1,	0,	FALSE },
    { "force-output",		LLForceoutput,		1,	0,	FALSE },
    { "format",			LLFormat,		2,	0,	TRUE },
    { "fourth",			LLFourth,		1,	0,	FALSE },
    { "fresh-line",		LLFreshline,		0,	1,	FALSE },
    { "fround",			LLFround,		1,	1,	FALSE },
    { "ftruncate",		LLFtruncate,		1,	1,	FALSE },
    { "funcall",		LLFuncall,		1,	0,	TRUE },
    { "functionp",		LLFunctionp,		1,	0,	FALSE },
    { ">=",			LLGe,			1,	0,	TRUE },
    { "gensym",			LLGensym,		0,	1,	FALSE },
    { ">",			LLGt,			1,	0,	TRUE },
    { "identity",		LLIdentity,		1,	0,	FALSE },
    { "input-stream-p",		LLInputstreamp,		1,	0,	FALSE },
    { "integerp",		LLIntegerp,		1,	0,	FALSE },
    { "isqrt",			LLIsqrt,		1,	0,	FALSE },
    { "last",			LLLast,			1,	0,	FALSE },
    { "<=",			LLLe,			1,	0,	TRUE },
    { "length",			LLLength,		1,	0,	FALSE },
    { "list*",			LLListstar,		1,	0,	TRUE },
    { "list",			LLList,			0,	0,	TRUE },
    { "listp",			LLListp,		1,	0,	FALSE },
    { "list-length",		LLListlength,		1,	0,	FALSE },
    { "load",			LLLoad,			1,	0,	FALSE },
    { "log",			LLLog,			1,	1,	FALSE },
    { "<",			LLLt,			1,	0,	TRUE },
    { "macro-function",		LLMacrofunction,	1,	0,	FALSE },
    { "macroexpand",		LLMacroexpand,		1,	1,	FALSE },
    { "macroexpand-1",		LLMacroexpand1,		1,	1,	FALSE },
    { "make-symbol",		LLMakesymbol,		1,	0,	FALSE },
    { "makunbound",		LLMakunbound,		1,	0,	FALSE },
    { "map",			LLMap,			3,	0,	TRUE },
    { "max",			LLMax,			1,	0,	TRUE },
    { "member",			LLMember,		2,	0,	FALSE },
    { "min",			LLMin,			1,	0,	TRUE },
    { "-",			LLMinus,		1,	0,	TRUE },
    { "minusp",			LLMinusp,		1,	0,	FALSE },
    { "mod",			LLMod,			2,	0,	FALSE },
    { "nconc",			LLNconc,		0,	0,	TRUE },
    { "ninth",			LLNinth,		1,	0,	FALSE },
    { "not",			LLNull,			1,	0,	FALSE },
    { "nth",			LLNth,			2,	0,	FALSE },
    { "nthcdr",			LLNthcdr,		2,	0,	FALSE },
    { "null",			LLNull,			1,	0,	FALSE },
    { "numberp",		LLNumberp,		1,	0,	FALSE },
    { "=",			LLNumeq,		1,	0,	TRUE },
    { "/=",			LLNumnoteq,		1,	0,	TRUE },
    { "oblist",			LLOblist,		0,	0,	FALSE },
    { "oddp",			LLOddp,			1,	0,	FALSE },
    { "open",			LLOpen,			1,	0,	TRUE },
    { "output-stream-p",	LLOutputstreamp,	1,	0,	FALSE },
    { "pairlis",		LLPairlis,		2,	1,	FALSE },
    { "+",			LLPlus,			0,	0,	TRUE },
    { "plusp",			LLPlusp,		1,	0,	FALSE },
    { "pprint",			LLPprint,		1,	1,	FALSE },
    { "prin1",			LLPrin1,		1,	1,	FALSE },
    { "princ",			LLPrinc,		1,	1,	FALSE },
    { "print",			LLPrint,		1,	1,	FALSE },
    { "random",			LLRandom,		1,	1,	FALSE },
    { "read",			LLRead,			0,	4,	FALSE },
    { "read-from-string",	LLReadfromstring,	1,	2,	FALSE },
    { "rem",			LLRem,			2,	0,	FALSE },
    { "rest",			LLRest,			1,	0,	FALSE },
    { "round",			LLRound,		1,	1,	FALSE },
    { "rplaca",			LLRplaca,		2,	0,	FALSE },
    { "rplacd",			LLRplacd,		2,	0,	FALSE },
    { "second",			LLSecond,		1,	0,	FALSE },
    { "set",			LLSet,			2,	0,	FALSE },
    { "seventh",		LLSeventh,		1,	0,	FALSE },
    { "signum",			LLSignum,		1,	0,	FALSE },
    { "sin",			LLSin,			1,	0,	FALSE },
    { "sixth",			LLSixth,		1,	0,	FALSE },
    { "special-form-p",		LLSpecialformp,		1,	0,	FALSE },
    { "sqrt",			LLSqrt,			1,	0,	FALSE },
    { "streamp",		LLStreamp,		1,	0,	FALSE },
    { "stringp",		LLStringp,		1,	0,	FALSE },
    { "subseq",			LLSubseq,		2,	1,	FALSE },
    { "symbol-function",	LLSymbolfunction,	1,	0,	FALSE },
    { "symbol-name",		LLSymbolname,		1,	0,	FALSE },
    { "symbol-value",		LLSymbolvalue,		1,	0,	FALSE },
    { "symbolp",		LLSymbolp,		1,	0,	FALSE },
    { "tan",			LLTan,			1,	0,	FALSE },
    { "tenth",			LLTenth,		1,	0,	FALSE },
    { "terpri",			LLTerpri,		0,	1,	FALSE },
    { "third",			LLThird,		1,	0,	FALSE },
    { "*",			LLTimes,		0,	0,	TRUE },
    { "throw",			LLThrow,		2,	0,	FALSE },
    { "truncate",		LLTruncate,		1,	1,	FALSE },
    { "type-of",		LLTypeof,		1,	0,	FALSE },
    { "zerop",			LLZerop,		1,	0,	FALSE },
/* the sequent and hp 4.3 (utah) don't define these */
#if ! (OPSYS_MFR == MD_SEQUENT || \
	(OPSYS_MFR == MD_HP && OPSYS_VERS == MD_BSD_4_3)) /* { */
    { "atanh",			LLAtanh,		1,	0,	FALSE },
    { "asinh",			LLAsinh,		1,	0,	FALSE },
    { "acosh",			LLAcosh,		1,	0,	FALSE },
    { "cosh",			LLCosh,			1,	0,	FALSE },
    { "sinh",			LLSinh,			1,	0,	FALSE },
    { "tanh",			LLTanh,			1,	0,	FALSE },
#endif /* } */
};

/*
 *    LLInitialize - initialize LightLisp
 *
 *    return {LL_ok, LL_notok} on {success, fail}
 */
LLstatus
LLInitialize()
{
	int i;

	if (LLInitializedP)
	    return (LL_ok);

	if (LL_Mach_Initialize() != LL_ok)
	    return (LL_notok);
	if (LL_Print_Initialize() != LL_ok)
	    return (LL_notok);
	if (LL_Read_Initialize() != LL_ok)
	    return (LL_notok);

	LLSetValue(LLMakeSymbol("*standard-output*"), LL_valueSlot,
		LLMakeStream(LLStandardStream), LL_environment);
	LLSetValue(LLMakeSymbol("*error-output*"), LL_valueSlot,
		LLMakeStream(LLErrorStream), LL_environment);
	LLSetValue(LLMakeSymbol("call-arguments-limit"), LL_valueSlot,
		LLMakeInteger(LL_MAXARGS), LL_environment);
	LLSetValue(LLMakeSymbol("lambda-parameters-limit"), LL_valueSlot,
		LLMakeInteger(LL_MAXARGS), LL_environment);

	LLSetValue(LLMakeSymbol("e"), LL_valueSlot,
		LLMakeReal(M_E), LL_environment);
	LLSetValue(LLMakeSymbol("pi"), LL_valueSlot,
		LLMakeReal(M_PI), LL_environment);

	LLSetValue(LLMakeSymbol("lisp-implementation-type"), LL_valueSlot,
		LLMakeString(LISP_IMPLEMENTATION_TYPE), LL_environment);
	LLSetValue(LLMakeSymbol("lisp-implementation-version"), LL_valueSlot,
		LLMakeString(LISP_IMPLEMENTATION_VERSION), LL_environment);

	for (i=0; i<sizeof_el (functions); i++) {
	    LLSetValue(LLMakeSymbol(functions[i].pname),
		    LL_functionSlot,
		    LLMakeFunction(functions[i].pname, functions[i].function,
			    functions[i].nrequired,
			    functions[i].noptional,
			    functions[i].restP),
		    LL_environment);
	}

	for (i=0; i<sizeof_el (special_forms); i++) {
	    LLSetValue(LLMakeSymbol(special_forms[i].pname),
		    LL_functionSlot,
		    LLMakeSpecialForm(special_forms[i].pname,
			    special_forms[i].special_form),
		    LL_environment);
	}

	LLInitializedP = TRUE;

	return (LL_ok);
}
