#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *body();

/*
 *    LLLet - let
 *
 *    (let ( var | (var value) ... )
 *           forms ... )
 *
 *    return the result
 */
LLobject *
LLLet(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLLet_tr(&arguments, &environment)))
	    return (arguments);
	
	return (LLEvaluator(arguments, environment));
}

/*
 *    LLLet_tr - let
 *
 *    (let ( var | (var value) ... )
 *           forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLLet_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (body(arguments, environment,
		"(let ( var | (var value) ... ) forms ... )",
		/* starP */ FALSE, LL_valueSlot, NULL));
}

/*
 *    LLLetstar - let*
 *
 *    (let* ( var | (var value) ... )
 *           forms ... )
 *
 *    return the result
 */
LLobject *
LLLetstar(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLLetstar_tr(&arguments, &environment)))
	    return (arguments);
	
	return (LLEvaluator(arguments, environment));
}

/*
 *    LLLetstar_tr - let*
 *
 *    (let* ( var | (var value) ... )
 *           forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLLetstar_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (body(arguments, environment,
		"(let* ( var | (var value) ... ) forms ... )",
		/* starP */ TRUE, LL_valueSlot, NULL));
}

/*
 *    LLFlet - flet
 *
 *    (flet ((name lambda-list [ documentation ] form ... ) ... )
 *          forms ... )
 *
 *    return the result
 */
LLobject *
LLFlet(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLFlet_tr(&arguments, &environment)))
	    return (arguments);
	
	return (LLEvaluator(arguments, environment));
}

/*
 *    LLFlet_tr - flet
 *
 *    (flet ((name lambda-list [ documentation ] form ... ) ... )
 *          forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLFlet_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (body(arguments, environment,
		"(flet ((name lambda-list [ documentation ] form ... ) ... ) forms ... )",
		/* starP */ FALSE, LL_functionSlot, LLFunctionClosure));
}

/*
 *    LLLabels - labels
 *
 *    (labels ((name lambda-list [ documentation ] form ... ) ... )
 *            forms ... )
 *
 *    return the result
 */
LLobject *
LLLabels(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLLabels_tr(&arguments, &environment)))
	    return (arguments);
	
	return (LLEvaluator(arguments, environment));
}

/*
 *    LLLabels_tr - labels
 *
 *    (labels ((name lambda-list [ documentation ] form ... ) ... )
 *            forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLLabels_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (body(arguments, environment,
		"(labels ((name lambda-list [ documentation ] form ... ) ... ) forms ... )",
		/* starP */ TRUE, LL_functionSlot, LLFunctionClosure));
}

/*
 *    LLMacrolet - macrolet
 *
 *    (macrolet ((name lambda-list [ documentation ] form ... ) ... )
 *              forms ... )
 *
 *    return the result
 */
LLobject *
LLMacrolet(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLMacrolet_tr(&arguments, &environment)))
	    return (arguments);
	
	return (LLEvaluator(arguments, environment));
}

/*
 *    LLMacrolet_tr - macrolet
 *
 *    (macrolet ((name lambda-list [ documentation ] form ... ) ... )
 *               forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLMacrolet_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (body(arguments, environment,
		"(macrolet ((name lambda-list [ documentation ] form ... ) ... ) forms ... )",
		/* starP */ FALSE, LL_functionSlot, LLMacroClosure));
}

/*
 *    WorkHorse ------------------------
 *
 *    LLobject *
 *    closure(code, environment)
 *    LLobject *code;
 *    LLobject *environment;
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
static LLobject *
body(arguments, environment, form, starP, slot, closure)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
String form;			/* e.g. (let () ... ) */
Boolean starP;			/* let* vs let - labels vs flet */
LLslot slot;			/* flet vs let */
LLobject *(*closure)();		/* only if slot == LL_functionSlot */
{
	LLobject *initializer, *i,
		*finit, *var, *symbol, *value,
		*letenv, **evalenv;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: ~A\n\
\tillegal number of arguments to let\n\
\tform: ~S\n\
", form, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	initializer = LL_first(*arguments);

	if ( ! LL_listp(initializer) ) {
	    LLErrorF("\
Form: ~A\n\
\tinitializer argument must be a list\n\
\tinitializer: ~S\n\
\tform:        ~S\n\
", form, initializer, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	letenv = *environment;
	evalenv = starP ? &letenv: environment;

	for (i=initializer; LL_consp(i); i=LL_cdr(i)) {
	    if (slot == LL_functionSlot) {
		finit = LL_car(i);

		if ( ! LL_consp(finit) || LLGetLength(finit) < 2) {
		    LLErrorF("\
Form: ~A\n\
\tbadly formed function declaration\n\
\tvar-lambda-list: ~S\n\
\tform:            ~S\n\
", form, finit, *arguments);
		    LLThrow(LL_lightlisp, LL_nil);
		}

		/*
		 *    The lambda-list etc is checked for okayness
		 *    when it is used in LLLambdaArgs
		 */
		symbol = LL_first(finit);
		if (starP) {
		    /* labels will be closured over later */
		    value = LL_rest(finit);
		} else {
		    /* flet and macrolet */
		    value = (*closure)(LL_rest(finit), *evalenv);
		}
	    } else {
		var = LL_car(i);

		if ( ! LL_consp(var) ) {
		    symbol = var;
		    value = LL_nil;
		} else {
		    if (LLGetLength(var) >= 3) {
			LLErrorF("\
Form: ~A\n\
\ttoo many members in var-value list\n\
\tvar-value: ~S\n\
\tform:      ~S\n\
", form, var, *arguments);
			LLThrow(LL_lightlisp, LL_nil);
		    }

		    symbol = LL_first(var);
		    value = LLEvaluator(LL_second(var), *evalenv);
		}
	    }

	    if ( ! LL_symbolp(symbol) ) {
		LLErrorF("\
Form: ~A\n\
\tname is not a symbol\n\
\tsymbol: ~S\n\
\tform:   ~S\n\
", form, symbol, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    letenv = LLExtendEnvironment(symbol, slot, value, letenv);
	}

	if (slot == LL_functionSlot && starP) {
	    /*
	     *    Now go back in and for each member of the labels,
	     *    take the closure over environment the whole labels
	     *    so that they can define recursive and mutually recursive
	     *    functions
	     */
	    for (i=initializer; LL_consp(i); i=LL_cdr(i)) {
		finit = LL_car(i);
		symbol = LL_first(finit);

		value = (*closure)(LLGetValue(symbol, LL_functionSlot, letenv), letenv);
		LLSetValue(symbol, LL_functionSlot, value, letenv);
	    }
	}

	*environment = letenv;
	*arguments = LL_rest(*arguments);
	return (LLProgn_tr(arguments, environment));
}
