#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "reader.h"
#include "internal.h"

/*
 *    LLLoad - load
 *
 *    (load path)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
LLLoad(path)
LLobject *path;
{
	LLobject *sexp, *promise;
	STREAM *stream;

	/*
	 *    For now strings are the only path components
	 */
	if ( ! LL_stringp(path) ) {
	    LLErrorF("\
Function: (load path)\n\
\tpath is not a string\n\
\tpath: ~S\n\
", path);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	stream = STREAMCreate(STREAM_filename, path->value.a_string, path->value.a_string,
		STREAM_STREAM, LLErrorStream, NULL);
	if (stream == NULL) {
	    LLErrorF("\
Function: (load path)\n\
\tcould not open the file because ~A\n\
\tpath: ~S\n\
", sys_errlist[errno], path);
	    return (LL_nil);	/* No, don't throw */
	}

	promise = LLCons(LLMakeSymbol("close"), LLCons(LLMakeStream(stream), LL_nil));
	promise = LLCons(promise, LL_nil);

	/* not a top-level catch b/c of the promise */
	if (LLCatchVoid(LL_alwayscatchable, promise, LL_environment))
	    return (LL_nil);

	while ((sexp = LLReadValue(stream, FALSE, LL_eofvalue, FALSE, LL_lightlisp)) != LL_eofvalue) {
	    LLEvaluator(sexp, LL_environment);
	    /* DON'T LLGarbageCollect(); */
	}

	LLEndCatch();	 /* evaluate the promise */

	return (LL_t);
}
