#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "machine.h"
#include "internal.h"

static LLobject *expand();

/*
 *    LLMacroexpand - macroexpand
 *
 *    (macroexpand form &optional env)
 *
 *    return the expanded form
 */
LLobject *
LLMacroexpand(form, env)
LLobject *form;
LLobject *env;		/* &optional */
{
	return (expand(form, env, /* tilldoneP */ TRUE));
}

/*
 *    LLMacroexpand1 - macroexpand1
 *
 *    (macroexpand-1 form &optional env)
 *
 *    return the expanded form
 */
LLobject *
LLMacroexpand1(form, env)
LLobject *form;
LLobject *env;		/* &optional */
{
	return (expand(form, env, /* tilldoneP */ FALSE));
}

static LLobject *
expand(form, env, tilldoneP)
LLobject *form;
LLobject *env;		/* &optional */
Boolean tilldoneP;
{
	LLobject *macro, *environment, *sexp, *ret;

	if (LL_null(env))
	    env = LL_environment;

	do {
	    if ( ! LL_consp(form) ) {
		ret = form;
		break;	 /* done */
	    }

	    macro = LL_car(form);
	    /*
	     *    If its a symbol that has a function binding
	     *    get the function binding, else (ultimately)
	     *    quit the expansion
	     */
	    if (LL_symbolp(macro) && ! LL_null(LLFboundp(macro)))
		macro = LLGetValue(macro, LL_functionSlot, LL_environment);

	    if (LL_macrop(macro)) {
		ret = form = LL_function_call(macro)(LL_cdr(form), env);
		/* go again? */
	    } else if (LL_macro_closurep(macro)) {
		environment = LLLambdaArgs(
			LL_first(LL_code(macro)), LL_env(macro),
			LL_rest(form), LL_environment,      /* global env */
			/* evaluator */ NULL, /* destructuringP */ TRUE);

		sexp = LL_rest(LL_code(macro));

		/*
		 *   Walk over the optional documentation string
		 */
		if (LL_stringp(LL_first(sexp)) && ! LL_null(LL_cdr(sexp)))
		    sexp = LL_cdr(sexp);

		if ( ! LL_null(LLProgn_tr(&sexp, &environment)) )
		    sexp = LLEvaluator(sexp, environment);

		ret = form = sexp;
		/* go again? */
	    } else {
		ret = form;
		break;	 /* done */
	    }
	} while (tilldoneP);

	return (ret);
}
