#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLMap - map
 *
 *    (map result-type function sequence &rest sequences)
 *
 *    return the mapping
 */
LLobject *
LLMap(result_type, function, sequence, sequences)
LLobject *result_type;		/* may be nil */
LLobject *function;
LLobject *sequence;
LLobject *sequences;		/* &rest */
{
	LLobject *s, *thislen, *tmp, *ret, *last, *args, *lastargs, *piece;
	LLsequence *seq;
	int minlen, i;

	if ( ! LL_null(result_type) ) {
	    if ( ! LL_symbolp(result_type) && ! LL_typep(result_type, LLTypeCell) ) {
		LLErrorF("\
Function: (map result-type function sequence &rest sequences)\n\
\tresult-type must be given by a type-name symbol\n\
\tresult-type: ~S\n\
", result_type);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    if ((seq = LLLookupSequence(result_type)) == NULL) {
		LLErrorF("\
Function: (map result-type function sequence &rest sequences)\n\
\tresult-type is not a sequence\n\
\tform: ~S\n\
", result_type);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	}

	if (LL_symbolp(function))
	    function = LLGetValue(function, LL_functionSlot, LL_environment);

	if ( ! LL_functionp(function) && ! LL_function_closurep(function) ) {
	    LLErrorF("\
Function: (map result-type function sequence &rest sequences)\n\
\tfunction is not a function\n\
\tform: ~S\n\
", function);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	sequences = LLCons(sequence, sequences);
	for (s=sequences, minlen = MAXINT; ! LL_null(s); s=LL_cdr(s)) {
	    sequence = LL_car(s);
	    if ( ! LL_sequencep(sequence) ) {
		LLErrorF("\
Function: (map result-type function sequence &rest sequences)\n\
\tsequence in sequences is not a sequence\n\
\tform: ~S\n\
", sequence);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    thislen = LLLength(sequence);
	    if ( ! LL_integerp(thislen) ) {
		LLErrorF("\
Function: (map result-type function sequence &rest sequences)\n\
\tlength of sequence returned is not an integer\n\
\tsequence: ~S\n\
\tlength:   ~S\n\
", sequence, thislen);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    minlen = MIN(minlen, thislen->value.a_integer);
	}

	ret = LL_nil;
	last = LL_nil;

	for (i=0; i<minlen; i++) {
	    args =  LL_nil;
	    lastargs = LL_nil;

	    for (s=sequences; ! LL_null(s); s=LL_cdr(s)) {
		sequence = LL_car(s);
		tmp = LLCons(LLElt(sequence, LLMakeInteger(i)), LL_nil);
		if (LL_null(args))
		    args = tmp;
		else
		    LL_rplacd(lastargs, tmp);
		lastargs = tmp;
	    }

	    /*
	     *    (funcall function &rest args) - so funcall is appropriate here
	     *    or LLApply(function args, LL_nil);
	     */
	    piece = LLFuncall(function, args);
	    if (LL_null(result_type))
		continue;		/* side-effect only */

	    tmp = LLCons(piece, LL_nil);
	    if (LL_null(ret)) {
		ret = tmp;
	    } else {
		LL_rplacd(last, tmp);
	    }
	    last = tmp;
	}

	/*
	 *    If the result-type is nil, then don't call coerce
	 *    and further, seq is probably totally bogus!
	 */
	return (LL_null(result_type) ? LL_nil: LLCoerce(ret, seq->type->object));
}
