#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *func();

/*
 *    LLMax - max
 *
 *    (max number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLMax(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, "max", /* maxP */ TRUE));
}

/*
 *    LLMin - min
 *
 *    (min number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLMin(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, "min", /* maxP */ FALSE));
}

static LLobject *
func(number, more_numbers, name, maxP)
LLobject *number;
LLobject *more_numbers;	/* &rest */
String name;
Boolean maxP;
{
	Boolean integerP;
	LLobject *n, *this;
	double rret, rtest;
	int iret;
	
	integerP = LL_integerp(number);
	if (integerP) {
	    iret = number->value.a_integer;
	} else {
	    rret = number->value.a_real;
	}

	for (n=more_numbers; LL_consp(n); n=LL_cdr(n)) {
	    this = LL_car(n);

	    if ( ! LL_numberp(this) ) {
		LLErrorF("\
Function: (~A number &rest more-numbers)\n\
\tnumber is not a number\n\
\toffender: ~S\n\
\tnumber:   ~S\n\
\tnumbers:  ~S\n\
", name, this, number, more_numbers);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    if (LL_realp(this) && integerP) {
		rret = iret;
		integerP = FALSE;
	    }

	    if (maxP) {
		if (integerP) {
		    if (iret < this->value.a_integer)
			iret = this->value.a_integer;
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rret < rtest)
			rret = rtest;
		}
	    } else {
		if (integerP) {
		    if (this->value.a_integer < iret)
			iret = this->value.a_integer;
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rtest < rret)
			rret = rtest;
		}
	    }
	}

	return (integerP ? LLMakeInteger(iret): LLMakeReal(rret));
}
