#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *func();

/*
 *    LLMinus - -
 *
 *    (- number &rest numbers)
 *
 *    return the result
 */
LLobject *
LLMinus(number, numbers)
LLobject *number;
LLobject *numbers;		/* &rest */
{
	return (func(number, numbers, "-", /* subP */ TRUE));
}

/*
 *    LLDiv - /
 *
 *    (/ number &rest numbers)
 *
 *    return the result
 */
LLobject *
LLDiv(number, numbers)
LLobject *number;
LLobject *numbers;		/* &rest */
{
	return (func(number, numbers, "/", /* subP */ FALSE));
}

static LLobject *
func(number, numbers, name, subP)
LLobject *number;
LLobject *numbers;	/* &rest */
String name;
Boolean subP;
{
	Boolean integerP;
	LLobject *n, *this;
	double rret;
	int iret, itest;
	
	if (LL_null(numbers)) {
	    if (subP) {
		if (LL_realp(number))
		    return (LLMakeReal(- number->value.a_real));
		else
		    return (LLMakeInteger(- number->value.a_integer));
	    } else {
		if (LL_realp(number) ? (number->value.a_real == 0.0): (number->value.a_integer == 0)) {
		    LLErrorF("\
Function: (~A number &rest numbers)\n\
\tdivisor is zero\n\
\toffender: ~S\n\
\tnumber:   ~S\n\
\tnumbers:  ~S\n\
", name, number, number, numbers);
		    LLThrow(LL_lightlisp, LL_nil);
		}
		if (LL_realp(number))
		    return (LLMakeReal( 1.0 / number->value.a_real));
		else
		    return (LLMakeReal( 1.0 / number->value.a_integer));
	    }
	    /*NOTREACHED*/
	}

	integerP = LL_integerp(number);
	if (integerP) {
	    iret = number->value.a_integer;
	} else {
	    rret = number->value.a_real;
	}
	
	for (n=numbers; LL_consp(n); n=LL_cdr(n)) {
	    this = LL_car(n);

	    if ( ! LL_numberp(this) ) {
		LLErrorF("\
Function: (~A number &rest numbers)\n\
\tnumber is not a number\n\
\toffender: ~S\n\
\tnumber:   ~S\n\
\tnumbers:  ~S\n\
", name, this, number, numbers);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    if (LL_realp(this) && integerP) {
		rret = iret;
		integerP = FALSE;
	    }

	    if (subP) {
		if (integerP) {
		    iret -= this->value.a_integer;
		} else {
		    rret -= LL_realp(this) ? this->value.a_real: this->value.a_integer;
		}
	    } else {
		if (LL_realp(this) ? (this->value.a_real == 0.0): (this->value.a_integer == 0)) {
		    LLErrorF("\
Function: (~A number &rest numbers)\n\
\tdivisor is zero\n\
\toffender: ~S\n\
\tnumber:   ~S\n\
\tnumbers:  ~S\n\
", name, this, number, numbers);
			LLThrow(LL_lightlisp, LL_nil);
		}

		if (integerP) {
		    itest = iret / this->value.a_integer;
		    if (itest * iret == this->value.a_integer) {
			iret = itest;
		    } else {
			rret = iret;
			integerP = FALSE;
			rret /= LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    }
		} else {
		    rret /= LL_realp(this) ? this->value.a_real: this->value.a_integer;
		}
	    }
	}

	return (integerP ? LLMakeInteger(iret): LLMakeReal(rret));
}
