#include <sys/types.h>
#include <setjmp.h>
#include <math.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

typedef enum function_op function_op;
enum function_op {
    mod,
    rem,
};
static LLobject *mod_rem();

/*
 *    LLMod - mod
 *
 *    (mod number divisor)
 *
 *    return the mod
 */
LLobject *
LLMod(number, divisor)
LLobject *number;
LLobject *divisor;
{
	return (mod_rem(number, divisor, mod, "mod"));
}

/*
 *    LLRem - rem
 *
 *    (rem number divisor)
 *
 *    return the rem
 */
LLobject *
LLRem(number, divisor)
LLobject *number;
LLobject *divisor;
{
	return (mod_rem(number, divisor, rem, "rem"));
}

/* pretend that zero is a positive number for now */
#define sign(n)	(n == 0 ? 1: (n/ABS(n)))	/* -1, 1 (or floating-point) */

static LLobject *
mod_rem(number, divisor, op, name)
LLobject *number;
LLobject *divisor;
function_op op;
String name;
{
	LLobject *ret;
	double modrem_r, divisor_r, number_r, numerator_r;
	int modrem_i, divisor_i, number_i, numerator_i;

	if ( ! LL_numberp(number) ) {
	    LLErrorF("\
Function: (~A number divisor)\n\
\tnumber is not a number\n\
\tnumber:  ~S\n\
\tdivisor: ~S\n\
", name, number, divisor);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if ( ! LL_numberp(divisor) ) {
	    LLErrorF("\
Function: (~A number divisor)\n\
\tdivisor is not a number\n\
\tnumber:  ~S\n\
\tdivisor: ~S\n\
", name, number, divisor);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LL_realp(divisor) ? (divisor->value.a_real == 0,0): (divisor->value.a_integer == 0)) {
	    LLErrorF("\
Function: (~A number divisor)\n\
\tdivisor is zero\n\
\tnumber:  ~S\n\
\tdivisor: ~S\n\
", name, number, divisor);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LL_integerp(number) && LL_integerp(divisor)) {
	    number_i = number->value.a_integer;
	    divisor_i = divisor->value.a_integer;

	    if (op == mod) {
		numerator_i = sign(number_i) != sign(divisor_i) ?
			number_i - divisor_i: number_i;

		modrem_i = number_i - (numerator_i / divisor_i) * divisor_i;
	    } else {	/* op == rem */
		modrem_i = number_i - (number_i / divisor_i) * divisor_i;
	    }

	    ret = LLMakeInteger(modrem_i);
	} else {
	    number_r = LL_realp(number) ?
		    number->value.a_real: (double) number->value.a_integer;
	    divisor_r = LL_realp(divisor) ?
		    divisor->value.a_real: (double) divisor->value.a_integer;

	    if (op == mod) {
		numerator_r = sign(number_r) != sign(divisor_r) ?
			number_r - divisor_r: number_r;

		modrem_r = number_r - ((int) (numerator_r / divisor_r)) * divisor_r;
	    } else {		/* op == rem */
		modrem_r = number_r - ((int) (number_r / divisor_r)) * divisor_r;
	    }

	    ret = LLMakeReal(modrem_r);
	}

	return (ret);
}
