#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    evenp,
    minusp,
    oddp,
    plusp,
    zerop,
};

static LLobject *func();

/*
 *    LLEvenp - evenp
 *
 *    (evenp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLEvenp(exp)
LLobject *exp;
{
	return (func(exp, "evenp", evenp));
}

/*
 *    LLMinusp - minusp
 *
 *    (minusp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLMinusp(exp)
LLobject *exp;
{
	return (func(exp, "minusp", minusp));
}

/*
 *    LLOddp - oddp
 *
 *    (oddp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLOddp(exp)
LLobject *exp;
{
	return (func(exp, "oddp", oddp));
}

/*
 *    LLPlusp - plusp
 *
 *    (plusp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLPlusp(exp)
LLobject *exp;
{
	return (func(exp, "plusp", plusp));
}

/*
 *    LLZerop - zerop
 *
 *    (zerop exp)
 *
 *    return {t, nil}
 */
LLobject *
LLZerop(exp)
LLobject *exp;
{
	return (func(exp, "zerop", zerop));
}

/*
 *    Workhorse --------------------------------
 */
static LLobject *
func(exp, name, op)
LLobject *exp;
String name;
func_op op;
{
	LLobject *ret;

	if ( ! LL_numberp(exp) ) {
	    LLErrorF("\
Function: (~A exp)\n\
\texp is not a number\n\
\texp: ~S\n\
", name, exp);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if ((op == evenp || op == oddp) && ! LL_integerp(exp) ) {
	    LLErrorF("\
Function: (~A exp)\n\
\texp is not an integer\n\
\texp: ~S\n\
", name, exp);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	switch (op) {
        case evenp:
	    ret = (exp->value.a_integer & 01) == 0 ? LL_t: LL_nil;
	    break;
        case minusp:
	    if (LL_realp(exp))
		ret = exp->value.a_real < 0.0 ? LL_t: LL_nil;
	    else
		ret = exp->value.a_integer < 0 ? LL_t: LL_nil;
	    break;
        case oddp:
	    ret = (exp->value.a_integer & 01) == 1 ? LL_t: LL_nil;
	    break;
        case plusp:
	    if (LL_realp(exp))
		ret = exp->value.a_real > 0.0 ? LL_t: LL_nil;
	    else
		ret = exp->value.a_integer > 0 ? LL_t: LL_nil;
	    break;
        case zerop:
	    if (LL_realp(exp))
		ret = exp->value.a_real == 0.0 ? LL_t: LL_nil;
	    else
		ret = exp->value.a_integer == 0 ? LL_t: LL_nil;
	    break;
	}

	return (ret);
}
