#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    eq,		/* = */
    lt,		/* < */
    le,		/* <= */
    gt,		/* > */
    ge,		/* >= */
};
static LLobject *func();

/*
 *    LLNumeq - =
 *
 *    (= number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLNumeq(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, "=", eq));
}

/*
 *    LLLt - <
 *
 *    (< number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLLt(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, "<", lt));
}

/*
 *    LLLe - <=
 *
 *    (<= number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLLe(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, "<=", le));
}

/*
 *    LLGt - >
 *
 *    (> number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLGt(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, ">", gt));
}

/*
 *    LLGe - >=
 *
 *    (>= number &rest more-numbers)
 *
 *    return the result
 */
LLobject *
LLGe(number, more_numbers)
LLobject *number;
LLobject *more_numbers;		/* &rest */
{
	return (func(number, more_numbers, ">=", ge));
}

static LLobject *
func(number, more_numbers, name, op)
LLobject *number;
LLobject *more_numbers;	/* &rest */
String name;
func_op op;
{
	Boolean integerP, retP;
	LLobject *n, *this;
	double rret, rtest;
	int iret;
	
	integerP = LL_integerp(number);
	if (integerP) {
	    iret = number->value.a_integer;
	} else {
	    rret = number->value.a_real;
	}

	retP = TRUE;
	for (n=more_numbers; LL_consp(n); n=LL_cdr(n)) {
	    this = LL_car(n);

	    if ( ! LL_numberp(this) ) {
		LLErrorF("\
Function: (~A number &rest more-numbers)\n\
\tnumber is not a number\n\
\toffender: ~S\n\
\tnumber:   ~S\n\
\tnumbers:  ~S\n\
", name, this, number, more_numbers);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    if (LL_realp(this) && integerP) {
		rret = iret;
		integerP = FALSE;
	    }

	    switch (op) {
	    case eq:
		/*
		 *    =
		 */
		if (integerP) {
		    if (iret != this->value.a_integer) {
			retP = FALSE;
			goto out;
		    }
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rret != rtest) {
			retP = FALSE;
			goto out;
		    }
		}
		break;
	    case lt:
		/*
		 *    <
		 */
		if (integerP) {
		    if (iret >= this->value.a_integer) {
			retP = FALSE;
			goto out;
		    }
		    iret = this->value.a_integer;
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rret >= rtest) {
			retP = FALSE;
			goto out;
		    }
		    rret = rtest;
		}
		break;
	    case le:
		/*
		 *    <=
		 */
		if (integerP) {
		    if (iret > this->value.a_integer) {
			retP = FALSE;
			goto out;
		    }
		    iret = this->value.a_integer;
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rret > rtest) {
			retP = FALSE;
			goto out;
		    }
		    rret = rtest;
		}
		break;
	    case gt:
		/*
		 *    >
		 */
		if (integerP) {
		    if (iret <= this->value.a_integer) {
			retP = FALSE;
			goto out;
		    }
		    iret = this->value.a_integer;
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rret <= rtest) {
			retP = FALSE;
			goto out;
		    }
		    rret = rtest;
		}
		break;
	    case ge:
		/*
		 *    >=
		 */
		if (integerP) {
		    if (iret < this->value.a_integer) {
			retP = FALSE;
			goto out;
		    }
		    iret = this->value.a_integer;
		} else {
		    rtest = LL_realp(this) ? this->value.a_real: this->value.a_integer;
		    if (rret < rtest) {
			retP = FALSE;
			goto out;
		    }
		    rret = rtest;
		}
		break;
	    }
	}
out:	;

	return (retP ? LL_t: LL_nil);
}
