#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "reader.h"
#include "internal.h"

/*
 *    LLOpen - open
 *
 *    (open path &keys :direction [ :input | :output ] )
 *
 *    b/c LightLisp does not do &keys yet
 *        use (open path &rest keys)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
LLOpen(path, keys)
LLobject *path;
LLobject *keys;
{
	Boolean directionP, inputP;
	LLobject *key, *keyval;
	STREAM *s;

	/*
	 *    For now strings are the only path components
	 */
	if ( ! LL_stringp(path) ) {
	    LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tpath is not a string\n\
\tpath: ~S\n\
", path);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	directionP = FALSE;
	inputP = TRUE;		/* by default, read it */

	while ( ! LL_null(keys) ) {
	    key = LL_first(keys);
	    keyval = LL_second(keys);
	    if ( ! LL_symbolp(key) || ! LL_symbolp(keyval) ) {
		LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tkeyword/value is not a keyword\n\
\tkeyword: ~S\n\
\tvalue:   ~S\n\
", key, keyval);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    keys = LL_cdr(LL_cdr(keys));

	    if (LL_pname(key) == LLInternSymbol(":direction")) {
		/*
		 *    :direction
		 */
		if (directionP) {
		    LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tthe :direction has already been set\n\
\tkeyword: ~S\n\
\tvalue:   ~S\n\
", key, keyval);
		    LLThrow(LL_lightlisp, LL_nil);
		}

		if (LL_pname(keyval) == LLInternSymbol(":input")) {
		    inputP = TRUE;
		} else if (LL_pname(keyval) == LLInternSymbol(":output")) {
		    inputP = FALSE;
		} else {
		    LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tthat :direction option is not supported (:input or :output only)\n\
\tkeyword: ~S\n\
\tvalue:   ~S\n\
", key, keyval);
		    LLThrow(LL_lightlisp, LL_nil);
		}

		directionP = TRUE;
	    } else {
		LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tkeyword is not supported\n\
\tkeyword: ~S\n\
\tvalue:   ~S\n\
", key, keyval);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	}

	if ( ! directionP ) {
	    LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tnot enough keywords specified to open the file\n\
\tpath:     ~S\n\
\tkeywords: ~S\n\
", path, keys);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (inputP) {
	    s = STREAMCreate(STREAM_filename, path->value.a_string, path->value.a_string,
		    STREAM_void, NULL, NULL);
	} else {
	    FILE *f;

	    if ((f = fopen(path->value.a_string, "w")) == NULL) {
		s = NULL;
	    } else {
		s = STREAMCreate(STREAM_void, NULL, NULL,
			STREAM_FILE, f, path->value.a_string);
	    }
	}

	if (s == NULL) {
	    LLErrorF("\
Function: (open path &keys :direction [ :input | :output ] )\n\
\tcould not the file because ~A\n\
\tpath: ~S\n\
", sys_errlist[errno], path);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	return (LLMakeStream(s));
}
