#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "machine.h"
#include "internal.h"

/*
 *    LLAtom - atom
 *
 *    (atom exp)
 *
 *    return {t, nil}
 */
LLobject *
LLAtom(exp)
LLobject *exp;
{
	return (! LL_consp(exp) ? LL_t: LL_nil);
}

/*
 *    LLConsp - is a cons
 *
 *    (consp exp)
 *
 *    return {LL_t, LL_nil}
 */
LLobject *
LLConsp(exp)
LLobject *exp;
{
	return (LL_consp(exp) ? LL_t: LL_nil);
}

/*
 *    LLFloatp - floatp
 *
 *    (floatp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLFloatp(exp)
LLobject *exp;
{
	return (LL_realp(exp) ? LL_t: LL_nil);
}

/*
 *    LLFunctionp - functionp
 *
 *    (functionp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLFunctionp(exp)
LLobject *exp;
{
	if (LL_function_closurep(exp))
	    return (LL_t);
	if (LL_functionp(exp))
	    return (LL_t);
        if (LL_symbolp(exp))
	    return (LL_t);
	if (LL_lambdap(exp))
	    return (LL_t);

	return (LL_nil);
}

/*
 *    LLInputstreamp - input-stream-p
 *
 *    (input-stream-p exp)
 *
 *    return {t, nil}
 */
LLobject *
LLInputstreamp(exp)
LLobject *exp;
{
	return (LL_streamp(exp) &&
		exp->value.a_stream != NULL && 
		exp->value.a_stream->input.type != STREAM_void ?
		LL_t: LL_nil);
}

/*
 *    LLIntegerp - integerp
 *
 *    (integerp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLIntegerp(exp)
LLobject *exp;
{
	return (LL_integerp(exp) ? LL_t: LL_nil);
}

/*
 *    LLListp - is a list
 *
 *    (listp exp)
 *
 *    return {LL_t, LL_nil}
 */
LLobject *
LLListp(exp)
LLobject *exp;
{
	return (LL_listp(exp) ? LL_t: LL_nil);
}


/*
 *    LLNumberp - numberp
 *
 *    (numberp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLNumberp(exp)
LLobject *exp;
{
	return (LL_integerp(exp) || LL_realp(exp) ? LL_t: LL_nil);
}

/*
 *    LLOutputstreamp - output-stream-p
 *
 *    (output-stream-p exp)
 *
 *    return {t, nil}
 */
LLobject *
LLOutputstreamp(exp)
LLobject *exp;
{
	return (LL_streamp(exp) &&
		exp->value.a_stream != NULL &&
		exp->value.a_stream->output.type != STREAM_void ?
		LL_t: LL_nil);
}

/*
 *    LLStreamp - streamp
 *
 *    (streamp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLStreamp(exp)
LLobject *exp;
{
	return (LL_streamp(exp) ? LL_t: LL_nil);
}

/*
 *    LLStringp - stringp
 *
 *    (stringp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLStringp(exp)
LLobject *exp;
{
	return (LL_stringp(exp) ? LL_t: LL_nil);
}

/*
 *    LLSymbolp - symbolp
 *
 *    (symbolp exp)
 *
 *    return {t, nil}
 */
LLobject *
LLSymbolp(exp)
LLobject *exp;
{
	return (LL_symbolp(exp) ? LL_t: LL_nil);
}
