#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *print();

/*
 *    LLPrin1 - prin1
 *
 *    (prin1 object &optional stream)
 *
 *    return the object
 */
LLobject *
LLPrin1(object, stream)
LLobject *object;
LLobject *stream;	/* &optional */
{
	return (print(object, stream, "prin1",
		LL_type_of(object)->prin1[0],
		/* newlineP */ FALSE, /* spaceP */ FALSE));
}

/*
 *    LLPrinc - princ
 *
 *    (princ object &optional stream)
 *
 *    return the object
 */
LLobject *
LLPrinc(object, stream)
LLobject *object;
LLobject *stream;	/* &optional */
{
	return (print(object, stream, "princ",
		LL_type_of(object)->princ[0],
		/* newlineP */ FALSE, /* spaceP */ FALSE));
}

/*
 *    LLPrint - print
 *
 *    (print object &optional stream)
 *
 *    return the object
 */
LLobject *
LLPrint(object, stream)
LLobject *object;
LLobject *stream;	/* &optional */
{
	return (print(object, stream, "print",
		LL_type_of(object)->print[0],
		/* newlineP */ TRUE, /*  spaceP */ TRUE));
}

/*
 *    LLPprint - pprint
 *
 *    (pprint object &optional stream)
 *
 *    return the object
 */
LLobject *
LLPprint(object, stream)
LLobject *object;
LLobject *stream;	/* &optional */
{
	return (print(object, stream, "pprint",
		LL_type_of(object)->pprint[0],
		/* newlineP */ TRUE, /* spaceP */ FALSE));
}

/*
 *    void
 *    printer(object, stream)
 *    LLobject *object;
 *    LLobject *stream;	- required
 */
static LLobject *
print(object, stream, name, printer, newlineP, spaceP)
LLobject *object;
LLobject *stream;	/* &optional */
String name;
void (*printer)();
Boolean newlineP;
Boolean spaceP;
{
	if (LL_null(stream))
	    stream = LLGetValue(LLMakeSymbol("*standard-output*"), LL_valueSlot, LL_environment);

	if ( ! LL_streamp(stream) ) {
	    LLErrorF("\
Function: (~A object &optional stream)\n\
\tstream is not a stream\n\
\tstream: ~S\n\
", name, stream);
	    LLThrow(LL_lightlisp, LL_nil);
	}
	if ( ! LLOutputstreamp(stream) ) {
	    LLErrorF("\
Function: (~A object &optional stream)\n\
\tstream is not an output-capable stream\n\
\tstream: ~S\n\
", name, stream);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (newlineP)
	    STREAMPutc(stream->value.a_stream, '\n');

	(*printer)(object, stream->value.a_stream);

	if (spaceP)
	    STREAMPutc(stream->value.a_stream, ' ');

	return (object);
}
