#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *progn_1_2();

/*
 *    LLProg1 - prog1
 *
 *    (prog1
 *     form
 *     forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return result
 */
LLobject *
LLProg1(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLProg1_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLProg1_tr - prog1
 *
 *    (prog1 
 *     form
 *     forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive)
 */
LLobject *
LLProg1_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (progn_1_2(arguments, environment, 1, "prog1"));
}

/*
 *    LLProg2 - prog2
 *
 *    (prog2
 *     form
 *     form
 *     forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return result
 */
LLobject *
LLProg2(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLProg2_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLProg2_tr - prog2
 *
 *    (prog2 
 *     form
 *     form
 *     forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive)
 */
LLobject *
LLProg2_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (progn_1_2(arguments, environment, 1, "prog2"));
}

/*
 *    LLProgn - progn
 *
 *    (progn 
 *     forms ... )
 *
 *    return result
 */
LLobject *
LLProgn(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLProgn_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLProgn_tr - progn
 *
 *    (progn 
 *     forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive)
 */
LLobject *
LLProgn_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (progn_1_2(arguments, environment, 0, "progn"));
}

/*
 *    return {t, nil} on {tail-recursive, non-tail-recursive)
 */
static LLobject *
progn_1_2(arguments, environment, number, name)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
int number;
String name;
{
	LLobject *sexps;

	if (number != 0 && (LL_consp(*arguments) ? LLGetLength(*arguments): 0) < number) {
	    LLErrorF("\
Form: (~A form forms ... )\n\
\tinvalid number of arguments to ~A\n\
\tform: ~S\n\
", name, name, *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	sexps = *arguments;

	if ( ! LL_consp(sexps) ) {
	    /* this is what cl does if sexps is not a cons */
	    *arguments = LL_nil;
	    return (LL_nil);
	}

	switch (number) {
	case 0:
	    break;
	case 1:
	    *arguments = LLEvaluator(LL_car(sexps), *environment);
	    sexps = LL_cdr(sexps);
	    break;
	case 2:
	    LLEvaluator(LL_car(sexps), *environment);
	    sexps = LL_cdr(sexps);
	    *arguments = LLEvaluator(LL_car(sexps), *environment);
	    sexps = LL_cdr(sexps);
	    break;
	}

	if ( ! LL_consp(sexps) ) {
	    /* this is what cl does if sexps is not a cons */
	    *arguments = LL_nil;
	    return (LL_nil);
	}

	while ( ! LL_null(LL_cdr(sexps)) ) {
	    LLEvaluator(LL_car(sexps), *environment);
	    sexps = LL_cdr(sexps);
	}

	if (number == 0)	/* else *arguments is already set */
	    *arguments = LL_car(sexps);
	return (LL_t);
}
