#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "reader.h"
#include "internal.h"

/*
 *    LLReadfromstring - read-from-string
 *
 *    (read-from-string string &optional eof-error-p eof-value)
 *
 *    return {result, eof-value} on {success, eof}
 */
LLobject *
LLReadfromstring(string, eof_error_p, eof_value)
LLobject *string;
LLobject *eof_error_p;
LLobject *eof_value;
{
	LLobject *sexp, *promise;
	STREAM *stream;

	if ( ! LL_stringp(string) ) {
	    LLErrorF("\
Function: (read-from-string string &optional eof-error-p eof-value)\n\
\tstring is not a string\n\
\tstring:      ~S\n\
\teof-error-p: ~S\n\
\teof-value:   ~S\n\
", string, eof_error_p, eof_value);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	stream = STREAMCreate(STREAM_string, string->value.a_string, "string",
		STREAM_STREAM, LLErrorStream, NULL);
	if (stream == NULL) {
	    LLErrorF("\
Function: (read-from-string string &optional eof-error-p eof-value)\n\
\tcould not read the string\n\
\tstring:      ~S\n\
\teof-error-p: ~S\n\
\teof-value:   ~S\n\
", string, eof_error_p, eof_value);
	    return (LL_nil);	/* No, don't throw */
	}

	promise = LLCons(LLMakeSymbol("close"), LLCons(LLMakeStream(stream), LL_nil));
	promise = LLCons(promise, LL_nil);

	/* not a top-level catch b/c of the promise */
	if (LLCatchVoid(LL_uncatchable, promise, LL_environment))
	    return (LL_nil);

	sexp = LLReadValue(stream,
		LL_null(eof_error_p) ? FALSE: TRUE,
		eof_value,
		FALSE,
		LL_lightlisp);

	LLEndCatch();		/* evaluate the promise */

	return (sexp);
}
