#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLSetq - setq
 *
 *    (setq var form ... )
 *
 *    return result
 */
LLobject *
LLSetq(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLSetq_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLSetq_tr - setq
 *
 *    (setq var form ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLSetq_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *sexps, *inter, *ret, *symbol, *value;

	sexps = *arguments;

	ret = LL_nil;
	while ( ! LL_null(sexps) ) {
	    symbol = LL_first(sexps);
	    value = LL_second(sexps);
	    if ( ! LL_symbolp(symbol) ) {
		LLErrorF("\
Form: (setq var form ... )\n\
\tvar is not a symbol\n\
\tvar:   ~S\n\
\tvalue: ~S\n\
\tform:  ~S\n\
", symbol, value, *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    ret = LLSetValue(symbol, LL_valueSlot, 
		    LLEvaluator(value, *environment), *environment);

	    inter = LL_cdr(sexps);
	    sexps = LL_cdr(inter);

	    if (LL_null(inter)) {
		LLErrorF("\
Form: (setq var form ... )\n\
\tnot an even number of arguments to setq\n\
\tform: ~S\n\
", *arguments);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	}

	*arguments = ret;
	return (LL_nil);
}
