#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLUnwindprotect - unwind-protect
 *
 *    (unwind-protect protected
 *           forms ... )
 *
 *    return the result
 */
LLobject *
LLUnwindprotect(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	if (LL_null(LLUnwindprotect_tr(&arguments, &environment)))
	    return (arguments);
	
	return (LLEvaluator(arguments, environment));
}

/*
 *    LLUnwindprotect_tr - unwind-protect
 *
 *    (unwind-protect protected
 *           forms ... )
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLUnwindprotect_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *args;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: (unwind-protect protected forms ... )\n\
\tillegal number of arguments to unwind-protect\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	args = *arguments;

	/*
	 *    cleanup-code is considered an implicit progn
	 *    LLCatchVoid does a lexical closure over progn
	 *    for us.
	 *
	 *   Install an uncatchable catcher so that the cleanup
	 *   code will be fired.  The cleanup code is a promise of
	 *   the code in the implicit-progn part
	 */
	if (LLCatchVoid(LL_uncatchable, LL_rest(args), *environment))
	    return (LL_nil);	/* will never to through here */

	*arguments = LLEvaluator(LL_first(args), *environment);

	LLEndCatch();		/* evaluate the promise */
	/* ... to here */

	return (LL_nil);
}
