#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static LLobject *when_or_unless();

/*
 *    LLWhen - when
 *
 *    (when test forms ... )
 *
 *    return the result
 */
LLobject *
LLWhen(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* valuet */
{
	if (LL_null(LLWhen_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLWhen_tr - when
 *
 *    (when test forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLWhen_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (when_or_unless(arguments, environment, /* when */ TRUE));
}

/*
 *    LLUnless - unless
 *
 *    (unless test forms ... )
 *
 *    return the result
 */
LLobject *
LLUnless(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* valuet */
{
	if (LL_null(LLUnless_tr(&arguments, &environment)))
	    return (arguments);

	return (LLEvaluator(arguments, environment));
}

/*
 *    LLUnless_tr - unless
 *
 *    (unless test forms ... )
 *
 *    Every CommonLisp implementation has a right to implement
 *       a macro as a special form for speed
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLUnless_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	return (when_or_unless(arguments, environment, /* ! when */ FALSE));
}

/*
 *    WorkHorse ----------------------------------------
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
static LLobject *
when_or_unless(arguments, environment, whenP)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
Boolean whenP;
{
	LLobject *test;

	if (LLGetLength(*arguments) < 1) {
	    LLErrorF("\
Form: (~A test forms ... )\n\
\tinvalid number of arguments to ~A\n\
\tform: ~S\n\
", whenP ? "when": "unless", whenP ? "when": "unless", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	test = LLEvaluator(LL_first(*arguments), *environment);
	if (whenP ? LL_null(test): ! LL_null(test)) {
	    *arguments = LL_nil;
	    return (LL_nil);
	}

	*arguments = LL_rest(*arguments);
	return (LLProgn_tr(arguments, environment));
}
