#include "sumachine.h"

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ar.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"

#include "archive.h"
#include "internal.h"

/*
 *    ARCHAppendFile - append the file to the archive
 *
 *    return {ARCH_ok, ARCH_notok}
 */
ARCHstatus
ARCHAppendFile(arch, filename)
ARCHarchive *arch;
String filename;
{
	char buf[1024];
	ARCHfile file;
	int fd, cc;
	struct stat statbuf;
	String f;

	if ((fd = open(filename, O_RDONLY)) < 0)
	    return (ARCHErrno = ARCH_ESEE_ERRNO, ARCH_notok);

	fstat(fd, &statbuf);

	file.name = ARCHCopyFilename(filename);
	file.size = statbuf.st_size;
	file.offset = lseek(arch->work_fd, 0, L_XTND);

	if ((f = rindex(filename, '/')) != NULL)
	    filename = &f[1];

#if OPSYS_MFR == MD_SUN && CPU_TYPE == MD_I386 || \
	OPSYS_MFR == MD_APOLLO || \
	OPSYS_SUBTYPE == MD_AIX || \
        OPSYS_SUBTYPE == MD_HPUX || \
        OPSYS_SUBTYPE == MD_SYS5 /* { */
        /*
         *    Copy one less in order to insure that there is a padding
         *    character at the end of the filename.  This is MANDATORY
         *    for the ar format on these machines
         */
        strncpy(file.header.ar_name, filename, sizeof (file.header.ar_name)-1);
#else /* } else { */
        /*
         *    Can copy the full size on these machines b/c the ar format
         *    does not require a terminating character.
         */
        strncpy(file.header.ar_name, filename, sizeof (file.header.ar_name));
#endif /* } */
	sprintf(file.header.ar_date, "%d", statbuf.st_mtime);
	sprintf(file.header.ar_uid, "%d", statbuf.st_uid);
	sprintf(file.header.ar_gid, "%d", statbuf.st_gid);
	sprintf(file.header.ar_mode, "%o", statbuf.st_mode);
	sprintf(file.header.ar_size, "%d", statbuf.st_size);
	strncpy(file.header.ar_fmag, ARFMAG, sizeof (file.header.ar_fmag));

	ARCHBlankSlashPad(file.header.ar_name, sizeof (file.header.ar_name));
	ARCHBlankPad(file.header.ar_date, sizeof (file.header.ar_date));
	ARCHBlankPad(file.header.ar_uid, sizeof (file.header.ar_uid));
	ARCHBlankPad(file.header.ar_gid, sizeof (file.header.ar_gid));
	ARCHBlankPad(file.header.ar_mode, sizeof (file.header.ar_mode));
	ARCHBlankPad(file.header.ar_size, sizeof (file.header.ar_size));

	while ((cc = read(fd, buf, sizeof (buf))) > 0) {
	    if (write(arch->work_fd, buf, cc) < 0) {
		close(fd);
		return (ARCHErrno = ARCH_ESEE_ERRNO, ARCH_notok);
	    }
	}
	close(fd);

	if (cc != 0)
	    return (ARCHErrno = ARCH_ESEE_ERRNO, ARCH_notok);

	/* if the position is odd */
	if (lseek(arch->work_fd, 0, L_XTND) & 01) {
	    if (write(arch->work_fd, "\n", 1) < 0)
		return (ARCHErrno = ARCH_ESEE_ERRNO, ARCH_notok);
	}

	DYNpAtEndPut(&arch->files, &file);

	return (arch->flushP = TRUE, ARCH_ok);
}
