#if ! defined(archive_h)
#define archive_h

typedef enum ARCHstatus ARCHstatus;
enum ARCHstatus {
    ARCH_ok,
    ARCH_notok,
};

typedef enum ARCHmode ARCHmode;
enum ARCHmode {
    ARCH_readonly,
    ARCH_readwrite,
};

/*
 *    For move and replace (new files)
 */
typedef enum ARCHwhere ARCHwhere;
enum ARCHwhere {
    ARCH_begin,		/* move/replace at/to the begin */
    ARCH_end,		/* move/replace at/to the end */
    ARCH_here,		/* move/replace here */
    ARCH_before,	/* move/replace before here */
    ARCH_after,		/* move/replace after here */
};

typedef enum ARCHextract ARCHextract;
enum ARCHextract {
    ARCH_now,		/* use the modtime now */
    ARCH_then,		/* use the modtime then */
};

typedef struct ARCHfile ARCHfile;
struct ARCHfile {
    String name;		/* a '\0' terminated version of header.ar_name */
    time_t date;		/* integer version of header.ar_date */
    uid_t uid;			/* integer version of header.ar_uid */
    gid_t gid;			/* integer version of header.ar_gid */
    unsigned short mode;	/* integer version of header.ar_mode */
    size_t size;		/* integer version of header.ar_size */
    struct ar_hdr header;
    off_t offset;		/* offset to this header */
};

typedef struct ARCHarchive ARCHarchive;
struct ARCHarchive {
    int fd;
    ARCHmode mode;
    Boolean flushP;
    time_t modified;
    DYNgeneralArray files;	/* of ARCHfile */
    int work_fd;
};

#define ARCH_ENO_ERROR		0	/* no error */
#define ARCH_ESEE_ERRNO		1	/* see errno */
#define ARCH_EDOTA_SHORT	2	/* archive is short */
#define ARCH_ENOT_ARCHIVE	3	/* not an archive */
#define ARCH_EBAD_WRITE		4	/* could not write */
#define ARCH_EREAD_ONLY		5	/* archive is read only */
#define ARCH_ESHORT_TMP		6	/* short tmp file */
#define ARCH_ENO_MEM		7	/* not enough memory */
#define ARCH_EOUT_OF_RANGE	8	/* index is not in range */
#define ARCH_EUNKNOWN_PLACE	9	/* symbolic place is not valid */
#define ARCH_ENOT_A_DIRECTORY	10	/* not a directory */

extern int ARCHErrno;
extern String ARCHErrList[];
extern int ARCHNErrs;

extern ARCHarchive *ARCHOpen();
extern ARCHstatus ARCHWrite();
extern ARCHstatus ARCHClose();
extern ARCHstatus ARCHFree();		/* like close, but won't write */
extern ARCHstatus ARCHWriteTo();

extern ARCHstatus ARCHUseForTmp();

extern int ARCHSizeOfArchive();
extern ARCHfile *ARCHAt();	/* pointer not good across the below routines */

extern ARCHstatus ARCHDelete();
extern ARCHstatus ARCHExtract();
extern ARCHstatus ARCHExtractTo();
extern ARCHstatus ARCHExtractToBuf();
extern ARCHstatus ARCHExtractToFd();
extern ARCHstatus ARCHAppendFromArchive();
extern ARCHstatus ARCHAppendFile();
extern ARCHstatus ARCHReplaceFromArchive();
extern ARCHstatus ARCHReplaceFile();
extern ARCHstatus ARCHMoveTo();
extern ARCHstatus ARCHMoveToEnd();
extern ARCHstatus ARCHRenameFile();

#if OPSYS_MFR == MD_SUN && CPU_TYPE == MD_I386 /* { */
#if ! defined(PORTAR) /* { */
This should not compile

    We are assuming PORTAR (not PORT5AR or standard) for the sun 386i machine.
    You must have already defined this (in sumchine.h for instance)
    See conslibs for the ``ar ts'' that must be done to archive object files

#endif /* } */
#endif /* } */

#endif /* archive_h */
