#include <sys/param.h>
#include <sys/file.h>
#include <stdio.h>
#include <ar.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "archive.h"

#include "internal.h"

#if ! TEST_APOLLO_HACK && OPSYS_MFR != MD_APOLLO /* { */

typedef struct pair pair;
struct pair {
    String name;
    int index;
};
static pair *pv;

static int compare();
static void rename_members();

/*
 *    CNSLProcess - process a library on a normal Unix
 *
 *    A normal unix is any unix that is not an apollo
 *    If Apollo EVER says ``we are unix compatibile''
 *        then they are lying.
 *
 *    return {CNSL_ok, CNSL_notok} on {success, fail}
 */
CNSLstatus
CNSLProcess(argc, argv, metalibrary, appendP, localP, symdefP, uniqueP)
int argc;
Stringv argv;
String metalibrary;
Boolean appendP;
Boolean localP;
Boolean symdefP;
Boolean uniqueP;
{
	char buf[MAXPATHLEN];
	ARCHarchive *meta_a, *a;
	int i, f, nfiles, m, nmeta, om;

	if ( ! appendP && access(metalibrary, F_OK) == 0 && unlink(metalibrary) < 0) {
	    fprintf(stderr, "conslibs: cannot remove metalibrary %s because %s\n", 
		    metalibrary, sys_errlist[errno]);
	    return (CNSL_notok);
	}

	if (localP && ARCHUseForTmp(".") == ARCH_notok) {
	    fprintf(stderr, "conslibs: cannot use . for temporaries because %s\n",
		    ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
	    return (CNSL_notok);
	}

	if ((meta_a = ARCHOpen(metalibrary, ARCH_readwrite)) == NULL) {
	    fprintf(stderr, "conslibs: cannot open metalibrary %s because %s\n", metalibrary,
		    ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
	    return (CNSL_notok);
	}

	for (i=0; i<argc; i++) {
	    if ((a = ARCHOpen(argv[i], ARCH_readonly)) != NULL) {
		nfiles = ARCHSizeOfArchive(a);
		for (f=0; f<nfiles; f++) {
		    /*
		     *    Don't copy over symdef files
		     */
		    if ( ! symdefP ) {
#if OPSYS_MFR == MD_SUN && CPU_TYPE == MD_I386 || \
	OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
			/*
			 *    Symdef files on SysV-derived machines are
			 *    the first file in the archive - with no name
			 */
			if (strequal(ARCHAt(a, f)->name, ""))
			    continue;
#else /* } else { */
#if OPSYS_SUBTYPE == MD_ULTRIX && CPU_TYPE == MD_MIPS /* { */
			/*
			 *    Symdef files on DEC's Ultrix/MIPS are
			 *    for some reason named this. (erg!)
			 */
			if (strequal(ARCHAt(a, f)->name, "__________ELEL_"))
			    continue;
#else /* } else { */
			/*
			 *    Symdef files as you would expect
			 */
			if (strequal(ARCHAt(a, f)->name, "__.SYMDEF"))
			    continue;
#endif /* } */
#endif /* } */
		    }

		    /*
		     *    Append the archive
		     */
		    if (ARCHAppendFromArchive(meta_a, a, f) == ARCH_notok) {
			fprintf(stderr, "conslibs: cannot append %s to metalibrary %s because %s\n", 
				argv[i], metalibrary,
				ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
			break;
		    }
		}
		ARCHFree(a);
	    } else if (ARCHErrno == ARCH_ENOT_ARCHIVE) {
		if (ARCHAppendFile(meta_a, argv[i]) == ARCH_notok) {
		    fprintf(stderr, "conslibs: cannot append %s to metalibrary %s because %s\n", 
			    argv[i], metalibrary,
			    ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
		    continue;
		}
	    } else {
		fprintf(stderr, "conslibs: cannot open library %s because %s\n", argv[i],
			ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
		continue;
	    }
	}

	if (uniqueP) {
	    nmeta = ARCHSizeOfArchive(meta_a);
	    pv = ALLOC_ARRAY(nmeta, pair);
	    for (m=0; m<nmeta; m++) {
		pv[m].name = ARCHAt(meta_a, m)->name; pv[m].index = m;
	    }

	    qsort(pv, nmeta, sizeof (pair), compare);

	    for (om=0, m=1; m<nmeta; m++) {
		if (strequal(pv[m].name, pv[m-1].name))
		    continue;
		
		if (om+1 < m)
		    rename_members(meta_a, om, m-1, metalibrary);
		om = m;
	    }
	}

	if (om != nmeta-1)
	    rename_members(meta_a, om, nmeta-1, metalibrary);

	if (ARCHClose(meta_a) == ARCH_notok) {
	    fprintf(stderr, "conslibs: cannot close metalibrary %s because %s\n", metalibrary,
		    ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
	    return (CNSL_notok);
	}

	return (CNSL_ok);
}

static int
compare(p1, p2)
pair *p1;
pair *p2;
{
	return (strcmp(p1->name, p2->name));
}

static void
rename_members(a, low, high, metalibrary)
ARCHarchive *a;
int low;
int high;
String metalibrary;	/* for the error message */
{
	char buf[MAXPATHLEN];
	String name;
	int i, c;

	name = ARCHAt(a, pv[low].index)->name;

	for (c=1, i=low+1; i<=high; c++, i++) {
	    /*
	     *    don't use any char that is evaluable by /bin/sh  (such as '#')
	     *    lest the filename fails to go through ranlib which does
	     *    a ``system(sprintf(buf, "ar rllb %s %s %s", first, archive, "__.SYMDEF"))''
	     */
	    sprintf(buf, "%d:%s", c, name);	

	    if (ARCHRenameFile(a, pv[i].index, buf) == ARCH_notok) {
		fprintf(stderr, "conslibs: could not rename %s to %s in metalibrary %s because %s\n", 
			name, buf, metalibrary,
			ARCHErrno != ARCH_ESEE_ERRNO ? ARCHErrList[ARCHErrno]: sys_errlist[errno]);
	    }
	}
}
#endif /* } */
