#ifdef SOURCE_HEADER
#endif SOURCE_HEADER
#include <sys/param.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstr.h"
#include "filesys.h"

#include "filename.h"
#include "internal.h"

static DYNgeneralArray glob_pgsv;		/* of DYNStrgeneralString (pre globbed) */
static DYNgeneralArray glob_gsv;		/* of DYNStrgeneralString (globbed) */
static DYNgeneralArray glob_sv;			/* of String (DYNStrTheString(glbo_gsv[i]) */
static DYNgeneralArray words;			/* of String */

static int curlies();
static Boolean glob_neededP();
static int glob();

/*
 *    FNGlobFilename - glob the filename.
 *
 *    The word "glob" has been handed down from the ancients, and
 *    god only knows what it means.  It is use in contexts where
 *    tilde expansions, curlies {word,word,word}, and pattern matching
 *    via '*', '?' and [<set>].
 *
 *    return a {NULL terminated string of matches, NULL}
 */
Stringv
FNGlobFilename(unglobbed)
String unglobbed;
{
	Boolean backslashP, oneP, gooberP;
	DYNStrgeneralString *sp, *new, dummy;
	String path, opath, begin, tilde, hostname, burst, root;
	int i;
	register char pc;
	register String p;

	/*
	 *    DYNInitClear is not done because DYNpAtEndPut is
	 *    not used (nor can it be because SPgeneralSpace has
	 *    pointers associated with each member).  So the 
	 *    array is extended and new members DYNStrInitClear-ed.
	 */
	DYNInitClear(&glob_pgsv, DYNStrgeneralString);
	DYNInitClear(&glob_gsv, DYNStrgeneralString);
	DYNInitClear(&glob_sv, String);
	DYNInitClear(&words, String);

	/*
	 *    Handle the curlies  Make one string, and copy into
	 *    that.  If a curlie is seen, then make extra bufs with
	 *    copies of the first and append in parallel.
	 */
	DYNInsureSize(&glob_pgsv, 1);
	DYNSetCurrent(&glob_pgsv, 1);
	DYNStrInitClear(DYNpLast(&glob_pgsv, DYNStrgeneralString));

	if (curlies(unglobbed) < 0)
	    return (FNErrno = FN_EMISSING_LCURL, (Stringv) NULL);

	/*
	 *    Now tilde expand.
	 */
	for (i=0; i<DYNSizeOfEl(&glob_pgsv); i++) {
	    sp = DYNpAt(&glob_pgsv, i, DYNStrgeneralString);
	    path = DYNStrTheString(sp);

	    /*
	     *    If ~<thing> or word:~<thing>, then replace with
	     *    the tilde expanded path.
	     */
	    if (path[0] == '~') {
		if ((tilde = FNTildeExpand(path)) == NULL)
		    return (FNErrno = FN_EBAD_TILDE, (Stringv) NULL);
		DYNStrClear(sp);		/* reinstall the new ~expanded string */
		DYNStrAppendString(sp, tilde);
	    }
	}
	/* glob_pgsv is the vector to use */

	/*
	 *    And now glob expand.
	 *    For now, remote paths don't get glob expanded
	 */
	for (i=0; i<DYNSizeOfEl(&glob_pgsv); i++) {
	    path = DYNStrTheString( DYNpAt(&glob_pgsv, i, DYNStrgeneralString) );

	    if (glob_neededP(path)) {
		begin = path;

		/*
		 *    glob (copies over by itself)
		 */
		DYNClear(&words);

		/*
		 *    Must special case full paths; get rid of the first
		 *    / b/c you start there in the traversal (FI_preroot)
		 */
		gooberP = FALSE;
		root = path;		/* in case path == "/" */
		while (*path == '/')
		    path++;
		oneP = FALSE;

		while (*path != '\0') {
		    opath = path;
		    while (*path != '\0' && *path != '/')
			path++;
		    if (*path == '\0')		/* don't add the word if you are done */
			break;
		    *path++ = '\0';		/* terminate the previous string */

		    if (gooberP || glob_neededP(opath)) {
			if ( ! gooberP ) {
			    if (oneP)
				root = begin;
			    else
				root = *begin == '/' ? "/": ".";
			}
			gooberP = TRUE;
			DYNAtEndPut(&words, opath);
		    } else {
			*--path = '/';		/* terminate the previous string */
		    }
		    while (*path == '/')
			    path++;
		    oneP = TRUE;
		}

		if (glob(root, DYNTheArray(&words, String), DYNSizeOfEl(&words)) < 0)
		    return (NULL);	/* FNErrno set below */
	    } else {
		/*
		 *    Copy over manually
		 */
		DYNInsureIncrement(&glob_gsv, 1);
		DYNSetCurrent(&glob_gsv, DYNSizeOfEl(&glob_gsv) + 1);

		new = DYNpLast(&glob_gsv, DYNStrgeneralString);
		DYNStrInitClear(new);

		backslashP = FALSE;
		for (p=path; *p != '\0'; p++) {
		    pc = *p;
		    if (backslashP) {
			/* if in backslash mode, and a backslash, then 1 backslash */
			if (pc == '\\')
			    DYNStrAppendC(new, '\\');
			backslashP = pc == '\\' ? FALSE: TRUE;	/* change the mode */
		    } else {
			/* else,  in backslash mode, and a backslash, then 1 backslash */
			DYNStrAppendC(new, pc);
			backslashP = FALSE;
		    }
		}
	    }
	}
	/* glob_gsv is the vector to use */

	/*
	 *    Now make the string vector, and NULL terminate it.
	 */
	for (i=0; i<DYNSizeOfEl(&glob_gsv); i++)
	    DYNAtEndPut(&glob_sv, DYNStrTheString( DYNpAt(&glob_gsv, i, DYNStrgeneralString) ));
	DYNAtEndPut(&glob_sv, NULL);

	return (DYNTheArray(&glob_sv, String));
}

/*
 *    curlies - expand curlies (recursively too).
 *
 *    return {0, -1} on {success, fail}.
 */
static int
curlies(glob)
String glob;
{
	String g;

	for (g=glob; *g != '\0'; g++) {
	    if (*g != '{') {
		int i;

		for (i=0; i<DYNSizeOfEl(&glob_pgsv); i++)
		    DYNStrAppendC(DYNpAt(&glob_pgsv, i, DYNStrgeneralString), *g);
	    } else {	/* *g == '{' */
		int len, start_i;

		start_i = 0;
		len = DYNStrLength( DYNpAt(&glob_pgsv, start_i, DYNStrgeneralString) );
		if (curl_bursting(len, start_i, &g) < 0)
		    return (-1);
	    }
	}
	return (0);
}

/*
 *    curl_bursting - recursively handle curlies.
 *
 *    start at '{' and end on '}' (on success) adding the appropriate
 *    strings to the glob_spv.
 *
 *    return {0, -1} on {success, fail}.
 */
static int
curl_bursting(len, start_i, p_g)
int len;
int start_i;
String *p_g;
{
	for ( (*p_g)++; **p_g != '\0' && **p_g != '}'; (*p_g)++) {
	    if (**p_g == '{') {
		int len1;

		len1 = DYNStrLength( DYNpAt(&glob_pgsv, start_i, DYNStrgeneralString) );
		if (curl_bursting(len1, start_i, p_g) < 0)
		    return (-1);
	    } else if (**p_g != ',') {
		int i;

		for (i=start_i; i<DYNSizeOfEl(&glob_pgsv); i++)
		    DYNStrAppendC(DYNpAt(&glob_pgsv, i, DYNStrgeneralString), **p_g);
	    } else {		/* **p_g == ',' */
		DYNStrgeneralString *new, *orig;
		int before, big, i;

		before = DYNSizeOfEl(&glob_pgsv);
		big = (before - start_i);

		DYNInsureIncrement(&glob_pgsv, big);
		DYNSetCurrent(&glob_pgsv, before + big);

		for (i=0; i<before-start_i; i++) {
		    orig = DYNpAt(&glob_pgsv, start_i+i, DYNStrgeneralString);
		    new = DYNpAt(&glob_pgsv, before+i, DYNStrgeneralString);
		    DYNStrInitClear(new);
		    DYNStrAppendNChars(new, DYNStrTheString(orig), len);
		}
		start_i = before;
	    }
	}

	return (**p_g == '}' ? 0: -1);
}

/*
 *    glob_neededP - determine if globbingh must be done.
 *
 *    return {TRUE, FALSE} on {is, is not}.
 */
static Boolean
glob_neededP(path)
String path;
{
	Boolean backslashP;
	register char pc;
	register String p;

	backslashP = FALSE;
	for (p=path; *p != '\0'; p++) {
	    pc = *p;
	    if (pc == '\\') {
		/* if in backslash mode, then get out */
		backslashP = backslashP ? FALSE: TRUE;	
	    } else if ((pc == '*' || pc == '?' || pc == '[') && ! backslashP) {
		/* if not in backslash mode and you see a glob char, then glob */
		return (TRUE);
	    } else {
		/* else, you aren't in backslash mode any more */
		backslashP = FALSE;
	    }
	}

	return (FALSE);
}

static int thing();
static Boolean matchP();

static FIactions acts = {
    thing, 	/* preroot */
    NULL, 	/* postroot */
    thing, 	/* predirectory */
    thing, 	/* postdirectory */
    thing,	/* file */
    NULL,	/* onerror */
    FIDefaultDirectorySort, 	/* sort */
    matchP	/* selectP */
};
static Stringv s_patternv;
static int s_npatternv;
static Boolean errorP;

static int
glob(root, patternv, npatternv)
String root;
Stringv patternv;
int npatternv;
{
	int ret;

	s_patternv = patternv;
	s_npatternv = npatternv;
	errorP = FALSE;

	/*
	 *    root is to have either a '.' or a '/' in it
	 *    as the place to start the traversal. 
	 */
	if ((ret = FITraverseFileSystem(root, &acts)) < 0)
	    FNErrno = FN_ESEE_ERRNO;

	return (errorP ? -1: ret);
}

static int recursion;

static int
thing(tail, stat, path, type)
String tail;
struct stat *stat;
String path;
FIfunctionType type;
{
	DYNStrgeneralString *new;

	switch (type) {
	case FI_preroot:
	    recursion = 0;
	    return;
	case FI_predirectory:
	    recursion++;
	    return;		/* b/c recursion is one too many */
	case FI_postdirectory:
	    recursion--;
	    break;
	case FI_file:
	    break;
	}

	/*
	 *    If you have reached the depth of recursion, and you are called 
	 *    here (the file has been selected (e.g. matched)) then you have 
	 *    globbed correctly.  Select the file.
	 */
	if (recursion == s_npatternv-1) {
	    DYNInsureIncrement(&glob_gsv, 1);
	    DYNSetCurrent(&glob_gsv, DYNSizeOfEl(&glob_gsv) + 1);

	    new = DYNpLast(&glob_gsv, DYNStrgeneralString);
	    DYNStrInitClear(new);

	    DYNStrAppendString(new, path);
	}
	    
}

static Boolean _matchP();

static Boolean
matchP(tail, stat, path)
String tail;
struct stat *stat;
String path;
{
	if (recursion == s_npatternv)		/* too far to select */
	    return (FALSE);
	return (_matchP(s_patternv[recursion], tail));
}

static Boolean
_matchP(pattern, name)
String pattern;
String name;
{
	Boolean setP[256];
	Boolean exceptP;
	String ket, begin;
	char namec, patternc;

	begin = pattern;
	for ( ;; ) {
	    namec = *name++;
	    patternc = *pattern++;
	    if (patternc == '*') {
		/*
		 *     *
		 */
		if (&pattern[-1] == begin && namec == '.')
		    return (FALSE);		/* a leading '.' doesn't match a "*" */
		if (*pattern == '\0')		/* pattern is "*" only */
		    return (TRUE);
		/*
		 *    Handle pattern is "*word" by getting rid of the
		 *    * and seeing if any member of name will match
		 *    the remainder of the pattern
		 */
		for (name--; *name != '\0'; name++)
		    if (_matchP(pattern, name))
			return (TRUE);
		return (FALSE);
	    } else if (patternc == '?') {
		/*
		 *    ?
		 */
		if (namec == '\0')	/* if pattern says ? and name is done */
		    return (FALSE);
		continue;	/* else continue - this char is matched */
	    } else if (patternc == '[') {
		/*
		 *    [set] or [^set]
		 */
		if (*pattern == '^') {
		    exceptP = TRUE;
		    pattern++;
		} else {
		    exceptP = FALSE;
		}
		if ((ket = index(pattern, ']')) == NULL)
		    return (FNErrno = FN_EMISSING_LBRA, errorP = TRUE, FALSE);
		*ket = '\0';
		if (ExpandSet(setP, pattern) < 0)
		    return (FNErrno = FN_EBADSET, errorP = TRUE, FALSE);
		*ket = ']';
		/*
		 *    If its not in the set (or if [^set] if is in the set)
		 *    then this is bad, and it doesn't match
		 */
		if (exceptP ? ! setP[ (u_char) namec]: setP[ (u_char) namec])
		    return (FALSE);
		continue;
	    } else if (patternc == '\0') {
		/*
		 *    done
		 */
		return (namec == '\0' ? TRUE: FALSE);
	    } else {
		/*
		 *    regular char
		 */
		if (patternc != namec)
		    return (FALSE);
		continue;
	    }
	}
}
