#ifdef SOURCE_HEADER
#endif SOURCE_HEADER
#include <sys/types.h>
#include <stdio.h>
#include <pwd.h>

extern struct passwd *getpwuid();
extern struct passwd *getpwnam();

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstr.h"

#include "filename.h"
#include "internal.h"

/*
 *    FNTildeExpand - do tilde expansions into a static buffer.
 *
 *    <filename>
 *    ~/<filename>
 *    ~user/<filename>
 *    host:~user/<filename>		- unimplemented
 *
 *    The string returned is static.
 *
 *    return {re-expanded-string, NULL}
 */
String
FNTildeExpand(path)
String path;
{
	static DYNStrgeneralString space, hack;
	String hack_test;

	/*
	 *    Prevent some interesting global buffer problems by
	 *    copying to a temporary buffer.
	 */
	if (DYNStrInitP(&space)) {
	    hack_test = DYNStrTheString(&space);
	    if (path > hack_test && path < &hack_test[DYNStrLength(&space)]) {
		DYNStrInit(&hack);
		DYNStrAppendString(&hack, path);
		path = DYNStrTheString(&hack);
	    }
	}
	DYNStrInitClear(&space);

	/*
	 *    Note FNGlobFilename uses these conventions too.
	 */
	if (path[0] != '~') {
	    String colon;

	    /*
	     *    <filename>
	     */
	    DYNStrAppendString(&space, path);
	} else {
	    String appender,
		   home_directory;

	    if (path[1] == '/' || path[1] == '\0') {

		/*
		 *    ~/<filename>
		 */
		/* handle the case of just '~' being expanded */
		appender = path[1] != '\0' ? &path[1]: NULL;
		home_directory = FNFromTilde(NULL);;
	    } else {
		struct passwd *pw;
		String slash;

		/*
		 *    ~user/<filename>
		 */
		if ((slash = index(path, '/')) != NULL)
		    *slash = '\0';
		pw = getpwnam(&path[1]);	/* skip over ~ in ~user */
		if (slash != NULL)		/* fix path */
		    *slash = '/';
		if (pw == NULL) 	/* don't return till path is fixed */
		    return (FNErrno = FN_EBAD_TILDE, (String) NULL);
		/* handles the case of just '~user' being expanded */
		appender = slash;
		home_directory = pw->pw_dir;
	    }
	    DYNStrAppendString(&space, home_directory);
	    if (appender != NULL) 
		DYNStrAppendString(&space, appender);
	}

	/*
	 *    If you had to copy to a local buffer, then get rid of 
	 *    the global buffer.
	 */
	if (DYNStrInitP(&hack))
	    DYNStrFree(&hack);

	return (DYNStrTheString(&space));
}
