;;;;
;;;;    Run the Gabriel Benchmarks for LightLisp
;;;;

(defparameter *benchmarks*
  '("boyer.ll" "browse.ll" "ctak.ll" "dderiv.ll" "deriv.ll" "destruct.ll"
	       "div.ll fft.ll" "fprint.ll" "fread.ll" "frpoly.ll" "puzzle.ll"
	       "stak.ll" "tak.ll" "takl.ll" "takr.ll" "timer.ll" "tprint.ll" 
	       "traverse.ll" "triangle.ll")
  "the benchmarks themselves")

(defun run-benchmark (name torun)
  "to run the benchmark"
  (let (ret)
       (format t "running ~S~%" name)
       (setq ret (time (eval torun)))
       (format t "    returns ~S~%" ret)))

;;;
;;;    To fake what LightLisp doesn't have
;;;
(defconstant LightLisp t)
(when 'LightLisp
  (defvar *gabriel-property-lists* nil
	  "an a-list of p-lists")
  
  (defun get (x y)
	 (getf (symbol-plist x) y))
  
  (defun getf (place indicator &optional default)
	 (let* ((plist (assoc place *gabriel-property-lists*))
		(sub-plist (member indicator plist)))
	       (if (or (null plist) (null sub-plist))
		   default)
	       (second sub-plist)))
  
  (defmacro mapcar (function list &rest more-lists)
	    "like map, but the result is always a list"
	    `(map 'list ,function ,list ,@more-lists))
  
  (defmacro maplist (function list &rest more-lists)
	    "call the function on successive cdrs of the lists
	    damn hairy, damn hairy."
	    (let* ((lists (cons list more-lists))
		   (do-savevar (gensym))
		   (do-init (do* ((init `(,do-savevar nil))
				  (l lists (cdr l))
				  (var (gensym) (gensym))
				  (var-init (car l)))
				 ((null l) init)
				 (setq init (cons `(,var var-init (cdr var)) init))))
		   (do-end-test `((null ,(caadr do-init)) ,do-savevar))
		   (list-of-lists (gensym))
		   (save-cars (gensym))
		   (save-lists (gensym))
		   (first-of `(lambda (,list-of-lists)
				      (do ((,save-cars nil)
					   (,save-lists list-of-lists (cdr ,save-lists)))
					  ((null ,save-lists) ,save-cars)
					  (setq ,save-cars (caar ,save-lists))))))
		  `(do ,do-init ,do-end-test
		       (setq ,do-savevar (apply ,function (funcall ,first-of ,lists))))))
  
  (defmacro mapc (function list &rest more-lists)
	    "like mapcar but don't accumlate a result"
	    `(map nil ,function ,list ,@more-lists))
  
  (defmacro pop (place)
	    `(setf ,place (cdr ,place)))
  
  (defmacro push (item place)
	    `(setf ,place (cons ,item ,place)))
  
  (defmacro setf (place value)
	    (case (car place)
		  ((car first)
		   `(rplaca ,(second place) ,value))
		  (cdr
		   `(rplacd ,(second place) ,value))
		  (get
		   `(setf (getf ,@(cdr place)) ,value))
		  (getf 
		   `(let* ((the-place ,place)
			   (getf-place (second the-place))
			   (getf-indicator (third the-place))
			   (plist (assoc getf-place *gabriel-property-lists*))
			   (sub-plist (member getf-indicator plist)))
			  (cond ((null plist)
				 (acons getf-indicator (list getf-indicator ,value)
					*gabriel-property-lists*))
				((null sub-plist)
				 (rplacd plist (cons ,value nil)))
				(t
				 (rplaca (cdr sub-plist) ,value)))))
		  (otherwise
		   (format t "setf failure with ~S and ~S~%" place value)
		   (throw 'lightlisp nil))))
  
  (defun symbol-plist (place)
	 (assoc place *gabriel-property-lists*)))

;;;
;;;    load em up
;;;

(dolist (b *benchmarks*)
  (format t "Loading ~A ... ~%" b)
  (if (null (load (format nil "benchmarks/gabriel/~A" b)))
      (format t "Could not load ~A~%" b)))
