;;;;
;;;;    The sorting benchmark
;;;;
;;;;    Taken directly from rudell's Monumental lsort.h in C
;;;;

(defun sort (list compare &optional (length (length list)))
  "sort via a merge-sort - if length is not given, then it is calculated"
  (labels ((merge-sort (list compare count)
		       (if (<= count 1)
			   list
			   (let ((p list) list1 list2 i)
				;; break the list in half
				(do ((i (truncate (/ count 2)) (1- i)))
				    ((<= i 1))
				    (setq p (cdr p)))
				(setq list1 list
				      list2 (cdr p))
				(rplacd p nil)
				;; recursively sort the sub-lists
				;; (unless only 1 element)
				(if (> (setq i (truncate (/ count 2))) 1)
				    (setq list1 (merge-sort list1 compare i)))
				(if (> (setq i (- count i)) 1)
				    (setq list2 (merge-sort list2 compare i)))
				;; merge the two sorted sub-lists
				(do* ((end nil)
				      (ret (cons nil nil))
				      (last ret))
				     (end (cdr ret))
				     (if (funcall compare (car list1) (car list2))
					 (progn
					  (rplacd last list1)
					  (setq last (cdr last))
					  (when (null (setq list1 (cdr list1)))
						(rplacd last list2)
						(setq end t)))
					 (progn
					  (rplacd last list2)
					  (setq last (cdr last))
					  (when (null (setq list2 (cdr list2)))
						(rplacd last list1)
						(setq end t)))))))))
	  (merge-sort list compare length)))

(labels ((number-string (n &optional (max 10000))
			(if (= n 0) nil
			    (cons (random max) (number-string (1- n))))))
  (dolist (n '(0 5 10 50 100 500 1000))
	  (format t "Sort ~D~%" n)
	  (format t "    result is sorted? - ~S~%"
		  (let ((ret (time (sort (number-string n)  #'>)))>)
		       (if (< (length ret) 2)
			   t
			   (apply #'>= ret))))))
