#include <sys/time.h>	/* because we are dealing with struct tm et al */

#include "LightLispMachine.h"

/*
 *    cc gmtime.c -lLightLisp -lst
 */

/*
 *    lisp_gmtime - gmtime
 *
 *    (gmtime &optional seconds)
 *
 *    return a list of the gmtime
 */
static LLobject *
lisp_gmtime(seconds)
LLobject *seconds;		/* &optional */
{
	LLobject *ret;
	struct tm *tm;
	time_t now;

	if (LL_null(seconds)) {
	    time(&now);
	} else if (LL_integerp(seconds)) {
	    now = seconds->value.a_integer;
	} else {
	    LLErrorF("\
Function: (gmtime &optional seconds)\n\
\tseconds is not a integer\n\
\tseconds: ~S\n\
", seconds);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	tm = gmtime(&now);

	/*
	 *    Cons up (isdst yday wday year mon mday hour min sec)
	 */
	ret = LLCons(LLMakeInteger(tm->tm_isdst),
		    LLCons(LLMakeInteger(tm->tm_yday),
		    LLCons(LLMakeInteger(tm->tm_wday),
		    LLCons(LLMakeInteger(tm->tm_year),
		    LLCons(LLMakeInteger(tm->tm_mon),
		    LLCons(LLMakeInteger(tm->tm_mday),
		    LLCons(LLMakeInteger(tm->tm_hour),
		    LLCons(LLMakeInteger(tm->tm_min),
		    LLCons(LLMakeInteger(tm->tm_sec),
		    LL_nil)))))))));

	return (ret);
}

main(argc, argv)
int argc;
char **argv;
{
	int ecode;

	if (LLInitialize() != LL_ok) {
	    fprintf(stderr, "demo: could not initialize LightLisp\n");
	    exit (-1);
	}

	/*
	 *    Make a binding to gmtime in the functionSlot
	 *    of the function-descriptor of 0 requireds, 1 optionals
	 */
	LLSetValue(LLMakeSymbol("gmtime"), LL_functionSlot,
		LLMakeFunction("gmtime", lisp_gmtime,
			/* nrequired */ 0, 
			/* noptional */ 1, 
			/* &rest? */ FALSE),
		LL_environment);

	ecode = LLMain(argc, argv, 
		"demo", /* binary */
		"demo", /* package/prompt */
		NULL, /* libdir */
		NULL /* user_initialize */);

	if (LLUninitialize() != LL_ok) {
	    fprintf(stderr, "demo: could not uninitialize LightLisp\n");
	    exit (-1);
	}

	exit (ecode);
}
