;;;;
;;;;    Demonstrate the lexical closure things
;;;;

;;    in order to test
(defmacro print-eval (&body forms)
  "make a progn that will print out the execution of each form"
  (cons 'progn
	(let ((code nil) tmp)
	     (dolist (f forms code)
		     (setq tmp `(format t "--------------------~%~
					form: ~S~%    ~
					result: ~S~%~
					--------------------~%~%"
					',f ,f))
		     ; get it in the correct order (no matter how expensive)
		     (setq code (append code (cons tmp nil)))))))

(print-eval
  ;;
  ;;   First just let
  ;;
  (let ((list '((aa ab ac ad ae) b c d e f g h)))
       (list list list))
  ;;
  ;;   Then just let*
  ;;
  (let* ((list '((aa* ab* ac* ad* ae*) b* c* d* e* f* g* h*)))
	(list list list))
  ;;
  ;;   Then flet (in conjunction with let)
  ;;
  (let ((list '((faa fab fac fad fae) fb fc fd fe ff fg fh)))
       (flet ((list (&rest members)
		    (caar members)))
	     (list list list)))
  ;;
  ;;   Then a fancy flet (in conjunction with let)
  ;;        note that caar is used in the redefinition of caar
  ;;
  (let ((list '((faa1 fab1 fac1 fad1 fae1) fb1 fc1 fd1 fe1 ff1 fg1 fh1)))
       (flet ((list (&rest members)
		    (caar members))
	      (caar (list)
		    (list* (caar list) (third list) (fifth list) (cons (seventh list) nil))))
	     (list list list)))
  ;;
  ;;   Then a fancy flet (in conjunction with let) but eval at the top level
  ;;        with the top-level definition of list (the usual definition)
  ;;        note that caar is used in the redefinition of caar
  ;;
  (let ((list '((faa2 fab2 fac2 fad2 fae2) fb2 fc2 fd2 fe2 ff2 fg2 fh2)))
       (flet ((list (&rest members)
		    (caar members))
	      (caar (list)
		    (eval `(list ,(caar list) ,(third list) ,(fifth list) ,(seventh list)))))
	     (list list list)))
  ;;;
  ;;;   Now do it all over again, but this time with labels
  ;;;
  ;;
  ;;   Then labels (in conjunction with let)
  ;;
  (let ((list '((faa fab fac fad fae) fb fc fd fe ff fg fh)))
       (labels ((list (&rest members)
		      (caar members)))
	       (list list list)))
  ;;
  ;;   Then a fancy labels (in conjunction with let)
  ;;        note that caar is used (skillfully) in the redefinition of caar
  ;;
  (let ((list '((faa1 fab1 fac1 fad1 fae1) fb1 fc1 fd1 fe1 ff1 fg1 fh1)))
       (labels ((list (&rest members)
		      (caar members))
		(caar (list)
		      (if (null list)
			  nil
			  (list* (caar list) (third list) (fifth list) (cons (seventh list) nil)))))
	       (list list list)))
  ;;
  ;;   Then a fancy labels (in conjunction with let) but eval at the top level
  ;;        with the top-level definition of list (the usual definition)
  ;;        note that caar is used in the redefinition of caar
  ;;
  (let ((list '((faa2 fab2 fac2 fad2 fae2) fb2 fc2 fd2 fe2 ff2 fg2 fh2)))
       (labels ((list (&rest members)
		      (caar members))
		(caar (list)
		      (if (null list)
			  nil
			  (eval `(list ,(caar list) ,(third list) ,(fifth list) ,(seventh list))))))
	       (list list list))))
