#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

/*
 *    evaluate_args - evaluate the &rest args
 *
 *    LLobject *
 *    evaluator(sexp, environment)
 *    LLobject *sexp;
 *    LLobject *environment;
 *
 *    LLEvaluator will work.
 *
 *    return the args
 */
static LLobject *
evaluate_args(args, environment)
LLobject *args;
LLobject *environment;
{
	LLobject *ret, *last, *tmp;

	ret = LL_nil;
	last = LL_nil;
	for (/* args=inherit */; LL_consp(args); args=LL_cdr(args)) {
	    tmp = LLCons(LLEvaluator(LL_car(args), environment), LL_nil);
	    if (LL_null(ret))
		ret = tmp;
	    else
		LL_rplacd(last, tmp);
	    last = tmp;
	}

	return (ret);
}

/*
 *    LLApplyC - apply the LightLisp function to the C function
 *
 *    HACK it because of the way C is (be compiler independent)
 *
 *    LLobject *
 *    evaluator(sexp, environment)
 *    LLobject *sexp;
 *    LLobject *environment;
 *
 *    LLEvaluator will work.
 *
 *    return whatever the function returns
 */
LLobject *
LLApplyC(function, args, environment, evaluator)
LLobject *function;
LLobject *args;
LLobject *environment;
LLobject *(*evaluator)();
{
	LLobject *arguments[LL_MAXARGS], *original_args, *rest, *arg;
	int reqopt, arg_i, shit;

	original_args = args;
	reqopt = LL_function_nrequired(function) + LL_function_noptional(function);

	for (arg_i=0; arg_i<reqopt && LL_consp(args); arg_i++, args=LL_cdr(args)) {
	    arg = LL_car(args);
	    arguments[arg_i] = evaluator == NULL ? arg: (*evaluator)(arg, environment);
	}
	if (arg_i < LL_function_nrequired(function)) {
	    LLErrorF("\
Not enough arguments for function call\n\
\tfunction:  ~A of ~D required args ~D optional args~A\n\
\tcall:      ~S\n\
", LL_pname(function),
	LL_function_nrequired(function),
	LL_function_noptional(function),
	LL_function_restp(function) ? " &rest": "",
	original_args);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	for (/*arg_i=inherit*/; arg_i<reqopt; arg_i++)
	    arguments[arg_i] = LL_nil;

	if (LL_function_restp(function)) {
	    arguments[arg_i++] = LL_null(args) || evaluator == NULL?
		    args: evaluate_args(args, environment, evaluator);
	} else if (arg_i != reqopt || ! LL_null(args)) {
	    LLErrorF("\
Too many arguments for function call\n\
\tfunction:  ~A of ~D required args ~D optional args~A\n\
\tcall:      ~S\n\
", LL_pname(function),
	LL_function_nrequired(function),
	LL_function_noptional(function),
	LL_function_restp(function) ? " &rest": "",
	original_args);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	/*
	 *    The following egregious hack is because
	 *    C does not support varargs at all.
	 */
#include "applyc.i"
}
