#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static LLobject *expand();

/*
 *    LLBackquote - expand the template
 *
 *    `form
 *
 *    return the result
 */
LLobject *
LLBackquote(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLBackquote_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLBackquote_tr - expand the template
 *
 *    `form
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLBackquote_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *expanded;

	if (LLGetLength(*arguments) != 1) {
	    LLErrorF("\
Function: `form\n\
\tinvalid number of arguments to backquote\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (LLCatchVoid(LL_lightlisp, LL_nil, LL_environment)) {
	    /*
	     *    If an error,  print out the
	     *    context of the failure and re-throw
	     */
	    /* It is never a good thing to print environment */
	    LLErrorF("\
\tfull form:   ~S\n\
", LL_first(*arguments));
	    LLThrow(LL_lightlisp, LL_nil);	/* re-throw */
	}

	expanded = expand(LL_first(*arguments), *environment);
	LLEndCatch();

	*arguments = expanded;
	return (LL_nil);
}

static LLobject *
expand(arg, env)
LLobject *arg;
LLobject *env;
{
	LLobject *first, *second, *ret, *evaled, *expanded, *comma, *atsign;

	if ( ! LL_consp(arg) ) {
	    /*
	     *   form
	     */
	    return (arg);
	}

	if (LL_commap(arg)) {
	    /*
	     *    ,form
	     *
	     *    (comma form)
	     */
	    if (LLGetLength(arg) != 2) {
		LLErrorF("\
Function: ,form\n\
\tinvalid number of args to ,\n\
\tform:       ~S\n\
", arg);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    second = LL_second(arg);
	    if (LL_atsignp(second)) {
		LLErrorF("\
Function: ,form\n\
\timproper occurrence of @ - not inside a list\n\
\tform:       ~S\n\
", arg);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    return (LLEvaluator(second, env));
	}

	comma = LL_first(arg);
	if (LL_commap(comma)) {
	    /*
	     *    ,form
	     *
	     *    (comma form)
	     */
	    if (LLGetLength(comma) != 2) {
		LLErrorF("\
Function: ,form\n\
\tinvalid number of args to ,\n\
\tform:       ~S\n\
", arg);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    atsign = LL_second(comma);
	    if (LL_atsignp(atsign)) {
		/*
		 *    ,@form
		 *
		 *    (comma (atsign form))
		 */
		if (LLGetLength(atsign) != 2) {
		    LLErrorF("\
Function: ,@form\n\
\tinvalid number of args to @\n\
\tform:       ~S\n\
", arg);
		    LLThrow(LL_lightlisp, LL_nil);
		}

		evaled = LLEvaluator(LL_second(atsign), env);
		if ( ! LL_listp(evaled) ) {
		    LLErrorF("\
Function: ,@form\n\
\timproper result of ,@ - does not result in a list\n\
\tform:       ~S\n\
", arg);
		    LLThrow(LL_lightlisp, LL_nil);
		}

		/* (first arg) is what derived comma, and atsign */
		expanded = expand(LL_rest(arg), env);

		if (LL_null(evaled)) {
		    ret = expanded;
		} else { /* LL_consp(evaled) */
		    /* (nconc &rest lists) */
		    ret = LLNconc(LLCons(evaled, LLCons(expanded, LL_nil)));
		}
	    } else {
		/*
		 *    ,form 
		 *
		 *    (comma form)
		 */
		evaled = LLEvaluator(LL_second(comma), env);
		expanded = expand(LL_rest(arg), env);
		ret = LLCons(evaled, expanded);
	    }
	} else {
	    /*
	     *    (form ... )
	     */
	    ret = LLCons(expand(LL_first(arg), env), expand(LL_rest(arg), env));
	}
	return (ret);
}
