#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

/*
 *    LLEvaluator - the evaluator
 *
 *    return the result
 */
LLobject *
LLEvaluator(sexp, environment)
LLobject *sexp;
LLobject *environment;
{
	LLobject *function, *mac_env;

loop:	/* to handle tail recursion nicely */

	if (LL_null(sexp))
	    return (sexp);

	if (LL_symbolp(sexp)) {
	    /*
	     *    symbol
	     */
	    return (LLGetValue(sexp, LL_valueSlot, environment));
	} else if (LL_consp(sexp)) {
	    /*
	     *    cons
	     */
	    function = LL_first(sexp);
	subloop:
	    if (LL_symbolp(function)) {
		/*
		 *    (symbol ... )
		 */
		function = LLGetValue(function, LL_functionSlot, environment);
		goto subloop;
	    } else if (LL_lambdap(function)) {
		/*
		 *    ((lambda (formals ... ) [ documentation ] body ) actuals ... )
		 */
		function = LLFunctionClosure(LL_cdr(function), environment);
		goto subloop;
	    } else if (LL_closurep(function)) {
		/*
		 *    (closure ... )
		 */
		if (LL_function_closurep(function)) {
		    environment = LLLambdaArgs(
			    LL_first(LL_code(function)), LL_env(function),
			    LL_rest(sexp), environment,
			    /* evaluator */ LLEvaluator, /* destructuringP */ FALSE);

		    sexp = LL_rest(LL_code(function));

		    /*
		     *   Walk over the optional documentation string
		     *   See also LLApply and LLFuncall
		     */
		    if (LL_stringp(LL_first(sexp)) && ! LL_null(LL_cdr(sexp)))
			sexp = LL_cdr(sexp);

		    if (LL_null(LLProgn_tr(&sexp, &environment)))
			return (sexp);

		    /* (expanded) sexp, environment */
		} else {	/* if (LL_macro_closurep(function)) */
		    mac_env = LLLambdaArgs(
			    LL_first(LL_code(function)), LL_env(function),
			    LL_rest(sexp), LL_environment,	/* global env */
			    /* evaluator */ NULL, /* destructuringP */ TRUE);

		    sexp = LL_rest(LL_code(function));

		    /*
		     *   Walk over the optional documentation string
		     */
		    if (LL_stringp(LL_first(sexp)) && ! LL_null(LL_cdr(sexp)))
			sexp = LL_cdr(sexp);

		    if ( ! LL_null(LLProgn_tr(&sexp, &mac_env)) )
			sexp = LLEvaluator(sexp, mac_env);

		    /* (expanded) sexp, environment */
		}
		goto loop;	/* tail-recur */
	    } else if (LL_special_formp(function)) {
		/*
		 *    (function :address-of args :address-of environment)
		 *
		 *    returns
		 *      nil -> s-exp is the result
		 *      t -> evaluate s-exp/environment again
		 */
		sexp = LL_cdr(sexp);	/* get rid of the ``function'' slot */
		if (LL_null(LL_function_call(function)(&sexp, &environment)))
		    return (sexp);

		goto loop;	/* tail-recur */
	    } else if (LL_functionp(function)) {
		/*
		 *    (function)
		 *    (function arg)
		 *    (function arg arg)
		 *    (function arg arg ... )
		 */

		/* LLApplyC does argschecking */
		return (LLApplyC(function, LL_cdr(sexp), environment, LLEvaluator));
	    } else if (LL_macrop(function)) {
		/*
		 *    (expander args environment)
		 *
		 *    expand in the global environment
		 */
		sexp = LL_function_call(function)(LL_cdr(sexp), LL_environment);
		goto loop;
	    } else {
		LLErrorF("\
Function slot is not a symbol, function or closure\n\
\tsexp: ~S\n\
", sexp);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    /*NOTREACHED*/
	}
	/* } else { */

	/*
	 *    ... else ...
	 */
	return (sexp);
}
