#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

/*
 *    LLFunction - function
 *
 *    (function sexp)
 *
 *    return the result
 */
LLobject *
LLFunction(arguments, environment)
LLobject *arguments;		/* value */
LLobject *environment;		/* value */
{
	LLFunction_tr(&arguments, &environment);
	return (arguments);
}

/*
 *    LLFunction_tr - function
 *
 *    (function sexp)
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLFunction_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* value/return */
{
	LLobject *sexp;

	if (LLGetLength(*arguments) != 1) {
	    LLErrorF("\
Function: (function sexp)\n\
\tillegal number of arguments to function\n\
\tform: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	sexp = LL_first(*arguments);

	if (LL_symbolp(sexp)) {
	    sexp = LLGetValue(sexp, LL_functionSlot, *environment);
	    if ( ! LL_functionp(sexp) && ! LL_function_closurep(sexp) ) {
		LLErrorF("\
Function: (function sexp)\n\
\tinappropriate result for function - must be function or function-closure\n\
\tresult: ~S\n\
", sexp);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    *arguments = sexp;
	    return (LL_nil);
	}

	if (LL_function_closurep(sexp)) {
	    *arguments = sexp;
	    return (LL_nil);
	}

	if (LL_functionp(sexp)) {
	    *arguments = LLFunctionClosure(sexp, *environment);
	    return (LL_nil);
	}

	if (LL_lambdap(sexp)) {
	    /*
	     *    (lambda ( symbol ... ) [ documentation ] body ... )
	     */
	    *arguments = LLFunctionClosure(LL_cdr(sexp), *environment);
	    return (LL_nil);
	}

	LLErrorF("\
Function: (function sexp)\n\
\tinvalid type of argument - must be a symbol, function, function-closure or a lambda form\n\
\tform: ~S\n\
", sexp);
	LLThrow(LL_lightlisp, LL_nil);
	/*NOTREACHED*/
}
