#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static LLobject *relocate();
static void scan_registers();
static void scan_newhalf();

/*
 *    LLGarbageCollect - do garbage collection
 *
 *    return nothing, just do it.
 */
void
LLGarbageCollect()
{
	LLtime start, done;
	st_generator *g;
	LLstring s;
	LLobjectType *t;

	start = LLStopWatch();

	if (LLHeap.this == 0) {
	    LLHeap.this = 1;
	    LLHeap.that = 0;
	} else {
	    LLHeap.this = 0;
	    LLHeap.that = 1;
	}

	LLHeap.halves[LLHeap.this].used = 0;
	LLHeap.halves[LLHeap.this].pointer.block = LLHeap.halves[LLHeap.this].top;
	LLHeap.halves[LLHeap.this].pointer.index = 0;

	g = st_init_gen(LLMachine.types);
	while (st_gen(g, &s, &t))
	    t->count = 0;
	st_free_gen(g);

	scan_registers();
	scan_newhalf();

	done = LLStopWatch();
	LLHeap.gctime += (done - start);
	LLHeap.gc++;
}

static void relocatev();

static void
scan_registers()
{
	st_generator *g;
	LLstring s;
	int i, h;
	LLobjectType *t;
	LLobject *v, *o;
	LLcatcher *c;
	LLxregister *x;
	STREAM *S;

	/*
	 *    do the oblist first b/c creating identifiers uses the oblist
	 *    manuall relocate any register you wish to know the location of
	 */

	st_free_table(LLMachine.integers);
	LLMachine.integers = st_init_table(st_numcmp, st_numhash);

	for (i=0; i<DYNStkSp(&LLMachine.catchers); i++) {
	    c = DYNStkpNth(&LLMachine.catchers, i, LLcatcher);
	    c->tag = relocate(c->tag);
	    c->cleanup = relocate(c->cleanup);
	}

	g = st_init_gen(LLMachine.types);
	while (st_gen(g, &s, &t))
	    t->object = relocate(t->object);
	st_free_gen(g);

	g = st_init_gen(LLMachine.oblist);
	while (st_gen(g, &s, &o))
	    st_insert(LLMachine.oblist, s, relocate(o));
	st_free_gen(g);

	g = st_init_gen(LLMachine.handles);
	while (st_gen(g, &h, &v))
	    st_insert(LLMachine.handles, h, relocate(v));
	st_free_gen(g);

	g = st_init_gen(LLMachine.streams);
	while (st_gen(g, &S, &v))
	    st_insert(LLMachine.streams, S, relocate(v));
	st_free_gen(g);

	for (i=0; i<DYNSizeOfEl(&LLMachine.externals); i++) {
	    x = DYNpAt(&LLMachine.externals, i, LLxregister);
	    (*x->scan)(x->userData, relocate);
	}

	for (i=0; i<sizeof_el(LLMachine.readtable); i++)
	    LLMachine.readtable[i] = relocate(LLMachine.readtable[i]);
	for (i=0; i<sizeof_el(LLMachine.dispatchreadtable); i++)
	    LLMachine.dispatchreadtable[i] = relocate(LLMachine.dispatchreadtable[i]);

	LLMachine.alwayscatchable = relocate(LLMachine.alwayscatchable);
	LLMachine.atsign = relocate(LLMachine.atsign);
	LLMachine.backquote = relocate(LLMachine.backquote);
	LLMachine.comma = relocate(LLMachine.comma);
	LLMachine.function = relocate(LLMachine.function);
	LLMachine.environment = relocate(LLMachine.environment);
	LLMachine.eofvalue = relocate(LLMachine.eofvalue);
	LLMachine.lightlisp = relocate(LLMachine.lightlisp);
	LLMachine.lambda = relocate(LLMachine.lambda);
	LLMachine.nil = relocate(LLMachine.nil);
	LLMachine.t = relocate(LLMachine.t);
	LLMachine.quote = relocate(LLMachine.quote);
	LLMachine.unbound = relocate(LLMachine.unbound);
	LLMachine.uncatchable = relocate(LLMachine.uncatchable);
}

static void
scan_newhalf()
{
	LLblock *b;
	LLobject *o;
	int i, limit;
	Boolean pastP;

	for (b=LLHeap.halves[LLHeap.this].top, pastP=FALSE; ! pastP; b=b->next) {
	    pastP = b == LLHeap.halves[LLHeap.this].pointer.block ? TRUE: FALSE;
	    limit = pastP ? LLHeap.halves[LLHeap.this].pointer.index: sizeof_el (b->subheap);

	    for (i=0; i<limit; i++) {
		o = &b->subheap[i];

		if (o->type->garbage_collect.rescan == NULL)
		    continue;

		(*o->type->garbage_collect.rescan)(o, relocate);

		/* must evaluate this again because relocate() changed pointer.{block,index} */
		pastP = b == LLHeap.halves[LLHeap.this].pointer.block ? TRUE: FALSE;
		limit = pastP ? LLHeap.halves[LLHeap.this].pointer.index: sizeof_el (b->subheap);
	    }

	}
}

static LLobject *
relocate(old)
LLobject *old;
{
	LLobject *new;

	if (LL_null(old))
	    return (LL_nil);

	new = (*old->type->garbage_collect.scan)(old);

	if (old->type != LLGarbageCell) {
	    old->type = LLGarbageCell;
	    old->value.a_garbage.broken_heart = new;
	}

	return (new);
}
