#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

/*
 *    LLGetBlock - get a new block for both heaps
 *
 *    return nothing, just do it.
 */
void
LLGetBlock()
{
	LLblock *b[2];

	if (LLMaxBlocks > 0 && LLHeap.nblocks == LLMaxBlocks) {
	    LLErrorF("\
LightLisp: hit a user-specifiable system limit of ~D heap blocks (~D objects)\n\
", LLHeap.nblocks, LLHeap.nblocks * sizeof_el (b[0]->subheap));
	    LLPrintFullStats(LLErrorStream);
	    /* throwing to nil which cannot be caught - forcing a top-level gc */
	    LLGarbageCollect();
	    LLThrow(LL_uncatchable, LL_nil);	/* not to LL_lightlisp */
	}

	b[0] = ALLOC_TYPE(LLblock);
	b[1] = ALLOC_TYPE(LLblock);

	if (b[0] == NULL || b[1] == NULL) {
	    /* TODO - do something nice here... */
	    LLErrorF("\
LightLisp: hit a runtime-dependent system limit of ~D heap blocks (~D objects)\n\
", LLHeap.nblocks, LLHeap.nblocks * sizeof_el (b[0]->subheap));
	    LLPrintFullStats(LLErrorStream);
	    /* throwing to that which cannot be caught - forcing a top-level gc */
	    LLGarbageCollect();
	    LLThrow(LL_uncatchable, LL_nil);	/* not to LL_lightlisp */
	}

	b[0]->next = NULL;
	*LLHeap.halves[0].tail = b[0];
	LLHeap.halves[0].tail = &b[0]->next;
	LLHeap.halves[0].total += sizeof_el (b[0]->subheap);

	b[1]->next = NULL;
	*LLHeap.halves[1].tail = b[1];
	LLHeap.halves[1].tail = &b[1]->next;
	LLHeap.halves[1].total += sizeof_el (b[1]->subheap);

	LLHeap.halves[LLHeap.this].pointer.block = b[LLHeap.this];
	LLHeap.halves[LLHeap.this].pointer.index = 0;

	LLHeap.nblocks++;
}
