#include <sys/types.h>
#include <setjmp.h>
#include <ctype.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static struct {
    LLobjectType **place;	/* place to put the objectType */
    LLobjectType template;
} standard_templates[] = {
    { &LLNilCell,
	{ "nil",	"null",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ LL_Pr_Nil, LL_Pr_Nil, LL_Pr_Nil },
		{ LL_Pr_Nil, LL_Pr_Nil, LL_Pr_Nil },
		{ LL_Pr_Nil, LL_Pr_Nil, LL_Pr_Nil },
		{ LL_Pr_Nil, LL_Pr_Nil, LL_Pr_Nil },
		{ LL_GcScan_Nil,	NULL }, NULL } },
    { &LLIntegerCell,
	{ "integer",	"integer",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ LL_Pr_Integer, LL_Pr_Integer1, LL_Pr_Integer2 },
		{ LL_Pr_Integer, LL_Pr_Integer1, LL_Pr_Integer2 },
		{ LL_Pr_Integer, LL_Pr_Integer1, LL_Pr_Integer2 },
		{ LL_Pr_Integer, LL_Pr_Integer1, LL_Pr_Integer2 },
		{ LL_GcScan_Integer,	NULL }, NULL } },
    { &LLRealCell,
	{ "flonum",	"flonum",	NULL, 
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Real, LL_Pr_Real1, LL_Pr_Real2 },
		{ LL_Pr_Real, LL_Pr_Real1, LL_Pr_Real2 },
		{ LL_Pr_Real, LL_Pr_Real1, LL_Pr_Real2 },
		{ LL_Pr_Real, LL_Pr_Real1, LL_Pr_Real2 },
		{ LL_GcScan_Real,	NULL }, NULL } },
    { &LLStringCell,
	{ "string",	"string",	NULL,
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_String2, LL_Pr_String2, LL_Pr_String2 },
		{ LL_Pr_String, LL_Pr_String, LL_Pr_String },
		{ LL_Pr_String1, LL_Pr_String1, LL_Pr_String1 },
		{ LL_Pr_String2, LL_Pr_String2, LL_Pr_String2 },
		{ LL_GcScan_String,	NULL }, &LLStringAsASequence } },
    { &LLSymbolCell,
	{ "symbol",	"symbol",	NULL,
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Symbol, LL_Pr_Symbol, LL_Pr_Symbol },
		{ LL_Pr_Symbol, LL_Pr_Symbol, LL_Pr_Symbol },
		{ LL_Pr_Symbol, LL_Pr_Symbol, LL_Pr_Symbol },
		{ LL_Pr_Symbol, LL_Pr_Symbol, LL_Pr_Symbol },
		{ LL_GcScan_Symbol,	LL_GcReScan_Symbol }, NULL } },
    { &LLFunctionClosureCell,
	{ "function-closure",	"function",	NULL,
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_GcScan_Closure,	LL_GcReScan_Closure }, NULL } },
    { &LLMacroClosureCell,
	{ "macro-closure",	"macro",	NULL,
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		{ LL_Pr_Closure, LL_Pr_Closure, LL_Pr_Closure },
		  { LL_GcScan_Closure,	LL_GcReScan_Closure }, NULL } },
    { &LLConsCell,
	{ "cons",	"cons",	NULL, 
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Cons, LL_Pr_Cons, LL_Pr_Cons },
		{ LL_Pr_Cons, LL_Pr_Cons, LL_Pr_Cons },
		{ LL_Pr_Cons, LL_Pr_Cons, LL_Pr_Cons },
		{ LL_Pr_Cons1, LL_Pr_Cons1, LL_Pr_Cons1 },
		{ LL_GcScan_Cons,	LL_GcReScan_Cons }, &LLConsAsASequence } },
    { &LLMacroCell,
	{ "compiled-macro",	"macro",	NULL, 
		{ LL_MACRO }, 0, 
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_GcScan_Function,	NULL }, NULL } },
    { &LLSpecialFormCell,
	{ "special-form",	"special-form",	NULL, 
		{ LL_SPECIAL_FORM }, 0, 
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function, },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function, },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function, },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_GcScan_Function,	NULL }, NULL } },
    { &LLDistinguishedCell,
	{ "distinguished",	"distinguished",	NULL, 
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Distinguished, LL_Pr_Distinguished, LL_Pr_Distinguished },
		{ LL_Pr_Distinguished, LL_Pr_Distinguished, LL_Pr_Distinguished },
		{ LL_Pr_Distinguished, LL_Pr_Distinguished, LL_Pr_Distinguished },
		{ LL_Pr_Distinguished, LL_Pr_Distinguished, LL_Pr_Distinguished },
		{ LL_GcScan_Distinguished,	NULL }, NULL } },
    { &LLStreamCell,
	{ "stream",	"stream",	NULL, 
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Stream, LL_Pr_Stream, LL_Pr_Stream },
		{ LL_Pr_Stream, LL_Pr_Stream, LL_Pr_Stream },
		{ LL_Pr_Stream, LL_Pr_Stream, LL_Pr_Stream },
		{ LL_Pr_Stream1, LL_Pr_Stream1, LL_Pr_Stream1 },
		{ LL_GcScan_Stream,	NULL }, NULL } },
    { &LLDeadCell,
	{ "dead",	"dead",	NULL, 
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Dead, LL_Pr_Dead, LL_Pr_Dead },
		{ LL_Pr_Dead, LL_Pr_Dead, LL_Pr_Dead },
		{ LL_Pr_Dead, LL_Pr_Dead, LL_Pr_Dead },
		{ LL_Pr_Dead, LL_Pr_Dead, LL_Pr_Dead },
		{ LL_GcScan_Dead,	NULL }, NULL } },
    { &LLGarbageCell,
	{ "garbage",	"garbage",	NULL, 
		{ LL_NON_FUNCTION, 0x0 }, 0, 
		{ LL_Pr_Garbage, LL_Pr_Garbage, LL_Pr_Garbage },
		{ LL_Pr_Garbage, LL_Pr_Garbage, LL_Pr_Garbage },
		{ LL_Pr_Garbage, LL_Pr_Garbage, LL_Pr_Garbage },
		{ LL_Pr_Garbage, LL_Pr_Garbage, LL_Pr_Garbage },
		{ LL_GcScan_Garbage,	NULL }, NULL } },
};

/*
 *   Also in make-function.c there is a template for LLFunctionCells[][][]
 */

/*
 *    LLInitialize - initialize things
 *
 *    return {LL_ok, LL_notok} on {is, is not}.
 */
LLstatus
LL_Mach_Initialize()
{
	int i;

	LLSymbolTable = LLCaseSensitivityP ?
		st_init_table(strcmp, st_strhash):
		st_init_table(nocase_strcmp, nocase_strhash);
	LLStringTable = st_init_table(strcmp, st_strhash);

	LLLightLisp = LLInternSymbol(LLLightLisp);
	LLLambda = LLInternSymbol(LLLambda);
	LLAndOptional = LLInternSymbol(LLAndOptional);
	LLAndRest = LLInternSymbol(LLAndRest);
	LLAndBody = LLInternSymbol(LLAndBody);

	srandom(time(0l));

	/*
	 *    Pre-Initialize
	 */
	LLMachine.nil = NULL;		/* LL_nil */
	LLMachine.t = NULL;

	LLMaxBlocks = 0;		/* unlimited */
	LLHeap.this = 0;
	LLHeap.that = 1;

	LLHeap.halves[0].top = NULL;
	LLHeap.halves[0].tail = &LLHeap.halves[0].top;

	LLHeap.halves[1].top = NULL;
	LLHeap.halves[1].tail = &LLHeap.halves[1].top;

	LLHeap.nblocks = 0;
	LLHeap.nstringb = 0;
	LLHeap.gc = 0;
	LLHeap.gctime = 0;
	LLHeap.starttime = LLStopWatch();

	LLGetBlock();

	/*
	 *    Initialize a little bit
	 */
	DYNStkInit(&LLMachine.catchers, LLcatcher);
	LLMachine.exit = NULL;
	LLMachine.types = st_init_table(st_ptrcmp, st_ptrhash);
	LLMachine.sequences = st_init_table(st_ptrcmp, st_ptrhash);
	LLMachine.oblist = st_init_table(st_ptrcmp, st_ptrhash);
	LLMachine.handles = st_init_table(st_ptrcmp, st_ptrhash);
	/* if you st_init_table_with_params, see LLGarbageCollect */
	LLMachine.integers = st_init_table(st_numcmp, st_numhash);
	LLMachine.streams = st_init_table(st_ptrcmp, st_ptrhash);
	DYNInit(&LLMachine.externals, LLxregister);
	for (i=0; i<sizeof_el(LLMachine.readtable); i++)
	    LLMachine.readtable[i] = LL_nil;
	for (i=0; i<sizeof_el(LLMachine.dispatchreadtable); i++)
	    LLMachine.dispatchreadtable[i] = LL_nil;
	LLMachine.nil = NULL;

	if (LLTopLevelCatchVoid())
	    return (LL_notok);

	/*
	 *    Set up the types
	 *
	 *    MAKE SURE YOU MAKE TWO PASSES
	 *    the first one bootstraps the TypeCell
	 *    the second one sets up the Type system
	 *
	 *    Pass I
	 */
	LLTypeCell = ALLOC_TYPE(LLobjectType);
	LLTypeCell->tname = LLInternSymbol("type");
	LLTypeCell->pname = LLTypeCell->tname;
	LLTypeCell->object = LLMakeObject();
	LLTypeCell->object->type = LLTypeCell;
	LLTypeCell->object->value.a_type = LLTypeCell;
	LLTypeCell->count = 1;
	LLTypeCell->flags[0] = LL_NON_FUNCTION;
	for (i=0; i<sizeof_el (LLTypeCell->print); i++) {
	    LLTypeCell->prin1[i] = LL_Pr_Type;
	    LLTypeCell->princ[i] = LL_Pr_Type;
	    LLTypeCell->print[i] = LL_Pr_Type;
	    LLTypeCell->pprint[i] = LL_Pr_Type;
	}
	LLTypeCell->garbage_collect.scan = LL_GcScan_Type;
	LLTypeCell->garbage_collect.rescan = NULL;
	LLTypeCell->sequence = NULL;
	st_add_direct(LLMachine.types, LLTypeCell->tname, LLTypeCell);

	/*
	 *    Pass II
	 */
	for (i=0; i<sizeof_el(standard_templates); i++) {
	    *standard_templates[i].place = LLMakeObjectType(standard_templates[i].template);
	}
	/*
	 *    hack it so that (sequencep nil) -> t
	 *    Check out the corresponding hack in type_killer (uninit.c)
	 */
	LLNilCell->sequence = &LLConsAsASequence;

	/*
	 *    before ANY further calls 
	 */
	LLMachine.uncatchable = LLMakeDistinguished("uncatchable");	
	LLMachine.alwayscatchable = LLMakeDistinguished("alwayscatchable");
	LLMachine.unbound = LLMakeDistinguished("unbound");
	LLMachine.lightlisp = LLMakeSymbol("lightlisp");
	LLMachine.lambda = LLMakeSymbol("lambda");

	/*
	 *    Initialize the rest of the way
	 */
	LLMachine.atsign = LLMakeSpecialForm("@", LLAtsign_tr);
	LLMachine.backquote = LLMakeSpecialForm("`", LLBackquote_tr);
	LLMachine.comma = LLMakeSpecialForm(",", LLComma_tr);
	LLMachine.function = LLMakeSpecialForm("function", LLFunction_tr);
	LLMachine.eofvalue = LLMakeDistinguished("eofvalue");
	LLMachine.quote = LLMakeSpecialForm("quote", LLQuote_tr);
	LLMachine.t = LLMakeSymbol("t");

	LLSetValue(LLMakeSymbol("acons"), LL_functionSlot,
		LLMakeFunction("acons", LLAcons, 3, 0, FALSE), LL_nil);
	LLSetValue(LLMakeSymbol("assoc"), LL_functionSlot,
		LLMakeFunction("assoc", LLAssoc, 2, 0, FALSE), LL_nil);
	LLSetValue(LLMakeSymbol("cons"), LL_functionSlot,
		LLMakeFunction("cons", LLCons, 2, 0, FALSE), LL_nil);
	LLSetValue(LLMakeSymbol("nconc"), LL_functionSlot,
		LLMakeFunction("nconc", LLNconc, 2, 0, FALSE), LL_nil);
	LLSetValue(LLMakeSymbol("throw"), LL_functionSlot,
		LLMakeFunction("throw", LLThrow, 2, 0, FALSE), LL_nil);

	LLSetValue(LLMakeSymbol("t"), LL_valueSlot, LL_t, LL_nil);
	LLSetValue(LLMakeSymbol("lambda-list-keywords"), LL_valueSlot,
		LLCons(LLMakeSymbol(LLAndOptional),
			LLCons(LLMakeSymbol(LLAndRest),
			LLCons(LLMakeSymbol(LLAndBody), LL_nil))), LL_environment);

	LLSetValue(LLMakeSymbol("function"), LL_functionSlot, LL_function, LL_environment);
	LLSetValue(LLMakeSymbol("quote"), LL_functionSlot, LL_quote, LL_environment);

	LLMachine.dispatchreadtable['\''] = LL_function;
	LLMachine.readtable['\''] = LL_quote;
	LLMachine.readtable[','] = LL_comma;
	LLMachine.readtable['`'] = LL_backquote;
	LLMachine.readtable['@'] = LL_atsign;

	return (LL_ok);
}
