/*
 *    The symbol table
 */
extern Boolean LLCaseSensitivityP;

extern st_table *LLSymbolTable;	/* of (String, LLstring) pairs (case insensitive) */
extern st_table *LLStringTable;	/* of (String, String) pairs (case sensitive) */

/*
 *    The heap management stuff
 *
 *    WATCHOUT - assume here that you are working with a malloc
 *    that does power of two binning.   So you want to have as
 *    the following difference as small as possible
 *
 *    2 ^ k - (sizeof (LLblock) + <internal-malloc-header>)
 *
 *    Assume (conservatively) that you have a malloc with an 8 byte header
 */
#define LL_MALLOC_HEADER	8
#define LL_BLOCKSIZE		((/* 32K bytes */ 0x8000 - sizeof (LLblock *)-LL_MALLOC_HEADER) / sizeof (LLobject))

typedef struct LLblock LLblock;
struct LLblock {
    LLobject subheap[LL_BLOCKSIZE];
    LLblock *next;
};

typedef struct LLheap LLheap;
struct LLheap {
    int this;			/* if ``this'' is 1, the ``that'' is 0 */
    int that;			/* if ``that'' is 1, the ``this'' is 0 */
    struct {
	struct {
	    LLblock *block;		/* block with next available slot */
	    int index;			/* index of the next available slot */
	} pointer;
	int used;			/* number of used objects */
	int total;			/* number of total objects */
	LLblock *top;
	LLblock **tail;			/* &NULL so appending is quick */
    } halves[2];
    int nblocks;			/* nblocks used */
    int nstringb;			/* bytes of string storage */
    int gc;				/* number of gc's */
    LLtime gctime;			/* time in gc (in milliseconds) */
    LLtime starttime;			/* all time (in milliseconds) */
};

extern int LLMaxBlocks;		/* a handle so you can set it at runtime */
extern LLheap LLHeap;

extern void LLGetBlock();
extern LLtime LLStopWatch();

/*
 *   Various type things
 *    Don't export these to the world
 */
extern LLsequence LLConsAsASequence;
extern LLsequence LLStringAsASequence;

extern LLobjectType *LLDistinguishedCell;
extern LLobjectType *LLGarbageCell;

/*
 *    Various garbage collection things
 *    Don't export these to the world
 */
extern LLobject *LL_GcScan_Nil();
extern LLobject *LL_GcScan_Integer();
extern LLobject *LL_GcScan_Real();
extern LLobject *LL_GcScan_String();
extern LLobject *LL_GcScan_Symbol();
extern void LL_GcReScan_Symbol();
extern LLobject *LL_GcScan_Type();
extern LLobject *LL_GcScan_Closure();
extern void LL_GcReScan_Closure();
extern LLobject *LL_GcScan_Cons();
extern void LL_GcReScan_Cons();
extern LLobject *LL_GcScan_Function();
extern LLobject *LL_GcScan_Distinguished();
extern LLobject *LL_GcScan_Dead();
extern LLobject *LL_GcScan_Stream();
extern LLobject *LL_GcScan_Garbage();

/*
 *   Tail-recursive sensitive pecial forms
 */
extern LLobject *LLAtsign_tr();		/* special form */
extern LLobject *LLBackquote_tr();	/* special form */
extern LLobject *LLComma_tr();		/* special form */
