#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static LLobject *coerce();
static LLobject *concatenate();
static LLobject *copy_seq();
static LLobject *elt();
static LLobject *length();
static LLobject *subseq();

LLsequence LLConsAsASequence = {
    "list",
    NULL,	/* filled in later */
    coerce,
    concatenate,
    copy_seq,
    elt,
    length,
    subseq,
};

static LLobject *
coerce(sequence, result_type)
LLobject *sequence;
LLobject *result_type;		/* typeCell (with sequence) */
{
	if (LL_typep(sequence, LLConsCell) || LL_typep(sequence, LLNilCell)) {
	    return (sequence);
	} else if (LL_typep(sequence, LLStringCell)) {
	    LLobject *ret, *last, *tmp;
	    String s;
	    char chars[2];

	    ret = LL_nil;
	    last = LL_nil;
	    for (s=sequence->value.a_string; *s != '\0'; s++) {
		chars[0] = *s;
		chars[1] = '\0';

		tmp = LLCons(LLMakeString(chars), LL_nil);
		if (LL_null(ret))
		    ret = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;
	    }
	    return (ret);
	}

	LLErrorF("\
Function: (coerce sequence result-type)\n\
\tcannot coerce a ~A into a list\n\
\tsource: ~S\n\
", LL_type_of(sequence)->pname, sequence);
	LLThrow(LL_lightlisp, LL_nil);
	/*NOTREACHED*/
}

static LLobject *
concatenate(result_type, sequences)
LLobject *result_type;		/* typeCell */
LLobject *sequences;		/* &rest */
{
	LLobject *sequence, *ret, *last, *tmp, *r, *subret, *sublast, *subr;

	if (LL_null(sequences))
	    return (LL_nil);

	ret = LL_nil;
	last = LL_nil;

	/* this is not LLAppend because ALL arguments are copied */
	for (r=sequences; LL_consp(r); r=LL_cdr(r)) {
	    sequence = LL_car(r);

	    if (LL_null(sequence))
		continue;

	    if ( ! LL_consp(sequence) ) {
		LLErrorF("\
Function: (concatenate result-type &rest sequences)\n\
\tsequence is not a list\n\
\tsequence:  ~S\n\
\tsequences: ~S\n\
", sequence, sequences);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    /*
	     *    => (concatenate 'list '(a b) nil)
	     *    (a b)
	     *
	     *    => (concatenate 'list '(a . b) '(c . d))
	     *    (a c)
	     *
	     *    => (concatenate 'list '(a b) 'd)
	     *    (a b)
	     */
	    subret = LL_nil;
	    sublast = LL_nil;

	    for (subr=sequence; LL_consp(subr); subr=LL_cdr(subr)) {
		tmp = LLCons(LL_car(subr), LL_nil);

		if (LL_null(subret))
		    subret = tmp;
		else
		    LL_rplacd(sublast, tmp);
		sublast = tmp;
	    }

	    if (LL_null(ret)) {
		ret = subret;
		last = subret;
	    } else {
		LL_rplacd(last, subret);
	    }
	    while ( ! LL_null(LL_cdr(last)) )
		last = LL_cdr(last);
	}

	return (ret);
}

static LLobject *
copy_seq(sequence)
LLobject *sequence;
{
	return (LLCopylist(sequence));
}

static LLobject *
elt(sequence, index)
LLobject *sequence;
LLobject *index;
{
	int index_i;

	index_i = index->value.a_integer;
	if (LLGetLength(sequence) < index_i) {
	    LLErrorF("\
Function: (elt sequence index)\n\
\treference beyond the end of the list\n\
\tlist:  ~S\n\
\tindex: ~S\n\
", sequence, index);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	return (LLGetNth(index_i, sequence));
}

static LLobject *
length(sequence)
LLobject *sequence;
{
	return (LLListlength(sequence));
}

static LLobject *
subseq(sequence, start, end)
LLobject *sequence;
LLobject *start;
LLobject *end;		/* &optional */
{
	int start_i, end_i, i;
	LLobject *s, *tmp, *ret, *last;

	start_i = start->value.a_integer;
	if (LL_null(end)) {
	    for (i=0, s=sequence; i<start_i && LL_consp(s); i++, s=LL_cdr(s))
		;

	    ret = LL_null(s) ? LL_nil: LLCopylist(s);
	} else {
	    end_i = end->value.a_integer;
	    for (i=0, s=sequence; i<start_i && LL_consp(s); i++, s=LL_cdr(s))
		;

	    ret = LL_nil;
	    last = LL_nil;
	    for ( /* inherit */; i<end_i && LL_consp(s); i++, s=LL_cdr(s)) {
		tmp = LLCons(LL_car(s), LL_nil);
		if (LL_null(ret))
		    ret = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;
	    }
	}

	return (ret);
}
