#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static LLobjectType function_template = {
	"" /* filled in below */, "function",	NULL,
		{ LL_FUNCTION /* ... */ }, 0,
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_Pr_Function, LL_Pr_Function, LL_Pr_Function },
		{ LL_GcScan_Function, NULL }, NULL
};

/*
 *    LLMakeFunction - create an function
 *
 *    LLobject *
 *    function( ... arguments ... )
 *    LLobject *argument0;
 *    LLobject *argument1;
 *    ...
 *
 *    see LightLispMachine.h for the distinction among function types
 *
 *    return the function cell
 */
LLobject *
LLMakeFunction(prepname, function, nrequired, noptional, restP)
String prepname;
LLobject *(*function)();
int nrequired;
int noptional;
Boolean restP;
{
	int total;
	LLobject *new;
	char buf[100];
	LLobjectType template;
	LLstring pname;

	total = nrequired + noptional + (restP ? 1: 0);
	if (total < 0 || LL_MAXARGS <= total) {
	    LLErrorF("\
Too many arguments requested for a function\n\
\trequest: ~D\t(limit ~D)\n\
\tfunction: ~A\n\
", total, LL_MAXARGS, prepname);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	pname = LLInternSymbol(prepname);
	new = LLMakeObject();

	if (LLFunctionCells[nrequired][noptional][restP ? 1: 0] == NULL) {
	    template = function_template;
	    template.flags[1] = nrequired;
	    template.flags[2] = noptional;
	    template.flags[3] = restP;
            sprintf(buf, "function%d-%d%s", nrequired, noptional, restP ? "-rest": "");
	    template.tname = buf;
            LLFunctionCells[nrequired][noptional][restP ? 1: 0] = LLMakeObjectType(template);
	}

	new->type = LLFunctionCells[nrequired][noptional][restP ? 1: 0];
	new->value.a_function.pname = pname;
	new->value.a_function.function = function;
	new->type->count++;

	return (new);
}
