#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

/*
 *    LLMakeObjectType - make a new objectType
 *
 *    Stick it in the table
 *    If you change this, be sure to modify the manual
 *    definition of the type type (LLTypeCell) in init.c
 *
 *    return the objectType
 */
LLobjectType *
LLMakeObjectType(template)
LLobjectType template;
{
	LLobjectType *new;
	LLsequence *new_sequence;

	template.tname = LLInternSymbol(template.tname);
	if (st_is_member(LLMachine.types, template.tname)) {
	    LLErrorF("\
Function: LLMakeObjectType\n\
\tthe type name has already been defined\n\
\ttype name: ~A\n\
", template.tname);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (template.sequence != NULL) {
	    template.sequence->sname = LLInternSymbol(template.sequence->sname);
	    if (st_is_member(LLMachine.sequences, template.sequence->sname)) {
		LLErrorF("\
Function: LLMakeObjectType\n\
\tthe sequence name has already been defined\n\
\tsequence name: ~A\n\
", template.sequence->sname);
		LLThrow(LL_lightlisp, LL_nil);
	    }
	}

	new = ALLOC_TYPE(LLobjectType);
	new_sequence = template.sequence == NULL ? NULL: ALLOC_TYPE(LLsequence);
	
	if (new == NULL || (template.sequence != NULL && new_sequence == NULL)) {
	    LLErrorF("\
LightLisp is out of memory - forcing garbage collection\n\
\ttrying to make the type ~A\n\
", template.tname);
	    LLThrow(LL_uncatchable, LL_nil);
	}

	*new = template;
	new->sequence = new_sequence;
	if (new->sequence != NULL) {
	    *new_sequence = *template.sequence;
	    new_sequence->type = new;
	    st_add_direct(LLMachine.sequences, new_sequence->sname, new_sequence);
	}

	new->tname = template.tname;
	new->pname = LLInternSymbol(template.pname);
	new->count = 0;

	/* new->prin1[0] must exist */
	if (new->prin1[1] == NULL)
	    new->prin1[1] = new->prin1[0];
	if (new->prin1[2] == NULL)
	    new->prin1[2] = new->prin1[0];

	/* new->princ[0] must exist */
	if (new->princ[1] == NULL)
	    new->princ[1] = new->princ[0];
	if (new->princ[2] == NULL)
	    new->princ[2] = new->princ[0];

	/* new->print[0] must exist */
	if (new->print[1] == NULL)
	    new->print[1] = new->print[0];
	if (new->print[2] == NULL)
	    new->print[2] = new->print[0];

	/* new->pprint[0] must exist */
	if (new->pprint[1] == NULL)
	    new->pprint[1] = new->pprint[0];
	if (new->pprint[2] == NULL)
	    new->pprint[2] = new->pprint[0];

	st_add_direct(LLMachine.types, new->tname, new);

	new->object = LLMakeObject();
	new->object->type = LLTypeCell;
	new->object->value.a_type = new;
	new->object->type->count++;

	return (new);
}
