#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

/*
 *    LLNconc - destructively append lists
 *
 *    (nconc &rest lists)
 *
 *    return the new list
 */
LLobject *
LLNconc(lists)
LLobject *lists;	/* &rest */
{
	LLobject *l, *ret, *last, *this;

	if (LL_null(lists))
	    return (LL_nil);
	if (LL_null(LL_cdr(lists)))	/* one arg only */
	    return (LL_car(lists));

	/*
	 *    => (nconc '(a b) 'c)
	 *    (a b . c)
	 *
	 *    don't check the last one (it need not be a cons)
	 */
	for (l=lists; LL_consp(LL_cdr(l)); l=LL_cdr(l)) {
	    if ( ! LL_listp(LL_car(l)) ) {
		LLErrorF("\
Function: (nconc &rest lists)\n\
\tlist is not a list\n\
\tlist: ~S\n\
", LL_car(l));
		LLThrow(LL_lightlisp, LL_nil);
	    }
	}

	ret = LL_car(lists);
	for (last=LL_car(lists); LL_consp(LL_cdr(last)); last=LL_cdr(last))
	    ;

	for (l=LL_cdr(lists); LL_consp(LL_cdr(l)); l=LL_cdr(l)) {
	    for (this=LL_car(l); LL_consp(LL_cdr(this)); this=LL_cdr(this))
		;

	    LL_rplacd(last, this);
	    while (LL_consp(LL_cdr(last)))
		last = LL_cdr(last);
	}

	/*
	 *    last guaranteed to be valid
	 *    LL_car(l) is last member of the args (may not be a list)
	 */
	LL_rplacd(last, LL_car(l));

	return (ret);
}
