#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static LLobject *coerce();
static LLobject *concatenate();
static LLobject *copy_seq();
static LLobject *elt();
static LLobject *length();
static LLobject *subseq();

LLsequence LLStringAsASequence = {
    "string",
    NULL,
    coerce,
    concatenate,
    copy_seq,
    elt,
    length,
    subseq,
};

static LLobject *
coerce(sequence, result_type)
LLobject *sequence;
LLobject *result_type;		/* typeCell (with sequence) */
{
	if (LL_typep(sequence, LLStringCell)) {
	    return (sequence);
	} else if (LL_typep(sequence, LLIntegerCell)) {
	    char buf[100];

	    sprintf(buf, "%ld", sequence->value.a_integer);
	    return (LLMakeString(buf));
	} else if (LL_typep(sequence, LLRealCell)) {
	    char buf[100];

	    sprintf(buf, "%lg", sequence->value.a_real);
	    return (LLMakeString(buf));
	} else if (LL_typep(sequence, LLConsCell)) {
	    LLobject *l;

	    for (l=sequence; LL_consp(l); l=LL_cdr(l)) { 
		if ( ! LL_stringp(LL_car(l)) ) {
		    LLErrorF("\
Function: (coerce: sequence result-type)\n\
\tcan coerce a list only if it is made of strings only\n\
\tsource: ~S\n\
", sequence);
		    LLThrow(LL_lightlisp, LL_nil);
		}
	    }

	    return (LLConcatenate(LLStringCell->object, sequence));
	}

	LLErrorF("\
Function: (coerce sequence result-type)\n\
\tcannot coerce a ~A into a string\n\
\tsource: ~S\n\
", LL_type_of(sequence)->pname, sequence);
	LLThrow(LL_lightlisp, LL_nil);
	/*NOTREACHED*/
}

static LLobject *
concatenate(result_type, sequences)
LLobject *result_type;	/* typeCell */
LLobject *sequences;	/* &rest */
{
	DYNStrgeneralString buf;
	LLobject *s, *sequence, *ret;

	DYNStrInit(&buf);
	for (s=sequences; LL_consp(s); s=LL_cdr(s)) {
	    sequence = LL_car(s);
	    if ( ! LL_stringp(sequence) ) {
		DYNFree(&buf);
		LLErrorF("\
Function: (concatenate result-type &rest sequences)\n\
\tattempt to coerce a non-string member into a string\n\
\tform: ~S\n\
", sequence);
		LLThrow(LL_lightlisp, LL_nil);
	    }

	    DYNStrAppendString(&buf, sequence->value.a_string);
	}

	/* TODO (sortof) - if LLMakeString throws, the buf is lost */
	ret = LLMakeString(DYNStrTheString(&buf));

	DYNStrFree(&buf);

	return (ret);
}

static LLobject *
copy_seq(sequence)
LLobject *sequence;
{
	return (LLMakeString(sequence->value.a_string));
}

static LLobject *
elt(sequence, index)
LLobject *sequence;
LLobject *index;
{
	char chars[2];
	int index_i;

	index_i = index->value.a_integer;
	if (strlen(sequence->value.a_string) < index_i) {
	    LLErrorF("\
Function: (elt sequence index)\n\
\treference beyond the end of the string\n\
\tstring: ~S\n\
\tindex:  ~S\n\
", sequence, index);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	chars[0] = sequence->value.a_string[index_i];
	chars[1] = '\0';

	return (LLMakeString(chars));
}

static LLobject *
length(sequence)
LLobject *sequence;
{
	return (LLMakeInteger(strlen(sequence->value.a_string)));
}

static LLobject *
subseq(sequence, start, end)
LLobject *sequence;
LLobject *start;
LLobject *end;		/* &optional */
{
	String string;
	int start_i, end_i, real_end, len;
	DYNStrgeneralString buf;
	LLobject *ret;

	string = sequence->value.a_string;
	real_end = strlen(string);
	start_i = start->value.a_integer;
	end_i = LL_null(end) ? real_end: MIN(real_end, end->value.a_integer);

	if (start_i >= real_end || start_i >= end_i) {
	    ret = LLMakeString("");
	} else {
	    len = end_i - start_i;

	    DYNStrInit(&buf);
	    DYNStrAppendNChars(&buf, &string[start_i], len);

	    /* TODO (sortof) - if LLMakeString throws, the buf is lost */
	    ret = LLMakeString(DYNStrTheString(&buf));

	    DYNStrFree(&buf);
	}

	return (ret);
}
