#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "machine.h"
#include "printer.h"
#include "internal.h"

static enum st_retval symbol_killer();
static enum st_retval string_killer();
static enum st_retval type_killer();

/*
 *    LL_Mach_Uninitialize - uninitialize things
 *
 *    return {LL_ok, LL_notok} on {is, is not}.
 */
LLstatus
LL_Mach_Uninitialize()
{
	LLblock *b, *nb;

	DYNStkFree(&LLMachine.catchers);

	DYNFree(&LLMachine.externals);
	st_free_table(LLMachine.integers);
	st_free_table(LLMachine.handles);
	st_free_table(LLMachine.oblist);

	st_foreach(LLMachine.types, type_killer, NULL);
	st_free_table(LLMachine.types);
	st_free_table(LLMachine.sequences);		/* killed with the types */

	st_foreach(LLStringTable, string_killer, NULL);
	st_free_table(LLStringTable);

	st_foreach(LLSymbolTable, symbol_killer, NULL);
	st_free_table(LLSymbolTable);

	/*
	 *    Toast all of the blocks
	 */
	for (b=LLHeap.halves[LLHeap.this].top; b != NULL; b=nb) {
	    nb = b->next;
	    FREE_TYPE(b, LLblock);
	}
	for (b=LLHeap.halves[LLHeap.that].top; b != NULL; b=nb) {
	    nb = b->next;
	    FREE_TYPE(b, LLblock);
	}

	return (LL_ok);
}

static enum st_retval 
type_killer(symbol, type, bullshit)
LLstring symbol;
LLobjectType *type;
FitAny bullshit;                        /* b/c st (3) requires it */
{
	if (type != LLNilCell) {
	    /*
	     *    hack it so that (sequencep nil) -> t
	     *    Check out the corresponding hack in LL_Mach_Initialize (init.c)
	     *
	     *    LLNilCell->sequence = &LLConsAsASequence;
	     */

	    if (type->sequence != NULL)
		FREE_TYPE(type->sequence, LLsequence);
	}

        FREE_TYPE(type, LLobjectType);

        return (ST_CONTINUE);
}

static enum st_retval 
symbol_killer(string, symbol, bullshit)
String string;
LLstring symbol;
FitAny bullshit;                        /* b/c st (3) requires it */
{
        FREE_STRING(symbol);

        return (ST_CONTINUE);
}

static enum st_retval 
string_killer(string, sstring, bullshit)
String string;
String sstring;
FitAny bullshit;                        /* b/c st (3) requires it */
{
        FREE_STRING(sstring);

        return (ST_CONTINUE);
}
