#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>
#include <varargs.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLFormatCv - format to a stream with a vector of arguments
 *
 *    The args are C variables
 *
 *    format may have in it:
 *        ~S
 *        ~A
 *        ~D, ~O ~X
 *        ~F, ~G
 *        ~C
 *
 *    return nothing, just do it.
 */
void
LLFormatCv(stream, format, ap)
STREAM *stream;
String format;
va_list ap;
{
	char *f;
	char buf[100];
	LLobject *sexp;
	String string;
	int32 integer;
	double real;
	int character;

	/* insure that errors are in sync with the real output */
	if (stream == LLErrorStream && LLStandardStream != NULL)
	    STREAMFlush(LLStandardStream);

	for (f=format; *f != '\0'; f++) {
	    if (*f != '~') {
		STREAMPutc(stream, *f);
	    } else if (LLTwiddleP(f, 'S')) {
		/*
		 *    ~S
		 */
		sexp = va_arg(ap, LLobject *);
		LL_prin1(sexp, 0, stream);	/* suitable for (read ... )  */
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'A')) {
		/*
		 *    ~A
		 */
		string = va_arg(ap, String);
		STREAMPuts(stream, string);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'D')) {
		/*
		 *    ~D
		 */
		integer = va_arg(ap, int32);
		sprintf(buf, "%ld", integer);
		STREAMPuts(stream, buf);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'O')) {
		/*
		 *    ~O
		 */
		integer = va_arg(ap, int32);
		sprintf(buf, "%#lo", integer);
		STREAMPuts(stream, buf);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'X')) {
		/*
		 *    ~X
		 */
		integer = va_arg(ap, int32);
		sprintf(buf, "%#lx", integer);
		STREAMPuts(stream, buf);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'E')) {
		/*
		 *    ~E
		 */
		real = va_arg(ap, double);
		sprintf(buf, "%le", real);
		STREAMPuts(stream, buf);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'F')) {
		/*
		 *    ~F
		 */
		real = va_arg(ap, double);
		sprintf(buf, "%lf", real);
		STREAMPuts(stream, buf);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'G')) {
		/*
		 *    ~G
		 */
		real = va_arg(ap, double);
		sprintf(buf, "%lg", real);
		STREAMPuts(stream, buf);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'C')) {
		/*
		 *    ~C
		 */
		character = va_arg(ap, int);
		STREAMPutc(stream, (char) character);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, '~')) {
		STREAMPutc(stream, '~');
	    } else if (LLTwiddleP(f, '&')) {
		if (STREAMGetColumnno(stream, STREAM_output) != 0)
		    STREAMPutc(stream, '\n');
		f++;	/* f++ above */
            } else if (LLTwiddleP(f, '%')) {
		STREAMPutc(stream, '\n');
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, '\n')) {
		f++;	/* over the '~' */
		f++;	/* over the '\n' */
		while (*f != '\n' && isspace(*f))
		    f++;
		f--;		/* b/c of the f++ above */
	    } else {
		STREAMPutc(stream, *f);
	    }
	}
}
