#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>
#include <varargs.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

static void badarg();
static void toomany();

/*
 *    LLFormatS - format to a stream
 *
 *    format may have in it:
 *        ~S
 *        ~A
 *        ~D, ~O ~X
 *        ~F, ~G
 *        ~C
 *
 *    return nothing, just do it.
 */
void
LLFormatS(stream, format, sexps)
STREAM *stream;
String format;
LLobject *sexps;		/* &rest style */
{
	char *f;
	LLobject *rexps, *sexp;

	/* insure that errors are in sync with the real output */
	if (stream == LLErrorStream && LLStandardStream != NULL)
	    STREAMFlush(LLStandardStream);

	for (f=format, rexps=sexps; *f != '\0'; f++) {
	    if (*f != '~') {
		STREAMPutc(stream, *f);
	    } else if (LLTwiddleP(f, 'S')) {
		/*
		 *    ~S
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		LL_prin1(sexp, 0, stream);	/* suitable for (read ... ) */

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'A')) {
		/*
		 *    ~A
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		LL_princ(sexp, 0, stream);	/* print it verbatim */

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'D')) {
		/*
		 *    ~D
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		if ( ! LL_integerp(sexp) )
		    badarg(format, f, sexp);

		LL_print(sexp, 0, stream);

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'O')) {
		/*
		 *    ~O
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		if ( ! LL_integerp(sexp) )
		    badarg(format, f, sexp);

		LL_print(sexp, 1, stream);

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'X')) {
		/*
		 *    ~X
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		if ( ! LL_integerp(sexp) )
		    badarg(format, f, sexp);

		LL_print(sexp, 2, stream);

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'G')) {
		/*
		 *    ~G
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		if ( ! LL_realp(sexp) )
		    badarg(format, f, sexp);

		LL_print(sexp, 0, stream);

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'F')) {
		/*
		 *    ~F
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		if ( ! LL_realp(sexp) )
		    badarg(format, f, sexp);

		LL_print(sexp, 1, stream);

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, 'E')) {
		/*
		 *    ~E
		 */
		if (LL_null(rexps))
		    toomany(format, rexps);

		sexp = LL_car(rexps);
		if ( ! LL_realp(sexp) )
		    badarg(format, f, sexp);

		LL_print(sexp, 2, stream);

		rexps = LL_cdr(rexps);
		f++;	/* f++ above */
            } else if (LLTwiddleP(f, '~')) {
		STREAMPutc(stream, '~');
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, '&')) {
		if (STREAMGetColumnno(stream, STREAM_output) != 0)
		    STREAMPutc(stream, '\n');
		f++;	/* f++ above */
            } else if (LLTwiddleP(f, '%')) {
		STREAMPutc(stream, '\n');
		f++;	/* f++ above */
	    } else if (LLTwiddleP(f, '\n')) {
		f++;	/* over the '~' */
		f++;	/* over the '\n' */
		while (*f != '\n' && isspace(*f))
		    f++;
		f--;		/* b/c of the f++ above */
	    } else {
		STREAMPutc(stream, *f);
	    }
	}

	if ( ! LL_null(rexps) ) {
	    LLErrorF("\
Too few directives in format\n\
\tformat:    \"~A\"\n\
\targuments: ~S\n\
", format, sexps);
	    LLThrow(LL_lightlisp, LL_nil);
	}
}

static void
badarg(format, f, sexp)
String format;
char *f;
LLobject *sexp;
{
	LLErrorF("\
Badly argument to format\n\
\ttype violation\n\
\tformat:      \"~A\"\n\
\tunprocessed:   ... ~A\"\n\
\targument:    ~S\n\
", format, f, sexp);
	LLThrow(LL_lightlisp, LL_nil);
}

static void
toomany(format, sexps)
String format;
LLobject *sexps;
{
	LLErrorF("\
Too many directives in format\n\
\tformat:    \"~A\"\n\
\targuments: ~S\n\
", format, sexps);
	LLThrow(LL_lightlisp, LL_nil);
}
