#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

#define a_good_place_to_break	40

static void in_line();	/* inline is a reserved word for gcc */
static void pretty();

void
LL_Pr_Nil(obj, stream)
LLobject *obj;
STREAM *stream;
{
	STREAMPuts(stream, "nil");
}

void
LL_Pr_Integer(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~D", obj->value.a_integer);
}

void
LL_Pr_Integer1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~O", obj->value.a_integer);
}


void
LL_Pr_Integer2(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~X", obj->value.a_integer);
}

void
LL_Pr_Real(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~G", obj->value.a_real);
}

void
LL_Pr_Real1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~F", obj->value.a_real);
}

void
LL_Pr_Real2(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~E", obj->value.a_real);
}

void
LL_Pr_String(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~A", obj->value.a_string);
}

void
LL_Pr_String1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "\"~A\"", obj->value.a_string);
}

void
LL_Pr_String2(obj, stream)
LLobject *obj;
STREAM *stream;
{
	char *s;

	STREAMPutc(stream, '"');
	for (s=obj->value.a_string; *s != '\0'; s++) {
	    if (*s == '\\' || *s == '"') {
		STREAMPutc(stream, '\\');
		STREAMPutc(stream, *s);
	    } else if (isprint(*s)) {
		STREAMPutc(stream, *s);
	    } else {
		STREAMPutc(stream, '\\');
		switch (*s) {
		case '\b':
		    STREAMPutc(stream, 'b');
		    break;
		case '\f':
		    STREAMPutc(stream, 'f');
		    break;
		case '\n':
		    STREAMPutc(stream, 'n');
		    break;
		case '\r':
		    STREAMPutc(stream, 'r');
		    break;
		case '\t':
		    STREAMPutc(stream, 't');
		    break;
		case '\v':
		    STREAMPutc(stream, 'v');
		    break;
		default:
		    LLFormatC(stream, "~D~D~D", 	/* not ~O (sticks an O in front) */
			    (((int) *s)>>6) & 0x7,
			    (((int) *s)>>3) & 0x7,
			    ((int) *s) & 0x7);
		    break;
		}
	    }
	}
	STREAMPutc(stream, '"');
}

void
LL_Pr_Symbol(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~A~A", LLInternedSymbolP(obj) ? "": "#:", LL_pname(obj));
}

void
LL_Pr_Type(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "<~A>", obj->value.a_type->pname);	/* LL_pname doesn't work on these */
}

void
LL_Pr_Closure(obj, stream)
LLobject *obj;
STREAM *stream;
{
	STREAMPuts(stream, "<closure ");
	LL_pprint(LL_first(LL_code(obj)), 0, stream);
	STREAMPuts(stream, " code ... env ");
	LL_pprint(LL_env(obj), 0, stream);
	STREAMPutc(stream, '>');
}

void
LL_Pr_Cons(obj, stream)
LLobject *obj;
STREAM *stream;
{
	in_line(stream, obj);
}

void
LL_Pr_Cons1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	pretty(stream, obj, 10, 0, FALSE);	/* 10 levels of nesting */
}

void
LL_Pr_Function(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "~A", LL_pname(obj));
}

void
LL_Pr_Distinguished(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "<The ~A>", LL_pname(obj));
}

void
LL_Pr_Dead(obj, stream)
LLobject *obj;
STREAM *stream;
{
	STREAMPuts(stream, "dead");
}

void
LL_Pr_Stream(obj, stream)
LLobject *obj;
STREAM *stream;
{
	STREAM *s;

	s = obj->value.a_stream;
	if (s == NULL) {
	    /*
	     *    closed
	     */
	    LLFormatC(stream, "closed stream");
	} else if (s->input.type != STREAM_void && s->output.type != STREAM_void) {
	    /*
	     *    Bidirectional
	     */
	    LLFormatC(stream, "bidirectional stream on (\"~A\", \"~A\")",
		    STREAMGetName(s, STREAM_input),
		    STREAMGetName(s, STREAM_output));
	} else if (s->input.type != STREAM_void) {
	    /*
	     *    Input stream (only) 
	     */
	    LLFormatC(stream, "input stream on \"~A\"",
		    STREAMGetName(s, STREAM_input));
	} else if (s->output.type != STREAM_void) {
	    /*
	     *    Output stream (only) 
	     */
	    LLFormatC(stream, "output stream on \"~A\"",
		    STREAMGetName(s, STREAM_output));
	} else {
	    /*
	     *    Its dead jim
	     */
	    LLFormatC(stream, "vacuous stream");
	}
}

void
LL_Pr_Stream1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	STREAM *s;

	s = obj->value.a_stream;
	if (s->input.type != STREAM_void && s->output.type != STREAM_void) {
	    /*
	     *    Bidirectional
	     */
	    LLFormatC(stream, "bidirectional stream on (\"~A\" (line ~D, column ~D), \"~A\" (line ~D, column ~D))",
		    STREAMGetName(s, STREAM_input),
		    STREAMGetLineno(s, STREAM_input),
		    STREAMGetColumnno(s, STREAM_input),
		    STREAMGetName(s, STREAM_output),
		    STREAMGetLineno(s, STREAM_output),
		    STREAMGetColumnno(s, STREAM_output));
	} else if (s->input.type != STREAM_void) {
	    /*
	     *    Input stream (only) 
	     */
	    LLFormatC(stream, "input stream on \"~A\" (line ~D, column ~D)",
		    STREAMGetName(s, STREAM_input),
		    STREAMGetLineno(s, STREAM_input),
		    STREAMGetColumnno(s, STREAM_input));
	} else if (s->output.type != STREAM_void) {
	    /*
	     *    Output stream (only) 
	     */
	    LLFormatC(stream, "output stream on \"~A\" (line ~D, column ~D)",
		    STREAMGetName(s, STREAM_output),
		    STREAMGetLineno(s, STREAM_output),
		    STREAMGetColumnno(s, STREAM_output));
	} else {
	    /*
	     *    Its dead jim
	     */
	    LLFormatC(stream, "vacuous stream");
	}
}

void
LL_Pr_Garbage(obj, stream)
LLobject *obj;
STREAM *stream;
{
	LLFormatC(stream, "garbage at ~X", obj);
}

/*
 *    workhorse functions
 */
static void
in_line(stream, sexp)		/* suitable for reading back in */
STREAM *stream;
LLobject *sexp;
{
	LLobject *s, *subsexp, *macro, *value;

	if (LL_atom(sexp)) {
	    LL_prin1(sexp, 0, stream);
	} else if (LL_null(LL_cdr(LL_cdr(sexp)))) {
	    /*
	     *    (<something> <something-else>)
	     *
	     *    Only inline those forms for which there is not
	     *    a corresponding lisp function.  This means
	     *   ` , and ,@
	     */
	    macro = LL_first(sexp);
	    value = LL_second(sexp);

	    if (LL_eq(macro, LL_backquote)) {
		/*
		 *    (` f) -> `f
		 */
		STREAMPutc(stream, '`');
		in_line(stream, value);
	    } else if (LL_eq(macro, LL_comma)) {
		/*
		 *    (, f) -> ,f
		 */
		STREAMPutc(stream, ',');
		if (LL_eq(LL_first(value), LL_atsign)) {
		    /*
		     *    (@ f) -> @f
		     *    for use in (, (@ f)) -> ,@f
		     */
		    STREAMPutc(stream, '@');
		    in_line(stream, LL_second(value));
		} else {
		    /*
		     *    (, f) -> ,f
		     */
		    in_line(stream, value);
		}
	    } else {
		/* must have been a mistake - handle like a regular cons */
		goto I_hate_gotos;
	    }
	} else {	/* a cons */
I_hate_gotos:
	    STREAMPutc(stream, '(');
	    for (s=sexp; LL_consp(s); s=LL_cdr(s)) {
		subsexp = LL_car(s);

		if (LL_eq(subsexp, LL_backquote)) {
		    /*
		     *    handle the cases
		     *
		     *    (list list list . `f)
		     *     which is really (list list list ` f)
		     */
		    STREAMPuts(stream, ". `");
		    in_line(stream, LL_second(s));
		    s = LL_nil;
		} else if (LL_eq(subsexp, LL_comma)) {
		    /*
		     *    handle the cases
		     *
		     *    (list list list . ,f)
		     *     which is really (list list list , f)
		     */
		    STREAMPuts(stream, ". ,");
		    in_line(stream, LL_second(s));
		    s = LL_nil;
		    break;
		}

		in_line(stream, subsexp);

		if ( ! LL_null(LL_cdr(s)) )
		    STREAMPutc(stream, ' ');
	    }

	    if ( ! LL_null(s) ) {
		STREAMPuts(stream, ". ");
		in_line(stream, s);
	    }

	    STREAMPutc(stream, ')');
	}
}

static void
pretty(stream, sexp, level, column, firstcolumn, ignoreP)
STREAM *stream;
LLobject *sexp;
int level;		/* level > 0 implies recur */
int column;
int firstcolumn;
Boolean ignoreP;
{
	LLobject *s, *subsexp, *macro, *value;
	int i;
	Boolean firstP = TRUE;

	if ( ! ignoreP && STREAMGetColumnno(stream, STREAM_output) >= a_good_place_to_break) {
	    STREAMPutc(stream, '\n');

	    /* combination of tabs and spaces */
	    for (i=0; i<column/8; i++)
		STREAMPutc(stream, '\t');
	    for (i=0; i<column%8; i++)
		STREAMPutc(stream, ' ');
	}

	if (LL_atom(sexp)) {
	    LL_pprint(sexp, 0, stream);
	} else if (level == 0) {
	    STREAMPuts(stream, "###");
	} else if (LL_null(LL_cdr(LL_cdr(sexp)))) {
	    /*
	     *    (<something> <something-else>)
	     *
	     *    Inline any well-known read macro character
	     *    this means ` , ' #' and and ,@
	     */
	    macro = LL_first(sexp);
	    value = LL_second(sexp);

	    if (LL_eq(macro, LL_quote)) {
		/*
		 *    (quote f) -> 'f
		 */
		STREAMPutc(stream, '\'');
		pretty(stream, value, level-1, column, column, TRUE);
	    } else if (LL_eq(macro, LL_function)) {
		/*
		 *    (function f) -> #'f
		 */
		STREAMPutc(stream, '#');
		STREAMPutc(stream, '\'');
		pretty(stream, value, level-1, column, column, TRUE);
	    } else if (LL_eq(macro, LL_backquote)) {
		/*
		 *    (` f) -> `f
		 */
		STREAMPutc(stream, '`');
		pretty(stream, value, level-1, column, column, TRUE);
	    } else if (LL_eq(macro, LL_comma)) {
		/*
		 *    (, f) -> ,f
		 */
		STREAMPutc(stream, ',');
		if (LL_eq(LL_first(value), LL_atsign)) {
		    /*
		     *    (@ f) -> @f
		     *    for use in (, (@ f)) -> ,@f
		     */
		    STREAMPutc(stream, '@');
		    pretty(stream, LL_second(value), level-1, column, column, TRUE);
		} else {
		    /*
		     *    (, f) -> ,f
		     */
		    pretty(stream, value, level-1, column, column, TRUE);
		}
	    } else {
		goto I_hate_gotos;
	    }
	} else {
I_hate_gotos:
	    STREAMPutc(stream, '(');
	    firstcolumn = STREAMGetColumnno(stream, STREAM_output);

	    for (s=sexp; LL_consp(s); s=LL_cdr(s)) {
		subsexp = LL_car(s);

		if (LL_eq(subsexp, LL_backquote)) {
		    /*
		     *    handle the cases
		     *
		     *    (list list list . `f)
		     *     which is really (list list list ` f)
		     */
		    STREAMPuts(stream, ". `");
		    pretty(stream, LL_second(s), level-1, column, firstcolumn, TRUE);
		    s = LL_nil;
		    break;
		} else if (LL_eq(subsexp, LL_comma)) {
		    /*
		     *    handle the cases
		     *
		     *    (list list list . ,f)
		     *     which is really (list list list , f)
		     */
		    STREAMPuts(stream, ". ,");
		    pretty(stream, LL_second(s), level-1, column, firstcolumn, TRUE);
		    s = LL_nil;
		    break;
		}

		if (LL_consp(subsexp)) {
		    if (firstP)
			column = STREAMGetColumnno(stream, STREAM_output);

		    pretty(stream, subsexp, level-1, column, firstcolumn, FALSE);
		} else {
		    /* don't call pretty so it can recur - just call this directly */
		    LL_pprint(subsexp, 0, stream);
		}

		if ( ! LL_null(LL_cdr(s)) ) {
		    if (STREAMGetColumnno(stream, STREAM_output) < a_good_place_to_break) {
			STREAMPutc(stream, ' ');
		    } else {
			STREAMPutc(stream, '\n');

			/* combination of tabs and spaces */
			for (i=0; i<column/8; i++)
			    STREAMPutc(stream, '\t');
			for (i=0; i<column%8; i++)
			    STREAMPutc(stream, ' ');
		    }
		}

		if (firstP) {
		    column = STREAMGetColumnno(stream, STREAM_output);
		    firstP = FALSE;
		}
	    }

	    if ( ! LL_null(s) ) {
		STREAMPuts(stream, ". ");

		if (LL_eq(s, LL_cdr(sexp)))
		    column = firstcolumn;

		/* TRUE - try to keep it on the same line */
		pretty(stream, s, level-1, column, firstcolumn, TRUE);
	    }
	    STREAMPutc(stream, ')');
	}
}
