#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>
#include <varargs.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "internal.h"

/*
 *    LLPrintFullStats - get and print the full stats
 *
 *    return nothing, just do it.
 */
void
LLPrintFullStats(stream)
STREAM *stream;		/* may be NULL */
{
	LLfullStats mem;
	LLcellsByType c;
	int i;

	LLGetFullStats(&mem);

	if (stream == NULL)
	    stream = LLErrorStream;

	STREAMPuts(stream, "Full LightLisp Stats\n");
	STREAMPutc(stream, '\n');

	STREAMPuts(stream, "Breakdown By Storage Monitoring\n");
	LLFormatC(stream, "  ~Dk bytes total core size\n", mem.totalCoreBytes/1024);
	LLFormatC(stream, "  ~Dk bytes LightLisp storage\n", mem.totalBytes/1024 + 1);
	STREAMPuts(stream, "--------------\n");
	LLFormatC(stream, "  ~Dk bytes not monitored\n", (mem.totalCoreBytes - mem.totalBytes)/1024 + 1);
	STREAMPutc(stream, '\n');

	STREAMPuts(stream, "Breakdown By Storage Class\n");
	LLFormatC(stream, "  ~Dk bytes for cell storage\n", mem.cellBytes/1024 + 1);
	LLFormatC(stream, "  ~Dk bytes for string storage\n", mem.stringBytes/1024 + 1);
	STREAMPuts(stream, "--------------\n");
	LLFormatC(stream, "  ~Dk bytes LightLisp storage\n", mem.totalBytes/1024 + 1);
	STREAMPutc(stream, '\n');

	STREAMPuts(stream, "Breakdown By Cell Status\n");
	LLFormatC(stream, "  ~D cells used\n", mem.cellsUsed);
	LLFormatC(stream, "  ~D cells free\n", mem.cellsFree);
	STREAMPuts(stream, "--------------\n");
	LLFormatC(stream, "  ~D cells total\n", mem.cellsUsed + mem.cellsFree);
	if (mem.maximumCellsAllowed != 0)
	    LLFormatC(stream, "  ~D cells maximum\n", mem.maximumCellsAllowed);
	STREAMPutc(stream, '\n');

	STREAMPuts(stream, "Breakdown By Cell Type\n");
	for (i=0; i<DYNSizeOfEl(&mem.cellsByType); i++) {
	    c = DYNAt(&mem.cellsByType, i, LLcellsByType);
	    LLFormatC(stream, "  ~D ~A cell~A\n", c.count, c.pname, c.count == 1 ? "": "s");
	}
	STREAMPuts(stream, "--------------\n");
	LLFormatC(stream, "  ~D total cells\n", mem.totalCells);
	STREAMPutc(stream, '\n');

	STREAMPuts(stream, "Breakdown of Time\n");
	LLFormatC(stream, "  ~F sec for LightLisp\n", mem.totalTime/1000.0);
	LLFormatC(stream, "  ~F sec for ~D garbage collect operations\n", mem.garbageCollectTime/1000.0, mem.garbageCollections);
	STREAMPuts(stream, "--------------\n");
	LLFormatC(stream, "  ~F% of LightLisp was spent in garbage collection\n", (100.0 * mem.garbageCollectTime) / mem.totalTime);

	LLFreeFullStats(&mem);
}
