#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "reader.h"
#include "internal.h"

static LLobject *read();
static LLobject *read_one();

/*
 *    LLReadValue - read an LightLisp form from a stream
 *
 *    return the form
 */
LLobject *
LLReadValue(stream, eof_error_p, eof_value, recursive_p, error)
STREAM *stream;
Boolean eof_error_p;
LLobject *eof_value;
Boolean recursive_p;
LLobject *error;	/* tag to throw to on error (can be LL_nil) */
{
	LLobject *ret;
	int c;

	LL_Re_STREAM = stream;
	LL_Re_CP = FALSE;
	LL_Re_CatcherTag = error;

again:
	LL_ridwhite();
	c = LL_peekc();

	switch (c) {
	case EOF:
	    if (eof_error_p || recursive_p)
		ret = read(recursive_p);	/* will cause an error */
	    /* else */
	    return (eof_value);
	case ';':
	    /*
	     *    ; comment
	     */
	    while ((c = LL_getc()) != '\n' && c != EOF)
		;
	    /* c == '\n' */
	    goto again;
	default:
	    ret = read_one(FALSE);
	    break;
	}

	return (ret);
}

static LLobject *
read_one(listP)
Boolean listP;
{
	LLobject *ret;
	int c;

again:
	LL_ridwhite();
	c = LL_peekc();

	switch (c) {
	case ';':
	    /*
	     *    ; comment
	     */
	    while ((c = LL_getc()) != '\n' && c != EOF)
		;
	    /* c == '\n' */
	    goto again;
	case ')':
	    LLFormatC(LL_Re_STREAM, "\
~A, line ~D: unmatched ')'\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input));
	    LLThrow(LL_Re_CatcherTag, LL_nil);
	    break;
	case EOF:
	    ret = read(listP);	/* will cause an error */
	    break;
	default:
	    ret = read(listP);
	    break;
	}

	return (ret);
}

/*
 *    read - read a token, or the car, and then the cdr
 *
 *    Forestall as much as possible till the lint phase.
 *
 *    return the object
 */
static LLobject *
read(listP)
Boolean listP;
{
	LLobject *car, *cdr, *ret;
	int c;	/* really a char */

again:
	LL_ridwhite();

	switch (c = LL_getc()) {
	case EOF:
	    LLFormatC(LL_Re_STREAM, "\
~A, line ~D: unexpected eof\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input));
	    LLThrow(LL_Re_CatcherTag, LL_nil);
	    break;
	case ';':
	    /*
	     *    ; comment
	     */
	    while ((c = LL_getc()) != '\n' && c != EOF)
		;
	    /* c == '\n' */
	    goto again;
	case ')':
	    /*
	     *    The rest of a list
	     */
	    return (LL_nil);
	case '(':
	    /*
	     *    (form ...  )
	     *    (car . cdr)
	     */
	    LL_ridwhite();
	    if ((c = LL_peekc()) == ')') {
		LL_getc();
		ret = LL_nil;
	    } else {
		/* DON'T LLCons(read(FALSE), read(TRUE)) - the args are evaluated BACKWARDS */
		car = read(FALSE);
		cdr = read(TRUE);
		ret = LLCons(car, cdr);
	    }
	    break;
	case '.':
	    /* DON'T LL_ridwhite(); */
	    c = LL_peekc();
	    if (isspace(c)) {
		/*
		 *   (value value value . value)
		 */
		c = LL_getc();		/* get rid of the '.' */
		cdr = read(FALSE);

		LL_ridwhite();
		if ((c = LL_getc()) != ')') {
		    LLFormatC(LL_Re_STREAM, "\
~A, line ~D: must be a single form after .\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input));
		    LLThrow(LL_Re_CatcherTag, LL_nil);
		}
		return (cdr);
	    } else if (isdigit(c)) {
		/*
		 *    .4
		 */
		ret = LL_Re_Number('.');
	    } else {
		/*
		 *    .apple
		 */
		ret = LL_Re_Symbol('.');
	    }
	    break;
	case '"':
	    /*
	     *    stringToken
	     */
	    ret = LL_Re_String(c);
	    break;
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	case '-':
	    /*
	     *    integer | real
	     */
	    ret = LL_Re_Number(c);
	    break;
	case '#':
	    /*
	     *    #readmacro
	     */
	    if ((c = LL_getc()) == EOF) {
		LLFormatC(LL_Re_STREAM, "\
~A, line ~D: unexpected eof after `#'\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input));
		LLThrow(LL_Re_CatcherTag, LL_nil);
	    }

	    if (LL_null(LLMachine.dispatchreadtable[c])) {
		LLFormatC(LL_Re_STREAM, "\
~A, line ~D: undefined dispatch macro character `~C'\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input),
	c);
		LLThrow(LL_Re_CatcherTag, LL_nil);
	    }

	    ret = LLCons(LLMachine.dispatchreadtable[c], LLCons(read_one(FALSE), LL_nil));
	    break;
	default:
	    if ( ! LL_null(LLMachine.readtable[c]) ) {
		/*
		 *    readmacro
		 */
		ret = LLCons(LLMachine.readtable[c], LLCons(read_one(FALSE), LL_nil));
	    } else {
		/*
		 *    symbol
		 */
		ret = LL_Re_Symbol(c);
	    }
	    break;
	}

	return (listP ? LLCons(ret, read(TRUE)): ret);
}
