#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <ctype.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "printer.h"
#include "reader.h"
#include "internal.h"

/*
 *    LL_Re_String - parse a string
 *
 *    return the string
 */
LLobject *
LL_Re_String(start)
char start;		/* may be '\0' */
{
	int c, i, odigit;
	Boolean backP;

	c = start == '\0' ? LL_getc(): start;		/* c == '"' */

	DYNStrClear(&LL_Re_StringToken);

	backP = FALSE;
	do {
	    c = LL_getc();
	    if (c == '"' && ! backP) {
		/* time to go */
		break;
	    } else if (c == EOF) {
		LLFormatC(LL_Re_STREAM, "\
~A, line ~D: unexpected eof while reading string\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input));
		LLThrow(LL_Re_CatcherTag, LL_nil);
	    } else if ( ! backP ) {
		backP = c == '\\';
		if ( ! backP )
		    DYNStrAppendC(&LL_Re_StringToken, c);
	    } else {	/* backP */
		switch (c) {
		case '\\':
		    DYNStrAppendC(&LL_Re_StringToken, '\\');
		    backP = FALSE;
		    break;
		case '\b':
		case 'b':
		    DYNStrAppendC(&LL_Re_StringToken, '\b');
		    backP = FALSE;
		    break;
		case '\f':
		case 'f':
		    DYNStrAppendC(&LL_Re_StringToken, '\f');
		    backP = FALSE;
		    break;
		case '\n':
		    /* a backslashed newline means ignore it */
		    backP = FALSE;
		    break;
		case 'n':
		    DYNStrAppendC(&LL_Re_StringToken, '\n');
		    backP = FALSE;
		    break;
		case '\r':
		    /* a backslashed return means ignore it */
		    backP = FALSE;
		    break;
		case 'r':
		    DYNStrAppendC(&LL_Re_StringToken, '\r');
		    backP = FALSE;
		    break;
		case '\t':
		case 't':
		    DYNStrAppendC(&LL_Re_StringToken, '\t');
		    backP = FALSE;
		    break;
		case '\v':
		case 'v':
		    DYNStrAppendC(&LL_Re_StringToken, '\v');
		    backP = FALSE;
		    break;
		case ' ':
		    DYNStrAppendC(&LL_Re_StringToken, ' ');
		    backP = FALSE;
		    break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		    odigit = c - '0';
		    for (i=0; i < 2 && (c = LL_getc()) >= '0' && c <= '7'; i++)
			odigit = 8 * odigit + c - '0';
		    if (i < 1)		/* took one too many */
			LL_ungetc(c);
		    if (odigit == 0) {
			LLFormatC(LL_Re_STREAM, "\
~A, line ~D: attempt to place `\\0' in a string (next char is `~C')\n\
", STREAMGetName(LL_Re_STREAM, STREAM_input),
	STREAMGetLineno(LL_Re_STREAM, STREAM_input), c);
			LLThrow(LL_Re_CatcherTag, LL_nil);
		    }
		    DYNStrAppendC(&LL_Re_StringToken, (char) odigit);
		    backP = FALSE;
		    break;
		case '"':
		    DYNStrAppendC(&LL_Re_StringToken, '"');
		    backP = FALSE;
		    break;
		default:
		    DYNStrAppendC(&LL_Re_StringToken, c);
		    backP = FALSE;
		    break;
		}
	    }
	} while (1);

	return (LLMakeString(DYNStrTheString(&LL_Re_StringToken)));
}
