#include <sys/param.h>
#include <sys/file.h>
#include <stdio.h>

#include "stuff.h"

#include "LightLisp.h"

static void usage();

/*
 *    LLMain - a main for terminal based LightLisp programs
 *
 *    LLstatus
 *    user_initialize()
 *
 *    return stuff for exit() from main()
 */
int
LLMain(argc, argv, binary, package, libdir, user_initialize)
int argc;
char **argv;
String binary;			/* e.g. argv[0] but may be NULL */
String package;			/* e.g. "LightLisp" but may be NULL */
String libdir;			/* e.g. "~cad/lib/LightLisp" but may be NULL */
LLstatus (*user_initialize)();	/* may be NULL */
{
	extern char *getenv();	/* damnit I wish they'd do this for us */
	Boolean no_lisprcp, no_preamblep, interactivep, preamble_directory_suppliedp;
	String word, preamble_directory, preamble_filename;
	char basic_preamble_filename[100];
	char buf[2*MAXPATHLEN];	/* big enuf */
	char *b;
	LLstatus status;
	int i;

	if (binary == NULL) {
	    binary = "lightlisp";
	} else {
	    if ((b = rindex(binary, '/')) != NULL)
		binary = &b[1];
	}
	if (package == NULL) {
	    package = "LightLisp";
	}
	if (libdir == NULL) {
	    if ((libdir = getenv(package)) == NULL)
		libdir = LightLisp_LIBDIR;
	}

	no_lisprcp = FALSE;
	no_preamblep = FALSE;
	interactivep = FALSE;

	preamble_directory = libdir;
	preamble_directory_suppliedp = FALSE;

	sprintf(basic_preamble_filename, "%s.ll", binary);
	preamble_filename = basic_preamble_filename;

	for (i=1; i<argc; i++) {
	    if (argv[i][0] != '-')
		break;

	    word = &argv[i][1];
	    if (nocase_lead_strequal("usage", word)) {
		usage(binary, NULL);
	    } else if (nocase_lead_strequal("nolisprc", word)) {
		no_lisprcp = TRUE;
	    } else if (nocase_lead_strequal("nopreamble", word)) {
		no_preamblep = TRUE;
	    } else if (nocase_lead_strequal("interactive", word)) {
		interactivep = TRUE;
	    } else if (nocase_lead_strequal("preambleDirectory", word)) {
		if (i+1 >= argc)
		    usage(binary, argv[i]);
		preamble_directory_suppliedp = TRUE;
		preamble_directory = argv[++i];
	    } else if (nocase_lead_strequal("preambleFilename", word)) {
		if (i+1 >= argc)
		    usage(binary, argv[i]);
		preamble_filename = argv[++i];
	    } else {
		usage(binary, argv[i]);
	    }
	}

        if (LLInitialize() != LL_ok) {
            fprintf(stderr, "%s: could not initialize LightLisp\n", binary);
            exit (1);
        }

	if (user_initialize != NULL && (*user_initialize)() != LL_ok) {
	    fprintf(stderr, "%s: could not initialize %s\n", binary, package);
	    exit (1);
	}

	if ( ! no_lisprcp ) {
	    sprintf(buf, "~/.%src", binary);
	    /* okay if its not there (be silent) */
	    if (access(FNTildeExpand(buf), F_OK) == 0)
		status = LLSource(buf);
	}

	if ( ! no_preamblep ) {
	    Boolean minep = FALSE;

	    if ( ! preamble_directory_suppliedp ) {
		sprintf(buf, "~/lib/%s/%s", package, preamble_filename);
		minep = access(FNTildeExpand(buf), F_OK) == 0 ? TRUE: FALSE;
	    }

	    if ( ! minep )
		sprintf(buf, "%s/%s", preamble_directory, preamble_filename);

	    /* NOT okay if its not there (DON'T be silent) */
	    status = LLSource(buf);
	}

	/*
	 *    If args left, they must be filenames; source them
	 */
	if (i < argc) {
	    for ( /* i = i */; i<argc; i++) {
		if (argv[i][0] == '-') {
		    fprintf(stderr, "%s: options cannot appear after filenames\n", package);
		    usage(binary, argv[i]);
		}

		status = LLSource(argv[i]);
		if (status != LL_ok)
		    exit (1);
	    }

	    if (interactivep)
		status = LLShell(binary);
	} else {
	    status = LLShell(binary);
	}

	return (status == LL_ok ? 0: 1);
}

static void 
usage(binary, argv)
String binary;
String argv;
{
	if (argv != NULL)
	    fprintf(stderr, "%s error at %s\n", binary, argv);

	fprintf(stderr, "\
usage: %s\n\
    [ -nopreamble ]\n\
    [ -nolisprc ]\n\
    [ -interactive ]\n\
    [ -preambleDirectory preamble_directory ]\n\
    [ -preambleFilename preamble_filename ]\n\
    [ filename ... ]\n\
", binary);
	exit (1);
}
