#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <signal.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "reader.h"
#include "printer.h"
#include "internal.h"

static STREAMstatus prompter();
static int totop();

/*
 *    LLShell - an interactive shell
 *
 *    CAUTION - calls garbage-collect
 *
 *    return {LL_ok, LL_notok} on {success, fail}
 */
LLstatus
LLShell(prompt)
String prompt;
{
	LLobject *sexp;
	STREAM *stream;
	int neofs;
	STREAMcharCounts counts;
	Boolean signaledP = FALSE;
/* shit, I hate system dependencies */
#if OPSYS_MFR == MD_DEC && OPSYS_VERS >= MD_ULTRIX32_2_4
	void (*sigint)(), (*sigfpe)();
#else
	int (*sigint)(), (*sigfpe)();
#endif

	if (LLInitialize() != LL_ok)
	    return (LL_notok);

	stream = STREAMSplice(LL_Re_InputStream, LLStandardStream);
	if (stream == NULL) {
	    LLErrorF("Shell: could not splice input and output streams\n");
	    return (LL_notok);
	}
	STREAMSetPrompt(stream, STREAM_tty, prompter, prompt, NULL);

	if (LLTopLevelCatchVoid()) {
	    /*
	     *    signals end up here
	     */
	    /* DON'T make the mistake of trying to print sexp */
	    STREAMClear(stream, STREAM_input);
	}

	if ( ! signaledP ) {
	    sigint = signal(SIGINT, totop);
	    sigfpe = signal(SIGFPE, totop);
	    signaledP = TRUE;
	}

	for (neofs=0; neofs < 20; neofs++) {

	    if (LLCatchVoid(LL_alwayscatchable, LL_nil, LL_environment)) {
		STREAMClear(stream, STREAM_input);

		/*
		 *   Normal error throws end up here
		 */
		/* don't print anything - printed below */
		continue;    /* make SURE to reinstantiate the catcher */
	    }

	    while ((sexp = LLReadValue(stream, FALSE, LL_eofvalue, FALSE, LL_lightlisp)) != LL_eofvalue) {
		neofs = 0;	/* successfully read something */
		sexp = LLEvaluator(sexp, LL_environment);

		LLPprint(sexp, LL_nil);	/* will pretty-print it */
		LLTerpri(LL_nil);

		LLGarbageCollect();
	    }

	    LLEndCatch();
	    LLPrintF("\nEOF - use (exit) to quit\n", sexp);
	    STREAMForcePrompt(stream);
	}

	/*
	 *    replace
	 */
	signal(SIGINT, sigint);
	signal(SIGFPE, sigfpe);

	return (LL_ok);
}

static STREAMstatus
prompter(stream, prompt)
STREAM *stream;
String prompt;
{
	STREAMcharCounts counts;

	STREAMGetCharCounts(stream, STREAM_input, &counts);

	if (counts.parens > 0) {
	    LLFormatC(stream, "~A-~D> ", prompt, counts.parens);
	} else {
	    LLFormatC(stream, "~A> ", prompt);
	}

	STREAMFlush(stream, STREAM_output);
}

static int
totop(signo)
int signo;
{
	if (signo != SIGINT) {
	    LLFormatC(LLStandardStream, "\n~A\n", sys_siglist[signo]);
	} else {
	    LLFormatC(LLStandardStream, "\n");
	}

	LLThrow(LL_uncatchable, LL_nil);
}
