#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "reader.h"
#include "printer.h"
#include "internal.h"

/*
 *    LLSourceString - source a string (e.g. load, but uses a string as arg)
 *
 *    CAUTION - calls garbage-collect
 *
 *    return {LL_ok, LL_notok} on {success, fail}
 */
LLstatus
LLSourceString(string_sexp)
String string_sexp;
{
	LLobject *sexp, *promise;
	STREAM *stream;

	if (LLInitialize() != LL_ok)
	    return (LL_notok);

	if (LLTopLevelCatchVoid())
	    return (LL_notok);

	stream = STREAMCreate(STREAM_string, string_sexp, string_sexp, STREAM_STREAM, LLErrorStream, NULL);
	if (stream == NULL) {
	    LLErrorF("source-string: could not read \"~A\"\n", string_sexp);
	    return (LL_notok);
	}

	/*
	 *    The promise must be an implicit progn.
	 */
	promise = LLCons(LLMakeSymbol("close"), LLCons(LLMakeStream(stream), LL_nil));
	promise = LLCons(promise, LL_nil);

        /*
         *    Use a scheme with two catchers in order to be able
         *    get the stream and the line number before the promise kills it
         */
        /* not a top-level catch b/c of the promise */
        if (LLCatchVoid(LL_uncatchable, promise, LL_environment))
            return (LL_notok);  /* never reached */

        if (LLCatchVoid(LL_alwayscatchable, LL_nil, LL_environment)) {
            LLErrorF("source: error in reading \"~A\" (aborting read at line ~D)\n",
                    string_sexp, STREAMGetLineno(stream, STREAM_input));
            LLEndCatch();       /* evaluate the promise */
            return (LL_notok);
        }

	promise = LL_nil;	/* so it can't be used after the gc */

	while ((sexp = LLReadValue(stream, FALSE, LL_eofvalue, FALSE, LL_lightlisp)) != LL_eofvalue) {
	    LLEvaluator(sexp, LL_environment);
	    LLGarbageCollect();
	}

	LLEndCatch();	 /* always catchable */
	LLEndCatch();	 /* evaluate the promise */

	return (LL_ok);
}
