#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "reader.h"
#include "printer.h"
#include "internal.h"

/*
 *    CAUTION - DO NOT call garbage-collect
 *    This is meant to be called from the lower levels
 */

/*
 *    LLSubshell - an interactive subshell
 *
 *    (subshell [ prompt ])
 *
 *    return the last value evaluated
 */
LLobject *
LLSubshell(arguments, environment)
LLobject *arguments;
LLobject *environment;
{
	LLSubshell_tr(&arguments, &environment);
	return (arguments);
}

static STREAMstatus prompter();

/*
 *    LLSubshell - an interactive subshell
 *
 *    (subshell [ prompt ])
 *
 *    return {t, nil} on {tail-recursive, non-tail-recursive}
 */
LLobject *
LLSubshell_tr(arguments, environment)
LLobject **arguments;		/* value/return */
LLobject **environment;		/* vaue/return */
{
	LLobject *ret, *sexp, *promise;
	String prompt;
	STREAM *stream;
	int len;

	if ((len = LLGetLength(*arguments)) > 1) {
	    LLErrorF("\
Form: (subshell [ prompt ] )\n\
\tinvalid number of arguments to subshell\n\
\targuments: ~S\n\
", *arguments);
	    LLThrow(LL_lightlisp, LL_nil);
	}

	if (len == 0) {
	    prompt = "shell";
	} else {
	    if ( ! LL_stringp(LL_first(*arguments)) ) {
		LLErrorF("\
Form: (subshell [ prompt ] )\n\
\tinvalid number of arguments to subshell\n\
\tprompt: ~S\n\
", LL_first(*arguments));
		LLThrow(LL_lightlisp, LL_nil);
	    }
	    prompt = LL_first(*arguments)->value.a_string;
	}

	STREAMFlush(LLStandardStream);
	STREAMFlush(LLErrorStream);

	/* copy it - so you can stick a prompt on it */
	stream = STREAMSplice(LL_Re_InputStream, LLStandardStream);
	if (stream == NULL) {
	    LLErrorF("\
Form: (subshell [ prompt ] )\n\
\tcould not splice stream\n\
");
	    *arguments = LL_nil;
	    return (LL_nil);
	}

	STREAMSetPrompt(stream, STREAM_tty, prompter, prompt, NULL);

	promise = LLCons(LLMakeSymbol("close"), LLCons(LLMakeStream(stream), LL_nil));
	promise = LLCons(promise, LL_nil);

	LLCatchVoid(LL_uncatchable, promise, LL_environment);	/* to evaluate the promise */

	ret = LL_nil;	/* in case no args are evaluated */
	do {
	    if (LLCatchVoid(LL_alwayscatchable, LL_nil, LL_environment)) {
		STREAMClear(stream, STREAM_input);

		/*
		 *   Normal error throws end up here
		 */
		/* don't print anything - printed below */
		continue;    /* make SURE to reinstantiate the catcher */
	    }

	    while ((sexp = LLReadValue(stream, FALSE, LL_eofvalue, FALSE, LL_lightlisp)) != LL_eofvalue) {
		ret = LLEvaluator(sexp, *environment);

		LLPprint(ret, LL_nil);	/* will pretty-print it */
		LLTerpri(LL_nil);

	    }

	    STREAMPuts(stream, "\nEOF\n");
	    break;
	} while (1);

	LLEndCatch();	/* LL_alwayscatchable */
	LLEndCatch();	/* LL_uncatchable - evaluate the promise */

	*arguments = ret;
	return (LL_nil);
}

static STREAMstatus
prompter(stream, prompt)
STREAM *stream;
String prompt;
{
	STREAMcharCounts counts;

	STREAMGetCharCounts(stream, STREAM_input, &counts);

	if (counts.parens > 0) {
	    LLFormatC(stream, "sub~A-~D> ", prompt, counts.parens);
	} else {
	    LLFormatC(stream, "sub~A> ", prompt);
	}

	STREAMFlush(stream, STREAM_output);
}
