#include "sumachine.h"

#include <sys/types.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstr.h"

#include "stream.h"

#include "internal.h"

/*
 *    STREAMPuts - put a string (no extra newline)
 *
 *    return {STREAM_ok, STREAM_notok} on {ok, notok}
 */
STREAMstatus
STREAMPuts(stream, s)
STREAM *stream;
String s;
{
	char *cs;
	STREAMstatus status;

	switch (stream->output.type) {
	case STREAM_void:
	    status = STREAM_ok;
	    break;
	case STREAM_FILE:
	    status = STREAM_ok;
	    switch (stream->newline.type) {
	    case STREAM_beforenewline:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if (*cs == '\n')
			STREAMCallNewlineDemon(stream);
		    putc(*cs, stream->output.value.a_FILE.file);
		}
		break;
	    case STREAM_afternewline:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    putc(*cs, stream->output.value.a_FILE.file);
		    if (*cs == '\n')
			STREAMCallNewlineDemon(stream);
		}
		break;
	    case STREAM_column0:
		for (cs=s; *cs != '\0'; cs++) {
		    if (stream->output.counts.columnno == 0)
			STREAMCallNewlineDemon(stream);
		    STREAMCount(&stream->output, *cs);
		    putc(*cs, stream->output.value.a_FILE.file);
		}
		break;
	    case STREAM_nonewline:
	    default:
		for (cs=s; *cs != '\0'; cs++)
		    STREAMCount(&stream->output, *cs);
		fputs(s, stream->output.value.a_FILE.file);
		break;
	    }
	    break;
	case STREAM_string:
	    status = STREAM_ok;
	    switch (stream->newline.type) {
	    case STREAM_beforenewline:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if (*cs == '\n')
			STREAMCallNewlineDemon(stream);
		    DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
		}
		break;
	    case STREAM_afternewline:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
		    if (*cs == '\n')
			STREAMCallNewlineDemon(stream);
		}
		break;
	    case STREAM_column0:
		for (cs=s; *cs != '\0'; cs++) {
		    if (stream->output.counts.columnno == 0)
			STREAMCallNewlineDemon(stream);
		    STREAMCount(&stream->output, *cs);
		    DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
		}
		break;
	    case STREAM_nonewline:
	    default:
		for (cs=s; *cs != '\0'; cs++)
		    STREAMCount(&stream->output, *cs);
		DYNStrAppendString(&stream->output.value.a_string.buf, s);
		break;
	    }
	    break;
	case STREAM_charback:
	    status = STREAM_ok;
	    switch (stream->newline.type) {
	    case STREAM_beforenewline:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if (*cs == '\n')
			STREAMCallNewlineDemon(stream);
		    if ((status = (*stream->output.value.a_callback.function)(*cs)) == STREAM_notok)
			break;
		}
		break;
	    case STREAM_afternewline:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if ((status = (*stream->output.value.a_callback.function)(*cs)) == STREAM_notok)
			break;
		    if (*cs == '\n')
			STREAMCallNewlineDemon(stream);
		}
		break;
	    case STREAM_column0:
		for (cs=s; *cs != '\0'; cs++) {
		    if (stream->output.counts.columnno == 0)
			STREAMCallNewlineDemon(stream);
		    STREAMCount(&stream->output, *cs);
		    if ((status = (*stream->output.value.a_callback.function)(*cs)) == STREAM_notok)
			break;
		}
		break;
	    case STREAM_nonewline:
	    default:
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if ((status = (*stream->output.value.a_callback.function)(*cs)) == STREAM_notok)
			break;
		}
		break;
	    }
	    break;
	case STREAM_lineback:
	    status = STREAM_ok;
	    for (cs=s; *cs != '\0'; cs++) {
		STREAMCount(&stream->output, *cs);

		if (stream->output.last == '\n')
		    DYNStrClear(&stream->output.value.a_string.buf);
		stream->output.last = *cs;

		DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
		if (*cs != '\n') {
		    status = STREAM_ok;
		} else {
		    switch (stream->newline.type) {
		    case STREAM_beforenewline:
			if (*cs == '\n')
			    STREAMCallNewlineDemon(stream);
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			break;
		    case STREAM_afternewline:
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			if (*cs == '\n')
			    STREAMCallNewlineDemon(stream);
			break;
		    case STREAM_column0:
			if (stream->output.counts.columnno == 0)
			    STREAMCallNewlineDemon(stream);
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			break;
		    case STREAM_nonewline:
		    default:
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			break;
		    }
		}
	    }
	    break;
	case STREAM_xactback:
	    status = STREAM_ok;
	    switch (stream->newline.type) {
	    case STREAM_beforenewline:
		DYNStrClear(&stream->output.value.a_string.buf);
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if (*cs != '\n') {
			DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
		    } else {
			/* throw the string */
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			DYNStrClear(&stream->output.value.a_string.buf);
			/* then call the newline function */
			STREAMCallNewlineDemon(stream);
			/* then throw the '\n' */
			DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			DYNStrClear(&stream->output.value.a_string.buf);
		    }
		}
		if (DYNStrLength(&stream->output.value.a_string.buf) != 0)
		    status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
		break;
	    case STREAM_afternewline:
		DYNStrClear(&stream->output.value.a_string.buf);
		for (cs=s; *cs != '\0'; cs++) {
		    STREAMCount(&stream->output, *cs);
		    if (*cs != '\n') {
			DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
		    } else {
			/* throw the string */
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			DYNStrClear(&stream->output.value.a_string.buf);
			/* then throw the '\n' */
			DYNStrAppendC(&stream->output.value.a_string.buf, *cs);
			status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
			DYNStrClear(&stream->output.value.a_string.buf);
			/* then call the newline function */
			STREAMCallNewlineDemon(stream);
		    }
		}
		if (DYNStrLength(&stream->output.value.a_string.buf) != 0)
		    status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
		break;
	    case STREAM_nonewline:
	    default:
		for (cs=s; *cs != '\0'; cs++)
		    STREAMCount(&stream->output, *cs);
		DYNStrClear(&stream->output.value.a_string.buf);
		DYNStrAppendString(&stream->output.value.a_string.buf, s);
		status = (*stream->output.value.a_callback.function)(&stream->output.value.a_string.buf);
		DYNStrClear(&stream->output.value.a_string.buf);
		break;
	    }
	    break;
	default:
	    status = STREAM_notok;
	    break;
	}

	return (status);
}
