#include "sumachine.h"

#include <sys/types.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstr.h"
#include "filename.h"

#include "stream.h"

#include "internal.h"

static STREAMstatus setup();

/*
 *    STREAMSplice - create a stream
 *
 *    return the created stream
 */
STREAM *
STREAMSplice(input, output)
STREAM *input;
STREAM *output;
{
	STREAM *new;
	STREAMstatus input_status, output_status;

	new = ALLOC_TYPE(STREAM);
	if (new == NULL)
	    return (NULL);

	input_status = setup(&new->input, STREAM_input, &input->input);

	new->firstReadP = input->firstReadP;
	DYNStrInit(&new->unget);
	DYNStrAppendString(&new->unget, DYNStrTheString(&input->unget));
	new->prompt.type = STREAM_noprompt; /* set it later */
	new->newline.type = STREAM_nonewline; /* set it later */

	output_status = setup(&new->output, STREAM_output, &output->output);
	if (input_status != STREAM_ok || output_status != STREAM_ok) {
	    STREAMDestroy(new);
	    return (NULL);
	}

	/*
	 *    Note - Don't destroy the userData (e.g. its the master's)
	 *    TODO - find a way to copy userData's
	 */
	STREAMSetPrompt(new, input->prompt.type, input->prompt.callback, input->prompt.userData, NULL);
	
	return (new);
}

static STREAMstatus
setup(new, sidename, side)
STREAMside *new;
STREAMsideName sidename;
STREAMside *side;
{
	int fd;

	new->type = side->type;

	new->counts = side->counts;

	new->name = COPY_STRING(side->name);

	switch (side->type) {
	case STREAM_void:
	    break;
	case STREAM_FILE:
	    fd = dup(fileno(side->value.a_FILE.file));
	    new->value.a_FILE.destroyP = TRUE;	/* a new copy */
	    new->value.a_FILE.file = fdopen(fd, sidename == STREAM_output ? "a": "r");
	    if (new->value.a_FILE.file == NULL)
		return (STREAM_notok);
	    break;
	case STREAM_string:
	    DYNStrInit(&new->value.a_string.buf);
	    DYNStrAppendString(&new->value.a_string.buf, DYNStrTheString(&new->value.a_string.buf));
	    new->value.a_string.index = side->value.a_string.index;
	    break;
	case STREAM_charback:
	    new->value.a_callback.function = side->value.a_callback.function;
	    break;
	case STREAM_lineback:
	case STREAM_xactback:
	    new->value.a_callback.function = side->value.a_callback.function;
	    DYNStrInit(&new->value.a_callback.buf);
	    break;
	default:
	    return (STREAM_notok);
	}

	return (STREAM_ok);
}
