#include "sumachine.h"

#include <sys/param.h>
#include <stdio.h>
#include "signal.h"

#include "stuff.h"
#include "alloc.h"
#include "tostring.h"
#include "toargv.h"
#include "filename.h"

static Boolean keyboardP;
static FILE *tty;

static int collectend;
static Stringv collect;

static void initl();
static void appendl();

static void suckwarts();

static int continuer();

Stringv
CollectArgv(init)
Stringv init;
{
#if OPSYS_SUBTYPE != MD_AIX && \
	OPSYS_SUBTYPE != MD_HPUX && \
	OPSYS_SUBTYPE != MD_SYS5 /* { */
#if (OPSYS_MFR == MD_DEC && OPSYS_VERS >= MD_ULTRIX32_2_4) || \
	(OPSYS_MFR == MD_SUN && OPSYS_VERS >= MD_SUN_4_0) /* { */
	void (*oldcontinuer)();
#else /* } else { */
	int (*oldcontinuer)();
#endif /* } */
#endif /* } */

	tty = fopen("/dev/tty", "w");

	if (tty == NULL) {
	    keyboardP = FALSE;
	} else {
	    keyboardP = isatty(fileno(stdin));
#if OPSYS_MFR == MD_DEC /* { */
	    setvbuf(tty, NULL, _IOLBF, BUFSIZ);
#endif /* } */
	}

#if OPSYS_SUBTYPE != MD_AIX && \
	OPSYS_SUBTYPE != MD_HPUX && \
	OPSYS_SUBTYPE != MD_SYS5 /* { */
	oldcontinuer = signal(SIGCONT, continuer);
#endif /* } */

	/*
	 *    If there was initialization
	 */
	if (init == NULL) {
	    initl();
	} else {
	    initl();

	    while (*init != NULL)
		appendl(COPY_STRING(*init));
	}

	suckwarts();

#if OPSYS_SUBTYPE != MD_AIX && \
	OPSYS_SUBTYPE != MD_HPUX && \
	OPSYS_SUBTYPE != MD_SYS5 /* { */
	signal(SIGCONT, oldcontinuer);
#endif /* } */

	fclose(tty);

	return (collect);
}

static char *line;
static int lineend;

static void initc();
static void appendc();

/*
 *    True, this is a bit X rated here, but you'll live.
 */
static void suckraw();
static void sucktty();

static void
suckwarts()
{
	if (keyboardP)
	    sucktty();
	else
	    suckraw(stdin);
}

static void
suckraw(file)
FILE *file;
{
	int c;

	initc();

	/*
	 *    Sure its inefficient, but who asked you.
	 */
	while ((c = getc(file)) != EOF) {
	    if (c == '\n') {
		appendl(line);
		initc();
	    } else {
		appendc(c);
	    }
	}
}

static void waste_rest();

static void
sucktty()
{
	char filename[MAXPATHLEN];
	Boolean first_charP;
	int c, nlines;
	Stringv edited;
	FILE *f;

	initc();

	first_charP = TRUE;

	while ((c = getc(stdin)) != EOF) {
	    switch (c) {
	    case '.':
		if ( ! first_charP ) {
		    appendc(c);
		} else {
		    if ((c = getc(stdin)) == '\n')
			goto out;	/* '.' alone on a line */
		    else
			ungetc(c, stdin);
		}
		break;
	    case '~':
		if ( ! first_charP ) {
		    appendc(c);
		} else {
		    switch (c = getc(stdin)) {
		    case '~':
			appendc('~');
			break;
		    case 'p':
			PrintArgv(tty, collect);
#if OPSYS_MFR != MD_DEC
			fflush(tty);
#endif
			waste_rest(c);
			break;
		    case 'r':
			if (fscanf(stdin, "%s", filename) != 1) {
			    fprintf(tty, "badly formed filename\n");
			    continue;
			}

			waste_rest(c);

			/* TODO - shit - this file should be its own package b/c of this dependency on filename */
			if ((f = FNTildeOpen(filename, "r")) == NULL) {
			    fprintf(tty, "could not open %s because %s\n", filename, sys_errlist[errno]);
			    continue;
			}

			nlines = lineend;

			suckraw(f);
			fclose(f);

			nlines = lineend - nlines;
			fprintf(tty, "\"%s\" %d line%s\n", filename, nlines, nlines == 1 ? "": "s"); 
			break;
		    case 'e':
		    case 'v':
			waste_rest(c);

			if ((edited = EditArgv(collect)) == NULL) {
			    fprintf(tty, "could not edit because %s\n", sys_errlist[errno]);
			    continue;
			}

			FreeArgv(collect);
			collectend = CountArgv(edited);
			collect = edited;

			fputs("(continue)\n", tty);
			break;
		    case 'w':
			if (fscanf(stdin, "%s", filename) != 1) {
			    fprintf(tty, "badly formed filename\n");
			    continue;
			}
			waste_rest(c);

			/* TODO - shit - this file should be its own package b/c of this dependency on filename */
			if ((f = FNTildeOpen(filename, "w")) == NULL) {
			    fprintf(tty, "could not open %s because %s\n", filename, sys_errlist[errno]);
			    continue;
			}

			PrintArgv(f, collect);
			fclose(f);

			fprintf(tty, "\"%s\" %d line%s\n", filename, lineend, lineend == 1 ? "": "s");
			break;
		    case '?':
		    default:
			waste_rest(c);
			fprintf(tty, "\
----------------------------------------\n\
~~		Quote a single tilde\n\
~e		Same as ~v\n\
~p		Print the buffer\n\
~r		Read in a file\n\
~v		Edit the buffer\n\
~w		Write to a file\n\
~?		Give this display\n\
----------------------------------------\n\
");
			break;
		    }
		}
		break;
	    case '\n':
		/*
		 *    Don't save the newlines - EditArgv doesn't want them.
		 *    And spit() knows to add them.
		 */
		appendl(line);
		initc();
		first_charP = TRUE;
		break;
	    default:
		appendc(c);
		first_charP = FALSE;
		break;
	    }
	}
out:	;

	if (lineend != 0)
	    appendl();

	fputs("EOT\n", tty);
}

static void
initc()
{
	line = ALLOC_ARRAY(1, char);
	lineend = 0;
	line[lineend] = '\0';
}

static void
appendc(c)
char c;
{
	/* +1 for new, +1 for '\0' */
	line = REALLOC_ARRAY(line, lineend+2, char);
	line[lineend++] = c;
	line[lineend] = '\0';
}

static void
initl()
{
	collect = ALLOC_ARRAY(1, String);
	collectend = 0;
	collect[collectend] = NULL;
}

static void
appendl(line)
String line;
{
	/* +1 for new, +1 for NULL */
	collect = REALLOC_ARRAY(collect, collectend+2, String);
	collect[collectend++] = line;
	collect[collectend] = NULL;
}

static void
waste_rest(c)
int c;	/* really a char */
{
	if (c != '\n') {
	    while ((c = getc(stdin)) != EOF && c != '\n')
		/* waste the rest of the line */;
	}
}

static int
continuer()
{
	if (keyboardP)
	    fputs("(continue)\n", tty);
}
