#include "sumachine.h"

#include <sys/types.h>

#include "stuff.h"

/* must include AFTER sumachine.h (stuff.h) */
#if OPSYS_TYPE == MD_UNIX /* { */
#include <sys/file.h>
#include <sys/stat.h>
#if OPSYS_SUBTYPE != MD_AIX && \
	OPSYS_SUBTYPE != MD_HPUX && \
	OPSYS_SUBTYPE != MD_SYS5 /* { */
#include <sys/wait.h>	/* no wait.h or sys/wait.h on aix */
#endif /* } */
#include <signal.h>
#endif /* } */
#if OPSYS_TYPE == MD_VMS /* { */
#include <file.h>
#include <stat.h>
#include <wait.h>
#include <signal.h>
#endif /* } */

#include "alloc.h"
#include "tostring.h"
#include "toargv.h"

extern String getenv();
extern String mktemp();

/*
 *    EditArgv - take a string, and edit it with vi.
 *
 *    return {a copy of the edited Stringv, NULL} on {success, error}.
 */
Stringv
EditArgv(argv)
Stringv argv;
{
	static char template[] = "/tmp/EditXXXXXX";
	char name[sizeof (template)];
#if OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	int status;
#else /* } else { */
	union wait status;
#if (OPSYS_MFR == MD_DEC && OPSYS_VERS >= MD_ULTRIX32_2_4) || \
	(OPSYS_MFR == MD_SUN && OPSYS_VERS >= MD_SUN_4_0) /* { */
	void (*sigcont)();
#else /* } else { */
	int (*sigcont)();
#endif /* } */
#endif /* } */
	int i, fd, tty, len, mask, ntable, pid;
	String tmp, buf, editor;
	Stringv new_vec;
	struct stat st;

	strcpy(name, template);
	tmp = mktemp(name);
	if ((fd = open(tmp, O_CREAT|O_RDWR|O_TRUNC, 0600)) < 0)
	    return (NULL);
	if ((tty = open("/dev/tty", O_RDWR)) < 0) {
	    close(fd);
	    return (NULL);
	}

	if (argv != NULL) {
	    for (i=0; argv[i] != NULL; i++) {
		len = strlen(argv[i]);
		if (len > 0)
		    write(fd, argv[i], len);
		if (len == 0 || argv[i][len-1] != '\n')
		    write(fd, "\n", 1);
	    }
	}
	close(fd);

#if OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	pid = fork();
#else /* } else { */
	pid = vfork();
#endif /* } */

	switch (pid) {
	case -1:
	    close(tty);
	    unlink(name);		/* error */
	    return (NULL);
	case 0:		/* child */
	    for (i=0; i<3; i++) {
		if (i != tty)
		    dup2(tty, i);
	    }

	    ntable = getdtablesize();
	    for (i=3; i<ntable; i++)
		close(i);

	    if ((editor = getenv("EDITOR")) != NULL)
		execlp(editor, editor, name, NULL);
	    execl("/usr/ucb/vi", "vi", name, NULL);
	    execl("/usr/ucb/ex", "ex", name, NULL);
	    execl("/bin/ed", "ed", name, NULL);
	    _exit (1);
	    break;
	default:	/* parent */
	    close(tty);

#if OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	    mask = sigblock(sigmask(SIGQUIT)|sigmask(SIGINT));
#else /* } else { */
	    /*
	     *    Can't block sigcont, so must save the handler
	     *    So the vi can be stopped, and foregrounded
	     */
	    sigcont = signal(SIGCONT, SIG_DFL);
	    mask = sigblock(sigmask(SIGQUIT)|sigmask(SIGINT)|sigmask(SIGSTOP));
#endif /* } */

	    wait(&status);

#if OPSYS_SUBTYPE != MD_AIX && \
	OPSYS_SUBTYPE != MD_HPUX && \
	OPSYS_SUBTYPE != MD_SYS5 /* { */
	    signal(SIGCONT, sigcont);
#endif /* } */
	    sigsetmask(mask);

#if OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	    if (status != 0)
		return (NULL);
#else /* } else { */
	    if (status.w_retcode != 0)
		return (NULL);
#endif /* } */
	    break;
	}

	/*
	 *    Remember vi writes a new file, so you must close the 
	 *    old one, and re-read the file of the same name.
	 *
	 *    Figure out how big the string should be, and then
	 *    allocate some space for it, and read it in.
	 */
	if ((fd = open(tmp, O_RDONLY)) < 0)
	    return (NULL);
	unlink(name);

	fstat(fd, &st);
	buf = ALLOC_ARRAY(st.st_size+1, char);

	if (lseek(fd, 0l, L_SET) < 0 || read(fd, buf, st.st_size) != st.st_size) {
	    close(fd);
	    FREE_ARRAY(buf, st.st_size+1, char);
	    return (NULL);
	}
	close(fd);

	/*
	 *    zero terminate it
	 */
	buf[st.st_size] = '\0';
	new_vec = NewlinesToArgv(buf);
	FREE_ARRAY(buf, st_size+1, char);

	return (new_vec);
}
