#if ! defined(OctLispMachine_h)
#define OctLispMachine_h

extern octObject OLNullObject;	/* use as a substitute for oct_null_id */

/*
 *    The OctLisp machine
 */
#define OL_OTYPE_LOW	((int) OCT_FACET)
#define OL_OTYPE_HIGH	((int) OCT_EDGE)
#define OL_OTYPE_RANGE	(OL_OTYPE_HIGH - OL_OTYPE_LOW + 1)

#define OL_PTYPE_LOW	((int) OCT_NULL)
#define OL_PTYPE_HIGH	((int) OCT_INTEGER_ARRAY)
#define OL_PTYPE_RANGE	(OL_PTYPE_HIGH - OL_PTYPE_LOW + 1)

#define OL_STATUS_LOW	((int) OCT_NO_BB)
#define OL_STATUS_HIGH	((int) OCT_ALREADY_ATTACHED)
#define OL_STATUS_RANGE	(OL_STATUS_HIGH - OL_STATUS_LOW + 1)

#define OL_TTYPE_LOW	((int) OCT_NO_TRANSFORM)
#define OL_TTYPE_HIGH	((int) OCT_FULL_TRANSFORM)
#define OL_TTYPE_RANGE	(OL_TTYPE_HIGH - OL_TTYPE_LOW + 1)

extern LLstring OLObject;

/*
 *    If you add fields here, be sure to initialize them in init.c
 *    and provide garbage collection services there also
 */
typedef struct OLmachine OLmachine;
struct OLmachine {
    st_table *objects;          /* of (objectId, LLobject *) pairs */
    LLobject *object_types[OL_OTYPE_RANGE];
    LLobject *property_types[OL_PTYPE_RANGE];
    LLobject *stati[OL_STATUS_RANGE];
    LLobject *transform_types[OL_TTYPE_RANGE];
    LLobject *object;
    LLobject *octlisp;
};

extern OLmachine OLMachine;

#define OL_object	(OLMachine.object)
#define OL_octlisp	(OLMachine.octlisp)

/*
 *    The various types
 */
extern LLobjectType *OLBoxCell;
extern LLobjectType *OLGeneratorCell;
extern LLobjectType *OLObjectCell;
extern LLobjectType *OLObjectMaskCell;
extern LLobjectType *OLObjectTypeCell;
extern LLobjectType *OLOctStatusCell;
extern LLobjectType *OLPointCell;
extern LLobjectType *OLPropertyTypeCell;
extern LLobjectType *OLTransformCell;
extern LLobjectType *OLTransformMaskCell;
extern LLobjectType *OLTransformTypeCell;

/*
 *    type predicates
 */
#define OL_boxp(o)		((o)->type == OLBoxCell)
#define OL_generatorp(o)	((o)->type == OLGeneratorCell)
#define OL_objectp(o)		((o)->type == OLObjectCell)
#define OL_object_maskp(o)	((o)->type == OLObjectMaskCell)
#define OL_object_typep(o)	((o)->type == OLObjectTypeCell)
#define OL_pointp(o)		((o)->type == OLPointCell)
#define OL_property_typep(o)	((o)->type == OLPropertyTypeCell)
#define OL_oct_statusp(o)	((o)->type == OLOctStatusCell)
#define OL_transformp(o)	((o)->type == OLTransformCell)
#define OL_transform_maskp(o)	((o)->type == OLTransformMaskCell)
#define OL_transform_typep(o)	((o)->type == OLTransformTypeCell)

#define OL_object_formp(f)	(LL_consp(f) && LL_symbolp(LL_first(f)) && \
				    LL_eq(LL_first(f), OL_object) && \
				    LLGetLength(f) == 3)

/*
 *    translate objects back to structures
 */
extern void OLExtractBox();
extern void OLExtractGenerator();
extern void OLExtractObject();
extern void OLExtractObjectMask();
extern void OLExtractObjectType();
extern void OLExtractOctStatus();
extern void OLExtractPoint();
extern void OLExtractPropertyType();
extern void OLExtractTransform();
extern void OLExtractTransformMask();
extern void OLExtractTransformType();

/*
 *    Make primitive objects
 */
extern LLobject *OLMakeBox();
extern LLobject *OLMakeGenerator();
extern LLobject *OLMakeObject();
extern LLobject *OLMakeObjectMask();
extern LLobject *OLMakeObjectType();
extern LLobject *OLMakeOctStatus();
extern LLobject *OLMakePoint();
extern LLobject *OLMakePropertyType();
extern LLobject *OLMakeTransform();
extern LLobject *OLMakeTransformMask();
extern LLobject *OLMakeTransformType();

/*
 *    The functions
 */
extern LLobject *OLAttach();
extern LLobject *OLAttachonce();
extern LLobject *OLBb();
extern LLobject *OLBox();
extern LLobject *OLBoxof();
extern LLobject *OLBoxp();
extern LLobject *OLBreadthfirsttraversal();
extern LLobject *OLCellof();
extern LLobject *OLCenterof();
extern LLobject *OLClosefacet();
extern LLobject *OLCopyfacet();
extern LLobject *OLCreate();
#if ! defined(OCT_JUST_RIGHT) /* { */
/* oct 1.0 */
extern LLobject *OLCreateandattach();
extern LLobject *OLCreatenotattached();
#endif /* } */
extern LLobject *OLCreateormodify();
extern LLobject *OLDelete();
extern LLobject *OLDepthfirsttraversal();
extern LLobject *OLDetach();
extern LLobject *OLEndingangleof();
extern LLobject *OLEndof();
extern LLobject *OLExtractmasks();
extern LLobject *OLFacetof();
extern LLobject *OLFlushfacet();
extern LLobject *OLFreefacet();
extern LLobject *OLFreegenerator();
extern LLobject *OLGeneraltransformof();
extern LLobject *OLGenerate();
extern LLobject *OLGeneratorp();
extern LLobject *OLGenfirstcontainer();
extern LLobject *OLGenfirstcontent();
extern LLobject *OLGetbyname();
extern LLobject *OLGetcontainerbyname();
extern LLobject *OLGetfacet();
extern LLobject *OLGetname();
extern LLobject *OLGetorcreate();
extern LLobject *OLInitgencontainers();
extern LLobject *OLInitgencontents();
extern LLobject *OLInnerradiusof();
extern LLobject *OLInstanceof();
extern LLobject *OLIsattached();
extern LLobject *OLLabelof();
extern LLobject *OLLowerleftof();
extern LLobject *OLMaskmember();
extern LLobject *OLMasterof();
extern LLobject *OLMergemasks();
extern LLobject *OLModeof();
extern LLobject *OLModify();
extern LLobject *OLNameof();
extern LLobject *OLObjectformp();
extern LLobject *OLObjectmaskp();
extern LLobject *OLObjectof();
extern LLobject *OLObjectp();
extern LLobject *OLObjecttypeof();
extern LLobject *OLObjecttypep();
extern LLobject *OLOcterrorstatus();
extern LLobject *OLOcterrorstring();
extern LLobject *OLOctstatusp();
extern LLobject *OLOpenfacet();
extern LLobject *OLOpenfacetbyname();
extern LLobject *OLOpenfacetpair();
extern LLobject *OLOpenmaster();
extern LLobject *OLOuterradiusof();
extern LLobject *OLPoint();
extern LLobject *OLPointof();
extern LLobject *OLPointp();
extern LLobject *OLPointsof();
#if ! defined(OCT_JUST_RIGHT) /* { */
/* oct 1.0 */
extern LLobject *OLPositionof();
#endif /* } */
extern LLobject *OLPropertytypeof();
extern LLobject *OLPropertytypep();
extern LLobject *OLStartingangleof();
extern LLobject *OLStartof();
extern LLobject *OLTransform();
extern LLobject *OLTransformmaskp();
extern LLobject *OLTransformof();
extern LLobject *OLTransformp();
extern LLobject *OLTransformtypeof();
extern LLobject *OLTransformtypep();
extern LLobject *OLTranslationof();
extern LLobject *OLUpperrightof();
extern LLobject *OLValueof();
extern LLobject *OLVersionof();
extern LLobject *OLViewof();
extern LLobject *OLWidthof();
extern LLobject *OLXof();
extern LLobject *OLYof();

#endif /* OctLispMachine_h */
