#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    attach,
    attach_once,
    detach,
    is_attached,
};

static LLobject *func();

/*
 *    OLAttach - attach
 *
 *    (attach container object)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLAttach(container, object)
LLobject *container;
LLobject *object;
{
	return (func(container, object, "attach", attach));
}

/*
 *    OLAttachonce - attach-once
 *
 *    (attach-once container object1)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLAttachonce(container, object)
LLobject *container;
LLobject *object;
{
	return (func(container, object, "attach-once", attach_once));
}

/*
 *    OLDetach - detach
 *
 *    (detach container object)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLDetach(container, object)
LLobject *container;
LLobject *object;
{
	return (func(container, object, "detach", detach));
}

/*
 *    OLIsattached - is-attached
 *
 *    (is-attached container object)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLIsattached(container, object)
LLobject *container;
LLobject *object;
{
	return (func(container, object, "is-attached", is_attached));
}

static LLobject *
func(container, object, name, op)
LLobject *container;
LLobject *object;
String name;
func_op op;
{
	octObject o, o1;
	LLobject *ret;

	if ( ! OL_objectp(container) || ! OL_objectp(object) ) {
	    LLErrorF("\
Function: (~A container object)\n\
\tboth container and object must be objects\n\
\tcontainer: ~S\n\
\tobject:    ~S\n\
", name, container, object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(container, &o);
	OLExtractObject(object, &o1);
	switch (op) {
	case attach:
	    OLGlobalOctStatus = octAttach(&o, &o1);
	    ret = OLGlobalOctStatus < OCT_OK ? LL_nil: LL_t;
	    break;
	case attach_once:
	    OLGlobalOctStatus = octAttachOnce(&o, &o1);
	    ret = OLGlobalOctStatus < OCT_OK ? LL_nil: LL_t;
	    break;
	case detach:
	    OLGlobalOctStatus = octDetach(&o, &o1);
	    ret = OLGlobalOctStatus < OCT_OK ? LL_nil: LL_t;
	    break;
	case is_attached:
	    OLGlobalOctStatus = octIsAttached(&o, &o1);
	    ret = OLGlobalOctStatus == OCT_NOT_ATTACHED ? LL_nil: LL_t;
	    break;
	}

	return (ret);
}
