#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    create,
    create_and_attach,
    create_not_attached,
    create_or_modify,
    get_or_create,
    modify,
};

static LLobject *func();

/*
 *    OLCreate - create
 *
 *    (create container object-form)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLCreate(container, object_form)
LLobject *container;
LLobject *object_form;
{
	return (func(container, object_form, "create", "container", create));
}

#if ! defined(OCT_JUST_RIGHT) /* { */
/* oct 1.0 */
/*
 *    OLCreateandattach - create-and-attach
 *
 *    (create-and-attach container object-form)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLCreateandattach(container, object_form)
LLobject *container;
LLobject *object_form;
{
	return (func(container, object_form, "create-and-attach", "container", create_and_attach));
}

/*
 *    OLCreatenotattached - create-not-attached
 *
 *    (create-not-attached facet object-form)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLCreatenotattached(facet, object_form)
LLobject *facet;
LLobject *object_form;
{
	return (func(facet, object_form, "create-not-attached", "facet", create_not_attached));
}
#endif /* } */

/*
 *    OLCreateormodify - create-or-modify
 *
 *    (create-or-modify container object-form)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLCreateormodify(container, object_form)
LLobject *container;
LLobject *object_form;
{
	return (func(container, object_form, "create-or-modify", "container", create_or_modify));
}

/*
 *    OLGetorcreate - get-or-create
 *
 *    (get-or-create container object-form)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLGetorcreate(container, object_form)
LLobject *container;
LLobject *object_form;
{
	return (func(container, object_form, "get-or-create", "container", get_or_create));
}

/*
 *    OLModify - modify
 *
 *    (modify container object-form)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLModify(container, object_form)
LLobject *container;
LLobject *object_form;
{
	return (func(container, object_form, "modify", "container", modify));
}

static LLobject *
func(container, object_form, name, container_name, op)
LLobject *container;
LLobject *object_form;
String name;
String container_name;
func_op op;
{
	struct octPoint *p;
	int np;
	octObject c, o;

	if ( ! OL_objectp(container) ) {
	    LLErrorF("\
Function: (~A ~A object-form)\n\
\t~A is not an object\n\
\t~A:   ~S\n\
\tobject-form: ~S\n\
", name, container_name, container_name, container_name, container, object_form);
	    LLThrow(OL_octlisp, LL_nil);
	}
	/* object_form checked by OLObjectForm */

	OLExtractObject(container, &c);
	OLObjectForm(object_form, &o, &np, &p);

	switch (op) {
	case create:
	    OLGlobalOctStatus = octCreate(&c, &o);
	    break;
#if ! defined(OCT_JUST_RIGHT) /* { */
/* oct 1.0 */
	case create_and_attach:
	    OLGlobalOctStatus = octCreateAndAttach(&c, &o);
	    break;
	case create_not_attached:
	    OLGlobalOctStatus = octCreateNotAttached(&c, &o);
	    break;
#endif /* } */
	case create_or_modify:
	    OLGlobalOctStatus = octCreateOrModify(&c, &o);
	    break;
	case get_or_create:
	    OLGlobalOctStatus = octGetOrCreate(&c, &o);
	    break;
	case modify:
	    o.objectId = c.objectId;
	    OLGlobalOctStatus = octModify(&o);
	    break;
	}

	if (o.type == OCT_PATH || o.type == OCT_POLYGON) {
	    if (OLGlobalOctStatus >= OCT_OK)
		OLGlobalOctStatus = octPutPoints(&o, np, p);
	}

	return (OLGlobalOctStatus < OCT_OK ? LL_nil: OLMakeObject(o));
}
