#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

static LLobject *func();

/*
 *    OLMergemasks - merge-masks
 *
 *    (merge-masks mask-or-type &rest more-masks-or-types)
 *
 *    return the merged mask
 */
LLobject *
OLMergemasks(mask_or_type, more_masks_or_types)
LLobject *mask_or_type;
LLobject *more_masks_or_types;
{
	return (func("merge-masks", /* mergeP */ TRUE, mask_or_type, more_masks_or_types));
}

/*
 *    OLExtractmasks - extract-masks
 *
 *    (extract-masks mask-or-type &rest more-masks-or-types)
 *
 *    return the extracted mask
 */
LLobject *
OLExtractmasks(mask_or_type, more_masks_or_types)
LLobject *mask_or_type;
LLobject *more_masks_or_types;
{
	return (func("extract-masks", /* mergeP */ FALSE, mask_or_type, more_masks_or_types));
}

static LLobject *
func(name, mergeP, mask_or_type, more_masks_or_types)
String name;
Boolean mergeP;
LLobject *mask_or_type;
LLobject *more_masks_or_types;
{
	LLobjectType *typeCell, *altTypeCell;
	String combo;
	octObjectMask ret_om, om;
	int /* octType */ tot;
	int /* octTransformMask */ ret_tm, tm;
	octTransformType ttt;
	LLobject *m, *ma;

	if (OL_object_maskp(mask_or_type) || OL_object_typep(mask_or_type)) {
	    typeCell = OLObjectMaskCell;
	    altTypeCell = OLObjectTypeCell;
	    combo = "an object-mask or object-type";
	} else if (OL_transform_maskp(mask_or_type) || OL_transform_typep(mask_or_type)) {
	    typeCell = OLTransformMaskCell;
	    altTypeCell = OLTransformTypeCell;
	    combo = "a transform-mask or transform-type";
	} else {
	    LLErrorF("\
Function: (~A mask_or_type &rest more-masks)\n\
\tmask is not an object-mask, object-type, transform-mask or transform-type\n\
\tmask:       ~S\n\
\tmore-masks: ~S\n\
", name, mask_or_type, more_masks_or_types);
	    LLThrow(OL_octlisp, LL_nil);
	}

	for (m=more_masks_or_types; LL_consp(m); m=LL_cdr(m)) {
	    ma = LL_car(m);
	    if ( ! LL_typep(ma, typeCell) && ! LL_typep(ma, altTypeCell) ) {
		LLErrorF("\
Function: (~A mask_or_type &rest more-masks)\n\
\tmask in more-masks is not ~A\n\
\tmask in more-masks: ~S\n\
\tmask:               ~S\n\
\tmore-masks:         ~S\n\
", name, combo, ma, mask_or_type, more_masks_or_types);
		LLThrow(OL_octlisp, LL_nil);
	    }
	}

	if (typeCell == OLObjectMaskCell) {
	    if (OL_object_maskp(mask_or_type)) {
		OLExtractObjectMask(mask_or_type, &ret_om);
	    } else {
		OLExtractObjectMask(mask_or_type, &tot);
		ret_om = (1<<(int) tot);
	    }

	    for (m=more_masks_or_types; LL_consp(m); m=LL_cdr(m)) {
		ma = LL_car(m);

		if (OL_object_maskp(ma)) {
		    OLExtractObjectMask(ma, &om);
		} else {
		    OLExtractObjectType(ma, &tot);
		    om = (1<<(int) tot);
		}

		if (mergeP)
		    ret_om |= om;
		else
		    ret_om &= ~om;
	    }

	    return (OLMakeObjectMask(ret_om));
	} else {	/* typeCell == OLObjectMaskCell */
	    if (OL_transform_maskp(mask_or_type)) {
		OLExtractTransformMask(mask_or_type, &ret_tm);
	    } else {
		OLExtractTransformType(mask_or_type, &ttt);
		ret_tm = (1<<(int) ttt);
	    }

	    for (m=more_masks_or_types; LL_consp(m); m=LL_cdr(m)) {
		ma = LL_car(m);

		if (OL_transform_maskp(ma)) {
		    OLExtractTransformMask(ma, &tm);
		} else {
		    OLExtractTransformType(ma, &ttt);
		    tm = (1<<(int) ttt);
		}

		if (mergeP)
		    ret_tm |= tm;
		else
		    ret_tm &= ~tm;
	    }

	    return (OLMakeTransformMask(ret_tm));
	}

	/*NOTREACHED*/
}
