#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    close_facet,
    delete,
    flush_facet,
    free_facet,
};

static LLobject *func();

/*
 *    OLClosefacet - close-facet
 *
 *    (close-facet facet)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLClosefacet(facet)
LLobject *facet;
{
	return (func(facet, "close-facet", close_facet));
}

/*
 *    OLDelete - delete
 *
 *    (delete object)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLDelete(object)
LLobject *object;
{
	return (func(object, "delete", delete));
}

/*
 *    OLFlushfacet - flush-facet
 *
 *    (flush-facet facet)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLFlushfacet(facet)
LLobject *facet;
{
	return (func(facet, "flush-facet", flush_facet));
}

/*
 *    OLFreefacet - free-facet
 *
 *    (free-facet facet)
 *
 *    return {t, nil} on {success, fail}
 */
LLobject *
OLFreefacet(facet)
LLobject *facet;
{
	return (func(facet, "free-facet", free_facet));
}

static LLobject *
func(object, name, op)
LLobject *object;
String name;
func_op op;
{
	octObject o;

	if ( ! OL_objectp(object) ) {
	    LLErrorF("\
Function: (~A object)\n\
\tobject must be an object\n\
\tobject: ~S\n\
", name, object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(object, &o);
	switch (op) {
	case close_facet:
	    OLGlobalOctStatus = octCloseFacet(&o);
	    break;
	case delete:
	    OLGlobalOctStatus = octDelete(&o);
	    break;
	case flush_facet:
	    OLGlobalOctStatus = octFlushFacet(&o);
	    break;
	case free_facet:
	    OLGlobalOctStatus = octFreeFacet(&o);
#ifdef DONT /* there is no need (objectId's provide protection against segfaults) */
	    if (OLGlobalOctStatus >= OCT_OK) {
		generator->type->count--;
		generator->type = LLDeadCell;
		generator->type->count++;
	    }
#endif
	    break;
	}

	return (OLGlobalOctStatus < OCT_OK ? LL_nil: LL_t);
}
