#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    get_by_name,
    get_container_by_name,
};

static LLobject *func();

/*
 *    OLGetByName - get-by-name
 *
 *    (get-by-name container object-type name)
 *
 *    return {object, nil} on {success, fail}
 */
LLobject *
OLGetbyname(container, object_type, name)
LLobject *container;
LLobject *object_type;
LLobject *name;
{
	return (func(container, object_type, name, "get-by-name", "container", get_by_name));
}

/*
 *    OLGetcontainerbyname - get-container-by-name
 *
 *    (get-container-by-name content object-type name)
 *
 *    return {object, nil} on {success, fail}
 */
LLobject *
OLGetcontainerbyname(content, object_type, name)
LLobject *content;
LLobject *object_type;
LLobject *name;
{
	return (func(content, object_type, name, "get-container-by-name", "content", get_container_by_name));
}

static LLobject *
func(container, object_type, name, function_name, container_name, op)
LLobject *container;
LLobject *object_type;
LLobject *name;
String function_name;
String container_name;
func_op op;
{
	octObject c, o;
	int /* octType */ t;

	if ( ! OL_objectp(container) ) {
	    LLErrorF("\
Function: (~A ~A name)\n\
\t~A is not an object\n\
\t~A:   ~S\n\
\tobject-type: ~S\n\
\tname:        ~S\n\
", function_name, container_name, container_name, container_name, container, object_type, name);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! OL_object_typep(object_type) ) {
	    LLErrorF("\
Function: (~A ~A name)\n\
\tobject-type is not an object-type\n\
\t~A:   ~S\n\
\tobject-type: ~S\n\
\tname:        ~S\n\
", function_name, container_name, container_name, container, object_type, name);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! LL_stringp(name) ) {
	    LLErrorF("\
Function: (~A ~A name)\n\
\tname is not a string\n\
\t~A:   ~S\n\
\tobject-type: ~S\n\
\tname:        ~S\n\
", function_name, container_name, container_name, container, object_type, name);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(container, &c);
	OLExtractObjectType(object_type, &t);

        switch (t) {
        case OCT_TERM:
            o.type = OCT_TERM;
            o.contents.term.name = name->value.a_string;
            break;
        case OCT_NET:
            o.type = OCT_NET;
            o.contents.net.name = name->value.a_string;
            break;
        case OCT_INSTANCE:
            o.type = OCT_INSTANCE;
            o.contents.instance.name = name->value.a_string;
            break;
        case OCT_PROP:
            o.type = OCT_PROP;
            o.contents.prop.name = name->value.a_string;
            break;
        case OCT_LAYER:
            o.type = OCT_LAYER;
            o.contents.layer.name = name->value.a_string;
            break;
        case OCT_BAG:
            o.type = OCT_BAG;
            o.contents.bag.name = name->value.a_string;
            break;
        case OCT_LABEL:
            o.type = OCT_LABEL;
            o.contents.label.label = name->value.a_string;
            break;
        default:
	    LLErrorF("\
Function: (~A ~A name)\n\
\tthe object does not have a name defined for it\n\
\t~A:   ~S\n\
\tobject-type: ~S\n\
\tname:        ~S\n\
", function_name, container_name, container_name, container, object_type, name);
	    LLThrow(OL_octlisp, LL_nil);
        }

	if (op == get_by_name) {
	    OLGlobalOctStatus = octGetByName(&c, &o);
	} else {
	    OLGlobalOctStatus = octGetContainerByName(&c, &o);
	}

	return (OLGlobalOctStatus < OCT_OK ? LL_nil: OLMakeObject(o));
}
