#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "oct_helperlib.h"
#include "version.h"
#include "internal.h"

static struct {
    LLobjectType **place;	/* place to put the objectType */
    LLobjectType template;
} standard_templates[] = {
    { &OLBoxCell,
	{ "box",	"box",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_Box1, OL_Pr_Box1, OL_Pr_Box1 },
		{ OL_Pr_Box, OL_Pr_Box, OL_Pr_Box },
		{ OL_Pr_Box, OL_Pr_Box, OL_Pr_Box },
		{ OL_Pr_Box, OL_Pr_Box, OL_Pr_Box },
		{ OL_GcScan_Box,	OL_GcReScan_Box }, NULL } },
    { &OLGeneratorCell,
	{ "generator",	"generator",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_Generator, OL_Pr_Generator, OL_Pr_Generator },
		{ OL_Pr_Generator, OL_Pr_Generator, OL_Pr_Generator },
		{ OL_Pr_Generator, OL_Pr_Generator, OL_Pr_Generator },
		{ OL_Pr_Generator, OL_Pr_Generator, OL_Pr_Generator },
		{ OL_GcScan_Generator,	OL_GcReScan_Generator }, NULL } },
    { &OLGenerator2Cell,
	{ "generator2",	"generator2",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ NULL, NULL, NULL },
		{ NULL, NULL, NULL },
		{ NULL, NULL, NULL },
		{ NULL, NULL, NULL },
		{ OL_GcScan_Generator2,	NULL }, NULL } },
    { &OLObjectCell,
	{ "object",	"object",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_Object1, OL_Pr_Object1, OL_Pr_Object1 },
		{ OL_Pr_Object, OL_Pr_Object, OL_Pr_Object },
		{ OL_Pr_Object, OL_Pr_Object, OL_Pr_Object },
		{ OL_Pr_Object2, OL_Pr_Object2, OL_Pr_Object2 },
		{ OL_GcScan_Object,	NULL }, NULL } },
    { &OLObjectMaskCell,
	{ "object-mask",	"object-mask",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_ObjectMask, OL_Pr_ObjectMask, OL_Pr_ObjectMask },
		{ OL_Pr_ObjectMask, OL_Pr_ObjectMask, OL_Pr_ObjectMask },
		{ OL_Pr_ObjectMask, OL_Pr_ObjectMask, OL_Pr_ObjectMask },
		{ OL_Pr_ObjectMask, OL_Pr_ObjectMask, OL_Pr_ObjectMask },
		{ OL_GcScan_ObjectMask,	NULL }, NULL } },
    { &OLObjectTypeCell,
	{ "object-type",	"object-type",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_ObjectType, OL_Pr_ObjectType, OL_Pr_ObjectType },
		{ OL_Pr_ObjectType, OL_Pr_ObjectType, OL_Pr_ObjectType },
		{ OL_Pr_ObjectType, OL_Pr_ObjectType, OL_Pr_ObjectType },
		{ OL_Pr_ObjectType, OL_Pr_ObjectType, OL_Pr_ObjectType },
		{ OL_GcScan_ObjectType,	NULL }, NULL } },
    { &OLPointCell,
	{ "point",	"point",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_Point1, OL_Pr_Point1, OL_Pr_Point1 },
		{ OL_Pr_Point, OL_Pr_Point, OL_Pr_Point },
		{ OL_Pr_Point, OL_Pr_Point, OL_Pr_Point },
		{ OL_Pr_Point, OL_Pr_Point, OL_Pr_Point },
		{ OL_GcScan_Point,	NULL }, NULL } },
    { &OLPropertyTypeCell,
	{ "property-type",	"property-type",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_PropertyType, OL_Pr_PropertyType, OL_Pr_PropertyType },
		{ OL_Pr_PropertyType, OL_Pr_PropertyType, OL_Pr_PropertyType },
		{ OL_Pr_PropertyType, OL_Pr_PropertyType, OL_Pr_PropertyType },
		{ OL_Pr_PropertyType, OL_Pr_PropertyType, OL_Pr_PropertyType },
		{ OL_GcScan_PropertyType,	NULL }, NULL } },
    { &OLOctStatusCell,
	{ "oct-status",	"oct-status",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_OctStatus, OL_Pr_OctStatus, OL_Pr_OctStatus },
		{ OL_Pr_OctStatus, OL_Pr_OctStatus, OL_Pr_OctStatus },
		{ OL_Pr_OctStatus, OL_Pr_OctStatus, OL_Pr_OctStatus },
		{ OL_Pr_OctStatus, OL_Pr_OctStatus, OL_Pr_OctStatus },
		{ OL_GcScan_OctStatus,	NULL }, NULL } },
    { &OLTransformCell,
	{ "transform",	"transform",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_Transform1, OL_Pr_Transform1, OL_Pr_Transform1 },
		{ OL_Pr_Transform, OL_Pr_Transform, OL_Pr_Transform },
		{ OL_Pr_Transform, OL_Pr_Transform, OL_Pr_Transform },
		{ OL_Pr_Transform, OL_Pr_Transform, OL_Pr_Transform },
		{ OL_GcScan_Transform,	OL_GcReScan_Transform }, NULL } },
    { &OLTransformMaskCell,
	{ "transform-mask",	"transform-mask",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_TransformMask, OL_Pr_TransformMask, OL_Pr_TransformMask },
		{ OL_Pr_TransformMask, OL_Pr_TransformMask, OL_Pr_TransformMask },
		{ OL_Pr_TransformMask, OL_Pr_TransformMask, OL_Pr_TransformMask },
		{ OL_Pr_TransformMask, OL_Pr_TransformMask, OL_Pr_TransformMask },
		{ OL_GcScan_TransformMask,	NULL }, NULL } },
    { &OLTransformTypeCell,
	{ "transform-type",	"transform-type",	NULL,
		{ LL_NON_FUNCTION, 0x0, 0x0 }, 0, 
		{ OL_Pr_TransformType, OL_Pr_TransformType, OL_Pr_TransformType },
		{ OL_Pr_TransformType, OL_Pr_TransformType, OL_Pr_TransformType },
		{ OL_Pr_TransformType, OL_Pr_TransformType, OL_Pr_TransformType },
		{ OL_Pr_TransformType, OL_Pr_TransformType, OL_Pr_TransformType },
		{ OL_GcScan_TransformType,	NULL }, NULL } },
};

static void scan();

static struct {
    String name;
    int /* octType */ type;
} object_types[] = {
    { "OCT-FACET",	OCT_FACET	},
    { "OCT-TERM",	OCT_TERM	},
    { "OCT-NET",	OCT_NET		},
    { "OCT-INSTANCE",	OCT_INSTANCE	},
    { "OCT-POLYGON",	OCT_POLYGON	},
    { "OCT-BOX",	OCT_BOX		},
    { "OCT-CIRCLE",	OCT_CIRCLE	},
    { "OCT-PATH",	OCT_PATH	},
    { "OCT-LABEL",	OCT_LABEL	},
    { "OCT-PROP",	OCT_PROP	},
    { "OCT-BAG",	OCT_BAG		},
    { "OCT-LAYER",	OCT_LAYER	},
    { "OCT-POINT",	OCT_POINT	},
    { "OCT-EDGE",	OCT_EDGE	},
};

static struct {
    String name;
    octObjectMask mask;
} object_masks[] = {
    { "OCT-FACET-MASK",		OCT_FACET_MASK		},
    { "OCT-TERM-MASK",		OCT_TERM_MASK		},
    { "OCT-NET-MASK",		OCT_NET_MASK		},
    { "OCT-INSTANCE-MASK",	OCT_INSTANCE_MASK	},
    { "OCT-PROP-MASK",		OCT_PROP_MASK		},
    { "OCT-BAG-MASK",		OCT_BAG_MASK		},
    { "OCT-POLYGON-MASK",	OCT_POLYGON_MASK	},
    { "OCT-BOX-MASK",		OCT_BOX_MASK		},
    { "OCT-CIRCLE-MASK",	OCT_CIRCLE_MASK		},
    { "OCT-PATH-MASK",		OCT_PATH_MASK		},
    { "OCT-LABEL-MASK",		OCT_LABEL_MASK		},
    { "OCT-LAYER-MASK",		OCT_LAYER_MASK		},
    { "OCT-POINT-MASK",		OCT_POINT_MASK		},
    { "OCT-EDGE-MASK",		OCT_EDGE_MASK		},
    { "OCT-GEO-MASK",		OCT_GEO_MASK		},
    { "OCT-ALL-MASK",		OCT_ALL_MASK		},
};

static struct {
    String name;
    octPropType type;
} property_types[] = {
    { "OCT-NULL",		OCT_NULL		},
    { "OCT-INTEGER",		OCT_INTEGER		},
    { "OCT-REAL",		OCT_REAL		},
    { "OCT-STRING",		OCT_STRING		},
    { "OCT-ID",			OCT_ID			},
    { "OCT-STRANGER",		OCT_STRANGER		},
    { "OCT-REAL-ARRAY",		OCT_REAL_ARRAY		},
    { "OCT-INTEGER-ARRAY",	OCT_INTEGER_ARRAY	},
};

static struct {
    String name;
    octStatus status;
} stati[] = {
    { "OCT-OK",			OCT_OK			},
    { "OCT-NEW-FACET",		OCT_NEW_FACET		},
    { "OCT-OLD-FACET",		OCT_OLD_FACET		},
    { "OCT-GEN-DONE",		OCT_GEN_DONE		},
    { "OCT-ALREADY-OPEN",	OCT_ALREADY_OPEN	},
    { "OCT-NOT-ATTACHED",	OCT_NOT_ATTACHED	},
    { "OCT-ALREADY-ATTACHED",	OCT_ALREADY_ATTACHED	},
    { "OCT-ERROR",		OCT_ERROR		},
    { "OCT-NO-EXIST",		OCT_NO_EXIST		},
    { "OCT-NO-PERM",		OCT_NO_PERM		},
    { "OCT-CORRUPTED-OBJECT",	OCT_CORRUPTED_OBJECT	},
    { "OCT-NOT-FOUND",		OCT_NOT_FOUND		},
    { "OCT-TOO-SMALL",		OCT_TOO_SMALL		},
    { "OCT-ILL-OP",		OCT_ILL_OP		},
    { "OCT-NO-BB",		OCT_NO_BB		},
};

static struct {
    String name;
    octTransformType transformType;
} transform_types[] = {
    { "OCT-NO-TRANSFORM",	OCT_NO_TRANSFORM	},
    { "OCT-MIRROR-X",		OCT_MIRROR_X		},
    { "OCT-MIRROR-Y",		OCT_MIRROR_Y		},
    { "OCT-ROT90",		OCT_ROT90		},
    { "OCT-ROT180",		OCT_ROT180		},
    { "OCT-ROT270",		OCT_ROT270		},
    { "OCT-MX-ROT90",		OCT_MX_ROT90		},
    { "OCT-MY-ROT90",		OCT_MY_ROT90		},
    { "OCT-FULL-TRANSFORM",	OCT_FULL_TRANSFORM	},
};

static struct {
    String name;
    int /* octTransformMask */ transformMask;
} transform_masks[] = {
    { "OCT-NO-TRANSFORM-MASK",		(1<< (int) OCT_NO_TRANSFORM)	},
    { "OCT-MIRROR-X-MASK",		(1<< (int) OCT_MIRROR_X)	},
    { "OCT-MIRROR-Y-MASK",		(1<< (int) OCT_MIRROR_Y)	},
    { "OCT-ROT90-MASK",			(1<< (int) OCT_ROT90)		},
    { "OCT-ROT180-MASK",		(1<< (int) OCT_ROT180)		},
    { "OCT-ROT270-MASK",		(1<< (int) OCT_ROT270)		},
    { "OCT-MX-ROT90-MASK",		(1<< (int) OCT_MX_ROT90)	},
    { "OCT-MY-ROT90-MASK",		(1<< (int) OCT_MY_ROT90)	},
    { "OCT-FULL-TRANSFORM-MASK",	(1<< (int) OCT_FULL_TRANSFORM)	},
};

/*
 *    LLobject *
 *    function(args ... )
 *    LLobject *args;
 */
static struct {
    String pname;
    LLobject *(*function)();
    int nrequired;
    int noptional;
    Boolean restP;
} functions[] = {
    { "attach",			OLAttach,		2,	0,	FALSE	},
    { "attach-once",		OLAttachonce,		2,	0,	FALSE	},
    { "bb",			OLBb,			1,	0,	FALSE	},
    { "box",			OLBox,			2,	0,	FALSE	},
    { "boxp",			OLBoxp,			1,	0,	FALSE	},
    { "box-of",			OLBoxof,		1,	0,	FALSE	},
    { "breadth-first-traversal",OLBreadthfirsttraversal,2,	0,	FALSE	},
    { "cell-of",		OLCellof,		1,	0,	FALSE	},
    { "center-of",		OLCenterof,		1,	0,	FALSE	},
    { "close-facet",		OLClosefacet,		1,	0,	FALSE	},
    { "copy-facet",		OLCopyfacet,		2,	0,	FALSE	},
    { "create",			OLCreate,		2,	0,	FALSE	},
#if ! defined(OCT_JUST_RIGHT) /* { */
/* oct 1.0 */
    { "create-and-attach",	OLCreateandattach,	2,	0,	FALSE	},
    { "create-not-attached",	OLCreatenotattached,	2,	0,	FALSE	},
#endif /* } */
    { "create-or-modify",	OLCreateormodify,	2,	0,	FALSE	},
    { "delete",			OLDelete,		1,	0,	FALSE	},
    { "detach",			OLDetach,		2,	0,	FALSE	},
    { "depth-first-traversal",	OLDepthfirsttraversal,	2,	0,	FALSE	},
    { "end-of",			OLEndof,		1,	0,	FALSE	},
    { "endingAngle-of",		OLEndingangleof,	1,	0,	FALSE	},
    { "extract-masks",		OLExtractmasks,		1,	0,	TRUE	},
    { "facet-of",		OLFacetof,		1,	0,	FALSE	},
    { "flush-facet",		OLFlushfacet,		1,	0,	FALSE	},
    { "free-facet",		OLFreefacet,		1,	0,	FALSE	},
    { "free-generator",		OLFreegenerator,	1,	0,	FALSE	},
    { "gen-first-container",	OLGenfirstcontainer,	2,	0,	FALSE	},
    { "gen-first-content",	OLGenfirstcontent,	2,	0,	FALSE	},
    { "generalTransform-of",	OLGeneraltransformof,	1,	0,	FALSE	},
    { "generate",		OLGenerate,		1,	0,	FALSE	},
    { "generatorp",		OLGeneratorp,		1,	0,	FALSE	},
    { "get-by-name",		OLGetbyname,		3,	0,	FALSE	},
    { "get-container-by-name",	OLGetcontainerbyname,	3,	0,	FALSE	},
    { "get-facet",		OLGetfacet,		1,	0,	FALSE	},
    { "get-name",		OLGetname,		1,	0,	FALSE	},
    { "get-or-create",		OLGetorcreate,		2,	0,	FALSE	},
    { "init-gen-containers",	OLInitgencontainers,	2,	0,	FALSE	},
    { "init-gen-contents",	OLInitgencontents,	2,	0,	FALSE	},
    { "innerRadius-of",		OLInnerradiusof,	1,	0,	FALSE	},
    { "instance-of",		OLInstanceof,		1,	0,	FALSE	},
    { "is-attached",		OLIsattached,		2,	0,	FALSE	},
    { "label-of",		OLLabelof,		1,	0,	FALSE	},
    { "lowerLeft-of",		OLLowerleftof,		1,	0,	FALSE	},
    { "mask-member",		OLMaskmember,		2,	0,	TRUE	},
    { "master-of",		OLMasterof,		1,	0,	FALSE	},
    { "merge-masks",		OLMergemasks,		1,	0,	TRUE	},
    { "mode-of",		OLModeof,		1,	0,	FALSE	},
    { "modify",			OLModify,		2,	0,	FALSE	},
    { "name-of",		OLNameof,		1,	0,	FALSE	},
    { "objectp",		OLObjectp,		1,	0,	FALSE	},
    { "object-form-p",		OLObjectformp,		1,	0,	FALSE	},
    { "object-mask-p",		OLObjectmaskp,		1,	0,	FALSE	},
    { "object-of",		OLObjectof,		1,	0,	FALSE	},
    { "object-type-of",		OLObjecttypeof,		1,	0,	FALSE	},
    { "object-type-p",		OLObjecttypep,		1,	0,	FALSE	},
    { "oct-error-status",	OLOcterrorstatus,	0,	0,	FALSE	},
    { "oct-error-string",	OLOcterrorstring,	0,	0,	FALSE	},
    { "oct-status-p",		OLOctstatusp,		1,	0,	FALSE	},
    { "open-facet",		OLOpenfacet,		2,	0,	FALSE	},
    { "open-facet-by-name",	OLOpenfacetbyname,	5,	0,	FALSE	},
    { "open-facet-pair",	OLOpenfacetpair,	3,	0,	FALSE	},
    { "open-master",		OLOpenmaster,		3,	0,	FALSE	},
    { "outerRadius-of",		OLOuterradiusof,	1,	0,	FALSE	},
    { "point",			OLPoint,		2,	0,	FALSE	},
    { "pointp",			OLPointp,		1,	0,	FALSE	},
    { "point-of",		OLPointof,		1,	0,	FALSE	},
    { "points-of",		OLPointsof,		1,	0,	FALSE	},
#if ! defined(OCT_JUST_RIGHT) /* { */
    { "position-of",		OLPositionof,		1,	0,	FALSE	},
#endif /* } */
    { "property-type-of",	OLPropertytypeof,	1,	0,	FALSE	},
    { "property-type-p",	OLPropertytypep,	1,	0,	FALSE	},
    { "start-of",		OLStartof,		1,	0,	FALSE	},
    { "startingAngle-of",	OLStartingangleof,	1,	0,	FALSE	},
    { "transform",		OLTransform,		2,	1,	FALSE	},
    { "transformp",		OLTransformp,		1,	0,	FALSE	},
    { "transform-mask-p",	OLTransformmaskp,	1,	0,	FALSE	},
    { "transform-of",		OLTransformof,		1,	0,	FALSE	},
    { "transform-type-p",	OLTransformtypep,	1,	0,	FALSE	},
    { "transformType-of",	OLTransformtypeof,	1,	0,	FALSE	},
    { "translation-of",		OLTranslationof,	1,	0,	FALSE	},
    { "upperRight-of",		OLUpperrightof,		1,	0,	FALSE	},
    { "value-of",		OLValueof,		1,	0,	FALSE	},
    { "version-of",		OLVersionof,		1,	0,	FALSE	},
    { "view-of",		OLViewof,		1,	0,	FALSE	},
    { "width-of",		OLWidthof,		1,	0,	FALSE	},
    { "x-of",			OLXof,			1,	0,	FALSE	},
    { "y-of",			OLYof,			1,	0,	FALSE	},
};

/*
 *    OLInitialize - initialize OctLisp
 *
 *    return {OL_ok, OL_notok} on {success, fail}
 */
OLstatus
OLInitialize()
{
	int i;

	if (OLInitializedP)
	    return (OL_ok);

	if (LLInitialize() != LL_ok)
	    return (OL_notok);

	if (sizeof (octGenerator) != 12) {
	    /*
	     *    To fix this, just go into generator_cell.c
	     *    and fix up the OLExtractGenerator function
	     *    and the garbage_collector function.  You will
	     *    probably have to add OLGenerator3Cell, OLGenerator4Cell etc.
	     */
	    LLErrorF("\
OLInitialize - INTERNAL ERROR - the size of the octGenerator has changed from 12 to ~D\n\
", sizeof (octGenerator));
	    return (OL_notok);
	}

	/* set this FIRST */
	OLNullObject.objectId = oct_null_id;

	OLObject = LLInternSymbol(OLObject);

        for (i=0; i<sizeof_el(standard_templates); i++) {
	    *standard_templates[i].place = LLMakeObjectType(standard_templates[i].template);
        }

	OLMachine.objects = st_init_table(st_numcmp, st_numhash);
	for (i=0; i<sizeof_el (OLMachine.object_types); i++)
	    OLMachine.object_types[i] = LL_nil;
	for (i=0; i<sizeof_el (OLMachine.stati); i++)
	    OLMachine.stati[i] = LL_nil;
	for (i=0; i<sizeof_el (OLMachine.property_types); i++)
	    OLMachine.property_types[i] = LL_nil;
	for (i=0; i<sizeof_el (OLMachine.transform_types); i++)
	    OLMachine.transform_types[i] = LL_nil;

	OLMachine.object = LLMakeSymbol(OLObject);
	OLMachine.octlisp = LLMakeSymbol("octlisp");

	LLMakeRegister(scan, &OLMachine);

	for (i=0; i<sizeof_el (object_types); i++) {
	    LLSetValue(LLMakeSymbol(object_types[i].name), LL_valueSlot,
		    OLMakeObjectType(object_types[i].type), LL_environment);
	}
	for (i=0; i<sizeof_el (object_masks); i++) {
	    LLSetValue(LLMakeSymbol(object_masks[i].name), LL_valueSlot,
		    OLMakeObjectMask(object_masks[i].mask), LL_environment);
	}
	for (i=0; i<sizeof_el (property_types); i++) {
	    LLSetValue(LLMakeSymbol(property_types[i].name), LL_valueSlot,
		    OLMakePropertyType(property_types[i].type), LL_environment);
	}
	for (i=0; i<sizeof_el (stati); i++) {
	    LLSetValue(LLMakeSymbol(stati[i].name), LL_valueSlot,
		    OLMakeOctStatus(stati[i].status), LL_environment);
	}
	for (i=0; i<sizeof_el (transform_types); i++) {
	    LLSetValue(LLMakeSymbol(transform_types[i].name), LL_valueSlot,
		    OLMakeTransformType(transform_types[i].transformType), LL_environment);
	}
	for (i=0; i<sizeof_el (transform_masks); i++) {
	    LLSetValue(LLMakeSymbol(transform_masks[i].name), LL_valueSlot,
		    OLMakeTransformMask(transform_masks[i].transformMask), LL_environment);
	}

	LLSetValue(LLMakeSymbol("OCT-CURRENT-VERSION"), LL_valueSlot,
		LLMakeString(OCT_CURRENT_VERSION), LL_environment);
	LLSetValue(LLMakeSymbol("oct-null-id"), LL_valueSlot,
		OLMakeObject(OLNullObject), LL_environment);
	LLSetValue(LLMakeSymbol("lisp-implementation-type"), LL_valueSlot,
		LLMakeString(LISP_IMPLEMENTATION_TYPE), LL_environment);
	{
		LLobject *version = LLMakeSymbol("lisp-implementation-version");
		char version_buf[40960];

		sprintf(version_buf, "%s\n%s", 
			LLGetValue(version, LL_valueSlot, LL_environment)->value.a_string,
			LISP_IMPLEMENTATION_VERSION);
		LLSetValue(version, LL_valueSlot, LLMakeString(version_buf), LL_environment);
	}

	for (i=0; i<sizeof_el (functions); i++) {
	    LLSetValue(LLMakeSymbol(functions[i].pname),
		    LL_functionSlot,
		    LLMakeFunction(functions[i].pname, functions[i].function,
			    functions[i].nrequired,
			    functions[i].noptional,
			    functions[i].restP),
		    LL_nil);
	}

	OLGlobalOctStatus = OCT_OK;

	if (OL_Help_Initialize() != OL_ok)
	    return (OL_notok);
	if (OL_Octh_Initialize() != OL_ok)
	    return (OL_notok);

	OLInitializedP = TRUE;

	return (OL_ok);
}

static void
scan(machine, relocate)
OLmachine *machine;
LLobject *(*relocate)();
{
	st_generator *g;
	octId id;
	int i;
	LLobject *o;

	g = st_init_gen(machine->objects);
	while (st_gen(g, &id, &o))
	    st_insert(machine->objects, id, (*relocate)(o));
	st_free_gen(g);

	for (i=0; i<sizeof_el (machine->object_types); i++)
	    machine->object_types[i] = (*relocate)(machine->object_types[i]);

	for (i=0; i<sizeof_el (machine->stati); i++)
	    machine->stati[i] = (*relocate)(machine->stati[i]);

	for (i=0; i<sizeof_el(machine->transform_types); i++)
	    machine->transform_types[i] = (*relocate)(machine->transform_types[i]);

	machine->object = (*relocate)(machine->object);
	machine->octlisp = (*relocate)(machine->octlisp);
}
