#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    init_gen_containers,
    init_gen_contents,
};

static LLobject *func();

/*
 *    OLInitgencontainers - init-gen-containers
 *
 *    (init-gen-containers content mask)
 *
 *    return the {generator, nil} on {success, fail}
 */
LLobject *
OLInitgencontainers(content, mask)
LLobject *content;
LLobject *mask;
{
	return (func(content, mask, "init-gen-containers", "content", init_gen_containers));
}

/*
 *    OLInitgencontents - init-gen-contents
 *
 *    (init-gen-contents container mask)
 *
 *    return the {generator, nil} on {success, fail}
 */
LLobject *
OLInitgencontents(container, mask)
LLobject *container;
LLobject *mask;
{
	return (func(container, mask, "init-gen-contents", "containner", init_gen_contents));
}

static LLobject *
func(container, mask, name, container_name, op)
LLobject *container;
LLobject *mask;
String name;
String container_name;
func_op op;
{
	octObject c;
	octObjectMask m;
	octGenerator g;

	if ( ! OL_objectp(container) ) {
	    LLErrorF("\
Function: (~A ~A mask)\n\
\t~A is not an object\n\
\t~A: ~S\n\
\tmask:      ~S\n\
", name, container_name, container_name, container, mask);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! OL_object_maskp(mask) ) {
	    LLErrorF("\
Function: (~A ~A mask)\n\
\tmask is not an object-mask\n\
\t~A: ~S\n\
\tmask:      ~S\n\
", name, container_name, container_name, container, mask);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(container, &c);
	OLExtractObjectMask(mask, &m);

	if (op == init_gen_containers) {
	    OLGlobalOctStatus = octInitGenContainers(&c, m, &g);
	} else {
	    OLGlobalOctStatus = octInitGenContents(&c, m, &g);
	}

	return (OLGlobalOctStatus < OCT_OK ? LL_nil: OLMakeGenerator(g));
}
