#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLIntegersof - integers-of
 *
 *    (integers-of object)
 *
 *    return the list of integers
 */
LLobject *
OLIntegersof(object)
LLobject *object;
{
	octObject o;
	int i;
	LLobject *list, *last, *tmp;

	if ( ! OL_objectp(object) ) {
	    LLErrorF("\
Function: (integers-of object)\n\
\tobject must be an object\n\
\tobject: ~S\n\
", object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(object, &o);

	if (o.type != OCT_PROP || o.contents.prop.type != OCT_INTEGER_ARRAY) {
	    LLErrorF("\
Function: (integers-of object)\n\
\tobject must be an OCT-PROP of type OCT-INTEGER-ARRAY\n\
\tobject: ~S\n\
", object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	list = LL_nil;
	for (i=0; i<o.contents.prop.value.integer_array.length; i++) {
	    tmp = LLCons(LLMakeInteger(o.contents.prop.value.integer_array.array[i]), LL_nil);
	    if (LL_null(list))
		list = tmp;
	    else
		LL_rplacd(last, tmp);
	    last = tmp;
	}

	return (list);
}

/*
 *    OLRealsof - reals-of
 *
 *    (reals-of object)
 *
 *    return the list of reals
 */
LLobject *
OLRealsof(object)
LLobject *object;
{
	octObject o;
	int i;
	LLobject *list, *last, *tmp;

	if ( ! OL_objectp(object) ) {
	    LLErrorF("\
Function: (reals-of object)\n\
\tobject must be an object\n\
\tobject: ~S\n\
", object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(object, &o);

	if (o.type != OCT_PROP || o.contents.prop.type != OCT_REAL_ARRAY) {
	    LLErrorF("\
Function: (reals-of object)\n\
\tobject must be an OCT-PROP of type OCT-REAL-ARRAY\n\
\tobject: ~S\n\
", object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	list = LL_nil;
	for (i=0; i<o.contents.prop.value.integer_array.length; i++) {
	    tmp = LLCons(LLMakeReal(o.contents.prop.value.real_array.array[i]), LL_nil);
	    if (LL_null(list))
		list = tmp;
	    else
		LL_rplacd(last, tmp);
	    last = tmp;
	}

	return (list);
}
