#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    for internal purposes only (mostly for the printers)
 *    in case we need to get to extract a object without
 *    the risk of being thrown.
 */
#define extract_object(obj, object)	\
	((object)->objectId = (obj)->value.a_userDefined.value[0])

void
OLExtractObject(obj, object)
LLobject *obj;
octObject *object;	/* return */
{
	/*
	 *    A userDefined.
	 *
	 *    obj->value.a_userDefined.value[0] = objectId
	 */

	extract_object(obj, object);
	if (object->objectId != oct_null_id) {
	    if (octGetById(object) < OCT_OK) {
		LLErrorF("\
Function: OLExtractObject(obj, object)\n\
\tinvalid object because ~A\n\
\tobject: <objectId ~D>\n\
", octErrorString(), object->objectId);
		LLThrow(OL_octlisp, LL_nil);
	    }
        }
}

LLobject *
OLMakeObject(oct_object)
octObject oct_object;
{
	LLobject *new, *object;

	if (st_lookup(OLMachine.objects, oct_object.objectId, &object))
	    return (object);

	new = LLMakeObject();
	new->type = OLObjectCell;
	new->type->count++;

	new->value.a_userDefined.value[0] = (FitAny) oct_object.objectId;

	st_add_direct(OLMachine.objects, oct_object.objectId, new);

	return (new);
}

void
OL_Pr_Object(obj, stream)
LLobject *obj;
STREAM *stream;
{
	octObject o;

	extract_object(obj, &o);
	if (o.objectId == oct_null_id) {
	    LLFormatC(stream, "oct-null-id");
	} else if (octGetById(&o) < OCT_OK) {
	    LLFormatC(stream, "<invalid objectId ~D>", o.objectId);
	} else {
	    LLFormatC(stream, "<objectId ~D>", o.objectId);
	}
}

void
OL_Pr_Object1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	octObject o;

	extract_object(obj, &o);
	if (o.objectId == oct_null_id) {
	    LLFormatC(stream, "oct-null-id");
	} else if (octGetById(&o) < OCT_OK) {
	    LLFormatC(stream, "'(invalid-objectId ~D)", o.objectId);
	} else {
	    LLFormatC(stream, "'~S", OLObjectof(obj));
	}
}

void
OL_Pr_Object2(obj, stream)
LLobject *obj;
STREAM *stream;
{
	octObject o;

	extract_object(obj, &o);
	if (o.objectId == oct_null_id) {
	    LLFormatC(stream, "oct-null-id");
	} else if (octGetById(&o) < OCT_OK) {
	    LLFormatC(stream, "<invalid-objectId ~D>", o.objectId);
	} else {
	    LLFormatC(stream, "~A", octhStringifyObject(&o));
	}
}

LLobject *
OL_GcScan_Object(old)
LLobject *old;
{
	LLobject *new;

	new = LLMakeObject();

	new->type = old->type;
	new->value.a_userDefined = old->value.a_userDefined;
	new->type->count++;

	return (new);
}
