#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

void
OLExtractObjectMask(obj, mask)
LLobject *obj;
octObjectMask *mask;	/* return */
{
	LLobject *one, *two;

	/*
	 *    A userDefined.
	 *
	 *    obj->value.a_userDefined.value[0] = objectId
	 */

	*mask = (octObjectMask) obj->value.a_userDefined.value[0];
}

LLobject *
OLMakeObjectMask(mask)
octObjectMask mask;
{
	LLobject *new;

	new = LLMakeObject();
	new->type = OLObjectMaskCell;
	new->type->count++;

	/* in case of throw below */
	new->value.a_userDefined.value[0] = (FitAny) mask;

	return (new);
}

static struct {
    octObjectMask mask;
    String name;
} map[] = {
    { OCT_FACET_MASK,		"OCT-FACET-MASK"	},
    { OCT_TERM_MASK,		"OCT-TERM-MASK"		},
    { OCT_NET_MASK,		"OCT-NET-MASK"		},
    { OCT_INSTANCE_MASK,	"OCT-INSTANCE-MASK"	},
    { OCT_PROP_MASK,		"OCT-PROP-MASK"		},
    { OCT_BAG_MASK,		"OCT-BAG-MASK"		},
    { OCT_POLYGON_MASK,		"OCT-POLYGON-MASK"	},
    { OCT_BOX_MASK,		"OCT-BOX-MASK"		},
    { OCT_CIRCLE_MASK,		"OCT-CIRCLE-MASK"	},
    { OCT_PATH_MASK,		"OCT-PATH-MASK"		},
    { OCT_LABEL_MASK,		"OCT-LABEL-MASK"	},
    { OCT_LAYER_MASK,		"OCT-LAYER-MASK"	},
    { OCT_POINT_MASK,		"OCT-POINT-MASK"	},
    { OCT_EDGE_MASK,		"OCT-EDGE-MASK"		},
};

void
OL_Pr_ObjectMask(obj, stream)
LLobject *obj;
STREAM *stream;
{
	octObjectMask m, cmask;
        int i, count;

	OLExtractObjectMask(obj, &m);

        if (m == 0x0) {
            LLFormatC(stream, "(extract-masks OCT-ALL-MASK OCT-ALL-MASK)");
	    return;
	}

        if (m == OCT_ALL_MASK) {
            LLFormatC(stream, "OCT-ALL-MASK");
	    return;
	}

        if (m == OCT_GEO_MASK) {
            LLFormatC(stream, "OCT-GEO-MASK");
	    return;
	}

        for (count=0, cmask=m, i=0; i<32; i++, cmask>>=1)
            count += (cmask & 01) == 01 ? 1: 0;

        if (count > 1)
            LLFormatC(stream, "(merge-masks");

        if ((m & OCT_GEO_MASK) == OCT_GEO_MASK) {
            LLFormatC(stream, " OCT-GEO-MASK");
	}
        for (i=0; i<sizeof_el (map); i++) {
            if ((map[i].mask & m) != 0x0)
		LLFormatC(stream, " ~A", map[i].name);
        }

        if (count > 1)
            LLFormatC(stream, ")");
}

LLobject *
OL_GcScan_ObjectMask(old)
LLobject *old;
{
	LLobject *new;

	new = LLMakeObject();

	new->type = old->type;
	new->value.a_userDefined = old->value.a_userDefined;
	new->type->count++;

	return (new);
}
