#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLOpenfacetbyname - open-facet-by-name
 *
 *    (open-facet-by-name cell view facet version mode)
 *
 *    return {facet, nil} on {success, fail}
 */
LLobject *
OLOpenfacetbyname(cell, view, facet, version, mode)
LLobject *cell;
LLobject *view;
LLobject *facet;
LLobject *version;
LLobject *mode;
{
	octObject f;

	if ( ! LL_stringp(cell) || ! LL_stringp(view) ||
		! LL_stringp(facet) || ! LL_stringp(version) ) {
	    LLErrorF("\
Function: (open-facet-by-name cell view facet version mode)\n\
\tall of cell view facet version must be strings\n\
\tcell:    ~S\n\
\tview:    ~S\n\
\tfacet:   ~S\n\
\tversion: ~S\n\
\tmode:    ~S\n\
", cell, view, facet, version, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! LL_symbolp(mode) ||
		(LL_pname(mode) != OLReadonly &&
		 LL_pname(mode) != OLReadwrite &&
		 LL_pname(mode) != OLReadwriteappend) ) {
	    LLErrorF("\
Function: (open-facet-by-name cell view facet version mode)\n\
\tmode must be a symbol :readonly :readwrite or :readwrite-append\n\
\tcell:    ~S\n\
\tview:    ~S\n\
\tfacet:   ~S\n\
\tversion: ~S\n\
\tmode:    ~S\n\
", cell, view, facet, version, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}

	f.type = OCT_FACET;
	f.contents.facet.cell = cell->value.a_string;
	f.contents.facet.view = view->value.a_string;
	f.contents.facet.facet = facet->value.a_string;
	f.contents.facet.version = version->value.a_string;
	if (LL_pname(mode) == OLReadonly)
	    f.contents.facet.mode = "r";
	else if (LL_pname(mode) == OLReadwrite)
	    f.contents.facet.mode = "w";
	else if (LL_pname(mode) == OLReadwriteappend)
	    f.contents.facet.mode = "a";

	OLGlobalOctStatus = octOpenFacet(&f);
	return (OLGlobalOctStatus < OCT_OK ? LL_nil: OLMakeObject(f));
}
