#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLOpenfacetpair - open-facet-pair
 *
 *    (open-facet-pair facet alternate mode)
 *
 *    return {facet, nil} on {success, fail}
 */
LLobject *
OLOpenfacetpair(facet, alternate, mode)
LLobject *facet;
LLobject *alternate;
LLobject *mode;
{
	octObject f, a;

	if ( ! OL_objectp(facet) && ! OL_object_formp(facet) ) {
	    LLErrorF("\
Function: (open-facet-pair facet alternate mode)\n\
\tfacet must be an object\n\
\tfacet:     ~S\n\
\talternate: ~S\n\
\tmode:      ~S\n\
", facet, alternate, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! LL_stringp(alternate) ) {
	    LLErrorF("\
Function: (open-facet-pair facet alternate mode)\n\
\talternate must be a string\n\
\tfacet:     ~S\n\
\talternate: ~S\n\
\tmode:      ~S\n\
", facet, alternate, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! LL_symbolp(mode) ||
		(LL_pname(mode) != OLReadonly &&
		 LL_pname(mode) != OLReadwrite &&
		 LL_pname(mode) != OLReadwriteappend) ) {
	    LLErrorF("\
Function: (open-facet-pair facet alternate mode)\n\
\tmode must be a symbol :readonly :readwrite or :readwrite-append\n\
\tfacet:     ~S\n\
\talternate: ~S\n\
\tmode:      ~S\n\
", facet, alternate, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(facet, &f);

	a = f;
	a.contents.facet.facet = alternate->value.a_string;

	if (LL_pname(mode) == OLReadonly)
	    a.contents.facet.mode = "r";
	else if (LL_pname(mode) == OLReadwrite)
	    a.contents.facet.mode = "w";
	else if (LL_pname(mode) == OLReadwriteappend)
	    a.contents.facet.mode = "a";

	OLGlobalOctStatus = octOpenFacet(&a);
	return (OLGlobalOctStatus < OCT_OK ? LL_nil: OLMakeObject(a));
}
