#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLOpenmaster - open-master
 *
 *    (open-master instance facet mode)
 *
 *    return {facet, nil} on {success, fail}
 */
LLobject *
OLOpenmaster(instance, facet, mode)
LLobject *instance;
LLobject *facet;
LLobject *mode;
{
	octObject i, m;

	/*
	 *    don't use an object-form b/c octOpenMaster
	 *    needs the objectId for the name lookup context
	 */
	if ( ! OL_objectp(instance) ) {
	    LLErrorF("\
Function: (open-master instance facet mode)\n\
\tinstance must be an object\n\
\tinstance: ~S\n\
\tfacet:    ~S\n\
\tmode:     ~S\n\
", instance, facet, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! LL_stringp(facet) ) {
	    LLErrorF("\
Function: (open-master instance facet mode)\n\
\tfacet must be a string\n\
\tinstance: ~S\n\
\tfacet:    ~S\n\
\tmode:     ~S\n\
", instance, facet, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}
	if ( ! LL_symbolp(mode) ||
		(LL_pname(mode) != OLReadonly &&
		 LL_pname(mode) != OLReadwrite &&
		 LL_pname(mode) != OLReadwriteappend) ) {
	    LLErrorF("\
Function: (open-master instance facet mode)\n\
\tmode must be a symbol :readonly :readwrite or :readwrite-append\n\
\tinstance: ~S\n\
\tfacet:    ~S\n\
\tmode:     ~S\n\
", instance, facet, mode);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(instance, &i);

	m.type = OCT_FACET;
	m.contents.facet.facet = facet->value.a_string;

	if (LL_pname(mode) == OLReadonly)
	    m.contents.facet.mode = "r";
	else if (LL_pname(mode) == OLReadwrite)
	    m.contents.facet.mode = "w";
	else if (LL_pname(mode) == OLReadwriteappend)
	    m.contents.facet.mode = "a";

	OLGlobalOctStatus = octOpenMaster(&i, &m);
	return (OLGlobalOctStatus < OCT_OK ? LL_nil: OLMakeObject(m));
}
