#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLTransform - transform
 *
 *    (transform translation transformType &optional generalTransform)
 *
 *    return the transform
 */
LLobject *
OLTransform(translation, transformType, generalTransform)
LLobject *translation;
LLobject *transformType;
LLobject *generalTransform;	/* &optional */
{
	struct octTransform transform;
	LLobject *g00, *g01, *g10, *g11;

	if ( ! OL_pointp(translation) ) {
	    LLErrorF("\
Function: (transform translation transformType &optional generalTransform)\n\
\tthe translation must be a point\n\
\ttranslation:      ~S\n\
\ttransformType:    ~S\n\
\tgeneralTransform: ~S\n\
", translation, transformType, generalTransform);
	    LLThrow(OL_octlisp, LL_nil);
	}

	if ( ! OL_transform_typep(transformType) ) {
	    LLErrorF("\
Function: (transform translation transformType &optional generalTransform)\n\
\tthe transformType must be a transform-type\n\
\ttranslation:      ~S\n\
\ttransformType:    ~S\n\
\tgeneralTransform: ~S\n\
", translation, transformType, generalTransform);
	    LLThrow(OL_octlisp, LL_nil);
	}

	if ( ! LL_null(generalTransform) &&
		 ! (LL_consp(generalTransform) && LLGetLength(generalTransform) == 4) ) {
	    LLErrorF("\
Function: (transform translation transformType &optional generalTransform)\n\
\tthe generalTransform must be unspecified or a 4 member list of floating-points\n\
\ttranslation:      ~S\n\
\ttransformType:    ~S\n\
\tgeneralTransform: ~S\n\
", translation, transformType, generalTransform);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractPoint(translation, &transform.translation);
	OLExtractTransformType(transformType, &transform.transformType);

	if (transform.transformType == OCT_FULL_TRANSFORM) {
	    if (LL_null(generalTransform)) {
		LLErrorF("\
Function: (transform translation transformType &optional generalTransform)\n\
\tif OCT-FULL-TRANSFORM, then generalTransform must be specified as a 4 member list of numbers\n\
\ttranslation:      ~S\n\
\ttransformType:    ~S\n\
\tgeneralTransform: ~S\n\
", translation, transformType, generalTransform);
		LLThrow(OL_octlisp, LL_nil);
	    }
	    g00 = LL_first(generalTransform);
	    g01 = LL_second(generalTransform);
	    g10 = LL_third(generalTransform);
	    g11 = LL_fourth(generalTransform);

	    if ( ! LL_numberp(g00) && ! LL_numberp(g01) && ! LL_numberp(g10) && ! LL_numberp(g11) ) {
		LLErrorF("\
Function: (transform translation transformType &optional generalTransform)\n\
\tall the members of generalTransform must be numbers\n\
\ttranslation:      ~S\n\
\ttransformType:    ~S\n\
\tgeneralTransform: ~S\n\
", translation, transformType, generalTransform);
		LLThrow(OL_octlisp, LL_nil);
	    }

	    transform.generalTransform[0][0] = LL_realp(g00) ? g00->value.a_real: g00->value.a_integer;
	    transform.generalTransform[0][1] = LL_realp(g01) ? g01->value.a_real: g01->value.a_integer;
	    transform.generalTransform[1][0] = LL_realp(g10) ? g10->value.a_real: g10->value.a_integer;
	    transform.generalTransform[1][1] = LL_realp(g11) ? g11->value.a_real: g11->value.a_integer;
	} else {
	    if ( ! LL_null(generalTransform) ) {
		LLErrorF("\
Function: (transform translation transformType &optional generalTransform)\n\
\tif not OCT-FULL-TRANSFORM, then generalTransform must be left unspecified\n\
\ttranslation:      ~S\n\
\ttransformType:    ~S\n\
\tgeneralTransform: ~S\n\
", translation, transformType, generalTransform);
		LLThrow(OL_octlisp, LL_nil);
	    }
	}

	return (OLMakeTransform(transform));
}
