#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

void
OLExtractTransform(obj, transform)
LLobject *obj;
struct octTransform *transform;	/* return */
{
	LLobject *car, *cdr, *g00, *g01, *g10, *g11;

	/*
	 *    A cons-like formation of a point and a transform-type or a list
	 *
	 *    (transform-cons point transform-type)
	 *    (transform-cons point (list 00 01 10 11))
	 *
	 *    car->value.a_userDefined.value[0] = transform.x
	 *    car->value.a_userDefined.value[1] = transform.y
	 *    cdr = transformType
	 *        or
	 *    cdr = (cons (cons 00 01) (cons 10 11))
	 *
	 *    obj->value.a_userDefined.value[0] = car;
	 *    obj->value.a_userDefined.value[1] = cdr;
	 */
	car = (LLobject *) obj->value.a_userDefined.value[0];
	cdr = (LLobject *) obj->value.a_userDefined.value[1];

	transform->translation.x = (octCoord) car->value.a_userDefined.value[0];
	transform->translation.y = (octCoord) car->value.a_userDefined.value[1];

	if (LL_atom(cdr)) {
	    /*
	     *    cdr = transformType
	     */
	    transform->transformType = (octTransformType) cdr->value.a_userDefined.value[0];
	} else {
	    /*
	     *    cdr = (cons (cons 00 01) (cons 10 11))
	     */
	    g00 = LL_car(LL_car(cdr));
	    g01 = LL_cdr(LL_car(cdr));
	    g10 = LL_car(LL_cdr(cdr));
	    g11 = LL_cdr(LL_cdr(cdr));
	    transform->transformType = OCT_FULL_TRANSFORM;
	    transform->generalTransform[0][0] = LL_realp(g00) ? g00->value.a_real: (double) g00->value.a_integer;
	    transform->generalTransform[0][1] = LL_realp(g01) ? g01->value.a_real: (double) g01->value.a_integer;
	    transform->generalTransform[1][0] = LL_realp(g10) ? g10->value.a_real: (double) g10->value.a_integer;
	    transform->generalTransform[1][1] = LL_realp(g11) ? g11->value.a_real: (double) g11->value.a_integer;
	}
}

LLobject *
OLMakeTransform(transform)
struct octTransform transform;
{
	LLobject *new;

	new = LLMakeObject();
	new->type = OLTransformCell;
	new->type->count++;

	/* in case of throw below */
	new->value.a_userDefined.value[0] = (FitAny) LL_nil;
	new->value.a_userDefined.value[1] = (FitAny) LL_nil;

	new->value.a_userDefined.value[0] = (FitAny) OLMakePoint(transform.translation);
	if (transform.transformType != OCT_FULL_TRANSFORM) {
	    new->value.a_userDefined.value[1] = (FitAny) OLMakeTransformType(transform.transformType);
	} else {
	    new->value.a_userDefined.value[1] = (FitAny) LLCons(
			LLCons(LLMakeReal(transform.generalTransform[0][0]),
				LLMakeReal(transform.generalTransform[0][1])),
			LLCons(LLMakeReal(transform.generalTransform[1][0]),
				LLMakeReal(transform.generalTransform[1][1])));
	}

	return (new);
}

void
OL_Pr_Transform(obj, stream)
LLobject *obj;
STREAM *stream;
{
	struct octTransform t;
	LLobject *car, *cdr;

	car = (LLobject *) obj->value.a_userDefined.value[0];
	cdr = (LLobject *) obj->value.a_userDefined.value[1];

	if (LL_atom(cdr)) {
	    LLFormatC(stream, "~S at ~S", car, cdr);
	} else {
	    /* GODDAMN pcc can't handle the macro version of this */
	    LLFormatC(stream, "transform [ [ ~S, ~S ], [ ~S, ~S ] ] at ~S",
		     LLCar(LLCar(cdr)),
		     LLCdr(LLCar(cdr)),
		     LLCar(LLCdr(cdr)),
		     LLCdr(LLCdr(cdr)),
		     car);
	}
}

void
OL_Pr_Transform1(obj, stream)
LLobject *obj;
STREAM *stream;
{
	struct octTransform t;
	LLobject *car, *cdr;

	car = (LLobject *) obj->value.a_userDefined.value[0];
	cdr = (LLobject *) obj->value.a_userDefined.value[1];

	if (LL_atom(cdr)) {
	    LLFormatC(stream, "(transform ~S ~S)", car, cdr);
	} else {
	    LLFormatC(stream, "(transform ~S OCT-FULL-TRANSFORM ~S)", car, cdr);
	}
}

LLobject *
OL_GcScan_Transform(old)
LLobject *old;
{
	LLobject *new;

	new = LLMakeObject();

	new->type = old->type;
	new->value.a_userDefined = old->value.a_userDefined;
	new->type->count++;

	return (new);
}

void
OL_GcReScan_Transform(new, relocate)
LLobject *new;
LLobject *(*relocate)();
{
	new->value.a_userDefined.value[0] = (FitAny) (*relocate)(new->value.a_userDefined.value[0]);
	new->value.a_userDefined.value[1] = (FitAny) (*relocate)(new->value.a_userDefined.value[1]);
}
