#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

void
OLExtractTransformMask(obj, transformMask)
LLobject *obj;
int /* octTransformMask */ *transformMask;	/* return */
{
	LLobject *one, *two;

	/*
	 *    A userDefined.
	 *
	 *    obj->value.a_userDefined.value[0] = objectId
	 */

	*transformMask = (int) /* octTransformMask */ obj->value.a_userDefined.value[0];
}

LLobject *
OLMakeTransformMask(transformMask)
int /* octTransformMask */ transformMask;
{
	LLobject *new;

	new = LLMakeObject();
	new->type = OLTransformMaskCell;
	new->type->count++;

	/* in case of throw below */
	new->value.a_userDefined.value[0] = (FitAny) transformMask;

	return (new);
}

static struct {
    int /* octTransformMask */ transformMask;
    String name;
} map[] = {
    { (1<<OCT_NO_TRANSFORM),	"OCT-NO-TRANSFORM"	},
    { (1<<OCT_MIRROR_X),	"OCT-MIRROR-X"		},
    { (1<<OCT_MIRROR_Y),	"OCT-MIRROR-Y"		},
    { (1<<OCT_ROT90),		"OCT-ROT90"		},
    { (1<<OCT_ROT180),		"OCT-ROT180"		},
    { (1<<OCT_ROT270),		"OCT-ROT270"		},
    { (1<<OCT_MX_ROT90),	"OCT-MX-ROT90"		},
    { (1<<OCT_MY_ROT90),	"OCT-MY-ROT90"		},
    { (1<<OCT_FULL_TRANSFORM),	"OCT-FULL-TRANSFORM"	},
};

void
OL_Pr_TransformMask(obj, stream)
LLobject *obj;
STREAM *stream;
{
	int /* octTransformMask */ m, cmask;
        int i, count;

	OLExtractTransformMask(obj, &m);

        if (m == 0x0) {
            LLFormatC(stream, "(extract-masks OCT-NO-TRANSFORM OCT-NO-TRANSFORM)");
	    return;
	}

        for (count=0, cmask=m, i=0; i<32; i++, cmask>>=1)
            count += (cmask & 01) == 01 ? 1: 0;

        if (count > 1)
            LLFormatC(stream, "(merge-masks");

        for (i=0; i<sizeof_el (map); i++) {
            if ((map[i].transformMask & m) != 0x0)
		LLFormatC(stream, " ~A", map[i].name);
        }

        if (count > 1)
            LLFormatC(stream, ")");
}

LLobject *
OL_GcScan_TransformMask(old)
LLobject *old;
{
	LLobject *new;

	new = LLMakeObject();

	new->type = old->type;
	new->value.a_userDefined = old->value.a_userDefined;
	new->type->count++;

	return (new);
}
