#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    translation_of,
    transformType_of,
    generalTransform_of,
};

static LLobject *func();

/*
 *    OLTranslationof - translation-of
 *
 *    (translation-of box)
 *
 *    return the point
 */
LLobject *
OLTranslationof(transform)
LLobject *transform;
{
	return (func(transform, "translation-of", translation_of));
}

/*
 *    OLTransformtypeof - transformType-of
 *
 *    (transformType-of box)
 *
 *    return the transformType
 */
LLobject *
OLTransformtypeof(transform)
LLobject *transform;
{
	return (func(transform, "transformType-of", transformType_of));
}

/*
 *    OLGeneraltransformof - generalTransform-of
 *
 *    (generalTransform-of box)
 *
 *    return the generalTransform (list)
 */
LLobject *
OLGeneraltransformof(transform)
LLobject *transform;
{
	return (func(transform, "generalTransform-of", generalTransform_of));
}

static LLobject *
func(transform, name, op)
LLobject *transform;
String name;
func_op op;
{
	struct octTransform t;

	if ( ! OL_transformp(transform) ) {
	    LLErrorF("\
Function: (~A transform)\n\
\ttransform must be a transform\n\
\ttransform: ~S\n\
", name, transform);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractTransform(transform, &t);
	switch (op) {
	case translation_of:
	    return (OLMakePoint(t.translation));
	case transformType_of:
	    return (OLMakeTransformType(t.transformType));
	case generalTransform_of:
	    if (t.transformType != OCT_FULL_TRANSFORM) {
		LLErrorF("\
Function: (~A transform)\n\
\tcannot take the generalTransform of a non OCT-FULL-TRANSFORM\n\
\ttransform: ~S\n\
", name, transform);
		LLThrow(OL_octlisp, LL_nil);
	    }
	    return (LLCons(LLCons(LLMakeReal(t.generalTransform[0][0]),
				    LLCons(LLMakeReal(t.generalTransform[0][1]), LL_nil)),
			    LLCons(LLMakeReal(t.generalTransform[1][0]),
				    LLCons(LLMakeReal(t.generalTransform[1][1]), LL_nil)),
				LL_nil));
	}
	/*NOTREACHED*/
}
