#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

typedef enum func_op func_op;
enum func_op {
    object_type_of,
    property_type_of,
};

static LLobject *func();

/*
 *    OLObjecttypeof - object-type-of
 *
 *    (object-type-of object)
 *
 *    return the object-type
 */
LLobject *
OLObjecttypeof(object)
LLobject *object;
{
	return (func(object, "object-type-of", object_type_of));
}

/*
 *    OLPropertytypeof - property-type-of
 *
 *    (property-type-of object)
 *
 *    return the property-type
 */
LLobject *
OLPropertytypeof(object)
LLobject *object;
{
	return (func(object, "property-type-of", property_type_of));
}

static LLobject *
func(object, name, op)
LLobject *object;
String name;
func_op op;
{
	octObject o;
	LLobject *value;

	if ( ! OL_objectp(object) && ! OL_object_formp(object) ) {
	    LLErrorF("\
Function: (~A object)\n\
\tobject must be an object or object-form\n\
\tobject: ~S\n\
", name, object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	if (OL_object_formp(object)) {
	    switch (op) {
	    case object_type_of:
		return (LL_second(object));
	    case property_type_of:
		value = OLValueof(object);

		if (OL_objectp(value)) {
		    return (OLMakePropertyType(OCT_ID));
		} else if (LL_integerp(value)) {
		    return (OLMakePropertyType(OCT_INTEGER));
		} else if (LL_null(value)) {
		    return (OLMakePropertyType(OCT_NULL));
		} else if (LL_realp(value)) {
		    return (OLMakePropertyType(OCT_REAL));
		} else if (LL_stringp(value)) {
		    return (OLMakePropertyType(OCT_STRING));
		} else if (LL_consp(value)) {
		    Boolean ip, rp;
		    LLobject *v, *va;

		    for (v=value, ip=TRUE, rp=TRUE; LL_consp(v); v=LL_cdr(v)) {
			va = LL_car(v);

			if ( ! LL_integerp(va) )
			    ip = FALSE;
			if ( ! LL_numberp(va) )
			    rp = FALSE;
		    }

		    if (ip) {
			return (OLMakePropertyType(OCT_INTEGER_ARRAY));
		    } else if (rp) {
			return (OLMakePropertyType(OCT_REAL_ARRAY));
		    } else {
			return (OLMakePropertyType(OCT_STRANGER));
		    }
		}
	    }
	    /*NOTREACHED*/
	}

	OLExtractObject(object, &o);
	switch (op) {
	case object_type_of:
	    return (OLMakeObjectType(o.type, o.type == OCT_PROP ? o.contents.prop.type: OCT_NULL));
	case property_type_of:
	    if (o.type != OCT_PROP) {
		LLErrorF("\
Function: (~A object)\n\
\tcannot take the property-type-of of a non OCT-PROP\n\
\tobject: ~S\n\
", name, object);
		LLThrow(OL_octlisp, LL_nil);
	    }
	    return (OLMakePropertyType(o.contents.prop.type));
	}

	/*NOTREACHED*/
}
