#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLValueof - value-of
 *
 *    (value-of object)
 *
 *    return the property value
 */
LLobject *
OLValueof(object)
LLobject *object;
{
	octObject o, i;
	LLobject *ret, *list, *last, *tmp;
	int ii;
	char *c, *begin, *storage;

	if (OLOfForm(object, OCT_PROP_MASK, "value-of", "OCT-PROP", "value", &ret) == OL_ok)
	    return (ret);

	if ( ! OL_objectp(object) ) {
	    LLErrorF("\
Function: (value-of object)\n\
\tobject must be an object\n\
\tobject: ~S\n\
", object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	OLExtractObject(object, &o);

	if (o.type != OCT_PROP) {
	    LLErrorF("\
Function: (value-of object)\n\
\tobject must be an OCT-PROP\n\
\tobject: ~S\n\
", object);
	    LLThrow(OL_octlisp, LL_nil);
	}

	switch (o.contents.prop.type) {
	case OCT_ID:
	    i.objectId = o.contents.prop.value.id;
	    if (o.contents.prop.value.id == oct_null_id || octGetById(&i) < OCT_OK)
		return (LLGetValue(LLMakeSymbol("oct-null-id"), LL_valueSlot, LL_environment));

	    return (OLMakeObject(i));
	case OCT_INTEGER:
	    return (LLMakeInteger(o.contents.prop.value.integer));
	case OCT_REAL:
	    return (LLMakeReal(o.contents.prop.value.real));
	case OCT_STRING:
	    return (LLMakeString(o.contents.prop.value.string));
	case OCT_INTEGER_ARRAY:
	    list = LL_nil;
	    for (ii=0; ii<o.contents.prop.value.integer_array.length; ii++) {
		tmp = LLCons(LLMakeInteger(o.contents.prop.value.integer_array.array[ii]), LL_nil);
		if (LL_null(list))
		    list = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;
	    }

	    return (list);
	case OCT_REAL_ARRAY:
	    list = LL_nil;
	    for (ii=0; ii<o.contents.prop.value.integer_array.length; ii++) {
		tmp = LLCons(LLMakeReal(o.contents.prop.value.real_array.array[ii]), LL_nil);
		if (LL_null(list))
		    list = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;
	    }

	    return (list);
	case OCT_STRANGER:
	    list = LL_nil;
	    begin = o.contents.prop.value.stranger.contents;
	    for (ii=0, c=o.contents.prop.value.stranger.contents;
		    ii<o.contents.prop.value.stranger.length; ii++, c++) {
		if (*c == '\0') {
		    /* cons on begin (if it isn't zero length) */
		    if (begin != c) {
			tmp = LLCons(LLMakeString(begin), LL_nil);
			if (LL_null(list))
			    list = tmp;
			else
			    LL_rplacd(last, tmp);
			last = tmp;
		    }

		    /* cons on the '\0' as an integer */
		    tmp = LLCons(LLMakeInteger(0), LL_nil);
		    if (LL_null(list))
			list = tmp;
		    else
			LL_rplacd(last, tmp);
		    last = tmp;

		    begin = &c[1];
		}
	    }

	    if (begin != c) {
		/* add a '\0' at the end */
		storage = ALLOC_ARRAY(o.contents.prop.value.stranger.length+1, char);
		bcopy(o.contents.prop.value.stranger.contents, storage, sizeof (char));
		storage[o.contents.prop.value.stranger.length] = '\0';

		tmp = LLCons(LLMakeString(storage), LL_nil);
		if (LL_null(list))
		    list = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;

		FREE_ARRAY(storage, o.contents.prop.value.stranger.length+1, char);
	    }

	    return (list);
	}

	/*NOTREACHED*/
}
