#include "sumachine.h"

#include <sys/param.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ar.h>
#include <ctype.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"

#include "archive.h"
#include "internal.h"

static ARCHstatus read_archive();
static ARCHstatus read_contents();
static int get_i();
static int get_o();

/*
 *    ARCHOpen - open up the archive
 *
 *    Create if it doesn't exist just yet
 *    Copy to a tmp file
 *
 *    return {the archive, NULL} on {success, fail}
 */
ARCHarchive *
ARCHOpen(archivename, mode)
String archivename;
ARCHmode mode;
{
	DYNgeneralArray files;
	ARCHarchive *arch;
	Boolean readP;
	char tmpname[MAXPATHLEN];
	int fd, work_fd, filemode, creat;
	struct stat statbuf;

	if (access(archivename, F_OK) == 0) {
	    readP = TRUE;
	    creat = 0x0;
	} else {
	    readP = FALSE;
	    creat = O_CREAT;
	}

	filemode = mode == ARCH_readonly ? O_RDONLY: creat|O_RDWR;
	if ((fd = open(archivename, filemode, 0666)) < 0)
	    return (ARCHErrno = ARCH_ESEE_ERRNO, (ARCHarchive *) NULL);

	fstat(fd, &statbuf);

	if (readP && statbuf.st_size == 0)
	    return (ARCHErrno = ARCH_EDOTA_SHORT, (ARCHarchive *) NULL);

	DYNInit(&files, ARCHfile);

	if (mode == ARCH_readonly) {
	    work_fd = fd;

	    if (read_contents(fd, &files) == ARCH_notok) {
		close(fd);
		DYNFree(&files);
		return (NULL);
	    }
	} else {
	    sprintf(tmpname, "%s/archXXXXXX", ARCHTmp);
	    if ((work_fd = mkstemp(tmpname)) < 0) {
		close(fd);
		DYNFree(&files);
		return (NULL);
	    }
	    unlink(tmpname);

	    if (readP && read_archive(fd, work_fd, &files) == ARCH_notok) {
		close(fd);
		close(work_fd);
		DYNFree(&files);
		return (NULL);
	    }
	}

	arch = ALLOC_TYPE(ARCHarchive);

	arch->fd = fd;
	arch->work_fd = work_fd;
	arch->flushP = FALSE;
	arch->mode = mode;
	arch->modified = statbuf.st_mtime;
	arch->files = files;

	return (arch);
}

static ARCHstatus
read_archive(fd, work_fd, files)
int fd;
int work_fd;
DYNgeneralArray *files;
{
	char buf[1024];		/* big enuf */
	int cc;

	/*
	 *    Copy the file to the work area
	 */
	while ((cc = read(fd, buf, sizeof (buf))) > 0) {
	    if (write(work_fd, buf, cc) < 0)
		return (ARCHErrno = ARCH_ESEE_ERRNO, ARCH_notok);
	}

	if (cc < 0)
	    return (ARCHErrno = ARCH_ESEE_ERRNO, ARCH_notok);

	if (cc != 0)
	    return (ARCHErrno = ARCH_EDOTA_SHORT, ARCH_notok);

	/*
	 *    Now read the contents of the archive
	 */
	lseek(work_fd, 0, L_SET);
	return (read_contents(work_fd, files));
}

static ARCHstatus
read_contents(fd, files)
int fd;
DYNgeneralArray *files;
{
	Boolean cpP;
	ARCHfile o;
	char archive_magic[SARMAG+1];
	char old_cp, term_ch;
	off_t here, incr_seek;
	int cc;
	RefAny cp;

	/*
	 *    Now read the individual arch headers
	 */
	lseek(fd, 0, L_SET);
	if (read(fd, archive_magic, SARMAG) != SARMAG)
	    return (ARCHErrno = ARCH_EDOTA_SHORT, ARCH_notok);

	if ( ! strnequal(archive_magic, ARMAG, SARMAG) )
	    return (ARCHErrno = ARCH_ENOT_ARCHIVE, ARCH_notok);

#if OPSYS_MFR == MD_SUN && CPU_TYPE == MD_I386 || \
	OPSYS_MFR == MD_APOLLO || \
	OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	/* 
	 *    These guys terminate their names with a slash,
	 */
	term_ch = '/';
#else /* } else { */
	term_ch = ' ';
#endif /* } */

	here = lseek(fd, 0, L_INCR);
	while ((cc = read(fd, &o.header, sizeof (o.header))) == sizeof (o.header)) {

	    cpP = FALSE;
	    for (cp=o.header.ar_name; cp<&o.header.ar_name[sizeof (o.header.ar_name)]; cp++) {
		if (*cp == term_ch) {
		    cpP = TRUE;
		    old_cp = *cp;
		    *cp = '\0';
		    break;
		}
	    }

	    if ( ! cpP ) {
		old_cp = *cp;
		*cp = '\0';
	    }

	    o.name = ARCHCopyFilename(o.header.ar_name);

	    *cp = old_cp;	/* put it back */

	    o.date = get_i(o.header.ar_date);
	    o.uid = get_i(o.header.ar_uid);	
	    o.gid = get_i(o.header.ar_gid);	
	    o.mode = get_o(o.header.ar_mode);	
	    o.size = get_i(o.header.ar_size);	
	    o.offset = here + sizeof (o.header);

	    DYNpAtEndPut(files, &o);

	    incr_seek = o.size + (((here+sizeof (o.header)+o.size) & 01) ? 1: 0);
	    here = lseek(fd, incr_seek, L_INCR);
	}

	return (ARCH_ok);
}

static int
get_i(word)
String word;
{
	int count;

	for (count=0; isdigit(*word); word++)
	    count = 10 * count + *word - '0';

	return (count);
}

static int
get_o(word)
String word;
{
	int count;

	for (count=0; isdigit(*word); word++)
	    count = 8 * count + *word - '0';

	return (count);
}
