#include "sumachine.h"

#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <ar.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"

#include "archive.h"
#include "internal.h"

/*
 *    ARCHWrite - write the archive back
 *
 *    return {ARCH_ok, ARCH_notok} on {success, fail}
 */
ARCHstatus
ARCHWrite(arch)
ARCHarchive *arch;
{
	char buf[1024];
	int i, cc, toread, sofar, total_sofar;
	ARCHfile *o;

	if (arch->mode == ARCH_readonly)
	    return (ARCHErrno = ARCH_EREAD_ONLY, ARCH_notok);

	ftruncate(arch->fd, 0);
	lseek(arch->fd, 0, L_SET);

	if (write(arch->fd, ARMAG, SARMAG) != SARMAG)
	    return (ARCHErrno = ARCH_EBAD_WRITE, ARCH_notok);

	total_sofar = SARMAG;

	for (i=0; i<DYNSizeOfEl(&arch->files); i++) {
	    o = DYNpAt(&arch->files, i, ARCHfile);
	    if (write(arch->fd, &o->header, sizeof (o->header)) != sizeof (o->header))
		return (ARCHErrno = ARCH_EBAD_WRITE, ARCH_notok);

	    total_sofar += sizeof (o->header);

	    lseek(arch->work_fd, o->offset, L_SET);
	    sofar = 0;
	    do {
	       toread = MIN(sizeof (buf), o->size - sofar);

	       if ((cc = read(arch->work_fd, buf, toread)) != toread)
		    return (ARCHErrno = ARCH_ESHORT_TMP, ARCH_notok);
		if (cc == 0)
		    break;

		if (write(arch->fd, buf, cc) != cc)
		    return (ARCHErrno = ARCH_EBAD_WRITE, ARCH_notok);

		sofar += cc;
		if (sofar == o->size)
		    break;
	    } while (1);
	    if (sofar != o->size)
		return (ARCHErrno = ARCH_EBAD_WRITE, ARCH_notok);

	    total_sofar += o->size;

	    /*
	     *    Files must start on even boundaries
	     */
	    if (o->size & 01) {
		if (write(arch->fd, "\n", 1) != 1)
		    return (ARCHErrno = ARCH_EBAD_WRITE, ARCH_notok);
		total_sofar++;
	    }
	}

	arch->flushP = FALSE;
	return (ARCH_ok);
}
