#if ! defined(dyna_h)
#define dyna_h

#ifdef INCLUDE_HEADER
/*
 *    			RCS Control Information:
 *    $Author: saj $				
 *    $Date: 2002/12/31 15:51:18 $
 *
 *    $Revision: 1.1.1.1 $		
 *    $Source: /home/cvs/octtools/src/OctLisp/src/dyna/dyna.h,v $
 *    $State: Exp $
 */
#endif INCLUDE_HEADER

/*
 *    And now for some space management.
 *    Most often used to manage buffers of characters.
 *
 *    Packages built upon this may trust that the "type" argument to any macro
 *    here will never be used in a sizeof (type) construct - rather
 *    array->sizeof_element will be used.  That way generic routines may
 *    be written to access specific-typed arrays.
 */

typedef struct DYNgeneralArray DYNgeneralArray;
struct DYNgeneralArray {
    u_int sizeof_element;
    u_int current;		/* current end (elements)	*/
    u_int nelements;		/* sizeof space (elements)	*/
    RefAny space;		/* space for things		*/
};

#define DYNInit(dyna, element) 	(			\
	bzero(dyna, sizeof (*dyna)),			\
	(dyna)->sizeof_element = sizeof (element)	\
)

#define DYNInitP(dyna)		((dyna)->sizeof_element != 0)

#define DYNClear(dyna)		((dyna)->current = 0)

#define DYNInitClear(dyna, element)	(			\
	DYNInitP(dyna) ? DYNClear(dyna): DYNInit(dyna, element)	\
)

#define DYNTheArray(dyna, type)			((type *) (dyna)->space)

/*
 *    Four sizes:
 *
 *    DYNSizeOfOne - bytes per element
 *    DYNSizeOf - how much memory for the array.
 *    DYNSizeOfEl - how many elements in the array.
 *    DYNTotalSizeOf - how many (total) bytes in the array.
 *    DYNTotalSizeOfEl - how many (total) elements in the array.
 *
 *    Cast everything to an int, so that when (dyna)->current == 0
 *    for instance, (DYNSizeOfEl(dyna)-1) == -1, not ~1 == 0xffffffff
 */
#define DYNSizeOfOne(dyna)				\
	((int) (dyna)->sizeof_element)

#define DYNSizeOf(dyna)					\
	((int) ((dyna)->current * (dyna)->sizeof_element))

#define DYNSizeOfEl(dyna)				\
	((int) (dyna)->current)

#define DYNTotalSizeOf(dyna)				\
	((int) (dyna)->nelements * (dyna)->sizeof_element)

#define DYNTotalSizeOfEl(dyna)				\
	((int) (dyna)->nelements)

#define DYNSetCurrent(dyna, index)					\
	(dyna)->current = MAX(index, 0)

#define DYNpAt(dyna, at, type)						\
	((type *) &(dyna)->space[(at) * (dyna)->sizeof_element])

#define DYNAt(dyna, at, type)						\
	(*((type *) &(dyna)->space[(at) * (dyna)->sizeof_element]))

#define DYNLast(dyna, type)						\
	DYNAt(dyna, (dyna)->current-1, type)

#define DYNpLast(dyna, type)						\
	DYNpAt(dyna, (dyna)->current-1, type)

#define DYNAtEndPut(dyna, put)						\
	DYNAtPut(dyna, (dyna)->current++, put)

#define DYNpAtEndPut(dyna, put)						\
	DYNpAtPut(dyna, (dyna)->current++, put)

extern void DYNFree();
extern DYNgeneralArray *DYNCopy();
extern int DYNPointerToIndex();
extern void DYNAtPut();
extern void DYNpAtPut();
extern void DYNCopyInto();
extern void DYNInsureSize();
extern void DYNInsureIncrement();
extern void DYNMakeArray();

#endif /* dyna_h */
