#include <sys/param.h>

#include <sys/stat.h>

#include "stuff.h"

#if OPSYS_MFR == MD_MASSCOMP || \
    (OPSYS_MFR == MD_HP && OPSYS_VERS == MD_HPUX)
#include <ndir.h>	/* if this isn't there, you're outta luck */
#else
#include <sys/dir.h>
#endif

#include "alloc.h"
#include "dyna.h"
#include "dynstr.h"

#include "filesys.h"

extern int stat();
extern int lstat();

static int (*s_stat)();		/* either stat() or lstat()	*/

static int traverse_file_system();

/*
 *    FITraverseFileSystem - traverse the file system.
 *
 *    Failure is defined as having an error initially on the
 *    file system.  Errors deep in the file system are handled
 *    by the error function.
 *
 *    On fail, errno is set.
 *
 *    return {0, -1} on {success, fail}
 */
int 
FITraverseFileSystem(path, todo)
String path;
FIactions *todo;
{
	s_stat = stat;
	return (traverse_filesystem(path, todo));
}

/*
 *    FITraverseLinkFileSystem - traverse the file-link system.
 *
 *    Failure is defined as having an error initially on the
 *    file system.  Errors deep in the file system are handled
 *    by the error function.
 *
 *    On fail, errno is set.
 *
 *    return {0, -1} on {success, fail}
 */
int 
FITraverseLinkFileSystem(path, todo)
String path;
FIactions *todo;
{
	s_stat = lstat;
	return (traverse_filesystem(path, todo));
}

static FIactions *s_todo;

static void traverse();
static DYNStrgeneralString currentpath;

/*
 *    traverse_filesystem - traverse the file system.
 *
 *    Failure is defined as having an error initially on the
 *    file system.  Errors deep in the file system are handled
 *    by the error function.
 *
 *    On fail, errno is set.
 *
 *    return {0, -1} on {success, fail}
 */
static int 
traverse_filesystem(path, todo)
String path;
FIactions *todo;
{
	String tail;
	struct stat statbuf;

	s_todo = todo;
	tail = (tail = rindex(path, '/')) != NULL ? tail: path;

	if ((*s_stat)(path, &statbuf) < 0)
	    return (-1);

	if (s_todo->pre_root != NULL)
	    (*s_todo->pre_root)(tail, &statbuf, path, FI_preroot);

	/*
	 *    Special case when path == file
	 */
	if ((statbuf.st_mode & S_IFMT) != S_IFDIR) {
	    /*
	     *    Do this by hand
	     */
	    if (s_todo->file != NULL) {
		if (s_todo->selectP == NULL || (*s_todo->selectP)(tail, &statbuf, path))
		    (*s_todo->file)(tail, &statbuf, path, FI_file);
	    }
	} else {
	    DYNStrInit(&currentpath);		/* free later */
	    DYNStrAppendString(&currentpath, path);
	    /*
	     *    If path == "/" then don't concat - if path == "//",
	     *    then concat - don't try to do path analysis here.
	     *
	     *    A simple policy decision is best  - FNMassageToFullPath
	     *    can be called to get the path == "/" bit done.
	     */
	    if (path[0] != '/' || path[1] != '\0')
		DYNStrAppendC(&currentpath, '/');

	    traverse();
	}

	if (s_todo->post_root != NULL)
	    (*s_todo->post_root)(tail, &statbuf, path, FI_postroot);

	/*
	 *    Free the space the path was in.
	 *    return Success.
	 */
	DYNStrFree(&currentpath);
	return (0);
}

/*
 *    traverse the filesystem embodied in currentpath
 *
 *    this filesystem is assumed to be /\0 terminated
 */
static void
traverse()
{
#ifdef DEBUG
	static DYNgeneralArray *saved_layer = NULL;
#endif DEBUG
	int i,
	    concat;
	DYNgeneralArray sorted;
	DIR *dirp;
	struct direct *d;
	struct stat statbuf;

	/*
	 *    Must suck in the whole directory so that you can
	 *    go deeper than the number of file descriptors that
	 *    are available.  Rediculous, but true.
	 */

	if ((dirp = opendir(DYNStrTheString(&currentpath))) == NULL) {
	    /*
	     *    Get rid of the / at the end - it looks terrible.
	     */
	    DYNStrShrinkC(&currentpath, '\0');
	    if (s_todo->onerror != NULL) {
		String fullpath,
		       filename;
		struct stat *what;

		fullpath = DYNStrTheString(&currentpath);
		filename = (filename = rindex(fullpath, '/')) != NULL ?
			filename+1: fullpath;
		what = (*s_stat)(fullpath, &statbuf) < 0 ?
			NULL: &statbuf;
		(*s_todo->onerror)(filename, what, fullpath, FI_onerror);
	    }
	    return;
	}

	DYNInit(&sorted, String);	/* free later */

	while ((d = readdir(dirp)) != NULL)
	    DYNAtEndPut(&sorted, COPY_STRING(d->d_name));
	closedir(dirp);

	/*
	 *    Malloc doesn't tolerate such stupidity.
	 */
	if (DYNSizeOfEl(&sorted) == 0)
	    return;

	if (s_todo->sort != NULL) {
	    qsort(DYNTheArray(&sorted, String), DYNSizeOfEl(&sorted), DYNSizeOfOne(&sorted), s_todo->sort);
	}

	concat = DYNStrLength(&currentpath);

	for (i=0; i<DYNSizeOfEl(&sorted); i++) {
	    String file;

	    file = DYNAt(&sorted, i, String);
	    DYNStrSetTail(&currentpath, concat);
	    DYNStrAppendString(&currentpath, file);

	    if ((*s_stat)(DYNStrTheString(&currentpath), &statbuf) < 0) {
		if (s_todo->onerror != NULL)
		    (*s_todo->onerror)(DYNAt(&sorted, i, String), NULL, DYNStrTheString(&currentpath), FI_onerror);
		continue;
	    }

	    /*
	     *    If you reject it out of hand.
	     */
	    if (s_todo->selectP != NULL && ! (*s_todo->selectP)(file, &statbuf,
		    DYNStrTheString(&currentpath)) ) {
		continue;
	    }

	    if ((statbuf.st_mode & S_IFMT) != S_IFDIR) {
		if (s_todo->file != NULL)
		    (*s_todo->file)(file, &statbuf, DYNStrTheString(&currentpath), FI_file);
	    } else {
		if (s_todo->pre_directory != NULL) {
		    (*s_todo->pre_directory)(file, &statbuf, DYNStrTheString(&currentpath), FI_predirectory);
		}

		if ( ! FIDotOrDotDotP(file) ) {
		    int concat;

		    concat = DYNStrLength(&currentpath);
		    DYNStrAppendC(&currentpath, '/');
		    DYNStrReplaceEndC(&currentpath, '\0');

		    traverse();

		    DYNStrSetTail(&currentpath, concat);
		    DYNStrReplaceEndC(&currentpath, '\0');
		}

		if (s_todo->post_directory != NULL) {
		    (*s_todo->post_directory)(file, &statbuf, DYNStrTheString(&currentpath), FI_postdirectory);
		}
	    }
	}

	for (i=0; i<DYNSizeOfEl(&sorted); i++) 
	    FREE_STRING(DYNAt(&sorted, i, String));
	DYNFree(&sorted);
}
