#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <struct.h>		/* structure offsets */

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

LLstring OLReadonly = ":readonly";
LLstring OLReadwrite = ":readwrite";
LLstring OLReadwriteappend = ":readwrite-append";

OLtypeName OLTypeNameTable[] = {
    { "octInteger",	OL_octInteger	},
    { "octReal",	OL_octReal	},
    { "octString",	OL_octString	},
    { "octNullString",	OL_octNullString},
    { "octCoord",	OL_octCoord	},
    { "octId",		OL_octId	},
    { "octTransform",	OL_octTransform	},
    { "octPoint",	OL_octPoint	},
    { "octStranger",	OL_octStranger	},
    { "octBox",		OL_octBox	},
    { "array",		OL_array	},
};
int OLNTypeNameTable = sizeof_el (OLTypeNameTable);

/*
 *    octBag ----------------------------------------
 */
static OLattribute def_octBag[] = {
    { "name",		TRUE,	OL_octNullString,fldoff(octObject, contents.bag.name)		},
};

/*
 *    octBox ----------------------------------------
 */
static OLattribute def_octBox[] = {
    { "box",		TRUE,	OL_octBox,	fldoff(octObject, contents.box)			},
};

/*
 *    octCircle ----------------------------------------
 */
static OLattribute def_octCircle[] = {
    { "startingAngle",	TRUE,	OL_octCoord,	fldoff(octObject, contents.circle.startingAngle)},
    { "endingAngle",	TRUE,	OL_octCoord,	fldoff(octObject, contents.circle.endingAngle)	},
    { "innerRadius",	TRUE,	OL_octCoord,	fldoff(octObject, contents.circle.innerRadius)	},
    { "outerRadius",	TRUE,	OL_octCoord,	fldoff(octObject, contents.circle.outerRadius)	},
    { "center",		TRUE,	OL_octPoint,	fldoff(octObject, contents.circle.center)	},
};

/*
 *    octEdge ----------------------------------------
 */
static OLattribute def_octEdge[] = {
    { "start",		TRUE,	OL_octPoint,	fldoff(octObject, contents.edge.start)		},
    { "end",		TRUE,	OL_octPoint,	fldoff(octObject, contents.edge.end)		},
};

/*
 *    octFacet ----------------------------------------
 */
static OLattribute def_octFacet[] = {
    { "cell",		TRUE,	OL_octString,	fldoff(octObject, contents.facet.cell)		},
    { "view",		TRUE,	OL_octString,	fldoff(octObject, contents.facet.view)		},
    { "facet",		TRUE,	OL_octString,	fldoff(octObject, contents.facet.facet)		},
    { "version",	TRUE,	OL_octString,	fldoff(octObject, contents.facet.version)	},
    { "mode",		TRUE,	OL_octModeSymbol,	fldoff(octObject, contents.facet.mode)		},
#ifdef BAD
    /* peter says that using this is dangerous (and indeed it is)_ - he never fills it on copyout! */
    { "path",		TRUE,	OL_octString,	fldoff(octObject, contents.facet.path)		},
#endif
};

/*
 *    octLabel ----------------------------------------
 */
static OLattribute def_octLabel[] = {
    { "label",		TRUE,	OL_octNullString,fldoff(octObject, contents.label.label)		},
#if ! defined(OCT_JUST_RIGHT)
    /* oct 1.0 */
    { "position",	TRUE,	OL_octPoint,	fldoff(octObject, contents.label.position)	},
#else
    /* oct 2.0 */
    { "region",		TRUE,	OL_octBox,	fldoff(octObject, contents.label.region)	},
#endif
};

/*
 *    octLayer ----------------------------------------
 */
static OLattribute def_octLayer[] = {
    { "name",		TRUE,	OL_octNullString,fldoff(octObject, contents.layer.name)		},
};

/*
 *    octInstance ----------------------------------------
 */
static OLattribute def_octInstance[] = {
    { "transform",	TRUE,	OL_octTransform,fldoff(octObject, contents.instance.transform)	},
    { "name",		TRUE,	OL_octNullString,fldoff(octObject, contents.instance.name)	},
    { "master",		TRUE,	OL_octString,	fldoff(octObject, contents.instance.master)	},
    { "view",		TRUE,	OL_octString,	fldoff(octObject, contents.instance.view)	},
    { "version",	TRUE,	OL_octString,	fldoff(octObject, contents.instance.version)	},
};

/*
 *    octNet ----------------------------------------
 */
static OLattribute def_octNet[] = {
    { "name",		TRUE,	OL_octNullString,fldoff(octObject, contents.net.name)	},
};

/*
 *    octPath ----------------------------------------
 */
static OLattribute def_octPath[] = {
    { "width",		TRUE,	OL_octCoord,	fldoff(octObject, contents.path.width)	},
    { "points",		TRUE,	OL_array,	0,	 OL_octPoint,	0		},
};

/*
 *    octPoint ----------------------------------------
 */
static OLattribute def_octPoint[] = {
    { "point",		TRUE,	OL_octPoint,	fldoff(octObject, contents.point)	},
};

/*
 *    octPolygon ----------------------------------------
 */
static OLattribute def_octPolygon[] = {
    { "points",		TRUE,	OL_array,	0,	OL_octPoint,	0		}
};

/*
 *    octNullProp ----------------------------------------
 */
static OLattribute def_octNullProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
};

/*
 *    octIntegerProp ----------------------------------------
 */
static OLattribute def_octIntegerProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_octInteger,	fldoff(octObject, contents.prop.value.integer)	},
};

/*
 *    octRealProp ----------------------------------------
 */
static OLattribute def_octRealProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_octReal,		fldoff(octObject, contents.prop.value.real)	},
};

/*
 *    octStringProp ----------------------------------------
 */
static OLattribute def_octStringProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.value.string)	},
};

/*
 *    octIdProp ----------------------------------------
 */
static OLattribute def_octIdProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_octId,		fldoff(octObject, contents.prop.value.id)	},
};

/*
 *    octIntegerArrayProp ----------------------------------------
 */
static OLattribute def_octIntegerArrayProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_array,		
				fldoff(octObject, contents.prop.value.integer_array.array),
				OL_octInteger,
				fldoff(octObject, contents.prop.value.integer_array.length)		},
};

/*
 *    octRealArrayProp ----------------------------------------
 */
static OLattribute def_octRealArrayProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_array,		
				fldoff(octObject, contents.prop.value.real_array.array),
				OL_octReal,
				fldoff(octObject, contents.prop.value.real_array.length)		},
};

/*
 *    octStrangerProp ----------------------------------------
 *
 *    show me a use for it and maybe I'll give you a better interface...
 */
static OLattribute def_octStrangerProp[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.prop.name)		},
    { "value",		TRUE,	OL_octStranger,		fldoff(octObject, contents.prop.value.stranger)	},
};

/*
 *    octTerm ----------------------------------------
 */
static OLattribute def_octTerm[] = {
    { "name",		TRUE,	OL_octNullString,	fldoff(octObject, contents.term.name)		},
    { "instance",	TRUE,	OL_octId,		fldoff(octObject, contents.term.instanceId)	},
};

OLattributeTable OLAttributeTable[] = {
    { OCT_BAG,		OCT_NULL,		sizeof_el(def_octBag),			def_octBag		},
    { OCT_BOX,		OCT_NULL,		sizeof_el(def_octBox),			def_octBox		},
    { OCT_CIRCLE,	OCT_NULL,		sizeof_el(def_octCircle),		def_octCircle		},
    { OCT_EDGE,		OCT_NULL,		sizeof_el(def_octEdge),			def_octEdge		},
    { OCT_FACET,	OCT_NULL,		sizeof_el(def_octFacet),		def_octFacet		},
    { OCT_LABEL,	OCT_NULL,		sizeof_el(def_octLabel),		def_octLabel		},
    { OCT_LAYER,	OCT_NULL,		sizeof_el(def_octLayer),		def_octLayer		},
    { OCT_INSTANCE,	OCT_NULL,		sizeof_el(def_octInstance),		def_octInstance		},
    { OCT_NET,		OCT_NULL,		sizeof_el(def_octNet),			def_octNet		},
    { OCT_PATH,		OCT_NULL,		sizeof_el(def_octPath),			def_octPath		},
    { OCT_POINT,	OCT_NULL,		sizeof_el(def_octPoint),		def_octPoint		},
    { OCT_POLYGON,	OCT_NULL,		sizeof_el(def_octPolygon),		def_octPolygon		},
    { OCT_PROP,		OCT_NULL,		sizeof_el(def_octNullProp),		def_octNullProp		},
    { OCT_PROP,		OCT_INTEGER,		sizeof_el(def_octIntegerProp),		def_octIntegerProp	},
    { OCT_PROP,		OCT_REAL,		sizeof_el(def_octRealProp),		def_octRealProp		},
    { OCT_PROP,		OCT_STRING,		sizeof_el(def_octStringProp),		def_octStringProp	},
    { OCT_PROP,		OCT_ID,			sizeof_el(def_octIdProp),		def_octIdProp		},
    { OCT_PROP,		OCT_INTEGER_ARRAY,	sizeof_el(def_octIntegerArrayProp),	def_octIntegerArrayProp	},
    { OCT_PROP,		OCT_REAL_ARRAY,		sizeof_el(def_octRealArrayProp),	def_octRealArrayProp	},
    { OCT_PROP,		OCT_STRANGER,		sizeof_el(def_octStrangerProp),		def_octStrangerProp	},
    { OCT_TERM,		OCT_NULL,		sizeof_el(def_octTerm),			def_octTerm		},
};
int OLNAttributeTable = sizeof_el(OLAttributeTable);

DYNgeneralArray OLPointArray;
DYNgeneralArray OLRealArray;
DYNgeneralArray OLIntegerArray;
DYNgeneralArray OLCharArray;
