#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <struct.h>		/* structure offsets */

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

/*
 *    OLObjectForm - handle the object form
 *
 *    (object object-type a-list)
 *
 *    return nothing, just do it.
 */
void
OLObjectForm(form, object, nparray, parray)
LLobject *form;
octObject *object;		/* return */
int *nparray;			/* return */
struct octPoint **parray;	/* return */
{
	LLobject *object_type, *a_list, *pair, *value, *v, *va;
	int nattributes, i, ii;
	String need;
	octPropType prop_type;
	struct octTransform t;
	struct octBox b;
	struct octPoint p;
	struct octUserValue uv;
	octObject o;
	Boolean foundP;
	char *s, *c, *mode;
	Boolean okayP;
	OLattribute *attributes, *a;

	if (OL_objectp(form) ) {
	    /*
	     *    If form is an object, just take the object as the template itself
	     *    As appropriate, get the points.
	     */
	    OLExtractObject(form, object);
	    if (object->type == OCT_PATH || object->type == OCT_POLYGON) {
		if (OLGetPoints(object, nparray, parray) == OL_notok) {
		    LLErrorF("\
Function: (object object-type a-list)\n\
\t~S\n\
\tobject: ~S\n\
", octErrorString(), form);
		    LLThrow(OL_octlisp, LL_nil);
		}
	    }

	    return;
	}

	if ( ! OL_object_formp(form) ) {
	    LLErrorF("\
Form: (object object-type a-list)\n\
\tbadly formed object form\n\
\tform: ~S\n\
", form);
	    LLThrow(OL_octlisp, LL_nil);
	}

	object_type = LL_second(form);
	a_list = LL_third(form);

	OLExtractObjectType(object_type, &object->type);
	if (object->type != OCT_PROP) {
	    prop_type = OCT_NULL;
	} else {
	    value = LLAssoc(LLMakeSymbol("value"), a_list);
	    value = LL_cdr(value);

	    if (OL_objectp(value)) {
		prop_type = OCT_ID;
	    } else if (LL_integerp(value)) {
		prop_type = OCT_INTEGER;
	    } else if (LL_null(value)) {
		prop_type = OCT_NULL;
	    } else if (LL_realp(value)) {
		prop_type = OCT_REAL;
	    } else if (LL_stringp(value)) {
		prop_type = OCT_STRING;
	    } else if (LL_consp(value)) {
		Boolean ip, rp;
		LLobject *v, *va;

		for (v=value, ip=TRUE, rp=TRUE; LL_consp(v); v=LL_cdr(v)) {
		    va = LL_car(v);

		    if ( ! LL_integerp(va) )
			ip = FALSE;
		    if ( ! LL_numberp(va) )
			rp = FALSE;
		}

		if (ip) {
		    prop_type = OCT_INTEGER_ARRAY;
		} else if (rp) {
		    prop_type = OCT_REAL_ARRAY;
		} else {
		    prop_type = OCT_STRANGER;
		}
	    }
	    object->contents.prop.type = prop_type;
	}

	for (i=0, foundP=FALSE; i<OLNAttributeTable; i++) {
	    if (OLAttributeTable[i].type == object->type &&
		    OLAttributeTable[i].subtype == prop_type) {
		foundP = TRUE;
		break;
	    }
	}

	if ( ! foundP ) {
	    LLErrorF("\
Form: (object object-type a-list)\n\
\tbadly formed object form - no such oct object type\n\
\tform: ~S\n\
", form);
	    LLThrow(OL_octlisp, LL_nil);
	}

	nattributes = OLAttributeTable[i].nattributes;
	attributes = OLAttributeTable[i].attributes;
	for (i=0; i<nattributes; i++) {
	    a = &attributes[i];

	    if ( ! a->requiredP )
		continue;

	    pair = LLAssoc(LLMakeSymbol(a->name), a_list);
	    if (LL_null(pair)) {
		LLErrorF("\
Form: (object object-type a-list)\n\
\tbadly formed object form - no ~A member was specified\n\
\tform: ~S\n\
", a->name, form);
		LLThrow(OL_octlisp, LL_nil);
	    }

	    value = LL_cdr(pair);

	    switch (a->type) {
	    case OL_octInteger:
		okayP = LL_integerp(value);
		need = "an integer";
		break;
	    case OL_octReal:
		okayP = LL_numberp(value);
		need = "a number";
		break;
	    case OL_octString:
		okayP = LL_stringp(value);
		need = "a string";
		break;
	    case OL_octNullString:
		okayP = LL_null(value) || LL_stringp(value);
		need = "a string or nil";
		break;
	    case OL_octCoord:
		okayP = LL_integerp(value);
		need = "an integer";
		break;
	    case OL_octId:
		okayP = OL_objectp(value);
		need = "an object";
		break;
	    case OL_octTransform:
		okayP = OL_transformp(value);
		need = "a transform";
		break;
	    case OL_octPoint:
		okayP = OL_pointp(value);
		need = "a point";
		break;
	    case OL_octStranger:
		okayP = LL_consp(value);
		for (v=value; LL_consp(v) && okayP; v=LL_cdr(v)) {
		    va = LL_car(v);
		    okayP = ! LL_stringp(va) &&
			    ! (LL_integerp(va) &&
				(0 <= va->value.a_integer && va->value.a_integer <= 255));
		}
		need = "a list of small integers (0 .. 255) and at least one string";
		break;
	    case OL_octModeSymbol:
		okayP = LL_symbolp(value) &&
			(LL_pname(value) == OLReadonly ||
				LL_pname(value) == OLReadwrite ||
				LL_pname(value) == OLReadwriteappend);
		need = "the keyword :readonly, :readwrite, or :readwrite-append";
		break;
	    case OL_octBox:
		okayP = OL_boxp(value);
		need = "a box";
		break;
	    case OL_array:
		okayP = LL_consp(value);

		switch (a->subtype) {
		case OL_octInteger:
		    need = "a list of integers";
		    break;
		case OL_octReal:
		    need = "a list of numbers";
		    break;
		case OL_octPoint:
		    need = "a list of points";
		    break;
		default:
		    need = "unexpected type for array (INTERNAL ERROR)";
		    okayP = FALSE;
		    break;
		}

		if (okayP) {
		    switch (a->subtype) {
		    case OL_octInteger:
			for (v=value; LL_consp(v) && okayP; v=LL_cdr(v)) {
			    va = LL_car(v);
			    okayP = LL_integerp(va);
			}
			break;
		    case OL_octReal:
			for (v=value; LL_consp(v) && okayP; v=LL_cdr(v)) {
			    va = LL_car(v);
			    okayP = LL_numberp(va);
			}
			break;
		    case OL_octPoint:
			for (v=value; LL_consp(v) && okayP; v=LL_cdr(v)) {
			    va = LL_car(v);
			    okayP = OL_pointp(va);
			}
			break;
		    /* default handled above */
		    }
		}
		break;
	    }

	    if ( ! okayP ) {
		LLErrorF("\
Form: (object object-type a-list)\n\
\tvalue is not the correct type for the ~A field - need ~A\n\
\tvalue: ~S\n\
\tform:  ~S\n\
", a->name, need, value, form);
		LLThrow(OL_octlisp, LL_nil);
	    }

	    switch (a->type) {
	    case OL_octInteger:
		*((int *) &((RefAny) object)[a->offset]) = value->value.a_integer;
		break;
	    case OL_octReal:
		*((double *) &((RefAny) object)[a->offset]) =
			LL_realp(value) ? value->value.a_real: value->value.a_integer;
		break;
	    case OL_octString:
		*((String *) &((RefAny) object)[a->offset]) = value->value.a_string;
		break;
	    case OL_octNullString:
		if (LL_stringp(value)) {
		    *((String *) &((RefAny) object)[a->offset]) = value->value.a_string;
		} else { /* LL_null(value) */
		    *((String *) &((RefAny) object)[a->offset]) = NULL;
		}
		break;
	    case OL_octCoord:
		*((octCoord *) &((RefAny) object)[a->offset]) = value->value.a_integer;
		break;
	    case OL_octId:
		OLExtractObject(value, &o);
		*((octId *) &((RefAny) object)[a->offset]) = o.objectId;
		break;
	    case OL_octTransform:
		OLExtractTransform(value, &t);
		*((struct octTransform *) &((RefAny) object)[a->offset]) = t;
		break;
	    case OL_octPoint:
		OLExtractPoint(value, &p);
		*((struct octPoint *) &((RefAny) object)[a->offset]) = p;
		break;
	    case OL_octStranger:
		DYNClear(&OLCharArray);
		for (v=value; LL_consp(v); v=LL_cdr(v)) {
		    va = LL_car(v);
		    if (LL_stringp(va)) {
			for (s=va->value.a_string; *s == '\0'; s++)
			    DYNAtEndPut(&OLCharArray, *s);
		    } else {	/* LL_integerp(va) */
			DYNAtEndPut(&OLCharArray, (char) va->value.a_integer);
		    }
		}
		uv.length = DYNSizeOfEl(&OLCharArray);
		uv.contents = DYNTheArray(&OLCharArray,char);
		*((struct octUserValue *) &((RefAny) object)[a->offset]) = uv;
		break;
	    case OL_octModeSymbol:
		if (LL_pname(value) == OLReadonly)
		    *((char **) &((RefAny) object)[a->offset]) = "r";
		else if (LL_pname(value) == OLReadwrite)
		    *((char **) &((RefAny) object)[a->offset]) = "w";
		else if (LL_pname(value) == OLReadwriteappend)
		    *((char **) &((RefAny) object)[a->offset]) = "a";
		break;
	    case OL_octBox:
		OLExtractBox(value, &b);
		*((struct octBox *) &((RefAny) object)[a->offset]) = b;
		break;
	    case OL_array:
		okayP = LL_consp(value);
		switch (a->subtype) {
		case OL_octInteger:
		    DYNClear(&OLIntegerArray);
		    for (v=value; LL_consp(v); v=LL_cdr(v)) {
			va = LL_car(v);
			DYNAtEndPut(&OLIntegerArray, va->value.a_integer);
		    }
		    *((int32 **) &((RefAny) object)[a->offset]) = DYNTheArray(&OLIntegerArray, int32);
		    *((int *) &((RefAny) object)[a->number_offset]) = DYNSizeOfEl(&OLIntegerArray);
		    break;
		case OL_octReal:
		    DYNClear(&OLRealArray);
		    for (v=value; LL_consp(v); v=LL_cdr(v)) {
			va = LL_car(v);
			DYNAtEndPut(&OLRealArray, LL_realp(v) ? va->value.a_real: (double) va->value.a_integer);
		    }
		    *((double **) &((RefAny) object)[a->offset]) = DYNTheArray(&OLRealArray, double);
		    *((int *) &((RefAny) object)[a->number_offset]) = DYNSizeOfEl(&OLRealArray);
		    break;
		case OL_octPoint:
		    DYNClear(&OLPointArray);
		    for (v=value; LL_consp(v); v=LL_cdr(v)) {
			struct octPoint pit;

			va = LL_car(v);
			OLExtractPoint(va, &pit);
			DYNAtEndPut(&OLPointArray, pit);
		    }
		    *parray = DYNTheArray(&OLPointArray, struct octPoint);
		    *nparray = DYNSizeOfEl(&OLPointArray);
		    break;
		}
		break;
	    default:
		LLErrorF("\
Form: (object object-type a-list)\n\
\tINTERNAL ERROR - this case was handled above, but not here ...\n\
\tvalue: ~S\n\
\tform:  ~S\n\
");
		LLThrow(OL_octlisp, LL_nil);
		/*NOTREACHED*/
	    }
	}
}
