#include <sys/types.h>
#include <setjmp.h>
#include <stdio.h>
#include <struct.h>		/* structure offsets */

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstk.h"
#include "dynstr.h"
#include "st.h"
#include "stream.h"

#include "LightLisp.h"
#include "LightLispMachine.h"

#include "oct.h"
#include "oct_helper.h"

#include "OctLisp.h"
#include "OctLispMachine.h"

#include "machine.h"
#include "help.h"
#include "internal.h"

static LLobject *make_stranger();

/*
 *    OLObjectReform - create the object form
 *
 *    (object object-type a-list)
 *
 *    return the form
 */
LLobject *
OLObjectReform(object)
octObject object;		/* return */
{
	LLobject *form, *object_type, *a_list, *value;
	int nattributes, i, ii;
	Boolean foundP;
	int integer;
	double real;
	String string;
	octCoord coord;
	struct octTransform transform;
	struct octPoint point;
	struct octUserValue stranger;
	struct octBox box;
	char *mode;
	int narray;
	int32 *iarray;
	double *rarray;
	struct octPoint *parray;
	OLattribute *attributes, *a;

	for (i=0, foundP=FALSE; i<OLNAttributeTable; i++) {
	    if (OLAttributeTable[i].type == object.type &&
		    OLAttributeTable[i].subtype == 
			    (object.type == OCT_PROP ? object.contents.prop.type: OCT_NULL)) {
		foundP = TRUE;
		break;
	    }
	}

	if ( ! foundP ) {
	    LLErrorF("\
Form: (object object-type a-list)\n\
\tno such object type\n\
\tobject: ~A\n\
", octhStringifyObject(&object));
	    LLThrow(OL_octlisp, LL_nil);
	}

	a_list = LL_nil;
	nattributes = OLAttributeTable[i].nattributes;
	attributes = OLAttributeTable[i].attributes;
	for (i=0; i<nattributes; i++) {
	    a = &attributes[i];

	    if ( ! a->requiredP )
		continue;

	    switch (a->type) {
	    case OL_octInteger:
		integer = *((int *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), LLMakeInteger(integer), a_list);
		break;
	    case OL_octReal:
		real = *((double *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), LLMakeReal(real), a_list);
		break;
	    case OL_octString:
		string = *((String *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), LLMakeString(string), a_list);
		break;
	    case OL_octNullString:
		string = *((String *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name),
			string == NULL ? LL_nil: LLMakeString(string),
			a_list);
		break;
	    case OL_octCoord:
		coord = *((octCoord *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), LLMakeInteger(coord), a_list);
		break;
	    case OL_octTransform:
		transform = *((struct octTransform *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), OLMakeTransform(transform), a_list);
		break;
	    case OL_octPoint:
		point = *((struct octPoint *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), OLMakePoint(point), a_list);
		break;
	    case OL_octStranger:
		stranger = *((struct octUserValue *) &((RefAny) &object)[a->offset]);
		/* TODO - make a OLStrangerCell */
		a_list = LLAcons(LLMakeSymbol(a->name), make_stranger(stranger), a_list);
		break;
	    case OL_octModeSymbol:
		mode = *((char **) &((RefAny) &object)[a->offset]);
		if (strequal(mode, "r"))
		    a_list = LLAcons(LLMakeSymbol(a->name), LLMakeSymbol(OLReadonly), a_list);
		else if (strequal(mode, "w"))
		    a_list = LLAcons(LLMakeSymbol(a->name), LLMakeSymbol(OLReadwrite), a_list);
		else if (strequal(mode, "a"))
		    a_list = LLAcons(LLMakeSymbol(a->name), LLMakeSymbol(OLReadwriteappend), a_list);
		break;
	    case OL_octBox:
		box = *((struct octBox *) &((RefAny) &object)[a->offset]);
		a_list = LLAcons(LLMakeSymbol(a->name), OLMakeBox(box), a_list);
		break;
	    case OL_array:
		switch (a->subtype) {
		case OL_octInteger:
		    narray = *(int *) &((RefAny) &object)[a->number_offset];
		    iarray = *(int32 **) &((RefAny) &object)[a->offset];

		    for (ii=0, value=LL_nil; ii<narray; ii++)
			value = LLCons(LLMakeInteger(iarray[ii]), value);
		    break;
		case OL_octReal:
		    narray = *(int *) &((RefAny) &object)[a->number_offset];
		    rarray = *(double **) &((RefAny) &object)[a->offset];

		    for (ii=0, value=LL_nil; ii<narray; ii++)
			value = LLCons(LLMakeReal(rarray[ii]), value);
		    break;
		case OL_octPoint:
		    if (OLGetPoints(&object, &narray, &parray) == OL_notok) {
			LLErrorF("\
Form: (object object-type a-list)\n\
\tcould not get points because ~A\n\
\tobject: ~A\n\
", octErrorString(), octhStringifyObject(&object));
			LLThrow(OL_octlisp, LL_nil);
		    }

		    for (ii=0, value=LL_nil; ii<narray; ii++)
			value = LLCons(OLMakePoint(parray[ii]), value);
		    break;
		}
		a_list = LLAcons(LLMakeSymbol(a->name), value, a_list);
		break;
	    }
	}

	return (LLCons(LLMakeSymbol(OLObject),
		LLCons(OLMakeObjectType(object.type),
		LLCons(a_list, LL_nil))));
}

static LLobject *
make_stranger(stranger)
struct octUserValue *stranger;
{
	LLobject *list, *last, *tmp;
	char *c, *begin, *storage;
	int i;

	list = LL_nil;
	begin = stranger->contents;
	for (i=0, c=stranger->contents; i<stranger->length; i++, c++) {
	    if (*c == '\0') {
		/* cons on begin (if it isn't zero length) */
		if (begin != c) {
		    tmp = LLCons(LLMakeString(begin), LL_nil);
		    if (LL_null(list))
			list = tmp;
		    else
			LL_rplacd(last, tmp);
		    last = tmp;
		}

		/* cons on the '\0' as an integer */
		tmp = LLCons(LLMakeInteger(0), LL_nil);
		if (LL_null(list))
		    list = tmp;
		else
		    LL_rplacd(last, tmp);
		last = tmp;

		begin = &c[1];
	    }
	}

	if (begin != c) {
	    /* add a '\0' at the end */
	    storage = ALLOC_ARRAY(stranger->length+1, char);
	    bcopy(stranger->contents, storage, sizeof (char));
	    storage[stranger->length] = '\0';

	    tmp = LLCons(LLMakeString(storage), LL_nil);
	    if (LL_null(list))
		list = tmp;
	    else
		LL_rplacd(last, tmp);
	    last = tmp;

	    FREE_ARRAY(storage, stranger->length+1, char);
	}

	return (list);
}
