#include <sys/types.h>

#include "stuff.h"
#include "list.h"
#include "list.h"
#include "st.h"

#include "oct.h"

#include "oct_helper.h"
#include "internal.h"

/*
 *    octhBreadthFirstTraversal - do a breadth first traversal
 *
 *    void
 *    anyObject(object, state)
 *    octObject *object;
 *    FitAny state;
 *    
 *    return {OCTH_OK, OCTH_NOTOK} on {success, fail}
 */
octhStatus
octhBreadthFirstTraversal(object, anyObject, state)
octObject *object;
void (*anyObject)();
FitAny state;
{
	lsGen gen;
	octObject swing_obj, nextswing_obj;
	octGenerator oct_gen;
	st_table *visited;	/* of (octId, octId) */
	lsList tovisit, tovisit1;

	tovisit = lsCreate();
	lsNewEnd(tovisit, object->objectId, NULL);

	visited = st_init_table(st_numcmp, st_numhash);

	while (lsLength(tovisit) > 0) {
	    tovisit1 = lsCreate();

	    gen = lsStart(tovisit);
	    while (lsNext(gen, &swing_obj.objectId, NULL) == LS_OK) {

		if (octGetById(&swing_obj) < 0) {
		    /* destroy these before throwing */
		    lsDestroy(tovisit, NULL);
		    lsDestroy(tovisit1, NULL);
		    st_free_table(visited);
		    return (OCTH_NOTOK);
		}

		(*anyObject)(&swing_obj, state);

		/*
		 *    Find the next wave (up)
		 *    Don't be stupid enough to try to gen containers on a facet
		 */
		if (swing_obj.type != OCT_FACET) {
		    if (octInitGenContainers(&swing_obj, OCT_ALL_MASK, &oct_gen) < OCT_OK) {
			/* destroy these before throwing */
			lsDestroy(tovisit, NULL);
			lsDestroy(tovisit1, NULL);
			st_free_table(visited);
			return (OCTH_NOTOK);
		    }

		    while (octGenerate(&oct_gen, &nextswing_obj) == OCT_OK) {
			if (st_is_member(visited, nextswing_obj.objectId))
			    continue;

			/* mark it as visited so that future generate's won't also add it */
			st_insert(visited, nextswing_obj.objectId, nextswing_obj.objectId);
			lsNewEnd(tovisit1, nextswing_obj.objectId, NULL);
		    }
		}

		/*
		 *    Find the next wave (down)
		 */
		if (octInitGenContents(&swing_obj, OCT_ALL_MASK, &oct_gen) < OCT_OK) {
		    /* destroy these before throwing */
		    lsDestroy(tovisit, NULL);
		    lsDestroy(tovisit1, NULL);
		    st_free_table(visited);
		    return (OCTH_NOTOK);
		}

		while (octGenerate(&oct_gen, &nextswing_obj) == OCT_OK) {
		    if (st_is_member(visited, nextswing_obj.objectId))
			continue;

		    /* mark it as visited so that future generate's won't also add it */
		    st_insert(visited, nextswing_obj.objectId, nextswing_obj.objectId);
		    lsNewEnd(tovisit1, nextswing_obj.objectId, NULL);
		}
	    }
	    lsFinish(gen);

	    /*
	     *    The next wave becomes the current wave
	     */
	    lsDestroy(tovisit, NULL);
	    tovisit = tovisit1;
	}

	lsDestroy(tovisit, NULL);
	st_free_table(visited);

	return (OCTH_OK);
}
