#include <sys/types.h>

#include "stuff.h"
#include "list.h"
#include "st.h"

#include "oct.h"

#include "oct_helper.h"
#include "internal.h"

static void dive();

/*
 *    octhDepthFirstTraversal - do a depth first traversal
 *
 *    void
 *    anyObject(object, state)
 *    octObject *object;
 *    FitAny state;
 *
 *    return {OCTH_OK, OCTH_NOTOK} on {success, fail}
 */
octhStatus
octhDepthFirstTraversal(object, anyObject, state)
octObject *object;
void (*anyObject)();
FitAny state;
{
	st_table *visited;

	visited = st_init_table(st_numcmp, st_numhash);
	dive(object, visited, anyObject, state);
	st_free_table(visited);

	return (OCTH_OK);
}

static void subdive();

/*
 *    dive - do a dfs
 *
 *    void
 *    anyObject(object, state)
 *    octObject *object;
 *    FitAny state;
 *
 *    carry enough state along that another depth-first-traversal
 *    can be called by (lambda dive-object) on the dive-object with impunity.
 *
 *    return nothing, just do it.
 */
static void
dive(object, visited, anyObject, state)
octObject *object;
st_table *visited;
void (*anyObject)();
FitAny state;
{
	subdive(object, visited, anyObject, state, octInitGenContainers);
	subdive(object, visited, anyObject, state, octInitGenContents);
}

/*
 *    void
 *    anyObject(object, state)
 *    octObject *object;
 *    FitAny state;
 */
static void
subdive(parent, visited, anyObject, state, init_gen_whatchemagoo)
octObject *parent;
st_table *visited;
void (*anyObject)();
FitAny state;
octStatus (*init_gen_whatchemagoo)();
{
	octGenerator gen;
	octObject object_obj;

	/*
	 *    Don't be stupid enough to try to gen containers on a facet
	 */
	if (parent->type == OCT_FACET && init_gen_whatchemagoo == octInitGenContainers)
	    return;

	if ((*init_gen_whatchemagoo)(parent, OCT_ALL_MASK, &gen) < OCT_OK)
	    return;

	while (octGenerate(&gen, &object_obj) == OCT_OK) {

	    /*
	     *    If you've already visited here, DON'T DO IT AGAIN.
	     */
	    if (st_is_member(visited, object_obj.objectId))
		continue;

	    st_insert(visited, object_obj.objectId, object_obj.objectId);

	    (*anyObject)(&object_obj, state);
	    dive(&object_obj, visited, anyObject, state);
	}
}
