#include <sys/types.h>

#include "stuff.h"
#include "list.h"

#include "oct.h"

#include "oct_helper.h"
#include "internal.h"

/* 
 *    octhGetTerminalImplementationBB - get the terminal's implementation bb
 *
 *    this has the same functionality a utFindImpl (only you must be inside
 *    of vem (rpc) to use utFindImpl - and we are not inside of vem).
 *
 *    return {OCTH_OK, OCTH_NOTOK} on {success, fail}
 */
octhStatus
octhGetTerminalImplementationBB(terminal, bb)
octObject *terminal;
struct octBox *bb;
{
	Boolean oneP;
	octGenerator box_gen;
	octStatus box_gen_status;
	octObject box, instance, master, formalTerminal, actualTerminal;

	if (terminal->type != OCT_TERM)
	    return (OCTH_NOTOK);
	
	if (terminal->contents.term.instanceId == oct_null_id) {
	    bb->lowerLeft.x = MAXINT;
	    bb->lowerLeft.y = MAXINT;
	    bb->upperRight.x = MININT;
	    bb->upperRight.y = MININT;

	    if (octInitGenContents(terminal, OCT_BOX_MASK, &box_gen) < OCT_OK)
		return (OCTH_NOTOK);

	    oneP = FALSE;	/* had better have an implementation */
	    while ((box_gen_status = octGenerate(&box_gen, &box)) == OCT_OK) {
		bb->lowerLeft.x = MIN(bb->lowerLeft.x, box.contents.box.lowerLeft.x);
		bb->lowerLeft.y = MIN(bb->lowerLeft.y, box.contents.box.lowerLeft.y);
		bb->upperRight.x = MAX(bb->upperRight.x, box.contents.box.upperRight.x);
		bb->upperRight.y = MAX(bb->upperRight.y, box.contents.box.upperRight.y);
		oneP = TRUE;
	    }

	    if (oneP) {
		/*
		 *    If there was a geometry attached, then this is the
		 *    implementation of the formalTerminal.
		 */
		if (box_gen_status != OCT_GEN_DONE)
		    return (OCTH_NOTOK);
	    } else {
		/*
		 *    It must be a promoted connectorTerminal or instanceTerminal.
		 *    Get the actualTerminal, and find its bb.
		 */
		if (octhGetUniqueContent(terminal, OCT_TERM_MASK, &actualTerminal) != OCTH_OK)
		    return (OCTH_NOTOK);
		if (octhGetTerminalImplementationBB(&actualTerminal, bb) != OCTH_OK)
		    return (OCTH_NOTOK);
	    }
	} else {
	    instance.type = OCT_INSTANCE;
	    instance.objectId = terminal->contents.term.instanceId;

	    if (octGetById(&instance) < OCT_OK)
		return (OCTH_NOTOK);

	    if (octhOpenMaster(&instance, "r", &master) != OCTH_OK)
		return (OCTH_NOTOK);

	    formalTerminal = *terminal;	/* the same, just in the master (don't modify the arg) */
	    if (octGetByName(&master, &formalTerminal) < OCT_OK) {
		octCloseFacet(&master);
		return (OCTH_NOTOK);
	    }

	    if (octhGetTerminalImplementationBB(&formalTerminal, bb) != OCTH_OK) {
		octCloseFacet(&master);
		return (OCTH_NOTOK);
	    }

	    if (octCloseFacet(&master) < OCT_OK)
		return (OCTH_NOTOK);
	}

	return (OCTH_OK);
}
