#include <sys/types.h>

#include "stuff.h"
#include "list.h"
#include "alloc.h"
#include "dyna.h"

#include "oct.h"

#include "oct_helper.h"
#include "internal.h"

/* 
 *    octhNetList - produce a list of the nets and terminals
 *
 *    void
 *    connectivity(net, nterms, terms, state)
 *    octObject *net;
 *    int nterms;
 *    octObject *terms;
 *    FitAny state;
 *
 *    return {OCTH_OK, OCTH_NOTOK} on {success, fail}
 */
octhStatus
octhNetList(facet, connectivity, state)
octObject *facet;
void (*connectivity)();
FitAny state;
{
	DYNgeneralArray terms;	/* of octObject */
	octGenerator net_gen, term_gen;
	octStatus net_gen_status, term_gen_status;
	octObject net, term;


	if (octInitGenContents(facet, OCT_NET_MASK, &net_gen) < OCT_OK)
	    return (OCTH_NOTOK);

	DYNInit(&terms, octObject);
	while ((net_gen_status = octGenerate(&net_gen, &net)) == OCT_OK) {
	    if (octInitGenContents(&net, OCT_TERM_MASK, &term_gen) < OCT_OK)
		break;

	    DYNClear(&terms);
	    while ((term_gen_status = octGenerate(&term_gen, &term)) == OCT_OK)
		DYNpAtEndPut(&terms, &term);

	    if (term_gen_status != OCT_GEN_DONE)
		break;

	    (*connectivity)(&net, DYNSizeOfEl(&terms), DYNTheArray(&terms, octObject), state);
	}
	DYNFree(&terms);

	return (net_gen_status == OCT_GEN_DONE ? OCTH_OK: OCTH_NOTOK);
}
