#include <sys/types.h>

#include "stuff.h"
#include "list.h"
#include "tostring.h"

#include "oct.h"

#include "oct_helper.h"
#include "internal.h"

#define MAX_ELS		5	/* maximum elements to be shown of arrays or containers

/*
 *    octhStringifyObject - like octPrintObject - but return a string
 *
 *    A pointer to static space is returned.
 *
 *    return the string
 */
String
octhStringifyObject(object)
octObject *object;
{
	static char idbuf[30], buf[2048];
	octGenerator gen;
	octObject sub_object;
	octStatus status;
	octTransformType type;
	char *b;
	
	if (object->objectId == oct_null_id)
	    strcpy(idbuf, "oct-null-id");
	else
	    sprintf(idbuf, "(id %d)", object->objectId);

	buf[0] = '\0';
	switch (object->type) {
	case OCT_BOX:
	    sprintf(&buf[strlen(buf)], "Box %s, (%d, %d) (%d, %d)", 
		idbuf,
		object->contents.box.lowerLeft.x,
		object->contents.box.lowerLeft.y,
		object->contents.box.upperRight.x,
		object->contents.box.upperRight.y);
	    break;
	case OCT_NET:
	    sprintf(&buf[strlen(buf)], "Net %s named %.1024s",
		    idbuf,
		    (object->contents.net.name == NIL(char) ?
		     "<NO-NAME>": object->contents.net.name));
	    break;
	case OCT_BAG:
	    sprintf(&buf[strlen(buf)], "Bag %s named %.1024s",
		    idbuf,
		    (object->contents.bag.name == NIL(char) ?
		     "<NO-NAME>": object->contents.bag.name));
	    break;
	case OCT_LAYER:
	    sprintf(&buf[strlen(buf)], "Layer %s named %.1024s",
		    idbuf,
		    (object->contents.layer.name == NIL(char) ?
		     "<NO-NAME>": object->contents.layer.name));
	    break;
	case OCT_PATH: {
	    struct octPoint pt_array[MAX_ELS];
	    int num_points = sizeof_el (pt_array);
	    int retval, i;
	    
	    sprintf(&buf[strlen(buf)], "Path %s of width %d",
		    idbuf, object->contents.path.width);

	    retval = octGetPoints(object, &num_points, pt_array);
	    if (retval == OCT_OK) {
		sprintf(&buf[strlen(buf)], " with %d points:", num_points);

		for (i=0; i<num_points; i++)
		    sprintf(&buf[strlen(buf)], " (%d, %d)", pt_array[i]);

	    } else if (retval == OCT_TOO_SMALL) {
		sprintf(&buf[strlen(buf)], " with %d points (too many to print)", num_points);
	    } else {
		sprintf(&buf[strlen(buf)], " %s", octErrorString());
	    }
	} break;
	case OCT_POLYGON: {
	    struct octPoint pt_array[MAX_ELS];
	    int num_points = sizeof_el (pt_array);
	    int retval, i;
	    
	    sprintf(&buf[strlen(buf)], "Polygon %s", idbuf);

	    retval = octGetPoints(object, &num_points, pt_array);
	    if (retval == OCT_OK) {
		sprintf(&buf[strlen(buf)], " with %d points:", num_points);

		for (i=0; i<num_points; i++)
		    sprintf(&buf[strlen(buf)], " (%d, %d)", pt_array[i]);

	    } else if (retval == OCT_TOO_SMALL) {
		sprintf(&buf[strlen(buf)], " with %d points (too many to print)", num_points);
	    } else {
		sprintf(&buf[strlen(buf)], " %s", octErrorString());
	    }
	} break;
	case OCT_LABEL:
#if ! defined(OCT_JUST_RIGHT)
	    /* oct 1.0 */
	    sprintf(&buf[strlen(buf)], "Label %s labeled %.1024s at (%d, %d)",
		    idbuf,
		    (object->contents.label.label == NIL(char) ?
		     "<NO-NAME>": object->contents.label.label),
		    object->contents.label.position.x,
		    object->contents.label.position.y);
#else
	    /* oct 2.0 */
	    sprintf(&buf[strlen(buf)], "Label %s labeled %.1024s in { (%d, %d), (%d, %d) } of height %d, justified %s vertically, %s horizontally, and %s per line",
		    idbuf,
		    (object->contents.label.label == NIL(char) ?
		     "<NO-NAME>": object->contents.label.label),
		    object->contents.label.region.lowerLeft.x,
		    object->contents.label.region.lowerLeft.y,
		    object->contents.label.region.upperRight.x,
		    object->contents.label.region.upperRight.y,
		    object->contents.label.textHeight,
		    object->contents.label.vertJust == OCT_JUST_TOP ? "top":
			object->contents.label.vertJust == OCT_JUST_CENTER ? "center":
			"bottom",
		    object->contents.label.horizJust == OCT_JUST_LEFT ? "left":
			object->contents.label.horizJust == OCT_JUST_CENTER ? "center":
			"right",
		    object->contents.label.lineJust == OCT_JUST_LEFT ? "left":
			object->contents.label.lineJust == OCT_JUST_CENTER ? "center":
			"right");
#endif
	    break;
	case OCT_POINT:
	    sprintf(&buf[strlen(buf)], "Point %s at (%d, %d)", idbuf,
		    object->contents.point.x,
		    object->contents.point.y);
	    break;
	case OCT_EDGE:
	    sprintf(&buf[strlen(buf)], "Edge %s from (%d, %d) to (%d, %d)", idbuf,
		    object->contents.edge.start.x, object->contents.edge.start.y,
		    object->contents.edge.end.x, object->contents.edge.end.y);
	    break;
	case OCT_CIRCLE:
	    sprintf(&buf[strlen(buf)], "Circle %s centered at (%d, %d) from angle %d to %d, and radius %d to %d",
		    idbuf,
		    object->contents.circle.center.x,
		    object->contents.circle.center.y,
		    object->contents.circle.startingAngle,
		    object->contents.circle.endingAngle,
		    object->contents.circle.innerRadius,
		    object->contents.circle.outerRadius);
	    break;
	case OCT_TERM:
	    sprintf(&buf[strlen(buf)], "Term %s named %.1024s",
		    idbuf, object->contents.term.name == NIL(char) ?
			    "<NO-NAME>": object->contents.term.name);
	    if (object->contents.term.instanceId == oct_null_id) {
		strcat(&buf[strlen(buf)], " is a formalTerminal");
	    } else {
		sub_object.objectId = object->contents.term.instanceId;
		if (octGetById(&sub_object) < OCT_OK) {
		    sprintf(&buf[strlen(buf)], " on invalid instance (id %d) because %s",
			    octErrorString());
		} else {
		    sprintf(&buf[strlen(buf)], " on instance (id %d) named %.1024s",
			    object->contents.term.instanceId,
			    sub_object.contents.instance.name == NIL(char) ?
				    "<NO-NAME>": sub_object.contents.instance.name);
		}
	    }
	    break;
	case OCT_INSTANCE: 
	    type = object->contents.instance.transform.transformType;
	    sprintf(&buf[strlen(buf)], "Instance %s named %.1024s of %s:%s at (%d, %d) with transform %s",
		   idbuf,
		    (object->contents.instance.name == NIL(char) ?
		     "<NO-NAME>": object->contents.instance.name),
		   object->contents.instance.master,
		   object->contents.instance.view,
		   object->contents.instance.transform.translation.x,
		   object->contents.instance.transform.translation.y,
		    ((int) type <= (int) OCT_FULL_TRANSFORM ? octhOctTransformTypeNames[(int)type]:
		      "<NO-VALID-TRANSFORM>"));

	    if (type == OCT_FULL_TRANSFORM) {
		double (*matrix)[2];
		matrix = object->contents.instance.transform.generalTransform;

		sprintf(&buf[strlen(buf)], " with matrix [%g, %g], [%g, %g]",
		       matrix[0][0], matrix[0][1], matrix[1][0],matrix[1][1]);
	    }
	    break;
	case OCT_FACET: 
	    sprintf(&buf[strlen(buf)], "A Facet %s %s:%s:%s:%s",
		   idbuf,
		    (object->contents.facet.cell == NIL(char) ?
		     "": object->contents.facet.cell),
		    (object->contents.facet.view == NIL(char) ?
		     "": object->contents.facet.view),
		    (object->contents.facet.facet == NIL(char) ?
		     "": object->contents.facet.facet),
		    (object->contents.facet.version == NIL(char) ?
		     "": object->contents.facet.version));
	    break;
	case OCT_PROP:
	    sprintf(&buf[strlen(buf)], "Prop %s named %.1024s, value ",
		    idbuf, object->contents.prop.name);

	    switch (object->contents.prop.type) {
	    case OCT_INTEGER:
		sprintf(&buf[strlen(buf)], "Integer %d", object->contents.prop.value.integer);
		break;
	    case OCT_REAL:
		sprintf(&buf[strlen(buf)], "Real %g", object->contents.prop.value.real);
		break;
	    case OCT_STRING:
		sprintf(&buf[strlen(buf)], "String %.1024s", object->contents.prop.value.string);
		break;
	    case OCT_STRANGER: {
		int length = object->contents.prop.value.stranger.length;
		RefAny stranger = object->contents.prop.value.stranger.contents;
		String pc;
		
		sprintf(&buf[strlen(buf)], "%d bytes of stranger data: ", length);

		b = &buf[strlen(buf)];
		while (length-- > 0 && &b[5] /* estimate strlen(pc) */ < &buf[sizeof_el (buf)-1]) {
		    pc = PrintableChar(*stranger++);
		    while (*pc != '\0')
			*b++ = *pc++;
		}
		*b++ = '\0';
	    } break;
	    case OCT_REAL_ARRAY: {
		int length = object->contents.prop.value.real_array.length;
		double *real = object->contents.prop.value.real_array.array;
	    
		sprintf(&buf[strlen(buf)], "%d doubles in array: ", length);

		b = &buf[strlen(buf)];
		while (length-- > 0 && &b[MAX_ELS] /* estimate length %g */ < &buf[sizeof_el (buf)-1]) {
		    sprintf(b, " %g", *real++);
		    b = &b[strlen(b)];
		}
	    } break;
	    case OCT_INTEGER_ARRAY: {
		int length = object->contents.prop.value.integer_array.length;
		int *integer = object->contents.prop.value.integer_array.array;
		
		sprintf(&buf[strlen(buf)], "%d ints in array: ", length);

		b = &buf[strlen(buf)];
		while (length-- > 0 && &b[10] /* estimate length %d */ < &buf[sizeof_el (buf)-1]) {
		    sprintf(b, " %d", *integer++);
		    b = &b[strlen(b)];
		}
	    } break;
	    case OCT_ID:
		sprintf(&buf[strlen(buf)], "Object id %d", object->contents.prop.value.id);
		break;
	    }
	    break;
	default:
	    sprintf(&buf[strlen(buf)], "An object of (unknown) type %d", object->type);
	    return;	/* don't even attempt to do any more characterization */
	}
	
	if (object->objectId == oct_null_id) {
	    strcat(&buf[strlen(buf)], ", no attachments possible for null-id");
	} else {
	    if (octInitGenContents(object, OCT_ALL_MASK, &gen) != OCT_OK) {
		strcat(&buf[strlen(buf)], octErrorString());
	    } else {
		int i = 0;

		status = octGenerate(&gen, &sub_object);
		
		if (status == OCT_OK) {
		    strcat(&buf[strlen(buf)], ", containing:");
		    do {
			sprintf(&buf[strlen(buf)], " %d", sub_object.objectId);
		    } while ((status = octGenerate(&gen, &sub_object)) == OCT_OK && i++ < MAX_ELS);
		    if (status == OCT_OK) {
			strcat(&buf[strlen(buf)], " more ... ");
			status = OCT_GEN_DONE;
		    }
		}
	    }
	    if (status != OCT_GEN_DONE)
		strcat(&buf[strlen(buf)], octErrorString());
	    
	    if (object->type != OCT_FACET) {
		if (octInitGenContainers(object, OCT_ALL_MASK, &gen) != OCT_OK) {
		    strcat(&buf[strlen(buf)], octErrorString());
		} else {
		    int i = 0;

		    status = octGenerate(&gen, &sub_object);
		    
		    if (status == OCT_OK) {
			strcat(&buf[strlen(buf)], ", attached to:");
			do {
			    sprintf(&buf[strlen(buf)], " %d", sub_object.objectId);
			} while ((status = octGenerate(&gen, &sub_object)) == OCT_OK && i++ < MAX_ELS);
			if (status == OCT_OK) {
			    strcat(&buf[strlen(buf)], " more ... ");
			    status = OCT_GEN_DONE;
			}
		    }
		    if (status != OCT_GEN_DONE)
			strcat(&buf[strlen(buf)], octErrorString());
		}
	    }
	}

	return (buf);
}
