;;;;
;;;;    Packing and unpacking facet names
;;;;

;;;
;;;    (setq f (open-facet-by-name "yes" "no"
;;;                                "contents" OCT-CURRENT-VERSION
;;;                                :readwrite))
;;;    (pack-facet-name f)
;;;

(defun pack-facet-name (facet)
  "pack a facet name into a string"
  (format nil "~a:~a:~a:~a"
	  (cell-of facet)
	  (view-of facet)
	  (facet-of facet)
	  (version-of facet)))

;;;
;;;    (setq facet-form (unpack-facet-name "yes:no:contents"))
;;;    (setq facet (open-facet facet-form :readwrite))
;;;
;;;    (setq facet-form
;;;          (unpack-facet-name "yes"
;;;                             `(object ,OCT-FACET
;;;                                      ((cell . "anything")
;;;                                       (view . "symbolic")
;;;                                       (facet . "contents")
;;;                                       (version . ,OCT-CURRENT-VERSION)
;;;                                       (mode . :readwrite-append)))))
;;;    (setq facet (open-facet facet-form :readwrite))
;;;

(defun unpack-facet-name (packed &optional default-form)
  "the hard one - unpack a facet
  this is all rather jockish, but unpacking a facet is a pain"
  (let ((burst (if (null packed) nil
		   (let ((ret nil))
			(dotimes (i (length packed) ret)
				 (setq ret (cons (elt packed i) ret))))))
	(cell nil)
	(view nil)
	(facet nil)
	(version nil))
       (macrolet ((while (test &body body)
			 "a while loop"
			 `(do ()
			      ((null ,test))
			      ,@body))
		  (not-: (string-list)
			 "not a ``:''"
			 `(and ,string-list
			       (not (equal (car ,string-list) ":"))))
		  (push (symbol string-list)
			"help in cons-ing up a list"
			`(setq ,symbol (cons (car ,string-list) ,symbol)))
		  (next (symbol)
			"help in cdr-ing down a list"
			`(setq ,symbol (cdr ,symbol))))
		 
		 ;; remember, do it backwards because consing up a list
		 ;; reverses the list (reconsing reverses the reversal)
		 (while (not-: burst)
			(push version burst)
			(next burst))
		 (next burst)	; over the :
		 (while (not-: burst)
			(push facet burst)
			(next burst))
		 (next burst)	; over the :
		 (while (not-: burst)
			(push view burst)
			(next burst))
		 (next burst)	; over the :
		 (while (not-: burst)
			(push cell burst)
			(next burst))
		 
		 ;; beause its backwards, if the form was underspecified,
		 ;; then the cell will be nil, but the verison will be set.
		 ;; must shift down appropriately
		 (let ((i 4))
		      (while (and (null cell) (< 0 (setq i (1- i))))
			     (setq cell view
				   view facet
				   facet version
				   version nil))))
       (if cell
	   (setq cell (apply #'concatenate 'string cell)))
       (if view
	   (setq view (apply #'concatenate 'string view)))
       (if facet
	   (setq facet (apply #'concatenate 'string facet)))
       (if version
	   (setq version (apply #'concatenate 'string version)))
       `(object ,OCT-FACET
		((cell . ,(cond (cell cell)
				(default-form (cell-of default-form))
				(t "cell")))
		 (view . ,(cond (view view)
				(default-form (view-of default-form))
				(t "symbolic")))
		 (facet . ,(cond (facet facet)
				 (default-form (facet-of default-form))
				 (t "contents")))
		 (version . ,(cond (version version)
				   (default-form (version-of default-form))
				   (t OCT-CURRENT-VERSION)))
		 (mode . ,(cond (default-form (mode-of default-form))
				(t :readwrite)))))))

;;;
;;;    (setq facet (open-facet-by-packed-name "/tmp/zippy:symbolic") :readwrite)
;;;

(defun open-facet-by-packed-name (packed mode)
  "open a facet by its packed facetname"
  (open-facet (unpack-facet-name packed) mode))
