;;;;(
;;;;    Make a checkerboard
;;;;
;;;;    (checkerboard "tricky-dick-nixon" "checker")
;;;;    (checkerboard "nabisco-trademark")
;;;;

(defun rcs (symbol)
  "use the symbol as a key to what to pick out of rcs-list,
  and then remove the leading and trailing $ signs
  symbol should be 'revision, 'date, 'author or 'source"
  (let ((rcs-list '((Revision . "$Revision: 1.1.1.1 $")
		    (Date . "$Date: 2002/12/31 15:51:27 $")
		    (Author . "$Author: saj $")
		    (Source . "$Source: /home/cvs/octtools/src/OctLisp/src/octlisp/demo/checkerboard.ll,v $"))))
       (let ((item (assoc symbol rcs-list))
	     string)
	    (when (null item)
		  (format t "Function (rcs symbol)~%    ~
			  where symbol is Revision, Date, Author, or Source~%    ~
			  symbol: ~S~%" symbol)
		  (throw 'lightlisp nil))
	    (setq string (cdr item))
	    (subseq string 1 (1- (length string))))))

(defparameter TECHNOLOGY "scmos"
  "the name of the technology to use")

(defparameter BLACK "NDIF"
  "the name of the black squares")
(defparameter WHITE "PDIF"
  "the name of the layer of white squares")

(defparameter WIDTH 10
  "the width of the board in squares")
(defparameter SQUARE (box (point 0 0)
			 (point 100 100))
  "the size of a square")

(defun width (box)
  "return the width of any box"
  (- (x-of (upperRight-of box)) (x-of (lowerLeft-of box))))

(defun height (box)
  "return the height of any box"
  (- (y-of (upperRight-of box)) (y-of (lowerLeft-of box))))

(defun safe-open-facet-by-name (cell view facet version mode)
  "a safe version of open-facet-by-name"
  (let ((facet (open-facet-by-name cell view facet version mode)))
       (unless (objectp facet)
	       (format t "open-facet-by-name could not open ~a:~a:~a:~a for ~s because ~a~%"
		       cell view facet version mode (oct-error-string))
	       (throw 'octlisp (oct-error-status)))
       ;; return the facet
       facet))

(defun checkerboard (cellname &optional (viewname "checkerboard"))
  "make a checkerboard in the checkerboard view
  the facet is returned"
  (let* ((facet (safe-open-facet-by-name cellname viewname
			    "contents" OCT-CURRENT-VERSION :readwrite))
	 (black-layer (create facet `(object ,OCT-LAYER ((name . ,BLACK)))))
	 (white-layer (create facet `(object ,OCT-LAYER ((name . ,WHITE))))))
	
	;; set up the required property stuff
	(create facet `(object ,OCT-PROP ((name . "EDITSTYLE")
					  (value . "PHYSICAL"))))
	(create facet `(object ,OCT-PROP ((name . "VIEWTYPE")
					  (value . "PHYSICAL"))))
	(create facet `(object ,OCT-PROP ((name . "TECHNOLOGY")
					  (value . ,TECHNOLOGY))))
	
	;; define local functions with an flet
	(flet ((checker (x-offset y-offset)
			"create a new square that is offset by x-offset
			and y-offset (which are both in units of SQUARE's)"
			(let ((width (width SQUARE))
			      (height (height SQUARE)))
			     (box (point (* x-offset width)
					 (* y-offset height))
				  (point (* (1+ x-offset) width)
					 (* (1+ y-offset) height))))))
	      ;; get down to business
	      ;; foreach x and foreach y, make a square alternately
	      ;; on black or white, chosing as appropriate for
	      ;; even or odd rows.
	      (dotimes (x WIDTH)
		       (dotimes (y WIDTH)
				(create (if (or (and (oddp x) (evenp y))
						(and (evenp x) (oddp y)))
					    black-layer
					    white-layer)
					`(object ,OCT-BOX
						 ((box . ,(checker x y))))))))
	;; write the facet to disk, and return it
	(flush-facet facet)
	facet))

(defun demo ()
  "run a demo"
  (setq demo (checkerboard "/tmp/tricky-dick-nixon" "checker")))

(format t "checkerboard -    ~A~%    ~
  use (demo) to demo~%" (rcs 'Revision))

;)
