;;;;
;;;;    Get a list of all formalTerminals on a facet
;;;;    that obey some predicate or another
;;;;

(defun formalTerminals-of-such-that (facet &optional predicate)
  "get all the formalTerminals on a facet that obey some predicate
  or another; by default get all terminals.

  predicate is called as (funcall predicate term) "
  (do* ((gen (init-gen-contents facet OCT-TERM-MASK))
	(term (generate gen) (generate gen))
	(list nil))
       ;; (null term) is the end-condition, list is the return value
       ((null term) list)
       
       ;; cons up the list if the predicate 
       (if (or (null predicate)	; &optional args are nil by default
	       (funcall predicate term))
	   (setq list (cons term list)))))

(defun INPUT-p (term)
  "is the terminal an input terminal?"
  (equal "INPUT"
	 (value-of (get-by-name term OCT-PROP "DIRECTION"))))

(defun print-objs (object-list)
  "print a list of objects nicely"
  (princ "The Object List:") (terpri)
  (dolist (o object-list)
	  (princ o)
	  (terpri)))

(defun demo ()
  "a demo function"
  (setq demo
	(let ((facet (open-facet-by-name "/tmp/whipping" "symbolic" "contents"
				 OCT-CURRENT-VERSION :readwrite-append)))
	     (when (null facet)
		   ;; print out an ultra-informative error message
		   ;; and throw up (urp)
		   (format t "Could not open the demo facet because ~a~%"
			   (oct-error-string))
		   (throw 'octlisp nil))
	     
	     ;; create some terminals on the facet
	     (dolist (term-name '("eany" "meany" "miney" "moe"
					 "this" "little" "piggy" "had" "roast-beef"))
		     ;; create the terminal and assign an arbitrary
		     ;; (randomly chosen) direction to it
		     (let ((term (create facet `(object ,OCT-TERM
							((name . ,term-name)
							 (instance . ,oct-null-id)))))
			   (direction (if (> (random 1000) 500) "INPUT" "OUTPUT")))
			  (create term `(object ,OCT-PROP
						((name . "DIRECTION")
						 (value . ,direction))))))
	     ;; return the facet as the value to setq
	     facet))
  
  ;; take the default args
  (print-objs (formalTerminals-of-such-that demo))
  
  ;; pass a predicate
  (print-objs (formalTerminals-of-such-that demo #'INPUT-p))
  
  ;; pass a lambda-predicate (a literal function)
  (print-objs (formalTerminals-of-such-that demo
					    #'(lambda (term)
						      (subshell "lambda")
						      (let* ((property (get-by-name term OCT-PROP "DIRECTION"))
							     (value (value-of property)))
							    (equal "OUTPUT" value))))))

(format t "Generate Formal Terminals Demo~%    use (demo) to get a demo~%")
