;;;;
;;;;    Test to see that all of the modes are implemented correctly
;;;;

(defun create-stuff (facet)
  "create some random stuff in the facet"
  (let ((layer (get-or-create facet `(object ,OCT-LAYER
				      ((name . "larry"))))))
       (dotimes (i 2)	; not too many
		(create layer `(object ,OCT-CIRCLE
				       ((center . ,(point (random 300) (random 300)))
					(innerRadius . 0)
					(outerRadius . ,(random 20))
					(startingAngle . 0)
					(endingAngle . 0)))))))

(defun traverse-stuff (container &optional (indent ""))
  "assume the facet is a tree - traverse the tree"
  (format t "~A~S~%" container indent)
  (setq indent (concatenate 'string "  " indent))
  (do-contents (o container)
	       (traverse-stuff o indent)))

(defun genname (&optional x)
  "generate a name in the form of gensym"
  (symbol-name (if x (gensym x) (gensym))))

(defun test-mode (mode)
  "test the behavior of open-facet-by-name and close-facet relative to ``mode''"
  (catch 'damn
	 (let* ((cell (format nil "/tmp/~A" (genname)))
		(view (genname))
		(facet (open-facet-by-name cell view "contents"
					   OCT-CURRENT-VERSION mode)))
	       (dotimes (i 3)
			(format t "Iteration  #~D for ~A:~A:contents~%" i cell view)
			(when (null facet)
			      (format t "Could not open ~A:~A:contents as ~S because ~A~%"
				      cell view mode (oct-error-string))
			      (throw 'damn nil))
			(create-stuff facet)
			(traverse-stuff facet)
			(if (not (close-facet facet))
			    (format t "Could not close (~S) facet ~A:~A:contents because ~A~%"
				    mode cell view (oct-error-string))
			    (format t "COULD close (~S) facet ~A:~A:contents because ~A~%"
				    mode cell view (oct-error-string)))))))

(test-mode :readonly)
(test-mode :readwrite)
(test-mode :readwrite-append)
