#include "sumachine.h"

#include <sys/types.h>
#include <stdio.h>

#include "stuff.h"
#include "alloc.h"
#include "dyna.h"
#include "dynstr.h"
#include "filename.h"

#include "stream.h"

#include "internal.h"

static STREAMstatus setup();

/*
 *    STREAMCreate - create a stream
 *
 *    return the created stream
 */
STREAM *
STREAMCreate(inputType, inputvalue, inputname, outputType, outputvalue, outputname)
STREAMtype inputType;
RefAny inputvalue;		/* depends on what inputType is */
String inputname;
STREAMtype outputType;
RefAny outputvalue;		/* depends on what outputType is */
String outputname;
{
	STREAM *new;
	STREAMstatus input_status, output_status;

	new = ALLOC_TYPE(STREAM);
	if (new == NULL)
	    return (NULL);

	input_status = setup(&new->input, STREAM_input, inputType, inputvalue, inputname);

	new->firstReadP = TRUE;
	DYNStrInit(&new->unget);
	new->prompt.type = STREAM_noprompt;
	new->newline.type = STREAM_nonewline;

	output_status = setup(&new->output, STREAM_output, outputType, outputvalue, outputname);
	if (input_status != STREAM_ok || output_status != STREAM_ok) {
	    STREAMDestroy(new);
	    return (NULL);
	}
	
	return (new);
}

static STREAMstatus
setup(side, sidename, type, value, name)
STREAMside *side;
STREAMsideName sidename;
STREAMtype type;
RefAny value;		/* depends on what type is */
String name;
{
	STREAMside *copy;
	int fd;
	FILE *f;

	if (type == STREAM_STREAM) {
	    copy = sidename == STREAM_output ?
		    &((STREAM *) value)->output: &((STREAM *) value)->input;

	    side->type = type;
	    side->counts = copy->counts;

	    side->name = COPY_STRING(name != NULL ? name: copy->name);

	    switch (copy->type) {
	    case STREAM_FILE:
		side->value.a_FILE.destroyP = TRUE;	/* a new copy */
		fd = dup(fileno(copy->value.a_FILE.file));
		side->value.a_FILE.file = fdopen(fd, sidename == STREAM_output ? "a": "r");
		if (side->value.a_FILE.file == NULL)
		    return (STREAM_notok);
		break;
	    case STREAM_string:
		DYNStrInit(&side->value.a_string.buf);
		DYNStrAppendString(&side->value.a_string.buf, 
			DYNStrTheString(&copy->value.a_string.buf));
		side->value.a_string.index = copy->value.a_string.index;
		break;
	    case STREAM_charback:
		side->value.a_callback.function = copy->value.a_callback.function;
		break;
	    case STREAM_lineback:
	    case STREAM_xactback:
		side->value.a_callback.function = copy->value.a_callback.function;
		DYNStrInit(&side->value.a_callback.buf);
		break;
	    default:
		return (STREAM_notok);
	    }

	    /* don't change type till AFTER success */
	    side->type = copy->type;
	    return (STREAM_ok);
	}

	side->type = type;

	side->counts.parens = 0;
	side->counts.curls = 0;
	side->counts.squares = 0;
	side->counts.bracks = 0;

	side->counts.lineno = 1;
	side->counts.columnno = 0;

	side->name = COPY_STRING(name != NULL ? name: "<anonymous>");

	switch (type) {
	case STREAM_void:
	    break;
	case STREAM_filename:
	    if ((f = FNTildeOpen((char *) value,
		    sidename == STREAM_output ? "a": "r")) == NULL) {
		return (STREAM_notok);
	    }

	    /* don't change type till AFTER success */
	    side->type = STREAM_FILE;
	    side->value.a_FILE.destroyP = TRUE;
	    side->value.a_FILE.file = f;
	    break;
	case STREAM_FILEWRAP:
	    side->type = STREAM_FILE;
	    side->value.a_FILE.destroyP = FALSE;
	    side->value.a_FILE.file = (FILE *) value;
	    break;
	case STREAM_FILE:
	    side->value.a_FILE.destroyP = TRUE;
	    side->value.a_FILE.file = (FILE *) value;
	    break;
	case STREAM_string:
	    side->value.a_string.index = 0;
	    DYNStrInit(&side->value.a_string.buf);
	    if (value != NULL)
		DYNStrAppendString(&side->value.a_string.buf, (char *) value);
	    break;
	case STREAM_charback:
	    side->value.a_callback.function = (STREAMstatus (*)()) value;
	    break;
	case STREAM_lineback:
	case STREAM_xactback:
	    side->value.a_callback.function = (STREAMstatus (*)()) value;
	    side->value.a_callback.index = 0;
	    DYNStrInit(&side->value.a_callback.buf);
	    break;
	default:
	    return (STREAM_notok);
	}

	return (STREAM_ok);
}
