#if ! defined(stream_h)
#define stream_h

typedef enum STREAMtype STREAMtype;
enum STREAMtype {
    STREAM_void,	/* nothing */
    STREAM_STREAM,	/* converted into a copy */
    STREAM_filename,	/* turns into a STREAM_FILE */
    STREAM_FILE,
    STREAM_FILEWRAP,	/* like STREAM_FILE, but don't fclose on destroy */
    STREAM_string,
    STREAM_charback,	/* per char (character buffered) */
    STREAM_lineback,	/* per line (line buffered) */
    STREAM_xactback,	/* per transaction */
};

typedef enum STREAMstatus STREAMstatus;
enum STREAMstatus {
    STREAM_ok,
    STREAM_eof,
    STREAM_notok,
};

typedef struct STREAMcharCounts STREAMcharCounts;
struct STREAMcharCounts {
    int parens;		/* ( ) - level of nesting */
    int curls;		/* { } - level of nesting */
    int squares;	/* [ ] - level of nesting */
    int bracks;		/* < > - level of nesting */
    int lineno;		/* the current line number */
    int columnno;	/* the current column number */
};

/*
 *    Printing and reading 
 *
 *    STREAMstatus
 *    callback_read(c)	- read a char
 *    char *c;			- return
 *
 *    STREAMstatus
 *    callback_write(c)	- write a char
 *    char c;			- value
 *
 *    --------------------
 *
 *    STREAMstatus
 *    callback_read(buf)	- read a line
 *    DYNStrgeneralString *buf;	- fill (clear)
 *
 *    STREAMstatus
 *    callback_write(buf)	- write a line
 *    DYNStrgeneralString *buf;	- write 
 *
 *    --------------------
 *
 *    STREAMstatus
 *    callback_read(buf)	- read a transaction
 *    DYNStrgeneralString *buf;	- fill (clear)
 *
 *    STREAMstatus
 *    callback_write(buf)	- write a transaction
 *    DYNStrgeneralString *buf;	- write 
 */
typedef struct STREAMside STREAMside;
struct STREAMside {
    STREAMtype type;
    String name;
    STREAMcharCounts counts;
    char last;
    union {
	struct {
	    Boolean destroyP;		/* fclose on destroy */
	    FILE *file;
	} a_FILE;
	struct {
	    int index;		 	/* if reading */
	    DYNStrgeneralString buf;	/* of char */
	} a_string;
	struct {
	    int index;		 	/* if reading */
	    DYNStrgeneralString buf;	/* of char */
	    STREAMstatus (*function)();
	} a_callback;
    } value;
};

typedef enum STREAMsideName STREAMsideName;
enum STREAMsideName {
    STREAM_input,
    STREAM_output,
};

typedef enum STREAMpromptType STREAMpromptType;
enum STREAMpromptType {
    STREAM_noprompt,
    STREAM_tty,		/* tty driver is on */
    STREAM_raw,		/* tty in raw mode or X based */
};

typedef enum STREAMnewlineType STREAMnewlineType;
enum STREAMnewlineType {
    STREAM_nonewline,		/* no newline processing */
    STREAM_beforenewline,	/* before newline processing */
    STREAM_afternewline,	/* after newline processing */
    STREAM_column0,		/* when in column 0 */
};

/*
 *    STREAMstatus
 *    prompt.callback(stream, userData)
 *    STREAM *stream;			- prompt stream (output)
 *    RefAny userData;
 *
 *    void
 *    prompt.destructor(userData)	- free would qualify
 *    RefAny userData;
 *
 *    STREAMstatus
 *    newline.callback(stream, userData)
 *    STREAM *stream;			- the stream
 *    RefAny userData;
 *
 *    void
 *    newline.destructor(userData)	- free would qualify
 *    RefAny userData;
 */
typedef struct STREAM STREAM;	/* to match FILE */
struct STREAM {
    STREAMside input;
    Boolean firstReadP;
    DYNStrgeneralString unget;		/* of char (for input only) */
    int debits;			/* how many '\n' do we owe (don't prompt twice) */
    struct {
	STREAMpromptType type;
	RefAny userData;
	void (*destructor)();
	STREAMstatus (*callback)();
    } prompt;
    struct {
	STREAMnewlineType type;
	RefAny userData;
	void (*destructor)();
	STREAMstatus (*callback)();
	Boolean inhibitP;	/* prevent infinite death-recursion */
    } newline;
    STREAMside output;
};

extern STREAM *STREAMCreate();
extern STREAM *STREAMSplice();
extern void STREAMDestroy();

extern String STREAMGetName();
extern int STREAMGetLineno();
extern int STREAMGetColumnno();

extern String STREAMGetString();

extern void STREAMGetCharCounts();
extern void STREAMSetCharCounts();

extern void STREAMForcePrompt();
extern void STREAMSetPrompt();

extern void STREAMFlush();		/* flush pending output */
extern void STREAMClear();		/* clear pending output */

/*
 *    Input operations
 */
extern STREAMstatus STREAMScanf();	/* TODO - this one is hard */
extern STREAMstatus STREAMGets();	/* TODO - this is mildy hard (in RCS) */
extern STREAMstatus STREAMGetc();
extern void STREAMUngetc();

/*
 *    Output operations
 */
extern STREAMstatus STREAMPrintf();
extern STREAMstatus STREAMPuts();		/* no '\n' is appended */
extern STREAMstatus STREAMPutc();

#endif /* stream_h */
