#include "sumachine.h"

#include <sys/types.h>
#include <stdio.h>

#include <ctype.h>

#include "stuff.h"
#include "alloc.h"

#include "toargv.h"

static int count_words();
static void copy_a_word();
static int length_of_word();

#define isquote(ch)	 ((ch) == '"' || (ch) == '\'' || (ch) == '`')

/*
 *    ToArgv - return a NULL terminated argument vector in the manner of csh.
 *
 *    Respect csh quoting conventions.
 *
 *    return the argv.
 */
Stringv
ToArgv(line)
String line;
{
	int i,
	    word_len,		/* length of a word	*/
	    words;		/* words on a line	*/
	Stringv argv;

	words = count_words(line);
	argv = ALLOC_ARRAY(words+1, String);
	argv[words] = NULL;
	for ( i=0; i<words; i++) {
	    /*
	     *    While whitespace,  skip it.
	     *    Quoting / escapes will take care of white in words.
	     */
	    while (*line && isspace(*line))
		line++;
	    word_len = length_of_word(line);
	    argv[i] = ALLOC_ARRAY(word_len+1, char);
	    copy_a_word(argv[i], &line);	/* line is incremented */
	}
	return (argv);
}

/*
 *    The purpose of this function is to count all of the words
 *    on the line, remembering the backslashes...
 *
 *    Words are delimited by whitespace.  Leading and trailing
 *    whitespace is ignored.
 *
 *    return the count of words on this line.
 */
static int
count_words(buf)
register char *buf;
{
	register char quot = '\0';
	register int count = 0;		/* count of words	*/

	while (*buf && isspace(*buf))	/* skip initial white	*/
	    buf++;
	while (*buf) {
	    if (isquote(*buf)) {
		quot = *buf++;			/* save the quote	*/
		while (*buf && *buf != quot) {
		    buf += (*buf == '\\');	/* skip escape		*/
		    buf += (*buf != '\0');	/* maybe at eoln	*/
		}
		buf += (*buf != '\0');	/* over the quote (maybe)	*/
	    } else {
		while (*buf && !isspace(*buf)) {
		    buf += (*buf == '\\');	/* skip escape		*/
		    buf += (*buf != '\0');		/* maybe at eoln	*/
		}
	    }
	    /*
	     *    You will now be at eoln or whitespace.
	     *    You covered a word.  Look for the next one.
	     */
	    count++;
	    while (*buf && isspace(*buf))
		buf++;
	}
	return (count);
}

/*
 *    length_of_word - find the squeezed length of the current word.
 *
 *    The squeezed length is the length without the quotes and
 *    without '\\' escapes.   It is this length that needs
 *    to be malloc()-ed in chars to save the word.
 *
 *    return the squeezed length.
 */
static int
length_of_word(buf)
char *buf;
{
	register char quot = '\0';
	register int squeeze_len = 0;		/* length of squeezed word	*/

	while (*buf && !isspace(*buf)) {
	    if (*buf == '\\' || !isquote(*buf)) {
		/*
		 *    If skipping or not a quote char
		 *    If esc char, then inc approprately
		 *    count one more char.
		 */
		buf += (*buf == '\\');
		squeeze_len++, buf++;
	    } else {
		/*
		 *    If its a quoted character and not skipping
		 *    Remember the quote;
		 *    While not eoln, and unfound "the" quote
		 *        if skip, go over it.
		 *        count the current char.
		 *   Now after the quote (or at eoln), maybe.
		 */
		quot = *buf++;
		while (*buf && *buf != quot) {
		    buf += *buf == '\\';	/* move over escape	*/
		    squeeze_len += (*buf != '\0');/* maybe count one	*/
		    buf += (*buf != '\0');	/* maybe over one more	*/
		}
		/*
		 *    Off of final quote (maybe)
		 */
		buf += (*buf != '\0');	
	    } 
	}
	return (squeeze_len);
}

/*
 *    copy_a_word - Copy a word, skipping magic and quotes.
 *
 *    Also advance the from guy because he really doesn't know
 *    how long the word REALLY is, quotings, escapes and all.
 *  
 *    Side effects of filling to and saveQuote, and moving *from.
 */
static void
copy_a_word(to, from)
register char *to;			/* put at to			*/
register char **from;			/* word is at *from		*/
{
	register char quote = '\0';

	if (!isquote(**from)) {
	    while (**from && !isspace(**from)) {
		*from += (**from == '\\');
		*to++ = *(*from)++;
	    } 
	} else {
	    /*
	     *    save the quote, (*from)++.
	     *    while not eoln, and not at "the" quote
	     *        if skipping, then from++
	     *        copy one more char
	     */
	    quote = *(*from)++;
	    while (**from && **from != quote) {
		*from += (**from == '\\');	/* move over escape	*/
		*to = **from;
		to += (**from != '\0');		/* at eoln maybe	*/
		*from += (**from != '\0');		/* at eoln maybe	*/
	    }
	    *from += (**from != '\0');	/* skip over last quote (maybe)	*/
	} 
	*to = '\0';			/* NULL terminate it	*/
}
