#include "sumachine.h"

#include <sys/types.h>

#include <signal.h>
#include <stdio.h>
#include <errno.h>

#include "stuff.h"

/* must include AFTER sumachine.h (stuff.h) */
#if OPSYS_TYPE == MD_UNIX /* { */
#include <sys/socket.h>
#if OPSYS_VERS == MD_BSD_4_3 || OPSYS_VERS == MD_BSD_4_2 /* { */
#include <netinet/in.h>
#endif /* } */
#if OPSYS_VERS == MD_MC_2_2 /* { */
#include <net/in.h>
#endif /* } */
#if OPSYS_SUBTYPE != MD_AIX && \
	OPSYS_SUBTYPE != MD_HPUX && \
	OPSYS_SUBTYPE != MD_SYS5 /* { */
#include <sys/wait.h>	/* no wait.h or sys/wait.h on aix */
#endif /* } */
#endif /* } */
#if OPSYS_TYPE == MD_VMS /* { */
#include <socket.h>
#include <wait.h>	/* only wait.h on vms */
#endif /* } */

#include "alloc.h"
#include "twoway.h"

static String shell = "/bin/sh";
static int *close_pids;

#if OPSYS_VERS == MD_MC_2_2 /* { */
static int fake_socketpair();
static int fake_socketpair_accept();
static int fake_socketpair_connect();
#endif /* } */

/*
 *    TwoWayOpen - sortof popen, except that it creates a bidirectional stream
 *
 *    TwoWay ~ bidirectional
 *
 *    return {fd, NULL} on {success, fail}
 */
int
TwoWayOpen(shell_line)
String shell_line;
{
	int ntable, i, pid, got_pid, sv[2];

	if (close_pids == NULL) {
	    ntable = getdtablesize();
	    close_pids = ALLOC_ARRAY(ntable, int);
	    for (i=0; i<ntable; i++)
		close_pids[i] = -1;
	}

#if OPSYS_VERS == MD_BSD_4_3 /* { */
	if (socketpair(AF_UNIX, SOCK_STREAM, PF_UNSPEC, sv) < 0)
	    return (-1);
#endif /* } */
#if OPSYS_VERS == MD_BSD_4_2 /* { */
	if (socketpair(AF_UNIX, SOCK_STREAM, /* PF_UNSPEC */ NULL, sv) < 0)
	    return (-1);
#endif /* } */
#if OPSYS_VERS == MD_MC_2_2 /* { */
	if (fake_socketpair(sv) < 0)
	    return (-1);
#endif /* } */

	/*
	 *    DO NOT USE vfork() even if it exists
	 *    b/c both process must be free to run.
	 */
	switch (pid = fork()) {
	case -1:
	    close(sv[0]);
	    close(sv[1]);
	    return (-1);
	case 0:		/* child */
#if OPSYS_VERS == MD_MC_2_2 /* { */
	    if (fake_socketpair_accept(sv[0]) < 0)
		_exit (1);
#endif /* } */
	    ntable = getdtablesize();
	    for (i=0; i<ntable; i++) {
		if (i != sv[0])
		    close(i);
	    }
	    for (i=0; i<3; i++) {
		if (sv[0] != i)
		    dup2(sv[0], i);
	    }
	    if (sv[0] > 2)
		close(sv[0]);
	    execl(shell, "sh", "-c", shell_line, NULL);
	    _exit (1);
	    break;
	default:	/* parent */
	    close(sv[0]);
#if OPSYS_VERS == MD_MC_2_2 /* { */
	    if (fake_socketpair_connect(sv[1]) < 0) {
		kill(pid, SIGKILL);	/* blast it if its not dead already */
		errno = 0;
		while((got_pid = wait(&status)) != pid && got_pid != -1 && errno != EINTR)
		    errno = 0;
		return (-1);
	    }
#endif /* } */
	    close_pids[sv[1]] = pid;
	    break;
	}

	return (sv[1]);
}

/*
 *    TwoWayClose - close then 2 way channel
 *
 *    wait on the child process responsible.
 *
 *    return {w_retcode, -1} on {caught child, not caught child}
 */
int
TwoWayClose(fd)
int fd;
{
	int want_pid, got_pid;
#if OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	int status;
#else /* } else { */
	union wait status;
#endif /* } */

	want_pid = close_pids[fd];
	close_pids[fd] = -1;
	close(fd);

	errno = 0;
	while((got_pid = wait(&status)) != want_pid && got_pid != -1 && errno != EINTR)
	    errno = 0;

#if OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
	return (got_pid < 0 ? -1: status);
#else /* } else { */
	return (got_pid < 0 ? -1: status.w_retcode);
#endif /* } */
}

#if OPSYS_VERS == MD_MC_2_2 /* { */
static struct sockaddr_in child_sin;

static int
fake_socketpair(sv)
int sv[2];
{
	int ret;

	/*
	 *    This is so painful...
	 *    Let the system assign the address.
	 *
	 *    Be sure and get 0 vs 1 correct (see case 0 in switch (pid) above))
	 */
	sv[0] = socket(SOCK_STREAM, /* PF_UNSPEC */ NULL, /* sockaddr */ NULL, SO_ACCEPTCONN);
	sv[1] = socket(SOCK_STREAM, /* PF_UNSPEC */ NULL, /* sockaddr */ NULL, 0x0);

	/*
	 *    socketaddr doesn't quite do the whole job
	 */
	ret = socketaddr(sv[0], &child_sin);
	child_sin.sin_addr.s_addr = htonl(gethostid());

#ifdef DEBUG
	fprintf(stderr, "socket: sv = { %d, %d }, %d (%s)\n", sv[0], sv[1], ret, sys_errlist[errno]);
#endif DEBUG

	if (sv[0] < 0 || sv[1] < 0 || ret < 0) {
	    close(sv[0]);
	    close(sv[1]);
	    return (-1);
	}
	return (0);
}

static int
fake_socketpair_accept(at)
int at;
{
	int ret;
	struct sockaddr_in sin;

	ret = accept(at, &sin);
#ifdef DEBUG
	fprintf(stderr, "accept: ret = %d (%s)\tfrom sin = { %d, %d, %#x }\n",
		ret, sys_errlist[errno],
		sin.sin_family, ntohs(sin.sin_port), ntohl(sin.sin_addr.s_addr));
#endif DEBUG
	return (ret);
}

static int
fake_socketpair_connect(to)
int to;
{
	int ret;
	struct sockaddr_in sin;

	ret = socketaddr(to, &sin);
#ifdef DEBUG
	fprintf(stderr, "\
pre connect: (connect to) child_sin = { %d, %d, %#x }\n\
\t(connect_from) to_sin = { %d, %d, %#x }\n",
		child_sin.sin_family, ntohs(child_sin.sin_port), ntohl(child_sin.sin_addr.s_addr),
		sin.sin_family, ntohs(sin.sin_port), ntohl(sin.sin_addr.s_addr));
#endif DEBUG

	ret = connect(to, &child_sin);

#ifdef DEBUG
	fprintf(stderr, "connect: ret = %d (%s)\n", ret, sys_errlist[errno]);
#endif DEBUG
	return (ret);
}
#endif /* } */
