#if ! defined(sumachine_h) /* { */
#define sumachine_h

/*
 *    Define a ``Standard Unix Machine''
 *
 *    By whatever means are available in /lib/cpp, figure out
 *    what type of operating system, maker, machine architecture
 *    and architecture instance we are on.
 *
 *    OPSYS_TYPE	- the type of operating system (e.g. vms)
 *    OPSYS_SUBTYPE	- the subtype of operating system (e.g. bsd, sys5, etc)
 *    OPSYS_MFR		- the manufacturer (e.g. Hewlett-Packard)
 *    OPSYS_VERS	- the operating system version (e.g. sun 4.0)
 *
 *    CPU_TYPE		- the architecture (e.g. 68K)
 *    CPU_MODEL		- the architecture instance (e.g. 68010)
 *
 *    $Revision: 1.1.1.1 $
 *    $Date: 2002/12/31 15:51:27 $
 *    $Source: /home/cvs/octtools/src/OctLisp/src/sumachine/sumachine.h,v $
 *
 *    This file is recursively #included by "stuff.h", so every single src
 *    file can have access to its defines.   This file should be included in
 *    each and every src file compiled b/c it ensures that the environment which
 *    those src files expect exists either by fiat or by faking it.
 */

/*
 *    The OPSYS specifics
 *
 *    Parameterized with the best known /lib/cpp predefines
 */
#if defined(bsd42) && ! defined(bsd_4_2) /* { */
/* for the sony news */
#define bsd_4_2	bsd42
#endif /* } */
#if defined(ultrix) /* { */
#define bsd_4_2	bsd4_2 
#endif /* } */

#if defined(unix) /* { */
#define OPSYS_TYPE	MD_UNIX
#else /* } else { */
#if defined(vms) /* { */
#define OPSYS_TYPE	MD_VMS
#define OPSYS_MFR	MD_DEC
#define OPSYS_VERS	MD_VMS_500
#endif /* } */
/* else undefined */
#endif /* } */
#if defined(u370) /* { */
#define OPSYS_TYPE	MD_IBM370_OS
#define OPSYS_MFR	MD_IBM
#define OPSYS_VERS	1	/* dunno */
#endif /* } */

/*
 *    WATCHOUT - the newer gcc doesn't define __GNU__
 *    So we may have to define __GNU__ ourselves
 *
 *    __GNU__ is the old way
 *    __GNUC__ is the new way
 *
 *    Use gcc -v somefile.c to determine what is being passed.
 */
#if defined(__GNU__) || defined(__GNUC__) /* { */
/*
 *    GNU doesn't know about any opsys-specific
 *    cpp tokens, so we must figure them out for it.
 *    GNU does however know about architecture-tokens
 */
#if defined(vax) /* { */
#if ! defined(ultrix) /* { */
/* just in case it was defined before... */
#define ultrix	1	/* what the hell (all vaxes running gnu are ultrix) */
#endif /* } */
/*
 *    yes of course this is not true, but how can I figure out
 *    ultrix vs 4.2 vs 4.3 - there ARE differences which must be handled
 */
#endif /* } */
#endif /* } */

#if defined(ultrix) /* { */
/* then defined(vax) || defined(mips) */
/* the r2000 and r3000 are indistinguishable (so far) */
#define OPSYS_MFR	MD_DEC
#define OPSYS_SUBTYPE	MD_ULTRIX
#include <sys/types.h>	/* can be multiply included safely */
#include <signal.h>	/* can be multiply included safely */
#if defined(_SIZE_T_) /* { */
#define OPSYS_VERS	MD_ULTRIX32_4_0
#else /* } else { */
/* ULTRIX32_3_1 is indistinguishable from 3.0 (so far) */
#if defined(SIGLOST) /* { */
#define OPSYS_VERS	MD_ULTRIX32_3_0
#else /* } else { */
#if defined(MAXSELFD) /* { */
#define OPSYS_VERS	MD_ULTRIX32_2_4
#else /* } else { */
#define OPSYS_VERS	MD_ULTRIX32_2_2
#endif /* } */
#endif /* } */
#endif /* } */
#endif /* } */
#if defined(vax) && ! defined(ultrix) /* { */
#if defined(tahoe) /* { */
#define OPSYS_MFR	MD_UCB
#define OPSYS_SUBTYPE	MD_BSD
#define OPSYS_VERS	MD_BSD_4_3
#else /* } else { */
#include <sys/types.h>	/* can be multiply included safely */
#if defined(FD_SETSIZE) /* { */
#define OPSYS_MFR	MD_UCB
#define OPSYS_SUBTYPE	MD_BSD
#define OPSYS_VERS	MD_BSD_4_3
#else /* } else { */
#define OPSYS_MFR	MD_UCB
#define OPSYS_SUBTYPE	MD_BSD
#define OPSYS_VERS	MD_BSD_4_2
#endif /* } */
#endif /* } */
#endif /* } */
#if defined(sun) /* { */
#define OPSYS_MFR	MD_SUN
#define OPSYS_SUBTYPE	MD_BSD
#include <sys/types.h>	/* can be multiply included safely */
/* what if yer on a sun2/sun3? - the following takes care of most of it */
#if defined(FD_SETSIZE) /* { */
#define OPSYS_VERS	MD_SUN_4_0
#else /* } else { */
#define OPSYS_VERS	MD_SUN_3_5
#endif /* } */
#endif /* } */
#if defined(sony_news) /* { */	/* sony_news == news8000 */
#define OPSYS_MFR	MD_SONY
#define OPSYS_SUBTYPE	MD_BSD
/*
 *    WATCHOUT - sony should have its own OPSYS_VERS space just in case
 *    WATCHOUT - they decide to diverge from the 4.2/4.3 development line
 *    WATCHOUT - everybody does after a while...
 */
#include <signal.h>	/* can be multiply included safely? */
#if ! defined(sigmask) /* { */
#define OPSYS_VERS	MD_BSD_4_2	/* for lack of a better version number */
#else /* } else { */
#define OPSYS_VERS	MD_BSD_4_3	/* for lack of a better version number */
#endif /* } */
#endif /* } */
#if defined(hp300) || defined(hp9000s200) || defined(hp9000s300) /* { */
#if defined(hpux) /* { */
#define OPSYS_MFR	MD_HP
#define OPSYS_SUBTYPE	MD_HPUX
/*
 *    Apparently they are all the way up on hpux 8.0?
 *    We don't run hpux here at Berkeley any more so I can't reall tell...
 */
#include <sys/types.h>		/* can do this multiple times */
#if defined(TYPES_INCLUDED) /* { */
#define OPSYS_VERS	MD_HPUX_5_5	/* uses TYPES_INCLUDED */
#else /* } else { */
#define OPSYS_VERS	MD_HPUX_6_5	/* use _TYPES_INCLUDED */
#endif /* } */
#else /* } else { */
/* the (utah?) bsd port */
#define OPSYS_MFR	MD_HP
#define OPSYS_SUBTYPE	MD_BSD
#define OPSYS_VERS	MD_BSD_4_3
#endif /* } */
#endif /* } */
#if defined(sequent) /* { */
#define OPSYS_MFR	MD_SEQUENT
#define OPSYS_SUBTYPE	MD_DYNIX
/*
 *    There is a 2.1 and we are now (here at berkeley) on either 3.0.12 or 3.0.16
 *    But I can't detect anything different about the two right now (erg).
 */
#define OPSYS_VERS	MD_DYNIX_3_0_12
#endif /* } */
#if defined(aiws) /* { */
#define OPSYS_MFR	MD_IBM
#define OPSYS_SUBTYPE	MD_AIX
#include <sys/types.h>	/* can be multiply included safely */
#if defined(NBBY) /* { */
/*
 *    Define BSD_INCLUDES for all compilation (grr!)
 */
#define OPSYS_VERS	MD_AIX_2_2_1
#else /* } else { */
#define OPSYS_VERS	MD_AIX_2_2
#endif /* } */
#endif /* } */
#if defined(ibm032) /* { */
#define OPSYS_MFR	MD_IBM
#define OPSYS_SUBTYPE	MD_AICS
#define OPSYS_VERS	MD_AICS_4_3_2
#endif /* } */
#if defined(u3b) /* { */
#define OPSYS_MFR	MD_ATT
#define OPSYS_SUBTYPE	MD_SYS5
#define OPSYS_VERS	MD_SYS5_3_11
#endif /* } */
#if defined(u3b5) /* { */
#define OPSYS_MFR	MD_ATT
#define OPSYS_SUBTYPE	MD_SYS5
#define OPSYS_VERS	MD_SYS5_3_11
#endif /* } */
#if defined(gould) /* { */
#define OPSYS_MFR	MD_GOULD
#define OPSYS_SUBTYPE	1	/* dunno anything about goulds */
#define OPSYS_VERS	1	/* dunno anything about goulds */
#endif /* } */
#if defined(apollo) /* { */
#define OPSYS_MFR	MD_APOLLO
#define OPSYS_SUBTYPE	MD_DOMAIN
#include <sys/types.h>	/* can be multiply included safely */
/*
 * there is also an apollo 10.1, but there is no way to distinguish
 * between it and apollo 10.0.1 from the #include file level
 */
#if defined(FD_SETSIZE) /* { */
#define OPSYS_VERS	MD_APOLLO_10_0_1
#else /* } else { */
/* there is also a 9.2.3, but how to tell that from 9.7 ? */
/* there is also a 9.5, but how to tell that from 9.7 ? */
/* there is also a 9.7.5, but how to tell that from 9.7 ? */
#define OPSYS_VERS	MD_APOLLO_9_7
#endif /* } */
#endif /* } */
#if defined(uts) /* { */
#define OPSYS_MFR	MD_AMDAHL
#define OPSYS_SUBTYPE	MD_UTS
#define OPSYS_VERS	MD_AMDAHL_UTS
#endif /* } */

/*
 *    THE CPU specifics
 *
 *    Parameterized with the best known /lib/cpp predefines
 *    Set up for a middle-of-the-road machine.
 */
#if defined(mips) /* { */
/*
 *    DECstation	Processor
 *       2100		  R2000 (slower clock)
 *       3100		  R2000
 *       5000		  R3000
 *       5400		  R3000
 *       ????		  R6000
 */
/*
 *    We might be on the R3000 machine, but there is really
 *    no way of distinguishing that from the src-code level.
 */
#define CPU_TYPE	MD_MIPS
#define CPU_MODEL	MD_R2000
#endif /* } */
#if defined(vax) /* { */
#define CPU_TYPE	MD_VAX
/* not clear how to figure any further - probably not necessary though */
#define CPU_MODEL	MD_VAX8650
#endif /* } */
#if defined(sun) /* { */
#if defined(mc68000) && ! (defined(mc68010) || defined(mc68020)) /* { */
#define CPU_TYPE	MD_MC68K
#define CPU_MODEL	MD_MC68000
#endif /* } */
#if defined(mc68000) && ! defined(mc68020) /* { */
#define CPU_TYPE	MD_MC68K
#define CPU_MODEL	MD_MC68010
#endif /* } */
#if defined(mc68000) && ! defined(mc68010) /* { */
#define CPU_TYPE	MD_MC68K
#define CPU_MODEL	MD_MC68020
#endif /* } */
#if defined(sparc) /* { */
#define CPU_TYPE	MD_SPARC
#define CPU_MODEL	MD_SPARC_CPU
#endif /* } */
#if defined(sun386) /* { */
/* well sun386 is the cpp token anyway */
#define CPU_TYPE	MD_I386
#define CPU_MODEL	MD_I80386
#endif /* } */
#endif /* } */
#if defined(sony_news) /* { */
#define CPU_TYPE	MD_MC68K
#define CPU_MODEL	MD_MC68020
#endif /* } */
#if defined(hp300) || defined(hp9000s200) || defined(hp9000s300) /* { */
#define CPU_TYPE	MD_MC68K
#define CPU_MODEL	MD_MC68010
#endif /* } */
#if defined(sequent) /* { */
#if defined(i386) /* { */
#define CPU_TYPE	MD_I386
#define CPU_MODEL	MD_I80386
#endif /* } */
#if defined(ns32000) /* { */
#define CPU_TYPE	MD_NS32
#define CPU_MODEL	MD_NS32032
#endif /* } */
#endif /* } */
#if defined(aiws) || defined(ibm032) /* { */
#define CPU_TYPE	MD_IBM_RT
#define CPU_MODEL	MD_IBM_RT_CPU
#endif /* } */
#if defined(gould) /* { */
#define CPU_TYPE	MD_GOULD
#define CPU_MODEL	MD_GOULD_CPU
#endif /* } */
#if defined(u370) /* { */
#define CPU_TYPE	MD_IBM370
#define CPU_MODEL	MD_IBM370_CPU
#endif /* } */
#if defined(apollo) /* { */
#define CPU_TYPE	MD_MC68K
#define CPU_MODEL	MD_MC68020
#endif /* } */
#if defined(national) /* { */
#define CPU_TYPE	MD_NSC32
#define CPU_MODEL	MD_NSC32032
#endif /* } */
#if defined(uts) /* { */
#define CPU_TYPE	MD_AMDAHL_N
#define CPU_MODEL	MD_AMDAHL_N_CPU
#endif /* } */

#if ! defined(OPSYS_TYPE) || \
	! defined(OPSYS_MFR) || \
	! defined(OPSYS_VERS) || \
	! defined(CPU_TYPE) || \
	! defined(CPU_MODEL) /* { */
_OPSYS_TYPE_	OPSYS_TYPE
_OPSYS_MFR_	OPSYS_MFR
_OPSYS_VERS_	OPSYS_VERS
_CPU_TYPE_	CPU_TYPE
_CPU_MODEL_	CPU_MODEL
@@ ABORT @@  No sumachine.h incomplete: reinstall
#endif /* } */

/*
 *    The following MUST be literals else /lib/cpp cannot do
 *    == operations on them.  Remember cpp does strcmps, not MATH
 *
 *    OPSYS_TYPE
 *
 *    #if OPSYS_TYPE == MD_opsystype
 *    #endif
 */
#define MD_UNIX		1001000	/* a master type */
#define MD_VMS		1002000	/* a master type */
#define MD_IBM370_OS	1003000	/* a master type */
#define MD_IBM_PC_AT_OS	1004000	/* a master type */

/*
 *    OPSYS_SUBTYPE
 *
 *    #if OPSYS_SUBTYPE == MD_opsystype
 *    #endif
 */
#define MD_BSD		1001001	/* berkeley's subtype of unix */
#define MD_ULTRIX	1001002	/* dec's subtype of unix */
#define MD_SYS5		1001003	/* att's subtype of unix */
#define MD_HPUX		1001004	/* hp's subtype of unix */
#define MD_DYNIX	1001005	/* sequent's subtype of unix */
#define MD_AIX		1001006	/* ibm's subtype of unix */
#define MD_UTS		1001007	/* amdahl's subtype of unix */
#define MD_DOMAIN	1001008	/* apollo's subtype of unix */
#define MD_AICS		1001009	/* ibm's academic subtype of unix */

/*
 *    OPSYS_MFR
 *
 *    #if OPSYS_MFR == MD_opsysmfr
 *    #endif
 */
#define MD_UCB		2000000
#define MD_ATT		2000001
#define MD_MASSCOMP	2000002
#define MD_DEC		2000003
#define MD_SUN		2000004
#define MD_WOLLONGONG	2000005
#define MD_IBM		2000007
#define MD_HP		2000008
#define MD_APOLLO	2000009
#define MD_SONY		2000010
#define MD_SEQUENT	2000011
#define MD_AMDAHL	2000012
#define MD_GOULD	2000012

/*
 *    OPSYS_VERS
 *
 *    #if OPSYS_VERS == MD_opsysvers
 *    #endif
 *    #if OPSYS_MFR == MD_opsysmfr && OPSYS_VERS <= MD_opsysvers
 *    #endif
 */
#define MD_BSD_4_0		3000400
#define MD_BSD_4_1		3000410
#define MD_BSD_4_2		3000420
#define MD_BSD_4_3		3000430
#define MD_BSD_4_4		3000440

#define MD_ULTRIX32_1_0		3010100
#define MD_ULTRIX32_1_1		3010110
#define MD_ULTRIX32_1_2		3010120
#define MD_ULTRIX32_1_3		3010130
#define MD_ULTRIX32_1_4		3010140
#define MD_ULTRIX32_2_0		3010200
#define MD_ULTRIX32_2_1		3010210
#define MD_ULTRIX32_2_2		3010220
#define MD_ULTRIX32_2_3		3010230
#define MD_ULTRIX32_2_4		3010240
#define MD_ULTRIX32_2_5		3010250
#define MD_ULTRIX32_2_6		3010260
#define MD_ULTRIX32_3_0		3010300
#define MD_ULTRIX32_3_1		3010310
#define MD_ULTRIX32_4_0		3010400
#define MD_ULTRIX32_4_1		3010401
#define MD_ULTRIX32_4_2		3010402
#define MD_ULTRIX32_4_3		3010402

#define MD_SUN_1_4		3020140
#define MD_SUN_2_0		3020200
#define MD_SUN_3_0		3020300
#define MD_SUN_3_1		3020310
#define MD_SUN_3_2		3020320
#define MD_SUN_3_3		3020330
#define MD_SUN_3_4		3020340
#define MD_SUN_3_5		3020350
#define MD_SUN_4_0		3020400
#define MD_SUN_4_1		3020410
#define MD_SUN_4_2		3020420
#define MD_SUN_4_3		3020430

#define MD_SYS5_2_04		3030204
#define MD_SYS5_3_11		3030311
#define MD_SYS5_5_0		3030500

#define MD_MC_2_1		3040210
#define MD_MC_2_2		3040220
#define MD_MC_3_0		3040300

#define MD_IBM_PC_AT		3060000

#define MD_HPUX_5_02		3070502
#define MD_HPUX_5_5		3070550
#define MD_HPUX_5_6		3070560
#define MD_HPUX_5_7		3070570
#define MD_HPUX_5_8		3070580
#define MD_HPUX_6_0		3070600
#define MD_HPUX_6_1		3070610
#define MD_HPUX_6_2		3070620
#define MD_HPUX_6_3		3070630
#define MD_HPUX_6_4		3070640
#define MD_HPUX_6_5		3070650
#define MD_HPUX_6_6		3070660
#define MD_HPUX_7_0		3070700	/* I've never seen any of these */
#define MD_HPUX_7_1		3070710
#define MD_HPUX_7_2		3070720
#define MD_HPUX_8_0		3070800

#define MD_APOLLO_5_9		3080590	/* plus ancien */
#define MD_APOLLO_6_0		3080600	/* ancien */
#define MD_APOLLO_9_2_3		3080923
#define MD_APOLLO_9_5		3080950
#define MD_APOLLO_9_7		3080970
#define MD_APOLLO_10_0_1	3081001
#define MD_APOLLO_10_1		3081010
#define MD_APOLLO_10_2		3081020

#define MD_VMS_500		3090500
#define MD_VMS_600		3090600

#define MD_DYNIX_2_0		3100200
#define MD_DYNIX_2_1		3100210
#define MD_DYNIX_2_2		3100220
#define MD_DYNIX_3_0		3100300
#define MD_DYNIX_3_0_1		3100301
#define MD_DYNIX_3_0_2		3100302
#define MD_DYNIX_3_0_3		3100303
#define MD_DYNIX_3_0_4		3100304
#define MD_DYNIX_3_0_5		3100305
#define MD_DYNIX_3_0_6		3100306
#define MD_DYNIX_3_0_7		3100307
#define MD_DYNIX_3_0_8		3100308
#define MD_DYNIX_3_0_9		3100309
#define MD_DYNIX_3_0_10		3100310
#define MD_DYNIX_3_0_11		3100311
#define MD_DYNIX_3_0_12		3100312
#define MD_DYNIX_3_0_13		3100313
#define MD_DYNIX_3_0_14		3100314
#define MD_DYNIX_3_0_15		3100315
#define MD_DYNIX_3_0_16		3100316
#define MD_DYNIX_3_0_17		3100317
#define MD_DYNIX_3_0_18		3100318
#define MD_DYNIX_3_0_19		3100319
#define MD_DYNIX_3_0_20		3100320

#define MD_AIX_2_0		3110200
#define MD_AIX_2_1		3110210
#define MD_AIX_2_2		3110220
#define MD_AIX_2_2_1		3110221
#define MD_AIX_2_3		3110230
#define MD_AIX_2_4		3110240

#define MD_AMDAHL_UTS		3120000

#define MD_AICS_4_3		3110430
#define MD_AICS_4_3_1		3110431
#define MD_AICS_4_3_2		3110432
#define MD_AICS_4_3_3		3110433
#define MD_AICS_4_3_4		3110434

/*
 *    CPU_TYPE
 *
 *    #if CPU_TYPE == MD_cputype
 *    #endif
 */
#define MD_VAX		4000000
#define MD_MC68K	4000001
#define MD_NSC32	4000002
#define MD_NSC16	4000003
#define MD_SPARC	4000004
#define MD_MC88K	4000005
#define MD_IBM_RT	4000006
#define MD_I286		4000007		/* these are all different right? */
#define MD_I386		4000008		/* these are all different right? */
#define MD_I486		4000009		/* these are all different right? */
#define MD_I586		4000010		/* these are all different right? */
#define MD_3B		4000011
#define MD_PDP11	4000012
#define MD_MIPS		4000013
#define MD_AMDAHL_N	4000014		/* there is some number here */

/*
 *    CPU_MODEL
 *
 *    #if CPU_MODEL == MD_modelname
 *    #endif
 */
#define MD_VAX730	5000000		/* for the most part non-differentiable */
#define MD_VAX750	5000001
#define MD_VAX780	5000002
#define MD_VAX785	5000003
#define MD_VAX8600	5000004
#define MD_VAX8650	5000005
#define MD_VAX8800	5000006
#define MD_VAX8850	5000007
#define MD_VAX8900	5000008
#define MD_VAX8950	5000009
#define MD_UVAX_1	5000010
#define MD_UVAX_2	5000011
#define MD_UVAX_3	5000012
#define MD_VAXSTAR	5000013

#define MD_MC68000	5001000
#define MD_MC68010	5001010
#define MD_MC68020	5001020
#define MD_MC68030	5001030
#define MD_MC68040	5001040
#define MD_MC68050	5001050		/* hah */

#define MD_NSC16032	5002000
#define MD_NSC32032	5003001

#define MD_SPARC1	5004001
#define MD_SPARC2	5004002
#define MD_SPARC3	5004003

#define MD_MC88000	5005000

#define MD_IBM_RT_CPU	5006000

#define MD_I80286	5007020
#define MD_I80386	5008030
#define MD_I80486	5009040
#define MD_I80586	5010040		/* hah (again) */

#define MD_3B2		5011020
#define MD_3B5		5011050
#define MD_3B20		5011200

#define MD_PDP11_44	5012044
#define MD_PDP11_70	5012070

#define MD_MIPS_R1000	5013000		/* did this ever exist? */
#define MD_MIPS_R2000	5013002
#define MD_MIPS_R3000	5013003
#define MD_MIPS_R6000	5013006

#define MD_AMDAHL_N_CPU	5016000

#if ! defined(display_sh) /* { */
/* don't add this stuff in if the query program display.sh is being run */

/*******************************************************************************
 *
 *    Below here are the hacks and warts that sumachine performs in order
 *    to provide a comon environment to all systems.   Whatever tricks
 *    and hacks are required are used to make the environment look ``enough''
 *    like a BSD/Ultrix/Sun system to allow the code to compile.
 *
 ******************************************************************************/

#if (OPSYS_MFR == MD_IBM && OPSYS_SUBTYPE == MD_AICS) /* { */
#if defined(__HIGHC__) /* { */
/* metaware's hc ``HighC'' compiler */
#undef __STDC__	/* their __STDC__ option is from hell - everything breaks */
#endif /* } */
#endif /* } */

/* most machines don't give us a header file for these */
#if (OPSYS_MFR == MD_HP && OPSYS_SUBTYPE != MD_HPUX) || \
	OPSYS_MFR == MD_UCB || \
	(OPSYS_MFR == MD_IBM && OPSYS_SUBTYPE == MD_AICS) || \
	OPSYS_SUBTYPE == MD_DYNIX || \
	OPSYS_MFR == MD_SUN || \
	OPSYS_MFR == MD_SONY /* { */
/*
 *    vanilla 4.3 || ibm aics 4.3 || sequent ||
 *    hp 4.3 (utah) || sun 3.5/4.0, || sony 4.2/4.3
 */
extern int abort();
extern int free();
extern int exit();
extern int perror();
#else /* } else { */
#if OPSYS_MFR == MD_HP && OPSYS_SUBTYPE == MD_HPUX /* { */
/* hpux 5.0 */
extern int abort();
extern void free();
extern void exit();
extern void perror();
#else /* } else { */
/* ``modern'' machines  */
extern void abort();
extern void free();
extern void exit();
extern void perror();
#endif /* } */
#endif /* } */
extern char *getenv();
extern void setenv();

extern char *sprintf();

extern char *malloc();
extern char *realloc();
extern char *sbrk();
extern char *brk();

extern int system();
extern double atof();	/* math.h */

#if ! defined(__STDC__) || OPSYS_MFR == MD_APOLLO || defined(__GNU__) /* { */
#if OPSYS_MFR != MD_APOLLO /* { */
/* you can't mess with __DATE__ or __TIME__ on an apollo without a warning */
/*
 *    remember both apollo and gnu lie to us about the ansii world
 *    external to the compiler that they provide.   Fake it for them.
 */
#if ! defined(__DATE__) /* { */
/* Pretend ANSII compatibility __STDC__ is supposed to do this for us */
#if defined(CUR_DATE) /* { */
#define __DATE__        CUR_DATE
#else /* } else { */
#define __DATE__        "some-date"
#endif /* } */
#endif /* } */
#if ! defined(__TIME__) /* { */
/* Pretend ANSII compatibility __STDC__ is supposed to do this for us */
#if defined(CUR_TIME) /* { */
#define __TIME__        CUR_TIME
#else /* } else { */
#define __TIME__        "some-time"
#endif /* } */
#endif /* } */
#endif /* } */
#endif /* } */

#if defined(FILE) /* { */
#include <stdio.h>
/* Some machines fail to define some functions in stdio.h */
extern FILE *popen(), *tmpfile();
extern int pclose();
#endif /* } */

#if OPSYS_MFR == MD_DEC && OPSYS_VERS >= MD_ULTRIX32_4_0 /* { */
/* X/Open compat */
#include <string.h>
#endif /* } */

extern int mkstemp();
extern char *mktemp();

extern char *index();
extern char *rindex();

/* ANSI C string.h -- 1/11/88 Draft Standard */
extern char *strcpy();
extern char *strncpy();
extern char *strcat();
extern char *strncat();
extern char *strerror();
extern char *strpbrk();
extern char *strtok();
extern char *strchr();
extern char *strrchr();
extern char *strstr();
#if OPSYS_MFR == MD_DEC && OPSYS_VERS >= MD_ULTRIX32_4_0 /* { */
/* X/Open compat */
extern void *memmove();
extern void *memccpy();
extern void *memchr();
extern void *memcpy();
extern void *memset();
#else /* } else { */
extern char *memmove();
extern char *memccpy();
extern char *memchr();
extern char *memcpy();
extern char *memset();
#endif /* } */
extern int memcmp();
#if OPSYS_TYPE == MD_UNIX /* { */
extern int strcmp();
extern int strncmp();
extern int strcoll();
#if OPSYS_MFR == MD_DEC && OPSYS_VERS >= MD_ULTRIX32_4_0 /* { */
/* X/Open compat */
extern size_t strxfrm();
extern size_t strlen();
extern size_t strspn();
extern size_t strcspn();
#else /* } else { */
extern int strxfrm();
extern int strlen();
extern int strspn();
extern int strcspn();
#endif /* } */
#endif /* } */

#if OPSYS_TYPE == MD_VMS /* { */
#include <errno.h>	/* this seems to be the only safe way to do this */
#include <perror.h>	/* this seems to be the only safe way to do this */
#else /* } else { */
#if OPSYS_TYPE == MD_UNIX /* { */
/*
 *    errno.h and perror.h exist under some unixes also, but not ultrix for
 *    example, only errno.h exists under ultrix.   We are allowed define these
 *    here without problems such was not the case under VMS
 */
extern char *sys_errlist[];
extern int errno;
extern int sys_nerr;
extern char *sys_siglist[];
/* NSIG is the sizeof_el () the array */
#endif /* } */
#endif /* } */

/*
 *    int32		- an integer that holds (at least) 32 bits
 *    int16		- an integer that holds (at least) 16 bits
 *
 *    Make some code here which will define these typedefs or die.
 */
#if CPU_TYPE == MD_VAX || \
	CPU_TYPE == MD_MIPS || \
	CPU_TYPE == MD_MC68K || \
	CPU_TYPE == MD_SPARC || \
	CPU_TYPE == MD_I386 || \
	CPU_TYPE == MD_IBM_RT /* { */
typedef long    int32;
typedef short   int16;
#define MACHDEP_INCLUDED	/* now short circuit oct's #include "machdep.h" */
#else /* } else { */
DONT KNOW WHAT AN INT vs A SHORT IS ...
this should not compile ever!@!
#endif /* } */

/*
 *    hacks and things.
 *
 *    Note that it possible to do just about anything here
 *    EXCEPT conditionally define things based on the inclusion
 *    of #include files.  Typically these include files are
 *    system-dependent (e.g. <dir.h> vs <sys/dir.h>) and so must
 *    be included based on sumachine.h!
 *
 *    If you must supply something for a machine, then do it
 *    at all times.  Tougho for the machines that don't provide
 *    the right stuff - they'll compile a bit slower that's all.
 */
#if OPSYS_MFR == MD_MASSCOMP /* { */
#include <sys/types.h>		/* can we do this multiple times? */
/* 
 *    This is disgusting.
 *    no_u_ lets you turn this off (<ndir.h> creates some, but not all) 
 *    This is done because some things define some but not all of these
 *    for compatiblity with other (better) systems.
 */
#ifndef no_u_char /* { */
typedef unsigned char	u_char;
#endif no_u_char /* } */
#ifndef no_u_short /* { */
typedef unsigned short	u_short;
#endif no_u_short /* } */
#ifndef no_u_int /* { */
typedef unsigned int	u_int;
#endif no_u_int /* } */
#ifndef no_u_long /* { */
typedef unsigned long	u_long;
#endif no_u_long /* } */
#endif /* } */
#if (OPSYS_SUBTYPE == MD_AIX && OPSYS_VERS <= MD_AIX_2_2) || \
    OPSYS_SUBTYPE == MD_SYS5 || \
    OPSYS_TYPE == MD_VMS /* { */
/* NOT OPSYS_SUBTYPE == MD_HPUX b/c they take care of this in <sys/types.h> */
typedef unsigned char	u_char;
typedef unsigned short	u_short;
typedef unsigned int	u_int;
typedef unsigned long	u_long;
#endif /* } */
#if OPSYS_SUBTYPE == MD_AIX /* { */
#if defined(_h_file) /* { */	/* if sys/file.h has been included */
#include <fcntl.h>		/* can be multiply included */
#endif /* } */
#endif /* } */
#if OPSYS_SUBTYPE == MD_HPUX /* { */
/*
 *    WATCHOUT - Assumes that sumachine.h is included after all the
 *    system includes.  This is no longer valid; sumachine.h is to
 *    be included first; this will require a bit of porting later...
 */
#if defined(FOPEN) /* { */	/* if sys/file.h has been included */
#include <fcntl.h>		/* can be multiply included */
#endif /* } */
#endif /* } */
#if OPSYS_SUBTYPE == MD_SYS5 /* { */
#define	O_RDONLY	000
#define	O_WRONLY	001
#define	O_RDWR		002
#define	O_NDELAY	FNDELAY
#define	O_APPEND	FAPPEND
#define	O_CREAT		FCREAT
#define	O_TRUNC		FTRUNC
#define	O_EXCL		FEXCL
#endif /* } */
#if OPSYS_SUBTYPE == MD_AIX || \
    OPSYS_SUBTYPE == MD_HPUX || \
    OPSYS_SUBTYPE == MD_SYS5 /* { */
/* can make a pointer to struct passwd without knowing what it is */
/* can't multiply include pwd.h */
struct passwd *getpwent(), *getpwuid(), *getpwnam();
#endif /* } */
#if OPSYS_MFR == MD_MASSCOMP || \
	OPSYS_SUBTYPE == MD_HPUX || \
	OPSYS_SUBTYPE == MD_SYS5 /* { */
#define MAXPATHLEN	1024
#endif /* } */
#if OPSYS_TYPE == MD_VMS /* { */
#define MAXPATHLEN	256	/* YEP! - that's right, not 1024, but 256 */
#endif /* } */
#if OPSYS_MFR == MD_MASSCOMP || \
	OPSYS_SUBTYPE == MD_AIX || \
	OPSYS_TYPE == MD_VMS /* { */
#define L_SET		0
#define L_INCR		1
#define L_XTND		2
#if (OPSYS_SUBTYPE == MD_AIX && OPSYS_VERS <= MD_AIX_2_2) /* { */
#define F_OK		0x0
#define X_OK		0x1
#define W_OK		0x2
#define R_OK		0x4
#endif /* } */
#endif /* } */
#if (OPSYS_SUBTYPE == MD_AIX && OPSYS_VERS <= MD_AIX_2_2) /* { */
/* aix doesn't provide these */
#include <stdio.h>	/* can be multiply included */
#include <sys/dir.h>	/* can be multiply included */
typedef struct {
    int fd;
    struct direct buf;
} DIR;
extern DIR *opendir();
extern struct direct *readdir();
extern long telldir();
extern void seekdir();
#define rewinddir(dirp) seekdir((dirp), (long)0)
extern	void closedir();
#endif /* } */
#if ! defined(_U) /* { */
#include <ctype.h>		/* can't be multiply included - damnit */
#endif /* } */
#if OPSYS_MFR == MD_DEC && \
	(OPSYS_VERS >= MD_ULTRIX32_2_0 && OPSYS_VERS < MD_ULTRIX32_4_0) /* { */
/*
 *    The idea is to don't call the routines - just call the macros 
 *    but this trick goes away under Ultrix 4.0, but that's okay,
 *    it doesn't matter any more either.
 */
#define tolower(x) _tolower(x)
#define toupper(x) _toupper(x)
#else /* } else { */
#if (OPSYS_MFR == MD_SONY && OPSYS_VERS == MD_BSD_4_2) /* { */
/* sony ctype.h is the old 4.2 ctype.h - make it look like 4.3's */
#undef toupper
#undef tolower
#define isgraph(x)	(isprint(x) && ! isspace(x))
#define _toupper(c)	((unsigned char)(c)-'a'+'A')
#define _tolower(c)	((unsigned char)(c)-'A'+'a')
#define toupper(c)	(islower(c) ? _toupper(c): (c))
#define tolower(c)	(isupper(c) ? _tolower(c): (c))
#else /* } else { */
/*
 *    On suns and the like which don't have the
 *    distinction between tolower and _tolower etc.
 */
#if ! defined(_tolower) /* { */
/* if the macros don't exist, call the routines (which will be macros) */
#define _tolower(x) tolower(x)
#define _toupper(x) toupper(x)
#endif /* } */
#endif /* } */
#endif /* } */
#if (OPSYS_MFR == MD_DEC && OPSYS_VERS < MD_ULTRIX32_2_0) || \
	(OPSYS_MFR == MD_SONY && OPSYS_VERS == MD_BSD_4_2) /* { */
#define sigmask(m)	(1 << ((m)-1))		/* straight from 4.3 */
#endif /* } */
#if (OPSYS_MFR == MD_DEC && OPSYS_VERS < MD_ULTRIX32_2_4) || \
	OPSYS_MFR == MD_SUN || \
	OPSYS_MFR == MD_SONY || \
	(OPSYS_MFR == MD_UCB && OPSYS_VERS == MD_BSD_4_2)  /* { */
#include <sys/param.h>		/* can be multiply included safely */
#ifndef MAXHOSTNAMELEN /* { */
#define MAXHOSTNAMELEN		256		/* from the hostname man page */
#endif /* } */

#if ! (OPSYS_MFR == MD_SUN && OPSYS_VERS >= MD_SUN_4_0) && \
	! (OPSYS_MFR == MD_SONY && OPSYS_VERS >= MD_BSD_4_3) /* { */
/*
 *    A little different than 4.3 as the typedef in types.h is
 *    not for FD_SETSIZE, but rather for 1.  (the 32 bit restriction applies)
 */
#ifndef NBBY		/* in sys/param.h */
#define	NBBY		8		/* number of bits in a byte */
#endif NBBY 
#ifndef howmany		/* in sys/param.h */
#define	howmany(x, y)	(((x)+((y)-1))/(y))
#endif howmany

#define	FD_SETSIZE	1
#define NFDBITS	(sizeof(int) * NBBY)	/* bits per mask */

#define	FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define	FD_CLR(n, p)	((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define	FD_ISSET(n, p)	((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)	bzero((char *)(p), sizeof(*(p)))
#endif /* } */
#endif /* } */
#if (OPSYS_MFR == MD_DEC && OPSYS_VERS < MD_ULTRIX32_2_4) || \
    (OPSYS_MFR == MD_SUN && OPSYS_VERS < MD_SUN_4_0) || \
    (OPSYS_MFR == MD_SONY && OPSYS_VERS < MD_BSD_4_3) || \
    (OPSYS_MFR == MD_UCB && OPSYS_VERS < MD_BSD_4_2) || \
    (OPSYS_MFR == MD_APOLLO && OPSYS_VERS < MD_APOLLO_10_0_1) || \
    (OPSYS_SUBTYPE == MD_AIX && OPSYS_VERS <= MD_AIX_2_2) || \
    OPSYS_SUBTYPE == MD_DYNIX || \
    OPSYS_SUBTYPE == MD_HPUX || \
    OPSYS_SUBTYPE == MD_SYS5 /* { */
#include <sys/types.h>		/* I think can we do this multiple times here */
/*
 *    These are a bit problematic - they have tended to get
 *    defined more and more frequently as versions increase.
 *    There is probably some standards conformance at work here.
 */
typedef u_short uid_t;
typedef u_short gid_t;
#endif /* } */
#if (OPSYS_SUBTYPE == MD_AIX && OPSYS_VERS <= MD_AIX_2_2) /* { */
typedef int size_t;
#endif /* } */
#if OPSYS_MFR == MD_APOLLO && OPSYS_VERS < MD_APOLLO_10_0_1 /* { */
/*
 *    b/c apollo doesn't provide the required #include files!!!
 *
 *    Examples: <stdargs.h> (versus varargs.h)
 */
#undef __STDC__	/* on Apollo 10.1 you can't undef this */
#endif /* } */
#if OPSYS_MFR == MD_DEC && OPSYS_VERS == MD_ULTRIX32_4_0 /* { */
/*
 *    Only use ``OPSYS_VERS == MD_ULTRIX32_4_0'' instead of ``>=''
 *    b/c I hope that they'll fix this one by the next release.
 *    How can they break ANSI-compatibility this way?
 *    They should know better - erg!
 *
 *    Here we have to include all of /usr/include/math.h and prevent
 *    the REAL math.h from being included.  This is because math.h on
 *    ultrix 4.0 uses the cpp token ``CC$gfloat'' which is not accepted
 *    by gcc or other ansi-compatible sompilers.   Only DEC's compilers
 *    accept dollar-signs in the cpp tokens.
 */
#if defined(__GNUC__) /* { */
/*
 *    For now only do this as a workaround for gcc.  If there are other C
 *    compilers out there that have this problem on dec equipment then we'll
 *    have to handle that later.
 *
 *    Remember, we can't #include <math.h> so here it is literally.
 *    A single line has been fixed to prevent a syntax error under gcc.
 *    Portability is Hell.
 */
#ifndef _POLY9 /* { */
extern int errno, signgam;

extern double atof(), frexp(), ldexp(), modf(), strtod();
extern double j0(), j1(), jn(), y0(), y1(), yn();
extern double erf(), erfc();
extern double exp(), log(), log10(), pow(), sqrt();
extern double floor(), ceil(), fmod(), fabs();
extern double gamma();
extern double lgamma();
extern double hypot();
extern int matherr();
extern int isnan();
extern double sinh(), cosh(), tanh();
extern double sin(), cos(), tan(), asin(), acos(), atan(), atan2();
extern long atol();

/* some useful constants */
#define M_E	2.7182818284590452354
#define M_LOG2E	1.4426950408889634074
#define M_LOG10E	0.43429448190325182765
#define M_LN2	0.69314718055994530942
#define M_LN10	2.30258509299404568402
#define M_PI	3.14159265358979323846
#define M_PI_2	1.57079632679489661923
#define M_PI_4	0.78539816339744830962
#define M_1_PI	0.31830988618379067154
#define M_2_PI	0.63661977236758134308
#define M_2_SQRTPI	1.12837916709551257390
#define M_SQRT2	1.41421356237309504880
#define M_SQRT1_2	0.70710678118654752440
#ifdef	vax /* { */
#define MAXFLOAT	((float)1.701411733192644299e+38)
#if	defined(GFLOAT) /* { */
/* THIS IS WHAT WAS CHANGED HERE */
/* #if	defined(GFLOAT) || CC$gfloat */
#define HUGE	8.9884656743115790e+307 
#define HUGE_VAL	8.9884656743115790e+307 
#else /* } else { */
#define HUGE	1.701411834604692293e+38 
#define HUGE_VAL	1.701411834604692293e+38 
#endif /* } */ /* GFLOAT */
#endif /* } */
#ifdef	mips /* { */ /* This number generates +Infinity */
#define	HUGE	1.8e+308
#define	HUGE_VAL	1.8e+308
#define	MAXFLOAT	((float)3.40282346638528860e+38)
#endif /* } */

#define _ABS(x)	((x) < 0 ? -(x) : (x))
#define _REDUCE(TYPE, X, XN, C1, C2)	{ \
	double x1 = (double)(TYPE)X, x2 = X - x1; \
	X = x1 - (XN) * (C1); X += x2; X -= (XN) * (C2); }
#define _POLY1(x, c)	((c)[0] * (x) + (c)[1])
#define _POLY2(x, c)	(_POLY1((x), (c)) * (x) + (c)[2])
#define _POLY3(x, c)	(_POLY2((x), (c)) * (x) + (c)[3])
#define _POLY4(x, c)	(_POLY3((x), (c)) * (x) + (c)[4])
#define _POLY5(x, c)	(_POLY4((x), (c)) * (x) + (c)[5])
#define _POLY6(x, c)	(_POLY5((x), (c)) * (x) + (c)[6])
#define _POLY7(x, c)	(_POLY6((x), (c)) * (x) + (c)[7])
#define _POLY8(x, c)	(_POLY7((x), (c)) * (x) + (c)[8])
#define _POLY9(x, c)	(_POLY8((x), (c)) * (x) + (c)[9])

struct exception {
	int type;
	char *name;
	double arg1;
	double arg2;
	double retval;
};

#define DOMAIN		1
#define	SING		2
#define	OVERFLOW	3
#define	UNDERFLOW	4
#define	TLOSS		5
#define	PLOSS		6
#endif /* } */
#endif /* } */
#endif /* } */
#if ((OPSYS_MFR == MD_UCB || OPSYS_MFR == MD_SONY) && \
	    (OPSYS_VERS == MD_BSD_4_2 || OPSYS_VERS == MD_BSD_4_3)) || \
	(OPSYS_MFR == MD_IBM && OPSYS_SUBTYPE == MD_AICS) || \
	OPSYS_SUBTYPE == MD_DYNIX || \
	OPSYS_MFR == MD_APOLLO || \
	OPSYS_TYPE == MD_VMS /* { */
/* some useful constants they don't define */
#define M_E	2.7182818284590452354
#define M_LOG2E	1.4426950408889634074
#define M_LOG10E	0.43429448190325182765
#define M_LN2	0.69314718055994530942
#define M_LN10	2.30258509299404568402
#define M_PI	3.14159265358979323846
#define M_PI_2	1.57079632679489661923
#define M_PI_4	0.78539816339744830962
#define M_1_PI	0.31830988618379067154
#define M_2_PI	0.63661977236758134308
#define M_2_SQRTPI	1.12837916709551257390
#define M_SQRT2	1.41421356237309504880
#define M_SQRT1_2	0.70710678118654752440
#if CPU_TYPE == MD_3B /* { */
#define MAXFLOAT	((float)3.40282346638528860e+38)
#endif /* } */
#if CPU_TYPE == MD_PDP11 || CPU_TYPE == MD_VAX /* { */
#define MAXFLOAT	((float)1.701411733192644299e+38)
#endif /* } */
#ifndef HUGE /* { */ /* may already be defined from math.h */
#define HUGE	MAXFLOAT
#endif /* } */
#endif /* } */
#if OPSYS_MFR == MD_SUN && CPU_TYPE == MD_I386 /* { */
#ifdef COFF /* { */
#ifndef PORTAR /* { */
#define PORTAR
#endif /* } */
#endif /* } */
#endif /* } */

/*
 *    Because these are provided anywhere sumachine is used.
 */
typedef struct Sumachine Sumachine;
struct Sumachine {
    char *opsys_type;
    char *opsys_mfr;
    char *opsys_vers;
    char *cpu_type;
    char *cpu_model;
};

extern char *Revision();
extern void GetSumachine();

#endif /* } */ /* display_sh */
#endif /* } */ /* sumachine_h */
