#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "tr.h"
#include "utility.h"

/* bdf2oct.c
 * 3/16/91
 * Kristofer S. J. Pister
 */

#define TECHNOLOGY "scmos"
#define TEXTLAYERNAME "MET1"
#define BBLAYERNAME "PLACE"
#define atox(c) (('0' <= c) && (c <= '9') ? c-'0': 10 + c-'a')

char *getValueByName();
/*
 * ptext
 */
main(ac, av)
int ac;
char *av[];
{
    octObject output, tmpProp;
    char *inputName, *outputName;
    FILE *fpin;

    octBegin();
    errProgramName(av[0]);

    /* parse the arguments */
    if (ac == 3) {
	inputName = av[1];
	outputName = av[2];
    } else {
	fprintf(stderr,"usage: %s <bdf file> <cell>\n",av[0]);
	exit(1);
    }

    /* open the bdf file */
    fpin = fopen(inputName,"r");
    if (fpin == NULL) {
	fprintf(stderr,"%s: can't open bdf file %s\n",av[0],inputName);
	exit(1);
    }
    /* open the output facet with mode == "w" */
    ohUnpackDefaults(&output,"w",":physical:contents");
    ohUnpackFacetName(&output,outputName);
    OH_ASSERT( octOpenFacet(&output));

    /* set properties of output facet */
    ohCreatePropStr(&output,&tmpProp,"TECHNOLOGY",TECHNOLOGY);
    ohCreatePropStr(&output,&tmpProp,"VIEWTYPE", "PHYSICAL");
    ohCreatePropStr(&output,&tmpProp,"EDITSTYLE","PHYSICAL");

    /* create the new instance */
    convertBdf2Oct(fpin,&output);

    /* close everybody up */
    (void) octCloseFacet(&output);
    octEnd();
    exit(0);
}

#define MAXSTRLEN 1000

convertBdf2Oct(fp,out)
FILE *fp;
octObject *out;
{
    char line[MAXSTRLEN], name[MAXSTRLEN], *value, *symbol, *ptr;
    int w,h,left,bottom;
    int ury, llx, digit, bit;
    int items = 0;
    int xoffset=0, yoffset=0, maxRowHeight=0;
    tr_stack *xfStack;
    octTransform xform;
    octObject box, letterBag, formBag, textLayer, bbLayer, tmpProp;

    /* initialize transformation stack for tr */
    xfStack = tr_create();

    /* create the layers we will use */
    ohCreateLayer(out,&textLayer,TEXTLAYERNAME);
    ohCreateLayer(out,&bbLayer,BBLAYERNAME);

    /* get the font name */
    value = getValueByName(fp,"FONT");
    ohCreatePropStr(out,&tmpProp,"FONT",value);

    /* get the font size */
    value = getValueByName(fp,"SIZE");
    ohCreatePropStr(out,&tmpProp,"SIZE",value);

    /* get the font bounding box */
    value = getValueByName(fp,"FONTBOUNDINGBOX");
    ohCreatePropStr(out,&tmpProp,"FONTBOUNDINGBOX",value);

    /* get the font properties, if they exist */
    fgets(line,MAXSTRLEN,fp);
    if (!strncmp(line,"STARTPROPERTIES",strlen("STARTPROPERTIES"))) {
	while (fgets(line,MAXSTRLEN,fp) != NULL) {
	    /* check for the end of the properties list */
	    if (!strncmp(line,"ENDPROPERTIES",strlen("ENDPROPERTIES"))) break;
	}
    }
   
    /* get the begining of the character segments section */
    (void) getValueByName(fp,"CHARS");

    /* look for character descriptions */
    while ((symbol = getValueByName(fp,"STARTCHAR") ) != NULL) {

	/* map some names to characters (like "space" to " ") */
	lookupAndModify(symbol);
	
	/* see if we need a new row */
	if (items % 16 == 0) {
	    xoffset = 0;
	    yoffset += maxRowHeight * 1.5;
	    maxRowHeight = 0;
	}
	items++;

	/* create a bag for this character, and a form bag in it */
	ohCreateBag(out,&letterBag,symbol);
	ohCreateBag(&letterBag,&formBag,"form");

	/* set up the transformation to put this character in the facet */
	tr_identity(xfStack);
	tr_translate(xfStack, xoffset, yoffset,0);
	tr_get_oct_transform(xfStack,&xform);

	/* set encoding, swidth, dwidth properties of the character */
	value = getValueByName(fp,"ENCODING");
	ohCreatePropStr(&letterBag,&tmpProp,"ENCODING",value);
	value = getValueByName(fp,"SWIDTH");
	ohCreatePropStr(&letterBag,&tmpProp,"SWIDTH",value);
	value = getValueByName(fp,"DWIDTH");
	ohCreatePropStr(&letterBag,&tmpProp,"DWIDTH",value);

	/* get the bounding box info, transform and attach bounding box to the
	 * appropriate form bag and the appropriate layer bag.
	 *
	 * THIS MAY NOT BE THE RIGHT WAY TO DO THINGS FOR VARIABLE WIDTH FONTS
	 */
	value = getValueByName(fp,"BBX");
	ohCreatePropStr(&letterBag,&tmpProp,"BBX",value);
	sscanf(value,"%d %d %d %d",&w,&h,&left,&bottom);
	ohCreateBox(&letterBag,&box,left,bottom,left+w,bottom+h);
	octTransformGeo(&box,&xform);
	octModify(&box);
	octAttach(&bbLayer,&box);
	xoffset += w;
	if (h > maxRowHeight) maxRowHeight = h;

	/* get the ATTRIBUTES, if they exist */

	/* find the begining of the bitmap */
	while (strncmp(line,"BITMAP",strlen("BITMAP"))) fgets(line,MAXSTRLEN,fp);
	fgets(line,MAXSTRLEN,fp);

	/* start at the top of the character and work down */
	ury = bottom+h;
	while (strncmp(line,"ENDCHAR",strlen("ENDCHAR"))) {
	    /* remove the newline character, start at left of character */
	    line[strlen(line)-1] = '\0';
	    llx = left;

	    /* process each (hex) character in this line */
	    for(ptr=line ; *ptr ; ptr++) {
		digit = atox(*ptr);
		for (bit = 0; bit <4 ; bit++) {
		    if (digit & (1 << (3-bit))) {
			ohCreateBox(&textLayer,&box,llx+bit,ury-1,llx+bit+1,ury);
			octTransformGeo(&box,&xform);
			octModify(&box);
			octAttach(&formBag,&box);
		    }
		}

		/* move to the next set of four bits */
		llx += 4;
	    }

	    /* get the next line, update the height of the bits */
	    fgets(line,MAXSTRLEN,fp);
	    ury--;
	}
    }
}

/* this list should be much larger, and should probably be hashed.
 */
static struct pairs {
    char *name, *value;
} Pairs[] = {
   {"asterisk", "*"},{"at", "@"},{"backslash", "\\"},{"bar", "|"},
   {"braceleft", "{"},{"braceright", "}"},{"bracketright", "["},
   {"bracketleft", "]"},{"colon", ":"},{"comma", ","},{"dollar", "$"},
   {"equal", "="},{"exclam", "!"},{"hyphen", "-"},{"less", "<"},
   {"more", ">"},{"numbersign", "#"},{"parenleft", "("},{"parenright", ")"},
   {"percent", "%"},{"period", "."},{"plus", "+"},{"question", "?"},
   {"quotedbl", "\""},{"quoteleft", "`"},{"quoteright", "'"},
   {"semicolon", ";"},{"space", " "},{"underscore", "_"},{"zero", "0"},
   {"one", "1"},{"two", "2"},{"three", "3"},{"four", "4"},{"five", "5"},
   {"six", "6"},{"seven","7"},{"eight", "8"},{"nine", "9"},
{NULL, NULL}
};

lookupAndModify(s)
char *s;
{
    int i, val = 0;

    /* if it is a single character, take it as is */
    if (s[1] == '\0') return(0);

    /* if it is a number, and if its printable, and return the character */
    if (*s == 'C') {
	/* convert from octal */
	sscanf(s+1,"%o",&val);
	if (isprint(val)) {
	    sprintf(s,"%c",val);
	    return(1);
	}
    }

    /* check in the pairs list */
    for (i=0; Pairs[i].name != NULL; i++) {
	if (!strcmp(Pairs[i].name, s) ) strcpy(s,Pairs[i].value);
    }

    /* none of the above: leave it as is */
}

char *getValueByName(fp,token)
FILE *fp;
char *token;
{
    char line[MAXSTRLEN], *value;

    /* keep grabbing lines until one matches */
    while (fgets(line,MAXSTRLEN,fp) != NULL) {
	if ( !strncmp(line,"ENDFONT",strlen("ENDFONT")) ) {
	    return(NULL);
	}
	
	if ( !strncmp(line,token,strlen(token)) ) {
	    /* copy the rest of the string (minus the space) into value */
	    value = line + strlen(token);
	    while (isspace(*value)) value++;

	    /* strip off the newline */
	    value[strlen(value)-1] = '\0';

	    /* make some permanent storage, and return */
	    return(util_strsav(value));
	}
    }

    /* if we get here we didn't find it in the file */
    errRaise("getValueByName",1,"encountered EOF before finding token '%s' in input file\n",token);
}
