#include "boss_int.h"

extern "C"
{
#include "tap.h"
/* #include <sys/types.h> */

int tapGetStrangeConnector();
}

#define OCT_TO_LAMBDA 20
#define CONTACTS_DIR "special_contacts"

create_new_via( char* lay1,
	       char* lay2,
	       int wid,
	       int height)
{
	char command[1024];

	(void) sprintf(command, "mkdir %s; cd %s; prim %s-%s %d %d;",
		       CONTACTS_DIR, CONTACTS_DIR, lay1, lay2,
		       wid, height);

	if (system(command) != 0) {
		return 0;
	}
	return 1;
}

/*********************************************************************/

int checkViaExist(char *name)
{
  octObject facet;
  struct octFacetInfo info;
  
  facet.type = OCT_FACET;
  facet.contents.facet.cell = name;
  facet.contents.facet.view = "physical";
  facet.contents.facet.facet = "contents";
  facet.contents.facet.version = OCT_CURRENT_VERSION;
  facet.contents.facet.mode = "r";
  
  if (octGetFacetInfo(&facet, &info) < OCT_OK) {
    return 0;
  }
  return 1;
}

/****************************************************************/

void make_via_name(char *buffer,
	      char* lay1,
	      char* lay2,
	      int wid, 
	      int height)
{
  (void) sprintf(buffer,"%s-%s.%dx%d", lay1, lay2, wid, height);
}

/****************************************************************/

char *getOrCreateVia(char* tech_dir,
		     int h_wid, int v_wid,
		     octObject* lay1, octObject* lay2,
		     octTransformType *trans)
{
  char* name;
  char buf[1024];
  char via_name[1024];
  int tmp;
  BossLayer layer1();

  if(h_wid == 0)
    {
      h_wid = tapGetMinWidth(lay1) >? tapGetMinWidth(lay2);
    }
  if (v_wid == 0)
    {
      v_wid = tapGetMinWidth(lay1) >? tapGetMinWidth(lay2);
    }
  
  /* make sure it is at least of min size */
  if(h_wid < 2 * OCT_TO_LAMBDA)
    h_wid = 2 * OCT_TO_LAMBDA;
  if(v_wid < 2 * OCT_TO_LAMBDA)
    v_wid = 2 * OCT_TO_LAMBDA;
  
  /* convert to lambda and round up */
  
  h_wid = (h_wid + OCT_TO_LAMBDA - 1) / OCT_TO_LAMBDA;
  v_wid = (v_wid + OCT_TO_LAMBDA - 1) / OCT_TO_LAMBDA;
  
  /* change sizes so that the horizintal one is always the largest */
  
  if(h_wid >= v_wid) {
    (*trans) = OCT_NO_TRANSFORM;
  } else {
    tmp = v_wid;
    v_wid = h_wid;
    h_wid = tmp;
    (*trans) = OCT_ROT90;
  }
  
  make_via_name(via_name, 
		(strcmp(lay1->contents.layer.name, 
			lay2->contents.layer.name) <= 0) ? 
		lay1->contents.layer.name:lay2->contents.layer.name,
		(strcmp(lay1->contents.layer.name, 
			lay2->contents.layer.name) <= 0) ? 
		lay2->contents.layer.name:lay1->contents.layer.name,
		h_wid, v_wid);
  
  (void) sprintf(buf,"%s/contacts/%s", tech_dir, via_name);
  
  if(!checkViaExist(buf)) {
    (void) sprintf(buf,"%s/%s", CONTACTS_DIR, via_name);
    if (!checkViaExist(buf)) {
      if (!create_new_via((strcmp(lay1->contents.layer.name,
				  lay2->contents.layer.name) <= 0)? 
			  lay1->contents.layer.name:lay2->contents.layer.name,
			  (strcmp(lay1->contents.layer.name, 
				  lay2->contents.layer.name) <= 0)?
			lay2->contents.layer.name:lay1->contents.layer.name,
			  h_wid, v_wid)) {
	return 0;
      }
    }
  }
  name = bossStrsav(buf);
  return name;
}

int tapGetStrangeConnector(int nLayers, 
			   tapLayerListElement* layerList,
			   octObject* instance)
{
  int horiz = 0, vert = 0, i;
  enum octTransformType trans;
  octObject* lay1 = 0;
  octObject* lay2 = 0;
  
  for(i = 0; i < nLayers;i++) {
    if((lay1 == 0) || 
       (strcmp(lay1->contents.layer.name,
	       layerList[i].layer.contents.layer.name) == 0))
      lay1 = &(layerList[i].layer);
    else
      if((lay2 == 0) || 
	 (strcmp(lay2->contents.layer.name,
		 layerList[i].layer.contents.layer.name) == 0))
	lay2 = &(layerList[i].layer);
      else {
	return 0;
      }
    
    if (layerList[i].direction & TAP_HORIZONTAL) {
      vert = vert >? layerList[i].width;
    }
    if (layerList[i].direction & TAP_VERTICAL) {
      horiz = horiz >? layerList[i].width;
    }
    if(layerList[i].direction & TAP_WEIRD_DIR) {
      printf(" Can not have non-mahattan geometry\n");
      return 0;
    }
  }
  if(lay1 == 0) {
    return 0;
  }
  if(lay2 == 0)
    lay2 = lay1;
  
  instance->type = OCT_INSTANCE;
  instance->contents.instance.master = 
    getOrCreateVia(tapGetDirectory(&(layerList[0].layer)),horiz,vert,lay1,lay2,&trans);
  instance->contents.instance.view = "physical";
  instance->contents.instance.facet = "contents";
  instance->contents.instance.version = OCT_CURRENT_VERSION;
  instance->contents.instance.transform.transformType = trans;
  if (instance->contents.instance.master == 0) {
    return 0;
  }
  return 1;
}

/****************************************************************/



