#include "boss_int.h"

extern "C"
{
#include "tap.h"
typedef struct tapLayerListElement bossTapLayerListElement;
}


BossContact::BossContact(octObject* facet) : ()
{
(void) fill_in(facet, "MET1", "MET1", 80, 80);
create( *facet );
}

BossContact::BossContact(octObject& facet, char* layer1, char* layer2, int height, int width) : ()
{
(void) fill_in( &facet, layer1, layer2, height, width);
create( facet );
}


void BossContact::fill_in(octObject* facet, char* layer1, char* layer2, int height, int width)
{
    BossLayer lay1(facet, layer1);
    BossLayer lay2(facet, layer2);

    bossTapLayerListElement  layerList[2];

    layerList[0].layer = *(octObject*)&lay1;
    layerList[0].width = height;
    layerList[0].direction = TAP_LEFT;
    layerList[0].angle = 0;

    layerList[1].layer = *(octObject*)&lay2;
    layerList[1].width = width;
    layerList[1].direction = TAP_BOTTOM;
    layerList[1].angle = 0;

    if(tapGetConnector(2, layerList, this) == 0)
      errRaise( BOSS_PKG_NAME, 29, form("Problem in finding proper connector %s %s %dx%d",
					layer1, layer2, height/20, width/20));
    contents.instance.name = "CONTACT";
}

void BossContact::create( octObject& facet)
{
  if(contents.instance.transform.transformType != OCT_NO_TRANSFORM)
    {
      // gotta do this to get the BB correct

      octTransformType tr = contents.instance.transform.transformType;
      contents.instance.transform.transformType = OCT_NO_TRANSFORM;
      BossInstance::create(facet);
      transform(tr);
    }
  else
    BossInstance::create(facet);

  octObject con_bag, inst_bag;
  if(ohGetOrCreateBag(&facet, &con_bag, "CONNECTORS") < OCT_OK)
    errRaise( BOSS_PKG_NAME, 34, "Problem in getting the connector bag");
  if(octAttach(&con_bag, this) < OCT_OK)
    errRaise( BOSS_PKG_NAME, 35, "Problem in attaching to the connector bag");

  if(ohGetOrCreateBag(&facet, &inst_bag, "INSTANCES") < OCT_OK)
    errRaise( BOSS_PKG_NAME, 36, "Problem in getting the instance bag");
  if(octDetach(&inst_bag, this) < OCT_OK)
    errRaise( BOSS_PKG_NAME, 37, "Problem in detaching from the instance bag");
}

BossTerm& BossContact::term()
{
  return term("t");
}

void BossContact::get(octId objectId)
{
  if(ohGetById(this, objectId) < OCT_OK)
    errRaise( BOSS_PKG_NAME, 38, "Problem in restoring contact instance");

// set up BB

  octTransformType tr = contents.instance.transform.transformType;
  int x = contents.instance.transform.translation.x;
  int y = contents.instance.transform.translation.y;

  contents.instance.transform.transformType = OCT_NO_TRANSFORM;
  contents.instance.transform.translation.x = 0;
  contents.instance.transform.translation.y = 0;
  modify();

  if(octBB(this, &BB) < OCT_OK)
    errRaise( BOSS_PKG_NAME, 23, "Problems getting instance Bounding Box");
  contents.instance.transform.transformType = tr;
  contents.instance.transform.translation.x = x;
  contents.instance.transform.translation.y = y;
  modify();
}

// taken from symlib and modified 

static void myReplaceContact( BossContact& cont_old, BossContact& cont_new )
{
    octObject oterm, nterm;
    octGenerator tgen;
    octStatus status;

    /* fix terminal attachments */
    OH_ASSERT(octInitGenContents(&cont_new, OCT_TERM_MASK, &tgen));
	
    while ((status = octGenerate(&tgen, &nterm)) == OCT_OK) {
	/* XXX fix this (XID) */
	status = ohGetByTermName(&cont_old, &oterm, nterm.contents.term.name);
	switch (status) {
	    case OCT_OK:
	    break;

	    case OCT_NOT_FOUND:
	    errRaise(BOSS_PKG_NAME, 44, form("missing terminal correspondence (%s)",
		     nterm.contents.term.name));

	    default:
	    errRaise( BOSS_PKG_NAME, 39, "Problems replacing instance");
	}
	    
	ohAttachContainers(&oterm, &nterm, OCT_ALL_MASK&~OCT_INSTANCE_MASK);
	ohAttachContents(&oterm, &nterm, OCT_ALL_MASK);
    }

    if (status != OCT_GEN_DONE) {
      errRaise( BOSS_PKG_NAME, 40, "Problems replacing instance");
    }

    ohAttachContainers(&cont_old, &cont_new, OCT_ALL_MASK&~OCT_FACET_MASK);
    ohAttachContents(&cont_old, &cont_new, OCT_ALL_MASK&~OCT_TERM_MASK);

    OH_ASSERT(octDelete(&cont_old));

    octObject con_bag, facet;
    octGetFacet( &cont_new, &facet );
    if(ohGetOrCreateBag( &facet, &con_bag, "CONNECTORS") < OCT_OK)
      errRaise( BOSS_PKG_NAME, 34, "Problem in getting the connector bag");

    if(octDetach( &con_bag, &cont_new) != OCT_OK)
      errRaise( BOSS_PKG_NAME, 48, "Problem in removing connector bag attachment");
    if(octIsAttached( &con_bag, &cont_new) != OCT_OK)
      errRaise( BOSS_PKG_NAME, 49, "Problem contact is not in connector bag");
}

void BossContact::add_layer(char* name)
{
BossTerm& term = this->term();

if(term.num_layers() > 1)
  errRaise( BOSS_PKG_NAME, 41, form("Cannot add another layer to contact, %d layers already", term.num_layers()));
if(term.num_layers() == 0)
  errRaise( BOSS_PKG_NAME, 42, "Contact is not on any layer????");

octObject facet;
octGetFacet(this, &facet);
BossContact cont(facet, term.layer(0), name, term.extent_y(), term.extent_x());
cont.align_x(*this);
cont.align_y(*this);
myReplaceContact(*this, cont);
(*this) = cont;
}

