#include "String.h"
#include "boss_int.h"
#include "vov.h"

extern "C"
{
#include "symlib.h"
#include "vulcan.h"
}


extern int bossInitialized();

BossFacet::BossFacet(  char* cell,
		     char* view, 
		     char* technology,
		     char* facet, 
		     char* mode,
		     char* version)
{	
    if ( ! bossInitialized() ) {
	errProgramName( "boss" );
        errRaise( BOSS_PKG_NAME, 1, "\n\n\tBOSS POLICY ERROR!\n\tbossBegin(argc,argv); must be called before doing anything.\n\tPlease add bossBegin() and bossEnd() to your code.\n" );
    }

    type = OCT_FACET;
    contents.facet.cell = bossStrsav(cell);
    contents.facet.view = bossStrsav(view);
    contents.facet.facet = bossStrsav(facet);
    contents.facet.version = bossStrsav(version);
    symOpenFacet( this, bossStrsav( mode ), bossStrsav( technology ));

    VOVoutputFacet( this );
}

BossFacet::~BossFacet()
{
    OH_ASSERT(octCloseFacet(this));
    //  createInterface();
}

void BossFacet::print_name()
{  
    cout.form("facet %s:%s:%s\n", cell(), view(), facet());
}

void BossFacet::flush()
{  
    OH_ASSERT( octFlushFacet(this) );
}

void BossFacet::createInterface()
{
    bossMessage("Creating interface facet");
    
    flush();
    octObject outFacet;
    outFacet.type = OCT_FACET;
    outFacet.contents.facet.cell = bossStrsav( this->cell() );
    outFacet.contents.facet.view = bossStrsav( this->view() );
    outFacet.contents.facet.facet = bossStrsav( "interface" );
    outFacet.contents.facet.mode = bossStrsav( "w" );
    outFacet.contents.facet.version = bossStrsav( OCT_CURRENT_VERSION );
    if(octOpenFacet(&outFacet) < OCT_OK)
      errRaise( BOSS_PKG_NAME, 27, "problems in opening interface facet");
    
    vulcanMessageType vmt = vulcan(this, &outFacet);
    if((vmt == VULCAN_SEVERE) || (vmt == VULCAN_FATAL))
      errRaise( BOSS_PKG_NAME, 28, "Problems in creating the interface facet");
    else
      bossMessage("Done");
    VOVoutputFacet( this );
}

