#include "String.h"
#include "boss_int.h"
#include "vov.h"

char* bossStrsav( const char* s  )
{
    char* ns = new char[ strlen(s) + 1 ];
    return strcpy( ns , s );
}

BossInstance::BossInstance( octObject& facet, 
			   char* name,
			   char* master,
			   char* view, 
			   char* facet_name, 
			   char* version)
{
    contents.instance.name = bossStrsav( name );
    contents.instance.master = bossStrsav( master );
    contents.instance.view = bossStrsav( view );
    contents.instance.facet = bossStrsav( facet_name );
    contents.instance.transform.transformType = OCT_NO_TRANSFORM,
    
    contents.instance.version = version;
    create(facet);
}

void BossInstance::create(octObject& facet)
{
    this->type = OCT_INSTANCE;
    this->contents.instance.transform.translation.x = 0;
    this->contents.instance.transform.translation.y = 0;
    bossMessage(form("Creating %s ", format_name()));
    if(octCreate(&facet, this) != OCT_OK)
      errRaise( BOSS_PKG_NAME, 16, "Problems creating instance");
    
    octObject inst_bag;
    if(ohGetOrCreateBag(&facet, &inst_bag, "INSTANCES") < OCT_OK)
      errRaise( BOSS_PKG_NAME, 17, "Problem in getting the connector bag");
    if(octAttach(&inst_bag, this) < OCT_OK)
      errRaise( BOSS_PKG_NAME, 18, "Problem in attaching to the instance bag");
    
    if(octBB(this, &BB) < OCT_OK)
      errRaise( BOSS_PKG_NAME, 23, "Problems getting instance Bounding Box");
    move_to_origin();
    
    VOVinput( VOV_OCT_FACET, form( "%s:%s:interface", 
				  contents.instance.master,
				contents.instance.view ) );
}

char* BossInstance::format_name()
{
    return form("inst %s:%s:%s from %s",
		contents.instance.name, 
		contents.instance.view, 
		contents.instance.facet, 
		contents.instance.master);
}

void BossInstance::print_name()
{
    cout << format_name();
}

BossTerm& BossInstance::term(char *name)
{
    BossTerm* term = new BossTerm(*this, name);
    return *term;
}

void bossPromoteAll( BossInstance& inst, char* extension)
{
    octGenerator gen;
    octObject term;
    
    OH_ASSERT(octInitGenContents(&inst, OCT_TERM_MASK, &gen));
    while(octGenerate(&gen, &term) == OCT_OK)
      {
	  if(extension == 0)
	    bossPromote( term );
	  else
	    {
		BossFormalTerm t( BossTerm(term), form("%s.%s", term.contents.term.name, extension));
	}
    }
}

void BossInstance::offsetBB(int top, int bot, int left, int right)
{
    BB.upperRight.y += top;
    BB.upperRight.x += right;
    BB.lowerLeft.y -= bot;
    BB.lowerLeft.x -= left;
}



