#include "boss_int.h"


extern "C"
{
#include "symlib.h"
}

BossInstArray::BossInstArray(int rows, int cols) 
{
    first_col = 0;
    first_row = 0;
    final_row = rows - 1;
    final_col = cols - 1;
    num_rows = rows;
    num_cols = cols;
    array = new BossInstance*[num_rows*num_cols];
    for ( int i = 0; i < num_rows*num_cols; i++ ) array[i] = 0;// Make sure it's all zeroes.
}

BossInstance& BossInstArray::elmt(int row, int col)
{
  if(row < 0)
    errRaise( BOSS_PKG_NAME, 48, form("Array row index %d too small -- min %d",row, 0));
  if(col < 0)
    errRaise( BOSS_PKG_NAME, 48, form("Array column index %d too small -- min %d", col, 0));
  if(row > last_row())
    errRaise( BOSS_PKG_NAME, 48, form("Array row index %d too large -- max %d", row, last_row()));
  if(col > last_col())
    errRaise( BOSS_PKG_NAME, 48, form("Array column index %d too large -- max %d", col, last_col()));
BossInstance*& inst = (array[(row + first_row) * num_cols + (col + first_col)]);
if (inst == 0)
  errRaise( BOSS_PKG_NAME, 99, form("element (%d, %d) of array is empty -- cannot perform operations on it", row, col));
return *inst;
}

BossInstance*& BossInstArray::get_elmt(int row, int col)
{
  if(row < 0)
    errRaise( BOSS_PKG_NAME, 48, form("Array row index %d too small -- min %d",row, 0));
  if(col < 0)
    errRaise( BOSS_PKG_NAME, 48, form("Array column index %d too small -- min %d", col, 0));
  if(row > last_row())
    errRaise( BOSS_PKG_NAME, 48, form("Array row index %d too large -- max %d", row, last_row()));
  if(col > last_col())
    errRaise( BOSS_PKG_NAME, 48, form("Array column index %d too large -- max %d", col, last_col()));

  return array[(row + first_row) * num_cols + (col + first_col)];
}

void BossInstArray::populate_all( BossFacet& facet, char* master, char* view, char* name,char* facet_name )
{
  for(int i = 0;i <= last_row(); i++)
    for(int j = 0;j <= last_col(); j++) 
      {
	if(get_elmt(i,j) != 0)
	  {
	    bossWarning(form("Deleting old instance in array at loc %d %d", i,j));
	    OH_ASSERT(octDelete(get_elmt(i,j)));
	  }
	get_elmt(i, j) = new BossInstance(facet, form("%s%d.%d", name, i, j), master, view, facet_name);
      }
}

void BossInstArray::populate_row( BossFacet& facet, int row_num, char* master, char* view, char* name,char* facet_name )
{
  for(int j = 0;j <= last_col(); j++) 
    {
      if(get_elmt(row_num,j) != 0)
	{
	  bossWarning(form("Deleting old instance in array at loc %d %d", row_num,j));
	  OH_ASSERT(octDelete(get_elmt(row_num,j)));
	}
      get_elmt(row_num, j) = new BossInstance(facet, form("%s%d.%d", name, row_num, j), master, view, facet_name);
    }
}

void BossInstArray::populate_col( BossFacet& facet, int col_num, char* master, char* view, char* name,char* facet_name )
{
  for(int i = 0;i <= last_row(); i++)
    {
      if(get_elmt(i, col_num) != 0)
	{
	  bossWarning(form("Deleting old instance in array at loc %d %d", i,col_num));
	  OH_ASSERT(octDelete(get_elmt(i, col_num)));
	}
      get_elmt(i, col_num) = new BossInstance(facet, form("%s%d.%d", name, i, col_num), master, view, facet_name);
    }
}

void BossInstArray::populate_loc( BossFacet& facet, int row_num, int col_num, char* master, char* view, char* name,char* facet_name )
{
  if(get_elmt(row_num, col_num) != 0)
    {
      bossWarning(form("Deleting old instance in array at loc %d %d", row_num, col_num));
      OH_ASSERT(octDelete(get_elmt(row_num, col_num)));
    }
  get_elmt(row_num, col_num) = new BossInstance(facet, form("%s%d.%d", name, row_num, col_num), master, view, facet_name);
}

void BossInstArray::transform_all(octTransformType trans )
{
  for(int i = 0;i <= last_row(); i++)
    for(int j = 0;j <= last_col(); j++) 
      {
	if(get_elmt(i,j) != 0)
	  get_elmt(i,j)->transform(trans);
      }
}

void BossInstArray::transform_row(octTransformType trans, int row)
{
  for(int j = 0;j <= last_col(); j++) {
    if(get_elmt(row,j) != 0)
      get_elmt(row,j)->transform(trans);
  }  
}

void BossInstArray::transform_col(octTransformType trans, int col )
{
  for(int i = 0;i <= last_row(); i++) {
    if(get_elmt(i,col) != 0)
      get_elmt(i,col)->transform(trans);
  }
}

void BossInstArray::place_all()
{
  for(int i = 0;i <= last_row(); i++) {
    if(i != 0) {
      if((get_elmt(i,0) != 0) && (get_elmt(i-1,0) != 0))
	{
	  get_elmt(i, 0)->align_x(*get_elmt(i-1, 0), 0 );
	  get_elmt(i, 0)->above(*get_elmt(i-1, 0), 0 );
	}
    }
    for(int j = 0+1;j <= last_col(); j++) {
	if((get_elmt(i,j) != 0) && (get_elmt(i,j - 1) != 0))
	  {
	    get_elmt(i, j)->right_of(*get_elmt(i, j - 1), 0 );
	    get_elmt(i, j)->align_y(*get_elmt(i, j - 1), 0 );
	  }
    }      
  }
}

void BossInstArray::place_row( int row_num, int delta_x, int delta_y )
{
  for(int j = 1;j <= last_col(); j++) {
    if((get_elmt(row_num, j) != 0) && (get_elmt(row_num, j - 1) != 0))
      {
	get_elmt(row_num, j)->right_of(*get_elmt(row_num, j - 1), delta_x);
	get_elmt(row_num, j)->align_y(*get_elmt(row_num, 0), delta_y);
      }
  }      
}

void BossInstArray::place_col( int col_num, int delta_x, int delta_y )
{
  for(int i = 1;i <= last_row(); i++) {
    if((get_elmt(i, col_num) != 0) && (get_elmt(i-1, col_num) != 0))
      {
	get_elmt(i, col_num)->align_x(*get_elmt(0, col_num), delta_x);
	get_elmt(i, col_num)->above(*get_elmt(i-1, col_num), delta_y);
      }
    }      
}

void BossInstArray::promote_all(BossFacet& facet, char* term_name, char*f_term_name)
{
  int cnt = 0;
  for(int row = 0;row <= last_row(); row++) 
    for(int col = 0;col <= last_col(); col++) 
      {
	octObject term;
	if(get_elmt(row, col) != 0)
	  if(ohGetByTermName(get_elmt(row, col), &term, term_name) == OCT_OK)
	    {
	      octObject new_term;
	      symCreateFormalTerm(&facet, &new_term, form("%s<%d>", f_term_name, cnt++));
	      symImplementFormalTerm(&new_term, &term);
	    }
      }
}

void BossInstArray::promote_row(BossFacet& facet, int row, char* term_name, char*f_term_name)
{
  int cnt = 0;
  for(int col = 0;col <= last_col(); col++) {
    octObject term;
    if(get_elmt(row, col) != 0)
      if(ohGetByTermName(get_elmt(row, col), &term, term_name) == OCT_OK)
	{
	  octObject new_term;
	  symCreateFormalTerm(&facet, &new_term, form("%s<%d>", f_term_name, cnt++));
	  symImplementFormalTerm(&new_term, &term);
	}
  }
}

void BossInstArray::promote_col(BossFacet& facet, int col, char* term_name, char*f_term_name)
{
  int cnt = 0;
  for(int row = 0;row <= last_row(); row++) {
    octObject term;
    if(get_elmt(row, col) != 0)
      if(ohGetByTermName(get_elmt(row, col), &term, term_name) == OCT_OK)
	{
	  octObject new_term;
	  symCreateFormalTerm(&facet, &new_term, form("%s<%d>", f_term_name, cnt++));
	  symImplementFormalTerm(&new_term, &term);
	}
  }
}

void BossInstArray::connect_all( char* term_name, BossNet& net)
{
for(int row = 0;row <= last_row(); row++)
  for(int col = 0;col <= last_col(); col++)
    if(get_elmt(row, col) != 0)
      symAddTermToNet(&(get_elmt(row, col)->term( term_name)), &net);	
}

void BossInstArray::connect_row(int row, char* term_name, BossNet& net)
{
for(int i = 0;i <= last_col(); i++)
  if(get_elmt(row, i) != 0)
    symAddTermToNet(&(get_elmt(row, i)->term( term_name)), &net);	
}

void BossInstArray::connect_col(int col, char* term_name, BossNet& net)
{
for(int i = 0;i <= last_row(); i++)
  if(get_elmt( i, col) != 0)
    symAddTermToNet(&(get_elmt(i, col)->term( term_name)), &net);	
}


void BossInstArray::offsetBB_all( int top, int bot, int left, int right)
{
  for(int row = 0; row <= last_row();row++)
    for(int col = 0; col <= last_col();col++)
      if(get_elmt(row, col) != 0)
	get_elmt(row, col)->offsetBB( top, bot, left, right);
}

void BossInstArray::offsetBB_row(int row, int top, int bot, int left, int right)
{
  for(int col = 0; col <= last_col();col++)
    if(get_elmt(row, col) != 0)
      get_elmt(row, col)->offsetBB( top, bot, left, right);
}


void BossInstArray::offsetBB_col(int col, int top, int bot, int left, int right)
{
  for(int row = 0; row <= last_row();row++)
    if(get_elmt(row, col) != 0)
      get_elmt(row, col)->offsetBB( top, bot, left, right);
}

void BossInstArray::zipper_by_col( BossFacet& facet, int col1, char* term1_name, int col2, char* term2_name)
{
for(int row = 0;row <= last_row(); row++)
  if((get_elmt( row, col1) != 0) && (get_elmt( row, col2) != 0))
    {
      BossNet net( facet );
      net.add_term(  get_elmt( row, col1)->term(term1_name));
      net.add_term(  get_elmt( row, col2)->term(term2_name));
    }
}

void BossInstArray::zipper_by_row( BossFacet& facet, int row1, char* term1_name, int row2, char* term2_name)
{
for(int col = 0;col <= last_col(); col++)
  if((get_elmt( row1, col) != 0) && (get_elmt( row2, col) != 0))
    {
      BossNet net( facet );
      net.add_term(  get_elmt( row1, col)->term(term1_name));
      net.add_term(  get_elmt( row2, col)->term(term2_name));
    }
}

void BossInstArray::move_x ( int delta )
{
  for(int i = 0;i <= last_row(); i++)
    for(int j = 0;j <= last_col(); j++) 
      if(get_elmt(i, j) != 0)	
	{
	  get_elmt(i, j)->move_x(delta);
	}
}

void BossInstArray::move_y ( int delta )
{
  for(int i = 0;i <= last_row(); i++)
    for(int j = 0;j <= last_col(); j++) 
      if(get_elmt(i, j) != 0)	
	{
	  get_elmt(i, j)->move_y(delta);
	}
}
