#include "boss_int.h"

extern "C"
{
#include "region.h"
#include "tr.h"
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static void getInstBB( BossPhysicalObject* inst, octBox& bb )
{
octBox box = inst->BB;
tr_oct_transform(&(inst->get_transform()), &(box.lowerLeft.x), &(box.lowerLeft.y));
tr_oct_transform(&(inst->get_transform()), &(box.upperRight.x), &(box.upperRight.y));

// transforms will sometimes make the box upside down

bb.lowerLeft.x = box.lowerLeft.x <? box.upperRight.x;
bb.upperRight.x = box.lowerLeft.x >? box.upperRight.x;
bb.lowerLeft.y = box.lowerLeft.y <? box.upperRight.y;
bb.upperRight.y = box.lowerLeft.y >? box.upperRight.y;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

octBox& BossPhysicalObject::getBB()
{
octBox* box = new( octBox );

if(this->type == OCT_INSTANCE)
  getInstBB( this, *box );
else
  if(this->type == OCT_TERM)
    {
      if(regFindImpl(this, box) < OCT_OK)
	errRaise( BOSS_PKG_NAME, 3, "Problems getting instance Bounding Box");
    }
  else
    if(this->type == OCT_BOX)    
      return this->contents.box;
    else
      errRaise( BOSS_PKG_NAME, 22, "Unknown object type in getBB");
return *box;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::lowerLeft_x()
{
return (getBB()).lowerLeft.x;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::lowerLeft_y()
{
return (getBB()).lowerLeft.y;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::upperRight_x()
{
return (getBB()).upperRight.x;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::upperRight_y()
{
return (getBB()).upperRight.y;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::right_of( BossPhysicalObject& obj, int offset )
{
int delta = obj.upperRight_x() + offset - lowerLeft_x();
move_x(delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::left_of( BossPhysicalObject& obj, int offset )
{
int delta = obj.lowerLeft_x() - offset - upperRight_x();
move_x(delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::above( BossPhysicalObject& obj, int offset )
{
int delta = obj.upperRight_y() + offset - lowerLeft_y();
move_y(delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::below( BossPhysicalObject& obj, int offset )
{
int delta = obj.lowerLeft_y() - offset - upperRight_y();
move_y(delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::align_x( BossPhysicalObject& obj, int offset )
{
int delta = obj.center_x() + offset - center_x();
move_x(delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::align_y( BossPhysicalObject& obj, int offset )
{
int delta = obj.center_y() + offset - center_y();
move_y(delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::center_x()
{
return (upperRight_x() + lowerLeft_x()) / 2;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::center_y()
{
return (upperRight_y() + lowerLeft_y()) / 2;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::between_x ( BossPhysicalObject& obj, BossPhysicalObject& obj2 , int offset = 0 )
{
int delta = ((obj.lowerLeft_x() >? obj2.lowerLeft_x()) - 
  (obj.upperRight_x() <? obj2.upperRight_x())) / 2;

delta -= extent_x() / 2;

if(obj.upperRight_x() < obj2.upperRight_x())
  above(obj, delta);
else
  above(obj2, delta);
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::between_y ( BossPhysicalObject& obj, BossPhysicalObject& obj2 , int offset = 0 )
{
int delta = ((obj.lowerLeft_y() >? obj2.lowerLeft_y()) - 
  (obj.upperRight_y() <? obj2.upperRight_y())) / 2;

delta -= extent_y() / 2;

if(obj.upperRight_y() < obj2.upperRight_y())
  above(obj, delta);
else
  above(obj2, delta);

}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::extent_x()
{
  return upperRight_x() - lowerLeft_x();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

int BossPhysicalObject::extent_y()
{
  return upperRight_y() - lowerLeft_y();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::modify()
{
octStatus stat;

if((stat = octModify(this)) < OCT_OK)
  errRaise( BOSS_PKG_NAME, 22, form("Problems in modifying inst location -- status %d",stat));
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::move_x(int delta)
{
if(this->type == OCT_INSTANCE)
  transform_x() += delta;
else
  if(this->type == OCT_TERM)
    {
      octObject inst;
      OH_ASSERT(ohGetById( &inst, BossTerm(*this).instId()));
      inst.contents.instance.transform.translation.x += delta;

      octStatus stat;
      
      if((stat = octModify(&inst)) < OCT_OK)
	errRaise( BOSS_PKG_NAME, 22, form("Problems in modifying terminal location -- status %d",stat));
      return;
    }
  else
    if(this->type == OCT_BOX)  
      {
	(getBB()).lowerLeft.x += delta;
	(getBB()).upperRight.x += delta;
      }
    else
      errRaise( BOSS_PKG_NAME, 22, "Unknown object type in move_x");
modify();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::move_y(int delta)
{
if(this->type == OCT_INSTANCE)
  transform_y() += delta;
else
  if(this->type == OCT_TERM)
    {
      octObject inst;
      OH_ASSERT(ohGetById( &inst, BossTerm(*this).instId()));
      inst.contents.instance.transform.translation.y += delta;

      octStatus stat;
      
      if((stat = octModify(&inst)) < OCT_OK)
	errRaise( BOSS_PKG_NAME, 22, form("Problems in modifying terminal location -- status %d",stat));
      return;
    }
  else
    if(this->type == OCT_BOX)  
      {
	(getBB()).lowerLeft.y += delta;
	(getBB()).upperRight.y += delta;
      }
    else
      errRaise( BOSS_PKG_NAME, 22, "Unknown object type in move_y");
modify();
}
 
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::transform(octTransformType trans)
{
octTransform t;

if(this->type == OCT_TERM)
  errRaise( BOSS_PKG_NAME, 22, "Can not transform a terminal");  
t.transformType = trans;
octTransformAndModifyGeo( this, &t);
move_to_origin();
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void BossPhysicalObject::move_to_origin()
{
move_x(-lowerLeft_x());
move_y(-lowerLeft_y());
}
