#include "boss_int.h"

extern "C"
{
#include "symlib.h"
#include "region.h"
#include "tap.h"

/* next three lines so that st.h will compile */
#define ANSI_H     /* So ansi.h will not be included. */
#undef ARGS
#define ARGS(args)  ()
#include <st.h>

extern int tapGetStrangeConnector();
}

static void callSymCreateSegment( octObject& layer, 
				      octPoint& p1, octObject& term1, int type1,
				      octPoint& p2, octObject& term2, int type2,
				      int width, int dir)
{
  static int functionRegistered = 0;

  if(functionRegistered == 0)
    {
      tapSetUserConnectorGenerator(tapGetStrangeConnector);
      functionRegistered = 1;
    }
  symCreateSegment(&layer, &p1, &term1, type1, 
		   &p2, &term2, type2, 
		   width, dir);
}

static void _bossStraightRoute( BossTerm &term1, BossTerm &term2, int width, char *layer_name)
{
  octObject facet;
  octGetFacet( &term1, &facet);

  BossLayer layer( &facet, layer_name);
  int dir = SYM_HORIZONTAL;

  if((term1.upperRight_x() <= term2.lowerLeft_x()) || 
     (term2.upperRight_x() <= term1.lowerLeft_x()))
    {
      dir = SYM_HORIZONTAL;
      if((term1.upperRight_y() <= term2.lowerLeft_y()) || 
	 (term2.upperRight_y() <= term1.lowerLeft_y()))
	bossWarning( form("There is no way to connect terminals %s and %s \
              by a manhattan segment\n	CONNECTING ANYWAY!!!", term1.term_name(), term2.term_name));
    }
  else
    {
      if((term1.upperRight_y() <= term2.lowerLeft_y()) || 
	 (term2.upperRight_y() <= term1.lowerLeft_y()))
	dir = SYM_VERTICAL;
      else
	{
	  if(((term1.lowerLeft_x() >? term2.lowerLeft_x()) <= (term1.upperRight_x() <? term2.upperRight_x())) &&
	     ((term1.lowerLeft_y() >? term2.lowerLeft_y()) <= (term1.upperRight_y() <? term2.upperRight_y())))
	    bossWarning("in creating segment: TERMINALS OVERLAP -- CONNECTING ANYWAY");
	  else
	    bossWarning( form("There is no way to connect terminals %s and %s \
              by a manhattan segment\n	CONNECTING ANYWAY!!", term1.term_name(), term2.term_name));	
	}
    }
  bossMessage( form("Creating %s segment on layer %s",
		    (dir == SYM_HORIZONTAL)?"HORIZONTAL":"VERTICAL", layer_name));
  octPoint p1, p2;
  callSymCreateSegment(layer, 
		       p1, term1, SYM_USE_EXISTING, 
		       p2, term2, SYM_USE_EXISTING, 
		       width, dir);
}


overload bossStraightRoute;

void bossStraightRoute(  BossTerm &term1, BossTerm &term2, int width, char *layer_name)
{
  _bossStraightRoute( term1, term2, width, layer_name);
}

void bossStraightRoute(  BossTerm &term1, BossContact &cont2, int width, char *layer_name)
{
  BossTerm term2 = cont2.term();
  _bossStraightRoute( term1, term2, width, layer_name);
  cont2.get( term2.instId() );
}

void bossStraightRoute(  BossContact &cont1, BossTerm &term2, int width, char *layer_name)
{
  BossTerm term1 = cont1.term();
  _bossStraightRoute( term1, term2, width, layer_name);
  cont1.get(term1.instId());
}

void bossStraightRoute(  BossContact &cont1, BossContact &cont2, int width, char *layer_name)
{
  BossTerm term1 = cont1.term();
  BossTerm term2 = cont2.term();
  _bossStraightRoute( term1, term2, width, layer_name);
  cont1.get(term1.instId());
  cont2.get(term2.instId());
}

void _bossZRoute( BossTerm &term1, BossTerm &term2, int width,
	     enum BossRouteType center_seg_dir, char *layer_name1, char *layer_name2 = 0, 
	     char *layer_name3 = 0)
{
  octObject facet;
  octGetFacet( &term1, &facet);

  BossLayer layer1( &facet, layer_name1);
  BossLayer *layer2 = 0;
  BossLayer *layer3 = 0;

  if(layer_name2 == 0)
    layer2 = &layer1;
  else
    layer2 = new BossLayer( &facet, layer_name2);    

  if(layer_name3 == 0)
    layer3 = &layer1;
  else
    layer3 = new BossLayer( &facet, layer_name3);    
  
  if(center_seg_dir == BOSS_HORIZONTAL)
    {
      int y = (term1.center_y() +
	       term2.center_y()) / 2;
      
      octPoint point;
      octPoint pt1, pt2;  // dummy points
      point.y = y;
      point.x = term1.center_x();
      octObject term3;
      callSymCreateSegment(layer1, 
		       pt1, term1, SYM_USE_EXISTING, 
		       point, term3, SYM_CREATE, 
		       width, SYM_VERTICAL);
      octObject term4;
      point.x = term2.center_x();
      callSymCreateSegment(*layer3, 
		       pt1, term2, SYM_USE_EXISTING, 
		       point, term4, SYM_CREATE, 
		       width, SYM_VERTICAL);
      callSymCreateSegment(*layer2, 
		       pt1, term3, SYM_USE_EXISTING, 
		       pt2, term4, SYM_USE_EXISTING, 
		       width, SYM_HORIZONTAL);
    }
  else
    {
      int x = (term1.center_x() +
	       term2.center_x()) / 2;
      
      octPoint point;
      octPoint pt1, pt2;  // dummy points
      point.x = x;
      point.y = term1.center_y();
      octObject term3;
      callSymCreateSegment(layer1, 
		       pt1, term1, SYM_USE_EXISTING, 
		       point, term3, SYM_CREATE, 
		       width, SYM_HORIZONTAL);
      octObject term4;
      point.y = term2.center_y();
      callSymCreateSegment(*layer3, 
		       pt1, term2, SYM_USE_EXISTING, 
		       point, term4, SYM_CREATE, 
		       width, SYM_HORIZONTAL);
      callSymCreateSegment(*layer2, 
		       pt1, term3, SYM_USE_EXISTING, 
		       pt2, term4, SYM_USE_EXISTING, 
		       width, SYM_VERTICAL);
    }
}


overload bossZRoute;

void bossZRoute( BossTerm &term1, BossTerm &term2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0)
{
  _bossZRoute( term1, term2, width, center_seg_dir, 
	      layer_name1, 
	      layer_name2,
	      layer_name3);
}

void bossZRoute(  BossTerm &term1, BossContact &cont2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0)
{
  BossTerm term2 = cont2.term();
  _bossZRoute( term1, 
	      term2,
	      width, center_seg_dir, 
	      layer_name1, 
	      layer_name2,
	      layer_name3);
  cont2.get(term2.instId());
}

void bossZRoute(  BossContact &cont1, BossTerm &term2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0)
{
  BossTerm term1 = cont1.term();
  _bossZRoute( term1,
	      term2, 
	      width, center_seg_dir, 
	      layer_name1, 
	      layer_name2,
	      layer_name3);
  cont1.get(term1.instId());
}

void bossZRoute(  BossContact &cont1, BossContact &cont2, int width,
		 enum BossRouteType center_seg_dir, 
		 char *layer_name1, 
		 char *layer_name2 = 0, 
		 char *layer_name3 = 0)
{
  BossTerm term1 = cont1.term();
  BossTerm term2 = cont2.term();
  _bossZRoute( term1,
	      term2,
	      width, center_seg_dir, 
	      layer_name1, 
	      layer_name2,
	      layer_name3);
  cont1.get(term1.instId());
  cont2.get(term2.instId());
}

static BossContact& _bossStub( BossTerm& term, enum BossDirection dir, int offset = 0, char* layer = 0, int width = 0)
{
  if(layer == 0) {
    if(term.num_layers() != 1)
      bossWarning("Cannot determine which layer to implement bossStub\nTaking the first one");
    layer = term.layer(0);
  }
  octObject facet;
  octGetFacet( &term, &facet);
  BossContact& cont = *(new BossContact(&facet));

  int term_width = width;
  
  switch ( dir ) {
  case BOSS_UP:
    cont.above( term, offset );
    cont.align_x( term );
    if(term_width == 0)
      term_width = term.extent_x();
    break;
  case BOSS_DOWN:
    cont.below( term, offset );
    cont.align_x( term );
    if(term_width == 0)
      term_width = term.extent_x();
    break;
  case BOSS_LEFT:
    cont.left_of( term, offset );
    cont.align_y( term );
    if(term_width == 0)
      term_width = term.extent_y();
    break;
  case BOSS_RIGHT:
    cont.right_of( term, offset );
    cont.align_y( term );
    if(term_width == 0)
      term_width = term.extent_y();
    break;
  default: 
    errRaise( BOSS_PKG_NAME, 12, "WEIRD routing direction");
  }

  bossStraightRoute( cont, term, term_width, layer );
  return cont;
}

BossContact& bossStub( BossTerm& term, enum BossDirection dir, int offset = 0, char* layer = 0, int width = 0)
{
return _bossStub( term, dir, offset, layer, width );
}

BossContact& bossStub( BossContact& cont, enum BossDirection dir, int offset = 0, char* layer = 0, int width = 0)
{
return  _bossStub( cont.term(), dir, offset, layer, width );
}

static int bossAlreadyConnected(octObject* curr, BossTerm& target, int first )
{
  static st_table* table = 0;
  if(curr->objectId == target.objectId)
    return 1;

  if(first == 1)
    {
      if(table != 0)
	st_free_table(table);
      table = st_init_table( st_numcmp, st_numhash );
    }
  st_insert( table, (char*)curr->objectId, (char*)0 );

  octGenerator gen;
  if(curr->type == OCT_TERM)
    OH_ASSERT(octInitGenContainers( curr, OCT_PATH_MASK, &gen));
  else
    OH_ASSERT(octInitGenContents( curr, OCT_TERM_MASK, &gen));

  octObject obj;
  while(octGenerate(&gen, &obj) == OCT_OK)
    {
      if(st_lookup( table, (char*)obj.objectId, (char **)0 ) == 1)
	continue;
      if( bossAlreadyConnected( &obj, target, 0) == 1)
	return 1;
    }
  return 0;
}

int bossTryRoute( BossTerm &term1, BossTerm &term2 )
{
  octObject facet;
  octGetFacet( &term1, &facet);

  for(int i = term1.num_layers();i > 0;i--)
    {
      if( (strcmp(term1.layer(i - 1), "COND") != 0) &&
	 (strcmp(term1.layer(i - 1), "COPO") != 0) &&
	 (strcmp(term1.layer(i - 1), "COPD") != 0) &&
	 (strcmp(term1.layer(i - 1), "COPS") != 0) &&
	 (strcmp(term1.layer(i - 1), "CONS") != 0) &&
	 (strcmp(term1.layer(i - 1), "COM2") != 0))
	for(int j = term2.num_layers();j > 0;j--)
	  {
	    if((strcmp(term1.layer(i - 1), term2.layer(j - 1)) == 0) &&
	       (bossAlreadyConnected( &term1, term2, 1 ) == 0))
	      {
		BossLayer layer( &facet, term1.layer(i - 1) );
		
		int width = term1.extent_x() <? term2.extent_x();
		int height = term1.extent_y() <? term2.extent_y();
		int max_lower_x = term1.lowerLeft_x() >? term2.lowerLeft_x();
		int min_upper_x = term1.upperRight_x() <? term2.upperRight_x();
		int max_lower_y = term1.lowerLeft_y() >? term2.lowerLeft_y();
		int min_upper_y = term1.upperRight_y() <? term2.upperRight_y();
		if( (max_lower_x + width <= min_upper_x) && (width >= layer.min_width()))
		  {
		    octBox box;
		    box.lowerLeft.x = max_lower_x - layer.min_spacing();
		    box.lowerLeft.y = min_upper_y; 
		    box.upperRight.x = min_upper_x + layer.min_spacing();
		    box.upperRight.y = max_lower_y;
		    box.lowerLeft.x++;
		    box.lowerLeft.y++;
		    box.upperRight.x--;
		    box.upperRight.y--;
		    if(bossSearchRegion( facet, box, term1.layer(i - 1)) == 0)
		      {
			bossMessage(form("Making auto connection on layer %s",term1.layer(i - 1)));
			bossStraightRoute( term1, term2, width, term1.layer(i - 1));
			return 1;
		      }
		  }
		else
		  if((max_lower_y + height <= min_upper_y) && (height >= layer.min_width()))
		    {
		      octBox box;
		      box.lowerLeft.x = min_upper_x;
		      box.lowerLeft.y = max_lower_y - layer.min_spacing();
		      box.upperRight.x = max_lower_x;
		      box.upperRight.y = min_upper_y + layer.min_spacing();
		      box.lowerLeft.x++;
		      box.lowerLeft.y++;
		      box.upperRight.x--;
		      box.upperRight.y--;
		      if(bossSearchRegion( facet, box, term1.layer(i - 1)) == 0)
			{
			  bossMessage(form("Making auto connection on layer %s",term1.layer(i - 1)));
			  bossStraightRoute( term1, term2, height, term1.layer(i - 1));
			  return 1;
			}
		    }
	      }
	  }
    }
return 0;
}


