#include "copyright.h"
#include "jumbo.h"

/*
 * assignCells
 *
 */
double Temp;
double Cost;
double DelC;
double RangeLimitFactor = 1.0;
extern int NumCells;
extern s_cell **Cells;
extern int Verbose;

anneal()
{

    int numGen;		/* number of states generated at each temp 	*/
    int numAcc;		/* number of states accepted at each temp 	*/
    int numRej;		/* number of states rejected at each temp 	*/
    double initialTemp(),updateTemp();
    double costFn();
    double delCostFn();
    long random();
    s_cell *cell;

    if( NumCells == 0 )return;
    initialPlacement();
    Temp = initialTemp(); 
    Cost = costFn(0);
    if( Verbose ){
	fprintf(stdout,"Starting annealing - initial cost = %1.3g\n",Cost);
	fprintf(stdout,"                     initial temp = %1.2g\n",Temp);
    }
    while ( !stopCriterion()) {
	numGen = 0;
	numAcc = 0;
	numRej = 0;
	while ( !innerLoopCriterion( numGen, numAcc )){

	    /* generate a new move					*/
	    cell = Cells[ randomInt( 0, NumCells - 1 ) ];
	    genMove( cell, RangeLimitFactor );
	    numGen++ ;
	    /* accept or reject the move				*/
	    DelC = delCostFn( cell );
	    if (((DelC <= 0)||
		(( Temp > 0 )&&
		 ((double)random())/(double)2147483647 < exp( -DelC / Temp )))){
		accMove( cell );
		Cost += DelC;
		numAcc++;
		
#ifdef PARANOID
		if( Cost != costFn(0) ){
		    fprintf(stdout,"Caught an error, numGen = %d\n",numGen );
		    Cost = costFn(0);
		}
#endif
		
	    } else {
		numRej++;
	    }
	}
	if( Verbose ){
	    fprintf(stdout,
		    "At T = %1.2g accumulated cost = %1.3g real cost = %1.3g\n",
		    Temp, Cost, costFn(0));
	    fprintf( stdout,
		     "%d states accepted %d states rejected\n",
		     numAcc, numRej); 
	}
	/*
	if( numAcc / numGen >= .5 )Temp *= 0.8;
	else 			   Temp *= 0.9;
	*/
	Temp *= 0.9;
	RangeLimitFactor *= .95;
    }
    if( Verbose ){
	fprintf(stdout,"Finished annealing - final cost = %1.3g\n", Cost);
        printCosts();
	(void)costFn(1);
    }
    return;
}

/*
 * innerLoopCriterion()
 *
 * Returns true if the annealing should break out of the inner loop and
 * update the temperature, FALSE if the inner loop should continue
 */
innerLoopCriterion( numGen, numAcc )
int numGen; 		/* number of states which have been generated 	*/
{
    static int maxStates;	/* number of states at a given temp 	*/
    static int maxAcc;		/* number of accepted states at a term  */

    /* the temp has changed if numGen == 0 - reset MaxStates	*/
    if( numGen == 0 ){
	maxStates = numStates();
	maxAcc = (int)(maxStates * .6) ;
    }
    if ( numAcc >= maxAcc ) return( TRUE );
    if ( numGen < maxStates ) return(FALSE);
    return(TRUE);
}

/*
 * stoppingCriterion()
 *
 * Returns TRUE if the annealing is finished and should break out of the 
 * outer loop, FALSE if it should continue
 *
 * Presently returns TRUE if the cost hasn't changed for four temperatures.
 */
stopCriterion()
{
    static double cost1 = -1;
    static double cost2 = -2;
    static double cost3 = -3;
    if(( Cost == 0)||((Cost == cost1)&&(Cost == cost2)&&(Cost == cost3))){
	return(TRUE);
    } 
    cost3 = cost2;
    cost2 = cost1;
    cost1 = Cost;
    return(FALSE);
}

/*
 * randomInt
 *
 * return a pseudo random int between lower and upper
 */
int
randomInt( lower, upper )
int lower;
int upper;
{
    if( lower == upper )return( lower );
    return(( random()%( upper + 1 - lower )) + lower );
}
/*
 * initialTemp()
 *
 * calculates the initial temperature
 */
double
initialTemp()
{

    return( 10 * ( double )NumCells );
}
int 
numStates()
{
    return(  10 * NumCells ); 
}
