#include "copyright.h"
#include "jumbo.h"

extern int NumRows;
extern int NumCols;
extern int NumCells;
extern int TargetNumBins;
extern int TargetBinSize;
extern int Verbose;
extern double TargetAspectRatio;
extern s_bin **Bin;
extern s_cell **Cells;
extern double AverageBinCount;
extern s_net *Nets;
/*
 * createBins
 *
 * allocates the grid of bins
 *
 * If the number of rows and columns are explicitly given, then the
 * grid size is set to that.  If one dimension is unknown, it is 
 * determined by the numberOfCells/targetBinSize or the targetAspectRatio.  
 * If both dimensions are unknown, the number of bins is determined
 * by the numberOfCells/targetBinSize and they are divided into rows and
 * columns according to the targetAspectRatio
 */
createBins()
{
    int row, col, index;
    double numBins;
    double tmp;

    AverageBinCount = 0;
    for( index = 0; index < NumCells; index++ ){
	AverageBinCount += Cells[ index ]->weight;
    }
    if(( NumRows < 1 )&&( NumCols < 1 )){ 
	if( TargetNumBins > 0 ){
	    numBins = TargetNumBins;
	} else if( TargetBinSize > 0 ){
	    numBins = AverageBinCount / TargetBinSize;
	} else {
	    fprintf( stderr,
	             "The target number of bins or bin size must be given\n");
	    exit( -1 );
	}
	tmp = sqrt( numBins * TargetAspectRatio );
	if( tmp - floor( tmp ) > 0.5 ){
	    NumRows = ceil( tmp );
	} else {
	    NumRows = floor( tmp );
	}
	tmp = numBins / NumRows;
	if( tmp - floor( tmp ) > 0.5 ){
	    NumCols = ceil( tmp );
	} else {
	    NumCols = floor( tmp );
	}
    } else if ( NumRows < 1 ){
	if( TargetNumBins > 0 ){
	    tmp = TargetNumBins / NumCols;
	} else {
	    if( TargetBinSize > 0 ){
		tmp = ( AverageBinCount / TargetBinSize ) / NumCols;
	    } else {
		tmp = NumCols * TargetAspectRatio;
	    }
	} 
	if( tmp - floor( tmp ) > 0.5 ){
	    NumRows = ceil( tmp );
	} else {
	    NumRows = floor( tmp );
	}
    } else if ( NumCols < 1 ){
	if( TargetNumBins > 0 ){
	    tmp = TargetNumBins / NumRows;
	} else {
	    if( TargetBinSize > 0 ){
		tmp = ( AverageBinCount / TargetBinSize ) / NumRows;
	    } else {
		tmp = NumRows / TargetAspectRatio;
	    }
	}
	if( tmp - floor( tmp ) > 0.5 ){
	    NumCols = ceil( tmp );
	} else {
	    NumCols = floor( tmp );
	}
    }
    if( NumRows < 1 ) NumRows = 1;
    if( NumCols < 1 ) NumCols = 1;
    if(( NumRows == 1 )&&( NumCols == 1 )){
	writeNoPartitionProp();
	if( Verbose )
	   fprintf( stdout,"No partition was done, only one bin\n" );
	exit( 0 );
    }
    Bin = ALLOC( s_bin *, NumRows );
    for( row = 0; row < NumRows; row++ ){
	Bin[ row ] = ALLOC( s_bin, NumCols );
	for( col = 0; col < NumCols; col++ ){
	    Bin[ row ][ col ].count = 0;
	    Bin[ row ][ col ].marked = FALSE;
	}
    }
    AverageBinCount /= ( NumRows * NumCols );
    if( Verbose )
	fprintf( stdout,"Partitioning into %d rows x %d cols\n",
		 NumRows, NumCols );
}
