#include "copyright.h"
#include "jumbo.h"

double nlCost = 0.0;
double ntCost = 0.0;
double bCost = 0.0;
extern double NlFactor;
extern double NtFactor;
extern double BFactor;
extern int NumRows;
extern int NumCols;
extern int NumCells;
extern int NumNets;
extern int NewRow;
extern int NewCol;
extern double AverageBinCount;
extern s_net *Nets;
extern s_bin **Bin;
/*
 * costFn
 *
 * returns the total cost of the configuration
 */
double
costFn(verbose)
{
    double cost;
    int netlength;
    int netTerms;
    int row, col;
    s_net *net;
    s_cell *cell;
    int lowerRow, lowerCol, upperRow, upperCol;
    int index;
    int maxNetLength = -1;
    int maxNetTerms = -1;
    int *netVector;
    int *termVector;

    /* the bin irregularity cost					*/

    if( verbose ){
	fprintf(stdout,"\nThere are %d cells and %d nets\n", NumCells, NumNets);
	fprintf(stdout,"The average bin count is %1.3g\n", AverageBinCount );
	fprintf(stdout,"The bin counts:\n");
    }
    bCost = 0.0;
    for( row = 0; row < NumRows; row++ ){
	for( col = 0; col < NumCols; col++ ){
	    Bin[ row ][ col ].marked = FALSE;
	    cost = Bin[ row ][ col ].count -  AverageBinCount;
	    bCost += cost * cost;
	    if( verbose )fprintf(stdout,"\t%d", Bin[ row ][ col ].count );
	}
	if( verbose )fprintf(stdout,"\n");
    }

    /* the net length and netTerms costs				*/

    nlCost = ntCost = 0;
    for( net = Nets; net != NIL( s_net ); net = net->next ){
	lowerRow = lowerCol = INFINITY;
	upperRow = upperCol =  -INFINITY;
	for( index = 0; index < net->numCells; index++ ){
	    cell = net->cell[ index ];
	    Bin[ cell->row ][ cell->col ].marked = TRUE;
	    if( cell->row > upperRow )upperRow = cell->row;
	    if( cell->col > upperCol )upperCol = cell->col;
	    if( cell->row < lowerRow )lowerRow = cell->row;
	    if( cell->col < lowerCol )lowerCol = cell->col;
	}
	netlength = upperRow - lowerRow + upperCol - lowerCol;
	nlCost += ( netlength * netlength );
	netTerms = 0;
	for( row = 0; row < NumRows; row++ ){
	    for( col = 0; col < NumCols; col++ ){
		if( Bin[ row ][ col ].marked ){
		    netTerms++;
		    Bin[ row ][ col ].marked = FALSE;
		}
	    }
	}
	/* a net either has 0 or more than 2 terms			*/
	if( netTerms == 1 )netTerms = 0;
	ntCost += netTerms;
	if( verbose ){
	    if( maxNetLength == -1 ){
		netVector = ALLOC( int, netlength + 1 );
		while( maxNetLength < netlength )
		    netVector[ ++maxNetLength ] = 0;
	    } else if( netlength > maxNetLength ){
		netVector = REALLOC( int, netVector, netlength + 1 );
		while( maxNetLength < netlength )
		    netVector[ ++maxNetLength ] = 0;
	    }
	    netVector[ netlength ]++;
	    if( maxNetTerms == -1 ){
		termVector = ALLOC( int, netTerms + 1 );
		while( maxNetTerms < netTerms )
		    termVector[ ++maxNetTerms ] = 0;
	    } else if( netTerms > maxNetTerms ){
		termVector = REALLOC( int, termVector, netTerms + 1 );
		while( maxNetTerms < netTerms )
		    termVector[ ++maxNetTerms ] = 0;
	    }
	    termVector[ netTerms ]++;
	}
    }
    if( verbose ){
	fprintf( stdout,"The nets by net length:\n\t");
	for( netlength = 0; netlength <= maxNetLength; netlength++ ){
	    fprintf( stdout,"%-5d", netlength );
	}
	fprintf( stdout,"\n\t");
	for( netlength = 0; netlength <= maxNetLength; netlength++ ){
	    fprintf( stdout,"%-5d", netVector[ netlength ] );
	}
	fprintf( stdout,"\nThe nets by number of terminals:\n\t");
	for( netTerms = 0; netTerms <= maxNetTerms; netTerms++ ){
	    fprintf( stdout,"%-5d", netTerms );
	}
	fprintf( stdout,"\n\t");
	for( netTerms = 0; netTerms <= maxNetTerms; netTerms++ ){
	    fprintf( stdout,"%-5d", termVector[ netTerms ] );
	}
	fprintf( stdout,"\n");
	FREE( netVector );
	FREE( termVector );
    }
    cost = nlCost * NlFactor + ntCost * NtFactor + bCost * BFactor;
    return( cost );
}
/* 
 * delCostFn()
 *
 * calculates the change in cost due to a move
 */
double
delCostFn( cell )
s_cell *cell;
{
    double cost;
    int lowerRow, upperRow, lowerCol, upperCol;
    s_cell *cellP;
    s_net *net;
    int cellInBin, newCellInBin;
    int nIndex, cIndex;
    int netlength;

    /* the bin irregularity cost					*/

    bCost = 2 * cell->weight * ( Bin[ NewRow ][ NewCol ].count - 
	                         Bin[ cell->row ][cell->col ].count + 
				 cell->weight );

    /* the net length and netTerms costs				*/

    nlCost = ntCost = 0.0;
    for( nIndex = 0; nIndex < cell->numNets; nIndex++ ){
	net = cell->net[ nIndex ];
	if( net->numCells == 1 )continue;
	cellInBin = newCellInBin = 0;
	lowerRow = lowerCol = INFINITY;
	upperRow = upperCol = -INFINITY;

	for( cIndex = 0; cIndex < net->numCells; cIndex++ ){
	    if(( cellP = net->cell[ cIndex ] ) == cell )continue;
	    if(( !cellInBin )&&
	       ( cellP->row == cell->row )&&( cellP->col == cell->col )){
		cellInBin = 1;
		ntCost = ntCost + 1.0;
	    }
	    if(( !newCellInBin )&&
	       ( cellP->row == NewRow )&&( cellP->col == NewCol )){
		newCellInBin = 1;
		ntCost = ntCost - 1.0;
	    }
	    /*
	     * lowerRow, upperRow, lowerCol and upperCol define the core area 
	     * of the net without the cell
	     */
	    if( cellP->row > upperRow )upperRow = cellP->row;
	    if( cellP->col > upperCol )upperCol = cellP->col;
	    if( cellP->row < lowerRow )lowerRow = cellP->row;
	    if( cellP->col < lowerCol )lowerCol = cellP->col;
	}
	/* this is due to the fact that a net cannot have 1 term	*/
	if( upperRow - lowerRow + upperCol - lowerCol == 0 ){
	    /* these conditions are exclusive				*/
	    if( newCellInBin )ntCost = ntCost - 1.0;
	    if( cellInBin )ntCost = ntCost + 1.0;
	}
	/* 
	 * nlcost is the incremental netlength of adding the cell to the 
	 * new position - the incremental netlength of adding the cell to
	 * the old position
	 */
	if( cell->row > upperRow )netlength = cell->row - lowerRow;
	else if( cell->row < lowerRow )netlength = upperRow - cell->row;
	else netlength = upperRow - lowerRow;

	if( cell->col > upperCol )netlength += cell->col - lowerCol;
	else if( cell->col < lowerCol )netlength += upperCol - cell->col;
	else netlength += upperCol - lowerCol;

	nlCost -= (netlength * netlength);

	if( NewRow > upperRow )netlength = NewRow - lowerRow;
	else if( NewRow < lowerRow )netlength = upperRow - NewRow;
	else netlength = upperRow - lowerRow;

	if( NewCol > upperCol )netlength += NewCol - lowerCol;
	else if( NewCol < lowerCol )netlength += upperCol - NewCol;
	else netlength += upperCol - lowerCol;

	nlCost += (netlength * netlength);

    }

    cost = nlCost * NlFactor + ntCost * NtFactor + bCost * BFactor;
    return( cost );
}
/* 
 * printCosts()
 *
 * print the cost factors to stdout
 */
printCosts()
{
    fprintf( stdout,"The cost factors are:\n");
    fprintf( stdout,"\tnet length cost:        %4f\n", nlCost * NlFactor );
    fprintf( stdout,"\tnet terminal cost:      %4f\n", ntCost * NtFactor );
    fprintf( stdout,"\tbin irregularity cost:  %4f\n", bCost * BFactor );
}
