#include "copyright.h"
#include "jumbo.h"
#include "st.h"

extern octObject oFacet;
extern double TargetAspectRatio;
extern st_table *NetTbl;
extern int NumRows, NumCols;
extern octObject oInterface;
int first = 1;
/*
 * readAspectRatio
 * 
 * finds a box on the interface facet in the FLOORPLAN bag which
 * specifies the desired aspect ratio of the cell.  If the 
 * targetAspectRatio has been given ( its not -2 ), it is multipled
 * by the aspect ratio io the box, otherwise it is set to the 
 * aspect ratio of the box,
 */
readAspectRatio()
{
    octObject oBag, oBox;
    if( first ) openInterface();
    oBag.type = OCT_BAG;
    oBag.contents.bag.name = "FLOORPLAN";
    if( octGetByName( &oInterface, &oBag ) < OCT_OK ){
	fprintf( stderr,
		 "Warning: Could not open get %s bag from %s:%s:%s: %s: %s\n",
		 oBag.contents.bag.name,
		 oInterface.contents.facet.cell,
		 oInterface.contents.facet.view,
		 oInterface.contents.facet.facet,
		 "the floorplanner aspect ratio will not be read",
		 octErrorString());
    }
    if( octGenFirstContent( &oBag, OCT_BOX_MASK, &oBox ) < OCT_OK ){
	fprintf( stderr,
		 "Warning: Could not get a box from the %s bag: %s: %s\n",
		 oBag.contents.bag.name,
		 "the floorplanner aspect ratio will not be read",
		 octErrorString());
    }
    if( TargetAspectRatio == -1 )TargetAspectRatio = 1;
    TargetAspectRatio *= ( oBox.contents.box.upperRight.y -
                           oBox.contents.box.lowerLeft.y ) /
                         ( oBox.contents.box.upperRight.x -
                           oBox.contents.box.lowerLeft.x );
    if( TargetAspectRatio < 0 ) TargetAspectRatio = -TargetAspectRatio;
    if( TargetAspectRatio < 0.05  ){
	fprintf( stderr,
		 "Warning: The target aspect of %1.3g ratio is too small, %s\n",
		 "setting it to 1.0");
    }
}
/*
 * readConstrainedTerms
 *
 * reads terminal constraints and creates dummy cells to account for them
 */
readConstrainedTerms()
{
    octObject oTerm, oNet;
    octObject oTerm_Edge, oTerm_RelPos;
    octGenerator gTerm;
    octStatus sTerm;
    s_cell *cell;
    s_net *net;
    char *ptr;

    if( first ) openInterface();
    oTerm_Edge.type = OCT_PROP;
    oTerm_Edge.contents.prop.name = "TERM_EDGE";
    oTerm_RelPos.type = OCT_PROP;
    oTerm_Edge.contents.prop.name = "TERM_RELATIVE_POSITION";
    OH_ASSERT( octInitGenContents( &oFacet, OCT_TERM_MASK, &gTerm ));
    while(( sTerm = octGenerate( &gTerm, &oTerm )) == OCT_OK ){
	if( octGenFirstContainer( &oTerm, OCT_TERM_MASK, &oNet ) < OCT_OK ){
	    fprintf( stderr,
		     "Warning: Formal term %s is not attached to a net: %s\n",
		     oTerm.contents.term.name,
		     octErrorString());
	    continue;
	}

	/* probably due to a net not connected to any instances		*/
	if (!st_lookup(NetTbl,(char *) &oNet.objectId, &ptr)) {
	    continue;
	}
    
        net = (s_net *) ptr;
	if( octGetByName( &oInterface, &oTerm ) < OCT_OK ){
	    fprintf( stderr,
		     "Warning: Could not match term %s %s: %s\n",
		     oTerm.contents.term.name,
		     "to a term on the interface facet",
		     octErrorString());
	    continue;
	}
	
	/* the TERM_EDGE and TERM_RELATIVE_POSITION must be there	*/
	if(( octGetByName( &oTerm, &oTerm_Edge ) < OCT_OK )||
	   ( oTerm_Edge.contents.prop.type != OCT_STRING )||
	   ( octGetByName( &oTerm, &oTerm_RelPos ) < OCT_OK )||
	   ( oTerm_RelPos.contents.prop.type != OCT_REAL ))continue;

	cell = ALLOC( s_cell, 1 );
	cell->objectId = oct_null_id;
	cell->weight = 0;
	cell->net = NIL( s_net * );
	if( strcmp( oTerm_Edge.contents.prop.value.string, "LEFT" )){
	    cell->row = ( NumRows * oTerm_RelPos.contents.prop.value.real ) - 1;
	    cell->col = 0;
	} else if( strcmp( oTerm_Edge.contents.prop.value.string, "RIGHT" )){
	    cell->row = ( NumRows * oTerm_RelPos.contents.prop.value.real ) - 1;
	    cell->col = NumCols - 1;
	} else if( strcmp( oTerm_Edge.contents.prop.value.string, "TOP" )){
	    cell->row = NumRows - 1;
	    cell->col = ( NumCols * oTerm_RelPos.contents.prop.value.real ) - 1;
	} else if( strcmp( oTerm_Edge.contents.prop.value.string, "BOTTOM" )){
	    cell->row = 0;
	    cell->col = ( NumCols * oTerm_RelPos.contents.prop.value.real ) - 1;
	} else {
	    fprintf( stderr, "Warning: %s %s value of %s property on term %s\n",
			     "Could not understand",
			     oTerm_Edge.contents.prop.value.string,
			     oTerm_Edge.contents.prop.name,
			     oTerm.contents.term.name );
	    FREE( cell );
	    continue;
	}
	net->cell = REALLOC( s_cell *, net->cell, ++net->numCells);
	net->cell[ net->numCells - 1 ] = cell;
    } 
    OH_ASSERT( sTerm );
    st_free_table( NetTbl );
}
/*
 * openInterface
 *
 * opens the interface facet
 */
openInterface()
{
    first = 0;
    oInterface.contents.facet.cell = oFacet.contents.facet.cell;
    oInterface.contents.facet.view = oFacet.contents.facet.view;
    oInterface.contents.facet.facet = "interface";
    oInterface.contents.facet.mode = "r";
    oInterface.contents.facet.version = OCT_CURRENT_VERSION;
    if( octOpenFacet( &oFacet ) != OCT_OLD_FACET ){
	fprintf( stderr,"Fatal Error: Could not open facet %s:%s:%s: %s\n",
		         oInterface.contents.facet.cell,
		         oInterface.contents.facet.view,
		         oInterface.contents.facet.facet,
			 octErrorString());
	exit(-1);
    }
}
