#include "copyright.h"
#include "jumbo.h"

/*
 * jumbo - a simulated annealing based partitioner
 *
 * Instances in a cell are assigned to bins in a two dimensional grid so
 * to minimize the netlength ( manhattan "size" of net bb in grid coords ),
 * the number of bins each net touches, and the difference in size between
 * the bins.
 *
 * Constrained formal terminals are given fixed cells which are assigned to
 * the appropriate bin, so the position of constrained terminals is taken 
 * into account in the cost function.
 */
/* command line options set in parseComLine				*/
int Verbose = FALSE;	
int ReadFloorplanning = FALSE;	
int NumRows = -1;		/* if unset, set in createBin		*/
int NumCols = -1;		/* if unset, set in createBin		*/
double TargetAspectRatio = 1.0;		
double NlFactor = 1.0;
double NtFactor = 1.0;
double BFactor = 1.0;
int HideSubcells = 0;
int Undo = 0;
int TargetNumBins = -1;
int TargetBinSize = -1;
int NoImplement = 0;
int CellWeightType = (unsigned)0;
char *InName = "";		/* name of input facet; 		*/
char *OutName = "";		/* if not "", name of output facet      */
char *PartName = "";		/* name of the new cells		*/

/* globals								*/
struct net *Nets;		/* set in readOct			*/
struct cell **Cells;		/* set in readOct			*/
s_bin **Bin;			/* created in createBin, modified in accMove */
int NumCells = 0;		/* set in readOct			*/
int NumNets = 0;		/* set in readOct			*/
double AverageBinCount;		/* set in createBin			*/
int NewRow;			/* repeatedly set in genMove		*/
int NewCol;			/* repeatedly set in genMove		*/

main( argc, argv )
int argc;
char **argv;
{
    parseComLine( argc, argv );

    openFacet();
    if( Undo ){
	undoPrevRun();
	exit( 0 );
    }
    readOct();
    if( ReadFloorplanning )readAspectRatio();
    createBins();
    if( ReadFloorplanning )readConstrainedTerms();
    anneal();
    if( NoImplement != 1 )writeOct();
    exit( 0 );
}

halt()
{
    int stop = 1;
    if( stop ) abort();
}
