#include "copyright.h"
#include "jumbo.h"
#include "st.h"
#include "nle.h"

extern s_cell **Cells;
extern int NumCells;
extern int NumNets;
extern s_net *Nets;
extern char *InName;
extern char *OutName;
extern char *PartName;
extern int NumRows, NumCols;
extern int CellWeightType;
extern int HideSubcells;
char *cellNames;
octObject oFacet;
octObject oInstances;
octObject oInterface;
char *Path;
extern s_bin **Bin;
st_table *NetTbl;
extern int ReadFloorplanning;
/*
 * readOct
 *
 */
openFacet()
{
    octObject oIn;
    char *c;

    oIn.type = OCT_FACET;
    oIn.contents.facet.cell = "";
    oIn.contents.facet.view = "";
    oIn.contents.facet.facet = "contents";
    oIn.contents.facet.mode = "a";
    oIn.contents.facet.version = OCT_CURRENT_VERSION;
    (void) ohUnpackFacetName(&oIn, InName);

    oFacet = oIn;
    (void) ohUnpackFacetName(&oFacet, OutName);
    if(( strcmp( oIn.contents.facet.cell, oFacet.contents.facet.cell ) != 0 )||
       ( strcmp( oIn.contents.facet.view, oFacet.contents.facet.view ) != 0 )){
	if( octCopyFacet( &oFacet, &oIn ) < OCT_OK ){
	    fprintf( stderr, "Could not copy facet %s:%s:%s to %s:%s:%s: %s\n",
			     oIn.contents.facet.cell, 
			     oIn.contents.facet.view,
			     oIn.contents.facet.facet,
			     oFacet.contents.facet.cell,
			     oFacet.contents.facet.view,
			     oFacet.contents.facet.facet,
			     octErrorString());
	    halt();
	}
    }
    oInterface.contents.facet = oIn.contents.facet;
    oInterface.contents.facet.facet = "interface";

    /* this is particularly disgusting, but had to be done to get around*/
    /* the difficulty of creating subcells in oct.			*/
    /* if this causes problems, comment out the code and the		*/
    /* subcells will be made in the oFacet's directory			*/
    Path = oFacet.contents.facet.cell;
    for( c = Path + strlen( Path ); c != Path; c-- ){
	if( *c == '/' ){
	    Path = c + 1;
	    break;
	}
	/*
	    if( chdir( Path ) != 0 ){
		fprintf( stderr, "Could not chdir to %s: exiting", Path );
		exit( -1 );
	    }
	*/
    }
    if( octOpenFacet( &oFacet ) != OCT_OLD_FACET ){
	fprintf( stderr,"Fatal Error: Could not open facet %s:%s:%s: %s\n",
		 oFacet.contents.facet.cell,
		 oFacet.contents.facet.view,
		 oFacet.contents.facet.facet,
		 octErrorString());
	exit(-1);
    }
}
readOct()
{
    octObject oInst;
    octObject oTerm;
    octObject oNet;
    octGenerator gInst;
    octGenerator gTerm;
    octStatus sInst;
    octStatus sTerm;
    s_cell *cell;
    s_net *net;
    extern int octIdCmp(), octIdHash();
    char *ptr;

    oInstances.type = OCT_BAG;
    oInstances.contents.bag.name = "INSTANCES";
    if( octGetByName( &oFacet, &oInstances ) < OCT_OK ){
	fprintf( stderr, "Fatal Error: Could not get INSTANCES bag: %s\n",
			  octErrorString() );
	exit( -1 );
    }

    NetTbl = st_init_table(octIdCmp, octIdHash);

    OH_ASSERT( octInitGenContents( &oInstances, OCT_INSTANCE_MASK, &gInst ));
    while(( sInst = octGenerate( &gInst, &oInst )) == OCT_OK ){

	if( NumCells == 0 ){
	    Cells = ALLOC( s_cell *, 1 );
	    NumCells = 1;
	} else {
	    Cells = REALLOC( s_cell *, Cells, ++NumCells );
	}
	cell = Cells[ NumCells - 1 ] = ALLOC( s_cell, 1 );

	cell->objectId = oInst.objectId;
	cell->numNets = 0;
	cell->row = -1;
	cell->col = -1;
	cell->net = NIL( s_net * );

	/* allocate the array of nets and fill it in			*/
	OH_ASSERT( octInitGenContents( &oInst, OCT_TERM_MASK, &gTerm ));
	while(( sTerm = octGenerate( &gTerm, &oTerm )) == OCT_OK ){
	    if( octGenFirstContainer( &oTerm, OCT_NET_MASK, &oNet ) == OCT_OK ){
		if (!st_lookup(NetTbl,(char *) &oNet.objectId, &ptr)) {
		    octId *id = ALLOC(octId, 1);
		    net = ALLOC( s_net, 1 );
		    net->objectId = oNet.objectId;
		    net->next = Nets;
		    Nets = net;
		    net->numCells = 1;
		    net->cell = ALLOC( s_cell *, 1 );
		    *id = net->objectId;
		    assert(st_insert(NetTbl, (char *) id, (char *)net) != 1);
		    NumNets++;
		} else {
		    net = (s_net *) ptr;
		    net->cell = REALLOC( s_cell *, net->cell, ++net->numCells);
		}
		net->cell[ net->numCells - 1 ] = cell;

		if( cell->net == NIL( s_net * )){
		    cell->net = ALLOC( s_net *, 1 );
		    cell->numNets = 1;
		} else {
		    cell->net = REALLOC( s_net *, cell->net, ++cell->numNets );
		}
		cell->net[ cell->numNets - 1 ] = net;
	    }
	}
	OH_ASSERT( sTerm );
	if( CellWeightType == WEIGHT_TERMS ) cell->weight = cell->numNets;
	else  cell->weight = 1;
	
    }
    OH_ASSERT( sInst );
    if( !ReadFloorplanning )st_free_table( NetTbl );
}

writeOct()
{
    octObject **oBag;
    octObject oCell;
    octObject oPart;
    octObject oProp;
    octObject oPartBag;
    s_cell *cell;
    int row, col, index;
    char baseCellName[ 100 ], cellNameBuffer[ 128 ];

    oBag = ALLOC( octObject *, NumRows );
    for( row = 0; row < NumRows; row++ ){
	oBag[ row ] = ALLOC( octObject, NumCols );
	for( col = 0; col < NumCols; col++ ){
	    oBag[ row ][ col ].type = OCT_BAG;
	    oBag[ row ][ col ].contents.bag.name = "";
	    OH_ASSERT( octCreate( &oFacet, &oBag[ row ][ col ] ));
	}
    }
    for( index = 0; index < NumCells; index++ ){
	cell = Cells[ index ];
	oCell.objectId = cell->objectId;
	OH_ASSERT( octGetById( &oCell ));
	OH_ASSERT( octAttach( &oBag[ cell->row ][ cell->col ], &oCell ));
    }
    oPart.type = OCT_FACET;
    oPart.contents.facet.cell = oFacet.contents.facet.cell;
    oPart.contents.facet.view = oFacet.contents.facet.view;
    oPart.contents.facet.facet = "contents";
    oPart.contents.facet.mode = "w";
    oPart.contents.facet.version = OCT_CURRENT_VERSION;
    (void) ohUnpackFacetName(&oPart, PartName);
    if( HideSubcells ){
	(void) sprintf( baseCellName, "%s/%s", Path,
					oPart.contents.facet.cell );
    } else {
	(void) sprintf( baseCellName, "%s", oPart.contents.facet.cell );
    }
    oPart.contents.facet.cell = cellNameBuffer;

    oPartBag.type = OCT_BAG;
    oPartBag.contents.bag.name = "PARTITIONS";
    OH_ASSERT( octGetOrCreate( &oFacet, &oPartBag ));
    cellNames = ALLOC( char, (unsigned)( NumRows * NumCols * 
				 ( strlen( baseCellName ) + 9 )));
    *cellNames = '\0';

    for( row = 0; row < NumRows; row++ ){
	for( col = 0; col < NumCols; col++ ){
	    if( Bin[ row][col].count == 0 ){
		OH_ASSERT( octDelete( &oBag[ row ][ col ] ));
		fprintf( stderr,"Bin %d.%d was empty\n", row, col );
		continue;
	    }
	    oPart.type = OCT_FACET;
	    (void) sprintf( cellNameBuffer, "%s.%d.%d", baseCellName, row, col );
	    (void) sprintf( cellNames + strlen( cellNames)," %s",cellNameBuffer );
	    if( octOpenFacet( &oPart ) < OCT_OK ){
		fprintf( stderr,"Could not open cell %s:%s:%s: %s\n",
				oPart.contents.facet.cell,
				oPart.contents.facet.view,
				oPart.contents.facet.facet,
				octErrorString());
		halt();
	    }
	    if( nleMakeMacro( &oPart, &oBag[ row ][ col ], 
			      &oCell, NLE_SOFT ) < NLE_OK ){
		fprintf( stderr,"Problem with nle on partition %s:%s:%s: %s\n",
				oPart.contents.facet.cell,
				oPart.contents.facet.view,
				oPart.contents.facet.facet,
				nleErrorString());
		halt();
	    }
	    OH_ASSERT( octAttach( &oPartBag, &oCell ));
	    OH_ASSERT( octDelete( &oBag[ row ][ col ] ));
	}
    }
    oProp.type = OCT_PROP;
    oProp.contents.prop.name = "PARTITION_NAMES";
    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.value.string = cellNames;
    OH_ASSERT( octCreate( &oFacet, &oProp ));
    if( octFlushFacet( &oFacet ) < OCT_OK ){
	fprintf( stderr,"Could not flush cell %s/%s:%s:%s: %s\n",
			Path,
			oFacet.contents.facet.cell,
			oFacet.contents.facet.view,
			oFacet.contents.facet.facet,
			octErrorString());
	halt();
    }
}

/*
 * gives the PARTITION_NAMES prop the value "no_partitions"
 */
writeNoPartitionProp()
{
    octObject oProp;

    oProp.type = OCT_PROP;
    oProp.contents.prop.name = "PARTITION_NAMES";
    oProp.contents.prop.type = OCT_STRING;
    oProp.contents.prop.value.string = "no_partitions";
    OH_ASSERT( octCreateOrModify( &oFacet, &oProp ));
    if( octFlushFacet( &oFacet ) < OCT_OK ){
	fprintf( stderr,"Could not flush cell %s/%s:%s:%s: %s\n",
			Path,
			oFacet.contents.facet.cell,
			oFacet.contents.facet.view,
			oFacet.contents.facet.facet,
			octErrorString());
	halt();
    }
}
