#include "copyright.h"
#include "jumbo.h"
#include "options.h"

optionStruct optionList[] = {
    { "v",    	 "", 		"verbose mode"},
    { "n",    	 "", 		"do not implement the partition"},
    { "t",     	 "",	 	"weight cells by number of terminals"},
    { "f",    	 "", 		"read floorplanning information"},
    { "h",    	 "", 		"hide subcells under the output cell"},
    { "u",    	 "", 		"undo the results of a previous run"},
    { "r",     	 "rows",	"number of rows"},
    { "c",     	 "cols",	"number of columns"},
    { "s",     	 "bin size",	"desired size of each bin"},
    { "b",    	 "number bins",	"number of bins in the partition"},
    { "a",     	 "aspect ratio",	"desired rows / cols"},
    { "N",     	 "cost factor",	"relative cost of the net length"},
    { "T",     	 "cost factor",	"relative cost of net terminals"},
    { "B",     	 "cost factor",	"relative cost of irregular bin sizes"},
    { "p",    	 "cell:view:facet", 	"name and view of the partitions"}, 
    { "o",    	 "cell:view:facet", 	"name and view of the output facet"}, 
    { OPT_RARG,	 "cell:view:facet", 	"name and view of the input facet"},
    { 0,        0,               0}
};

/* global variables set by the command line				*/

extern int Verbose;
extern int NumRows;
extern int NumCols;
extern int ReadFloorplanning;
extern double TargetAspectRatio;
extern double NlFactor;
extern double NtFactor;
extern double BFactor;
extern int TargetBinSize;
extern int TargetNumBins;
extern int CellWeightType;
extern int NoImplement;
extern int Undo;
extern int HideSubcells;
extern char *InName;
extern char *OutName;
extern char *PartName;


/*
 * parseComLine
 *
 * parse the command line 
 */
parseComLine( argc, argv )
int argc;
char **argv;
{

    extern char *optarg;
    int option;


    while(( option = optGetOption( argc,argv)) != EOF ){
	switch ( option ) {
	    case 'v':
		Verbose = TRUE;
		break;
	    case 'f':
		ReadFloorplanning = TRUE;
		break;
	    case 'n':
		NoImplement = TRUE;
		break;
	    case 'h':
		HideSubcells = TRUE;
		break;
	    case 'u':
		Undo = TRUE;
		break;
	    case 'r':
		NumRows = atoi( optarg );
		if(( NumRows != -1 )&&( NumRows < 1 )){
		    fprintf(stderr, "At least one row must be specified\n");
		    optUsage();
		}
		break;
	    case 'c':
		NumCols = atoi( optarg );
		if(( NumCols != -1 )&&( NumCols < 1 )){
		    fprintf(stderr, "At least one column must be specified\n");
		    optUsage();
		}
		break;
	    case 'a':
		TargetAspectRatio = atof( optarg );
		if( TargetAspectRatio <= 0 ){
		    fprintf(stderr, "The aspect ratio must be > 0\n");
		    optUsage();
		}
		break;
	    case 's':
		TargetBinSize = atoi( optarg );
		break;
	    case 'b':
		TargetNumBins = atoi( optarg );
		break;
	    case 'B':
		BFactor = atof( optarg );
		if( BFactor < 0 ){
		    fprintf(stderr, "The bin factor must be >= 0\n");
		    optUsage();
		}
		break;
	    case 'N':
		NlFactor = atof( optarg );
		if( NlFactor < 0 ){
		    fprintf(stderr, "The netlength factor must be >= 0\n");
		    optUsage();
		}
		break;
	    case 'T':
		NtFactor = atof( optarg );
		if( NtFactor < 0 ){
		    fprintf(stderr, "The net fragmentation factor must be >= 0\n");
		    optUsage();
		}
		break;
	    case 't':
		CellWeightType = WEIGHT_TERMS;
		break;
	    case 'p':
		PartName = util_strsav( optarg );
		break;
	    case 'o':
		OutName = util_strsav( optarg );
		break;
	    default:
		optUsage();
		break;
	}
    }
    if(( argc == 1 )||( optind == argc )){
	optUsage();
	halt();
    }
    InName = util_strsav( argv[ argc - 1 ] );
}
