#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "utility.h"
#include "options.h"
#include "vov.h"

/* phyt.c
 * 3/16/91
 * Kristofer S. J. Pister
 */
optionStruct optionList[] = {
    { "h",      "lambda",      "desired text height (integer) in lambda [100]"},
    { "t",      "text_layer",  "what layer to place the text on [MET1]"},
    { "b",      "bb_layer",    "where to place the bounding box [PLACE]"},
    { "T",      "tech_name",   "name of the technology for the new facet [scmos]"},
    { "v",      "",            "be verbose about what's going on"},
    { OPT_RARG, "text_string", "text to be generated"},
    { OPT_OARG, "palette",     "palette to use for letters and symbols [8x13]"},
    { "o",      "cell",        "output cell name"},
    { OPT_DESC,  NULL, "phyt generates text on the layer text_layer, inside a bounding box \
on layer bb_layer.  Setting bb_layer to '-' suppresses bounding box generation.\n\
The default output cell name is the input text string (minus white space)."},
    {NULL, NULL, NULL}
};

#define DEF_PAL_NAME "$OCTTOOLS/lib/phyt/8x13:physical"
#define DEF_TEXT_LAYER "MET1"
#define DEF_BB_LAYER "PLACE"
#define DEF_TECHNOLOGY "scmos"
#define DEF_TEXT_HEIGHT 100

/*
 * ptext
 */
main(ac, av)
int ac;
char *av[];
{
    octObject palette, output, tmpProp;
    char *paletteName= DEF_PAL_NAME,     /* where to find the symbols */
         *outputName=  NULL,             /* what to call the output facet */
         *textString=  NULL,             /* what to put in the output facet */
         *technology=  DEF_TECHNOLOGY,   /* technology for new facet */
	 *textLayer=   DEF_TEXT_LAYER,   /* where to put text */
	 *bbLayer=     DEF_BB_LAYER;     /* where to put bounding box */
    int textHeight = DEF_TEXT_HEIGHT,    /* in lambda (desired) */
        verbose = 0;                     /* talk about it a lot */

    char * genGoodName();
    int option, scale, h,w,bx,by;

    octBegin();

    /* parse the command line */
    while ((option = optGetOption(ac,av)) != EOF) {
	switch (option) {
	case 'h':
	    textHeight = atoi(optarg);
	    break;
	case 't':
	    textLayer = util_strsav(optarg);
	    break;
	case 'b':
	    bbLayer = util_strsav(optarg);
	    break;
	case 'T':
	    technology = util_strsav(optarg);
	    break;
	case 'o':
	    outputName = util_strsav(optarg);
	    break;
	case 'v':
	    verbose++;
	    break;
	default:
	    optUsage();
	}
    }

    if (optind >= ac) {
	optUsage();
    }

    /* process the remaining args: <text> [palette] */
    textString = util_strsav(av[optind]);
    optind++;
    if (optind < ac) paletteName = util_strsav(av[optind]);

    /* generate an appropriate name from the text string */
    if (outputName == NULL) outputName = genGoodName(textString);

    VOVbegin( ac, av );
    if (verbose) {
	fprintf(stdout,"Generating facet %s in technology %s\n", outputName, technology);
	fprintf(stdout,"using text string `%s'\n",textString);
	fprintf(stdout,"and palette %s\n",paletteName);
	fprintf(stdout,"in %d lambda font on layer %s with bounding boxes on layer %s\n"
		,textHeight,textLayer,bbLayer);
	if (!strcmp(bbLayer,"-")) fprintf(stdout,"(no bounding boxes)\n");
    }

    /* open the palette (input) facet with mode == "r" */
    ohUnpackDefaults(&palette,"r",":physical:contents");
    ohUnpackFacetName(&palette,paletteName);
    VOVinputFacet( &palette );
    OH_ASSERT( octOpenFacet(&palette));

    /* open the phytext (output) facet with mode == "w" */
    ohUnpackDefaults(&output,"w",":physical:contents");
    ohUnpackFacetName(&output,outputName);
    VOVoutputFacet( &output );
    OH_ASSERT( octOpenFacet(&output) );

    /* set properties of output facet */
    ohCreatePropStr(&output,&tmpProp,"TECHNOLOGY",technology);
    ohCreatePropStr(&output,&tmpProp,"VIEWTYPE", "PHYSICAL");
    ohCreatePropStr(&output,&tmpProp,"EDITSTYLE","PHYSICAL");

    /* calculate scale based on desired height in lambda 
     * scale is desired text height of actual divided by the actual height of the
     * font (based on bbox).
     * The fonts are stored w/ one oct unit per pixel, so I need to multiply
     * the desired height be 20 to get the actual size in lambda.
     * I only use integers to avoid having 'non-lambda' sized boxes.
     * To get the closest integer, add 1/2 of denom. to numerator before division.
     */
    if (ohGetByPropName(&palette,&tmpProp,"FONTBOUNDINGBOX") != OCT_OK) {
	fprintf(stderr,"property FONTBOUNDINGBOX doesn't exist. Assuming 1 1 0 0\n");
	h = w = 1;
	bx = by = 0;
    } else {
	sscanf(tmpProp.contents.prop.value.string,"%d %d %d %d",&w,&h,&bx,&by);
    }
    textHeight *= 20;
    scale = (textHeight + (h-by) / 2) / (h-by);
    if (textHeight < (h-by)) {
	fprintf(stderr,"desired height (%d lambda) less than minimum possible (%d lambda).  Using minimum.\n",textHeight,h-by);
	scale = 1;
    }

    /* create the new instance */
    ptCreateTextInstance(&palette,&output,textString,textLayer,bbLayer,(double) scale);

    /* close everybody up */
    (void) octCloseFacet(&output);
    (void) octCloseFacet(&palette);
    octEnd();
    VOVend( 0 );
}












