#include "copyright.h"
#include "port.h"
#include "oct.h"
#include "oh.h"
#include "tr.h"
#include "errtrap.h"
#include "utility.h"

/* phytlib.c
 * 3/16/91
 * Kristofer S. J. Pister
 */

#define MAXFILENAMECHARS 20

ptCreateTextInstance(pal,out,text,textLayerName,bbLayerName,scale)
octObject *pal,*out;
char *text, *textLayerName, *bbLayerName;
double scale;
{
    octGenerator gen;
    char symbol[100], *textPtr;
    octObject letterBag, geo, formBag, BBox, bbLayer, textLayer;
    int xoffset=0, yoffset=0, llx, lly, i;
    tr_stack *xfStack;
    octTransform xform, scaleXform;

    /* create the appropriate layers in the new facet
     * *** only create and use bbLayer if bbLayerName is not "-"
     */
    ohCreateLayer(out,&textLayer,textLayerName);
    if (strcmp(bbLayerName,"-")) ohCreateLayer(out,&bbLayer,bbLayerName);

    /* create the transformation stack, and the scale transform */
    xfStack = tr_create();
    tr_scale(xfStack,scale,1);
    tr_get_oct_transform(xfStack,&scaleXform);

    /* step through each symbol in text */
    textPtr = text;
    while (*textPtr != '\0') {
	/* get the symbol name */
	if (*textPtr == '\\') {
	    if (textPtr[1] == '\\') {
		/* just an escaped backslash */
		strcpy(symbol,"\\");
	    } else {
		/* read symbol name until space or end of string */
		textPtr++;
		for (i=0; *textPtr != '\0' && !isspace(*textPtr); i++, textPtr++) {
		    symbol[i] = *textPtr;
		}
		symbol[i] = '\0';
		fprintf(stderr,"found symbol %s\n",symbol);
	    }
	} else {
	    /* just a regular character */
	    *symbol = *textPtr;
	    symbol[1] = '\0';
	    textPtr++;
	}

	/* get the bag corresponding to the symbol */
	if (ohGetByBagName(pal,&letterBag,symbol) != OCT_OK) {
	    errRaise("ptCreateTextInstance",1,"can't find bag containing symbol '%s' in text palette\n",symbol);
	}

	/* get the bounding box, translate to new location, create, update offsets */
	octGenFirstContent(&letterBag,OCT_BOX_MASK,&BBox);
	octTransformGeo(&BBox,&scaleXform);
	llx = BBox.contents.box.lowerLeft.x;
	lly = BBox.contents.box.lowerLeft.y;
	tr_identity(xfStack);
	tr_translate(xfStack, xoffset - llx, yoffset - lly,1);
	tr_get_oct_transform(xfStack,&xform);
	octTransformGeo(&BBox,&xform);
	if (strcmp(bbLayerName,"-")) octCreate(&bbLayer,&BBox);
	xoffset += BBox.contents.box.upperRight.x - BBox.contents.box.lowerLeft.x;

	/* get the form bag */
	if (ohGetByBagName(&letterBag,&formBag,"form") != OCT_OK) {
	    errRaise("ptCreateTextInstance",2,"can't find form bag for symbol %s in text palette\n",symbol);
	    continue;
	}

	/* generate all the geos in the form bag and transform */
	octInitGenContents(&formBag, OCT_GEO_MASK, &gen);
	while ( OCT_OK == octGenerate(&gen, &geo)) {
	    octTransformGeo(&geo,&scaleXform);
	    octTransformGeo(&geo,&xform);
	    octCreate(&textLayer,&geo);
	}
    }
}

char *genGoodName(s)
char *s;
{
    int i;
    char *d;

    if (s == NULL) errRaise("genGoodName",1,"no input string\n");

    /* create space for the new string */
    d = util_strsav(s);

    /* only take the first N alphanumerics */
    for (i=0; i<MAXFILENAMECHARS && *s != '\0' ; s++) {
	if (isalnum(*s)) {
	    d[i++] = *s;
	}
    }
    d[i] = '\0';

    return(d);
}
