#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "rpc.h"
#include "tap.h"
#include "oh.h"

/* usermain.c
 * 3/16/91
 * Kristofer S.J. Pister
 */
#define DEF_PAL_NAME "$OCTTOOLS/lib/phyt/8x13:physical"
#define DEF_TEXT_LAYER "MET1"
#define DEF_BB_LAYER "PLACE"
#define DEF_TECHNOLOGY "scmos"
#define DEF_TEXT_HEIGHT 100

#define MAXFILENAMECHARS 20

rpcStatus ptSetOptions(), ptCreateRPC(), exitApp();
RPCFunction CommandArray[] = {
    {ptCreateRPC, "miscellaneous", "phyt-create-text", "C", 0},
    {ptSetOptions, "miscellaneous", "phyt-set-options", "S", 0},
    {exitApp, "miscellaneous", "exit", "X", 0}};
    
octObject palette, output;
char *paletteName= DEF_PAL_NAME,     /* where to find the symbols */
     *outputName=  NULL,             /* what to call the output facet */
     *textString=  NULL,             /* what to put in the output facet */
     *technology=  DEF_TECHNOLOGY,   /* technology for new facet */
     *textLayer=   DEF_TEXT_LAYER,   /* where to put text */
     *bbLayer=     DEF_BB_LAYER;     /* where to put bounding box */
int textHeight = DEF_TEXT_HEIGHT;    /* in lambda (desired) */
int scale, verbose=1;

char *genGoodName();

long
UserMain(display, spot, cmdList, userOptionWord, array)
char *display;
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
RPCFunction **array;
{
    octObject facet, tmpProp;
    char buffer[4096];
    int h,w,bx,by;

    /* open the default font */
    ptNewFont();

    VEMMSG("rpcphyt initialized\n");
    vemPrompt();
    
    *array = CommandArray;
    return(sizeof(CommandArray) / sizeof(struct RPCFunction));
}

ptCreateRPC(spot,cmdList,uow)
RPCSpot * spot;
lsList cmdList;
long uow;
{
    Pointer pointer;
    RPCArg *arg;
    struct octPoint point;
    octObject master, tmpProp, instance;
    int instX, instY;

    if (lsLength(cmdList) != 2) {
	ptUsage();
	return(RPC_OK);
    }

    /* grab the first argument (point) */
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if ((arg->argType != VEM_POINT_ARG) || (arg->length != 1)) {
	ptUsage();
	return(RPC_OK);
    }

    /* get the point and the facet that the point is in */
    point = arg->argData.points[0].thePoint;
    master.objectId = arg->argData.points[0].facet;
    OH_ASSERT(octGetById(&master));
    instX = point.x;
    instY = point.y;
    
    /* grab the second arguement (text) */
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_TEXT_ARG) {
	ptUsage();
	vemPrompt();
	return(RPC_OK);
    }
    textString = util_strsav(arg->argData.string);
    outputName = genGoodName(textString);

    /* open the phytext (output) facet with mode == "w" */
    ohUnpackDefaults(&output,"w",":physical:contents");
    ohUnpackFacetName(&output,outputName);
    OH_ASSERT( octOpenFacet(&output) );

    /* set properties of output facet */
    ohCreatePropStr(&output,&tmpProp,"TECHNOLOGY",technology);
    ohCreatePropStr(&output,&tmpProp,"VIEWTYPE", "PHYSICAL");
    ohCreatePropStr(&output,&tmpProp,"EDITSTYLE","PHYSICAL");

    /* generate the new facet, instantiate it, and flush master */
    ptCreateTextInstance(&palette,&output,textString,textLayer,bbLayer, (double) scale);
    octFlushFacet(&output);

    OH_ASSERT(ohCreateInstance(&master,&instance,outputName,outputName,"physical",instX,instY,OCT_NO_TRANSFORM));
    octFlushFacet(&master);
    vemWnFlush();

    OH_ASSERT(octCloseFacet(&output));
    return(RPC_OK);
}



ptUsage()
{
    VEMMSG("\nformat: point \"string\" : phyt-create-text\n");
    vemPrompt();
}

/* ptSetOptions
 * this should be done w/ a dialog box w/ lots of useful bells and
 * whistles.  I just don't have time to learn how to create those.
 */
#define STRING 0
#define INT 1
static struct pairs {
    char *name;
    void *variable;
    int type, newpalette;
} Pairs[] = {
    {"textLayer", &textLayer, STRING,0},
    {"bbLayer", &bbLayer, STRING,0},
    {"textHeight", &textHeight, INT,1},
    {"paletteName",&paletteName, STRING,1},
    {"technology", &technology, STRING,0},
    { NULL, NULL, 0, 0 }
};

ptSetOptions(spot,cmdList,uow)
RPCSpot * spot;
lsList cmdList;
long uow;
{
    RPCArg *arg;
    Pointer pointer;
    char *optstr, *value, mesg[100];
    int i;

    if (lsLength(cmdList) != 1) {
	VEMMSG("\nformat: \"<option> <value>\" : phyt-set-options\n");
	VEMMSG("Options:\n\ttextLayer\n\tbbLayer\n\ttextHeight\n\tpaletteName\n\ttechnology\n");
	vemPrompt();
	return(RPC_OK);
    }
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_TEXT_ARG) {
	ptUsage();
	vemPrompt();
	return(RPC_OK);
    }
    optstr = util_strsav(arg->argData.string);

    /* look through all the options for a match */
    for (i=0 ; Pairs[i].name != NULL ; i++) {
	if ( !strncmp(optstr,Pairs[i].name,strlen(Pairs[i].name)) ) {
	    /* point value to the rest of the string, minus white space */
	    value = optstr + strlen(Pairs[i].name);
	    while (isspace(*value)) value++;

	    /* make assignment */
	    switch (Pairs[i].type) {
	    case STRING:
		Pairs[i].variable = value;
		sprintf(mesg,"%s now equals %s\n",Pairs[i].name,(char *)Pairs[i].variable);
		VEMMSG(mesg);
		vemPrompt();
		break;
	    case INT:
		*((int *)Pairs[i].variable) = atoi(value);
		sprintf(mesg,"%s now equals %d\n",Pairs[i].name,*((int *) Pairs[i].variable));
		VEMMSG(mesg);
		vemPrompt();
		break;
	    default:
		VEMMSG("unknown variable type\n");
		vemPrompt();
		break;
	    }

	    /* close the old font, open the new one */
	    if (Pairs[i].newpalette) {
		OH_ASSERT ( octCloseFacet(&palette));
		ptNewFont();
	    }

	    return(RPC_OK);
	}
    }

    sprintf(mesg,"couldn't find any matching options\ncall phyt-set-options with no arguments for a list of options\n");
    VEMMSG(mesg);
    vemPrompt();
    return(RPC_OK);

}


/* ptNewFont
 * open new palette, calculate new scale
 */
ptNewFont()
{
    octObject tmpProp;
    int h,w,bx,by;

    /* open the palette (input) facet with mode == "r" */
    ohUnpackDefaults(&palette,"r",":physical:contents");
    ohUnpackFacetName(&palette,paletteName);
    OH_ASSERT( octOpenFacet(&palette));

    if (ohGetByPropName(&palette,&tmpProp,"FONTBOUNDINGBOX") != OCT_OK) {
	fprintf(stderr,"property FONTBOUNDINGBOX doesn't exist. Assuming 1 1 0 0\n");
	h = w = 1;
	bx = by = 0;
    } else {
	sscanf(tmpProp.contents.prop.value.string,"%d %d %d %d",&w,&h,&bx,&by);
    }
     /* calculate scale based on desired height in lambda 
     * scale is desired text height of actual divided by the actual height of the
     * font (based on bbox).
     * The fonts are stored w/ one oct unit per pixel, so I need to multiply
     * the desired height be 20 to get the actual size in lambda.
     * I only use integers to avoid having 'non-lambda' sized boxes.
     * To get the closest integer, add 1/2 of denom. to numerator before division.
     */
   textHeight *= 20;
    scale = (textHeight + (h-by) / 2) / (h-by);
    if (textHeight < (h-by)) {
	fprintf(stderr,"desired height (%d lambda) less than minimum possible (%d lambda).  Using minimum.\n",textHeight,h-by);
	scale = 1;
    }
}

exitApp(spot,cmdList,uow)
RPCSpot * spot;
lsList cmdList;
long uow;
{
    (void) octCloseFacet(&palette);
    RPCExit();
}


