
#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "smfont.h"	/* font information	*/
#include "rpc.h"
#include "st.h"			/* symbol (hash) table package	*/
#include "tap.h"
#include "oh.h"
#include <sys/file.h>
#include "sm.h"

font *fontptr[MAXGLYPH];
char *FontFile;
char *FontDirectory;

st_table *fonts = (st_table *) 0;

char SMDisplayName[1024];

extern char *loadFont();

int GenLetters = 1;

/*
 * SlideMaker
 *
 * RPC program for slide creation
 *
 * Rick L Spickelmier, 10/9/86
 *
 * last modification, 3/1/87
 *
 * Clean up and additional features by: Chris Guthrie
 *
 *     need to move the origin of arrows so center-? works on them
 *
 */

RPCFunction CommandArray[] = {
    {horizBox, "creation", "horizontal-box", "H", 0},
    {vertBox, "creation", "vertical-box", "V", 0},
    {newString, "creation", "string", "S", 0},
    {newBox, "creation", "box", "B", 0},
    {newArrow, "creation", "arrow"},
    {placeInstance, "creation", "place-instance"},
    {chlayer, "editing", "change-color",},
    {chscale, "editing", "change-scale"},
    {chjust, "editing", "change-just"},
    {changeString, "editing", "change-string", "E"},
    {lineUpX, "moving", "line-up-x"},
    {lineUpY, "moving", "line-up-y"},
    {centerY, "moving", "center-y"},
    {spaceY, "moving", "space-y"},
    {group, "grouping", "group"},
    {ungroup, "grouping", "ungroup"},
    {setSequenceFile, "sequencing", "set-sequence-file"},
    {initScreen, "sequencing", "initialize-screen"},
    {loadUp, "sequencing", "load-up"},
    {nextSlide, "sequencing", "next-slide", "N", 0},
    {previousSlide, "sequencing", "previous-slide", "P", 0},
    {continuous, "sequencing", "continuous"},
    {listfonts, "miscellaneous", "fonts"},
    {exitApp, "miscellaneous", "exit", "X", 0}};
    

/*ARGSUSED*/
long
UserMain(display, spot, cmdList, userOptionWord, array)
char *display;
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
RPCFunction **array;
{
    octObject facet, prop;
    char buffer[4096];

    (void) strcpy(SMDisplayName, display);
    
    (void) sprintf(buffer, "%s/%s", "$OCTTOOLS", FONTFILE);
    FontFile = util_file_expand(buffer);

    if (access(FontFile, F_OK) != 0) {
        char errmsg[4096];
        (void) sprintf(errmsg, "The letter files (%s, %s) are not installed\n",
                       FontFile, buffer);
	VEMMSG(errmsg);
	RPCExit(0);
    }

    getfonts(FontFile);

    /* set up the hash table */
    if ((fonts = st_init_table(strcmp, st_strhash)) == NULL) {
	VEMMSG("initialization error\n");
	RPCExit(0);
    }

    (void) sprintf(buffer, "%s/%s", "$OCTTOOLS", FONTDIRECTORY);
    FontDirectory = util_file_expand(buffer);

    if (access(FontDirectory, F_OK) != 0) {
        char errmsg[4096];
        (void) sprintf(errmsg, "The font definition files (%s, %s) are not installed\n",
                       FontDirectory, buffer);
	VEMMSG(errmsg);
	RPCExit(0);
    }

    (void) st_insert(fonts, "math",    (char *) loadFont(FontDirectory, "math.hmp"));
    (void) st_insert(fonts, "greekc",  (char *) loadFont(FontDirectory, "greekc.hmp"));
    (void) st_insert(fonts, "italict", (char *) loadFont(FontDirectory, "italict.hmp"));
    (void) st_insert(fonts, "romans",  (char *) loadFont(FontDirectory, "romans.hmp"));
    (void) st_insert(fonts, "romand",  (char *) loadFont(FontDirectory, "romand.hmp"));
    (void) st_insert(fonts, "romant",  (char *) loadFont(FontDirectory, "romant.hmp"));
    
    /* make sure the cell has the right properties */
    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");

    if (ohCreateOrModifyPropStr(&facet, &prop, "TECHNOLOGY", "slide") < OCT_OK) {
	(void) octError("can not update the TECHNOLOGY property");
	RPCExit(0);
    }
    
    if (ohCreateOrModifyPropStr(&facet, &prop, "VIEWTYPE", "SLIDE") < OCT_OK) {
	(void) octError("can not update the VIEWTYPE property");
	RPCExit(0);
    }
    
    if (ohCreateOrModifyPropStr(&facet, &prop, "EDITSTYLE", "PHYSICAL") < OCT_OK) {
	(void) octError("can not update the EDITSTYLE property");
	RPCExit(0);
    }

    vemFlushTechInfo(&facet);

    /* make sure unreadable letters are not created */
    (void) umask(000);

    (void) sprintf(buffer, "\n2SLIDE-MAKER, Version %d\nproblems/bugs to slide-maker@eros.berkeley.edu0\n",
	    MAJORVERSION);
    VEMMSG(buffer);
    
    VEMMSG("\n2Initialized0\n");
    vemPrompt();

    *array = CommandArray;
    return(sizeof(CommandArray) / sizeof(struct RPCFunction));
}


#ifdef lint
main(argc, argv)
int argc;
char **argv;
{
    (void) UserMain("", (RPCSpot *) 0, (lsList) 0, (long) 0, (RPCFunction **) 0);
}
#endif lint

