#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "oct.h"
#include "oh.h"
#include "options.h"
#include "vov.h"

optionStruct optionList[] = {
    { "f",		"list",		"file containing slide list" },
    { "x",		"xcount",	"number of slides across" },
    { "y",		"ycount",	"number of slides down" },
    { OPT_RARG,		"head",		"head of the names of the palettes" },
    { OPT_RARG,		"< list",	"list of slides" },
    { OPT_DESC,		0,		"creates head-1:slide, head-2:slide,... until the list of slides is exhausted" },
    { 0, 0, 0},
};



void
newPalette(palette, head)
    octObject *palette;
    char *head;
{
    static int count = 0;
    static char name[1024];
    octObject prop;

    if (palette->objectId != oct_null_id) {
	OH_ASSERT(octCloseFacet(palette));
    }

    (void) sprintf(name, "%s-%d", head, ++count);

    palette->contents.facet.cell = name;

    VOVoutputFacet( palette );

    OH_ASSERT(octOpenFacet(palette));

    (void) fprintf(stderr, "starting palette: %s\n", name);

    OH_ASSERT(ohCreatePropStr(palette, &prop, "TECHNOLOGY", "slide"));
    OH_ASSERT(ohCreatePropStr(palette, &prop, "VIEWTYPE", "SLIDE"));
    OH_ASSERT(ohCreatePropStr(palette, &prop, "EDITSTYLE", "PHYSICAL"));

    return;
}

char *optProgName;

#define WIDTH 11000.0
#define HEIGHT 8500.0

main(argc, argv)
    int argc;
    char **argv;
{
    char string[4096];
    char *filename = 0;
    octObject palette, slide, instance;
    struct octBox box;
    int x = 0;
    int xmax, ymax, xsize, ysize;
    int xmaxorig = 3;
    int ymaxorig = 3;
    double scale, xscale, yscale;
    int option;
    extern int optind;
    extern char *optarg;

    optProgName = argv[0];

    while ((option = optGetOption(argc, argv)) != EOF) {
	switch(option) {
	case 'x': xmaxorig = atoi(optarg); break;
	case 'y': ymaxorig = atoi(optarg); break;
	case 'f': filename = util_strsav(optarg); break;
	default: optUsage();
	}
    }

    if (argc == optind+1) {
	palette.type = OCT_FACET;
	palette.objectId = oct_null_id;
	palette.contents.facet.view = "slide";
	palette.contents.facet.facet = "contents";
	palette.contents.facet.version = OCT_CURRENT_VERSION;
	palette.contents.facet.mode = "w";
    } else {
	optUsage();
    }

    VOVbegin( argc, argv );
    newPalette(&palette, argv[optind]);
    xmax = xmaxorig;
    ymax = ymaxorig;

    if ( filename ) {
	VOVinput( VOV_UNIX_FILE, filename );
	if ( freopen( filename, "r", stdin ) == 0 ) {
	    errRaise( "SP",1,"Cannot open file %s", filename );
	}
    } 
    while (gets(string) != NULL) {

	ohUnpackDefaults(&slide, "r", ":slide:contents");
	ohUnpackFacetName(&slide, string);

	VOVinputFacet( &slide );
	if (octOpenFacet(&slide) < OCT_OK) {
	    errRaise( "SP", 1, "opening %s", ohFormatName(&slide));
	}

	if (x >= xmax) {
	    x = 0;
	    ymax--;
	}

	if (ymax == 0) {
	    newPalette(&palette, argv[optind]);
	    x = 0;
	    xmax = xmaxorig;
	    ymax = ymaxorig;
	}

	OH_ASSERT(octBB(&slide, &box));

	xsize = box.upperRight.x - box.lowerLeft.x;
	ysize = box.upperRight.y - box.lowerLeft.y;

	/* scale to HxW */
	
	xscale = WIDTH / (double) xsize;
	yscale = HEIGHT / (double) ysize;
	scale = MIN(xscale, yscale);

	instance.type = OCT_INSTANCE;
	instance.contents.instance.master = slide.contents.facet.cell;
	instance.contents.instance.view = slide.contents.facet.view;
	instance.contents.instance.facet = "contents";	
	instance.contents.instance.name = string;
	instance.contents.instance.transform.translation.x
	  = (x * (int) WIDTH) - scale * (box.upperRight.x + box.lowerLeft.x)/2;
	instance.contents.instance.transform.translation.y
	  = (ymax * (int) HEIGHT) - scale * (box.upperRight.y + box.lowerLeft.x)/2;
	instance.contents.instance.transform.transformType = OCT_FULL_TRANSFORM;
	instance.contents.instance.transform.generalTransform[0][0] = scale;
	instance.contents.instance.transform.generalTransform[0][1] = 0;
	instance.contents.instance.transform.generalTransform[1][0] = 0;
	instance.contents.instance.transform.generalTransform[1][1] = scale;

	OH_ASSERT(octCreate(&palette, &instance));

	x++;
    }
    OH_ASSERT(octCloseFacet(&palette));
    VOVend( 0 ) ;
}

