/*
 * slide-preen
 *
 *   remove all unused slide-maker facets from a group of slides
 *
 *   usage: slide-preen slide1 slide2 .... sliden
 *
 *   Rick L Spickelmier, 11/5/87
 *
 *
 * Note: this program makes many assumptions about how slide-maker slides
 * (oct facets) are stored.  DO NOT ATTEMPT TO RUN THIS ON NON-SLIDE-MAKER
 * (OR TEXT-MAKER) OCT FACETS.
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/dir.h>
#include "oct.h"
#include "oh.h"
#include "st.h"


#define NIL(type)   (type *) 0

#define OCTASSERT(function, msg) \
  if (function < OCT_OK) {       \
      (void) octError(msg);      \
      exit(-1);                  \
  }

st_table *Table;
extern int strcmp();
extern int strncmp();


main(argc, argv)
int argc;
char **argv;
{
    octObject facet, tech;
    int i;
    DIR *dirp;
    struct direct *dp;
    char *cell;
    char command[1024];
    st_generator *stgen;

    
    for (i = 1; i < argc; i++) {
	
	facet.type = OCT_FACET;
	facet.contents.facet.cell = argv[i];
	facet.contents.facet.view = "slide";
	facet.contents.facet.facet = "contents";
	facet.contents.facet.version = OCT_CURRENT_VERSION;
	facet.contents.facet.mode = "r";

	if (octOpenFacet(&facet) < OCT_OK) {
	    (void) fprintf(stderr, "%s can not be opened, skipping\n", ohFormatName(&facet));
	    continue;
	}

	/* get the TECHNOLOGY property, if not slide, skip */
	tech.type = OCT_PROP;
	tech.contents.prop.name = "TECHNOLOGY";
	if ((octGetByName(&facet, &tech) < OCT_OK) ||
	    (strcmp(tech.contents.prop.value.string, "slide") != 0)) {
	    (void) fprintf(stderr, "`%s' is not a slide-maker facet, skipping\n", argv[i]);
	    continue;
	}
	
	Table = st_init_table(strcmp, st_strhash);
    
	dirp = opendir(argv[i]);
	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
	    if ((strcmp(dp->d_name, "slide") == 0) ||
		(strcmp(dp->d_name, ".") == 0) ||
		(strcmp(dp->d_name, "..") == 0)) {
		
		continue;
	    }

	    if ((strncmp(dp->d_name, "string-", 7) != 0) &&
		(strncmp(dp->d_name, "group-", 6) != 0) &&
		(strncmp(dp->d_name, "box-", 4) != 0) &&
		(strncmp(dp->d_name, "arrow-", 6) != 0)) {

		continue;
	    }
	    
	    (void) st_insert(Table, dp->d_name, NIL(char));
	}

	(void) closedir(dirp);

	processFacet(&facet);
	
	/* remove to garbage */
	stgen = st_init_gen(Table);

	chdir(argv[i]);
	while (st_gen(stgen, &cell, (char **) 0) != 0) {
	    (void) sprintf(command,"/bin/rm -rf %s", cell);
	    (void) fprintf(stderr, "executing: %s\n", command);
	    system(command);
	}
	chdir("..");
    }
}


processFacet(facet)
octObject *facet;
{
    octGenerator gen;
    octObject inst, master;
    char *dummy;
    st_generator *stgen;

    
    OCTASSERT(octInitGenContents(facet, OCT_INSTANCE_MASK, &gen), "init gen");

    while (octGenerate(&gen, &inst) == OCT_OK) {
	st_delete(Table, &inst.contents.instance.master, &dummy);
	if (strncmp(inst.contents.instance.master, "group-", 6) == 0) {
	    master.type = OCT_FACET;
	    master.contents.facet.facet = "contents";
	    master.contents.facet.mode = "r";
	    OCTASSERT(octOpenMaster(&inst, &master), "open master");
	    processFacet(&master);
	}
    }
}

