#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <math.h>
#include "smfont.h"
#include "rpc.h"
#include "st.h"
#include "tap.h"
#include "oh.h"
#include "sm.h"
  
font *fontptr[MAXGLYPH];

extern st_table *fonts;

static dmTextItem features[] = {{ "TEXT", 5, 60, 0, 0 },
				{ "FONT", 1, 20, 0, 0 },
				{ "COLOR", 1, 20, 0, 0 },
				{ "SCALE", 1, 20, 0, 0 },
				{ "JUST", 1, 20, 0, 0 } } ;

/*
 * SlideMaker
 *
 * RPC program for slide creation
 *
 * Rick L Spickelmier, 10/9/86
 *
 * last modification, 11/6/88
 *
 * Clean up and additional features by: Chris Guthrie
 *
 */

/*ARGSUSED*/
int
newString(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg;
    Pointer pointer;
    octObject facet, instance, prop;
    char string[1024], fontName[1024], colorName[1024], justification[1024];
    double scale;
    struct octPoint point;
    int len;
    static int initialized = 0;
    
    
    len = lsLength(cmdList);
    
    if ((len != 1) && (len != 2)) {
	VEMMSG("format: point \"string\" : string (cursor over another string)\n");
	VEMMSG("format: point : string\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if ((arg->argType != VEM_POINT_ARG) || (arg->length != 1)) {
	VEMMSG("format: point \"string\" : string (cursor over another string)\n");
	VEMMSG("format: point : string\n");
	vemPrompt();
	return(RPC_OK);
    }

    point = arg->argData.points[0].thePoint;

    /* get the facet that the point is in */
    facet.objectId = arg->argData.points[0].facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");
    
    if (len == 1) {

	if (initialized == 0) {
	    features[0].value = "";
	    features[1].value = "";
	    features[2].value = "";
	    features[3].value = "";
	    features[4].value = "";
	    initialized = 1;
	}
	
	if (dmMultiText("String Creation Menu",
			sizeof(features) / sizeof(dmTextItem), features) == VEM_NOSELECT) {
	    VEMMSG("Aborting...\n");
	    vemPrompt();
	    return(RPC_OK);
	}

	stripNewLines(features[0].value);

	if (!st_lookup(fonts, features[1].value, (char **) 0)) {
	    VEMMSG("Bad Font\n");
	    vemPrompt();
	    return(RPC_OK);
	}

	if (tapIsLayerDefined(&facet, features[2].value) != OCT_OK) {
	    VEMMSG("Bad Color\n");
	    vemPrompt();
	    return(RPC_OK);
	}
	    
	(void) sscanf(features[3].value, "%lf", &scale);

	if (justValid(features[4].value) == 0) {
	    vemPrompt();
	    return(RPC_OK);
	}
	
	return(createString(&facet, point,
			    features[0].value,
			    features[1].value,
			    features[2].value,
			    scale,
			    features[4].value));
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_TEXT_ARG) {
	VEMMSG("format: point \"string\" : string (cursor over another string)\n");
	VEMMSG("format: point : string\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) strcpy(string, arg->argData.string);
    
    if (vuFindSpot(spot, &instance, OCT_INSTANCE_MASK) != VEM_OK) {
	VEMMSG("no string under the cursor\n");
	vemPrompt();
	return(RPC_OK);
    }

    /* find properties on the instance */
    if ((ohGetByPropName(&instance, &prop, "TYPE") < OCT_OK)
	|| (strcmp(prop.contents.prop.value.string, "TEXT") != 0)) {
	VEMMSG("not a TEXT object\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) GetStringProperty(&instance, "FONT", fontName, "romant");
    if (!st_lookup(fonts, fontName, (char **) 0)) {
	VEMMSG("Bad Font\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) GetStringProperty(&instance, "LAYER", colorName, "cyan");
    if (tapIsLayerDefined(&facet, colorName) != OCT_OK) {
	VEMMSG("Bad Color\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) GetStringProperty(&instance, "JUSTIFICATION", justification, "left");
    if (justValid(justification) == 0) {
	vemPrompt();
	return(RPC_OK);
    }
	
    if (ohGetByPropName(&instance, &prop, "SCALE") < OCT_OK) {
	scale = 1.0;
    } else {
	if (prop.contents.prop.type == OCT_REAL) {
	    scale = prop.contents.prop.value.real;
	} else {
	    scale = (double) prop.contents.prop.value.integer;
	}
    }
    
    return(createString(&facet, point, string, fontName, colorName,
			scale, justification));
}


/*ARGSUSED*/
int
newBox(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject facet, instance;
    struct octBox box;
    char colorName[1024];
    int len, i;
    

    len = lsLength(cmdList);
    
    if ((len != 2) && (len != 1)) {
	VEMMSG("format: box \"color\" : box\n");
	VEMMSG("        box : box\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_BOX_ARG) {
	VEMMSG("format: box \"color\" : box\n");
	VEMMSG("        box : box\n");
	vemPrompt();
	return(RPC_OK);
    }

    facet.type = OCT_FACET;
    facet.objectId = arg->argData.points[0].facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");

    if (len == 2) {
	
	(void) lsDelBegin(cmdList, &pointer);
	arg2 = (RPCArg *) pointer;
	if (arg2->argType != VEM_TEXT_ARG) {
	    VEMMSG("format: box \"color\" : box\n");
	    VEMMSG("        box : box\n");
	    return(RPC_OK);
	}
    
	(void) strcpy(colorName, arg2->argData.string);

    } else {
	
	if (vuFindSpot(spot, &instance, OCT_INSTANCE_MASK) != VEM_OK) {
	    VEMMSG("No color specified\n");
	    vemPrompt();
	    return(RPC_OK);
	}

	(void) GetStringProperty(&instance, "LAYER", colorName, "cyan");
    }
        
    if (tapIsLayerDefined(&facet, colorName) != OCT_OK) {
	VEMMSG("Bad Color\n");
	vemPrompt();
	return(RPC_OK);
    }

    for (i = 0; i < arg->length * 2; i += 2) {
	box.lowerLeft = arg->argData.points[i].thePoint;
	box.upperRight = arg->argData.points[i+1].thePoint;
	(void) createBox(spot, &facet, &box, colorName, "left");
    }
    return(RPC_OK);
}


/*ARGSUSED*/
int
createBox(spot, facet, box, colorName, justification)
RPCSpot *spot;
octObject *facet;
struct octBox *box;
char *colorName;
char *justification;
{
    octObject newfacet, color, path, instance, prop;
    struct octPoint points[5];
    struct octBox bb;
    char newfacetName[1024];
    int x, y;
    long loc;

    (void) time(&loc);

    if (tapIsLayerDefined(facet, colorName) != OCT_OK) {
	VEMMSG("Bad Color\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (octFlushFacet(facet) < OCT_OK) {  /* to force directory creation */
	(void) octError("can not flush the facet");
	return(RPC_OK);
    }

    (void) sprintf(newfacetName, "%s/box-%ld", facet->contents.facet.cell, loc);
    newfacet.type = OCT_FACET;
    newfacet.contents.facet.cell = newfacetName;
    newfacet.contents.facet.view = "slide";
    newfacet.contents.facet.facet = "contents";
    newfacet.contents.facet.version = OCT_CURRENT_VERSION;
    newfacet.contents.facet.mode = "w";

    if (octOpenFacet(&newfacet) < OCT_OK) {
	(void) octError("can not open the facet");
	return(RPC_OK);
    }
        
    color.type = OCT_LAYER;
    color.contents.layer.name = colorName;
    if (octGetOrCreate(&newfacet, &color) < OCT_OK) {
	(void) octError("can not create the layer");
	return(RPC_OK);
    }

    path.type = OCT_PATH;
    path.contents.path.width = 0;
    
    if (octCreate(&color, &path) < OCT_OK) {
	(void) octError("can not create the path");
	return(RPC_OK);
    }

    /* translate the points */
    x = SMMIN(box->lowerLeft.x, box->upperRight.x);
    y = SMMIN(box->lowerLeft.y, box->upperRight.y);
    
    points[0].x = box->lowerLeft.x - x;
    points[0].y = box->lowerLeft.y - y;
    points[1].x = box->upperRight.x - x;
    points[1].y = box->lowerLeft.y - y;
    points[2].x = box->upperRight.x - x;
    points[2].y = box->upperRight.y - y;
    points[3].x = box->lowerLeft.x - x;
    points[3].y = box->upperRight.y - y;
    points[4].x = points[0].x;
    points[4].y = points[0].y;

    OCTASSERT(octPutPoints(&path, 5, points), "can not put points");

    if (addPolicyProperties(&newfacet) < OCT_OK) {
	VEMMSG("can not create the policy properties\n");
	return(RPC_OK);
    }
    
    if (createInterface(&newfacet, colorName, "box") != OCT_OK) {
	return(RPC_OK);
    }
    
    if (octFlushFacet(&newfacet) < OCT_OK) {
	(void) octError("close failed");
	return(RPC_OK);
    }

    /* place the instance */
    instance.type = OCT_INSTANCE;
    instance.contents.instance.name = "box";
    instance.contents.instance.master = celltail(newfacetName);
    instance.contents.instance.view = "slide";
    instance.contents.instance.facet = "contents";
    instance.contents.instance.version = OCT_CURRENT_VERSION;

    instance.contents.instance.transform.translation.x
      = x - justOffset(justification, &newfacet);
    instance.contents.instance.transform.translation.y = y;
    instance.contents.instance.transform.transformType = OCT_NO_TRANSFORM;

    if (octCreate(facet, &instance) < OCT_OK) {
	(void) octError("can not create the instance");
	return(RPC_OK);
    }

    OCTASSERT(ohCreatePropStr(&instance, &prop, "TYPE", "BOX"),
	      "can not create the TYPE");
    OCTASSERT(ohCreatePropStr(&instance, &prop, "LAYER", colorName),
	      "can not create the LAYER");
    OCTASSERT(ohCreatePropStr(&instance, &prop, "JUSTIFICATION", justification),
	      "can not create the JUSTIFCATION");
          
    if (octBB(&instance, &bb) < OCT_OK) {
	(void) octError("can not get the bounding box");
	return(RPC_OK);
    }

    WNQREGION(facet, &bb);
    (void) vemWnFlush();
    
    VEMMSG("Finished\n");
    vemPrompt();

    return(RPC_OK);
}


/*ARGSUSED*/
int
newArrow(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg1, *arg2;
    Pointer pointer;
    octObject facet, instance;
    struct octPoint point;
    int stalk, head;
    double stalkAngle, headAngle;
    char colorName[1024];
    int len;
    

    len = lsLength(cmdList);
    
    if ((len != 2) && (len != 1)) {
	VEMMSG("format: line(s) \"color\" : arrow\n");
	VEMMSG("        line(s) : arrow\n");
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg1 = (RPCArg *) pointer;
    if ((arg1->argType != VEM_LINE_ARG) || (arg1->length < 2) || (arg1->length > 3)) {
	VEMMSG("format: line(s) \"color\" : arrow\n");
	VEMMSG("        line(s) : arrow\n");
	return(RPC_OK);
    }

    facet.type = OCT_FACET;
    facet.objectId = arg1->argData.points[0].facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");
	
    if (len == 2) {
	
	(void) lsDelBegin(cmdList, &pointer);
	arg2 = (RPCArg *) pointer;
	if (arg2->argType != VEM_TEXT_ARG) {
	    VEMMSG("format: line(s) \"color\" : arrow\n");
	    VEMMSG("        line(s) : arrow\n");
	    return(RPC_OK);
	}

	(void) strcpy(colorName, arg2->argData.string);
    
    } else {

	if (vuFindSpot(spot, &instance, OCT_INSTANCE_MASK) != VEM_OK) {
	    VEMMSG("nothing under the spot\n");
	    return(RPC_OK);
	}

	(void) GetStringProperty(&instance, "LAYER", colorName, "cyan");
    }
        
    if (tapIsLayerDefined(&facet, colorName) != OCT_OK) {
	VEMMSG("Bad Color\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    point = arg1->argData.points[1].thePoint;

    stalk = (int) hypot((double) (arg1->argData.points[0].thePoint.x
				  - arg1->argData.points[1].thePoint.x),
			(double) (arg1->argData.points[0].thePoint.y
				  - arg1->argData.points[1].thePoint.y));

    if (arg1->length == 3) {
	head = (int) hypot((double) (arg1->argData.points[2].thePoint.x
				     - arg1->argData.points[1].thePoint.x),
			   (double) (arg1->argData.points[2].thePoint.y
				     - arg1->argData.points[1].thePoint.y));
    } else {
	head = stalk / 10;
	if (head == 0) {
	    head = 1;
	}
    }

    /* must substract points[1] */
    stalkAngle = atan2((double) (arg1->argData.points[0].thePoint.x
				 - arg1->argData.points[1].thePoint.x),
		       (double) (arg1->argData.points[0].thePoint.y
				 - arg1->argData.points[1].thePoint.y));

    if (arg1->length == 3) {
	headAngle = stalkAngle - atan2((double) (arg1->argData.points[2].thePoint.x
						 - arg1->argData.points[1].thePoint.x),
				       (double) (arg1->argData.points[2].thePoint.y
						 - arg1->argData.points[1].thePoint.y));
    } else {
	headAngle = 3.1415926 / 6.0;
    }
    
    return(createArrow(spot, &facet, point, stalk, head, stalkAngle, headAngle,
		       colorName));
}


/*ARGSUSED*/
int
createArrow(spot, facet, point, stalk, head, stalkAngle, headAngle, colorName)
RPCSpot *spot;
octObject *facet;
struct octPoint point;
int stalk, head;
double stalkAngle, headAngle;
char *colorName;
{
    octObject newfacet, color, instance, path, prop;
    struct octPoint points[2];
    struct octBox bb;
    double c, s;
    long loc;
    char newfacetName[1024];
    
    (void) time(&loc);

    if (tapIsLayerDefined(facet, colorName) != OCT_OK) {
	VEMMSG("Bad Color\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    /* OCTASSERT(octFlushFacet(facet), "can not flush the facet"); */
    sm_create_dir( facet );

    (void) sprintf(newfacetName, "%s/arrow-%ld", facet->contents.facet.cell, loc);
    
    newfacet.type = OCT_FACET;
    newfacet.contents.facet.cell = newfacetName;
    newfacet.contents.facet.view = "slide";
    newfacet.contents.facet.facet = "contents";
    newfacet.contents.facet.version = OCT_CURRENT_VERSION;
    newfacet.contents.facet.mode = "w";
    OCTASSERT(octOpenFacet(&newfacet), "can not open the facet");
    
    color.type = OCT_LAYER;
    color.contents.layer.name = colorName;
    OCTASSERT(octGetOrCreate(&newfacet, &color), "can not get or create the LAYER");

    /* create the STALK */

    path.type = OCT_PATH;
    path.contents.path.width = 0;
    OCTASSERT(octCreate(&color, &path), "can not create the PATH");

    points[0].x = 0;
    points[0].y = 0;
    points[1].x = 0;
    points[1].y = stalk;

    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");
    
    /* create the "vee" */
    
    OCTASSERT(octCreate(&color, &path), "can not create the PATH");
    
    s = sin(headAngle);
    c = cos(headAngle);

    points[0].x = 0;
    points[0].y = 0;
    points[1].x = (int) (head * s);
    points[1].y = (int) (head * c);

    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");
    
    OCTASSERT(octCreate(&color, &path), "can not create the PATH");
        
    points[1].x = -points[1].x;
  
    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");

    if (addPolicyProperties(&newfacet) < OCT_OK) {
	VEMMSG("can not create the policy properties\n");
	return(RPC_OK);
    }

    if (createInterface(&newfacet, colorName, "arrow") != OCT_OK) {
	return(RPC_OK);
    }

    OCTASSERT(octFlushFacet(&newfacet), "can not flush the facet");
        
    /* place the instance */
    instance.type = OCT_INSTANCE;
    instance.contents.instance.name = "arrow";
    instance.contents.instance.master = celltail(newfacetName);
    instance.contents.instance.view = "slide";
    instance.contents.instance.facet = "contents";
    instance.contents.instance.version = OCT_CURRENT_VERSION;

    instance.contents.instance.transform.translation.x = point.x;
    instance.contents.instance.transform.translation.y = point.y;
    instance.contents.instance.transform.transformType = OCT_FULL_TRANSFORM;
    instance.contents.instance.transform.generalTransform[0][0] = cos(stalkAngle);
    instance.contents.instance.transform.generalTransform[0][1] = sin(stalkAngle);
    instance.contents.instance.transform.generalTransform[1][0] = -sin(stalkAngle);
    instance.contents.instance.transform.generalTransform[1][1] = cos(stalkAngle);

    OCTASSERT(octCreate(facet, &instance), "can not create the instance");
    
    OCTASSERT(ohCreatePropStr(&instance, &prop, "TYPE", "ARROW"),
	      "can not create TYPE");
    OCTASSERT(ohCreatePropStr(&instance, &prop, "LAYER", colorName),
	      "can not create LAYER");
    OCTASSERT(ohCreatePropInt(&instance, &prop, "STALK-LENGTH", stalk),
	      "can not create STALK-LENGTH");
    OCTASSERT(ohCreatePropInt(&instance, &prop, "HEAD-LENGTH", head),
	      "can not create HEAD-LENGTH");
    OCTASSERT(ohCreatePropReal(&instance, &prop, "STALK-ANGLE", stalkAngle),
	      "can not create STALK-ANGLE");
    OCTASSERT(ohCreatePropReal(&instance, &prop, "HEAD-ANGLE", headAngle),
	      "can not create HEAD-ANGLE");
    
    OCTASSERT(octBB(&instance, &bb), "can not get the bounding box");
    
    WNQREGION(facet, &bb);

    (void) vemWnFlush();

    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
        
}


int
horizBox(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    octObject facet, layer, path;
    struct octPoint points[2];
    

    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");

    layer.type = OCT_LAYER;
    layer.contents.layer.name = "pageFrame";
    OCTASSERT(octGetOrCreate(&facet, &layer), "can not find the pageFrame layer");
    
    path.type = OCT_PATH;
    path.contents.path.width = 0;
    OCTASSERT(octCreate(&layer, &path), "can not create the path");

    points[0].x = OFFSET;
    points[0].y = OFFSET;
    points[1].x = (octCoord) (PAGEHEIGHT * UNITSPERINCH) + OFFSET;
    points[1].y = points[0].y;

    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");

    OCTASSERT(octCreate(&layer, &path), "can not create the path");

    points[0].x = OFFSET;
    points[0].y = (octCoord) (PAGEWIDTH * UNITSPERINCH) + OFFSET;
    points[1].x = (octCoord) (PAGEHEIGHT * UNITSPERINCH) + OFFSET;
    points[1].y = points[0].y;

    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");
    
    OCTASSERT(octCreate(&layer, &path), "can not create the path");
    
    points[0].x = OFFSET;
    points[0].y = OFFSET;
    points[1].x = points[0].x;
    points[1].y = (octCoord) (PAGEWIDTH * UNITSPERINCH) + OFFSET;
    
    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");

    OCTASSERT(octCreate(&layer, &path), "can not create the path");
    
    points[0].x = (octCoord) (PAGEHEIGHT * UNITSPERINCH) + OFFSET;
    points[0].y = OFFSET;
    points[1].x = points[0].x;
    points[1].y = (octCoord) (PAGEWIDTH * UNITSPERINCH) + OFFSET;
    
    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");
    
    (void) vemCommand("show-all", spot, cmdList, userOptionWord);

    VEMMSG("Finished\n");
    vemPrompt();

    return(RPC_OK);
}


int
vertBox(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    octObject facet, layer, path;
    struct octPoint points[2];
    

    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");

    layer.type = OCT_LAYER;
    layer.contents.layer.name = "pageFrame";
    OCTASSERT(octGetOrCreate(&facet, &layer), "can not find the pageFrame layer");
    
    path.type = OCT_PATH;
    path.contents.path.width = 0;
    OCTASSERT(octCreate(&layer, &path), "can not create the path");

    points[0].x = OFFSET;
    points[0].y = OFFSET;
    points[1].x = (octCoord) (PAGEWIDTH * UNITSPERINCH)+ OFFSET;
    points[1].y = points[0].y;

    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");

    OCTASSERT(octCreate(&layer, &path), "can not create the path");

    points[0].x = OFFSET;
    points[0].y = (octCoord) (PAGEHEIGHT * UNITSPERINCH) + OFFSET;
    points[1].x = (octCoord) (PAGEWIDTH * UNITSPERINCH) + OFFSET;
    points[1].y = points[0].y;

    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");
    
    OCTASSERT(octCreate(&layer, &path), "can not create the path");
    
    points[0].x = OFFSET;
    points[0].y = OFFSET;
    points[1].x = points[0].x;
    points[1].y = (octCoord) (PAGEHEIGHT * UNITSPERINCH) + OFFSET;
    
    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");

    OCTASSERT(octCreate(&layer, &path), "can not create the path");
    
    points[0].x = (octCoord) (PAGEWIDTH * UNITSPERINCH) + OFFSET;
    points[0].y = OFFSET;
    points[1].x = points[0].x;
    points[1].y = (octCoord) (PAGEHEIGHT * UNITSPERINCH) + OFFSET;
    
    OCTASSERT(octPutPoints(&path, 2, points), "can not put points");
    
    (void) vemCommand("show-all", spot, cmdList, userOptionWord);

    VEMMSG("Finished\n");
    vemPrompt();

    return(RPC_OK);
}


#define NOT_ZERO(val) ((val) != 0 ? (val) : 1)

#define ASPECT_RATIO(b) \
(double) (b.upperRight.x-b.lowerLeft.x)/NOT_ZERO(b.upperRight.y-b.lowerLeft.y)

#define DO_SCALE(fill, bb, field)\
 ((double) (fill.upperRight.field - fill.lowerLeft.field)/\
  NOT_ZERO(bb.upperRight.field - bb.lowerLeft.field))


/*ARGSUSED*/
int
placeInstance(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    octObject master, facet, instance, inst;
    struct octTransform *trans;
    struct RPCArg *arg1, *arg2;
    Pointer pointer;
    struct octBox fill, bb;
    double scale;
    char *cell, *view, *ptr;
    int len;
    

    len = lsLength(cmdList);
    
    if ((len < 1) || (len > 2)) {
	VEMMSG("box [\"cell:view\"] : place-instance\n");
	vemPrompt();
	return RPC_OK;
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg1 = (RPCArg *) pointer;
    if (arg1->argType != VEM_BOX_ARG) {
	VEMMSG("box [\"cell:view\"] : place-instance\n");
	vemPrompt();
	return RPC_OK;
    }

    if (len > 1) {
	(void) lsDelBegin(cmdList, &pointer);
	arg2 = (RPCArg *) pointer;
	if (arg2->argType != VEM_TEXT_ARG) {
	    VEMMSG("box [\"cell:view\"] : place-instance\n");
	    vemPrompt();
	    return RPC_OK;
	}

	if ((ptr = strchr(arg2->argData.string, ':')) == NULL) {
	    VEMMSG("box [\"cell:view\"] : place-instance\n");
	    vemPrompt();
	    return RPC_OK;
	}

	*ptr = '\0';
	cell = util_strsav(arg2->argData.string);
	view = util_strsav(++ptr);
	
    } else {
	
	if (vuFindSpot(spot, &instance, OCT_INSTANCE_MASK) != VEM_OK) {
	    VEMMSG("nothing under the spot\n");
	    return(RPC_OK);
	}
	
	cell = instance.contents.instance.master;
	view = instance.contents.instance.view;
    }
    
    facet.objectId = arg1->argData.points[0].facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");

    fill.lowerLeft.x = SMMIN(arg1->argData.points[0].thePoint.x,
			     arg1->argData.points[1].thePoint.x);
    fill.lowerLeft.y = SMMIN(arg1->argData.points[0].thePoint.y,
			     arg1->argData.points[1].thePoint.y);
    fill.upperRight.x = SMMAX(arg1->argData.points[0].thePoint.x,
			      arg1->argData.points[1].thePoint.x);
    fill.upperRight.y = SMMAX(arg1->argData.points[0].thePoint.y,
			      arg1->argData.points[1].thePoint.y);


    inst.type = OCT_INSTANCE;
    inst.contents.instance.name = "";
    inst.contents.instance.master = cell;
    inst.contents.instance.view = view;
    inst.contents.instance.facet = "contents";
    inst.contents.instance.version = OCT_CURRENT_VERSION;
    trans = &(inst.contents.instance.transform);
    
    master.type = OCT_FACET;
    master.contents.facet.cell = cell;
    master.contents.facet.view = view;
    master.contents.facet.facet = "contents";
    master.contents.facet.version = OCT_CURRENT_VERSION;
    master.contents.facet.mode = "r";
		    
    if (octOpenFacet(&master) < OCT_OK) {
	VEMMSG("Can not open facet\n");
	vemPrompt();
	return RPC_OK;
    }
    
    if (octBB(&master, &bb) < OCT_OK) {
	VEMMSG("Can not find bounding box for instance\n");
	return RPC_OK;
    }
    
    if (ASPECT_RATIO(bb) > ASPECT_RATIO(fill)) {
	scale = DO_SCALE(fill,bb,x);
    } else {
	scale = DO_SCALE(fill,bb,y);
    }
    
    /*
     * place the center of the instance at the center of the
     * the box
     */
    trans->translation.x = (fill.upperRight.x + fill.lowerLeft.x)/2 -
      scale*(bb.upperRight.x + bb.lowerLeft.x)/2;
    trans->translation.y = (fill.upperRight.y + fill.lowerLeft.y)/2 -
      scale*(bb.upperRight.y + bb.lowerLeft.y)/2;
    trans->transformType = OCT_FULL_TRANSFORM;
    trans->generalTransform[0][0] = scale;
    trans->generalTransform[0][1] = 0;
    trans->generalTransform[1][0] = 0;
    trans->generalTransform[1][1] = scale;

    if (octCreate(&facet, &inst) < OCT_OK) {
	(void) octError("can not create the instance\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    if (octBB(&inst, &bb) < OCT_OK) {
	(void) octError("can not get the bounding box");
	vemPrompt();
	return(RPC_OK);
    }
    
    WNQREGION(&facet, &bb);
    (void) vemWnFlush();

    VEMMSG("Finished\n");
    vemPrompt();
    return RPC_OK;
}

