#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "smfont.h"
#include "rpc.h"
#include "st.h"
#include "tap.h"
#include "oh.h"
#include "sm.h"
  
static dmTextItem features[] = {{ "TEXT", 5, 60, 0, 0 },
				{ "FONT", 1, 20, 0, 0 },
				{ "COLOR", 1, 20, 0, 0 },
				{ "SCALE", 1, 20, 0, 0 },
				{ "JUST", 1, 20, 0, 0 } } ;

extern st_table *fonts;


/*ARGSUSED*/
int
chlayer(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject facet, instance, prop, bag;
    octGenerator gen;
    char layerName[32], just[32];
    int len;
    
    len = lsLength(cmdList);

    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");
    
    if (len != 2) {
	VEMMSG("format: objects \"color\" : change-color\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects \"color\" : change-color\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg2 = (RPCArg *) pointer;
    if (arg2->argType != VEM_TEXT_ARG) {
	VEMMSG("format: objects \"color\" : change-color\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) strcpy(layerName, arg2->argData.string);

    if (tapIsLayerDefined(&facet, layerName) != OCT_OK) {
	VEMMSG("Bad Color\n");
	vemPrompt();
	return(RPC_OK);
    }

    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");

    OCTASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");

    while (octGenerate(&gen, &instance) == OCT_OK) {
	
	(void) GetStringProperty(&instance, "JUSTIFICATION", just, "left");
	
	/*
	 * point should always be lower left, so back out of any justification
	 */
	instance.contents.instance.transform.translation.x
		+= justOffset(just, &instance);
	
	OCTASSERT(ohCreateOrModifyPropStr(&instance, &prop, "LAYER", layerName), "Can't change Layer type");
	(void) rerecreate(&facet, &instance, spot);
    }
    VEMMSG("Finally Finished\n");
    vemPrompt();
    return(RPC_OK);
}


/*ARGSUSED*/
int
chscale(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject facet, instance, prop, bag;
    octGenerator gen;
    double scale;
    char just[32];
    int len;
    
    len = lsLength(cmdList);

    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");
    
    if (len != 2) {
	VEMMSG("format: objects \"scale\" : change-scale\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects \"scale\" : change-scale\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg2 = (RPCArg *) pointer;
    if (arg2->argType != VEM_TEXT_ARG) {
	VEMMSG("format: objects \"scale\" : change-scale\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) sscanf(arg2->argData.string, "%lf", &scale);
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");

    while (octGenerate(&gen, &instance) == OCT_OK) {

	(void) GetStringProperty(&instance, "JUSTIFICATION", just, "left");
	
	/*
	 * point should always be lower left, so back out of any justification
	 */
	instance.contents.instance.transform.translation.x
		+= justOffset(just, &instance);
	
	/* get the properties */
	OCTASSERT(ohCreateOrModifyPropReal(&instance, &prop, "SCALE", scale), "Can't change scale");
	(void) rerecreate(&facet, &instance, spot);
    }
    VEMMSG("Finally Finished\n");
    vemPrompt();
    return(RPC_OK);
}


/*ARGSUSED*/
int
chjust(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject facet, instance, prop, bag;
    octGenerator gen;
    char just[32];
    int len;
    
    len = lsLength(cmdList);

    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");
    
    if (len != 2) {
	VEMMSG("format: objects \"justification\" : change-just\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects \"justification\" : change-just\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg2 = (RPCArg *) pointer;
    if (arg2->argType != VEM_TEXT_ARG) {
	VEMMSG("format: objects \"justification\" : change-just\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (justValid(arg2->argData.string) == 0) {
	vemPrompt();
	return(RPC_OK);
    }
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");

    while (octGenerate(&gen, &instance) == OCT_OK) {

	(void) GetStringProperty(&instance, "JUSTIFICATION", just, "left");
	
	/*
	 * point should always be lower left, so back out of any justification
	 */
	instance.contents.instance.transform.translation.x
		+= justOffset(just, &instance);
	
	OCTASSERT(ohCreateOrModifyPropStr(&instance, &prop, "JUSTIFICATION", arg2->argData.string), "Can't change the justification");
	(void) rerecreate(&facet, &instance, spot);
    }
    VEMMSG("Finally Finished\n");
    vemPrompt();
    return(RPC_OK);
}


/*ARGSUSED*/
int
changeString(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg;
    Pointer pointer;
    octObject facet, textprop, fontprop, layerprop, scaleprop, justprop;
    octObject instance, bag;
    octGenerator gen;
    char scalebuf[32];
    char buffer[1024];
    int len;
    int i,j;
    
    
    len = lsLength(cmdList);

    facet.type = OCT_FACET;
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "can not get the facet");

    if (len != 1) {
	VEMMSG("format: objects : change-string\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects : change-string\n");
	vemPrompt();
	return(RPC_OK);
    }

    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");

    while (octGenerate(&gen, &instance) == OCT_OK) {

	(void) GetStringProperty(&instance, "TYPE", buffer, "empty");

	if (!STREQ(buffer, "TEXT")) {
	    VEMMSG("Not a Text String\n");
	    continue;
	}
	
	(void) GetStringProperty(&instance, "TEXT", buffer, "SlideMaker");
	for (i = 1; i < (strlen(buffer) / 45) + 1; i++) {
	    for (j = strlen(buffer)+1; j >= i * 45; j--) {
		buffer[j+1] = buffer[j];
	    }
	    buffer[i * 45] = '\n';
	}
	
	features[0].value = util_strsav(buffer);

	(void) GetStringProperty(&instance, "FONT", buffer, "romant");
	features[1].value = util_strsav(buffer);

	(void) GetStringProperty(&instance, "LAYER", buffer, "cyan");
	features[2].value = util_strsav(buffer);

	(void) GetStringProperty(&instance, "JUSTIFICATION", buffer, "left");
	features[4].value = util_strsav(buffer);

	/*
	 * point should always be lower left, so back out of any justification
	 */
	instance.contents.instance.transform.translation.x
		+= justOffset(features[4].value, &instance);

	OCTASSERT(ohGetByPropName(&instance, &scaleprop, "SCALE"), "can not get scale prop");
	(void) sprintf(scalebuf, "%.2lf", scaleprop.contents.prop.value.real);
	features[3].value = scalebuf;
	
	if (dmMultiText("String Change Menu", sizeof(features) / sizeof(dmTextItem), features) == VEM_NOSELECT) {
	    VEMMSG("Aborting current string, continuing on...\n");
	    continue;
	}

	/* We have all the new properties.  Let's push them now. */
	textprop.type = OCT_PROP;
	textprop.contents.prop.name = "TEXT";
	textprop.contents.prop.type = OCT_STRING;
	stripNewLines(features[0].value);
	textprop.contents.prop.value.string = features[0].value;

	fontprop.type = OCT_PROP;
	fontprop.contents.prop.name = "FONT";
	fontprop.contents.prop.type = OCT_STRING;
	fontprop.contents.prop.value.string = features[1].value;

	layerprop.type = OCT_PROP;
	layerprop.contents.prop.name = "LAYER";
	layerprop.contents.prop.type = OCT_STRING;
	layerprop.contents.prop.value.string = features[2].value;

	justprop.type = OCT_PROP;
	justprop.contents.prop.name = "JUSTIFICATION";
	justprop.contents.prop.type = OCT_STRING;
	justprop.contents.prop.value.string = features[4].value;

	scaleprop.type = OCT_PROP;
	scaleprop.contents.prop.name = "SCALE";
	scaleprop.contents.prop.type = OCT_REAL;
	(void) sscanf(features[3].value, "%lf",
	    &scaleprop.contents.prop.value.real);

	if (!st_lookup(fonts, features[1].value, (char **) 0)) {
	    VEMMSG("No such font type, setting to `romant'\n");
	    (void) strcpy(features[1].value, "romant");
	}

	if (tapIsLayerDefined(&facet, features[2].value) != OCT_OK) {
	    VEMMSG("Bad Color\n");
	    vemPrompt();
	    return(RPC_OK);
	}
	
	if (justValid(justprop.contents.prop.value.string) == 0) {
	    vemPrompt();
	    return(RPC_OK);
	}

	if (scaleprop.contents.prop.value.real <= 0.0) {
	    VEMMSG("Illegal scale value\n");
	    vemPrompt();
	    return(RPC_OK);
	}

	/* verified all of the entries, now update */
	OCTASSERT(octCreateOrModify(&instance, &fontprop), "Can't change font");
	OCTASSERT(octCreateOrModify(&instance, &layerprop), "Can't change Layer type");
	OCTASSERT(octCreateOrModify(&instance, &justprop), "Can't change the justification");
	OCTASSERT(octCreateOrModify(&instance, &textprop), "Can't change Text");
	OCTASSERT(octCreateOrModify(&instance, &scaleprop), "Can't change Scale type");
	
	(void) rerecreate(&facet, &instance, spot);
    }
    VEMMSG("Finally Finished\n");
    vemPrompt();
    return(RPC_OK);
}


int
rerecreate(facet, instance, spot)
octObject *facet, *instance;
RPCSpot *spot;
{
    octObject prop;
    struct octPoint point;
    struct octBox bb;
    char string[1024], fontName[32], layerName[32], justification[32];
    char type[20];
    double scale;
    int noBoundingBox;
    int stalk, head;
    double stalkAngle, headAngle;
    

    point = instance->contents.instance.transform.translation;

    /* get the properties */
    (void) GetStringProperty(instance, "TYPE", type, "NOTYPE");

    if STREQ(type, "NOTYPE") {
	VEMMSG("No Type property for the instance, ignoring....\n");
	return(RPC_OK);
    }

    if STREQ(type, "GROUP") {

	VEMMSG("Groups are not handled yet, ignoring....\n");
	return(RPC_OK);

    } else if STREQ(type, "TEXT") {

	(void) GetStringProperty(instance, "TEXT", string, "SlideMaker");

	(void) GetStringProperty(instance, "FONT", fontName, "romant");
	
	if (!st_lookup(fonts, fontName, (char **) 0)) {
	    VEMMSG("bad font, setting to `romant'\n");
	    (void) strcpy(fontName, "romant");
	}

	(void) GetStringProperty(instance, "LAYER", layerName, "cyan");
	
	if (ohGetByPropName(instance, &prop, "SCALE") < OCT_OK) {
	    scale = 1.0;
	} else {
	    if (prop.contents.prop.type == OCT_REAL) {
		scale = prop.contents.prop.value.real;
	    } else {
		scale = (double) prop.contents.prop.value.integer;
	    }		    
	}

	(void) GetStringProperty(instance, "JUSTIFICATION", justification, "left");

	point = instance->contents.instance.transform.translation;

	noBoundingBox = 0;
	if (octBB(instance, &bb) < OCT_OK) {
	    noBoundingBox = 1;
	}

	if (octDelete(instance) < OCT_OK) {
	    (void) octError("can not delete");
	    return(RPC_OK);
	}

	if (noBoundingBox == 0) {
	    WNQREGION(facet, &bb);
	}
	
	if (createString(facet, point, string, fontName, layerName, scale, justification) != RPC_OK) {
	    return(RPC_OK);
	}

    } else if STREQ(type, "BOX") {

	(void) GetStringProperty(instance, "JUSTIFICATION", justification, "left");
	(void) GetStringProperty(instance, "LAYER", layerName, "cyan");
	
	noBoundingBox = 0;
	if (octBB(instance, &bb) < OCT_OK) {
	    noBoundingBox = 1;
	}

	OCTASSERT(octDelete(instance), "can not delete");

	if (noBoundingBox == 0) {
	    WNQREGION(facet, &bb);
	}	    
	
	if (createBox(spot, facet, &bb, layerName, justification) != RPC_OK) {
	    return(RPC_OK);
	}

    } else if STREQ(type, "ARROW") {

	(void) GetStringProperty(instance, "LAYER", layerName, "cyan");
	(void) GetIntegerProperty(instance, "STALK-LENGTH", &stalk, 20);
	(void) GetIntegerProperty(instance, "HEAD-LENGTH", &head, 20);
	(void) GetRealProperty(instance, "STALK-ANGLE", &stalkAngle, 0.0);
	(void) GetRealProperty(instance, "HEAD-ANGLE", &headAngle, 0.0);
	
	noBoundingBox = 0;
	if (octBB(instance, &bb) < OCT_OK) {
	    noBoundingBox = 1;
	}

	OCTASSERT(octDelete(instance), "can not delete");

	if (noBoundingBox == 0) {
	    WNQREGION(facet, &bb);
	}
	
	if (createArrow(spot, facet, point, stalk, head, stalkAngle, headAngle,
			layerName) != RPC_OK) {
	    return(RPC_OK);
	}
	
    } else {
	
	VEMMSG("bad TYPE property, skipping\n");
    }
    return(RPC_OK);
}

