#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "rpc.h"
#include "oh.h"
#include "sm.h"


/*ARGSUSED*/
int
group(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg;
    Pointer pointer;
    octObject bag, facet, object, newfacet, instance, prop;
    octGenerator gen;
    struct octBox bb;
    char newfacetName[1024];
    long loc;
    int x, y;
    octId oldId;
    
    if (lsLength(cmdList) != 1) {
	VEMMSG("format: objects : group\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects : group\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "failed to get the facet");
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    /* figure out the lower left */
    x = MAXINTEGER;
    y = MAXINTEGER;
    
    OCTASSERT(octInitGenContentsSpecial(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    while (octGenerate(&gen, &object) == OCT_OK) {
	x = SMMIN(x, object.contents.instance.transform.translation.x);
	y = SMMIN(y, object.contents.instance.transform.translation.y);
    }

    /* copy and move the items */
    (void) time(&loc);

    (void) sprintf(newfacetName, "%s/group-%d", facet.contents.facet.cell, loc);
    newfacet.type = OCT_FACET;
    newfacet.contents.facet.cell = newfacetName;
    newfacet.contents.facet.view = "slide";
    newfacet.contents.facet.facet = "contents";
    newfacet.contents.facet.version = OCT_CURRENT_VERSION;
    newfacet.contents.facet.mode = "w";

    OCTASSERT(octOpenFacet(&newfacet), "can not open the facet");

    OCTASSERT(octInitGenContentsSpecial(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    while (octGenerate(&gen, &object) == OCT_OK) {
	oldId = object.objectId;
	object.contents.instance.transform.translation.x -= x;
	object.contents.instance.transform.translation.y -= y;
	OCTASSERT(octCreate(&newfacet, &object), "create failed");
	
	object.objectId = oldId;
	OCTASSERT(octDelete(&object), "delete failed");
    }

    if (addPolicyProperties(&newfacet) < OCT_OK) {
	VEMMSG("can not create the policy properties\n");
	return(RPC_OK);
    }
        
    if (createInterface(&newfacet, "MET2", "group") != OCT_OK) {
	vemPrompt();
	return(RPC_OK);
    }

    OCTASSERT(octFlushFacet(&newfacet), "flush failed");

    /* place the instance */
    instance.type = OCT_INSTANCE;
    instance.contents.instance.name = "string";
    instance.contents.instance.master = celltail(newfacetName);
    instance.contents.instance.view = "slide";
    instance.contents.instance.facet = "contents";
    instance.contents.instance.version = OCT_CURRENT_VERSION;

    instance.contents.instance.transform.translation.x = x;
    instance.contents.instance.transform.translation.y = y;
    instance.contents.instance.transform.transformType = OCT_NO_TRANSFORM;

    OCTASSERT(octCreate(&facet, &instance), "can not create the instance");

    if (ohCreatePropStr(&instance, &prop, "TYPE", "GROUP") < OCT_OK) {
	(void) octError("can not create the TYPE");
	vemPrompt();
	return(RPC_OK);
    }

    if (ohCreatePropStr(&instance, &prop, "JUSTIFICATION", "left") < OCT_OK) {
	(void) octError("can not create the JUSTIFICATION");
	vemPrompt();
	return(RPC_OK);
    }
    
    OCTASSERT(octBB(&instance, &bb), "can not get the bounding box");
    
    WNQREGION(&facet, &bb);
    
    (void) vemWnFlush();
    
    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
}


/*ARGSUSED*/
int
ungroup(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg;
    Pointer pointer;
    octObject bag, facet, newobject, master, instance, prop;
    struct octBox bb;
    octGenerator gen, instgen;
    int x, y;
    
    if (lsLength(cmdList) != 1) {
	VEMMSG("format: object : ungroup\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "failed to get the facet");
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects : group\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContents(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    while (octGenerate(&gen, &instance) == OCT_OK) {

	x = instance.contents.instance.transform.translation.x;
	y = instance.contents.instance.transform.translation.y;

	/* verify that this is a group */
	OCTASSERT(ohGetByPropName(&instance, &prop, "TYPE"), "no TYPE property");
	
	if (!STREQ(prop.contents.prop.value.string, "GROUP")) {
	    VEMMSG("not a GROUP\n");
	    vemPrompt();
	    return(RPC_OK);
	}
	if (openMaster(&instance, &master, "r") < OCT_OK) {
	    (void) octError("can not open the master");
	    vemPrompt();
	    return(RPC_OK);
	}
	
	/* copy the instances */
	OCTASSERT(octInitGenContents(&master, OCT_INSTANCE_MASK, &instgen), "init gen failed");
	
	while (octGenerate(&instgen, &newobject) == OCT_OK) {
	    newobject.contents.instance.transform.translation.x += x;
	    newobject.contents.instance.transform.translation.y += y;
	    OCTASSERT(octCreate(&facet, &newobject), "create failed");
	}
	
	OCTASSERT(octBB(&instance, &bb), "can not get the bounding box");
	
	OCTASSERT(octDelete(&instance), "can not delete the GROUP container");
	
	WNQREGION(&facet, &bb);
    }
    
    (void) vemWnFlush();
        
    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
}

