#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "rpc.h"
#include "oh.h"
#include "sm.h"


/*ARGSUSED*/
int
lineUpX(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject bag, facet, object, justprop;
    struct octBox bb;
    octGenerator gen;
    int x;
    int len;

    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "failed to get the facet");
    
    len = lsLength(cmdList);
    
    if ((len > 2) || (len < 1)) {
	VEMMSG("format: objects [point] : line-up-x\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects [point] : line-up-x\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (len == 2) {
	(void) lsDelBegin(cmdList, &pointer);
	arg2 = (RPCArg *) pointer;
	if (arg2->argType != VEM_POINT_ARG) {
	    VEMMSG("format: objects [point] : line-up-x\n");
	    vemPrompt();
	    return(RPC_OK);
	}
	x = arg2->argData.points[0].thePoint.x;
    } else {
	/* point is center of bb of the facet */
	OCTASSERT(octBB(&facet, &bb), "failed to get the bb of the facet\n");
	x = (bb.lowerLeft.x + bb.upperRight.x) / 2;
    }

    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContentsSpecial(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    while (octGenerate(&gen, &object) == OCT_OK) {
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");

	OCTASSERT(ohGetOrCreatePropStr(&object, &justprop, "JUSTIFICATION", "left"), "can not create or modify the master");
	
	object.contents.instance.transform.translation.x
	  = x - justOffset(justprop.contents.prop.value.string, &object);
	
	OCTASSERT(octModify(&object), "modify failed");
	

	WNQREGION(&facet, &bb);
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	WNQREGION(&facet, &bb);
    }

    
    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
}


/*ARGSUSED*/
int
lineUpY(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject bag, facet, object;
    struct octBox bb;
    octGenerator gen;
    int y;
    
    if (lsLength(cmdList) != 2) {
	VEMMSG("format: objects point : line-up-y\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects point : line-up-y\n");
	vemPrompt();
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg2 = (RPCArg *) pointer;
    if (arg2->argType != VEM_POINT_ARG) {
	VEMMSG("format: objects point : line-up-y\n");
	vemPrompt();
	return(RPC_OK);
    }

    y = arg2->argData.points[0].thePoint.y;

    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "failed to get the facet");
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContentsSpecial(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    while (octGenerate(&gen, &object) == OCT_OK) {
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	object.contents.instance.transform.translation.y = y;
	OCTASSERT(octModify(&object), "modify failed");
	
	WNQREGION(&facet, &bb);
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	WNQREGION(&facet, &bb);
	
    }

    (void) vemWnFlush();

    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
}


/*ARGSUSED*/
int
spaceY(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject bag, facet, object;
    struct octBox bb;
    octGenerator gen;
    int y, topy, deltay;
    
    if (lsLength(cmdList) != 2) {
	VEMMSG("format: objects point point : space-y\n");
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects point point : space-y\n");
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg2 = (RPCArg *) pointer;
    if (arg2->argType != VEM_POINT_ARG) {
	VEMMSG("format: objects point point : space-y\n");
	return(RPC_OK);
    }

    if (arg2->length != 2) {
	VEMMSG("format: objects point point : space-y\n");
	return(RPC_OK);
    }

    topy = arg2->argData.points[0].thePoint.y;

    deltay = ABS(arg2->argData.points[1].thePoint.y
		 - arg2->argData.points[0].thePoint.y);
    
    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "failed to get the facet");
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");
    
    OCTASSERT(octInitGenContentsSpecial(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    y = topy;
    
    while (octGenerate(&gen, &object) == OCT_OK) {
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	object.contents.instance.transform.translation.y = y;
	OCTASSERT(octModify(&object), "modify failed");
	y -= deltay;
	
	WNQREGION(&facet, &bb);
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	WNQREGION(&facet, &bb);
    }

    (void) vemWnFlush();

    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
}


/*ARGSUSED*/
int
centerY(spot, cmdList, userOptionWord)
RPCSpot *spot;
lsList cmdList;
long userOptionWord;
{
    RPCArg *arg, *arg2;
    Pointer pointer;
    octObject bag, facet, object;
    octGenerator gen;
    struct octBox bb;
    int y;
    int ysize;
    
    if (lsLength(cmdList) != 2) {
	VEMMSG("format: objects point : center-y\n");
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_OBJ_ARG) {
	VEMMSG("format: objects point : center-y\n");
	return(RPC_OK);
    }
    
    (void) lsDelBegin(cmdList, &pointer);
    arg2 = (RPCArg *) pointer;
    if (arg2->argType != VEM_POINT_ARG) {
	VEMMSG("format: objects point : center-y\n");
	return(RPC_OK);
    }

    y = arg2->argData.points[0].thePoint.y;

    facet.objectId = spot->facet;
    OCTASSERT(octGetById(&facet), "failed to get the facet");
    
    bag.objectId = arg->argData.objArg.theBag;
    OCTASSERT(octGetById(&bag), "failed to get the object bag");

    OCTASSERT(octInitGenContentsSpecial(&bag, OCT_INSTANCE_MASK, &gen), "init gen failed");
    
    while (octGenerate(&gen, &object) == OCT_OK) {
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	ysize = ABS(bb.lowerLeft.y - bb.upperRight.y);
	object.contents.instance.transform.translation.y = y - ysize / 2;
	OCTASSERT(octModify(&object), "modify failed");
	
	WNQREGION(&facet, &bb);
	
	OCTASSERT(octBB(&object, &bb), "can not get the bounding box");
	
	WNQREGION(&facet, &bb);
    }

    (void) vemWnFlush();

    VEMMSG("Finished\n");
    vemPrompt();
    
    return(RPC_OK);
}
