#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <math.h>
#include "smfont.h"
#include "rpc.h"
#include "oh.h"
#include "st.h"
#include "sm.h"
#include <sys/types.h>
#include <sys/stat.h>

extern font *fontptr[MAXGLYPH];
extern st_table *fonts;

#define SM_OKAY 1
#define SM_ERROR 0
  
/* state is global, stack is global */
struct curstate {
    char *font;
    char *color;
    double scale;
    double base;
    int intercharacterspace;
    int capitalXheight;
    int capitalXwidth;
    int spacesize;
};

struct curstate curstate;

struct curstate statestack[100];
int stackptr = -1;

int Translation;

extern char *strchr();


void sm_create_dir( facet )
    octObject *facet;
{
    char *dirName = facet->contents.facet.cell;
    struct stat buf;
    if ( stat( dirName, &buf  ) < 0 ) {
	/* fprintf( stderr, "Creating dir %s\n", dirName ); */
	mkdir( dirName, 0777 );
    }
}


void
pushstate()
{
    statestack[++stackptr] = curstate;
    return;
}


void
popstate()
{
    curstate = statestack[stackptr--];
    return;
}


static int
newfont(ptr)
char **ptr;
/* starting at [ */
{
    char fontname[20];
    char *loc;

    (*ptr)++;				/* skip [		*/
    if ((loc = strchr(*ptr, ']')) == NIL(char)) {	/* find ]		*/
	VEMMSG("Missing ']' in font change request\n");
	return(SM_ERROR);
    }
    (void) strncpy(fontname, *ptr, loc - *ptr);	/* get font		*/
    fontname[loc - *ptr] = '\0';	/* make it a string	*/
    curstate.font = util_strsav(fontname);
    *ptr = loc + 1;			/* set pointer past ] */

    return(SM_OKAY);
}


static int
newcolor(ptr)
char **ptr;
/* starting at [ */
{
    char color[20];
    char *loc;

    (*ptr)++;				/* skip [		*/
    if ((loc = strchr(*ptr, ']')) == NIL(char)) {	/* find ]		*/
	VEMMSG("Missing ']' in color change request\n");
	return(SM_ERROR);
    }
    (void) strncpy(color, *ptr, loc - *ptr);	/* get color		*/
    color[loc - *ptr] = '\0';		/* make it a string	*/
    *ptr = loc + 1;			/* set pointer past ] */
    curstate.color = util_strsav(color);

    return(SM_OKAY);
}


static int
newscale(ptr)
char **ptr;
/* starting at [ */
{
    char *loc;
    double scale;

    (*ptr)++;				/* skip [		*/
    (void) sscanf(*ptr, "%lf", &scale);	/* get scale		*/
    if ((loc = strchr(*ptr, ']')) == NIL(char)) {	/* find ]		*/
	VEMMSG("Missing ']' in scale change request\n");
	return(SM_ERROR);
    }
    *ptr = loc + 1;			/* set pointer past ] */
    curstate.scale *= scale;

    return(SM_OKAY);
}


static int
newbase(ptr)
char **ptr;
/* starting at [ */
{
    char *loc;
    double base;

    (*ptr)++;				/* skip [		*/
    (void) sscanf(*ptr, "%lf", &base);		/* get base		*/
    if ((loc = strchr(*ptr, ']')) == NIL(char)) {	/* find ]		*/
	VEMMSG("Missing ']' in baseline change request\n");
	return(SM_ERROR);
    }
    *ptr = loc + 1;			/* set pointer past ] */
    curstate.base += base;

    return(SM_OKAY);
}


static int
newmove(ptr)
char **ptr;
/* starting at [ */
{
    char *loc;
    double move;

    (*ptr)++;				/* skip [		*/
    (void) sscanf(*ptr, "%lf", &move);		/* get move		*/
    if ((loc = strchr(*ptr, ']')) == NIL(char)) {	/* find ]		*/
	VEMMSG("Missing ']' in move request\n");
	return(SM_ERROR);
    }
    *ptr = loc + 1;			/* set pointer past ] */
    Translation += ((int) (curstate.scale
			     * (curstate.capitalXwidth * move)));

    return(SM_OKAY);
}


static void
forceDir(name)
    char* name;
{
    char buffer[1024], *end;
    octObject facet;

    (void) strcpy(buffer, name);
    end = strrchr(buffer, '/');
    end = strrchr(--end, '/');
    *end = '\0';

    if (ohOpenFacet(&facet, buffer, "force", "contents", "a") < OCT_OK) {
	octError("forcing font directory");
	return;
    }
    (void) octFlushFacet(&facet);

    *end = '/';
    end = strrchr(buffer, '/');
    *end = '\0';

    if (ohOpenFacet(&facet, buffer, "force", "contents", "a") < OCT_OK) {
	octError("forcing font directory");
	return;
    }
    (void) octFlushFacet(&facet);
    return;
}



static int
createLetter(character, newfacet)
char character;
octObject *newfacet;
{
    int glyph, lettersize;
    char buffer[1024];
    octObject letterfacet, instance;
    char letterName[1024];

    if (character == ' ') {
	Translation += (int) (curstate.scale * curstate.spacesize);
	return(SM_OKAY);
    }

    if (character != BULLET) {
	if ((glyph = findGlyph(character, curstate.font)) == -1) {
	    (void) sprintf(buffer, "Can not find information for %c in font %s\n",
			   character, curstate.font);
	    VEMMSG(buffer);
	    return(SM_ERROR);;
	}

	SIZEOFLETTER(character, curstate.font, lettersize);
	lettersize *= SCALE;

    } else {
	/* handle bullet */
	lettersize = (int) (curstate.capitalXwidth * BULLETSCALE);
    }
    
    Translation += ((int) (curstate.scale
			   * (lettersize + curstate.intercharacterspace))) / 2;

    if (character == BULLET) {
	(void) sprintf(letterName, "%s/%s/%s/%s/bullet-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, MAJORVERSION);
    } else if (character == '/') {
	(void) sprintf(letterName, "%s/%s/%s/%s/slash-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, MAJORVERSION);
    } else if (character == '.') {
	(void) sprintf(letterName, "%s/%s/%s/%s/period-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, MAJORVERSION);
    } else if (character == ':') {
	(void) sprintf(letterName, "%s/%s/%s/%s/colon-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, MAJORVERSION);
    } else if (character == ';') {
	(void) sprintf(letterName, "%s/%s/%s/%s/semicolon-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, MAJORVERSION);
    } else if (character == '\\') {
	(void) sprintf(letterName, "%s/%s/%s/%s/backslash-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, MAJORVERSION);
    } else {	    
	(void) sprintf(letterName, "%s/%s/%s/%s/%c-%d", "$OCTTOOLS", LETTERDIR,
		       curstate.font, curstate.color, character, MAJORVERSION);
    }

    /* forceDir(letterName); */

    /* check for existance of of the letter */

    letterfacet.type = OCT_FACET;
    letterfacet.contents.facet.cell = letterName;
    letterfacet.contents.facet.view = "slide";
    letterfacet.contents.facet.facet = "contents";
    letterfacet.contents.facet.version = OCT_CURRENT_VERSION;
    letterfacet.contents.facet.mode = "r";

	
    if (octOpenFacet(&letterfacet) < OCT_OK) {

	if (GenLetters) {
	    if (character != BULLET) {
	    
		/* generate the letter */

		genletter(glyph, letterName, "slide", SCALE, curstate.color, SCALE * WIDTH);
	    } else {
		/* build the bullet */
		octObject circle, layer;

		layer.type = OCT_LAYER;
		layer.contents.layer.name = curstate.color;
		
		circle.type = OCT_CIRCLE;
		circle.contents.circle.startingAngle = 0;
		circle.contents.circle.endingAngle = 3600;
		circle.contents.circle.innerRadius = 0;
		circle.contents.circle.outerRadius = lettersize;
		circle.contents.circle.center.x = 0;
		circle.contents.circle.center.y = 0;

		letterfacet.type = OCT_FACET;
		letterfacet.contents.facet.cell = letterName;
		letterfacet.contents.facet.view = "slide";
		letterfacet.contents.facet.facet = "contents";
		letterfacet.contents.facet.version = OCT_CURRENT_VERSION;
		letterfacet.contents.facet.mode = "w";
	    
		if (octOpenFacet(&letterfacet) < OCT_OK) {
		    (void) octError("can not open the letter facet - no glyph\n");
		    return(SM_ERROR);
		}

		if (addPolicyProperties(&letterfacet) < OCT_OK) {
		    VEMMSG("can not create the policy properties\n");
		    return(SM_ERROR);
		}

		if (octCreate(&letterfacet, &layer) < OCT_OK) {
		    (void) octError("can not create the layer");
		    return(SM_ERROR);
		}

		if (octCreate(&layer, &circle) < OCT_OK) {
		    (void) octError("can not create the circle");
		    return(SM_ERROR);
		}

		(void) octCloseFacet(&letterfacet);
		letterfacet.contents.facet.mode = "r";
	    }

	    /* open the letter cell */
	    if (octOpenFacet(&letterfacet) < OCT_OK) {
		(void) octError("can not open the letter facet - no glyph\n");
		return(SM_ERROR);
	    }
	    
	    (void) sprintf(buffer, "%c", character);
	    if (createInterface(&letterfacet, curstate.color, buffer) != OCT_OK) {
		return(SM_ERROR);
	    }
	} else {
	    fprintf(stderr, "FATAL ERROR: missing letter\n");
	    exit(-1);
	}
    }
	
    /* place the instance */
    instance.type = OCT_INSTANCE;
    instance.contents.instance.name = "letter";
    instance.contents.instance.master = letterName;
    instance.contents.instance.view = "slide";
    instance.contents.instance.facet = "contents";
    instance.contents.instance.version = OCT_CURRENT_VERSION;
	
    instance.contents.instance.transform.translation.x = Translation;
    instance.contents.instance.transform.translation.y
      = curstate.base * curstate.capitalXheight;
    instance.contents.instance.transform.transformType = OCT_FULL_TRANSFORM;
    instance.contents.instance.transform.generalTransform[0][0] = curstate.scale;
    instance.contents.instance.transform.generalTransform[1][1] = curstate.scale;
    instance.contents.instance.transform.generalTransform[0][1] = 0.0;
    instance.contents.instance.transform.generalTransform[1][0] = 0.0;

    if (octCreate(newfacet, &instance) < OCT_OK) {
	(void) octError("can not create the instance\n");
	return(SM_ERROR);
    }

    Translation += ((int) (curstate.scale
			   * (lettersize + curstate.intercharacterspace))) / 2;

    return(SM_OKAY);
}    


static int
newcommand(ptr, newfacet)
char **ptr;
octObject *newfacet;
/* **ptr = '\' */
{
    if (**ptr != '\\') {
	VEMMSG("Missing backslash\n");
	return(SM_ERROR);
    }

    (*ptr)++;

    if (strncmp(*ptr, "font", 4) == 0) {
	*ptr = *ptr + 4;
	return(newfont(ptr));
    }
    if (strncmp(*ptr, "color", 5) == 0) {
	*ptr = *ptr + 5;
	return(newcolor(ptr));
    }
    if (strncmp(*ptr, "scale", 5) == 0) {
	*ptr = *ptr + 5;
	return(newscale(ptr));
    }
    if (strncmp(*ptr, "base", 4) == 0) {
	*ptr = *ptr + 4;
	return(newbase(ptr));
    }
    if (strncmp(*ptr, "move", 4) == 0) {
	*ptr = *ptr + 4;
	return(newmove(ptr));
    }
    
    if (strncmp(*ptr, "bu", 2) == 0) {
	*ptr = *ptr + 2;
	return(createLetter(BULLET, newfacet));
    }
    if (**ptr == '\\') {
	*ptr = *ptr + 1;
	return(createLetter('\\', newfacet));
    }
    if (**ptr == '{') {
	*ptr = *ptr + 1;
	return(createLetter('{', newfacet));
    }
    if (**ptr == '}') {
	*ptr = *ptr + 1;
	return(createLetter('}', newfacet));
    }

    return(SM_OKAY);
}


void
setcurstateextras()
{
    SIZEOFLETTER('m', curstate.font, curstate.intercharacterspace);
    curstate.intercharacterspace *= SCALE;
    curstate.intercharacterspace /= 6;
    
    SIZEOFLETTER('a', curstate.font, curstate.spacesize);
    curstate.spacesize *= SCALE;

    HEIGHTOFLETTER('X', curstate.font, curstate.capitalXheight);
    curstate.capitalXheight *= SCALE;
    if (curstate.capitalXheight < 0) {
	curstate.capitalXheight = -curstate.capitalXheight;
    }
    SIZEOFLETTER('X', curstate.font, curstate.capitalXwidth);
    curstate.capitalXwidth *= SCALE;
    
    return;
}


int
createString(facet, point, input, fontName, colorName, scale, justification)
octObject *facet;
struct octPoint point;
char *input;
char *fontName;
char *colorName;
double scale;
char *justification;
{
    char *string;
    time_t loc;
    char newfacetName[1024];
    octObject newfacet, instance, prop;
    struct octBox bb;
    static int counter = 0;
    octStatus status;
    
    
    /* create a new master */
    (void) time(&loc);
    
    /* simulation of oct_path = '.' */
    /* OCTASSERT(octFlushFacet(facet), "can not flush the facet"); */
    sm_create_dir( facet );
    (void) sprintf(newfacetName, "%s/string-%d-%d", facet->contents.facet.cell, loc, ++counter);


    newfacet.type = OCT_FACET;
    newfacet.contents.facet.cell = newfacetName;
    newfacet.contents.facet.view = "slide";
    newfacet.contents.facet.facet = "contents";
    newfacet.contents.facet.version = OCT_CURRENT_VERSION;
    newfacet.contents.facet.mode = "w";

    openit:
    
    /* handle the possiblity of creating two strings with the same 'ctime' */
    
    if ((status = octOpenFacet(&newfacet)) < OCT_OK) {
	
	if (status == OCT_ERROR) {
	    octError( ohFormatName( &newfacet ) );
	    (void) sprintf(newfacetName, "%s-%d", newfacetName, ++counter);
	    VEMMSG("Whoa!  Slow Down, Take it Easy!\n");
	    newfacet.contents.facet.cell = newfacetName;
	    goto openit;
	} else {
	    (void) octError("can not open the facet");
	    return(RPC_OK);
	}
    }	
    
    if (addPolicyProperties(&newfacet) < OCT_OK) {
	VEMMSG("can not create the policy properties\n");
	return(RPC_OK);
    }
    
    curstate.font = fontName;
    curstate.color = colorName;
    curstate.scale = scale;
    curstate.base = 0.0;
    setcurstateextras();

    string = input;
    Translation = 0;

    while(*string != '\0') {

	if (*string == '\\') {
	    if (newcommand(&string, &newfacet) != SM_OKAY) {
		return(RPC_OK);
	    }
	    setcurstateextras();
	} else if (*string == '{') {
	    string++;
	    pushstate();
	} else if (*string == '}') {
	    string++;
	    popstate();
	} else {
	    if (createLetter(*string, &newfacet) != SM_OKAY) {
		return(RPC_OK);
	    }
	    string++;
	}
    }

    if (createInterface(&newfacet, curstate.color, input) != OCT_OK) {
	return(RPC_OK);
    }

    if (octFlushFacet(&newfacet) < OCT_OK) {
	(void) octError("can not flush the new facet");
	return(RPC_OK);
    }
    
    /* place the instance */
    instance.type = OCT_INSTANCE;
    instance.contents.instance.name = "string";
    instance.contents.instance.master = celltail(newfacetName);
    instance.contents.instance.view = "slide";
    instance.contents.instance.facet = "contents";
    instance.contents.instance.version = OCT_CURRENT_VERSION;

    instance.contents.instance.transform.translation.x
      = point.x - justOffset(justification, &newfacet);
    instance.contents.instance.transform.translation.y = point.y;
    instance.contents.instance.transform.transformType = OCT_NO_TRANSFORM;

    OCTASSERT(octCreate(facet, &instance), "can not create the instance");

    if (ohCreatePropStr(&instance, &prop, "TEXT", input) < OCT_OK) {
	(void) octError("can not create the TEXT");
	return(RPC_OK);
    }
    
    if (ohCreatePropStr(&instance, &prop, "FONT", fontName) < OCT_OK) {
	(void) octError("can not create the FONT");
	return(RPC_OK);
    }
    
    if (ohCreatePropStr(&instance, &prop, "LAYER", colorName) < OCT_OK) {
	(void) octError("can not create the LAYER");
	return(RPC_OK);
    }

    if (ohCreatePropReal(&instance, &prop, "SCALE", scale) < OCT_OK) {
	(void) octError("can not create the SCALE");
	return(RPC_OK);
    }
    
    if (ohCreatePropStr(&instance, &prop, "JUSTIFICATION", justification) < OCT_OK) {
	(void) octError("can not create the JUSTIFICATION");
	return(RPC_OK);
    }
    
    if (ohCreatePropStr(&instance, &prop, "TYPE", "TEXT") < OCT_OK) {
	(void) octError("can not create the TYPE");
	return(RPC_OK);
    }
    
    OCTASSERT(octBB(&instance, &bb), "can not get the bounding box");
    
    WNQREGION(facet, &bb);
    (void) vemWnFlush();
    
    VEMMSG("Finished\n");
    vemPrompt();

    return(RPC_OK);
}

