#include "copyright.h"
#include "port.h"
#include "utility.h"
#include <X11/Xlib.h>
#define XId
#include "rpc.h"
#include "oh.h"
#include "sm.h"


extern char SMDisplayName[];
int initialized = 0;
FILE *fp;

lsGen slideGen;
lsList slideList;


/*ARGSUSED*/
int
setSequenceFile(spot, cmdList, uow)
RPCSpot *spot;
lsList cmdList;
long uow;
{
    RPCArg *arg;
    Pointer pointer;
    char *item;
    char buffer[1024], name[1024];
    
    
    /* get the argument */
    if (lsLength(cmdList) > 1) {
	VEMMSG("[\"file\"] : set-sequence-file\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (lsLength(cmdList) == 0) {
	if (initialized == 0) {
	    VEMMSG("a file has not been previously set\n");
	    vemPrompt();
	    return(RPC_OK);
	}
	slideGen = lsStart(slideList);
	VEMMSG("resetting to the beginning of the current sequence file\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_TEXT_ARG) {
	VEMMSG("[\"file\"] : set-sequence-file\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (initialized == 1) {
	(void) fclose(fp);
    }

    if ((fp = fopen(util_file_expand(arg->argData.string), "r")) == NULL) {
	VEMMSG("can not open the file\n");
	vemPrompt();
	return(RPC_OK);
    }

    slideList = lsCreate();

    while(fscanf(fp, "%s", name) == 1) {
	(void) sprintf(buffer, "Adding `%s' to the sequence list\n", name);
	VEMMSG(buffer);
	item = util_strsav(name);
	(void) lsNewEnd(slideList, (Pointer) item, (lsHandle *) 0);
    }

    (void) fclose(fp);
    
    slideGen = lsStart(slideList);
    
    VEMMSG("Finished\n");
    vemPrompt();

    initialized = 1;

    return(RPC_OK);
}


/*ARGSUSED*/
int
initScreen(spot, cmdList, uow)
RPCSpot *spot;
lsList cmdList;
long uow;
{
    Display *display;
    Colormap cm;
    int screen;
    Window root;
    XColor scr, exact;
    Window theWin;
    wnOpts options;
    

    if ((display = XOpenDisplay(SMDisplayName)) == NULL) {
	VEMMSG("can not open the display, will not be able to use the sequencer\n");
	return(RPC_OK);
    }

    screen = DefaultScreen(display);
    root = RootWindow(display, screen);
    
    theWin = spot->theWin;

    (void) vemWnTurnOffLayer(theWin, "pageFrame");
    (void) vemWnGetOptions(theWin, &options);
    options.disp_options &= ~VEM_GRIDLINES;
    options.disp_options &= ~VEM_TITLEDISP;
    (void) vemWnSetOptions(theWin, &options);

    if (lsLength(cmdList) == 1) {
	XMoveResizeWindow(display, theWin, 0, 0, 
			  XDisplayWidth(display, screen),
			  XDisplayHeight(display, screen));
    } else {
	cm = DefaultColormap(display, screen);
	XAllocNamedColor(display, cm, "DimGray", &scr, &exact);
	XSetWindowBackground(display, root, scr.pixel);
	XClearWindow(display, root);
	XMoveResizeWindow(display, theWin, 140, 170, 750, 500);
    }
        
    XSync(display, False);
        
    (void) vemCommand("show-all", spot, lsCreate(), (long) 0);

    VEMMSG("Finished\n");
    vemPrompt();

    return(RPC_OK);
}


/*ARGSUSED*/
int
nextSlide(spot, cmdList, uow)
RPCSpot *spot;
lsList cmdList;
long uow;
{
    char *name;
    RPCArg *arg;
    Pointer pointer;
    lsList argList;
    char buffer[1024];
    char cell[1024], view[1024];
    

    if (initialized == 0) {
	VEMMSG("Not Initialized\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (lsNext(slideGen, &pointer, (lsHandle *) 0) != VEM_OK) {
	VEMMSG("Finished\n");
	vemPrompt();
	return(RPC_OK);
    }

    name = (char *) pointer;
    
    (void) sprintf(buffer, "Displaying slide `%s'\n", name);
    VEMMSG(buffer);

    arg = (struct RPCArg *) malloc(sizeof(struct RPCArg));

    argList = lsCreate();
    arg->argType = VEM_TEXT_ARG;
    
    if (sscanf(name, "%[^:]:%s", cell, view) == 2) {
	/* the view was specified */
	(void) strcpy(buffer, name);
    } else {
	/* no view was specified */
	(void) sprintf(buffer, "%s:slide", name);
    }

    arg->argData.string = util_strsav(buffer);
    (void) lsNewBegin(argList, (Pointer) arg, (lsHandle *) 0);
    
    (void) vemCommand("switch-facet", spot, argList, uow);

    VEMMSG("Finished\n");
    vemPrompt();
    return(RPC_OK);
}


/*ARGSUSED*/
int
previousSlide(spot, cmdList, uow)
RPCSpot *spot;
lsList cmdList;
long uow;
{
    char *name;
    Pointer pointer;
    RPCArg *arg;
    lsList argList;
    char buffer[1024], cell[1024], view[1024];

    if (initialized == 0) {
	VEMMSG("Not Initialized\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (lsPrev(slideGen, &pointer, (lsHandle *) 0) != VEM_OK) {
	VEMMSG("Finished\n");
	vemPrompt();
	return(RPC_OK);
    }
    name = (char *) pointer;

    arg = (struct RPCArg *) malloc(sizeof(struct RPCArg));

    argList = lsCreate();
    arg->argType = VEM_TEXT_ARG;
    
    if (sscanf(name, "%[^:]:%s", cell, view) == 2) {
	/* the view was specified */
	(void) strcpy(buffer, name);
    } else {
	/* no view was specified */
	(void) sprintf(buffer, "%s:slide", name);
    }
    
    arg->argData.string = util_strsav(buffer);
    (void) lsNewBegin(argList, (Pointer) arg, (lsHandle *) 0);
    
    (void) vemCommand("switch-facet", spot, argList, uow);

    VEMMSG("Finished\n");
    vemPrompt();
    return(RPC_OK);
}


/*ARGSUSED*/
int
continuous(spot, cmdList, uow)
RPCSpot *spot;
lsList cmdList;
long uow;
{
    char *name;
    RPCArg *arg;
    Pointer pointer;
    lsList argList;
    char buffer[1024], cell[1024], view[1024];
    int count, interval;
    lsGen gen;

    if (initialized == 0) {
	VEMMSG("Not Initialized\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_TEXT_ARG) {
	VEMMSG("\"count\" \"interval\" : continuous\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (sscanf(arg->argData.string, "%d", &count) != 1) {
	VEMMSG("'count' must be an integer\n");
	VEMMSG("\"count\" \"interval\" : continuous\n");
	vemPrompt();
	return(RPC_OK);
    }

    (void) lsDelBegin(cmdList, &pointer);
    arg = (RPCArg *) pointer;
    if (arg->argType != VEM_TEXT_ARG) {
	VEMMSG("\"count\" \"interval\" : continuous\n");
	vemPrompt();
	return(RPC_OK);
    }

    if (sscanf(arg->argData.string, "%d", &interval) != 1) {
	VEMMSG("'interval' must be an integer\n");
	VEMMSG("\"count\" \"interval\" : continuous\n");
	vemPrompt();
	return(RPC_OK);
    }

    while (count-- > 0) {
	
	VEMMSG("Starting Sequence\n");

	gen = lsStart(slideList);

	while (lsNext(gen, &pointer, (lsHandle *) 0) == VEM_OK) {

	    name = (char *) pointer;

	    (void) sprintf(buffer, "Displaying slide %s\n", name);
	    VEMMSG(buffer);

	    arg = (struct RPCArg *) malloc(sizeof(struct RPCArg));

	    argList = lsCreate();
	    arg->argType = VEM_TEXT_ARG;
	    
	    if (sscanf(name, "%[^:]:%s", cell, view) == 2) {
		/* the view was specified */
		(void) strcpy(buffer, name);
	    } else {
		/* no view was specified */
		(void) sprintf(buffer, "%s:slide", name);
	    }
	    
	    arg->argData.string = util_strsav(buffer);
	    lsNewBegin(argList, (Pointer) arg, (lsHandle *) 0);
	    
	    (void) vemCommand("switch-facet", spot, argList, uow);
	    (void) vemWnFlush();

	    sleep((unsigned) interval);
	}

	sleep((unsigned) interval);
    }

    VEMMSG("Finished\n");
    vemPrompt();
    return(RPC_OK);
}


static void
open_all(facet)
struct octObject *facet;
{
    octObject master, instance;
    octGenerator gen;
    int retval;
    int status;

    master.type = OCT_FACET;
    master.contents.facet.facet = "contents";
    master.contents.facet.mode = "r";
    
    if (octInitGenContents(facet,OCT_INSTANCE_MASK, &gen) != OCT_OK) {
	(void) octError("generating instances");
	return;
    }

    while ((status = octGenerate(&gen,&instance)) == OCT_OK) {
	retval = octOpenMaster(&instance, &master);
	if (retval < OCT_OK) {
	    (void) octError("Generating the instances");
	    return;
	} else if (retval == OCT_ALREADY_OPEN) {
	    continue;
	} else {
	    open_all(&master);
	}
	
    }

    if (status != OCT_GEN_DONE) {
	(void) octError("generating instances");
	return;
    }
    return;
}


/*ARGSUSED*/
int
loadUp(spot, cmdList, uow)
RPCSpot *spot;
lsList cmdList;
long uow;
{
    octObject facet;
    octStatus retval;
    char *name;
    Pointer pointer;
    char buffer[1024], cell[1024], view[1024];
    lsGen gen;

    if (initialized == 0) {
	VEMMSG("Not initialized\n");
	return(RPC_OK);
    }

    gen = lsStart(slideList);
    
    while (lsNext(gen, &pointer, (lsHandle *) 0) == VEM_OK) {

	name = (char *) pointer;

	(void) sprintf(buffer, "Loading up %s\n", name);
	VEMMSG(buffer);
	
	if (sscanf(name, "%[^:]:%s", cell, view) == 2) {
	    /* the view was specified */
	    facet.contents.facet.cell = cell;
	    facet.contents.facet.view = view;
	} else {
	    /* no view was specified */
	    facet.contents.facet.cell = name;
	    facet.contents.facet.view = "slide";
	}
	
	facet.type = OCT_FACET;
	facet.contents.facet.facet = "contents";
	facet.contents.facet.version = OCT_CURRENT_VERSION;
	facet.contents.facet.mode = "r";
	
	retval = octOpenFacet(&facet);
	if (retval < OCT_OK) {
	    (void) octError(name);
	    continue;
	} else if (retval == OCT_ALREADY_OPEN) {
	    continue;
	} else if (retval == OCT_NEW_FACET) {
	    VEMMSG("\n2Warning, non-existant facet\n");
	    VEMMSG(name);
	    VEMMSG("\n0");
	    continue;
	} else {
	    open_all(&facet);
	}
    }

    VEMMSG("Finished\n");
    vemPrompt();
    return(RPC_OK);
}


