#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "rpc.h"
#include "oh.h"
#include "st.h"
#include <pwd.h>
#include <sys/file.h>
#include "sm.h"

extern st_table *fonts;


int
findGlyph(c,fontName)
char c;
char *fontName;
{
    int *item;
    char *pointer;

    if ((int) c < 0) {
	return(-1);
    }
    if (st_lookup(fonts, fontName, &pointer)) {
	item = (int *) pointer;
	return(item[(int) c]);
    }
    return(-1);
}


octStatus
openMaster(instance, master, mode)
octObject *instance;
octObject *master;
char *mode;
{
    master->contents.facet.facet = "contents";
    master->contents.facet.mode = mode;
    master->type = OCT_FACET;
    return(octOpenMaster(instance, master));
}


octStatus
createInterface(facet, layerName, string)
octObject *facet;
char *layerName;
char *string;
{
    octObject interface, layer, label, geo, place;
    struct octBox box;
       
    interface = *facet;
    interface.contents.facet.facet = "interface";
    interface.contents.facet.mode = "w";
        
    if (octOpenFacet(&interface) < OCT_OK) {
	(void) octError("can not create the interface");
	return(OCT_ERROR);
    }

    if (addPolicyProperties(&interface) < OCT_OK) {
	VEMMSG("can not create the policy properties\n");
	return(RPC_OK);
    }

    layer.type = OCT_LAYER;
    layer.contents.layer.name = layerName;
    if (octCreate(&interface, &layer) < OCT_OK) {
	(void) octError("can not create the LAYER");
	return(OCT_ERROR);
    }
    
    place.type = OCT_LAYER;
    place.contents.layer.name = "PLACE";
    if (octCreate(&interface, &place) < OCT_OK) {
	(void) octError("can not create the LAYER");
	return(OCT_ERROR);
    }
        
    if (octBB(facet, &box) < OCT_OK) {
	(void) octError("bad bb call");
	if (octFlushFacet(&interface) < OCT_OK) {
	    (void) octError("can not flush the facet");
	    return(OCT_ERROR);
	}
	return(OCT_OK);
    }
    
    label.type = OCT_LABEL;
    label.contents.label.label = string;
    label.contents.label.region = box;
    label.contents.label.textHeight = 20;
    label.contents.label.vertJust = OCT_JUST_CENTER;
    label.contents.label.horizJust = OCT_JUST_CENTER;
    label.contents.label.lineJust = OCT_JUST_CENTER;
    if (octCreate(&layer, &label) < OCT_OK) {
	(void) octError("can not create the LABEL");
	return(OCT_ERROR);
    }
    
    geo.type = OCT_BOX;
    geo.contents.box = box;
    if (octCreate(&place, &geo) < OCT_OK) {
	(void) octError("can not create the interface BOX");
	return(OCT_ERROR);
    }
    
    if (octFlushFacet(&interface) < OCT_OK) {
	(void) octError("can not flush the facet");
	return(OCT_ERROR);
    }
    	    
    return(OCT_OK);
}


octStatus
addPolicyProperties(facet)
octObject *facet;
{
    octObject prop;
    
    if (ohCreatePropStr(facet, &prop, "EDITSTYLE", "PHYSICAL") < OCT_OK) {
	(void) octError("can not create the EDITSTYLE");
	return(OCT_ERROR);
    }
    
    if (ohCreatePropStr(facet, &prop, "VIEWTYPE", "slide") < OCT_OK) {
	(void) octError("can not create the VIEWTYPE");
	return(OCT_ERROR);
    }

    if (ohCreatePropStr(facet, &prop, "TECHNOLOGY", "slide") < OCT_OK) {
	(void) octError("can not create the TECHNOLOGY");
	return(OCT_ERROR);
    }
    
    return(OCT_OK);
}


int
justValid(justification)
char *justification;
{
    if (!STREQ(justification, "left") &&
	!STREQ(justification, "center") &&
	!STREQ(justification, "right")) {
	VEMMSG("illegal justification, must be one of 'left', 'center', or 'right'\n");
	return(0);
    }
    return(1);
}


int
justOffset(justification, instance)
char *justification;
octObject *instance;
{
    struct octBox bb;
    
    if STREQ(justification, "left") {
	return(0);
    } else if STREQ(justification, "center") {
	if (octBB(instance, &bb) < OCT_OK) {
	    return(0);
	}
	return((bb.upperRight.x - bb.lowerLeft.x) / 2);
    } else if STREQ(justification, "right") {
	if (octBB(instance, &bb) < OCT_OK) {
	    return(0);
	}
	return(bb.upperRight.x - bb.lowerLeft.x);
    }
    return(0);
}


char *
celltail(name)
char *name;
{
    char *newname, *p;

    newname = util_strsav(name);
    
    p = strrchr(newname, '/');

    if (p == 0) {
	return(newname);
    } else {
	return(++p);
    }
}

    
octStatus
GetStringProperty(container, name, string, dflt)
octObject *container;
char *name;
char *string;
char *dflt;
{
    octObject prop;

    if (ohGetByPropName(container, &prop, name) < OCT_OK) {
	(void) strcpy(string, dflt);
    } else {
	(void) strcpy(string, prop.contents.prop.value.string);
    }
    return(OCT_OK);
}    
    

octStatus
GetRealProperty(container, name, real, dflt)
octObject *container;
char *name;
double *real;
double dflt;
{
    octObject prop;

    if (ohGetByPropName(container, &prop, name) < OCT_OK) {
	*real = dflt;
    } else {
	*real = prop.contents.prop.value.real;
    }
    return(OCT_OK);
}    
    

octStatus
GetIntegerProperty(container, name, integer, dflt)
octObject *container;
char *name;
int *integer;
int dflt;
{
    octObject prop;

    if (ohGetByPropName(container, &prop, name) < OCT_OK) {
	*integer = dflt;
    } else {
	*integer = prop.contents.prop.value.integer;
    }
    return(OCT_OK);
}    


void
stripNewLines(string)
char *string;
{
    char *p;
    
    while ((p = strrchr(string, '\n')) != 0) {
	(void) strcpy(p, (p+1));
    }
    return;
}

