static char rcs_id[] = "$Id: files.c,v 1.1.1.1 2002/12/31 15:51:04 saj Exp $";
#include "copyright.h"
#include "port.h"
#include <sys/file.h>
#include "internal.h"
#include "io_internal.h"
#include "fsys.h"

octStatus
oct_make_key(user_facet, file_desc, parent, location)
struct octFacet *user_facet;
char **file_desc;
struct facet *parent;
int location;
{
    char *parent_desc =
      (parent == NIL(struct facet) ? NIL(char) : parent->file_desc);

    if (!fsys_make_key(user_facet, file_desc, parent_desc, location)) {
	oct_error(fsys_error_message);
	return OCT_ERROR;
    }

    return OCT_OK;
}
       
octStatus
oct_save_key(file_desc, copy_desc)
char *file_desc;
char **copy_desc;
{
    if (!fsys_save_key(file_desc, copy_desc)) {
	oct_error(fsys_error_message);
	return OCT_ERROR;
    }

    return OCT_OK;
}
       
octStatus
oct_resolve(user_facet, mode, desc)
struct octFacet *user_facet;
int mode;
char *desc;
{
    int umode, fmode;
    
    switch (mode&~OCT_REVERT) {
    case OCT_READ:
	umode = FSYS_READABLE;
	break;
    case OCT_OVER_WRITE:
	umode = FSYS_WRITABLE;
	break;
    case OCT_APPEND:
	umode = FSYS_READABLE|FSYS_WRITABLE;
	break;
    default:
	oct_error("Illegal facet mode (%d)", mode);
	errRaise(OCT_PKG_NAME, OCT_ERROR, octErrorString());
    }

    fmode = fsys_resolve(desc);
    
    if (!(fmode&FSYS_EXISTS)) {
	if (fmode&FSYS_CREATABLE) {
	    if (umode&FSYS_WRITABLE) {
		return OCT_NEW_FACET;
	    } else {
		oct_error("`%s' doesn't exist",
			  oct_facet_name(user_facet));
		return OCT_NO_EXIST;
	    }
	} else {
	    oct_error("Cannot create facet: %s", fsys_error_message);
	    return OCT_NO_PERM;
	}
    }
    
    if ((fmode&umode) != umode) {
	oct_error("`%s' : permission denied", oct_facet_name(user_facet));
	return OCT_NO_PERM;
    }
    
    return OCT_OLD_FACET;
}

octStatus
oct_open_facet_stream(desc, mode, file_handle)
char *desc;
char *mode;
char **file_handle;
{
    IOFILE *ptr, *fsys_open();
    IOFILE **file_desc = (IOFILE **) file_handle;

    ptr = fsys_open(desc, mode);
    if (ptr == NULL) {
	oct_error(fsys_error_message);
	return OCT_ERROR;
    }
    
    *file_desc = ptr;
    return OCT_OK;
}

octStatus
oct_commit(desc)
char *desc;
{
    if (!fsys_commit(desc)){
	oct_error("Cannot commit facet: %s", fsys_error_message);
	return OCT_ERROR;
    }
    return OCT_OK;
}

octStatus
oct_lock(facet)
struct facet *facet;
{
    if (facet->facet_flags.locked) {
	return OCT_OK;
    }
	
    if (!fsys_lock(facet->file_desc)){
	oct_error("Cannot lock facet: %s", fsys_error_message);
	return OCT_ERROR;
    }

    facet->facet_flags.locked = 1;
    return OCT_OK;
}

octStatus
oct_unlock(facet)
struct facet *facet;
{
    if (!(facet->facet_flags.locked)) {
	return OCT_OK;
    }
    
    if (!fsys_unlock(facet->file_desc)){
	oct_error("Cannot unlock facet: %s", fsys_error_message);
	return OCT_ERROR;
    }
    facet->facet_flags.locked = 0;
    return OCT_OK;
}
