static char rcs_id[] = "$Id: polygon.c,v 1.1.1.1 2002/12/31 15:51:04 saj Exp $";
/*
 * Implementations of the the oct functions for polygons.
 */

#include "copyright.h"
#include "port.h"
#include "internal.h"
#include "geo.h"
#include "io.h"

static octStatus polygon_bb();
static octStatus polygon_read_fields();

extern struct object_desc oct_geo_points_desc;
static struct object_desc *super = &oct_geo_points_desc;

oct_polygon_desc_set(object_desc)
struct object_desc *object_desc;
{
    *object_desc = *super;
    
    object_desc->internal_size = sizeof(struct polygon);
    object_desc->external_size = 0;
    object_desc->flags |= HAS_BB_FIELD|HAS_BB_FUNC;
    object_desc->contain_mask = (octObjectMask) OCT_PROP_MASK|OCT_BAG_MASK;
    object_desc->user_offset = ILLEGAL_OFFSET;
    object_desc->bb_offset = FIELD_OFFSET(struct polygon, bbox);
    object_desc->bb_func = polygon_bb;
    object_desc->read_fields_func = polygon_read_fields;
}

static octStatus
polygon_bb(pptr, user_box)
generic *pptr;
struct octBox *user_box;
{
    struct octPoint *ptr;
    int32 i;
    struct polygon *poly = (struct polygon *) pptr;
    struct octBox *box = &poly->bbox;

    if (VALID_BB(&poly->bbox)) {
	if (user_box != NIL(struct octBox)) {
	    *user_box = poly->bbox;
	}
	return OCT_OK;
    }
	
    i = poly->num_points;

    if (i < 1) {
	oct_error("The polygon has no points, thus no bounding box");
	return OCT_NO_BB;
    }
    
    ptr = poly->points;
    box->lowerLeft = box->upperRight = *ptr;
    
    for(--i,++ptr; i > 0; i--,ptr++) {
	if (box->lowerLeft.x > ptr->x) {
	    box->lowerLeft.x = ptr->x;
	} else if (box->upperRight.x < ptr->x) {
	    box->upperRight.x = ptr->x;
	}
	if (box->lowerLeft.y > ptr->y) {
	    box->lowerLeft.y = ptr->y;
	} else if (box->upperRight.y < ptr->y) {
	    box->upperRight.y = ptr->y;
	}
    }
    if (user_box != NIL(struct octBox)) {
	*user_box = *box;
    }
    return OCT_OK;
}

static octStatus
polygon_read_fields(pptr)
generic *pptr;
{
    struct polygon *polygon = (struct polygon *) pptr;
    int retval;

    retval = (*super->read_fields_func)(pptr);

    if (retval) {
	MAKE_BB_INVALID(&polygon->bbox); /* force the bb computation */
	(void) oct_bb_get(pptr, NIL(struct octBox));
    }
    return retval;
}
