#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yybdsyn.h"

extern int number_of_assigns;


/*
 * This routine will "clean up the flattened tree before sending it
 * to MLespresso for flattening.  It will:
 * 1) forward substitute all constants.
 * 2) throw out all "statements" that are not really statements.
 *            ( they don't do anything )
 * 3) re-EVAL everything to simplify it down.
 * 4) Throw out assignments flagged as useless by revassign
 */
node *clean_up( flattree )
node *flattree;
{
    node *result;
    node *place;
    node *lastplace;
    node *nextplace;

    result = place = flattree;
    lastplace = NULL;

    while ( place != NULL ) {
	nextplace = place->next;
	place->next = NULL;
	
	if ((place->type == T_EQ) && 
		(place->un.assign.lexpr->type == T_IS_STATE)) {
	    number_of_assigns++;
	    /* node_eval rhs */
	    place->un.assign.rexpr = node_eval(place->un.assign.rexpr);

	    /* link it into tree */
	    if ( lastplace == NULL ) {
		lastplace = result = place;
	    } else {
		lastplace->next = place;
		lastplace = lastplace->next;
	    }
	    lastplace->next = place = nextplace;

	} else {
	    /* throw out the "statement" and move on */
	    node_dispose( place );
	    if ( lastplace == NULL ) {
		result = place = nextplace;
	    } else {
		lastplace->next = place = nextplace;
	    }
	}
    }
    return result;
}
