#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yybdsyn.h"
#include "eval.h"

extern FILE *outfile;
extern int saw_complex;		/* flag the substitution */
extern int update;
char *map_token_to_name();

char *functname();
node *macro_ex();

node *complex( root, constant, width )
node *root;
int *constant;
int *width;
{
    int const1, const2;
    int width1, width2;
    int concconst;
    int concwidth;
    node *stmt;
    node *laststmt;
    node *arg1, *arg2;

    if ( root == NULL ) {
	*constant = 1;
	*width = 0;
	return root;
    }

    laststmt = NULL;
    for ( stmt = root; stmt != NULL; stmt = stmt->next ) {
	if ( stmt->conc != NULL ) {
	    stmt->conc = complex( stmt->conc, &concconst, &concwidth );
	} else {
	    concconst = 1;
	    concwidth = 0;
	}
	switch ( stmt->type ) {
	    case T_NOT: case T_BUF:
	    case T_UNARY_PLUS: case T_UNARY_MINUS:
	    case T_UNARY_OR: case T_UNARY_AND:
		stmt->un.unary.expr = complex( stmt->un.unary.expr,
			&const1, &width1 );
		*constant = const1 & concconst;
		break;

	    case T_WIDTH:
		stmt->un.unary.expr = complex( stmt->un.unary.expr,
			&const1, &width1 );
		*constant = concconst;
		break;

	    case T_DIV: case T_MOD:
	    case T_AND: case T_OR: case T_NAND:
	    case T_NOR: case T_XOR: case T_EQV:
		stmt->un.binary.lexpr = complex( stmt->un.binary.lexpr,
			&const1, &width1 );
		stmt->un.binary.rexpr = complex( stmt->un.binary.rexpr,
			&const2, &width2 );
		*constant = const1 & const2 & concconst;
		break;

	    case T_SXT: case T_ZXT: case T_OXT:
		stmt->un.binary.lexpr = complex( stmt->un.binary.lexpr,
			&const1, &width1 );
		stmt->un.binary.rexpr = complex( stmt->un.binary.rexpr,
			&const2, &width2 );
		if ( !const2 ) {
		    error( "Can only extend by constant", stmt );
		}
		*constant = const1 & concconst;
		break;

	    case T_EQL: case T_NEQ:
		stmt->un.binary.lexpr = complex( stmt->un.binary.lexpr,
			&const1, &width1 );
		stmt->un.binary.rexpr = complex( stmt->un.binary.rexpr,
			&const2, &width2 );
		/* complex is when both are variable and the
		 * comparison is several bits */
		if (!const1 && !const2 &&
			((width1 > 3) || (width2 > 3))) {
		    arg1 = stmt->un.binary.lexpr;
		    arg2 = stmt->un.binary.rexpr;
		    stmt->un.binary.lexpr = stmt->un.binary.rexpr = NULL;
		    if ( laststmt == NULL ) {
			root = stmt = macro_ex( stmt, arg1, arg2,
				width1, width2 );
		    } else {
			laststmt->next = stmt = macro_ex( stmt, arg1,
				arg2, width1, width2 );
		    }
		}
		*constant = const1 & const2 & concconst;
		break;

	    case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
	    case T_PLUS: case T_MINUS: case T_TIMES: 
		stmt->un.binary.lexpr = complex( stmt->un.binary.lexpr,
			&const1, &width1 );
		stmt->un.binary.rexpr = complex( stmt->un.binary.rexpr,
			&const2, &width2 );
		if ( !( const1 & const2 ) ) {
		    arg1 = stmt->un.binary.lexpr;
		    arg2 = stmt->un.binary.rexpr;
		    stmt->un.binary.lexpr = stmt->un.binary.rexpr = NULL;
		    if ( laststmt == NULL ) {
			root = stmt = macro_ex( stmt, arg1, arg2,
				width1, width2 );
		    } else {
			laststmt->next = stmt = macro_ex( stmt, arg1,
				arg2, width1, width2 );
		    }
		}
		*constant = const1 & const2 & concconst;
		break;

	    case T_SL0: case T_SR0: case T_SL1:
	    case T_SR1: case T_SLR: case T_SRR:
		stmt->un.binary.lexpr = complex( stmt->un.binary.lexpr,
			&const1, &width1 );
		stmt->un.binary.rexpr = complex( stmt->un.binary.rexpr,
			&const2, &width2);
		if ( !const2 ) {
			if ((int) (1 << width2) <= 0) {
				error( "Shift width exceeds machine word width ", stmt );
			}
		    arg1 = stmt->un.binary.lexpr;
		    arg2 = stmt->un.binary.rexpr;
		    stmt->un.binary.lexpr = stmt->un.binary.rexpr = NULL;
		    if ( laststmt == NULL ) {
			root = stmt = macro_ex( stmt, arg1, arg2,
				width1, width2);
		    } else {
			laststmt->next = stmt = macro_ex( stmt, arg1,
				arg2, width1, width2 );
		    }
		}
		*constant = const1 & concconst;
		break;

	    case T_BIT:
		stmt->un.ternary.lexpr = complex( stmt->un.ternary.lexpr,
			&const1, &width1 );
		stmt->un.ternary.mexpr = complex( stmt->un.ternary.mexpr,
			&const2, &width2 );
		if ( !const2 ) {
		    if ( stmt->un.ternary.rexpr == NULL ) {
			/* do complex expansion */
			arg1 = stmt->un.ternary.lexpr;
			arg2 = stmt->un.ternary.mexpr;
			stmt->un.ternary.lexpr = stmt->un.ternary.mexpr = NULL;
			if ( laststmt == NULL ) {
			    root = stmt = macro_ex( stmt, arg1, arg2,
				    width1, width2);
			} else {
			    laststmt->next = stmt = macro_ex( stmt, arg1,
				    arg2, width1, width2 );
			}
		    } else {
			error( "Can't evaluate complex bit range", stmt );
		    }
		} else {
		    if ( stmt->un.ternary.rexpr != NULL ) {
			stmt->un.ternary.rexpr = complex(
				stmt->un.ternary.rexpr, &const2, &width2 );
			if ( !const2 ) {
			    error( "Can't evaluate complex bit range", stmt );
			}
		    } else {	/* make a rexpr so eval can work */
			stmt->un.ternary.rexpr =
				node_dup( stmt->un.ternary.mexpr );
		    }
		}
		*constant = const1 & concconst;
		break;

	    case T_EQ:
		if ( (stmt->un.assign.lexpr->type == T_BIT) &&
			(stmt->un.assign.lexpr->un.ternary.rexpr == NULL) ) {
		    stmt->un.assign.lexpr->un.ternary.rexpr = 
			    node_dup( stmt->un.assign.lexpr->un.ternary.mexpr );
		}
		stmt->un.assign.rexpr = complex( stmt->un.assign.rexpr,
			constant, width );
		*constant = 0;
		break;

	    case T_IS_STATE:
		*constant = 0;
		break;

	    case T_IS_CONSTANT:
		*constant = concconst;
		break;

	    case T_LABEL:
		stmt->un.label.stmt = complex( stmt->un.label.stmt,
			constant, width );
		*constant = 0;
		break;

	    case T_IF:
		stmt->un.bdif.thenb = complex( stmt->un.bdif.thenb,
			constant, width );
		stmt->un.bdif.elseb = complex( stmt->un.bdif.elseb,
			constant, width );
		*constant = 0;
		break;

	    case T_FOR:
	    case T_FOR_DOWN:
		stmt->un.bdfor.stmt = complex( stmt->un.bdfor.stmt,
			constant, width );
		*constant = 0;
		break;

	    case T_CASE:
		stmt->un.bdcase.stmt = complex( stmt->un.bdcase.stmt,
			constant, width );
		*constant = 0;
		break;

	    case T_SELECT:
		stmt->un.select.bdcase = complex( stmt->un.select.bdcase,
			constant, width );
		*constant = 0;
		break;

	    case T_LEAVE:
		*constant = 0;
		break;

	    default:
		ierror( "Unknown type in complex", stmt );

	}

	switch ( stmt->type ) {
	    case T_NOT: case T_BUF: case T_WIDTH:
	    case T_UNARY_PLUS: case T_UNARY_MINUS:
	    case T_UNARY_OR: case T_UNARY_AND:
		*width = width1 + concwidth;
		break;

	    case T_TIMES:
		*width = width1 + width2 + concwidth;
		break;

	    case T_EQL: case T_NEQ:
	    case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
		*width = 1 + concwidth;
		break;

	    case T_PLUS: case T_MINUS:
		*width = (width1 > width2) ? width1 + 1 + concwidth :
			width2 + 1 + concwidth;
		break;

	    case T_AND: case T_OR: case T_NAND:
	    case T_NOR: case T_XOR: case T_EQV:
		*width = (width1 > width2) ? width1 + concwidth :
			width2 + concwidth;
		break;

	    case T_SXT: case T_ZXT: case T_OXT:
		stmt->un.binary.rexpr = node_eval( stmt->un.binary.rexpr );
		*width = stmt->un.binary.rexpr->un.constant.num + concwidth;
		break;

	    case T_DIV: case T_MOD:
	    case T_SL0: case T_SR0: case T_SL1:
	    case T_SR1: case T_SLR: case T_SRR:
		*width = width1 + concwidth;
		break;

	    case T_BIT:
		if ( stmt->un.ternary.rexpr == NULL ) {
		    *width = 1 + concwidth;
		} else {
		    stmt->un.ternary.mexpr =
			    node_eval( stmt->un.ternary.mexpr );
		    stmt->un.ternary.rexpr =
			    node_eval( stmt->un.ternary.rexpr );
		    *width = stmt->un.ternary.mexpr->un.constant.num -
			    stmt->un.ternary.rexpr->un.constant.num +
			    1 + concwidth;
		}
		break;

	    case T_IS_STATE:
		*width = stmt->un.bdvar.hbit - stmt->un.bdvar.lbit + 1 +
			concwidth;
		break;

	    case T_IS_CONSTANT:
		*width = stmt->un.constant.hbit + 1 + concwidth;
		break;

	    case T_EQ:
	    case T_LABEL:
	    case T_IF:
	    case T_FOR:
	    case T_FOR_DOWN:
	    case T_CASE:
	    case T_SELECT:
	    case T_LEAVE:
		*width = 0;
		break;

	    case T_FCALL:
		*width = stmt->un.fcall.hbit + 1 + concwidth;
		break;

	    default:
		ierror( "Unknown type in complex 2", stmt );

	}
    }
    return root;
}


node *macro_ex( op, arg1, arg2, width1, width2 )
node *op;		/* operation to be replaced */
node *arg1, *arg2;
int width1, width2;	/* width of args */
{
    node *nextsave, *concsave;
    char opname[8];		/* name of operator */
    char name[16];		/* name of function to create */
    char parsestring[256];	/* string to send to parser */
    char buf[80];
    int a;			/* high bit (a) */
    int b;			/* high bit (b) */
    int max;			/* max (a,b) */
    int maxwidth;		/* max +1 */
    int power;			/* 2 ^ width2 --> for bit select */
    node *call;			/* function call result */
    int hibit;			/* logical hbit of created function */

    saw_complex = 1;
    nextsave = op->next;
    concsave = op->conc;
    op->next = op->conc = NULL;
    a = width1 - 1;
    b = width2 - 1;
    max = ( a > b ) ? a : b;
    maxwidth = max + 1;

    (void) strcpy( opname, map_token_to_name( op->type ) );

    switch ( op->type ) {

	case T_TIMES:
	    (void) strcpy( opname, "TIMES" );
	    hibit = width1 + width2 - 1;
	    (void) sprintf( name, "$%s_%d_%d", opname, width1, width2 );
	    break;

	case T_PLUS:
	    (void) strcpy( opname, "PLUS" );
	    (void) sprintf( name, "$%s_%d", opname, maxwidth + 1 );
	    hibit = max + 1;
	    break;

	case T_MINUS:
	    (void) strcpy( opname, "MINUS" );
	    (void) sprintf( name, "$%s_%d", opname, maxwidth );
	    hibit = max + 1;
	    break;

	case T_LSS: case T_LEQ: case T_GTR: case T_GEQ:
	case T_EQL: case T_NEQ:
	    hibit = 0;
	    (void) sprintf( name, "$%s_%d", opname, maxwidth );
	    break;

	case T_SL0: case T_SL1: case T_SLR:
	case T_SR0: case T_SR1: case T_SRR:
	    hibit = width1 - 1;
	    (void) sprintf( name, "$%s_%d_%d", opname, width1, width2 );
	    break;

	case T_BIT:
	    (void) strcpy( opname, "BIT" );
	    hibit = 0;
    	    power = (int) pow( (double) 2, (double) width2 ) - 1;
	    (void) sprintf( name, "$%s_%d_%d", opname, width1, width2 );
	    break;
    }

    (void) sprintf( parsestring, "MACRO \"$max\" = %d $ENDMACRO;\n", max );
    if (op->type == T_BIT)	/* pow only defined where it's needed */
	{
    	    (void) sprintf( buf, "MACRO \"$pow\" = %d $ENDMACRO;\n", power );
    	    (void) strcat( parsestring, buf );
	}
    (void) sprintf( buf, "MACRO \"$a\" = %d $ENDMACRO;\n", a );
    (void) strcat( parsestring, buf );
    (void) sprintf( buf, "MACRO \"$b\" = %d $ENDMACRO;\n", b );
    (void) strcat( parsestring, buf );
    (void) sprintf( buf, "MACRO \"$name\" = %s $ENDMACRO;\n$%s\n", name, opname );
    (void) strcat( parsestring, buf );

    if ( update ) {
	(void) fprintf( outfile, "# Inserting %s in %s at line %d\n", name,
		op->file, op->line );
    }
    call = node_create_fi( T_FCALL, op );
    call->un.fcall.def = symtab_lookup( global_syms, name, 0 );
    if ( call->un.fcall.def == NULL ) {
	yyreparse( parsestring );

	call->un.fcall.def = symtab_lookup( global_syms, name, 0 );
	if ( call->un.fcall.def == NULL ) {
	    ierror( "Lookup fails in complex", call );
	}
    }

    /* link on args in reverse order. */
    call->un.fcall.actual_args = arg2;
    call->un.fcall.actual_args->next = arg1;
    call->un.fcall.hbit = hibit;
    call->next = nextsave;
    call->conc = concsave;
    NODE_DISPOSE( op );
    return call;
}
