#include "copyright.h"
#include "port.h"
#include "utility.h"
#include "symtab.h"
#include "node.h"
#include "yybdsyn.h"
#include "mult.h"

#define UNDEFINED -1
char *map_token_to_name();

int functnum;			/* LIF number for a function */
extern int watch_espresso;	/* trace level */
extern int nodes_alloced;
extern int make_unspec_set;
extern FILE *bdsyntemp;
extern int short_1_bit;

void
node_init()
{
    functnum = 0;
    nodes_alloced = 0;
}


/*
 * This routine will copy a node and recursively call itself to
 * copy all subtrees of the node.
 */
node *node_dup(template)
node *template;		/* node to copy */
{
    node *newnode;

    if ( template == NULL ) {
	return NULL;
    } else {
        newnode = node_create_fi( template->type, template );
        newnode->next = node_dup( template->next );
        newnode->conc = node_dup( template->conc );
        switch ( template->type ) {

	    case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	    case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
	        newnode->un.unary.expr = node_dup( template->un.unary.expr );
	        break;
    
	    case T_AND: case T_OR: case T_NAND: case T_NOR: case T_XOR: case T_EQV:
	    case T_EQL: case T_NEQ: case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
	    case T_SL0: case T_SR0: case T_SL1: case T_SR1: case T_SLR: case T_SRR:
	    case T_SXT: case T_ZXT: case T_OXT:
	    case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	        newnode->un.binary.lexpr = node_dup( template->un.binary.lexpr );
	        newnode->un.binary.rexpr = node_dup( template->un.binary.rexpr );
	        break;
    
	    case T_BIT:
	        newnode->un.ternary.lexpr = node_dup( template->un.ternary.lexpr );
	        newnode->un.ternary.mexpr = node_dup( template->un.ternary.mexpr );
	        newnode->un.ternary.rexpr = node_dup( template->un.ternary.rexpr );
	        break;
    
	    case T_EQ:
		newnode->un.assign.lexpr = node_dup( template->un.assign.lexpr );

		newnode->un.assign.rexpr = node_dup( template->un.assign.rexpr );
		newnode->un.assign.needlif = template->un.assign.needlif;
		newnode->un.assign.prelifed = template->un.assign.prelifed;
		newnode->un.assign.lifed = template->un.assign.lifed;
		break;

	    case T_IS_STATE:
	        newnode->un.bdvar.name = template->un.bdvar.name;
	        newnode->un.bdvar.lbit = template->un.bdvar.lbit;
	        newnode->un.bdvar.hbit = template->un.bdvar.hbit;
		newnode->un.bdvar.rev = template->un.bdvar.rev;
		newnode->un.bdvar.var = template->un.bdvar.var;
	        break;
    
	    case T_IS_CONSTANT:
	        newnode->un.constant.num = template->un.constant.num;
	        newnode->un.constant.dont_care = template->un.constant.dont_care;
	        newnode->un.constant.hbit = template->un.constant.hbit;
	        break;

	    case T_LEAVE:
		newnode->un.leave.name = template->un.leave.name;
	        newnode->un.leave.conds = node_dup( template->un.leave.conds );
		break;

	    case T_LABEL:
		newnode->un.label.name = template->un.label.name;
	        newnode->un.label.stmt = node_dup( template->un.label.stmt );
		break;

	    case T_IF:
	        newnode->un.bdif.expr = node_dup( template->un.bdif.expr );
	        newnode->un.bdif.thenb = node_dup( template->un.bdif.thenb );
	        newnode->un.bdif.elseb = node_dup( template->un.bdif.elseb );
	        break;

	    case T_FOR:
	    case T_FOR_DOWN:
	        newnode->un.bdfor.var = node_dup( template->un.bdfor.var );
	        newnode->un.bdfor.from = node_dup( template->un.bdfor.from );
	        newnode->un.bdfor.to = node_dup( template->un.bdfor.to );
	        newnode->un.bdfor.by = node_dup( template->un.bdfor.by );
	        newnode->un.bdfor.stmt = node_dup( template->un.bdfor.stmt );
	        break;
    
	    case T_CASE:
	        newnode->un.bdcase.expr = node_dup( template->un.bdcase.expr);
	        newnode->un.bdcase.stmt = node_dup( template->un.bdcase.stmt);
		break;
    
	    case T_SELECT: case T_SELECTONE: case T_SELECTALL:
	        newnode->un.select.expr = node_dup( template->un.select.expr);
	        newnode->un.select.bdcase = node_dup( template->un.select.bdcase );
		break;
    
	    case T_FCALL:
	        newnode->un.fcall.actual_args =
			node_dup( template->un.fcall.actual_args );
	        newnode->un.fcall.def = template->un.fcall.def;
		newnode->un.fcall.hbit = template->un.fcall.hbit;
		break;
    
        }
	return( newnode );
    }
}


/*
 * This routine takes an int, and returns a constant node with
 * a 32 bit number.  It also takes an infonode as in node_create.
 */
node *node_mkconst( number, infonode )
int number;
node *infonode;
{
    node *result;

    result = node_create_fi( T_IS_CONSTANT, infonode );
    result->un.constant.hbit = 31;
    result->un.constant.num = number;
    return( result );
}


/*
 * This routine will allocate a node that is initialized correctly
 * for its particular type.  The origin information is copied
 * from the second argument.
 */
node *node_create_fi(ntype, infonode)
int ntype;		/* type of new node */
node *infonode;		/* node to copy info from */
{
    return node_create(ntype, infonode->file, infonode->line);
}
    

/*
 * This returns a one bit constant node
 */
node *node_create_bit( bit, infonode )
int bit;
node *infonode;
{
    node *result;

    result = node_create_fi( T_IS_CONSTANT, infonode );
    result->un.constant.hbit = 0;
    result->un.constant.num = bit;
    return ( result );
}


/*
 * This routine will allocate a node that is initialized correctly
 * for its particular type.
 */
node *node_create(ntype, filename, line)
int ntype;
char *filename;
int line;
{
    node *newnode;

    newnode = node_alloc();
    nodes_alloced++;
    newnode->type = ntype;
    newnode->file = filename;
    newnode->line = line;
    newnode->next = NULL;
    newnode->conc = NULL;
    switch ( ntype ) {

        case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
	    newnode->un.unary.expr = NULL;
	    break;

        case T_AND: case T_OR: case T_NAND: case T_NOR: case T_XOR: case T_EQV:
        case T_EQL: case T_NEQ: case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
        case T_SL0: case T_SR0: case T_SL1: case T_SR1: case T_SLR: case T_SRR:
        case T_SXT: case T_ZXT: case T_OXT: case T_COLON:
        case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	    newnode->un.binary.lexpr = NULL;
	    newnode->un.binary.rexpr = NULL;
	    break;

	case T_BIT:
	    newnode->un.ternary.lexpr = NULL;
	    newnode->un.ternary.mexpr = NULL;
	    newnode->un.ternary.rexpr = NULL;
	    break;

	case T_EQ:
	    newnode->un.assign.lexpr = NULL;
	    newnode->un.assign.rexpr = NULL;
	    newnode->un.assign.needlif = 0;
	    newnode->un.assign.prelifed = 0;
	    newnode->un.assign.lifed = 0;
	    break;

	case T_IS_META:		/* Included for hack in parser */
	case T_IS_STATE:
	    newnode->un.bdvar.name = NULL;
	    newnode->un.bdvar.nbit = UNDEFINED;
	    newnode->un.bdvar.lbit = UNDEFINED;
	    newnode->un.bdvar.hbit = UNDEFINED;
	    newnode->un.bdvar.rev = NIL(usage);
	    newnode->un.bdvar.var = 0;
	    break;

	case T_IS_CONSTANT:
	    newnode->un.constant.num = 0;
	    newnode->un.constant.dont_care = 0;
	    newnode->un.constant.hbit = UNDEFINED;
	    break;

	case T_LEAVE:
	    newnode->un.leave.name = NULL;
	    newnode->un.leave.conds = NULL;
	    break;

	case T_LABEL:
	    newnode->un.label.name = NULL;
	    newnode->un.label.stmt = NULL;
	    break;

	case T_IF:
	    newnode->un.bdif.expr = NULL;
	    newnode->un.bdif.thenb = NULL;
	    newnode->un.bdif.elseb = NULL;
	    break;

	case T_FOR:
	case T_FOR_DOWN:
	    newnode->un.bdfor.var = NULL;
	    newnode->un.bdfor.from = NULL;
	    newnode->un.bdfor.to = NULL;
	    newnode->un.bdfor.by = NULL;
	    newnode->un.bdfor.stmt = NULL;
	    break;

	case T_CASE:
	    newnode->un.bdcase.expr = NULL;
	    newnode->un.bdcase.stmt = NULL;
	    break;

	case T_SELECT: case T_SELECTONE: case T_SELECTALL:
	    newnode->un.select.expr = NULL;
	    newnode->un.select.bdcase = NULL;
	    break;

	case T_FCALL:
	    newnode->un.fcall.def = NULL;
	    newnode->un.fcall.actual_args = NULL;
	    newnode->un.fcall.hbit = UNDEFINED;
	    break;

	case T_LIF:
	    newnode->un.lif.names = NULL;
	    newnode->un.lif.table = NULL;
	    break;

	case T_OTHERWISE:
	    break;
	    
	default:
	    ierror("Unknown type in node_create", newnode);
    }
    return newnode;
}


/*
 * This routine will free a node back to the allocator and recursively
 * free all of the node's subtrees, nexttrees, and conctrees.
 * CARE MUST BE TAKEN when using this routine.  The trees linked to
 * doomed node WILL BE DESTROYED, if the pointer is not first set to
 * NULL.
 */
void
node_dispose(root)
node *root;
{
    node *doomed, *temp;

    for ( doomed = root; doomed != NULL; ) {
        NODE_DISPOSE ( doomed->conc );
        switch ( doomed->type ) {

	    case T_IS_CONSTANT:
	    case T_OTHERWISE:
		break;

	    case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	    case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
	        NODE_DISPOSE ( doomed->un.unary.expr );
	        break;

	    case T_AND: case T_OR: case T_NAND: case T_NOR: case T_XOR: case T_EQV:
	    case T_EQL: case T_NEQ: case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
	    case T_SL0: case T_SR0: case T_SL1: case T_SR1: case T_SLR: case T_SRR:
	    case T_SXT: case T_ZXT: case T_OXT:
	    case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	        NODE_DISPOSE ( doomed->un.binary.lexpr );
	        NODE_DISPOSE ( doomed->un.binary.rexpr );
	        break;

	    case T_BIT:
	        NODE_DISPOSE ( doomed->un.ternary.lexpr );
	        NODE_DISPOSE ( doomed->un.ternary.mexpr );
	        NODE_DISPOSE ( doomed->un.ternary.rexpr );
	        break;

	    case T_EQ:
	        NODE_DISPOSE ( doomed->un.assign.lexpr );
	        NODE_DISPOSE ( doomed->un.assign.rexpr );
	        break;

	    case T_LEAVE:
		NODE_DISPOSE ( doomed->un.leave.conds );
		break;

	    case T_LABEL:
		NODE_DISPOSE ( doomed->un.label.stmt );
		break;

	    case T_IF:
	        NODE_DISPOSE ( doomed->un.bdif.expr );
	        NODE_DISPOSE ( doomed->un.bdif.thenb );
	        NODE_DISPOSE ( doomed->un.bdif.elseb );
	        break;

	    case T_FOR:
	    case T_FOR_DOWN:
	        NODE_DISPOSE ( doomed->un.bdfor.var );
	        NODE_DISPOSE ( doomed->un.bdfor.from );
	        NODE_DISPOSE ( doomed->un.bdfor.to );
	        NODE_DISPOSE ( doomed->un.bdfor.by );
	        NODE_DISPOSE ( doomed->un.bdfor.stmt );
	        break;

	    case T_CASE:
	        NODE_DISPOSE ( doomed->un.bdcase.expr );
	        NODE_DISPOSE ( doomed->un.bdcase.stmt );
		break;

	    case T_SELECT: case T_SELECTONE: case T_SELECTALL:
	        NODE_DISPOSE ( doomed->un.select.expr );
	        NODE_DISPOSE ( doomed->un.select.bdcase );
		break;

	    case T_FCALL:
	        NODE_DISPOSE ( doomed->un.fcall.actual_args );
		break;

	    case T_IS_ARG:
	    case T_IS_STATE:
		break;

	    case T_LIF:
		if ( doomed->un.lif.names != NULL )
		    FREE( doomed->un.lif.names );
		if ( doomed->un.lif.table != NULL )
		    FREE( doomed->un.lif.table );
		break;

	    default:
		ierror("node dispose on ?", doomed);

        }
	temp = doomed->next;
        FREE( doomed );		/* return node to alloc list */
	nodes_alloced--;
	doomed = temp;
    }
}


/*
 * indents lines for node_print
 */
void
node_print_indent( indent, string )
int indent;
char *string;
{
    int i;

    /* do proper indent */
    if ( indent > 0 ) {
	for( i=0; i < indent; i++ ) {
	    putchar(' ');
	}
    }
    (void) printf("%s ",string);
}


/*
 * This routine will print (to stdout) a representation of the tree
 * attatched to the given node.  When calling this from the outside,
 * indent should be zero.
 */
void
node_print(pnode, indent, string)
node *pnode;
int indent;
char *string;
{

    node_print_indent( indent, string );

    if ( pnode == NULL ) {

	(void) printf("NULL\n");

    } else {

        switch ( pnode->type ) {

	    case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	    case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
		(void) printf("UNARY %s \n", 
		    map_token_to_name(pnode->type));
	        node_print(pnode->un.unary.expr, indent+2, "uexp" );
	        break;

	    case T_AND: case T_OR: case T_NAND: case T_NOR: case T_XOR: case T_EQV:
	    case T_EQL: case T_NEQ: case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
	    case T_SL0: case T_SR0: case T_SL1: case T_SR1: case T_SLR: case T_SRR:
	    case T_SXT: case T_ZXT: case T_OXT:
	    case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
		(void) printf("BINARY %s \n",
		    map_token_to_name(pnode->type));
	        node_print ( pnode->un.binary.lexpr, indent+2 , "left");
	        node_print ( pnode->un.binary.rexpr, indent+2 , "righ");
	        break;

	    case T_BIT:
		(void) printf("BIT\n");
	        node_print ( pnode->un.ternary.lexpr, indent+2, "left" );
	        node_print ( pnode->un.ternary.mexpr, indent+2, "midd" );
	        node_print ( pnode->un.ternary.rexpr, indent+2, "righ" );
	        break;

	    case T_EQ:
		(void) printf("ASSIGN %d\n", pnode->un.assign.needlif );
	        node_print ( pnode->un.assign.lexpr, indent+2, "left" );
	        node_print ( pnode->un.assign.rexpr, indent+2, "righ" );
	        break;

	    case T_IS_ARG:
	    case T_IS_STATE:
		if (pnode->un.bdvar.rev == NIL(usage)) {
		    (void) printf("%s<%d:%d>\n", pnode->un.bdvar.name->name,
			    pnode->un.bdvar.hbit, pnode->un.bdvar.lbit);
		} else {
		    (void) printf("%s<%d:%d>%d.%d\n", pnode->un.bdvar.name->name,
			    pnode->un.bdvar.hbit, pnode->un.bdvar.lbit,
			    pnode->un.bdvar.rev->block->id,
			    pnode->un.bdvar.rev->rev);

		}
	        break;

	    case T_IS_CONSTANT:
		if ( pnode->un.constant.dont_care ) {
		    (void) printf("x<%d:0>\n", pnode->un.constant.hbit);
		} else {
		    (void) printf("%d<%d:0>\n", pnode->un.constant.num,
				pnode->un.constant.hbit);
		}
		break;

	    case T_LEAVE:
		(void) printf( "LEAVE %s\n", pnode->un.leave.name->name );
		node_print(pnode->un.leave.conds, indent+2, "cond" );
		break;

	    case T_LABEL:
		(void) printf( "LABEL %s\n", pnode->un.label.name->name );
		node_print(pnode->un.label.stmt, indent+2, "stmt" );
		break;

	    case T_IF:
		(void) printf("IF\n");
		node_print(pnode->un.bdif.expr, indent+2, "ifex" );
	        node_print(pnode->un.bdif.thenb, indent+2, "then" );
	        node_print(pnode->un.bdif.elseb, indent+2, "else" );
	        break;

	    case T_FOR:
	    case T_FOR_DOWN:
		(void) printf("FOR\n");
	        node_print ( pnode->un.bdfor.var, indent+2, "var " );
	        node_print ( pnode->un.bdfor.from, indent+2, "from" );
	        node_print ( pnode->un.bdfor.to, indent+2, "to  " );
	        node_print ( pnode->un.bdfor.by, indent+2, "by  " );
	        node_print ( pnode->un.bdfor.stmt, indent+2, "stmt" );
	        break;

	    case T_CASE:
		(void) printf("CASE\n");
	        node_print ( pnode->un.bdcase.expr, indent+2, "cexp" );
	        node_print ( pnode->un.bdcase.stmt, indent+2, "stmt" );
		break;

	    case T_SELECT: case T_SELECTONE: case T_SELECTALL:
		(void) printf("SELECT\n");
	        node_print ( pnode->un.select.expr, indent+2, "sexp" );
	        node_print ( pnode->un.select.bdcase, indent+2, "case" );
		break;

	    case T_FCALL:
		(void) printf( "FCALL %s<%d>\n", pnode->un.fcall.def->name,
			pnode->un.fcall.hbit );
	        node_print ( pnode->un.fcall.actual_args, indent+2, "args" );
		break;

	    case T_OTHERWISE:
		(void) printf("OTHERWISE\n");
		break;

	    case T_LIF:
		(void) printf("LIF\n%s%s", pnode->un.lif.names, pnode->un.lif.table);
		break;

	    default:
		(void) printf("%d",pnode->type );
		ierror("Type unknown in node_print",pnode);
        }

	if (pnode->conc != NULL)
	    node_print( pnode->conc, indent, " &  " );
	if (pnode->next != NULL)
	    node_print( pnode->next, indent, " >  " );
    }
}


/* This routine is like node_print.  However, it only accepts
 * expressions.  It will print a complex expression on a single line.
 */
void
node_print_expr( pnode )
node *pnode;
{
    if ( pnode == NULL ) {

	(void) printf("NULL");

    } else {

        switch ( pnode->type ) {

	    case T_NOT: case T_BUF: case T_WIDTH: case T_UNARY_OR:
	    case T_UNARY_AND: case T_UNARY_PLUS: case T_UNARY_MINUS:
	        (void) printf("%s(", map_token_to_name(pnode->type));
	        node_print_expr(pnode->un.unary.expr);
	        (void) printf(")");
	        break;

	    case T_AND: case T_OR: case T_NAND: case T_NOR: case T_XOR: case T_EQV:
	    case T_EQL: case T_NEQ: case T_LEQ: case T_GEQ: case T_LSS: case T_GTR:
	    case T_SL0: case T_SR0: case T_SL1: case T_SR1: case T_SLR: case T_SRR:
	    case T_SXT: case T_ZXT: case T_OXT:
	    case T_PLUS: case T_MINUS: case T_TIMES: case T_DIV: case T_MOD:
	        (void) printf("(");
	        node_print_expr( pnode->un.binary.lexpr);
	        (void) printf(" %s ", map_token_to_name(pnode->type));
		if( pnode->type == T_OR )
		    (void) printf("\n        ");
	        node_print_expr ( pnode->un.binary.rexpr);
	        (void) printf(")");
	        break;

	    case T_BIT:
		node_print_expr ( pnode->un.ternary.lexpr);
		(void) printf("<");
		node_print_expr ( pnode->un.ternary.mexpr);
		(void) printf(":");
		node_print_expr ( pnode->un.ternary.rexpr);
		(void) printf(">");
		break;

	    case T_EQ:
		node_print_expr ( pnode->un.assign.lexpr);
		(void) printf(" = ");
		node_print_expr ( pnode->un.assign.rexpr);
		(void) printf("\n\n");
	        break;

	    case T_IS_ARG:
	    case T_IS_STATE:
		if (pnode->un.bdvar.rev == NIL(usage)) {
		    (void) printf("%s<%d>", pnode->un.bdvar.name->name,
			    pnode->un.bdvar.hbit);
		} else {
		    (void) printf("%s<%d>%d.%d", pnode->un.bdvar.name->name,
			    pnode->un.bdvar.hbit,
			    pnode->un.bdvar.rev->block->id,
			    pnode->un.bdvar.rev->rev);
		}
	        break;

	    case T_IS_CONSTANT:
		if (pnode->un.constant.dont_care) {
		    (void) printf("?");
		} else {
		    (void) printf("%d", pnode->un.constant.num);
		    /* XXX used to have: pnode->un.constant.hbit */
		}
	        break;

	    case T_FCALL:
		(void) printf("FCALL %s\n",pnode->un.fcall.def->name);
	        node_print_expr ( pnode->un.fcall.actual_args);
		break;

	    case T_LIF:
		(void) printf("LIF\n%s\n%s", pnode->un.lif.names, pnode->un.lif.table);
		break;

	    default:
		(void) fprintf(stderr,"%d",pnode->type);
	        ierror("Non expr in node_print_expr",pnode);

        }
        if (pnode->conc != NULL) {
	    (void) printf(" & ");
	    node_print_expr ( pnode->conc );
	}
	if (pnode->next != NULL) {
	    node_print_expr ( pnode->next );
        }
    }
}


/*
 * This will take a node and pipe the correct input to mlespresso
 * flattener.  It returns the name of the output from its LIF
 * table.
 */
char *node_lif( root, varname, bit, assign_num )
node *root;
char *varname;			/* name of lhs */
int bit;			/* bit of lhs */
int assign_num;			/* What assignment statement are we in */
{
    int numops;			/* number of ops in ORR or ANDR */
    node *op;			/* opererand of ORR or ANDR */
    char myname[1024];		/* temporary name of a function */
    char *opname;		/* name of operands to functions */
    char *op2name;
    char *namelist;		/* dynamically alloced .names line */
    symbol *symbolent;		/* symbol table entry */
    int i;			/* index */
    node *lastx;		/* pointer to hashed rhs for forward subst */
    int totlen;			/* length of .names string */

    switch ( root->type ) {
	
	case T_IS_STATE:
	    /* set up default name */
	    symbolent = root->un.bdvar.name;
	    if (root->un.bdvar.rev == NIL(usage)) {
   		    if (short_1_bit && symbolent->u.state.nobit) {
   		        (void) sprintf( myname, "%s", symbolent->name );
   		    }
			else { 
				(void) sprintf( myname,"%s<%d>",
					symbolent->name, root->un.bdvar.lbit);
			}
	    } else {
		(void) sprintf( myname,"%s<%d>%d.%d", symbolent->name,
			root->un.bdvar.lbit,
			root->un.bdvar.rev->block->id,
			root->un.bdvar.rev->rev);
	    }

	    /* process Dont_care */
	    if ( upper_strcmp( symbolent->name, "DONT_CARE" ) == 0 ) {
		(void) strcpy( myname, "DONT_CARE" );

#ifdef FOO
/* For now we will keep "temporaries" -- they are valuable to mis */

	    /* check if temporary var => should be collapsed out */
	    /* temporary variable are used in routine param passing,
	     * and complicated IF and SELECT conditions ("$$COND") */
	    } else if (symbolent->u.state.is_temp &&
		    (root->un.bdvar.rev != NIL(usage)) ) {
		lastx = root->un.bdvar.rev->assign;

		if (lastx->un.assign.lifed != assign_num ) {
		    rhs = lastx->un.assign.rexpr;
		    if ( rhs->type == T_LIF ) {
			(void) fputs( rhs->un.lif.names, bdsyntemp );
			(void) fputs( rhs->un.lif.table, bdsyntemp );
			if ( watch_espresso >= 3 ) {
			    (void) printf( "%s%s", rhs->un.lif.names,
				    rhs->un.lif.table );
			}
		    } else {
			opname = node_lif( rhs, varname, bit, assign_num );
			(void) fputs( ".names ", bdsyntemp );
			(void) fputs( opname, bdsyntemp );
			putc( ' ', bdsyntemp );
			(void) fputs( myname, bdsyntemp );
			(void) fputs( "\n1 1\n", bdsyntemp );
			if ( watch_espresso >=3 ) {
			    (void) printf( ".names %s %s\n1 1\n", opname, myname );
			}
			FREE( opname );
		    }
		    lastx->un.assign.lifed = assign_num;
		}
#endif

	    /* check for forward subst condition */
	    } else if ((upper_strcmp(varname, symbolent->name) == 0) &&
		    (root->un.bdvar.rev != NIL(usage)) &&
		    (root->un.bdvar.lbit == bit)) {
		lastx = root->un.bdvar.rev->assign;

		/*
		 * If lifed = assign_num, we've already sent this
		 * subtree to ml.  If the lastx has a needlif, the LIF
		 * table is done, and we just point to it.
		 */
		if ((lastx->un.assign.lifed != assign_num) &&
			!lastx->un.assign.needlif) {
		    opname = node_lif(lastx->un.assign.rexpr, varname,
			    bit, assign_num);
		    (void) fputs( ".names ", bdsyntemp );
		    (void) fputs( opname, bdsyntemp );
		    putc( ' ', bdsyntemp );
		    (void) fputs( myname, bdsyntemp );
		    (void) fputs( "\n1 1\n", bdsyntemp );
		    if ( watch_espresso >=3 ) {
			(void) printf(".names %s %s\n1 1\n", opname, myname);
		    }
		    FREE(opname);
		    lastx->un.assign.lifed = assign_num;
		}
	    
	    } else {
		if ((symbolent->type == T_IS_STATE) ||
			(symbolent->type == T_IS_OUTPUT_PORT)) {
		    if (root->un.bdvar.rev == NIL(usage)) {
			if (make_unspec_set) {
			    (void) strcat( myname, "**" );
			} else {
			    /* set unspecified variable = 0 */
			    (void) sprintf(myname,"_U%d", functnum++);
			    (void) fputs( ".names ", bdsyntemp );
			    (void) fputs( myname, bdsyntemp );
			    putc( '\n', bdsyntemp );
			    if ( watch_espresso >=4 ) {
				(void) printf( ".names %s\n", myname );
			    }
			}
		    } else if ( (symbolent->type == T_IS_OUTPUT_PORT) &&
			    islastupdate( root ) ) {
			/* specified output port */
   			    if (short_1_bit && symbolent->u.state.nobit) {
   			        (void) sprintf( myname, "%s", symbolent->name );
   			    }
				else { 
					(void) sprintf( myname,"%s<%d>",
						symbolent->name, root->un.bdvar.lbit);
				}
		    } else {
			/* specified intermediate variable */
			/* just pass forward "myname" */
		    }
		} else {
		    /* input port */
		    /* just pass forward "myname" */
		}
	    }
	    break;

        case T_AND: case T_OR: case T_NAND:
	case T_NOR: case T_XOR: case T_EQV:
	    opname = node_lif( root->un.binary.lexpr, varname, bit,assign_num );
	    op2name = node_lif( root->un.binary.rexpr, varname, bit,assign_num);
	    (void) sprintf(myname,"_B%d", functnum++);
	    (void) fputs( ".names ", bdsyntemp );
	    (void) fputs( opname, bdsyntemp );
	    putc( ' ', bdsyntemp );
	    (void) fputs( op2name, bdsyntemp );
	    putc( ' ', bdsyntemp );
	    (void) fputs( myname, bdsyntemp );
	    putc( '\n', bdsyntemp );
	    if ( watch_espresso >=4 ) {
		(void) printf(".names %s %s %s\n", opname, op2name, myname);
	    }
	    switch ( root->type ) {
		case T_AND:
		    (void) fputs( "11 1\n", bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf("11 1\n");
		    }
		    break;
		case T_OR:
		    (void) fputs( "1- 1\n-1 1\n", bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf("1- 1\n-1 1\n");
		    }
		    break;
		case T_NAND:
		    (void) fputs( "0- 1\n-0 1\n", bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf("0- 1\n-0 1\n");
		    }
		    break;
		case T_NOR:
		    (void) fputs( "00 1\n", bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf("00 1\n");
		    }
		    break;
		case T_EQV:
		    (void) fputs( "11 1\n00 1\n", bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf("11 1\n00 1\n");
		    }
		    break;
		case T_XOR:
		    (void) fputs( "10 1\n01 1\n", bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf("10 1\n01 1\n");
		    }
		    break;
	    }
	    FREE( opname );
	    FREE( op2name );
	    break;

        case T_NOT:
	    opname = node_lif(root->un.unary.expr, varname, bit,assign_num);
	    (void) sprintf( myname,"_N%d", functnum++ );
	    (void) fputs( ".names ", bdsyntemp );
	    (void) fputs( opname, bdsyntemp );
	    putc( ' ', bdsyntemp );
	    (void) fputs( myname, bdsyntemp );
	    (void) fputs( "\n0 1\n", bdsyntemp );
	    if ( watch_espresso >=4 ) {
		(void) printf(".names %s %s\n0 1\n", opname, myname );
	    }
	    FREE( opname );
	    break;

        case T_UNARY_OR: case T_UNARY_AND:
	    numops = 0;
	    namelist = ALLOC(char, 256);	/* init namelist is 256 chars */
	    (void) strcpy( namelist,".names");
	    totlen = strlen(".names");
	    for( op = root->un.unary.expr; op != NULL; op = op->conc ) {
		opname = node_lif( op, varname, bit, assign_num );
		namelist = REALLOC(char, namelist,
			totlen + strlen( opname ) + 2);
		*(namelist + totlen) = ' ';
		(void) strcpy( namelist + totlen + 1, opname );
		totlen += strlen( opname ) + 1;
		FREE( opname );
		numops++;
	    }
	    (void) sprintf( myname,"_M%d",functnum++ );
	    (void) fputs( namelist, bdsyntemp );
	    putc( ' ', bdsyntemp );
	    (void) fputs( myname, bdsyntemp );
	    putc( '\n', bdsyntemp );
	    if ( watch_espresso >=4 ) {
		(void) printf("%s %s\n", namelist, myname );
	    }
	    FREE( namelist );
	    for( i=1; i <=numops ; i++ ) {
		if ( root->type == T_UNARY_AND ) {
		    putc( '1', bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf( "1" );
		    }
		} else {
		    putc( '0', bdsyntemp );
		    if ( watch_espresso >=4 ) {
			(void) printf( "0" );
		    }
		}
	    }
	    if ( root->type == T_UNARY_AND ) {
		(void) fputs( " 1\n", bdsyntemp );
		if ( watch_espresso >=4 ) {
		    (void) printf( " 1\n" );
		}
	    } else {
		(void) fputs( " 0\n", bdsyntemp );
		if ( watch_espresso >=4 ) {
		    (void) printf( " 0\n" );
		}
	    }
	    break;

	case T_IS_CONSTANT:
	    if (root->un.constant.dont_care )
		error("Wildcard not used in compare",root);
	    (void) sprintf( myname,"_C%d",functnum++);
	    if (root->un.constant.num == 0) {
		(void) fputs( ".names ", bdsyntemp );
		(void) fputs( myname, bdsyntemp );
		putc( '\n', bdsyntemp );
		if ( watch_espresso >=4 ) {
		    (void) printf( ".names %s\n", myname );
		}
	    } else {
		(void) fputs( ".names ", bdsyntemp );
		(void) fputs( myname, bdsyntemp );
		(void) fputs( "\n1\n", bdsyntemp );
		if ( watch_espresso >=4 ) {
		    (void) printf( ".names %s\n1\n", myname );
		}
	    }
	    break;
	
	case T_LIF:
	    ierror( "node_lif on LIF", root );
	    break;

	default:
	    ierror("node_lif on ?", root);

    }
    return util_strsav(myname);
}
