
/* RCS Info: $Revision: 1.1.1.1 $ on $Date: 2002/12/31 15:50:54 $
 *           $Source: /home/cvs/octtools/src/Synthesis/misII/graphics/build.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "utility.h"
#include "graphics_int.h"

void free_node();

dg_graph *
dg_alloc(title)
    char *title;
{
    dg_graph *graph;

    graph = ALLOC(dg_graph, 1);

    if (title)
        graph->title = util_strsav(title);

    graph->nodes = lsCreate();
    graph->arcs = lsCreate();

    return (graph);
}

void
dg_free(graph)
    dg_graph *graph;
{
    (void) lsDestroy(graph->nodes, free_node);

    assert(lsLength(graph->arcs) == 0);

    (void) lsDestroy(graph->arcs, (void (*)()) NULL);

    if (graph->title)
        FREE(graph->title);

    FREE(graph);

    return;
}



dg_node *
dg_create_node(graph)
    dg_graph *graph;
{
    dg_node *node;

    node = ALLOC(dg_node, 1);

    (void) lsNewEnd(graph->nodes, (lsGeneric) node,  (lsHandle *) NULL);

    node->graph = graph;

    node->type = DG_UNKNOWN;
    node->color = util_strsav(DEF_COLOR);
    node->shape = DEF_SHAPE;
    node->font = util_strsav(DEF_FONT);
    node->numlines = 0;
    node->lines = NULL;
    node->fanins = lsCreate();
    node->fanouts = lsCreate();

    return (node);
}

/* ARGSUSED */
static lsStatus
killarc(arc, data)
    lsGeneric arc;
    lsGeneric data;
{
    dg_delete_arc((dg_arc *) arc);

    return (LS_OK);
}

static void
free_lines(node)
    dg_node *node;
{
    int i;

    if (node->lines) {
        for (i = 0; i < node->numlines; i++)
            FREE(node->lines[i]);
        FREE(node->lines);
        node->lines = NULL;
        node->numlines = 0;
    }

    return;
}

void
dg_delete_node(node)
    dg_node *node;
{
    dg_graph *graph = node->graph;

    free_node(node);
    (void) lsDelValue(graph->nodes, (lsGeneric) node);

    return;
}

void
free_node(node)
    dg_node *node;
{
    (void) lsForeach(node->fanins, killarc, (lsGeneric) NULL);
    (void) lsForeach(node->fanouts, killarc, (lsGeneric) NULL);

    assert(lsLength(node->fanins) == 0);
    assert(lsLength(node->fanouts) == 0);

    (void) lsDestroy(node->fanins, (void (*)()) NULL);
    (void) lsDestroy(node->fanouts, (void (*)()) NULL);

    FREE(node->color);
    FREE(node->font);
    free_lines(node);
    
    FREE(node);

    return;
}



dg_arc *
dg_create_arc(from, to)
    dg_node *from;
    dg_node *to;
{
    dg_arc *arc;

    arc = ALLOC(dg_arc, 1);

    assert(from->graph == to->graph);

    arc->graph = from->graph;
    arc->from = from;
    arc->to = to;

    (void) lsNewEnd(arc->graph->arcs, (lsGeneric) arc, (lsHandle *) NULL);

    (void) lsNewEnd(from->fanouts, (lsGeneric) arc, (lsHandle *) NULL);

    (void) lsNewEnd(to->fanins, (lsGeneric) arc, (lsHandle *) NULL);

    arc->color = util_strsav(DEF_COLOR);
    arc->style = DEF_LINESTYLE;
    arc->label = NULL;

    return (arc);
}

void
dg_delete_arc(arc)
    dg_arc *arc;
{
    (void) lsDelValue(arc->from->fanouts, (lsGeneric) arc);
    (void) lsDelValue(arc->to->fanins, (lsGeneric) arc);

    FREE(arc->color);

    (void) lsDelValue(arc->graph->arcs, (lsGeneric) arc);

    FREE(arc);

    return;
}



void
dg_set_node_props(node, type, color, shape, font, numlines, lines)
    dg_node *node;
    dg_nodetype type;
    char *color;
    dg_nodeshape shape;
    char *font;
    int numlines;
    char **lines;
{
    int i;

    node->type = type;

    FREE(node->color);
    node->color = util_strsav(color);

    node->shape = shape;

    FREE(node->font);
    node->font = util_strsav(font);

    free_lines(node);
    if (lines) {
        node->lines = ALLOC(char *, numlines);
        for (i = 0; i < numlines; i++)
            node->lines[i] = util_strsav(lines[i]);

        node->numlines = numlines;
    }

    return;
}

void
dg_get_node_props(node, type, color, shape, font, numlines, lines)
    dg_node *node;
    dg_nodetype *type;
    char **color;
    dg_nodeshape *shape;
    char **font;
    int *numlines;
    char ***lines;
{
    *type = node->type;
    *color = node->color;
    *shape = node->shape;
    *font = node->font;
    *numlines = node->numlines;
    *lines = node->lines;

    return;
}



void
dg_set_arc_props(arc, color, style, label)
    dg_arc *arc;
    char *color;
    dg_linestyle style;
    char *label;
{
    FREE(arc->color);
    arc->color = util_strsav(color);

    arc->style = style;

    if (arc->label)
        FREE(arc->label);
    if (label)
        arc->label = util_strsav(label);
    else
        arc->label = NULL;

    return;
}

void
dg_get_arc_props(arc, color, style, label)
    dg_arc *arc;
    char **color;
    dg_linestyle *style;
    char **label;
{
    *color = arc->color;
    *style = arc->style;
    *label = arc->label;

    return;
}



void
dg_dump(graph)
    dg_graph *graph;
{
    lsGen gen, gen2;
    dg_node *node;
    dg_arc *arc;
    int i;

    (void)  printf("Graph: %s\n", graph->title);
    (void)  printf("Nodes:\n");
    dg_foreach_node(graph, gen, node) {
        (void)  printf("\t%lx: type %s, color %s, shape %s, font %s, text: ",
                (int) node, dg_type_name(node->type),
                node->color, dg_shape_name(node->shape),
                node->font);
        if (node->numlines == 0)
            (void)  printf("<NULL>\n");
        else if (node->numlines == 1)
            (void)  printf("%s\n", node->lines[0]);
        else {
            (void)  printf("\n");
            for (i = 0; i < node->numlines; i++)
                (void)  printf("\t\t%s\n", node->lines[i]);
        }

        (void)  printf("\t\tlevel = %d, vertpos = %d, fanins %d, fanouts %d\n",
                node->level, node->vertpos, num_fanins(node),
                num_fanouts(node));
        (void)  printf("\t\tfanouts:");
        dg_foreach_fanout(node, gen2, arc)
            (void)  printf(" %lx", (int) arc->to);
        (void)  printf("\n\t\tfanins:");
        dg_foreach_fanin(node, gen2, arc)
            (void)  printf(" %lx", (int) arc->from);
        (void)  printf("\n");
    }
    (void)  printf("\nArcs:\n");
    dg_foreach_arc(graph, gen, arc) {
        (void)  printf("\t%lx: %lx -> %lx, color %s, style %s, label %s\n",
                (int) arc, (int) arc->from, (int) arc->to,
                arc->color, dg_style_name(arc->style),
                arc->label);
    }
    return;
}



lsStatus
lsDelValue(list, data)
    lsList list;
    lsGeneric data;
{
    lsGen gen;
    lsGeneric this;

    gen = lsStart(list);

    while (lsNext(gen, &this, (lsHandle *) NULL) == LS_OK)
        if (this == data)
            (void) lsDelBefore(gen, &this);
    
    (void) lsFinish(gen);

    return (LS_OK);
}
