/* RCS Info: $Revision: 1.1.1.1 $ on $Date: 2002/12/31 15:50:54 $
 *           $Source: /home/cvs/octtools/src/Synthesis/misII/graphics/clipshape.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by  Rick L. Spickelmier
 *
 */

#include "port.h"
#include "graphics_int.h"
#include <math.h>

extern int line_intersect(), line_intersect_ellipse();

void
dg_clipshape(shape, width, height, fromx, fromy, toxp, toyp)
    dg_nodeshape shape;
    int width, height;
    int fromx, fromy;
    int *toxp, *toyp;
{
    int tox = *toxp, toy = *toyp;
    double dm = (double) height / width;            
    double m = (double) (fromy - toy) / (fromx - tox);

    switch (shape) {
        case DG_SQUARE:
        case DG_RECTANGLE:
        if ((m > dm) || (m < - dm)) {
            /* Either top or bottom. */
            if (fromy > toy) {
                /* Top. */
                assert(line_intersect(fromx, fromy, tox, toy, tox - width / 2, toy + height / 2, tox + width / 2, toy + height / 2, toxp, toyp));
            } else {
                /* Bottom. */
                assert(line_intersect(fromx, fromy, tox, toy, tox - width / 2, toy - height / 2, tox + width / 2, toy - height / 2, toxp, toyp));
            }
        } else {
            /* Either left or right. */
            if (fromx < tox) {
                /* Left. */
                assert(line_intersect(fromx, fromy, tox, toy, tox - width / 2, toy - height / 2, tox - width / 2, toy + height / 2, toxp, toyp));
            } else {
                /* Right. */
                assert(line_intersect(fromx, fromy, tox, toy, tox + width / 2, toy - height / 2, tox + width / 2, toy + height / 2, toxp, toyp));
            }
        }
        break;

        case DG_CIRCLE:
        case DG_ELLIPSE:
        assert(line_intersect_ellipse(fromx, fromy, tox, toy, width, height, toxp, toyp));
        if (((fromx < tox) && ((*toxp > tox) || (*toxp < fromx))) ||
                ((fromx > tox) && ((*toxp < tox) ||
                (*toxp > fromx)))) {
            *toxp -= (*toxp - tox);
            *toyp -= (*toyp - toy);
        }
        
        break;

        case DG_TRIANGLE:
        break;

        case DG_DIAMOND:
        break;

        case DG_NO_SHAPE:
        break;
    
        default:
        assert(0);
    }

    return;
}

static int
line_intersect(x1_a, y1_a, x2_a, y2_a, x1_b, y1_b, x2_b, y2_b, xp, yp)
    int x1_a, y1_a;
    int x2_a, y2_a;
    int x1_b, y1_b;
    int x2_b, y2_b;
    int *xp, *yp;
{
    double ma = (double) (y2_a - y1_a) / (x2_a - x1_a);
    double mb = (double) (y2_b - y1_b) / (x2_b - x1_b);
    int ba = y1_a - ma * x1_a;
    int bb = y1_b - mb * x1_b;

    if (fabs(ma - mb) < 0.00001)
        return (FALSE);

    *xp = (bb - ba) / (ma - mb);
    *yp = ma * *xp + ba;

    return (TRUE);
}

static int
line_intersect_ellipse(fromx, fromy, tox, toy, width, height, toxp, toyp)
    int fromx, fromy;
    int tox, toy;
    int width, height;
    int *toxp, *toyp;
{
    double alpha = atan2((double) fromx - tox, (double) fromy - toy);
    int x = (width / 2) * cos(alpha);
    int y = (height / 2) * sin(alpha);

    *toxp = tox + x;
    *toyp = toy + y;

    return (TRUE);
}
