
#ifndef NO_X

/* RCS Info: $Revision: 1.1.1.1 $ on $Date: 2002/12/31 15:50:54 $
 *           $Source: /home/cvs/octtools/src/Synthesis/misII/graphics/curve.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "utility.h"
#include <X11/Xlib.h>

#ifdef notdef
static void do_spline();

int
dg_curve_verts(verts, vcount, nvlist)
    Vertex *verts;
    int vcount;
    Vertex **nvlist;
{
    int diff, x, i, j, k;
    Vertex *newverts;
    double a, b, c, d;
    double slope1, slope2;
    int pos, numnewverts;

    /* How many vertices do we need? */
    numnewverts = 0;
    for (i = 0; i < vcount - 1; i++)
        if (verts[i].flags & VertexCurved) {
            diff = verts[i].x - verts[i + 1].x;
            if (diff < 0)
                diff = - diff;

            numnewverts += diff / 10 + 1;
        } else
            numnewverts += 1;
    numnewverts++;
    
    newverts = ALLOC(Vertex, numnewverts);

    pos = 0;
    for (i = 0; i < vcount - 1; i++) {
        if (verts[i].flags & VertexCurved) {
            if (i == 0) {
                if (verts[1].x == verts[0].x)
                    slope1 = 0;
                else
                    slope1 = (double) (verts[1].y -
                        verts[0].y) /
                        (verts[1].x - verts[0].x);
            } else {
                if (newverts[pos - 1].x == verts[i].x)
                    slope1 = 0;
                else
                    slope1 = (double) (newverts[pos - 1].y -
                        verts[i].y)
                        / (newverts[pos - 1].x -
                        verts[i].x);
            }
            if (i == vcount - 2)
                slope2 = 0;
            else {
                if (verts[i + 2].x == verts[i + 1].x)
                    slope2 = 0;
                else
                    slope2 = (double) (verts[i + 2].y -
                        verts[i + 1].y) /
                        (verts[i + 2].x -
                        verts[i + 1].x);
            }
            do_spline(verts[i].x, verts[i + 1].x, verts[i].y,
                    verts[i + 1].y, slope1, slope2, &a, &b,
                    &c, &d);

            newverts[pos] = verts[i];
            newverts[pos].flags &= ~VertexCurved;
            pos++;

            diff = verts[i + 1].x - verts[i].x;
            k = ((diff > 0) ? diff : - diff) / 10;
            x = verts[i].x;
            for (j = 0; j < k; j++) {
                x += diff / 10;
                newverts[pos].x = x;
                newverts[pos].y = a * x * x * x + b * x * x +
                        c * x + d;
                newverts[pos].flags = verts[i].flags &
                        ~VertexCurved;
                pos++;
            }
        } else {
            newverts[pos] = verts[i];
            newverts[pos].flags &= ~VertexCurved;
            pos++;
        }
    }
    newverts[pos] = verts[i];
    newverts[pos].flags &= ~VertexCurved;
    pos++;

    assert(pos == numnewverts);

    *nvlist = newverts;
    return (numnewverts);
}
#endif

static void
do_spline(x1, x2, y1, y2, y1_p, y2_p, ap, bp, cp, dp)
    int x1, x2;
    int y1, y2;
    double y1_p, y2_p;
    double *ap, *bp, *cp, *dp;
{
    double **matrix;
    double *vals;
    int i;

    matrix = ALLOC(double *,4);
    for (i = 0; i < 4; i++)
        matrix[i] = ALLOC(double, 4);
    vals = ALLOC(double, 4);

    matrix[0][0] = x1 * x1 * x1;
    matrix[0][1] = x1 * x1;
    matrix[0][2] = x1;
    matrix[0][3] = 1;

    matrix[1][0] = x2 * x2 * x2;
    matrix[1][1] = x2 * x2;
    matrix[1][2] = x2;
    matrix[1][3] = 1;

    matrix[2][0] = 3 * x1 * x1;
    matrix[2][1] = 2 * x1;
    matrix[2][2] = 1;
    matrix[2][3] = 0;

    matrix[3][0] = 3 * x2 * x2;
    matrix[3][1] = 2 * x2;
    matrix[3][2] = 1;
    matrix[3][3] = 0;

    vals[0] = y1;
    vals[1] = y2;
    vals[2] = y1_p;
    vals[3] = y2_p;

    dg_solvemat(matrix, vals, 4);

    *ap = vals[0];
    *bp = vals[1];
    *cp = vals[2];
    *dp = vals[3];

    FREE(vals);
    for (i = 0; i < 4; i++)
        FREE(matrix[i]);
    FREE(matrix);

    return;
}

#endif /* NO_X */

