/* RCS Info: $Revision: 1.1.1.1 $ on $Date: 2002/12/31 15:50:54 $
 *           $Source: /home/cvs/octtools/src/Synthesis/misII/graphics/graphics_int.h,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *	 faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "graphics.h"
#include "list.h"
#include "st.h"
#include "utility.h"

struct dg_graph_s {
	char *title;
	lsList nodes;
	lsList arcs;
	st_table x_colors;
	st_table x_fonts;
	st_table x_patterns;
} ;

struct dg_node_s {
	dg_graph *graph;
	dg_nodetype type;
	char *color;
	dg_nodeshape shape;
	char *font;
	int numlines;
	char **lines;
	lsList fanins;
	lsList fanouts;
	int level;
	int vertpos;
	int grids_high;
	int width, height;
} ;

struct dg_arc_s {
	dg_graph *graph;
	dg_node *from;
	dg_node *to;
	char *color;
	dg_linestyle style;
	char *label;
} ;

typedef struct dg_outdev_s {
	dg_plot *(*open)();
	void (*calc_sizes)();
	void (*draw_node)();
	void (*draw_arc)();
	void (*update)();
	void (*close)();
} dg_outdev;

#define DEF_COLOR	"black"
#define DEF_SHAPE	DG_SQUARE
#define DEF_FONT	"6x10"
#define DEF_LINESTYLE	DG_SOLID
#ifndef NULL
#define NULL		0
#endif
#ifndef TRUE
#define TRUE		1
#endif
#ifndef FALSE
#define FALSE		0
#endif

#define dg_lsForeachItem(list, gen, data)				\
	for (gen = lsStart(list);					\
		(lsNext(gen, (lsGeneric *) &data, LS_NH) == LS_OK)	\
		|| ((void) lsFinish(gen), 0); )

#define dg_foreach_node(graph, gen, node)	\
	dg_lsForeachItem((graph)->nodes, gen, node)

#define dg_foreach_arc(graph, gen, arc)	\
	dg_lsForeachItem((graph)->arcs, gen, arc)

#define dg_foreach_fanin(node, gen, arc)	\
	dg_lsForeachItem((node)->fanins, gen, arc)

#define dg_foreach_fanout(node, gen, arc)	\
	dg_lsForeachItem((node)->fanouts, gen, arc)

#define num_fanins(node)	lsLength((node)->fanins)

#define num_fanouts(node)	lsLength((node)->fanouts)

#define eq(str1, str2) (strcmp((str1), (str2)) == 0)

extern void XBox();
extern void XBoxFilled();
extern void XEllipse();
extern void XEllipseFilled();
extern void XDiamond();
extern void XDiamondFilled();
extern void XTriangle();
extern void XTriangleFilled();

extern void dg_level();
extern void dg_place();
extern void dg_make_uniform();
extern void dg_dump();

extern void dg_setdev();
extern dg_plot *dg_open();
extern void dg_draw_node();
extern void dg_draw_arc();
extern void dg_update();
extern void dg_close();

extern int dg_debug;

extern char *dg_type_name();
extern char *dg_style_name();
extern char *dg_out_name();
extern char *dg_shape_name();

extern dg_nodetype dg_parse_type();
extern dg_linestyle dg_parse_style();
extern dg_outtype dg_parse_out();
extern dg_nodeshape dg_parse_shape();

extern void dg_solvemat();

extern void dg_calc_sizes();

#define DEF_CONSTANT_COLOR	"black"
#define DEF_CONSTANT_FONT	"6x10"
#define DEF_CONSTANT_SHAPE	DG_DIAMOND

#define DEF_INPUT_COLOR		"red"
#define DEF_INPUT_FONT		"6x10"
#define DEF_INPUT_SHAPE		DG_ELLIPSE

#define DEF_OUTPUT_COLOR	"green"
#define DEF_OUTPUT_FONT		"6x10"
#define DEF_OUTPUT_SHAPE	DG_ELLIPSE

#define DEF_FUNCTION_COLOR	"blue"
#define DEF_FUNCTION_FONT	"6x10"
#define DEF_FUNCTION_SHAPE	DG_RECTANGLE

#define DEF_ARC_COLOR		"black"
#define DEF_ARC_STYLE		DG_SOLID
#define DEF_ARC_LABEL		NULL
