/* RCS Info: $Revision: 1.1.1.1 $ on $Date: 2002/12/31 15:50:54 $
 *           $Source: /home/cvs/octtools/src/Synthesis/misII/graphics/place.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "utility.h"
#include "graphics_int.h"

/* Here's the real stuff.  Assign relative positions of the nodes in all
 * the levels in the graph.  Assume the order of column 0 is
 * fixed.  Then determine the best posttions for column 1 based on this,
 * then column 2, etc.  The primary inputs are fixed also.  The go back
 * towards the primary outputs, using the orders of both the previous and
 * next columns.
 */

extern int assignPositions();
extern int nodeCompareFanouts(), nodeCompareBoth();
extern int nodeHeight();

void
dg_place(graph)
    dg_graph *graph;
{
    int level, i;
    lsGen gen;
    dg_node *node;
    int maxverts = 0;
    int maxlevel = 0;

    dg_foreach_node(graph, gen, node) {
        if (node->level > maxlevel)
            maxlevel = node->level;
        node->grids_high = nodeHeight(node);
    }

    for (level = 0; level < maxlevel; level++) {
        i = 0;
        dg_foreach_node(graph, gen, node)
            if (node->level == level)
                i += node->grids_high;
        if (i > maxverts)
            maxverts = i;
    }

    /* First go from the outputs to the inputs... */
    maxverts = assignPositions(graph, 0, maxverts, 1, 1);
    for (i = 1; i < maxlevel; i++)
        maxverts = assignPositions(graph, i, maxverts, 1, 0);
    maxverts = assignPositions(graph, maxlevel, maxverts, 1, 1);

    /* ... and now from the inputs back to the outputs. */
    for (i = maxlevel - 1; i > 0; i--)
        maxverts = assignPositions(graph, i, maxverts, 0, 0);
    maxverts = assignPositions(graph, 0, maxverts, 0, 1);

    return;
}

/* ARGSUSED */
static int
assignPositions(graph, level, maxverts, prevonly, keeporder)
    dg_graph *graph;
    int level;
    int maxverts;
    int prevonly;
    int keeporder;
{
    lsList lev;
    lsGen gen, gen2;
    dg_node *node, *fanin, *fanout, *node2;
    dg_arc *arc;
    int *fixed;
    int i, j, num, avg, try, size;

    /* Fixed is 0 if there is nothing here, 1 if there are only
     * dummy nodes here, and 2 if there is a real node here.
     */
    fixed = ALLOC(int, maxverts);
    for (i = 0; i < maxverts; i++)
        fixed[i] = 0;

    lev = lsCreate();

    dg_foreach_node(graph, gen, node)
        if (node->level == level)
            (void) lsNewEnd(lev, (lsGeneric) node, (lsHandle *) NULL);

    if (prevonly)
        (void) lsSort(lev, nodeCompareFanouts);
    else
        (void) lsSort(lev, nodeCompareBoth);
    
    dg_lsForeachItem(lev, gen, node) {
        avg = 0;
        num = 0;
        dg_foreach_fanout(node, gen2, arc) {
            fanout = arc->to;
            avg += fanout->vertpos;
            num++;
        }
        if (!prevonly) {
            dg_foreach_fanin(node, gen2, arc) {
                fanin = arc->from;
                avg += fanin->vertpos * 2;
                num += 2;
            }
        }
        if (!num)
            avg = 0;
        else
            avg /= num;

        if (node->type == DG_DUMMY) {
            size = node->grids_high;
            if (!fits(fixed, size, avg, maxverts))
                size = 0;
            try = avg;
        } else {
            for (size = node->grids_high; size >= 0;
                    size -= (size > 1) ? 2 : 1) {
                try = findPlace(fixed, size, avg, maxverts);
                if (try >= 0)
                    break;
            }
            assert(try >= 0);
            if ((try == 0) && (node->grids_high > 1))
                try += node->grids_high / 2;
        }

        if (size < node->grids_high) {
            /* Make room... */
            i = node->grids_high - size;
            maxverts += i;
            fixed = REALLOC(int, fixed, maxverts);
            for (j = maxverts - 1; j > try + i; j--)
                fixed[j] = fixed[j - i];
            dg_lsForeachItem(lev, gen2, node2) {
                if (node2 == node) {
                    (void) lsFinish(gen2);
                    break;
                }
                if (node2->vertpos > try)
                    node2->vertpos += i;
            }
        }

        for (i = 0; i < node->grids_high; i++) {
            assert(try - node->grids_high / 2 + i >= 0);
            assert(try - node->grids_high / 2 + i < maxverts);
            fixed[try - node->grids_high / 2 + i] = 2;
        }

        node->vertpos = try + node->grids_high / 2;
    }

    (void) lsDestroy(lev, (void (*)()) NULL);
    FREE(fixed);

    return (maxverts);
}

static int
findPlace(fixed, size, avg, maxverts)
    int *fixed;
    int size;
    int avg;
    int maxverts;
{
    int i;

    if (fits(fixed, size, avg, maxverts))
        return (avg);
    else {
        for (i = 1; i < maxverts; i++) {
            if ((avg + i < maxverts) &&
                    fits(fixed, size, avg + i, maxverts))
                return (avg + i);
            if ((avg - i >= 0) &&
                    fits(fixed, size, avg - i, maxverts))
                return (avg - i);
        }
    }
    return (-1);
}

static int
fits(map, size, place, maxverts)
    int *map;
    int size;
    int place;
    int maxverts;
{
    int i;

    if (size == 0)
        return (1);

    if (place + size / 2 >= maxverts)
        return(0);

    for (i = 0; i < size; i++)
        if (map[place - size / 2 + i])
            return (0);
    
    return (1);
}

#define dw(n)    ((((dg_node *) n2)->type == DG_DUMMY) ? 0 : 0)

static int
nodeCompareFanouts(n1, n2)
    dg_node *n1, *n2;
{
    return ((num_fanouts(n2) + n2->grids_high + dw(n2)) -
        (num_fanouts(n1) + n1->grids_high + dw(n1)));
}

static int
nodeCompareBoth(n1, n2)
    dg_node *n1, *n2;
{
    return ((num_fanouts(n2) + num_fanins(n2) +
            n2->grids_high + dw(n2)) -
        (num_fanouts(n1) + num_fanins(n1) +
            n1->grids_high + dw(n1)));
}

void
dg_old_place(graph)
    dg_graph *graph;
{
    lsList lev, prev;
    int level, i;
    lsGen gen, gen2;
    dg_node *node, *fanin;
    dg_arc *arc;
    int maxverts = 0;
    int maxlevel = 0;

    dg_foreach_node(graph, gen, node)
        if (node->level > maxlevel)
            maxlevel = node->level;

    for (level = 0; level < maxlevel; level++) {
        i = 0;
        dg_foreach_node(graph, gen, node)
            if (node->level == level)
                i++;
        if (i > maxverts)
            maxverts = i;
    }

    i = 0;
    prev = lsCreate();
    dg_foreach_node(graph, gen, node)
        if (node->level == 0) {
            (void) lsNewEnd(prev, (lsGeneric) node, (lsHandle *) NULL);
            node->vertpos = i++;
        }

    for (level = 1; level < maxlevel; level++) {
        lev = lsCreate();

        dg_foreach_node(graph, gen, node)
            if (node->level == level) {
                (void) lsNewEnd(lev, (lsGeneric) node,
                        (lsHandle *) NULL);
                node->vertpos = -1;
            }
        
        i = 0;
        dg_lsForeachItem(prev, gen, node) {
            dg_foreach_fanin(node, gen2, arc) {
                fanin = arc->from;
                if (fanin->vertpos == -1)
                    fanin->vertpos = i++;
            }
        }

        dg_lsForeachItem(lev, gen, node)
            if (node->vertpos == -1)
                node->vertpos = i++;


        (void) lsDestroy(prev, (void (*)()) NULL);
        prev = lev;
    }

    (void) lsDestroy(prev, (void (*)()) NULL);

    return;
}

static int
nodeHeight(node)
    dg_node *node;
{
    int height;

    if (node->type == DG_DUMMY)
        return (1);
    
    height = node->numlines;

    switch (node->shape) {
         case DG_SQUARE:
         case DG_RECTANGLE:
        height += 1;
        break;

         case DG_CIRCLE:
         case DG_ELLIPSE:
        height += 2;
        break;

         case DG_TRIANGLE:
         case DG_DIAMOND:
        height += 2;
    }

    if (!(height % 2))
        height++;

    return (height);
}
