
#ifndef NO_X

/* RCS Info: $Revision: 1.1.1.1 $ on $Date: 2002/12/31 15:50:54 $
 *           $Source: /home/cvs/octtools/src/Synthesis/misII/graphics/xwindow.c,v $
 * Copyright (c) 1988 Wayne A. Christopher, U. C. Berkeley CAD Group 
 *     faustus@cad.berkeley.edu, ucbvax!faustus
 * Permission is granted to modify and re-distribute this code in any manner
 * as long as this notice is preserved.  All standard disclaimers apply.
 *
 * Converted to X11 by Rick L. Spickelmier
 *
 */

#include "port.h"
#include "utility.h"

#include "port.h"
#include <math.h>
#include <X11/Xlib.h>
#include "graphics_int.h"

#define X_PAD         4
#define Y_PAD         4

#define LABEL_FONT    "6x10"
#define TITLE_FONT    "9x15"

#define SQRT_2        1.414214

#define ARROW_ANGLE    (3.1415926 / 10)

#define ARROW_LENGTH    10

static struct xdata {
    Display *display;
    st_table *color_table;
    st_table *font_table;
    GC graphics_context;
    Colormap color_map;
    XFontStruct *label_font;
    XFontStruct *title_font;
    Window win;
    int maxx, maxy;
    int xlo, xhi, ylo, yhi;
    int xspacing, yspacing;
    /* current settings */
    XFontStruct *current_font;
    char *current_color;
    dg_linestyle current_style;
};

extern int shape_width();
void arrow();

static Display *display_opened = NULL;
static int refcount = 0;

#define x_trans(xd, xpos) (((xpos) - xd->xlo + 1) * xd->xspacing)
#define y_trans(xd, ypos) (xd->maxy - ((ypos) - xd->ylo + 1) * xd->yspacing)

static dg_plot *
open_win(window, arg, xlo, ylo, xhi, yhi, label)
    dg_plot *window;
    char *arg;
    int xlo, ylo;
    int xhi, yhi;
    char *label;
{
    struct xdata *xd;
    int screen;
    int i;
    /* dummy variables for XGetGeometry */
    Window root;
    int x, y;
    unsigned int border, depth;
    XEvent event;

    if (!window) {
        xd = ALLOC(struct xdata, 1);

        if (!display_opened) {
            display_opened = XOpenDisplay(arg);
            if (!display_opened) {
                (void) fprintf(stderr, "Error: can't open display '%s'\n",
                        arg ? arg : "NULL");
                return (NULL);
            }
        }

        xd->display = display_opened;
        refcount++;

        xd->color_table = st_init_table(strcmp, st_strhash);
        xd->font_table = st_init_table(strcmp, st_strhash);
	xd->current_color = "";
	xd->current_font = (XFontStruct *) 0;
	xd->current_style = DG_BAD_STYLE;

	screen = DefaultScreen(xd->display);
	xd->win = XCreateSimpleWindow(xd->display,
				      DefaultRootWindow(xd->display),
				      200, 200,
				      400, 400,
				      3,
				      BlackPixel(xd->display, screen),
				      WhitePixel(xd->display, screen));

	XStoreName(xd->display, xd->win, label ? label : "misII plot");

	xd->graphics_context = DefaultGC(xd->display, screen);
	xd->color_map = DefaultColormap(xd->display, screen);

        if (!(xd->label_font = XLoadQueryFont(xd->display, LABEL_FONT)))
            (void) fprintf(stderr, "Error: can't open font '%s'\n",
                    LABEL_FONT);

        if (!(xd->title_font = XLoadQueryFont(xd->display, TITLE_FONT)))
            (void) fprintf(stderr, "Error: can't open font '%s'\n",
                    TITLE_FONT);

        XMapRaised(xd->display, xd->win);

	/* wait for the exposure event */
	XSelectInput(xd->display, xd->win, ExposureMask);
	XNextEvent(xd->display, &event);

    } else {
        xd = (struct xdata *) window;
        XClearWindow(xd->display, xd->win);
    }

    XGetGeometry(xd->display, xd->win, &root, &x, &y, 
	    (unsigned int *) &xd->maxx, (unsigned int *) &xd->maxy,
	    (unsigned int *) &border, (unsigned int *) &depth);

    xd->xspacing = xd->maxx / (xhi - xlo + 2);
    xd->yspacing = xd->maxy / (yhi - ylo + 2);

    xd->xlo = xlo;
    xd->xhi = xhi;
    xd->ylo = ylo;
    xd->yhi = yhi;

    i = XTextWidth(xd->title_font, label, strlen(label));

    if (xd->current_font != xd->title_font) {
	XSetFont(xd->display, xd->graphics_context, xd->title_font->fid);
	xd->current_font = xd->title_font;
    }
    XDrawImageString(xd->display, xd->win, xd->graphics_context,
		     (xd->maxx - i) / 2,
		     (xd->title_font->max_bounds.ascent + xd->title_font->max_bounds.descent),
		     label, strlen(label));
    return ((dg_plot *) xd);
}

/* ARGSUSED */
static void
calc_sizes(win, shape, font, numlines, lines, widthp, heightp)
    dg_plot *win;
    dg_nodeshape shape;
    char *font;
    int numlines;
    char **lines;
    int *widthp, *heightp;
{
    struct xdata *xd = (struct xdata *) win;
    int i, j;
    int height, width;
    XFontStruct *fontInfo;

    if (!st_lookup(xd->font_table, font, (char **) &fontInfo)) {
        if (!(fontInfo = XLoadQueryFont(xd->display, font))) {
            (void) fprintf(stderr, "Error: no such font '%s'\n", font);
            return;
        }
        (void) st_insert(xd->font_table, util_strsav(font), (char *) fontInfo);
    }

    if (numlines) {
	height = (fontInfo->max_bounds.ascent + fontInfo->max_bounds.descent) * numlines;
        width = 0;
        for (i = 0; i < numlines; i++) {
            j = XTextWidth(fontInfo, lines[i], strlen(lines[i]));
            if (j > width)
                width = j;
        }
    } else {
	height = fontInfo->max_bounds.ascent + fontInfo->max_bounds.descent;
        width = fontInfo->max_bounds.rbearing - fontInfo->min_bounds.lbearing;
    }

    *widthp = width + X_PAD;
    *heightp = height + Y_PAD;

    return;
}

static void
draw_node(win, xpos, ypos, shape, color, font, numlines, lines, width, height)
    dg_plot *win;
    int xpos;
    int ypos;
    dg_nodeshape shape;
    char *color;
    char *font;
    int numlines;
    char **lines;
    int width, height;
{
    struct xdata *xd = (struct xdata *) win;
    int cx, cy, i, j;
    int realheight;
    XFontStruct *fontInfo;
    XColor *xColor;
    XColor dummy;

    cx = x_trans(xd, xpos);
    cy = y_trans(xd, ypos);

    if (!st_lookup(xd->font_table, font, (char **) &fontInfo)) {
        if (!(fontInfo = XLoadQueryFont(xd->display, font))) {
            (void) fprintf(stderr, "Error: no such font '%s'\n", font);
            return;
        }
        (void) st_insert(xd->font_table, util_strsav(font), (char *) fontInfo);
    }

    if (!st_lookup(xd->color_table, color, (char **) &xColor)) {
        xColor = ALLOC(XColor, 1);
        if (!XAllocNamedColor(xd->display, xd->color_map, color, xColor, &dummy)) {
            (void) fprintf(stderr, "Error: no such color '%s'\n", color);
            return;
        }
        (void) st_insert(xd->color_table, util_strsav(color), (char *) xColor);
    }

    realheight = height;

    XSetForeground(xd->display, xd->graphics_context, xColor->pixel);
    
    switch (shape) {
        case DG_SQUARE:
        if (width > height)
            height = width;
        else
            width = height;
        /* FALLTHRU */

        case DG_RECTANGLE:
	XClearArea(xd->display, xd->win,
		   cx - width / 2, cy - height / 2, width, height, False);
        XDrawRectangle(xd->display, xd->win, xd->graphics_context,
		       cx - width / 2 - 1, cy - height / 2 - 1,
		       width + 2, height + 2);
        break;

        case DG_CIRCLE:
        if (width > height)
            height = width;
        else
            width = height;
        /* FALLTHRU */

        case DG_ELLIPSE:
	/* clear area */
        XDrawArc(xd->display, xd->win, xd->graphics_context,
	         cx - width / 2 - 2, cy - height / 2 - 2,
		 width + 4, height + 4, 0, 360 * 64);

        break;

        case DG_TRIANGLE:
	/* clear area */
        XTriangle(xd->display, xd->win, xd->graphics_context, cx, cy, width + 2, height + 2);
        break;

        case DG_DIAMOND:
	/* clear area */
        XDiamond(xd->display, xd->win, xd->graphics_context, cx, cy, width + 2, height + 2);
        break;

        case DG_NO_SHAPE:
        break;
    
        default:
        assert(0);
    }

    XSetForeground(xd->display, xd->graphics_context, 
		   BlackPixel(xd->display, DefaultScreen(xd->display)));

    if (xd->current_font != xd->label_font) {
	XSetFont(xd->display, xd->graphics_context, xd->label_font->fid);
	xd->current_font = xd->label_font;
    }
    if (numlines) {
        for (i = 0; i < numlines; i++) {
            j = XTextWidth(fontInfo, lines[i], strlen(lines[i]));
            XDrawImageString(xd->display, xd->win, xd->graphics_context,
			     cx - j / 2,
			     cy + i * numlines + Y_PAD / 2 + (height - realheight) / 2,
			     lines[i], strlen(lines[i]));
        }
    }

    return;
}

static struct pattern {
    dg_linestyle style;
    int x_style;
} patterns[] = {
    { DG_BAD_STYLE, 0 },
    { DG_SOLID, LineSolid },
    { DG_DASHED, LineOnOffDash },
    { DG_DBLDASHED, LineDoubleDash }
} ;

static void
draw_arc(win, fx, fy, tx, ty, color, style, label, arr, save, fromshape,
        fromwidth, fromheight, toshape, towidth, toheight)
    dg_plot *win;
    int fx, fy;
    int tx, ty;
    char *color;
    dg_linestyle style;
    char *label;
    int arr;
    int save;
    dg_nodeshape fromshape;
    int fromwidth, fromheight;
    dg_nodeshape toshape;
    int towidth, toheight;
{
    struct xdata *xd = (struct xdata *) win;
    XPoint verts[2];
    /*
    XPoint *newverts;
    int numnewverts;
    */
    static XPoint *saved = NULL;
    static int numsaved = 0;
    XColor *xColor;
    XColor dummy;
    int i, cx, cy;

    if (style == DG_NO_LINE)
        return;

    if (!st_lookup(xd->color_table, color, (char **) &xColor)) {
        xColor = ALLOC(XColor, 1);
        if (!XAllocNamedColor(xd->display, xd->color_map, color, xColor, &dummy)) {
            (void) fprintf(stderr, "Error: no such color '%s'\n", color);
            return;
        }
        (void) st_insert(xd->color_table, util_strsav(color), (char *) xColor);
    }

    if (xd->current_style != style) {
	for (i = 0; i < sizeof (patterns) / sizeof (struct pattern); i++) {
	    if (patterns[i].style == style) break;
	}
	XSetLineAttributes(xd->display, xd->graphics_context,
		      1, patterns[i].x_style, CapButt, JoinMiter);
	xd->current_style = style;
    }

    verts[0].x = x_trans(xd, fx);
    verts[0].y = y_trans(xd, fy);
    verts[1].x = x_trans(xd, tx);
    verts[1].y = y_trans(xd, ty);

    verts[0].x += shape_width(fromshape, fromwidth, fromheight) / 2;
    verts[1].x -= shape_width(toshape, towidth, toheight) / 2;

    if (save || numsaved) {
        if (numsaved)
            saved = REALLOC(XPoint, saved, numsaved + 2);
        else {
            saved = ALLOC(XPoint, 3);
            saved[numsaved++] = verts[0];
        }
        saved[numsaved].x = (verts[0].x + verts[1].x) / 2;
        saved[numsaved].y = (verts[0].y + verts[1].y) / 2;
        numsaved++;
        saved[numsaved++] = verts[1];

        if (numsaved == 2)
            goto print_label;

        if (save)
            return;
    }

    XSetForeground(xd->display, xd->graphics_context, xColor->pixel);

    if (numsaved) {
        saved = REALLOC(XPoint, saved, numsaved + 1);
        saved[numsaved] = saved[numsaved - 1];
        numsaved++;
	/*
        numnewverts = dg_curve_verts(saved, numsaved, &newverts);
	*/
	/*
        XDrawLines(xd->display, xd->win, xd->graphics_context, newverts, numnewverts, CoordModeOrigin);
	*/
        XDrawLines(xd->display, xd->win, xd->graphics_context, saved, numsaved, CoordModeOrigin);
        numsaved = 0;
        FREE(saved);
        /* FREE(newverts); */
    } else {
        XDrawLines(xd->display, xd->win, xd->graphics_context, verts, 2, CoordModeOrigin);
    }

    if (arr) {
        arrow(xd->display, xd->win, xd->graphics_context,
	      verts[0].x, verts[0].y, verts[1].x, verts[1].y,
              ARROW_ANGLE, ARROW_LENGTH);
    }
    
print_label:
    if (label) {
        cx = (x_trans(xd, fx) * 4 + x_trans(xd, tx)) / 5;
        cy = (y_trans(xd, fy) * 4 + y_trans(xd, ty)) / 5;

	if (xd->current_font != xd->label_font) {
	    XSetFont(xd->display, xd->graphics_context, xd->label_font->fid);
	    xd->current_font = xd->label_font;
	}
	XSetForeground(xd->display, xd->graphics_context, 
		       BlackPixel(xd->display, DefaultScreen(xd->display)));
	XDrawImageString(xd->display, xd->win, xd->graphics_context,
			 cx, cy, label, strlen(label));
    }

    return;
}

/* ARGSUSED */
void
arrow(display, win, gc, fromx, fromy, tox, toy, arrangle, arrlength)
    Display *display;
    Window win;
    GC gc;
    int fromx, fromy;
    int tox, toy;
    double arrangle;
    int arrlength;
{
    XPoint verts[3];
    double angle;
    
    angle = atan2((double) fromy - toy, (double) fromx - tox);

    verts[0].x = arrlength * cos(angle - arrangle) + tox;
    verts[0].y = arrlength * sin(angle - arrangle) + toy;
    verts[1].x = tox;
    verts[1].y = toy;
    verts[2].x = arrlength * cos(angle + arrangle) + tox;
    verts[2].y = arrlength * sin(angle + arrangle) + toy;
    XDrawLines(display, win, gc, verts, 3, CoordModeOrigin);

    return;
}

/* ARGSUSED */
static void
update_win(win)
    dg_plot *win;
{
    struct xdata *xd = (struct xdata *) win;
    XFlush(xd->display);
    return;
}

/* ARGSUSED */
static enum st_retval
color_free(k, v, a)
    char *k, *v, *a;
{
    FREE(k);
    FREE(v);
    return (ST_CONTINUE);
}

/* ARGSUSED */
static enum st_retval
font_free(k, v, a)
    char *k, *v, *a;
{
    struct xdata *xd = (struct xdata *) a;
    FREE(k);
    XFreeFont(xd->display, (XFontStruct *) v);
    return (ST_CONTINUE);
}

static void
close_win(win)
    dg_plot *win;
{
    struct xdata *xd = (struct xdata *) win;

    XDestroyWindow(xd->display, xd->win);
    XFlush(xd->display);

    (void) st_foreach(xd->color_table, color_free, (char *) xd);
    (void) st_foreach(xd->font_table, font_free, (char *) xd);

    st_free_table(xd->color_table);
    st_free_table(xd->font_table);

    FREE(xd);

    if (--refcount == 0) {
        XCloseDisplay(display_opened);
        display_opened = NULL;
    }

    return;
}

static int
shape_width(shape, width, height)
    dg_nodeshape shape;
    int width, height;
{
    switch (shape) {
        case DG_NO_SHAPE:
        case DG_SQUARE:
        case DG_RECTANGLE:
        return (width);

        case DG_ELLIPSE:
        case DG_CIRCLE:
        return (width + (SQRT_2 - 1) * width / height / 2);

        case DG_TRIANGLE:
        return (width + height * 2);

        case DG_DIAMOND:
        return (width + height);
    }

    return (0);
}

dg_outdev dg_xwindow_dev = {
    open_win,
    calc_sizes,
    draw_node,
    draw_arc,
    update_win,
    close_win
} ;

#endif /* NO_X */

